#
#  Copyright (C) 2008-2014 National Institute For Space Research (INPE) - Brazil.
#
#  This file is part of the TerraLib - a Framework for building GIS enabled applications.
#
#  TerraLib is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License,
#  or (at your option) any later version.
#
#  TerraLib is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with TerraLib. See COPYING. If not, write to
#  TerraLib Team at <terralib-team@terralib.org>.
#
#
#  Description: Auxiliary macros and functions.
#
#  Author: Gilberto Ribeiro de Queiroz <gribeiro@dpi.inpe.br>
#          Juan Carlos P. Garrido <juan@dpi.inpe.br>
#          Frederico Augusto T. Bede <frederico.bede@funcate.org.br>
#

MACRO(TeInstallPlugins plugins location)

  set(_files)

  foreach(plugin ${plugins})
    get_target_property(_loc ${plugin} LOCATION)
    if(EXISTS ${_loc})
        list(APPEND _files ${_loc})
    endif()
  endforeach()

  install(FILES ${_files}
           DESTINATION "${TERRALIB_BASE_DESTINATION_DIR}qtplugins/${location}"
           COMPONENT runtime)

ENDMACRO(TeInstallPlugins)


MACRO(TeInstallQt5Plugins)
  find_package(Qt5 COMPONENTS Sql Svg PrintSupport Widgets)
  
# Installing image plugins
  set(_plugins Qt5::QGifPlugin Qt5::QICOPlugin Qt5::QJpegPlugin Qt5::QTiffPlugin Qt5::QSvgPlugin)
  TeInstallPlugins("${_plugins}" "imageformats")

# Installing svg plugins
  set(_plugins Qt5::QSvgIconPlugin)
  TeInstallPlugins("${_plugins}" "iconengines")

# Installing sql plugins
  set(_plugins Qt5::QSQLiteDriverPlugin)
  TeInstallPlugins("${_plugins}" "sqldrivers")

# Installing printer support plugin
  if(WIN32)
    set(_plugins Qt5::QWindowsPrinterSupportPlugin)
    TeInstallPlugins("${_plugins}" "printsupport")
  endif()

# Installing platform plugins
  if(WIN32)
    set(_plugins Qt5::QWindowsIntegrationPlugin Qt5::QMinimalIntegrationPlugin)
    TeInstallPlugins("${_plugins}" "platforms")
  elseif(APPLE)
    set(_plugins Qt5::QCocoaIntegrationPlugin Qt5::QMinimalIntegrationPlugin)
    TeInstallPlugins("${_plugins}" "platforms")
  elseif(UNIX)
  	set(_plugins Qt5::QXcbIntegrationPlugin Qt5::QMinimalIntegrationPlugin)
    TeInstallPlugins("${_plugins}" "platforms")
  endif()
  
 # Installing style plugins
  if(WIN32)
    set(_plugins Qt5::QWindowsVistaStylePlugin)
    TeInstallPlugins("${_plugins}" "styles")
  endif()
  
ENDMACRO(TeInstallQt5Plugins)

MACRO(TeInstallQt5Fonts)
  if(${CMAKE_SYSTEM_NAME} MATCHES "Linux")
    set (QT_FONT_DIR "${Qt5_DIR}/../../../lib/fonts")
    if(EXISTS ${QT_FONT_DIR})
      install(DIRECTORY ${QT_FONT_DIR}
        DESTINATION "${TERRALIB_BASE_DESTINATION_DIR}lib/"
        COMPONENT runtime)
    endif()
  endif()
ENDMACRO(TeInstallQt5Fonts)

#
# Macro installQtPlugins
#
# Description: Installs the required Qt plugins.
#
# param plugs List of the names of plugins to be installed.
#
MACRO(TeInstallQtPlugins plgs)

  if ("${QT_PLUGINS_DIR}" STREQUAL "")
    set (QT_PLUGINS_DIR "${Qt5_DIR}/../../../plugins")
  endif()

  set (_regex_exp "")

  set( _first TRUE)

  foreach(plg ${plgs})
    if(NOT _first)
      set (_regex_exp ${_regex_exp}|${plg})
    else()
      set (_regex_exp ${plg})
      set (_first FALSE)
    endif()
  endforeach()

  set (_regex_exp "(${_regex_exp})?(${CMAKE_SHARED_LIBRARY_SUFFIX})$")

  set (_dest "${TERRALIB_BASE_DESTINATION_DIR}qtplugins")

if(QT4_FOUND)
  set (_plugin_dirs "imageformats;iconengines;sqldrivers")
else()
    set (_plugin_dirs "imageformats;iconengines;sqldrivers;platforms")
endif()

  foreach(_plugin_dir ${_plugin_dirs})

    install (DIRECTORY ${QT_PLUGINS_DIR}/${_plugin_dir}
             DESTINATION ${_dest}  COMPONENT runtime
             FILES_MATCHING REGEX "${_regex_exp}")
  endforeach()

ENDMACRO(TeInstallQtPlugins)

#
# Macro TeInstallTerraLibPlugins
#
# Description: Installs the required TerraLib plugins. (Used by outter projects.)
#
# param plugins List of the names of plugins to be installed.
#
# param location Path where the plugins must be installed.
#
MACRO(TeInstallTerraLibPlugins plugins location)


  set(_files "")

  foreach(plugin ${plugins})

    get_target_property(_loc ${plugin} LOCATION_RELEASE)

    list(APPEND _files "${_loc}")

    get_filename_component(_lib_dir ${_loc} DIRECTORY)
    
    set(_sym_loc "${_lib_dir}/${CMAKE_SHARED_LIBRARY_PREFIX}${plugin}${CMAKE_SHARED_LIBRARY_SUFFIX}")

    list(APPEND _files "${_sym_loc}")
    
  endforeach()

  install (FILES ${_files}
    DESTINATION ${location}
    CONFIGURATIONS Release
    COMPONENT runtime
  )

  set(_filesd "")

  foreach(plugin ${plugins})

    get_target_property(_loc ${plugin} LOCATION_DEBUG)

    list(APPEND _filesd "${_loc}")

    get_filename_component(_lib_dir ${_loc} DIRECTORY)
    
    set(_sym_loc "${_lib_dir}/${CMAKE_SHARED_LIBRARY_PREFIX}${plugin}${CMAKE_SHARED_LIBRARY_SUFFIX}")

    list(APPEND _filesd "${_sym_loc}")
    
  endforeach()

  install (FILES ${_filesd}
    DESTINATION ${location}
    CONFIGURATIONS Debug
    COMPONENT runtime
  )

  if(UNIX OR APPLE)
    install (FILES ${_files}
      DESTINATION ${location}
      COMPONENT runtime
    )
  endif()

  
ENDMACRO(TeInstallTerraLibPlugins)

MACRO(GetTerraLibModules _modules)
  set(allModules "${TERRALIB_MODULES}" "${TERRALIB_PLUGINS}")

  foreach(module ${allModules})
    set (_mod_name terralib_mod_${module})

#    string(FIND ${_mod_name} binding_python _py_ok)

#    if(${_py_ok} GREATER 0)
#      set(_mod_name _${_mod_name})
#    endif()

    list(APPEND ${_modules} ${_mod_name})
  endforeach()
ENDMACRO(GetTerraLibModules)

#
# Macro TeInstallTerraPrint
#
# Description: Installs the required TerraPrint plugins.
#
# param plugins List of the names of plugins to be installed.
#
# param location Path where the plugins must be installed.
#

MACRO(TeLayoutInstallPlugins location)

  install(FILES ${TERRALIB_LAYOUT_LIBRARIES}
           DESTINATION ${location}
           COMPONENT runtime
           )

  if(APPLE)
    install(FILES ${terralib_layout_DIR}/../../${CMAKE_SHARED_LIBRARY_PREFIX}terralib_mod_layout${CMAKE_SHARED_LIBRARY_SUFFIX}
                  ${terralib_layout_DIR}/../../${CMAKE_SHARED_LIBRARY_PREFIX}terralib_mod_qt_plugins_layout${CMAKE_SHARED_LIBRARY_SUFFIX}
            DESTINATION ${location}
            COMPONENT runtime)
  endif()

  if(EXISTS ${terralib_layout_DIR}/CMakeCache.txt)
    install(
      DIRECTORY ${terralib_layout_INCLUDE_DIRS}/../share/
      DESTINATION ${location}/../share
      COMPONENT runtime
    )
  else()
    install(
      DIRECTORY ${terralib_layout_DIR}/../../../share/
      DESTINATION ${location}/../share
      COMPONENT runtime
    )
  endif()
ENDMACRO(TeLayoutInstallPlugins)

#
# Macro: TERRALIB_ADD_PYTHON_BINDING
#
# Description: Register a python binding for a terralib module
#
# param _module terralib module to be used in the binding
#
# param _source SWIG source file

MACRO(TERRALIB_ADD_PYTHON_BINDING _module _source)
file(GLOB TERRALIB_BINDINGS_PYTHON_SRC_FILES ${_source})

source_group("Source Files" FILES ${TERRALIB_BINDINGS_PYTHON_SRC_FILES})

string(REPLACE "terralib_mod_" "" py_module ${_module})

if(PYTHONLIBS_VERSION_STRING VERSION_GREATER "3")
set_source_files_properties(${TERRALIB_BINDINGS_PYTHON_SRC_FILES}
                            PROPERTIES CPLUSPLUS ON CMAKE_SWIG_FLAGS "-includeall")
else()
set_source_files_properties(${TERRALIB_BINDINGS_PYTHON_SRC_FILES}
                            PROPERTIES CPLUSPLUS ON CMAKE_SWIG_FLAGS "-includeall")
endif()

set(TERRALIB_LIBRARIES_DEPENDENCIES ${PYTHON_LIBRARIES})

if(TERRALIB_DEFINE_RPATH)
  ADD_EXTERNAL_LIBRARY_TO_RPATH(${py_module} ${TERRALIB_LIBRARIES_DEPENDENCIES})
endif()

if(${CMAKE_VERSION} VERSION_GREATER "3.10")
  SWIG_ADD_LIBRARY(
    ${py_module}
    TYPE SHARED
    LANGUAGE python
    SOURCES ${TERRALIB_BINDINGS_PYTHON_SRC_FILES}
  )
else()
  SWIG_ADD_MODULE(${py_module} python ${TERRALIB_BINDINGS_PYTHON_SRC_FILES})
endif()

SWIG_LINK_LIBRARIES(${py_module} "${_module}" ${PYTHON_LIBRARIES})

set_target_properties("_${py_module}"
                      PROPERTIES
                      LIBRARY_OUTPUT_DIRECTORY ${CMAKE_SWIG_OUTDIR}
                      LIBRARY_OUTPUT_DIRECTORY_RELEASE ${CMAKE_SWIG_OUTDIR}
                      LIBRARY_OUTPUT_DIRECTORY_DEBUG ${CMAKE_SWIG_OUTDIR}
                      DEBUG_POSTFIX "")

if(APPLE)
  set_target_properties("_${py_module}"
                        PROPERTIES
                        MACOSX_RPATH TRUE
                        INSTALL_RPATH "@executable_path/../lib;${CMAKE_INSTALL_PREFIX}/lib;${TERRALIB_DESTINATION_RUTINME}"
                        INSTALL_RPATH_USE_LINK_PATH TRUE
                        BUILD_WITH_INSTALL_RPATH FALSE)
endif()


if(TERRALIB_PYTHON_BINDINGS)
  set(TERRALIB_PYTHON_BINDINGS "${TERRALIB_PYTHON_BINDINGS}, ${py_module}")
else()
  set(TERRALIB_PYTHON_BINDINGS "${py_module}")
endif()

if(NOT TERRALIB_MOD_BINDING_PYTHON_INSTALL_DIR)
 set(TERRALIB_MOD_BINDING_PYTHON_INSTALL_DIR "${TERRALIB_BASE_DESTINATION_DIR}${TERRALIB_MOD_BINDING_PYTHON_DIR}")
endif()

install(TARGETS "_${py_module}" DESTINATION ${TERRALIB_MOD_BINDING_PYTHON_INSTALL_DIR})
install(FILES "${CMAKE_SWIG_OUTDIR}/${py_module}.py" DESTINATION ${TERRALIB_MOD_BINDING_PYTHON_INSTALL_DIR})
ENDMACRO(TERRALIB_ADD_PYTHON_BINDING)

#
# Macro: TERRALIB_ADD_C_WRAPPER
#
# Description: Register a C wrapper for a terralib module
#
# param _module terralib module to be used in the wrapper
#
# param _dir C source dir
MACRO(TERRALIB_ADD_C_BINDING _module _dir)

file(GLOB HDR_FILES ${_dir}/*.h)
file(GLOB SRC_FILES ${_dir}/*.cpp)

string(REPLACE "terralib_mod_" "" folder ${_module})

source_group("Header Files\\${folder}" FILES ${HDR_FILES})
source_group("Source Files\\${folder}" FILES ${SRC_FILES})

set(c_module "${_module}_c")


add_library(${c_module} SHARED ${HDR_FILES}
                               ${SRC_FILES})

set_target_properties(${c_module}
                     PROPERTIES VERSION ${TERRALIB_VERSION_MAJOR}.${TERRALIB_VERSION_MINOR}
                                SOVERSION ${TERRALIB_VERSION_MAJOR}.${TERRALIB_VERSION_MINOR})
target_link_libraries(${c_module} ${_module})


ENDMACRO(TERRALIB_ADD_C_BINDING)

#
# Macro: TERRALIB_REGISTER_TRANSLATION
#
# Description: Register the ".mo" file for a given ".po" file
#
# param proj_name The module to be translated. This will be your text domain.
#
# param locale The locale of the ".po" file
#
# param po_dir Where the ".po" file is located. The ".po" file named must be: "proj_name"_"locale".po
#
# param mo_dir Output directory for the ".mo" file
#

MACRO(TERRALIB_REGISTER_TRANSLATION proj_name locale po_dir mo_dir)
  if(XGETTEXT_FOUND)
    if(TERRALIB_TRANSLATOR_ENABLED)
      set(_po_file "${po_dir}/${proj_name}_${locale}.po")
      set(_mo_file "${mo_dir}/${locale}/LC_MESSAGES/${proj_name}.mo")
      if(EXISTS ${po_dir}/${proj_name}_${locale}.po)
        add_custom_command(
          TARGET ${proj_name}
          POST_BUILD
          COMMAND ${CMAKE_COMMAND} -E make_directory ${mo_dir}/${locale}/LC_MESSAGES/
          COMMAND ${GETTEXT_MSGFMT_EXECUTABLE} -o ${_mo_file} ${_po_file}
          COMMENT "Generating translation binary file for ${proj_name}...")
      endif()
    endif(TERRALIB_TRANSLATOR_ENABLED)
  endif(XGETTEXT_FOUND)
ENDMACRO(TERRALIB_REGISTER_TRANSLATION)

MACRO(TERRALIB_CREATE_TRANSLATION proj_name keyword_s keyword_p locale directory po_dir)
  if(XGETTEXT_FOUND)
    if(TERRALIB_TRANSLATOR_ENABLED)
      get_filename_component(_absDir ${directory} ABSOLUTE)
      set(_pot_file ${CMAKE_CURRENT_BINARY_DIR}/${proj_name}.pot)
      set(_po_file "${po_dir}/${proj_name}_${locale}.po")
      set(_list_of_cpp_files "${CMAKE_CURRENT_BINARY_DIR}/${proj_name}_cpp_files.txt")
      file(GLOB_RECURSE _srcs ${_absDir} ${_absDir}/*.cpp)
	  file(WRITE ${_list_of_cpp_files} "")
	  set(_absolutePathToSRC "${TERRALIB_ABSOLUTE_ROOT_DIR}/src/")
	  FOREACH(cppfile ${_srcs})
		file(RELATIVE_PATH relativeCppFile ${_absolutePathToSRC} "${cppfile}")
	  	file(APPEND ${_list_of_cpp_files} "${relativeCppFile}\n")
	  ENDFOREACH()      
      set(encoding "UTF-8")
      if(NOT EXISTS "${po_dir}/${proj_name}_${locale}.po")
        configure_file(${po_dir}/template_translation.po.in
                       ${po_dir}/${proj_name}_${locale}.po @ONLY)
      endif()
      add_custom_command(
        TARGET ${proj_name}
        PRE_BUILD
        COMMAND ${XGETTEXT_EXECUTABLE} --directory=${_absolutePathToSRC}  --from-code=${encoding} --no-wrap --c++ --keyword=${keyword_s} --keyword=${keyword_p}:1,2 -o ${_pot_file} --files-from=${_list_of_cpp_files}
        COMMAND ${GETTEXT_MSGMERGE_EXECUTABLE} -U --backup=none -q --lang=${locale} ${_po_file} ${_pot_file}
        COMMENT "Generating translation source file for ${proj_name}...")
    endif(TERRALIB_TRANSLATOR_ENABLED)
  endif(XGETTEXT_FOUND)
ENDMACRO(TERRALIB_CREATE_TRANSLATION)


#
# Macro: TERRALIB_SETUP_PYTHON_DEPS
#
# Description: Prepare python dependencies on Windows to installation
#
MACRO(TERRALIB_SETUP_PYTHON_DEPS)

find_package(PythonInterp REQUIRED)
set(PYTHON_ZIP "${CMAKE_BINARY_DIR}/python${PYTHON_VERSION_MAJOR}${PYTHON_VERSION_MINOR}.zip")

get_filename_component(PYTHON_DIR "${PYTHON_INCLUDE_DIR}/.." ABSOLUTE)
set(PYTHON_DLL "${PYTHON_DIR}/python${PYTHON_VERSION_MAJOR}${PYTHON_VERSION_MINOR}.dll")

if(NOT EXISTS ${PYTHON_ZIP})

  set(PYTHON_COMPILEALL -O -m compileall)
  if(${PYTHON_VERSION_STRING} VERSION_GREATER "3.2")
    set(PYTHON_COMPILEALL ${PYTHON_COMPILEALL} -b)
  endif()

  MESSAGE(STATUS "Compiling python sources...")
  execute_process(COMMAND ${PYTHON_EXECUTABLE} ${PYTHON_COMPILEALL} "Lib/"
                  WORKING_DIRECTORY "${PYTHON_DIR}"
                  OUTPUT_QUIET)						
  MESSAGE(STATUS "Finished compiling python sources...")

  set(PYTHON_LIST ${CMAKE_BINARY_DIR}/python_list.txt)
  FILE(WRITE ${PYTHON_LIST} "")

  get_filename_component(PYTHON_LIB_DIR "${PYTHON_INCLUDE_DIR}/../Lib" ABSOLUTE)	
  FILE(GLOB_RECURSE PYTHON_FILES ABSOLUTE "${PYTHON_LIB_DIR}/*")

  FOREACH(item ${PYTHON_FILES})
    IF(NOT(${item} MATCHES "__pycache__") AND NOT(${item} MATCHES "(.*).py$")
      AND NOT(${item} MATCHES "(.*)[^doc|unit]test(.*)")
      OR (${item} MATCHES "asyncio/test_utils") AND NOT(${item} MATCHES "venv")
      AND NOT(${item} MATCHES "pip") OR (${item} MATCHES "pipes")
      AND NOT(${item} MATCHES "tkinter") AND NOT(${item} MATCHES "turtledemo") 
      AND NOT(${item} MATCHES "pydoc_data") AND NOT(${item} MATCHES "site-packages") 
      AND NOT(${item} MATCHES "idlelib") AND NOT(${item} MATCHES "(.*).exe$")
      )
      set(_item)
      STRING(REPLACE "${PYTHON_LIB_DIR}/" "" _item ${item})
      FILE(APPEND ${PYTHON_LIST} "${_item}\n")
    ENDIF()
  ENDFOREACH()
    

  
  MESSAGE(STATUS "Compacting python byte-code...")
  execute_process(COMMAND ${CMAKE_COMMAND} -E tar c ${PYTHON_ZIP} --files-from=${PYTHON_LIST} --format=zip 
                  WORKING_DIRECTORY ${PYTHON_LIB_DIR}
                  OUTPUT_QUIET)
  MESSAGE(STATUS "Finished compacting python byte-code...")

endif()
install(FILES ${PYTHON_ZIP}
        DESTINATION ${TERRALIB_DESTINATION_RUNTIME}
        COMPONENT runtime)
install(FILES ${PYTHON_DLL}
        DESTINATION ${TERRALIB_DESTINATION_RUNTIME}
        COMPONENT runtime)
ENDMACRO(TERRALIB_SETUP_PYTHON_DEPS)


#
# Macro: TERRALIB_INSTALL_BINDINGS
#
# Description: Install generated bindings from a given directory to a destinatination
#
# param bindings_dir Where the bindings are located
#
# param destination Where the bindings will be installed
#
MACRO(TERRALIB_INSTALL_BINDINGS bindings_dir destination)
  INSTALL(DIRECTORY ${bindings_dir}
                   DESTINATION ${destination}
                   COMPONENT runtime)
ENDMACRO(TERRALIB_INSTALL_BINDINGS)
