#include "RasterExamples.h"

// TerraLib
#include <terralib/dataaccess.h>
#include <terralib/geometry.h>
#include <terralib/raster.h>

// STL
#include <iostream>
#include <string>

/*
  This is an example of how to access a raster in the postgis database using the GDAL library.
  The GDAL PostGIS raster driver is read-only support.

  For testing, a postgis specific tool was used to import a raster into the database: raster2pgsql

  The command line to import is:

    raster2pgsql raster_options_go_here raster_file someschema.sometable > out.sql

    e.g.: raster2pgsql -s 4236 -I -C -M *.tif -F -t 100x100 public.demelevation > elev.sql

  As a result an output file is generated with the script to be executed in the database. 

  The command line to execute the script is:

    psql -d db_name -f out.sql

    e.g.: psql -d gisdb -f elev.sql

Ref: https://trac.osgeo.org/gdal/wiki/frmts_wtkraster.html
     http://postgis.refractions.net/documentation/manual-svn/using_raster.xml.html#RT_Raster_Loader

  NOTE: The GDAL library needs to be compiled with postgis driver support.
*/

#include <terralib/dataaccess/dataset/DataSet.h>

void PostGISRaster_RasterAPI();
void PostGISRaster_DataSourceAPI();

void PostGISRaster()
{
  PostGISRaster_RasterAPI();

  PostGISRaster_DataSourceAPI();
}

void PostGISRaster_RasterAPI()
{
  try
  {
    std::cout << "This example shows how to use the GDAL PostGIS Raster support using Raster API." << std::endl << std::endl;

    // 1 - define raster info and load
    std::map<std::string, std::string> rinfo;
    rinfo["URI"] = "PG:host=localhost port=5432 dbname='raster_db' user='postgres' password='****' schema='public' table=cbers mode=2";

    te::rst::RasterPtr inraster(te::rst::RasterFactory::open(rinfo));

    // 2 - Copy image to file format
    std::string uriOut = TERRALIB_DATA_DIR "/geotiff/raster_pgis_rasterAPI.tif";

    te::rst::RasterPtr outraster = te::rst::CreateCopy(*(inraster.get()), uriOut);

    if (outraster)
    {
      std::cout << "End. Check the PostGIS Raster copy to file: raster_pgis_rasterAPI.tif." << std::endl << std::endl;
    }
    else
    {
      std::cout << "PostGIS Raster copy failed!" << std::endl << std::endl;
    }

    std::cout << "Done!" << std::endl << std::endl;
  }
  catch (const std::exception& e)
  {
    std::cout << std::endl << "An exception has occurred in PostGISRaster(): " << e.what() << std::endl;
  }
  catch (...)
  {
    std::cout << std::endl << "An unexpected exception has occurred in PostGISRaster()!" << std::endl;
  }
}

void PostGISRaster_DataSourceAPI()
{
  try
  {
    std::cout << "This example shows how to use the GDAL PostGIS Raster support using DataSource API." << std::endl << std::endl;

    // 1 - define raster info and load
    std::map<std::string, std::string> rinfo;

    std::string uri = "PG://postgres:*****@localhost:5432/raster_db?schema=public&table=cbers&mode=2";

    rinfo["URI"] = uri;

    std::unique_ptr<te::da::DataSource> ds = te::da::DataSourceFactory::make("GDAL", uri);
    ds->open();
    ds->isValid();

    std::vector<std::string> dsNames = ds->getDataSetNames();

    std::unique_ptr<te::da::DataSet> dataSet = ds->getDataSet(dsNames[0]);

    std::size_t rpos = te::da::GetFirstPropertyPos(dataSet.get(), te::dt::RASTER_TYPE);

    std::unique_ptr<te::rst::Raster> inraster = dataSet->getRaster(rpos);


    // 2 - Copy image to file format
    std::string uriOut = TERRALIB_DATA_DIR "/geotiff/raster_pgis_dataSourceAPI.tif";

    te::rst::RasterPtr outraster = te::rst::CreateCopy(*(inraster.get()), uriOut);

    if (outraster)
    {
      std::cout << "End. Check the PostGIS Raster copy to file: raster_pgis_dataSourceAPI.tif." << std::endl << std::endl;
    }
    else
    {
      std::cout << "PostGIS Raster copy failed!" << std::endl << std::endl;
    }

    std::cout << "Done!" << std::endl << std::endl;
  }
  catch (const std::exception& e)
  {
    std::cout << std::endl << "An exception has occurred in PostGISRaster(): " << e.what() << std::endl;
  }
  catch (...)
  {
    std::cout << std::endl << "An unexpected exception has occurred in PostGISRaster()!" << std::endl;
  }
}
