#include "../Config.h"
#include <terralib/common.h>
#include <terralib/dataaccess.h>
#include <terralib/dataaccess/datasource/DataSourceFactory.h>
#include <terralib/dataaccess/datasource/DataSourceInfo.h>
#include <terralib/dataaccess/datasource/DataSourceInfoManager.h>
#include <terralib/dataaccess/query/Field.h>
#include <terralib/dataaccess/query/Fields.h>
#include <terralib/dataaccess/query/From.h>
#include <terralib/dataaccess/query/FromItem.h>
#include <terralib/dataaccess/query/Select.h>
#include <terralib/datatype/SimpleData.h>
#include <terralib/geometry/GeometryProperty.h>
#include <terralib/vp/ComplexData.h>
#include <terralib/vp/Dissolve.h>
#include <terralib/vp/Utils.h>

// Boost
#include <boost/uuid/random_generator.hpp>
#include <boost/uuid/uuid_io.hpp>

// STL
#include <iostream>
#include <map>
#include <memory>
#include <string>
#include <vector>

//OGR to OGR
bool DissolveOGRToOGR()
{
  std::string data_dir = TERRALIB_DATA_DIR;

  std::string filename(data_dir + "/shape/SP_cities.shp");
  
  std::string srcInfo ("file://" + filename);
  
  te::da::DataSourcePtr srcDs(te::da::DataSourceFactory::make("OGR", srcInfo));
  srcDs->open();
  
  std::string inDsetName = "SP_cities";
  if (!srcDs->dataSetExists(inDsetName))
  {
    std::cout << "Input dataset not found: " << inDsetName << std::endl;
    return false;
  }

  te::da::DataSet* inDset = srcDs->getDataSet(inDsetName).release();
  te::da::DataSetType* inDsetType = srcDs->getDataSetType(inDsetName).release();

  // Set specific parameters.
  std::vector<std::string> selProperties;
  selProperties.push_back("NOMEMESO");

  std::map<std::string, te::dt::AbstractData*> specificParams;
  specificParams["DISSOLVE"] =
      new te::vp::ComplexData<std::vector<std::string> >(selProperties);

  specificParams["IS_COLLECTION"] =
      new te::dt::SimpleData<bool, te::dt::BOOLEAN_TYPE>(true);

  std::map<te::dt::Property*, std::vector<te::stat::StatisticalSummary> > stats;

  std::unique_ptr<te::dt::Property> prop1 =
      srcDs->getProperty(inDsetName, "POPULACA");

  std::vector<te::stat::StatisticalSummary> stat1;
  stat1.push_back(te::stat::SUM);
  stat1.push_back(te::stat::MIN_VALUE);
  stat1.push_back(te::stat::MODE);
  stats.insert(std::make_pair(prop1.get(), stat1));

  specificParams["SUMMARY"] = new te::vp::ComplexData<
      std::map<te::dt::Property*, std::vector<te::stat::StatisticalSummary> > >(
      stats);

  te::gm::GeometryProperty* geomInputProp =
      te::da::GetFirstGeomProperty(inDsetType);

  int inputSRID = geomInputProp->getSRID();

  std::string filename2(data_dir + "/shape/dissolve_ogrtoogr.shp");
  std::string tgrInfo ("file://" + filename2);

  te::da::DataSourcePtr trgDs(te::da::DataSourceFactory::make("OGR", tgrInfo));
  trgDs->open();
  
  std::string outDS = "dissolve_ogrtoogr";
  
  if (trgDs->dataSetExists(outDS))
  {
    std::cout << "A dataset with the same requested output dataset name "
                 "already exists: " << outDS << std::endl;
    return false;
  }

  // Can be added more than one input parameter.
  std::vector<te::vp::InputParams> inputParamsVec;

  // Declare the input parameters
  te::vp::InputParams structInputParams;

  // Set the inputLayer parameters
  structInputParams.m_inputDataSource = srcDs;
  structInputParams.m_inputDataSetType = inDsetType;
  structInputParams.m_inputDataSet = inDset;

  inputParamsVec.push_back(structInputParams);

  // Set parameters (Input/Output).
  te::vp::AlgorithmParams* params = new te::vp::AlgorithmParams();
  params->setInputParams(inputParamsVec);
  params->setOutputDataSource(trgDs);
  params->setOutputDataSetName(outDS);
  params->setOutputSRID(inputSRID);
  params->setSpecificParams(specificParams);

  te::vp::Dissolve dissolve;
  bool result = dissolve.executeMemory(params);

  delete params;
  
  return result;
}

//OGR to Postgis
bool DissolveOGRToPGIS()
{
  std::string data_dir = TERRALIB_DATA_DIR;
  
  std::string filename(data_dir + "/shape/SP_cities.shp");
  
  std::string srcInfo("file://" + filename);

  te::da::DataSourcePtr srcDs(te::da::DataSourceFactory::make("OGR", srcInfo));
  srcDs->open();
  
  std::string inDsetName = "SP_cities";
  if (!srcDs->dataSetExists(inDsetName))
  {
    std::cout << "Input dataset not found: " << inDsetName << std::endl;
    return false;
  }
  
  te::da::DataSet* inDset = srcDs->getDataSet(inDsetName).release();
  te::da::DataSetType* inDsetType = srcDs->getDataSetType(inDsetName).release();

  // Set specific parameters.
  std::vector<std::string> selProperties;
  selProperties.push_back("NOMEMESO");

  std::map<std::string, te::dt::AbstractData*> specificParams;
  specificParams["DISSOLVE"] =
      new te::vp::ComplexData<std::vector<std::string> >(selProperties);

  specificParams["IS_COLLECTION"] =
      new te::dt::SimpleData<bool, te::dt::BOOLEAN_TYPE>(true);
  
  std::map<te::dt::Property*, std::vector<te::stat::StatisticalSummary> > stats;

  std::unique_ptr<te::dt::Property> prop1 =
      srcDs->getProperty(inDsetName, "POPULACA");

  std::vector<te::stat::StatisticalSummary> stat1;
  stat1.push_back(te::stat::SUM);
  stat1.push_back(te::stat::MIN_VALUE);
  stat1.push_back(te::stat::MODE);
  stats.insert(std::make_pair(prop1.get(), stat1));

  specificParams["SUMMARY"] = new te::vp::ComplexData<
      std::map<te::dt::Property*, std::vector<te::stat::StatisticalSummary> > >(
      stats);

  te::gm::GeometryProperty* geomInputProp =
      te::da::GetFirstGeomProperty(inDsetType);

  int inputSRID = geomInputProp->getSRID();

  std::string connInfo(
      "ppgsql://postgres:postgres@atlas.dpi.inpe.br:5433/testPostGIS");

  te::da::DataSourcePtr trgDs(
      te::da::DataSourceFactory::make("POSTGIS", connInfo));
  trgDs->open();
  
  std::string outDS = "dissolve_ogrtopgis";
  
  if (trgDs->dataSetExists(outDS))
  {
    std::cout << "A dataset with the same requested output dataset name "
                 "already exists: " << outDS << std::endl;
    return false;
  }
  
  // Can be added more than one input parameter.
  std::vector<te::vp::InputParams> inputParamsVec;

  // Declare the input parameters
  te::vp::InputParams structInputParams;

  // Set the inputLayer parameters
  structInputParams.m_inputDataSource = srcDs;
  structInputParams.m_inputDataSetType = inDsetType;
  structInputParams.m_inputDataSet = inDset;

  inputParamsVec.push_back(structInputParams);

  // Set parameters (Input/Output).
  te::vp::AlgorithmParams* params = new te::vp::AlgorithmParams();
  params->setInputParams(inputParamsVec);
  params->setOutputDataSource(trgDs);
  params->setOutputDataSetName(outDS);
  params->setOutputSRID(inputSRID);
  params->setSpecificParams(specificParams);

  te::vp::Dissolve dissolve;
  bool result = dissolve.executeMemory(params);

  delete params;

  return result;
}

//Postgis to Postgis
bool DissolvePGISToPGIS()
{
  boost::uuids::basic_random_generator<boost::mt19937> gen;
  boost::uuids::uuid u = gen();
  std::string dsId = boost::uuids::to_string(u);

  std::string connInfo(
      "ppgsql://postgres:postgres@atlas.dpi.inpe.br:5433/testPostGIS");

  te::da::DataSourcePtr srcDs(
      te::da::DataSourceFactory::make("POSTGIS", connInfo));
  srcDs->setId(dsId);

  te::da::DataSourceInfoPtr dsInfoPtr(new te::da::DataSourceInfo());
  dsInfoPtr->setId(dsId);
  dsInfoPtr->setConnInfo(connInfo);
  dsInfoPtr->setType("POSTGIS");

  te::da::DataSourceInfoManager::getInstance().add(dsInfoPtr);

  srcDs->open();

  std::string inDsetName = "sp_cities";
  if (!srcDs->dataSetExists(inDsetName))
  {
    std::cout << "Input dataset not found: " << inDsetName << std::endl;
    return false;
  }
  
  te::da::DataSet* inDset = srcDs->getDataSet(inDsetName).release();
  te::da::DataSetType* inDsetType = srcDs->getDataSetType(inDsetName).release();

  std::vector<std::string> selProperties;
  selProperties.push_back("nomemeso");

  std::map<std::string, te::dt::AbstractData*> specificParams;
  specificParams["DISSOLVE"] =
      new te::vp::ComplexData<std::vector<std::string> >(selProperties);

  specificParams["IS_COLLECTION"] =
      new te::dt::SimpleData<bool, te::dt::BOOLEAN_TYPE>(true);

  std::map<te::dt::Property*, std::vector<te::stat::StatisticalSummary> > stats;

  std::unique_ptr<te::dt::Property> prop1 =
      srcDs->getProperty(inDsetName, "populaca");

  std::vector<te::stat::StatisticalSummary> stat1;
  stat1.push_back(te::stat::SUM);
  stat1.push_back(te::stat::MIN_VALUE);
  stat1.push_back(te::stat::MODE);
  stats.insert(std::make_pair(prop1.get(), stat1));

  specificParams["SUMMARY"] = new te::vp::ComplexData<
      std::map<te::dt::Property*, std::vector<te::stat::StatisticalSummary> > >(
          stats);

  te::gm::GeometryProperty* geomInputProp =
      te::da::GetFirstGeomProperty(inDsetType);

  int inputSRID = geomInputProp->getSRID();

  te::da::Select* inputSelect = new te::da::Select();
  te::da::Fields* fields = new te::da::Fields();

  te::da::Field* f_all = new te::da::Field("*");
  fields->push_back(f_all);
  inputSelect->setFields(fields);

  te::da::From* from = new te::da::From;

  te::da::FromItem* fromItem = new te::da::DataSetName(inDsetName);
  from->push_back(fromItem);
  inputSelect->setFrom(from);

  std::string outDS = "dissolve_pgistopgis";

  if (srcDs->dataSetExists(outDS))
  {
    std::cout << "A dataset with the same requested output dataset name "
                 "already exists: " << outDS << std::endl;
    return false;
  }
  
  // Can be added more than one input parameter.
  std::vector<te::vp::InputParams> inputParamsVec;

  // Declare the input parameters

  // Set the inputLayer parameters
  te::vp::InputParams structInputParams;

  structInputParams.m_inputDataSource = srcDs;
  structInputParams.m_inputDataSetType = inDsetType;
  structInputParams.m_inputDataSet = inDset;
  structInputParams.m_inputQuery = inputSelect;

  inputParamsVec.push_back(structInputParams);

  // Set parameters (Input/Output).
  te::vp::AlgorithmParams* params = new te::vp::AlgorithmParams();
  params->setInputParams(inputParamsVec);
  params->setOutputDataSource(srcDs);
  params->setOutputDataSetName(outDS);
  params->setOutputSRID(inputSRID);
  params->setSpecificParams(specificParams);

  te::vp::Dissolve dissolve;
  bool result = dissolve.executeQuery(params);

  delete params;

  return result;
}

//Postgis to OGR
bool DissolvePGISToOGR()
{
  boost::uuids::basic_random_generator<boost::mt19937> gen;
  boost::uuids::uuid u = gen();
  std::string dsId = boost::uuids::to_string(u);

  std::string connInfo(
      "ppgsql://postgres:postgres@atlas.dpi.inpe.br:5433/testPostGIS");

  te::da::DataSourcePtr srcDs(
      te::da::DataSourceFactory::make("POSTGIS", connInfo));
  srcDs->setId(dsId);

  te::da::DataSourceInfoPtr dsInfoPtr(new te::da::DataSourceInfo());
  dsInfoPtr->setId(dsId);
  dsInfoPtr->setConnInfo(connInfo);
  dsInfoPtr->setType("POSTGIS");

  te::da::DataSourceInfoManager::getInstance().add(dsInfoPtr);

  srcDs->open();
  
  std::string inDsetName = "sp_cities";
  if (!srcDs->dataSetExists(inDsetName))
  {
    std::cout << "Input dataset not found: " << inDsetName << std::endl;
    return false;
  }
  
  te::da::DataSet* inDset = srcDs->getDataSet(inDsetName).release();
  te::da::DataSetType* inDsetType = srcDs->getDataSetType(inDsetName).release();

  std::vector<std::string> selProperties;
  selProperties.push_back("nomemeso");

  std::map<std::string, te::dt::AbstractData*> specificParams;
  specificParams["DISSOLVE"] =
      new te::vp::ComplexData<std::vector<std::string> >(selProperties);

  specificParams["IS_COLLECTION"] =
      new te::dt::SimpleData<bool, te::dt::BOOLEAN_TYPE>(true);

  std::map<te::dt::Property*, std::vector<te::stat::StatisticalSummary> > stats;

  std::unique_ptr<te::dt::Property> prop1 =
      srcDs->getProperty(inDsetName, "populaca");

  std::vector<te::stat::StatisticalSummary> stat1;
  stat1.push_back(te::stat::SUM);
  stat1.push_back(te::stat::MIN_VALUE);
  stat1.push_back(te::stat::MODE);
  stats.insert(std::make_pair(prop1.get(), stat1));

  specificParams["SUMMARY"] = new te::vp::ComplexData<
      std::map<te::dt::Property*, std::vector<te::stat::StatisticalSummary> > >(
          stats);

  te::gm::GeometryProperty* geomInputProp =
      te::da::GetFirstGeomProperty(inDsetType);

  int inputSRID = geomInputProp->getSRID();

  te::da::Select* inputSelect = new te::da::Select();
  te::da::Fields* fields = new te::da::Fields();

  te::da::Field* f_all = new te::da::Field("*");
  fields->push_back(f_all);
  inputSelect->setFields(fields);

  te::da::From* from = new te::da::From;

  te::da::FromItem* fromItem = new te::da::DataSetName(inDsetName);
  from->push_back(fromItem);
  inputSelect->setFrom(from);

  std::string data_dir = TERRALIB_DATA_DIR;
  
  std::string uriResult(data_dir + "/shape/dissolve_pgistoogr.shp");
  std::string tgrInfo("file://" + uriResult);

  te::da::DataSourcePtr trgDs(te::da::DataSourceFactory::make("OGR", tgrInfo));
  trgDs->open();

  std::string outDS = "dissolve_pgistoogr";

  if (trgDs->dataSetExists(outDS))
  {
    std::cout << "A dataset with the same requested output dataset name "
                 "already exists: " << outDS << std::endl;
    return false;
  }

  // Can be added more than one input parameter.
  std::vector<te::vp::InputParams> inputParamsVec;

  // Declare the input parameters
  te::vp::InputParams structInputParams;

  // Set the inputLayer parameters
  structInputParams.m_inputDataSource = srcDs;
  structInputParams.m_inputDataSetType = inDsetType;
  structInputParams.m_inputDataSet = inDset;
  structInputParams.m_inputQuery = inputSelect;

  inputParamsVec.push_back(structInputParams);

  // Set parameters (Input/Output).
  te::vp::AlgorithmParams* params = new te::vp::AlgorithmParams();
  params->setInputParams(inputParamsVec);
  params->setOutputDataSource(trgDs);
  params->setOutputDataSetName(outDS);
  params->setOutputSRID(inputSRID);
  params->setSpecificParams(specificParams);

  te::vp::Dissolve dissolve;
  bool result = dissolve.executeQuery(params);

  delete params;

  return result;
}


