/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

 This file is part of the TerraLib - a Framework for building GIS enabled applications.

 TerraLib is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, either version 3 of the License,
 or (at your option) any later version.

 TerraLib is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with TerraLib. See COPYING. If not, write to
 TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
 \file terralib/itest/ogr/ITsOGR.cpp

 \brief A integration test suit for the TerraLib OGR Module.

 \author Mario Rocco Pettinati
 */

// TerraLib
#include <terralib_buildconfig.h>

#include <terralib/core/filesystem/FileSystem.h>
#include <terralib/dataaccess/dataset/DataSet.h>
#include <terralib/dataaccess/dataset/DataSetItem.h>
#include <terralib/dataaccess/datasource/DataSource.h>
#include <terralib/dataaccess/query/Expression.h>
#include <terralib/dataaccess/query/Field.h>
#include <terralib/dataaccess/query/Fields.h>
#include <terralib/dataaccess/query/From.h>
#include <terralib/dataaccess/query/FromItem.h>
#include <terralib/dataaccess/query/Select.h>
#include <terralib/dataaccess/query/Where.h>
#include <terralib/dataaccess/utils/BasicQueries.h>
#include <terralib/dataaccess/utils/DataAccess.h>
#include <terralib/dataaccess/utils/Utils.h>
#include <terralib/geometry/Envelope.h>
#include <terralib/geometry/Geometry.h>

// GTest and GMock
#include <gmock/gmock.h>
#include <gtest/gtest.h>

//STL include files
#include <memory>

namespace
{
  class ITsOGR : public ::testing::Test
  {
    virtual void SetUp()
    {
      
      
    }

    virtual void TearDown()
    {
    }
  };

  struct DataInfo
  {
    DataInfo(const std::string& fileName)
    {
      m_uri = "file://" + fileName;
      m_path = te::core::FileSystem::parentPath(fileName);
      m_fileName = fileName;
      m_datasetName = te::core::FileSystem::stem(fileName);
    }

    std::string m_uri;
    std::string m_path;
    std::string m_fileName;
    std::string m_datasetName;
  };

  DataInfo getBaseFileName()
  {
    std::string baseFileName = std::string(TERRALIB_DATA_DIR) + "/shape/municipios_desmatamento.shp";
    DataInfo baseDataInfo(baseFileName);

    return baseDataInfo;
  }

  DataInfo getOutputDataInfo(const std::string& fileName)
  {
    std::string outputPath = te::core::FileSystem::tempDirectoryPath() + "/terralib/ogr/test_results";
    outputPath = te::core::FileSystem::adjustPathToGenericForm(outputPath);
    outputPath += "/" + fileName;

    DataInfo outputFile(outputPath);
    return outputFile;
  }

  std::unique_ptr<te::da::DataAccess> getDataAccess(const DataInfo& dataInfo)
  {
    std::unique_ptr<te::da::DataAccess> dataAccess(new te::da::DataAccess(te::core::URI(dataInfo.m_uri), "OGR", dataInfo.m_datasetName));
    return dataAccess;
  }

  std::unique_ptr<te::da::DataAccess> createDataSet(te::da::DataAccess* baseDataAccess, const std::string& dataSetName, const std::string& extension)
  {
    std::unique_ptr<te::da::DataAccess> outputDataAccess;
    try
    {
      DataInfo outputDataInfo = getOutputDataInfo(dataSetName + "." + extension);
      std::string filePath = outputDataInfo.m_path;
      if (te::core::FileSystem::exists(filePath) == false)
      {
        te::core::FileSystem::createDirectories(filePath);
      }

      if (te::da::DataSource::exists("OGR", outputDataInfo.m_fileName) == true)
      {
        te::da::DataSource::drop("OGR", outputDataInfo.m_fileName);
      }

      std::unique_ptr<te::da::DataSetType> outputDataSetType(static_cast<te::da::DataSetType*>(baseDataAccess->getDataSetType()->clone()));
      outputDataSetType->setName(outputDataInfo.m_datasetName);

      outputDataAccess = getDataAccess(outputDataInfo);
      outputDataAccess->createDataSet(outputDataSetType.get());
    }
    catch (const std::exception&)
    {
      return std::unique_ptr<te::da::DataAccess>();
    }

    return outputDataAccess;
  }

  bool copyDataSet(te::da::DataAccess* inputDataAccess, te::da::DataAccess* outputDataAccess)
  {
    try
    {
      std::unique_ptr<te::da::DataSet> dataSet = inputDataAccess->getDataSet();

      outputDataAccess->saveDataSet(dataSet.release(), false);

      outputDataAccess->flush();
    }
    catch (const std::exception&)
    {
      return false;
    }

    return true;
  }

  void compareDataSetTypeFromDifferentGetApproaches(te::da::DataAccess* inputDataAccess)
  {
    //1 - acquires the dataSetType using GetDataSetType
    std::unique_ptr<te::da::DataSet> dataSetFromGet = inputDataAccess->getDataSet();
    std::vector<std::string> vecPropertyNamesGet;
    std::vector<int> vecPropertyTypesGet;
    for (std::size_t i = 0; i < dataSetFromGet->getNumProperties(); ++i)
    {
      vecPropertyNamesGet.push_back(dataSetFromGet->getPropertyName(i));
      vecPropertyTypesGet.push_back(dataSetFromGet->getPropertyDataType(i));
    }

    //2 - acquires the dataSetType using a query select * from dataSetName using implicit column names
    std::string query1 = "select * from " + inputDataAccess->getDataSetName();
    std::unique_ptr<te::da::DataSet> dataSetGromQueryAllImplicit = inputDataAccess->query(query1);

    std::vector<std::string> vecPropertyNamesQueryAllImplicit;
    std::vector<int> vecPropertyTypesQueryAllImplicit;
    for (std::size_t i = 0; i < dataSetGromQueryAllImplicit->getNumProperties(); ++i)
    {
      vecPropertyNamesQueryAllImplicit.push_back(dataSetGromQueryAllImplicit->getPropertyName(i));
      vecPropertyTypesQueryAllImplicit.push_back(dataSetGromQueryAllImplicit->getPropertyDataType(i));
    }

    //3 - acquires the dataSetType using a query select column1, column2 from dataSetName using implicit column names
    std::unique_ptr<te::da::Select> query2 = te::da::BasicQueries::GetSimpleQuery(inputDataAccess->getDataSetType());
    std::unique_ptr<te::da::DataSet> dataSetGromQueryAllExplicit = inputDataAccess->query(query2.get());

    std::vector<std::string> vecPropertyNamesQueryAllExplicit;
    std::vector<int> vecPropertyTypesQueryAllExplicit;
    for (std::size_t i = 0; i < dataSetGromQueryAllExplicit->getNumProperties(); ++i)
    {
      vecPropertyNamesQueryAllExplicit.push_back(dataSetGromQueryAllExplicit->getPropertyName(i));
      vecPropertyTypesQueryAllExplicit.push_back(dataSetGromQueryAllExplicit->getPropertyDataType(i));
    }

    //and now we compare the results

    ASSERT_TRUE(vecPropertyNamesGet == vecPropertyNamesQueryAllImplicit);
    ASSERT_TRUE(vecPropertyTypesGet == vecPropertyTypesQueryAllImplicit);

    ASSERT_TRUE(vecPropertyNamesGet == vecPropertyNamesQueryAllExplicit);
    ASSERT_TRUE(vecPropertyTypesGet == vecPropertyTypesQueryAllExplicit);
  }

  void validatePropertyAccessFunctions(te::da::DataAccess* dataAccess)
  {
    //check if column exists
    te::da::DataSourcePtr dataSource = dataAccess->getDataSource();
    std::string dataSetName = dataAccess->getDataSetName();

    boost::ptr_vector<te::dt::Property> vecProperties = dataSource->getProperties(dataSetName);
    for (std::size_t referencePropertyIndex = 0; referencePropertyIndex < vecProperties.size(); ++referencePropertyIndex)
    {
      const te::dt::Property& referenceProperty = vecProperties.at(referencePropertyIndex);
      std::string referencePropertyName = referenceProperty.getName();

      ASSERT_TRUE(dataSource->propertyExists(dataSetName, referencePropertyName));

      std::unique_ptr<te::dt::Property> propertyFroGetPropertyByName = dataSource->getProperty(dataSetName, referencePropertyName);
      ASSERT_EQ(referencePropertyName, propertyFroGetPropertyByName->getName());

      std::unique_ptr<te::dt::Property> propertyFroGetPropertyByIndex = dataSource->getProperty(dataSetName, referencePropertyIndex);
      ASSERT_EQ(referencePropertyName, propertyFroGetPropertyByIndex->getName());
    }
  }

  void transverseDataSet(te::da::DataSet* dataSet, std::size_t expectedCount)
  {
    ASSERT_FALSE(dataSet == nullptr);

    std::size_t geometryColumnPos = te::da::GetFirstSpatialPropertyPos(dataSet);
    std::size_t dataSetSize = 0;

    dataSet->moveBeforeFirst();
    while (dataSet->moveNext() == true)
    {
      std::unique_ptr<te::da::DataSetItem> item = dataSet->getItem();
      ASSERT_TRUE(item != nullptr);

      if (geometryColumnPos != std::string::npos)
      {
        std::unique_ptr<te::gm::Geometry> geometry = item->getGeometry(geometryColumnPos);
        ASSERT_FALSE(geometry == nullptr);

        ASSERT_TRUE(geometry->isValid());
      }
      
      ++dataSetSize;
    }
    ASSERT_EQ(expectedCount, dataSet->size());
    ASSERT_EQ(expectedCount, dataSetSize);
  }

  void transverseTwoDataSets(te::da::DataSet* dataSet1, te::da::DataSet* dataSet2)
  {
    ASSERT_FALSE(dataSet1 == nullptr);
    ASSERT_FALSE(dataSet2 == nullptr);

    std::size_t geometryColumnPos = te::da::GetFirstSpatialPropertyPos(dataSet1);

    dataSet1->moveBeforeFirst();
    dataSet2->moveBeforeFirst();
    while ((dataSet1->moveNext() == true) & (dataSet2->moveNext() == true))
    {
      std::unique_ptr<te::da::DataSetItem> item1 = dataSet1->getItem();
      std::unique_ptr<te::da::DataSetItem> item2 = dataSet2->getItem();
      ASSERT_TRUE(item1 != nullptr);
      ASSERT_TRUE(item2 != nullptr);

      if (geometryColumnPos != std::string::npos)
      {
        std::unique_ptr<te::gm::Geometry> geometry1 = item1->getGeometry(geometryColumnPos);
        std::unique_ptr<te::gm::Geometry> geometry2 = item2->getGeometry(geometryColumnPos);
        ASSERT_FALSE(geometry1 == nullptr);
        ASSERT_FALSE(geometry2 == nullptr);

        ASSERT_TRUE(geometry1->equals(geometry2.get()));
      }
    }

    ASSERT_TRUE(dataSet1->isAfterEnd());
    ASSERT_TRUE(dataSet2->isAfterEnd());
  }
  
  void transverseDataSet(te::da::DataAccess* dataAccess, std::size_t expectedCount)
  {
    std::unique_ptr<te::da::DataSet> dataSet = dataAccess->getDataSet();
    transverseDataSet(dataSet.get(), expectedCount);
  }

  void executeNonSpatialFilter(te::da::DataAccess* dataAccess, std::size_t expectedCount)
  {
    std::unique_ptr<te::da::Select> select(new te::da::Select());

    te::da::Fields* fields = new te::da::Fields();
    fields->push_back(new te::da::Field("id"));
    select->setFields(fields);
    select->setFrom(te::da::BasicQueries::GetFrom(dataAccess->getDataSetType()).release());
    select->setWhere(te::da::BasicQueries::GetWhereEqualTo("n_munic1", "PARAGOMINAS").release());

    std::unique_ptr<te::da::DataSet> dataSet = dataAccess->query(*select);
    transverseDataSet(dataSet.get(), expectedCount);
  }

  void executeAttributeFilter(te::da::DataAccess* dataAccess, std::size_t expectedCount)
  {
    std::unique_ptr<te::da::Select> select = te::da::BasicQueries::GetSimpleQueryEqualTo(dataAccess->getDataSetType(), "n_munic1", "PARAGOMINAS");

    std::unique_ptr<te::da::DataSet> dataSet = dataAccess->query(*select);
    transverseDataSet(dataSet.get(), expectedCount);
  }

  void executeSpatialFilter(te::da::DataAccess* dataAccess, std::size_t expectedCount)
  {
    double x = -46.91161000000000314;
    double y = -2.20822289999999999;
    double d = 0.001;
    te::gm::Envelope envelope(x - d, y - d, x + d, y + d);

    te::da::Select select;
    select.setFields(te::da::BasicQueries::GetFields(dataAccess->getDataSetType()).release());
    select.setFrom(te::da::BasicQueries::GetFrom(dataAccess->getDataSetType()).release());
    select.setWhere(new te::da::Where(te::da::BasicQueries::GetIntersectsExpression(dataAccess->getDataSetType(), envelope).release()));

    std::unique_ptr<te::da::DataSet> dataSet = dataAccess->query(select);
    transverseDataSet(dataSet.get(), expectedCount);
  }

  void readsTwoDataSetsSingleThread(te::da::DataAccess* dataAccess)
  {
    std::unique_ptr<te::da::DataSet> dataSet1 = dataAccess->getDataSet();
    std::unique_ptr<te::da::DataSet> dataSet2 = dataAccess->getDataSet();

    transverseTwoDataSets(dataSet1.get(), dataSet2.get());
  }


  void writeWhileResultsetIsOpened(te::da::DataAccess* baseDataAccess, te::da::DataAccess* dataAccess)
  {
    std::unique_ptr<te::da::DataSet> dataSet1 = baseDataAccess->getDataSet();
    std::unique_ptr<te::da::DataSet> dataSet2 = dataAccess->getDataSet();

    dataSet2->moveBeforeFirst();
    dataSet2->moveNext();
    dataSet2->moveNext();

    dataAccess->beginTransaction();

    dataAccess->saveDataSet(dataSet1.release());

    dataAccess->commitTransaction();
  }

  void runTestSequence(const std::string& extension)
  {
    ASSERT_FALSE(extension.empty());

    DataInfo baseDataInfo = getBaseFileName();

    //reads the base input;
    std::unique_ptr<te::da::DataAccess> baseDataAccess = getDataAccess(baseDataInfo);
    std::size_t baseInputSize = baseDataAccess->getDataSet()->size();
    ASSERT_TRUE(baseDataAccess->dataSetExists());

    //creates the output dataset in the given format
    std::unique_ptr<te::da::DataAccess> specificDataAccess = createDataSet(baseDataAccess.get(), "output_" + extension, extension);
    ASSERT_FALSE(specificDataAccess == nullptr);
    ASSERT_TRUE(specificDataAccess->dataSetExists());

    //copies the base dataSet into the output dataSet
    bool result = copyDataSet(baseDataAccess.get(), specificDataAccess.get());
    ASSERT_TRUE(result);

    ASSERT_EQ(specificDataAccess->getDataSetType()->size(), 14);

    //compares different approaches of getting the dataSetType
    compareDataSetTypeFromDifferentGetApproaches(specificDataAccess.get());

    //validates the functions that get the properties from the dataSource
    validatePropertyAccessFunctions(specificDataAccess.get());

    //reads the whole data in the given format (getDataSet)
    transverseDataSet(specificDataAccess.get(), baseInputSize);

    //queries the data using a command that requests only the id, withou any spatial asttribute
    executeNonSpatialFilter(specificDataAccess.get(), 1);

    //queries the data in the given format using attribute filter
    executeAttributeFilter(specificDataAccess.get(), 1);

    //queries the data in the given format using spatial filter
    executeSpatialFilter(specificDataAccess.get(), 1);

    //reads two dataSets pointing to the same dataSetName
    readsTwoDataSetsSingleThread(specificDataAccess.get());

    //transaction with resultset opened
    writeWhileResultsetIsOpened(baseDataAccess.get(), specificDataAccess.get());
  }
  
  TEST_F(ITsOGR, ShapeFile)
  {
    runTestSequence("shp");
  }
  
  TEST_F(ITsOGR, SQLite)
  {
    runTestSequence("sqlite");
  }
  
  TEST_F(ITsOGR, GeoPackage)
  {
    runTestSequence("gpkg");
  }
  
  TEST_F(ITsOGR, GeoJson)
  {
    runTestSequence("geojson");
  }
}