/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

 This file is part of the TerraLib - a Framework for building GIS enabled applications.

 TerraLib is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, either version 3 of the License,
 or (at your option) any later version.

 TerraLib is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with TerraLib. See COPYING. If not, write to
 TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
 \file terralib/itest/process/ITsBasicProcess.cpp

 \brief A integration test suit for the TerraLib Process Module.

 \author Mario Rocco Pettinati
 */

// TerraLib
#include <terralib_buildconfig.h>

#include <terralib/datatype/SimpleData.h>
#include <terralib/process/Connector.h>
#include <terralib/process/Task.h>
#include <terralib/process/TaskCapabilities.h>
#include <terralib/process/ParameterDescriptor.h>
#include <terralib/process/Process.h>
#include <terralib/process/Serializer.h>
#include <terralib/process/SimpleDataParameterValue.h>

// GTest and GMock
#include <gmock/gmock.h>
#include <gtest/gtest.h>

//Boost
#include <boost/property_tree/xml_parser.hpp>
#include <boost/property_tree/info_parser.hpp>
#include <boost/property_tree/json_parser.hpp>

#include <iostream>
#include <cstdio>

namespace
{
  class ITsBasicProcess : public ::testing::Test
  {
    virtual void SetUp()
    {
    }

    virtual void TearDown()
    {
    }
  };

  class ReadIntTask : public te::process::Task
  {
  public:

    ReadIntTask()
      : te::process::Task("ReadInt")
    {
      te::process::TaskCapabilities* capabilities = new te::process::TaskCapabilities();
      capabilities->addParameterDescriptor(te::process::ParameterDescriptor("number", te::dt::INT32_TYPE, true));
      capabilities->addParameterDescriptor(te::process::ParameterDescriptor("result", te::dt::INT32_TYPE, false, te::process::ParameterDescriptor::Direction::OUTPUT));
      init(capabilities);
    }

  protected:

    void runImpl() 
    {
      const te::process::ParameterValue* inputParam = this->getParameter(0);
      te::process::ParameterValue* outputParam = inputParam->clone();

      this->setParameter(1, outputParam, true);
    }
  };

  class ReadIntTaskFactory : public te::process::TaskFactory
  {
  public:

    ReadIntTaskFactory() : TaskFactory("ReadInt", "Calc")
    {
    }

  protected:

    te::process::Task* build() override
    {
      return new ReadIntTask();
    }
  };

  class SumIntTask : public te::process::Task
  {
  public:

    SumIntTask()
      : te::process::Task("SumInt")
    {
      te::process::TaskCapabilities* capabilities = new te::process::TaskCapabilities();
      capabilities->addParameterDescriptor(te::process::ParameterDescriptor("number1", te::dt::INT32_TYPE, true));
      capabilities->addParameterDescriptor(te::process::ParameterDescriptor("number2", te::dt::INT32_TYPE, true));
      capabilities->addParameterDescriptor(te::process::ParameterDescriptor("result", te::dt::INT32_TYPE, false, te::process::ParameterDescriptor::Direction::OUTPUT));
      init(capabilities);
    }

  protected:

    void runImpl() 
    {
      const te::process::ParameterValue* pValue1 = this->getParameter(0);
      const te::process::ParameterValue* pValue2 = this->getParameter(1);

      const te::dt::AbstractData* absData1 = (static_cast<const te::process::SimpleDataParameterValue*>(pValue1))->getValue();
      const te::dt::AbstractData* absData2 = (static_cast<const te::process::SimpleDataParameterValue*>(pValue2))->getValue();

      const te::dt::Int32* data1 = static_cast<const te::dt::Int32*>(absData1);
      const te::dt::Int32* data2 = static_cast<const te::dt::Int32*>(absData2);

      boost::int32_t sumResult = data1->getValue() + data2->getValue();
      te::dt::Int32* sumResultData = new te::dt::Int32(sumResult);
      te::process::SimpleDataParameterValue* sumResultValue = new te::process::SimpleDataParameterValue(sumResultData);

      this->setParameter(2, sumResultValue, true);
    }
  };

  class SumIntTaskFactory : public te::process::TaskFactory
  {
  public:

    SumIntTaskFactory() : TaskFactory("SumInt", "Calc")
    {
    }

  protected:

    te::process::Task* build() override
    {
      return new SumIntTask();
    }
  };

  class ShowIntSumResultTask : public te::process::Task
  {
  public:

    ShowIntSumResultTask()
      : te::process::Task("ShowIntSumResult")
    {
      te::process::TaskCapabilities* capabilities = new te::process::TaskCapabilities();
      capabilities->addParameterDescriptor(te::process::ParameterDescriptor("number", te::dt::INT32_TYPE, true));
      init(capabilities);
    }

  protected:

    void runImpl() 
    {
      const te::process::ParameterValue* pValue = this->getParameter(0);
      const te::dt::AbstractData* absData = (static_cast<const te::process::SimpleDataParameterValue*>(pValue))->getValue();

      const te::dt::Int32* data = static_cast<const te::dt::Int32*>(absData);
      int result = data->getValue();
      std::cout << "THE RESULT IS: " << result << std::endl;
    }
  };

  class ShowIntSumResultTaskFactory : public te::process::TaskFactory
  {
  public:

    ShowIntSumResultTaskFactory() : TaskFactory("ShowIntSumResult", "Calc")
    {
    }

  protected:

    te::process::Task* build() override
    {
      return new ShowIntSumResultTask();
    }
  };

  static ReadIntTaskFactory  ReadIntTaskFactoryInstance;
  static SumIntTaskFactory  SumIntTaskFactoryInstance;
  static ShowIntSumResultTaskFactory ShowIntSumResultTaskFactoryInstance;

  TEST_F(ITsBasicProcess, Sum)
  {
    //we first create the tasks for reading the input data
    te::process::TaskSharedPtr taskInput1(new ReadIntTask());
    taskInput1->setParameter(0, new te::process::SimpleDataParameterValue(new te::dt::Int32(10)), true);

    te::process::TaskSharedPtr taskInput2(new ReadIntTask());
    taskInput2->setParameter(0, new te::process::SimpleDataParameterValue(new te::dt::Int32(15)), true);

    //then we create the task to sum the data
    te::process::TaskSharedPtr sumTask(new SumIntTask());

    //here we connect the output from the read data tasks to the input of the sum task
    te::process::ConnectorSharedPtr connector1(new te::process::Connector());
    connector1->associate(taskInput1.get(), 1, sumTask.get(), 0);

    te::process::ConnectorSharedPtr connector2(new te::process::Connector());
    connector2->associate(taskInput2.get(), 1, sumTask.get(), 1);

    //finally, we create a task to show the results and connect it to the output of the sum task
    te::process::TaskSharedPtr showResulTask(new ShowIntSumResultTask());

    te::process::ConnectorSharedPtr connector3(new te::process::Connector());
    connector3->associate(sumTask.get(), 2, showResulTask.get(), 0);

    //now we create a process and add all the tasks to it
    std::unique_ptr<te::process::Process> process(new te::process::Process());
    process->addTask(taskInput1);
    process->addTask(taskInput2);
    process->addTask(sumTask);
    process->addTask(showResulTask);
    process->addConnector(connector1);
    process->addConnector(connector2);
    process->addConnector(connector3);

    //and finally we run it
    ASSERT_EQ(process->run(nullptr), true);

    const te::process::ParameterValue* pValue = showResulTask->getParameter(0);
    ASSERT_NE(pValue, nullptr);

    ASSERT_EQ(pValue->getDataType(), te::dt::INT32_TYPE);

    const te::dt::AbstractData* absData = (static_cast<const te::process::SimpleDataParameterValue*>(pValue))->getValue();
    ASSERT_NE(absData, nullptr);

    ASSERT_EQ(absData->getTypeCode(), te::dt::INT32_TYPE);

    const te::dt::Int32* intData = static_cast<const te::dt::Int32*>(absData);

    ASSERT_TRUE(intData->getValue() == 25);
  }

  std::unique_ptr<te::process::ParameterValue> Reconstruct(te::process::ParameterValue* inputParameterValue)
  {
    std::string stringValue = inputParameterValue->toString();
    int dataType = inputParameterValue->getDataType();

    //then we try to reconstruct it
    std::unique_ptr<te::process::ParameterValue> paramReconstructed(te::process::CreateParameterValue(dataType));

    //initializes the reconstructed object based on its string representation
    paramReconstructed->fromString(stringValue);

    return paramReconstructed;
  }

  TEST_F(ITsBasicProcess, ParameterValueReconstructionInt)
  {
    //we first convert the parameter value to its string representation
    std::unique_ptr<te::process::ParameterValue> param(new te::process::SimpleDataParameterValue(new te::dt::Int32(10)));
    std::unique_ptr<te::process::ParameterValue> paramReconstructed = Reconstruct(param.get());
    
    te::process::SimpleDataParameterValue* paramReconstructedCasted = dynamic_cast<te::process::SimpleDataParameterValue*>(paramReconstructed.get());
    ASSERT_NE(paramReconstructedCasted, nullptr);

    const te::dt::Int32* data = paramReconstructedCasted->getValueAs<te::dt::Int32>();
    ASSERT_NE(data, nullptr);

    ASSERT_EQ(data->getValue(), 10);
   }

  TEST_F(ITsBasicProcess, ParameterValueReconstructionDouble)
  {
    //we first convert the parameter value to its string representation
    std::unique_ptr<te::process::ParameterValue> param(new te::process::SimpleDataParameterValue(new te::dt::Double(10.5)));
    std::unique_ptr<te::process::ParameterValue> paramReconstructed = Reconstruct(param.get());

    te::process::SimpleDataParameterValue* paramReconstructedCasted = dynamic_cast<te::process::SimpleDataParameterValue*>(paramReconstructed.get());
    ASSERT_NE(paramReconstructedCasted, nullptr);

    const te::dt::Double* data = paramReconstructedCasted->getValueAs<te::dt::Double>();
    ASSERT_NE(data, nullptr);

    ASSERT_EQ(data->getValue(), 10.5);
  }

  TEST_F(ITsBasicProcess, SerializeTaskCapabilities)
  {
    te::process::TaskCapabilities capabilities;

    {
      te::process::ParameterDescriptor descriptor("parameter1", te::dt::DOUBLE_TYPE, true);
      descriptor.setDescription("This is parameter1");
      descriptor.setDefaultValue(new te::process::SimpleDataParameterValue(new te::dt::Double(10.5)));
      capabilities.addParameterDescriptor(descriptor);
    }
    {
      te::process::ParameterDescriptor descriptor("parameter2", te::dt::BOOLEAN_TYPE, true);
      descriptor.setDescription("This is parameter2");
      capabilities.addParameterDescriptor(descriptor);
    }

    te::process::Serializer serializer;
    boost::property_tree::ptree ptree = serializer.encodeTaskCapabilities(&capabilities);

    std::ostringstream xml;
    std::ostringstream json;
    std::ostringstream info;
    
    try
    {
#if BOOST_VERSION > 105600
      boost::property_tree::xml_writer_settings<std::string> settingsXML('\t', 1);
      boost::property_tree::info_writer_settings<char> settingsINFO('\t', 1);
#else
      boost::property_tree::xml_writer_settings<char> settings('\t', 1);
#endif

      boost::property_tree::write_xml(xml, ptree, settingsXML);
      boost::property_tree::write_json(json, ptree, true);
      boost::property_tree::write_info(info, ptree, settingsINFO);
    }
    catch (const boost::property_tree::xml_parser_error &)
    {
    }

    std::string xmlStr = xml.str();
    std::string jsonStr = json.str();
    std::string infoStr = info.str();

    ASSERT_NE(xmlStr.empty(), true);
    ASSERT_NE(jsonStr.empty(), true);
    ASSERT_NE(infoStr.empty(), true);


    std::unique_ptr<te::process::TaskCapabilities> resultCapabilites(serializer.decodeTaskCapabilities(ptree));
    ASSERT_EQ(capabilities.size(), resultCapabilites->size());
  }

  TEST_F(ITsBasicProcess, SerializationProcess)
  {
    //we first create the tasks for reading the input data
    te::process::TaskSharedPtr taskInput1(new ReadIntTask());
    taskInput1->setParameter(0, new te::process::SimpleDataParameterValue(new te::dt::Int32(10)), true);

    te::process::TaskSharedPtr taskInput2(new ReadIntTask());
    taskInput2->setParameter(0, new te::process::SimpleDataParameterValue(new te::dt::Int32(15)), true);

    //then we create the task to sum the data
    te::process::TaskSharedPtr sumTask(new SumIntTask());

    //here we connect the output from the read data tasks to the input of the sum task
    te::process::ConnectorSharedPtr connector1(new te::process::Connector());
    connector1->associate(taskInput1.get(), 1, sumTask.get(), 0);

    te::process::ConnectorSharedPtr connector2(new te::process::Connector());
    connector2->associate(taskInput2.get(), 1, sumTask.get(), 1);

    //finally, we create a task to show the results and connect it to the output of the sum task
    te::process::TaskSharedPtr showResulTask(new ShowIntSumResultTask());

    te::process::ConnectorSharedPtr connector3(new te::process::Connector());
    connector3->associate(sumTask.get(), 2, showResulTask.get(), 0);

    //now we create a process and add all the tasks to it
    std::unique_ptr<te::process::Process> process(new te::process::Process());
    process->addTask(taskInput1);
    process->addTask(taskInput2);
    process->addTask(sumTask);
    process->addTask(showResulTask);
    process->addConnector(connector1);
    process->addConnector(connector2);
    process->addConnector(connector3);

    te::process::Serializer serializer;
    boost::property_tree::ptree ptree = serializer.encodeProcess(process.get());

    std::ostringstream xml;
    std::ostringstream json;
    std::ostringstream info;

    try
    {
#if BOOST_VERSION > 105600
      boost::property_tree::xml_writer_settings<std::string> settingsXML('\t', 1);
      boost::property_tree::info_writer_settings<char> settingsINFO('\t', 1);
#else
      boost::property_tree::xml_writer_settings<char> settings('\t', 1);
#endif

      boost::property_tree::write_xml(xml, ptree, settingsXML);
      boost::property_tree::write_json(json, ptree, true);
      boost::property_tree::write_info(info, ptree, settingsINFO);
    }
    catch (const boost::property_tree::xml_parser_error&)
    {
    }

    std::string xmlStr = xml.str();
    std::string jsonStr = json.str();
    std::string infoStr = info.str();

    ASSERT_NE(xmlStr.empty(), true);
    ASSERT_NE(jsonStr.empty(), true);
    ASSERT_NE(infoStr.empty(), true);

    std::unique_ptr<te::process::Process> resultProcess(serializer.decodeProcess(ptree));
    ASSERT_EQ(process->getTasks().size(), resultProcess->getTasks().size());
    ASSERT_EQ(process->getConnectors().size(), resultProcess->getConnectors().size());

    ASSERT_EQ(resultProcess->run(nullptr), true);
  }
 }
 
