/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

 This file is part of the TerraLib - a Framework for building GIS enabled applications.

 TerraLib is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, either version 3 of the License,
 or (at your option) any later version.

 TerraLib is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with TerraLib. See COPYING. If not, write to
 TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
 \file terralib/itest/raster/ITsRaster.cpp

 \brief A integration test suit for the TerraLib Raster Module.

 \author Eric Silva Abreu
 */

// TerraLib
#include <terralib_buildconfig.h>
#include <terralib/core/utils/Platform.h>
#include <terralib/color/RGBAColor.h>
#include <terralib/common.h>
#include <terralib/dataaccess.h>
#include <terralib/core/plugin.h>
#include <terralib/gdal/Band.h>
#include <terralib/gdal/Raster.h>
#include <terralib/geometry/Envelope.h>
#include <terralib/memory/CachedRaster.h>
#include <terralib/raster/Grid.h>
#include <terralib/raster/RasterFactory.h>
#include <terralib/raster/Utils.h>
#include <terralib/srs/Converter.h>

// STL
#include <memory>
#include <map>
#include <string>

// GTest and GMock
#include <gmock/gmock.h>
#include <gtest/gtest.h>

std::string imageName = TERRALIB_DATA_DIR "/geotiff/Landsat8_22162_07102015_8bits_rec.tif";
std::string imageNameOut = TERRALIB_DATA_DIR "/geotiff/Landsat8_22162_07102015_8bits_rec_out.tif";

namespace
{
  std::unique_ptr<te::rst::Raster> OpenRaster(std::string path);

  void GetValuesByBlockDraw(te::rst::Raster* raster, te::rst::Grid* gridCanvas);
  
  void GetValuesByGDALDraw(te::rst::Raster* raster, te::rst::Grid* gridCanvas);

  class RasterTest : public ::testing::Test
  {
    virtual void SetUp()
    {
      static bool terralibInitialized = false;
      if (terralibInitialized == false)
      {
        TerraLib::getInstance().initialize();
        loadTL5Modules();
        terralibInitialized = true;
      }
    }

    virtual void TearDown()
    {

    }

    virtual void loadTL5Modules()
    {
      std::vector<std::string> vecPluginsToLoad;
      vecPluginsToLoad.push_back("te.da.gdal.teplg.json");

      std::string plugins_path = te::core::FindInTerraLibPath("share/terralib/plugins");
      for (std::size_t i = 0; i < vecPluginsToLoad.size(); ++i)
      {
        const std::string& pluginFileName = vecPluginsToLoad.at(i);

        te::core::PluginInfo info = te::core::JSONPluginInfoSerializer(plugins_path + "/" + pluginFileName);
        if (te::core::PluginManager::instance().isLoaded(info.name) == true)
        {
          continue;
        }

        te::core::PluginManager::instance().insert(info);
        te::core::PluginManager::instance().load(info.name);
      }
    }
  };

  TEST_F(RasterTest, RasterAccess_GetValuesBy_Block_Draw)
  {
    std::unique_ptr<te::rst::Raster> raster = OpenRaster(imageName);

    te::gm::Envelope* extent = new te::gm::Envelope(*raster->getGrid()->getExtent());
    std::unique_ptr<te::rst::Grid> gridCanvas(new te::rst::Grid((unsigned int)800, (unsigned int)600, extent, raster->getSRID()));

    GetValuesByBlockDraw(raster.get(), gridCanvas.get());

    ASSERT_TRUE(raster != nullptr);
  }

  TEST_F(RasterTest, RasterAccess_GetValuesBy_GDAL_Draw)
  {
    std::unique_ptr<te::rst::Raster> raster = OpenRaster(imageName);

    te::gm::Envelope* extent = new te::gm::Envelope(*raster->getGrid()->getExtent());
    std::unique_ptr<te::rst::Grid> gridCanvas(new te::rst::Grid((unsigned int)800, (unsigned int)600, extent, raster->getSRID()));

    GetValuesByGDALDraw(raster.get(), gridCanvas.get());

    ASSERT_TRUE(raster != nullptr);
  }

  TEST_F(RasterTest, RasterAccess_Resample_Draw_TerraLib)
  {
    std::unique_ptr<te::rst::Raster> raster = OpenRaster(imageName);

    ASSERT_TRUE(raster != nullptr);

    te::rst::Band* band1 = raster->getBand(0);
    te::rst::Band* band2 = raster->getBand(1);
    te::rst::Band* band3 = raster->getBand(2);

    te::rst::GetBufferValueFPtr m_getBuff;   //!< A pointer to a function that helps to extract a double or complex value from a specific buffer data type (char, int16, int32, float, ...).
    te::rst::GetBufferValueFPtr m_getBuffI;  //!< A pointer to a function that helps to extract the imaginary part value from a specific buffer data type (cint16, cint32, cfloat, cdouble).
    te::rst::SetBufferValueFPtr m_setBuff;   //!< A pointer to a function that helps to insert a double or complex value into a specific buffer data type (char, int16, int32, float, ...).
    te::rst::SetBufferValueFPtr m_setBuffI;  //!< A pointer to a function that helps to insert the imaginary part value into a specific buffer data type (cint16, cint32, cfloat, cdouble).

    te::rst::SetBlockFunctions(&m_getBuff, &m_getBuffI, &m_setBuff, &m_setBuffI, band1->getProperty()->getType());

    //-----------------------------------------------------

    int outputSizeWidth = 800;
    int outputSizeHeight = 600;

    int pxlsize = te::rst::GetPixelSize(band1->getProperty()->getType());
    int blksize = outputSizeWidth * outputSizeHeight * pxlsize;

    int blockWidth = band1->getProperty()->m_blkw;
    int blockHeight = band1->getProperty()->m_blkh;

    void* buffer1 = new unsigned char[blksize];
    void* buffer2 = new unsigned char[blksize];
    void* buffer3 = new unsigned char[blksize];

    int inputSizeWidth = 2498;
    int inputSizeHeight = 2523;

    int startX = 0;
    int startY = 0;

    band1->te::rst::Band::getBlockBuffer(buffer1, startX, startY, inputSizeWidth, inputSizeHeight, outputSizeWidth, outputSizeHeight, band1->getProperty()->getType());
    band2->te::rst::Band::getBlockBuffer(buffer2, startX, startY, inputSizeWidth, inputSizeHeight, outputSizeWidth, outputSizeHeight, band2->getProperty()->getType());
    band3->te::rst::Band::getBlockBuffer(buffer3, startX, startY, inputSizeWidth, inputSizeHeight, outputSizeWidth, outputSizeHeight, band3->getProperty()->getType());

    //-------------------------------

    std::map<std::string, std::string> rinfo;
    rinfo["URI"] = imageNameOut;

    te::rst::Grid* grid = new te::rst::Grid(outputSizeWidth, outputSizeHeight);

    std::vector<te::rst::BandProperty*> bands;

    for (std::size_t b = 0; b < 3; b++)
      bands.push_back(new te::rst::BandProperty(*(raster->getBand(b)->getProperty())));

    te::rst::Raster* rout = te::rst::RasterFactory::make(grid, bands, rinfo);

    //--------------------------------------
    for (unsigned int r = 0; r < rout->getNumberOfRows(); ++r)
    {
      for (unsigned int c = 0; c < rout->getNumberOfColumns(); ++c)
      {
        double dValueA = 0;
        double dValueB = 0;
        double dValueC = 0;

        m_getBuff(c + r * outputSizeWidth, buffer1, &dValueA);
        m_getBuff(c + r * outputSizeWidth, buffer2, &dValueB);
        m_getBuff(c + r * outputSizeWidth, buffer3, &dValueC);

        if (dValueA != 0. || dValueB != 0. || dValueC != 0.)
        {
          int a = 0;
        }

        rout->setValue(c, r, dValueA, 0);
        rout->setValue(c, r, dValueB, 1);
        rout->setValue(c, r, dValueC, 2);
      }
    }

    delete rout;
  }

  TEST_F(RasterTest, RasterAccess_Resample_Draw_GDAL)
  {
    std::unique_ptr<te::rst::Raster> raster = OpenRaster(imageName);

    ASSERT_TRUE(raster != nullptr);

    te::rst::Band* band1 = raster->getBand(0);
    te::rst::Band* band2 = raster->getBand(1);
    te::rst::Band* band3 = raster->getBand(2);

    te::rst::GetBufferValueFPtr m_getBuff;   //!< A pointer to a function that helps to extract a double or complex value from a specific buffer data type (char, int16, int32, float, ...).
    te::rst::GetBufferValueFPtr m_getBuffI;  //!< A pointer to a function that helps to extract the imaginary part value from a specific buffer data type (cint16, cint32, cfloat, cdouble).
    te::rst::SetBufferValueFPtr m_setBuff;   //!< A pointer to a function that helps to insert a double or complex value into a specific buffer data type (char, int16, int32, float, ...).
    te::rst::SetBufferValueFPtr m_setBuffI;  //!< A pointer to a function that helps to insert the imaginary part value into a specific buffer data type (cint16, cint32, cfloat, cdouble).

    te::rst::SetBlockFunctions(&m_getBuff, &m_getBuffI, &m_setBuff, &m_setBuffI, band1->getProperty()->getType());

    //-----------------------------------------------------

    int outputSizeWidth = 800;
    int outputSizeHeight = 600;

    int pxlsize = te::rst::GetPixelSize(band1->getProperty()->getType());
    int blksize = outputSizeWidth * outputSizeHeight * pxlsize;

    int blockWidth = band1->getProperty()->m_blkw;
    int blockHeight = band1->getProperty()->m_blkh;

    void* buffer1 = new unsigned char[blksize];
    void* buffer2 = new unsigned char[blksize];
    void* buffer3 = new unsigned char[blksize];

    int inputSizeWidth = 2498;
    int inputSizeHeight = 2523;

    int startX = 0;
    int startY = 0;

    ((te::gdal::Band*)band1)->getBlockBuffer(buffer1, startX, startY, inputSizeWidth, inputSizeHeight, outputSizeWidth, outputSizeHeight, band1->getProperty()->getType());
    ((te::gdal::Band*)band2)->getBlockBuffer(buffer2, startX, startY, inputSizeWidth, inputSizeHeight, outputSizeWidth, outputSizeHeight, band2->getProperty()->getType());
    ((te::gdal::Band*)band3)->getBlockBuffer(buffer3, startX, startY, inputSizeWidth, inputSizeHeight, outputSizeWidth, outputSizeHeight, band3->getProperty()->getType());

    //-------------------------------

    std::map<std::string, std::string> rinfo;
    rinfo["URI"] = imageNameOut;

    te::rst::Grid* grid = new te::rst::Grid(outputSizeWidth, outputSizeHeight);

    std::vector<te::rst::BandProperty*> bands;

    for (std::size_t b = 0; b < 3; b++)
      bands.push_back(new te::rst::BandProperty(*(raster->getBand(b)->getProperty())));

    te::rst::Raster* rout = te::rst::RasterFactory::make(grid, bands, rinfo);

    //--------------------------------------
    for (unsigned int r = 0; r < rout->getNumberOfRows(); ++r)
    {
      for (unsigned int c = 0; c < rout->getNumberOfColumns(); ++c)
      {
        double dValueA = 0;
        double dValueB = 0;
        double dValueC = 0;

        m_getBuff(c + r * outputSizeWidth, buffer1, &dValueA);
        m_getBuff(c + r * outputSizeWidth, buffer2, &dValueB);
        m_getBuff(c + r * outputSizeWidth, buffer3, &dValueC);

        if (dValueA != 0. || dValueB != 0. || dValueC != 0.)
        {
          int a = 0;
        }

        rout->setValue(c, r, dValueA, 0);
        rout->setValue(c, r, dValueB, 1);
        rout->setValue(c, r, dValueC, 2);
      }
    }

    delete rout;
  }


  std::unique_ptr<te::rst::Raster> OpenRaster(std::string path)
  {
    std::map<std::string, std::string> rinfo;
    rinfo["URI"] = path;
    te::rst::Raster* raster = te::rst::RasterFactory::open(rinfo);

    std::unique_ptr<te::rst::Raster> raster_ptr;
    raster_ptr.reset(raster);

    return raster_ptr;
  }

  void GetValuesByBlockDraw(te::rst::Raster* raster, te::rst::Grid* gridCanvas)
  {
    // create a SRS converter
    std::unique_ptr<te::srs::Converter> converter;

    if (raster->getSRID() != gridCanvas->getSRID())
    {
      converter.reset(new te::srs::Converter());

      converter->setSourceSRID(raster->getSRID());
      converter->setTargetSRID(gridCanvas->getSRID());
    }

    //band property information
    te::rst::BandProperty* bandProp = raster->getBand(0)->getProperty();

    std::size_t nBands = raster->getNumberOfBands();
    const int bandType = bandProp->getType();
    const int nblocksX = bandProp->m_nblocksx;
    const int nblocksY = bandProp->m_nblocksy;
    const int blockHeight = bandProp->m_blkh;
    const int blockWidht = bandProp->m_blkw;

    //function to get value from buffer
    te::rst::GetBufferValueFPtr m_getBuff;   //!< A pointer to a function that helps to extract a double or complex value from a specific buffer data type (char, int16, int32, float, ...).
    te::rst::GetBufferValueFPtr m_getBuffI;  //!< A pointer to a function that helps to extract the imaginary part value from a specific buffer data type (cint16, cint32, cfloat, cdouble).
    te::rst::SetBufferValueFPtr m_setBuff;   //!< A pointer to a function that helps to insert a double or complex value into a specific buffer data type (char, int16, int32, float, ...).
    te::rst::SetBufferValueFPtr m_setBuffI;  //!< A pointer to a function that helps to insert the imaginary part value into a specific buffer data type (cint16, cint32, cfloat, cdouble).

    te::rst::SetBlockFunctions(&m_getBuff, &m_getBuffI, &m_setBuff, &m_setBuffI, bandType);


    te::color::RGBAColor** rgba = new te::color::RGBAColor*[gridCanvas->getNumberOfRows()];

    for (unsigned int i = 0; i < gridCanvas->getNumberOfRows(); ++i)
    {
      rgba[i] = new te::color::RGBAColor[gridCanvas->getNumberOfColumns()];
    }

    //get raster values
    int* rgbValues = new int[nBands];

    for (int blkYIdx = 0; blkYIdx < nblocksY; ++blkYIdx)
    {
      for (int blkXIdx = 0; blkXIdx < nblocksX; ++blkXIdx)
      {
        void** arrayBuffer = new void*[nBands];

        for (std::size_t b = 0; b < nBands; ++b)
        {
          te::rst::Band* band = raster->getBand(b);

          void* buffer = band->read(blkXIdx, blkYIdx);

          arrayBuffer[b] = buffer;
        }

        for (int r = 0; r < blockHeight; ++r)
        {
          for (int c = 0; c < blockWidht; ++c)
          {
            for (std::size_t b = 0; b < nBands; ++b)
            {
              double dValue = 0;
              m_getBuff(c + r * blockWidht, arrayBuffer[b], &dValue);

              rgbValues[b] = static_cast<int>(dValue);
            }

            te::color::RGBAColor color(rgbValues[0], rgbValues[1], rgbValues[2], TE_OPAQUE);

            //remap
            int row = (blkYIdx * blockHeight) + r;
            int column = (blkXIdx * blockWidht) + c;

            te::gm::Coord2D inputGeo = raster->getGrid()->gridToGeo(row, column);

            if (converter.get())
              converter->convert(inputGeo.x, inputGeo.y, inputGeo.x, inputGeo.y);

            te::gm::Coord2D outputGrid = gridCanvas->geoToGrid(inputGeo.x, inputGeo.y);

            int x = te::rst::Round(outputGrid.x);
            int y = te::rst::Round(outputGrid.y);

            if ((x >= 0 && x < gridCanvas->getNumberOfColumns()) && (y >= 0 && y < gridCanvas->getNumberOfRows()))
              rgba[y][x] = color;
          }
        }

        delete arrayBuffer;
      }
    }

    delete rgbValues;
  }

  void GetValuesByGDALDraw(te::rst::Raster* raster, te::rst::Grid* gridCanvas)
  {
    //get band info
    int bandType = raster->getBandDataType(0);

    // get GDAL Data
    te::gdal::Raster* gdalRaster = dynamic_cast<te::gdal::Raster*>(raster);

    ASSERT_TRUE(gdalRaster != nullptr);

    //buffer size
    unsigned int canvasWidth = gridCanvas->getNumberOfColumns();
    unsigned int canvasHeight = gridCanvas->getNumberOfRows();
    
    int pxlsize = te::rst::GetPixelSize(bandType);
    int blksize = canvasWidth * canvasHeight * pxlsize;

    std::size_t nBands = gdalRaster->getNumberOfBands();

    void** arrayBuffer = new void*[nBands];

    //get buffers
    for (std::size_t b = 0; b < nBands; ++b)
    {
      te::gdal::Band* gdalBand = dynamic_cast<te::gdal::Band*>(gdalRaster->getBand(b));

      ASSERT_TRUE(gdalBand != nullptr);

      void* buffer = new unsigned char[blksize];

      gdalBand->getBlockBuffer(buffer, 0, 0, raster->getNumberOfColumns(), raster->getNumberOfRows(), canvasWidth, canvasHeight, bandType);

      arrayBuffer[b] = buffer;
    }

    //function to get value from buffer
    te::rst::GetBufferValueFPtr m_getBuff;   //!< A pointer to a function that helps to extract a double or complex value from a specific buffer data type (char, int16, int32, float, ...).
    te::rst::GetBufferValueFPtr m_getBuffI;  //!< A pointer to a function that helps to extract the imaginary part value from a specific buffer data type (cint16, cint32, cfloat, cdouble).
    te::rst::SetBufferValueFPtr m_setBuff;   //!< A pointer to a function that helps to insert a double or complex value into a specific buffer data type (char, int16, int32, float, ...).
    te::rst::SetBufferValueFPtr m_setBuffI;  //!< A pointer to a function that helps to insert the imaginary part value into a specific buffer data type (cint16, cint32, cfloat, cdouble).

    te::rst::SetBlockFunctions(&m_getBuff, &m_getBuffI, &m_setBuff, &m_setBuffI, bandType);

    //create output rgba matrix
    te::color::RGBAColor** rgba = new te::color::RGBAColor*[gridCanvas->getNumberOfRows()];

    for (unsigned int i = 0; i < gridCanvas->getNumberOfRows(); ++i)
    {
      rgba[i] = new te::color::RGBAColor[gridCanvas->getNumberOfColumns()];
    }

    //get raster values
    int* rgbValues = new int[nBands];

    for (int r = 0; r < canvasHeight; ++r)
    {
      for (int c = 0; c < canvasWidth; ++c)
      {
        for (std::size_t b = 0; b < nBands; ++b)
        {
          double dValue = 0;
          m_getBuff(c + r * canvasWidth, arrayBuffer[b], &dValue);

          rgbValues[b] = static_cast<int>(dValue);
        }

        te::color::RGBAColor color(rgbValues[0], rgbValues[1], rgbValues[2], TE_OPAQUE);

        rgba[r][c] = color;
      }
    }
  }
}
