/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

 This file is part of the TerraLib - a Framework for building GIS enabled applications.

 TerraLib is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, either version 3 of the License,
 or (at your option) any later version.

 TerraLib is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with TerraLib. See COPYING. If not, write to
 TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
 \file terralib/itest/raster/ITsRaster.cpp

 \brief A integration test suit for the TerraLib VP Module.

 \author Mario Rocco Pettinati
 */

// TerraLib
#include <terralib_buildconfig.h>

// TerraLib
#include <terralib/common/StringUtils.h>
#include <terralib/common/STLUtils.h>
#include <terralib/geometry.h>
#include <terralib/geometry/CommonDataStructures.h>
#include <terralib/dataaccess.h>
#include <terralib/datatype/SimpleProperty.h>
#include <terralib/datatype/StringProperty.h>
#include <terralib/geometry/GeometryProperty.h>
#include <terralib/vp/CommonFunctions.h>
#include <terralib/vp/Feature.h>
#include <terralib/vp/GeometrySnapper.h>
#include <terralib/vp/OverlapFixer.h>
#include <terralib/vp/Utils.h>
#include "TestUtils.h"

//GEOS include Files
#include <geos/algorithm/RayCrossingCounter.h>
#include <geos/geom.h>
#include <geos/geom/Coordinate.h>
#include <geos/geomgraph.h>
#include <geos/geomgraph/DirectedEdge.h>
#include <geos/geomgraph/Position.h>
#include <geos/geom/Geometry.h>
#include <geos/geom/GeometryFactory.h>
#include <geos/geom/MultiPolygon.h>
#include <geos/geom/Polygon.h>
#include <geos/io/WKTReader.h>
#include <geos/operation/overlay/OverlayOp.h>
#include <geos/operation/overlay/PolygonBuilder.h>

// GTest and GMock
#include <gmock/gmock.h>
#include <gtest/gtest.h>

//STL include files
#include <thread>

namespace
{
  class ItsFixOverlaps : public ::testing::Test
  {
    virtual void SetUp()
    {
    }

    virtual void TearDown()
    {
    }
  };

  TEST_F(ItsFixOverlaps, PointInPolygon)
  {
    geos::geom::Coordinate coordinateGEOS(5, -1);

    std::vector<geos::geom::Coordinate> vecCoordinates;
    vecCoordinates.push_back(geos::geom::Coordinate(5, 0));
    vecCoordinates.push_back(geos::geom::Coordinate(5, 5));
    vecCoordinates.push_back(geos::geom::Coordinate(0, 5));

    geos::algorithm::RayCrossingCounter pointInPolygonTester(coordinateGEOS);
    for (std::size_t i = 1; i < vecCoordinates.size(); ++i)
    {
      pointInPolygonTester.countSegment(vecCoordinates.at(i - 1), vecCoordinates.at(i));
    }

    bool pointOnPolygon = pointInPolygonTester.isPointInPolygon();
    ASSERT_FALSE(pointOnPolygon);
  }

  TEST_F(ItsFixOverlaps, TestPolygonizer)
  {
    te::gtest::DataInfo redDataInfo = te::gtest::GetRedDataInfo();

    te::vp::FeatureSet featureSet;
    te::gtest::GetFeaturesFromShapefile(redDataInfo.m_uri, featureSet);

    ASSERT_FALSE(featureSet.empty());

    te::gm::GeometryVectorConst inputGeometryVector = te::vp::GetGeometriesFromFeatureSet(featureSet);

    te::gm::GeometryVector vecSnapped = te::vp::GeometrySnapper::snapGeometries(inputGeometryVector, te::gm::GetAppropriatePrecision(inputGeometryVector.at(0)->getSRID()));

    te::gm::GeometryVector outputGeometryVector;
    te::gm::Polygonizer(vecSnapped, outputGeometryVector);

    te::gtest::SimpleSaveGeometriesToShapefile(te::gtest::GetOutputDir(), redDataInfo.m_datasetName + "_polygonizer", te::gm::ToConstVector(outputGeometryVector));

    te::common::FreeContents(featureSet);
    te::common::FreeContents(vecSnapped);
    te::common::FreeContents(outputGeometryVector);
  }

  TEST_F(ItsFixOverlaps, MaybeHasOverlapsTestCases)
  {
    std::vector< std::pair< std::string, std::string> > vecTestCases;
    std::vector<bool> vecExpectedResult;
    {
      //overlaps without crossing any segment
      std::string wktA("POLYGON ((0 0, 0 2, 0 8, 0 10, 10 10, 10 0, 0 0))");
      std::string wktB("POLYGON ((-2 2, -2 8, 0 8, 2 8, 2 2, 0 2, -2 2))");
      vecTestCases.emplace_back(std::pair< std::string, std::string>::pair(wktA, wktB));
      vecExpectedResult.push_back(true);
    }

    {
      //touches
      std::string wktA("POLYGON ((0 0, 0 2, 0 8, 0 10, 10 10, 10 0, 0 0))");
      std::string wktB("POLYGON ((-2 2, -2 8, 0 8, 0 2, -2 2))");
      vecTestCases.emplace_back(std::pair< std::string, std::string>::pair(wktA, wktB));
      vecExpectedResult.push_back(false);
    }

    {
      //overlaps with one crossing segment (both crossing segments have equal envelopes)
      std::string wktA("POLYGON((0 0, 0 2, 0 6, 6 6, 6 0, 3 2, 0 0))");
      std::string wktB("POLYGON ((0 0, 0 2, 3 0, 0 0))");
      vecTestCases.emplace_back(std::pair< std::string, std::string>::pair(wktA, wktB));
      vecExpectedResult.push_back(true);
    }

    {
      //two equal geometries
      std::string wktA("POLYGON ((-57.20354796178087 -14.338209298045475, -57.2030099617809 -14.338209298045507, -57.20300996178088 -14.338747298045176, -57.2024719617809 -14.338747298045211, -57.20247196178094 -14.338209298045536, -57.20193396178097 -14.338209298045578, -57.20193396178107 -14.336595298046568, -57.20247196178104 -14.336595298046534, -57.202471961781086 -14.336057298046855, -57.203009961781056 -14.336057298046834, -57.203009961781056 -14.335788298046994, -57.202026611646005 -14.335788298047055, -57.202026709997604 -14.335898132080683, -57.20174850138856 -14.335898368738551, -57.20147029277441 -14.335898605070307, -57.2011920841552 -14.335898841075943, -57.20091387553094 -14.335899076755455, -57.2008579617812 -14.335899124056258, -57.20085796178116 -14.336595298046628, -57.20058896178117 -14.336595298046648, -57.20058896178112 -14.337133298046327, -57.20031996178115 -14.337133298046336, -57.200319961781126 -14.337671298046011, -57.20005096178114 -14.33767129804602, -57.20005096178108 -14.338478298045528, -57.199781961781106 -14.338478298045548, -57.19978196178106 -14.339016298045216, -57.19951296178108 -14.339016298045232, -57.19951296178109 -14.338865588453302, -57.199181624321085 -14.339362594643019, -57.199180937437426 -14.339362136722432, -57.19916462247484 -14.339381580100147, -57.199137603998345 -14.339397179193622, -57.199106879719174 -14.339402596701937, -57.1989749617811 -14.339402596701953, -57.19897496178105 -14.340092298044608, -57.198705961781066 -14.340092298044626, -57.19870596178106 -14.34036129804446, -57.19843696178108 -14.340361298044476, -57.198436961781056 -14.340630298044308, -57.19816796178107 -14.34063029804432, -57.19816796178105 -14.340899298044162, -57.19789896178108 -14.340899298044182, -57.19789896178106 -14.341168298044021, -57.197629961781075 -14.34116829804403, -57.19762996178104 -14.341706298043698, -57.19736096178106 -14.341706298043714, -57.19736096178103 -14.341975298043554, -57.19682296178107 -14.34197529804359, -57.19682296178111 -14.341437298043914, -57.19655396178112 -14.341437298043937, -57.196553961781156 -14.340899298044262, -57.196284961781195 -14.340899298044281, -57.19628496178121 -14.340361298044613, -57.19601596178122 -14.340361298044625, -57.19601596178125 -14.340092298044798, -57.19574696178128 -14.340092298044814, -57.19574696178129 -14.339823298044978, -57.1956313101113 -14.339823298044985, -57.1956314393101 -14.339972253133581, -57.195631674589016 -14.340243503779343, -57.19563190987284 -14.340514754418916, -57.19563214516161 -14.340786005052294, -57.1953539304157 -14.340786234290718, -57.19535416537476 -14.341057484922416, -57.19535440033874 -14.341328735547913, -57.19535463530764 -14.341599986167223, -57.1956328510574 -14.34159975691525, -57.19563308636585 -14.341871007523842, -57.195633288648494 -14.342104182916056, -57.195633321679225 -14.342142258126247, -57.19563355699751 -14.342413508722448, -57.19568909162275 -14.342413462893814, -57.19568909162275 -14.342687611049419, -57.19550751608085 -14.342687611049419, -57.19550751608085 -14.342869186591312, -57.19532594053896 -14.34305076213321, -57.19532594053896 -14.343232337675106, -57.19514436499706 -14.343232337675106, -57.19496278945517 -14.343413913217002, -57.19496278945517 -14.343595488758897, -57.194781213913274 -14.343777064300793, -57.19459963837137 -14.343958639842691, -57.19441806282948 -14.344140215384584, -57.19441806282948 -14.344321790926482, -57.19423648728758 -14.344503366468379, -57.19423648728758 -14.344684942010272, -57.19405491174569 -14.34486651755217, -57.193873336203794 -14.34486651755217, -57.193691760661906 -14.34522966863596, -57.19351018512 -14.345411244177857, -57.19314703403621 -14.345411244177857, -57.192965458494314 -14.345592819719752, -57.19242073186862 -14.345592819719752, -57.19223915632673 -14.345411244177857, -57.192057580784834 -14.345411244177857, -57.19187600524294 -14.34522966863596, -57.191694429701045 -14.345048093094062, -57.19151285415914 -14.345048093094062, -57.191331278617255 -14.34486651755217, -57.191149703075354 -14.344684942010272, -57.190786551991565 -14.344684942010272, -57.19042340090777 -14.344503366468379, -57.190241825365874 -14.344321790926482, -57.188970796572605 -14.344321790926482, -57.188970796572605 -14.344503366468379, -57.188426069946914 -14.344503366468379, -57.188426069946914 -14.34486651755217, -57.18840198464709 -14.34489060285199, -57.18840195959232 -14.344860619711943, -57.18812373970094 -14.344860840533489, -57.187845519804824 -14.344861061028706, -57.18756729990396 -14.344861281197588, -57.18728907999839 -14.344861501040137, -57.1870108600881 -14.344861720556358, -57.18673264017312 -14.344861939746242, -57.18645442025343 -14.344862158609802, -57.186176200329044 -14.344862377147027, -57.18589798039997 -14.34486259535792, -57.185898204052336 -14.34513384604638, -57.185898427709375 -14.345405096728655, -57.18589857338865 -14.345581772472837, -57.18693290081666 -14.345581772472837, -57.18693290081666 -14.34576253921499, -57.18711366755881 -14.345943305957146, -57.187294434300966 -14.346124072699302, -57.18747520104312 -14.346304839441455, -57.18837903475389 -14.346304839441455, -57.18855980149605 -14.346485606183611, -57.1887405682382 -14.346485606183611, -57.18892133498036 -14.346666372925764, -57.18892133498036 -14.347027906410071, -57.187655967785275 -14.347027906410071, -57.187655967785275 -14.349016340573773, -57.18783673452743 -14.349377874058085, -57.188017501269584 -14.349558640800241, -57.18837903475389 -14.349920174284549, -57.18855980149605 -14.349920174284549, -57.18855980149605 -14.350100941026701, -57.1887405682382 -14.350100941026701, -57.18892133498036 -14.350281707768858, -57.18946363520682 -14.350281707768858, -57.18946363520682 -14.350462474511014, -57.190909769144056 -14.350462474511014, -57.191271302628365 -14.350643241253167, -57.19145206937052 -14.350824007995323, -57.19145206937052 -14.351727841706095, -57.191271302628365 -14.35190860844825, -57.19109053588622 -14.35190860844825, -57.190909769144056 -14.352089375190403, -57.190909769144056 -14.353173975643331, -57.19109053588622 -14.353354742385484, -57.192898203307756 -14.353354742385484, -57.19325973679207 -14.353173975643331, -57.19344050353423 -14.353173975643331, -57.19362127027638 -14.352993208901177, -57.19362127027638 -14.352812442159024, -57.193802037018536 -14.352812442159024, -57.19398280376069 -14.352631675416868, -57.194163570502845 -14.352631675416868, -57.194344337245 -14.352450908674715, -57.194525103987154 -14.352450908674715, -57.194525103987154 -14.352270141932559, -57.19506740421362 -14.352089375190403, -57.19560970444008 -14.352089375190403, -57.195790471182235 -14.352270141932559, -57.195790471182235 -14.352450908674715, -57.19597123792439 -14.352631675416868, -57.19597123792439 -14.352812442159024, -57.19601463716095 -14.352855841395584, -57.19605224270654 -14.352855841395584, -57.19623381824844 -14.353037416937482, -57.196370376954285 -14.353173975643331, -57.19687507163516 -14.353173975643331, -57.19705583837732 -14.352993208901177, -57.19723660511948 -14.352812442159024, -57.19723660511948 -14.350643241253167, -57.19705583837732 -14.350643241253167, -57.19687507163516 -14.350462474511014, -57.19669430489301 -14.350281707768858, -57.19651353815085 -14.350100941026701, -57.19560970444008 -14.350100941026701, -57.19542893769793 -14.349920174284549, -57.19542893769793 -14.349377874058085, -57.19560970444008 -14.349197107315929, -57.19560970444008 -14.348654807089465, -57.19542893769793 -14.348474040347313, -57.19524817095577 -14.348293273605156, -57.19506740421362 -14.348293273605156, -57.19488663747146 -14.348112506863004, -57.19488663747146 -14.347931740120847, -57.19470587072931 -14.347931740120847, -57.194525103987154 -14.347750973378693, -57.194344337245 -14.34757020663654, -57.194344337245 -14.347389439894384, -57.194525103987154 -14.347208673152227, -57.19850197231455 -14.347208673152227, -57.19886350579886 -14.347027906410071, -57.19922503928317 -14.346847139667918, -57.19958657276748 -14.346847139667918, -57.19976733950964 -14.346666372925764, -57.19976733950964 -14.346485606183611, -57.19994810625179 -14.346304839441455, -57.19994810625179 -14.34576253921499, -57.20012887299394 -14.345581772472837, -57.2003096397361 -14.345401005730682, -57.20049040647825 -14.34522023898853, -57.200671173220414 -14.34522023898853, -57.20085193996256 -14.345039472246373, -57.20121347344687 -14.345039472246373, -57.20121347344687 -14.3421472043719, -57.20284037412627 -14.3421472043719, -57.20292641703796 -14.342104182916056, -57.203184186783695 -14.341975298043186, -57.20300996178068 -14.341975298043195, -57.2030099617807 -14.34159355905784, -57.20300996178071 -14.341437298043527, -57.20193396178077 -14.341437298043596, -57.20193396178084 -14.34036129804426, -57.20247196178081 -14.340361298044227, -57.20247196178084 -14.339823298044552, -57.20300996178079 -14.33982329804452, -57.20300996178084 -14.339285298044846, -57.203547961780814 -14.339285298044814, -57.20354796178087 -14.338209298045475))");
      std::string wktB("POLYGON ((-57.20354796178087 -14.338209298045475, -57.2030099617809 -14.338209298045507, -57.20300996178088 -14.338747298045176, -57.2024719617809 -14.338747298045211, -57.20247196178094 -14.338209298045536, -57.20193396178097 -14.338209298045578, -57.20193396178107 -14.336595298046568, -57.20247196178104 -14.336595298046534, -57.202471961781086 -14.336057298046855, -57.203009961781056 -14.336057298046834, -57.203009961781056 -14.335788298046994, -57.202026611646005 -14.335788298047055, -57.202026709997604 -14.335898132080683, -57.20174850138856 -14.335898368738551, -57.20147029277441 -14.335898605070307, -57.2011920841552 -14.335898841075943, -57.20091387553094 -14.335899076755455, -57.2008579617812 -14.335899124056258, -57.20085796178116 -14.336595298046628, -57.20058896178117 -14.336595298046648, -57.20058896178112 -14.337133298046327, -57.20031996178115 -14.337133298046336, -57.200319961781126 -14.337671298046011, -57.20005096178114 -14.33767129804602, -57.20005096178108 -14.338478298045528, -57.199781961781106 -14.338478298045548, -57.19978196178106 -14.339016298045216, -57.19951296178108 -14.339016298045232, -57.19951296178109 -14.338865588453302, -57.199181624321085 -14.339362594643019, -57.199180937437426 -14.339362136722432, -57.19916462247484 -14.339381580100147, -57.199137603998345 -14.339397179193622, -57.199106879719174 -14.339402596701937, -57.1989749617811 -14.339402596701953, -57.19897496178105 -14.340092298044608, -57.198705961781066 -14.340092298044626, -57.19870596178106 -14.34036129804446, -57.19843696178108 -14.340361298044476, -57.198436961781056 -14.340630298044308, -57.19816796178107 -14.34063029804432, -57.19816796178105 -14.340899298044162, -57.19789896178108 -14.340899298044182, -57.19789896178106 -14.341168298044021, -57.197629961781075 -14.34116829804403, -57.19762996178104 -14.341706298043698, -57.19736096178106 -14.341706298043714, -57.19736096178103 -14.341975298043554, -57.19682296178107 -14.34197529804359, -57.19682296178111 -14.341437298043914, -57.19655396178112 -14.341437298043937, -57.196553961781156 -14.340899298044262, -57.196284961781195 -14.340899298044281, -57.19628496178121 -14.340361298044613, -57.19601596178122 -14.340361298044625, -57.19601596178125 -14.340092298044798, -57.19574696178128 -14.340092298044814, -57.19574696178129 -14.339823298044978, -57.1956313101113 -14.339823298044985, -57.1956314393101 -14.339972253133581, -57.195631674589016 -14.340243503779343, -57.19563190987284 -14.340514754418916, -57.19563214516161 -14.340786005052294, -57.1953539304157 -14.340786234290718, -57.19535416537476 -14.341057484922416, -57.19535440033874 -14.341328735547913, -57.19535463530764 -14.341599986167223, -57.1956328510574 -14.34159975691525, -57.19563308636585 -14.341871007523842, -57.195633288648494 -14.342104182916056, -57.195633321679225 -14.342142258126247, -57.19563355699751 -14.342413508722448, -57.19568909162275 -14.342413462893814, -57.19568909162275 -14.342687611049419, -57.19550751608085 -14.342687611049419, -57.19550751608085 -14.342869186591312, -57.19532594053896 -14.34305076213321, -57.19532594053896 -14.343232337675106, -57.19514436499706 -14.343232337675106, -57.19496278945517 -14.343413913217002, -57.19496278945517 -14.343595488758897, -57.194781213913274 -14.343777064300793, -57.19459963837137 -14.343958639842691, -57.19441806282948 -14.344140215384584, -57.19441806282948 -14.344321790926482, -57.19423648728758 -14.344503366468379, -57.19423648728758 -14.344684942010272, -57.19405491174569 -14.34486651755217, -57.193873336203794 -14.34486651755217, -57.193691760661906 -14.34522966863596, -57.19351018512 -14.345411244177857, -57.19314703403621 -14.345411244177857, -57.192965458494314 -14.345592819719752, -57.19242073186862 -14.345592819719752, -57.19223915632673 -14.345411244177857, -57.192057580784834 -14.345411244177857, -57.19187600524294 -14.34522966863596, -57.191694429701045 -14.345048093094062, -57.19151285415914 -14.345048093094062, -57.191331278617255 -14.34486651755217, -57.191149703075354 -14.344684942010272, -57.190786551991565 -14.344684942010272, -57.19042340090777 -14.344503366468379, -57.190241825365874 -14.344321790926482, -57.188970796572605 -14.344321790926482, -57.188970796572605 -14.344503366468379, -57.188426069946914 -14.344503366468379, -57.188426069946914 -14.34486651755217, -57.18840198464709 -14.34489060285199, -57.18840195959232 -14.344860619711943, -57.18812373970094 -14.344860840533489, -57.187845519804824 -14.344861061028706, -57.18756729990396 -14.344861281197588, -57.18728907999839 -14.344861501040137, -57.1870108600881 -14.344861720556358, -57.18673264017312 -14.344861939746242, -57.18645442025343 -14.344862158609802, -57.186176200329044 -14.344862377147027, -57.18589798039997 -14.34486259535792, -57.185898204052336 -14.34513384604638, -57.185898427709375 -14.345405096728655, -57.18589857338865 -14.345581772472837, -57.18693290081666 -14.345581772472837, -57.18693290081666 -14.34576253921499, -57.18711366755881 -14.345943305957146, -57.187294434300966 -14.346124072699302, -57.18747520104312 -14.346304839441455, -57.18837903475389 -14.346304839441455, -57.18855980149605 -14.346485606183611, -57.1887405682382 -14.346485606183611, -57.18892133498036 -14.346666372925764, -57.18892133498036 -14.347027906410071, -57.187655967785275 -14.347027906410071, -57.187655967785275 -14.349016340573773, -57.18783673452743 -14.349377874058085, -57.188017501269584 -14.349558640800241, -57.18837903475389 -14.349920174284549, -57.18855980149605 -14.349920174284549, -57.18855980149605 -14.350100941026701, -57.1887405682382 -14.350100941026701, -57.18892133498036 -14.350281707768858, -57.18946363520682 -14.350281707768858, -57.18946363520682 -14.350462474511014, -57.190909769144056 -14.350462474511014, -57.191271302628365 -14.350643241253167, -57.19145206937052 -14.350824007995323, -57.19145206937052 -14.351727841706095, -57.191271302628365 -14.35190860844825, -57.19109053588622 -14.35190860844825, -57.190909769144056 -14.352089375190403, -57.190909769144056 -14.353173975643331, -57.19109053588622 -14.353354742385484, -57.192898203307756 -14.353354742385484, -57.19325973679207 -14.353173975643331, -57.19344050353423 -14.353173975643331, -57.19362127027638 -14.352993208901177, -57.19362127027638 -14.352812442159024, -57.193802037018536 -14.352812442159024, -57.19398280376069 -14.352631675416868, -57.194163570502845 -14.352631675416868, -57.194344337245 -14.352450908674715, -57.194525103987154 -14.352450908674715, -57.194525103987154 -14.352270141932559, -57.19506740421362 -14.352089375190403, -57.19560970444008 -14.352089375190403, -57.195790471182235 -14.352270141932559, -57.195790471182235 -14.352450908674715, -57.19597123792439 -14.352631675416868, -57.19597123792439 -14.352812442159024, -57.19601463716095 -14.352855841395584, -57.19605224270654 -14.352855841395584, -57.19623381824844 -14.353037416937482, -57.196370376954285 -14.353173975643331, -57.19687507163516 -14.353173975643331, -57.19705583837732 -14.352993208901177, -57.19723660511948 -14.352812442159024, -57.19723660511948 -14.350643241253167, -57.19705583837732 -14.350643241253167, -57.19687507163516 -14.350462474511014, -57.19669430489301 -14.350281707768858, -57.19651353815085 -14.350100941026701, -57.19560970444008 -14.350100941026701, -57.19542893769793 -14.349920174284549, -57.19542893769793 -14.349377874058085, -57.19560970444008 -14.349197107315929, -57.19560970444008 -14.348654807089465, -57.19542893769793 -14.348474040347313, -57.19524817095577 -14.348293273605156, -57.19506740421362 -14.348293273605156, -57.19488663747146 -14.348112506863004, -57.19488663747146 -14.347931740120847, -57.19470587072931 -14.347931740120847, -57.194525103987154 -14.347750973378693, -57.194344337245 -14.34757020663654, -57.194344337245 -14.347389439894384, -57.194525103987154 -14.347208673152227, -57.19850197231455 -14.347208673152227, -57.19886350579886 -14.347027906410071, -57.19922503928317 -14.346847139667918, -57.19958657276748 -14.346847139667918, -57.19976733950964 -14.346666372925764, -57.19976733950964 -14.346485606183611, -57.19994810625179 -14.346304839441455, -57.19994810625179 -14.34576253921499, -57.20012887299394 -14.345581772472837, -57.2003096397361 -14.345401005730682, -57.20049040647825 -14.34522023898853, -57.200671173220414 -14.34522023898853, -57.20085193996256 -14.345039472246373, -57.20121347344687 -14.345039472246373, -57.20121347344687 -14.3421472043719, -57.20284037412627 -14.3421472043719, -57.20292641703796 -14.342104182916056, -57.203184186783695 -14.341975298043186, -57.20300996178068 -14.341975298043195, -57.2030099617807 -14.34159355905784, -57.20300996178071 -14.341437298043527, -57.20193396178077 -14.341437298043596, -57.20193396178084 -14.34036129804426, -57.20247196178081 -14.340361298044227, -57.20247196178084 -14.339823298044552, -57.20300996178079 -14.33982329804452, -57.20300996178084 -14.339285298044846, -57.203547961780814 -14.339285298044814, -57.20354796178087 -14.338209298045475))");
      vecTestCases.emplace_back(std::pair< std::string, std::string>::pair(wktA, wktB));
      vecExpectedResult.push_back(true);
    }

    for (std::size_t i = 0; i < vecTestCases.size(); ++i)
    {
      std::unique_ptr<te::gm::Geometry> geometryA(te::gm::WKTReader::read(vecTestCases.at(i).first.c_str()));
      std::unique_ptr<te::gm::Geometry> geometryB(te::gm::WKTReader::read(vecTestCases.at(i).second.c_str()));

      te::gm::SpatialRelation relation = te::vp::FastRelationCalculator(geometryA.get(), geometryB.get());
      bool result = (relation == te::gm::SpatialRelation::OVERLAPS);
      ASSERT_TRUE(result == vecExpectedResult.at(i));
    }
  }

  TEST_F(ItsFixOverlaps, MaybeHasOverlapsRealData)
  {
    te::gtest::DataInfo redDataInfo = te::gtest::GetRedDataInfo();

    te::vp::FeatureSet featureSet;
    te::gtest::GetFeaturesFromShapefile(redDataInfo.m_uri, featureSet);

    ASSERT_FALSE(featureSet.empty());

    te::gm::GeometryVectorConst inputGeometryVector = te::vp::GetGeometriesFromFeatureSet(featureSet);

    std::vector<te::gm::Geometry*> vecSnapped = te::vp::GeometrySnapper::snapGeometries(inputGeometryVector, te::gm::GetAppropriatePrecision(inputGeometryVector.at(0)->getSRID()));

    int optimizationCount = 0;
    int missedCount = 0;
    for (std::size_t i = 0; i < vecSnapped.size() / 10; ++i)
    {
      for (std::size_t j = i+1; j < vecSnapped.size() / 10; ++j)
      {
        const te::gm::Geometry* geometryA = vecSnapped.at(i);
        const te::gm::Geometry* geometryB = vecSnapped.at(j);

        const te::gm::Envelope* mbrA = geometryA->getMBR();
        const te::gm::Envelope* mbrB = geometryB->getMBR();
        if (mbrA->intersects(*mbrB) == true)
        {
          te::gm::SpatialRelation relation = te::vp::FastRelationCalculator(geometryA, geometryB);

          bool maybeHasOverlaps = (relation == te::gm::SpatialRelation::OVERLAPS);
          bool hasOverlaps = te::vp::OverlapFixer::hasOverlaps(geometryA, geometryB);

          if (maybeHasOverlaps == false && hasOverlaps == false)
          {
            ++optimizationCount;
          }

          if (maybeHasOverlaps == true && hasOverlaps == false)
          {
            ++missedCount;
          }

          //detecting false positives
          ASSERT_FALSE(maybeHasOverlaps == false && hasOverlaps == true);
        }
      }
    }

    te::common::FreeContents(featureSet);
    te::common::FreeContents(vecSnapped);
  }

  TEST_F(ItsFixOverlaps, DetectOverlaps)
  {
    te::gtest::DataInfo redDataInfo = te::gtest::GetRedDataInfo();

    te::vp::FeatureSet featureSet;
    te::gtest::GetFeaturesFromShapefile(redDataInfo.m_uri, featureSet);

    ASSERT_FALSE(featureSet.empty());

    te::gm::GeometryVectorConst inputGeometryVector = te::vp::GetGeometriesFromFeatureSet(featureSet);

    te::gm::GeometryVector vecSnapped = te::vp::GeometrySnapper::snapGeometries(inputGeometryVector, te::gm::GetAppropriatePrecision(inputGeometryVector.at(0)->getSRID()));

    bool wasChanged = false;
    te::gm::GeometryVector vecIntersection = te::vp::OverlapFixer::detectOverlaps(vecSnapped);

    te::gtest::SimpleSaveGeometriesToShapefile(te::gtest::GetOutputDir(), redDataInfo.m_datasetName + "_fragments", te::gm::ToConstVector(vecIntersection));

    te::common::FreeContents(featureSet);
    te::common::FreeContents(vecSnapped);
    te::common::FreeContents(vecIntersection);
  }

  TEST_F(ItsFixOverlaps, FixOverlaps)
  {
    te::gtest::DataInfo redDataInfo = te::gtest::GetRedDataInfo();

    te::vp::FeatureSet featureSet;
    te::gtest::GetFeaturesFromShapefile(redDataInfo.m_uri, featureSet);

    ASSERT_FALSE(featureSet.empty());

    te::gm::GeometryVectorConst inputGeometryVector = te::vp::GetGeometriesFromFeatureSet(featureSet);

    te::gm::GeometryVector vecSnapped = te::vp::GeometrySnapper::snapGeometries(inputGeometryVector, te::gm::GetAppropriatePrecision(inputGeometryVector.at(0)->getSRID()));

    std::vector<te::vp::OverlapFixerReport*> vecReport = te::vp::OverlapFixer::fix(vecSnapped);
    
    te::gtest::DataToSave attributeOriginalId("fix_original_id");
    te::gtest::DataToSave attributeOriginalId2("fix_original_id_big_column_name");
    te::gtest::DataToSave attributeResult("fix_result");

    te::gm::GeometryVectorConst vecOutputGeometries;
    for (std::size_t i = 0; i < vecReport.size(); ++i)
    {
      te::vp::OverlapFixerReport* report = vecReport.at(i);

      te::gm::Geometry* geometry = report->m_geometry.get();
      if (geometry == nullptr || geometry->isEmpty() == true)
      {
        continue;
      }

      std::string resultStr = "UNCHANGED";
      if (report->m_result == te::vp::OverlapFixerReport::FixResult::FIXED)
      {
        resultStr = "FIXED";
      }
      else if (report->m_result == te::vp::OverlapFixerReport::FixResult::REMOVED)
      {
        resultStr = "REMOVED";
      }
      else if (report->m_result == te::vp::OverlapFixerReport::FixResult::NEW)
      {
        resultStr = "NEW";
      }

      vecOutputGeometries.push_back(geometry);
      attributeOriginalId.m_vecValues.push_back(te::common::Convert2String(static_cast<boost::uint32_t>(i)));
      attributeOriginalId2.m_vecValues.push_back(te::common::Convert2String(static_cast<boost::uint32_t>(i)));
      attributeResult.m_vecValues.push_back(resultStr);
    }

    std::vector<te::gtest::DataToSave> vecDataToSave;
    vecDataToSave.push_back(attributeOriginalId);
    vecDataToSave.push_back(attributeOriginalId2);
    vecDataToSave.push_back(attributeResult);

    te::gtest::SimpleSaveGeometriesToShapefile(te::gtest::GetOutputDir(), redDataInfo.m_datasetName + "_fixed", vecOutputGeometries, vecDataToSave);

    te::common::FreeContents(featureSet);
    te::common::FreeContents(vecSnapped);
    te::common::FreeContents(vecReport);
  }
}

