/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

 This file is part of the TerraLib - a Framework for building GIS enabled applications.

 TerraLib is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, either version 3 of the License,
 or (at your option) any later version.

 TerraLib is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with TerraLib. See COPYING. If not, write to
 TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
 \file terralib/itest/raster/ITsGroupThreadManager.cpp

 \brief A integration test suit for the TerraLib VP Module.

 \author Mario Rocco Pettinati
 */

// TerraLib
#include <terralib_buildconfig.h>

// TerraLib
#include "TestUtils.h"
#include <terralib/common/TerraLib.h>
#include <terralib/common/StringUtils.h>
#include <terralib/common/STLUtils.h>
#include <terralib/core/plugin.h>
#include <terralib/core/utils.h>
#include <terralib/core/filesystem/FileSystem.h>
#include <terralib/dataaccess/query_h.h>
#include <terralib/dataaccess/dataset/DataSet.h>
#include <terralib/dataaccess/dataset/DataSetType.h>
#include <terralib/dataaccess/utils/BasicQueries.h>
#include <terralib/dataaccess/utils/DataAccess.h>
#include <terralib/dataaccess/utils/Utils.h>
#include <terralib/datatype.h>
#include <terralib/memory/PreparedItemCopy.h>
#include <terralib/vp/AbstractDataHandler.h>
#include <terralib/vp/AbstractOperation.h>
#include <terralib/vp/CommonFunctions.h>
#include <terralib/vp/Feature.h>
#include <terralib/vp/GeometrySnapper.h>
#include <terralib/vp/IntersectionOperation.h>
#include <terralib/vp/NoGroupingDataHandler.h>
#include <terralib/vp/OperationThreadManager.h>
#include <terralib/vp/SubdivideDataHandler.h>
#include <terralib/vp/SubdivideByBorderTillingStrategy.h>
#include <terralib/vp/Utils.h>

//boost
#include <boost/thread/mutex.hpp>
#include <boost/thread/lock_guard.hpp>

// GTest and GMock
#include <gmock/gmock.h>
#include <gtest/gtest.h>

//STL include files
#include <thread>
#include <vector>

namespace
{
  class ITsGroupThreadManager : public ::testing::Test
  {
    virtual void SetUp()
    {
    }

    virtual void TearDown()
    {
    }
  };

  class CopyOperation : public te::vp::AbstractOperation
  {
  public:

    CopyOperation()
      : AbstractOperation("Copy")
    {
    }

    virtual std::unique_ptr<te::vp::AbstractOperationCapabilities> getCapabilities() const
    {
      std::unique_ptr<te::vp::AbstractOperationCapabilities> capabilities(new te::vp::AbstractOperationCapabilities(getName(), 1, 1, "CopyDataHandler", false, false));
      return capabilities;
    }

    std::vector<te::vp::FeatureSet> executeImpl(const std::vector<te::vp::FeatureSet>& vecInput)
    {
      te::mem::PreparedItemCopy copy;

      std::vector<te::vp::FeatureSet> vecOutput;
      for (std::size_t i = 0; i < vecInput.size(); ++i)
      {
        const te::vp::FeatureSet& inputFeatureSet = vecInput.at(i);
        te::vp::FeatureSet outputFeatureSet;
        for (std::size_t j = 0; j < inputFeatureSet.size(); ++j)
        {
          te::vp::Feature* inputFeature = inputFeatureSet.at(j);
          const te::gm::Geometry* geometry = inputFeature->getFirstGeometryPtr();

          //create the output feature based on the resulting geometry and the input attributes
          te::gm::GeometryVector vecAdjustedOutput = te::vp::AdjustToGeomType(geometry, getOutputGeometryType());
          for (std::size_t k = 0; k < vecAdjustedOutput.size(); ++k)
          {
            te::vp::Feature* outputFeature = this->createOutputFeature(i);
            copy.copyValues(inputFeature, outputFeature, false);
            outputFeature->setFirstGeometry(vecAdjustedOutput.at(k));

            outputFeatureSet.push_back(outputFeature);
          }
        }

        vecOutput.push_back(outputFeatureSet);
      }

      return vecOutput;
    }
  };

  TEST_F(ITsGroupThreadManager, CopyingNoSQLite)
  {
    te::gtest::DataInfo redDataInfo = te::gtest::GetRedDataInfo();

    //loads the input data access
    std::unique_ptr<te::da::DataAccess> inputRedDataAccess(new te::da::DataAccess(te::core::URI(redDataInfo.m_uri), "OGR", redDataInfo.m_datasetName));

    //creates the output
    std::unique_ptr<te::da::DataSetType> outputDataSetType(static_cast<te::da::DataSetType*>(inputRedDataAccess->getDataSetType()->clone()));
    std::string outputDataSetName = redDataInfo.m_datasetName + "_copy_no_sqlite";
    std::string outputFileName = te::gtest::GetOutputDir() + "/" + outputDataSetName + ".shp";
    if (te::core::FileSystem::exists(outputFileName) == true)
    {
      te::core::FileSystem::remove(outputFileName);
    }

    std::unique_ptr<te::da::DataAccess> outputDataAccess(new te::da::DataAccess(te::core::URI(outputFileName), "OGR", outputDataSetName));
    
    te::vp::AbstractParameters* params = new te::vp::AbstractParameters();
    params->addInputDataAccess(inputRedDataAccess.get());
    params->addOutputDataAccess(outputDataAccess.get());

    te::vp::AbstractOperation* operation = new CopyOperation();
    operation->setParameters(params);
    te::vp::AbstractDataHandler* dataHandler = new te::vp::NoGroupingDataHandler();

    te::vp::OperationThreadManager threadManager(operation, dataHandler);
    threadManager.start();
  }

  TEST_F(ITsGroupThreadManager, CopyingSQLiteWithSubdivision)
  {
    te::gtest::DataInfo redDataInfo = te::gtest::GetRedDataInfo();

    //loads the input data access
    std::unique_ptr<te::da::DataAccess> inputRedDataAccess(new te::da::DataAccess(te::core::URI(redDataInfo.m_uri), "OGR", redDataInfo.m_datasetName));

    //creates the output
    std::unique_ptr<te::da::DataSetType> outputDataSetType(static_cast<te::da::DataSetType*>(inputRedDataAccess->getDataSetType()->clone()));
    std::string outputDataSetName = redDataInfo.m_datasetName + "_copy_sqlite_with_subdivide";
    std::string outputFileName = te::gtest::GetOutputDir() + "/" + outputDataSetName + ".shp";
    if (te::core::FileSystem::exists(outputFileName) == true)
    {
      te::core::FileSystem::remove(outputFileName);
    }

    std::unique_ptr<te::da::DataAccess> outputDataAccess(new te::da::DataAccess(te::core::URI(outputFileName), "OGR", outputDataSetName));
    
    te::vp::AbstractParameters* params = new te::vp::AbstractParameters();
    params->addInputDataAccess(inputRedDataAccess.get());
    params->addOutputDataAccess(outputDataAccess.get());

    te::vp::AbstractOperation* operation = new CopyOperation();
    operation->setParameters(params);
    te::vp::AbstractDataHandler* dataHandler = new te::vp::SubdivideDataHandler();

    te::vp::OperationThreadManager threadManager(operation, dataHandler);
    threadManager.start();
  }
}

TEST_F(ITsGroupThreadManager, LazyIntersectionNoSQLite)
{
  te::gtest::DataInfo redDataInfo = te::gtest::GetRedDataInfo();
  te::gtest::DataInfo blueDataInfo = te::gtest::GetBlueDataInfo();

  //loads the input data access
  std::unique_ptr<te::da::DataAccess> inputRedDataAccess(new te::da::DataAccess(te::core::URI(redDataInfo.m_uri), "OGR", redDataInfo.m_datasetName));
  std::unique_ptr<te::da::DataAccess> inputBlueDataAccess(new te::da::DataAccess(te::core::URI(blueDataInfo.m_uri), "OGR", blueDataInfo.m_datasetName));

  //creates the output
  std::unique_ptr<te::da::DataSetType> outputDataSetType(static_cast<te::da::DataSetType*>(inputRedDataAccess->getDataSetType()->clone()));
  std::string outputDataSetName = redDataInfo.m_datasetName + "_intersection_no_sqlite";
  std::string outputFileName = te::gtest::GetOutputDir() + "/" + outputDataSetName + ".shp";
  if (te::core::FileSystem::exists(outputFileName) == true)
  {
    te::core::FileSystem::remove(outputFileName);
  }

  std::unique_ptr<te::da::DataAccess> outputDataAccess(new te::da::DataAccess(te::core::URI(outputFileName), "OGR", outputDataSetName));
  
  te::vp::AbstractParameters* params = new te::vp::AbstractParameters();
  params->addInputDataAccess(inputRedDataAccess.get());
  params->addInputDataAccess(inputBlueDataAccess.get());
  params->addOutputDataAccess(outputDataAccess.get());

  te::vp::AbstractOperation* operation = new te::vp::IntersectionOperation();
  operation->setParameters(params);
  te::vp::AbstractDataHandler* dataHandler = new te::vp::NoGroupingDataHandler();

  te::vp::OperationThreadManager threadManager(operation, dataHandler);
  threadManager.start();
}

TEST_F(ITsGroupThreadManager, LazyIntersectionSQLiteWithSubdivide)
{
  te::gtest::DataInfo redDataInfo = te::gtest::GetRedDataInfo();
  te::gtest::DataInfo blueDataInfo = te::gtest::GetBlueDataInfo();

  //loads the input data access
  std::unique_ptr<te::da::DataAccess> inputRedDataAccess(new te::da::DataAccess(te::core::URI(redDataInfo.m_uri), "OGR", redDataInfo.m_datasetName));
  std::unique_ptr<te::da::DataAccess> inputBlueDataAccess(new te::da::DataAccess(te::core::URI(blueDataInfo.m_uri), "OGR", blueDataInfo.m_datasetName));

  //creates the output
  std::unique_ptr<te::da::DataSetType> outputDataSetType(static_cast<te::da::DataSetType*>(inputRedDataAccess->getDataSetType()->clone()));
  std::string outputDataSetName = redDataInfo.m_datasetName + "_intersection_sqlite_with_subdivide";
  std::string outputFileName = te::gtest::GetOutputDir() + "/" + outputDataSetName + ".shp";
  if (te::core::FileSystem::exists(outputFileName) == true)
  {
    te::core::FileSystem::remove(outputFileName);
  }

  std::unique_ptr<te::da::DataAccess> outputDataAccess(new te::da::DataAccess(te::core::URI(outputFileName), "OGR", outputDataSetName));

  te::vp::AbstractParameters* params = new te::vp::AbstractParameters();
  params->addInputDataAccess(inputRedDataAccess.get());
  params->addInputDataAccess(inputBlueDataAccess.get());
  params->addOutputDataAccess(outputDataAccess.get());

  te::vp::AbstractOperation* operation = new te::vp::IntersectionOperation();
  operation->setParameters(params);
  te::vp::AbstractDataHandler* dataHandler = new te::vp::SubdivideDataHandler();

  te::vp::OperationThreadManager threadManager(operation, dataHandler);
  threadManager.start();
}
