/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

 This file is part of the TerraLib - a Framework for building GIS enabled applications.

 TerraLib is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, either version 3 of the License,
 or (at your option) any later version.

 TerraLib is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with TerraLib. See COPYING. If not, write to
 TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
 \file terralib/itest/raster/ItsSpatialIndex.cpp

 \brief A integration test suit for the TerraLib VP Module.

 \author Mario Rocco Pettinati
 */

// TerraLib
#include <terralib_buildconfig.h>

// TerraLib
#include <terralib/common/StringUtils.h>
#include <terralib/core/uri/URI.h>
#include <terralib/dataaccess/utils/DataAccess.h>
#include <terralib/vp/AbstractOperation.h>
#include <terralib/vp/OperationThreadManager.h>
#include <terralib/vp/SubdivideDataHandler.h>

#include <terralib/geometry/algorithms/CellTilling.h>
#include <terralib/geometry/LinearRing.h>
#include <terralib/geometry/Polygon.h>
#include <terralib/geometry/GeometryCollection.h>
#include <terralib/geometry/Utils.h>
#include <terralib/vp/SubdivideOperation.h>
#include <terralib/vp/DissolveSubdivisionOperation.h>
#include <terralib/vp/Feature.h>
#include <terralib/vp/SubdivideByBorderTillingStrategy.h>
#include <terralib/vp/Utils.h>

#include <boost/lexical_cast.hpp>
#include <boost/numeric/ublas/matrix.hpp>

#include "TestUtils.h"

// GTest and GMock
#include <gmock/gmock.h>
#include <gtest/gtest.h>

namespace
{
  class ITsOperations : public ::testing::Test
  {
    virtual void SetUp()
    {
    }

    virtual void TearDown()
    {
    }
  };

  void execute1(te::vp::AbstractOperation* operation, const std::vector<te::da::DataAccess*>& vecInputDataAccess, const std::vector<te::da::DataAccess*>& vecOutputDataAccess, te::vp::AbstractParameters* params)
  {
    if (params == nullptr)
    {
      params = new te::vp::AbstractParameters();
    }

    params->setInputDataAccessVector(vecInputDataAccess);
    params->setOutputDataAccessVector(vecOutputDataAccess);

    params->setMaxCoordinates(1000);

    operation->setParameters(params);
    te::vp::AbstractDataHandler* dataHandler = te::vp::CreateDataHandler(operation->getCapabilities()->getDataHandlerName());

    te::vp::OperationThreadManager threadManager(operation, dataHandler);
    threadManager.start();
  }

  
  std::vector<te::da::DataAccess*> ToVectorPtr(const std::vector<te::da::DataAccessPtr>& vecDataAccessPtr)
  {
    std::vector<te::da::DataAccess*> vecDataAccess;
    for (std::size_t i = 0; i < vecDataAccessPtr.size(); ++i)
    {
      vecDataAccess.push_back(vecDataAccessPtr.at(i).get());
    }

    return vecDataAccess;
  }

  te::da::DataAccessPtr CreateTemporaryDataAccess()
  {
    std::string dir = te::core::FileSystem::tempDirectoryPath() + "/terralib/vp/test_results";

    std::string tempds = "tempda_" + te::common::GetUniqueID();

    te::da::DataSourcePtr dataSource = te::vp::CreateTemporaryDataSource();
    te::da::DataAccessPtr dataAccess(new te::da::DataAccess(dataSource, tempds));

    return dataAccess;
  }


  std::vector<te::da::DataAccessPtr> execute(te::vp::AbstractOperation* operation, const std::vector<te::gtest::DataInfo>& vecDataInfo, te::vp::AbstractParameters* params = nullptr)
  {
    std::string outputDataSetName = operation->getName();

    //loads the input data access
    std::vector<te::da::DataAccess*> vecInputDataAccess;
    for (std::size_t i = 0; i < vecDataInfo.size(); ++i)
    {
      const te::gtest::DataInfo& dataInfo = vecDataInfo.at(i);
      te::da::DataAccess* dataAccess = new te::da::DataAccess(te::core::URI(dataInfo.m_uri), "OGR", dataInfo.m_datasetName);

      vecInputDataAccess.push_back(dataAccess);
    }

    //creates the output
    std::size_t numberOfOutputs = operation->getCapabilities()->getNumberOfOutputs();
    std::vector<te::da::DataAccessPtr> vecOutputDataAccess;
    for (std::size_t i = 0; i < numberOfOutputs; ++i)
    {
      std::string name = outputDataSetName;
      if (numberOfOutputs > 1)
      {
        name += "_o" + te::common::Convert2String(static_cast<boost::uint32_t>(i + 1));
      }

      std::string outputFileName = te::gtest::GetOutputDir() + "/" + name + ".shp";
      outputFileName = te::core::FileSystem::adjustPathToNativeForm(outputFileName);
      if (te::da::DataSource::exists("OGR", outputFileName) == true)
      {
        te::da::DataSource::drop("OGR", outputFileName);
      }      
      
      if (te::core::FileSystem::exists(outputFileName) == true)
      {
        te::core::FileSystem::remove(outputFileName);
      }

      

      te::da::DataAccessPtr outputDataAccess(new te::da::DataAccess(te::core::URI(outputFileName), "OGR", name));
      vecOutputDataAccess.push_back(std::move(outputDataAccess));
    }
 
    execute1(operation, vecInputDataAccess, ToVectorPtr(vecOutputDataAccess), params);

    te::common::FreeContents(vecInputDataAccess);

    return vecOutputDataAccess;
  }

  TEST_F(ITsOperations, MakeValid)
  {
    std::vector<te::gtest::DataInfo> vecDataInfo;
    vecDataInfo.push_back(te::gtest::GetRedDataInfo());
    
    te::vp::AbstractOperation* operation = te::vp::OperationFactory::make("MakeValid");
    execute(operation, vecDataInfo);
  }

  TEST_F(ITsOperations, Intersection)
  {
    std::vector<te::gtest::DataInfo> vecDataInfo;
    vecDataInfo.push_back(te::gtest::GetRedDataInfo());
    vecDataInfo.push_back(te::gtest::GetBlueDataInfo());

    te::vp::AbstractOperation* operation = te::vp::OperationFactory::make("Intersection");
    execute(operation, vecDataInfo);
  }

  TEST_F(ITsOperations, Difference)
  {
    std::vector<te::gtest::DataInfo> vecDataInfo;
    vecDataInfo.push_back(te::gtest::GetRedDataInfo());
    vecDataInfo.push_back(te::gtest::GetBlueDataInfo());

    te::vp::AbstractOperation* operation = te::vp::OperationFactory::make("Difference");
    execute(operation, vecDataInfo);
  }

  TEST_F(ITsOperations, Identity)
  {
    std::vector<te::gtest::DataInfo> vecDataInfo;
    vecDataInfo.push_back(te::gtest::GetRedDataInfo());
    vecDataInfo.push_back(te::gtest::GetBlueDataInfo());

    te::vp::AbstractOperation* operation = te::vp::OperationFactory::make("Identity");
    execute(operation, vecDataInfo);
  }  

  TEST_F(ITsOperations, Union)
  {
    std::vector<te::gtest::DataInfo> vecDataInfo;
    vecDataInfo.push_back(te::gtest::GetRedDataInfo());
    vecDataInfo.push_back(te::gtest::GetBlueDataInfo());

    te::vp::AbstractOperation* operation = te::vp::OperationFactory::make("Union");
    execute(operation, vecDataInfo);
  }

  TEST_F(ITsOperations, Dissolve)
  {
    std::vector<te::gtest::DataInfo> vecDataInfo;
    vecDataInfo.push_back(te::gtest::GetRedDataInfo());

    te::vp::AbstractOperation* operation = te::vp::OperationFactory::make("Dissolve");
    execute(operation, vecDataInfo);
  }

  TEST_F(ITsOperations, Subdivide)
  {
    te::gtest::DataInfo redDataInfo = te::gtest::GetRedDataInfo();

    std::unique_ptr<te::da::DataAccess> dataAccess(new te::da::DataAccess(te::core::URI(redDataInfo.m_uri), "OGR", redDataInfo.m_datasetName));
    te::gm::Envelope envelope = dataAccess->getMBR();

    std::vector<te::gtest::DataInfo> vecDataInfo;
    vecDataInfo.push_back(redDataInfo);

    te::gm::CellTilling* cellTilling = new te::gm::CellTilling(envelope, envelope.getWidth() / 20., envelope.getHeight() / 20.);
    te::vp::SubdivideOperationParameters* params = new te::vp::SubdivideOperationParameters(500, cellTilling);

    te::vp::AbstractOperation* operation = te::vp::OperationFactory::make("Subdivide");
    execute(operation, vecDataInfo, params);
  }

  TEST_F(ITsOperations, DissolveSubdivision)
  {
    te::gtest::DataInfo dummyDataInfo("");

    std::vector<te::gtest::DataInfo> vecDataInfo;
    vecDataInfo.push_back(te::gtest::DataInfo(te::gtest::GetOutputDir() + "/Subdivide.shp"));

    te::vp::AbstractOperation* operation = te::vp::OperationFactory::make("DissolveSubdivision");

    std::string rowPropertyname = te::vp::SubdivideByBorderTillingStrategy::getRowPropertyName();
    std::string columnPropertyname = te::vp::SubdivideByBorderTillingStrategy::getColumnPropertyName();
    te::vp::DissolveSubdivisionOperationParameters* params = new te::vp::DissolveSubdivisionOperationParameters(rowPropertyname, columnPropertyname, true);

    execute(operation, vecDataInfo, params);
  }
 }
 
