/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

 This file is part of the TerraLib - a Framework for building GIS enabled applications.

 TerraLib is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, either version 3 of the License,
 or (at your option) any later version.

 TerraLib is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with TerraLib. See COPYING. If not, write to
 TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
 \file terralib/itest/raster/ItsSpatialIndex.cpp

 \brief A integration test suit for the TerraLib VP Module.

 \author Mario Rocco Pettinati
 */

// TerraLib
#include <terralib_buildconfig.h>

// TerraLib
#include <terralib/common/STLUtils.h>
#include <terralib/vp/KDTreeMultiIndex.h>
#include <terralib/vp/RTreeIndex.h>
#include <terralib/vp/Utils.h>

#include "TestUtils.h"

// GTest and GMock
#include <gmock/gmock.h>
#include <gtest/gtest.h>

namespace
{
  class ItsSpatialIndex : public ::testing::Test
  {
    virtual void SetUp()
    {
    }

    virtual void TearDown()
    {
    }
  };

  TEST_F(ItsSpatialIndex, RTreeBasic)
  {
    te::vp::RTreeIndex<std::size_t> index;
    index.insert(te::gm::Envelope(0, 0, 10, 10), 0);
    index.insert(te::gm::Envelope(5, 5, 15, 15), 1);
    index.insert(te::gm::Envelope(10, 10, 20, 20), 2);

    {
      te::gm::Envelope s1(0, 0, 20, 20);
      ASSERT_EQ(index.search(s1)->size(), 3);
    }
    {
      te::gm::Envelope s1(10, 10, 10, 10);
      ASSERT_EQ(index.search(s1)->size(), 3);
    }
    {
      te::gm::Envelope s1(-5, -5, 5, 5);
      ASSERT_EQ(index.search(s1)->size(), 2);
    }

    {
      std::vector<std::size_t> vecReport = index.getData()->getAllData();
      ASSERT_EQ(vecReport.size(), 3);
    }
    
    index.remove(te::gm::Envelope(5, 5, 15, 15), 1);

    {
      te::gm::Envelope s1(0, 0, 20, 20);
      ASSERT_EQ(index.search(s1)->size(), 2);
    }
    {
      te::gm::Envelope s1(10, 10, 10, 10);
      ASSERT_EQ(index.search(s1)->size(), 2);
    }
    {
      te::gm::Envelope s1(-5, -5, 5, 5);
      ASSERT_EQ(index.search(s1)->size(), 1);
    }

    {
      std::vector<std::size_t> vecReport = index.getData()->getAllData();
      ASSERT_EQ(vecReport.size(), 2);
    }
  }

  TEST_F(ItsSpatialIndex, KDTreeBasic)
  {
    te::vp::KDTreeMultiIndex<std::size_t> index;
    index.insert(te::gm::Coord2D(0, 0), 0);
    index.insert(te::gm::Coord2D(5, 5), 1);
    index.insert(te::gm::Coord2D(10, 10), 2);
    index.insert(te::gm::Coord2D(10, 10), 3); //this tree allows duplicated keys

    {
      te::gm::Envelope s1(0, 0, 20, 20);
      ASSERT_EQ(index.search(s1)->size(), 4);
    }
    {
      te::gm::Envelope s1(10, 10, 10, 10);
      ASSERT_EQ(index.search(s1)->size(), 2);
    }
    {
      te::gm::Envelope s1(-5, -5, 5, 5);
      ASSERT_EQ(index.search(s1)->size(), 2);
    }

    {
      std::vector<std::size_t> vecReport = index.getData()->getAllData();
      ASSERT_EQ(vecReport.size(), 4);
    }

    index.remove(te::gm::Coord2D(5, 5), 1);

    {
      te::gm::Envelope s1(0, 0, 20, 20);
      ASSERT_EQ(index.search(s1)->size(), 3);
    }
    {
      te::gm::Envelope s1(10, 10, 10, 10);
      ASSERT_EQ(index.search(s1)->size(), 2);
    }
    {
      te::gm::Envelope s1(-5, -5, 5, 5);
      ASSERT_EQ(index.search(s1)->size(), 1);
    }

    {
      std::vector<std::size_t> vecReport = index.getData()->getAllData();
      ASSERT_EQ(vecReport.size(), 3);
    }
  }
}
 
