/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

 This file is part of the TerraLib - a Framework for building GIS enabled applications.

 TerraLib is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, either version 3 of the License,
 or (at your option) any later version.

 TerraLib is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with TerraLib. See COPYING. If not, write to
 TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
 \file terralib/itest/raster/ITsSubDivide.cpp

 \brief A integration test suit for the TerraLib VP Module.

 \author Mario Rocco Pettinati
 */

// TerraLib
#include <terralib_buildconfig.h>

// TerraLib
#include <terralib/common/STLUtils.h>
#include <terralib/geometry/algorithms/CellTilling.h>
#include <terralib/geometry/algorithms/GeometrySubdivider.h>
#include <terralib/geometry/CommonDataStructures.h>
#include <terralib/geometry/Geometry.h>
#include <terralib/geometry/Polygon.h>
#include <terralib/geometry/Utils.h>
#include <terralib/geometry/WKTReader.h>
#include "TestUtils.h"


// GTest and GMock
#include <gmock/gmock.h>
#include <gtest/gtest.h>

//STL include files
#include <thread>

namespace
{
  class ItsSubDivide : public ::testing::Test
  {
    virtual void SetUp()
    {
    }

    virtual void TearDown()
    {
    }
  };

  TEST_F(ItsSubDivide, SubdivideByMaxCoordinates)
  {
    std::size_t maxCoordinates = 10;

    te::gm::Envelope envelope(0, 0, 10, 10);
    te::gm::Polygon polygon = te::gm::CreatePolygon(envelope, 0, 9);

    std::string wkt = polygon.toString();

    te::gm::GeometryVector geometryVector = te::gm::GeometrySubdivider::subdivideByMaxCoordinates(&polygon, maxCoordinates);
    te::gm::GeometryPtr result = te::gm::CreateCollectionFromGeometryVector(geometryVector);
    std::string wktResult = result->toString();

    for (std::size_t i = 0; i < geometryVector.size(); ++i)
    {
      const te::gm::Geometry* geometry = geometryVector.at(i);
      ASSERT_LE(geometry->getNPoints(), maxCoordinates);
    }

    te::common::FreeContents(geometryVector);
  }

  TEST_F(ItsSubDivide, SubdivideByMaxCoordinatesInvalid)
  {
    std::size_t maxCoordinates = 10;

    std::string wkt = "POLYGON((300 200, 350 200, 400 200, 450 200, 500 200, 500 250, 500 300, 450 300, 450 250, 400 250, 400 300, 450 300, 450 350, 400 350, 350 350, 300 350, 300 300, 300 250, 300 200))";
    std::unique_ptr<te::gm::Geometry> geometryA(te::gm::WKTReader::read(wkt.c_str()));

    te::gm::GeometryVector geometryVector = te::gm::GeometrySubdivider::subdivideByMaxCoordinates(geometryA.get(), maxCoordinates);
    te::gm::GeometryPtr result = te::gm::CreateCollectionFromGeometryVector(geometryVector);
    std::string wktResult = result->toString();

    for (std::size_t i = 0; i < geometryVector.size(); ++i)
    {
      const te::gm::Geometry* geometry = geometryVector.at(i);
      ASSERT_LE(geometry->getNPoints(), maxCoordinates);
    }

    te::common::FreeContents(geometryVector);
  }

  TEST_F(ItsSubDivide, SubdivideByCellTilling)
  {
    std::string wkt = "POLYGON((300 200, 350 200, 400 200, 450 200, 500 200, 500 250, 500 300, 450 300, 450 250, 400 250, 400 300, 450 300, 450 350, 400 350, 350 350, 300 350, 300 300, 300 250, 300 200))";
    std::unique_ptr<te::gm::Geometry> geometryA(te::gm::WKTReader::read(wkt.c_str()));

    te::gm::CellTilling cellTilling(*geometryA->getMBR(), geometryA->getMBR()->getWidth() / 10., geometryA->getMBR()->getHeight() / 10.);
    std::vector<te::gm::GeometrySubdivider::ColumnRow> vecColumnRow;

    te::gm::GeometryVector geometryVector = te::gm::GeometrySubdivider::subdivideByCellTilling(geometryA.get(), cellTilling, vecColumnRow);
    te::gm::GeometryPtr result = te::gm::CreateCollectionFromGeometryVector(geometryVector);
    std::string wktResult = result->toString();

    ASSERT_EQ(geometryVector.size(), 90);

    for (std::size_t i = 0; i < geometryVector.size(); ++i)
    {
      const te::gm::Geometry* geometry = geometryVector.at(i);
    }

    te::common::FreeContents(geometryVector);
  }
}
 
