/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

 This file is part of the TerraLib - a Framework for building GIS enabled applications.

 TerraLib is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, either version 3 of the License,
 or (at your option) any later version.

 TerraLib is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with TerraLib. See COPYING. If not, write to
 TerraLib Team at <terralib-team@terralib.org>.
 */

#include "TestUtils.h"

 // TerraLib
#include <terralib_buildconfig.h>

#include <terralib/common/STLUtils.h>
#include <terralib/core/filesystem/FileSystem.h>
#include <terralib/dataaccess.h>
#include <terralib/datatype.h>
#include <terralib/geometry.h>
#include <terralib/geometry/CommonDataStructures.h>
#include <terralib/memory.h>
#include <terralib/memory/PreparedItemCopy.h>
#include <terralib/vp/Feature.h>

#include <iostream>
#include <fstream>

void te::gtest::GetFeaturesFromShapefile(const std::string& uri, te::vp::FeatureSet& vecFeatures)
{
  std::unique_ptr<te::da::DataSource> dataSource = te::da::DataSourceFactory::make("OGR", uri);
  dataSource->open();

  std::vector<std::string> vecNames = dataSource->getDataSetNames();
  if (vecNames.empty() == true)
  {
    return;
  }

  const std::string& dataSetName = vecNames.at(0);

  std::unique_ptr<te::da::DataSetType> dataSetType = dataSource->getDataSetType(dataSetName);
  std::size_t numProperties = dataSetType->getProperties().size();

  std::unique_ptr<te::da::DataSet> dataSet = dataSource->getDataSet(dataSetName);

  std::size_t geomIdx = te::da::GetFirstSpatialPropertyPos(dataSet.get());

  //get data
  bool isValid = true;

  dataSet->moveBeforeFirst();

  while (dataSet->moveNext())
  {
    std::unique_ptr<te::vp::Feature> feature(te::vp::Feature::createFeatureFromCurrent(dataSet.get()));
    vecFeatures.push_back(feature.release());
  }
}

void te::gtest::SimpleSaveGeometriesToShapefile(const std::string& outputPath, const std::string& dataSetName, const te::gm::GeometryVectorConst& geometryVectpr, const std::vector<DataToSave>& dataToAppend)
{
  if (geometryVectpr.empty() == true)
  {
    return;
  }

  te::gm::GeomType geomTypeId = geometryVectpr.at(0)->getGeomTypeId();
  int srid = geometryVectpr.at(0)->getSRID();

  std::unique_ptr<te::da::DataSetType> outputDataSetType(new te::da::DataSetType(dataSetName));
  outputDataSetType->setDatasetName(dataSetName);

  //id
  te::dt::SimpleProperty* propId = new te::dt::SimpleProperty("id", te::dt::INT32_TYPE, true);
  outputDataSetType->add(propId);

  //geom
  te::gm::GeometryProperty* propGeom = new te::gm::GeometryProperty("geom", 0, geomTypeId, true);
  propGeom->setSRID(srid);
  outputDataSetType->add(propGeom);

  for (std::size_t i = 0; i < dataToAppend.size(); ++i)
  {
    te::dt::Property* prop = new te::dt::StringProperty(dataToAppend.at(i).m_attributeName, te::dt::STRING, true);
    outputDataSetType->add(prop);

  }

  //then we create the features
  te::vp::FeatureSet featureSet;
  for (std::size_t i = 0; i < geometryVectpr.size(); ++i)
  {
    te::vp::Feature* feature = new te::vp::Feature(outputDataSetType.get());
    feature->setInt32("id", static_cast<boost::int32_t>(i));
    feature->setGeometry("geom", static_cast<te::gm::Geometry*>(geometryVectpr.at(i)->clone()));

    for (std::size_t j = 0; j < dataToAppend.size(); ++j)
    {
      feature->setString(dataToAppend.at(j).m_attributeName, dataToAppend.at(j).m_vecValues.at(i));
    }

    featureSet.push_back(feature);
  }

  SimpleSaveGeometriesToShapefile(outputPath, outputDataSetType.get(), featureSet);

  te::common::FreeContents(featureSet);
}

void te::gtest::SimpleSaveGeometriesToShapefile(const std::string& outputPath, te::da::DataSetType* outputDataSetType, const te::vp::FeatureSet& featureSet)
{
  if (outputPath.empty() == true)
  {
    return;
  }

  if (featureSet.empty() == true)
  {
    return;
  }

  std::string outputFileName = outputPath + "/" + outputDataSetType->getDatasetName() + ".shp";
  if (te::core::FileSystem::exists(outputFileName) == true)
  {
    if (te::core::FileSystem::remove(outputFileName) == false)
    {
      return;
    }
  }

  std::unique_ptr<te::gm::Geometry> firstGeometry = featureSet.at(0)->getFirstGeometry();
  std::string geomType = firstGeometry->getGeometryType();
  int srid = firstGeometry->getSRID();

  //create output data

  std::unique_ptr<te::da::DataSource> outDataSource = te::da::DataSourceFactory::make("OGR", outputFileName);
  outDataSource->open();

  std::unique_ptr<te::mem::DataSet> outDataSet(new te::mem::DataSet(outputDataSetType));

  te::mem::PreparedItemCopy copy;

  for (std::size_t i = 0; i < featureSet.size(); ++i)
  {
    const te::vp::Feature* inputFeature = featureSet.at(i);
    std::unique_ptr<te::gm::Geometry> resultGeometry = inputFeature->getFirstGeometry();

    //creates the output dataSet item
    std::unique_ptr<te::vp::Feature> outputFeature(new te::vp::Feature(outputDataSetType));
    copy.copyValues(inputFeature, outputFeature.get());
    
    outDataSet->add(outputFeature.release());
  }

  outDataSet->moveBeforeFirst();

  te::da::Create(outDataSource.get(), outputDataSetType, outDataSet.get());

  outDataSource->close();
}

void te::gtest::CreateShapeFile(const std::string& uri, te::da::DataSetType* dataSetType)
{
  if (te::core::FileSystem::exists(uri) == true)
  {
    if (te::core::FileSystem::remove(uri) == false)
    {
      return;
    }
  }

  std::unique_ptr<te::da::DataSource> outDataSource = te::da::DataSourceFactory::make("OGR", uri);

  std::map<std::string, std::string> options;
  outDataSource->createDataSet(dataSetType, options);
}

te::gtest::DataInfo te::gtest::GetRedDataInfo()
{
  te::gtest::DataInfo dataInfo(std::string(TERRALIB_DATA_DIR) + "/shape/desmatamento.shp");
  return dataInfo;
}

te::gtest::DataInfo te::gtest::GetBlueDataInfo()
{
  te::gtest::DataInfo dataInfo(std::string(TERRALIB_DATA_DIR) + "/shape/municipios_desmatamento.shp");
  return dataInfo;
}


std::string te::gtest::GetOutputDir()
{
  std::string outputPath = te::core::FileSystem::tempDirectoryPath() + "/terralib/vp/test_results";
  outputPath = te::core::FileSystem::adjustPathToGenericForm(outputPath);

  return outputPath;
}

std::string te::gtest::GetWktFromFile(const std::string& fileName)
{
  std::string completeFileName = std::string(TERRALIB_DATA_DIR) + "/" + fileName;

  std::string wkt;
  std::string line;
  std::ifstream myfile(completeFileName);
  if (myfile.is_open())
  {
    while (getline(myfile, line))
    {
      wkt += line;
    }
    myfile.close();
  }

  return wkt;
}