/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/attributefill/RasterToVectorDialog.cpp

  \brief Raster to vector attributefill dialog.
*/

// TerraLib
#include "../../core/filesystem/FileSystem.h"
#include "../../core/logger/Logger.h"
#include "../../core/translator/Translator.h"
#include "../../common/progress/ProgressManager.h"
#include "../../common/StringUtils.h"
#include "../../dataaccess/dataset/DataSetTypeConverter.h"
#include "../../dataaccess/datasource/DataSourceInfoManager.h"
#include "../../dataaccess/datasource/DataSourceFactory.h"
#include "../../dataaccess/datasource/DataSourceManager.h"
#include "../../dataaccess/utils/Utils.h"
#include "../../datatype/Enums.h"
#include "../../geometry/GeometryProperty.h"
#include "../../maptools/AbstractLayer.h"
#include "../../qt/widgets/datasource/selector/DataSourceSelectorDialog.h"
#include "../../qt/widgets/layer/utils/DataSet2Layer.h"
#include "../../qt/widgets/progress/ProgressViewerDialog.h"
#include "../../qt/widgets/Utils.h"
#include "../../qt/widgets/utils/FileDialog.h"
#include "../../raster/Band.h"
#include "../../raster/BandProperty.h"
#include "../../raster/Grid.h"
#include "../../raster/RasterProperty.h"
#include "../Config.h"
#include "../Exception.h"
#include "../RasterToVector.h"
#include "RasterToVectorDialog.h"
#include "ui/ui_RasterToVectorDialogForm.h"

// Qt
#include <QFileDialog>
#include <QIntValidator>
#include <QMessageBox>

// Boost
#include <boost/filesystem.hpp>
#include <boost/algorithm/string.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/uuid/random_generator.hpp>
#include <boost/uuid/uuid_io.hpp>
#include <boost/regex.hpp>

te::attributefill::RasterToVectorDialog::RasterToVectorDialog(QWidget* parent, Qt::WindowFlags f)
  : QDialog(parent, f),
    m_ui(new Ui::RasterToVectorDialogForm),
    m_layers(std::list<te::map::AbstractLayerPtr>())
{
  // add controls
  m_ui->setupUi(this);

  m_ui->m_noDataLineEdit->setValidator( new QIntValidator);

  // add icons
  m_ui->m_imgLabel->setPixmap(QIcon::fromTheme("raster-vector-hint").pixmap(112,48));
  m_ui->m_targetDatasourceToolButton->setIcon(QIcon::fromTheme("datasource"));

  connect(m_ui->m_inRasterComboBox, SIGNAL(currentIndexChanged(int)), this, SLOT(onRasterComboBoxChanged(int)));
  connect(m_ui->m_inVectorComboBox, SIGNAL(currentIndexChanged(int)), this, SLOT(onVectorComboBoxChanged(int)));

  connect(m_ui->m_noDataCheckBox, SIGNAL(toggled(bool)), m_ui->m_noDataLineEdit, SLOT(setEnabled(bool)));

  connect(m_ui->m_targetDatasourceToolButton, SIGNAL(pressed()), this, SLOT(onTargetDatasourceToolButtonPressed()));
  connect(m_ui->m_targetFileToolButton, SIGNAL(pressed()), this,  SLOT(onTargetFileToolButtonPressed()));

  connect(m_ui->m_okPushButton, SIGNAL(clicked()), this, SLOT(onOkPushButtonClicked()));
  connect(m_ui->m_cancelPushButton, SIGNAL(clicked()), this, SLOT(onCancelPushButtonClicked()));

  m_ui->m_helpPushButton->setNameSpace("dpi.inpe.br.plugins"); 
  m_ui->m_helpPushButton->setPageReference("plugins/attributefill/attrfill_raster_to_vector.html");

}

te::attributefill::RasterToVectorDialog::~RasterToVectorDialog() = default;

void te::attributefill::RasterToVectorDialog::setLayers(std::list<te::map::AbstractLayerPtr> layers)
{
  m_layers = layers;
  
  std::list<te::map::AbstractLayerPtr>::iterator it = m_layers.begin();

  while(it != m_layers.end())
  {
    std::unique_ptr<te::da::DataSetType> dsType = it->get()->getSchema();
    if (dsType->hasRaster())
    {
      m_ui->m_inRasterComboBox->addItem(QString(it->get()->getTitle().c_str()), QVariant(it->get()->getId().c_str()));
    }
    if (dsType->hasGeom())
    {
      te::gm::GeometryProperty* geomProp = te::da::GetFirstGeomProperty(dsType.get());
      if (geomProp->getGeometryType() == te::gm::PolygonType ||
          geomProp->getGeometryType() == te::gm::MultiPolygonType ||
          geomProp->getGeometryType() == te::gm::PointType ||
          geomProp->getGeometryType() == te::gm::MultiPointType)
      {
        m_ui->m_inVectorComboBox->addItem(QString(it->get()->getTitle().c_str()), QVariant(it->get()->getId().c_str()));
      }
    }
    ++it;
  }
}

te::map::AbstractLayerPtr te::attributefill::RasterToVectorDialog::getLayer()
{
  return m_outLayer;
}

std::vector<unsigned int> te::attributefill::RasterToVectorDialog::getSelectedBands()
{
  std::vector<unsigned int> vecBands;
  
  for(int i = 0; i < m_ui->m_bandsListWidget->count(); ++i)
  {
    if(m_ui->m_bandsListWidget->isItemSelected(m_ui->m_bandsListWidget->item(i)))
    {
      vecBands.push_back(i);
    }
  }
  return vecBands;
}

std::vector<te::stat::StatisticalSummary> te::attributefill::RasterToVectorDialog::getSelectedStatistics()
{
  std::vector<te::stat::StatisticalSummary> vecStatistics;
  
  for(int i = 0; i < m_ui->m_statisticsListWidget->count(); ++i)
  {
    if(m_ui->m_statisticsListWidget->isItemSelected(m_ui->m_statisticsListWidget->item(i)))
    {
      switch(i)
      {
        case 0:
          vecStatistics.push_back(te::stat::MIN_VALUE);
          break;
        case 1:
          vecStatistics.push_back(te::stat::MAX_VALUE);
          break;
        case 2:
          vecStatistics.push_back(te::stat::MEAN);
          break;
        case 3:
          vecStatistics.push_back(te::stat::SUM);
          break;
        case 4:
          vecStatistics.push_back(te::stat::COUNT);
          break;
        case 5:
          vecStatistics.push_back(te::stat::VALID_COUNT);
          break;
        case 6:
          vecStatistics.push_back(te::stat::STANDARD_DEVIATION);
          break;
        case 7:
          vecStatistics.push_back(te::stat::VARIANCE);
          break;
        case 8:
          vecStatistics.push_back(te::stat::SKEWNESS);
          break;
        case 9:
          vecStatistics.push_back(te::stat::KURTOSIS);
          break;
        case 10:
          vecStatistics.push_back(te::stat::AMPLITUDE);
          break;
        case 11:
          vecStatistics.push_back(te::stat::MEDIAN);
          break;
        case 12:
          vecStatistics.push_back(te::stat::VAR_COEFF);
          break;
        case 13:
          vecStatistics.push_back(te::stat::MODE);
          break;
        case 14:
          vecStatistics.push_back(te::stat::PERCENT_EACH_CLASS_BY_AREA);
          break;
        case 15:
          vecStatistics.push_back(te::stat::TOTAL_AREA_BY_CLASS);
          break;
        default:
          continue;
      }
    }
  }
  return vecStatistics;
}

bool te::attributefill::RasterToVectorDialog::getValueOption()
{
  return !m_ui->m_statisticsListWidget->selectedItems().isEmpty();
}

void te::attributefill::RasterToVectorDialog::onRasterComboBoxChanged(int index)
{
  std::list<te::map::AbstractLayerPtr>::iterator it = m_layers.begin();
  
  std::string layerID = m_ui->m_inRasterComboBox->itemData(index, Qt::UserRole).toString().toUtf8().data();

  while(it != m_layers.end())
  {
    if(layerID == it->get()->getId())
      m_rasterLayer = it->get();
    
    ++it;
  }

  m_ui->m_bandsListWidget->clear();

  std::unique_ptr<te::da::DataSetType> dsType = m_rasterLayer->getSchema();
  std::unique_ptr<te::da::DataSet> dsRaster = m_rasterLayer->getData();

  te::rst::RasterProperty* rasterProp = te::da::GetFirstRasterProperty(dsType.get());

  std::unique_ptr<te::rst::Raster> raster = dsRaster->getRaster(rasterProp->getName());
  std::size_t n_bands = raster->getNumberOfBands();

  double noDataValue = raster->getBand(0)->getProperty()->m_noDataValue;

  if(noDataValue == std::numeric_limits<double>::max())
  {
    m_ui->m_noDataLineEdit->setText(
        boost::lexical_cast<std::string>(noDataValue).c_str());
  }

  for(std::size_t b = 0; b < n_bands; ++b)
    m_ui->m_bandsListWidget->addItem(boost::lexical_cast<std::string>(b).c_str());

}

void te::attributefill::RasterToVectorDialog::onVectorComboBoxChanged(int index)
{
  std::list<te::map::AbstractLayerPtr>::iterator it = m_layers.begin();
  
  std::string layerID = m_ui->m_inVectorComboBox->itemData(index, Qt::UserRole).toString().toUtf8().data();

  while(it != m_layers.end())
  {
    if(layerID == it->get()->getId())
    {
      m_vectorLayer = it->get();
      std::unique_ptr<te::da::DataSetType> dsType = it->get()->getSchema();
        
      m_ui->m_statisticsListWidget->clear();

      te::gm::GeometryProperty* geomProp = te::da::GetFirstGeomProperty(dsType.get());

      if (geomProp->getGeometryType() == te::gm::PolygonType || geomProp->getGeometryType() == te::gm::MultiPolygonType)
      {
        m_ui->m_statisticsListWidget->addItem("Minimum value");
        m_ui->m_statisticsListWidget->addItem("Maximum value");
        m_ui->m_statisticsListWidget->addItem("Mean");
        m_ui->m_statisticsListWidget->addItem("Sum of values");
        m_ui->m_statisticsListWidget->addItem("Total number of values");
        m_ui->m_statisticsListWidget->addItem("Total not null values");
        m_ui->m_statisticsListWidget->addItem("Standard deviation");
        m_ui->m_statisticsListWidget->addItem("Variance");
        m_ui->m_statisticsListWidget->addItem("Skewness");
        m_ui->m_statisticsListWidget->addItem("Kurtosis");
        m_ui->m_statisticsListWidget->addItem("Amplitude");
        m_ui->m_statisticsListWidget->addItem("Median");
        m_ui->m_statisticsListWidget->addItem("Coefficient variation");
        m_ui->m_statisticsListWidget->addItem("Mode");
        m_ui->m_statisticsListWidget->addItem("Percent of each class by area");
        m_ui->m_statisticsListWidget->addItem("Total area by class");

        m_isStatistical = true;
        m_ui->m_textureCheckBox->setEnabled(true);
      }
      else
      {
        m_ui->m_statisticsListWidget->addItem("Value");

        m_isStatistical = false;
        m_ui->m_textureCheckBox->setChecked(false);
        m_ui->m_textureCheckBox->setEnabled(false);
      }
    }
    ++it;
  }
}

void te::attributefill::RasterToVectorDialog::onTargetDatasourceToolButtonPressed()
{
  m_ui->m_newLayerNameLineEdit->clear();
  m_ui->m_newLayerNameLineEdit->setEnabled(true);
  te::qt::widgets::DataSourceSelectorDialog dlg(this);
  dlg.exec();

  std::list<te::da::DataSourceInfoPtr> dsPtrList = dlg.getSelecteds();

  if(dsPtrList.empty())
    return;

  std::list<te::da::DataSourceInfoPtr>::iterator it = dsPtrList.begin();

  if(!te::core::FileSystem::isDirectory(it->get()->getTitle()))
  {
    QMessageBox::warning(this, tr("File information"), tr("Select a datasource folder."));
    return;
  }

  bool isValidaty = false;

  std::map<QString, QString> mappedExtensions;

  std::vector< std::string > tokens;
  te::common::Tokenize( te::qt::widgets::GetDiskVectorFileSelFilter( true ).toUtf8().data(), tokens, ";;" );
  for(auto extension : tokens)
  {
    boost::regex expression("[.]+[a-z0-9]+");
    boost::smatch extMatched;
    if(boost::regex_search(extension, extMatched, expression))
    {
      mappedExtensions[QString::fromStdString(extension)] = QString::fromStdString(extMatched[0]);
    }
  }

  for(auto currentLayerPath : te::core::FileSystem::listDirectory(it->get()->getTitle()))
  {
    std::string layerExt = te::core::FileSystem::extension(currentLayerPath);

    for(auto currentExt : mappedExtensions)
    {
      if(boost::iequals(layerExt, currentExt.second.toUtf8().data()))
         isValidaty = true;
    }

  }

  if(!isValidaty)
  {
    QMessageBox::warning(this, tr("File information"), tr("Select a datasource folder with vectorial files."));
    m_ui->m_repositoryLineEdit->clear();
    return;
  }

  m_ui->m_repositoryLineEdit->setText(QString(it->get()->getTitle().c_str()));

  m_outputDatasource = *it;
  
  m_toFile = false;
}

void te::attributefill::RasterToVectorDialog::onTargetFileToolButtonPressed()
{
  m_ui->m_newLayerNameLineEdit->clear();
  m_ui->m_repositoryLineEdit->clear();
  
  te::qt::widgets::FileDialog fileDialog(this, te::qt::widgets::FileDialog::VECTOR);

  try {
    fileDialog.exec();
  }
  catch (te::common::Exception& ex) {
    QMessageBox::warning(this, tr("File information"), ex.what());
    return;
  }

  m_ui->m_repositoryLineEdit->setText(fileDialog.getPath().c_str());
  m_ui->m_newLayerNameLineEdit->setText(fileDialog.getFileName().c_str());
  
  m_toFile = true;
  m_ui->m_newLayerNameLineEdit->setEnabled(false);
}

void te::attributefill::RasterToVectorDialog::onOkPushButtonClicked()
{
  if(m_ui->m_inRasterComboBox->count() == 0)
  {
    QMessageBox::information(this, tr("Raster to Vector"), tr("Select an input raster layer."));
    return;
  }

  if(m_ui->m_inVectorComboBox->count() == 0)
  {
    QMessageBox::information(this, tr("Raster to Vector"), tr("Select an input vector layer."));
    return;
  }

  std::unique_ptr<te::da::DataSet> rasterDataSet = m_rasterLayer->getData();

  std::size_t rpos = te::da::GetFirstPropertyPos(rasterDataSet.get(), te::dt::RASTER_TYPE);

  std::unique_ptr<te::rst::Raster> inputRst = rasterDataSet->getRaster(rpos);

  std::string sridMessage;
  bool isValidSRIDs = te::qt::widgets::isValidSRIDs(m_rasterLayer->getSRID(), m_vectorLayer->getSRID(), sridMessage);

  if (!isValidSRIDs)
  {
    QMessageBox::information(this, tr("Raster to Vector"), sridMessage.c_str());
    return;
  }

  inputRst->getGrid()->setSRID(m_rasterLayer->getSRID());

  if(m_ui->m_noDataCheckBox->isChecked())
  {
    if(!m_ui->m_noDataLineEdit->text().isEmpty())
    {
      double noDataValue = m_ui->m_noDataLineEdit->text().toDouble();
      te::rst::BandProperty* bProp = inputRst->getBand(0)->getProperty();
      bProp->m_noDataValue = noDataValue;
    }
  }

  te::map::DataSetLayer* dsVectorLayer = dynamic_cast<te::map::DataSetLayer*>(m_vectorLayer.get());

  if(!dsVectorLayer)
  {
    QMessageBox::information(this, tr("Raster to Vector"), tr("Can not execute this operation on this type of layer."));
    return;
  }

  const te::da::ObjectIdSet* oidSet = nullptr;

  if(m_ui->m_onlySelectedCheckBox->isChecked())
  {
    oidSet = dsVectorLayer->getSelected();
    if(!oidSet)
    {
      QMessageBox::information(this, tr("Raster to Vector"), tr("Select the layer objects to perform the raster to vector operation."));
      return;
    }
  }

  te::da::DataSourcePtr inVectorDataSource = te::da::GetDataSource(dsVectorLayer->getDataSourceId(), true);
  if (!inVectorDataSource.get())
  {
    QMessageBox::information(this, tr("Raster to Vector"), tr("The selected vector data source can not be accessed."));
    return;
  }

  std::vector<unsigned int> vecBands = getSelectedBands();
  if(vecBands.empty())
  {
    QMessageBox::information(this, tr("Raster to Vector"), tr("Select at least one band."));
    return;
  }

  std::vector<te::stat::StatisticalSummary> vecStatistics;
  
  if (m_isStatistical)
    vecStatistics = getSelectedStatistics();

  m_texture = m_ui->m_textureCheckBox->isChecked();

  bool isValueOptionSelected = getValueOption();
  
  if (isValueOptionSelected == false && vecStatistics.empty() && m_texture == false)
  {
    QMessageBox::information(this, tr("Raster to Vector"), tr("Select at least one statistic operation or select the texture checkbox."));
    return;
  }

  if(m_ui->m_repositoryLineEdit->text().isEmpty())
  {
    QMessageBox::information(this, tr("Raster to Vector"), tr("Define a repository for the result."));
    return;
  }

  if(m_ui->m_newLayerNameLineEdit->text().isEmpty())
  {
    QMessageBox::information(this, tr("Raster to Vector"), tr("Define a name for the resulting layer."));
    return;
  }
  
  std::string outputdataset = m_ui->m_newLayerNameLineEdit->text().toUtf8().data();

  //progress
  te::qt::widgets::ProgressViewerDialog v(this);

  try
  {
    bool res;

    if(m_toFile)
    {
      boost::filesystem::path uri(m_ui->m_repositoryLineEdit->text().toUtf8().data());
      
      if (te::core::FileSystem::exists(uri.string()))
      {
        QMessageBox::information(this, tr("Raster to Vector"), tr("Output file already exists. Remove it or select a new name and try again."));
        return;
      }

      std::size_t idx = outputdataset.find(".");
      if (idx != std::string::npos)
        outputdataset=outputdataset.substr(0,idx);

      std::string dsinfo("file://" + uri.string());

      te::da::DataSourcePtr dsOGR(te::da::DataSourceFactory::make("OGR", dsinfo).release());
      dsOGR->open();
      if (dsOGR->dataSetExists(outputdataset))
      {
        QMessageBox::information(this, tr("Raster to Vector"), tr("There is already a dataset with the requested name in the output data source. Remove it or select a new name and try again."));
        return;
      }

      std::unique_ptr<te::da::DataSetTypeConverter> converterVector(
            new te::da::DataSetTypeConverter(m_vectorLayer->getSchema().get(),
                                             inVectorDataSource->getCapabilities(),
                                             inVectorDataSource->getEncoding()));

      te::da::AssociateDataSetTypeConverterSRID(converterVector.get(), m_vectorLayer->getSRID());

      this->setCursor(Qt::WaitCursor);

      te::attributefill::RasterToVector* rst2vec = new te::attributefill::RasterToVector();
      rst2vec->setInput(inputRst.get(),
                        inVectorDataSource, 
                        dsVectorLayer->getDataSetName(),
                        converterVector.release(),
                        oidSet);

      rst2vec->setParams(vecBands,
                         vecStatistics,
                         m_ui->m_boxPixelRadioButton->isChecked(),
                         m_texture);

      rst2vec->setOutput(dsOGR, outputdataset);
      
      if (!rst2vec->paramsAreValid())
        res = false;
      else
        res = rst2vec->run();

      if (!res)
      {
        this->setCursor(Qt::ArrowCursor);
        dsOGR->close();
        QMessageBox::information(this, tr("Raster to Vector"), tr("Error: could not generate the operation."));
        reject();
      }
      dsOGR->close();

      delete rst2vec;

      // let's include the new datasource in the managers
      boost::uuids::basic_random_generator<boost::mt19937> gen;
      boost::uuids::uuid u = gen();
      std::string id_ds = boost::uuids::to_string(u);
      
      te::da::DataSourceInfoPtr ds(new te::da::DataSourceInfo);
      ds->setConnInfo(dsinfo);
      ds->setTitle(uri.stem().string());
      ds->setAccessDriver("OGR");
      ds->setType("OGR");
      ds->setDescription(uri.string());
      ds->setId(id_ds);
      
      te::da::DataSourcePtr newds = te::da::DataSourceManager::getInstance().get(id_ds, "OGR", ds->getConnInfo());
      newds->open();
      te::da::DataSourceInfoManager::getInstance().add(ds);
      m_outputDatasource = ds;
    }
    else
    {
      te::da::DataSourcePtr aux = te::da::GetDataSource(m_outputDatasource->getId());
      if (!aux)
      {
        QMessageBox::information(this, tr("Raster to Vector"), tr("The selected output datasource can not be accessed."));
        return;
      }

      if (aux->dataSetExists(outputdataset))
      {
        QMessageBox::information(this, tr("Raster to Vector"), tr("Dataset already exists. Remove it or select a new name and try again."));
        return;
      }
      this->setCursor(Qt::WaitCursor);

      std::unique_ptr<te::da::DataSetTypeConverter> converterVector(
            new te::da::DataSetTypeConverter(m_vectorLayer->getSchema().get(),
                                             inVectorDataSource->getCapabilities(),
                                             inVectorDataSource->getEncoding()));

      te::da::AssociateDataSetTypeConverterSRID(converterVector.get(), m_vectorLayer->getSRID());

      te::attributefill::RasterToVector* rst2vec = new te::attributefill::RasterToVector();
      
      rst2vec->setInput(inputRst.get(),
                        inVectorDataSource,
                        dsVectorLayer->getDataSetName(),
                        converterVector.release(),
                        oidSet);

      rst2vec->setParams(vecBands,
                         vecStatistics,
                         m_ui->m_boxPixelRadioButton->isChecked(),
                         m_texture);

      rst2vec->setOutput(aux, outputdataset);

      if (!rst2vec->paramsAreValid())
        res = false;
      else
        res = rst2vec->run();

      delete rst2vec;

      if (!res)
      {
        this->setCursor(Qt::ArrowCursor);
        QMessageBox::information(this, tr("Raster to Vector"), tr("Error: could not generate the operation."));

        reject();
      }
    }

    // creating a layer for the result
    te::da::DataSourcePtr outDataSource = te::da::GetDataSource(m_outputDatasource->getId());
    
    te::qt::widgets::DataSet2Layer converter(m_outputDatasource->getId());
      
    te::da::DataSetTypePtr dt(outDataSource->getDataSetType(outputdataset).release());
    m_outLayer = converter(dt);
  }
  catch(const std::exception& e)
  {
    this->setCursor(Qt::ArrowCursor);

    QMessageBox::information(this, tr("Raster to Vector"), e.what());

#ifdef TERRALIB_LOGGER_ENABLED
    TE_CORE_LOG_DEBUG("attributefill", e.what());
#endif // TERRALIB_LOGGER_ENABLED

    return;
  }

  this->setCursor(Qt::ArrowCursor);
  accept();
}

void te::attributefill::RasterToVectorDialog::onCancelPushButtonClicked()
{
  reject();
}
