/*
 * File: Random.i
 *
 * Random specific to Lua
 */

%{
#include <boost/random.hpp>
#include <boost/math/distributions/beta.hpp>

namespace te
{
  namespace random
  {
    class MersenneTwister
    {
      public:
        MersenneTwister(int seed) : mt(seed) { }
        virtual ~MersenneTwister() {}
        MersenneTwister(const MersenneTwister& other) : mt(other.mt) {}

        int operator()()
        {
            return mt();
        }

        boost::mt19937& getEngine()
        {
          return mt;
        }

      private:
	    boost::mt19937 mt;
    };

    class NormalDistribution
    {
      public:
        NormalDistribution(MersenneTwister& e, double mean = 0.0, double stddev = 1.0)
          : d(e.getEngine(), boost::normal_distribution<>(mean, stddev)) { }
        virtual ~NormalDistribution() {}
        NormalDistribution(const NormalDistribution& other) : d(other.d) { }

        double operator()()
        {
          return d();
        }

      private:
		boost::variate_generator<boost::mt19937, boost::normal_distribution<> > d;
    };

    class LogNormalDistribution
    {
      public:
        LogNormalDistribution(MersenneTwister& e, double mean = 0.0, double stddev = 1.0)
          : d(e.getEngine(), boost::lognormal_distribution<>(mean, stddev)) { }
        virtual ~LogNormalDistribution() {}
        LogNormalDistribution(const LogNormalDistribution& other) : d(other.d) { }

        double operator()()
        {
          return d();
        }

      private:
        boost::variate_generator<boost::mt19937, boost::lognormal_distribution<> > d;
    };

    class DiscreteDistribution
    {
      public:
        DiscreteDistribution(MersenneTwister& e, const std::vector<double>& weights)
          : d(e.getEngine(), boost::random::discrete_distribution<>(weights.begin(), weights.end()))
        {
          // std::size_t i(0);
          // std::discrete_distribution<> d(weights.size(), 0.0, 0.0,
            // [&weights,&i](double)
            // {
              // auto w = weights[i];
              // ++i;
              // return w;
            // });

          // this->d.param(d.param());
        }
        virtual ~DiscreteDistribution() {}
        DiscreteDistribution(const DiscreteDistribution& other) : d(other.d) { }

        int operator()()
        {
          return d();
        }

      private:
        boost::variate_generator<boost::mt19937, boost::random::discrete_distribution<> > d;
    };

    class UniformIntDistribution
    {
      public:
        UniformIntDistribution(MersenneTwister& e, int min = 0, int max = std::numeric_limits<int>::max())
          : d(e.getEngine(), boost::random::uniform_int_distribution<>(min, max)) { }
        virtual ~UniformIntDistribution() {}
        UniformIntDistribution(const UniformIntDistribution& other) : d(other.d) { }

        int operator()()
        {
            return d();
        }

        private:
           boost::variate_generator<boost::mt19937, boost::random::uniform_int_distribution<> > d;
    };

    class UniformRealDistribution
    {
      public:
        UniformRealDistribution(MersenneTwister& e, double min = 0.0, double max = 1.0)
          : d(e.getEngine(), boost::random::uniform_real_distribution<>(min, max)) { }
        virtual ~UniformRealDistribution() {}
        UniformRealDistribution(const UniformRealDistribution& other) : d(other.d) { }

        double operator()()
        {
          return d();
        }

      private:
		boost::variate_generator<boost::mt19937, boost::random::uniform_real_distribution<> > d;
    };

    class WeibullDistribution
    {
      public:
        WeibullDistribution(MersenneTwister& e, double a = 1.0, double b = 1.0)
          : d(e.getEngine(), boost::random::weibull_distribution<>(a, b)) { }
        virtual ~WeibullDistribution() {}
        WeibullDistribution(const WeibullDistribution& other) : d(other.d) { }

        double operator()()
        {
          return d();
        }

      private:
		boost::variate_generator<boost::mt19937, boost::random::weibull_distribution<> > d;
    };

    class BetaDistribution
    {
      public:
        BetaDistribution(double alpha = 1.0, double beta = 1.0)
          : d(boost::math::beta_distribution<>(alpha, beta)) { }
        virtual ~BetaDistribution() {}
        BetaDistribution(const BetaDistribution& other) : d(other.d) { }

        double operator()(double rand)
        {
          return boost::math::quantile(d, rand);
        }

      private:
		boost::math::beta_distribution<> d;
    };

    class BernoulliDistribution
    {
      public:
        BernoulliDistribution(MersenneTwister& e, double p = 0.5)
          : d(e.getEngine(), boost::bernoulli_distribution<>(p)) { }
        virtual ~BernoulliDistribution() {}
        BernoulliDistribution(const BernoulliDistribution& other) : d(other.d) { }

        bool operator()()
        {
          return d();
        }

      private:
		boost::variate_generator<boost::mt19937, boost::bernoulli_distribution<> > d;
    };

    class PoissonDistribution
    {
      public:
        PoissonDistribution(MersenneTwister& e, double mean = 1.0)
          : d(e.getEngine(), boost::poisson_distribution<>(mean)) { }
        virtual ~PoissonDistribution() {}
        PoissonDistribution(const PoissonDistribution& other) : d(other.d) { }

        int operator()()
        {
          return d();
        }

      private:
		boost::variate_generator<boost::mt19937, boost::poisson_distribution<> > d;
    };

    class ExponentialDistribution
    {
      public:
        ExponentialDistribution(MersenneTwister& e, double lambda = 1.0)
          : d(e.getEngine(), boost::exponential_distribution<>(lambda)) { }
        virtual ~ExponentialDistribution() {}
        ExponentialDistribution(const ExponentialDistribution& other) : d(other.d) { }

        double operator()()
        {
          return d();
        }

      private:
		boost::variate_generator<boost::mt19937, boost::exponential_distribution<> > d;
    };

  } // namespace random
} // namespace te
%}

%nspace te::random::MersenneTwister;
%nspace te::random::NormalDistribution;
%nspace te::random::LogNormalDistribution;
%nspace te::random::DiscreteDistribution;
%nspace te::random::UniformIntDistribution;
%nspace te::random::UniformRealDistribution;
%nspace te::random::WeibullDistribution;
%nspace te::random::BetaDistribution;
%nspace te::random::BernoulliDistribution;
%nspace te::random::PoissonDistribution;
%nspace te::random::ExponentialDistribution;

namespace te
{
  namespace random
  {
    class MersenneTwister
    {
      public:
        MersenneTwister(int seed) : mt(seed);
		int operator()();
    };

    class NormalDistribution
    {
      public:
        NormalDistribution(MersenneTwister& e, double mean = 0.0, double stddev = 1.0);
        double operator()();
    };

    class LogNormalDistribution
    {
      public:
        LogNormalDistribution(MersenneTwister& e, double mean = 0.0, double stddev = 1.0);
        double operator()();
    };

    class DiscreteDistribution
    {
      public:
        DiscreteDistribution(MersenneTwister& e, const std::vector<double>& weights);
        int operator()();
    };

    class UniformIntDistribution
    {
      public:
        UniformIntDistribution(MersenneTwister& e, int a = 0, int b = std::numeric_limits<int>::max());
        int operator()();
    };

    class UniformRealDistribution
    {
      public:
        UniformRealDistribution(MersenneTwister& e, double a = 0.0, double b = 1.0);
        double operator()();
    };

    class WeibullDistribution
    {
      public:
        WeibullDistribution(MersenneTwister& e, double a = 1.0, double b = 1.0);
        double operator()();
    };
	
    class BetaDistribution
    {
      public:
        BetaDistribution(double alpha = 1.0, double beta = 1.0);
        double operator()(double rand);
    };	

    class BernoulliDistribution
    {
      public:
        BernoulliDistribution(MersenneTwister& e, double p = 0.5);
        bool operator()();
    };

    class PoissonDistribution
    {
      public:
        PoissonDistribution(MersenneTwister& e, double mean = 1.0);
        int operator()();
    };

    class ExponentialDistribution
    {
      public:
        ExponentialDistribution(MersenneTwister& e, double lambda = 1.0);
        double operator()();
    };
  }
}