/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/classification/ISOData.cpp

  \brief ISOData (Maximum a Posteriori) strategy for classification.
*/

#include "ISOData.h"
#include "../common/progress/TaskProgress.h"

#include <set>

namespace te
{
  namespace cl
  {
    ISOData::Parameters::Parameters()
    {
      reset();
    }

    ISOData::Parameters::~Parameters() = default;

    const ISOData::Parameters& ISOData::Parameters::operator=(
      const ISOData::Parameters& rhs)
    {
      reset();

      m_clustersNumber = rhs.m_clustersNumber;
      m_maxIterationsNumber = rhs.m_maxIterationsNumber;
      m_maxMergesPerIteration = rhs.m_maxMergesPerIteration;
      m_maxSplitsPerIteration = rhs.m_maxSplitsPerIteration;
      m_minClusterSizePercent = rhs.m_minClusterSizePercent;
      m_stdDevSplitThreshold = rhs.m_stdDevSplitThreshold;
      m_mergeDistThreshold = rhs.m_mergeDistThreshold;

      return *this;
    }

    void ISOData::Parameters::reset()
    {
      m_clustersNumber = 2;
      m_maxIterationsNumber = 0;
      m_maxMergesPerIteration = 1;
      m_maxSplitsPerIteration = 1;
      m_minClusterSizePercent = 0;
      m_stdDevSplitThreshold = 0.05;
      m_mergeDistThreshold = 0.1;
    }

    te::common::AbstractParameters* ISOData::Parameters::clone() const
    {
      return new ISOData::Parameters(*this);
    }

    //  ------------------------------------------------------------------------------------------------------------------------------------------

    ISOData::ClusterData::ClusterData()
    {
    }

    ISOData::ClusterData::~ClusterData() = default;

    //  ------------------------------------------------------------------------------------------------------------------------------------------

    ISOData::ISOData()
    {
      reset();
    }

    ISOData::~ISOData() = default;

    bool ISOData::initialize(const ISOData::Parameters& params)
    {
      reset();

      if( params.m_clustersNumber < 2 ) return false;
      
      if(
          ( params.m_minClusterSizePercent < 0 )
          ||
          ( params.m_minClusterSizePercent > 100 )
        )
      {
        return false;
      }

      if( params.m_stdDevSplitThreshold < 0 ) return false;

      if( params.m_mergeDistThreshold < 0 ) return false;

      m_parameters = params;

      m_isInitialized = true;

      return true;
    }

    void ISOData::reset()
    {
      m_isInitialized = false;
      m_parameters.reset();
    }

    bool ISOData::classify(const InputAdaptor<double>& input,
                       const std::vector<unsigned int>& attributesIndices,
                       const std::vector<double>& inputNoDataValues,
                       InputOutputAdaptor<ClusterData::ClusterLabelT>& output,
                       const unsigned int outputIndex,
                       const unsigned int outputNoDataValue,
                       const bool enableProgressInterface,
                       std::vector< ClusterData >* clustersDataPtr )
    {
      if( !m_isInitialized ) return false;
      if( input.getElementsCount() != output.getElementsCount() ) return false;
      if( input.getElementsCount() < m_parameters.m_clustersNumber ) return false;
      if( outputIndex >= output.getFeaturesCount() ) return false;
      if( inputNoDataValues.size() != attributesIndices.size() ) return false;
      for( std::size_t attributesIndicesIdx = 0 ; attributesIndicesIdx <
        attributesIndices.size() ; ++attributesIndicesIdx  )
      {
        if( attributesIndices[ attributesIndicesIdx ] >= input.getFeaturesCount() )
        {
          return false;
        }
      }
      
      // Progress interface
      
      const unsigned int maxIterationsNumber = 
        m_parameters.m_maxIterationsNumber ?
            (
              m_parameters.m_maxIterationsNumber 
              + 
              ( m_parameters.m_maxIterationsNumber % 2 )
            )
          :
            ( 2 * m_parameters.m_clustersNumber );      

      std::unique_ptr< te::common::TaskProgress > progressPtr;
      if( enableProgressInterface )
      {
        progressPtr.reset( new te::common::TaskProgress );
        progressPtr->setTotalSteps( maxIterationsNumber + 2 );
        progressPtr->setMessage( "Classifying" );
      }      

      // defining the initial clusters centers

      ClustersContainerT clustersMap;
      ClusterData::ClusterLabelT nextAvaliableClusterLabel = 0;
      std::vector< std::pair< double, double > > inputElementsOffsetsAndGains;
      if( ! initializeClusters( input, attributesIndices, clustersMap,
        nextAvaliableClusterLabel, inputNoDataValues, inputElementsOffsetsAndGains ) )
        return false;
      
      if( enableProgressInterface )
      {
        progressPtr->pulse();
        if( ! progressPtr->isActive() ) return false;
      }      

      // Main classification loop

      const unsigned int clustersNumberMin = m_parameters.m_clustersNumber -
        ( m_parameters.m_clustersNumber / 2 );
      const unsigned int clustersNumberMax = m_parameters.m_clustersNumber +
        ( m_parameters.m_clustersNumber / 2 );


      unsigned int currentIterationIdx = 0;

      while( currentIterationIdx < maxIterationsNumber )
      {
//         std::cout << std::endl << "Iteration " << currentIterationIdx << std::endl;
//         std::cout << std::endl << "Start";
//         printClusters( clustersMap );

        if( enableProgressInterface )
        {
          progressPtr->pulse();
          if( ! progressPtr->isActive() ) return false;
        }

        // Distributing elements among the current clusters

        if( ! classifyElements( input, attributesIndices, inputNoDataValues,
          clustersMap, output, outputIndex, outputNoDataValue,
          inputElementsOffsetsAndGains ) )
        {
          return false;
        }

        // Updating clusters statistical data

        if( ! updateClustersStats( input, attributesIndices, inputNoDataValues,
          clustersMap, output, outputIndex, inputElementsOffsetsAndGains ) )
        {
          return false;
        }

//         std::cout << std::endl << "After update stats";
//         printClusters( clustersMap );

        // Removing too small clusters

        if( !removeSmallClusters( input.getElementsCount(), clustersMap ) )
        {
          return false;
        }

//         std::cout << std::endl << "After remove small";
//         printClusters( clustersMap );

        // split & merge

        if( 
            ( currentIterationIdx % 2 == 0  )
            &&
            ( clustersMap.size() < clustersNumberMax )
          )
        {
          // split
//           const std::size_t clustersMapSizeBeforeSplit = clustersMap.size();
          if( ! splitClusters( 
                  input.getElementsCount(),
                  m_parameters.m_maxSplitsPerIteration ?
                    std::min(
                      m_parameters.m_maxSplitsPerIteration,
                      ( clustersNumberMax - ((unsigned int)clustersMap.size()) )
                    )
                    :
                    ( clustersNumberMax - ((unsigned int)clustersMap.size()) ),
                  nextAvaliableClusterLabel, clustersMap
                )
            )
          {
            return false;
          }
          
//           std::cout << std::endl << "After split";
//           printClusters( clustersMap );          
        }
        else if( 
                 ( currentIterationIdx % 2 != 0  )
                 &&
                 ( clustersMap.size() > clustersNumberMin )
               )
        {  // merge
          if( ! mergeClusters(
                  m_parameters.m_mergeDistThreshold,
                  m_parameters.m_maxMergesPerIteration ?
                    std::min(
                      m_parameters.m_maxMergesPerIteration
                      ,
                      ( ((unsigned int)clustersMap.size()) - clustersNumberMin )
                    )
                    :
                    (
                      ((unsigned int)clustersMap.size())
                      -
                      clustersNumberMin
                    ),
                  clustersMap
                )
            )
          {
            return false;
          }
          
/*          std::cout << std::endl << "After merge";
          printClusters( clustersMap );         */ 
        }

        // Going to the next iteration

        ++currentIterationIdx;
      }
      
      // merging equal clusters
      
      if( ! mergeClusters( 
          0,
          std::numeric_limits< unsigned int >::max(),
          clustersMap ) )
      {
        return false;
      }
            
//         std::cout << std::endl << "After last merge";
//         printClusters( clustersMap );
            
      // Recoding clusters
      
      recodeClustersLabels( clustersMap );

      if( ! classifyElements( input, attributesIndices, inputNoDataValues,
        clustersMap, output, outputIndex, outputNoDataValue,
        inputElementsOffsetsAndGains ) )
      {
        return false;
      }        
      
      if( enableProgressInterface )
      {
        progressPtr->pulse();
        if( ! progressPtr->isActive() ) return false;
      }      
      
      // Output of clusters data

      if( clustersDataPtr )
      {
        clustersDataPtr->clear();

        ClustersContainerT::const_iterator clustersMapIt = clustersMap.begin();
        const ClustersContainerT::const_iterator clustersMapItEnd = clustersMap.end();

        while( clustersMapIt != clustersMapItEnd )
        {
          clustersDataPtr->push_back( clustersMapIt->second );

          ++clustersMapIt;
        }
      }

      return true;
    }

    bool ISOData::initializeClusters( const InputAdaptor<double>& input,
      const std::vector<unsigned int>& attributesIndices,
      ClustersContainerT& clustersMap,
      ClusterData::ClusterLabelT& nextAvaliableClusterLabel,
      const std::vector<double>& inputNoDataValues,
      std::vector< std::pair< double, double > >& inputElementsOffsetsAndGains ) const
    {
      // Globals
      
      const unsigned int elementsCount = input.getElementsCount();
      const std::size_t attributesIndicesSize = attributesIndices.size();      
      
      // read min, max, mean values
      
      std::vector< double > elementsMin( attributesIndicesSize,
        std::numeric_limits< double >::max() );
      std::vector< double > elementsMax( attributesIndicesSize,
        -1.0 * std::numeric_limits< double >::max() );
      std::vector< double > elementsMean( attributesIndicesSize, 0 );
      unsigned int validElementsCount = 0;
      
      {
        unsigned int elementIdx = 0;
        bool isValidSample = false;
        std::size_t attributesIndicesIdx = 0;
        std::vector< double > element( attributesIndicesSize, 0 );
        
        while( elementIdx < elementsCount )
        {
          isValidSample = true;

          for( attributesIndicesIdx = 0 ; attributesIndicesIdx <
            attributesIndicesSize ; ++attributesIndicesIdx )
          {
            input.getFeature( elementIdx, attributesIndices[ attributesIndicesIdx ],
              element[ attributesIndicesIdx ] );

            if(
                element[ attributesIndicesIdx ]
                ==
                inputNoDataValues[ attributesIndicesIdx ]
              )
            {
              isValidSample = false;
              break;
            }
          }

          if( isValidSample )
          {
            for( attributesIndicesIdx = 0 ; attributesIndicesIdx <
              attributesIndicesSize ; ++attributesIndicesIdx )
            {
              elementsMin[ attributesIndicesIdx ] = std::min(
                elementsMin[ attributesIndicesIdx ],
                element[ attributesIndicesIdx ] );
              elementsMax[ attributesIndicesIdx ] = std::max(
                elementsMax[ attributesIndicesIdx ],
                element[ attributesIndicesIdx ] );
              
              elementsMean[ attributesIndicesIdx ] += element[ attributesIndicesIdx ];
            }
            
            ++validElementsCount;
          }

          ++elementIdx;
        }
        
        if( validElementsCount == 0 ) return false;
        
        for( attributesIndicesIdx = 0 ; attributesIndicesIdx <
          attributesIndicesSize ; ++attributesIndicesIdx )
        {
          if( 
              (
                elementsMin[ attributesIndicesIdx ]
                ==
                std::numeric_limits< double >::max()
              )
              ||
              (
                 elementsMax[ attributesIndicesIdx ]
                 ==
                 (-1.0 * std::numeric_limits< double >::max())
              )
            )
          {
            return false;
          }
          
          elementsMean[ attributesIndicesIdx ] /= ((double)validElementsCount);
        }        
      }
      
      // gain and offset

      {
        inputElementsOffsetsAndGains.clear();
        inputElementsOffsetsAndGains.resize( attributesIndicesSize );
        
        for( std::size_t attributesIndicesIdx = 0 ; attributesIndicesIdx <
          attributesIndicesSize ; ++attributesIndicesIdx )
        {
          inputElementsOffsetsAndGains[ attributesIndicesIdx ].second =
            elementsMax[ attributesIndicesIdx ] -
            elementsMin[ attributesIndicesIdx ];
          if( inputElementsOffsetsAndGains[ attributesIndicesIdx ].second == 0 )
            return false;          
          inputElementsOffsetsAndGains[ attributesIndicesIdx ].second = 1.0 /
            inputElementsOffsetsAndGains[ attributesIndicesIdx ].second;

          inputElementsOffsetsAndGains[ attributesIndicesIdx ].first = (-1.0) *
            elementsMin[ attributesIndicesIdx ];
        }
      }
      
      // normalized mean
      
      std::vector< double > elementsNormMean( attributesIndicesSize, 0 );
      
      {
        for( std::size_t elementsMeanIdx = 0 ; elementsMeanIdx < attributesIndicesSize ;
          ++elementsMeanIdx )
        {
          elementsNormMean[ elementsMeanIdx ] = ( elementsMean[ elementsMeanIdx ] + 
            inputElementsOffsetsAndGains[ elementsMeanIdx ].first ) *
            inputElementsOffsetsAndGains[ elementsMeanIdx ].second;
          assert( elementsNormMean[ elementsMeanIdx ] >= 0.0 );
          assert( elementsNormMean[ elementsMeanIdx ] <= 1.0 );
        }
      }
      
      // read normalized stddev
      
      std::vector< double > elementsNormStddev( attributesIndicesSize, 0 );
      
      {
        unsigned int elementIdx = 0;
        bool isValidSample = false;
        std::size_t attributesIndicesIdx = 0;
        std::vector< double > element( attributesIndicesSize, 0 );
        double diff = 0;

        while( elementIdx < elementsCount )
        {
          isValidSample = true;

          for( attributesIndicesIdx = 0 ; attributesIndicesIdx <
            attributesIndicesSize ; ++attributesIndicesIdx )
          {
            input.getFeature( elementIdx, attributesIndices[ attributesIndicesIdx ],
              element[ attributesIndicesIdx ] );

            if(
                element[ attributesIndicesIdx ]
                ==
                inputNoDataValues[ attributesIndicesIdx ]
              )
            {
              isValidSample = false;
              break;
            }
          }

          if( isValidSample )
          {
            for( attributesIndicesIdx = 0 ; attributesIndicesIdx <
              attributesIndicesSize ; ++attributesIndicesIdx )
            {
              diff = 
                ( 
                  ( 
                    element[ attributesIndicesIdx ] 
                    + 
                    inputElementsOffsetsAndGains[ attributesIndicesIdx ].first
                  )
                  *
                  inputElementsOffsetsAndGains[ attributesIndicesIdx ].second 
                )
                -
                elementsNormMean[ attributesIndicesIdx ];
                
              elementsNormStddev[ attributesIndicesIdx ] += ( diff * diff );
            }
          }

          ++elementIdx;
        }
        
        for( attributesIndicesIdx = 0 ; attributesIndicesIdx <
          attributesIndicesSize ; ++attributesIndicesIdx )
        {
          elementsNormStddev[ attributesIndicesIdx ] /= ((double)validElementsCount);
          elementsNormStddev[ attributesIndicesIdx ] = std::sqrt( 
            elementsNormStddev[ attributesIndicesIdx ] );
        }
      }
      
      // initial clusters centers     
      
      {
        ClusterData auxClusterData;
        auxClusterData.m_avgDistToClusterCenter = 0;
        auxClusterData.m_size = 1;
        auxClusterData.m_label = 0;
        auxClusterData.m_means.resize( attributesIndicesSize, 0 );
        auxClusterData.m_sums.resize( attributesIndicesSize, 0 );
        auxClusterData.m_stdDevs.resize( attributesIndicesSize, 0 );
        
        std::size_t attributesIndicesIdx = 0;
        double rangeMin = 0;
        double rangeMax = 0;
        double rangeStep = 0;
        
        for( unsigned int clusterIdx = 0 ; clusterIdx < m_parameters.m_clustersNumber ;
          ++clusterIdx )
        {
          auxClusterData.m_label = nextAvaliableClusterLabel++;
          
          for( attributesIndicesIdx = 0 ; attributesIndicesIdx <
            attributesIndicesSize ; ++attributesIndicesIdx )
          {
            rangeMin = elementsNormMean[ attributesIndicesIdx ] - 
              elementsNormStddev[ attributesIndicesIdx ];
            rangeMax = elementsNormMean[ attributesIndicesIdx ] +
              elementsNormStddev[ attributesIndicesIdx ];
            rangeStep = ( rangeMax - rangeMin ) / 
              ((double)m_parameters.m_clustersNumber + 1);
            
            auxClusterData.m_means[ attributesIndicesIdx ] = rangeMin +
              ( ((double)(clusterIdx + 1)) * rangeStep );
          }          
          
          clustersMap[ auxClusterData.m_label ] = auxClusterData;
        }
      }

      if( clustersMap.size() == m_parameters.m_clustersNumber )
      {
        return true;
      }
      else
      {
        return false;
      }
    }

    bool ISOData::classifyElements( const InputAdaptor<double>& input,
      const std::vector<unsigned int>& attributesIndices,
      const std::vector<double>& inputNoDataValues,
      ClustersContainerT& clustersMap,
      InputOutputAdaptor<ClusterData::ClusterLabelT>& output,
      const unsigned int outputIndex,
      const ClusterData::ClusterLabelT outputNoDataValue,
      const std::vector< std::pair< double, double > >& inputElementsOffsetsAndGains ) const
    {
      // Globals
      
      const std::size_t attributesIndicesSize = attributesIndices.size();
      const unsigned int elementsCount = input.getElementsCount();
      unsigned int attributesIndicesIdx = 0 ;
      unsigned int elementIdx = 0;
      double elementDistance = 0;
      double featureDistance = 0;
      std::vector< double > inputFeatureValues( attributesIndicesSize, 0 );
      ClustersContainerT::iterator clustersMapIt;
      ClustersContainerT::iterator minDistClustersMapIt;        
      const ClustersContainerT::iterator clustersMapItEnd = clustersMap.end();
      double minClusterDist = 0;
      bool isInvalidSample = false;
      
      // Reseting clusters sums and size

      clustersMapIt = clustersMap.begin();

      while( clustersMapIt != clustersMapItEnd )
      {
        std::fill( clustersMapIt->second.m_sums.begin(),
          clustersMapIt->second.m_sums.end(), 0 );
        clustersMapIt->second.m_size = 0;

        ++clustersMapIt;
      }      

      // Classifying elements
      // Updating means, sums and size
      
      for( elementIdx = 0; elementIdx < elementsCount ; ++elementIdx )
      {
        // load sample

        isInvalidSample = false;

        for( attributesIndicesIdx = 0 ; attributesIndicesIdx <
          attributesIndicesSize ; ++attributesIndicesIdx )
        {
          input.getFeature( elementIdx, attributesIndices[ attributesIndicesIdx ],
            inputFeatureValues[ attributesIndicesIdx ] );

          if(
              inputFeatureValues[ attributesIndicesIdx ]
              ==
              inputNoDataValues[ attributesIndicesIdx ]
            )
          {
            isInvalidSample = true;
            break;
          }
        }

        if( isInvalidSample )
        {
          output.setFeature( elementIdx, outputIndex, outputNoDataValue );
        }
        else
        {
          // apply gains and offset

          for( attributesIndicesIdx = 0 ; attributesIndicesIdx <
            attributesIndicesSize ; ++attributesIndicesIdx )
          {
            inputFeatureValues[ attributesIndicesIdx ] +=
              inputElementsOffsetsAndGains[ attributesIndicesIdx ].first;
            inputFeatureValues[ attributesIndicesIdx ] *=
              inputElementsOffsetsAndGains[ attributesIndicesIdx ].second;
            assert( inputFeatureValues[ attributesIndicesIdx ] >= 0.0 );
            assert( inputFeatureValues[ attributesIndicesIdx ] <= 1.0 );
          }

          // locate the closest cluster

          minClusterDist = std::numeric_limits< double >::max();
          clustersMapIt = clustersMap.begin();

          while( clustersMapIt != clustersMapItEnd )
          {
            const std::vector< double >& clusterMmeans =
              clustersMapIt->second.m_means;

            elementDistance = 0;
            for( attributesIndicesIdx = 0 ; attributesIndicesIdx <
              attributesIndicesSize ; ++attributesIndicesIdx )
            {
              featureDistance = inputFeatureValues[ attributesIndicesIdx ] -
                clusterMmeans[ attributesIndicesIdx ];
              elementDistance += ( featureDistance * featureDistance );
            }
            elementDistance = std::sqrt( elementDistance );

            if( elementDistance < minClusterDist )
            {
              minClusterDist = elementDistance;
              minDistClustersMapIt = clustersMapIt;
            }

            ++clustersMapIt;
          }

          if( minClusterDist != std::numeric_limits< double >::max() )
          {
            output.setFeature( elementIdx, outputIndex, 
              minDistClustersMapIt->second.m_label );
            
            ++( minDistClustersMapIt->second.m_size );
            
            for( attributesIndicesIdx = 0 ; attributesIndicesIdx <
              attributesIndicesSize ; ++attributesIndicesIdx )
            {
              minDistClustersMapIt->second.m_sums[ attributesIndicesIdx ] 
                += inputFeatureValues[ attributesIndicesIdx ];
            }            
          }
        }
      }
      
      clustersMapIt = clustersMap.begin();

      while( clustersMapIt != clustersMapItEnd )
      {
        if( clustersMapIt->second.m_size )
        {
          for( attributesIndicesIdx = 0 ; attributesIndicesIdx <
            attributesIndicesSize ; ++attributesIndicesIdx )
          {
            clustersMapIt->second.m_means[ attributesIndicesIdx ] =
              clustersMapIt->second.m_sums[ attributesIndicesIdx ] 
              / ((double)clustersMapIt->second.m_size);
          }
        }

        ++clustersMapIt;
      }         

      return true;
    }

    bool ISOData::removeSmallClusters(
      const unsigned int inputElelementsCount,
      ClustersContainerT& clustersMap ) const
    {
      ClustersContainerT::iterator clustersMapIt = clustersMap.begin();
      ClustersContainerT::iterator auxClustersMapIt;
      const unsigned int minClusterSize = (unsigned int)(
        m_parameters.m_minClusterSizePercent * ((double)inputElelementsCount)
        / 100.0 );

      while( clustersMapIt != clustersMap.end() )
      {
        if( clustersMapIt->second.m_size < minClusterSize )
        {
//           std::cout << std::endl << "Removing small cluster " << 
//             clustersMapIt->first << std::endl;
//             ;
          auxClustersMapIt = clustersMapIt;
          ++auxClustersMapIt;
          clustersMap.erase( clustersMapIt );
          clustersMapIt = auxClustersMapIt;
        }
        else
        {
          ++clustersMapIt;
        }
      }

      return true;
    }

    bool ISOData::updateClustersStats(
      const InputAdaptor<double>& input,
      const std::vector<unsigned int>& attributesIndices,
      const std::vector<double>& inputNoDataValues,
      ClustersContainerT& clustersMap,
      const InputOutputAdaptor<ClusterData::ClusterLabelT>& output,
      const unsigned int outputIndex,
      const std::vector< std::pair< double, double > >& inputElementsOffsetsAndGains ) const
    {
      // Globals

      const std::size_t attributesIndicesSize = attributesIndices.size();
      const unsigned int elementsCount = input.getElementsCount();
      unsigned int attributesIndicesIdx = 0 ;
      unsigned int elementIdx = 0;
      double elementDistance = 0;
      double featureDistance = 0;
      std::vector< double > inputFeatureValues( attributesIndicesSize );
      ClusterData::ClusterLabelT outputFeatureValue = 0;
      ClustersContainerT::iterator
        clustersMapIt;
      bool isInvalidSample = false;

      // Reseting clusters statistical data

      clustersMapIt = clustersMap.begin();

      while( clustersMapIt != clustersMap.end() )
      {
        std::fill( clustersMapIt->second.m_stdDevs.begin(),
          clustersMapIt->second.m_stdDevs.end(), 0 );
        clustersMapIt->second.m_avgDistToClusterCenter = 0;

        ++clustersMapIt;
      }

      // updating Average distance from all samples to the cluster center.
      // updating the standart deviation vectors

      for( elementIdx = 0; elementIdx < elementsCount ; ++elementIdx )
      {
        // load sample

        isInvalidSample = false;

        for( attributesIndicesIdx = 0 ; attributesIndicesIdx <
          attributesIndicesSize ; ++attributesIndicesIdx )
        {
          input.getFeature( elementIdx, attributesIndices[ attributesIndicesIdx ],
            inputFeatureValues[ attributesIndicesIdx ] );

          if(
              inputFeatureValues[ attributesIndicesIdx ]
              ==
              inputNoDataValues[ attributesIndicesIdx ]
            )
          {
            isInvalidSample = true;
            break;
          }
        }

        // update statistical values

        if( !isInvalidSample )
        {
          // apply gains and offset

          for( attributesIndicesIdx = 0 ; attributesIndicesIdx <
            attributesIndicesSize ; ++attributesIndicesIdx )
          {
            inputFeatureValues[ attributesIndicesIdx ] +=
              inputElementsOffsetsAndGains[ attributesIndicesIdx ].first;
            inputFeatureValues[ attributesIndicesIdx ] *=
              inputElementsOffsetsAndGains[ attributesIndicesIdx ].second;
            assert( inputFeatureValues[ attributesIndicesIdx ] >= 0.0 );
            assert( inputFeatureValues[ attributesIndicesIdx ] <= 1.0 );
          }

          // The related cluster ID
          output.getFeature( elementIdx, outputIndex, outputFeatureValue );
          
          ClusterData& currentCluster= clustersMap[ (ClusterData::ClusterLabelT)
            outputFeatureValue ];          

          elementDistance = 0;
          for( attributesIndicesIdx = 0 ; attributesIndicesIdx <
            attributesIndicesSize ; ++attributesIndicesIdx )
          {
            featureDistance = inputFeatureValues[ attributesIndicesIdx ] -
              currentCluster.m_means[ attributesIndicesIdx ];
            featureDistance = featureDistance * featureDistance;
            currentCluster.m_stdDevs[ attributesIndicesIdx ] += featureDistance;
            elementDistance += featureDistance;
          }
          elementDistance = std::sqrt( elementDistance );
          
          currentCluster.m_avgDistToClusterCenter += elementDistance;
        }
      }
      
      clustersMapIt = clustersMap.begin();

      while( clustersMapIt != clustersMap.end() )
      {
        if( clustersMapIt->second.m_size )
        {
          clustersMapIt->second.m_avgDistToClusterCenter /=
             ((double)clustersMapIt->second.m_size);

          for( attributesIndicesIdx = 0 ; attributesIndicesIdx <
            attributesIndicesSize ; ++attributesIndicesIdx )
          {
            clustersMapIt->second.m_stdDevs[ attributesIndicesIdx ] /=
               ((double)clustersMapIt->second.m_size);
            clustersMapIt->second.m_stdDevs[ attributesIndicesIdx ] =
              std::sqrt( clustersMapIt->second.m_stdDevs[ attributesIndicesIdx ] );
          }
        }

        ++clustersMapIt;
      }

      return true;
    }

    bool ISOData::splitClusters(
      const unsigned int inputElementsCount,
      const unsigned int maxNumberOfSplits,
      ClusterData::ClusterLabelT& nextAvaliableClusterLabel,
      ClustersContainerT& clustersMap ) const
    {
      if( clustersMap.size() == 0 ) return false;

      std::multimap< double, ClusterData::ClusterLabelT > clustersStdDevMap;
      double meanOfAvgDistToClusterCenters = 0;
      {
        ClustersContainerT::const_iterator clustersMapIt = clustersMap.begin();
        const ClustersContainerT::const_iterator clustersMapItEnd = clustersMap.end();
        const std::size_t stdDevsSize = ( clustersMap.empty() ? 0 :
          clustersMap.begin()->second.m_stdDevs.size() );
        std::size_t stdDevsIdx = 0 ;
        double clusterStdDev = 0;
        double meanOfAvgDistToClusterCentersDiv = 0;

        while( clustersMapIt != clustersMapItEnd )
        {
          if( clustersMapIt->second.m_size )
          {
            const std::vector< double >& stdDevs =
              clustersMapIt->second.m_stdDevs;

            clusterStdDev = 0;
            for( stdDevsIdx = 0 ; stdDevsIdx < stdDevsSize ; ++stdDevsIdx )
            {
              clusterStdDev += ( stdDevs[ stdDevsIdx ] * stdDevs[ stdDevsIdx ] );
            }
            clusterStdDev = std::sqrt( clusterStdDev );

            clustersStdDevMap.insert( std::make_pair( clusterStdDev, clustersMapIt->first ) );

            meanOfAvgDistToClusterCenters +=
              ( clustersMapIt->second.m_avgDistToClusterCenter *
              ((double)clustersMapIt->second.m_size) );

            meanOfAvgDistToClusterCentersDiv += ((double)clustersMapIt->second.m_size);
          }

          ++clustersMapIt;
        }

        if( meanOfAvgDistToClusterCentersDiv == 0 ) return false;
        meanOfAvgDistToClusterCenters /= meanOfAvgDistToClusterCentersDiv;
      }

      {
        const std::size_t attributesIndicesSize = clustersMap.empty() ? 0:
          clustersMap.begin()->second.m_means.size();
        std::multimap< double, ClusterData::ClusterLabelT >::const_reverse_iterator
          clustersStdDevMapIt = clustersStdDevMap.rbegin();
        const std::multimap< double, ClusterData::ClusterLabelT >::const_reverse_iterator
          clustersStdDevMapItEnd = clustersStdDevMap.rend();
        const unsigned int minClusterSize = 2 * ( 1 + (unsigned int)(
          m_parameters.m_minClusterSizePercent * ((double)inputElementsCount)
          / 100.0 ) );
        unsigned int splitsCount = 0;
        unsigned int attributesIndicesIdx = 0;
        std::set< ClusterData::ClusterLabelT > processedClustersLabels;

        ClusterData auxClusterData;
        auxClusterData.m_avgDistToClusterCenter = 0;
        auxClusterData.m_size = 0;
        auxClusterData.m_means.resize( attributesIndicesSize, 0 );
        auxClusterData.m_sums.resize( attributesIndicesSize, 0 );
        auxClusterData.m_stdDevs.resize( attributesIndicesSize, 0 );

        while (
                ( clustersStdDevMapIt != clustersStdDevMapItEnd )
                &&
                ( splitsCount < maxNumberOfSplits )
              )
        {
          if(
              processedClustersLabels.find( clustersStdDevMapIt->second )
              ==
              processedClustersLabels.end()
            )
          {
            if(
                (
                  clustersStdDevMapIt->first
                  >
                  m_parameters.m_stdDevSplitThreshold
                )
                &&
                (
                  clustersMap[ clustersStdDevMapIt->second ].m_avgDistToClusterCenter
                  >
                  meanOfAvgDistToClusterCenters
                )
                &&
                (
                  clustersMap[ clustersStdDevMapIt->second ].m_size
                  >
                  minClusterSize
                )
              )
            {
              ++splitsCount;

              auxClusterData.m_label = nextAvaliableClusterLabel++;

//               std::cout << std::endl << "Spliting cluster " << clustersStdDevMapIt->second
//                 << " into cluster " << auxClusterData.m_label
//                 << std::endl;

              for( attributesIndicesIdx = 0 ; attributesIndicesIdx <
                attributesIndicesSize ; ++attributesIndicesIdx )
              {
                auxClusterData.m_means[ attributesIndicesIdx ] =
                  clustersMap[ clustersStdDevMapIt->second ].m_means[ attributesIndicesIdx ] -
                  clustersMap[ clustersStdDevMapIt->second ].m_stdDevs[ attributesIndicesIdx ];
                clustersMap[ clustersStdDevMapIt->second ].m_means[ attributesIndicesIdx ] =
                  clustersMap[ clustersStdDevMapIt->second ].m_means[ attributesIndicesIdx ] +
                  clustersMap[ clustersStdDevMapIt->second ].m_stdDevs[ attributesIndicesIdx ];
              }

              clustersMap[ auxClusterData.m_label ] = auxClusterData;

              processedClustersLabels.insert( auxClusterData.m_label );
              processedClustersLabels.insert( clustersStdDevMapIt->second );
            }
          }

          ++clustersStdDevMapIt;
        }
      }

      return true;
    }

    bool ISOData::mergeClusters(
      const double mergeDistThreshold,
      const unsigned int maxNumberOfMerges,
      ClustersContainerT& clustersMap )
    {
      if( clustersMap.size() < 2 ) return false;

      std::multimap< double, std::pair< ClusterData::ClusterLabelT,
        ClusterData::ClusterLabelT > > interCluDists;
      {
        ClustersContainerT::iterator
          clustersMapIt = clustersMap.begin();
        ClustersContainerT::iterator
          nextClustersMapIt;
        const std::size_t attributesIndicesSize = clustersMap.empty() ? 0:
          clustersMap.begin()->second.m_means.size();
        double clusterDistance = 0;
        double featureDistance = 0;
        unsigned int attributesIndicesIdx = 0 ;

        while( clustersMapIt != clustersMap.end() )
        {
          nextClustersMapIt = clustersMapIt;
          ++nextClustersMapIt;

          const std::vector< double >& cluster1Mmeans =
            clustersMapIt->second.m_means;

          while( nextClustersMapIt != clustersMap.end() )
          {
            const std::vector< double >& cluster2Mmeans =
              nextClustersMapIt->second.m_means;

            clusterDistance = 0;
            for( attributesIndicesIdx = 0 ; attributesIndicesIdx <
              attributesIndicesSize ; ++attributesIndicesIdx )
            {
              featureDistance = cluster1Mmeans[ attributesIndicesIdx ] -
                cluster2Mmeans[ attributesIndicesIdx ];
              clusterDistance += ( featureDistance * featureDistance );
            }
            clusterDistance = std::sqrt( clusterDistance );

            interCluDists.insert( std::make_pair( clusterDistance,
              std::pair< ClusterData::ClusterLabelT,
              ClusterData::ClusterLabelT >( clustersMapIt->first,
              nextClustersMapIt->first ) ) );;

            ++nextClustersMapIt;
          }

          ++clustersMapIt;
        }
      }

      {
        const std::size_t attributesIndicesSize = clustersMap.empty() ? 0:
          clustersMap.begin()->second.m_means.size();
        std::set< ClusterData::ClusterLabelT > processedClustersLabels;
        std::multimap< double, std::pair< ClusterData::ClusterLabelT,
          ClusterData::ClusterLabelT > >::const_iterator interCluDistsIt =
          interCluDists.begin();
        const std::multimap< double, std::pair< ClusterData::ClusterLabelT,
          ClusterData::ClusterLabelT > >::const_iterator interCluDistsItEnd =
          interCluDists.end();
        unsigned int attributesIndicesIdx = 0 ;
        unsigned int mergesCount = 0;

        while(
                ( interCluDistsIt != interCluDistsItEnd )
                &&
                ( mergesCount < maxNumberOfMerges )
              )
        {
          if(
              ( interCluDistsIt->first < mergeDistThreshold )
              &&
              (
                processedClustersLabels.find( interCluDistsIt->second.second )
                ==
                processedClustersLabels.end()
              )
              &&
              (
                processedClustersLabels.find( interCluDistsIt->second.first )
                ==
                processedClustersLabels.end()
              )
            )
          {
//             std::cout << std::endl << "Merging cluster " << interCluDistsIt->second.first
//               << " into cluster " << interCluDistsIt->second.second
//               << std::endl;

            ++mergesCount;

            ClusterData& targetCluster =
              clustersMap[ interCluDistsIt->second.second ];
            const ClusterData& sourceCluster =
              clustersMap[ interCluDistsIt->second.first ];

            processedClustersLabels.insert( targetCluster.m_label );
            processedClustersLabels.insert( sourceCluster.m_label );

            std::vector< double >& targetMeans = targetCluster.m_means;
            const std::vector< double >& sourceMeans = sourceCluster.m_means;

            targetCluster.m_size += sourceCluster.m_size;

            for( attributesIndicesIdx = 0 ; attributesIndicesIdx <
              attributesIndicesSize ; ++attributesIndicesIdx )
            {
              targetMeans[ attributesIndicesIdx ] =
                (
                  (
                    (
                      targetMeans[ attributesIndicesIdx ]
                      *
                      ((double)targetCluster.m_size)
                    )
                    +
                    (
                      sourceMeans[ attributesIndicesIdx ]
                      *
                      ((double)sourceCluster.m_size)
                    )
                  )
                  /
                  (
                    ((double)targetCluster.m_size)
                    +
                    ((double)sourceCluster.m_size)
                  )
                );
             }

             clustersMap.erase( sourceCluster.m_label );
          }

          ++interCluDistsIt;
        }
      }

      return true;
    }

    void ISOData::printClusters( const ClustersContainerT& clustersMap )
    {
      std::cout << std::endl << "Clusters Size = " << clustersMap.size();
      for( ClustersContainerT::const_iterator
        clustersMapIt = clustersMap.begin() ; clustersMapIt != clustersMap.end() ;
        ++clustersMapIt )
      {
        std::cout << std::endl << "ClusterLabel=" << clustersMapIt->first
          << " ClusterSize=" << clustersMapIt->second.m_size
          << " AvgDistToClusterCenter=" << clustersMapIt->second.m_avgDistToClusterCenter
          ;

        std::cout << " Means=[";
        for( std::size_t meanIdx = 0 ; meanIdx < clustersMapIt->second.m_means.size() ;
          ++meanIdx )
        {
          if( meanIdx ) std::cout << ";";
          std::cout << clustersMapIt->second.m_means[ meanIdx ];
        }
        std::cout << "]";
        
        std::cout << " StdDevs=[";
        for( std::size_t stdDevsIdx = 0 ; stdDevsIdx < clustersMapIt->second.m_stdDevs.size() ;
          ++stdDevsIdx )
        {
          if( stdDevsIdx ) std::cout << ";";
          std::cout << clustersMapIt->second.m_stdDevs[ stdDevsIdx ];
        }
        std::cout << "]";        
      }
      std::cout << std::endl;
    }
    
    void ISOData::recodeClustersLabels( ClustersContainerT& clustersMap )
    {
      ClustersContainerT intrealClustersMap = clustersMap;
      clustersMap.clear();
      
      ClusterData::ClusterLabelT nextAvaliableClusterLabel = 0;
      ClusterData auxClusterData;
      
      for( ClustersContainerT::iterator it = intrealClustersMap.begin() ; it !=
        intrealClustersMap.end() ; ++it )
      {
        auxClusterData = it->second;
        auxClusterData.m_label = nextAvaliableClusterLabel++;
        
        clustersMap[ auxClusterData.m_label ] = auxClusterData;
      }
    }

  }  // end namespace cl
}    // end namespace te
