/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/classification/MaxLikelihood.cpp

  \brief A maximum likelihood estimation strategy for classification (a.k.a. MaxVer in portuguese).
*/

// TerraLib
#include "MaxLikelihood.h"

// STL
#include <limits>

namespace te
{
  namespace cl
  {
    MaxLikelihood::Parameters::Parameters()
    {
      reset();
    }

    MaxLikelihood::Parameters::~Parameters() = default;

    const MaxLikelihood::Parameters& MaxLikelihood::Parameters::operator=(
      const MaxLikelihood::Parameters& rhs)
    {
      reset();

      return *this;
    }

    void MaxLikelihood::Parameters::reset()
    {
    }

    te::common::AbstractParameters* MaxLikelihood::Parameters::clone() const
    {
      return new MaxLikelihood::Parameters(*this);
    }

    //  ------------------------------------------------------------------------------------------------------------------------------------------

    MaxLikelihood::MaxLikelihood()
    {
      reset();
    }

    MaxLikelihood::~MaxLikelihood() = default;

    bool MaxLikelihood::initialize(const MaxLikelihood::Parameters& params)
    {
      reset();

      m_parameters = params;

      m_isInitialized = true;

      return true;
    }

    void MaxLikelihood::reset()
    {
      m_isInitialized = false;
      m_parameters.reset();
      m_classesMeans.clear();
      m_classesCovarianceMatrixes.clear();
      m_classesCovarianceInvMatrixes.clear();
      m_classLabels.clear();
      m_classesOptizedMaxLikelihoodDiscriminantTerm.clear();
    }

    bool MaxLikelihood::train( const InputAdaptor< double >& samples,
      const std::vector<unsigned int>& attributesIndices,
      const std::vector<unsigned int>& sampleLabels,
      const bool enableProgressInterface)
    {
      if( ! m_isInitialized ) return false;
      if( samples.getElementsCount() == 0 ) return false;
      if( sampleLabels.size() != samples.getElementsCount() ) return false;
      if( attributesIndices.empty() ) return false;

      const unsigned int attributesIndicesSize = static_cast<unsigned int>(attributesIndices.size());

      std::unique_ptr< te::common::TaskProgress > progressPtr;
      if( enableProgressInterface )
      {
        progressPtr.reset( new te::common::TaskProgress );
        progressPtr->setTotalSteps( 3 );
        progressPtr->setMessage( "Trainning" );
      }

      // grouping samples by class

      std::vector< std::vector< std::vector< double > > > samplesByClass;

      {
        m_classLabels.clear();

        const unsigned int samplesCount = samples.getElementsCount();
        std::map< unsigned int, unsigned int > labels2ClassIndexMap;
        std::map< unsigned int, unsigned int >::iterator labels2ClassIndexMapIt;
        unsigned int attributesIndicesIdx = 0;

        for( unsigned int sampleIdx = 0 ; sampleIdx < samplesCount ; ++sampleIdx )
        {
          const unsigned int& sampleLabel = sampleLabels[ sampleIdx ];

          labels2ClassIndexMapIt = labels2ClassIndexMap.find( sampleLabel );

          if( labels2ClassIndexMapIt  == labels2ClassIndexMap.end() )
          {
            labels2ClassIndexMap[sampleLabel] = static_cast<unsigned int>(m_classLabels.size());
            m_classLabels.push_back( sampleLabel );

            samplesByClass.push_back( std::vector< std::vector< double > >() );
            samplesByClass.back().push_back( std::vector< double >() );

            std::vector< double >& sample = samplesByClass.back().back();
            sample.resize( attributesIndicesSize, 0.0 );

            for( attributesIndicesIdx = 0; attributesIndicesIdx < attributesIndicesSize;
              ++attributesIndicesIdx )
            {
              samples.getFeature( sampleIdx, attributesIndicesIdx, sample[ attributesIndicesIdx ] );
            }
          }
          else
          {
            samplesByClass[ labels2ClassIndexMapIt->second ].push_back( std::vector< double >() );
            std::vector< double >& sample = samplesByClass[ labels2ClassIndexMapIt->second ].back();
            sample.resize( attributesIndicesSize, 0.0 );

            for( attributesIndicesIdx = 0; attributesIndicesIdx < attributesIndicesSize;
              ++attributesIndicesIdx )
            {
              samples.getFeature( sampleIdx, attributesIndicesIdx, sample[ attributesIndicesIdx ] );
            }
          }
        }
      }

      if( enableProgressInterface )
      {
        progressPtr->pulse();
        if( ! progressPtr->isActive() ) return false;
      }

      // classes means calcule

      {
        const std::size_t samplesByClassSize = samplesByClass.size();

        m_classesMeans.resize( samplesByClassSize );
        std::vector< double > dymmyMeansVec( attributesIndicesSize, 0.0 );
        std::fill( m_classesMeans.begin(), m_classesMeans.end(), dymmyMeansVec );

        unsigned int attributeIdx = 0;
        double mean = 0;
        std::size_t classSamplesIdx = 0;
        std::size_t classSamplesSize = 0;

        for( std::size_t samplesByClassIdx = 0 ; samplesByClassIdx < samplesByClassSize ;
          ++samplesByClassIdx )
        {
          const std::vector< std::vector< double > >& classSamples = samplesByClass[
            samplesByClassIdx ];
          classSamplesSize = classSamples.size();

          for( attributeIdx = 0; attributeIdx < attributesIndicesSize; ++attributeIdx )
          {
            mean = 0;

            for( classSamplesIdx = 0 ; classSamplesIdx < classSamplesSize ; ++classSamplesIdx )
            {
              mean += classSamples[ classSamplesIdx ][ attributeIdx ];
            }

            mean /= (double)classSamplesSize;

            m_classesMeans[ samplesByClassIdx ][ attributeIdx ] = mean;
          }
        }
      }

      if( enableProgressInterface )
      {
        progressPtr->pulse();
        if( ! progressPtr->isActive() ) return false;
      }

      // covariance matrixes calcule

      {
        const std::size_t samplesByClassSize = samplesByClass.size();
        const boost::numeric::ublas::matrix< double > dummyCovarianceMatrix(
          attributesIndicesSize, attributesIndicesSize );

        m_classesCovarianceMatrixes.clear();
        m_classesCovarianceMatrixes.resize( samplesByClassSize, dummyCovarianceMatrix );

        m_classesCovarianceInvMatrixes.clear();
        m_classesCovarianceInvMatrixes.resize( samplesByClassSize, dummyCovarianceMatrix );

        m_classesOptizedMaxLikelihoodDiscriminantTerm.clear();
        m_classesOptizedMaxLikelihoodDiscriminantTerm.resize( samplesByClassSize, 0.0 );

        unsigned int attributeIdx1 = 0;
        unsigned int attributeIdx2 = 0;
        std::size_t classSamplesIdx = 0;
        std::size_t classSamplesSize = 0;
        double mean1 = 0;
        double mean2 = 0;
        double covariance = 0;

        for( std::size_t samplesByClassIdx = 0 ; samplesByClassIdx < samplesByClassSize ;
          ++samplesByClassIdx )
        {
          const std::vector< std::vector< double > >& classSamples = samplesByClass[
            samplesByClassIdx ];
          classSamplesSize = classSamples.size();

          for( attributeIdx1 = 0; attributeIdx1 < attributesIndicesSize; ++attributeIdx1 )
          {
            mean1 = m_classesMeans[ samplesByClassIdx ][ attributeIdx1 ];

            for( attributeIdx2 = 0; attributeIdx2 < attributesIndicesSize; ++attributeIdx2 )
            {
              mean2 = m_classesMeans[ samplesByClassIdx ][ attributeIdx2 ];
              covariance = 0.0;

              for( classSamplesIdx = 0 ; classSamplesIdx < classSamplesSize ; ++classSamplesIdx )
              {
                covariance +=
                  (
                    ( classSamples[ classSamplesIdx ][ attributeIdx1 ] - mean1 )
                    *
                    ( classSamples[ classSamplesIdx ][ attributeIdx2 ] - mean2 )
                  );
              }

              covariance /= (double)( classSamplesSize - 1 );

              m_classesCovarianceMatrixes[ samplesByClassIdx ]( attributeIdx1,
                attributeIdx2 ) = covariance;
            }
          }

          if( ! te::common::GetInverseMatrix( m_classesCovarianceMatrixes[ samplesByClassIdx ],
            m_classesCovarianceInvMatrixes[ samplesByClassIdx ] ) )
          {
            m_classesCovarianceInvMatrixes[ samplesByClassIdx ] =
              boost::numeric::ublas::identity_matrix< double >( attributesIndicesSize );
          }

          double classCovarianceMatrixDet = 0;
          if( ! te::common::GetDeterminant( m_classesCovarianceMatrixes[ samplesByClassIdx ],
            classCovarianceMatrixDet ) )
          {
            return false;
          }

          if( classCovarianceMatrixDet > 0.0 )
          {
            m_classesOptizedMaxLikelihoodDiscriminantTerm[ samplesByClassIdx ] =
              ( -1.0 * std::log( classCovarianceMatrixDet ) );
          }
          else
          {
            m_classesOptizedMaxLikelihoodDiscriminantTerm[ samplesByClassIdx ] =
              0.0;
          }
        }
      }

      if( enableProgressInterface )
      {
        progressPtr->pulse();
        if( ! progressPtr->isActive() ) return false;
      }

      return true;
    }

    bool MaxLikelihood::classify(const InputAdaptor<double>& input,
                       const std::vector<unsigned int>& attributesIndices,
                       const std::vector<double>& inputNoDataValues,
                       OutputAdaptor<unsigned int>& output,
                       const unsigned int outputIndex,
                       const unsigned int outputNoDataValue,
                       const bool enableProgressInterface )
    {
      if( !m_isInitialized ) return false;
      if( m_classesCovarianceMatrixes.empty() ) return false;
      if( input.getElementsCount() != output.getElementsCount() ) return false;
      if( attributesIndices.size() != m_classesMeans[ 0 ].size() ) return false;
      if( outputIndex >= output.getFeaturesCount() ) return false;
      if( inputNoDataValues.size() != attributesIndices.size() ) return false;

      // globals

      const unsigned int attributesIndicesSize = static_cast<unsigned int>(attributesIndices.size());
      const unsigned int inputCount = input.getElementsCount();
      const unsigned int inputFeaturesCount = input.getFeaturesCount();
      const unsigned int classesNmb = static_cast<unsigned int>(m_classesMeans.size());
      const unsigned int featuresCnt = static_cast<unsigned int>(m_classesMeans[0].size());

      // classifying

      boost::numeric::ublas::matrix< double > sample( attributesIndicesSize, 1 );
      boost::numeric::ublas::matrix< double > sampleMinusMean( attributesIndicesSize, 1 );
      boost::numeric::ublas::matrix< double > sampleMinusMeanT( 1, attributesIndicesSize );
      boost::numeric::ublas::matrix< double > auxMatrix;
      boost::numeric::ublas::matrix< double > mahalanobisDistanceMatrix;
      unsigned int attributesIndicesIdx = 0;
      unsigned int featureIdx = 0;
      unsigned int classIdx = 0;
      double closestClassdiscriminantFunctionValue = 0;
      double discriminantFunctionValue = 0;
      unsigned int closestClassIdx = 0;
      bool isValidInput = false;
      
      std::unique_ptr< te::common::TaskProgress > progressPtr;
      if( enableProgressInterface )
      {
        progressPtr.reset( new te::common::TaskProgress );
        progressPtr->setTotalSteps( inputCount );
        progressPtr->setMessage( "Classifying" );
      }      

      for( unsigned int inputIdx = 0 ; inputIdx < inputCount ; ++inputIdx )
      {
        // load sample

        isValidInput = true;

        for( attributesIndicesIdx = 0 ; attributesIndicesIdx < attributesIndicesSize ;
          ++attributesIndicesIdx )
        {
          featureIdx = attributesIndices[ attributesIndicesIdx ];
          if( featureIdx >= inputFeaturesCount ) return false;

          input.getFeature( inputIdx, featureIdx, sample( attributesIndicesIdx, 0 ) );

          if( sample( attributesIndicesIdx, 0 ) == inputNoDataValues[ attributesIndicesIdx ] )
          {
            isValidInput = false;
            break;
          }
        }

        if( isValidInput )
        {
          // finding the closest class

          closestClassdiscriminantFunctionValue = -1.0 * std::numeric_limits< double >::max();

          for( classIdx = 0 ; classIdx < classesNmb ; ++classIdx )
          {
            for( featureIdx = 0 ; featureIdx < featuresCnt ; ++featureIdx )
            {
              sampleMinusMean( featureIdx, 0 ) = sampleMinusMeanT( 0, featureIdx ) =
                ( sample( featureIdx, 0 ) - m_classesMeans[ classIdx ][ featureIdx ] );
            }

            //calculate the mahalanobis distance: (x-mean)T * CovMatrizInvert * (x-mean)

            auxMatrix = boost::numeric::ublas::prod( sampleMinusMeanT,
              m_classesCovarianceInvMatrixes[ classIdx ] );

            mahalanobisDistanceMatrix = boost::numeric::ublas::prod( auxMatrix, sampleMinusMean );
            assert( mahalanobisDistanceMatrix.size1() == 1 );
            assert( mahalanobisDistanceMatrix.size2() == 1 );

            discriminantFunctionValue = m_classesOptizedMaxLikelihoodDiscriminantTerm[ classIdx ]
              - ( 0.5 * mahalanobisDistanceMatrix( 0, 0 ) );

            if( discriminantFunctionValue > closestClassdiscriminantFunctionValue )
            {
              closestClassdiscriminantFunctionValue = discriminantFunctionValue;
              closestClassIdx = classIdx;
            }
          }

          output.setFeature( inputIdx, 0, m_classLabels[ closestClassIdx ] );
        }
        else
        {
          output.setFeature(inputIdx, 0, outputNoDataValue);
        }
        
        if( enableProgressInterface )
        {
          progressPtr->pulse();
          if( ! progressPtr->isActive() ) return false;
        }        
      }

      return true;
    }
    
    bool MaxLikelihood::getCovMatrixes( 
      std::vector< boost::numeric::ublas::matrix< double > >& covMatrixes ) const
    {
      if( m_isInitialized )
      {
        covMatrixes = m_classesCovarianceMatrixes;
        return true;        
      }
      else
      {
        return false;
      }
    }
    
    bool MaxLikelihood::getInverseCovMatrixes( 
      std::vector< boost::numeric::ublas::matrix< double > >& invCovMatrixes ) const
    {
      if( m_isInitialized )
      {
        invCovMatrixes = m_classesCovarianceInvMatrixes;
        return true;        
      }
      else
      {
        return false;
      }
    }    
    
    bool MaxLikelihood::getMeans( std::vector< std::vector< double > >& classesMeans ) const
    {
      if( m_isInitialized )
      {
        classesMeans = m_classesMeans;
        return true;        
      }
      else
      {
        return false;
      }
    }
    
    bool MaxLikelihood::getLables( std::vector<unsigned int>& classLabels ) const
    {
      if( m_isInitialized )
      {
        classLabels = m_classLabels;
        return true;
      }
      else
      {
        return false;
      }
    }

  }  // end namespace cl
}    // end namespace te
