/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/color/Utils.cpp

  \brief Implementation of utility functions for color bar.
*/

// TerraLib
#include "../common/Exception.h"
#include "../core/filesystem/FileSystem.h"
#include "../core/translator/Translator.h"
#include "../core/utils/Platform.h"
#include "ColorScheme.h"
#include "ColorSchemeCatalog.h"
#include "ColorSchemeCatalogManager.h"
#include "ColorSchemeGroup.h"
#include "Utils.h"

//STL
#include <string>
#include <iostream>
#include <fstream>

// Boost
#include <boost/property_tree/json_parser.hpp>
#include <boost/property_tree/ptree.hpp>
#include <boost/filesystem/fstream.hpp>
#include <boost/filesystem.hpp>

te::color::ColorSchemeCatalog* te::color::loadColorBar(const std::string& jsonfile, const std::string& catalogName)
{
  //create color scheme catalog
  te::color::ColorSchemeCatalog* csc = new te::color::ColorSchemeCatalog(catalogName);

  try
  {
    if (!te::core::FileSystem::exists(jsonfile))
      return nullptr;

    boost::filesystem::path path(boost::filesystem::canonical(jsonfile));

    boost::filesystem::ifstream f(path);

    if (!f.is_open())
      return nullptr;

    boost::property_tree::ptree pt;
    boost::property_tree::json_parser::read_json(f, pt);

    for (boost::property_tree::ptree::value_type &v : pt.get_child("ramps"))
    {
      std::string name = v.second.get<std::string>("name");

      //create color scheme group
      te::color::ColorSchemeGroup* csg = new te::color::ColorSchemeGroup(name);

      csc->push_back(csg);

      for (boost::property_tree::ptree::value_type &t : v.second.get_child("schemes"))
      {
        std::string name = t.second.get<std::string>("name");

        te::color::ColorScheme* cs = new te::color::ColorScheme(name);

        std::vector<te::color::RGBAColor>* rgbaVec = new std::vector<te::color::RGBAColor>();

        for (boost::property_tree::ptree::value_type &c : t.second.get_child("values"))
        {
          unsigned int red = c.second.get<unsigned int>("red");
          unsigned int green = c.second.get<unsigned int>("green");
          unsigned int blue = c.second.get<unsigned int>("blue");

          rgbaVec->push_back(te::color::RGBAColor(red, green, blue, TE_OPAQUE));
        }

        cs->push_back(rgbaVec);

        csg->push_back(cs);
      }
    }

    f.close();
  }
  catch (boost::property_tree::json_parser::json_parser_error &je)
  {
    std::string errmsg = "Error parsing: " + je.filename() + ": " + je.message();
    te::common::Exception ex(TE_TR(errmsg));
    throw(ex);
  }
  catch (std::exception const& e)
  {
    std::cerr << e.what() << std::endl;

    return nullptr;
  }

  return csc;
}

bool te::color::saveCustomColorBar(const std::string& groupName, const std::string& schemeName, const std::vector<te::color::RGBAColor> rgbaVec)
{
  te::color::ColorSchemeCatalog* csc = ColorSchemeCatalogManager::getInstance().findByName("Custom");

  if (csc)
  {
    //create color custom scheme group
    std::vector<te::color::ColorSchemeGroup*> colorSchemeGroups = csc->getColorSchemeGroups();

    te::color::ColorSchemeGroup* csg = nullptr;
    te::color::ColorScheme* cs = nullptr;

    for (unsigned int i = 0; i < colorSchemeGroups.size(); ++i)
    {
      if (colorSchemeGroups[i]->getName() == groupName)
      {
        csg = colorSchemeGroups[i];
        break;
      }
    }
    if (!csg)
    {
      csg = new te::color::ColorSchemeGroup(groupName);
      csc->push_back(csg);

      cs = new te::color::ColorScheme(schemeName);
      std::vector<te::color::RGBAColor>* rgba = new std::vector<te::color::RGBAColor>(rgbaVec);
      cs->push_back(rgba);

      csg->push_back(cs);
    }
    else
    {
      std::vector<te::color::ColorScheme*> colorSchemes = csg->getColorSchemes();

      for (unsigned int i = 0; i < colorSchemes.size(); ++i)
      {
        if (colorSchemes[i]->getName() == schemeName)
        {
          cs = colorSchemes[i];
          break;
        }
      }

      if (!cs)
      {
        cs = new te::color::ColorScheme(schemeName);
        std::vector<te::color::RGBAColor>* rgba = new std::vector<te::color::RGBAColor>(rgbaVec);
        cs->push_back(rgba);

        csg->push_back(cs);
      }
    }
  }
  boost::property_tree::ptree pt;
  boost::property_tree::ptree children;

  std::fstream f;
  std::string jsonf = te::core::GetUserDirectory() + "/custom_color_ramps.json";
  if (!te::core::FileSystem::exists(jsonf))
  {
    boost::property_tree::ptree ramps_child;

    pt.put("name", groupName);
    boost::property_tree::ptree group_children;
    boost::property_tree::ptree group_child;
    group_child.put("name", schemeName);

    boost::property_tree::ptree scheme_children;

    for (std::size_t i = 0; i < rgbaVec.size(); ++i)
    {
      boost::property_tree::ptree child;
      child.put("red", rgbaVec[i].getRed());
      child.put("green", rgbaVec[i].getGreen());
      child.put("blue", rgbaVec[i].getBlue());
      scheme_children.push_back(std::make_pair("", child));
    }

    group_child.put_child("values", scheme_children);
    group_children.push_back(std::make_pair("", group_child));

    pt.put_child("schemes", group_children);

    ramps_child.push_back(std::make_pair("", pt));
    children.push_back(std::make_pair("ramps", ramps_child));

    boost::property_tree::json_parser::write_json(jsonf, children);

    return true;
  }

  f.open(jsonf.c_str());

  if (!f.is_open())
    return false;

  boost::property_tree::json_parser::read_json(f, pt);

  boost::property_tree::ptree ramps_child;

  bool exitGroup = false;
  bool existScheme = false;

  for (boost::property_tree::ptree::value_type& v : pt.get_child("ramps"))
  {
    boost::property_tree::ptree group_children;
    boost::property_tree::ptree pt1;

    std::string group_name = v.second.get<std::string>("name");
    pt1.put("name", group_name);

    if (group_name == groupName)
    {
      exitGroup = true;
    }

    for (boost::property_tree::ptree::value_type& t :
      v.second.get_child("schemes"))
    {
      std::string scheme_name = t.second.get<std::string>("name");
      boost::property_tree::ptree group_child;
      group_child.put("name", scheme_name);

      if (group_name == groupName && scheme_name == schemeName)
      {
        existScheme = true;
      }

      boost::property_tree::ptree scheme_children;

      if (group_name == groupName && scheme_name == schemeName)
      {//if scheme already exist, replace it
        for (std::size_t i = 0; i < rgbaVec.size(); ++i)
        {
          boost::property_tree::ptree child;
          child.put("red", rgbaVec[i].getRed());
          child.put("green", rgbaVec[i].getGreen());
          child.put("blue", rgbaVec[i].getBlue());
          scheme_children.push_back(std::make_pair("", child));
        }
      }
      else
      {
        for (boost::property_tree::ptree::value_type& c :
          t.second.get_child("values"))
        {
          unsigned int red = c.second.get<unsigned int>("red");
          unsigned int green = c.second.get<unsigned int>("green");
          unsigned int blue = c.second.get<unsigned int>("blue");

          boost::property_tree::ptree child;
          child.put("red", red);
          child.put("green", green);
          child.put("blue", blue);

          scheme_children.push_back(std::make_pair("", child));
        }
      }

      group_child.put_child("values", scheme_children);
      group_children.push_back(std::make_pair("", group_child));
    }

    if (!existScheme && groupName == group_name)
    {
      boost::property_tree::ptree group_child;
      group_child.put("name", schemeName);

      boost::property_tree::ptree scheme_children;

      for (std::size_t i = 0; i < rgbaVec.size(); ++i)
      {
        boost::property_tree::ptree child;
        child.put("red", rgbaVec[i].getRed());
        child.put("green", rgbaVec[i].getGreen());
        child.put("blue", rgbaVec[i].getBlue());
        scheme_children.push_back(std::make_pair("", child));
      }

      group_child.put_child("values", scheme_children);
      group_children.push_back(std::make_pair("", group_child));
    }

    pt1.add_child("schemes", group_children);
    ramps_child.push_back(std::make_pair("", pt1));
  }

  //if group not exist
  if (!exitGroup && !existScheme)
  {
    boost::property_tree::ptree group_children;
    boost::property_tree::ptree pt1;

    pt1.put("name", groupName);

    boost::property_tree::ptree group_child;
    group_child.put("name", schemeName);

    boost::property_tree::ptree scheme_children;

    for (std::size_t i = 0; i < rgbaVec.size(); ++i)
    {
      boost::property_tree::ptree child;
      child.put("red", rgbaVec[i].getRed());
      child.put("green", rgbaVec[i].getGreen());
      child.put("blue", rgbaVec[i].getBlue());
      scheme_children.push_back(std::make_pair("", child));
    }

    group_child.put_child("values", scheme_children);
    group_children.push_back(std::make_pair("", group_child));

    pt1.add_child("schemes", group_children);
    ramps_child.push_back(std::make_pair("", pt1));
  }

  children.push_back(std::make_pair("ramps", ramps_child));

  boost::property_tree::json_parser::write_json(jsonf, children);

  f.close();

  return true;
}

te::color::ColorSchemeCatalog* te::color::GetSchemeCatalog(std::string catalogName)
{
  return te::color::ColorSchemeCatalogManager::getInstance().findByName(catalogName);
}
