/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralib/common/ThreadGroup.h

   \brief This class represents a thread group

   \ingroup common
 */

#ifndef __TERRALIB_COMMON_INTERNAL_THREADGROUP_H
#define __TERRALIB_COMMON_INTERNAL_THREADGROUP_H

 // TerraLib
#include "Config.h"

//Boost include files
#include <boost/thread.hpp>
#include <boost/asio/io_service.hpp>

//STL
#include <functional>
#include <string>

namespace te
{
  namespace common
  {
    /*!
      \class ThreadGroup
      \brief This class represents a thread group
    */
    class TECOMMONEXPORT ThreadGroup
    {
    public:

      //!< Constructor. If the given thread count is 0, it will detect and use all available threads
      ThreadGroup(std::size_t threadCount = 0);
      
      //!< Destructor
      ~ThreadGroup();

      //!< Gets the number of threads that are defined to be used by this class
      std::size_t getThreadCount();

      //!< Creates a new job and binds it to the given function. Use std::bind(function, object, params...) as argument
      template <typename Function>
      void addJob(Function func);
      
      //!< Waits for all threads to finish their jobs
      void waitToFinish();

    private:

      //!< Starts the execution of the thread group. Note that the jobs must be added before the calling of this function
      void run();

    private:

      std::size_t m_threadCount;
      std::unique_ptr<boost::asio::io_service> m_ioService; //!< Boost IO service
      std::unique_ptr<boost::thread_group> m_threadGroup; //!< Boost thread group
      std::unique_ptr<boost::asio::io_service::work> m_work; //!< Boost work
    };

    //implementation of the templated function
    template <typename Function>
    void ThreadGroup::addJob(Function func)
    {
      m_ioService->post(func);
    }

  } // end namespace common
}   // end namespace te

#endif  // __TERRALIB_COMMON_INTERNAL_THREADGROUP_H

