/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled
   applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/core/filesystem/FileSystem.cpp

  \brief A class for handling system files and paths.

  \author Matheus Cavassan Zaglia
*/

// TerraLib
#include "FileSystem.h"
#include "../Exception.h"
#include "../encoding/CharEncoding.h"
#include "../logger/Logger.h"

// Boost
#include <boost/filesystem.hpp>

//STL
#include <vector>
#include <stdio.h>
#include <iostream>
#include <sstream>
#include <fstream>
#include <algorithm>
#include <cctype>
#include <ctype.h>
#include <exception>

// System dependent includes
#if TE_PLATFORMCODE_MSWINDOWS == TE_PLATFORM
#include <windows.h>
#elif TE_PLATFORMCODE_LINUX == TE_PLATFORM
#include <limits.h>
#include <unistd.h>
#elif TE_PLATFORMCODE_APPLE == TE_PLATFORM
#include <mach-o/dyld.h>
#endif

boost::mutex te::core::FileSystem::mutexFileSystem;

std::string te::core::FileSystem::currentPath()
{
  return te::core::CharEncoding::toUTF8(
      boost::filesystem::current_path().string());
}

std::string te::core::FileSystem::systemCompletePath(const std::string& path)
{
  std::string complete =
      boost::filesystem::system_complete(te::core::CharEncoding::fromUTF8(path))
          .string();
  return te::core::CharEncoding::toUTF8(complete);
}

std::string te::core::FileSystem::tempDirectoryPath()
{
  return te::core::CharEncoding::toUTF8(
      boost::filesystem::temp_directory_path().string());
}

std::string te::core::FileSystem::uniquePath(const std::string& format)
{
  return te::core::CharEncoding::toUTF8(
      boost::filesystem::unique_path(format).string());
}

std::string te::core::FileSystem::absolutePath(const std::string& path)
{
  return te::core::CharEncoding::toUTF8(
      boost::filesystem::absolute(path).string());
}

std::string te::core::FileSystem::extension(const std::string& path)
{
  return boost::filesystem::extension(path);
}

std::string te::core::FileSystem::fileName(const std::string& path)
{
  return te::core::CharEncoding::toUTF8(boost::filesystem::path(path).filename().string());
}

std::string te::core::FileSystem::stem(const std::string& path)
{
  return te::core::CharEncoding::toUTF8(boost::filesystem::path(path).stem().string());
}

std::string te::core::FileSystem::parentPath(const std::string& path)
{
  return te::core::CharEncoding::toUTF8(boost::filesystem::path(path).parent_path().string());
}

bool te::core::FileSystem::isDirectory(const std::string& path)
{
  return boost::filesystem::is_directory(
      te::core::CharEncoding::fromUTF8(path));
}

bool te::core::FileSystem::isEmpty(const std::string& path)
{
  return boost::filesystem::is_empty(te::core::CharEncoding::fromUTF8(path));
}

bool te::core::FileSystem::isRegularFile(const std::string& path)
{
  return boost::filesystem::is_regular_file(
      te::core::CharEncoding::fromUTF8(path));
}

std::string te::core::FileSystem::executableDirectory()
{
#if TE_PLATFORMCODE_MSWINDOWS == TE_PLATFORM
  char exePath[1024];
  if(GetModuleFileName(NULL, exePath, sizeof(exePath)) == 0)
    exePath[0] = '\0';
#elif TE_PLATFORMCODE_SOLARIS == TE_PLATFORM
  char exePath[PATH_MAX];
  if(realpath(getexecname(), exePath) == NULL)
    exePath[0] = '\0';
#elif TE_PLATFORMCODE_LINUX == TE_PLATFORM
  char exePath[PATH_MAX];
  ssize_t len = ::readlink("/proc/self/exe", exePath, sizeof(exePath));
  if(len == -1 || len == sizeof(exePath))
    len = 0;
  exePath[len] = '\0';
#elif TE_PLATFORMCODE_APPLE == TE_PLATFORM
  char exePath[PATH_MAX];
  uint32_t len = sizeof(exePath);
  if(_NSGetExecutablePath(exePath, &len) != 0)
  {
    exePath[0] = '\0';  // buffer too small (!)
  }
  else
  {
    char* canonicalPath = realpath(exePath, NULL);
    if(canonicalPath != NULL)
    {
      strncpy(exePath, canonicalPath, len);
      free(canonicalPath);
    }
  }
#endif

  boost::filesystem::path p(exePath);
  return te::core::CharEncoding::toUTF8(p.parent_path().string());
}

bool te::core::FileSystem::exists(const std::string& path)
{
  return boost::filesystem::exists(te::core::CharEncoding::fromUTF8(path));
}

bool te::core::FileSystem::createDirectory(const std::string& path)
{
  return boost::filesystem::create_directory(
      te::core::CharEncoding::fromUTF8(path));
}

bool te::core::FileSystem::createDirectories(const std::string& path)
{
  return boost::filesystem::create_directories(
      te::core::CharEncoding::fromUTF8(path));
}

void te::core::FileSystem::copyFile(const std::string& from,
                                    const std::string& to)
{
  boost::filesystem::copy_file(te::core::CharEncoding::fromUTF8(from),
                               te::core::CharEncoding::fromUTF8(to));
}

bool te::core::FileSystem::remove(const std::string& path)
{
  return boost::filesystem::remove(te::core::CharEncoding::fromUTF8(path));
}

void te::core::FileSystem::rename(const std::string& old_p,
                                  const std::string& new_p)
{
  boost::filesystem::rename(te::core::CharEncoding::fromUTF8(old_p),
                            te::core::CharEncoding::fromUTF8(new_p));
}

std::vector<std::string> te::core::FileSystem::listDirectory(
    const std::string& path)
{
  if(!isDirectory(path))
    throw InvalidArgumentException()
        << ErrorDescription("The given path is not a directory.");

  std::vector<std::string> result;

  boost::filesystem::path p(te::core::CharEncoding::fromUTF8(path));

  for(boost::filesystem::directory_iterator it(p), it_end; it != it_end; ++it)
    result.push_back(te::core::CharEncoding::toUTF8(it->path().string()));

  return result;
}

uintmax_t te::core::FileSystem::fileSize(const std::string &path)
{
  return boost::filesystem::file_size(te::core::CharEncoding::fromUTF8(path));
}

std::string te::core::FileSystem::adjustPathToNativeForm(const std::string& filePath)
{
  boost::filesystem::path boostPath(filePath);
  boostPath = boostPath.make_preferred();
  
  return boostPath.string();
}

std::string te::core::FileSystem::adjustPathToGenericForm(const std::string& filePath)
{
  boost::filesystem::path boostPath(filePath);
  return boostPath.generic_string();
}

std::string te::core::FileSystem::getNativeSeparator()
{
  boost::filesystem::path path;
  path += boost::filesystem::path::preferred_separator;

  return path.string();
}

std::string te::core::FileSystem::readFromFile(const std::string& filePath)
{
  boost::mutex::scoped_lock scoped_lock(te::core::FileSystem::mutexFileSystem);

  if (exists(filePath) == false)
  {
    throw InvalidArgumentException() << ErrorDescription("The given file name does not exists");
  }
  std::string tempFilePath = filePath + ".temp";

  //if the temporary file still exists, we must check if the original file also exists  
  if (exists(tempFilePath) == true)
  {
    //if the orignal file does not exist, we must use the temporary file as there might occurred an error when writing it previously
    if (exists(filePath) == false)
    {
      rename(tempFilePath, filePath);
    }
    else
    {
      // remove the temporary file
      remove(tempFilePath);
    }
  }

  std::string content = getFileContent(filePath);
  return content;
}

void te::core::FileSystem::writeToFile(const std::string& filePath, const std::string& content)
{
  boost::mutex::scoped_lock scoped_lock(te::core::FileSystem::mutexFileSystem);
  
  if (isInvalid(content)) {

    TE_LOG_WARN("The content to be written to a file is invalid.");
    return;
  }

  std::string tempFilePath = filePath + ".temp";

  //if there is an old temporary file, we just remove it
  if (exists(tempFilePath) == true)
  {
    remove(tempFilePath);
  }

  //we first write the contents to a temporary file
  {
    //this scope if to force the flush of the content in the destruction of the stream
    std::ofstream out(te::core::CharEncoding::fromUTF8(tempFilePath).c_str());
    out << content;
  }

  //then we remove the original file (if it already exists)
  if (exists(filePath) == true)
  {
    remove(filePath);
  }

  //and finally we rename the temporary file to the orginal file name
  rename(tempFilePath, filePath);
}

std::string te::core::FileSystem::getFileContent(const std::string& filePath)
{
  std::string content;

  try
  {
    if (exists(filePath))
    {
      std::ifstream file(te::core::CharEncoding::fromUTF8(filePath).c_str());

      if (isBlank(filePath))
      {
        TE_LOG_WARN("Contains full non-printable character content.");
        return content;
      }

      content = std::string((std::istreambuf_iterator<char>(file)), std::istreambuf_iterator<char>());

      return content;
    }
  }
  catch (const std::exception& e)
  {
    TE_LOG_ERROR(("File System - Get file content - STD Exception: ") + e.what());
    return content;
  }
  catch (...)
  {
    TE_LOG_ERROR("Could not read file.");
    return content;
  }

  return content;
}

std::string te::core::FileSystem::createTempFilePathString(const std::string& filePath)
{
  std::string sufix("_copy");
  
  std::string fNameWithoutExtension = stem(filePath);
  
  std::string tempDir = createTempFileDirectoryPathString();
  std::string fileExtension = extension(filePath);

  std::stringstream ssTempFileName;
  ssTempFileName << tempDir << fNameWithoutExtension << sufix << fileExtension;
  std::string tempFilePath = ssTempFileName.str();

  return tempFilePath;
}

std::string te::core::FileSystem::createTempFileDirectoryPathString()
{
  std::string tempDir = tempDirectoryPath();
  std::string slash = getNativeSeparator();

  std::stringstream ss;
  ss << tempDir << slash << "terralib" << slash << "filesystem" << slash;
  std::string path = ss.str();

  return path;
}

bool te::core::FileSystem::isBlank(const std::string& filePath)
{
  bool is_blank = true;
  std::string line;
 
  std::ifstream file(te::core::CharEncoding::fromUTF8(filePath).c_str());
  
  while (std::getline(file, line)) 
  {
    std::ptrdiff_t printableCharacter = std::count_if(line.begin(), line.end(), [](unsigned char c) { return static_cast<bool>(std::isgraph(c)); });

    // checks if the given character is printable (excluding space)
    if (!line.empty() && printableCharacter > 0)
    {
      is_blank = false;
      break;
    }
  }

  return is_blank;
}

bool te::core::FileSystem::isInvalid(const std::string& content)
{
  if (std::all_of(content.begin(), content.end(), [](char c) { return !std::isgraph(c); }))
  {
    return true;
  }

  return false;
}

bool te::core::FileSystem::allocateTmpDiskFile( unsigned long int size, FILE** fileptr,
   std::string& fullFileName )
{
  boost::mutex::scoped_lock scoped_lock(te::core::FileSystem::mutexFileSystem);
  
  bool returnValue = true;
  const boost::filesystem::path randShortFileNamePath( 
    "TerralibTmpFile_%%%%-%%%%-%%%%-%%%%" );  
  boost::filesystem::path filePath;

  // trying to use the user home path
  
  {
    fullFileName = te::core::FileSystem::uniquePath( 
      ( boost::filesystem::path( te::core::GetUserDirectory() ) /= 
      randShortFileNamePath ).string() );
      
    if( fullFileName.empty() )
    {
      returnValue = false;
    }
    else       
    {
      filePath = boost::filesystem::path( fullFileName );

      try
      {
        (*fileptr) = std::fopen( te::core::CharEncoding::fromUTF8(fullFileName).c_str(),  "w+b"  );
      }
      catch(...)
      {
        (*fileptr) = nullptr;
        returnValue = false;
      }

      if( (*fileptr) != nullptr )
      {
        std::fclose( (*fileptr) );

        try
        {
          boost::filesystem::resize_file( filePath, (boost::uintmax_t)size );
        }
        catch(...)
        {
          boost::filesystem::remove( filePath );
          returnValue = false;
        }

        try
        {
          (*fileptr) = std::fopen( te::core::CharEncoding::fromUTF8(fullFileName).c_str(),  "r+b"  );
        }
        catch(...)
        {
          (*fileptr) = nullptr;
        }

        if( (*fileptr) == nullptr )
        {
          boost::filesystem::remove( filePath );
          returnValue = false;
        }
      }
    }
  }  
  
  // trying to use the system temp path
  
  if( ! returnValue )
  {
    fullFileName = te::core::FileSystem::uniquePath( 
      ( boost::filesystem::path( te::core::FileSystem::tempDirectoryPath() ) /=
      randShortFileNamePath).string() );
      
    if( fullFileName.empty() )
    {
      returnValue = false;
    }
    else
    {
      filePath = boost::filesystem::path( fullFileName );

      try
      {
        (*fileptr) = std::fopen( te::core::CharEncoding::fromUTF8(fullFileName).c_str(),  "w+b"  );
      }
      catch(...)
      {
        (*fileptr) = nullptr;
        returnValue = false;
      }

      if( (*fileptr) != nullptr )
      {
        std::fclose( (*fileptr) );

        try
        {
          boost::filesystem::resize_file( filePath, (boost::uintmax_t)size );
        }
        catch(...)
        {
          boost::filesystem::remove( filePath );
          returnValue = false;
        }

        try
        {
          (*fileptr) = std::fopen( te::core::CharEncoding::fromUTF8(fullFileName).c_str(),  "r+b"  );
        }
        catch(...)
        {
          (*fileptr) = nullptr;
        }

        if( (*fileptr) == nullptr )
        {
          boost::filesystem::remove( filePath );
          returnValue = false;
        }
      }
    }
  }
  
  return returnValue;
}
