/*
  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

  This file is part of the TerraLib - a Framework for building GIS enabled
  applications.

  TerraLib is free software: you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation, either version 3 of the License,
  or (at your option) any later version.

  TerraLib is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with TerraLib. See COPYING. If not, write to
  TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/core/plugin/DefaultSerializers.cpp

  \brief A singleton that can be used to register plugin finders.

  \author Gilberto Ribeiro de Queiroz
  \author Matheus Cavassan Zaglia
 */

// TerraLib
#include "Serializers.h"
#include "Exception.h"
#include "../filesystem/FileSystem.h"
#include "../encoding/CharEncoding.h"
#include "../translator/Translator.h"

// Boost
#include <boost/filesystem.hpp>
#include <boost/format.hpp>
#include <boost/property_tree/json_parser.hpp>
#include <boost/property_tree/ptree.hpp>

te::core::PluginInfo te::core::JSONPluginInfoSerializer(
    const std::string& file_name)
{
  boost::property_tree::ptree doc;

  if (!te::core::FileSystem::exists(file_name))
  {
    boost::format err_msg(TE_TR("The file %1% doesn't exist."));
    throw InvalidArgumentException()
        << ErrorDescription((err_msg % file_name).str());
  }
  boost::property_tree::json_parser::read_json(file_name, doc);

  PluginInfo plugin;

  plugin.name = doc.get<std::string>("name");
  plugin.display_name = doc.get<std::string>("display_name");
  plugin.description = doc.get<std::string>("description");
  plugin.version = doc.get<std::string>("version");
  plugin.release = doc.get<std::string>("release");
  plugin.engine = doc.get<std::string>("engine");
  plugin.license_description = doc.get<std::string>("license_description");
  plugin.license_URL = doc.get<std::string>("license_URL");
  plugin.site = doc.get<std::string>("site");

  for(const boost::property_tree::ptree::value_type& v :
       doc.get_child("provider"))
  {
    if(v.first == "name")
      plugin.provider.name = v.second.get_value<std::string>();
    else if(v.first == "site")
      plugin.provider.site = v.second.get_value<std::string>();
    else
      plugin.provider.email = v.second.get_value<std::string>();
  }

  for (const boost::property_tree::ptree::value_type& v :
       doc.get_child("dependencies"))
    plugin.dependencies.push_back(v.second.get_value<std::string>());
  for(const boost::property_tree::ptree::value_type& v :
       doc.get_child("resources"))
  {
    Resource r;

    r.first = v.first;
    r.second = v.second.get_value<std::string>();

    plugin.resources.push_back(r);
  }

  // plugin.description = doc.get<std::string>("parameters");
  // plugin.description = doc.get<std::string>("host_application");

  return plugin;
}
