/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/core/translator/Translator.cpp

  \brief This class is designed for dealing with multi-language text translation in TerraLib.

  \author Matheus Cavassan Zaglia
  \author Gilberto Ribeiro de Queiroz
 */

// Boost
#include <boost/locale.hpp>
#include <boost/format.hpp>
#include <boost/filesystem.hpp>

// TerraLib
#include "Translator.h"
#include "../Exception.h"
#include "../utils/Platform.h"
#include "../logger/Logger.h"

te::core::Translator& te::core::Translator::instance()
{
  static Translator instance;

  return instance;
}

te::core::Translator::Translator() 
  : m_loadTranslationFiles(false)
  , m_localeGenerator(new boost::locale::generator())
{
  m_localeGenerator->locale_cache_enabled(true);
}

te::core::Translator::~Translator()
{
  delete m_localeGenerator;
}

std::string te::core::Translator::translate(const std::string& message)
{
  return translate(message.c_str());
}

std::string te::core::Translator::translate(const char* message)
{
#ifdef TERRALIB_TRANSLATOR_ENABLED

  updateTranslationFiles();

  for(const auto& dir : m_textDomainMap)
  {
    for(const auto& domain : dir.second)
    {
      std::string translated = boost::locale::translate(message).str(m_translationLocale, domain);
      if (translated != message)
      {
        return translated;
      }
    }
  }
  return message;
#else
  return message;
#endif
}

std::string
te::core::Translator::translate(const std::string& msg1,
                                const std::string& msg2,
                                unsigned int n)
{
  return translate(msg1.c_str(), msg2.c_str(), n);
}

std::string
te::core::Translator::translate(const char* msg1,
                                const char* msg2,
                                unsigned int n)
{
#ifdef TERRALIB_TRANSLATOR_ENABLED

  updateTranslationFiles();

  for(const auto& dir : m_textDomainMap)
  {
    for(const auto& domain : dir.second)
    {
      std::string translated = boost::locale::translate(msg1, msg2, n).str(m_translationLocale, domain);

      if(n == 1 && translated != msg1)
      {
        return translated;
      }
      else if(n > 1 && translated != msg2)
      {
        return translated;
      }
    }
  }
  return ((n == 1) ? msg1 : msg2);
#else
  return ((n == 1) ? msg1 : msg2);
#endif
}

void te::core::Translator::setLocale(const std::string& locale)
{
  m_locale = locale + ".UTF-8";
}

void te::core::Translator::addTextDomain(const std::string& textDomain, const std::string& dir)
{
  if(exist(textDomain))
  {
    boost::format err_msg(TE_TR("The text domain %1% already exist."));
    throw Exception() << te::ErrorDescription((err_msg % textDomain).str());
  }

  if(m_textDomainMap.find(dir) == m_textDomainMap.end())
  {
    m_localeGenerator->add_messages_path(dir);
  }

  m_localeGenerator->add_messages_domain(textDomain);

  m_textDomainMap[dir].push_back(textDomain);

  m_loadTranslationFiles = true;
}


bool te::core::Translator::exist(const std::string& textDomain)
{
  for(const auto& dir : m_textDomainMap)
  {
    auto domain = std::find(dir.second.begin(), dir.second.end(), textDomain);
    if(domain != dir.second.end())
      return true;
  }
  return false;
}

void te::core::Translator::updateTranslationFiles()
{
  if(m_loadTranslationFiles == true)
  {
    m_loadTranslationFiles = false;
    m_translationLocale = m_localeGenerator->generate(m_locale);
  }
}

void te::core::Translator::loadAllTranslations(const std::string& translationsDir)
{
  std::string translationExtention = ".mo";

  boost::filesystem::path translationPath(translationsDir);

  if(boost::filesystem::exists(translationPath))
  {
    boost::filesystem::recursive_directory_iterator end_itr;

    for (boost::filesystem::recursive_directory_iterator itr (translationPath); itr != end_itr; ++itr)
    {
      if (is_regular_file (itr->path ()))
      {
        if(itr->path().extension()==translationExtention)
        {
          std::string translationFile = itr->path ().string ();

          boost::filesystem::path translationFilePath(translationFile);

          std::string textDomain = translationFilePath.stem().string();

          addTextDomain(textDomain, translationsDir);

          TE_LOG_INFO("Loading translation file: " + translationFile);
        }
      }
    }
  }
}
