/*
  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

  This file is part of the TerraLib - a Framework for building GIS enabled applications.

  TerraLib is free software: you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation, either version 3 of the License,
  or (at your option) any later version.

  TerraLib is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with TerraLib. See COPYING. If not, write to
  TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/core/uri/ProxyInfo.h

  \brief A class to store the proxy information that must be used to access data located in URIs.

  \author Mario Rocco Pettinati
 */

#ifndef __TERRALIB_CORE_URI_PROXYINFO_H__
#define __TERRALIB_CORE_URI_PROXYINFO_H__

// TerraLib
#include "../Config.h"

// STL
#include <string>

namespace te
{
  namespace core
  {
    /*!
      \class URI

      \brief A class to store the proxy information that must be used to access data located in URIs.
     */
    class TECOREEXPORT ProxyInfo
    {
      public:

        /*! \brief Default constructor. */
        ProxyInfo();

        /*! \brief Gets Proxy Host information. 
            \return The Proxy Host information
        */
        const std::string& getHost() const;

        /*! \brief Sets Proxy Host information.
            \param host The Proxy Host information
        */
        void setHost(const std::string& host);

        /*! \brief Gets Proxy Port information.
            \return The Proxy Port information
        */
        int getPort() const;

        /*! \brief Sets Proxy Port information.
            \param port The Proxy Port information
        */
        void setPort(int port);

        /*! \brief Gets Proxy UserName information.
            \return The Proxy UserName information
        */
        const std::string& getUserName() const;

        /*! \brief Sets Proxy User Name information.
            \param userName The Proxy User Name information
        */
        void setUserName(const std::string& userName);

        /*! \brief Gets Proxy Password information.
            \return The Proxy Password information
        */
        const std::string& getPassword() const;

        /*! \brief Sets Proxy Password information.
            \param password The Proxy Password information
        */
        void setPassword(const std::string& password);

        /*! \brief Checks if the use of the proxy settings is enabled and must be used in requests.
            \return TRUE if the of the proxy settings is enabled; FALSE otherwise
        */
        bool isEnabled() const;

        /*! \brief Sets Enabled or Disabled the use of the proxy settings.
            \param enabled The new state of the use of the proxy settings.
        */
        void setEnabled(bool enabled);

        /*! \brief Checks if the proxy settings has authentication information sets (userName and password). */
        bool isAuthenticated() const;

        /*! \brief Gets the proxy settings host component (host:port). */
        std::string getHostPort() const;

        /*! \brief Gets the proxy settings authentication component (user:password). */
        std::string getAuthentication() const;

        /*! \brief Gets the proxy settings as a string.
            \return The proxy settings as a string.
        */
        std::string toString() const;

      protected:

        std::string m_host; //!< Stores information related to the proxy host
        int m_port; //!< Stores information related to the proxy port
        std::string m_userName; //!< Stores information related to the proxy user
        std::string m_password; //!< Stores information related to the proxy password
        bool m_isEnabled; //!< Stores information related to the enabling of the use of these proxy settings

    };

  }  // end namespace core
}    // end namespace te

#endif  // __TERRALIB_CORE_URI_PROXYINFO_H__
