/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/memory/DataSetItem.cpp

  \brief An implementation of the DatasetItem class for the TerraLib In-Memory Data Access driver.
*/

// TerraLib
#include "../../common/StringUtils.h"
#include "../../core/translator/Translator.h"
#include "../../datatype/AbstractData.h"
#include "../../datatype/ByteArray.h"
#include "../../datatype/DateTime.h"
#include "../../datatype/Property.h"
#include "../../datatype/SimpleData.h"
#include "../../geometry/Geometry.h"
#include "../../raster/Grid.h"
#include "../../raster/RasterProperty.h"
#include "../../raster/Raster.h"
#include "../Exception.h"
#include "../utils/Utils.h"

#include "DataSetItem.h"
#include "DataSetType.h"

// STL
#include <cassert>

te::da::DataSetItem::DataSetItem(const std::vector<std::string>& vecNames, const std::vector<int>& vecTypes)
  : m_vecNames(vecNames)
  , m_vecTypes(vecTypes)
{
  m_data.resize(m_vecNames.size(), nullptr);
}

te::da::DataSetItem::DataSetItem(const te::da::DataSet* dataSet)
{
  if (dataSet == nullptr)
    return;

  std::size_t numProperties = dataSet->getNumProperties();

  m_vecNames.reserve(numProperties);
  m_vecTypes.reserve(numProperties);  
  for (std::size_t i = 0; i < numProperties; ++i)
  {
    m_vecNames.push_back(dataSet->getPropertyName(i));
    m_vecTypes.push_back(dataSet->getPropertyDataType(i));
  }

  m_data.resize(numProperties, nullptr);
}

te::da::DataSetItem::DataSetItem(const te::da::DataSetType* dataSetType)
{
  if(dataSetType == nullptr)
    return;

  std::size_t numProperties = dataSetType->getProperties().size();

  m_vecNames.reserve(numProperties);
  m_vecTypes.reserve(numProperties);
  for (std::size_t i = 0; i < numProperties; ++i)
  {
    m_vecNames.push_back(dataSetType->getProperty(i)->getName());
    m_vecTypes.push_back(dataSetType->getProperty(i)->getType());
  }

  m_data.resize(numProperties, nullptr);
}

te::da::DataSetItem::DataSetItem(const DataSetItem& rhs)
{
  m_vecNames = rhs.m_vecNames;
  m_vecTypes = rhs.m_vecTypes;
  m_data.reserve(rhs.getNumProperties());

  for (std::size_t i = 0; i < rhs.getNumProperties(); ++i)
  {
    if (rhs.isNull(i) == true)
    {
      m_data.push_back(nullptr);
    }
    else
    {
      m_data.push_back(rhs.getValue(i).release());
    }
  }
}

te::da::DataSetItem::~DataSetItem()
{
  m_data.clear();
}


te::da::DataSetItem* te::da::DataSetItem::createFromCurrent(const te::da::DataSet* dataSet)
{
  if (dataSet == nullptr)
    return nullptr;

  std::size_t numProperties = dataSet->getNumProperties();

  std::vector<std::string> vecNames;
  std::vector<int> vecTypes;

  for (std::size_t i = 0; i < numProperties; ++i)
  {
    vecNames.push_back(dataSet->getPropertyName(i));
    vecTypes.push_back(dataSet->getPropertyDataType(i));
  }

  te::da::DataSetItem* newItem = new te::da::DataSetItem(vecNames, vecTypes);
  newItem->m_data.resize(numProperties, nullptr);
  for (std::size_t i = 0; i < numProperties; ++i)
  {
    if (dataSet->isNull(i) == false)
    {
      newItem->m_data.replace(i, dataSet->getValue(i).release());
    }
  }

  return newItem;
}

te::da::DataSetItem& te::da::DataSetItem::operator=(const DataSetItem& rhs)
{
  assert(getNumProperties() == rhs.getNumProperties());

  if (this != &rhs)
  {
    m_vecNames = rhs.m_vecNames;
    m_vecTypes = rhs.m_vecTypes;
    m_data = rhs.m_data;
  }

  return *this;
}

te::da::DataSetItem* te::da::DataSetItem::clone() const
{
  return new DataSetItem(*this);
}

const std::vector<std::string>& te::da::DataSetItem::getPropertyNames() const
{
  return m_vecNames;
}

const std::vector<int>& te::da::DataSetItem::getPropertyTypes() const
{
  return m_vecTypes;
}

std::size_t te::da::DataSetItem::getNumProperties() const
{
  return m_vecNames.size();
}

int te::da::DataSetItem::getPropertyDataType(std::size_t pos) const
{
  return m_vecTypes.at(pos);
}

const std::string& te::da::DataSetItem::getPropertyName(std::size_t pos) const
{
  return m_vecNames.at(pos);
}

std::size_t te::da::DataSetItem::getPropertyPos(const std::string& name) const
{
  for (std::size_t i = 0; i < m_vecNames.size(); ++i)
  {
    if (te::common::IsStringEqual(m_vecNames.at(i), name, false) == true)
    {
      return i;
    }
  }

  return std::string::npos;
}

char te::da::DataSetItem::getChar(std::size_t i) const
{
  assert(isNull(i) == false);
  return static_cast<const te::dt::Char*>(&m_data[i])->getValue();
}

unsigned char te::da::DataSetItem::getUChar(std::size_t i) const
{
  assert(isNull(i) == false);
  return static_cast<const te::dt::UChar*>(&m_data[i])->getValue();
}

boost::int16_t te::da::DataSetItem::getInt16(std::size_t i) const
{
  assert(isNull(i) == false);
  return static_cast<const te::dt::Int16*>(&m_data[i])->getValue();
}

boost::int32_t te::da::DataSetItem::getInt32(std::size_t i) const
{
  assert(isNull(i) == false);
  return static_cast<const te::dt::Int32*>(&m_data[i])->getValue();
}

boost::int64_t te::da::DataSetItem::getInt64(std::size_t i) const
{
  assert(isNull(i) == false);
  return static_cast<const te::dt::Int64*>(&m_data[i])->getValue();
}

bool te::da::DataSetItem::getBool(std::size_t i) const
{
  assert(isNull(i) == false);
  return static_cast<const te::dt::Boolean*>(&m_data[i])->getValue();
}

float te::da::DataSetItem::getFloat(std::size_t i) const
{
  assert(isNull(i) == false);
  return static_cast<const te::dt::Float*>(&m_data[i])->getValue();
}

double te::da::DataSetItem::getDouble(std::size_t i) const
{
  assert(isNull(i) == false);
  return static_cast<const te::dt::Double*>(&m_data[i])->getValue();
}

std::string te::da::DataSetItem::getNumeric(std::size_t i) const
{
  assert(isNull(i) == false);
  return static_cast<const te::dt::Numeric*>(&m_data[i])->getValue();
}

std::string te::da::DataSetItem::getString(std::size_t i) const
{
  assert(isNull(i) == false);
  return static_cast<const te::dt::String*>(&m_data[i])->getValue();
}

std::unique_ptr<te::dt::ByteArray> te::da::DataSetItem::getByteArray(std::size_t i) const
{
  assert(isNull(i) == false);
  return std::unique_ptr<te::dt::ByteArray>(static_cast<te::dt::ByteArray*>(m_data[i].clone()));
}

std::unique_ptr<te::gm::Geometry> te::da::DataSetItem::getGeometry(std::size_t i) const
{
  assert(isNull(i) == false);
  return std::unique_ptr<te::gm::Geometry>(static_cast<te::gm::Geometry*>(m_data[i].clone()));
}

std::unique_ptr<te::rst::Raster> te::da::DataSetItem::getRaster(std::size_t i) const
{
  assert(isNull(i) == false);
  return std::unique_ptr<te::rst::Raster>(static_cast<te::rst::Raster*>(m_data[i].clone()));
}

std::unique_ptr<te::dt::DateTime> te::da::DataSetItem::getDateTime(std::size_t i) const
{
  assert(isNull(i) == false);
  return std::unique_ptr<te::dt::DateTime>(static_cast<te::dt::DateTime*>(m_data[i].clone()));
}

std::unique_ptr<te::dt::AbstractData> te::da::DataSetItem::getValue(std::size_t i) const
{
  assert(isNull(i) == false);

  switch(getPropertyDataType(i))
  {
    case te::dt::CHAR_TYPE:
      return std::unique_ptr<te::dt::AbstractData>(new te::dt::Char(getChar(i)));

    case te::dt::UCHAR_TYPE:
      return std::unique_ptr<te::dt::AbstractData>(new te::dt::UChar(getUChar(i)));

    case te::dt::INT16_TYPE:
      return std::unique_ptr<te::dt::AbstractData>(new te::dt::Int16(getInt16(i)));

    case te::dt::UINT16_TYPE:
      return std::unique_ptr<te::dt::AbstractData>(new te::dt::UInt16(getInt16(i)));

    case te::dt::INT32_TYPE:
      return std::unique_ptr<te::dt::AbstractData>(new te::dt::Int32(getInt32(i)));

    case te::dt::UINT32_TYPE:
      return std::unique_ptr<te::dt::AbstractData>(new te::dt::UInt32(getInt32(i)));

    case te::dt::INT64_TYPE:
      return std::unique_ptr<te::dt::AbstractData>(new te::dt::Int64(getInt64(i)));

    case te::dt::UINT64_TYPE:
      return std::unique_ptr<te::dt::AbstractData>(new te::dt::UInt64(getInt64(i)));

    case te::dt::BOOLEAN_TYPE:
      return std::unique_ptr<te::dt::AbstractData>(new te::dt::Boolean(getBool(i)));

    case te::dt::FLOAT_TYPE:
      return std::unique_ptr<te::dt::AbstractData>(new te::dt::Float(getFloat(i)));

    case te::dt::DOUBLE_TYPE:
      return std::unique_ptr<te::dt::AbstractData>(new te::dt::Double(getDouble(i)));

    case te::dt::NUMERIC_TYPE:
      return std::unique_ptr<te::dt::AbstractData>(new te::dt::Numeric(getNumeric(i)));

    case te::dt::STRING_TYPE:
      return std::unique_ptr<te::dt::AbstractData>(new te::dt::String(getString(i)));

    case te::dt::BYTE_ARRAY_TYPE:
      return getByteArray(i);

    case te::dt::GEOMETRY_TYPE:
      return getGeometry(i);

    case te::dt::DATETIME_TYPE:
      return getDateTime(i);

    case te::dt::RASTER_TYPE:
      return getRaster(i);

    default:
      return std::unique_ptr<te::dt::AbstractData>();
  }
}

bool te::da::DataSetItem::isNull(std::size_t i) const
{
  return boost::is_null(m_data.begin() + i);
}
