 /*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/memory/DataSetItem.h

  \brief An implementation of the DatasetItem class for the TerraLib In-Memory Data Access driver.
*/

#ifndef __TERRALIB_DATAACCESS_INTERNAL_DATASETITEM_H
#define __TERRALIB_DATAACCESS_INTERNAL_DATASETITEM_H

// TerraLib
#include "../Config.h"

#include <boost/cstdint.hpp>
#include <boost/noncopyable.hpp>
#include <boost/ptr_container/ptr_vector.hpp>
#include <boost/shared_ptr.hpp>

#include <string>
#include <vector>

// Forward declarations
namespace te { namespace dt { class AbstractData;  class Array; class ByteArray; class DateTime; } }
namespace te { namespace gm { class Geometry; } }
namespace te { namespace rst { class Raster; } }

namespace te
{
  namespace da
  {
    // Forward declaration
    class DataSet;
    class DataSetType;

    /*!
      \class DataSetItem

      \brief An implementation of the DatasetItem class for the TerraLib In-Memory Data Access driver.
    */
    class TEDATAACCESSEXPORT DataSetItem
    {
    public:

      /*!
        \brief It creates a new item having the same schema as the given column names and types

        \param vecNames The names of the properties of this item
        \param vecTypes The types of the properties of this item
      */
      explicit DataSetItem(const std::vector<std::string>& vecNames, const std::vector<int>& vecTypes);

      /*!
        \brief It creates a new item having the same schema as the parent dataset.

        \param parent The dataset to which this item is associated.

        \note The dataset item will NOT take the ownership of the given pointer.
      */
      explicit DataSetItem(const te::da::DataSet* dataSet);

      /*!
        \brief It creates a new item having the same schema as the dataset type.

        \param parent The dataset to which this item is associated.

        \note The dataset item will NOT take the ownership of the given pointer.
      */
      explicit DataSetItem(const te::da::DataSetType* dataSetType);

      /*!
        \brief It creates a new item by cloning the values in the source item (rhs).

        \param rhs The right-hand-side object to copy its values.
      */
      explicit DataSetItem(const DataSetItem& rhs);

      /*! \brief Destructor. */
      virtual ~DataSetItem();

      /*!
        \brief Creates a dataSetItem based on the current row of the givem DataSet
      */
      static te::da::DataSetItem* createFromCurrent(const te::da::DataSet* dataSet);

      /*!
        \brief Assignment operator.

        It copies the values from the rhs item.

        \param rhs The right-hand-side ibject to copy its values.

        \return A reference to this item.

        \pre The rhs item must be of a compatible type with the item being assigned.
      */

      DataSetItem& operator=(const DataSetItem& rhs);

      /*!
        \brief It returns a clone of the DataSetItem.
      */
      virtual DataSetItem* clone() const;

      /*!
        \brief It returns the number of properties.
      */
      std::size_t getNumProperties() const;

      /*!
        \brief It returns the name of properties.
      */
      const std::vector<std::string>& getPropertyNames() const;

      /*!
        \brief It returns the types of properties.
      */
      const std::vector<int>& getPropertyTypes() const;

      /*!
        \brief It returns the type of the pos-th property.
      */
      int getPropertyDataType(std::size_t pos) const;

      /*!
        \brief It returns the name of the pos-th property.
      */
      const std::string& getPropertyName(std::size_t pos) const;

      /*!
        \brief It returns the name of the pos-th property.
      */
      std::size_t getPropertyPos(const std::string& name) const;

      /*!
        \brief It returns the value of the i-th property.
      */
      char getChar(std::size_t i) const;

      /*!
        \brief It returns the value of the i-th property.
      */
      unsigned char getUChar(std::size_t i) const;

      /*!
        \brief It returns the value of the i-th property.
      */
      boost::int16_t getInt16(std::size_t i) const;

      /*!
        \brief It returns the value of the i-th property.
      */
      boost::int32_t getInt32(std::size_t i) const;

      /*!
        \brief It returns the value of the i-th property.
      */
      boost::int64_t getInt64(std::size_t i) const;

      /*!
        \brief It returns the value of the i-th property.
      */
      bool getBool(std::size_t i) const;

      /*!
        \brief It returns the value of the i-th property.
      */
      float getFloat(std::size_t i) const;

      /*!
        \brief It returns the value of the i-th property.
      */
      double getDouble(std::size_t i) const;

      /*!
        \brief It returns the value of the i-th property.
      */
      std::string getNumeric(std::size_t i) const;

      /*!
        \brief It returns the value of the i-th property.
      */
      std::string getString(std::size_t i) const;

      /*!
        \brief It returns the value of the i-th property.
      */
      std::unique_ptr<te::dt::ByteArray> getByteArray(std::size_t i) const;

      /*!
        \brief It returns the value of the i-th property.
      */
      std::unique_ptr<te::gm::Geometry> getGeometry(std::size_t i) const;

      /*!
        \brief It returns the value of the i-th property.
      */
      std::unique_ptr<te::rst::Raster> getRaster(std::size_t i) const;

      /*!
        \brief It returns the value of the i-th property.
      */
      std::unique_ptr<te::dt::DateTime> getDateTime(std::size_t i) const;

      /*!
        \brief It returns the value of the i-th property.
      */
      std::unique_ptr<te::dt::AbstractData> getValue(std::size_t i) const;

      /*!
        \brief Checks if the value in the given index is null
      */
      bool isNull(std::size_t i) const;

    protected:

      std::vector<std::string> m_vecNames;  //!< Vector containing the names of the attributes of the dataSet item
      std::vector<int> m_vecTypes;          //!< Vector containing the types of the attributes of the dataSet item
      boost::ptr_vector<boost::nullable<te::dt::AbstractData> > m_data;  //!< The data values of the 
    };
  }// end namespace da
}// end namespace te

#endif  // __TERRALIB_DATAACCESS_INTERNAL_DATASETITEM_H
