/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/dataaccess/dataset/ForeignKey.cpp

  \brief It models a foreign key constraint for a DataSetType.
*/

// TerraLib
#include "../../common/STLUtils.h"
#include "../../datatype/Property.h"
#include "DataSetType.h"
#include "ForeignKey.h"

te::da::ForeignKey::ForeignKey(unsigned int id)
  : Constraint(id),
    m_onDelete(NO_ACTION),
    m_onUpdate(NO_ACTION),
    m_refDt(nullptr)    
{
}

te::da::ForeignKey::ForeignKey(const std::string& name,
                               unsigned int id)
  : Constraint(name, id),
    m_onDelete(NO_ACTION),
    m_onUpdate(NO_ACTION),
    m_refDt(nullptr)
{
}

te::da::ForeignKey::ForeignKey(const ForeignKey& rhs)
  : Constraint(rhs.getName(), 0)
{

  m_onDelete = rhs.m_onDelete;

  m_onUpdate = rhs.m_onUpdate;

  m_refDt = rhs.m_refDt;

  for (std::size_t t = 0; t < rhs.m_properties.size(); ++t)
  {
    m_properties.push_back(rhs.m_properties[t]->clone());
  }

  for (std::size_t t = 0; t < rhs.m_refProperties.size(); ++t)
  {
    m_refProperties.push_back(rhs.m_refProperties[t]->clone());
  }
}

te::da::ForeignKey::~ForeignKey()
{
  te::common::FreeContents(m_properties);
  m_properties.clear();

  te::common::FreeContents(m_refProperties);
  m_refProperties.clear();
}

te::da::ForeignKey& te::da::ForeignKey::operator=(const ForeignKey& rhs)
{
  if(this != &rhs)
  {
    Constraint::operator=(rhs);

    m_onDelete = rhs.m_onDelete;

    m_onUpdate = rhs.m_onUpdate;

    m_refDt = rhs.m_refDt;

    for (std::size_t t = 0; t < rhs.m_properties.size(); ++t)
    {
      m_properties.push_back(rhs.m_properties[t]->clone());
    }

    for (std::size_t t = 0; t < rhs.m_refProperties.size(); ++t)
    {
      m_refProperties.push_back(rhs.m_refProperties[t]->clone());
    }
  }

  return *this;
}

bool te::da::ForeignKey::has(const std::string& propertyName)
{
  size_t size = m_properties.size();

  for (size_t i = 0; i < size; ++i)
  {
    if (m_properties[i]->getName() == propertyName)
    {
      return true;
    }
  }

  return false;
}

bool te::da::ForeignKey::isReferenced(const std::string& propertyName)
{
  size_t size = m_refProperties.size();

  for (size_t i = 0; i < size; ++i)
  {
    if (m_refProperties[i]->getName() == propertyName)
    {
      return true;
    }
  }

  return false;
}

void te::da::ForeignKey::replace(const std::string& propName, const te::dt::Property* pp)
{
  {
    std::size_t size = m_properties.size();

    for (std::size_t i = 0; i < size; ++i)
    {
      if (m_properties[i]->getName() == propName)
      {
        delete  m_properties[i];

        m_properties[i] = pp->clone();

        break;
      }
    }
  }

  {
    std::size_t size = m_refProperties.size();

    for (std::size_t i = 0; i < size; ++i)
    {
      if (m_refProperties[i]->getName() == propName)
      {
        delete  m_refProperties[i];

        m_refProperties[i] = pp->clone();

        break;
      }
    }
  }
}

te::da::Constraint* te::da::ForeignKey::clone()
{
  return new ForeignKey(*this);
}

