#include "BasicQueries.h"

#include "../../common/Exception.h"
#include "../../core/translator/Translator.h"
#include "../../dataaccess/dataset/DataSetType.h"
#include "../../datatype/AbstractData.h"
#include "../../datatype/SimpleData.h"
#include "../../geometry/GeometryProperty.h"

#include "../query/And.h"
#include "../query/BinaryOp.h"
#include "../query/DataSetName.h"
#include "../query/Distinct.h"
#include "../query/EqualTo.h"
#include "../query/Field.h"
#include "../query/Fields.h"
#include "../query/From.h"
#include "../query/FunctionNames.h"
#include "../query/In.h"
#include "../query/IsNull.h"
#include "../query/Literal.h"
#include "../query/LiteralEnvelope.h"
#include "../query/LiteralInt32.h"
#include "../query/LiteralString.h"
#include "../query/PropertyName.h"
#include "../query/Select.h"
#include "../query/ST_Intersects.h"
#include "../query/Where.h"
#include "Utils.h"


std::unique_ptr<te::da::Select> te::da::BasicQueries::GetQuery(bool distinct, const std::vector<std::string>& vecColumns, const std::string& tableName, const std::vector<std::string>& vecWhereColumns, const std::vector<te::dt::AbstractData*>& vecWhereValues)
{
  std::unique_ptr<te::da::Fields> fields = GetFields(vecColumns);

  te::da::From* from = new te::da::From();
  from->push_back(new te::da::DataSetName(tableName));

  te::da::Expression* whereExpression = nullptr;
  if (vecWhereColumns.empty() == false)
  {
    for (std::size_t i = 0; i < vecWhereColumns.size(); ++i)
    {
      std::string whereColumnName = vecWhereColumns.at(i);
      const te::dt::AbstractData* whereColumnValue = vecWhereValues.at(i);

      te::da::Expression* equal = nullptr;
      if (whereColumnValue == nullptr)
      {
        equal = new te::da::IsNull(new te::da::PropertyName(whereColumnName));
      }
      else
      {
        equal = new te::da::EqualTo(new te::da::PropertyName(whereColumnName), te::da::GetLiteralFromAbstractData(whereColumnValue));
      }

      
      if (whereExpression == nullptr)
      {
        whereExpression = equal;
      }
      else
      {
        te::da::And* andExpression = new te::da::And(whereExpression, equal);
        whereExpression = andExpression;
      }
    }
  }

  std::unique_ptr<te::da::Select> selectPtr(new te::da::Select());
  if (distinct == true)
  {
    te::da::Distinct* distinctClause = new te::da::Distinct;
    selectPtr->setDistinct(distinctClause);
  }

  selectPtr->setFields(fields.release());
  selectPtr->setFrom(from);

  if (whereExpression != nullptr)
  {
    te::da::Where* where = new te::da::Where(whereExpression);
    selectPtr->setWhere(where);
  }

  return selectPtr;
}

std::unique_ptr<te::da::Select> te::da::BasicQueries::GetQuery(bool distinct, const std::vector<std::string>& vecColumns, const std::string& tableName)
{
  std::vector<std::string> vecWhereColumns;
  std::vector<te::dt::AbstractData*> vecWhereValues;
  return GetQuery(distinct, vecColumns, tableName, vecWhereColumns, vecWhereValues);
}

std::unique_ptr<te::da::Fields> te::da::BasicQueries::GetFields(const te::da::DataSetType* dataSetType)
{
  std::vector<std::string> vecColumns;
  for (std::size_t i = 0; i < dataSetType->size(); ++i)
  {
    vecColumns.push_back(dataSetType->getProperty(i)->getName());
  }

  return GetFields(vecColumns);
}

std::unique_ptr<te::da::Fields> te::da::BasicQueries::GetFields(const std::vector<std::string>& vecColumns)
{
  std::unique_ptr<te::da::Fields> fields(new te::da::Fields());
  for (const std::string& columnName : vecColumns)
  {
    fields->push_back(new te::da::Field(columnName));
  }

  return fields;
}

std::unique_ptr<te::da::From> te::da::BasicQueries::GetFrom(const te::da::DataSetType* dataSetType)
{
  std::unique_ptr<te::da::From> from(new te::da::From());
  from->push_back(new te::da::DataSetName(dataSetType->getName()));

  return from;
}

std::unique_ptr<te::da::Expression> te::da::BasicQueries::GetRelationalExpression(const std::string& propertyName, int propertyValue, const std::string& relationalOperator)
{
  te::da::PropertyName* pName = new te::da::PropertyName(propertyName);

  te::da::Literal* literal = new te::da::LiteralInt32(propertyValue);

  std::unique_ptr<te::da::Expression> expression(new te::da::BinaryOp(relationalOperator, pName, literal));
  return expression;
}

std::unique_ptr<te::da::Expression> te::da::BasicQueries::GetRelationalExpression(const std::string& propertyName, const std::string& propertyValue, const std::string& relationalOperator)
{
  te::da::PropertyName* pName = new te::da::PropertyName(propertyName);

  te::da::Literal* literal = new te::da::LiteralString(propertyValue);

  std::unique_ptr<te::da::Expression> expression(new te::da::BinaryOp(relationalOperator, pName, literal));
  return expression;
}

std::unique_ptr<te::da::Expression> te::da::BasicQueries::GetIntersectsExpression(const te::da::DataSetType* dataSetType, const te::gm::Envelope& envelope)
{
  te::gm::GeometryProperty* geometryProperty = te::da::GetFirstGeomProperty(dataSetType);
  if (geometryProperty == nullptr)
  {
    throw te::common::Exception(TE_TR("te::da::BasicQueries::GetIntersectsExpression - The given dataSetType does not have any spatial property"));
  }

  te::da::PropertyName* geometryPropertyName = new te::da::PropertyName(geometryProperty->getName());
  te::da::LiteralEnvelope* literalEnvelope = new te::da::LiteralEnvelope(envelope, geometryProperty->getSRID());

  std::unique_ptr<te::da::Expression> expression(new te::da::ST_Intersects(geometryPropertyName, literalEnvelope));

  return expression;
}

std::unique_ptr<te::da::Expression> te::da::BasicQueries::GetInExpression(const std::string& propertyName, const std::vector<int>& vecValues)
{
  te::da::In* in = new te::da::In(propertyName);
  for (int value : vecValues)
  {
    in->add(new te::da::LiteralInt32(value));
  }

  return std::unique_ptr<te::da::Expression>(in);
}

std::unique_ptr<te::da::Expression> te::da::BasicQueries::GetInExpression(const std::string& propertyName, const std::vector<std::string>& vecValues)
{
  te::da::In* in = new te::da::In(propertyName);
  for (const std::string& value : vecValues)
  {
    in->add(new te::da::LiteralString(value));
  }

  return std::unique_ptr<te::da::Expression>(in);
}

void te::da::BasicQueries::AppendExpressionToWhereClause(te::da::Select* select, te::da::Expression* expression)
{
  te::da::Where* whereClause = select->getWhere();
  if (whereClause != nullptr)
  {
    // Original restriction expression
    std::unique_ptr<te::da::Expression> currentExpression(whereClause->getExp()->clone());

    // The final restriction: original restriction expression + the oids restriction
    whereClause->setExp(new te::da::And(currentExpression.release(), expression));
  }
  else
  {
    whereClause = new te::da::Where(expression);
    select->setWhere(whereClause);
  }
}

std::unique_ptr<te::da::Where> te::da::BasicQueries::GetWhereEqualTo(const std::string& propertyName, int propertyValue)
{
  std::unique_ptr<te::da::Expression> expression = GetRelationalExpression(propertyName, propertyValue, te::da::FunctionNames::sm_EqualTo);
  std::unique_ptr<te::da::Where> where(new te::da::Where(expression.release()));
  return where;
}

std::unique_ptr<te::da::Where> te::da::BasicQueries::GetWhereEqualTo(const std::string& propertyName, const std::string& propertyValue)
{
  std::unique_ptr<te::da::Expression> expression = GetRelationalExpression(propertyName, propertyValue, te::da::FunctionNames::sm_EqualTo);
  std::unique_ptr<te::da::Where> where(new te::da::Where(expression.release()));
  return where;
}

std::unique_ptr<te::da::Select> te::da::BasicQueries::GetSimpleQuery(const te::da::DataSetType* dataSetType)
{
  std::unique_ptr<te::da::Select> selectPtr(new te::da::Select());

  selectPtr->setFields(GetFields(dataSetType).release());
  selectPtr->setFrom(GetFrom(dataSetType).release());

  return selectPtr;
}

std::unique_ptr<te::da::Select> te::da::BasicQueries::GetSimpleQueryEqualTo(const te::da::DataSetType* dataSetType, const std::string& propertyName, int propertyValue)
{
  std::unique_ptr<te::da::Select> select = te::da::BasicQueries::GetSimpleQuery(dataSetType);
  select->setWhere(te::da::BasicQueries::GetWhereEqualTo(propertyName, propertyValue).release());

  return select;
}

std::unique_ptr<te::da::Select> te::da::BasicQueries::GetSimpleQueryEqualTo(const te::da::DataSetType* dataSetType, const std::string& propertyName, const std::string& propertyValue)
{
  std::unique_ptr<te::da::Select> select = te::da::BasicQueries::GetSimpleQuery(dataSetType);
  select->setWhere(te::da::BasicQueries::GetWhereEqualTo(propertyName, propertyValue).release());

  return select;
}

std::unique_ptr<te::da::Select> te::da::BasicQueries::GetSimpleQueryIntersects(const te::da::DataSetType* dataSetType, const te::gm::Envelope& envelope)
{
  std::unique_ptr<te::da::Expression> expression = te::da::BasicQueries::GetIntersectsExpression(dataSetType, envelope);

  std::unique_ptr<te::da::Select> select = te::da::BasicQueries::GetSimpleQuery(dataSetType);
  AppendExpressionToWhereClause(select.get(), expression.release());

  return select;
}

