/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
\file dataaccess/utils/DataAccess.cpp

\brief Class used to centralize and control access to data in terralib.
       It aims to create a high-level interface for simplifying the
       concepts defined in the terralib DataAccess module.
*/

#include "DataAccess.h"
#include "Utils.h"

// TerraLib
#include "../../core/translator/Translator.h"
#include "../../common/Exception.h"
#include "../../common/StringUtils.h"
#include "../../geometry/Envelope.h"
#include "../../geometry/GeometryProperty.h"
#include "../datasource/DataSource.h"
#include "../datasource/DataSourceFactory.h"
#include "../datasource/DataSourceTransactor.h"
#include "../dataset/DataSet.h"
#include "../dataset/DataSetAdapter.h"
#include "../dataset/DataSetType.h"
#include "../dataset/DataSetTypeConverter.h"
#include "../dataset/ObjectIdSet.h"
#include "../query_h.h"
#include "BasicQueries.h"

te::da::DataAccess::DataAccess(const te::core::URI& uri, const std::string& dataSourceDriver, const std::string& dataSetName):
  m_dataSetName(dataSetName)
{
  m_dataSource = te::da::DataSourceFactory::make(dataSourceDriver, uri);

  if (m_dataSource.get() == nullptr)
    throw te::common::Exception(TE_TR("Could not create the required data source instance!"));

  if (m_dataSource->isOpened() == false)
  {
    m_dataSource->open();
  }

  if (dataSetName.empty() == false)
  {
    DataSourceTransactorPtr transactor = getTransactor();
    m_dataSetType = transactor->getDataSetType(dataSetName);
  }
}

te::da::DataAccess::DataAccess(te::da::DataSourcePtr dataSource, const std::string& dataSetName)
  : m_dataSource(dataSource)
  , m_dataSetName(dataSetName)
{
  if (m_dataSource->isOpened() == false)
  {
    m_dataSource->open();
  }

  if (dataSetExists() == true)
  {
    DataSourceTransactorPtr transactor = getTransactor();
    m_dataSetType = transactor->getDataSetType(dataSetName);
  }
}

te::da::DataAccess::DataAccess(te::da::DataSourcePtr dataSource, const te::da::DataSetType* dataSetType)
  : m_dataSource(dataSource)
  , m_dataSetName(dataSetType->getName())
{
  if (m_dataSource->isOpened() == false)
  {
    m_dataSource->open();
  }
  m_dataSetType.reset(static_cast<te::da::DataSetType*>(dataSetType->clone()));
}

te::da::DataAccess::DataAccess(const te::da::DataAccess& rhs)
{
  this->m_dataSetName = rhs.m_dataSetName;
  this->m_dataSource = rhs.m_dataSource;
  if (rhs.m_dataSetType != nullptr)
  {
    this->m_dataSetType.reset(static_cast<te::da::DataSetType*>(rhs.m_dataSetType->clone()));
  }
  
  if (rhs.m_dataSetQuery != nullptr)
  {
    this->m_dataSetQuery.reset(static_cast<te::da::Select*>(rhs.m_dataSetQuery->clone()));
  }
  
  if (rhs.m_objectIdSetFilter != nullptr)
  {
    this->m_objectIdSetFilter.reset(static_cast<te::da::ObjectIdSet*>(rhs.m_objectIdSetFilter->clone()));
  }
}

te::da::DataAccess::~DataAccess()
{
  
}

te::dt::AbstractData* te::da::DataAccess::clone() const
{
  te::dt::AbstractData* newDataAccess = new te::da::DataAccess(*this);
  return newDataAccess;
}

void te::da::DataAccess::beginTransaction()
{
  if (m_dataSource == nullptr)
  {
    return;
  }

  te::da::DataSourceTransactorPtr transactor = getTransactor();
  transactor->begin();
}

void te::da::DataAccess::commitTransaction()
{
  if (m_dataSource == nullptr)
  {
    return;
  }

  te::da::DataSourceTransactorPtr transactor = getTransactor();
  transactor->commit();
}

void te::da::DataAccess::rollbackTransaction()
{
  if (m_dataSource == nullptr)
  {
    return;
  }

  te::da::DataSourceTransactorPtr transactor = getTransactor();
  transactor->rollBack();
}

void te::da::DataAccess::flush()
{
  if (m_dataSource == nullptr)
  {
    return;
  }

  if (m_dataSource->isOpened() == true)
  {
    if (m_transactor->isInTransaction() == false)
    {
      m_dataSource->close();
      m_dataSource->open();
    }
  }
}

int te::da::DataAccess::getTypeCode() const
{
  return te::dt::DATAACCESS_TYPE;
}

std::string te::da::DataAccess::toString() const
{
  return getDataSourceConnectionInfo();
}


bool te::da::DataAccess::dataSetExists() const
{
  if (m_dataSetName.empty() == true)
  {
    return false;
  }

  DataSourceTransactorPtr transactor = getTransactor();
  bool exists = transactor->dataSetExists(m_dataSetName);
  return exists;
}

std::string te::da::DataAccess::getDataSourceConnectionInfo() const
{
  if (m_dataSource.get())
  {
    const te::core::URI uri(m_dataSource->getConnectionInfo());
    std::string uriTxt;

    uriTxt += "Scheme: " + uri.scheme() + " ";
    uriTxt += "User: " + uri.user() + " ";
    uriTxt += "Host: " + uri.host() + " ";
    uriTxt += "Port: " + uri.port() + " ";
    uriTxt += "Path: " + uri.path() + " ";

    return uriTxt;
  }

  return std::string();
}

void te::da::DataAccess::setConverter(te::da::DataSetTypeConverter* converter)
{
  m_dataSetTypeConverter.reset(converter);
}

te::da::DataSetTypeConverter* te::da::DataAccess::getConverter() const
{
  return m_dataSetTypeConverter.get();
}

std::string te::da::DataAccess::getDataSetName() const
{
  return m_dataSetName;
}

const te::da::DataSetType* te::da::DataAccess::getDataSetType() const
{
  if (m_dataSetTypeConverter != nullptr)
  {
    return m_dataSetTypeConverter->getResult();
  }
  return m_dataSetType.get();
}

void te::da::DataAccess::setQuery(te::da::Select* select)
{
  m_dataSetQuery.reset(select);
}

te::da::Select* te::da::DataAccess::getQuery() const
{
  return m_dataSetQuery.get();
}

std::string te::da::DataAccess::getStrQuery() const
{
  if (m_dataSetQuery.get() && m_dataSource.get())
  {
    std::string strSelect;

    te::da::SQLVisitor visitor(*m_dataSource->getDialect(), strSelect);
    m_dataSetQuery->accept(visitor);

    return strSelect;
  }

  return std::string();
}

std::unique_ptr<te::da::DataSet> te::da::DataAccess::getDataSet() const
{
  return getDataSet(nullptr);
}

std::unique_ptr<te::da::DataSet> te::da::DataAccess::getDataSet(const te::da::Expression* filter) const
{
  //we first handle the filter concatenation, if needed
  std::unique_ptr<te::da::Expression> filterExpression;
  if (filter != nullptr)
  {
    filterExpression.reset(filter->clone());
  }

  if (m_objectIdSetFilter != nullptr)
  {
    if (filterExpression == nullptr)
    {
      //if the restriction is empty, we just assign the expression
      filterExpression.reset(m_objectIdSetFilter->getExpression());
    }
    else
    {
      //if the restriction is not empty, we concatenate the expression
      filterExpression.reset(new te::da::And(filterExpression.release(), m_objectIdSetFilter->getExpression()));
    }
  }

  //then we handle the query creation, if needed
  std::unique_ptr<te::da::Select> select;

  //if we are based on a query, we concatenate the filter restriction
  if (m_dataSetQuery != nullptr)
  {
    select.reset(static_cast<te::da::Select*>(m_dataSetQuery->clone()));
    if (filterExpression != nullptr)
    {
      te::da::BasicQueries::AppendExpressionToWhereClause(select.get(), filterExpression.get());
    }
  }
  else if (filterExpression != nullptr)
  {
    select = te::da::BasicQueries::GetSimpleQuery(getDataSetType());
    if (filterExpression != nullptr)
    {
      te::da::BasicQueries::AppendExpressionToWhereClause(select.get(), filterExpression.release());
    }
  }

  //then we run the query or just aquire the non filtered dataSet
  DataSourceTransactorPtr transactor = getTransactor();
  std::unique_ptr<te::da::DataSet> dataset;

  if (select != nullptr)
  {
    dataset = transactor->query(*select.get());
  }
  else
  {
    dataset = transactor->getDataSet(m_dataSetName);
  }

  if (m_dataSetTypeConverter.get())
  {
    te::da::DataSetAdapter* dsAdapted = te::da::CreateAdapter(dataset.release(), m_dataSetTypeConverter.get(), true);

    std::unique_ptr<te::da::DataSet> result(dsAdapted);

    return result;
  }

  return dataset;
}

te::da::DataSourcePtr te::da::DataAccess::getDataSource() const
{
  return m_dataSource;
}

std::unique_ptr<te::da::DataSet> te::da::DataAccess::query(const te::da::Select& select)
{
  DataSourceTransactorPtr transactor = getTransactor();

  std::unique_ptr<te::da::DataSet> dataset = transactor->query(select);

  if (m_dataSetTypeConverter.get())
  {
    te::da::DataSetAdapter* dsAdapted = te::da::CreateAdapter(dataset.release(), m_dataSetTypeConverter.get(), true);

    std::unique_ptr<te::da::DataSet> result(dsAdapted);

    return result;
  }

  return dataset;
}

std::unique_ptr<te::da::DataSet> te::da::DataAccess::query(const std::string& query)
{
  DataSourceTransactorPtr transactor = getTransactor();

  std::unique_ptr<te::da::DataSet> dataset = transactor->query(query);

  if (m_dataSetTypeConverter.get())
  {
    te::da::DataSetAdapter* dsAdapted = te::da::CreateAdapter(dataset.release(), m_dataSetTypeConverter.get(), true);

    std::unique_ptr<te::da::DataSet> result(dsAdapted);

    return result;
  }

  return dataset;
}

void te::da::DataAccess::execute(const std::string& query)
{
  DataSourceTransactorPtr transactor = getTransactor();
  transactor->execute(query);
}

void te::da::DataAccess::createDataSet(te::da::DataSetType* dataSetType)
{
  std::unique_ptr<te::da::DataSetType> adjustedDataSetType = te::da::PrepareDateSetType(dataSetType, &m_dataSource->getCapabilities());
  
  DataSourceTransactorPtr transactor = getTransactor();

  //create output data
  std::map<std::string, std::string> options;

  try
  {
    transactor->begin();

    transactor->createDataSet(adjustedDataSetType.get(), options);

    transactor->commit();
  }
  catch (te::common::Exception& e)
  {
    transactor->rollBack();
    throw e;
  }
  catch (std::exception& e)
  {
    transactor->rollBack();
    throw e;
  }

  m_dataSetName = dataSetType->getName();

  if (m_dataSetName.empty() == true)
  {
    throw Exception(TE_TR("DataSet Name cannot be empty!"));
  }

  //we must ensure that the dataSetType have the correct columns names as the creation may change it
  m_dataSetType = transactor->getDataSetType(m_dataSetName);
}

void te::da::DataAccess::saveDataSet(te::da::DataSet* dataSet, bool enableProgress)
{
  //Creates an adapter to handle correctly the primary key and autonumber atributes. The adapter will handle the destruction of the dataSet
  std::unique_ptr<te::da::DataSet> adaptedDataSet = CreateAdapterToHandlePersistence(this, dataSet);

  // save data
  DataSourceTransactorPtr transactor = getTransactor();

  std::map<std::string, std::string> options;

  try
  {
    transactor->begin();

    adaptedDataSet->moveBeforeFirst();

    transactor->add(m_dataSetName, adaptedDataSet.get(), options, 0, enableProgress);

    transactor->commit();
  }
  catch (te::common::Exception& e)
  {
    transactor->rollBack();
    throw e;
  }
  catch (std::exception& e)
  {
    transactor->rollBack();
    throw e;
  }
}

std::vector<std::string> te::da::DataAccess::getDistinctValues(const std::string& attributeName)
{
  std::vector<std::string> values;

  std::unique_ptr<te::da::Select> selectPtr;

  if (m_dataSetQuery.get())
  {
    //attribute name
    std::string aliasName = GetAliasName(attributeName);

    selectPtr.reset(new te::da::Select(m_dataSetQuery.get()));

    te::da::Fields* fields = new te::da::Fields();

    te::da::Field* fieldName = new te::da::Field(aliasName);
    fields->push_back(fieldName);

    te::da::Distinct* dist = new te::da::Distinct;

    selectPtr->setDistinct(dist);
    selectPtr->setFields(fields);
  }
  else
  {
    //attribute name
    std::string aliasName = GetAliasName(attributeName);

    te::da::Fields* fields = new te::da::Fields;
    te::da::Field* fieldName = new te::da::Field(aliasName);
    fields->push_back(fieldName);

    te::da::Distinct* dist = new te::da::Distinct;

    //dataset Name
    aliasName = GetAliasName(m_dataSetName);

    te::da::From* from = new te::da::From;
    te::da::FromItem* fi = new te::da::DataSetName(m_dataSetName, aliasName);
    from->push_back(fi);

    selectPtr.reset(new te::da::Select());

    selectPtr->setDistinct(dist);
    selectPtr->setFields(fields);
    selectPtr->setFrom(from);
  }

  DataSourceTransactorPtr transactor = getTransactor();

  std::unique_ptr<te::da::DataSet> dataset;

  dataset = transactor->query(selectPtr.get());

  dataset->moveBeforeFirst();

  std::set<std::string> valuesStr;

  while (dataset->moveNext())
  {
    if (!dataset->isNull(attributeName))
    {
      std::string value = dataset->getAsString(attributeName);

      std::set<std::string>::iterator it = valuesStr.find(value);

      if (it == valuesStr.end())
        valuesStr.insert(value);
    }
  }

  std::set<std::string>::iterator it = valuesStr.begin();

  while (it != valuesStr.end())
  {
    values.push_back((*it).c_str());

    ++it;
  }

  return values;
}

te::da::Literal* te::da::DataAccess::getLiteral(const std::string& attributeName, const std::string& value)
{
  const te::da::DataSetType* dataSetType = getDataSetType();

  te::dt::Property* prop = dataSetType->getProperty(attributeName);

  te::da::Literal* literal = nullptr;

  if (prop->getType() == te::dt::STRING_TYPE)
  {
    literal = new te::da::LiteralString(value);
  }
  else if (prop->getType() == te::dt::INT16_TYPE)
  {
    literal = new te::da::LiteralInt16(atoi(value.c_str()));
  }
  else if (prop->getType() == te::dt::INT32_TYPE)
  {
    literal = new te::da::LiteralInt32(atoi(value.c_str()));
  }
  else if (prop->getType() == te::dt::INT64_TYPE)
  {
    literal = new te::da::LiteralInt64(atoi(value.c_str()));
  }
  else if (prop->getType() == te::dt::NUMERIC_TYPE)
  {
    literal = new te::da::LiteralDouble(atof(value.c_str()));
  }

  return literal;
}

te::gm::Envelope te::da::DataAccess::getMBR() const
{
  DataSourceTransactorPtr transactor = getTransactor();

  const te::da::DataSetType* dataSetType = getDataSetType();

  te::gm::GeometryProperty* geometryProperty = te::da::GetFirstGeomProperty(dataSetType);

  std::unique_ptr<te::gm::Envelope> envelope = transactor->getExtent(m_dataSetName, geometryProperty->getName());
  if (envelope == nullptr)
  {
    return te::gm::Envelope();
  }

  te::gm::Envelope mbr(*envelope);
  return mbr;
}

int te::da::DataAccess::getSRID() const
{
  const te::da::DataSetType* dataSetType = getDataSetType();
  if (dataSetType == nullptr)
  {
    return 0;
  }

  te::gm::GeometryProperty* geometryProperty = te::da::GetFirstGeomProperty(dataSetType);
  return geometryProperty->getSRID();
}

void te::da::DataAccess::setObjectIdSetFilter(const te::da::ObjectIdSet* objectIdSetFilter)
{
  if (objectIdSetFilter == nullptr)
  {
    m_objectIdSetFilter.reset();
  }
  else
  {
    m_objectIdSetFilter.reset(objectIdSetFilter->clone());
  }
}

te::da::DataSourceTransactorPtr te::da::DataAccess::getTransactor() const
{
  if (m_transactor == nullptr)
  {
    m_transactor.reset(m_dataSource->getTransactor().release());
  }

  return m_transactor;
}

void te::da::DataAccess::setTransactor(te::da::DataSourceTransactorPtr transactor)
{
  m_transactor = transactor;
}
