/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralib/dataaccess/utils/FieldNameValidator.cpp

   \brief This class is responsible for validating field names. It can handle size limits and changes the form 'tableName.propertyName' into the form 'tableName_propertyName'
 */

#include "FieldNameValidator.h"

#include "../../common/Exception.h"
#include "../../common/StringUtils.h"
#include "../../core/translator/Translator.h"

// Boost
#include <boost/algorithm/string/replace.hpp>

te::da::FieldNameValidator::FieldNameValidator(int fieldNameMaxSize, bool replaceDotForUnderline, bool replaceSpecialChars)
  : m_fieldNameMaxSize(fieldNameMaxSize)
  , m_replaceDotForUnderline(replaceDotForUnderline)
  , m_replaceSpecialChars(replaceSpecialChars)
{
  
}

std::string te::da::FieldNameValidator::adjustFieldNameMaxSize(const std::string& fieldName, int fieldNameMaxSize)
{
  std::string adjustedFieldName(fieldName);
  if (fieldNameMaxSize != -1)
  {
    if (adjustedFieldName.size() > fieldNameMaxSize)
    {
      adjustedFieldName = adjustedFieldName.substr(0, static_cast<std::size_t>(fieldNameMaxSize));
    }
  }

  return adjustedFieldName;
}

std::string te::da::FieldNameValidator::replaceDotForUnderline(const std::string& fieldName)
{
  std::string adjustedFieldName(fieldName);
  boost::replace_all(adjustedFieldName, ".", "_");

  return adjustedFieldName;
}

std::string te::da::FieldNameValidator::replaceSpecialCharacters(const std::string& fieldName)
{
  bool changed = false;
  std::string adjustedFieldName = te::common::ReplaceSpecialChars(fieldName, changed);

  return adjustedFieldName;
}

bool te::da::FieldNameValidator::addFieldName(const std::string& fieldName)
{
  std::string adjustedFieldName;
  return addFieldName(fieldName, adjustedFieldName);
}

bool te::da::FieldNameValidator::exists(const std::string& fieldName) const
{
  if (m_setAdjustedNames.find(fieldName) == m_setAdjustedNames.end())
  {
    return false;
  }

  return true;
}

bool te::da::FieldNameValidator::addFieldName(const std::string& fieldName, std::string& adjustedFieldName)
{
  //adjustedFieldName = te::common::Convert2LCase(fieldName);
  adjustedFieldName = fieldName;

  //we first replace '.' for '_'
  if (m_replaceDotForUnderline == true)
  {
    adjustedFieldName = replaceDotForUnderline(adjustedFieldName);
  }

  //then we must replace any special characters
  if (m_replaceSpecialChars == true)
  {
    adjustedFieldName = replaceSpecialCharacters(adjustedFieldName);
  }
  
  //then we must ensure that the field name size respects the defined maximum size
  if (m_fieldNameMaxSize != -1)
  {
    adjustedFieldName = adjustFieldNameMaxSize(adjustedFieldName, m_fieldNameMaxSize);
  }

  //now we must check if the ajusted name already exists. If so, we try to ajust it be replacing the last characters with numbers
  std::string baseName(adjustedFieldName);
  int renameIncrement = 1;
  while (exists(te::common::Convert2LCase(adjustedFieldName)) == true)
  {
    std::string tail = te::common::Convert2String(static_cast<boost::int32_t>(renameIncrement));
    adjustedFieldName = baseName.substr(0, baseName.size() - tail.size());
    adjustedFieldName += tail;

    ++renameIncrement;
    if (renameIncrement == 100)
    {
      throw te::common::Exception(std::string("te::ogr::FieldNameValidator::addFieldName - ") + TE_TR("Error ajusting the name of the column"));
    }
  }

  m_vecAdjustedNames.push_back(adjustedFieldName);
  m_setAdjustedNames.insert(te::common::Convert2LCase(adjustedFieldName));

  if (te::common::IsStringEqual(fieldName, adjustedFieldName, false) == true)
  {
    //if no change has been needed, we return the original column name with its original case
    adjustedFieldName = fieldName;
    return false;
  }

  return true;
}

const std::vector<std::string>& te::da::FieldNameValidator::getAdjustedNames() const
{
  return m_vecAdjustedNames;
}
