/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/datatype/Date.cpp

  \brief A base class for date data types.
*/

// TerraLib
#include "Date.h"

// Boost
#include <boost/date_time/gregorian/gregorian.hpp>
#include <boost/date_time/posix_time/posix_time.hpp>

class te::dt::DateImpl
{
public:
  DateImpl(const boost::gregorian::date& date)
    : m_date(date)
    , m_day(date.day())
    , m_month(date.month())
    , m_year(date.year())
  {

  }

  boost::gregorian::date m_date;
  boost::gregorian::greg_day m_day;
  boost::gregorian::greg_month m_month;
  boost::gregorian::greg_year m_year;
};

te::dt::Date::Date()
{
} 

te::dt::Date::Date(const boost::gregorian::greg_year& year,
                   const boost::gregorian::greg_month& month,
                   const boost::gregorian::greg_day& day)
  : m_dateImpl(new DateImpl(boost::gregorian::date(year, month, day)))
{  
}

te::dt::Date::Date(const boost::gregorian::date& d)
  : m_dateImpl(new DateImpl(d))
{ 
}

te::dt::Date::Date(const te::dt::Date& rhs)
  : m_dateImpl(new DateImpl(rhs.m_dateImpl->m_date))
{
}

const boost::gregorian::date& te::dt::Date::getDate() const
{
  return m_dateImpl->m_date;
}

boost::gregorian::date& te::dt::Date::getDate()
{ 
  return m_dateImpl->m_date;
}

const boost::gregorian::greg_day& te::dt::Date::getDay() const
{ 
  return m_dateImpl->m_day;
}

const boost::gregorian::greg_month& te::dt::Date::getMonth() const
{ 
  return m_dateImpl->m_month;
}

const boost::gregorian::greg_year& te::dt::Date::getYear() const
{ 
  return m_dateImpl->m_year;
}

te::dt::Date& te::dt::Date::operator=(const te::dt::Date& rhs)
{
  m_dateImpl.reset(new DateImpl(rhs.m_dateImpl->m_date));
  return *this;
}

bool te::dt::Date::operator==(const DateTime& rhs) const
{
  const te::dt::Date* t = dynamic_cast<const te::dt::Date*>(&rhs);
  return getDate() == t->getDate();
}
        
bool te::dt::Date::operator!=(const DateTime& rhs) const
{
  const te::dt::Date* t = dynamic_cast<const te::dt::Date*>(&rhs);
  return getDate() != t->getDate();
}
       
bool te::dt::Date::operator<(const DateTime& rhs) const
{
  const te::dt::Date* t = dynamic_cast<const te::dt::Date*>(&rhs);
  return getDate() < t->getDate();
}
        
bool te::dt::Date::operator>(const DateTime& rhs) const
{
  const te::dt::Date* t = dynamic_cast<const te::dt::Date*>(&rhs);
  return getDate() > t->getDate();
}

long te::dt::Date::operator-(const Date& rhs) const
{
  boost::gregorian::date_duration dd(getDate() - rhs.getDate());

  return dd.days();
}

te::dt::AbstractData* te::dt::Date::clone() const
{
  return new Date(*this);
}

std::string te::dt::Date::toString() const
{
  return boost::gregorian::to_simple_string(getDate());
}

te::dt::Date* te::dt::Date::now()
{
  boost::posix_time::ptime boostLocalTime = boost::posix_time::second_clock::local_time();
  te::dt::Date* date = new te::dt::Date(boostLocalTime.date());

  return date;
}

te::dt::Date::~Date() = default;
