/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/datatype/DoubleConverters.h

  \brief A set of function that convert a Double type to other types.
*/

#ifndef __TERRALIB_DATATYPE_INTERNAL_DOUBLECONVERTERS_H
#define __TERRALIB_DATATYPE_INTERNAL_DOUBLECONVERTERS_H

// TerraLib
#include "Exception.h"
#include "Config.h"

namespace te
{
  namespace dt
  {
// Forward declarations
    class AbstractData;

    /** @name Double Methods
      *  Methods related to Double conversion.
      */
    //@{

    /*!
      \brief It converts an Double data value to a char data value.

      \param d The input data value.

      \return A new data value converted to char data type. The caller will take the ownership of the returned data.

      \exception Exception It throws an exception if the input abstract data is not an Double type.
    */
    TEDATATYPEEXPORT AbstractData* DoubleToCharConverter(AbstractData* d) ;

    /*!
      \brief It converts an Double data value to an unsigned char data value.

      \param d The input data value.

      \return A new data value converted to unsigned char data type. The caller will take the ownership of the returned data.

      \exception Exception It throws an exception if the input abstract data is not an Double type.
    */
    TEDATATYPEEXPORT AbstractData* DoubleToUCharConverter(AbstractData* d) ;

    /*!
      \brief It converts an Double data value to a Int16 data value.

      \param d The input data value.

      \return A new data value converted to Int16 data type. The caller will take the ownership of the returned data.

      \exception Exception It throws an exception if the input abstract data is not an Double type.
    */
    TEDATATYPEEXPORT AbstractData* DoubleToInt16Converter(AbstractData* d) ;

    /*!
      \brief It converts an Double data value to an unsigned Int16 data value.

      \param d The input data value.

      \return A new data value converted to unsigned Int16 data type. The caller will take the ownership of the returned data.

      \exception Exception It throws an exception if the input abstract data is not an Double type.
    */
    TEDATATYPEEXPORT AbstractData* DoubleToUInt16Converter(AbstractData* d) ;

    /*!
      \brief It converts an Double data value to an Int32 data value.

      \param d The input data value.

      \return A new data value converted to Int32 data type. The caller will take the ownership of the returned data.

      \exception Exception It throws an exception if the input abstract data is not an Double type.
    */
    TEDATATYPEEXPORT AbstractData* DoubleToInt32Converter(AbstractData* d) ;

    /*!
      \brief It converts an Double data value to an unsigned Int32 data value.

      \param d The input data value.

      \return A new data value converted to unsigned Int32 data type. The caller will take the ownership of the returned data.

      \exception Exception It throws an exception if the input abstract data is not an Double type.
    */
    TEDATATYPEEXPORT AbstractData* DoubleToUInt32Converter(AbstractData* d) ;

    /*!
      \brief It converts an Double data value to a Int64 data value.

      \param d The input data value.

      \return A new data value converted to Int64 data type. The caller will take the ownership of the returned data.

      \exception Exception It throws an exception if the input abstract data is not an Double type.
    */
    TEDATATYPEEXPORT AbstractData* DoubleToInt64Converter(AbstractData* d) ;

    /*!
      \brief It converts an Double data value to an unsigned Int64 data value.

      \param d The input data value.

      \return A new data value converted to unsigned Int64 data type. The caller will take the ownership of the returned data.

      \exception Exception It throws an exception if the input abstract data is not an Double type.
    */
    TEDATATYPEEXPORT AbstractData* DoubleToUInt64Converter(AbstractData* d) ;

    /*!
      \brief It converts an Double data value to a boolean data value.

      \param d The input data value.

      \return A new data value converted to boolean data type. The caller will take the ownership of the returned data.

      \exception Exception It throws an exception if the input abstract data is not an Double type.
    */
    TEDATATYPEEXPORT AbstractData* DoubleToBooleanConverter(AbstractData* d) ;

    /*!
      \brief It converts an Double data value to a float value.

      \param d The input data value.

      \return A new data value converted to float data type. The caller will take the ownership of the returned data.

      \exception Exception It throws an exception if the input abstract data is not an Double type.
    */
    TEDATATYPEEXPORT AbstractData* DoubleToFloatConverter(AbstractData* d) ;

    /*!
      \brief It converts an Double data value to a double value.

      \param d The input data value.

      \return A new data value converted to double data type. The caller will take the ownership of the returned data.

      \exception Exception It throws an exception if the input abstract data is not an Double type.
    */
    TEDATATYPEEXPORT AbstractData* DoubleToDoubleConverter(AbstractData* d) ;

    /*!
      \brief It converts an Double data value to Numeric data value.

      \param d The input data value.

      \return A new data value converted to Numeric data type. The caller will take the ownership of the returned data.

      \exception Exception It throws an exception if the input abstract data is not an Double type.
    */
    TEDATATYPEEXPORT AbstractData* DoubleToNumericConverter(AbstractData* d) ;

    /*!
      \brief It converts an Double data value to a string data value.

      \param d The input data value.

      \return A new data value converted to string data type. The caller will take the ownership of the returned data.

      \exception Exception It throws an exception if the input abstract data is not an Double type.
    */
    TEDATATYPEEXPORT AbstractData* DoubleToStringConverter(AbstractData* d) ;

    /*!
      \brief Auxiliary method that asserts that the given data is an Double type.

      \param d The input data value.

      \exception Exception It throws an exception if the input abstract data is not an Double type.
    */
    TEDATATYPEEXPORT void AssertDouble(AbstractData* d) ;

    //@}

  } // end namespace dt
}   // end namespace te

#endif  // __TERRALIB_DATATYPE_INTERNAL_DOUBLECONVERTERS_H
