/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/edit/qt/tools/RotateGeometryTool.cpp

  \brief This class implements a concrete tool to rotate geometries.
*/

// TerraLib
#include "../../../qt/widgets/canvas/MapDisplay.h"
#include "../../Feature.h"
#include "../../RepositoryManager.h"
#include "../Renderer.h"
#include "../Utils.h"
#include "RotateGeometryTool.h"

// Qt
#include <QMessageBox>

te::edit::RotateGeometryTool::RotateGeometryTool(te::qt::widgets::MapDisplay* display, const te::map::AbstractLayerPtr& layer, QObject* parent)
  : GeometriesUpdateTool(display, layer.get(), parent),
    m_rotateStarted(false),
    m_angle(0)
{
  setCursor(Qt::ArrowCursor);

  draw();
}

te::edit::RotateGeometryTool::~RotateGeometryTool() = default;

bool te::edit::RotateGeometryTool::mousePressEvent(QMouseEvent* e)
{
  if (e->button() != Qt::LeftButton)
    return false;

  pickFeature(GetPosition(e));

  if (m_feature)
    m_rotateStarted = true;

  m_origin = GetPosition(e);

  m_axisX = 0.;

  return true;
}

bool te::edit::RotateGeometryTool::mouseMoveEvent(QMouseEvent* e)
{
  if(!m_rotateStarted)
    return false;

  QPointF currentPosition = GetPosition(e);

  // Calculates the axis X value
  m_axisX = currentPosition.x() - m_origin.x();

  // Rotate geometry using the current axis X
  if (m_feature)
  {
    if (m_axisX > 0 )
      m_angle += 1;
    else
      m_angle -= 1;

    if (std::abs(m_angle) > 180)
      m_angle = 0;

    Rotate(m_feature->getGeometry(), (m_axisX > 0 ? -1 : 1));
  }

  draw();

  m_origin = currentPosition;

  return true;
}

bool te::edit::RotateGeometryTool::mouseReleaseEvent(QMouseEvent* e)
{
  if (e->button() != Qt::RightButton)
    return false;

  m_rotateStarted = false;

  return editingFinished();
}

void te::edit::RotateGeometryTool::reset()
{
  delete m_feature;

  m_rotateStarted = false;

  m_origin *= 0;
  m_axisX = 0.;

  m_angle = 0;
}

void te::edit::RotateGeometryTool::pickFeature(const QPointF& pos)
{
  reset();

  te::gm::Envelope env = buildEnvelope(pos);

  try
  {
    m_feature = PickFeature(m_layer, env, m_display->getSRID());

    draw();
  }
  catch(std::exception& e)
  {
    QMessageBox::critical(m_display, tr("Error"), QString(tr("The geometry cannot be selected from the layer. Details:") + " %1.").arg(e.what()));
  }
}

te::gm::Envelope te::edit::RotateGeometryTool::buildEnvelope(const QPointF& pos)
{
  QPointF pixelOffset(4.0, 4.0);

  QRectF rect(pos - pixelOffset, pos + pixelOffset);

  // Converts rect boundary to world coordinates
  QPointF ll(rect.left(), rect.bottom());
  QPointF ur(rect.right(), rect.top());
  ll = m_display->transform(ll);
  ur = m_display->transform(ur);

  te::gm::Envelope env(ll.x(), ll.y(), ur.x(), ur.y());

  return env;
}

void te::edit::RotateGeometryTool::draw()
{
  const te::gm::Envelope& env = m_display->getExtent();
  if(!env.isValid())
    return;

  // Clear!
  QPixmap* draft = m_display->getDraftPixmap();
  draft->fill(Qt::transparent);

   // Initialize the renderer
  Renderer& renderer = Renderer::getInstance();
  renderer.begin(draft, env, m_display->getSRID());

  // Draw the layer edited geometries
  renderer.drawRepository(m_layer->getId(), env, m_display->getSRID());

  if (m_feature == nullptr)
  {
    renderer.end();
    m_display->repaint();
    return;
  }

  // Draw the vertexes
  if (RepositoryManager::getInstance().hasIdentify(m_layer->getId(), m_feature->getId()) == false)
    renderer.draw(m_feature, true);
  else
    renderer.drawVertexes(m_feature->getGeometry());

  // Draw the angle
  te::gm::Coord2D centroid = m_feature->getGeometry()->getCentroid();
  std::unique_ptr<te::gm::Point> point(new te::gm::Point(centroid.x, centroid.y, m_display->getSRID()));

  renderer.drawText((QString::number(m_angle) + tr(" deg")).toUtf8().data(), point.get());
  
  renderer.end();

  m_display->repaint();
}

bool te::edit::RotateGeometryTool::editingFinished()
{
  if (m_feature)
    storeFeature();

  emit geometriesEdited();

  return true;
}

void te::edit::RotateGeometryTool::storeFeature()
{
  if (m_feature == nullptr)
    return;

  RepositoryManager::getInstance().addFeature(m_layer->getId(), m_feature->clone());
}
