/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/edit/qt/tools/SplitPolygonTool.cpp

  \brief This class implements a concrete tool to split polygons.
*/

// TerraLib
#include "../../../common/STLUtils.h"
#include "../../../dataaccess/dataset/ObjectId.h"
#include "../../../geometry/Envelope.h"
#include "../../../dataaccess/utils/Utils.h"
#include "../../../qt/widgets/canvas/MapDisplay.h"
#include "../../../qt/widgets/Utils.h"
#include "../../Feature.h"
#include "../../Repository.h"
#include "../../RepositoryManager.h"
#include "../../Utils.h"
#include "../Renderer.h"
#include "../Utils.h"
#include "SplitPolygonTool.h"

// Qt
#include <QMessageBox>
#include <QMouseEvent>
#include <QPainter>
#include <QPixmap>

// STL
#include <cassert>
#include <memory>

te::edit::SplitPolygonTool::SplitPolygonTool(te::qt::widgets::MapDisplay* display, const te::map::AbstractLayerPtr& layer, QObject* parent)
  : CreateLineTool(display, layer, Qt::ArrowCursor, parent),
  m_tol(0.000001),
  m_oidSet(nullptr),
  m_vecFeatures(0)
{
  setCursor(Qt::ArrowCursor);

  pickFeatures();

  draw();
}

te::edit::SplitPolygonTool::~SplitPolygonTool()
{
  te::common::FreeContents(m_vecFeatures);
  m_vecFeatures.clear();

  delete m_oidSet;

  m_layer->clearSelected();
}

bool te::edit::SplitPolygonTool::mousePressEvent(QMouseEvent* e)
{
  if (e->button() != Qt::LeftButton)
    return false;

  if (m_isFinished) // Is Finished?! So, start again...
  {
    resetVisualizationTool();
    m_isFinished = false;

    pickFeatures();
  }

  return te::edit::CreateLineTool::mousePressEvent(e);
}

bool te::edit::SplitPolygonTool::mouseMoveEvent(QMouseEvent* e)
{
  return te::edit::CreateLineTool::mouseMoveEvent(e);
}

bool te::edit::SplitPolygonTool::mouseDoubleClickEvent(QMouseEvent* e)
{
  if (Qt::LeftButton != e->button())
    return false;

  editingFinished();

  return true;
}

bool te::edit::SplitPolygonTool::mouseReleaseEvent(QMouseEvent* e)
{
  if (Qt::RightButton != e->button())
    return false;

  editingFinished();

  return true;
}

void te::edit::SplitPolygonTool::editingFinished()
{
  if (m_coords.size() < 2) // Can not stop yet...
    return;

  m_isFinished = true;

  if (m_oidSet == nullptr)
    m_oidSet = new te::da::ObjectIdSet();

  for (std::size_t i = 0; i < m_vecFeatures.size(); ++i)
  {
    if (RepositoryManager::getInstance().hasIdentify(m_layer->getId(), m_vecFeatures[i]->getId()))
    {
      RepositoryManager::getInstance().removeFeature(m_layer->getId(), m_vecFeatures[i]->getId());
      splitPolygon(i);
    }
  }

  te::edit::CreateLineTool::clear();

  if (!m_oidSet->size())
    return;

  emit geometriesEdited();
}

void te::edit::SplitPolygonTool::splitPolygon(std::size_t index)
{
  setCursor(Qt::WaitCursor);

  try
  {
    int srid = m_display->getSRID();

    //gets the input geometry
    te::gm::Geometry* inputGeometry = m_vecFeatures.at(index)->getGeometry();
    if (m_layer->getSRID() != m_display->getSRID())
    {
      inputGeometry->transform(srid);
    }

    //we get the line that will be used to split the geometry
    std::unique_ptr<te::gm::Geometry> blade_in;
    blade_in.reset(te::edit::CreateLineTool::buildLine());
    blade_in->setSRID(srid);

    std::vector<te::gm::Geometry*> vecResult = te::gm::SplitGeometry(inputGeometry, blade_in.get());
    for (std::size_t i = 0; i < vecResult.size(); ++i)
    {
      //we use the same pointer, so we dont need to delete it
      te::gm::Geometry* newGeometry = vecResult.at(i);

      //than we clone the current feature, as all the parts must have the same attributes
      Feature* newFeature = m_vecFeatures.at(index)->clone();
      newFeature->setGeometry(newGeometry);

      //for the first part, we will just update it in the database. And so we keep its ID
      //for all the other parts, we create a new ID and so it will be added to the database
      if (i >= 1)
      {
        newFeature->setId(GenerateId());
        newFeature->setOperationTypeId(INSERT);
      }

      //and finally we update the repository
      RepositoryManager::getInstance().addFeature(m_layer->getId(), newFeature);
      m_oidSet->add(newFeature->getId()->clone());
    }

    setCursor(Qt::ArrowCursor);
  }
  catch (std::exception& e)
  {
    setCursor(Qt::ArrowCursor);
    QMessageBox::critical(m_display, tr("Error"), QString(tr("Could not split.") + " %1.").arg(e.what()));
    return;
  }
}

void te::edit::SplitPolygonTool::draw()
{
  const te::gm::Envelope& env = m_display->getExtent();
  if (!env.isValid())
    return;

  // Clear!
  QPixmap* draft = m_display->getDraftPixmap();
  draft->fill(Qt::transparent);

  // Initialize the renderer
  Renderer& renderer = Renderer::getInstance();
  renderer.begin(draft, env, m_display->getSRID());

  // Draw the layer edited geometries
  renderer.drawRepository(m_layer->getId(), env, m_display->getSRID());

  renderer.end();

  m_display->repaint();
}

void te::edit::SplitPolygonTool::resetVisualizationTool()
{
  te::edit::Repository* repo = te::edit::RepositoryManager::getInstance().getRepository(m_layer->getId());

  if (m_oidSet)
  {
    std::set<te::da::ObjectId*, te::common::LessCmp<te::da::ObjectId*> >::const_iterator it;
    for (it = m_oidSet->begin(); it != m_oidSet->end(); ++it)
    {
      if (repo->hasIdentifier(*it))
        repo->remove((*it));
    }
  }

  te::edit::CreateLineTool::clear();
  m_isFinished = true;

  delete m_oidSet;
  m_oidSet = nullptr;
}

void te::edit::SplitPolygonTool::pickFeatures()
{
  try
  {
    if(m_layer->getSelected() == nullptr || m_layer->getSelected()->size() == 0)
    {
      QMessageBox::information(m_display,
                               tr("TerraAmazon"),
                               tr("To use this tool, you must select at least one geometry."));
      return;
    }

    std::unique_ptr<te::da::DataSet> ds(m_layer->getData(m_layer->getSelected()));
    std::size_t gpos = te::da::GetFirstSpatialPropertyPos(ds.get());

    std::vector<std::string> oidPropertyNames;
    te::da::GetOIDPropertyNames(m_layer->getSchema().get(), oidPropertyNames);

    te::common::FreeContents(m_vecFeatures);
    m_vecFeatures.clear();

    ds->moveBeforeFirst();
    while (ds->moveNext())
    {
      //we first get all the attributes
      std::map<std::string, te::dt::AbstractData*> mapFields;
      for (std::size_t i = 0; i < ds->getNumProperties(); ++i)
      {
        std::string pName = ds->getPropertyName(i);
        std::size_t pos = te::da::GetPropertyPos(m_layer->getSchema().get(), pName);
        if (!ds->isNull(pos) && ds->getPropertyDataType(pos) != te::dt::GEOMETRY_TYPE)
        {
          std::unique_ptr<te::dt::AbstractData> data(ds->getValue(pos));
          mapFields[pName] = data.release();
        }
      }

      //then we get the geometry
      std::unique_ptr<te::gm::Geometry> g(ds->getGeometry(gpos));

      //and finally we create the feature
      Feature* f = new Feature(te::da::GenerateOID(ds.get(), oidPropertyNames), g.release());
      f->setData(mapFields);
      f->setOperationTypeId(UPDATE);
      
      m_vecFeatures.push_back(f);
      RepositoryManager::getInstance().addFeature(m_layer->getId(), f->clone());
    }

  }
  catch (std::exception& e)
  {
    QMessageBox::critical(m_display, tr("Error"), QString(tr("The geometry cannot be selected from the layer. Details:") + " %1.").arg(e.what()));
  }
}
