/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/gdal/ScopedDataSetHandle.cpp

  \brief This encapsulates a GDAL dataSet and closes it when it runs out of scope. It also creates a wrapper to the dataSet in the cases which the file has a RPC metadata.
*/

// TerraLib
#include "ScopedDataSetHandle.h"

#include <gdal.h>
#include <gdal_priv.h>
#include <gdalwarper.h>
#include <ogr_spatialref.h>

te::gdal::ScopedDataSetHandle::ScopedDataSetHandle(GDALDataset* gdalDataSetHandle)
  : m_gdalDataSetWrappedHandle(gdalDataSetHandle)
  , m_gdalDataSetHandle(gdalDataSetHandle)
{
  if (m_gdalDataSetHandle == nullptr)
  {
    return;
  }

  //GDAL explanation
  //Many modern raw satellite products are distributed with RPCs, including products from GeoEye, and DigitalGlobe.
  //RPCs provide a higher systematic description of georeferencing over an image, and also contain information on the viewing geometry that in theory makes orthocorrection (given a DEM) and some 3D operations like building height computation possible.

  //For this reason, if the GDALDataSet has this RPC information, we must add a wrapper the help reading it
  double mGeoTransform[6];
  bool hasGeoTransform = GDALGetGeoTransform(m_gdalDataSetHandle, mGeoTransform) == CE_None;
  bool hasInValidRasterTransformParameters = true;
  if (hasGeoTransform == true)
  {
    hasInValidRasterTransformParameters = (mGeoTransform[1] < 0.0 || mGeoTransform[2] != 0.0 || mGeoTransform[4] != 0.0 || mGeoTransform[5] > 0.0);
  }
  if ((hasGeoTransform == true && hasInValidRasterTransformParameters == true) || GDALGetGCPCount(m_gdalDataSetHandle) > 0 || GDALGetMetadata(m_gdalDataSetHandle, "RPC"))
  {
    m_gdalDataSetWrappedHandle = static_cast<GDALDataset*>(GDALAutoCreateWarpedVRT(m_gdalDataSetHandle, nullptr, nullptr, GRA_NearestNeighbour, 0.2, nullptr));
    if (m_gdalDataSetWrappedHandle != nullptr)
    {
      if (GDALGetMetadata(m_gdalDataSetHandle, "RPC"))
      {
        std::unique_ptr< OGRSpatialReference > poSRS(new OGRSpatialReference());
        poSRS->importFromEPSG(4326);

        char* projWKTPtr = nullptr;
        OGRErr osrsErrorReturn = poSRS->exportToWkt(&projWKTPtr);
        if (osrsErrorReturn == OGRERR_NONE)
        {
          osrsErrorReturn = m_gdalDataSetWrappedHandle->SetProjection(projWKTPtr);
        }

        CPLFree(projWKTPtr);
      }
    }
  }
}

te::gdal::ScopedDataSetHandle::~ScopedDataSetHandle()
{
  if (m_gdalDataSetHandle != m_gdalDataSetWrappedHandle)
  {
    GDALDereferenceDataset(m_gdalDataSetHandle);
  }

  GDALClose(m_gdalDataSetWrappedHandle);
}

GDALDataset* te::gdal::ScopedDataSetHandle::getGDALDataSetHandle() const
{
  return m_gdalDataSetWrappedHandle;
}

