/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
 \file SnapCoordinates.cpp
 */

#include "CoordinateSnapper.h"

#include "../common/STLUtils.h"
#include "../sam/kdtree/Index.h"
#include "../sam/rtree/Index.h"

#include "Geometry.h"
#include "GeometryCollection.h"
#include "GeometryFunctions.h"
#include "LineString.h"
#include "Polygon.h"
#include "Utils.h"
#include "Validation.h"

struct CoordIndexData
{
  te::gm::Coord2D m_coord;
  std::size_t m_data;

  CoordIndexData()
    : m_data(0)
  {

  }
  CoordIndexData(const te::gm::Coord2D& coord, std::size_t data)
    : m_coord(coord)
    , m_data(data)
  {

  }
};

struct CoordIndexInternal
{
  te::sam::rtree::Index<std::size_t> m_rTree;
  std::vector<te::gm::Coord2D> m_vecIndexedCoords;
};

struct CoordIndex
{
  //CoordIndex
  CoordIndex()
    : m_internal(new CoordIndexInternal())
  {

  }

  void insert(const te::gm::Coord2D& coord)
  {
    std::size_t newData = m_internal->m_vecIndexedCoords.size();
    te::gm::Envelope mbr = te::gm::CreateEnvelope(coord, 0);

    m_internal->m_rTree.insert(mbr, newData);
    m_internal->m_vecIndexedCoords.push_back(coord);
  }

  void remove(const te::gm::Coord2D& coord, std::size_t data)
  {
    te::gm::Envelope mbr = te::gm::CreateEnvelope(coord, 0);
    m_internal->m_rTree.remove(mbr, data);
  }

  void remove(const CoordIndexData& coordIndexData)
  {
    te::gm::Envelope mbr = te::gm::CreateEnvelope(coordIndexData.m_coord, 0);
    m_internal->m_rTree.remove(mbr, coordIndexData.m_data);
  }

  const std::vector<te::gm::Coord2D>& getCoords() const
  {
    return m_internal->m_vecIndexedCoords;
  }

  std::size_t search(const te::gm::Envelope& mbr, std::vector<CoordIndexData>& vecData)
  {
    std::vector<std::size_t> vecIndexesReport;
    m_internal->m_rTree.search(mbr, vecIndexesReport);

    for (std::size_t r = 0; r < vecIndexesReport.size(); ++r)
    {
      std::size_t index = vecIndexesReport.at(r);
      const te::gm::Coord2D& coord = m_internal->m_vecIndexedCoords.at(index);
      CoordIndexData indexData(coord, index);
      vecData.push_back(indexData);
    }

    return vecData.size();
  }

  std::size_t search(const te::gm::Coord2D& referenceCoord, double distance, std::vector<CoordIndexData>& vecData)
  {
    te::gm::Envelope mbr = te::gm::CreateEnvelope(referenceCoord, distance);

    std::vector<std::size_t> vecIndexesReport;
    m_internal->m_rTree.search(mbr, vecIndexesReport);

    for (std::size_t r = 0; r < vecIndexesReport.size(); ++r)
    {
      std::size_t index = vecIndexesReport.at(r);
      const te::gm::Coord2D& candidateCoord = m_internal->m_vecIndexedCoords.at(index);
      if(te::gm::Distance::IsWithinDistance(referenceCoord, candidateCoord, distance) == true)
      {
        CoordIndexData indexData(candidateCoord, index);
        vecData.push_back(indexData);
      }
    }

    return vecData.size();
  }

  protected:

    std::unique_ptr<CoordIndexInternal> m_internal;
};

//SnapCoordinates
te::gm::GeometryVector te::gm::CoordinateSnapper::snapCoordinates(const te::gm::GeometryVectorConst& vecGeometries, double distance, bool validateResult)
{
  CoordIndex coordIndex;

  populateCoordIndex(vecGeometries, coordIndex);
  snapToClusters(distance, coordIndex);
  te::gm::GeometryVector vecOutputGeometries = snapCoordinates(vecGeometries, distance, validateResult, coordIndex);
  return vecOutputGeometries;
}

//SnapCoordinates
te::gm::GeometryPtr te::gm::CoordinateSnapper::snapCoordinates(const te::gm::Geometry* geometry, double distance, bool validateResult)
{
  te::gm::GeometryVectorConst inputVector;
  inputVector.push_back(geometry);

  te::gm::GeometryVector outputVector = snapCoordinates(inputVector, distance, validateResult);

  return outputVector.at(0);
}

void te::gm::CoordinateSnapper::populateCoordIndex(const te::gm::Geometry* geometry, CoordIndex& coordIndex)
{
  if (geometry->isCollection() == true)
  {
    const te::gm::GeometryCollection* geometryCollection = static_cast<const te::gm::GeometryCollection*>(geometry);
    for (std::size_t c = 0; c < geometryCollection->getNumGeometries(); ++c)
    {
      populateCoordIndex(geometryCollection->getGeometryN(c), coordIndex);
    }
  }
  else if (geometry->get2DGeomTypeId() == te::gm::PolygonType)
  {
    const te::gm::Polygon* polygon = static_cast<const te::gm::Polygon*>(geometry);
    std::vector<te::gm::Curve*> vecCurves = polygon->getRings();
    for (std::size_t r = 0; r < vecCurves.size(); ++r)
    {
      populateCoordIndex(vecCurves.at(r), coordIndex);
    }
  }
  else if (geometry->get2DGeomTypeId() == te::gm::LineStringType)
  {
    const te::gm::LineString* line = static_cast<const te::gm::LineString*>(geometry);
    std::size_t nPoints = line->getNPoints();
    if (line->isClosed() == true)
    {
      //we dont need to handle the same point twice
      --nPoints;
    }
    for (std::size_t i = 0; i < nPoints; ++i)
    {
      const double& x = line->getX(i);
      const double& y = line->getY(i);

      te::gm::Coord2D coord(x, y);
      coordIndex.insert(coord);
    }
  }
}

void te::gm::CoordinateSnapper::populateCoordIndex(const te::gm::GeometryVectorConst& vecGeometries, CoordIndex& coordIndex)
{
  for (std::size_t i = 0; i < vecGeometries.size(); ++i)
  {
    const te::gm::Geometry* geometry = vecGeometries.at(i);
    if (te::gm::IsNullOrEmpty(geometry) == true)
    {
      continue;
    }
    populateCoordIndex(geometry, coordIndex);
  }
}

void te::gm::CoordinateSnapper::snapToClusters(double distance, CoordIndex& coordIndex)
{
  const std::vector<te::gm::Coord2D>& vecIndexedCoords = coordIndex.getCoords();
  for (std::size_t currentIndex = 0; currentIndex < vecIndexedCoords.size(); ++currentIndex)
  {
    const te::gm::Coord2D& coord = vecIndexedCoords.at(currentIndex);

    //get the candidates
    std::vector<CoordIndexData> vecDataLevel1;
    coordIndex.search(coord, distance, vecDataLevel1);

    std::size_t myNeighborsCount = 0;
    for (std::size_t j = 0; j < vecDataLevel1.size(); ++j)
    {
      const std::size_t& candidateIndex = vecDataLevel1.at(j).m_data;
      if (currentIndex == candidateIndex)
      {
        continue;
      }

      ++myNeighborsCount;
    }

    if (myNeighborsCount == 0)
    {
      continue;
    }

    for (std::size_t j = 0; j < vecDataLevel1.size(); ++j)
    {
      const std::size_t& candidateIndex = vecDataLevel1.at(j).m_data;
      const te::gm::Coord2D& candidateCoord = vecDataLevel1.at(j).m_coord;

      //get the candidates neighbors
      std::vector<CoordIndexData> vecDataLevel2;
      coordIndex.search(candidateCoord, distance, vecDataLevel2);

      std::size_t myCandidateNeighborsCount = 0;
      for (std::size_t k = 0; k < vecDataLevel2.size(); ++k)
      {
        const std::size_t& candidateIndex2 = vecDataLevel2.at(k).m_data;
        if (candidateIndex == candidateIndex2)
        {
          continue;
        }

        ++myCandidateNeighborsCount;
      }

      //if my candidate has more or a equal equal number of neighbors than me, I discard myself
      if (myNeighborsCount <= myCandidateNeighborsCount)
      {
        coordIndex.remove(coord, currentIndex);
        break;
      }
    }
  }
}

te::gm::Geometry* te::gm::CoordinateSnapper::snapCoordinates(const te::gm::Geometry* geometry, double distance, CoordIndex& coordIndex, bool& wasChanged)
{
  te::gm::Geometry* resultGeometry = static_cast<te::gm::Geometry*>(geometry->clone());

  if (resultGeometry->isCollection() == true)
  {
    te::gm::GeometryCollection* geometryCollection = static_cast<te::gm::GeometryCollection*>(resultGeometry);
    for (std::size_t c = 0; c < geometryCollection->getNumGeometries(); ++c)
    {
      te::gm::Geometry* adjustedGeometry = snapCoordinates(geometryCollection->getGeometryN(c), distance, coordIndex, wasChanged);
      geometryCollection->setGeometryN(c, adjustedGeometry);
    }
  }
  else if (resultGeometry->get2DGeomTypeId() == te::gm::PolygonType)
  {
    te::gm::Polygon* newPolygon = static_cast<te::gm::Polygon*>(resultGeometry);
    for (std::size_t r = 0; r < newPolygon->getNumRings(); ++r)
    {
      te::gm::Curve* curve = static_cast<te::gm::Curve*>(snapCoordinates(newPolygon->getRingN(r), distance, coordIndex, wasChanged));
      newPolygon->setRingN(r, curve);
    }
  }
  else if (resultGeometry->get2DGeomTypeId() == te::gm::LineStringType)
  {
    te::gm::LineString* line = static_cast<te::gm::LineString*>(resultGeometry);

    std::size_t nPoints = line->getNPoints();
    for (std::size_t i = 0; i < nPoints; ++i)
    {
      const double& x = line->getX(i);
      const double& y = line->getY(i);

      te::gm::Coord2D coord(x, y);

      std::vector<CoordIndexData> vecData;
      std::size_t count = coordIndex.search(coord, distance, vecData);
      if (count == 1)
      {
        const te::gm::Coord2D& referenceCoord = vecData.at(0).m_coord;
        if (te::gm::Distance::IsEqual(coord.x, coord.y, referenceCoord.x, referenceCoord.y) == false)
        {
          line->setPoint(i, referenceCoord.x, referenceCoord.y);
          wasChanged = true;
        }
      }
    }
  }

  return resultGeometry;
}

te::gm::GeometryVector te::gm::CoordinateSnapper::snapCoordinates(const te::gm::GeometryVectorConst& vecGeometries, double distance, bool validateResult, CoordIndex& coordIndex)
{
  te::gm::GeometryVector vecResult;
  for (std::size_t i = 0; i < vecGeometries.size(); ++i)
  {
    const te::gm::Geometry* geometry = vecGeometries.at(i);
    if (te::gm::IsNullOrEmpty(geometry) == true)
    {
      vecResult.push_back(nullptr);
      continue;
    }

    bool wasChanged1 = false;
    te::gm::GeometryPtr adjustedGeometry(snapCoordinates(geometry, distance, coordIndex, wasChanged1));

    if (validateResult == true)
    {
      if (adjustedGeometry != nullptr && adjustedGeometry->isEmpty() == false && adjustedGeometry->getDimension() == te::gm::A)
      {
        //we must ensure that the geometry stays valid
        if ((wasChanged1 == true) && adjustedGeometry->isValid() == false)
        {
          adjustedGeometry.reset(te::gm::Validation::MakeValid(adjustedGeometry.get()).release());
        }
      }
    }

    vecResult.push_back(adjustedGeometry.release());
  }

  return vecResult;
}


