/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/geometry/CoordinateSnapper.h
  \brief Algorithms to snap existing "close" coordinates to each other based uppon a given tolerance, ensuring that they now have the exactly same location.
 */

#ifndef __TERRALIB_GEOMETRY_INTERNAL_COORDINATESNAPPER_H
#define __TERRALIB_GEOMETRY_INTERNAL_COORDINATESNAPPER_H

#include "Config.h"

#include "CommonDataStructures.h"

#include <vector>

struct CoordIndexInternal;
struct CoordIndexData;
struct CoordIndex;

namespace te
{
  namespace gm
  {
    struct Coord2D;
    class Envelope;
    class Geometry;
    class GeometryPtr;

    /*!
      \class CoordinateSnapper

      \brief Algorithm to snap existing "close" coordinates to each other based uppon a given tolerance, ensuring that they now have the exactly same location.

      \ingroup vp

      \sa GeometrySnapper
    */
    class TEGEOMEXPORT CoordinateSnapper
    {
    public:

      /*!
        \brief Algorithm to snap existing "close" coordinates to each other based uppon a given tolerance, ensuring that they now have the exactly same location.

        \param vecGeometries The geometry list to be analysed
        \param distance The distance to be considered for the snapping of the coordinates. 

        \return The resulting snapped geometries
      */
      static te::gm::GeometryVector snapCoordinates(const te::gm::GeometryVectorConst& vecGeometries, double distance, bool validateResult = true);

      /*!
        \brief Algorithm to snap existing "close" coordinates to each other based uppon a given tolerance, ensuring that they now have the exactly same location.

        \param geometry The geometry to be analysed
        \param distance The distance to be considered for the snapping of the coordinates.

        \return The resulting snapped geometry
      */
      static te::gm::GeometryPtr snapCoordinates(const te::gm::Geometry* geometry, double distance, bool validateResult = true);

    protected:

      /*!
          \brief Populates the coordinate index considering the coordinates of the given geometry

          \param geometry The geometry from which the coordinates will be added to the coordinate index
          \param coordIndex The index to be populated
        */
      static void populateCoordIndex(const te::gm::Geometry* geometry, CoordIndex& coordIndex);

      /*!
          \brief Populates the coordinate index considering the coordinates of the given geometry vector

          \param vecGeometries The geometries from which the coordinates will be added to the coordinate index
          \param coordIndex The index to be populated
        */
      static void populateCoordIndex(const te::gm::GeometryVectorConst& vecGeometries, CoordIndex& coordIndex);

      /*!
          \brief Analyses all the coordinates in the coordinate index, detecting clusters and the snapping them to each detected cluster

          \param distance The distance to be considered when detecting clusters and snapping geometries
          \param coordIndex The index to be populated
        */
      static void snapToClusters(double distance, CoordIndex& coordIndex);
      
      /*!
          \brief Snaps the geometry coordinates to the existing clusters in the coordinate index

          \param geometry The geometry to be snapped
          \param distance The distance to be considered when detecting clusters and snapping geometries
          \param coordIndex The coordinate index containing the clusters
        */
      static te::gm::Geometry* snapCoordinates(const te::gm::Geometry* geometry, double distance, CoordIndex& coordIndex, bool& wasChanged);

      /*!
         \brief Snaps the geometries coordinates to the existing clusters in the coordinate index

         \param vecGeometries The geometries to be snapped
         \param distance The distance to be considered when detecting clusters and snapping geometries
         \param coordIndex The coordinate index containing the clusters
       */
      static te::gm::GeometryVector snapCoordinates(const te::gm::GeometryVectorConst& vecGeometries, double distance, bool validateResult, CoordIndex& coordIndex);
    };
  }
}
#endif // __TERRALIB_GEOMETRY_INTERNAL_COORDINATESNAPPER_H
