/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
 \file FixGeometryTopology.h
 
 \brief Functions to fix geometry topology.
 */

#ifndef __TERRALIB_GEOMETRY_INTERNAL_FIXGEOMETRYTOPOLOGY_H
#define __TERRALIB_GEOMETRY_INTERNAL_FIXGEOMETRYTOPOLOGY_H

//Terralib

#include "../common/Static.h"

#include "Enums.h"
#include "Validation.h"

#include "Config.h"

#include <vector>

namespace te
{
  namespace gm
  {
    class Geometry;

    /*!
      \class FixGeometryTopology

      \brief A static class with methods to fix geometry topology.
    */
    class TEGEOMEXPORT FixGeometryTopology : public te::common::Static
    {
      public:

      /*!
        \brief  This method gets a topologically inconsistent geometry and calls specific functions to fix this geometry.

        \param geometryToFix  Input Geometry to fix.
        \param fixedGeometryVector A vector with the result of fixed geometries.
        \param topologyError  If any error occurs, TopologyValidationError struct is populated with error message and geometry coordinate.

        \return True, if the function was executed successfully and the geometry was fixed correctly.
      */
       static bool fixTopology(const te::gm::Geometry& geometry,
                               std::vector<te::gm::Geometry*>& fixedGeometryVector,
                               TopologyValidationError topologyError);

      private:

        /*!
          \brief  This function closes an opened geometry ring.

          \param geometryToFix  Input Geometry to fix.
          \param topologyError  If any error occurs, TopologyValidationError struct is populated with error message and geometry coordinate.

          \return True, if the the function was executed successfully and the geometry was fixed correctly.

          \note This function is only used for Geometries of Polygon and MultiPolygon types.
        */
        static bool closeGeometryRing(const te::gm::Geometry& geometry,
                                      TopologyValidationError topologyError);

        /*!
          \brief  This function fixes geometries with self-intersection.

          \param geometryToFix  Input Geometry to fix.
          \param fixedGeometryVector A vector with the result of fixed geometries.
          \param topologyError  If any error occurs, TopologyValidationError struct is populated with error message and geometry coordinate.

          \return True, if the the function was executed successfully and the geometry was fixed correctly.
        */
        static bool fixSelfIntersection(const te::gm::Geometry& geometry,
                                        std::vector<te::gm::Geometry*>& fixedGeometryVector,
                                        TopologyValidationError topologyError);
    };
  }
}
#endif // __TERRALIB_GEOMETRY_INTERNAL_FIXGEOMETRYTOPOLOGY_H
