/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/geometry/GTFilter.cpp

  \brief 2D Geometric transformation outliers remotion filter.
*/ 

// TerraLib
#include "../common/PlatformUtils.h"
#include "GTFactory.h"
#include "GTFilter.h"
#include "MultiPoint.h"
#include "Point.h"
#include "Surface.h"

// boost
#include <boost/random.hpp>

// STL
#include <cmath>
#include <cstdlib>
#include <limits>
#include <memory>

// ---------------------------------------------------------------------------

te::gm::GTFilter::RansacSetData::RansacSetData()
{
  m_maxDirectMapError = 0;
  m_maxInverseMapError = 0;
  m_area1 = 0;
  m_area2 = 0;
}

te::gm::GTFilter::RansacSetData::~RansacSetData()
{
}

// ---------------------------------------------------------------------------

te::gm::GTFilter::ApplyRansacThreadEntryThreadParams::ApplyRansacThreadEntryThreadParams()
{
  m_inputGTParamsPtr = nullptr;
  m_allowedMaxDirectMapError = 0.0;
  m_allowedMaxInverseMapError = 0.0;
  m_useDynamicIterationsNumber = 0;
  m_dynamicMaxIterationsPtr = nullptr;
  m_globalIterationsCounterPtr = nullptr;
  m_procsNumber = 1;
  m_returnValuePtr = nullptr;
  m_mutexPtr = nullptr;
  m_keepRunningFlagPtr = nullptr;
  m_tpsMapPtr = nullptr;
  m_bestSetDataPtr = nullptr;
  m_maxFoundGlobalItDMapErrorPtr = nullptr;
  m_maxFoundGlobalItIMapErrorPtr = nullptr;
  m_areaRestT = RANSAC_BBOX_AREA_REST;
  m_globalConsecutiveLowQualitySetsCounterPtr = nullptr;
}

te::gm::GTFilter::ApplyRansacThreadEntryThreadParams::ApplyRansacThreadEntryThreadParams(
  const te::gm::GTFilter::ApplyRansacThreadEntryThreadParams& other )
{
  operator=( other );
}

te::gm::GTFilter::ApplyRansacThreadEntryThreadParams::
    ~ApplyRansacThreadEntryThreadParams() = default;

const te::gm::GTFilter::ApplyRansacThreadEntryThreadParams& 
  te::gm::GTFilter::ApplyRansacThreadEntryThreadParams::operator=(
  const te::gm::GTFilter::ApplyRansacThreadEntryThreadParams& )
{
  return *this;
}

// ---------------------------------------------------------------------------

te::gm::GTFilter::~GTFilter() = default;

te::gm::GTFilter::GTFilter() = default;

bool te::gm::GTFilter::applyRansac(const std::string& transfName,
                                   const GTParameters& inputParams,
                                   const double allowedMaxDirectMapError,
                                   const double allowedMaxInverseMapError,
                                   const RansacIntegerCounterT& maxIterations,
                                   const bool enableMultiThread,
                                   const RANSACAreaRestrictionType areaRestT,
                                   const std::vector<double>& tiePointsWeights,
                                   std::vector< te::gm::GTParameters::TiePoint >& outTiePoints,
                                   std::unique_ptr< GeometricTransformation >& outTransf
                                   )
{
  if(allowedMaxDirectMapError < 0)
    return false;

  if(allowedMaxInverseMapError < 0)
    return false;

  outTiePoints.clear();
  outTransf.reset();  

  // generating the tie-points accumulated probabilities map
  // with positive values between 0 and 1
  std::map< double, GTParameters::TiePoint > tpsMap;
  {
    const unsigned int inputTPNmb = (unsigned int)
      inputParams.m_tiePoints.size();
    double tiePointsWeightsSum = 0;
      
    if( tiePointsWeights.size() > 0 )
    {
      if( tiePointsWeights.size() != inputParams.m_tiePoints.size () )
        return false;        
      
      // finding normalization factors

      unsigned int tpIdx = 0;

      for( tpIdx = 0 ; tpIdx < inputTPNmb ; ++tpIdx )
      {
        if( tiePointsWeights[ tpIdx ] < 0.0 ) return false;
        
        tiePointsWeightsSum += tiePointsWeights[ tpIdx ];
      }

      if( tiePointsWeightsSum <= 0.0 ) return false;

      // map fill 
      
      double newWSum = 0.0;
      double newW = 0.0;
      
      for( tpIdx = 0 ; tpIdx < inputTPNmb ; ++tpIdx )
      {
        newW = tiePointsWeights[ tpIdx ];
        newW /= tiePointsWeightsSum;

        newWSum += newW;

        tpsMap[ newWSum ] = inputParams.m_tiePoints[ tpIdx ];
      }
    }
    else
    {
      double wSum = 0;
      const double increment = 1.0 / 
        ( (double)inputParams.m_tiePoints.size() );
      
      for( unsigned int tpIdx = 0 ; tpIdx < inputTPNmb ; ++tpIdx )
      {
        wSum += increment;
        tpsMap[ wSum ] = inputParams.m_tiePoints[ tpIdx ];
      }
    }
  }

  // creating global execution parameters

  std::unique_ptr< const GeometricTransformation > auxTransfPtr;
  try
  {
    auxTransfPtr.reset( te::gm::GTFactory::make( transfName ) );
    if( auxTransfPtr.get() == nullptr ) return false;
  }
  catch(...)
  {
    return false;
  }

  RansacSetData bestSetData;
  bestSetData.m_maxDirectMapError = allowedMaxDirectMapError;
  bestSetData.m_maxInverseMapError = allowedMaxInverseMapError;
  bestSetData.m_area1 = 0;
  bestSetData.m_area2 = 0;

  const unsigned int procsNumber = te::common::GetPhysProcNumber();
  std::mutex syncMutex;
  bool returnValue = true;
  bool keepRunningFlag = true;
  RansacIntegerCounterT dynamicMaxIterations =
    maxIterations ?
      maxIterations
    :
      std::numeric_limits< RansacIntegerCounterT >::max();
  RansacIntegerCounterT dynamicMaxConsLowQualityIterations =
    getExcpectedIterationsNumber( auxTransfPtr->getMinRequiredTiePoints(),
    inputParams.m_tiePoints.size(), auxTransfPtr->getMinRequiredTiePoints() );
  RansacIntegerCounterT dynamicMaxConsInvalidIterations =
    getExcpectedIterationsNumber( auxTransfPtr->getMinRequiredTiePoints(),
    inputParams.m_tiePoints.size(), auxTransfPtr->getMinRequiredTiePoints() );
  RansacIntegerCounterT globalIterationsCounter = 0;
  double maxFoundGlobalItDMapError = 0.0;
  double maxFoundGlobalItIMapError = 0.0;
  RansacIntegerCounterT globalConsecutiveLowQualitySetsCounter = 0;
  RansacIntegerCounterT globalConsInvalidIterationsCounter = 0;
  
  ApplyRansacThreadEntryThreadParams baseThreadParams;
  baseThreadParams.m_transfName = transfName;
  baseThreadParams.m_inputGTParamsPtr = &inputParams;
  baseThreadParams.m_allowedMaxDirectMapError = allowedMaxDirectMapError;
  baseThreadParams.m_allowedMaxInverseMapError = allowedMaxInverseMapError;
  baseThreadParams.m_useDynamicIterationsNumber = ( maxIterations == 0 );
  baseThreadParams.m_dynamicMaxIterationsPtr = &dynamicMaxIterations;
  baseThreadParams.m_dynamicMaxConsLowQualityIterationsPtr = &dynamicMaxConsLowQualityIterations;
  baseThreadParams.m_dynamicMaxConsInvalidIterationsPtr = &dynamicMaxConsInvalidIterations;
  baseThreadParams.m_procsNumber = enableMultiThread ?
    ((RansacIntegerCounterT)procsNumber) : 1;
  baseThreadParams.m_returnValuePtr = &returnValue;
  baseThreadParams.m_mutexPtr = &syncMutex;
  baseThreadParams.m_keepRunningFlagPtr = &keepRunningFlag;
  baseThreadParams.m_tpsMapPtr = &tpsMap;
  baseThreadParams.m_bestSetDataPtr = &bestSetData;
    baseThreadParams.m_maxFoundGlobalItDMapErrorPtr = &maxFoundGlobalItDMapError;
  baseThreadParams.m_maxFoundGlobalItIMapErrorPtr = &maxFoundGlobalItIMapError;
  baseThreadParams.m_areaRestT = areaRestT;
  baseThreadParams.m_globalIterationsCounterPtr = &globalIterationsCounter;
  baseThreadParams.m_globalConsecutiveLowQualitySetsCounterPtr = &globalConsecutiveLowQualitySetsCounter;
  baseThreadParams.m_globalConsInvalidIterationsCounterPtr = &globalConsInvalidIterationsCounter;

  // Calling the ransac thread entry

  if( enableMultiThread )
  {
    std::vector< std::shared_ptr< std::thread > > threads;

    for( unsigned int threadIdx = 0 ; threadIdx < procsNumber ;
      ++threadIdx )
    {
      std::shared_ptr< std::thread > threadPtr( new std::thread(
        applyRansacThreadEntry, &baseThreadParams ) );

      threads.push_back( threadPtr );
    }

    for( unsigned int threadIdx = 0 ; threadIdx < procsNumber ;
      ++threadIdx )
    {
      threads[ threadIdx ]->join();
    }
  }
  else
  {
    applyRansacThreadEntry( &baseThreadParams );
  }

  if( returnValue )
  {
    try
    {
      outTransf.reset( te::gm::GTFactory::make( transfName ) );
    }
    catch(...)
    {
      return false;
    }

    if( outTransf.get() )
    {
      if( ! outTransf->initialize( bestSetData.m_transformationParams ) )
      {
        return false;
      }

      outTiePoints = bestSetData.m_tiePoins;

      return true;
    }
    else
    {
      return false;
    }
  }
  else
  {    
    return false;
  }

  return false;
}

void te::gm::GTFilter::applyRansacThreadEntry(te::gm::GTFilter::ApplyRansacThreadEntryThreadParams* paramsPtr)
{
  // locked region start -------------------------------------------------

  // accessing global shared objects

  std::unique_lock< std::mutex > locker( *paramsPtr->m_mutexPtr );

  const std::string transfName = paramsPtr->m_transfName;
  const GTParameters inputGTParams = *paramsPtr->m_inputGTParamsPtr;
  const double allowedMaxDirectMapError = paramsPtr->m_allowedMaxDirectMapError;
  const double allowedMaxInverseMapError = paramsPtr->m_allowedMaxInverseMapError;
  const bool useDynamicIterationsNumber = paramsPtr->m_useDynamicIterationsNumber;
  RansacIntegerCounterT& globalDynamicMaxIterations = (*(paramsPtr->m_dynamicMaxIterationsPtr));
  RansacIntegerCounterT& globalDynamicMaxConsLowQualityIterations =
    (*(paramsPtr->m_dynamicMaxConsLowQualityIterationsPtr));
  RansacIntegerCounterT& globalDynamicMaxConsInvalidIterations =
    (*(paramsPtr->m_dynamicMaxConsInvalidIterationsPtr));
  bool& globalKeepRunningFlag = (*(paramsPtr->m_keepRunningFlagPtr));
  std::map< double, GTParameters::TiePoint > tpsMap = *(paramsPtr->m_tpsMapPtr);
  RansacSetData& globalBestSetData = (*(paramsPtr->m_bestSetDataPtr));
  double& maxFoundGlobalItDMapError = (*(paramsPtr->m_maxFoundGlobalItDMapErrorPtr));
  double& maxFoundGlobalItIMapError = (*(paramsPtr->m_maxFoundGlobalItIMapErrorPtr));
  const RANSACAreaRestrictionType areaRestT = paramsPtr->m_areaRestT;
  RansacIntegerCounterT& globalIterationsCounter = (*(paramsPtr->m_globalIterationsCounterPtr));
  RansacIntegerCounterT& globalConsecutiveLowQualitySetsCounter =
    (*(paramsPtr->m_globalConsecutiveLowQualitySetsCounterPtr));
  RansacIntegerCounterT& globalConsInvalidIterationsCounter =
    (*(paramsPtr->m_globalConsInvalidIterationsCounterPtr));

  // initializing the random number generator to assure that each thread has
  // a different seed
  boost::random::mt19937 generator( (boost::random::mt19937::result_type)time(nullptr) );

  const std::vector< te::gm::GTParameters::TiePoint > tiePoints =
    paramsPtr->m_inputGTParamsPtr->m_tiePoints;

  locker.unlock();

  // locked region end -------------------------------------------------
  
  // globals

  std::unique_ptr< const GeometricTransformation > auxTransfPtr;
  try
  {
    auxTransfPtr.reset( te::gm::GTFactory::make( transfName ) );
  }
  catch(...)
  {
    auxTransfPtr.reset();
  }
  if( auxTransfPtr.get() == nullptr )
  {
    locker.lock();
    *(paramsPtr->m_returnValuePtr) = false;
    *(paramsPtr->m_keepRunningFlagPtr) = false;
    return;
  }

  const unsigned int reqTPsNmb = auxTransfPtr->getMinRequiredTiePoints();

  const unsigned int inputTPNmb = (unsigned int)tiePoints.size();
  if( inputTPNmb < reqTPsNmb )
  {
    locker.lock();
    *(paramsPtr->m_returnValuePtr) = false;
    *(paramsPtr->m_keepRunningFlagPtr) = false;
    return;
  }
  
  boost::random::uniform_01< double > distribution; 

  GTParameters consensusSetParams = inputGTParams;
  consensusSetParams.m_directParameters.clear();
  consensusSetParams.m_inverseParameters.clear();
  consensusSetParams.m_tiePoints.clear();

  std::vector< te::gm::GTParameters::TiePoint > consensusSetTiePoints;
  double consensusSetMaxDMapError = 0;
  double consensusSetMaxIMapError = 0;
  unsigned int consensusSetSize = 0;
  double consensusSetArea1 = 0.0;
  double consensusSetArea2 = 0.0;
  double tiePointDMapErr = 0;
  double tiePointIMapErr = 0;    
  std::map< double, GTParameters::TiePoint >::const_iterator tpsMapIt;
  unsigned int inputTpIdx = 0;
  RansacIntegerCounterT selectedTpsCounter = 0;
  std::vector< GTParameters::TiePoint* > selectedTpsPtrsVec;
  selectedTpsPtrsVec.resize( reqTPsNmb, nullptr );
  unsigned int selectedTpsPtrsVecIdx = 0;
  bool selectedTpsNotSelectedBefore = false;
  RansacIntegerCounterT dynamicMaxIterationsEstimation = 0;
  RansacIntegerCounterT dynamicMaxConsLowQualityIterationsEst = 0;
  RansacIntegerCounterT dynamicMaxConsInvalidterationsEst = 0;
  double randomValue = 0;
  double consensusSetMinXCoord1 = 0;
  double consensusSetMaxXCoord1 = 0;
  double consensusSetMinYCoord1 = 0;
  double consensusSetMaxYCoord1 = 0;
  double consensusSetMinXCoord2 = 0;
  double consensusSetMaxXCoord2 = 0;
  double consensusSetMinYCoord2 = 0;
  double consensusSetMaxYCoord2 = 0;

  while( true )
  {
//     {
//       locker.lock();
//       std::cout << std::endl
//         << " | globalIterationsCounter =" << globalIterationsCounter
//         << " | globalConsecutiveLowQualitySetsCounter =" << globalConsecutiveLowQualitySetsCounter
//         << " | globalConsInvalidIterationsCounter =" << globalConsInvalidIterationsCounter
//         << std::endl
//         ;
//       std::cout << std::endl
//         << " | globalDynamicMaxIterations =" << globalDynamicMaxIterations
//         << " | globalDynamicMaxConsLowQualityIterations =" << globalDynamicMaxConsLowQualityIterations
//         << " | globalDynamicMaxConsInvalidIterations =" << globalDynamicMaxConsInvalidIterations
//         << " | maxFoundGlobalItDMapError =" << maxFoundGlobalItDMapError
//         << " | maxFoundGlobalItIMapError =" << maxFoundGlobalItIMapError
//         << std::endl
//         ;
//       locker.unlock();
//     }

    // Trying to find a valid base consensus set
    // with the minimum number of required tie-points
    // Random selecting n distinc tpoints from the original set

    consensusSetParams.m_directParameters.clear();
    consensusSetParams.m_inverseParameters.clear();
    consensusSetParams.m_tiePoints.clear();

    selectedTpsCounter = 0;

    for( selectedTpsPtrsVecIdx = 0 ; selectedTpsPtrsVecIdx <
      reqTPsNmb ; ++selectedTpsPtrsVecIdx )
    {
      selectedTpsPtrsVec[ selectedTpsPtrsVecIdx ] = nullptr;
    }

    while( selectedTpsCounter < reqTPsNmb )
    {
      randomValue = distribution( generator  );
      tpsMapIt = tpsMap.upper_bound( randomValue );
      assert( tpsMapIt != tpsMap.end() );
        
      // Checking if this TP was already selected before
      
      selectedTpsNotSelectedBefore = true;
      
      for( selectedTpsPtrsVecIdx = 0 ; selectedTpsPtrsVecIdx <
        selectedTpsCounter ; ++selectedTpsPtrsVecIdx )
      {
        if( selectedTpsPtrsVec[ selectedTpsPtrsVecIdx ] ==
          &(tpsMapIt->second) )
        {
          selectedTpsNotSelectedBefore = false;
          break;
        }
      }
      
      if( selectedTpsNotSelectedBefore )
      {
        consensusSetParams.m_tiePoints.push_back( tpsMapIt->second );
      }
      
      ++selectedTpsCounter;
    }

    /* Trying to generate a valid base consensus transformation with the 
    selected points */      
    
    if( auxTransfPtr->computeParameters( consensusSetParams ) )
    {
      // finding those tie-points in agreement with the generated
      // consensus basic transformation

      consensusSetTiePoints.clear();
      consensusSetMaxDMapError = 0;
      consensusSetMaxIMapError = 0;
      consensusSetMinXCoord1 = std::numeric_limits< double >::max();
      consensusSetMaxXCoord1 = -1.0 * consensusSetMinXCoord1;
      consensusSetMinYCoord1 = consensusSetMinXCoord1;
      consensusSetMaxYCoord1 = consensusSetMaxXCoord1;
      consensusSetMinXCoord2 = consensusSetMinXCoord1;
      consensusSetMaxXCoord2 = consensusSetMaxXCoord1;
      consensusSetMinYCoord2 = consensusSetMinXCoord2;
      consensusSetMaxYCoord2 = consensusSetMaxXCoord2;

      for( inputTpIdx = 0 ; inputTpIdx < inputTPNmb ; ++inputTpIdx )
      {
        const GTParameters::TiePoint& curTP = tiePoints[ inputTpIdx ];

        tiePointDMapErr = auxTransfPtr->getDirectMappingError( curTP,
          consensusSetParams );
        tiePointIMapErr = auxTransfPtr->getInverseMappingError( curTP,
          consensusSetParams );

        if( ( tiePointDMapErr <= allowedMaxDirectMapError ) &&
          ( tiePointIMapErr <= allowedMaxInverseMapError ) )
        {
          consensusSetTiePoints.push_back( curTP );

          consensusSetMaxDMapError = std::max( tiePointDMapErr,
            consensusSetMaxDMapError );
          consensusSetMaxIMapError = std::max( tiePointIMapErr,
            consensusSetMaxIMapError );

          consensusSetMinXCoord1 = std::min( consensusSetMinXCoord1,
            curTP.first.x );
          consensusSetMaxXCoord1 = std::max( consensusSetMaxXCoord1,
            curTP.first.x );

          consensusSetMinYCoord1 = std::min( consensusSetMinYCoord1,
            curTP.first.y );
          consensusSetMaxYCoord1 = std::max( consensusSetMaxYCoord1,
            curTP.first.y );

          consensusSetMinXCoord2 = std::min( consensusSetMinXCoord2,
            curTP.second.x );
          consensusSetMaxXCoord2 = std::max( consensusSetMaxXCoord2,
            curTP.second.x );

          consensusSetMinYCoord2 = std::min( consensusSetMinYCoord2,
            curTP.second.y );
          consensusSetMaxYCoord2 = std::max( consensusSetMaxYCoord2,
            curTP.second.y );
        }
      }

      consensusSetSize = (unsigned int)consensusSetTiePoints.size();

      if( areaRestT == RANSAC_NO_AREA_REST )
      {
        consensusSetArea1 = 0;
        consensusSetArea2 = 0;
      }
      else if( areaRestT == RANSAC_BBOX_AREA_REST )
      {
        consensusSetArea1 = ( consensusSetMaxXCoord1 - consensusSetMinXCoord1 )
          * ( consensusSetMaxYCoord1 - consensusSetMinYCoord1 );
        consensusSetArea2 = ( consensusSetMaxXCoord2 - consensusSetMinXCoord2 )
          * ( consensusSetMaxYCoord2 - consensusSetMinYCoord2 );
      }
      else if( areaRestT == RANSAC_CHULL_AREA_REST )
      {
        getTPsConvexHullArea( consensusSetTiePoints, consensusSetArea1,
          consensusSetArea2 );
      }

      // locked region start -------------------------------------------------

      locker.lock();

      globalConsInvalidIterationsCounter = 0;

      /* Is this an acceptable consensus set ?? */

      if(
          ( consensusSetSize > globalBestSetData.m_tiePoins.size() )
          ||
          (
            ( consensusSetSize == globalBestSetData.m_tiePoins.size() )
            &&
            ( consensusSetMaxDMapError <= globalBestSetData.m_maxDirectMapError )
            &&
            ( consensusSetMaxIMapError <= globalBestSetData.m_maxInverseMapError )
            &&
            ( consensusSetArea1 >= globalBestSetData.m_area1 )
            &&
            ( consensusSetArea2 >= globalBestSetData.m_area2 )
          )
        )
      {
        globalConsecutiveLowQualitySetsCounter = 0;

//         std::cout << std::endl << std::endl
//           << " | global m_maxDirectMapError =" << globalBestSetData.m_maxDirectMapError
//           << " | global m_maxInverseMapError =" << globalBestSetData.m_maxInverseMapError
//           << " | global m_area1 =" << globalBestSetData.m_area1
//           << " | global m_area2 =" << globalBestSetData.m_area2
//           << " | global m_tiePoins.size =" << globalBestSetData.m_tiePoins.size()
//           << std::endl
//           ;
//         std::cout << std::endl
//           << " | consensusSetSize =" << consensusSetSize
//           << " | consensusSetMaxDMapError =" << consensusSetMaxDMapError
//           << " | consensusSetMaxIMapError =" << consensusSetMaxIMapError
//           << " | consensusSetArea1 =" << consensusSetArea1
//           << " | consensusSetArea2 =" << consensusSetArea2
//           << std::endl
//           ;

        maxFoundGlobalItDMapError = std::max( maxFoundGlobalItDMapError,
          consensusSetMaxDMapError );
        maxFoundGlobalItIMapError = std::max( maxFoundGlobalItIMapError,
          consensusSetMaxIMapError );

        globalBestSetData.m_maxDirectMapError = consensusSetMaxDMapError;
        globalBestSetData.m_maxInverseMapError = consensusSetMaxIMapError;
        globalBestSetData.m_area1 = consensusSetArea1;
        globalBestSetData.m_area2 = consensusSetArea2;
        globalBestSetData.m_transformationParams = consensusSetParams;
        globalBestSetData.m_tiePoins = consensusSetTiePoints;

        if( useDynamicIterationsNumber )
        {
          dynamicMaxConsLowQualityIterationsEst = getExcpectedIterationsNumber(
            globalBestSetData.m_tiePoins.size(), inputTPNmb, reqTPsNmb );

          globalDynamicMaxConsLowQualityIterations =
            ( globalDynamicMaxConsLowQualityIterations +
             dynamicMaxConsLowQualityIterationsEst ) / 2;

          globalDynamicMaxConsInvalidIterations =
            globalDynamicMaxConsLowQualityIterations;
        }
      }
      else
      {
        ++globalConsecutiveLowQualitySetsCounter;
      }

      locker.unlock();

      // locked region end -------------------------------------------------
    }
    else
    {
      locker.lock();
      ++globalConsInvalidIterationsCounter;
      locker.unlock();
    }

    // locked region start -------------------------------------------------

    locker.lock();

    // Updating iteration related variables

    ++globalIterationsCounter;

    if(
        ( globalBestSetData.m_tiePoins.size() == inputTPNmb )
        ||
        ( globalIterationsCounter > globalDynamicMaxIterations )
        ||
        ( globalConsecutiveLowQualitySetsCounter > globalDynamicMaxConsLowQualityIterations )
        ||
        ( globalConsInvalidIterationsCounter > globalDynamicMaxConsInvalidIterations )
      )
    {
      globalKeepRunningFlag = false;
    }

    if( ! globalKeepRunningFlag ) break;

    locker.unlock();

    // locked region end -------------------------------------------------
  } // global loop
}

void te::gm::GTFilter::getTPsConvexHullArea(
  const std::vector< GTParameters::TiePoint >& tiePoints,
  double& area1, double& area2 )
{
  area1 = 0;
  area2 = 0;

  if( tiePoints.size() >= 3 )
  {
    te::gm::MultiPoint points1( 0, te::gm::MultiPointType );
    te::gm::MultiPoint points2( 0, te::gm::MultiPointType );
    
    for( unsigned int tiePointsIdx = 0 ; tiePointsIdx < tiePoints.size() ;
      ++tiePointsIdx )
    {
      points1.add(new te::gm::Point(tiePoints[tiePointsIdx].first.x,
        tiePoints[ tiePointsIdx ].first.y ) );
      points2.add(new te::gm::Point(tiePoints[tiePointsIdx].second.x,
        tiePoints[ tiePointsIdx ].second.y ) );
    }
    
    te::gm::Surface* convexHull1Ptr = (te::gm::Surface*)points1.convexHull();
    if( convexHull1Ptr )
    {
      area1 = convexHull1Ptr->getArea();
      delete convexHull1Ptr;
    }

    te::gm::Surface* convexHull2Ptr = (te::gm::Surface*)points2.convexHull();
    if( convexHull2Ptr )
    {
      area2 = convexHull2Ptr->getArea();
      delete convexHull2Ptr;
    }
  }
}

te::gm::GTFilter::RansacIntegerCounterT te::gm::GTFilter::getExcpectedIterationsNumber(
  const unsigned int goodTPNumber,
  const unsigned int totalTPNumber,
  const unsigned int modelRequiredTPNumber
)
{
  RansacFloatCounterT expectedIterations = ((RansacFloatCounterT)goodTPNumber)
    / ((RansacFloatCounterT)totalTPNumber);
  expectedIterations = std::pow( expectedIterations,
    (RansacFloatCounterT)modelRequiredTPNumber );

  if( expectedIterations == 0.0 )
  {
    return std::numeric_limits< RansacIntegerCounterT >::max();
  }
  else
  {
    return (RansacIntegerCounterT)( 1.0 / expectedIterations );
  }
}
