/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/geometry/GeometryFunctions.cpp

  \brief A list of common used Geometry functions for the Geometry Module.
*/

// TerraLib
#include "GeometryFunctions.h"

#include "../sam/kdtree/Index.h"
#include "../sam/kdtree/Partition.h"

#include "CommonDataStructures.h"
#include "GeometryCollection.h"
#include "LinearRing.h"
#include "LineString.h"
#include "MultiLineString.h"
#include "Point.h"
#include "Polygon.h"
#include "Utils.h"
#include "Validation.h"
#include "validation/AddProjectedCoordinates.h"

//geos
#include <geos/algorithm/LineIntersector.h>
#include <geos/geom/Coordinate.h>
#include <geos/geom/LineSegment.h>

//STL
#include <algorithm>
#include <map>
#include <vector>

bool te::gm::Distance::IsEqual(double x1, double y1, double x2, double y2)
{
  if (x1 == x2 && y1 == y2)
  {
    return true;
  }

  return false;
}

bool te::gm::Distance::IsEqual(const te::gm::Coord2D& c1, const te::gm::Coord2D& c2)
{
  return IsEqual(c1.x, c1.y, c2.x, c2.y);
}

bool te::gm::Distance::IsEqual(const te::gm::Point& p1, const te::gm::Point& p2)
{
  return IsEqual(p1.getX(), p1.getY(), p2.getX(), p2.getY());
}

double te::gm::Distance::CalculateDistance(const te::gm::Coord2D& c1, const te::gm::Coord2D& c2)
{
  te::gm::Point point1(c1.x, c1.y);
  te::gm::Point point2(c2.x, c2.y);
  double distance = point1.distance(&point2);

  return distance;
}

double te::gm::Distance::CalculateSquaredDistance(const te::gm::Coord2D& c1, const te::gm::Coord2D& c2)
{
  //we do this to avoid calculating the root
  double deltaX = c1.x - c2.x;
  double deltaY = c1.y - c2.y;

  double deltaX2 = deltaX * deltaX;
  double deltaY2 = deltaY * deltaY;

  double squaredDistance = deltaX2 + deltaY2;
  return squaredDistance;
}

double te::gm::Distance::CalculateSquaredDistance(const te::gm::Point& p1, const te::gm::Point& p2)
{
  //we do this to avoid calculating the root
  double deltaX = p1.getX() - p2.getX();
  double deltaY = p1.getY() - p2.getY();

  double deltaX2 = deltaX * deltaX;
  double deltaY2 = deltaY * deltaY;

  double squaredDistance = deltaX2 + deltaY2;
  return squaredDistance;
}


bool te::gm::Distance::IsWithinDistance(const te::gm::Coord2D& c1, const te::gm::Coord2D& c2, double distance)
{
  //we do this to avoid calculating the root
  double distance2 = distance * distance;
  double squaredDistance = CalculateSquaredDistance(c1, c2);

  //avoinding the use of square root
  if (squaredDistance <= distance2)
  {
    return true;
  }

  return false;
}

bool te::gm::Distance::IsWithinDistance(const te::gm::Point& p1, const te::gm::Point& p2, double distance)
{
  //we do this to avoid calculating the root
  double distance2 = distance * distance;
  double squaredDistance = CalculateSquaredDistance(p1, p2);

  //avoinding the use of square root
  if (squaredDistance <= distance2)
  {
    return true;
  }

  return false;
}

bool te::gm::Distance::IsOnLine(const te::gm::Coord2D& coordToCheck, const te::gm::Coord2D& c1, const te::gm::Coord2D& c2)
{
#ifdef TERRALIB_GEOS_ENABLED

  geos::geom::Coordinate coordToCheckGeos(coordToCheck.x, coordToCheck.y);
  geos::geom::LineSegment segment(c1.x, c1.y, c2.x, c2.y);
  double currentDistance = segment.distance(coordToCheckGeos);
  if (currentDistance == 0)
  {
    return true;
  }

  return false;

#else
  throw Exception(TE_TR("IsOnLine routine is supported by GEOS! Please, enable the GEOS support."));
#endif
}

double te::gm::Distance::PerpendicularSquaredDistance(const te::gm::Coord2D& first, const te::gm::Coord2D& last, const te::gm::Coord2D& pin, te::gm::Coord2D& projectedCoordinate, bool& isOnSegment)
{
#ifdef TERRALIB_GEOS_ENABLED

  isOnSegment = false;

  double dx = last.x - first.x;
  double dy = last.y - first.y;
  //checks if the segment first and last coordinates are equal
  if (dx == 0. && dy == 0.)
  {
    projectedCoordinate.x = first.x;
    projectedCoordinate.y = first.y;

    isOnSegment = true;

    double squaredDistance = CalculateSquaredDistance(pin, first);
    return squaredDistance;
  }

  geos::geom::LineSegment segment(first.x, first.y, last.x, last.y);

  geos::geom::Coordinate geosPin(pin.x, pin.y);
  geos::geom::Coordinate geosRet;
  segment.project(geosPin, geosRet);

  double projectionFactor = segment.projectionFactor(geosRet);
  if (projectionFactor >= 0.0 && projectionFactor <= 1.0)
  {
    isOnSegment = true;
  }

  projectedCoordinate.x = geosRet.x;
  projectedCoordinate.y = geosRet.y;

  double squaredDistance = CalculateSquaredDistance(pin, projectedCoordinate);
  return squaredDistance;

#else
  throw Exception(TE_TR("PerpendicularDistance routine is supported by GEOS! Please, enable the GEOS support."));
#endif
}

double te::gm::Distance::PerpendicularDistance(const te::gm::Coord2D& first, const te::gm::Coord2D& last, const te::gm::Coord2D& pin, te::gm::Coord2D& projectedCoordinate, bool& isOnSegment)
{
  double squaredDistance = te::gm::Distance::PerpendicularSquaredDistance(first, last, pin, projectedCoordinate, isOnSegment);
  return std::sqrt(squaredDistance);
}

double te::gm::Distance::PerpendicularSquaredDistance(const te::gm::Point& first, const te::gm::Point& last, const te::gm::Point& pin, te::gm::Point& projectedPoint, bool& isOnSegment)
{
  te::gm::Coord2D cFirst(first.getX(), first.getY());
  te::gm::Coord2D cLast(last.getX(), last.getY());
  te::gm::Coord2D cPin(pin.getX(), pin.getY());
  te::gm::Coord2D cPinter;

  double distance = te::gm::Distance::PerpendicularSquaredDistance(cFirst, cLast, cPin, cPinter, isOnSegment);

  projectedPoint.setX(cPinter.getX());
  projectedPoint.setY(cPinter.getY());
  projectedPoint.setSRID(pin.getSRID());

  return distance;
}
double te::gm::Distance::PerpendicularDistance(const te::gm::Point& first, const te::gm::Point& last, const te::gm::Point& pin, te::gm::Point& projectedPoint, bool& isOnSegment)
{
  double squaredDistance = te::gm::Distance::PerpendicularSquaredDistance(first, last, pin, projectedPoint, isOnSegment);
  return std::sqrt(squaredDistance);
}

te::gm::GeometryVector te::gm::Distance::AddProjectedCoordinates(const te::gm::GeometryVectorConst& geometryVector, double distance, bool& wasChanged)
{
  te::gm::GeometryVector outputGeometryVector;

  te::gm::Envelope emptyEnvelope;
  std::unique_ptr<te::gm::KdTreeSingle> index(new te::gm::KdTreeSingle(emptyEnvelope));

  //we first create an index of all the points
  te::gm::SpatialIndex::populateCoordinateIndex(geometryVector, *index);

  std::unique_ptr<te::gm::AddProjectedCoordinates> validator(new te::gm::AddProjectedCoordinates());
  validator->setPrecision(distance);
  validator->setCoordinateIndex(index.release());

  //then we process each geometry
  for (std::size_t i = 0; i < geometryVector.size(); ++i)
  {
    bool currentWasChanged = false;
    std::unique_ptr<te::gm::Geometry> result(validator->validate(geometryVector.at(i), currentWasChanged));
    if (currentWasChanged == true)
    {
      if (te::gm::IsNullOrEmpty(result.get()) == false && result->isValid() == false)
      {
        result.reset(te::gm::Validation::MakeValidWithSnap(result.get()).release());
      }
    }

    wasChanged = wasChanged | currentWasChanged;

    outputGeometryVector.push_back(result.release());
  }

  return outputGeometryVector;
}

struct SegmentInfo
{
  SegmentInfo(const te::gm::Coord2D& c1, const te::gm::Coord2D& c2, std::size_t index)
    : m_index(index), m_c1(c1), m_c2(c2)
  {

  }
  
  te::gm::Coord2D m_c1;
  te::gm::Coord2D m_c2;
  std::size_t m_index;
};

void te::gm::SpatialIndex::populateCoordinateIndex(const te::gm::Geometry* geometry, KdTreeSingle& spatialIndex)
{
  std::vector<te::gm::Coord2D> vecCoords;
  getCoordinates(geometry, vecCoords);

  buildIndex(vecCoords, spatialIndex);
}

void te::gm::SpatialIndex::getCoordinates(const te::gm::Geometry* geometry, std::vector<te::gm::Coord2D>& vecCoords)
{
  if (geometry->isCollection() == true)
  {
    const te::gm::GeometryCollection* collection = static_cast<const te::gm::GeometryCollection*>(geometry);
    for (std::size_t i = 0; i < collection->getNumGeometries(); ++i)
    {
      getCoordinates(collection->getGeometryN(i), vecCoords);
    }
  }
  else if (geometry->get2DGeomTypeId() == te::gm::PolygonType)
  {
    const te::gm::Polygon* polygon = static_cast<const te::gm::Polygon*>(geometry);
    for (std::size_t i = 0; i < polygon->getNumRings(); ++i)
    {
      getCoordinates(polygon->getRingN(i), vecCoords);
    }
  }
  else if (geometry->get2DGeomTypeId() == te::gm::LineStringType)
  {
    const te::gm::LineString* lineString = static_cast<const te::gm::LineString*>(geometry);
    for (std::size_t i = 0; i < lineString->getNPoints(); ++i)
    {
      const double& x = lineString->getX(i);
      const double& y = lineString->getY(i);

      vecCoords.push_back(te::gm::Coord2D(x, y));
    }
  }
  else if (geometry->get2DGeomTypeId() == te::gm::PointType)
  {
    const te::gm::Point* point = static_cast<const te::gm::Point*>(geometry);
    const double& x = point->getX();
    const double& y = point->getY();

    vecCoords.push_back(te::gm::Coord2D(x, y));
  }
}

void te::gm::SpatialIndex::populateCoordinateIndex(const te::gm::GeometryVectorConst& geometryVector, KdTreeSingle& spatialIndex)
{
  std::vector<te::gm::Coord2D> vecCoords;
  for (std::size_t i = 0; i < geometryVector.size(); ++i)
  {
    getCoordinates(geometryVector.at(i), vecCoords);
  }

  buildIndex(vecCoords, spatialIndex);
}

struct LessThanX
{
  bool operator()(const te::gm::Coord2D& c1, const te::gm::Coord2D& c2) const
  {
    if (c1.x < c2.x)
    {
      return true;
    }
    return false;
  }
};

struct LessThanY
{
  bool operator()(const te::gm::Coord2D& c1, const te::gm::Coord2D& c2) const
  {
    if (c1.y < c2.y)
    {
      return true;
    }
    return false;
  }
};

void te::gm::SpatialIndex::buildIndex(const std::vector<te::gm::Coord2D>& vecCoords, te::gm::KdTreeSingle& spatialIndex)
{
  std::vector<std::pair<KdTreeSingle::kdKey, KdTreeSingle::kdData> > vecIndexInput;
  vecIndexInput.reserve(vecCoords.size());

  LessThanX lessThanX;
  LessThanY lessThanY;
  
  std::vector<te::gm::Coord2D> vecCoordsCopy(vecCoords.begin(), vecCoords.end());

  te::sam::kdtree::kdsort(vecCoordsCopy, 0, vecCoordsCopy.size()-1, 'x', lessThanX, lessThanY);

  for (std::size_t i = 0; i < vecCoordsCopy.size(); ++i)
  {
    std::pair<KdTreeSingle::kdKey, KdTreeSingle::kdData> keyValue(vecCoordsCopy.at(i), i);
    vecIndexInput.push_back(keyValue);
  }

  spatialIndex.buildOptimized(vecIndexInput);
}
