/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/geometry/Point.cpp

  \brief A point with x and y coordinate values.
*/

// TerraLib
#include "../core/translator/Translator.h"
#include "../Defines.h"
#include "../srs/Converter.h"
#include "Envelope.h"
#include "Exception.h"
#include "Point.h"


// STL
#include <cassert>

const std::string te::gm::Point::sm_typeName("Point");

const double te::gm::Point::sm_notNumber(TE_DOUBLE_NOT_A_NUMBER);

te::gm::Point::Point(int srid, GeomType t, Envelope* mbr)
  : Geometry(t, srid, mbr),
  m_x(sm_notNumber),
  m_y(sm_notNumber),
  m_z(sm_notNumber),
  m_m(sm_notNumber)
{
}

te::gm::Point::Point(const double& x, const double& y, int srid, GeomType t, Envelope* mbr)
  : Geometry(t, srid, mbr),
  m_x(x),
  m_y(y),
  m_z(sm_notNumber),
  m_m(sm_notNumber)
{
}

te::gm::Point::Point(const Point& rhs)

    = default;

te::gm::Point& te::gm::Point::operator=(const Point& rhs)
{
  if(this != &rhs)
  {
    Geometry::operator=(rhs);

    m_x = rhs.m_x;
    m_y = rhs.m_y;
    m_z = rhs.m_z;
    m_m = rhs.m_m;
  }

  return *this;
}

te::dt::AbstractData* te::gm::Point::clone() const
{
  return new Point(*this);
}

bool te::gm::Point::operator<(const Point& rhs)
{
  if (m_x < rhs.m_x) return true;
  if (m_x > rhs.m_x) return false;
  if (m_y < rhs.m_y) return true;
  if (m_y > rhs.m_y) return false;
  return false;
}

te::gm::Dimensionality te::gm::Point::getDimension() const _NOEXCEPT_OP(true)
{
  return te::gm::P;
}

const std::string& te::gm::Point::getGeometryType() const _NOEXCEPT_OP(true)
{
  return sm_typeName;
}

void te::gm::Point::setSRID(int srid) _NOEXCEPT_OP(true)
{
  m_srid = srid;
}

void te::gm::Point::convert(te::srs::Converter* converter)
{
#ifdef TERRALIB_MOD_SRS_ENABLED
  if (converter == nullptr)
  {
    throw Exception(TE_TR("Invalid Converter!"));
  }

  if (converter->getTargetSRID() == m_srid)
  {
    return;
  }

  if (converter->getSourceSRID() != m_srid)
  {
    converter->setSourceSRID(m_srid);
  }

  double x = getX();
  double y = getY();

  converter->convert(x, y);

  setX(x);
  setY(y);

  m_srid = converter->getTargetSRID();

  if (m_mbr)
    computeMBR(false);
#else
  throw Exception(TE_TR("transform method is not supported!"));
#endif // TERRALIB_MOD_SRS_ENABLED
}

void te::gm::Point::computeMBR(bool /*cascade*/) const _NOEXCEPT_OP(true)
{
  if (m_mbr == nullptr)
    m_mbr = new Envelope(getX(), getY(), getX(), getY());
  else
    m_mbr->init(getX(), getY(), getX(), getY());
}
