/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/geometry/Utils.cpp

  \brief Utility functions for the Geometry Module.
*/

// TerraLib
#include "../common/STLUtils.h"
#include "../common/UnitOfMeasure.h"
#include "../core/translator/Translator.h"
#include "../sam/rtree/Index.h"
#include "../srs/SpatialReferenceSystem.h"
#include "../srs/SpatialReferenceSystemManager.h"
#include "algorithms/SafeOverlay.h"
#include "CommonDataStructures.h"
#include "Curve.h"
#include "Envelope.h"
#include "Exception.h"
#include "Geometry.h"
#include "GeometryFactory.h"
#include "GeometryFunctions.h"
#include "GeometryCollection.h"
#include "GEOSGeometryFactory.h"
#include "GEOSReader.h"
#include "GEOSWriter.h"
#include "InternalConfig.h"
#include "Line.h"
#include "LinearRing.h"
#include "LineString.h"
#include "MultiLineString.h"
#include "MultiPolygon.h"
#include "Point.h"
#include "Polygon.h"
#include "Utils.h"
#include "Validation.h"

// STL
#include <cmath>
#include <memory>

#include <geos/algorithm/InteriorPointArea.h>
#include <geos/algorithm/LineIntersector.h>
#include <geos/algorithm/RayCrossingCounter.h>
#include <geos/geom/Coordinate.h>
#include <geos/geom/Geometry.h>
#include <geos/geom/IntersectionMatrix.h>
#include <geos/geom/LineSegment.h>
#include <geos/geom/LinearRing.h>
#include <geos/geom/Point.h>
#include <geos/geom/Polygon.h>
#include <geos/geom/LinearRing.h>
#if GEOS_FULLVERSION >= 30900
  #include <geos/geom/Position.h>
#else
  #include <geos/geomgraph/Position.h>
#endif
#include <geos/operation/intersection/Rectangle.h>
#include <geos/operation/intersection/RectangleIntersection.h>
#include <geos/operation/polygonize/Polygonizer.h>
#include <geos/operation/union/CascadedPolygonUnion.h>
#include <geos/operation/union/UnaryUnionOp.h>
#include <geos/operation/distance/DistanceOp.h>
#include <geos/operation/overlay/snap/GeometrySnapper.h>
#include <geos/operation/valid/IsValidOp.h>
#include <geos/operation/buffer/OffsetCurveBuilder.h>
#include <geos/operation/buffer/BufferParameters.h>
#include <geos/util/GEOSException.h>
#if ( GEOS_FULLVERSION < 30904 )
  // Version 3.9.4
  // Remove linear CascadedUnion #412  
  #include <geos/operation/union/CascadedUnion.h>
#endif
#include <geos/geom/GeometryFactory.h>

const std::string te::gm::Get2DGeometryType(const te::gm::GeomType& type)
{
  switch (type)
  {
  case te::gm::GeometryType:
  case te::gm::GeometryZType:
  case te::gm::GeometryMType:
  case te::gm::GeometryZMType:
    return "Geometry";

  case te::gm::PointType:
  case te::gm::PointZType:
  case te::gm::PointMType:
  case te::gm::PointZMType:
    return "Point";

  case te::gm::LineStringType:
  case te::gm::LineStringZType:
  case te::gm::LineStringMType:
  case te::gm::LineStringZMType:
    return "LineString";

  case te::gm::PolygonType:
  case te::gm::PolygonZType:
  case te::gm::PolygonMType:
  case te::gm::PolygonZMType:
    return "Polygon";

  case te::gm::GeometryCollectionType:
  case te::gm::GeometryCollectionZType:
  case te::gm::GeometryCollectionMType:
  case te::gm::GeometryCollectionZMType:
    return "GeometryCollection";

  case te::gm::MultiPointType:
  case te::gm::MultiPointZType:
  case te::gm::MultiPointMType:
  case te::gm::MultiPointZMType:
    return "MultiPoint";

  case te::gm::MultiLineStringType:
  case te::gm::MultiLineStringZType:
  case te::gm::MultiLineStringMType:
  case te::gm::MultiLineStringZMType:
    return "MultiLineString";

  case te::gm::MultiPolygonType:
  case te::gm::MultiPolygonZType:
  case te::gm::MultiPolygonMType:
  case te::gm::MultiPolygonZMType:
    return "MultiPolygon";

  case te::gm::CircularStringType:
  case te::gm::CircularStringZType:
  case te::gm::CircularStringMType:
  case te::gm::CircularStringZMType:
    return "CircularString";

  case te::gm::CompoundCurveType:
  case te::gm::CompoundCurveZType:
  case te::gm::CompoundCurveMType:
  case te::gm::CompoundCurveZMType:
    return "CompoundCurve";

  case te::gm::CurvePolygonType:
  case te::gm::CurvePolygonZType:
  case te::gm::CurvePolygonMType:
  case te::gm::CurvePolygonZMType:
    return "CurvePolygon";

  case te::gm::MultiSurfaceType:
  case te::gm::MultiSurfaceZType:
  case te::gm::MultiSurfaceMType:
  case te::gm::MultiSurfaceZMType:
    return "MultiSurface";

  default:
    return "UnknownGeometry";
  }
}

te::gm::GeomType te::gm::Get2DGeomTypeId(const te::gm::GeomType& type)
{
  switch (type)
  {
  case te::gm::GeometryType:
  case te::gm::GeometryZType:
  case te::gm::GeometryMType:
  case te::gm::GeometryZMType:
    return te::gm::GeometryType;

  case te::gm::PointType:
  case te::gm::PointZType:
  case te::gm::PointMType:
  case te::gm::PointZMType:
    return te::gm::PointType;

  case te::gm::LineStringType:
  case te::gm::LineStringZType:
  case te::gm::LineStringMType:
  case te::gm::LineStringZMType:
    return te::gm::LineStringType;

  case te::gm::PolygonType:
  case te::gm::PolygonZType:
  case te::gm::PolygonMType:
  case te::gm::PolygonZMType:
    return te::gm::PolygonType;

  case te::gm::GeometryCollectionType:
  case te::gm::GeometryCollectionZType:
  case te::gm::GeometryCollectionMType:
  case te::gm::GeometryCollectionZMType:
    return te::gm::GeometryCollectionType;

  case te::gm::MultiPointType:
  case te::gm::MultiPointZType:
  case te::gm::MultiPointMType:
  case te::gm::MultiPointZMType:
    return te::gm::MultiPointType;

  case te::gm::MultiLineStringType:
  case te::gm::MultiLineStringZType:
  case te::gm::MultiLineStringMType:
  case te::gm::MultiLineStringZMType:
    return te::gm::MultiLineStringType;

  case te::gm::MultiPolygonType:
  case te::gm::MultiPolygonZType:
  case te::gm::MultiPolygonMType:
  case te::gm::MultiPolygonZMType:
    return te::gm::MultiPolygonType;

  case te::gm::CircularStringType:
  case te::gm::CircularStringZType:
  case te::gm::CircularStringMType:
  case te::gm::CircularStringZMType:
    return te::gm::CircularStringType;

  case te::gm::CompoundCurveType:
  case te::gm::CompoundCurveZType:
  case te::gm::CompoundCurveMType:
  case te::gm::CompoundCurveZMType:
    return te::gm::CompoundCurveType;

  case te::gm::CurvePolygonType:
  case te::gm::CurvePolygonZType:
  case te::gm::CurvePolygonMType:
  case te::gm::CurvePolygonZMType:
    return te::gm::CurvePolygonType;

  case te::gm::MultiSurfaceType:
  case te::gm::MultiSurfaceZType:
  case te::gm::MultiSurfaceMType:
  case te::gm::MultiSurfaceZMType:
    return te::gm::MultiSurfaceType;

  default:
    return te::gm::UnknownGeometryType;
  }
}

te::gm::Envelope te::gm::CreateEnvelope(double x1, double y1, double x2, double y2)
{
  te::gm::Envelope envelope(std::min(x1, x2)
    , std::min(y1, y2)
    , std::max(x1, x2)
    , std::max(y1, y2));

  return envelope;
}

te::gm::Envelope te::gm::CreateEnvelope(const te::gm::Point& p1, const te::gm::Point& p2)
{
  te::gm::Envelope envelope(std::min(p1.getX(), p2.getX())
    , std::min(p1.getY(), p2.getY())
    , std::max(p1.getX(), p2.getX())
    , std::max(p1.getY(), p2.getY()));

  return envelope;
}

te::gm::Envelope te::gm::CreateEnvelope(const te::gm::Coord2D& coord, double extension)
{
  te::gm::Envelope envelope(coord.x - extension, coord.y - extension, coord.x + extension, coord.y + extension);
  return envelope;
}

std::unique_ptr<te::gm::Geometry> te::gm::GetGeometryUnion(const te::gm::GeometryVectorConst& geomVec, double tolerance)
{
  std::unique_ptr<te::gm::Geometry> geometry;

  if (geomVec.size() == 1)
  {
    te::dt::AbstractData* abs = geomVec[0]->clone();
    geometry.reset(static_cast<te::gm::Geometry*>(abs));
  }
  else if (geomVec.size() > 1)
  {
    bool wasChanged = false;
    te::gm::GeometryVector geomVecCopy = te::gm::PrepareGeometriesToIntersection(geomVec, wasChanged, tolerance);

    bool success = false;

    std::unique_ptr<te::gm::GeometryCollection> gc(new te::gm::GeometryCollection(0, te::gm::GeometryCollectionType, geomVec[0]->getSRID()));

    for (std::size_t i = 0; i < geomVecCopy.size(); ++i)
    {
      if (geomVecCopy[i] != nullptr)
      {
        gc->add(geomVecCopy[i]);
      }
    }

    try
    {
      geometry.reset(te::gm::UnaryUnion(gc.get()));
      success = true;
    }
    catch(...)
    {
    }

    if(success == false)
    {
      try
      {
        //buffer(0) will do the union
        std::unique_ptr<te::gm::Geometry> bufferedGeometry(gc->buffer(0));
      }
      catch(geos::util::GEOSException& e)
      {
        throw Exception(e.what());
      }
    }
  }

  return geometry;
}

te::gm::Geometry* te::gm::GetGeomFromEnvelope(const Envelope* const e, int srid)
{
// create an outer ring with the same envelope as our envelope
  LinearRing* r = new LinearRing(5, LineStringType, srid, new Envelope(*e));

  r->setPoint(0, e->m_llx, e->m_lly);
  r->setPoint(1, e->m_urx, e->m_lly);
  r->setPoint(2, e->m_urx, e->m_ury);
  r->setPoint(3, e->m_llx, e->m_ury);
  r->setPoint(4, e->m_llx, e->m_lly);

// create the polygon
  Polygon* p = new Polygon(1, PolygonType, srid, new Envelope(*e));
  p->setRingN(0, r);

  return p;
}

bool te::gm::SatisfySpatialRelation(const Geometry* g1,
                                    const Geometry* g2,
                                    SpatialRelation relation)
{
  switch(relation)
  {
    case CONTAINS:
      return g1->contains(g2);

    case COVEREDBY:
      return g1->coveredBy(g2);

    case COVERS:
      return g1->covers(g2);

    case CROSSES:
      return g1->crosses(g2);

    case DISJOINT:
      return g1->disjoint(g2);

    case EQUALS:
      return g1->equals(g2);

    case INTERSECTS:
      return g1->intersects(g2);

    case OVERLAPS:
      return g1->overlaps(g2);

    case TOUCHES:
      return g1->touches(g2);

    case WITHIN:
      return g1->within(g2);

    default:
      throw Exception(TE_TR("Invalid spatial relation!"));
  }
}

bool te::gm::RelationMatches(const std::string& relation, const std::string& relationPattern)
{
  bool matches = geos::geom::IntersectionMatrix::matches(relation, relationPattern);
  return matches;
}

te::gm::Envelope te::gm::AdjustToCut(const Envelope & env, double bWidth, double bHeight)
{
  double auxD;
  int auxI;

  int magicX;
  auxD = env.m_llx/bWidth;
  auxI = static_cast<int>(env.m_llx/bWidth);
  if (env.m_llx < 0 && (auxD - auxI) != 0)
    magicX = static_cast<int>(env.m_llx/bWidth - 1);
  else
    magicX = auxI;

  int magicY;
  auxD = env.m_lly/bHeight;
  auxI = static_cast<int>(env.m_lly/bHeight);
  if (env.m_lly < 0 && (auxD - auxI) != 0)
    magicY  = static_cast<int>(env.m_lly/bHeight - 1);
  else
    magicY  = auxI;

  double xi = magicX*bWidth;
  double yi = magicY*bHeight;

  int magicX2;
  auxD = env.m_urx/bWidth;
  auxI = static_cast<int>(env.m_urx/bWidth);
  if ((env.m_urx < 0) || (auxD - auxI) == 0)
    magicX2 = static_cast<int>(env.m_urx/bWidth);
  else
    magicX2 = static_cast<int>(env.m_urx/bWidth + 1);

  int magicY2;
  auxD = env.m_ury/bHeight;
  auxI = static_cast<int>(env.m_ury/bHeight);
  if ((env.m_ury < 0) || (auxD - auxI) == 0)
    magicY2 = static_cast<int>(env.m_ury/bHeight);
  else
    magicY2 = static_cast<int>(env.m_ury/bHeight + 1);

  double xf = (magicX2)*bWidth;
  double yf = (magicY2)*bHeight;

  return te::gm::Envelope(xi,yi,xf,yf);
}

te::gm::Coord2D* te::gm::locateAlong(const LineString* line, double initial, double final, double target)
{
  double tTof; // Distance of target to fist point

  std::map<int, double> pointLenghtFromFirst;

  double fullLineLenght = 0;

  pointLenghtFromFirst[0] = 0;
  for(std::size_t i = 1; i < line->getNPoints(); i++)
  {
    std::unique_ptr<Point> pA = line->getPointN(i);
    std::unique_ptr<Point> pB = line->getPointN(i - 1);

    fullLineLenght += pA->distance(pB.get());
    pointLenghtFromFirst[static_cast<int>(i)] = fullLineLenght;
  }

  double diference = final-initial;

  tTof = ((target-initial)*fullLineLenght)/diference;

  int onTheFly = -1;

  int pointBeforeTarget = -1;

  for(std::size_t i = 0; i < pointLenghtFromFirst.size(); i++)
  {
    if(pointLenghtFromFirst[static_cast<int>(i)] == tTof)
    {
      onTheFly = static_cast<int>(i);
      break;
    }
    if(tTof >= pointLenghtFromFirst[static_cast<int>(i)] && tTof < pointLenghtFromFirst[static_cast<int>(i+1)])
    {
      pointBeforeTarget = static_cast<int>(i);
      break;
    }
  }

  te::gm::Coord2D* targetCoord = new te::gm::Coord2D();

  if(onTheFly >= 0)
  {
    std::unique_ptr<Point> pOnTheFly = line->getPointN(static_cast<std::size_t>(onTheFly));
    targetCoord->x = pOnTheFly->getX();
    targetCoord->y = pOnTheFly->getY();

    return targetCoord;
  }

  std::unique_ptr<Point> p1 = line->getPointN(static_cast<std::size_t>(pointBeforeTarget));
  std::unique_ptr<Point> p2 = line->getPointN(static_cast<std::size_t>(pointBeforeTarget + 1));

  // Partial distance: target to p1
  double pd = tTof - pointLenghtFromFirst[pointBeforeTarget];

  // Total distance: p2 to p1
  double td = pointLenghtFromFirst[pointBeforeTarget+1] - pointLenghtFromFirst[pointBeforeTarget];

  targetCoord->x = ((pd*(p2->getX()-p1->getX()))/td)+p1->getX();

  targetCoord->y = ((pd*(p2->getY()-p1->getY()))/td)+p1->getY();

  return targetCoord;
}

void te::gm::Multi2Single(const te::gm::Geometry* g, std::vector<te::gm::Geometry*>& geoms)
{
  const te::gm::GeometryCollection* gc = dynamic_cast<const te::gm::GeometryCollection*>(g);
  if (gc)
  {
    for (std::size_t i = 0; i < gc->getNumGeometries(); ++i)
    {
      te::gm::Geometry* currentGeom = gc->getGeometryN(i);
      currentGeom->setSRID(gc->getSRID());
      Multi2Single(currentGeom, geoms);
    }
  }
  else
    geoms.push_back(dynamic_cast<te::gm::Geometry*>(g->clone()));
}

te::gm::Geometry* te::gm::CascadedPolygonUnion(const std::vector<Polygon*>& polygonVector)
{
#ifdef TERRALIB_GEOS_ENABLED

  std::vector<geos::geom::Polygon*> geosPolygonVector;

  try
  {
    if(polygonVector.empty())
      return nullptr;

    int srid = polygonVector[0]->getSRID();

    for(auto i : polygonVector)
    {
      if(i->getGeomTypeId() == te::gm::PolygonType)
      {
        geos::geom::Polygon* geosPolygon = GEOSWriter::write(i);
        geosPolygonVector.push_back(geosPolygon);
      }
    }

    std::unique_ptr<geos::geom::Geometry> unionGeom(geos::operation::geounion::CascadedPolygonUnion::Union(&geosPolygonVector));

    te::common::FreeContents(geosPolygonVector);

    if(unionGeom == nullptr)
      return nullptr;

    unionGeom->setSRID(srid);

    return GEOSReader::read(unionGeom.get());
  }
  catch(geos::util::GEOSException& e)
  {
    te::common::FreeContents(geosPolygonVector);
    throw Exception(e.what());
  }

#else
  throw te::common::Exception(TE_TR("CascadedPolygonUnion routine is supported by GEOS! Please, enable the GEOS support."));
#endif
}

te::gm::Geometry* te::gm::CascadedUnion(const std::vector<te::gm::Geometry*>& vecGeometries)
{
#ifdef TERRALIB_GEOS_ENABLED

  if (vecGeometries.empty() == true)
  {
    return nullptr;
  }

  std::vector<geos::geom::Geometry*> vecGeometriesGEOS;
  
  try
  {
    int srid = vecGeometries.at(0)->getSRID();
    
    for(std::size_t i = 0; i < vecGeometries.size(); ++i)
    {
      te::gm::Geometry* geometry = vecGeometries.at(i);
      geos::geom::Geometry* geosPolygon = GEOSWriter::write(geometry);
      vecGeometriesGEOS.push_back(geosPolygon);
    }        

    #if ( GEOS_FULLVERSION >= 30904 )
      // Version 3.9.4
      // Remove linear CascadedUnion #412  
      std::unique_ptr<geos::geom::Geometry> unionGeom(geos::operation::geounion::UnaryUnionOp::Union(  vecGeometriesGEOS ) );
    #else
      std::unique_ptr<geos::geom::Geometry> unionGeom(geos::operation::geounion::CascadedUnion::Union(&vecGeometriesGEOS));
    #endif 
      
    te::common::FreeContents(vecGeometriesGEOS);  

    if (unionGeom == nullptr)
      return nullptr;

    unionGeom->setSRID(srid);

    return GEOSReader::read(unionGeom.get());
  }
  catch (geos::util::GEOSException& e)
  {
    te::common::FreeContents(vecGeometriesGEOS);
    throw Exception(e.what());
  }

#else
  throw te::common::Exception(TE_TR("CascadedUnion routine is supported by GEOS! Please, enable the GEOS support."));
#endif
}

te::gm::Geometry* te::gm::UnaryUnion(te::gm::Geometry* geom)
{
#ifdef TERRALIB_GEOS_ENABLED

  try
  {
    std::unique_ptr<geos::geom::Geometry> geomGeos(GEOSWriter::write(geom));

    std::unique_ptr<geos::geom::Geometry> unionGeom = geos::operation::geounion::UnaryUnionOp::Union(*(geomGeos.get()));

    unionGeom->setSRID(geom->getSRID());

    return GEOSReader::read(unionGeom.get());
  }
  catch(geos::util::GEOSException& e)
  {
    throw Exception(e.what());
  }

#else
  throw te::common::Exception(TE_TR("Union routine is supported by GEOS! Please, enable the GEOS support."));
#endif
}

void te::gm::ClosestPoints(te::gm::Geometry* geomA, te::gm::Geometry* geomB,
  te::gm::Coord2D& coordA, te::gm::Coord2D& coordB)
{
#ifdef TERRALIB_GEOS_ENABLED
  std::unique_ptr<geos::geom::Geometry> thisGeomA(GEOSWriter::write(geomA));
  std::unique_ptr<geos::geom::Geometry> thisGeomB(GEOSWriter::write(geomB));

  geos::operation::distance::DistanceOp op(thisGeomA.get(), thisGeomB.get());

  std::unique_ptr<geos::geom::CoordinateSequence> cs(op.nearestPoints());

  if (cs->getSize() == 2)
  {
    coordA = te::gm::Coord2D(cs->getAt(0).x, cs->getAt(0).y);
    coordB = te::gm::Coord2D(cs->getAt(1).x, cs->getAt(1).y);
  }

#else
  throw te::common::Exception(TE_TR("getCentroid routine is supported by GEOS! Please, enable the GEOS support."));
#endif
}

te::gm::Line* te::gm::GetIntersectionLine(te::gm::Geometry* geom, te::gm::Coord2D coord)
{
  std::vector<te::gm::Geometry*> geomVec;
  std::unique_ptr<te::gm::Point> pRef(new te::gm::Point(coord.x, coord.y, geom->getSRID()));

  te::gm::Multi2Single(geom, geomVec);

  te::gm::Line* line = nullptr;

  for (std::size_t t = 0; t < geomVec.size(); ++t)
  {
    te::gm::Geometry* g = geomVec[t];

    if (g->getGeomTypeId() == te::gm::LineStringType)
    {
      te::gm::LineString* ls = static_cast<te::gm::LineString*>(g);

      if (ls->getNPoints() == 2)
      {
        std::unique_ptr<te::gm::Point> p0 = ls->getPointN(0);
        std::unique_ptr<te::gm::Point> p1 = ls->getPointN(1);

        line = new te::gm::Line(te::gm::LineStringType, geom->getSRID());

        line->setCoord(0, p0->getX(), p0->getY());
        line->setCoord(1, p1->getX(), p1->getY());

        break;
      }
      else
      {
        for (std::size_t p = 0; p < ls->getNPoints() - 1; ++p)
        {
          std::unique_ptr<te::gm::Point> p0 = ls->getPointN(p);
          std::unique_ptr<te::gm::Point> p1 = ls->getPointN(p + 1);

          std::unique_ptr<te::gm::Line> lineAux(new te::gm::Line(te::gm::LineStringType, geom->getSRID()));
          lineAux->setCoord(0, p0->getX(), p0->getY());
          lineAux->setCoord(1, p1->getX(), p1->getY());

          std::unique_ptr<te::gm::Geometry> geomMBR(te::gm::GetGeomFromEnvelope(lineAux->getMBR(), geom->getSRID()));

          if (geomMBR->intersects(pRef.get()))
          {
            line = lineAux.release();
            break;
          }
        }
      }
    }
  }

  return line;
}

double te::gm::GetAngle(te::gm::Coord2D coordA, te::gm::Coord2D coordB)
{
  if (coordA.y == coordB.y)
  {
    return 0.0;
  }

  double dx = (coordA.x - coordB.x);
  double dy = (coordA.y - coordB.y);
  
  double rad = std::atan2(dy, dx);

  double pi = atan(1) * 4;
  double angle = rad * 180 / pi + 180;
  
  return angle;
}

bool te::gm::Rotate(te::gm::Coord2D pr, te::gm::LineString* l, double angle, te::gm::LineString* lOut)
{
  double alfa;
  double dx, dy;
  double x, y;
  double xr, yr;

  try
  {
    if (l->size() < 2)
      return (false);

    //transform Degree to Radius
    alfa = (4.*std::atan(1.)*angle) / 180.;

    dx = pr.getX();
    dy = pr.getY();

    for (std::size_t count = 0; count < l->size(); count++)
    {
      std::unique_ptr<te::gm::Point> curPoint = l->getPointN(count);

      x = curPoint->getX() - dx;
      y = curPoint->getY() - dy;

      xr = x * std::cos(alfa) - y * std::sin(alfa);
      yr = x * std::sin(alfa) + y * std::cos(alfa);

      lOut->setPoint(count, xr + dx, yr + dy);
    }
  }
  catch (...)
  {
    return false;
  }

  return true;
}

bool te::gm::AdjustSegment(te::gm::Point* P0, te::gm::Point* P1, double d0, te::gm::Coord2D& P0out, te::gm::Coord2D& P1out)
{
  try
  {
    if (P0->getX() == P1->getX() && P0->getY() == P1->getY())
      return false;

    te::gm::Coord2D vL1((P1->getX() - P0->getX()), (P1->getY() - P0->getY()));
    te::gm::Coord2D vL2(-1 * (P1->getX() - P0->getX()), -1 * (P1->getY() - P0->getY()));

    double vL_norm1 = std::sqrt(vL1.getX() * vL1.getX() + vL1.getY() * vL1.getY());
    double vL_norm2 = std::sqrt(vL2.getX() * vL2.getX() + vL2.getY() * vL2.getY());

    te::gm::Coord2D uL1((vL1.getX() / vL_norm1), (vL1.getY() / vL_norm1));
    te::gm::Coord2D uL2((vL2.getX() / vL_norm2), (vL2.getY() / vL_norm2));

    std::unique_ptr<te::gm::Point> pFim1(new te::gm::Point(P0->getX() + uL1.getX() * d0, P0->getY() + uL1.getY() * d0, P0->getSRID()));
    std::unique_ptr<te::gm::Point> pFim2(new te::gm::Point(P0->getX() + uL2.getX() * d0, P0->getY() + uL2.getY() * d0, P1->getSRID()));

    if (pFim1->distance(P1) <= pFim2->distance(P1))
    {
      P0out.x = P0->getX();
      P0out.y = P0->getY();
      P1out.x = pFim1->getX();
      P1out.y = pFim1->getY();
    }
    else
    {
      P0out.x = P0->getX();
      P0out.y = P0->getY();
      P1out.x = pFim2->getX();
      P1out.y = pFim2->getY();
    }
  }
  catch (...)
  {
    return false;
  }

  return true;
}

void te::gm::Polygonizer(te::gm::Geometry* g, std::vector<te::gm::Polygon*>& pols)
{
#ifdef TERRALIB_GEOS_ENABLED

  geos::operation::polygonize::Polygonizer polygonizer;

  std::unique_ptr<geos::geom::Geometry> geomGeos(GEOSWriter::write(g));
  polygonizer.add(geomGeos.get());
  
  std::unique_ptr<std::vector<std::unique_ptr<geos::geom::Polygon>>> vecpolGeos;
#if GEOS_FULLVERSION >= 30900
  vecpolGeos.reset( new std::vector<std::unique_ptr<geos::geom::Polygon> >() );
  *vecpolGeos = polygonizer.getPolygons();
#elif GEOS_FULLVERSION >= 30800
  vecpolGeos = polygonizer.getPolygons();  
#else
  vecpolGeos.reset( new std::vector<std::unique_ptr<geos::geom::Polygon>() );
  
  std::vector<geos::geom::Polygon*>* vecTempPols = polygonizer.getPolygons();

  for (geos::geom::Polygon* tempPolygon : *vecTempPols)
  {
    vecpolGeos->emplace_back(std::unique_ptr<geos::geom::Polygon>(tempPolygon));
  }

  //we just delete the external vector. The internal elements have been assigned to an unique_ptr
  delete vecTempPols;
#endif

  for (std::size_t i = 0; i < vecpolGeos->size(); i++)
  {
    vecpolGeos->at(i)->setSRID(g->getSRID());
    pols.push_back(GEOSReader::read(vecpolGeos->at(i).get()));
  }

#else
  throw te::common::Exception(TE_TR("Polygonizer routine is supported by GEOS! Please, enable the GEOS support."));
#endif
}

void te::gm::Polygonizer(const te::gm::GeometryVector& vecInput, te::gm::GeometryVector& vecOutput)
{
  if (vecInput.empty() == true)
  {
    return;
  }

#ifdef TERRALIB_GEOS_ENABLED

  try
  {
    geos::operation::polygonize::Polygonizer polygonizer;

    int srid = vecInput.at(0)->getSRID();

    std::vector<std::unique_ptr<geos::geom::Geometry>> vecGeosGeometries;
    for (std::size_t i = 0; i < vecInput.size(); ++i)
    {
      const te::gm::Geometry* inputGeometry = vecInput.at(i);
      std::unique_ptr<geos::geom::Geometry> geomGeos(GEOSWriter::write(inputGeometry));
      polygonizer.add(geomGeos.get());


      vecGeosGeometries.push_back(std::move(geomGeos));
    }

    std::unique_ptr<std::vector<std::unique_ptr<geos::geom::Polygon>>> vecpolGeos;
#if GEOS_FULLVERSION >= 30900
    vecpolGeos.reset( new std::vector<std::unique_ptr<geos::geom::Polygon> >() );
    *vecpolGeos = polygonizer.getPolygons();
#elif GEOS_FULLVERSION >= 30800
    vecpolGeos = polygonizer.getPolygons();      
#else
    vecpolGeos.reset( new std::vector<std::unique_ptr<geos::geom::Polygon>() );
    
    std::vector<geos::geom::Polygon*>* vecTempPols = polygonizer.getPolygons();

    for (geos::geom::Polygon* tempPolygon : *vecTempPols)
    {
      vecpolGeos->emplace_back(std::unique_ptr<geos::geom::Polygon>(tempPolygon));
    }

    //we just delete the external vector. The internal elements have been assigned to an unique_ptr
    delete vecTempPols;
#endif

    for (std::size_t i = 0; i < vecpolGeos->size(); ++i)
    {
      te::gm::Geometry* geometry = GEOSReader::read(static_cast<geos::geom::Geometry*>(vecpolGeos->at(i).get()));
      geometry->setSRID(srid);

      vecOutput.push_back(geometry);
    }
  }
  catch (geos::util::GEOSException& e)
  {
    std::string what = e.what();
    throw Exception("Exception thrown in 'Polygonizer': " + what);
  }

#else
  throw te::common::Exception(TE_TR("Polygonizer routine is supported by GEOS! Please, enable the GEOS support."));
#endif
}

void te::gm::AddPolygon(te::gm::Polygon* polygon, te::gm::GeometryVector& pAdd)
{
#ifdef TERRALIB_GEOS_ENABLED

  AddLineString(dynamic_cast<te::gm::LineString*>(polygon->getExteriorRing()), pAdd);

  for (std::size_t n = polygon->getNumInteriorRings(); n-- > 0;){
    AddLineString(dynamic_cast<te::gm::LineString*>(polygon->getInteriorRingN(n)), pAdd);
  }

#else
  throw Exception(TE_TR("addPolygon routine is supported by GEOS! Please, enable the GEOS support."));
#endif
}

void te::gm::AddLineString(te::gm::LineString* lineString, te::gm::GeometryVector& pAdd)
{
#ifdef TERRALIB_GEOS_ENABLED
  std::unique_ptr<geos::geom::LineString> ls;

  te::gm::LinearRing* teLinearRing = dynamic_cast<te::gm::LinearRing*>(lineString);
  if (teLinearRing)
  {
    std::unique_ptr<geos::geom::LinearRing> geosLinearRing(GEOSWriter::write(teLinearRing));

    std::unique_ptr<geos::geom::CoordinateSequence> geosCoordSeq(geosLinearRing->getCoordinates());

    ls.reset(GEOSGeometryFactory::getGeomFactory()->createLineString(geosCoordSeq.release()));
  }
  else
  {
    ls.reset((GEOSWriter::write(lineString)));
  }

  // unioning the linestring with the point makes any self intersections explicit.
  std::unique_ptr<geos::geom::Point> point(GEOSGeometryFactory::getGeomFactory()->createPoint(ls->getCoordinateN(0)));

  std::unique_ptr<geos::geom::Geometry> toAdd(ls->Union(point.get()));

  pAdd.push_back(GEOSReader::read(toAdd.get()));

#else
  throw Exception(TE_TR("addLineString routine is supported by GEOS! Please, enable the GEOS support."));
#endif
}

te::gm::Geometry* te::gm::PrepareGeometriesToIntersection(const te::gm::Geometry* geom_A, te::gm::Geometry* geomB, bool& wasChanged, double tolerance)
{
  return PrepareGeometriesToIntersection(geom_A, geomB, true, wasChanged, tolerance);
}

te::gm::Geometry* te::gm::PrepareGeometriesToIntersection(const te::gm::Geometry* geom_A, te::gm::Geometry* geomB, bool checkValidityAndFix, bool& wasChanged, double tolerance)
{
  te::gm::GeometryVector vecGeomB;
  vecGeomB.push_back(geomB);

  return te::gm::PrepareGeometriesToIntersection(geom_A, vecGeomB, checkValidityAndFix, wasChanged, tolerance);
}

te::gm::Geometry* te::gm::PrepareGeometriesToIntersection(const te::gm::Geometry* geom_A, const te::gm::GeometryVector& vecGeomB, bool& wasChanged, double tolerance)
{
  return te::gm::PrepareGeometriesToIntersection(geom_A, vecGeomB, true, wasChanged, tolerance);
}

te::gm::Geometry* te::gm::PrepareGeometriesToIntersection(const te::gm::Geometry* geom_A, const te::gm::GeometryVector& vecGeomB, bool checkValidityAndFix, bool& wasChanged, double tolerance)
{
  wasChanged = false;

#ifdef TERRALIB_GEOS_ENABLED

  if (te::gm::IsNullOrEmpty(geom_A) == true)
  {
    return nullptr;
  }

  te::gm::Geometry* geomResult = nullptr;

  GeomType geomTypeA = geom_A->get2DGeomTypeId();
  if (geom_A->getDimension() == te::gm::P)
  {
    geomResult = dynamic_cast<te::gm::Geometry*>(geom_A->clone());
    return geomResult;
  }

  // Check if MBRs of both LineStrings have intersection.
  const te::gm::Envelope* env_A = geom_A->getMBR();

  //for optimizations purposes, we are going to use a fast clipping in all adjacent geometries to reduce the number of points
  //in this clipping, to avoid discarding any point that will be useful in the next steped, we must expand the clipping envelope considering some issues:
  //1 - we need to double the tolerance (as radius the radious area going to be used)
  //2 - to be extra conservative, we expand the result tolerance in 1% to avoid discarding points exactly in the same radious location
  double expandingDelta = 2 * tolerance * 1.01;
  te::gm::Envelope expandedEnvelopeGeometryA(
    env_A->m_llx - expandingDelta,
    env_A->m_lly - expandingDelta,
    env_A->m_urx + expandingDelta,
    env_A->m_ury + expandingDelta);

  std::vector<te::gm::MultiLineString*> multiLineStringVectorA = GetAsMultiLineStringVector(*geom_A);

  // Creating the RTree with the ls_B segment boxes intersects ls_A segment box
  te::sam::rtree::Index<std::size_t, 8> rtree;
  std::vector<te::gm::LineString> rtreeSegments;

  //we decompose all geometries in vecGeomB into line segments, and then we index it
  for (std::size_t i = 0; i < vecGeomB.size(); ++i)
  {
    te::gm::Geometry* geometryB = vecGeomB[i];
    if (te::gm::IsNullOrEmpty(geometryB) == true)
    {
      continue;
    }

    std::unique_ptr<te::gm::Geometry> geometryBCloned(static_cast<te::gm::Geometry*>(geometryB->clone()));
    if (geometryBCloned->getMBR()->within(expandedEnvelopeGeometryA) == false)
    {
      //here we clip the geometry B. We do not need to ensule validation because we are just going to use its segments list
      geometryBCloned.reset(te::gm::SafeOverlay::ClipGeometry(geometryBCloned.get(), expandedEnvelopeGeometryA));
      if (te::gm::IsNullOrEmpty(geometryBCloned.get()) == true)
      {
        continue;
      }
    }

    std::vector<te::gm::MultiLineString*> multiLineStringVectorB = GetAsMultiLineStringVector(*geometryBCloned.get());
    for (std::size_t j = 0; j < multiLineStringVectorB.size(); ++j)
    {
      te::gm::MultiLineString* multiLineB = multiLineStringVectorB[j];
      const std::vector<te::gm::Geometry*>& vecGeometries = multiLineB->getGeometries();

      for (std::size_t k = 0; k < vecGeometries.size(); ++k)
      {
        te::gm::LineString* lineB = static_cast<te::gm::LineString*>(vecGeometries[k]);

        //finally we get all the segments
        std::size_t nSegmentsB = lineB->size() - 1;
        for (std::size_t s = 0; s < nSegmentsB; ++s)
        {
          te::gm::Envelope currentSegmentEnvelope = CreateEnvelope(lineB->getX(s), lineB->getY(s), lineB->getX(s + 1), lineB->getY(s + 1));

          te::gm::LineString segment_B(2, te::gm::LineStringType, geom_A->getSRID());

          segment_B.setPoint(0, lineB->getX(s), lineB->getY(s));
          segment_B.setPoint(1, lineB->getX(s + 1), lineB->getY(s + 1));

          rtree.insert(currentSegmentEnvelope, rtreeSegments.size());
          rtreeSegments.push_back(segment_B);
        }
      }
    }
    te::common::FreeContents(multiLineStringVectorB);
  }

  for (std::size_t i = 0; i < multiLineStringVectorA.size(); ++i)
  {
    bool currentWasChanged = false;
    std::unique_ptr<te::gm::MultiLineString> currentResult(PrepareGeometriesToIntersection(*multiLineStringVectorA[i], rtree, rtreeSegments, currentWasChanged, tolerance));

    if (currentWasChanged == true)
    {
      wasChanged = true;
      delete multiLineStringVectorA[i];
      multiLineStringVectorA[i] = currentResult.release();
    }
  }

  if (wasChanged)
  {
    std::unique_ptr<te::gm::Geometry> preparedGeometryResult(MultiLineToDefinedType(multiLineStringVectorA, geomTypeA));

    if (checkValidityAndFix == true)
    {
      te::gm::TopologyValidationError error;
      bool result = te::gm::Validation::CheckValidity(preparedGeometryResult.get(), error);
      if (result == false)
      {
        preparedGeometryResult.reset(te::gm::Validation::MakeValidWithSnap(preparedGeometryResult.get()).release());
      }
    }

    geomResult = preparedGeometryResult.release();
  }
  else
  {
    geomResult = dynamic_cast<te::gm::Geometry*>(geom_A->clone());
  }


  te::common::FreeContents(multiLineStringVectorA);

  return geomResult;

#else
  throw Exception(TE_TR("PrepareGeometriesToIntersection routine is supported by GEOS! Please, enable the GEOS support."));
#endif
}

te::gm::GeometryVector te::gm::PrepareGeometriesToIntersection(const te::gm::GeometryVectorConst& vecGeometries, bool& wasChanged, double tolerance)
{
  //creates the geometry spatial index
  te::gm::GeometryVector vecResultGeometries;
  te::common::Clone(vecGeometries, vecResultGeometries);

  wasChanged = false;
  //we first create an index containing all the geometries
  te::sam::rtree::Index<std::size_t, 8> rtree;
  
  for (std::size_t i = 0; i < vecResultGeometries.size(); ++i)
  {
    te::gm::Geometry* geometryA = vecResultGeometries.at(i);
    if (te::gm::IsNullOrEmpty(geometryA) == true)
    {
      continue;
    }

    if (geometryA->getDimension() == te::gm::P)
    {
      continue;
    }

    // Check if MBRs of both LineStrings have intersection.
    const te::gm::Envelope* envelopeA = geometryA->getMBR();
    te::gm::Envelope expandedEnvelopeA(envelopeA->m_llx - tolerance,
      envelopeA->m_lly - tolerance,
      envelopeA->m_urx + tolerance,
      envelopeA->m_ury + tolerance);

    rtree.insert(expandedEnvelopeA, i);
  }

  for (std::size_t i = 0; i < vecResultGeometries.size(); ++i)
  {
    te::gm::Geometry* geometryA = vecResultGeometries.at(i);
    if (te::gm::IsNullOrEmpty(geometryA) == true)
    {
      continue;
    }

    if (geometryA->getDimension() == te::gm::P)
    {
      continue;
    }

    const te::gm::Envelope* envelopeA = geometryA->getMBR();
    te::gm::Envelope expandedEnvelopeA(envelopeA->m_llx - tolerance,
      envelopeA->m_lly - tolerance,
      envelopeA->m_urx + tolerance,
      envelopeA->m_ury + tolerance);

    std::vector<std::size_t> vecIndexes;
    rtree.search(expandedEnvelopeA, vecIndexes);

    te::gm::GeometryVector vecCandidates;
    for (std::size_t j = 0; j < vecIndexes.size(); ++j)
    {
      std::size_t currentCandidateIndex = vecIndexes.at(j);
      if (currentCandidateIndex == i)
      {
        continue;
      }

      te::gm::Geometry* candidateGeometry = vecResultGeometries.at(currentCandidateIndex);
      if (te::gm::IsNullOrEmpty(candidateGeometry) == false)
      {
        vecCandidates.push_back(candidateGeometry);
      }
    }

    if (vecCandidates.empty() == true)
    {
      continue;
    }

    bool currentWasChanged = false;
    std::unique_ptr<te::gm::Geometry> geometryAFixed(te::gm::PrepareGeometriesToIntersection(geometryA, vecCandidates, currentWasChanged, tolerance));
    if (currentWasChanged == true)
    {
      if (te::gm::IsNullOrEmpty(geometryAFixed.get()) == false && geometryAFixed->isValid() == false)
      {
        geometryAFixed.reset(te::gm::Validation::MakeValidWithSnap(geometryAFixed.get()).release());
      }

      wasChanged = currentWasChanged;

      //then we replace the resulted geometry into our base geometry vector
      delete vecResultGeometries[i];
      vecResultGeometries[i] = geometryAFixed.release();
    }
  }
  return vecResultGeometries;
}

te::gm::MultiLineString* te::gm::PrepareGeometriesToIntersection(const te::gm::MultiLineString& mline_A,
                                                                 const te::sam::rtree::Index<std::size_t, 8>& rtree,
                                                                 std::vector<te::gm::LineString>& vecRtreeSegments,
                                                                 bool& wasChanged, double tolerance)
{
  wasChanged = false;

#ifdef TERRALIB_GEOS_ENABLED

  te::gm::MultiLineString* multiLineStringResult = new te::gm::MultiLineString(0, mline_A.getGeomTypeId(), mline_A.getSRID());

  const std::vector<te::gm::Geometry*>& geometryVectorA = mline_A.getGeometries();

  for (std::size_t i = 0; i < geometryVectorA.size(); ++i)
  {
    bool currentWasChanged = false;

    te::gm::Geometry* singleResult = PrepareGeometriesToIntersection(*dynamic_cast<te::gm::LineString*>(geometryVectorA[i]), rtree, vecRtreeSegments, currentWasChanged, tolerance);
    multiLineStringResult->add(singleResult);

    if (currentWasChanged == true)
    {
      wasChanged = true;
    }
  }

  return multiLineStringResult;

#else
  throw Exception(TE_TR("PrepareGeometriesToIntersection routine is supported by GEOS! Please, enable the GEOS support."));
#endif
}

te::gm::LineString* te::gm::PrepareGeometriesToIntersection(const te::gm::LineString& ls_A,
                                                            const te::sam::rtree::Index<std::size_t, 8>& rtree,
                                                            std::vector<te::gm::LineString>& vecRtreeSegments, 
                                                            bool& wasChanged, double tolerance)
{
  wasChanged = false;

#ifdef TERRALIB_GEOS_ENABLED

  // First try to snap the points from ls_A in the points of ls_B
  bool wasSnapped = false;

  //if there are none candidate segments, we avoid some processing
  std::vector<std::size_t> report;
  rtree.search(*ls_A.getMBR(), report);

  if (report.empty())
  {
    return static_cast<te::gm::LineString*>(ls_A.clone());
  }

  std::size_t n_pointsA = ls_A.getNPoints();
  std::size_t n_segments_A = n_pointsA - 1;

  std::vector<te::gm::Point> vecOutputPoints;
  vecOutputPoints.reserve(n_pointsA);

  std::set<std::size_t> setIndexesIgnored;
  for (std::size_t i = 0; i < n_segments_A; ++i)
  {
    te::gm::Point segmentP1(ls_A.getX(i), ls_A.getY(i));
    te::gm::Point segmentP2(ls_A.getX(i+1), ls_A.getY(i+1));

    te::gm::Envelope segmentEnvelope = CreateEnvelope(segmentP1.getX(), segmentP1.getY(), segmentP2.getX(), segmentP2.getY());

    //now we create an ignore list based in all candidate segments that are EQUAL to the segments in the prepared line
    std::vector<std::size_t> segmentReport;
    rtree.search(segmentEnvelope, segmentReport);

    for (std::size_t i = 0; i < segmentReport.size(); ++i)
    {
      std::size_t index = segmentReport[i];
      //if it is already ignored, we go to the next one
      if (setIndexesIgnored.find(index) != setIndexesIgnored.end())
      {
        continue;
      }

      const te::gm::LineString& candidate = vecRtreeSegments[index];

      //if the boxes arent igual, the segment must not be ignored
      if(segmentEnvelope.equals(*candidate.getMBR()) == false)
      {
        continue;
      }

      //if the boxes are equal, we must check the coordinates  
      //if one coordinate is equal, necessarely the other one is tool
      //so we can check only  the first coordinate
      te::gm::Point candidateP1(candidate.getX(0), candidate.getY(0));
      te::gm::Point candidateP2(candidate.getX(1), candidate.getY(1));
      
      //if the segment fist point is not equal any point of the candidate line, the segment must not be ignored
      if (Distance::IsEqual(segmentP1, candidateP1) == false && Distance::IsEqual(segmentP1, candidateP2) == false)
      {
        continue;
      }

      //if we got here, both segments are equal and so they can be added to the ignore list
      setIndexesIgnored.insert(index);
    }
  }

  std::unique_ptr<te::gm::LineString> lr_Result(dynamic_cast<te::gm::LineString*>(ls_A.clone()));
  lr_Result.reset(SnapLineToPoints(rtree, vecRtreeSegments, setIndexesIgnored, *lr_Result.get(), wasSnapped, tolerance));

  lr_Result.reset(AddIntersectionPoints(rtree, vecRtreeSegments, setIndexesIgnored, *lr_Result.get(), true, tolerance));
  lr_Result.reset(AddIntersectionPoints(rtree, vecRtreeSegments, setIndexesIgnored, *lr_Result.get(), false, tolerance));

  if (wasSnapped == true || lr_Result->size() != ls_A.size())
  {
    wasChanged = true;
  }

  return lr_Result.release();

#else
  throw Exception(TE_TR("PrepareGeometriesToIntersection routine is supported by GEOS! Please, enable the GEOS support."));
#endif
}

te::gm::LineString* te::gm::SnapLineToPoints(const te::sam::rtree::Index<std::size_t, 8>& rtree,
                                             const std::vector<te::gm::LineString>& referenceSegments,
                                             const std::set<std::size_t>& setIndexesIgnored,
                                             const LineString& linearRingToSnap, bool& wasChanged,
                                             double tolerance)
{
  wasChanged = false;

#ifdef TERRALIB_GEOS_ENABLED

  te::gm::LineString* lr_Result = dynamic_cast<te::gm::LineString*>(linearRingToSnap.clone());

  std::size_t npoints_A = linearRingToSnap.getNPoints();

  for (std::size_t i = 0; i < npoints_A; ++i)
  {
    te::gm::Coord2D currentPoint(lr_Result->getX(i), lr_Result->getY(i));

    te::gm::Envelope expandedEnv;
    expandedEnv.init(
        currentPoint.x- tolerance, currentPoint.y - tolerance,
        currentPoint.x + tolerance, currentPoint.y + tolerance);

    std::vector<size_t> report;
    rtree.search(expandedEnv, report);

    if (report.empty())
      continue;

    // Extracts all the points.
    std::vector<te::gm::Coord2D> candidates;
    for (std::size_t ls = 0; ls < report.size(); ++ls)
    {
      std::size_t index = report[ls];

      //we the candidate is in the ignore list, we go to the next one
      if (setIndexesIgnored.find(index) != setIndexesIgnored.end())
      {
        continue;
      }

      te::gm::Coord2D pStart(referenceSegments[index].getX(0), referenceSegments[index].getY(0));
      //std::unique_ptr<Point> pStart = referenceSegments[index].getPointN(0);
      candidates.push_back(pStart);

      //std::unique_ptr<Point> pEnd = referenceSegments[index].getPointN(1);
      te::gm::Coord2D pEnd(referenceSegments[index].getX(1), referenceSegments[index].getY(1));
      candidates.push_back(pEnd);
    }

    if (candidates.empty() == true)
    {
      continue;
    }

    //Checks if there is on point that is within the tolerance.
    //If there is more than one, is found the closest.
    te::gm::CoordDistanceOrderFunctor func(currentPoint);
    std::sort(candidates.begin(), candidates.end(), func);

    if (te::gm::Distance::IsEqual(currentPoint, candidates.at(0)) == false)
    {
      if (te::gm::Distance::IsWithinDistance(currentPoint, candidates.at(0), tolerance) == true)
      {
        wasChanged = true;

        lr_Result->setPoint(i, candidates[0].x, candidates[0].y);

        if (i == 0)
        {
          lr_Result->setPoint(npoints_A - 1, candidates[0].x, candidates[0].y);
        }
      }
    }
  }

  return lr_Result;

#else
  throw Exception(TE_TR("SnapLineToPoints routine is supported by GEOS! Please, enable the GEOS support."));
#endif
}

te::gm::LineString* te::gm::AddIntersectionPoints(const te::sam::rtree::Index<std::size_t, 8>& rtree,
                                                  const std::vector<te::gm::LineString>& candidateSegments,
                                                  const std::set<std::size_t>& setIndexesIgnored,
                                                  const LineString& lr_Reference,
                                                  const bool& usePerpendicularDistance,
                                                  double tolerance)
{
#ifdef TERRALIB_GEOS_ENABLED

  bool addedIntersectionPoints = false;

  std::size_t n_pointsA = lr_Reference.getNPoints();
  std::size_t n_segments_A = n_pointsA - 1;

  std::vector<te::gm::Coord2D> vecOutputPoints;
  vecOutputPoints.reserve(n_pointsA);

  for (std::size_t i = 0; i < n_segments_A; ++i)
  {
    // Search for candidates.
    te::gm::Envelope currentSegmentEnvelope = CreateEnvelope(lr_Reference.getX(i), lr_Reference.getY(i), lr_Reference.getX(i + 1), lr_Reference.getY(i + 1));

    te::gm::Coord2D cCurrentStartCoord(lr_Reference.getX(i), lr_Reference.getY(i));

    vecOutputPoints.push_back(cCurrentStartCoord);

    te::gm::Envelope expandedEnv(currentSegmentEnvelope.getLowerLeftX() - tolerance,
      currentSegmentEnvelope.getLowerLeftY() - tolerance,
      currentSegmentEnvelope.getUpperRightX() + tolerance,
      currentSegmentEnvelope.getUpperRightY() + tolerance);

    std::vector<size_t> report;
    rtree.search(expandedEnv, report);

    if (report.empty())
      continue;

    //creates the curent segment
    te::gm::LineString segment_A(2, te::gm::LineStringType, lr_Reference.getSRID());
    segment_A.setPoint(0, lr_Reference.getX(i), lr_Reference.getY(i));
    segment_A.setPoint(1, lr_Reference.getX(i + 1), lr_Reference.getY(i + 1));

    std::vector<te::gm::LineString> candidates;
    candidates.reserve(report.size());
    for(std::size_t ls = 0; ls < report.size(); ++ls)
    {
      std::size_t index = report[ls];

      //we the candidate is in the ignore list, we go to the next one
      if (setIndexesIgnored.find(index) != setIndexesIgnored.end())
      {
        continue;
      }
      
      const te::gm::LineString& segmentCandidate = candidateSegments[index];
      candidates.push_back(segmentCandidate);
    }

    if (candidates.empty() == true)
    {
      continue;
    }

    // Executes the intersection for each candidate point, and get the coordinates.
    std::vector<te::gm::Coord2D> vecIntersectionPoints;

    if(usePerpendicularDistance)
      vecIntersectionPoints = GetIntersectionPointsByPerpendicularDistance(segment_A, candidates);
    else
      vecIntersectionPoints = GetIntersectionPointsByOperators(segment_A, candidates);

    CoordDistanceOrderFunctor func(cCurrentStartCoord);
    std::sort(vecIntersectionPoints.begin(), vecIntersectionPoints.end(), func);

    for (std::size_t j = 0; j < vecIntersectionPoints.size(); ++j)
    {
      const te::gm::Coord2D& currentIntersectionPoint = vecIntersectionPoints[j];
      const te::gm::Coord2D& lastPoint = vecOutputPoints[vecOutputPoints.size() - 1];

      if (te::gm::Distance::IsWithinDistance(currentIntersectionPoint, lastPoint, tolerance) == true)
      {
        continue;
      }
      if (te::gm::Distance::IsWithinDistance(currentIntersectionPoint, cCurrentStartCoord, tolerance) == true)
      {
        continue;
      }
      te::gm::Coord2D cCurrentEndCoord(lr_Reference.getX(i + 1), lr_Reference.getY(i + 1));
      if (te::gm::Distance::IsWithinDistance(currentIntersectionPoint, cCurrentEndCoord, tolerance) == true)
      {
        continue;
      }

      addedIntersectionPoints = true;
      vecOutputPoints.push_back(vecIntersectionPoints[j]);
    }
  }

  //we onlye create anew line if we detected at least one distinct intersection point
  if (addedIntersectionPoints == false)
  {
    return dynamic_cast<te::gm::LineString*>(lr_Reference.clone());
  }

  // Set the end point in the output vector.
  te::gm::Coord2D cStart(lr_Reference.getX(0), lr_Reference.getY(0));
  vecOutputPoints.push_back(cStart);

  te::gm::LinearRing* result = new te::gm::LinearRing(vecOutputPoints.size(), te::gm::LineStringType, lr_Reference.getSRID());

  for (std::size_t i = 0; i < vecOutputPoints.size(); ++i)
  {
    result->setPoint(i, vecOutputPoints[i].x, vecOutputPoints[i].y);
  }

  return result;

#else
  throw Exception(TE_TR("GetIntersectionPoints routine is supported by GEOS! Please, enable the GEOS support."));
#endif
}

std::vector<te::gm::Coord2D> te::gm::GetIntersectionPointsByOperators(
    const te::gm::LineString& lsReference,
    const std::vector<te::gm::LineString>& candidates,
    double tolerance)
{
#ifdef TERRALIB_GEOS_ENABLED

  std::vector<te::gm::Coord2D> intersectionPoints;
  std::unique_ptr<te::gm::Coord2D> startPoint;
  std::unique_ptr<te::gm::Coord2D> endPoint;

  geos::geom::Coordinate referenceP1(lsReference.getX(0), lsReference.getY(0));
  geos::geom::Coordinate referenceP2(lsReference.getX(1), lsReference.getY(1));

  for (std::size_t i = 0; i < candidates.size(); ++i)
  {
    const te::gm::LineString& lineCandidate = candidates[i];

    geos::geom::Coordinate candidateP1(lineCandidate.getX(0), lineCandidate.getY(0));
    geos::geom::Coordinate candidateP2(lineCandidate.getX(1), lineCandidate.getY(1));

    geos::algorithm::LineIntersector intersector;
    intersector.computeIntersection(referenceP1, referenceP2, candidateP1, candidateP2);
    if (intersector.hasIntersection() == false)
    {
      continue;
    }

    std::vector<te::gm::Coord2D> candidatesIntersectionPoints;
    std::size_t numberOfIntersections = static_cast<std::size_t>(intersector.getIntersectionNum());
    for (std::size_t j = 0; j < numberOfIntersections; ++j)
    {
      const geos::geom::Coordinate& interCoord = intersector.getIntersection(j);

      candidatesIntersectionPoints.push_back(te::gm::Coord2D(interCoord.x, interCoord.y));
    }

    for (std::size_t j = 0; j < candidatesIntersectionPoints.size(); ++j)
    {
      if (startPoint == nullptr || endPoint == nullptr)
      {
        std::size_t numCoords = lsReference.size();
        startPoint.reset(new te::gm::Coord2D(lsReference.getX(0), lsReference.getY(0)));
        endPoint.reset(new te::gm::Coord2D(lsReference.getX(numCoords-1), lsReference.getY(numCoords-1)));
      }      

      const te::gm::Coord2D& candidateIntersectionPoint = candidatesIntersectionPoints[j];
      
      //if the candidate intersection point is "near" to the start or end point of the current segment, we drop it
      if (te::gm::Distance::IsWithinDistance(candidateIntersectionPoint, *startPoint.get(), tolerance) == true || te::gm::Distance::IsWithinDistance(candidateIntersectionPoint, *endPoint.get(), tolerance) == true)
      {
        continue;
      }

      std::size_t lineCandidateNumCoords = lineCandidate.size();
      te::gm::Coord2D candidateStartPoint(lineCandidate.getX(0), lineCandidate.getY(0));
      te::gm::Coord2D candidateEndPoint(lineCandidate.getX(lineCandidateNumCoords - 1), lineCandidate.getY(lineCandidateNumCoords - 1));

      // If the calculated intersection point is "near" the start or end point of the current segment, just snap it.
      if(te::gm::Distance::IsWithinDistance(candidateIntersectionPoint, candidateStartPoint, tolerance) == true)
      {
        intersectionPoints.push_back(candidateStartPoint);
      }
      else if (te::gm::Distance::IsWithinDistance(candidateIntersectionPoint, candidateEndPoint, tolerance) == true)
      {
        intersectionPoints.push_back(candidateEndPoint);
      }
      else
      {
        intersectionPoints.push_back(candidateIntersectionPoint);
      }
    }
  }

  return intersectionPoints;

#else
  throw Exception(TE_TR("GetIntersectionPointsByOperators routine is supported by GEOS! Please, enable the GEOS support."));
#endif
}

std::vector<te::gm::Coord2D> te::gm::GetIntersectionPointsByPerpendicularDistance(
    const te::gm::LineString& lsReference,
    const std::vector<te::gm::LineString>& candidates,
    double tolerance)
{
#ifdef TERRALIB_GEOS_ENABLED

  std::vector<te::gm::Coord2D> intersectionPoints;

  std::size_t lineReferenceNumCoords = lsReference.size();
  te::gm::Coord2D startPoint(lsReference.getX(0), lsReference.getY(0));
  te::gm::Coord2D endPoint(lsReference.getX(lineReferenceNumCoords - 1), lsReference.getY(lineReferenceNumCoords - 1));

  for (std::size_t i = 0; i < candidates.size(); ++i)
  {
    std::size_t lineCandidateNumCoords = candidates[i].size();
    te::gm::Coord2D candidateStartPoint(candidates[i].getX(0), candidates[i].getY(0));
    te::gm::Coord2D candidateEndPoint(candidates[i].getX(lineCandidateNumCoords - 1), candidates[i].getY(lineCandidateNumCoords - 1));

    std::unique_ptr<geos::geom::LineSegment> lineSegment(
        new geos::geom::LineSegment(startPoint.x, startPoint.y,
                                    endPoint.x, endPoint.y));

    //we try to find the an intersection point touching our current segment
    //we first compare our segment to the candidate start point
    te::gm::Coord2D pInterStart;
    bool isOnSegment = false;
    double startDistance = te::gm::Distance::PerpendicularDistance(startPoint, endPoint, candidateStartPoint, pInterStart, isOnSegment);

    if (te::gm::Distance::IsWithinDistance(startPoint, pInterStart, tolerance) == false && te::gm::Distance::IsWithinDistance(endPoint, pInterStart, tolerance) == false)
    {
      if (startDistance <= tolerance && isOnSegment == true)
      {
        intersectionPoints.push_back(candidateStartPoint);
      }
    }

    //then we compare our segment to the candidate end point
    te::gm::Coord2D pInterEnd;
    double endDistance = te::gm::Distance::PerpendicularDistance(startPoint, endPoint, candidateEndPoint, pInterEnd, isOnSegment);

    if (te::gm::Distance::IsWithinDistance(startPoint, pInterEnd, tolerance) == false && te::gm::Distance::IsWithinDistance(endPoint, pInterEnd, tolerance) == false)
    {
      if (endDistance <= tolerance && isOnSegment == true)
      {
        intersectionPoints.push_back(candidateEndPoint);
      }
    }
  }

  return intersectionPoints;

#else
  throw Exception(TE_TR("GetIntersectionPointsByPerpendicularDistance routine is supported by GEOS! Please, enable the GEOS support."));
#endif
}

std::vector<te::gm::MultiLineString*> te::gm::GetAsMultiLineStringVector(const te::gm::Geometry& geometry)
{
  std::vector<te::gm::MultiLineString*> multiLineStringVector;

  switch(geometry.getGeomTypeId())
  {
    case te::gm::LineStringType:
    case te::gm::LineStringMType:
    case te::gm::LineStringZType:
    case te::gm::LineStringZMType:
    {
      te::gm::MultiLineString* multiLineString = new te::gm::MultiLineString(0, GetMultiLineStringType(geometry), geometry.getSRID());

      te::gm::LineString* lineString = dynamic_cast<te::gm::LineString*>(geometry.clone());

      multiLineString->add(lineString);

      multiLineStringVector.push_back(multiLineString);
    }
    break;

    case te::gm::MultiLineStringType:
    case te::gm::MultiLineStringMType:
    case te::gm::MultiLineStringZType:
    case te::gm::MultiLineStringZMType:
    {
      te::gm::MultiLineString* multiLineString = new te::gm::MultiLineString(0, GetMultiLineStringType(geometry), geometry.getSRID());

      std::vector<te::gm::Geometry*> geometryVector;
      te::gm::Multi2Single(&geometry, geometryVector);

      for(auto g : geometryVector)
      {
        te::gm::LineString* lineString = dynamic_cast<te::gm::LineString*>(g->clone());

        multiLineString->add(lineString);

        multiLineStringVector.push_back(multiLineString);
      }

      te::common::FreeContents(geometryVector);
    }
    break;

    case te::gm::PolygonType:
    case te::gm::PolygonMType:
    case te::gm::PolygonZType:
    case te::gm::PolygonZMType:
    {
      te::gm::MultiLineString* multiLineString = new te::gm::MultiLineString(0, GetMultiLineStringType(geometry), geometry.getSRID());

      const te::gm::Polygon* polygon = dynamic_cast<const te::gm::Polygon*>(&geometry);

      const std::vector<te::gm::Curve*>& curveVector = polygon->getRings();

      for (auto curve : curveVector)
      {
        te::gm::LineString* lineString = dynamic_cast<te::gm::LineString*>(curve->clone());

        multiLineString->add(lineString);
      }

      multiLineStringVector.push_back(multiLineString);
    }
    break;

    case te::gm::MultiPolygonType:
    case te::gm::MultiPolygonMType:
    case te::gm::MultiPolygonZType:
    case te::gm::MultiPolygonZMType:
    {
      std::vector<te::gm::Geometry*> geometryVector;
      Multi2Single(&geometry, geometryVector);

      for(auto g : geometryVector)
      {
        std::vector<te::gm::MultiLineString*> multiLineStringVectorTemp = GetAsMultiLineStringVector(*g);

        multiLineStringVector.insert(multiLineStringVector.end(), multiLineStringVectorTemp.begin(), multiLineStringVectorTemp.end());
      }

      te::common::FreeContents(geometryVector);
    }
    break;

    default:
      throw Exception("This type of geometry is not composed by LineString.");
  }

  return multiLineStringVector;
}

te::gm::Geometry* te::gm::MultiLineToDefinedType(const std::vector<te::gm::MultiLineString*>& multiLineStringVector, te::gm::GeomType geometryType)
{
  te::gm::Geometry* geometry = nullptr;

  if(multiLineStringVector.empty())
    return geometry;

  int srid = multiLineStringVector[0]->getSRID();

  switch (geometryType)
  {
    case te::gm::LineStringType:
    case te::gm::LineStringZType:
    case te::gm::LineStringMType:
    case te::gm::LineStringZMType:
    {
      te::gm::MultiLineString* mls = multiLineStringVector[0];

      if(mls->isEmpty())
        return geometry;

      geometry = dynamic_cast<te::gm::Geometry*>(mls->getGeometryN(0)->clone());
    }
    break;

    case te::gm::PolygonType:
    case te::gm::PolygonZType:
    case te::gm::PolygonMType:
    case te::gm::PolygonZMType:
    {
      te::gm::MultiLineString* mls = multiLineStringVector[0];

      if(mls->isEmpty())
        return geometry;

      te::gm::Polygon* polygon = new te::gm::Polygon(0, geometryType, srid);

      for(auto g : mls->getGeometries())
        polygon->add(dynamic_cast<te::gm::Curve*>(g->clone()));

      geometry = polygon;
    }
    break;

    case te::gm::MultiLineStringType:
    case te::gm::MultiLineStringZType:
    case te::gm::MultiLineStringMType:
    case te::gm::MultiLineStringZMType:
    {
      te::gm::MultiLineString* outMls = new te::gm::MultiLineString(0, geometryType, srid);

      for(auto inMls : multiLineStringVector)
      {
        if(inMls->isEmpty())
          continue;

        for(auto g : inMls->getGeometries())
          outMls->add(dynamic_cast<te::gm::Geometry*>(g->clone()));
      }

      geometry = outMls;
    }
    break;

    case te::gm::MultiPolygonType:
    case te::gm::MultiPolygonZType:
    case te::gm::MultiPolygonMType:
    case te::gm::MultiPolygonZMType:
    {
      te::gm::MultiPolygon* outMPolygon = new te::gm::MultiPolygon(0, geometryType, srid);

      for(auto inMls : multiLineStringVector)
      {
        if(inMls->getGeometries().empty())
          continue;

        std::size_t numRings = inMls->getGeometries().size();

        te::gm::Polygon* polygon = new te::gm::Polygon(numRings, GetSimpleType(geometryType), srid);

        for (std::size_t j = 0; j < numRings; ++j)
        {
          polygon->setRingN(j, dynamic_cast<te::gm::Curve*>(inMls->getGeometryN(j)->clone()));
        }

        outMPolygon->add(polygon);
      }

      geometry = outMPolygon;
    }
    break;

    default:
      throw Exception("This type of geometry is not composed by LineString.");
  }

  return geometry;
}

te::gm::GeomType te::gm::GetMultiLineStringType(const te::gm::Geometry& geometry)
{
  switch (geometry.getGeomTypeId())
  {
    case te::gm::LineStringZType:
    case te::gm::PolygonZType:
    case te::gm::MultiLineStringZType:
    case te::gm::MultiPolygonZType:
      return te::gm::MultiLineStringZType;

    case te::gm::LineStringMType:
    case te::gm::PolygonMType:
    case te::gm::MultiLineStringMType:
    case te::gm::MultiPolygonMType:
      return te::gm::MultiLineStringMType;

    case te::gm::LineStringZMType:
    case te::gm::PolygonZMType:
    case te::gm::MultiLineStringZMType:
    case te::gm::MultiPolygonZMType:
      return te::gm::MultiLineStringZMType;

    default:
      return te::gm::MultiLineStringType;
  }
}

bool te::gm::IsMultiType(te::gm::GeomType geomType)
{
  switch(geomType)
  {
    case te::gm::MultiLineStringType:
    case te::gm::MultiLineStringMType:
    case te::gm::MultiLineStringZType:
    case te::gm::MultiLineStringZMType:
    case te::gm::MultiPointType:
    case te::gm::MultiPointMType:
    case te::gm::MultiPointZType:
    case te::gm::MultiPointZMType:
    case te::gm::MultiPolygonType:
    case te::gm::MultiPolygonMType:
    case te::gm::MultiPolygonZType:
    case te::gm::MultiPolygonZMType:
    case te::gm::GeometryCollectionType:
    case te::gm::GeometryCollectionMType:
    case te::gm::GeometryCollectionZType:
    case te::gm::GeometryCollectionZMType:
      return true;
    default:
      return false;
  }
}

te::gm::GeomType te::gm::GetSimpleType(te::gm::GeomType geomType)
{
  if (IsMultiType(geomType) == false)
  {
    return geomType;
  }

  switch(geomType)
  {
    case te::gm::MultiLineStringType:
      return te::gm::LineStringType;
    case te::gm::MultiLineStringMType:
      return te::gm::LineStringMType;
    case te::gm::MultiLineStringZType:
      return te::gm::LineStringZType;
    case te::gm::MultiLineStringZMType:
      return te::gm::LineStringZMType;
    case te::gm::MultiPointType:
      return te::gm::PointType;
    case te::gm::MultiPointMType:
      return te::gm::PointMType;
    case te::gm::MultiPointZType:
      return te::gm::PointZType;
    case te::gm::MultiPointZMType:
      return te::gm::PointZMType;
    case te::gm::MultiPolygonType:
      return te::gm::PolygonType;
    case te::gm::MultiPolygonMType:
      return te::gm::PolygonMType;
    case te::gm::MultiPolygonZType:
      return te::gm::PolygonZType;
    case te::gm::MultiPolygonZMType:
      return te::gm::PolygonZMType;
    case te::gm::GeometryCollectionType:
      return te::gm::GeometryType;
    case te::gm::GeometryCollectionMType:
      return te::gm::GeometryMType;
    case te::gm::GeometryCollectionZType:
      return te::gm::GeometryZType;
    case te::gm::GeometryCollectionZMType:
      return te::gm::GeometryZMType;
    default:
      return te::gm::UnknownGeometryType;
  }
}

te::gm::GeomType te::gm::GetMultiType(te::gm::GeomType geomType)
{
  if (IsMultiType(geomType) == true)
  {
    return geomType;
  }

  switch (geomType)
  {
  case te::gm::LineStringType:
    return te::gm::MultiLineStringType;
  case te::gm::LineStringMType:
    return te::gm::MultiLineStringMType;
  case te::gm::LineStringZType:
    return te::gm::MultiLineStringZType;
  case te::gm::LineStringZMType:
    return te::gm::MultiLineStringZMType;
  case te::gm::PointType:
    return te::gm::MultiPointType;
  case te::gm::PointMType:
    return te::gm::MultiPointMType;
  case te::gm::PointZType:
    return te::gm::MultiPointZType;
  case te::gm::PointZMType:
    return te::gm::MultiPointZMType;
  case te::gm::PolygonType:
    return te::gm::MultiPolygonType;
  case te::gm::PolygonMType:
    return te::gm::MultiPolygonMType;
  case te::gm::PolygonZType:
    return te::gm::MultiPolygonZType;
  case te::gm::PolygonZMType:
    return te::gm::MultiPolygonZMType;
  case te::gm::GeometryType:
    return te::gm::GeometryCollectionType;
  case te::gm::GeometryMType:
    return te::gm::GeometryCollectionMType;
  case te::gm::GeometryZType:
    return te::gm::GeometryCollectionZType;
  case te::gm::GeometryZMType:
    return te::gm::GeometryCollectionZMType;
  default:
    return te::gm::UnknownGeometryType;
  }
}

te::gm::LineString te::gm::CreateLine(const te::gm::Coord2D& startCoord, const te::gm::Coord2D& endCoord, const int& srid, const int& numberOfIntermediateCoords)
{
  double gapX = (endCoord.getX() - startCoord.getX()) / static_cast<double>(numberOfIntermediateCoords + 1);
  double gapY = (endCoord.getY() - startCoord.getY()) / static_cast<double>(numberOfIntermediateCoords + 1);

  std::vector<te::gm::Coord2D> coords;

  coords.push_back(startCoord);
  for (std::size_t i = 0; i < static_cast<std::size_t>(numberOfIntermediateCoords); ++i)
  {
    std::size_t coordIndex = i + 1;
    te::gm::Coord2D coord(startCoord.getX() + (coordIndex * gapX), startCoord.getY() + (coordIndex * gapY));
    coords.push_back(coord);
  }
  coords.push_back(endCoord);

  te::gm::LineString resultLine(coords.size(), te::gm::LineStringType, srid);

  for (std::size_t t = 0; t < coords.size(); ++t)
    resultLine.setPoint(t, coords[t].getX(), coords[t].getY());

  return resultLine;
}

te::gm::Polygon te::gm::CreatePolygon(const te::gm::Envelope& box, const int& srid, const int& numberOfIntermediateCoords)
{
  te::gm::Coord2D lowerLeft(box.m_llx, box.m_lly);
  te::gm::Coord2D lowerRight(box.m_urx, box.m_lly);

  te::gm::Coord2D upperLeft(box.m_llx, box.m_ury);
  te::gm::Coord2D upperRight(box.m_urx, box.m_ury);

  te::gm::LineString bottomLine = CreateLine(lowerLeft, lowerRight, srid, numberOfIntermediateCoords);
  te::gm::LineString rightLine = CreateLine(lowerRight, upperRight, srid, numberOfIntermediateCoords);
  te::gm::LineString topLine = CreateLine(upperRight, upperLeft, srid, numberOfIntermediateCoords);
  te::gm::LineString leftLine = CreateLine(upperLeft, lowerLeft, srid, numberOfIntermediateCoords);

  std::vector<te::gm::Coord2D> coords;

  //we do not add the last coord to avoid duplicity
  for (std::size_t i = 0; i < bottomLine.size() - 1; ++i)
    coords.push_back(te::gm::Coord2D(bottomLine.getX(i), bottomLine.getY(i)));

  for (std::size_t i = 0; i < rightLine.size() - 1; ++i)
    coords.push_back(te::gm::Coord2D(rightLine.getX(i), rightLine.getY(i)));

  for (std::size_t i = 0; i < topLine.size() - 1; ++i)
    coords.push_back(te::gm::Coord2D(topLine.getX(i), topLine.getY(i)));

  for (std::size_t i = 0; i < leftLine.size() - 1; ++i)
    coords.push_back(te::gm::Coord2D(leftLine.getX(i), leftLine.getY(i)));

  //now we must add the first point again to ensure that we have a closed ring
  coords.push_back(*coords.begin());

  te::gm::LinearRing* fullLine = new te::gm::LinearRing(coords.size(), te::gm::LineStringType);

  for (std::size_t t = 0; t < coords.size(); ++t)
    fullLine->setPoint(t, coords[t].getX(), coords[t].getY());

  te::gm::Polygon polygon(1, te::gm::PolygonType, srid);
  polygon.setRingN(0, fullLine);

  return polygon;
}

te::gm::LineString* te::gm::ParallelLine(const te::gm::LineString* line, double distance)
{
  te::gm::LineString* newLine = nullptr;

  //handles te::gm::LinearRing
  bool isOriginalLineRing = false;
  if (dynamic_cast<const te::gm::LinearRing*>(line) != nullptr)
  {
    isOriginalLineRing = true;
  }

  te::gm::Point p1(*line->getStartPoint());
  te::gm::Point p2(*line->getEndPoint());

  bool isClosed = false;
  if (te::gm::Distance::IsEqual(p1, p2) == true)
  {
    isClosed = true;
  }

#ifdef TERRALIB_GEOS_ENABLED
  try
  {
    std::unique_ptr<geos::geom::Geometry> g(GEOSWriter::write(line));

    std::unique_ptr<geos::geom::CoordinateSequence> inputCoordinates(g->getCoordinates());

    geos::operation::buffer::BufferParameters bufferParams;
    
    std::vector<geos::geom::CoordinateSequence*> outputLineList;
    
    // parallel line
    std::unique_ptr<geos::operation::buffer::OffsetSegmentGenerator> generator;
    
    generator.reset(new geos::operation::buffer::OffsetSegmentGenerator(g->getPrecisionModel(), bufferParams, distance));

    std::size_t n1 = inputCoordinates->getSize() - 1;

    #if GEOS_FULLVERSION >= 30900
      generator->initSideSegments(inputCoordinates->getAt(0), 
        inputCoordinates->getAt(1), geos::geom::Position::LEFT);
    #else
      generator->initSideSegments(inputCoordinates->getAt(0), inputCoordinates->getAt(1), geos::geomgraph::Position::LEFT);
    #endif

    generator->addFirstSegment();

    for (std::size_t i = 2 ; i <= n1 ; ++i) 
    {
      generator->addNextSegment(inputCoordinates->getAt(i), true);
    }

    generator->addLastSegment();

    generator->getCoordinates(outputLineList);

    // result in strange lines
    if (isOriginalLineRing == false)
    {
      if (isClosed == false)
      {
        geos::geom::CoordinateSequence* currentCS = outputLineList.at(0);
        geos::geom::CoordinateArraySequence* adjustedCS = new geos::geom::CoordinateArraySequence();
        for(int c = 0; c < currentCS->size() - 1; ++c)
        {
          const geos::geom::Coordinate& coordinate = currentCS->getAt(c);
          adjustedCS->add(coordinate);
        }
        delete outputLineList[0];
        outputLineList[0] = adjustedCS;
      }
      
      std::unique_ptr<geos::geom::LineString> geosLine(GEOSGeometryFactory::getGeomFactory()->createLineString(outputLineList.at(0)));
      geosLine->setSRID(line->getSRID());
      newLine = GEOSReader::read(geosLine.get());
    }
    else
    {
      std::unique_ptr<geos::geom::LinearRing> geosRing(GEOSGeometryFactory::getGeomFactory()->createLinearRing(outputLineList.at(0)));
      geosRing->setSRID(line->getSRID());
      te::gm::LinearRing* linearRing = GEOSReader::read(geosRing.get());
      if (linearRing)
      {
        newLine = dynamic_cast<te::gm::LineString*>(linearRing);
      }
    }    

    if (outputLineList.size() > 1)
    {
      outputLineList.erase(outputLineList.begin());

      te::common::FreeContents(outputLineList);
    }
  }
  catch(geos::util::GEOSException& e)
  {
    throw Exception(e.what());
  }

#else
  throw Exception(TE_TR("getLength routine is supported by GEOS! Please, enable the GEOS support."));
#endif  

  return newLine;
}

bool te::gm::IsPointOnPolygon(const te::gm::Coord2D& coord, const te::gm::Geometry* geometry)
{
#ifdef TERRALIB_GEOS_ENABLED
  if (geometry->isCollection() == true)
  {
    const te::gm::GeometryCollection* collection = static_cast<const te::gm::GeometryCollection*>(geometry);
    for (std::size_t i = 0; i < collection->getNumGeometries(); ++i)
    {
      if (IsPointOnPolygon(coord, collection->getGeometryN(i)) == true)
      {
        return true;
      }
    }

    return false;
  }
  else if (geometry->get2DGeomTypeId() == te::gm::PolygonType)
  {
    geos::geom::Coordinate coordinate(coord.x, coord.y);
    geos::algorithm::RayCrossingCounter pointInPolygonTester(coordinate);

    const te::gm::Polygon* polygon = static_cast<const te::gm::Polygon*>(geometry);
    for (std::size_t i = 0; i < polygon->getNumRings(); ++i)
    {
      const te::gm::LineString* ring = static_cast<const te::gm::LineString*>(polygon->getRingN(i));
    
      for (std::size_t j = 1; j < ring->getNPoints(); ++j)
      {
        geos::geom::Coordinate coordinate1(ring->getX(j - 1), ring->getY(j - 1));
        geos::geom::Coordinate coordinate2(ring->getX(j), ring->getY(j));

        pointInPolygonTester.countSegment(coordinate1, coordinate2);
      }
    }

    return pointInPolygonTester.isPointInPolygon();
  }

  throw Exception(TE_TR("isPointOnPolygon must be used only for Polygon and MultiPolygon geometry types."));

#else
throw Exception(TE_TR("GetIntersectionPointsByPerpendicularDistance routine is supported by GEOS! Please, enable the GEOS support."));
#endif
}

te::gm::Coord2D te::gm::GetInteriorPoint(const te::gm::Geometry* geometry)
{
#ifdef TERRALIB_GEOS_ENABLED
  te::gm::Coord2D interiorPoint(TE_DOUBLE_NOT_A_NUMBER, TE_DOUBLE_NOT_A_NUMBER);

  std::unique_ptr<geos::geom::Geometry> geometryGEOS(GEOSWriter::write(geometry));

  if (geometry->getDimension() == te::gm::A)
  {
    geos::algorithm::InteriorPointArea algorithm(geometryGEOS.get());

    geos::geom::Coordinate geosInteriorPoint;
    if (algorithm.getInteriorPoint(geosInteriorPoint) == true)
    {
      interiorPoint.x = geosInteriorPoint.x;
      interiorPoint.y = geosInteriorPoint.y;
      return interiorPoint;
    }
  }
  
  try
  {
    std::unique_ptr<geos::geom::Point> pointGeos(geometryGEOS->getInteriorPoint());
    if (pointGeos != nullptr)
    {
      interiorPoint.x = pointGeos->getX();
      interiorPoint.y = pointGeos->getY();
    }
  }
  catch (geos::util::GEOSException & e)
  {
    throw Exception(e.what());
  }

  return interiorPoint;

#else
  throw Exception(TE_TR("GetInteriorPoint routine is supported by GEOS! Please, enable the GEOS support."));
#endif
}

te::gm::Geometry* te::gm::NormalizeGeometryByType(const te::gm::Geometry* geometry, te::gm::GeomType singleGeometryType)
{
  if (geometry == nullptr)
  {
    return nullptr;
  }

  if (singleGeometryType == te::gm::GeometryType)
  {
    return static_cast<te::gm::Geometry*>(geometry->clone());
  }

  te::gm::GeomType multiGeometryType = te::gm::GetMultiType(singleGeometryType);

  //test if the geometry is already from the given simple type or its equivalent collection type
  if (geometry->get2DGeomTypeId() == multiGeometryType)
  {
    return static_cast<te::gm::Geometry*>(geometry->clone());
  }
  else if (geometry->get2DGeomTypeId() == singleGeometryType)
  {
    return static_cast<te::gm::Geometry*>(geometry->clone());
  }

  //if the geometry is not from the expected type, we check if it is a collectiion and look inside
  if (geometry->isCollection() == true)
  {
    const te::gm::GeometryCollection* collection = dynamic_cast<const te::gm::GeometryCollection*>(geometry);

    //if we are a geometry collection, we must create an output collection based on the given singleGeometryType
    te::gm::Geometry* newGeometry = te::gm::GeometryFactory::make(multiGeometryType, geometry->getSRID());
    std::unique_ptr<te::gm::GeometryCollection> newCollection(static_cast<te::gm::GeometryCollection*>(newGeometry));

    //then we recursevely analyse all the geometries inside the collection
    for (std::size_t i = 0; i < collection->getNumGeometries(); ++i)
    {
      std::unique_ptr<te::gm::Geometry> normalizedGeometry(te::gm::NormalizeGeometryByType(collection->getGeometryN(i), singleGeometryType));
      if (normalizedGeometry != nullptr)
      {
        te::gm::GeometryVector vecSingleGeometries;
        te::gm::Multi2Single(normalizedGeometry.get(), vecSingleGeometries);

        for (te::gm::Geometry* simpleGeometry : vecSingleGeometries)
        {
          newCollection->add(simpleGeometry);
        }
      }
    }

    //if at least one geometry inside of the collection was from the given type, we return it
    if (newCollection->getNumGeometries() > 0)
    {
      //we must ensure that the result polygon or multipolygon geometry is valid
      te::gm::GeometryPtr fixedCollection(newCollection.release());
      if (fixedCollection->isValid() == false)
      {
        fixedCollection.reset(te::gm::Validation::MakeValidWithSnap(fixedCollection.get()).release());
      }

      te::gm::GeomType resultSimpleType = te::gm::GetSimpleType(fixedCollection->get2DGeomTypeId());
      if (resultSimpleType != singleGeometryType)
      {
        throw Exception("The result of the 'NormalizeGeometryByType' is not the expeteced type");
      }

      return fixedCollection.release();
    }
  }

  //if we get here it means that the given geometry was not or had not geometries related to the given singleGeometryType
  return nullptr;
}

double te::gm::GetAppropriatePrecision(int srid)
{
  //we cache the precision based on the cached srid
  //this is done because getUnit is an expensive function as it must parse strings
  static double precision = 0.000000001;
  static int cachedSRID = TE_UNKNOWN_SRS;

  if (srid == cachedSRID)
  {
    return precision;
  }

  cachedSRID = srid;

  te::common::UnitOfMeasurePtr uom = te::srs::SpatialReferenceSystemManager::getInstance().getUnit(srid);
  if (uom != nullptr)
  {
    switch (uom->getId())
    {
    case te::common::UOM_Metre:
      precision = 0.001;
      break;
    case te::common::UOM_Degree:
      precision = 0.000000001;
      break;
    default:
      break;
    }
  }

  return precision;
}

te::gm::Geometry* te::gm::ClipGeometry(const te::gm::Geometry* geometry, const te::gm::Envelope& clipArea)
{
#ifdef TERRALIB_GEOS_ENABLED
  try
  {
    std::unique_ptr<geos::geom::Geometry> geometryGEOS(GEOSWriter::write(geometry));
    geos::operation::intersection::Rectangle rectangle(clipArea.m_llx, clipArea.m_lly, clipArea.m_urx, clipArea.m_ury);

    std::unique_ptr<geos::geom::Geometry> resultGeometry(geos::operation::intersection::RectangleIntersection::clip(*geometryGEOS, rectangle));
    if (resultGeometry == nullptr)
    {
      return nullptr;
    }

    resultGeometry->setSRID(geometry->getSRID());

    return GEOSReader::read(resultGeometry.get());
  }
  catch (geos::util::GEOSException& e)
  {
    std::string what = e.what();
    throw Exception("Exception thrown in 'ClipGeometry': " + what);
  }

#else
  throw Exception(TE_TR("GetInteriorPoint routine is supported by GEOS! Please, enable the GEOS support."));
#endif
}

bool te::gm::IsNullOrEmpty(const te::gm::GeometryPtr& geometry)
{
  if (geometry.get() == nullptr)
  {
    return true;
  }

  if (geometry->isEmpty() == true)
  {
    return true;
  }

  return false;
}

bool te::gm::IsNullOrEmpty(const te::gm::Geometry* geometry)
{
  if (geometry == nullptr)
  {
    return true;
  }

  if (geometry->isCollection() == true)
  {
    const te::gm::GeometryCollection* collection = static_cast<const te::gm::GeometryCollection*>(geometry);
    for (std::size_t i = 0; i < collection->getNumGeometries(); ++i)
    {
      if (IsNullOrEmpty(collection->getGeometryN(i)) == false)
      {
        return false;
      }
    }
  }
  else if (geometry->get2DGeomTypeId() == te::gm::PolygonType)
  {
    const te::gm::Polygon* polygon = static_cast<const te::gm::Polygon*>(geometry);
    for (std::size_t i = 0; i < polygon->getNumRings(); ++i)
    {
      if (IsNullOrEmpty(polygon->getRingN(i)) == false)
      {
        return false;
      }
    }
  }
  else if (geometry->get2DGeomTypeId() == te::gm::LineStringType)
  {
    const te::gm::LineString* lineString = static_cast<const te::gm::LineString*>(geometry);
    if (lineString->size() > 0)
    {
      return false;
    }
  }
  else
  {
    return geometry->isEmpty();
  }

  return true;
}

te::gm::GeometryPtr te::gm::CreatePolygon(const te::gm::Curve* curve)
{
  te::gm::Polygon* polygon = new te::gm::Polygon(0, te::gm::PolygonType, curve->getSRID());
  polygon->add(static_cast<te::gm::Curve*>(curve->clone()));

  te::gm::GeometryPtr result(polygon);
  return result;
}

bool te::gm::isCCW(const te::gm::LinearRing* ring)
{
  double S2 = 0.0;

  for (std::size_t i = 0; i < ring->getNPoints() - 1; ++i)
  {
    double xi = ring->getX(i);
    double yi = ring->getY(i);

    double xf = ring->getX(i+1);
    double yf = ring->getY(i + 1);

    S2 += (xi + xf) * (yf - yi);
  }

  if (S2 > 0.0)
  {
    return true;
  }
  else
  {
    return false;
  }
}
double te::gm::GetArea(const te::gm::Geometry* geometry)
{
  if (te::gm::IsNullOrEmpty(geometry) == true)
  {
    return 0.0;
  }

  if (geometry->isCollection() == true)
  {
    const te::gm::GeometryCollection* collection = static_cast<const te::gm::GeometryCollection*>(geometry);

    double area = 0.0;
    for (std::size_t i = 0; i < collection->getNumGeometries(); ++i)
    {
      area += GetArea(collection->getGeometryN(i));
    }
    return area;
  }
  else if (geometry->get2DGeomTypeId() == te::gm::PolygonType)
  {
    const te::gm::Polygon* polygon = static_cast<const te::gm::Polygon*>(geometry);
    return polygon->getArea();
  }

  return 0.0;
}

te::gm::GeometryPtr te::gm::CreateCollectionFromGeometryVector(const te::gm::GeometryVector& geometryVector)
{
  if (geometryVector.empty() == true)
  {
    return nullptr;
  }

  int srid = geometryVector.at(0)->getSRID();

  te::gm::GeometryCollection* collection = new te::gm::GeometryCollection(0, te::gm::GeometryCollectionType, srid);
  for (std::size_t i = 0; i < geometryVector.size(); ++i)
  {
    collection->add(static_cast<te::gm::Geometry*>(geometryVector.at(i)->clone()));
  }

  te::gm::GeometryPtr geometry(collection);
  return geometry;
}

te::gm::GeometryVectorConst te::gm::ToConstVector(const te::gm::GeometryVector& vecGeometries)
{
  te::gm::GeometryVectorConst vecGeometriesConst;
  vecGeometriesConst.reserve(vecGeometries.size());
  for (te::gm::Geometry* geometry : vecGeometries)
  {
    vecGeometriesConst.push_back(geometry);
  }

  return vecGeometriesConst;
}

std::vector<te::gm::Geometry*> te::gm::SplitGeometry(const te::gm::Geometry* inputGeometry, const te::gm::Geometry* bladeLineGeometry)
{
  //we first extract the boundary of the polygon
  std::unique_ptr<te::gm::Geometry> inputGeometryBoundary(inputGeometry->getBoundary());

  //then we make the union between the geometry to be splitted and the split line, creating a subdivision
  std::unique_ptr<te::gm::Geometry> subdividedGeometry(inputGeometryBoundary->Union(bladeLineGeometry));

  //then we execute a polygonyze algorithm to create one polygon to each part of the subivision

  std::vector<te::gm::Polygon*> vecParts;
  te::gm::Polygonizer(subdividedGeometry.get(), vecParts);

  //as we have polygonized the borders, we could have created artefacts inside holes
  //for this reason, we must discard all the parts that are not covered by the original geometry
  //it is similar to executing an intersection operation with the original geometry
  std::vector<te::gm::Geometry*> vecResult;
  int step = 0;
  for (te::gm::Polygon* part : vecParts)
  {
    ++step;
    te::gm::Coord2D interiorPoint = te::gm::GetInteriorPoint(part);

    bool considerAsResult = false;
    if (interiorPoint.x == TE_DOUBLE_NOT_A_NUMBER || interiorPoint.y == TE_DOUBLE_NOT_A_NUMBER)
    {
      //i the interior point was not sucessfully calculated, we make check if the part si within the original polygion
      //this operation is much more expensive
      considerAsResult = part->within(inputGeometry);
    }
    else
    {
      //if the interior point was sucessfully calculated, we use is as it is faster
      considerAsResult = te::gm::IsPointOnPolygon(interiorPoint, inputGeometry);
    }

    if (considerAsResult == true)
    {
      vecResult.push_back(static_cast<te::gm::Geometry*>(part));
    }
    else
    {
      //discard the result
      delete part;
    }
  }

  return vecResult;
}


