/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralib/geometry/algorithms/CellTilling.cpp

   \brief Algorithm to help controlling creating and iterating in a tile
 */

#include "CellTilling.h"

#include "../../common/MathUtils.h"
#include "../../geometry/Coord2D.h"
#include "../../geometry/Envelope.h"

#include <cmath>

te::gm::CellTilling::CellTilling(const te::gm::Envelope& envelope, double resolutionX, double resolutionY)
  : m_envelope(new te::gm::Envelope(envelope))
  , m_resolutionX(resolutionX)
  , m_resolutionY(resolutionY)
{
  m_numColumns = static_cast<std::size_t>(std::ceil(m_envelope->getWidth() / m_resolutionX));
  m_numRows = static_cast<std::size_t>(std::ceil(m_envelope->getHeight() / m_resolutionY));
}

te::gm::CellTilling::~CellTilling() = default;

std::size_t te::gm::CellTilling::getNumColumns() const
{
  return m_numColumns;
}

std::size_t te::gm::CellTilling::getNumRows() const
{
  return m_numRows;
}

std::size_t te::gm::CellTilling::getNumTiles() const
{
  std::size_t numTiles = m_numColumns * m_numRows;
  return numTiles;
}

te::gm::Envelope te::gm::CellTilling::getTile(std::size_t column, std::size_t row) const
{
  double x1 = m_envelope->m_llx + (column * m_resolutionX);
  double y1 = m_envelope->m_lly + (row * m_resolutionY);

  double x2 = m_envelope->m_llx + ((column + 1) * m_resolutionX);
  double y2 = m_envelope->m_lly + ((row + 1) * m_resolutionY);

  te::gm::Envelope envelope(x1, y1, x2, y2);
  return envelope;
}

std::size_t te::gm::CellTilling::getTileColumn(const te::gm::Coord2D& coord) const
{
  double distanceFromOrigin = coord.x - m_envelope->m_llx;
  std::size_t columnIndex = static_cast<std::size_t>(std::floor(distanceFromOrigin / m_resolutionX));
  return columnIndex;
}

std::size_t te::gm::CellTilling::getTileRow(const te::gm::Coord2D& coord) const
{
  double distanceFromOrigin = coord.y - m_envelope->m_lly;
  std::size_t rowIndex = static_cast<std::size_t>(std::floor(distanceFromOrigin / m_resolutionY));
  return rowIndex;
}