/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralib/geometry/algorithms/CellTilling.h

   \brief Algorithm to help controlling creating and iterating in a tile
 */

#ifndef __TERRALIB_GEOMETRY_INTERNAL_CELLTILLING_H
#define __TERRALIB_GEOMETRY_INTERNAL_CELLTILLING_H

 // TerraLib
#include "../Config.h"

//STL
#include <cstdio>
#include <memory>

namespace te
{
  namespace gm
  {
    struct Coord2D;
    class Envelope;

    /*!
     \class CellTilling

     \brief Algorithm to help controlling creating and iterating in a tile
     */
    class TEGEOMEXPORT CellTilling
    {
    public:

      /*!
        * \brief Constructor
        *
        * \param envelope The envelope to be used for tilling
        * \param resolution The resolution of each tile
        */
      CellTilling(const te::gm::Envelope& envelope, double resolutionX, double resolutionY);

      //!< Destructor
      ~CellTilling();

      /*!
        * \brief Returns the number of columns of the tilling
        *
        * \return The number of columns of the tilling
        */
      std::size_t getNumColumns() const;

      /*!
        * \brief Returns the number of rows of the tilling
        *
        * \return The number of rows of the tilling
        */
      std::size_t getNumRows() const;

      /*!
        * \brief Returns the number of tiles of the tilling
        *
        * \return The number of tiles of the tilling
        */
      std::size_t getNumTiles() const;

      /*!
        * \brief Returns the tile located in the given row and column
        *
        * \return The tile located in the given row and column
        */
      te::gm::Envelope getTile(std::size_t column, std::size_t row) const;

      //!< Find the tile column index where the given coordinate is located
      std::size_t getTileColumn(const te::gm::Coord2D & coord) const;

      //!< Find the tile row index where the given coordinate is located
      std::size_t getTileRow(const te::gm::Coord2D & coord) const;

    protected:

      std::unique_ptr<te::gm::Envelope> m_envelope; //!< The reference envelope
      double m_resolutionX; //!< The resolution of the tilling in the x-axis
      double m_resolutionY; //!< The resolution of the tilling in the y-axis
      std::size_t m_numColumns; //!< The number of columns of the tilling
      std::size_t m_numRows; //!< The number of rows of the tilling
    };
  } //namespace gm
} //namespace te

#endif // __TERRALIB_GEOMETRY_INTERNAL_CELLTILLING_H
