/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralib/geometry/algorithms/GeometrySubdivider.h

   \brief This class contains functions to help subdivide a geometry into small parts
 */

#include "GeometrySubdivider.h"

#include "../../common/Exception.h"
#include "../../common/STLUtils.h"
#include "../validation/FixNoding.h"
#include "../Geometry.h"
#include "../Polygon.h"
#include "../Utils.h"
#include "CellTilling.h"


te::gm::GeometrySubdivider::GeometrySubdivider() = default;

te::gm::GeometrySubdivider::~GeometrySubdivider() = default;

te::gm::GeometryPtr te::gm::GeometrySubdivider::subdivideByClipping(const te::gm::Geometry* geometry, const te::gm::Envelope & envelope)
{
  te::gm::GeometryPtr result = subdivideByClippingImpl(geometry, envelope);
  result.reset(fixNoding(result.get()));

  return result;
}

te::gm::GeometryVector te::gm::GeometrySubdivider::subdivideByCellTilling(const te::gm::Geometry* geometry, const te::gm::CellTilling& cellTilling, std::vector<ColumnRow>& vecColumnRow)
{
  te::gm::GeometryVector vecResult = subdivideByCellTillingImpl(geometry, cellTilling, vecColumnRow);
  te::gm::GeometryVector vecResultFixed = fixNoding(vecResult);

  te::common::FreeContents(vecResult);

  return vecResultFixed;
}

te::gm::GeometryVector te::gm::GeometrySubdivider::subdivideByMaxCoordinates(const te::gm::Geometry* geometry, std::size_t maxCoordinates)
{
  te::gm::GeometryVector vecResult = subdivideByMaxCoordinatesImpl(geometry, maxCoordinates);
  te::gm::GeometryVector vecResultFixed = fixNoding(vecResult);

  te::common::FreeContents(vecResult);

  return vecResultFixed;
}

te::gm::Geometry* te::gm::GeometrySubdivider::fixNoding(const te::gm::Geometry* geometry)
{
  te::gm::GeometryPtr result;
  if (te::gm::IsNullOrEmpty(geometry) == false)
  {
    if (geometry->getDimension() == te::gm::A)
    {
      //we must ensure that all geometries are valid
      //in some cases, this fast algorithm can lead to self-intersectiong geometries.
      te::gm::FixNoding fixNoding;
      result.reset(fixNoding.validate(geometry).release());
    }
    else
    {
      result.reset(static_cast<te::gm::Geometry*>(geometry->clone()));
    }
  }

  return result.release();
}

te::gm::GeometryVector te::gm::GeometrySubdivider::fixNoding(const te::gm::GeometryVector& vecGeometries)
{
  te::gm::GeometryVector vecResult;
  vecResult.reserve(vecGeometries.size());
  for (std::size_t i = 0; i < vecGeometries.size(); ++i)
  {
    vecResult.push_back(fixNoding(vecGeometries.at(i)));
  }
  return vecResult;
}

te::gm::GeometryPtr te::gm::GeometrySubdivider::subdivideByClippingImpl(const te::gm::Geometry* geometry, const te::gm::Envelope& envelope)
{
  te::gm::GeometryPtr result;
  try
  {
    result.reset(te::gm::ClipGeometry(geometry, envelope));
  }
  catch (const std::exception&)
  {
    //if we get an exception, we try to execute an intersection. This is much slower, but ir more robust
    te::gm::Polygon polygon = te::gm::CreatePolygon(envelope, geometry->getSRID(), 0);

    try
    {
      result.reset(polygon.intersection(geometry));
      if (result != nullptr && result->isEmpty() == false)
      {
        //we must discard the geometries that does not have the same type from the input
        result.reset(te::gm::NormalizeGeometryByType(result.get(), te::gm::GetSimpleType(geometry->get2DGeomTypeId())));
      }
    }
    catch (const std::exception&)
    {
      throw te::common::Exception("Error subdividing geometries");
    }
  }

  return result;
}

te::gm::GeometryVector te::gm::GeometrySubdivider::subdivideByCellTillingImpl(const te::gm::Geometry* geometry, const te::gm::CellTilling& cellTilling, std::vector<ColumnRow>& vecColumnRow)
{
  const te::gm::Envelope* mbr = geometry->getMBR();

  std::size_t initRow = cellTilling.getTileRow(mbr->getLowerLeft());
  std::size_t initColumn = cellTilling.getTileColumn(mbr->getLowerLeft());

  std::size_t finalRow = cellTilling.getTileRow(mbr->getUpperRight());
  std::size_t finalColumn = cellTilling.getTileColumn(mbr->getUpperRight());

  te::gm::GeometryVector vecResult;
  for (std::size_t row = initRow; row <= finalRow; ++row)
  {
    for (std::size_t column = initColumn; column <= finalColumn; ++column)
    {
      te::gm::Envelope currentTileEnvelope = cellTilling.getTile(column, row);
      te::gm::GeometryPtr result = subdivideByClipping(geometry, currentTileEnvelope);
      if (te::gm::IsNullOrEmpty(result) == false)
      {
        ColumnRow columnRow(column, row);

        vecResult.push_back(result.release());
        vecColumnRow.push_back(columnRow);
      }
    }
  }

  return vecResult;
}

te::gm::GeometryVector te::gm::GeometrySubdivider::subdivideByMaxCoordinatesImpl(const te::gm::Geometry* geometry, std::size_t maxCoordinates)
{
  te::gm::GeometryVector vecResult;
  //we cannot handle subdivision with less than 5 coordinates
  if (maxCoordinates < 5)
  {
    vecResult.push_back(static_cast<te::gm::Geometry*>(geometry->clone()));
    return vecResult;
  }

  //if the given geometry already has less than the maximum number of coordinates, we just return it
  if (geometry->getNPoints() <= maxCoordinates)
  {
    vecResult.push_back(static_cast<te::gm::Geometry*>(geometry->clone()));
    return vecResult;
  }

  double precision = te::gm::GetAppropriatePrecision(geometry->getSRID());

  te::gm::Envelope envelope(*geometry->getMBR());
  envelope.m_llx -= precision;
  envelope.m_lly -= precision;
  envelope.m_urx += precision;
  envelope.m_ury += precision;

  return subdivideByMaxCoordinatesImpl(geometry, envelope, maxCoordinates);
}

te::gm::GeometryVector te::gm::GeometrySubdivider::subdivideByMaxCoordinatesImpl(const te::gm::Geometry* geometry, const te::gm::Envelope& envelope, std::size_t maxCoordinates)
{
  //if the given geometry already has less than the maximum number of coordinates, we just return it
  te::gm::GeometryVector vecResult;

  te::gm::GeometryPtr result;
  te::gm::GeometryVector vecSingleResult;
  if (geometry->getMBR()->within(envelope) == true)
  {
    te::gm::Multi2Single(geometry, vecSingleResult);
  }
  else
  {
    result.reset(subdivideByClipping(geometry, envelope).release());
    if (te::gm::IsNullOrEmpty(result) == true)
    {
      return vecResult;
    }

    te::gm::Multi2Single(result.get(), vecSingleResult);
  }

  for (std::size_t i = 0; i < vecSingleResult.size(); ++i)
  {
    //this will destroy the geometry after its use
    te::gm::GeometryPtr singleGeometry(vecSingleResult.at(i));
    vecSingleResult.at(i) = nullptr;

    //if the given geometry already has less than the maximum number of coordinates, we just return it
    if (singleGeometry->getNPoints() <= maxCoordinates)
    {
      vecResult.push_back(singleGeometry.release());
      continue;
    }

    std::size_t splitCount = 2;

    double width = envelope.getWidth();
    double height = envelope.getHeight();
    double widthStep = width / splitCount;
    double heightStep = height / splitCount;

    for (std::size_t row = 0; row < splitCount; ++row)
    {
      for (std::size_t column = 0; column < splitCount; ++column)
      {
        double x1 = envelope.m_llx + (widthStep * column);
        double x2 = envelope.m_llx + (widthStep * (column + 1));
        double y1 = envelope.m_lly + (heightStep * row);
        double y2 = envelope.m_lly + (heightStep * (row + 1));

        te::gm::Envelope subEnvelope(x1, y1, x2, y2);
        te::gm::GeometryVector vecCurrentSingleResult = subdivideByMaxCoordinatesImpl(singleGeometry.get(), subEnvelope, maxCoordinates);
        vecResult.insert(vecResult.end(), vecCurrentSingleResult.begin(), vecCurrentSingleResult.end());
      }
    }
  }

  return vecResult;
}
