/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralib/geometry/algorithms/GeometrySubdivider.h

   \brief This class contains functions to help subdivide a geometry into small parts
 */

#ifndef __TERRALIB_GEOMETRY_INTERNAL_GEOMETRYSUBDIVIDER_H
#define __TERRALIB_GEOMETRY_INTERNAL_GEOMETRYSUBDIVIDER_H

 // TerraLib
#include "../Config.h"

#include "../CommonDataStructures.h"

namespace te
{
  namespace gm
  {
    class CellTilling;
    class Geometry;
    class GeometryPtr;

    /*!
     \class GeometrySubdivider

     \brief This class contains functions to help subdivide a geometry into small parts
     */
    class TEGEOMEXPORT GeometrySubdivider
    {
    public:

      struct ColumnRow
      {
        ColumnRow(std::size_t column, std::size_t row)
          : m_column(column)
          , m_row(row)
        {
        }

        std::size_t m_column;
        std::size_t m_row;
      };


      /*!
       * \brief Constructor
       */
      GeometrySubdivider();

      /*!
       * \brief Destructor
       */
      virtual ~GeometrySubdivider();

      static te::gm::GeometryPtr subdivideByClipping(const te::gm::Geometry* geometry, const te::gm::Envelope& envelope);

      static te::gm::GeometryVector subdivideByCellTilling(const te::gm::Geometry* geometry, const te::gm::CellTilling& cellTilling, std::vector<ColumnRow>& vecColumnRow);

      /*!
        * \brief Subdivides the given geometry recursivelly using its envelope spplited in 4 parts until all the subparts have less then the given 'maxCoordinates' coordinates
        *
        * \param geometry The input geometry to be subdivided
        * \param maxCoordinates The maximum nunber of coordinates of each part
        *
        * \return A vector containg all the subparts of the geometry
        */
      static te::gm::GeometryVector subdivideByMaxCoordinates(const te::gm::Geometry* geometry, std::size_t maxCoordinates);

    protected:

      static te::gm::Geometry* fixNoding(const te::gm::Geometry* geometry);

      static te::gm::GeometryVector fixNoding(const te::gm::GeometryVector& vecGeometries);

      static te::gm::GeometryPtr subdivideByClippingImpl(const te::gm::Geometry* geometry, const te::gm::Envelope& envelope);

      static te::gm::GeometryVector subdivideByCellTillingImpl(const te::gm::Geometry* geometry, const te::gm::CellTilling& cellTilling, std::vector<ColumnRow>& vecColumnRow);

      /*!
        * \brief Subdivides the given geometry recursivelly using its envelope spplited in 4 parts until all the subparts have less then the given 'maxCoordinates' coordinates
        *
        * \param geometry The input geometry to be subdivided
        * \param maxCoordinates The maximum nunber of coordinates of each part
        *
        * \return A vector containg all the subparts of the geometry
        */
      static te::gm::GeometryVector subdivideByMaxCoordinatesImpl(const te::gm::Geometry* geometry, std::size_t maxCoordinates);

      /*!
        * \brief Subdivides the given geometry recursivelly using its envelope spplited in 4 parts until all the subparts have less then the given 'maxCoordinates' coordinates
        *
        * \param geometry The input geometry to be subdivided
        * \param envelope The envelope to be used to clip the geometry.
        * \param maxCoordinates The maximum nunber of coordinates of each part
        *
        * \return A vector containg all the subparts of the geometry
        */
      static te::gm::GeometryVector subdivideByMaxCoordinatesImpl(const te::gm::Geometry* geometry, const te::gm::Envelope& envelope, std::size_t maxCoordinates);

    };
  } //namespace gm
} //namespace te

#endif // __TERRALIB_GEOMETRY_INTERNAL_GEOMETRYSUBDIVIDER_H
