/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralib/geometry/algorithms/SafeOverlay.cpp

   \brief Algorithm to help controlling creating and iterating in a tile
 */

#include "SafeOverlay.h"

#include "../Geometry.h"
#include "../Polygon.h"
#include "../Validation.h"
#include "../Utils.h"

te::gm::Geometry* te::gm::SafeOverlay::ClipGeometry(const te::gm::Geometry* geometry, const te::gm::Envelope& envelope)
{
  std::unique_ptr<te::gm::Geometry> result;
  bool tryAgain = false;
  try
  {
    result.reset(te::gm::ClipGeometry(geometry, envelope));
  }
  catch (...)
  {
    tryAgain = true;
  }

  if (tryAgain == true)
  {
    int srid = geometry->getSRID();
    te::gm::Polygon polygon = te::gm::CreatePolygon(envelope, srid, 0);
    result.reset(te::gm::SafeOverlay::Intersection(geometry, static_cast<te::gm::Geometry*>(&polygon)).release());
  }

  if (te::gm::IsNullOrEmpty(result.get()) == false)
  {
    //we must discard the geometries that does not have the same type from the input
    result.reset(te::gm::NormalizeGeometryByType(result.get(), te::gm::GetSimpleType(geometry->get2DGeomTypeId())));
  }
  else
  {
    result.reset();
  }

  return result.release();
}

te::gm::GeometryPtr te::gm::SafeOverlay::Difference(const te::gm::Geometry* g1, const te::gm::Geometry* g2)
{
  te::gm::GeometryPtr result;
  try
  {
    result.reset(g1->difference(g2));
    return result.release();
  }
  catch (const std::exception&)
  {
  }

  std::unique_ptr<GeometryPair> geometryPair = te::gm::SafeOverlay::MakeValid(g1, g2);
  if (te::gm::IsNullOrEmpty(geometryPair->first) == true)
  {
    return nullptr;
  }
  if (te::gm::IsNullOrEmpty(geometryPair->second) == true)
  {
    return geometryPair->first.release();
  }

  try
  {
    result.reset(geometryPair->first->difference(geometryPair->second.get()));
  }
  catch (const std::exception&)
  {
    //re-throw the exception
    throw;
  }

  return result.release();
}

te::gm::GeometryPtr te::gm::SafeOverlay::Intersection(const te::gm::Geometry* g1, const te::gm::Geometry* g2)
{
  te::gm::GeometryPtr result;
  try
  {
    result.reset(g1->intersection(g2));
    return result.release();
  }
  catch (const std::exception&)
  {
  }

  std::unique_ptr<te::gm::GeometryPair> geometryPair = te::gm::SafeOverlay::MakeValid(g1, g2);
  if (te::gm::IsNullOrEmpty(geometryPair->first) == true || te::gm::IsNullOrEmpty(geometryPair->second) == true)
  {
    return nullptr;
  }

  try
  {
    result.reset(geometryPair->first->intersection(geometryPair->second.get()));
  }
  catch (const std::exception&)
  {
    //re-throw the exception
    throw;
  }

  return result.release();
}

std::unique_ptr<te::gm::GeometryPair> te::gm::SafeOverlay::MakeValid(const te::gm::Geometry* g1, const te::gm::Geometry* g2)
{
  std::unique_ptr<GeometryPair> geometryPair(new te::gm::GeometryPair());

  try
  {
    //if we get here, we try to fix both geometries and try again
    geometryPair->first.reset(te::gm::Validation::MakeValidWithSnap(g1).release());
    geometryPair->second.reset(te::gm::Validation::MakeValidWithSnap(g2).release());
  }
  catch (const std::exception&)
  {
    //re-throw the exception
    throw;
  }

  return geometryPair;
}
