/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralib/geometry/algorithms/SafeOverlay.h

   \brief A class containing helper methods to make overlay algorithms more robust and safer to execute when dealing with invalid geometries.
 */

#ifndef __TERRALIB_GEOMETRY_INTERNAL_SAFEOVERLAY_H
#define __TERRALIB_GEOMETRY_INTERNAL_SAFEOVERLAY_H

 // TerraLib
#include "../Config.h"

//STL
#include <memory>
#include <utility>

namespace te
{
  namespace gm
  {
    class Envelope;
    class Geometry;
    class GeometryPtr;

    using GeometryPair = std::pair<te::gm::GeometryPtr, te::gm::GeometryPtr>;

    /*!
     \class SafeOverlay

     \brief A class containing helper methods to make overlay algorithms more robust and safer to execute when dealing with invalid geometries.
     \description When some exception is thrown during the execution of the basic overlay algorithms,
                  this class will handle it, try to make the geoemtries valid, and try again
     */
    class TEGEOMEXPORT SafeOverlay
    {
    public:

      //!< Constructor
      SafeOverlay() = delete;

      //!< Destructor
      ~SafeOverlay() = delete;

      //!< Tries to clip the geometry. If an exception is thrown, it tries to make the geometry valid and retry again
      static te::gm::Geometry* ClipGeometry(const te::gm::Geometry* geometry, const te::gm::Envelope& envelope);

      //!< Tries to make the difference. If an exception is thrown, it tries to make the geometry valid and retry again
      static te::gm::GeometryPtr Difference(const te::gm::Geometry* g1, const te::gm::Geometry* g2);

      //!< Tries to make the intersection. If an exception is thrown, it tries to make the geometry valid and retry again
      static te::gm::GeometryPtr Intersection(const te::gm::Geometry* g1, const te::gm::Geometry* g2);

      //!< Tries to make the geometry pair valid. If it is unable to do that, it will throw an exception
      static std::unique_ptr<te::gm::GeometryPair> MakeValid(const te::gm::Geometry* g1, const te::gm::Geometry* g2);
    };
  } //namespace gm
} //namespace te

#endif // __TERRALIB_GEOMETRY_INTERNAL_SAFEOVERLAY_H
