#include "AbstractValidator.h"

#include "../../common/STLUtils.h"
#include "../CommonDataStructures.h"
#include "../Geometry.h"
#include "../GeometryCollection.h"
#include "../LineString.h"
#include "../LinearRing.h"
#include "../Polygon.h"
#include "../Utils.h"

te::gm::AbstractValidator::AbstractValidator()
  : m_precision(0.0)
{

}


te::gm::GeometryPtr te::gm::AbstractValidator::validate(const te::gm::Geometry* geometry) const
{
  bool wasChanged = false;
  return validate(geometry, wasChanged);
}

te::gm::GeometryPtr te::gm::AbstractValidator::validate(const te::gm::Geometry* geometry, bool& wasChanged) const
{
  if (te::gm::IsNullOrEmpty(geometry) == true)
  {
    te::gm::GeometryPtr result;
    return result;
  }

  if (geometry->isCollection() == true)
  {
    //if we have a collection, we must open it to analyse each single geometry
    const te::gm::GeometryCollection* inputCollection = static_cast<const  te::gm::GeometryCollection*>(geometry);
    return validateCollection(inputCollection, wasChanged);
  }
  else if (geometry->get2DGeomTypeId() == te::gm::PolygonType)
  {
    //if we have a polygon, we must open each to analyse each single line
    const te::gm::Polygon* inputPolygon = static_cast<const  te::gm::Polygon*>(geometry);
    return validatePolygon(inputPolygon, wasChanged);
  }
  else if (geometry->get2DGeomTypeId() == te::gm::LineStringType)
  {
    //if we have a linestring, we must can now validate it
    const te::gm::LineString* inputLineString = static_cast<const  te::gm::LineString*>(geometry);
    return validateLine(inputLineString, wasChanged);
  }

  //we dont know how to fix this type, so we clone it and return
  te::gm::GeometryPtr result(static_cast<te::gm::Geometry*>(geometry->clone()));
  return result;
}

void te::gm::AbstractValidator::setPrecision(double precision)
{
  m_precision = precision;
}

te::gm::GeometryPtr te::gm::AbstractValidator::validateCollection(const te::gm::GeometryCollection* collection, bool& wasChanged) const
{
  wasChanged = false;

  //if we have a collection, we must open it to analyse each single geometry
  std::unique_ptr<te::gm::GeometryCollection> outputCollection(static_cast<te::gm::GeometryCollection*>(collection->clone()));
  outputCollection->clear();

  for (std::size_t i = 0; i < collection->getNumGeometries(); ++i)
  {
    bool currentWasChanged = false;
    te::gm::GeometryPtr currentResult = validate(collection->getGeometryN(i), currentWasChanged);

    wasChanged = wasChanged | currentWasChanged;

    if (te::gm::IsNullOrEmpty(currentResult) == false)
    {
      if (currentResult->isCollection() == true)
      {
        te::gm::GeometryVector vecSingle;
        te::gm::Multi2Single(currentResult.get(), vecSingle);
        for (std::size_t j = 0; j < vecSingle.size(); ++j)
        {
          outputCollection->add(vecSingle.at(j));
        }
      }
      else
      {
        outputCollection->add(currentResult.release());
      }
    }
  }

  te::gm::GeometryPtr result;
  if (te::gm::IsNullOrEmpty(outputCollection.get()) == true)
  {
    return result;
  }

  result.reset(outputCollection.release());
  return result;
}

te::gm::GeometryPtr te::gm::AbstractValidator::validatePolygon(const te::gm::Polygon* polygon, bool& wasChanged) const
{
  if (polygon->getNumRings() == 0)
  {
    return nullptr;
  }

  wasChanged = false;

  //we handle all the rings
  te::gm::GeometryVector vecOuterRings;
  te::gm::GeometryVector vecInnerRings;

  bool handleMultiType = false;
  for (std::size_t i = 0; i < polygon->getNumRings(); ++i)
  {
    te::gm::LinearRing* originalRing = static_cast<te::gm::LinearRing*>(polygon->getRingN(i));
    bool isOriginalRingCCW = te::gm::isCCW(originalRing);

    bool currentWasChanged = false;
    te::gm::GeometryPtr currentResult = validate(polygon->getRingN(i), currentWasChanged);

    wasChanged = wasChanged | currentWasChanged;

    if (te::gm::IsNullOrEmpty(currentResult) == false)
    {
      std::vector<te::gm::Geometry*> vecSinglePart;
      if (currentResult->isCollection() == true)
      {
        handleMultiType = true;
        te::gm::Multi2Single(currentResult.get(), vecSinglePart);
      }
      else
      {
        vecSinglePart.push_back(currentResult.release());
      }

      //then we annalyse the orientation of the rings
      for (std::size_t j = 0; j < vecSinglePart.size(); ++j)
      {
        te::gm::LinearRing* newRing = static_cast<te::gm::LinearRing*>(vecSinglePart.at(j));
        bool isNewRingCCW = te::gm::isCCW(newRing);

        //if the outer ring has been splited, we must check its orientation. Maybe part of it must be kept as a hole
        if (isOriginalRingCCW == isNewRingCCW)
        {
          if (i == 0)
          {
            vecOuterRings.push_back(vecSinglePart.at(j));
          }
          else
          {
            vecInnerRings.push_back(vecSinglePart.at(j));
          }
        }
        else
        {
          if (i == 0)
          {
            vecInnerRings.push_back(vecSinglePart.at(j));
          }
          else
          {
            vecOuterRings.push_back(vecSinglePart.at(j)); 
          }
        }
      }
    }
  }

  if (wasChanged == false)
  {
    te::common::FreeContents(vecOuterRings);
    te::common::FreeContents(vecInnerRings);

    te::gm::GeometryPtr result(static_cast<te::gm::Geometry*>(polygon->clone())); 
    return result;
  }

  if (vecOuterRings.empty() == true)
  {
    te::common::FreeContents(vecOuterRings);
    te::common::FreeContents(vecInnerRings);

    te::gm::GeometryPtr result;
    return result;
  }

  te::gm::GeometryPtr result;
  if (handleMultiType == true)
  {
    result.reset(mountTopology(vecOuterRings, vecInnerRings).release());
    
    te::common::FreeContents(vecOuterRings);
    te::common::FreeContents(vecInnerRings);
  }
  else
  {
    te::gm::Polygon* outputPolygon = new te::gm::Polygon(0, polygon->get2DGeomTypeId(), polygon->getSRID());
    outputPolygon->add(static_cast<te::gm::Curve*>(vecOuterRings.at(0)));

    for (std::size_t i = 0; i < vecInnerRings.size(); ++i)
    {
      outputPolygon->add(static_cast<te::gm::Curve*>(vecInnerRings.at(i)));
    }

    result.reset(outputPolygon);
  }

  return result;
}

struct lesserEnvelopeSize
{
  bool operator()(const te::gm::Geometry* g1, const te::gm::Geometry* g2)
  {
    if (g1->getMBR()->within(*g2->getMBR()) && g1->getMBR()->equals(*g2->getMBR()) == false)
    {
      return true;
    }
    return false;
  }
};

te::gm::GeometryPtr te::gm::AbstractValidator::mountTopology(const te::gm::GeometryVector& vecOuterRings, const te::gm::GeometryVector& vecInnerRings) const
{
  te::gm::GeometryVector vecOuterRingsCopySorted(vecOuterRings.begin(), vecOuterRings.end());
  te::gm::GeometryVector vecInnerRingsCopy(vecInnerRings.begin(), vecInnerRings.end());

  std::sort(vecOuterRingsCopySorted.begin(), vecOuterRingsCopySorted.end(), lesserEnvelopeSize());


  te::gm::GeometryVector vecOuterPolygons;

  //we first create the outer polygons
  for (std::size_t i = 0; i < vecOuterRingsCopySorted.size(); ++i)
  {
    te::gm::GeometryPtr outerRingAsPolygon = CreatePolygon(static_cast<te::gm::Curve*>(vecOuterRingsCopySorted.at(i)));
    vecOuterPolygons.push_back(outerRingAsPolygon.release());
  }

  te::gm::GeometryVector undefinedHoleVector;
  std::vector< te::gm::GeometryVector > vecHolesPerPolgygon(vecOuterPolygons.size());;
  for (std::size_t i = 0; i < vecInnerRingsCopy.size(); ++i)
  {
    te::gm::GeometryPtr innerRingAsPolygon = CreatePolygon(static_cast<te::gm::Curve*>(vecInnerRingsCopy.at(i)));
    
    bool found = false;
    for (std::size_t j = 0; j < vecOuterPolygons.size(); ++j)
    {
      const te::gm::Geometry* outerPolygon = vecOuterPolygons.at(j);
      if (innerRingAsPolygon->within(outerPolygon) == true)
      {
        vecHolesPerPolgygon[j].push_back(innerRingAsPolygon.release());
        found = true;
        break;
      }
    }

    if (found == false)
    {
      undefinedHoleVector.push_back(innerRingAsPolygon.release());
    }
  }

  te::gm::GeometryVector outputPolygons;
  for (std::size_t i = 0; i < vecOuterPolygons.size(); ++i)
  {
    te::gm::GeometryPtr outerPolygon(vecOuterPolygons.at(i));
    const te::gm::GeometryVector& vecHoles = vecHolesPerPolgygon.at(i);

    if (vecHoles.empty() == false)
    {
      te::gm::GeometryPtr innerPolygon = te::gm::CascadedUnion(vecHoles);
      outerPolygon.reset(outerPolygon->difference(innerPolygon.get()));
      outerPolygon.reset(te::gm::NormalizeGeometryByType(outerPolygon.get(), te::gm::PolygonType));
    }

    outputPolygons.push_back(outerPolygon.release());
  }

  te::gm::GeometryPtr outerPolygon = te::gm::CascadedUnion(outputPolygons);
  if (undefinedHoleVector.empty() == false)
  {
    te::gm::GeometryPtr undefinedHoles = te::gm::CascadedUnion(undefinedHoleVector);
    outerPolygon.reset(outerPolygon->difference(undefinedHoles.get()));
  }
  outerPolygon.reset(te::gm::NormalizeGeometryByType(outerPolygon.get(), te::gm::PolygonType));

  te::common::FreeContents(outputPolygons);
  te::common::FreeContents(undefinedHoleVector);

  return outerPolygon;
}
