/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralib/geometry/validation/AbstractValidator.h

   \brief An abstract class to represent an algorithm that validates an geometry
 */

#ifndef __TERRALIB_GEOMETRY_INTERNAL_ABSTRACTVALIDATOR_H
#define __TERRALIB_GEOMETRY_INTERNAL_ABSTRACTVALIDATOR_H

 // TerraLib
#include "../Config.h"

#include "../CommonDataStructures.h"

#include <memory>

namespace te
{
  namespace gm
  {
    class Geometry;
    class GeometryCollection;
    class GeometryPtr;
    class LineString;
    class Polygon;

    /*!
      \class AbstractValidator

      \brief An abstract class to represent an algorithm that validates an geometry
    */
    class TEGEOMEXPORT AbstractValidator
    {
    public:

      /*!
       * \brief Constructor
       */
      AbstractValidator();

      /*!
       * \brief Destructor
       */
      virtual ~AbstractValidator() = default;

      /*!
       * \brief Validates a geometry
       *
       * \param geometry The input geometry to be validate
       *
       * \return The valdiated result geometry
       */
      virtual te::gm::GeometryPtr validate(const te::gm::Geometry* geometry) const;

      /*!
       * \brief Validates a geometry, an returns additional variable to inform if the geometry has been changed or not 
       *
       * \param geometry The input geometry to be validate
       * \param wasChanged output param. TRUE if the geometry has been changed. FALSE otherwise
       *
       * \return The valdiated result geometry
       */
      virtual te::gm::GeometryPtr validate(const te::gm::Geometry* geometry, bool& wasChanged) const;

      /*!
       * \brief When appliable, sets the precision to be used in the validation
       *
       * \param precision When appliable, the precision to be used in the validation
       */
      virtual void setPrecision(double precision);

    protected:

      /*!
       * \brief Validates a geometry collection, an returns additional variable to inform if the geometry has been changed or not
       *
       * \param geometry The input geometry collection to be validate
       * \param wasChanged output param. TRUE if the geometry has been changed. FALSE otherwise
       *
       * \return The valdiated result geometry collection
       */
      virtual te::gm::GeometryPtr validateCollection(const te::gm::GeometryCollection* collection, bool& wasChanged) const;

      /*!
       * \brief Validates a polygon, an returns additional variable to inform if the geometry has been changed or not
       *
       * \param geometry The input polygon to be validate
       * \param wasChanged output param. TRUE if the geometry has been changed. FALSE otherwise
       *
       * \return The valdiated result polygon
       */
      virtual te::gm::GeometryPtr validatePolygon(const te::gm::Polygon* polygon, bool& wasChanged) const;

      /*!
       * \brief Validates a line, an returns additional variable to inform if the geometry has been changed or not
       *
       * \param geometry The input line to be validate
       * \param wasChanged output param. TRUE if the geometry has been changed. FALSE otherwise
       *
       * \return The valdiated result line
       */
      virtual te::gm::GeometryPtr validateLine(const te::gm::LineString* lineString, bool& wasChanged) const = 0;

      /*!
       * \brief This function recreates the rings topology of a polygon, distributing the given inner rings among all the given outer rings.
       *
       * \param vecOuterRings A vector containing all the outer rings
       * \param vecInnerRings A vector containing all the inner rings
       *
       * \return The geometry with the ring topology fixed
       */
      te::gm::GeometryPtr mountTopology(const te::gm::GeometryVector& vecOuterRings, const te::gm::GeometryVector& vecInnerRings) const;

    protected:

      double m_precision; //!< When appliable, the precision to be used in the validation
    };

    //Typedef
    class AbstractValidatorPtr : public std::unique_ptr<AbstractValidator>
    {
    public:
      AbstractValidatorPtr(te::gm::AbstractValidator* validator = nullptr) : std::unique_ptr<te::gm::AbstractValidator>(validator) {};
    };

    class AbstractValidatorSharedPtr : public std::shared_ptr<AbstractValidator>
    {
    public:
      AbstractValidatorSharedPtr(te::gm::AbstractValidator* validator = nullptr) : std::shared_ptr<te::gm::AbstractValidator>(validator) {};
    };

  } // end namespace gm
}   // end namespace te

#endif  // __TERRALIB_GEOMETRY_INTERNAL_ABSTRACTVALIDATOR_H
