#include "AddProjectedCoordinates.h"

#include "../CommonDataStructures.h"
#include "../GeometryFunctions.h"
#include "../Geometry.h"
#include "../LineString.h"
#include "../Utils.h"

#include <algorithm>

te::gm::AddProjectedCoordinates::AddProjectedCoordinates()
{

}

te::gm::GeometryPtr te::gm::AddProjectedCoordinates::validate(const te::gm::Geometry* geometry, bool& wasChanged) const
{
  if (geometry == nullptr)
  {
    te::gm::GeometryPtr result;
    return result;
  }

  if (m_coordinateIndex == nullptr)
  {
    m_coordinateIndex.reset(new KdTreeSingle(te::gm::Envelope()));

    //we first create an index of all the points
    te::gm::SpatialIndex::populateCoordinateIndex(geometry, *m_coordinateIndex);
  }

  return te::gm::AbstractValidator::validate(geometry, wasChanged);
}

void te::gm::AddProjectedCoordinates::setCoordinateIndex(KdTreeSingle* coordinateIndex)
{
  m_coordinateIndex.reset(coordinateIndex);
}

te::gm::GeometryPtr te::gm::AddProjectedCoordinates::validateLine(const te::gm::LineString* lineString, bool& wasChanged) const
{
  bool currentWasChanged = false;
  std::vector<te::gm::Coord2D> vecLineCoordinates;
  for (std::size_t i = 1; i < lineString->size(); ++i)
  {
    te::gm::Coord2D c1(lineString->getX(i - 1), lineString->getY(i - 1));
    te::gm::Coord2D c2(lineString->getX(i), lineString->getY(i));

    if (c1 == c2)
    {
      vecLineCoordinates.push_back(c1);
      continue;
    }

    std::vector<te::gm::Coord2D> vecIntermediatedProjectedCoordinates = getIntermediateProjectedCoordinates(c1, c2, *m_coordinateIndex, m_precision);

    vecLineCoordinates.push_back(c1);

    if (vecIntermediatedProjectedCoordinates.empty() == false)
    {
      currentWasChanged = true;
    }

    vecLineCoordinates.insert(vecLineCoordinates.end(), vecIntermediatedProjectedCoordinates.begin(), vecIntermediatedProjectedCoordinates.end());
  }

  vecLineCoordinates.push_back(lineString->getEndPoint()->getCentroid());

  //we must clone the line string as output type should be keeped the same as the input type
  te::gm::LineString* outputLineString = static_cast<te::gm::LineString*>(lineString->clone());
  if (currentWasChanged == false)
  {
    return outputLineString;
  }

  wasChanged = true;

  outputLineString->setNumCoordinates(vecLineCoordinates.size());
  for (std::size_t i = 0; i < vecLineCoordinates.size(); ++i)
  {
    const te::gm::Coord2D& coord = vecLineCoordinates.at(i);
    outputLineString->setX(i, coord.x);
    outputLineString->setY(i, coord.y);
  }

  return outputLineString;
}

std::vector<te::gm::Coord2D> te::gm::AddProjectedCoordinates::getIntermediateProjectedCoordinates(const te::gm::Coord2D& c1, const te::gm::Coord2D& c2, const te::gm::KdTreeSingle& spatialIndex, double distance) const
{
  te::gm::Envelope envelope = te::gm::CreateEnvelope(c1.x, c1.y, c2.x, c2.y);
  //creates a buffer using the given distance
  envelope.m_llx -= distance;
  envelope.m_lly -= distance;
  envelope.m_urx += distance;
  envelope.m_ury += distance;

  //then searchs for coordinates that intersects the box
  std::vector<te::gm::KD_SINGLE_NODE*> vecReport;
  spatialIndex.search(envelope, vecReport);

  std::vector<te::gm::Coord2D> vecIntermediatedProjectedCoordinates;
  if (vecReport.empty() == true)
  {
    return vecIntermediatedProjectedCoordinates;
  }

  //if we have candidates, we must check
  //1 - if it is different from c1 and c2
  //2 - if the perpendicular distance is within the given 'distance'
  //3 - if the perpendicular projected coordinate is different from c1 and c2
  //4 - if the perpendicular projected coordinate lies within the segment represented by 'c1' and 'c2'

  for (std::size_t i = 0; i < vecReport.size(); ++i)
  {
    const te::gm::Coord2D& candidateCoord = vecReport.at(i)->getKey();

    //1 - if the are different from c1 and c2
    if (candidateCoord == c1 || candidateCoord == c2)
    {
      continue;
    }

    //2 - if  the perpendicular distance is within the given 'distance'
    te::gm::Coord2D projectedCoord;
    bool isOnSegment = false;
    double currentDistance = te::gm::Distance::PerpendicularDistance(c1, c2, candidateCoord, projectedCoord, isOnSegment);

    if (currentDistance > distance)
    {
      continue;
    }

    //3 - if the perpendicular projected coordinate is different from c1 and c2
    if (te::gm::Distance::IsEqual(projectedCoord, c1) == true)
    {
      continue;
    }
    if (te::gm::Distance::IsEqual(projectedCoord, c2) == true)
    {
      continue;
    }

    //4 - if the perpendicular projected coordinates lies within the segment represented by 'c1' and 'c2'
    if (isOnSegment == true)
    {
      vecIntermediatedProjectedCoordinates.push_back(projectedCoord);
    }
  }

  //now we must sort the vector based on the distance from each coordinate to c1
  if (vecIntermediatedProjectedCoordinates.size() > 1)
  {
    te::gm::CoordDistanceOrderFunctor functor(c1);
    std::sort(vecIntermediatedProjectedCoordinates.begin(), vecIntermediatedProjectedCoordinates.end(), functor);
  }

  return vecIntermediatedProjectedCoordinates;
}
