#include "RemoveCollapsedSegments.h"

#include "../../sam/rtree/Index.h"
#include "../Geometry.h"
#include "../GeometryFunctions.h"
#include "../LineString.h"
#include "../LinearRing.h"
#include "../MultiLineString.h"
#include "../Utils.h"

//Boost
#include <boost/bimap.hpp>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/connected_components.hpp>

te::gm::GeometryPtr te::gm::RemoveCollapsedSegments::validateLine(const te::gm::LineString* lineString, bool& wasChanged) const
{
  if(lineString == nullptr)
  {
    te::gm::GeometryPtr result;
    return result;
  }

  VecFragments fragments;
  bool hasCollpasedSegments = removeCollapsedSegmentsAndGetFragments(lineString, fragments);

  if (hasCollpasedSegments == false)
  {
    te::gm::GeometryPtr result(static_cast<te::gm::LineString*>(lineString->clone()));
    return result;
  }

  //check if the line has been completely collapsed
  if (fragments.empty() == true)
  {
    te::gm::GeometryPtr result;
    return result;
  }

  wasChanged = true;

  typedef
    boost::adjacency_list<
    boost::vecS           // edge list
    , boost::vecS         // vertex list
    , boost::directedS    // directedness
    , float               // property associated with vertices
    >
    Graph;

  Graph graph;
  std::map<te::gm::Coord2D, std::size_t> mapSegments;

  for (std::size_t i = 0; i < fragments.size(); ++i)
  {
    const VecCoords& vecCoords = fragments.at(i);
    std::size_t fragmentIndexVertex = boost::add_vertex(graph);

    //we first add the vertex to the graph
    const te::gm::Coord2D& startCoord = vecCoords.at(0);


    //then we check if the is any other segment ending on this vertex
    std::map<te::gm::Coord2D, std::size_t>::const_iterator itStart = mapSegments.find(startCoord);
    if (itStart != mapSegments.end())
    {
      std::size_t segmentToAssociate = itStart->second;
      boost::add_edge(segmentToAssociate, fragmentIndexVertex, graph);
    }
    else
    {
      mapSegments[startCoord] = fragmentIndexVertex;
    }

    //we first add the vertex to the graph
    const te::gm::Coord2D& endCoord = vecCoords.at(vecCoords.size() - 1);

    //then we check if the is any other segment ending on this vertex
    std::map<te::gm::Coord2D, std::size_t>::const_iterator itEnd = mapSegments.find(endCoord);
    if (itEnd != mapSegments.end())
    {
      std::size_t segmentToAssociate = itEnd->second;
      boost::add_edge(fragmentIndexVertex, segmentToAssociate, graph);
    }
    {
      mapSegments[endCoord] = fragmentIndexVertex;
    }
  }

  std::vector<int> component(fragments.size());
  size_t num_components = boost::connected_components(graph, &component[0]);

  std::vector<VecFragments> groupedFragments(num_components);
  for (std::size_t i = 0; i < fragments.size(); ++i)
  {
    groupedFragments[component.at(i)].push_back(fragments.at(i));
  }

  std::vector<te::gm::LineString*> vecLines;
  bool createRings = (lineString->isClosed() && dynamic_cast<const te::gm::LinearRing*>(lineString) != nullptr);
  for (std::size_t i = 0; i < groupedFragments.size(); ++i)
  {
    VecCoords fullLine = joinFragments(groupedFragments.at(i));

    std::unique_ptr<te::gm::LineString> newLine;
    if (createRings == true)
    {
      if (fullLine.size() < 4)
      {
        //the ring has collpased
        continue;
      }

      newLine.reset(new te::gm::LinearRing(fullLine.size(), te::gm::LineStringType, lineString->getSRID()));
    }
    else
    {
      if (fullLine.size() < 2)
      {
        //the line has collpased
        continue;
      }

      newLine.reset(new te::gm::LineString(fullLine.size(), te::gm::LineStringType, lineString->getSRID()));
    }

    for (std::size_t j = 0; j < fullLine.size(); ++j)
    {
      newLine->setX(j, fullLine.at(j).x);
      newLine->setY(j, fullLine.at(j).y);
    }

    vecLines.push_back(newLine.release());
  }

  //now we create the approprite return type
  te::gm::GeometryPtr result;
  if (vecLines.size() == 1)
  {
    result.reset(vecLines.at(0));
  }
  else if (vecLines.size() > 1)
  {
    te::gm::MultiLineString* multiLineString = new te::gm::MultiLineString(0, te::gm::MultiLineStringType, lineString->getSRID());
    for (std::size_t i = 0; i < vecLines.size(); ++i)
    {
      multiLineString->add(vecLines.at(i));
    }
    result.reset(multiLineString);
  }

  return result;
}

bool te::gm::RemoveCollapsedSegments::removeCollapsedSegmentsAndGetFragments(const te::gm::LineString* lineString, VecFragments& fragments) const
{
  //we need to remove the segments that have and opposite segment exactly equal to it
  //for it, the first step is to index all the start coordinates of each segment
  te::sam::rtree::Index<std::size_t> index;

  std::set<std::size_t> segmentsToRemove;
  for (std::size_t currentSegmentIndex = 0; currentSegmentIndex < lineString->size() - 1; ++currentSegmentIndex)
  {
    te::gm::Coord2D c1(lineString->getX(currentSegmentIndex), lineString->getY(currentSegmentIndex));
    te::gm::Coord2D c2(lineString->getX(currentSegmentIndex + 1), lineString->getY(currentSegmentIndex + 1));

    te::gm::Envelope envelope = te::gm::CreateEnvelope(c1.x, c1.y, c2.x, c2.y);

    //we first check if there is a, euqla opposite segment
    std::vector<std::size_t> vecReport;
    index.search(envelope, vecReport);

    bool addSegment = true;
    if (vecReport.size() >= 1)
    {
      for (std::size_t j = 0; j < vecReport.size(); ++j)
      {
        std::size_t candidateSegmentIndex = vecReport.at(j);
        te::gm::Coord2D candidateC1(lineString->getX(candidateSegmentIndex), lineString->getY(candidateSegmentIndex));
        te::gm::Coord2D candidateC2(lineString->getX(candidateSegmentIndex + 1), lineString->getY(candidateSegmentIndex + 1));

        //if we have a, equal segment in the opposite direction, we must remove both
        if (c1 == candidateC2 && c2 == candidateC1)
        {
          index.remove(envelope, currentSegmentIndex);
          index.remove(envelope, candidateSegmentIndex);

          segmentsToRemove.insert(currentSegmentIndex);
          segmentsToRemove.insert(candidateSegmentIndex);
          addSegment = false;
          break;
        }
      }
    }

    if (addSegment == true)
    {
      index.insert(envelope, currentSegmentIndex);
    }
  }

  if (segmentsToRemove.empty() == true)
  {
    return false;
  }

  //as we removed at least one segment, we must recreate the topology by merging the fragments
  VecCoords vecCoords;
  for (std::size_t currentSegmentIndex = 0; currentSegmentIndex < lineString->size() - 1; ++currentSegmentIndex)
  {
    if (segmentsToRemove.find(currentSegmentIndex) != segmentsToRemove.end())
    {
      if (vecCoords.empty() == false)
      {
        fragments.push_back(vecCoords);
        vecCoords = VecCoords();
      }

      continue;
    }

    if (vecCoords.empty() == true)
    {
      te::gm::Coord2D c1(lineString->getX(currentSegmentIndex), lineString->getY(currentSegmentIndex));
      te::gm::Coord2D c2(lineString->getX(currentSegmentIndex + 1), lineString->getY(currentSegmentIndex + 1));
      vecCoords.push_back(c1);
      vecCoords.push_back(c2);
    }
    else
    {
      te::gm::Coord2D c2(lineString->getX(currentSegmentIndex + 1), lineString->getY(currentSegmentIndex + 1));
      vecCoords.push_back(c2);
    }
  }

  if (vecCoords.empty() == false)
  {
    fragments.push_back(vecCoords);
  }

  return true;
}

te::gm::VecCoords te::gm::RemoveCollapsedSegments::joinFragments(const VecFragments& vecFragments) const
{
  VecCoords vecCoords;
  if (vecFragments.size() == 1)
  {
    //just create the line and return
    vecCoords = vecFragments.at(0);
    return vecCoords;
  }

  VecFragments vecFragmentsCopy(vecFragments);
  while (vecFragmentsCopy.empty() == false)
  {
    bool erased = false;
    if (vecCoords.empty() == true)
    {
      vecCoords = *vecFragmentsCopy.begin();
      vecFragmentsCopy.erase(vecFragmentsCopy.begin());
      erased = true;
    }
    else
    {
      for (std::size_t i = 0; i < vecFragmentsCopy.size(); ++i)
      {
        const VecCoords& otherFragment = vecFragmentsCopy.at(i);

        const te::gm::Coord2D& endCoord = vecCoords.at(vecCoords.size() - 1);
        const te::gm::Coord2D& currentStartCoord = otherFragment.at(0);

        if (te::gm::Distance::IsEqual(endCoord, currentStartCoord) == true)
        {
          vecCoords.insert(vecCoords.end(), otherFragment.begin() + 1, otherFragment.end());
          vecFragmentsCopy.erase(vecFragmentsCopy.begin() + i);
          erased = true;
          break;
        }
      }
    }
    //check for error
    if (erased == false)
    {
      return vecCoords;
    }
  }

  return vecCoords;
}