/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/maptools/DataSetLayer.cpp

  \brief A layer with reference to a dataset.
*/

// TerraLib
#include "../core/translator/Translator.h"
#include "../dataaccess/query/DataSetName.h"
#include "../dataaccess/query/Field.h"
#include "../dataaccess/query/Fields.h"
#include "../dataaccess/query/From.h"
#include "../dataaccess/query/FromItem.h"
#include "../dataaccess/query/Select.h"
#include "../dataaccess/query/SpatialQueryProcessor.h"
#include "../dataaccess/query/Where.h"
#include "../dataaccess/utils/Utils.h"
#include "../geometry/GeometryProperty.h"
#include "../raster/Grid.h"
#include "../raster/RasterProperty.h"
//#include "../se/Style.h"
#include "DataSetLayer.h"
//#include "Chart.h"
#include "Exception.h"
//#include "Grouping.h"
//#include "RasterContrast.h"
#include "RendererFactory.h"
#include "Utils.h"

// Boost
#include <boost/format.hpp>

// STL
#include <memory>

const std::string te::map::DataSetLayer::sm_type("DATASETLAYER");

te::map::DataSetLayer::DataSetLayer(AbstractLayer* parent)
  : AbstractLayer(parent),
    m_schema(nullptr)
{
}

te::map::DataSetLayer::DataSetLayer(const std::string& id, AbstractLayer* parent)
  : AbstractLayer(id, parent),
    m_schema(nullptr)
{
}

te::map::DataSetLayer::DataSetLayer(const std::string& id,
                                    const std::string& title,
                                    AbstractLayer* parent)
  : AbstractLayer(id, title, parent),
    m_schema(nullptr)
{
}

te::map::DataSetLayer::~DataSetLayer()
{
  delete m_schema;
}

te::map::AbstractLayer* te::map::DataSetLayer::clone()
{
  te::map::DataSetLayer* layer = new te::map::DataSetLayer;
  
  te::map::CopyAbstractLayerInfo(this, layer);

  layer->setRendererType(m_rendererType);

  return layer; 
}

void te::map::DataSetLayer::setSRID(int srid)
{
  te::map::AbstractLayer::setSRID(srid);
  this->setOutOfDate();
}

void te::map::DataSetLayer::loadSchema() const
{
  if(m_schema == nullptr)
  {
    te::da::DataSourcePtr ds = te::da::GetDataSource(m_datasourceId, true);
    
    std::unique_ptr<LayerSchema> type = ds->getDataSetType(m_datasetName);
    
    // Cache the schema from datasource
    if (type.get())
    {
      m_schema = static_cast<LayerSchema*>(type->clone());

      // propagate it to my cached dataset schema
      if (m_schema->hasGeom())
      {
        gm::GeometryProperty* myGeom = te::da::GetFirstGeomProperty(m_schema);
        myGeom->setSRID(m_srid);
      }
      else if (m_schema->hasRaster())
      {
        rst::RasterProperty* rstProp = te::da::GetFirstRasterProperty(m_schema);
        rstProp->getGrid()->setSRID(m_srid);
      }
    }
  }
}

std::unique_ptr<te::map::LayerSchema> te::map::DataSetLayer::getSchema() const
{
  if(m_schema == nullptr && !m_datasourceId.empty() && !m_datasetName.empty())
  {
    loadSchema();
  }

  std::unique_ptr<LayerSchema> schemaCloned;
  if (m_schema != nullptr)
  {
    schemaCloned.reset(static_cast<LayerSchema*>(m_schema->clone()));
  }

  return schemaCloned;
}

std::unique_ptr<te::da::DataSet> te::map::DataSetLayer::getData(te::common::TraverseType travType,
                                                              const te::common::AccessPolicy accessPolicy) const
{
  assert(!m_datasetName.empty());

  te::da::DataSourcePtr ds = te::da::GetDataSource(m_datasourceId, true);

  ds->setEncoding(this->getEncoding());

  return ds->getDataSet(m_datasetName, travType, accessPolicy);
}

std::unique_ptr<te::da::DataSet> te::map::DataSetLayer::getData(const std::string& propertyName,
                                                              const te::gm::Envelope* e,
                                                              te::gm::SpatialRelation r,
                                                              te::common::TraverseType travType,
                                                              const te::common::AccessPolicy accessPolicy) const
{
  assert(!m_datasetName.empty());

  te::da::DataSourcePtr ds = te::da::GetDataSource(m_datasourceId, true);

  ds->setEncoding(this->getEncoding());

  return ds->getDataSet(m_datasetName, propertyName, e, r, travType, accessPolicy);
}

std::unique_ptr<te::da::DataSet> te::map::DataSetLayer::getData(const std::string& propertyName,
                                                              const te::gm::Geometry* g,
                                                              te::gm::SpatialRelation r,
                                                              te::common::TraverseType travType,
                                                              const te::common::AccessPolicy accessPolicy) const
{
  assert(!m_datasetName.empty());

  te::da::DataSourcePtr ds = te::da::GetDataSource(m_datasourceId, true);

  ds->setEncoding(this->getEncoding());

  return ds->getDataSet(m_datasetName, propertyName, g, r, travType, accessPolicy);
}

std::unique_ptr<te::da::DataSet> te::map::DataSetLayer::getData(
    te::da::Expression* restriction, te::common::TraverseType travType,
    const te::common::AccessPolicy /*accessPolicy*/) const
{
  assert(restriction);
  assert(!m_datasetName.empty());

  te::da::DataSourcePtr ds = te::da::GetDataSource(m_datasourceId, true);

  ds->setEncoding(this->getEncoding());

  // Where clause
  te::da::Where* filter = new te::da::Where(restriction);
  
  // All fields (?)
  te::da::Fields* all = new te::da::Fields;
  all->push_back(new te::da::Field("*"));
  
  // From the data set
  te::da::FromItem* fromItem = new te::da::DataSetName(m_datasetName);
  te::da::From* from = new te::da::From;
  from->push_back(fromItem);

  // The final Select
  std::unique_ptr<te::da::Select> select(new te::da::Select(all, from, filter));

  te::da::SpatialQueryProcessor sqp;
  return sqp.getDataSet(ds, select.get(), travType);
}

std::unique_ptr<te::da::DataSet> te::map::DataSetLayer::getData(const te::da::ObjectIdSet* oids,
                                                              te::common::TraverseType travType,
                                                              const te::common::AccessPolicy accessPolicy) const
{
  assert(oids);

  te::da::DataSourcePtr ds = te::da::GetDataSource(m_datasourceId, true);

  ds->setEncoding(this->getEncoding());

  return ds->getDataSet(m_datasetName, oids, travType, accessPolicy);
}

bool te::map::DataSetLayer::isValid() const
{
  if (m_datasourceId.empty() || m_datasetName.empty())
  {
    m_isValid = false;
    return false;
  }

  te::da::DataSourcePtr ds;
  try
  {
    ds = te::da::GetDataSource(m_datasourceId, true);
    
    //ensures that the schema is also valid
    this->getSchema();
  }
  catch(...)
  {
    m_isValid = false;
    return false;
  }

  if (ds.get() == nullptr || !ds->isValid() || !ds->isOpened() || m_schema == nullptr)
  {
    m_isValid = false;
    return false;
  }

  m_isValid = true;

  return true;
}

void te::map::DataSetLayer::draw(Canvas* canvas, const te::gm::Envelope& bbox, int srid, const double& scale, bool* cancel)
{
  if(m_rendererType.empty())
    throw Exception((boost::format(TE_TR("Could not draw the data set layer %1%. The renderer type is empty!")) % getTitle()).str());

  // Try get the defined renderer
  std::unique_ptr<AbstractRenderer> renderer(RendererFactory::make(m_rendererType));
  if(renderer.get() == nullptr)
    throw Exception((boost::format(TE_TR("Could not draw the data set layer %1%. The renderer %2% could not be created!")) % getTitle() % m_rendererType).str());

  renderer->draw(this, canvas, bbox, srid, scale, cancel);
}

const std::string& te::map::DataSetLayer::getType() const
{
  return sm_type;
}

const std::string& te::map::DataSetLayer::getRendererType() const
{
  return m_rendererType;
}

void te::map::DataSetLayer::setRendererType(const std::string& t)
{
  m_rendererType = t;
}

void te::map::DataSetLayer::setOutOfDate()
{
  delete m_schema;
  m_schema = nullptr;
}
