/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/maptools/QueryEncoder.cpp
  
  \brief A visitor that converts a OGC Filter Expression to TerraLib Expression.
 */

// TerraLib
#include "FilterEncoder.h"

#include "../dataaccess/query_h.h" 
#include "../datatype.h"
#include "../fe.h"
#include "../gml/Envelope.h"


std::map<std::string, std::string> te::map::FilterEncoder::sm_fnameMap;

te::map::FilterEncoder::FilterEncoder()
  : m_expression(nullptr)
  , m_op(nullptr)
{
}

te::map::FilterEncoder::~FilterEncoder() = default;

te::fe::Filter* te::map::FilterEncoder::getFilter(const te::da::Expression* exp)
{
  exp->accept(*this);

  te::fe::Filter* filter = new te::fe::Filter();
  filter->setOp(m_op);

  return filter;
}

void te::map::FilterEncoder::visit(const te::da::Expression& /*visited*/)
{

}

void te::map::FilterEncoder::visit(const te::da::DataSetName& /*visited*/)
{

}

void te::map::FilterEncoder::visit(const te::da::FromItem& /*visited*/)
{

}

void te::map::FilterEncoder::visit(const te::da::Function& visited)
{
  if (visited.getName() == te::da::FunctionNames::sm_EqualTo ||
      visited.getName() == te::da::FunctionNames::sm_NotEqualTo ||
      visited.getName() == te::da::FunctionNames::sm_GreaterThan ||
      visited.getName() == te::da::FunctionNames::sm_GreaterThanOrEqualTo ||
      visited.getName() == te::da::FunctionNames::sm_LessThan ||
      visited.getName() == te::da::FunctionNames::sm_LessThanOrEqualTo)
  {
    te::fe::BinaryComparisonOp* bcop = new te::fe::BinaryComparisonOp(sm_fnameMap[visited.getName()].c_str());

    visited.getArg(0)->accept(*this);
    bcop->setFirst(m_expression);
    visited.getArg(1)->accept(*this);
    bcop->setSecond(m_expression);

    m_op = bcop;
  }
  else if (visited.getName() == te::da::FunctionNames::sm_And ||
           visited.getName() == te::da::FunctionNames::sm_Or)
  {
    te::fe::BinaryLogicOp* blop = new te::fe::BinaryLogicOp(sm_fnameMap[visited.getName()].c_str());

    for (std::size_t t = 0; t < visited.getNumArgs(); ++t)
    {
      visited.getArg(t)->accept(*this);
      blop->add(m_op);
    }

    m_op = blop;
  }
}

void te::map::FilterEncoder::visit(const te::da::Insert& /*visited*/)
{

}

void te::map::FilterEncoder::visit(const te::da::Join& /*visited*/)
{

}

void te::map::FilterEncoder::visit(const te::da::JoinCondition& /*visited*/)
{

}

void te::map::FilterEncoder::visit(const te::da::JoinConditionOn& /*visited*/)
{

}

void te::map::FilterEncoder::visit(const te::da::JoinConditionUsing& /*visited*/)
{

}

void te::map::FilterEncoder::visit(const te::da::Like& visited)
{
  te::fe::PropertyIsLike* comp = new te::fe::PropertyIsLike();
  visited.getArg(0)->accept(*this);
  
  te::fe::PropertyName* propertyName = dynamic_cast<te::fe::PropertyName*>(m_expression);
  if (propertyName != nullptr)
  {
    comp->setPropertyName(propertyName);
  }
  
  visited.getArg(1)->accept(*this);
  te::fe::Literal* literal = dynamic_cast<te::fe::Literal*>(m_expression);
  if (literal != nullptr)
  {
    comp->setLiteral(literal);
  }

  comp->setSingleChar(visited.getSingleChar());
  comp->setWildCard(visited.getWildCard());
  comp->setEscapeChar(visited.getEscapeChar());

  m_op = comp;
}

void te::map::FilterEncoder::visit(const te::da::Literal& visited)
{
  te::fe::Literal* literal = new te::fe::Literal(visited.getValue()->toString());

  m_expression = literal;
}

void te::map::FilterEncoder::visit(const te::da::LiteralBool& visited)
{
  te::fe::Literal* literal = new te::fe::Literal(visited.getValue()->toString());

  m_expression = literal;
}

void te::map::FilterEncoder::visit(const te::da::LiteralByteArray& visited)
{
  te::fe::Literal* literal = new te::fe::Literal(visited.getValue()->toString());

  m_expression = literal;
}

void te::map::FilterEncoder::visit(const te::da::LiteralDateTime& visited)
{
  te::fe::Literal* literal = new te::fe::Literal(visited.getValue()->toString());

  m_expression = literal;
}

void te::map::FilterEncoder::visit(const te::da::LiteralDouble& visited)
{
  te::fe::Literal* literal = new te::fe::Literal(visited.getValue()->toString());

  m_expression = literal;
}

void te::map::FilterEncoder::visit(const te::da::LiteralEnvelope& /*visited*/)
{

}

void te::map::FilterEncoder::visit(const te::da::LiteralGeom& visited)
{
  te::fe::Literal* literal = new te::fe::Literal(visited.getValue()->toString());

  m_expression = literal;
}

void te::map::FilterEncoder::visit(const te::da::LiteralInt16& visited)
{
  te::fe::Literal* literal = new te::fe::Literal(visited.getValue()->toString());

  m_expression = literal;
}

void te::map::FilterEncoder::visit(const te::da::LiteralInt32& visited)
{
  te::fe::Literal* literal = new te::fe::Literal(visited.getValue()->toString());

  m_expression = literal;
}

void te::map::FilterEncoder::visit(const te::da::LiteralInt64& visited)
{
  te::fe::Literal* literal = new te::fe::Literal(visited.getValue()->toString());

  m_expression = literal;
}

void te::map::FilterEncoder::visit(const te::da::LiteralString& visited)
{
  te::fe::Literal* literal = new te::fe::Literal(visited.getValue()->toString());

  m_expression = literal;
}

void te::map::FilterEncoder::visit(const te::da::PropertyName& visited)
{
  te::fe::PropertyName* pName = new te::fe::PropertyName(visited.getName());

  m_expression = pName;
}

void te::map::FilterEncoder::visit(const te::da::Query& /*visited*/)
{

}

void te::map::FilterEncoder::visit(const te::da::Select& /*visited*/)
{

}

void te::map::FilterEncoder::visit(const te::da::SelectExpression& /*visited*/)
{

}

void te::map::FilterEncoder::visit(const te::da::SubSelect& /*visited*/)
{

}

void te::map::FilterEncoder::visit(const te::da::In& visited)
{
  te::da::PropertyName* propertyName = visited.getPropertyName();

  for (std::size_t i = 0; i < visited.getNumArgs(); ++i)
  {
    te::fe::BinaryComparisonOp* equalComparissonOperator = new te::fe::BinaryComparisonOp(te::fe::Globals::sm_propertyIsEqualTo);

    visited.getPropertyName()->accept(*this);
    equalComparissonOperator->setFirst(m_expression);
    visited.getArg(i)->accept(*this);
    equalComparissonOperator->setSecond(m_expression);
    
    if (i > 0)
    {
      te::fe::BinaryLogicOp* orLogicOperator = new te::fe::BinaryLogicOp(te::fe::Globals::sm_or);
      orLogicOperator->add(m_op);
      orLogicOperator->add(equalComparissonOperator);      

      m_op = orLogicOperator;
    }
    else
    {
      m_op = equalComparissonOperator;
    }
  }
}

void te::map::FilterEncoder::visit(const te::da::Cast& /*visited*/)
{

}

void te::map::FilterEncoder::initialize()
{
  // BinaryComparisonOp
  sm_fnameMap[te::da::FunctionNames::sm_EqualTo] = te::fe::Globals::sm_propertyIsEqualTo;
  sm_fnameMap[te::da::FunctionNames::sm_NotEqualTo] = te::fe::Globals::sm_propertyIsNotEqualTo;
  sm_fnameMap[te::da::FunctionNames::sm_GreaterThan] = te::fe::Globals::sm_propertyIsGreaterThan;
  sm_fnameMap[te::da::FunctionNames::sm_GreaterThanOrEqualTo] = te::fe::Globals::sm_propertyIsGreaterThanOrEqualTo;
  sm_fnameMap[te::da::FunctionNames::sm_LessThan] = te::fe::Globals::sm_propertyIsLessThan;
  sm_fnameMap[te::da::FunctionNames::sm_LessThanOrEqualTo] = te::fe::Globals::sm_propertyIsLessThanOrEqualTo;

  // BinaryLogicOp
  sm_fnameMap[te::da::FunctionNames::sm_And] = te::fe::Globals::sm_and;
  sm_fnameMap[te::da::FunctionNames::sm_Or] = te::fe::Globals::sm_or;

  // BinarySpatialOp
  sm_fnameMap[te::da::FunctionNames::sm_ST_Crosses] = te::fe::Globals::sm_crosses;
  sm_fnameMap[te::da::FunctionNames::sm_ST_Disjoint] = te::fe::Globals::sm_disjoint;
  sm_fnameMap[te::da::FunctionNames::sm_ST_Equals] = te::fe::Globals::sm_equals;
  sm_fnameMap[te::da::FunctionNames::sm_ST_Intersects] = te::fe::Globals::sm_intersects;
  sm_fnameMap[te::da::FunctionNames::sm_ST_Overlaps] = te::fe::Globals::sm_overlaps;
  sm_fnameMap[te::da::FunctionNames::sm_ST_Touches] = te::fe::Globals::sm_touches;
  sm_fnameMap[te::da::FunctionNames::sm_ST_Within] = te::fe::Globals::sm_within;

  // DistanceBuffer
  sm_fnameMap[te::da::FunctionNames::sm_ST_DWithin] = te::fe::Globals::sm_dWithin;
  sm_fnameMap[te::da::FunctionNames::sm_ST_Beyond] = te::fe::Globals::sm_beyond;

  // UnaryLogicOp
  sm_fnameMap[te::da::FunctionNames::sm_Not] = te::fe::Globals::sm_not;

  // BinaryOperator
  sm_fnameMap[te::da::FunctionNames::sm_Add] = te::fe::Globals::sm_add;
  sm_fnameMap[te::da::FunctionNames::sm_Sub] = te::fe::Globals::sm_sub;
  sm_fnameMap[te::da::FunctionNames::sm_Mul] = te::fe::Globals::sm_mul;
  sm_fnameMap[te::da::FunctionNames::sm_Div] = te::fe::Globals::sm_div;
}
