/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/maptools/FilterEncoder.h
  
  \brief A visitor that converts a TerraLib Expression to OGC Filter Expression.
 */

#ifndef __TERRALIB_MAPTOOLS_INTERNAL_FILTERENCODER_H
#define __TERRALIB_MAPTOOLS_INTERNAL_FILTERENCODER_H

// TerraLib
#include "Config.h"
#include "../dataaccess/query/QueryVisitor.h"

// STL
#include <map>
#include <string>

namespace te
{
// Forward declarations
  namespace fe
  {
    class AbstractOp;
    class Expression;
    class Filter;
  }

  namespace map
  {
    /*!
      \class FilterEncoder
      
      \brief A visitor that converts a TerraLib Expression to OGC Filter Expression.

      \sa Visitor
    */
    class TEMAPEXPORT FilterEncoder : public te::da::QueryVisitor
    {
      public:

        /** @name Initializer Methods
         *  Methods related to instantiation and destruction of a Filter visitor.
         */
        //@{

        /*! \brief Default constructor. */
        FilterEncoder();

        /*! \brief Virtual destructor. */
        ~FilterEncoder();

        //@}

        /** @name Conversion Methods
         *  Methods that can be used to convert a OGC Filter Expression to a TerraLib Expression.
         */
        //@{

        /*!
          \brief It converts the TerraLib Expression to a OGC Filter Expression.

          \param exp A valid Terralib expression.

          \return An ogc filter for the given expression.

          \note The caller of this method will take the ownership of the returned expression.
        */
        te::fe::Filter* getFilter(const te::da::Expression* exp);

        //@}

        /** @name Visitor Methods
         *  All concrete visitors must implement these methods.
         */
        //@{

        void visit(const te::da::Expression& visited);
        void visit(const te::da::DataSetName& visited);
        void visit(const te::da::FromItem& visited);
        void visit(const te::da::Function& visited);
        void visit(const te::da::Insert& visited);
        void visit(const te::da::Join& visited);
        void visit(const te::da::JoinCondition& visited);
        void visit(const te::da::JoinConditionOn& visited);
        void visit(const te::da::JoinConditionUsing& visited);
        void visit(const te::da::Like& visited) override;
        void visit(const te::da::Literal& visited);
        void visit(const te::da::LiteralBool& visited);
        void visit(const te::da::LiteralByteArray& visited);
        void visit(const te::da::LiteralDateTime& visited);
        void visit(const te::da::LiteralDouble& visited);
        void visit(const te::da::LiteralEnvelope& visited);
        void visit(const te::da::LiteralGeom& visited);
        void visit(const te::da::LiteralInt16& visited);
        void visit(const te::da::LiteralInt32& visited);
        void visit(const te::da::LiteralInt64& visited);
        void visit(const te::da::LiteralString& visited);
        void visit(const te::da::PropertyName& visited);
        void visit(const te::da::Query& visited);
        void visit(const te::da::Select& visited);
        void visit(const te::da::SelectExpression& visited);
        void visit(const te::da::SubSelect& visited);
        void visit(const te::da::In& visited);
        void visit(const te::da::Cast& visited);

        //@}

      private:

        /*! \brief Static method to initialize the FilterEncoder. */
        static void initialize();

      private:

        /** @name Not Allowed Methods
         *  No copy allowed. 
         */
        //@{

        /*!
          \brief No copy constructor allowed.

          \param rhs The other object.
        */
        FilterEncoder(const FilterEncoder& rhs);

        /*!
          \brief No assignment operator allowed.

          \param rhs The other object.

          \return A reference for this.
        */
        FilterEncoder& operator=(const FilterEncoder& rhs);

        //@}

      private:

        te::fe::Expression* m_expression;                       //!< Expression used during conversion.
        te::fe::AbstractOp* m_op;                               //!< Operation used during conversion.

        static std::map<std::string, std::string> sm_fnameMap;  //!< A map that associates Filter Operator names to Query Functions names.

        friend class Module;
    };

  } // end namespace map
}   // end namespace te

#endif  // __TERRALIB_MAPTOOLS_INTERNAL_FILTERENCODER_H
