/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
\file RasterContrast.cpp

\brief This class contains the parameters needed to apply dynamic contrast over a raster.
*/

// TerraLib
#include "../common/STLUtils.h"
#include "RasterContrast.h"


te::map::RasterContrast::RasterContrast(const te::map::RasterTransform::ContrastType& type, const std::size_t& nBands)
  : m_type(type),
   m_nBands(nBands)
{
  m_gain.resize(m_nBands, 1.);
  m_offset1.resize(m_nBands, 0.);
  m_offset2.resize(m_nBands, 0.);
  m_min.resize(m_nBands, 0.);
  m_max.resize(m_nBands, 0.);
}

te::map::RasterContrast::RasterContrast(const RasterContrast& rhs)
  : m_type((te::map::RasterTransform::ContrastType)rhs.m_type),
  m_nBands(rhs.m_nBands)
{
  m_gain.reserve(rhs.m_nBands);
  std::copy(rhs.m_gain.begin(), rhs.m_gain.end(), std::back_inserter(m_gain));

  m_offset1.reserve(rhs.m_nBands);
  std::copy(rhs.m_offset1.begin(), rhs.m_offset1.end(), std::back_inserter(m_offset1));

  m_offset2.reserve(rhs.m_nBands);
  std::copy(rhs.m_offset2.begin(), rhs.m_offset2.end(), std::back_inserter(m_offset2));

  m_min.reserve(rhs.m_nBands);
  std::copy(rhs.m_min.begin(), rhs.m_min.end(), std::back_inserter(m_min));

  m_max.reserve(rhs.m_nBands);
  std::copy(rhs.m_max.begin(), rhs.m_max.end(), std::back_inserter(m_max));
}

te::map::RasterContrast::~RasterContrast()
{
  m_gain.clear();
  m_offset1.clear();
  m_offset2.clear();
  m_min.clear();
  m_max.clear();
}

const te::map::RasterTransform::ContrastType te::map::RasterContrast::getType() const
{
  return m_type;
}

std::size_t te::map::RasterContrast::getNBands()
{
  return m_nBands;
}

void te::map::RasterContrast::setValues(const double& gain, const double& offset1, const double& offset2, const double& min, const double& max, const std::size_t& band)
{
  if (band >= m_nBands)
    throw;

  m_gain[band] = gain;
  m_offset1[band] = offset1;
  m_offset2[band] = offset2;
  m_min[band] = min;
  m_max[band] = max;
}

void te::map::RasterContrast::getValues(std::vector<double>& gain, std::vector<double>& offset1, std::vector<double>& offset2, std::vector<double>& min, std::vector<double>& max)
{
  gain = m_gain;
  offset1 = m_offset1;
  offset2 = m_offset2;
  min = m_min;
  max = m_max;
}
