/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file RasterContrast.h

  \brief This class contains the parameters needed to apply dynamic contrast over a raster.
*/

#ifndef __TERRALIB_MAPTOOLS_INTERNAL_RASTERCONTRAST_H
#define __TERRALIB_MAPTOOLS_INTERNAL_RASTERCONTRAST_H

// TerraLib
#include "Config.h"
#include "Enums.h"
#include "RasterTransform.h"

// STL
#include <cstdlib>
#include <string>
#include <vector>

namespace te
{
  namespace map
  {
    /*!
      \class RasterContrast

      \brief This class contains the parameters needed to apply dynamic contrast over a raster.

      \ingroup map

      \sa RasterTransform
    */
    class TEMAPEXPORT RasterContrast
    {
      public:

        /*!
          \brief It constructs a new RasterContrast instance.

          \param type          The contrast type.
        */
        RasterContrast(const te::map::RasterTransform::ContrastType& type, const std::size_t& nBands);

        /*! \brief Destructor. */
        ~RasterContrast();

        /*! \brief Copy constructor. */
        RasterContrast(const RasterContrast& rhs);

        /*!
          \brief It gets the contrast type.

          \return The contrast type.
        */
        const te::map::RasterTransform::ContrastType getType() const;

        /*!
          \brief It gets the contrast number of bands.

          \return  The number of bands
        */
        std::size_t getNBands();

        /*!
          \brief It set the transformation values for contranst of one band.

          \param gain     The gain value of transformation.
          \param offset1  The offset1 value of transformation.
          \param offset2  The offset2 value of transformation.
          \param min      The minimun band value for this transformation.
          \param max      The maximum band value for this transformation.
          \param band     The band index.
        */
        void setValues(const double& gain, const double& offset1, const double& offset2, const double& min, const double& max, const std::size_t& band);

        /*!
        \brief It get the contrast transformation values.

        \param gain     The vector with gain values of transformation.
        \param offset1  The vector with offset1 values of transformation.
        \param offset2  The vector with offset2 values of transformation.
        \param min      The vector with minimun band values for this transformation.
        \param max      The vector with maximum band values for this transformation.
        */
        void getValues(std::vector<double>& gain, std::vector<double>& offset1, std::vector<double>& offset2, std::vector<double>& min, std::vector<double>& max);


      private:

        te::map::RasterTransform::ContrastType m_type;                         //!< The contrast type.

        std::size_t m_nBands;

        std::vector<double> m_gain;
        std::vector<double> m_offset1;
        std::vector<double> m_offset2;
        std::vector<double> m_min;
        std::vector<double> m_max;
    };

  } // end namespace map
}   // end namespace te

#endif  // __TERRALIB_MAPTOOLS_INTERNAL_RASTERCONTRAST_H

