/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - am_arrayBuffer Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file RasterTransform.h

  \brief A Raster Transform is a class that defines functions to transform a styled raster.
*/

// TerraLib
#include "RasterTransform.h"

#include "../raster/RasterProperty.h"
#include "../se/RasterSymbolizer.h"


// Boost
#include <boost/math/special_functions/fpclassify.hpp>

// STL
#include <limits>
#include <algorithm>

te::map::RasterTransform::RasterTransform(te::rst::Raster* input, te::rst::Raster* output) :
  m_rasterIn(input),
  m_rasterOut(output),
  m_inputNBands(input->getNumberOfBands()),
  m_monoBand(-1),
  m_redBand(-1),
  m_greenBand(-1),
  m_blueBand(-1),
  m_alphaBand(-1),
  m_transp(TE_OPAQUE),
  m_currentContrastType(LINEAR_CONTRAST),
  m_currentChannelDomain( REAL_CHANNEL_DOMAIN ),
  m_channelDomainGetFuncPtr( &RasterTransform::getRealValueChannelDomainGetFunction ),
  m_contrastTransfFuncPtr(&RasterTransform::applyLinearContrast),
  m_transfFuncPtr(&RasterTransform::setExtractRGB),
  m_RGBAFuncPtr(&RasterTransform::getExtractRGB),
  m_useBufferCache(false),
  m_arrayBuffer( nullptr )
{
  m_transformBandInfo = new TransformBandInfo*[m_inputNBands];

  for (std::size_t b = 0; b < m_inputNBands; ++b)
  {
    te::rst::Band* band = m_rasterIn->getBand(b);

    TransformBandInfo* tbi = new TransformBandInfo();
    tbi->m_noDataValue = input->getBand(b)->getProperty()->m_noDataValue;

    m_transformBandInfo[b] = tbi;
  }
}

te::map::RasterTransform::~RasterTransform()
{
  for (std::size_t t = 0; t < m_inputNBands; ++t)
  {
    if( m_transformBandInfo ) delete m_transformBandInfo[t];
    if( m_arrayBuffer ) delete m_arrayBuffer[t];
  }

  delete[] m_transformBandInfo;
  delete[] m_arrayBuffer;
}

void te::map::RasterTransform::setGain(double value, int band)
{
  m_transformBandInfo[band]->m_contrastGain = value;
}

std::vector< double > te::map::RasterTransform::getGain() const
{
  std::vector< double > gain;

  for (std::size_t t = 0; t < m_inputNBands; ++t)
  {
    gain.push_back(m_transformBandInfo[t]->m_contrastGain);
  }

  return gain;
}

void te::map::RasterTransform::setOffset(double value, int band)
{
  m_transformBandInfo[band]->m_contrastOffset1 = value;
}

std::vector< double > te::map::RasterTransform::getOffset()
{
  std::vector< double > offset;

  for (std::size_t t = 0; t < m_inputNBands; ++t)
  {
    offset.push_back(m_transformBandInfo[t]->m_contrastOffset1);
  }

  return offset;
}

void te::map::RasterTransform::setContrastType( const ContrastType& newType )
{
  m_currentContrastType = newType;
  setContrastFuncPtr( m_currentContrastType );
}

void te::map::RasterTransform::setGammaR(double value)
{
  assert(m_inputNBands > m_redBand);
  m_transformBandInfo[m_redBand]->m_gamma = value;
}

double te::map::RasterTransform::getGammaR()
{
  assert(m_inputNBands > m_redBand);
  return m_transformBandInfo[m_redBand]->m_gamma;
}

void te::map::RasterTransform::setGammaG(double value)
{
  assert(m_inputNBands > m_greenBand);
  m_transformBandInfo[m_greenBand]->m_gamma = value;
}

double te::map::RasterTransform::getGammaG()
{
  assert(m_inputNBands > m_greenBand);
  return m_transformBandInfo[m_greenBand]->m_gamma;
}

void te::map::RasterTransform::setGammaB(double value)
{
  assert(m_inputNBands > m_blueBand);
  m_transformBandInfo[m_blueBand]->m_gamma = value;
}

double te::map::RasterTransform::getGammaB()
{
  assert(m_inputNBands > m_blueBand);
  return m_transformBandInfo[m_blueBand]->m_gamma;
}

void te::map::RasterTransform::setGammaM(double value)
{
  assert(m_inputNBands > m_monoBand );
  m_transformBandInfo[m_monoBand]->m_gamma = value;
}

double te::map::RasterTransform::getGammaM()
{
  assert(m_inputNBands > m_monoBand);
  return m_transformBandInfo[m_monoBand]->m_gamma;
}

void te::map::RasterTransform::setRGBMap(std::map<RGBChannels, short>& rgbMap)
{
  clearRGBMap();

  std::map<RGBChannels,short>::iterator it = rgbMap.begin();
  while (it != rgbMap.end())
  {
    if (it->first == RED_CHANNEL)
      m_redBand = it->second;

    else if (it->first == GREEN_CHANNEL)
      m_greenBand = it->second;

    else if (it->first == BLUE_CHANNEL)
      m_blueBand = it->second;

    else if (it->first == ALPHA_CHANNEL)
      m_alphaBand = it->second;

    else if (it->first == MONO_CHANNEL)
      m_monoBand = it->second;

    ++it;
  }
}

void te::map::RasterTransform::setBChannelMapping(short bIn, te::map::RasterTransform::RGBChannels bOut)
{
  if (bOut == RED_CHANNEL)
    m_redBand = bIn;

  else if (bOut == GREEN_CHANNEL)
    m_greenBand = bIn;

  else if (bOut == BLUE_CHANNEL)
    m_blueBand = bIn;

  else if (bOut == ALPHA_CHANNEL)
    m_alphaBand = bIn;

  else if (bOut == MONO_CHANNEL)
    m_monoBand = bIn;
}

short te::map::RasterTransform::getBChannelMapping(RGBChannels channel)
{
  if (channel == RED_CHANNEL)
    return m_redBand;

  else if (channel == GREEN_CHANNEL)
    return m_greenBand;

  else if (channel == BLUE_CHANNEL)
    return m_blueBand;

  else if (channel == ALPHA_CHANNEL)
    return m_alphaBand;

  else if (channel == MONO_CHANNEL)
    return m_monoBand;

  return -1;
}

void te::map::RasterTransform::clearRGBMap()
{
  m_monoBand = -1;
  m_redBand = -1;
  m_greenBand = -1;
  m_blueBand = -1;
  m_alphaBand = -1;
}

std::map<te::map::RasterTransform::RGBChannels, short> te::map::RasterTransform::getRGBMap()
{
  std::map<te::map::RasterTransform::RGBChannels, short> rgbChannels;

  rgbChannels[RED_CHANNEL] = m_redBand;
  rgbChannels[GREEN_CHANNEL] = m_greenBand;
  rgbChannels[BLUE_CHANNEL] = m_blueBand;
  rgbChannels[ALPHA_CHANNEL] = m_alphaBand;
  rgbChannels[MONO_CHANNEL] = m_monoBand;

  return rgbChannels;
}

void te::map::RasterTransform::setLinearTransfParameters(double vmin, double vmax, double rmin, double rmax)
{
  double gain = 1.0;
  double offset1 = 0.0;
  double offset2 = 0.0;

  if(vmax != vmin)
  {
    const double outRange = rmax - rmin;
    const double inRange = vmax - vmin;

    gain = outRange / inRange;

    if(!(boost::math::isnan(gain) || boost::math::isinf(gain)))
    {
      offset1 = (-1.0) * vmin;
      offset2 = rmin;
    }
  }

  for (std::size_t b = 0; b < m_inputNBands; ++b)
  {
    TransformBandInfo* tbi = m_transformBandInfo[b];

    tbi->m_rstMinValue = rmin;
    tbi->m_rstMaxValue = rmax;
    tbi->m_contrastGain = gain;
    tbi->m_contrastOffset1 = offset1;
    tbi->m_contrastOffset2 = offset2;
  }
}

void te::map::RasterTransform::setLinearTransfParameters(double vmin, double vmax, double rmin, double rmax, int band)
{
  double gain = 1.0;
  double offset1 = 0.0;
  double offset2 = 0.0;

  if (vmax != vmin)
  {
    const double outRange = rmax - rmin;
    const double inRange = vmax - vmin;

    gain = outRange / inRange;

    if (!(boost::math::isnan(gain) || boost::math::isinf(gain)))
    {
      offset1 = (-1.0) * vmin;
      offset2 = rmin;
    }
  }

  TransformBandInfo* tbi = m_transformBandInfo[band];

  tbi->m_rstMinValue = rmin;
  tbi->m_rstMaxValue = rmax;
  tbi->m_contrastGain = gain;
  tbi->m_contrastOffset1 = offset1;
  tbi->m_contrastOffset2 = offset2;
}

void te::map::RasterTransform::setTransfParameters(double rmin, double rmax, double gain, double offset1, double offset2, int band)
{
  TransformBandInfo* tbi = m_transformBandInfo[band];

  tbi->m_rstMinValue = rmin;
  tbi->m_rstMaxValue = rmax;
  tbi->m_contrastGain = gain;
  tbi->m_contrastOffset1 = offset1;
  tbi->m_contrastOffset2 = offset2;
}

void te::map::RasterTransform::setInputRasterResampleSize(int inputX0, int inputY0, int inputWidth, int inputHeight, int outWidth, int outHeight)
{
  for (std::size_t t = 0; t < m_inputNBands; ++t)
  {
    if( m_arrayBuffer ) delete m_arrayBuffer[t];
  }
  delete[] m_arrayBuffer;

  m_arrayBuffer = new BandBufferInfo*[m_rasterIn->getNumberOfBands()];

  void* buffer = nullptr;

  for (std::size_t b = 0; b < m_rasterIn->getNumberOfBands(); ++b)
  {
    te::rst::Band* band = m_rasterIn->getBand(b);

    int pxlsize = te::rst::GetPixelSize(band->getProperty()->getType());
    int blksize = outWidth * outHeight * pxlsize;

    if(
        ( b == m_monoBand ) || ( b == m_redBand ) || ( b == m_greenBand ) ||
        ( b == m_blueBand ) || ( b == m_alphaBand )
      )
    {
      buffer = new unsigned char[blksize];
      band->getBlockBuffer(buffer, inputX0, inputY0, inputWidth, inputHeight, outWidth, outHeight, band->getProperty()->getType());
    }
    else
    {
      buffer = nullptr;
    }

    BandBufferInfo* bbi = new BandBufferInfo(band->getProperty()->getType(), 
      outWidth, outHeight, buffer, m_currentChannelDomain );

    m_arrayBuffer[b] = bbi;
  }

  m_useBufferCache = true;
}

te::map::RasterTransform::RasterTransfFunctions te::map::RasterTransform::getTransfFunction()
{
  if (m_transfFuncPtr == nullptr || m_RGBAFuncPtr)
  {
    return NO_TRANSF;
  }
  else  if (m_transfFuncPtr == &RasterTransform::setMono2ThreeBand &&
            m_RGBAFuncPtr == &RasterTransform::getMono2ThreeBand)
  {
    return MONO2THREE_TRANSF;
  }
  else if (m_transfFuncPtr == &RasterTransform::setExtractRGB &&
           m_RGBAFuncPtr == &RasterTransform::getExtractRGB)
  {
    return EXTRACT2RGB_TRANSF;
  }
  else  if (m_transfFuncPtr == &RasterTransform::setRed2ThreeBand &&
            m_RGBAFuncPtr == &RasterTransform::getRed2ThreeBand)
  {
    return RED2THREE_TRANSF;
  }
  else  if (m_transfFuncPtr == &RasterTransform::setGreen2ThreeBand &&
            m_RGBAFuncPtr == &RasterTransform::getGreen2ThreeBand)
  {
    return GREEN2THREE_TRANSF;
  }
  else  if (m_transfFuncPtr == &RasterTransform::setBlue2ThreeBand &&
            m_RGBAFuncPtr == &RasterTransform::getBlue2ThreeBand)
  {
    return BLUE2THREE_TRANSF;
  }
  else  if (m_transfFuncPtr == &RasterTransform::setCategorize &&
            m_RGBAFuncPtr == &RasterTransform::getCategorize)
  {
    return CATEGORIZE_TRANSF;
  }
  else  if (m_transfFuncPtr == &RasterTransform::setInterpolate &&
            m_RGBAFuncPtr == &RasterTransform::getInterpolate)
  {
    return INTERPOLATE_TRANSF;
  }
  else  if (m_transfFuncPtr == &RasterTransform::setRecode &&
    m_RGBAFuncPtr == &RasterTransform::getRecode)
  {
    return RECODE_TRANSF;
  }
  else  if (m_transfFuncPtr == &RasterTransform::setBand2Band)
  {
    return BAND2BAND_TRANSF;
  }
  else
  {
    return NO_TRANSF;
  }
}

void te::map::RasterTransform::setTransfFunction(RasterTransfFunctions func)
{
  if (func == MONO2THREE_TRANSF)
  {
    m_transfFuncPtr = &RasterTransform::setMono2ThreeBand;
    m_RGBAFuncPtr = &RasterTransform::getMono2ThreeBand;
  }
  else if (func == EXTRACT2RGB_TRANSF)
  {
    m_transfFuncPtr = &RasterTransform::setExtractRGB;
    m_RGBAFuncPtr = &RasterTransform::getExtractRGB;
  }
  else if (func == RED2THREE_TRANSF)
  {
    m_transfFuncPtr = &RasterTransform::setRed2ThreeBand;
    m_RGBAFuncPtr = &RasterTransform::getRed2ThreeBand;
  }
  else if (func == GREEN2THREE_TRANSF)
  {
    m_transfFuncPtr = &RasterTransform::setGreen2ThreeBand;
    m_RGBAFuncPtr = &RasterTransform::getGreen2ThreeBand;
  }
  else if (func == BLUE2THREE_TRANSF)
  {
    m_transfFuncPtr = &RasterTransform::setBlue2ThreeBand;
    m_RGBAFuncPtr = &RasterTransform::getBlue2ThreeBand;
  }
  else if (func == CATEGORIZE_TRANSF)
  {
    m_transfFuncPtr = &RasterTransform::setCategorize;
    m_RGBAFuncPtr = &RasterTransform::getCategorize;
  }
  else if (func == INTERPOLATE_TRANSF)
  {
    m_transfFuncPtr = &RasterTransform::setInterpolate;
    m_RGBAFuncPtr = &RasterTransform::getInterpolate;
  }
  else if (func == RECODE_TRANSF)
  {
    m_transfFuncPtr = &RasterTransform::setRecode;
    m_RGBAFuncPtr = &RasterTransform::getRecode;
  }
  else if (func == BAND2BAND_TRANSF)
  {
    m_transfFuncPtr = &RasterTransform::setBand2Band;
    m_RGBAFuncPtr = nullptr;
  }
  else if (func == EXTRACT2RGBA_TRANSF)
  {
    m_transfFuncPtr = &RasterTransform::setExtractRGBA;
    m_RGBAFuncPtr = &RasterTransform::getExtractRGBA;
  }
  else
  {
    m_transfFuncPtr = nullptr;
    m_RGBAFuncPtr = nullptr;
  }
}

void te::map::RasterTransform::setContrastGainsAndOffsets(
  const std::vector< double >& contrastGains,
  const std::vector< double >& contrastOffsets1,
  const std::vector< double >& contrastOffsets2 )
{
  if(
      ( contrastGains.size() != m_rasterIn->getNumberOfBands() )
      ||
      ( contrastOffsets1.size() != m_rasterIn->getNumberOfBands() )
      ||
      ( contrastOffsets2.size() != m_rasterIn->getNumberOfBands() )
    )
  {
    throw Exception(TE_TR("Invalid vector size"));
  }

  for (std::size_t b = 0; b < m_inputNBands; ++b)
  {
    TransformBandInfo* tbi = m_transformBandInfo[b];

    tbi->m_contrastGain = contrastGains[b];
    tbi->m_contrastOffset1 = contrastOffsets1[b];
    tbi->m_contrastOffset2 = contrastOffsets2[b];
  }
}

void te::map::RasterTransform::setContrastFuncPtr(
  const RasterTransform::ContrastType contrastType )
{
  switch( contrastType )
  {
    case LINEAR_CONTRAST :
    {
      m_contrastTransfFuncPtr = &RasterTransform::applyLinearContrast;
      break;
    }
    case SQUARE_CONTRAST :
    {
      m_contrastTransfFuncPtr = &RasterTransform::applySquareContrast;
      break;
    }
    case SQUARE_ROOT_CONTRAST :
    {
      m_contrastTransfFuncPtr = &RasterTransform::applySquareRootContrast;
      break;
    }
    case LOG_CONTRAST :
    {
      m_contrastTransfFuncPtr = &RasterTransform::applyLogContrast;
      break;
    }
  }
}

void te::map::RasterTransform::setMono2ThreeBand(double icol, double ilin, double ocol, double olin)
{
  double val;

  (this->*(m_channelDomainGetFuncPtr))(static_cast<unsigned int>(icol), static_cast<unsigned int>(ilin), val, static_cast<size_t>(m_monoBand));

  if(checkNoValue(val, m_monoBand) == false)
  {
    (this->*(m_contrastTransfFuncPtr))( static_cast<unsigned int>(m_monoBand), val, val );

    val *= m_transformBandInfo[m_monoBand]->m_gamma;

    fixValue(val, m_monoBand);

    std::vector<double> vecValues;
    vecValues.resize(3, val);

    m_rasterOut->setValues(static_cast<unsigned int>(ocol), static_cast<unsigned int>(olin), vecValues);
  }
}

te::color::RGBAColor te::map::RasterTransform::getMono2ThreeBand(double icol, double ilin)
{
  double val;

  if (m_useBufferCache)
  {
    m_arrayBuffer[m_monoBand]->getValue( (unsigned int)icol, (unsigned int)ilin, 
      val );
  }
  else
  {
    (this->*(m_channelDomainGetFuncPtr))(static_cast<unsigned int>(icol), static_cast<unsigned int>(ilin), val, static_cast<unsigned int>(m_monoBand));
  }

  if(checkNoValue(val, m_monoBand) == false)
  {
    (this->*(m_contrastTransfFuncPtr))( static_cast<unsigned int>(m_monoBand), val, val );

    val *= m_transformBandInfo[m_monoBand]->m_gamma;

    fixValue(val, m_monoBand);

    if(std::isnan(val))
    {
      return te::color::RGBAColor();
    }

    double transp = checkAlphaValue(icol, ilin);

    te::color::RGBAColor c(static_cast<int>(val), static_cast<int>(val), static_cast<int>(val), static_cast<int>(transp));

    return c;
  }

  return te::color::RGBAColor();
}

void te::map::RasterTransform::setExtractRGB(double icol, double ilin, double ocol, double olin)
{
  double valR, valG, valB;

  (this->*(m_channelDomainGetFuncPtr))(static_cast<unsigned int>(icol), static_cast<unsigned int>(ilin), valR, static_cast<size_t>(m_redBand));
  (this->*(m_channelDomainGetFuncPtr))(static_cast<unsigned int>(icol), static_cast<unsigned int>(ilin), valG, static_cast<size_t>(m_greenBand));
  (this->*(m_channelDomainGetFuncPtr))(static_cast<unsigned int>(icol), static_cast<unsigned int>(ilin), valB, static_cast<size_t>(m_blueBand));

  if(checkNoValue(valR, m_redBand) == false ||
     checkNoValue(valG, m_greenBand) == false ||
     checkNoValue(valB, m_blueBand) == false)
  {
    std::vector<double> vecValues;

    (this->*(m_contrastTransfFuncPtr))(static_cast<unsigned int>(m_redBand), valR, valR);
    valR *= m_transformBandInfo[m_redBand]->m_gamma;
    fixValue(valR, m_redBand);
    vecValues.push_back(valR);

    (this->*(m_contrastTransfFuncPtr))(static_cast<unsigned int>(m_greenBand), valG, valG);
    valG *= m_transformBandInfo[m_greenBand]->m_gamma;
    fixValue(valG, m_greenBand);
    vecValues.push_back(valG);

    (this->*(m_contrastTransfFuncPtr))(static_cast<unsigned int>(m_blueBand), valB, valB);
    valB *= m_transformBandInfo[m_blueBand]->m_gamma;
    fixValue(valB, m_blueBand);
    vecValues.push_back(valB);

    m_rasterOut->setValues(static_cast<unsigned int>(ocol), static_cast<unsigned int>(olin), vecValues);
  }
}

te::color::RGBAColor te::map::RasterTransform::getExtractRGB(double icol, double ilin)
{
  double valR, valG, valB;

  if (m_useBufferCache)
  {
    m_arrayBuffer[m_redBand]->getValue( (unsigned int)icol, (unsigned int)ilin, 
      valR );
    m_arrayBuffer[m_greenBand]->getValue( (unsigned int)icol, (unsigned int)ilin, 
      valG );    
    m_arrayBuffer[m_blueBand]->getValue( (unsigned int)icol, (unsigned int)ilin, 
      valB );    
  }
  else
  {
    (this->*(m_channelDomainGetFuncPtr))(static_cast<unsigned int>(icol), static_cast<unsigned int>(ilin), valR, static_cast<size_t>(m_redBand));
    (this->*(m_channelDomainGetFuncPtr))(static_cast<unsigned int>(icol), static_cast<unsigned int>(ilin), valG, static_cast<size_t>(m_greenBand));
    (this->*(m_channelDomainGetFuncPtr))(static_cast<unsigned int>(icol), static_cast<unsigned int>(ilin), valB, static_cast<size_t>(m_blueBand));
  }

  if(checkNoValue(valR, m_redBand) == false ||
     checkNoValue(valG, m_greenBand) == false ||
     checkNoValue(valB, m_blueBand) == false)
  {
    (this->*(m_contrastTransfFuncPtr))(static_cast<unsigned int>(m_redBand), valR, valR);
    valR *= m_transformBandInfo[m_redBand]->m_gamma;
    fixValue(valR, m_redBand);

    (this->*(m_contrastTransfFuncPtr))(static_cast<unsigned int>(m_greenBand), valG, valG);
    valG *= m_transformBandInfo[m_greenBand]->m_gamma;
    fixValue(valG, m_greenBand);

    (this->*(m_contrastTransfFuncPtr))(static_cast<unsigned int>(m_blueBand), valB, valB);
    valB *= m_transformBandInfo[m_blueBand]->m_gamma;
    fixValue(valB, m_blueBand);

    double transp = checkAlphaValue(icol, ilin);

    te::color::RGBAColor c(static_cast<int>(valR), static_cast<int>(valG), static_cast<int>(valB), static_cast<int>(transp));

    return c;
  }

  return te::color::RGBAColor();
}

void te::map::RasterTransform::setExtractRGBA(double icol, double ilin, double ocol, double olin)
{
  double valR, valG, valB, valA;

  (this->*(m_channelDomainGetFuncPtr))(static_cast<unsigned int>(icol), static_cast<unsigned int>(ilin), valR, static_cast<size_t>(m_redBand));
  (this->*(m_channelDomainGetFuncPtr))(static_cast<unsigned int>(icol), static_cast<unsigned int>(ilin), valG, static_cast<size_t>(m_greenBand));
  (this->*(m_channelDomainGetFuncPtr))(static_cast<unsigned int>(icol), static_cast<unsigned int>(ilin), valB, static_cast<size_t>(m_blueBand));
  (this->*(m_channelDomainGetFuncPtr))(static_cast<unsigned int>(icol), static_cast<unsigned int>(ilin), valA, static_cast<size_t>(m_alphaBand));

  if(checkNoValue(valR, m_redBand) == false ||
     checkNoValue(valG, m_greenBand) == false ||
     checkNoValue(valB, m_blueBand) == false ||
     checkNoValue(valA, m_alphaBand) == false)
  {
    std::vector<double> vecValues;

    (this->*(m_contrastTransfFuncPtr))(static_cast<unsigned int>(m_redBand), valR, valR);
    valR *= m_transformBandInfo[m_redBand]->m_gamma;
    fixValue(valR, m_alphaBand);
    vecValues.push_back(valR);

    (this->*(m_contrastTransfFuncPtr))(static_cast<unsigned int>(m_greenBand), valG, valG);
    valG *= m_transformBandInfo[m_greenBand]->m_gamma;
    fixValue(valG, m_greenBand);
    vecValues.push_back(valG);

    (this->*(m_contrastTransfFuncPtr))(static_cast<unsigned int>(m_blueBand), valB, valB);
    valB *= m_transformBandInfo[m_blueBand]->m_gamma;
    fixValue(valB, m_blueBand);
    vecValues.push_back(valB);

    fixValue(valA, m_alphaBand);
    vecValues.push_back(valB);

    m_rasterOut->setValues(static_cast<unsigned int>(ocol), static_cast<unsigned int>(olin), vecValues);
  }
}

te::color::RGBAColor te::map::RasterTransform::getExtractRGBA(double icol, double ilin)
{
  double valR, valG, valB, valA;

  if (m_useBufferCache)
  {
    m_arrayBuffer[m_redBand]->getValue( (unsigned int)icol, (unsigned int)ilin, 
      valR );    
    m_arrayBuffer[m_greenBand]->getValue( (unsigned int)icol, (unsigned int)ilin, 
      valG );    
    m_arrayBuffer[m_blueBand]->getValue( (unsigned int)icol, (unsigned int)ilin, 
      valB );    
    m_arrayBuffer[m_alphaBand]->getValue( (unsigned int)icol, (unsigned int)ilin, 
      valA );    
    
  }
  else
  {
    (this->*(m_channelDomainGetFuncPtr))(static_cast<unsigned int>(icol), static_cast<unsigned int>(ilin), valR, static_cast<size_t>(m_redBand));
    (this->*(m_channelDomainGetFuncPtr))(static_cast<unsigned int>(icol), static_cast<unsigned int>(ilin), valG, static_cast<size_t>(m_greenBand));
    (this->*(m_channelDomainGetFuncPtr))(static_cast<unsigned int>(icol), static_cast<unsigned int>(ilin), valB, static_cast<size_t>(m_blueBand));
    (this->*(m_channelDomainGetFuncPtr))(static_cast<unsigned int>(icol), static_cast<unsigned int>(ilin), valA, static_cast<size_t>(m_alphaBand));
  }

  if(checkNoValue(valR, m_redBand) == false ||
     checkNoValue(valG, m_greenBand) == false ||
     checkNoValue(valB, m_blueBand) == false ||
     checkNoValue(valA, m_alphaBand) == false)
  {
    (this->*(m_contrastTransfFuncPtr))(static_cast<unsigned int>(m_redBand), valR, valR);
    valR *= m_transformBandInfo[m_redBand]->m_gamma;
    fixValue(valR, m_redBand);

    (this->*(m_contrastTransfFuncPtr))(static_cast<unsigned int>(m_greenBand), valG, valG);
    valG *= m_transformBandInfo[m_greenBand]->m_gamma;
    fixValue(valG, m_greenBand);

    (this->*(m_contrastTransfFuncPtr))(static_cast<unsigned int>(m_blueBand), valB, valB);
    valB *= m_transformBandInfo[m_blueBand]->m_gamma;
    fixValue(valB, m_blueBand);

    fixValue(valA, m_alphaBand);

    if(valA < m_transp)
      return te::color::RGBAColor(static_cast<int>(valR), static_cast<int>(valG), static_cast<int>(valB), static_cast<int>(valA));
    else
      return te::color::RGBAColor(static_cast<int>(valR), static_cast<int>(valG), static_cast<int>(valB), static_cast<int>(m_transp));
  }

  return te::color::RGBAColor();
}

void te::map::RasterTransform::setRed2ThreeBand(double icol, double ilin, double ocol, double olin)
{
  double val;

  (this->*(m_channelDomainGetFuncPtr))(static_cast<unsigned int>(icol), static_cast<unsigned int>(ilin), val, static_cast<size_t>(m_redBand));

  if(checkNoValue(val, m_redBand) == false)
  {
    (this->*(m_contrastTransfFuncPtr))(static_cast<unsigned int>(m_redBand), val, val);
    val *= m_transformBandInfo[m_redBand]->m_gamma;
    fixValue(val, m_redBand);

    std::vector<double> vecValues;
    vecValues.push_back(val);
    vecValues.push_back(0.);
    vecValues.push_back(0.);

    m_rasterOut->setValues(static_cast<unsigned int>(ocol), static_cast<unsigned int>(olin), vecValues);
  }
}

te::color::RGBAColor te::map::RasterTransform::getRed2ThreeBand(double icol, double ilin)
{
  double val;

  if (m_useBufferCache)
  {
    m_arrayBuffer[m_redBand]->getValue( (unsigned int)icol, (unsigned int)ilin, 
      val );       
  }
  else
  {
    (this->*(m_channelDomainGetFuncPtr))(static_cast<unsigned int>(icol), static_cast<unsigned int>(ilin), val, static_cast<size_t>(m_redBand));
  }

  if(checkNoValue(val, m_redBand) == false)
  {
    (this->*(m_contrastTransfFuncPtr))(static_cast<unsigned int>(m_redBand), val, val);
    val *= m_transformBandInfo[m_redBand]->m_gamma;
    fixValue(val, m_redBand);

    double transp = checkAlphaValue(icol, ilin);

    te::color::RGBAColor c(static_cast<int>(val), static_cast<int>(0.), static_cast<int>(0.), static_cast<int>(transp));

    return c;

  }

  return te::color::RGBAColor();
}

void te::map::RasterTransform::setGreen2ThreeBand(double icol, double ilin, double ocol, double olin)
{
  double val;

  (this->*(m_channelDomainGetFuncPtr))(static_cast<unsigned int>(icol), static_cast<unsigned int>(ilin), val, static_cast<size_t>(m_greenBand));

  if(checkNoValue(val, m_greenBand) == false)
  {
    (this->*(m_contrastTransfFuncPtr))(static_cast<unsigned int>(m_greenBand), val, val);
    val *= m_transformBandInfo[m_greenBand]->m_gamma;
    fixValue(val, m_greenBand);

    std::vector<double> vecValues;
    vecValues.push_back(0.);
    vecValues.push_back(val);
    vecValues.push_back(0.);

    m_rasterOut->setValues(static_cast<unsigned int>(ocol), static_cast<unsigned int>(olin), vecValues);
  }
}

te::color::RGBAColor te::map::RasterTransform::getGreen2ThreeBand(double icol, double ilin)
{
  double val;

  if (m_useBufferCache)
  {
    m_arrayBuffer[m_greenBand]->getValue( (unsigned int)icol, (unsigned int)ilin, 
      val );           
  }
  else
  {
    (this->*(m_channelDomainGetFuncPtr))(static_cast<unsigned int>(icol), static_cast<unsigned int>(ilin), val, static_cast<size_t>(m_greenBand));
  }

  if(checkNoValue(val, m_greenBand) == false)
  {
    (this->*(m_contrastTransfFuncPtr))(static_cast<unsigned int>(m_greenBand), val, val);
    val *= m_transformBandInfo[m_greenBand]->m_gamma;
    fixValue(val, m_greenBand);

    double transp = checkAlphaValue(icol, ilin);

    te::color::RGBAColor c(static_cast<int>(0.), static_cast<int>(val), static_cast<int>(0.), static_cast<int>(transp));

    return c;

  }

  return te::color::RGBAColor();
}

void te::map::RasterTransform::setBlue2ThreeBand(double icol, double ilin, double ocol, double olin)
{
  double val;

  (this->*(m_channelDomainGetFuncPtr))(static_cast<unsigned int>(icol), static_cast<unsigned int>(ilin), val, static_cast<size_t>(m_blueBand));

  if(checkNoValue(val, m_blueBand) == false)
  {
    (this->*(m_contrastTransfFuncPtr))(static_cast<unsigned int>(m_blueBand), val, val);
    val *= m_transformBandInfo[m_blueBand]->m_gamma;
    fixValue(val, m_blueBand);

    std::vector<double> vecValues;
    vecValues.push_back(0.);
    vecValues.push_back(0.);
    vecValues.push_back(val);

    m_rasterOut->setValues(static_cast<unsigned int>(ocol), static_cast<unsigned int>(olin), vecValues);
  }
}

te::color::RGBAColor te::map::RasterTransform::getBlue2ThreeBand(double icol, double ilin)
{
  double val;

  if (m_useBufferCache)
  {
    m_arrayBuffer[m_blueBand]->getValue( (unsigned int)icol, (unsigned int)ilin, 
      val );         
  }
  else
  {
    (this->*(m_channelDomainGetFuncPtr))(static_cast<unsigned int>(icol), static_cast<unsigned int>(ilin), val, static_cast<size_t>(m_blueBand));
  }

  if(checkNoValue(val, m_blueBand) == false)
  {
    (this->*(m_contrastTransfFuncPtr))(static_cast<unsigned int>(m_blueBand), val, val);
    val *= m_transformBandInfo[m_blueBand]->m_gamma;
    fixValue(val, m_blueBand);

    double transp = checkAlphaValue(icol, ilin);

    te::color::RGBAColor c(static_cast<int>(0.), static_cast<int>(0.), static_cast<int>(val), static_cast<int>(transp));

    return c;

  }

  return te::color::RGBAColor();
}

void te::map::RasterTransform::setCategorize(double icol, double ilin, double ocol, double olin)
{
   double val;

  (this->*(m_channelDomainGetFuncPtr))(static_cast<unsigned int>(icol), static_cast<unsigned int>(ilin), val, static_cast<size_t>(m_monoBand));

  te::color::RGBAColor c=  getCategorizedColor(val);

  std::vector<double> vecValues;
  vecValues.push_back(c.getRed());
  vecValues.push_back(c.getGreen());
  vecValues.push_back(c.getBlue());

  m_rasterOut->setValues(static_cast<unsigned int>(ocol), static_cast<unsigned int>(olin), vecValues);
}

te::color::RGBAColor te::map::RasterTransform::getCategorize(double icol, double ilin)
{
  double val;

  if (m_useBufferCache)
  {
    m_arrayBuffer[m_monoBand]->getValue( (unsigned int)icol, (unsigned int)ilin, 
      val );             
  }
  else
  {
    (this->*(m_channelDomainGetFuncPtr))(static_cast<unsigned int>(icol), static_cast<unsigned int>(ilin), val, static_cast<size_t>(m_monoBand));
  }

  if(checkNoValue(val, m_monoBand) == false)
    return getCategorizedColor(val);

  return te::color::RGBAColor();
}

void te::map::RasterTransform::setInterpolate(double icol, double ilin, double ocol, double olin)
{
   double val;

  (this->*(m_channelDomainGetFuncPtr))(static_cast<unsigned int>(icol), static_cast<unsigned int>(ilin), val, static_cast<size_t>(m_monoBand));

  te::color::RGBAColor c=  getInterpolatedColor(val);

  std::vector<double> vecValues;
  vecValues.push_back(c.getRed());
  vecValues.push_back(c.getGreen());
  vecValues.push_back(c.getBlue());

  m_rasterOut->setValues(static_cast<unsigned int>(ocol), static_cast<unsigned int>(olin), vecValues);
}

te::color::RGBAColor te::map::RasterTransform::getInterpolate(double icol, double ilin)
{
  double val;

  if (m_useBufferCache)
  {
    m_arrayBuffer[m_monoBand]->getValue( (unsigned int)icol, (unsigned int)ilin, 
      val );      
  }
  else
  {
    (this->*(m_channelDomainGetFuncPtr))(static_cast<unsigned int>(icol), static_cast<unsigned int>(ilin), val, static_cast<size_t>(m_monoBand));
  }

  if(checkNoValue(val, m_monoBand) == false)
    return getInterpolatedColor(val);

  return te::color::RGBAColor();
}

void te::map::RasterTransform::setRecode(double icol, double ilin, double ocol, double olin)
{
  double val;

  (this->*(m_channelDomainGetFuncPtr))(static_cast<unsigned int>(icol), static_cast<unsigned int>(ilin), val, static_cast<size_t>(m_monoBand));

  te::color::RGBAColor c = getRecodedColor(val);

  std::vector<double> vecValues;
  vecValues.push_back(c.getRed());
  vecValues.push_back(c.getGreen());
  vecValues.push_back(c.getBlue());

  m_rasterOut->setValues(static_cast<unsigned int>(ocol), static_cast<unsigned int>(olin), vecValues);
}

te::color::RGBAColor te::map::RasterTransform::getRecode(double icol, double ilin)
{
  double val;

  if (m_useBufferCache)
  {
    m_arrayBuffer[m_monoBand]->getValue( (unsigned int)icol, (unsigned int)ilin, 
      val );          
  }
  else
  {
    (this->*(m_channelDomainGetFuncPtr))(static_cast<unsigned int>(icol), static_cast<unsigned int>(ilin), val, static_cast<size_t>(m_monoBand));
  }

  if (checkNoValue(val, m_monoBand) == false)
    return getRecodedColor(val);

  return te::color::RGBAColor();
}

void te::map::RasterTransform::setBand2Band(double icol, double ilin, double ocol, double olin)
{
 double val;

  for(std::size_t t = 0; t < m_rasterIn->getNumberOfBands(); ++t)
  {
    (this->*(m_channelDomainGetFuncPtr))(static_cast<unsigned int>(icol), static_cast<unsigned int>(ilin), val, t);

    if(checkNoValue(val, static_cast<int>(t) == false))
    {
      (this->*(m_contrastTransfFuncPtr))( static_cast<unsigned int>(t), val, val );
      val *= m_transformBandInfo[t]->m_gamma;

      short band = static_cast<short>(t);
      fixValue(val, band);

      m_rasterOut->setValue(static_cast<unsigned int>(ocol), static_cast<unsigned int>(olin), val, t);
    }
  }
}

double te::map::RasterTransform::checkAlphaValue(double icol, double ilin)
{
  if (m_alphaBand != -1)
  {
    double transp = 0.;

    if (m_useBufferCache)
    {
      m_arrayBuffer[m_alphaBand]->getValue( (unsigned int)icol, (unsigned int)ilin, 
        transp );                
    }
    else
    {
      (this->*(m_channelDomainGetFuncPtr))(static_cast<unsigned int>(icol), static_cast<unsigned int>(ilin), transp, static_cast<size_t>(m_alphaBand));
    }

    if (transp < m_transp)
      return transp;
    else
      return m_transp;
  }

  return m_transp;
}

te::color::RGBAColor te::map::RasterTransform::getInterpolatedColor(double value)
{
  InterpolatedMap::iterator it = m_interpolateMap.begin();

  while(it != m_interpolateMap.end())
  {
    RasterThreshold rt = it->first;

    if(rt.first<= value && rt.second > value)
    {
      int distance = int(value - it->first.first);

      if(distance < 0)
      {
        distance = 0;
      }
      else if(distance >= static_cast<int>(it->second.getColorBar().size()))
      {
        distance = static_cast<int>(it->second.getColorBar().size() - 1);
      }

      return it->second.getColorBar()[static_cast<size_t>(distance)];
    }

    ++it;
  }

  return te::color::RGBAColor();
}

te::color::RGBAColor te::map::RasterTransform::getCategorizedColor(double value)
{
  CategorizedMap::iterator it = m_categorizeMap.begin();

  while(it != m_categorizeMap.end())
  {
    RasterThreshold rt = it->first;

    if(rt.first<= value && rt.second > value)
    {
      return it->second;
    }

    ++it;
  }

  return te::color::RGBAColor();
}

te::color::RGBAColor te::map::RasterTransform::getRecodedColor(double value)
{
  return m_recodeMap[static_cast<int>(value)];
}

void te::map::RasterTransform::setChannelDomain( const ChannelDomain domain )
{
  if( domain == REAL_CHANNEL_DOMAIN )
  {
    m_currentChannelDomain = domain;
    m_channelDomainGetFuncPtr = &RasterTransform::getRealValueChannelDomainGetFunction;
  }
  else if( domain == IMAGINARY_CHANNEL_DOMAIN )
  {
    m_currentChannelDomain = domain;
    m_channelDomainGetFuncPtr = &RasterTransform::getImaginaryValueChannelDomainGetFunction;
  }
  else if( domain == SQRT_CHANNEL_DOMAIN )
  {
    m_currentChannelDomain = domain;
    m_channelDomainGetFuncPtr = &RasterTransform::getSQRTChannelDomainGetFunction;
  }
  else
  {
    throw Exception(TE_TR("Invalid channel domain"));
  }      
}

te::map::RasterTransform::ChannelDomain te::map::RasterTransform::getChannelDomain() const
{
  return m_currentChannelDomain;
}