/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file src/terralib/maptools/serialization/xml/Utils.cpp
   
  \brief Auxiliary functions to read layer information from a XML document.
*/

// TerraLib
#include "../../../core/uri/Utils.h"
#include "../../../se/serialization/xml/Symbolizer.h"
#include "../../../srs/Config.h"
#include "../../../srs/Exception.h"
#include "../../../srs/SpatialReferenceSystemManager.h"
#include "../../../xml/AbstractWriter.h"
#include "../../../xml/Reader.h"
#include "../../AbstractLayer.h"
#include "../../Chart.h"
#include "../../Enums.h"
#include "../../Grouping.h"
#include "../../RasterContrast.h"
#include "Utils.h"

// Boost
#include <boost/lexical_cast.hpp>

// STL
#include <cassert>

te::map::Visibility te::map::serialize::GetVisibility(const std::string& visible)
{
  return (visible == "VISIBLE") ? te::map::VISIBLE :
         (visible == "NOT_VISIBLE") ? te::map::NOT_VISIBLE :
         te::map::PARTIALLY_VISIBLE;
}

std::string te::map::serialize::GetVisibility(const te::map::Visibility& visible)
{
  return (visible == te::map::VISIBLE) ? "VISIBLE" :
         (visible == te::map::NOT_VISIBLE) ? "NOT_VISIBLE" :
         "PARTIALLY_VISIBLE";
}

te::map::GroupingType te::map::serialize::GetGroupingType(const std::string& type)
{
  if(type == "EQUAL_STEPS")
    return te::map::EQUAL_STEPS;

  if(type == "QUANTIL")
    return te::map::QUANTIL;

  if(type == "STD_DEVIATION")
    return te::map::STD_DEVIATION;

  assert(type == "UNIQUE_VALUE");

  return te::map::UNIQUE_VALUE;
}

std::string te::map::serialize::GetGroupingType(const te::map::GroupingType& type)
{
  switch(type)
  {
    case te::map::EQUAL_STEPS:
      return "EQUAL_STEPS";

    case te::map::QUANTIL:
      return "QUANTIL";

    case te::map::STD_DEVIATION:
      return "STD_DEVIATION";

    default:
    {
      assert(type == te::map::UNIQUE_VALUE);
      return "UNIQUE_VALUE";
    }
  }
}

std::string te::map::serialize::ReadLayerTitle(te::xml::Reader& reader)
{
  assert(reader.getNodeType() == te::xml::START_ELEMENT);
  assert(reader.getElementLocalName() == "Title");
  reader.next();
  assert(reader.getNodeType() == te::xml::VALUE);
  std::string title = te::core::URIDecode(reader.getElementValue());
  reader.next();
  assert(reader.getNodeType() == te::xml::END_ELEMENT);

  return title;
}

std::string te::map::serialize::ReadLayerEncoding(te::xml::Reader& reader)
{
  std::string encoding;

  if (reader.getNodeType() == te::xml::START_ELEMENT &&
      reader.getElementLocalName() == "Encoding")
  {
    reader.next();
    assert(reader.getNodeType() == te::xml::VALUE);
    encoding = reader.getElementValue();
    reader.next();
    assert(reader.getNodeType() == te::xml::END_ELEMENT);
  }

  return encoding;
}

std::string te::map::serialize::ReadDataSetName(te::xml::Reader& reader)
{
  std::string dataset;

  assert(reader.getNodeType() == te::xml::START_ELEMENT);
  assert(reader.getElementLocalName() == "DataSetName");
  reader.next();
  if (reader.getNodeType() == te::xml::VALUE)
  {
    dataset = te::core::URIDecode(reader.getElementValue());
    reader.next();
  }

  assert(reader.getNodeType() == te::xml::END_ELEMENT);

  return dataset;
}

std::string te::map::serialize::ReadDataSourceId(te::xml::Reader& reader)
{
  std::string datasourceId;

  assert(reader.getNodeType() == te::xml::START_ELEMENT);
  assert(reader.getElementLocalName() == "DataSourceId");
  reader.next();
  if (reader.getNodeType() == te::xml::VALUE)
  {
    datasourceId = reader.getElementValue();
    reader.next();
  }
  assert(reader.getNodeType() == te::xml::END_ELEMENT);

  return datasourceId;
}

std::string te::map::serialize::ReadLayerVisibility(te::xml::Reader& reader)
{
  assert(reader.getNodeType() == te::xml::START_ELEMENT);
  assert(reader.getElementLocalName() == "Visible");
  reader.next();
  assert(reader.getNodeType() == te::xml::VALUE);
  std::string visible = reader.getElementValue();
  reader.next();
  assert(reader.getNodeType() == te::xml::END_ELEMENT);

  return visible;
}

te::map::Grouping* te::map::serialize::ReadLayerGrouping(te::xml::Reader& reader)
{
  if(reader.getElementLocalName() != "Grouping")
    return nullptr;

  /* Property Name */
  reader.next();
  assert(reader.getNodeType() == te::xml::START_ELEMENT);
  assert(reader.getElementLocalName() == "PropertyName");
  reader.next();
  assert(reader.getNodeType() == te::xml::VALUE);
  std::string propertyName = reader.getElementValue();
  reader.next();
  assert(reader.getNodeType() == te::xml::END_ELEMENT);

  /* Property Data Type */
  reader.next();
  assert(reader.getNodeType() == te::xml::START_ELEMENT);
  assert(reader.getElementLocalName() == "PropertyDataType");
  reader.next();
  assert(reader.getNodeType() == te::xml::VALUE);
  int propertyType = reader.getElementValueAsInt32();
  reader.next();
  assert(reader.getNodeType() == te::xml::END_ELEMENT);

  /* Grouping Type */
  reader.next();
  assert(reader.getNodeType() == te::xml::START_ELEMENT);
  assert(reader.getElementLocalName() == "Type");
  reader.next();
  assert(reader.getNodeType() == te::xml::VALUE);
  std::string type = reader.getElementValue();
  reader.next();
  assert(reader.getNodeType() == te::xml::END_ELEMENT);

  /* Precision */
  reader.next();
  assert(reader.getNodeType() == te::xml::START_ELEMENT);
  assert(reader.getElementLocalName() == "Precision");
  reader.next();
  assert(reader.getNodeType() == te::xml::VALUE);
  std::size_t precision = static_cast<std::size_t>(reader.getElementValueAsInt32());
  reader.next();
  assert(reader.getNodeType() == te::xml::END_ELEMENT);
  reader.next();

  std::unique_ptr<te::map::Grouping> g(new te::map::Grouping(propertyName, GetGroupingType(type), precision));
  g->setPropertyType(propertyType);

  /* Number of Slices */
  if(reader.getElementLocalName() == "NumSlices")
  {
    assert(reader.getNodeType() == te::xml::START_ELEMENT);
    reader.next();
    assert(reader.getNodeType() == te::xml::VALUE);
    std::size_t numSlices =static_cast<std::size_t>(reader.getElementValueAsInt32());
    reader.next();
    assert(reader.getNodeType() == te::xml::END_ELEMENT);

    g->setNumSlices(numSlices);

    reader.next();
  }

  /* Smmary */
  if(reader.getElementLocalName() == "Summary")
  {
    assert(reader.getNodeType() == te::xml::START_ELEMENT);
    reader.next();
    assert(reader.getNodeType() == te::xml::VALUE);
    std::string summary = reader.getElementValue();
    reader.next();
    assert(reader.getNodeType() == te::xml::END_ELEMENT);

    g->setSummary(summary);
    
    reader.next();
  }

  if(reader.getElementLocalName() == "StandardDeviation")
  {
    assert(reader.getNodeType() == te::xml::START_ELEMENT);
    reader.next();
    assert(reader.getNodeType() == te::xml::VALUE);
    double stdDeviation = reader.getElementValueAsDouble();
    reader.next();
    assert(reader.getNodeType() == te::xml::END_ELEMENT);
    
    g->setStdDeviation(stdDeviation);

    reader.next();
  }

  ////////////////////////////////// DEPRECATED////////////////////////
  /* Grouping Items */
  while (reader.getNodeType() != te::xml::END_ELEMENT &&
        reader.getElementLocalName() == "Grouping")
  {
    reader.next();
  }
  ////////////////////////////////////////////////////////////////////

  reader.next();

  return g.release();
}

te::map::RasterContrast* te::map::serialize::ReadLayerRasterContrast(te::xml::Reader& reader)
{
  if (reader.getElementLocalName() != "RasterContrast")
    return nullptr;

  assert(reader.getNodeType() == te::xml::START_ELEMENT);
  assert(reader.getElementLocalName() == "RasterContrast");

  reader.next();

  /* Type */
  assert(reader.getElementLocalName() == "ContrastType");
  reader.next();
  assert(reader.getNodeType() == te::xml::VALUE);
  int type = reader.getElementValueAsInt32();
  reader.next();
  assert(reader.getNodeType() == te::xml::END_ELEMENT);

  reader.next();

  /* Type */
  assert(reader.getElementLocalName() == "NBands");
  reader.next();
  assert(reader.getNodeType() == te::xml::VALUE);
  int nBands = reader.getElementValueAsInt32();
  reader.next();
  assert(reader.getNodeType() == te::xml::END_ELEMENT);

  std::unique_ptr<te::map::RasterContrast> rc(new te::map::RasterContrast(static_cast<te::map::RasterTransform::ContrastType>(type), static_cast<unsigned long>(nBands)));

  reader.next();

  /* Contrast Values */
  for (int i = 0; i < nBands; ++i)
  {
    assert(reader.getNodeType() == te::xml::START_ELEMENT);
    assert(reader.getElementLocalName() == "BandContrast");
    reader.next();

    /* gain */
    assert(reader.getElementLocalName() == "GainContrast");
    reader.next();
    assert(reader.getNodeType() == te::xml::VALUE);
    double gain = reader.getElementValueAsDouble();
    reader.next();
    assert(reader.getNodeType() == te::xml::END_ELEMENT);

    reader.next();

    /* offset1 */
    assert(reader.getElementLocalName() == "Offset1Contrast");
    reader.next();
    assert(reader.getNodeType() == te::xml::VALUE);
    double offset1 = reader.getElementValueAsDouble();
    reader.next();
    assert(reader.getNodeType() == te::xml::END_ELEMENT);

    reader.next();

    /* offset2 */
    assert(reader.getElementLocalName() == "Offset2Contrast");
    reader.next();
    assert(reader.getNodeType() == te::xml::VALUE);
    double offset2 = reader.getElementValueAsDouble();
    reader.next();
    assert(reader.getNodeType() == te::xml::END_ELEMENT);

    reader.next();

    /* min */
    assert(reader.getElementLocalName() == "minContrast");
    reader.next();
    assert(reader.getNodeType() == te::xml::VALUE);
    double min = reader.getElementValueAsDouble();
    reader.next();
    assert(reader.getNodeType() == te::xml::END_ELEMENT);

    reader.next();

    /* max */
    assert(reader.getElementLocalName() == "maxContrast");
    reader.next();
    assert(reader.getNodeType() == te::xml::VALUE);
    double max = reader.getElementValueAsDouble();
    reader.next();
    assert(reader.getNodeType() == te::xml::END_ELEMENT);

    reader.next();

    assert(reader.getNodeType() == te::xml::END_ELEMENT);
    reader.next();

    rc->setValues(gain, offset1, offset2, min, max, static_cast<unsigned long>(i));
  }

  assert(reader.getNodeType() == te::xml::END_ELEMENT);
  reader.next();

  return rc.release();
}

std::unique_ptr<te::map::Chart> te::map::serialize::ReadLayerChart(te::xml::Reader& reader)
{
  if(reader.getElementLocalName() != "Chart")
    return std::unique_ptr<te::map::Chart>();

  assert(reader.getNodeType() == te::xml::START_ELEMENT);
  assert(reader.getElementLocalName() == "Chart");

  reader.next();

  assert(reader.getNodeType() == te::xml::START_ELEMENT);
  assert(reader.getElementLocalName() == "Type");

  reader.next();

  assert(reader.getNodeType() == te::xml::VALUE);

  te::map::ChartType type = reader.getElementValue() == "PIE" ? te::map::Pie : te::map::Bar;

  reader.next();

  assert(reader.getNodeType() == te::xml::END_ELEMENT); // Type

  reader.next();

  assert(reader.getNodeType() == te::xml::START_ELEMENT);
  assert(reader.getElementLocalName() == "ContourColor");

  reader.next();

  assert(reader.getNodeType() == te::xml::VALUE);
  
  std::string hexColor = reader.getElementValue();

  te::color::RGBAColor contourColor(hexColor);

  reader.next();

  assert(reader.getNodeType() == te::xml::END_ELEMENT); // ContourColor

  reader.next();

  assert(reader.getNodeType() == te::xml::START_ELEMENT);
  assert(reader.getElementLocalName() == "ContourWidth");

  reader.next();

  assert(reader.getNodeType() == te::xml::VALUE);

  std::size_t contourWidth = static_cast<size_t>(reader.getElementValueAsInt32());

  reader.next();

  assert(reader.getNodeType() == te::xml::END_ELEMENT); // ContourWidth

  reader.next();

  assert(reader.getNodeType() == te::xml::START_ELEMENT);
  assert(reader.getElementLocalName() == "Height");

  reader.next();

  assert(reader.getNodeType() == te::xml::VALUE);

  std::size_t height = static_cast<size_t>(reader.getElementValueAsInt32());

  reader.next();

  assert(reader.getNodeType() == te::xml::END_ELEMENT); // Height

  reader.next();

  std::size_t barWidth = std::string::npos;

  if(reader.getNodeType() == te::xml::START_ELEMENT &&
     reader.getElementLocalName() == "BarWidth")
  {
    reader.next();

    assert(reader.getNodeType() == te::xml::VALUE);

    barWidth = static_cast<size_t>(reader.getElementValueAsInt32());

    reader.next();

    assert(reader.getNodeType() == te::xml::END_ELEMENT); // BarWidth

    reader.next();
  }

  assert(reader.getNodeType() == te::xml::START_ELEMENT);
  assert(reader.getElementLocalName() == "IsVisible");

  reader.next();

  assert(reader.getNodeType() == te::xml::VALUE);

  bool isVisible = reader.getElementValueAsBoolean();

  reader.next();

  assert(reader.getNodeType() == te::xml::END_ELEMENT); // IsVisible

  reader.next();

  double maxValue = 0;
  if(type == te::map::Bar)
  {
    assert(reader.getNodeType() == te::xml::START_ELEMENT);
    assert(reader.getElementLocalName() == "MaxValue");

    reader.next();

    assert(reader.getNodeType() == te::xml::VALUE);

    maxValue = reader.getElementValueAsDouble();

    reader.next();

    assert(reader.getNodeType() == te::xml::END_ELEMENT); // MaxValue

    reader.next();
  }

  std::string summary  = "NONE";
  if(reader.getNodeType() == te::xml::START_ELEMENT &&
     reader.getElementLocalName() == "Summary")
  {
    reader.next();

    assert(reader.getNodeType() == te::xml::VALUE);

    summary = reader.getElementValue();

    reader.next();

    assert(reader.getNodeType() == te::xml::END_ELEMENT); // Summary

    reader.next();
  }

  std::vector<std::string> properties;
  std::vector<te::color::RGBAColor> colors;

  while(reader.getNodeType() == te::xml::START_ELEMENT &&
        reader.getElementLocalName() == "Slice")
  {
    reader.next();

    assert(reader.getNodeType() == te::xml::START_ELEMENT);
    assert(reader.getElementLocalName() == "PropertyName");

    reader.next();

    assert(reader.getNodeType() == te::xml::VALUE);

    std::string propName = reader.getElementValue();

    reader.next();

    assert(reader.getNodeType() == te::xml::END_ELEMENT); // PropertyName

    reader.next();

    assert(reader.getNodeType() == te::xml::START_ELEMENT);
    assert(reader.getElementLocalName() == "Color");

    reader.next();

    assert(reader.getNodeType() == te::xml::VALUE);

    std::string hColor = reader.getElementValue();
    te::color::RGBAColor color(hColor);

    reader.next();
    assert(reader.getNodeType() == te::xml::END_ELEMENT); // Color

    reader.next();
    assert(reader.getNodeType() == te::xml::END_ELEMENT); // Slice

    reader.next();

    properties.push_back(propName);
    colors.push_back(hColor);
  }

  assert(reader.getNodeType() == te::xml::END_ELEMENT); // Chart

  reader.next();

  std::unique_ptr<te::map::Chart> chart(new te::map::Chart(type, properties, colors));
  chart->setContourColor(contourColor);
  chart->setContourWidth(contourWidth);
  chart->setHeight(height);
  chart->setVisibility(isVisible);
  chart->setSummary(summary);
  if(barWidth != std::string::npos)
    chart->setBarWidth(barWidth);
  if(type == te::map::Bar)
    chart->setMaxValue(maxValue);

  return chart;
}

int te::map::serialize::ReadSRIDValue(te::xml::Reader& reader)
{
  assert(reader.getNodeType() == te::xml::START_ELEMENT);
  assert(reader.getElementLocalName() == "SRSInfo");

  reader.next();

  //srid
  assert(reader.getNodeType() == te::xml::START_ELEMENT);
  assert(reader.getElementLocalName() == "SRID");
  reader.next();
  assert(reader.getNodeType() == te::xml::VALUE);
  int srid = reader.getElementValueAsInt32();
  reader.next();
  assert(reader.getNodeType() == te::xml::END_ELEMENT);

  reader.next();

  //proj4
  assert(reader.getNodeType() == te::xml::START_ELEMENT);
  assert(reader.getElementLocalName() == "Proj4");
  reader.next();
  std::string proj4;
  if(reader.getNodeType() != te::xml::END_ELEMENT)
  {
    proj4 = reader.getElementValue();
    reader.next();
  }

  assert(reader.getNodeType() == te::xml::END_ELEMENT);
  reader.next();

  if (srid >= TE_SRS_USER_DEFINED_START_ID && srid != TE_SRS_GOOGLE)
  {
    if (srid != 0)
    {
      if(te::srs::SpatialReferenceSystemManager::getInstance().isValidP4TxtID(proj4))
      {
        std::pair<std::string, unsigned int> sridInfo = te::srs::SpatialReferenceSystemManager::getInstance().getIdFromP4Txt(proj4);
        srid = static_cast<int>(sridInfo.second);
      }
      else
      {
        srid = te::srs::SpatialReferenceSystemManager::getInstance().getNewUserDefinedSRID();
        std::string newName = "USER:" + boost::lexical_cast<std::string>(srid);
        te::srs::SpatialReferenceSystemManager::getInstance().add(newName, proj4, "", static_cast<unsigned int>(srid), "USER");
      }
    }
  }

  return srid;
}

void te::map::serialize::WriteLayerChart(te::map::Chart* chart, te::xml::AbstractWriter& writer)
{
  writer.writeStartElement("te_map:Chart");

  te::map::ChartType type = chart->getType();

  writer.writeElement("te_map:Type", (type == te::map::Pie ? "PIE" : "BAR"));

  writer.writeElement("te_map:ContourColor", chart->getContourColor().getColor());

  writer.writeElement("te_map:ContourWidth", boost::lexical_cast<int>(chart->getContourWidth()));

  writer.writeElement("te_map:Height", boost::lexical_cast<int>(chart->getHeight()));

  if(chart->getType() == te::map::Bar)
    writer.writeElement("te_map:BarWidth", boost::lexical_cast<int>(chart->getBarWidth()));

  writer.writeElement("te_map:IsVisible", (chart->isVisible() ? "true" : "false"));

  if(chart->getType() == te::map::Bar)
    writer.writeElement("te_map:MaxValue", chart->getMaxValue());

  if(chart->getSummary() != "NONE")
    writer.writeElement("te_map:Summary", chart->getSummary());

  std::vector<std::string> properties = chart->getProperties();

  for(std::size_t i = 0; i < properties.size(); ++i)
  {
    writer.writeStartElement("te_map:Slice");

    writer.writeElement("te_map:PropertyName", properties[i]);

    writer.writeElement("te_map:Color", chart->getColor(i).getColor());

    writer.writeEndElement("te_map:Slice");
  }

  writer.writeEndElement("te_map:Chart");
}

void te::map::serialize::WriteLayerGrouping(te::map::Grouping* g, te::xml::AbstractWriter& writer)
{
  writer.writeStartElement("te_map:Grouping");

  te::map::GroupingType type = g->getType();

  writer.writeElement("te_map:PropertyName", g->getPropertyName());
  writer.writeElement("te_map:PropertyDataType", g->getPropertyType());
  writer.writeElement("te_map:Type", GetGroupingType(type));
  writer.writeElement("te_map:Precision", static_cast<unsigned int>(g->getPrecision()));
  writer.writeElement("te_map:NumSlices", static_cast<unsigned int>(g->getNumSlices()));

  writer.writeElement("te_map:Summary", g->getSummary());

  if(type == te::map::STD_DEVIATION)
    writer.writeElement("te_map:StandardDeviation", g->getStdDeviation());

  writer.writeEndElement("te_map:Grouping");
}

void te::map::serialize::WriteLayerRasterContrast(te::map::RasterContrast* rc, te::xml::AbstractWriter& writer)
{
  writer.writeStartElement("te_map:RasterContrast");

  writer.writeElement("te_map:ContrastType", static_cast<int>(rc->getType()));
  writer.writeElement("te_map:NBands", static_cast<int>(rc->getNBands()));

  std::size_t nBands = rc->getNBands();
  std::vector<double> gain, offset1, offset2, min, max;

  rc->getValues(gain, offset1, offset2, min, max);

  for (std::size_t t = 0; t < nBands; ++t)
  {
    writer.writeStartElement("te_map:BandContrast");

    writer.writeElement("te_map:GainContrast", gain[t]);
    writer.writeElement("te_map:Offset1Contrast", offset1[t]);
    writer.writeElement("te_map:Offset2Contrast", offset2[t]);
    writer.writeElement("te_map:minContrast", min[t]);
    writer.writeElement("te_map:maxContrast", max[t]);

    writer.writeEndElement("te_map:BandContrast");
  }

  writer.writeEndElement("te_map:RasterContrast");
}

void te::map::serialize::WriteAbstractLayer(const te::map::AbstractLayer* layer, te::xml::AbstractWriter& writer)
{
  writer.writeAttribute("id", layer->getId());
  writer.writeElement("te_map:Title", te::core::URIEncode(layer->getTitle()));
  writer.writeElement("te_map:Visible", GetVisibility(layer->getVisibility()));
  writer.writeElement("te_map:Encoding", te::core::CharEncoding::getEncodingName(layer->getEncoding()));


  //grouping
  te::map::Grouping* g = layer->getGrouping();

  if(g != nullptr)
    WriteLayerGrouping(g, writer);

  //chart
  te::map::Chart* chart = layer->getChart();

  if(chart != nullptr)
    WriteLayerChart(chart, writer);

  //raster contrast
  te::map::RasterContrast* rc = layer->getRasterContrast();

  if (rc != nullptr)
    WriteLayerRasterContrast(rc, writer);
}

void te::map::serialize::WriteSRIDValue(const int& srid, te::xml::AbstractWriter& writer)
{
  std::string proj4 = te::srs::SpatialReferenceSystemManager::getInstance().getP4Txt(static_cast<unsigned int>(srid));

  writer.writeStartElement("te_map:SRSInfo");

  writer.writeElement("te_map:SRID", srid);

  writer.writeElement("te_map:Proj4", proj4);

  writer.writeEndElement("te_map:SRSInfo");
}

