/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/memory/DataSetItem.cpp

  \brief An implementation of the DatasetItem class for the TerraLib In-Memory Data Access driver.
*/

// TerraLib
#include "../dataaccess/dataset/DataSetType.h"
#include "../dataaccess/utils/Utils.h"
#include "../datatype/AbstractData.h"
#include "../datatype/ByteArray.h"
#include "../datatype/DataConverterManager.h"
#include "../datatype/DateTime.h"
#include "../datatype/Property.h"
#include "../datatype/SimpleData.h"
#include "../geometry/Geometry.h"
#include "../raster/RasterProperty.h"
#include "DataSet.h"
#include "DataSetItem.h"
#include "PreparedItemCopy.h"
#include "Raster.h"

// STL
#include <cassert>
#include <iterator>

te::mem::DataSetItem::DataSetItem(const std::vector<std::string>& vecNames, const std::vector<int>& vecTypes)
  : te::da::DataSetItem(vecNames, vecTypes)
{

}

te::mem::DataSetItem::DataSetItem(const te::da::DataSet* dataSet)
  : te::da::DataSetItem(dataSet)
{
}

te::mem::DataSetItem::DataSetItem(const te::da::DataSetType* dataSetType)
  : te::da::DataSetItem(dataSetType)
{
  
}

te::mem::DataSetItem::DataSetItem(const te::mem::DataSetItem& rhs)
  : te::da::DataSetItem(rhs)
{
}

te::mem::DataSetItem::~DataSetItem() = default;

te::mem::DataSetItem& te::mem::DataSetItem::operator=(const DataSetItem& rhs)
{
  assert(getNumProperties() == rhs.getNumProperties());

  if (this != &rhs)
  {
    m_vecNames = rhs.m_vecNames;
    m_vecTypes = rhs.m_vecTypes;
    m_data = rhs.m_data;
  }

  return *this;
}

te::da::DataSetItem* te::mem::DataSetItem::clone() const
{
  return new DataSetItem(*this);
}

void te::mem::DataSetItem::setChar(std::size_t i, char value) 
{
  m_data.replace(i, new te::dt::Char(value));
}

void te::mem::DataSetItem::setChar(const std::string& name, char value)
{
  std::size_t pos = getPropertyPos(name);
  setChar(pos, value);
}

void te::mem::DataSetItem::setUChar(std::size_t i, unsigned char value)
{
  m_data.replace(i, new te::dt::UChar(value));
}

void te::mem::DataSetItem::setUChar(const std::string& name, unsigned char value)
{
  std::size_t pos = getPropertyPos(name);
  setUChar(pos, value);
}

void te::mem::DataSetItem::setInt16(std::size_t i, boost::int16_t value)
{
  m_data.replace(i, new te::dt::Int16(value));
}

void te::mem::DataSetItem::setInt16(const std::string& name, boost::int16_t value)
{
  std::size_t pos = getPropertyPos(name);
  setInt16(pos, value);
}

void te::mem::DataSetItem::setInt32(std::size_t i, boost::int32_t value)
{
  m_data.replace(i, new te::dt::Int32(value));
}

void te::mem::DataSetItem::setInt32(const std::string& name, boost::int32_t value)
{
  std::size_t pos = getPropertyPos(name);
  setInt32(pos, value);
}

void te::mem::DataSetItem::setInt64(std::size_t i, boost::int64_t value)
{
  m_data.replace(i, new te::dt::Int64(value));
}

void te::mem::DataSetItem::setInt64(const std::string& name, boost::int64_t value)
{
  std::size_t pos = getPropertyPos(name);
  setInt64(pos, value);
}

void te::mem::DataSetItem::setBool(std::size_t i, bool value) 
{
  m_data.replace(i, new te::dt::Boolean(value));
}

void te::mem::DataSetItem::setBool(const std::string& name, bool value) 
{
  std::size_t pos = getPropertyPos(name);
  setBool(pos, value);
}

void te::mem::DataSetItem::setFloat(std::size_t i, float value) 
{
  m_data.replace(i, new te::dt::Float(value));
}

void te::mem::DataSetItem::setFloat(const std::string& name, float value) 
{
  std::size_t pos = getPropertyPos(name);
  setFloat(pos, value);
}

void te::mem::DataSetItem::setDouble(std::size_t i, double value) 
{
  m_data.replace(i, new te::dt::Double(value));
}

void te::mem::DataSetItem::setDouble(const std::string& name, double value) 
{
  std::size_t pos = getPropertyPos(name);
  setDouble(pos, value);
}

void te::mem::DataSetItem::setNumeric(std::size_t i, const std::string& value) 
{
  m_data.replace(i, new te::dt::Numeric(value));
}

void te::mem::DataSetItem::setNumeric(const std::string& name, const std::string& value) 
{
  std::size_t pos = getPropertyPos(name);
  setNumeric(pos, value);
}

void te::mem::DataSetItem::setString(std::size_t i, const std::string& value) 
{
  m_data.replace(i, new te::dt::String(value));
}

void te::mem::DataSetItem::setString(const std::string& name, const std::string& value) 
{
  std::size_t pos = getPropertyPos(name);
  setString(pos, value);
}

void te::mem::DataSetItem::setByteArray(std::size_t i, te::dt::ByteArray* value) 
{
  m_data.replace(i, value);
}

void te::mem::DataSetItem::setByteArray(const std::string& name, te::dt::ByteArray* value) 
{
  std::size_t pos = getPropertyPos(name);
  setByteArray(pos, value);
}

void te::mem::DataSetItem::setGeometry(std::size_t i, te::gm::Geometry* value)
{
  m_data.replace(i, value);
}

void te::mem::DataSetItem::setGeometry(const std::string& name, te::gm::Geometry* value)
{
  std::size_t pos = getPropertyPos(name);
  setGeometry(pos, value);
}

void te::mem::DataSetItem::setRaster(std::size_t i, te::rst::Raster* value)
{
  m_data.replace(i, value);
}

void te::mem::DataSetItem::setRaster(const std::string& name, te::rst::Raster* value)
{
  std::size_t pos = getPropertyPos(name);
  setRaster(pos, value);
}

void te::mem::DataSetItem::setDateTime(std::size_t i, te::dt::DateTime* value) 
{
  m_data.replace(i, value);
}

void te::mem::DataSetItem::setDateTime(const std::string& name, te::dt::DateTime* value)
{
  std::size_t pos = getPropertyPos(name);
  setDateTime(pos, value);
}

void te::mem::DataSetItem::setValue(std::size_t i, te::dt::AbstractData* value)
{
  if (value != nullptr)
  {
    //if the datatype of the given value is different from the expected datatype, we try to convert it
    if (value->getTypeCode() != m_vecTypes.at(i))
    {
      int srcType = value->getTypeCode();
      int destType = m_vecTypes.at(i);
      try
      {
        const te::dt::DataTypeConverter& converter = te::dt::DataConverterManager::getInstance().get(srcType, destType);
        te::dt::AbstractData* convertedValue = converter(value);
        delete value;
        value = convertedValue;
      }
      catch (const std::exception&)
      {
        throw;
      }
    }
  }

  m_data.replace(i, value);
}

void te::mem::DataSetItem::setValue(const std::string& name, te::dt::AbstractData* value)
{
  std::size_t pos = getPropertyPos(name);
  setValue(pos, value);
} 

void te::mem::DataSetItem::copyValuesFrom(const te::da::DataSetItem* source, bool copyGeometries)
{
  assert(source);

  PreparedItemCopy copy;
  copy.copyValues(source, this, copyGeometries);
}
