 /*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/memory/DataSetItem.h

  \brief An implementation of the DatasetItem class for the TerraLib In-Memory Data Access driver.
*/

#ifndef __TERRALIB_MEMORY_INTERNAL_DATASETITEM_H
#define __TERRALIB_MEMORY_INTERNAL_DATASETITEM_H

// TerraLib
#include "Config.h"

#include "../dataaccess/dataset/DataSetItem.h"

// Forward declarations
namespace te { namespace dt { class AbstractData;  class Array; class ByteArray; class DateTime; } }
namespace te { namespace gm { class Geometry; } }
namespace te { namespace rst { class Raster; } }

namespace te
{
  namespace da
  {
// Forward declaration
    class DataSet;
  }

  namespace mem
  {
    /*!
      \class DataSetItem

      \brief An implementation of the DatasetItem class for the TerraLib In-Memory Data Access driver.
    */
    class TEMEMORYEXPORT DataSetItem : public te::da::DataSetItem
    {
      public:

        explicit DataSetItem(const std::vector<std::string>& vecNames, const std::vector<int>& vecTypes);

        /*!
          \brief It creates a new item having the same schema as the parent dataset.

          \param parent The dataset to which this item is associated.

          \note The dataset item will NOT take the ownership of the given pointer.
        */
        explicit DataSetItem(const te::da::DataSet* dataSet);

        /*!
          \brief It creates a new item having the same schema as the dataset type.

          \param parent The dataset to which this item is associated.

          \note The dataset item will NOT take the ownership of the given pointer.
        */
        explicit DataSetItem(const te::da::DataSetType* dataSetType);

        /*!
        \brief It creates a new item by cloning the values in the source item (rhs).

        \param rhs The right-hand-side object to copy its values.
      */
        explicit DataSetItem(const DataSetItem& rhs);

        /*! \brief Destructor. */
        virtual ~DataSetItem();

        /*!
          \brief Assignment operator.

          It copies the values from the rhs item.

          \param rhs The right-hand-side ibject to copy its values.

          \return A reference to this item.

          \pre The rhs item must be of a compatible type with the item being assigned.
        */

        DataSetItem& operator=(const DataSetItem& rhs);

        /*!
          \brief It returns a clone of the DataSetItem.
        */
        virtual te::da::DataSetItem* clone() const override;

        /*!
          \brief It sets the value of the i-th property.
        */
        void setChar(std::size_t i, char value);

        /*!
          \brief It sets the value of the property, indicating its name. 
        */
        void setChar(const std::string& name, char value);

        /*!
          \brief It sets the value of the i-th property.
        */
        void setUChar(std::size_t i, unsigned char value);

        /*!
          \brief It sets the value of the property, indicating its name. 
        */
        void setUChar(const std::string& name, unsigned char value);

        /*!
          \brief It sets the value of the i-th property.
        */
        void setInt16(std::size_t i, boost::int16_t value);

        /*!
          \brief It sets the value of the property, indicating its name. 
        */
        void setInt16(const std::string& name, boost::int16_t value);

        /*!
          \brief It sets the value of the i-th property.
        */
        void setInt32(std::size_t i, boost::int32_t value);

        /*!
          \brief It sets the value of the property, indicating its name. 
        */
        void setInt32(const std::string& name, boost::int32_t value);

        /*!
          \brief It sets the value of the i-th property.
        */
        void setInt64(std::size_t i, boost::int64_t value);

        /*!
          \brief It sets the value of the property, indicating its name. 
        */
        void setInt64(const std::string& name, boost::int64_t value);

        /*!
          \brief It sets the value of the i-th property.
        */
        void setBool(std::size_t i, bool value);

        /*!
          \brief It sets the value of the property, indicating its name. 
        */
        void setBool(const std::string& name, bool value);


        /*!
          \brief It sets the value of the i-th property.
        */
        void setFloat(std::size_t i, float value);

        /*!
          \brief It sets the value of the property, indicating its name. 
        */
        void setFloat(const std::string& name, float value);

        /*!
          \brief It sets the value of the i-th property.
        */
        void setDouble(std::size_t i, double value);

        /*!
          \brief It sets the value of the property, indicating its name. 
        */
        void setDouble(const std::string& name, double value);

        /*!
          \brief It sets the value of the i-th property.
        */
        void setNumeric(std::size_t i, const std::string& value);

        /*!
          \brief It sets the value of the property, indicating its name. 
        */
        void setNumeric(const std::string& name, const std::string& value);

        /*!
          \brief It sets the value of the i-th property.
        */
        void setString(std::size_t i, const std::string& value);

        /*!
          \brief It sets the value of the property, indicating its name. 
        */
        void setString(const std::string& name, const std::string& value);

        /*!
          \brief It sets the value of the i-th property.

          \note It will take the ownership of the given pointer.
        */
        void setByteArray(std::size_t i, te::dt::ByteArray* value);

        /*!
          \brief It sets the value of the property, indicating its name. 

          \note It will take the ownership of the given pointer.
        */
        void setByteArray(const std::string& name, te::dt::ByteArray* value);

        /*!
          \brief It sets the value of the i-th property.

          \note It will take the ownership of the given pointer.
        */
        void setGeometry(std::size_t i, te::gm::Geometry* value);

        /*!
          \brief It sets the value of the property, indicating its name. 

          \note It will take the ownership of the given pointer.
        */
        void setGeometry(const std::string& name, te::gm::Geometry* value);

        /*!
          \brief It sets the value of the i-th property.

          \note It will take the ownership of the given pointer.
        */
        void setRaster(std::size_t i, te::rst::Raster* value);

        /*!
          \brief It sets the value of the property, indicating its name. 
        */
        void setRaster(const std::string& name, te::rst::Raster* value);

        /*!
          \brief It sets the value of the i-th property.

          \note It will take the ownership of the given pointer.
        */
        void setDateTime(std::size_t i, te::dt::DateTime* value);

        /*!
          \brief It sets the value of the property, indicating its name. 

          \note It will take the ownership of the given pointer.
        */
        void setDateTime(const std::string& name, te::dt::DateTime* value);

        /*!
          \brief It sets the value of the i-th property.

          \note It will take the ownership of the given pointer.
        */
        void setValue(std::size_t i, te::dt::AbstractData* value);

        /*!
          \brief It sets the value of the property, indicating its name. 

          \note It will take the ownership of the given pointer.
        */
        void setValue(const std::string& name, te::dt::AbstractData* value);

        /*!
        \brief Copies the values from the source dataSet item to the destination dataSet item by matching the column names. If a column from the source does not exist in the destination, it will not be copied
      */
        void copyValuesFrom(const te::da::DataSetItem* source, bool copyGeometries = true);
    };
  } // end namespace mem
}   // end namespace te

#endif  // __TERRALIB_MEMORY_INTERNAL_DATASETITEM_H
