 /*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/memory/PreparedItemCopy.h

  \brief This class is used to optimize the copy of several dataSetItems by caching the mapping between the columns from the input and the output
*/

#include "PreparedItemCopy.h"

#include "../datatype/AbstractData.h"
#include "../datatype/Enums.h"
#include "DataSetItem.h"


te::mem::PreparedItemCopy::PreparedItemCopy()
{

}

te::mem::PreparedItemCopy::~PreparedItemCopy()
{

}

void te::mem::PreparedItemCopy::reset()
{
  m_columnMapping.clear();
}

void te::mem::PreparedItemCopy::copyValues(const te::da::DataSetItem* source, te::mem::DataSetItem* destination, bool copyGeometries) const
{
  if (m_columnMapping.empty() == true)
  {
    m_columnMapping = prepareCopyValues(source, destination, copyGeometries);
  }

  copyValues(source, destination, m_columnMapping);
}

te::mem::ColumnMapping te::mem::PreparedItemCopy::prepareCopyValues(const te::da::DataSetItem* source, te::mem::DataSetItem* destination, bool copyGeometries) const
{
  assert(source);
  assert(destination);

  ColumnMapping columnMapping;

  const std::vector<int>& vecSourceDataTypes = source->getPropertyTypes();
  std::size_t lastFoundGeometryPropertyIndex = 0;
  for (std::size_t outputPos = 0; outputPos < destination->getNumProperties(); ++outputPos)
  {
    std::size_t inputPos = std::string::npos;

    if (destination->getPropertyDataType(outputPos) == te::dt::GEOMETRY_TYPE)
    {
      if (copyGeometries == false)
      {
        continue;
      }

      //this part of the algorithm must be done this way becausa the geometry columns may have different names depending on the dataSource. 
      std::vector<int>::const_iterator it = std::find(vecSourceDataTypes.begin() + lastFoundGeometryPropertyIndex, vecSourceDataTypes.end(), te::dt::GEOMETRY_TYPE);
      if (it != vecSourceDataTypes.end())
      {
        lastFoundGeometryPropertyIndex = std::distance(vecSourceDataTypes.begin(), it);

        inputPos = lastFoundGeometryPropertyIndex;

        ++lastFoundGeometryPropertyIndex;
      }
    }
    else
    {
      std::string columnName = destination->getPropertyName(outputPos);
      inputPos = source->getPropertyPos(columnName);
    }

    if (inputPos == std::string::npos)
    {
      continue;
    }

    std::pair<std::size_t, std::size_t> columnMap;
    columnMap.first = inputPos;
    columnMap.second = outputPos;

    columnMapping.push_back(columnMap);
  }

  return columnMapping;
}

void te::mem::PreparedItemCopy::copyValues(const te::da::DataSetItem* source, te::mem::DataSetItem* destination, const ColumnMapping& vecColumnsMapping) const
{
  assert(source);
  assert(destination);

  for (std::size_t i = 0; i < vecColumnsMapping.size(); ++i)
  {
    std::size_t inputPos = vecColumnsMapping.at(i).first;
    std::size_t outputPos = vecColumnsMapping.at(i).second;

    if (source->isNull(inputPos) == true)
    {
      continue;
    }
    destination->setValue(outputPos, source->getValue(inputPos).release());
  }
}