 /*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/memory/PreparedItemCopy.h

  \brief This class is used to optimize the copy of several dataSetItems by caching the mapping between the columns from the input and the output
*/

#ifndef __TERRALIB_MEMORY_INTERNAL_PREPAREDITEMCOPY_H
#define __TERRALIB_MEMORY_INTERNAL_PREPAREDITEMCOPY_H

// TerraLib
#include "Config.h"

#include <utility>
#include <vector>

namespace te
{
  namespace da
  {
// Forward declaration
    class DataSetItem;
  }

  namespace mem
  {
    using ColumnMapping = std::vector< std::pair<std::size_t, std::size_t> >;

    class DataSetItem;

    /*!
      \class PreparedItemCopy

      \brief This class is used to optimize the copy of several dataSetItems by caching the mapping between the columns from the input and the output
    */
    class TEMEMORYEXPORT PreparedItemCopy
    {
      public:

        PreparedItemCopy();

        ~PreparedItemCopy();

        void reset();

        /*!
        \brief Copies the values from the source dataSet item to the destination dataSet item by matching the column names. If a column from the source does not exist in the destination, it will not be copied
        */
        void copyValues(const te::da::DataSetItem* source, te::mem::DataSetItem* destination, bool copyGeometries = true) const;

    protected:
      
      /*!
        \brief Creates a list mapping the input columns to the output columns. This is used to optmize significantly when you need to copy several items with the same column order
        */
      ColumnMapping prepareCopyValues(const te::da::DataSetItem* source, te::mem::DataSetItem* destination, bool copyGeometries) const;

      /*!
        \brief Copies the values from the source dataSet item to the destination dataSet item using the given pre calculated vecColumnsMapping.
        */
      void copyValues(const te::da::DataSetItem* source, te::mem::DataSetItem* destination, const ColumnMapping& vecColumnsMapping) const;

    protected:

      mutable ColumnMapping m_columnMapping;
    };
  } // end namespace mem
}   // end namespace te

#endif  // __TERRALIB_MEMORY_INTERNAL_PREPAREDITEMCOPY_H
