/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/ogr/SQLVisitor.cpp

  \brief A visitor for building an SQL statement using OGR dialect.
*/

// TerraLib
#include "../common/StringUtils.h"
#include "../geometry/Envelope.h"
#include "../dataaccess/query/DataSetName.h"
#include "../dataaccess/query/Function.h"
#include "../dataaccess/query/LiteralString.h"
#include "../dataaccess/query/Having.h"
#include "../dataaccess/query/In.h"
#include "../dataaccess/query/Join.h"
#include "../dataaccess/query/JoinConditionOn.h"
#include "../dataaccess/query/LiteralBool.h"
#include "../dataaccess/query/LiteralEnvelope.h"
#include "../dataaccess/query/LiteralGeom.h"
#include "../dataaccess/query/LiteralDateTime.h"
#include "../dataaccess/query/PropertyName.h"
#include "../dataaccess/query/Select.h"
#include "../dataaccess/query/Where.h"
#include "../dataaccess/Enums.h"
#include "../datatype/AbstractData.h"
#include "../datatype/Enums.h"
#include "../datatype/SimpleData.h"
#include "../datatype/DateTime.h"
#include "SQLVisitor.h"

//Boost include files
#include <boost/algorithm/string/replace.hpp>

// STL
#include <cassert>
#include <string>
#include <vector>

te::ogr::SQLVisitor::SQLVisitor(const te::da::SQLDialect& dialect, std::string& sql)
  : te::da::SQLVisitor(dialect, sql),
    m_bbox(nullptr)
{
}

void te::ogr::SQLVisitor::visit(const te::da::DataSetName& visited)
{
  m_sql += "\"" + visited.getName() + "\"";

  const std::string& alias = visited.getAlias();
  if (alias.empty() == false)
  {
    m_sql += " " + alias;
  }
}

void te::ogr::SQLVisitor::visit(const te::da::LiteralEnvelope& visited)
{
  m_bbox = visited.getValue();

  const te::gm::Envelope* envelope = visited.getValue();

  m_sql += "BuildMBR";
  m_sql += "(" + te::common::Convert2String(envelope->getLowerLeftX(), 15);
  m_sql += "," + te::common::Convert2String(envelope->getLowerLeftY(), 15);
  m_sql += "," + te::common::Convert2String(envelope->getUpperRightX(), 15);
  m_sql += "," + te::common::Convert2String(envelope->getUpperRightY(), 15);
  m_sql += ")";
}

void te::ogr::SQLVisitor::visit(const te::da::PropertyName& visited)
{
  //ogr can handle full names. So we just must quote them separately
  std::string value = visited.getName();
  if (value != "*")
  {
    std::vector<std::string> vecTokens;
    te::common::Tokenize(value, vecTokens, ".");

    value = "\"" + vecTokens.at(0) + "\"";
    for (std::size_t i = 1; i < vecTokens.size(); ++i)
    {
      if (i != 0)
      {
        value += ".";
      }
      std::string currentValue = vecTokens.at(i);
      if (currentValue != "*")
      {
        currentValue = "\"" + currentValue + "\"";
      }

      value += currentValue;
    }
  }

  m_sql += value;
}

void te::ogr::SQLVisitor::visit(const te::da::Select& visited)
{
  m_sql += "SELECT ";

  if(visited.getDistinct())
  {
    visitDistinct(*(visited.getDistinct()));
    m_sql += " ";
  }

  if(visited.getFields())
  {
    te::da::SQLVisitor::visit(*(visited.getFields()));
    m_sql += " ";
  }


  if(visited.getFrom())
  {
    te::da::SQLVisitor::visit(*(visited.getFrom()));
    m_sql += " ";
  }

  if(visited.getWhere())
  {
    m_sql += "WHERE ";
    visited.getWhere()->getExp()->accept(*this);
    m_sql += " ";
  }

  if(visited.getGroupBy())
  {
    te::da::SQLVisitor::visit(*(visited.getGroupBy()));
    m_sql += " ";
  }

  if(visited.getHaving())
  {
    m_sql += "HAVING ";
    visited.getHaving()->getExp()->accept(*this);
    m_sql += " ";
  }

  if(visited.getOrderBy())
    te::da::SQLVisitor::visit(*(visited.getOrderBy()));
}

void te::ogr::SQLVisitor::visit(const te::da::Join& visited)
{
  assert(visited.getFirst() && visited.getSecond());

  visited.getFirst()->accept(*this);

  if(visited.isNatural())
    m_sql += " NATURAL ";

  switch(visited.getType())
  {
    case te::da::JOIN :
      m_sql += " JOIN ";
    break;

    case te::da::INNER_JOIN :
      m_sql += " INNER JOIN ";
    break;

    case te::da::LEFT_JOIN :
      m_sql += " LEFT JOIN ";
    break;

    case te::da::RIGHT_JOIN :
      m_sql += " RIGHT JOIN ";
    break;

    case te::da::FULL_OUTER_JOIN :
      m_sql += " FULL OUTER JOIN ";
    break;

    case te::da::CROSS_JOIN :
      m_sql += " CROSS JOIN ";
    break;

    case te::da::NATURAL_JOIN :
      m_sql += " NATURAL JOIN ";
    break;
  }

  visited.getSecond()->accept(*this);

  if(visited.getCondition())
  {
    m_sql += " ";
    visited.getCondition()->accept(*this);
  }
}

void te::ogr::SQLVisitor::visit(const te::da::JoinConditionOn& visited)
{
  assert(visited.getCondition());  
  m_sql += "ON ";

  te::da::Function func = dynamic_cast< te::da::Function&>(*visited.getCondition());
  const std::string& fname = func.getName();

  assert(func.getNumArgs() == 2);
  func[0]->accept(*this);
  m_sql += " ";
  m_sql += fname;
  m_sql += " ";
  func[1]->accept(*this);
}

void te::ogr::SQLVisitor::visit(const te::da::LiteralBool& visited)
{
  te::dt::AbstractData* absData = visited.getValue();
  if (absData == nullptr)
  {
    return;
  }

  std::string strBoolValue = "0";
  if (absData->getTypeCode() == te::dt::BOOLEAN_TYPE)
  {
    te::dt::Boolean* value = static_cast<te::dt::Boolean*>(absData);
    if (value->getValue() == true)
    {
      strBoolValue = "1";
    }
  }
  else if (absData->getTypeCode() == te::dt::BOOLEAN_TYPE)
  {
    te::dt::Int32* value = static_cast<te::dt::Int32*>(absData);
    if (value->getValue() != 0)
    {
      strBoolValue = "1";
    }
  }

  m_sql += strBoolValue;
}

void te::ogr::SQLVisitor::visit(const te::da::LiteralString& visited)
{
  if (visited.getValue())
  {
    std::string strEscape(visited.getValue()->toString());
    boost::replace_all(strEscape, "'", "''");

    m_sql += "'";
    m_sql += strEscape;
    m_sql += "'";
  }
}

void te::ogr::SQLVisitor::visit(const te::da::LiteralGeom& visited)
{
  if (visited.getValue())
    m_sql += "GeomFromText('" + visited.getValue()->toString() + "')";
}

void te::ogr::SQLVisitor::visit(const te::da::LiteralDateTime& visited)
{
  te::dt::DateTime const* dtPtr = ( te::dt::DateTime const* )
    visited.getValue();

  if( dtPtr != nullptr )
  {
    std::string strEscape(visited.getValue()->toString());
    boost::replace_all(strEscape, "'", "''");

    switch( dtPtr->getDateTimeType() )
    {
      case te::dt::DATE :
        m_sql += "cast( '" + strEscape + "' as date )";
        break;
      case te::dt::TIME_DURATION :
        m_sql += "cast( '" + strEscape + "' as time )";
        break;
      case te::dt::TIME_INSTANT :
      case te::dt::TIME_INSTANT_TZ :
        m_sql += "cast( '" + strEscape + "' as timestamp )";
        break;
      default :
        m_sql += "'" + strEscape + "'";
        break;
    }
  }
}

void te::ogr::SQLVisitor::visit(const te::da::In& visited)
{
  //to avoid 'memory exhauted' OGR error, we must split the in-clauses that contains to values
  const std::size_t BUFFER_SIZE = 25000;
  std::size_t lastSQLSize = m_sql.size();
  
  for (size_t i = 0; i < visited.getNumArgs(); ++i)
  {
    if( i == 0 )
    {
      assert(visited.getPropertyName());
      visited.getPropertyName()->accept(*this);
      
      m_sql += " IN(";
    }
    else
    {
      if( ( m_sql.size() - lastSQLSize ) >= BUFFER_SIZE )
      {
        lastSQLSize = m_sql.size();
        
        m_sql += ") OR ";
        
        assert(visited.getPropertyName());
        visited.getPropertyName()->accept(*this);        

        m_sql += " IN(";
      }
      else
      {
        m_sql += ",";
      }
    }

    assert(visited[i]);
    visited[i]->accept(*this);
  }

  m_sql += ")";
}

te::gm::Envelope* te::ogr::SQLVisitor::getMBR()
{
  return m_bbox;
}

