/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

// TerraLib
#include "../common/progress/TaskProgress.h"
#include "../common/StringUtils.h"
#include "../core/logger/Logger.h"
#include "../core/translator/Translator.h"
#include "../core/uri/URI.h"
#include "../core/uri/Utils.h"
#include "../dataaccess/dataset/ObjectId.h"
#include "../dataaccess/dataset/ObjectIdSet.h"
#include "../dataaccess/query/DataSetName.h"
#include "../dataaccess/query/Query.h"
#include "../dataaccess/query/Select.h"
#include "../dataaccess/utils/FieldNameValidator.h"
#include "../dataaccess/utils/Utils.h"
#include "../datatype/ByteArray.h"
#include "../datatype/Date.h"
#include "../datatype/TimeInstant.h"
#include "../datatype/TimeInstantTZ.h"
#include "../geometry/Envelope.h"
#include "../geometry/GeometryProperty.h"
#include "../srs/SpatialReferenceSystemManager.h"
#include "../srs/Config.h"
#include "DataSource.h"
#include "DataSet.h"
#include "SQLVisitor.h"
#include "Transactor.h"
#include "Utils.h"

//BOOST
#include <boost/thread/locks.hpp>

// OGR
#include <ogrsf_frmts.h>

OGRFieldType GetOGRType(int te_type)
{
  switch (te_type)
  {
    case te::dt::CHAR_TYPE:
    case te::dt::UCHAR_TYPE:
    case te::dt::STRING_TYPE:
      return OFTString;

    case te::dt::INT16_TYPE:
    case te::dt::UINT16_TYPE:
    case te::dt::INT32_TYPE:
    case te::dt::UINT32_TYPE:
      return OFTInteger;

    case te::dt::INT64_TYPE:
    case te::dt::UINT64_TYPE:
      return OFTInteger64;

    case te::dt::FLOAT_TYPE:
    case te::dt::DOUBLE_TYPE:
    case te::dt::NUMERIC_TYPE:
      return OFTReal;

    case te::dt::DATETIME_TYPE:
      return OFTDateTime;
  };

  return OFTInteger;
}

struct ScopedOGRLayer
{
  ScopedOGRLayer(GDALDataset* gdalDataSet, OGRLayer* ogrLayer, bool releaseOnDestruction)
    : m_gdalDataSet(gdalDataSet)
    , m_ogrLayer(ogrLayer)
    , m_release(releaseOnDestruction)
  {

  }

  ScopedOGRLayer(ScopedOGRLayer&& other)
    : m_gdalDataSet(other.m_gdalDataSet)
    , m_ogrLayer(other.m_ogrLayer)
    , m_release(other.m_release)
  {
    other.m_gdalDataSet = nullptr;
    other.m_ogrLayer = nullptr;
    other.m_release = false;
  }

  ~ScopedOGRLayer()
  {
    if (m_ogrLayer != nullptr && m_release == true)
    {
      m_gdalDataSet->ReleaseResultSet(m_ogrLayer);
    }
  }

  OGRLayer* operator->() const
  {
    return m_ogrLayer;
  }

  OGRLayer* release()
  {
    OGRLayer* returnOGRLayer = m_ogrLayer;

    m_ogrLayer = nullptr;

    return returnOGRLayer;
  }

  ScopedOGRLayer(const ScopedOGRLayer&) = delete;
  ScopedOGRLayer& operator=(const ScopedOGRLayer&) = delete;

  GDALDataset* m_gdalDataSet;
  OGRLayer* m_ogrLayer;
  bool m_release;
};

//!< Local function to handle the way of accessing OGRLayer depending on the current OGRDriver
ScopedOGRLayer GetLayer(GDALDataset* gDataSet, const std::string& dataSetName)
{
  OGRLayer* ogrLayer = gDataSet->GetLayerByName(dataSetName.c_str());
  ScopedOGRLayer scoperLayer(gDataSet, ogrLayer, false);
  return std::move(scoperLayer);
}

te::ogr::Transactor::Transactor(DataSource* ds) : m_ogrDs(ds), m_fid(0), m_transactionCount(0), m_putFIDAttributeOption("putFIDAttribute")
{
}

te::ogr::Transactor::~Transactor()
{
}

te::da::DataSource* te::ogr::Transactor::getDataSource() const
{
  return m_ogrDs;
}

void te::ogr::Transactor::begin()
{
  m_ogrDs->begin();
}

void te::ogr::Transactor::commit()
{
  m_ogrDs->commit();
}

void te::ogr::Transactor::rollBack()
{
  m_ogrDs->rollBack();
}

bool te::ogr::Transactor::isInTransaction() const
{
  return m_ogrDs->isInTransaction();
}

std::unique_ptr<te::da::DataSet> te::ogr::Transactor::getDataSet(const std::string& name,
                                          te::common::TraverseType /*travType*/,
                                          bool /*connected*/,
                                          const te::common::AccessPolicy pol)
{
  boost::unique_lock< boost::mutex > lockGuard(getStaticMutex());

  if (!m_ogrDs->getOGRDataSource())
  {
    lockGuard.release();
    getStaticMutex().unlock();
    return std::unique_ptr<te::da::DataSet>(nullptr);
  }

  CPLSetConfigOption("SHAPE_ENCODING", te::core::CharEncoding::getEncodingName(m_ogrDs->getEncoding()).c_str());

  bool releaseDataSet = false;
  GDALDataset* ds = nullptr;
  if (this->isInTransaction() == true)
  {
    ds = m_ogrDs->getOGRDataSource();
  }
  else
  {
    releaseDataSet = true;
    unsigned int ogropenflags = (pol == te::common::RWAccess || pol == te::common::WAccess) ? GDAL_OF_UPDATE : GDAL_OF_READONLY;
    ds = CreateGDALDataSet(m_ogrDs->getOGRDataSource()->GetDescription(), ogropenflags);
  }
  if (!ds)
  {
    CPLSetConfigOption("SHAPE_ENCODING", te::core::CharEncoding::getEncodingName(te::core::EncodingType::UTF8).c_str());
    throw Exception(TE_TR("Could not retrieve the data source."));
  }

  ScopedOGRLayer scopedLayer = GetLayer(ds, name);

  CPLSetConfigOption("SHAPE_ENCODING", te::core::CharEncoding::getEncodingName(te::core::EncodingType::UTF8).c_str());

  lockGuard.release();
  getStaticMutex().unlock();

  if(scopedLayer.m_ogrLayer == nullptr)
    throw Exception(TE_TR("The informed data set could not be found in the data source."));

  return std::unique_ptr<te::da::DataSet>(new DataSet(ds, scopedLayer.release(), releaseDataSet, scopedLayer.m_release));
}

std::unique_ptr<te::da::DataSet> te::ogr::Transactor::getDataSet(const std::string& name,
                                          const std::string& /*propertyName*/,
                                          const te::gm::Envelope* e,
                                          te::gm::SpatialRelation /*r*/,
                                          te::common::TraverseType /*travType*/,
                                          bool /*connected*/,
                                          const te::common::AccessPolicy pol)
{
  boost::unique_lock< boost::mutex > lockGuard(getStaticMutex());

  if (!m_ogrDs->getOGRDataSource())
  {
    lockGuard.release();
    getStaticMutex().unlock();
    return std::unique_ptr<te::da::DataSet>(nullptr);
  }

  CPLSetConfigOption("SHAPE_ENCODING", te::core::CharEncoding::getEncodingName(m_ogrDs->getEncoding()).c_str());

  bool releaseDataSet = false;
  GDALDataset* ds = nullptr;
  if (this->isInTransaction() == true)
  {
    ds = m_ogrDs->getOGRDataSource();
  }
  else
  {
    releaseDataSet = true;
    unsigned int ogropenflags = (pol == te::common::RWAccess || pol == te::common::WAccess) ? GDAL_OF_UPDATE : GDAL_OF_READONLY;
    ds = CreateGDALDataSet(m_ogrDs->getOGRDataSource()->GetDescription(), ogropenflags);
  }

  if (!ds)
  {
    CPLSetConfigOption("SHAPE_ENCODING", te::core::CharEncoding::getEncodingName(te::core::EncodingType::UTF8).c_str());
    throw Exception(TE_TR("Could not retrieve the data source."));
  }

  ScopedOGRLayer scopedLayer = GetLayer(ds, name);

  CPLSetConfigOption("SHAPE_ENCODING", te::core::CharEncoding::getEncodingName(te::core::EncodingType::UTF8).c_str());

  lockGuard.release();
  getStaticMutex().unlock();

  if(scopedLayer.m_ogrLayer == nullptr)
    throw Exception(TE_TR("The informed data set could not be found in the data source."));

  scopedLayer->SetSpatialFilterRect(e->m_llx, e->m_lly, e->m_urx, e->m_ury);

  return std::unique_ptr<te::da::DataSet>(new DataSet(ds, scopedLayer.release(), releaseDataSet, scopedLayer.m_release));
}

std::unique_ptr<te::da::DataSet> te::ogr::Transactor::getDataSet(const std::string& name,
                                          const std::string& /*propertyName*/,
                                          const te::gm::Geometry* g,
                                          te::gm::SpatialRelation /*r*/,
                                          te::common::TraverseType /*travType*/,
                                          bool /*connected*/,
                                          const te::common::AccessPolicy pol)
{
  boost::unique_lock< boost::mutex > lockGuard(getStaticMutex());

  if (!m_ogrDs->getOGRDataSource())
  {
    lockGuard.release();
    getStaticMutex().unlock();
    return std::unique_ptr<te::da::DataSet>(nullptr);
  }

  CPLSetConfigOption("SHAPE_ENCODING", te::core::CharEncoding::getEncodingName(m_ogrDs->getEncoding()).c_str());

  bool releaseDataSet = false;
  GDALDataset* ds = nullptr;
  if (this->isInTransaction() == true)
  {
    ds = m_ogrDs->getOGRDataSource();
  }
  else
  {
    releaseDataSet = true;
    unsigned int ogropenflags = (pol == te::common::RWAccess || pol == te::common::WAccess) ? GDAL_OF_UPDATE : GDAL_OF_READONLY;
    ds = CreateGDALDataSet(m_ogrDs->getOGRDataSource()->GetDescription(), ogropenflags);
  }

  if (!ds)
  {
    CPLSetConfigOption("SHAPE_ENCODING", te::core::CharEncoding::getEncodingName(te::core::EncodingType::UTF8).c_str());
    throw Exception(TE_TR("Could not retrieve the data source."));
  }

  ScopedOGRLayer scopedLayer = GetLayer(ds, name);

  CPLSetConfigOption("SHAPE_ENCODING", te::core::CharEncoding::getEncodingName(te::core::EncodingType::UTF8).c_str());

  lockGuard.release();
  getStaticMutex().unlock();

  if(scopedLayer.m_ogrLayer == nullptr)
    throw Exception(TE_TR("The informed data set could not be found in the data source."));

  OGRGeometry* ogrg = Convert2OGR(g);

  scopedLayer->SetSpatialFilter(ogrg);

  OGRGeometryFactory::destroyGeometry(ogrg);

  return std::unique_ptr<te::da::DataSet>(new DataSet(ds, scopedLayer.release(), releaseDataSet, scopedLayer.m_release));
}

std::unique_ptr<te::da::DataSet> te::ogr::Transactor::query(const te::da::Select& q,
                                      te::common::TraverseType /*travType*/,
                                      bool /*connected*/,
                                      const te::common::AccessPolicy pol)
{
  boost::unique_lock< boost::mutex > lockGuard(getStaticMutex());

  if (!m_ogrDs->getOGRDataSource())
  {
    lockGuard.release();
    getStaticMutex().unlock();
    return std::unique_ptr<te::da::DataSet>(nullptr);
  }

  CPLSetConfigOption("SHAPE_ENCODING", te::core::CharEncoding::getEncodingName(m_ogrDs->getEncoding()).c_str());

  bool releaseDataSet = false;
  GDALDataset* ds = nullptr;
  if (this->isInTransaction() == true)
  {
    ds = m_ogrDs->getOGRDataSource();
  }
  else
  {
    releaseDataSet = true;
    unsigned int ogropenflags = (pol == te::common::RWAccess || pol == te::common::WAccess) ? GDAL_OF_UPDATE : GDAL_OF_READONLY;
    ds = CreateGDALDataSet(m_ogrDs->getOGRDataSource()->GetDescription(), ogropenflags);
  }
  
  if (!ds)
  {
    std::string errorMessage = CPLGetLastErrorMsg();
    TE_LOG_ERROR(errorMessage);

    CPLSetConfigOption("SHAPE_ENCODING", te::core::CharEncoding::getEncodingName(te::core::EncodingType::UTF8).c_str());
    lockGuard.release();
    getStaticMutex().unlock();
    throw Exception(TE_TR("Could not retrieve the data source."));
  }

  std::string sql;

  SQLVisitor visitor(*m_ogrDs->getDialect(), sql);

  q.accept(visitor);

  if (m_ogrDs->getCapabilities().getQueryCapabilities().supportsSpatialSQLDialect() == false)
  {
    sql = RemoveSpatialSql(sql);
  }  

  OGRLayer* layer = te::ogr::ExecuteSQL(ds, sql);

  CPLSetConfigOption("SHAPE_ENCODING", te::core::CharEncoding::getEncodingName(te::core::EncodingType::UTF8).c_str());

  lockGuard.release();
  getStaticMutex().unlock();

  if (layer == nullptr)
  {
    throw Exception(TE_TR("Could not retrieve the DataSet from data source."));
  }

  if (m_ogrDs->getCapabilities().getQueryCapabilities().supportsSpatialSQLDialect() == false)
  {
    te::gm::Envelope* e = visitor.getMBR();
    if (e != nullptr)
      layer->SetSpatialFilterRect(e->m_llx, e->m_lly, e->m_urx, e->m_ury);
  }

  return std::unique_ptr<te::da::DataSet>(new DataSet(ds, layer, releaseDataSet, true));
}

std::unique_ptr<te::da::DataSet> te::ogr::Transactor::query(const std::string& query,
                                      te::common::TraverseType /*travType*/,
                                      bool /*connected*/,
                                      const te::common::AccessPolicy pol)
{
  boost::unique_lock< boost::mutex > lockGuard(getStaticMutex());

  if (!m_ogrDs->getOGRDataSource())
  {
    lockGuard.release();
    getStaticMutex().unlock();
    return std::unique_ptr<te::da::DataSet>();
  }

  CPLSetConfigOption("SHAPE_ENCODING", te::core::CharEncoding::getEncodingName(m_ogrDs->getEncoding()).c_str());

  bool releaseDataSet = false;
  GDALDataset* ds = nullptr;
  if (this->isInTransaction() == true)
  {
    ds = m_ogrDs->getOGRDataSource();
  }
  else
  {
    releaseDataSet = true;
    unsigned int ogropenflags = (pol == te::common::RWAccess || pol == te::common::WAccess) ? GDAL_OF_UPDATE : GDAL_OF_READONLY;
    ds = CreateGDALDataSet(m_ogrDs->getOGRDataSource()->GetDescription(), ogropenflags);
  }

  if (!ds)
  {
    CPLSetConfigOption("SHAPE_ENCODING", te::core::CharEncoding::getEncodingName(te::core::EncodingType::UTF8).c_str());
    lockGuard.release();
    getStaticMutex().unlock();
    throw Exception(TE_TR("Could not retrieve the data source."));
  }

  OGRLayer* layer = te::ogr::ExecuteSQL(ds, query);

  CPLSetConfigOption("SHAPE_ENCODING", te::core::CharEncoding::getEncodingName(te::core::EncodingType::UTF8).c_str());

  lockGuard.release();
  getStaticMutex().unlock();

  if (layer == nullptr)
  {
    throw Exception(TE_TR("Could not retrieve the DataSet from data source."));
  }

  return std::unique_ptr<te::da::DataSet>(new DataSet(ds, layer, releaseDataSet, true));
}

void te::ogr::Transactor::execute(const te::da::Query& command)
{
  std::string sql;
  SQLVisitor v(*m_ogrDs->getDialect(), sql);

  command.accept(v);

  execute(sql);
}

void te::ogr::Transactor::execute(const std::string& command)
{
  if (!m_ogrDs->getOGRDataSource())
    return;

  boost::unique_lock< boost::mutex > lockGuard(getStaticMutex());

  OGRLayer* layer = te::ogr::ExecuteSQL(m_ogrDs->getOGRDataSource(), command);

  if(layer != nullptr)
    m_ogrDs->getOGRDataSource()->ReleaseResultSet(layer);
}

std::unique_ptr<te::da::PreparedQuery> te::ogr::Transactor::getPrepared(const std::string& /*qName*/)
{
  return std::unique_ptr<te::da::PreparedQuery>();
}

std::unique_ptr<te::da::BatchExecutor> te::ogr::Transactor::getBatchExecutor()
{
  return std::unique_ptr<te::da::BatchExecutor>();
}

void te::ogr::Transactor::cancel()
{
}

boost::int64_t te::ogr::Transactor::getLastGeneratedId()
{
  return m_fid;
}

std::string te::ogr::Transactor::escape(const std::string& value)
{
  return value;
}

std::vector<std::string> te::ogr::Transactor::getDataSetNames()
{
  std::vector<std::string> names;

  if (!m_ogrDs->getOGRDataSource())
    return names;

  for(int i=0; i<m_ogrDs->getOGRDataSource()->GetLayerCount(); i++)
    names.push_back(m_ogrDs->getOGRDataSource()->GetLayer(i)->GetName());

  return names;
}

std::size_t te::ogr::Transactor::getNumberOfDataSets()
{
  if (!m_ogrDs->getOGRDataSource())
    return 0;

  return static_cast<size_t>(m_ogrDs->getOGRDataSource()->GetLayerCount());
}

std::unique_ptr<te::da::DataSetType> te::ogr::Transactor::getDataSetType(const std::string& name)
{
  if (!m_ogrDs->getOGRDataSource())
    return std::unique_ptr<te::da::DataSetType>();

  boost::unique_lock< boost::mutex > lockGuard(getStaticMutex());

  ScopedOGRLayer scopedLayer = GetLayer(m_ogrDs->getOGRDataSource(), name);

  if (scopedLayer.m_ogrLayer == nullptr)
    return std::unique_ptr<te::da::DataSetType>();

  std::unique_ptr<te::da::DataSetType> type(GetDataSetType(m_ogrDs->getOGRDataSource(), scopedLayer.m_ogrLayer));
  type->setName(name);
  
  return type;
}

std::unique_ptr<te::da::DataSetTypeCapabilities> te::ogr::Transactor::getCapabilities(const std::string &name)
{
  std::unique_ptr<te::da::DataSetTypeCapabilities> cap(new te::da::DataSetTypeCapabilities);

  OGRLayer* l = m_ogrDs->getOGRDataSource()->GetLayerByName(name.c_str());

  if(l != nullptr)
  {
    cap->setSupportAddColumn((l->TestCapability(OLCCreateField) == 0) ? false : true);
    cap->setSupportRemoveColumn((l->TestCapability(OLCDeleteField) == 0) ? false : true);
    cap->setSupportDataEdition((l->TestCapability(OLCRandomWrite) == 0) ? false : true);
  }

  return cap;
}

boost::ptr_vector<te::dt::Property> te::ogr::Transactor::getProperties(const std::string& datasetName)
{
  boost::ptr_vector<te::dt::Property> ps;

  if (!m_ogrDs->getOGRDataSource())
    return ps;

  ScopedOGRLayer scopedLayer = GetLayer(m_ogrDs->getOGRDataSource(), datasetName);

  if(scopedLayer.m_ogrLayer != nullptr)
  {
    int srs = te::ogr::Convert2TerraLibProjection(scopedLayer->GetSpatialRef());
    std::unique_ptr<te::da::DataSetType> dt(GetDataSetType(scopedLayer.m_gdalDataSet, scopedLayer.m_ogrLayer, srs));
    std::vector<te::dt::Property*> props = dt->getProperties();
    std::vector<te::dt::Property*>::iterator it;

    for(it=props.begin(); it!=props.end(); ++it)
      ps.push_back((*it)->clone());
  }

  return ps;
}

std::unique_ptr<te::dt::Property> te::ogr::Transactor::getProperty(const std::string& datasetName, const std::string& name)
{
  std::unique_ptr<te::dt::Property> propertyFound;

  std::unique_ptr<te::da::DataSetType> dataSetType = getDataSetType(datasetName);
  if (dataSetType != nullptr)
  {
    te::dt::Property* property = dataSetType->getProperty(name);
    if (property != nullptr)
    {
      propertyFound.reset(property->clone());
    }
  }

  return propertyFound;
}

std::unique_ptr<te::dt::Property> te::ogr::Transactor::getProperty(const std::string& datasetName, std::size_t propertyPos)
{
  std::unique_ptr<te::dt::Property> propertyFound;

  std::unique_ptr<te::da::DataSetType> dataSetType = getDataSetType(datasetName);
  if (dataSetType != nullptr)
  {
    te::dt::Property* property = dataSetType->getProperty(propertyPos);
    if (property != nullptr)
    {
      propertyFound.reset(property->clone());
    }
  }

  return propertyFound;
}

std::vector<std::string> te::ogr::Transactor::getPropertyNames(const std::string& datasetName)
{
  std::vector<std::string> vecNames;

  std::unique_ptr<te::da::DataSetType> dataSetType = getDataSetType(datasetName);
  if (dataSetType != nullptr)
  {
    for (const te::dt::Property* property : dataSetType->getProperties())
    {
      vecNames.push_back(property->getName());
    }
  }

  return vecNames;
}

std::size_t te::ogr::Transactor::getNumberOfProperties(const std::string& datasetName)
{
  std::vector<std::string> vecNames = getPropertyNames(datasetName);
  return vecNames.size();
}

bool te::ogr::Transactor::propertyExists(const std::string& datasetName, const std::string& name)
{
  std::unique_ptr<te::dt::Property> property = getProperty(datasetName, name);
  if (property != nullptr)
  {
    return true;
  }

  return false;
}

void te::ogr::Transactor::addProperty(const std::string& datasetName, te::dt::Property* p)
{
  if (!m_ogrDs->getOGRDataSource())
    return;

  OGRLayer* l = m_ogrDs->getOGRDataSource()->GetLayerByName(datasetName.c_str());
  if (l == nullptr)
  {
    return;
  }
  
  te::da::FieldNameValidator validator = te::da::CreateValidator(&m_ogrDs->getCapabilities());
  InitValidator(validator, l->GetLayerDefn());

  std::string originalFieldName = p->getName();
  std::string adjustedFiledName;
  validator.addFieldName(originalFieldName, adjustedFiledName);

  p->setName(adjustedFiledName);

  CPLErrorReset();

  OGRErr error = OGRERR_FAILURE;
  std::string expectedName;
  std::string actualName;
  if (p->getType() == te::dt::GEOMETRY_TYPE)
  {
    std::unique_ptr<OGRGeomFieldDefn> newField(Convert2OGRGeom(p));
    error = l->CreateGeomField(newField.get(), FALSE);
    if (error != OGRERR_NONE)
    {
      std::string errorMessage = CPLGetLastErrorMsg();
      throw Exception(TE_TR("Error when attempting add the property: ") + p->getName() + TE_TR(" with the following message: ") + errorMessage);
    }

    expectedName = newField->GetNameRef();

    OGRGeomFieldDefn* createdField = l->GetLayerDefn()->GetGeomFieldDefn(
      l->GetLayerDefn()->GetGeomFieldCount() - 1);
    actualName = createdField->GetNameRef();
  }
  else
  {
    std::unique_ptr<OGRFieldDefn> newField(Convert2OGR(p));
    error = l->CreateField(newField.get());
    if (error != OGRERR_NONE)
    {
      std::string errorMessage = CPLGetLastErrorMsg();
      throw Exception(TE_TR("Error when attempting add the property: ") + p->getName() + TE_TR(" with the following message: ") + errorMessage);
    }

    expectedName = newField->GetNameRef();

    int fieldCount = l->GetLayerDefn()->GetFieldCount();
    OGRFieldDefn* createdField = l->GetLayerDefn()->GetFieldDefn(fieldCount - 1);
    actualName = createdField->GetNameRef();
  }

  //we must check if the field has been renamed. This can happen for ShapeFiles

  if (te::common::IsStringEqual(expectedName, actualName, false) == false)
  {
    p->setName(actualName);
  }

  m_ogrDs->getOGRDataSource()->FlushCache();
}

void te::ogr::Transactor::dropProperty(const std::string& datasetName, const std::string& name)
{
  OGRLayer* l = m_ogrDs->getOGRDataSource()->GetLayerByName(datasetName.c_str());

  if(l != nullptr)
  {
    if(!l->TestCapability(OLCDeleteField))
      throw Exception(TE_TR("This dataset do not support remove properties operation."));

    int fPos = l->GetLayerDefn()->GetFieldIndex(name.c_str());

    if(fPos < 0)
      throw Exception(TE_TR("Field not found."));

    OGRErr error = l->DeleteField(fPos);

    if(error != OGRERR_NONE)
      throw Exception(TE_TR("Error when attempting remove the property."));
  }
}

void te::ogr::Transactor::renameProperty(const std::string& datasetName,
                            const std::string& propertyName,
                            const std::string& newPropertyName)
{
  if (!m_ogrDs->getOGRDataSource())
    return;

  OGRLayer* l = m_ogrDs->getOGRDataSource()->GetLayerByName(datasetName.c_str());

  if(l != nullptr)
  {
    int idx = l->GetLayerDefn()->GetFieldIndex(propertyName.c_str());

    if(idx == -1)
      throw Exception(TE_TR("Field to be renamed does not exists."));

    OGRFieldDefn* df = l->GetLayerDefn()->GetFieldDefn(idx);

    OGRFieldDefn* dfn = new OGRFieldDefn(df);

    dfn->SetName(newPropertyName.c_str());

    OGRErr err = l->AlterFieldDefn(idx, dfn, ALTER_NAME_FLAG);

    if(err != OGRERR_NONE)
      throw Exception(TE_TR("Fail to rename field."));

    m_ogrDs->getOGRDataSource()->FlushCache();
  }
}

void te::ogr::Transactor::changePropertyDefinition(const std::string& datasetName, const std::string& propName, te::dt::Property* newProp)
{
  if (!m_ogrDs->getOGRDataSource())
    return;

  std::unique_ptr<te::dt::Property> p;

  p.reset(newProp);

  OGRLayer* l = m_ogrDs->getOGRDataSource()->GetLayerByName(datasetName.c_str());

  if(l != nullptr)
  {
    if(!l->TestCapability(OLCAlterFieldDefn))
      throw Exception(TE_TR("This data source do not support the operation of alter columns type."));

    int idx = l->GetLayerDefn()->GetFieldIndex(propName.c_str());

    if(idx == -1)
      throw Exception(TE_TR("Field to be renamed does not exists."));

    OGRFieldDefn* dfn = new OGRFieldDefn(l->GetLayerDefn()->GetFieldDefn(idx));

    dfn->SetType(GetOGRType(newProp->getType()));

    OGRErr err = l->AlterFieldDefn(idx, dfn, ALTER_TYPE_FLAG);

    if(err != OGRERR_NONE)
      throw Exception(TE_TR("Fail to to change field type."));

    m_ogrDs->getOGRDataSource()->FlushCache();
  }
}

std::unique_ptr<te::da::PrimaryKey> te::ogr::Transactor::getPrimaryKey(const std::string& datasetName)
{
  std::unique_ptr<te::da::DataSetType> dataSetType = this->getDataSetType(datasetName);

  std::unique_ptr<te::da::PrimaryKey> primaryKey;
  if (dataSetType != nullptr)
  {
    te::da::PrimaryKey* pk = dataSetType->getPrimaryKey();
    if (pk != nullptr)
    {
      primaryKey.reset(static_cast<te::da::PrimaryKey*>(pk->clone()));
    }
  }
  return primaryKey;
}

bool te::ogr::Transactor::primaryKeyExists(const std::string& /*datasetName*/, const std::string& /*name*/)
{
  return false;
}

void te::ogr::Transactor::addPrimaryKey(const std::string& /*datasetName*/, te::da::PrimaryKey* /*pk*/)
{
}

void te::ogr::Transactor::dropPrimaryKey(const std::string& /*datasetName*/)
{
}

std::unique_ptr<te::da::ForeignKey> te::ogr::Transactor::getForeignKey(const std::string& /*datasetName*/, const std::string& /*name*/)
{
  return std::unique_ptr<te::da::ForeignKey>();
}

std::vector<std::string> te::ogr::Transactor::getForeignKeyNames(const std::string& /*datasetName*/)
{
  return std::vector<std::string>();
}

bool te::ogr::Transactor::foreignKeyExists(const std::string& /*datasetName*/, const std::string& /*name*/)
{
  return false;
}

void te::ogr::Transactor::addForeignKey(const std::string& /*datasetName*/, te::da::ForeignKey* /*fk*/)
{
}

void te::ogr::Transactor::dropForeignKey(const std::string& /*datasetName*/, const std::string& /*fkName*/)
{
}

std::unique_ptr<te::da::UniqueKey> te::ogr::Transactor::getUniqueKey(const std::string& /*datasetName*/, const std::string& /*name*/)
{
  return std::unique_ptr<te::da::UniqueKey>();
}

std::vector<std::string> te::ogr::Transactor::getUniqueKeyNames(const std::string& /*datasetName*/)
{
  return std::vector<std::string>();
}

bool te::ogr::Transactor::uniqueKeyExists(const std::string& /*datasetName*/, const std::string& /*name*/)
{
  return false;
}

void te::ogr::Transactor::addUniqueKey(const std::string& /*datasetName*/, te::da::UniqueKey* /*uk*/)
{
}

void te::ogr::Transactor::dropUniqueKey(const std::string& /*datasetName*/, const std::string& /*name*/)
{
}

std::unique_ptr<te::da::CheckConstraint> te::ogr::Transactor::getCheckConstraint(const std::string& /*datasetName*/, const std::string& /*name*/)
{
  return std::unique_ptr<te::da::CheckConstraint>();
}

std::vector<std::string> te::ogr::Transactor::getCheckConstraintNames(const std::string& /*datasetName*/)
{
  return std::vector<std::string>();
}

bool te::ogr::Transactor::checkConstraintExists(const std::string& /*datasetName*/, const std::string& /*name*/)
{
  return false;
}

void te::ogr::Transactor::addCheckConstraint(const std::string& /*datasetName*/, te::da::CheckConstraint* /*cc*/)
{
}

void te::ogr::Transactor::dropCheckConstraint(const std::string& /*datasetName*/, const std::string& /*name*/)
{
}

std::unique_ptr<te::da::Index> te::ogr::Transactor::getIndex(const std::string& /*datasetName*/, const std::string& /*name*/)
{
  return std::unique_ptr<te::da::Index>();
}

std::vector<std::string> te::ogr::Transactor::getIndexNames(const std::string& /*datasetName*/)
{
  return std::vector<std::string>();
}

bool te::ogr::Transactor::indexExists(const std::string& /*datasetName*/, const std::string& /*name*/)
{
  return false;
}

void te::ogr::Transactor::addIndex(const std::string& datasetName, te::da::Index* idx,
                      const std::map<std::string, std::string>& /*options*/)
{
  std::string idxName = idx->getName();

  // Check if the index is associated to a UK or PK
  std::unique_ptr<te::da::PrimaryKey> pk = getPrimaryKey(datasetName);

  if (pk.get() && (pk->getAssociatedIndex() == idx))
    return;

  std::vector<std::string> ukNames = getUniqueKeyNames(datasetName);

  for (std::size_t i = 0; i < ukNames.size(); ++i)
    if (getUniqueKey(datasetName, ukNames[i])->getAssociatedIndex() == idx)
      return;

  // If there is not a uk or pk associated, let's create the index!
  std::string sql("CREATE INDEX IF NOT EXISTS ");
  sql += "\"" + idxName + "\"";
  sql += " ON ";
  sql += "\"" + datasetName + "\"";
  sql += " (";

  std::size_t numProperties = idx->getProperties().size();

  for (size_t i = 0; i < numProperties; ++i)
  {
    if (i != 0)
      sql += ", ";

    sql += "\"" + idx->getProperties()[i]->getName() + "\"";
  }

  sql += ")";

  execute(sql);

  m_ogrDs->getOGRDataSource()->FlushCache();
}

void te::ogr::Transactor::dropIndex(const std::string& /*datasetName*/, const std::string& idxName)
{
  std::string sql("DROP INDEX " + idxName);
  execute(sql);
}

std::unique_ptr<te::da::Sequence> te::ogr::Transactor::getSequence(const std::string& /*name*/)
{
  return std::unique_ptr<te::da::Sequence>();
}

std::vector<std::string> te::ogr::Transactor::getSequenceNames()
{
  return std::vector<std::string>();
}

bool te::ogr::Transactor::sequenceExists(const std::string& /*name*/)
{
  return false;
}

void te::ogr::Transactor::addSequence(te::da::Sequence* /*sequence*/)
{
}

void te::ogr::Transactor::dropSequence(const std::string& /*name*/)
{
}

std::unique_ptr<te::gm::Envelope> te::ogr::Transactor::getExtent(const std::string& datasetName,
                                                  const std::string& propertyName)
{
  if (!m_ogrDs->getOGRDataSource())
    return std::unique_ptr<te::gm::Envelope>();

  std::unique_ptr<te::gm::Envelope> res;
  std::string sql("SELECT * ");
  sql +=  " FROM \"";
  sql += datasetName + "\"";

  OGRLayer* l = te::ogr::ExecuteSQL(m_ogrDs->getOGRDataSource(), sql);
  if(l != nullptr)
  {
    if (l->GetFeatureCount() > 0)
    {
      std::unique_ptr<OGREnvelope> env(new OGREnvelope);
      if (l->GetExtent(env.get()) != OGRERR_NONE)
      {
        m_ogrDs->getOGRDataSource()->ReleaseResultSet(l);
        throw Exception(TE_TR("Error when attempting get extent."));
      }
      if (env->IsInit() != 0)
      {
        res.reset(Convert2TerraLib(env.get()));
      }
    }
    else
    {
      res.reset(new te::gm::Envelope());
    }
    
    m_ogrDs->getOGRDataSource()->ReleaseResultSet(l);
  }

  return res;
}

std::unique_ptr<te::gm::Envelope> te::ogr::Transactor::getExtent(const std::string& datasetName,
                                                  std::size_t /*propertyPos*/)
{
  return getExtent(datasetName, "");
}

std::size_t te::ogr::Transactor::getNumberOfItems(const std::string& datasetName)
{
  if (!m_ogrDs->getOGRDataSource())
    return 0;

  OGRLayer* l = m_ogrDs->getOGRDataSource()->GetLayerByName(datasetName.c_str());

  if(l != nullptr)
    return static_cast<size_t>(l->GetFeatureCount());

  return 0;
}

bool te::ogr::Transactor::hasDataSets()
{
  if (!m_ogrDs->getOGRDataSource())
    return false;

  return (m_ogrDs->getOGRDataSource()->GetLayerCount() > 0);
}

bool te::ogr::Transactor::dataSetExists(const std::string& name)
{
  if (!m_ogrDs->getOGRDataSource())
    return false;

  return (m_ogrDs->getOGRDataSource()->GetLayerByName(name.c_str()) != nullptr);
}

void te::ogr::Transactor::createDataSet(te::da::DataSetType* dt, const std::map<std::string, std::string>& /*options*/)
{
  if (!m_ogrDs->getOGRDataSource())
  {
    m_ogrDs->createOGRDataSource();
  }

  if (!m_ogrDs->getOGRDataSource())
    return;

  if(!m_ogrDs->getOGRDataSource()->TestCapability(ODsCCreateLayer))
    throw Exception(TE_TR("This driver does not support dataset creation."));

  OGRwkbGeometryType geomType = wkbUnknown;
  OGRSpatialReference* srs = nullptr;
  std::string strSRID = "";
  if(dt->hasGeom())
  {
    geomType = Convert2OGR(te::da::GetFirstGeomProperty(dt)->getGeometryType());
    int srid = te::da::GetFirstGeomProperty(dt)->getSRID();
    if (srid != TE_UNKNOWN_SRS)
    {
      srs  = Convert2OGRProjection( srid );
      strSRID = boost::lexical_cast<std::string>(srid);
    }
  }

  char** papszOptions = nullptr;

  std::map<std::string, std::string> kvp = te::core::Expand(m_ogrDs->getConnectionInfo().query());

  //for compatibility purposes, we must force the name of the FID column to be created
  std::string driverName = m_ogrDs->getOGRDataSource()->GetDriverName();

  std::string primaryKeyColumnName = "";
  bool ignorePrimaryKeyColumn = false;

  if(driverName == "SQLite" )
  {
    te::da::PrimaryKey* primaryKey = dt->getPrimaryKey();
    if (primaryKey != nullptr && primaryKey->getProperties().size() == 1)
    {
      te::dt::Property* property = primaryKey->getProperties().at(0);
      if (property->getType() == te::dt::INT32_TYPE || property->getType() == te::dt::INT64_TYPE)
      {
        ignorePrimaryKeyColumn = true;
        primaryKeyColumnName = property->getName();
        kvp["FID"] = primaryKeyColumnName;
      }
    }

    if( dt->hasPropertyOfType( te::dt::GEOMETRY_TYPE ) )
    {
      te::dt::Property const * const geomPropPtr =
        dt->findFirstPropertyOfType( te::dt::GEOMETRY_TYPE );
      assert( geomPropPtr != nullptr );

      kvp["GEOMETRY_NAME"] = geomPropPtr->getName();
    }

    kvp["FORMAT"] = "SPATIALITE";

    if (strSRID.empty() == false)
    {
      kvp["SRID"] = strSRID;
    }
  }
  else if(driverName == "GPKG")
  {
    te::da::PrimaryKey* primaryKey = dt->getPrimaryKey();
    if (primaryKey != nullptr && primaryKey->getProperties().size() == 1)
    {
      te::dt::Property* property = primaryKey->getProperties().at(0);
      if (property->getType() == te::dt::INT32_TYPE || property->getType() == te::dt::INT64_TYPE)
      {
        ignorePrimaryKeyColumn = true;
        primaryKeyColumnName = property->getName();
        kvp["FID"] = primaryKeyColumnName;
      }
    }

    if( dt->hasPropertyOfType( te::dt::GEOMETRY_TYPE ) )
    {
      te::dt::Property const * const geomPropPtr =
        dt->findFirstPropertyOfType( te::dt::GEOMETRY_TYPE );
      assert( geomPropPtr != nullptr );

      kvp["GEOMETRY_NAME"] = geomPropPtr->getName();
    }
  }
  else if (driverName == "GeoJSON")
  {
    kvp["RFC7946"] = "YES";
  }
  else if (driverName == "ESRI Shapefile")
  {
    primaryKeyColumnName = "FID";
    ignorePrimaryKeyColumn = true;
  }

  std::map<std::string, std::string>::const_iterator it = kvp.begin();
  while(it != kvp.end())
  {
    papszOptions = CSLSetNameValue(papszOptions, it->first.c_str(), it->second.c_str());
    ++it;
  }

  OGRLayer* newLayer = m_ogrDs->getOGRDataSource()->CreateLayer(dt->getName().c_str(), srs, geomType, papszOptions);

  if (srs)
    srs->Release();

  if(papszOptions)
    CSLDestroy(papszOptions);

  if(newLayer == nullptr)
    throw Exception(TE_TR("Error when attempting create the dataset type."));

  m_ogrDs->updateCapabilities();

  dt->setName(newLayer->GetName());

// add the properties

  int numberOfDtGeomProperties = 0;

  for (size_t i = 0; i < dt->size(); ++i)
  {
    if (dt->getProperty(i)->getType() == te::dt::GEOMETRY_TYPE)
      ++numberOfDtGeomProperties;
  }

  int numberOfLayerGeomProperties = m_ogrDs->getOGRDataSource()->GetLayerByName(
    dt->getName().c_str())->GetLayerDefn()->GetGeomFieldCount();

  for (size_t i = 0; i < dt->size(); ++i)
  {
    te::dt::Property* p = dt->getProperty(i);

    if (p->getType() == te::dt::GEOMETRY_TYPE)
    {
      if (
            (
              numberOfDtGeomProperties
              ==
              numberOfLayerGeomProperties
            )
            ||
            (
              m_ogrDs->getOGRDataSource()->GetLayerByName(
               dt->getName().c_str())->GetLayerDefn()->GetGeomFieldIndex(
               p->getName().c_str() )
              !=
              (-1)
            )
         )
      {
        continue;
      }
    }

    if (ignorePrimaryKeyColumn == true)
    {
      if (te::common::IsStringEqual(p->getName(), primaryKeyColumnName, false) == true)
      {
        continue;
      }
    }

    addProperty(dt->getName(), p);
  }

  m_ogrDs->getOGRDataSource()->FlushCache();
}

void te::ogr::Transactor::cloneDataSet(const std::string& name,
                          const std::string& cloneName,
                          const std::map<std::string, std::string>& /*options*/)
{
  if (!m_ogrDs->getOGRDataSource())
    return;

  if(!m_ogrDs->getOGRDataSource()->TestCapability(ODsCCreateLayer))
    throw Exception(TE_TR("This driver does not support creates a dataset."));

  OGRLayer* l = m_ogrDs->getOGRDataSource()->GetLayerByName(name.c_str());

  if(l == nullptr)
    throw Exception(TE_TR("Could not retrieve the DataSet from data source."));

  OGRLayer* cl = m_ogrDs->getOGRDataSource()->CopyLayer(l, cloneName.c_str());

  if(cl == nullptr)
    throw Exception(TE_TR("Error when attempting clone the dataset."));
}

void te::ogr::Transactor::dropDataSet(const std::string& name)
{
  if (!m_ogrDs->getOGRDataSource())
    return;

  if(!m_ogrDs->getOGRDataSource()->TestCapability(ODsCDeleteLayer))
    throw Exception(TE_TR("This driver does not support remove a dataset."));

  int i=0;

  for(; i<m_ogrDs->getOGRDataSource()->GetLayerCount(); i++)
    if(name.compare(m_ogrDs->getOGRDataSource()->GetLayer(i)->GetName()) == 0)
      break;

  if(i == m_ogrDs->getOGRDataSource()->GetLayerCount())
    throw Exception(TE_TR("Could not retrieve the DataSet from data source."));

  if(m_ogrDs->getOGRDataSource()->DeleteLayer(i) != OGRERR_NONE)
    throw Exception(TE_TR("Error when attempting to remove the dataset."));
}

void te::ogr::Transactor::renameDataSet(const std::string& /*name*/, const std::string& /*newName*/)
{
}

void te::ogr::Transactor::add(const std::string& datasetName,
                  te::da::DataSet* d,
                  const std::map<std::string, std::string>& options,
                  std::size_t limit,
                  bool enableProgress)
{
  if(limit == 0)
    limit = std::string::npos;

  if (!m_ogrDs->getOGRDataSource())
    return;

  OGRLayer* layer = m_ogrDs->getOGRDataSource()->GetLayerByName(datasetName.c_str());
  if(layer == nullptr)
    throw Exception(TE_TR("Could not retrieve the DataSet from data source."));

  layer->ResetReading();

  OGRFeatureDefn* featureDefinition = layer->GetLayerDefn();

  //verify geometry property and get OGRSpatialReference information
  OGRSpatialReference* srs = nullptr;

  te::da::FieldNameValidator validator = te::da::CreateValidator(&m_ogrDs->getCapabilities());

  unsigned int numberOfDtGeomFields = 0;
  for (std::size_t i = 0; i < d->getNumProperties(); ++i)
  {
    std::string originalFieldName = d->getPropertyName(i);
    validator.addFieldName(originalFieldName);

    if( d->getPropertyDataType(i) == te::dt::GEOMETRY_TYPE )
    {
      ++numberOfDtGeomFields;
    }    
  }

  const std::vector<std::string>& vecAdjustedNames = validator.getAdjustedNames();
  std::vector<int> vecIndexes;

  for (std::size_t i = 0; i < d->getNumProperties(); ++i)
  {
    if( d->getPropertyDataType(i) == te::dt::GEOMETRY_TYPE )
    {
      if( 
          ( numberOfDtGeomFields == 1 )
          &&
          ( featureDefinition->GetGeomFieldCount() == 1 )
        )
      {
        vecIndexes.push_back( 0 );
      }
      else
      {
        vecIndexes.push_back(featureDefinition->GetGeomFieldIndex(
          vecAdjustedNames[ i ].c_str()) );
      }
    }
    else
    {
      vecIndexes.push_back(featureDefinition->GetFieldIndex(
        vecAdjustedNames[ i ].c_str()));
    }
  }

  const std::string FID_COLUMN = layer->GetFIDColumn();

  const bool putFIDAttribute = te::da::HasOption(options, m_putFIDAttributeOption);

  try
  {
    begin();

    std::size_t nproperties = d->getNumProperties();

    std::size_t nProcessedRows = 0;

    std::unique_ptr< te::common::TaskProgress > progressPtr;
    if(enableProgress == true)
    {
      progressPtr.reset( new te::common::TaskProgress );
      progressPtr->setTotalSteps(1 + static_cast<int>(d->size()));
      progressPtr->setMessage("Saving...");
      progressPtr->useTimer(true);
      progressPtr->pulse();
    }

    OGRErr resultCode = 0;

    while(d->moveNext() && (nProcessedRows != limit))
    {
      OGRFeature* feat = OGRFeature::CreateFeature(featureDefinition);

      for(std::size_t sourceIndex = 0; sourceIndex != nproperties; ++sourceIndex)
      {
        int destinationIndex = vecIndexes.at(sourceIndex);
        if( destinationIndex < 0 )
        {
          continue;
        }

        std::string propertyName = vecAdjustedNames.at(sourceIndex);
        int propertyDataType = d->getPropertyDataType(sourceIndex);

        if (putFIDAttribute == false && te::common::Convert2UCase(propertyName) == "FID")
        {
          continue;
        }

        if (d->isNull(sourceIndex) == true)
        {
          continue;
        }

        if (FID_COLUMN == propertyName)
        {
          if (d->getPropertyDataType(sourceIndex) == te::dt::INT32_TYPE)
          {
            feat->SetFID(static_cast<GIntBig>(d->getInt32(sourceIndex)));
          }
          else if (d->getPropertyDataType(sourceIndex) == te::dt::INT64_TYPE)
          {
            feat->SetFID(static_cast<GIntBig>(d->getInt64(sourceIndex)));
          }
          continue;
        }

        switch(propertyDataType)
        {
          case te::dt::INT16_TYPE:
            feat->SetField(static_cast<int>(destinationIndex), d->getInt16(sourceIndex));
          break;

          case te::dt::INT32_TYPE:
            feat->SetField(static_cast<int>(destinationIndex), d->getInt32(sourceIndex));
          break;

          case te::dt::INT64_TYPE:
            feat->SetField(static_cast<int>(destinationIndex), static_cast<GIntBig>(d->getInt64(sourceIndex)));
            break;

          case te::dt::STRING_TYPE:
            feat->SetField(static_cast<int>(destinationIndex), d->getAsString(sourceIndex).c_str());
          break;

          case te::dt::DOUBLE_TYPE:
            feat->SetField(static_cast<int>(destinationIndex), d->getDouble(sourceIndex));
           break;

          case te::dt::NUMERIC_TYPE:
            feat->SetField(static_cast<int>(destinationIndex), atof(d->getNumeric(sourceIndex).c_str()));
          break;

          case te::dt::BOOLEAN_TYPE:
            feat->SetField(static_cast<int>(destinationIndex), static_cast<int>(d->getBool(sourceIndex)));
          break;

          case te::dt::BYTE_ARRAY_TYPE:
            {
              std::unique_ptr<te::dt::ByteArray> ba(d->getByteArray(sourceIndex));
              feat->SetField(static_cast<int>(destinationIndex), static_cast<int>(ba->bytesUsed()), reinterpret_cast<unsigned char*>(ba->getData()));
            }
          break;

          case te::dt::DATETIME_TYPE:
            {
              std::unique_ptr<te::dt::DateTime> dtm(d->getDateTime(sourceIndex));

              te::dt::Date* dtime = dynamic_cast<te::dt::Date*>(dtm.get());
              if(dtime)
              {
                feat->SetField(static_cast<int>(destinationIndex),
                               static_cast<int>(dtime->getYear()),
                               static_cast<int>(dtime->getMonth()),
                               static_cast<int>(dtime->getDay()));
                break;
              }

              te::dt::TimeDuration* tduration = dynamic_cast<te::dt::TimeDuration*>(dtm.get());

              if(tduration)
              {
                feat->SetField(static_cast<int>(destinationIndex), 0, 0, 0,
                               static_cast<int>(tduration->getHours()),
                               static_cast<int>(tduration->getMinutes()),
                               static_cast<float>(tduration->getSeconds()));
                break;
              }

              te::dt::TimeInstant* tinst = dynamic_cast<te::dt::TimeInstant*>(dtm.get());

              if(tinst)
              {
                feat->SetField(static_cast<int>(destinationIndex),
                               static_cast<int>(tinst->getDate().getYear()),
                               static_cast<int>(tinst->getDate().getMonth()),
                               static_cast<int>(tinst->getDate().getDay()),
                               static_cast<int>(tinst->getTime().getHours()),
                               static_cast<int>(tinst->getTime().getMinutes()),
                               static_cast<float>(tinst->getTime().getSeconds()));
                break;
              }

              te::dt::TimeInstantTZ* timeInstantTZ = dynamic_cast<te::dt::TimeInstantTZ*>(dtm.get());
              if (timeInstantTZ != nullptr)
              {
                feat->SetField(static_cast<int>(destinationIndex),
                  static_cast<int>(timeInstantTZ->getDate().getYear()),
                  static_cast<int>(timeInstantTZ->getDate().getMonth()),
                  static_cast<int>(timeInstantTZ->getDate().getDay()),
                  static_cast<int>(timeInstantTZ->getTime().getHours()),
                  static_cast<int>(timeInstantTZ->getTime().getMinutes()),
                  static_cast<float>(timeInstantTZ->getTime().getSeconds()));
                break;
              }
              rollBack();
              throw Exception (TE_TR("Unsupported date and time type by OGR."));
            }

          case te::dt::GEOMETRY_TYPE:
            {
              std::unique_ptr<te::gm::Geometry> geom(d->getGeometry(sourceIndex));
              assert( geom.get() != nullptr );

              if (srs == nullptr)
              {
                srs = Convert2OGRProjection(geom->getSRID());
              }

              OGRGeometry* OGRgeom = Convert2OGR(geom.get(), srs);

              resultCode = feat->SetGeomFieldDirectly(destinationIndex, OGRgeom);
              if( resultCode != OGRERR_NONE )
              {
                std::string errorMessage = CPLGetLastErrorMsg();
                OGRFeature::DestroyFeature(feat);
                rollBack();
                throw Exception(errorMessage);
              }
            }
          break;

          default:
            rollBack();
            throw Exception(TE_TR("Unsupported data type by OGR."));
        }
      }

      resultCode = layer->CreateFeature(feat);
      if(resultCode != OGRERR_NONE)
      {
        std::string errorMessage = CPLGetLastErrorMsg();
        OGRFeature::DestroyFeature(feat);
        rollBack();
        throw Exception(errorMessage);
      }

      m_fid = static_cast<long>(feat->GetFID());

      OGRFeature::DestroyFeature(feat);
      nProcessedRows++;

      if(progressPtr.get())
      {
        progressPtr->pulse();

        if (progressPtr->isActive() == false)
        {
          rollBack();
          throw Exception(TE_TR("Save canceled!"));
        }
      }
    }
    commit();
  }
  catch(Exception& e)
  {
    if (srs)
      srs->Release();

    rollBack();
    throw e;
  }

  if (srs)
    srs->Release();
}

void te::ogr::Transactor::remove(const std::string& datasetName, const te::da::ObjectIdSet* oids)
{
  if(!m_ogrDs->getOGRDataSource())
    return;

  OGRLayer* l = m_ogrDs->getOGRDataSource()->GetLayerByName(datasetName.c_str());

  if(l == nullptr)
    throw Exception(TE_TR("Could not retrieve the DataSet from data source."));

  if(!l->TestCapability(OLCDeleteFeature))
    throw Exception(TE_TR("Driver does not support removal of features."));

  std::vector<GIntBig> oidsInt;

  if (oids)
  {
    std::set<te::da::ObjectId*, te::common::LessCmp<te::da::ObjectId*> >::const_iterator it = oids->begin();

    while (it != oids->end())
    {
      std::string id = (*it)->getValueAsString();

      char* endptr = nullptr;

      GIntBig idLongLong = strtoll(id.c_str(), &endptr, 10);

      oidsInt.push_back(idLongLong);

      ++it;
    }
  }
  else
  {
    GIntBig count = l->GetFeatureCount();

    for (GIntBig i = 0; i < count; ++i)
    {
      oidsInt.push_back(i);
    }
  }

  begin();

  //We must sort the vector in decending order to avoid problems in delete Feature
  std::sort(oidsInt.rbegin(), oidsInt.rend());

  std::size_t numItems = oidsInt.size();

  for (std::size_t i = 0; i < numItems; ++i)
  {
    OGRErr result = l->DeleteFeature(oidsInt.at(i));
    if (result != OGRERR_NONE)
    {
      rollBack();
      throw Exception(TE_TR("Error when attempting to remove the feature."));
    }
  }

  commit();
  
  m_ogrDs->getOGRDataSource()->FlushCache();
}


void te::ogr::Transactor::update(const std::string&/* datasetName*/,
                    te::da::DataSet* /*dataset*/,
                    const std::vector<std::size_t>&/* properties*/,
                    const te::da::ObjectIdSet* /*oids*/,
                    const std::map<std::string, std::string>& /*options*/,
                    std::size_t /*limit*/)
{
}

void te::ogr::Transactor::update(const std::string &datasetName, te::da::DataSet *dataset, const std::vector< std::set<int> >& properties,
                                 const std::vector<size_t>& ids)
{
  if(m_ogrDs->getOGRDataSource() == nullptr)
    throw Exception(TE_TR("Data source failure"));

  OGRLayer* l = m_ogrDs->getOGRDataSource()->GetLayerByName(datasetName.c_str());

  if(l == nullptr)
    throw Exception(TE_TR("Could not retrieve dataset"));

  OGRFeatureDefn* ogrFeatureDefinition = l->GetLayerDefn();

  dataset->moveFirst();
  int i = 0;

  //verify geometry property and get OGRSpatialReference information
  OGRSpatialReference* srs = nullptr;

  std::vector<std::string> vecAdjustedNames;
  te::da::FieldNameValidator validator = te::da::CreateValidator(&m_ogrDs->getCapabilities());
  
  for (std::size_t i = 0; i < dataset->getNumProperties(); ++i)
  {
    std::string originalFieldName = dataset->getPropertyName(i);
    std::string adjustedFieldName;
    validator.addFieldName(originalFieldName, adjustedFieldName);

    vecAdjustedNames.push_back(adjustedFieldName);
  }

  do
  {
    size_t idIndex = ids[0];
    int idValue = -1;

    int idDataType = dataset->getPropertyDataType(idIndex);

    switch (idDataType)
    {
      case te::dt::INT16_TYPE:
        idValue = dataset->getInt16(idIndex);
      break;

      case te::dt::INT32_TYPE:
        idValue = dataset->getInt32(idIndex);
      break;

      default:
        idValue = static_cast<int>(dataset->getInt64(idIndex));
      break;
    }

    if (idValue < 0)
      continue;

    OGRFeature* feat = l->GetFeature(idValue)->Clone();

    std::set<int> ls = properties[static_cast<size_t>(i)];
    std::set<int>::iterator it;

    for(it = ls.begin(); it != ls.end(); ++it)
    {
      int sourceIndex = *it;
      std::string fieldName = vecAdjustedNames.at(sourceIndex);
      int propertyDataType = dataset->getPropertyDataType(sourceIndex);
      
      //int fpos_o = fpos - 1;

      int destinationIndex = ogrFeatureDefinition->GetFieldIndex(fieldName.c_str());
      //if the field is not found, we must check if it is the primary key
      if (destinationIndex == -1 && propertyDataType != te::dt::GEOMETRY_TYPE)
      {
        //if the not found field is not the primary key, an error has ocurred
        if (sourceIndex != idIndex)
        {
          throw Exception(TE_TR("Fail updating data. Index mismatch"));
        }

        continue;
      }

      if (dataset->isNull(sourceIndex) == true)
      {
#if GDAL_VERSION_NUM >= 2020000
        //introduced in version 2.2
        feat->SetFieldNull(destinationIndex);
#else
        feat->UnsetField(destinationIndex);
#endif
        continue;
      }

      switch(dataset->getPropertyDataType(static_cast<size_t>(sourceIndex)))
      {
        case te::dt::INT32_TYPE:
          feat->SetField(destinationIndex, dataset->getInt32(static_cast<size_t>(sourceIndex)));
        break;

        case te::dt::DOUBLE_TYPE:
        case te::dt::NUMERIC_TYPE:
          feat->SetField(destinationIndex, dataset->getDouble(static_cast<size_t>(sourceIndex)));
        break;

        case te::dt::STRING_TYPE:
          feat->SetField(destinationIndex, dataset->getString(static_cast<size_t>(sourceIndex)).c_str());
        break;

        case te::dt::BOOLEAN_TYPE:
          feat->SetField(destinationIndex, static_cast<int>(dataset->getBool(sourceIndex)));
         break;

        case te::dt::BYTE_ARRAY_TYPE:
        {
          std::unique_ptr<te::dt::ByteArray> ba(dataset->getByteArray(sourceIndex));
          feat->SetField(destinationIndex, static_cast<int>(ba->bytesUsed()), reinterpret_cast<unsigned char*>(ba->getData()));
        }
        break;

        case te::dt::DATETIME_TYPE:
        {
          std::unique_ptr<te::dt::DateTime> dtm(dataset->getDateTime(sourceIndex));

          te::dt::Date* dtime = dynamic_cast<te::dt::Date*>(dtm.get());
          if (dtime)
          {
            feat->SetField(static_cast<int>(destinationIndex),
              static_cast<int>(dtime->getYear()),
              static_cast<int>(dtime->getMonth()),
              static_cast<int>(dtime->getDay()));
            break;
          }

          te::dt::TimeDuration* tduration = dynamic_cast<te::dt::TimeDuration*>(dtm.get());

          if (tduration)
          {
            feat->SetField(static_cast<int>(destinationIndex), 0, 0, 0,
              static_cast<int>(tduration->getHours()),
              static_cast<int>(tduration->getMinutes()),
              static_cast<float>(tduration->getSeconds()));
            break;
          }

          te::dt::TimeInstant* tinst = dynamic_cast<te::dt::TimeInstant*>(dtm.get());

          if (tinst)
          {
            feat->SetField(static_cast<int>(destinationIndex),
              static_cast<int>(tinst->getDate().getYear()),
              static_cast<int>(tinst->getDate().getMonth()),
              static_cast<int>(tinst->getDate().getDay()),
              static_cast<int>(tinst->getTime().getHours()),
              static_cast<int>(tinst->getTime().getMinutes()),
              static_cast<float>(tinst->getTime().getSeconds()));
            break;
          }

          te::dt::TimeInstantTZ* timeInstantTZ = dynamic_cast<te::dt::TimeInstantTZ*>(dtm.get());
          if (timeInstantTZ != nullptr)
          {
            feat->SetField(static_cast<int>(destinationIndex),
              static_cast<int>(timeInstantTZ->getDate().getYear()),
              static_cast<int>(timeInstantTZ->getDate().getMonth()),
              static_cast<int>(timeInstantTZ->getDate().getDay()),
              static_cast<int>(timeInstantTZ->getTime().getHours()),
              static_cast<int>(timeInstantTZ->getTime().getMinutes()),
              static_cast<float>(timeInstantTZ->getTime().getSeconds()));
            break;
          }
          throw Exception(TE_TR("Unsupported date and time type by OGR."));
        }

        case te::dt::GEOMETRY_TYPE:
        {
          std::unique_ptr<te::gm::Geometry> gm = dataset->getGeometry(static_cast<size_t>(sourceIndex));
          
          if (srs == nullptr)
          {
            srs = Convert2OGRProjection(gm->getSRID());
          }

          feat->SetGeometry(te::ogr::Convert2OGR(gm.get(), srs));
        }
      }
    }

    OGRErr error = l->SetFeature(feat);
    if(error != OGRERR_NONE)
    {
      if (srs != nullptr)
      {
        srs->Release();
      }

      std::string errorMessage = CPLGetLastErrorMsg();
      throw Exception(TE_TR("Fail updating data: " + errorMessage));
    }

    i++;
  } while (dataset->moveNext());

  m_ogrDs->getOGRDataSource()->FlushCache();

  if (srs != nullptr)
  {
    srs->Release();
  }
    
}

void te::ogr::Transactor::optimize(const std::map<std::string, std::string>& /*opInfo*/)
{
}
