/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/ogr/Utils.cpp
   
  \brief Utility functions for OGR support.
*/

// TerraLib
#include "../common/Exception.h"
#include "../common/Globals.h"
#include "../core/logger/Logger.h"
#include "../core/translator/Translator.h"
#include "../dataaccess/dataset/DataSetType.h"
#include "../dataaccess/utils/FieldNameValidator.h"
#include "../dataaccess/utils/Utils.h"
#include "../datatype/ArrayProperty.h"
#include "../datatype/DateTime.h"
#include "../datatype/DateTimeProperty.h"
#include "../datatype/NumericProperty.h"
#include "../datatype/Property.h"
#include "../datatype/SimpleProperty.h"
#include "../datatype/StringProperty.h"
#include "../geometry/Envelope.h"
#include "../geometry/Geometry.h"
#include "../geometry/GeometryProperty.h"
#include "../geometry/Utils.h"
#include "../geometry/WKBReader.h"
#include "../srs/SpatialReferenceSystemManager.h"
#include "../srs/Config.h"
#include "Utils.h"
#include "../core/logger/Logger.h"

// OGR
#include <ogrsf_frmts.h>
#include <ogr_spatialref.h>

// Boost
#include <boost/algorithm/string.hpp>
#include <boost/filesystem.hpp>

te::gm::Geometry* te::ogr::Convert2TerraLib(OGRGeometry* ogrGeom)
{
  int wkbSize = ogrGeom->WkbSize();

  unsigned char* wkbArray = new unsigned char[wkbSize];

  ogrGeom->exportToWkb(wkbNDR, wkbArray);

  te::gm::Geometry* teGeom = nullptr;

  try
  {
    teGeom = te::gm::WKBReader::read((const char*)wkbArray);
  }
  catch(...)
  {
    delete [] wkbArray;
    throw;
  }

  delete [] wkbArray;

  return teGeom;
}

OGRGeometry* te::ogr::Convert2OGR(const te::gm::Geometry* teGeom)
{
  //Creates the Spatial Referecnce for the geometry
  OGRSpatialReference* srs = Convert2OGRProjection(teGeom->getSRID());

  //We created the geometry and give it the 'srs' shared pointer
  OGRGeometry* ogrGeom = Convert2OGR(teGeom, srs);

  //Now we can release the shared pointed. The geometry will destroy it after using it
  srs->Release();

  return ogrGeom;
}

OGRGeometry* te::ogr::Convert2OGR(const te::gm::Geometry* teGeom, OGRSpatialReference* srs)
{
  size_t size = teGeom->getWkbSize();

  char* wkbArray = new char[size];

  teGeom->getWkb(wkbArray, te::common::Globals::sm_machineByteOrder);

  OGRGeometry* ogrGeom = nullptr;

  OGRErr result = OGRGeometryFactory::createFromWkb((unsigned char*)wkbArray, srs, &ogrGeom, static_cast<int>(size));

  delete[] wkbArray;

  if (result == OGRERR_NONE)
    return ogrGeom;

  if (result == OGRERR_NOT_ENOUGH_DATA)
    throw te::common::Exception(TE_TR("Error when attempting convert the geometry to OGR. Not enough data."));

  if (result == OGRERR_UNSUPPORTED_GEOMETRY_TYPE)
    throw te::common::Exception(TE_TR("Error when attempting convert the geometry to OGR. Unsupported geometry type."));

  if (result == OGRERR_CORRUPT_DATA)
    throw te::common::Exception(TE_TR("Error when attempting convert the geometry to OGR. Corrupt data."));

  return nullptr;
}

te::gm::Envelope* te::ogr::Convert2TerraLib(const OGREnvelope* env)
{
   return new te::gm::Envelope(env->MinX, env->MinY, env->MaxX, env->MaxY);
}

OGREnvelope* te::ogr::Convert2OGR(const te::gm::Envelope* env)
{
  OGREnvelope* envOGR = new OGREnvelope();
  envOGR->MinX = env->m_llx;
  envOGR->MinY = env->m_lly;
  envOGR->MaxX = env->m_urx;
  envOGR->MaxY = env->m_ury;
  return envOGR;
}

int te::ogr::Convert2TerraLibProjection(OGRSpatialReference* osrs)
{
  if(osrs == nullptr)
    return TE_UNKNOWN_SRS;
  
  int srid = 0;
  
  OGRErr ogrReturn = osrs->AutoIdentifyEPSG();
  if( ogrReturn == OGRERR_NONE )
  {
    const char* srsAuth = osrs->GetAuthorityCode(nullptr);
    
    if (srsAuth)
    {
      srid = atoi(srsAuth);
    }
  }  

  std::string proj4;
  if( srid == TE_UNKNOWN_SRS )
  {
    char* proj4StrPtr = nullptr;
    ogrReturn = osrs->exportToProj4( &proj4StrPtr );

    if( ogrReturn == OGRERR_NONE )
    {
      std::pair< std::string, unsigned int > customSRID;
      proj4 = std::string ( proj4StrPtr );

      OGRFree( proj4StrPtr );

      if(te::srs::SpatialReferenceSystemManager::getInstance().isValidP4TxtID(proj4))
      {
        customSRID = te::srs::SpatialReferenceSystemManager::getInstance().getIdFromP4Txt(
          proj4);
        srid = (int)customSRID.second;
      }
    }
  }
  
  std::string wkt;
  if( srid == TE_UNKNOWN_SRS )
  {
    char* wktPtr = nullptr;
    ogrReturn = osrs->exportToWkt( &wktPtr );
  
    if( ogrReturn == OGRERR_NONE )
    {    
      std::pair< std::string, unsigned int > customSRID;
      wkt = std::string( wktPtr );

      OGRFree( wktPtr );

      if(te::srs::SpatialReferenceSystemManager::getInstance().isValidWktID(wkt))
      {
        customSRID = te::srs::SpatialReferenceSystemManager::getInstance().getIdFromWkt(
          wkt);
        srid = (int)customSRID.second;
      }
    }
  }
  
  if( srid == TE_UNKNOWN_SRS )
  {
    // geographic SIRGAS 2000 Datum
    std::string straux(osrs->GetRoot()->GetChild(0)->GetValue());
    
    if (boost::find_first(straux, "SIRGAS"))
    {
      if (osrs->IsGeographic())
        srid = TE_SRS_SIRGAS2000;
    }
    else if (boost::find_first(straux, "UTM "))
    { // UTM using SIRGAS 2000 Datum
      double centralm = osrs->GetProjParm(SRS_PP_CENTRAL_MERIDIAN,-1);
      if (centralm != -1)
      {
        int zone = (int)(centralm/6 + 31);
        
        double fsnorth = osrs->GetProjParm(SRS_PP_FALSE_NORTHING,-1);
        if (fsnorth > 0)
          srid = 31960+zone;
        else if (fsnorth == 0)
          srid = 31954+zone;
      }
    }
  }

  if (srid == TE_UNKNOWN_SRS)
  {
    //if we have an Unkwnown SRID, but the WKT and PROJ4 representations of the SRS were correctly decoded,
    //we have a valid user defined projection. Lets register it
    if (wkt.empty() == false && proj4.empty() == false)
    {
      srid = te::srs::SpatialReferenceSystemManager::getInstance().getNewUserDefinedSRID();
      std::string newName = "USER:" + boost::lexical_cast<std::string>(srid);
      te::srs::SpatialReferenceSystemManager::getInstance().add(newName, proj4, wkt, srid, "USER");
    }
  }
  
  return srid;
}

OGRSpatialReference* te::ogr::Convert2OGRProjection(int srid)
{
  std::unique_ptr< OGRSpatialReference > osrs( new OGRSpatialReference() );
  
  OGRErr error = OGRERR_NONE;
  try
  {
    error = osrs->importFromEPSG(srid);
    if (error != OGRERR_NONE)
    {
      std::string proj4Str =
        te::srs::SpatialReferenceSystemManager::getInstance().getP4Txt(srid);

      if (!proj4Str.empty())
      {
        char* proj4StrPtr = (char*)proj4Str.c_str();
        error = osrs->importFromProj4(proj4StrPtr);
      }
    }
  }
  catch( te::common::Exception& )
  {
    error = OGRERR_UNSUPPORTED_SRS;
  }
  catch (std::exception&)
  {
    error = OGRERR_UNSUPPORTED_SRS;
  }
  
  if( error != OGRERR_NONE )
  {
    try
    {
      std::string wktStr = 
        te::srs::SpatialReferenceSystemManager::getInstance().getWkt( srid );
      
      if( !wktStr.empty() )
      {
        char* wktStrPtr = (char*)wktStr.c_str();
        error = osrs->importFromWkt( &wktStrPtr );
      }
    }
    catch( te::common::Exception& )
    {
      error = OGRERR_UNSUPPORTED_SRS;
    }
  }
  
  if(error != OGRERR_NONE)
    throw(te::common::Exception(TE_TR("Error converting spatial reference system.")));
  
  return osrs.release();
}

void te::ogr::Convert2TerraLib(OGRFeatureDefn* featDef,  te::da::DataSetType* dt,
  const std::string& geometryPropertyName, int srs)
{
  assert(dt);

  int nFields = featDef->GetFieldCount();
  
  for(int i = 0; i < nFields; i++)
  {
    OGRFieldDefn* fieldDef = featDef->GetFieldDefn(i); 
    te::dt::Property* p = Convert2TerraLib(fieldDef);
    dt->add(p);
  }

  std::string ogrFieldName;
  int terraLibSRID = 0;

  for( int geomFieldIdx = 0 ; geomFieldIdx < featDef->GetGeomFieldCount() ;
    ++geomFieldIdx )
  {
    const OGRGeomFieldDefn * ogrFieldDefPtr = featDef->GetGeomFieldDefn(
      geomFieldIdx );
    assert( ogrFieldDefPtr );

    const OGRwkbGeometryType ogrGeomType = ogrFieldDefPtr->GetType();

    if( ogrGeomType != wkbNone )
    {
      if( geometryPropertyName.empty() )
      {
        ogrFieldName = std::string( ogrFieldDefPtr->GetNameRef() );

        if( ogrFieldName.empty() )
        {
          ogrFieldName = "OGR_GEOMETRY";

          if( geomFieldIdx > 0 )
          {
            ogrFieldName.append( "_" + std::to_string( geomFieldIdx ) );
          }
        }
      }
      else
      {
        ogrFieldName = geometryPropertyName;

        if( geomFieldIdx > 0 )
        {
          ogrFieldName.append( "_" + std::to_string( geomFieldIdx ) );
        }
      }

      if( srs == 0 )
      {
        OGRSpatialReference * spatRefPtr = ogrFieldDefPtr->GetSpatialRef();
        if( spatRefPtr )
        {
          terraLibSRID = te::ogr::Convert2TerraLibProjection( spatRefPtr );
        }
        else
        {
          terraLibSRID = 0;
        }
      }
      else
      {
        terraLibSRID = srs;
      }

      te::gm::GeomType terralibGeomType = Convert2TerraLib(ogrGeomType);

      te::gm::GeometryProperty* geometryProperty = new te::gm::GeometryProperty(
        ogrFieldName, terraLibSRID, terralibGeomType);
      te::dt::Property* existingGeometryProperty = dt->getProperty(
        geometryProperty->getName() );

      if (existingGeometryProperty == nullptr)
      {
        //if the geometry propery does not exist in the dataSet, we must add it
        dt->add(geometryProperty);
      }
      else
      {
        //if the geometry already exists, we must replace it with the correct metadata
        dt->replace(existingGeometryProperty, geometryProperty);
        delete existingGeometryProperty;
      }
    }
  }
}

te::da::DataSetType* te::ogr::Convert2TerraLib(OGRFeatureDefn* featDef, const std::string& geometryPropertyName, int srs)
{
  te::da::DataSetType* dt = new  te::da::DataSetType(featDef->GetName());
  
  dt->setTitle(featDef->GetName());
  
  te::ogr::Convert2TerraLib(featDef,dt, geometryPropertyName, srs);

  return dt;
}

int GetTerraLibDataTypeFromSQLiteDataType(const std::string& sqliteDataType)
{
  std::string upperTypeName(sqliteDataType);
  std::transform(upperTypeName.begin(), upperTypeName.end(), upperTypeName.begin(), ::toupper);

  static std::map<std::string, int> mapTypes;
  if (mapTypes.empty())
  {
    mapTypes["INT"] = te::dt::INT32_TYPE;
    mapTypes["INTEGER"] = te::dt::INT32_TYPE;
    mapTypes["BIGINT"] = te::dt::INT64_TYPE;
    mapTypes["UNSIGNED BIG INT"] = te::dt::UINT64_TYPE;

    mapTypes["TEXT"] = te::dt::STRING_TYPE;
    mapTypes["CLOB"] = te::dt::STRING_TYPE;

    mapTypes["REAL"] = te::dt::DOUBLE_TYPE;
    mapTypes["DOUBLE"] = te::dt::DOUBLE_TYPE;
    mapTypes["DOUBLE PRECISION"] = te::dt::DOUBLE_TYPE;
    mapTypes["FLOAT"] = te::dt::DOUBLE_TYPE;

    mapTypes["BOOLEAN"] = te::dt::BOOLEAN_TYPE;
    mapTypes["DATE"] = te::dt::DATETIME_TYPE;
    mapTypes["DATETIME"] = te::dt::DATETIME_TYPE;
    
    mapTypes["GEOMETRY"] = te::dt::GEOMETRY_TYPE;
    mapTypes["POINT"] = te::dt::GEOMETRY_TYPE;
    mapTypes["MULTIPOINT"] = te::dt::GEOMETRY_TYPE;
    mapTypes["LINE"] = te::dt::GEOMETRY_TYPE;
    mapTypes["MULTILINE"] = te::dt::GEOMETRY_TYPE;
    mapTypes["LINESTRING"] = te::dt::GEOMETRY_TYPE;
    mapTypes["MULTILINESTRING"] = te::dt::GEOMETRY_TYPE;
    mapTypes["POLYGON"] = te::dt::GEOMETRY_TYPE;
    mapTypes["MULTIPOLYGON"] = te::dt::GEOMETRY_TYPE;
    mapTypes["GEOMETRYCOLLECTION"] = te::dt::GEOMETRY_TYPE;
  }

  std::map<std::string, int>::iterator it = mapTypes.find(upperTypeName);
  if (it != mapTypes.end())
  {
    return it->second;
  }

  if (upperTypeName.find("CHAR") != std::string::npos)
  {
    return te::dt::STRING_TYPE;
  }

  if (upperTypeName.find("INT") != std::string::npos)
  {
    return te::dt::INT32_TYPE;
  }

  return te::dt::UNKNOWN_TYPE;
}

bool TableHasAutoIncrement(GDALDataset* gdalDataSet, const std::string& dataSetName)
{
  std::string query = "SELECT * FROM sqlite_master";
  query += " WHERE tbl_name = '" + dataSetName + "' AND sql LIKE '%AUTOINCREMENT%'";

  OGRLayer* ogrLayerMetadata = gdalDataSet->ExecuteSQL(query.c_str(), nullptr, "SQLITE");
  if (ogrLayerMetadata == nullptr)
  {
    return false;
  }

  bool hasAutoIncrement = false;

  OGRFeature* ogrFeature = ogrLayerMetadata->GetNextFeature();
  if (ogrFeature != nullptr)
  {
    hasAutoIncrement = true;
  }

  OGRFeature::DestroyFeature(ogrFeature);
  gdalDataSet->ReleaseResultSet(ogrLayerMetadata);

  return hasAutoIncrement;
}

te::gm::GeometryProperty* GetGeometryPropertySQLite(GDALDataset* gdalDataSet, const std::string& dataSetName)
{
  std::string query = "Select * from geometry_columns where upper(f_table_name) = upper('" + dataSetName + "')";

  OGRLayer* ogrLayerMetadata = gdalDataSet->ExecuteSQL(query.c_str(), nullptr, "SQLITE");
  if (ogrLayerMetadata == nullptr)
  {
    gdalDataSet->ReleaseResultSet(ogrLayerMetadata);
    return nullptr;
  }

  OGRFeature* ogrFeature = ogrLayerMetadata->GetNextFeature();
  if (ogrFeature == nullptr)
  {
    return nullptr;
  }

  int srid = ogrFeature->GetFieldAsInteger("srid");
  te::gm::GeomType geometryType = static_cast<te::gm::GeomType>(ogrFeature->GetFieldAsInteger("geometry_type"));
  std::string geometryColumn = ogrFeature->GetFieldAsString("f_geometry_column");
  
  OGRFeature::DestroyFeature(ogrFeature);
  gdalDataSet->ReleaseResultSet(ogrLayerMetadata);

  te::gm::GeometryProperty* geometryProperty = new te::gm::GeometryProperty(geometryColumn, srid, geometryType);
  return geometryProperty;
}

te::gm::GeometryProperty* GetGeometryPropertyGPKG(GDALDataset* gdalDataSet, const std::string& dataSetName)
{
  std::string query = "Select * from gpkg_geometry_columns where upper(table_name) = upper('" + dataSetName + "')";

  OGRLayer* ogrLayerMetadata = gdalDataSet->ExecuteSQL(query.c_str(), nullptr, "SQLITE");
  if (ogrLayerMetadata == nullptr)
  {
    return nullptr;
  }

  OGRFeature* ogrFeature = ogrLayerMetadata->GetNextFeature();
  if (ogrFeature == nullptr)
  {
    return nullptr;
  }

  int srid = ogrFeature->GetFieldAsInteger("srs_id");
  std::string geometryTypeName = ogrFeature->GetFieldAsString("geometry_type_name");
  std::string geometryColumn = ogrFeature->GetFieldAsString("column_name");
  te::gm::GeomType geometryType = te::gm::Geometry::getGeomTypeId(geometryTypeName);

  OGRFeature::DestroyFeature(ogrFeature);
  gdalDataSet->ReleaseResultSet(ogrLayerMetadata);

  te::gm::GeometryProperty* geometryProperty = new te::gm::GeometryProperty(geometryColumn, srid, geometryType);
  return geometryProperty;
}

void FixDataTypes(GDALDataset* gdalDataSet, const std::string& dataSetName, te::da::DataSetType* dataSetType)
{
  std::string driverName = gdalDataSet->GetDriverName();
  if (driverName != "SQLite" )
  {
    return;
  }

  if (dataSetName == "SELECT")
  {
    return;
  }

  std::string sql = "PRAGMA table_info(" + dataSetName + ")";
  OGRLayer* ogrLayerMetadata = gdalDataSet->ExecuteSQL(sql.c_str(), nullptr, "SQLITE");
  if (ogrLayerMetadata == nullptr)
  {
    return;
  }

  te::da::PrimaryKey* primaryKey = nullptr;

  ogrLayerMetadata->ResetReading();
  OGRFeature* ogrFeature = ogrLayerMetadata->GetNextFeature();
  while (ogrFeature != nullptr)
  {
    std::string propertyName = ogrFeature->GetFieldAsString("name");
    std::string propertyTypeStr = ogrFeature->GetFieldAsString("type");
    bool isPK = static_cast<bool>(ogrFeature->GetFieldAsInteger("pk"));
    bool isNotNull = static_cast<bool>(ogrFeature->GetFieldAsInteger("notnull"));

    int propertyType = GetTerraLibDataTypeFromSQLiteDataType(propertyTypeStr);

    te::dt::Property* property = dataSetType->getProperty(propertyName);
    if (property == nullptr && isPK == true)
    {
      //if the property is not there and it is a primary key, we must add it
      te::dt::SimpleProperty* simpleProperty = new te::dt::SimpleProperty(propertyName, te::dt::INT32_TYPE);
      if (TableHasAutoIncrement(gdalDataSet, dataSetName) == true)
      {
        simpleProperty->setAutoNumber(true);
        simpleProperty->setRequired(true);
      }

      property = simpleProperty;
      dataSetType->insert(0, property);
    }

    if((propertyType != te::dt::UNKNOWN_TYPE && propertyType != property->getType()) || property->getType() == te::dt::GEOMETRY_TYPE )
    {
      te::dt::SimpleProperty* newProperty = nullptr;
      if (propertyType == te::dt::GEOMETRY_TYPE)
      {
        if (driverName == "SQLite")
        {
          newProperty = GetGeometryPropertySQLite(gdalDataSet, dataSetName);
        }
        else
        {
          newProperty = GetGeometryPropertyGPKG(gdalDataSet, dataSetName);
        }
        newProperty->setRequired(isNotNull);
      }
      else
      {
        newProperty = new te::dt::SimpleProperty(propertyName, propertyType, isNotNull);
      }

      dataSetType->replace(property, newProperty);
      delete property;

      property = newProperty;
    }

    if (isPK == true)
    {
      if (primaryKey == nullptr)
      {
        primaryKey = new te::da::PrimaryKey("pk_" + dataSetName, dataSetType);
      }

      primaryKey->add(property);
    }

    OGRFeature::DestroyFeature(ogrFeature);
    ogrFeature = ogrLayerMetadata->GetNextFeature();
  }

  gdalDataSet->ReleaseResultSet(ogrLayerMetadata);
}

te::da::DataSetType* te::ogr::GetDataSetType(GDALDataset* gdalDataSet, OGRLayer* ogrLayer, int srs)
{
  std::string driverName = gdalDataSet->GetDriverName();
  std::string dataSetName = ogrLayer->GetName();
  std::string fidColumn = ogrLayer->GetFIDColumn();

  //this column name '_ogr_geometry_' is created when we run queries like 'select * from dataSetName' in a ShapeFile
  //So we must adjust this name because the name considered in the OGR_DIALECT is 'OGR_GEOMETRY'
  std::string ogrFieldName;
  if(
      ( ogrLayer->GetLayerDefn()->GetGeomFieldCount() == 1 )
      &&
      (
        std::strcmp( ogrLayer->GetLayerDefn()->GetGeomFieldDefn(0 )->GetNameRef(),
          "_ogr_geometry_" ) == 0
      )
    )
  {
    ogrFieldName = "OGR_GEOMETRY";
  }

  OGRFeatureDefn* featDef = ogrLayer->GetLayerDefn();

  te::da::DataSetType* dataSetType = Convert2TerraLib(featDef, ogrFieldName, srs);

  FixDataTypes(gdalDataSet, dataSetName, dataSetType);

  //handles wkbUnknown
  OGRwkbGeometryType ogrGeomType = featDef->GetGeomType();
  if (ogrGeomType == wkbUnknown)
  {
    ogrLayer->ResetReading();
    OGRFeature* ogrFeature = ogrLayer->GetNextFeature();
    if (ogrFeature != nullptr)
    {
      OGRGeometry* ogrGeometry = ogrFeature->GetGeometryRef();
      if (ogrGeometry != nullptr)
      {
        ogrGeomType = ogrGeometry->getGeometryType();
      }
      OGRFeature::DestroyFeature(ogrFeature);
    }

    if (ogrGeomType != wkbUnknown)
    {
      te::gm::GeometryProperty* geometryProperty = te::da::GetFirstGeomProperty(dataSetType);
      if (geometryProperty != nullptr)
      {
        geometryProperty->setGeometryType(Convert2TerraLib(ogrGeomType));
      }
    }
    ogrLayer->ResetReading();
  }

  //if we have information related to the FID but we dont have any spatial column, we do not interfere on the FID
  te::da::PrimaryKey* primaryKey = dataSetType->getPrimaryKey();
  if (primaryKey == nullptr && driverName != "SQLite")
  {
    if (fidColumn.empty() == true)
    {
      fidColumn = "FID";
    }

    std::size_t fidPos = te::da::GetPropertyPos(dataSetType, fidColumn);
    if (fidPos != std::string::npos)
    {
      //we must ensure to use the same existing name (INCLUSING THE CASE)
      te::dt::Property* fidProperty = dataSetType->getProperty(fidPos);
      fidColumn = fidProperty->getName();
    }


    te::dt::SimpleProperty* fidSimpleProperty = new te::dt::SimpleProperty(fidColumn, te::dt::INT32_TYPE, true);
    fidSimpleProperty->setAutoNumber(true);

    if (fidPos == std::string::npos)
    {
      dataSetType->insert(0, fidSimpleProperty);
    }
    else
    {
      te::dt::Property* fidProperty = dataSetType->getProperty(fidPos);
      dataSetType->replace(fidProperty, fidSimpleProperty);
      delete fidProperty;
    }

    te::da::PrimaryKey* pk = new te::da::PrimaryKey(fidColumn, dataSetType);
    pk->add(fidSimpleProperty);
  }

  //handles srid
  te::gm::GeometryProperty* geometryProperty = te::da::GetFirstGeomProperty(dataSetType);
  if (geometryProperty != nullptr && geometryProperty->getSRID() == TE_UNKNOWN_SRS)
  {
    OGRSpatialReference* ogrSpatialReference = ogrLayer->GetSpatialRef();
    if (ogrSpatialReference != nullptr)
    {
      srs = te::ogr::Convert2TerraLibProjection(ogrSpatialReference);
      geometryProperty->setSRID(srs);
    }
  }

  return dataSetType;
}

OGRFeatureDefn* te::ogr::Convert2OGR(te::da::DataSetType* dt)
{
  OGRFeatureDefn* featDef = new OGRFeatureDefn(dt->getName().c_str());
  std::vector<te::dt::Property*> props = dt->getProperties();

  for(unsigned int i = 0; i < props.size(); i++)
  {
    te::dt::Property* p = props[i];
    if(p->getType() != te::dt::GEOMETRY_TYPE)
      featDef->AddFieldDefn(Convert2OGR(p));
  }

  if(dt->hasGeom())
  {
    te::gm::GeometryProperty* geom = te::da::GetFirstGeomProperty(dt);
    featDef->SetGeomType(Convert2OGR(geom->getGeometryType()));
  }

  return featDef;
}

te::dt::Property* te::ogr::Convert2TerraLib(OGRFieldDefn* fieldDef)
{
  te::dt::Property* p = nullptr;
  std::string name = fieldDef->GetNameRef();
  bool isRequired = !fieldDef->IsNullable();
  switch(fieldDef->GetType())
  {
    case OFTInteger:
      p = new te::dt::SimpleProperty(name, te::dt::INT32_TYPE, isRequired);
    break;

    case OFTInteger64:
      p = new te::dt::SimpleProperty(name, te::dt::INT64_TYPE, isRequired);
      break;

    case OFTIntegerList:
      p = new te::dt::ArrayProperty(name, new te::dt::SimpleProperty(name, te::dt::INT32_TYPE, isRequired));
    break;

    case OFTString:
    {
      te::dt::StringProperty* sp = nullptr;

      if(fieldDef->GetWidth() == 0)
        sp = new te::dt::StringProperty(name, te::dt::STRING, 0, isRequired);
      else
        sp = new te::dt::StringProperty(name, te::dt::VAR_STRING, fieldDef->GetWidth(), isRequired);

      p = sp;
    }
    break;

    case OFTStringList:
      p = new te::dt::ArrayProperty(name, new te::dt::StringProperty(name, te::dt::STRING, 0, isRequired));
    break;

    case OFTReal:
      p = new te::dt::SimpleProperty(name, te::dt::DOUBLE_TYPE, isRequired);
    break;

    case OFTRealList:
      p = new te::dt::ArrayProperty(name, new te::dt::SimpleProperty(name, te::dt::DOUBLE_TYPE, isRequired));
    break;

    case OFTBinary:
      p = new te::dt::SimpleProperty(name, te::dt::BYTE_ARRAY_TYPE, isRequired);
    break;

    case OFTDate:
      p = new te::dt::DateTimeProperty(name, te::dt::DATE, 0, isRequired);
    break;

    case OFTTime:
      p = new te::dt::DateTimeProperty(name, te::dt::TIME_DURATION, 0, isRequired);
    break;

    case OFTDateTime:
      p = new te::dt::DateTimeProperty(name, te::dt::TIME_INSTANT, 0, isRequired);
    break;

    default:
      throw(te::common::Exception(TE_TR("Unexpected data type.")));
  }

  return p;
}

OGRFieldDefn* te::ogr::Convert2OGR(te::dt::Property* p)
{
  OGRFieldDefn* fieldDef = new OGRFieldDefn(p->getName().c_str(), OFTInteger);
  switch(p->getType())
  {
    case te::dt::INT32_TYPE:
      return fieldDef;
    break;

    case te::dt::INT64_TYPE:
    {
      fieldDef->SetType(OFTInteger64);
      return fieldDef;
      break;
    }

    case te::dt::ARRAY_TYPE:
    {
      te::dt::ArrayProperty* at = dynamic_cast<te::dt::ArrayProperty*>(p);
      if(!at)
        throw(te::common::Exception(TE_TR("Unexpected data type.")));

      int elementType = at->getElementType()->getType();
      
      if(elementType == te::dt::INT32_TYPE)
        fieldDef->SetType(OFTIntegerList);
      else if(elementType == te::dt::STRING_TYPE)
        fieldDef->SetType(OFTStringList);
      else if(elementType == te::dt::DOUBLE_TYPE)
        fieldDef->SetType(OFTRealList);
      else
        throw(te::common::Exception(TE_TR("Unsupported data type by OGR.")));
      break;
    }

    case te::dt::STRING_TYPE:
    {
      fieldDef->SetType(OFTString);

      te::dt::StringProperty* sp = dynamic_cast<te::dt::StringProperty*>(p);
      if (!sp)
      {
        throw(te::common::Exception(TE_TR("Unexpected data type.")));
      }

      fieldDef->SetWidth(static_cast<int>(sp->size()));
      break;
    } 

    case te::dt::DOUBLE_TYPE:
      fieldDef->SetType(OFTReal);
    break;

    case te::dt::NUMERIC_TYPE:
    {
      te::dt::NumericProperty* np = dynamic_cast<te::dt::NumericProperty*>(p);
      if(!np)
      {
        throw(te::common::Exception(TE_TR("Unexpected data type.")));
      }

      fieldDef->SetType(OFTReal);
      fieldDef->SetPrecision(np->getScale());
      break;
    }
      

    case te::dt::BYTE_ARRAY_TYPE:
      fieldDef->SetType(OFTBinary);
    break;

    case te::dt::DATETIME_TYPE:
    {
      const te::dt::DateTimeProperty* dp = dynamic_cast<const te::dt::DateTimeProperty*>(p);
      if(!dp)
      {
        throw(te::common::Exception(TE_TR("Unexpected data type.")));
      }

      te::dt::DateTimeType elementType = dp->getSubType();
      if(elementType == te::dt::DATE)
        fieldDef->SetType(OFTDate);
      else if(elementType == te::dt::TIME_DURATION)
        fieldDef->SetType(OFTTime);
      else if(elementType == te::dt::TIME_INSTANT)
        fieldDef->SetType(OFTDateTime);
      else
        throw(te::common::Exception(TE_TR("Unsupported data type by OGR.")));
      break;
    }

    default:
      throw(te::common::Exception(TE_TR("Unsupported data type by OGR.")));
  }
  
  return fieldDef;
}

OGRGeomFieldDefn* te::ogr::Convert2OGRGeom(te::dt::Property* p)
{
  if( p->getType() != te::dt::GEOMETRY_TYPE )
  {
    throw(te::common::Exception(TE_TR("Unsupported data type by OGR.")));
  }

  te::gm::GeometryProperty* at = dynamic_cast<te::gm::GeometryProperty*>(p);
  if(!at)
    throw(te::common::Exception(TE_TR("Unexpected data type.")));

  return new OGRGeomFieldDefn( at->getName().c_str(),
    te::ogr::Convert2OGR( at->getGeometryType() ) );
}

te::gm::GeomType te::ogr::Convert2TerraLib(OGRwkbGeometryType ogrGeomType)
{
  switch(ogrGeomType)
  {
    case wkbPoint:
      return te::gm::MultiPointType;

    case wkbPointM:
      return te::gm::MultiPointMType;

    case wkbPointZM:
      return te::gm::MultiPointZMType;

    case wkbPoint25D:
      return te::gm::MultiPointZType;

    case wkbMultiPoint:
      return te::gm::MultiPointType;

    case wkbMultiPointM:
      return te::gm::MultiPointMType;

    case wkbMultiPointZM:
      return te::gm::MultiPointZMType;

    case wkbMultiPoint25D:
      return te::gm::MultiPointZType;


    case wkbLinearRing:
      return te::gm::MultiLineStringType;

    case wkbLineString:
      return te::gm::MultiLineStringType;

    case wkbLineStringM:
      return te::gm::MultiLineStringMType;

    case wkbLineStringZM:
      return te::gm::MultiLineStringZMType;

    case wkbLineString25D:
      return te::gm::MultiLineStringZType;

    case wkbMultiLineString:
      return te::gm::MultiLineStringType;

    case wkbMultiLineStringM:
      return te::gm::MultiLineStringMType;

    case wkbMultiLineStringZM:
      return te::gm::MultiLineStringZMType;

    case wkbMultiLineString25D:
      return te::gm::MultiLineStringZType;
 
    case wkbCurvePolygon:
    case wkbPolygon:
      return te::gm::MultiPolygonType;

    case wkbCurvePolygonM:
    case wkbPolygonM:
      return te::gm::MultiPolygonMType;

    case wkbCurvePolygonZM:
    case wkbPolygonZM:
      return te::gm::MultiPolygonZMType;

    case wkbPolygon25D:
      return te::gm::MultiPolygonZType;

    case wkbMultiPolygon:
      return te::gm::MultiPolygonType;

    case wkbMultiPolygonM:
      return te::gm::MultiPolygonMType;

    case wkbMultiPolygonZM:
      return te::gm::MultiPolygonZMType;

    case wkbMultiPolygon25D:
      return te::gm::MultiPolygonZType;


    case wkbGeometryCollection:
      return te::gm::GeometryCollectionType;

    case wkbGeometryCollectionM:
      return te::gm::GeometryCollectionMType;

    case wkbGeometryCollectionZM:
      return te::gm::GeometryCollectionZMType;

   case wkbGeometryCollection25D:
      return te::gm::GeometryCollectionZType;


    case wkbMultiSurface:
      return te::gm::MultiPolygonType;

    case wkbMultiSurfaceM:
      return te::gm::MultiPolygonMType;

    case wkbMultiSurfaceZM:
      return te::gm::MultiPolygonZMType;

    case wkbMultiSurfaceZ:
      return te::gm::MultiPolygonZType;


    case wkbPolyhedralSurface:
      return te::gm::PolyhedralSurfaceType;

    case wkbPolyhedralSurfaceM:
      return te::gm::PolyhedralSurfaceMType;

    case wkbPolyhedralSurfaceZM:
      return te::gm::PolyhedralSurfaceZMType;

    case wkbPolyhedralSurfaceZ:
      return te::gm::PolyhedralSurfaceZType;


    case wkbTIN:
      return te::gm::TINType;

    case wkbTINM:
      return te::gm::TINMType;

    case wkbTINZM:
      return te::gm::TINZMType;

    case wkbTINZ:
      return te::gm::TINZType;

    case wkbTriangle:
      return te::gm::TriangleType;

    case wkbTriangleM:
      return te::gm::TriangleMType;

    case wkbTriangleZM:
      return te::gm::TriangleZMType;

    case wkbTriangleZ:
      return te::gm::TriangleZType;


    case wkbUnknown:
      return te::gm::GeometryType;
      
    default:  
      return te::gm::UnknownGeometryType;
  }
}

OGRwkbGeometryType te::ogr::Convert2OGR(te::gm::GeomType geomType)
{
  switch(geomType)
  {
    case te::gm::UnknownGeometryType:
      return wkbUnknown;

    case te::gm::PointType:
      return wkbPoint;

    case te::gm::PointZType:
      return wkbPoint25D;

    case te::gm::PointMType:
      return wkbPointM;

    case te::gm::PointZMType:
      return wkbPointZM;


    case te::gm::LineStringType:
      return wkbLineString;

    case te::gm::LineStringZType:
      return wkbLineString25D;

    case te::gm::LineStringMType:
      return wkbLineStringM;

    case te::gm::LineStringZMType:
      return wkbLineStringZM;


    case te::gm::PolygonType:
      return wkbPolygon;

    case te::gm::PolygonZType:
      return wkbPolygon25D;

    case te::gm::PolygonMType:
      return wkbPolygonM;

    case te::gm::PolygonZMType:
      return wkbPolygonZM;


    case te::gm::MultiPointType:
      return wkbMultiPoint;

    case te::gm::MultiPointZType:
      return wkbMultiPoint25D;

    case te::gm::MultiPointMType:
      return wkbMultiPointM;

    case te::gm::MultiPointZMType:
      return wkbMultiPointZM;


    case te::gm::MultiLineStringType:
      return wkbMultiLineString;

    case te::gm::MultiLineStringZType:
      return wkbMultiLineString25D;

    case te::gm::MultiLineStringMType:
      return wkbMultiLineStringM;

    case te::gm::MultiLineStringZMType:
      return wkbMultiLineStringZM;

    
    case te::gm::MultiPolygonType:
      return wkbMultiPolygon;

    case te::gm::MultiPolygonZType:
      return wkbMultiPolygon25D;

    case te::gm::MultiPolygonMType:
      return wkbMultiPolygonM;

    case te::gm::MultiPolygonZMType:
      return wkbMultiPolygonZM;


    case te::gm::GeometryCollectionType:
      return wkbGeometryCollection;

    case te::gm::GeometryCollectionZType:
      return wkbGeometryCollection25D;

    case te::gm::GeometryCollectionMType:
      return wkbGeometryCollectionM;

    case te::gm::GeometryCollectionZMType:
      return wkbGeometryCollectionZM;


    case te::gm::MultiSurfaceType:
      return wkbMultiSurface;

    case te::gm::MultiSurfaceMType:
      return wkbMultiSurfaceM;

    case te::gm::MultiSurfaceZMType:
      return wkbMultiSurfaceZM;

    case te::gm::MultiSurfaceZType:
      return wkbMultiSurfaceZ;


    case te::gm::PolyhedralSurfaceType:
      return wkbPolyhedralSurface;

    case te::gm::PolyhedralSurfaceZType:
      return wkbPolyhedralSurfaceZ;

    case te::gm::PolyhedralSurfaceMType:
      return wkbPolyhedralSurfaceM;

    case te::gm::PolyhedralSurfaceZMType:
      return wkbPolyhedralSurfaceZM;


    case te::gm::TINType:
      return wkbTIN;

    case te::gm::TINMType:
      return wkbTINM;

    case te::gm::TINZMType:
      return wkbTINZM;

    case te::gm::TINZType:
      return wkbTINZ;


    case te::gm::TriangleType:
      return wkbTriangle;

    case te::gm::TriangleMType:
      return wkbTriangleM;

    case te::gm::TriangleZMType:
      return wkbTriangleZM;

    case te::gm::TriangleZType:
      return wkbTriangleZ;

    default:
      throw(te::common::Exception(TE_TR("Unsupported geometry type by OGR Driver.")));
  }
}

std::string te::ogr::GetDriverName(const std::string& path)
{
  boost::filesystem::path mpath(path.c_str());

  std::string ext = mpath.extension().string();
  
  if(ext == ".shp" || ext == ".SHP")
    return std::string("ESRI Shapefile");

  if(ext == ".mif" || ext == ".MIF")
    return std::string("Mapinfo File");
  
  if(ext == ".kml" || ext == ".KML")
    return std::string("KML");
  
  if(ext == ".geojson" || ext == ".GEOJSON")
    return std::string("GeoJSON");
  
  if(ext == ".gml" || ext == ".GML")
    return std::string("GML");
  
  if(ext == ".dxf" || ext == ".DXF")
    return std::string("DXF");

  if(ext == ".dgn" || ext == ".DGN")
    return std::string("DGN");

  if (ext == ".csv" || ext == ".CSV")
    return std::string("CSV");

  if (ext == ".sqlite" || ext == ".SQLITE")
    return std::string("SQLite");

  if (ext == ".gpkg" || ext == ".GPKG")
    return std::string("GPKG");

  return "";
}

std::vector<std::string> te::ogr::GetOGRDrivers(bool filterCreate)
{
  std::vector<std::string> drivernames;

  int ndrivers = GetGDALDriverManager()->GetDriverCount();

  for (int i = 0; i < ndrivers; ++i)
  {
    GDALDriver* driver = GetGDALDriverManager()->GetDriver(i);
    if (filterCreate && !OGR_Dr_TestCapability(driver, ODrCCreateDataSource))
        continue;
    drivernames.push_back(driver->GetDescription());
  }

  return drivernames;
}

std::string te::ogr::GetOGRConnectionInfo(const std::map<std::string, std::string>& connInfo)
{
  std::map<std::string, std::string>::const_iterator it = connInfo.find("URI");

  if(it != connInfo.end())
    return it->second;

  it = connInfo.find("SOURCE");

  if(it != connInfo.end())
    return it->second;

  throw te::common::Exception(TE_TR("Invalid data source connection information!"));
}

std::string te::ogr::RemoveSpatialSql(const std::string& sql)
{
  // Try find AND
  std::size_t pos = sql.find("AND st_intersects");

  // Try find without AND
  if(pos == std::string::npos)
    pos = sql.find("WHERE st_intersects");

  if(pos == std::string::npos)
    return sql;

  std::string newQuery;

  std::size_t pos2 = sql.find("))", pos);
  newQuery = sql.substr(0, pos);
  newQuery += sql.substr(pos2 + 2);

  return newQuery;
}

TEOGREXPORT boost::mutex & te::ogr::getStaticMutex()
{
  static boost::mutex getStaticMutexStaticMutex;
  return getStaticMutexStaticMutex;
}

void te::ogr::OGRErrorHandler(CPLErr eErrClass, int /*errNo*/,
                              const char* msg)
{
  std::string message(msg);
  switch (eErrClass)
  {
  case CE_None:
    TE_LOG_INFO(message);
    break;
  case CE_Debug:
    TE_LOG_DEBUG(message);
    break;
  case CE_Warning:
    TE_LOG_WARN(message);
    break;
  case CE_Failure:
    TE_LOG_ERROR(message);
    break;
  case CE_Fatal:
    TE_LOG_FATAL(message);
    break;

  default:
    break;
  }
}

GDALDataset* te::ogr::CreateGDALDataSet(const std::string& fileName, unsigned int nOpenFlags)
{
  GDALDataset* ds = nullptr;

  std::string driverName = GetDriverName(fileName);
  if (driverName.empty() == true)
  {
    ds = static_cast<GDALDataset*>(GDALOpenEx(fileName.c_str(), nOpenFlags, nullptr, nullptr, nullptr));
  }
  else
  {
    const char* papszAllowedDrivers[] = { driverName.c_str(), nullptr };
    ds = static_cast<GDALDataset*>(GDALOpenEx(fileName.c_str(), nOpenFlags, papszAllowedDrivers, nullptr, nullptr));
  }

  return ds;
}

OGRLayer* te::ogr::ExecuteSQL(GDALDataset* gdalDataSet, const std::string& query)
{
  std::string queryCopy(query);

  std::string driverName = gdalDataSet->GetDriverName();
  if (driverName == "ESRI Shapefile")
  {
    CPLSetConfigOption("SHAPE_ENCODING", te::core::CharEncoding::getEncodingName(te::core::EncodingType::UTF8).c_str());
  }

  std::string dialect;

  CPLErrorReset();

  OGRLayer* ogrLayer = nullptr;
  try
  {
    ogrLayer = gdalDataSet->ExecuteSQL(queryCopy.c_str(), nullptr, dialect.c_str());
  }
  catch(const std::exception&)
  {
    CPLErr error = CPLGetLastErrorType();
    if (error != CE_None)
    {
      std::string errorMessage = CPLGetLastErrorMsg();
      throw te::common::Exception(errorMessage);
    }
  }

  if (driverName == "ESRI Shapefile")
  {
    CPLSetConfigOption("SHAPE_ENCODING", "UTF-8");
  }

  if (ogrLayer == nullptr)
  {
    CPLErr error = CPLGetLastErrorType();
    if (error != CE_None)
    {
      std::string errorMessage = CPLGetLastErrorMsg();
      throw te::common::Exception(errorMessage);
    }
  }

  return ogrLayer;
}

bool te::ogr::InitValidator(te::da::FieldNameValidator& validator, OGRFeatureDefn* ogrFeatureDefinition)
{
  bool adjusted = false;
  for (int i = 0; i < ogrFeatureDefinition->GetFieldCount(); ++i)
  {
    OGRFieldDefn* fieldDefinition = ogrFeatureDefinition->GetFieldDefn(i);
    std::string fieldName = fieldDefinition->GetNameRef();

    if (validator.addFieldName(fieldName) == true)
    {
      adjusted = true;
    }
  }

  return adjusted;
}
