/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/postgis/Connection.cpp

  \brief A class that implements a connection to a PostgreSQL database.
*/

// TerraLib
#include "../core/translator/Translator.h"
#include "Connection.h"
#include "Exception.h"

// STL
#include <cassert>
#include <string>

// libpq
#include <libpq-fe.h>

// Boost
#include <boost/format.hpp>

/*! \brief This will just close the PostgreSQL notice processor. */
static void PostGISNoticeProcessor(void* /*arg*/, const char* /*pszMessage*/);

PGresult* te::pgis::Connection::query(const std::string& query)
{
  PGresult* result = PQexecParams(m_pgconn, query.c_str(), 0, nullptr, nullptr, nullptr, nullptr, 1);

  if( result )
  {
    if( PQresultStatus(result) != PGRES_TUPLES_OK )
    {
      boost::format errmsg(TE_TR("Could not retrieve the dataset due to the following error: %1%."));
      errmsg = errmsg % PQerrorMessage(m_pgconn);

      PQclear(result);

      throw Exception(errmsg.str());
    }
  }
  else
  {
    throw Exception( TE_TR("Could not retrieve the dataset" ) );
  }

  return result;
}

void te::pgis::Connection::execute(const std::string& command)
{
  PGresult* result = PQexec(m_pgconn, command.c_str());

  if((PQresultStatus(result) != PGRES_COMMAND_OK) &&
     (PQresultStatus(result) != PGRES_TUPLES_OK))
  {
    boost::format errmsg(TE_TR("Could not execute the sql statement due to the following error: %1%."));
    
    errmsg = errmsg % PQerrorMessage(m_pgconn);

    std::string msg = errmsg.str();

    std::size_t idx = msg.find("invalid SRID");

    if(idx != std::string::npos)
       msg+= " PostGIS doens't recognize this projection. "
                                "Please, reproject your data to a correct one. "
                                "Or save your result as Vector File, and afterwards use Exchange to put the result "
                                "in PGIS changing the USER SRID to a valid one.";

    PQclear(result);

    throw Exception(msg);
  }

  PQclear(result);
}

te::pgis::Connection::~Connection()
{
  if(m_pgconn)
    PQfinish(m_pgconn);
}

te::pgis::Connection::Connection(ConnectionPool* pool, const std::string& conninfo, const std::string& cencoding, bool inuse) :
  m_id(0),
  m_pool(pool),
  m_pgconn(nullptr),
  m_inuse(inuse),
  m_lastuse(boost::posix_time::second_clock::local_time())
{
  static int newId = 0;
  m_id = ++newId;

  // it tries to create the connection object
  m_pgconn = PQconnectdb(conninfo.c_str());

  ConnStatusType status = PQstatus(m_pgconn);

  if(status != CONNECTION_OK)
  {
    boost::format errmsg(TE_TR("It was not possible to create a connection to the given data source due to the following error: %1%."));

    errmsg = errmsg % PQerrorMessage(m_pgconn);

    PQfinish(m_pgconn);

    m_pgconn = nullptr;

    throw Exception(errmsg.str());
  }

  PQsetNoticeProcessor(m_pgconn, PostGISNoticeProcessor, this);

  if(cencoding.empty())
    return;

  if(PQsetClientEncoding(m_pgconn, cencoding.c_str()) == -1)
  {
    boost::format errmsg(TE_TR("It was not possible to set the client encoding for the PostGIS data source due to the following error: %1%."));

    errmsg = errmsg % PQerrorMessage(m_pgconn);

    PQfinish(m_pgconn);

    m_pgconn = nullptr;

    throw Exception(errmsg.str());
  }
}

static void PostGISNoticeProcessor(void* /*arg*/, const char* /*pszMessage*/)
{
}
