/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralib/process/Connector.cpp

   \brief Abstraction that represents a Connector between parameters of a Task
 */

#include "Connector.h"
#include "../common/Exception.h"
#include "../common/StringUtils.h"
#include "../core/translator/Translator.h"

#include "ParameterValue.h"
#include "Task.h"
#include "TaskCapabilities.h"
#include "TaskParameters.h"

#include <string>

te::process::Connector::Connector()
  : m_id(te::common::GetUniqueID())
  , m_taskFrom(nullptr)
  , m_paramIndexFrom(std::string::npos)
  , m_taskTo(nullptr)
  , m_paramIndexTo(std::string::npos)
  , m_isValid(false)
{

}

te::process::Connector::~Connector()
{
}

const std::string& te::process::Connector::getId() const
{
  return m_id;
}

void te::process::Connector::setId(const std::string& id)
{
  m_id = id;
}

void te::process::Connector::clear()
{
  m_taskFrom = nullptr;
  m_paramIndexFrom = std::string::npos;
  m_taskTo = nullptr;
  m_paramIndexTo = std::string::npos;
  m_isValid = false;
}

bool te::process::Connector::isValid() const
{
  return m_isValid;
}

bool te::process::Connector::isReady() const
{
  if (isValid() == false)
  {
    return false;
  }

  //the connector is only ready if the task from has already finished its execution
  if (m_taskFrom->getState() != te::process::Task::State::FINISHED)
  {
    return false;
  }

  return true;
}

const te::process::Task* te::process::Connector::getTaskFrom() const
{
  return m_taskFrom;
}

std::size_t te::process::Connector::getParamIndexFrom() const
{
  return m_paramIndexFrom;
}

const te::process::Task* te::process::Connector::getTaskTo() const
{
  return m_taskTo;
}

std::size_t te::process::Connector::getParamIndexTo() const
{
  return m_paramIndexTo;
}

bool te::process::Connector::canAssociate(te::process::Task* taskFrom, std::size_t paramIndexFrom, te::process::Task* taskTo, std::size_t paramIndexTo, std::string& info)
{
  if (taskFrom == nullptr || taskTo == nullptr)
  {
    info = TE_TR("Invalid Task");

    return false;
  }

  if (taskFrom->getId() == taskTo->getId())
  {
    info = TE_TR("Source and Destination can not be the same task");

    return false;
  }

  if (paramIndexFrom == std::string::npos || paramIndexTo == std::string::npos)
  {
    info = TE_TR("Invalid Parameter Index");

    return false;
  }

  //now we check if the parameters data type are equal
  const te::process::TaskCapabilities* taskFromCapabilities = taskFrom->getCapabilities();
  const te::process::TaskCapabilities* taskToCapabilities = taskTo->getCapabilities();

  if (paramIndexFrom >= taskFromCapabilities->size())
  {
    info = TE_TR("Source Task does not have the given parameter index");

    return false;
  }

  if (paramIndexTo >= taskToCapabilities->size())
  {
    info = TE_TR("Destination Task does not have the given parameter index");

    return false;
  }

  const te::process::ParameterDescriptor& taskFromParameter = taskFromCapabilities->getParameterDescriptor(paramIndexFrom);
  const te::process::ParameterDescriptor& taskToParameter = taskToCapabilities->getParameterDescriptor(paramIndexTo);

  if (taskFromParameter.getDataType() != taskToParameter.getDataType())
  {
    info = TE_TR("The parameters datatype from both task must be the same");

    return false;
  }

  if (taskFromParameter.getDirection() == taskToParameter.getDirection())
  {
    info = TE_TR("Invalid connector direction");

    return false;
  }

  return true;
}

void te::process::Connector::associate(te::process::Task* taskFrom, std::size_t paramIndexFrom, te::process::Task* taskTo, std::size_t paramIndexTo)
{
  clear();

  std::string associationInfo;

  bool canAssociate = Connector::canAssociate(taskFrom, paramIndexFrom, taskTo, paramIndexTo, associationInfo);

  if (canAssociate == false)
  {
    throw te::common::Exception(associationInfo);
  }

  m_taskFrom = taskFrom;
  m_paramIndexFrom = paramIndexFrom;
  m_taskTo = taskTo;
  m_paramIndexTo = paramIndexTo;
  m_isValid = true;

  m_taskFrom->addConnector(this);
  m_taskTo->addConnector(this);
}

void te::process::Connector::disassociate()
{
  if (m_isValid)
  {
    m_taskFrom->removeConnector(this);
    m_taskTo->removeConnector(this);
  }

  clear();
}

void te::process::Connector::apply()
{
  if (isValid() == false)
  {
    throw te::common::Exception(TE_TR("The connector is not valid"));
  }

  if (isReady() == false)
  {
    throw te::common::Exception(TE_TR("Input task not ready"));
  }

  const te::process::ParameterValue* value = m_taskFrom->getParameter(m_paramIndexFrom);
  te::process::ParameterValue* newValue = nullptr;
  if (value != nullptr)
  {
    newValue = value->clone();
  }

  m_taskTo->setParameter(m_paramIndexTo, newValue, true);
}

te::process::ConnectorPtr::ConnectorPtr(te::process::Connector* pointer)
  : std::unique_ptr<Connector>(pointer)
{

}

te::process::ConnectorPtr::~ConnectorPtr() = default;

te::process::ConnectorSharedPtr::ConnectorSharedPtr(te::process::Connector* pointer)
  : std::shared_ptr<Connector>(pointer)
{

}

te::process::ConnectorSharedPtr::~ConnectorSharedPtr() = default;
