#include "DataStoreTaskProcess.h"

#include "Task.h"
#include "TaskCapabilities.h"
#include "SimpleDataParameterValue.h"

#include "../core/uri/URI.h"
#include "../datatype/Enums.h"
#include "../datatype/SimpleData.h"
#include "../dataaccess/dataset/DataSetAdapter.h"
#include "../dataaccess/dataset/DataSetTypeConverter.h"
#include "../dataaccess/datasource/DataSource.h"
#include "../dataaccess/datasource/DataSourceFactory.h"
#include "../dataaccess/utils/DataAccess.h"
#include "../dataaccess/utils/Utils.h"

te::process::DataStoreTask::DataStoreTask() : te::process::Task("DataStoreTask")
{
  te::process::TaskCapabilities* capabilities = new te::process::TaskCapabilities();
  capabilities->addParameterDescriptor(te::process::ParameterDescriptor("Data Source URI", te::dt::STRING_TYPE, true));
  capabilities->addParameterDescriptor(te::process::ParameterDescriptor("Data Set Name", te::dt::STRING_TYPE, true));
  capabilities->addParameterDescriptor(te::process::ParameterDescriptor("Input Data Access", te::dt::DATAACCESS_TYPE, true));
  capabilities->addParameterDescriptor(te::process::ParameterDescriptor("Output Data Access", te::dt::DATAACCESS_TYPE, false, te::process::ParameterDescriptor::Direction::OUTPUT));
  init(capabilities);
}

void te::process::DataStoreTask::setDataSourceType(const std::string& type)
{
  m_driver = type;
}

std::string te::process::DataStoreTask::getDataSourceType()
{
  return m_driver;
}


void te::process::DataStoreTask::runImpl()
{
  const te::process::ParameterValue* pValue1 = this->getParameter(0);
  const te::process::ParameterValue* pValue2 = this->getParameter(1);
  const te::process::ParameterValue* pValue3 = this->getParameter(2);

  const te::dt::AbstractData* absData1 = (static_cast<const te::process::SimpleDataParameterValue*>(pValue1))->getValue();
  const te::dt::AbstractData* absData2 = (static_cast<const te::process::SimpleDataParameterValue*>(pValue2))->getValue();
  const te::dt::AbstractData* absData3 = (static_cast<const te::process::SimpleDataParameterValue*>(pValue3))->getValue();

  const te::dt::String* data1 = static_cast<const te::dt::String*>(absData1);
  const te::dt::String* data2 = static_cast<const te::dt::String*>(absData2);
  const te::da::DataAccess* dataAccess = static_cast<const te::da::DataAccess*>(absData3);

  std::string uriStr = data1->getValue();
  std::string dataSetName =  data2->getValue();

  //get input data
  const te::da::DataSetType* inputDataSetType = dataAccess->getDataSetType();

  //open output datasource
  std::unique_ptr<te::da::DataSource> outputDataSource = te::da::DataSourceFactory::make(m_driver, uriStr);

  if (outputDataSource.get() == nullptr)
    throw te::common::Exception(TE_TR("Output invalid."));

  outputDataSource->open();

  //converter
  std::unique_ptr<te::da::DataSetTypeConverter> converter(new te::da::DataSetTypeConverter(inputDataSetType, outputDataSource->getCapabilities(), outputDataSource->getEncoding()));

  te::da::DataSetType* dsTypeResult = converter->getResult();

  dsTypeResult->setName(dataSetName);

  //exchange
  std::map<std::string, std::string> nopt;

  std::unique_ptr<te::da::DataSet> dataset = dataAccess->getDataSet();

  outputDataSource->createDataSet(dsTypeResult, nopt);

  std::unique_ptr<te::da::DataSetAdapter> dsAdapter(te::da::CreateAdapter(dataset.get(), converter.get()));

  if (dataset->moveBeforeFirst())
    outputDataSource->add(dsTypeResult->getName(), dsAdapter.get(), nopt);

  outputDataSource->close();
};

te::process::DataStoreTaskFactory::DataStoreTaskFactory(): te::process::TaskFactory("DataStoreTask", "Exchanger")
{
}

te::process::DataStoreTaskFactory::~DataStoreTaskFactory() = default;

te::process::Task* te::process::DataStoreTaskFactory::build()
{
  return new DataStoreTask();
}
