/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralib/process/ParameterDescriptor.h

   \brief This class represents the description of a parameter
 */

#ifndef __TERRALIB_PROCESS_INTERNAL_PARAMETERDESCRIPTOR_H
#define __TERRALIB_PROCESS_INTERNAL_PARAMETERDESCRIPTOR_H

 // TerraLib
#include "Config.h"

#include "../datatype/Enums.h"

//STL include files
#include <memory>
#include <string>

namespace te
{
  namespace process
  {
    class ParameterValue;

    /*!
      \class Task

      \brief This class represents the description of a parameter
    */
    class TEPROCESSEXPORT ParameterDescriptor
    {
    public:

      enum class Direction { INPUT, OUTPUT };

      //!< Constructor
      ParameterDescriptor(const std::string& name = "", int dataType = te::dt::UNKNOWN_TYPE, bool isRequired = true, Direction type = Direction::INPUT);

      //!<  Copy constructor
      ParameterDescriptor(const ParameterDescriptor& rhs);

      /*! \brief Virtual destructor. */
      virtual ~ParameterDescriptor();

      //!< Assigment operator
      virtual ParameterDescriptor& operator=(const ParameterDescriptor& rhs);

      //!< Gets the parameter name
      const std::string getName() const;

      //!< Sets the parameter name
      void setName(const std::string& name);

      //!< Gets the parameter description
      const std::string getDescription() const;

      //!< Sets the parameter description
      void setDescription(const std::string& description);

      //!< Gets the parameter data type
      int getDataType() const;

      //!< Sets the parameter data type
      void setDataType(int dataType);

      //!< Checks if the parameter is required
      bool isRequired() const;

      //!< Sets the parameter to be required
      void setRequired(bool required);

      //!< Gets the parameter type
      Direction getDirection() const;

      //!< Sets the parameter type
      void setDirection(Direction direction);

      //!< Gets the default value. If not set, returns null
      const ParameterValue* getDefaultValue() const;

      //!< Sets the default value. This object wille TAKE THE OWNERSHIP  of the pointer
      void setDefaultValue(ParameterValue* defaultValue);

      //!< Checks if the given value is valid basedon the descriptor metadata
      bool isValueValid(const te::process::ParameterValue* value) const;

    protected:

      std::string m_name;
      std::string m_description;
      int m_dataType;
      bool m_isRequired;
      Direction m_direction;
      std::unique_ptr<ParameterValue> m_defaultValue;
    };
  } // end namespace process
}   // end namespace te

#endif  // __TERRALIB_PROCESS_INTERNAL_PARAMETERDESCRIPTOR_H

