/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralib/process/Process.cpp

   \brief This class represents a container of tasks and connectors and is responsible for controlling its execution flow
 */

#include "Process.h"

#include "../common/ThreadGroup.h"

#include "Connector.h"
#include "ProcessListener.h"
#include "Task.h"

//static functions for thread use
void runTaskThread(te::process::TaskSharedPtr task, te::process::ProcessListener* listener, bool* result);

te::process::Process::Process()
  : m_thread(nullptr)
{

}

te::process::Process::~Process()
{
  requestStop();
}

void te::process::Process::clear()
{
  m_vecTasks.clear();
}

bool te::process::Process::isValid(te::process::ProcessListener* listener) const
{
  for (TaskSharedPtr task : m_vecTasks)
  {
    std::string errorMessage;

    if (task->isValid(errorMessage) == false)
    {
      if (listener != nullptr)
      {
        listener->onProcessAborted(errorMessage);
      }

      return false;
    }
  }

  return true;
}

void te::process::Process::addTask(te::process::TaskSharedPtr task)
{
  m_vecTasks.push_back(task);
}

void te::process::Process::addConnector(te::process::ConnectorSharedPtr connectorPtr)
{
  m_vecConnectors.push_back(connectorPtr);
}

te::process::TaskSharedPtr te::process::Process::getTask(const std::string& id) const
{
  for (te::process::TaskSharedPtr task : m_vecTasks)
  {
    if (task->getId() == id)
    {
      return task;
    }
  }

  return te::process::TaskSharedPtr(nullptr);
}

te::process::ConnectorSharedPtr te::process::Process::getConnector(const std::string& id) const
{
  for (te::process::ConnectorSharedPtr connector : m_vecConnectors)
  {
    if (connector->getId() == id)
    {
      return connector;
    }
  }

  return te::process::ConnectorSharedPtr(nullptr);
}

void te::process::Process::removeTask(const std::string& id)
{
  te::process::TaskSharedPtr task = getTask(id);

  std::vector<te::process::Connector*> inputVec = task->getInputConnectors();

  for (te::process::Connector* connector : inputVec)
  {
    removeConnector(connector->getId());
  }

  std::vector<te::process::Connector*> outputVec = task->getOutputConnectors();

  for (te::process::Connector* connector : outputVec)
  {
    removeConnector(connector->getId());
  }

  std::vector<te::process::TaskSharedPtr>::iterator it = m_vecTasks.begin();

  while (it != m_vecTasks.end())
  {
    if (*it == task)
    {
      m_vecTasks.erase(it);
      break;
    }

    ++it;
  }
}

void te::process::Process::removeConnector(const std::string& id)
{
  te::process::ConnectorSharedPtr connector = getConnector(id);

  connector->disassociate();

  std::vector<te::process::ConnectorSharedPtr>::iterator it = m_vecConnectors.begin();

  while (it != m_vecConnectors.end())
  {
    if (*it == connector)
    {
      m_vecConnectors.erase(it);
      break;
    }

    ++it;
  }
}

const std::vector<te::process::TaskSharedPtr>& te::process::Process::getTasks() const
{
  return m_vecTasks;
}

const std::vector<te::process::ConnectorSharedPtr>& te::process::Process::getConnectors() const
{
  return m_vecConnectors;
}

bool te::process::Process::run(te::process::ProcessListener* listener)
{
  if (listener != nullptr)
  {
    listener->onProcessStarted();
  }

  std::size_t taskProcessedCount = 0;
  while (taskProcessedCount != m_vecTasks.size())
  {
    std::vector<te::process::TaskSharedPtr> vecReadyToRunTasks = getReadyToRunTasks();
    if (vecReadyToRunTasks.empty() == true)
    {
      if (listener != nullptr)
      {
        listener->onProcessAborted("No tasks are ready to run");
      }
      return false;
    }

    //we process all taks using all available threads
    bool* result = new bool[vecReadyToRunTasks.size()];
    std::fill_n(result, vecReadyToRunTasks.size(), false);

    te::common::ThreadGroup threadGroup;
    for(std::size_t i = 0; i < vecReadyToRunTasks.size(); ++i)
    {
      TaskSharedPtr task = vecReadyToRunTasks.at(i);
      threadGroup.addJob(std::bind(runTaskThread, task, listener, &result[i]));
    }

    threadGroup.waitToFinish();

    //then we check the results
    for (std::size_t i = 0; i < vecReadyToRunTasks.size(); ++i)
    {
      if (result[i] == false)
      {
        delete[] result;
        return false;
      }
      ++taskProcessedCount;
    }

    delete[] result;
  }

  if (listener != nullptr)
  {
    listener->onProcessFinished();
  }

  return true;
}

void te::process::Process::runAsync(te::process::ProcessListener* listener)
{
  m_thread = new std::thread(&te::process::Process::run, this, listener);
}

void te::process::Process::requestStop()
{
  if (m_thread == nullptr)
  {
    return;
  }

  try
  {
    if (m_thread->joinable() == true)
    {
      m_thread->join();
    }

    delete m_thread;
    m_thread = nullptr;
  }
  catch (...)
  {

  }
}

std::vector<te::process::TaskSharedPtr> te::process::Process::getReadyToRunTasks() const
{
  std::vector<te::process::TaskSharedPtr> vecReadyToRunTasks;
  for (std::size_t i = 0; i < m_vecTasks.size(); ++i)
  {
    te::process::TaskSharedPtr task = m_vecTasks.at(i);
    if (task->getState() == te::process::Task::State::READY_TO_RUN)
    {
      vecReadyToRunTasks.push_back(task);
    }
  }

  return vecReadyToRunTasks;
}

void runTaskThread(te::process::TaskSharedPtr task, te::process::ProcessListener* listener, bool* result)
{
  *result = false;
  if (listener != nullptr)
  {
    listener->onTaskStarted(task);
  }

  try
  {
    task->run();
    *result = true;
  }
  catch (const std::exception & e)
  {
    if (listener != nullptr)
    {
      listener->onProcessAborted(e.what());
    }
    return;
  }

  if (listener != nullptr)
  {
    listener->onTaskFinished(task);
  }
}
