/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralib/process/Process.h

   \brief This class represents a container of tasks and connectors and is responsible for controlling its execution flow
 */

#ifndef __TERRALIB_PROCESS_INTERNAL_PROCESS_H
#define __TERRALIB_PROCESS_INTERNAL_PROCESS_H

 // TerraLib
#include "Config.h"

//STL include files
#include <memory>
#include <thread>
#include <string>
#include <vector>

namespace te
{
  namespace process
  {
    class ConnectorSharedPtr;
    class ProcessListener;
    class TaskSharedPtr;

    /*!
      \class Progress

      \brief This class represents a container of tasks and connectors and is responsible for controlling its execution flow
    */
    class TEPROCESSEXPORT Process
    {
    public:

      /*!
        \brief Constructor
      */
      Process();

      /*! \brief Virtual destructor. */
      virtual ~Process();

      //!< Clears the association
      void clear();

      //!< Checks if the process is valid
      bool isValid(te::process::ProcessListener* listener) const;

      //!< Adds a task to the process. Process will be the holder of the given shared pointers
      void addTask(te::process::TaskSharedPtr taskPtr);

      //!< Adds a connector to the process. Process will be the holder of the given shared pointers
      void addConnector(te::process::ConnectorSharedPtr connectorPtr);

      //!< Gets the task associated to the given id
      te::process::TaskSharedPtr getTask(const std::string& id) const;

      //!< Gets the connector associated to the given id
      te::process::ConnectorSharedPtr getConnector(const std::string& id) const;

      //!< Removed the task associated to the given id
      void removeTask(const std::string& id);

      //!< Removes the connector associated to the given id
      void removeConnector(const std::string& id);
      
      //!< Gets all taks from the process
      const std::vector<te::process::TaskSharedPtr>& getTasks() const;

      //!< Gets all connectors from the process
      const std::vector<te::process::ConnectorSharedPtr>& getConnectors() const;

      //!< Runs the process synchronously
      bool run(te::process::ProcessListener* listener);

      //!< Runs the process asynchronously
      void runAsync(te::process::ProcessListener* listener);

      //!< Requests that the asynchronously execution be stopped
      void requestStop();

    protected:

      //!< Gets all tasks that are ready to rund
      std::vector<te::process::TaskSharedPtr> getReadyToRunTasks() const;

    protected:
      std::thread* m_thread;
      std::vector<te::process::TaskSharedPtr> m_vecTasks;
      std::vector<te::process::ConnectorSharedPtr> m_vecConnectors;
      
    };
  } // end namespace process
}   // end namespace te

#endif  // __TERRALIB_PROCESS_INTERNAL_PROCESS_H
