/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralib/process/Serializer.cpp

   \brief This class serializes the common data structures from Process Module
 */

#include "Serializer.h"

#include "../common/Exception.h"
#include "../core/translator/Translator.h"

#include "Connector.h"
#include "ParameterDescriptor.h"
#include "ParameterValue.h"
#include "Process.h"
#include "Task.h"
#include "TaskCapabilities.h"
#include "TaskParameters.h"

te::process::Serializer::Serializer()
{

}

te::process::Serializer::~Serializer()
{

}

boost::property_tree::ptree te::process::Serializer::encodeParameterDescriptor(const te::process::ParameterDescriptor& descriptor, bool addHeader) const
{
  boost::property_tree::ptree descriptorNode;
  descriptorNode.put("name", descriptor.getName());
  descriptorNode.put("description", descriptor.getDescription());
  descriptorNode.put("dataType", descriptor.getDataType());
  descriptorNode.put("direction", static_cast<int>(descriptor.getDirection()));
  descriptorNode.put("required", descriptor.isRequired());
  descriptorNode.put_child("default_value", encodeParameterValue(descriptor.getDefaultValue(), true));

  if (addHeader == true)
  {
    boost::property_tree::ptree ptree;
    ptree.push_back(std::make_pair("parameter_descriptor", descriptorNode));

    return ptree;
  }

  return descriptorNode;
}

te::process::ParameterDescriptor te::process::Serializer::decodeParameterDescriptor(const boost::property_tree::ptree& ptree) const
{
  te::process::ParameterDescriptor descriptor;

  boost::optional<const boost::property_tree::ptree&> descriptorNode = ptree.get_child_optional("parameter_descriptor");
  if (!descriptorNode)
  {
    descriptorNode = ptree;
  }

  descriptor.setName(descriptorNode->get<std::string>("name"));
  descriptor.setDescription(descriptorNode->get<std::string>("description"));
  descriptor.setDataType(descriptorNode->get<int>("dataType"));
  descriptor.setDirection(static_cast<te::process::ParameterDescriptor::Direction>(descriptorNode->get<int>("direction")));
  descriptor.setRequired(descriptorNode->get<bool>("required"));
  descriptor.setDefaultValue(decodeParameterValue(descriptorNode->get_child("default_value")));

  return descriptor;
}

boost::property_tree::ptree te::process::Serializer::encodeParameterValue(const te::process::ParameterValue* value, bool addHeader) const
{
  boost::property_tree::ptree ptree;
  if (value != nullptr)
  {
    boost::property_tree::ptree valueNode;
    valueNode.add("datatype", value->getDataType());
    valueNode.add("value", value->toString());

    if (addHeader == false)
    {
      return valueNode;
    }

    ptree.push_back(std::make_pair("parameter_value", valueNode));
  }

  return ptree;
}

te::process::ParameterValue* te::process::Serializer::decodeParameterValue(const boost::property_tree::ptree& ptree) const
{
  boost::optional<const boost::property_tree::ptree&> valueNode = ptree.get_child_optional("parameter_value");
  if (!valueNode)
  {
    valueNode = ptree;
  }

  te::process::ParameterValue* parameterValue = nullptr;
  if (valueNode && valueNode->size() > 0)
  {
    int dataType = valueNode->get<int>("datatype");
    std::string strValue = valueNode->get<std::string>("value");

    parameterValue = te::process::CreateParameterValue(dataType);
    parameterValue->fromString(strValue);
  }

  return parameterValue;
}

boost::property_tree::ptree te::process::Serializer::encodeTaskCapabilities(const te::process::TaskCapabilities* capabilities) const
{
  const std::vector<te::process::ParameterDescriptor>& vecDescriptors = capabilities->getParameterDescriptors();

  boost::property_tree::ptree allDescriptorsNode;

  for (const ParameterDescriptor& descriptor : vecDescriptors)
  {
    allDescriptorsNode.add_child("parameter_descriptor", encodeParameterDescriptor(descriptor, false));
  }

  boost::property_tree::ptree descriptorsTree;
  descriptorsTree.add_child("parameter_descriptors", allDescriptorsNode);

  boost::property_tree::ptree ptree;
  ptree.push_back(std::make_pair("task_capabilities", descriptorsTree));

  return ptree;
}

te::process::TaskCapabilities* te::process::Serializer::decodeTaskCapabilities(const boost::property_tree::ptree& ptree) const
{
  te::process::TaskCapabilities* taskCapabilities = new te::process::TaskCapabilities();

  boost::optional<const boost::property_tree::ptree&> taskCapabilitiesNode = ptree.get_child_optional("task_capabilities");
  if (!taskCapabilitiesNode)
  {
    taskCapabilitiesNode = ptree;
  }

  const boost::property_tree::ptree& allDescriptorsNode = taskCapabilitiesNode->get_child("parameter_descriptors");

  boost::property_tree::ptree::const_iterator itDescriptor = allDescriptorsNode.begin();
  while (itDescriptor != allDescriptorsNode.end())
  {
    te::process::ParameterDescriptor descriptor = decodeParameterDescriptor(itDescriptor->second);
    taskCapabilities->addParameterDescriptor(descriptor);

    ++itDescriptor;
  }

  return taskCapabilities;
}

boost::property_tree::ptree te::process::Serializer::encodeTaskParameters(const te::process::TaskParameters* taskParameters, bool addHeader) const
{
  boost::property_tree::ptree allValuesNode;
  for (std::size_t i = 0; i < taskParameters->size(); ++i)
  {
    allValuesNode.add_child("parameter_value", encodeParameterValue(taskParameters->getValue(i), false));
  }

  if (addHeader == false)
  {
    return allValuesNode;
  }

  boost::property_tree::ptree ptree;
  ptree.push_back(std::make_pair("task_parameters", allValuesNode));

  return ptree;
}

te::process::TaskParameters* te::process::Serializer::decodeTaskParameters(const boost::property_tree::ptree& ptree, const te::process::TaskCapabilities* taskCapabilities) const
{
  te::process::TaskParameters* taskParameters = new te::process::TaskParameters(*taskCapabilities);

  boost::optional<const boost::property_tree::ptree&> taskParametersNode = ptree.get_child_optional("task_parameters");
  if (!taskParametersNode)
  {
    taskParametersNode = ptree;
  }

  std::size_t paramIndex = 0;
  boost::property_tree::ptree::const_iterator itDescriptor = taskParametersNode->begin();
  while (itDescriptor != taskParametersNode->end())
  {
    te::process::ParameterValue* value = decodeParameterValue(itDescriptor->second);
    taskParameters->setValue(paramIndex, value, true);

    ++paramIndex;
    ++itDescriptor;
  }

  return taskParameters;
}

boost::property_tree::ptree te::process::Serializer::encodeTask(const te::process::Task* task, bool addHeader) const
{
  boost::property_tree::ptree taskTree;

  taskTree.put("task_type", task->getType());
  taskTree.put("task_id", task->getId());
  taskTree.put_child("task_parameters", encodeTaskParameters(task->getTaskParameters(), false));

  if (addHeader == true)
  {
    boost::property_tree::ptree header;
    header.push_back(std::make_pair("task", taskTree));

    return header;
  }

  return taskTree;
}

te::process::Task* te::process::Serializer::decodeTask(const boost::property_tree::ptree& ptree) const
{
  boost::optional<const boost::property_tree::ptree&> taskNode = ptree.get_child_optional("task");
  if (!taskNode)
  {
    taskNode = ptree;
  }

  std::string typeName = taskNode->get<std::string>("task_type");
  std::string taskId = taskNode->get<std::string>("task_id");

  //decodes the task
  te::process::Task* task = te::process::CreateTask(typeName);

  //decodes its parameters
  te::process::TaskParameters* taskParameters = decodeTaskParameters(*taskNode, task->getCapabilities());

  //sets the parameters to the task
  task->setId(taskId);
  task->setTaskParameters(taskParameters);

  return task;
}

boost::property_tree::ptree te::process::Serializer::encodeConnector(const te::process::Connector* connector, bool addHeader) const
{
  boost::property_tree::ptree item;
  item.put("id", connector->getId());
  item.put("task_from", connector->getTaskFrom()->getId());
  item.put("index_from", connector->getParamIndexFrom());
  item.put("task_to", connector->getTaskTo()->getId());
  item.put("index_to", connector->getParamIndexTo());

  if (addHeader == true)
  {
    boost::property_tree::ptree header;
    header.push_back(std::make_pair("connector", item));

    return header;
  }
  
  return item;
}

te::process::Connector* te::process::Serializer::decodeConnector(const boost::property_tree::ptree& ptree, const std::map<std::string, te::process::Task*>& mapTasks) const
{
  boost::optional<const boost::property_tree::ptree&> connectorNode = ptree.get_child_optional("connector");
  if (!connectorNode)
  {
    connectorNode = ptree;
  }

  std::string connectorId = connectorNode->get<std::string>("id");
  std::string taskFromName = connectorNode->get<std::string>("task_from");
  std::size_t indexFrom = connectorNode->get<std::size_t>("index_from");
  std::string taskToName = connectorNode->get<std::string>("task_to");
  std::size_t indexTo = connectorNode->get<std::size_t>("index_to");

  std::map<std::string, te::process::Task*>::const_iterator itTaskFrom = mapTasks.find(taskFromName);
  if (itTaskFrom == mapTasks.end())
  {
    throw te::common::Exception(TE_TR("te::process::Serializer::decodeConnector - Task not found: ") + taskFromName);
  }
  std::map<std::string, te::process::Task*>::const_iterator itTaskTo = mapTasks.find(taskToName);
  if (itTaskTo == mapTasks.end())
  {
    throw te::common::Exception(TE_TR("te::process::Serializer::decodeConnector - Task not found: ") + taskToName);
  }

  te::process::Task* taskFrom = itTaskFrom->second;
  te::process::Task* taskTo = itTaskTo->second;

  te::process::Connector* connector = new te::process::Connector();
  connector->setId(connectorId);
  connector->associate(taskFrom, indexFrom, taskTo, indexTo);

  return connector;
}

boost::property_tree::ptree te::process::Serializer::encodeProcess(const te::process::Process* process) const
{
  boost::property_tree::ptree tree;

  const std::vector<te::process::TaskSharedPtr>& vecTasks = process->getTasks();
  const std::vector<te::process::ConnectorSharedPtr>& vecConnectors = process->getConnectors();

  boost::property_tree::ptree allTasks;
  for (te::process::TaskSharedPtr task : vecTasks)
  {
    allTasks.add_child("task", encodeTask(task.get(), false));
  }

  boost::property_tree::ptree allConnectors;
  for (te::process::ConnectorSharedPtr connector : vecConnectors)
  {
    allConnectors.add_child("connector", encodeConnector(connector.get(), false));
  }

  tree.push_back(std::make_pair("tasks", allTasks));
  tree.push_back(std::make_pair("connectors", allConnectors));

  boost::property_tree::ptree processTree;
  processTree.push_back(std::make_pair("process", tree));

  return processTree;
}

te::process::Process* te::process::Serializer::decodeProcess(const boost::property_tree::ptree& ptree) const
{
  te::process::Process* process = new te::process::Process();

  boost::optional<const boost::property_tree::ptree&> processNode = ptree.get_child_optional("process");
  if (!processNode)
  {
    processNode = ptree;
  }

  //decode the tasks
  boost::optional<const boost::property_tree::ptree&> tasksNode = processNode->get_child("tasks");

  std::map<std::string, te::process::Task*> mapTasks;

  boost::property_tree::ptree::const_iterator itTask = tasksNode->begin();
  while (itTask != tasksNode->end())
  {
    te::process::TaskSharedPtr task(decodeTask(itTask->second));
    process->addTask(task);

    mapTasks[task->getId()] = task.get();

    ++itTask;
  }

  //decode the connectors
  boost::optional<const boost::property_tree::ptree&> connectorsNode = processNode->get_child_optional("connectors");
  if (connectorsNode)
  {
    boost::property_tree::ptree::const_iterator itConnector = connectorsNode->begin();
    while (itConnector != connectorsNode->end())
    {
      te::process::ConnectorSharedPtr connector(decodeConnector(itConnector->second, mapTasks));
      process->addConnector(connector);

      ++itConnector;
    }
  }

  return process;
}
