/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralib/process/Serializer.h

   \brief This class serializes the common data structures from Process Module
 */

#ifndef __TERRALIB_PROCESS_INTERNAL_SERIALIZER_H
#define __TERRALIB_PROCESS_INTERNAL_SERIALIZER_H

 // TerraLib
#include "Config.h"

// Boost
#include <boost/property_tree/ptree.hpp>

//STL
#include <map>
#include <string>

namespace te
{
  namespace process
  {
    class Connector;
    class ParameterDescriptor;
    class ParameterValue;
    class Process;
    class Task;
    class TaskCapabilities;
    class TaskParameters;

    /*!
      \class Serializer

      \brief This class serializes the common data structures from Process Module
    */
    class TEPROCESSEXPORT Serializer
    {
    public:

      /*!
        \brief Constructor
      */
      Serializer();

      /*! \brief Virtual destructor. */
      virtual ~Serializer();

      //!< Encodes the 'ParameterDescriptor' into a 'boost::property_tree::ptree' object
      boost::property_tree::ptree encodeParameterDescriptor(const te::process::ParameterDescriptor& descriptor, bool addHeader) const;

      //!< Encodes a 'boost::property_tree::ptree' into a 'ParameterDescriptor' object
      te::process::ParameterDescriptor decodeParameterDescriptor(const boost::property_tree::ptree& ptree) const;

      //!< Encodes the 'ParameterValue' into a 'boost::property_tree::ptree' object
      boost::property_tree::ptree encodeParameterValue(const te::process::ParameterValue* descriptor, bool addHeader) const;

      //!< Encodes a 'boost::property_tree::ptree' into a 'ParameterValue' object
      te::process::ParameterValue* decodeParameterValue(const boost::property_tree::ptree& ptree) const;

      //!< Encodes the 'TaskCapabilities' into a 'boost::property_tree::ptree' object
      boost::property_tree::ptree encodeTaskCapabilities(const te::process::TaskCapabilities* capabilities) const;

      //!< Encodes a 'boost::property_tree::ptree' into a 'TaskCapabilities' object
      te::process::TaskCapabilities* decodeTaskCapabilities(const boost::property_tree::ptree& ptree) const;

      //!< Encodes the 'TaskParameters' into a 'boost::property_tree::ptree' object
      boost::property_tree::ptree encodeTaskParameters(const te::process::TaskParameters* taskParameters, bool addHeader) const;

      //!< Encodes a 'boost::property_tree::ptree' into a 'TaskParameters' object
      te::process::TaskParameters* decodeTaskParameters(const boost::property_tree::ptree& ptree, const te::process::TaskCapabilities* capabilities) const;

      //!< Encodes the 'Task' into a 'boost::property_tree::ptree' object
      boost::property_tree::ptree encodeTask(const te::process::Task* task, bool addHeader) const;

      //!< Encodes a 'boost::property_tree::ptree' into a 'Task' object
      te::process::Task* decodeTask(const boost::property_tree::ptree& ptree) const;

      //!< Encodes the 'Connector' into a 'boost::property_tree::ptree' object
      boost::property_tree::ptree encodeConnector(const te::process::Connector* connector, bool addHeader) const;

      //!< Encodes a 'boost::property_tree::ptree' into a 'Connector' object
      te::process::Connector* decodeConnector(const boost::property_tree::ptree& ptree, const std::map<std::string, te::process::Task*>& mapTasks) const;

      //!< Encodes the 'Process' into a 'boost::property_tree::ptree' object
      boost::property_tree::ptree encodeProcess(const te::process::Process* process) const;

      //!< Encodes a 'boost::property_tree::ptree' into a 'Process' object
      te::process::Process* decodeProcess(const boost::property_tree::ptree& ptree) const;

    };

  } // end namespace process
}   // end namespace te

#endif  // __TERRALIB_PROCESS_INTERNAL_SERIALIZER_H
