/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralib/process/SimpleDataParameterValue.cpp

   \brief This class represents a simple data parameter value
 */

#include "SimpleDataParameterValue.h"

#include "../common/Exception.h"
#include "../core/translator/Translator.h"
#include "../datatype/AbstractData.h"
#include "../datatype/DataConverterManager.h"
#include "../datatype/Enums.h"
#include "../datatype/SimpleData.h"

namespace
{
  static te::process::SimpleDataParameterValueFactory SimpleDataParameterValueFactoryInstanceInt16(te::dt::INT16_TYPE);
  static te::process::SimpleDataParameterValueFactory SimpleDataParameterValueFactoryInstanceInt32(te::dt::INT32_TYPE);
  static te::process::SimpleDataParameterValueFactory SimpleDataParameterValueFactoryInstanceString(te::dt::STRING_TYPE);
  static te::process::SimpleDataParameterValueFactory SimpleDataParameterValueFactoryInstanceFloat(te::dt::FLOAT_TYPE);
  static te::process::SimpleDataParameterValueFactory SimpleDataParameterValueFactoryInstanceDouble(te::dt::DOUBLE_TYPE);
  static te::process::SimpleDataParameterValueFactory SimpleDataParameterValueFactoryInstanceBoolean(te::dt::BOOLEAN_TYPE);
}

te::process::SimpleDataParameterValue::SimpleDataParameterValue(int dataType)
  : te::process::ParameterValue(dataType)
{

}

te::process::SimpleDataParameterValue::SimpleDataParameterValue(te::dt::AbstractData* absData)
  : te::process::ParameterValue(absData->getTypeCode())
  , m_absData(absData)
{

}

te::process::SimpleDataParameterValue::~SimpleDataParameterValue() = default;

te::process::ParameterValue* te::process::SimpleDataParameterValue::clone() const
{
  SimpleDataParameterValue* newValue = new SimpleDataParameterValue(this->getDataType());
  if (this->m_absData != nullptr)
  {
    newValue->setValue(this->m_absData->clone());
  }

  return newValue;
}

const te::dt::AbstractData* te::process::SimpleDataParameterValue::getValue() const
{
  return m_absData.get();
}

void te::process::SimpleDataParameterValue::setValue(te::dt::AbstractData* absData)
{
  m_absData.reset(absData);
}

std::string te::process::SimpleDataParameterValue::toString() const
{
  try
  {
    const te::dt::DataTypeConverter& converter = te::dt::DataConverterManager::getInstance().get(this->getDataType(), te::dt::STRING_TYPE);
    te::dt::AbstractData* convertedData = converter(m_absData.get());
    if (convertedData == nullptr)
    {
      throw te::common::Exception(std::string("te::process::SimpleDataParameterValue::toString()") + TE_TR("Invalid Convertion"));
    }

    te::dt::String* strData = dynamic_cast<te::dt::String*>(convertedData);
    if (strData == nullptr)
    {
      throw te::common::Exception(std::string("te::process::SimpleDataParameterValue::toString()") + TE_TR("Invalid Convertion"));
    }
    
    return strData->getValue();
  }
  catch (const std::exception&)
  {
    throw te::common::Exception(std::string("te::process::SimpleDataParameterValue::toString()") + TE_TR("Invalid Convertion"));
  }
}

void te::process::SimpleDataParameterValue::fromString(const std::string& strValue)
{
  try
  {
    std::unique_ptr<te::dt::AbstractData> stringData(new te::dt::String(strValue));

    const te::dt::DataTypeConverter& converter = te::dt::DataConverterManager::getInstance().get(te::dt::STRING_TYPE, this->getDataType());
    te::dt::AbstractData* convertedData = converter(stringData.get());
    if (convertedData == nullptr)
    {
      throw te::common::Exception(std::string("te::process::SimpleDataParameterValue::fromString()") + TE_TR("Invalid Convertion"));
    }

    m_absData.reset(convertedData);
  }
  catch (const std::exception&)
  {
    throw te::common::Exception(std::string("te::process::SimpleDataParameterValue::fromString()") + TE_TR("Invalid Convertion"));
  }
}

te::process::SimpleDataParameterValueFactory::SimpleDataParameterValueFactory(int dataType)
  : te::process::ParameterValueFactory(dataType)
{

}


te::process::SimpleDataParameterValueFactory::~SimpleDataParameterValueFactory()
{

}

te::process::ParameterValue* te::process::SimpleDataParameterValueFactory::build()
{
  te::process::SimpleDataParameterValue* newValue = new te::process::SimpleDataParameterValue(this->getKey());
  return newValue;
}
