/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralib/process/SimpleDataParameterValue.h

   \brief This class represents a simple data parameter value
 */

#ifndef __TERRALIB_PROCESS_INTERNAL_SIMPLEDATAPARAMETERVALUE_H
#define __TERRALIB_PROCESS_INTERNAL_SIMPLEDATAPARAMETERVALUE_H

 // TerraLib
#include "Config.h"

#include "ParameterValue.h"

//STL include files
#include <memory>
#include <string>

namespace te
{
  namespace dt
  {
    class AbstractData;
  }

  namespace process
  {
    /*!
      \class Task

      \brief This class represents the value of a parameter.
    */
    class TEPROCESSEXPORT SimpleDataParameterValue : public te::process::ParameterValue
    {
    public:

      //!< Constructor
      explicit SimpleDataParameterValue(int dataType);

      //!< Constructor
      explicit SimpleDataParameterValue(te::dt::AbstractData* absData);

      /*! \brief Virtual destructor. */
      virtual ~SimpleDataParameterValue();

      //!< Clones the object
      virtual ParameterValue* clone() const override;

      //!< Returns the string representation of the parameter value
      virtual std::string toString() const override;

      //!< Decodes the given string representation and sets the value into the parameter value object
      virtual void fromString(const std::string& strValue) override;

      //!< Gets the value associated to this parameter
      virtual const te::dt::AbstractData* getValue() const;

      //!< Gets the value associated to this parameter, casting it to the requested type
      template<typename T>
      const T* getValueAs() const;

      //!< Sets the value to be associated to this parameter. This object will take the ownership of the given pointer
      virtual void setValue(te::dt::AbstractData* absData);

    protected:

      std::unique_ptr<te::dt::AbstractData> m_absData; //!< The data of the parameter value
    };

    class SimpleDataParameterValueFactory : public te::process::ParameterValueFactory
    {
    public:

      //!< Constructor
      SimpleDataParameterValueFactory(int dataType);

      //!< Destructor
      virtual ~SimpleDataParameterValueFactory();

    protected:

      te::process::ParameterValue* build() override;
    };


    //implementation
    template<typename T>
    const T* SimpleDataParameterValue::getValueAs() const
    {
      const te::dt::AbstractData* absData = getValue();
      if (absData == nullptr)
      {
        return nullptr;
      }

      return dynamic_cast<const T*>(absData);
    }

  } // end namespace process
}   // end namespace te

#endif  // __TERRALIB_PROCESS_INTERNAL_SIMPLEDATAPARAMETERVALUE_H

