/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralib/process/Task.cpp

   \brief Abstraction that represents a Task that can be executed
 */

#include "Task.h"

#include "../common/Exception.h"
#include "../common/StringUtils.h"
#include "../core/translator/Translator.h"

#include "Connector.h"
#include "ParameterValue.h"
#include "TaskCapabilities.h"
#include "TaskParameters.h"

te::process::Task::Task(const std::string& taskType)
  : m_taskType(taskType)
  , m_id(te::common::GetUniqueID())
  , m_state(State::NOT_INITIALIZED)
{
  
}

te::process::Task::~Task() = default;

const std::string& te::process::Task::getId() const
{
  return m_id;
}

void te::process::Task::setId(const std::string& id)
{
  m_id = id;
}

const std::string& te::process::Task::getType() const
{
  return m_taskType;
}

const te::process::TaskCapabilities* te::process::Task::getCapabilities() const
{
  return m_capabilities.get();
}

te::process::Task::State te::process::Task::getState() const
{
  return m_state;
}

const te::process::ParameterValue* te::process::Task::getParameter(std::size_t index) const
{
  if (m_state == State::NOT_INITIALIZED)
  {
    throw te::common::Exception(TE_TR("Task not initialized"));
  }

  return m_parameters->getValue(index);
}

void te::process::Task::setParameter(std::size_t index, te::process::ParameterValue* value, bool takeOwnership)
{
  if (m_state == State::NOT_INITIALIZED)
  {
    throw te::common::Exception(TE_TR("Task not initialized"));
  }

  m_parameters->setValue(index, value, takeOwnership);

  const ParameterDescriptor& descriptor = m_capabilities->getParameterDescriptor(index);
  if (descriptor.getDirection() == ParameterDescriptor::Direction::INPUT)
  {
    checkForReadyToRun();
  }
}

const te::process::TaskParameters* te::process::Task::getTaskParameters() const
{
  return m_parameters.get();
}

void te::process::Task::setTaskParameters(TaskParameters* taskParameters)
{
  m_parameters.reset(taskParameters);
  checkForReadyToRun();
}

void te::process::Task::addConnector(te::process::Connector* connector)
{
  m_connectors.push_back(connector);
}

void te::process::Task::removeConnector(Connector* connector)
{
  std::vector<te::process::Connector*>::iterator it = m_connectors.begin();

  while (it != m_connectors.end())
  {
    if (*it == connector)
    {
      m_connectors.erase(it);
      break;
    }

    ++it;
  }
}

std::vector<te::process::Connector*> te::process::Task::getInputConnectors() const
{
  std::vector<te::process::Connector*> vecConnectors;
  for (te::process::Connector* connector : m_connectors)
  {
    if (connector->getTaskTo() == this)
    {
      vecConnectors.push_back(connector);
    }
  }

  return vecConnectors;
}

std::vector<te::process::Connector*> te::process::Task::getOutputConnectors() const
{
  std::vector<te::process::Connector*> vecConnectors;
  for (te::process::Connector* connector : m_connectors)
  {
    if (connector->getTaskFrom() == this)
    {
      vecConnectors.push_back(connector);
    }
  }

  return vecConnectors;
}

bool te::process::Task::isValid(std::string& errorMessage) const
{
  std::vector<Connector*> vecInputConectors = getInputConnectors();

  //we must check if all the required input parameters have values OR connectors associated to them
  for (std::size_t paramIndex = 0; paramIndex < m_capabilities->size(); ++paramIndex)
  {
    const te::process::ParameterDescriptor param = m_capabilities->getParameterDescriptor(paramIndex);

    if (param.isRequired() == false)
    {
      continue;
    }

    //if a value is required, we must check if it already has a value
    if (m_parameters->getValue(paramIndex) != nullptr)
    {
      continue;
    }

    //if the parameter value is null, we must check if a conector will push this value in the future
    bool found = false;
    for (Connector* connector : vecInputConectors)
    {
      if (paramIndex == connector->getParamIndexTo())
      {
        found = true;
        break;
      }
    }

    if (found == false)
    {
      errorMessage = "Parameter: " + param.getName() + " from Task: " + this->getId() + " not defined.";
      return false;
    }
  }

  return true;
}

void te::process::Task::run()
{
  checkForReadyToRun();

  if (m_state != State::READY_TO_RUN)
  {
    throw te::common::Exception(TE_TR("Task not ready to run"));
  }

  m_state = State::RUNNING;

  try
  {
    runImpl();

    m_state = State::FINISHED;

    pushParametersToConnectors();
  }
  catch (const std::exception&)
  {
    m_state = State::FINISHED_WITH_ERRORS;
  }
  catch (...)
  {
    m_state = State::FINISHED_WITH_ERRORS;
  }
}

void te::process::Task::init(TaskCapabilities* capabilities)
{
  m_capabilities.reset(capabilities);
  if (capabilities == nullptr)
  {
    return;
  }

  m_parameters.reset(new te::process::TaskParameters(*capabilities));

  checkForReadyToRun();
}

bool te::process::Task::checkConnectorsAreReady()
{
  std::vector<te::process::Connector*> vecConnectors = getInputConnectors();
  for (te::process::Connector* connector : vecConnectors)
  {
    if (connector->isReady() == false)
    {
      return false;
    }
  }

  return true;
}

void te::process::Task::pushParametersToConnectors()
{
  std::vector<te::process::Connector*> vecConnectors = getOutputConnectors();
  for (te::process::Connector* connector : vecConnectors)
  {
    connector->apply();
  }
}

void te::process::Task::checkForReadyToRun()
{
  //we first fetch all parameters that comes from input connector (if appliable)
  if (checkConnectorsAreReady() == false)
  {
    m_state = State::MISSING_PARAMETERS;
    return;
  }

  bool readyToRun = m_parameters->isValid();
  if (readyToRun == true)
  {
    m_state = State::READY_TO_RUN;
  }
  else
  {
    m_state = State::MISSING_PARAMETERS;
  }
}

te::process::TaskPtr::TaskPtr(te::process::Task* task)
  : std::unique_ptr<Task>(task)
{

}

te::process::TaskPtr::~TaskPtr() = default;

te::process::TaskSharedPtr::TaskSharedPtr(te::process::Task* task)
  : std::shared_ptr<Task>(task)
{

}

te::process::TaskSharedPtr::~TaskSharedPtr() = default;



te::process::TaskFactory::TaskFactory(const std::string& taskType, const std::string& category)
  : te::common::AbstractFactory<te::process::Task, std::string>(taskType)
{
  m_mapCategories[category].push_back(taskType);
}

te::process::TaskFactory::~TaskFactory() = default;

std::vector<std::string> te::process::TaskFactory::getCategories()
{
  std::vector<std::string> vecCategories;
  vecCategories.reserve(m_mapCategories.size());

  std::map<std::string, std::vector<std::string> >::iterator it = m_mapCategories.begin();
  while (it != m_mapCategories.end())
  {
    vecCategories.push_back(it->first);
    ++it;
  }

  return vecCategories;
}

std::vector<std::string> te::process::TaskFactory::getTasksFromCategory(const std::string& category)
{
  std::vector<std::string> vecTasks;

  std::map<std::string, std::vector<std::string> >::iterator it = m_mapCategories.find(category);
  if (it != m_mapCategories.end())
  {
    vecTasks = it->second;
  }

  return vecTasks;
}


te::process::Task* te::process::CreateTask(const std::string& taskType)
{
  return TaskFactory::make(taskType);
}
