/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralib/process/TaskParameters.cpp

   \brief This class is a holder for all the task input and output parameters
 */

#include "TaskParameters.h"

#include "../common/Exception.h"
#include "../core/translator/Translator.h"
#include "../datatype/Enums.h"

#include "ParameterValue.h"
#include "TaskCapabilities.h"


te::process::TaskParameters::TaskParameters(const te::process::TaskCapabilities& taskCapabilities)
  : m_capabilities(taskCapabilities)
{
  const std::vector<te::process::ParameterDescriptor>& vecDescriptors = taskCapabilities.getParameterDescriptors();

  for (std::size_t i = 0; i < vecDescriptors.size(); ++i)
  {
    const te::process::ParameterDescriptor& descriptor = vecDescriptors.at(i);

    te::process::ParameterValue* value = nullptr;
    bool isOwner = false;

    if (descriptor.getDefaultValue() != nullptr)
    {
      value = descriptor.getDefaultValue()->clone();
      isOwner = true;
    }

    m_vecValues.push_back(value);
    m_vecOwnership.push_back(isOwner);
  }
}

te::process::TaskParameters::TaskParameters(const te::process::TaskParameters& rhs)
{
  *this = rhs;
}

te::process::TaskParameters::~TaskParameters()
{
  clear();
}

te::process::TaskParameters& te::process::TaskParameters::operator=(const TaskParameters& rhs)
{
  clear();

  for (std::size_t i = 0; i < rhs.m_vecValues.size(); ++i)
  {
    bool isOwner = rhs.m_vecOwnership.at(i);

    te::process::ParameterValue* value = nullptr;
    if (isOwner == true)
    {
      if (rhs.m_vecValues.at(i) != nullptr)
      {
        value = rhs.m_vecValues.at(i)->clone();
      }
    }
    else
    {
      value = rhs.m_vecValues.at(i);
    }

    m_vecValues.push_back(value);
    m_vecOwnership.push_back(isOwner);
  }

  return *this;
}

std::size_t te::process::TaskParameters::size() const
{
  return m_vecValues.size();
}

bool te::process::TaskParameters::isValid() const
{
  if (m_capabilities.size() != m_vecValues.size())
  {
    return false;
  }

  for (std::size_t i = 0; i < m_capabilities.size(); ++i)
  {
    const te::process::ParameterDescriptor& descriptor = m_capabilities.getParameterDescriptor(i);
    if (descriptor.isValueValid(m_vecValues.at(i)) == false)
    {
      return false;
    }
  }

  return true;
}

const te::process::ParameterValue* te::process::TaskParameters::getValue(std::size_t index) const
{
  if (index >= m_vecValues.size())
  {
    throw te::common::Exception(TE_TR("Invalid value index in") + std::string(": te::process::TaskParameters::getValue"));
  }

  return m_vecValues.at(index);
}

void te::process::TaskParameters::setValue(std::size_t index, te::process::ParameterValue* value, bool takeOwnership)
{
  if (index >= m_vecValues.size())
  {
    throw te::common::Exception(TE_TR("Invalid value index in") + std::string(": te::process::TaskParameters::setValue"));
  }

  delete m_vecValues.at(index);
  
  m_vecValues.at(index) = value;
  m_vecOwnership.at(index) = takeOwnership;
}

void te::process::TaskParameters::clear()
{
  for (std::size_t i = 0; i < m_vecValues.size(); ++i)
  {
    if (m_vecOwnership.at(i) == false)
    {
      continue;
    }

    delete m_vecValues.at(i);
  }

  m_vecValues.clear();
  m_vecOwnership.clear();
}
