#include "ProcessBuilder.h"

#include "ui/ui_ProcessBuilderForm.h"

#include "Config.h"
#include "ProcessManager.h"
#include "ProcessRunner.h"
#include "explorer/DataAccessProcessTreeGroup.h"
#include "explorer/DataStoreProcessTreeGroup.h"
#include "explorer/ProcessExplorerTree.h"
#include "graphics/ProcessGraphicsScene.h"
#include "graphics/ProcessGraphicsView.h"
#include "graphics/TaskGraphicsItem.h"
#include "property/ConnectorsProperty.h"
#include "property/ProcessInspector.h"
#include "property/TaskProperty.h"

#include "../Connector.h"
#include "../Process.h"
#include "../Task.h"

// TerraLib
#include "../../qt/widgets/Utils.h"

// Qt
#include <QAction>
#include <QMenu>
#include <QMenuBar>
#include <QMessageBox>

te::process::qt::ProcessBuilder::ProcessBuilder(QWidget* parent) :
  QMainWindow(parent)
{
  RegisterFactories();

  setupUI();
}

te::process::qt::ProcessBuilder::~ProcessBuilder()
{
  delete m_ui;
}

te::process::qt::ProcessExplorerTree* te::process::qt::ProcessBuilder::getExplorerTree()
{
  return m_explorerTree;
}

void te::process::qt::ProcessBuilder::setupUI()
{
  m_ui = new Ui::ProcessBuilderForm;
  m_ui->setupUi(this);

  //define dock corner position
  setCorner(Qt::BottomLeftCorner, Qt::LeftDockWidgetArea);
  setCorner(Qt::BottomRightCorner, Qt::RightDockWidgetArea);

  //manager
  m_manager = new ProcessManager(this);

  //runner
  m_processRunner = new ProcessRunner(this);

  //explorer dock
  m_explorerTree = new te::process::qt::ProcessExplorerTree(this);
  createDock(m_explorerTree, m_explorerDock, Qt::LeftDockWidgetArea, "Tasks Explorer");

  DataAccessProcessTreeGroup* dataAccessTreeGroup = new DataAccessProcessTreeGroup(m_explorerTree);
  DataStoreProcessTreeGroup* dataStoreTreeGroup = new DataStoreProcessTreeGroup(m_explorerTree);

  //process inspector
  m_processInspector = new te::process::qt::ProcessInspector(m_manager, this);
  createDock(m_processInspector, m_inspectorDock, Qt::RightDockWidgetArea, "Process Inspector");

  //task property dock
  m_taskProperty = new te::process::qt::TaskProperty(m_manager, this);
  createDock(m_taskProperty, m_taskPropertyDock, Qt::RightDockWidgetArea, "Task Property");

  //connector property dock
  m_connectorsProperty = new te::process::qt::ConnectorsProperty(m_manager, this);
  createDock(m_connectorsProperty, m_connectorPropertyDock, Qt::RightDockWidgetArea, "Task Connectors");

  //processlog dock
  createDock(m_processRunner->getLoggerWidget(), m_processRunnerDock, Qt::BottomDockWidgetArea, "Process Runner Log");

  //graphics view
  m_scene = new ProcessGraphicsScene(m_manager, this);

  ProcessGraphicsView* processGraphicsView = new ProcessGraphicsView(m_scene, this);
  processGraphicsView->scale(2, 2);
  this->setCentralWidget(processGraphicsView);

  //connects
  connect(m_scene, SIGNAL(selectionChanged()), this, SLOT(sceneitemSelectionChanged()));

  //create menus
  setupMenus();
}

void te::process::qt::ProcessBuilder::setupMenus()
{
  //application menu bar
  QMenuBar* menuBar = new QMenuBar(this);
  setMenuBar(menuBar);

  //file menu
  QMenu* fileMenu = new QMenu(tr("File"), menuBar);
  menuBar->addMenu(fileMenu);

  QAction* fileExit = new QAction(tr("Exit"), menuBar);
  fileExit->setIcon(QIcon::fromTheme("system-log-out"));

  fileMenu->addAction(fileExit);

  connect(fileExit, SIGNAL(triggered()), SLOT(close()));

  //menu project
  QMenu* projMenu = new QMenu(tr("Project"), menuBar);
  menuBar->addMenu(projMenu);

  QAction* projRun = new QAction(tr("Run"), menuBar);
  QAction* projValidate = new QAction(tr("Validate"), menuBar);
  QAction* projReset = new QAction(tr("Reset"), menuBar);

  projMenu->addAction(projRun);
  projMenu->addAction(projValidate);
  projMenu->addAction(projReset);

  connect(projRun, SIGNAL(triggered()), this, SLOT(onProjectRunTriggered()));
  connect(projValidate, SIGNAL(triggered()), this, SLOT(onProjectValidateTriggered()));
  connect(projReset, SIGNAL(triggered()), this, SLOT(onProjectResetTriggered()));

  //menu format
  QMenu* formatMenu = new QMenu(tr("Format"), menuBar);
  menuBar->addMenu(formatMenu);

  QMenu* alignMenu = new QMenu(tr("Alignment"), menuBar);
  formatMenu->addMenu(alignMenu);

  QAction* leftAlignAction = new QAction(tr("Left Alignment"), alignMenu);
  QAction* rightAlignAction = new QAction(tr("Right Alignment"), alignMenu);
  QAction* topAlignAction = new QAction(tr("Top Alignment"), alignMenu);
  QAction* bottomtAlignAction = new QAction(tr("Bottom Alignment"), alignMenu);
  QAction* hCenterAlignAction = new QAction(tr("Horizontal Center Alignment"), alignMenu);
  QAction* vCenterAlignAction = new QAction(tr("Vertical Center Alignment"), alignMenu);

  leftAlignAction->setIcon(QIcon::fromTheme("layout-alignleft"));
  rightAlignAction->setIcon(QIcon::fromTheme("layout-alignright"));
  topAlignAction->setIcon(QIcon::fromTheme("layout-aligntop"));
  bottomtAlignAction->setIcon(QIcon::fromTheme("layout-alignbottom"));
  hCenterAlignAction->setIcon(QIcon::fromTheme("layout-alignvrtcenter"));
  vCenterAlignAction->setIcon(QIcon::fromTheme("layout-alignhrzcenter"));

  alignMenu->addAction(leftAlignAction);
  alignMenu->addAction(rightAlignAction);
  alignMenu->addAction(topAlignAction);
  alignMenu->addAction(bottomtAlignAction);
  alignMenu->addAction(hCenterAlignAction);
  alignMenu->addAction(vCenterAlignAction);

  connect(leftAlignAction, SIGNAL(triggered()), this, SLOT(onFormatLeftAlignTriggered()));
  connect(rightAlignAction, SIGNAL(triggered()), this, SLOT(onFormatRightAlignTriggered()));
  connect(topAlignAction, SIGNAL(triggered()), this, SLOT(onFormatTopAlignTriggered()));
  connect(bottomtAlignAction, SIGNAL(triggered()), this, SLOT(onFormatBottomAlignTriggered()));
  connect(hCenterAlignAction, SIGNAL(triggered()), this, SLOT(onFormatHCenterAlignTriggered()));
  connect(vCenterAlignAction, SIGNAL(triggered()), this, SLOT(onFormatVCenterAlignTriggered()));

  //menu window
  QMenu* windowMenu = new QMenu(tr("Window"), menuBar);
  menuBar->addMenu(windowMenu);

  m_windowExplorerDock = new QAction(tr("Tasks Explorer"), menuBar);
  m_windowExplorerDock->setIcon(QIcon::fromTheme("view-layer-explorer"));
  m_windowExplorerDock->setCheckable(true);
  m_windowExplorerDock->setChecked(true);

  m_windowInspectorDock = new QAction(tr("Process Inspector"), menuBar);
  m_windowInspectorDock->setIcon(QIcon::fromTheme("zoom"));
  m_windowInspectorDock->setCheckable(true);
  m_windowInspectorDock->setChecked(true);

  m_windowTaskDock = new QAction(tr("Task Property"), menuBar);
  m_windowTaskDock->setIcon(QIcon::fromTheme("view-data-table"));
  m_windowTaskDock->setCheckable(true);
  m_windowTaskDock->setChecked(true);

  m_windowConnectorDock = new QAction(tr("Task Connectors"), menuBar);
  m_windowConnectorDock->setIcon(QIcon::fromTheme("connector_direction"));
  m_windowConnectorDock->setCheckable(true);
  m_windowConnectorDock->setChecked(true);

  m_windowRunnerDock = new QAction(tr("Process Runner"), menuBar);
  m_windowRunnerDock->setIcon(QIcon::fromTheme("terminal"));
  m_windowRunnerDock->setCheckable(true);
  m_windowRunnerDock->setChecked(true);

  windowMenu->addAction(m_windowExplorerDock);
  windowMenu->addAction(m_windowInspectorDock);
  windowMenu->addAction(m_windowTaskDock);
  windowMenu->addAction(m_windowConnectorDock);
  windowMenu->addAction(m_windowRunnerDock);

  connect(m_windowExplorerDock, SIGNAL(toggled(bool)), m_explorerDock, SLOT(setVisible(bool)));
  connect(m_explorerDock, SIGNAL(visibilityChanged(bool)), this, SLOT(onWindowExplorerDockVisibilityChanged(bool)));

  connect(m_windowInspectorDock, SIGNAL(toggled(bool)), m_inspectorDock, SLOT(setVisible(bool)));
  connect(m_inspectorDock, SIGNAL(visibilityChanged(bool)), this, SLOT(onWindowInspectorDockVisibilityChanged(bool)));

  connect(m_windowTaskDock, SIGNAL(toggled(bool)), m_taskPropertyDock, SLOT(setVisible(bool)));
  connect(m_taskPropertyDock, SIGNAL(visibilityChanged(bool)), this, SLOT(onWindowTaskDockVisibilityChanged(bool)));

  connect(m_windowConnectorDock, SIGNAL(toggled(bool)), m_connectorPropertyDock, SLOT(setVisible(bool)));
  connect(m_connectorPropertyDock, SIGNAL(visibilityChanged(bool)), this, SLOT(onWindowConnectorDockVisibilityChanged(bool)));

  connect(m_windowRunnerDock, SIGNAL(toggled(bool)), m_processRunnerDock, SLOT(setVisible(bool)));
  connect(m_processRunnerDock, SIGNAL(visibilityChanged(bool)), this, SLOT(onWindowRunnerDockVisibilityChanged(bool)));

  //menu help
  QMenu* helpMenu = new QMenu(tr("Help"), menuBar);
  menuBar->addMenu(helpMenu);

  QAction* helpAction = new QAction(tr("About"), menuBar);
  helpAction->setIcon(QIcon::fromTheme("help-about-browser"));

  helpMenu->addAction(helpAction);

  connect(helpAction, SIGNAL(triggered()), this, SLOT(onHelpAboutTriggered()));
}

void te::process::qt::ProcessBuilder::createDock(QWidget* widget, QDockWidget*& dockWidget, Qt::DockWidgetArea area, const std::string& dockTitle)
{
  dockWidget = new QDockWidget(this);
  dockWidget->setWindowTitle(te::qt::widgets::Convert2QString(dockTitle));
  dockWidget->setWidget(widget);

  addDockWidget(area, dockWidget);
}

void te::process::qt::ProcessBuilder::sceneitemSelectionChanged()
{
  QList<QGraphicsItem*> graphicItems =  m_scene->selectedItems();

  if (graphicItems.size() == 1)
  {
    QGraphicsItem* item = graphicItems.at(0);

    TaskGraphicsItem* taskItem = dynamic_cast<TaskGraphicsItem*>(item);

    if (taskItem)
    {
      m_taskProperty->setTask(taskItem->getTaskId());

      m_taskProperty->setTaskItem(taskItem);

      m_connectorsProperty->setTask(taskItem->getTaskId());
    }
  }
}

void te::process::qt::ProcessBuilder::onProjectRunTriggered()
{
  m_processRunner->run(m_manager->getProcess());
}

void te::process::qt::ProcessBuilder::onProjectValidateTriggered()
{
  m_processRunner->validate(m_manager->getProcess());
}

void te::process::qt::ProcessBuilder::onProjectResetTriggered()
{
  m_processRunner->reset(m_manager->getProcess());
}

void te::process::qt::ProcessBuilder::onFormatLeftAlignTriggered()
{
  m_scene->alignSelecteds(Qt::AlignLeft);
}

void te::process::qt::ProcessBuilder::onFormatRightAlignTriggered()
{
  m_scene->alignSelecteds(Qt::AlignRight);
}

void te::process::qt::ProcessBuilder::onFormatTopAlignTriggered()
{
  m_scene->alignSelecteds(Qt::AlignTop);
}

void te::process::qt::ProcessBuilder::onFormatBottomAlignTriggered()
{
  m_scene->alignSelecteds(Qt::AlignBottom);
}

void te::process::qt::ProcessBuilder::onFormatHCenterAlignTriggered()
{
  m_scene->alignSelecteds(Qt::AlignHCenter);
}

void te::process::qt::ProcessBuilder::onFormatVCenterAlignTriggered()
{
  m_scene->alignSelecteds(Qt::AlignVCenter);
}

void te::process::qt::ProcessBuilder::onWindowExplorerDockVisibilityChanged(bool state)
{
  m_windowExplorerDock->blockSignals(true);
  m_windowExplorerDock->setChecked(state);
  m_windowExplorerDock->blockSignals(false);
}

void te::process::qt::ProcessBuilder::onWindowInspectorDockVisibilityChanged(bool state)
{
  m_windowInspectorDock->blockSignals(true);
  m_windowInspectorDock->setChecked(state);
  m_windowInspectorDock->blockSignals(false);
}

void te::process::qt::ProcessBuilder::onWindowTaskDockVisibilityChanged(bool state)
{
  m_windowTaskDock->blockSignals(true);
  m_windowTaskDock->setChecked(state);
  m_windowTaskDock->blockSignals(false);
}

void te::process::qt::ProcessBuilder::onWindowConnectorDockVisibilityChanged(bool state)
{
  m_windowConnectorDock->blockSignals(true);
  m_windowConnectorDock->setChecked(state);
  m_windowConnectorDock->blockSignals(false);
}

void te::process::qt::ProcessBuilder::onWindowRunnerDockVisibilityChanged(bool state)
{
  m_windowRunnerDock->blockSignals(true);
  m_windowRunnerDock->setChecked(state);
  m_windowRunnerDock->blockSignals(false);
}

void te::process::qt::ProcessBuilder::onHelpAboutTriggered()
{
  QMessageBox::information(this, tr("About"), tr("Process Builder 1.0 :)"));
}
