#include "ProcessManager.h"
#include "capabilities/DefaultTaskCapabilities.h"
#include "capabilities/TaskGraphicalCapabilities.h"

#include "../Connector.h"
#include "../Process.h"
#include "../Task.h"



te::process::qt::ProcessManager::ProcessManager(QObject *parent) :
  QObject(parent),
  m_process(nullptr)
{
  m_process = new te::process::Process();
}

te::process::qt::ProcessManager::~ProcessManager()
{
  m_processItemMapCount.clear();

  delete m_process;
}

void te::process::qt::ProcessManager::start()
{
  if (m_process)
    delete m_process;

  m_process = new te::process::Process();
}

void te::process::qt::ProcessManager::reset()
{

}

std::string te::process::qt::ProcessManager::addTask(te::process::TaskSharedPtr task)
{
  std::string itemName = createProcessItemName(task->getType());

  task->setId(itemName);

  m_process->addTask(task);

  emit createdTask(itemName);

  return itemName;
}

std::string te::process::qt::ProcessManager::createTask(const std::string& taskType)
{
  te::process::TaskSharedPtr task(te::process::CreateTask(taskType));

  std::string itemName = createProcessItemName(taskType);

  task->setId(itemName);

  m_process->addTask(task);

  emit createdTask(itemName);

  return itemName;
}

void te::process::qt::ProcessManager::removeTask(const std::string& id)
{
  std::string idToRemove = id;

  te::process::TaskSharedPtr task = m_process->getTask(id);

  std::vector<te::process::Connector*> inputVec = task->getInputConnectors();

  for (te::process::Connector* connector : inputVec)
  {
    removeConnector(connector->getId());
  }

  std::vector<te::process::Connector*> outputVec = task->getOutputConnectors();

  for (te::process::Connector* connector : outputVec)
  {
    removeConnector(connector->getId());
  }

  m_process->removeTask(id);

  emit removedTask(idToRemove);
}

void te::process::qt::ProcessManager::resetTaskId(const std::string& id, const std::string& newId)
{
  std::string currentTaskId = id;
  std::string newTaskId = newId;

  te::process::TaskSharedPtr task = m_process->getTask(id);

  task->setId(newId);

  emit resetedTaskId(currentTaskId, newTaskId);
}

std::string te::process::qt::ProcessManager::createConnector(const std::string& taskFrom, std::size_t paramIndexFrom, const std::string& taskTo, std::size_t paramIndexTo)
{
  te::process::TaskSharedPtr taskFromPtr = m_process->getTask(taskFrom);
  te::process::TaskSharedPtr taskToPtr = m_process->getTask(taskTo);

  te::process::ConnectorSharedPtr connector(new Connector());

  connector->associate(taskFromPtr.get(), paramIndexFrom, taskToPtr.get(), paramIndexTo);

  m_process->addConnector(connector);

  std::string connectorId = connector->getId();

  emit createdConnector(connectorId);

  return connectorId;
}

void te::process::qt::ProcessManager::removeConnector(const std::string& id)
{
  std::string idToRemove = id;

  m_process->removeConnector(id);

  emit removedConnector(idToRemove);
}

te::process::Process* te::process::qt::ProcessManager::getProcess() const
{
  return m_process;
}

bool te::process::qt::ProcessManager::isProcessItemNameValid(const std::string& itemType, const std::string& itemName)
{
  const std::vector<te::process::TaskSharedPtr> tasks = m_process->getTasks();

  for (te::process::TaskSharedPtr task : tasks)
  {
    if (task->getId() == itemName)
    {
      return false;
    }
  }

  return true;
}

std::unique_ptr<te::process::qt::TaskGraphicalCapabilities> te::process::qt::ProcessManager::getTaskGraphicalCapabilities(const std::string& taskType)
{
  te::process::qt::TaskGraphicalCapabilities* task = nullptr;

  try
  {
    task = te::process::qt::TaskGraphicalCapabilitiesFactory::make(taskType);
  }
  catch (...)
  {
    task = te::process::qt::TaskGraphicalCapabilitiesFactory::make("DefaultTask");

    te::process::qt::DefaultTaskCapabilities* defaultTask = dynamic_cast<te::process::qt::DefaultTaskCapabilities*>(task);

    if (defaultTask)
    {
      defaultTask->setTaskType(taskType);
    }
  }

  std::unique_ptr<te::process::qt::TaskGraphicalCapabilities> taskPtr(task);

  return taskPtr;
}

std::string te::process::qt::ProcessManager::createProcessItemName(const std::string& itemType)
{
  std::string newItemName = itemType;

  std::map<std::string, int>::iterator it = m_processItemMapCount.find(itemType);

  if (it == m_processItemMapCount.end())
  {
    m_processItemMapCount[itemType] = 1;
  }
  else
  {
    ++m_processItemMapCount[itemType];
  }

  newItemName += " " + std::to_string(m_processItemMapCount[itemType]);

  bool isValid = isProcessItemNameValid(itemType, newItemName);

  while (isValid != true)
  {
    newItemName = itemType;

    ++m_processItemMapCount[itemType];

    newItemName += " " + std::to_string(m_processItemMapCount[itemType]);

    isValid = isProcessItemNameValid(itemType, newItemName);
  }

  return newItemName;
}