/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

This file is part of the TerraLib - a Framework for building GIS enabled applications.

TerraLib is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

TerraLib is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with TerraLib. See COPYING. If not, write to
TerraLib Team at <terralib-team@terralib.org>.
*/

/*!
 * \file src/terralib/process/qt/TaskGraphicalCapabilities.h
 *
 * \brief Defines a task capabilities of a task used in graphical components
 */

#ifndef __TERRALIB_PROCESS_QT_INTERNAL_TASKGRAPHICALCAPABILITIES_H
#define __TERRALIB_PROCESS_QT_INTERNAL_TASKGRAPHICALCAPABILITIES_H

#include "../../Config.h"

#include "../../../common/AbstractFactory.h"

//QT Includes
#include <QIcon>
#include <QPainterPath>

//STL includes
#include <memory>
#include <string>

namespace te
{
  namespace process
  {
    namespace qt
    {
      /*!
        \class TaskGraphicalCapabilities

        \brief Defines a task capabilities of a task used in graphical components
      */
      class TEPROCESSEXPORT TaskGraphicalCapabilities
      {
      public:
        TaskGraphicalCapabilities(
            const std::string& operationName
          , const std::string& description
          , const std::string& iconName
          , const std::string& iconPath
          );

        //!< Destructor
        virtual ~TaskGraphicalCapabilities();

        //!< Returns the name of the operation
        std::string getName() const;

        //!< Returns the description of the operation
        std::string getDescription() const;

        //!< Returns the icon path
        std::string getIconPath() const;

        //!< Returns the icon of the operation
        const QIcon& getIcon() const;

        virtual QPainterPath getShape(QList<QPointF>& inHotPoints, QList<QPointF>& outHotPoints) const = 0;

        virtual QRectF shapeBoundingRect() const = 0;

      protected:

        std::string m_operationName;  //!< The name of the operation
        std::string m_description;    //!< The caption of the widget
        std::string m_iconName;       //!< The icon name of the operation
        std::string m_iconPath;       //!< The icon path of the operation
        QIcon       m_qIcon;          //!< The icon of the operation

      public:

        QColor m_brushColor;
      };


      /*!
        \class VectorProcessingDialogCapabilitiesFactory

        \brief Abstract factory used to create VectorProcessingDialogCapabilities.
      */
      class TEPROCESSEXPORT TaskGraphicalCapabilitiesFactory : public te::common::AbstractFactory < TaskGraphicalCapabilities, std::string >
      {
      public:

        /*!
          \brief Constructor.

          \param factoryKey The key that identifies the factory.
        */
        TaskGraphicalCapabilitiesFactory(const std::string& factoryKey);

        //!< Destructor
        virtual ~TaskGraphicalCapabilitiesFactory();
      };

    } //end namespace qt
  }//end namespace process
}//end namespace te


#endif //__TERRALIB_PROCESS_QT_INTERNAL_TASKGRAPHICALCAPABILITIES_H
