/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralib/process/qt/AbstractGraphicsItem.h

   \brief An Abstract graphicItem to represents a process Builder item.
 */

#ifndef __TERRALIB_PROCESS_QT_INTERNAL_ABSTRACTGRAPHICSITEM_H
#define __TERRALIB_PROCESS_QT_INTERNAL_ABSTRACTGRAPHICSITEM_H

#include "../../Config.h"

#include "HotPointGraphicsItem.h"

 // Qt Includes
#include <QBrush>
#include <QFont>
#include <QGraphicsObject>
#include <QPainterPath>
#include <QPen>

namespace te
{
  namespace process
  {
    namespace qt
    {
      /*!
      \class AbstractGraphicsItem

      \brief An Abstract graphicItem to represents a process Builder item.

      */
      class TEPROCESSEXPORT AbstractGraphicsItem : public QGraphicsObject
      {
      public:

        /*! \brief Default constructor. */
        AbstractGraphicsItem();

        /*! \brief Virtual destructor. */
        virtual ~AbstractGraphicsItem();

      protected:

        /*!
          \brief Virtual function used to notify when user clicks over a hotpoint.

          \param hotpoint Pointer to the hotpoint pressed.
        */
        virtual void hotPointPressed(HotPointGraphicsItem* hotpoint) = 0;

      protected:

        /*!
          \brief  Virtual function that defines the outer bounds of the item as a rectangle.

          \return QRectF for this item.
        */
        virtual QRectF boundingRect() const = 0;

        /*!
          \brief Function used to draw the process item.

          \param painter Qt Painter used to draw the process item.

          \param option Qt style options used to draw the process item.
        */
        virtual void paint(QPainter* painter, const QStyleOptionGraphicsItem* option, QWidget* widget = nullptr);

        /*!
          \brief Receive hover enter events for this item. The default implementation calls update().

          \param event Qt event.
        */
        void hoverEnterEvent(QGraphicsSceneHoverEvent *event) override;

        /*!
          \brief Receive hover move events for this item. The default implementation does nothing.

          \param event Qt event.
        */
        void hoverMoveEvent(QGraphicsSceneHoverEvent *event) override;

        /*!
          \brief Receive hover leave events for this item. The default implementation calls update().

          \param event Qt event.
        */
        void hoverLeaveEvent(QGraphicsSceneHoverEvent *event) override;

        /*!
          \brief Receive mouse press events for this item.

          \param event Qt event.
        */
        void mousePressEvent(QGraphicsSceneMouseEvent *event) override;

        /*!
          \brief Receive mouse move events for this item.

          \param event Qt event.
        */
        void mouseMoveEvent(QGraphicsSceneMouseEvent *event) override;

        /*!
          \brief Receive mouse release events for this item.

          \param event Qt event.
        */
        void mouseReleaseEvent(QGraphicsSceneMouseEvent *event) override;

      public:

        QPen getPen();

        QBrush getBrush();

        QFont getFont();

        void setPenColor(QColor color);

        void setPenWidth(int width);

        void setPenStyle(Qt::PenStyle style);

        void setBrushColor(QColor color);

        void setBrushStyle(Qt::BrushStyle style);

        void setFont(QFont font);

      protected:

        /*!
          \brief Function used to get hotpoint given a item position.

          \Return Pointer to the hotpoint if in this position has a hotpoint and nullptr in other case.
        */
        HotPointGraphicsItem* getHotPoint(const QPointF& pos);

      protected:

        QPainterPath m_path;                                        //!< Paintr path that represents this item.

        std::vector<HotPointGraphicsItem*> m_hotPoints;             //!< List of all hotpoints for this item.

        bool m_drawHotPoints;                                       //!< Flag used to indicate if the hot points has to be painted.

      protected:

        QPen    m_pen;
        QBrush  m_brush;
        QFont   m_font;
      };
    }
  }
}

#endif // __TERRALIB_PROCESS_QT_INTERNAL_PROCESSSQUAREGRAPHICSITEM_H
