/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*! 
  \file terralib/process/qt/ConnectorGraphicsItem.h

  \brief A GraphicItem to represents connector between two Tasks parameters (hot points).
*/

#ifndef __TERRALIB_PROCESS_QT_INTERNAL_CONNECTORGRAPHICSITEM_H
#define __TERRALIB_PROCESS_QT_INTERNAL_CONNECTORGRAPHICSITEM_H

#include "../../Config.h"

// Qt Includes
#include <QColor>
#include <QGraphicsObject>
#include <QPointF>

namespace te
{
  namespace process
  {
    namespace qt
    {
      class HotPointGraphicsItem;
      class TaskGraphicsItem;

      /*!
      \class ConnectorGraphicsItem

      \brief  A GraphicItem to represents connector between two Tasks parameters (hot points).

      */
      class TEPROCESSEXPORT ConnectorGraphicsItem : public QGraphicsObject
      {
      public:

        /*! 
          \brief Default constructor. 

          \param taskFrom QGraophicsItem that represents FROM task

          \param hotPointFrom QGraphicsItem that represents FROM hotpoint

          \param taskTo QGraophicsItem that represents TO task

          \param hotPointTo QGraphicsItem that represents TO hotpoint
        */
        ConnectorGraphicsItem(const std::string& connectorId, TaskGraphicsItem* taskFrom, HotPointGraphicsItem* hotPointFrom, TaskGraphicsItem* taskTo, HotPointGraphicsItem* hotPointTo);

        /*! \brief Virtual destructor. */
        ~ConnectorGraphicsItem();

        /*! \brief Get function to access connector internal pointer. */
        const std::string getConnectorId() const;

        /*! \brief Get function to notify that the connector was removed. */
        void removed();

      public:

        /*!
          \brief  Virtual function that defines the outer bounds of the item as a rectangle.

          \return QRectF for this item.
        */
        virtual QRectF boundingRect() const;

        /*!
          \brief Function used to draw the process item.

          \param painter Qt Painter used to draw the process item.

          \param option Qt style options used to draw the process item.
        */
        virtual void paint(QPainter* painter, const QStyleOptionGraphicsItem* option, QWidget* widget = nullptr);

      protected:

        /*!
          \brief Function used to get the hot points position in scene coordinate and create a QPainterPath

          \return a valid QPainterPath if connector hotpoints was defined
        */
        QPainterPath getConnectorPainterPath() const;

        QPainterPath getConnectorArrowPainterPath() const;

      protected:

        std::string m_connectorId;                    //!< TerraLib connector object id

        TaskGraphicsItem* m_taskFrom;                 //!< QGraphicsItem that represents FROM Task
        HotPointGraphicsItem* m_hotPointFrom;         //!< QGraphicsItem that represents FROM HotPoint

        TaskGraphicsItem* m_taskTo;                   //!< QGraphicsItem that represents TO Task
        HotPointGraphicsItem* m_hotPointTo;           //!< QGraphicsItem that represents TO HotPoint
      };
    }
  }
}

#endif // __TERRALIB_PROCESS_QT_INTERNAL_CONNECTORGRAPHICSITEM_H
