/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*! 
  \file terralib/process/qt/HotPointGraphicsItem.h

  \brief A GraphicItem to represents Task parameters as hot point.
*/

#ifndef __TERRALIB_PROCESS_QT_INTERNAL_HOTPOINTGRAPHICSITEM_H
#define __TERRALIB_PROCESS_QT_INTERNAL_HOTPOINTGRAPHICSITEM_H

#include "../../Config.h"

// Qt Includes
#include <QColor>
#include <QGraphicsObject>
#include <QPointF>

namespace te
{
  namespace process
  {
    namespace qt
    {
      /*!
      \class HotPointGraphicsItem

      \brief A GraphicItem to represents Task parameters as hot point.

      */
      class TEPROCESSEXPORT HotPointGraphicsItem : public QGraphicsObject
      {
      public:

        /*! \brief Default constructor. */
        HotPointGraphicsItem(const QPointF& position, const std::size_t& index, const std::string& taskId);

        /*! \brief Virtual destructor. */
        ~HotPointGraphicsItem();

      public:

        /*!
          \brief Function used to get the position of this hotpoint

          \return A hotpoint position inside the TaskGraphicsItem.
        */
        QPointF getHotPointPos();

        /*!
          \brief  Virtual function that defines the outer bounds of the item as a rectangle.

          \return QRectF for this item.
        */
        virtual QRectF boundingRect() const;

        /*!
          \brief Function used to draw the process item.

          \param painter Qt Painter used to draw the process item.

          \param option Qt style options used to draw the process item.
        */
        virtual void paint(QPainter* painter, const QStyleOptionGraphicsItem* option, QWidget* widget = nullptr);

        /*!
          \brief Function used to define the color used to draw the hotpoint.

          \param color Qt color value.
        */
        void setHotPointColor(QColor color);

        /*!
          \brief Function used to get the param index from task associated to this hotpoint

          \return The m_index attribute value
        */
        std::size_t getHotPointIndex() const;

      protected:

        QPointF m_pos;                              //!< Position that represents this hot point.

        std::size_t m_index;                        //!< Parameter index from task.

        std::string m_taskId;                       //!< Task id for this hot point parameter.

        QColor m_color;                             //!< Color used to draw this hotpoint.
      };
    }
  }
}

#endif // __TERRALIB_PROCESS_QT_INTERNAL_HOTPOINTGRAPHICSITEM_H
