#include "ProcessGraphicsScene.h"

#include "ConnectorGraphicsItem.h"
#include "HotPointGraphicsItem.h"
#include "TaskGraphicsItem.h"

#include "../explorer/TaskMimeData.h"

#include "../ProcessManager.h"


//Qt Includes
#include <QGraphicsSceneDragDropEvent>
#include <QGraphicsView>
#include <QKeyEvent>
#include <QMimeData>
#include <QPointF>

te::process::qt::ProcessGraphicsScene::ProcessGraphicsScene(ProcessManager* manager, QObject *parent) :
  QGraphicsScene(parent),
  m_manager(manager)
{
  //connects
  connect(m_manager, SIGNAL(removedTask(const std::string&)), this, SLOT(onRemovedTask(const std::string&)));
  connect(m_manager, SIGNAL(resetedTaskId(const std::string&, const std::string&)), this, SLOT(onResetedTaskId(const std::string&, const std::string&)));

  connect(m_manager, SIGNAL(removedConnector(const std::string&)), this, SLOT(onRemovedConnector(const std::string&)));
}

te::process::qt::ProcessGraphicsScene::~ProcessGraphicsScene() = default;

void te::process::qt::ProcessGraphicsScene::createTaskItem(te::process::TaskSharedPtr task, QPointF pos)
{
  std::string taskId = m_manager->addTask(task);

  TaskGraphicsItem* item = new TaskGraphicsItem(m_manager, taskId);

  if (pos.isNull() == false)
  {
    item->setPos(pos);
  }

  addItem(item);
}

void te::process::qt::ProcessGraphicsScene::createConnectorItem(te::process::qt::TaskGraphicsItem* taskFromItem, HotPointGraphicsItem* hotPointFromItem,
                                                                te::process::qt::TaskGraphicsItem* taskToItem, HotPointGraphicsItem* hotPointToItem)
{
  std::string connectorId = m_manager->createConnector(taskFromItem->getTaskId(), hotPointFromItem->getHotPointIndex(), taskToItem->getTaskId(), hotPointToItem->getHotPointIndex());

  ConnectorGraphicsItem* item = new ConnectorGraphicsItem(connectorId, taskFromItem, hotPointFromItem, taskToItem, hotPointToItem);

  //add connector to task TO connectors list.
  taskToItem->addConnector(item);

  addItem(item);
}

void te::process::qt::ProcessGraphicsScene::alignSelecteds(Qt::Alignment alignment)
{
  QList<QGraphicsItem*> graphicItems = selectedItems();

  //check number of selected items
  if (graphicItems.size() < 2)
  {
    return;
  }

  //find rect of all selected items
  QRectF rect;

  for (int i = 0; i < graphicItems.size(); ++i)
  {
    QGraphicsItem* item = graphicItems.at(i);

    QRectF rectItem = item->sceneBoundingRect();

    if (rect.isEmpty())
    {
      rect = rectItem;
    }
    else
    {
      rect = rect.united(rectItem);
    }
  }

  //find point of alignment
  double value;

  switch (alignment)
  {
  case Qt::AlignLeft:
    value = rect.topLeft().x();
    break;
  case Qt::AlignRight:
    value = rect.bottomRight().x();
    break;
  case Qt::AlignTop:
    value = rect.topLeft().y();
    break;
  case Qt::AlignBottom:
    value = rect.bottomLeft().y();
    break;
  case Qt::AlignHCenter:
    value = rect.center().y();
    break;
  case Qt::AlignVCenter:
    value = rect.center().x();
    break;
  }

  //move selected items
  for (int i = 0; i < graphicItems.size(); ++i)
  {
    QGraphicsItem* item = graphicItems.at(i);

    QPointF delta;

    switch (alignment)
    {
      case Qt::AlignLeft: 
      {
        QPointF point = QPointF(item->mapToScene(item->boundingRect().topLeft()).x(), 0);
        delta = QPointF(value, point.y()) - point;
        break;
      }
      case Qt::AlignRight: 
      {
        QPointF point = QPointF(item->mapToScene(item->boundingRect().bottomRight()).x(), 0);
        delta = QPointF(value, point.y()) - point;
        break;
      }
      case Qt::AlignTop: 
      {
        QPointF point = QPointF(0, item->mapToScene(item->boundingRect().topLeft()).y());
        delta = QPointF(point.x(), value)- point;
        break;
      }
      case Qt::AlignBottom: 
      {
        QPointF point = QPointF(0, item->mapToScene(item->boundingRect().bottomLeft()).y());
        delta = QPointF(point.x(), value) - point;
        break;
      }
      case Qt::AlignHCenter:
      {
        QPointF point = QPointF(0, item->mapToScene(item->boundingRect().center()).y());
        delta = QPointF(point.x(), value) - point;
        break;
      }
      case Qt::AlignVCenter:
      {
        QPointF point = QPointF(item->mapToScene(item->boundingRect().center()).x(), 0);
        delta = QPointF(value, point.y()) - point;
        break;
      }
    }

    item->moveBy(delta.x(), delta.y());
  }
}

void te::process::qt::ProcessGraphicsScene::onRemovedTask(const std::string& taskId)
{
  QList<QGraphicsItem*> items = this->items();

  for (int i = 0; i < items.size(); ++i)
  {
    QGraphicsItem* item = items.at(i);

    TaskGraphicsItem* itemGraph = dynamic_cast<TaskGraphicsItem*>(item);

    if (itemGraph)
    {
      if (itemGraph->getTaskId() == taskId)
      {
        this->removeItem(itemGraph);

        break;
      }
    }
  }
}

void te::process::qt::ProcessGraphicsScene::onResetedTaskId(const std::string& id, const std::string& newId)
{
  QList<QGraphicsItem*> items = this->items();

  for (int i = 0; i < items.size(); ++i)
  {
    QGraphicsItem* item = items.at(i);

    TaskGraphicsItem* itemGraph = dynamic_cast<TaskGraphicsItem*>(item);

    if (itemGraph)
    {
      if (itemGraph->getTaskId() == id)
      {
        itemGraph->resetTaskId(newId);

        break;
      }
    }
  }
}

void te::process::qt::ProcessGraphicsScene::onRemovedConnector(const std::string& connectorId)
{
  QList<QGraphicsItem*> items = this->items();

  for (int i = 0; i < items.size(); ++i)
  {
    QGraphicsItem* item = items.at(i);

    ConnectorGraphicsItem* itemGraph = dynamic_cast<ConnectorGraphicsItem*>(item);

    if (itemGraph)
    {
      if (itemGraph->getConnectorId() == connectorId)
      {
        itemGraph->removed();
        this->removeItem(itemGraph);

        break;
      }
    }
  }
}


void te::process::qt::ProcessGraphicsScene::dragEnterEvent(QGraphicsSceneDragDropEvent* event)
{
  const QMimeData* mimeData = event->mimeData();

  QByteArray encodedData = mimeData->data("MimeDataType");

  std::string str = encodedData.constData();

  if (str == "Task")
  {
    event->acceptProposedAction();
  }
  else
  {
    QGraphicsScene::dragEnterEvent(event);
  }
}

void te::process::qt::ProcessGraphicsScene::dragMoveEvent(QGraphicsSceneDragDropEvent *event)
{
  const QMimeData* mimeData = event->mimeData();

  QByteArray encodedData = mimeData->data("MimeDataType");

  std::string str = encodedData.constData();

  if (str == "Task")
  {
    event->acceptProposedAction();
  }
  else
  {
    QGraphicsScene::dragMoveEvent(event);
  }
}

void te::process::qt::ProcessGraphicsScene::dragLeaveEvent(QGraphicsSceneDragDropEvent *event)
{
  const QMimeData* mimeData = event->mimeData();

  QByteArray encodedData = mimeData->data("MimeDataType");

  std::string str = encodedData.constData();

  if (str == "Task")
  {
    event->acceptProposedAction();
  }
  else
  {
    QGraphicsScene::dragLeaveEvent(event);
  }
}

void te::process::qt::ProcessGraphicsScene::dropEvent(QGraphicsSceneDragDropEvent* event)
{
  const QMimeData* mimeData = event->mimeData();

  QByteArray encodedData = mimeData->data("MimeDataType");

  std::string str = encodedData.constData();

  if (str == "Task")
  {
    event->acceptProposedAction();

    const TaskMimeData* taskMimeData = dynamic_cast<const TaskMimeData*>(mimeData);

    if (taskMimeData)
    {
      createTaskItem(taskMimeData->getTask(), event->scenePos());
    }    
  }
  else
  {
    QGraphicsScene::dropEvent(event);
  }
}

void te::process::qt::ProcessGraphicsScene::keyPressEvent(QKeyEvent* event)
{
  if (event->key() == Qt::Key_Delete)
  {
    QList<QGraphicsItem*> graphicItems = selectedItems();

    for (int i = 0; i < graphicItems.size(); ++i)
    {
      QGraphicsItem* selected = graphicItems.at(i);

      TaskGraphicsItem* itemGraph = dynamic_cast<TaskGraphicsItem*>(selected);

      if (itemGraph)
      {
        m_manager->removeTask(itemGraph->getTaskId());
      }
    }
  }
}
