#include "TaskGraphicsItem.h"

#include "ConnectorGraphicsItem.h"
#include "HotPointMimeData.h"
#include "ProcessGraphicsScene.h"

#include "../ProcessManager.h"

#include "../../Connector.h"
#include "../../Process.h"
#include "../../Task.h"
#include "../../TaskCapabilities.h"

// TerraLib
#include "../../../qt/widgets/Utils.h"

// Qt Includes
#include <QBrush>
#include <QDrag>
#include <QFont>
#include <QGraphicsTextItem>
#include <QGraphicsView>
#include <QGraphicsSceneDragDropEvent>
#include <QMessageBox>
#include <QPainter>
#include <QSvgRenderer>

void DrawIcon(QPainter* painter, QRectF rect, QString path)
{
  QRectF iconRect(rect.center().x() - 8, rect.center().y() - 8, 16, 16);

  QSvgRenderer renderer(path);

  renderer.render(painter, iconRect);
}

void DrawTaskText(const std::string& taskName, QPainter* painter, QFont font, QPointF center)
{
  QString text(te::qt::widgets::Convert2QString(taskName));

  QFontMetrics fm(font);
  
  QPainterPath textPath;
  textPath.addText(0, 0, font, text);

  QRectF textRect = textPath.boundingRect();
  center.setX(center.x() - textPath.boundingRect().width() / 2.);
  center.setY(center.y() + textPath.boundingRect().height() / 2.);

  textPath.translate(center);

  painter->fillPath(textPath, QBrush(Qt::black));
}

void DrawParamText(const std::size_t& index, QPainter* painter, QPointF corner)
{
  QString text = QString::number(index);
  
  QFont font("Helvetica", 4);

  QPainterPath textPath;
  textPath.addText(0, 0, font, text);
  textPath.translate(QPointF(corner.x(), corner.y() - 2));
  
  painter->fillPath(textPath, QBrush(Qt::black));
}


te::process::qt::TaskGraphicsItem::TaskGraphicsItem(ProcessManager* manager, const std::string& taskId) :
  AbstractGraphicsItem(),
  m_manager(manager),
  m_taskId(taskId)
{
  m_font = QFont("Helvetica", 4);
  m_pen = QPen(Qt::black);

  setAcceptDrops(true);

  te::process::TaskSharedPtr task = m_manager->getProcess()->getTask(m_taskId);

  m_taskGraphicalCapabilities = m_manager->getTaskGraphicalCapabilities(task->getType());

  m_brush = m_taskGraphicalCapabilities->m_brushColor;

  //get capabilities to check the number of input/output parameters
  const te::process::TaskCapabilities* taskCap = task->getCapabilities();

  std::vector<te::process::ParameterDescriptor> inputParams = taskCap->filter(te::process::ParameterDescriptor::Direction::INPUT);
  std::vector<te::process::ParameterDescriptor> outputParams = taskCap->filter(te::process::ParameterDescriptor::Direction::OUTPUT);

  QList<QPointF> hotPointsIn;
  QList<QPointF> hotPointsOut;

  m_path = m_taskGraphicalCapabilities->getShape(hotPointsIn, hotPointsOut);

  //set shape for input parameters
  for (std::size_t i = 0; i < inputParams.size(); ++i)
  {
    std::size_t index = taskCap->getParameterIndex(inputParams[i].getName());

    m_hotPoints.push_back(new HotPointGraphicsItem(hotPointsIn.at(i) , index, m_taskId));
  }

  //set shape for output parameters
  for (std::size_t i = 0; i < outputParams.size(); ++i)
  {
    std::size_t index = taskCap->getParameterIndex(outputParams[i].getName());

    m_hotPoints.push_back(new HotPointGraphicsItem(hotPointsOut.at(i), index, m_taskId));
  }
}

te::process::qt::TaskGraphicsItem::~TaskGraphicsItem() = default;

void te::process::qt::TaskGraphicsItem::resetTaskId(const std::string& newTaskId)
{
  m_taskId = newTaskId;
}

const std::string te::process::qt::TaskGraphicsItem::getTaskId() const
{
  return m_taskId;
}

void te::process::qt::TaskGraphicsItem::addConnector(ConnectorGraphicsItem* connector)
{
  m_connectors.push_back(connector);
}

void te::process::qt::TaskGraphicsItem::hotPointPressed(HotPointGraphicsItem* hotpoint)
{
  HotPointMimeData* mimeData = new HotPointMimeData(this, hotpoint);
  mimeData->setText("TaskHotPoint");
  
  QDrag *drag = new QDrag(this);
  drag->setMimeData(mimeData);

  Qt::DropAction dropAction = drag->exec();
}

QRectF te::process::qt::TaskGraphicsItem::boundingRect() const
{
  return m_taskGraphicalCapabilities->shapeBoundingRect();
}

void te::process::qt::TaskGraphicsItem::paint(QPainter* painter, const QStyleOptionGraphicsItem* option, QWidget* widget)
{
  //draw default part of this item
  AbstractGraphicsItem::paint(painter, option, widget);

  painter->setPen(m_pen);
  painter->setBrush(m_brush);
  painter->drawPath(m_path);

  //draw task connectors
  for (std::size_t i = 0; i < m_connectors.size(); ++i)
  {
    m_connectors[i]->update();
  }

  //draw task name
  te::process::TaskSharedPtr task = m_manager->getProcess()->getTask(m_taskId);

  if (task)
  {
    //DrawTaskText(task->getType(), painter, m_font, boundingRect().center());

    DrawIcon(painter, m_taskGraphicalCapabilities->shapeBoundingRect(), te::qt::widgets::Convert2QString(m_taskGraphicalCapabilities->getIconPath()));

    //draw task properties names
    const te::process::TaskCapabilities* taskCap = task->getCapabilities();

    for (std::size_t i = 0; i < m_hotPoints.size(); ++i)
    {
      DrawParamText(m_hotPoints[i]->getHotPointIndex(), painter, m_hotPoints[i]->getHotPointPos());
    }
  }
}

void te::process::qt::TaskGraphicsItem::dragEnterEvent(QGraphicsSceneDragDropEvent* event)
{
  m_drawHotPoints = true;

  const QMimeData* mimeData = event->mimeData();

  const te::process::qt::HotPointMimeData* hotPointMimeMdata = dynamic_cast<const te::process::qt::HotPointMimeData*>(mimeData);

  if (hotPointMimeMdata != nullptr)
  {
    event->setAccepted(true);
  }
  else
  {
    event->setAccepted(false);
    QGraphicsObject::dragEnterEvent(event);
  }

  update();
}

void te::process::qt::TaskGraphicsItem::dragMoveEvent(QGraphicsSceneDragDropEvent *event)
{
  const QMimeData* mimeData = event->mimeData();

  const te::process::qt::HotPointMimeData* hotPointMimeMdata = dynamic_cast<const te::process::qt::HotPointMimeData*>(mimeData);

  if (hotPointMimeMdata != nullptr)
  {
    HotPointGraphicsItem* hotPoint = getHotPoint(event->pos());

    if (hotPoint)
    {
      te::process::TaskSharedPtr taskFrom = m_manager->getProcess()->getTask(hotPointMimeMdata->getTask()->getTaskId());
      std::size_t paramFromIndex = hotPointMimeMdata->getHotPoint()->getHotPointIndex();

      te::process::TaskSharedPtr taskTo = m_manager->getProcess()->getTask(this->getTaskId());
      std::size_t paramToIndex = hotPoint->getHotPointIndex();

      std::string msg;

      bool connect = Connector::canAssociate(taskFrom.get(), paramFromIndex, taskTo.get(), paramToIndex, msg);

      if (connect)
      {
        hotPoint->setHotPointColor(Qt::green);
      }
      else
      {
        hotPoint->setHotPointColor(Qt::gray);
      }

      update();
    }

    event->setAccepted(true);
  }
  else
  {
    QGraphicsObject::dragMoveEvent(event);
  }
}

void te::process::qt::TaskGraphicsItem::dragLeaveEvent(QGraphicsSceneDragDropEvent *event)
{
  m_drawHotPoints = false;

  update();
}

void te::process::qt::TaskGraphicsItem::dropEvent(QGraphicsSceneDragDropEvent* event)
{
  const QMimeData* mimeData = event->mimeData();

  const te::process::qt::HotPointMimeData* hotPointMimeMdata = dynamic_cast<const te::process::qt::HotPointMimeData*>(mimeData);

  if (hotPointMimeMdata != nullptr)
  {
    //get hotpoints from this connection
    HotPointGraphicsItem* hotPointTo = getHotPoint(event->pos());

    if (hotPointTo)
    {
      HotPointGraphicsItem* hotPointFrom = hotPointMimeMdata->getHotPoint();

      te::process::TaskSharedPtr taskFrom = m_manager->getProcess()->getTask(hotPointMimeMdata->getTask()->getTaskId());
      std::size_t paramFromIndex = hotPointFrom->getHotPointIndex();

      te::process::TaskSharedPtr taskTo = m_manager->getProcess()->getTask(this->getTaskId());
      std::size_t paramToIndex = hotPointTo->getHotPointIndex();

      std::string msg;

      bool connect = Connector::canAssociate(taskFrom.get(), paramFromIndex, taskTo.get(), paramToIndex, msg);

      if (connect)
      {
        //add to scene
        ProcessGraphicsScene* scene = static_cast<ProcessGraphicsScene*>(this->scene());

        scene->createConnectorItem(hotPointMimeMdata->getTask(), hotPointFrom, this, hotPointTo);

        event->setAccepted(true);
      }
      else
      {
        QMessageBox::warning(nullptr, tr("Warning"), te::qt::widgets::Convert2QString(msg));
      }
    }
  }

  update();
}

