/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*! 
  \file terralib/process/qt/TaskGraphicsItem.h

  \brief A GraphicItem to represents a Task Process Builder item.
*/

#ifndef __TERRALIB_PROCESS_QT_INTERNAL_TASKGRAPHICSITEM_H
#define __TERRALIB_PROCESS_QT_INTERNAL_TASKGRAPHICSITEM_H

#include "../../Config.h"

// TerraLib
#include "AbstractGraphicsItem.h"
#include "../capabilities/TaskGraphicalCapabilities.h"

// STL
#include <memory>

namespace te
{
  namespace process
  {
    namespace qt
    {
      class ConnectorGraphicsItem;

      class ProcessManager;

      /*!
      \class TaskGraphicsItem

      \brief A GraphicItem to represents a Task Process Builder item.

      */
      class TEPROCESSEXPORT TaskGraphicsItem : public AbstractGraphicsItem
      {
      public:

        /*! \brief Default constructor. */
        TaskGraphicsItem(ProcessManager* manager, const std::string& taskId);

        /*! \brief Virtual destructor. */
        ~TaskGraphicsItem();

        /*! \brief Function used to reset the task id. */
        void resetTaskId(const std::string& newTaskId);

        /*! \brief Function used to get the task id. */
         const std::string getTaskId() const;

        /*! \brief Function used to add a new connector item to this task. */
        void addConnector(ConnectorGraphicsItem* connector);

      protected:

        /*!
          \brief Virtual function used to notify when user clicks over a hotpoint.

          \param hotpoint Pointer to the hotpoint pressed.
        */
        virtual void hotPointPressed(HotPointGraphicsItem* hotpoint);

      protected:

        /*!
          \brief  Virtual function that defines the outer bounds of the item as a rectangle.

          \return QRectF for this item.
        */
        virtual QRectF boundingRect() const;

        /*!
          \brief Function used to draw the process item.

          \param painter Qt Painter used to draw the process item.

          \param option Qt style options used to draw the process item.
        */
        virtual void paint(QPainter* painter, const QStyleOptionGraphicsItem* option, QWidget* widget = nullptr);

        /*!
          \brief Virtual function to reimplement qt event.
        */
        virtual void dragEnterEvent(QGraphicsSceneDragDropEvent* event) override;

        /*!
          \brief Virtual function to reimplement qt event.
        */
        virtual void dragMoveEvent(QGraphicsSceneDragDropEvent *event) override;

        /*!
          \brief Virtual function to reimplement qt event.
        */
        virtual void dragLeaveEvent(QGraphicsSceneDragDropEvent *event) override;

        /*!
          \brief Virtual function to reimplement qt event.
        */
        virtual void dropEvent(QGraphicsSceneDragDropEvent* event) override;


      protected:

        ProcessManager* m_manager;

        std::string m_taskId;                               //!< Task id that represents this item.

        std::vector<ConnectorGraphicsItem*> m_connectors;   //!< All connectors that is linked with this task.

        std::unique_ptr<TaskGraphicalCapabilities> m_taskGraphicalCapabilities;
      };
    }
  }
}

#endif // __TERRALIB_PROCESS_QT_INTERNAL_TASKGRAPHICSITEM_H
