#include "ProcessInspector.h"

#include "../ProcessManager.h"

#include "../../Connector.h"
#include "../../Process.h"
#include "../../Task.h"
#include "../../TaskCapabilities.h"

// TerraLib
#include "../../../qt/widgets/Utils.h"

// Qt
#include <QGridLayout>

void RemoveChildren(QTreeWidgetItem* item)
{
  int count = item->childCount();

  for (int i = 0; i < count; ++i)
  {
    QTreeWidgetItem* child = item->takeChild(i);

    delete child;
  }
}

te::process::qt::ProcessInspector::ProcessInspector(ProcessManager* manager, QWidget *parent) :
  QWidget(parent)
  , m_manager(manager)
{
  this->setMinimumWidth(400);

  //setup tree
  QStringList headerlabels;
  headerlabels.append(tr("Item"));
  headerlabels.append(tr("Name"));

  m_processTree = new QTreeWidget(this);
  m_processTree->setAlternatingRowColors(true);
  m_processTree->setColumnCount(2);
  m_processTree->setHeaderLabels(headerlabels);

  m_processTreeItem = new QTreeWidgetItem(m_processTree);
  m_processTreeItem->setExpanded(true);
  m_processTreeItem->setText(0, tr("Process"));

  m_processTree->addTopLevelItem(m_processTreeItem);

  //create main layout
  QGridLayout* layout = new QGridLayout(this);
  layout->setContentsMargins(0, 0, 0, 0);
  layout->addWidget(m_processTree);

  setLayout(layout);

  //connects
  connect(m_manager, SIGNAL(createdTask(const std::string&)), this, SLOT(onAddedTask(const std::string&)));
  connect(m_manager, SIGNAL(removedTask(const std::string&)), this, SLOT(onRemovedTask(const std::string&)));
  connect(m_manager, SIGNAL(resetedTaskId(const std::string&, const std::string&)), this, SLOT(onResetedTaskId(const std::string&, const std::string&)));


  connect(m_manager, SIGNAL(createdConnector(const std::string&)), this, SLOT(onAddedConnector(const std::string&)));
  connect(m_manager, SIGNAL(removedConnector(const std::string&)), this, SLOT(onRemovedConnector(const std::string&)));
}

te::process::qt::ProcessInspector::~ProcessInspector()
{
  m_taskItemMap.clear();
}

void te::process::qt::ProcessInspector::setProcess(te::process::Process* process)
{
  //clear data
  m_taskItemMap.clear();

  RemoveChildren(m_processTreeItem);

  //fill items
  std::vector<te::process::TaskSharedPtr> tasks = process->getTasks();

  for (te::process::TaskSharedPtr task : tasks)
  {
    onAddedTask(task->getId());
  }

  std::vector<te::process::ConnectorSharedPtr> connectors = process->getConnectors();

  for (te::process::ConnectorSharedPtr connector : connectors)
  {
    onAddedConnector(connector->getId());
  }
}

void te::process::qt::ProcessInspector::onAddedTask(const std::string& taskId)
{
  te::process::TaskSharedPtr task = m_manager->getProcess()->getTask(taskId);

  QTreeWidgetItem* taskItem = new QTreeWidgetItem(m_processTreeItem);
  taskItem->setExpanded(true);

  taskItem->setText(0, te::qt::widgets::Convert2QString(task->getType()));
  taskItem->setText(1, te::qt::widgets::Convert2QString(task->getId()));
  taskItem->setData(0, Qt::UserRole, QVariant(te::qt::widgets::Convert2QString(taskId)));
  taskItem->setIcon(0, QIcon::fromTheme("task"));
 
  m_processTreeItem->addChild(taskItem);

  m_taskItemMap[task->getId()] = taskItem;

  listParameters(taskItem, task.get());

  listConnectors(taskItem, task.get());

  m_processTree->resizeColumnToContents(0);
}

void te::process::qt::ProcessInspector::onAddedConnector(const std::string& connectorId)
{
  te::process::ConnectorSharedPtr connector = m_manager->getProcess()->getConnector(connectorId);

  std::map<std::string, QTreeWidgetItem*>::iterator itTaskFrom = m_taskItemMap.find(connector->getTaskFrom()->getId());

  if (itTaskFrom != m_taskItemMap.end())
  {
    te::process::TaskSharedPtr task = m_manager->getProcess()->getTask(itTaskFrom->first);

    listConnectors(itTaskFrom->second, task.get());
  }

  std::map<std::string, QTreeWidgetItem*>::iterator itTaskTo = m_taskItemMap.find(connector->getTaskTo()->getId());

  if (itTaskTo != m_taskItemMap.end())
  {
    te::process::TaskSharedPtr task = m_manager->getProcess()->getTask(itTaskTo->first);

    listConnectors(itTaskTo->second, task.get());
  }
}

void te::process::qt::ProcessInspector::onResetedTaskId(const std::string& id, const std::string& newId)
{
  std::map<std::string, QTreeWidgetItem*>::iterator itTask = m_taskItemMap.find(id);

  if (itTask != m_taskItemMap.end())
  {
    te::process::TaskSharedPtr task = m_manager->getProcess()->getTask(newId);

    QTreeWidgetItem* taskItem = itTask->second;
    
    taskItem->setText(0, te::qt::widgets::Convert2QString(task->getType()));
    taskItem->setText(1, te::qt::widgets::Convert2QString(task->getId()));
    taskItem->setData(0, Qt::UserRole, QVariant(te::qt::widgets::Convert2QString(newId)));
    
    m_taskItemMap[newId] = taskItem;

    m_taskItemMap.erase(id);
  }
}

void te::process::qt::ProcessInspector::onRemovedTask(const std::string& taskId)
{
  std::map<std::string, QTreeWidgetItem*>::iterator it = m_taskItemMap.find(taskId);

  if (it != m_taskItemMap.end())
  {
    m_processTreeItem->removeChild(it->second);
  }
  else
  {
    setProcess(m_manager->getProcess());
  }
}

void te::process::qt::ProcessInspector::onRemovedConnector(const std::string& connectorId)
{
  std::map<std::string, QTreeWidgetItem*>::iterator it = m_taskItemMap.begin();

  while (it != m_taskItemMap.end())
  {
    QTreeWidgetItem* taskItem = it->second;

    for (int i = 0; i < taskItem->childCount(); ++i)
    {
      QTreeWidgetItem* taskParamItem = taskItem->child(i);

      for (int j = 0; j < taskParamItem->childCount(); ++j)
      {
        QTreeWidgetItem* connectorItem = taskParamItem->child(j);

        QVariant var = connectorItem->data(0, Qt::UserRole);

        if (var.isValid())
        {
          std::string curConnectorId = te::qt::widgets::Convert2StdString(var.toString());

          if (curConnectorId == connectorId)
          {
            taskParamItem->removeChild(connectorItem);

            break;
          }
        }
      }
    }

    ++it;
  }
}

void te::process::qt::ProcessInspector::listParameters(QTreeWidgetItem* taskItemParent, const te::process::Task* task)
{
  const te::process::TaskCapabilities* capabilities = task->getCapabilities();

  const std::vector<te::process::ParameterDescriptor> params = capabilities->getParameterDescriptors();

  for (std::size_t i = 0; i < params.size(); ++i)
  {
    te::process::ParameterDescriptor param = params[i];

    std::string direction = "parameter_output";

    if (param.getDirection() == te::process::ParameterDescriptor::Direction::INPUT)
    {
      direction = "parameter_input";
    }

    QTreeWidgetItem* paramItem = new QTreeWidgetItem(taskItemParent);

    paramItem->setIcon(0, QIcon::fromTheme(direction.c_str()));
    paramItem->setText(1, te::qt::widgets::Convert2QString(param.getName()));
    paramItem->setData(0, Qt::UserRole, QVariant(static_cast<int>(i)));

    taskItemParent->addChild(paramItem);
  }
}

void te::process::qt::ProcessInspector::listConnectors(QTreeWidgetItem* taskItemParent, const te::process::Task* task)
{
  std::vector<te::process::Connector*> inputConnectors = task->getInputConnectors();

  for (std::size_t i = 0; i < inputConnectors.size(); ++i)
  {
    addConnector(taskItemParent, inputConnectors[i], true);
  }

  std::vector<te::process::Connector*> outputConnectors = task->getOutputConnectors();

  for (std::size_t i = 0; i < outputConnectors.size(); ++i)
  {
    addConnector(taskItemParent, outputConnectors[i], false);
  }
}

void te::process::qt::ProcessInspector::addConnector(QTreeWidgetItem* taskItemParent, te::process::Connector* connector, bool inputConnectors)
{
  std::string taskId;

  QVariant taskVar = taskItemParent->data(0, Qt::UserRole);

  if (taskVar.isValid())
  {
    taskId = te::qt::widgets::Convert2StdString(taskVar.toString());
  }

  std::size_t index = std::string::npos;

  if (taskId == connector->getTaskFrom()->getId())
  {
    index = connector->getParamIndexFrom();
  }
  else if (taskId == connector->getTaskTo()->getId())
  {
    index = connector->getParamIndexTo();
  }

  std::string direction = "connector_sender";
  std::string taskDir = connector->getTaskTo()->getId();
  
  if (inputConnectors)
  {
    direction = "connector_receiver";
    taskDir = connector->getTaskFrom()->getId();
  }

  //get taskItemChild
  if (index != std::string::npos)
  {
    QTreeWidgetItem* taskParamItem = taskItemParent->child(index);

    if (taskParamItem)
    {
      RemoveChildren(taskParamItem);

      QTreeWidgetItem* connectorItem = new QTreeWidgetItem(taskParamItem);

      connectorItem->setIcon(0, QIcon::fromTheme(direction.c_str()));
      connectorItem->setText(1, te::qt::widgets::Convert2QString(taskDir));
      connectorItem->setData(0, Qt::UserRole, QVariant(te::qt::widgets::Convert2QString(connector->getId())));

      taskParamItem->addChild(connectorItem);

      taskParamItem->setExpanded(true);

      m_processTree->resizeColumnToContents(0);
    }
  }
}
