#include "TaskPropertyHandler.h"

#include "../ProcessManager.h"

#include "../../Connector.h"
#include "../../Process.h"
#include "../../TaskCapabilities.h"
#include "../../SimpleDataParameterValue.h"


// TerraLib Includes
#include "../../../datatype/SimpleData.h"
#include "../../../datatype/Utils.h"
#include "../../../qt/widgets/propertybrowser/AbstractPropertyManager.h"
#include "../../../qt/widgets/Utils.h"

te::process::qt::TaskPropertyHandler::TaskPropertyHandler(const std::string& taskType)
  : m_taskType(taskType)
  , m_manager(nullptr)
{

}

te::process::qt::TaskPropertyHandler::~TaskPropertyHandler()
{
  m_propertyMap.clear();
}

void te::process::qt::TaskPropertyHandler::setTask(const std::string& taskId, ProcessManager* manager)
{
  m_taskId = taskId;
  m_manager = manager;
}

const std::string& te::process::qt::TaskPropertyHandler::getType() const
{
  return m_taskType;
}

void te::process::qt::TaskPropertyHandler::setDefaultProperties(QtProperty* parent)
{
  te::process::TaskSharedPtr task = m_manager->getProcess()->getTask(m_taskId);

  QtProperty* type = te::qt::widgets::AbstractPropertyManager::getInstance().m_stringManager->addProperty(te::qt::widgets::Convert2QString("Type"));
  type->setEnabled(false);
  parent->addSubProperty(type);

  QtProperty* name = te::qt::widgets::AbstractPropertyManager::getInstance().m_stringManager->addProperty(te::qt::widgets::Convert2QString("Name"));
  parent->addSubProperty(name);

  te::qt::widgets::AbstractPropertyManager::getInstance().m_stringManager->setValue(type, te::qt::widgets::Convert2QString(task->getType())); //type
  te::qt::widgets::AbstractPropertyManager::getInstance().m_stringManager->setValue(name, te::qt::widgets::Convert2QString(task->getId())); //name
}

void te::process::qt::TaskPropertyHandler::setInputProperties(QtProperty* parent)
{
  setProperty(parent, te::process::ParameterDescriptor::Direction::INPUT);

}

void te::process::qt::TaskPropertyHandler::setOutputProperties(QtProperty* parent)
{
  setProperty(parent, te::process::ParameterDescriptor::Direction::OUTPUT);
}

void te::process::qt::TaskPropertyHandler::valueChanged(QtProperty* p, const QVariant& qVariant)
{
  te::process::TaskSharedPtr task = m_manager->getProcess()->getTask(m_taskId);

  std::string propName = te::qt::widgets::Convert2StdString(p->propertyName());

  std::map<std::string, QtProperty*>::iterator it = m_propertyMap.find(propName);

  if (it != m_propertyMap.end())
  {
    const te::process::TaskCapabilities* taskCap = task->getCapabilities();

    std::size_t paramIndex = taskCap->getParameterIndex(propName);

    te::process::ParameterDescriptor paramDesc = taskCap->getParameterDescriptor(paramIndex);

    te::dt::AbstractData* ad = createPropertyData(paramDesc, qVariant);

    te::process::SimpleDataParameterValue* Value = new te::process::SimpleDataParameterValue(ad);

    task->setParameter(paramIndex, Value, true);
  }
}

void te::process::qt::TaskPropertyHandler::setProperty(QtProperty* parent, te::process::ParameterDescriptor::Direction dir)
{
  te::process::TaskSharedPtr task = m_manager->getProcess()->getTask(m_taskId);

  const te::process::TaskCapabilities* taskCap = task->getCapabilities();

  std::vector<te::process::ParameterDescriptor> params = taskCap->filter(dir);

  for (te::process::ParameterDescriptor param : params)
  {
    QtProperty* prop = createProperty(task, param);

    prop->setToolTip(getParameterToolTip(task, param));

    //check if alredy has value defined
    te::dt::AbstractData* ad = getParamData(param.getName());

    if (ad)
    {
      setPropertyData(ad, prop, param.getDataType());
    }

    parent->addSubProperty(prop);

    m_propertyMap[param.getName()] = prop;
  }
}

te::dt::AbstractData* te::process::qt::TaskPropertyHandler::getParamData(const std::string& paramName)
{
  te::dt::AbstractData* ad = nullptr;

  te::process::TaskSharedPtr task = m_manager->getProcess()->getTask(m_taskId);

  const te::process::TaskCapabilities* taskCap = task->getCapabilities();

  std::size_t paramIndex = taskCap->getParameterIndex(paramName);

  te::process::ParameterDescriptor paramDesc = taskCap->getParameterDescriptor(paramIndex);

  //check if parameter has a default value
  if (paramDesc.getDefaultValue())
  {
    const te::process::ParameterValue* pValue = paramDesc.getDefaultValue();
    const te::dt::AbstractData* absData = (static_cast<const te::process::SimpleDataParameterValue*>(pValue))->getValue();
    ad = absData->clone();
  }

  //check if already has a property defined
  if (task->getParameter(paramIndex))
  {
    delete ad;

    const te::process::ParameterValue* pValue = task->getParameter(paramIndex);
    const te::dt::AbstractData* absData = (static_cast<const te::process::SimpleDataParameterValue*>(pValue))->getValue();

    ad = absData->clone();
  }

  return ad;
}

te::dt::AbstractData* te::process::qt::TaskPropertyHandler::createPropertyData(te::process::ParameterDescriptor& paramDesc, const QVariant& var)
{
  te::dt::AbstractData* ad = nullptr;

  int dataType = paramDesc.getDataType();

  switch (dataType)
  {
  case te::dt::DOUBLE_TYPE:
  {
    double val = var.toDouble();
    ad = new te::dt::SimpleData<double, te::dt::DOUBLE_TYPE>(val);

    break;
  }
  case te::dt::STRING_TYPE:
  {
    std::string val = te::qt::widgets::Convert2StdString(var.toString());
    ad = new te::dt::SimpleData<std::string, te::dt::STRING_TYPE>(val);

    break;
  }
  case te::dt::INT16_TYPE:
  case te::dt::INT32_TYPE:
  case te::dt::INT64_TYPE:
  {
    int val = var.toInt();
    ad = new te::dt::SimpleData<int32_t, te::dt::INT32_TYPE>(val);

    break;
  }
  case te::dt::UINT16_TYPE:
  case te::dt::UINT32_TYPE:
  case te::dt::UINT64_TYPE:
  {
    int val = var.toInt();
    ad = new te::dt::SimpleData<int32_t, te::dt::INT32_TYPE>(val);

    break;
  }
  default:
  {
    throw te::common::Exception(TE_TR("Unexpected data type!"));
  }
  }

  return ad;
}

QString te::process::qt::TaskPropertyHandler::getParameterToolTip(te::process::TaskSharedPtr task, te::process::ParameterDescriptor& paramDesc)
{
  const te::process::TaskCapabilities* taskCap = task->getCapabilities();

  std::size_t paramIndex = taskCap->getParameterIndex(paramDesc.getName());

  QString propName;
  propName.append("Index: ");
  propName.append(QString::number(paramIndex));
  propName.append("\n");
  propName.append("Type:");
  propName.append(te::qt::widgets::Convert2QString(te::dt::ConvertDataTypeToString(paramDesc.getDataType())));

  return propName;
}

te::process::qt::TaskPropertyHandlerFactory::TaskPropertyHandlerFactory(const std::string& taskType)
  : te::common::AbstractFactory<te::process::qt::TaskPropertyHandler, std::string>(taskType)
{
}

te::process::qt::TaskPropertyHandlerFactory::~TaskPropertyHandlerFactory() = default;

te::process::qt::TaskPropertyHandler* te::process::qt::CreateTaskPropertyHandler(const std::string& taskType, const std::string& taskId, ProcessManager* manager)
{
  te::process::qt::TaskPropertyHandler*  taskHandler = nullptr;
  
  try
  {
    taskHandler = TaskPropertyHandlerFactory::make(taskType);
  }
  catch (...)
  {
    taskHandler = TaskPropertyHandlerFactory::make("DefaultTaskPropertyHandler");
  }
  
  taskHandler->setTask(taskId, manager);

  return taskHandler;
}