/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralib/qt/af/ApplicationDocker.cpp

   \brief The base API for docker controllers of TerraLib applications.
 */

#include "ApplicationDocker.h"

// TerraLib Includes
#include "../widgets/Utils.h"

// Qt Includes
#include <QAction>
#include <QDockWidget>
#include <QEvent>
#include <QToolBar>


te::qt::af::ApplicationDocker::ApplicationDocker(QWidget* parent, Qt::ToolBarArea area)
  : QMainWindow(parent),
  m_currentAction(nullptr),
  m_minimumWidth(0)
{
  this->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Preferred);

  m_sideBar = new QToolBar(this);
  m_sideBar->setWindowTitle(tr("Dock Side Bar"));
  m_sideBar->setMovable(false);

  m_sideBar->installEventFilter(this);
 
  this->addToolBar(area, m_sideBar);
}

te::qt::af::ApplicationDocker::~ApplicationDocker()
{
  std::map<QAction*, ApplicationDockerInfo>::iterator it = m_dockInfoActionMap.begin();

  while (it != m_dockInfoActionMap.end())
  {
    if (it->second.m_isCloseable)
    {
      it->second.m_dockWidget->close();
    }

    ++it;
  }

  m_dockInfoActionMap.clear();
}

bool te::qt::af::ApplicationDocker::eventFilter(QObject* /*watched*/, QEvent* e)
{
  switch (e->type())
  {
    case QEvent::ContextMenu:
    {
      return true;
    }
    break;

    default:
      break;
  }

  return false;
}

QSize te::qt::af::ApplicationDocker::sizeHint() const
{
  return m_sizeHint;
}

QSize te::qt::af::ApplicationDocker::minimumSizeHint() const
{
  return QSize(m_sideBar->width(), 100);
}

bool te::qt::af::ApplicationDocker::hasActiveDock()
{
  if (m_currentAction)
  {
    return true;
  }

  return false;
}

bool te::qt::af::ApplicationDocker::hasDockWidget(QDockWidget* dock)
{
  std::map<QAction*, ApplicationDockerInfo>::iterator it = m_dockInfoActionMap.begin();

  while (it != m_dockInfoActionMap.end())
  {
    if (it->second.m_dockWidget == dock)
    {
      return true;
    }

    ++it;
  }

  return false;
}

void te::qt::af::ApplicationDocker::raiseDockWidget(QDockWidget* dock)
{
  std::map<QAction*, ApplicationDockerInfo>::iterator it = m_dockInfoActionMap.begin();

  while (it != m_dockInfoActionMap.end())
  {
    if (it->second.m_dockWidget == dock)
    {
      it->first->setChecked(true);
    }

    ++it;
  }
}

void te::qt::af::ApplicationDocker::addDock(const ApplicationDockerInfo& dockInfo, Qt::DockWidgetArea area)
{
  dockInfo.m_dockWidget->setTitleBarWidget(new QWidget(dockInfo.m_dockWidget));
  dockInfo.m_dockWidget->setVisible(false);

  addDockWidget(area, dockInfo.m_dockWidget);

  QAction* actionDock = new QAction(QString("..."), nullptr);
  actionDock->setIcon(dockInfo.m_icon);
  actionDock->setCheckable(true);
  actionDock->setToolTip(dockInfo.m_dockName);

  connect(actionDock, SIGNAL(toggled(bool)), this, SLOT(onDockActionVisibilityChanged(bool)));

  m_sideBar->addAction(actionDock);

  m_dockInfoActionMap.insert(std::map<QAction*, ApplicationDockerInfo>::value_type(actionDock, dockInfo));

  //set minimum width
  if (dockInfo.m_dockWidget->minimumWidth() > m_minimumWidth)
  {
    m_minimumWidth = dockInfo.m_dockWidget->minimumWidth();
  }
}

void te::qt::af::ApplicationDocker::takeDock(ApplicationDockerInfo& dockInfo)
{
  dockInfo = m_dockInfoActionMap[m_currentAction];

  removeDockWidget(dockInfo.m_dockWidget);

  m_sideBar->removeAction(m_currentAction);

  m_dockInfoActionMap.erase(m_currentAction);

  delete m_currentAction;

  m_currentAction = nullptr;

  this->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Preferred);

  m_sizeHint = m_sideBar->size();

  emit clearDockSelected();

  computeMinimumSize();
}

void te::qt::af::ApplicationDocker::removeDock(QDockWidget* dock)
{
  std::map<QAction*, ApplicationDockerInfo>::iterator it = m_dockInfoActionMap.begin();

  while (it != m_dockInfoActionMap.end())
  {
    if (it->second.m_dockWidget == dock)
    {
      break;
    }

    ++it;
  }

  if (it == m_dockInfoActionMap.end())
  {
    return;
  }

  QAction* action = it->first;

  removeDockWidget(it->second.m_dockWidget);

  m_sideBar->removeAction(action);

  m_dockInfoActionMap.erase(action);

  if (action == m_currentAction)
  {
    delete m_currentAction;

    m_currentAction = nullptr;

    this->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Preferred);

    m_sizeHint = m_sideBar->size();

    emit clearDockSelected();
  }

  computeMinimumSize();
}

bool te::qt::af::ApplicationDocker::isCloseable()
{
  if (m_currentAction)
  {
    return m_dockInfoActionMap[m_currentAction].m_isCloseable;
  }

  return false;
}

int te::qt::af::ApplicationDocker::getMinimumWidth()
{
  return computeGlobalMinSize() + m_sideBar->width();
}

void te::qt::af::ApplicationDocker::computeMinimumSize()
{
  if (m_currentAction == nullptr)
  {
    this->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Preferred);

    m_sizeHint = m_sideBar->size();

    return;
  }

  m_minimumWidth = m_sideBar->width();

  std::map<QAction*, ApplicationDockerInfo>::iterator it = m_dockInfoActionMap.begin();

  while (it != m_dockInfoActionMap.end())
  {
    int width = it->second.m_dockWidget->minimumWidth();

    if (width > m_minimumWidth)
    {
      m_minimumWidth = it->second.m_dockWidget->width();
    }

    ++it;
  }

  int minGlobalSize = computeGlobalMinSize();

  if (m_minimumWidth < minGlobalSize)
  {
    m_minimumWidth = minGlobalSize;
  }

  this->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Preferred);

  m_sizeHint = QSize(m_sideBar->width() + m_minimumWidth, m_sideBar->height());
}

int te::qt::af::ApplicationDocker::computeGlobalMinSize()
{
  int  minSize = 0;

  std::map<QAction*, ApplicationDockerInfo>::iterator it = m_dockInfoActionMap.begin();

  while (it != m_dockInfoActionMap.end())
  {
    int width = it->second.m_dockWidget->minimumWidth();

    if (width > minSize)
    {
      minSize = width;
    }

    ++it;
  }

  return minSize;
}

void te::qt::af::ApplicationDocker::resizeEvent(QResizeEvent *event)
{
  computeMinimumSize();
}

void te::qt::af::ApplicationDocker::onDockActionVisibilityChanged(bool state)
{
  QAction* action = dynamic_cast<QAction*>(QObject::sender());

  if (action)
  {
    if (m_currentAction)
    {
      QDockWidget* curDockWidget = m_dockInfoActionMap[m_currentAction].m_dockWidget;

      curDockWidget->setVisible(false);

      m_currentAction->blockSignals(true);
      m_currentAction->setChecked(false);
      m_currentAction->blockSignals(false);
    }

    QDockWidget* newDockWidget = m_dockInfoActionMap[action].m_dockWidget;

    if (state)
    {
      m_currentAction = action;

      this->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Preferred);

      newDockWidget->resize(QSize(m_minimumWidth, m_sideBar->height()));

      newDockWidget->setVisible(state);

      newDockWidget->repaint();

      this->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Preferred);

      m_sizeHint = QSize(m_sideBar->width() + m_minimumWidth, m_sideBar->height());

      emit dockSelected(m_currentAction->toolTip());
    }
    else
    {
      m_currentAction = nullptr;

      this->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Preferred);

      m_sizeHint = m_sideBar->size();

      emit clearDockSelected();
    }
  }
}

void te::qt::af::ApplicationDocker::onCloseDock()
{
  if (m_currentAction)
  {
    removeDockWidget(m_dockInfoActionMap[m_currentAction].m_dockWidget);

    m_dockInfoActionMap[m_currentAction].m_dockWidget->close();

    m_sideBar->removeAction(m_currentAction);

    m_dockInfoActionMap.erase(m_currentAction);

    delete m_currentAction;

    m_currentAction = nullptr;

    this->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Preferred);

    m_sizeHint = m_sideBar->size();

    emit clearDockSelected();

    computeMinimumSize();
  }
}
