/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralib/qt/af/ApplicationDocker.cpp

   \brief The base API for docker controllers of TerraLib applications.
 */

#include "ApplicationDockerController.h"


te::qt::af::ApplicationDockerController::ApplicationDockerController(QMainWindow* mainWindow)
  : QObject(mainWindow),
  m_mainWindow(mainWindow)
{
  startDocker(m_leftDockWidget, Qt::LeftDockWidgetArea, m_leftAppDocker, m_leftAppDockerTitleBar, m_leftDockVisibleAction, tr("Left Dock"));
  
  startDocker(m_rightDockWidget, Qt::RightDockWidgetArea, m_rightAppDocker, m_rightAppDockerTitleBar, m_rightDockVisibleAction, tr("Right Dock"));
}

te::qt::af::ApplicationDockerController::~ApplicationDockerController()
{
  delete m_leftAppDocker;
  delete m_rightAppDocker;
}

void te::qt::af::ApplicationDockerController::addDockWidget(const ApplicationDockerInfo& dockInfo, Qt::DockWidgetArea area)
{
  if (area == Qt::LeftDockWidgetArea)
  {
    m_leftAppDocker->addDock(dockInfo, Qt::LeftDockWidgetArea);
  }
  else if (area == Qt::RightDockWidgetArea)
  {
    m_rightAppDocker->addDock(dockInfo, Qt::RightDockWidgetArea);
  }
}

void te::qt::af::ApplicationDockerController::removeDockWidget(QDockWidget* dock)
{
  if (m_leftAppDocker->hasDockWidget(dock))
  {
    m_leftAppDocker->removeDock(dock);
    return;
  }

  if (m_rightAppDocker->hasDockWidget(dock))
  {
    m_rightAppDocker->removeDock(dock);
    return;
  }
}

void te::qt::af::ApplicationDockerController::addDockerVisibilityActions(QMenu* menu)
{
  menu->addAction(m_leftDockVisibleAction);
  menu->addAction(m_rightDockVisibleAction);
}

void te::qt::af::ApplicationDockerController::raiseDockWidget(QDockWidget* dock)
{
  if (m_leftAppDocker->hasDockWidget(dock))
  {
    m_leftAppDocker->raiseDockWidget(dock);
    return;
  }

  if (m_rightAppDocker->hasDockWidget(dock))
  {
    m_rightAppDocker->raiseDockWidget(dock);
    return;
  }
}

te::qt::af::ApplicationDockerInfo te::qt::af::ApplicationDockerController::createAppDockInfo(QDockWidget* dock, const QIcon& icon, const QString& dockName, bool isCloseable)
{
  te::qt::af::ApplicationDockerInfo appDockInfo;

  appDockInfo.m_dockName = dockName;
  appDockInfo.m_dockWidget = dock;
  appDockInfo.m_icon = icon;
  appDockInfo.m_isCloseable = isCloseable;

  return appDockInfo;
}

void te::qt::af::ApplicationDockerController::onDockSelected(QString dockName)
{
  ApplicationDocker* appDocker = dynamic_cast<ApplicationDocker*>(QObject::sender());

  if (appDocker && appDocker == m_leftAppDocker)
  {
    m_leftAppDockerTitleBar->setDockInfo(dockName, appDocker->isCloseable(), m_leftAppDocker->getMinimumWidth());

    m_leftAppDockerTitleBar->resize(m_leftAppDocker->getMinimumWidth(), m_leftAppDockerTitleBar->height());
  }
  else if (appDocker && appDocker == m_rightAppDocker)
  {
    m_rightAppDockerTitleBar->setDockInfo(dockName, appDocker->isCloseable(), m_rightAppDocker->getMinimumWidth());

    m_rightAppDockerTitleBar->resize(m_rightAppDocker->getMinimumWidth(), m_rightAppDockerTitleBar->height());
  }

  appDocker->adjustSize();
}

void te::qt::af::ApplicationDockerController::onClearDockSelected()
{
  ApplicationDocker* appDocker = dynamic_cast<ApplicationDocker*>(QObject::sender());

  if (appDocker && appDocker == m_leftAppDocker)
  {
    m_leftAppDockerTitleBar->reset();
  }
  else if (appDocker && appDocker == m_rightAppDocker)
  {
    m_rightAppDockerTitleBar->reset();
  }
}

void te::qt::af::ApplicationDockerController::onPinDock()
{
  ApplicationDockerTitleBar* appDockerTitleBar = dynamic_cast<ApplicationDockerTitleBar*>(QObject::sender());

  if (appDockerTitleBar && appDockerTitleBar == m_leftAppDockerTitleBar)
  {
    m_leftDockWidget->setFloating(!m_leftDockWidget->isFloating());
  }
  else if (appDockerTitleBar && appDockerTitleBar == m_rightAppDockerTitleBar)
  {
    m_rightDockWidget->setFloating(!m_rightDockWidget->isFloating());
  }
}

void te::qt::af::ApplicationDockerController::onMoveDock()
{
  ApplicationDockerTitleBar* appDockerTitleBar = dynamic_cast<ApplicationDockerTitleBar*>(QObject::sender());

  if (appDockerTitleBar && appDockerTitleBar == m_leftAppDockerTitleBar)
  {
    if (m_leftAppDocker->hasActiveDock())
    {
      te::qt::af::ApplicationDockerInfo dockerInfo;
      m_leftAppDocker->takeDock(dockerInfo);

      m_rightAppDocker->addDock(dockerInfo, Qt::RightDockWidgetArea);

      m_rightAppDocker->raiseDockWidget(dockerInfo.m_dockWidget);
    }
  }
  else if (appDockerTitleBar && appDockerTitleBar == m_rightAppDockerTitleBar)
  {
    if (m_rightAppDocker->hasActiveDock())
    {
      te::qt::af::ApplicationDockerInfo dockerInfo;
      m_rightAppDocker->takeDock(dockerInfo);

      m_leftAppDocker->addDock(dockerInfo, Qt::RightDockWidgetArea);

      m_leftAppDocker->raiseDockWidget(dockerInfo.m_dockWidget);
    }
  }
}

void te::qt::af::ApplicationDockerController::onDockVisibilityChanged(bool state)
{
  QDockWidget* dock = dynamic_cast<QDockWidget*>(QObject::sender());

  if (dock && dock == m_leftDockWidget)
  {
    m_leftDockVisibleAction->blockSignals(true);
    m_leftDockVisibleAction->setChecked(state);
    m_leftDockVisibleAction->blockSignals(false);
  }
  else if (dock && dock == m_rightDockWidget)
  {
    m_rightDockVisibleAction->blockSignals(true);
    m_rightDockVisibleAction->setChecked(state);
    m_rightDockVisibleAction->blockSignals(false);
  }
}

void te::qt::af::ApplicationDockerController::startDocker(QDockWidget*& dockWidget, Qt::DockWidgetArea area, ApplicationDocker*& appDocker, ApplicationDockerTitleBar*& appDockerTitleBar, QAction*& action, QString title)
{
  //docker
  dockWidget = new QDockWidget(m_mainWindow);
  dockWidget->setWindowTitle(title);
  dockWidget->setAllowedAreas(area);

  appDockerTitleBar = new ApplicationDockerTitleBar(dockWidget, area);

  dockWidget->setTitleBarWidget(appDockerTitleBar);

  if (area == Qt::RightDockWidgetArea)
  {
    appDocker = new ApplicationDocker(m_mainWindow, Qt::RightToolBarArea);
  }
  else if (area == Qt::LeftDockWidgetArea)
  {
    appDocker = new ApplicationDocker(m_mainWindow, Qt::LeftToolBarArea);
  }

  dockWidget->setWidget(appDocker);

  m_mainWindow->addDockWidget(area, dockWidget);

  // actions
  action = new QAction(this);
  action->setText(title);
  action->setCheckable(true);
  action->setToolTip(tr("Set dock visibility."));

  if (area == Qt::RightDockWidgetArea)
  {
    action->setIcon(QIcon::fromTheme("document-import"));
  }
  else if (area == Qt::LeftDockWidgetArea)
  {
    action->setIcon(QIcon::fromTheme("document-export"));
  }

  //connects
  connect(action, SIGNAL(toggled(bool)), dockWidget, SLOT(setVisible(bool)));
  connect(dockWidget, SIGNAL(visibilityChanged(bool)), this, SLOT(onDockVisibilityChanged(bool)));
  connect(appDocker, SIGNAL(dockSelected(QString)), this, SLOT(onDockSelected(QString)));
  connect(appDocker, SIGNAL(clearDockSelected()), this, SLOT(onClearDockSelected()));
  connect(appDockerTitleBar, SIGNAL(pinTriggered()), this, SLOT(onPinDock()));
  connect(appDockerTitleBar, SIGNAL(moveTriggered()), this, SLOT(onMoveDock()));
  connect(appDockerTitleBar, SIGNAL(closeTriggered()), appDocker, SLOT(onCloseDock()));
}
