/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralib/qt/af/ApplicationDockerTitleBar.cpp

   \brief The base API for docker title bar for TerraLib Docker.
 */

#include "ApplicationDockerTitleBar.h"

// Qt Includes
#include <QApplication>
#include <QAction>
#include <QFrame>
#include <QGridLayout>
#include <QHBoxLayout>
#include <QLabel>
#include <QStyle>
#include <QToolButton>


te::qt::af::ApplicationDockerTitleBar::ApplicationDockerTitleBar(QWidget* parent, Qt::DockWidgetArea area)
  : QWidget(parent)
{
  //build ui
  QGridLayout* widgetLayout = new QGridLayout(this);
  widgetLayout->setObjectName("widgetLayout");
  widgetLayout->setContentsMargins(0, 0, 0, 0);
  this->setLayout(widgetLayout);

  QFrame* frame = new QFrame(this);
  frame->setFrameShape(QFrame::Panel);
  frame->setFrameShadow(QFrame::Raised);

  widgetLayout->addWidget(frame);

  QGridLayout* frameLayout = new QGridLayout(frame);
  frameLayout->setObjectName("frameLayout");
  frameLayout->setContentsMargins(1, 1, 1, 1);
  frameLayout->setHorizontalSpacing(1);
  frame->setLayout(frameLayout);

  QAction* pinAction = new QAction(this);
  pinAction->setIcon(QIcon::fromTheme("pin"));
  pinAction->setCheckable(false);
  pinAction->setToolTip(tr("Pin Dock"));

  QToolButton* pinToolButton = new QToolButton(this);
  pinToolButton->setDefaultAction(pinAction);
  pinToolButton->setIconSize(QSize(8, 8));
  pinToolButton->setAutoRaise(true);
 
  QAction* moveAction = new QAction(this);
  moveAction->setCheckable(false);
  moveAction->setToolTip(tr("Move Dock"));

  if (area == Qt::LeftDockWidgetArea)
  {
    moveAction->setIcon(QIcon::fromTheme("right-arrow"));
  }
  else if (area == Qt::RightDockWidgetArea)
  {
    moveAction->setIcon(QIcon::fromTheme("left-arrow"));
  }

  m_moveToolButton = new QToolButton(this);
  m_moveToolButton->setDefaultAction(moveAction);
  m_moveToolButton->setAutoRaise(true);
  m_moveToolButton->setIconSize(QSize(8, 8));

  m_titleLabel = new QLabel(this);
  m_titleLabel->setAlignment(Qt::AlignCenter);
  m_titleLabel->setText("DOCK TITLE");

  QStyle* style = qApp->style();
  QIcon closeIcon = style->standardIcon(QStyle::SP_TitleBarCloseButton);

  QAction* closeAction = new QAction(this);
  closeAction->setIcon(closeIcon);
  closeAction->setCheckable(false);
  closeAction->setToolTip(tr("Close Dock"));

  m_closeToolButton = new QToolButton(this);
  m_closeToolButton->setDefaultAction(closeAction);
  m_closeToolButton->setIconSize(QSize(8, 8));
  m_closeToolButton->setAutoRaise(true);

  if (area == Qt::LeftDockWidgetArea)
  {
    frameLayout->addWidget(pinToolButton, 0, 0);
    frameLayout->addWidget(m_moveToolButton, 0, 1);
    frameLayout->addWidget(m_titleLabel, 0, 2);
    frameLayout->addWidget(m_closeToolButton, 0, 3);
  }
  else if (area == Qt::RightDockWidgetArea)
  {
    frameLayout->addWidget(m_closeToolButton, 0, 0);
    frameLayout->addWidget(m_titleLabel, 0, 1);
    frameLayout->addWidget(m_moveToolButton, 0, 2);
    frameLayout->addWidget(pinToolButton, 0, 3);
  }

  m_moveToolButton->setEnabled(false);
  m_titleLabel->setVisible(false);
  m_closeToolButton->setVisible(false);

  //connects
  connect(pinToolButton->defaultAction(), SIGNAL(triggered()), this, SIGNAL(pinTriggered()));
  connect(m_moveToolButton->defaultAction(), SIGNAL(triggered()), this, SIGNAL(moveTriggered()));
  connect(m_closeToolButton->defaultAction(), SIGNAL(triggered()), this, SIGNAL(closeTriggered()));

  this->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Preferred);

  m_minWidth = 35;
}

te::qt::af::ApplicationDockerTitleBar::~ApplicationDockerTitleBar() = default;

QSize te::qt::af::ApplicationDockerTitleBar::sizeHint() const
{
  return QSize(m_minWidth, 16);
}

QSize te::qt::af::ApplicationDockerTitleBar::minimumSizeHint() const
{
  return QSize(m_minWidth, 16);
}

void te::qt::af::ApplicationDockerTitleBar::setDockInfo(QString title, bool isCloseable, int minWidth)
{
  m_moveToolButton->setEnabled(true);

  m_titleLabel->setVisible(true);

  m_titleLabel->setText(title);

  m_closeToolButton->setVisible(isCloseable);

  m_minWidth = minWidth;

  this->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Preferred);
}

void te::qt::af::ApplicationDockerTitleBar::reset()
{
  m_moveToolButton->setEnabled(false);

  m_titleLabel->setVisible(false);

  m_closeToolButton->setVisible(false);

  this->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Preferred);

  m_minWidth = 35;
}
