/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

This file is part of the TerraLib - a Framework for building GIS enabled applications.

TerraLib is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

TerraLib is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with TerraLib. See COPYING. If not, write to
TerraLib Team at <terralib-team@terralib.org>.
*/

/*!
\file terralib/qt/af/ThirdPartyInfo.cpp

\brief A class used to get information about terralib apps dependencies.
*/

#include "ThirdPartyInfo.h"

// TerraLib Includes
#include "terralib_buildconfig.h"

// STL Includes
#include <algorithm>

// ThirdParty Includes
#ifdef TERRALIB_MOD_GDAL_ENABLED
#include <gdal_version.h>
#endif

#ifdef TERRALIB_BOOST_ENABLED
#include <boost/version.hpp>
#endif

#ifdef TERRALIB_GEOS_ENABLED
#include <geos/version.h>
#endif

#ifdef TERRALIB_QT_ENABLED
#include <qconfig.h>
#endif

#ifdef TERRALIB_PROJ4_ENABLED
  #include <proj.h>

  #ifndef PROJ_VERSION_MAJOR
    #error PROJ_VERSION_MAJOR not declared
  #endif

  #if PROJ_VERSION_MAJOR < 8
    //this must be added to allow the include of the file proj_api.h. This file is deprecated in proj 6 and will be removed in proj 7.
    #ifndef ACCEPT_USE_OF_DEPRECATED_PROJ_API_H
      #define ACCEPT_USE_OF_DEPRECATED_PROJ_API_H
    #endif

    #include <proj_api.h>
  #endif
#endif

#ifdef TERRALIB_QWT_ENABLED
#include <qwt_global.h>
#endif

#ifdef TERRALIB_MOD_GDAL_ENABLED
#include <sqlite3.h>
#endif

#ifdef TERRALIB_MOD_POSTGIS_ENABLED
#include <libpq-fe.h>
#endif

te::qt::af::ThirdPartyInfo::ThirdPartyInfo()
{

}

te::qt::af::ThirdPartyInfo::~ThirdPartyInfo() = default;

std::string te::qt::af::ThirdPartyInfo::getQtVersion()
{
  std::string version = "Not Defined.";

#ifdef TERRALIB_QT_ENABLED
  version = QT_VERSION_STR;
#endif

  return version;
}

std::string te::qt::af::ThirdPartyInfo::getGDALVersion()
{
  std::string version = "Not Defined.";

#ifdef TERRALIB_MOD_GDAL_ENABLED
  version = GDAL_RELEASE_NAME;
#endif

  return version;
}

std::string te::qt::af::ThirdPartyInfo::getBoostVersion()
{
  std::string version = "Not Defined.";

#ifdef TERRALIB_BOOST_ENABLED
  version = BOOST_LIB_VERSION;

  std::replace(version.begin(), version.end(), '_', '.');
#endif

  return version;
}

std::string te::qt::af::ThirdPartyInfo::getGeosVersion()
{
  std::string version = "Not Defined.";

#ifdef TERRALIB_GEOS_ENABLED
  version = GEOS_VERSION;
#endif

  return version;
}

std::string te::qt::af::ThirdPartyInfo::getProjVersion()
{
  std::string version = "Not Defined.";

#ifdef TERRALIB_PROJ4_ENABLED
  #ifndef PROJ_VERSION_MAJOR
    #error PROJ_VERSION_MAJOR not declared
  #endif

  #if PROJ_VERSION_MAJOR < 8
    version = std::to_string(PJ_VERSION);
  #else
    version = std::to_string(PROJ_VERSION_NUMBER);
  #endif
#endif

  return version;
}

std::string te::qt::af::ThirdPartyInfo::getQwtVersion()
{
  std::string version = "Not Defined.";

#ifdef TERRALIB_QWT_ENABLED
  version = QWT_VERSION_STR;
#endif

  return version;
}

std::string te::qt::af::ThirdPartyInfo::getSQLiteVersion()
{
  std::string version = "Not Defined.";

#ifdef TERRALIB_MOD_GDAL_ENABLED
  version = SQLITE_VERSION;
#endif

  return version;
}

std::string te::qt::af::ThirdPartyInfo::getPostgreSQLVersion()
{
  std::string version = "Not Defined";

#ifdef TERRALIB_MOD_POSTGIS_ENABLED
  version = std::to_string(PQlibVersion());
#endif

  return version;
}
