#include "../ApplicationController.h"
#include "../events/MapEvents.h"
#include "../events/ApplicationEvents.h"
#include "../Utils.h"
#include "../../widgets/canvas/MapDisplay.h"
#include "../../widgets/Utils.h"

#include "../../widgets/layer/explorer/LayerItemView.h"

#include "ui/ui_DisplayWidgetForm.h"
#include "DisplayWidget.h"

// Qt
#include <QColor>
#include <QColorDialog>
#include <QFrame>
#include <QPalette>
#include <QIntValidator>
#include <QApplication>
#include <QDesktopWidget>
#include <QRect>

// Boost
#include <boost/lexical_cast.hpp>

te::qt::af::DisplayWidget::DisplayWidget(QWidget* parent) :
AbstractSettingWidget(parent),
m_ui(new Ui::DisplayWidgetForm)
{
  m_ui->setupUi(this);

  m_resumeText = tr("Configurations of the display");

  QSettings sett(QSettings::IniFormat, QSettings::UserScope, qApp->organizationName(), qApp->applicationName());
  QString hexColor = sett.value("display/defaultDisplayColor").toString();
  QColor dColor;
  dColor.setNamedColor(hexColor);

  QPalette palette = m_ui->m_backColor->palette();

  if(!dColor.isValid())
  {
    palette.setColor(QPalette::Window, Qt::white);
  }
  else
  {
    palette.setColor(QPalette::Window, dColor);
  }

  m_ui->m_backColor->setPalette(palette);
  m_ui->m_backColor->setAutoFillBackground(true);

  QIntValidator* validatorWidth = new QIntValidator(this);
  m_ui->m_screenWidthLineEdit->setValidator(validatorWidth);

  QIntValidator* validatorHeight = new QIntValidator(this);
  m_ui->m_screenHeightLineEdit->setValidator(validatorHeight);

  bool isChecked = sett.value("display/autoRedraw").toBool();
  m_ui->m_autoRedrawChkBox->setChecked(isChecked);

  this->resetScreenSizeState();

  connect(m_ui->m_changeBackgroundPushButton, SIGNAL(clicked()), SLOT(onChangeBackgroundPushButton()));
  connect(m_ui->m_applyScreenSizePushButton, SIGNAL(clicked()), SLOT(onApplyScreenSizeButtonClicked()));
  connect(m_ui->m_restoreScreenSizePushButton, SIGNAL(clicked()), SLOT(onRestoreScreenSizeButtonClicked()));
  connect(m_ui->m_autoRedrawChkBox, SIGNAL(stateChanged(int)), SLOT(onAutoRedrawChkBoxStateChanged(int)));
}

te::qt::af::DisplayWidget::~DisplayWidget()
{
  delete m_ui;
}

void te::qt::af::DisplayWidget::onChangeBackgroundPushButton()
{
  QColor color = QColorDialog::getColor(m_ui->m_backColor->palette().color(QPalette::Window), this);

  if(!color.isValid())
    return;

  QPalette palette = m_ui->m_backColor->palette();
  palette.setColor(QPalette::Window, color);
  m_ui->m_backColor->setPalette(palette);
  m_ui->m_backColor->setAutoFillBackground(true);

  changeApplyButtonState(true);
}

void te::qt::af::DisplayWidget::onApplyScreenSizeButtonClicked()
{

  QString widthMMStr = m_ui->m_screenWidthLineEdit->text();
  QString heightMMStr = m_ui->m_screenHeightLineEdit->text();

  if(widthMMStr.isEmpty() || heightMMStr.isEmpty())
  {
    QMessageBox::warning(this,
                       tr("Display Options"),
                       tr("Fill width and height!"));
    return;
  }

  int widthMM = widthMMStr.toInt();
  int heightMM = heightMMStr.toInt();

  QRect screenRect = QApplication::desktop()->screenGeometry();
  int width = screenRect.width();
  int height = screenRect.height();

  int dpiX = (double)width/((double)widthMM/25.4);
  int dpiY = (double)height/((double)heightMM/25.4);

  te::qt::af::evt::GetMapDisplay getMapDisplay;

  m_app->trigger(&getMapDisplay);

  te::qt::widgets::MapDisplay* mapDisplay = getMapDisplay.m_display->getDisplay();

  if(mapDisplay == nullptr)
  {
    QMessageBox::warning(this,
                       tr("Display Options"),
                       tr("Map Display Configuration Failed!"));
    return;
  }

  mapDisplay->setOverrideDPI(dpiX, dpiY);

  te::qt::af::evt::GetLayerExplorer layerExEvent;

  m_app->trigger(&layerExEvent);

  std::list<te::map::AbstractLayerPtr> ls;

  te::qt::widgets::GetValidLayers(layerExEvent.m_layerExplorer->model(), QModelIndex(), ls);

  if(ls.empty())
    return;

  getMapDisplay.m_display->fit(ls);

  this->saveScreenSizeChanges();

  QMessageBox::information(this,
                           tr("Display Options"),
                           tr("Screen Size Configuration Applied!"));
}

void te::qt::af::DisplayWidget::onRestoreScreenSizeButtonClicked()
{

  QString widthMMStr = m_ui->m_screenWidthLineEdit->text();
  QString heightMMStr = m_ui->m_screenHeightLineEdit->text();

  if(widthMMStr.isEmpty() && heightMMStr.isEmpty())
  {
    return;
  }

  te::qt::af::evt::GetMapDisplay getMapDisplay;

  m_app->trigger(&getMapDisplay);

  te::qt::widgets::MapDisplay* mapDisplay = getMapDisplay.m_display->getDisplay();

  if(mapDisplay == nullptr)
  {
    QMessageBox::warning(this,
                       tr("Display Options"),
                       tr("Map Display Configuration Failed!"));
    return;
  }

  mapDisplay->setOverrideDPI(-1, -1);

  te::qt::af::evt::GetLayerExplorer layerExEvent;

  m_app->trigger(&layerExEvent);

  std::list<te::map::AbstractLayerPtr> ls;

  te::qt::widgets::GetValidLayers(layerExEvent.m_layerExplorer->model(), QModelIndex(), ls);

  m_ui->m_screenWidthLineEdit->clear();
  m_ui->m_screenHeightLineEdit->clear();


  this->saveScreenSizeChanges();

  QMessageBox::information(this,
                           tr("Display Options"),
                           tr("Default Screen Size Restored!"));

  if(ls.empty())
    return;

  getMapDisplay.m_display->fit(ls);

}

void te::qt::af::DisplayWidget::saveChanges()
{
  QSettings sett(QSettings::IniFormat, QSettings::UserScope, qApp->organizationName(), qApp->applicationName());

  sett.setValue("display/defaultDisplayColor", m_ui->m_backColor->palette().color(QPalette::Window).name());

  te::qt::af::evt::MapColorChanged mapColorChanged(m_ui->m_backColor->palette().color(QPalette::Window));

  m_app->trigger(&mapColorChanged);

  sett.setValue("display/autoRedraw", m_ui->m_autoRedrawChkBox->isChecked());

  changeApplyButtonState(false);
}

void te::qt::af::DisplayWidget::saveScreenSizeChanges()
{
  QSettings sett(QSettings::IniFormat, QSettings::UserScope, qApp->organizationName(), qApp->applicationName());

  sett.setValue("display/screenWidthMM", m_ui->m_screenWidthLineEdit->text());
  sett.setValue("display/screenHeightMM", m_ui->m_screenHeightLineEdit->text());
}

void te::qt::af::DisplayWidget::resetState()
{
  QSettings sett(QSettings::IniFormat, QSettings::UserScope, qApp->organizationName(), qApp->applicationName());
  QString hexColor = sett.value("display/defaultDisplayColor").toString();
  QColor dColor;
  dColor.setNamedColor(hexColor);

  QPalette palette = m_ui->m_backColor->palette();

  if(!dColor.isValid())
  {
    palette.setColor(QPalette::Window, Qt::white);
  }
  else
  {
    palette.setColor(QPalette::Window, dColor);
  }

  m_ui->m_backColor->setPalette(palette);
  m_ui->m_backColor->setAutoFillBackground(true);

  m_ui->m_autoRedrawChkBox->setChecked(false);
}

void te::qt::af::DisplayWidget::resetScreenSizeState()
{
  QSettings sett(QSettings::IniFormat, QSettings::UserScope, qApp->organizationName(), qApp->applicationName());

  QString widthMM = sett.value("display/screenWidthMM").toString();
  QString heightMM = sett.value("display/screenHeightMM").toString();

  m_ui->m_screenWidthLineEdit->setText(widthMM);
  m_ui->m_screenHeightLineEdit->setText(heightMM);
}

void te::qt::af::DisplayWidget::onApplyButtonClicked()
{
  saveChanges();
}

void te::qt::af::DisplayWidget::getHelpInformations(QString& ns, QString& helpFile)
{
  ns = "dpi.inpe.br.apf";
  helpFile = "apf/settings/display/DisplayConfig.html";
}

void te::qt::af::DisplayWidget::onAutoRedrawChkBoxStateChanged(int state)
{
  if (state == Qt::Checked)
  {

  }
  changeApplyButtonState(true);
}
