/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/plugins/datasource/ogr/OGRConnectorDialog.cpp

  \brief A dialog window for showing the OGR connector widget.
*/

// TerraLib
#include "../../../../core/translator/Translator.h"
#include "../../../../core/uri/URI.h"
#include "../../../../core/uri/Utils.h"
#include "../../../../dataaccess/datasource/DataSource.h"
#include "../../../../dataaccess/datasource/DataSourceFactory.h"
#include "../../../../dataaccess/datasource/DataSourceInfo.h"
#include "../../../../dataaccess/datasource/DataSourceManager.h"
#include "../../../af/Utils.h"
#include "../../../widgets/Exception.h"
#include "../../../widgets/Utils.h"
#include "../../../widgets/utils/FileDialog.h"
#include "OGRConnectorDialog.h"
#include "ui/ui_OGRConnectorDialogForm.h"
#include "Utils.h"

// Boost
#include <boost/uuid/random_generator.hpp>
#include <boost/uuid/uuid_io.hpp>
#include <boost/filesystem.hpp>
#include <boost/lexical_cast.hpp>

// Qt
#include <QFileInfo>
#include <QFileDialog>
#include <QMessageBox>

// STL
#include <cassert>

te::qt::plugins::ogr::OGRConnectorDialog::OGRConnectorDialog(QWidget* parent, Qt::WindowFlags f)
  : QDialog(parent, f),
    m_ui(new Ui::OGRConnectorDialogForm)
{
// add controls
  m_ui->setupUi(this);

// connect signal and slots
  connect(m_ui->m_openPushButton, SIGNAL(pressed()), this, SLOT(openPushButtonPressed()));
  connect(m_ui->m_testPushButton, SIGNAL(pressed()), this, SLOT(testPushButtonPressed()));
  connect(m_ui->m_searchFeatureToolButton, SIGNAL(pressed()), this, SLOT(searchFeatureToolButtonPressed()));

  m_ui->m_helpPushButton->setNameSpace("dpi.inpe.br.plugins");
  m_ui->m_helpPushButton->setPageReference("plugins/ogr/ogr.html");
}

te::qt::plugins::ogr::OGRConnectorDialog::~OGRConnectorDialog() = default;

const te::da::DataSourceInfoPtr& te::qt::plugins::ogr::OGRConnectorDialog::getDataSource() const
{
  return m_datasource;
}

const te::da::DataSourcePtr& te::qt::plugins::ogr::OGRConnectorDialog::getDriver() const
{
  return m_driver;
}

void te::qt::plugins::ogr::OGRConnectorDialog::set(const te::da::DataSourceInfoPtr& ds)
{
  m_datasource = ds;

  if(m_datasource.get() != nullptr)
  {
    setConnectionInfo(m_datasource->getConnInfoAsString());

    m_ui->m_datasourceTitleLineEdit->setText(QString::fromUtf8(m_datasource->getTitle().c_str()));

    m_ui->m_datasourceDescriptionTextEdit->setText(QString::fromUtf8(m_datasource->getDescription().c_str()));
  }
}

void te::qt::plugins::ogr::OGRConnectorDialog::openPushButtonPressed()
{
  try
  {
    // check if driver is loaded
    
    if(te::da::DataSourceFactory::find("OGR") == 0)
      throw te::qt::widgets::Exception(TE_TR("Sorry! No data access driver loaded for OGR data sources!"));

    // get data source connection info based on form data
    
    const std::string dsInfo = getConnectionInfo();

    // perform connection
    
    std::unique_ptr<te::da::DataSource> ds = te::da::DataSourceFactory::make("OGR", dsInfo);
    ds->open();
    m_driver.reset(ds.release());

    if(m_driver.get() == nullptr)
      throw te::qt::widgets::Exception(TE_TR("Could not open dataset via OGR due to an unknown error!"));

    QString title = m_ui->m_datasourceTitleLineEdit->text().trimmed();

    if(title.isEmpty())
      title = m_ui->m_featureRepoLineEdit->text();

    if(m_datasource.get() == nullptr)
    {
      // create a new data source based on form data
      
      m_datasource.reset(new te::da::DataSourceInfo);

      m_datasource->setConnInfo(dsInfo);

      boost::uuids::basic_random_generator<boost::mt19937> gen;
      boost::uuids::uuid u = gen();
      std::string dsId = boost::uuids::to_string(u);

      m_datasource->setId(dsId);
      m_driver->setId(dsId);
      m_datasource->setTitle(title.toUtf8().data());
      m_datasource->setDescription(m_ui->m_datasourceDescriptionTextEdit->toPlainText().toUtf8().data());
      m_datasource->setAccessDriver("OGR");
      m_datasource->setType("OGR");
    }
    else
    {
      m_driver->setId(m_datasource->getId());
      m_datasource->setConnInfo(dsInfo);
      m_datasource->setTitle(title.toUtf8().data());
      m_datasource->setDescription(m_ui->m_datasourceDescriptionTextEdit->toPlainText().trimmed().toUtf8().data());
    }
  }
  catch(const std::exception& e)
  {
    QMessageBox::warning(this,
                         tr("TerraLib Qt Components"),
                         tr(e.what()));
    return;
  }
  catch(...)
  {
    QMessageBox::warning(this,
                         tr("TerraLib Qt Components"),
                         tr("Unknown error while opening feature via OGR!"));
    return;
  }

  accept();
}

void te::qt::plugins::ogr::OGRConnectorDialog::testPushButtonPressed()
{
  try
  {
    // check if driver is loaded
    if(te::da::DataSourceFactory::find("OGR") == 0)
      throw te::qt::widgets::Exception(TE_TR("Sorry! No data access driver loaded for OGR data sources!"));

    // perform connection
    std::unique_ptr<te::da::DataSource> ds(te::da::DataSourceFactory::make("OGR", getConnectionInfo()));
    ds->open();

    if(ds.get() == nullptr)
      throw te::qt::widgets::Exception(TE_TR("Could not open feature repository via OGR!"));
    
    if (m_ui->m_dirRadioButton->isChecked() && ds->getNumberOfDataSets() <= 0)
      QMessageBox::information(this,
                               tr("TerraLib Qt Components"),
                               tr("Directory does not contain datasets!"));
    else
      QMessageBox::information(this,
                       tr("TerraLib Qt Components"),
                       tr("Data source is ok!"));
  }
  catch(const std::exception& e)
  {
    QMessageBox::warning(this,
                         tr("TerraLib Qt Components"),
                         tr(e.what()));
  }
  catch(...)
  {
    QMessageBox::warning(this,
                         tr("TerraLib Qt Components"),
                         tr("Unknown error while testing OGR data source!"));
  }
}

void te::qt::plugins::ogr::OGRConnectorDialog::searchFeatureToolButtonPressed()
{
  
  m_ui->m_featureRepoLineEdit->clear();
  
  if(m_ui->m_fileRadioButton->isChecked())
  {
    te::qt::widgets::FileDialog fileDialog(this, te::qt::widgets::FileDialog::VECTOR,
      "vector", te::qt::widgets::FileDialog::OPEN_FILE_MODE );

    try
    {
      fileDialog.exec();
    }
    catch(te::common::Exception& ex)
    {
      QMessageBox::warning(0, tr("File information"), ex.what());
      return;
    }
    
    std::string fileName = fileDialog.getPath();
    if(fileName.empty()) return;
    
    m_ui->m_featureRepoLineEdit->setText(fileName.c_str());

    QFileInfo info(fileName.c_str());

    te::qt::widgets::AddFilePathToSettings(info.absolutePath(), "vector");
  }
  else if(m_ui->m_dirRadioButton->isChecked())
  {
    QString dirName = QFileDialog::getExistingDirectory(this, tr("Select a directory with shape files"),
      te::qt::widgets::GetFilePathFromSettings("vector"), QFileDialog::ShowDirsOnly);

    if(dirName.isEmpty())
      return;

    te::qt::widgets::AddFilePathToSettings(dirName, "vector");

    m_ui->m_featureRepoLineEdit->setText(dirName);
  }
  else
  {
    QMessageBox::warning(this,
                         tr("TerraLib Qt Components"),
                         tr("Sorry, network files are not implemented yet!\nWe will provide it soon!"));
  }
  
  updateDataSetsList();
}

const std::string te::qt::plugins::ogr::OGRConnectorDialog::getConnectionInfo() const
{
  QString qstr; // Auxiliary string used to hold temporary data

  std::string strURI = "file://"; // The base of the URI

  qstr = m_ui->m_featureRepoLineEdit->text().trimmed();
  
  if(qstr.isEmpty())
    throw te::qt::widgets::Exception(TE_TR("Please select a feature file first!"));
  
  strURI += qstr.toUtf8().data();

  return strURI;
}

void te::qt::plugins::ogr::OGRConnectorDialog::setConnectionInfo(const std::string& connInfo)
{
  const te::core::URI uri(connInfo);
  std::string path = te::core::URIDecode(uri.host() + uri.path());

  if(!path.empty())
  {
    m_ui->m_featureRepoLineEdit->setText(QString::fromUtf8(path.c_str()));

    if(boost::filesystem::is_directory(path))
    {
      m_ui->m_dirRadioButton->setChecked(true);
    }
    else if(boost::filesystem::is_regular_file(path))
    {
      m_ui->m_fileRadioButton->setChecked(true);
    }
    else
    {
      QMessageBox::warning(this,
                           tr("TerraLib Qt Components"),
                           tr("Sorry, network files are not implemented yet!\nWe will provide it soon!"));
    }    
  }
}

void te::qt::plugins::ogr::OGRConnectorDialog::getSelectedDataSetNames ( 
    std::vector< std::string >& selectedDataSetNames ) const
{
  selectedDataSetNames.clear();
  
  const QList<QListWidgetItem *> selectedDataSetNamesItems = 
    m_ui->m_dataSetsListWidget->selectedItems();
    
  QList<QListWidgetItem *>::const_iterator it = selectedDataSetNamesItems.begin();
  const QList<QListWidgetItem *>::const_iterator itEnd = 
    selectedDataSetNamesItems.end();
    
  while( it != itEnd )
  {
    selectedDataSetNames.push_back( (*it)->text().toStdString() );
    
    ++it;
  }
}

void te::qt::plugins::ogr::OGRConnectorDialog::enableDataSetSelection( const bool enable )
{
  m_ui->m_dataSetsListWidget->setEnabled( enable );
}

void te::qt::plugins::ogr::OGRConnectorDialog::updateDataSetsList()
{
  m_ui->m_dataSetsListWidget->clear();
  
  if( !m_ui->m_featureRepoLineEdit->text().isEmpty() )
  {
    if(m_ui->m_fileRadioButton->isChecked())
    {
      try
      {
        // check if driver is loaded
        if(te::da::DataSourceFactory::find("OGR") == 0)
        {
          QMessageBox::critical(this,
                               tr("TerraLib Qt Components"),
                               tr("No data access driver loaded") );       
          return;
        }

        // get data source connection info based on form data
        const std::string dsInfo = "file://" + 
          m_ui->m_featureRepoLineEdit->text().toStdString();

        // perform connection
        
        std::unique_ptr<te::da::DataSource> ds = te::da::DataSourceFactory::make("OGR", dsInfo);
        ds->open();

        if(ds.get() == nullptr)
        {
          QMessageBox::critical(this,
                               tr("TerraLib Qt Components"),
                               tr("Could not open dataset via OGR due to an unknown error") );       
          return;
        }

        const std::vector<std::string> dataSetNames = ds->getDataSetNames();
        
        for( std::size_t dataSetNamesIdx = 0 ; dataSetNamesIdx < dataSetNames.size() ;
          ++dataSetNamesIdx )
        {
          m_ui->m_dataSetsListWidget->addItem( dataSetNames[ dataSetNamesIdx ].c_str() );
        }
        
        m_ui->m_dataSetsListWidget->selectAll();
      }
      catch(const std::exception& e)
      {
        QMessageBox::warning(this,
                             tr("TerraLib Qt Components"),
                             tr(e.what()));
        return;
      }
      catch(...)
      {
        QMessageBox::warning(this,
                             tr("TerraLib Qt Components"),
                             tr("Unknown error while opening feature via GPKG!"));
        return;
      }        
    }
    else if(m_ui->m_dirRadioButton->isChecked())
    {
      // nothing to be done
    }
    else
    {
      QMessageBox::warning(this,
                           tr("TerraLib Qt Components"),
                           tr("Sorry, network files are not implemented yet!\nWe will provide it soon!"));
    }
  }  
}