/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/plugins/datasource/ogr/Plugin.cpp

  \brief Plugin implementation for the OGR data source widget.
*/

// TerraLib
#include "../../../../common/Config.h"
#include "../../../../core/logger/Logger.h"
#include "../../../../core/translator/Translator.h"
#include "../../../../srs/Config.h"
#include "../../../../dataaccess/dataset/DataSetType.h"
#include "../../../../dataaccess/datasource/DataSourceInfoManager.h"
#include "../../../../dataaccess/datasource/DataSourceManager.h"
#include "../../../../dataaccess/utils/Utils.h"
#include "../../../../maptools/AbstractLayer.h"
#include "../../../widgets/canvas/MultiThreadMapDisplay.h"
#include "../../../widgets/datasource/core/DataSourceTypeManager.h"
#include "../../../widgets/layer/explorer/LayerItemView.h"
#include "../../../widgets/Utils.h"
#include "../../../widgets/utils/FileDialog.h"
#include "../../../af/ApplicationController.h"
#include "../../../af/BaseApplication.h"
#include "../../../af/Utils.h"
#include "../../../af/events/ApplicationEvents.h"
#include "GeoFileDragAndDropHandler.h"
#include "OGRType.h"
#include "Plugin.h"
#include "Utils.h"
#include "OGRConnectorDialog.h"

// OGR
#include <ogrsf_frmts.h>

// Qt 
#include <QFileInfo>
#include <QAction>
#include <QApplication>
#include <QFileDialog>
#include <QMenu>
#include <QMessageBox>
#include <QToolBar>

QString GetFileExtensionName(const std::string& type)
{
  if(type.compare("ESRI Shapefile") == 0)
    return QObject::tr("ESRI Shapefile (*.shp *.SHP)");
  else if(type.compare("MapInfo File") == 0)
    return QObject::tr("MapInfo File (*.mif *.MIF)");
  else if(type.compare("DGN") == 0)
    return QObject::tr("DGN File (*.dgn *.DGN)");
  else if(type.compare("CSV") == 0)
    return QObject::tr("CSV File (*.csv *.CSV)");
  else if(type.compare("GML") == 0)
    return QObject::tr("GML File (*.gml *.GML)");
  else if(type.compare("KML") == 0)
    return QObject::tr("KML File (*.kml *.KML)");
  else if(type.compare("GeoJSON") == 0)
    return QObject::tr("GeoJSON File (*.geojson *.GEOJSON)");
  else if(type.compare("DXF") == 0)
    return QObject::tr("DXF File (*.dxf *.DXF)");

  return "";
}

QString GetSupportedFiles()
{
  QString res;
  GDALDriverManager mger;

  int count = mger.GetDriverCount();

  for(int i=0; i<count; i++)
  {
    std::string dName = mger.GetDriver(i)->GetDescription();

    QString drv = GetFileExtensionName(dName);

    if(!drv.isEmpty())
      res += drv + ";;";
  }

  return res;
}

te::qt::plugins::ogr::Plugin::Plugin(const te::core::PluginInfo& pluginInfo)
    :

      te::core::CppPlugin(pluginInfo),
      m_showWindow(nullptr),
      m_handlerLayerExplorer(nullptr),
      m_handlerMapDisplay(nullptr)
{
  te::qt::af::AppCtrlSingleton::getInstance().addListener(this, te::qt::af::SENDER);
}

te::qt::plugins::ogr::Plugin::~Plugin() = default;

void te::qt::plugins::ogr::Plugin::startup()
{
  if(m_initialized)
    return;

  te::qt::widgets::DataSourceTypeManager::getInstance().add(new OGRType);

  TE_LOG_TRACE(TE_TR("TerraLib Qt OGR widget startup!"));

  m_initialized = true;
  
  //  //Initializing action
  //  QAction* act = te::qt::af::AppCtrlSingleton::getInstance().findAction("Project.Add Layer.Tabular File");
  //  QMenu* mnu = te::qt::af::AppCtrlSingleton::getInstance().findMenu("Project.Add Layer");

  //  if(act != 0 && mnu != 0)
  {
    //    QWidget* parent = act->parentWidget();
    m_showWindow = new QAction(QIcon::fromTheme("file-vector"), tr("Vector File..."), this);
    m_showWindow->setToolTip(tr("Add new vector file as a layer."));
    m_showWindow->setObjectName("Project.Add Layer.Vector File");

    te::qt::af::evt::NewActionsAvailable e;
    e.m_category = "Dataaccess";
    e.m_actions << m_showWindow;

    emit triggered(&e);

    connect (m_showWindow, SIGNAL(triggered()), SLOT(showWindow()));

    //register actions into application tool bar
    QToolBar* toolBar = te::qt::af::AppCtrlSingleton::getInstance().getToolBar("File Tool Bar");

    if (toolBar)
      toolBar->addAction(m_showWindow);

    //install handler
    m_handlerLayerExplorer = new GeoFileDragAndDropHandler(this);
    m_handlerMapDisplay = new GeoFileDragAndDropHandler(this);

    te::qt::af::AppCtrlSingleton::getInstance().addListener(m_handlerMapDisplay, te::qt::af::SENDER);

    te::qt::af::BaseApplication* ba = dynamic_cast<te::qt::af::BaseApplication*>(te::qt::af::AppCtrlSingleton::getInstance().getMainWindow());
    if(ba != nullptr)
    {
      ba->getLayerExplorer()->installEventFilter(m_handlerLayerExplorer);
      ba->getMapDisplay()->installEventFilter(m_handlerMapDisplay);
    }
  }
}

void te::qt::plugins::ogr::Plugin::shutdown()
{
  if(!m_initialized)
    return;

  //remove event handler
  te::qt::af::AppCtrlSingleton::getInstance().removeListener(m_handlerMapDisplay);

  te::qt::af::BaseApplication* ba = dynamic_cast<te::qt::af::BaseApplication*>(te::qt::af::AppCtrlSingleton::getInstance().getMainWindow());

  if (ba)
  {
    ba->getLayerExplorer()->removeEventFilter(m_handlerLayerExplorer);
    ba->getMapDisplay()->removeEventFilter(m_handlerMapDisplay);
  }

  te::da::DataSourceInfoManager::getInstance().removeByType("OGR");
  te::qt::widgets::DataSourceTypeManager::getInstance().remove("OGR");

  TE_LOG_TRACE(TE_TR("TerraLib Qt OGR widget shutdown!"));

  //  delete m_showWindow;

  m_initialized = false;
}

void te::qt::plugins::ogr::Plugin::showWindow()
{
  OGRConnectorDialog diag( nullptr );
  diag.enableDataSetSelection( true );

  if(diag.exec() == QDialog::Rejected)
  {
    return;
  }
  
  std::vector< std::string > selectedDataSetNames;
  diag.getSelectedDataSetNames ( selectedDataSetNames );
  if( selectedDataSetNames.empty() )
  {
    return;
  }
  
  te::da::DataSourceInfoPtr dataSourceInfoPtr = diag.getDataSource();
  if( dataSourceInfoPtr.get() == nullptr )
  {
    return;
  }
  
  std::list<te::map::AbstractLayerPtr> layers;
  te::qt::plugins::ogr::CreateSelectedLayers( dataSourceInfoPtr, 
    selectedDataSetNames, layers );
}

TERRALIB_PLUGIN_CALL_BACK_IMPL(te::qt::plugins::ogr::Plugin)