/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/plugins/datasource/pgisRaster/Plugin.cpp

  \brief Plugin implementation for the PostGIS Raster data source widget.
*/

// TerraLib
#include "../../../../common/Config.h"
#include "../../../../core/translator/Translator.h"
#include "../../../../core/logger/Logger.h"
#include "../../../../dataaccess/datasource/DataSourceInfoManager.h"
#include "../../../widgets/datasource/core/DataSourceTypeManager.h"
#include "../../../widgets/datasource/selector/DataSourceSelectorDialog.h"
#include "../../../widgets/datasource/selector/DataSourceSelectorWidget.h"
#include "../../../af/ApplicationController.h"
#include "../../../af/events/ApplicationEvents.h"
#include "PostGISDataSetSelectorDialog.h"
#include "PostGISType.h"
#include "Plugin.h"
#include "Utils.h"

// Qt 
#include <QAction>
#include <QMessageBox>
#include <QToolBar>

te::qt::plugins::pgisRaster::Plugin::Plugin(
    const te::core::PluginInfo& pluginInfo)
    :

      te::core::CppPlugin(pluginInfo),
      m_openPostGISRasterSelector(nullptr)
{
  te::qt::af::AppCtrlSingleton::getInstance().addListener(this, te::qt::af::SENDER);
}

te::qt::plugins::pgisRaster::Plugin::~Plugin() = default;

void te::qt::plugins::pgisRaster::Plugin::startup()
{
  if(m_initialized)
    return;

  TE_LOG_TRACE(TE_TR("TerraLib Qt PostGIS widget startup!"));

  m_initialized = true;

  //Initializing action
  m_openPostGISRasterSelector = new QAction(QIcon::fromTheme("datasource-postgis"), tr("PostGIS Raster..."), this);
  m_openPostGISRasterSelector->setToolTip(tr("Add new PostGIS Raster as a layer."));
  m_openPostGISRasterSelector->setObjectName("Project.Add Layer.PostGIS Raster");

  te::qt::af::evt::NewActionsAvailable e;
  e.m_category = "Dataaccess";
  e.m_actions << m_openPostGISRasterSelector;

  emit triggered(&e);

  connect(m_openPostGISRasterSelector, SIGNAL(triggered()), SLOT(openDataSourceSelector()));

  //register actions into application tool bar
  QToolBar* toolBar = te::qt::af::AppCtrlSingleton::getInstance().getToolBar("File Tool Bar");

  if (toolBar)
    toolBar->addAction(m_openPostGISRasterSelector);

}

void te::qt::plugins::pgisRaster::Plugin::shutdown()
{
  if(!m_initialized)
    return;

  te::da::DataSourceInfoManager::getInstance().removeByType("POSTGISRASTER");
  te::da::DataSourceInfoManager::getInstance().removeByType("GDALPOSTGISRASTER");

  TE_LOG_TRACE(TE_TR("TerraLib Qt PostGIS Raster widget shutdown!"));

  m_initialized = false;
}

void te::qt::plugins::pgisRaster::Plugin::openDataSourceSelector()
{
  te::qt::widgets::DataSourceTypeManager::getInstance().add(new PostGISType);

  std::unique_ptr<te::qt::widgets::DataSourceSelectorDialog> dselector(new te::qt::widgets::DataSourceSelectorDialog(te::qt::af::AppCtrlSingleton::getInstance().getMainWindow()));

  dselector->getSelectorWidget()->showSpecificDataSource("POSTGISRASTER");

  if (dselector->exec() == QDialog::Accepted)
  {
    std::list<te::da::DataSourceInfoPtr> selectedDatasources = dselector->getSelecteds();

    if (!selectedDatasources.empty())
    {
      te::da::DataSourceInfoPtr dsInfo = selectedDatasources.front();

      std::unique_ptr<PostGISDataSetSelectorDialog> selector(new PostGISDataSetSelectorDialog(te::qt::af::AppCtrlSingleton::getInstance().getMainWindow()));

      selector->set(dsInfo);

      if (selector->exec() == QDialog::Accepted)
      {
        PostGISRasterInfoVec vec;
        selector->getSelecteds(vec);

        if (vec.empty())
        {
          QMessageBox::warning(te::qt::af::AppCtrlSingleton::getInstance().getMainWindow(),
            tr("PostGIS Raster"), tr("No PostGIS Raster selected."));
        }
        else
        {
          CreateLayers(vec);
        }
      }
      
    }
  }

  te::qt::widgets::DataSourceTypeManager::getInstance().remove("POSTGISRASTER");
}

TERRALIB_PLUGIN_CALL_BACK_IMPL(te::qt::plugins::pgisRaster::Plugin)
