/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
\file terralib/qt/plugins/datasource/pgis/PostGISDataSetSelectorDialog.cpp

\brief A dialog window for showing the PostGIS data set selector widget.
*/

// TerraLib
#include "../../../../dataaccess/datasource/DataSourceFactory.h"
#include "../../../../dataaccess/datasource/DataSourceManager.h"
#include "../../../../dataaccess/utils/Utils.h"
#include "../../../../qt/widgets/datasource/core/DataSourceType.h"
#include "../../../../qt/widgets/datasource/core/DataSourceTypeManager.h"
#include "../../../../qt/widgets/mapdisplay/DataSetDisplay.h"
#include "PostGISDataSetSelectorDialog.h"
#include "ui/ui_PostGISDataSetSelectorDialogForm.h"

// Qt
#include <QMessageBox>
#include <QTreeWidget>
#include <QTreeWidgetItem>

te::qt::plugins::pgisRaster::PostGISDataSetSelectorDialog::PostGISDataSetSelectorDialog(QWidget* parent, Qt::WindowFlags f)
  : QDialog(parent, f),
  m_ui(new Ui::PostGISDataSetSelectorDialogForm),
  m_rootItem(nullptr),
  m_propertyBrowser(nullptr),
  m_rasterPropertyInfo(nullptr),
  m_timer(new QTimer(this))
{
// add controls
  m_ui->setupUi(this);

// init components

  m_mapPreview.reset(new te::qt::widgets::DataSetDisplay(m_ui->m_mapPreviewGroupBox));
  QVBoxLayout* mapPreviewGroupBoxLayout = new QVBoxLayout(m_ui->m_mapPreviewGroupBox);
  mapPreviewGroupBoxLayout->addWidget(m_mapPreview.get(), 1);
  m_mapPreview->show();

  m_ui->m_mapPreviewGroupBox->setChecked(true);

  m_propertyLayout = new QGridLayout(m_ui->m_dataPreviewGroupBox);

  // connect signals and slots
  connect(m_ui->m_mapPreviewGroupBox, SIGNAL(toggled(bool)), this, SLOT(onMapPreviewToggled(bool)));
  connect(m_ui->m_treeWidget, SIGNAL(itemPressed(QTreeWidgetItem*, int)), this, SLOT(onItemPressed(QTreeWidgetItem*, int)));

  m_timer->setSingleShot(true);
  connect(m_timer, SIGNAL(timeout()), this, SLOT(onResizeTimeout()));

  m_ui->m_helpPushButton->setPageReference("widgets/dataset_selector/dataset_selector.html");
}

te::qt::plugins::pgisRaster::PostGISDataSetSelectorDialog::~PostGISDataSetSelectorDialog()
{
  delete m_propertyBrowser;
  delete m_rasterPropertyInfo;
}

void te::qt::plugins::pgisRaster::PostGISDataSetSelectorDialog::set(const te::da::DataSourceInfoPtr& ds)
{
  m_datasource = ds;

  if (ds.get() == nullptr)
    return;

  //set interface labels
  QString title(tr("Dataset Selection - %1"));
  title = title.arg(QString::fromUtf8(ds->getTitle().c_str()));

  const te::qt::widgets::DataSourceType* dsType = te::qt::widgets::DataSourceTypeManager::getInstance().get(ds->getAccessDriver());

  if(dsType == nullptr)
    return;

  title = tr("%1 - %2");
  title = title.arg(QString::fromUtf8(dsType->getTitle().c_str())).arg(QString::fromUtf8(ds->getTitle().c_str()));

  m_ui->m_titleLabel->setText(title);

  //list raster datasets
  listDataSets();
}

void te::qt::plugins::pgisRaster::PostGISDataSetSelectorDialog::getSelecteds(PostGISRasterInfoVec& vec)
{
  if (!m_rootItem)
    return;

  for (int i = 0; i < m_rootItem->childCount(); ++i)
  {
    QTreeWidgetItem* item = m_rootItem->child(i);

    if (item->checkState(0) == Qt::Checked)
    {
      PostGISRasterInfo info;
      info.m_tableName = item->text(0).toUtf8().data();
      info.m_schemaName = item->text(1).toUtf8().data();
      info.m_uri = getGDALURI(info.m_tableName, info.m_schemaName);

      vec.push_back(info);
    }
  }
}

void te::qt::plugins::pgisRaster::PostGISDataSetSelectorDialog::listDataSets()
{
  //clear and set root item
  m_ui->m_treeWidget->clear();
    
  m_rootItem = new QTreeWidgetItem(m_ui->m_treeWidget, ROOT_ITEM);
  m_rootItem->setText(0, tr("Rasters"));
  m_rootItem->setExpanded(true);

  m_ui->m_treeWidget->addTopLevelItem(m_rootItem);

  //get data source
  te::da::DataSourcePtr ds = te::da::DataSourceManager::getInstance().get(m_datasource->getId(), m_datasource->getAccessDriver(), m_datasource->getConnInfo());

  if (!ds.get())
  {
    QMessageBox::warning(this, tr("PostGIS Raster"), tr("Error openning data source."));
    return;
  }

  if (!ds->isOpened())
    ds->open();

  
  //get list
  std::unique_ptr<te::da::DataSet> dataSet;
  
  try
  {
    dataSet = ds->query("SELECT r_table_name, r_table_schema FROM raster_columns");
  }
  catch (...)
  {
    QMessageBox::warning(this, tr("PostGIS Raster"), tr("Error getting data sets."));
    return;
  }

  if (!dataSet.get())
  {
    QMessageBox::warning(this, tr("PostGIS Raster"), tr("Error getting data sets."));
    return;
  }

  dataSet->moveBeforeFirst();

  //set items
  while (dataSet->moveNext())
  {
    std::string name = dataSet->getString(0);
    std::string schema = dataSet->getString(1);

    QTreeWidgetItem* item = new QTreeWidgetItem(m_rootItem, TABLE_ITEM);

    item->setFlags(Qt::ItemIsEnabled |Qt::ItemIsUserCheckable | Qt::ItemIsSelectable);
    item->setText(0, name.c_str());
    item->setText(1, schema.c_str());
    item->setCheckState(0, Qt::Unchecked);

    m_rootItem->addChild(item);
  }
}

void te::qt::plugins::pgisRaster::PostGISDataSetSelectorDialog::drawDataSet(std::string dataSetName, std::string schema)
{
  if (!m_ui->m_mapPreviewGroupBox->isChecked())
    return;

  //build GDAL URI
  std::map<std::string, std::string> rinfo;

  std::string uri = getGDALURI(dataSetName, schema);

  rinfo["URI"] = uri;

  //get raster
  te::da::DataSourcePtr ds = te::da::DataSourceFactory::make("GDAL", uri);
  ds->open();
  ds->isValid();

  std::vector<std::string> dsNames = ds->getDataSetNames();

  std::unique_ptr<te::da::DataSet> dataSet = ds->getDataSet(dsNames[0]);

  te::da::DataSetTypePtr dataSetType = ds->getDataSetType(dsNames[0]);

  //draw raster
  m_mapPreview->draw(dataSetType, ds, dataSet.get());
}

void te::qt::plugins::pgisRaster::PostGISDataSetSelectorDialog::getDataSetInfo(std::string dataSetName, std::string schema)
{
  std::map<std::string, std::string> rinfo;

  std::string uri = getGDALURI(dataSetName, schema);

  rinfo["URI"] = uri;

  std::unique_ptr<te::da::DataSource> ds = te::da::DataSourceFactory::make("GDAL", uri);
  ds->open();
  ds->isValid();

  std::vector<std::string> dsNames = ds->getDataSetNames();

  std::unique_ptr<te::da::DataSet> dataSet = ds->getDataSet(dsNames[0]);

  std::size_t rpos = te::da::GetFirstPropertyPos(dataSet.get(), te::dt::RASTER_TYPE);

  std::unique_ptr<te::rst::Raster> inraster = dataSet->getRaster(rpos);

  buildDataSetInfoProperties(inraster.get());
}

void te::qt::plugins::pgisRaster::PostGISDataSetSelectorDialog::buildDataSetInfoProperties(te::rst::Raster* raster)
{
  if (m_propertyBrowser)
    delete m_propertyBrowser;

  if (m_rasterPropertyInfo)
    delete m_rasterPropertyInfo;

  m_propertyBrowser = new QtTreePropertyBrowser(this);
  m_propertyBrowser->setPropertiesWithoutValueMarked(true);
  m_propertyBrowser->setResizeMode(QtTreePropertyBrowser::ResizeToContents);

  m_rasterPropertyInfo = new RasterPropertiesInfo(m_propertyBrowser, raster);

  m_propertyLayout->addWidget(m_propertyBrowser);
}

void te::qt::plugins::pgisRaster::PostGISDataSetSelectorDialog::resizeEvent(QResizeEvent* e)
{
  QWidget::resizeEvent(e);

  m_timer->start(200);
}

std::string te::qt::plugins::pgisRaster::PostGISDataSetSelectorDialog::getGDALURI(std::string dataSetName, std::string schema)
{
  //"PG://postgres:tdk696@localhost:5432/raster_db?schema=public&table=cbers&mode=2";

  std::string uri = "PG://";
  uri += m_datasource->getConnInfo().user();
  uri += ":";
  uri += m_datasource->getConnInfo().password();
  uri += "@";
  uri += m_datasource->getConnInfo().host();
  uri += ":";
  uri += m_datasource->getConnInfo().port();
  uri += m_datasource->getConnInfo().path();
  uri += "?";
  uri += "schema=";
  uri += schema;
  uri += "&table=";
  uri += dataSetName;
  uri += "&mode=2";

  return uri;
}

void te::qt::plugins::pgisRaster::PostGISDataSetSelectorDialog::onMapPreviewToggled(bool on)
{
  if (on)
  {
    QList<QTreeWidgetItem*> list = m_ui->m_treeWidget->selectedItems();

    if (!list.isEmpty())
    {
      QTreeWidgetItem* item = list.first();

      onItemPressed(item, 0);
    }
  }
  else
  {
    m_mapPreview->clear();
  }
}

void te::qt::plugins::pgisRaster::PostGISDataSetSelectorDialog::onItemPressed(
    QTreeWidgetItem* item, int /*column*/)
{
  if (!item || item->type() == ROOT_ITEM)
    return;

  std::string dataSetName = item->text(0).toUtf8().data();
  std::string schemaName = item->text(1).toUtf8().data();

  getDataSetInfo(dataSetName, schemaName);

  drawDataSet(dataSetName, schemaName);
}

void te::qt::plugins::pgisRaster::PostGISDataSetSelectorDialog::onResizeTimeout()
{
  onMapPreviewToggled(m_ui->m_mapPreviewGroupBox->isChecked());
}

