/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
\file terralib/qt/plugins/datasource/pgis/RasterPropertiesInfo.cpp

\brief A QtPropertyBrowser component to show raster properties.
*/

#include "../../../../datatype/Utils.h"
#include "../../../../raster/RasterSummary.h"
#include "../../../../raster/RasterSummaryManager.h"
#include "../../../../raster/Band.h"
#include "../../../../raster/BandProperty.h"
#include "../../../../raster/Utils.h"
#include "../../../../srs/SpatialReferenceSystemManager.h"
#include "../../../../qt/widgets/propertybrowser/AbstractPropertyManager.h"

#include "RasterPropertiesInfo.h"

te::qt::plugins::pgisRaster::RasterPropertiesInfo::RasterPropertiesInfo(QtTreePropertyBrowser* pb, te::rst::Raster* raster) :
  te::qt::widgets::AbstractPropertyItem(pb)
{
  QtProperty* rasterInfo_prop = te::qt::widgets::AbstractPropertyManager::getInstance().m_groupManager->addProperty(tr("Raster properties"));

  addProperty(rasterInfo_prop, QColor(255, 255, 175));

  //SRID Property
  QtProperty* srid_prop = te::qt::widgets::AbstractPropertyManager::getInstance().m_stringManager->addProperty(tr("SRID"));
  rasterInfo_prop->addSubProperty(srid_prop);
  srid_prop->setPropertyName("srid");
  srid_prop->setEnabled(false);

 //SRID setting value
  QString sridStr = QString::number(raster->getSRID());
  sridStr += QObject::tr(" -  ");
  sridStr += QString(te::srs::SpatialReferenceSystemManager::getInstance().getName(raster->getSRID()).c_str());
  te::qt::widgets::AbstractPropertyManager::getInstance().m_stringManager->setValue(srid_prop, sridStr);

  //add properties
  QtProperty* nLinesProp = te::qt::widgets::AbstractPropertyManager::getInstance().m_intManager->addProperty(tr("Number of Lines"));
  QtProperty* nColumnsProp = te::qt::widgets::AbstractPropertyManager::getInstance().m_intManager->addProperty(tr("Number of Columns"));
  QtProperty* nBandsProp = te::qt::widgets::AbstractPropertyManager::getInstance().m_intManager->addProperty(tr("Number of Bands"));
  QtProperty* nResXProp = te::qt::widgets::AbstractPropertyManager::getInstance().m_doubleManager->addProperty(tr("Resolution X"));
  QtProperty* nResYProp = te::qt::widgets::AbstractPropertyManager::getInstance().m_doubleManager->addProperty(tr("Resolution Y"));
  QtProperty* nAccessProp = te::qt::widgets::AbstractPropertyManager::getInstance().m_stringManager->addProperty(tr("Access Policy"));
  QtProperty* nMultiResProp = te::qt::widgets::AbstractPropertyManager::getInstance().m_intManager->addProperty(tr("Multi Resolution Levels"));

  //add properties to parent property
  rasterInfo_prop->addSubProperty(nLinesProp);
  rasterInfo_prop->addSubProperty(nColumnsProp);
  rasterInfo_prop->addSubProperty(nBandsProp);
  rasterInfo_prop->addSubProperty(nResXProp);
  rasterInfo_prop->addSubProperty(nResYProp);
  rasterInfo_prop->addSubProperty(nAccessProp);
  rasterInfo_prop->addSubProperty(nMultiResProp);

  //set the property state
  nLinesProp->setEnabled(false);
  nColumnsProp->setEnabled(false);
  nBandsProp->setEnabled(false);
  nResXProp->setEnabled(false);
  nResYProp->setEnabled(false);
  nAccessProp->setEnabled(false);
  nMultiResProp->setEnabled(false);

  te::common::UnitOfMeasurePtr unitPtr = te::srs::SpatialReferenceSystemManager::getInstance().getUnit(raster->getSRID());

  if (unitPtr.get())
  {
    if (unitPtr->getId() == te::common::UOM_Degree)
    {
      te::qt::widgets::AbstractPropertyManager::getInstance().m_doubleManager->setDecimals(nResXProp, 6);
      te::qt::widgets::AbstractPropertyManager::getInstance().m_doubleManager->setDecimals(nResYProp, 6);
    }
    else if (unitPtr->getId() == te::common::UOM_Metre)
    {
      te::qt::widgets::AbstractPropertyManager::getInstance().m_doubleManager->setDecimals(nResXProp, 2);
      te::qt::widgets::AbstractPropertyManager::getInstance().m_doubleManager->setDecimals(nResYProp, 2);
    }
  }

  //set properties values
  te::qt::widgets::AbstractPropertyManager::getInstance().m_intManager->setValue(nLinesProp, (int)raster->getNumberOfRows());
  te::qt::widgets::AbstractPropertyManager::getInstance().m_intManager->setValue(nColumnsProp, (int)raster->getNumberOfColumns());
  te::qt::widgets::AbstractPropertyManager::getInstance().m_intManager->setValue(nBandsProp, (int)raster->getNumberOfBands());
  te::qt::widgets::AbstractPropertyManager::getInstance().m_doubleManager->setValue(nResXProp, raster->getResolutionX());
  te::qt::widgets::AbstractPropertyManager::getInstance().m_doubleManager->setValue(nResYProp, raster->getResolutionY());
  te::qt::widgets::AbstractPropertyManager::getInstance().m_intManager->setValue(nMultiResProp, (int)raster->getMultiResLevelsCount());


  if (raster->getAccessPolicy() == te::common::NoAccess)
    te::qt::widgets::AbstractPropertyManager::getInstance().m_stringManager->setValue(nAccessProp, tr("No Access"));
  else if (raster->getAccessPolicy() == te::common::RAccess)
    te::qt::widgets::AbstractPropertyManager::getInstance().m_stringManager->setValue(nAccessProp, tr("Read Only Access"));
  else if (raster->getAccessPolicy() == te::common::WAccess)
    te::qt::widgets::AbstractPropertyManager::getInstance().m_stringManager->setValue(nAccessProp, tr("Write Only Access"));
  else if (raster->getAccessPolicy() == te::common::RWAccess)
    te::qt::widgets::AbstractPropertyManager::getInstance().m_stringManager->setValue(nAccessProp, tr("Read and Write Access"));

  //set information for each band
  for (std::size_t t = 0; t < raster->getNumberOfBands(); ++t)
  {
    //get band property
    te::rst::Band* band = raster->getBand(t);

    //create band top property
    QtProperty* bandProp = te::qt::widgets::AbstractPropertyManager::getInstance().m_groupManager->addProperty(tr("Band ") + QString::number(t));

    //create band properties
    QtProperty* descriptionProp = te::qt::widgets::AbstractPropertyManager::getInstance().m_stringManager->addProperty(tr("Description"));
    QtProperty* dataTypeProp = te::qt::widgets::AbstractPropertyManager::getInstance().m_stringManager->addProperty(tr("Data Type"));
    QtProperty* dummyValueProp = te::qt::widgets::AbstractPropertyManager::getInstance().m_doubleManager->addProperty(tr("Dummy Value"));
    QtProperty* colorInterpProp = te::qt::widgets::AbstractPropertyManager::getInstance().m_stringManager->addProperty(tr("Color Interpretation"));
    QtProperty* palleteInterpProp = te::qt::widgets::AbstractPropertyManager::getInstance().m_stringManager->addProperty(tr("Pallete Interpretation"));

    //add band properties to band top property
    bandProp->addSubProperty(descriptionProp);
    bandProp->addSubProperty(dataTypeProp);
    bandProp->addSubProperty(dummyValueProp);
    bandProp->addSubProperty(colorInterpProp);
    bandProp->addSubProperty(palleteInterpProp);

    //set the band properties state
    descriptionProp->setEnabled(false);
    dataTypeProp->setEnabled(false);
    dummyValueProp->setEnabled(false);
    colorInterpProp->setEnabled(false);
    palleteInterpProp->setEnabled(false);

    //set the band properties values
    te::qt::widgets::AbstractPropertyManager::getInstance().m_stringManager->setValue(descriptionProp, band->getProperty()->m_description.c_str());
    te::qt::widgets::AbstractPropertyManager::getInstance().m_stringManager->setValue(dataTypeProp, te::dt::ConvertDataTypeToString(band->getProperty()->getType()).c_str());
    te::qt::widgets::AbstractPropertyManager::getInstance().m_doubleManager->setValue(dummyValueProp, band->getProperty()->m_noDataValue);
    te::qt::widgets::AbstractPropertyManager::getInstance().m_stringManager->setValue(colorInterpProp, te::rst::ConvertColorInterpTypeToString(band->getProperty()->m_colorInterp).c_str());
    te::qt::widgets::AbstractPropertyManager::getInstance().m_stringManager->setValue(palleteInterpProp, te::rst::ConvertPalleteInterpTypeToString(band->getProperty()->m_paletteInterp).c_str());

    //add block information
    QtProperty* blockProp = te::qt::widgets::AbstractPropertyManager::getInstance().m_groupManager->addProperty(tr("Block properties"));

    QtProperty* nBlockXProp = te::qt::widgets::AbstractPropertyManager::getInstance().m_intManager->addProperty(tr("Number of blocks in X"));
    QtProperty* nBlockYProp = te::qt::widgets::AbstractPropertyManager::getInstance().m_intManager->addProperty(tr("Number of blocks in Y"));
    QtProperty* blockWProp = te::qt::widgets::AbstractPropertyManager::getInstance().m_intManager->addProperty(tr("Block Width (pixels)"));
    QtProperty* blockHProp = te::qt::widgets::AbstractPropertyManager::getInstance().m_intManager->addProperty(tr("Block Height (pixels)"));

    blockProp->addSubProperty(nBlockXProp);
    blockProp->addSubProperty(nBlockYProp);
    blockProp->addSubProperty(blockWProp);
    blockProp->addSubProperty(blockHProp);

    nBlockXProp->setEnabled(false);
    nBlockYProp->setEnabled(false);
    blockWProp->setEnabled(false);
    blockHProp->setEnabled(false);

    te::qt::widgets::AbstractPropertyManager::getInstance().m_intManager->setValue(nBlockXProp, band->getProperty()->m_nblocksx);
    te::qt::widgets::AbstractPropertyManager::getInstance().m_intManager->setValue(nBlockYProp, band->getProperty()->m_nblocksy);
    te::qt::widgets::AbstractPropertyManager::getInstance().m_intManager->setValue(blockWProp, band->getProperty()->m_blkw);
    te::qt::widgets::AbstractPropertyManager::getInstance().m_intManager->setValue(blockHProp, band->getProperty()->m_blkh);

    bandProp->addSubProperty(blockProp);

    //get category names
    if (!band->getProperty()->m_categoryNames.empty())
    {
      QtProperty* categProp = te::qt::widgets::AbstractPropertyManager::getInstance().m_groupManager->addProperty(tr("Name of categories"));

      //create a property for each category name
      for (std::size_t c = 0; c < band->getProperty()->m_categoryNames.size(); ++c)
      {
        QtProperty* cNameProp = te::qt::widgets::AbstractPropertyManager::getInstance().m_stringManager->addProperty(tr("Category ") + QString::number(c));

        categProp->addSubProperty(cNameProp);

        cNameProp->setEnabled(false);

        te::qt::widgets::AbstractPropertyManager::getInstance().m_stringManager->setValue(cNameProp, band->getProperty()->m_categoryNames[c].c_str());
      }

      bandProp->addSubProperty(categProp);
    }

    //get metadata
    if (!band->getProperty()->m_metadata.empty())
    {
      QtProperty* metaProp = te::qt::widgets::AbstractPropertyManager::getInstance().m_groupManager->addProperty(tr("Metadata"));

      //create a property for each metadata name
      for (std::size_t c = 0; c < band->getProperty()->m_metadata.size(); ++c)
      {
        QtProperty* metaItemProp =
          te::qt::widgets::AbstractPropertyManager::getInstance().m_stringManager->addProperty(
          QString::fromStdString(band->getProperty()->m_metadata[c].first));

        metaProp->addSubProperty(metaItemProp);

        metaItemProp->setEnabled(false);

        te::qt::widgets::AbstractPropertyManager::getInstance().m_stringManager->setValue(
          metaItemProp, QString::fromStdString(
          band->getProperty()->m_metadata[c].second));
      }

      bandProp->addSubProperty(metaProp);
    }

    //add band top property
    rasterInfo_prop->addSubProperty(bandProp);
  }
}

te::qt::plugins::pgisRaster::RasterPropertiesInfo::~RasterPropertiesInfo() =
    default;
