/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

This file is part of the TerraLib - a Framework for building GIS enabled applications.

TerraLib is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

TerraLib is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with TerraLib. See COPYING. If not, write to
TerraLib Team at <terralib-team@terralib.org>.
*/

/*!
\file terralib/qt/plugins/datasource/gdal/Utils.cpp

\brief Utility functions for the GDAL data source widget plugin.
*/

// TerraLib
#include "../../../../core/logger/Logger.h"
#include "../../../../core/translator/Translator.h"
#include "../../../../dataaccess/dataset/DataSetType.h"
#include "../../../../dataaccess/datasource/DataSourceManager.h"
#include "../../../../dataaccess/utils/Utils.h"
#include "../../../../maptools/AbstractLayer.h"
#include "../../../../srs/SpatialReferenceSystemManager.h"
#include "../../../widgets/layer/utils/DataSet2Layer.h"
#include "../../../af/ApplicationController.h"
#include "../../../af/events/LayerEvents.h"
#include "Utils.h"

// Qt
#include <QApplication>
#include <QDir>
#include <QFileInfo>
#include <QMessageBox>

// STL
#include <cassert>

// Boost
#include <boost/uuid/random_generator.hpp>
#include <boost/uuid/uuid_io.hpp>
#include <boost/filesystem.hpp>

void te::qt::plugins::pgisRaster::CreateLayers(te::qt::plugins::pgisRaster::PostGISRasterInfoVec& vec)
{
  std::list<te::map::AbstractLayerPtr> layers;

  for (std::size_t t = 0; t < vec.size(); ++t)
  {
    PostGISRasterInfo info = vec[t];

    //generate id
    boost::uuids::basic_random_generator<boost::mt19937> gen;
    boost::uuids::uuid u = gen();
    std::string id = boost::uuids::to_string(u);

    //create datasource info
    te::da::DataSourceInfoPtr dsInfo(new te::da::DataSourceInfo);

    dsInfo->setAccessDriver("GDAL");
    dsInfo->setConnInfo(info.m_uri);
    dsInfo->setDescription("A PostGIS Raster");
    dsInfo->setId(id);
    dsInfo->setTitle(info.m_tableName);
    dsInfo->setType("GDALPOSTGISRASTER");

    if (!te::da::DataSourceInfoManager::getInstance().add(dsInfo))
      dsInfo = te::da::DataSourceInfoManager::getInstance().getByConnInfo(dsInfo->getConnInfoAsString());

    //create layers
    te::da::DataSourcePtr ds = te::da::DataSourceManager::getInstance().get(dsInfo->getId(), dsInfo->getAccessDriver(), dsInfo->getConnInfo());
    
    te::da::DataSetTypePtr dataSetType = ds->getDataSetType(info.m_tableName);

    try
    {
      te::qt::widgets::DataSet2Layer ds2layer(dsInfo->getId());
      te::map::AbstractLayerPtr l = ds2layer(dataSetType);

      layers.push_back(l);
    }
    catch (std::exception&)
    {
      QMessageBox::warning(nullptr, TE_TR("PostGIS Raster"), TE_TR("There were issues while reading the layer from the requested PostGIS Raster."));
    }
  }

  te::map::AbstractLayerPtr parentLayer(nullptr);
  te::qt::af::evt::LayersAdded added(layers, parentLayer);
  te::qt::af::AppCtrlSingleton::getInstance().trigger(&added);
}
