/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralib/raster/ExifGpsMetadata.cpp

   \brief Exchangeable image file format for digital still cameras.
 */


#include "ExifGpsMetadata.h"

 // TerraLib
#include <terralib/common/StringUtils.h>

te::qt::plugins::photolayer::ExifGpsMetadata::ExifGpsMetadata(te::rst::Raster* raster) : m_raster(raster)
{
  m_rasterInfo = m_raster->getInfo();
}

te::qt::plugins::photolayer::ExifGpsMetadata::~ExifGpsMetadata() = default;

bool te::qt::plugins::photolayer::ExifGpsMetadata::HasExifGpsInfo(te::rst::Raster* raster)
{
  std::map<std::string, std::string> rasterInfo = raster->getInfo();

  //get gps information
  std::map<std::string, std::string>::iterator it = rasterInfo.find("EXIF_GPSVersionID");

  if (it == rasterInfo.end())
  {
    return false;
  }

  return true;
}

double te::qt::plugins::photolayer::ExifGpsMetadata::getAltitude()
{
  //get altitude info
  std::map<std::string, std::string>::iterator it = m_rasterInfo.find("EXIF_GPSAltitude");

  if (it == m_rasterInfo.end())
  {
    return 0.;
  }

  double altitude = atof(fixValue(it->second.c_str()).c_str());

  //get reference
  std::map<std::string, std::string>::iterator itRef = m_rasterInfo.find("EXIF_GPSAltitudeRef");

  if (itRef != m_rasterInfo.end())
  {
    if (fixValue(itRef->second.c_str()) == "1")
    {
      altitude = altitude * -1;
    }
  }

  return altitude;
}

std::string te::qt::plugins::photolayer::ExifGpsMetadata::getDateTime()
{
  std::string dateTime;

  std::map<std::string, std::string>::iterator it = m_rasterInfo.find("EXIF_DateTimeOriginal");

  if (it != m_rasterInfo.end())
  {
    dateTime = it->second;
  }

  return dateTime;
}

void te::qt::plugins::photolayer::ExifGpsMetadata::getLatitudeDMS(double& degree, double& minute, double& second)
{
  //get latitude info
  std::map<std::string, std::string>::iterator itLat = m_rasterInfo.find("EXIF_GPSLatitude");

  if (itLat != m_rasterInfo.end())
  {
    std::string fullLatInfo = itLat->second;

    getDMS(fullLatInfo, degree, minute, second);

    //check reference
    std::map<std::string, std::string>::iterator itLatRef = m_rasterInfo.find("EXIF_GPSLatitudeRef");

    if (itLatRef != m_rasterInfo.end())
    {
      if (fixValue(itLatRef->second.c_str()) == "S")
      {
        degree = degree * -1;
      }
    }
  }
}

void te::qt::plugins::photolayer::ExifGpsMetadata::getLongitudeDMS(double& degree, double& minute, double& second)
{
  //get latitude info
  std::map<std::string, std::string>::iterator itLong = m_rasterInfo.find("EXIF_GPSLongitude");

  if (itLong != m_rasterInfo.end())
  {
    std::string fullLongInfo = itLong->second;

    getDMS(fullLongInfo, degree, minute, second);

    //check reference
    std::map<std::string, std::string>::iterator itLongRef = m_rasterInfo.find("EXIF_GPSLongitudeRef");

    if (itLongRef != m_rasterInfo.end())
    {
      if (fixValue(itLongRef->second.c_str()) == "W")
      {
        degree = degree * -1;
      }
    }
  }
}

te::gm::Point* te::qt::plugins::photolayer::ExifGpsMetadata::getLocation()
{
  //get lat info
  double latDegree = 0.;
  double latMinute = 0.;
  double latSecond = 0.;

  getLatitudeDMS(latDegree, latMinute, latSecond);

  //get long info
  double longDegree = 0.;
  double longMinute = 0.;
  double longSecond = 0.;

  getLongitudeDMS(longDegree, longMinute, longSecond);

  //transform to decimal degrees
  double latValue = dms2dd(latDegree, latMinute, latSecond);
  double longValue = dms2dd(longDegree, longMinute, longSecond);

  //srs
  int srid = 4326;

  //create point
  te::gm::Point* point = new te::gm::Point(longValue, latValue, srid);

  return point;
}

std::string te::qt::plugins::photolayer::ExifGpsMetadata::fixValue(const std::string& value)
{
  std::string newValue = value;

  newValue.erase(std::remove(newValue.begin(), newValue.end(), '('), newValue.end());
  newValue.erase(std::remove(newValue.begin(), newValue.end(), ')'), newValue.end());

  return newValue;
}

void te::qt::plugins::photolayer::ExifGpsMetadata::getDMS(const std::string& value, double& degree, double& minute, double& second)
{
  std::vector<std::string> dmsToken;
  te::common::Tokenize(value, dmsToken, " ");

  if (dmsToken.size() == 3)
  {
    for (std::size_t i = 0; i < dmsToken.size(); ++i)
    {
      std::string value = fixValue(dmsToken[i]);

      dmsToken[i] = value;
    }

    degree = atof(dmsToken[0].c_str());
    minute = atof(dmsToken[1].c_str());
    second = atof(dmsToken[2].c_str());
  }
}

double te::qt::plugins::photolayer::ExifGpsMetadata::dms2dd(double degree, double minute, double second)
{
  bool negative = false;

  if (degree < 0)
  {
    degree = degree * -1;

    negative = true;
  }

  double dd = degree + (minute / 60.) + (second / 3600.);

  if (negative)
  {
    dd = dd * -1;
  }

  return dd;
}
