/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/plugins/photolayer/ExifGpsMetadata.h

  \brief Exchangeable image file format for digital still cameras. 
*/

#ifndef __TE_QT_PLUGINS_PHOTOLAYER_INTERNAL_EXIFGPSMETADATA_H
#define __TE_QT_PLUGINS_PHOTOLAYER_INTERNAL_EXIFGPSMETADATA_H

// TerraLib
#include <terralib/raster/Raster.h>
#include <terralib/geometry/Point.h>

// STL
#include <map>
#include <string>

namespace te
{
  namespace qt
  {
    namespace plugins
    {
      namespace photolayer
      {
        /*!
          \class ExifGpsMetadata

          \brief Exchangeable image file format for digital still cameras.

          \ingroup rst
        */
        class ExifGpsMetadata
        {
        public:

          /*!
            \brief Constructor.

            \param r The raster to get gps information.
          */
          ExifGpsMetadata(te::rst::Raster* raster);

          /*! \brief Destructor. */
          ~ExifGpsMetadata();

        public:

          /*!
            \brief Static function used to check if raster has exif gps metadata information.

            \param r The raster to get gps information.

            \note  GPSVersionID - Indicates the version of GPSInfoIFD. The version is given as 2.2.0.0.
                                  This tag is mandatory when GPSInfo tag is present.
          */
          static bool HasExifGpsInfo(te::rst::Raster* raster);

          /*!
            \brief Function used to get the information about altitude.

            \return Double value with altitude information.

            \note   GPSAltitude - Indicates the altitude based on the reference in GPSAltitudeRef.
                                  Altitude is expressed as one RATIONAL value. The reference unit is meters.

                    GPSAltitudeRef - Indicates the altitude used as the reference altitude.If the reference is sea level
                                     and the altitude is above sea level, 0 is given.If the altitude is below sea level,
                                     a value of 1 is given and the altitude is indicated as an absolute value in
                                     the GPSAltitude tag.The reference unit is meters.
          */
          double getAltitude();

          /*!
            \brief Function used to get the date time information

            \return String value with date time information.

            \note   DateTimeOriginal - The date and time when the original image data was generated.
                                       For a DSC the date and time the picture was taken are recorded.
                                       The format is "YYYY:MM:DD HH:MM:SS" with time shown in 24-hour format, and the date and time
                                       separated by one blank character [20.H]. When the date and time are unknown, all the character spaces
                                       except colons (":") may be filled with blank characters, or else the Interoperability field may be
                                       filled with blank characters.
          */
          std::string getDateTime();

          /*!
            \brief Function used to get the latitude information.

            \param degree Reference value to get degree information
            \param minute Reference value to get minute information
            \param second Reference value to get second information

            \note   GPSLatitude - Indicates the latitude. The latitude is expressed as three RATIONAL values giving
                                  the degrees, minutes, and seconds, respectively.

                    GPSLatitudeRef - Indicates whether the latitude is north or south latitude. The ASCII value 'N'
                                     indicates north latitude, and 'S' is south latitude.
          */
          void getLatitudeDMS(double& degree, double& minute, double& second);

          /*!
            \brief Function used to get the longitude information.

            \param degree Reference value to get degree information
            \param minute Reference value to get minute information
            \param second Reference value to get second information

            \note   GPSLongitude - Indicates the longitude. The longitude is expressed as three RATIONAL values giving
                                  the degrees, minutes, and seconds, respectively.

                    GPSLongitudeRef - Indicates whether the longitude is east or west longitude. ASCII 'E'
                                      indicates east longitude, and 'W' is west longitude.
          */
          void getLongitudeDMS(double& degree, double& minute, double& second);

          /*!
            \brief Function used to get a geometry with raster gps location.

            \return Point geometry with raster location if it was readed correctly.

            \note   Using default srs value: Lat Long with datum 'WGS-84' - SRID: 4326
          */
          te::gm::Point* getLocation();

        protected:

          /*!
            \brief Function used to remove "(" and ")" from string

            \return string with only value contents
          */
          std::string fixValue(const std::string& value);

          /*!
            \brief Function used to get the degree, minute, second information from a exif string value

            \param value Exif string value with dms information
            \param degree Reference value to get degree information
            \param minute Reference value to get minute information
            \param second Reference value to get second information
          */
          void getDMS(const std::string& value, double& degree, double& minute, double& second);

          /*!
            \brief Function used to convert from degree, minute, second to decimal degree

            \param degree Reference value to degree information
            \param minute Reference value to minute information
            \param second Reference value to second information

            \return double value with decimal degree information
          */
          double dms2dd(double degree, double minute, double second);

        protected:

          te::rst::Raster* m_raster;                          //!< Input raster.

          std::map<std::string, std::string> m_rasterInfo;    //!< Raster metadata information.
        };
      } // end namespace photolayer
    } // end namespace plugins
  } // end namespace qt
} // end namespace te

#endif  // __TE_QT_PLUGINS_PHOTOLAYER_INTERNAL_EXIFGPSMETADATA_H
