/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralib/qt/plugins/photolayer/PhotoLayerBuilder.cpp

   \brief Class used to build a vector layer of points from a set of images with EXIF GPS metadata.
 */


#include "PhotoLayerBuilder.h"
#include "ExifGpsMetadata.h"

 // TerraLib
#include <terralib/common/StringUtils.h>
#include <terralib/dataaccess/datasource/DataSource.h>
#include <terralib/dataaccess/datasource/DataSourceFactory.h>
#include <terralib/dataaccess/utils/Utils.h>
#include <terralib/datatype/StringProperty.h>
#include <terralib/geometry/GeometryProperty.h>
#include <terralib/geometry/Point.h>
#include <terralib/memory/DataSet.h>
#include <terralib/memory/DataSetItem.h>
#include <terralib/raster/RasterProperty.h>

te::qt::plugins::photolayer::PhotoLayerBuilder::PhotoLayerBuilder(te::da::DataSourcePtr ds, const std::string& dataSetName, const std::string& path): 
  m_dataSource(ds),
  m_dataSetName(dataSetName),
  m_path(path)
{

}

te::qt::plugins::photolayer::PhotoLayerBuilder::~PhotoLayerBuilder() = default;

std::size_t te::qt::plugins::photolayer::PhotoLayerBuilder::GetValidFilesCount(const std::string& path)
{
  std::size_t count = 0;

  std::string connInfo("file://");
  connInfo += path;

  std::unique_ptr<te::da::DataSource> dsptr = te::da::DataSourceFactory::make("GDAL", connInfo);
  dsptr->open();

  std::vector<std::string> dsNames = dsptr->getDataSetNames();

  for (std::size_t i = 0; i < dsNames.size(); ++i)
  {
    std::string curDataSet = dsNames[i];

    std::unique_ptr<te::da::DataSetType> dataSetType = dsptr->getDataSetType(curDataSet);

    if (dataSetType == nullptr || dataSetType->hasRaster() == false)
    {
      continue;
    }

    std::unique_ptr<te::da::DataSet> dataSet = dsptr->getDataSet(curDataSet);

    te::rst::RasterProperty* rasterProp = te::da::GetFirstRasterProperty(dataSetType.get());

    std::unique_ptr<te::rst::Raster> raster = dataSet->getRaster(rasterProp->getName());

    bool isValidRaster = te::qt::plugins::photolayer::ExifGpsMetadata::HasExifGpsInfo(raster.get());

    if (isValidRaster == true)
    {
      ++count;
    }
  }

  return count;
}

void te::qt::plugins::photolayer::PhotoLayerBuilder::execute()
{
  std::string connInfo("file://");
  connInfo += m_path;

  std::unique_ptr<te::da::DataSource> dsptr = te::da::DataSourceFactory::make("GDAL", connInfo);
  dsptr->open();

  std::vector<std::string> dsNames = dsptr->getDataSetNames();

  if (dsNames.empty())
  {
    return;
  }

  //create dataset type
  std::unique_ptr<te::da::DataSetType> dsType = createDataSetType();

  //create dataset
  std::unique_ptr<te::mem::DataSet> dataSet(new te::mem::DataSet(dsType.get()));

  int count = 0;

  //iterate for each image
  for (std::size_t i = 0; i < dsNames.size(); ++i)
  {
    std::string curDataSet = dsNames[i];

    std::unique_ptr<te::da::DataSetType> dataSetType = dsptr->getDataSetType(curDataSet);

    if (dataSetType == nullptr || dataSetType->hasRaster() == false)
    {
      continue;
    }

    std::unique_ptr<te::da::DataSet> dataSetRaster = dsptr->getDataSet(curDataSet);

    te::rst::RasterProperty* rasterProp = te::da::GetFirstRasterProperty(dataSetType.get());

    //get raster
    std::unique_ptr<te::rst::Raster> raster = dataSetRaster->getRaster(rasterProp->getName());

    bool isValidRaster = te::qt::plugins::photolayer::ExifGpsMetadata::HasExifGpsInfo(raster.get());

    if (isValidRaster == false)
    {
      continue;
    }

    //get exif gps information
    te::qt::plugins::photolayer::ExifGpsMetadata exifGps(raster.get());

    //create dataset item
    te::mem::DataSetItem* item = new te::mem::DataSetItem(dataSet.get());

    //set id
    item->setInt32(0, count);

    //set fileName
    std::string fileName = m_path + "/" + curDataSet;

    item->setString(1, fileName);

    //set datetime
    std::string dateTime = exifGps.getDateTime();

    item->setString(2, dateTime);

    //set altitude
    double altitude = exifGps.getAltitude();

    item->setDouble(3, altitude);

    //set geometry
    te::gm::Point* point = exifGps.getLocation();

    item->setGeometry(4, point);

    //add item
    dataSet->add(item);


    ++count;
  }

  saveDataSet(dataSet.get(), dsType.get());
}


std::unique_ptr<te::da::DataSetType> te::qt::plugins::photolayer::PhotoLayerBuilder::createDataSetType()
{
  std::unique_ptr<te::da::DataSetType> dsType(new te::da::DataSetType(m_dataSetName));

  //create id property
  te::dt::SimpleProperty* idProperty = new te::dt::SimpleProperty("id", te::dt::INT32_TYPE);
  dsType->add(idProperty);

  te::dt::SimpleProperty* fileProperty = new te::dt::StringProperty("filename", te::dt::VAR_STRING, 256);
  dsType->add(fileProperty);

  te::dt::SimpleProperty* dateProperty = new te::dt::StringProperty("date_time", te::dt::VAR_STRING, 256);
  dsType->add(dateProperty);

  te::dt::SimpleProperty* altitudeProperty = new te::dt::SimpleProperty("altitude", te::dt::DOUBLE_TYPE, false);
  dsType->add(altitudeProperty);

  //create geometry property
  te::gm::GeometryProperty* geomProperty = new te::gm::GeometryProperty("geom", 4326, te::gm::PointType);
  dsType->add(geomProperty);

  //create primary key
  std::string pkName = "pk_id";
  pkName += "_" + m_dataSetName;
  te::da::PrimaryKey* pk = new te::da::PrimaryKey(pkName, dsType.get());
  pk->add(idProperty);

  //create spatial index
  std::string idxSpatial = m_dataSetName;
  idxSpatial += "_spatial_idx";
  te::da::Index* idx = new te::da::Index(idxSpatial, te::da::R_TREE_TYPE, dsType.get());
  idx->add(geomProperty);

  return dsType;
}

void te::qt::plugins::photolayer::PhotoLayerBuilder::saveDataSet(te::mem::DataSet* dataSet, te::da::DataSetType* dsType)
{
  //save dataset
  dataSet->moveBeforeFirst();

  std::map<std::string, std::string> options;

  m_dataSource->createDataSet(dsType, options);

  m_dataSource->add(m_dataSetName, dataSet, options);
}