/*  Copyright (C) 2011-2012 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
\file terralib/qt/widgets/layer/utils/CreateLayerDialog.cpp

\brief This interface is used to create new layer operation.
*/


#include "PhotoLayerDialog.h"
#include "ui/ui_PhotoLayerDialogForm.h"

#include "PhotoLayerBuilder.h"

// TerraLib
#include <terralib/dataaccess/datasource/DataSource.h>
#include <terralib/dataaccess/datasource/DataSourceFactory.h>
#include <terralib/dataaccess/datasource/DataSourceInfoManager.h>
#include <terralib/dataaccess/datasource/DataSourceManager.h>
#include <terralib/dataaccess/utils/Utils.h>
#include <terralib/qt/widgets/datasource/selector/DataSourceSelectorDialog.h>
#include <terralib/qt/widgets/layer/utils/DataSet2Layer.h>
#include <terralib/qt/widgets/utils/ScopedCursor.h>
#include <terralib/qt/widgets/Utils.h>

// Qt
#include <QFileDialog>
#include <QMessageBox>

// Boost
#include <boost/filesystem.hpp>
#include <boost/uuid/random_generator.hpp>
#include <boost/uuid/uuid_io.hpp>

te::qt::plugins::photolayer::PhotoLayerDialog::PhotoLayerDialog(QWidget* parent, Qt::WindowFlags f)
  : QDialog(parent, f),
  m_ui(new Ui::PhotoLayerDialogForm)
{
  // add controls
  m_ui->setupUi(this);
  m_ui->m_imgLabel->setPixmap(QIcon::fromTheme("photo-layer-hint").pixmap(112, 48));

  m_ui->m_inputFileToolButton->setIcon(QIcon::fromTheme("folder"));
  m_ui->m_targetFileToolButton->setIcon(QIcon::fromTheme("folder"));
  m_ui->m_targetDatasourceToolButton->setIcon(QIcon::fromTheme("datasource"));

  //connects
  connect(m_ui->m_inputFileToolButton, SIGNAL(pressed()), this, SLOT(onInputFileToolButtonPressed()));
  connect(m_ui->m_targetDatasourceToolButton, SIGNAL(pressed()), this, SLOT(onTargetDatasourceToolButtonPressed()));
  connect(m_ui->m_targetFileToolButton, SIGNAL(pressed()), this, SLOT(onTargetFileToolButtonPressed()));
  connect(m_ui->m_okPushButton, SIGNAL(clicked()), this, SLOT(onOkPushButtonClicked()));
}

te::qt::plugins::photolayer::PhotoLayerDialog::~PhotoLayerDialog() = default;

te::map::AbstractLayerPtr te::qt::plugins::photolayer::PhotoLayerDialog::getLayer()
{
  return m_outputLayer;
}

void te::qt::plugins::photolayer::PhotoLayerDialog::onInputFileToolButtonPressed()
{
  QString fileName = QFileDialog::getExistingDirectory(this, tr("Select Photo Directory..."), QString());

  if (fileName.isEmpty())
  {
    m_ui->m_inputPathLineEdit->clear();
    m_ui->m_fileCountLabel->clear();

    return;
  }

  m_ui->m_inputPathLineEdit->setText(fileName);

  std::size_t count = 0;
  {
    te::qt::widgets::ScopedCursor cursor(Qt::WaitCursor);

    count = te::qt::plugins::photolayer::PhotoLayerBuilder::GetValidFilesCount(te::qt::widgets::Convert2StdString(fileName));
  }

  m_ui->m_fileCountLabel->setText(QString::number(count));
}

void te::qt::plugins::photolayer::PhotoLayerDialog::onTargetDatasourceToolButtonPressed()
{
  m_ui->m_newLayerNameLineEdit->clear();
  m_ui->m_newLayerNameLineEdit->setEnabled(true);

  te::qt::widgets::DataSourceSelectorDialog dlg(this);
  dlg.exec();

  std::list<te::da::DataSourceInfoPtr> dsPtrList = dlg.getSelecteds();

  if (dsPtrList.size() <= 0)
    return;

  std::list<te::da::DataSourceInfoPtr>::iterator it = dsPtrList.begin();

  m_ui->m_repositoryLineEdit->setText(QString(it->get()->getTitle().c_str()));

  m_outputDatasource = *it;

  te::da::DataSourcePtr outputDataSource = te::da::DataSourceManager::getInstance().get(m_outputDatasource->getId(), m_outputDatasource->getType(), m_outputDatasource->getConnInfo());

  outputDataSource->open();

  m_toFile = false;
}

void te::qt::plugins::photolayer::PhotoLayerDialog::onTargetFileToolButtonPressed()
{
  m_ui->m_newLayerNameLineEdit->clear();
  m_ui->m_repositoryLineEdit->clear();

  QString fileName = QFileDialog::getSaveFileName(this, tr("Save as..."), QString(), tr("Shapefile (*.shp *.SHP);;"), nullptr, QFileDialog::DontConfirmOverwrite);

  if (fileName.isEmpty())
    return;

  boost::filesystem::path outfile(fileName.toUtf8().data());

  m_ui->m_repositoryLineEdit->setText(outfile.string().c_str());

  m_ui->m_newLayerNameLineEdit->setText(outfile.leaf().string().c_str());

  m_ui->m_newLayerNameLineEdit->setEnabled(false);

  m_toFile = true;

  //create new data source
  boost::filesystem::path uri(m_ui->m_repositoryLineEdit->text().toUtf8().data());

  const std::string connInfo("file://" + uri.string());
  std::unique_ptr<te::da::DataSource> ds = te::da::DataSourceFactory::make("OGR", connInfo);
  ds->open();

  boost::uuids::basic_random_generator<boost::mt19937> gen;
  boost::uuids::uuid u = gen();
  std::string id_ds = boost::uuids::to_string(u);

  te::da::DataSourceInfoPtr dsInfoPtr(new te::da::DataSourceInfo);
  dsInfoPtr->setConnInfo(connInfo);
  dsInfoPtr->setTitle(uri.stem().string());
  dsInfoPtr->setAccessDriver("OGR");
  dsInfoPtr->setType("OGR");
  dsInfoPtr->setDescription(uri.string());
  dsInfoPtr->setId(id_ds);

  te::da::DataSourceInfoManager::getInstance().add(dsInfoPtr);

  m_outputDatasource = dsInfoPtr;

  te::da::DataSourcePtr outputDataSource = te::da::DataSourceManager::getInstance().get(m_outputDatasource->getId(), m_outputDatasource->getType(), m_outputDatasource->getConnInfo());

  outputDataSource->open();
}

void te::qt::plugins::photolayer::PhotoLayerDialog::onOkPushButtonClicked()
{
  // check input parameters
  if (m_ui->m_inputPathLineEdit->text().isEmpty())
  {
    QMessageBox::information(this, tr("Warning"), tr("Input path not defined."));
    return;
  }

  std::string path = te::qt::widgets::Convert2StdString(m_ui->m_inputPathLineEdit->text());

  //check if any valid file was found
  if(m_ui->m_fileCountLabel->text().isEmpty())
  {
    QMessageBox::information(this, tr("Warning"), tr("No valid files were found."));
    return;
  }

  bool checkStr = false;

  int nFiles = m_ui->m_fileCountLabel->text().toInt(&checkStr);

  if (checkStr == false || nFiles == 0)
  {
    QMessageBox::information(this, tr("Warning"), tr("No valid files were found."));
    return;
  }

  //get output information
  if (m_ui->m_repositoryLineEdit->text().isEmpty())
  {
    QMessageBox::information(this, tr("Warning"), tr("Define a repository for the result."));
    return;
  }

  if (m_ui->m_newLayerNameLineEdit->text().isEmpty())
  {
    QMessageBox::information(this, tr("Warning"), tr("Define a name for the resulting layer."));
    return;
  }

  //fix layer name
  std::string layerName = te::qt::widgets::Convert2StdString(m_ui->m_newLayerNameLineEdit->text());

  std::size_t idx = layerName.find(".");

  if (idx != std::string::npos)
    layerName = layerName.substr(0, idx);

  //start operation
  te::da::DataSourcePtr outDataSource = te::da::GetDataSource(m_outputDatasource->getId());

  te::qt::plugins::photolayer::PhotoLayerBuilder builder(outDataSource, layerName, path);

  try
  {
    //execute
    te::qt::widgets::ScopedCursor cursor(Qt::WaitCursor);

    builder.execute();

    //create output layer
    te::qt::widgets::DataSet2Layer converter(m_outputDatasource->getId());

    te::da::DataSetTypePtr dt(outDataSource->getDataSetType(layerName).release());

    m_outputLayer = converter(dt);
  }
  catch (const std::exception& e)
  {
    QMessageBox::warning(this, tr("Warning"), e.what());

    return;
  }
  catch (...)
  {
    QMessageBox::warning(this, tr("Warning"), tr("An exception has occurred!"));

    return;
  }
 
  QMessageBox::information(this, tr("Information"), tr("Photo Layer was created!"));

  accept();
}
