/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralib/qt/plugins/photolayer/PhotoViewerDialog.cpp

   \brief This file has the PhotoViewerDialog class.
 */

#include "PhotoViewerDialog.h"
#include "ui/ui_PhotoViewerDialogForm.h"

// TerraLib
#include "../../../dataaccess/datasource/DataSourceInfo.h"
#include "../../../dataaccess/datasource/DataSourceInfoManager.h"
#include "../../../dataaccess/datasource/DataSourceManager.h"
#include "../../../maptools/Utils.h"
#include "../../widgets/layer/utils/DataSet2Layer.h"
#include "../../widgets/tools/PanMiddleClick.h"
#include "../../widgets/tools/ZoomWheel.h"
#include "../../widgets/canvas/Canvas.h"
#include "../../widgets/canvas/MultiThreadMapDisplay.h"
#include "../../widgets/Utils.h"

// Qt
#include <QGridLayout>
#include <QMessageBox>

// Boost
#include <boost/uuid/random_generator.hpp>
#include <boost/uuid/uuid_io.hpp>
#include <boost/filesystem.hpp>

te::qt::plugins::photolayer::PhotoViewerDialog::PhotoViewerDialog(QWidget* parent, Qt::WindowFlags f)
  : QDialog(parent, f),
    m_ui(new Ui::PhotoViewerDialogForm),
    m_panTool(nullptr), 
    m_zoomTool(nullptr)
{
  // setup inteface
  m_ui->setupUi(this);

//build map display
  QGridLayout* displayLayout = new QGridLayout(m_ui->m_frame);
  m_mapDisplay = new te::qt::widgets::MultiThreadMapDisplay(m_ui->m_frame->size(), m_ui->m_frame);

  displayLayout->addWidget(m_mapDisplay);
  displayLayout->setContentsMargins(0,0,0,0);

//start tools
  m_panTool = new te::qt::widgets::PanMiddleClick(m_mapDisplay);
  m_zoomTool = new te::qt::widgets::ZoomWheel(m_mapDisplay, 1.5);

  m_mapDisplay->installEventFilter(m_panTool);
  m_mapDisplay->installEventFilter(m_zoomTool);

//connects
  connect(m_ui->m_listWidget, SIGNAL(itemActivated(QListWidgetItem *)), this, SLOT(onItemActivated(QListWidgetItem *)));


  this->setMinimumSize(800, 600);
}

te::qt::plugins::photolayer::PhotoViewerDialog::~PhotoViewerDialog()
{
  delete m_panTool;
  delete m_zoomTool;
}

void te::qt::plugins::photolayer::PhotoViewerDialog::setListOfFiles(std::vector<std::string>& files)
{
  m_ui->m_listWidget->clear();

  if (files.empty())
  {
    return;
  }

  for (std::size_t i = 0; i < files.size(); ++i)
  {
    m_ui->m_listWidget->addItem(files[i].c_str());
  }

  QListWidgetItem* item = m_ui->m_listWidget->item(0);
  item->setSelected(true);

  onItemActivated(item);

  if (files.size() == 1)
  {
    m_ui->m_filesGroupBox->setVisible(false);
  }
}

void te::qt::plugins::photolayer::PhotoViewerDialog::setURI(const std::string& uri)
{
  te::da::DataSourceInfoPtr dsInfo(new te::da::DataSourceInfo);

  dsInfo->setAccessDriver("GDAL");

  dsInfo->setConnInfo("file://" + uri);

  dsInfo->setDescription("A photo image file.");

  boost::uuids::basic_random_generator<boost::mt19937> gen;
  boost::uuids::uuid u = gen();
  std::string id = boost::uuids::to_string(u);

  dsInfo->setId(id);

  boost::filesystem::path mpath(uri);

  std::string fileBaseName = mpath.stem().string();

  dsInfo->setTitle(fileBaseName);

  dsInfo->setType("GDAL");

  if (!te::da::DataSourceInfoManager::getInstance().add(dsInfo))
    dsInfo = te::da::DataSourceInfoManager::getInstance().getByConnInfo(dsInfo->getConnInfoAsString());

  //get datasource and layer
  te::da::DataSourcePtr ds = te::da::DataSourceManager::getInstance().get(dsInfo->getId(), dsInfo->getType(), dsInfo->getConnInfo());

  std::vector<std::string> dsets = ds->getDataSetNames();

  if (dsets.empty())
  {
    QMessageBox::warning(this, tr("Warning"), tr("File not found."));
    return;
  }

  te::da::DataSetTypePtr dsType = ds->getDataSetType(dsets[0]);

  te::qt::widgets::DataSet2Layer dataset2layer(dsInfo->getId());

  te::map::AbstractLayerPtr layer = dataset2layer(dsType);
  layer->setVisibility(te::map::VISIBLE);

  //show data
  std::list<te::map::AbstractLayerPtr> list;
  list.push_back(layer);

  te::gm::Envelope e = layer->getExtent();

  m_mapDisplay->setMouseTracking(true);
  m_mapDisplay->setSRID(layer->getSRID(), false);
  m_mapDisplay->setExtent(e, false);
  m_mapDisplay->setLayerList(list);
}

void te::qt::plugins::photolayer::PhotoViewerDialog::onItemActivated(QListWidgetItem *item)
{
  if (item == nullptr)
  {
    return;
  }

  std::string uri = te::qt::widgets::Convert2StdString(item->text());

  setURI(uri);

  m_mapDisplay->refresh(true);
}

void te::qt::plugins::photolayer::PhotoViewerDialog::resizeEvent(QResizeEvent* e)
{
  QDialog::resizeEvent(e);
}
