/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralib/qt/plugins/photolayer/Plugin.cpp

   \brief Plugin implementation for the TerraLib Photo Layer Qt Plugin.
 */

#include "Plugin.h"
#include "PhotoLayerDialog.h"
#include "PhotoViewerDialog.h"

// TerraLib
#include "../../../common/Config.h"
#include "../../../common/TerraLib.h"
#include "../../../core/translator/Translator.h"
#include "../../../core/logger/Logger.h"
#include "../../../dataaccess/utils/Utils.h"
#include "../../../geometry/GeometryProperty.h"
#include "../../../maptools/AbstractLayer.h"
#include "../../../maptools/WorldDeviceTransformer.h"
#include "../../af/ApplicationController.h"
#include "../../af/BaseApplication.h"
#include "../../af/events/ApplicationEvents.h"
#include "../../af/events/LayerEvents.h"
#include "../../widgets/canvas/MapDisplay.h"
#include "../../widgets/layer/explorer/LayerItem.h"
#include "../../widgets/layer/explorer/LayerItemView.h"
#include "../../widgets/Utils.h"

// QT
#include <QAction>
#include <QObject>
#include <QMenu>
#include <QMenuBar>
#include <QMessageBox>

bool IsPhotoLayer(te::map::AbstractLayerPtr layer);

te::qt::plugins::photolayer::Plugin::Plugin(const te::core::PluginInfo& pluginInfo)
    : te::core::CppPlugin(pluginInfo), m_pointPickerTool(nullptr)
{
}

te::qt::plugins::photolayer::Plugin::~Plugin()
{
  te::qt::af::BaseApplication* ba = dynamic_cast<te::qt::af::BaseApplication*>(te::qt::af::AppCtrlSingleton::getInstance().getMainWindow());

  te::qt::widgets::MapDisplay* mapDisplay = ba->getMapDisplay();

  if (mapDisplay->getCurrentTool() == m_pointPickerTool)
  {
    mapDisplay->setCurrentTool(nullptr);
  }
}

void te::qt::plugins::photolayer::Plugin::startup()
{
  if(m_initialized)
    return;

  // Register the application framework listener
  te::qt::af::AppCtrlSingleton::getInstance().addListener(this, te::qt::af::SENDER);

  // Get plugins menu
  QMenu* pluginsMenu = te::qt::af::AppCtrlSingleton::getInstance().getMenu("Plugins");

  m_menu = new QMenu(tr("Photo Layer"));

  pluginsMenu->addMenu(m_menu);

  //add action into plugin menu
  m_action = new QAction(m_menu);
  m_action->setText("Photo Layer Creator...");
  m_action->setIcon(QIcon::fromTheme("photo"));

  m_menu->addAction(m_action);

  connect(m_action, SIGNAL(triggered(bool)), this, SLOT(onActionActivated(bool)));

  //toolbar visibility
  m_toolVisibleAction = new QAction(m_menu);
  m_toolVisibleAction->setText("Photo Layer Tool Bar");
  m_toolVisibleAction->setIcon(QIcon::fromTheme("photo_view"));
  m_toolVisibleAction->setCheckable(true);

  m_menu->addAction(m_toolVisibleAction);

  connect(m_toolVisibleAction, SIGNAL(triggered(bool)), this, SLOT(onToolVisibleActionActivated(bool)));

  //start toolbar
  m_toolBar = new QToolBar(tr("Photo Layer Tool Bar"));

  m_toolAction = new QAction(m_toolBar);
  m_toolAction->setText("Photo Layer Tool Viewer...");
  m_toolAction->setIcon(QIcon::fromTheme("photo_view"));
  m_toolAction->setCheckable(true);

  connect(m_toolAction, SIGNAL(triggered(bool)), this, SLOT(onToolActionActivated(bool)));

  m_toolBar->addAction(m_toolAction);
  m_toolBar->setVisible(false);

  QActionGroup* actionGroup = te::qt::af::AppCtrlSingleton::getInstance().findActionGroup("Map.ToolsGroup");

  actionGroup->addAction(m_toolAction);

  //add new toolbar into application
  te::qt::af::evt::NewActionsAvailable e;
  e.m_toolbar = m_toolBar;
  e.m_category = "PhotoLayer";

  emit triggered(&e);
  
  TE_LOG_TRACE(TE_TR("TerraLib Photo Layer Qt Plugin startup!"));

  m_initialized = true;
}

void te::qt::plugins::photolayer::Plugin::shutdown()
{
  if(!m_initialized)
    return;

  delete m_menu;

  delete m_toolBar;

  TE_LOG_TRACE(TE_TR("TerraLib Photo Layer Qt Plugin shutdown!"));

  m_initialized = false;

  te::qt::af::AppCtrlSingleton::getInstance().removeListener(this);
}

void te::qt::plugins::photolayer::Plugin::onActionActivated(bool)
{
  QWidget* parent = te::qt::af::AppCtrlSingleton::getInstance().getMainWindow();

  te::qt::plugins::photolayer::PhotoLayerDialog dlg(parent);
  int res = dlg.exec();

  if (res == QDialog::Accepted)
  {
    te::map::AbstractLayerPtr layer = dlg.getLayer();

    if (!layer)
      return;

    te::qt::af::evt::LayerAdded evt(layer);

    emit triggered(&evt);
  }
}

void te::qt::plugins::photolayer::Plugin::onToolVisibleActionActivated(bool state)
{
  m_toolBar->setVisible(state);
}

void te::qt::plugins::photolayer::Plugin::onToolActionActivated(bool state)
{
  if (!state)
    return;

  te::qt::af::BaseApplication* ba = dynamic_cast<te::qt::af::BaseApplication*>(te::qt::af::AppCtrlSingleton::getInstance().getMainWindow());

  te::qt::widgets::MapDisplay* mapDisplay = ba->getMapDisplay();

  m_pointPickerTool = new te::qt::widgets::PointPicker(mapDisplay, Qt::CrossCursor);
  mapDisplay->setCurrentTool(m_pointPickerTool);

  connect(m_pointPickerTool, SIGNAL(pointPicked(QPointF&)), this, SLOT(onGetPointCoordinate(QPointF&)));
}

void te::qt::plugins::photolayer::Plugin::onGetPointCoordinate(QPointF& point)
{
  te::qt::af::BaseApplication* ba = dynamic_cast<te::qt::af::BaseApplication*>(te::qt::af::AppCtrlSingleton::getInstance().getMainWindow());

  std::list<te::map::AbstractLayerPtr> selectedLayers = te::qt::widgets::GetSelectedLayersOnly(ba->getLayerExplorer());

  if (selectedLayers.size() != 1)
  {
    QString msg = tr("Select a photo layer to accomplish this operation!");
    QMessageBox::warning(ba, tr("Photo Layer"), msg);
    return;
  }

  //check if layer is valid
  te::map::AbstractLayerPtr layer = selectedLayers.front();

  if (IsPhotoLayer(layer) == false)
  {
    QString msg = tr("Select a photo layer to accomplish this operation!");
    QMessageBox::warning(ba, tr("Photo Layer"), msg);
    return;
  }

  te::qt::widgets::MapDisplay* mapDisplay = ba->getMapDisplay();
  te::gm::Envelope envelope = mapDisplay->getExtent();

  //create envelope
  double deviceX, deviceY;

  te::map::WorldDeviceTransformer transformer(envelope.m_llx, envelope.m_lly, envelope.m_urx, envelope.m_ury, mapDisplay->getWidth(), mapDisplay->getHeight());
  transformer.world2Device(point.x(), point.y(), deviceX, deviceY);

  QPointF pixelOffset(4.0, 4.0);
  QPointF localPos(deviceX, deviceY);
  QRectF rect(localPos - pixelOffset, localPos + pixelOffset);
 
  QPointF ll(rect.left(), rect.bottom());
  QPointF ur(rect.right(), rect.top());
  ll = mapDisplay->transform(ll);
  ur = mapDisplay->transform(ur);

  //adjust envelope
  te::gm::Envelope env(ll.x(), ll.y(), ur.x(), ur.y());
   
  if (mapDisplay->getSRID() != 4326)
  {
    env.transform(mapDisplay->getSRID(), 4326);
  }

  //get dataset
  std::unique_ptr<te::da::DataSetType> dsType(layer->getSchema());

  te::gm::GeometryProperty* gp = te::da::GetFirstGeomProperty(dsType.get());

  std::unique_ptr<te::da::DataSet> dataset = layer->getData(gp->getName(), &env, te::gm::INTERSECTS);

  //filter result
  std::vector<std::string> files;

  while (dataset->moveNext())
  {
    std::string fileName = dataset->getString("filename");

    files.push_back(fileName);
  }

  if (files.empty() == false)
  {
    te::qt::plugins::photolayer::PhotoViewerDialog* dataViewer = new te::qt::plugins::photolayer::PhotoViewerDialog(m_menu, Qt::Tool);
    dataViewer->setAttribute(Qt::WA_DeleteOnClose);
    dataViewer->show();

    auto hostRect = ba->geometry();
    dataViewer->move(hostRect.center() - dataViewer->rect().center());

    dataViewer->setListOfFiles(files);
  }
}

bool IsPhotoLayer(te::map::AbstractLayerPtr layer)
{
  std::unique_ptr<te::da::DataSetType> dsType = layer->getSchema();

  if (dsType->hasGeom() == false)
  {
    return false;
  }

  te::gm::GeometryProperty* gmProp = te::da::GetFirstGeomProperty(dsType.get());

  if (gmProp == nullptr)
  {
    return false;
  }

  if (gmProp->getGeometryType() != te::gm::PointType && gmProp->getGeometryType() != te::gm::MultiPointType)
  {
    return false;
  }

  te::dt::Property* prop = dsType->getProperty("filename");

  if (prop == nullptr)
  {
    return false;
  }

  return true;
}

TERRALIB_PLUGIN_CALL_BACK_IMPL(te::qt::plugins::photolayer::Plugin)
