/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/plugins/rp/ContrastAction.cpp

  \brief This file defines the Contrast class
*/

// Terralib
#include "../../../dataaccess/utils/Utils.h"
#include "../../../raster/BandProperty.h"
#include "../../../raster/RasterProperty.h"
#include "../../af/events/LayerEvents.h"
#include "../../af/ApplicationController.h"
#include "../../af/BaseApplication.h"
#include "ContrastAction.h"

// Qt
#include <QObject>
#include <QMessageBox>
#include <QApplication>

// STL
#include <memory>

te::qt::plugins::rp::ContrastAction::ContrastAction(QMenu* menu, QMenu* popupMenu):te::qt::plugins::rp::AbstractAction(menu, popupMenu)
{
  createAction(tr("Contrast...").toUtf8().data(), "histogram");

  m_action->setObjectName("Processing.Raster Processing.Contrast");
}

te::qt::plugins::rp::ContrastAction::~ContrastAction() = default;

void te::qt::plugins::rp::ContrastAction::onActionActivated(bool /*checked*/)
{
  std::list<te::map::AbstractLayerPtr> layersList = getLayers();

  if(layersList.empty())
  {
     QMessageBox::warning(te::qt::af::AppCtrlSingleton::getInstance().getMainWindow(), tr("Contrast"), tr("There is no selected raster layer."));
     return;
  }

  // Register the application framework listener
  te::qt::af::AppCtrlSingleton::getInstance().addListener(this, te::qt::af::RECEIVER);

  m_contrastDlg.reset(new te::qt::widgets::ContrastDialogForm(te::qt::af::AppCtrlSingleton::getInstance().getMainWindow()));

  connect(m_contrastDlg.get(), SIGNAL(addLayer(te::map::AbstractLayerPtr)), this, SLOT(addLayer(te::map::AbstractLayerPtr)));
  connect(m_contrastDlg.get(), SIGNAL(closeTool()), this, SLOT(closeTool()));

  te::qt::af::BaseApplication* ba = dynamic_cast<te::qt::af::BaseApplication*>(te::qt::af::AppCtrlSingleton::getInstance().getMainWindow());

  QActionGroup* mapEditionTools = te::qt::af::AppCtrlSingleton::getInstance().findActionGroup("Map.ToolsGroup");
  assert(mapEditionTools);

  m_contrastDlg->setMapDisplay(ba->getMapDisplay());
  m_contrastDlg->setActionGroup(mapEditionTools);
  m_contrastDlg->setModal(false);

  try
  {
    m_contrastDlg->setList(layersList);
  }
  catch(te::common::Exception& e)
  {
    QMessageBox::warning(te::qt::af::AppCtrlSingleton::getInstance().getMainWindow(), tr("Contrast"), tr(e.what()));
    return;
  }

  m_contrastDlg->show();
}

void te::qt::plugins::rp::ContrastAction::onApplicationTriggered(te::qt::af::evt::Event* e)
{
  switch (e->m_id)
  {
    case te::qt::af::evt::DRAWING_FINISHED:
    {
      m_contrastDlg->onMapDisplayExtentChanged();
    }
    break;

    default:
      break;
  }
}

void te::qt::plugins::rp::ContrastAction::addLayer(te::map::AbstractLayerPtr outputLayer)
{
  //add new layer
  addNewLayer(outputLayer);
}

void te::qt::plugins::rp::ContrastAction::closeTool()
{
  te::qt::af::AppCtrlSingleton::getInstance().removeListener(this);

  m_contrastDlg.reset(nullptr);
}
