/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/plugins/rp/Plugin.cpp

  \brief Plugin implementation for the RP Qt Plugin widget.
*/

// TerraLib
#include "../../../common/Config.h"
#include "../../../core/translator/Translator.h"
#include "../../../core/logger/Logger.h"
#include "../../af/ApplicationController.h"
#include "../../af/events/LayerEvents.h"
#include "../../af/Utils.h"
#include "RasterConfigWidgetFactory.h"
#include "Plugin.h"

#ifdef TE_QT_PLUGIN_RP_HAVE_ARITHMETICOP
  #include "ArithmeticOpAction.h"
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_CLASSIFIER
  #include "ClassifierAction.h"
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_CLIPPING
  #include "ClippingAction.h"
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_CLOUDDETECTION
#include "CloudDetectionAction.h"
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_COLORTRANSFORM
  #include "ColorTransformAction.h"
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_COMPOSEBANDS
  #include "ComposeBandsAction.h"
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_CONTRAST
  #include "ContrastAction.h"
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_FILTER
  #include "FilterAction.h"
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_FUSION
  #include "FusionAction.h"
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_MAJORITYFILTER
  #include "MajorityFilterAction.h"
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_MIXTUREMODEL
  #include "MixtureModelAction.h"
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_MOSAIC
  #include "MosaicAction.h"
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_PALETTE
#include "PaletteAction.h"
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_PCA
#include "PCAAction.h"
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_POSTCLASSIFICATION
#include "PostClassificationAction.h"
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_RASTEREDIT
#include "RasterEditAction.h"
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_RASTERSLICING
  #include "RasterSlicingAction.h"
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_RASTERIZATION
  #include "RasterizationAction.h"
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_SEGMENTER
  #include "SegmenterAction.h"
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_REGISTER
  #include "RegisterAction.h"
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_VECTORIZATION
#include "VectorizationAction.h"
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_BOUNDARYEXTRACTION
  #include "BoundaryExtractionAction.h"
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_RASTERSTATISTICS
  #include "RasterStatisticsAction.h"
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_RASTERTRANSFORM
  #include "RasterTransformAction.h"
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_RESTORATION
  #include "RestorationAction.h"
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_RADAR_MULTISOURCECLASSIFIER
  #include "radar/MultiSourceClassifierAction.h"
#endif

// QT
#include <QAction>
#include <QMenu>
#include <QMenuBar>

te::qt::plugins::rp::Plugin::Plugin(const te::core::PluginInfo& pluginInfo)
    : te::core::CppPlugin(pluginInfo), m_rpMenu(nullptr)
{
}

te::qt::plugins::rp::Plugin::~Plugin() = default;

void te::qt::plugins::rp::Plugin::startup()
{
  if(m_initialized)
    return;

  te::qt::af::AppCtrlSingleton::getInstance().addListener(this, te::qt::af::SENDER);

// it initializes the Translator support for the TerraLib RP Qt Plugin
  //TE_ADD_TEXT_DOMAIN(TE_QT_PLUGIN_RP_TEXT_DOMAIN, TE_QT_PLUGIN_RP_TEXT_DOMAIN_DIR, "UTF-8");

  TE_LOG_TRACE(TE_TR("TerraLib Qt RP Plugin startup!"));

// add plugin menu

  QMenu* pluginMenu = te::qt::af::AppCtrlSingleton::getInstance().getMenu("Processing");

  m_rpMenu = new QMenu(pluginMenu);
  m_rpMenu->setIcon(QIcon::fromTheme("rp-rasterprocessing-icon"));

  m_rpRadarMenu = new QMenu(pluginMenu);
  m_rpRadarMenu->setIcon(QIcon::fromTheme("rp-radar-rasterprocessing-icon"));

  // Insert action before plugin manager action

  QAction* pluginsSeparator = te::qt::af::AppCtrlSingleton::getInstance().findAction("ManagePluginsSeparator");

  pluginMenu->insertMenu(pluginsSeparator, m_rpMenu);
  pluginMenu->insertMenu(pluginsSeparator, m_rpRadarMenu);

  m_rpMenu->setTitle(tr("Raster Processing"));
  m_rpRadarMenu->setTitle(tr("Radar Raster Processing"));

  // add pop up menu

  m_popupActionRP = new QAction(m_rpMenu);
  m_popupActionRP->setText(tr("Raster Processing"));

  m_popupActionRPRadar = new QAction(m_rpRadarMenu);
  m_popupActionRPRadar->setText(tr("Radar Raster Processing"));

//add to application layer tree pop up menu
  //te::qt::af::evt::LayerPopUpAddAction evt(m_popupAction, "RASTER_LAYER_ITEM", 2 /*UNIQUE_ITEM_SELECTED*/);
  //te::qt::af::AppCtrlSingleton::getInstance().trigger(&evt);

  m_rpPopupMenu = new QMenu(m_rpMenu);
  m_popupActionRP->setMenu(m_rpPopupMenu);

  m_rpRadarPopupMenu = new QMenu(m_rpRadarMenu);
  m_popupActionRPRadar->setMenu(m_rpRadarPopupMenu);

// register actions
  registerActions();

  RasterConfigWidgetFactory::initialize();

  m_initialized = true;
}

void te::qt::plugins::rp::Plugin::shutdown()
{
  if(!m_initialized)
    return;

//remove from application layer tree pop up menu
  //te::qt::af::evt::LayerPopUpRemoveAction evt(m_popupAction);
  //te::qt::af::AppCtrlSingleton::getInstance().trigger(&evt);

// remove menu
  delete m_rpMenu;

// unregister actions
  unRegisterActions();

  RasterConfigWidgetFactory::finalize();

  TE_LOG_TRACE(TE_TR("TerraLib Qt RP Plugin shutdown!"));

  m_initialized = false;
}

void te::qt::plugins::rp::Plugin::registerActions()
{
#ifdef TE_QT_PLUGIN_RP_HAVE_ARITHMETICOP
  m_arithmeticOp = new te::qt::plugins::rp::ArithmeticOpAction(m_rpMenu, m_rpPopupMenu);
  te::qt::af::AddActionToCustomToolbars(&te::qt::af::AppCtrlSingleton::getInstance(), m_arithmeticOp->getAction());
  connect(m_arithmeticOp, SIGNAL(triggered(te::qt::af::evt::Event*)), SIGNAL(triggered(te::qt::af::evt::Event*)));
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_CLASSIFIER
    m_classifier = new te::qt::plugins::rp::ClassifierAction(m_rpMenu, m_rpPopupMenu);
    te::qt::af::AddActionToCustomToolbars(&te::qt::af::AppCtrlSingleton::getInstance(), m_classifier->getAction());
    connect(m_classifier, SIGNAL(triggered(te::qt::af::evt::Event*)), SIGNAL(triggered(te::qt::af::evt::Event*)));
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_CLIPPING
    m_clipping = new te::qt::plugins::rp::ClippingAction(m_rpMenu, m_rpPopupMenu);
    te::qt::af::AddActionToCustomToolbars(&te::qt::af::AppCtrlSingleton::getInstance(), m_clipping->getAction());
    connect(m_clipping, SIGNAL(triggered(te::qt::af::evt::Event*)), SIGNAL(triggered(te::qt::af::evt::Event*)));
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_CLOUDDETECTION
    m_cloudDetection = new te::qt::plugins::rp::CloudDetectionAction(m_rpMenu, m_rpPopupMenu);
    te::qt::af::AddActionToCustomToolbars(&te::qt::af::AppCtrlSingleton::getInstance(), m_cloudDetection->getAction());
    connect(m_cloudDetection, SIGNAL(triggered(te::qt::af::evt::Event*)), SIGNAL(triggered(te::qt::af::evt::Event*)));
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_COLORTRANSFORM
    m_colorTrans = new te::qt::plugins::rp::ColorTransformAction(m_rpMenu, m_rpPopupMenu);
    te::qt::af::AddActionToCustomToolbars(&te::qt::af::AppCtrlSingleton::getInstance(), m_colorTrans->getAction());
    connect(m_colorTrans, SIGNAL(triggered(te::qt::af::evt::Event*)), SIGNAL(triggered(te::qt::af::evt::Event*)));
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_COMPOSEBANDS
   m_composeBands = new te::qt::plugins::rp::ComposeBandsAction(m_rpMenu, m_rpPopupMenu);
   te::qt::af::AddActionToCustomToolbars(&te::qt::af::AppCtrlSingleton::getInstance(), m_composeBands->getAction());
   connect(m_composeBands, SIGNAL(triggered(te::qt::af::evt::Event*)), SIGNAL(triggered(te::qt::af::evt::Event*)));
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_CONTRAST
    m_contrast = new te::qt::plugins::rp::ContrastAction(m_rpMenu, m_rpPopupMenu);
    te::qt::af::AddActionToCustomToolbars(&te::qt::af::AppCtrlSingleton::getInstance(), m_contrast->getAction());
    connect(m_contrast, SIGNAL(triggered(te::qt::af::evt::Event*)), SIGNAL(triggered(te::qt::af::evt::Event*)));
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_FILTER
    m_filter = new te::qt::plugins::rp::FilterAction(m_rpMenu, m_rpPopupMenu);
    te::qt::af::AddActionToCustomToolbars(&te::qt::af::AppCtrlSingleton::getInstance(), m_filter->getAction());
    connect(m_filter, SIGNAL(triggered(te::qt::af::evt::Event*)), SIGNAL(triggered(te::qt::af::evt::Event*)));
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_FUSION
    m_fusion = new te::qt::plugins::rp::FusionAction(m_rpMenu, m_rpPopupMenu);
    te::qt::af::AddActionToCustomToolbars(&te::qt::af::AppCtrlSingleton::getInstance(), m_fusion->getAction());
    connect(m_fusion, SIGNAL(triggered(te::qt::af::evt::Event*)), SIGNAL(triggered(te::qt::af::evt::Event*)));
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_MAJORITYFILTER
    m_majorityFilter = new te::qt::plugins::rp::MajorityFilterAction(m_rpMenu, m_rpPopupMenu);
    te::qt::af::AddActionToCustomToolbars(&te::qt::af::AppCtrlSingleton::getInstance(), m_majorityFilter->getAction());
    connect(m_majorityFilter, SIGNAL(triggered(te::qt::af::evt::Event*)), SIGNAL(triggered(te::qt::af::evt::Event*)));
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_MIXTUREMODEL
  m_mixtureModel = new te::qt::plugins::rp::MixtureModelAction(m_rpMenu, m_rpPopupMenu);
  te::qt::af::AddActionToCustomToolbars(&te::qt::af::AppCtrlSingleton::getInstance(), m_mixtureModel->getAction());
  connect(m_mixtureModel, SIGNAL(triggered(te::qt::af::evt::Event*)), SIGNAL(triggered(te::qt::af::evt::Event*)));
#endif

  #ifdef TE_QT_PLUGIN_RP_HAVE_MOSAIC
  m_mosaic = new te::qt::plugins::rp::MosaicAction(m_rpMenu, m_rpPopupMenu);
  te::qt::af::AddActionToCustomToolbars(&te::qt::af::AppCtrlSingleton::getInstance(), m_mosaic->getAction());
  connect(m_mosaic, SIGNAL(triggered(te::qt::af::evt::Event*)), SIGNAL(triggered(te::qt::af::evt::Event*)));
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_PALETTE
  m_palette = new te::qt::plugins::rp::PaletteAction(m_rpMenu, m_rpPopupMenu);
  te::qt::af::AddActionToCustomToolbars(&te::qt::af::AppCtrlSingleton::getInstance(), m_palette->getAction());
  connect(m_palette, SIGNAL(triggered(te::qt::af::evt::Event*)), SIGNAL(triggered(te::qt::af::evt::Event*)));
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_PCA
  m_PCA = new te::qt::plugins::rp::PCAAction(m_rpMenu, m_rpPopupMenu);
  te::qt::af::AddActionToCustomToolbars(&te::qt::af::AppCtrlSingleton::getInstance(), m_PCA->getAction());
  connect(m_PCA, SIGNAL(triggered(te::qt::af::evt::Event*)), SIGNAL(triggered(te::qt::af::evt::Event*)));
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_POSTCLASSIFICATION
  m_postClass = new te::qt::plugins::rp::PostClassificationAction(m_rpMenu, m_rpPopupMenu);
  te::qt::af::AddActionToCustomToolbars(&te::qt::af::AppCtrlSingleton::getInstance(), m_postClass->getAction());
  connect(m_postClass, SIGNAL(triggered(te::qt::af::evt::Event*)), SIGNAL(triggered(te::qt::af::evt::Event*)));
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_RASTEREDIT
  m_rasterEdit = new te::qt::plugins::rp::RasterEditAction(m_rpMenu, m_rpPopupMenu);
  te::qt::af::AddActionToCustomToolbars(&te::qt::af::AppCtrlSingleton::getInstance(), m_rasterEdit->getAction());
  connect(m_rasterEdit, SIGNAL(triggered(te::qt::af::evt::Event*)), SIGNAL(triggered(te::qt::af::evt::Event*)));
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_RASTERSLICING
  m_rasterSlicing = new te::qt::plugins::rp::RasterSlicingAction(m_rpMenu, m_rpPopupMenu);
  te::qt::af::AddActionToCustomToolbars(&te::qt::af::AppCtrlSingleton::getInstance(), m_rasterSlicing->getAction());
  connect(m_rasterSlicing, SIGNAL(triggered(te::qt::af::evt::Event*)), SIGNAL(triggered(te::qt::af::evt::Event*)));
#endif  
  
#ifdef TE_QT_PLUGIN_RP_HAVE_RASTERIZATION
  m_raster = new te::qt::plugins::rp::RasterizationAction(m_rpMenu, m_rpPopupMenu);
  te::qt::af::AddActionToCustomToolbars(&te::qt::af::AppCtrlSingleton::getInstance(), m_raster->getAction());
  connect(m_raster, SIGNAL(triggered(te::qt::af::evt::Event*)), SIGNAL(triggered(te::qt::af::evt::Event*)));
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_REGISTER
  m_register = new te::qt::plugins::rp::RegisterAction(m_rpMenu, m_rpPopupMenu);
  te::qt::af::AddActionToCustomToolbars(&te::qt::af::AppCtrlSingleton::getInstance(), m_register->getAction());
  connect(m_register, SIGNAL(triggered(te::qt::af::evt::Event*)), SIGNAL(triggered(te::qt::af::evt::Event*)));
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_SEGMENTER
    m_segmenter = new te::qt::plugins::rp::SegmenterAction(m_rpMenu, m_rpPopupMenu);
    te::qt::af::AddActionToCustomToolbars(&te::qt::af::AppCtrlSingleton::getInstance(), m_segmenter->getAction());
    connect(m_segmenter, SIGNAL(triggered(te::qt::af::evt::Event*)), SIGNAL(triggered(te::qt::af::evt::Event*)));
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_VECTORIZATION
    m_vector = new te::qt::plugins::rp::VectorizationAction(m_rpMenu, m_rpPopupMenu);
    te::qt::af::AddActionToCustomToolbars(&te::qt::af::AppCtrlSingleton::getInstance(), m_vector->getAction());
    connect(m_vector, SIGNAL(triggered(te::qt::af::evt::Event*)), SIGNAL(triggered(te::qt::af::evt::Event*)));
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_BOUNDARYEXTRACTION
  m_boundary = new te::qt::plugins::rp::BoundaryExtractionAction(m_rpMenu, m_rpPopupMenu);
  te::qt::af::AddActionToCustomToolbars(&te::qt::af::AppCtrlSingleton::getInstance(), m_boundary->getAction());
  connect(m_boundary, SIGNAL(triggered(te::qt::af::evt::Event*)), SIGNAL(triggered(te::qt::af::evt::Event*)));
#endif
  
#ifdef TE_QT_PLUGIN_RP_HAVE_RASTERSTATISTICS
  m_rasterStats = new te::qt::plugins::rp::RasterStatisticsAction(
    m_rpMenu, m_rpPopupMenu);
  te::qt::af::AddActionToCustomToolbars(
    &te::qt::af::AppCtrlSingleton::getInstance(), 
    m_rasterStats->getAction());
  connect(m_rasterStats, SIGNAL(triggered(te::qt::af::evt::Event*)), 
    SIGNAL(triggered(te::qt::af::evt::Event*)));
#endif  
  
#ifdef TE_QT_PLUGIN_RP_HAVE_RASTERTRANSFORM
  m_rasterTransf = new te::qt::plugins::rp::RasterTransformAction(
    m_rpMenu, m_rpPopupMenu);
  te::qt::af::AddActionToCustomToolbars(
    &te::qt::af::AppCtrlSingleton::getInstance(), 
    m_rasterTransf->getAction());
  connect(m_rasterTransf, SIGNAL(triggered(te::qt::af::evt::Event*)), 
    SIGNAL(triggered(te::qt::af::evt::Event*)));
#endif    
  
#ifdef TE_QT_PLUGIN_RP_HAVE_RESTORATION
  m_restoration = new te::qt::plugins::rp::RestorationAction(
    m_rpMenu, m_rpPopupMenu);
  te::qt::af::AddActionToCustomToolbars(
    &te::qt::af::AppCtrlSingleton::getInstance(), 
    m_restoration->getAction());
  connect(m_restoration, SIGNAL(triggered(te::qt::af::evt::Event*)), 
    SIGNAL(triggered(te::qt::af::evt::Event*)));
#endif    

#ifdef TE_QT_PLUGIN_RP_HAVE_RADAR_MULTISOURCECLASSIFIER
  m_multiSourceClassifier = new te::qt::plugins::rp::MultiSourceClassifierAction(
    m_rpRadarMenu, m_rpRadarPopupMenu);
  te::qt::af::AddActionToCustomToolbars(
    &te::qt::af::AppCtrlSingleton::getInstance(),
    m_multiSourceClassifier->getAction());
  connect(m_multiSourceClassifier, SIGNAL(triggered(te::qt::af::evt::Event*)),
    SIGNAL(triggered(te::qt::af::evt::Event*)));
#endif
  
}

void  te::qt::plugins::rp::Plugin::unRegisterActions()
{
#ifdef TE_QT_PLUGIN_RP_HAVE_ARITHMETICOP
    delete m_arithmeticOp;
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_CLASSIFIER
    delete m_classifier;
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_CLIPPING
    delete m_clipping;
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_CLOUDDETECTION
    delete m_cloudDetection;
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_COLORTRANSFORM
    delete m_colorTrans;
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_COMPOSEBANDS
    delete m_composeBands;
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_CONTRAST
    delete m_contrast;
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_FILTER
    delete m_filter;
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_FUSION
    delete m_fusion;
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_MIXTUREMODEL
    delete m_mixtureModel;
#endif

    #ifdef TE_QT_PLUGIN_RP_HAVE_MOSAIC
    delete m_mosaic;
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_PCA
    delete m_PCA;
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_POSTCLASSIFICATION
    delete m_postClass;
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_RASTEREDIT
    delete m_rasterEdit;
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_RASTERSLICING
    delete m_rasterSlicing;
#endif    

#ifdef TE_QT_PLUGIN_RP_HAVE_RASTERIZATION
    delete m_raster;
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_REGISTER
  delete m_register;
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_SEGMENTER
    delete m_segmenter;
#endif

#ifdef TE_QT_PLUGIN_RP_HAVE_VECTORIZATION
  delete m_vector;
#endif
  
#ifdef TE_QT_PLUGIN_RP_HAVE_RASTERSTATISTICS
  delete m_rasterStats;
#endif  
  
#ifdef TE_QT_PLUGIN_RP_HAVE_RASTERTRANSFORM
  delete m_rasterTransf;
#endif    
  
#ifdef TE_QT_PLUGIN_RP_HAVE_RESTORATION
  delete m_restoration;
#endif      

#ifdef TE_QT_PLUGIN_RP_HAVE_RADAR_MULTISOURCECLASSIFIER
  delete m_multiSourceClassifier;
#endif
}

TERRALIB_PLUGIN_CALL_BACK_IMPL(te::qt::plugins::rp::Plugin)
