/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file RasterConfigWidget.cpp

  \brief A frame for setting raster processing options.
*/

// TerraLib
#include "../../../common/SystemApplicationSettings.h"
#include "../../../common/UserApplicationSettings.h"
#include "../../../core/utils/Platform.h"
#include "../../../srs/Config.h"
#include "../../../raster/Raster.h"
#include "../../af/ApplicationController.h"
#include "../../af/Utils.h"
#include "RasterConfigWidget.h"
#include "ui/ui_RasterConfigWidgetForm.h"


te::qt::plugins::rp::RasterConfigWidget::RasterConfigWidget(QWidget* parent)
  : AbstractSettingWidget(parent),
    m_ui(new Ui::RasterConfigWidgetForm)
{
  m_resumeText = tr("Changes the raster processing configurations of the application.");

  m_ui->setupUi(this);

  fillInterpolator();

  initialize();

  if(!m_ui->m_questionCheckBox->isChecked() && !m_ui->m_defaultCheckBox->isChecked())
    m_ui->m_defaultCheckBox->setEnabled(false);
}

te::qt::plugins::rp::RasterConfigWidget::~RasterConfigWidget() = default;

void te::qt::plugins::rp::RasterConfigWidget::saveChanges()
{
  QSettings settings(QSettings::IniFormat, QSettings::UserScope, qApp->organizationName(), qApp->applicationName());

  settings.setValue("raster_config/default_interpolator", m_ui->m_interpolatorComboBox->itemData(m_ui->m_interpolatorComboBox->currentIndex()).toInt());
  settings.setValue("raster_config/default_level", m_ui->m_levelSpinBox->value());
  settings.setValue("raster_config/default_question", m_ui->m_questionCheckBox->isChecked());
  settings.setValue("raster_config/default_option", m_ui->m_defaultCheckBox->isChecked());

  changeApplyButtonState(false);
}

void te::qt::plugins::rp::RasterConfigWidget::resetState()
{
}

void te::qt::plugins::rp::RasterConfigWidget::getHelpInformations(QString& ns, QString& helpFile)
{
  ns = "dpi.inpe.br.apf";
  helpFile = "";
}

void te::qt::plugins::rp::RasterConfigWidget::initialize()
{
  QSettings set(QSettings::IniFormat, QSettings::UserScope, qApp->organizationName(), qApp->applicationName());

  set.beginGroup("raster_config");
  set.endGroup();

  // Get user settings
  QSettings settings(QSettings::IniFormat, QSettings::UserScope, QApplication::instance()->organizationName(), QApplication::instance()->applicationName());

  int interpolator = settings.value("raster_config/default_interpolator", te::common::SystemApplicationSettings::getInstance().getValue("Application.rasterConfigDefaultInterpolator").c_str()).toInt();
  int level = settings.value("raster_config/default_level", te::common::SystemApplicationSettings::getInstance().getValue("Application.rasterConfigDefaultLevel").c_str()).toInt();
  bool question = settings.value("raster_config/default_question", te::common::SystemApplicationSettings::getInstance().getValue("Application.rasterConfigDefaultQuestion").c_str()).toBool();
  bool option = settings.value("raster_config/default_option", te::common::SystemApplicationSettings::getInstance().getValue("Application.rasterConfigDefaultOption").c_str()).toBool();

  m_ui->m_interpolatorComboBox->setCurrentIndex(m_ui->m_interpolatorComboBox->findData(interpolator));
  m_ui->m_levelSpinBox->setValue(level);
  m_ui->m_questionCheckBox->setChecked(question);
  m_ui->m_defaultCheckBox->setChecked(option);

  // Signals & slots
  connect(m_ui->m_interpolatorComboBox, SIGNAL(currentIndexChanged(int)), SLOT(onInterpolatorComboBoxChanged(int)));
  connect(m_ui->m_levelSpinBox, SIGNAL(valueChanged(int)), SLOT(onLevelChanged(int)));
  connect(m_ui->m_questionCheckBox, SIGNAL(toggled(bool)), SLOT(onQuestionChecked(bool)));
  connect(m_ui->m_defaultCheckBox, SIGNAL(toggled(bool)), SLOT(onSetMultiResolutionDefaultChecked(bool)));
}

void te::qt::plugins::rp::RasterConfigWidget::fillInterpolator()
{
  m_ui->m_interpolatorComboBox->clear();

  m_ui->m_interpolatorComboBox->addItem(tr("Nearest Neighbor"), te::rst::NearestNeighbor);
  m_ui->m_interpolatorComboBox->addItem(tr("Bilinear"), te::rst::Bilinear);
  m_ui->m_interpolatorComboBox->addItem(tr("Bicubic"), te::rst::Bicubic);
}

void te::qt::plugins::rp::RasterConfigWidget::onInterpolatorComboBoxChanged(
    int /*index*/)
{
  changeApplyButtonState(true);
}

void te::qt::plugins::rp::RasterConfigWidget::onLevelChanged(int /*value*/)
{
  changeApplyButtonState(true);
}

void te::qt::plugins::rp::RasterConfigWidget::onQuestionChecked(bool checked)
{
  if(checked)
  {
    m_ui->m_defaultCheckBox->setEnabled(true);
    m_ui->m_defaultCheckBox->setChecked(false);
  }
  else if(!m_ui->m_defaultCheckBox->isChecked())
    m_ui->m_defaultCheckBox->setEnabled(false);

  changeApplyButtonState(true);
}

void te::qt::plugins::rp::RasterConfigWidget::onSetMultiResolutionDefaultChecked(bool checked)
{
  if(checked)
    m_ui->m_questionCheckBox->setChecked(false);
  else if(!m_ui->m_questionCheckBox->isChecked())
    m_ui->m_defaultCheckBox->setEnabled(false);

  changeApplyButtonState(true);
}
