#include "DrawThread.h"

// TerraLib
#include "../Utils.h"
#include "../../../core/logger/Logger.h"
#include "../../../core/translator/Translator.h"
#include "../../../common/Exception.h"
#include "../../../common/StringUtils.h"
#include "../../../geometry/Envelope.h"
#include "../../../maptools/AbstractLayer.h"
#include "../../../srs/Config.h"
#include "Canvas.h"

// STL
#include <ctime>

te::qt::widgets::DrawThread::DrawThread(QPaintDevice* dev,
                                        te::map::AbstractLayer* layer,
                                        te::gm::Envelope* env,
                                        const QColor& bckGround, int srid,
                                        double scale, te::map::AlignType hAlign,
                                        te::map::AlignType vAlign)
    :

      m_device(dev),
      m_layer(layer),
      m_envelope(env),
      m_bckGround(bckGround),
      m_srid(srid),
      m_scale(scale),
      m_hAlign(hAlign),
      m_vAlign(vAlign),
      m_cancel(false)
{
  setAutoDelete(false);
}

te::qt::widgets::DrawThread::~DrawThread() = default;

void te::qt::widgets::DrawThread::run()
{
  m_finished = false;

  try
  {
    time_t startTime, endTime;

    time(&startTime);

    Canvas canvas(m_device);
    canvas.clear();

    m_layer->draw(&canvas, *m_envelope, m_srid, m_scale, &m_cancel);

    time(&endTime);

    // difference in seconds
    double diffInSec = difftime(endTime, startTime);

    std::string strTime = TE_TR("The layer ");
                strTime += m_layer->getTitle();
                strTime += TE_TR(" has been drawn in ");
                strTime += te::common::Convert2String(diffInSec, 2);
                strTime += TE_TR(" seconds");

    TE_LOG_INFO(strTime);

    if(!m_cancel)
      m_finished = true;
  }
  catch(const te::common::Exception& e)
  {
    if(e.code() == te::common::NO_CONNECTION_AVAILABLE)
    {
      //try again
      //msleep(100);
    }
    else
      m_errorMessage = QString(tr("The layer") + " %1 " + tr("could not be drawn! Details:") + " %2").arg(m_layer->getTitle().c_str()).arg(e.what());
  }
  catch(const std::exception& e)
  {
    m_errorMessage = QString(tr("The layer") + " %1 " + tr("could not be drawn! Details:") + " %2").arg(m_layer->getTitle().c_str()).arg(e.what());
  }
  catch(...)
  {
    m_errorMessage = QString(tr("The layer") + " %1 " + tr("could not be drawn!")).arg(m_layer->getTitle().c_str());
  }

  if (!m_errorMessage.isEmpty())
    TE_LOG_INFO(m_errorMessage.toUtf8().data());

  emit finished();
}

bool te::qt::widgets::DrawThread::hasFinished() const
{
  return m_finished;
}

QString te::qt::widgets::DrawThread::errorMessage() const
{
  return m_errorMessage;
}

QString te::qt::widgets::DrawThread::layerId() const
{
  return QString::fromUtf8(m_layer->getId().c_str());
}
