/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

This file is part of the TerraLib - a Framework for building GIS enabled applications.

TerraLib is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

TerraLib is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with TerraLib. See COPYING. If not, write to
TerraLib Team at <terralib-team@terralib.org>.
*/

/*!
 * \file canvas/Grid.h
 *
 * \brief Contains an implementation of a grid.
 */

#ifndef __TERRALIB_QT_WIDGETS_INTERNAL_GRID_H
#define __TERRALIB_QT_WIDGETS_INTERNAL_GRID_H

//TerraLib
#include "../Config.h"

// Qt
#include <QFrame>

// STL
#include <vector>

// Forward Declarations
class QPixmap;

namespace te
{
  namespace gm
  {
    // Forward Declarations
    class LineString;
  }

  namespace qt
  {
    namespace widgets
    {
      // Forward declaration
      class MapDisplay;

      /*!
       * \enum GridTypes
       *
       * \brief Defines the types of grid that can be used.
       */
      enum GridTypes
      {
        PLANAR,     //!< Planar grid.
        GEOGRAPHIC  //!< Geographic grid.
      };

      /*!
       * \class Grid
       *
       * \brief This class implements the grid, planar or geographic, to be used with the MapDisplay
       *
       * \ingroup widgets
       */
      class TEQTWIDGETSEXPORT Grid : public QFrame
      {
        Q_OBJECT

      public:

        /** @name Constructor and destructor methods.
        *  Methods related to instantiation and destruction.
        */
        //@{

        /*!
         * \brief Constructor.
         *
         * \param parent The display over what the scale will reflect.
         */
        Grid(MapDisplay* parent);

        /*!
         * \brief Destuctor.
         */
        ~Grid();
        //@}

        /** @name Public methods
        *  Methods to access the internal attributes of the class.
        */
        //@{

        /*!
         * \brief Returns the current color being used.
         *
         * \return The current color of the grid and texts.
         */
        QColor getColor() const;

        /*!
         * \brief Returns the current font being used.
         *
         * \return The current font of the texts.
         */
        QFont getFont() const;

        /*!
         * \brief Returns the current type of grid being used.
         *
         * \return The current type of grid being used.
         *
         * \sa GridTypes
         */
        int getType() const;
        //@}

      public slots:

        /** @name Public slots.
        *  Methods to update the internal attributes of the class, using the signal-slot mechanism of QT.
        */
        //@{

        /*!
         * \brief Updates the color of the lines and texts.
         *
         * \param color The new color to be used.
         */
        void setColor(const QColor& color);

        /*!
         * \brief Updates the font of the texts.
         *
         * \param font New font to be used.
         */
        void setFont(const QFont& font);

        /*!
         * \brief Updates the type of grid: planar or geographic.
         * \param type The new type.
         */
        void setGridType(const int& type);

        /*!
         * \brief Refresh the grid image. This method MUST BE called when the area of visualization changes in MapDisplay.
         */
        void redraw();
        //@}

      signals:

        /** @name Signals
        *  Signals emmited by the widget.
        */
        //@{

        /*!
         * Signal emmited when the widget is about to be closed.
         */
        void aboutToBeclosed();
        //@}

        /** @name Protected methods
        *  Methods used internally by the object.
        */
        //@{

      protected:
        /*!
         * \brief Initializes the object.
         */
        void initialize();

        /*!
         * \brief Builds the planar grid.
         */
        void makePlanarGrid();

        /*!
         * \brief Builds the geographic grid.
         */
        void makeGeographicGrid();

        /*!
         * \brief Reimplementation of the QFrame method.
         *
         * \param event The event generated.
         */
        void paintEvent(QPaintEvent* event);

        /*!
         * \brief Reimplementation of the QFrame method.
         *
         * \param event The event generated.
         */
        void closeEvent(QCloseEvent *event);
        //@}

        MapDisplay* m_display;                        //!< Pointer to a MapDisplay.
        int m_maxGap;                                 //!< Maximum number of lines to be presented on the grid.
        double m_verticalGap;                         //!< Distance between lines in y-axis.
        double m_horizontalGap;                       //!< Distance between lines in x-axis.
        double m_initialX;                            //!< Initial coordinate in x-axis.
        double m_initialY;                            //!< Initial coordinate in y-axis.
        QPixmap* m_backGround;                        //!< Pixmap with the grid image.
        std::vector<te::gm::LineString> m_hlines;     //!< Grid lines.
        std::vector<te::gm::LineString> m_vlines;     //!< Grid lines.
        int m_srid;                                   //!< SRID being used to reproject the world box to a planar coordinate system.
        int m_geoSrid;                                //!< Geographic SRID being used to reproject the world box to a geographic coordinate system.
        QColor m_color;                               //!< Color to be used.
        QFont m_font;                                 //!< Font to be used.
        GridTypes m_type;                             //!< Type of the grid (planar or geographic).
      };
    }
  }
}

#endif //__TERRALIB_QT_WIDGETS_INTERNAL_GRID_H
