/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

This file is part of the TerraLib - a Framework for building GIS enabled applications.

TerraLib is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

TerraLib is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with TerraLib. See COPYING. If not, write to
TerraLib Team at <terralib-team@terralib.org>.
*/

/*!
 * \file widgets/canvas/GridPropertyWidget.h
 *
 * \brief Defines a class that presents a widget with the grid properties to be edited.
 */

#ifndef __TERRALIB_QT_WIDGETS_INTERNAL_GRIDPROPERTYWIDGET_H
#define __TERRALIB_QT_WIDGETS_INTERNAL_GRIDPROPERTYWIDGET_H

//TerraLib
#include "../Config.h"
#include "../propertybrowser/AbstractPropertyItem.h"

// Qt
#include <QDockWidget>

//Forward declarations
class QAction;

namespace te
{
  namespace qt
  {
    namespace widgets
    {
      // Forward declaration
      class Grid;

      /*!
       * \class GridPropertyWidget
       *
       * \brief a QDockWidget that contains a property browser object with grid properties.
       *
       * \ingroup widgets
      */
      class TEQTWIDGETSEXPORT GridPropertyWidget : public QDockWidget
      {
        Q_OBJECT

      public:
        /** @name Constructor and destructor methods.
        *  Methods related to instantiation and destruction.
        */
        //@{

        /*!
         * \brief Constructor.
         *
         * \param wid A pointer to a graphic scale object.
         *
         * \param editAction The action used to show edit widget.
         */
        GridPropertyWidget(Grid* wid, QAction* editAction);

        /*!
         * \brief Destructor.
         */
        ~GridPropertyWidget();
        //@}

        /** @name Public slots.
        *  Methods to update the internal attributes of the class, using the signal-slot mechanism of QT.
        */
        //@{

      protected slots:

        /*!
         * \brief This function must be called when the Grid related to it was closed.
         */
        void gridClosed();
        //@}

      protected:

        QAction* m_editAction;   //!< Action used to show the widget.
        bool m_gridStatus;       //!< Grid widget enabling status.
      };


      /*!
       * \class GridPropertyBrowser
       *
       * \brief Widget with propeties of a Grid for the users to change the presentation of the grid on the display.
       *
       * \ingroup widgets
       */
      class GridPropertyBrowser : public AbstractPropertyItem
      {
        Q_OBJECT

      public:

        /** @name Constructor and destructor methods.
        *  Methods related to instantiation and destruction.
        */
        //@{

        /*!
         * \brief Constructor.
         *
         * \param pb Property browser.
         *
         * \param wid ScaleWidget object to be used.
         */
        GridPropertyBrowser(QtTreePropertyBrowser* pb, Grid* wid);

        /*!
         * \brief Destructor.
         */
        ~GridPropertyBrowser();
        //@}

      protected slots:

        /** @name Protected slots.
        *  Methods to update the internal attributes of the class, using the signal-slot mechanism of QT.
        */
        //@{

        /*!
         * \brief Called when a property of color type changed.
         *
         * \param p The property that was changed.
         *
         * \param value The new value.
         */
        void valueChanged(QtProperty* p, const QColor& value);

        /*!
         * \brief Called when a property of font type changed.
         *
         * \param p The property that was changed.
         *
         * \param value The new value.
         */
        void valueChanged(QtProperty* p, const QFont& value);

        /*!
         * \brief Called when a property of int type changed.
         *
         * \param p The property that was changed.
         *
         * \param value The new value.
         */
        void valueChanged(QtProperty* p, int value);
        //@}

      Q_SIGNALS :

        /** @name Signals.
        *  Emmited to the grid to sinalizes the attributes that was changed.
        */
        //@{

        /*!
         * \brief Emmited when the color was changed.
         */
        void colorChanged(const QColor&);

        /*!
         * \brief Emmited when the font was changed.
         */
        void fontChanged(const QFont&);

        /*!
         * \brief Emmited when the type was changed.
         */
        void gridTypeChanged(const int&);
        //@}
      };
    }
  }
}

#endif //__TERRALIB_QT_WIDGETS_INTERNAL_GRIDPROPERTYWIDGET_H
