/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

This file is part of the TerraLib - a Framework for building GIS enabled applications.

TerraLib is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

TerraLib is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with TerraLib. See COPYING. If not, write to
TerraLib Team at <terralib-team@terralib.org>.
*/

/*!
\file terralib/qt/widgets/canvas/ScalePropertyWidget.h

\brief Defines a class that presents a widget with the graphic scale properties to be edited.
*/

#ifndef __TERRALIB_QT_WIDGETS_INTERNAL_SCALEPROPERTYWIDGET_H
#define __TERRALIB_QT_WIDGETS_INTERNAL_SCALEPROPERTYWIDGET_H

//TerraLib
#include "../Config.h"
#include "../propertybrowser/AbstractPropertyItem.h"

// Qt
#include <QAction>
#include <QDockWidget>

namespace te
{
  namespace qt
  {
    namespace widgets
    {
      // Forward declaration
      class ScaleWidget;

      /*!
      \class ScalePropertyWidget
      
      \brief a QDockWidget that contains a property browser object with graphic scale properties.

      \ingroup widgets
      */
      class TEQTWIDGETSEXPORT ScalePropertyWidget : public QDockWidget
      {
        Q_OBJECT

      public:
        /** @name Constructor and destructor methods.
        *  Methods related to instantiation and destruction.
        */
        //@{

        /*!
        \brief Constructor.
        \param wid A pointer to a graphic scale object.
        \param parent The parent widget (Used by Qt to control the life cicle of the object).
        */
        ScalePropertyWidget(ScaleWidget* wid, QAction* editScale);

        /*!
        \brief Destructor.
        */
        ~ScalePropertyWidget();

      protected slots:

        /*!
         * \brief This function must be called when the Grid related to it was closed.
         */
        void scaleWidgetClosed();
        //@}

      protected:

        QAction* m_editAction;   //!< Action used to show the widget.
        bool m_scaleStatus;      //!< Scale widget enabling status.
      };


      /*!
      \class ScalePropertyBrowser

      \brief Widget with propeties of a ScaleWidget for the users to change the presentation of the graphic scale.

      \ingroup widgets
      */
      class ScalePropertyBrowser : public AbstractPropertyItem
      {
        Q_OBJECT

      public:

        /** @name Constructor and destructor methods.
        *  Methods related to instantiation and destruction.
        */
        //@{

        /*!
        \brief Constructor.
        \param pb Property browser.
        \param wid ScaleWidget object to be used.
        */
        ScalePropertyBrowser(QtTreePropertyBrowser* pb, ScaleWidget* wid);

        /*!
        \brief Destructor.
        */
        ~ScalePropertyBrowser();
        //@}

      protected slots:

      /** @name Protected slots.
      *  Methods to update the internal attributes of the class, using the signal-slot mechanism of QT.
      */
      //@{
  
        /*!
        \brief Called when a property of boolean type changed.
        \param p The property that was changed.
        \param value The new value.
        */
        void valueChanged(QtProperty* p, bool value);
      
        /*!
        \brief Called when a property of color type changed.
        \param p The property that was changed.
        \param value The new value.
        */
        void valueChanged(QtProperty* p, const QColor& value);

        /*!
        \brief Called when a property of font type changed.
        \param p The property that was changed.
        \param value The new value.
        */
        void valueChanged(QtProperty* p, const QFont& value);

        /*!
        \brief Called when a property of int type changed.
        \param p The property that was changed.
        \param value The new value.
        */
        void valueChanged(QtProperty* p, int value);
        //@}

      Q_SIGNALS :

        /** @name Signals.
        *  Emmited to the graphic scale to sinalizes the attributes that was changed.
        */
        //@{

        /*!
        \brief Emmited when the interaction mode was changed.
        */
        void interactHandleChanged(const bool&);

        /*!
        \brief Emmited when the header color was changed.
        */
        void headerColorChanged(const QColor&);

        /*!
        \brief Emmited when the ruler color was changed.
        */
        void rulerColorChanged(const QColor&);

        /*!
        \brief Emmited when the footer color was changed.
        */
        void footerColorChanged(const QColor&);

        /*!
        \brief Emmited when the header font was changed.
        */
        void headerFontChanged(const QFont&);

        /*!
        \brief Emmited when the footer font was changed.
        */
        void footerFontChanged(const QFont&);

        /*!
        \brief Emmited when the ruler type was changed.
        */
        void scaleTypeChanged(const int&);

        /*!
        \brief Emmited when the footer visibility was changed.
        */
        void numericScaleVisibilityChanged(const bool&);

        /*!
        \brief Emmited when the unit of measure of the graphic scale was changed.
        */
        void unitChanged(const QString&);

        /*!
        \brief Emmited when the visibility of the intermediary values of the header was changed.
        */
        void allIntermediaryValuesChanged(const bool&);

        /*!
        \brief Emmited when the height of the ruler was changed.
        */
        void rulerHeightChanged(const int&);

        /*!
        \brief Emmited when the number of intervals in the ruler was changed.
        */
        void numberOfIntervalsChanged(const int&);

        /*!
        \brief Emmited when the units per interval was changed.
        */
        void unitsPerGapChanged(const int&);
        //@}
      };
    }
  }
}

#endif //__TERRALIB_QT_WIDGETS_INTERNAL_SCALEPROPERTYWIDGET_H
