/*
  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

  This file is part of the TerraLib - a Framework for building GIS enabled
  applications.

  TerraLib is free software: you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation, either version 3 of the License,
  or (at your option) any later version.

  TerraLib is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with TerraLib. See COPYING. If not, write to
  TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/widgets/ceditor/IOWidget.cpp

  \brief A widget that can be used to do input and out
*/

// TerraLib
#include "IOWidget.h"

// QT
#include <QKeyEvent>
#include <QTextBlock>
#include <QTextCursor>
#include <QTimer>

te::qt::widgets::IOWidget::IOWidget(QWidget *parent)
    : QPlainTextEdit(parent), userPrompt(QString(""))
{
  setLineWrapMode(WidgetWidth);
  lastIndex = getIndex(this->textCursor());
}

void te::qt::widgets::IOWidget::keyPressEvent(QKeyEvent *e)
{
  switch(e->key())
  {
    case Qt::Key_Return:
    case Qt::Key_Enter:
      handleEnter();
      break;
    case Qt::Key_Backspace:
      handleLeft(e);
      break;
    case Qt::Key_Left:
      handleLeft(e);
      break;
    case Qt::Key_Home:
      handleHome();
      break;
    default:
      if(isWritable(this->textCursor()))
        QPlainTextEdit::keyPressEvent(e);
      break;
  }
}

void te::qt::widgets::IOWidget::handleEnter()
{
  QString cmd = getCommand();

  QPlainTextEdit::moveCursor(QTextCursor::EndOfLine);

  if(cmd.length() > 0)
  {
    setFocus();
    insertPlainText("\n");
    lastIndex = getIndex(this->textCursor());
    emit command(cmd);
  }
  else
  {
    appendPlainText(prompt());
    ensureCursorVisible();
  }
}

void te::qt::widgets::IOWidget::addPrompt()
{
  insertPlainText(userPrompt);
  ensureCursorVisible();
  QTimer::singleShot(0, this, SLOT(setFocus()));
  locked = false;
}

void te::qt::widgets::IOWidget::append(QString text)
{
  insertPlainText(text);
  ensureCursorVisible();
  lastIndex = getIndex(this->textCursor());
  locked = true;
}

void te::qt::widgets::IOWidget::clearLine()
{
  QTextCursor c = this->textCursor();
  c.select(QTextCursor::LineUnderCursor);
  c.removeSelectedText();
}

QString te::qt::widgets::IOWidget::getCommand() const
{
  QTextCursor c = this->textCursor();
  c.select(QTextCursor::LineUnderCursor);

  QString text = c.selectedText();
  text.remove(0, lastIndex + userPrompt.length());

  return text;
}

void te::qt::widgets::IOWidget::handleLeft(QKeyEvent *event)
{
  if(getIndex(textCursor()) > userPrompt.length())
  {
    QPlainTextEdit::keyPressEvent(event);
  }
}

void te::qt::widgets::IOWidget::handleHome()
{
  QTextCursor c = textCursor();
  c.movePosition(QTextCursor::StartOfLine);
  c.movePosition(QTextCursor::Right, QTextCursor::MoveAnchor,
                 userPrompt.length());
  setTextCursor(c);
}

int te::qt::widgets::IOWidget::getIndex(const QTextCursor &cursor)
{
  QTextBlock b;
  int column = 1;
  b = cursor.block();
  column = cursor.position() - b.position();
  return column;
}

void te::qt::widgets::IOWidget::setPrompt(const QString &prompt)
{
  userPrompt = prompt;
  clearLine();
}

QString te::qt::widgets::IOWidget::prompt() const
{
  return userPrompt;
}

bool te::qt::widgets::IOWidget::isWritable(const QTextCursor &cursor)
{
  QTextCursor tempCursor = textCursor();
  tempCursor.movePosition(QTextCursor::End);
  return (getIndex(cursor) >= prompt().length()) &&
         (cursor.blockNumber() + 1) == (tempCursor.blockNumber() + 1) &&
         !locked;
}
