/*
  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

  This file is part of the TerraLib - a Framework for building GIS enabled
  applications.

  TerraLib is free software: you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation, either version 3 of the License,
  or (at your option) any later version.

  TerraLib is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with TerraLib. See COPYING. If not, write to
  TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/widgets/ceditor/IOWidget.h

  \brief A widget that can be used to do input and out
*/

#ifndef __TERRALIB_QTWIDGETS_CEDITOR_IOWIDGET_H__
#define __TERRALIB_QTWIDGETS_CEDITOR_IOWIDGET_H__

// TerraLib
#include "../Config.h"

// QT
#include <QPlainTextEdit>

class QKeyEvent;
class QTextCursor;

namespace te
{
  namespace qt
  {
    namespace widgets
    {
      class TEQTWIDGETSEXPORT IOWidget : public QPlainTextEdit
      {
        Q_OBJECT

       public:
        IOWidget(QWidget *parent = 0);

        /*!
          \brief prompt Retrieves the current prompt.
         */
        QString prompt() const;

        /*!
          \brief setPrompt Sets a new prompt from a given string.
          \param prompt String of the new prompt
         */
        void setPrompt(const QString &prompt);

       protected:
        void keyPressEvent(QKeyEvent *e);

       private:
        /*!
          \brief handleLeft Handles Left arrow key press.
          \param event The received event.
         */
        void handleLeft(QKeyEvent *event);

        /*!
          \brief handleEnter Handles Enter and Return key press.
         */
        void handleEnter();

        /*!
          \brief handleHome Handles Home key press.
         */
        void handleHome();

        /*!
          \brief clearLine Clears the line under the cursor.
         */
        void clearLine();

        /*!
          \brief getCommand Gets the input from the user.
         */
        QString getCommand() const;

        /*!
          \brief isWritable Checks if a given cursor is at a writable position.
          \param cursor Cursor that will be checked.
         */
        bool isWritable(const QTextCursor &cursor);

        /*!
          \brief getIndex Gets the index of a given cursor.
          \param cursor Cursor used to retrieve the index.
         */
        int getIndex(const QTextCursor &cursor);

        /*!
          \brief userPrompt prompt to indicate that the text edit is waiting
           for input
         */
        QString userPrompt;

        /*!
          \brief locked False if the text edit can receive input, true
           otherwise.
         */
        bool locked;

        /*!
          \brief lastIndex The index of the last character.
         */
        int lastIndex;

       signals:
        /*!
          \brief command Signal emitted when input is entered.
          \param command String to be emitted.
         */
        void command(QString command);

       public slots:
        /*!
          \brief addPrompt Adds the prompt to the text edit and unlocks the
          widget for input
         */
        void addPrompt();

        /*!
          \brief append Appends the given text to the text edit and locks the
           widget for input
          \param text String to be appended
         */
        void append(QString text);
      };
    } // end namespace widgets
  }   // end namespace qt
}     // end namespace te
#endif // __TERRALIB_QTWIDGETS_CEDITOR_IOWIDGET_H__
