/*
  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

  This file is part of the TerraLib - a Framework for building GIS enabled
  applications.

  TerraLib is free software: you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation, either version 3 of the License,
  or (at your option) any later version.

  TerraLib is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with TerraLib. See COPYING. If not, write to
  TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/widgets/ceditor/ScriptWidget.cpp

  \brief A widget that can be used to edit a script.
*/

#include "ScriptWidget.h"
#include "Utils.h"

// TerraLib
#include "../../../Exception.h"
#include "../../../vm/core/VirtualMachineManager.h"

// boost
#include <boost/format.hpp>

// Qt
#include <QApplication>
#include <QFile>
#include <QFileInfo>
#include <QMessageBox>
#include <QTextStream>

te::qt::widgets::ScriptWidget::ScriptWidget(QWidget* parent)
    : QsciScintilla(parent), m_fileName(""), m_dirty(false)
{
  setAutoCompletionSource(QsciScintilla::AcsAll);
  setMarginLineNumbers(1, true);
  setFolding(QsciScintilla::PlainFoldStyle, 2);
  setAutoCompletionThreshold(1);
  QTimer::singleShot(0, this, SLOT(setFocus()));
}

te::qt::widgets::ScriptWidget::~ScriptWidget() = default;

QString te::qt::widgets::ScriptWidget::getScriptType() const
{
  if(m_fileName.isEmpty())
    return QString(tr("lang-unknown"));

  QFileInfo info(m_fileName);

  return info.suffix();
}

QString te::qt::widgets::ScriptWidget::getScriptName() const
{
  QFileInfo info(m_fileName);
  return info.baseName();
}

void te::qt::widgets::ScriptWidget::open(const QString& fileName)
{
  QFile file(fileName);

  if(!file.open(QFile::ReadOnly))
  {
    QMessageBox::warning(
        this, tr("TerraLib"),
        tr("Cannot read file %1:\n%2.").arg(fileName).arg(file.errorString()));
    return;
  }

  QTextStream ifile(&file);

  QApplication::setOverrideCursor(Qt::WaitCursor);

  setText(ifile.readAll());

  QApplication::restoreOverrideCursor();

  m_fileName = fileName;

  file.close();

  setLexer(te::qt::widgets::LexerFactory(getScriptType()));
}

void te::qt::widgets::ScriptWidget::save(const QString& fileName)
{
  QFile file(fileName);

  if(!file.open(QFile::WriteOnly))
  {
    throw InvalidArgumentException()
        << te::ErrorDescription(tr("Cannot write file %1:\n%2.")
                                    .arg(fileName)
                                    .arg(file.errorString())
                                    .toUtf8()
                                    .data());
  }

  m_fileName = fileName;

  QTextStream ofile(&file);

  QApplication::setOverrideCursor(Qt::WaitCursor);

  ofile << text();

  QApplication::restoreOverrideCursor();
  file.close();
  m_dirty = false;

  setLexer(te::qt::widgets::LexerFactory(getScriptType()));
}

void te::qt::widgets::ScriptWidget::setHighlightMode(const QString& mode)
{
  setLexer(te::qt::widgets::LexerFactory(mode));
}

void te::qt::widgets::ScriptWidget::setDirty(bool dirty)
{
  m_dirty = dirty;
}

void te::qt::widgets::ScriptWidget::insert(const QString &text)
{
  int line, pos;

  getCursorPosition(&line, &pos);

  QsciScintilla::insert(text);

  setCursorPosition(line, pos + text.size() + 1);
}
