/*
  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

  This file is part of the TerraLib - a Framework for building GIS enabled
  applications.

  TerraLib is free software: you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation, either version 3 of the License,
  or (at your option) any later version.

  TerraLib is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with TerraLib. See COPYING. If not, write to
  TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/widgets/ceditor/ScriptWidget.h

  \brief A widget that can be used to edit a script.
*/

#ifndef __TERRALIB_QTWIDGETS_CEDITOR_SCRIPTWIDGET_H__
#define __TERRALIB_QTWIDGETS_CEDITOR_SCRIPTWIDGET_H__

// TerraLib
#include "../../../vm/core/VirtualMachine.h"
#include "../Config.h"

// QScintilla
#include <Qsci/qsciscintilla.h>

// Qt
#include <QTextEdit>
#include <QWidget>

// Forward declaration
class QString;

namespace te
{
  namespace qt
  {
    namespace widgets
    {
      /*!
        \class ScriptWidget

        \brief A widget that can be used to edit a script.
       */
      class TEQTWIDGETSEXPORT ScriptWidget : public QsciScintilla
      {
        Q_OBJECT

       public:
        /*!
          \brief ScriptWidget

          \param parent
         */
        ScriptWidget(QWidget* parent = 0);

        /*! \brief Destructor. */
        ~ScriptWidget();

        /*!
          \brief Returns the file name of the script.

          \return File name.
         */
        QString getFileName() const
        {
          return m_fileName;
        }

        /*!
          \brief getScriptType

          \return Returns the type of the script based on its extension.
         */
        QString getScriptType() const;

        /*!
          \brief getScriptName

          \return Return the script name.
         */
        QString getScriptName() const;

        /*!
          \brief Open the script file and presents it on the text edit.

          \param fileName File name.
         */
        void open(const QString& fileName);

        /*!
          \brief Saves the content of the script into a file.

          \param fileName Name of the file.
         */
        void save(const QString& fileName);

        /*!
          \brief Sets the highlight mode for the script.

          \param mode Name of the mode.

          \note Currently supported modes are: py, lua, sql.
         */
        void setHighlightMode(const QString& mode);

        /*!
          \brief Sets the status of the widget. True if the text was modified,
          false otherwise.

          \param dirty Status to be set.
         */
        void setDirty(bool dirty);


       public slots:

        /*!
        \brief Insert the text \a text at the current position 
        and move foward the cursor position

        */
        virtual void insert(const QString &text);

       private:
        QString m_fileName;  //!< File name.
        bool m_dirty; //!< Text was changed status.
      };

    }  // end namespace widgets
  }    // end namespace qt
}  // end namespace te

#endif  // __TERRALIB_QTWIDGETS_CEDITOR_SCRIPTWIDGET_H__
