/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/widgets/charts/HistogramDataWidget.cpp

  \brief A widget used to adjust a histogram's input data.
*/

//Terralib
#include "../../../dataaccess/dataset/DataSet.h"
#include "../../../dataaccess/dataset/DataSetType.h"
#include "../../../dataaccess/utils/Utils.h"
#include "../../../maptools/Utils.h"
#include "../../../qt/widgets/charts/Utils.h"
#include "../../../raster.h"
#include "../../../raster/RasterSummaryManager.h"
#include "../../../statistics/core/Enums.h"
#include "../../../statistics/core/Utils.h"
#include "../utils/ScopedCursor.h"
#include "Histogram.h"
#include "HistogramDataWidget.h"
#include "ui/ui_HistogramDataWidgetForm.h"

void updateSummary(te::da::DataSet* dataSet, Ui::HistogramDataWidgetForm* ui)
{
  ui->m_summaryComboBox->clear();

  size_t selectedPropertyIdx = 0;
  for (size_t i = 0; i < dataSet->getNumProperties(); i++)
  {
    if(ui->m_propertyComboBox->currentText().toUtf8().data() == dataSet->getPropertyName(i))
      selectedPropertyIdx = i;
  }

  int propType = dataSet->getPropertyDataType(selectedPropertyIdx);

  if(propType == te::dt::DATETIME_TYPE || propType == te::dt::STRING_TYPE)
  {
    ui->m_summaryComboBox->addItem(QString::fromUtf8("None"), QVariant(-1));
    ui->m_summaryComboBox->addItem(QString(te::stat::GetStatSummaryFullName(te::stat::MIN_VALUE).c_str()), QVariant(te::stat::MIN_VALUE));
    ui->m_summaryComboBox->addItem(QString(te::stat::GetStatSummaryFullName(te::stat::MAX_VALUE).c_str()), QVariant(te::stat::MAX_VALUE));
  }
  else
  {
    ui->m_summaryComboBox->addItem(QString::fromUtf8("None"), QVariant(-1));
    ui->m_summaryComboBox->addItem(QString(te::stat::GetStatSummaryFullName(te::stat::MIN_VALUE).c_str()), QVariant(te::stat::MIN_VALUE));
    ui->m_summaryComboBox->addItem(QString(te::stat::GetStatSummaryFullName(te::stat::MAX_VALUE).c_str()), QVariant(te::stat::MAX_VALUE));
    ui->m_summaryComboBox->addItem(QString(te::stat::GetStatSummaryFullName(te::stat::MEAN).c_str()), QVariant(te::stat::MEAN));
    ui->m_summaryComboBox->addItem(QString(te::stat::GetStatSummaryFullName(te::stat::SUM).c_str()), QVariant(te::stat::SUM));
    ui->m_summaryComboBox->addItem(QString(te::stat::GetStatSummaryFullName(te::stat::STANDARD_DEVIATION).c_str()), QVariant(te::stat::STANDARD_DEVIATION));
    ui->m_summaryComboBox->addItem(QString(te::stat::GetStatSummaryFullName(te::stat::VARIANCE).c_str()), QVariant(te::stat::VARIANCE));
    ui->m_summaryComboBox->addItem(QString(te::stat::GetStatSummaryFullName(te::stat::SKEWNESS).c_str()), QVariant(te::stat::SKEWNESS));
    ui->m_summaryComboBox->addItem(QString(te::stat::GetStatSummaryFullName(te::stat::KURTOSIS).c_str()), QVariant(te::stat::KURTOSIS));
    ui->m_summaryComboBox->addItem(QString(te::stat::GetStatSummaryFullName(te::stat::AMPLITUDE).c_str()), QVariant(te::stat::AMPLITUDE));
    ui->m_summaryComboBox->addItem(QString(te::stat::GetStatSummaryFullName(te::stat::MEDIAN).c_str()), QVariant(te::stat::MEDIAN));
    ui->m_summaryComboBox->addItem(QString(te::stat::GetStatSummaryFullName(te::stat::VAR_COEFF).c_str()), QVariant(te::stat::VAR_COEFF));
  }
  ui->m_summaryComboBox->setCurrentIndex(0);
  
  ui->m_slicesSpinBox->setValue( (int)
    te::rst::RasterSummaryManager::getInstance().getDefaultRasterHistSize() );
}

te::qt::widgets::HistogramDataWidget::HistogramDataWidget(te::map::AbstractLayerPtr layer, QWidget* parent, Qt::WindowFlags f)
  : QWidget(parent, f),
    m_ui(new Ui::HistogramDataWidgetForm),
   m_layer(layer)
{
  m_ui->setupUi(this);

  QString item;
  std::unique_ptr<te::da::DataSet> dataset = m_layer->getData();
  std::unique_ptr<te::da::DataSetType> dataType = m_layer->getSchema();

  if(dataType->hasRaster())
    {
      m_ui->m_readAllcheckBox->show();

      //Adjusting the widget to work with a raster file.
      std::unique_ptr<te::rst::Raster> raster (te::map::GetRaster(m_layer.get()));
      
      te::rst::RasterSummary rasterSummary;
      if( ! te::rst::RasterSummaryManager::getInstance().get(
        raster.get(), false, rasterSummary ) )
      {
        throw te::common::Exception(TE_TR("Raster summary Error"));
      }         

      const std::complex<double>* cmin = rasterSummary.at(0).m_minVal;
      const std::complex<double>* cmax = rasterSummary.at(0).m_maxVal;

      double min = cmin->real();
      double max = cmax->real();

      size_t size = raster->getNumberOfBands();
      m_ui->m_slicesSpinBox->setMinimum(0);

      int type = raster->getBandDataType(0);

      if((type >= te::dt::UCHAR_TYPE) && (type <= te::dt::UINT32_TYPE))
      {
        if (min >= 0 && max <= 255)
          m_ui->m_slicesSpinBox->setMaximum(255);
        else
          m_ui->m_slicesSpinBox->setMaximum(max);
      }

      for (size_t i = 0; i < size; i++)
      {
        item = QString::number(i);
        m_ui->m_propertyComboBox->addItem((QString::fromUtf8("Band: ") + item), QVariant::fromValue(i));
      }
    }
  else
  {
    m_ui->m_readAllcheckBox->hide();
    for (std::size_t i = 0; i < dataset->getNumProperties(); i++)
    {
      if(dataset->getPropertyDataType(i) != te::dt::GEOMETRY_TYPE)
      {
        item = QString::fromUtf8(dataset->getPropertyName(i).c_str());
        m_ui->m_propertyComboBox->addItem(item, QVariant::fromValue(i));
      }
    }
  }

  updateSummary(dataset.get(), getForm());

  if(te::da::HasLinkedTable(dataType.get()))
  {
    m_ui->m_summaryComboBox->show();
    m_ui->m_summaryLabel->show();
  }
  else
  {
    m_ui->m_summaryComboBox->hide();
    m_ui->m_summaryLabel->hide();
  }

// connect signal and slots
  connect(m_ui->m_propertyComboBox, SIGNAL(currentIndexChanged(QString)), this, SLOT(onPropertyComboBoxIndexChanged(QString)));
  connect(m_ui->m_readAllcheckBox, SIGNAL(stateChanged(int)), this, SLOT(onReadAllCheckBoxStateChanged(int)));
}

te::qt::widgets::HistogramDataWidget::~HistogramDataWidget() = default;

Ui::HistogramDataWidgetForm* te::qt::widgets::HistogramDataWidget::getForm()
{
  return m_ui.get();
}

te::qt::widgets::Histogram* te::qt::widgets::HistogramDataWidget::getHistogram() 
{
  std::unique_ptr<te::da::DataSet> dataset = m_layer->getData();
  std::unique_ptr<te::da::DataSetType> dataType = m_layer->getSchema();
  te::qt::widgets::Histogram* histogram;

  if(dataType->hasRaster())
  {
    histogram = te::qt::widgets::createHistogram(m_layer, m_ui->m_propertyComboBox->itemData(m_ui->m_propertyComboBox->currentIndex()).toInt(),
      m_ui->m_slicesSpinBox->value(), -1, m_ui->m_readAllcheckBox->isChecked());
  }
  else
  {
    //Getting the Columns that will be used to populate the chart

    size_t selectedPropertyIdx = 0;

    for (size_t i = 0; i < dataset->getNumProperties(); i++)
    {
      if(m_ui->m_propertyComboBox->currentText().toUtf8().data() == dataset->getPropertyName(i))
      {
        selectedPropertyIdx = i;
      }
    }

    int propType = dataset->getPropertyDataType(selectedPropertyIdx);
    int stat = m_ui->m_summaryComboBox->itemData(m_ui->m_summaryComboBox->currentIndex()).toInt();

    if(propType == te::dt::DATETIME_TYPE || propType == te::dt::STRING_TYPE)
    {
      histogram = te::qt::widgets::createHistogram(m_layer, (int)selectedPropertyIdx, stat);
    }
    else
    {
      histogram = te::qt::widgets::createHistogram(m_layer, (int)selectedPropertyIdx, m_ui->m_slicesSpinBox->value(), stat);
    }
  }
  return histogram;
}

void te::qt::widgets::HistogramDataWidget::setHistogramProperty(int propId)
{
  m_ui->m_propertyComboBox->setCurrentIndex(m_ui->m_propertyComboBox->findData(propId));
  m_ui->m_propertyComboBox->setEnabled(false);
}

void te::qt::widgets::HistogramDataWidget::onPropertyComboBoxIndexChanged(
    QString /*text*/)
{
  te::qt::widgets::ScopedCursor c(Qt::WaitCursor);
  std::unique_ptr<te::da::DataSet> dataset = m_layer->getData();
  std::unique_ptr<te::da::DataSetType> dataType = m_layer->getSchema();

  if(!dataType->hasRaster())
  {
    size_t selectedPropertyIdx= te::da::GetPropertyPos(dataset.get(), m_ui->m_propertyComboBox->currentText().toUtf8().data());
    int propType = dataset->getPropertyDataType(selectedPropertyIdx);

    if(propType == te::dt::DATETIME_TYPE || propType == te::dt::STRING_TYPE)
    {
      m_ui->m_slicesSpinBox->setEnabled(false);
    }
    else
    {
      m_ui->m_slicesSpinBox->setEnabled(true);
    }
  }
  else
  {
    std::unique_ptr<te::rst::Raster> raster (te::map::GetRaster(m_layer.get()));

    te::rst::RasterSummary rasterSummary;
    if( ! te::rst::RasterSummaryManager::getInstance().get(
      raster.get(), false, rasterSummary ) )
    {
      throw te::common::Exception(TE_TR("Raster summary Error"));
    }     

    const std::complex<double>* cmin = rasterSummary.at(m_ui->m_propertyComboBox->currentIndex()).m_minVal;
    const std::complex<double>* cmax = rasterSummary.at(m_ui->m_propertyComboBox->currentIndex()).m_maxVal;

    double min = cmin->real();
    double max = cmax->real();

    if (min >= 0 && max <= 255)
      m_ui->m_slicesSpinBox->setMaximum(255);
    else
      m_ui->m_slicesSpinBox->setMaximum(max);

    m_ui->m_slicesSpinBox->setValue(255);
  }

  updateSummary(dataset.get(), getForm());

  if(te::da::HasLinkedTable(dataType.get()))
  {
    m_ui->m_summaryComboBox->show();
    m_ui->m_summaryLabel->show();
  }
  else
  {
    m_ui->m_summaryComboBox->hide();
    m_ui->m_summaryLabel->hide();
  }
}

void te::qt::widgets::HistogramDataWidget::onReadAllCheckBoxStateChanged(int state)
{
  if (state == Qt::Checked)
    m_ui->m_slicesSpinBox->setEnabled(true);
  else if(state == Qt::CheckState::Unchecked)
    m_ui->m_slicesSpinBox->setEnabled(false);
}

std::string te::qt::widgets::HistogramDataWidget::getSummaryFunction()
{
  return m_ui->m_summaryComboBox->currentText().toUtf8().data();
}
