/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralib/qt/widgets/dataset/utils/DataSetTypeFilter.h

   \brief Custom widget used filter a dataset type, creating a DataSetTypeConverter.
 */

#include "DataSetTypeFilter.h"

#include "../../utils/DoubleListWidget.h"
#include "../../../../common/StringUtils.h"
#include "../../../../dataaccess/dataset/DataSetTypeConverter.h"
#include "../../../../dataaccess/datasource/DataSourceCapabilities.h"
#include "../../../../dataaccess/utils/DataAccess.h"
#include "../../../../dataaccess/utils/Utils.h"
#include "../../../../geometry/GeometryProperty.h"

te::qt::widgets::DataSetTypeFilter::DataSetTypeFilter(QWidget* parent, Qt::WindowFlags f)
  : QWidget(parent, f)
{
  //add double list widget to this form
  m_doubleListWidget.reset(new te::qt::widgets::DoubleListWidget(this));
  m_doubleListWidget->setLeftLabel(tr("Input Attributes"));
  m_doubleListWidget->setRightLabel(tr("Output Attributes"));
  m_doubleListWidget->setVisibledOrderButtons(false);

  QGridLayout* layout = new QGridLayout(this);
  layout->addWidget(m_doubleListWidget.get());
  layout->setContentsMargins(0, 0, 0, 0);
}

te::qt::widgets::DataSetTypeFilter::~DataSetTypeFilter() = default;

void te::qt::widgets::DataSetTypeFilter::init(const te::da::DataSetType* dataSetType, bool selectAllAttrs)
{
  m_doubleListWidget->clearInputValues();
  m_doubleListWidget->clearOutputValues();

  std::set<std::string> fixedProperties;

  //added as fixed output pk properties
  if (dataSetType->getPrimaryKey())
  {
    const std::vector<te::dt::Property*>& properties = dataSetType->getPrimaryKey()->getProperties();

    std::vector<std::string> pkProperties;

    for (size_t t = 0; t < properties.size(); ++t)
    {
      pkProperties.push_back(properties[t]->getName());

      fixedProperties.insert(properties[t]->getName());
    }

    m_doubleListWidget->setFixedOutputValues(pkProperties, "key");
  }

  //added as fixed output geom prop
  te::gm::GeometryProperty* geomProp = te::da::GetFirstGeomProperty(dataSetType);

  if (geomProp)
  {
    std::vector<std::string> geomProperties;
    geomProperties.push_back(geomProp->getName());

    fixedProperties.insert(geomProp->getName());

    m_doubleListWidget->setFixedOutputValues(geomProperties, "geometry");
  }

  //add other properties as input
  std::vector<te::dt::Property*> props = dataSetType->getProperties();

  std::vector<std::string> inputValues;

  for (std::size_t i = 0; i < props.size(); ++i)
  {
    std::set<std::string>::iterator it = fixedProperties.find(props[i]->getName());

    if (it == fixedProperties.end())
    {
      inputValues.push_back(props[i]->getName());
    }
  }

  if (selectAllAttrs)
  {
    m_doubleListWidget->addOutputValues(inputValues);
  }
  else
  {
    m_doubleListWidget->setInputValues(inputValues);
  }
}

void te::qt::widgets::DataSetTypeFilter::applyFilterToConverter(te::da::DataSetTypeConverter* converter) const
{
  assert(converter);

  //then we must filter all selected attributes. All the attributes not in this list will be removed from the converter
  std::vector<std::string> attrs = m_doubleListWidget->getOutputValues();

  std::unique_ptr<te::da::DataSetType> dataSetTypeConverted(static_cast<te::da::DataSetType*>(converter->getResult()->clone()));
  for (std::size_t i = 0; i < dataSetTypeConverted->getProperties().size(); ++i)
  {
    te::dt::Property* currentProperty = dataSetTypeConverted->getProperties().at(i);
    bool remove = true;
    for (std::size_t j = 0; j < attrs.size(); ++j)
    {
      if (te::common::IsStringEqual(currentProperty->getName(), attrs.at(j), false) == true)
      {
        remove = false;
        break;
      }
    }

    if (remove == true)
    {
      converter->remove(currentProperty->getName());
    }
  }
}

void te::qt::widgets::DataSetTypeFilter::applyFilterToDataAccess(te::da::DataAccess* dataAccess) const
{
  //if the dataAccess has no converter, we must create a new one
  te::da::DataSetTypeConverter* converter = dataAccess->getConverter();
  if (converter == nullptr)
  {
    converter = getConverter(dataAccess->getDataSetType(), dataAccess->getDataSource()->getCapabilities(), dataAccess->getDataSetName());
    dataAccess->setConverter(converter);
  }
  else
  {
    applyFilterToConverter(converter);
  }
}

te::da::DataSetTypeConverter* te::qt::widgets::DataSetTypeFilter::getConverter(const te::da::DataSetType* dataSetType, const te::da::DataSourceCapabilities& capabilities, const std::string& resultDataSetName) const
{
  te::da::DataSetTypeConverter* converter = new te::da::DataSetTypeConverter(dataSetType, capabilities);
  applyFilterToConverter(converter);

  converter->getResult()->setDatasetName(resultDataSetName);
  converter->getResult()->setTitle(resultDataSetName);

  return converter;
}

