/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file DataSourceTreeWidget.h

   \brief A tree widget for the data sources of an application.
 */

#include "DataSourceTreeWidget.h"

// TerraLib Includes
#include "../../Exception.h"
#include "../../Utils.h"
#include "../../datasource/connector/AbstractDataSourceConnector.h"
#include "../../layer/utils/DataSet2Layer.h"
#include "../../../../common/STLUtils.h"
#include "../../../../core/translator/Translator.h"
#include "../../../../dataaccess/datasource/DataSourceInfo.h"
#include "../../../../dataaccess/datasource/DataSourceInfoManager.h"
#include "../../../../dataaccess/datasource/DataSource.h"
#include "../../../../dataaccess/datasource/DataSourceManager.h"
#include "../../../../dataaccess/utils/Utils.h"
#include "../../../../geometry/GeometryProperty.h"
#include "../../../../geometry/Utils.h"

// QT Includes
#include <QMessageBox>
#include <QMetaType>

Q_DECLARE_METATYPE(te::da::DataSetTypePtr)
Q_DECLARE_METATYPE(te::da::DataSourcePtr)


te::qt::widgets::DataSourceTreeWidget::DataSourceTreeWidget(QWidget* parent) : QTreeWidget(parent)
{
  setHeaderHidden(true);
  setContextMenuPolicy(Qt::CustomContextMenu);
  setSortingEnabled(true);

  connect(this, SIGNAL(itemDoubleClicked(QTreeWidgetItem*, int)), this, SLOT(slotItemDoubleClicked(QTreeWidgetItem*, int)));
  connect(this, SIGNAL(customContextMenuRequested(const QPoint&)), this, SLOT(slotCustomContextMenu(const QPoint&)));

  initMenus();
}

te::qt::widgets::DataSourceTreeWidget::~DataSourceTreeWidget()
{
  m_dsTypes.clear();
}

void te::qt::widgets::DataSourceTreeWidget::setDataSourceTypes(std::vector<te::qt::widgets::DataSourceType*> dsTypes)
{
  //clear tree
  clear();

  m_dsTypes = dsTypes;

  //add top level items as data source types
  for (std::size_t i = 0; i < dsTypes.size(); ++i)
  {
    QTreeWidgetItem* item = new QTreeWidgetItem(this, te::qt::widgets::DataSourceTreeWidget::DATASOURCETYPE);
    item->setText(0, te::qt::widgets::Convert2QString(dsTypes[i]->getTitle()));
    item->setIcon(0, dsTypes[i]->getIcon(te::qt::widgets::DataSourceType::ICON_DATASOURCE_SMALL));
    item->setData(0, Qt::UserRole, QVariant(te::qt::widgets::Convert2QString(dsTypes[i]->getName())));

    addTopLevelItem(item);

    loadDataSources(item);
  }

  sortItems(0, Qt::AscendingOrder);
}

void te::qt::widgets::DataSourceTreeWidget::initMenus()
{
  //datasourcetype menu
  m_dataSourceTypeMenu = new QMenu(this);

  QAction* actionNewConnection = m_dataSourceTypeMenu->addAction(tr("New Connection..."));
  m_dataSourceTypeMenu->addSeparator();
  QAction* actionDataSourceTypeUpdate = m_dataSourceTypeMenu->addAction(tr("Update"));

  connect(actionNewConnection, SIGNAL(triggered()), this, SLOT(newConnection()));
  connect(actionDataSourceTypeUpdate, SIGNAL(triggered()), this, SLOT(dataSourceTypeUpdate()));

  //datasource menu
  m_dataSourceMenu = new QMenu(this);

  QAction* actionDataSourceUpdate = m_dataSourceMenu->addAction(tr("Connect"));

  connect(actionDataSourceUpdate, SIGNAL(triggered()), this, SLOT(dataSourceUpdate()));

  //dataset menu
  m_dataSetMenu = new QMenu(this);

  QAction* actionDataSetCreateLayer = m_dataSetMenu->addAction(tr("Create Layer"));

  connect(actionDataSetCreateLayer, SIGNAL(triggered()), this, SLOT(createLayer()));
}

void te::qt::widgets::DataSourceTreeWidget::loadDataSources(QTreeWidgetItem* item)
{
  if (item == nullptr || item->type() != te::qt::widgets::DataSourceTreeWidget::DATASOURCETYPE)
  {
    return;
  }

  //clear child items
  for (int i = 0; i < item->childCount(); ++i)
  {
    QTreeWidgetItem* child = item->takeChild(i);

    delete child;
  }

  //load datasources from specific type
  std::string dataSourceType = te::qt::widgets::Convert2StdString(item->data(0, Qt::UserRole).toString());

  std::vector<te::da::DataSourceInfoPtr> datasources;

  te::da::DataSourceInfoManager::getInstance().getByType(dataSourceType, datasources);

  for (std::size_t i = 0; i < datasources.size(); ++i)
  {
    QTreeWidgetItem* subItem = new QTreeWidgetItem(item, te::qt::widgets::DataSourceTreeWidget::DATASOURCE);
    subItem->setText(0, te::qt::widgets::Convert2QString(datasources[i]->getTitle()));
    subItem->setIcon(0, QIcon::fromTheme("datasource"));
    subItem->setData(0, Qt::UserRole, QVariant(te::qt::widgets::Convert2QString(datasources[i]->getId())));

    item->addChild(subItem);
  }

  item->sortChildren(0, Qt::AscendingOrder);
}

void te::qt::widgets::DataSourceTreeWidget::loadDataSets(QTreeWidgetItem* item)
{
  if (item == nullptr || item->type() != te::qt::widgets::DataSourceTreeWidget::DATASOURCE)
  {
    return;
  }

  //clear child items
  QList<QTreeWidgetItem *> childList = item->takeChildren();

  for (int i = 0; i < childList.count(); ++i)
  {
    delete childList.at(i);
  }

  childList.clear();

  //get data source
  std::string dataSourceInfoId = te::qt::widgets::Convert2StdString(item->data(0, Qt::UserRole).toString());

  te::da::DataSourceInfoPtr dsInfo = te::da::DataSourceInfoManager::getInstance().get(dataSourceInfoId);

  te::da::DataSourcePtr ds = te::da::DataSourceManager::getInstance().get(dsInfo->getId(), dsInfo->getAccessDriver(), dsInfo->getConnInfo());

  if (ds.get() == nullptr)
  {
    return;
  }

  if (ds->isValid() == false)
  {
    return;
  }

  if (ds->isOpened() == false)
  {
    ds->open();
  }
  
  //load datasets from specific datasource
  std::vector<std::string> datasetNames = ds->getDataSetNames();

  std::sort(datasetNames.begin(), datasetNames.end());

  for (std::size_t i = 0; i < datasetNames.size(); ++i)
  {
    //check datasettype geom type
    te::da::DataSetTypePtr dsType(ds->getDataSetType(datasetNames[i]).release());

    QIcon icon;

    if (dsType->hasRaster())
    {
      icon = QIcon::fromTheme("file-raster");
    }
    else if (dsType->hasGeom())
    {
      te::gm::GeomType geomType = te::gm::UnknownGeometryType;

      te::dt::Property* prop = te::da::GetFirstSpatialProperty(dsType.get());

      if (prop)
      {
        te::gm::GeometryProperty* geomProp = dynamic_cast<te::gm::GeometryProperty*>(prop);

        geomType =  te::gm::Get2DGeomTypeId(geomProp->getGeometryType());
      }

      if (geomType == te::gm::PointType || geomType == te::gm::MultiPointType)
      {
        icon = QIcon::fromTheme("geom-point");
      }
      else if (geomType == te::gm::LineStringType || geomType == te::gm::MultiLineStringType)
      {
        icon = QIcon::fromTheme("geom-line");
      }
      else if (geomType == te::gm::PolygonType || geomType == te::gm::MultiPolygonType)
      {
        icon = QIcon::fromTheme("geom-poly");
      }
    }
    else
    {
      continue;
    }

    QTreeWidgetItem* subItem = new QTreeWidgetItem(item, te::qt::widgets::DataSourceTreeWidget::DATASET);
    subItem->setText(0, te::qt::widgets::Convert2QString(datasetNames[i]));
    subItem->setIcon(0, icon);
    subItem->setData(0, Qt::UserRole, QVariant::fromValue(dsType));
    subItem->setData(1, Qt::UserRole, QVariant::fromValue(ds));

    item->addChild(subItem);
  }
}

void te::qt::widgets::DataSourceTreeWidget::slotItemDoubleClicked(QTreeWidgetItem* item, int column)
{
  if (item == nullptr)
  {
    return;
  }

  if (item->type() == te::qt::widgets::DataSourceTreeWidget::DATASOURCE)
  {
    dataSourceUpdate();
  }
  else if (item->type() == te::qt::widgets::DataSourceTreeWidget::DATASET)
  {
    createLayer();
  }
}

void te::qt::widgets::DataSourceTreeWidget::slotCustomContextMenu(const QPoint &point)
{
  QTreeWidgetItem* item = itemAt(point);

  if (item == nullptr)
  {
    return;
  }

  QMenu* itemMenu = nullptr;

  switch (item->type())
  {
    case te::qt::widgets::DataSourceTreeWidget::DATASOURCETYPE:
    {
      itemMenu = m_dataSourceTypeMenu;
    }
    break;

    case te::qt::widgets::DataSourceTreeWidget::DATASOURCE:
    {
      itemMenu = m_dataSourceMenu;
    }
    break;

    case te::qt::widgets::DataSourceTreeWidget::DATASET:
    {
      itemMenu = m_dataSetMenu;
    }
    break;

    default:
      break;
  }

  if (itemMenu)
  {
    itemMenu->exec(QCursor::pos());
  }
}

void te::qt::widgets::DataSourceTreeWidget::newConnection()
{
  //get selected item
  QList<QTreeWidgetItem*> selectedItems = this->selectedItems();

  if (selectedItems.count() != 1)
  {
    return;
  }

  //check if selected item is a dataset item
  QTreeWidgetItem* itemFirst = selectedItems.first();

  if (itemFirst->type() != te::qt::widgets::DataSourceTreeWidget::DATASOURCETYPE)
  {
    return;
  }

  //get datasourcetype 
  std::string datasourceTypeName = te::qt::widgets::Convert2StdString(itemFirst->data(0, Qt::UserRole).toString());

  te::qt::widgets::DataSourceType* dsType = nullptr;

  for (std::size_t i = 0; i < m_dsTypes.size(); ++i)
  {
    if (m_dsTypes[i]->getName() == datasourceTypeName)
    {
      dsType = m_dsTypes[i];
      break;
    }
  }

  //show datasourcetype connector
  if (dsType == nullptr)
  {
    return;
  }

  std::unique_ptr<QWidget> connectorw(dsType->getWidget(te::qt::widgets::DataSourceType::WIDGET_DATASOURCE_CONNECTOR));

  if (connectorw.get() == nullptr)
  {
    throw Exception(TE_TR("This type of data source hasn't provide a dialog for adding a new data source"));
  }

  te::qt::widgets::AbstractDataSourceConnector* connector = dynamic_cast<te::qt::widgets::AbstractDataSourceConnector*>(connectorw.get());

  if (connector == nullptr)
  {
    throw Exception(TE_TR("Wrong type of object for adding a new data source!"));
  }

  std::list<te::da::DataSourceInfoPtr> datasources;

  connector->connect(datasources);

  //update datasource list of this datasource type
  loadDataSources(itemFirst);
}

void te::qt::widgets::DataSourceTreeWidget::dataSourceTypeUpdate()
{

}

void te::qt::widgets::DataSourceTreeWidget::dataSourceUpdate()
{
  //get selected item
  QList<QTreeWidgetItem*> selectedItems = this->selectedItems();

  if (selectedItems.count() != 1)
  {
    return;
  }

  //check if selected item is a dataset item
  QTreeWidgetItem* itemFirst = selectedItems.first();

  if (itemFirst->type() != te::qt::widgets::DataSourceTreeWidget::DATASOURCE)
  {
    return;
  }

  loadDataSets(itemFirst);
}

void te::qt::widgets::DataSourceTreeWidget::createLayer()
{
  //get selected item
  QList<QTreeWidgetItem*> selectedItems = this->selectedItems();

  if (selectedItems.count() != 1)
  {
    return;
  }

  //check if selected item is a dataset item
  QTreeWidgetItem* itemFirst = selectedItems.first();

  if (itemFirst->type() != te::qt::widgets::DataSourceTreeWidget::DATASET)
  {
    return;
  }

  //get info
  QVariant varDataSetType = itemFirst->data(0, Qt::UserRole);
  te::da::DataSetTypePtr dataSetType = varDataSetType.value<te::da::DataSetTypePtr>();

  QVariant varDataSource = itemFirst->data(1, Qt::UserRole);
  te::da::DataSourcePtr dataSource = varDataSource.value<te::da::DataSourcePtr>();
  
  //create layer
  te::qt::widgets::DataSet2Layer ds2layer(dataSource->getId());

  te::map::AbstractLayerPtr layer = ds2layer(dataSetType);

  std::list<te::map::AbstractLayerPtr> layers;
  layers.push_back(layer);

  emit layersCreated(layers);

  QMessageBox::information(this, tr("Information"), tr("Layer created."));
}
