/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file FileSystemTreeView.cpp

  \brief A tree view for the file system of an application.
*/

// TerraLib
#include "FileSystemTreeView.h"
#include "../../Utils.h"
#include "../../layer/utils/DataSet2Layer.h"
#include "../../../../common/StringUtils.h"
#include "../../../../dataaccess/datasource/DataSource.h"
#include "../../../../dataaccess/datasource/DataSourceManager.h"
#include "../../../../dataaccess/datasource/DataSourceFactory.h"
#include "../../../../dataaccess/datasource/DataSourceInfo.h"
#include "../../../../dataaccess/datasource/DataSourceInfoManager.h"

// Qt Includes
#include <QMessageBox>
#include <QMenu>

// Boost
#include <boost/uuid/random_generator.hpp>
#include <boost/uuid/uuid_io.hpp>
#include <boost/filesystem.hpp>

te::qt::widgets::FileSystemTreeView::FileSystemTreeView(QWidget* parent)
  : QTreeView(parent)
{
  setContextMenuPolicy(Qt::CustomContextMenu);

  connect(this, SIGNAL(clicked(const QModelIndex&)), this, SLOT(itemClicked(const QModelIndex&)));
  connect(this, SIGNAL(doubleClicked(const QModelIndex&)), this, SLOT(itemDoubleClicked(const QModelIndex&)));
  connect(this, SIGNAL(customContextMenuRequested(const QPoint&)), this, SLOT(slotCustomContextMenu(const QPoint&)));

  //set model
  m_model = new QFileSystemModel;
  m_model->setRootPath("");
  
  setModel(m_model);

  for (int i = 1; i < m_model->columnCount(); ++i)
  {
    hideColumn(i);
  }
}

te::qt::widgets::FileSystemTreeView::~FileSystemTreeView() = default;

void te::qt::widgets::FileSystemTreeView::setNameFilters(QStringList list)
{
  m_model->setNameFilters(list);
  m_model->setNameFilterDisables(false);
}

void te::qt::widgets::FileSystemTreeView::setDataSourceTypesNames(std::vector<std::string> dsTypesNames)
{
  m_dataSourceTypesNames = dsTypesNames;
}

void te::qt::widgets::FileSystemTreeView::itemClicked(const QModelIndex & index)
{
  if (index.isValid() == false)
  {
    return;
  }

  QFileInfo info = m_model->fileInfo(index);

  emit clicked(info);
}

void te::qt::widgets::FileSystemTreeView::itemDoubleClicked(const QModelIndex & index)
{
  if (index.isValid() == false)
  {
    return;
  }

  QFileInfo info = m_model->fileInfo(index);

  if (info.isFile())
  {
    std::list<te::map::AbstractLayerPtr> layers = createLayer(info);

    if (layers.empty() == false)
    {
      emit layersCreated(layers);

      QMessageBox::information(this, tr("Information"), tr("Layer created."));
    }
  }
}


void te::qt::widgets::FileSystemTreeView::slotCustomContextMenu(const QPoint &point)
{
  QModelIndex index = this->indexAt(point);

  if (index.isValid() == false)
  {
    return;
  }

  QFileInfo info = m_model->fileInfo(index);

  //QMenu* itemMenu;

  //itemMenu->exec(QCursor::pos());
}

std::list<te::map::AbstractLayerPtr> te::qt::widgets::FileSystemTreeView::createLayer(const QFileInfo& file)
{
  std::list<te::map::AbstractLayerPtr> layers;

  std::string fileName = te::qt::widgets::Convert2StdString(file.absoluteFilePath());

  //try each data source type to create layer
  for (std::size_t i = 0; i < m_dataSourceTypesNames.size(); ++i)
  {
    std::string connInfo("file://" + fileName);

    try
    {
      //open data source
      te::da::DataSourcePtr dataSource(te::da::DataSourceFactory::make(m_dataSourceTypesNames[i], connInfo));

      dataSource->open();

      //get data sets
      std::list<te::da::DataSetTypePtr> dataSetTypes;

      std::vector<std::string> dataSets = dataSource->getDataSetNames();

      std::vector<std::string>::iterator it;

      for (it = dataSets.begin(); it != dataSets.end(); ++it)
      {
        te::da::DataSetTypePtr dsType;

        dsType = dataSource->getDataSetType(*it);

        if (dsType.get() != nullptr)
        {
          dataSetTypes.push_back(dsType);
        }
      }

      if (dataSetTypes.empty() == false)
      {
        //add data source to manager
        boost::uuids::basic_random_generator<boost::mt19937> gen;
        boost::uuids::uuid u = gen();
        std::string id = boost::uuids::to_string(u);

        te::da::DataSourceInfoPtr dsInfo(new te::da::DataSourceInfo);

        dsInfo->setAccessDriver(m_dataSourceTypesNames[i]);
        dsInfo->setConnInfo(connInfo);
        dsInfo->setId(id);
        dsInfo->setTitle(te::qt::widgets::Convert2StdString(file.baseName()));
        dsInfo->setType(m_dataSourceTypesNames[i]);

        if (!te::da::DataSourceInfoManager::getInstance().add(dsInfo))
        {
          dsInfo = te::da::DataSourceInfoManager::getInstance().getByConnInfo(dsInfo->getConnInfoAsString());
        }

        //force to add datasource to manager
        te::da::DataSourcePtr ds = te::da::DataSourceManager::getInstance().get(dsInfo->getId(), dsInfo->getType(), dsInfo->getConnInfo());

        if (!ds->isOpened())
        {
          ds->open();
        }

        //create layers
        std::transform(dataSetTypes.begin(), dataSetTypes.end(), std::back_inserter(layers), te::qt::widgets::DataSet2Layer(dsInfo->getId()));
      }
    }
    catch (...)
    {
      continue;
    }
  }

  return layers;
}

