#include "LayerItem.h"

#include "ChartItem.h"
#include "StyleItem.h"

#include "../../se/SymbologyPreview.h"
#include "../../Utils.h"

#include "../../../../core/uri/URI.h"
#include "../../../../core/uri/Utils.h"
#include "../../../../dataaccess/datasource/DataSourceInfoManager.h"
#include "../../../../maptools/DataSetLayer.h"
#include "../../../../maptools/DataSetAdapterLayer.h"
#include "../../../../maptools/Utils.h"

#include "../../../../se/Style.h"
#include "../../../../se/Utils.h"

#include "../../../../srs/SpatialReferenceSystemManager.h"

QString GetDataSetName(te::map::AbstractLayerPtr l)
{
  QString toolTip = QObject::tr("<h3><b>Layer Info</b></h3>");

  toolTip += QObject::tr("<b>Name: </b>") + te::qt::widgets::Convert2QString(l->getTitle()) + QString("<br/>");
  toolTip += QObject::tr("<b>DataSet: </b>") + te::qt::widgets::Convert2QString(l->getDataSetName()) + QString("<br/>");

  return toolTip;
}

QString GetDataSourceInfo(te::map::AbstractLayerPtr l)
{
  // Gets the connection info
  const std::string& id = l->getDataSourceId();

  te::da::DataSourceInfoPtr info = te::da::DataSourceInfoManager::getInstance().get(id);

  if (info.get() == nullptr)
  {
    return QString(QObject::tr("<b>Information unavailable.</b>"));
  }

  const te::core::URI& connInfo = info->getConnInfo();

  QString toolTip = QObject::tr("<h3><b>Connection Info</b></h3>");

  toolTip += QObject::tr("<b>Scheme: </b>") + te::qt::widgets::Convert2QString(connInfo.scheme());

  toolTip += QString("<br/>");

  if (!connInfo.user().empty())
  {
    toolTip += QObject::tr("<b>tUser: </b>") + te::qt::widgets::Convert2QString(connInfo.user());
    toolTip += QString("<br/>");
  }

  if (!connInfo.host().empty())
  {
    toolTip += QObject::tr("<b>Host: </b>") + te::qt::widgets::Convert2QString(connInfo.host());
    toolTip += QString("<br/>");
  }

  if (!connInfo.port().empty())
  {
    toolTip += QObject::tr("<b>Port: </b>") + te::qt::widgets::Convert2QString(connInfo.port());
    toolTip += QString("<br/>");
  }

  if (!connInfo.path().empty())
  {
    toolTip += QObject::tr("<b>Path: </b>") + te::qt::widgets::Convert2QString(connInfo.path());
    toolTip += QString("<br/>");
  }

  std::map<std::string, std::string> kvp = te::core::Expand(connInfo.query());
  std::map<std::string, std::string>::iterator itBegin = kvp.begin();
  std::map<std::string, std::string>::iterator itEnd = kvp.end();

  while (itBegin != itEnd)
  {
    toolTip += QObject::tr("<b>");
    toolTip += te::qt::widgets::Convert2QString(itBegin->first);
    toolTip += ": ";
    toolTip += QObject::tr("</b>");
    toolTip += te::qt::widgets::Convert2QString(itBegin->second);
    toolTip += QString("<br/>");
    ++itBegin;
  }

  return toolTip;
}

QString GetSRSInfo(te::map::AbstractLayerPtr l)
{
  QString toolTip;

  if (l->getSRID() != TE_UNKNOWN_SRS)
  {
    toolTip += QObject::tr("<h3><b>Spatial Reference System</b></h3>");
    toolTip += QObject::tr("<b>SRID: </b>") + QString::number(l->getSRID()) + QString("<br/>");
    toolTip += QObject::tr("<b>Name: </b>") + te::qt::widgets::Convert2QString(te::srs::SpatialReferenceSystemManager::getInstance().getName(l->getSRID())) + QString("<br/>");
    toolTip += QObject::tr("<b>Proj4: </b>") + te::qt::widgets::Convert2QString(te::srs::SpatialReferenceSystemManager::getInstance().getP4Txt(l->getSRID())) + QString("<br/>");
  }

  return toolTip;
}

QString GetWarnings(te::map::AbstractLayerPtr l)
{
  bool hasWarnings = false;

  QString warningsTip = QObject::tr("<h3><b>Warnings</b></h3>");

  std::unique_ptr<te::da::DataSetType> schema = l->getSchema();

  if (schema.get())
  {
    // check PK
    if (schema->hasRaster() == false && schema->getPrimaryKey() == nullptr)
    {
      hasWarnings = true;

      warningsTip += QObject::tr("<b>Primary Key: </b>Not found!");
    }

    // check SRID
    if (l->getSRID() == TE_UNKNOWN_SRS)
    {
      hasWarnings = true;

      warningsTip += QObject::tr("<b>SRID: </b>Not defined!");
    }
  }

  if (hasWarnings == false)
  {
    return QString();
  }

  return warningsTip;
}

QString GetTooltip(te::map::AbstractLayerPtr l)
{
  QString toolTip = "<p style='white-space:pre'>";

  // Gets dataset name
  toolTip += GetDataSetName(l);

  // Gets datasource info
  toolTip += GetDataSourceInfo(l);

  // Gets srs info
  if (l->getSRID() != TE_UNKNOWN_SRS)
  {
    toolTip += GetSRSInfo(l);
  }
  
  //check for layer warnings
  toolTip += GetWarnings(l);

  toolTip += "</p>";

  return toolTip;
}


QString BuildToolIip(te::qt::widgets::LayerItem const* item)
{
  te::map::AbstractLayerPtr l = item->getLayer();

  if(!l->getLastValidatedValue())
    return QObject::tr("Invalid Layer");

  QString toolTip = GetTooltip(l);

  return toolTip;
}

te::qt::widgets::LayerItem::LayerItem(te::map::AbstractLayerPtr layer) :
  TreeItem("LAYER"),
  m_layer(layer)
{
  if(!m_layer->isValid())
    return;

  if (m_layer->getStyle() != nullptr)
    addChild(new StyleItem(m_layer->getStyle()));

  if (m_layer->getSelectionStyle() != nullptr)
    addChild(new StyleItem(m_layer->getSelectionStyle()));

  updateChart();
}

te::qt::widgets::LayerItem::~LayerItem()
{
  std::vector<TreeItem*> items;
  getChildren(items, "CHART");

  if(!items.empty())
    ((*items.begin())->setParent(nullptr));

  items.clear();
  getChildren(items, "STYLE");

  if(!items.empty())
    ((*items.begin())->setParent(nullptr));

  if (m_layer->getParent() != nullptr)
    m_layer->getParent()->remove(m_layer->getIndex());
}

std::string te::qt::widgets::LayerItem::getAsString() const
{
  return m_layer->getTitle();
}

te::qt::widgets::VISIBLE te::qt::widgets::LayerItem::isVisible() const
{
  return (m_layer->getVisibility() == te::map::VISIBLE) ? TOTALLY : NONE;
}

void te::qt::widgets::LayerItem::setVisible(const VISIBLE& visible, const bool& updateAncestors, const bool&)
{
  m_layer->setVisibility((visible == TOTALLY) ? te::map::VISIBLE : te::map::NOT_VISIBLE);

  if(updateAncestors)
    m_parent->setVisible(visible, true);
}

Qt::ItemFlags te::qt::widgets::LayerItem::flags()
{
  return TreeItem::flags() | Qt::ItemIsEditable | Qt::ItemIsDragEnabled | Qt::ItemIsUserCheckable;
}

te::map::AbstractLayerPtr te::qt::widgets::LayerItem::getLayer() const
{
  return m_layer;
}

void te::qt::widgets::LayerItem::updateChart()
{
  std::vector<TreeItem*> items;

  getChildren(items, "LEGEND");

  if(m_layer->getChart() != nullptr)
    insertChild(new ChartItem(m_layer->getChart()), items.size());
}

int te::qt::widgets::LayerItem::updateLegend()
{
  StyleItem* si = new StyleItem(m_layer->getStyle());

  addChild(si);

  if (m_layer->getSelectionStyle())
  {
    StyleItem* siSel = new StyleItem(m_layer->getSelectionStyle());

    addChild(siSel);
  }

  return si->getPosition();
}

std::string te::qt::widgets::LayerItem::getToolTip() const
{
  return BuildToolIip(this).toUtf8().data();
}
