/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

This file is part of the TerraLib - a Framework for building GIS enabled applications.

TerraLib is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

TerraLib is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with TerraLib. See COPYING. If not, write to
TerraLib Team at <terralib-team@terralib.org>.
*/

/*!
* \file layer/explorer/TreeItem.cpp
*/

#include "TreeItem.h"

te::qt::widgets::TreeItem::TreeItem(const std::string &type) :
  m_type(type)
{
  m_children.reset(new Queue);
  m_parent = nullptr;
}

te::qt::widgets::TreeItem::~TreeItem()
{
  while(m_children->getSize() > 0)
    delete m_children->remove(0);
}

void te::qt::widgets::TreeItem::setParent(TreeItem* item)
{
  m_parent = item;
}

void te::qt::widgets::TreeItem::addChild(TreeItem* item)
{
  m_children->add(item);

  item->setParent(this);
}

void te::qt::widgets::TreeItem::insertChild(TreeItem* item, const size_t& pos) 
{
  if(pos > m_children->getSize())
    throw te::common::Exception("Index out of boundaries.");

  m_children->insert(item, (unsigned int) pos);

  item->setParent(this);
}

te::qt::widgets::TreeItem* te::qt::widgets::TreeItem::getChild(const size_t& pos) const 
{
  if(pos >= m_children->getSize())
    throw te::common::Exception("Index out of boundaries.");

  return m_children->getValue((unsigned int) pos);
}

te::qt::widgets::TreeItem* te::qt::widgets::TreeItem::getParent() const
{
  return m_parent;
}

te::qt::widgets::TreeItem* te::qt::widgets::TreeItem::removeChild(const size_t& pos) 
{
  if(pos >= m_children->getSize())
    throw te::common::Exception("Index out of boundaries.");

  return m_children->remove((unsigned int)pos);
}

size_t te::qt::widgets::TreeItem::getChildrenCount(const std::string& type) const
{
  if(type.empty())
    return (size_t) m_children->getSize();

  std::vector<TreeItem*> items;

  getChildren(items, type);

  return items.size();
}

void te::qt::widgets::TreeItem::getChildren(std::vector<TreeItem*>& items, const std::string& type) const
{
  te::common::NodeT<TreeItem*>* aux = m_children->m_first;

  items.clear();

  while(aux != nullptr)
  {
    if(aux->m_value->m_type == type || type.empty())
      items.push_back(aux->m_value);

    aux = aux->m_next;
  }
}


bool te::qt::widgets::TreeItem::hasChildren() const
{
  return (getChildrenCount("") != 0);
}

int te::qt::widgets::TreeItem::getPosition()
{
  return (m_parent == nullptr) ? -1 : (int) m_parent->m_children->position(this);
}

te::qt::widgets::VISIBLE te::qt::widgets::TreeItem::isVisible() const
{
  return NONE;
}

void te::qt::widgets::TreeItem::setVisible(const VISIBLE&, const bool&, const bool&)
{
}

std::string te::qt::widgets::TreeItem::getType() const
{
  return m_type;
}

std::string te::qt::widgets::TreeItem::getToolTip() const
{
  return std::string();
}

void te::qt::widgets::TreeItem::removeAllChilds()
{
  while(m_children->getSize() > 0)
    delete m_children->remove(0);
}

Qt::ItemFlags te::qt::widgets::TreeItem::flags()
{
  return Qt::ItemIsSelectable | Qt::ItemIsDropEnabled | Qt::ItemIsEnabled;
}
