#include "DefaultFeatureInfoMaker.h"

// TerraLib
#include "../../../../common/STLUtils.h"
#include "../../../../dataaccess/dataset/DataSet.h"
#include "../../../../dataaccess/dataset/DataSetType.h"
#include "../../../../dataaccess/utils/Utils.h"
#include "../../../../geometry/Coord2D.h"
#include "../../../../geometry/Envelope.h"
#include "../../../../geometry/Geometry.h"
#include "../../../../geometry/GeometryProperty.h"
#include "../../../../geometry/Point.h"
#include "../../../../geometry/Utils.h"
#include "../../../../geometry/Validation.h"
#include "../../../../maptools/AbstractMarkRenderer.h"
#include "../../../../maptools/MarkRendererManager.h"
#include "../../../../raster/Raster.h"
#include "../../../../raster/RasterProperty.h"
#include "../../../../raster/Utils.h"
#include "../../../../se/Fill.h"
#include "../../../../se/Mark.h"
#include "../../../../se/Stroke.h"
#include "../../../../se/Utils.h"
#include "../../../../srs/Config.h"
#include "../../../widgets/canvas/MapDisplay.h"
#include "../../Utils.h"
#include "../../canvas/Canvas.h"
#include "../../canvas/MapDisplay.h"

// Qt
#include <QLabel>
#include <QMessageBox>
#include <QMouseEvent>
#include <QObject>
#include <QtCore/QPointF>
#include <QtCore/QString>
#include <QtCore/QStringList>

// STL
#include <cassert>
#include <memory>

Q_DECLARE_METATYPE(std::shared_ptr<te::gm::Geometry>)

te::qt::widgets::DefaultFeatureInfoMaker::DefaultFeatureInfoMaker() = default;

te::qt::widgets::DefaultFeatureInfoMaker::~DefaultFeatureInfoMaker() = default;

void te::qt::widgets::DefaultFeatureInfoMaker::getInfo(
    QTreeWidgetItem* layerItem, MapDisplay* display,
    const te::map::AbstractLayerPtr& layer, const te::gm::Envelope& e)
{
  if(layer->getVisibility() != te::map::VISIBLE)
    return;

  bool needRemap = false;
  te::gm::Envelope reprojectedEnvelope(e);

  if((layer->getSRID() != TE_UNKNOWN_SRS) &&
     (display->getSRID() != TE_UNKNOWN_SRS) &&
     (layer->getSRID() != display->getSRID()))
  {
    needRemap = true;
    reprojectedEnvelope.transform(display->getSRID(), layer->getSRID());
  }

  if(!reprojectedEnvelope.intersects(layer->getExtent()))
    return;

  try
  {
    // Gets the Layer Schema
    std::unique_ptr<const te::map::LayerSchema> ls(layer->getSchema());

    // Gets the name of the referenced spatial property
    std::string spatialPropertyName = layer->getGeomPropertyName();

    if(ls->hasGeom())
    {
      te::gm::GeometryProperty* gp = nullptr;

      if(spatialPropertyName.empty())
        gp = te::da::GetFirstGeomProperty(ls.get());
      else
        gp = dynamic_cast<te::gm::GeometryProperty*>(
            ls->getProperty(spatialPropertyName));

      assert(gp);

      // Retrieves the data from layer
      std::unique_ptr<te::da::DataSet> dataset(
          layer
              ->getData(gp->getName(), &reprojectedEnvelope, te::gm::INTERSECTS)
              .release());

      getGeometryInfo(layerItem, display, dataset.get(),
                      layer->getGeomPropertyName(), reprojectedEnvelope,
                      layer->getSRID(), needRemap);
    }

    if(ls->hasRaster())
    {
      te::rst::RasterProperty* rp = nullptr;

      if(spatialPropertyName.empty())
        rp = te::da::GetFirstRasterProperty(ls.get());
      else
        rp = dynamic_cast<te::rst::RasterProperty*>(
            ls->getProperty(spatialPropertyName));

      assert(rp);

      // Retrieves the data from layer
      std::unique_ptr<te::da::DataSet> dataset(
          layer
              ->getData(rp->getName(), &reprojectedEnvelope, te::gm::INTERSECTS)
              .release());

      if(!dataset->moveNext())
        return;

      std::unique_ptr<te::rst::Raster> raster(
          dataset->getRaster(rp->getName()));
      assert(raster.get());

      getRasterInfo(layerItem, display, raster.get(), reprojectedEnvelope,
                    layer->getSRID(), needRemap);
    }
  }
  catch(...)
  {
    delete layerItem;
    // TODO: catch the exceptions...
  }
}

void te::qt::widgets::DefaultFeatureInfoMaker::getGeometryInfo(
    QTreeWidgetItem* layerItem, MapDisplay* display, te::da::DataSet* dataset,
    const std::string& geomPropertyName, const te::gm::Envelope& e, int srid,
    bool needRemap)
{
  // Generates a geometry from the given extent. It will be used to refine the
  // results
  std::unique_ptr<te::gm::Geometry> geometryFromEnvelope(
      te::gm::GetGeomFromEnvelope(&e, srid));

  // The restriction point. It will be used to refine the results
  te::gm::Coord2D center = e.getCenter();
  te::gm::Point point(center.x, center.y, srid);

  std::size_t gpos = std::string::npos;
  geomPropertyName.empty()
      ? gpos = te::da::GetFirstPropertyPos(dataset, te::dt::GEOMETRY_TYPE)
      : gpos = te::da::GetPropertyPos(dataset, geomPropertyName);

  while(dataset->moveNext())
  {
    std::unique_ptr<te::gm::Geometry> g(dataset->getGeometry(gpos));

    if(g.get() == nullptr)
      continue;

    g->setSRID(srid);

    bool showInfo = false;
    bool hadError = false;
    //we first try to check the geometry relation without checking the geomtry validity
    //this will work on almost all situations
    try
    {
      std::unique_ptr<te::gm::Geometry> geometryCopy(te::gm::ClipGeometry(g.get(), e));
      if (te::gm::IsNullOrEmpty(geometryCopy.get()) == false)
      {
        showInfo = geometryCopy->contains(&point) || geometryCopy->crosses(geometryFromEnvelope.get()) || geometryFromEnvelope->contains(geometryCopy.get());
      }
    }
    catch (std::exception& )
    {
      hadError = true;
    }

    //if we got an error, we try to make the geometry valid and try again
    if (hadError == true)
    {
      try
      {
        std::unique_ptr<te::gm::Geometry> geometryCopy(te::gm::ClipGeometry(g.get(), e));
        geometryCopy.reset(te::gm::Validation::MakeValidWithSnap(geometryCopy.get()).release());
        if (te::gm::IsNullOrEmpty(geometryCopy.get()) == false)
        {
          showInfo = geometryCopy->contains(&point) || geometryCopy->crosses(geometryFromEnvelope.get()) || geometryFromEnvelope->contains(geometryCopy.get());
        }
      }
      catch (std::exception& e)
      {
        QString message = QObject::tr("Invalid geometry: ") + e.what();
        QMessageBox::warning(display, QObject::tr("Information"), message);
        return;
      }
    }
    

    if(showInfo == true)
    {
      QTreeWidgetItem* rootItem = nullptr;

      // Feature found! Building the list of property values...
      for(std::size_t i = 0; i < dataset->getNumProperties(); ++i)
      {
        if(dataset->getPropertyDataType(i) == te::dt::RASTER_TYPE)
          continue;

        if(i == 0)
        {
          rootItem = createItemFromGeometryInfo(i, layerItem, dataset);
        }
        else
        {
          createItemFromGeometryInfo(i, rootItem, dataset);
        }
      }

      if(needRemap)
      {
        g->setSRID(srid);
        g->transform(display->getSRID());
      }

      drawGeometry(display, g.get());  // to show feedback!
    }
  }
}

QTreeWidgetItem*
te::qt::widgets::DefaultFeatureInfoMaker::createItemFromGeometryInfo(
    std::size_t index, QTreeWidgetItem* layerItem, te::da::DataSet* dataset)
{
  QTreeWidgetItem* propertyItem = new QTreeWidgetItem(layerItem);
  propertyItem->setText(
      0, te::qt::widgets::Convert2QString(dataset->getPropertyName(index)));

  if(dataset->getPropertyDataType(index) == te::dt::GEOMETRY_TYPE)
    propertyItem->setIcon(0, QIcon::fromTheme("geometry"));

  if(!dataset->isNull(index))
  {
    QString qvalue;

    if(dataset->getPropertyDataType(index) == te::dt::STRING_TYPE)
    {
      std::string value = dataset->getString(index);
      qvalue = te::qt::widgets::Convert2QString(value);
    }
    else if(dataset->getPropertyDataType(index) == te::dt::GEOMETRY_TYPE)
    {
      std::shared_ptr<te::gm::Geometry> geometry(dataset->getGeometry(index).release());
      propertyItem->setData(1, Qt::UserRole, QVariant::fromValue< std::shared_ptr<te::gm::Geometry> >(geometry));

      qvalue = QObject::tr("Geometry String");
    }
    else
    {
      qvalue = te::qt::widgets::Convert2QString(dataset->getAsString(index, 6));
    }

    //sets the text to the base item
    propertyItem->setText(1, qvalue);

    //to allow word wrap, we must use a QLabel together with the QTreeWidgetItem
    QTreeWidget* treeWidget = propertyItem->treeWidget();

    QLabel* qLabel = new QLabel(qvalue, treeWidget);

    //eables the word wrap
    qLabel->setWordWrap(true);
 
    //sets the QLabel to be used together with the QTreeWidgetItem
    treeWidget->setItemWidget(propertyItem, 1, qLabel);

    //Hack: We want to keep out text also in the QTreeWidgetItem for compatibility reason,
    //but the default behavior is to display both QTreeWidgetItem and QLabel texts
    //This messes up the component
    //For this reason, we "paint" the QTreeWidgetItem text color with the same color as its background
    //This makes the text from the QTreeWidgetItem to be invisible
    propertyItem->setTextColor(1, QColor(0, 0, 0, 0));
  }
  else  // property null value!
  {
    propertyItem->setText(1, "");
  }

  return propertyItem;
}

void te::qt::widgets::DefaultFeatureInfoMaker::getRasterInfo(
    QTreeWidgetItem* layerItem, MapDisplay* /*display*/,
    te::rst::Raster* raster, const te::gm::Envelope& e, int /*srid*/,
    bool /*needRemap*/)
{
  te::gm::Coord2D center = e.getCenter();
  te::gm::Coord2D coord = raster->getGrid()->geoToGrid(center.x, center.y);

  int x = te::rst::Round(coord.x);
  int y = te::rst::Round(coord.y);

  if((x < 0 || x >= static_cast<int>(raster->getNumberOfColumns())) ||
     (y < 0 || y >= static_cast<int>(raster->getNumberOfRows())))
    return;

  // Raster location informations
  QTreeWidgetItem* lineItem = new QTreeWidgetItem(layerItem);
  lineItem->setText(0, QObject::tr("Line"));
  lineItem->setText(1, QString::number(y));

  QTreeWidgetItem* columnItem = new QTreeWidgetItem(layerItem);
  columnItem->setText(0, QObject::tr("Column"));
  columnItem->setText(1, QString::number(x));

  QTreeWidgetItem* coordXItem = new QTreeWidgetItem(layerItem);
  coordXItem->setText(0, QObject::tr("Coordinate X"));
  coordXItem->setText(1, QString::number(center.x));

  QTreeWidgetItem* coordYItem = new QTreeWidgetItem(layerItem);
  coordYItem->setText(0, QObject::tr("Coordinate Y"));
  coordYItem->setText(1, QString::number(center.y));

  // Raster value informations for each band
  double value = 0.0;
  for(std::size_t b = 0; b < raster->getNumberOfBands(); ++b)
  {
    raster->getValue(x, y, value, b);

    QTreeWidgetItem* bandValueItem = new QTreeWidgetItem(layerItem);
    bandValueItem->setIcon(0, QIcon::fromTheme("channel-gray"));
    bandValueItem->setText(0, QObject::tr("Band") + " " + QString::number(b));
    bandValueItem->setText(1, QString::number(value, 'f', 2));
  }
}

void te::qt::widgets::DefaultFeatureInfoMaker::drawGeometry(MapDisplay* display,
                                                            te::gm::Geometry* g)
{
  QPixmap* draft = display->getDraftPixmap();

  const te::gm::Envelope& displayExtent = display->getExtent();

  bool deleteGeometry = false;
  if (g->getMBR()->within(displayExtent) == false)
  {
    g = te::gm::ClipGeometry(g, displayExtent);
    deleteGeometry = true;
  }

  Canvas canvas(draft);
  canvas.setWindow(displayExtent.m_llx, displayExtent.m_lly,
                   displayExtent.m_urx, displayExtent.m_ury);
  canvas.setRenderHint(QPainter::Antialiasing, true);

  switch(g->getGeomTypeId())
  {
    case te::gm::PolygonType:
    case te::gm::PolygonZType:
    case te::gm::PolygonMType:
    case te::gm::PolygonZMType:
    case te::gm::MultiPolygonType:
    case te::gm::MultiPolygonZType:
    case te::gm::MultiPolygonMType:
    case te::gm::MultiPolygonZMType:
    {
      canvas.setPolygonContourWidth(2);
      canvas.setPolygonContourColor(te::color::RGBAColor(255, 0, 0, 128));
      canvas.setPolygonFillColor(te::color::RGBAColor(255, 255, 255, 128));
    }
    break;

    case te::gm::LineStringType:
    case te::gm::LineStringZType:
    case te::gm::LineStringMType:
    case te::gm::LineStringZMType:
    case te::gm::MultiLineStringType:
    case te::gm::MultiLineStringZType:
    case te::gm::MultiLineStringMType:
    case te::gm::MultiLineStringZMType:
    {
      canvas.setLineColor(te::color::RGBAColor(255, 0, 0, 128));
      canvas.setLineWidth(6);
    }
    break;

    case te::gm::PointType:
    case te::gm::PointZType:
    case te::gm::PointMType:
    case te::gm::PointZMType:
    case te::gm::MultiPointType:
    case te::gm::MultiPointZType:
    case te::gm::MultiPointMType:
    case te::gm::MultiPointZMType:
    {
      std::size_t size = 24;

      te::se::Stroke* stroke = te::se::CreateStroke("#FF0000", "2", "0.5");
      te::se::Fill* fill = te::se::CreateFill("#FFFFFF", "0.5");
      te::se::Mark* mark = te::se::CreateMark("square", stroke, fill);

      te::map::MarkRendererParams params(mark, size);
      te::color::RGBAColor** rgba =
          te::map::MarkRendererManager::getInstance().render(params);

      canvas.setPointColor(te::color::RGBAColor(0, 0, 0, TE_TRANSPARENT));
      canvas.setPointPattern(rgba, static_cast<int>(size),
                             static_cast<int>(size));

      te::common::Free(rgba, size);
      delete mark;
    }
    break;

    default:
      return;
  }

  canvas.draw(g);

  if (deleteGeometry == true)
  {
    delete g;
  }
}