/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/widgets/layer/utils/DataSet2Layer.cpp

  \brief ....
*/

// TerraLib
#include "../../../../common/StringUtils.h"
#include "../../../../core/translator/Translator.h"
#include "../../../../dataaccess/dataset/DataSetType.h"
#include "../../../../dataaccess/datasource/DataSourceManager.h"
#include "../../../../dataaccess/datasource/DataSourceTransactor.h"
#include "../../../../dataaccess/utils/Utils.h"
#include "../../../../geometry/Envelope.h"
#include "../../../../geometry/GeometryProperty.h"
#include "../../../../maptools/AbstractLayer.h"
#include "../../../../maptools/DataSetLayer.h"
#include "../../../../raster/Grid.h"
#include "../../../../raster/RasterProperty.h"
#include "../../../../se/Utils.h"
#include "../../../../srs/Config.h"
#include "../../Exception.h"
#include "DataSet2Layer.h"

te::qt::widgets::DataSet2Layer::DataSet2Layer(const std::string& datasourceId)
  : m_datasourceId(datasourceId)
{
}

te::map::DataSetLayerPtr te::qt::widgets::DataSet2Layer::operator()(const te::da::DataSetTypePtr& dataset) const
{
  return operator()(dataset, "");
}

te::map::DataSetLayerPtr te::qt::widgets::DataSet2Layer::operator()(const te::da::DataSetTypePtr& dataset, const std::string& geomPropertyName) const
{
  if(dataset.get() == nullptr)
    throw Exception(TE_TR("Can not convert a NULL dataset to a layer!"));

  std::string id = te::common::GetUniqueID();

  std::string title = dataset->getTitle().empty() ? dataset->getName() : dataset->getTitle();
  
  te::map::DataSetLayerPtr layer(new te::map::DataSetLayer(id, title));
  layer->setDataSetName(dataset->getName());
  layer->setDataSourceId(m_datasourceId);
  layer->setVisibility(te::map::VISIBLE);
  layer->setRendererType("ABSTRACT_LAYER_RENDERER");

  if(dataset->size() == 0)
  {
    te::da::DataSourcePtr ds(te::da::DataSourceManager::getInstance().find(m_datasourceId));
    te::da::LoadProperties(dataset.get(), m_datasourceId);
  }

  if(dataset->hasRaster())
  {
    te::rst::RasterProperty* rp = nullptr;
    if (geomPropertyName.empty() == true)
    {
      rp = te::da::GetFirstRasterProperty(dataset.get());
    }
    else
    {
      rp = dynamic_cast<te::rst::RasterProperty*>(dataset->getProperty(geomPropertyName));
    }
    assert(rp);

    te::rst::Grid* grid = rp->getGrid();

    layer->setSRID(grid->getSRID());
    layer->setExtent(*(grid->getExtent()));
    layer->setStyle(te::se::CreateCoverageStyle(te::da::GetFirstRasterProperty(dataset.get())->getBandProperties()));
    if (geomPropertyName.empty() == false)
    {
      layer->setGeomPropertytName(geomPropertyName);
    }
  }  
  else if(dataset->hasGeom())
  {
    te::gm::GeometryProperty* gp = nullptr;
    if (geomPropertyName.empty() == true)
    {
      gp = te::da::GetFirstGeomProperty(dataset.get());
    }
    else
    {
      gp = dynamic_cast<te::gm::GeometryProperty*>(dataset->getProperty(geomPropertyName));
    }

    assert(gp);

    std::unique_ptr<te::gm::Envelope> mbr;
    try
    {
      mbr.reset(te::da::GetExtent(dataset->getName(), gp->getName(), m_datasourceId));
    }
    catch (const std::exception&)
    {
    }

    layer->setSRID(gp->getSRID());
    if (mbr.get() != nullptr)
    {
      layer->setExtent(*mbr);
    }
      
    layer->setStyle(te::se::CreateFeatureTypeStyle(gp->getGeometryType()));
    if (geomPropertyName.empty() == false)
    {
      layer->setGeomPropertytName(geomPropertyName);
    }  
  }
  else
  {
    layer->setSRID(TE_UNKNOWN_SRS);
  }

  return layer;
}

te::map::DataSetLayerPtr te::qt::widgets::CreateLayerFromDataSet(const std::string& datasourceId, const std::string& dataSetName, const std::string& geomPropertyName, int sridIfUnknown)
{
  // creating a layer for the result
  te::da::DataSourcePtr dataSource = te::da::GetDataSource(datasourceId);
  if (dataSource->isOpened() == false)
  {
    dataSource->open();
  }

  te::qt::widgets::DataSet2Layer converter(datasourceId);

  te::da::DataSetTypePtr dt(dataSource->getDataSetType(dataSetName).release());
  te::map::DataSetLayerPtr newLayer= converter(dt);
  if (newLayer->getSRID() == TE_UNKNOWN_SRS && sridIfUnknown != TE_UNKNOWN_SRS)
  {
    newLayer->setSRID(sridIfUnknown);
  }

  return newLayer;
}

te::map::DataSetLayerPtr te::qt::widgets::CreateLayerFromDataSet(const std::string& datasourceId, const std::string& dataSetName, int sridIfUnknown)
{
  return CreateLayerFromDataSet(datasourceId, dataSetName, "", sridIfUnknown);
}

