/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/widgets/layer/utils/RasterContrastMenuWidget.cpp

  \brief This file defines a class for a RasterContrastMenuWidget.
*/

// TerraLib
#include "../../../../dataaccess/utils/Utils.h"
#include "../../../../maptools/RasterContrast.h"
#include "../../../../maptools/Utils.h"
#include "../../../../se/RasterSymbolizer.h"
#include "../../../../se/CoverageStyle.h"
#include "../../../../se/Rule.h"
#include "../../../../se/RasterSymbolizer.h"
#include "../../../../se/ChannelSelection.h"
#include "../../../../se/SelectedChannel.h"
#include "../../../../rp/Contrast.h"
#include "../../canvas/MapDisplay.h"
#include "../../utils/ScopedCursor.h"
#include "RasterContrastMenuWidget.h"

// Qt
#include <QActionGroup>

te::qt::widgets::RasterContrastMenuWidget::RasterContrastMenuWidget(te::qt::widgets::MapDisplay* mapDisplay, QWidget* parent, Qt::WindowFlags f): 
  QWidget(parent, f), 
  m_mapDisplay(mapDisplay)
{
  //create menu
  m_menu = new QMenu(tr("Auto Contrast"), this);

  //create actions
  QAction* actMinMax = new QAction(tr("Min Max"), this);
  QAction* act2Percent = new QAction(tr("2% Enhancement"), this);
  m_actClear = new QAction(tr("Clear"), this);
    
  //add actions to menu
  m_menu->addAction(actMinMax);
  m_menu->addAction(act2Percent);
  m_menu->addAction(m_actClear);

  //connections
  connect(actMinMax, SIGNAL(triggered()), this, SLOT(onContrastMinMaxClicked()));
  connect(act2Percent, SIGNAL(triggered()), this, SLOT(onContrast2PercentClicked()));
  connect(m_actClear, SIGNAL(triggered()), this, SLOT(onContrastClearClicked()));
}

te::qt::widgets::RasterContrastMenuWidget::~RasterContrastMenuWidget() = default;

void te::qt::widgets::RasterContrastMenuWidget::setLayer(te::map::AbstractLayerPtr layer)
{
  m_layer = layer;

  if (m_layer->getRasterContrast() == nullptr)
  {
    m_actClear->setEnabled(false);
  }
  else
  {
    m_actClear->setEnabled(true);
  }
}

QMenu* te::qt::widgets::RasterContrastMenuWidget::getMenu()
{
  return m_menu;
}

void te::qt::widgets::RasterContrastMenuWidget::onContrastMinMaxClicked()
{
  te::qt::widgets::ScopedCursor cursor(Qt::WaitCursor);

  std::unique_ptr< te::rst::Raster > raster( te::map::GetRaster(m_layer.get()) );

  if (raster.get() != nullptr)
  {
    std::vector< int > bands;
    getRasterSelectedBands( bands );

    te::map::RasterContrast* contrast = new te::map::RasterContrast(
      te::map::RasterTransform::LINEAR_CONTRAST, raster->getNumberOfBands() );

    const unsigned int maxRowsCols = std::max( raster->getNumberOfColumns(),
      raster->getNumberOfRows() );
    unsigned int sampleStep = std::max( 1u, maxRowsCols / 32 );

    for( std::size_t bandsIdx = 0 ; bandsIdx < bands.size() ; ++bandsIdx )
    {
      const int bandIdx = bands[ bandsIdx ];

      te::rst::BandSummary summary;
      if( ! te::rst::RasterSummaryManager::getInstance().get(
        raster.get(),
        bandIdx,
        0, // rowStart
        0, // colStart
        raster->getNumberOfRows() - 1,
        raster->getNumberOfColumns() - 1,
        0, // histoBins
        sampleStep,
        false, // readall
        summary ) )
      {
        throw te::common::Exception(TE_TR("Raster summary Error"));
      }

      double min = summary.m_minVal->real();
      double max = summary.m_maxVal->real();

      double gain = 1.;
      double offset1 = 0.;
      double offset2 = 0.;

      if (te::rp::Contrast::getGainAndOffset(
        te::rp::Contrast::InputParameters::LinearContrastT, min, max, 0., 255.,
        gain, offset1, offset2))
      {
        contrast->setValues(gain, offset1, offset2, min, max, bandIdx);
      }
    }

    m_layer->setRasterContrast(contrast);
  }

  m_mapDisplay->updateLayer(m_layer, true);
}

void te::qt::widgets::RasterContrastMenuWidget::onContrast2PercentClicked()
{
  te::qt::widgets::ScopedCursor cursor(Qt::WaitCursor);

  std::unique_ptr< te::rst::Raster > raster( te::map::GetRaster(m_layer.get()) );

  if (raster.get() != nullptr)
  {
    std::vector< int > bands;
    getRasterSelectedBands( bands );

    te::map::RasterContrast* contrast = new te::map::RasterContrast(
      te::map::RasterTransform::LINEAR_CONTRAST, raster->getNumberOfBands() );

    const unsigned int maxRowsCols = std::max( raster->getNumberOfColumns(),
      raster->getNumberOfRows() );
    unsigned int sampleStep = std::max( 1u, maxRowsCols / 32 );

    for( std::size_t bandsIdx = 0 ; bandsIdx < bands.size() ; ++bandsIdx )
    {
      const int bandIdx = bands[ bandsIdx ];

      te::rst::BandSummary summary;
      if( ! te::rst::RasterSummaryManager::getInstance().get(
        raster.get(),
        bandIdx,
        0, // rowStart
        0, // colStart
        raster->getNumberOfRows() - 1,
        raster->getNumberOfColumns() - 1,
        0, // histoBins
        sampleStep,
        false, // readall
        summary ) )
      {
        throw te::common::Exception(TE_TR("Raster summary Error"));
      }

      // get 2% histogram values
      std::map<double, unsigned>& realHist( *summary.m_histogramR );

      double totalCount = 0.;

      for (std::map<double, unsigned>::iterator it = realHist.begin(); it != realHist.end(); ++it)
      {
        totalCount += it->second;
      }

      double minCount = totalCount * 0.02;
      double maxCount = totalCount * 0.98;

      double count = 0.;
      double min = 0.;
      double max = 0.;

      for (std::map<double, unsigned>::iterator it = realHist.begin(); it != realHist.end(); ++it)
      {
        count += it->second;

        if (count < minCount)
        {
          min = (*it).first;
        }

        if (maxCount <= count)
        {
          max = (*it).first;
          break;
        }
      }

      double gain = 1.;
      double offset1 = 0.;
      double offset2 = 0.;

      if (te::rp::Contrast::getGainAndOffset(
        te::rp::Contrast::InputParameters::LinearContrastT, min, max, 0., 255.,
        gain, offset1, offset2))
      {
        contrast->setValues(gain, offset1, offset2, min, max, bandIdx);
      }
    }

    m_layer->setRasterContrast(contrast);
  }

  m_mapDisplay->updateLayer(m_layer, true);
}

void te::qt::widgets::RasterContrastMenuWidget::onContrastClearClicked()
{
  m_layer->setRasterContrast(nullptr);

  m_mapDisplay->updateLayer(m_layer, true);
}

void te::qt::widgets::RasterContrastMenuWidget::getRasterSelectedBands( std::vector< int >& bands )
{
  bands.clear();

  if( m_layer.get() != nullptr )
  {
    if( m_layer->getStyle() != nullptr )
    {
      te::se::CoverageStyle* stylePtr = dynamic_cast< te::se::CoverageStyle* >(
        m_layer->getStyle() );

      if( stylePtr )
      {
        const te::se::Rule* rulePtr = stylePtr->getRule(0);

        if( rulePtr != nullptr )
        {
          const std::vector<te::se::Symbolizer*>& symbolizers = rulePtr->getSymbolizers();

          if( !symbolizers.empty() )
          {
            te::se::RasterSymbolizer* rasterSymbolizerPtr =
              dynamic_cast<te::se::RasterSymbolizer*>(symbolizers[0]);

            if( rasterSymbolizerPtr != nullptr )
            {
              te::se::ChannelSelection* csPtr = rasterSymbolizerPtr->getChannelSelection();

              if( csPtr != nullptr )
              {
                te::se::SelectedChannel* rCsPtr = csPtr->getRedChannel();
                te::se::SelectedChannel* gCsPtr = csPtr->getGreenChannel();
                te::se::SelectedChannel* bCsPtr = csPtr->getBlueChannel();
                te::se::SelectedChannel* grayCsPtr = csPtr->getGrayChannel();

                if(
                    ( rCsPtr != nullptr )
                    &&
                    (!rCsPtr->getSourceChannelName().empty() )
                  )
                {
                  const int bandIndex = std::atoi(rCsPtr->getSourceChannelName().c_str());
                  bands.push_back( bandIndex );
                }

                if(
                    ( gCsPtr != nullptr )
                    &&
                    (!gCsPtr->getSourceChannelName().empty() )
                  )
                {
                  const int bandIndex = std::atoi(gCsPtr->getSourceChannelName().c_str());
                  bands.push_back( bandIndex );
                }

                if(
                    ( bCsPtr != nullptr )
                    &&
                    (!bCsPtr->getSourceChannelName().empty() )
                  )
                {
                  const int bandIndex = std::atoi(bCsPtr->getSourceChannelName().c_str());
                  bands.push_back( bandIndex );
                }

                if(
                    ( grayCsPtr != nullptr )
                    &&
                    (!grayCsPtr->getSourceChannelName().empty() )
                  )
                {
                  const int bandIndex = std::atoi(grayCsPtr->getSourceChannelName().c_str());
                  bands.push_back( bandIndex );
                }
              }
            }
          }
        }
      }
    }
  }
}
