/*
  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

  This file is part of the TerraLib - a Framework for building GIS enabled
  applications.

  TerraLib is free software: you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation, either version 3 of the License,
  or (at your option) any later version.

  TerraLib is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with TerraLib. See COPYING. If not, write to
  TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/widgets/ceditor/LogTextEdit.h

  \brief A TextEdit that can be used to watch a log file.
*/

// TerraLib
#include "LogTextEdit.h"

// STL
#include <functional>

// QT
#include <QColor>
#include <QFile>
#include <QTextStream>
#if(QT_VERSION >= 0x050000)
#include <QStandardPaths>
#else
#include <QDesktopServices>
#endif

te::qt::widgets::LogTextEdit::LogTextEdit(QWidget *parent,
                                          const QString &fileName)
    : QPlainTextEdit(parent), m_watcher(nullptr)
{
  setReadOnly(true);
  setTextInteractionFlags(Qt::TextSelectableByMouse |
                          Qt::TextSelectableByKeyboard);
  setLineWrapMode(QPlainTextEdit::NoWrap);
  QString tempName;
  tempName = fileName;

  if (tempName.isEmpty())
  {
    QString userDataDir;
#if(QT_VERSION >= 0x050000)
    userDataDir =
        QStandardPaths::writableLocation(QStandardPaths::DataLocation);
#else
    userDataDir =
        QDesktopServices::storageLocation(QDesktopServices::DataLocation);
#endif
    tempName = userDataDir + "/log/TerraLib.log";
  }

  m_watcher.reset(new te::core::FileWatcher(
    tempName.toUtf8().data(),
      std::bind(&te::qt::widgets::LogTextEdit::textReady, this)));

  m_watcher->watch();

  m_file.setFileName(tempName);
  QTextStream ifile(&m_file);

  m_file.open(QFile::ReadOnly);

  setPlainText(ifile.readAll());

  m_pos = ifile.pos();

  moveCursor(QTextCursor::End);

  connect(this, SIGNAL(textPending()), this, SLOT(read()));
}

te::qt::widgets::LogTextEdit::~LogTextEdit()
{
  m_watcher->unwatch();
  m_file.close();
}

void te::qt::widgets::LogTextEdit::read()
{
  QTextStream ifile(&m_file);

  ifile.seek(m_pos);

  moveCursor(QTextCursor::End);

  while (!ifile.atEnd())
  {
    insertPlainText(ifile.readLine());
    insertPlainText("\n");
  }

  m_pos = ifile.pos();
}

void te::qt::widgets::LogTextEdit::textReady()
{                       
  emit textPending();   
}

void te::qt::widgets::LogTextEdit::findAll(const QString &str)
{
  QList<QTextEdit::ExtraSelection> matches;

  moveCursor(QTextCursor::Start);

  auto bgColor = QColor(Qt::gray).lighter(135);

  while(find(str))
  {
    QTextEdit::ExtraSelection extra;

    extra.format.setBackground(bgColor);

    extra.cursor = textCursor();

    matches.append(extra);
  }

  setExtraSelections(matches);

  moveCursor(QTextCursor::End);
}
