/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/widgets/progress/ProgressViewerDialog.cpp

  \brief A progress dialog.
*/

// Terralib
#include "ProgressViewerDialog.h"
#include "../Utils.h"

#include "../../../core/translator/Translator.h"

te::qt::widgets::ProgressViewerDialog::ProgressViewerDialog(QWidget* parent, QString title)
    : AbstractQtProgressViewer(parent), 
  m_totalSteps(0), 
  m_currentStep(0), 
  m_propStep(0), 
  m_message("")
{
  m_dlgProgress = new QProgressDialog(this);
  m_dlgProgress->setWindowModality(Qt::WindowModal);
  m_dlgProgress->setAutoClose(true);
  m_dlgProgress->setRange(0, 100);

  connect(m_dlgProgress, SIGNAL(canceled()), this, SLOT(cancel()));
}

te::qt::widgets::ProgressViewerDialog::~ProgressViewerDialog()
{
  delete m_dlgProgress;
}

void te::qt::widgets::ProgressViewerDialog::addTask(te::common::TaskProgress* t, int id)
{
  m_reset = false;

  m_tasks.insert(std::map<int, te::common::TaskProgress*>::value_type(id, t));

  updateMessage(-1);
}

void te::qt::widgets::ProgressViewerDialog::removeTask(int taskId)
{
  cancelTask(taskId);

  std::map<int, te::common::TaskProgress*>::iterator it = m_tasks.find(taskId);

  if(it != m_tasks.end())
  {
    m_tasks.erase(it);
    updateMessage(-1);
  }

  if(m_tasks.empty())
  {
    m_totalSteps = 0;
    m_currentStep = 0;
    m_propStep = 0;

    emit setProgressRange(0, 100);

    emit resetProgress();
  }
}

void te::qt::widgets::ProgressViewerDialog::cancelTask(int taskId)
{
  std::map<int, te::common::TaskProgress*>::iterator it = m_tasks.find(taskId);

  if(it != m_tasks.end())
  {
    // update total and current values
    m_totalSteps -= it->second->getTotalSteps();
    m_currentStep -= it->second->getCurrentStep();

    if (m_totalSteps > 0)
    {
      double aux = static_cast<double> (m_currentStep) / static_cast<double> (m_totalSteps);

      m_propStep = static_cast<int> (100.0 * aux);
    }
    else
    {
      m_propStep = 0;
    }
  }
}

void te::qt::widgets::ProgressViewerDialog::setTotalValues(int taskId)
{
  m_totalSteps += m_tasks[taskId]->getTotalSteps();

  if (m_totalSteps == 0)
  {
    emit setProgressRange(0, 0);
  }
  else
  {
    emit setProgressRange(0, 100);
  }
}

void te::qt::widgets::ProgressViewerDialog::setTitle(QString title)
{
  m_dlgProgress->setWindowTitle(title);
}

void te::qt::widgets::ProgressViewerDialog::updateValue(int taskId)
{
  std::lock_guard<std::mutex> mutexGuard(m_mutex);

  std::map<int, te::common::TaskProgress*>::iterator it;

  double currentStep = 0;

  for (it = m_tasks.begin(); it != m_tasks.end(); ++it)
    currentStep += it->second->getCurrentStep();

  it = m_tasks.find(taskId);

  double aux = static_cast<double>(currentStep) / static_cast<double>(m_totalSteps);

  int val = static_cast<int>(100.0 * aux);

  if(val != m_propStep && val >= 0.0)
  {
    m_propStep = val;

    emit setProgressValue(m_propStep);
  }
}

void te::qt::widgets::ProgressViewerDialog::updateMessage(int /*taskId*/)
{
  if(m_tasks.size() == 1)
  {
    m_message = m_tasks.begin()->second->getMessage();
  }
  else
  {
    m_message = TE_TR("Multi Tasks");
  }

  emit setProgressMessage(m_message);
}

void te::qt::widgets::ProgressViewerDialog::onSetProgressMessage(const std::string& message)
{
  m_dlgProgress->show();
  m_dlgProgress->setLabelText(te::qt::widgets::Convert2QString(message));
}

void te::qt::widgets::ProgressViewerDialog::onSetProgressRange(int min, int max)
{
  m_dlgProgress->setRange(min, max);
}

void te::qt::widgets::ProgressViewerDialog::onSetProgressValue(int value)
{
  if (value >= m_propStep && m_reset == false)
  {
    m_dlgProgress->show();
    m_dlgProgress->setValue(value);
  }
}

void te::qt::widgets::ProgressViewerDialog::onResetProgress()
{
  m_dlgProgress->reset();

  m_reset = true;
}

void te::qt::widgets::ProgressViewerDialog::cancel()
{
  std::map<int, te::common::TaskProgress*>::iterator it = m_tasks.begin();

  while(it != m_tasks.end())
  {
    it->second->cancel();
    ++it;
  }
}