/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
\file terralib/qt/widgets/progress/ProgressViewerTaskWidget.cpp

\brief A progress dialog.
*/

// Terralib
#include "../../../core/translator/Translator.h"
#include "../Utils.h"
#include "ProgressViewerTaskWidget.h"


// Qt
#include <QtCore/QCoreApplication>
#include <QApplication>
#include <QMessageBox>
#include <QStyle>

te::qt::widgets::ProgressViewerTaskWidget::ProgressViewerTaskWidget(
    QWidget* parent, bool hideToolBar, std::string message, std::string iconName)
    : AbstractQtProgressViewer(parent),
      m_taskId(-1),
      m_totalSteps(0),
      m_currentStep(0),
      m_propStep(0),
      m_message(message),
      m_task(nullptr)
{
  // build custom widget
  m_mainGridLayout = new QGridLayout(this);
  m_mainGridLayout->setContentsMargins(0, 0, 0, 0);
  m_mainGridLayout->setHorizontalSpacing(2);

  m_button = new QToolButton(this);

  if (iconName.empty())
  {
    m_button->setIcon(m_button->style()->standardIcon(QStyle::SP_TitleBarCloseButton));
  }
  else
  {
    m_button->setIcon(QIcon::fromTheme(te::qt::widgets::Convert2QString(iconName)));
  }

  m_label = new QLabel(this);

  if (hideToolBar)
  {
    m_mainGridLayout->addWidget(m_button, 0, 0);
    m_mainGridLayout->addWidget(m_label, 0, 1);

    m_progressBar = nullptr;
  }
  else
  {
    m_progressBar = new QProgressBar(this);

    m_mainGridLayout->addWidget(m_button, 1, 1);
    m_mainGridLayout->addWidget(m_progressBar, 1, 0);
    m_mainGridLayout->addWidget(m_label, 0, 0, 1, 2);

    // set default range
    m_progressBar->setRange(0, 100);
    m_progressBar->setTextVisible(false);
  }

  m_button->setEnabled(false);

  if (!m_message.empty())
    m_label->setText(te::qt::widgets::Convert2QString(m_message));

  // connect signal cancel
  connect(m_button, SIGNAL(released()), this, SLOT(cancel()));
}

te::qt::widgets::ProgressViewerTaskWidget::~ProgressViewerTaskWidget() =
    default;

void te::qt::widgets::ProgressViewerTaskWidget::addTask(
    te::common::TaskProgress* /*t*/, int /*id*/)
{
}

void te::qt::widgets::ProgressViewerTaskWidget::removeTask(int taskId)
{
  cancelTask(taskId);

  if (m_task && m_task->getId() == taskId)
  {

    m_task = nullptr;
    m_button->setEnabled(false);

    if (m_progressBar != nullptr)
    {
      emit setProgressRange(0, 100);
    }

    emit resetProgress();
  }
}

void te::qt::widgets::ProgressViewerTaskWidget::cancelTask(int taskId)
{
  if (m_task && m_task->getId() == taskId)
  {
    m_totalSteps = 0;
    m_currentStep = 0;
    m_propStep = 0;
  }
}

void te::qt::widgets::ProgressViewerTaskWidget::setTotalValues(int taskId)
{
  if (m_task && m_task->getId() == taskId)
  {
    m_totalSteps = m_task->getTotalSteps();

    if (m_totalSteps == 0)
    {
      emit setProgressRange(0, 0);
    }
    else
    {
      emit setProgressRange(0, 100);
    }
  }
}

void te::qt::widgets::ProgressViewerTaskWidget::updateValue(int taskId)
{
  double currentStep = 0;

  if (m_task && m_task->getId() == taskId)
  {
    currentStep += m_task->getCurrentStep();

    double aux = static_cast<double>(currentStep) / static_cast<double>(m_totalSteps);

    int val = static_cast<int>(100.0 * aux);

    if (val != m_propStep && val >= 0.0)
    {
      m_propStep = val;

      emit setProgressValue(m_propStep);
    }
  }
}

void te::qt::widgets::ProgressViewerTaskWidget::updateMessage(int taskId)
{
  if (m_task && m_task->getId() == taskId)
  {
    m_message = m_task->getMessage();

    emit setProgressMessage(m_message);
  }
}

void te::qt::widgets::ProgressViewerTaskWidget::setTask(te::common::TaskProgress* task)
{
  m_task = task;
  m_taskId = task->getId();

  m_button->setEnabled(true);
}

void te::qt::widgets::ProgressViewerTaskWidget::setMessage(const std::string& message)
{
  m_message = message;

  if (m_label != nullptr)
  {
    m_label->setText(te::qt::widgets::Convert2QString(message));
  }
}

void te::qt::widgets::ProgressViewerTaskWidget::onSetProgressMessage(const std::string& message)
{
  m_label->setText(message.c_str());
}

void te::qt::widgets::ProgressViewerTaskWidget::onSetProgressRange(int min, int max)
{
  m_progressBar->setRange(min, max);
}

void te::qt::widgets::ProgressViewerTaskWidget::onSetProgressValue(int value)
{
  if (m_progressBar->isTextVisible() == false)
  {
    m_progressBar->setTextVisible(true);
  }

  m_progressBar->setValue(value);
}

void te::qt::widgets::ProgressViewerTaskWidget::onResetProgress()
{
  if (m_progressBar)
  {
    m_progressBar->reset();
    m_progressBar->setTextVisible(false);
  }
}

void te::qt::widgets::ProgressViewerTaskWidget::cancel()
{
  int reply = QMessageBox::question(nullptr, tr("Task in Progress"), tr("Are you sure you want to cancel this task?"), QMessageBox::No, QMessageBox::Yes);
  if (reply == QMessageBox::No)
  {
    return;
  }

  if (m_task)
  {
    m_task->cancel();
  }

  emit taskCancelled();
}
