/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/widgets/progress/ProgressWidgetItem.cpp

  \brief Custom widget used to represent a single task.
*/

// TerraLib
#include "ProgressWidgetItem.h"
#include "../Utils.h"

// Qt
#include <QtCore/QCoreApplication>
#include <QSizePolicy>
#include <QStyle>

te::qt::widgets::ProgressWidgetItem::ProgressWidgetItem(QWidget* parent, int taskId, int totalSteps, QString label)
  : QWidget(parent),
  m_taskId(taskId),
  m_totalSteps(0),
  m_currentStep(0),
  m_propStep(0)
{
  // build custom widget
  m_mainGridLayout = new QGridLayout(this);
  m_mainGridLayout->setContentsMargins(1, 1, 1, 1);
  m_mainGridLayout->setHorizontalSpacing(2);
  
  m_frame = new QFrame(this);
  m_frame->setFrameShape(QFrame::StyledPanel);
  m_frame->setFrameShadow(QFrame::Sunken);

  m_mainGridLayout->addWidget(m_frame);

  m_frameGridLayout = new QGridLayout(m_frame);
  m_frameGridLayout->setContentsMargins(1, 1, 1, 1);
  m_frameGridLayout->setVerticalSpacing(1);

  m_label = new QLabel(m_frame);
  m_label->setText(label);

  m_progressBar = new QProgressBar(m_frame);

  m_button = new QToolButton(this);
  m_button->setIcon(m_button->style()->standardIcon(QStyle::SP_TitleBarCloseButton));
  
  m_frameGridLayout->addWidget(m_label, 0, 0, 1, 2);
  m_frameGridLayout->addWidget(m_progressBar, 1, 1);
  m_frameGridLayout->addWidget(m_button, 1, 0);

  // set default range
  if (totalSteps != 0)
    m_progressBar->setRange(0, 100);
  else
    m_progressBar->setRange(0, 0);

  // connect signal cancel
  connect(m_button, SIGNAL(released()), this, SLOT(cancel()));

  connect(this, SIGNAL(setProgressMessage(const std::string&)), this, SLOT(onSetProgressMessage(const std::string&)));
  connect(this, SIGNAL(setProgressRange(int, int)), this, SLOT(onSetProgressRange(int, int)));
  connect(this, SIGNAL(setProgressValue(int)), this, SLOT(onSetProgressValue(int)));
  connect(this, SIGNAL(resetProgress()), this, SLOT(onResetProgress()));

  this->setSizePolicy(QSizePolicy::MinimumExpanding, QSizePolicy::Fixed);
}

te::qt::widgets::ProgressWidgetItem::~ProgressWidgetItem() = default;

void te::qt::widgets::ProgressWidgetItem::setValue(int step)
{
  m_currentStep = step;

  if (m_currentStep != m_propStep && m_currentStep >= 0.0)
  {
    //fix progress bar range
    if (m_progressBar->maximum() != 100)
    { 
      emit setProgressRange(0, 100);
    }
     
    m_propStep = m_currentStep;

    emit setProgressValue(m_propStep);
  }
}

void te::qt::widgets::ProgressWidgetItem::setTotalValues(int values)
{
  m_totalSteps = values;

  if (m_totalSteps == 0)
  {
    emit setProgressRange(0, 0);
  }
  else
  {
    emit setProgressRange(0, 100);
  }
}

void te::qt::widgets::ProgressWidgetItem::setLabel(const std::string& message)
{
  emit setProgressMessage(message);
}

void te::qt::widgets::ProgressWidgetItem::reset()
{
  emit resetProgress();
}

void te::qt::widgets::ProgressWidgetItem::cancel()
{
  emit taskCanceled(m_taskId);
}

void te::qt::widgets::ProgressWidgetItem::onSetProgressMessage(const std::string& message)
{
  m_label->setText(te::qt::widgets::Convert2QString(message));
}

void te::qt::widgets::ProgressWidgetItem::onSetProgressRange(int min, int max)
{
  m_progressBar->setRange(min, max);
}

void te::qt::widgets::ProgressWidgetItem::onSetProgressValue(int value)
{
  if (m_progressBar->minimum() == 0 && m_progressBar->maximum() == 0)
    return;

  m_progressBar->setValue(value);
}

void te::qt::widgets::ProgressWidgetItem::onResetProgress()
{
  m_progressBar->reset();
}
