/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/widgets/se/AbstractPropertyItem.cpp

  \brief A widget used to define the main property items
         that can be used to describe a se object.
*/

#include "AbstractPropertyItem.h"

#include "AbstractPropertyManager.h"



te::qt::widgets::AbstractPropertyItem::AbstractPropertyItem(QtTreePropertyBrowser* propertyBrowser, QColor c)
  : QObject()
  , m_propertyBrowser(propertyBrowser)
  , m_colorParent(c)
  , m_enableListeners(false)
  , m_itemProperty(nullptr)
{
  if (m_propertyBrowser)
  {
    te::qt::widgets::AbstractPropertyManager::getInstance().setFactories(m_propertyBrowser);
  }

  connect(te::qt::widgets::AbstractPropertyManager::getInstance().m_intManager, SIGNAL(valueChanged(QtProperty *, int)), this, SLOT(valueChanged(QtProperty *, int)));
  connect(te::qt::widgets::AbstractPropertyManager::getInstance().m_intSliderManager, SIGNAL(valueChanged(QtProperty *, int)), this, SLOT(valueChanged(QtProperty *, int)));
  connect(te::qt::widgets::AbstractPropertyManager::getInstance().m_doubleManager, SIGNAL(valueChanged(QtProperty *, double)), this, SLOT(valueChanged(QtProperty *, double)));
  connect(te::qt::widgets::AbstractPropertyManager::getInstance().m_boolManager, SIGNAL(valueChanged(QtProperty *, bool)), this, SLOT(valueChanged(QtProperty *, bool)));
  connect(te::qt::widgets::AbstractPropertyManager::getInstance().m_stringManager, SIGNAL(valueChanged(QtProperty *, const QString &)), this, SLOT(valueChanged(QtProperty *, const QString &)));
  connect(te::qt::widgets::AbstractPropertyManager::getInstance().m_strDlgManager, SIGNAL(valueChanged(QtProperty*, const QString&)), this, SLOT(valueChanged(QtProperty*, const QString&)));
  connect(te::qt::widgets::AbstractPropertyManager::getInstance().m_colorManager, SIGNAL(valueChanged(QtProperty *, const QColor &)), this, SLOT(valueChanged(QtProperty *, const QColor &)));
  connect(te::qt::widgets::AbstractPropertyManager::getInstance().m_fontManager, SIGNAL(valueChanged(QtProperty *, const QFont &)), this, SLOT(valueChanged(QtProperty *, const QFont &)));
  connect(te::qt::widgets::AbstractPropertyManager::getInstance().m_sizeManager, SIGNAL(valueChanged(QtProperty *, const QSize &)), this, SLOT(valueChanged(QtProperty *, const QSize &)));
  connect(te::qt::widgets::AbstractPropertyManager::getInstance().m_sizeFManager, SIGNAL(valueChanged(QtProperty *, const QSizeF &)), this, SLOT(valueChanged(QtProperty *, const QSizeF &)));
  connect(te::qt::widgets::AbstractPropertyManager::getInstance().m_enumManager, SIGNAL(valueChanged(QtProperty *, int)), this, SLOT(valueChanged(QtProperty *, int)));
  connect(te::qt::widgets::AbstractPropertyManager::getInstance().m_pointFManager, SIGNAL(valueChanged(QtProperty *, const QPointF &)), this, SLOT(valueChanged(QtProperty *, const QPointF &)));
}

te::qt::widgets::AbstractPropertyItem::~AbstractPropertyItem()
{
  delete m_itemProperty;
}

void te::qt::widgets::AbstractPropertyItem::init(QtProperty* property)
{
  m_enableListeners = true;
  m_itemProperty = property;
}

QtProperty* te::qt::widgets::AbstractPropertyItem::getItemProperty() const
{
  return m_itemProperty;
}

void te::qt::widgets::AbstractPropertyItem::setPropertyColor(QtProperty *property, QColor c)
{
  QList<QtBrowserItem*> list = m_propertyBrowser->items(property);
  for (int i = 0; i < list.size(); ++i)
  {
    m_propertyBrowser->setBackgroundColor(list[i], c);
  }
}

void te::qt::widgets::AbstractPropertyItem::addProperty(QtProperty* qProperty, const QColor& qColor)
{
  if (qProperty == nullptr)
  {
    return;
  }

  QtBrowserItem* item = m_propertyBrowser->addProperty(qProperty);
  if (m_colorParent.isValid())
  {
    m_propertyBrowser->setBackgroundColor(item, m_colorParent);
  }
  else
  {
    m_propertyBrowser->setBackgroundColor(item, qColor);
  }
}

void te::qt::widgets::AbstractPropertyItem::addSubProperty(QtProperty* qPropertyParent, QtProperty* qProperty, const QColor& qColor)
{
  if(qPropertyParent == nullptr)
  {
    return;
  }

  if (qProperty == nullptr)
  {
    return;
  }

  qPropertyParent->addSubProperty(qProperty);
  setPropertyColor(qProperty, qColor);
}

void te::qt::widgets::AbstractPropertyItem::setExpandedStatus(QtBrowserItem *item, bool status, bool applyToChildren)
{
  if (m_propertyBrowser == nullptr)
  {
    return;
  }

  m_propertyBrowser->setExpanded(item, status);

  if (applyToChildren == false)
  {
    return;
  }

  QList<QtBrowserItem*> list = item->children();
  for(int i = 0; i < list.size(); ++i)
  {
    setExpandedStatus(list[i], status, applyToChildren);
  }
}

void te::qt::widgets::AbstractPropertyItem::setExpandedStatus(bool status, bool applyToChildren)
{
  if (m_propertyBrowser == nullptr)
  {
    return;
  }

  if (m_itemProperty == nullptr)
  {
    return;
  }

  QList<QtBrowserItem*> list = m_propertyBrowser->items(m_itemProperty);
  for (int i = 0; i < list.size(); ++i)
  {
    setExpandedStatus(list.at(i), status, applyToChildren);
  }
}

void te::qt::widgets::AbstractPropertyItem::setChildrenEnabled(bool enabled)
{
  if (m_itemProperty == nullptr)
  {
    return;
  }

  QList<QtProperty*> properties = m_itemProperty->subProperties();
  for (int i = 0; i < properties.size(); ++i)
  {
    properties[i]->setEnabled(enabled);
  }
}

QString te::qt::widgets::AbstractPropertyItem::getValueFromEnumProperty(QtProperty* enumPropertyItem)
{
  int index = te::qt::widgets::AbstractPropertyManager::getInstance().m_enumManager->value(enumPropertyItem);
  
  QString qValue;

  if (index >= 0)
  {
    QStringList qEnumNamesList = te::qt::widgets::AbstractPropertyManager::getInstance().m_enumManager->enumNames(enumPropertyItem);
    qValue = qEnumNamesList.at(index);
  }
  return qValue;
}

void te::qt::widgets::AbstractPropertyItem::setValueToEnumProperty(QtProperty* enumPropertyItem, const QString& value)
{
  QStringList qEnumNamesList = te::qt::widgets::AbstractPropertyManager::getInstance().m_enumManager->enumNames(enumPropertyItem);
  int index = qEnumNamesList.indexOf(value);
  if (index >= 0)
  {
    te::qt::widgets::AbstractPropertyManager::getInstance().m_enumManager->setValue(enumPropertyItem, index);
  }
}

void te::qt::widgets::AbstractPropertyItem::valueChanged(QtProperty* p, const QVariant& qVariant)
{
  
}

void te::qt::widgets::AbstractPropertyItem::valueChanged(QtProperty* p, int value)
{
  QVariant qVariant(value);
  valueChanged(p, qVariant);
}

void te::qt::widgets::AbstractPropertyItem::valueChanged(QtProperty* p, double value)
{
  QVariant qVariant(value);
  valueChanged(p, qVariant);
}

void te::qt::widgets::AbstractPropertyItem::valueChanged(QtProperty* p, bool value)
{
  QVariant qVariant(value);
  valueChanged(p, qVariant);
}

void te::qt::widgets::AbstractPropertyItem::valueChanged(QtProperty* p, const QString& value)
{
  QVariant qVariant(value);
  valueChanged(p, qVariant);
}

void te::qt::widgets::AbstractPropertyItem::valueChanged(QtProperty* p, const QColor& value)
{
  QVariant qVariant(value);
  valueChanged(p, qVariant);
}

void te::qt::widgets::AbstractPropertyItem::valueChanged(QtProperty* p, const QFont& value)
{
  QVariant qVariant(value);
  valueChanged(p, qVariant);
}

void te::qt::widgets::AbstractPropertyItem::valueChanged(QtProperty* p, const QSize& value)
{
  QVariant qVariant(value);
  valueChanged(p, qVariant);
}

void te::qt::widgets::AbstractPropertyItem::valueChanged(QtProperty* p, const QSizeF& value)
{
  QVariant qVariant(value);
  valueChanged(p, qVariant);
}

void te::qt::widgets::AbstractPropertyItem::valueChanged(QtProperty* p, const QPointF& value)
{
  QVariant qVariant(value);
  valueChanged(p, qVariant);
}
