/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/widgets/query/QueryBuilderDialog.cpp

  \brief This file defines a class for a Query Builder Dialog.
*/

// TerraLib
#include "QueryBuilderDialog.h"
#include "ui/ui_QueryBuilderDialogForm.h"

#include "../../../common/StringUtils.h"
#include "../../../dataaccess/dataset/ObjectIdSet.h"
#include "../../../dataaccess/utils/Utils.h"
#include "../../../dataaccess/query_h.h"
#include "../../../geometry/GeometryProperty.h"
#include "../../../maptools/DataSetLayer.h"
#include "../../../maptools/QueryLayer.h"
#include "../../../se/Utils.h"
#include "../utils/ScopedCursor.h"

// Boost
#include <boost/uuid/random_generator.hpp>
#include <boost/uuid/uuid_io.hpp>

// Qt
#include <QGridLayout>
#include <QMessageBox>

Q_DECLARE_METATYPE(te::map::AbstractLayerPtr)

te::qt::widgets::QueryBuilderDialog::QueryBuilderDialog(QWidget* parent, Qt::WindowFlags f)
  : QDialog(parent, f),
    m_ui(new Ui::QueryBuilderDialogForm)
{
  m_ui->setupUi(this);

// Signals slots
  connect(m_ui->m_inputLayerComboBox, SIGNAL(activated(QString)), this, SLOT(onInputLayerActivated(QString)));
  connect(m_ui->m_applyPushButton, SIGNAL(clicked()), this, SLOT(onApplyPushButtonClicked()));
  connect(m_ui->m_createLayerPushButton, SIGNAL(clicked()), this, SLOT(onCreateLayerPushButtonClicked()));


  //create query builder widget
  QGridLayout* layout = new QGridLayout(m_ui->m_widget);
  m_queryWidget = new te::qt::widgets::QueryBuilderWidget(m_ui->m_widget);
  layout->addWidget(m_queryWidget);
  layout->setContentsMargins(0, 0, 0, 0);

  connect(m_queryWidget, SIGNAL(strWhereClause(std::string)), this, SLOT(onStrWhereClause(std::string)));

  //
  m_ui->m_helpPushButton->setPageReference("widgets/query/query_dialog.html");
}

te::qt::widgets::QueryBuilderDialog::~QueryBuilderDialog() = default;

void te::qt::widgets::QueryBuilderDialog::setLayerList(std::list<te::map::AbstractLayerPtr>& layerList)
{
  m_ui->m_inputLayerComboBox->clear();

  std::list<te::map::AbstractLayerPtr>::iterator it = layerList.begin();

  while (it != layerList.end())
  {
    te::map::AbstractLayerPtr l = *it;

    std::unique_ptr<te::da::DataSetType> dsType = l->getSchema();

    if (dsType->hasGeom())
      m_ui->m_inputLayerComboBox->addItem(l->getTitle().c_str(), QVariant::fromValue(l));

    ++it;
  }

  if (m_ui->m_inputLayerComboBox->count() > 0)
  {
    QString s = m_ui->m_inputLayerComboBox->currentText();

    onInputLayerActivated(s);
  }
}

void te::qt::widgets::QueryBuilderDialog::setCurrentLayer(te::map::AbstractLayerPtr layer)
{
  for (int i = 0; i < m_ui->m_inputLayerComboBox->count(); ++i)
  {
    QVariant varLayer = m_ui->m_inputLayerComboBox->itemData(i, Qt::UserRole);
    te::map::AbstractLayerPtr l = varLayer.value<te::map::AbstractLayerPtr>();

    if (layer == l)
    {
      m_ui->m_inputLayerComboBox->setCurrentIndex(i);
      break;
    }
  }

  if (m_ui->m_inputLayerComboBox->count() > 0)
  {
    QString s = m_ui->m_inputLayerComboBox->currentText();

    onInputLayerActivated(s);
  }
}

void te::qt::widgets::QueryBuilderDialog::startWhereClause()
{
  m_queryWidget->startWhereClause();
}

void te::qt::widgets::QueryBuilderDialog::setWhereClause(te::da::Where* where)
{
  m_queryWidget->setWhereClause(where);
}

te::da::Where* te::qt::widgets::QueryBuilderDialog::getWhereClause()
{
  return m_queryWidget->getWhereClause();
}

std::string te::qt::widgets::QueryBuilderDialog::getWhereString()
{
  return m_queryWidget->getWhereString();
}

void te::qt::widgets::QueryBuilderDialog::onInputLayerActivated(QString /*value*/)
{
  // Gets the input layer
  int idxLayer = m_ui->m_inputLayerComboBox->currentIndex();

  QVariant varLayer = m_ui->m_inputLayerComboBox->itemData(idxLayer, Qt::UserRole);

  te::map::AbstractLayerPtr layer = varLayer.value<te::map::AbstractLayerPtr>();

  //set output query layer
  m_ui->m_queryLayerNameLineEdit->setText(QString::fromUtf8(layer->getTitle().c_str()));

  // Sets the DataSource
  std::string dataSourceName = layer->getDataSourceId();
  std::string dataSetName = layer->getDataSetName();

  std::unique_ptr<te::da::DataSetType> dsType = layer->getSchema();

  te::da::DataSourcePtr dsPtr = te::da::GetDataSource(dataSourceName);

  if (!dsPtr->isOpened())
    dsPtr->open();

  te::map::QueryLayer* queryLayer = dynamic_cast<te::map::QueryLayer*>(layer.get());

  if (queryLayer != nullptr)
  {
    m_queryWidget->setInputParameters(dsPtr, "", std::move(dsType));

    te::da::Select* sel = queryLayer->getQuery();

    if (sel->getWhere())
    {
      m_queryWidget->setWhereClause(sel->getWhere());
    }
    else
    {
      m_queryWidget->startWhereClause();
    }

    //change button label
    m_ui->m_createLayerPushButton->setText(tr("Update Layer"));
    m_ui->m_queryLayerNameLineEdit->setEnabled(false);
  }
  else
  {
    m_queryWidget->setInputParameters(dsPtr, dataSetName, std::move(dsType));

    m_queryWidget->startWhereClause();

    //change button label
    m_ui->m_createLayerPushButton->setText(tr("Create Layer"));
    m_ui->m_queryLayerNameLineEdit->setEnabled(true);
  }
}

void te::qt::widgets::QueryBuilderDialog::onCreateLayerPushButtonClicked()
{
  // Gets the defined restriction
  te::da::Where* wh = getWhereClause();

  if (wh == nullptr || wh->getExp() == nullptr)
  {
    QMessageBox::information(this, tr("Query"), tr("Invalid restriction."));
    return;
  }

  if (m_ui->m_queryLayerNameLineEdit->text().isEmpty())
  {
    QMessageBox::information(this, tr("Query"), tr("Define the output layer name."));
    return;
  }

  std::string queryLayerName = m_ui->m_queryLayerNameLineEdit->text().toUtf8().data();

  // Gets the selected layer
  int index = m_ui->m_inputLayerComboBox->currentIndex();

  QVariant varLayer = m_ui->m_inputLayerComboBox->itemData(index, Qt::UserRole);

  te::map::AbstractLayerPtr layer = varLayer.value<te::map::AbstractLayerPtr>();

  te::map::QueryLayer* queryLayer = dynamic_cast<te::map::QueryLayer*>(layer.get());

  if (queryLayer != nullptr)
  {
    te::da::Select* sel = queryLayer->getQuery();

    sel->setWhere(wh);

    queryLayer->computeExtent();

    QMessageBox::information(this, tr("Query"), tr("Layer updated."));
  }
  else
  {

    std::unique_ptr<te::da::DataSetType> schema = layer->getSchema();

    std::vector<std::string> tokens;
    te::common::Tokenize(schema->getName(), tokens, ".");

    std::string alias;

    if (tokens.size() > 1)
      alias = tokens[1];
    else
      alias = schema->getName();

    assert(layer.get());

    static boost::uuids::basic_random_generator<boost::mt19937> gen;
    boost::uuids::uuid u = gen();
    std::string id = boost::uuids::to_string(u);

    te::map::AbstractLayerPtr aLayer(new te::map::QueryLayer(id, queryLayerName));
    aLayer->setDataSourceId(layer->getDataSourceId());

    te::map::QueryLayer* qLayer = dynamic_cast<te::map::QueryLayer*>(aLayer.get());
    qLayer->setDataSetName(layer->getDataSetName());
    qLayer->setRendererType("QUERY_LAYER_RENDERER");

    //fields
    te::da::Fields* fields = new te::da::Fields;

    std::vector<te::dt::Property*> props = schema->getProperties();
    for (auto p : props)
    {
      fields->push_back(new te::da::Field(alias + "." + p->getName()));
    }

    //from
    te::da::From* from = new te::da::From;

    te::da::FromItem* fromItem = new te::da::DataSetName(schema->getName(), alias);
    from->push_back(fromItem);

    //build the select object
    te::da::Select* s = new te::da::Select(fields, from, wh);

    qLayer->setQuery(s);

    te::gm::GeometryProperty* gp = te::da::GetFirstGeomProperty(schema.get());
    if (gp)
    {
      qLayer->computeExtent();
      qLayer->setSRID(gp->getSRID());

      // style
      qLayer->setStyle(te::se::CreateFeatureTypeStyle(gp->getGeometryType()));
    }

    emit createLayer(aLayer);

    QMessageBox::information(this, tr("Query"), tr("Layer created."));
  }
}

void te::qt::widgets::QueryBuilderDialog::onApplyPushButtonClicked()
{
  // Gets the selected layer
  int index = m_ui->m_inputLayerComboBox->currentIndex();

  QVariant varLayer = m_ui->m_inputLayerComboBox->itemData(index, Qt::UserRole);

  te::map::AbstractLayerPtr layer = varLayer.value<te::map::AbstractLayerPtr>();

  assert(layer.get());

  //check if the action is clear layer selection
  if (m_ui->m_noneSelRadioButton->isChecked())
  {
    layer->clearSelected();

    emit layerSelectedObjectsChanged(layer);

    return;
  }

  // Gets the defined restriction
  te::da::Where* wh = getWhereClause();

  if (wh == nullptr || wh->getExp() == nullptr)
  {
    QMessageBox::information(this, tr("Query"), tr("Add a restriction expression first."));
    return;
  }

  // Let's execute the query
  try
  {
    te::qt::widgets::ScopedCursor scopedCursor(Qt::WaitCursor);

    // The filter expression
    te::da::Expression* e = wh->getExp()->clone();

    // Gets the layer schema
    std::unique_ptr<const te::map::LayerSchema> schema(layer->getSchema());

    // Gets the dataset
    std::unique_ptr<te::da::DataSet> dataset = layer->getData(e);
    assert(dataset.get());

    if (m_ui->m_newSelRadioButton->isChecked())
    {
      // Generates the oids
      dataset->moveBeforeFirst();
      te::da::ObjectIdSet* oids = te::da::GenerateOIDSet(dataset.get(), schema.get());

      oids->setExpression(wh->getExp()->clone(), false);

      layer->clearSelected();
      layer->select(oids);

      emit layerSelectedObjectsChanged(layer);
    }
    else if (m_ui->m_addSelRadioButton->isChecked())
    {
      // Generates the oids
      dataset->moveBeforeFirst();
      te::da::ObjectIdSet* oids = te::da::GenerateOIDSet(dataset.get(), schema.get());

      oids->setExpression(wh->getExp()->clone(), false);

      layer->select(oids);

      emit layerSelectedObjectsChanged(layer);
    }
  }
  catch (std::exception& e)
  {
    QMessageBox::information(this, tr("Query"), e.what());
  }
}

void te::qt::widgets::QueryBuilderDialog::onStrWhereClause(std::string str)
{
  m_ui->m_sqlResultTextEdit->clear();

  m_ui->m_sqlResultTextEdit->append(str.c_str());
}

