/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/widgets/query/QueryBuilderWidget.cpp

  \brief This file has the QueryBuilderWidget class.
*/

// TerraLib
#include "QueryBuilderWidget.h"
#include "ui/ui_QueryBuilderWidgetForm.h"

#include "QueryBuilderTreeWidget.h"
#include "QueryConditionWidget.h"
#include "QueryConnectorWidget.h"

#include "../../../dataaccess/datasource/DataSourceCapabilities.h"
#include "../../../dataaccess/query_h.h"

// Qt
#include <QGridLayout>
#include <QMessageBox>
#include <QTreeWidgetItem>
#include <QMessageBox>
#include <QHeaderView>

te::qt::widgets::QueryBuilderWidget::QueryBuilderWidget(QWidget* parent, Qt::WindowFlags f)
  : QWidget(parent, f),
    m_ui(new Ui::QueryBuilderWidgetForm),
    m_starterWidget(nullptr)
{
  m_ui->setupUi(this);
    
  //create tree
  QGridLayout* layout = new QGridLayout(m_ui->m_widget);
  m_treeWidget = new te::qt::widgets::QueryBuilderTreeWidget(m_ui->m_widget);
  layout->addWidget(m_treeWidget);
  layout->setContentsMargins(0, 0, 0, 0);

  QHeaderView* header = m_treeWidget->header();
  header->setVisible(false);
  m_treeWidget->setHeader(header);
  m_treeWidget->setSelectionMode(QAbstractItemView::NoSelection);
}

te::qt::widgets::QueryBuilderWidget::~QueryBuilderWidget()
{
  if (m_widgetMap.empty() == false)
  {
    std::map<QTreeWidgetItem*, QWidget*>::iterator it = m_widgetMap.begin();

    while (it != m_widgetMap.end())
    {
      //delete it->second;

      ++it;
    }
  }

  m_widgetMap.clear();
}

void te::qt::widgets::QueryBuilderWidget::setInputParameters(te::da::DataSourcePtr dataSource, std::string dataSetName, std::unique_ptr<te::da::DataSetType> dsType)
{
  m_dataSource = dataSource;
  m_dataSetName = dataSetName;
  m_dsType.reset(dsType.release());

  m_treeWidget->clear();

  m_widgetMap.clear();
}

void te::qt::widgets::QueryBuilderWidget::startWhereClause()
{
  createCondition(true, nullptr, true);

  getWhereString();
}

void te::qt::widgets::QueryBuilderWidget::setWhereClause(te::da::Where* where)
{
  m_treeWidget->clear();

  m_widgetMap.clear();

  m_starterWidget = nullptr;

  std::unique_ptr<te::da::Expression> exp(where->getExp()->clone());

  buildTree(exp.get(), nullptr, true);

  getWhereString();
}

te::da::Where* te::qt::widgets::QueryBuilderWidget::getWhereClause()
{
  te::da::Where* w = nullptr;

  if (m_starterWidget)
  {
    te::qt::widgets::QueryConnectorWidget* connector = dynamic_cast<te::qt::widgets::QueryConnectorWidget*>(m_starterWidget);

    if (connector)
    {
      te::da::Expression* exp = connector->getExpression();

      if (exp)
      {
        w = new te::da::Where(exp);
      }
    }
    else
    {
      te::qt::widgets::QueryConditionWidget* condition = dynamic_cast<te::qt::widgets::QueryConditionWidget*>(m_starterWidget);

      if (condition)
      {
        te::da::Expression* exp = condition->getExpression();

        if (exp)
        {
          w = new te::da::Where(exp);
        }
      }
    }
  }

  return w;
}

QWidget* te::qt::widgets::QueryBuilderWidget::createCondition(bool startCondition, QWidget* parent, bool right)
{
  te::da::DataSetType* dsType = dynamic_cast<te::da::DataSetType*>(m_dsType->clone());

  te::qt::widgets::QueryConditionWidget* condition = new te::qt::widgets::QueryConditionWidget(m_dataSource, m_dataSetName, dsType);

  connect(condition, SIGNAL(conditionChanged()), this, SLOT(onChanged()));

  QTreeWidgetItem* item = nullptr;

  if (parent)
  {
    te::qt::widgets::QueryConnectorWidget* connector = dynamic_cast<te::qt::widgets::QueryConnectorWidget*>(parent);

    if (connector)
    {
      condition->setConnectorParent(connector);

      if (right)
      {
        connector->setConditionRight(condition);

        if (connector->getConditionLeft())
        {
          item = new QTreeWidgetItem(connector->m_item, connector->getConditionLeft()->m_item);
        }
        else if (connector->getConnectorLeft())
        {
          item = new QTreeWidgetItem(connector->m_item, connector->getConnectorLeft()->m_item);
        }
        else
        {
          throw;
        }
      }
      else
      {
        connector->setConditionLeft(condition);

        item = new QTreeWidgetItem(connector->m_item);
      }
    }
  }
  else
  {
    item = new QTreeWidgetItem(m_treeWidget);
  }

  m_widgetMap.insert(std::map<QTreeWidgetItem*, QWidget*>::value_type(item, condition));

  m_treeWidget->setItemWidget(item, 0, condition);

  condition->m_item = item;

  connect(condition, SIGNAL(addConnectorClicked(QWidget*, QWidget*)), this, SLOT(onAddConnectorClicked(QWidget*, QWidget*)));
  connect(condition, SIGNAL(removeConnectorClicked(QWidget*, QWidget*)), this, SLOT(onRemoveConnectorClicked(QWidget*, QWidget*)));

  if (startCondition)
  {
    m_starterWidget = condition;
  }

  return condition;
}

QWidget* te::qt::widgets::QueryBuilderWidget::createConnector(bool startCondition, QWidget* owner, QWidget* parent, bool autoCreateCondition)
{
  te::qt::widgets::QueryConnectorWidget* connector = new te::qt::widgets::QueryConnectorWidget(m_dataSource, m_dataSetName);

  connect(connector, SIGNAL(connectorChanged()), this, SLOT(onChanged()));

  if (startCondition)
  {
    m_starterWidget = connector;
  }

  te::qt::widgets::QueryConditionWidget* condition = dynamic_cast<te::qt::widgets::QueryConditionWidget*>(owner);

  QTreeWidgetItem* item = nullptr;

  if (parent)
  {
    te::qt::widgets::QueryConnectorWidget* connectorParent = dynamic_cast<te::qt::widgets::QueryConnectorWidget*>(parent);

    if (connectorParent)
    {
      connector->setConnectorParent(connectorParent);

      if (condition)
      {
        if (condition == connectorParent->getConditionLeft())
        {
          connectorParent->setConnectorLeft(connector);

          item = new QTreeWidgetItem(connectorParent->m_item, connectorParent->m_item);
        }
        else if (condition == connectorParent->getConditionRight())
        {
          connectorParent->setConnectorRight(connector);

          item = new QTreeWidgetItem(connectorParent->m_item);
        }
      }
      else
      {
        item = new QTreeWidgetItem(connectorParent->m_item);

        if (connectorParent->getConnectorLeft() == nullptr && connectorParent->getConditionLeft() == nullptr)
        {
          connectorParent->setConnectorLeft(connector);
        }
        else
        {
          if (connectorParent->getConnectorRight() == nullptr && connectorParent->getConditionRight() == nullptr)
          {
            connectorParent->setConnectorRight(connector);
          }
        }
      }
    }
  }
  else
  {
    item = new QTreeWidgetItem(m_treeWidget);
  }

  m_widgetMap.insert(std::map<QTreeWidgetItem*, QWidget*>::value_type(item, connector));

  m_treeWidget->setItemWidget(item, 0, connector);

  connector->m_item = item;

  if (condition)
  {
    QTreeWidgetItem* conditionParentItem = condition->m_item->parent();

    if (conditionParentItem == nullptr)
    {
      int index = m_treeWidget->indexOfTopLevelItem(condition->m_item);

      QTreeWidgetItem* conditionItem = m_treeWidget->takeTopLevelItem(index);

      connector->m_item->insertChild(0, conditionItem);

      connector->setConditionLeft(condition);

      condition->setConnectorParent(connector);
    }
    else
    {
      int index = conditionParentItem->indexOfChild(condition->m_item);

      QTreeWidgetItem* conditionItem = conditionParentItem->takeChild(index);

      connector->m_item->insertChild(0, conditionItem);

      connector->setConditionLeft(condition);

      condition->setConnectorParent(connector);
    }
  }

  if (autoCreateCondition)
  {
    createCondition(false, connector, true);
  }

  connector->m_item->setExpanded(true);

  return connector;
}

void te::qt::widgets::QueryBuilderWidget::removeCondition(te::qt::widgets::QueryConditionWidget* condition)
{
  m_treeWidget->removeItemWidget(condition->m_item, 0);

  delete condition->m_item;

  //delete condition;

}

void te::qt::widgets::QueryBuilderWidget::removeConnector(te::qt::widgets::QueryConnectorWidget* connector)
{
  m_treeWidget->removeItemWidget(connector->m_item, 0);

  delete connector->m_item;

  //delete connector;
}

void te::qt::widgets::QueryBuilderWidget::rebuildTree(QTreeWidgetItem* item)
{
  item->setExpanded(false);

  for (int i = 0; i < item->childCount(); ++i)
  {
    QTreeWidgetItem* childItem = item->child(i);

    m_treeWidget->setItemWidget(childItem, 0, m_widgetMap[childItem]);

    if (childItem->childCount() > 0)
    {
      rebuildTree(childItem);
    }
  }

  item->setExpanded(true);
}

void te::qt::widgets::QueryBuilderWidget::buildTree(te::da::Expression* exp, QWidget* parent, bool right)
{
  // Check Expression Type
  te::da::Function* func = dynamic_cast<te::da::Function*>(exp);

  if (func)
  {
    if (functionsIsACondition(func))
    {
      bool starter = (parent == nullptr) ? true : false;

      QWidget* condition = createCondition(starter, parent, right);

      te::qt::widgets::QueryConditionWidget* condWidget = dynamic_cast<te::qt::widgets::QueryConditionWidget*>(condition);

      if (condWidget)
      {
        condWidget->setCurrentCondition(func);
      }

      if (starter)
      {
        m_starterWidget = condition;
      }
    }
    else if (functionsIsAConnector(func))
    {
      if (func->getNumArgs() == 2)
      {
        bool starter = (parent == nullptr) ? true : false;

        //create connector
        QWidget* connector = createConnector(starter, nullptr, parent, false);

        te::qt::widgets::QueryConnectorWidget* connWidget = dynamic_cast<te::qt::widgets::QueryConnectorWidget*>(connector);

        if (connWidget)
        {
          connWidget->setCurrentConnector(func);
        }

        if (starter)
        {
          m_starterWidget = connector;
        }

        //call to arg1
        buildTree(func->getArg(0), connector, false);

        //call to arg2
        buildTree(func->getArg(1), connector, true);
      }
      else
      {
        QMessageBox::warning(this, tr("Query Builder"), tr("Error: Expression Type Undefined!"));
      }
    }
    else
    {
      QMessageBox::warning(this, tr("Query Builder"), tr("Error: Expression Type Undefined!"));
    }
  }
  else
  {
    QMessageBox::warning(this, tr("Query Builder"), tr("Error: Expression Type Undefined!"));
  }
}

bool te::qt::widgets::QueryBuilderWidget::functionsIsAConnector(te::da::Function* func)
{
  te::da::DataSourceCapabilities dsCap = m_dataSource->getCapabilities();
  te::da::QueryCapabilities queryCap = dsCap.getQueryCapabilities();

  std::set<std::string>::iterator it = queryCap.getLogicalOperators().begin();

  while (it != queryCap.getLogicalOperators().end())
  {
    if (func->getName() == *it)
    {
      return true;
    }

    ++it;
  }

  return false;
}

bool te::qt::widgets::QueryBuilderWidget::functionsIsACondition(te::da::Function* func)
{
  te::da::DataSourceCapabilities dsCap = m_dataSource->getCapabilities();
  te::da::QueryCapabilities queryCap = dsCap.getQueryCapabilities();

  std::set<std::string>::iterator it;

  it = queryCap.getArithmeticOperators().begin();
  while (it != queryCap.getArithmeticOperators().end())
  {
    if (func->getName() == *it)
    {
      return true;
    }

    ++it;
  }

  it = queryCap.getComparsionOperators().begin();
  while (it != queryCap.getComparsionOperators().end())
  {
    if (func->getName() == *it)
    {
      return true;
    }

    ++it;
  }

  return false;
}

std::string te::qt::widgets::QueryBuilderWidget::getWhereString()
{
  std::string sql;

  try
  {
    std::unique_ptr<te::da::Where> where(getWhereClause());

    if (where.get())
    {
      te::da::SQLVisitor visitor(*m_dataSource->getDialect(), sql);

      where->getExp()->accept(visitor);
    }
  }
  catch (te::common::Exception& ex)
  {
    return ex.what();
  }

  emit strWhereClause(sql);

  return sql;
}

void te::qt::widgets::QueryBuilderWidget::onAddConnectorClicked(QWidget* owner, QWidget* parent)
{
  bool starterCondition = false;

  if (owner == m_starterWidget)
  {
    starterCondition = true;
  }

  createConnector(starterCondition, owner, parent, true);

  getWhereString();
}

void te::qt::widgets::QueryBuilderWidget::onRemoveConnectorClicked(QWidget* owner, QWidget* parent)
{
  if (owner == m_starterWidget)
  {
    QMessageBox::warning(this, tr("Query Builder"), tr("Can not remove this element."));
    return;
  }

  te::qt::widgets::QueryConditionWidget* condition = dynamic_cast<te::qt::widgets::QueryConditionWidget*>(owner);
  te::qt::widgets::QueryConnectorWidget* connectorParent = dynamic_cast<te::qt::widgets::QueryConnectorWidget*>(parent);

  if (condition && connectorParent)
  {
    if (condition == connectorParent->getConditionRight())
    {
      removeCondition(condition);

      if (connectorParent->getConnectorParent() == nullptr)
      {
        te::qt::widgets::QueryConditionWidget* conditionLeft = connectorParent->getConditionLeft();

        if (conditionLeft)
        {
          int index = connectorParent->m_item->indexOfChild(conditionLeft->m_item);

          QTreeWidgetItem* conditionLeftItem = connectorParent->m_item->takeChild(index);

          removeConnector(connectorParent);

          m_treeWidget->insertTopLevelItem(0, conditionLeftItem);

          m_starterWidget = conditionLeft;

          conditionLeft->setConnectorParent(nullptr);
        }
        else
        {
          te::qt::widgets::QueryConnectorWidget* connectorLeft = connectorParent->getConnectorLeft();

          if (connectorLeft)
          {
            int index = connectorParent->m_item->indexOfChild(connectorLeft->m_item);

            QTreeWidgetItem* connectorLeftItem = connectorParent->m_item->takeChild(index);

            removeConnector(connectorParent);

            m_treeWidget->insertTopLevelItem(m_treeWidget->topLevelItemCount(), connectorLeftItem);

            m_starterWidget = connectorLeft;

            connectorLeft->setConnectorParent(nullptr);

            rebuildTree(connectorLeftItem);
          }
        }
      }
      else
      {
        te::qt::widgets::QueryConnectorWidget* connectorParentParent = connectorParent->getConnectorParent();

        te::qt::widgets::QueryConditionWidget* conditionLeft = connectorParent->getConditionLeft();

        if (conditionLeft)
        {
          //left is a condition
          int index = connectorParent->m_item->indexOfChild(conditionLeft->m_item);

          QTreeWidgetItem* conditionLeftItem = connectorParent->m_item->takeChild(index);

          if (connectorParentParent->getConnectorLeft() == connectorParent)
          {
            removeConnector(connectorParent);

            connectorParentParent->m_item->insertChild(0, conditionLeftItem);

            connectorParentParent->setConditionLeft(conditionLeft);
          }
          else if (connectorParentParent->getConnectorRight() == connectorParent)
          {
            removeConnector(connectorParent);

            connectorParentParent->m_item->insertChild(1, conditionLeftItem);

            connectorParentParent->setConditionRight(conditionLeft);
          }

          conditionLeft->setConnectorParent(connectorParentParent);

          rebuildTree(conditionLeftItem);
        }
        else
        {
          //left is another connector
          te::qt::widgets::QueryConnectorWidget* connectorLeft = connectorParent->getConnectorLeft();

          if (connectorLeft)
          {
            int index = connectorParent->m_item->indexOfChild(connectorLeft->m_item);

            QTreeWidgetItem* connectorLeftItem = connectorParent->m_item->takeChild(index);

            if (connectorParentParent->getConnectorLeft() == connectorParent)
            {
              removeConnector(connectorParent);

              connectorParentParent->m_item->insertChild(0, connectorLeftItem);

              connectorParentParent->setConnectorLeft(connectorLeft);
            }
            else if (connectorParentParent->getConnectorRight() == connectorParent)
            {
              removeConnector(connectorParent);

              connectorParentParent->m_item->insertChild(1, connectorLeftItem);

              connectorParentParent->setConnectorRight(connectorLeft);
            }

            connectorLeft->setConnectorParent(connectorParentParent);

            rebuildTree(connectorLeftItem);
          }
        }
      }
    }
    else if (condition == connectorParent->getConditionLeft())
    {
      removeCondition(condition);

      if (connectorParent->getConnectorParent() == nullptr)
      {
        te::qt::widgets::QueryConditionWidget* conditionRight = connectorParent->getConditionRight();

        if (conditionRight)
        {
          int index = connectorParent->m_item->indexOfChild(conditionRight->m_item);

          QTreeWidgetItem* conditionRightItem = connectorParent->m_item->takeChild(index);

          removeConnector(connectorParent);

          m_treeWidget->insertTopLevelItem(0, conditionRightItem);

          m_starterWidget = conditionRight;

          conditionRight->setConnectorParent(nullptr);
        }
        else
        {
          te::qt::widgets::QueryConnectorWidget* connectorRight = connectorParent->getConnectorRight();

          if (connectorRight)
          {
            int index = connectorParent->m_item->indexOfChild(connectorRight->m_item);

            QTreeWidgetItem* connectorRightItem = connectorParent->m_item->takeChild(index);

            removeConnector(connectorParent);

            m_treeWidget->insertTopLevelItem(0, connectorRightItem);

            m_starterWidget = connectorRight;

            connectorRight->setConnectorParent(nullptr);

            rebuildTree(connectorRightItem);
          }
        }
      }
      else
      {
        te::qt::widgets::QueryConnectorWidget* connectorParentParent = connectorParent->getConnectorParent();

        te::qt::widgets::QueryConditionWidget* conditionRight = connectorParent->getConditionRight();

        if (conditionRight)
        {
          //left is a condition
          int index = connectorParent->m_item->indexOfChild(conditionRight->m_item);

          QTreeWidgetItem* conditionRightItem = connectorParent->m_item->takeChild(index);

          if (connectorParentParent->getConnectorLeft() == connectorParent)
          {
            removeConnector(connectorParent);

            connectorParentParent->m_item->insertChild(0, conditionRightItem);

            connectorParentParent->setConditionLeft(conditionRight);
          }
          else if (connectorParentParent->getConnectorRight() == connectorParent)
          {
            removeConnector(connectorParent);

            connectorParentParent->m_item->insertChild(1, conditionRightItem);

            connectorParentParent->setConditionRight(conditionRight);
          }

          conditionRight->setConnectorParent(connectorParentParent);

          rebuildTree(conditionRightItem);
        }
        else
        {
          //left is another connector
          te::qt::widgets::QueryConnectorWidget* connectorRight = connectorParent->getConnectorRight();

          if (connectorRight)
          {
            int index = connectorParent->m_item->indexOfChild(connectorRight->m_item);

            QTreeWidgetItem* connectorRightItem = connectorParent->m_item->takeChild(index);

            if (connectorParentParent->getConnectorLeft() == connectorParent)
            {
              removeConnector(connectorParent);

              connectorParentParent->m_item->insertChild(0, connectorRightItem);

              connectorParentParent->setConnectorLeft(connectorRight);
            }
            else if (connectorParentParent->getConnectorRight() == connectorParent)
            {
              removeConnector(connectorParent);

              connectorParentParent->m_item->insertChild(1, connectorRightItem);

              connectorParentParent->setConnectorRight(connectorRight);
            }

            connectorRight->setConnectorParent(connectorParentParent);

            rebuildTree(connectorRightItem);
          }
        }
      }
    }
  }

  getWhereString();
}

void te::qt::widgets::QueryBuilderWidget::onChanged()
{
  getWhereString();
}
