/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/widgets/query/QueryConditionWidget.cpp

  \brief This file has the Query Condition Widget class.
*/

// TerraLib
#include "QueryConditionWidget.h"
#include "QueryConnectorWidget.h"
#include "ui/ui_QueryConditionWidgetForm.h"

#include "../Utils.h"
#include "../../../dataaccess/query/Function.h"
#include "../../../dataaccess/query/Like.h"
#include "../../../dataaccess/query/LiteralDateTime.h"
#include "../../../dataaccess/dataset/DataSetType.h"
#include "../../../dataaccess/datasource/DataSourceCapabilities.h"
#include "../../../dataaccess/query_h.h"
#include "../../../datatype/Date.h"
#include "../../../datatype/TimeInstant.h"
#include "../../../datatype/DateTimeUtils.h"


// Qt
#include <QMessageBox>
#include <QSortFilterProxyModel>
#include <QStringListModel>
#include <QDate>

#include <boost/date_time.hpp>

Q_DECLARE_METATYPE(te::dt::Property*);

te::qt::widgets::QueryConditionWidget::QueryConditionWidget(te::da::DataSourcePtr dataSource, std::string dataSetName, te::da::DataSetType* dsType, QWidget* parent, Qt::WindowFlags f)
  : QWidget(parent, f),
    m_ui(new Ui::QueryConditionWidgetForm),
    m_parent(nullptr),
    m_dataSource(dataSource),
    m_dataSetName(dataSetName),
    m_item(nullptr)
{
  m_ui->setupUi(this);

  m_dsType.reset(dsType);

  m_ui->m_addToolButton->setIcon(QIcon::fromTheme("list-add"));
  m_ui->m_removeToolButton->setIcon(QIcon::fromTheme("list-remove"));
  m_ui->m_loadToolButton->setIcon(QIcon::fromTheme("zoom"));
  m_ui->m_loadToolButton->setCheckable(true);

  //connects
  connect(m_ui->m_addToolButton, SIGNAL(clicked()), this, SLOT(onAddToolButtonClicked()));
  connect(m_ui->m_removeToolButton, SIGNAL(clicked()), this, SLOT(onRemoveToolButtonClicked()));
  connect(m_ui->m_loadToolButton, SIGNAL(clicked()), this, SLOT(onLoadAttrsToolButtonClicked()));

  connect(m_ui->m_functionComboBox, SIGNAL(activated(int)), this, SIGNAL(conditionChanged()));
  connect(m_ui->m_propertyComboBox, SIGNAL(activated(int)), this, SIGNAL(conditionChanged()));
  connect(m_ui->m_valueComboBox, SIGNAL(editTextChanged(const QString&)), this, SIGNAL(conditionChanged()));
  
  //fill interface
  setOperatorsList();

  setAttributeList();

  if (dataSetName.empty())
  {
    m_ui->m_loadToolButton->setEnabled(false);
  }

  const te::da::DataSourceCapabilities& capabilities = dataSource->getCapabilities();
  const te::da::QueryCapabilities& queryCapabilities = capabilities.getQueryCapabilities();
  if (queryCapabilities.supportsSelect() == false)
  {
    m_ui->m_loadToolButton->setEnabled(false);
  }
}

te::qt::widgets::QueryConditionWidget::~QueryConditionWidget() = default;

void te::qt::widgets::QueryConditionWidget::setCurrentCondition(te::da::Function* func)
{
  if (func)
  {
    if (func->getNumArgs() == 2)
    {
      //get function
      std::string connector = func->getName();

      int idx = m_ui->m_functionComboBox->findText(connector.c_str());

      if (idx != (-1))
      {
        m_ui->m_functionComboBox->setCurrentIndex(idx);
      }
      else
      {
        m_ui->m_functionComboBox->addItem(connector.c_str());
        m_ui->m_functionComboBox->setCurrentIndex(m_ui->m_functionComboBox->findText(connector.c_str()));
      }

      //get property
      te::da::PropertyName* pName = dynamic_cast<te::da::PropertyName*>(func->getArg(0));
      
      //get value
      te::da::Literal* lit = dynamic_cast<te::da::Literal*>(func->getArg(1));

      if (pName && lit)
      {
        std::string propertyName = getAliasName(pName->getName());
        idx = m_ui->m_propertyComboBox->findText(QString::fromUtf8(propertyName.c_str()));

        if (idx != (-1))
        {
          m_ui->m_propertyComboBox->setCurrentIndex(idx);
        }
        else
        {
          m_ui->m_propertyComboBox->addItem(QString::fromUtf8(propertyName.c_str()));
          m_ui->m_propertyComboBox->setCurrentIndex(m_ui->m_propertyComboBox->findText(QString::fromUtf8(propertyName.c_str())));
        }


        std::string valueString = lit->getValue()->toString();

        if (connector == "like")
        {
          valueString.erase(std::remove(valueString.begin(), valueString.end(), '%'), valueString.end());
        }

        idx = m_ui->m_valueComboBox->findText(QString::fromUtf8(valueString.c_str()));

        if (idx != (-1))
        {
          m_ui->m_valueComboBox->setCurrentIndex(idx);
        }
        else
        {
          m_ui->m_valueComboBox->addItem(QString::fromUtf8(valueString.c_str()));
          m_ui->m_valueComboBox->setCurrentIndex(m_ui->m_valueComboBox->findText(QString::fromUtf8(valueString.c_str())));
        }
      }
    }
  }
}

te::da::Expression* te::qt::widgets::QueryConditionWidget::getExpression()
{
  std::string functionStr = m_ui->m_functionComboBox->currentText().toUtf8().data();

  std::string propertyStr = m_ui->m_propertyComboBox->currentText().toUtf8().data();
  int idx = m_ui->m_propertyComboBox->currentIndex();
  te::dt::Property* property = m_ui->m_propertyComboBox->itemData(idx).value<te::dt::Property*>();
  int type = property->getType();

  //value
  QString value = m_ui->m_valueComboBox->currentText();

  te::da::Literal* l = nullptr;

  if (type == te::dt::DOUBLE_TYPE || type == te::dt::FLOAT_TYPE || type == te::dt::NUMERIC_TYPE)
  {
    l = new te::da::LiteralDouble(value.toDouble());
  }
  else if (type == te::dt::INT16_TYPE)
  {
    l = new te::da::LiteralInt16(value.toInt());
  }
  else if (type == te::dt::INT32_TYPE)
  {
    l = new te::da::LiteralInt32(value.toInt());
  }
  else if (type == te::dt::INT64_TYPE)
  {
    l = new te::da::LiteralInt64(value.toInt());
  }
  else if (type == te::dt::STRING_TYPE)
  {
    l = new te::da::LiteralString(value.toUtf8().data());
  }
  else if (type == te::dt::DATETIME_TYPE)
  {
    std::unique_ptr< te::dt::DateTime > dTPtr;

    try
    {
      dTPtr.reset( te::dt::CreateDateTimeFromString( value.toStdString() ) );
    }
    catch(...)
    {
      dTPtr.reset();
    }

    if( dTPtr.get() == nullptr )
    {
      try
      {
        dTPtr.reset( te::dt::CreateDateFromString( value.toStdString() ) );
      }
      catch(...)
      {
        dTPtr.reset();
      }
    }

    if( dTPtr.get() != nullptr )
    {
      l = new te::da::LiteralDateTime( *dTPtr );
    }
    else
    {
      return nullptr;
    }
  }
  else
  {
    return nullptr;
  }

  //property
  te::da::PropertyName* pName = new te::da::PropertyName(property->getName());

  te::da::Function* func = nullptr;
  if (functionStr == "like")
  {
    std::string strValue = l->getValue()->toString();

    func = new te::da::Like(pName, strValue);
  }
  else
  {
    func = new te::da::Function(functionStr);
    func->add(pName);
    func->add(l);
  }
  
  return func;
}

void te::qt::widgets::QueryConditionWidget::setConnectorParent(te::qt::widgets::QueryConnectorWidget* parent)
{
  m_parent = parent;
}

te::qt::widgets::QueryConnectorWidget* te::qt::widgets::QueryConditionWidget::getConnectorParent()
{
  return m_parent;
}

void te::qt::widgets::QueryConditionWidget::setOperatorsList()
{
  te::da::DataSourceCapabilities dsCap = m_dataSource->getCapabilities();
  te::da::QueryCapabilities queryCap = dsCap.getQueryCapabilities();

  std::vector<std::string> vecOperators;

  std::set<std::string>::iterator it;

  it = queryCap.getComparsionOperators().begin();
  while (it != queryCap.getComparsionOperators().end())
  {
    vecOperators.push_back(*it);
    ++it;
  }

  for (size_t t = 0; t < vecOperators.size(); ++t)
  {
    m_ui->m_functionComboBox->addItem(vecOperators[t].c_str());
  }
}

void te::qt::widgets::QueryConditionWidget::setAttributeList()
{
  if (m_dsType.get())
  {
    for (size_t i = 0; i < m_dsType->size(); ++i)
    {
      std::string propName = m_dsType->getProperty(i)->getName();

      if (m_dsType->getProperty(i)->getType() != te::dt::GEOMETRY_TYPE)
      {
        std::string alias = getAliasName(propName);
        m_ui->m_propertyComboBox->addItem(QString::fromUtf8(alias.c_str()), QVariant::fromValue(m_dsType->getProperty(i)));
      }
    }
  }
}

void te::qt::widgets::QueryConditionWidget::fillValuesComboBox(QStringList& list)
{
  QStringListModel* model = new QStringListModel;

  for(int i = 0; i < list.size(); ++i)
  {
    model->insertRow(model->rowCount());
    model->setData(model->index(model->rowCount() - 1, 0), list.at(i), Qt::DisplayRole);
  }

  QSortFilterProxyModel* proxy = new QSortFilterProxyModel;
  proxy->setSourceModel(model);

  m_ui->m_valueComboBox->setModel(proxy);
  m_ui->m_valueComboBox->setEditable(true);
  m_ui->m_valueComboBox->setCompleter(0);

  // When the edit text changes, use it to filter the proxy model.
  connect(m_ui->m_valueComboBox, SIGNAL(editTextChanged(QString)), proxy, SLOT(setFilterWildcard(QString)));
}

std::string te::qt::widgets::QueryConditionWidget::getAliasName(std::string fullName)
{
  std::string name = fullName;
  std::string aliasName = fullName;

  size_t pos = name.find(".");

  if (pos != std::string::npos)
  {
    aliasName = name.substr(pos + 1, name.size() - 1);
  }

  return aliasName;
}

void te::qt::widgets::QueryConditionWidget::onAddToolButtonClicked()
{
  emit addConnectorClicked(this, m_parent);
}

void te::qt::widgets::QueryConditionWidget::onRemoveToolButtonClicked()
{
  emit removeConnectorClicked(this, m_parent);
}

void te::qt::widgets::QueryConditionWidget::onLoadAttrsToolButtonClicked()
{
  if (m_ui->m_loadToolButton->isChecked())
  {
    std::string propertyName =
        m_ui->m_propertyComboBox->currentText().toUtf8().data();

    te::da::Fields* fields = new te::da::Fields;
    te::da::Field* f =
        new te::da::Field(new te::da::PropertyName(propertyName));
    fields->push_back(f);

    // get alias name
    std::string dataSetName = m_dataSetName;
    std::string aliasName = getAliasName(m_dataSetName);

    te::da::From* from = new te::da::From;
    te::da::FromItem* fi = new te::da::DataSetName(dataSetName, aliasName);
    from->push_back(fi);

    std::unique_ptr<te::da::Select> select(new te::da::Select());

    select->setFields(fields);
    select->setFrom(from);

    std::unique_ptr<te::da::DataSet> dataset;

    try
    {
      dataset = m_dataSource->query(*select);
    }
    catch(const std::exception& e)
    {
      QString msg = tr("An exception has occurred: ");
      msg += Convert2QString(e.what());

      QMessageBox::warning(nullptr, tr("Query Error: "), msg);

      return;
    }
    catch(...)
    {
      QString msg = tr("An unexpected exception has occurred!");
      QMessageBox::warning(nullptr, tr("Query Error: "), msg);

      return;
    }

    if(dataset == nullptr)
    {
      QString msg = tr("An unexpected exception has occurred!");
      QMessageBox::warning(nullptr, tr("Query Error: "), msg);
      return;
    }

    size_t pos = propertyName.find(".");
    if(pos != std::string::npos)
    {
      std::string dataSetAliasName = propertyName.substr(0, pos);
      propertyName = propertyName.substr(pos + 1, propertyName.size() - 1);
    }

    std::set<std::string> valuesStr;
    std::set<double> valuesDouble;
    std::set<int> valuesInt;

    bool found = false;
    std::size_t propertyPos;
    for(std::size_t t = 0; t < dataset->getNumProperties(); ++t)
    {
      if(dataset->getPropertyName(t) == propertyName)
      {
        propertyPos = t;
        found = true;
        break;
      }
    }

    if(found == false)
    {
      return;
    }

    int propertyType = dataset->getPropertyDataType(propertyPos);

    while(dataset->moveNext())
    {
      if(!dataset->isNull(propertyName))
      {
        if(propertyType == te::dt::INT32_TYPE)
        {
          int value = dataset->getInt32(propertyName);

          std::set<int>::iterator it = valuesInt.find(value);

          if(it == valuesInt.end())
            valuesInt.insert(value);
        }
        else if(propertyType == te::dt::INT16_TYPE)
        {
          int value = dataset->getInt16(propertyName);

          std::set<int>::iterator it = valuesInt.find(value);

          if(it == valuesInt.end())
            valuesInt.insert(value);
        }
        else if(propertyType == te::dt::INT64_TYPE)
        {
          int value = dataset->getInt64(propertyName);

          std::set<int>::iterator it = valuesInt.find(value);

          if(it == valuesInt.end())
            valuesInt.insert(value);
        }
        else if(propertyType == te::dt::DOUBLE_TYPE ||
                propertyType == te::dt::FLOAT_TYPE)
        {
          double value = dataset->getDouble(propertyName);

          std::set<double>::iterator it = valuesDouble.find(value);

          if(it == valuesDouble.end())
            valuesDouble.insert(value);
        }
        else if(propertyType == te::dt::NUMERIC_TYPE)
        {
          std::string value = dataset->getNumeric(propertyName);

          std::set<std::string>::iterator it = valuesStr.find(value);

          if(it == valuesStr.end())
            valuesStr.insert(value);
        }
        else
        {
          std::string value = dataset->getAsString(propertyName);

          std::set<std::string>::iterator it = valuesStr.find(value);

          if(it == valuesStr.end())
            valuesStr.insert(value);
        }
      }
    }

    QStringList list;

    if(propertyType == te::dt::INT32_TYPE ||
       propertyType == te::dt::INT16_TYPE || propertyType == te::dt::INT64_TYPE)
    {
      std::set<int>::iterator it = valuesInt.begin();

      while(it != valuesInt.end())
      {
        list.append(QString::number(*it));

        ++it;
      }
    }
    else if(propertyType == te::dt::DOUBLE_TYPE ||
            propertyType == te::dt::FLOAT_TYPE)
    {
      std::set<double>::iterator it = valuesDouble.begin();

      while(it != valuesDouble.end())
      {
        list.append(QString::number(*it, 'f', 5));

        ++it;
      }
    }
    else
    {
      std::set<std::string>::iterator it = valuesStr.begin();

      while(it != valuesStr.end())
      {
        list.append(QString::fromUtf8((*it).c_str()));

        ++it;
      }
    }

    fillValuesComboBox(list);
  }
  else
  {
    QAbstractItemModel* itemModel = m_ui->m_valueComboBox->model();
    QSortFilterProxyModel* proxy =
        dynamic_cast<QSortFilterProxyModel*>(itemModel);

    disconnect(m_ui->m_valueComboBox, SIGNAL(editTextChanged(QString)), proxy,
               SLOT(setFilterWildcard(QString)));

    m_ui->m_valueComboBox->setModel(0);
  }  
}
