/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/widgets/query/QueryConnectorWidget.cpp

  \brief This file has the Query Condition Widget class.
*/

// TerraLib
#include "QueryConditionWidget.h"
#include "QueryConnectorWidget.h"
#include "ui/ui_QueryConnectorWidgetForm.h"

#include "../../../dataaccess/datasource/DataSourceCapabilities.h"



// Qt
#include <QMessageBox>

te::qt::widgets::QueryConnectorWidget::QueryConnectorWidget(te::da::DataSourcePtr dataSource, std::string dataSetName, QWidget* parent, Qt::WindowFlags f)
  : QWidget(parent, f),
    m_ui(new Ui::QueryConnectorWidgetForm),
    m_connectorParent(nullptr),
    m_connectorLeft(nullptr),
    m_connectorRight(nullptr),
    m_conditionLeft(nullptr),
    m_conditionRight(nullptr),
    m_dataSource(dataSource),
    m_dataSetName(dataSetName),
    m_item(nullptr)
{
  m_ui->setupUi(this);

  setConnectorsList();

  connect(m_ui->m_connectorComboBox, SIGNAL(activated(int)), this, SIGNAL(connectorChanged()));
}

te::qt::widgets::QueryConnectorWidget::~QueryConnectorWidget() = default;

void te::qt::widgets::QueryConnectorWidget::setCurrentConnector(te::da::Function* func)
{
  if (func)
  {
    std::string connector = func->getName();

    int idx = m_ui->m_connectorComboBox->findText(connector.c_str());

    if (idx != (-1))
    {
      m_ui->m_connectorComboBox->setCurrentIndex(idx);
    }
    else
    {
      m_ui->m_connectorComboBox->addItem(connector.c_str());
      m_ui->m_connectorComboBox->setCurrentIndex(m_ui->m_connectorComboBox->findText(connector.c_str()));
    }
  }
}

te::da::Expression* te::qt::widgets::QueryConnectorWidget::getExpression()
{
  std::string connectorStr = m_ui->m_connectorComboBox->currentText().toUtf8().data();

  te::da::Function* func = new te::da::Function(connectorStr);

  te::qt::widgets::QueryConditionWidget* conditionLeftWidget = getConditionLeft();
  te::qt::widgets::QueryConnectorWidget* connectorLeftWidget = getConnectorLeft();

  if (conditionLeftWidget)
  {
    te::da::Expression* exp = conditionLeftWidget->getExpression();

    if (exp)
    {
      func->add(exp);
    }
    else
    {
      delete func;
      return nullptr;
    }
  }
  else if (connectorLeftWidget)
  {
    te::da::Expression* exp = connectorLeftWidget->getExpression();

    if (exp)
    {
      func->add(exp);
    }
    else
    {
      delete func;
      return nullptr;
    }
  }
  else
  {
    delete func;
    return nullptr;
  }

  te::qt::widgets::QueryConditionWidget* conditionRightWidget = getConditionRight();
  te::qt::widgets::QueryConnectorWidget* connectorRightWidget = getConnectorRight();

  if (conditionRightWidget)
  {
    te::da::Expression* exp = conditionRightWidget->getExpression();

    if (exp)
    {
      func->add(exp);
    }
    else
    {
      delete func;
      return nullptr;
    }
  }
  else if (connectorRightWidget)
  {
    te::da::Expression* exp = connectorRightWidget->getExpression();

    if (exp)
    {
      func->add(exp);
    }
    else
    {
      delete func;
      return nullptr;
    }
  }
  else
  {
    delete func;
    return nullptr;
  }

  return func;
}

void te::qt::widgets::QueryConnectorWidget::setConnectorParent(te::qt::widgets::QueryConnectorWidget* connectorParent)
{
  m_connectorParent = connectorParent;
}

te::qt::widgets::QueryConnectorWidget* te::qt::widgets::QueryConnectorWidget::getConnectorParent()
{
  return m_connectorParent;
}

void te::qt::widgets::QueryConnectorWidget::setConnectorLeft(te::qt::widgets::QueryConnectorWidget* connectorChild)
{
  m_connectorLeft = connectorChild;

  m_conditionLeft = nullptr;
}

te::qt::widgets::QueryConnectorWidget* te::qt::widgets::QueryConnectorWidget::getConnectorLeft()
{
  return m_connectorLeft;
}

void te::qt::widgets::QueryConnectorWidget::setConnectorRight(te::qt::widgets::QueryConnectorWidget* connectorChild)
{
  m_connectorRight = connectorChild;

  m_conditionRight = nullptr;
}

te::qt::widgets::QueryConnectorWidget* te::qt::widgets::QueryConnectorWidget::getConnectorRight()
{
  return m_connectorRight;
}

void te::qt::widgets::QueryConnectorWidget::setConditionLeft(te::qt::widgets::QueryConditionWidget* condition)
{
  m_conditionLeft = condition;

  m_connectorLeft = nullptr;
}

te::qt::widgets::QueryConditionWidget* te::qt::widgets::QueryConnectorWidget::getConditionLeft()
{
  return m_conditionLeft;
}

void te::qt::widgets::QueryConnectorWidget::setConditionRight(te::qt::widgets::QueryConditionWidget* condition)
{
  m_conditionRight = condition;
 
  m_connectorRight = nullptr;
}

te::qt::widgets::QueryConditionWidget* te::qt::widgets::QueryConnectorWidget::getConditionRight()
{
  return m_conditionRight; 
}

void te::qt::widgets::QueryConnectorWidget::setConnectorsList()
{
  te::da::DataSourceCapabilities dsCap = m_dataSource->getCapabilities();
  te::da::QueryCapabilities queryCap = dsCap.getQueryCapabilities();

  std::set<std::string>::iterator it = queryCap.getLogicalOperators().begin();

  std::vector<std::string> vecConnectors;

  while (it != queryCap.getLogicalOperators().end())
  {
    vecConnectors.push_back(*it);
    ++it;
  }

  for (size_t t = 0; t < vecConnectors.size(); ++t)
  {
    m_ui->m_connectorComboBox->addItem(vecConnectors[t].c_str());
  }
}
