/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/widgets/query/QueryDataSourceDialog.cpp

  \brief This file defines a class for a Query DataSource Dialog Dialog.
*/

// TerraLib
#include "../../../common/StringUtils.h"
#include "../../../dataaccess/dataset/DataSet.h"
#include "../../../dataaccess/dataset/DataSetAdapter.h"
#include "../../../dataaccess/dataset/DataSetTypeConverter.h"
#include "../../../dataaccess/dataset/DataSetType.h"
#include "../../../dataaccess/dataset/ObjectIdSet.h"
#include "../../../dataaccess/datasource/DataSourceCapabilities.h"
#include "../../../dataaccess/datasource/DataSourceFactory.h"
#include "../../../dataaccess/datasource/DataSourceInfoManager.h"
#include "../../../dataaccess/datasource/DataSourceInfo.h"
#include "../../../dataaccess/datasource/DataSourceManager.h"
#include "../../../dataaccess/query/SQLDialect.h"
#include "../../../dataaccess/query/SQLFunctionEncoder.h"
#include "../../../dataaccess/utils/Utils.h"
#include "../../../datatype/DateTimeProperty.h"
#include "../../../datatype/NumericProperty.h"
#include "../../../datatype/SimpleProperty.h"
#include "../../../datatype/StringProperty.h"
#include "../../../geometry/GeometryProperty.h"
#include "../../../maptools/DataSetLayer.h"
#include "../datasource/core/DataSourceType.h"
#include "../datasource/core/DataSourceTypeManager.h"
#include "../datasource/selector/DataSourceSelectorDialog.h"
#include "../layer/utils/DataSet2Layer.h"
#include "../utils/ScopedCursor.h"
#include "../ceditor/ScriptWidget.h"
#include "QueryDataSourceDialog.h"
#include "ui/ui_QueryDataSourceDialogForm.h"
#include "QuickSelectDialog.h"

// Qt
#include <QCompleter>
#include <QCursor>
#include <QFileDialog>
#include <QGridLayout>
#include <QMessageBox>
#include <QTextStream>
#include <QListWidgetItem>
#include <QSortFilterProxyModel>
#include <QStringListModel>

// STL
#include <cassert>
#include <memory>

// Boost
#include <boost/filesystem.hpp>
#include <boost/uuid/random_generator.hpp>
#include <boost/uuid/uuid_io.hpp>

Q_DECLARE_METATYPE(te::map::AbstractLayerPtr)

te::qt::widgets::QueryDataSourceDialog::QueryDataSourceDialog(QWidget* parent, Qt::WindowFlags f)
  : QDialog(parent, f),
    m_ui(new Ui::QueryDataSourceDialogForm)
{
  m_completer = nullptr;

  m_ui->setupUi(this);

  m_ui->m_applyToolButton->setIcon(QIcon::fromTheme("media-playback-start-green"));
  m_ui->m_executeToolButton->setIcon(QIcon::fromTheme("media-playback-start-green-execute"));
  m_ui->m_clearToolButton->setIcon(QIcon::fromTheme("edit-clear"));
  m_ui->m_quickSelectToolButton->setIcon(QIcon::fromTheme("wand"));
  m_ui->m_applySelToolButton->setIcon(QIcon::fromTheme("pointer-selection"));
  m_ui->m_targetDatasourceToolButton->setIcon(QIcon::fromTheme("datasource"));
  m_ui->m_createLayerlToolButton->setIcon(QIcon::fromTheme("layer-new"));
  
  m_ui->m_saveCsvToolButton->setIcon(QIcon::fromTheme("view-data-table"));
  m_ui->m_saveSqlToolButton->setIcon(QIcon::fromTheme("document-save-as"));
  m_ui->m_openSqlToolButton->setIcon(QIcon::fromTheme("document-open"));

  //table view
  m_tableModel = new te::qt::widgets::DataSetTableModel(m_ui->m_tableView);
  m_ui->m_tableView->setModel(m_tableModel);

  //dataset display
  m_dataSetDisplay = new te::qt::widgets::DataSetDisplay(this);

  m_appMapDisplay = nullptr;

  QGridLayout* displayGridLayout = new QGridLayout(m_ui->m_displayWidget);
  displayGridLayout->setContentsMargins(0, 0, 0, 0);
  displayGridLayout->addWidget(m_dataSetDisplay);

  m_ui->m_sqlEditor->setHighlightMode("sql");

  // Signals  slots
  connect(m_ui->m_dataSourceTypeComboBox, SIGNAL(activated(int)), this, SLOT(onDataSourceTypeSelected(int)));
  connect(m_ui->m_dataSourceComboBox, SIGNAL(activated(int)), this, SLOT(onDataSourceSelected(int)));
  connect(m_ui->m_baseDataSetComboBox, SIGNAL(activated(int)), this, SLOT(onBaseDataSetSelected(int)));
  connect(m_ui->m_dataSetListWidget, SIGNAL(itemClicked(QListWidgetItem*)), this, SLOT(onDataSetItemClicked(QListWidgetItem*)));
  connect(m_ui->m_applyToolButton, SIGNAL(clicked()), this, SLOT(onApplyPushButtonClicked()));
  connect(m_ui->m_executeToolButton, SIGNAL(clicked()), this, SLOT(onExecutePushButtonClicked()));
  connect(m_ui->m_clearToolButton, SIGNAL(clicked()), this, SLOT(onClearPushButtonClicked()));
  connect(m_ui->m_saveCsvToolButton, SIGNAL(clicked()), this, SLOT(onSaveCsvToolButtonClicked()));
  connect(m_ui->m_saveSqlToolButton, SIGNAL(clicked()), this, SLOT(onSaveSqlToolButtonClicked()));
  connect(m_ui->m_openSqlToolButton, SIGNAL(clicked()), this, SLOT(onOpenSqlToolButtonClicked()));
  connect(m_ui->m_quickSelectToolButton, SIGNAL(clicked()), this, SLOT(onQuickSelectToolButtonClicked()));
  connect(m_ui->m_applySelToolButton, SIGNAL(clicked()), this, SLOT(onApplySelToolButtonClicked()));
  connect(m_ui->m_createLayerlToolButton, SIGNAL(pressed()), this, SLOT(onCreateLayerToolButtonClicked()));
  connect(m_ui->m_targetDatasourceToolButton, SIGNAL(pressed()), this, SLOT(onTargetDatasourceToolButtonPressed()));
  connect(m_ui->m_targetFileToolButton, SIGNAL(pressed()), this,  SLOT(onTargetFileToolButtonPressed()));
  connect(m_ui->m_pkTableComboBox, SIGNAL(activated(int)), this, SLOT(onPkTableComboBoxSelected(int)));
  connect(m_ui->m_attrDataSetListWidget, SIGNAL(itemDoubleClicked(QListWidgetItem*)), this, SLOT(onListWidgetDoubleClicked(QListWidgetItem*)));
  connect(m_ui->m_dataSetListWidget, SIGNAL(itemDoubleClicked(QListWidgetItem*)), this, SLOT(onListWidgetDoubleClicked(QListWidgetItem*)));

  //load data sources type
  loadDataSourcesType();

  //load data sources information
  loadDataSourcesInformation("");

  //
  m_ui->m_helpPushButton->setPageReference("widgets/query/query_datasource.html");
}

te::qt::widgets::QueryDataSourceDialog::~QueryDataSourceDialog()
{
  m_keyWords.clear();
}

void te::qt::widgets::QueryDataSourceDialog::setLayerList(std::list<te::map::AbstractLayerPtr> layerList)
{
  m_layerList = layerList;

  if(m_ui->m_baseDataSetComboBox->count() > 0)
    onBaseDataSetSelected(0);
}

void te::qt::widgets::QueryDataSourceDialog::setAppMapDisplay(te::qt::widgets::MapDisplay* appMapDisplay)
{
  m_appMapDisplay = appMapDisplay;
}

void te::qt::widgets::QueryDataSourceDialog::loadDataSourcesType()
{
  m_ui->m_dataSourceTypeComboBox->addItem("All");

  std::map<std::string, DataSourceType*>::const_iterator it = DataSourceTypeManager::getInstance().begin();
  std::map<std::string, DataSourceType*>::const_iterator itend = DataSourceTypeManager::getInstance().end();

  while (it != itend)
  {
    if (it->second->hasDatabaseSupport() || it->second->hasVectorialSupport())
    {
      m_ui->m_dataSourceTypeComboBox->addItem(it->second->getTitle().c_str(), QVariant(it->second->getName().c_str()));
    }

    ++it;
  }

    m_ui->m_dataSourceTypeComboBox->setCurrentIndex(0);
}

void te::qt::widgets::QueryDataSourceDialog::loadDataSourcesInformation(std::string dsType)
{
  if (!m_completer)
    m_completer = new QCompleter(this);

  QStringListModel* model = new QStringListModel(m_completer);

  m_ui->m_dataSourceComboBox->clear();

  std::map<std::string, te::da::DataSourceInfoPtr>::const_iterator it = te::da::DataSourceInfoManager::getInstance().begin();
  std::map<std::string, te::da::DataSourceInfoPtr>::const_iterator itend =te::da::DataSourceInfoManager::getInstance().end();

  while(it != itend)
  {
    try
    {
      if (dsType == "" || it->second->getType() == dsType)
      {
        te::da::DataSourcePtr ds = te::da::GetDataSource(it->second->getId(), false);

        const te::da::SQLDialect* dialect = ds->getDialect();

        if (dialect)
        {
          model->insertRow(model->rowCount());
          model->setData(model->index(model->rowCount() - 1, 0), it->second->getTitle().c_str(), Qt::DisplayRole);
          m_ui->m_dataSourceComboBox->addItem(it->second->getTitle().c_str(), QVariant(it->second->getId().c_str()));
        }
      }
    }
    catch (...)
    {
      ++it;
      continue;
    }

    ++it;
  }

  QSortFilterProxyModel* proxy = new QSortFilterProxyModel(m_ui->m_dataSourceComboBox);
  proxy->setSourceModel(model);
  proxy->setFilterCaseSensitivity(Qt::CaseInsensitive);

  m_completer->setCaseSensitivity(Qt::CaseInsensitive);
  m_completer->setModel(proxy);
  m_completer->setCompletionMode(QCompleter::UnfilteredPopupCompletion);
  m_completer->setModelSorting(QCompleter::CaseInsensitivelySortedModel);

  m_ui->m_dataSourceComboBox->setEditable(true);
  m_ui->m_dataSourceComboBox->setModelColumn(0);
  m_ui->m_dataSourceComboBox->setCompleter(m_completer);

  // When the edit text changes, use it to filter the proxy model.
  connect(m_ui->m_dataSourceComboBox, SIGNAL(editTextChanged(QString)), proxy, SLOT(setFilterWildcard(QString)));
}

void te::qt::widgets::QueryDataSourceDialog::buildMap()
{
  m_keyWords.clear();

  // add defaul SQL key words
  m_keyWords.insert(std::map<std::string, Qt::GlobalColor>::value_type("SELECT", Qt::blue));
  m_keyWords.insert(std::map<std::string, Qt::GlobalColor>::value_type("FROM", Qt::blue));
  m_keyWords.insert(std::map<std::string, Qt::GlobalColor>::value_type("WHERE", Qt::blue));
  m_keyWords.insert(std::map<std::string, Qt::GlobalColor>::value_type("JOIN", Qt::blue));
  m_keyWords.insert(std::map<std::string, Qt::GlobalColor>::value_type("INNER", Qt::blue));
  m_keyWords.insert(std::map<std::string, Qt::GlobalColor>::value_type("LEFT", Qt::blue));
  m_keyWords.insert(std::map<std::string, Qt::GlobalColor>::value_type("RIGHT", Qt::blue));
  m_keyWords.insert(std::map<std::string, Qt::GlobalColor>::value_type("AS", Qt::blue));
  m_keyWords.insert(std::map<std::string, Qt::GlobalColor>::value_type("GROUP", Qt::blue));
  m_keyWords.insert(std::map<std::string, Qt::GlobalColor>::value_type("ORDER", Qt::blue));
  m_keyWords.insert(std::map<std::string, Qt::GlobalColor>::value_type("BY", Qt::blue));

  //add data source specific key words
  std::string dataSourceId = m_ui->m_dataSourceComboBox->itemData(m_ui->m_dataSourceComboBox->currentIndex()).toString().toUtf8().data();

  te::da::DataSourcePtr ds = te::da::GetDataSource(dataSourceId);

  const te::da::SQLDialect* dialect = ds->getDialect();

  std::vector<std::string> names = dialect->getRegisteredNames();

  for(std::size_t t = 0; t < names.size(); ++t)
  {
    QString s = names[t].c_str();
    QString sUpper = s.toUpper();

    m_keyWords.insert(std::map<std::string, Qt::GlobalColor>::value_type(sUpper.toUtf8().data(), Qt::red));
  }

}

void te::qt::widgets::QueryDataSourceDialog::onDataSourceTypeSelected(int index)
{
  std::string type = "";
  if (index != 0)
    type = m_ui->m_dataSourceTypeComboBox->itemData(index).toString().toUtf8().data();

  //load data sources information
  loadDataSourcesInformation(type);
}

void te::qt::widgets::QueryDataSourceDialog::onDataSourceSelected(int index)
{
  m_ui->m_baseDataSetComboBox->clear();
  m_ui->m_dataSetListWidget->clear();
  m_ui->m_attrDataSetListWidget->clear();
  m_ui->m_pkTableComboBox->clear();
  m_ui->m_pkAttrComboBox->clear();

  std::string dataSourceId = m_ui->m_dataSourceComboBox->itemData(index).toString().toUtf8().data();

  te::da::DataSourcePtr ds = te::da::GetDataSource(dataSourceId);

  if(!ds->isOpened())
    ds->open();

  std::vector<std::string> dataSetNames;
  te::da::GetDataSetNames(dataSetNames, dataSourceId);

  for(std::size_t t = 0; t < dataSetNames.size(); ++t)
  {
    m_ui->m_baseDataSetComboBox->addItem(dataSetNames[t].c_str());
    m_ui->m_dataSetListWidget->addItem(dataSetNames[t].c_str());
    m_ui->m_pkTableComboBox->addItem(dataSetNames[t].c_str());
  }

  if(m_ui->m_baseDataSetComboBox->count() > 0)
    onBaseDataSetSelected(0);

  if (m_ui->m_pkTableComboBox->count() > 0)
    onPkTableComboBoxSelected(0);

  buildMap();
}

void te::qt::widgets::QueryDataSourceDialog::onBaseDataSetSelected(int index)
{
  std::string dataSet = m_ui->m_baseDataSetComboBox->itemText(index).toUtf8().data();

  m_ui->m_layerComboBox->clear();

  std::list<te::map::AbstractLayerPtr>::iterator it = m_layerList.begin();

  while(it != m_layerList.end())
  {
    te::map::AbstractLayerPtr l = *it;

    std::unique_ptr<te::da::DataSetType> dsType = l->getSchema();

    te::map::DataSetLayer* dsLayer = dynamic_cast<te::map::DataSetLayer*>(l.get());

    if(dsLayer && dsType->getName() == dataSet)
      m_ui->m_layerComboBox->addItem(l->getTitle().c_str(), QVariant::fromValue(l));

    ++it;
  }

  m_ui->m_pkTableComboBox->setCurrentIndex(index);
  onPkTableComboBoxSelected(index);
}

void te::qt::widgets::QueryDataSourceDialog::onDataSetItemClicked(QListWidgetItem* item)
{
  m_ui->m_attrDataSetListWidget->clear();

  if(!item)
    return;

  std::string dataSourceId = m_ui->m_dataSourceComboBox->itemData(m_ui->m_dataSourceComboBox->currentIndex()).toString().toUtf8().data();

  std::string dataSetName = item->text().toUtf8().data();

  te::da::DataSetType* dsType = te::da::GetDataSetType(dataSetName, dataSourceId);

  std::vector<te::dt::Property*> propVec = dsType->getProperties();

  for(std::size_t t = 0; t < propVec.size(); ++t)
  {
    m_ui->m_attrDataSetListWidget->addItem(propVec[t]->getName().c_str());
  }

  m_ui->m_attrDataSetListWidget->addItem("*");

  delete dsType;
}

void te::qt::widgets::QueryDataSourceDialog::onPkTableComboBoxSelected(int index)
{
  std::string dataSetName = m_ui->m_pkTableComboBox->itemText(index).toUtf8().data();

  std::string dataSourceId = m_ui->m_dataSourceComboBox->itemData(m_ui->m_dataSourceComboBox->currentIndex()).toString().toUtf8().data();

  te::da::DataSetType* dsType = te::da::GetDataSetType(dataSetName, dataSourceId);

  std::vector<te::dt::Property*> propVec = dsType->getProperties();

  for (std::size_t t = 0; t < propVec.size(); ++t)
  {
    m_ui->m_pkAttrComboBox->addItem(propVec[t]->getName().c_str());
  }

  delete dsType;
}

void te::qt::widgets::QueryDataSourceDialog::onApplyPushButtonClicked()
{
  te::qt::widgets::ScopedCursor cursor(Qt::WaitCursor);

  m_ui->m_sqlEditor->setFocus();

  if(m_ui->m_sqlEditor->text().isEmpty())
  {
    return;
  }

  std::string sql;

  if (m_ui->m_sqlEditor->selectedText().isEmpty())
    sql = m_ui->m_sqlEditor->text().toUtf8().data();
  else
    sql = m_ui->m_sqlEditor->selectedText().toUtf8().data();

  std::vector<std::string> tokens;
  te::common::Tokenize(sql, tokens, " ");

  if (te::common::Convert2UCase(tokens[0]) != te::common::Convert2UCase("select"))
  {
    QString errorMessage = "SQL Error: ";
    errorMessage += tr("Use the \"Execute\" option to execute a instruction.");
    errorMessage += "\n";
    errorMessage += "-------------------------------------------------------------------------\n";

    m_ui->m_logPlainTextEdit->appendPlainText(errorMessage);

    m_ui->m_previewTabWidget->setCurrentIndex(1);
    
    return;
  }

  std::string dataSourceId = m_ui->m_dataSourceComboBox->itemData(m_ui->m_dataSourceComboBox->currentIndex()).toString().toUtf8().data();

  te::da::DataSourcePtr ds = te::da::GetDataSource(dataSourceId);

  //get dataset
  std::unique_ptr<te::da::DataSet> dataSet;
  te::da::DataSetTypePtr dsType;

  try
  {
   dataSet = ds->query(sql);
   dsType.reset(te::da::CreateDataSetType("result", dataSet.get()).release());
  }
  catch(const std::exception& e)
  {
    m_dataSetDisplay->clear();
    m_tableModel->setDataSet(nullptr);

    std::string errorMessage = "SQL Error: ";
    errorMessage += e.what();
    errorMessage += "\n";
    errorMessage += "-------------------------------------------------------------------------\n";

    m_ui->m_logPlainTextEdit->appendPlainText(errorMessage.c_str());

    m_ui->m_previewTabWidget->setCurrentIndex(2);

    return;
  }

  std::string message = "SQL Done: ";
  message += sql;
  message += "\n";
  message += "-------------------------------------------------------------------------\n";

  m_ui->m_logPlainTextEdit->appendPlainText(message.c_str());

  //draw dataset
  m_dataSetDisplay->clear();

  bool draw = false;
  for(std::size_t t = 0;  t < dataSet->getNumProperties(); ++t)
  {
    int type = dataSet->getPropertyDataType(t);

    if(type == te::dt::GEOMETRY_TYPE)
    {
      draw = true;
      break;
    }
  }

  if (draw)
  {
    m_dataSetDisplay->draw(dsType, ds, dataSet.get());

    m_ui->m_previewTabWidget->setCurrentIndex(0);
  }
  else
  {
    m_dataSetDisplay->clear();

    m_ui->m_previewTabWidget->setCurrentIndex(1);
  }

  //show dataset on table
  m_tableModel->setDataSet(dataSet.release());

}

void te::qt::widgets::QueryDataSourceDialog::onExecutePushButtonClicked()
{
  te::qt::widgets::ScopedCursor cursor(Qt::WaitCursor);

  m_ui->m_sqlEditor->setFocus();

  if (m_ui->m_sqlEditor->text().isEmpty())
  {
    return;
  }

  std::string sql;

  if (m_ui->m_sqlEditor->selectedText().isEmpty())
    sql = m_ui->m_sqlEditor->text().toUtf8().data();
  else
    sql = m_ui->m_sqlEditor->selectedText().toUtf8().data();

  std::vector<std::string> tokens;
  te::common::Tokenize(sql, tokens, " ");

  if (te::common::Convert2UCase(tokens[0]) == te::common::Convert2UCase("select"))
  {
    QString errorMessage = "SQL Error: ";
    errorMessage += tr("Use the \"Query\" option to execute a selection.");
    errorMessage += "\n";
    errorMessage += "-------------------------------------------------------------------------\n";

    m_ui->m_logPlainTextEdit->appendPlainText(errorMessage);

    m_ui->m_previewTabWidget->setCurrentIndex(2);

    return;
  }

  std::string dataSourceId = m_ui->m_dataSourceComboBox->itemData(m_ui->m_dataSourceComboBox->currentIndex()).toString().toUtf8().data();

  te::da::DataSourcePtr ds = te::da::GetDataSource(dataSourceId);

  std::string dataSetName = m_ui->m_baseDataSetComboBox->currentText().toUtf8().data();

  te::da::DataSetTypePtr dsType(te::da::GetDataSetType(dataSetName, dataSourceId));
  
  try
  {
    ds->execute(sql);
  }
  catch (const std::exception& e)
  {
    m_dataSetDisplay->clear();
    m_tableModel->setDataSet(nullptr);

    std::string errorMessage = "SQL Error: ";
    errorMessage += e.what();
    errorMessage += "\n";
    errorMessage += "-------------------------------------------------------------------------\n";

    m_ui->m_logPlainTextEdit->appendPlainText(errorMessage.c_str());

    m_ui->m_previewTabWidget->setCurrentIndex(2);

    return;
  }

  std::string message = "SQL Done: ";
  message += sql;
  message += "\n";
  message += "-------------------------------------------------------------------------\n";

  m_ui->m_logPlainTextEdit->appendPlainText(message.c_str());

  m_ui->m_previewTabWidget->setCurrentIndex(2);
}

void te::qt::widgets::QueryDataSourceDialog::onClearPushButtonClicked()
{
  m_ui->m_sqlEditor->setText("");

  m_dataSetDisplay->clear();

  m_tableModel->setDataSet(nullptr);
}

void te::qt::widgets::QueryDataSourceDialog::onSQLEditorTextChanged()
{
}

void te::qt::widgets::QueryDataSourceDialog::onSaveSqlToolButtonClicked()
{
  //select file
  QString path = QFileDialog::getSaveFileName(this, tr("Set a SQL file..."), "", tr("SQL File *.sql"));
  
  if(path.isNull())
    return;

  if (!path.endsWith(".sql"))
    path += ".sql";

 //open file
  QFile file(path);

  if(!file.open(QIODevice::WriteOnly | QIODevice::Text)) 
  {
    QMessageBox::warning(this, tr("Query DataSource"), file.errorString());
    return;
  }

  // save to file
  QTextStream out(&file);

  out << m_ui->m_sqlEditor->text();

  file.close();
}

void te::qt::widgets::QueryDataSourceDialog::onSaveCsvToolButtonClicked()
{
  // check input parameters
  if (m_ui->m_sqlEditor->text().isEmpty())
  {
    QMessageBox::information(this, tr("Warning"), tr("SQL not defined."));
    return;
  }

  //select file
  QString path = QFileDialog::getSaveFileName(this, tr("Set a CSV file..."), "", tr("CSV File *.csv"));

  if (path.isNull())
    return;

  if (!path.endsWith(".csv"))
    path += ".csv";

  te::qt::widgets::ScopedCursor cursor(Qt::WaitCursor);

  //create dataset
  std::string dataSourceId = m_ui->m_dataSourceComboBox->itemData(m_ui->m_dataSourceComboBox->currentIndex()).toString().toUtf8().data();

  te::da::DataSourcePtr ds = te::da::GetDataSource(dataSourceId);

  std::string sql;

  if (m_ui->m_sqlEditor->selectedText().isEmpty())
    sql = m_ui->m_sqlEditor->text().toUtf8().data();
  else
    sql = m_ui->m_sqlEditor->selectedText().toUtf8().data();

  std::unique_ptr<te::da::DataSet> dataSet;

  try
  {
    dataSet = ds->query(sql);
  }
  catch (...)
  {
    QMessageBox::warning(this, tr("Query DataSource"), tr("Error executing SQL."));
    return;
  }

  if (dataSet->size() == 0)
  {
    QMessageBox::warning(this, tr("Query DataSource"), tr("Query result is empty."));
    return;
  }

  te::da::DataSourcePtr outputDataSource;

  //create new data source
  boost::filesystem::path uri(path.toUtf8().data());

  std::string connInfo("file://" + uri.string());

  std::string filename = uri.filename().string();
  //save data
  std::unique_ptr<te::da::DataSetType> dsType(new te::da::DataSetType(filename.substr(0, filename.find('.'))));//dataSetName));

  dataSet->moveFirst();

  std::set<std::string> names;

  int srid = 0;

  for (std::size_t t = 0; t < dataSet->getNumProperties(); ++t)
  {
    //check if the property name its duplicated
    std::string propName = dataSet->getPropertyName(t);

    int count = 1;
    while (names.find(propName) != names.end())
    {
      propName += "_";
      propName += te::common::Convert2String(count);
    }

    names.insert(propName);

    //create output property
    te::dt::Property* p = nullptr;
    if (dataSet->getPropertyDataType(t) != te::dt::GEOMETRY_TYPE)
    {
      if (dataSet->getPropertyDataType(t) == te::dt::STRING_TYPE)
      {
        p = new te::dt::StringProperty(propName, te::dt::VAR_STRING, 255, false);
      }
      else if (dataSet->getPropertyDataType(t) == te::dt::NUMERIC_TYPE)
      {
        p = new te::dt::NumericProperty(propName, 0, 0, false);
      }
      else if (dataSet->getPropertyDataType(t) == te::dt::DATETIME_TYPE)
      {
        std::unique_ptr<te::dt::DateTime> dateTime = dataSet->getDateTime(t);
        
        p = new te::dt::DateTimeProperty(propName, dateTime->getDateTimeType());
      }
      else
      {
        p = new te::dt::SimpleProperty(propName, dataSet->getPropertyDataType(t));
      }
    }
    else
    {
      std::unique_ptr<te::gm::Geometry> geom = dataSet->getGeometry(t);

      srid = geom->getSRID();

      p = new te::gm::GeometryProperty(propName, srid, geom->getGeomTypeId());
    }

    //add property to output datasetype
    if (p)
    {
      dsType->add(p);
    }
    else
    {
      QMessageBox::warning(this, tr("Query DataSource"), tr("Error creating output dataset."));
      return;
    }
  }

  dataSet->moveBeforeFirst();

  std::unique_ptr<te::da::DataSource> dsOGR = te::da::DataSourceFactory::make("OGR", connInfo);
  dsOGR->open();

  te::da::Create(dsOGR.get(), dsType.get(), dataSet.get());

  dsOGR->close();
  connInfo.clear();

  QMessageBox::information(this, tr("Query DataSource"), tr("Data was exported."));
}

void te::qt::widgets::QueryDataSourceDialog::onOpenSqlToolButtonClicked()
{
  //select file
  QString path = QFileDialog::getOpenFileName(this, tr("Select a SQL file..."), "", tr("SQL File *.sql"));
  
  if(path.isNull())
    return;
            
  m_ui->m_sqlEditor->open(path);
}

void te::qt::widgets::QueryDataSourceDialog::onQuickSelectToolButtonClicked()
{
  te::qt::widgets::QuickSelectDialog dlg(this);

  //set current datasource
  std::string dataSourceId = m_ui->m_dataSourceComboBox->itemData(m_ui->m_dataSourceComboBox->currentIndex()).toString().toUtf8().data();

  te::da::DataSourcePtr ds = te::da::GetDataSource(dataSourceId);

  dlg.setInputDataSource(ds);

  if (dlg.exec() == QDialog::Accepted)
  {
    std::string selectStr = dlg.getSelectString();

    m_ui->m_sqlEditor->insert(selectStr.c_str());
  }
}

void te::qt::widgets::QueryDataSourceDialog::onApplySelToolButtonClicked()
{
  if(m_ui->m_sqlEditor->text().isEmpty())
  {
    QMessageBox::information(this, tr("Warning"), tr("SQL not defined."));
    return;
  }

  QVariant varLayer = m_ui->m_layerComboBox->itemData(m_ui->m_layerComboBox->currentIndex(), Qt::UserRole);
  te::map::AbstractLayerPtr layer = varLayer.value<te::map::AbstractLayerPtr>();

  if(!layer.get())
  {
    QMessageBox::warning(this, tr("Query DataSource"), tr("No layer selected."));
    return;
  }

  te::qt::widgets::ScopedCursor cursor(Qt::WaitCursor);

  std::string dataSourceId = m_ui->m_dataSourceComboBox->itemData(m_ui->m_dataSourceComboBox->currentIndex()).toString().toUtf8().data();

  te::da::DataSourcePtr ds = te::da::GetDataSource(dataSourceId);

  std::string dataSetName = m_ui->m_baseDataSetComboBox->currentText().toUtf8().data();

  te::da::DataSetTypePtr dsType(te::da::GetDataSetType(dataSetName, dataSourceId));

  std::string sql;

  if (m_ui->m_sqlEditor->selectedText().isEmpty())
    sql = m_ui->m_sqlEditor->text().toUtf8().data();
  else
    sql = m_ui->m_sqlEditor->selectedText().toUtf8().data();

  //get dataset
  std::unique_ptr<te::da::DataSet> dataSet;
  
  try
  {
    dataSet = ds->query(sql);
  }
  catch(...)
  {
    QMessageBox::warning(this, tr("Query DataSource"), tr("Error executing SQL."));
    return;
  }

  try
  {
    if(m_ui->m_newSelRadioButton->isChecked())
    {
      // Generates the oids
      dataSet->moveBeforeFirst();
      te::da::ObjectIdSet* oids = te::da::GenerateOIDSet(dataSet.get(), dsType.get());

      oids->setExpressionByInClause();

      layer->clearSelected();
      layer->select(oids);
    }
    else if(m_ui->m_addSelRadioButton->isChecked())
    {
      // Generates the oids
      dataSet->moveBeforeFirst();
      te::da::ObjectIdSet* oids = te::da::GenerateOIDSet(dataSet.get(), dsType.get());

      oids->setExpressionByInClause();

      layer->select(oids);
    }

    if(m_appMapDisplay)
    {
      m_appMapDisplay->refresh();
    }
  }
  catch(te::common::Exception& e)
  {
    QMessageBox::warning(this, tr("Query DataSource"), tr("Error selecting objects: ") + e.what());
    return;
  }
  QMessageBox::information(this, tr("Query DataSource"), tr("Selection done."));
}

void te::qt::widgets::QueryDataSourceDialog::onCreateLayerToolButtonClicked()
{
  // check input parameters
  if(m_ui->m_sqlEditor->text().isEmpty())
  {
    QMessageBox::information(this, tr("Warning"), tr("SQL not defined."));
    return;
  }

  if(m_ui->m_repositoryLineEdit->text().isEmpty())
  {
    QMessageBox::information(this, tr("Warning"), tr("Define a repository for the result."));
    return;
  }
       
  if(m_ui->m_newLayerNameLineEdit->text().isEmpty())
  {
    QMessageBox::information(this, tr("Warning"), tr("Define a name for the resulting layer."));
    return;
  }

  te::qt::widgets::ScopedCursor cursor(Qt::WaitCursor);

  //create dataset
  std::string dataSourceId = m_ui->m_dataSourceComboBox->itemData(m_ui->m_dataSourceComboBox->currentIndex()).toString().toUtf8().data();

  te::da::DataSourcePtr ds = te::da::GetDataSource(dataSourceId);

  std::string inputDataSetName = m_ui->m_baseDataSetComboBox->currentText().toUtf8().data();

  te::da::DataSetTypePtr inputDataSetType(te::da::GetDataSetType(inputDataSetName, dataSourceId));

  std::string sql;

  if(m_ui->m_sqlEditor->selectedText().isEmpty())
    sql = m_ui->m_sqlEditor->text().toUtf8().data();
  else
    sql = m_ui->m_sqlEditor->selectedText().toUtf8().data();

  std::unique_ptr<te::da::DataSet> dataSet;

  try
  {
    dataSet = ds->query(sql);
  }
  catch(...)
  {
    QMessageBox::warning(this, tr("Query DataSource"), tr("Error executing SQL."));
    return;
  }

  if (dataSet->size() == 0)
  {
    QMessageBox::warning(this, tr("Query DataSource"), tr("Query result is empty."));
    return;
  }

  //create / get data source
  te::da::DataSourcePtr outputDataSource;

  if(m_toFile)
  {
    //create new data source
    boost::filesystem::path uri(m_ui->m_repositoryLineEdit->text().toUtf8().data());

    const std::string connInfo("file://" + uri.string());

    boost::uuids::basic_random_generator<boost::mt19937> gen;
    boost::uuids::uuid u = gen();
    std::string id_ds = boost::uuids::to_string(u);

    te::da::DataSourceInfoPtr dsInfoPtr(new te::da::DataSourceInfo);
    dsInfoPtr->setConnInfo(connInfo);
    dsInfoPtr->setTitle(uri.stem().string());
    dsInfoPtr->setAccessDriver("OGR");
    dsInfoPtr->setType("OGR");
    dsInfoPtr->setDescription(uri.string());
    dsInfoPtr->setId(id_ds);

    te::da::DataSourceInfoManager::getInstance().add(dsInfoPtr);

    outputDataSource = te::da::DataSourceManager::getInstance().get(id_ds, "OGR", dsInfoPtr->getConnInfo());
  }
  else
  {
    outputDataSource = te::da::GetDataSource(m_outputDatasource->getId());
  }

  //get output dataset name
  std::string dataSetName = m_ui->m_newLayerNameLineEdit->text().toUtf8().data();

  std::size_t idx = dataSetName.find(".");
  if (idx != std::string::npos)
        dataSetName=dataSetName.substr(0,idx);

  //save data
  std::unique_ptr<te::da::DataSetType> dsType(new te::da::DataSetType(dataSetName));

  dataSet->moveFirst();

  std::set<std::string> names;

  bool hasGeomProp = false;

  int srid = TE_UNKNOWN_SRS;

  for(std::size_t t = 0; t < dataSet->getNumProperties(); ++t)
  {
    //check if the property name its duplicated
    std::string propName = dataSet->getPropertyName(t);

    int count = 1;
    while(names.find(propName) != names.end())
    {
      propName += "_";
      propName += te::common::Convert2String(count);
    }

    names.insert(propName);

    //create output property
    te::dt::Property* p = nullptr;
    if(dataSet->getPropertyDataType(t) != te::dt::GEOMETRY_TYPE)
    {
      if (dataSet->getPropertyDataType(t) == te::dt::STRING_TYPE)
      {
        p = new te::dt::StringProperty(propName, te::dt::VAR_STRING, 255, false);
      }
      else if (dataSet->getPropertyDataType(t) == te::dt::NUMERIC_TYPE)
      {
        p = new te::dt::NumericProperty(propName, 0, 0, false);
      }
      else if (dataSet->getPropertyDataType(t) == te::dt::DATETIME_TYPE)
      {
        p = new te::dt::DateTimeProperty(propName);
      }
      else
        p = new te::dt::SimpleProperty(propName, dataSet->getPropertyDataType(t));
    }
    else
    {
      hasGeomProp = true;

      std::unique_ptr<te::gm::Geometry> geom = dataSet->getGeometry(t);

      srid = geom->getSRID();

      p = new te::gm::GeometryProperty(propName, srid, geom->getGeomTypeId());
    }
    
    //add property to output datasetype
    if(p)
    {
      dsType->add(p);

      //check primary key
      if (m_ui->m_pkCheckBox->isChecked())
      {
        std::string pkAttrName = m_ui->m_pkAttrComboBox->itemText(m_ui->m_pkAttrComboBox->currentIndex()).toUtf8().data();

        if (te::common::Convert2UCase(pkAttrName) == te::common::Convert2UCase(p->getName()))
        {
          std::string pkName = dataSetName + "_" + p->getName() + "_pk";
          te::da::PrimaryKey* pk = new te::da::PrimaryKey(pkName, dsType.get());
          pk->add(p);
        }
      }
    }
    else
    {
       QMessageBox::warning(this, tr("Query DataSource"), tr("Error creating output dataset."));
      return;
    }
  }

  dataSet->moveBeforeFirst();

  //create converter in case property name changed
  te::da::DataSetTypeConverter* converter = new te::da::DataSetTypeConverter(dsType.get(), outputDataSource->getCapabilities(), outputDataSource->getEncoding());

  if (hasGeomProp)
  {
    te::da::AssociateDataSetTypeConverterSRID(converter, srid);
  }
  
  std::unique_ptr<te::da::DataSetAdapter> dsAdapter(te::da::CreateAdapter(dataSet.get(), converter));
  
  //save data
  std::map<std::string, std::string> options;

  outputDataSource->createDataSet(dsType.get(), options);

  outputDataSource->add(dataSetName, dsAdapter.get(), options);

  //create layer
  try
  {
    te::qt::widgets::DataSet2Layer converter(outputDataSource->getId());
      
    te::da::DataSetTypePtr dt(outputDataSource->getDataSetType(dataSetName).release());

    te::map::AbstractLayerPtr layer = converter(dt);

    emit createNewLayer(layer);
  }
  catch(te::common::Exception& e)
  {
    QMessageBox::warning(this, tr("Query DataSource"), tr("Error creating layer. ") + e.what());
    return;
  }
  QMessageBox::information(this, tr("Query DataSource"), tr("Layer created."));
}

void te::qt::widgets::QueryDataSourceDialog::onTargetDatasourceToolButtonPressed()
{
  m_ui->m_newLayerNameLineEdit->clear();
  m_ui->m_newLayerNameLineEdit->setEnabled(true);

  te::qt::widgets::DataSourceSelectorDialog dlg(this);
  dlg.exec();

  std::list<te::da::DataSourceInfoPtr> dsPtrList = dlg.getSelecteds();

  if(dsPtrList.empty())
    return;

  std::list<te::da::DataSourceInfoPtr>::iterator it = dsPtrList.begin();

  m_ui->m_repositoryLineEdit->setText(QString(it->get()->getTitle().c_str()));

  m_outputDatasource = *it;
  
  m_toFile = false;
}

void te::qt::widgets::QueryDataSourceDialog::onTargetFileToolButtonPressed()
{
  m_ui->m_newLayerNameLineEdit->clear();
  m_ui->m_repositoryLineEdit->clear();
  
  QString fileName = QFileDialog::getSaveFileName(this, tr("Save as..."), QString(), tr("Shapefile (*.shp *.SHP);;"),nullptr, QFileDialog::DontConfirmOverwrite);
  
  if (fileName.isEmpty())
    return;
  
  boost::filesystem::path outfile(fileName.toUtf8().data());

  m_ui->m_repositoryLineEdit->setText(outfile.string().c_str());

  m_ui->m_newLayerNameLineEdit->setText(outfile.leaf().string().c_str());

  m_ui->m_newLayerNameLineEdit->setEnabled(false);
  
  m_toFile = true;
}

void te::qt::widgets::QueryDataSourceDialog::onListWidgetDoubleClicked(QListWidgetItem* item)
{
  m_ui->m_sqlEditor->insert(item->text());
  m_ui->m_sqlEditor->insert(" ");
}

