/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
 \file terralib/qt/widgets/query/QuickSelectDialog.cpp

 \brief This file defines a class for a Select Dialog.
 */

// TerraLib

#include "QuickSelectDialog.h"
#include "QueryBuilderDialog.h"
#include "ui/ui_QuickSelectDialogForm.h"

#include "../../../dataaccess/datasource/DataSourceCapabilities.h"
#include "../../../dataaccess/query/DataSetName.h"
#include "../../../dataaccess/query/EqualTo.h"
#include "../../../dataaccess/query/Expression.h"
#include "../../../dataaccess/query/Field.h"
#include "../../../dataaccess/query/Fields.h"
#include "../../../dataaccess/query/From.h"
#include "../../../dataaccess/query/FromItem.h"
#include "../../../dataaccess/query/FunctionNames.h"
#include "../../../dataaccess/query/GroupBy.h"
#include "../../../dataaccess/query/GroupByItem.h"
#include "../../../dataaccess/query/LiteralInt16.h"
#include "../../../dataaccess/query/OrderBy.h"
#include "../../../dataaccess/query/OrderByItem.h"
#include "../../../dataaccess/query/PropertyName.h"
#include "../../../dataaccess/query/QueryCapabilities.h"
#include "../../../dataaccess/query/SQLVisitor.h"
#include "../../../dataaccess/query/UnaryFunction.h"
#include "../../../dataaccess/query/Where.h"
#include "../../../dataaccess/utils/Utils.h"

te::qt::widgets::QuickSelectDialog::QuickSelectDialog(QWidget* parent, Qt::WindowFlags f)
  : QDialog(parent, f),
  m_ui(new Ui::QuickSelectDialogForm),
  m_queryWidget(nullptr)
{
  m_ui->setupUi(this);

  m_ui->m_addAttrToolButton->setIcon(QIcon::fromTheme("list-add"));
  m_ui->m_removeAttrToolButton->setIcon(QIcon::fromTheme("list-remove"));
  m_ui->m_checkToolButton->setIcon(QIcon::fromTheme("check"));

  m_queryLayout = new QGridLayout(m_ui->m_whereClauseWidget);

  m_queryWidget = new te::qt::widgets::QueryBuilderWidget(m_ui->m_whereClauseWidget);

  m_queryLayout->addWidget(m_queryWidget);
  m_queryLayout->setContentsMargins(0, 0, 0, 0);
   
  // Signals & slots
  connect(m_ui->m_datasetComboBox, SIGNAL(activated(int)), this, SLOT(onDataSetComboBoxSelected(int)));
  connect(m_ui->m_addAttrToolButton, SIGNAL(clicked()), this, SLOT(onAddAttrToolButtonClicked()));
  connect(m_ui->m_removeAttrToolButton, SIGNAL(clicked()), this, SLOT(onRemoveAttrToolButtonClicked()));
  connect(m_ui->m_checkToolButton, SIGNAL(clicked()), this, SLOT(onCheckToolButtonClicked()));
  connect(m_ui->m_okPushButton, SIGNAL(clicked()), this, SLOT(onOkPushButtonClicked()));

  //help information
  m_ui->m_helpPushButton->setPageReference("widgets/query/quickselect_dialog.html");
}

te::qt::widgets::QuickSelectDialog::~QuickSelectDialog() = default;

void te::qt::widgets::QuickSelectDialog::setInputDataSource(te::da::DataSourcePtr dataSource)
{
  m_dataSource = dataSource;

  if (!m_dataSource->isOpened())
    m_dataSource->open();

  //fill dataset list
  m_ui->m_datasetComboBox->clear();

  std::vector<std::string> dataSetNames;
  te::da::GetDataSetNames(dataSetNames, m_dataSource->getId());

  for (std::size_t t = 0; t < dataSetNames.size(); ++t)
    m_ui->m_datasetComboBox->addItem(dataSetNames[t].c_str());

  if (m_ui->m_datasetComboBox->count() > 0)
    onDataSetComboBoxSelected(0);
}


std::set<std::string> te::qt::widgets::QuickSelectDialog::getFunctionsNames()
{
  std::set<std::string> functions;

  functions.insert(te::da::FunctionNames::sm_Sum);
  functions.insert(te::da::FunctionNames::sm_Avg);
  functions.insert(te::da::FunctionNames::sm_Count);
  functions.insert(te::da::FunctionNames::sm_Max);
  functions.insert(te::da::FunctionNames::sm_Min);
  functions.insert(te::da::FunctionNames::sm_StdDev);
  functions.insert(te::da::FunctionNames::sm_Variance);
  functions.insert(te::da::FunctionNames::sm_IsNull);

  return functions;
}

void te::qt::widgets::QuickSelectDialog::onDataSetComboBoxSelected(int index)
{
  //clear interface
  m_ui->m_tableWidget->setRowCount(0);
  m_ui->m_tableWidget->resizeColumnsToContents();
  m_ui->m_resultTextEdit->clear();

  //create query builder widget
  std::string dataSetName = m_ui->m_datasetComboBox->currentText().toUtf8().data();
  std::unique_ptr<te::da::DataSetType> dsType = m_dataSource->getDataSetType(dataSetName);

  m_queryWidget->setInputParameters(m_dataSource, dataSetName, std::move(dsType));

  m_queryWidget->startWhereClause();

  m_ui->m_whereClauseGroupBox->setChecked(false);
}

void te::qt::widgets::QuickSelectDialog::onAddAttrToolButtonClicked()
{
  if (m_ui->m_datasetComboBox->count() == 0)
    return;

  //new entry
  int newrow = m_ui->m_tableWidget->rowCount();

  m_ui->m_tableWidget->insertRow(newrow);

  //function combo
  QComboBox* cmbFunction = new QComboBox(m_ui->m_tableWidget);
  m_ui->m_tableWidget->setCellWidget(newrow, 0, cmbFunction);

  std::set<std::string> functions = getFunctionsNames();

  std::set<std::string>::iterator it = functions.begin();

  cmbFunction->addItem("");

  while (it != functions.end())
  {
    cmbFunction->addItem(it->c_str());
    ++it;
  }

  //attribute combo
  QComboBox* cmbAttribute = new QComboBox(m_ui->m_tableWidget);
  m_ui->m_tableWidget->setCellWidget(newrow, 1, cmbAttribute);

  std::string dataSetName = m_ui->m_datasetComboBox->currentText().toUtf8().data();

  std::unique_ptr<te::da::DataSetType> dsType = m_dataSource->getDataSetType(dataSetName);

  std::vector<te::dt::Property*> propVec = dsType->getProperties();

  //cmbAttribute->addItem("*");

  for (std::size_t t = 0; t < propVec.size(); ++t)
  {
    cmbAttribute->addItem(propVec[t]->getName().c_str());
  }

  //attribute combo
  QComboBox* cmbOrder = new QComboBox(m_ui->m_tableWidget);
  m_ui->m_tableWidget->setCellWidget(newrow, 2, cmbOrder);

  cmbOrder->addItem("");
  cmbOrder->addItem("ASC", te::da::SortOrder::ASC);
  cmbOrder->addItem("DESC", te::da::SortOrder::DESC);

  //alias
  QTableWidgetItem* itemAlias = new QTableWidgetItem("");
  m_ui->m_tableWidget->setItem(newrow, 3, itemAlias);

#if (QT_VERSION >= 0x050000)
  m_ui->m_tableWidget->horizontalHeader()->setSectionResizeMode(0, QHeaderView::Stretch);
  m_ui->m_tableWidget->horizontalHeader()->setSectionResizeMode(1, QHeaderView::Stretch);
  m_ui->m_tableWidget->horizontalHeader()->setSectionResizeMode(2, QHeaderView::Stretch);
  m_ui->m_tableWidget->horizontalHeader()->setSectionResizeMode(3, QHeaderView::Stretch);
#else
  m_ui->m_tableWidget->horizontalHeader()->setResizeMode(QHeaderView::ResizeToContents);
#endif
}

void te::qt::widgets::QuickSelectDialog::onRemoveAttrToolButtonClicked()
{
  if (m_ui->m_tableWidget->currentRow() == -1)
    return;

  m_ui->m_tableWidget->removeRow(m_ui->m_tableWidget->currentRow());

#if (QT_VERSION >= 0x050000)
  m_ui->m_tableWidget->horizontalHeader()->setSectionResizeMode(0, QHeaderView::Stretch);
  m_ui->m_tableWidget->horizontalHeader()->setSectionResizeMode(1, QHeaderView::Stretch);
  m_ui->m_tableWidget->horizontalHeader()->setSectionResizeMode(2, QHeaderView::Stretch);
  m_ui->m_tableWidget->horizontalHeader()->setSectionResizeMode(3, QHeaderView::Stretch);
#else
  m_ui->m_tableWidget->horizontalHeader()->setResizeMode(QHeaderView::ResizeToContents);
#endif
}

void te::qt::widgets::QuickSelectDialog::onCheckToolButtonClicked()
{
  try
  {
    te::da::Select s = getSelect();

    m_dataSource->query(s);
  }
  catch (te::common::Exception& ex)
  {
    m_ui->m_resultTextEdit->setText(ex.what());

    return;
  }

  std::string selStr = getSelectString();

  m_ui->m_resultTextEdit->setText(selStr.c_str());

  return;
}

void te::qt::widgets::QuickSelectDialog::onOkPushButtonClicked()
{

}

te::da::Select te::qt::widgets::QuickSelectDialog::getSelect()
{
  int nRows = m_ui->m_tableWidget->rowCount();

  if (nRows == 0)
  {
    throw te::common::Exception("Undefined parameters.");
  }

  bool useGroupingBy = m_ui->m_groupingCheckBox->isChecked();

  te::da::Fields* fields = new te::da::Fields();

  te::da::From* from = new te::da::From();
  te::da::FromItem* fromItem = new te::da::DataSetName(m_ui->m_datasetComboBox->currentText().toUtf8().data());
  from->push_back(fromItem);

  te::da::OrderBy* orderBy = new te::da::OrderBy();

  te::da::GroupBy* groupBy = new te::da::GroupBy();

  for (int i = 0; i < nRows; ++i)
  {
    //get function
    QComboBox* cmbFunction = dynamic_cast<QComboBox*>(m_ui->m_tableWidget->cellWidget(i, 0));
    std::string function;
      
    if(cmbFunction->currentText().isEmpty() == false)
      function = cmbFunction->currentText().toUtf8().data();

    //get attribute
    QComboBox* cmbAttribute = dynamic_cast<QComboBox*>(m_ui->m_tableWidget->cellWidget(i, 1));
    std::string attribute;

    if (cmbAttribute->currentText().isEmpty() == false)
      attribute = cmbAttribute->currentText().toUtf8().data();

    //get alias
    QTableWidgetItem* item = m_ui->m_tableWidget->item(i, 3);
    std::string alias;
    
    if(item->text().isEmpty() == false)
      alias = item->text().toUtf8().data();

    if (alias.empty())
    {
      if (function.empty())
      {
        if (attribute == "*")
          alias = "all";
        else
          alias = attribute;
      }
      else
      {
        if (attribute != "*")
          alias = function + "_" + attribute;
        else
          alias = function + "_all";
      }

      item->setText(alias.c_str());
    }

    //get order by
    QComboBox* cmbOrder = dynamic_cast<QComboBox*>(m_ui->m_tableWidget->cellWidget(i, 2));
    std::string orderStrType;

    if(cmbOrder->currentText().isEmpty() == false)
      orderStrType = cmbOrder->currentText().toUtf8().data();

    if (function.empty())
    {
      te::da::Field* field = new te::da::Field(attribute, alias);
      fields->push_back(field);
    }
    else
    {
      te::da::UnaryFunction* func = new te::da::UnaryFunction(function, new te::da::PropertyName(attribute));

      te::da::Field* funcFiled = new te::da::Field(*func, alias);
      fields->push_back(funcFiled);
    }

    if (orderStrType.empty() == false)
    {
      int idx = cmbOrder->currentIndex();

      te::da::SortOrder orderType = static_cast<te::da::SortOrder>(cmbOrder->itemData(idx).toInt());
      te::da::OrderByItem* orderByItem = new te::da::OrderByItem(alias, orderType);
      orderBy->push_back(orderByItem);
    }

    if (useGroupingBy)
    {
      if (function.empty() && attribute != "*")
      {
        te::da::GroupByItem* groupByItem = new te::da::GroupByItem(alias);
        groupBy->push_back(groupByItem);
      }
      else if(attribute != "*")
      {
        te::da::GroupByItem* groupByItem = new te::da::GroupByItem(attribute);
        groupBy->push_back(groupByItem);
      }
    }
  }

  te::da::Select select;

  select.setFields(fields);
  select.setFrom(from);

  if (groupBy->empty() == false)
    select.setGroupBy(groupBy);

  if (orderBy->empty() == false)
    select.setOrderBy(orderBy);

  if (m_ui->m_whereClauseGroupBox->isChecked())
  {
    te::da::Where* w = m_queryWidget->getWhereClause();

    select.setWhere(w);
  }
   
  return select;
}

std::string te::qt::widgets::QuickSelectDialog::getSelectString()
{
  std::string sql;

  try
  {
    te::da::Select s = getSelect();

    te::da::SQLVisitor visitor(*m_dataSource->getDialect(), sql);

    s.accept(visitor);
  }
  catch (te::common::Exception& ex)
  {
    m_ui->m_resultTextEdit->setText(ex.what());

    return "";
  }

  return sql;
}

std::string te::qt::widgets::QuickSelectDialog::setAliasName(std::string value)
{
  if (value.empty() == false)
  {
    std::string dataSetName = value;
    std::string aliasName = value;

    size_t pos = dataSetName.find(".");

    if (pos != std::string::npos)
    {
      aliasName = dataSetName.substr(pos + 1, dataSetName.size() - 1);
    }

    return aliasName;
  }
  return "";
}
