/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.
  
    This file is part of the TerraLib - a Framework for building GIS enabled applications.
    
    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.
    
    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.
    
    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */


//Terralib
#include "../../../dataaccess/dataset/DataSet.h"
#include "../../../dataaccess/utils/Utils.h"
#include "../../widgets/utils/FileDialog.h"
#include "../../widgets/datasource/selector/DataSourceSelectorDialog.h"
#include "BoundaryExtractionWizardPage.h"
#include "ui/ui_BoundaryExtractionWizardPageForm.h"

// Qt
#include <QCheckBox>
#include <QFileDialog>
#include <QIntValidator>
#include <QMessageBox>

// BOOST
#include <boost/algorithm/string.hpp>
#include <boost/filesystem.hpp>
#include <boost/lexical_cast.hpp>

te::qt::widgets::BoundaryExtractionWizardPage::BoundaryExtractionWizardPage(QWidget *parent) :
  QWizardPage(parent)
{
  m_ui.reset(new Ui::BoundaryExtractionWizardPageForm);

  m_ui->setupUi(this);

//configure page
  this->setTitle(tr("Boundary Extraction"));
  this->setSubTitle(tr("Define the boundary extraction parameters."));

  m_ui->m_targetDatasourceToolButton->setIcon(QIcon::fromTheme("datasource"));

  connect(m_ui->m_targetDatasourceToolButton, SIGNAL(pressed()), this, SLOT(onTargetDatasourceToolButtonPressed()));
  connect(m_ui->m_targetFileToolButton, SIGNAL(pressed()), this,  SLOT(onTargetFileToolButtonPressed()));
}

te::qt::widgets::BoundaryExtractionWizardPage::~BoundaryExtractionWizardPage() = default;

bool te::qt::widgets::BoundaryExtractionWizardPage::isComplete() const
{
  return true;
}

void te::qt::widgets::BoundaryExtractionWizardPage::setLayer(te::map::AbstractLayerPtr layer)
{
  m_layer = layer;

  m_ui->m_layerLineEdit->setText(m_layer->getTitle().c_str());

  //list bands
  m_ui->m_bandComboBox->clear();

  //get input raster
  std::unique_ptr<te::da::DataSet> ds = m_layer->getData();

  if(ds.get())
  {
    std::size_t rpos = te::da::GetFirstPropertyPos(ds.get(), te::dt::RASTER_TYPE);

    std::unique_ptr<te::rst::Raster> inputRst = ds->getRaster(rpos);

    if(inputRst.get())
    {
      for(unsigned int i = 0; i < inputRst->getNumberOfBands(); ++i)
      {
        m_ui->m_bandComboBox->addItem(QString::number(i));
      }
    }
  }
}

std::unique_ptr<te::rst::Raster> te::qt::widgets::BoundaryExtractionWizardPage::getRaster()
{
  assert(m_layer.get());

   //get input raster
  std::unique_ptr<te::da::DataSet> ds = m_layer->getData();

  std::size_t rpos = te::da::GetFirstPropertyPos(ds.get(), te::dt::RASTER_TYPE);

  return ds->getRaster(rpos);
}

unsigned int te::qt::widgets::BoundaryExtractionWizardPage::getBand()
{
  return m_ui->m_bandComboBox->currentText().toUInt();
}

double te::qt::widgets::BoundaryExtractionWizardPage::getNoDataValue()
{
  return  m_ui->m_noDataValueLineEdit->text().toDouble();

}

bool te::qt::widgets::BoundaryExtractionWizardPage::outputDataSourceToFile()
{
  return m_toFile;
}

te::da::DataSourceInfoPtr te::qt::widgets::BoundaryExtractionWizardPage::getDataSourceInfo()
{
  return m_outputDatasource;
}

std::string te::qt::widgets::BoundaryExtractionWizardPage::getLayerName()
{
  if(m_ui->m_newLayerNameLineEdit->text().isEmpty())
    return "";

  return m_ui->m_newLayerNameLineEdit->text().toUtf8().data();
}

std::string te::qt::widgets::BoundaryExtractionWizardPage::getRepositoryName()
{
  if(m_ui->m_repositoryLineEdit->text().isEmpty())
    return "";

  return m_ui->m_repositoryLineEdit->text().toUtf8().data();
}

void te::qt::widgets::BoundaryExtractionWizardPage::onTargetDatasourceToolButtonPressed()
{
  m_ui->m_newLayerNameLineEdit->clear();
  m_ui->m_newLayerNameLineEdit->setEnabled(true);
  te::qt::widgets::DataSourceSelectorDialog dlg(this);
  dlg.exec();

  std::list<te::da::DataSourceInfoPtr> dsPtrList = dlg.getSelecteds();

  if(dsPtrList.size() <= 0)
    return;

  std::list<te::da::DataSourceInfoPtr>::iterator it = dsPtrList.begin();

  m_ui->m_repositoryLineEdit->setText(QString(it->get()->getTitle().c_str()));

  m_outputDatasource = *it;

  m_toFile = false;
}

void te::qt::widgets::BoundaryExtractionWizardPage::onTargetFileToolButtonPressed()
{
  m_ui->m_newLayerNameLineEdit->clear();
  m_ui->m_repositoryLineEdit->clear();

  te::qt::widgets::FileDialog fileDialog(this, te::qt::widgets::FileDialog::VECTOR);

  try
  {
    fileDialog.exec();
  }
  catch(te::common::Exception& ex)
  {
    QMessageBox::warning(this, tr("File information"), ex.what());
    return;
  }
  catch(...)
  {
    QMessageBox::warning(this, tr("File information"), tr("Output layer is invalid."));
    return;
  }

  m_ui->m_newLayerNameLineEdit->setText(fileDialog.getFileName().c_str());
  m_ui->m_repositoryLineEdit->setText(fileDialog.getPath().c_str());
  
  m_toFile = true;
  m_ui->m_newLayerNameLineEdit->setEnabled(false);
}


