
/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/widgets/rp/ClassifierWizard.cpp

  \brief A Qt dialog that allows users to run a classifier operation defined by RP module.
*/

// TerraLib
#include "../../../common/progress/ProgressManager.h"
#include "../../../color/ColorBar.h"
#include "../../../color/ColorScheme.h"
#include "../../../color/ColorSchemeCatalog.h"
#include "../../../color/ColorSchemeCatalogManager.h"
#include "../../../color/ColorSchemeGroup.h"
#include "../../../color/RGBAColor.h"
#include "../../../dataaccess/dataset/DataSet.h"
#include "../../../dataaccess/utils/Utils.h"
#include "../../../maptools/Utils.h"
#include "../../../maptools/GroupingAlgorithms.h"
#include "../../../raster/Grid.h"
#include "../../../raster/Raster.h"
#include "../../../raster/RasterSummary.h"
#include "../../../raster/RasterSummaryManager.h"
#include "../../../rp/Classifier.h"
#include "../../../rp/Module.h"
#include "../../../se/Categorize.h"
#include "../../../se/MapItem.h"
#include "../../../se/Recode.h"
#include "../../../se/ColorMap.h"
#include "../../../se/ParameterValue.h"
#include "../../../se/RasterSymbolizer.h"
#include "../../../se/Rule.h"
#include "../../../se/Utils.h"
#include "../help/HelpPushButton.h"
#include "../layer/search/LayerSearchWidget.h"
#include "../layer/search/LayerSearchWizardPage.h"
#include "../progress/ProgressViewerDialog.h"
#include "../raster/RasterInfoWidget.h"
#include "../utils/ScopedCursor.h"
#include "ClassifierWizard.h"
#include "ClassifierWizardPage.h"
#include "RasterInfoWizardPage.h"
#include "Utils.h"

// STL
#include <cassert>

// Qt
#include <QMessageBox>
#include <QApplication>
#include <QActionGroup>

te::qt::widgets::ClassifierWizard::ClassifierWizard(QWidget* parent)
  : QWizard(parent)
{
  //configure the wizard
  this->setWizardStyle(QWizard::ModernStyle);
  this->setWindowTitle(tr("Classifier"));
  //this->setFixedSize(640, 480);

  this->setOption(QWizard::HaveHelpButton, true);
  this->setOption(QWizard::HelpButtonOnRight, false);

  te::qt::widgets::HelpPushButton* helpButton = new te::qt::widgets::HelpPushButton(this);

  this->setButton(QWizard::HelpButton, helpButton);

  helpButton->setPageReference("plugins/rp/rp_classifier.html");

  connect(this, SIGNAL(currentIdChanged(int)), SLOT(onPageChanged(int)));

  addPages();
}

te::qt::widgets::ClassifierWizard::~ClassifierWizard()
{
  m_classifierPage->clearCanvas();
}

bool te::qt::widgets::ClassifierWizard::validateCurrentPage()
{
  if(currentPage() ==  m_layerSearchPage.get())
  {
    std::list<te::map::AbstractLayerPtr> list = m_layerSearchPage->getSearchWidget()->getSelecteds();

    if(list.empty() == false)
    {
      te::map::AbstractLayerPtr l = *list.begin();

      m_classifierPage->set(l);
    }

    return m_layerSearchPage->isComplete();
  }
  else if(currentPage() ==  m_classifierPage.get())
  {
    return m_classifierPage->isComplete();
  }
  else if(currentPage() ==  m_rasterInfoPage.get())
  {
    return execute();
  }

  return true;
}

void te::qt::widgets::ClassifierWizard::setList(std::list<te::map::AbstractLayerPtr>& layerList)
{
  m_layerSearchPage->getSearchWidget()->setList(layerList);
  m_layerSearchPage->getSearchWidget()->filterOnlyByRaster();

  m_classifierPage->setList(layerList);
}

void te::qt::widgets::ClassifierWizard::setMapDisplay(te::qt::widgets::MapDisplay* mapDisplay)
{
  m_classifierPage->setMapDisplay(mapDisplay);
}

void te::qt::widgets::ClassifierWizard::setActionGroup(QActionGroup* actionGroup)
{
  m_classifierPage->setActionGroup(actionGroup);
}

te::map::AbstractLayerPtr te::qt::widgets::ClassifierWizard::getOutputLayer()
{
  return m_outputLayer;
}

void te::qt::widgets::ClassifierWizard::onMapDisplayExtentChanged()
{
  m_classifierPage->onMapDisplayExtentChanged();
}

void te::qt::widgets::ClassifierWizard::addPages()
{
  m_layerSearchPage.reset(new te::qt::widgets::LayerSearchWizardPage(this));
  m_classifierPage.reset(new te::qt::widgets::ClassifierWizardPage(this));
  m_rasterInfoPage.reset(new te::qt::widgets::RasterInfoWizardPage(this));

  addPage(m_layerSearchPage.get());
  addPage(m_classifierPage.get());
  addPage(m_rasterInfoPage.get());

  //for contrast only one layer can be selected
  m_layerSearchPage->getSearchWidget()->enableMultiSelection(false);
}

bool te::qt::widgets::ClassifierWizard::execute()
{
  if(m_rasterInfoPage->getWidget()->fileExists())
  {
    QMessageBox::warning(this, tr("Classifier"), tr("File already exists."));
    return false;
  }

  //get layer
  std::list<te::map::AbstractLayerPtr> list = m_layerSearchPage->getSearchWidget()->getSelecteds();
  te::map::AbstractLayerPtr l = *list.begin();

  std::unique_ptr<te::rst::Raster> inputRst(te::map::GetRaster(l.get()));

  te::rp::Classifier algorithmInstance;

  te::rp::Classifier::InputParameters algoInputParams = m_classifierPage->getInputParams();

  algoInputParams.m_inputRasterPtr = inputRst.get();

  if( algoInputParams.m_createRasterPalette )
  {
    te::cl::ROISet* rs = m_classifierPage->getROISet();

    if( ( rs != 0 ) && ( ! rs->getROISet().empty() ) )
    {
      std::map<std::string, te::cl::ROI*>::const_iterator it = rs->getROISet().begin();
      std::map<std::string, te::cl::ROI*>::const_iterator itEnd = rs->getROISet().end();

      while( it != itEnd )
      {
        te::color::RGBAColor rgbColor( it->second->getColor() );

        te::rst::BandProperty::ColorEntry cEntry;
        cEntry.c1 = rgbColor.getRed();
        cEntry.c2 = rgbColor.getGreen();
        cEntry.c3 = rgbColor.getBlue();
        cEntry.c4 = rgbColor.getAlpha();

        algoInputParams.m_userOutputPalette.push_back( cEntry );

        ++it;
      }
    }
  }

  te::rp::Classifier::OutputParameters algoOutputParams = m_classifierPage->getOutputParams();
  algoOutputParams.m_rInfo = m_rasterInfoPage->getWidget()->getInfo();
  algoOutputParams.m_rType = m_rasterInfoPage->getWidget()->getType();

  try
  {
    {
      te::qt::widgets::ProgressViewerDialog v(this);

      te::qt::widgets::ScopedCursor cursor(Qt::WaitCursor);

      if (algorithmInstance.initialize(algoInputParams))
      {
        if (algorithmInstance.execute(algoOutputParams))
        {
          algoOutputParams.reset();

          //set output layer
          m_outputLayer = te::qt::widgets::createLayer(m_rasterInfoPage->getWidget()->getType(),
                                                     m_rasterInfoPage->getWidget()->getInfo());
        }
        else
        {
            std::string what = TE_TR("Classifier execution error.");
            what += " " + te::rp::Module::getLastLogStr();
            throw te::common::Exception(what);
        }
      }
      else
      {
          std::string what = TE_TR("Classifier execution error.");
          what += " " + te::rp::Module::getLastLogStr();
          throw te::common::Exception(what);
      }
    }

    //create legend

    if (
         m_classifierPage->getPallete()
         &&
         ( m_classifierPage->getROISet() != 0 )
         &&
         ( m_classifierPage->getROISet()->getROISet().empty() == false)
       )
    {
      te::cl::ROISet* rs = m_classifierPage->getROISet();

      te::se::ColorMap* cm = new te::se::ColorMap();

      te::se::Recode* r = new te::se::Recode();
      r->setFallbackValue("#000000");
      r->setLookupValue(new te::se::ParameterValue("Rasterdata"));

      std::map<std::string, te::cl::ROI*> roiSetMap = rs->getROISet();
      std::map<std::string, te::cl::ROI*>::const_iterator it = roiSetMap.begin();
      const std::map<std::string, te::cl::ROI*>::const_iterator itEnd = roiSetMap.end();
      unsigned int count = 0;

      while (it != itEnd )
      {
        const std::string color = it->second->getColor();
        const std::string label = it->second->getLabel();
        te::se::MapItem* m = new te::se::MapItem();

        m->setValue(new te::se::ParameterValue(color));
        m->setData((double)count);
        m->setTitle(label);

        r->add(m);

        ++count;
        ++it;
      }

      cm->setCategorize(nullptr);
      cm->setInterpolate(nullptr);
      cm->setRecode(r);

      te::se::RasterSymbolizer* rasterSymb = te::se::GetRasterSymbolizer(m_outputLayer->getStyle());
      rasterSymb->setColorMap(cm);
    }
  }
  catch (const std::exception& e)
  {
    QMessageBox::critical(this, tr("Classifier"), e.what());

    return false;
  }

  QMessageBox::information(this, tr("Classifier"), tr("Classifier ended sucessfully."));

  te::qt::widgets::applyRasterMultiResolution(tr("Classifier"), te::map::GetRaster(m_outputLayer.get()));

  emit addLayer(m_outputLayer);

  return true;
}

void te::qt::widgets::ClassifierWizard::buildLegend()
{
  //Calc min and max
  
  te::rst::RasterSummary rasterSummary;
  if( ! te::rst::RasterSummaryManager::getInstance().get(
    te::map::GetRaster(m_outputLayer.get()), false, rasterSummary ) )
  {
    throw te::common::Exception(TE_TR("Raster summary Error"));
  }      
  
  const std::complex<double>* cmin = rasterSummary.at(0).m_minVal;
  const std::complex<double>* cmax = rasterSummary.at(0).m_maxVal;
  double min = cmin->real();
  double max = cmax->real();

  std::vector<double> vec;
  vec.push_back(min);
  vec.push_back(max);

  //Get color
  std::vector<te::color::RGBAColor> colorVec;

  te::color::ColorSchemeCatalog* csc = te::color::ColorSchemeCatalogManager::getInstance().findByName("Default");

  if(!csc)
    return;

  std::vector<te::color::ColorSchemeGroup*> csgVec = csc->getColorSchemeGroups();

  for(std::size_t t = 0; t < csgVec.size(); ++t)
  {
    te::color::ColorSchemeGroup* csg = csgVec[t];

    if(csg->getName() == "Basic")
    {
      std::vector<te::color::ColorScheme*> csVec = csg->getColorSchemes();

      for(std::size_t p = 0; p < csVec.size(); ++p)
      {
        te::color::ColorScheme* cs = csVec[p];

        if(cs->getName() == "Red")
        {
          std::vector<te::color::RGBAColor>* colors = cs->getColors()[0];     //WARNING... Always getting the first position

          std::vector<te::color::RGBAColor>::iterator it = colors->begin();

          // create color bar
          te::color::ColorBar* cb = new te::color::ColorBar(*(colors->begin()), *(colors->end() - 1), 256);

          int count = 0;

          //fill color bar
          while(it != colors->end())
          {
            if(count != 0 && count != static_cast<int>(colors->size()) - 1)
            {
              double pos = (1. / (colors->size() - 1)) * count;

              cb->addColor(*it, pos);
            }

            ++count;
            ++it;
          }
          colorVec = cb->getSlices(2);
        }
      }
    }
  }

  //Get legend
  std::vector<te::se::Rule*> legVec;

  te::map::GroupingByEqualSteps("raster",vec.begin(), vec.end(), 2, legVec);

  te::se::Categorize* c = new te::se::Categorize();

  c->setFallbackValue("#000000");
  c->setLookupValue(new te::se::ParameterValue("Rasterdata"));

  QColor cWhite(Qt::white);
  std::string colorWhiteStr = cWhite.name().toUtf8().data();

  //added dummy color for values < than min values...
  c->addValue(new te::se::ParameterValue(colorWhiteStr));

  for(std::size_t i = 0; i < legVec.size(); ++i)
  {
    std::string rangeStrLower;
    std::string rangeStrUpper;

    te::fe::GetFilterStepValues(legVec[i]->getFilter(), rangeStrLower, rangeStrUpper);

    QColor color(colorVec[i].getRed(), colorVec[i].getGreen(), colorVec[i].getBlue(), colorVec[i].getAlpha());

    std::string colorStr = color.name().toUtf8().data();

    c->addThreshold(new te::se::ParameterValue(rangeStrLower));
    c->addValue(new te::se::ParameterValue(colorStr));

    if(legVec[i] == legVec.back())
    {
      c->addThreshold(new te::se::ParameterValue(rangeStrUpper));
    }
  }

  c->addValue(new te::se::ParameterValue(colorWhiteStr));

  c->setThresholdsBelongTo(te::se::Categorize::SUCCEEDING);

  te::se::ColorMap* cm = new te::se::ColorMap();

  cm->setCategorize(c);
  cm->setInterpolate(nullptr);
  cm->setRecode(nullptr);

  te::se::RasterSymbolizer* rasterSymb = te::se::GetRasterSymbolizer(m_outputLayer->getStyle());
  rasterSymb->setColorMap(cm);
}

void te::qt::widgets::ClassifierWizard::onPageChanged(int /*index*/)
{
  this->adjustSize();
}
