/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/widgets/rp/ColorTransformDialog.h

  \brief This file defines a class for a ColorTransform dialog.
*/

#ifndef __TERRALIB_QT_WIDGETS_RP_INTERNAL_COLORTRANSFORMDIALOG_H
#define __TERRALIB_QT_WIDGETS_RP_INTERNAL_COLORTRANSFORMDIALOG_H

// TerraLib
#ifndef Q_MOC_RUN
#include "../../../maptools/AbstractLayer.h"
#endif
#include "../Config.h"
#include "OutputRasterWidget.h"

// STL
#include <memory>

// Qt
#include <QComboBox>
#include <QDialog>


// Forward declaration
namespace Ui { class ColorTransformDialogForm; }

namespace te
{
  namespace qt
  {
    namespace widgets
    {
      //forward declaration
      class Canvas;
      class MapDisplay;

      /*!
        \class ColorTransformDialog

        \brief This class is GUI used to define the colorTransform parameters for the RP colorTransform operation.
      */
      class TEQTWIDGETSEXPORT ColorTransformDialog : public QDialog
      {
        Q_OBJECT

           enum ColorTransformTypes
          {
            COLORTRANSFORM_RGB2IHS,
            COLORTRANSFORM_IHS2RGB,
            COLORTRANSFORM_RGB2HLS,
            COLORTRANSFORM_HLS2RGB
          };

        public:

          ColorTransformDialog(QWidget* parent = 0);

          ~ColorTransformDialog();

        public:

          bool isRGB2IHS();

          bool isIHS2RGB();

          bool isRGB2HLS();

          bool isHLS2RGB();

          void setList(std::list<te::map::AbstractLayerPtr>& layerList);

          void setMapDisplay(te::qt::widgets::MapDisplay* mapDisplay);

        public slots:

          void colorTransformTypeComboBoxActivated(int index);

          void rgbRComboBoxActivated(int index);
          void rgbGComboBoxActivated(int index);
          void rgbBComboBoxActivated(int index);

          void ihsIComboBoxActivated(int index);
          void ihsHComboBoxActivated(int index);
          void ihsSComboBoxActivated(int index);

          void rgb2hlsRComboBoxActivated(int index);
          void rgb2hlsGComboBoxActivated(int index);
          void rgb2hlsBComboBoxActivated(int index);

          void hlsHComboBoxActivated(int index);
          void hlsLComboBoxActivated(int index);
          void hlsSComboBoxActivated(int index);

          void onNewROIPushButtonClicked();

          void onOkPushButtonClicked();

          void onEnvelopeAcquired(te::gm::Envelope env);
          void onMapDisplayExtentChanged();

          void onSearchLayerToolButtonClicked();

        signals:

          void geomAquired(te::gm::Polygon* poly);

          void addLayer(te::map::AbstractLayerPtr layer);

          void closeTool();

        protected:

          bool execute();

          bool prepareOutputLayer();

          void prepareOutputLayerName(std::map<std::string, unsigned int> rasterSelectedBands);

          te::rst::Raster* buildOutputRaster();

          te::rst::Raster* buildMemRaster();

          void fillColorTransformTypes();

          te::rst::Raster* getRRaster();

          te::rst::Raster* getGRaster();

          te::rst::Raster* getBRaster();

          te::rst::Raster* getIRaster();

          te::rst::Raster* getHRaster();

          te::rst::Raster* getSRaster();

          te::rst::Raster* getLRaster();
          
          int getNBits();

          void getRasterBands(QComboBox* layer, int index, QComboBox* band);

          std::map<std::string, unsigned int> getSelectedBands();

          void getRasterFromROI();

          void applyPreview();

          void applyPreviewRGB2IHS();

          void applyPreviewIHS2RGB();

          void applyPreviewRGB2HLS();

          void applyPreviewHLS2RGB();

          void drawPreview(te::rst::Raster* raster);

          void drawGeom();

          void clearCanvas();

          void closeEvent(QCloseEvent* e);
          
          void getRGBMinMax( double& min, double& max ) const;

        private:

          std::unique_ptr<Ui::ColorTransformDialogForm> m_ui;
          
          std::list<te::map::AbstractLayerPtr> m_layerList;

          te::map::AbstractLayerPtr m_layerR;
          te::map::AbstractLayerPtr m_layerG;
          te::map::AbstractLayerPtr m_layerB;
          te::rst::Raster* m_rasterPreview;
          te::rst::Raster* m_rasterRPreview;
          te::rst::Raster* m_rasterGPreview;
          te::rst::Raster* m_rasterBPreview;
          te::gm::Geometry* m_geom;
          te::qt::widgets::MapDisplay* m_mapDisplay;
          te::qt::widgets::OutputRasterWidget* m_outputWidget;
          std::vector< std::map<std::string, std::string> > m_outputRastersInfos;

      };

    } // end namespace widgets
  }   // end namespace qt
}     // end namespace te

#endif  // __TERRALIB_QT_WIDGETS_RP_INTERNAL_COLORTRANSFORMDIALOG_H
