/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/widgets/rp/ContrastDialogForm.cpp

  \brief This file defines a class for a Contrast dialog.
*/

// TerraLib
#include "../../af/ApplicationController.h"
#include "../../../raster/Grid.h"
#include "../../../common/progress/ProgressManager.h"
#include "../../../common/StringUtils.h"
#include "../../../core/translator/Translator.h"
#include "../../../dataaccess/dataset/DataSet.h"
#include "../../../dataaccess/utils/Utils.h"
#include "../../../geometry/GeometryProperty.h"
#include "../../../geometry/Utils.h"
#include "../../../memory/DataSet.h"
#include "../../../memory/DataSetItem.h"
#include "../../../raster/Raster.h"
#include "../../../raster/RasterProperty.h"
#include "../../../raster/RasterSummaryManager.h"
#include "../../../raster/Utils.h"
#include "../../../maptools/Utils.h"
#include "../../../rp/Module.h"
#include "../../../qt/widgets/rp/Utils.h"
#include "../../../qt/widgets/utils/ScopedCursor.h"
#include "../../../se/CoverageStyle.h"
#include "../../../se/RasterSymbolizer.h"
#include "../../../se/Rule.h"
#include "../../../se/SelectedChannel.h"
#include "../../../se/Symbolizer.h"
#include "../../../se/Utils.h"
#include "../../af/ApplicationController.h"
#include "../../af/events/ApplicationEvents.h"
#include "../../af/events/LayerEvents.h"
#include "../../widgets/tools/ExtentAcquire.h"
#include "../canvas/Canvas.h"
#include "../progress/ProgressViewerDialog.h"
#include "Utils.h"
#include "ContrastDialogForm.h"
#include "RasterHistogramWidget.h"
#include "RasterNavigatorWidget.h"
#include "OutputRasterWidget.h"
#include "ui/ui_ContrastDialogForm.h"

// Qt
#include <QGridLayout>
#include <QMessageBox>
#include <QCheckBox>

// Boost
#include <boost/lexical_cast.hpp>

//STL
#include <memory>

te::qt::widgets::ContrastDialogForm::ContrastDialogForm(QWidget* parent)
  : QDialog(parent),
  m_ui(new Ui::ContrastDialogForm),
  m_layer(nullptr),
  m_outputWidget(nullptr),
  m_mapDisplay(nullptr),
  m_updatePreview(false)
{
// setup controls
  m_ui->setupUi(this);

  fillContrastTypes();

  //build histograms
  
  QGridLayout* histogramLayout = new QGridLayout(m_ui->m_histWidget);
  m_histogramWidget.reset(new te::qt::widgets::RasterHistogramWidget(m_ui->m_histWidget));
  m_histogramWidget->setSizePolicy( QSizePolicy( QSizePolicy::Preferred, QSizePolicy::Preferred ) );
  m_histogramWidget->setBaseSize( 200, 100 );
  m_histogramWidget->setMinimumSize( 60, 60 );
  resetInputHistogram();
  histogramLayout->addWidget(m_histogramWidget.get());
  histogramLayout->setContentsMargins(0,0,0,0);
  histogramLayout->setSizeConstraint( QLayout::SetNoConstraint );
  
  QGridLayout* histogramLayout2 = new QGridLayout(m_ui->m_histWidget2);
  m_histogramWidget2.reset(new te::qt::widgets::RasterHistogramWidget(m_ui->m_histWidget2));
  m_histogramWidget2->setSizePolicy( QSizePolicy( QSizePolicy::Preferred, QSizePolicy::Preferred ) );
  m_histogramWidget2->setBaseSize( 200, 100 );
  m_histogramWidget2->setMinimumSize( 60, 60 );
  resetOutputHistogram();
  histogramLayout2->addWidget(m_histogramWidget2.get());
  histogramLayout2->setContentsMargins(0,0,0,0);
  histogramLayout2->setSizeConstraint( QLayout::SetNoConstraint );

  QGridLayout* displayLayout = new QGridLayout(m_ui->m_navigatorWidget);
  m_navigator.reset(new te::qt::widgets::RpToolsWidget(m_ui->m_navigatorWidget));
  displayLayout->addWidget(m_navigator.get());
  displayLayout->setContentsMargins(0, 0, 0, 0);

  m_ui->m_navigatorWidget->setEnabled(true);

  m_ui->m_resetToolButton->setIcon(QIcon::fromTheme("edit-undo"));
  m_ui->m_applyContrastToolButton->setIcon(QIcon::fromTheme("check"));
  m_ui->m_infoToolButton->setIcon(QIcon::fromTheme("help-about-browser"));
  m_ui->m_enableDisableHistoPushButton->setIcon(QIcon::fromTheme("histogram"));

  m_ui->m_resetToolButton->setEnabled(false);
  m_ui->m_applyContrastToolButton->setEnabled(false);
  m_ui->m_infoToolButton->setEnabled(false);

  //build input layer
  QGridLayout* inputLayout = new QGridLayout(m_ui->m_inputLayerWidget);
  inputLayout->setContentsMargins(0, 0, 0, 0);
  m_inputWidget = new te::qt::widgets::InputLayerWidget(m_ui->m_inputLayerWidget);
  inputLayout->addWidget(m_inputWidget);

  //build output parameters
  QGridLayout* outputLayout = new QGridLayout(m_ui->m_outputWidget);
  outputLayout->setContentsMargins(0, 0, 0, 0);
  m_outputWidget = new te::qt::widgets::OutputRasterWidget(m_ui->m_outputWidget);
  outputLayout->addWidget(m_outputWidget);

//connects
  connect(m_ui->m_contrastTypeComboBox, SIGNAL(activated(int)), this, SLOT(onContrastTypeComboBoxActivated(int)));
  connect(m_ui->m_bandTableWidget, SIGNAL(cellClicked(int, int)), this, SLOT(onCellClicked(int, int)));
  connect(m_ui->m_bandTableWidget, SIGNAL(cellDoubleClicked(int, int)), this, SLOT(onCellDoubleClicked(int, int)));
  connect(m_ui->m_bandTableWidget, SIGNAL(cellChanged(int, int)), this, SLOT(onCellChanged(int,int)));

  connect(m_ui->m_applyContrastToolButton, SIGNAL(clicked()), this, SLOT(onApplyPreviewClicked()));
  connect(m_ui->m_resetToolButton, SIGNAL(clicked()), this, SLOT(onResetValuesToolButtonClicked()));
  connect(m_ui->m_infoToolButton, SIGNAL(clicked()), this, SLOT(onInfoToolButtonClicked()));
  connect(m_ui->m_enableDisableHistoPushButton, SIGNAL(clicked(bool)), this, SLOT(onEableDisableHistoToolButtonClicked(bool)));

  connect(m_ui->m_allImageRadioButton, SIGNAL(toggled(bool)), this, SLOT(onAllImageRadioButtonToggled(bool)));
  connect(m_ui->m_visibleAreaRadioButton, SIGNAL(toggled(bool)), this, SLOT(onVisibleAreaRadioButtonToggled(bool)));
  connect(m_ui->m_roiRadioButton, SIGNAL(toggled(bool)), this, SLOT(onRoiRadioButtonToggled(bool)));
  connect(m_ui->m_okPushButton, SIGNAL(clicked()), this, SLOT(onOkPushButtonClicked()));
  connect(m_inputWidget, SIGNAL(layerChanged()), this, SLOT(layerChanged()));
  connect(m_navigator.get(), SIGNAL(envelopeAcquired(te::gm::Envelope)), this, SLOT(onEnvelopeAcquired(te::gm::Envelope)));

  m_ui->m_helpPushButton->setNameSpace("dpi.inpe.br.plugins");
  m_ui->m_helpPushButton->setPageReference("plugins/rp/rp_contrast.html");

  m_ui->m_sampledCheckBox->hide();
  m_ui->m_sampledCheckBox->setChecked(false);

  resize( 630, 730 );
}

te::qt::widgets::ContrastDialogForm::~ContrastDialogForm() = default;

void te::qt::widgets::ContrastDialogForm::setList(std::list<te::map::AbstractLayerPtr>& layerList)
{
  std::list<te::map::AbstractLayerPtr> layersRasterList;

  std::list<te::map::AbstractLayerPtr>::iterator it = layerList.begin();
  //Filter only raster layers
  while (it != layerList.end())
  {
    if(it->get()->getVisibility() == te::map::VISIBLE)
    {
      std::unique_ptr<te::da::DataSetType> dsType = it->get()->getSchema();

      if (dsType->hasRaster())
      {
        te::rst::RasterProperty* rasterProp = te::da::GetFirstRasterProperty(dsType.get());

        if (!rasterProp->getBandProperties().empty())
        {
          if(!(rasterProp->getBandProperties()[0]->m_colorInterp == te::rst::PaletteIdxCInt))
            layersRasterList.push_back(it->get());
        }
      }
    }

    ++it;
  }

  if(layersRasterList.empty())
  {
    throw te::common::Exception(TE_TR("There are no valid raster layers available for the contrast operation."));
    return;
  }

  m_inputWidget->setLayerList(layersRasterList);

  m_layer = layersRasterList.begin()->get();

  std::unique_ptr<te::rst::Raster> raster(te::map::GetRaster(m_layer.get()));

  if (raster->getBand(0)->getProperty()->getType() == te::dt::UCHAR_TYPE || 
      raster->getBand(0)->getProperty()->getType() == te::dt::INT16_TYPE ||
      raster->getBand(0)->getProperty()->getType() == te::dt::INT32_TYPE ||
      raster->getBand(0)->getProperty()->getType() == te::dt::INT64_TYPE ||
      raster->getBand(0)->getProperty()->getType() == te::dt::UINT16_TYPE ||
      raster->getBand(0)->getProperty()->getType() == te::dt::UINT32_TYPE || 
      raster->getBand(0)->getProperty()->getType() == te::dt::UINT64_TYPE)
  {
    double outRangeMin, outRangeMax;
    te::rst::GetDataTypeRanges(raster->getBandDataType(0), outRangeMin, outRangeMax);

    m_ui->m_outRangeMinLineEdit->setText(te::common::Convert2String(outRangeMin, 2).c_str());
    m_ui->m_outRangeMaxLineEdit->setText(te::common::Convert2String(outRangeMax, 2).c_str());

    connect(m_histogramWidget.get(), SIGNAL(minValueSelected(int, int)), this, SLOT(onMinValueSelected(int, int)));
    connect(m_histogramWidget.get(), SIGNAL(maxValueSelected(int, int)), this, SLOT(onMaxValueSelected(int, int)));
  }
  else
  {
    double vmin = std::numeric_limits<double>::max();
    double vmax = std::numeric_limits<double>::min();
    
    te::rst::RasterSummary rasterSummary;
    if( ! te::rst::RasterSummaryManager::getInstance().get(
      raster.get(), false, rasterSummary ) )
    {
      throw te::common::Exception(TE_TR("Raster summary Error"));
    }        

    for (std::size_t b = 0; b < raster->getNumberOfBands(); ++b)
    {
      const std::complex<double>* cmin = rasterSummary.at(b).m_minVal;
      const std::complex<double>* cmax = rasterSummary.at(b).m_maxVal;
      double min = cmin->real();
      double max = cmax->real();

      vmin = std::min(vmin, min);
      vmax = std::max(vmax, max);
    }

    double range = vmax - vmin;
    unsigned int nbits = (unsigned int)(std::ceil(log(range) / log(2.)));
    double maxValue = (double)(pow(2.0, (double)(nbits <= 8) ? 8 : nbits)) - 1;

    m_ui->m_outRangeMinLineEdit->setText("0.00");
    m_ui->m_outRangeMaxLineEdit->setText(te::common::Convert2String(maxValue, 2).c_str());

    connect(m_histogramWidget.get(), SIGNAL(minValueSelected(double, int)), this, SLOT(onMinValueSelected(double, int)));
    connect(m_histogramWidget.get(), SIGNAL(maxValueSelected(double, int)), this, SLOT(onMaxValueSelected(double, int)));
  }

  m_mapDisplay->setLayerList(layerList);

  te::gm::Envelope e = m_mapDisplay->getExtent();
  m_mapDisplay->setExtent(e, true);

  m_histogramWidget->setEnableNavigationTools(true);
}

void te::qt::widgets::ContrastDialogForm::setMapDisplay(te::qt::widgets::MapDisplay* mapDisplay)
{
  m_navigator->setMapDisplay(mapDisplay);
  m_mapDisplay = mapDisplay;
}

void te::qt::widgets::ContrastDialogForm::setActionGroup(QActionGroup* actionGroup)
{
  //configure tools
  m_navigator->setActionGroup(actionGroup);
  m_navigator->enableBoxAction();
}

te::map::AbstractLayerPtr te::qt::widgets::ContrastDialogForm::getOutputLayer()
{
  return m_outputLayer;
}

te::map::AbstractLayerPtr te::qt::widgets::ContrastDialogForm::get()
{
  return m_layer;
}

te::rp::Contrast::InputParameters te::qt::widgets::ContrastDialogForm::getInputParams()
{
  //get contrast type
  int index = m_ui->m_contrastTypeComboBox->currentIndex();
  int contrastType = m_ui->m_contrastTypeComboBox->itemData(index).toInt();
  
  std::vector< unsigned int > selectedBandIndexes;
    getCheckedInputBandIndexes( selectedBandIndexes );
  
  int nBands = m_ui->m_bandTableWidget->rowCount();

  te::rp::Contrast::InputParameters algoInputParams;
  
    getCheckedInputBandIndexes( algoInputParams.m_inRasterBands );  

  if(contrastType == te::rp::Contrast::InputParameters::LinearContrastT)
  {
    algoInputParams.m_type = te::rp::Contrast::InputParameters::LinearContrastT;

    for(int i = 0; i < nBands; ++i)
    {
      QCheckBox* bandCheckBox = (QCheckBox*)m_ui->m_bandTableWidget->cellWidget(i, 0);

      if(bandCheckBox->isChecked())
      {
        QTableWidgetItem* item = m_ui->m_bandTableWidget->item(i, 2);
        QTableWidgetItem* item1 = m_ui->m_bandTableWidget->item(i, 3);

        if (item)
        {
          QString valueMin = item->text();
          algoInputParams.m_lCMinInput.push_back(valueMin.toDouble());
        }

        if (item1)
        {
          QString valueMax = item1->text();
          algoInputParams.m_lCMaxInput.push_back(valueMax.toDouble());
        }
      }
    }
  }
  else if(contrastType == te::rp::Contrast::InputParameters::HistogramEqualizationContrastT)
  {
    algoInputParams.m_type = te::rp::Contrast::InputParameters::HistogramEqualizationContrastT;

    for(int i = 0; i < nBands; ++i)
    {
      QCheckBox* bandCheckBox = (QCheckBox*)m_ui->m_bandTableWidget->cellWidget(i, 0);

      if(bandCheckBox->isChecked())
      {
        QTableWidgetItem* item = m_ui->m_bandTableWidget->item(i, 2);

        if (item)
        {
          QString valueMax = item->text();
          algoInputParams.m_hECMaxInput.push_back(valueMax.toDouble());
        }
      }
    }
  }
  else if(contrastType == te::rp::Contrast::InputParameters::MeanAndStdContrastT)
  {
    algoInputParams.m_type = te::rp::Contrast::InputParameters::MeanAndStdContrastT;

    for(int i = 0; i < nBands; ++i)
    {
      QCheckBox* bandCheckBox = (QCheckBox*)m_ui->m_bandTableWidget->cellWidget(i, 0);

      if(bandCheckBox->isChecked())
      {
        QTableWidgetItem* item = m_ui->m_bandTableWidget->item(i, 2);
        QTableWidgetItem* item1 = m_ui->m_bandTableWidget->item(i, 3);

        if (item)
        {
          QString valueMean = item->text();
          algoInputParams.m_sMASCMeanInput.push_back(valueMean.toDouble());
        }

        if (item1)
        {
          QString valueStdDev = item1->text();
          algoInputParams.m_sMASCStdInput.push_back(valueStdDev.toDouble());
        }
      }
    }
  }
  else if(contrastType == te::rp::Contrast::InputParameters::SquareContrastT)
  {
    algoInputParams.m_type = te::rp::Contrast::InputParameters::SquareContrastT;

    for(int i = 0; i < nBands; ++i)
    {
      QCheckBox* bandCheckBox = (QCheckBox*)m_ui->m_bandTableWidget->cellWidget(i, 0);

      if(bandCheckBox->isChecked())
      {
        QTableWidgetItem* item = m_ui->m_bandTableWidget->item(i, 2);
        QTableWidgetItem* item1 = m_ui->m_bandTableWidget->item(i, 3);

        if (item)
        {
          QString valueMin = item->text();
          algoInputParams.m_squareCMinInput.push_back(valueMin.toDouble());
        }

        if (item1)
        {
          QString valueMax = item1->text();
          algoInputParams.m_squareCMaxInput.push_back(valueMax.toDouble());
        }
      }
    }
  }
  else if(contrastType == te::rp::Contrast::InputParameters::SquareRootContrastT)
  {
    algoInputParams.m_type = te::rp::Contrast::InputParameters::SquareRootContrastT;

    for(int i = 0; i < nBands; ++i)
    {
      QCheckBox* bandCheckBox = (QCheckBox*)m_ui->m_bandTableWidget->cellWidget(i, 0);

      if(bandCheckBox->isChecked())
      {
        QTableWidgetItem* item = m_ui->m_bandTableWidget->item(i, 2);
        QTableWidgetItem* item1 = m_ui->m_bandTableWidget->item(i, 3);

        if (item)
        {
          QString valueMin = item->text();
          algoInputParams.m_squareRootCMinInput.push_back(valueMin.toDouble());
        }

        if (item1)
        {
          QString valueMax = item1->text();
          algoInputParams.m_squareRootCMaxInput.push_back(valueMax.toDouble());
        }
      }
    }
  }
  else if(contrastType == te::rp::Contrast::InputParameters::LogContrastT)
  {
    algoInputParams.m_type = te::rp::Contrast::InputParameters::LogContrastT;

    for(int i = 0; i < nBands; ++i)
    {
      QCheckBox* bandCheckBox = (QCheckBox*)m_ui->m_bandTableWidget->cellWidget(i, 0);

      if(bandCheckBox->isChecked())
      {
        QTableWidgetItem* item = m_ui->m_bandTableWidget->item(i, 2);

        if (item)
        {
          QString valueMin = item->text();
          algoInputParams.m_logCMinInput.push_back(valueMin.toDouble());
        }

        QTableWidgetItem* item1 = m_ui->m_bandTableWidget->item(i, 3);

        if (item1)
        {
          QString valueMax = item1->text();
          algoInputParams.m_logCMaxInput.push_back(valueMax.toDouble());
        }
      }
    }
  }
  else if(contrastType == te::rp::Contrast::InputParameters::DecorrelationEnhancementT)
  {
    algoInputParams.m_type = te::rp::Contrast::InputParameters::DecorrelationEnhancementT;
  }
  else if (contrastType == te::rp::Contrast::InputParameters::CumulativeCut2EnhancementT)
  {
    algoInputParams.m_type = te::rp::Contrast::InputParameters::CumulativeCut2EnhancementT;
  }
  else if(contrastType == te::rp::Contrast::InputParameters::PercentCutUDLinearT)
  {
    algoInputParams.m_type = te::rp::Contrast::InputParameters::PercentCutUDLinearT;
    algoInputParams.m_cCutUDCSelecWideRange = true;

    for(int i = 0; i < nBands; ++i)
    {
      QCheckBox* bandCheckBox = (QCheckBox*)m_ui->m_bandTableWidget->cellWidget(i, 0);

      if(bandCheckBox->isChecked())
      {
        QTableWidgetItem* item = m_ui->m_bandTableWidget->item(i, 2);
        QTableWidgetItem* item1 = m_ui->m_bandTableWidget->item(i, 3);

        if (item)
        {
          QString valueMean = item->text();
          algoInputParams.m_cCutUDCMinInput.push_back(valueMean.toDouble());
        }

        if (item1)
        {
          QString valueStdDev = item1->text();
          algoInputParams.m_cCutUDCMaxInput.push_back(valueStdDev.toDouble());
        }
      }
    }
  }  

  algoInputParams.m_sampled = m_ui->m_sampledCheckBox->isChecked();

  return algoInputParams;
}

void te::qt::widgets::ContrastDialogForm::onAllImageRadioButtonToggled(bool isChecked)
{
  if(isChecked)
  {
    clearCanvas();

    resetInputHistogram();
    resetOutputHistogram();

    if(!m_mapDisplay->getExtent().isValid())
      return;

    if(m_layer)
    {
      m_raster.reset(te::map::GetRaster(m_layer.get()));

      m_histogramWidget->setInputRaster(m_raster.get());

      m_ui->m_resetToolButton->setEnabled(true);
      m_ui->m_applyContrastToolButton->setEnabled(true);
    }

    drawInputHistogram();

    onContrastTypeComboBoxActivated(m_ui->m_contrastTypeComboBox->currentIndex());
  }
}

void te::qt::widgets::ContrastDialogForm::fillContrastTypes()
{
  m_ui->m_contrastTypeComboBox->clear();

  m_ui->m_contrastTypeComboBox->addItem(tr("Linear"), te::rp::Contrast::InputParameters::LinearContrastT);
  m_ui->m_contrastTypeComboBox->addItem(tr("Histogram Equalization"), te::rp::Contrast::InputParameters::HistogramEqualizationContrastT);
  m_ui->m_contrastTypeComboBox->addItem(tr("Square Constrast"), te::rp::Contrast::InputParameters::SquareContrastT);
  m_ui->m_contrastTypeComboBox->addItem(tr("Square Root Constrast"), te::rp::Contrast::InputParameters::SquareRootContrastT);
  m_ui->m_contrastTypeComboBox->addItem(tr("Log Constrast"), te::rp::Contrast::InputParameters::LogContrastT);
  m_ui->m_contrastTypeComboBox->addItem(tr("Mean and Standard Deviation"), te::rp::Contrast::InputParameters::MeanAndStdContrastT);
  m_ui->m_contrastTypeComboBox->addItem(tr("Decorrelation Enhancement"), te::rp::Contrast::InputParameters::DecorrelationEnhancementT);
  m_ui->m_contrastTypeComboBox->addItem(tr("Cumulative 2% Enhancement"), te::rp::Contrast::InputParameters::CumulativeCut2EnhancementT);
  m_ui->m_contrastTypeComboBox->addItem(tr("Linear Percentual Cut"), te::rp::Contrast::InputParameters::PercentCutUDLinearT);
}

void te::qt::widgets::ContrastDialogForm::listBands()
{
  assert(m_layer.get());

  if(m_layer)
  {
    std::unique_ptr<te::rst::Raster> inputRst(te::map::GetRaster(m_layer.get()));

    if(inputRst.get())
    {
      m_ui->m_bandTableWidget->setRowCount((int)inputRst->getNumberOfBands());

      for (std::size_t b = 0; b < inputRst->getNumberOfBands(); ++b)
      {
        QCheckBox* bandCheckBox = new QCheckBox(this);
        bandCheckBox->setChecked(true);

        connect(bandCheckBox, SIGNAL(stateChanged(int)), this, SLOT(onBandChecked()));

        m_ui->m_bandTableWidget->setCellWidget(static_cast<int>(b), 0, bandCheckBox);

        QTableWidgetItem* band = new QTableWidgetItem(te::common::Convert2String(static_cast<int>(b)).c_str());
        m_ui->m_bandTableWidget->setItem(static_cast<int>(b), 1, band);
      }
    }
  }

  m_ui->m_bandTableWidget->resizeColumnsToContents();

  m_ui->m_bandTableWidget->selectRow(0);

#if (QT_VERSION >= 0x050000)
  m_ui->m_bandTableWidget->horizontalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
#else
  m_ui->m_bandTableWidget->horizontalHeader()->setResizeMode(QHeaderView::ResizeToContents);
#endif
}

void te::qt::widgets::ContrastDialogForm::applyPreview()
{
  if (!m_raster.get())
    return;

  if(m_ui->m_bandTableWidget->rowCount() == 0)
  {
    QMessageBox::warning(this, tr("Warning"), tr("Draw histogram before to apply contrast preview."));
    return;
  }

  try
  {
    //set contrast parameters
    te::rp::Contrast::InputParameters algoInputParams = getInputParams();

    algoInputParams.m_inRasterPtr = m_raster.get();

    double minValue = m_ui->m_outRangeMinLineEdit->text().toDouble();
    double maxValue = m_ui->m_outRangeMaxLineEdit->text().toDouble();

    algoInputParams.m_outRangeMin = minValue;
    algoInputParams.m_outRangeMax = maxValue;

    te::rp::Contrast::OutputParameters algoOutputParams;

    std::map<std::string, std::string> rinfo;
    rinfo["MAXMEMPERCENTUSED"] = "50";

    algoOutputParams.m_createdOutRasterDSType = "EXPANSIBLE";
    algoOutputParams.m_createdOutRasterInfo = rinfo;

    //run contrast
    te::rp::Contrast algorithmInstance;

    if (algorithmInstance.initialize(algoInputParams))
    {
      bool execResult = false;
      {
        te::qt::widgets::ScopedCursor cursor(Qt::WaitCursor);
        execResult = algorithmInstance.execute(algoOutputParams);
      }
      
      if( execResult )
      {
        m_previewRasterPtr.reset( algoOutputParams.m_createdOutRasterPtr.release() );
        
        drawPreviewRaster();
        
        if (algoInputParams.m_type == te::rp::Contrast::InputParameters::CumulativeCut2EnhancementT)
        {
          for (unsigned int i = 0; i < m_previewRasterPtr->getNumberOfBands(); ++i)
          {
            m_ui->m_bandTableWidget->item(i, 2)->setText(QString::number(
              m_previewRasterPtr->getBand(i)->getMinValue().real()));
            m_ui->m_bandTableWidget->item(i, 3)->setText(QString::number(
              m_previewRasterPtr->getBand(i)->getMaxValue().real()));
          }
        }        
        
        drawOutputHistogram();
      }
      else
      {
        QMessageBox::warning(this, tr("Warning"), tr("Contrast error :") +
          QString( algorithmInstance.getErrorMessage().c_str() ) );
        return;
      }
    }
    else
    {
      QMessageBox::warning(this, tr("Warning"), tr("Contrast error :") +
        QString( algorithmInstance.getErrorMessage().c_str() ) );
      return;
    }
  }
  catch (std::exception& e )
  {
    QMessageBox::warning(this, tr("Warning"), tr("Contrast error ") +
      e.what() );
    return;
  }
}

void te::qt::widgets::ContrastDialogForm::getRasterFromROI()
{
  if (m_geom->getGeomTypeId() == te::gm::PolygonType)
  {
    try
    {
      std::unique_ptr<te::gm::Envelope> env(new te::gm::Envelope(*m_geom->getMBR()));

      std::unique_ptr<te::rst::Raster> raster(te::map::GetRaster(m_layer.get()));

      std::map<std::string, std::string> info;
      info["RTYPE"] = "EXPANSIBLE";
      info["MAXMEMPERCENTUSED"] = "50";

      //execute clipping
      
      {
        te::qt::widgets::ScopedCursor cursor(Qt::WaitCursor);
      
        m_raster.reset(raster->trim(env.get(), info));
      }

      resetInputHistogram();
      m_histogramWidget->setInputRaster(m_raster.get());
    }
    catch (std::exception& e)
    {
      QMessageBox::warning(this, tr("Warning"), tr("Constrast error.") + e.what());
      return;
    }
  }
}

bool te::qt::widgets::ContrastDialogForm::execute()
{
  if (m_outputWidget->fileExists())
  {
    QMessageBox::warning(this, tr("Contrast"), tr("File already exists."));
    return false;
  }

  if(m_ui->m_roiRadioButton->isChecked() || m_ui->m_visibleAreaRadioButton->isChecked())
  {
    if(!m_geom.get() || !m_geom->isValid())
    {
      QMessageBox::warning(this, tr("Contrast"), tr("The region of interest informed is invalid."));
      return false;
    }
  }

  if(!m_raster)
  {
    QMessageBox::warning(this, tr("Contrast"), tr("Histogram area and Contrast preview are invalid."));
    return false;
  }

  if(m_minValue.empty() || m_maxValue.empty())
     return false;

  std::unique_ptr<te::rst::Raster> inputRst(te::map::GetRaster(m_layer.get()));

  //run contrast
  te::rp::Contrast algorithmInstance;

  te::rp::Contrast::InputParameters algoInputParams = getInputParams();
  algoInputParams.m_inRasterPtr = inputRst.get();
  algoInputParams.m_enableProgress = true;

  double minValue = m_ui->m_outRangeMinLineEdit->text().toDouble();
  double maxValue = m_ui->m_outRangeMaxLineEdit->text().toDouble();

  algoInputParams.m_outRangeMin = minValue;
  algoInputParams.m_outRangeMax = maxValue;

  te::rp::Contrast::OutputParameters algoOutputParams;
  algoOutputParams.m_createdOutRasterDSType = m_outputWidget->getType();
  algoOutputParams.m_createdOutRasterInfo = m_outputWidget->getInfo();
  algoOutputParams.m_outRasterBands = algoInputParams.m_inRasterBands;


  try
  {
    if (algorithmInstance.initialize(algoInputParams))
    {
      bool execResult = false;
      {
        te::qt::widgets::ScopedCursor cursor(Qt::WaitCursor);
        execResult = algorithmInstance.execute(algoOutputParams);
      }
      
      if( execResult )
      {
        algoOutputParams.reset();

        //set output layer
        m_outputLayer = te::qt::widgets::createLayer(m_outputWidget->getType(),
          m_outputWidget->getInfo());

      }
      else
      {
        std::string what = TE_TR("Contrast enhencement execution error.");
        what += " " + algorithmInstance.getErrorMessage();
        throw te::common::Exception(what);
      }
    }
    else
    {
      std::string what = TE_TR("Contrast enhencement initialization error.");
      what += " " + algorithmInstance.getErrorMessage();
      throw te::common::Exception(what);
    }
  }
  catch (const std::exception& e)
  {
    QMessageBox::warning(this, tr("Contrast"), e.what());
    return false;
  }

  QMessageBox::information(this, tr("Contrast"), tr("Contrast enhencement ended sucessfully"));

  te::qt::widgets::applyRasterMultiResolution(tr("Contrast"), te::map::GetRaster(m_outputLayer.get()));

  emit addLayer(m_outputLayer);

  return true;
}

void te::qt::widgets::ContrastDialogForm::setDisableInterface(bool enabled)
{
  m_ui->m_typeGroupBox->setEnabled(enabled);
  m_ui->m_bandsGroupBox->setEnabled(enabled);
  m_ui->m_histogramGroupBox->setEnabled(enabled);
  m_ui->m_histWidget->setEnabled(enabled);
  m_ui->m_histAreaGroupBox->setEnabled(enabled);
  m_ui->m_outputWidget->setEnabled(enabled);

  resetInputHistogram();
  resetOutputHistogram();

  m_ui->m_bandTableWidget->setRowCount(0);

  if(!enabled)
    m_layer = nullptr;
}

void te::qt::widgets::ContrastDialogForm::clearCanvas()
{
  te::qt::widgets::Canvas canvasInstance(m_mapDisplay->getDraftPixmap());

  canvasInstance.clear();

  m_geom.reset(nullptr);

  m_mapDisplay->repaint();
}

te::se::ChannelSelection* te::qt::widgets::ContrastDialogForm::getChannelSelection()
{
  te::se::Style* style = m_layer->getStyle();
  assert(style);

  te::se::CoverageStyle* cs = dynamic_cast<te::se::CoverageStyle*>(style);
  assert(cs);

  // for while, consider one rule
  const te::se::Rule* rule = cs->getRule(0);

  const std::vector<te::se::Symbolizer*>& symbolizers = rule->getSymbolizers();
  assert(!symbolizers.empty());

  // for while, consider one raster symbolizer
  te::se::RasterSymbolizer* symbolizer = dynamic_cast<te::se::RasterSymbolizer*>(symbolizers[0]);
  assert(symbolizer);

  return symbolizer->getChannelSelection();
}

void te::qt::widgets::ContrastDialogForm::closeEvent(QCloseEvent*)
{
  clearCanvas();

  emit closeTool();
}

void te::qt::widgets::ContrastDialogForm::resetWindow(te::map::AbstractLayerPtr layer)
{
  std::unique_ptr<te::da::DataSetType> dsType = layer->getSchema();
  //Checking if the layer contains a raster property
  if (!dsType->hasRaster())
  {
    QMessageBox::warning(this, tr("Warning"), tr("There is no selected raster layer."));

    setDisableInterface(false);

    return;
  }

  te::rst::RasterProperty* rasterProp = te::da::GetFirstRasterProperty(dsType.get());

  if (rasterProp->getBandProperties().empty())
  {
    QMessageBox::warning(this, tr("Warning"), tr("Layers without Band information!"));

    setDisableInterface(false);

    return;
  }

  if (rasterProp->getBandProperties()[0]->m_colorInterp == te::rst::PaletteIdxCInt)
  {
    QMessageBox::warning(this, tr("Warning"), tr("Layers with raster palette are not able to apply contrast!"));

    setDisableInterface(false);

    return;
  }

  setDisableInterface(true);

  clearCanvas();

  m_ui->m_visibleAreaRadioButton->setChecked(true);

  std::list<te::map::AbstractLayerPtr> layerList;
  layerList.push_back(layer);

  setList(layerList);

  te::gm::Envelope displayExtent = te::map::GetExtent(layerList, m_mapDisplay->getSRID(), true);

  m_mapDisplay->setExtent(displayExtent, false);

  m_mapDisplay->setLayerList(layerList);

  m_mapDisplay->refresh(true);
}

void te::qt::widgets::ContrastDialogForm::drawPreviewRaster()
{
  if( m_previewRasterPtr.get() != nullptr )
  {
    m_ui->m_bandTableWidget->setEnabled(false);

    const te::gm::Envelope& envRaster = *m_previewRasterPtr->getExtent();
    const te::gm::Envelope& env = m_mapDisplay->getExtent();

    std::unique_ptr< te::se::Style > style( te::se::CreateCoverageStyle(
      m_previewRasterPtr->getNumberOfBands()) );

    te::se::CoverageStyle* cs = dynamic_cast<te::se::CoverageStyle*>(style.get());
    assert(cs);

    std::unique_ptr< te::qt::widgets::Canvas > canvasInstance( 
      new te::qt::widgets::Canvas(m_mapDisplay->width(), m_mapDisplay->height()) );
    canvasInstance->setDevice(m_mapDisplay->getDraftPixmap(), false);
    canvasInstance->setWindow(env.m_llx, env.m_lly, env.m_urx, env.m_ury);

    canvasInstance->setPolygonContourWidth(2);
    canvasInstance->setPolygonContourColor(te::color::RGBAColor(100, 177, 216, TE_OPAQUE));
    canvasInstance->setPolygonFillColor(te::color::RGBAColor(100, 177, 216, 0));

    // Draw raster
    bool cancel = false;
    {
      te::qt::widgets::ScopedCursor cursor(Qt::WaitCursor);
      te::map::DrawRaster(m_previewRasterPtr.get(), canvasInstance.get(), envRaster, m_previewRasterPtr->getSRID(), 
        env, m_mapDisplay->getSRID(), cs, nullptr, m_mapDisplay->getScale(), &cancel);
    }

    if(m_ui->m_roiRadioButton->isChecked() || m_ui->m_visibleAreaRadioButton->isChecked())
    {
      if (!m_geom.get() || !m_geom->isValid())
        return;

      if(m_geom->getSRID() != m_mapDisplay->getSRID())
        m_geom->transform(m_mapDisplay->getSRID());

      canvasInstance->draw(m_geom.get());
    }

    m_ui->m_bandTableWidget->setEnabled(true);

    m_mapDisplay->repaint();
  }
}

void te::qt::widgets::ContrastDialogForm::drawGeom()
{
  if (!m_geom.get() || !m_geom->isValid())
    return;

  const te::gm::Envelope& env = m_mapDisplay->getExtent();

  te::qt::widgets::Canvas canvasInstance(m_mapDisplay->getDraftPixmap());
  canvasInstance.setWindow(env.m_llx, env.m_lly, env.m_urx, env.m_ury);

  canvasInstance.setPolygonContourWidth(2);
  canvasInstance.setPolygonContourColor(te::color::RGBAColor(100, 177, 216, TE_OPAQUE));

  if(m_geom->getSRID() != m_mapDisplay->getSRID())
    m_geom->transform(m_mapDisplay->getSRID());

  canvasInstance.draw(m_geom.get());

  m_mapDisplay->repaint();
}

void te::qt::widgets::ContrastDialogForm::setDefaultValues(int contrastType)
{
  int row = m_ui->m_bandTableWidget->currentRow();

  m_ui->m_sampledCheckBox->hide();
  m_ui->m_sampledCheckBox->setChecked(false);
  m_ui->m_infoToolButton->setEnabled(true);

  if(row == -1)
    return;

  if(contrastType == te::rp::Contrast::InputParameters::LinearContrastT ||
     contrastType == te::rp::Contrast::InputParameters::SquareContrastT ||
     contrastType == te::rp::Contrast::InputParameters::SquareRootContrastT ||
     contrastType == te::rp::Contrast::InputParameters::LogContrastT ||
    contrastType == te::rp::Contrast::InputParameters::CumulativeCut2EnhancementT)
  {
    bool reset = true;

    if(m_ui->m_bandTableWidget->horizontalHeaderItem(3) && m_ui->m_bandTableWidget->horizontalHeaderItem(3)->text() == tr("Minimum") &&
       m_ui->m_bandTableWidget->horizontalHeaderItem(4) && m_ui->m_bandTableWidget->horizontalHeaderItem(4)->text() == tr("Maximum"))
       reset = false;

    if(reset)
    {
      QStringList list;
      list.append(tr(""));
      list.append(tr("Band"));
      list.append(tr("Minimum"));
      list.append(tr("Maximum"));

      m_ui->m_bandTableWidget->setColumnCount(4);
      m_ui->m_bandTableWidget->setHorizontalHeaderLabels(list);

      int nBands = m_ui->m_bandTableWidget->rowCount();

      for(int i = 0; i < nBands; ++i)
      {
        QTableWidgetItem* band = new QTableWidgetItem(te::common::Convert2String(i).c_str());
        m_ui->m_bandTableWidget->setItem(i, 1, band);

        QTableWidgetItem* itemMin = new QTableWidgetItem(te::common::Convert2String(m_minValue[i], 2).c_str());
        itemMin->setFlags(Qt::ItemIsEnabled | Qt::ItemIsEditable | Qt::ItemIsSelectable);
        m_ui->m_bandTableWidget->setItem(i, 2, itemMin);

        QTableWidgetItem* itemMax = new QTableWidgetItem(te::common::Convert2String(m_maxValue[i], 2).c_str());
        itemMax->setFlags(Qt::ItemIsEnabled | Qt::ItemIsEditable | Qt::ItemIsSelectable);
        m_ui->m_bandTableWidget->setItem(i, 3, itemMax);
      }
    }

    resetInputHistogram();
    resetOutputHistogram();
    m_histogramWidget->setMinimumValueEnabled(true);
    m_histogramWidget->setMaximumValueEnabled(true);
    m_histogramWidget->updateMinimumValueLabel(tr("Minimum"));
    m_histogramWidget->updateMaximumValueLabel(tr("Maximum"));
    m_histogramWidget->updateMinimumValueLine(m_minValue[row], false);
    m_histogramWidget->updateMaximumValueLine(m_maxValue[row], false);

    if (contrastType == te::rp::Contrast::InputParameters::CumulativeCut2EnhancementT)
    {
      m_ui->m_sampledCheckBox->show();
      m_ui->m_sampledCheckBox->setChecked(true);
    }
    else
      m_information = tr("Use left button to set minimum value and right button to define maximum value over the histogram.");
  }
  else if(contrastType == te::rp::Contrast::InputParameters::HistogramEqualizationContrastT)
  {
    QStringList list;
    list.append(tr(""));
    list.append(tr("Band"));
    list.append(tr("Maximum"));

    m_ui->m_bandTableWidget->setColumnCount(3);
    m_ui->m_bandTableWidget->setHorizontalHeaderLabels(list);

    std::unique_ptr<te::rst::Raster> raster(te::map::GetRaster(m_layer.get()));
    int nBands = m_ui->m_bandTableWidget->rowCount();
    double currentRowMaxValue = 0;

    for(int i = 0; i < nBands; ++i)
    {
      const int bandDataType = raster->getBand( i )->getProperty()->m_type;
      const bool isRealValuesBand = ( bandDataType == te::dt::FLOAT_TYPE ) ||
        ( bandDataType == te::dt::DOUBLE_TYPE );
      double bandMaxValue = isRealValuesBand ? 255.0 : m_maxValue[i];

      QTableWidgetItem* band = new QTableWidgetItem(te::common::Convert2String(i).c_str());
      m_ui->m_bandTableWidget->setItem(i, 1, band);

      QTableWidgetItem* itemMax = new QTableWidgetItem(te::common::Convert2String(bandMaxValue, 2).c_str());
      itemMax->setFlags(Qt::ItemIsEnabled | Qt::ItemIsEditable | Qt::ItemIsSelectable);
      m_ui->m_bandTableWidget->setItem(i, 2, itemMax);

      if( i == row )
        currentRowMaxValue = bandMaxValue;
    }

    m_histogramWidget->setMinimumValueEnabled(true);
    m_histogramWidget->setMaximumValueEnabled(false);
    m_histogramWidget->updateMinimumValueLabel(tr("Maximum"));
    m_histogramWidget->updateMinimumValueLine(currentRowMaxValue, false);

    m_information = tr("Use left button to set maximum value over the histogram.");
  }
  else if(contrastType == te::rp::Contrast::InputParameters::MeanAndStdContrastT)
  {
    QStringList list;
    list.append(tr(""));
    list.append(tr("Band"));
    list.append(tr("Mean"));
    list.append(tr("Std Dev"));

    m_ui->m_bandTableWidget->setColumnCount(4);
    m_ui->m_bandTableWidget->setHorizontalHeaderLabels(list);

    int nBands = m_ui->m_bandTableWidget->rowCount();

    for(int i = 0; i < nBands; ++i)
    {
      QTableWidgetItem* band = new QTableWidgetItem(te::common::Convert2String(i).c_str());
      m_ui->m_bandTableWidget->setItem(i, 1, band);

      QTableWidgetItem* itemMean = new QTableWidgetItem("127");
      itemMean->setFlags(Qt::ItemIsEnabled | Qt::ItemIsEditable | Qt::ItemIsSelectable);
      m_ui->m_bandTableWidget->setItem(i, 2, itemMean);

      QTableWidgetItem* itemStdDev = new QTableWidgetItem("50");
      itemStdDev->setFlags(Qt::ItemIsEnabled | Qt::ItemIsEditable | Qt::ItemIsSelectable);
      m_ui->m_bandTableWidget->setItem(i, 3, itemStdDev);
    }

    m_histogramWidget->setMinimumValueEnabled(true);
    m_histogramWidget->setMaximumValueEnabled(true);
    m_histogramWidget->updateMinimumValueLabel(tr("Mean"));
    m_histogramWidget->updateMaximumValueLabel(tr("Std Dev"));
    m_histogramWidget->updateMinimumValueLine(127, false);
    m_histogramWidget->updateMaximumValueLine(50, false);

    m_information = tr("Use left button to set Mean value and right button to define Std Dev value over the histogram.");
  }
  else if(contrastType == te::rp::Contrast::InputParameters::DecorrelationEnhancementT)
  {
    QStringList list;
    list.append(tr(""));
    list.append(tr("Band"));

    m_ui->m_bandTableWidget->setColumnCount(2);
    m_ui->m_bandTableWidget->setHorizontalHeaderLabels(list);

    m_histogramWidget->setMinimumValueEnabled(false);
    m_histogramWidget->setMaximumValueEnabled(false);
    
    m_information = tr("");
  }
  else if(contrastType == te::rp::Contrast::InputParameters::PercentCutUDLinearT)
  {
    QStringList list;
    list.append(tr(""));
    list.append(tr("Band"));
    list.append(tr("Min %"));
    list.append(tr("Max %"));

    m_ui->m_bandTableWidget->setColumnCount(4);
    m_ui->m_bandTableWidget->setHorizontalHeaderLabels(list);

    int nBands = m_ui->m_bandTableWidget->rowCount();

    for(int i = 0; i < nBands; ++i)
    {
      QTableWidgetItem* band = new QTableWidgetItem(te::common::Convert2String(i).c_str());
      m_ui->m_bandTableWidget->setItem(i, 1, band);

      QTableWidgetItem* itemMean = new QTableWidgetItem("0");
      itemMean->setFlags(Qt::ItemIsEnabled | Qt::ItemIsEditable | Qt::ItemIsSelectable);
      m_ui->m_bandTableWidget->setItem(i, 2, itemMean);

      QTableWidgetItem* itemStdDev = new QTableWidgetItem("100");
      itemStdDev->setFlags(Qt::ItemIsEnabled | Qt::ItemIsEditable | Qt::ItemIsSelectable);
      m_ui->m_bandTableWidget->setItem(i, 3, itemStdDev);
    }

    m_histogramWidget->setMinimumValueEnabled(false);
    m_histogramWidget->setMaximumValueEnabled(false);
    m_histogramWidget->updateMinimumValueLabel(tr("Minimum"));
    m_histogramWidget->updateMaximumValueLabel(tr("Maximum"));
    m_histogramWidget->updateMinimumValueLine(m_minValue[row], false);
    m_histogramWidget->updateMaximumValueLine(m_maxValue[row], false);

    m_information = tr("");
  }  

  m_ui->m_bandTableWidget->resizeColumnsToContents();
  #if (QT_VERSION >= 0x050000)
  m_ui->m_bandTableWidget->horizontalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
  #else
  m_ui->m_bandTableWidget->horizontalHeader()->setResizeMode(QHeaderView::ResizeToContents);
  #endif
}

void te::qt::widgets::ContrastDialogForm::onContrastTypeComboBoxActivated(int index)
{
  if(m_ui->m_bandTableWidget->rowCount() == 0)
     return;

  int contrastType = m_ui->m_contrastTypeComboBox->itemData(index).toInt();

  setDefaultValues(contrastType);
}

void te::qt::widgets::ContrastDialogForm::onCellClicked(int row, int /*column*/)
{
  QCheckBox* bandCheckBox = (QCheckBox*)m_ui->m_bandTableWidget->cellWidget(row, 0);

  if(bandCheckBox->isChecked())
        drawInputHistogram();
  
  resetOutputHistogram();
}

void te::qt::widgets::ContrastDialogForm::onCellDoubleClicked(int row, int column)
{
  if (column == 0)
    return;

  m_lastText = m_ui->m_bandTableWidget->item(row, column)->text().toUtf8().data();
  m_updatePreview = true;
  
  resetOutputHistogram();
}

void te::qt::widgets::ContrastDialogForm::onCellChanged(int row, int column)
{
  if(row == -1)
     return;

  bool isNumeric;

  m_ui->m_bandTableWidget->item(row, column)->text().toDouble(&isNumeric);

  if(!isNumeric)
  {
    m_ui->m_bandTableWidget->item(row, column)->setText(m_lastText.c_str());
    return;
  }

  if(m_updatePreview)
  {
    std::unique_ptr<te::rst::Raster> raster(te::map::GetRaster(m_layer.get()));

    int index = m_ui->m_contrastTypeComboBox->currentIndex();

    int contrastType = m_ui->m_contrastTypeComboBox->itemData(index).toInt();

    m_updatePreview = false;

    if(column == 2)
    {
      if(contrastType == te::rp::Contrast::InputParameters::LinearContrastT ||
         contrastType == te::rp::Contrast::InputParameters::SquareContrastT ||
         contrastType == te::rp::Contrast::InputParameters::SquareRootContrastT ||
         contrastType == te::rp::Contrast::InputParameters::LogContrastT)
      {
        if(raster->getBand(0)->getProperty()->getType() == te::dt::UCHAR_TYPE || raster->getBand(0)->getProperty()->getType() == te::dt::UINT16_TYPE ||
           raster->getBand(0)->getProperty()->getType() == te::dt::UINT32_TYPE || raster->getBand(0)->getProperty()->getType() == te::dt::UINT64_TYPE)
        {
          int value = m_ui->m_bandTableWidget->item(row, column)->text().toDouble();
          int maxValue = m_ui->m_bandTableWidget->item(row, 3)->text().toDouble();

          if(value > maxValue)
          {
            m_ui->m_bandTableWidget->item(row, column)->setText(m_lastText.c_str());
            return;
          }

          m_histogramWidget->updateMinimumValueLine(value, true);
        }
        else
        {
          double value = m_ui->m_bandTableWidget->item(row, column)->text().toDouble();
          double maxValue = m_ui->m_bandTableWidget->item(row, 3)->text().toDouble();

          if(value > maxValue)
          {
            m_ui->m_bandTableWidget->item(row, column)->setText(m_lastText.c_str());
            return;
          }

          m_histogramWidget->updateMinimumValueLine(value, true);
        }
      }
      else if(contrastType == te::rp::Contrast::InputParameters::HistogramEqualizationContrastT ||
              contrastType == te::rp::Contrast::InputParameters::MeanAndStdContrastT)
      {
        if(raster->getBand(0)->getProperty()->getType() == te::dt::UCHAR_TYPE || raster->getBand(0)->getProperty()->getType() == te::dt::UINT16_TYPE ||
           raster->getBand(0)->getProperty()->getType() == te::dt::UINT32_TYPE || raster->getBand(0)->getProperty()->getType() == te::dt::UINT64_TYPE)
        {
           int minValue = m_ui->m_bandTableWidget->item(row, column)->text().toDouble();

           m_histogramWidget->updateMinimumValueLine(minValue, true);
        }
        else
        {
          double minValue = m_ui->m_bandTableWidget->item(row, column)->text().toDouble();

          m_histogramWidget->updateMinimumValueLine(minValue, true);
        }
      }
    }
    else if(column == 3)
    {
      if(contrastType == te::rp::Contrast::InputParameters::LinearContrastT ||
         contrastType == te::rp::Contrast::InputParameters::SquareContrastT ||
         contrastType == te::rp::Contrast::InputParameters::SquareRootContrastT ||
         contrastType == te::rp::Contrast::InputParameters::LogContrastT)
      {
        if(raster->getBand(0)->getProperty()->getType() == te::dt::UCHAR_TYPE || raster->getBand(0)->getProperty()->getType() == te::dt::UINT16_TYPE ||
           raster->getBand(0)->getProperty()->getType() == te::dt::UINT32_TYPE || raster->getBand(0)->getProperty()->getType() == te::dt::UINT64_TYPE)
        {
          int value = m_ui->m_bandTableWidget->item(row, column)->text().toDouble();
          int minValue = m_ui->m_bandTableWidget->item(row, 2)->text().toDouble();

          if(value < minValue)
          {
            m_ui->m_bandTableWidget->item(row, column)->setText(m_lastText.c_str());
            return;
          }

          m_histogramWidget->updateMaximumValueLine(value, true);
        }
        else
        {
          double value = m_ui->m_bandTableWidget->item(row, column)->text().toDouble();
          double minValue = m_ui->m_bandTableWidget->item(row, 2)->text().toDouble();

          if(value < minValue)
          {
            m_ui->m_bandTableWidget->item(row, column)->setText(m_lastText.c_str());
            return;
          }

          m_histogramWidget->updateMaximumValueLine(value, true);
        }
      }
      else if(contrastType == te::rp::Contrast::InputParameters::MeanAndStdContrastT)
      {
        if(raster->getBand(0)->getProperty()->getType() == te::dt::UCHAR_TYPE || raster->getBand(0)->getProperty()->getType() == te::dt::UINT16_TYPE ||
           raster->getBand(0)->getProperty()->getType() == te::dt::UINT32_TYPE || raster->getBand(0)->getProperty()->getType() == te::dt::UINT64_TYPE)
        {
           int maxValue = m_ui->m_bandTableWidget->item(row, column)->text().toDouble();

           m_histogramWidget->updateMaximumValueLine(maxValue, true);
        }
        else
        {
          double maxValue = m_ui->m_bandTableWidget->item(row, column)->text().toDouble();

          m_histogramWidget->updateMaximumValueLine(maxValue, true);
        }
      }
    }
  }
  
  drawInputHistogram();
  resetOutputHistogram();
}

void te::qt::widgets::ContrastDialogForm::onMinValueSelected(int value,
                                                             int band )
{
  onMinValueSelected( (double)value, band );
}

void te::qt::widgets::ContrastDialogForm::onMinValueSelected(double value,
                                                             int /*band*/)
{
  if (!m_layer || !m_raster)
    return;

  int row = m_ui->m_bandTableWidget->currentRow();

  if(row == -1)
    return;

  m_updatePreview = true;

  int index = m_ui->m_contrastTypeComboBox->currentIndex();

  int contrastType = m_ui->m_contrastTypeComboBox->itemData(index).toInt();

  if (contrastType == te::rp::Contrast::InputParameters::LinearContrastT ||
    contrastType == te::rp::Contrast::InputParameters::SquareContrastT ||
    contrastType == te::rp::Contrast::InputParameters::SquareRootContrastT ||
    contrastType == te::rp::Contrast::InputParameters::LogContrastT ||
    contrastType == te::rp::Contrast::InputParameters::CumulativeCut2EnhancementT)
  {
    double max = m_ui->m_bandTableWidget->item(row, 3)->text().toDouble();

    if(value > max)
      return;

    m_ui->m_bandTableWidget->item(row, 2)->setText(QString::number(value));
    m_ui->m_bandTableWidget->setCurrentCell(row, 2);
  }
  else if (contrastType == te::rp::Contrast::InputParameters::HistogramEqualizationContrastT)
  {
    m_ui->m_bandTableWidget->item(row, 2)->setText(QString::number(value));
    m_ui->m_bandTableWidget->setCurrentCell(row, 2);
  }
  else if (contrastType == te::rp::Contrast::InputParameters::MeanAndStdContrastT)
  {
    m_ui->m_bandTableWidget->item(row, 2)->setText(QString::number(value));
    m_ui->m_bandTableWidget->setCurrentCell(row, 2);
  }
  else
  {
    return;
  }
}

void te::qt::widgets::ContrastDialogForm::onMaxValueSelected(int value,
                                                             int band )
{
  onMaxValueSelected( (double)value, band);
}

void te::qt::widgets::ContrastDialogForm::onMaxValueSelected(double value,
                                                             int /*band*/)
{
  if (!m_layer || !m_raster)
    return;

  int row = m_ui->m_bandTableWidget->currentRow();

  if(row == -1)
    return;

  m_updatePreview = true;

  int index = m_ui->m_contrastTypeComboBox->currentIndex();

  int contrastType = m_ui->m_contrastTypeComboBox->itemData(index).toInt();

  if (contrastType == te::rp::Contrast::InputParameters::LinearContrastT ||
    contrastType == te::rp::Contrast::InputParameters::SquareContrastT ||
    contrastType == te::rp::Contrast::InputParameters::SquareRootContrastT ||
    contrastType == te::rp::Contrast::InputParameters::LogContrastT)
  {
    double min = m_ui->m_bandTableWidget->item(row, 2)->text().toDouble();

    if(value < min)
      return;

    m_ui->m_bandTableWidget->item(row, 3)->setText(QString::number(value));
    m_ui->m_bandTableWidget->setCurrentCell(row, 3);
  }
  else if (contrastType == te::rp::Contrast::InputParameters::MeanAndStdContrastT)
  {
    if (!m_ui->m_bandTableWidget->item(row, 2)->text().isEmpty())
    {
      double mean = m_ui->m_bandTableWidget->item(row, 2)->text().toDouble();

      double stdDev = std::abs(mean - value);

      m_ui->m_bandTableWidget->item(row, 3)->setText(QString::number(stdDev));
      m_ui->m_bandTableWidget->setCurrentCell(row, 3);
    }
  }
  else
  {
    return;
  }
}

void te::qt::widgets::ContrastDialogForm::onOkPushButtonClicked()
{
  //check output parameters
  if (m_outputWidget->getOutputRaster().empty())
  {
    QMessageBox::information(this, tr("Contrast"), tr("Output image is not defined."));
    return;
  }

  std::vector< unsigned int> checkedBandIndexes;
    getCheckedInputBandIndexes( checkedBandIndexes );  
  if( checkedBandIndexes.empty() )
  {
    QMessageBox::information(this, tr("Contrast"), tr("Invalid number of bands."));
    return;
  }

  if (!execute())
  {
  //  QMessageBox::information(this, tr("Contrast"), tr("Error to apply contrast."));
    return;
  }

  clearCanvas();

  emit closeTool();
}

void te::qt::widgets::ContrastDialogForm::layerChanged()
{
  m_layer = m_inputWidget->getInputLayer();

  m_minValue.clear();
  m_maxValue.clear();

  onAllImageRadioButtonToggled(m_ui->m_allImageRadioButton->isChecked());
  onVisibleAreaRadioButtonToggled(m_ui->m_visibleAreaRadioButton->isChecked());
  onRoiRadioButtonToggled(m_ui->m_roiRadioButton->isChecked());
}

void te::qt::widgets::ContrastDialogForm::onApplyPreviewClicked()
{
  if(m_ui->m_bandTableWidget->rowCount() == 0)
  {
    QMessageBox::warning(this, tr("Warning"), tr("Draw histogram before to apply contrast preview."));
    return;
  }

  if(m_ui->m_roiRadioButton->isChecked() || m_ui->m_visibleAreaRadioButton->isChecked())
  {
    if(!m_geom.get() || !m_geom->isValid())
    {
      QMessageBox::warning(this, tr("Warning"), tr("ROI is invalid."));
      return;
    }
  }

  applyPreview();
}

void te::qt::widgets::ContrastDialogForm::drawInputHistogram()
{
  if(!(m_ui->m_allImageRadioButton->isChecked() || m_ui->m_visibleAreaRadioButton->isChecked() ||
       m_ui->m_roiRadioButton->isChecked()))
  {
    QMessageBox::warning(this, tr("Warning"), tr("Select a Histogram Area to draw histogram."));
    return;
  }

  m_minValue.clear();
  m_maxValue.clear();
  
  te::rst::RasterSummary rasterSummary;
  {
    te::qt::widgets::ScopedCursor cursor(Qt::WaitCursor);
    
    if( ! te::rst::RasterSummaryManager::getInstance().get(
      m_raster.get(), false, rasterSummary ) )
    {
      throw te::common::Exception(TE_TR("Raster summary Error"));
    }
  }
  
  for (std::size_t b = 0; b < m_raster->getNumberOfBands(); ++b)
  {
    const std::complex<double>* cmin = rasterSummary.at(b).m_minVal;
    const std::complex<double>* cmax = rasterSummary.at(b).m_maxVal;
    double min = cmin->real();
    double max = cmax->real();

    m_minValue.push_back(min);
    m_maxValue.push_back(max);
  }

  if(m_ui->m_bandTableWidget->rowCount() == 0)
     listBands();

  applyHistogram();
}

void te::qt::widgets::ContrastDialogForm::drawOutputHistogram()
{
  resetOutputHistogram();
  
  if( m_previewRasterPtr.get() != nullptr )
  {
    unsigned int currentBand = 0;
    if( ! getSelectOutputBandIdx( currentBand ) ) return;
    
    te::qt::widgets::ScopedCursor cursor(Qt::WaitCursor);
    
    m_histogramWidget2->setInputRaster( m_previewRasterPtr.get() );
    
    m_histogramWidget2->drawHistogram((int)currentBand);    
  }
}

void te::qt::widgets::ContrastDialogForm::onRoiRadioButtonToggled(bool isChecked)
{
  if(isChecked)
  {
    clearCanvas();

    onResetValuesToolButtonClicked();

    m_ui->m_resetToolButton->setEnabled(true);
    m_ui->m_applyContrastToolButton->setEnabled(true);
  }

  m_ui->m_navigatorWidget->setEnabled(isChecked);
  m_navigator->setEnvelopeTool(isChecked);
}

void te::qt::widgets::ContrastDialogForm::onVisibleAreaRadioButtonToggled(bool isChecked)
{
  if(isChecked)
  {   
    if(!m_mapDisplay->getExtent().isValid())
      return;

    clearCanvas();

    resetInputHistogram();
    resetOutputHistogram();

    m_ui->m_navigatorWidget->setEnabled(false);

    try
    {
      te::gm::Envelope reprojectedBBOX(m_mapDisplay->getExtent());
      reprojectedBBOX.transform(m_mapDisplay->getSRID(), m_layer->getSRID());

      if (!reprojectedBBOX.intersects(m_layer->getExtent()))
         throw te::common::Exception(TE_TR("Visible area doesn't intersect with the layer."));

      te::gm::Envelope ibbox = reprojectedBBOX.intersection(m_layer->getExtent());

      if (ibbox.isValid())
        m_geom.reset(te::gm::GetGeomFromEnvelope(&ibbox, m_mapDisplay->getSRID()));

      getRasterFromROI();

      drawGeom();

      m_ui->m_resetToolButton->setEnabled(true);
      m_ui->m_applyContrastToolButton->setEnabled(true);

      drawInputHistogram();

      onContrastTypeComboBoxActivated(m_ui->m_contrastTypeComboBox->currentIndex());
    }
    catch(te::common::Exception &e)
    { 
      onResetValuesToolButtonClicked();

      QMessageBox::warning(this, tr("Contrast"), e.what());
      return;
    }
  }
}

void te::qt::widgets::ContrastDialogForm::applyHistogram()
{
  unsigned int bandIdx = 0;
  if( ! getSelectInputBandIdx( bandIdx ) )
  {
    return;
  }

  if( m_raster.get() == nullptr ) return;

  resetInputHistogram();
  m_histogramWidget->setInputRaster(m_raster.get());

  int index = m_ui->m_contrastTypeComboBox->currentIndex();

  int contrastType = m_ui->m_contrastTypeComboBox->itemData(index).toInt();

  if(contrastType == te::rp::Contrast::InputParameters::LinearContrastT ||
     contrastType == te::rp::Contrast::InputParameters::SquareContrastT ||
     contrastType == te::rp::Contrast::InputParameters::SquareRootContrastT ||
     contrastType == te::rp::Contrast::InputParameters::LogContrastT)
  {
    QTableWidgetItem* item = m_ui->m_bandTableWidget->item(bandIdx, 2);
    QTableWidgetItem* item1 = m_ui->m_bandTableWidget->item(bandIdx, 3);

    if (item && item1)
    {
      double min = item->text().toDouble();
      double max = item1->text().toDouble();

      m_histogramWidget->updateMinimumValueLine(min, false);
      m_histogramWidget->updateMaximumValueLine(max, false);
    }
  }
  else if(contrastType == te::rp::Contrast::InputParameters::HistogramEqualizationContrastT)
  {
    QTableWidgetItem* item = m_ui->m_bandTableWidget->item(bandIdx, 2);
    if (item)
    {
      double min = item->text().toDouble();

      m_histogramWidget->updateMinimumValueLine(min, false);
    }
  }
  else if(contrastType == te::rp::Contrast::InputParameters::MeanAndStdContrastT)
  {
    QTableWidgetItem* item = m_ui->m_bandTableWidget->item(bandIdx, 2);
    QTableWidgetItem* item1 = m_ui->m_bandTableWidget->item(bandIdx, 3);

    if (item && item1)
    {
      double min = item->text().toDouble();
      double max = item1->text().toDouble();

      m_histogramWidget->updateMinimumValueLine(min, false);
      m_histogramWidget->updateMaximumValueLine(max, false);
    }
  }

  m_histogramWidget->drawHistogram(bandIdx);
}

void te::qt::widgets::ContrastDialogForm::onApplicationTriggered(te::qt::af::evt::Event* e)
{
  if (e->m_id == te::qt::af::evt::LAYER_SELECTED)
  {
    te::qt::af::evt::LayerSelected* evt = static_cast<te::qt::af::evt::LayerSelected*>(e);

    te::map::AbstractLayerPtr layer = evt->m_layer;
    assert(layer);

    if (layer == nullptr)
      return;

    resetWindow(layer);
  }
}

void te::qt::widgets::ContrastDialogForm::onMapDisplayExtentChanged()
{
  drawGeom();

  applyPreview();
}

void te::qt::widgets::ContrastDialogForm::onEnvelopeAcquired(te::gm::Envelope env)
{
  if(!env.isValid())
     return;

  if(m_mapDisplay->getSRID() != m_layer->getSRID())
    env.transform(m_mapDisplay->getSRID(), m_layer->getSRID());

  if(!env.intersects(m_layer->getExtent()))
    return;

  m_geom.reset(te::gm::GetGeomFromEnvelope(&env, m_layer->getSRID()));

  getRasterFromROI();

  drawGeom();

  drawInputHistogram();
  resetOutputHistogram();

  onContrastTypeComboBoxActivated(m_ui->m_contrastTypeComboBox->currentIndex());
}

void te::qt::widgets::ContrastDialogForm::onBandChecked()
{
  std::vector< unsigned int> bandIndexes;
    getCheckedInputBandIndexes( bandIndexes );
  
  if( ! bandIndexes.empty() )
  {
    applyPreview();
  }
}

void te::qt::widgets::ContrastDialogForm::onResetValuesToolButtonClicked()
{
  m_ui->m_bandTableWidget->setRowCount(0);

  resetInputHistogram();
  resetOutputHistogram();

  m_minValue.clear();
  m_maxValue.clear();

  clearCanvas();

  if(!m_ui->m_roiRadioButton->isChecked())
  {
    if(m_ui->m_visibleAreaRadioButton->isChecked())
    {
      te::gm::Envelope reprojectedBBOX(m_mapDisplay->getExtent());
      reprojectedBBOX.transform(m_mapDisplay->getSRID(), m_layer->getSRID());

      if (!reprojectedBBOX.intersects(m_layer->getExtent()))
         throw te::common::Exception(TE_TR("Visible area doesn't intersect with the layer."));

      te::gm::Envelope ibbox = reprojectedBBOX.intersection(m_layer->getExtent());

      if (ibbox.isValid())
        m_geom.reset(te::gm::GetGeomFromEnvelope(&ibbox, m_mapDisplay->getSRID()));

      getRasterFromROI();

      drawGeom();
    }

    drawInputHistogram();
    resetOutputHistogram();
    
    onContrastTypeComboBoxActivated(m_ui->m_contrastTypeComboBox->currentIndex());
  }
}

void te::qt::widgets::ContrastDialogForm::onInfoToolButtonClicked()
{
  QMessageBox::information(this, tr("Contrast"), tr(m_information.toUtf8().data()));
}

void te::qt::widgets::ContrastDialogForm::resetInputHistogram()
{
  if( m_histogramWidget )
  {
    m_histogramWidget->clear();
    m_histogramWidget->setHistogramInputTitle("");
    m_histogramWidget->setHistogramOutputTitle("");
    m_histogramWidget->setChartInputColor( 255, 0, 0 );
    m_histogramWidget->setChartOutputColor( 255, 0, 0 );
    m_histogramWidget->clearMinMaxLines();
  }
}

void te::qt::widgets::ContrastDialogForm::resetOutputHistogram()
{
  if( m_histogramWidget2 )
  {
    m_histogramWidget2->clear();
    m_histogramWidget2->setHistogramInputTitle("");
    m_histogramWidget2->setHistogramOutputTitle("");  
    m_histogramWidget2->setChartInputColor( 0, 0, 255 );
    m_histogramWidget2->setChartOutputColor( 0, 0, 255 );
    m_histogramWidget2->setEnableTools( false );
    m_histogramWidget2->setEnableNavigationTools( false );
    m_histogramWidget2->setMinimumValueEnabled(false);  
    m_histogramWidget2->setMaximumValueEnabled(false);
    m_histogramWidget2->clearMinMaxLines();
  }
}

bool te::qt::widgets::ContrastDialogForm::getSelectInputBandIdx( unsigned int& bandIdx )
{
  QList<QTableWidgetItem *> selectedItens = m_ui->m_bandTableWidget->selectedItems();
  
  if( selectedItens.isEmpty() ) return false;
  
  bandIdx = (unsigned int)selectedItens[ 0 ]->row();
  
  if( ((QCheckBox*)m_ui->m_bandTableWidget->cellWidget(bandIdx, 0))->isChecked() )
  {
    return true;
  }
  else
  {
    return false;
  }
}

bool te::qt::widgets::ContrastDialogForm::getSelectOutputBandIdx( unsigned int& bandIdx )
{
  int rowsCount = m_ui->m_bandTableWidget->rowCount();
  unsigned int checkedCount = 0;
  
  for(int i = 0; i < rowsCount; ++i)
  {
    if( ((QCheckBox*)m_ui->m_bandTableWidget->cellWidget(i, 0))->isChecked() )
    {
      if( ( m_ui->m_bandTableWidget->item( i, 1 )->isSelected() ) )
      {
        bandIdx = checkedCount;
        return true;
      }
      
      ++checkedCount;
    }
  }
  
  return false;
}

void te::qt::widgets::ContrastDialogForm::getCheckedInputBandIndexes(
  std::vector< unsigned int> & bandIndexes )
{
  bandIndexes.clear();
  
  int rowsCount = m_ui->m_bandTableWidget->rowCount();
  
  for(int i = 0; i < rowsCount; ++i)
  {
    QCheckBox* bandCheckBox = (QCheckBox*)m_ui->m_bandTableWidget->cellWidget(i, 0);

    if(bandCheckBox->isChecked())
    {
      bandIndexes.push_back( (unsigned int)i );
    }
  }  
}

void te::qt::widgets::ContrastDialogForm::onEableDisableHistoToolButtonClicked(bool checked)
{
  if( 
      m_ui->m_histWidget->isVisible()
      &&
      m_ui->m_histWidget2->isVisible()
    )
  {
    m_ui->m_histWidget2->setVisible( false );
    m_ui->m_outLabel->setVisible( false );
  }
  else if( 
            m_ui->m_histWidget->isVisible()
            &&
           (!m_ui->m_histWidget2->isVisible())
         )
  {
    m_ui->m_histWidget->setVisible( false );
    m_ui->m_inLabel->setVisible( false );
    
    m_ui->m_histWidget2->setVisible( true );
    m_ui->m_outLabel->setVisible( true );
  }
  else
  {
    m_ui->m_histWidget->setVisible( true );
    m_ui->m_inLabel->setVisible( true );
    
    m_ui->m_histWidget2->setVisible(  true );
    m_ui->m_outLabel->setVisible(  true );
  }
}
