/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/widgets/rp/MajorityFilterDialogForm.h

  \brief This file has the MajorityFilterDialogForm class.
*/

// TerraLib
#include "../../../common/progress/ProgressManager.h"
#include "../../../common/StringUtils.h"
#include "../../../core/translator/Translator.h"
#include "../../../dataaccess/dataset/DataSet.h"
#include "../../../dataaccess/dataset/DataSetType.h"
#include "../../../dataaccess/utils/Utils.h"
#include "../../../geometry/Utils.h"
#include "../../../maptools/Utils.h"
#include "../../../qt/widgets/utils/ScopedCursor.h"
#include "../../../raster/BandProperty.h"
#include "../../../raster/RasterProperty.h"
#include "../../../rp/MajorityFilter.h"
#include "../../../rp/Module.h"
#include "../../../se/CoverageStyle.h"
#include "../../../se/RasterSymbolizer.h"
#include "../../../se/Rule.h"
#include "../../../se/Utils.h"
#include "../../af/events/LayerEvents.h"
#include "../../widgets/tools/ExtentAcquire.h"
#include "../canvas/Canvas.h"
#include "../canvas/MapDisplay.h"
#include "../progress/ProgressViewerDialog.h"
#include "MajorityFilterDialog.h"
#include "Utils.h"
#include "ui/ui_MajorityFilterDialogForm.h"

//QT
#include <QApplication>
#include <QGridLayout>
#include <QMessageBox>
#include <QCheckBox>

//Boost
#include <boost/lexical_cast.hpp>

te::qt::widgets::MajorityFilterDialog::MajorityFilterDialog(QWidget* parent)
  : QDialog(parent),
    m_ui(new Ui::MajorityFilterDialogForm),
    m_rasterPreview(nullptr),
    m_previewRaster(nullptr),
    m_geom(nullptr),
    m_mapDisplay(nullptr)
{
//build form
  m_ui->setupUi(this);

  QGridLayout* displayLayout = new QGridLayout(m_ui->m_navigatorWidget);
  m_navigator.reset(new te::qt::widgets::RpToolsWidget(m_ui->m_navigatorWidget));
  displayLayout->addWidget(m_navigator.get());
  displayLayout->setContentsMargins(0,0,0,0);

  //build input layer
  QGridLayout* inputLayout = new QGridLayout(m_ui->m_inputWidget);
  inputLayout->setContentsMargins(0, 0, 0, 0);
  m_inputWidget = new te::qt::widgets::InputLayerWidget(m_ui->m_inputWidget);
  inputLayout->addWidget(m_inputWidget);

  //build output parameters
  QGridLayout* outputLayout = new QGridLayout(m_ui->m_outputWidget);
  outputLayout->setContentsMargins(0, 0, 0, 0);
  m_outputWidget = new te::qt::widgets::OutputRasterWidget(m_ui->m_outputWidget);
  outputLayout->addWidget(m_outputWidget);

  m_ui->m_npixelsLineEdit->setValidator(new QIntValidator(this));

  m_ui->m_helpPushButton->setNameSpace("dpi.inpe.br.plugins");
  //m_ui->m_helpPushButton->setPageReference("plugins/rp/rp_filter.html");

  m_ui->m_navigatorWidget->setEnabled(false);

  m_ui->m_simpleGroupBox->show();
  m_ui->m_fullGroupBox->hide();

  //connects
  connect(m_ui->m_typeComboBox, SIGNAL(activated(int)), SLOT(onFilterTypeComboBoxActivated(int)));
  connect(m_ui->m_previewCheckBox, SIGNAL(toggled(bool)), SLOT(onPreviewCheckBoxToggled(bool)));
  connect(m_inputWidget, SIGNAL(layerChanged()), this, SLOT(layerChanged()));
  connect(m_ui->m_okPushButton, SIGNAL(clicked()), SLOT(onOkPushButtonClicked()));
  connect(m_navigator.get(), SIGNAL(envelopeAcquired(te::gm::Envelope)), this, SLOT(onEnvelopeAcquired(te::gm::Envelope)));
  connect(m_ui->m_neighborsComboBox, SIGNAL(activated(int)), SLOT(onNeighborsComboBoxActivated(int)));
  connect(m_ui->m_thresholdComboBox, SIGNAL(activated(int)), SLOT(onThresholdComboBoxActivated(int)));
  connect(m_ui->m_npixelsLineEdit, SIGNAL(textEdited(const QString&)), SLOT(onNPixelsLineEditTextEdited(const QString&)));

  listFilterTypes();
}

te::qt::widgets::MajorityFilterDialog::~MajorityFilterDialog()
{
  if(m_geom)
    delete m_geom;

  if (m_rasterPreview)
    delete m_rasterPreview;

  if (m_previewRaster)
    delete m_previewRaster;
}

void te::qt::widgets::MajorityFilterDialog::setList(std::list<te::map::AbstractLayerPtr>& layerList, te::map::AbstractLayerPtr selectedlayer)
{
  std::list<te::map::AbstractLayerPtr> layersRasterList;

  std::list<te::map::AbstractLayerPtr>::iterator it = layerList.begin();
  //Filter only raster layers
  while (it != layerList.end())
  {
    std::unique_ptr<te::da::DataSetType> dsType = it->get()->getSchema();
    if (dsType->hasRaster())
    {
      te::dt::Property* property = dsType->findFirstPropertyOfType(te::dt::RASTER_TYPE);
      if (property != nullptr)
      {
        te::rst::RasterProperty* rasterproperty = static_cast<te::rst::RasterProperty*>(property);
        const std::vector<te::rst::BandProperty*>& vecBandProperties = rasterproperty->getBandProperties();
        if (vecBandProperties.empty() == false)
        {
          te::rst::BandProperty* bandProperty = vecBandProperties.at(0);
          if (bandProperty->m_colorInterp == te::rst::PaletteIdxCInt)
          {
            layersRasterList.push_back(it->get());
          }
        }
      }
    }

    ++it;
  }

  m_inputWidget->setLayerList(layersRasterList);

  if (!layersRasterList.empty())
  {
    m_layer = layersRasterList.begin()->get();
  }
}

void te::qt::widgets::MajorityFilterDialog::setMapDisplay(te::qt::widgets::MapDisplay* mapDisplay)
{
  m_navigator->setMapDisplay(mapDisplay);
  m_mapDisplay = mapDisplay;
}

void te::qt::widgets::MajorityFilterDialog::setActionGroup(QActionGroup* actionGroup)
{
  //configure tools
  m_navigator->setActionGroup(actionGroup);
  m_navigator->enableBoxAction();
}

te::rp::MajorityFilter::InputParameters  te::qt::widgets::MajorityFilterDialog::getInputParams()
{
  int idx = m_ui->m_typeComboBox->currentIndex();

  te::rp::MajorityFilter::InputParameters algoInputParams;

  algoInputParams.m_enableProgress = true;
  algoInputParams.m_filterType = (te::rp::MajorityFilter::InputParameters::MajorityFilterType)m_ui->m_typeComboBox->itemData(idx).toInt();

  if (algoInputParams.m_filterType == te::rp::MajorityFilter::InputParameters::MajorityFilterType::SimpleFilterT)
  {
    algoInputParams.m_npixels = m_ui->m_npixelsLineEdit->text().toInt();
  }
  else if (algoInputParams.m_filterType == te::rp::MajorityFilter::InputParameters::MajorityFilterType::FullFilterT)
  {
    int thresholdId = m_ui->m_thresholdComboBox->currentIndex();

    algoInputParams.m_neighbors = m_ui->m_neighborsComboBox->currentText().toInt();
    algoInputParams.m_threshold = (te::rp::MajorityFilter::InputParameters::MajorityFilterThreshold)m_ui->m_thresholdComboBox->itemData(thresholdId).toInt();
  }

  return algoInputParams;
}

void te::qt::widgets::MajorityFilterDialog::onFilterTypeComboBoxActivated(int index)
{
  int filterType = m_ui->m_typeComboBox->itemData(index).toInt();

  if (filterType == te::rp::MajorityFilter::InputParameters::SimpleFilterT)
  {
    m_ui->m_fullGroupBox->hide();
    m_ui->m_simpleGroupBox->show();
  }
  else
  {
    m_ui->m_simpleGroupBox->hide();
    m_ui->m_fullGroupBox->show();
  }

  this->adjustSize();

  if(m_ui->m_previewCheckBox->isChecked())
     applyPreview();
}

void te::qt::widgets::MajorityFilterDialog::onNeighborsComboBoxActivated(int index)
{
  if (m_ui->m_previewCheckBox->isChecked())
    applyPreview();
}

void te::qt::widgets::MajorityFilterDialog::onThresholdComboBoxActivated(int index)
{
  if (m_ui->m_previewCheckBox->isChecked())
    applyPreview();
}

void te::qt::widgets::MajorityFilterDialog::onNPixelsLineEditTextEdited(const QString& npixels)
{
  if (npixels.isEmpty())
    return;

  if (npixels.toInt() < 0)
  {
    QMessageBox::information(this, tr("Majority Filter"), tr("Number of pixels is invalid."));
    return;
  }

  if (m_ui->m_previewCheckBox->isChecked())
    applyPreview();
}

void te::qt::widgets::MajorityFilterDialog::onPreviewCheckBoxToggled(bool isChecked)
{
  if (isChecked)
  {
    m_ui->m_navigatorWidget->setEnabled(true);
    m_navigator->setEnvelopeTool(true);
  }
  else
  {
    m_ui->m_navigatorWidget->setEnabled(false);
    m_navigator->setEnvelopeTool(false);
  }
}

void te::qt::widgets::MajorityFilterDialog::onOkPushButtonClicked()
{
  if(!m_inputWidget->getInputLayer().get())
  {
    QMessageBox::information(this, tr("Majority Filter"), tr("Input image is not defined."));
    return;
  }

  //check output parameters
  if (m_outputWidget->getOutputRaster().empty())
  {
    QMessageBox::information(this, tr("Filter Majority"), tr("Output image is not defined."));
    return;
  }

  int idx = m_ui->m_typeComboBox->currentIndex();

  if ((te::rp::MajorityFilter::InputParameters::MajorityFilterType)m_ui->m_typeComboBox->itemData(idx).toInt() == te::rp::MajorityFilter::InputParameters::MajorityFilterType::SimpleFilterT)
  {
    if (m_ui->m_npixelsLineEdit->text().isEmpty())
    {
      QMessageBox::information(this, tr("Majority Filter"), tr("Number os pixels is not defined."));
      return;
    }
  }

  if (!execute())
  {
    QMessageBox::information(this, tr("Filter"), tr("Error to execute the filter."));
    return;
  }

  clearCanvas();

  emit closeTool();
}

void te::qt::widgets::MajorityFilterDialog::layerChanged()
{
  m_layer = m_inputWidget->getInputLayer();
}

void te::qt::widgets::MajorityFilterDialog::onEnvelopeAcquired(te::gm::Envelope env)
{
  if (!m_ui->m_previewCheckBox->isChecked() || !env.isValid())
    return;

  if (env.getLowerLeftX() == 0 && env.getLowerLeftY() == 0 && env.getUpperRightX() == 0 && env.getUpperRightY() == 0)
    return;

  // verify if is necessary convert the raster to the given srid
  bool needRemap = false;
  if ((m_mapDisplay->getSRID() != TE_UNKNOWN_SRS) && (m_layer->getSRID() != TE_UNKNOWN_SRS) && (m_mapDisplay->getSRID() != m_layer->getSRID()))
    needRemap = true;

  te::gm::Envelope envOut = env;

  // create a SRS converter
  std::unique_ptr<te::srs::Converter> converter(new te::srs::Converter());

  if (needRemap)
  {
    converter->setSourceSRID(m_mapDisplay->getSRID());
    converter->setTargetSRID(m_layer->getSRID());

    converter->convert(envOut.getLowerLeftX(), envOut.getLowerLeftY(), envOut.m_llx, envOut.m_lly);
    converter->convert(envOut.getUpperRightX(), envOut.getUpperRightY(), envOut.m_urx, envOut.m_ury);
  }

  if (!envOut.intersects(m_layer->getExtent()))
  {
    QMessageBox::warning(this, tr("Warning"), tr("ROI is invalid."));
    return;
  }
  if (m_geom)
    delete m_geom;

  m_geom = te::gm::GetGeomFromEnvelope(&env, m_mapDisplay->getSRID());

  if(!m_geom->isValid())
     return;

  getRasterFromROI();

  drawGeom();
}

void te::qt::widgets::MajorityFilterDialog::onMapDisplayExtentChanged()
{
  drawGeom();

  if (m_ui->m_previewCheckBox->isChecked() && m_previewRaster)
  {
    drawPreview(m_previewRaster);
  }
}

bool te::qt::widgets::MajorityFilterDialog::execute()
{
  if(m_outputWidget->fileExists())
  {
    QMessageBox::warning(this, tr("Majority Filter"), tr("File already exists."));
    return false;
  }

  //get layer
  std::unique_ptr<te::rst::Raster> inputRst(te::map::GetRaster(m_layer.get()));

  //run majority filter
  te::rp::MajorityFilter algorithmInstance;

  te::rp::MajorityFilter::InputParameters algoInputParams = getInputParams();
  algoInputParams.m_inRasterPtr = inputRst.get();

  te::rp::MajorityFilter::OutputParameters algoOutputParams;
  algoOutputParams.m_rInfo = m_outputWidget->getInfo();
  algoOutputParams.m_rType = m_outputWidget->getType();

  try
  {
    //progress
    te::qt::widgets::ProgressViewerDialog v(this);

    if (algorithmInstance.initialize(algoInputParams))
    {
      bool res = false;
      {
        te::qt::widgets::ScopedCursor cursor(Qt::WaitCursor);
        res = algorithmInstance.execute(algoOutputParams);
      }
      
      if ( res )
      {
        algoOutputParams.reset();

        //set output layer
        m_outputLayer = te::qt::widgets::createLayer(m_outputWidget->getType(),
                                                     m_outputWidget->getInfo());
      }
      else
      {
          std::string what = TE_TR("Filter execution error.");
          what += " " + te::rp::Module::getLastLogStr();
          throw te::common::Exception(what);
      }
    }
    else
    {
        std::string what = TE_TR("Filter execution error.");
        what += " " + te::rp::Module::getLastLogStr();
        throw te::common::Exception(what);
    }

    te::qt::widgets::applyRasterMultiResolution(tr("Majority Filter"), te::map::GetRaster(m_outputLayer.get()));

    m_outputLayer->setStyle(m_layer->getStyle()->clone());

    emit addLayer(m_outputLayer);
  }
  catch (const std::exception& e)
  {
    QMessageBox::critical(this, tr("Majority Filter"), e.what());
    return false;
  }

  QMessageBox::information(this, tr("Majority Filter"), tr("Filter ended sucessfully."));

  return true;
}

void te::qt::widgets::MajorityFilterDialog::listFilterTypes()
{
  m_ui->m_typeComboBox->clear();

  m_ui->m_typeComboBox->addItem(tr("Simple"), te::rp::MajorityFilter::InputParameters::SimpleFilterT);
  m_ui->m_typeComboBox->addItem(tr("Full"), te::rp::MajorityFilter::InputParameters::FullFilterT);

  m_ui->m_neighborsComboBox->addItem("4");
  m_ui->m_neighborsComboBox->addItem("8");

  m_ui->m_thresholdComboBox->addItem(tr("Majority"), te::rp::MajorityFilter::InputParameters::MajorityThreshold);
  m_ui->m_thresholdComboBox->addItem(tr("Half"), te::rp::MajorityFilter::InputParameters::HalfThreshold);

  onFilterTypeComboBoxActivated(0);
}

void te::qt::widgets::MajorityFilterDialog::applyPreview()
{
  if (!m_rasterPreview)
    return;

  //run majority filter
  te::rp::MajorityFilter algorithmInstance;

  te::rp::MajorityFilter::InputParameters algoInputParams = getInputParams();
  algoInputParams.m_inRasterPtr = m_rasterPreview;

  te::rp::MajorityFilter::OutputParameters algoOutputParams;

  std::map<std::string, std::string> rinfo;
  rinfo["MEM_RASTER_NROWS"] = boost::lexical_cast<std::string>(m_rasterPreview->getNumberOfRows());
  rinfo["MEM_RASTER_NCOLS"] = boost::lexical_cast<std::string>(m_rasterPreview->getNumberOfColumns());
  rinfo["MEM_RASTER_DATATYPE"] = boost::lexical_cast<std::string>(m_rasterPreview->getBandDataType(0));
  rinfo["MEM_RASTER_NBANDS"] = boost::lexical_cast<std::string>(m_rasterPreview->getNumberOfBands());

  algoOutputParams.m_rType = "MEM";
  algoOutputParams.m_rInfo = rinfo;

  try
  {
  //progress
  te::qt::widgets::ProgressViewerDialog v(this);

    te::qt::widgets::ScopedCursor cursor(Qt::WaitCursor);

  if (algorithmInstance.initialize(algoInputParams))
  {
    if (algorithmInstance.execute(algoOutputParams))
    {
      //set output layer
      te::dt::AbstractData* abs = algoOutputParams.m_outputRasterPtr->clone();

      if (m_previewRaster)
        delete m_previewRaster;

      te::rst::Raster* rst = static_cast<te::rst::Raster*>(abs);
      m_previewRaster = rst;

      if (m_ui->m_previewCheckBox->isChecked())
        drawPreview(m_previewRaster);

    }
    else
    {
        std::string what = TE_TR("Filter execution error.");
        what += " " + te::rp::Module::getLastLogStr();
        throw te::common::Exception(what);
    }
  }
  else
  {
      std::string what = TE_TR("Filter execution error.");
      what += " " + te::rp::Module::getLastLogStr();
      throw te::common::Exception(what);
  }
  }
  catch (const std::exception& e)
  {
    QMessageBox::critical(this, tr("Majority Filter"), e.what());
  }
}

void te::qt::widgets::MajorityFilterDialog::drawPreview(te::rst::Raster* raster)
{
  te::qt::widgets::ScopedCursor cursor(Qt::WaitCursor);

  const te::gm::Envelope& envRaster = *raster->getExtent();
  const te::gm::Envelope& env = m_mapDisplay->getExtent();

  te::se::Style* style = m_layer->getStyle();

  te::se::CoverageStyle* cs = dynamic_cast<te::se::CoverageStyle*>(style);
  assert(cs);

  te::qt::widgets::Canvas* canvasInstance = new Canvas(m_mapDisplay->width(), m_mapDisplay->height());
  canvasInstance->setDevice(m_mapDisplay->getDraftPixmap(), false);
  canvasInstance->setWindow(env.m_llx, env.m_lly, env.m_urx, env.m_ury);

  canvasInstance->setPolygonContourWidth(2);
  canvasInstance->setPolygonContourColor(te::color::RGBAColor(100, 177, 216, TE_OPAQUE));

  // Draw raster
  bool cancel = false;
  te::map::DrawRaster(raster, canvasInstance, envRaster, raster->getSRID(), env, m_mapDisplay->getSRID(), cs, nullptr, m_mapDisplay->getScale(), &cancel);
 
  if (!m_geom->isValid())
    return;

  canvasInstance->draw(m_geom);

  m_mapDisplay->repaint();

  delete canvasInstance;

}

void te::qt::widgets::MajorityFilterDialog::drawGeom()
{
  if(!m_geom || !m_geom->isValid())
    return;

  const te::gm::Envelope& env = m_mapDisplay->getExtent();

  te::qt::widgets::Canvas* canvasInstance = new Canvas(m_mapDisplay->width(), m_mapDisplay->height());
  canvasInstance->setDevice(m_mapDisplay->getDraftPixmap(), false);
  canvasInstance->setWindow(env.m_llx, env.m_lly, env.m_urx, env.m_ury);

  canvasInstance->setPolygonContourWidth(2);
  canvasInstance->setPolygonContourColor(te::color::RGBAColor(100, 177, 216, TE_OPAQUE));

  canvasInstance->draw(m_geom);

  m_mapDisplay->repaint();

  delete canvasInstance;
}

void te::qt::widgets::MajorityFilterDialog::getRasterFromROI()
{
  if (m_geom->getGeomTypeId() == te::gm::PolygonType)
  {
    {
      te::qt::widgets::ScopedCursor cursor(Qt::WaitCursor);

    te::gm::Envelope* env = new te::gm::Envelope(*m_geom->getMBR());

    // verify if is necessary convert the raster to the given srid
    bool needRemap = false;
    if ((m_mapDisplay->getSRID() != TE_UNKNOWN_SRS) && (m_layer->getSRID() != TE_UNKNOWN_SRS) && (m_mapDisplay->getSRID() != m_layer->getSRID()))
      needRemap = true;

    // create a SRS converter
    std::unique_ptr<te::srs::Converter> converter(new te::srs::Converter());

    if (needRemap)
    {
      converter->setSourceSRID(m_mapDisplay->getSRID());
      converter->setTargetSRID(m_layer->getSRID());

      converter->convert(env->getLowerLeftX(), env->getLowerLeftY(), env->m_llx, env->m_lly);
      converter->convert(env->getUpperRightX(), env->getUpperRightY(), env->m_urx, env->m_ury);
    }

    std::unique_ptr<te::rst::Raster> raster(te::map::GetRaster(m_layer.get()));

    std::map<std::string, std::string> info;
    info["FORCE_MEM_DRIVER"] = "TRUE";
    
    if (m_rasterPreview)
      delete m_rasterPreview;

    //execute clipping
    m_rasterPreview = raster->trim(env, info);

    delete env;
    }

    if (m_ui->m_previewCheckBox->isChecked())
    {
      applyPreview();
    }
  }
}

void te::qt::widgets::MajorityFilterDialog::clearCanvas()
{
  te::qt::widgets::Canvas canvasInstance(m_mapDisplay->getDraftPixmap());

  canvasInstance.clear();

  m_mapDisplay->repaint();
}

void te::qt::widgets::MajorityFilterDialog::closeEvent(QCloseEvent* /*e*/)
{
  clearCanvas();

  emit closeTool();
}
