/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/widgets/rp/MixtureModelWizardPage.h

  \brief This file defines a class for a MixtureModel Wizard page.

  /brief JSON File Structure used to save the mix model components information

  {
    "Band_Sensors":{
    [
      "Band": {
        "band": "****",
        "use": "****",
        "sensor": "****"
      },
      .
      .
      .
      "Band": {
        "band": "****",
        "use": "****",
        "sensor": "****"
      },
    ]
    },

    "MixModel_Components":{
      [
        "Component": {
          "name": "****",
          "coordGrid": {
            "xGrid": "****",
            "yGrid": "****"
          },
          "coordGeo": {
            "xGeo": "****",
            "yGeo": "****"
          },
          "Color": "****"
          },
        .
        .
        .
        "Component": {
          "name": "****",
          "coordGrid": {
            "xGrid": "****",
            "yGrid": "****"
          },
          "coordGeo": {
            "xGeo": "****",
            "yGeo": "****"
          },
          "Color":"****"
        },
      ]
    },

    "Mixture Model": {
      "Method": "****"
    },

    "Transform Matrix": [
      [
        "****",
        .
        .
        .
        "****"
      ],
      .
      .
      .
      [
        "****",
        .
        .
        .
        "****"
      ]
    ]
  }


  struct MixModelComponent
  {
    std::string m_name;
    te::gm::Geometry* m_geomGrid;
    te::gm::Geometry* m_geomGeo;
    std::vector<double> m_values;
    te::color::RGBAColor m_color;
  };
*/

#ifndef __TERRALIB_QT_WIDGETS_RP_INTERNAL_MIXTUREMODELWIZARDPAGE_H
#define __TERRALIB_QT_WIDGETS_RP_INTERNAL_MIXTUREMODELWIZARDPAGE_H

// TerraLib
#include "../../../classification/ROISet.h"
#include "../../../color/RGBAColor.h"
#include "../../../geometry/Geometry.h"
#ifndef Q_MOC_RUN
#include "../../../maptools/AbstractLayer.h"
#include "../../../qt/widgets/canvas/MapDisplay.h"
#include "../../../qt/widgets/charts/ChartDisplay.h"
#include "../../../qt/widgets/charts/ChartDisplayWidget.h"
#include "../../../qt/widgets/charts/ChartStyle.h"
#include "../../../raster/Raster.h"
#include "../../../rp/Functions.h"
#include "../../../rp/MixtureModel.h"
#endif
#include "../Config.h"


// STL
#include <memory>

// Qt
#include <QWizardPage>
#include <QTableWidget>
#include <QActionGroup>
#include <qgridlayout.h>
#include <QSignalMapper>
#include <QTreeWidget>
#include <QTreeWidgetItem>

//QWT
#include <qwt_plot_curve.h>

// Forward declaration
namespace Ui { class MixtureModelWizardPageForm; }

namespace te
{
  namespace se { class Mark; }

  namespace qt
  {
    namespace widgets
    {
      class RpToolsWidget;
      class SensorManagerDialog;
      class ROIManagerDialog;

      /*!
        \class MixtureModelWizardPage

        \brief This class is GUI used to define the mixture model parameters for the RP constast operation.
      */
      class TEQTWIDGETSEXPORT MixtureModelWizardPage : public QWizardPage
      {
        Q_OBJECT

          //!\enum MixModelTypes
          //!\brief Available Algorithms for Mixture Model
          enum MixModelTypes
          {
            MIXMODEL_LINEAR, //!< Linear (Least Square Method).
            MIXMODEL_PCA,    //!< Principal Components Mixture Model.
            MIXMODEL_LINEAR_RLS3T    //!< Linear (Resticted Least Square Method (only 3 components, 4 bands).
          };

        //!\struct MixModelComponent
        //!\brief Defines component attributes
          struct MixModelComponent
          {
            std::string m_name; //!< name
            te::gm::Geometry *m_geomGrid; //!< component coordinades in raster lineXcolumn
            te::gm::Geometry *m_geomGeo;  //!< component coordinates in geografic coordinates
            std::vector<double> m_values; //!< raster values
            te::color::RGBAColor m_color; //!< color
          };

        public:

          MixtureModelWizardPage(QWidget* parent = 0);

          ~MixtureModelWizardPage();

         bool isComplete() const;

        public:

          /*!
            \brief This method is used to set the selected layer for mixture model operation
            \param layer The layer ptr 
            \note This layer MUST HAVE a valid raster object.
          */
          void set(std::list<te::map::AbstractLayerPtr> layers);

          void setMapDisplay(te::qt::widgets::MapDisplay* mapDisplay);

          void setActionGroup(QActionGroup* actionGroup);

          std::list<te::map::AbstractLayerPtr> get();

          te::rp::MixtureModel::InputParameters getInputParams();

          bool isDecompose();

          te::rp::MixtureModel::OutputParameters getOutputParams();

          void saveMixtureModelComponents();

          void loadMixtureModelComponents(const std::string &fileName);

          void updateComponents();

          void clearCanvas();

          void onMapDisplayExtentChanged();

          boost::numeric::ublas::matrix<double> getTransfMatrix() {
            return m_transfMatrix;
          }

          te::cl::ROISet* getROISet();

          public slots:

          void onMixturetypeChanged();

          void onMixturetabChanged(int);

          void onSaveToolButtonClicked();

          void onLoadToolButtonClicked();

          void onPointPicked(double x, double y);

          void onEnvelopeAcquired(te::gm::Envelope env);

          void onGeomAquired(te::gm::Polygon* poly);

          void onRemoveToolButtonClicked();

          void onComponentItemClicked(QTreeWidgetItem * item, int column);
          void oncomponentChanged();
          void oncolorToolButtonClicked();
          void onselectedEnabled(bool);
          void onallEnabled(bool);

          void drawGeom();

          void onSensorToolButtonClicked(int);
          void onBandItemClicked();

          void showROIManager(bool show);

          void onROIManagerClosed(te::cl::ROISet* rs);

      protected:

          void fillMixtureModelTypes();

          void listBands();

          void drawMarks();

          void PlotSpectralSignature();

          double GetMediumWavelength(std::string sensor);

          void addGeometryComponent();
          void addPointComponent();
          void clearComponents();
          
          void printTransMatrix() const;

      private:
         std::unique_ptr<Ui::MixtureModelWizardPageForm> m_ui;

         QSignalMapper m_ButtonSignalMapper;

          te::qt::widgets::SensorManagerDialog *m_sensorManagerDialog; //!< Dialog to select/edit spectral sensors
          std::map<std::string, te::rp::SpectralSensorParams > m_SensorParams; //!< map with sensor parameters

          std::unique_ptr<te::qt::widgets::RpToolsWidget> m_navigator;
          std::unique_ptr<te::qt::widgets::ROIManagerDialog> m_roiMngDlg;

          int m_bandindex;

          std::map<std::string, MixModelComponent > m_components;   //!< The map of selected components
          unsigned int m_countComponents;                           //!< The maximum number of components inserted.

          std::list<te::map::AbstractLayerPtr>  m_layers; //!< list of input layers
          std::vector<double> m_maxValue; //!< vector to store maximum value of each band
          te::color::RGBAColor** m_rgbaMark;  //!< component color
          te::se::Mark* m_mark; //!< component point mark type

          te::qt::widgets::MapDisplay* m_mapDisplay;

          te::qt::widgets::ChartDisplayWidget *m_displayWidget;
          QGridLayout* m_layoutg;
          QColor m_color; //!< Current component color
          te::gm::Geometry* m_geom; //!< Current geometry/envelope component
          std::string m_comp; //!< Current component

          std::string m_json_out;  //!< Json filename to save components
          boost::numeric::ublas::matrix<double> m_transfMatrix;  //!< Transformation matrix;
          int m_type; //!< algorithm mixture model type (MIXMODEL_LINEAR or MIXMODEL_PCA)

          te::cl::ROISet* m_rs;
     };

    } // end namespace widgets
  }   // end namespace qt
}     // end namespace te

#endif  // __TERRALIB_QT_WIDGETS_RP_INTERNAL_MIXTUREMODELWIZARDPAGE_H
