/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

This file is part of the TerraLib - a Framework for building GIS enabled applications.

TerraLib is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

TerraLib is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with TerraLib. See COPYING. If not, write to
TerraLib Team at <terralib-team@terralib.org>.
*/

//Terralib
#include "../../../datatype/DataType.h"
#include "../../../dataaccess/utils/Utils.h"
#include "../../../raster/Band.h"
#include "../../../raster/Raster.h"
#include "../../../se/ColorMap.h"
#include "../charts/ChartDisplay.h"
#include "../charts/ChartStyle.h"
#include "../charts/Histogram.h"
#include "../charts/HistogramChart.h"
#include "../charts/HistogramStyle.h"
#include "../charts/Utils.h"
#include "../utils/ScopedCursor.h"
#include "PCAWizardPage.h"
#include "../se/ui/ui_SlicingColorMapForm.h"
#include "ui/ui_PCAWizardPageForm.h"

// QT
#include <QComboBox>
#include <QString>
#include <QFileDialog>

#include <algorithm>

te::qt::widgets::PCAWizardPage::PCAWizardPage(QWidget *parent) :
  QWizardPage(parent)
{
  m_ui.reset(new Ui::PCAWizardPageForm);

  m_ui->setupUi(this);

  //configure page
  this->setTitle(tr("Principal Components Analysis"));
  this->setSubTitle(tr("Define PCA parameters."));

  m_ui->m_outputDataTypeComboBox->addItem( "Unsigned Int 8 bits" );
  m_ui->m_outputDataTypeComboBox->addItem( "Unsigned Int 16 bits" );
  m_ui->m_outputDataTypeComboBox->addItem( "Unsigned Int 32 bits" );
  m_ui->m_outputDataTypeComboBox->addItem( "Float" );
  m_ui->m_outputDataTypeComboBox->addItem( "Double" );

  m_ui->m_outputDataTypeComboBox->setCurrentIndex( 4 );

  connect(m_ui->m_inversePCACheckBox, SIGNAL(toggled(bool)), this, SLOT(inversePCACheckBoxToggled(bool)));
  connect(m_ui->m_loadPCAMatrixFilePushButton, SIGNAL(clicked(bool)), this, SLOT(loadPCAMatrixFilePushButtonClicked(bool)));
}

te::qt::widgets::PCAWizardPage::~PCAWizardPage() = default;

std::vector< unsigned int > te::qt::widgets::PCAWizardPage::getSelectedBands() const
{
  std::vector< unsigned int > returnVector;

  QList<QListWidgetItem *> selectedItens = m_ui->m_layerBandsListWidget->selectedItems();

  QList<QListWidgetItem *>::iterator it = selectedItens.begin();
  QList<QListWidgetItem *>::iterator itEnd = selectedItens.end();

  while( it != itEnd )
  {
    returnVector.push_back( (*it)->text().toUInt() );
    ++it;
  }
  
  std::sort( returnVector.begin(), returnVector.end() );

  return returnVector;
}

std::string te::qt::widgets::PCAWizardPage::getMatrixFileName() const
{
  return m_ui->m_pcaMatrixFileLineEdit->text().toStdString();
}

int te::qt::widgets::PCAWizardPage::getOutputRasterDataType() const
{
  switch( m_ui->m_outputDataTypeComboBox->currentIndex() )
  {
    case 0 : // "Unsigned Int 8 bits"
    {
      return te::dt::UCHAR_TYPE;
      break;
    }
    case 1 : // "Unsigned Int 16 bits"
    {
      return te::dt::UINT16_TYPE;
      break;
    }
    case 2 : // "Unsigned Int 32 bits"
    {
      return te::dt::UINT32_TYPE;
      break;
    }
    case 3 : // "Float"
    {
      return te::dt::FLOAT_TYPE;
      break;
    }
    default : // "Double"
    {
      return te::dt::DOUBLE_TYPE;
      break;
    }
  }

  return te::dt::DOUBLE_TYPE;
}

void te::qt::widgets::PCAWizardPage::setLayer(te::map::AbstractLayerPtr layer)
{
  m_rasterLayer = layer;

  //get input raster
  std::unique_ptr<te::da::DataSet> ds = m_rasterLayer->getData();

  if (ds.get())
  {
    std::size_t rpos = te::da::GetFirstPropertyPos(ds.get(), te::dt::RASTER_TYPE);

    std::unique_ptr<te::rst::Raster> inputRst = ds->getRaster(rpos);

    if (inputRst.get())
    {
      m_raster.reset(inputRst.release());

      size_t bandsNumber = m_raster->getNumberOfBands();

      m_ui->m_layerBandsListWidget->clear();
      for (size_t band = 0; band < bandsNumber; ++band)
      {
        m_ui->m_layerBandsListWidget->addItem(QString::number(band));
      }

      m_ui->m_layerBandsListWidget->selectAll();
    }
  }
}

bool te::qt::widgets::PCAWizardPage::isComplete() const
{
  QList<QListWidgetItem *> selectedItens = m_ui->m_layerBandsListWidget->selectedItems();

  if (
        ( selectedItens.empty() )
        ||
        (
          ( m_ui->m_inversePCACheckBox->checkState() == Qt::Checked )
          &&
          ( m_ui->m_pcaMatrixFileLineEdit->text().isEmpty() )
        )
     )
  {
    return false;
  }
  else
  {
    return true;
  }
}

bool te::qt::widgets::PCAWizardPage::inversePCASelected() const
{
  return ( m_ui->m_inversePCACheckBox->checkState() == Qt::Checked );
}

void te::qt::widgets::PCAWizardPage::inversePCACheckBoxToggled(bool checked)
{
  m_ui->m_pcaMatrixFileLineEdit->clear();

  if( checked )
  {
    m_ui->m_loadPCAMatrixFilePushButton->setText( tr( "Load" ) );

    m_ui->m_outputDataTypeComboBox->setEnabled( true );
  }
  else
  {
    m_ui->m_loadPCAMatrixFilePushButton->setText( tr( "Save" ) );

    m_ui->m_outputDataTypeComboBox->setCurrentIndex( 4 );
    m_ui->m_outputDataTypeComboBox->setEnabled( false );
  }

  emit completeChanged();
}

void te::qt::widgets::PCAWizardPage::loadPCAMatrixFilePushButtonClicked(
    bool /*checked*/)
{
  QString fileName;

  if( m_ui->m_inversePCACheckBox->checkState() == Qt::Checked )
  {
    fileName = QFileDialog::getOpenFileName(this, tr("Select File"),
      ".", "*", nullptr, QFileDialog::ReadOnly);
  }
  else
  {
    fileName = QFileDialog::getSaveFileName(this, tr("Select the output file name"),
      ".", "*", nullptr , QFileDialog::DontConfirmOverwrite);
  }

  m_ui->m_pcaMatrixFileLineEdit->setText( fileName );

  emit completeChanged();
}

