/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
\file terralib/qt/widgets/rp/PaletteDialog.cpp

\brief This interface is used to create palette raster from rgb raster.
*/

//Terralib
#include "PaletteDialog.h"
#include "ui/ui_PaletteDialogForm.h"

#include <terralib/core/translator/Translator.h>
#include <terralib/dataaccess/utils/Utils.h>
#include <terralib/maptools/AbstractLayer.h>
#include <terralib/maptools/RasterContrast.h>
#include <terralib/maptools/RasterTransform.h>
#include <terralib/maptools/RasterTransformConfigurer.h>
#include <terralib/maptools/Utils.h>
#include <terralib/qt/widgets/Exception.h>
#include <terralib/qt/widgets/rp/Utils.h>
#include <terralib/raster/Band.h>
#include <terralib/raster/BandProperty.h>
#include <terralib/raster/Grid.h>
#include <terralib/raster/Raster.h>
#include <terralib/raster/RasterFactory.h>
#include <terralib/rp/PaletteRaster.h>
#include <terralib/rp/Macros.h>
#include <terralib/se/Categorize.h>
#include <terralib/se/ChannelSelection.h>
#include <terralib/se/ColorMap.h>
#include <terralib/se/CoverageStyle.h>
#include <terralib/se/ParameterValue.h>
#include <terralib/se/RasterSymbolizer.h>
#include <terralib/se/Rule.h>
#include <terralib/se/SelectedChannel.h>
#include <terralib/se/Style.h>
#include <terralib/se/Utils.h>

//Qt
#include <QMessageBox>

// Boost
#include <boost/filesystem.hpp>
#include <boost/format.hpp>
#include <boost/lexical_cast.hpp>

te::qt::widgets::PaletteDialog::PaletteDialog(QWidget* parent, Qt::WindowFlags f)
  : QDialog(parent, f),
  m_ui(new Ui::PaletteDialogForm)
{
  // add controls
  m_ui->setupUi(this);

  //build output parameters
  QGridLayout* outputLayout = new QGridLayout(m_ui->m_outputWidget);
  outputLayout->setContentsMargins(0, 0, 0, 0);
  m_outputWidget = new te::qt::widgets::OutputRasterWidget(m_ui->m_outputWidget);
  outputLayout->addWidget(m_outputWidget);

//connects
  connect(m_ui->m_okPushButton, SIGNAL(clicked()), this, SLOT(onOkPushButtonClicked()));
  connect(m_ui->m_cancelPushButton, SIGNAL(clicked()), this, SLOT(onCancelPushButtonClicked()));

  m_ui->m_palletesizeSpinBox->setValue(256);

  m_ui->m_helpPushButton->setNameSpace("dpi.inpe.br.plugins");
  m_ui->m_helpPushButton->setPageReference("plugins/rp/rp_palette.html");

}

te::qt::widgets::PaletteDialog::~PaletteDialog()
{
}

te::map::AbstractLayerPtr te::qt::widgets::PaletteDialog::getLayer()
{
  return m_layerResult;
}

void te::qt::widgets::PaletteDialog::setParameters(te::map::AbstractLayerPtr layer)
{
  m_layer = layer;

  std::unique_ptr<te::da::DataSetType> dsType = m_layer->getSchema();
  if (!dsType->hasRaster())
  {
    throw te::common::Exception(TE_TR("There is no selected raster layer."));
  }

  te::rst::RasterProperty* rasterProp = te::da::GetFirstRasterProperty(dsType.get());

  if (rasterProp->getBandProperties().empty())
  {
    throw te::common::Exception(TE_TR("Layers without Band information!"));
  }

  if (rasterProp->getBandProperties()[0]->m_colorInterp == te::rst::PaletteIdxCInt)
  {
    throw te::common::Exception(TE_TR("Layers with raster palette are not necessary to apply palette!"));
  }
}

void te::qt::widgets::PaletteDialog::onOkPushButtonClicked()
{
  std::vector< te::rst::BandProperty::ColorEntry > palette;

  try
  {
    //check output parameters
    if (m_outputWidget->getOutputRaster().empty())
    {
      throw te::common::Exception(TE_TR("Output image is not defined."));
    }

    std::string errorMessage;

    this->setCursor(Qt::WaitCursor);

    std::unique_ptr<te::rst::Raster> InRaster(te::map::GetRaster(m_layer.get()));

    // Gets the layer schema
    std::unique_ptr<te::map::LayerSchema> schema(m_layer->getSchema());

    // Gets the name of the referenced spatial property
    std::string spatialPropertyName = m_layer->getGeomPropertyName();

    te::rst::RasterProperty* rasterProperty = nullptr;

    if (spatialPropertyName.empty())
      rasterProperty = te::da::GetFirstRasterProperty(schema.get());
    else
      rasterProperty = dynamic_cast<te::rst::RasterProperty*>(schema->getProperty(spatialPropertyName));

    assert(rasterProperty);

    // Gets the layer style
    te::se::Style* style = m_layer->getStyle();
    if (style == nullptr)
    {
      // Try create an appropriate style. i.e. a CoverageStyle
      style = te::se::CreateCoverageStyle(rasterProperty->getBandProperties());

      if (style == nullptr)
        throw te::common::Exception((boost::format(TE_TR("Could not create a default coverage style to the layer %1%.")) % m_layer->getTitle()).str());

      m_layer->setStyle(style);
    }
    // Should I render this style?
    te::se::CoverageStyle* cs = dynamic_cast<te::se::CoverageStyle*>(style);
    if (cs == nullptr)
      throw te::common::Exception(TE_TR("The layer style is not a Coverage Style!"));

    // create a raster transform
    te::map::RasterTransform rasterTransform(InRaster.get(), nullptr);

    //check band data type
    if (InRaster->getBandDataType(0) != te::dt::UCHAR_TYPE)
    {
      te::rst::RasterSummary rasterSummary;
      if( ! te::rst::RasterSummaryManager::getInstance().get(
        InRaster.get(), false, rasterSummary ) )
      {
        throw te::common::Exception(TE_TR("Raster summary Error"));
      }          
      
      // raster min / max values
      for (int b = 0; b < static_cast<int>(InRaster->getNumberOfBands()); b++)
      {
        const std::complex<double>* cmin = rasterSummary.at(b).m_minVal;
        const std::complex<double>* cmax = rasterSummary.at(b).m_maxVal;
        double min = cmin->real();
        double max = cmax->real();

        rasterTransform.setLinearTransfParameters(min, max, 0, 255, b);
      }
    }
    else
    {
      rasterTransform.setLinearTransfParameters(0, 255, 0, 255);
    }

    // for while, consider one rule
    const te::se::Rule* rule = style->getRule(0);

    const std::vector<te::se::Symbolizer*>& symbolizers = rule->getSymbolizers();
    assert(!symbolizers.empty());

    // for while, consider one raster symbolizer
    te::se::RasterSymbolizer* rasterSymbolizer = dynamic_cast<te::se::RasterSymbolizer*>(symbolizers[0]);
    assert(rasterSymbolizer);

    // configure the raster transformation based on the raster symbolizer
    te::map::RasterTransformConfigurer rtc(rasterSymbolizer, &rasterTransform);
    rtc.configure();

    te::map::RasterContrast* rc = m_layer->getRasterContrast();

    if (rc)
    {
      int type = rc->getType();

      std::vector<double> gain;
      std::vector<double> offset1;
      std::vector<double> offset2;
      std::vector<double> min;
      std::vector<double> max;
      rc->getValues(gain, offset1, offset2, min, max);

      rasterTransform.setContrastType(static_cast<te::map::RasterTransform::ContrastType>(type));

      for (std::size_t t = 0; t < min.size(); ++t)
      {
        rasterTransform.setLinearTransfParameters(min[t], max[t], 0, 255, static_cast<int>(t));
      }
    }

    std::vector<te::rst::BandProperty*> bandsPropertiestmp;
    te::rst::BandProperty* tmpbprop = new te::rst::BandProperty(*(InRaster->getBand(
      0)->getProperty()));
    tmpbprop->m_type = te::dt::UINT16_TYPE;
    tmpbprop->m_noDataValue = 0;
    bandsPropertiestmp.push_back(tmpbprop);

    std::map<std::string, std::string> rinfotmp;
    rinfotmp["MEM_RASTER_NROWS"] = boost::lexical_cast<std::string>(InRaster->getNumberOfRows());
    rinfotmp["MEM_RASTER_NCOLS"] = boost::lexical_cast<std::string>(InRaster->getNumberOfColumns());
    rinfotmp["MEM_RASTER_DATATYPE"] = boost::lexical_cast<std::string>(InRaster->getBandDataType(0));
    rinfotmp["MEM_RASTER_NBANDS"] = 1;

    std::unique_ptr<te::rst::Raster> TmpRaster;
    try
    {
      TmpRaster.reset(te::rst::RasterFactory::make(
        "MEM",
        new te::rst::Grid(*(InRaster->getGrid())),
        bandsPropertiestmp,
        rinfotmp,
        nullptr,
        nullptr));
    }
    catch(...)
    {
      TERP_LOG_AND_THROW("Output raster creation error");
    }
    TERP_TRUE_OR_THROW(TmpRaster.get() != nullptr, "Raster creation error");    

    int palettesize = m_ui->m_palletesizeSpinBox->value();

    te::rp::Codifima cod(*InRaster, rasterTransform, *TmpRaster, palettesize);
    cod.Apply();

    std::vector<te::rst::BandProperty*> bandsProperties;
    te::rst::BandProperty* bprop = new te::rst::BandProperty(*(InRaster->getBand(
      0)->getProperty()));
    bprop->m_colorInterp = te::rst::PaletteIdxCInt;
    if (palettesize <= 256)
     bprop->m_type = te::dt::UCHAR_TYPE;
    else
      bprop->m_type = te::dt::UINT16_TYPE;
    bprop->m_paletteInterp = te::rst::RGBPalInt;
    bprop->m_palette = cod.getPalette();
    bprop->m_noDataValue = 0;
    bandsProperties.push_back(bprop);

    palette = cod.getPalette();

    std::unique_ptr<te::rst::Raster> OutRaster;
    try
    {
      OutRaster.reset(te::rst::RasterFactory::make(
        m_outputWidget->getType(),
        new te::rst::Grid(*(InRaster->getGrid())),
        bandsProperties,
        m_outputWidget->getInfo(),
        nullptr,
        nullptr));
    }
    catch(...)
    {
      TERP_LOG_AND_THROW("Output raster creation error");
    }
    TERP_TRUE_OR_THROW(OutRaster.get() != nullptr, "Output raster creation error");

    // fill the result RGBA array
    for (unsigned int r = 0; r < OutRaster->getNumberOfRows(); ++r)
    {
      for (unsigned int c = 0; c < OutRaster->getNumberOfColumns(); ++c)
      {
        std::complex<double> value;
        TmpRaster->getValue(c, r, value, 0);
        OutRaster->setValue(c, r, value.real(), 0);
      }
    }

  }
  catch (te::common::Exception &e)
  {
    throw e;
  }

  //set output layer
  m_layerResult = te::qt::widgets::createLayer(m_outputWidget->getType(),
    m_outputWidget->getInfo());

  //set legend
  te::se::RasterSymbolizer* rasterSymb = te::se::GetRasterSymbolizer(m_layerResult->getStyle());

  te::se::ColorMap* cm = new te::se::ColorMap();
  te::se::Categorize* c = new te::se::Categorize();
  c->setFallbackValue("#000000");
  c->setLookupValue(new te::se::ParameterValue("Rasterdata"));

  QColor cWhite(Qt::white);
  std::string colorWhiteStr = cWhite.name().toUtf8().data();

  //added dummy color for values < than min values...
  c->addThreshold(new te::se::ParameterValue("0"));
  c->addValue(new te::se::ParameterValue(colorWhiteStr));

  for (unsigned int i = 0; i < palette.size(); ++i)
  {
    QColor color(palette[i].c1, palette[i].c2,palette[i].c3);

    std::string rangeStr = std::to_string(i+1);
    std::string colorStr = color.name().toUtf8().data();

    c->addThreshold(new te::se::ParameterValue(rangeStr));
    c->addValue(new te::se::ParameterValue(colorStr));
  }

  //added dummy color for values > than max values...
  std::string rangeStr = std::to_string(palette.size()+1);
  c->addThreshold(new te::se::ParameterValue(rangeStr));
  c->addValue(new te::se::ParameterValue(colorWhiteStr));

  c->setThresholdsBelongTo(te::se::Categorize::SUCCEEDING);
  cm->setCategorize(c);
  cm->setInterpolate(nullptr);
  cm->setRecode(nullptr);

  rasterSymb->setColorMap(cm);
  te::se::ChannelSelection* cs = rasterSymb->getChannelSelection();
  te::se::SelectedChannel* scGray = cs->getGrayChannel();

  scGray->setSourceChannelName("0");
  cs->setColorCompositionType(te::se::GRAY_COMPOSITION);

  ////

  accept();
}

void te::qt::widgets::PaletteDialog::onCancelPushButtonClicked()
{
  reject();
}
