/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/widgets/rp/PostClassificationWidget.cpp

  \brief This file has the PostClassificationWidget class.
*/

// TerraLib
#include "../../../core/translator/Translator.h"
#include "../../../common/StringUtils.h"
#include "../../../common/progress/ProgressManager.h"
#include "../../../dataaccess/utils/Utils.h"
#include "../../../geometry/Utils.h"
#include "../../../maptools/Utils.h"
#include "../../../qt/widgets/utils/ScopedCursor.h"
#include "../../../raster/BandProperty.h"
#include "../../../raster/RasterProperty.h"
#include "../../../rp/Module.h"
#include "../../../rp/PostClassification.h"
#include "../../../se/CoverageStyle.h"
#include "../../../se/RasterSymbolizer.h"
#include "../../../se/Utils.h"
#include "../progress/ProgressViewerDialog.h"
#include "PostClassificationWidget.h"
#include "Utils.h"
#include "ui/ui_PostClassificationWidgetForm.h"

//Qt
#include <QMessageBox>

te::qt::widgets::PostClassificationWidget::PostClassificationWidget(QWidget* parent, Qt::WindowFlags f)
  : QWidget(parent, f),
  m_ui(new Ui::PostClassificationWidgetForm),
  m_layer(nullptr)
{
  m_ui->setupUi(this);

  m_ui->m_weightLineEdit->setText("1");
  m_ui->m_thresholdLineEdit->setText("1");

  //build input layer
  QGridLayout* inputLayout = new QGridLayout(m_ui->m_inputWidget);
  inputLayout->setContentsMargins(0, 0, 0, 0);
  m_inputWidget = new te::qt::widgets::InputLayerWidget(m_ui->m_inputWidget);
  inputLayout->addWidget(m_inputWidget);

  //build output parameters
  QGridLayout* outputLayout = new QGridLayout(m_ui->m_outputWidget);
  outputLayout->setContentsMargins(0, 0, 0, 0);
  m_outputWidget = new te::qt::widgets::OutputRasterWidget(m_ui->m_outputWidget);
  outputLayout->addWidget(m_outputWidget);

  //connects
  connect(m_ui->m_weightSlider, SIGNAL(valueChanged(int)), this, SLOT(weightSliderValueChanged(int)));
  connect(m_ui->m_thresholdSlider, SIGNAL(valueChanged(int)), this, SLOT(thresholdSliderValueChanged(int)));
}

te::qt::widgets::PostClassificationWidget::~PostClassificationWidget() =
    default;

Ui::PostClassificationWidgetForm* te::qt::widgets::PostClassificationWidget::getForm() const
{
  return m_ui.get();
}

void te::qt::widgets::PostClassificationWidget::setList(std::list<te::map::AbstractLayerPtr>& layerList)
{
  std::list<te::map::AbstractLayerPtr> layersRasterList;

  std::list<te::map::AbstractLayerPtr>::iterator it = layerList.begin();
  //Filter only raster layers
  while (it != layerList.end())
  {
    std::unique_ptr<te::da::DataSetType> dsType = it->get()->getSchema();

    if (dsType->hasRaster())
    {
      te::se::RasterSymbolizer* rs = te::se::GetRasterSymbolizer(it->get()->getStyle());

      if(rs && rs->getColorMap())
        layersRasterList.push_back(it->get());
    }

    ++it;
  }

  if(!layersRasterList.empty())
  {
    m_inputWidget->setLayerList(layersRasterList);

    m_layer = layersRasterList.begin()->get();
  }
}

bool te::qt::widgets::PostClassificationWidget::execute()
{
  if(!m_layer)
  {
    QMessageBox::warning(this, tr("Post Classification"), tr("Input image is not defined."));
    return false;
  }

  //check output parameters
  if (m_outputWidget->getOutputRaster().empty())
  {
    QMessageBox::warning(this, tr("Post Classification"), tr("Output image is not defined."));
    return false;
  }

  //get layer
  std::unique_ptr<te::da::DataSet> ds = m_layer->getData();

  std::size_t rpos = te::da::GetFirstPropertyPos(ds.get(), te::dt::RASTER_TYPE);

  std::unique_ptr<te::rst::Raster> inputRst = ds->getRaster(rpos);

  //run post classification
  te::rp::PostClassification algorithmInstance;

  te::rp::PostClassification::InputParameters algoInputParams;
  algoInputParams.m_inRasterPtr = inputRst.get();
  algoInputParams.m_weight = (unsigned int)m_ui->m_weightLineEdit->text().toUtf8().toInt();
  algoInputParams.m_threshold = (unsigned int)m_ui->m_thresholdLineEdit->text().toUtf8().toInt();

  te::rp::PostClassification::OutputParameters algoOutputParams;
  algoOutputParams.m_rInfo = m_outputWidget->getInfo();
  algoOutputParams.m_rType = m_outputWidget->getType();

  try
  {
  //progress
  te::qt::widgets::ProgressViewerDialog v(this);
    te::qt::widgets::ScopedCursor cursor(Qt::WaitCursor);

  if (algorithmInstance.initialize(algoInputParams))
  {
    if (algorithmInstance.execute(algoOutputParams))
    {
      algoOutputParams.reset();

      //set output layer
      m_outputLayer = te::qt::widgets::createLayer(m_outputWidget->getType(),
        m_outputWidget->getInfo());

      m_outputLayer->setStyle(m_layer->getStyle()->clone());
    }
    else
    {
        std::string what = TE_TR("Post Classification execution error.");
        what += " " + te::rp::Module::getLastLogStr();
        throw te::common::Exception(what);
    }
  }
  else
  {
      std::string what = TE_TR("Post Classification initialization error.");
      what += " " + te::rp::Module::getLastLogStr();
      throw te::common::Exception(what);
    }
  }
  catch (const std::exception& e)
  {
    QMessageBox::critical(this, tr("Post Classification"), e.what());
    return false;
  }

  QMessageBox::information(this, tr("Post Classification"), tr("Post Classification ended sucessfully."));

  emit addLayer(m_outputLayer);

  return true;
}

void te::qt::widgets::PostClassificationWidget::weightSliderValueChanged(int value)
{
  m_ui->m_weightLineEdit->setText(te::common::Convert2String(value).c_str());
}

void te::qt::widgets::PostClassificationWidget::thresholdSliderValueChanged(int value)
{
  m_ui->m_thresholdLineEdit->setText(te::common::Convert2String(value).c_str());
}
