/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/widgets/rp/RasterEditWizard.cpp

  \brief A Qt dialog that allows users to run a clipping operation defined by RP module.
*/

// TerraLib
#include "../../../common/progress/ProgressManager.h"
#include "../../../geometry/Envelope.h"
#include "../../../dataaccess/dataset/DataSet.h"
#include "../../../dataaccess/utils/Utils.h"
#include "../../../geometry/Geometry.h"
#include "../../../geometry/GeometryCollection.h"
#include "../../../raster/Grid.h"
#include "../../../raster/Interpolator.h"
#include "../../../raster/PositionIterator.h"
#include "../../../raster/Raster.h"
#include "../../../raster/RasterFactory.h"
#include "../../../raster/Utils.h"
#include "../../../raster/BandProperty.h"
#include "../../../maptools/Utils.h"
#include "../../../rp/Module.h"
#include "../../../rp/Functions.h"
#include "../../../se/ColorMap.h"
#include "../../../se/Recode.h"
#include "../../../se/MapItem.h"
#include "../../../se/RasterSymbolizer.h"
#include "../../../se/ParameterValue.h"
#include "../../../se/Utils.h"
#include "../../../color/RGBAColor.h"
#include "../help/HelpPushButton.h"
#include "../layer/search/LayerSearchWidget.h"
#include "../layer/search/LayerSearchWizardPage.h"
#include "../progress/ProgressViewerDialog.h"
#include "../utils/ScopedCursor.h"
#include "RasterEditWizard.h"
#include "RasterEditWizardPage.h"
#include "../raster/RasterInfoWidget.h"
#include "RasterInfoWizardPage.h"
#include "Utils.h"

// STL
#include <cassert>

// Qt
#include <QApplication>
#include <QMessageBox>
#include <QActionGroup>

te::qt::widgets::RasterEditWizard::RasterEditWizard(QWidget* parent,
  te::qt::widgets::MapDisplay* mainMapDispPtr,
  te::qt::widgets::LayerItemView* mainLayerExplorerPtr )
  : QWizard(parent), m_mainMapDispPtr( mainMapDispPtr ),
    m_mainLayerExplorerPtr( mainLayerExplorerPtr )
{
  //configure the wizard
  this->setWizardStyle(QWizard::ModernStyle);
  this->setWindowTitle(tr("Raster Edit"));
  //this->setFixedSize(640, 580);

  this->setOption(QWizard::HaveHelpButton, true);
  this->setOption(QWizard::HelpButtonOnRight, false);

  te::qt::widgets::HelpPushButton* helpButton = new te::qt::widgets::HelpPushButton(this);

  this->setButton(QWizard::HelpButton, helpButton);

  helpButton->setPageReference("plugins/rp/rp_rasteredit.html");

  connect(this, SIGNAL(currentIdChanged(int)), SLOT(onPageChanged(int)));

  addPages();
}

te::qt::widgets::RasterEditWizard::~RasterEditWizard()
{
}

bool te::qt::widgets::RasterEditWizard::validateCurrentPage()
{
  if(currentPage() ==  m_layerSearchPage.get())
  {
    std::list<te::map::AbstractLayerPtr> list = m_layerSearchPage->getSearchWidget()->getSelecteds();

    if(list.empty() == false)
    {
      te::map::AbstractLayerPtr l = *list.begin();

      if( !m_rasterEditPage->setLayer(l) )return false;
    }

    return m_layerSearchPage->isComplete();
  }
  else if(currentPage() ==  m_rasterEditPage.get())
  {
    bool res = m_rasterEditPage->isComplete();

    if(!res)
      QMessageBox::warning(this, tr("Warning"), tr("Incomplete parameters"));

    return res;
  }
  else if(currentPage() ==  m_rasterInfoPage.get())
  {
    return execute();
  }

  return true;
}

void te::qt::widgets::RasterEditWizard::setList(std::list<te::map::AbstractLayerPtr>& layerList)
{
  m_layerSearchPage->getSearchWidget()->setList(layerList);
  m_layerSearchPage->getSearchWidget()->filterOnlyByRaster();
}

void te::qt::widgets::RasterEditWizard::setLayer(te::map::AbstractLayerPtr layer)
{
  m_rasterEditPage->setLayer(layer);
}

void te::qt::widgets::RasterEditWizard::setActionGroup(QActionGroup* actionGroup)
{
}

te::map::AbstractLayerPtr te::qt::widgets::RasterEditWizard::getOutputLayer()
{
  return m_outputLayer;
}

void te::qt::widgets::RasterEditWizard::onMapDisplayExtentChanged()
{
}

void te::qt::widgets::RasterEditWizard::addPages()
{
  m_layerSearchPage.reset(new te::qt::widgets::LayerSearchWizardPage(this));
  m_rasterEditPage.reset(new te::qt::widgets::RasterEditWizardPage(this,
    m_mainMapDispPtr, m_mainLayerExplorerPtr ));
  m_rasterInfoPage.reset(new te::qt::widgets::RasterInfoWizardPage(this));

  addPage(m_layerSearchPage.get());
  addPage(m_rasterEditPage.get());
  addPage(m_rasterInfoPage.get());

  //for contrast only one layer can be selected
  m_layerSearchPage->getSearchWidget()->enableMultiSelection(false);
}

bool te::qt::widgets::RasterEditWizard::execute()
{
  std::shared_ptr< te::rst::Raster > outputMEMRasterPtr =
    m_rasterEditPage->getOutputRaster();
  if( outputMEMRasterPtr.get() == 0 )
  {
    QMessageBox::warning(this, tr("Raster Edit"), tr("Invalid output raster."));
    return false;
  }

  if(m_rasterInfoPage->getWidget()->fileExists())
  {
    QMessageBox::warning(this, tr("Raster Edit"), tr("File already exists."));
    return false;
  }

  const std::map< unsigned int , std::string >&
    palleteIndex2Description =  m_rasterEditPage->getPalleteIndex2Descriptions()[ 0 ];

  m_rasterEditPage->hide();
  m_rasterEditPage->setLayer( te::map::AbstractLayerPtr() );

  try
  {
    // Creating the output raster

    std::vector< te::rst::BandProperty * > bandProps;
    for( unsigned int bandIdx = 0 ; bandIdx < outputMEMRasterPtr->getNumberOfBands() ;
      ++bandIdx )
    {
      bandProps.push_back( new te::rst::BandProperty(
        *( outputMEMRasterPtr->getBand( bandIdx )->getProperty() ) ) );
    }

    const std::string rasterType = m_rasterInfoPage->getWidget()->getType();

    std::map<std::string, std::string> rasterInfo =
      m_rasterInfoPage->getWidget()->getInfo();

    te::rst::Grid* gridPtr = new te::rst::Grid( *outputMEMRasterPtr->getGrid() );

    std::unique_ptr< te::rst::Raster > outputRasterPtr;

    try
    {
      te::qt::widgets::ScopedCursor cursor(Qt::WaitCursor);
      outputRasterPtr.reset(te::rst::RasterFactory::make(rasterType, gridPtr, bandProps,
        rasterInfo));
    }
    catch(...)
    {
      outputRasterPtr.reset();
    }
    if( outputRasterPtr.get() == 0 )
    {
      QMessageBox::critical(this, tr("RasterEdit"), tr("Output raster creation error"));
      return false;
    }

    {
      te::qt::widgets::ScopedCursor cursor(Qt::WaitCursor);
      te::rst::Copy( *outputMEMRasterPtr, *outputRasterPtr );
    }

    outputRasterPtr.reset();
  }
  catch(const std::exception& e)
  {
    QMessageBox::warning(this, tr("Raster Edit"), e.what());

    return false;
  }
  catch(...)
  {
    QMessageBox::warning(this, tr("Raster Edit"), tr("An exception has occurred!"));

    return false;
  }

  // Creating the output layer

  const std::vector< te::rst::BandProperty::ColorEntry >& outputPalette =
    outputMEMRasterPtr->getBand( 0 )->getProperty()->m_palette;

  m_outputLayer = te::qt::widgets::createLayer(
    m_rasterInfoPage->getWidget()->getType(),
    m_rasterInfoPage->getWidget()->getInfo());

  te::se::ColorMap* cm = new te::se::ColorMap();

  te::se::Recode* r = new te::se::Recode();
  r->setFallbackValue("#000000");
  r->setLookupValue(new te::se::ParameterValue("Rasterdata"));

  std::map< unsigned int , std::string >::const_iterator palleteIndex2DescriptionIt;
  const std::map< unsigned int , std::string >::const_iterator palleteIndex2DescriptionItEnd =
    palleteIndex2Description.end();

  for( std::size_t outputPaletteIdx = 0 ; outputPaletteIdx < outputPalette.size() ;
    ++outputPaletteIdx )
  {
    const te::rst::BandProperty::ColorEntry& colorEntry = outputPalette[
      outputPaletteIdx ];
    te::color::RGBAColor rgbaColor( colorEntry.c1, colorEntry.c2,
      colorEntry.c3, colorEntry.c4 );

    te::se::MapItem* m = new te::se::MapItem();
    m->setValue(new te::se::ParameterValue( rgbaColor.getColor(
      te::color::RGBAColor::HexRgb ) ) );
    m->setData((double)outputPaletteIdx);

    palleteIndex2DescriptionIt = palleteIndex2Description.find( outputPaletteIdx );
    if( palleteIndex2DescriptionIt == palleteIndex2DescriptionItEnd )
    {
      m->setTitle( boost::lexical_cast< std::string >( outputPaletteIdx ) );
    }
    else
    {
      m->setTitle( palleteIndex2DescriptionIt->second );
    }

    r->add(m);
  }

  cm->setCategorize(nullptr);
  cm->setInterpolate(nullptr);
  cm->setRecode(r);

  te::se::RasterSymbolizer* rasterSymb = te::se::GetRasterSymbolizer(m_outputLayer->getStyle());
  rasterSymb->setColorMap(cm);
  
  te::qt::widgets::applyRasterMultiResolution(tr("Raster Edit"), te::map::GetRaster(m_outputLayer.get()));

  int addOutPutLayer = QMessageBox::question(this, tr("Raster Edit"),
    tr("Would you like to add the created raster layer on active project?"),
    QMessageBox::Yes | QMessageBox::No);
  if (addOutPutLayer == QMessageBox::Yes)
  {
    emit addLayer(m_outputLayer);
  }

  QMessageBox::information(this, tr("Raster Edit"), tr("Raster Edit ended sucessfully."));

  return true;
}

void te::qt::widgets::RasterEditWizard::onPageChanged(int /*index*/)
{
  this->adjustSize();
}
