/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/widgets/rp/RasterEditWizardPage.cpp

  \brief This file defines a class for a RasterEdit Wizard page.
*/

#include "RasterEditWizardPage.h"
#include "ui/ui_RasterEditWizardPageForm.h"
#include "../../../common/StringUtils.h"
#include "../../../dataaccess/dataset/DataSet.h"
#include "../../../dataaccess/dataset/ObjectIdSet.h"
#include "../../../dataaccess/utils/Utils.h"
#include "../../../dataaccess/datasource/DataSourceFactory.h"
#include "../../../dataaccess/datasource/DataSourceManager.h"
#include "../../../geometry/Geometry.h"
#include "../../../geometry/GeometryProperty.h"
#include "../../../geometry/GeometryCollection.h"
#include "../../../geometry/LinearRing.h"
#include "../../../geometry/Polygon.h"
#include "../../../geometry/MultiPolygon.h"
#include "../../../geometry/Utils.h"
#include "../../../raster/Grid.h"
#include "../../../raster/Raster.h"
#include "../../../raster/Band.h"
#include "../../../raster/BandProperty.h"
#include "../../../raster/RasterProperty.h"
#include "../../../raster/RasterFactory.h"
#include "../../../raster/ProxyRaster.h"
#include "../../../raster/Utils.h"
#include "../../../maptools/Utils.h"
#include "../../../maptools/AbstractLayerRenderer.h"
#include "../../../maptools/AbstractLayerRendererFactory.h"
#include "../../../maptools/DataSetLayer.h"
#include "../../../se/CoverageStyle.h"
#include "../../../se/Style.h"
#include "../../../se/Rule.h"
#include "../../../se/Symbolizer.h"
#include "../../../se/RasterSymbolizer.h"
#include "../../../se/ColorMap.h"
#include "../../../se/Recode.h"
#include "../../../se/MapItem.h"
#include "../../../se/ParameterValue.h"
#include "../../../se/Utils.h"
#include "../../../rp/Functions.h"
#include "../canvas/Canvas.h"
#include "../canvas/MapDisplay.h"
#include "../layer/explorer/TreeItem.h"
#include "../layer/explorer/LayerItem.h"
#include "../../../memory/DataSet.h"
#include "../../../memory/DataSource.h"
#include "../../../memory/DataSetItem.h"
#include "../Utils.h"
#include "../../af/BaseApplication.h"
#include "../../af/ApplicationController.h"
#include "RasterNavigatorDialog.h"
#include "RasterNavigatorWidget.h"
#include "../tools/PointPicker.h"
#include "../../widgets/utils/ScopedCursor.h"

// Qt
#include <QGridLayout>
#include <QCheckBox>
#include <QDoubleValidator>
#include <QIntValidator>
#include <QMessageBox>
#include <QBrush>
#include <QColorDialog>
#include <QFileDialog>

// STL
#include <memory>
#include <set>
#include <stack>

Q_DECLARE_METATYPE(te::map::AbstractLayerPtr)

te::qt::widgets::RasterEditWizardPage::RasterEditWizardPage(QWidget* parent,
  te::qt::widgets::MapDisplay* mainMapDispPtr,
  te::qt::widgets::LayerItemView* mainLayerExplorerPtr )
  : QWizardPage(parent),
    m_mainMapDispPtr( mainMapDispPtr ),
    m_mainLayerExplorerPtr( mainLayerExplorerPtr ),
    m_ui(new Ui::RasterEditWizardPageForm)
{
// setup controls
  m_ui->setupUi(this);
  m_ui->m_roiSubstitutionValuesTableWidget->setSortingEnabled(false);
  m_ui->m_roiTableWidget->setSortingEnabled(false);

//configure page
  this->setTitle(tr("Raster Edit"));
  this->setSubTitle(tr("Select the type of EDITION and set their specific parameters."));

// internal connects

  connect(m_ui->m_roiTableWidget, SIGNAL( itemClicked( QTableWidgetItem* ) ), this,
    SLOT(onRoiTableWidgetItemClicked( QTableWidgetItem* ) ) );
  connect(m_ui->m_removeRoiPushButton, SIGNAL(clicked()), this, SLOT(onRemoveRoiItem()));
  connect(m_ui->m_addRulePushButton, SIGNAL(clicked()), this, SLOT(addRulePushButtonClicked()));
  connect(m_ui->m_removeRulePushButton, SIGNAL(clicked()), this, SLOT(removeRulePushButtonClicked()));
  connect(m_ui->m_replaceAllPushButton, SIGNAL(clicked()), this, SLOT(onReplaceAllPushButtonClicked()));
  connect(m_ui->m_addNewValuePushButton, SIGNAL(clicked()), this, SLOT(onAddNewValuePushButtonClicked()));
  connect(m_ui->m_setNewValueColorPushButton, SIGNAL(clicked()), this, SLOT(onsetNewValueColorPushButtonClicked()));
  connect(m_ui->m_loadROISPushButton, SIGNAL(clicked()), this, SLOT(onLoadROISPushButtonClicked()));
  connect(m_ui->m_oldValueComboBox, SIGNAL(currentIndexChanged(int)), this, SLOT(oldValueComboBoxcurrentIndexChanged(int)));
  connect(m_ui->m_newValueComboBox, SIGNAL(currentIndexChanged(int)), this, SLOT(newValueComboBoxcurrentIndexChanged(int)));
  connect(m_ui->m_transferPushButton, SIGNAL(clicked()), this, SLOT(onTansferPushButtonClicked()));
}

te::qt::widgets::RasterEditWizardPage::~RasterEditWizardPage()
{
  setLayer( te::map::AbstractLayerPtr() );

  const std::size_t geomVecSize = m_geomVec.size();
  for( std::size_t geomVecIdx = 0 ; geomVecIdx < geomVecSize ; ++geomVecIdx )
  {
    delete m_geomVec[ geomVecIdx ];
  }
}

bool te::qt::widgets::RasterEditWizardPage::isComplete() const
{
  return ( m_outputRasterPtr.get() != 0 );
}

bool te::qt::widgets::RasterEditWizardPage::setLayer(te::map::AbstractLayerPtr layer)
{
  // removing old strctures related to the last layer

  if( m_rasterMavDialogPtr.get() )
  {
    m_rasterMavDialogPtr->hide();
    m_rasterMavDialogPtr->getWidget()->removeVectorial();
    m_rasterMavDialogPtr.reset();
  }

  if( !m_outputDataSourceID.empty() )
  {
    te::da::DataSourceManager::getInstance().detach( m_outputDataSourceID );
    m_outputDataSourceID.clear();
  }

  m_outputRasterPtr.reset();
  m_palleteIndex2Descriptions.clear();

  // clear ui

  m_ui->m_roiSubstitutionValuesTableWidget->clearContents();
  m_ui->m_roiTableWidget->clearContents();
  m_ui->m_bandComboBox->clear();

  if( layer.get() == 0 )
  {
    return false;
  }
  else
  {
    std::unique_ptr<te::rst::Raster> inputRst(te::map::GetRaster(layer.get()));

    if( inputRst.get() == 0 )
    {
      return false;
    }
    else
    {
      m_palleteIndex2Descriptions.clear();
      m_palleteIndex2Descriptions.resize( inputRst->getNumberOfBands() );

      std::vector< std::vector< te::rst::BandProperty::ColorEntry > > outputPalettes;

      // do we have a layer pallete ?

      {
        te::se::Style const * const layerStylePtr = layer->getStyle();

        if( layerStylePtr )
        {
          te::se::Rule const * const layerRulePtr = layerStylePtr->getRule(0);

          if( layerRulePtr )
          {
            te::se::Symbolizer const * const layerSymbPtr =
              layerRulePtr->getSymbolizer(0);

            if( ( layerSymbPtr != 0 ) && ( layerSymbPtr->getType() == "RasterSymbolizer" ) )
            {
              te::se::RasterSymbolizer const * const rasterSymbPtr =
                (te::se::RasterSymbolizer const * const)layerSymbPtr;

              te::se::ColorMap const * const layerColorMapPtr =
                rasterSymbPtr->getColorMap();

              if( layerColorMapPtr )
              {
                te::se::Recode const * const recodePtr = layerColorMapPtr->getRecode();

                if( recodePtr )
                {
                  std::vector< te::se::MapItem* > mapItens = recodePtr->getMapItems();
                  te::se::ParameterValue const * parValPtr = 0;
                  const std::size_t mapItensSize = mapItens.size();
                  te::color::RGBAColor rgbaColor;
                  te::rst::BandProperty::ColorEntry colorEntry;
                  const unsigned int nBands = inputRst->getNumberOfBands();
                  unsigned int bandIdx = 0;
                  te::se::MapItem* mapItemPtr = 0;
                  std::map< unsigned int , std::string > palleteIndex2DescriptionMap;

                  for( std::size_t mapItensIdx = 0 ; mapItensIdx < mapItensSize ;
                    ++mapItensIdx )
                  {
                    mapItemPtr = mapItens[ mapItensIdx ];
                    parValPtr = mapItemPtr->getValue();
                    rgbaColor = te::se::GetColor( parValPtr );

                    colorEntry.c1 = (short)rgbaColor.getRed();
                    colorEntry.c2 = (short)rgbaColor.getGreen();
                    colorEntry.c3 = (short)rgbaColor.getBlue();
                    colorEntry.c4 = (short)rgbaColor.getAlpha();

                    if( outputPalettes.empty() )
                    {
                      outputPalettes.resize( nBands );
                    }

                    for( bandIdx = 0 ; bandIdx < nBands ; ++bandIdx )
                    {
                      outputPalettes[ bandIdx ].push_back( colorEntry );
                    }

                    palleteIndex2DescriptionMap[ outputPalettes[ 0 ].size() - 1 ]
                      = mapItemPtr->getTitle();
                  }

                  m_palleteIndex2Descriptions.clear();
                  m_palleteIndex2Descriptions.resize( inputRst->getNumberOfBands(),
                    palleteIndex2DescriptionMap );
                }
              }
            }
          }
        }
      }

      // Checking raster type

      {
        const unsigned int nBands = inputRst->getNumberOfBands();

        for( unsigned int bandIdx = 0 ; bandIdx < nBands ; ++bandIdx )
        {
          const te::rst::Band& inBand = *inputRst->getBand( bandIdx );

          if(
              ( inBand.getProperty()->m_type != te::dt::UCHAR_TYPE )
              &&
              ( inBand.getProperty()->m_type != te::dt::UINT16_TYPE )
              &&
              ( inBand.getProperty()->m_type != te::dt::UINT32_TYPE )
              &&
              ( inBand.getProperty()->m_type != te::dt::UINT64_TYPE )
              &&
              ( inBand.getProperty()->m_type != te::dt::BOOLEAN_TYPE )
              &&
              ( inBand.getProperty()->m_type != te::dt::BIT_TYPE )
            )
          {
            QMessageBox::critical(this, tr("RasterEdit"), tr("Invalid input raster type"));
            return false;
          }
        }
      }

      // raster pallete

      if( outputPalettes.empty() )
      {
        const unsigned int nBands = inputRst->getNumberOfBands();
        const unsigned int nCols = inputRst->getNumberOfColumns();
        const unsigned int nRows = inputRst->getNumberOfRows();
        unsigned int row = 0 ;
        unsigned int col = 0;
        double value = 0;
        std::vector< te::rst::BandProperty::ColorEntry > bandPalette;
        std::set< double > uniqueRasterValues;

        for( unsigned int bandIdx = 0 ; bandIdx < nBands ; ++bandIdx )
        {
          const te::rst::Band& inBand = *inputRst->getBand( bandIdx );
          const double& noDataValue = inBand.getProperty()->m_noDataValue;

          if( inBand.getProperty()->m_palette.empty() )
          {
            uniqueRasterValues.clear();
            for( row = 0 ; row < nRows ; ++row )
            {
              for( col = 0 ; col < nCols ; ++col )
              {
                inBand.getValue( col, row, value );

                if( value != noDataValue )
                  uniqueRasterValues.insert( value );
              }
            }

            // creating a random pallete

            bandPalette.clear();

            if( !uniqueRasterValues.empty() )
            {
              te::rp::CreateFixedStepPalette( *uniqueRasterValues.rbegin(),
                true, bandPalette );
            }

            outputPalettes.push_back( bandPalette );
          }
          else
          {
            outputPalettes.push_back( inBand.getProperty()->m_palette );
          }
        }
      }

      // creating the memory raster

      try
      {
        std::vector< te::rst::BandProperty * > bandProps;
        for( unsigned int bandIdx = 0 ; bandIdx < inputRst->getNumberOfBands() ;
          ++bandIdx )
        {
          bandProps.push_back( new te::rst::BandProperty(
            *( inputRst->getBand( bandIdx )->getProperty() ) ) );

          if( !outputPalettes.empty() )
          {
            bandProps.back()->m_colorInterp = te::rst::PaletteIdxCInt;
            bandProps.back()->m_palette = outputPalettes[ bandIdx ];
          }
        }

        std::map<std::string, std::string> rasterInfo;
        rasterInfo[ "MAXMEMPERCENTUSED "] = 50;

        te::rst::Grid* gridPtr = new te::rst::Grid(
          *inputRst->getGrid() );

        te::qt::widgets::ScopedCursor cursor(Qt::WaitCursor);
        
        std::unique_ptr< te::rst::Raster > outputRasterPtr;
        outputRasterPtr.reset(te::rst::RasterFactory::make( "EXPANSIBLE",
          gridPtr, bandProps, rasterInfo));

        m_outputRasterPtr.reset( outputRasterPtr.release() );
      }
      catch(const std::exception& e)
      {
        QMessageBox::warning(this, tr("Output raster creation error"), e.what());
        return false;
      }
      catch(...)
      {
        QMessageBox::critical(this, tr("RasterEdit"), tr("Output raster creation error"));
        return false;
      }

      // raster data copy

      {
        te::qt::widgets::ScopedCursor cursor(Qt::WaitCursor);
        
        const unsigned int nBands = m_outputRasterPtr->getNumberOfBands();
        const unsigned int nCols = m_outputRasterPtr->getNumberOfColumns();
        const unsigned int nRows = m_outputRasterPtr->getNumberOfRows();
        unsigned int row = 0 ;
        unsigned int col = 0;
        double value = 0;

        for( unsigned int bandIdx = 0 ; bandIdx < nBands ; ++bandIdx )
        {
          const te::rst::Band& inBand = *inputRst->getBand( bandIdx );
          te::rst::Band& outBand = *m_outputRasterPtr->getBand( bandIdx );

          for( row = 0 ; row < nRows ; ++row )
          {
            for( col = 0 ; col < nCols ; ++col )
            {
              inBand.getValue( col, row, value );
              outBand.setValue( col, row, value );
            }
          }
        }
      }

      // memory data set creation

      {
        const void * rasterAddress = static_cast<const void*>(m_outputRasterPtr.get());
        std::stringstream rasterAddressSS ;
        rasterAddressSS << rasterAddress;
        m_outputDataSourceID = rasterAddressSS.str();

        std::vector< te::rst::BandProperty * > bandProps;
        for( unsigned int bandIdx = 0 ; bandIdx < m_outputRasterPtr->getNumberOfBands() ;
          ++bandIdx )
        {
          bandProps.push_back( new te::rst::BandProperty(
            *( m_outputRasterPtr->getBand( bandIdx )->getProperty() ) ) );
        }

        std::unique_ptr< te::rst::RasterProperty > propPtr(
          new te::rst::RasterProperty( new te::rst::Grid(
          *m_outputRasterPtr->getGrid() ),
          bandProps, std::map< std::string, std::string >() ) );

        std::unique_ptr< te::da::DataSetType > dataSetTypePtr1(
          new te::da::DataSetType(m_outputDataSourceID) );
        dataSetTypePtr1->add( propPtr.release() );

        std::unique_ptr< te::da::DataSetType > dataSetTypePtr2(
          (te::da::DataSetType*)dataSetTypePtr1->clone() );

        te::da::DataSourcePtr outputDataSourcePtr( te::da::DataSourceFactory::make(
          "MEM", "memory://?OPERATION_MODE=NON-SHARED").release() );
        outputDataSourcePtr->setId( m_outputDataSourceID );
        outputDataSourcePtr->createDataSet(dataSetTypePtr1.release(),
          std::map< std::string, std::string >() );

        std::unique_ptr< te::mem::DataSet > dataSetPtr(
          new te::mem::DataSet( dataSetTypePtr2.get() ) ); // The dataset will NOT take the ownership of the given DataSetType.

        std::vector< te::rst::Raster* > rastersPtrs;
        rastersPtrs.push_back( m_outputRasterPtr.get() );
        std::unique_ptr< te::rst::ProxyRaster > proxyRasterPtr(
          new te::rst::ProxyRaster( rastersPtrs, false ) );

        std::unique_ptr< te::mem::DataSetItem > dsItemPtr(
          new te::mem::DataSetItem( dataSetPtr.get() ) );
        dsItemPtr->setRaster( 0, proxyRasterPtr.release() );
        dataSetPtr->add( dsItemPtr.release() );

        dataSetPtr->moveBeforeFirst();
        outputDataSourcePtr->add( m_outputDataSourceID, dataSetPtr.get(),
          std::map< std::string, std::string >() );
        te::da::DataSourceManager::getInstance().insert( outputDataSourcePtr );
      }

      // raster navigator

      {
        std::unique_ptr< te::map::DataSetLayer > dataSetLayerPointer(
          new te::map::DataSetLayer(m_outputDataSourceID, m_outputDataSourceID) );
        dataSetLayerPointer->setDataSetName(m_outputDataSourceID);
        dataSetLayerPointer->setDataSourceId(m_outputDataSourceID);
        dataSetLayerPointer->setVisibility(te::map::VISIBLE);
        dataSetLayerPointer->setRendererType("ABSTRACT_LAYER_RENDERER");
        dataSetLayerPointer->setSRID(m_outputRasterPtr->getSRID());
        dataSetLayerPointer->setExtent(*(m_outputRasterPtr->getGrid()->getExtent()));
//        dataSetLayerPointer->setStyle( layer->getStyle()->clone() );

        m_rasterMavDialogPtr.reset( new te::qt::widgets::RasterNavigatorDialog(
          this) );
        m_rasterMavDialogPtr->set( te::map::AbstractLayerPtr(
          dataSetLayerPointer.release() ) );
        m_rasterMavDialogPtr->getWidget()->hideInfoTool( true );
        connect(m_rasterMavDialogPtr->getWidget(),
          SIGNAL(envelopeAcquired(te::gm::Envelope)), this,
          SLOT(onEnvelopeAcquired(te::gm::Envelope)));
        connect(m_rasterMavDialogPtr->getWidget(),
          SIGNAL(geomAquired(te::gm::Polygon*)), this,
          SLOT(onGeomAquired(te::gm::Polygon*)));
        connect(m_rasterMavDialogPtr->getWidget(),
          SIGNAL(pointPicked(double, double)), this,
          SLOT(onPointPicked(double, double)));
        connect(m_rasterMavDialogPtr->getWidget(),
          SIGNAL(mapDisplayExtentChanged()), this,
          SLOT(drawGeometries()));
        m_rasterMavDialogPtr->setWindowTitle(tr("Raster Edit"));
        m_rasterMavDialogPtr->show();
      }

      // UI update

      for( unsigned int band = 0 ; band < m_outputRasterPtr->getNumberOfBands() ; ++band )
      {
        m_ui->m_bandComboBox->addItem( QString::number( band ) );
      }

      updateValuesCombos();
    }
  }

  return true;
}

unsigned int te::qt::widgets::RasterEditWizardPage::getBand()
{
  if( m_ui->m_bandComboBox->count() )
  {
    return m_ui->m_bandComboBox->currentText().toUInt();
  }
  else
  {
    return 0;
  }
}

void te::qt::widgets::RasterEditWizardPage::getROIs(std::vector<te::gm::Geometry*> &geomVec)
{
  geomVec = m_geomVec;
}

void te::qt::widgets::RasterEditWizardPage::addRoiItem()
{
  int row = m_ui->m_roiTableWidget->rowCount();

  m_ui->m_roiTableWidget->insertRow(row);

  QTableWidgetItem* item = new QTableWidgetItem(
    "ROI Item " + QString::number( row + 1 ) );
  m_ui->m_roiTableWidget->setItem(row, 0, item);

  m_ui->m_roiTableWidget->resizeRowsToContents();
  m_ui->m_roiTableWidget->resizeColumnToContents(0);
}

void te::qt::widgets::RasterEditWizardPage::addRulePushButtonClicked()
{
  if( m_outputRasterPtr.get() == 0 ) return;
  if( m_ui->m_oldValueComboBox->count() == 0 ) return;
  if( m_ui->m_newValueComboBox->count() == 0 ) return;

  QString oldText = m_ui->m_oldValueComboBox->currentText();
  QVariant oldData = m_ui->m_oldValueComboBox->currentData();
  QString newText = m_ui->m_newValueComboBox->currentText();
  QVariant newData = m_ui->m_newValueComboBox->currentData();

  const unsigned int oldValue = oldData.toUInt();
  const unsigned int newValue = newData.toUInt();
  int rowsCount = m_ui->m_roiTableWidget->rowCount();
  const unsigned int bandIdx = getBand();
  const unsigned int palleteSize =
    m_outputRasterPtr->getBand( bandIdx )->getProperty()->m_palette.size();

  m_ui->m_roiSubstitutionValuesTableWidget->insertRow( rowsCount );

  QTableWidgetItem* oldItem = new QTableWidgetItem( oldText );
  oldItem->setData( Qt::UserRole, oldData );

  m_ui->m_roiSubstitutionValuesTableWidget->setItem(rowsCount, 0, oldItem);

  if( ( palleteSize != 0 ) && ( oldValue < palleteSize ) )
  {
    const te::rst::BandProperty::ColorEntry& rasterColor =
      m_outputRasterPtr->getBand(
      bandIdx )->getProperty()->m_palette[ (size_t)oldValue ];

    te::color::RGBAColor tlColor( rasterColor.c1, rasterColor.c2,
      rasterColor.c3, rasterColor.c4 );

    QColor qtColor = te::qt::widgets::Convert2Qt( tlColor );

    QBrush qtBrush( qtColor, Qt::SolidPattern );

    QTableWidgetItem* itemColor = new QTableWidgetItem();
    itemColor->setBackground( qtBrush );
    itemColor->setForeground( qtBrush );
    m_ui->m_roiSubstitutionValuesTableWidget->setItem(rowsCount, 1, itemColor);
  }

  QTableWidgetItem* newItem = new QTableWidgetItem( newText );
  newItem->setData( Qt::UserRole, newData );

  m_ui->m_roiSubstitutionValuesTableWidget->setItem(rowsCount, 2, newItem);

  if( ( palleteSize != 0 ) && ( newValue < palleteSize ) )
  {
    const te::rst::BandProperty::ColorEntry& rasterColor =
      m_outputRasterPtr->getBand(
      bandIdx )->getProperty()->m_palette[ (size_t)newValue ];

    te::color::RGBAColor tlColor( rasterColor.c1, rasterColor.c2,
      rasterColor.c3, rasterColor.c4 );

    QColor qtColor = te::qt::widgets::Convert2Qt( tlColor );

    QBrush qtBrush( qtColor, Qt::SolidPattern );

    QTableWidgetItem* itemColor = new QTableWidgetItem();
    itemColor->setBackground( qtBrush );
    itemColor->setForeground( qtBrush );
    m_ui->m_roiSubstitutionValuesTableWidget->setItem(rowsCount, 3, itemColor);
  }

//  m_ui->m_roiSubstitutionValuesTableWidget->resizeRowsToContents();
  m_ui->m_roiSubstitutionValuesTableWidget->resizeColumnsToContents();
}

void te::qt::widgets::RasterEditWizardPage::removeRulePushButtonClicked()
{
  QList<QTableWidgetItem *> selectedItens;
  int itemRow = 0;

  do
  {
    selectedItens = m_ui->m_roiSubstitutionValuesTableWidget->selectedItems();

    if( selectedItens.empty() )
    {
      break;
    }

    QTableWidgetItem * itemPtr = selectedItens.front();
    itemRow = itemPtr->row();
    m_ui->m_roiSubstitutionValuesTableWidget->removeRow( itemRow );
  } while( true );
}

void te::qt::widgets::RasterEditWizardPage::onEnvelopeAcquired(te::gm::Envelope env)
{
  if( m_outputRasterPtr.get() )
  {
    if(m_rasterMavDialogPtr->getWidget()->getDisplay()->getSRID() != m_outputRasterPtr->getSRID())
      env.transform(m_rasterMavDialogPtr->getWidget()->getDisplay()->getSRID(), m_outputRasterPtr->getSRID());

    if (!env.isValid() || !env.intersects( *m_outputRasterPtr->getGrid()->getExtent()) )
      return;

    m_geomVec.push_back(te::gm::GetGeomFromEnvelope(&env, m_outputRasterPtr->getSRID()));

    addRoiItem();

    drawGeometries();
  }
}

void te::qt::widgets::RasterEditWizardPage::onGeomAquired(te::gm::Polygon *poly)
{
  if( m_outputRasterPtr.get() )
  {
    if(m_rasterMavDialogPtr->getWidget()->getDisplay()->getSRID() != m_outputRasterPtr->getSRID())
      poly->transform(m_outputRasterPtr->getSRID());

    std::unique_ptr< te::gm::Geometry > extentGeomPtr( te::gm::GetGeomFromEnvelope(
      m_outputRasterPtr->getGrid()->getExtent(), m_outputRasterPtr->getSRID() ) );

    if(
        ( !poly->isValid() )
        ||
        ( !poly->intersects( extentGeomPtr.get() ) )
      )
    {
      return;
    }

    m_geomVec.push_back(poly);

    addRoiItem();

    drawGeometries();
  }
}

void te::qt::widgets::RasterEditWizardPage::onPointPicked(double x, double y)
{
  if( m_outputRasterPtr.get() != 0 )
  {
    double row = 0;
    double col = 0;
    m_outputRasterPtr->getGrid()->geoToGrid( x, y, col, row );

    if(
        ( col >= 0 )
        &&
        ( col < m_outputRasterPtr->getNumberOfColumns() )
        &&
        ( row >= 0 )
        &&
        ( row < m_outputRasterPtr->getNumberOfRows() )
      )
    {
      const unsigned int intCol = te::common::Round< double, unsigned int >( col );
      const unsigned int intRow = te::common::Round< double, unsigned int >( row );

      std::vector< std::pair< double, double > > targetValues;
      getReplaceValues( targetValues );

      unsigned int targetInputBand = getBand();
      double pixelValue = 0;


      for( std::size_t targetValuesIdx = 0 ; targetValuesIdx < targetValues.size() ;
        ++targetValuesIdx )
      {
        m_outputRasterPtr->getValue( intCol, intRow, pixelValue, targetInputBand );

        if( pixelValue == targetValues[ targetValuesIdx ].first )
        {
          bool res = false;
          {
            te::qt::widgets::ScopedCursor cursor(Qt::WaitCursor);
            res = te::rp::ReplaceContiguousSegmentValues( *m_outputRasterPtr,
              targetInputBand, intRow, intCol,
              targetValues[ targetValuesIdx ].second );
          }
          
          if( ! res )
          {
            QMessageBox::critical(this, tr("RasterEdit"), tr("Values replacement error"));
            return;
          }
        }
      }

      m_rasterMavDialogPtr->getWidget()->refresh();
    }
  }
}

void te::qt::widgets::RasterEditWizardPage::onSelectRoiItem()
{
  drawGeometries();
}

void te::qt::widgets::RasterEditWizardPage::onRemoveRoiItem()
{
  int row = m_ui->m_roiTableWidget->currentIndex().row();

  if(m_geomVec.empty())
  {
    return;
  }

  m_ui->m_roiTableWidget->removeRow(row);

  m_geomVec.erase(m_geomVec.begin()+row);

  drawGeometries();
}

void te::qt::widgets::RasterEditWizardPage::drawGeometries()
{
  if( m_rasterMavDialogPtr.get() )
  {
    int selectedRow = m_ui->m_roiTableWidget->currentIndex().row();

    m_rasterMavDialogPtr->getWidget()->getDisplay()->getDraftPixmap()->fill(Qt::transparent);

    const te::gm::Envelope& mapExt = m_rasterMavDialogPtr->getWidget()->getDisplay()->getExtent();

    te::qt::widgets::Canvas canvasInstance(m_rasterMavDialogPtr->getWidget()->getDisplay()->getDraftPixmap());
    canvasInstance.setWindow(mapExt.m_llx, mapExt.m_lly, mapExt.m_urx, mapExt.m_ury);

    te::gm::Geometry* geom = nullptr;

    const std::size_t geomVecSize = m_geomVec.size();

    canvasInstance.setPolygonContourWidth(1);
    canvasInstance.setPolygonContourColor(te::color::RGBAColor(100, 177, 216, TE_OPAQUE));
    canvasInstance.setPolygonFillColor(te::color::RGBAColor(100, 177, 216, 80));

    for(std::size_t i = 0; i < geomVecSize; ++i)
    {
      geom = m_geomVec[i];

      if( geom->getSRID() != m_rasterMavDialogPtr->getWidget()->getDisplay()->getSRID())
      {
        geom->transform(m_rasterMavDialogPtr->getWidget()->getDisplay()->getSRID());
      }

      if( selectedRow == (int)i )
      {
        canvasInstance.setPolygonContourWidth(2);
        canvasInstance.setPolygonContourColor(te::color::RGBAColor(100, 177, 216, 222));
        canvasInstance.setPolygonFillColor(te::color::RGBAColor(100, 177, 216, 220));

        canvasInstance.draw( geom );

        canvasInstance.setPolygonContourWidth(1);
        canvasInstance.setPolygonContourColor(te::color::RGBAColor(100, 177, 216, TE_OPAQUE));
        canvasInstance.setPolygonFillColor(te::color::RGBAColor(100, 177, 216, 80));
      }
      else
      {
        canvasInstance.draw( geom );
      }
    }

    m_rasterMavDialogPtr->getWidget()->getDisplay()->repaint();
  }
}

void te::qt::widgets::RasterEditWizardPage::getReplaceValues(
  std::vector< std::pair< double, double > >& replaceValues ) const
{
  replaceValues.clear();

  const int rowsCount = m_ui->m_roiSubstitutionValuesTableWidget->rowCount();
  std::pair< double, double > auxPair;
  QTableWidgetItem const* itemPtr = 0;

  for( int row = 0 ; row < rowsCount ; ++row )
  {
    itemPtr = m_ui->m_roiSubstitutionValuesTableWidget->item( row, 0 );
    auxPair.first = (double)itemPtr->data( Qt::UserRole ).toUInt();

    itemPtr = m_ui->m_roiSubstitutionValuesTableWidget->item( row, 2 );
    auxPair.second = (double)itemPtr->data( Qt::UserRole ).toUInt();

    replaceValues.push_back( auxPair );
  }
}

std::shared_ptr< te::rst::Raster > te::qt::widgets::RasterEditWizardPage::getOutputRaster()
{
  return m_outputRasterPtr;
}


void te::qt::widgets::RasterEditWizardPage::onReplaceAllPushButtonClicked()
{
  std::vector< std::pair< double, double > > targetValues;
  getReplaceValues( targetValues );

  const unsigned int targetInputBand = getBand();

  std::vector<te::gm::Polygon*> restrictionPols;
  const std::size_t geomVecSize = m_geomVec.size();
  for( std::size_t geomVecIdx = 0 ; geomVecIdx < geomVecSize ;
    ++geomVecIdx )
  {
    if( m_geomVec[ geomVecIdx ]->getGeomTypeId() == te::gm::PolygonType )
    {
      restrictionPols.push_back( (te::gm::Polygon*)m_geomVec[ geomVecIdx ] );
    }
  }

  bool res = false;
  {
    te::qt::widgets::ScopedCursor cursor(Qt::WaitCursor);
    res = te::rp::SelectiveReplaceValues( *m_outputRasterPtr, targetInputBand,
      targetValues, true, restrictionPols, *m_outputRasterPtr, targetInputBand );
  }
  
  if( ! res )
  {
    QMessageBox::critical(this, tr("RasterEdit"), tr("Values replacement error"));
    return;
  }

  m_rasterMavDialogPtr->getWidget()->refresh();
}

void te::qt::widgets::RasterEditWizardPage::onAddNewValuePushButtonClicked()
{
  const unsigned int bandIdx = getBand();
  const unsigned int palleteSize =
    m_outputRasterPtr->getBand( bandIdx )->getProperty()->m_palette.size();
  const double noDataValue =
    m_outputRasterPtr->getBand( bandIdx )->getProperty()->m_noDataValue;
  const unsigned int newValue = (unsigned int)m_ui->m_newValueDoubleSpinBox->value();
  const std::string newValueName = m_ui->m_newValueNameLineEdit->text().toStdString();

  if(
      ( newValue >= palleteSize )
      &&
      ( ((double)newValue) != noDataValue )
      &&
      ( !newValueName.empty() )
    )
  {
    QPalette qtPalette = m_ui->m_newValueColorLabel->palette();

    QColor qtColor = qtPalette.color( QPalette::Foreground );

    te::color::RGBAColor terralibColor = te::qt::widgets::Convert2TerraLib( qtColor );

    int red = 0;
    int green = 0;
    int blue = 0;
    int alpha = 0;
    terralibColor.getRgba( &red, &green, &blue, &alpha );

    te::rst::BandProperty::ColorEntry rasterColour;
    rasterColour.c1 = (short)red;
    rasterColour.c2 = (short)green;
    rasterColour.c3 = (short)blue;
    rasterColour.c4 = (short)alpha;

    const unsigned int newElementsCount = newValue - palleteSize + 1;
    std::vector< te::rst::BandProperty::ColorEntry >& outPalette = m_outputRasterPtr->getBand(
      bandIdx )->getProperty()->m_palette;

    for( unsigned int insertedElements = 0 ; insertedElements < newElementsCount ;
      ++insertedElements )
    {
      m_palleteIndex2Descriptions[ bandIdx ][ outPalette.size() ] = newValueName;
      outPalette.push_back( rasterColour );
    }

    updateValuesCombos();
    m_rasterMavDialogPtr->getWidget()->refresh();
  }
  else
  {
    QMessageBox::critical(this, tr("RasterEdit"), tr("Invalid value"));
    return;
  }
}

void te::qt::widgets::RasterEditWizardPage::onsetNewValueColorPushButtonClicked()
{
  QColor qtColor = QColorDialog::getColor( Qt::white, this, tr( "Select color") );

  QPalette qtPalette = m_ui->m_newValueColorLabel->palette();
  qtPalette.setColor(m_ui->m_newValueColorLabel->backgroundRole(), qtColor);
  qtPalette.setColor(m_ui->m_newValueColorLabel->foregroundRole(), qtColor);

  m_ui->m_newValueColorLabel->setAutoFillBackground(true);
  m_ui->m_newValueColorLabel->setPalette(qtPalette);
}

void te::qt::widgets::RasterEditWizardPage::onLoadROISPushButtonClicked()
{
  QString inputFile = QFileDialog::getOpenFileName(
      this, tr("Select a ROI file"),
      te::qt::widgets::GetFilePathFromSettings("roifile"),
      "Shape Files (*.shp *.SHP)");

  if(inputFile.isEmpty())
    return;

  const std::string fileName( inputFile.toUtf8().data() );

  std::string dsInf("file://" + fileName );

  te::da::DataSourcePtr dsPtr(
      te::da::DataSourceFactory::make("OGR", dsInf).release());

  dsPtr->open();

  if(! dsPtr->isValid() )
  {
    QMessageBox::warning(this, tr("Classifier"),
                         tr("Invalid file data source"));
    return;
  }

  std::vector<std::string> dataSetNames = dsPtr->getDataSetNames();
  if( dataSetNames.empty() )
  {
    QMessageBox::warning(this, tr("Classifier"),
                         tr("Invalid file data source"));
    return;
  }

  const std::string dataSetname = dataSetNames[ 0 ];

  std::unique_ptr<te::da::DataSetType> dsType = dsPtr->getDataSetType(
    dataSetname );
  if(!dsType->hasGeom())
  {
    QMessageBox::warning(this, tr("Classifier"),
                         tr("The informed file does not have geometries."));
    return;
  }

  te::dt::Property* p = dsType->findFirstPropertyOfType(te::dt::GEOMETRY_TYPE);

  std::size_t geomPropPosition = dsType->getPropertyPosition( p );

  std::unique_ptr<te::da::DataSet> ds = dsPtr->getDataSet(dataSetname);

  //move the data set to begin
  ds->moveBeforeFirst();

  std::unique_ptr< te::gm::Geometry > geomPtr;
  te::gm::MultiPolygon* mPolPtr = 0;
  std::unique_ptr< te::gm::Geometry > polPtr;
  const int rasterSRID = m_outputRasterPtr->getSRID();
  te::gm::Polygon rasterEnvelopePol = te::gm::CreatePolygon(
    *m_outputRasterPtr->getGrid()->getExtent(), rasterSRID, 0 );
  int rowsCount = 0;

  while(ds->moveNext())
  {
    geomPtr = ds->getGeometry( geomPropPosition );
    if( geomPtr->getGeomTypeId() != te::gm::MultiPolygonType )
    {
      QMessageBox::warning(this, tr("Classifier"),
                           tr("The layer contains not supported geometries"));
      return;
    }

    mPolPtr = (te::gm::MultiPolygon*)geomPtr.get();
    if( mPolPtr->getNumGeometries() != 1 )
    {
      QMessageBox::warning(this, tr("Classifier"),
                           tr("The layer contains not supported geometries"));
      return;
    }

    polPtr.reset( (te::gm::Geometry*)mPolPtr->getGeometryN( 0 )->clone() );
    polPtr->transform( rasterSRID );

    if( rasterEnvelopePol.intersects( polPtr.get() ) )
    {
      m_geomVec.push_back( polPtr.release() );

      rowsCount = m_ui->m_roiTableWidget->rowCount();
      m_ui->m_roiTableWidget->insertRow( rowsCount );
      QTableWidgetItem* item = new QTableWidgetItem(
        "ROI Item " + QString::number( rowsCount + 1 ) );
      m_ui->m_roiTableWidget->setItem( rowsCount, 0, item);
    }
  }

//  m_ui->m_roiTableWidget->resizeRowsToContents();
  m_ui->m_roiTableWidget->resizeColumnsToContents();

  drawGeometries();
}

void te::qt::widgets::RasterEditWizardPage::updateValuesCombos()
{
  m_ui->m_oldValueComboBox->clear();
  m_ui->m_newValueComboBox->clear();
  m_ui->m_newValueDoubleSpinBox->setRange(0,0);

  if( m_outputRasterPtr.get() )
  {
    const unsigned int bandIdx = getBand();
    const unsigned int palleteSize =
      m_outputRasterPtr->getBand( bandIdx )->getProperty()->m_palette.size();
    std::map< unsigned int , std::string >::const_iterator
      palleteIndex2DescriptionMapIt;

    for( unsigned int palleteIdx = 0 ; palleteIdx < palleteSize ;
      ++palleteIdx )
    {
      palleteIndex2DescriptionMapIt = m_palleteIndex2Descriptions[ bandIdx ].find(
        palleteIdx );
      QVariant data( palleteIdx );
      QString text = QString::number( palleteIdx ) +
        (
          ( palleteIndex2DescriptionMapIt == m_palleteIndex2Descriptions[ bandIdx ].end() )
          ?
            QString()
          :
            ( "(" + QString( palleteIndex2DescriptionMapIt->second.c_str() ) + ")")
        );

      m_ui->m_oldValueComboBox->addItem( text, data );
      m_ui->m_newValueComboBox->addItem( text, data );
    }

    if( palleteSize )
    {
      m_ui->m_oldValueComboBox->setCurrentIndex( 0 );
      m_ui->m_newValueComboBox->setCurrentIndex( 0 );
    }

    m_ui->m_newValueDoubleSpinBox->setMinimum( (int)palleteSize );
  }
}

void te::qt::widgets::RasterEditWizardPage::oldValueComboBoxcurrentIndexChanged(int index)
{
  if( ( index >= 0 ) && ( m_outputRasterPtr.get() != 0 ) )
  {
    const unsigned int bandIdx = getBand();
    const unsigned int palleteSize =
      m_outputRasterPtr->getBand( bandIdx )->getProperty()->m_palette.size();
    QVariant dataVariant = m_ui->m_oldValueComboBox->itemData( index, Qt::UserRole );
    const unsigned int paletteIndex = dataVariant.toUInt();

    if( (unsigned int)paletteIndex < palleteSize )
    {
      const te::rst::BandProperty::ColorEntry& rasterColor =
        m_outputRasterPtr->getBand(
        bandIdx )->getProperty()->m_palette[ (size_t)paletteIndex ];

      te::color::RGBAColor tlColor( rasterColor.c1, rasterColor.c2,
        rasterColor.c3, rasterColor.c4 );

      QColor qtColor = te::qt::widgets::Convert2Qt( tlColor );

      QPalette qtPalette = m_ui->m_oldColorLabel->palette();
      qtPalette.setColor(m_ui->m_oldColorLabel->backgroundRole(), qtColor);
      qtPalette.setColor(m_ui->m_oldColorLabel->foregroundRole(), qtColor);

      m_ui->m_oldColorLabel->setAutoFillBackground(true);
      m_ui->m_oldColorLabel->setPalette(qtPalette);
    }
  }
}

void te::qt::widgets::RasterEditWizardPage::newValueComboBoxcurrentIndexChanged(int index)
{
  if( ( index >= 0 ) && ( m_outputRasterPtr.get() != 0 ) )
  {
    const unsigned int bandIdx = getBand();
    const unsigned int palleteSize =
      m_outputRasterPtr->getBand( bandIdx )->getProperty()->m_palette.size();
    QVariant dataVariant = m_ui->m_newValueComboBox->itemData( index, Qt::UserRole );
    const unsigned int paletteIndex = dataVariant.toUInt();

    if( (unsigned int)paletteIndex < palleteSize )
    {
      const te::rst::BandProperty::ColorEntry& rasterColor =
        m_outputRasterPtr->getBand(
        bandIdx )->getProperty()->m_palette[ (size_t)paletteIndex ];

      te::color::RGBAColor tlColor( rasterColor.c1, rasterColor.c2,
        rasterColor.c3, rasterColor.c4 );

      QColor qtColor = te::qt::widgets::Convert2Qt( tlColor );

      QPalette qtPalette = m_ui->m_newColorLabel->palette();
      qtPalette.setColor(m_ui->m_newColorLabel->backgroundRole(), qtColor);
      qtPalette.setColor(m_ui->m_newColorLabel->foregroundRole(), qtColor);

      m_ui->m_newColorLabel->setAutoFillBackground(true);
      m_ui->m_newColorLabel->setPalette(qtPalette);
    }
  }
}

void te::qt::widgets::RasterEditWizardPage::onRoiTableWidgetItemClicked(
  QTableWidgetItem*)
{
  drawGeometries();
}

void te::qt::widgets::RasterEditWizardPage::onTansferPushButtonClicked()
{
  std::unique_ptr< te::qt::widgets::PointPicker > pointPToolPtr(
    new te::qt::widgets::PointPicker( m_mainMapDispPtr, Qt::ArrowCursor, this ) );
  pointPToolPtr->setParent( nullptr );

  connect( pointPToolPtr.get(), SIGNAL( pointPicked(QPointF&) ), this,
    SLOT( onPointPickedOnMainMapDisplay( QPointF& ) ) );

  m_mainMapDispPtr->setCurrentTool( pointPToolPtr.release() );
}

void te::qt::widgets::RasterEditWizardPage::onPointPickedOnMainMapDisplay(
  QPointF& point )
{
  if(
      ( m_mainLayerExplorerPtr != 0 )
      &&
      ( m_outputRasterPtr.get() != 0 )
      &&
      ( m_rasterMavDialogPtr.get() != 0 )
    )
  {
    std::list< te::qt::widgets::TreeItem * > selectedItems =
      m_mainLayerExplorerPtr->getSelectedItems();
    const std::list< te::qt::widgets::TreeItem * >::const_iterator selectedItemsIt =
      selectedItems.begin();

    if( selectedItems.size() != 1 )
    {
      QMessageBox::critical(this, tr("RasterEdit"), tr("Invalid layer selection"));
      return;
    }
    else
    {
      if( (*selectedItemsIt)->getType() != "LAYER" )
      {
        QMessageBox::critical(this, tr("RasterEdit"), tr("Invalid layer selection"));
        return;
      }
      else
      {
        te::map::AbstractLayerPtr layerPtr = ( (te::qt::widgets::LayerItem*)(
          *selectedItemsIt ) )->getLayer();
        if( layerPtr.get() == 0 ) throw te::common::Exception(
          TE_TR( "Invalid layer pointer" ) );

        std::unique_ptr<te::da::DataSet> dataSetPtr = layerPtr->getData();
        if( dataSetPtr.get() == 0 ) throw te::common::Exception(
          TE_TR( "Invalid dataset pointer" ) );

        if(
            ( dataSetPtr->getNumProperties() == 0 )
            ||
            ( dataSetPtr->getPropertyDataType( 0 ) != te::dt::RASTER_TYPE )
          )
        {
          QMessageBox::critical(this, tr("RasterEdit"), tr("Incompatible selected layer"));
          return;
        }
        else
        {
          std::unique_ptr<te::rst::Raster> inputRst( te::map::GetRaster(
            layerPtr.get()) );
          if( inputRst.get() == 0 ) throw te::common::Exception(
            TE_TR( "Invalid raster pointer" ) );

          if(
              inputRst->getBand( 0 )->getProperty()->m_type
              !=
              m_outputRasterPtr->getBand( getBand() )->getProperty()->m_type
            )
          {
            QMessageBox::critical(this, tr("RasterEdit"), tr("Incompatible selected layer"));
            return;
          }
          else
          {
            double rowDouble = 0;
            double colDouble =0;
            inputRst->getGrid()->geoToGrid( point.x(), point.y(), colDouble,
              rowDouble );

            unsigned int col = te::common::Round< double, unsigned int >( colDouble );
            unsigned int row = te::common::Round< double, unsigned int >( rowDouble );

            if(
                ( col >= 0 )
                &&
                ( col < inputRst->getNumberOfColumns() )
                &&
                ( row >= 0 )
                &&
                ( row < inputRst->getNumberOfRows() )
              )
            {
              double outputValue = 0;
              inputRst->getValue( col, row, outputValue, 0 );

              if( ((std::size_t)outputValue) >= m_outputRasterPtr->getBand( getBand()
                )->getProperty()->m_palette.size() )
              {
                QMessageBox::critical(this, tr("RasterEdit"), tr("Invalid input layer pixel value"));
                return;
              }
              else
              {
                bool res = false;
                {
                  te::qt::widgets::ScopedCursor cursor(Qt::WaitCursor);
                  res = te::rp::ReplaceContiguousSegmentValues( *inputRst, 0,
                  *m_outputRasterPtr, getBand(), row, col, outputValue );
                }
                if( ! res )
                {
                  QMessageBox::critical(this, tr("RasterEdit"), tr("Error importing segment"));
                  return;
                }

                m_rasterMavDialogPtr->getWidget()->refresh();
              }
            }
          }

          return;
        }
      }
    }
  }
}
