/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/widgets/rp/RasterEditWizardPage.h

  \brief This file defines a class for raster data edition.
*/

#ifndef __TERRALIB_QT_WIDGETS_RP_INTERNAL_RASTEREDITWIZARDPAGE_H
#define __TERRALIB_QT_WIDGETS_RP_INTERNAL_RASTEREDITWIZARDPAGE_H

// TerraLib
#include "../../../geometry/Envelope.h"
#ifndef Q_MOC_RUN
#include "../../../maptools/AbstractLayer.h"
#include "../../../qt/widgets/canvas/MapDisplay.h"
#include "../../../qt/widgets/layer/explorer/LayerItemView.h"
#endif
#include "../Config.h"

// STL
#include <memory>

// Qt
#include <QWizardPage>
#include <QActionGroup>
#include <QTableWidgetItem>

// Forward declaration
namespace Ui { class RasterEditWizardPageForm; }

namespace te
{
  namespace qt
  {
    namespace widgets
    {
      class RpToolsWidget;
      class RasterNavigatorDialog;

      /*!
        \class RasterEditWizardPage

        \brief This class is GUI used to define the parameters raster data edition.
      */
      class TEQTWIDGETSEXPORT RasterEditWizardPage : public QWizardPage
      {
        Q_OBJECT

        public:

          RasterEditWizardPage(QWidget* parent,
            te::qt::widgets::MapDisplay* mainMapDispPtr,
            te::qt::widgets::LayerItemView* mainLayerExplorerPtr );

          ~RasterEditWizardPage();

          bool isComplete() const;

          /*!
            \brief This method is used to set the selected layer for editing operation

            \param layer The layer ptr

            \return true if ok, false on errors.

            \note This layer MUST HAVE a valid raster object.
          */
           bool setLayer(te::map::AbstractLayerPtr layer);

          /*!
            \brief Return the current output raster.

            \return Return the current output raster.
          */
           std::shared_ptr< te::rst::Raster > getOutputRaster();

          /*!
            \brief Return the current output pallete indexes to their respetive descriptions.

            \brief Return the current output pallete indexes to their respetive descriptions.
          */
           const std::vector< std::map< unsigned int , std::string > >&
           getPalleteIndex2Descriptions() const { return m_palleteIndex2Descriptions; };

        protected slots:

          void addRulePushButtonClicked();

          void removeRulePushButtonClicked();

          void onEnvelopeAcquired(te::gm::Envelope env);

          void onGeomAquired(te::gm::Polygon* poly);

          void onPointPicked(double x, double y);

          void onSelectRoiItem();

          void onRemoveRoiItem();

          void onReplaceAllPushButtonClicked();

          void onAddNewValuePushButtonClicked();

          void onsetNewValueColorPushButtonClicked();

          void onLoadROISPushButtonClicked();

          void oldValueComboBoxcurrentIndexChanged(int index);

          void newValueComboBoxcurrentIndexChanged(int index);

          void drawGeometries();

          void onRoiTableWidgetItemClicked( QTableWidgetItem*);

          void onTansferPushButtonClicked();

          void onPointPickedOnMainMapDisplay( QPointF& );

        protected:

          unsigned int getBand();

          void getROIs(std::vector<te::gm::Geometry*> &geomVec);

          void getReplaceValues( std::vector< std::pair< double, double > >& replaceValues ) const;

          void updateValuesCombos();

          void addRoiItem();

        private:

          te::qt::widgets::MapDisplay* m_mainMapDispPtr;
          te::qt::widgets::LayerItemView* m_mainLayerExplorerPtr;
          std::unique_ptr<Ui::RasterEditWizardPageForm> m_ui;
          std::vector<te::gm::Geometry*> m_geomVec;
          std::unique_ptr< te::qt::widgets::RasterNavigatorDialog > m_rasterMavDialogPtr;
          std::shared_ptr< te::rst::Raster > m_outputRasterPtr;
          std::string m_outputDataSourceID;
          std::vector< std::map< unsigned int , std::string > > m_palleteIndex2Descriptions; // One map per band
      };

    } // end namespace widgets
  }   // end namespace qt
}     // end namespace te

#endif  // __TERRALIB_QT_WIDGETS_RP_INTERNAL_RASTEREDITWIZARDPAGE_H
