/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/widgets/rp/RasterHistogramWidget.cpp

  \brief This file has the RasterHistogramWidget class.
*/

// TerraLib
#include "../../../dataaccess/dataset/DataSet.h"
#include "../../../dataaccess/utils/Utils.h"
#include "../../../raster/Band.h"
#include "../../../raster/Raster.h"
#include "../../../raster/RasterSummaryManager.h"
#include "../../../raster/Utils.h"
#include "../../../core/translator/Translator.h"
#include "../charts/ChartDisplay.h"
#include "../charts/ChartStyle.h"
#include "../charts/Histogram.h"
#include "../charts/HistogramChart.h"
#include "../charts/HistogramStyle.h"
#include "RasterHistogramWidget.h"
#include "ui/ui_RasterHistogramWidgetForm.h"

// Qt
#include <QGridLayout>

//Qwt
#include <qwt_legend.h>

//STL
#include <memory>

te::qt::widgets::RasterHistogramWidget::RasterHistogramWidget(QWidget* parent, Qt::WindowFlags f)
  : QWidget(parent, f),
    m_histogramSampleStep(1),
    m_ui(new Ui::RasterHistogramWidgetForm)
{
  m_ui->setupUi(this);

  m_ui->m_applyToolButton->setIcon(QIcon::fromTheme("chart-bar"));

  m_inputRaster = nullptr;
  m_minValueLine = nullptr;
  m_maxValueLine = nullptr;

  //Creating and adjusting the chart Display's style.
  m_chartStyle = new te::qt::widgets::ChartStyle();
  m_chartStyle->setTitle(tr(""));
  m_chartStyle->setAxisX(tr("Gray Level"));
  m_chartStyle->setAxisY(tr("Frequency"));
  m_chartStyle->setGridChecked(true);

  //build form
  QGridLayout* layout = new QGridLayout(m_ui->m_widget);

  //Creating and adjusting the chart Display
  m_chartDisplay = new te::qt::widgets::ChartDisplay(m_ui->m_widget, QString::fromUtf8(""), m_chartStyle);
  m_chartDisplay->adjustDisplay();
  m_chartDisplay->show();
  m_chartDisplay->replot();

  layout->addWidget(m_chartDisplay, 0, 0);
  layout->setContentsMargins(0,0,0,0);

  m_histogramInput = new te::qt::widgets::Histogram();
  m_histogramChartInput = new te::qt::widgets::HistogramChart(m_histogramInput);
  m_histogramChartInput->setPen(Qt::black);
  m_histogramChartInput->setBrush(QBrush(Qt::blue));
  m_histogramChartInput->attach(m_chartDisplay);
  m_histogramChartInput->setTitle(tr("Input"));

  m_histogramOutput = new te::qt::widgets::Histogram();
  m_histogramChartOutput = new te::qt::widgets::HistogramChart(m_histogramOutput);
  m_histogramChartOutput->setPen(Qt::black, 3.);
  m_histogramChartOutput->setBrush(QBrush(QColor(255, 0, 0, 127)));
  m_histogramChartOutput->setStyle(QwtPlotHistogram::Outline);
  m_histogramChartOutput->attach(m_chartDisplay);
  m_histogramChartOutput->setTitle(tr("Output"));

  m_chartDisplay->insertLegend(new QwtLegend(), QwtPlot::RightLegend);
  m_chartDisplay->enableTools(false);

  //connects
  connect(m_ui->m_applyToolButton, SIGNAL(clicked()), this, SLOT(onApplyToolButtonClicked()));
  connect(m_chartDisplay, SIGNAL(leftPointSelected(const QPointF &)), this, SLOT(onLeftPointSelected(const QPointF &)));
  connect(m_chartDisplay, SIGNAL(rigthPointSelected(const QPointF &)), this, SLOT(onRigthPointSelected(const QPointF &)));

  //hide tool bar
  m_ui->m_frame->setVisible(false);
}

te::qt::widgets::RasterHistogramWidget::~RasterHistogramWidget() = default;

Ui::RasterHistogramWidgetForm* te::qt::widgets::RasterHistogramWidget::getForm() const
{
  return m_ui.get();
}

void te::qt::widgets::RasterHistogramWidget::setChartInputColor(int red, int green, int blue)
{
  if( m_histogramChartInput )
    m_histogramChartInput->setBrush(QBrush(QColor(red, green, blue)));
}

void te::qt::widgets::RasterHistogramWidget::setChartOutputColor(int red, int green, int blue)
{
  if( m_histogramChartOutput )
    m_histogramChartOutput->setBrush(QBrush(QColor(red, green, blue)));
}

void te::qt::widgets::RasterHistogramWidget::setEnableTools(bool enable)
{
  m_chartDisplay->enableTools(enable);
}

void te::qt::widgets::RasterHistogramWidget::setEnableNavigationTools(bool enable)
{
  m_chartDisplay->enableMultiSelection(enable);
  m_chartDisplay->enableMagnifier(enable);
  m_chartDisplay->enablePan(enable);
}

void te::qt::widgets::RasterHistogramWidget::setInputRaster(te::rst::Raster* raster, te::gm::Envelope env)
{
  if(!raster)
    return;

  m_inputRaster = raster;

  m_inputEnv = env;

  //set bands from input raster
  m_ui->m_bandComboBox->clear();

  for(unsigned int i = 0; i < m_inputRaster->getNumberOfBands(); ++i)
  {
    QString strBand;
    strBand.setNum(i);

    m_ui->m_bandComboBox->addItem(strBand);
  }
}

void te::qt::widgets::RasterHistogramWidget::setOutputRaster(te::rst::Raster* raster)
{
  m_outputRaster.reset(raster);
}

void te::qt::widgets::RasterHistogramWidget::drawHistogram(int band)
{
  QString toFind = QString::number(band);
  int idx = m_ui->m_bandComboBox->findText(toFind);
  m_ui->m_bandComboBox->setCurrentIndex(idx);

  if(m_inputRaster)
  {
    createHistogram(m_histogramInput, m_inputRaster, band, 0);
    
    m_histogramChartInput->setData();
    m_histogramChartInput->attach(m_chartDisplay);
  }
  else
    m_histogramChartInput->detach();

  if(m_outputRaster.get())
  {
    createHistogram(m_histogramOutput, m_outputRaster.get(), band, 0);

    m_histogramChartOutput->setData();
    m_histogramChartOutput->attach(m_chartDisplay);
  }
  else
    m_histogramChartOutput->detach();

  m_chartDisplay->updateLayout();
  m_chartDisplay->replot();
}

void te::qt::widgets::RasterHistogramWidget::setMinimumValueEnabled(bool enable)
{
  if(m_minValueLine)
  {
    if(enable)
      m_minValueLine->attach(m_chartDisplay);
    else
      m_minValueLine->detach();
  }
}

void te::qt::widgets::RasterHistogramWidget::updateMinimumValueLine(int value, bool replot)
{
  if(!m_minValueLine)
  {
    m_minValueLine = new QwtPlotMarker();
    m_minValueLine->setLabel(QString::fromUtf8("Minimum"));
    m_minValueLine->setLabelAlignment(Qt::AlignLeft | Qt::AlignVCenter);
    m_minValueLine->setLabelOrientation(Qt::Vertical);
    m_minValueLine->setLineStyle(QwtPlotMarker::VLine);
    m_minValueLine->setLinePen(Qt::darkRed, 2, Qt::DashDotLine);
    m_minValueLine->attach(m_chartDisplay);
  }

  m_minValueLine->setXValue(value);

  if(replot)
  {
    m_chartDisplay->updateLayout();

    m_chartDisplay->replot();
  }
}

void te::qt::widgets::RasterHistogramWidget::updateMinimumValueLine(double value, bool replot)
{
  if(!m_minValueLine)
  {
    m_minValueLine = new QwtPlotMarker();
    m_minValueLine->setLabel(QString::fromUtf8("Minimum"));
    m_minValueLine->setLabelAlignment(Qt::AlignLeft | Qt::AlignVCenter);
    m_minValueLine->setLabelOrientation(Qt::Vertical);
    m_minValueLine->setLineStyle(QwtPlotMarker::VLine);
    m_minValueLine->setLinePen(Qt::darkRed, 2, Qt::DashDotDotLine);
    m_minValueLine->attach(m_chartDisplay);
  }

  m_minValueLine->setXValue(value);

  if(replot)
  {
    m_chartDisplay->updateLayout();

    m_chartDisplay->replot();
  }
}

void te::qt::widgets::RasterHistogramWidget::updateMinimumValueLabel(QString value)
{
  if(m_minValueLine)
    m_minValueLine->setLabel(value);
}

void te::qt::widgets::RasterHistogramWidget::setMaximumValueEnabled(bool enable)
{
  if(m_maxValueLine)
  {
    if(enable)
      m_maxValueLine->attach(m_chartDisplay);
    else
      m_maxValueLine->detach();
  }
}

void te::qt::widgets::RasterHistogramWidget::updateMaximumValueLine(int value, bool replot)
{
  if(!m_maxValueLine)
  {
    m_maxValueLine = new QwtPlotMarker();
    m_maxValueLine->setLabel(QString::fromUtf8("Maximum"));
    m_maxValueLine->setLabelAlignment(Qt::AlignLeft | Qt::AlignVCenter);
    m_maxValueLine->setLabelOrientation(Qt::Vertical);
    m_maxValueLine->setLineStyle(QwtPlotMarker::VLine);
    m_maxValueLine->setLinePen(Qt::darkRed, 2, Qt::DashDotLine);
    m_maxValueLine->attach(m_chartDisplay);
  }

  m_maxValueLine->setXValue(value);

  if(replot)
  {
    m_chartDisplay->updateLayout();

    m_chartDisplay->replot();
  }
}

void te::qt::widgets::RasterHistogramWidget::updateMaximumValueLine(double value, bool replot)
{
  if(!m_maxValueLine)
  {
    m_maxValueLine = new QwtPlotMarker();
    m_maxValueLine->setLabel(QString::fromUtf8("Maximum"));
    m_maxValueLine->setLabelAlignment(Qt::AlignLeft | Qt::AlignVCenter);
    m_maxValueLine->setLabelOrientation(Qt::Vertical);
    m_maxValueLine->setLineStyle(QwtPlotMarker::VLine);
    m_maxValueLine->setLinePen(Qt::darkRed, 2, Qt::DashDotDotLine);
    m_maxValueLine->attach(m_chartDisplay);
  }

  m_maxValueLine->setXValue(value);

  if(replot)
  {
    m_chartDisplay->updateLayout();

    m_chartDisplay->replot();
  }
}

void te::qt::widgets::RasterHistogramWidget::updateMaximumValueLabel(QString value)
{
  if(m_maxValueLine)
    m_maxValueLine->setLabel(value);
}

void te::qt::widgets::RasterHistogramWidget::clearMinMaxLines()
{
  delete m_minValueLine;
  m_minValueLine = nullptr;

  delete m_maxValueLine;
  m_maxValueLine = nullptr;
}

void te::qt::widgets::RasterHistogramWidget::clear()
{
  if( m_histogramChartInput ) m_histogramChartInput->detach();
  if( m_histogramChartOutput ) m_histogramChartOutput->detach();
  if( m_histogramInput ) delete m_histogramInput;
  if( m_histogramOutput ) delete m_histogramOutput;

  clearMinMaxLines();

  m_inputRaster = nullptr;
  m_outputRaster.reset(nullptr);

  m_histogramInput = new te::qt::widgets::Histogram();
  m_histogramChartInput = new te::qt::widgets::HistogramChart(m_histogramInput);
  m_histogramChartInput->setPen(Qt::black);
  m_histogramChartInput->setBrush(QBrush(Qt::blue));
  m_histogramChartInput->attach(m_chartDisplay);
  m_histogramChartInput->setTitle(tr("Input"));

  m_histogramOutput = new te::qt::widgets::Histogram();
  m_histogramChartOutput = new te::qt::widgets::HistogramChart(m_histogramOutput);
  m_histogramChartOutput->setPen(Qt::black, 3.);
  m_histogramChartOutput->setBrush(QBrush(QColor(255, 0, 0, 127)));
  m_histogramChartOutput->setStyle(QwtPlotHistogram::Outline);
  m_histogramChartOutput->attach(m_chartDisplay);
  m_histogramChartOutput->setTitle(tr("Output"));

  m_chartDisplay->replot();
}

void te::qt::widgets::RasterHistogramWidget::createHistogram(te::qt::widgets::Histogram* histogram, te::rst::Raster* raster, int bandId, int slices)
{
  histogram->setValues(std::map<te::dt::AbstractData*, unsigned int>());
  
  std::map<double, unsigned int> histogramR;
  
  if (raster == m_inputRaster && m_inputEnv.isValid())
  {
    double llX = m_inputEnv.getLowerLeftX();
    double urY = m_inputEnv.getUpperRightY();
    double urX = m_inputEnv.getUpperRightX();
    double llY = m_inputEnv.getLowerLeftY();

    double ulCol = 0;
    double ulRow = 0;
    raster->getGrid()->geoToGrid( llX, urY, ulCol, ulRow );
    unsigned int firstCol = (unsigned int)std::max( 0.0, ulCol);
    firstCol = std::min( firstCol, raster->getNumberOfColumns() - 1);
    unsigned int firstRow = (unsigned int)std::max( 0.0, ulRow);
    firstRow = std::min( firstRow, raster->getNumberOfRows() - 1);    
    
    double lrCol = 0;
    double lrRow = 0;
    raster->getGrid()->geoToGrid( urX, llY, lrCol, lrRow );
    unsigned int lastCol = (unsigned int)std::max( 0.0, lrCol);
    lastCol = std::min( lastCol, raster->getNumberOfColumns() - 1);
    unsigned int lastRow = (unsigned int)std::max( 0.0, lrRow);
    lastRow = std::min( lastRow, raster->getNumberOfRows() - 1);    
    
    te::rst::BandSummary bandSummary;
    if( ! te::rst::RasterSummaryManager::getInstance().get(
      raster, bandId, firstRow, firstCol, lastRow, lastCol, (unsigned int)slices,
      m_histogramSampleStep, false, bandSummary ) )
    {
      throw te::common::Exception(TE_TR("Raster summary Error"));
    }       
    
    histogramR = *(bandSummary.m_histogramR);      
  }
  else
  {
    te::rst::BandSummary bandSummary;
    if( ! te::rst::RasterSummaryManager::getInstance().get(
      raster, bandId, 0, 0, raster->getNumberOfRows() - 1, 
      raster->getNumberOfColumns() - 1, (unsigned int)slices,
      m_histogramSampleStep, false, bandSummary ) )
    {
      throw te::common::Exception(TE_TR("Raster summary Error"));
    }       
    
    histogramR = *(bandSummary.m_histogramR);         
  }

  if( histogramR.empty() )
  {
    throw te::common::Exception(TE_TR("Raster summary Error"));
  }

  const double min = histogramR.begin()->first;
  const double max = histogramR.rbegin()->first;

  for (std::map<double, unsigned int>::iterator it = histogramR.begin(); it != histogramR.end(); ++it)
  {
    histogram->insert(std::make_pair(new te::dt::Double(it->first), it->second));
  }

  histogram->setMinValue(min);

  if (slices != 0)
    histogram->setInterval((max - min) / ((double)(slices - 1)));

  histogram->setSummarized(false);
}

void te::qt::widgets::RasterHistogramWidget::onApplyToolButtonClicked()
{
  int index = m_ui->m_bandComboBox->currentIndex();

  drawHistogram(index);
}

void te::qt::widgets::RasterHistogramWidget::onLeftPointSelected(const QPointF& point)
{
  emit minValueSelected((int)point.x(), m_ui->m_bandComboBox->currentIndex());

  emit minValueSelected((double)point.x(), m_ui->m_bandComboBox->currentIndex());
}

void te::qt::widgets::RasterHistogramWidget::onRigthPointSelected(const QPointF& point)
{
  emit maxValueSelected((int)point.x(), m_ui->m_bandComboBox->currentIndex());

  emit maxValueSelected((double)point.x(), m_ui->m_bandComboBox->currentIndex());
}

void te::qt::widgets::RasterHistogramWidget::setHistogramInputTitle( 
  const std::string& newTitle )
{
  if( m_histogramChartInput )
    m_histogramChartInput->setTitle(newTitle.c_str());
}

void te::qt::widgets::RasterHistogramWidget::setHistogramOutputTitle( 
  const std::string& newTitle )
{
  if( m_histogramChartOutput )
    m_histogramChartOutput->setTitle(newTitle.c_str());
}

void te::qt::widgets::RasterHistogramWidget::setHistogramSampleStep(
  unsigned int newHistogramSampleStep )
{
  m_histogramSampleStep = newHistogramSampleStep;
}
