/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.
  
    This file is part of the TerraLib - a Framework for building GIS enabled applications.
    
    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.
    
    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.
    
    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

#include "RasterSlicingWizard.h"

//Terralib
#include "../../../common/progress/ProgressManager.h"
#include "../../../common/STLUtils.h"
#include "../../../dataaccess/datasource/DataSourceManager.h"
#include "../../../dataaccess/utils/Utils.h"
#include "../../../rp/Functions.h"
#include "../../../maptools/Utils.h"
#include "../../../se/Categorize.h"
#include "../../../se/ColorMap.h"
#include "../../../se/ParameterValue.h"
#include "../../../se/RasterSymbolizer.h"
#include "../../../se/SelectedChannel.h"
#include "../../../se/Utils.h"
#include "../help/HelpPushButton.h"
#include "../layer/search/LayerSearchWidget.h"
#include "../layer/search/LayerSearchWizardPage.h"
#include "../progress/ProgressViewerDialog.h"
#include "../utils/ScopedCursor.h"
#include "RasterSlicingWizardPage.h"
#include "RasterInfoWizardPage.h"
#include "../raster/RasterInfoWidget.h"
#include "Utils.h"

// Qt
#include <QApplication>
#include <QMessageBox>

// STD
#include <cmath>

te::qt::widgets::RasterSlicingWizard::RasterSlicingWizard(QWidget* parent) :
QWizard(parent)
{
  this->setWizardStyle(QWizard::ModernStyle);
  this->setWindowTitle(tr("Raster Slicing"));

  this->setOption(QWizard::HaveHelpButton, true);
  this->setOption(QWizard::HelpButtonOnRight, false);

  te::qt::widgets::HelpPushButton* helpButton = new te::qt::widgets::HelpPushButton(this);

  this->setButton(QWizard::HelpButton, helpButton);

  helpButton->setPageReference("plugins/rp/rp_rasterslicing.html");

  connect(this, SIGNAL(currentIdChanged(int)), SLOT(onPageChanged(int)));

  addPages();
}

te::qt::widgets::RasterSlicingWizard::~RasterSlicingWizard()
{
  m_wizardPage->clearCanvas();
}

bool te::qt::widgets::RasterSlicingWizard::validateCurrentPage()
{
  if(currentPage() ==  m_layerSearchPage.get())
  {
    te::qt::widgets::ScopedCursor cursor(Qt::WaitCursor);
    std::list<te::map::AbstractLayerPtr> list = m_layerSearchPage->getSearchWidget()->getSelecteds();

    if(list.empty() == false)
    {
      te::map::AbstractLayerPtr layerPtr = *list.begin();
      
      if( layerPtr.get() != nullptr )
        m_wizardPage->setLayer(layerPtr);
    }
    
    return m_layerSearchPage->isComplete();
  }
  else if(currentPage() ==  m_wizardPage.get())
  {
    return m_wizardPage->isComplete();
  }
  else if(currentPage() ==  m_rasterInfoPage.get())
  {
    return execute();
  }  

  return true;
}

void te::qt::widgets::RasterSlicingWizard::setList(std::list<te::map::AbstractLayerPtr> &layerList)
{
  m_layerSearchPage->getSearchWidget()->setList(layerList);
  m_layerSearchPage->getSearchWidget()->filterOnlyByRaster();
}

te::map::AbstractLayerPtr te::qt::widgets::RasterSlicingWizard::getOutputLayer()
{
  return m_outputLayer;
}

void te::qt::widgets::RasterSlicingWizard::setExtent(const te::gm::Envelope& extent)
{
  m_wizardPage->setExtent(extent);
}

void te::qt::widgets::RasterSlicingWizard::setSRID(int srid)
{
  m_wizardPage->setSRID(srid);
}

void te::qt::widgets::RasterSlicingWizard::setMapDisplay(te::qt::widgets::MapDisplay* mapDisplay)
{
  m_wizardPage->setMapDisplay(mapDisplay);
}

void te::qt::widgets::RasterSlicingWizard::setActionGroup(QActionGroup *actionGroup)
{
  m_wizardPage->setActionGroup(actionGroup);
}

void te::qt::widgets::RasterSlicingWizard::onMapDisplayExtentChanged()
{
  m_wizardPage->applyPreview();
}

void te::qt::widgets::RasterSlicingWizard::addPages()
{
  m_layerSearchPage.reset(new LayerSearchWizardPage(this));
  m_layerSearchPage->getSearchWidget()->enableMultiSelection(false);
  addPage(m_layerSearchPage.get());

  m_wizardPage.reset(new RasterSlicingWizardPage(this));
  addPage(m_wizardPage.get());
  
  m_rasterInfoPage.reset(new te::qt::widgets::RasterInfoWizardPage(this));
  addPage(m_rasterInfoPage.get());
}

bool te::qt::widgets::RasterSlicingWizard::execute()
{
  if(m_rasterInfoPage->getWidget()->fileExists())
  {
    QMessageBox::warning(this, tr("Raster Slicing"), tr("File already exists."));
    return false;
  }

  //get layer
  std::list<te::map::AbstractLayerPtr> list = m_layerSearchPage->getSearchWidget()->getSelecteds();
  te::map::AbstractLayerPtr l = *list.begin();

  std::unique_ptr<te::rst::Raster> inputRst;
  
  unsigned int inputRasterBand = 0;
  bool eqHistogram = false;
  m_wizardPage->getParameters( inputRasterBand, eqHistogram );
  
  try
  {
    //progress
    te::qt::widgets::ProgressViewerDialog v(this);

    std::unique_ptr< te::rst::Raster > outRasterPtr;

    bool trimRaster = m_wizardPage->trimRaster();
    
    std::unique_ptr< te::se::ColorMap > colorMap( m_wizardPage->getColorMap() );

    if (trimRaster)
    {
      std::unique_ptr<te::rst::Raster> tempRaster(te::map::GetRaster(l.get()));

      std::map<std::string, std::string> rinfo = tempRaster->getInfo();
      rinfo["FORCE_MEM_DRIVER"] = "TRUE";

      try
      {
        inputRst.reset(tempRaster->trim(&m_wizardPage->getExtent(), rinfo));
      }
      catch(...)
      {
        QMessageBox::critical(this, tr("Raster slicing"), tr("Raster trim error") );
        return false;
      }
    }
    else
    {
      inputRst.reset(te::map::GetRaster(l.get()));
    }
    
    // Checking for continous integer output indexes    
    
    std::vector< double > outIndexes;
    bool isPositiveIntegerOut = true;
    unsigned int minIndex = std::numeric_limits< unsigned int >::max();
    unsigned int maxIndex = 0;
    
    {
      m_wizardPage->getIndexes( outIndexes );
      double truncDouble = 0;
      
      for( std::size_t outIndexesIdx = 0 ; outIndexesIdx < outIndexes.size() ;
        ++outIndexesIdx )
      {
        truncDouble = std::abs( std::trunc( outIndexes[ outIndexesIdx ] ) );
        
        if( outIndexes[ outIndexesIdx ] != truncDouble )
        {
          isPositiveIntegerOut = false;
        }
        
        minIndex = std::min( minIndex, (unsigned int)truncDouble);
        maxIndex = std::max( maxIndex, (unsigned int)truncDouble);
      }
    }
    
    // slices limits
    
    std::vector< double > slice_limits;
    
    {
      std::vector<te::se::ParameterValue*> values = colorMap->getCategorize()->getThresholds();
      
      for (size_t i = 0; i < values.size(); ++i)
      {
        std::string limits = te::se::GetString(values[i]);
        size_t pos = limits.find(" ");
        double min = std::stod(limits.substr(0, pos));
        double max = std::stod(limits.substr(pos));
        slice_limits.push_back( min );
      }
    }

    //Creating the color palette    
    
    std::vector< te::rst::BandProperty::ColorEntry > palette;
    
    if( isPositiveIntegerOut )
    {
      te::rst::BandProperty::ColorEntry color;
      color.c1 = 0;
      color.c2 = 0;
      color.c3 = 0;
      color.c4 = 0;
      
      palette.resize( maxIndex - minIndex + 1, color );     
      
      std::vector<te::se::ParameterValue*> colors = colorMap->getCategorize()->getThresholdValues();

      for (size_t i = 0; i < colors.size(); ++i)
      {
        //Acquiring each color and inserting it in the palette
        std::string colorName = te::se::GetString(colors[i]);
        te::color::RGBAColor rgbac(colorName);

        te::rst::BandProperty::ColorEntry color;
        color.c1 = rgbac.getRed();
        color.c2 = rgbac.getGreen();
        color.c3 = rgbac.getBlue();
        color.c4 = rgbac.getAlpha();
        
        palette[ i ] = color;
      }
    }
    
    // create raster

    {
      te::qt::widgets::ScopedCursor cursor(Qt::WaitCursor);

      if (te::rp::RasterSlicing(*inputRst, inputRasterBand, true,
        static_cast<int>(slice_limits.size()), eqHistogram, m_rasterInfoPage->getWidget()->getInfo(),
        m_rasterInfoPage->getWidget()->getType(), true, palette.empty() ? 0 : &palette,
        outRasterPtr, &slice_limits, &outIndexes))
      {
        outRasterPtr.reset();

        //set output layer
        m_outputLayer = te::qt::widgets::createLayer(m_rasterInfoPage->getWidget()->getType(),
          m_rasterInfoPage->getWidget()->getInfo());
      }
      else
      {
        std::string what = TE_TR("Raster slicing execution error.");
        what += " " + te::rp::Module::getLastLogStr();
        throw te::common::Exception(what);
      }

      te::qt::widgets::applyRasterMultiResolution(tr("Raster slicing"), te::map::GetRaster(m_outputLayer.get()));
    }
  }
  catch (const std::exception& e)
  {
    QMessageBox::critical(this, tr("Raster slicing"), e.what());
    return false;
  }

  QMessageBox::information(this, tr("Raster slicing"), tr("Raster slicing ended sucessfully"));

  emit addLayer(m_outputLayer);

  return true;
}

void te::qt::widgets::RasterSlicingWizard::onPageChanged(int /*index*/)
{
  this->adjustSize();
}
