/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.
  
    This file is part of the TerraLib - a Framework for building GIS enabled applications.
    
    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.
    
    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.
    
    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */


//Terralib
#include "RasterStatisticsWizardPage.h"
#include "ui/ui_RasterStatisticsWizardPageForm.h"
#include "ui/ui_RasterStatisticsWizardLogForm.h"
#include "../../../dataaccess/dataset/DataSet.h"
#include "../../../dataaccess/utils/Utils.h"
#include "../../../classification/ROISet.h"
#include "../../../raster/RasterSummaryManager.h"
#include "../../../rp/Functions.h"
#include "../../widgets/utils/FileDialog.h"
#include "../../widgets/datasource/selector/DataSourceSelectorDialog.h"
#include "../../widgets/charts/Histogram.h"
#include "../../widgets/charts/Scatter.h"
#include "../../widgets/charts/ScatterChart.h"
#include "../../widgets/charts/HistogramChart.h"
#include "../../widgets/charts/ChartStyle.h"
#include "../../widgets/charts/ChartDisplayWidget.h"
#include "../../widgets/charts/ChartDisplay.h"
#include "../../widgets/utils/ScopedCursor.h"
#include "../charts/HistogramDialog.h"
#include "../classification/ROIManagerDialog.h"
#include "../classification/ROIManagerWidget.h"

// Qt
#include <QMessageBox>
#include <QListView>
#include <QProgressDialog>
#include <QBoxLayout>

// STL
#include <algorithm>


te::qt::widgets::RasterStatisticsWizardPage::RasterStatisticsWizardPage(QWidget *parent) :
  QWizardPage(parent)
{
  m_uiStatPageForm.reset(new Ui::RasterStatisticsWizardPageForm);
  m_uiStatPageForm->setupUi(this);
  
  m_logDialogPtr.reset(new te::qt::widgets::RasterStatisticsWizardLog( this ) );

  this->setTitle(tr("Raster Statistics"));
  this->setSubTitle(tr("Calcule of raster statistics"));
  
  m_roiManagerPtr.reset( new te::qt::widgets::ROIManagerDialog( this, Qt::Tool ) );
  
  connect(m_uiStatPageForm->m_histogramPushButton, SIGNAL(clicked(bool)), this,
          SLOT(histogramPushButtonClicked(bool)));
  connect(m_uiStatPageForm->m_scatterPushButton, SIGNAL(clicked(bool)), this,
          SLOT(scatterPushButtonClicked(bool)));  
  connect(m_uiStatPageForm->m_statisticsPushButton, SIGNAL(clicked(bool)), this,
          SLOT(statisticsPushButtonClicked(bool)));  
  connect(m_uiStatPageForm->m_samplesPushButton, SIGNAL(clicked(bool)), this,
          SLOT(samplesPushButtonClicked(bool)));   
  connect(m_uiStatPageForm->m_coStatisticsPushButton, SIGNAL(clicked(bool)), this,
          SLOT(coStatisticsPushButtonClicked(bool)));
  connect(m_roiManagerPtr.get(), SIGNAL(roiManagerClosed(te::cl::ROISet*)), this, 
          SLOT(onROIManagerClosed(te::cl::ROISet*)));
}

te::qt::widgets::RasterStatisticsWizardPage::~RasterStatisticsWizardPage() = default;

bool te::qt::widgets::RasterStatisticsWizardPage::isComplete() const
{
  return true;
}

void te::qt::widgets::RasterStatisticsWizardPage::setLayer(te::map::AbstractLayerPtr layer)
{
  //list bands
  m_uiStatPageForm->m_bandsListWidget->clear();

  //get input raster
  std::unique_ptr<te::da::DataSet> ds = layer->getData();

  if(ds.get())
  {
    std::size_t rpos = te::da::GetFirstPropertyPos(ds.get(), te::dt::RASTER_TYPE);

    std::unique_ptr<te::rst::Raster> inputRst = ds->getRaster(rpos);

    if(inputRst.get())
    {
      m_layer = layer;
      
      m_roiManagerPtr->set(m_layer);

      m_uiStatPageForm->m_layerLineEdit->setText(m_layer->getTitle().c_str());      
      
      for(unsigned int i = 0; i < inputRst->getNumberOfBands(); ++i)
      {
        m_uiStatPageForm->m_bandsListWidget->addItem(QString::number(i));
      }
    }
    else
    {
      QMessageBox::critical(
          this, tr("Warning"),
          tr("Invalid input layer"));
    }
  }
}

void te::qt::widgets::RasterStatisticsWizardPage::histogramPushButtonClicked(
  bool checked )
{
  std::unique_ptr< te::qt::widgets::ScopedCursor > cursorPtr(
    new te::qt::widgets::ScopedCursor( Qt::WaitCursor) );  
  
  if( m_layer.get() == nullptr )
  {
    cursorPtr.reset();
    QMessageBox::critical( this, tr("Warning"), tr("No input layer"));
    return;
  }    
  
  // bands
  
  std::vector< unsigned int > bandsIndexes;
  getBandIndexes( bandsIndexes );
    
  if( bandsIndexes.size() != 1 )
  {
    cursorPtr.reset();
    QMessageBox::critical( this, tr("Warning"),
        tr("Invalid number of selected bands"));
    return;
  }  
  
  // ROIs
  
  std::vector< std::string > roisIDs = getROIsIDs();
  
  const std::string targetROIID = 
    m_uiStatPageForm->m_targetClassComboBox->currentText().toStdString();
    
  if( targetROIID.empty() && ( ! roisIDs.empty() ) )
  {
    cursorPtr.reset();
    QMessageBox::critical( this, tr("Error"),tr("Invalid target ROI label"));
    return;
  }
  
  // histograms
  
  std::map< std::pair< std::string, unsigned int >, std::map<double, 
    unsigned int> > realHistograms;
  std::map< std::pair< std::string, unsigned int >, std::map<double, 
    unsigned int> > imagHistograms;
  if( ! getHistograms( realHistograms, imagHistograms ) )
  {
    cursorPtr.reset();
    QMessageBox::critical( this, tr("Error"),tr("Histograms generation error"));
    return;
  }
  
  std::pair< std::string, unsigned int > auxPair;
  auxPair.first = targetROIID.empty() ? "ALL_IMAGE" : targetROIID;
  auxPair.second = bandsIndexes[ 0 ];  
  
  std::map< std::pair< std::string, unsigned int >, std::map<double, 
    unsigned int> >::const_iterator histogramsIt;
  if( m_uiStatPageForm->m_domainComboBox->currentText() == "Real" )
  {
    histogramsIt = realHistograms.find( auxPair );
    if( histogramsIt == realHistograms.end() )
    {
      cursorPtr.reset();
      QMessageBox::critical( this, tr("Error"),tr("Histograms generation error"));
      return;
    };
  }
  else
  {
    histogramsIt = imagHistograms.find( auxPair );
    if( histogramsIt == imagHistograms.end() )
    {
      cursorPtr.reset();
      QMessageBox::critical( this, tr("Error"),tr("Histograms generation error"));
      return;
    };
  }
    
  const std::map<double, unsigned int>& histogram = histogramsIt->second;
  
  // Histogram dialog
  
  const double min = histogram.begin()->first;
  const double max = histogram.rbegin()->first;
  
  std::unique_ptr< te::qt::widgets::Histogram > newHistogramPtr( 
    new te::qt::widgets::Histogram() );  

  for(std::map<double, unsigned int>::const_iterator it = histogram.begin(); 
    it != histogram.end(); ++it)
  {
     newHistogramPtr->insert( std::make_pair( 
       new te::dt::Double(it->first), it->second) );
  }  

  newHistogramPtr->setMinValue(min);

  newHistogramPtr->setSummarized(false);  
  
  std::unique_ptr< te::qt::widgets::HistogramChart > histoChartPtr( 
    new te::qt::widgets::HistogramChart( newHistogramPtr.release() ) );
  
  std::unique_ptr< te::qt::widgets::ChartStyle > chartStylePtr(
    new te::qt::widgets::ChartStyle() );
  const std::string charStyleWindowTitle = "Histogram - Band " + 
    std::to_string( bandsIndexes[ 0 ] ) + " - ROI " + auxPair.first;
  chartStylePtr->setTitle( charStyleWindowTitle.c_str() );
  chartStylePtr->setAxisX( tr( "Value" ) );
  chartStylePtr->setAxisY( tr( "Frequency" ) );  
  
  QDialog* dialogPtr = new QDialog( this, 0 );
  dialogPtr->setMinimumSize(300, 300);
  QBoxLayout* layout = new QBoxLayout(QBoxLayout::TopToBottom, dialogPtr);
  
  te::qt::widgets::ChartDisplay* chartDisplayPtr = new te::qt::widgets::ChartDisplay(
    dialogPtr, QString::fromUtf8("Histogram"), chartStylePtr.release() );
  histoChartPtr->attach(chartDisplayPtr);    
  chartDisplayPtr->adjustDisplay(); 
  
  layout->addWidget(chartDisplayPtr);
  
  cursorPtr.reset();
  
  dialogPtr->exec();
}

void te::qt::widgets::RasterStatisticsWizardPage::scatterPushButtonClicked(
  bool checked )
{
  std::unique_ptr< te::qt::widgets::ScopedCursor > cursorPtr(
    new te::qt::widgets::ScopedCursor( Qt::WaitCursor) );  
  
  if( m_layer.get() == nullptr )
  {
    cursorPtr.reset();
    QMessageBox::critical( this, tr("Warning"), tr("No input layer"));
    return;
  }    
  
  // bands
  
  std::vector< unsigned int > bandsIndexes;
  getBandIndexes( bandsIndexes );
    
  if( bandsIndexes.size() != 2 )
  {
    cursorPtr.reset();
    QMessageBox::critical( this, tr("Warning"),
        tr("Invalid number of selected bands"));
    return;
  }  
  
  // ROIs
  
  std::vector< std::string > roisIDs = getROIsIDs();
  
  const std::string targetROIID = 
    m_uiStatPageForm->m_targetClassComboBox->currentText().toStdString();
    
  if( targetROIID.empty() && ( ! roisIDs.empty() ) )
  {
    cursorPtr.reset();
    QMessageBox::critical( this, tr("Error"),tr("Invalid target ROI label"));
    return;
  }
    
  // histograms
  
  std::map< std::pair< std::string, unsigned int >, std::map<double, 
    unsigned int> > realHistograms;
  std::map< std::pair< std::string, unsigned int >, std::map<double, 
    unsigned int> > imagHistograms;
  if( ! getHistograms( realHistograms, imagHistograms ) )
  {
    cursorPtr.reset();
    QMessageBox::critical( this, tr("Error"),tr("Histograms generation error"));
    return;
  }
  
  // jointhistograms[ roi_id, band1, band2 ] = frequency
  
  std::map< std::tuple< std::string, unsigned int, unsigned int >,
    std::map< std::pair< double, double >, unsigned int> > realJointHistograms;
  std::map< std::tuple< std::string, unsigned int, unsigned int >,
     std::map< std::pair< double, double >, unsigned int> > imagJointHistograms;
     
  if( ! getJoinHistograms( realHistograms, imagHistograms, realJointHistograms,
    imagJointHistograms ) )
  {
    cursorPtr.reset();
    QMessageBox::critical( this, tr("Error"),tr("Joint histograms generation error"));
    return;
  }
  
  std::map< std::tuple< std::string, unsigned int, unsigned int >,
    std::map< std::pair< double, double >, unsigned int> >::const_iterator
    jointHistogramsIt;
  
  std::tuple< std::string, unsigned int, unsigned int > auxTuple;
  std::get< 0 >( auxTuple ) = targetROIID.empty() ? "ALL_IMAGE" : targetROIID;
  std::get< 1 >( auxTuple ) = bandsIndexes[ 0 ];
  std::get< 2 >( auxTuple ) = bandsIndexes[ 1 ];
  
  if( m_uiStatPageForm->m_domainComboBox->currentText() == "Real" )
  {
    jointHistogramsIt = realJointHistograms.find( auxTuple );
    if( jointHistogramsIt == realJointHistograms.end() )
    {
      cursorPtr.reset();
      QMessageBox::critical( this, tr("Error"),tr("Joint Histograms generation error"));
      return;
    };
  }
  else
  {
    jointHistogramsIt = imagJointHistograms.find( auxTuple );
    if( jointHistogramsIt == imagJointHistograms.end() )
    {
      cursorPtr.reset();
      QMessageBox::critical( this, tr("Error"),tr("Joint Histograms generation error"));
      return;
    };
  }  
  
  const std::map< std::pair< double, double >, unsigned int>& jointHistogram 
    = jointHistogramsIt->second;
  
  // Histogram dialog
  
  std::unique_ptr< te::qt::widgets::Scatter > newScatterPtr(
    new te::qt::widgets::Scatter() );

  for(std::map< std::pair< double, double >, unsigned int>::const_iterator it = 
    jointHistogram.begin(); it != jointHistogram.end(); ++it)
  {
    double x = it->first.first;
    double y = it->first.second;
    newScatterPtr->addX( x );
    newScatterPtr->addY( y );
  }
  
  newScatterPtr->calculateMinMaxValues();
  
  std::unique_ptr< te::qt::widgets::ScatterChart > scatterChartPtr( 
    new te::qt::widgets::ScatterChart( newScatterPtr.release() ) );
  
  std::unique_ptr< te::qt::widgets::ChartStyle > chartStylePtr(
    new te::qt::widgets::ChartStyle() );
  const std::string charStyleWindowTitle = "Scatter - Band " + 
    std::to_string( bandsIndexes[ 0 ] ) 
    + " X Band " + std::to_string( bandsIndexes[ 1 ] ) 
    + " - ROI " + std::get< 0 >( auxTuple );
  chartStylePtr->setTitle( charStyleWindowTitle.c_str() );
  std::string axisXText( "Band " + std::to_string( bandsIndexes[ 0 ] ) + " value" );
  std::string axisYText( "Band " + std::to_string( bandsIndexes[ 1 ] ) + " value" );
  chartStylePtr->setAxisX( axisXText.c_str() );
  chartStylePtr->setAxisY( axisYText.c_str() );  
  
  QDialog* dialogPtr = new QDialog( this, 0 );
  dialogPtr->setMinimumSize(300, 300);
  QBoxLayout* layout = new QBoxLayout(QBoxLayout::TopToBottom, dialogPtr);
  
  te::qt::widgets::ChartDisplay* chartDisplayPtr = new te::qt::widgets::ChartDisplay(
    dialogPtr, QString::fromUtf8("Histogram"), chartStylePtr.release() );
  scatterChartPtr->attach(chartDisplayPtr);    
  chartDisplayPtr->adjustDisplay(); 
  
  layout->addWidget(chartDisplayPtr);
  
  cursorPtr.reset();
  
  dialogPtr->exec();
}

void te::qt::widgets::RasterStatisticsWizardPage::statisticsPushButtonClicked(
  bool checked )
{
  std::unique_ptr< te::qt::widgets::ScopedCursor > cursorPtr(
    new te::qt::widgets::ScopedCursor( Qt::WaitCursor) );  
  
  if( m_layer.get() == nullptr )
  {
    cursorPtr.reset();
    QMessageBox::critical( this, tr("Warning"), tr("No input layer"));
    return;
  }    
  
  // bands
  
  std::vector< unsigned int > bandsIndexes;
  getBandIndexes( bandsIndexes );
    
  if( bandsIndexes.empty() )
  {
    cursorPtr.reset();
    QMessageBox::critical( this, tr("Warning"),
        tr("Invalid number of selected bands"));
    return;
  }

  // histograms
  
  std::map< std::pair< std::string, unsigned int >, std::map<double, 
    unsigned int> > realHistograms;
  std::map< std::pair< std::string, unsigned int >, std::map<double, 
    unsigned int> > imagHistograms;
  
  if( ! getHistograms( realHistograms, imagHistograms ) )
  {
    cursorPtr.reset();
    QMessageBox::critical( this, tr("Error"),tr("Histograms generation error"));
    return;
  }
  
  // ROIs
  
  std::vector< std::string > roisIDs = getROIsIDs();
  if( roisIDs.empty() ) roisIDs.push_back( "ALL_IMAGE" );
  
  const std::string targetROIID = 
    m_uiStatPageForm->m_targetClassComboBox->currentText().toStdString();   
  
  // Generate statistics text
  
  {
    m_logDialogPtr->clearAllText();
    
    std::map< std::pair< std::string, unsigned int >, std::map<double, 
      unsigned int> >::const_iterator histogramsIt;
    std::map< std::pair< std::string, unsigned int >, std::map<double, 
      unsigned int> >::const_iterator histogramsItEnd;
    std::pair< std::string, unsigned int > auxPair;
    
    for( std::size_t roisIDsIdx = 0 ; roisIDsIdx < roisIDs.size() ; ++roisIDsIdx )
    {
      const std::string& roiID = roisIDs[ roisIDsIdx ];
      
      if( 
          targetROIID.empty()
          ||
          (
            (!targetROIID.empty())
            &&
            ( roiID == targetROIID )
          )
        )
      {
        m_logDialogPtr->addTextLine( "---------------------------------------------" );
        m_logDialogPtr->addTextLine( "ROI ID: " + roiID );
        m_logDialogPtr->addTextLine( "Domain: " + 
          m_uiStatPageForm->m_domainComboBox->currentText().toStdString() );
        
        for( std::size_t bandsIndexesIdx = 0 ; bandsIndexesIdx < bandsIndexes.size() ;
          ++bandsIndexesIdx )
        {
          const unsigned int bandIndex = bandsIndexes[ bandsIndexesIdx ];
          
          auxPair.first = roiID;
          auxPair.second = bandIndex;
          if( m_uiStatPageForm->m_domainComboBox->currentText() == "Real" )
          {
            histogramsIt = realHistograms.find( auxPair );
            histogramsItEnd = realHistograms.end();
          }
          else
          {
            histogramsIt = imagHistograms.find( auxPair );
            histogramsItEnd = imagHistograms.end();
          }
          if( histogramsIt == histogramsItEnd )
          {
            cursorPtr.reset();
            QMessageBox::critical( this, tr("Error"),tr("Histograms generation error"));
            return;
          }        
          
          m_logDialogPtr->addTextLine( " " ); 
          m_logDialogPtr->addTextLine( "Band: " + std::to_string( bandIndex ) );
          
          const std::map<double, unsigned int>& histogram =
            histogramsIt->second;
            
          double min = 0;
          double max = 0;
          double mean = 0;
          double stdDev = 0;
          double mode = 0;
          double entropy = 0;
          double sum = 0;
          double sum2 = 0;
          double sum3 = 0;
          double sum4 = 0;
          double variance = 0;
          double median = 0;
          te::rp::getHistogramStats( 
            histogram, 
            min, max, mean, stdDev, mode, entropy, sum, sum2, 
            sum3, sum4, variance, median );
          
          m_logDialogPtr->addTextLine( "    Minimum:" + std::to_string( min ) );
          m_logDialogPtr->addTextLine( "    Maximum:" + std::to_string( max ) );
          m_logDialogPtr->addTextLine( "    Mean:" + std::to_string( mean ) );
          m_logDialogPtr->addTextLine( "    Standart Deviation:" + std::to_string( stdDev ) );
          m_logDialogPtr->addTextLine( "    Mode:" + std::to_string( mode ) );
          m_logDialogPtr->addTextLine( "    Entropy:" + std::to_string( entropy ) );
          m_logDialogPtr->addTextLine( "    Sum:" + std::to_string( sum ) );
          m_logDialogPtr->addTextLine( "    Quadratic sum:" + std::to_string( sum2 ) );
          m_logDialogPtr->addTextLine( "    Cubic sum:" + std::to_string( sum3 ) );
          m_logDialogPtr->addTextLine( "    Quartic sum:" + std::to_string( sum4 ) );
          m_logDialogPtr->addTextLine( "    Variance:" + std::to_string( variance ) );
          m_logDialogPtr->addTextLine( "    Median:" + std::to_string( median ) );
        }
      }
    }
  }
  
  cursorPtr.reset();
  
  m_logDialogPtr->exec();
}

void te::qt::widgets::RasterStatisticsWizardPage::samplesPushButtonClicked(
  bool checked)
{
  if( m_layer.get() == nullptr )
  {
    QMessageBox::critical( this, tr("Warning"), tr("No input layer"));
    return;
  }  
  
  m_roiManagerPtr->show();
}

void te::qt::widgets::RasterStatisticsWizardPage::setMapDisplay(te::qt::widgets::MapDisplay* mapDisplay)
{
  m_roiManagerPtr->setMapDisplay(mapDisplay);
}

void te::qt::widgets::RasterStatisticsWizardPage::coStatisticsPushButtonClicked(
  bool checked )
{
  std::unique_ptr< te::qt::widgets::ScopedCursor > cursorPtr(
    new te::qt::widgets::ScopedCursor( Qt::WaitCursor) );
  
  if( m_layer.get() == nullptr )
  {
    cursorPtr.reset();
    QMessageBox::critical( this, tr("Warning"), tr("No input layer"));
    return;
  }    
  
  // bands
  
  std::vector< unsigned int > bandsIndexes;
  std::string bandsIndexesStr;
  getBandIndexes( bandsIndexes );
    
  if( bandsIndexes.empty() )
  {
    cursorPtr.reset();
    QMessageBox::critical( this, tr("Warning"),
        tr("Invalid number of selected bands"));
    return;
  }
  else
  {
    for( std::size_t bandsIndexesIdx = 0 ; bandsIndexesIdx < bandsIndexes.size() ;
      ++bandsIndexesIdx )
    {
      if( bandsIndexesIdx ) bandsIndexesStr.append( " " );
      bandsIndexesStr.append( std::to_string( bandsIndexes[ bandsIndexesIdx ] ) );
    }
  }
  
  // ROIs
  
  std::vector< std::string > roisIDs = getROIsIDs();
  if( roisIDs.empty() ) roisIDs.push_back( "ALL_IMAGE" );
  
  const std::string targetROIID = 
    m_uiStatPageForm->m_targetClassComboBox->currentText().toStdString();     
  
  // histograms
  
  std::unique_ptr< QProgressDialog > progressPtr( new QProgressDialog( 
    "Processing data", QString(), 0, 100, this, 0 ) );
  progressPtr->setAutoClose( false );
  progressPtr->setAutoReset( false );
  progressPtr->setMinimumDuration( 0 );
  progressPtr->setValue( 1 );
  
  std::map< std::pair< std::string, unsigned int >, std::map<double, 
    unsigned int> > realHistograms;
  std::map< std::pair< std::string, unsigned int >, std::map<double, 
    unsigned int> > imagHistograms;
  
  if( ! getHistograms( realHistograms, imagHistograms ) )
  {
    cursorPtr.reset();
    QMessageBox::critical( this, tr("Error"),tr("Histograms generation error"));
    return;
  }
  
  // jointhistograms[ roi_id, band1, band2 ] = frequency
  
  std::map< std::tuple< std::string, unsigned int, unsigned int >,
    std::map< std::pair< double, double >, unsigned int> > realJointHistograms;
  std::map< std::tuple< std::string, unsigned int, unsigned int >,
     std::map< std::pair< double, double >, unsigned int> > imagJointHistograms;
     
  if( ! getJoinHistograms( realHistograms, imagHistograms, realJointHistograms,
    imagJointHistograms ) )
  {
    cursorPtr.reset();
    QMessageBox::critical( this, tr("Error"),tr("Joint histograms generation error"));
    return;
  }
  
  // Generate statistics text
  
  {
    m_logDialogPtr->clearAllText();
    
    for( std::size_t roisIDsIdx = 0 ; roisIDsIdx < roisIDs.size() ; ++roisIDsIdx )
    {
      const std::string& roiID = roisIDs[ roisIDsIdx ];    
      
      if( 
          targetROIID.empty()
          ||
          (
            ( !targetROIID.empty() )
            &&
            ( roiID == targetROIID )
          )
        )
      {      
        m_logDialogPtr->addTextLine( "---------------------------------------------" );
        m_logDialogPtr->addTextLine( "ROI ID: " + roiID );
        m_logDialogPtr->addTextLine( "Bands: " + bandsIndexesStr );
        m_logDialogPtr->addTextLine( "Domain: " + 
          m_uiStatPageForm->m_domainComboBox->currentText().toStdString() );
        m_logDialogPtr->addTextLine( " " );
        
        // Covariance matrix
        
        te::rp::Matrix< double > covarianceMatrix;
        if( m_uiStatPageForm->m_domainComboBox->currentText() == "Real" )
        {
          if( ! getCovarianceMatrix( roiID, realJointHistograms, realHistograms,
            covarianceMatrix ) )
          {
            cursorPtr.reset();
            QMessageBox::critical( this, tr("Error"),tr("Covariance matrix calcule error"));
            return;
          }
        }
        else
        {
          if( ! getCovarianceMatrix( roiID, imagJointHistograms, imagHistograms,
            covarianceMatrix ) )
          {
            cursorPtr.reset();
            QMessageBox::critical( this, tr("Error"),tr("Covariance matrix calcule error"));
            return;
          }
        }      
        m_logDialogPtr->addTextLine( " " );
        m_logDialogPtr->addTextLine( "Covariance matrix:" );
        printMatrix( covarianceMatrix );

        // Correlation matrix
        
        te::rp::Matrix< double > correlationMatrix;
        if( m_uiStatPageForm->m_domainComboBox->currentText() == "Real" )
        {
          if( ! getCorrelationMatrix( roiID, realJointHistograms, realHistograms,
            correlationMatrix ) )
          {
            cursorPtr.reset();
            QMessageBox::critical( this, tr("Error"),tr("Covariance matrix calcule error"));
            return;
          }
        }
        else
        {
          if( ! getCorrelationMatrix( roiID, imagJointHistograms, imagHistograms,
            correlationMatrix ) )
          {
            cursorPtr.reset();
            QMessageBox::critical( this, tr("Error"),tr("Covariance matrix calcule error"));
            return;
          }
        }      
        m_logDialogPtr->addTextLine( " " );
        m_logDialogPtr->addTextLine( "Correlation matrix:" );
        printMatrix( correlationMatrix );            
        
        // Kurtosis matrix
        
        te::rp::Matrix< double > kurtosisMatrix;
        if( m_uiStatPageForm->m_domainComboBox->currentText() == "Real" )
        {
          if( ! getKurtosisMatrix( roiID, realJointHistograms, realHistograms,
            kurtosisMatrix ) )
          {
            cursorPtr.reset();
            QMessageBox::critical( this, tr("Error"),tr("Kurtosis matrix calcule error"));
            return;
          }
        }
        else
        {
          if( ! getKurtosisMatrix( roiID, imagJointHistograms, imagHistograms,
            kurtosisMatrix ) )
          {
            cursorPtr.reset();
            QMessageBox::critical( this, tr("Error"),tr("Kurtosis matrix calcule error"));
            return;
          }
        }      
        m_logDialogPtr->addTextLine( " " );
        m_logDialogPtr->addTextLine( "Kurtosis matrix:" );
        printMatrix( kurtosisMatrix );
      }
    }
  }
  
  progressPtr.reset();
  
  cursorPtr.reset();
  
  m_logDialogPtr->exec();
}

void te::qt::widgets::RasterStatisticsWizardPage::getBandIndexes( 
  std::vector< unsigned int >& bandsIndexes ) const
{
  QList<QListWidgetItem *> selectedBandsItems =
    m_uiStatPageForm->m_bandsListWidget->selectedItems();
  
  QList<QListWidgetItem *>::const_iterator it = selectedBandsItems.begin();
  const QList<QListWidgetItem *>::const_iterator itEnd = selectedBandsItems.end();
  
  while( it != itEnd )
  {
    bandsIndexes.push_back( (*it)->text().toUInt() );
    ++it;
  }
  
  std::sort( bandsIndexes.begin(), bandsIndexes.end() );
}

bool te::qt::widgets::RasterStatisticsWizardPage::getHistograms(
  std::map< std::pair< std::string, unsigned int >, std::map<double, 
    unsigned int> >& realHistograms,
  std::map< std::pair< std::string, unsigned int >, std::map<double, 
    unsigned int> >& imagHistograms ) const
{
  realHistograms.clear();
  imagHistograms.clear();
  
  if( m_layer.get() == nullptr )
  {
    return false;
  }    
  
  // bands
  
  std::vector< unsigned int > bandsIndexes;
  getBandIndexes( bandsIndexes );
    
  if( bandsIndexes.empty() )
  {
    return false;
  }
  
  // ROIs
  
  std::map<std::string, te::cl::ROI*> roiMap;
  if( m_roiManagerPtr->getWidget()->getROISet() )
  {
    roiMap = m_roiManagerPtr->getWidget()->getROISet()->getROISet();
  }
  
  // histograms[ roi_id ][ band ][ pixel_value ] = frequency
  
  {  
    std::unique_ptr<te::da::DataSet> ds = m_layer->getData();
    std::size_t rpos = te::da::GetFirstPropertyPos(ds.get(), te::dt::RASTER_TYPE);
    std::unique_ptr<te::rst::Raster> inputRst = ds->getRaster(rpos);    
    
    if( roiMap.empty() )
    { // full image
      for( std::size_t bandsIndexesIdx = 0 ; bandsIndexesIdx < bandsIndexes.size() ;
        ++bandsIndexesIdx )
      {
        te::rst::BandSummary bandSummary;
        te::rst::RasterSummaryManager::getInstance().get( inputRst.get(),
          bandsIndexes[ bandsIndexesIdx ], false, bandSummary );
        realHistograms[ std::pair< std::string, unsigned int >( "ALL_IMAGE",
          bandsIndexes[ bandsIndexesIdx ] ) ] = (*bandSummary.m_histogramR );
        imagHistograms[ std::pair< std::string, unsigned int >( "ALL_IMAGE",
          bandsIndexes[ bandsIndexesIdx ] ) ] = ( *bandSummary.m_histogramI );
      }
    }
    else
    {
      const std::string targetROIID = 
        m_uiStatPageForm->m_targetClassComboBox->currentText().toStdString();          
      std::map<std::string, te::gm::Polygon*>::const_iterator roiPolygonsMapIt;
      std::map<std::string, te::gm::Polygon*>::const_iterator roiPolygonsMapItEnd;      
      std::vector< te::gm::Polygon const* > roisPols;
      std::string roiID;
      std::map<double, unsigned> rHistogram;
      std::map<double, unsigned> iHistogram;
      std::map<std::string, te::cl::ROI*>::const_iterator roiMapIt = roiMap.begin();
      const std::map<std::string, te::cl::ROI*>::const_iterator roiMapItEnd = 
        roiMap.end();
        
      while( roiMapIt != roiMapItEnd )
      {
        roiID = roiMapIt->first;
        
        if( 
            targetROIID.empty()
            ||
            (
              ( !targetROIID.empty() )
              &&
              ( roiID == targetROIID )
            )
          )
        {
          const std::map<std::string, te::gm::Polygon*>& roiPolygonsMap = 
            roiMapIt->second->getPolygons();
          roiPolygonsMapIt = roiPolygonsMap.begin();
          roiPolygonsMapItEnd = roiPolygonsMap.end();
          
          roisPols.clear();
          while( roiPolygonsMapIt != roiPolygonsMapItEnd )
          {
            roisPols.push_back( roiPolygonsMapIt->second );
            ++roiPolygonsMapIt;
          }
          
          for( std::size_t bandsIndexesIdx = 0 ; bandsIndexesIdx < bandsIndexes.size() ;
            ++bandsIndexesIdx )
          {
            if( ! te::rp::getHistograms( 
              *inputRst, 
              bandsIndexes[ bandsIndexesIdx ],
              roisPols, 
              0, // histoBins
              0, // maxThreads
              te::rst::PolygonIterator< double >::IterationType::ScanLineIterationT,
              rHistogram, 
              iHistogram ) )
            {
              return false;
            }
            
            realHistograms[ std::pair< std::string, unsigned int >( roiID,
              bandsIndexes[ bandsIndexesIdx ] ) ] = rHistogram;
            imagHistograms[ std::pair< std::string, unsigned int >( roiID,
              bandsIndexes[ bandsIndexesIdx ] ) ] = iHistogram;
          }
        }
        
        ++roiMapIt;
      }
    }
  }
  
  return true;
}

bool te::qt::widgets::RasterStatisticsWizardPage::getJoinHistograms(
  const std::map< std::pair< std::string, unsigned int >, std::map<double, 
    unsigned int> >& realHistograms,
  const std::map< std::pair< std::string, unsigned int >, std::map<double, 
    unsigned int> >& imagHistograms,
  std::map< std::tuple< std::string, unsigned int, unsigned int >,
    std::map< std::pair< double, double >, unsigned int> >& realJointHistograms,
  std::map< std::tuple< std::string, unsigned int, unsigned int >,
    std::map< std::pair< double, double >, unsigned int> >& imagJointHistograms ) const
{
  realJointHistograms.clear();
  imagJointHistograms.clear();
  
  // bands
  
  std::vector< unsigned int > bandsIndexes;
  getBandIndexes( bandsIndexes );
  if( bandsIndexes.empty() ) return false;
  
  // ROIs
  
  std::map<std::string, te::cl::ROI*> roiMap;
  if( m_roiManagerPtr->getWidget()->getROISet() )
  {
    roiMap = m_roiManagerPtr->getWidget()->getROISet()->getROISet();
  }  
  
  // joint histograms
   
  std::unique_ptr<te::da::DataSet> ds = m_layer->getData();
  std::size_t rpos = te::da::GetFirstPropertyPos(ds.get(), te::dt::RASTER_TYPE);
  std::unique_ptr<te::rst::Raster> inputRst = ds->getRaster(rpos);    

  if( roiMap.empty() )
  { // full image
    std::vector< te::gm::Polygon const* > roisPols;
    std::map< std::pair< std::string, unsigned int >, std::map<double, 
      unsigned int> >::const_iterator histoIt;
    
    for( std::size_t bandsIndexesIdx1 = 0 ; bandsIndexesIdx1 < bandsIndexes.size() ;
      ++bandsIndexesIdx1 )
    {
      for( std::size_t bandsIndexesIdx2 = 0 ; bandsIndexesIdx2 < bandsIndexes.size() ;
        ++bandsIndexesIdx2 )
      {
        histoIt = realHistograms.find( std::pair< std::string, unsigned int >( 
          "ALL_IMAGE", bandsIndexes[ bandsIndexesIdx1 ] ) );
        if(  histoIt == realHistograms.end() ) return false;
        const std::map<double, unsigned int>& realHist1 = 
          histoIt->second;
          
        histoIt = realHistograms.find( std::pair< std::string, unsigned int >( 
          "ALL_IMAGE", bandsIndexes[ bandsIndexesIdx2 ] ) );
        if(  histoIt == realHistograms.end() ) return false;
        const std::map<double, unsigned int>& realHist2 = 
          histoIt->second;
          
        histoIt = imagHistograms.find( std::pair< std::string, unsigned int >( 
          "ALL_IMAGE", bandsIndexes[ bandsIndexesIdx1 ] ) );
        if(  histoIt == imagHistograms.end() ) return false;
        const std::map<double, unsigned int>& imagHist1 = 
          histoIt->second;
          
        histoIt = imagHistograms.find( std::pair< std::string, unsigned int >( 
          "ALL_IMAGE", bandsIndexes[ bandsIndexesIdx2 ] ) );
        if(  histoIt == imagHistograms.end() ) return false;
        const std::map<double, unsigned int>& imagHist2 = 
          histoIt->second;          
        
        std::map< std::pair< double, double >, unsigned int> realJointHistogram;
        std::map< std::pair< double, double >, unsigned int> imagJointHistogram;
      
        if( ! te::rp::getJointHistograms( 
          *inputRst, 
          bandsIndexes[ bandsIndexesIdx1 ],
          bandsIndexes[ bandsIndexesIdx2 ],
          roisPols, 
          0, // histoBins
          0, // maxThreads
          te::rst::PolygonIterator< double >::IterationType::ScanLineIterationT,
          realJointHistogram, 
          imagJointHistogram,
          &realHist1,
          &imagHist1,
          &realHist2,
          &imagHist2 ) )
        {
          return false;
        }
        
        std::tuple< std::string, unsigned int, unsigned int > auxTuple;
        std::get< 0 >( auxTuple ) = "ALL_IMAGE";
        std::get< 1 >( auxTuple ) = bandsIndexes[ bandsIndexesIdx1 ];
        std::get< 2 >( auxTuple ) = bandsIndexes[ bandsIndexesIdx2 ];
        
        realJointHistograms[ auxTuple ] = realJointHistogram;
        imagJointHistograms[ auxTuple ] = imagJointHistogram;
      }
    }
  }
  else
  {
    std::map<std::string, te::gm::Polygon*>::const_iterator roiPolygonsMapIt;
    std::map<std::string, te::gm::Polygon*>::const_iterator roiPolygonsMapItEnd;   
    std::vector< te::gm::Polygon const* > roisPols;
    std::string roiID;
    std::map<std::string, te::cl::ROI*>::const_iterator roiMapIt = roiMap.begin();
    const std::map<std::string, te::cl::ROI*>::const_iterator roiMapItEnd = 
      roiMap.end();
    const std::string targetROIID = 
      m_uiStatPageForm->m_targetClassComboBox->currentText().toStdString();
      
    while( roiMapIt != roiMapItEnd )
    {
      roiID = roiMapIt->first;
      
      if( 
          targetROIID.empty()
          ||
          (
            ( !targetROIID.empty() )
            &&
            ( roiID == targetROIID )
          )
        )
      {
        const std::map<std::string, te::gm::Polygon*>& roiPolygonsMap = 
          roiMapIt->second->getPolygons();
        roiPolygonsMapIt = roiPolygonsMap.begin();
        roiPolygonsMapItEnd = roiPolygonsMap.end();
        
        roisPols.clear();
        while( roiPolygonsMapIt != roiPolygonsMapItEnd )
        {
          roisPols.push_back( roiPolygonsMapIt->second );
          ++roiPolygonsMapIt;
        }
        
        std::map< std::pair< std::string, unsigned int >, std::map<double, 
          unsigned int> >::const_iterator histoIt;
        
        for( std::size_t bandsIndexesIdx1 = 0 ; bandsIndexesIdx1 < bandsIndexes.size() ;
          ++bandsIndexesIdx1 )
        {
          for( std::size_t bandsIndexesIdx2 = 0 ; bandsIndexesIdx2 < bandsIndexes.size() ;
            ++bandsIndexesIdx2 )
          {
            histoIt = realHistograms.find( std::pair< std::string, unsigned int >( 
              roiID, bandsIndexes[ bandsIndexesIdx1 ] ) );
            if(  histoIt == realHistograms.end() ) return false;
            const std::map<double, unsigned int>& realHist1 = 
              histoIt->second;
              
            histoIt = realHistograms.find( std::pair< std::string, unsigned int >( 
              roiID, bandsIndexes[ bandsIndexesIdx2 ] ) );
            if(  histoIt == realHistograms.end() ) return false;
            const std::map<double, unsigned int>& realHist2 = 
              histoIt->second;
              
            histoIt = imagHistograms.find( std::pair< std::string, unsigned int >( 
              roiID, bandsIndexes[ bandsIndexesIdx1 ] ) );
            if(  histoIt == imagHistograms.end() ) return false;
            const std::map<double, unsigned int>& imagHist1 = 
              histoIt->second;
              
            histoIt = imagHistograms.find( std::pair< std::string, unsigned int >( 
              roiID, bandsIndexes[ bandsIndexesIdx2 ] ) );
            if(  histoIt == imagHistograms.end() ) return false;
            const std::map<double, unsigned int>& imagHist2 = 
              histoIt->second;          
            
            std::map< std::pair< double, double >, unsigned int> realJointHistogram;
            std::map< std::pair< double, double >, unsigned int> imagJointHistogram;
          
            if( ! te::rp::getJointHistograms( 
              *inputRst, 
              bandsIndexes[ bandsIndexesIdx1 ],
              bandsIndexes[ bandsIndexesIdx2 ],
              roisPols, 
              0, // histoBins
              0, // maxThreads
              te::rst::PolygonIterator< double >::IterationType::ScanLineIterationT,
              realJointHistogram, 
              imagJointHistogram,
              &realHist1,
              &imagHist1,
              &realHist2,
              &imagHist2 ) )
            {
              return false;
            }
            
            std::tuple< std::string, unsigned int, unsigned int > auxTuple;
            std::get< 0 >( auxTuple ) = roiID;
            std::get< 1 >( auxTuple ) = bandsIndexes[ bandsIndexesIdx1 ];
            std::get< 2 >( auxTuple ) = bandsIndexes[ bandsIndexesIdx2 ];
            
            realJointHistograms[ auxTuple ] = realJointHistogram;
            imagJointHistograms[ auxTuple ] = imagJointHistogram;
          }
        }
      }
      
      ++roiMapIt;
    }
  }
  
  return true;   
}

bool te::qt::widgets::RasterStatisticsWizardPage::getCovarianceMatrix(
  const std::string& roiID,
  const std::map< std::tuple< std::string, unsigned int, unsigned int >,
    std::map< std::pair< double, double >, unsigned int> >& jointHistograms,
  const std::map< std::pair< std::string, unsigned int >, std::map<double, 
    unsigned int> >& histograms,    
  te::rp::Matrix< double >& matrix )
{
  // bands
  
  std::vector< unsigned int > bandsIndexes;
  getBandIndexes( bandsIndexes );
    
  if( bandsIndexes.empty() )
  {
    QMessageBox::critical( this, tr("Warning"),
        tr("Invalid number of selected bands"));
    return false;
  }
  
  const std::size_t bandsIndexesSize = bandsIndexes.size()  ;
  
  if( !matrix.reset( bandsIndexes.size(), bandsIndexes.size() ) )
  {
    return false;
  }
  
  for( std::size_t bandsIndexesIdx1 = 0 ; bandsIndexesIdx1 < bandsIndexesSize ;
    ++bandsIndexesIdx1 )
  {
    const unsigned int bandIdx1 = bandsIndexes[ bandsIndexesIdx1 ];
    
    for( std::size_t bandsIndexesIdx2 = 0 ; bandsIndexesIdx2 < bandsIndexesSize ;
      ++bandsIndexesIdx2 )
    {
      const unsigned int bandIdx2 = bandsIndexes[ bandsIndexesIdx2 ];

    if( bandsIndexesIdx2 >= bandsIndexesIdx1 )
      {
        if( ! getCovariance( roiID, bandIdx1, bandIdx2, jointHistograms,
          histograms, matrix( bandsIndexesIdx1, bandsIndexesIdx2 ) ) )
        {
          return false;
        }
        
        matrix( bandsIndexesIdx2, bandsIndexesIdx1 ) =
          matrix( bandsIndexesIdx1, bandsIndexesIdx2 );
      }
    }
  }
  
  return true;
}

bool te::qt::widgets::RasterStatisticsWizardPage::getCovariance(
  const std::string& roiID,
  const unsigned int& bandIdx1,
  const unsigned int& bandIdx2,            
  const std::map< std::tuple< std::string, unsigned int, unsigned int >,
    std::map< std::pair< double, double >, unsigned int> >& jointHistograms,
  const std::map< std::pair< std::string, unsigned int >, std::map<double, 
    unsigned int> >& histograms,
  double& covariance )
{
  std::map< std::tuple< std::string, unsigned int, unsigned int >,
    std::map< std::pair< double, double >, unsigned int> >::const_iterator 
    jointHistogramsIt;
  std::map< std::pair< std::string, unsigned int >, std::map<double, 
    unsigned int> >::const_iterator histogramsIt1;
  std::map< std::pair< std::string, unsigned int >, std::map<double, 
    unsigned int> >::const_iterator histogramsIt2;    
  std::tuple< std::string, unsigned int, unsigned int > auxTuple;
  std::pair< std::string, unsigned int > auxPair;
  
  std::get< 0 >( auxTuple ) = roiID;
  std::get< 1 >( auxTuple ) = bandIdx1;
  std::get< 2 >( auxTuple ) = bandIdx2;
  
  jointHistogramsIt = jointHistograms.find( auxTuple );
  if( jointHistogramsIt == jointHistograms.end() )
  {
    return false;
  }
  
  auxPair.first = roiID;
  auxPair.second = bandIdx1;
  
  histogramsIt1 = histograms.find( auxPair );
  if( histogramsIt1 == histograms.end() )
  {
    return false;
  }
  
  auxPair.second = bandIdx2;
  
  histogramsIt2 = histograms.find( auxPair );
  if( histogramsIt2 == histograms.end() )
  {
    return false;
  }
  
  double corrlationCoef = 0;
  if( ! te::rp::getJointHistogramStats( jointHistogramsIt->second,
    histogramsIt1->second, histogramsIt2->second, covariance, corrlationCoef ) )
  {
    return false;
  }
  
  return true;
}

bool te::qt::widgets::RasterStatisticsWizardPage::getCorrelation(
  const std::string& roiID,
  const unsigned int& bandIdx1,
  const unsigned int& bandIdx2,            
  const std::map< std::tuple< std::string, unsigned int, unsigned int >,
    std::map< std::pair< double, double >, unsigned int> >& jointHistograms,
  const std::map< std::pair< std::string, unsigned int >, std::map<double, 
    unsigned int> >& histograms,
  double& correlation )
{
  std::map< std::tuple< std::string, unsigned int, unsigned int >,
    std::map< std::pair< double, double >, unsigned int> >::const_iterator 
    jointHistogramsIt;
  std::map< std::pair< std::string, unsigned int >, std::map<double, 
    unsigned int> >::const_iterator histogramsIt1;
  std::map< std::pair< std::string, unsigned int >, std::map<double, 
    unsigned int> >::const_iterator histogramsIt2;    
  std::tuple< std::string, unsigned int, unsigned int > auxTuple;
  std::pair< std::string, unsigned int > auxPair;
  
  std::get< 0 >( auxTuple ) = roiID;
  std::get< 1 >( auxTuple ) = bandIdx1;
  std::get< 2 >( auxTuple ) = bandIdx2;
  
  jointHistogramsIt = jointHistograms.find( auxTuple );
  if( jointHistogramsIt == jointHistograms.end() )
  {
    return false;
  }
  
  auxPair.first = roiID;
  auxPair.second = bandIdx1;
  
  histogramsIt1 = histograms.find( auxPair );
  if( histogramsIt1 == histograms.end() )
  {
    return false;
  }
  
  auxPair.second = bandIdx2;
  
  histogramsIt2 = histograms.find( auxPair );
  if( histogramsIt2 == histograms.end() )
  {
    return false;
  }
  
  double covariance = 0;
  if( ! te::rp::getJointHistogramStats( jointHistogramsIt->second,
    histogramsIt1->second, histogramsIt2->second, covariance, correlation ) )
  {
    return false;
  }
  
  return true;
}

bool te::qt::widgets::RasterStatisticsWizardPage::printMatrix( 
  te::rp::Matrix< double >& matrix )
{
  const unsigned int linesNmb = matrix.getLinesNumber();
  const unsigned int colsNmb = matrix.getColumnsNumber();
  std::string lineStr;
  
  for( unsigned int line = 0 ; line < linesNmb ; ++line )
  {
    lineStr = "  ";
    
    for( unsigned int col = 0 ; col < colsNmb ; ++col )
    {
      lineStr.append( " " );
      lineStr.append(  std::to_string( matrix( line, col ) ) );
    }
    
    m_logDialogPtr->addTextLine( lineStr );
  }
  
  return true;
}

bool te::qt::widgets::RasterStatisticsWizardPage::getCorrelationMatrix(
  const std::string& roiID,
  const std::map< std::tuple< std::string, unsigned int, unsigned int >,
    std::map< std::pair< double, double >, unsigned int> >& jointHistograms,
  const std::map< std::pair< std::string, unsigned int >, std::map<double, 
    unsigned int> >& histograms,    
  te::rp::Matrix< double >& matrix )
{
  // bands
  
  std::vector< unsigned int > bandsIndexes;
  getBandIndexes( bandsIndexes );
    
  if( bandsIndexes.empty() )
  {
    QMessageBox::critical( this, tr("Warning"),
        tr("Invalid number of selected bands"));
    return false;
  }
  
  const std::size_t bandsIndexesSize = bandsIndexes.size()  ;
  
  if( !matrix.reset( bandsIndexes.size(), bandsIndexes.size() ) )
  {
    return false;
  }
  
  for( std::size_t bandsIndexesIdx1 = 0 ; bandsIndexesIdx1 < bandsIndexesSize ;
    ++bandsIndexesIdx1 )
  {
    const unsigned int bandIdx1 = bandsIndexes[ bandsIndexesIdx1 ];
    
    for( std::size_t bandsIndexesIdx2 = 0 ; bandsIndexesIdx2 < bandsIndexesSize ;
      ++bandsIndexesIdx2 )
    {
      const unsigned int bandIdx2 = bandsIndexes[ bandsIndexesIdx2 ];

      if( bandsIndexesIdx2 == bandsIndexesIdx1 )
      {
        matrix( bandsIndexesIdx2, bandsIndexesIdx1 ) = 1.0;
      }
      else if( bandsIndexesIdx2 > bandsIndexesIdx1 )
      {
        if( ! getCorrelation( roiID, bandIdx1, bandIdx2, jointHistograms,
          histograms, matrix( bandsIndexesIdx1, bandsIndexesIdx2 ) ) )
        {
          return false;
        }
        
        matrix( bandsIndexesIdx2, bandsIndexesIdx1 ) =
          matrix( bandsIndexesIdx1, bandsIndexesIdx2 );
      }
    }
  }
  
  return true;
}

bool te::qt::widgets::RasterStatisticsWizardPage::getKurtosis(
  const std::string& roiID,
  const unsigned int& bandIdx1,
  const std::map< std::tuple< std::string, unsigned int, unsigned int >,
    std::map< std::pair< double, double >, unsigned int> >& jointHistograms,
  const std::map< std::pair< std::string, unsigned int >, std::map<double, 
    unsigned int> >& histograms,
  double& kurtosis )
{
  kurtosis = 0;
  
  std::map< std::tuple< std::string, unsigned int, unsigned int >,
    std::map< std::pair< double, double >, unsigned int> >::const_iterator 
    jointHistogramsIt;
  std::map< std::pair< std::string, unsigned int >, std::map<double, 
    unsigned int> >::const_iterator histogramsIt1;
  std::tuple< std::string, unsigned int, unsigned int > auxTuple;
  std::pair< std::string, unsigned int > auxPair;
  
  std::get< 0 >( auxTuple ) = roiID;
  std::get< 1 >( auxTuple ) = bandIdx1;
  std::get< 2 >( auxTuple ) = bandIdx1;  
  jointHistogramsIt = jointHistograms.find( auxTuple );
  if( jointHistogramsIt == jointHistograms.end() )
  {
    return false;
  }
  const std::map< std::pair< double, double >, unsigned int>& jointHistogram =
    jointHistogramsIt->second;  
  
  auxPair.first = roiID;
  auxPair.second = bandIdx1;  
  histogramsIt1 = histograms.find( auxPair );
  if( histogramsIt1 == histograms.end() )
  {
    return false;
  }
  const std::map<double, unsigned int>& histogram = histogramsIt1->second;
  
  double covariance = 0;
  double corrlationCoef = 0;
  if( ! te::rp::getJointHistogramStats( jointHistogramsIt->second,
    histogramsIt1->second, histogramsIt1->second, covariance, corrlationCoef ) )
  {
    return false;
  }
  
  double min = 0;
  double max = 0;
  double mean = 0;
  double stdDev = 0;
  double mode = 0;
  double entropy = 0;
  double sum = 0;
  double sum2 = 0;
  double sum3 = 0;
  double sum4 = 0;
  double variance = 0;
  double median = 0;
  te::rp::getHistogramStats( 
    histogram, 
    min, max, mean, stdDev, mode, entropy, sum, sum2, 
    sum3, sum4, variance, median );

  std::map< std::pair< double, double >, unsigned int>::const_iterator 
    jointHistogramIt =  jointHistogram.begin();
  const std::map< std::pair< double, double >, unsigned int>::const_iterator 
    jointHistogramItEnd =  jointHistogram.end();  
  double sumPB1B2 = 0;
  double elementsNmb = 0;
  
  while( jointHistogramIt != jointHistogramItEnd )
  {
    sumPB1B2 += ( jointHistogramIt->first.first * jointHistogramIt->first.second 
      * (double)jointHistogramIt->second );
    elementsNmb += (double)jointHistogramIt->second;
    ++jointHistogramIt;
  }
  
  const double sigma = std::sqrt( covariance );
  const double sigma4 = sigma * sigma * sigma * sigma;

  if( ( elementsNmb != 0 ) && ( sigma4 != 0 ) )
  {
    sumPB1B2 /= elementsNmb;
    
    kurtosis = 
      ( sum4 / elementsNmb - 4 * mean * sum3 /
      elementsNmb  + 6 * std::pow( mean, 2 ) * sumPB1B2 - 3 * pow( mean, 4 ) 
      ) / sigma4 - 3.;
  }
  
  return true;
}

bool te::qt::widgets::RasterStatisticsWizardPage::getKurtosisMatrix(
  const std::string& roiID,
  const std::map< std::tuple< std::string, unsigned int, unsigned int >,
    std::map< std::pair< double, double >, unsigned int> >& jointHistograms,
  const std::map< std::pair< std::string, unsigned int >, std::map<double, 
    unsigned int> >& histograms,
  te::rp::Matrix< double >& matrix )
{
  // bands
  
  std::vector< unsigned int > bandsIndexes;
  getBandIndexes( bandsIndexes );
    
  if( bandsIndexes.empty() )
  {
    QMessageBox::critical( this, tr("Warning"),
        tr("Invalid number of selected bands"));
    return false;
  }
  
  const std::size_t bandsIndexesSize = bandsIndexes.size()  ;
  
  if( !matrix.reset( 1, bandsIndexes.size() ) )
  {
    return false;
  }
  
  for( std::size_t bandsIndexesIdx1 = 0 ; bandsIndexesIdx1 < bandsIndexesSize ;
    ++bandsIndexesIdx1 )
  {
    const unsigned int bandIdx1 = bandsIndexes[ bandsIndexesIdx1 ];

    if( ! getKurtosis( roiID, bandIdx1, jointHistograms,
      histograms, matrix( 0, bandsIndexesIdx1 ) ) )
    {
      return false;
    }
  }
  
  return true;
}

std::vector< std::string > te::qt::widgets::RasterStatisticsWizardPage::getROIsIDs() const
{
  std::vector< std::string > roisIDs;
  
  if( m_roiManagerPtr->getWidget()->getROISet() )
  {
    std::map<std::string, te::cl::ROI*> roiMap;
    roiMap = m_roiManagerPtr->getWidget()->getROISet()->getROISet();
    
    std::map<std::string, te::cl::ROI*>::const_iterator roiMapIt = roiMap.begin();
    std::map<std::string, te::cl::ROI*>::const_iterator roiMapItEnd = roiMap.end();
    while( roiMapIt != roiMapItEnd )
    {
      roisIDs.push_back( roiMapIt->first );
      ++roiMapIt;
    }
  }
  
  return roisIDs;
}

void te::qt::widgets::RasterStatisticsWizardPage::setActionGroup(QActionGroup* actionGroup)
{
  m_roiManagerPtr->setActionGroup(actionGroup);
}

void te::qt::widgets::RasterStatisticsWizardPage::onROIManagerClosed(
  te::cl::ROISet* rs)
{
  std::map<std::string, te::cl::ROI*> roiMap;
  if( rs )
  {
    roiMap = rs->getROISet();
  }
    
  m_uiStatPageForm->m_targetClassComboBox->clear();
  m_uiStatPageForm->m_targetClassComboBox->addItem( "" );
  
  if( !roiMap.empty() )
  {
    std::map<std::string, te::cl::ROI*>::const_iterator roiMapIt = roiMap.begin();
    const std::map<std::string, te::cl::ROI*>::const_iterator roiMapItEnd = 
      roiMap.end();
    while( roiMapIt != roiMapItEnd )
    {
      m_uiStatPageForm->m_targetClassComboBox->addItem( roiMapIt->first.c_str() );
      ++roiMapIt;
    }
  }
}