/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.
  
    This file is part of the TerraLib - a Framework for building GIS enabled applications.
    
    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.
    
    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.
    
    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */


#ifndef TE_QT_WIDGETS_RASTERSTATISTICSWIZARDPAGE_H
#define TE_QT_WIDGETS_RASTERSTATISTICSWIZARDPAGE_H

//Terralib
#ifndef Q_MOC_RUN
#include "../../../dataaccess/datasource/DataSourceInfo.h"
#include "../../../maptools/AbstractLayer.h"
#endif
#include "../Config.h"
#include "../classification/ROIManagerDialog.h"
#include "../../../rp/Matrix.h"
#include "RasterStatisticsWizardLog.h"

//Qt
#include <QWizardPage>
#include <QActionGroup>

//STL
#include <memory>

//Forward declarations
namespace Ui
{
  class RasterStatisticsWizardPageForm;
}

namespace te 
{
  namespace rst { class Raster; }

  namespace qt 
  {
    namespace widgets 
    {
      
      class TEQTWIDGETSEXPORT RasterStatisticsWizardPage : public QWizardPage
      {
        Q_OBJECT

        public:
          
          RasterStatisticsWizardPage(QWidget* parent=0);

          ~RasterStatisticsWizardPage();

          bool isComplete() const;

          /*!
            \brief This method is used to set the selected layer for vectorization operation
            
            \param layer The layer ptr 

            \note This layer MUST HAVE a valid raster object.
          */
          void setLayer(te::map::AbstractLayerPtr layer);
          
          void setMapDisplay(te::qt::widgets::MapDisplay* mapDisplay);
          
          void setActionGroup(QActionGroup* actionGroup);

        protected:

          std::unique_ptr<Ui::RasterStatisticsWizardPageForm> m_uiStatPageForm;
          std::unique_ptr<te::qt::widgets::ROIManagerDialog> m_roiManagerPtr;
          std::unique_ptr<te::qt::widgets::RasterStatisticsWizardLog> m_logDialogPtr;
          te::map::AbstractLayerPtr m_layer;
          
          /*!
            \brief Return the current selected band indexes.
            \param bandsIndexes Current selected band indexes.
          */
          void getBandIndexes( std::vector< unsigned int >& bandsIndexes ) const;
          
          /*!
            \brief Return the current histograms.
            \param realHistograms Real histograms.
            \param imagHistograms Real histograms.
            \note Histograms map: map key.first=ROI Ids, map_key.second=band_index.
            \note Histograms map: map key.first="ALL_IMAGE" for the case where no ROIs are avaliable.
            \return true if ok, false on erros.
          */
          bool getHistograms(
            std::map< std::pair< std::string, unsigned int >, std::map<double, 
              unsigned int> >& realHistograms,
            std::map< std::pair< std::string, unsigned int >, std::map<double, 
              unsigned int> >& imagHistograms ) const;
            
          /*!
            \brief Return the current joint histograms.
            \param realHistograms Real histograms.
            \param imagHistograms Real histograms.
            \note Joint Histograms map key: std::tuple( 0 )=roi_id, std::tuple( 1 )=band_index_1, std::tuple( 2 )=band_index_2
            \note Joint Histograms map key: std::tuple( 0 )="ALL_IMAGE" for the case where no ROIs are avaliable.
            \return true if ok, false on erros.
          */
          bool getJoinHistograms(
            const std::map< std::pair< std::string, unsigned int >, std::map<double, 
              unsigned int> >& realHistograms,
            const std::map< std::pair< std::string, unsigned int >, std::map<double, 
              unsigned int> >& imagHistograms,
            std::map< std::tuple< std::string, unsigned int, unsigned int >,
              std::map< std::pair< double, double >, unsigned int> >& realJointHistograms,
            std::map< std::tuple< std::string, unsigned int, unsigned int >,
               std::map< std::pair< double, double >, unsigned int> >& imagJointHistograms ) const;
               
          /*!
            \brief Return the covariance value.
            \param roiID Target ROI id.
            \param bandIdx1 Target band 1 index.
            \param bandIdx2 Target band 2 index.
            \param jointHistograms Joint histograms.
            \param histograms Single band histograms.
            \param covariance Covariance value.
            \return true if ok, false on erros.
          */
          static bool getCovariance(
            const std::string& roiID,
            const unsigned int& bandIdx1,
            const unsigned int& bandIdx2,            
            const std::map< std::tuple< std::string, unsigned int, unsigned int >,
              std::map< std::pair< double, double >, unsigned int> >& jointHistograms,
            const std::map< std::pair< std::string, unsigned int >, std::map<double, 
              unsigned int> >& histograms,
            double& covariance );
          
          /*!
            \brief Return the correlation value.
            \param roiID Target ROI id.
            \param bandIdx1 Target band 1 index.
            \param bandIdx2 Target band 2 index.
            \param jointHistograms Joint histograms.
            \param histograms Single band histograms.
            \param correlation Correlation value.
            \return true if ok, false on erros.
          */
          static bool getCorrelation(
            const std::string& roiID,
            const unsigned int& bandIdx1,
            const unsigned int& bandIdx2,            
            const std::map< std::tuple< std::string, unsigned int, unsigned int >,
              std::map< std::pair< double, double >, unsigned int> >& jointHistograms,
            const std::map< std::pair< std::string, unsigned int >, std::map<double, 
              unsigned int> >& histograms,
            double& correlation );          
          
          /*!
            \brief Return the Kurtosis value.
            \param roiID Target ROI id.
            \param bandIdx1 Target band 1 index.
            \param jointHistograms Joint histograms.
            \param histograms Single band histograms.
            \param kurtosis Correlation value.
            \return true if ok, false on erros.
          */
          static bool getKurtosis(
            const std::string& roiID,
            const unsigned int& bandIdx1,
            const std::map< std::tuple< std::string, unsigned int, unsigned int >,
              std::map< std::pair< double, double >, unsigned int> >& jointHistograms,
            const std::map< std::pair< std::string, unsigned int >, std::map<double, 
              unsigned int> >& histograms,
            double& kurtosis );                    
               
          /*!
            \brief Return the covariance matrix.
            \param roiID Target ROI id.
            \param jointHistograms Joint histograms.
            \param histograms Single band histograms.
            \param matrix Output matrix.
            \return true if ok, false on erros.
          */
          bool getCovarianceMatrix(
            const std::string& roiID,
            const std::map< std::tuple< std::string, unsigned int, unsigned int >,
              std::map< std::pair< double, double >, unsigned int> >& jointHistograms,
            const std::map< std::pair< std::string, unsigned int >, std::map<double, 
              unsigned int> >& histograms,
            te::rp::Matrix< double >& matrix );
          
          /*!
            \brief Return the correlation matrix.
            \param roiID Target ROI id.
            \param jointHistograms Joint histograms.
            \param histograms Single band histograms.
            \param matrix Output matrix.
            \return true if ok, false on erros.
          */
          bool getCorrelationMatrix(
            const std::string& roiID,
            const std::map< std::tuple< std::string, unsigned int, unsigned int >,
              std::map< std::pair< double, double >, unsigned int> >& jointHistograms,
            const std::map< std::pair< std::string, unsigned int >, std::map<double, 
              unsigned int> >& histograms,
            te::rp::Matrix< double >& matrix );
          
          /*!
            \brief Return the kurtosis matrix.
            \param roiID Target ROI id.
            \param jointHistograms Joint histograms.
            \param histograms Single band histograms.
            \param matrix Output matrix.
            \return true if ok, false on erros.
          */
          bool getKurtosisMatrix(
            const std::string& roiID,
            const std::map< std::tuple< std::string, unsigned int, unsigned int >,
              std::map< std::pair< double, double >, unsigned int> >& jointHistograms,
            const std::map< std::pair< std::string, unsigned int >, std::map<double, 
              unsigned int> >& histograms,
            te::rp::Matrix< double >& matrix );          
          
          bool printMatrix( te::rp::Matrix< double >& matrix );
          
          /*!
            \brief Return the current loaded ROIs IDs.
            \return Return the current loaded ROIs IDs.
          */
          std::vector< std::string > getROIsIDs() const;
          
        protected slots:
          
          void histogramPushButtonClicked(bool checked);
          
          void scatterPushButtonClicked(bool checked);
          
          void statisticsPushButtonClicked(bool checked);
          
          void samplesPushButtonClicked(bool checked);
          
          void coStatisticsPushButtonClicked(bool checked);
          
          void onROIManagerClosed(te::cl::ROISet* rs);
      };
      
    } // namespace widgets
  } // namespace qt
} // namespace te

#endif // TE_QT_WIDGETS_RASTERSTATISTICSWIZARDPAGE_H
