/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

#include "RasterTransformWizard.h"

#include "RasterTransformWizardPage.h"
#include "RasterInfoWizardPage.h"
#include "../layer/search/LayerSearchWizardPage.h"
#include "../../../dataaccess/utils/Utils.h"
#include "../../../raster/Grid.h"
#include "../../../raster/Band.h"
#include "../../../raster/RasterFactory.h"
#include "../../../raster/Reprojection.h"
#include "../../../raster/BandProperty.h"
#include "../../../rp/Functions.h"
#include "../help/HelpPushButton.h"
#include "../layer/search/LayerSearchWidget.h"
#include "../progress/ProgressViewerDialog.h"
#include "../../../common/progress/TaskProgress.h"
#include "../utils/ScopedCursor.h"
#include "../raster/RasterInfoWidget.h"
#include "Utils.h"
#include "../../../maptools/Utils.h"

#include <QApplication>
#include <QMessageBox>
#include <QActionGroup>

te::qt::widgets::RasterTransformWizard::RasterTransformWizard(QWidget* parent )
  : QWizard(parent)
{
  //configure the wizard
  this->setWizardStyle(QWizard::ModernStyle);
  this->setWindowTitle(tr("Raster Transform"));
  //this->setFixedSize(640, 580);

  this->setOption(QWizard::HaveHelpButton, true);
  this->setOption(QWizard::HelpButtonOnRight, false);

  te::qt::widgets::HelpPushButton* helpButton = 
    new te::qt::widgets::HelpPushButton(this);

  this->setButton(QWizard::HelpButton, helpButton);

  helpButton->setPageReference("plugins/rp/rp_rastertransform.html");

  connect(this, SIGNAL(currentIdChanged(int)), SLOT(onPageChanged(int)));

  addPages();
}

te::qt::widgets::RasterTransformWizard::~RasterTransformWizard()
{
}

bool te::qt::widgets::RasterTransformWizard::validateCurrentPage()
{
  if(currentPage() ==  m_layerSearchPage.get())
  {
    std::list<te::map::AbstractLayerPtr> list = m_layerSearchPage->getSearchWidget()->getSelecteds();

    if(list.empty() == false)
    {
      te::map::AbstractLayerPtr l = *list.begin();

      m_mainWizardPagePage->setLayer(l);
    }

    return m_layerSearchPage->isComplete();
  }
  else if(currentPage() ==  m_mainWizardPagePage.get())
  {
    bool res = m_mainWizardPagePage->isComplete();

    if(!res)
      QMessageBox::critical(this, tr("Warning"), tr("Incomplete parameters"));

    return res;
  }
  else if(currentPage() ==  m_rasterInfoPage.get())
  {
    return execute();
  }

  return true;
}

void te::qt::widgets::RasterTransformWizard::setList(std::list<te::map::AbstractLayerPtr>& layerList)
{
  m_layerSearchPage->getSearchWidget()->setList(layerList);
  m_layerSearchPage->getSearchWidget()->filterOnlyByRaster();
}

void te::qt::widgets::RasterTransformWizard::setLayer(te::map::AbstractLayerPtr layer)
{
  m_mainWizardPagePage->setLayer(layer);
}

te::map::AbstractLayerPtr te::qt::widgets::RasterTransformWizard::getOutputLayer()
{
  return m_outputLayer;
}

void te::qt::widgets::RasterTransformWizard::addPages()
{
  m_layerSearchPage.reset(new te::qt::widgets::LayerSearchWizardPage(this));
  m_mainWizardPagePage.reset(new te::qt::widgets::RasterTransformWizardPage(
    this ) );
  m_rasterInfoPage.reset(new te::qt::widgets::RasterInfoWizardPage(this));

  addPage(m_layerSearchPage.get());
  addPage(m_mainWizardPagePage.get());
  addPage(m_rasterInfoPage.get());

  m_layerSearchPage->getSearchWidget()->enableMultiSelection(false);
}

bool te::qt::widgets::RasterTransformWizard::execute()
{
  if( m_mainWizardPagePage->getLayerPtr() == 0 ) return false;
  
  std::vector< unsigned int > selectedInBandsIndexes;
  m_mainWizardPagePage->getSelectedInBandIndexes( selectedInBandsIndexes );
  if( selectedInBandsIndexes.empty() )
  {
    QMessageBox::critical( this, tr("Error"), tr("No bands selected"));
    return false;
  }
  
  // progress
  
//  te::qt::widgets::ProgressViewerDialog v(this);
  std::unique_ptr< te::common::TaskProgress > taskProgressPtr( 
    new te::common::TaskProgress( "Transforming raster", 
    te::common::TaskProgress::UNDEFINED, 6 ) );
  
  // load original data
  
  taskProgressPtr->pulse();
  if( ! taskProgressPtr->isActive() )  
  {
    return false;
  }  
  
  std::unique_ptr<te::rst::Raster> outputMEMRasterPtr;
  
  if( ! executeLoad( outputMEMRasterPtr ) )
  {
    return false;
  }
  
  // data type change
  
  if( 
      m_mainWizardPagePage->m_uiPageForm->m_dataTypeConvEnabledCheckBox->checkState() 
      ==
      Qt::Checked
    )
  {
    if( ! executeDataTypeChange( outputMEMRasterPtr ) )
    {
      return false;
    }
  }
  
  taskProgressPtr->pulse();
  if( ! taskProgressPtr->isActive() )  
  {
    return false;
  }
  
  // no-data value change
  
  if( 
      m_mainWizardPagePage->m_uiPageForm->m_nodataRelaceCheckBox->checkState() 
      ==
      Qt::Checked
    )
  {
    if( ! executeNoDataValueChange( outputMEMRasterPtr ) )
    {
      return false;
    }
  }  
  
  taskProgressPtr->pulse();
  if( ! taskProgressPtr->isActive() )  
  {
    return false;
  }
  
  // resample
  
  if( 
      m_mainWizardPagePage->m_uiPageForm->m_ressampleCheckBox->checkState() 
      ==
      Qt::Checked
    )
  {
    if( ! executeResample( outputMEMRasterPtr ) )
    {
      return false;
    }
  }
  
  taskProgressPtr->pulse();
  if( ! taskProgressPtr->isActive() )  
  {
    return false;
  }
  
  // reprojection
  
  if( 
      m_mainWizardPagePage->m_uiPageForm->m_enableSRSChangeCheckBox->checkState() 
      ==
      Qt::Checked
    )
  {
    if( ! executeReprojection( outputMEMRasterPtr ) )
    {
      return false;
    }
  }      
  
  taskProgressPtr->pulse();
  if( ! taskProgressPtr->isActive() )  
  {
    return false;
  }
  
  // Save the output raster
  
  {
    std::vector< te::rst::BandProperty * > bandProps;
    for( unsigned int bandIdx = 0 ; bandIdx < outputMEMRasterPtr->getNumberOfBands() ;
      ++bandIdx )
    {
      bandProps.push_back( new te::rst::BandProperty(
        *( outputMEMRasterPtr->getBand( bandIdx )->getProperty() ) ) );
    }

    const std::string rasterType = m_rasterInfoPage->getWidget()->getType();

    std::map<std::string, std::string> rasterInfo =
      m_rasterInfoPage->getWidget()->getInfo();

    te::rst::Grid* gridPtr = new te::rst::Grid( *outputMEMRasterPtr->getGrid() );

    std::unique_ptr< te::rst::Raster > outputRasterPtr;

    try
    {
      outputRasterPtr.reset(te::rst::RasterFactory::make(rasterType, gridPtr, bandProps,
        rasterInfo));
    }
    catch(...)
    {
      outputRasterPtr.reset();
    }
    if( outputRasterPtr.get() == 0 )
    {
      QMessageBox::critical(this, tr("Raster Transform"), tr("Output raster creation error"));
      return false;
    }  
    
    const unsigned int nRows = outputRasterPtr->getNumberOfRows();
    const unsigned int nCols = outputRasterPtr->getNumberOfColumns();
    const unsigned int nOutBands = outputRasterPtr->getNumberOfBands();
    unsigned int outBandIdx = 0;
    unsigned int col = 0;
    unsigned int row = 0;
    std::complex< double > value;
    
    for( outBandIdx = 0 ; outBandIdx < nOutBands ; ++outBandIdx )
    {
      const te::rst::Band& inBand = *outputMEMRasterPtr->getBand( outBandIdx );
      te::rst::Band& outBand = *outputRasterPtr->getBand( outBandIdx );
      
      for( row = 0 ; row < nRows ; ++row )
      {
        for( col = 0 ; col < nCols ; ++col )
        {
          inBand.getValue( col, row, value );
          outBand.setValue( col, row, value );
        }
      }
    } 
  }
  
  taskProgressPtr->pulse();
  if( ! taskProgressPtr->isActive() )  
  {
    return false;
  }
  
  taskProgressPtr.reset();
  
  // create output layer
  
  m_outputLayer = te::qt::widgets::createLayer(
    m_rasterInfoPage->getWidget()->getType(),
    m_rasterInfoPage->getWidget()->getInfo());  
  
  emit addLayer(m_outputLayer);

  QMessageBox::information(this, tr("Raster Transform"), tr("Raster Transform ended sucessfully."));
  
  te::qt::widgets::applyRasterMultiResolution(tr("Raster Transform"), te::map::GetRaster(m_outputLayer.get()));

  return true;
}

void te::qt::widgets::RasterTransformWizard::onPageChanged(int /*index*/)
{
  this->adjustSize();
}

bool te::qt::widgets::RasterTransformWizard::executeLoad(
  std::unique_ptr< te::rst::Raster >& rasterPtr )
{
  if( m_mainWizardPagePage->getLayerPtr() == 0 )
  {
    QMessageBox::critical( this, tr("Error"), tr("Invalid input layer"));
    return false;
  }
  
  // selected bands
  
  std::vector< unsigned int > selectedInBandsIndexes;
  m_mainWizardPagePage->getSelectedInBandIndexes( selectedInBandsIndexes );
  if( selectedInBandsIndexes.empty() )
  {
    QMessageBox::critical( this, tr("Error"), tr("No bands selected"));
    return false;
  }  
  
  // get input raster
  
  std::unique_ptr<te::rst::Raster> inputRasterPtr;
  std::unique_ptr<te::da::DataSet> ds = m_mainWizardPagePage->getLayerPtr()->getData();
  
  {
    if(ds.get())
    {
      std::size_t rpos = te::da::GetFirstPropertyPos(ds.get(), 
        te::dt::RASTER_TYPE);

      inputRasterPtr = ds->getRaster(rpos);

      if(inputRasterPtr.get() == nullptr)
      {
        QMessageBox::critical( this, tr("Error"), tr("Invalid input raster"));
        return false;
      }
    }
  }  
  
  // creating the memory raster

  try
  {
    std::vector< te::rst::BandProperty * > bandProps;
    for( unsigned int selectedInBandsIndexesIdx = 0 ; selectedInBandsIndexesIdx < 
      selectedInBandsIndexes.size() ; ++selectedInBandsIndexesIdx )
    {
      bandProps.push_back( new te::rst::BandProperty(
        *( inputRasterPtr->getBand( selectedInBandsIndexes[ 
        selectedInBandsIndexesIdx ] )->getProperty() ) ) );
    }

    std::map<std::string, std::string> rasterInfo;
    rasterInfo[ "MAXMEMPERCENTUSED "] = 50;

    te::rst::Grid* gridPtr = new te::rst::Grid(
      *inputRasterPtr->getGrid() );

    te::qt::widgets::ScopedCursor cursor(Qt::WaitCursor);

    rasterPtr.reset(te::rst::RasterFactory::make( "EXPANSIBLE",
      gridPtr, bandProps, rasterInfo));
  }
  catch(const std::exception& e)
  {
    rasterPtr.reset();
    QMessageBox::critical(this, tr("Output raster creation error"), e.what());
    return false;
  }
  catch(...)
  {
    rasterPtr.reset();
    QMessageBox::critical(this, tr("Raster Transform"), tr("Output raster creation error"));
    return false;
  }  
  
  // data copy
  {
    const unsigned int nRows = rasterPtr->getNumberOfRows();
    const unsigned int nCols = rasterPtr->getNumberOfColumns();
    const unsigned int nOutBands = rasterPtr->getNumberOfBands();
    unsigned int outBandIdx = 0;
    unsigned int col = 0;
    unsigned int row = 0;
    std::complex< double > value;
    te::qt::widgets::ScopedCursor cursor(Qt::WaitCursor);
    
    for( outBandIdx = 0 ; outBandIdx < nOutBands ; ++outBandIdx )
    {
      const te::rst::Band& inBand = *inputRasterPtr->getBand( 
        selectedInBandsIndexes[ outBandIdx ] );
      te::rst::Band& outBand = *rasterPtr->getBand( outBandIdx );
      
      for( row = 0 ; row < nRows ; ++row )
      {
        for( col = 0 ; col < nCols ; ++col )
        {
          inBand.getValue( col, row, value );
          outBand.setValue( col, row, value );
        }
      }
    } 
  } 
  
  return true;
}

bool te::qt::widgets::RasterTransformWizard::executeDataTypeChange( 
  std::unique_ptr< te::rst::Raster >& rasterPtr )
{
  // gains and offs
  
  std::vector< int > inDataTypes;
  std::vector< double > inRangesMin;
  std::vector< double > inRangesMax;
  for( unsigned int inBandIdx = 0 ; inBandIdx < rasterPtr->getNumberOfBands() ;
    ++inBandIdx )
  {
    inDataTypes.push_back( rasterPtr->getBand( inBandIdx )->getProperty()->m_type );
    
    double min = 0;
    double max = 0;
    te::rp::GetDataTypeRange( inDataTypes[ inBandIdx ], min, max );
    inRangesMin.push_back( min );
    inRangesMax.push_back( max );
  }  
  
  const int outDataType = (int)m_mainWizardPagePage->getOutDataType();    
  double outRangeMin = 0;
  double outRangeMax = 0;
  te::rp::GetDataTypeRange( outDataType, outRangeMin, outRangeMax );  
  
  // out = ( ( in - in_off ) * gain ) + out_off
  std::vector< double > gains; 
  std::vector< double > inOff;
  std::vector< double > outOff;
  for( std::size_t inRangesMinIdx = 0 ; inRangesMinIdx < inRangesMin.size() ;
    ++inRangesMinIdx )
  {
    if( 
        (
          m_mainWizardPagePage->m_uiPageForm->m_normalizeOutputCheckBox->checkState() 
          !=
          Qt::Checked
        )
        ||
        ( inDataTypes[ inRangesMinIdx] == outDataType )
        ||
        ( inDataTypes[ inRangesMinIdx] == te::dt::DOUBLE_TYPE )
        ||
        ( outDataType == te::dt::DOUBLE_TYPE )   
        ||
        ( inDataTypes[ inRangesMinIdx] == te::dt::CDOUBLE_TYPE )
        ||
        ( outDataType == te::dt::CDOUBLE_TYPE )          
      )
    {
      gains.push_back( 1 );
      inOff.push_back( 0 );
      outOff.push_back( 0 );
    }
    else
    {
      double inRange = inRangesMax[ inRangesMinIdx ] - inRangesMin[ inRangesMinIdx ];
      double outRange = outRangeMax - outRangeMin;
      
      inOff.push_back( inRangesMin[ inRangesMinIdx ] );
      outOff.push_back( outRangeMin );       
      
      if( inRange == 0 )
      {
        gains.push_back( 0 );       
      }     
      else
      {
        gains.push_back( outRange / inRange );  
      }
    }
  }
  
  // creating the memory raster
  
  std::unique_ptr< te::rst::Raster > auxRasterPtr;

  try
  {
    std::vector< te::rst::BandProperty * > bandProps;
    for( unsigned int inRasterBandIdx = 0 ; inRasterBandIdx < 
      rasterPtr->getNumberOfBands() ; ++inRasterBandIdx )
    {
      bandProps.push_back( new te::rst::BandProperty(
        *( rasterPtr->getBand( inRasterBandIdx )->getProperty() ) ) );
      bandProps[ inRasterBandIdx ]->m_type = outDataType;
    }

    std::map<std::string, std::string> rasterInfo;
    rasterInfo[ "MAXMEMPERCENTUSED "] = 50;

    te::rst::Grid* gridPtr = new te::rst::Grid(
      *rasterPtr->getGrid() );

    te::qt::widgets::ScopedCursor cursor(Qt::WaitCursor);

    auxRasterPtr.reset(te::rst::RasterFactory::make( "EXPANSIBLE",
      gridPtr, bandProps, rasterInfo));
  }
  catch(const std::exception& e)
  {
    QMessageBox::critical(this, tr("Output raster creation error"), e.what());
    return false;
  }
  catch(...)
  {
    QMessageBox::critical(this, tr("Raster Transform"), tr("Output raster creation error"));
    return false;
  }    
  
  // Data copy
  
  {
    const unsigned int nRows = rasterPtr->getNumberOfRows();
    const unsigned int nCols = rasterPtr->getNumberOfColumns();
    const unsigned int nBands = rasterPtr->getNumberOfBands();
    unsigned int bandIdx = 0;
    unsigned int col = 0;
    unsigned int row = 0;
    std::complex< double > value;
    te::qt::widgets::ScopedCursor cursor(Qt::WaitCursor);
    
    for( bandIdx = 0 ; bandIdx < nBands ; ++bandIdx )
    {
      const te::rst::Band& inBand = *rasterPtr->getBand( bandIdx );
      te::rst::Band& outBand = *auxRasterPtr->getBand( bandIdx );
      
      for( row = 0 ; row < nRows ; ++row )
      {
        for( col = 0 ; col < nCols ; ++col )
        {
          inBand.getValue( col, row, value );
          
          value.real( 
            (
              ( value.real() - inOff[ bandIdx ] )
              * 
              gains[ bandIdx ] 
            )
            + 
            outOff[ bandIdx ]
          );
          value.real( std::min( outRangeMax, value.real() ) );
          value.real( std::max( outRangeMin, value.real() ) );          
          
          value.imag( 
            (
              ( value.imag() - inOff[ bandIdx ] )
              * 
              gains[ bandIdx ] 
            )
            + 
            outOff[ bandIdx ]
          );
          value.imag( std::min( outRangeMax, value.imag() ) );
          value.imag( std::max( outRangeMin, value.imag() ) );          
          
          outBand.setValue( col, row, value );
        }
      }
    }
  }
  
  rasterPtr.reset( auxRasterPtr.release() );
  
  return true;
}

bool te::qt::widgets::RasterTransformWizard::executeNoDataValueChange( 
  std::unique_ptr< te::rst::Raster >& rasterPtr )
{
  if( m_mainWizardPagePage->m_uiPageForm->m_newNoDataValueLineEdit->text().isEmpty() )
  {
    QMessageBox::critical(this, tr("Error"), tr("Invalid mew no-data value") );
    return false;
  }
  
  const double newNoDataValue = 
    m_mainWizardPagePage->m_uiPageForm->m_newNoDataValueLineEdit->text().toDouble();
  
  if( 
      m_mainWizardPagePage->m_uiPageForm->m_updatePixValuesCheckBox->checkState() 
      !=
      Qt::Checked
    )
  {  // changing onlye the property data
    for( unsigned int inBandIdx = 0 ; inBandIdx < rasterPtr->getNumberOfBands() ;
      ++inBandIdx )
    {
      rasterPtr->getBand( inBandIdx )->getProperty()->m_noDataValue =
        newNoDataValue;
    }  
  }
  else
  {
    // creating the memory raster
    
    std::unique_ptr< te::rst::Raster > auxRasterPtr;

    try
    {
      std::vector< te::rst::BandProperty * > bandProps;
      for( unsigned int inRasterBandIdx = 0 ; inRasterBandIdx < 
        rasterPtr->getNumberOfBands() ; ++inRasterBandIdx )
      {
        bandProps.push_back( new te::rst::BandProperty(
          *( rasterPtr->getBand( inRasterBandIdx )->getProperty() ) ) );
        bandProps[ inRasterBandIdx ]->m_noDataValue = newNoDataValue;
      }

      std::map<std::string, std::string> rasterInfo;
      rasterInfo[ "MAXMEMPERCENTUSED "] = 50;

      te::rst::Grid* gridPtr = new te::rst::Grid(
        *rasterPtr->getGrid() );

      te::qt::widgets::ScopedCursor cursor(Qt::WaitCursor);

      auxRasterPtr.reset(te::rst::RasterFactory::make( "EXPANSIBLE",
        gridPtr, bandProps, rasterInfo));
    }
    catch(const std::exception& e)
    {
      QMessageBox::critical(this, tr("Output raster creation error"), e.what());
      return false;
    }
    catch(...)
    {
      QMessageBox::critical(this, tr("Raster Transform"), tr("Output raster creation error"));
      return false;
    }    
    
    // Data copy
    
    {
      std::vector< double > oldNoDataValues;
      if( m_mainWizardPagePage->m_uiPageForm->m_oldNoDataValueLineEdit->text().isEmpty() )
      {
        for( unsigned int bandIdx = 0 ; bandIdx < rasterPtr->getNumberOfBands() ; 
          ++bandIdx )
        {
          oldNoDataValues.push_back( rasterPtr->getBand( 
            bandIdx )->getProperty()->m_noDataValue );
        }
      }
      else
      {
        const double oldNoDataValue = 
          m_mainWizardPagePage->m_uiPageForm->m_oldNoDataValueLineEdit->text().toDouble();
        oldNoDataValues.resize( rasterPtr->getNumberOfBands(), oldNoDataValue );
      }
      
      const unsigned int nRows = rasterPtr->getNumberOfRows();
      const unsigned int nCols = rasterPtr->getNumberOfColumns();
      const unsigned int nBands = rasterPtr->getNumberOfBands();
      unsigned int bandIdx = 0;
      unsigned int col = 0;
      unsigned int row = 0;
      std::complex< double > value;
      std::complex< double > newNoDataCValue( newNoDataValue, newNoDataValue );
      te::qt::widgets::ScopedCursor cursor(Qt::WaitCursor);
      
      for( bandIdx = 0 ; bandIdx < nBands ; ++bandIdx )
      {
        const te::rst::Band& inBand = *rasterPtr->getBand( bandIdx );
        te::rst::Band& outBand = *auxRasterPtr->getBand( bandIdx );
        
        for( row = 0 ; row < nRows ; ++row )
        {
          for( col = 0 ; col < nCols ; ++col )
          {
            inBand.getValue( col, row, value );
            
            if( value.real() == oldNoDataValues[ bandIdx ] )
            {
              outBand.setValue( col, row, newNoDataCValue );
            }
            else
            {
              outBand.setValue( col, row, value );
            }
          }
        }
      }
    }
    
    rasterPtr.reset( auxRasterPtr.release() );
  }
  
  return true;
}

bool te::qt::widgets::RasterTransformWizard::executeResample( 
  std::unique_ptr< te::rst::Raster >& rasterPtr )
{
  unsigned int outputNRows = 0;
  unsigned int outputNCols = 0;
  
  if( m_mainWizardPagePage->m_uiPageForm->m_ressampleByScaleRadioButton->isChecked() )
  {
    const double xScale = 
      m_mainWizardPagePage->m_uiPageForm->m_ressampleXScaleLinesEdit->text().toDouble();
    const double yScale = 
      m_mainWizardPagePage->m_uiPageForm->m_ressampleYScaleLinesEdit->text().toDouble();         
      
    outputNRows = (unsigned int)( ((double)rasterPtr->getNumberOfRows()) *
      yScale );
    outputNCols = (unsigned int)( ((double)rasterPtr->getNumberOfColumns()) *
      xScale );      
    
    if( 
        ( outputNRows == 0 )
        ||
        ( outputNCols == 0 )
      )
    {
      QMessageBox::critical(this, tr("Raster Transform"), tr("Invalid scale values"));
      return false;
    }
  }  
  
  if( m_mainWizardPagePage->m_uiPageForm->m_ressampleByLinesColsRadioButton->isChecked() )
  {
    outputNRows = m_mainWizardPagePage->m_uiPageForm->m_ressampleLinesEdit->text().toUInt();
    outputNCols = m_mainWizardPagePage->m_uiPageForm->m_ressampleColsLineEdit->text().toUInt();    
    
    if( 
        ( outputNRows == 0 )
        ||
        ( outputNCols == 0 )
      )
    {
      QMessageBox::critical(this, tr("Raster Transform"), tr("Invalid scale values"));
      return false;
    }
  }  
  
  if( m_mainWizardPagePage->m_uiPageForm->m_ressampleByResolutionRadioButton->isChecked() )
  {
    const double xRes = 
      m_mainWizardPagePage->m_uiPageForm->m_ressampleResolutionLXLineEdit->text().toDouble();
    const double yRes = 
      m_mainWizardPagePage->m_uiPageForm->m_ressampleResolutionLYLineEdit->text().toDouble();        
    
    if( 
        ( xRes <= 0 )
        ||
        ( yRes <= 0 )
      )
    {
      QMessageBox::critical(this, tr("Raster Transform"), tr("Invalid resolution values"));
      return false;
    }
    
    outputNRows = (unsigned int)( rasterPtr->getGrid()->getExtent()->getHeight() / yRes );
    outputNCols = (unsigned int)( rasterPtr->getGrid()->getExtent()->getWidth() / xRes );
    
    if( 
        ( outputNRows == 0 )
        ||
        ( outputNCols == 0 )
      )
    {
      QMessageBox::critical(this, tr("Raster Transform"), tr("Invalid resolution values"));
      return false;
    }    
  }
  
  // resampling
  
  std::unique_ptr< te::rst::Raster > auxRasterPtr;
  
  {
    std::vector< unsigned int > inputRasterBands;
    for( unsigned int bandIdx = 0 ; bandIdx < rasterPtr->getNumberOfBands() ; 
      ++bandIdx )
    {
      inputRasterBands.push_back( bandIdx );
    }
    
    const te::rst::Interpolator::Method interpMethod = 
      (te::rst::Interpolator::Method)m_mainWizardPagePage->getInterpolator();
    const unsigned int interpWindowRadius = 
      m_mainWizardPagePage->m_uiPageForm->m_interpWRadiusLineEdit->text().toUInt();
    
    std::map<std::string, std::string> rasterInfo;
    rasterInfo[ "MAXMEMPERCENTUSED "] = 50;  

    bool res = false;
    {
      te::qt::widgets::ScopedCursor cursor(Qt::WaitCursor);
    
      res = te::rp::RasterResample(
        *rasterPtr,
        inputRasterBands,
        interpMethod,
        interpWindowRadius,
        0, // first row
        0, // first col
        rasterPtr->getNumberOfRows(),
        rasterPtr->getNumberOfColumns(),
        outputNRows,
        outputNCols,
        rasterInfo,
        "EXPANSIBLE",
        auxRasterPtr );
    }

    if( !res )
    {
      QMessageBox::critical(this, tr("Raster Transform"), tr("Raster resample error"));
      return false;
    }
  }
  
  rasterPtr.reset( auxRasterPtr.release() );
  
  return true;
}

bool te::qt::widgets::RasterTransformWizard::executeReprojection( 
  std::unique_ptr< te::rst::Raster >& rasterPtr )
{
  if( m_mainWizardPagePage->m_uiPageForm->m_newSRSLineEdit->text().isEmpty() )
  {
    QMessageBox::critical(this, tr("Raster Transform"), tr("Invalid new SRID"));
    return false;
  }
  const int newSRID = m_mainWizardPagePage->m_uiPageForm->m_newSRSLineEdit->text().toInt();
  
  if( 
      m_mainWizardPagePage->m_uiPageForm->m_enableRasterReprojectionCheckBox->checkState() 
      !=
      Qt::Checked
    )
  {
    rasterPtr->getGrid()->setSRID( newSRID );
  }
  else
  {
    if( newSRID <= 0 ) 
    {
      QMessageBox::critical(this, tr("Raster Transform"), tr("Invalid new SRID"));
      return false;
    }    
    
    int oldSRID = m_mainWizardPagePage->m_uiPageForm->m_oldSRSLineEdit->text().toInt();
    if( oldSRID == 0 )
    {
      oldSRID = rasterPtr->getGrid()->getSRID();
    }
    if( oldSRID <= 0 ) 
    {
      QMessageBox::critical(this, tr("Raster Transform"), tr("Invalid old SRID"));
      return false;
    }
    
    rasterPtr->getGrid()->setSRID( oldSRID );

    std::map<std::string, std::string> rasterInfo;
    rasterInfo[ "RTYPE"] = "EXPANSIBLE";
    
    std::unique_ptr< te::rst::Raster > auxRasterPtr;
    try
    {
      te::qt::widgets::ScopedCursor cursor(Qt::WaitCursor);

      auxRasterPtr.reset( te::rst::Reproject( 
        rasterPtr.get(),
        newSRID,
        rasterPtr->getGrid()->getExtent()->getLowerLeftX(),
        rasterPtr->getGrid()->getExtent()->getLowerLeftY(),
        rasterPtr->getGrid()->getExtent()->getUpperRightX(),
        rasterPtr->getGrid()->getExtent()->getUpperRightY(),
        0, // output rex x = auto
        0, // output rex y = auto
        rasterInfo, 
        m_mainWizardPagePage->getInterpolator() ) );
    }
    catch(...)
    {
      auxRasterPtr.reset();
    }
    
    if( auxRasterPtr.get() == nullptr )
    {
      QMessageBox::critical(this, tr("Raster Transform"), tr("Reprojection error"));
      return false;
    }
    
    rasterPtr.reset( auxRasterPtr.release() );
  }
    
  return true;
}

