/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

#include "RestorationWizard.h"

#include "RestorationWizardPage.h"
#include "RasterInfoWizardPage.h"
#include "../layer/search/LayerSearchWizardPage.h"
#include "../../../dataaccess/utils/Utils.h"
#include "../../../rp/Restoration.h"
#include "../help/HelpPushButton.h"
#include "../layer/search/LayerSearchWidget.h"
#include "../utils/ScopedCursor.h"
#include "../raster/RasterInfoWidget.h"
#include "Utils.h"
#include "../../../maptools/Utils.h"

#include <QMessageBox>

te::qt::widgets::RestorationWizard::RestorationWizard(QWidget* parent )
  : QWizard(parent)
{
  //configure the wizard
  this->setWizardStyle(QWizard::ModernStyle);
  this->setWindowTitle(tr("Raster Restoration"));
  //this->setFixedSize(640, 580);

  this->setOption(QWizard::HaveHelpButton, true);
  this->setOption(QWizard::HelpButtonOnRight, false);

  te::qt::widgets::HelpPushButton* helpButton = 
    new te::qt::widgets::HelpPushButton(this);

  this->setButton(QWizard::HelpButton, helpButton);

  helpButton->setPageReference("plugins/rp/rp_rasterrestoration.html");

  connect(this, SIGNAL(currentIdChanged(int)), SLOT(onPageChanged(int)));

  addPages();
}

te::qt::widgets::RestorationWizard::~RestorationWizard()
{
}

bool te::qt::widgets::RestorationWizard::validateCurrentPage()
{
  if(currentPage() ==  m_layerSearchPage.get())
  {
    std::list<te::map::AbstractLayerPtr> list = m_layerSearchPage->getSearchWidget()->getSelecteds();

    if(list.empty() == false)
    {
      te::map::AbstractLayerPtr l = *list.begin();

      m_mainWizardPagePage->setLayer(l);
    }

    return m_layerSearchPage->isComplete();
  }
  else if(currentPage() ==  m_mainWizardPagePage.get())
  {
    return m_mainWizardPagePage->isComplete();;
  }
  else if(currentPage() ==  m_rasterInfoPage.get())
  {
    return execute();
  }

  return true;
}

void te::qt::widgets::RestorationWizard::setList(std::list<te::map::AbstractLayerPtr>& layerList)
{
  m_layerSearchPage->getSearchWidget()->setList(layerList);
  m_layerSearchPage->getSearchWidget()->filterOnlyByRaster();
}

void te::qt::widgets::RestorationWizard::setLayer(te::map::AbstractLayerPtr layer)
{
  m_mainWizardPagePage->setLayer(layer);
}

te::map::AbstractLayerPtr te::qt::widgets::RestorationWizard::getOutputLayer()
{
  return m_outputLayer;
}

void te::qt::widgets::RestorationWizard::addPages()
{
  m_layerSearchPage.reset(new te::qt::widgets::LayerSearchWizardPage(this));
  m_mainWizardPagePage.reset(new te::qt::widgets::RestorationWizardPage(
    this ) );
  m_rasterInfoPage.reset(new te::qt::widgets::RasterInfoWizardPage(this));

  addPage(m_layerSearchPage.get());
  addPage(m_mainWizardPagePage.get());
  addPage(m_rasterInfoPage.get());

  m_layerSearchPage->getSearchWidget()->enableMultiSelection(false);
}

bool te::qt::widgets::RestorationWizard::execute()
{
  //get input raster
  
  te::map::AbstractLayer* layerPtr = m_mainWizardPagePage->getLayerPtr();
  if( layerPtr == nullptr )
  {
    QMessageBox::critical( this, tr("Error"), tr("No input layer selected"));
    return false;
  }    
    
  std::unique_ptr<te::da::DataSet> ds(layerPtr->getData());
  
  std::size_t rpos = te::da::GetFirstPropertyPos(ds.get(), te::dt::RASTER_TYPE);

  std::unique_ptr<te::rst::Raster> inputRst = ds->getRaster(rpos);  
  if( inputRst.get() == nullptr )
  {
    QMessageBox::critical( this, tr("Error"), tr("Invalid input layer selected"));
    return false;
  }        

  //run the algorithm
  
  std::vector< unsigned int > selectedInBandsDesignationIndexes;
  m_mainWizardPagePage->getSelectedInBandDesignations( selectedInBandsDesignationIndexes );
  if( selectedInBandsDesignationIndexes.size() != inputRst->getNumberOfBands() )
  {
    QMessageBox::critical( this, tr("Error"), tr("Invalid number of selected bands"));
    return false;
  }  

  te::rp::Restoration::InputParameters algoInputParams;  
  algoInputParams.m_sensorType = m_mainWizardPagePage->getSensorType();
  algoInputParams.m_sensorBandDesignations = selectedInBandsDesignationIndexes;
  algoInputParams.m_inRasterPtr = inputRst.get();
  for( unsigned int bandIdx = 0 ; bandIdx < inputRst->getNumberOfBands() ; ++bandIdx )
  {
    algoInputParams.m_inRasterBands.push_back( bandIdx );
  }
  algoInputParams.m_samplingFactor = m_mainWizardPagePage->getSamplingFactor();
  algoInputParams.m_enableProgress = true;
  algoInputParams.m_enableRasterCache = true;
  
  te::rp::Restoration::OutputParameters algoOutputParams;
  algoOutputParams.m_rType = m_rasterInfoPage->getWidget()->getType();
  algoOutputParams.m_rInfo = m_rasterInfoPage->getWidget()->getInfo();
  
  try
  {
//    te::qt::widgets::ProgressViewerDialog v(this);

    
    
    te::rp::Restoration algorithmInstance;

    if(algorithmInstance.initialize(algoInputParams))
    {
      bool execReturnOK = false;
      {
        te::qt::widgets::ScopedCursor cursor(Qt::WaitCursor);
        execReturnOK = algorithmInstance.execute(algoOutputParams);
      }
      
      if( execReturnOK )
      {
        algoOutputParams.reset();
        
        //set output layer
        m_outputLayer = te::qt::widgets::createLayer(m_rasterInfoPage->getWidget()->getType(), 
                                                     m_rasterInfoPage->getWidget()->getInfo());
        emit addLayer(m_outputLayer);
      }
      else
      {
        QMessageBox::critical(this, tr("Restoration"), 
          tr("Restoration execution error") +
          QString( " : ") + algorithmInstance.getErrorMessage().c_str() );
        
        return false;
      }
    }
    else
    {
      QMessageBox::critical(this, tr("Restoration"), 
        tr("Restoration initialization error") +
        QString( " : ") + QString( algorithmInstance.getErrorMessage().c_str() ) );
      
      return false;
    }
  }
  catch(...)
  {
    QMessageBox::critical(this, tr("Restoration"), 
      tr("Restoration execution error") );
    
    return false;
  }

  QMessageBox::information(this, tr("Restoration"), 
    tr("Restoration execution finished" ) );
  
  te::qt::widgets::applyRasterMultiResolution(tr("Restoration"), te::map::GetRaster(m_outputLayer.get()));
  
  return true;
}

void te::qt::widgets::RestorationWizard::onPageChanged(int /*index*/)
{
  this->adjustSize();
}

