/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.
  
    This file is part of the TerraLib - a Framework for building GIS enabled applications.
    
    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.
    
    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.
    
    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */


//Terralib
#include "RestorationWizardPage.h"

#include "../../../dataaccess/utils/Utils.h"

#include <QMessageBox>

te::qt::widgets::RestorationWizardPage::RestorationWizardPage(QWidget *parent) :
  QWizardPage(parent)
{
  m_uiPageForm.reset(new Ui::RestorationWizardPageForm);
  m_uiPageForm->setupUi(this);

  this->setTitle(tr("Raster Restoration"));
  this->setSubTitle(tr("Raster Restoration Parameters"));
  
  // fill sensor types combo
  
  std::map< te::rp::Restoration::InputParameters::SensorType, std::string > 
    supportedSensorTypes;
  te::rp::Restoration::InputParameters::getSupportedSensorTypes(
    supportedSensorTypes );
  
  std::map< te::rp::Restoration::InputParameters::SensorType, std::string >::const_iterator 
    supportedSensorTypesIt = supportedSensorTypes.begin();
    
  while( supportedSensorTypesIt != supportedSensorTypes.end() )
  {
    m_uiPageForm->m_sensorTypeComboBox->addItem( supportedSensorTypesIt->second.c_str(), 
      QVariant( (int)supportedSensorTypesIt->first ) );
    
    ++supportedSensorTypesIt;
  }
  

  // fill sampling factor combo
  
  std::map< te::rp::Restoration::InputParameters::SamplingFactorType, std::string > 
    supportedSamplingFactorTypes;
  te::rp::Restoration::InputParameters::getSupportedSamplingFactorsTypes(
    supportedSamplingFactorTypes );
  
  std::map< te::rp::Restoration::InputParameters::SamplingFactorType, std::string >::const_iterator 
    supportedSamplingFactorTypesIt = supportedSamplingFactorTypes.begin();
    
  while( supportedSamplingFactorTypesIt != supportedSamplingFactorTypes.end() )
  {
    m_uiPageForm->m_samplingFactorComboBox->addItem( supportedSamplingFactorTypesIt->second.c_str(), 
      QVariant( (int)supportedSamplingFactorTypesIt->first ) );
    
    ++supportedSamplingFactorTypesIt;
  }
  
  // connections
  
  connect(m_uiPageForm->m_bandDesignationsListWidget, SIGNAL(itemSelectionChanged()), this, 
    SLOT(onBandsItemSelectionChanged()));
  connect(m_uiPageForm->m_sensorTypeComboBox, SIGNAL(currentIndexChanged(int)), this, 
    SLOT(onSensorsCurrentIndexChanged(int)));   
  connect(m_uiPageForm->m_samplingFactorComboBox, SIGNAL(currentIndexChanged(int)), this, 
    SLOT(onSamplingFactorCurrentIndexChanged(int)));     
}

te::qt::widgets::RestorationWizardPage::~RestorationWizardPage() = default;

bool te::qt::widgets::RestorationWizardPage::isComplete() const
{
  QList<QListWidgetItem *> selectedBandsItems =
    m_uiPageForm->m_bandDesignationsListWidget->selectedItems();
    
  return (!selectedBandsItems.isEmpty());
}

void te::qt::widgets::RestorationWizardPage::setLayer(te::map::AbstractLayerPtr layer)
{
  //list bands
  m_uiPageForm->m_bandDesignationsListWidget->clear();

  //get input raster
  std::unique_ptr<te::da::DataSet> ds = layer->getData();

  if(ds.get())
  {
    std::size_t rpos = te::da::GetFirstPropertyPos(ds.get(), te::dt::RASTER_TYPE);

    std::unique_ptr<te::rst::Raster> inputRst = ds->getRaster(rpos);

    if(inputRst.get())
    {
      m_layer = layer;

      for(unsigned int i = 1; i <= 16; ++i)
      {
        m_uiPageForm->m_bandDesignationsListWidget->addItem(QString::number(i));
      }
      
      m_uiPageForm->m_samplingFactorComboBox->setCurrentIndex(0);
      onSamplingFactorCurrentIndexChanged(0);
    }
    else
    {
      QMessageBox::critical(
          this, tr("Warning"),
          tr("Invalid input layer"));
    }
  }
}

void te::qt::widgets::RestorationWizardPage::getSelectedInBandDesignations( 
  std::vector< unsigned int >& bandsIndexes ) const
{
  QList<QListWidgetItem *> selectedBandsItems =
    m_uiPageForm->m_bandDesignationsListWidget->selectedItems();
  
  QList<QListWidgetItem *>::const_iterator it = selectedBandsItems.begin();
  const QList<QListWidgetItem *>::const_iterator itEnd = selectedBandsItems.end();
  
  while( it != itEnd )
  {
    bandsIndexes.push_back( (*it)->text().toUInt() );
    ++it;
  }
}

te::map::AbstractLayer* te::qt::widgets::RestorationWizardPage::getLayerPtr()
{
  return m_layer.get();
}

void te::qt::widgets::RestorationWizardPage::onBandsItemSelectionChanged()
{
  m_uiPageForm->m_bandDesigLineEdit->clear();
  
  QList<QListWidgetItem *> selectedBandsItems =
    m_uiPageForm->m_bandDesignationsListWidget->selectedItems();
    
  QString bandsText;
    
  QList<QListWidgetItem *>::const_iterator it = selectedBandsItems.begin();
  const QList<QListWidgetItem *>::const_iterator itEnd = selectedBandsItems.end();
  
  while( it != itEnd )
  {
    if( bandsText.isEmpty() )
    {
      bandsText = QString::number( (*it)->text().toUInt() );
    }
    else
    {
      bandsText = bandsText + " + " + QString::number( (*it)->text().toUInt() );
    }

    ++it;
  }
  
  m_uiPageForm->m_bandDesigLineEdit->setText( bandsText );
  
  emit completeChanged();
}

void te::qt::widgets::RestorationWizardPage::onSensorsCurrentIndexChanged(int)
{
  
}

void te::qt::widgets::RestorationWizardPage::onSamplingFactorCurrentIndexChanged(
  int )
{
  m_uiPageForm->m_inputXResLineEdit->clear();
  m_uiPageForm->m_inputYResLineEdit->clear();
  m_uiPageForm->m_outputXResLineEdit->clear();
  m_uiPageForm->m_outputYResLineEdit->clear();
  
  if( m_layer.get() != nullptr )
  {
    std::unique_ptr<te::da::DataSet> ds = m_layer->getData();

    if(ds.get() != nullptr )
    {
      std::size_t rpos = te::da::GetFirstPropertyPos(ds.get(), te::dt::RASTER_TYPE);

      std::unique_ptr<te::rst::Raster> inputRst = ds->getRaster(rpos);    
      
      if( inputRst.get() != nullptr )
      {
        te::rp::Restoration::InputParameters::SamplingFactorType currentSFactor =
          getSamplingFactor();
          
          
        double resSamplingFactor = 0;
        te::rp::Restoration::InputParameters::getResolutionSamplingFactor( 
          currentSFactor, resSamplingFactor );
        
        const double inResX = inputRst->getResolutionX();
        const double inResY = inputRst->getResolutionY();
        
        m_uiPageForm->m_inputXResLineEdit->setText( QString::number( inResX ) );
        m_uiPageForm->m_inputYResLineEdit->setText( QString::number( inResY ) );
        m_uiPageForm->m_outputXResLineEdit->setText( QString::number( inResX *
          resSamplingFactor ) );
        m_uiPageForm->m_outputYResLineEdit->setText( QString::number( inResY *
          resSamplingFactor ) );
      }
    }
  }
}

te::rp::Restoration::InputParameters::SensorType 
  te::qt::widgets::RestorationWizardPage::getSensorType() const
{
  return (te::rp::Restoration::InputParameters::SensorType)
    m_uiPageForm->m_sensorTypeComboBox->currentData( Qt::UserRole ).toInt();
}

te::rp::Restoration::InputParameters::SamplingFactorType
  te::qt::widgets::RestorationWizardPage::getSamplingFactor() const
{
  return (te::rp::Restoration::InputParameters::SamplingFactorType)
    m_uiPageForm->m_samplingFactorComboBox->currentData( Qt::UserRole ).toInt();
}
            