/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/widgets/rp/SensorManagerDialog.cpp

  \brief A dialog used to manage Spectral Sensors.
*/

#include "../../af/ApplicationController.h"

#include "MixtureModelWizard.h"
#include "MixtureModelWizardPage.h"
#include "SensorManagerDialog.h"

#include "ui/ui_SensorManagerDialogForm.h"
#include <QApplication>
#include <QMessageBox>
#include <QSettings>
#include <QString>

#include <iostream>

te::qt::widgets::SensorManagerDialog::SensorManagerDialog( QWidget* parent, Qt::WindowFlags f):
  QDialog(parent, f),
  m_ui(new Ui::SensorManagerDialogForm)
{
  m_ui->setupUi(this);
  
  // Assign the edit/add/remove button icons
  m_ui->m_addToolButton->setIcon(QIcon::fromTheme("list-add"));

  connect(m_ui->m_okPushButton, SIGNAL(clicked()), SLOT(onOkPushButtonClicked()));
  connect(m_ui->m_cancelPushButton, SIGNAL(clicked()), SLOT(onCancelPushButtonClicked()));
  connect(m_ui->m_searchedSensorLineEdit, SIGNAL(textChanged(const QString&)), SLOT(onSearchLineEditTextChanged(const QString&)));
  connect(m_ui->m_addToolButton, SIGNAL(clicked()), SLOT(onAddToolButtonClicked()));
  connect(m_ui->m_savePushButton, SIGNAL(clicked()), SLOT(onSavePushButtonClicked()));

  m_ui->m_helpPushButton->setPageReference("plugins/rp/rp_sensor.html");

  updateSensorParams();

  m_edited = false;
  m_modified = false;
}


te::qt::widgets::SensorManagerDialog::~SensorManagerDialog()
{
  delete m_ui;
  m_modified = false;
}

const QString& te::qt::widgets::SensorManagerDialog::getSelectedSensor() const
{
  return m_sensorselected;
}
//
//const te::rp::SpectralSensorParams& te::qt::widgets::SensorManagerDialog::getSelectedSensorParams() const
//{
//  return m_sensorselectedparams;
//}

void te::qt::widgets::SensorManagerDialog::onSearchLineEditTextChanged(const QString& text)
{
  QList<QTableWidgetItem*> items = m_ui->m_SensorTableWidget->findItems(text, Qt::MatchContains | Qt::MatchRecursive);
  items.append(m_ui->m_SensorTableWidget->findItems(text, Qt::MatchContains | Qt::MatchRecursive));
  filter(items);
}


void te::qt::widgets::SensorManagerDialog::filter(const QList<QTableWidgetItem*>& items)
{
  for(int i = 0; i < m_ui->m_SensorTableWidget->rowCount(); ++i)
  {
    QTableWidgetItem* sensor = m_ui->m_SensorTableWidget->item(i,0);
    bool sel = items.indexOf(sensor) == -1;
    m_ui->m_SensorTableWidget->setRowHidden(i, sel);
  }
  update();
}


void te::qt::widgets::SensorManagerDialog::onSensorTableWidgetItemClicked(QTableWidgetItem* item)
{
  int r =  item->row();
  QTableWidgetSelectionRange selrow(r, 0, r, 3);
  m_ui->m_SensorTableWidget->setRangeSelected(selrow, true);
}

void te::qt::widgets::SensorManagerDialog::onSensorTableWidgetItemChanged(QTableWidgetItem* item)
{
  try
  {
    int r = item->row();
    if (m_ui->m_SensorTableWidget->item(r, 2)->text().size() && m_ui->m_SensorTableWidget->item(r, 3)->text().size())
    {
      double lower = m_ui->m_SensorTableWidget->item(r, 2)->text().toDouble();
      double upper = m_ui->m_SensorTableWidget->item(r, 3)->text().toDouble();
      double min = m_ui->m_SensorTableWidget->item(r, 4)->text().toDouble();
      double max = m_ui->m_SensorTableWidget->item(r, 5)->text().toDouble();
      if (upper <= lower)
      {
        QMessageBox::warning(this, "", tr("Invalid limits"));
        return;
      }

      if (max <= min)
      {
        QMessageBox::warning(this, "", tr("Invalid digital limits"));
        return;
      }

      std::string s = m_ui->m_SensorTableWidget->item(r, 0)->text().toUtf8().data();
      std::map<std::string, te::rp::SpectralSensorParams > ::iterator it = m_Params.find(s);
      if (it != m_Params.end())
      {
        it->second.m_band = m_ui->m_SensorTableWidget->item(r, 1)->text().toInt();
        if (it->second.m_lower != lower)
        {
          it->second.m_lower = lower;
          m_edited = true;
          m_modified = true;
        }
        if (it->second.m_upper != upper)
        {
          it->second.m_upper = upper;
          m_edited = true;
          m_modified = true;
        }
        if (it->second.m_min != min)
        {
          it->second.m_min = min;
          m_edited = true;
          m_modified = true;
        }
        if (it->second.m_max != max)
        {
          it->second.m_max = max;
          m_edited = true;
          m_modified = true;
        }
      }
      else
      {
        m_edited = true;
        m_modified = true;
      }
    }
  }
  catch (const std::exception&)
  {
    std::cout << std::endl << "Invalid Value" << std::endl;
  }

}

void te::qt::widgets::SensorManagerDialog::onOkPushButtonClicked()
{
  if (m_edited)
  {
    int reply = QMessageBox::question(nullptr, tr("Mixture Model"), tr("There are spectral sensors edited. Save?"), QMessageBox::No, QMessageBox::Yes);
    if (reply == QMessageBox::Yes)
    {
      te::rp::SaveSensorParams(m_Params);
    }
  }

  int i = m_ui->m_SensorTableWidget->currentRow();
  QTableWidgetItem* sensor = m_ui->m_SensorTableWidget->item(i, 0);
  if (!sensor)
  {
    reject();
    return;
  }

  m_sensorselected = sensor->text();
  accept();
}

void te::qt::widgets::SensorManagerDialog::onCancelPushButtonClicked()
{
  m_sensorselected.clear();
  reject();
}

void te::qt::widgets::SensorManagerDialog::onHelpPushButtonClicked()
{
 QMessageBox::critical(this, "", tr("Not implemented yet!"));
}

void te::qt::widgets::SensorManagerDialog::onAddToolButtonClicked()
{
  QList<QTableWidgetItem*> items = m_ui->m_SensorTableWidget->findItems(m_ui->m_searchedSensorLineEdit->text(), Qt::MatchContains | Qt::MatchRecursive);
  if (items.size())
  {
    QMessageBox::warning(this, "", tr("Sensor already exists!"));
    return;
  }

  m_ui->m_SensorTableWidget->disconnect(SIGNAL(itemChanged(QTableWidgetItem*)));

  int newrow = 0;

  m_ui->m_SensorTableWidget->insertRow(newrow);

  QTableWidgetItem *item0 = new QTableWidgetItem(m_ui->m_searchedSensorLineEdit->text());
  QTableWidgetItem *item1 = new QTableWidgetItem("");
  QTableWidgetItem *item2 = new QTableWidgetItem("");
  QTableWidgetItem *item3 = new QTableWidgetItem("");
  QTableWidgetItem *item4 = new QTableWidgetItem("");
  QTableWidgetItem *item5 = new QTableWidgetItem("");

  m_ui->m_SensorTableWidget->setItem(newrow, 0, item0);
  m_ui->m_SensorTableWidget->setItem(newrow, 1, item1);
  m_ui->m_SensorTableWidget->setItem(newrow, 2, item2);
  m_ui->m_SensorTableWidget->setItem(newrow, 3, item3);
  m_ui->m_SensorTableWidget->setItem(newrow, 4, item4);
  m_ui->m_SensorTableWidget->setItem(newrow, 5, item5);
#if (QT_VERSION >= 0x050000)
  m_ui->m_SensorTableWidget->horizontalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
#else
  m_ui->m_SensorTableWidget->horizontalHeader()->setResizeMode(QHeaderView::ResizeToContents);
#endif

  m_Params.insert(std::pair<std::string, te::rp::SpectralSensorParams>(m_ui->m_searchedSensorLineEdit->text().toUtf8().data(), te::rp::SpectralSensorParams(0,0,0,0,0)));

  connect(m_ui->m_SensorTableWidget, SIGNAL(itemChanged(QTableWidgetItem*)), SLOT(onSensorTableWidgetItemChanged(QTableWidgetItem*)));

}

void te::qt::widgets::SensorManagerDialog::onSavePushButtonClicked()
{
  te::rp::SaveSensorParams(m_Params);

  m_edited = false;
}

void te::qt::widgets::SensorManagerDialog::updateSensorParams()
{
  // define sensor information
  m_Params = te::rp::getSensorParams();

  int newrow = 0;
  m_ui->m_SensorTableWidget->setRowCount(newrow);

  for (std::map<std::string, te::rp::SpectralSensorParams > ::iterator it = m_Params.begin(); it != m_Params.end(); ++it)
  {
    m_ui->m_SensorTableWidget->insertRow(newrow);

    QTableWidgetItem *item0 = new QTableWidgetItem(it->first.c_str());
    QTableWidgetItem *item1 = new QTableWidgetItem();
    QTableWidgetItem *item2 = new QTableWidgetItem();
    QTableWidgetItem *item3 = new QTableWidgetItem();
    QTableWidgetItem *item4 = new QTableWidgetItem();
    QTableWidgetItem *item5 = new QTableWidgetItem();
    item1->setData(Qt::DisplayRole, it->second.m_band);
    item2->setData(Qt::DisplayRole, it->second.m_lower);
    item3->setData(Qt::DisplayRole, it->second.m_upper);
    item4->setData(Qt::DisplayRole, it->second.m_min);
    item5->setData(Qt::DisplayRole, it->second.m_max);

    m_ui->m_SensorTableWidget->setItem(newrow, 0, item0);
    m_ui->m_SensorTableWidget->setItem(newrow, 1, item1);
    m_ui->m_SensorTableWidget->setItem(newrow, 2, item2);
    m_ui->m_SensorTableWidget->setItem(newrow, 3, item3);
    m_ui->m_SensorTableWidget->setItem(newrow, 4, item4);
    m_ui->m_SensorTableWidget->setItem(newrow, 5, item5);
#if (QT_VERSION >= 0x050000)
    m_ui->m_SensorTableWidget->horizontalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
#else
    m_ui->m_SensorTableWidget->horizontalHeader()->setResizeMode(QHeaderView::ResizeToContents);
#endif

    newrow++;
  }

  // Signals & slots
  connect(m_ui->m_SensorTableWidget, SIGNAL(itemChanged(QTableWidgetItem*)), SLOT(onSensorTableWidgetItemChanged(QTableWidgetItem*)));

}
