/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/widgets/se/BasicFillPropertyItem.cpp

  \brief A widget used to define the basic fill se object.
*/

#include "BasicFillPropertyItem.h"
#include "../Utils.h"
#include "../../../color.h"
#include "../../../raster/Utils.h"
#include "../../../se/Fill.h"
#include "../../../se/SvgParameter.h"
#include "../../../se/Utils.h"
#include "../propertybrowser/AbstractPropertyManager.h"


// STL
#include <cassert>

// QtPropertyBrowser


te::qt::widgets::BasicFillPropertyItem::BasicFillPropertyItem(QtProperty* parent, QtTreePropertyBrowser* propertyBrowser, StylePropertyItemListener* listener, QColor color)
  : te::qt::widgets::CheckboxStylePropertyItem(propertyBrowser, tr("Basic Fill"), listener, color)
  , m_colorProperty(nullptr)
  , m_opacityProperty(nullptr)
{
  //add property to parent
  QtProperty* itemProperty = this->getItemProperty();
  addSubProperty(parent, itemProperty, color);

  //color
  m_colorProperty = te::qt::widgets::AbstractPropertyManager::getInstance().m_colorManager->addProperty(tr("Color"));
  addListenerToProperty(m_colorProperty);

  //opacity
  m_opacityProperty = te::qt::widgets::AbstractPropertyManager::getInstance().m_intSliderManager->addProperty(tr("Opacity"));
  te::qt::widgets::AbstractPropertyManager::getInstance().m_intSliderManager->setMinimum(m_opacityProperty, 0);
  te::qt::widgets::AbstractPropertyManager::getInstance().m_intSliderManager->setMaximum(m_opacityProperty, 255);
  te::qt::widgets::AbstractPropertyManager::getInstance().m_intSliderManager->setSingleStep(m_opacityProperty, 10);
  addListenerToProperty(m_opacityProperty);
  
  itemProperty->addSubProperty(m_colorProperty);
  itemProperty->addSubProperty(m_opacityProperty);

  setFill(nullptr);
}

te::qt::widgets::BasicFillPropertyItem::~BasicFillPropertyItem()
{
  delete m_colorProperty;
  delete m_opacityProperty;
}

bool te::qt::widgets::BasicFillPropertyItem::setFill(const te::se::Fill* fill)
{
  te::color::RGBAColor rgba(TE_SE_DEFAULT_FILL_BASIC_COLOR, TE_OPAQUE);
  double opacity = TE_SE_DEFAULT_OPACITY;

  if (fill != nullptr)
  {
    te::se::GetColor(fill, rgba);
    if (fill->getOpacity())
    {
      opacity = te::se::GetDouble(fill->getOpacity());
    }

    this->setChecked(true);
  }

  m_enableListeners = false;

  //Fill Color
  int alpha = te::common::Round<double, int>(opacity * 255.);

  QColor color(rgba.getRgba());
  color.setAlpha(alpha);
  te::qt::widgets::AbstractPropertyManager::getInstance().m_colorManager->setValue(m_colorProperty, color);

  // Opacity
  te::qt::widgets::AbstractPropertyManager::getInstance().m_intSliderManager->setValue(m_opacityProperty, alpha);  

  m_enableListeners = true;

  return true;
}

te::se::Fill* te::qt::widgets::BasicFillPropertyItem::getFill() const
{
  //we always return a new instance
  te::se::Fill* newItem = new te::se::Fill();

  //Fill Color
  QColor color = te::qt::widgets::AbstractPropertyManager::getInstance().m_colorManager->value(m_colorProperty);

  // Opacity
  int alpha = te::qt::widgets::AbstractPropertyManager::getInstance().m_intSliderManager->value(m_opacityProperty);
  color.setAlpha(alpha);

  double opacity = static_cast<double>(alpha) / 255.;

  // Updating fill
  newItem->setColor(Convert2StdString(color.name()));
  newItem->setOpacity(Convert2StdString(QString::number(opacity, 'f', 2)));

  return newItem;
}

void te::qt::widgets::BasicFillPropertyItem::valueChanged(QtProperty* p, const QVariant& qVariant)
{
  if (m_enableListeners == false)
  {
    StylePropertyItem::valueChanged(p, qVariant);
    return;
  }

  m_enableListeners = false;

  if (p == m_colorProperty)
  {
    QColor color = qVariant.value<QColor>();
    int alpha = color.alpha();
    te::qt::widgets::AbstractPropertyManager::getInstance().m_intSliderManager->setValue(m_opacityProperty, alpha);
  }
  else if (p == m_opacityProperty)
  {
    int alpha = qVariant.toInt();
    QColor color = te::qt::widgets::AbstractPropertyManager::getInstance().m_colorManager->value(m_colorProperty);
    color.setAlpha(alpha);
    te::qt::widgets::AbstractPropertyManager::getInstance().m_colorManager->setValue(m_colorProperty, color);
  }

  m_enableListeners = true;

  StylePropertyItem::valueChanged(p, qVariant);
}
