/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/widgets/se/BasicStrokePropertyItem.cpp

  \brief A widget used to define the basic fill se object.
*/

// TerraLib
#include "BasicStrokePropertyItem.h"

#include "../../../common/MathUtils.h"
#include "../../../se/Config.h"
#include "../../../se/Stroke.h"
#include "../../../se/SvgParameter.h"
#include "../../../se/Utils.h"
#include "../../../maptools/Utils.h"
#include "../../../maptools/Enums.h"
#include "../propertybrowser/AbstractPropertyManager.h"
#include "../Utils.h"


// Qt
#include <QPainter>
#include <QIcon>

// QtPropertyBrowser
#include <QtPropertyBrowser/QtTreePropertyBrowser>
#include <QtPropertyBrowser/QtVariantPropertyManager>

// STL
#include <cassert>


te::qt::widgets::BasicStrokePropertyItem::BasicStrokePropertyItem(QtProperty* parent, QtTreePropertyBrowser* pb, StylePropertyItemListener* listener, QColor c)
  : te::qt::widgets::CheckboxStylePropertyItem(pb, tr("Basic Stroke"), listener, c)
  , m_colorProperty(nullptr)
  , m_opacityProperty(nullptr)
  , m_widthProperty(nullptr)
  , m_joinProperty(nullptr)
  , m_capProperty(nullptr)
  , m_dashOffsetProperty(nullptr)
{
  //add property to parent
  QtProperty* itemProperty = this->getItemProperty();
  addSubProperty(parent, itemProperty, c);

  //color
  m_colorProperty = te::qt::widgets::AbstractPropertyManager::getInstance().m_colorManager->addProperty(tr("Color"));
  addListenerToProperty(m_colorProperty);

  //opacity
  m_opacityProperty = te::qt::widgets::AbstractPropertyManager::getInstance().m_intSliderManager->addProperty(tr("Opacity"));
  te::qt::widgets::AbstractPropertyManager::getInstance().m_intSliderManager->setMinimum(m_opacityProperty, 0);
  te::qt::widgets::AbstractPropertyManager::getInstance().m_intSliderManager->setMaximum(m_opacityProperty, 255);
  te::qt::widgets::AbstractPropertyManager::getInstance().m_intSliderManager->setSingleStep(m_opacityProperty, 10);
  addListenerToProperty(m_opacityProperty);

  //width
  m_widthProperty = te::qt::widgets::AbstractPropertyManager::getInstance().m_doubleManager->addProperty(tr("Width"));
  te::qt::widgets::AbstractPropertyManager::getInstance().m_doubleManager->setMinimum(m_widthProperty, 0);
  te::qt::widgets::AbstractPropertyManager::getInstance().m_doubleManager->setMaximum(m_widthProperty, 100);
  te::qt::widgets::AbstractPropertyManager::getInstance().m_doubleManager->setSingleStep(m_widthProperty, 1);
  te::qt::widgets::AbstractPropertyManager::getInstance().m_doubleManager->setDecimals(m_widthProperty, 2);
  addListenerToProperty(m_widthProperty);

  //join
  m_joinProperty = te::qt::widgets::AbstractPropertyManager::getInstance().m_enumManager->addProperty(tr("Join"));
  te::qt::widgets::AbstractPropertyManager::getInstance().m_enumManager->setEnumNames(m_joinProperty, getJoinNames());
  te::qt::widgets::AbstractPropertyManager::getInstance().m_enumManager->setEnumIcons(m_joinProperty, getJoinIcons());
  addListenerToProperty(m_joinProperty);

  //cap
  m_capProperty = te::qt::widgets::AbstractPropertyManager::getInstance().m_enumManager->addProperty(tr("Cap"));
  te::qt::widgets::AbstractPropertyManager::getInstance().m_enumManager->setEnumNames(m_capProperty, getCapNames());
  te::qt::widgets::AbstractPropertyManager::getInstance().m_enumManager->setEnumIcons(m_capProperty, getCapIcons());
  addListenerToProperty(m_capProperty);

  //dash
  m_dashProperty = te::qt::widgets::AbstractPropertyManager::getInstance().m_enumManager->addProperty(tr("Dash"));
  te::qt::widgets::AbstractPropertyManager::getInstance().m_enumManager->setEnumNames(m_dashProperty, getDashIconsNames());
  te::qt::widgets::AbstractPropertyManager::getInstance().m_enumManager->setEnumIcons(m_dashProperty, getDashIcons());
  addListenerToProperty(m_dashProperty);

  //dash offset
  m_dashOffsetProperty = te::qt::widgets::AbstractPropertyManager::getInstance().m_doubleManager->addProperty(tr("Dash Offset"));
  te::qt::widgets::AbstractPropertyManager::getInstance().m_doubleManager->setMinimum(m_dashOffsetProperty, 0);
  te::qt::widgets::AbstractPropertyManager::getInstance().m_doubleManager->setMaximum(m_dashOffsetProperty, 100);
  te::qt::widgets::AbstractPropertyManager::getInstance().m_doubleManager->setSingleStep(m_dashOffsetProperty, 1);
  te::qt::widgets::AbstractPropertyManager::getInstance().m_doubleManager->setDecimals(m_dashOffsetProperty, 2);
  addListenerToProperty(m_dashOffsetProperty);

  itemProperty->addSubProperty(m_colorProperty);
  itemProperty->addSubProperty(m_opacityProperty);
  itemProperty->addSubProperty(m_widthProperty);
  itemProperty->addSubProperty(m_joinProperty);
  itemProperty->addSubProperty(m_capProperty);
  itemProperty->addSubProperty(m_dashProperty);
  itemProperty->addSubProperty(m_dashOffsetProperty);

  setStroke(nullptr);
}

te::qt::widgets::BasicStrokePropertyItem::~BasicStrokePropertyItem()
{
  delete m_colorProperty;
  delete m_opacityProperty;
  delete m_widthProperty;
  delete m_joinProperty;
  delete m_capProperty;
  delete m_dashProperty;
  delete m_dashOffsetProperty;
}

void te::qt::widgets::BasicStrokePropertyItem::setStroke(const te::se::Stroke* stroke)
{
  te::color::RGBAColor rgba(TE_SE_DEFAULT_STROKE_BASIC_COLOR, TE_OPAQUE);
  double opacity = TE_SE_DEFAULT_OPACITY;
  double width = TE_SE_DEFAULT_STROKE_BASIC_WIDTH;
  std::string join = TE_SE_MITRE_JOIN;
  std::string cap = TE_SE_BUTT_CAP;
  std::string dasharray = "";
  double dashOffset = 0.;

  if (stroke != nullptr)
  {
    te::se::GetColor(stroke, rgba);

    if (stroke->getOpacity())
    {
      opacity = te::se::GetDouble(stroke->getOpacity());
    }

    const te::se::SvgParameter* widthParameter = stroke->getWidth();
    if (widthParameter != nullptr)
    {
      width = te::se::GetDouble(widthParameter);
    }

    const te::se::SvgParameter* joinParameter = stroke->getLineJoin();
    if (joinParameter != nullptr)
    {
      join = te::se::GetString(joinParameter);
    }

    const te::se::SvgParameter* capParameter = stroke->getLineCap();
    if (capParameter != nullptr)
    {
      cap = te::se::GetString(capParameter);
    }

    const te::se::SvgParameter* dasharrayParameter = stroke->getDashArray();
    if (dasharrayParameter != nullptr)
    {
      dasharray = te::se::GetString(dasharrayParameter);
    }
    const te::se::SvgParameter* dashOffsetParameter = stroke->getDashOffset();
    if (dashOffsetParameter != nullptr)
    {
      dashOffset = te::se::GetDouble(dashOffsetParameter);
    }
  }

  //sets the values to the components
  m_enableListeners = false;

  //Color
  int alpha = te::common::Round<double, int>(opacity * 255.);

  QColor c = QColor(rgba.getRgba());
  c.setAlpha(alpha);
  te::qt::widgets::AbstractPropertyManager::getInstance().m_colorManager->setValue(m_colorProperty, c);

  // Opacity
  te::qt::widgets::AbstractPropertyManager::getInstance().m_intSliderManager->setValue(m_opacityProperty, alpha);

  // Width
  te::qt::widgets::AbstractPropertyManager::getInstance().m_doubleManager->setValue(m_widthProperty, width);

  //Join
  AbstractPropertyItem::setValueToEnumProperty(m_joinProperty, Convert2QString(join));

  //Cap
  AbstractPropertyItem::setValueToEnumProperty(m_capProperty, Convert2QString(cap));

  //Dasharray
  AbstractPropertyItem::setValueToEnumProperty(m_dashProperty, Convert2QString(dasharray));

  //Dash Offset
  te::qt::widgets::AbstractPropertyManager::getInstance().m_doubleManager->setValue(m_dashOffsetProperty, dashOffset);

  m_enableListeners = true;
}

te::se::Stroke* te::qt::widgets::BasicStrokePropertyItem::getStroke() const
{
  //Color
  QColor color = te::qt::widgets::AbstractPropertyManager::getInstance().m_colorManager->value(m_colorProperty);

  // Opacity
  int alpha = te::qt::widgets::AbstractPropertyManager::getInstance().m_intSliderManager->value(m_opacityProperty);
  color.setAlpha(alpha);
  double opacity = static_cast<double>(alpha) / 255.;

  //Width
  double width = te::qt::widgets::AbstractPropertyManager::getInstance().m_doubleManager->value(m_widthProperty);

  //Join
  QString qJoin = AbstractPropertyItem::getValueFromEnumProperty(m_joinProperty);
  std::string join = Convert2StdString(qJoin);

  //Cap
  QString qCap = AbstractPropertyItem::getValueFromEnumProperty(m_capProperty);
  std::string cap = Convert2StdString(qCap);

  //Dash array
  QString qDasharray = AbstractPropertyItem::getValueFromEnumProperty(m_dashProperty);
  std::string dasharray = Convert2StdString(qDasharray);

  //Dash Offset
  double dashOffset = te::qt::widgets::AbstractPropertyManager::getInstance().m_doubleManager->value(m_dashOffsetProperty);

  // Updating item
  te::se::Stroke* newItem = new te::se::Stroke();
  newItem->setColor(Convert2StdString(color.name()));
  newItem->setOpacity(Convert2StdString(QString::number(opacity, 'f', 2)));
  newItem->setWidth(Convert2StdString(QString::number(width, 'f', 2)));
  newItem->setLineJoin(join);
  newItem->setLineCap(cap);
  newItem->setDashArray(dasharray);
  newItem->setDashOffset(Convert2StdString(QString::number(dashOffset, 'f', 2)));

  return newItem;
}

void te::qt::widgets::BasicStrokePropertyItem::valueChanged(QtProperty* p, const QVariant& qVariant)
{
  if (m_enableListeners == false)
  {
    StylePropertyItem::valueChanged(p, qVariant);
    return;
  }

  m_enableListeners = false;

  if (p == m_colorProperty)
  {
    QColor color = qVariant.value<QColor>();
    int alpha = color.alpha();
    te::qt::widgets::AbstractPropertyManager::getInstance().m_intSliderManager->setValue(m_opacityProperty, alpha);
  }
  else if (p == m_opacityProperty)
  {
    int alpha = qVariant.toInt();
    QColor color = te::qt::widgets::AbstractPropertyManager::getInstance().m_colorManager->value(m_colorProperty);
    color.setAlpha(alpha);
    te::qt::widgets::AbstractPropertyManager::getInstance().m_colorManager->setValue(m_colorProperty, color);
  }

  m_enableListeners = true;

  StylePropertyItem::valueChanged(p, qVariant);
}

QMap<int, QIcon> te::qt::widgets::BasicStrokePropertyItem::getDashIcons()
{
  QMap<int, QIcon> map;

  // Dash graphical representation size
  QSize size(32, 16);

  // Line that will be draw with the dashe styles
  QLine line(0, size.height() * 0.5, size.width(), size.height() * 0.5);

  // Setuping...
  QPen pen;
  pen.setWidth(2);

  int dSize = static_cast<int>(m_dashes.size());

  for(int t = 0; t < dSize; ++t)
  {
    QPixmap pixmap(size);
    pixmap.fill(Qt::transparent);

    QPainter painter(&pixmap);

    std::string pattern = m_dashes[t];
    std::vector<double> dasharray;
    te::map::GetDashStyle(pattern, dasharray);
      
    QVector<qreal> qdasharray = QVector<qreal>::fromStdVector(dasharray);

    pen.setDashPattern(qdasharray);
    painter.setPen(pen);
    painter.drawLine(line);

    QIcon icon(pixmap);

    map[t] = icon;
  }

  return map;
}

QMap<int, QIcon> te::qt::widgets::BasicStrokePropertyItem::getJoinIcons()
{
  QMap<int, QIcon> map;

  map[0] = QIcon::fromTheme("stroke-join-miter").pixmap(16,16);
  map[1] = QIcon::fromTheme("stroke-join-round").pixmap(16,16);
  map[2] = QIcon::fromTheme("stroke-join-bevel").pixmap(16,16);

  return map;
}

QMap<int, QIcon> te::qt::widgets::BasicStrokePropertyItem::getCapIcons()
{
  QMap<int, QIcon> map;

  map[0] = QIcon::fromTheme("stroke-cap-butt").pixmap(16,16);
  map[1] = QIcon::fromTheme("stroke-cap-round").pixmap(16,16);
  map[2] = QIcon::fromTheme("stroke-cap-square").pixmap(16,16);

  return map;
}

QStringList te::qt::widgets::BasicStrokePropertyItem::getDashIconsNames()
{
  m_dashes.push_back("");
  m_dashes.push_back("1 2");
  m_dashes.push_back("4 2");
  m_dashes.push_back("4 2 1 2");
  m_dashes.push_back("4 2 1 2 1 2");

  QStringList l;

  for(size_t t = 0; t < m_dashes.size(); ++t)
  {
     l << Convert2QString(m_dashes[t]);
  }

  return l;
}

QStringList te::qt::widgets::BasicStrokePropertyItem::getJoinNames()
{
  m_joinMap.insert(std::map<int, std::string>::value_type(0, TE_SE_MITRE_JOIN));
  m_joinMap.insert(std::map<int, std::string>::value_type(1, TE_SE_ROUND_JOIN));
  m_joinMap.insert(std::map<int, std::string>::value_type(2, TE_SE_BEVEL_JOIN));

  QStringList l;

  l << Convert2QString(m_joinMap[0]);
  l << Convert2QString(m_joinMap[1]);
  l << Convert2QString(m_joinMap[2]);

  return l;
}

QStringList te::qt::widgets::BasicStrokePropertyItem::getCapNames()
{
  m_capMap.insert(std::map<int, std::string>::value_type(0, TE_SE_BUTT_CAP));
  m_capMap.insert(std::map<int, std::string>::value_type(1, TE_SE_ROUND_CAP));
  m_capMap.insert(std::map<int, std::string>::value_type(2, TE_SE_SQUARE_CAP));

  QStringList l;

  l << Convert2QString(m_capMap[0]);
  l << Convert2QString(m_capMap[1]);
  l << Convert2QString(m_capMap[2]);

  return l;
}
