/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralib/qt/widgets/se/CreateRuleDialog.cpp

   \brief A dialog used to create a rule and define the filter restriction.
 */

// TerraLib
#include "CreateRuleDialog.h"
#include "ui/ui_CreateRuleDialogForm.h"

#include "../../../dataaccess/utils/Utils.h"
#include "../../../geometry/GeometryProperty.h"
#include "../../../maptools/FilterEncoder.h"
#include "../../../maptools/QueryEncoder.h"
#include "../../../se/Utils.h"

// Qt
#include <QGridLayout>
#include <QMessageBox>

te::qt::widgets::CreateRuleDialog::CreateRuleDialog(QWidget* parent, Qt::WindowFlags f)
  : QDialog(parent, f),
    m_ui(new Ui::CreateRuleDialogForm),
    m_rule(nullptr)
{
  m_ui->setupUi(this);

  //create query builder widget
  QGridLayout* layout = new QGridLayout(m_ui->m_filterGroupBox);
  m_queryWidget = new te::qt::widgets::QueryBuilderWidget(m_ui->m_filterGroupBox);
  layout->addWidget(m_queryWidget);
  layout->setContentsMargins(0, 0, 0, 0);

  //connects
  connect(m_ui->m_okPushButton, SIGNAL(clicked()), this, SLOT(onOkPushButtonClicked()));
}

te::qt::widgets::CreateRuleDialog::~CreateRuleDialog()
{
  delete m_rule;
}

void te::qt::widgets::CreateRuleDialog::setDataInformation(std::string dataSourceId, std::string dataSetName)
{
  te::da::DataSourcePtr dsPtr = te::da::GetDataSource(dataSourceId);

  if (!dsPtr->isOpened())
    dsPtr->open();

  std::unique_ptr<te::da::DataSetType> dsType = dsPtr->getDataSetType(dataSetName);

  te::gm::GeometryProperty* geometryProperty = te::da::GetFirstGeomProperty(dsType.get());
  m_layerGeomType = geometryProperty->getGeometryType();

  m_queryWidget->setInputParameters(dsPtr, dataSetName, std::move(dsType));

  m_queryWidget->startWhereClause();
}

void te::qt::widgets::CreateRuleDialog::setRule(te::se::Rule* rule)
{
  QString name = QObject::tr("Default Rule").toUtf8().data();

  m_rule = rule;

  if (rule->getName())
  {
    name = (*rule->getName()).c_str();
  }

  m_ui->m_ruleLineEdit->setText(name);

  const te::fe::Filter* filter = rule->getFilter();
  if (filter != nullptr)
  {
    te::map::QueryEncoder qe;
    te::da::Expression* exp = qe.getExpression(filter);

    te::da::Where w(exp);

    m_queryWidget->setWhereClause(&w);

    setFilterGroupBoxChecked(true);
  }
}

te::se::Rule* te::qt::widgets::CreateRuleDialog::getRule()
{
  if (m_rule)
  {
    return m_rule->clone();
  }

  return nullptr;
}

void te::qt::widgets::CreateRuleDialog::onOkPushButtonClicked()
{
  if (m_ui->m_ruleLineEdit->text().isEmpty())
  {
    QMessageBox::warning(this, tr("Create Rule"), tr("Rule name not defined."));
    return;
  }

  std::string* ruleName = new std::string(m_ui->m_ruleLineEdit->text().toUtf8().data());

  if (m_rule == nullptr)
  {
    m_rule = new te::se::Rule;

    te::se::Symbolizer* symbolizer = te::se::CreateSymbolizer(m_layerGeomType);
    m_rule->push_back(symbolizer);
  }

  m_rule->setName(ruleName);

  if (isFilterGroupBoxCheckable() == true
    && isFilterGroupBoxChecked() == true)
  {
    //get where clause
    std::unique_ptr<te::da::Where> whereClause(m_queryWidget->getWhereClause());
    const te::da::Expression* whereExp = whereClause->getExp();

    //converte terralib query to filter
    te::map::FilterEncoder fe;
    te::fe::Filter* filter = fe.getFilter(whereExp);

    m_rule->setFilter(filter);
  }
  else
  {
    m_rule->setFilter(nullptr);
  }


  accept();
}

bool te::qt::widgets::CreateRuleDialog::isFilterGroupBoxCheckable()
{
  return m_ui->m_filterGroupBox->isCheckable();
}

void te::qt::widgets::CreateRuleDialog::setFilterGroupBoxCheckable(bool check)
{
  m_ui->m_filterGroupBox->setCheckable(check);
}

bool te::qt::widgets::CreateRuleDialog::isFilterGroupBoxChecked()
{
  return m_ui->m_filterGroupBox->isChecked();
}

void te::qt::widgets::CreateRuleDialog::setFilterGroupBoxChecked(bool check)
{
  m_ui->m_filterGroupBox->setChecked(check);
}
