/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/widgets/ExternalGraphicRenderer.cpp

  \brief A concrete renderer based on Qt4 for conversion of Symbology Enconding External Graphic elements to an image pattern.
*/

// TerraLib
#include "../../../se/ExternalGraphic.h"
#include "../../../se/InlineContent.h"
#include "../../../xlink/SimpleLink.h"
#include "../Utils.h"
#include "ExternalGraphicRenderer.h"

// Qt
#include <QImage>
#include <QMimeDatabase>
#include <QPainter>
#include <QSvgRenderer>

te::qt::widgets::ExternalGraphicRenderer::ExternalGraphicRenderer()

{
}

te::qt::widgets::ExternalGraphicRenderer::~ExternalGraphicRenderer() = default;

te::color::RGBAColor** te::qt::widgets::ExternalGraphicRenderer::render(const te::se::ExternalGraphic* eg, std::size_t height, double rotation, std::size_t& width)
{
  std::string format = eg->getFormat().c_str();

  QMimeDatabase mimeTypeDb;
  QMimeType mimeType = mimeTypeDb.mimeTypeForName(te::qt::widgets::Convert2QString(format));

  const te::se::InlineContent* inLine = eg->getInlineContent();

  if (inLine)
  {
    std::string sufix = te::qt::widgets::Convert2StdString(mimeType.preferredSuffix());
    if (sufix == "svg")
    {
      const std::string strXML = inLine->getData();

      QByteArray ba;
      ba.append(strXML.c_str());

      QSvgRenderer renderer(QByteArray::fromBase64(ba));

      QImage image(static_cast<int>(width), static_cast<int>(height), QImage::Format_ARGB32);
      image.fill(Qt::transparent);

      QPainter painter(&image);
      painter.rotate(-rotation);
      renderer.render(&painter);

      te::color::RGBAColor** rgba = te::qt::widgets::GetImage(&image);

      return rgba;
    }
    else
    {
      const std::string strBase64 = inLine->getData();

      QByteArray ba;
      ba.append(strBase64.c_str());

      QImage img = QImage::fromData(QByteArray::fromBase64(ba), sufix.c_str());
      if (rotation != 0)
      {
        QMatrix matrix;
        matrix.translate(static_cast<double>(img.width()) / 2., static_cast<double>(img.height()) / 2.);
        matrix.rotate(-rotation);

        img = img.transformed(matrix);
      }

      // Verifies the requested size
      if (img.height() != static_cast<int>(height))
        img = img.scaledToHeight(static_cast<int>(height), Qt::SmoothTransformation);

      width = static_cast<size_t>(img.width());

      // Converts QImage to te::color::RGBA**
      te::color::RGBAColor** rgba = te::qt::widgets::GetImage(&img);

      return rgba;
    }
  }

  const te::xl::SimpleLink* link = eg->getOnlineResource();
  if(link == nullptr)
    return nullptr;

  const std::string uri = link->getHref();
  if(uri.empty())
    return nullptr;

  // Try load the image
  QImage img;
  if(!img.load(Convert2QString(uri)))
    return nullptr;

  if (rotation != 0)
  {
    QMatrix matrix;
    matrix.translate(static_cast<double>(img.width()) / 2., static_cast<double>(img.height()) / 2.);
    matrix.rotate(-rotation);

    img = img.transformed(matrix);
  }

  // Verifies the requested size
  if(img.height() != static_cast<int>(height))
    img = img.scaledToHeight(static_cast<int>(height), Qt::SmoothTransformation);

  width = static_cast<size_t>(img.width());

  // Converts QImage to te::color::RGBA**
  te::color::RGBAColor** rgba = te::qt::widgets::GetImage(&img);

  return rgba;
}
