/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralib/qt/widgets/se/GraphicFillPropertyItem.cpp

   \brief A widget used to define the graphic fill se object.
 */

#include "GraphicFillPropertyItem.h"

#include "GroupPropertyItemHandler.h"
#include "MarkExternalGraphicPropertyItem.h"
#include "MarkGlyphPropertyItem.h"
#include "MarkWellKnownPropertyItem.h"

 // TerraLib Includes
#include "../propertybrowser/AbstractPropertyManager.h"
#include "../Utils.h"
#include "../../../common/MathUtils.h"
#include "../../../se/Graphic.h"
#include "../../../se/Mark.h"
#include "../../../se/ParameterValue.h"
#include "../../../se/Utils.h"

te::qt::widgets::GraphicFillPropertyItem::GraphicFillPropertyItem(QtProperty* parent, QtTreePropertyBrowser* propertyBrowser, StylePropertyItemListener* listener, QColor color, QColor subColor, QString caption)
  : te::qt::widgets::CheckboxStylePropertyItem(propertyBrowser, caption, listener, color)
  , m_groupHandler(new GroupPropertyItemHandler())
  , m_sizeProperty(nullptr)
  , m_angleProperty(nullptr)
  , m_opacityProperty(nullptr)
  , m_markWellKnownProperty(nullptr)
  , m_markGlyphProperty(nullptr)
  , m_markExternalGraphicProperty(nullptr)
{
  //add property to parent
  QtProperty* itemProperty = this->getItemProperty();
  addSubProperty(parent, itemProperty, color);

  //size
  m_sizeProperty = te::qt::widgets::AbstractPropertyManager::getInstance().m_doubleManager->addProperty(tr("Size"));
  te::qt::widgets::AbstractPropertyManager::getInstance().m_doubleManager->setMinimum(m_sizeProperty, 1);
  te::qt::widgets::AbstractPropertyManager::getInstance().m_doubleManager->setMaximum(m_sizeProperty, 1024);
  te::qt::widgets::AbstractPropertyManager::getInstance().m_doubleManager->setSingleStep(m_sizeProperty, 1);
  te::qt::widgets::AbstractPropertyManager::getInstance().m_doubleManager->setDecimals(m_sizeProperty, 2);
  addListenerToProperty(m_sizeProperty);
  itemProperty->addSubProperty(m_sizeProperty);

  //angle
  m_angleProperty = te::qt::widgets::AbstractPropertyManager::getInstance().m_doubleManager->addProperty(tr("Angle"));
  te::qt::widgets::AbstractPropertyManager::getInstance().m_doubleManager->setMinimum(m_angleProperty, 0);
  te::qt::widgets::AbstractPropertyManager::getInstance().m_doubleManager->setMaximum(m_angleProperty, 360);
  te::qt::widgets::AbstractPropertyManager::getInstance().m_doubleManager->setSingleStep(m_angleProperty, 10);
  te::qt::widgets::AbstractPropertyManager::getInstance().m_doubleManager->setDecimals(m_angleProperty, 2);
  addListenerToProperty(m_angleProperty);
  itemProperty->addSubProperty(m_angleProperty);

  //opacity
  m_opacityProperty = te::qt::widgets::AbstractPropertyManager::getInstance().m_intSliderManager->addProperty(tr("Opacity"));
  te::qt::widgets::AbstractPropertyManager::getInstance().m_intSliderManager->setMinimum(m_opacityProperty, 0);
  te::qt::widgets::AbstractPropertyManager::getInstance().m_intSliderManager->setMaximum(m_opacityProperty, 255);
  te::qt::widgets::AbstractPropertyManager::getInstance().m_intSliderManager->setSingleStep(m_opacityProperty, 10);
  addListenerToProperty(m_opacityProperty);
  itemProperty->addSubProperty(m_opacityProperty);

  //graphic
  m_markWellKnownProperty = new te::qt::widgets::MarkWellKnownPropertyItem(itemProperty, propertyBrowser, listener, subColor);
  m_groupHandler->addPropertyToGroup(m_markWellKnownProperty);

  //graphic
  m_markGlyphProperty = new te::qt::widgets::MarkGlyphPropertyItem(itemProperty, propertyBrowser, listener, subColor);
  m_groupHandler->addPropertyToGroup(m_markGlyphProperty);

  //graphic
  m_markExternalGraphicProperty = new te::qt::widgets::MarkExternalGraphicPropertyItem(itemProperty, propertyBrowser, listener, subColor);
  m_groupHandler->addPropertyToGroup(m_markExternalGraphicProperty);

  setGraphic(nullptr);
}

te::qt::widgets::GraphicFillPropertyItem::~GraphicFillPropertyItem()
{
  delete m_groupHandler;
  delete m_sizeProperty;
  delete m_angleProperty;
  delete m_opacityProperty;
  delete m_markWellKnownProperty;
  delete m_markGlyphProperty;
  delete m_markExternalGraphicProperty;
}

bool te::qt::widgets::GraphicFillPropertyItem::setGraphic(const te::se::Graphic* graphic)
{
  m_enableListeners = false;

  double size = TE_SE_DEFAULT_GRAPHIC_SIZE;
  double rotation = TE_SE_DEFAULT_ROTATION;
  double opacity = TE_SE_DEFAULT_OPACITY;

  bool isGlyphMark = false;
  bool isWKMMark = false;
  bool isExternalGraphic = false;
  te::se::Mark* mark = nullptr;
  te::se::ExternalGraphic* externalGraphic = nullptr;

  if (graphic != nullptr)
  {
    // Size
    const te::se::ParameterValue* sizeParam = graphic->getSize();
    if (sizeParam != nullptr)
    {
      size = te::se::GetDouble(sizeParam);
    }

    // Rotation
    const te::se::ParameterValue* rotationParam = graphic->getRotation();
    if (rotationParam != nullptr)
    {
      rotation = te::se::GetDouble(rotationParam);
    }

    // Opacity
    const te::se::ParameterValue* opacityParam = graphic->getOpacity();
    if (opacityParam)
    {
      opacity = te::se::GetDouble(opacityParam);
    }

    //Mark
    const std::vector<te::se::Mark*>& vecMarks = graphic->getMarks();
    if (vecMarks.empty() == false)
    {
      //we are able to handle just one mark at the moment
      mark = vecMarks.at(0);

      const std::string* name = mark->getWellKnownName();
      if (name != nullptr)
      {
        std::size_t found = name->find("ttf://");
        if (found != std::string::npos)
        {
          isGlyphMark = true;
        }

        found = name->find("://");
        if (found == std::string::npos)
        {
          isWKMMark = true;
        }
      }
    }

    //Exernal Graphic
    const std::vector<te::se::ExternalGraphic*>& vecExternalGraphics = graphic->getExternalGraphics();
    if (vecExternalGraphics.empty() == false)
    {
      //we are able to handle just one external graphic at the moment
      externalGraphic = vecExternalGraphics.at(0);
      isExternalGraphic = true;
    }
  }
  // Size
  te::qt::widgets::AbstractPropertyManager::getInstance().m_doubleManager->setValue(m_sizeProperty, size);

  // Rotation
  te::qt::widgets::AbstractPropertyManager::getInstance().m_doubleManager->setValue(m_angleProperty, rotation);

  // Opacity
  int alpha = te::common::Round<double, int>(opacity * 255.);
  te::qt::widgets::AbstractPropertyManager::getInstance().m_intSliderManager->setValue(m_opacityProperty, alpha);

  //Mark
  if (isWKMMark == true)
  {
    m_markWellKnownProperty->setMark(mark);
    m_markGlyphProperty->setMark(nullptr);
    m_markExternalGraphicProperty->setExternalGraphic(nullptr);
  }
  if (isGlyphMark == true)
  {
    m_markWellKnownProperty->setMark(nullptr);
    m_markGlyphProperty->setMark(mark);
    m_markExternalGraphicProperty->setExternalGraphic(nullptr);
  }
  if (isExternalGraphic == true)
  {
    m_markWellKnownProperty->setMark(nullptr);
    m_markGlyphProperty->setMark(nullptr);
    m_markExternalGraphicProperty->setExternalGraphic(externalGraphic);
    m_markExternalGraphicProperty->setGraphicSize(te::common::Round<double, int>(size));
  }

  m_markWellKnownProperty->setChecked(isWKMMark);
  m_markGlyphProperty->setChecked(isGlyphMark);
  m_markExternalGraphicProperty->setChecked(isExternalGraphic);

  m_markWellKnownProperty->setExpandedStatus(isWKMMark, true);
  m_markGlyphProperty->setExpandedStatus(isGlyphMark, true);
  m_markExternalGraphicProperty->setExpandedStatus(isExternalGraphic, true);

  m_enableListeners = true;

  return true;
}

te::se::Graphic* te::qt::widgets::GraphicFillPropertyItem::getGraphic() const
{
  te::se::Graphic* newItem = new te::se::Graphic();

  //Size
  double size = te::qt::widgets::AbstractPropertyManager::getInstance().m_doubleManager->value(m_sizeProperty);
  newItem->setSize(new te::se::ParameterValue(Convert2StdString(QString::number(size, 'f', 2))));

  // Rotation
  double rotation = te::qt::widgets::AbstractPropertyManager::getInstance().m_doubleManager->value(m_angleProperty);
  newItem->setRotation(new te::se::ParameterValue(Convert2StdString(QString::number(rotation, 'f', 2))));

  // Opacity
  int alpha = te::qt::widgets::AbstractPropertyManager::getInstance().m_intSliderManager->value(m_opacityProperty);
  double seOpacity = alpha / 255.;
  newItem->setOpacity(new te::se::ParameterValue(Convert2StdString(QString::number(seOpacity, 'f', 2))));

  if (m_markWellKnownProperty->isChecked())
  {
    newItem->add(m_markWellKnownProperty->getMark());
  }
  if (m_markGlyphProperty->isChecked() == true)
  {
    newItem->add(m_markGlyphProperty->getMark());
  }
  if (m_markExternalGraphicProperty->isChecked() == true)
  {
    newItem->add(m_markExternalGraphicProperty->getExternalGraphic());
  }

  return newItem;
}

void te::qt::widgets::GraphicFillPropertyItem::valueChanged(QtProperty* p, double value)
{
  if (p == m_sizeProperty)
  {
    if (m_markExternalGraphicProperty != nullptr)
    {
      m_markExternalGraphicProperty->setGraphicSize(te::common::Round<double, int>(value));
    }
  }

  te::qt::widgets::CheckboxStylePropertyItem::valueChanged(p, value);
}