/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/widgets/se/LineSymbolizerProperty.cpp

  \brief A widget used to define the line symbolizer
         properties, such as:
           - basic stroke
*/

// TerraLib
#include "LineSymbolizerProperty.h"

#include "../../../se/GraphicStroke.h"
#include "../../../se/LineSymbolizer.h"
#include "../../../se/ParameterValue.h"
#include "../../../se/Stroke.h"
#include "../../../se/Utils.h"
#include "../propertybrowser/AbstractPropertyManager.h"
#include "../Utils.h"
#include "LinePerpendicularOffsetPropertyItem.h"
#include "StrokePropertyItem.h"

// Qt
#include <QGridLayout>
#include <QToolBox>

// QtPropertyBrowser
#include <QtPropertyBrowser/QtTreePropertyBrowser>
#include <QtPropertyBrowser/QtVariantPropertyManager>

// STL
#include <cassert>

te::qt::widgets::LineSymbolizerProperty::LineSymbolizerProperty(QWidget* parent)
  : AbstractSymbolizerProperty(parent)
  , StylePropertyItemListener()
{
  QGridLayout* layout = new QGridLayout(this);
  this->setLayout(layout);

  QToolBox* tb = new QToolBox(this);
  layout->addWidget(tb);

  QtTreePropertyBrowser* basicPropBrowser = new QtTreePropertyBrowser(this);
  basicPropBrowser->setIndentation(10);
  basicPropBrowser->setPropertiesWithoutValueMarked(true);
  basicPropBrowser->setRootIsDecorated(false);
  basicPropBrowser->setResizeMode(QtTreePropertyBrowser::ResizeToContents);

  layout->addWidget(basicPropBrowser);

  // Stroke configuration
  m_strokePropItem = new te::qt::widgets::StrokePropertyItem(basicPropBrowser, this);

  // perpendicular offset: a widget used to define the Perpendicular Offset between a line and marks.
  m_perpendicularOffset = new te::qt::widgets::LinePerpendicularOffsetPropertyItem(basicPropBrowser, this);

  setSymbolizer(nullptr);
}

te::qt::widgets::LineSymbolizerProperty::~LineSymbolizerProperty()
{
  delete m_strokePropItem;
  delete m_perpendicularOffset;
}

void te::qt::widgets::LineSymbolizerProperty::setSymbolizer(const te::se::Symbolizer* symb)
{
  //read the properties of the symbolizer
  const te::se::Stroke* stroke = nullptr;
  double perpendicularOffset = 0.0;
  if (symb != nullptr)
  {
    const te::se::LineSymbolizer* lineSymbolizer = dynamic_cast<const te::se::LineSymbolizer*>(symb);
    if (lineSymbolizer != nullptr)
    {
      stroke = lineSymbolizer->getStroke();
      if (lineSymbolizer->getPerpendicularOffset() != nullptr)
      {
        perpendicularOffset = GetDouble(lineSymbolizer->getPerpendicularOffset());
      }
    }
  }

  //stroke
  m_strokePropItem->setStroke(stroke);

  // perpendicular offset: a widget used to define the Perpendicular Offset between a line and marks.
  m_perpendicularOffset->setLinePerpendicularOffset(perpendicularOffset);
}

te::se::Symbolizer* te::qt::widgets::LineSymbolizerProperty::getSymbolizer() const
{
  //we always return a new instance
  te::se::LineSymbolizer* newItem = new te::se::LineSymbolizer();;

  newItem->setStroke(m_strokePropItem->getStroke());

  double perpendicularOffset = m_perpendicularOffset->getLinePerpendicularOffset();
  newItem->setPerpendicularOffset(new te::se::ParameterValue(Convert2StdString(QString::number(perpendicularOffset, 'f', 2))));

  return newItem;
}

void te::qt::widgets::LineSymbolizerProperty::handleValueChanged(QtProperty* /*qProperty*/, const QVariant& /*qVariant*/)
{
  emit symbolizerChanged();
}
