/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralib/qt/widgets/se/MarkExternalGraphicPropertyItem.cpp

   \brief A widget used to define the graphic mark external graphic se object.
 */

#include "MarkExternalGraphicPropertyItem.h"

 // TerraLib Includes
#include "../Utils.h"
#include "../propertybrowser/AbstractPropertyManager.h"
#include "../../../se/ExternalGraphic.h"
#include "../../../se/InlineContent.h"
#include "../../../xlink/SimpleLink.h"

#include <QBuffer>
#include <QImageReader>
#include <QFileDialog>
#include <QMessageBox>
#include <QMimeDatabase>

te::qt::widgets::MarkExternalGraphicPropertyItem::MarkExternalGraphicPropertyItem(QtProperty* parent, QtTreePropertyBrowser* propertyBrowser, StylePropertyItemListener* listener, QColor c)
  : te::qt::widgets::CheckboxStylePropertyItem(propertyBrowser, tr("External Graphic"), listener, c)
  , m_onLineResourceProperty(nullptr)
  , m_inLineContentProperty(nullptr)
  , m_inlineContent(nullptr)
  , m_format("")
  , m_size(0)
{
  //add property to parent
  QtProperty* itemProperty = this->getItemProperty();
  addSubProperty(parent, itemProperty, c);

  //Online resource
  m_onLineResourceProperty = te::qt::widgets::AbstractPropertyManager::getInstance().m_strDlgManager->addProperty(tr("OnLine Resource"));
  addListenerToProperty(m_onLineResourceProperty);

  //Inline content
  m_inLineContentProperty = te::qt::widgets::AbstractPropertyManager::getInstance().m_strDlgManager->addProperty(tr("InLine Content"));
  addListenerToProperty(m_inLineContentProperty);


  itemProperty->addSubProperty(m_onLineResourceProperty);
  itemProperty->addSubProperty(m_inLineContentProperty);

  connect(te::qt::widgets::AbstractPropertyManager::getInstance().m_dlgFactory, SIGNAL(internalDlg(QWidget*, QtProperty*)), this, SLOT(setDlg(QWidget*, QtProperty*)));

  setExternalGraphic(nullptr);
}

te::qt::widgets::MarkExternalGraphicPropertyItem::~MarkExternalGraphicPropertyItem()
{
  delete m_onLineResourceProperty;
  delete m_inLineContentProperty;
  delete m_inlineContent;
}

bool te::qt::widgets::MarkExternalGraphicPropertyItem::setExternalGraphic(const te::se::ExternalGraphic* externalGraphic)
{
  // Updating path
  const te::xl::SimpleLink* linkOnlineResource = nullptr;
  const te::se::InlineContent* inLineContent = nullptr;
  if (externalGraphic != nullptr)
  {
    linkOnlineResource = externalGraphic->getOnlineResource();
    inLineContent = externalGraphic->getInlineContent();
    
    m_format = externalGraphic->getFormat();
  }

  m_enableListeners = false;

  if (linkOnlineResource != nullptr)
  {
    QString path = Convert2QString(linkOnlineResource->getHref());

    te::qt::widgets::AbstractPropertyManager::getInstance().m_strDlgManager->setValue(m_onLineResourceProperty, path);
    te::qt::widgets::AbstractPropertyManager::getInstance().m_strDlgManager->setValue(m_inLineContentProperty, tr("NOT DEFINED"));
  }

  if (inLineContent != nullptr)
  {
    m_inlineContent = inLineContent->clone();

    te::qt::widgets::AbstractPropertyManager::getInstance().m_strDlgManager->setValue(m_onLineResourceProperty, "");
    te::qt::widgets::AbstractPropertyManager::getInstance().m_strDlgManager->setValue(m_inLineContentProperty, tr("DEFINED"));
  }

  m_enableListeners = true;

  return true;
}

te::se::ExternalGraphic* te::qt::widgets::MarkExternalGraphicPropertyItem::getExternalGraphic() const
{
  te::xl::SimpleLink* linkOnlineResource = nullptr;
  std::string format;

  QString onlineResourcePath = te::qt::widgets::AbstractPropertyManager::getInstance().m_strDlgManager->value(m_onLineResourceProperty);
  if (onlineResourcePath.isEmpty() == false)
  {
    // Updating external graphic path
    linkOnlineResource = new te::xl::SimpleLink;
    linkOnlineResource->setHref(Convert2StdString(onlineResourcePath));
  }

  te::se::ExternalGraphic* externalGraphic = new te::se::ExternalGraphic();
  externalGraphic->setOnlineResource(linkOnlineResource);
  
  if (m_inlineContent != nullptr)
  {
    externalGraphic->setInlineContent(m_inlineContent->clone());
  }

  externalGraphic->setFormat(m_format);

  return externalGraphic;
}

void te::qt::widgets::MarkExternalGraphicPropertyItem::setGraphicSize(int size)
{ 
  m_size = size; 
}

void te::qt::widgets::MarkExternalGraphicPropertyItem::setDlg(QWidget* parent, QtProperty* prop)
{
  if (prop == m_onLineResourceProperty)
  {
    connect(parent, SIGNAL(showDlg()), this, SLOT(showOnLineResourceDlg()));
  }
  else if (prop == m_inLineContentProperty)
  {
    connect(parent, SIGNAL(showDlg()), this, SLOT(showInLineContentDlg()));
  }
}

void te::qt::widgets::MarkExternalGraphicPropertyItem::showOnLineResourceDlg()
{
  // Bulding the filter string
  QString filter = tr("Images") + " ( ";
  QList<QByteArray> formats = QImageReader::supportedImageFormats();
  for (int i = 0; i < formats.size() - 1; ++i)
    filter += "*." + formats[i] + " ";
  filter += ")";

  QString path = QFileDialog::getOpenFileName(nullptr, tr("Select an Image File"), "", filter);
  if (path.isNull())
    return;

  delete m_inlineContent;
  m_inlineContent = nullptr;

  // Updating external graphic format
  QFileInfo fInfo(path);
  QMimeDatabase mimeTypeDb;
  QMimeType mimeType = mimeTypeDb.mimeTypeForFile(fInfo);

  m_format = Convert2StdString(mimeType.name());

  // Updating external graphic path
  te::qt::widgets::AbstractPropertyManager::getInstance().m_strDlgManager->setValue(m_onLineResourceProperty, path);
  te::qt::widgets::AbstractPropertyManager::getInstance().m_strDlgManager->setValue(m_inLineContentProperty, tr("NOT DEFINED"));
}

void te::qt::widgets::MarkExternalGraphicPropertyItem::showInLineContentDlg()
{
  // Bulding the filter string
  QString filter = tr("Images") + " ( ";
  QList<QByteArray> formats = QImageReader::supportedImageFormats();
  for (int i = 0; i < formats.size() - 1; ++i)
    filter += "*." + formats[i] + " ";
  filter += ")";

  QString path = QFileDialog::getOpenFileName(nullptr, tr("Select an Image File"), "", filter);
  if (path.isNull())
    return;

  delete m_inlineContent;
  m_inlineContent = nullptr;

  std::string strBase64;

  QFileInfo fInfo(path);
  if (Convert2StdString(fInfo.completeSuffix()) == "svg")
  {
    QFile input(path);
    input.open(QIODevice::ReadOnly);

    QByteArray ba = input.readAll();

    QByteArray baBase64 = ba.toBase64();

    strBase64 = baBase64.data();
  }
  else
  {
    QImage img;

    // Try load image
    if (!img.load(path))
    {
      QMessageBox::critical(nullptr, tr("Error"), tr("The selected image cannot be loaded."));
      return;
    }

    int size = 16; //default size
    if (m_size != 0)
    {
      size = m_size;
    }

    QByteArray ba;
    QBuffer bu(&ba);

    bu.open(QIODevice::WriteOnly);
    QImage scaledImg = img.scaled(size, size);

    std::string completeSuffix = Convert2StdString(fInfo.completeSuffix());
    scaledImg.save(&bu, completeSuffix.c_str());

    QByteArray baBase64 = ba.toBase64();

    strBase64 = baBase64.data();
  }

  m_inlineContent = new te::se::InlineContent(te::se::InlineContent::BASE64);
  m_inlineContent->setData(strBase64);

  // Updating external graphic format
  QMimeDatabase mimeTypeDb;
  QMimeType mimeType = mimeTypeDb.mimeTypeForFile(fInfo);

  m_format = Convert2StdString(mimeType.name());

  // Updating external graphic path
  te::qt::widgets::AbstractPropertyManager::getInstance().m_strDlgManager->setValue(m_onLineResourceProperty, "");
  te::qt::widgets::AbstractPropertyManager::getInstance().m_strDlgManager->setValue(m_inLineContentProperty, fInfo.absoluteFilePath());
}
