/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralib/qt/widgets/se/MarkGlyphPropertyItem.h

   \brief A widget used to define the graphic mark glyph se object.
 */

#include "MarkGlyphPropertyItem.h"

#include "BasicFillPropertyItem.h"

 // TerraLib Includes
#include "CharMapWidget.h"
#include "GlyphMarkRenderer.h"
#include "../propertybrowser/AbstractPropertyManager.h"
#include "../Utils.h"
#include "../../../se/Mark.h"

#include <QDialog>
#include <QGridLayout>
#include <QPushButton>
#include <QScrollArea>

te::qt::widgets::MarkGlyphPropertyItem::MarkGlyphPropertyItem(QtProperty* parent, QtTreePropertyBrowser* propertyBrowser, StylePropertyItemListener* listener, QColor c)
  : te::qt::widgets::CheckboxStylePropertyItem(propertyBrowser, tr("Glyph"), listener, c)
  , m_fontProperty(nullptr)
  , m_charProperty(nullptr)
  , m_basicFillPropertyItem(nullptr)
{
  //add property to parent
  QtProperty* itemProperty = this->getItemProperty();
  addSubProperty(parent, itemProperty, c);

  //Font
  m_fontProperty = te::qt::widgets::AbstractPropertyManager::getInstance().m_fontManager->addProperty(tr("Font"));
  addListenerToProperty(m_fontProperty);

  //Char
  m_charProperty = te::qt::widgets::AbstractPropertyManager::getInstance().m_strDlgManager->addProperty(tr("Char"));
  connect(te::qt::widgets::AbstractPropertyManager::getInstance().m_dlgFactory, SIGNAL(internalDlg(QWidget*, QtProperty*)), this, SLOT(setDlg(QWidget*, QtProperty*)));
  addListenerToProperty(m_charProperty);  

  itemProperty->addSubProperty(m_fontProperty);
  itemProperty->addSubProperty(m_charProperty);

  //Fill
  m_basicFillPropertyItem = new te::qt::widgets::BasicFillPropertyItem(itemProperty, propertyBrowser, listener, c);

  setMark(nullptr);
}

te::qt::widgets::MarkGlyphPropertyItem::~MarkGlyphPropertyItem()
{
  delete m_fontProperty;
  delete m_charProperty;
  delete m_basicFillPropertyItem;
}

bool te::qt::widgets::MarkGlyphPropertyItem::setMark(const te::se::Mark* mark)
{
  std::string markName = "";
  QFont qFont;
  qFont.setPointSize(TE_SE_DEFAULT_FONT_SIZE);
  int charCode = 0;

  const te::se::Fill* fill = nullptr;

  if (mark != nullptr)
  {
    try
    {
      if (mark->getWellKnownName() != nullptr)
      {
        markName = *mark->getWellKnownName();
      }

      if (markName.empty() == false)
      {
        QString qMarkName(Convert2QString(markName));
        std::size_t found = markName.find("ttf://");
        if (found != std::string::npos)
        {
          QString qFontName;
          QChar qCharCode;
          te::qt::widgets::GlyphMarkRenderer::decode(qMarkName, qFontName, qCharCode);

          qFont = QFont(qFontName);
          charCode = qCharCode.unicode();
        }
      }
    }
    catch (...)
    {
    }

    fill = mark->getFill();
  }

  m_enableListeners = false;

  QString strCharCode;
  strCharCode.setNum(charCode);

  //font
  te::qt::widgets::AbstractPropertyManager::getInstance().m_fontManager->setValue(m_fontProperty, qFont);

  //char code
  te::qt::widgets::AbstractPropertyManager::getInstance().m_strDlgManager->setValue(m_charProperty, strCharCode);

  //fill
  m_basicFillPropertyItem->setFill(fill);
  m_basicFillPropertyItem->setChecked(fill != nullptr);

  m_enableListeners = true;

  return true;
}

te::se::Mark* te::qt::widgets::MarkGlyphPropertyItem::getMark() const
{
  //font 
  QFont qFont = te::qt::widgets::AbstractPropertyManager::getInstance().m_fontManager->value(m_fontProperty);

  //char code
  QString qCharCodeStr = te::qt::widgets::AbstractPropertyManager::getInstance().m_strDlgManager->value(m_charProperty);

  //mark name
  QString qMarkName = te::qt::widgets::GlyphMarkRenderer::encode(qFont.family(), qCharCodeStr.toInt());
  std::string markName = Convert2StdString(qMarkName);

  te::se::Mark* newItem = new te::se::Mark();
  newItem->setWellKnownName(new std::string(markName));

  if (m_basicFillPropertyItem != nullptr)
  {
    newItem->setFill(m_basicFillPropertyItem->getFill());
  }
  

  return newItem;
}

void te::qt::widgets::MarkGlyphPropertyItem::setDlg(QWidget* parent, QtProperty* prop)
{
  if (prop == m_charProperty)
  {
    connect(parent, SIGNAL(showDlg()), this, SLOT(showCharMapDlg()));
  }
}

void te::qt::widgets::MarkGlyphPropertyItem::showCharMapDlg()
{
  QFont f = te::qt::widgets::AbstractPropertyManager::getInstance().m_fontManager->value(m_fontProperty);

  //build form
  QDialog d;

  QGridLayout* l = new QGridLayout(&d);
  l->setObjectName("mainLayout");
  QScrollArea* sa = new QScrollArea(&d);
  sa->setFixedSize(300, 200);
  QGridLayout* lsa = new QGridLayout(&d);
  lsa->setObjectName("scrollAreaLayout");
  sa->setLayout(lsa);

  te::qt::widgets::CharMapWidget* cmw = new te::qt::widgets::CharMapWidget(sa);
  lsa->addWidget(cmw);
  l->addWidget(sa, 0, 0, 1, 2);
  connect(cmw, SIGNAL(charSelected(const unsigned int&)), this, SLOT(charMapSelected(const unsigned int&)));
  cmw->setFont(f);
  sa->setWidget(cmw);

  QSpacerItem* spacer = new QSpacerItem(16, 16, QSizePolicy::MinimumExpanding, QSizePolicy::Minimum);
  l->addItem(spacer, 1, 0);
  QPushButton* pb = new QPushButton(tr("Close"), &d);
  connect(pb, SIGNAL(clicked()), &d, SLOT(close()));
  l->addWidget(pb, 1, 1);
  l->setSizeConstraint(QLayout::SetFixedSize);

  d.setLayout(l);
  d.setWindowTitle(tr("Char Map"));
  d.setWindowFlags(Qt::Tool);
  d.exec();
}

void te::qt::widgets::MarkGlyphPropertyItem::charMapSelected(const unsigned int& charCode)
{
  QString str;
  str.setNum(charCode);

  te::qt::widgets::AbstractPropertyManager::getInstance().m_strDlgManager->setValue(m_charProperty, str);
}
