/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
\file terralib/qt/widgets/se/PointPlacementPropertyItem.cpp

\brief A widget used to define the PointPlacement se object.
*/


#include "../../../common/MathUtils.h"
#include "../../../common/StringUtils.h"
#include "../../../se/AnchorPoint.h"
#include "../../../se/Displacement.h"
#include "../../../se/SvgParameter.h"
#include "../../../se/PointPlacement.h"
#include "../../../se/Utils.h"
#include "../propertybrowser/AbstractPropertyManager.h"
#include "../Utils.h"
#include "PointPlacementPropertyItem.h"


te::qt::widgets::PointPlacementPropertyItem::PointPlacementPropertyItem(QtTreePropertyBrowser* pb, StylePropertyItemListener* listener, QColor c)
  : te::qt::widgets::CheckboxStylePropertyItem(pb, tr("Point Placement"), listener, c)
  , m_anchorProperty(nullptr)
  , m_displacementProperty(nullptr)
  , m_rotationProperty(nullptr)
{
  //build property browser basic fill
  QtProperty* itemProperty = this->getItemProperty();
  addProperty(itemProperty, QColor(241, 165, 61));

  //anchor
  m_anchorProperty = te::qt::widgets::AbstractPropertyManager::getInstance().m_sizeManager->addProperty(tr("Anchor"));
  te::qt::widgets::AbstractPropertyManager::getInstance().m_sizeManager->setMinimum(m_anchorProperty, QSize(0, 0));
  te::qt::widgets::AbstractPropertyManager::getInstance().m_sizeManager->setMaximum(m_anchorProperty, QSize(100, 100));
  addListenerToProperty(m_anchorProperty);

  //displacement
  m_displacementProperty = te::qt::widgets::AbstractPropertyManager::getInstance().m_sizeManager->addProperty(tr("Displacement"));
  te::qt::widgets::AbstractPropertyManager::getInstance().m_sizeManager->setMinimum(m_displacementProperty, QSize(-100, -100));
  te::qt::widgets::AbstractPropertyManager::getInstance().m_sizeManager->setMaximum(m_displacementProperty, QSize(100, 100));
  addListenerToProperty(m_displacementProperty);

  //rotation
  m_rotationProperty = te::qt::widgets::AbstractPropertyManager::getInstance().m_intSliderManager->addProperty(tr("Rotation"));
  te::qt::widgets::AbstractPropertyManager::getInstance().m_intSliderManager->setMinimum(m_rotationProperty, -180);
  te::qt::widgets::AbstractPropertyManager::getInstance().m_intSliderManager->setMaximum(m_rotationProperty, 180);
  addListenerToProperty(m_rotationProperty);

  itemProperty->addSubProperty(m_anchorProperty);
  itemProperty->addSubProperty(m_displacementProperty);
  itemProperty->addSubProperty(m_rotationProperty);

  setPointPlacement(nullptr);
}

te::qt::widgets::PointPlacementPropertyItem::~PointPlacementPropertyItem()
{
  delete m_anchorProperty;
  delete m_displacementProperty;
  delete m_rotationProperty;
}

void te::qt::widgets::PointPlacementPropertyItem::setPointPlacement(const te::se::PointPlacement* pointPlacenement)
{
  const te::se::AnchorPoint* anchorPoint = nullptr;
  const te::se::Displacement* displacement = nullptr;
  double rotation = 0.0;

  if (pointPlacenement != nullptr)
  {
    anchorPoint = pointPlacenement->getAnchorPoint();
    displacement = pointPlacenement->getDisplacement();
    const te::se::ParameterValue* rotationParameter = pointPlacenement->getRotation();
    if (rotationParameter != nullptr)
    {
      rotation = GetDouble(rotationParameter);
    }
  }

  m_enableListeners = false;

  //Anchor point
  QSize anchorSize(50, 50);
  if(anchorPoint != nullptr)
  {
    int x = static_cast<int>(te::se::GetDouble(anchorPoint->getAnchorPointX()) * 100.);
    int y = static_cast<int>(te::se::GetDouble(anchorPoint->getAnchorPointY()) * 100.);

    anchorSize.setWidth(x);
    anchorSize.setHeight(y);
  }
  te::qt::widgets::AbstractPropertyManager::getInstance().m_sizeManager->setValue(m_anchorProperty, anchorSize);

  //Displacement
  QSize dispSize(0, 0);
  if (displacement != nullptr)
  {
    dispSize.setWidth(te::se::GetInt(displacement->getDisplacementX()));
    dispSize.setHeight(te::se::GetInt(displacement->getDisplacementY()));
  }
  te::qt::widgets::AbstractPropertyManager::getInstance().m_sizeManager->setValue(m_displacementProperty, dispSize);

  //Rotation
  int rotationValue = te::common::Round<double, int>(rotation);
  te::qt::widgets::AbstractPropertyManager::getInstance().m_intSliderManager->setValue(m_rotationProperty, rotation);

  m_enableListeners = true;
}

te::se::PointPlacement* te::qt::widgets::PointPlacementPropertyItem::getPointPlacement() const
{
  //Anchor Point
  QSize qAnchorPoint = te::qt::widgets::AbstractPropertyManager::getInstance().m_sizeManager->value(m_anchorProperty);
  double x = qAnchorPoint.width() / 100.;
  double y = qAnchorPoint.height() / 100.;

  te::se::AnchorPoint* anchorPoint = new te::se::AnchorPoint();
  anchorPoint->setAnchorPointX(new te::se::ParameterValue(Convert2StdString(QString::number(x))));
  anchorPoint->setAnchorPointY(new te::se::ParameterValue(Convert2StdString(QString::number(y))));

  //Displacement
  QSize qDisplacement = te::qt::widgets::AbstractPropertyManager::getInstance().m_sizeManager->value(m_displacementProperty);

  te::se::Displacement* displacement = new te::se::Displacement();
  displacement->setDisplacementX(new te::se::ParameterValue(Convert2StdString(QString::number(qDisplacement.width()))));
  displacement->setDisplacementY(new te::se::ParameterValue(Convert2StdString(QString::number(qDisplacement.height()))));

  int rotation = te::qt::widgets::AbstractPropertyManager::getInstance().m_intSliderManager->value(m_rotationProperty);

  te::se::PointPlacement* newItem = new te::se::PointPlacement();
  newItem->setAnchorPoint(anchorPoint);
  newItem->setDisplacement(displacement);
  newItem->setRotation(new te::se::ParameterValue(Convert2StdString(QString::number(rotation, 2))));

  return newItem;
}
