/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/widgets/se/PolygonSymbolizerProperty.cpp

  \brief A widget used to define the polygon symbolizer
         properties, such as:
           - basic fill
           - basic stroke
*/

// Terralib
#include "../../../se/Fill.h"
#include "../../../se/GraphicStroke.h"
#include "../../../se/ParameterValue.h"
#include "../../../se/PolygonSymbolizer.h"
#include "../../../se/Stroke.h"
#include "../Utils.h"
#include "../propertybrowser/AbstractPropertyManager.h"
#include "BasicFillPropertyItem.h"
#include "BasicStrokePropertyItem.h"
#include "LinePerpendicularOffsetPropertyItem.h"
#include "LineInitialGapPropertyItem.h"
#include "LineGapPropertyItem.h"
#include "PolygonSymbolizerProperty.h"
#include "../../../se/Utils.h"

#include "FillPropertyItem.h"
#include "StrokePropertyItem.h"

// Qt
#include <QGridLayout>
#include <QToolBox>

// QtPropertyBrowser
#include <QtPropertyBrowser/QtTreePropertyBrowser>
#include <QtPropertyBrowser/QtVariantPropertyManager>

// STL
#include <cassert>

te::qt::widgets::PolygonSymbolizerProperty::PolygonSymbolizerProperty(QWidget* parent)
  : AbstractSymbolizerProperty(parent)
  , StylePropertyItemListener()
{
  QGridLayout* layout = new QGridLayout(this);
  this->setLayout(layout);

  QToolBox* tb = new QToolBox(this);
  layout->addWidget(tb);

  QtTreePropertyBrowser* basicPropBrowser = new QtTreePropertyBrowser(this);
  basicPropBrowser->setIndentation(10);
  basicPropBrowser->setPropertiesWithoutValueMarked(true);
  basicPropBrowser->setRootIsDecorated(false);
  basicPropBrowser->setResizeMode(QtTreePropertyBrowser::ResizeToContents);

  layout->addWidget(basicPropBrowser);

  //Fill property
  m_fillPropItem = new te::qt::widgets::FillPropertyItem(basicPropBrowser, this);

  //Stroke property
  m_strokePropItem = new te::qt::widgets::StrokePropertyItem(basicPropBrowser, this);
  
  // perpendicular offset: a widget used to define the Perpendicular Offset between a line and marks.
  m_perpendicularOffset = new te::qt::widgets::LinePerpendicularOffsetPropertyItem(basicPropBrowser, this);

  setSymbolizer(nullptr);
}

te::qt::widgets::PolygonSymbolizerProperty::~PolygonSymbolizerProperty()
{
  delete m_fillPropItem;
  delete m_strokePropItem;
  delete m_perpendicularOffset;
}

void te::qt::widgets::PolygonSymbolizerProperty::setSymbolizer(const te::se::Symbolizer* symb)
{
  //read the properties of the symbolizer
  const te::se::Fill* fill = nullptr;
  const te::se::Stroke* stroke = nullptr;
  double perpendicularOffset = 0.0;
  if (symb != nullptr)
  {
    const te::se::PolygonSymbolizer* polygonSymbolizer = dynamic_cast<const te::se::PolygonSymbolizer*>(symb);
    if(polygonSymbolizer != nullptr)
    {   
      fill = polygonSymbolizer->getFill();
      stroke = polygonSymbolizer->getStroke();
      if (polygonSymbolizer->getPerpendicularOffset() != nullptr)
      {
        perpendicularOffset = GetDouble(polygonSymbolizer->getPerpendicularOffset());
      }
    }
  }

  //fill
  m_fillPropItem->setFill(fill);

  //stroke
  m_strokePropItem->setStroke(stroke);

  // perpendicular offset: a widget used to define the Perpendicular Offset between a line and marks.
  m_perpendicularOffset->setLinePerpendicularOffset(perpendicularOffset);
}

te::se::Symbolizer* te::qt::widgets::PolygonSymbolizerProperty::getSymbolizer() const
{
  //we always return a new instance
  te::se::PolygonSymbolizer* newItem = new te::se::PolygonSymbolizer();;

  newItem->setFill(m_fillPropItem->getFill());
  newItem->setStroke(m_strokePropItem->getStroke());

  double perpendicularOffset = m_perpendicularOffset->getLinePerpendicularOffset();
  newItem->setPerpendicularOffset(new te::se::ParameterValue(Convert2StdString(QString::number(perpendicularOffset, 'f', 2))));

  return newItem;
}

void te::qt::widgets::PolygonSymbolizerProperty::handleValueChanged(QtProperty* qProperty, const QVariant& qVariant)
{
  emit symbolizerChanged();
}
