/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/widgets/se/StyleControllerWidget.cpp

  \brief A widget used to controll the style se element.
*/

// TerraLib
#include "StyleControllerWidget.h"
#include "../../../core/filesystem/FileSystem.h"
#include "../../../dataaccess/utils/Utils.h"
#include "../../../geometry/GeometryProperty.h"
#include "../../../fe.h"
#include "../../../maptools/FilterEncoder.h"
#include "../../../maptools/QueryEncoder.h"
#include "../../../se/FeatureTypeStyle.h"
#include "../../../se/Font.h"
#include "../../../se/Rule.h"
#include "../../../se/TextSymbolizer.h"
#include "../../../se/Utils.h"
#include "../../../se/Style.h"
#include "../../../se/Symbolizer.h"
#include "../../../se/serialization/xml/Style.h"
#include "../../../xml/AbstractWriter.h"
#include "../../../xml/AbstractWriterFactory.h"
#include "../../../xml/Reader.h"
#include "../../../xml/ReaderFactory.h"
#include "../../widgets/query/QueryBuilderDialog.h"
#include "../../widgets/Utils.h"
#include "CreateRuleDialog.h"
#include "StyleExplorer.h"
#include "Symbol.h"
#include "SymbolPreviewWidget.h"
#include "SymbolSelectorDialog.h"

// TerraLib - UIC
#include "ui/ui_StyleControllerWidgetForm.h"

// Qt
#include <QMessageBox>
#include <QFileDialog>
#include <QFile>
#include <QDir>
#include <QString>

// STL
#include <cassert>
#include <fstream>

// Boost
#include <boost/filesystem/operations.hpp>

te::qt::widgets::StyleControllerWidget::StyleControllerWidget(QWidget* parent, Qt::WindowFlags f)
  : QWidget(parent, f),
    m_ui(new Ui::StyleControllerWidgetForm),
    m_currentStyle(nullptr),
    m_currentLayer(nullptr),
    m_isSelectionStyle(false)
{
  m_ui->setupUi(this);

  //build explorer
  m_explorer = new te::qt::widgets::StyleExplorer(m_ui->m_styleFrame);

  QGridLayout* explorerLayout = new QGridLayout(m_ui->m_styleFrame);
  explorerLayout->addWidget(m_explorer);
  explorerLayout->setContentsMargins(0, 0, 0, 0);

  // Signals & slots
  connect(m_ui->m_iconSizeSlider, SIGNAL(valueChanged(int)), this, SLOT(changeLegendIconSize(int)));
  connect(m_ui->m_addRuleToolButton, SIGNAL(clicked()), this, SLOT(onAddRuleClicked()));
  connect(m_ui->m_editRuleToolButton, SIGNAL(clicked()), this, SLOT(onEditRuleClicked()));
  connect(m_ui->m_removeRuleToolButton, SIGNAL(clicked()), this, SLOT(onRemoveRuleClicked()));
  connect(m_ui->m_addTextSymbToolButton, SIGNAL(clicked()), this, SLOT(onAddTextSymbolizerClicked()));
  connect(m_ui->m_addSymbToolButton, SIGNAL(clicked()), this, SLOT(onAddSymbolizerClicked()));
  connect(m_ui->m_removeSymbToolButton, SIGNAL(clicked()), this, SLOT(onRemoveSymbolizerClicked()));
  connect(m_ui->m_upSymbToolButton, SIGNAL(clicked()), this, SLOT(onUpSymbolizerClicked()));
  connect(m_ui->m_downSymbToolButton, SIGNAL(clicked()), this, SLOT(onDownSymbolizerClicked()));
  connect(m_ui->m_libManagerToolButton, SIGNAL(clicked()), this, SLOT(onLibraryManagerClicked()));
  connect(m_ui->m_exportStyleToolButton, SIGNAL(clicked()), this, SLOT(onExportClicked()));
  connect(m_ui->m_importStyleToolButton, SIGNAL(clicked()), this, SLOT(onImportClicked()));
  connect(m_ui->m_mapRefreshToolButton, SIGNAL(clicked()), this, SLOT(onMapRefreshClicked()));

  m_ui->m_layerNameLabel->setText(tr("No Layer selected!"));

  m_ui->m_helpPushButton->setNameSpace("dpi.inpe.br.plugins");
  m_ui->m_helpPushButton->setPageReference("widgets/style/explorer.html");
}

te::qt::widgets::StyleControllerWidget::~StyleControllerWidget() = default;

void te::qt::widgets::StyleControllerWidget::setStyle(te::se::Style* style, std::string dataSourceId, std::string dataSetName)
{
  m_currentStyle = style;
  m_currentLayer = nullptr;

  m_explorer->setStyle(style);

  if (dataSourceId.empty() == false && dataSetName.empty() == false)
  {
    m_dataSourceId = dataSourceId;

    te::da::DataSourcePtr ds = te::da::GetDataSource(dataSourceId);

    if (!ds->isOpened())
      ds->open();

    m_dsType = ds->getDataSetType(dataSetName);
  }
}

void te::qt::widgets::StyleControllerWidget::setLayer(te::map::AbstractLayer* layer, te::se::Style* style, bool isSelectionStyle)
{
  m_currentLayer = layer;

  m_currentStyle = style;

  m_explorer->setStyle(style);

  m_isSelectionStyle = isSelectionStyle;

  if (layer)
  {
    if (isSelectionStyle)
    {
      m_ui->m_layerNameLabel->setText(tr("Selection Style: ") + te::qt::widgets::Convert2QString(m_currentLayer->getTitle()));
    }
    else
    {
      m_ui->m_layerNameLabel->setText(tr("Visualization Style: ") + te::qt::widgets::Convert2QString(m_currentLayer->getTitle()));
    }
  }

  //set info
  m_dataSourceId = m_currentLayer->getDataSourceId();
    
  m_dsType = m_currentLayer->getSchema();
}

te::qt::widgets::StyleExplorer* te::qt::widgets::StyleControllerWidget::getStyleExplorer() const
{
  return m_explorer;
}

void te::qt::widgets::StyleControllerWidget::onAddRuleClicked()
{
  if (m_dataSourceId.empty() || !m_dsType)
  {
    QMessageBox::information(this, tr("Style"), tr("Select a layer first."));
    return;
  }

  if (!m_currentStyle)
  {
    QMessageBox::information(this, tr("Style"), tr("Style not defined"));
    return;
  }

  te::qt::widgets::CreateRuleDialog dlg;

  dlg.setDataInformation(m_dataSourceId, m_dsType->getTitle());

  if (dlg.exec() == QDialog::Accepted)
  {
    te::se::Rule* rule = dlg.getRule();

    m_currentStyle->push_back(rule);

    m_explorer->updateStyleTree();

    emit updateLayer();
  }
}

void te::qt::widgets::StyleControllerWidget::onEditRuleClicked()
{
  if (m_dataSourceId.empty() || !m_dsType)
  {
    QMessageBox::information(this, tr("Style"), tr("Select a layer first."));
    return;
  }

  if (!m_currentStyle)
  {
    QMessageBox::information(this, tr("Style"), tr("Style not defined"));
    return;
  }

  te::se::Rule* rule = m_explorer->getCurrentRule();

  if (rule == nullptr)
  {
    QMessageBox::information(this, tr("Style"), tr("Select a rule first."));
    return;
  }

  te::qt::widgets::CreateRuleDialog dlg;

  dlg.setDataInformation(m_dataSourceId, m_dsType->getTitle());

  dlg.setRule(rule->clone());

  if (dlg.exec() == QDialog::Accepted)
  {
    std::unique_ptr<te::se::Rule> r(dlg.getRule());

    rule->setName(new std::string(*r->getName()));
    rule->setFilter(r->getFilter()->clone());

    m_explorer->updateStyleTree();

    emit updateLayer();
  }
}

void te::qt::widgets::StyleControllerWidget::onRemoveRuleClicked()
{
  if (!m_currentStyle)
  {
    QMessageBox::information(this, tr("Style"), tr("Style not defined"));
    return;
  }

  if (m_currentStyle->getRules().size() == 1)
  {
    QMessageBox::information(this, tr("Style"), tr("It is not possible to leave a style without rules."));
    return;
  }

  te::se::Rule* rule = m_explorer->getCurrentRule();

  if (rule == nullptr)
  {
    QMessageBox::information(this, tr("Style"), tr("Select a rule first."));
    return;
  }

  for (std::size_t t = 0; t < m_currentStyle->getRules().size(); ++t)
  {
    if (m_currentStyle->getRule(t) == rule)
    {
      m_currentStyle->removeRule(t);
      break;
    }
  }

  m_explorer->updateStyleTree();

  emit updateLayer();
}

void te::qt::widgets::StyleControllerWidget::onAddTextSymbolizerClicked()
{
  te::se::Rule* rule = m_explorer->getCurrentRule();

  if (rule == nullptr)
  {
    QMessageBox::information(this, tr("Style"), tr("Select a rule first."));
    return;
  }

  //creates new text symbolizer representation
  te::se::Fill* fill = te::se::CreateFill("#000000", "1.0");
  te::se::Font* font = te::se::CreateFont("Arial", "8", te::se::Font::StyleNormal, te::se::Font::WeightNormal);
  te::se::Symbolizer* symb = te::se::CreateTextSymbolizer("", fill, font);

  rule->push_back(symb); 

  m_explorer->updateStyleTree();
}

void te::qt::widgets::StyleControllerWidget::onAddSymbolizerClicked()
{
  te::se::Rule* rule = m_explorer->getCurrentRule();

  if(rule == nullptr)
  {
    QMessageBox::information(this, tr("Style"), tr("Select a rule first."));
    return;
  }

  te::se::Symbolizer* symb = m_explorer->getCurrentSymbolizer();
 
  if(symb)
    rule->push_back(symb->clone());
  else
  {
    assert(!rule->getSymbolizers().empty());

    const te::se::Symbolizer* symbolizer = rule->getSymbolizer(0);
    assert(symbolizer);

    rule->push_back(symbolizer->clone());
  }

  m_explorer->updateStyleTree();
}

void te::qt::widgets::StyleControllerWidget::onRemoveSymbolizerClicked()
{
   te::se::Symbolizer* symb = m_explorer->getCurrentSymbolizer();

  if(symb == nullptr)
  {
    QMessageBox::information(this, tr("Style"), tr("Select a symbol first."));
    return;
  }

  te::se::Rule* rule = m_explorer->getCurrentRule();
  assert(rule);

  if(rule->getSymbolizers().size() == 1)
  {
    QMessageBox::information(this, tr("Style"), tr("The rule must have at least one symbol."));
    return;
  }

  for(std::size_t i = 0; i < rule->getSymbolizers().size(); ++i)
  {
    if(rule->getSymbolizer(i) == symb)
    {
      rule->removeSymbolizer(i);
      break;
    }
  }

  m_explorer->updateStyleTree();
}

void te::qt::widgets::StyleControllerWidget::onUpSymbolizerClicked()
{
  te::se::Symbolizer* symb = m_explorer->getCurrentSymbolizer();

  if(symb == nullptr)
  {
    QMessageBox::information(this, tr("Style"), tr("Select a symbol first."));
    return;
  }

  m_explorer->goUpSymbolizer();
}

void te::qt::widgets::StyleControllerWidget::onDownSymbolizerClicked()
{
  te::se::Symbolizer* symb = m_explorer->getCurrentSymbolizer();

  if(symb == nullptr)
  {
    QMessageBox::information(this, tr("Style"), tr("Select a symbol first."));
    return;
  }

  m_explorer->goDownSymbolizer();
}

void te::qt::widgets::StyleControllerWidget::onLibraryManagerClicked()
{
  te::se::Rule* rule = m_explorer->getCurrentRule();

  if(!rule)
  {
    QMessageBox::information(this, tr("Style"), tr("Select a rule first."));
    return;
  }

  Symbol* symbol = nullptr;

  std::vector<std::string> propNames;

  for (std::size_t t = 0; t < m_dsType->getProperties().size(); ++t)
  {
    if (m_dsType->getProperty(t)->getType() != te::dt::GEOMETRY_TYPE)
      propNames.push_back(m_dsType->getProperty(t)->getName());
  }
  
  SymbolSelectorDialog dlg(propNames, this);
    
  if (dlg.exec() == QDialog::Accepted)
  {
    symbol = dlg.getSymbol();
  }

  if (symbol == nullptr)
  {
    return;
  }

  std::vector<te::se::Symbolizer*> symbolizers;
  for(std::size_t i = 0; i < symbol->getSymbolizersCount(); ++i)
    symbolizers.push_back(symbol->getSymbolizer(i)->clone());

  rule->setSymbolizers(symbolizers);

  m_explorer->updateStyleTree();
}

void te::qt::widgets::StyleControllerWidget::onExportClicked()
{  
  if (!m_currentStyle)
  {
    QMessageBox::information(this, tr("Style"), tr("Select a layer first."));
    return;
  }

  QString styleFile = QFileDialog::getSaveFileName(this, tr("Save File"), te::qt::widgets::GetFilePathFromSettings("style"), "Style file (*.sld)");

  if (styleFile.isEmpty())
    return;

  QFileInfo info(styleFile);

  te::qt::widgets::AddFilePathToSettings(info.absolutePath(), "style");

  if (info.suffix().isEmpty())
    styleFile.append(".sld");

  writeStyle(m_currentStyle, styleFile.toUtf8().data());
}

void te::qt::widgets::StyleControllerWidget::onImportClicked()
{
  QString styleFile = QFileDialog::getOpenFileName(this, tr("Select a style file"), te::qt::widgets::GetFilePathFromSettings("style"), "Style file (*.sld)");

  QFileInfo info(styleFile);

  te::qt::widgets::AddFilePathToSettings(info.absolutePath(), "style");

  if (!styleFile.isEmpty())
    readStyle(styleFile.toUtf8().data());
  else
    return;
}

void te::qt::widgets::StyleControllerWidget::onMapRefreshClicked()
{
  emit mapRefresh();
}

void te::qt::widgets::StyleControllerWidget::changeLegendIconSize(int size)
{
  m_explorer->setLegendIconSize(size);
}

void te::qt::widgets::StyleControllerWidget::onVisualStyleChecked(bool state)
{
  if (state && m_currentLayer && m_currentLayer->getStyle())
  {
    m_explorer->setStyle(m_currentLayer->getStyle());
    m_currentStyle = m_currentLayer->getStyle();
  }
}

void te::qt::widgets::StyleControllerWidget::onSelectionStyleChecked(bool state)
{
  if (state && m_currentLayer && m_currentLayer->getSelectionStyle())
  {
    m_explorer->setStyle(m_currentLayer->getSelectionStyle());
    m_currentStyle = m_currentLayer->getSelectionStyle();
  }
}

void te::qt::widgets::StyleControllerWidget::writeStyle(const te::se::Style* style, const std::string& path)
{
  boost::filesystem::wpath file = te::core::FileSystem::absolutePath(path);

  std::string layerName;

  if (m_dsType)
  {
    layerName = m_dsType->getTitle();
  }

  {
    std::unique_ptr<te::xml::AbstractWriter> writer(te::xml::AbstractWriterFactory::make());

    writer->setURI(file.string());
    writer->writeStartDocument("UTF-8", "no");

    writer->writeStartElement("StyledLayerDescriptor");

    writer->writeAttribute("xmlns", "http://www.opengis.net/sld");
    writer->writeAttribute("xmlns:ogc", "http://www.opengis.net/ogc");
    writer->writeAttribute("xmlns:se", "http://www.opengis.net/se");
    writer->writeAttribute("xmlns:xlink", "http://www.w3.org/1999/xlink");
    writer->writeAttribute("xmlns:xsi", "http://www.w3.org/2001/XMLSchema-instance");
    writer->writeAttribute("xsi:schemaLocation", "http://www.opengis.net/sld StyledLayerDescriptor.xsd");

    writer->writeAttribute("version", style->getVersion());

    writer->writeStartElement("NamedLayer");
    writer->writeElement("se:Name", layerName);
    writer->writeStartElement("UserStyle");
    writer->writeElement("se:Name", layerName);

    te::se::serialize::Style::getInstance().write(style, *writer.get());

    writer->writeEndElement("UserStyle");
    writer->writeEndElement("NamedLayer");

    writer->writeEndElement("StyledLayerDescriptor");
    writer->writeToFile();
  }
}

void te::qt::widgets::StyleControllerWidget::readStyle(const std::string& path)
{
  if(m_dsType.get())
  {
    if(m_dsType->hasGeom())
        readVectorStyle(path);
    else if(m_dsType->hasRaster())
        readRasterStyle(path);
    else
    {
      QMessageBox::information(this, tr("Style"), tr("The dataset type is invalid."));
      return;
    }
  }
  else
  {
    QMessageBox::information(this, tr("Style"), tr("Dataset type error."));
    return;
  }
}

void te::qt::widgets::StyleControllerWidget::readVectorStyle(const std::string& path)
{
  std::unique_ptr<te::se::Style> style;

  boost::filesystem::wpath file = te::core::FileSystem::absolutePath(path);
  {
    std::unique_ptr<te::xml::Reader> reader(te::xml::ReaderFactory::make());
    reader->setValidationScheme(false);

    reader->read(file.string());
    reader->next();

    if (!checkSymbolizer(file.string()))
    {
      QMessageBox::warning(this, tr("Style Explorer"), tr("The selected theme is not compatible with the current data."));
      return;
    }

    if ((reader->getNodeType() == te::xml::START_ELEMENT) &&
      (reader->getElementLocalName() == "StyledLayerDescriptor"))
      reader->next();

    if ((reader->getNodeType() == te::xml::START_ELEMENT) &&
      (reader->getElementLocalName() == "NamedLayer"))
      reader->next();

    if ((reader->getNodeType() == te::xml::START_ELEMENT) &&
      (reader->getElementLocalName() == "Name"))
    {
      reader->next();

      if(reader->getNodeType() == te::xml::VALUE)
        reader->next();

      if(reader->getNodeType() == te::xml::END_ELEMENT)
        reader->next();
    }

    if ((reader->getNodeType() == te::xml::START_ELEMENT) &&
      (reader->getElementLocalName() == "UserStyle"))
      reader->next();

    if ((reader->getNodeType() == te::xml::START_ELEMENT) &&
      (reader->getElementLocalName() == "Name"))
    {
      reader->next();

      if(reader->getNodeType() == te::xml::VALUE)
        reader->next();

      if(reader->getNodeType() == te::xml::END_ELEMENT)
        reader->next();
    }

    if ((reader->getNodeType() == te::xml::START_ELEMENT) &&
      (reader->getElementLocalName() == "FeatureTypeStyle"))
    {
      if (reader->getNodeType() == te::xml::START_ELEMENT)
        style.reset(te::se::serialize::Style::getInstance().read(*reader.get()));

      m_explorer->importStyle(style.release(), !m_isSelectionStyle);
    }
    else
    {
      QMessageBox::warning(this, tr("Style Explorer"), tr("The style contained unrecognized elements and could not be read."));
      return;
    }
  }
}

void te::qt::widgets::StyleControllerWidget::readRasterStyle(const std::string& path)
{
  std::unique_ptr<te::se::Style> style;

  boost::filesystem::wpath file = te::core::FileSystem::absolutePath(path);
  {
    std::unique_ptr<te::xml::Reader> reader(te::xml::ReaderFactory::make());
    reader->setValidationScheme(false);

    reader->read(file.string());
    reader->next();

    if (!checkSymbolizer(file.string()))
    {
      QMessageBox::warning(this, tr("Style Explorer"), tr("The selected theme is not compatible with the current data."));
      return;
    }

    if ((reader->getNodeType() == te::xml::START_ELEMENT) &&
      (reader->getElementLocalName() == "StyledLayerDescriptor"))
      reader->next();

    if ((reader->getNodeType() == te::xml::START_ELEMENT) &&
      (reader->getElementLocalName() == "NamedLayer"))
      reader->next();

    if ((reader->getNodeType() == te::xml::START_ELEMENT) &&
      (reader->getElementLocalName() == "Name"))
    {
      reader->next();

      if(reader->getNodeType() == te::xml::VALUE)
        reader->next();

      if(reader->getNodeType() == te::xml::END_ELEMENT)
        reader->next();
    }

    if ((reader->getNodeType() == te::xml::START_ELEMENT) &&
      (reader->getElementLocalName() == "UserStyle"))
      reader->next();

    if ((reader->getNodeType() == te::xml::START_ELEMENT) &&
      (reader->getElementLocalName() == "Name"))
    {
      reader->next();

      if(reader->getNodeType() == te::xml::VALUE)
        reader->next();

      if(reader->getNodeType() == te::xml::END_ELEMENT)
        reader->next();
    }

    if ((reader->getNodeType() == te::xml::START_ELEMENT) &&
      (reader->getElementLocalName() == "CoverageStyle"))
    {
      if (reader->getNodeType() == te::xml::START_ELEMENT)
        style.reset(te::se::serialize::Style::getInstance().read(*reader.get()));

      m_explorer->importStyle(style.release(), !m_isSelectionStyle);
    }
    else
    {
      QMessageBox::warning(this, tr("Style Explorer"), tr("The style contained unrecognized elements and could not be read."));
      return;
    }
  }
}

void te::qt::widgets::StyleControllerWidget::clear()
{
  m_currentStyle = nullptr;
  m_currentLayer = nullptr;

  m_explorer->setStyle(nullptr);
}

bool te::qt::widgets::StyleControllerWidget::checkSymbolizer(std::string file)
{
  std::string sldFile;
  std::ifstream t(file);
  std::stringstream buffer;
  buffer << t.rdbuf();

  sldFile = buffer.str();

  std::size_t found = sldFile.find(m_explorer->getCurrentSymbolizer()->getType());

  if (found != std::string::npos)
    return true;
  else
    return false;
}
