/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled
   applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/widgets/se/SymbolSelectorDialog.cpp

  \brief A dialog used to select a specific symbol.
*/

// TerraLib
#include "../../../common/Exception.h"
#include "../../../common/StringUtils.h"
#include "Symbol.h"
#include "Utils.h"
#include "../../../dataaccess/dataset/DataSet.h"
#include "../../../dataaccess/dataset/DataSetType.h"
#include "../../../maptools/AbstractLayer.h"
#include "../../../se/Symbolizer.h"
#include "../../../se/PointSymbolizer.h"
#include "../../../se/LineSymbolizer.h"
#include "../../../se/PolygonSymbolizer.h"
#include "../../../se/TextSymbolizer.h"
#include "../../../se/Utils.h"
#include "../../../se/ParameterValue.h"
#include "SymbologyPreview.h"
#include "SymbolEditInfoDialog.h"
#include "SymbolizerEditInfoDialog.h"
#include "PointSymbolizerProperty.h"
#include "LineSymbolizerProperty.h"
#include "PolygonSymbolizerProperty.h"
#include "TextSymbolizerProperty.h"
#include "ui/ui_SymbolEditInfoDialogForm.h"

// Qt
#include <QFileDialog>
#include <QMessageBox>
#include <QTableWidgetItem>
#include <QCloseEvent>

// STL
#include <cassert>

te::qt::widgets::SymbolEditInfoDialog::SymbolEditInfoDialog(const std::vector<std::string>& fieldsNames, te::qt::widgets::Symbol* currentSymbol, QWidget* parent,
                                                            Qt::WindowFlags f)
  : QDialog(parent, f), m_ui(new Ui::SymbolEditInfoDialogForm),
  m_currentSymbol(nullptr),
  m_fieldsNames(fieldsNames),
  m_currentSymbolizer(nullptr),
  m_currentIconSize(32, 32),
  m_symbolizerType("")
{
  // Setup
  m_ui->setupUi(this);
  
  if (currentSymbol != nullptr)
  {
    m_currentSymbol.reset(currentSymbol->clone());
  }
  else
  {
    m_currentSymbol.reset(new te::qt::widgets::Symbol());
  }

  // connect signal/slot
  connect(m_ui->m_okPushButton, SIGNAL(clicked(bool)), this, SLOT(onOkPushButtonClicked()));
  connect(m_ui->m_cancelPushButton, SIGNAL(clicked(bool)), this, SLOT(onCancelPushButtonClicked()));
  connect(m_ui->m_helpPushButton, SIGNAL(clicked(bool)), this, SLOT(onHelpPushButtonClicked()));
  connect(m_ui->m_addPushButton, SIGNAL(clicked(bool)), this, SLOT(onAddPushButtonClicked()));
  connect(m_ui->m_editPushButton, SIGNAL(clicked(bool)), this, SLOT(onEditPushButtonClicked()));
  connect(m_ui->m_removePushButton, SIGNAL(clicked(bool)), this, SLOT(onRemovePushButtonClicked()));
  connect(m_ui->m_tblSymbolizers, SIGNAL(currentItemChanged(QTableWidgetItem*, QTableWidgetItem*)), 
    this, SLOT(onCurrentItemChanged(QTableWidgetItem*, QTableWidgetItem*)));

  m_ui->m_tblSymbolizers->setIconSize(m_currentIconSize);  

  m_ui->m_tblSymbolizers->verticalHeader()->setVisible(false);
  m_ui->m_tblSymbolizers->horizontalHeader()->setVisible(true);
  m_ui->m_tblSymbolizers->horizontalHeader()->setStretchLastSection(true);

  // Disable editing
  m_ui->m_tblSymbolizers->setEditTriggers(QAbstractItemView::NoEditTriggers);
  m_ui->m_tblSymbolizers->setSelectionBehavior(QAbstractItemView::SelectRows);
  m_ui->m_tblSymbolizers->setSelectionMode(QAbstractItemView::SingleSelection);

  m_ui->m_editPushButton->setEnabled(false);
  m_ui->m_removePushButton->setEnabled(false);

  initialize();
}

te::qt::widgets::SymbolEditInfoDialog::~SymbolEditInfoDialog()
{
  m_currentSymbolizer.release();
}

void te::qt::widgets::SymbolEditInfoDialog::initialize()
{
  initializeInfo();
  initializeTable();
}

void te::qt::widgets::SymbolEditInfoDialog::initializeInfo()
{
  te::qt::widgets::SymbolInfo symbolInfo = m_currentSymbol->getInfo();

  if (symbolInfo.m_id.empty())
  {
    m_id = te::common::GetUniqueID();
    return;
  }

  QString author = QString::fromUtf8(symbolInfo.m_author.c_str());
  m_ui->m_authorLineEdit->setText(author);

  QString description = QString::fromUtf8(symbolInfo.m_description.c_str());
  m_ui->m_descriptionTextEdit->setText(description);

  m_id = symbolInfo.m_id;
  
  QString name = QString::fromUtf8(symbolInfo.m_name.c_str());
  m_ui->m_nameLineEdit->setText(name);

  QString tags = QString::fromUtf8(symbolInfo.m_tags.c_str());
  m_ui->m_tagsLineEdit->setText(tags);
}

void te::qt::widgets::SymbolEditInfoDialog::onOkPushButtonClicked()
{
  if (m_currentSymbol == nullptr)
  {
    QMessageBox::information(this, "Symbol", "To create a symbol, you need to create at least one symbolizer.");
    return;
  }

  if (m_currentSymbol.get()->getSymbolizersCount() == 0)
  {
    QMessageBox::information(this, "Symbol", "To create a symbol, you need to create at least one symbolizer.");
    return;
  }

  if (isInfoFieldsEmpty() == false)
  {
    Symbol* symbol = createOrUpdateSymbol();

    m_currentSymbol.reset(symbol);

    accept();
  }
  else
  {
    QMessageBox::information(this, "Symbol", "Blank text field.");
  }  
}

void te::qt::widgets::SymbolEditInfoDialog::onCancelPushButtonClicked()
{
  reject();
}

void te::qt::widgets::SymbolEditInfoDialog::onHelpPushButtonClicked()
{

}

void te::qt::widgets::SymbolEditInfoDialog::onAddPushButtonClicked()
{  
  SymbolizerEditInfoDialog symbolizerEditInfoDialog(m_fieldsNames, nullptr, this);  
  symbolizerEditInfoDialog.setWindowTitle(tr("Add Symbolizer"));
  symbolizerEditInfoDialog.setSizePolicy(QSizePolicy::Preferred, QSizePolicy::Expanding);
  symbolizerEditInfoDialog.setSymbolizerType(m_symbolizerType);

  if(symbolizerEditInfoDialog.exec() == QDialog::Accepted)
  {
    te::se::Symbolizer* symbolizer = symbolizerEditInfoDialog.getSymbolizer();
    addSymbolizer(symbolizer);
    initializeTable();
  }
}

void te::qt::widgets::SymbolEditInfoDialog::onEditPushButtonClicked()
{
  te::se::Symbolizer* currentSymbolizer = m_currentSymbolizer.release();

  SymbolizerEditInfoDialog symbolizerEditInfoDialog(m_fieldsNames, currentSymbolizer, this);  
  symbolizerEditInfoDialog.setWindowTitle(tr("Edit Symbolizer"));
  symbolizerEditInfoDialog.setSizePolicy(QSizePolicy::Preferred, QSizePolicy::Expanding);

  if(symbolizerEditInfoDialog.exec() == QDialog::Accepted)
  {
    te::se::Symbolizer* symbolizer = symbolizerEditInfoDialog.getSymbolizer();
    updateSymbolizer(symbolizer);
    initializeTable();
    m_currentSymbolizer.reset(symbolizer);
  }
  else
  {
    m_currentSymbolizer.reset(currentSymbolizer);
  }
}

void te::qt::widgets::SymbolEditInfoDialog::onRemovePushButtonClicked()
{
  int result = QMessageBox::information(this, tr("Symbolizer Selector"), "Remove the selected symbolizer?", QMessageBox::Yes | QMessageBox::No);

  if (result == QMessageBox::No)
  {
    return;
  }

  removeCurrentSymbolizer();

  initializeTable();
}

te::qt::widgets::Symbol* te::qt::widgets::SymbolEditInfoDialog::createOrUpdateSymbol()
{
  te::qt::widgets::Symbol* symbol = nullptr;
  bool toAdd = false;
    
  if (m_currentSymbol != nullptr)
  {
    symbol = m_currentSymbol.release();
  }
  else
  {
    symbol = new te::qt::widgets::Symbol();
  }

  te::qt::widgets::SymbolInfo symbolInfo = createSymbolInfo();
  symbol->setInfo(symbolInfo);
      
  return symbol;
}

te::qt::widgets::SymbolInfo te::qt::widgets::SymbolEditInfoDialog::createSymbolInfo()
{
  te::qt::widgets::SymbolInfo symbolInfo;
  symbolInfo.m_author = m_ui->m_authorLineEdit->text().toUtf8().constData();
  symbolInfo.m_description = m_ui->m_descriptionTextEdit->toPlainText().toUtf8().constData();
  symbolInfo.m_id = m_id;
  symbolInfo.m_name = m_ui->m_nameLineEdit->text().toUtf8().constData();
  symbolInfo.m_tags = m_ui->m_tagsLineEdit->text().toUtf8().constData();

  if (m_currentSymbol != nullptr)
  {
    symbolInfo.m_id = m_currentSymbol->getInfo().m_id;
  }
  
  return symbolInfo;
}

te::qt::widgets::Symbol* te::qt::widgets::SymbolEditInfoDialog::getSymbol()
{
  return m_currentSymbol->clone();
}

bool te::qt::widgets::SymbolEditInfoDialog::isInfoFieldsEmpty()
{
  bool isEmpty = false;

  if (m_ui->m_authorLineEdit->text().isEmpty() ||
    m_ui->m_authorLineEdit->text().isNull())
  {
    isEmpty = true;
  }

  if (m_ui->m_descriptionTextEdit->toPlainText().isEmpty() ||
    m_ui->m_descriptionTextEdit->toPlainText().isNull())
  {
    isEmpty = true;
  }

  if (m_ui->m_nameLineEdit->text().isEmpty() ||
    m_ui->m_nameLineEdit->text().isNull())
  {
    isEmpty = true;
  }

  if (m_ui->m_tagsLineEdit->text().isEmpty() ||
    m_ui->m_tagsLineEdit->text().isNull())
  {
    isEmpty = true;
  }

  return isEmpty;
}

bool te::qt::widgets::SymbolEditInfoDialog::addSymbolizer(te::se::Symbolizer* symbolizer)
{
  m_currentSymbol->addSymbolizer(symbolizer);
  return true;
}

bool te::qt::widgets::SymbolEditInfoDialog::updateSymbolizer(te::se::Symbolizer* symbolizer)
{ 
  int index = m_ui->m_tblSymbolizers->currentIndex().row();
  if (index > -1)
  {
    m_currentSymbolizer.release();
    m_currentSymbol->setSymbolizer(index, symbolizer);
    return true;
  }
  
  return false;
}

bool te::qt::widgets::SymbolEditInfoDialog::removeCurrentSymbolizer()
{
  int index = m_ui->m_tblSymbolizers->currentIndex().row();
  if (index > -1)
  {
    m_currentSymbolizer.release();
    m_currentSymbol->removeSymbolizer(index);
    return true;
  }

  return false;
}

void te::qt::widgets::SymbolEditInfoDialog::initializeTable() 
{
  if (m_currentSymbol == nullptr)
  {
    return;
  }

  int symbolCount = static_cast<int>(m_currentSymbol->getSymbolizersCount());

  m_ui->m_tblSymbolizers->setRowCount(symbolCount);
  m_ui->m_tblSymbolizers->setColumnCount(2);

  m_ui->m_tblSymbolizers->setHorizontalHeaderItem(0, new QTableWidgetItem(tr("Symbol")));
  m_ui->m_tblSymbolizers->setHorizontalHeaderItem(1, new QTableWidgetItem(tr("Symbolizers")));

  for (int i = 0 ; i < symbolCount ; ++i)
  {
    te::se::Symbolizer* symbolizer = m_currentSymbol->getSymbolizer(i);

    // Image
    QTableWidgetItem* itemSymbolImg = new QTableWidgetItem();
    QPixmap img = SymbologyPreview::build(symbolizer, m_ui->m_tblSymbolizers->iconSize());
    itemSymbolImg->setIcon(img);
    m_ui->m_tblSymbolizers->setItem(i, 0, itemSymbolImg);
    
    // Type
    QString type = QString::fromUtf8(symbolizer->getType().c_str());
    QTableWidgetItem* itemSymbolType = new QTableWidgetItem();
    type += " " + QString::number(i);
    itemSymbolType->setText(type);
    m_ui->m_tblSymbolizers->setItem(i, 1, itemSymbolType);
  }

  m_ui->m_tblSymbolizers->resizeRowsToContents();
}

void te::qt::widgets::SymbolEditInfoDialog::onCurrentItemChanged(QTableWidgetItem * current, QTableWidgetItem * previous)
{
  m_currentSymbolizer.release();

  m_ui->m_editPushButton->setEnabled(true);
  m_ui->m_removePushButton->setEnabled(true);

  if (current != nullptr)
  {
    int index = current->row();  
  m_currentSymbolizer.reset(m_currentSymbol->getSymbolizer(index));
  }
  else
  {
    m_ui->m_editPushButton->setEnabled(false);
    m_ui->m_removePushButton->setEnabled(false);
  }
}

void te::qt::widgets::SymbolEditInfoDialog::setEnableObligatoryFields(bool enable)
{
  m_ui->m_authorLineEdit->setEnabled(enable);
  m_ui->m_descriptionTextEdit->setEnabled(enable);
  m_ui->m_nameLineEdit->setEnabled(enable);
  m_ui->m_tagsLineEdit->setEnabled(enable);
}

void te::qt::widgets::SymbolEditInfoDialog::setAuthor(const std::string& author)
{
   QString qauthor = QString::fromUtf8(author.c_str());
   m_ui->m_authorLineEdit->setText(qauthor);
}

void te::qt::widgets::SymbolEditInfoDialog::setDescription(const std::string& description)
{
  QString qdescription = QString::fromUtf8(description.c_str());
  m_ui->m_descriptionTextEdit->setText(qdescription);
}

void te::qt::widgets::SymbolEditInfoDialog::setName(const std::string& name)
{
  QString qname = QString::fromUtf8(name.c_str());
  m_ui->m_nameLineEdit->setText(qname);
}

void te::qt::widgets::SymbolEditInfoDialog::setTags(const std::string& tags)
{
  QString qtags = QString::fromUtf8(tags.c_str());
  m_ui->m_tagsLineEdit->setText(qtags);
}

void te::qt::widgets::SymbolEditInfoDialog::setSymbolizerType(const std::string& symbolizerType)
{
  m_symbolizerType = symbolizerType;
}
