/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/widgets/se/SymbolSelectorDialog.h

  \brief A dialog used to create or edit a specific symbol.
*/

#ifndef __TERRALIB_QT_WIDGETS_SE_INTERNAL_SYMBOL_EDITINFO_WIDGET_H
#define __TERRALIB_QT_WIDGETS_SE_INTERNAL_SYMBOL_EDITINFO_WIDGET_H

// TerraLib
#include "../Config.h"

// Qt
#include <QWidget>
#include <QMap>
#include <QDialog>
#include <QSize>

// STL
#include <memory>
#include <map>
#include <utility>
#include <vector>

class QString;
class QGridLayout;
class QTableWidgetItem;

namespace Ui
{
  class SymbolEditInfoDialogForm;
}

namespace te
{
  // Forward declarations
  namespace se
  {
    class Rule;
    class TextSymbolizer;
  }
  namespace qt
  {
    namespace widgets
    {
      // Forward declarations
      class  Symbol;
      struct SymbolInfo;
      class  SymbolPreviewWidget;
      class  Symbolizer;
      class  TextSymbolizerProperty;

      /*!
        \class SymbolSelectorDialog

        \brief A dialog used to create or edit a specific symbol.
      */
      class TEQTWIDGETSEXPORT SymbolEditInfoDialog : public QDialog
      {
        Q_OBJECT

        public:

          /** @name Initializer Methods
           *  Methods related to instantiation and destruction.
           */
          //@{

          /*! 
          \brief Constructs a symbol selector dialog which is a child of parent, with widget flags set to f. 

          \note field names from dataset for text symbolizer (no geometry)
          */
         SymbolEditInfoDialog(const std::vector<std::string>& fieldsNames, te::qt::widgets::Symbol* currentSymbol = nullptr, QWidget* parent = nullptr, Qt::WindowFlags f = nullptr);

          /*! \brief Destructor. */
          virtual ~SymbolEditInfoDialog();

          /*!
          \brief Gets the selected symbol.

          \return The selected fill element.       

            \note The caller will take the ownership of the returned symbol.
            \note It return a NULL pointer if there is not a symbol yet created.
          */
          virtual te::qt::widgets::Symbol* getSymbol();

          virtual void setEnableObligatoryFields(bool enable);

          void setAuthor(const std::string& author);

          void setDescription(const std::string& description);

          void setName(const std::string& name);

          void setTags(const std::string& tags);

          virtual void setSymbolizerType(const std::string& symbolizerType);
                              
        protected slots:
          
          void onOkPushButtonClicked();

          void onCancelPushButtonClicked();

          void onHelpPushButtonClicked();

          void onAddPushButtonClicked();

          void onEditPushButtonClicked();

          void onRemovePushButtonClicked();

          void onCurrentItemChanged(QTableWidgetItem * current, QTableWidgetItem * previous);

        protected:

          /*! \brief Initialize symbol configuration frame. */
          virtual void initialize();

          /*! \brief Initialize symbol information fields. */
          virtual void initializeInfo();
                    
          /*! \brief Will create a new symbol */
          virtual te::qt::widgets::Symbol* createOrUpdateSymbol();

          /*! \brief Will create a new symbol infomation from interface fields. */
          virtual te::qt::widgets::SymbolInfo createSymbolInfo();
          
          virtual bool isInfoFieldsEmpty();

          virtual bool addSymbolizer(te::se::Symbolizer* symbolizer);

          virtual bool updateSymbolizer(te::se::Symbolizer* symbolizer);

          virtual bool removeCurrentSymbolizer();

          virtual void initializeTable();
                                              
        private:

          std::unique_ptr<Ui::SymbolEditInfoDialogForm>       m_ui; //!< Dialog form.
          std::unique_ptr<te::qt::widgets::Symbol>            m_currentSymbol; //!< Current symbol
          std::string                                         m_id; //!< Symbol ID
          std::vector<std::string>                            m_fieldsNames; //!< Dataset fields names for text symbolizer
          std::unique_ptr<te::se::Symbolizer>                 m_currentSymbolizer; //!< Current symbol
          QSize                                               m_currentIconSize; //!< current icons size
          std::string                                         m_symbolizerType;
      }; 

    } // end namespace widgets
  }   // end namespace qt
}     // end namespace te

#endif  // __TERRALIB_QT_WIDGETS_SE_INTERNAL_SYMBOL_EDITINFO_WIDGET_H
