/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/widgets/se/SymbolLibrary.cpp
 
  \brief This class represents a library of symbols.
*/

// TerraLib
#include "../../../common/Exception.h"
#include "../../../common/STLUtils.h"
#include "../../../core/translator/Translator.h"
#include "Symbol.h"
#include "SymbolLibrary.h"
#include "SymbolNode.h"

// STL
#include <cassert>

te::qt::widgets::SymbolLibrary::SymbolLibrary(const std::string& name) : 
  m_name(name),
  m_root(nullptr),
  m_version("1.0.0")
{
  assert(!name.empty());

  // symbol library root node
  m_root.reset(new SymbolNode("Library"));
}

te::qt::widgets::SymbolLibrary::~SymbolLibrary()
{
  m_root->eraseAll();
}

void te::qt::widgets::SymbolLibrary::addGroup(te::qt::widgets::SymbolNode* group, const std::string& groupParentId)
{
  assert(group);
  
  if (groupParentId.compare("") == 0 ||
    m_root->getID().compare(groupParentId) == 0)
  {
    m_root->add(group);
  }
  else
  {
    te::qt::widgets::SymbolNode* parent = findGroupById(groupParentId);
    if (parent)
    {
      parent->add(group);
    }
  }  
}

te::qt::widgets::SymbolNode* te::qt::widgets::SymbolLibrary::removeGroup(te::qt::widgets::SymbolNode* group)
{
  assert(group);

  return m_root->remove(group);
}

void te::qt::widgets::SymbolLibrary::eraseGroup(te::qt::widgets::SymbolNode* group)
{
  assert(group);

  m_root->erase(group);
}

te::qt::widgets::SymbolNode* te::qt::widgets::SymbolLibrary::findGroupById(const std::string& id) const
{
  te::qt::widgets::SymbolNode* node = m_root->findNodeById(id);
  return node;
}

te::qt::widgets::Symbol* te::qt::widgets::SymbolLibrary::findSymbolById(const std::string& id) const
{
  te::qt::widgets::Symbol* symbol = m_root->findSymbolById(id);
  return symbol;
}

te::qt::widgets::SymbolNode* te::qt::widgets::SymbolLibrary::findParentBySymbolId(const std::string& id)
{
  te::qt::widgets::SymbolNode* node = m_root->findParentBySymbolId(id);
  return node;
}

void te::qt::widgets::SymbolLibrary::addSymbol(te::qt::widgets::Symbol* symbol, const std::string& groupParentId)
{
  assert(symbol);

  if (groupParentId.compare("") == 0 ||
    m_root->getID().compare(groupParentId) == 0)
  {
    m_root->addSymbol(symbol);
  }
  else
  {
    te::qt::widgets::SymbolNode* parent = findGroupById(groupParentId);
    if (parent)
    {
      parent->addSymbol(symbol);
    }
  }
}

te::qt::widgets::Symbol* te::qt::widgets::SymbolLibrary::removeSymbol(te::qt::widgets::Symbol* symbol)
{
  assert(symbol);
  
  return m_root->removeSymbol(symbol);
}

void te::qt::widgets::SymbolLibrary::eraseSymbol(te::qt::widgets::Symbol* symbol)
{
  assert(symbol);
  
  m_root->eraseSymbol(symbol);
}

const std::string& te::qt::widgets::SymbolLibrary::getName() const
{
  return m_name;
}

bool te::qt::widgets::SymbolLibrary::isEmpty()
{
  if (!m_root->hasNodes() && !m_root->hasSymbols())
  {
    return true;
  }
  return false;
}

te::qt::widgets::SymbolNode* te::qt::widgets::SymbolLibrary::getRoot() const
{
  return m_root.get();
}

bool te::qt::widgets::SymbolLibrary::swapGroup(const std::string& id, const std::string& toParentId)
{
  SymbolNode* group = findGroupById(id);

  assert(group);

  SymbolNode* oldParent = group->getParent();
  SymbolNode* newParent = findGroupById(toParentId);
  
  if (oldParent && newParent)
  {
    if (oldParent->remove(group))
    {
      newParent->add(group);
      return true;
    }
  } 
  else if (newParent)
  {
    newParent->add(group);
    return true;
  }
  else if (oldParent)
  {
    oldParent->remove(group);
    m_root->add(group);
    return true;
  }

  return false;
}

bool te::qt::widgets::SymbolLibrary::swapSymbol(const std::string& id, const std::string& toParentId)
{
  Symbol* symbol = findSymbolById(id);
  SymbolNode* oldParent = findParentBySymbolId(id);
  SymbolNode* newParent = findGroupById(toParentId);

  if (oldParent && newParent)
  {
    if (oldParent->removeSymbol(symbol))
    {
      newParent->addSymbol(symbol);
      return true;
    }
  } 
  else if (newParent)
  {
    newParent->addSymbol(symbol);
    return true;
  }
  else if (oldParent)
  {
    oldParent->removeSymbol(symbol);
    m_root->addSymbol(symbol);
    return true;
  }

  return false;
}

std::string te::qt::widgets::SymbolLibrary::version()
{
  return m_version;
}
