/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralib/qt/widgets/se/SymbolNode.cpp

   \brief This class represents a SymbolNode.
 */

 // TerraLib
#include "SymbolNode.h"
#include "../../../common/StringUtils.h"
#include "../../../common/STLUtils.h"
#include "../../../common/Exception.h"
#include "../../../core/translator/Translator.h"
#include "Utils.h"

// STL
#include <cassert>

te::qt::widgets::SymbolNode::SymbolNode():
  m_parent(nullptr),
  m_type(LEAF),
  m_name("unknown")
{
  m_id = te::common::GetUniqueID();
}

te::qt::widgets::SymbolNode::SymbolNode(const std::string& name, te::qt::widgets::SymbolNode* parent):
  m_parent(parent),
  m_type(LEAF),
  m_name(name)
{
  m_id = te::common::GetUniqueID();
}

te::qt::widgets::SymbolNode::~SymbolNode()
{
  eraseAll();
}

void te::qt::widgets::SymbolNode::add(te::qt::widgets::SymbolNode* node)
{
  assert(node);
  
  std::string id = node->getID();
  if(findNodeById(id))
    throw te::common::Exception(TE_TR("There is already a symbol group with the given id."));

  std::string name = node->getName();
  if(findNodeFromTopLevelByName(name))
    throw te::common::Exception(TE_TR("There is already a symbol group with the given name."));

  m_nodes[id] = node;
  node->setParent(this);

  m_type = te::qt::widgets::SymbolNode::NODE;
}

te::qt::widgets::SymbolNode* te::qt::widgets::SymbolNode::remove(te::qt::widgets::SymbolNode* node)
{
  assert(node);

  std::string id = node->getID();
  std::map<std::string, SymbolNode*>::iterator it = m_nodes.find(id);
  SymbolNode* foundNode = nullptr;

  if (it != m_nodes.end())
  {
    m_nodes.erase(it);

    if (m_nodes.empty())
    {
      m_type = te::qt::widgets::SymbolNode::LEAF;
    }
    return node;
  }
  else
  {

    for (it = m_nodes.begin(); it != m_nodes.end(); ++it)
    {
      foundNode = it->second->remove(node);
    }
  }

  return foundNode;
}

const std::map<std::string, te::qt::widgets::SymbolNode*>& te::qt::widgets::SymbolNode::getNodes() const
{
  return m_nodes;
}

std::size_t te::qt::widgets::SymbolNode::getNodesCount() const
{
  return m_nodes.size();
}

te::qt::widgets::SymbolNode::SymbolNodeType te::qt::widgets::SymbolNode::getType()
{
  return m_type;
}

void te::qt::widgets::SymbolNode::addSymbol(te::qt::widgets::Symbol* symbol)
{
  assert(symbol);

  std::string id = symbol->getInfo().m_id;
  if(findSymbolById(id))
    throw te::common::Exception(TE_TR("There is already a symbol with the given id."));

  m_symbolMap[id] = symbol;
}

te::qt::widgets::Symbol* te::qt::widgets::SymbolNode::removeSymbol(te::qt::widgets::Symbol* symbol)
{
  assert(symbol);

  std::string id = symbol->getInfo().m_id;
  std::map<std::string, Symbol*>::iterator it = m_symbolMap.find(id);
  te::qt::widgets::Symbol* resultSymbol = nullptr;

  if (it != m_symbolMap.end())
  {
    m_symbolMap.erase(it);
    return symbol;
  }
  else
  {
    std::map<std::string, SymbolNode*>::const_iterator itn;

    for (itn = m_nodes.cbegin(); itn != m_nodes.cend(); ++itn)
    {
      resultSymbol = itn->second->removeSymbol(symbol);
    }
  }

  return resultSymbol;
}

const std::map<std::string, te::qt::widgets::Symbol*>& te::qt::widgets::SymbolNode::getSymbols() const
{
  return m_symbolMap;
}

te::qt::widgets::Symbol* te::qt::widgets::SymbolNode::findSymbolById(const std::string& id) const
{
  std::map<std::string, Symbol*>::const_iterator it = m_symbolMap.find(id);
  
  if (it != m_symbolMap.end())
  {
    return it->second;
  }

  std::map<std::string, SymbolNode*>::const_iterator itn;
  te::qt::widgets::Symbol* symbol = nullptr;

  for (itn = m_nodes.cbegin(); itn != m_nodes.cend(); ++itn)
  {
    symbol = itn->second->findSymbolById(id);
    if (symbol)
    {
      break;
    }
  }

  return symbol;
}

te::qt::widgets::SymbolNode* te::qt::widgets::SymbolNode::findNodeById(const std::string& id) const
{
  std::map<std::string, SymbolNode*>::const_iterator it = m_nodes.find(id);

  if (it != m_nodes.cend())
  {
    return it->second;
  }

  te::qt::widgets::SymbolNode* node = nullptr;

  for (it = m_nodes.cbegin(); it != m_nodes.cend(); ++it)
  {
    node = it->second->findNodeById(id);
    if (node)
    {
      break;
    }
  }

  return node;
}

te::qt::widgets::SymbolNode* te::qt::widgets::SymbolNode::findParentBySymbolId(const std::string& id)
{
  std::map<std::string, Symbol*>::const_iterator it = m_symbolMap.find(id);
  
  if (it != m_symbolMap.cend())
  {
    return this;
  }

  std::map<std::string, SymbolNode*>::const_iterator itn;
  te::qt::widgets::SymbolNode* parent = nullptr;

  for (itn = m_nodes.cbegin(); itn != m_nodes.cend(); ++itn)
  {
    parent = itn->second->findParentBySymbolId(id);
    if (parent)
    {
      break;
    }
  }

  return parent;
}

te::qt::widgets::SymbolNode* te::qt::widgets::SymbolNode::findNodeFromTopLevelByName(const std::string& name) const
{
  te::qt::widgets::SymbolNode* node = nullptr;

  if (!hasNodes())
  {
    return node;
  }

  std::map<std::string, SymbolNode*>::const_iterator it = m_nodes.cbegin();

  for (; it != m_nodes.cend(); ++it)
  {
    std::string nodeName = it->second->getName();
    if (nodeName.compare(name) == 0)
    {
      node = it->second;
      break;
    }
  }

  return node;
}

te::qt::widgets::SymbolNode* te::qt::widgets::SymbolNode::findNodeFromTopLevelById(const std::string& id) const
{
  te::qt::widgets::SymbolNode* node = nullptr;

  if (!hasNodes())
  {
    return node;
  }

  std::map<std::string, SymbolNode*>::const_iterator it = m_nodes.cbegin();

  for (; it != m_nodes.cend(); ++it)
  {
    std::string nodeId = it->second->getID();
    if (nodeId.compare(id) == 0)
    {
      node = it->second;
      break;
    }
  }

  return node;
}

void te::qt::widgets::SymbolNode::erase(te::qt::widgets::SymbolNode* node)
{
  SymbolNode* removedNode = remove(node);
  assert(removedNode);

  removedNode->clear();
  delete removedNode;
}

void te::qt::widgets::SymbolNode::eraseAll()
{
  std::map<std::string, SymbolNode*>::iterator it;

  for (it = m_nodes.begin(); it != m_nodes.end();)
  {
    te::qt::widgets::SymbolNode* node = it->second;
    node->eraseAll();    
    node->clear();

    delete node;
    it = m_nodes.erase(it);
  }

  clear();

  m_type = te::qt::widgets::SymbolNode::LEAF;
}

void te::qt::widgets::SymbolNode::eraseSymbol(te::qt::widgets::Symbol* symbol)
{
  Symbol* removedSymbol = removeSymbol(symbol);
  assert(removedSymbol);

  delete removedSymbol;
}

void te::qt::widgets::SymbolNode::clear()
{
  te::common::FreeContents(m_symbolMap);
  m_symbolMap.clear();
}

std::string te::qt::widgets::SymbolNode::getID() const
{
  return m_id;
}

std::string te::qt::widgets::SymbolNode::getName() const
{
  return m_name;
}

bool te::qt::widgets::SymbolNode::hasNodes() const
{
  return !m_nodes.empty();
}

bool te::qt::widgets::SymbolNode::hasSymbols() const
{
  return !m_symbolMap.empty();
}
void te::qt::widgets::SymbolNode::setParent(SymbolNode* parent)
{
  m_parent = parent;
}

te::qt::widgets::SymbolNode* te::qt::widgets::SymbolNode::getParent() const
{
  return m_parent;
}

void te::qt::widgets::SymbolNode::setName(const std::string& name)
{
  m_name = name;
}

