/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/widgets/se/SymbolNode.h

  \brief This class represents a SymbolNode.
*/

#ifndef __TERRALIB_QT_WIDGETS_SE_INTERNAL_SYMBOL_NODE_H
#define __TERRALIB_QT_WIDGETS_SE_INTERNAL_SYMBOL_NODE_H

// TerraLib
#include "../Config.h"
#include "Symbol.h"

// STL
#include <string>
#include <map>

namespace te
{
  namespace qt
  {
    namespace widgets
    {
      /*!
        \class Symbol

        \brief This class represents a SymbolNode. TODO: More description!
      */
      class TEQTWIDGETSEXPORT SymbolNode
      {
        private:

          /*!
            \enum NodeType

            \brief Auxiliary internal enumeration to control the tree nodes.
          */
          enum SymbolNodeType
          {
            NODE = 0, /*!< node */
            LEAF = 1,  /*!< leaf */
          };

        public:

          SymbolNode();

          /** @name Initializer Methods
            *  Methods related to instantiation and destruction.
            */
          //@{

          /*! \brief Default constructor. */
          SymbolNode(const std::string& name, te::qt::widgets::SymbolNode* parent = nullptr);

          /*! \brief Destructor. */
          virtual ~SymbolNode();
          
          virtual void add(te::qt::widgets::SymbolNode* node);

          virtual te::qt::widgets::SymbolNode* remove(te::qt::widgets::SymbolNode* node);

          virtual void erase(te::qt::widgets::SymbolNode* node);

          virtual void eraseAll();
                    
          /*!
          \brief 

          \return 
            
          \note 
          */
          const std::map<std::string, te::qt::widgets::SymbolNode*>& getNodes() const;
          
          /*!
          \brief 

          \return 
          */
          std::size_t getNodesCount() const;

          /*!
            \brief It inserts a new symbol to this library.

            \param symbol The new symbol that will be added to this library.
          */
          void addSymbol(Symbol* symbol);

          /*!
            \brief It removes the symbol from this library.

            \param symbol The symbol to be removed.

            \exception Exception If the symbol doesn't exist it will raise an exception.
          */
          Symbol* removeSymbol(Symbol* symbol);

          /*!
            \brief It removes the symbol from this library.

            \param symbol The symbol to be removed.

            \exception Exception If the symbol doesn't exist it will raise an exception.
          */
          void eraseSymbol(Symbol* symbol);

          /*!
          \brief 

          \return 
            
          \note 
          */
          const std::map<std::string, te::qt::widgets::Symbol*>& getSymbols() const;

          /*!
            \brief It returns the symbol identified by a given id or NULL if none is found.

            \param name The id of the symbol we are looking for.

            \return A pointer to a symbol or NULL if none is found.
          */
          te::qt::widgets::Symbol* findSymbolById(const std::string& id) const;

          /*!
            \brief It returns the symbol identified by a given id or NULL if none is found.

            \param name The id of the symbol we are looking for.

            \return A pointer to a symbol or NULL if none is found.
          */
          te::qt::widgets::SymbolNode* findParentBySymbolId(const std::string& id);

          /*!
            \brief It returns the group identified by a given id or NULL if none is found.

            \param name The id of the symbol we are looking for.

            \return A pointer to a symbol or NULL if none is found.
          */
          te::qt::widgets::SymbolNode* findNodeById(const std::string& id) const;
          
          /*!
            \brief It returns the group identified by a given id or NULL if none is found.

            \param name The name of the group we are looking for.

            \return A pointer to a group or NULL if none is found.
          */
          te::qt::widgets::SymbolNode* findNodeFromTopLevelByName(const std::string& name) const;

          /*!
            \brief It returns the group identified by a given id or NULL if none is found.

            \param name The name of the group we are looking for.

            \return A pointer to a group or NULL if none is found.
          */
          te::qt::widgets::SymbolNode* findNodeFromTopLevelById(const std::string& id) const;
          
          bool hasNodes() const;

          bool hasSymbols() const;

          std::string getID() const;

          void setName(const std::string& name);

          std::string getName() const;

          void setParent(SymbolNode* parent);

          SymbolNode* getParent() const;

        protected:

          SymbolNodeType getType();

          /*! \brief It removes all symbols from this library. */
          void clear();

        protected:

          std::map<std::string, te::qt::widgets::SymbolNode*>       m_nodes;
          te::qt::widgets::SymbolNode*                              m_parent;
          SymbolNodeType                                            m_type;
          std::map<std::string, Symbol*>                            m_symbolMap; //!< The set of symbol.
          std::string                                               m_id;
          std::string                                               m_name;
      };

    } // end namespace widgets
  }   // end namespace qt
}     // end namespace te

#endif  // __TERRALIB_QT_WIDGETS_SE_INTERNAL_SYMBOL_NODE_H
