/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled
   applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/widgets/se/SymbolSelectorDialog.cpp

  \brief A dialog used to select a specific symbol.
*/

// TerraLib
#include "../../../common/Exception.h"
#include "../Utils.h"
#include "Symbol.h"
#include "SymbolInfoDialog.h"
#include "SymbolLibrary.h"
#include "SymbolLibraryManager.h"
#include "SymbolPreviewWidget.h"
#include "SymbolSelectorDialog.h"
#include "SymbologyPreview.h"
#include "SymbolEditInfoDialog.h"
#include "SymbolGroupDialog.h"
#include "SymbolTreeFilter.h"
#include "Utils.h"
#include "SymbolNode.h"
#include "ui/ui_SymbolSelectorDialogForm.h"

// Qt
#include <QFileDialog>
#include <QMessageBox>
#include <QStringListIterator>
#include <QHeaderView>

// STL
#include <cassert>

te::qt::widgets::SymbolSelectorDialog::SymbolSelectorDialog(const std::vector<std::string>& fieldsNames, QWidget* parent,
                                                            Qt::WindowFlags f)
  : QDialog(parent, f), m_ui(new Ui::SymbolSelectorDialogForm),
  m_fieldsNames(fieldsNames),
  m_currentDragDropItem(nullptr)
{
  // Setup
  m_ui->setupUi(this);
  m_ui->m_symbolLibraryTreeWidget->setIconSize(QSize(32, 32));

  // Preview
  m_preview = new SymbolPreviewWidget(QSize(120, 120), this);

  // Adjusting...
  QGridLayout* previewLayout = new QGridLayout(m_ui->m_previewGroupBox);
  previewLayout->setContentsMargins(0, 0, 0, 0);
  previewLayout->addWidget(m_preview);

  // Signals & slots
  connect(m_ui->m_symbolLibraryTreeWidget,
          SIGNAL(currentItemChanged(QTreeWidgetItem*, QTreeWidgetItem*)),
          SLOT(onCurrentItemChanged(QTreeWidgetItem*, QTreeWidgetItem*)));
  connect(m_ui->m_showSymbolInfoPushButton, SIGNAL(pressed()),
          SLOT(onShowSymbolInfoPushButtonPressed()));
  connect(m_ui->m_importSymbolLibraryPushButton, SIGNAL(pressed()),
          SLOT(onImportButtonPressed()));
  connect(m_ui->m_exportSymbolInfoPushButton, SIGNAL(pressed()),
          SLOT(onExportButtonPressed()));
  connect(m_ui->m_searchLineEdit, SIGNAL(textChanged(const QString&)),
          SLOT(onSearchLineEditTextChanged(const QString&)));

  /* Manage symbol */
  connect(m_ui->m_createSymbolPushButton, SIGNAL(clicked(bool)), this, SLOT(onCreateSymbolPushButtonClicked()));
  connect(m_ui->m_editSymbolPushButton, SIGNAL(clicked(bool)), this, SLOT(onEditSymbolPushButtonClicked()));
  connect(m_ui->m_removeSymbolPushButton, SIGNAL(clicked(bool)), this, SLOT(onRemoveSymbolPushButtonClicked()));

  /* Manage group */
  connect(m_ui->m_createGroupPushButton, SIGNAL(clicked(bool)), this, SLOT(onCreateGroupPushButtonClicked()));
  connect(m_ui->m_renameGroupPushButton, SIGNAL(clicked(bool)), this, SLOT(onRenameGroupPushButtonClicked()));
  connect(m_ui->m_removeGroupPushButton, SIGNAL(clicked(bool)), this, SLOT(onRemoveGroupPushButtonClicked()));

  connect(m_ui->m_okPushButton, SIGNAL(clicked(bool)), this, SLOT(onOkPushButtonClicked()));
  connect(m_ui->m_closePushButton, SIGNAL(clicked(bool)), this, SLOT(onClosePushButtonClicked()));
  connect(m_ui->m_helpPushButton, SIGNAL(clicked(bool)), this, SLOT(onHelpPushButtonClicked()));
  
  enableSymbolItem(false);
  enableGroupItem(false);

  m_ui->m_symbolLibraryTreeWidget->header()->setStretchLastSection(false);
  te::qt::widgets::SetHeaderViewResizeMode(m_ui->m_symbolLibraryTreeWidget->header(), QHeaderView::ResizeToContents);

  // tree: enable drag and drop
  m_ui->m_symbolLibraryTreeWidget->setSelectionMode(QAbstractItemView::SingleSelection);
  m_ui->m_symbolLibraryTreeWidget->setDragEnabled(true);
  m_ui->m_symbolLibraryTreeWidget->viewport()->setAcceptDrops(true);
  m_ui->m_symbolLibraryTreeWidget->setDropIndicatorShown(true);
  m_ui->m_symbolLibraryTreeWidget->setDragDropMode(QAbstractItemView::InternalMove);

  // Install event filter for drag and drop event
  SymbolTreeFilter* filter = new SymbolTreeFilter();
  m_ui->m_symbolLibraryTreeWidget->viewport()->installEventFilter(filter);
  connect(filter, SIGNAL(drag()), this, SLOT(onDragEnterEvent()));
  connect(filter, SIGNAL(drop()), this, SLOT(onDropEvent()), Qt::QueuedConnection); // The tree needs to execute this event first

  initialize();
}

te::qt::widgets::SymbolSelectorDialog::~SymbolSelectorDialog()
{
}

void te::qt::widgets::SymbolSelectorDialog::onCurrentItemChanged(
    QTreeWidgetItem* current, QTreeWidgetItem* previous)
{
  if (current == nullptr)
  {
    enableSymbolItem(false);
    enableGroupItem(false);
    return;
  }
     
  // check if item is not a symbol 
  if(current->type() == GROUP)
  {
    enableSymbolItem(false);
    enableGroupItem(true);
    return;
  }

  Symbol* symbol = getSymbolFromItem(current);

  if(symbol == nullptr)
    return;

  m_preview->updatePreview(symbol);

  // enable buttons to edit and remove
  enableSymbolItem(true);
  enableGroupItem(false);
}

void te::qt::widgets::SymbolSelectorDialog::onShowSymbolInfoPushButtonPressed()
{
  Symbol* symbol = getSelectedSymbol();

  if(symbol == nullptr)
    return;

  SymbolInfoDialog info;
  info.setSymbolInfo(symbol->getInfo());
  info.setReadMode();

  info.exec();
}

void te::qt::widgets::SymbolSelectorDialog::onImportButtonPressed()
{
  QString qPath = te::qt::widgets::GetFilePathFromSettings("Symbology Encoding");

  QString path = QFileDialog::getOpenFileName(
      this, tr("Select a TerraLib Symbol Library File"), qPath,
      "TerraLib Symbol Library Files (*.xml)" );
  if(path.isNull())
    return;

  try
  {
    ReadSymbolLibrary(Convert2StdString(path));
    initialize();
  }
  catch(te::common::Exception& e)
  {
    QString message =
        tr("The selected symbol library could not be import.\n Details: )");
    message += e.what();
    QMessageBox::critical(this, tr("Error"), message);
  }
}

void te::qt::widgets::SymbolSelectorDialog::onExportButtonPressed()
{
  QString dir = QFileDialog::getExistingDirectory(this, tr("Select a Directory"),
     "", QFileDialog::ShowDirsOnly | QFileDialog::DontResolveSymlinks);

  if(dir.isNull())
    return;

  te::qt::widgets::AddFilePathToSettings(dir, "Symbology Encoding");

  try
  {
    dir += "/symbolLibrary.xml";

    WriteSymbolLibrary(Convert2StdString(dir));

    QMessageBox::information(this, tr("Export Symbol Library"), "Library successfully exported.");

    initialize();
  }
  catch(te::common::Exception& e)
  {
    QString message =
        tr("The selected symbol library could not be export.\n Details: )");
    message += e.what();
    QMessageBox::critical(this, tr("Error"), message);
  }
}

void te::qt::widgets::SymbolSelectorDialog::onSearchLineEditTextChanged(
    const QString& text)
{
  QList<QTreeWidgetItem*> items = m_ui->m_symbolLibraryTreeWidget->findItems(
      text, Qt::MatchContains | Qt::MatchRecursive, 0);
  items.append(m_ui->m_symbolLibraryTreeWidget->findItems(
      text, Qt::MatchContains | Qt::MatchRecursive, 1));
  items.append(m_ui->m_symbolLibraryTreeWidget->findItems(
      text, Qt::MatchContains | Qt::MatchRecursive, 2));

  filter(items);
}

void te::qt::widgets::SymbolSelectorDialog::initialize()
{
  m_ui->m_symbolLibraryTreeWidget->clear();

  SymbolLibrary* library = SymbolLibraryManager::getInstance().getLibrary();
  
  if (library->isEmpty() == true)
  {
	  m_ui->m_exportSymbolInfoPushButton->setEnabled(false);
  }
  else
  {
	  m_ui->m_exportSymbolInfoPushButton->setEnabled(true);
  }

  SymbolNode* rootNode = library->getRoot();  
  
  QTreeWidgetItem* rootParent = nullptr;

  buildTree(rootNode, rootParent);
  
  m_ui->m_symbolLibraryTreeWidget->sortItems(0, Qt::AscendingOrder);
  m_ui->m_symbolLibraryTreeWidget->expandAll();
}

void te::qt::widgets::SymbolSelectorDialog::buildTree(te::qt::widgets::SymbolNode* node, QTreeWidgetItem* parent)
{
  std::map<std::string, SymbolNode*> nodes = node->getNodes();
  
  std::map<std::string, SymbolNode*>::const_iterator it = nodes.cbegin();
  
  for (; it != nodes.cend() ; ++it)
  {
    SymbolNode* group = it->second;
    QTreeWidgetItem* groupItem = nullptr;
    if (parent)
    {
      groupItem = createItemFromGroup(group, parent);
    }
    else
    {
      groupItem = createItemFromGroup(group, m_ui->m_symbolLibraryTreeWidget);
    }
    buildTree(group, groupItem); // create sub group    
  }

  // Add all symbols from the current node
  std::map<std::string, Symbol*> symbols = node->getSymbols();

  std::map<std::string, Symbol*>::const_iterator itSymbol = symbols.cbegin();

  for (; itSymbol != symbols.cend() ; ++itSymbol)
  {
    Symbol* symbol = itSymbol->second;
    if (parent)
    {
      createItemFromSymbol(symbol, parent);
    }
    else
    {
      createItemFromSymbol(symbol, m_ui->m_symbolLibraryTreeWidget);
    }
  }
}

QTreeWidgetItem* te::qt::widgets::SymbolSelectorDialog::createItemFromGroup(SymbolNode* node, QTreeWidgetItem* parent)
{
  QTreeWidgetItem* item = new QTreeWidgetItem(parent, GROUP);

  // Bulding the Qt symbol item
  item->setText(0, Convert2QString(node->getName()));
  item->setData(0, Qt::UserRole, QVariant(Convert2QString(node->getID())));

  return item;
}
QTreeWidgetItem* te::qt::widgets::SymbolSelectorDialog::createItemFromGroup(SymbolNode* node, QTreeWidget* parent)
{
  QTreeWidgetItem* item = new QTreeWidgetItem(parent, GROUP);

  // Bulding the Qt symbol item
  item->setText(0, Convert2QString(node->getName()));
  item->setData(0, Qt::UserRole, QVariant(Convert2QString((node->getID()))));

  return item;
}

QTreeWidgetItem* te::qt::widgets::SymbolSelectorDialog::createItemFromSymbol(Symbol* symbol, QTreeWidgetItem* parent)
{
  QTreeWidgetItem* item = new QTreeWidgetItem(parent, SYMBOL);

  configItemFromSymbol(symbol, item);

  return item;
}

QTreeWidgetItem* te::qt::widgets::SymbolSelectorDialog::createItemFromSymbol(Symbol* symbol, QTreeWidget* parent)
{
  QTreeWidgetItem* item = new QTreeWidgetItem(parent, SYMBOL);

  configItemFromSymbol(symbol, item);

  return item;
}

void te::qt::widgets::SymbolSelectorDialog::configItemFromSymbol(Symbol* symbol, QTreeWidgetItem* item)
{
  // Bulding the Qt symbol item
  item->setText(0, Convert2QString(symbol->getInfo().m_name));
  item->setIcon(0, SymbologyPreview::build(
                              symbol->getSymbolizers(),
                              m_ui->m_symbolLibraryTreeWidget->iconSize()));
  item->setToolTip(0, formatSymbolInfo(symbol->getInfo()));
  item->setData(0, Qt::UserRole, QVariant(Convert2QString(symbol->getInfo().m_id)));
  item->setText(1, Convert2QString(symbol->getInfo().m_author));
  item->setText(2, Convert2QString(symbol->getInfo().m_tags));
  item->setText(3, Convert2QString(symbol->getInfo().m_description));

  // Symbol: couldn't drop
  auto flags = item->flags();
  flags = flags & (~Qt::ItemIsDropEnabled);
  
  item->setFlags(flags);
}

void te::qt::widgets::SymbolSelectorDialog::filter(
    const QList<QTreeWidgetItem*>& items)
{
  for(int i = 0; i < m_ui->m_symbolLibraryTreeWidget->topLevelItemCount(); ++i)
  {
    QTreeWidgetItem* library = m_ui->m_symbolLibraryTreeWidget->topLevelItem(i);
    assert(library && library->type() == GROUP);

    for(int j = 0; j < library->childCount(); ++j)
    {
      QTreeWidgetItem* symbol = library->child(j);
      assert(symbol && symbol->type() == SYMBOL);
      bool hide = items.indexOf(symbol) == -1;
      symbol->setHidden(hide);
    }
  }
  update();
}

te::qt::widgets::Symbol*
te::qt::widgets::SymbolSelectorDialog::getSelectedSymbol() const
{
  QList<QTreeWidgetItem*> selected =
      m_ui->m_symbolLibraryTreeWidget->selectedItems();

  if(selected.empty())
    return nullptr;

  return getSymbolFromItem(selected.at(0));
}

te::qt::widgets::Symbol*
te::qt::widgets::SymbolSelectorDialog::getSymbolFromItem(
    QTreeWidgetItem* item) const
{
  assert(item);

  if(item->type() != SYMBOL)
    return nullptr;
    
  SymbolLibrary* symbolLibrary =
    SymbolLibraryManager::getInstance().getLibrary();
  assert(symbolLibrary);

  // Gets the symbol
  QString id = item->data(0, Qt::UserRole).toString();
  assert(!id.isEmpty());

  Symbol* symbol = symbolLibrary->findSymbolById(Convert2StdString(id));
  assert(symbol);

  return symbol;
}

QString te::qt::widgets::SymbolSelectorDialog::formatSymbolInfo(
    const te::qt::widgets::SymbolInfo& info) const
{
  QString information("<h3>Symbol Information</h3><ul>");
  information += "<li><b>Name: </b>" + Convert2QString(info.m_name) + "</li>";
  information += "<li><b>Author: </b>" + Convert2QString(info.m_author) + "</li>";
  information += "<li><b>Tags: </b>" + Convert2QString(info.m_tags) + "</li>";
  information += "<li><b>Description: </b>" + Convert2QString(info.m_description) + "</li>";
  information += "</ul>";

  return information;
}

te::qt::widgets::Symbol* te::qt::widgets::SymbolSelectorDialog::getSymbol()
    const
{
  Symbol* symbol = getSelectedSymbol();

  if (symbol == nullptr)
  {
    return nullptr;
  }

  return symbol->clone();
}

void te::qt::widgets::SymbolSelectorDialog::onCreateSymbolPushButtonClicked()
{
  Symbol* symbol = nullptr;
  QString parentId = getCurrentGroupItemID();

  SymbolEditInfoDialog symbolEditInfoWidget(m_fieldsNames, symbol, this);  
  symbolEditInfoWidget.setWindowTitle(tr("Create Symbol"));
  symbolEditInfoWidget.setSizePolicy(QSizePolicy::Preferred, QSizePolicy::Expanding);

  if(symbolEditInfoWidget.exec() == QDialog::Accepted)
  {
    symbol = symbolEditInfoWidget.getSymbol();
    addSymbol(symbol, parentId);
    initialize();
  }
}

void te::qt::widgets::SymbolSelectorDialog::onEditSymbolPushButtonClicked()
{
  Symbol* symbol = getSelectedSymbol();

  SymbolEditInfoDialog symbolEditInfoWidget(m_fieldsNames, symbol, this);  
  symbolEditInfoWidget.setWindowTitle(tr("Edit Symbol"));
  symbolEditInfoWidget.setSizePolicy(QSizePolicy::Preferred, QSizePolicy::Expanding);

  if(symbolEditInfoWidget.exec() == QDialog::Accepted)
  {
    symbol = symbolEditInfoWidget.getSymbol();
    updateCurrentSymbol(symbol);
    initialize();
  }
}

void te::qt::widgets::SymbolSelectorDialog::onRemoveSymbolPushButtonClicked()
{
  int result = QMessageBox::information(this, tr("Symbol Selector"), "Remove the selected symbol?", QMessageBox::Yes | QMessageBox::No);

  if (result == QMessageBox::No)
  {
    return;
  }

  eraseCurrentSymbol();
  
  initialize();
}

void te::qt::widgets::SymbolSelectorDialog::onCreateGroupPushButtonClicked()
{
  QString parentId = getCurrentGroupItemID();

  SymbolGroupDialog symbolGroupDialog(this);  
  symbolGroupDialog.setWindowTitle(tr("Create Symbol Group"));

  if(symbolGroupDialog.exec() == QDialog::Accepted)
  {
    std::string name = symbolGroupDialog.getGroupName();
    QString qName = Convert2QString(name);
    addGroup(qName, parentId);
    initialize();
  }
}

void te::qt::widgets::SymbolSelectorDialog::onRenameGroupPushButtonClicked()
{
  QString id = getCurrentGroupItemID();
  QTreeWidgetItem* item = getCurrentGroupItem();
  std::string groupName = Convert2StdString(item->text(0));

  SymbolGroupDialog symbolGroupDialog(groupName, this);  
  symbolGroupDialog.setWindowTitle(tr("Rename Symbol Group"));

  if(symbolGroupDialog.exec() == QDialog::Accepted)
  {
    std::string name = symbolGroupDialog.getGroupName();
    QString qName = Convert2QString(name);
    updateCurrentGroup(qName, id);
    initialize();
  }
}

void te::qt::widgets::SymbolSelectorDialog::onRemoveGroupPushButtonClicked()
{
  int result = QMessageBox::information(this, tr("Symbol Selector"), "Remove the selected group?", QMessageBox::Yes | QMessageBox::No);

  if (result == QMessageBox::No)
  {
    return;
  }
      
  eraseCurrentGroup();
  
  initialize();
}

void te::qt::widgets::SymbolSelectorDialog::onOkPushButtonClicked()
{
  accept();
}

void te::qt::widgets::SymbolSelectorDialog::onClosePushButtonClicked()
{
  reject();
}

void te::qt::widgets::SymbolSelectorDialog::onHelpPushButtonClicked()
{

}

bool te::qt::widgets::SymbolSelectorDialog::addSymbol(Symbol* symbol, const QString& groupParentId)
{
  if (symbol == nullptr)
  {
    return false;
  }

  te::qt::widgets::SymbolLibrary* symbolLibrary = SymbolLibraryManager::getInstance().getLibrary();
  if (symbolLibrary != nullptr)
  {
    std::string parentId = Convert2StdString(groupParentId);
    symbolLibrary->addSymbol(symbol, parentId);
    return true;
  }

  return false;
}

bool te::qt::widgets::SymbolSelectorDialog::updateCurrentSymbol(Symbol* symbol)
{
  QString id = getCurrentGroupItemID();

  // remove by id
  removeSymbol(symbol);

  bool result = addSymbol(symbol, id);

  return result;
}

bool te::qt::widgets::SymbolSelectorDialog::eraseCurrentSymbol()
{
  Symbol* symbol = getSelectedSymbol();

  if (symbol == nullptr)
  {
    return false;
  }
  
  bool result = eraseSymbol(symbol);  

  return result;
}

bool te::qt::widgets::SymbolSelectorDialog::eraseSymbol(Symbol* symbol)
{
  te::qt::widgets::SymbolLibrary* symbolLibrary = SymbolLibraryManager::getInstance().getLibrary();
  if (symbolLibrary == nullptr)
  {
    return false;
  }

  symbolLibrary->eraseSymbol(symbol);
  
  return true;
}

bool te::qt::widgets::SymbolSelectorDialog::removeSymbol(Symbol* symbol)
{
  te::qt::widgets::SymbolLibrary* symbolLibrary = SymbolLibraryManager::getInstance().getLibrary();
  if (symbolLibrary == nullptr)
  {
    return false;
  }

  symbolLibrary->removeSymbol(symbol);
  
  return true;
}

bool te::qt::widgets::SymbolSelectorDialog::addGroup(const QString& groupName, const QString& groupParentId)
{
  te::qt::widgets::SymbolLibrary* symbolLibrary = SymbolLibraryManager::getInstance().getLibrary();
  if (symbolLibrary == nullptr)
  {
    return false;
  }

  std::string name = Convert2StdString(groupName);
  std::string parentId = Convert2StdString(groupParentId);

  SymbolNode* node = new SymbolNode(name);

  symbolLibrary->addGroup(node, parentId);

  return true;
}

bool te::qt::widgets::SymbolSelectorDialog::updateCurrentGroup(const QString& newGroupName, const QString& groupId)
{
  te::qt::widgets::SymbolLibrary* symbolLibrary = SymbolLibraryManager::getInstance().getLibrary();
  if (symbolLibrary == nullptr)
  {
    return false;
  }

  std::string name = Convert2StdString(newGroupName);
  std::string id = Convert2StdString(groupId);

  SymbolNode* group = symbolLibrary->findGroupById(id);
  if (!group)
  {
    return false;
  }

  group->setName(name);

  return true;
}

bool te::qt::widgets::SymbolSelectorDialog::eraseCurrentGroup()
{
  QTreeWidgetItem* item = getSelectedItem();

  if (item->type() != GROUP)
  {
    return false;
  }

  QString id = item->data(0, Qt::UserRole).toString();

  bool result = eraseGroup(id);

  return result;
}

bool te::qt::widgets::SymbolSelectorDialog::eraseGroup(const QString& groupId)
{
  te::qt::widgets::SymbolLibrary* symbolLibrary = SymbolLibraryManager::getInstance().getLibrary();
  if (symbolLibrary == nullptr)
  {
    return false;
  }

  std::string id = Convert2StdString(groupId);

  SymbolNode* group = symbolLibrary->findGroupById(id);
  if (group)
  {
    symbolLibrary->eraseGroup(group);
    
    return true;
  }

  return false;
}

QTreeWidgetItem* te::qt::widgets::SymbolSelectorDialog::getSelectedItem() const
{
  QList<QTreeWidgetItem*> selected = m_ui->m_symbolLibraryTreeWidget->selectedItems();

  if(selected.empty())
    return nullptr;

  return selected.at(0);
}

void te::qt::widgets::SymbolSelectorDialog::enableSymbolItem(bool enable)
{
  m_ui->m_editSymbolPushButton->setEnabled(enable);
  m_ui->m_removeSymbolPushButton->setEnabled(enable);
}

void te::qt::widgets::SymbolSelectorDialog::enableGroupItem(bool enable)
{
  m_ui->m_renameGroupPushButton->setEnabled(enable);
  m_ui->m_removeGroupPushButton->setEnabled(enable);
}

QTreeWidgetItem* te::qt::widgets::SymbolSelectorDialog::getCurrentGroupItem()
{
  QTreeWidgetItem* item = getSelectedItem();

  if (item == nullptr)
  {
    return item;
  }

  if (item->type() == SYMBOL)
  {
    item = searchParentGroup(item);
  }

  return item;
}

QTreeWidgetItem* te::qt::widgets::SymbolSelectorDialog::searchParentGroup(QTreeWidgetItem* item)
{
  if (item->parent() != nullptr)
  {
    if (item->parent()->type() == GROUP)
    {
      return item->parent();
    }
  }
  else
  {
    return nullptr;
  }

  return searchParentGroup(item->parent());
}

QString te::qt::widgets::SymbolSelectorDialog::getSelectedItemID()
{
  QString id = "";
  QTreeWidgetItem* item = getSelectedItem();
  if (item)
  {
    id = item->data(0, Qt::UserRole).toString();
  }

  return id;
}

QString te::qt::widgets::SymbolSelectorDialog::getCurrentGroupItemID()
{
  QString id = "";
  QTreeWidgetItem* item = getCurrentGroupItem();
  if (item)
  {
    id = item->data(0, Qt::UserRole).toString();
  }

  return id;
}

bool te::qt::widgets::SymbolSelectorDialog::isRootItem(QTreeWidgetItem* item)
{
  te::qt::widgets::SymbolLibrary* symbolLibrary = SymbolLibraryManager::getInstance().getLibrary();
  if (symbolLibrary == nullptr)
  {
    return false;
  }
    
  if (!item)
  {
    return false;
  }

  if (item->type() != GROUP)
  {
    return false;
  }

  QString itemId = item->data(0, Qt::UserRole).toString();
  std::string id = Convert2StdString(itemId);

  std::string rootId = symbolLibrary->getRoot()->getID();

  if (id.compare(rootId) == 0)
  {
    return true;
  }

  return false;
}

void te::qt::widgets::SymbolSelectorDialog::onDragEnterEvent()
{
  m_currentDragDropItem = getSelectedItem();
}

void te::qt::widgets::SymbolSelectorDialog::onDropEvent()
{
  te::qt::widgets::SymbolLibrary* symbolLibrary = SymbolLibraryManager::getInstance().getLibrary();

  if (symbolLibrary == nullptr
    || m_currentDragDropItem == nullptr)
  {
    return;
  }

  // drop item
  QString id = m_currentDragDropItem->data(0, Qt::UserRole).toString();
  std::string currentId = Convert2StdString(id);
  
  // to parent item
  QString newParentId = "";
  std::string toParentId = "";
  QTreeWidgetItem* parent = m_currentDragDropItem->parent();
  if (parent)
  {
    newParentId = parent->data(0, Qt::UserRole).toString();
    toParentId = Convert2StdString(newParentId);
  }
  
  if (m_currentDragDropItem->type() == GROUP)
  {
    symbolLibrary->swapGroup(currentId, toParentId);
  }
  else if(m_currentDragDropItem->type() == SYMBOL)
  {
    symbolLibrary->swapSymbol(currentId, toParentId);
  }  
}

