/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/widgets/se/SymbolSelectorDialog.h

  \brief A dialog used to select a specific symbol.
*/

#ifndef __TERRALIB_QT_WIDGETS_SE_INTERNAL_SYMBOLSELECTORDIALOG_H
#define __TERRALIB_QT_WIDGETS_SE_INTERNAL_SYMBOLSELECTORDIALOG_H

// TerraLib
#include "../Config.h"

// Qt
#include <QDialog>
#include <QStringListIterator>

// STL
#include <memory>
#include <string>
#include <vector>

// Forward declaraion
class QTreeWidgetItem;
class QTreeWidget;
class QString;

namespace Ui
{
  class SymbolSelectorDialogForm;
}

namespace te
{
  // Forward declarations
  namespace se
  {
    class Rule;
  }
  namespace qt
  {
    namespace widgets
    {
      // Forward declarations
      class  Symbol;
      struct SymbolInfo;
      class  SymbolPreviewWidget;
      class  SymbolNode;

      /*!
        \class SymbolSelectorDialog

        \brief A dialog used to select a specific symbol.
      */
      class TEQTWIDGETSEXPORT SymbolSelectorDialog : public QDialog
      {
        Q_OBJECT

        private:

          /*!
            \enum NodeType

            \brief Auxiliary internal enumeration to control the tree nodes.
          */
          enum NodeType
          {
            GROUP = 0, /*!< Library root node. */
            SYMBOL  = 1,  /*!< Symbol node.       */
          };

        public:

          /** @name Initializer Methods
           *  Methods related to instantiation and destruction.
           */
          //@{

          /*! 
          \brief Constructs a symbol selector dialog which is a child of parent, with widget flags set to f. 

          \note field names from dataset for text symbolizer (no geometry)
          
          */
         SymbolSelectorDialog(const std::vector<std::string>& fieldsNames, QWidget* parent = 0, Qt::WindowFlags f = 0);

          /*! \brief Destructor. */
          virtual ~SymbolSelectorDialog();

          /*! \brief Auxiliary internal method that returns the selected symbol.
           */
          Symbol* getSelectedSymbol() const;

          /*!
          \brief Gets the selected symbol.

          \return The selected fill element.       

            \note The caller will take the ownership of the returned symbol.
            \note It return a NULL pointer if there is not a selected symbol.
          */
          Symbol* getSymbol() const;
                    
        private slots:

          void onCurrentItemChanged(QTreeWidgetItem* current, QTreeWidgetItem* previous);

          void onShowSymbolInfoPushButtonPressed();
          
          void onImportButtonPressed();

          void onExportButtonPressed();

          void onSearchLineEditTextChanged(const QString& text);

          void onCreateSymbolPushButtonClicked();

          void onEditSymbolPushButtonClicked();

          void onRemoveSymbolPushButtonClicked();

           void onCreateGroupPushButtonClicked();

          void onRenameGroupPushButtonClicked();

          void onRemoveGroupPushButtonClicked();

          void onOkPushButtonClicked();

          void onClosePushButtonClicked();

          void onHelpPushButtonClicked();
          
          void onDropEvent();

          void onDragEnterEvent();

        private:

          /*! \brief Initialize the dialog. */
          void initialize();

          /*! \brief  Auxiliary internal method to filter the symbols. */
          void filter(const QList<QTreeWidgetItem*>& items);

          /*! \brief Auxiliary internal method that returns the symbol given a list widget item. */
          Symbol* getSymbolFromItem(QTreeWidgetItem* item) const;

          /*! \brief Auxiliary internal method to format a symbol info to be used on tool tips. */
          QString formatSymbolInfo(const SymbolInfo& info) const;

          /*! \brief Add new symbol to a symbol library. */
          bool addSymbol(Symbol* symbol, const QString& groupParentId);

          /*! 
          \brief Update the current selected symbol. 
          
            \note The symbol with the old state will be removed, and a new one with the same id and name will be added to the library. 
          */
          bool updateCurrentSymbol(Symbol* symbol);

          /*! 
          \brief Remove a symbol from a symbol library . 
          
            \note The symbol is removed, but the pointer is not deleted.
          */
          bool removeSymbol(Symbol* symbol);
                    
          /*! 
          \brief Remove a symbol from a symbol library . 
          
            \note The symbol is removed, but the pointer is not deleted.
          */
          bool eraseSymbol(Symbol* symbol);

          /*! 
          \brief Remove the current selected symbol from a symbol library. 
          
            \note The symbol is removed, but the pointer is not deleted.
          */
          bool eraseCurrentSymbol();

          /*! \brief Add new symbol to a symbol library. */
          bool addGroup(const QString& groupName, const QString& groupParentId);

          /*! 
          \brief Update the current selected symbol. 
          
            \note The symbol with the old state will be removed, and a new one with the same id and name will be added to the library. 
          */
          bool updateCurrentGroup(const QString& newGroupName, const QString& groupId);
          
          /*! 
          \brief Remove a symbol from a symbol library . 
          
            \note The symbol is removed, but the pointer is not deleted.
          */
          bool eraseGroup(const QString& groupId);

          /*! 
          \brief Remove the current selected symbol from a symbol library. 
          
            \note The symbol is removed, but the pointer is not deleted.
          */
          bool eraseCurrentGroup();
                    
          /*! 
          \brief Get the current selected symbol. 
          */
          QTreeWidgetItem* getSelectedItem() const;
          
          void buildTree(SymbolNode* node, QTreeWidgetItem* parent);

          QTreeWidgetItem* createItemFromGroup(SymbolNode* node, QTreeWidgetItem* parent = nullptr);

          QTreeWidgetItem* createItemFromGroup(SymbolNode* node, QTreeWidget* parent);

          QTreeWidgetItem* createItemFromSymbol(Symbol* symbol, QTreeWidgetItem* parent = nullptr);

          QTreeWidgetItem* createItemFromSymbol(Symbol* symbol, QTreeWidget* parent);

          void configItemFromSymbol(Symbol* symbol, QTreeWidgetItem* item);

          void enableSymbolItem(bool enable);

          void enableGroupItem(bool enable);

          QTreeWidgetItem* getCurrentGroupItem();

          QString getCurrentGroupItemID();

          QTreeWidgetItem* searchParentGroup(QTreeWidgetItem* item);

          QString getSelectedItemID();

          bool isRootItem(QTreeWidgetItem* item);
          
        private:

          std::unique_ptr<Ui::SymbolSelectorDialogForm> m_ui; //!< Dialog form.
          SymbolPreviewWidget*                          m_preview; //!< Preview Widget used to visualize the symbol.
          std::vector<std::string>                      m_fieldsNames; //!< Dataset fields names for text symbolizer (no geometry)
          QTreeWidgetItem*                              m_currentDragDropItem;
      }; 

    } // end namespace widgets
  }   // end namespace qt
}     // end namespace te

#endif  // __TERRALIB_QT_WIDGETS_SE_INTERNAL_SYMBOLSELECTORDIALOG_H
