/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled
   applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/widgets/se/SymbolSelectorDialog.cpp

  \brief A dialog used to select a specific symbol.
*/

// TerraLib
#include "../../../common/Exception.h"
#include "Symbol.h"
#include "Utils.h"
#include "../../../dataaccess/dataset/DataSet.h"
#include "../../../dataaccess/dataset/DataSetType.h"
#include "../../../maptools/AbstractLayer.h"
#include "../../../se/Symbolizer.h"
#include "../../../se/PointSymbolizer.h"
#include "../../../se/LineSymbolizer.h"
#include "../../../se/PolygonSymbolizer.h"
#include "../../../se/TextSymbolizer.h"
#include "../../../se/Utils.h"
#include "../../../se/ParameterValue.h"
#include "../Utils.h"
#include "SymbolizerEditInfoDialog.h"
#include "PointSymbolizerProperty.h"
#include "LineSymbolizerProperty.h"
#include "PolygonSymbolizerProperty.h"
#include "TextSymbolizerProperty.h"
#include "ui/ui_SymbolizerEditInfoDialogForm.h"

// Qt
#include <QFileDialog>
#include <QMessageBox>

// STL
#include <cassert>

// Boost
#include <boost/uuid/random_generator.hpp>
#include <boost/uuid/uuid_io.hpp>

te::qt::widgets::SymbolizerEditInfoDialog::SymbolizerEditInfoDialog(const std::vector<std::string>& fieldsNames, te::se::Symbolizer* symbolizer, QWidget* parent,
                                                            Qt::WindowFlags f)
  : QDialog(parent, f), m_ui(new Ui::SymbolizerEditInfoDialogForm),
  m_widget(nullptr),
  m_fieldsNames(fieldsNames),
  m_currentSymbolizer(nullptr),
  m_symbolizerType("")
{
  // Setup
  m_ui->setupUi(this);

  m_layoutFrame = new QGridLayout(m_ui->frame);
  m_layoutFrame->setContentsMargins(0,0,0,0);

  if (symbolizer != nullptr)
  {
    m_currentSymbolizer.reset(symbolizer->clone());
  }

  // connect signal/slot
  connect(m_ui->m_cmbSymbols, SIGNAL(currentIndexChanged(const QString &)), this, SLOT(onCmbSymbolsCurrentIndexChanged(const QString &)));
  connect(m_ui->m_okPushButton, SIGNAL(clicked(bool)), this, SLOT(onOkPushButtonClicked()));
  connect(m_ui->m_cancelPushButton, SIGNAL(clicked(bool)), this, SLOT(onCancelPushButtonClicked()));
  connect(m_ui->m_helpPushButton, SIGNAL(clicked(bool)), this, SLOT(onHelpPushButtonClicked()));

  initialize();
}

te::qt::widgets::SymbolizerEditInfoDialog::~SymbolizerEditInfoDialog()
{
  
}

void te::qt::widgets::SymbolizerEditInfoDialog::initialize()
{
  if (m_currentSymbolizer != nullptr)
  {
    loadSymbolizers();
  }
  else
  {
    createInterfaces();
  }

  initializeCombo();
}

void te::qt::widgets::SymbolizerEditInfoDialog::initializeCombo()
{
  if (m_interfaces.empty())
  {
    return;
  }
  
  std::string name = "";
  QWidget* widget = nullptr;

  if (m_currentSymbolizer == nullptr)
  {
    std::pair<QString, QWidget*> p = m_interfaces.begin()->second;
    name = m_interfaces.begin()->first;
    widget = p.second;
  }
  else
  {
    te::se::Symbolizer* symbolizer = m_currentSymbolizer.get();

    std::map<std::string, std::pair<QString, QWidget*>>::iterator it;

    for (it = m_interfaces.begin(); it != m_interfaces.end(); ++it)
    {
      std::pair<QString, QWidget*> p = it->second;
      std::string sText = p.first.toUtf8().constData();

      if (symbolizer->getType().compare(sText) == 0)
      {
        name = it->first;
        widget = p.second;
      }
    }

    m_ui->m_cmbSymbols->setEnabled(false);
  }

  // symbolizer widget type 
  setWidgetOnFrame(widget);

  QString qName = QString::fromUtf8(name.c_str());
  int index = m_ui->m_cmbSymbols->findText(qName);
  m_ui->m_cmbSymbols->setCurrentIndex(index);
}

void te::qt::widgets::SymbolizerEditInfoDialog::setWidgetOnFrame(QWidget* widget)
{
  if (widget == nullptr)
  {
    return;
  }

  removeWidgetFromFrame(widget);

  m_widget = widget;
  m_widget->setParent(this);

  m_layoutFrame->addWidget(m_widget);

  m_widget->setVisible(true);
}

void te::qt::widgets::SymbolizerEditInfoDialog::removeWidgetFromFrame(QWidget* widget)
{
  widget->setVisible(false);
  m_layoutFrame->removeWidget(widget);
}

void te::qt::widgets::SymbolizerEditInfoDialog::loadSymbolizers()
{
  QStringList list;

  if(m_currentSymbolizer)
  {
    te::se::Symbolizer* symbolizer = m_currentSymbolizer.get();
    if (symbolizer->getType() == "PointSymbolizer")
    {
      createPointSymbolizerInterface(symbolizer);
      list.append("Point");
    }
    else if (symbolizer->getType() == "LineSymbolizer")
    {
      createLineSymbolizerInterface(symbolizer);
      list.append("Line");
    }
    else if (symbolizer->getType() == "PolygonSymbolizer")
    {
      createPolygonSymbolizerInterface(symbolizer);
      list.append("Polygon");
    }
    else if (symbolizer->getType() == "TextSymbolizer")
    {
      createTextSymbolizerInterface(symbolizer);
      list.append("Text");
    }
  }  

  m_ui->m_cmbSymbols->blockSignals(true);
  m_ui->m_cmbSymbols->addItems(list);
  m_ui->m_cmbSymbols->blockSignals(false);
}

void te::qt::widgets::SymbolizerEditInfoDialog::createInterfaces()
{
  QStringList list;

  createPointSymbolizerInterface();
  list.append("Point");

  createLineSymbolizerInterface();
  list.append("Line");

  createPolygonSymbolizerInterface();
  list.append("Polygon");

  createTextSymbolizerInterface();
  list.append("Text");

  m_ui->m_cmbSymbols->blockSignals(true);
  m_ui->m_cmbSymbols->addItems(list);
  m_ui->m_cmbSymbols->blockSignals(false);
}

void te::qt::widgets::SymbolizerEditInfoDialog::createPointSymbolizerInterface(te::se::Symbolizer* symbolizer)
{
  // PointSymbolizer Property
  te::qt::widgets::PointSymbolizerProperty* pointSymbolizerProperty = new te::qt::widgets::PointSymbolizerProperty(this);
  pointSymbolizerProperty->setVisible(false);
  pointSymbolizerProperty->setObjectName("PointSymbolizer");
    
  m_interfaces["Point"] = std::make_pair("PointSymbolizer", pointSymbolizerProperty);
  connect(pointSymbolizerProperty, SIGNAL(symbolizerChanged()), this, SLOT(onSymbolizerChanged()));
    
  pointSymbolizerProperty->setSizePolicy(QSizePolicy::Preferred, QSizePolicy::Expanding);

  // We must have a valid symbolizer for changes in the widget take effect.
  if(symbolizer == nullptr)
  {
    symbolizer = te::se::CreateSymbolizer(te::gm::PointType);
  }

  pointSymbolizerProperty->setSymbolizer(dynamic_cast<te::se::PointSymbolizer*>(symbolizer));
}

void te::qt::widgets::SymbolizerEditInfoDialog::createLineSymbolizerInterface(te::se::Symbolizer* symbolizer)
{
  // LineSymbolizer Property
  te::qt::widgets::LineSymbolizerProperty* lineSymbolizerProperty = new te::qt::widgets::LineSymbolizerProperty(this);
  lineSymbolizerProperty->setVisible(false);
  lineSymbolizerProperty->setObjectName("LineSymbolizer");
    
  m_interfaces["Line"] = std::make_pair("LineSymbolizer", lineSymbolizerProperty);
  connect(lineSymbolizerProperty, SIGNAL(symbolizerChanged()), this, SLOT(onSymbolizerChanged()));
    
  lineSymbolizerProperty->setSizePolicy(QSizePolicy::Preferred, QSizePolicy::Expanding);

  // We must have a valid symbolizer for changes in the widget take effect.
  if(symbolizer == nullptr)
  {
    symbolizer = te::se::CreateSymbolizer(te::gm::LineStringType);
  }

  lineSymbolizerProperty->setSymbolizer(dynamic_cast<te::se::LineSymbolizer*>(symbolizer));
}

void te::qt::widgets::SymbolizerEditInfoDialog::createPolygonSymbolizerInterface(te::se::Symbolizer* symbolizer)
{
  // PolygonSymbolizer Property
  te::qt::widgets::PolygonSymbolizerProperty* polygonSymbolizerProperty = new te::qt::widgets::PolygonSymbolizerProperty(this);
  polygonSymbolizerProperty->setVisible(false);
  polygonSymbolizerProperty->setObjectName("PolygonSymbolizer");
    
  m_interfaces["Polygon"] = std::make_pair("PolygonSymbolizer", polygonSymbolizerProperty);
  connect(polygonSymbolizerProperty, SIGNAL(symbolizerChanged()), this, SLOT(onSymbolizerChanged()));
    
  polygonSymbolizerProperty->setSizePolicy(QSizePolicy::Preferred, QSizePolicy::Expanding);

  // We must have a valid symbolizer for changes in the widget take effect.
  if(symbolizer == nullptr)
  {
    symbolizer = te::se::CreateSymbolizer(te::gm::PolygonType);
  }

  polygonSymbolizerProperty->setSymbolizer(dynamic_cast<te::se::PolygonSymbolizer*>(symbolizer));
}

void te::qt::widgets::SymbolizerEditInfoDialog::createTextSymbolizerInterface(te::se::Symbolizer* symbolizer)
{
  // TextSymbolizer Property
  te::qt::widgets::TextSymbolizerProperty* textSymbolizerProperty = new te::qt::widgets::TextSymbolizerProperty(this);
  textSymbolizerProperty->setVisible(false);
  textSymbolizerProperty->setObjectName("TextSymbolizer");
    
  m_interfaces["Text"] = std::make_pair("TextSymbolizer", textSymbolizerProperty);
  connect(textSymbolizerProperty, SIGNAL(symbolizerChanged()), this, SLOT(onSymbolizerChanged()));
    
  textSymbolizerProperty->setSizePolicy(QSizePolicy::Preferred, QSizePolicy::Expanding);

  // We must have a valid symbolizer for changes in the widget take effect.
  if(symbolizer == nullptr)
  {
    symbolizer = createTextSymbolizer();
  }

  textSymbolizerProperty->setSymbolizer(dynamic_cast<te::se::TextSymbolizer*>(symbolizer)); 

  initializeTextSymbolizerLabels(textSymbolizerProperty);
}

void te::qt::widgets::SymbolizerEditInfoDialog::onSymbolizerChanged()
{
  
}

void te::qt::widgets::SymbolizerEditInfoDialog::onCmbSymbolsCurrentIndexChanged (const QString & text)
{
  QWidget* newWidget = m_widget;

  removeWidgetFromFrame(m_widget);
  
  std::map<std::string, std::pair<QString, QWidget*>>::iterator it;

  for (it = m_interfaces.begin(); it != m_interfaces.end(); ++it)
  {
    std::string key = it->first;
    std::pair<QString, QWidget*> p = it->second;
    std::string sText = text.toUtf8().constData();

    if (sText.compare(key) == 0)
    {
      newWidget = p.second;
    }
  }

  setWidgetOnFrame(newWidget);
}

void te::qt::widgets::SymbolizerEditInfoDialog::onOkPushButtonClicked()
{
  te::se::Symbolizer* symbol = createSymbolizer();

  m_currentSymbolizer.reset(symbol);

  accept(); 
}

void te::qt::widgets::SymbolizerEditInfoDialog::onCancelPushButtonClicked()
{
  reject();
}

void te::qt::widgets::SymbolizerEditInfoDialog::onHelpPushButtonClicked()
{

}

te::se::Symbolizer* te::qt::widgets::SymbolizerEditInfoDialog::createSymbolizer()
{
  te::se::Symbolizer* symbolizer = m_currentSymbolizer.release();
  
  if (m_widget == nullptr)
  {
    return symbolizer;
  }
  else
  {
    AbstractSymbolizerProperty* symbolizerProperty = dynamic_cast<AbstractSymbolizerProperty*>(m_widget);
    symbolizer = symbolizerProperty->getSymbolizer();
  }
        
  return symbolizer;
}

te::se::Symbolizer* te::qt::widgets::SymbolizerEditInfoDialog::getSymbolizer()
{
  return m_currentSymbolizer->clone();
}

void te::qt::widgets::SymbolizerEditInfoDialog::initializeTextSymbolizerLabels(te::qt::widgets::TextSymbolizerProperty* textSymbolizer)
{
  textSymbolizer->setLabels(m_fieldsNames);

  te::se::Symbolizer* symbol = textSymbolizer->getSymbolizer();
  te::se::TextSymbolizer* symbText = dynamic_cast<te::se::TextSymbolizer*>(symbol);

  if (symbText)
  {
    if (!symbText->getLabel() || te::se::GetString(symbText->getLabel()) == "")
    {
      if (!m_fieldsNames.empty())
      {
        symbText->setLabel(new te::se::ParameterValue(m_fieldsNames[0]));
      }
    }
  }
}

te::se::TextSymbolizer* te::qt::widgets::SymbolizerEditInfoDialog::createTextSymbolizer()
{
  //creates new text symbolizer representation
  te::se::Fill* fill = te::se::CreateFill("#000000", "1.0");
  te::se::Font* font = te::se::CreateFont("Arial", "8", te::se::Font::StyleNormal, te::se::Font::WeightNormal);
  te::se::TextSymbolizer* symbolizer = te::se::CreateTextSymbolizer("", fill, font);
  return symbolizer;
}

void te::qt::widgets::SymbolizerEditInfoDialog::setSymbolizerType(const std::string& symbolizerType)
{
  m_symbolizerType = symbolizerType;

  if (m_symbolizerType.compare("") != 0)
  {
    std::string type = getGeomType(symbolizerType);

    QString qName = Convert2QString(type);
    int index = m_ui->m_cmbSymbols->findText(qName);

    if (index != -1)
    {
      m_ui->m_cmbSymbols->setCurrentIndex(index);
    }
  }
}

std::string te::qt::widgets::SymbolizerEditInfoDialog::getGeomType(const std::string& symbolizerType)
{
  std::string type = "Line";

  if (symbolizerType.compare("PointSymbolizer") == 0)
  {
    type = "Point";
  }
  else if (symbolizerType.compare("LineSymbolizer") == 0)
  {
    type = "Line";
  }
  else if (symbolizerType.compare("PolygonSymbolizer") == 0)
  {
    type = "Polygon";
  }
  else if (symbolizerType.compare("TextSymbolizer") == 0)
  {
    type = "Text";
  }

  return type;
}


