/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/widgets/se/SymbolizerEditInfoDialog.h

  \brief A dialog used to create or edit a specific symbol.
*/

#ifndef __TERRALIB_QT_WIDGETS_SE_INTERNAL_SYMBOLIZER_EDITINFO_WIDGET_H
#define __TERRALIB_QT_WIDGETS_SE_INTERNAL_SYMBOLIZER_EDITINFO_WIDGET_H

// TerraLib
#include "../Config.h"

// Qt
#include <QWidget>
#include <QMap>
#include <QDialog>

// STL
#include <memory>
#include <map>
#include <utility>
#include <vector>
#include <string>

class QString;
class QGridLayout;

namespace Ui
{
  class SymbolizerEditInfoDialogForm;
}

namespace te
{
  // Forward declarations
  namespace se
  {
    class Rule;
    class TextSymbolizer;
  }
  namespace qt
  {
    namespace widgets
    {
      // Forward declarations
      class  Symbol;
      struct SymbolInfo;
      class  SymbolPreviewWidget;
      class  Symbolizer;
      class  TextSymbolizerProperty;

      /*!
        \class SymbolizerEditInfoDialog

        \brief A dialog used to create or edit a specific symbol.
      */
      class TEQTWIDGETSEXPORT SymbolizerEditInfoDialog : public QDialog
      {
        Q_OBJECT

        public:

          /** @name Initializer Methods
           *  Methods related to instantiation and destruction.
           */
          //@{

          /*! 
          \brief Constructs a symbol selector dialog which is a child of parent, with widget flags set to f. 

          \note field names from dataset for text symbolizer (no geometry)
          */
          SymbolizerEditInfoDialog(const std::vector<std::string>& fieldsNames, te::se::Symbolizer* symbolizer = nullptr, QWidget* parent = nullptr, Qt::WindowFlags f = nullptr);

          /*! \brief Destructor. */
          virtual ~SymbolizerEditInfoDialog();

          /*!
          \brief Gets the selected symbol.

          \return The selected fill element.       

            \note The caller will take the ownership of the returned symbol.
            \note It return a NULL pointer if there is not a symbol yet created.
          */
          virtual te::se::Symbolizer* getSymbolizer();

          virtual void setSymbolizerType(const std::string& symbolizerType);
                              
        protected slots:
          
          virtual void onCmbSymbolsCurrentIndexChanged ( const QString & text );

          virtual void onSymbolizerChanged();

          void onOkPushButtonClicked();

          void onCancelPushButtonClicked();

          void onHelpPushButtonClicked();

        protected:

          /*! \brief Initialize symbol configuration frame. */
          virtual void initialize();
          
          /*! \brief Initialize symbol options combobox. */
          virtual void initializeCombo();

          virtual void initializeTextSymbolizerLabels(te::qt::widgets::TextSymbolizerProperty* textSymbolizer);
          
          /*! \brief Set a symbol configuration widget inside a frame */
          virtual void setWidgetOnFrame(QWidget* widget);

          /*! \brief Remove a symbol configuration widget from the frame */
          virtual void removeWidgetFromFrame(QWidget* widget);

          /*! \brief If the symbol already exists, only the symbol configuration widget to which it belongs will be created. */
          virtual void loadSymbolizers();

          /*! \brief Will create the symbol configuration widgets of various types. */
          virtual void createInterfaces();

          /*! \brief Will create the symbol configuration widget for Point type. */
          virtual void createPointSymbolizerInterface(te::se::Symbolizer* symbolizer = nullptr);

          /*! \brief Will create the symbol configuration widget for Line type. */
          virtual void createLineSymbolizerInterface(te::se::Symbolizer* symbolizer = nullptr);

          /*! \brief Will create the symbol configuration widget for Polygon type. */
          virtual void createPolygonSymbolizerInterface(te::se::Symbolizer* symbolizer = nullptr);

          /*! \brief Will create the symbol configuration widget for Text type. */
          virtual void createTextSymbolizerInterface(te::se::Symbolizer* symbolizer = nullptr);

          /*! \brief Will create a new symbol */
          virtual te::se::Symbolizer* createSymbolizer();
          
          virtual te::se::TextSymbolizer* createTextSymbolizer();

          virtual std::string getGeomType(const std::string& symbolizerType);
                                              
        private:

          std::unique_ptr<Ui::SymbolizerEditInfoDialogForm>       m_ui; //!< Dialog form.
          QWidget*                                                m_widget; //!< Current symbol configuration widget
          std::map<std::string, std::pair<QString, QWidget*>>     m_interfaces; //!< symbol configuration widget map
          QGridLayout*                                            m_layoutFrame; //!< Layout, within a frame, which will contain the symbol configuration widget
          std::vector<std::string>                                m_fieldsNames; //!< Dataset fields names for text symbolizer
          std::unique_ptr<te::se::Symbolizer>                     m_currentSymbolizer; //!< current symbolizer
          std::string                                             m_symbolizerType;
      }; 

    } // end namespace widgets
  }   // end namespace qt
}     // end namespace te

#endif  // __TERRALIB_QT_WIDGETS_SE_INTERNAL_SYMBOLIZER_EDITINFO_WIDGET_H
