/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
\file terralib/qt/widgets/se/TextPropertyItem.cpp

    \brief A widget used to define the text se object.
*/
#include "TextPropertyItem.h"

#include "../../../common/StringUtils.h"
#include "../../../se/SvgParameter.h"
#include "../../../se/TextSymbolizer.h"
#include "../../../se/Utils.h"
#include "../propertybrowser/AbstractPropertyManager.h"
#include "../Utils.h"
#include "BasicFillPropertyItem.h"
#include "Utils.h"

// STL
#include <cassert>

// QtPropertyBrowser


te::qt::widgets::TextPropertyItem::TextPropertyItem(QtTreePropertyBrowser* pb, StylePropertyItemListener* listener, QColor c)
  : te::qt::widgets::CheckboxStylePropertyItem(pb, tr("Text Properties"), listener, c)
  , m_basicFillProperty(nullptr)
  , m_labelProperty(nullptr)
  , m_fontProperty(nullptr)
{
  //add property to tree propertybrowser
  QtProperty* itemProperty = this->getItemProperty();
  addProperty(itemProperty, QColor(89, 230, 226));

  //label
  QStringList emptyList;
  m_labelProperty = te::qt::widgets::AbstractPropertyManager::getInstance().m_enumManager->addProperty(tr("Label"));
  addListenerToProperty(m_labelProperty);
  te::qt::widgets::AbstractPropertyManager::getInstance().m_enumManager->setEnumNames(m_labelProperty, emptyList);
  
  //font
  m_fontProperty = te::qt::widgets::AbstractPropertyManager::getInstance().m_fontManager->addProperty(tr("Font"));
  addListenerToProperty(m_fontProperty);
 
  itemProperty->addSubProperty(m_labelProperty);
  itemProperty->addSubProperty(m_fontProperty);

  m_basicFillProperty = new te::qt::widgets::BasicFillPropertyItem(itemProperty, m_propertyBrowser, listener, QColor(89, 230, 226));

  // Setups initial fill
  setTextSymbolizer(nullptr);
}

te::qt::widgets::TextPropertyItem::~TextPropertyItem()
{
  delete m_basicFillProperty;
  delete m_labelProperty;
  delete m_fontProperty;
}

void te::qt::widgets::TextPropertyItem::setTextSymbolizer(const te::se::TextSymbolizer* symb)
{
  //read the properties of the symbolizer
  std::string label = "";
  const te::se::Font* font = nullptr;
  const te::se::Fill* fill = nullptr;
  if (symb != nullptr)
  {
    const te::se::TextSymbolizer* textSymbolizer = dynamic_cast<const te::se::TextSymbolizer*>(symb);
    if (textSymbolizer != nullptr)
    {
      const te::se::ParameterValue* labelParameter = textSymbolizer->getLabel();
      if (labelParameter != nullptr)
      {
        label = te::se::GetString(labelParameter);
      }

      font = textSymbolizer->getFont();
      fill = textSymbolizer->getFill();
    }
  }

  m_enableListeners = false;

  //Label
  QString qValue = Convert2QString(label);
  AbstractPropertyItem::setValueToEnumProperty(m_labelProperty, qValue);

  //Font
  QFont qFont;
  qFont.setPointSize(TE_SE_DEFAULT_FONT_SIZE);
  if (font != nullptr)
  {
    qFont = Convert2Qt(font);
  }
  te::qt::widgets::AbstractPropertyManager::getInstance().m_fontManager->setValue(m_fontProperty, qFont);

  //Fill
  m_basicFillProperty->setFill(fill);

  m_enableListeners = true;
}

te::se::TextSymbolizer* te::qt::widgets::TextPropertyItem::getTextSymbolizer() const
{
  te::se::TextSymbolizer* newItem = new te::se::TextSymbolizer();
  
  //Label
  QString qLabel = AbstractPropertyItem::getValueFromEnumProperty(m_labelProperty);
  std::string label = Convert2StdString(qLabel);
  newItem->setLabel(new te::se::ParameterValue(label));

  //Font
  QFont qFont = te::qt::widgets::AbstractPropertyManager::getInstance().m_fontManager->value(m_fontProperty);
  te::se::Font* font = Convert2TerraLib(qFont);
  newItem->setFont(font);

  //Fill
  newItem->setFill(m_basicFillProperty->getFill());

  return newItem;
}

void te::qt::widgets::TextPropertyItem::setLabels(const std::vector<std::string>& labels)
{
  QStringList list;

  for (std::size_t i = 0; i < labels.size(); ++i)
  {
    list << labels[i].c_str();
  }

  m_enableListeners = false;
  te::qt::widgets::AbstractPropertyManager::getInstance().m_enumManager->setEnumNames(m_labelProperty, list);
  m_enableListeners = true;
}
