/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
\file terralib/qt/widgets/se/TextSymbolizerProperty.cpp

      \brief A widget used to define the text symbolizer
      properties, such as:
      - basic fill
      - font
      - halo
*/

// Terralib
#include "../../../se/Halo.h"
#include "../../../se/Fill.h"
#include "../../../se/Font.h"
#include "../../../se/LabelPlacement.h"
#include "../../../se/LinePlacement.h"
#include "../../../se/ParameterValue.h"
#include "../../../se/PointPlacement.h"
#include "../../../se/TextSymbolizer.h"
#include "../propertybrowser/AbstractPropertyManager.h"
#include "GroupPropertyItemHandler.h"
#include "HaloPropertyItem.h"
#include "LinePlacementPropertyItem.h"
#include "PointPlacementPropertyItem.h"
#include "TextPropertyItem.h"
#include "TextSymbolizerProperty.h"

// Qt
#include <QGridLayout>
#include <QToolBox>

// QtPropertyBrowser
#include <QtPropertyBrowser/QtTreePropertyBrowser>
#include <QtPropertyBrowser/QtVariantPropertyManager>

// STL
#include <cassert>

te::qt::widgets::TextSymbolizerProperty::TextSymbolizerProperty(QWidget* parent)
  : AbstractSymbolizerProperty(parent)
  , StylePropertyItemListener()
  , m_groupHandler(new GroupPropertyItemHandler())
  , m_text(nullptr)
  , m_pp(nullptr)
  , m_lp(nullptr)
  , m_halo(nullptr)
{
  QGridLayout* layout = new QGridLayout(this);
  this->setLayout(layout);

  QToolBox* tb = new QToolBox(this);
  layout->addWidget(tb);

  QtTreePropertyBrowser* basicPropBrowser = new QtTreePropertyBrowser(this);
  basicPropBrowser->setIndentation(10);
  basicPropBrowser->setPropertiesWithoutValueMarked(true);
  basicPropBrowser->setRootIsDecorated(false);
  basicPropBrowser->setResizeMode(QtTreePropertyBrowser::ResizeToContents);

  layout->addWidget(basicPropBrowser);

  m_text = new te::qt::widgets::TextPropertyItem(basicPropBrowser, this);
  
  m_pp = new te::qt::widgets::PointPlacementPropertyItem(basicPropBrowser, this);

  m_lp = new te::qt::widgets::LinePlacementPropertyItem(basicPropBrowser, this);

  m_halo = new te::qt::widgets::HaloPropertyItem(basicPropBrowser, this);

  m_groupHandler->addPropertyToGroup(m_pp);
  m_groupHandler->addPropertyToGroup(m_lp);

  tb->addItem(basicPropBrowser, "Text Symbology");
}

te::qt::widgets::TextSymbolizerProperty::~TextSymbolizerProperty()
{
  delete m_groupHandler;
  delete m_text;
  delete m_pp;
  delete m_lp;
  delete m_halo;
}

void te::qt::widgets::TextSymbolizerProperty::setSymbolizer(const te::se::Symbolizer* symb)
{
  const te::se::TextSymbolizer* textSymbolyzer = nullptr;
  const te::se::LabelPlacement* labelPlacement = nullptr;
  const te::se::PointPlacement* pointPlacement = nullptr;
  const te::se::LinePlacement* linePlacement = nullptr;
  const te::se::Halo* halo = nullptr;

  if (symb != nullptr)
  {
    textSymbolyzer = dynamic_cast<const te::se::TextSymbolizer*>(symb);
    if (textSymbolyzer != nullptr)
    {
      labelPlacement = textSymbolyzer->getLabelPlacement();
      if (labelPlacement != nullptr)
      {
        pointPlacement = labelPlacement->getPointPlacement();
        linePlacement = labelPlacement->getLinePlacement();
      }

      halo = textSymbolyzer->getHalo();
    }
  }

  //Basic text symbolizer
  m_text->setTextSymbolizer(textSymbolyzer);
  m_text->setChecked(textSymbolyzer != nullptr);

  //Point Placement
  m_pp->setPointPlacement(pointPlacement);
  m_pp->setChecked(pointPlacement != nullptr);

  m_lp->setLinePlacement(linePlacement);
  m_lp->setChecked(linePlacement != nullptr);

  m_halo->setHalo(halo);
  m_halo->setChecked(halo != nullptr);
}

te::se::Symbolizer* te::qt::widgets::TextSymbolizerProperty::getSymbolizer() const
{
  te::se::PointPlacement* pointPlacement = nullptr;
  te::se::LinePlacement* linePlacement = nullptr;
  te::se::LabelPlacement* labelPlacement = nullptr;
  te::se::Halo* halo = nullptr;

  if (m_pp->isChecked() == true)
  {
    pointPlacement = m_pp->getPointPlacement();
  }
  if (m_lp->isChecked() == true)
  {
    linePlacement = m_lp->getLinePlacement();
  }
  if (m_halo->isChecked() == true)
  {
    halo = m_halo->getHalo();
  }

  if (pointPlacement != nullptr || linePlacement != nullptr)
  {
    labelPlacement = new te::se::LabelPlacement();
    labelPlacement->setPointPlacement(pointPlacement);
    labelPlacement->setLinePlacement(linePlacement);
  }

  te::se::TextSymbolizer* newItem = m_text->getTextSymbolizer();
  newItem->setLabelPlacement(labelPlacement);
  newItem->setHalo(halo);

  return newItem;
}

void te::qt::widgets::TextSymbolizerProperty::setLabels(const std::vector<std::string>& names)
{
  m_text->setLabels(names);
}

void te::qt::widgets::TextSymbolizerProperty::handleValueChanged(QtProperty* /*qProperty*/, const QVariant& /*qVariant*/)
{
  emit symbolizerChanged();
}