/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/widgets/se/Utils.cpp
   
  \brief This file contains utility functions for dealing with Symbology Enconding and Qt concepts.
 */

// TerraLib
#include "../../../color/RGBAColor.h"
#include "../../../maptools/Utils.h"
#include "../../../se/serialization/xml/Symbolizer.h"
#include "../../../se/Font.h"
#include "../../../se/Stroke.h"
#include "../../../se/SvgParameter.h"
#include "../../../se/Utils.h"
#include "../../../se/Symbolizer.h"
#include "../../../xml/Reader.h"
#include "../../../xml/ReaderFactory.h"
#include "../../../xml/AbstractWriter.h"
#include "../../../xml/AbstractWriterFactory.h"
#include "../Exception.h"
#include "../Utils.h"
#include "Symbol.h"
#include "SymbolLibrary.h"
#include "SymbolLibraryManager.h"
#include "Utils.h"
#include "SymbolNode.h"

// Boost
#include <boost/format.hpp>
#include <boost/uuid/random_generator.hpp>
#include <boost/uuid/uuid_io.hpp>

// Qt
#include <QtCore/QVector>
#include <QBrush>
#include <QFont>
#include <QPen>

// STL
#include <cassert>

void te::qt::widgets::Config(QPen& pen, const te::se::Stroke* stroke)
{
  if(stroke == nullptr)
  {
    pen.setColor(Qt::transparent);
    return;
  }

  // Color
  te::color::RGBAColor rgba(TE_SE_DEFAULT_STROKE_BASIC_COLOR, TE_OPAQUE);
  te::se::GetColor(stroke, rgba);
  QColor qrgba(rgba.getRgba());
  qrgba.setAlpha(rgba.getAlpha());
  pen.setColor(qrgba);

  // Width
  const te::se::SvgParameter* width = stroke->getWidth();
  if(width)
    pen.setWidth(te::se::GetInt(width));

  // Cap Style
  const te::se::SvgParameter* linecap = stroke->getLineCap();
  if(linecap)
  {
    std::string capValue = te::se::GetString(linecap);

    Qt::PenCapStyle capStyle = Qt::FlatCap;
    capValue == TE_SE_ROUND_CAP ? capStyle = Qt::RoundCap : capStyle = Qt::SquareCap;

    pen.setCapStyle(capStyle);
  }
  
  // Join Style
  const te::se::SvgParameter* linejoin = stroke->getLineJoin();
  if(linejoin)
  {
    std::string joinValue = te::se::GetString(linejoin);

    Qt::PenJoinStyle joinStyle = Qt::MiterJoin;
    joinValue == TE_SE_ROUND_JOIN ? joinStyle = Qt::RoundJoin : joinStyle = Qt::BevelJoin;

    pen.setJoinStyle(joinStyle);
  }

  // Dash Style
  const te::se::SvgParameter* dasharray = stroke->getDashArray();
  if(dasharray)
  {
    std::string value = te::se::GetString(dasharray);
    
    std::vector<double> pattern;
    te::map::GetDashStyle(value, pattern);
    
    pen.setDashPattern(QVector<qreal>::fromStdVector(pattern));
  }

  /* TODO: Is necessary verify stroke-dashoffset, <GraphicStroke>, and <GraphicFill> elements here ?! */
}

void te::qt::widgets::Config(QBrush& brush, const te::se::Fill* fill)
{
  if(fill == nullptr)
  {
    brush.setColor(Qt::transparent);
    return;
  }

  te::color::RGBAColor rgba(TE_SE_DEFAULT_FILL_BASIC_COLOR, TE_OPAQUE);
  te::se::GetColor(fill, rgba);
  QColor qrgba(rgba.getRgba());
  qrgba.setAlpha(rgba.getAlpha());
  brush.setColor(qrgba);

  /* TODO: Is necessary verify <GraphicFill> element here ?! */
}

void te::qt::widgets::ReadSymbolLibraryFromBuffer(const std::string& buffer)
{
  std::unique_ptr<te::xml::Reader> reader(te::xml::ReaderFactory::make("XERCES"));
  reader->readFromContent(buffer);

  if (!reader->next())
    throw Exception(TE_TR("Could not read symbol library information from buffer"));

  if (reader->getNodeType() != te::xml::START_ELEMENT)
    throw Exception(TE_TR("Error reading the document. The start element wasn't found."));

  if (reader->getElementLocalName() != "SymbolLibrary")
    throw Exception(TE_TR("The first tag in the document is not 'SymbolLibrary'."));

  std::string name = reader->getAttr("name");

  reader->next();

  SymbolLibrary* library = SymbolLibraryManager::getInstance().getLibrary();

  BuildGroupTreeInsideLibrary(*reader, library);
}


void te::qt::widgets::ReadSymbolLibrary(const std::string& path)
{
  std::unique_ptr<te::xml::Reader> reader(te::xml::ReaderFactory::make("XERCES"));
  reader->read(path);

  if(!reader->next())
    throw Exception((boost::format(TE_TR("Could not read symbol library information in file: %1%.")) % path).str());

  if(reader->getNodeType() != te::xml::START_ELEMENT)
    throw Exception((boost::format(TE_TR("Error reading the document %1%, the start element wasn't found.")) % path).str());

  if(reader->getElementLocalName() != "SymbolLibrary")
    throw Exception((boost::format(TE_TR("The first tag in the document %1% is not 'SymbolLibrary'.")) % path).str());

  std::string name = reader->getAttr("name");

  reader->next();

  SymbolLibrary* library = SymbolLibraryManager::getInstance().getLibrary();

  BuildGroupTreeInsideLibrary(*reader, library);
}

void te::qt::widgets::BuildGroupTreeInsideLibrary(te::xml::Reader& reader, SymbolLibrary* library)
{
  if ((reader.getNodeType() == te::xml::START_ELEMENT) &&
    (reader.getElementLocalName() == "Group"))
  {
    te::qt::widgets::SymbolNode* node = ReadGroup(reader);
    library->addGroup(node);
  }
  else if ((reader.getNodeType() == te::xml::START_ELEMENT) &&
	  (reader.getElementLocalName() == "Symbol"))
  {
    te::qt::widgets::Symbol* symbol = ReadSymbol(reader);
    library->addSymbol(symbol);
  }
  else
  {
	  // end element
	  reader.next();
  }

  if (reader.getNodeType() != te::xml::END_DOCUMENT
	  && reader.getNodeType() != te::xml::UNKNOWN)
  {
    BuildGroupTreeInsideLibrary(reader, library);
  }
}

te::qt::widgets::SymbolNode* te::qt::widgets::ReadGroup(te::xml::Reader& reader)
{
  std::unique_ptr<te::qt::widgets::SymbolNode> node(new te::qt::widgets::SymbolNode);

  // Symbol Name
  reader.next();
  assert(reader.getNodeType() == te::xml::START_ELEMENT);
  assert(reader.getElementLocalName() == "Name");
  reader.next();
  std::string name = reader.getElementValue();
  node->setName(name);
  reader.next();

  assert(reader.getNodeType() == te::xml::END_ELEMENT);
  reader.next();

  while ((reader.getNodeType() == te::xml::START_ELEMENT) &&
    (reader.getElementLocalName() == "Group"))
  {
    te::qt::widgets::SymbolNode* subNode = ReadGroup(reader);
    node->add(subNode);
  }

  while((reader.getNodeType() == te::xml::START_ELEMENT) &&
        (reader.getElementLocalName() == "Symbol"))
  {
    Symbol* symbol = ReadSymbol(reader);
    node->addSymbol(symbol);
  }

  assert(reader.getNodeType() == te::xml::END_ELEMENT);
  reader.next();

  return node.release();
}

te::qt::widgets::Symbol* te::qt::widgets::ReadSymbol(te::xml::Reader& reader)
{
  std::unique_ptr<te::qt::widgets::Symbol> symbol(new te::qt::widgets::Symbol);

  te::qt::widgets::SymbolInfo info;
  
  info = symbol->getInfo();

  // Symbol Name
  reader.next();

  // Symbol Id (compatibility with old XML)
  if (reader.getElementLocalName() == "Id")
  {
    reader.next();
    reader.next();
    reader.next();
  }

  assert(reader.getNodeType() == te::xml::START_ELEMENT);
  assert(reader.getElementLocalName() == "Name");
  reader.next();
  info.m_name = reader.getElementValue();
  reader.next();

  assert(reader.getNodeType() == te::xml::END_ELEMENT);
  reader.next();

  // Symbol Author
  if(reader.getElementLocalName() == "Author")
  {
    assert(reader.getNodeType() == te::xml::START_ELEMENT);
    reader.next();
    info.m_author = reader.getElementValue();
    reader.next();

    assert(reader.getNodeType() == te::xml::END_ELEMENT);
    reader.next();
  }

  // Symbol Tags
  if(reader.getElementLocalName() == "Tags")
  {
    assert(reader.getNodeType() == te::xml::START_ELEMENT);
    reader.next();
    info.m_tags = reader.getElementValue();
    reader.next();

    assert(reader.getNodeType() == te::xml::END_ELEMENT);
    reader.next();
  }

  // Symbol Description
  if(reader.getElementLocalName() == "Description")
  {
    assert(reader.getNodeType() == te::xml::START_ELEMENT);
    reader.next();
    info.m_description = reader.getElementValue();
    reader.next();

    assert(reader.getNodeType() == te::xml::END_ELEMENT);
    reader.next();
  }
  
  symbol->setInfo(info);

  assert(reader.getNodeType() == te::xml::START_ELEMENT);
  assert(reader.getElementLocalName().find("Symbolizer") != std::string::npos);

  /* TODO: *** Need move the OGC Symbology Enconding serialization methods 
              from <te::serialization> module to <te::se> module to avoid dependency with serialization module! ***/

  while((reader.getNodeType() == te::xml::START_ELEMENT) &&
        (reader.getElementLocalName().find("Symbolizer") != std::string::npos))
    symbol->addSymbolizer(te::se::serialize::Symbolizer::getInstance().read(reader));

  assert(reader.getNodeType() == te::xml::END_ELEMENT);
  reader.next();

  return symbol.release();
}

std::string te::qt::widgets::SerializeSymbolLibraryToBuffer()
{
  std::string buffer;

  SymbolLibrary* library = SymbolLibraryManager::getInstance().getLibrary();

  if (library->isEmpty())
  {
	  return buffer;
  }

  std::unique_ptr<te::xml::AbstractWriter> writer(te::xml::AbstractWriterFactory::make());
  writer->writeStartDocument("UTF-8", "no");
  writer->writeStartElement("SymbolLibrary");
  writer->writeAttribute("xmlns", "http://www.terralib.org/schemas/qt/se");
  writer->writeAttribute("xsd:schemaLocation", "http://www.terralib.org/schemas/qt/se symbol.xsd");
  writer->writeAttribute("xmlns:se", "http://www.opengis.net/se");
  writer->writeAttribute("xmlns:ogc", "http://www.opengis.net/ogc");
  writer->writeAttribute("xmlns:xsd", "http://www.w3.org/2001/XMLSchema-instance");
  writer->writeAttribute("xmlns:xlink", "http://www.w3.org/1999/xlink");
  writer->writeAttribute("name", "TerraLib 5 Default Symbol Library");
  writer->writeAttribute("version", library->version());

  SerializeSymbolLibrary(*writer);

  buffer = writer->getContent();

  return buffer;
}

void te::qt::widgets::WriteSymbolLibrary(const std::string& path)
{
  SymbolLibrary* library = SymbolLibraryManager::getInstance().getLibrary();

  if (library->isEmpty())
  {
	  return;
  }

  std::unique_ptr<te::xml::AbstractWriter> writer(te::xml::AbstractWriterFactory::make());

  writer->setURI(path);

  writer->writeStartDocument("UTF-8", "no");

  writer->writeStartElement("SymbolLibrary");

  writer->writeAttribute("xmlns", "http://www.terralib.org/schemas/qt/se");
  writer->writeAttribute("xsd:schemaLocation", "http://www.terralib.org/schemas/qt/se symbol.xsd");
  writer->writeAttribute("xmlns:se", "http://www.opengis.net/se");
  writer->writeAttribute("xmlns:ogc", "http://www.opengis.net/ogc");
  writer->writeAttribute("xmlns:xsd", "http://www.w3.org/2001/XMLSchema-instance");
  writer->writeAttribute("xmlns:xlink", "http://www.w3.org/1999/xlink");
  writer->writeAttribute("name", "TerraLib 5 Default Symbol Library");
  writer->writeAttribute("version", library->version());
  
  SerializeSymbolLibrary(*writer);

  writer->writeToFile();
}

void te::qt::widgets::SerializeSymbolLibrary(te::xml::AbstractWriter& writer)
{
  SymbolLibrary* library = SymbolLibraryManager::getInstance().getLibrary();

  if (library->isEmpty())
  {
	  return;
  }

  te::qt::widgets::SymbolNode* root = library->getRoot();

  WriteElements(writer, root);
}

void te::qt::widgets::WriteElements(te::xml::AbstractWriter& writer, te::qt::widgets::SymbolNode* root)
{
  const std::map<std::string, SymbolNode*>& nodes = root->getNodes();

  std::map<std::string, SymbolNode*>::const_iterator it = nodes.cbegin();

  for (; it != nodes.cend(); ++it)
  {
    std::string name = it->second->getName();
    SymbolNode* node = it->second;
    WriteElement(node, writer);
  }
  
  const std::map<std::string, Symbol*>& symbols = root->getSymbols();

  std::map<std::string, Symbol*>::const_iterator its = symbols.cbegin();
  for (; its != symbols.cend(); ++its)
  {
    Symbol* symbol = its->second;
    WriteElement(symbol, writer);
  }
}

void te::qt::widgets::WriteElement(te::qt::widgets::SymbolNode* group, te::xml::AbstractWriter& writer)
{
  writer.writeStartElement("Group");

  writer.writeElement("Name", group->getName());

  const std::map<std::string, SymbolNode*>& nodes = group->getNodes();

  std::map<std::string, SymbolNode*>::const_iterator it = nodes.cbegin();

  for (; it != nodes.cend(); ++it)
  {
    std::string name = it->second->getName();
    SymbolNode* node = it->second;
    WriteElement(node, writer);
  }
  
  const std::map<std::string, Symbol*>& symbols = group->getSymbols();

  std::map<std::string, Symbol*>::const_iterator its = symbols.cbegin();
  for (; its != symbols.cend(); ++its)
  {
    Symbol* symbol = its->second;
    WriteElement(symbol, writer);
  }

  writer.writeEndElement("Group");
}

void te::qt::widgets::WriteElement(te::qt::widgets::Symbol* symbol, te::xml::AbstractWriter& writer)
{
  writer.writeStartElement("Symbol");

  writer.writeElement("Name", symbol->getInfo().m_name);
  writer.writeElement("Author", symbol->getInfo().m_author);
  writer.writeElement("Tags", symbol->getInfo().m_tags);
  writer.writeElement("Description", symbol->getInfo().m_description);

  const std::vector<te::se::Symbolizer*>& symbolizers = symbol->getSymbolizers();
  
  std::vector<te::se::Symbolizer*>::const_iterator it = symbolizers.begin();
  for (; it != symbolizers.end(); ++it)
  {
    te::se::Symbolizer* symbolizer = (*it);
    WriteElement(symbolizer, writer);
  }

  writer.writeEndElement("Symbol");
}

void te::qt::widgets::WriteElement(te::se::Symbolizer* symbolizer, te::xml::AbstractWriter& writer)
{
  te::se::serialize::Symbolizer::getInstance().write(symbolizer, writer);
}

te::se::Font* te::qt::widgets::Convert2TerraLib(const QFont& qFont)
{
  te::se::Font* font = new te::se::Font();
  font->setFamily(Convert2StdString(qFont.family()));
  font->setSize(Convert2StdString(QString::number(qFont.pointSizeF(), 'f', 2)));
  font->setStyle(static_cast<te::se::Font::FontStyleType>(qFont.style()));

  QFont::Weight weightValue = static_cast<QFont::Weight>(qFont.weight());

  if (weightValue == QFont::Light)
    font->setWeight(te::se::Font::WeightLight);
  else if (weightValue == QFont::Normal)
    font->setWeight(te::se::Font::WeightNormal);
  else if (weightValue == QFont::DemiBold)
    font->setWeight(te::se::Font::WeightDemiBold);
  else if (weightValue == QFont::Bold)
    font->setWeight(te::se::Font::WeightBold);
  else if (weightValue == QFont::Black)
    font->setWeight(te::se::Font::WeightBlack);

  return font;
}

QFont te::qt::widgets::Convert2Qt(const te::se::Font* font)
{
  QFont qFont;
  qFont.setFamily(Convert2QString(te::se::GetString(font->getFamily())));
  qFont.setPointSize(te::se::GetDouble(font->getSize()));
  qFont.setStyle(static_cast<QFont::Style>(te::se::GetInt(font->getStyle())));

  std::string weightValue = te::se::GetString(font->getWeight());

  if (weightValue == "light")
    qFont.setWeight(QFont::Light);
  else if (weightValue == "normal")
    qFont.setWeight(QFont::Normal);
  else if (weightValue == "demibold")
    qFont.setWeight(QFont::DemiBold);
  else if (weightValue == "bold")
    qFont.setWeight(QFont::Bold);
  else if (weightValue == "black")
    qFont.setWeight(QFont::Black);

  return qFont;
}

