/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/widgets/tools/CoordTracking.cpp

  \brief This class implements a concrete tool to geographic coordinate tracking on mouse move operation.
*/

// TerraLib
#include "../../../srs/Converter.h"
#include "../canvas/MapDisplay.h"
#include "../Utils.h"
#include "CoordTracking.h"

// Qt
#include <QClipboard>
#include <QMouseEvent>

#define SRS_LATLONG 4326

te::qt::widgets::CoordTracking::CoordTracking(te::qt::widgets::MapDisplay* display, QObject* parent) 
  : AbstractTool(display, parent)
{
  m_display->setMouseTracking(true);

  m_conveter = new te::srs::Converter();
  m_conveter->setTargetSRID(SRS_LATLONG);
}

te::qt::widgets::CoordTracking::~CoordTracking()
{
  delete m_conveter;
}

void te::qt::widgets::CoordTracking::setDisplaySRID(int srid)
{
  if (srid != TE_UNKNOWN_SRS)
  {
    m_conveter->setSourceSRID(srid);
  }
  else
  {
    delete m_conveter;

    m_conveter = new te::srs::Converter();
    m_conveter->setTargetSRID(SRS_LATLONG);
  }
}

bool te::qt::widgets::CoordTracking::mouseMoveEvent(QMouseEvent* e)
{
#if (QT_VERSION >= 0x050000)
  QPointF p = e->localPos();
#else
  QPointF p = e->posF();
#endif

  m_currentPos.clear();
  
  p = m_display->transform(p);
  if(!p.isNull())
    emit coordTracked(p);

  QPointF pLL(0., 0.);

  try
  {
    if (!p.isNull() && m_conveter->getSourceSRID() != TE_UNKNOWN_SRS)
    {
      double llx = p.x();
      double lly = p.y();

      if(m_conveter->getSourceSRID() != m_conveter->getTargetSRID())
        m_conveter->convert(llx, lly);

      pLL.setX(llx);
      pLL.setY(lly);
    }
  }
  catch (...)
  {
  }

  if (!p.isNull())
    emit coordLLTracked(p, pLL);

  m_currentPos.append(QString::number(p.y(), 'f', 6));
  m_currentPos.append(" ");
  m_currentPos.append(QString::number(p.x(), 'f', 6));

  return false;
}

bool te::qt::widgets::CoordTracking::keyPressEvent(QKeyEvent* e)
{
  if (e->key() == Qt::Key_C)
  {
    if (e->modifiers() == Qt::ControlModifier)
    {
      QClipboard* clipboard = te::qt::widgets::GetClipboard();
      clipboard->setText(m_currentPos);

      return false;
    }
  }

  return true;
}
