/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/widgets/tools/Info.cpp

  \brief This class implements a concrete tool to get informations about a specified feature using pointing operation.
*/

// TerraLib
#include "../../../common/STLUtils.h"
#include "../../../dataaccess/dataset/DataSet.h"
#include "../../../dataaccess/dataset/DataSetType.h"
#include "../../../dataaccess/utils/Utils.h"
#include "../../../geometry/Coord2D.h"
#include "../../../geometry/Geometry.h"
#include "../../../geometry/GeometryProperty.h"
#include "../../../geometry/Point.h"
#include "../../../geometry/Utils.h"
#include "../../../maptools/MarkRendererManager.h"
#include "../../../raster/Raster.h"
#include "../../../raster/RasterProperty.h"
#include "../../../raster/Utils.h"
#include "../../../se/Fill.h"
#include "../../../se/Stroke.h"
#include "../../../se/Mark.h"
#include "../../../se/Utils.h"
#include "../../../srs/Config.h"
#include "../canvas/Canvas.h"
#include "../canvas/MapDisplay.h"
#include "../Utils.h"
#include "../layer/info/FeatureInfoMakerFactory.h"
#include "../layer/info/AbstractFeatureInfoMaker.h"
#include "Info.h"

// Qt
#include <QtCore/QPointF>
#include <QtCore/QStringList>
#include <QMessageBox>
#include <QMouseEvent>
#include <QMenu>
#include <QPoint>
#include <QClipboard>
#include <QTreeWidget>

// STL
#include <cassert>
#include <memory>

Q_DECLARE_METATYPE(std::shared_ptr<te::gm::Geometry>)

te::qt::widgets::Info::Info(QTreeWidget* infoTree, te::qt::widgets::MapDisplay* display, const QCursor& cursor, const std::list<te::map::AbstractLayerPtr>& layers, QObject* parent)
  : AbstractTool(display, parent),
    m_infoTreeWidget(infoTree),
    m_layers(layers),
    m_mode(""),
    m_copyAction(nullptr)
{
  setCursor(cursor);

  connect(m_infoTreeWidget, SIGNAL(customContextMenuRequested(const QPoint&)), this, SLOT(onCustomContextMenuRequested(const QPoint&)));

  m_copyAction = new QAction(QString(tr("Copy")), this);
  connect(m_copyAction, SIGNAL(changed()), this, SLOT(onCustomContextMenuCopy()));
  connect(m_display, SIGNAL(drawLayersFinished(const QMap<QString, QString>&)), SLOT(onRedraw()));
}

te::qt::widgets::Info::~Info()
{
  
}

bool te::qt::widgets::Info::mouseReleaseEvent(QMouseEvent* e)
{
  if(e->button() != Qt::LeftButton)
    return false;

  emit geometrySelected();

  Qt::KeyboardModifiers keys = e->modifiers();

  std::list<te::map::AbstractLayerPtr> layers;

  QString option = tr("Visible layers top to bottom");

  if (m_mode.compare(option) == 0
     || keys & Qt::ShiftModifier)
  {
    layers = m_display->getLayerList();
  }
  else
  {
    if (m_layers.empty())
      return false;

    layers = m_layers;
  }

  QPointF pixelOffset(4.0, 4.0);
#if (QT_VERSION >= 0x050000)
  QRectF rect = QRectF(e->localPos() - pixelOffset, e->localPos() + pixelOffset);
#else
  QRectF rect = QRectF(e->posF() - pixelOffset, e->posF() + pixelOffset);
#endif
      
  // Converts rect boundary to world coordinates
  QPointF ll(rect.left(), rect.bottom());
  QPointF ur(rect.right(), rect.top());
  ll = m_display->transform(ll);
  ur = m_display->transform(ur);

  // Bulding the query box
  m_currentBox = te::gm::Envelope(ll.x(), ll.y(), ur.x(), ur.y());

  buildTree(m_currentBox, layers);

  return true;
}

void te::qt::widgets::Info::buildTree(const te::gm::Envelope& envelope, const std::list<te::map::AbstractLayerPtr>& layers)
{
  // Clear draft!
  QPixmap* draft = m_display->getDraftPixmap();
  draft->fill(Qt::transparent);

  // Clear info widget!
  m_infoTreeWidget->clear();

  // For each layer, get info!
  std::list<te::map::AbstractLayerPtr>::const_iterator it;
  for (it = layers.begin(); it != layers.end(); ++it)
  {
    te::map::AbstractLayerPtr layer = *it;

    // The Layer item
    QTreeWidgetItem* layerItem = new QTreeWidgetItem;
    layerItem->setIcon(0, QIcon::fromTheme("layer"));
    layerItem->setText(0, tr("Layer"));
    layerItem->setText(1, layer->getTitle().c_str());

    m_infoTreeWidget->addTopLevelItem(layerItem);

    std::unique_ptr<AbstractFeatureInfoMaker> featureInfoMaker;

    featureInfoMaker.reset(te::qt::widgets::FeatureInfoMakerFactory::make(layer->getType()));

    featureInfoMaker->getInfo(layerItem, m_display, layer, envelope);

    if (layerItem->childCount() == 0)
      delete layerItem;
  }

  if (m_infoTreeWidget->topLevelItemCount() > 0)
  {
    m_infoTreeWidget->expandAll();
    m_infoTreeWidget->resizeColumnToContents(0);
    m_infoTreeWidget->show();
  }

  m_display->repaint();
}

void te::qt::widgets::Info::setLayers(const std::list<te::map::AbstractLayerPtr>& layers)
{
  m_layers = layers;
}

void te::qt::widgets::Info::setMode(const QString& mode)
{
  m_mode = mode;
}

void te::qt::widgets::Info::onCustomContextMenuRequested(const QPoint& pos)
{
  QMenu* menu = new QMenu(m_infoTreeWidget);

  menu->addAction(m_copyAction);
  menu->popup(m_infoTreeWidget->viewport()->mapToGlobal(pos));

  m_copyAction->setData(QVariant(pos)); // position data
}

void te::qt::widgets::Info::onCustomContextMenuCopy()
{
  QTreeWidgetItem* item = m_infoTreeWidget->itemAt(m_copyAction->data().toPoint());
  QString txt = item->text(1);

  if (txt.compare(tr("Layer")) == 0
      || txt.contains(tr("FID")) == 0)
  {
    QString qKey = item->text(0);
    QString qValue = item->text(1);

    std::shared_ptr<te::gm::Geometry> geometry = item->data(1, Qt::UserRole).value< std::shared_ptr<te::gm::Geometry> >();
    if (geometry != nullptr)
    {
      qValue = te::qt::widgets::Convert2QString(geometry->toString());
    }

    QStringList information;
    information.append(qKey + ": " + qValue);

    getChildInformation(item, information);

    txt = information.join("\n");
  }

  QClipboard* clipboard = te::qt::widgets::GetClipboard();
  clipboard->setText(txt);
}

void te::qt::widgets::Info::getChildInformation(QTreeWidgetItem* item, QStringList& information)
{
  int childCount = item->childCount();

  for (int i = 0; i < childCount; ++i)
  {
    QTreeWidgetItem* child = item->child(i);
    
    QString qKey = child->text(0);
    QString qValue = child->text(1);

    std::shared_ptr<te::gm::Geometry> geometry = item->data(1, Qt::UserRole).value< std::shared_ptr<te::gm::Geometry> >();
    if (geometry != nullptr)
    {
      qValue = te::qt::widgets::Convert2QString(geometry->toString());
    }

    information.append(qKey + ": " + qValue);
    if (child->childCount() > 0)
    {
      getChildInformation(child, information);
    }
  }
}

void te::qt::widgets::Info::onRedraw()
{
  if (m_layers.empty())
    return;

  std::list<te::map::AbstractLayerPtr> layers;

  QString option = tr("Visible layers top to bottom");

  if (m_mode.compare(option) == 0)
  {
    layers = m_display->getLayerList();
  }
  else
  {
    layers = m_layers;
  }

  if (m_currentBox.isValid())
  {
    buildTree(m_currentBox, layers);
  }
}

void te::qt::widgets::Info::onCurrentIndexChanged(const QString& text)
{
  setMode(text);
}
