/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/widgets/tools/PointMove.h

  \brief This class implements a concrete tool to geographic coordinate move a point on mouse operation.
*/

#ifndef __TERRALIB_QT_WIDGETS_INTERNAL_POINTMOVE_H
#define __TERRALIB_QT_WIDGETS_INTERNAL_POINTMOVE_H

// TerraLib
#include "../Config.h"
#include "AbstractTool.h"

// Forward declarations
class QPointF;

namespace te
{
  namespace qt
  {
    namespace widgets
    {
// Forward declarations
    class MapDisplay;

      /*!
        \class PointMove

        \brief This class implements a concrete tool to geographic coordinate move a point on mouse click operation.
      */
      class TEQTWIDGETSEXPORT PointMove : public AbstractTool
      {
        Q_OBJECT

        public:

          /** @name Initializer Methods
           *  Methods related to instantiation and destruction.
           */
          //@{

          /*!
            \brief It constructs a coordinate tracking tool associated with the given map display.

            \param display The map display associated with the tool.
            \param cursor The default tool cursor.
            \param parent The tool's parent.

            \note The tool will NOT take the ownership of the given pointers.
          */
          PointMove(MapDisplay* display, std::map<unsigned int, QPointF> *points, const QCursor& cursor, QObject* parent = 0);

          /*! \brief Destructor. */
          ~PointMove();

          //@}

          /** @name AbstractTool Methods
           *  Methods related with tool behavior.
           */
          //@{

          bool mousePressEvent(QMouseEvent* e);

          bool mouseMoveEvent(QMouseEvent* e);

          bool mouseReleaseEvent(QMouseEvent* e);

          //@}

      private:
        te::gm::Envelope buildEnvelope(const QPointF& pos);

        void draw();

        signals:

          /*! This signal is emitted when a coordinate was picked. */
          void pointMoved(QPointF& coordinate_original, QPointF& coordinate_new);

          /*! This signal is emitted when a mouse moving. */
          void pointMoveMoving(QPointF& coordinate);

      protected:

        bool m_moveStarted;                            //!< Flag that indicates if move operation was started.
        QPointF m_origin;                              //!< Origin point on mouse pressed.
        QPointF m_delta;                               //!< Difference between pressed point and destination point on mouse move.
        std::map< unsigned int, QPointF > *m_points;
        unsigned int  m_ptMoving;
      };

    } // end namespace widgets
  }   // end namespace qt
}     // end namespace te

#endif  // __TERRALIB_QT_WIDGETS_INTERNAL_POINTMOVE_H
