/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/widgets/utils/FileDialog.h

  \brief An instance of QFileDialog that includes file extention if necessary.
*/

#ifndef __TERRALIB_QT_WIDGETS_INTERNAL_FILEDIALOG_H
#define __TERRALIB_QT_WIDGETS_INTERNAL_FILEDIALOG_H

// TerraLib
#include "../Config.h"

// Qt
#include <QDialog>
#include <QFileDialog>

#include <set>

namespace te
{
  namespace qt
  {
    namespace widgets
    {
      /*!
        \class FileDialog

        \brief Defines a component for choose a file.

        This class is used to create a QFileDialog in order to facilitate the use of file path,
        file name and file extension.

        \ingroup widgets
      */
      class TEQTWIDGETSEXPORT FileDialog
      {
        public:

          enum FileType
          {
            RASTER,
            VECTOR
          };
          
          enum FileAccessMode
          {
            UNDEFINED_FILE_MODE,
            OPEN_FILE_MODE,
            SAVE_FILE_MODE
          };          

          /*!
            \brief Constructor.
            \param parent The QWidget parent argumnt.
            \param pathInfo The type of file opened. Actually we assume three values: "vector", "raster", "database". Any other value may be not used in the system.
            \param File Mode will be set to UNDEFINED_FILE_MODE.
          */
          FileDialog(QWidget* parent, FileType type, std::string pathInfo = "");
          
          /*!
            \brief Constructor.
            \param parent The QWidget parent argumnt.
            \param pathInfo The type of file opened. Actually we assume three values: "vector", "raster", "database". Any other value may be not used in the system.
            \fileMode File mode.
          */
          FileDialog(QWidget* parent, const FileType& type, 
            const std::string& pathInfo, const FileAccessMode& fileMode );

          /*!
            \brief Destructor.
          */
          ~FileDialog(){}

          /*!
            \brief This method will open the dialog of file selection and
            populate the class members with the chosen file path.
            \note te::common::Exception on errors.
          */
          void exec();

          /*!
            \brief This method will return the chosen path.
            \return The complete chosen path.
          */
          std::string getPath();

          /*!
            \brief Sets the file path.
            \param The file path.
          */
          void setPath(std::string filePath);

          /*!
            \brief This method will return the file name.
            \return The complete chosen file name.
          */
          std::string getFileName();

          /*!
            \brief Sets the file name.
            \param The file path with a file name.
          */
          void setFileName(std::string filePath);

          /*!
            \brief This method will return the file extension.
            \return The complete chosen file extension.
          */
          std::string getExtension();

          /*!
            \brief Sets the file extension.
            \param The file path with a file and an extension.
          */
          void setExtension(std::string filePath);

          /*!
            \brief Genarate an extensions map with the extensions returned by driver.
            \param The extensions returned by driver.
          */
          std::map<QString, QString> getMappedExtensions(QString extensions);

          /*!
            \brief Verify if the file name has special characters
            \param The name of the file.
          */
          bool hasInvalidCharacters(std::string filename);

          void setAllowedExtensions(std::set<std::string> allowedExtensions);

        private:

          FileAccessMode m_fileMode;
          QWidget*    m_parent;     //!< The QWidget parent argumnt.
          FileType    m_type;       //!< The type of file.
          std::string m_pathInfo;   //!< The file path info.
          std::string m_path;       //!< The path of the file.
          std::string m_fileName;   //!< The name of the file.
          std::string m_extension;  //!< The extension of the file.
          std::set<std::string> m_allowedExtensions; //!< Set of allowed extensions.
      };

    } // end namespace widgets
  }   // end namespace qt
}     // end namespace te

#endif  // __TERRALIB_QT_WIDGETS_INTERNAL_FILEDIALOG_H

