/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/raster/Band.cpp

  \brief It gives access to values in one band (dimension) of a raster.
*/

// TerraLib

#include "Band.h"
#include "BandProperty.h"
#include "PositionIterator.h"
#include "Utils.h"

#include "../geometry/Envelope.h"

// STL
#include <cassert>
#include <limits>

te::rst::Band::Band(BandProperty* p, std::size_t idx)
  : m_property(p),
    m_idx(idx)
{
}

te::rst::Band::~Band()
{
  delete m_property;
}

te::rst::Band& te::rst::Band::operator=(const Band& rhs)
{
  if(this != &rhs)
  {
    delete m_property;

    m_property = rhs.m_property ? new BandProperty(*rhs.m_property) : nullptr;

    m_idx = rhs.m_idx;
  }

  return *this;
}

void te::rst::Band::getValue(unsigned int c, unsigned int r, std::complex<double>& value) const
{
  double vr = 0.0;

  double vi = 0.0;

  getValue(c, r, vr);

  getIValue(c, r, vi);

  value = std::complex<double>(vr, vi);
}

void te::rst::Band::setValue(unsigned int c, unsigned int r, const std::complex<double>& value)
{
  setValue(c, r, value.real());

  setIValue(c, r, value.imag());
}

void te::rst::Band::getBlockBuffer(void* buffer, int x0, int y0, int width, int height, int outWidth, int outHeight, int dataType)
{
  //function to get value from buffer
  te::rst::GetBufferValueFPtr m_getBuff;   //!< A pointer to a function that helps to extract a double or complex value from a specific buffer data type (char, int16, int32, float, ...).
  te::rst::GetBufferValueFPtr m_getBuffI;  //!< A pointer to a function that helps to extract the imaginary part value from a specific buffer data type (cint16, cint32, cfloat, cdouble).
  te::rst::SetBufferValueFPtr m_setBuff;   //!< A pointer to a function that helps to insert a double or complex value into a specific buffer data type (char, int16, int32, float, ...).
  te::rst::SetBufferValueFPtr m_setBuffI;  //!< A pointer to a function that helps to insert the imaginary part value into a specific buffer data type (cint16, cint32, cfloat, cdouble).

  te::rst::SetBlockFunctions(&m_getBuff, &m_getBuffI, &m_setBuff, &m_setBuffI, m_property->getType());

  double scaleY = ((double)(height)) / ((double)(outHeight));
  double scaleX = ((double)(width)) / ((double)(outWidth));

  te::gm::Envelope inputTotalArea(x0, y0, x0 + width - 1, y0 + height - 1);

  int blkw = m_property->m_blkw;
  int blkh = m_property->m_blkh;

  int startBlockXIdx = x0 / blkw;
  int startBlockYIdx = y0 / blkh;
  int finalBlockXIdx = (x0 + width -1 ) / blkw;
  int finalBlockYIdx = (y0 + height -1 ) / blkh;

  for (int blkYIdx = startBlockYIdx; blkYIdx <= finalBlockYIdx; ++blkYIdx)
  {
    for (int blkXIdx = startBlockXIdx; blkXIdx <= finalBlockXIdx; ++blkXIdx)
    {
      void* bufferIn = read(blkXIdx, blkYIdx);

      int blockX0 = blkXIdx * blkw;
      int blockX1 = blockX0 + blkw - 1;
      int blockY0 = blkYIdx * blkh;
      int blockY1 = blockY0 + blkh - 1;

      te::gm::Envelope blockArea(blockX0, blockY0, blockX1, blockY1);

      te::gm::Envelope inter = blockArea.intersection(inputTotalArea);

      if (inter.isValid() == false)
        continue;

      int outX0 = te::rst::Round((inter.m_llx - x0) / scaleX);
      int outY0 = te::rst::Round((inter.m_lly - y0) / scaleY);
      int outX1 = te::rst::Round(((inter.m_urx - x0) + 1) / scaleX);
      int outY1 = te::rst::Round(((inter.m_ury - y0) + 1) / scaleY);

      for (int y = outY0; y < outY1; ++y)
      {
        for (int x = outX0; x < outX1; ++x)
        {
          int inX = static_cast<int>(x0 + (x * scaleX) - blockX0);
          int inY = static_cast<int>(y0 + (y * scaleY) - blockY0);

          if (inX < 0) 
            inX = 0;
          if (inY < 0) 
            inY = 0;
          if (inX >= blkw) 
            inX = blkw - 1;
          if (inY >= blkh) 
            inY = blkh - 1;

          double dValue = 0;

          int indexIn = inX + inY * blkw;

          m_getBuff(indexIn, bufferIn, &dValue);

          int indexOut = x + y * outWidth;

          m_setBuff(indexOut, buffer, &dValue);
        }
      }
    }
  }
}

std::complex<double> te::rst::Band::getMinValue(bool readall, unsigned int rs, unsigned int cs, unsigned int rf, unsigned int cf) const
{
  double minReal = 0;
  double maxReal = 0;
  double meanReal = 0;
  double stdDevReal = 0;
  double minImg = 0;
  double maxImg = 0;
  double meanImg = 0;
  double stdDevImg = 0; 
  
  if( getRaster() )
  {
    if( 
        ( rs == 0 ) 
        && 
        ( cs == 0 ) 
        && 
        ( rf == 0 ) 
        && 
        ( cf == 0 ) 
      )
    {
      rf = getRaster()->getNumberOfRows() - 1;
      cf = getRaster()->getNumberOfColumns() - 1;
    }
    
    std::map<double, unsigned> iHistogram;
    std::map<double, unsigned> rHistogram;
    
    if( ! getHistograms( *getRaster(), m_idx, rs, cs, rf, 
      cf, 0, 1, 0, rHistogram, iHistogram ) )
    {
      throw te::rst::Exception( "Histogram build error" );
    }

    getHistogramStats( rHistogram, minReal, maxReal, meanReal, stdDevReal );
    getHistogramStats( iHistogram, minImg, maxImg, meanImg, stdDevImg );
  }
    
  return std::complex<double>( minReal, minImg );
}

std::complex<double> te::rst::Band::getMaxValue(bool readall, unsigned int rs, unsigned int cs, unsigned int rf, unsigned int cf) const
{
  double minReal = 0;
  double maxReal = 0;
  double meanReal = 0;
  double stdDevReal = 0;
  double minImg = 0;
  double maxImg = 0;
  double meanImg = 0;
  double stdDevImg = 0; 
  
  if( getRaster() )
  {
    if( 
        ( rs == 0 ) 
        && 
        ( cs == 0 ) 
        && 
        ( rf == 0 ) 
        && 
        ( cf == 0 ) 
      )
    {
      rf = getRaster()->getNumberOfRows() - 1;
      cf = getRaster()->getNumberOfColumns() - 1;
    }    
    
    std::map<double, unsigned> iHistogram;
    std::map<double, unsigned> rHistogram;
    
    if( ! getHistograms( *getRaster(), m_idx, rs, cs, rf, 
      cf, 0, 1, 0, rHistogram, iHistogram ) )
    {
      throw te::rst::Exception( "Histogram build error" );
    }

    getHistogramStats( rHistogram, minReal, maxReal, meanReal, stdDevReal );
    getHistogramStats( iHistogram, minImg, maxImg, meanImg, stdDevImg );
  }
    
  return std::complex<double>( maxReal, maxImg );
}

std::complex<double> te::rst::Band::getStdValue(unsigned int rs, unsigned int cs, unsigned int rf, unsigned int cf) const
{
  double minReal = 0;
  double maxReal = 0;
  double meanReal = 0;
  double stdDevReal = 0;
  double minImg = 0;
  double maxImg = 0;
  double meanImg = 0;
  double stdDevImg = 0; 
  
  if( getRaster() )
  {
    if( 
        ( rs == 0 ) 
        && 
        ( cs == 0 ) 
        && 
        ( rf == 0 ) 
        && 
        ( cf == 0 ) 
      )
    {
      rf = getRaster()->getNumberOfRows() - 1;
      cf = getRaster()->getNumberOfColumns() - 1;
    }    
    
    std::map<double, unsigned> iHistogram;
    std::map<double, unsigned> rHistogram;
    
    if( ! getHistograms( *getRaster(), m_idx, rs, cs, rf, 
      cf, 0, 1, 0, rHistogram, iHistogram ) )
    {
      throw te::rst::Exception( "Histogram build error" );
    }

    getHistogramStats( rHistogram, minReal, maxReal, meanReal, stdDevReal );
    getHistogramStats( iHistogram, minImg, maxImg, meanImg, stdDevImg );
  }
    
  return std::complex<double>( stdDevReal, stdDevImg );
}

std::complex<double> te::rst::Band::getMeanValue(unsigned int rs, unsigned int cs, unsigned int rf, unsigned int cf) const
{
  double minReal = 0;
  double maxReal = 0;
  double meanReal = 0;
  double stdDevReal = 0;
  double minImg = 0;
  double maxImg = 0;
  double meanImg = 0;
  double stdDevImg = 0; 
  
  if( getRaster() )
  {
    if( 
        ( rs == 0 ) 
        && 
        ( cs == 0 ) 
        && 
        ( rf == 0 ) 
        && 
        ( cf == 0 ) 
      )
    {
      rf = getRaster()->getNumberOfRows() - 1;
      cf = getRaster()->getNumberOfColumns() - 1;
    }    
    
    std::map<double, unsigned> iHistogram;
    std::map<double, unsigned> rHistogram;
    
    if( ! getHistograms( *getRaster(), m_idx, rs, cs, rf, 
      cf, 0, 1, 0, rHistogram, iHistogram ) )
    {
      throw te::rst::Exception( "Histogram build error" );
    }

    getHistogramStats( rHistogram, minReal, maxReal, meanReal, stdDevReal );
    getHistogramStats( iHistogram, minImg, maxImg, meanImg, stdDevImg );
  }
    
  return std::complex<double>( meanReal, meanImg );
}

std::map<double, unsigned> te::rst::Band::getHistogramR(unsigned int rs, 
  unsigned int cs, unsigned int rf, unsigned int cf, unsigned int b) const
{
  std::map<double, unsigned> iHistogram;
  std::map<double, unsigned> rHistogram;
  
  if( getRaster() )
  {
    if( 
        ( rs == 0 ) 
        && 
        ( cs == 0 ) 
        && 
        ( rf == 0 ) 
        && 
        ( cf == 0 ) 
      )
    {
      rf = getRaster()->getNumberOfRows() - 1;
      cf = getRaster()->getNumberOfColumns() - 1;
    }    
    
    if( ! getHistograms( *getRaster(), m_idx, rs, cs, rf, 
      cf, b, 1, 0, rHistogram, iHistogram ) )
    {
      rHistogram.clear();
    }
  }  

  return rHistogram;
}

void te::rst::Band::getHistogramR( const unsigned int rowStart, const unsigned int colStart,
  const unsigned int finalRow, const unsigned int finalCol, const unsigned int histoBins,
  const unsigned int sampleStep, std::map<double, unsigned>& histogram ) const
{
  std::map<double, unsigned> iHistogram;
  
  if( getRaster() )
  {
    if( ! getHistograms( *getRaster(), m_idx, rowStart, colStart, finalRow, 
      finalCol, histoBins, sampleStep, 0, histogram, iHistogram ) )
    {
      throw te::rst::Exception( "Histogram build error" );
    }
  }
}

std::map<double, unsigned> te::rst::Band::getHistogramI(unsigned int rs, 
  unsigned int cs, unsigned int rf, unsigned int cf, unsigned int b) const
{
  std::map<double, unsigned> rHistogram;
  std::map<double, unsigned> iHistogram;
  
  if( getRaster() )
  {
    if( 
        ( rs == 0 ) 
        && 
        ( cs == 0 ) 
        && 
        ( rf == 0 ) 
        && 
        ( cf == 0 ) 
      )
    {
      rf = getRaster()->getNumberOfRows() - 1;
      cf = getRaster()->getNumberOfColumns() - 1;
    }    
    
    if( ! getHistograms( *getRaster(), m_idx, rs, cs, rf, cf, b, 1, 0, 
      rHistogram, iHistogram ) )
    {
      iHistogram.clear();
    }
  }
  
  return iHistogram;
}

std::complex<double> te::rst::Band::getScaleValue() const
{
  return m_property->m_valuesScale;
}

void te::rst::Band::setScaleValue(const std::complex<double> s)
{
  m_property->m_valuesScale = s;
}

std::complex<double> te::rst::Band::getOffsetValue() const
{
  return m_property->m_valuesOffset;
}

void te::rst::Band::setOffsetValue(const std::complex<double> o)
{
  m_property->m_valuesOffset = o;
}

te::rst::BandProperty* te::rst::Band::getProperty()
{
  return m_property;
}

const te::rst::BandProperty* te::rst::Band::getProperty() const
{
  return m_property;
}

te::rst::Band& te::rst::Band::callOperator(std::complex<double>(*f)(std::complex<double>, std::complex<double>), te::rst::Band& rhs)
{
  assert(getRaster()->getNumberOfRows() == rhs.getRaster()->getNumberOfRows());
  assert(getRaster()->getNumberOfColumns() == rhs.getRaster()->getNumberOfColumns());
  assert(getRaster()->getAccessPolicy() == te::common::RWAccess || getRaster()->getAccessPolicy() == te::common::WAccess);

  std::complex<double> lhsv, rhsv;

  std::vector<std::pair<unsigned, unsigned> > rcStPos, rcFPos;

  unsigned last_y;
  unsigned last_x;
  for (unsigned x = 0; x < (unsigned) getProperty()->m_nblocksx; x++)
  {
    for (unsigned y = 0; y < (unsigned) getProperty()->m_nblocksy; y++)
    {
      rcStPos.push_back(std::pair<unsigned, unsigned> (y * getProperty()->m_blkh, x * getProperty()->m_blkw));
      last_y = (y + 1) * getProperty()->m_blkh;
      last_x = (x + 1) * getProperty()->m_blkw;
      if (last_y > getRaster()->getNumberOfRows())
        last_y = getRaster()->getNumberOfRows();
      if (last_x > getRaster()->getNumberOfColumns())
        last_x = getRaster()->getNumberOfColumns();
      rcFPos.push_back(std::pair<unsigned, unsigned> (last_y, last_x));
    }
  }

// rasters without no data values
  if (getProperty()->m_noDataValue == std::numeric_limits<double>::max())
  {
    for (unsigned i = 0; i < rcStPos.size(); i++)
    {
      for (unsigned r = rcStPos[i].first; r < rcFPos[i].first; r++)
        for (unsigned c = rcStPos[i].second; c < rcFPos[i].second; c++)
        {
          getValue(c, r, lhsv);
          rhs.getValue(c, r, rhsv);

          lhsv = f(lhsv, rhsv);
          setValue(c, r, lhsv);
        }
      }
  }
// rasters with no data values
  else
  {
    for (unsigned i = 0; i < rcStPos.size(); i++)
    {
      for (unsigned r = rcStPos[i].first; r < rcFPos[i].first; r++)
        for (unsigned c = rcStPos[i].second; c < rcFPos[i].second; c++)
        {
// lhs data
          getValue(c, r, lhsv);
          if (lhsv.real() == getProperty()->m_noDataValue)
            continue;

// rhs data
          rhs.getValue(c, r, rhsv);
          if (rhsv.real() == rhs.getProperty()->m_noDataValue)
            continue;

          lhsv = f(lhsv, rhsv);
          setValue(c, r, lhsv);
        }
    }
  }

  return *this;
}

te::rst::Band& te::rst::Band::callOperator(std::complex<double>(*f)(std::complex<double>, std::complex<double>), std::complex<double>& cvalue)
{
  assert(getRaster()->getAccessPolicy() == te::common::RWAccess || getRaster()->getAccessPolicy() == te::common::WAccess);

  std::complex<double> lhsv;

  std::vector<std::pair<unsigned, unsigned> > rcStPos, rcFPos;

  unsigned last_y;
  unsigned last_x;
  for (unsigned x = 0; x < (unsigned) getProperty()->m_nblocksx; x++)
  {
    for (unsigned y = 0; y < (unsigned) getProperty()->m_nblocksy; y++)
    {
      rcStPos.push_back(std::pair<unsigned, unsigned> (y * getProperty()->m_blkh, x * getProperty()->m_blkw));
      last_y = (y + 1) * getProperty()->m_blkh;
      last_x = (x + 1) * getProperty()->m_blkw;
      if (last_y > getRaster()->getNumberOfRows())
        last_y = getRaster()->getNumberOfRows();
      if (last_x > getRaster()->getNumberOfColumns())
        last_x = getRaster()->getNumberOfColumns();
      rcFPos.push_back(std::pair<unsigned, unsigned> (last_y, last_x));
    }
  }

// rasters without no data values
  if (getProperty()->m_noDataValue == std::numeric_limits<double>::max())
  {
    for (unsigned i = 0; i < rcStPos.size(); i++)
    {
      for (unsigned r = rcStPos[i].first; r < rcFPos[i].first; r++)
        for (unsigned c = rcStPos[i].second; c < rcFPos[i].second; c++)
        {
          getValue(c, r, lhsv);

          lhsv = f(lhsv, cvalue);
          setValue(c, r, lhsv);
        }
      }
  }
// rasters with no data values
  else
  {
    for (unsigned i = 0; i < rcStPos.size(); i++)
    {
      for (unsigned r = rcStPos[i].first; r < rcFPos[i].first; r++)
        for (unsigned c = rcStPos[i].second; c < rcFPos[i].second; c++)
        {
// lhs data
          getValue(c, r, lhsv);
          if (lhsv.real() == getProperty()->m_noDataValue)
            continue;

          lhsv = f(lhsv, cvalue);
          setValue(c, r, lhsv);
        }
    }
  }

  return *this;
}

std::complex<double> plus(std::complex<double> lhs, std::complex<double> rhs)
{
  return lhs + rhs;
}

std::complex<double> minus(std::complex<double> lhs, std::complex<double> rhs)
{
  return lhs - rhs;
}

std::complex<double> times(std::complex<double> lhs, std::complex<double> rhs)
{
  return lhs * rhs;
}

std::complex<double> divide(std::complex<double> lhs, std::complex<double> rhs)
{
  return lhs / rhs;
}

te::rst::Band& te::rst::Band::operator+=(te::rst::Band& rhs)
{
  return callOperator(plus, rhs);
}

te::rst::Band& te::rst::Band::operator+=(std::complex<double>& cvalue)
{
  return callOperator(plus, cvalue);
}

te::rst::Band& te::rst::Band::operator-=(te::rst::Band& rhs)
{
  return callOperator(minus, rhs);
}

te::rst::Band& te::rst::Band::operator-=(std::complex<double>& cvalue)
{
  return callOperator(minus, cvalue);
}

te::rst::Band& te::rst::Band::operator*=(std::complex<double>& cvalue)
{
  return callOperator(times, cvalue);
}

te::rst::Band& te::rst::Band::operator*=(te::rst::Band& rhs)
{
  return callOperator(times, rhs);
}

te::rst::Band& te::rst::Band::operator/=(te::rst::Band& rhs)
{
  return callOperator(divide, rhs);
}

te::rst::Band& te::rst::Band::operator/=(std::complex<double>& cvalue)
{
  return callOperator(divide, cvalue);
}

int te::rst::Band::getBlockSize() const
{
  int blkw = m_property->m_blkw;

  int blkh = m_property->m_blkh;

  int pxlsize = GetPixelSize(m_property->getType());

  int blksize = blkw * blkh * pxlsize;

  return blksize;
}

te::rst::Band::Band(const Band& rhs)
  : m_property(nullptr),
    m_idx(rhs.m_idx)
{
  m_property = rhs.m_property ? new BandProperty(*rhs.m_property) : nullptr;
}

