/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/raster/BoundaryExtract.h

  \brief It implements the raster boundary extractor
*/

#ifndef __TERRALIB_RASTER_INTERNAL_BOUNDARYEXTRACT_H
#define __TERRALIB_RASTER_INTERNAL_BOUNDARYEXTRACT_H



// TerraLib
#include "../geometry/Coord2D.h"
#include "../geometry/LineString.h"
#include "Enums.h"
#include "Raster.h"
#include "Vectorizer.h"
#include "VectorizerPolygonStructure.h"

// STL
#include <vector>

namespace te
{
  namespace rst
  {
// Forward declaration.
    class Raster;

    /*!
      \class BoundaryExtract

      \brief It implements the Boundary Extract (vectorizer specialization - extract only one pixel value)

      This algorithm implements a vectorizer, which is an algorithm to convert an
      image with pixel-labels into a set of polygons. All connected pixels in the
      image with the same label will define an individual polygon. This algorithm
      is based on the implementation of the previous version of this library, the
      TerraLib 4.

      \ingroup rst

      \sa Raster, Geometry
    */
    class TERASTEREXPORT BoundaryExtract : Vectorizer
    {
      public:

        /*!
          \brief Constructor.

          \param r      The input raster.
          \param b      The selected band of the raster to be vectorized.
          \param mp     The maximum allowed number of polygons to be created (default = 0, unlimited).
          \param noDataExclusion If true, regions with no-data pixels will not generate geometries.
        */
        BoundaryExtract(Raster* r, std::size_t b, unsigned int mp = 0, const bool noDataExclusion = true );

        /*! \brief Destructor. */
        ~BoundaryExtract();

        
        /*!
          \brief Returns true if current algorithm implementation runs ok, false otherwise.

          \param polygons    The vector of polygons (will be cleared) to get the result of the vectorization.
          
          \param polygonsValues A pointer to a valid vector where the raster pixel values related to each polygon will be stored, or a NULL pointer.
          
          \note The caller of this method must take the ownership of the returned geometries and must delete them when necessary.
        */
        bool run(std::vector<te::gm::Geometry*>& polygons, double pixelvalue );

    protected:

      /*!
      \brief Detects an edge of a cell in Raster.

      \param segmentInitialCol abscissa (column) of the upper-left point of the shape
      \param segmentInitialRow ordinate (line) of the upper-left point of the shape
      \param line 2D Line.

      \return true if ok, otherwise false
      */
      bool detectEdge(long segmentInitialCol, long segmentInitialRow, te::gm::LinearRing& outputLine);

    };
  } // end namespace rst
}   // end namespace te

#endif  // __TERRALIB_RASTER_INTERNAL_BOUNDARYEXTRACT_H
