/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/raster/RasterSummaryManager.cpp

  \brief A singleton for keeping raster summaries (most statistics).
*/

// TerraLib
#include "../common/STLUtils.h"
#include "Band.h"
#include "BandProperty.h"
#include "Raster.h"
#include "RasterSummary.h"
#include "RasterSummaryManager.h"
#include "Utils.h"

// STL
#include <complex>

// ---------------------------------------------------------------------------

te::rst::RasterSummaryManager::SumManKey::SumManKey()
{
  m_bandIndex = 0;
  m_rowStart = 0;
  m_colStart = 0;
  m_finalRow = 0;
  m_finalCol = 0;
  m_histoBins = 0;
  m_sampleStep = 0;
  m_rasterPtr = 0;
  m_rasterPolicy = te::common::NoAccess;
}

te::rst::RasterSummaryManager::SumManKey::SumManKey(const SumManKey& other)
{
  operator=( other );
}

te::rst::RasterSummaryManager::SumManKey::~SumManKey()
{
}

bool te::rst::RasterSummaryManager::SumManKey::operator==( const SumManKey& other ) const
{
  return
    ( 
      ( m_bandIndex == other.m_bandIndex )
      &&
      ( m_rowStart == other.m_rowStart )
      &&
      ( m_colStart == other.m_colStart )
      &&
      ( m_finalRow == other.m_finalRow )
      &&
      ( m_finalCol == other.m_finalCol )
      &&
      ( m_histoBins == other.m_histoBins )
      &&
      ( m_sampleStep == other.m_sampleStep )
      &&
      ( m_connInfo == other.m_connInfo )
    );
}

bool te::rst::RasterSummaryManager::SumManKey::operator<( const SumManKey& other ) const
{
  return
    ( 
      ( m_bandIndex < other.m_bandIndex )
      &&
      ( m_rowStart < other.m_rowStart )
      &&
      ( m_colStart < other.m_colStart )
      &&
      ( m_finalRow < other.m_finalRow )
      &&
      ( m_finalCol < other.m_finalCol )
      &&
      ( m_histoBins < other.m_histoBins )
      &&
      ( m_sampleStep < other.m_sampleStep )
      &&
      ( m_connInfo < other.m_connInfo )
    );
}

te::rst::RasterSummaryManager::SumManKey& 
  te::rst::RasterSummaryManager::SumManKey::operator=(
  const SumManKey& other)
{
  m_bandIndex = other.m_bandIndex;
  m_rowStart = other.m_rowStart;
  m_colStart = other.m_colStart;
  m_finalRow = other.m_finalRow;
  m_finalCol = other.m_finalCol;
  m_histoBins = other.m_histoBins;
  m_sampleStep = other.m_sampleStep;
  m_rasterPtr = other.m_rasterPtr;
  m_rasterPolicy = other.m_rasterPolicy;
  m_connInfo = other.m_connInfo;
  m_summary = other.m_summary;
  
  return *this;
}

// ---------------------------------------------------------------------------

te::rst::RasterSummaryManager::~RasterSummaryManager()
{
}

te::rst::RasterSummaryManager::RasterSummaryManager()
{
  m_enableDeletedRastersCache = true;
    m_defRastersHistBins = 0;
}

bool te::rst::RasterSummaryManager::find(const Raster* raster, 
  RasterSummary& summary ) const
{
  assert( raster );
  const std::map<std::string, std::string> rInfo = raster->getInfo();
  
  summary.clear();
  
  int inputDataType = 0;
  std::unique_ptr< BandSummary > bandSummaryPtr;
    
  for( unsigned int bandIdx = 0 ; bandIdx < raster->getNumberOfBands() ; ++bandIdx )
  {
    inputDataType = raster->getBand( bandIdx )->getProperty()->getType();
    
    bandSummaryPtr.reset( new BandSummary() );
    if( ! find( raster, bandIdx, 0, 0, raster->getNumberOfRows() - 1,
      raster->getNumberOfColumns() - 1, m_defRastersHistBins, 1, *bandSummaryPtr ) )
    {
      return false;
    }
    
    summary.push_back( bandSummaryPtr.release() );
  }
  
  return true;
}

bool te::rst::RasterSummaryManager::find(
  const Raster* raster, const unsigned int bandIndex,
  const unsigned int rowStart, const unsigned int colStart,
  const unsigned int finalRow, const unsigned int finalCol,
  const unsigned int histoBins, const unsigned int sampleStep,
  BandSummary& summary ) const
{
  assert( raster );
 
  summary.clear();
  
  const std::map<std::string, std::string> m_connInfo = raster->getInfo();
  
  std::unique_lock< std::mutex > locker( m_syncMutex );
  
  SummaryContT::const_iterator it = m_rasterSummaries.begin();
  const SummaryContT::const_iterator itEnd = m_rasterSummaries.end();
    
  while( it != itEnd )
  {
    if(
        ( bandIndex == it->m_bandIndex )
        &&
        ( rowStart == it->m_rowStart )
        &&
        ( colStart == it->m_colStart )
        &&
        ( finalRow == it->m_finalRow )
        &&
        ( finalCol == it->m_finalCol )
        &&
        ( histoBins == it->m_histoBins )
        &&
        ( sampleStep == it->m_sampleStep )
        &&
        ( m_connInfo == it->m_connInfo )
      )
    {
      summary = it->m_summary;
      return true;
    }
    
    ++it;
  }
  
  return false;
}

void te::rst::RasterSummaryManager::remove(const Raster* raster)
{
  std::unique_lock< std::mutex > locker( m_syncMutex );  
  
  SummaryContT::const_iterator it = m_rasterSummaries.begin();
  SummaryContT::const_iterator it2;

  while( it != m_rasterSummaries.end() )
  {
    if( 
        ( raster == it->m_rasterPtr )
        &&
        ( 
          ( it->m_rasterPolicy & te::common::WAccess )
          ||
          ( !m_enableDeletedRastersCache )
        )
      )
    {
      it2 = it;
      ++it2;
      
      m_rasterSummaries.erase( it );
      it = it2;
    }
    else
    {
      ++it;
    }
  }
}

void te::rst::RasterSummaryManager::removeAll()
{
  std::unique_lock< std::mutex > locker( m_syncMutex ); 
  m_rasterSummaries.clear();
}

void te::rst::RasterSummaryManager::remove(const Raster* raster, 
  const unsigned int bandIndex,
  const unsigned int rowStart, const unsigned int colStart,
  const unsigned int finalRow, const unsigned int finalCol,
  const unsigned int histoBins, const unsigned int sampleStep)
{
  assert( raster ); 
  
  const std::map<std::string, std::string> connInfo = raster->getInfo();
  
  std::unique_lock< std::mutex > locker( m_syncMutex );  
  
  if( 
      ( raster->getAccessPolicy() & te::common::WAccess )
      ||
      ( !m_enableDeletedRastersCache )
    )
  {
    SummaryContT::const_iterator it =
      m_rasterSummaries.begin();
    SummaryContT::const_iterator it2;

    while( it != m_rasterSummaries.end() )
    {
      if(
          ( bandIndex == it->m_bandIndex )
          &&
          ( rowStart == it->m_rowStart )
          &&
          ( colStart == it->m_colStart )
          &&
          ( finalRow == it->m_finalRow )
          &&
          ( finalCol == it->m_finalCol )
          &&
          ( histoBins == it->m_histoBins )
          &&
          ( sampleStep == it->m_sampleStep )
          &&
          ( connInfo == it->m_connInfo )
          &&
          ( 
            ( it->m_rasterPolicy & te::common::WAccess )
            ||
            ( !m_enableDeletedRastersCache )
          )
        )
      {
        it2 = it;
        ++it2;
        
        m_rasterSummaries.erase( it );
        it = it2;
      }
      else
      {
        ++it;
      }
    }    
  }
}

bool te::rst::RasterSummaryManager::get( const Raster* raster,
  bool readall, RasterSummary& summary )
{
  assert( raster );
  
  summary.clear();
  
  std::unique_ptr< BandSummary > bSumPtr;
  
  for( unsigned int bandIdx = 0 ; bandIdx < raster->getNumberOfBands() ;
    ++bandIdx )
  {
    bSumPtr.reset( new BandSummary() );    
    
    if( ! get( raster, bandIdx, readall, *bSumPtr ) )
    {
      return false;
    }
    
    summary.push_back( bSumPtr.release() );
  }
  
  return true;
}

bool te::rst::RasterSummaryManager::get( const Raster* raster, 
  const unsigned int rowStart, 
  const unsigned int colStart, const unsigned int finalRow, 
  const unsigned int finalCol, const unsigned int histoBins, 
  const unsigned int sampleStep, bool readall, RasterSummary& summary )
{
  assert( raster );
  
  summary.clear();
  
  std::unique_ptr< BandSummary > bSumPtr;
  
  for( unsigned int bandIdx = 0 ; bandIdx < raster->getNumberOfBands() ;
    ++bandIdx )
  {
    bSumPtr.reset( new BandSummary() );
    
    if( ! get( raster, bandIdx, rowStart, colStart, finalRow,
      finalCol, histoBins, sampleStep, readall, *bSumPtr ) )
    {
      return false;
    }
    
    summary.push_back( bSumPtr.release() );
  }
  
  return true;
}

bool te::rst::RasterSummaryManager::get( const Raster* raster, 
  const unsigned int bandIndex, bool readall, BandSummary& summary )
{
  assert( raster );
  
  summary.clear();
  
  if( ! get( raster, bandIndex, 0, 0, raster->getNumberOfRows() -1,
    raster->getNumberOfColumns() - 1, m_defRastersHistBins, 1, readall, summary ) )
  {
    return false;
  }
  
  return true;
}

bool te::rst::RasterSummaryManager::get( 
  const Raster* raster, const unsigned int bandIndex,
  const unsigned int rowStart, const unsigned int colStart,
  const unsigned int finalRow, const unsigned int finalCol,
  const unsigned int histoBins, const unsigned int sampleStep,
  bool readall, BandSummary& summary )
{
  assert( raster );
  
  if( bandIndex >= raster->getNumberOfBands() ) return false;
  
  summary.clear();
  
  if( 
      ( !readall ) 
      && 
      find( raster, bandIndex, rowStart, colStart, finalRow,
        finalCol, histoBins, sampleStep, summary ) 
    )
  {
    return true;
  }
  else
  {
    remove( raster, bandIndex, rowStart, colStart, finalRow, finalCol,
      histoBins, sampleStep );
    
    const int inputDataType = raster->getBand( bandIndex )->getProperty()->getType();        
    
    SumManKey key;
    key.m_bandIndex = bandIndex;
    key.m_rowStart = rowStart;
    key.m_colStart = colStart;
    key.m_finalRow = finalRow;
    key.m_finalCol = finalCol;
    key.m_histoBins = histoBins;
    key.m_sampleStep = sampleStep;
    key.m_rasterPtr = raster;
    key.m_rasterPolicy = raster->getAccessPolicy();
    key.m_connInfo = raster->getInfo();
    key.m_summary.m_histogramR = new std::map<double, unsigned>();
    key.m_summary.m_histogramI = new std::map<double, unsigned>();
    
    if( ! getHistograms( *raster, bandIndex, key.m_rowStart, 
        key.m_colStart, key.m_finalRow, key.m_finalCol, key.m_histoBins,
        key.m_sampleStep, 0, *key.m_summary.m_histogramR,
        *key.m_summary.m_histogramI) )
    {
      return false;
    }
    
    double minR = 0;
    double maxR = 0;
    double meanR = 0;
    double stdDevR = 0;
    te::rst::getHistogramStats( *key.m_summary.m_histogramR, minR, maxR, meanR, stdDevR );    
    
    double minI = 0;
    double maxI = 0;
    double meanI = 0;
    double stdDevI = 0;
    te::rst::getHistogramStats( *key.m_summary.m_histogramI, minI, maxI, meanI, stdDevI );
     
    key.m_summary.m_minVal = new std::complex<double>( minR, minI );
    key.m_summary.m_maxVal = new std::complex<double>( maxR, maxI );
    key.m_summary.m_stdVal = new std::complex<double>( stdDevR, stdDevI );
    key.m_summary.m_meanVal = new std::complex<double>( meanR, meanI );
    
    summary = key.m_summary;
    
    std::unique_lock< std::mutex > locker( m_syncMutex );
    m_rasterSummaries.push_back( key );
  }

  return true;
}

void te::rst::RasterSummaryManager::enableDeletedRastersCache( const bool enabled )
{
  std::unique_lock< std::mutex > locker( m_syncMutex );
  m_enableDeletedRastersCache = enabled;
}

void te::rst::RasterSummaryManager::setDefaultRasterHistSize( const unsigned int binsNumber )
{
  std::unique_lock< std::mutex > locker( m_syncMutex );
    m_defRastersHistBins = binsNumber;
}

unsigned int te::rst::RasterSummaryManager::getDefaultRasterHistSize()
{
  std::unique_lock< std::mutex > locker( m_syncMutex );
  return m_defRastersHistBins;
}

