/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

#include "AlgorithmParametersSerializer.h"
#include "Macros.h"

#include <time.h>
#include <ctime>

#define TL_R_S_M_KEY "TERRALIB_RP_PARAMETERS"
#define TL_R_S_SEPARATOR_KEY "|"
#define TL_R_S_ALGO_NAME_KEY "RP_PARAM_ALGO_NAME"
#define TL_R_S_ALGO_DATE_KEY "RP_PARAM_DATE"
#define TL_R_S_PARAM_KEY "RP_PARAM"

namespace te
{
  namespace rp
  {
    AlgorithmParametersSerializer::AlgorithmParametersSerializer()
    {
    }

    AlgorithmParametersSerializer::~AlgorithmParametersSerializer()
    {
    }

    AlgorithmParametersSerializer::AlgorithmParametersSerializer(
      const AlgorithmParametersSerializer& other )
    {
      operator=( other );
    }

    AlgorithmParametersSerializer& AlgorithmParametersSerializer::operator=(
      const AlgorithmParametersSerializer& other )
    {
      m_algoName = other.m_algoName;
      m_params = other.m_params;

      return *this;
    }
    
    void AlgorithmParametersSerializer::setAlgorithmName( const std::string& name )
    {
      TERP_TRUE_OR_THROW( !name.empty(), "Invalid algorithm name" );
      m_algoName = name;
    }

    bool AlgorithmParametersSerializer::addParameter( const std::string& name,
      const std::string& value )
    {
      TERP_TRUE_OR_RETURN_FALSE( !name.empty(), "Invalid parameter name" );
      TERP_TRUE_OR_RETURN_FALSE( name.find( TL_R_S_SEPARATOR_KEY ) == std::string::npos,
        "Invalid parameter name" );
      TERP_TRUE_OR_RETURN_FALSE( value.find( TL_R_S_SEPARATOR_KEY ) == std::string::npos,
        "Invalid parameter value" );

      std::vector< std::string > auxVeluesVec;
      auxVeluesVec.push_back( value );

      m_params.insert( ParsContPairT( name, auxVeluesVec ) );

      return true;
    }

    bool AlgorithmParametersSerializer::addMultivaluedParameter( const std::string& name,
      const std::vector< std::string >& values )
    {
      TERP_TRUE_OR_RETURN_FALSE( !name.empty(), "Invalid parameter name" );
      TERP_TRUE_OR_RETURN_FALSE( name.find( TL_R_S_SEPARATOR_KEY ) == std::string::npos,
        "Invalid parameter name" );

      const std::size_t valuesSize = values.size();

      for( std::size_t valuesIdx = 0 ; valuesIdx < valuesSize ; ++valuesIdx )
      {
        TERP_TRUE_OR_RETURN_FALSE( values[ valuesIdx ].find(
          TL_R_S_SEPARATOR_KEY ) == std::string::npos, "Invalid parameter value" );
      }

      m_params.insert( ParsContPairT( name, values ) );

      return true;
    }

    void AlgorithmParametersSerializer::getMetaData( std::map<std::string,
      std::string>& metadata ) const
    {
      metadata.clear();

      if( m_algoName.empty() ) return;
      if( m_params.empty() ) return;

      std::time_t rawtime;
      std::time(&rawtime);
      struct std::tm * timeinfo = std::localtime(&rawtime);
      char strTime[80];
      std::strftime(strTime, 50, "%Y %m %d %H %M %S", timeinfo);

      metadata[ TL_R_S_M_KEY ].append( TL_R_S_ALGO_DATE_KEY );
      metadata[ TL_R_S_M_KEY ].append( TL_R_S_SEPARATOR_KEY TL_R_S_SEPARATOR_KEY );
      metadata[ TL_R_S_M_KEY ].append( strTime );

      metadata[ TL_R_S_M_KEY ].append( TL_R_S_SEPARATOR_KEY TL_R_S_SEPARATOR_KEY );
      metadata[ TL_R_S_M_KEY ].append( TL_R_S_ALGO_NAME_KEY );
      metadata[ TL_R_S_M_KEY ].append( TL_R_S_SEPARATOR_KEY TL_R_S_SEPARATOR_KEY );
      metadata[ TL_R_S_M_KEY ].append( m_algoName );

      ParsContT::const_iterator pIt = m_params.begin();
      const ParsContT::const_iterator pItEnd = m_params.end();
      std::size_t valuesSize = 0;
      std::size_t valuesIdx = 0;

      while( pIt != pItEnd )
      {
        metadata[ TL_R_S_M_KEY ].append( TL_R_S_SEPARATOR_KEY TL_R_S_SEPARATOR_KEY );
        metadata[ TL_R_S_M_KEY ].append( TL_R_S_PARAM_KEY );
        metadata[ TL_R_S_M_KEY ].append( TL_R_S_SEPARATOR_KEY TL_R_S_SEPARATOR_KEY );

        metadata[ TL_R_S_M_KEY ].append( "\"" );
        metadata[ TL_R_S_M_KEY ].append( pIt->first );
        metadata[ TL_R_S_M_KEY ].append( "\"=" );

        valuesSize = pIt->second.size();

        if( valuesSize )
        {
          for( valuesIdx = 0 ; valuesIdx < valuesSize ; ++valuesIdx )
          {
            if( valuesIdx ) metadata[ TL_R_S_M_KEY ].append( TL_R_S_SEPARATOR_KEY );

            metadata[ TL_R_S_M_KEY ].append( "\"" );
            metadata[ TL_R_S_M_KEY ].append( pIt->second[ valuesIdx ] );
            metadata[ TL_R_S_M_KEY ].append( "\"" );
          }
        }
        else
        {
          metadata[ TL_R_S_M_KEY ].append( "\"\"" );
        }

        ++pIt;
      }
    }

    void AlgorithmParametersSerializer::reset()
    {
      m_algoName.clear();
      m_params.clear();
    }
  } // end namespace rp
}   // end namespace te    

