/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/rp/Blender.cpp
  \brief Blended pixel value calculation for two overlaped rasters.
*/

#include "Blender.h"

#include "Macros.h"
#include "../geometry/LinearRing.h"
#include "../geometry/CircularString.h"
#include "../geometry/MultiPoint.h"
#include "../geometry/MultiLineString.h"
#include "../geometry/Point.h"
#include "../geometry/Coord2D.h"
#include "../geometry/CurvePolygon.h"
#include "../geometry/Envelope.h"
#include "../geometry/Enums.h"
#include "../raster/Raster.h"
#include "../raster/Grid.h"
#include "../raster/Band.h"
#include "../raster/BandProperty.h"
#include "../raster/Utils.h"
#include "../raster/SynchronizedRaster.h"
#include "../common/PlatformUtils.h"
#include "../common/MathUtils.h"
#include "../common/progress/TaskProgress.h"

#include <boost/graph/graph_concepts.hpp>
#include <boost/scoped_ptr.hpp>

#include <complex>
#include <limits>
#include <algorithm>
#include <memory>

// Get the perpendicular distance from a point P(pX,pY) from a line defined
// by the points A(lineAX,lineAY) and B(lineBX,lineBY)
// Requires two previously declared variables aux1 and aux2
#define getPerpendicularDistance( pX, pY, lineAX, lineAY, lineBX, lineBY, aux1, aux2, perpDist ) \
  aux1 = lineAX - lineBX; \
  aux2 = lineAY - lineBY; \
  if( aux1 == 0.0 ) \
  { \
    perpDist = std::abs( pX - lineAX ); \
  } \
  else if( aux2 == 0.0 ) \
  { \
    perpDist = std::abs( pY - lineAY ); \
  } \
  else \
  { \
    perpDist = \
      std::abs( \
        ( aux2 * pX ) - ( aux1 * pY ) + ( lineAX * lineBY ) - ( lineBX * lineAY ) \
      ) \
      / \
      std::sqrt( ( aux1 * aux1 ) + ( aux2 * aux2 ) ); \
  }

namespace te
{
  namespace rp
  {
    Blender::Blender() : 
      m_euclideanDistanceMethodImp_auxPoint(te::gm::Point(te::gm::PointType)),
      m_sumMethodImp_auxPoint(te::gm::Point(te::gm::PointType)),
      m_maxMethodImp_auxPoint(te::gm::Point(te::gm::PointType)),
      m_minMethodImp_auxPoint(te::gm::Point(te::gm::PointType))
    {


      initState();
    }
    
    Blender::~Blender()
    {
      clear();
    }    
    
    bool Blender::initialize( te::rst::Raster& raster1, 
      const std::vector< unsigned int >& raster1Bands, 
      const te::rst::Raster& raster2, 
      const std::vector< unsigned int >& raster2Bands,
      const BlendMethod& blendMethod, 
      const te::rst::Interpolator::Method& interpMethod1,
      const te::rst::Interpolator::Method& interpMethod2,
      const double& noDataValue,
      const bool forceRaster1NoDataValue,
      const bool forceRaster2NoDataValue,
      const std::vector< double >& pixelOffsets1,
      const std::vector< double >& pixelScales1,
      const std::vector< double >& pixelOffsets2,
      const std::vector< double >& pixelScales2,
      te::gm::MultiPolygon const * const r1ValidDataDelimiterPtr,
      te::gm::MultiPolygon const * const r2ValidDataDelimiterPtr,
      const te::gm::GeometricTransformation& geomTransformation,
      const unsigned int threadsNumber,
      const bool enableProgressInterface,
      BlenderRule const * const rulePtr,
      const bool enableRasterCache )
    {
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( 
        raster1.getAccessPolicy() & te::common::RAccess, 
        "Invalid raster 1" );        
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( 
        raster2.getAccessPolicy() & te::common::RAccess, 
        "Invalid raster 2" ); 
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( raster1Bands.size() > 0,
        "Invalid raster bands vector" );      
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( raster1Bands.size() ==
        raster2Bands.size(), "Invalid raster bands vector" );
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( pixelOffsets1.size() ==  
        raster1Bands.size(), "Invalid pixel offsets" );
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( pixelScales1.size() ==  
        raster1Bands.size(), "Invalid pixel scales" );        
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( pixelOffsets2.size() ==  
        raster2Bands.size(), "Invalid pixel offsets" );
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( pixelScales2.size() ==  
        raster2Bands.size(), "Invalid pixel scales" );        
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( ( r1ValidDataDelimiterPtr ?
        ( r1ValidDataDelimiterPtr->getNumGeometries() > 0 ) : true ),
        "Invalid geometry 1" )
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( ( r1ValidDataDelimiterPtr ?
        r1ValidDataDelimiterPtr->isValid() : true ),
        "Invalid geometry 1" )      
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( ( r2ValidDataDelimiterPtr ?
        ( r2ValidDataDelimiterPtr->getNumGeometries() > 0 ) : true ),
        "Invalid geometry 2" )
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( ( r2ValidDataDelimiterPtr ?
        r2ValidDataDelimiterPtr->isValid() : true ),
        "Invalid geometry 2" )      
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( geomTransformation.isValid(),
        "Invalid transformation" );
        
      clear();
      
      // defining the input rasters
      
      m_raster1Ptr = &raster1;
      m_raster2Ptr = &raster2;
      
      // Generating the valid data area points
      
      if( r1ValidDataDelimiterPtr )
      {
        m_r1ValidDataDelimiterPtr.reset( (te::gm::MultiPolygon*)r1ValidDataDelimiterPtr->clone() );
      }
      if( r2ValidDataDelimiterPtr )
      {
        m_r2ValidDataDelimiterPtr.reset( (te::gm::MultiPolygon*)r2ValidDataDelimiterPtr->clone() );
      }      
      
      // indexed under raster 1 lines/cols
      std::unique_ptr< te::gm::MultiPolygon > indexedDelimiter1Ptr(
        new te::gm::MultiPolygon( 0, te::gm::MultiPolygonType, 0, nullptr ) ); 
      
      if( r1ValidDataDelimiterPtr )
      {
        const te::rst::Grid& grid = (*raster1.getGrid());
        
        const std::size_t nGeoms = r1ValidDataDelimiterPtr->getNumGeometries();        
        
        for( std::size_t geomIdx = 0 ; geomIdx < nGeoms ; ++geomIdx )
        {
          te::gm::Polygon const* geomPtr = dynamic_cast< te::gm::Polygon* >(
            r1ValidDataDelimiterPtr->getGeometryN( geomIdx ) );
          TERP_DEBUG_TRUE_OR_THROW( geomPtr, "Invalid geometry pointer" );
          
          const std::size_t nRings = geomPtr->getNumRings();
          
          te::gm::Polygon* outPolPtr = new te::gm::Polygon( 0, te::gm::PolygonType, 0, nullptr );           
          
          for( std::size_t ringIdx = 0 ; ringIdx < nRings ; ++ringIdx )
          {
            te::gm::LinearRing const* inRingPtr = dynamic_cast< te::gm::LinearRing const* >( 
              geomPtr->getRingN( ringIdx ) );
            assert( inRingPtr );
            
            const std::size_t nPoints = inRingPtr->getNPoints();
            te::gm::Coord2D const * inCoordsPtr = inRingPtr->getCoordinates();
            te::gm::Coord2D auxCoord;        
            
            te::gm::LinearRing* outRingPtr = new te::gm::LinearRing( nPoints, 
               te::gm::LineStringType, 0, nullptr );
            
            for( std::size_t pIdx = 0 ; pIdx < nPoints ; ++pIdx )
            {
              grid.geoToGrid( inCoordsPtr[ pIdx ].x, inCoordsPtr[ pIdx ].y, 
                auxCoord.x, auxCoord.y ); 
              outRingPtr->setPoint( pIdx, auxCoord.x, auxCoord.y );
            }
            
            outPolPtr->add( outRingPtr );          
            
          }
          
          indexedDelimiter1Ptr->add( outPolPtr );
        }
      }
      else
      {
        te::gm::LinearRing* outRingPtr = new te::gm::LinearRing( 5, 
           te::gm::LineStringType, 0, nullptr );        
        
        outRingPtr->setPoint( 0, 
           -0.5, 
           -0.5 );
        outRingPtr->setPoint( 1, 
           ((double)raster1.getNumberOfColumns()) - 0.5, 
           -0.5 );
        outRingPtr->setPoint( 2, 
           ((double)raster1.getNumberOfColumns()) - 0.5, 
           ((double)raster1.getNumberOfRows()) - 0.5 );
        outRingPtr->setPoint( 3, 
           -0.5, 
           ((double)raster1.getNumberOfRows()) - 0.5 );
        outRingPtr->setPoint( 4, 
           -0.5, 
           -0.5 );
        
        te::gm::Polygon* outPolPtr = new te::gm::Polygon( 0, te::gm::PolygonType, 0, nullptr );           
        
        outPolPtr->add( outRingPtr );
           
        indexedDelimiter1Ptr->add( outPolPtr );
      }
      
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( indexedDelimiter1Ptr->isValid(),
        "Invalid indexed geometry 1" );       
      
      // indexed under raster 1 lines/cols
      std::unique_ptr< te::gm::MultiPolygon > indexedDelimiter2Ptr(
        new te::gm::MultiPolygon( 0, te::gm::MultiPolygonType, 0, nullptr ) ); 
      
      if( r2ValidDataDelimiterPtr )
      {
        const te::rst::Grid& grid = (*raster2.getGrid());
        
        const std::size_t nGeoms = r2ValidDataDelimiterPtr->getNumGeometries();
        
        for( std::size_t geomIdx = 0 ; geomIdx < nGeoms ; ++geomIdx )
        {        
          te::gm::Polygon const* geomPtr = dynamic_cast< te::gm::Polygon* >(
            r2ValidDataDelimiterPtr->getGeometryN( geomIdx ) );
          TERP_DEBUG_TRUE_OR_THROW( geomPtr, "Invalid geometry pointer" );          
          
          const std::size_t nRings = geomPtr->getNumRings();
          
          te::gm::Polygon* outPolPtr = new te::gm::Polygon( 0, te::gm::PolygonType, 0, nullptr ); 
          
          for( std::size_t ringIdx = 0 ; ringIdx < nRings ; ++ringIdx )
          {
            te::gm::LinearRing const* inRingPtr = dynamic_cast< te::gm::LinearRing const* >( 
              geomPtr->getRingN( ringIdx ) );
            assert( inRingPtr );
            
            const std::size_t nPoints = inRingPtr->getNPoints();
            te::gm::Coord2D const * inCoordsPtr = inRingPtr->getCoordinates();
            te::gm::Coord2D auxCoord;        
            te::gm::Coord2D auxCoord2;        
            
            te::gm::LinearRing* outRingPtr = new te::gm::LinearRing( nPoints, 
               te::gm::LineStringType, 0, nullptr );
            
            for( std::size_t pIdx = 0 ; pIdx < nPoints ; ++pIdx )
            {
              grid.geoToGrid( inCoordsPtr[ pIdx ].x, inCoordsPtr[ pIdx ].y, 
                auxCoord.x, auxCoord.y ); 
              geomTransformation.inverseMap( auxCoord.x, auxCoord.y, auxCoord2.x, auxCoord2.y );
              outRingPtr->setPoint( pIdx, auxCoord2.x, auxCoord2.y );
            }
            
            outPolPtr->add( outRingPtr );
          }
          
          indexedDelimiter2Ptr->add( outPolPtr );
        }
      }
      else
      {
        te::gm::LinearRing* outRingPtr = new te::gm::LinearRing( 5, 
           te::gm::LineStringType, 0, nullptr );          
        
        te::gm::Coord2D auxCoord;
        
        geomTransformation.inverseMap( 
          -0.5,
          -0.5,
          auxCoord.x, auxCoord.y );
        outRingPtr->setPoint( 0, auxCoord.x, auxCoord.y );
        outRingPtr->setPoint( 4, auxCoord.x, auxCoord.y ); 
        
        geomTransformation.inverseMap( 
          ((double)raster2.getNumberOfColumns()) - 0.5, 
           -0.5,
          auxCoord.x, auxCoord.y );
        outRingPtr->setPoint( 1, auxCoord.x, auxCoord.y );
        
        geomTransformation.inverseMap( 
          ((double)raster2.getNumberOfColumns()) - 0.5, 
          ((double)raster2.getNumberOfRows()) - 0.5,
          auxCoord.x, auxCoord.y );
        outRingPtr->setPoint( 2, auxCoord.x, auxCoord.y );
        
        geomTransformation.inverseMap( 
          -0.5, 
         ((double)raster2.getNumberOfRows()) - 0.5,
          auxCoord.x, auxCoord.y );
        outRingPtr->setPoint( 3, auxCoord.x, auxCoord.y );
        
        te::gm::Polygon* outPolPtr = new te::gm::Polygon( 0, te::gm::PolygonType, 0, nullptr );   
        
        outPolPtr->add( outRingPtr );
        
        indexedDelimiter2Ptr->add( outPolPtr );
      }
      
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( indexedDelimiter2Ptr->isValid(),
        "Invalid indexed geometry 1" );
      
      // Calculating the intersection (raster 1 lines/cols)
      
      try
      {
        m_intersectionPtr.reset( indexedDelimiter2Ptr->intersection( 
          indexedDelimiter1Ptr.get() ) );
      }
      catch( const std::exception& e )
      {
        TERP_INSTANCE_LOG_AND_RETURN_FALSE( "Indexed geometry intersection:" + 
          std::string( e.what() ) );
      }
      
      // Initializing the intersection tile indexer
      
      if( m_intersectionPtr.get() )
      {
        TERP_TRUE_OR_THROW( getTileIndexers( m_intersectionPtr.get(), 
          m_intersectionTileIndexers ), "Intersection tile indexers creation error" );
      }
      
      // Extracting the intersection segments points
      
      if( m_intersectionPtr.get() )
      {
        std::size_t ringIdx = 0;
        std::unique_ptr< te::gm::Geometry > ringIntersectionPtr;
        std::size_t nPols = 0;
        std::size_t polIdx = 0;
        
        nPols = indexedDelimiter2Ptr->getNumGeometries();           
        for( polIdx = 0 ; polIdx < nPols ; ++polIdx )
        {
          te::gm::Polygon const* polPtr = dynamic_cast< te::gm::Polygon* >(
            indexedDelimiter2Ptr->getGeometryN( polIdx ) );
          
          if( polPtr )
          {
            for( ringIdx = 0 ; ringIdx < polPtr->getNumRings() ;
              ++ringIdx )
            {
              ringIntersectionPtr.reset( indexedDelimiter1Ptr->intersection( 
                polPtr->getRingN( ringIdx ) ) );
              
              if( ringIntersectionPtr.get() != nullptr ) 
              {
                getSegments( ringIntersectionPtr.get(), m_r2IntersectionSegmentsPoints );
              }
            }
          }
        }
        
        nPols = indexedDelimiter1Ptr->getNumGeometries();          
        for( polIdx = 0 ; polIdx < nPols ; ++polIdx )
        {
          te::gm::Polygon const* polPtr = dynamic_cast< te::gm::Polygon* >(
            indexedDelimiter1Ptr->getGeometryN( polIdx ) );
          
          if( polPtr )
          {
            for( ringIdx = 0 ; ringIdx < polPtr->getNumRings() ;
              ++ringIdx )
            {
              ringIntersectionPtr.reset( indexedDelimiter2Ptr->intersection( 
                polPtr->getRingN( ringIdx ) ) );
              
              if( ringIntersectionPtr.get() != nullptr ) 
              {
                getSegments( ringIntersectionPtr.get(), m_r1IntersectionSegmentsPoints );
              }
            }    
          }
        }    

/*        std::cout << std::endl;
        for( unsigned int idx = 0 ; idx < m_r1IntersectionSegmentsPoints.size() ; ++idx )
        {
          std::cout << std::endl << "m_r1IntersectionSegmentsPoints[" << idx << "]=" 
            << m_r1IntersectionSegmentsPoints[ idx ].first.x
            << " " << m_r1IntersectionSegmentsPoints[ idx ].first.y
            << " " << m_r1IntersectionSegmentsPoints[ idx ].second.x
            << " " << m_r1IntersectionSegmentsPoints[ idx ].second.y;            
        }
        for( unsigned int idx = 0 ; idx < m_r2IntersectionSegmentsPoints.size() ; ++idx )
        {
          std::cout << std::endl << "m_r2IntersectionSegmentsPoints[" << idx << "]=" 
            << m_r2IntersectionSegmentsPoints[ idx ].first.x
            << " " << m_r2IntersectionSegmentsPoints[ idx ].first.y
            << " " << m_r2IntersectionSegmentsPoints[ idx ].second.x
            << " " << m_r2IntersectionSegmentsPoints[ idx ].second.y;            
        }   
        std::cout << std::endl;  */   
          
        m_r2IntersectionSegmentsPointsSize = (unsigned int)m_r2IntersectionSegmentsPoints.size();
        m_r1IntersectionSegmentsPointsSize = (unsigned int)m_r1IntersectionSegmentsPoints.size();
      }
      
      // defining the blending method
      
      switch( blendMethod )
      {
        case te::rp::blender::NoBlendMethod :
        {
          m_blendMethod = te::rp::blender::NoBlendMethod;
          break;
        }
        case te::rp::blender::EuclideanDistanceMethod :
        {
          if( ( m_intersectionPtr.get() != nullptr ) && 
            ( m_r1IntersectionSegmentsPointsSize > 1 ) && 
            ( m_r2IntersectionSegmentsPointsSize > 1 ) )
          {
            m_blendMethod = te::rp::blender::EuclideanDistanceMethod;
          }
          else
          {
            m_blendMethod = te::rp::blender::NoBlendMethod;
          }
          break;
        }
        case te::rp::blender::SumMethod :
        {
          if( ( m_intersectionPtr.get() != nullptr ) && 
            ( m_r1IntersectionSegmentsPointsSize > 1 ) && 
            ( m_r2IntersectionSegmentsPointsSize > 1 ) )
          {
            m_blendMethod = te::rp::blender::SumMethod;
          }
          else
          {
            m_blendMethod = te::rp::blender::NoBlendMethod;
          }
          break;
        }
        case te::rp::blender::MaxMethod :
        {
          if( ( m_intersectionPtr.get() != nullptr ) && 
            ( m_r1IntersectionSegmentsPointsSize > 1 ) && 
            ( m_r2IntersectionSegmentsPointsSize > 1 ) )
          {
            m_blendMethod = te::rp::blender::MaxMethod;
          }
          else
          {
            m_blendMethod = te::rp::blender::NoBlendMethod;
          }
          break;
        }        
        case te::rp::blender::MinMethod :
        {
          if( ( m_intersectionPtr.get() != nullptr ) && 
            ( m_r1IntersectionSegmentsPointsSize > 1 ) && 
            ( m_r2IntersectionSegmentsPointsSize > 1 ) )
          {
            m_blendMethod = te::rp::blender::MinMethod;
          }
          else
          {
            m_blendMethod = te::rp::blender::NoBlendMethod;
          }
          break;
        }         
        case te::rp::blender::RuleMethod :
        {
          m_blendMethod = te::rp::blender::RuleMethod;
          break;
        }        
        default :
        {
          TERP_LOG_AND_THROW( "Invalid blend method" );
          break;
        }
      }      
      
      // Defining the blending function pointers
      
      setBlendFunctionsPonters( m_blendMethod );
              
      // defining the geometric transformation  
        
      m_geomTransformationPtr = geomTransformation.clone();
      
      // defining the interpolators
      
      if( forceRaster1NoDataValue )
      {
        std::vector< std::complex<double> > noDataValues1( raster1.getNumberOfBands(),
          std::complex<double>( noDataValue, 0.0 ) ); 
        m_interp1Ptr.reset( new te::rst::Interpolator( &raster1, interpMethod1,
          noDataValues1 ) );
      }
      else
      {
        m_interp1Ptr.reset( new te::rst::Interpolator( &raster1, interpMethod1 ) );
      }      
      
      if( forceRaster2NoDataValue )
      {
        std::vector< std::complex<double> > noDataValues2( raster2.getNumberOfBands(),
          std::complex<double>( noDataValue, 0.0 ) );
        m_interp2Ptr.reset( new te::rst::Interpolator( &raster2, interpMethod2,
          noDataValues2 ) );
      }
      else
      {
        m_interp2Ptr.reset( new te::rst::Interpolator( &raster2, interpMethod2 ) );
      }
        
      m_interpMethod1 = interpMethod1;
      m_interpMethod2 = interpMethod2;
      
      // defining dummy values
      
      m_forceRaster1NoDataValue = forceRaster1NoDataValue;
      m_forceRaster2NoDataValue = forceRaster2NoDataValue;
      
      for( std::vector< unsigned int >::size_type rasterBandsIdx = 0 ; 
        rasterBandsIdx < raster1Bands.size() ; ++rasterBandsIdx )
      {
        TERP_INSTANCE_TRUE_OR_RETURN_FALSE( raster1Bands[ rasterBandsIdx ] <
          raster1.getNumberOfBands(), "Invalid band" );
        TERP_INSTANCE_TRUE_OR_RETURN_FALSE( raster2Bands[ rasterBandsIdx ] <
          raster2.getNumberOfBands(), "Invalid band" );            
        
        if( forceRaster1NoDataValue )
        {
          m_raster1NoDataValues.push_back( noDataValue );
        }
        else
        {
          m_raster1NoDataValues.push_back( raster1.getBand( raster1Bands[ 
            rasterBandsIdx ] )->getProperty()->m_noDataValue );
        }
        
        if( forceRaster2NoDataValue )
        {
          m_raster2NoDataValues.push_back( noDataValue );
        }
        else
        {
          m_raster2NoDataValues.push_back( raster2.getBand( raster2Bands[ 
            rasterBandsIdx ] )->getProperty()->m_noDataValue );
        }        
      }
      
      m_outputNoDataValue = noDataValue;
      
      // defining raster bands
      
      m_raster1Bands = raster1Bands;
      m_raster1BandsSize = m_raster1Bands.size();
      
      m_raster2Bands = raster2Bands;
      m_raster2BandsSize = m_raster2Bands.size();
      
      // defining pixel offsets
      
      m_pixelOffsets1 = pixelOffsets1;
      m_pixelScales1 = pixelScales1;
      
      m_pixelOffsets2 = pixelOffsets2;
      m_pixelScales2 = pixelScales2;
      
      // threads
      
      if( threadsNumber == 0 )
      {
        m_threadsNumber = te::common::GetPhysProcNumber();
      }
      else
      {
        m_threadsNumber = threadsNumber;
      }
      
      // Rules
      
      if( blendMethod == te::rp::blender::RuleMethod ) 
      {
        TERP_INSTANCE_TRUE_OR_RETURN_FALSE( rulePtr != nullptr, "No rule provided" );
        m_rulePtr.reset( rulePtr->clone() );
        TERP_INSTANCE_TRUE_OR_RETURN_FALSE( 
          m_rulePtr->initialize(
            raster1,
            raster1Bands,
            raster2,
            raster2Bands,
            interpMethod1,
            interpMethod2,
            noDataValue,
            forceRaster1NoDataValue,
            forceRaster2NoDataValue,
            pixelOffsets1,
            pixelScales1,
            pixelOffsets1,
            pixelScales2,
            r1ValidDataDelimiterPtr,
            r2ValidDataDelimiterPtr,
            geomTransformation
          ),
          "Rule initialization error - " + m_rulePtr->getErrorMessage() );
      }
      
      // cache 
      
      m_enableRasterCache = enableRasterCache;
      
      // internal variables used by internal methods
      
      m_noBlendMethodImp_pixel1Values.resize( m_raster1BandsSize );
      m_noBlendMethodImp_pixel2Values.resize( m_raster2BandsSize );
      
      m_euclideanDistanceMethodImp_pixel1Values.resize( m_raster1BandsSize );
      m_euclideanDistanceMethodImp_pixel2Values.resize( m_raster2BandsSize );
      
      // progress interface
      
      m_enableProgressInterface = enableProgressInterface;
      
      return true;
    }
    
    void Blender::initState()
    {
      m_enableProgressInterface = false;
      m_forceRaster1NoDataValue = false;
      m_forceRaster2NoDataValue = false;
      m_threadsNumber = 0;
      m_blendMethod = te::rp::blender::InvalidBlendMethod;
      m_primaryBlendFuncPtr = nullptr;
      m_raster1Ptr = nullptr;
      m_raster2Ptr = nullptr;
      m_r1IntersectionSegmentsPointsSize = 0;
      m_r2IntersectionSegmentsPointsSize = 0;
      m_geomTransformationPtr = nullptr;
      m_interpMethod1 = te::rst::NearestNeighbor;
      m_interpMethod2 = te::rst::NearestNeighbor;
      m_outputNoDataValue = 0;
      m_enableRasterCache = true;
    }
    
    void Blender::clear()
    {
      m_r1ValidDataDelimiterPtr.reset();
      m_r2ValidDataDelimiterPtr.reset();
      m_intersectionPtr.reset();
      m_r1IntersectionSegmentsPoints.clear();
      m_r2IntersectionSegmentsPoints.clear();
      if( m_geomTransformationPtr ) delete m_geomTransformationPtr;
      m_interp1Ptr.reset();
      m_interp2Ptr.reset();
      m_raster1Bands.clear();
      m_raster2Bands.clear();
      m_pixelOffsets1.clear();
      m_pixelScales1.clear();
      m_pixelOffsets2.clear();
      m_pixelScales2.clear();
      m_raster1NoDataValues.clear();
      m_raster2NoDataValues.clear();
      m_intersectionTileIndexers.clear();
      m_rulePtr.reset();
      m_errorMessage.clear();

      initState();
    }
    
    void Blender::setBlendFunctionsPonters( const BlendMethod blendMethod )
    {
      switch( blendMethod )
      {
        case te::rp::blender::NoBlendMethod :
        {
          m_primaryBlendFuncPtr = &Blender::noBlendMethodImp;
          break;
        }
        case te::rp::blender::EuclideanDistanceMethod :
        {
          m_primaryBlendFuncPtr = &Blender::euclideanDistanceMethodImp;
          break;
        }  
        case te::rp::blender::SumMethod :
        {
          m_primaryBlendFuncPtr = &Blender::sumMethodImp;
          break;
        }
        case te::rp::blender::MaxMethod :
        {
          m_primaryBlendFuncPtr = &Blender::maxMethodImp;
          break;
        }
        case te::rp::blender::MinMethod :
        {
          m_primaryBlendFuncPtr = &Blender::minMethodImp;
          break;
        }
        case te::rp::blender::RuleMethod :
        {
          m_primaryBlendFuncPtr = &Blender::ruleMethodImp;
          break;
        }                        
        default :
        {
          TERP_LOG_AND_THROW( "Invalid blend method" );
          break;
        }
      }  
    }
    
    void Blender::noBlendMethodImp( const double& line, const double& col,
      double* const values )
    {
      // Finding the point over the second raster
      
      m_geomTransformationPtr->directMap( col, line, m_noBlendMethodImp_Point2Col,
        m_noBlendMethodImp_Point2Line );
      
      // Blending values
      
      m_noBlendMethodImp_pixel1IsValid = true;
      m_noBlendMethodImp_pixel2IsValid = true;
      
      for( m_noBlendMethodImp_BandIdx = 0 ; m_noBlendMethodImp_BandIdx <
        m_raster1BandsSize ; ++m_noBlendMethodImp_BandIdx )
      {
        m_interp1Ptr->getValue( col, line, 
          m_noBlendMethodImp_pixel1Values[ m_noBlendMethodImp_BandIdx ], 
          m_raster1Bands[ m_noBlendMethodImp_BandIdx ] );
        if( 
            m_noBlendMethodImp_pixel1Values[ m_noBlendMethodImp_BandIdx ]
            ==
             m_raster1NoDataValues[ m_noBlendMethodImp_BandIdx ]
          )
        {
          m_noBlendMethodImp_pixel1IsValid = false;
        }
        
        m_interp2Ptr->getValue( m_noBlendMethodImp_Point2Col, 
          m_noBlendMethodImp_Point2Line,
          m_noBlendMethodImp_pixel2Values[ m_noBlendMethodImp_BandIdx ],
          m_raster2Bands[ m_noBlendMethodImp_BandIdx ]);        
        
        if( 
            m_noBlendMethodImp_pixel2Values[ m_noBlendMethodImp_BandIdx ]
            ==
            m_raster2NoDataValues[ m_noBlendMethodImp_BandIdx ]
          )
        {
           m_noBlendMethodImp_pixel2IsValid = false;
        }
      }
      
      if( m_noBlendMethodImp_pixel1IsValid )
      {        
        for( m_noBlendMethodImp_BandIdx = 0 ; m_noBlendMethodImp_BandIdx <
          m_raster1BandsSize ; ++m_noBlendMethodImp_BandIdx )
        {                        
          values[ m_noBlendMethodImp_BandIdx ] = 
            (
              m_noBlendMethodImp_pixel1Values[ m_noBlendMethodImp_BandIdx ].real()
              *
              m_pixelScales1[ m_noBlendMethodImp_BandIdx ]
            )
            +
            m_pixelOffsets1[ m_noBlendMethodImp_BandIdx ];
        }
      }
      else if( m_noBlendMethodImp_pixel2IsValid )
      {        
        for( m_noBlendMethodImp_BandIdx = 0 ; m_noBlendMethodImp_BandIdx <
          m_raster2BandsSize ; ++m_noBlendMethodImp_BandIdx )
        {                        
          values[ m_noBlendMethodImp_BandIdx ] =
            (
              m_noBlendMethodImp_pixel2Values[ m_noBlendMethodImp_BandIdx ].real()
              *
              m_pixelScales2[ m_noBlendMethodImp_BandIdx ]
            )
            +
            m_pixelOffsets2[ m_noBlendMethodImp_BandIdx ];
        }
      }          
      else
      {
        for( m_noBlendMethodImp_BandIdx = 0 ; m_noBlendMethodImp_BandIdx <
          m_raster1BandsSize ; ++m_noBlendMethodImp_BandIdx )
        {                        
          values[ m_noBlendMethodImp_BandIdx ] = m_outputNoDataValue;
        }
      }
    }
    
    void Blender::euclideanDistanceMethodImp( const double& line, const double& col,
      double* const values )
    {
      TERP_DEBUG_TRUE_OR_THROW( m_intersectionPtr.get(), "Invalid intersection pointer" );
      TERP_DEBUG_TRUE_OR_THROW( m_r1IntersectionSegmentsPointsSize > 1, "Invalid intersection points" );
      TERP_DEBUG_TRUE_OR_THROW( m_r2IntersectionSegmentsPointsSize > 1, "Invalid intersection points" );
      
      // Checking if it is inside the intersection
      
      m_euclideanDistanceMethodImp_auxPoint.setX( col );
      m_euclideanDistanceMethodImp_auxPoint.setY( line );
      
      m_euclideanDistanceMethodImp_PointInsideIntersection = false;
      for( m_euclideanDistanceMethodImp_IntersectionTileIndexersIdx = 0;
        m_euclideanDistanceMethodImp_IntersectionTileIndexersIdx < m_intersectionTileIndexers.size() ;
        ++m_euclideanDistanceMethodImp_IntersectionTileIndexersIdx )
      {
        if( m_intersectionTileIndexers[ 
          m_euclideanDistanceMethodImp_IntersectionTileIndexersIdx ].within( 
          m_euclideanDistanceMethodImp_auxPoint ) )
        {
          m_euclideanDistanceMethodImp_PointInsideIntersection = true;
          break;
        }
      }
      
      // Blending if the point is inside the intersection
      
      if( m_euclideanDistanceMethodImp_PointInsideIntersection )
      {
        // Finding distances to both rasters valid area delimiters
              
        m_euclideanDistanceMethodImp_dist1 = std::numeric_limits<double>::max();
        for( m_euclideanDistanceMethodImp_vecIdx = 0 ; 
          m_euclideanDistanceMethodImp_vecIdx < m_r1IntersectionSegmentsPointsSize ; 
          ++m_euclideanDistanceMethodImp_vecIdx )
        {
          
          getPerpendicularDistance( 
            col,
            line,
            m_r1IntersectionSegmentsPoints[ m_euclideanDistanceMethodImp_vecIdx ].first.x,
            m_r1IntersectionSegmentsPoints[ m_euclideanDistanceMethodImp_vecIdx ].first.y, 
            m_r1IntersectionSegmentsPoints[ m_euclideanDistanceMethodImp_vecIdx ].second.x,
            m_r1IntersectionSegmentsPoints[ m_euclideanDistanceMethodImp_vecIdx ].second.y,           
            m_euclideanDistanceMethodImp_aux1,
            m_euclideanDistanceMethodImp_aux2,
            m_euclideanDistanceMethodImp_currDist );
            
          if( m_euclideanDistanceMethodImp_currDist < m_euclideanDistanceMethodImp_dist1 )
          {
            m_euclideanDistanceMethodImp_dist1 = m_euclideanDistanceMethodImp_currDist;
          }
        }     
        
        m_euclideanDistanceMethodImp_dist2 = std::numeric_limits<double>::max();
        for( m_euclideanDistanceMethodImp_vecIdx = 0 ; 
          m_euclideanDistanceMethodImp_vecIdx < m_r2IntersectionSegmentsPointsSize ; 
          ++m_euclideanDistanceMethodImp_vecIdx )
        {
          
          getPerpendicularDistance( 
            col,
            line,
            m_r2IntersectionSegmentsPoints[ m_euclideanDistanceMethodImp_vecIdx ].first.x,
            m_r2IntersectionSegmentsPoints[ m_euclideanDistanceMethodImp_vecIdx ].first.y, 
            m_r2IntersectionSegmentsPoints[ m_euclideanDistanceMethodImp_vecIdx ].second.x,
            m_r2IntersectionSegmentsPoints[ m_euclideanDistanceMethodImp_vecIdx ].second.y,           
            m_euclideanDistanceMethodImp_aux1,
            m_euclideanDistanceMethodImp_aux2,
            m_euclideanDistanceMethodImp_currDist );
            
          if( m_euclideanDistanceMethodImp_currDist < m_euclideanDistanceMethodImp_dist2 )
          {
            m_euclideanDistanceMethodImp_dist2 = m_euclideanDistanceMethodImp_currDist;
          }
        } 
        
        // Finding the point over the second raster
        
        m_geomTransformationPtr->directMap( col, line, m_euclideanDistanceMethodImp_Point2Col,
          m_euclideanDistanceMethodImp_Point2Line );      
        
        // Blending values
        
        m_euclideanDistanceMethodImp_pixel1IsValid = true;
        m_euclideanDistanceMethodImp_pixel2IsValid = true;
        
        for( m_euclideanDistanceMethodImp_BandIdx = 0 ; m_euclideanDistanceMethodImp_BandIdx <
          m_raster1BandsSize ; ++m_euclideanDistanceMethodImp_BandIdx )
        {
          m_interp1Ptr->getValue( col, line, 
            m_euclideanDistanceMethodImp_pixel1Values[ m_euclideanDistanceMethodImp_BandIdx ], 
            m_raster1Bands[ m_euclideanDistanceMethodImp_BandIdx ] );
          if( 
              m_euclideanDistanceMethodImp_pixel1Values[ m_euclideanDistanceMethodImp_BandIdx ]
              ==
               m_raster1NoDataValues[ m_euclideanDistanceMethodImp_BandIdx ]
            )
          {
            m_euclideanDistanceMethodImp_pixel1IsValid = false;
          }
          
          m_interp2Ptr->getValue( m_euclideanDistanceMethodImp_Point2Col, 
            m_euclideanDistanceMethodImp_Point2Line, 
            m_euclideanDistanceMethodImp_pixel2Values[ m_euclideanDistanceMethodImp_BandIdx ], 
            m_raster2Bands[ m_euclideanDistanceMethodImp_BandIdx ] );            
          if( 
              m_euclideanDistanceMethodImp_pixel2Values[ m_euclideanDistanceMethodImp_BandIdx ]
              ==
               m_raster2NoDataValues[ m_euclideanDistanceMethodImp_BandIdx ]
            )
          {
            m_euclideanDistanceMethodImp_pixel2IsValid = false;
          }
        }        
          
        if( 
            m_euclideanDistanceMethodImp_pixel1IsValid
            &&
            m_euclideanDistanceMethodImp_pixel2IsValid
          )
        {
          for( m_euclideanDistanceMethodImp_BandIdx = 0 ; m_euclideanDistanceMethodImp_BandIdx <
            m_raster1BandsSize ; ++m_euclideanDistanceMethodImp_BandIdx )
          {
            values[ m_euclideanDistanceMethodImp_BandIdx ] =
              (
                (
                  (
                    (
                      m_euclideanDistanceMethodImp_pixel1Values[ m_euclideanDistanceMethodImp_BandIdx ].real()  
                      *
                      m_pixelScales1[ m_euclideanDistanceMethodImp_BandIdx ] 
                      
                    )
                    +
                    m_pixelOffsets1[ m_euclideanDistanceMethodImp_BandIdx ]
                  )
                  *
                  m_euclideanDistanceMethodImp_dist1
                )
                +
                (
                  (
                    (
                      m_euclideanDistanceMethodImp_pixel2Values[ m_euclideanDistanceMethodImp_BandIdx ].real()
                      * 
                      m_pixelScales2[ m_euclideanDistanceMethodImp_BandIdx ] 
                    )
                    +
                    m_pixelOffsets2[ m_euclideanDistanceMethodImp_BandIdx ]
                  )
                  *
                  m_euclideanDistanceMethodImp_dist2
                )
              )
              /
              ( 
                m_euclideanDistanceMethodImp_dist1 
                +
                m_euclideanDistanceMethodImp_dist2
              );
          }
        }
        else if( m_euclideanDistanceMethodImp_pixel1IsValid )
        {
          for( m_euclideanDistanceMethodImp_BandIdx = 0 ; m_euclideanDistanceMethodImp_BandIdx <
            m_raster1BandsSize ; ++m_euclideanDistanceMethodImp_BandIdx )
          {
            values[ m_euclideanDistanceMethodImp_BandIdx ] = 
              ( 
                m_euclideanDistanceMethodImp_pixel1Values[ m_euclideanDistanceMethodImp_BandIdx ].real() 
                *
                m_pixelScales1[ m_euclideanDistanceMethodImp_BandIdx ]
              )
              +
              m_pixelOffsets1[ m_euclideanDistanceMethodImp_BandIdx ];
          }
        }
        else if( m_euclideanDistanceMethodImp_pixel2IsValid )
        {
          for( m_euclideanDistanceMethodImp_BandIdx = 0 ; m_euclideanDistanceMethodImp_BandIdx <
            m_raster2BandsSize ; ++m_euclideanDistanceMethodImp_BandIdx )
          {
            values[ m_euclideanDistanceMethodImp_BandIdx ] = 
              ( 
                m_euclideanDistanceMethodImp_pixel2Values[ m_euclideanDistanceMethodImp_BandIdx ].real() 
                *
                m_pixelScales2[ m_euclideanDistanceMethodImp_BandIdx ]
              )
              +
              m_pixelOffsets2[ m_euclideanDistanceMethodImp_BandIdx ];
          }
        }          
        else
        {
          for( m_euclideanDistanceMethodImp_BandIdx = 0 ; m_euclideanDistanceMethodImp_BandIdx <
            m_raster1BandsSize ; ++m_euclideanDistanceMethodImp_BandIdx )
          {
            values[ m_euclideanDistanceMethodImp_BandIdx ] = m_outputNoDataValue;
          }
        }
      }
      else
      {
        noBlendMethodImp( line, col, values );
      }
    }    

    void Blender::sumMethodImp( const double& line, const double& col,
      double* const values )
    {
      TERP_DEBUG_TRUE_OR_THROW( m_intersectionPtr.get(), "Invalid intersection pointer" );
      TERP_DEBUG_TRUE_OR_THROW( m_r1IntersectionSegmentsPointsSize > 1, "Invalid intersection points" );
      TERP_DEBUG_TRUE_OR_THROW( m_r2IntersectionSegmentsPointsSize > 1, "Invalid intersection points" );
      
      // Checking if it is inside the intersection
      
      m_sumMethodImp_auxPoint.setX( col );
      m_sumMethodImp_auxPoint.setY( line );
      
      m_sumMethodImp_PointInsideIntersection = false;
      for( m_sumMethodImp_IntersectionTileIndexersIdx = 0;
        m_sumMethodImp_IntersectionTileIndexersIdx < m_intersectionTileIndexers.size() ;
        ++m_sumMethodImp_IntersectionTileIndexersIdx )
      {
        if( m_intersectionTileIndexers[ 
          m_sumMethodImp_IntersectionTileIndexersIdx ].within( 
          m_sumMethodImp_auxPoint ) )
        {
          m_sumMethodImp_PointInsideIntersection = true;
          break;
        }
      }
      
      // Blending if the point is inside the intersection      
      
      if( m_sumMethodImp_PointInsideIntersection )
      {
        // Finding the point over the second raster
        
        m_geomTransformationPtr->directMap( col, line, m_sumMethodImp_Point2Col,
          m_sumMethodImp_Point2Line );      
        
        // Blending values

        for( m_sumMethodImp_BandIdx = 0 ; m_sumMethodImp_BandIdx <
          m_raster1Bands.size() ; ++m_sumMethodImp_BandIdx )
        {
          m_interp1Ptr->getValue( col, line, m_sumMethodImp_cValue1, 
            m_raster1Bands[ m_sumMethodImp_BandIdx ] ); 
          m_interp2Ptr->getValue( m_sumMethodImp_Point2Col, 
            m_sumMethodImp_Point2Line, m_sumMethodImp_cValue2, 
            m_raster2Bands[ m_sumMethodImp_BandIdx ] );
      
          if( m_sumMethodImp_cValue1.real() == m_raster1NoDataValues[ m_sumMethodImp_BandIdx ] )
          {
            if( m_sumMethodImp_cValue2.real() == m_raster2NoDataValues[ m_sumMethodImp_BandIdx ] )
            {
              values[ m_sumMethodImp_BandIdx ] = m_outputNoDataValue;
            }
            else
            {
              values[ m_sumMethodImp_BandIdx ] = 
                ( 
                  m_sumMethodImp_cValue2.real()
                  *
                  m_pixelScales2[ m_sumMethodImp_BandIdx ]
                )
                +
                m_pixelOffsets2[ m_sumMethodImp_BandIdx ]; 
            }
          }
          else
          {
            if( m_sumMethodImp_cValue2.real() == m_raster2NoDataValues[ m_sumMethodImp_BandIdx ] )
            {
              values[ m_sumMethodImp_BandIdx ] =  
                ( 
                  m_sumMethodImp_cValue1.real()
                  
                  *
                  m_pixelScales1[ m_sumMethodImp_BandIdx ]
                )
                +
                m_pixelOffsets1[ m_sumMethodImp_BandIdx ]; 
            }
            else
            {
              values[ m_sumMethodImp_BandIdx ] =
                (
                  (
                    m_sumMethodImp_cValue1.real()  
                    * 
                    m_pixelScales1[ m_sumMethodImp_BandIdx ] 
                  )
                  +
                  m_pixelOffsets1[ m_sumMethodImp_BandIdx ]
                )
                +
                (
                  (
                    m_sumMethodImp_cValue2.real()
                    * 
                    m_pixelScales2[ m_sumMethodImp_BandIdx ]
                  )
                  +
                  m_pixelOffsets2[ m_sumMethodImp_BandIdx ]
                );
            }          
          }      
        }
      }
      else
      {
        noBlendMethodImp( line, col, values );
      }
    }    
    
    void Blender::maxMethodImp( const double& line, const double& col,
      double* const values )
    {
      TERP_DEBUG_TRUE_OR_THROW( m_intersectionPtr.get(), "Invalid intersection pointer" );
      TERP_DEBUG_TRUE_OR_THROW( m_r1IntersectionSegmentsPointsSize > 1, "Invalid intersection points" );
      TERP_DEBUG_TRUE_OR_THROW( m_r2IntersectionSegmentsPointsSize > 1, "Invalid intersection points" );
      
      // Checking if it is inside the intersection
      
      m_maxMethodImp_auxPoint.setX( col );
      m_maxMethodImp_auxPoint.setY( line );
      
      m_maxMethodImp_PointInsideIntersection = false;
      for( m_maxMethodImp_IntersectionTileIndexersIdx = 0;
        m_maxMethodImp_IntersectionTileIndexersIdx < m_intersectionTileIndexers.size() ;
        ++m_maxMethodImp_IntersectionTileIndexersIdx )
      {
        if( m_intersectionTileIndexers[ 
          m_maxMethodImp_IntersectionTileIndexersIdx ].within( 
          m_maxMethodImp_auxPoint ) )
        {
          m_maxMethodImp_PointInsideIntersection = true;
          break;
        }
      }
      
      // Blending if the point is inside the intersection      
      
      if( m_maxMethodImp_PointInsideIntersection )
      {
        // Finding the point over the second raster
        
        m_geomTransformationPtr->directMap( col, line, m_maxMethodImp_Point2Col,
          m_maxMethodImp_Point2Line );      
        
        // Blending values

        for( m_maxMethodImp_BandIdx = 0 ; m_maxMethodImp_BandIdx <
          m_raster1Bands.size() ; ++m_maxMethodImp_BandIdx )
        {
          m_interp1Ptr->getValue( col, line, m_maxMethodImp_cValue1, 
            m_raster1Bands[ m_maxMethodImp_BandIdx ] ); 
          m_interp2Ptr->getValue( m_maxMethodImp_Point2Col, 
            m_maxMethodImp_Point2Line, m_maxMethodImp_cValue2, 
            m_raster2Bands[ m_maxMethodImp_BandIdx ] );
      
          if( m_maxMethodImp_cValue1.real() == m_raster1NoDataValues[ m_maxMethodImp_BandIdx ] )
          {
            if( m_maxMethodImp_cValue2.real() == m_raster2NoDataValues[ m_maxMethodImp_BandIdx ] )
            {
              values[ m_maxMethodImp_BandIdx ] = m_outputNoDataValue;
            }
            else
            {
              values[ m_maxMethodImp_BandIdx ] = 
                ( 
                  m_maxMethodImp_cValue2.real()
                  *
                  m_pixelScales2[ m_maxMethodImp_BandIdx ]
                )
                +
                m_pixelOffsets2[ m_maxMethodImp_BandIdx ]; 
            }
          }
          else
          {
            if( m_maxMethodImp_cValue2.real() == m_raster2NoDataValues[ m_maxMethodImp_BandIdx ] )
            {
              values[ m_maxMethodImp_BandIdx ] =  
                ( 
                  m_maxMethodImp_cValue1.real()
                  
                  *
                  m_pixelScales1[ m_maxMethodImp_BandIdx ]
                )
                +
                m_pixelOffsets1[ m_maxMethodImp_BandIdx ]; 
            }
            else
            {
              values[ m_maxMethodImp_BandIdx ] =
                std::max(
                  (
                    (
                      m_maxMethodImp_cValue1.real()  
                      * 
                      m_pixelScales1[ m_maxMethodImp_BandIdx ] 
                    )
                    +
                    m_pixelOffsets1[ m_maxMethodImp_BandIdx ]
                  )
                  ,
                  (
                    (
                      m_maxMethodImp_cValue2.real()
                      * 
                      m_pixelScales2[ m_maxMethodImp_BandIdx ]
                    )
                    +
                    m_pixelOffsets2[ m_maxMethodImp_BandIdx ]
                  )
                );
            }          
          }      
        }
      }
      else
      {
        noBlendMethodImp( line, col, values );
      }
    }    

    void Blender::minMethodImp( const double& line, const double& col,
      double* const values )
    {
      TERP_DEBUG_TRUE_OR_THROW( m_intersectionPtr.get(), "Invalid intersection pointer" );
      TERP_DEBUG_TRUE_OR_THROW( m_r1IntersectionSegmentsPointsSize > 1, "Invalid intersection points" );
      TERP_DEBUG_TRUE_OR_THROW( m_r2IntersectionSegmentsPointsSize > 1, "Invalid intersection points" );
      
      // Checking if it is inside the intersection
      
      m_minMethodImp_auxPoint.setX( col );
      m_minMethodImp_auxPoint.setY( line );
      
      m_minMethodImp_PointInsideIntersection = false;
      for( m_minMethodImp_IntersectionTileIndexersIdx = 0;
        m_minMethodImp_IntersectionTileIndexersIdx < m_intersectionTileIndexers.size() ;
        ++m_minMethodImp_IntersectionTileIndexersIdx )
      {
        if( m_intersectionTileIndexers[ 
          m_minMethodImp_IntersectionTileIndexersIdx ].within( 
          m_minMethodImp_auxPoint ) )
        {
          m_minMethodImp_PointInsideIntersection = true;
          break;
        }
      }
      
      // Blending if the point is inside the intersection      
      
      if( m_minMethodImp_PointInsideIntersection )
      {
        // Finding the point over the second raster
        
        m_geomTransformationPtr->directMap( col, line, m_minMethodImp_Point2Col,
          m_minMethodImp_Point2Line );      
        
        // Blending values

        for( m_minMethodImp_BandIdx = 0 ; m_minMethodImp_BandIdx <
          m_raster1Bands.size() ; ++m_minMethodImp_BandIdx )
        {
          m_interp1Ptr->getValue( col, line, m_minMethodImp_cValue1, 
            m_raster1Bands[ m_minMethodImp_BandIdx ] ); 
          m_interp2Ptr->getValue( m_minMethodImp_Point2Col, 
            m_minMethodImp_Point2Line, m_minMethodImp_cValue2, 
            m_raster2Bands[ m_minMethodImp_BandIdx ] );
      
          if( m_minMethodImp_cValue1.real() == m_raster1NoDataValues[ m_minMethodImp_BandIdx ] )
          {
            if( m_minMethodImp_cValue2.real() == m_raster2NoDataValues[ m_minMethodImp_BandIdx ] )
            {
              values[ m_minMethodImp_BandIdx ] = m_outputNoDataValue;
            }
            else
            {
              values[ m_minMethodImp_BandIdx ] = 
                ( 
                  m_minMethodImp_cValue2.real()
                  *
                  m_pixelScales2[ m_minMethodImp_BandIdx ]
                )
                +
                m_pixelOffsets2[ m_minMethodImp_BandIdx ]; 
            }
          }
          else
          {
            if( m_minMethodImp_cValue2.real() == m_raster2NoDataValues[ m_minMethodImp_BandIdx ] )
            {
              values[ m_minMethodImp_BandIdx ] =  
                ( 
                  m_minMethodImp_cValue1.real()
                  
                  *
                  m_pixelScales1[ m_minMethodImp_BandIdx ]
                )
                +
                m_pixelOffsets1[ m_minMethodImp_BandIdx ]; 
            }
            else
            {
              values[ m_minMethodImp_BandIdx ] =
                std::min(
                  (
                    (
                      m_minMethodImp_cValue1.real()  
                      * 
                      m_pixelScales1[ m_minMethodImp_BandIdx ] 
                    )
                    +
                    m_pixelOffsets1[ m_minMethodImp_BandIdx ]
                  )
                  ,
                  (
                    (
                      m_minMethodImp_cValue2.real()
                      * 
                      m_pixelScales2[ m_minMethodImp_BandIdx ]
                    )
                    +
                    m_pixelOffsets2[ m_minMethodImp_BandIdx ]
                  )
                );
            }          
          }      
        }
      }
      else
      {
        noBlendMethodImp( line, col, values );
      }
    }         
        
    void Blender::ruleMethodImp( const double& line, const double& col,
       double* const values )
    {
      // Finding the point over the second raster
      
      m_geomTransformationPtr->directMap( col, line, m_ruleMethodImp_Point2Col,
        m_ruleMethodImp_Point2Line );
      
      // Calling the rule to get the blended values
      
      TERP_DEBUG_TRUE_OR_THROW( m_rulePtr.get() != nullptr, "Invalid rule pointer" );
      m_rulePtr->getBlendedValues( line, col, m_ruleMethodImp_Point2Line,
        m_ruleMethodImp_Point2Col, values );
    }
        
    bool Blender::blendIntoRaster1()
    {
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( m_raster1Ptr->getAccessPolicy() & 
        te::common::WAccess, "Invalid output raster access policy" );      
      
      // Locating raster2 over the raster1
      
      unsigned int firstOutputRasterCol = 0;
      unsigned int lastOutputRasterRow = 0;
      unsigned int lastOutputRasterCol = 0;
      unsigned int firstOutputRasterRow = 0;
      
      {
        const double raster2LastRowIdx = 
          (double)( m_raster2Ptr->getNumberOfRows() - 1 );
        const double raster2LastColIdx =
          (double)( m_raster2Ptr->getNumberOfColumns() - 1 );
        double raster2LLColOverRaster1 = 0;
        double raster2LLRowOverRaster1 = 0;
        double raster2LRColOverRaster1 = 0;
        double raster2LRRowOverRaster1 = 0;
        double raster2URColOverRaster1 = 0;
        double raster2URRowOverRaster1 = 0;
        double raster2ULColOverRaster1 = 0;
        double raster2ULRowOverRaster1 = 0;          

        m_geomTransformationPtr->inverseMap( 
          0.0,
          raster2LastRowIdx,
          raster2LLColOverRaster1,
          raster2LLRowOverRaster1);
        m_geomTransformationPtr->inverseMap( 
          raster2LastColIdx,
          raster2LastRowIdx,
          raster2LRColOverRaster1,
          raster2LRRowOverRaster1);                
        m_geomTransformationPtr->inverseMap( 
          raster2LastColIdx,
          0.0,
          raster2URColOverRaster1,
          raster2URRowOverRaster1); 
        m_geomTransformationPtr->inverseMap( 
          0.0,
          0.0,
          raster2ULColOverRaster1,
          raster2ULRowOverRaster1);  
        
        firstOutputRasterCol = (unsigned int)
          std::max( 0l,
            std::min( (long int)( m_raster1Ptr->getNumberOfColumns() - 1 ),
              te::common::Round< double, long int >(
                std::min( raster2LLColOverRaster1, 
                  std::min( raster2LRColOverRaster1, 
                    std::min( raster2URColOverRaster1, 
                                raster2ULColOverRaster1
                    ) 
                  ) 
                )
              )
            )
          );
          
        lastOutputRasterRow = (unsigned int)
          std::max( 0l,
            std::min( (long int)( m_raster1Ptr->getNumberOfRows() - 1 ),
              te::common::Round< double, long int >(
                std::max( raster2LLRowOverRaster1, 
                  std::max( raster2LRRowOverRaster1, 
                    std::max( raster2URRowOverRaster1, 
                                raster2ULRowOverRaster1
                    ) 
                  ) 
                ) 
              )
            )
          );
          
        lastOutputRasterCol = (unsigned int)
          std::max( 0l,
            std::min( (long int)( m_raster1Ptr->getNumberOfColumns() - 1 ),
              te::common::Round< double, long int >(
                std::max( raster2LLColOverRaster1, 
                  std::max( raster2LRColOverRaster1, 
                    std::max( raster2URColOverRaster1, 
                                raster2ULColOverRaster1
                    ) 
                  ) 
                )
              )
            )
          );
          
        firstOutputRasterRow = (unsigned int)
          std::max( 0l,
            std::min( (long int)( m_raster1Ptr->getNumberOfRows() - 1 ),
              te::common::Round< double, long int >(
                std::min( raster2LLRowOverRaster1, 
                  std::min( raster2LRRowOverRaster1, 
                    std::min( raster2URRowOverRaster1, 
                                raster2ULRowOverRaster1
                    ) 
                  ) 
                ) 
              )
            )
          );
          
        assert( firstOutputRasterCol >= 0 );
        assert( firstOutputRasterCol <= 
          ( m_raster1Ptr->getNumberOfColumns() - 1 ) );    
        assert( lastOutputRasterRow >= 0 );
        assert( lastOutputRasterRow <= 
          ( m_raster1Ptr->getNumberOfRows() - 1 ) );                 
        assert( lastOutputRasterCol >= 0 );
        assert( lastOutputRasterCol <= 
          ( m_raster1Ptr->getNumberOfColumns() - 1 ) ); 
        assert( firstOutputRasterRow >= 0 );
        assert( firstOutputRasterRow <= 
          ( m_raster1Ptr->getNumberOfRows() - 1 ) );
      }  
      
      // Discovering the raster 1 blocks we need to process
      
      std::vector< RasterBlockInfo > raster1BlocksInfos;
      bool allRaster1BandsWithSameBlocking = true;
      
      {
        const te::rst::Band& firstBand = *( m_raster1Ptr->getBand(
           m_raster1Bands[ 0 ] ) );        
        
        for( unsigned int raster1BandsIdx = 0 ; raster1BandsIdx < 
          m_raster1Bands.size() ; ++raster1BandsIdx )
        {
          const te::rst::Band& band = *( m_raster1Ptr->getBand(
            m_raster1Bands[ raster1BandsIdx ] ) ); 
          
          if(
              ( band.getProperty()->m_blkh != firstBand.getProperty()->m_blkh )
              ||
              ( band.getProperty()->m_blkw != firstBand.getProperty()->m_blkw )
              ||
              ( band.getProperty()->m_nblocksx != firstBand.getProperty()->m_nblocksx )
              ||
              ( band.getProperty()->m_nblocksy != firstBand.getProperty()->m_nblocksy )
            )
          {
            allRaster1BandsWithSameBlocking = false;
            break;
          }
        }
          
        unsigned int firstBlockX = firstOutputRasterCol / 
          firstBand.getProperty()->m_blkw;
        unsigned int lastBlockX = lastOutputRasterCol / 
          firstBand.getProperty()->m_blkw;
        unsigned int firstBlockY = firstOutputRasterRow / 
          firstBand.getProperty()->m_blkh;
        unsigned int lastBlockY = lastOutputRasterRow / 
          firstBand.getProperty()->m_blkh;                        
          
        for( unsigned int blkY = firstBlockY ; blkY <= lastBlockY ; ++blkY )
        {
          for( unsigned int blkX = firstBlockX ; blkX <= lastBlockX ; ++blkX )
          {
            raster1BlocksInfos.push_back( RasterBlockInfo() );
            
            RasterBlockInfo& rBInfo = raster1BlocksInfos.back();
            
            rBInfo.m_wasProcessed = false;            
            rBInfo.m_blkX = blkX;
            rBInfo.m_blkY = blkY;
            rBInfo.m_blkTotalPixelsNumber = firstBand.getProperty()->m_blkh *
              firstBand.getProperty()->m_blkw;
            
            rBInfo.m_firstRasterRow2Process = blkY * firstBand.getProperty()->m_blkh;
            rBInfo.m_firstRasterRow2Process = std::max( firstOutputRasterRow,
              rBInfo.m_firstRasterRow2Process );
            rBInfo.m_firstRasterRow2Process = std::min( lastOutputRasterRow,
              rBInfo.m_firstRasterRow2Process );
            
            rBInfo.m_rasterRows2ProcessBound = ( blkY + 1 ) * firstBand.getProperty()->m_blkh;
            rBInfo.m_rasterRows2ProcessBound = std::max( firstOutputRasterRow,
              rBInfo.m_rasterRows2ProcessBound );
            rBInfo.m_rasterRows2ProcessBound = std::min( lastOutputRasterRow + 1,
              rBInfo.m_rasterRows2ProcessBound );              
            
            rBInfo.m_firstRasterCol2Process = blkX * firstBand.getProperty()->m_blkw;
            rBInfo.m_firstRasterCol2Process = std::max( firstOutputRasterCol,
              rBInfo.m_firstRasterCol2Process );
            rBInfo.m_firstRasterCol2Process = std::min( lastOutputRasterCol,
              rBInfo.m_firstRasterCol2Process );              
            
            rBInfo.m_rasterCols2ProcessBound = ( blkX + 1 ) * firstBand.getProperty()->m_blkw;          
            rBInfo.m_rasterCols2ProcessBound = std::max( firstOutputRasterCol,
              rBInfo.m_rasterCols2ProcessBound );
            rBInfo.m_rasterCols2ProcessBound = std::min( lastOutputRasterCol + 1,
              rBInfo.m_rasterCols2ProcessBound );                
          }
        }
      }          
      
      // blending
      
      bool returnValue = true;
      
      {
        // Guessing memory resources

        const double totalPhysMem = (double)te::common::GetTotalPhysicalMemory();
        const double usedVMem = (double)te::common::GetUsedVirtualMemory();
        const double totalVMem = ( (double)te::common::GetTotalVirtualMemory() );
        const double maxVMem2Use = 0.75 * std::min( totalPhysMem, ( totalVMem - usedVMem ) );
        
        
        // Creating thread exec params
        
        bool abortValue = false;
        std::mutex mutex;
        te::rst::RasterSynchronizer sync1( *m_raster1Ptr, te::common::RWAccess );
        te::rst::RasterSynchronizer sync2( *((te::rst::Raster*)m_raster2Ptr), te::common::RAccess );
        std::mutex blockProcessedSignalMutex;
        std::condition_variable blockProcessedSignal;
        unsigned int runningThreadsCounter = 0;
        unsigned int processedBlocksNmb = 0;
        
        BlendIntoRaster1ThreadParams auxThreadParams;
        auxThreadParams.m_returnValuePtr = &returnValue;
        auxThreadParams.m_abortValuePtr = &abortValue;
        auxThreadParams.m_sync1Ptr = &sync1;
        auxThreadParams.m_sync2Ptr = &sync2;
        auxThreadParams.m_raster1BlocksInfosPtr = &raster1BlocksInfos;
        auxThreadParams.m_mutexPtr = &mutex;
        auxThreadParams.m_blockProcessedSignalMutexPtr = &blockProcessedSignalMutex;
        auxThreadParams.m_blockProcessedSignalPtr = &blockProcessedSignal;
        auxThreadParams.m_runningThreadsCounterPtr = &runningThreadsCounter;
        auxThreadParams.m_raster1BandsPtr = &m_raster1Bands;
        auxThreadParams.m_raster2BandsPtr = &m_raster2Bands;
        auxThreadParams.m_blendMethod = m_blendMethod;
        auxThreadParams.m_interpMethod2 = m_interpMethod2;
        auxThreadParams.m_noDataValue = m_outputNoDataValue;
        auxThreadParams.m_forceRaster1NoDataValue = m_forceRaster1NoDataValue;
        auxThreadParams.m_forceRaster2NoDataValue = m_forceRaster2NoDataValue;
        auxThreadParams.m_pixelOffsets1Ptr = &m_pixelOffsets1;
        auxThreadParams.m_pixelScales1Ptr = &m_pixelScales1;
        auxThreadParams.m_pixelOffsets2Ptr = &m_pixelOffsets2;
        auxThreadParams.m_pixelScales2Ptr = &m_pixelScales2;
        auxThreadParams.m_r1ValidDataDelimiterPtr = m_r1ValidDataDelimiterPtr.get();
        auxThreadParams.m_r2ValidDataDelimiterPtr = m_r2ValidDataDelimiterPtr.get();
        auxThreadParams.m_geomTransformationPtr = m_geomTransformationPtr;
        auxThreadParams.m_maxRasterCachedBlocks = 
          m_enableRasterCache
          ?
            (unsigned int) std::max(
              1.0, 
              (
                maxVMem2Use /
                (
                  (double)
                  ( 
                    m_raster2Ptr->getBand( m_raster1Bands[ 0 ] )->getBlockSize() 
                    *
                    m_raster1Bands.size()
                    *
                    m_threadsNumber
                  )
                )
              )
            )
          : m_raster1Bands.size();
        
        auxThreadParams.m_rulePtr = m_rulePtr.get();
        auxThreadParams.m_processedBlocksNmbPtr = &processedBlocksNmb;

        // creating threads
        
        if( ( m_threadsNumber == 1 ) || (!allRaster1BandsWithSameBlocking) )
        {
          runningThreadsCounter = 1;
          
          auxThreadParams.m_useProgress = m_enableProgressInterface;
          
          blendIntoRaster1Thread( &auxThreadParams );
        }
        else
        {
          std::vector< std::unique_ptr< std::thread > > threads;
          runningThreadsCounter = m_threadsNumber;
          
          for( unsigned int threadIdx = 0 ; threadIdx < m_threadsNumber ;
            ++threadIdx )
          {
            auxThreadParams.m_useProgress = false;
            
            threads.push_back( std::unique_ptr< std::thread >( 
              new std::thread( blendIntoRaster1Thread, 
               &auxThreadParams ) ) );
          };    
          
          // progress stuff
          
          std::unique_ptr< te::common::TaskProgress > progressPtr;
          if( m_enableProgressInterface )
          {
            std::unique_lock<std::mutex> locker( mutex );
            
            progressPtr.reset( new te::common::TaskProgress );
            progressPtr->setTotalSteps(static_cast<int>(raster1BlocksInfos.size()));
            progressPtr->setMessage( "Blending" );
            
            locker.unlock();
          
            while( true )
            {
              locker.lock();
              
              if( abortValue )
              {
                locker.unlock();
                break;
              }
              
              if( runningThreadsCounter == 0 )
              {
                locker.unlock();
                break;
              }
              
              if( progressPtr->isActive() )
              {
                if( ((int)processedBlocksNmb) > progressPtr->getCurrentStep() )
                {
                  progressPtr->pulse();
                }                
                
                locker.unlock();
                
                std::unique_lock<std::mutex> blockProcessedSignalLocker( 
                  blockProcessedSignalMutex );
                blockProcessedSignal.wait_for( blockProcessedSignalLocker, 
                  std::chrono::seconds( 1 ) );
              }
              else
              {
                abortValue = true;
                
                locker.unlock();
                
                break;
              }
            }
          }
          
          // Joining threads

          for( unsigned int threadIdx = 0 ; threadIdx < m_threadsNumber ;
            ++threadIdx )
          {                    
            threads[ threadIdx ]->join();
          }
        }
      }
      
      return returnValue;
    }
    
    void Blender::blendIntoRaster1Thread( BlendIntoRaster1ThreadParams* paramsPtr )
    {
      // locker
      
      std::unique_lock<std::mutex> locker(*paramsPtr->m_mutexPtr);      
      
      // local parameters copy
      
      const std::vector< unsigned int > raster1Bands = 
        *paramsPtr->m_raster1BandsPtr;
      const std::vector< unsigned int > raster2Bands = 
        *paramsPtr->m_raster2BandsPtr;              
      
      te::rst::SynchronizedRaster raster1( raster1Bands.size(), // to avoid deadlocks
        *( paramsPtr->m_sync1Ptr ) );
      te::rst::SynchronizedRaster raster2( paramsPtr->m_maxRasterCachedBlocks,
        *( paramsPtr->m_sync2Ptr ) );
      
      const unsigned int raster1BlocksInfosSize = static_cast<unsigned int>(
        paramsPtr->m_raster1BlocksInfosPtr->size());
      
      const std::vector< double > pixelOffsets1 =*paramsPtr->m_pixelOffsets1Ptr;
      const std::vector< double > pixelScales1 =*paramsPtr->m_pixelScales1Ptr;
      const std::vector< double > pixelOffsets2 = *paramsPtr->m_pixelOffsets2Ptr;
      const std::vector< double > pixelScales2 = *paramsPtr->m_pixelScales2Ptr;
      
      std::unique_ptr< te::gm::MultiPolygon > r1ValidDataDelimiterPtr;
      if( paramsPtr->m_r1ValidDataDelimiterPtr )
      {
        r1ValidDataDelimiterPtr.reset( 
          (te::gm::MultiPolygon*)paramsPtr->m_r1ValidDataDelimiterPtr->clone() );
      }
      
      std::unique_ptr< te::gm::MultiPolygon > r2ValidDataDelimiterPtr;
      if( paramsPtr->m_r2ValidDataDelimiterPtr )
      {
        r2ValidDataDelimiterPtr.reset( 
          (te::gm::MultiPolygon*)paramsPtr->m_r2ValidDataDelimiterPtr->clone() );
      }
      
      std::unique_ptr< te::gm::GeometricTransformation >
        geomTransformationPtr;
      if( paramsPtr->m_geomTransformationPtr )
      {
        geomTransformationPtr.reset( 
          paramsPtr->m_geomTransformationPtr->clone() );      
      }
      
      std::unique_ptr< te::common::TaskProgress > progressPtr;      
      if( paramsPtr->m_useProgress )
      {
        progressPtr.reset( new te::common::TaskProgress );
        progressPtr->setTotalSteps( raster1BlocksInfosSize );
        progressPtr->setMessage( "Blending" );
      }       
      
      const double noDataValue = paramsPtr->m_noDataValue;
      
      std::unique_ptr< BlenderRule > rulePtr;
      if( paramsPtr->m_rulePtr )
      {
        rulePtr.reset( paramsPtr->m_rulePtr->clone() );
      }
      
      locker.unlock();
      
      // local blender
      
      Blender blender;      
      if( ! blender.initialize( 
        raster1,
        raster1Bands,
        raster2,
        raster2Bands,
        paramsPtr->m_blendMethod,
        te::rst::NearestNeighbor,
        paramsPtr->m_interpMethod2,
        noDataValue,
        paramsPtr->m_forceRaster1NoDataValue,
        paramsPtr->m_forceRaster2NoDataValue,
        pixelOffsets1,
        pixelScales1,
        pixelOffsets2,
        pixelScales2,
        r1ValidDataDelimiterPtr.get(),
        r2ValidDataDelimiterPtr.get(),
        *geomTransformationPtr,
        1,
        false,
        rulePtr.get(),
        false ) )
      {
        locker.lock();
        
        *(paramsPtr->m_abortValuePtr) = true;
        *(paramsPtr->m_returnValuePtr) = false;
        --( *(paramsPtr->m_runningThreadsCounterPtr) );
        
        return;
      }            
      
      // Guessing the output raster channels ranges
      
      const unsigned int raster1BandsSize = static_cast<unsigned int>(raster1Bands.size());
      
      std::vector< double > raster1BandsRangeMin( raster1BandsSize, 0 );
      std::vector< double > raster1BandsRangeMax( raster1BandsSize, 0 ); 
      
      {
        for( unsigned int raster1BandsIdx = 0 ;  raster1BandsIdx < 
          raster1BandsSize ; ++raster1BandsIdx )
        {
          unsigned int bandIdx = raster1Bands[ raster1BandsIdx ];
          
          te::rst::GetDataTypeRanges( raster1.getBand( bandIdx )->getProperty()->m_type,
            raster1BandsRangeMin[ raster1BandsIdx ],
            raster1BandsRangeMax[ raster1BandsIdx ]);             
        }
      }      
      
      // loocking for the next raster block to blend
      
      boost::scoped_array< double > blendedValuesHandler( new double[ raster1BandsSize ] );
      double* blendedValuesHandlerPtr = blendedValuesHandler.get();
      
      for( unsigned int raster1BlocksInfosIdx = 0 ; raster1BlocksInfosIdx <
        raster1BlocksInfosSize ; ++raster1BlocksInfosIdx )
      {
        locker.lock();
                
        if( paramsPtr->m_raster1BlocksInfosPtr->operator[]( raster1BlocksInfosIdx ).m_wasProcessed )
        {
          locker.unlock();
        }
        else
        {
          RasterBlockInfo& rBInfo = paramsPtr->m_raster1BlocksInfosPtr->operator[]( 
            raster1BlocksInfosIdx );
          rBInfo.m_wasProcessed = true;
          
          locker.unlock();
          
          //blending block data          
          
          unsigned int raster1Row = 0;
          unsigned int raster1Col = 0;
          unsigned int raster1BandsIdx = 0;
            
          for( raster1Row = rBInfo.m_firstRasterRow2Process ; raster1Row < rBInfo.m_rasterRows2ProcessBound ;
            ++raster1Row )
          {
            for( raster1Col = rBInfo.m_firstRasterCol2Process ; raster1Col < rBInfo.m_rasterCols2ProcessBound ;
              ++raster1Col )
            {
              blender.getBlendedValues( (double)raster1Row, (double)raster1Col,
                blendedValuesHandlerPtr );  
              
              for( raster1BandsIdx = 0 ; raster1BandsIdx < raster1BandsSize ; ++raster1BandsIdx )
              {          
                double& blendedValue = blendedValuesHandlerPtr[ raster1BandsIdx ];
                
                if( blendedValue != noDataValue )
                {
                  blendedValue = std::max( blendedValue , 
                    raster1BandsRangeMin[ raster1BandsIdx ] );
                  blendedValue = std::min( blendedValue , 
                    raster1BandsRangeMax[ raster1BandsIdx ] );

                  raster1.setValue( raster1Col, raster1Row, blendedValue,
                    raster1Bands[ raster1BandsIdx ] );
                }
              }
            }
          }
          
          locker.lock();
          
          ++( *(paramsPtr->m_processedBlocksNmbPtr) );
          
          locker.unlock();
          
          // progress stuff
          
          if( progressPtr.get() )
          {
            if( progressPtr->isActive() )
            {
              progressPtr->pulse();
            }
            else
            {
              locker.lock();
              
              *(paramsPtr->m_abortValuePtr) = true;
              *(paramsPtr->m_returnValuePtr) = false;
              --( *(paramsPtr->m_runningThreadsCounterPtr) );
              
              return;
            }
          }
          else
          {
            // notifying the main thread with the block processed signal
            
            paramsPtr->m_blockProcessedSignalPtr->notify_all();
          }
        }
        
        // do we must continue processing ?
        
        locker.lock();
        
        if( *(paramsPtr->m_abortValuePtr) )
        {
          locker.unlock();
          break;
        }
        else
        {
          locker.unlock();
        }
      }
      
      locker.lock();
      
      --( *(paramsPtr->m_runningThreadsCounterPtr) );
      
      locker.unlock();
    }
    
    void Blender::getSegments( te::gm::Geometry const * const geometryPtr,
      std::vector< std::pair< te::gm::Coord2D, te::gm::Coord2D > >& segments ) const
    {
      if( dynamic_cast< te::gm::LineString const * >( geometryPtr ) )
      {
        te::gm::LineString const* castGeomPtr = 
          dynamic_cast< te::gm::LineString const * >( geometryPtr );
        
        std::size_t nPoints = castGeomPtr->size();
        te::gm::Coord2D const* coodsPtr = castGeomPtr->getCoordinates();
        
        for( std::size_t pIdx = 1 ; pIdx < nPoints ; ++pIdx )
        {
          segments.push_back( std::pair< te::gm::Coord2D, te::gm::Coord2D >(
            coodsPtr[ pIdx - 1 ], coodsPtr[ pIdx ] ) );
        }
      } 
      else if( dynamic_cast< te::gm::CircularString const * >( geometryPtr ) )
      {
        te::gm::CircularString const* castGeomPtr =
          dynamic_cast< te::gm::CircularString const * >( geometryPtr );

        std::size_t nPoints = castGeomPtr->size();
        const std::vector< te::gm::Coord2D >& coods = castGeomPtr->getCoordinates();

        for( std::size_t pIdx = 1 ; pIdx < nPoints ; ++pIdx )
        {
          segments.push_back( std::pair< te::gm::Coord2D, te::gm::Coord2D >(
            coods[ pIdx - 1 ], coods[ pIdx ] ) );
        }
      }
      else if( dynamic_cast< te::gm::CurvePolygon const * >( geometryPtr ) )
      {
        te::gm::CurvePolygon const * castGeomPtr = 
          dynamic_cast< te::gm::CurvePolygon const * >( geometryPtr );
        
        std::size_t numGeoms = castGeomPtr->getNumRings();
        
        for( std::size_t gIdx = 0 ; gIdx < numGeoms ; ++gIdx )
        {
          getSegments( castGeomPtr->getRingN( gIdx ), segments );
        }
      }              
      else if( dynamic_cast< te::gm::GeometryCollection const * >( geometryPtr ) )
      {
        te::gm::GeometryCollection const * castGeomPtr = 
          dynamic_cast< te::gm::GeometryCollection const * >( geometryPtr );
        
        std::size_t numGeoms = castGeomPtr->getNumGeometries();
        
        for( std::size_t gIdx = 0 ; gIdx < numGeoms ; ++gIdx )
        {
          getSegments( castGeomPtr->getGeometryN( gIdx ), segments );
        }
      }
    }
    
    bool Blender::getTileIndexers( te::gm::Geometry const * const geometryPtr, 
      boost::ptr_vector< te::rst::TileIndexer >& tileIndexers ) const
    {
      if( dynamic_cast< te::gm::Polygon const * >( geometryPtr ) )
      {
        te::gm::Polygon const * castGeomPtr = 
          dynamic_cast< te::gm::Polygon const * >( geometryPtr );
          
        tileIndexers.push_back( new te::rst::TileIndexer( *castGeomPtr, 1.0 ) );
      }              
      else if( dynamic_cast< te::gm::GeometryCollection const * >( geometryPtr ) )
      {
        te::gm::GeometryCollection const * castGeomPtr = 
          dynamic_cast< te::gm::GeometryCollection const * >( geometryPtr );
        
        std::size_t numGeoms = castGeomPtr->getNumGeometries();
        
        for( std::size_t gIdx = 0 ; gIdx < numGeoms ; ++gIdx )
        {
          if( ! ( getTileIndexers( castGeomPtr->getGeometryN( gIdx ), tileIndexers ) ) )
          {
            return false;
          }
        }
      }
      
      return true;
    }    
    
    const std::string& Blender::getErrorMessage() const
    {
      return m_errorMessage;
    }
    
    void Blender::setErrorMessage( const std::string& newErrorMessage )
    {
      m_errorMessage = newErrorMessage;
    }
    
  } // end namespace rp
}   // end namespace te    

