/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/rp/ClassifierEMStrategy.cpp

  \brief EM (Expectation-Maximization) strategy for pixel-based classification.
*/

// TerraLib
#include "../common/MatrixUtils.h"
#include "../common/progress/TaskProgress.h"
#include "../raster/Grid.h"
#include "../raster/PositionIterator.h"
#include "../raster/RasterIterator.h"
#include "../raster/Utils.h"
#include "../memory/CachedRaster.h"
#include "ClassifierEMStrategy.h"
#include "Macros.h"
#include "Functions.h"
#include "AlgorithmParametersSerializer.h"

// STL
#include <complex>
#include <iostream>
#include <map>
#include <ctime>

// Boost
#include <boost/numeric/ublas/io.hpp>
#include <boost/numeric/ublas/lu.hpp>
#include <boost/numeric/ublas/matrix.hpp>

namespace
{
  static te::rp::ClassifierEMStrategyFactory classifierEMStrategyFactoryInstance;
}

te::rp::ClassifierEMStrategy::Parameters::Parameters()
{
  reset();
}

te::rp::ClassifierEMStrategy::Parameters::~Parameters() = default;

const te::rp::ClassifierEMStrategy::Parameters& te::rp::ClassifierEMStrategy::Parameters::operator=(const te::rp::ClassifierEMStrategy::Parameters& rhs)
{
  reset();

  m_useRandomSamples = rhs.m_useRandomSamples;
  m_numberOfClusters = rhs.m_numberOfClusters;
  m_maxIterations = rhs.m_maxIterations;
  m_maxInputPoints = rhs.m_maxInputPoints;
  m_epsilon = rhs.m_epsilon;
  m_clustersMeans = rhs.m_clustersMeans;

  return *this;
}

void te::rp::ClassifierEMStrategy::Parameters::reset() 
{
  m_useRandomSamples = true;
  m_numberOfClusters = 0;
  m_maxIterations = 0;
  m_maxInputPoints = 0;
  m_epsilon = 0.0;
  m_clustersMeans.clear();
}

te::common::AbstractParameters* te::rp::ClassifierEMStrategy::Parameters::clone() const
{
  return new te::rp::ClassifierEMStrategy::Parameters(*this);
}

bool te::rp::ClassifierEMStrategy::Parameters::serialize(
  AlgorithmParametersSerializer& serializer ) const
{
  TERP_TRUE_OR_RETURN_FALSE( ClassifierStrategyParameters::serialize( serializer ),
    "Serialization error" );

  TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "useRandomSamples",
    m_useRandomSamples ), "Error adding parameter useRandomSamples " );

  TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "numberOfClusters",
    m_numberOfClusters ), "Error adding parameter numberOfClusters" );

  TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "maxIterations",
    m_maxIterations ), "Error adding parameter maxIterations" );

  TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "maxInputPoints",
    m_maxInputPoints ), "Error adding parameter maxInputPoints" );

  TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "epsilon",
    m_epsilon ), "Error adding parameter epsilon" );

  return true;
}

te::rp::ClassifierEMStrategy::ClassifierEMStrategy()
{
  m_isInitialized = false;
}

te::rp::ClassifierEMStrategy::~ClassifierEMStrategy() = default;

bool te::rp::ClassifierEMStrategy::initialize(te::rp::ClassifierStrategyParameters const* const strategyParams) 
{
  m_isInitialized = false;

  te::rp::ClassifierEMStrategy::Parameters const* paramsPtr = dynamic_cast<te::rp::ClassifierEMStrategy::Parameters const*>(strategyParams);

  if(!paramsPtr)
    return false;

  m_parameters = *(paramsPtr);

  TERP_INSTANCE_TRUE_OR_RETURN_FALSE(m_parameters.m_numberOfClusters > 1, TE_TR("The number of clusters must be at least 2."))
  
  TERP_INSTANCE_TRUE_OR_RETURN_FALSE(m_parameters.m_maxIterations > 0, TE_TR("The number of iterations must be at least 1."))
  
  if (m_parameters.m_maxInputPoints == 0)
  {
    m_parameters.m_maxInputPoints = 1000;
  }
  
  TERP_INSTANCE_TRUE_OR_RETURN_FALSE(m_parameters.m_maxInputPoints >= m_parameters.m_numberOfClusters, TE_TR("The maximum number of points must be at least higher than the number of clusters."))
  
  TERP_INSTANCE_TRUE_OR_RETURN_FALSE(m_parameters.m_epsilon > 0, TE_TR("The stop criteria must be higher than 0."))

  m_isInitialized = true;

  return true;
}

bool te::rp::ClassifierEMStrategy::execute() 
{
  TERP_INSTANCE_TRUE_OR_RETURN_FALSE(m_isInitialized, TE_TR("Instance not initialized"))
  
  std::srand(std::time(nullptr));

  // input raster cache

  te::rst::Raster const * inputRasterPtr = m_inputRasterPtr;
  std::unique_ptr< te::mem::CachedRaster > cachedInputRasterPtr;

  if( m_enableRasterCache  )
  {
    assert( m_inputRasterPtr );
    cachedInputRasterPtr.reset( new te::mem::CachedRaster( *m_inputRasterPtr, 25, 0 ) );
    inputRasterPtr = cachedInputRasterPtr.get();
  }

  // Creating the output raster

  te::rst::Raster* outputRasterPtr = 0;
  std::unique_ptr< te::mem::CachedRaster > cachedOutputRasterPtr;

  {
    std::vector< int > dt;
    dt.push_back( te::dt::UINT32_TYPE );

    std::vector< double > noDataValues;
    noDataValues.push_back( (double)( m_parameters.m_numberOfClusters + 2 ) );

    TERP_INSTANCE_TRUE_OR_RETURN_FALSE( createOutputRaster( dt, noDataValues ),
      "Output raster creation error" );

    TERP_INSTANCE_TRUE_OR_RETURN_FALSE( setOutputRasterPalette( m_parameters.m_numberOfClusters + 1 ),
      "Output raster palette creation error" );

    outputRasterPtr = m_outputRasterPtr.get();
    if( m_enableRasterCache  )
    {
      assert( m_outputRasterPtr.get() );
      cachedOutputRasterPtr.reset( new te::mem::CachedRaster( *m_outputRasterPtr, 33, 0 ) );
      outputRasterPtr = cachedOutputRasterPtr.get();
    }
  }


// create a vector of points with random positions inside raster to obtain input data
  std::vector<te::gm::Point*> randomPoints = te::rst::GetRandomPointsInRaster(*inputRasterPtr, m_parameters.m_maxInputPoints);

  boost::numeric::ublas::matrix<double> Xk;
  double min_pixel_value = std::numeric_limits< double >::max();
  double max_pixel_value = -1.0 * min_pixel_value;
  std::vector< double > max_pixels_values( m_inputRasterBands.size(), max_pixel_value );
  std::vector< double > min_pixels_values( m_inputRasterBands.size(), min_pixel_value );
  
  if( m_parameters.m_useRandomSamples )
  {
    Xk = boost::numeric::ublas::matrix<double>( 
      m_parameters.m_maxInputPoints, m_inputRasterBands.size() );
    
    const unsigned int inputBandsSize = static_cast<unsigned int>(m_inputRasterBands.size()); 
    
    std::vector<te::gm::Point*> randomPoints = te::rst::GetRandomPointsInRaster(*inputRasterPtr, m_parameters.m_maxInputPoints);
    
    te::rst::PointSetIterator<double> pit = te::rst::PointSetIterator<double>::begin(inputRasterPtr, randomPoints);
    te::rst::PointSetIterator<double> pitend = te::rst::PointSetIterator<double>::end(inputRasterPtr, randomPoints);
    unsigned int k = 0;
    
    while (pit != pitend)
    {
      for (unsigned int inputBandsIdx = 0; inputBandsIdx < inputBandsSize; inputBandsIdx++)
      {
        inputRasterPtr->getValue(pit.getColumn(), pit.getRow(), 
          Xk(k, inputBandsIdx ), m_inputRasterBands[inputBandsIdx]);
        
        max_pixel_value = std::max( max_pixel_value, Xk(k, inputBandsIdx ) );
        min_pixel_value = std::min( min_pixel_value, Xk(k, inputBandsIdx ) );
        max_pixels_values[ inputBandsIdx ] = std::max( max_pixels_values[ inputBandsIdx ], 
          Xk(k, inputBandsIdx ) );
        min_pixels_values[ inputBandsIdx ] = std::min( min_pixels_values[ inputBandsIdx ], 
          Xk(k, inputBandsIdx ) );        
      }

      ++k;
      ++pit;
    }  
  }
  else
  {
    const unsigned int inputBandsSize = (unsigned int)m_inputRasterBands.size(); 
    const unsigned int nCols = inputRasterPtr->getNumberOfColumns();
    const unsigned int nRows = inputRasterPtr->getNumberOfRows();    
    const unsigned int maxNRowsNCols = std::max( nRows, nCols );
    const unsigned int samplesStep =  
      (unsigned int)(
        ((double)maxNRowsNCols) 
        /
        std::sqrt( (double)m_parameters.m_maxInputPoints ) 
      );
    const unsigned int totalSamplesNumber = 
      ( 
        (unsigned int)std::ceil( (double)nCols / (double)samplesStep ) 
      )
      * 
      ( 
        (unsigned int)std::ceil( (double)nRows / (double)samplesStep ) 
      );
    unsigned int col = 0;
    unsigned int inputBandsIdx = 0;
    unsigned int sampleIdx = 0;
    
    Xk = boost::numeric::ublas::matrix<double>( totalSamplesNumber, 
      inputBandsSize );
   
    for( unsigned int row = 0 ; row < nRows ; row += samplesStep )
    {
      for( col = 0 ; col < nCols ; col += samplesStep )
      {
        for ( inputBandsIdx = 0; inputBandsIdx < inputBandsSize; ++inputBandsIdx)
        {
          assert( sampleIdx < totalSamplesNumber );
          inputRasterPtr->getValue( col, row, Xk(sampleIdx, inputBandsIdx ), 
            m_inputRasterBands[inputBandsIdx]);          
          
          max_pixel_value = std::max( max_pixel_value, Xk(sampleIdx, inputBandsIdx ) );
          min_pixel_value = std::min( min_pixel_value, Xk(sampleIdx, inputBandsIdx ) );
          max_pixels_values[ inputBandsIdx ] = std::max( 
            max_pixels_values[ inputBandsIdx ], 
            Xk(sampleIdx, inputBandsIdx ) );
          min_pixels_values[ inputBandsIdx ] = std::min( 
            min_pixels_values[ inputBandsIdx ], 
            Xk(sampleIdx, inputBandsIdx ) );                  
        }
        
        ++sampleIdx;
      }
    }
    
    assert( sampleIdx == totalSamplesNumber );
  }
  
  // the parameter vector of means for each cluster
  
  boost::numeric::ublas::matrix<double> clustersMeans = 
    boost::numeric::ublas::matrix<double>( m_parameters.m_numberOfClusters, 
    m_inputRasterBands.size() );
  boost::numeric::ublas::matrix<double> prevClustersMeans = 
    boost::numeric::ublas::matrix<double>( m_parameters.m_numberOfClusters, 
    m_inputRasterBands.size() );
  
  {
    const unsigned int inputBandsSize = static_cast<unsigned int>(m_inputRasterBands.size()); 
    
    if (m_parameters.m_clustersMeans.size() > 0)
    {
      for (unsigned int clusterIdx = 0; clusterIdx < m_parameters.m_numberOfClusters; 
        clusterIdx++)
      {
        for (unsigned int inputBandsIdx = 0; inputBandsIdx < inputBandsSize; inputBandsIdx++)
        {
          clustersMeans( clusterIdx, inputBandsIdx ) = 
            m_parameters.m_clustersMeans[clusterIdx][inputBandsIdx];
        }
      }
    }
    else
    {
      // define vector of means randomly, in the interval [0, max_pixel_value].
      
      for (unsigned int clusterIdx = 0; clusterIdx < m_parameters.m_numberOfClusters; clusterIdx++)
      {
        for (unsigned int inputBandsIdx = 0; inputBandsIdx < inputBandsSize; inputBandsIdx++)
        {
           clustersMeans( clusterIdx, inputBandsIdx ) = 
             min_pixels_values[ inputBandsIdx ] 
             +
             (
                ( (double)clusterIdx )
                *
                ( max_pixels_values[ inputBandsIdx ] - min_pixels_values[ inputBandsIdx ] )
                /
                ( (double)(  m_parameters.m_numberOfClusters - 1 ) )
             );
        }
      }
    }
    
    prevClustersMeans = clustersMeans;
  }

  // the parameter vector of covariance matrices for each cluster

  std::vector<boost::numeric::ublas::matrix<double> > clustersCovMatrixes;
  
  {
    const unsigned int inputBandsSize = static_cast<unsigned int>(m_inputRasterBands.size()); 
    boost::numeric::ublas::matrix<double> tmp_sigma( inputBandsSize, inputBandsSize );
    
    for (unsigned int clusterIdx = 0; clusterIdx < m_parameters.m_numberOfClusters; clusterIdx++)
    {
      for (unsigned int inputBandsIdx = 0; inputBandsIdx < inputBandsSize; inputBandsIdx++)
      {
        for (unsigned int inputBandsIdx2 = 0; inputBandsIdx2 < inputBandsSize; inputBandsIdx2++)
        {
          tmp_sigma( inputBandsIdx, inputBandsIdx2 ) = 0.0;
        }
        
        tmp_sigma( inputBandsIdx, inputBandsIdx ) = 1.0;
      }
      
      clustersCovMatrixes.push_back(tmp_sigma);
    }
  }

  // variables used to estimate the cluster's probabilities

  boost::numeric::ublas::matrix<double> Pj = boost::numeric::ublas::matrix<double>( 
    m_parameters.m_numberOfClusters, 1);  
  
  {
    const std::size_t xkSize1 = Xk.size1();
    
    boost::numeric::ublas::matrix<double> PCj_Xk = 
      boost::numeric::ublas::matrix<double>( m_parameters.m_numberOfClusters, 
      xkSize1 );
    for (unsigned int clusterIdx = 0; clusterIdx < m_parameters.m_numberOfClusters; 
      clusterIdx++)
    {
      Pj( clusterIdx, 0) = 1 / (double) m_parameters.m_numberOfClusters;
      
      for (unsigned int randomPointsIdx = 0; randomPointsIdx < xkSize1; 
        randomPointsIdx++)
      {
        PCj_Xk( clusterIdx, randomPointsIdx ) = 0.0;
      }
    }
    
    double sum_PCj_Xk = 0;
    double numerator_PCj_Xk = 0;
    double denominator_PCj_Xk = 0;
    double detCovMtx = 0;
    double det2CovMtx = 0;
    double distance_MUj = 0;

    const unsigned int inputBandsSize = static_cast<unsigned int>(m_inputRasterBands.size()); 
    boost::numeric::ublas::matrix<double> Xk_minus_MUj( inputBandsSize, 1);
    boost::numeric::ublas::matrix<double> Xk_minus_MUj_T(1, inputBandsSize );
    boost::numeric::ublas::matrix<double> product_NETAj(1, 1);
    boost::numeric::ublas::matrix<double> product_NETAj2(1, 1);
    boost::numeric::ublas::matrix<double> inverseCovMtx( inputBandsSize, inputBandsSize );
    boost::numeric::ublas::matrix<double> inverse2CovMtx( inputBandsSize, inputBandsSize );
    boost::numeric::ublas::matrix<double> product_Xk_minusMUj( inputBandsSize, inputBandsSize );
    boost::numeric::ublas::matrix<double> sum_product_Xk_minusMUj( inputBandsSize, inputBandsSize );
    boost::numeric::ublas::matrix<double> newClustersMeans;

    std::unique_ptr< te::common::TaskProgress > task;
    if( m_progressInterfaceEnabled )
    {
      task.reset( new te::common::TaskProgress( 
        TE_TR("Estimating clusters"), 
        te::common::TaskProgress::UNDEFINED, m_parameters.m_maxIterations) );
    }    
    
    for (unsigned int iterationIdx = 0; iterationIdx < m_parameters.m_maxIterations; iterationIdx++)
    {
      // computing PCj_Xk
      
      for (unsigned int clusterIdx = 0; clusterIdx < m_parameters.m_numberOfClusters; clusterIdx++)
      {
        te::common::GetDeterminant(clustersCovMatrixes[clusterIdx], detCovMtx);
        
        if (detCovMtx == 0.0)
        {
          for(unsigned int randomPointsIdx = 0; randomPointsIdx < 
            xkSize1; randomPointsIdx++ )
          {
            PCj_Xk( clusterIdx, randomPointsIdx ) = 0.0;
          }
        } 
        else
        {
          detCovMtx = std::pow(detCovMtx, -0.5);
          
          TERP_INSTANCE_TRUE_OR_RETURN_FALSE( te::common::GetInverseMatrix(
            clustersCovMatrixes[clusterIdx], inverseCovMtx ),
            "Inverse covariance matrix error" );
          
          for(unsigned int randomPointsIdx = 0; randomPointsIdx < 
            xkSize1; randomPointsIdx++ )
          {
            // computing numerator of equation for PCj_Xk
            
            for (unsigned int inputBandsIdx = 0; inputBandsIdx < inputBandsSize; 
              inputBandsIdx++)
            {
              Xk_minus_MUj( inputBandsIdx, 0) = 
                Xk( randomPointsIdx, inputBandsIdx ) - 
                clustersMeans( clusterIdx, inputBandsIdx );
            }
            
            Xk_minus_MUj_T = boost::numeric::ublas::trans(Xk_minus_MUj);

            product_NETAj = prod(Xk_minus_MUj_T, inverseCovMtx);
            product_NETAj = prod(product_NETAj, Xk_minus_MUj);
            product_NETAj *= -0.5;

            numerator_PCj_Xk = detCovMtx * std::exp( product_NETAj(0, 0) ) 
              * Pj( clusterIdx, 0);

            // computing denominator of equation for PCj_Xk
            
            denominator_PCj_Xk = 0.0;
            
            for (unsigned int clusterIdx2 = 0; clusterIdx2 < m_parameters.m_numberOfClusters; 
              clusterIdx2++)
            {
              te::common::GetDeterminant(clustersCovMatrixes[clusterIdx2], det2CovMtx);
              
              if (det2CovMtx != 0.0)
              {
                det2CovMtx = pow(det2CovMtx, -0.5);
                
                TERP_INSTANCE_TRUE_OR_RETURN_FALSE( te::common::GetInverseMatrix(
                  clustersCovMatrixes[clusterIdx2], inverse2CovMtx ),
                  "Inverse covariance matrix error" );            

                for (unsigned int inputBandsIdx = 0; inputBandsIdx < inputBandsSize; inputBandsIdx++)
                {
                  Xk_minus_MUj( inputBandsIdx, 0) = Xk( randomPointsIdx, inputBandsIdx ) - 
                    clustersMeans( clusterIdx2, inputBandsIdx );
                }
                
                Xk_minus_MUj_T = boost::numeric::ublas::trans(Xk_minus_MUj);

                product_NETAj2 = prod(Xk_minus_MUj_T, inverse2CovMtx);
                product_NETAj2 = prod(product_NETAj2, Xk_minus_MUj);
                product_NETAj2 *= -0.5;

                denominator_PCj_Xk += det2CovMtx * exp( product_NETAj2(0, 0)) * 
                  Pj( clusterIdx2, 0);
              }
            }
            
            // computing PCj_Xk
            
            if (denominator_PCj_Xk == 0.0)
            {
              PCj_Xk( clusterIdx, randomPointsIdx ) = 0.0;
            }
            else
            {
              PCj_Xk( clusterIdx, randomPointsIdx ) = numerator_PCj_Xk / denominator_PCj_Xk;
            }
          }
        }
      }

      // computing t + 1 for sigma and means
      
      for (unsigned int clusterIdx = 0; clusterIdx < m_parameters.m_numberOfClusters; clusterIdx++)
      {
        sum_PCj_Xk = 0.0;
        
        for (unsigned int randomPointsIdx = 0; randomPointsIdx < xkSize1; 
          randomPointsIdx++)
        {
          sum_PCj_Xk += PCj_Xk( clusterIdx, randomPointsIdx );
        }        
        
        if( sum_PCj_Xk != 0.0 )
        {
          // next sigma calcule
          
          for (unsigned int inputBandsIdx = 0; inputBandsIdx < inputBandsSize; inputBandsIdx++)
          {
            for (unsigned int inputBandsIdx2 = 0; inputBandsIdx2 < inputBandsSize; inputBandsIdx2++)
            {
              sum_product_Xk_minusMUj( inputBandsIdx, inputBandsIdx2 ) = 0.0;
            }
          }
          
          for (unsigned int randomPointsIdx = 0; randomPointsIdx < xkSize1; 
            randomPointsIdx++)
          {
            for (unsigned int l = 0; l < inputBandsSize; l++)
            {
              Xk_minus_MUj(l, 0) = Xk( randomPointsIdx, l) - clustersMeans( clusterIdx, l);
            }
            
            Xk_minus_MUj_T = boost::numeric::ublas::trans(Xk_minus_MUj);

            product_Xk_minusMUj = prod(Xk_minus_MUj, Xk_minus_MUj_T);
            product_Xk_minusMUj *= PCj_Xk( clusterIdx, randomPointsIdx );

            sum_product_Xk_minusMUj += product_Xk_minusMUj;
          }
          
          clustersCovMatrixes[clusterIdx] = sum_product_Xk_minusMUj / sum_PCj_Xk;
          
          // next means calcule 
          
          for (unsigned int inputBandsIdx = 0; inputBandsIdx < inputBandsSize; inputBandsIdx++)
          {
            clustersMeans( clusterIdx, inputBandsIdx ) = 0.0;
          }
          
          for (unsigned int randomPointsIdx = 0; randomPointsIdx < xkSize1; 
            randomPointsIdx++)
          {
            for (unsigned int inputBandsIdx = 0; inputBandsIdx < inputBandsSize; 
               inputBandsIdx++)
            {
              clustersMeans( clusterIdx, inputBandsIdx ) += 
                PCj_Xk( clusterIdx, randomPointsIdx ) 
                * 
                Xk( randomPointsIdx, inputBandsIdx );
            }
          }
          
          for (unsigned int inputBandsIdx = 0; inputBandsIdx < inputBandsSize; inputBandsIdx++)
          {
            clustersMeans( clusterIdx, inputBandsIdx ) /= sum_PCj_Xk;
          }
          
          // computing Pj for t + 1
          
          Pj( clusterIdx, 0) = 0.0;
          
          for (unsigned int randomPointsIdx = 0; randomPointsIdx < xkSize1; randomPointsIdx++)
          {
            Pj( clusterIdx, 0) += PCj_Xk( clusterIdx, randomPointsIdx );
          }
          
          Pj( clusterIdx, 0) /= (double)xkSize1;
        }
      }

      // checking convergence
      
      distance_MUj = 0.0;
      double a_minus_b;
      
      for (unsigned int clusterIdx = 0; clusterIdx < m_parameters.m_numberOfClusters; clusterIdx++)
      {
        for (unsigned int inputBandsIdx = 0; inputBandsIdx < inputBandsSize; inputBandsIdx++)
        {
          a_minus_b = clustersMeans( clusterIdx, inputBandsIdx ) - prevClustersMeans( clusterIdx, inputBandsIdx );
          distance_MUj += a_minus_b * a_minus_b;
        }
      }
      
      distance_MUj = sqrt(distance_MUj);
      
      if (distance_MUj < m_parameters.m_epsilon)
      {
        break;
      }
       
      prevClustersMeans = clustersMeans;

      if( m_progressInterfaceEnabled )
      {
        TERP_INSTANCE_TRUE_OR_RETURN_FALSE( task->isActive(),
          "Canceled" );
        task->pulse();
      }
    }
  }
  
  // pre calc of statistical values
  
  
  std::vector<boost::numeric::ublas::matrix<double> > inverseClustersCovMatrixes;
  std::vector< double > clustersCovMatrixesDet;
  
  {
    double det = 0;
    boost::numeric::ublas::matrix<double> inverse;
    
    for( std::size_t clustersCovMatrixesIdx = 0 ; clustersCovMatrixesIdx <
      clustersCovMatrixes.size() ; ++clustersCovMatrixesIdx )
    {
      te::common::GetDeterminant(clustersCovMatrixes[clustersCovMatrixesIdx], det);
      clustersCovMatrixesDet.push_back( det );
      
      if( det == 0 )
      {
        inverseClustersCovMatrixes.push_back( 
          boost::numeric::ublas::identity_matrix< double >( m_inputRasterBands.size() ) );
      }
      else
      {
        TERP_INSTANCE_TRUE_OR_RETURN_FALSE( te::common::GetInverseMatrix(
          clustersCovMatrixes[clustersCovMatrixesIdx], inverse ),
          "Inverse covariance matrix error" );
        inverseClustersCovMatrixes.push_back( inverse );
      }
    }
  }

    // classifying image
  
  {
    const unsigned int inputBandsSize = static_cast<unsigned int>(m_inputRasterBands.size());     
    const double outNoDataValue = outputRasterPtr->getBand( 0 )->getProperty()->m_noDataValue;
    const double inNoDataValue = inputRasterPtr->getBand( m_inputRasterBands[ 0 ] )->getProperty()->m_noDataValue;
    te::rst::RasterIterator<double> it = te::rst::RasterIterator<double>::begin((te::rst::Raster*) inputRasterPtr, m_inputRasterBands);
    te::rst::RasterIterator<double> itend = te::rst::RasterIterator<double>::end((te::rst::Raster*) inputRasterPtr, m_inputRasterBands);
    boost::numeric::ublas::matrix<double> X = boost::numeric::ublas::matrix<double>(1, inputBandsSize );
    boost::numeric::ublas::matrix<double> PCj_X = boost::numeric::ublas::matrix<double>( m_parameters.m_numberOfClusters, 1);
    boost::numeric::ublas::matrix<double> product_NETAj(1, 1);
    boost::numeric::ublas::matrix<double> product_NETAj2(1, 1);    
    double max_PCj_X;
    unsigned int cluster;
    double numerator_PCj_X;
    double denominator_PCj_X;    
    boost::numeric::ublas::matrix<double> X_minus_MUj( inputBandsSize, 1);
    boost::numeric::ublas::matrix<double> X_minus_MUj_T(1, inputBandsSize );

    std::unique_ptr< te::common::TaskProgress > task;
    if( m_progressInterfaceEnabled )
    {
      task.reset( new te::common::TaskProgress( "Classifying", 
        te::common::TaskProgress::UNDEFINED, inputRasterPtr->getNumberOfColumns() 
        * inputRasterPtr->getNumberOfRows()) );
    }
    
    while (it != itend)
    {
      for (unsigned int inputBandsIdx = 0; inputBandsIdx < inputBandsSize; inputBandsIdx++)
      {
        X(0, inputBandsIdx ) = (*it)[ m_inputRasterBands[ inputBandsIdx ]];
      }

      if( X(0, 0) == inNoDataValue )
      {
        outputRasterPtr->setValue(it.getColumn(), it.getRow(), outNoDataValue, 0);
      }
      else
      {
        // computing PCj_X
        
        for (unsigned int clusterIdx = 0; clusterIdx < m_parameters.m_numberOfClusters; clusterIdx++)
        {
          // computing numerator of equation for PCj_X
          
          for (unsigned int inputBandsIdx = 0; inputBandsIdx < inputBandsSize; inputBandsIdx++)
          {
            X_minus_MUj( inputBandsIdx, 0) = X(0, inputBandsIdx ) - 
              clustersMeans( clusterIdx, inputBandsIdx );
          }
          
          X_minus_MUj_T = boost::numeric::ublas::trans(X_minus_MUj);        

          product_NETAj = prod(X_minus_MUj_T, inverseClustersCovMatrixes[ clusterIdx ] );
          product_NETAj = prod(product_NETAj, X_minus_MUj);
          product_NETAj *= -0.5;

          numerator_PCj_X = clustersCovMatrixesDet[ clusterIdx ] * 
            exp(product_NETAj(0, 0)) * Pj( clusterIdx, 0);

          // computing denominator of equation for PCj_X
          
          denominator_PCj_X = 0.0;
          
          for (unsigned int clusterIdx2 = 0; clusterIdx2 < m_parameters.m_numberOfClusters; 
            clusterIdx2++)
          {
            for (unsigned int inputBandsIdx = 0; inputBandsIdx < inputBandsSize; inputBandsIdx++)
            {
              X_minus_MUj( inputBandsIdx, 0) = X(0, inputBandsIdx ) - 
                clustersMeans( clusterIdx2, inputBandsIdx );
            }
            
            X_minus_MUj_T = boost::numeric::ublas::trans(X_minus_MUj);          

            product_NETAj2 = prod(X_minus_MUj_T, inverseClustersCovMatrixes[ clusterIdx2 ]);
            product_NETAj2 = prod(product_NETAj2, X_minus_MUj);
            product_NETAj2 *= -0.5;

            denominator_PCj_X += clustersCovMatrixesDet[ clusterIdx2 ] * 
              exp(product_NETAj2(0, 0)) * Pj( clusterIdx2, 0);
          }
          
          if (denominator_PCj_X == 0.0)
          {
            PCj_X( clusterIdx, 0) = 0;
          }
          else
          {
            PCj_X( clusterIdx, 0) = numerator_PCj_X / denominator_PCj_X;
          }        
        }

        max_PCj_X = 0.0;
        cluster = 0;
        for (unsigned int j = 0; j < m_parameters.m_numberOfClusters; j++)
        {
          if (PCj_X(j, 0) > max_PCj_X)
          {
            max_PCj_X = PCj_X(j, 0);
            cluster = j;
          }
        }

        // save cluster information in output raster
        
        if( max_PCj_X == 0)
        {
          outputRasterPtr->setValue(it.getColumn(), it.getRow(), outNoDataValue, 0);
        }
        else
        {
          outputRasterPtr->setValue(it.getColumn(), it.getRow(), cluster, 0);
        }
      }

      ++it;
      
      if( m_progressInterfaceEnabled )
      {
        TERP_INSTANCE_TRUE_OR_RETURN_FALSE( task->isActive(),
          "Canceled" );
        task->pulse();
      }
    }
  }

  return true;
}

te::rp::ClassifierEMStrategyFactory::ClassifierEMStrategyFactory()
  : te::rp::ClassifierStrategyFactory("em")
{
}

te::rp::ClassifierEMStrategyFactory::~ClassifierEMStrategyFactory() = default;

te::rp::ClassifierStrategy* te::rp::ClassifierEMStrategyFactory::build()
{
  return new te::rp::ClassifierEMStrategy();
}
