/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/rp/ClassifierISODataStrategy.cpp

  \brief ISOData strategy for image classification.
*/

// TerraLib

#include "ClassifierISODataStrategy.h"
#include "../memory/CachedRaster.h"
#include "Macros.h"
#include "ClassifierAdaptors.h"

#include <boost/lexical_cast.hpp>

namespace
{
  static te::rp::ClassifierISODataStrategyFactory classifierISODataStrategyFactoryInstance;
}

// --------------------------------------------------------------------------

te::rp::ClassifierISODataStrategy::Parameters::Parameters()
{
  reset();
}

te::rp::ClassifierISODataStrategy::Parameters::~Parameters() = default;

const te::rp::ClassifierISODataStrategy::Parameters& 
  te::rp::ClassifierISODataStrategy::Parameters::operator=(
  const te::rp::ClassifierISODataStrategy::Parameters& rhs)
{
  m_parameters = rhs.m_parameters;
  return *this;
}

void te::rp::ClassifierISODataStrategy::Parameters::reset() _NOEXCEPT_OP(false)
{
  m_parameters.reset();
}

te::common::AbstractParameters* te::rp::ClassifierISODataStrategy::Parameters::clone() const
{
  return new te::rp::ClassifierISODataStrategy::Parameters( *this );
}

// --------------------------------------------------------------------------

te::rp::ClassifierISODataStrategy::ClassifierISODataStrategy()
{
  m_isInitialized = false;
}

te::rp::ClassifierISODataStrategy::~ClassifierISODataStrategy() = default;

bool te::rp::ClassifierISODataStrategy::initialize(
  te::rp::ClassifierStrategyParameters const* const strategyParams) _NOEXCEPT_OP(false)
{
  m_isInitialized = false;

  te::rp::ClassifierISODataStrategy::Parameters const* paramsPtr = 
    dynamic_cast<te::rp::ClassifierISODataStrategy::Parameters const*>(strategyParams);
  TERP_INSTANCE_TRUE_OR_RETURN_FALSE( ( paramsPtr != 0 ), "Invalid parameters" );

  m_parameters = *(paramsPtr);

  TERP_INSTANCE_TRUE_OR_RETURN_FALSE( m_genericClassifierInstance.initialize(
    m_parameters.m_parameters ), "Initialization error" );

  m_isInitialized = true;

  return true;
}

bool te::rp::ClassifierISODataStrategy::execute() _NOEXCEPT_OP(false)
{
  TERP_INSTANCE_TRUE_OR_RETURN_FALSE(m_isInitialized, TE_TR("Instance not initialized"))

  // input raster cache

  te::rst::Raster const * inputRasterPtr = m_inputRasterPtr;
  std::unique_ptr< te::mem::CachedRaster > cachedInputRasterPtr;

  if( m_enableRasterCache  )
  {
    assert( m_inputRasterPtr );
    cachedInputRasterPtr.reset( new te::mem::CachedRaster( *m_inputRasterPtr, 25, 0 ) );
    inputRasterPtr = cachedInputRasterPtr.get();
  }

  // Creating the EXPANSIBLE output raster

  {
    std::vector< int > dt;
    dt.push_back( te::dt::UINT32_TYPE );

    std::vector< double > noDataValues;
    noDataValues.push_back( (double)( m_parameters.m_parameters.m_clustersNumber + 2 ) ) ;

    TERP_INSTANCE_TRUE_OR_RETURN_FALSE( createOutputRaster( dt, noDataValues ),
      "Output raster creation error" );
  }

  // classifying
  
  std::vector< te::cl::ISOData::ClusterData > clustersData;
  
  {
    te::rp::ca::RasterInputAdaptor< double > inputRasterAdp( *inputRasterPtr );
    te::rp::ca::RasterInputOutputAdaptor< unsigned int > outputRasterAdp( *m_outputRasterPtr );  
    
    std::vector< double > inputNoDataValues;
    for( unsigned int inputRasterBandsIdx = 0 ; inputRasterBandsIdx < m_inputRasterBands.size() ;
      ++inputRasterBandsIdx )
    {
      inputNoDataValues.push_back( inputRasterPtr->getBand(
        m_inputRasterBands[ inputRasterBandsIdx ] )->getProperty()->m_noDataValue );
    }
    
    TERP_INSTANCE_TRUE_OR_RETURN_FALSE( m_genericClassifierInstance.classify( 
      inputRasterAdp,
      m_inputRasterBands, 
      inputNoDataValues,
      outputRasterAdp,
      0,
      m_outputRasterPtr->getBand( 0 )->getProperty()->m_noDataValue,
      m_progressInterfaceEnabled,
      &clustersData ), "Classifier execution error" );
  }
  
  // Updating the pallete to match the dynamic number of produced clusters
  
  TERP_INSTANCE_TRUE_OR_RETURN_FALSE( setOutputRasterPalette( 
    clustersData.size() ),
    "Output raster palette creation error" );
  
  // Create output metadata
  
  {
    std::vector< te::cl::ISOData::ClusterData >::const_iterator clustersDataIt =
      clustersData.begin();
    const std::vector< te::cl::ISOData::ClusterData >::const_iterator clustersDataItBegin =
      clustersData.begin();      
    const std::vector< te::cl::ISOData::ClusterData >::const_iterator clustersDataItEnd =
      clustersData.end();
    std::string metadataValueString;
      
    // CLUSTERS_LABELS
    
    while( clustersDataIt != clustersDataItEnd )
    {
      if( clustersDataIt != clustersDataItBegin )
        metadataValueString.append( ";" );
      metadataValueString.append( boost::lexical_cast< std::string >( 
        clustersDataIt->m_label ) );
      ++clustersDataIt;
    }
    m_metaData[ "CLUSTERS_LABELS" ] = metadataValueString;
    
    // CLUSTERS_SIZES
    metadataValueString.clear();
    clustersDataIt = clustersData.begin();
    while( clustersDataIt != clustersDataItEnd )
    {
      if( clustersDataIt != clustersDataItBegin )
        metadataValueString.append( ";" );
      metadataValueString.append( boost::lexical_cast< std::string >( 
        clustersDataIt->m_size ) );
      ++clustersDataIt;
    }
    m_metaData[ "CLUSTERS_SIZES" ] = metadataValueString;
    
    // CLUSTERS_MEANS
    metadataValueString.clear();
    clustersDataIt = clustersData.begin();
    while( clustersDataIt != clustersDataItEnd )
    {
      if( clustersDataIt != clustersDataItBegin )
        metadataValueString.append( ";" );
      for( std::size_t meansIdx = 0 ; meansIdx < clustersDataIt->m_means.size() ;
        ++meansIdx )
      {
        if( meansIdx )
          metadataValueString.append( ";" );
        metadataValueString.append( boost::lexical_cast< std::string >( 
          clustersDataIt->m_means[ meansIdx ] ) );
      }
      ++clustersDataIt;
    }
    m_metaData[ "CLUSTERS_MEANS" ] = metadataValueString;    
  }

  return true;
}

// --------------------------------------------------------------------------

te::rp::ClassifierISODataStrategyFactory::ClassifierISODataStrategyFactory()
  : te::rp::ClassifierStrategyFactory("isodata")
{
}

te::rp::ClassifierISODataStrategyFactory::~ClassifierISODataStrategyFactory() =
    default;

te::rp::ClassifierStrategy* te::rp::ClassifierISODataStrategyFactory::build()
{
  return new te::rp::ClassifierISODataStrategy();
}
