/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/rp/ClassifierISODataStrategy.h

  \brief ISOData strategy for image classification.
*/

#ifndef __TERRALIB_RP_INTERNAL_CLASSIFIERISODATASTRATEGY_H
#define __TERRALIB_RP_INTERNAL_CLASSIFIERISODATASTRATEGY_H

#include "Config.h"
#include "ClassifierStrategy.h"
#include "ClassifierStrategyFactory.h"

#include "../classification/ISOData.h"

namespace te
{
  namespace rp
  {
    /*!
      \class ClassifierISODataStrategy

      \note Based on Reference: IsoData Algorithm - Pattern Recognition Principles - Rafael Gonzalez, Julius T. Tou - Addison Wesley Publishing Company - 1974.
      \note Avaliable metadata: CLUSTERS_LABELS a semicolon-separated list of clusters lables.
      \note Avaliable metadata: CLUSTERS_SIZES a semicolon-separated list of clusters sizes (following the same order as the CLUSTERS_LABELS metadata).
      \note Avaliable metadata: CLUSTERS_MEANS a semicolon-separated list of clusters means (following the same order as the CLUSTERS_LABELS metadata).
             
       \ingroup rp_class
    */
    class TERPEXPORT ClassifierISODataStrategy : public ClassifierStrategy
    {
      public:

        /*!
          \class Parameters

          \brief Classifier Parameters
        */
        class TERPEXPORT Parameters : public ClassifierStrategyParameters
        {
          public:
            
            /*!
              \brief Generic Classifier Parameters.
            */
            te::cl::ISOData::Parameters m_parameters;

            Parameters();

            ~Parameters();

            //overload
            const Parameters& operator=(const Parameters& params);

            //overload
            void reset() _NOEXCEPT_OP(false);

            //overload
            AbstractParameters* clone() const;
        };

      public:

        ClassifierISODataStrategy();

        ~ClassifierISODataStrategy();

        //overload
        bool initialize(ClassifierStrategyParameters const* const strategyParams) _NOEXCEPT_OP(false);

        //overload
        bool execute() _NOEXCEPT_OP(false);

      protected:

        bool m_isInitialized;                                        //!< True if this instance is initialized.
        ClassifierISODataStrategy::Parameters m_parameters;           //!< Internal execution parameters.
        te::cl::ISOData m_genericClassifierInstance;  //!< Internal generic classifier instance.

    };

    /*!
      \class ClassifierISODataStrategyFactory

      \brief Raster ISOData Classifier strategy factory.

      \note Factory key: RegionGrowing
    */
    class TERPEXPORT ClassifierISODataStrategyFactory : public ClassifierStrategyFactory
    {
      public:

        ClassifierISODataStrategyFactory();

        ~ClassifierISODataStrategyFactory();

        //overload
        te::rp::ClassifierStrategy* build();
    };

  } // end namespace rp
}   // end namespace te

#endif // __TERRALIB_RP_INTERNAL_CLASSIFIERISODATASTRATEGY_H

