/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/rp/ClassifierMAPStrategy.cpp

  \brief Maximum a posteriori probability strategy.
*/

// TerraLib

#include "ClassifierMAPStrategy.h"
#include "Macros.h"
#include "../raster/Band.h"
#include "../memory/CachedRaster.h"
// #include "../common/MatrixUtils.h"
#include "ClassifierAdaptors.h"
#include "AlgorithmParametersSerializer.h"

#include <cfloat>

namespace
{
  static te::rp::ClassifierMAPStrategyFactory classifierMAPStrategyFactoryInstance;
}

//-----------------------------------------------------------------------------

te::rp::ClassifierMAPStrategy::Parameters::Parameters()
{
  reset();
}

te::rp::ClassifierMAPStrategy::Parameters::~Parameters() = default;

const te::rp::ClassifierMAPStrategy::Parameters& 
  te::rp::ClassifierMAPStrategy::Parameters::operator=(
  const te::rp::ClassifierMAPStrategy::Parameters& rhs)
{
  reset();

  m_trainSamplesPtr = rhs.m_trainSamplesPtr;
  m_prioriProbs = rhs.m_prioriProbs;
  m_prioriCalcSampleStep = rhs.m_prioriCalcSampleStep;
  m_MAPModelParams = rhs.m_MAPModelParams;

  return *this;
}

void te::rp::ClassifierMAPStrategy::Parameters::reset() 
{
  m_trainSamplesPtr.reset();
  m_prioriProbs.clear();
  m_prioriCalcSampleStep = 5;
  m_MAPModelParams.reset();
}

te::common::AbstractParameters* te::rp::ClassifierMAPStrategy::Parameters::clone() const
{
  return new te::rp::ClassifierMAPStrategy::Parameters( *this );
}

bool te::rp::ClassifierMAPStrategy::Parameters::serialize(
  AlgorithmParametersSerializer& serializer ) const
{
  TERP_TRUE_OR_RETURN_FALSE( ClassifierStrategyParameters::serialize( serializer ),
    "Serialization error" );

  TERP_TRUE_OR_RETURN_FALSE( serializer.addMultivaluedParameter( "prioriProbs",
    m_prioriProbs ), "Error adding parameter prioriProbs" );

  TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "prioriCalcSampleStep",
    m_prioriCalcSampleStep ), "Error adding parameter prioriCalcSampleStep" );
  
  TERP_TRUE_OR_RETURN_FALSE( serializer.addMultivaluedParameter( "classLabels",
    m_MAPModelParams.m_classLabels ), "Error adding parameter classLabels" );
  
  TERP_TRUE_OR_RETURN_FALSE( serializer.addMultivaluedParameters( 
    "classesMeans", m_MAPModelParams.m_classesMeans ), 
    "Error adding parameter classesMeans" );  

  TERP_TRUE_OR_RETURN_FALSE( serializer.addMultivaluedParameters( 
    "classesCovarianceMatrixes", m_MAPModelParams.m_classesCovarianceMatrixes ), 
    "Error adding parameter classesCovarianceMatrixes" );  

  return true;
}

//-----------------------------------------------------------------------------

te::rp::ClassifierMAPStrategy::OutputParameters::OutputParameters()
{
  reset();
}

te::rp::ClassifierMAPStrategy::OutputParameters::~OutputParameters() = default;

const te::rp::ClassifierMAPStrategy::OutputParameters& 
  te::rp::ClassifierMAPStrategy::OutputParameters::operator=(
  const te::rp::ClassifierMAPStrategy::OutputParameters& rhs)
{
  reset();

  m_MAPModelParams = rhs.m_MAPModelParams;

  return *this;
}

void te::rp::ClassifierMAPStrategy::OutputParameters::reset() 
{
  m_MAPModelParams.reset();
}

te::common::AbstractParameters* te::rp::ClassifierMAPStrategy::OutputParameters::clone() const
{
  return new OutputParameters( *this );
}

bool te::rp::ClassifierMAPStrategy::OutputParameters::serialize(
  AlgorithmParametersSerializer& serializer ) const
{
  TERP_TRUE_OR_RETURN_FALSE( ClassifierStrategyOutParameters::serialize( serializer ),
    "Serialization error" );

  TERP_TRUE_OR_RETURN_FALSE( serializer.addMultivaluedParameter( "classLabels",
    m_MAPModelParams.m_classLabels ), "Error adding parameter classLabels" );
  
  TERP_TRUE_OR_RETURN_FALSE( serializer.addMultivaluedParameters( 
    "classesMeans", m_MAPModelParams.m_classesMeans ), 
    "Error adding parameter classesMeans" );  

  TERP_TRUE_OR_RETURN_FALSE( serializer.addMultivaluedParameters( 
    "classesCovarianceMatrixes", m_MAPModelParams.m_classesCovarianceMatrixes ), 
    "Error adding parameter classesCovarianceMatrixes" );
  
  return true;
}

//-----------------------------------------------------------------------------

te::rp::ClassifierMAPStrategy::ClassifierMAPStrategy()
{
  m_isInitialized = false;
  m_highestTrainClassID = 0;
}

te::rp::ClassifierMAPStrategy::~ClassifierMAPStrategy() = default;

bool te::rp::ClassifierMAPStrategy::initialize(
  te::rp::ClassifierStrategyParameters const* const strategyParams) 
{
  m_isInitialized = false;
  m_initParams.reset();
  m_outParams.reset();
  m_highestTrainClassID = 0;
  
  ClassifierMAPStrategy::Parameters const * const castParamsPtr = 
    dynamic_cast< ClassifierMAPStrategy::Parameters const * >( strategyParams );
  TERP_INSTANCE_TRUE_OR_RETURN_FALSE( castParamsPtr, "Invalid parameters" );
  
  m_initParams = (*castParamsPtr);

  // Groupping samples

  std::vector< std::vector< double > > samples;
  std::vector<unsigned int> samplesLabels;
  std::vector<unsigned int> samplesAttributesIndices;

  if( castParamsPtr->m_trainSamplesPtr != nullptr )
  {
    TERP_INSTANCE_TRUE_OR_RETURN_FALSE( castParamsPtr->m_trainSamplesPtr->size() > 0,
      "Invalid classes samples number" )
    TERP_INSTANCE_TRUE_OR_RETURN_FALSE( castParamsPtr->m_trainSamplesPtr->begin()->second.size() > 0,
      "Invalid classes samples number" )        
    
    Parameters::MClassesSamplesCT::const_iterator classesSamplesIt = castParamsPtr->m_trainSamplesPtr->begin();
    const Parameters::MClassesSamplesCT::const_iterator classesSamplesItEnd = castParamsPtr->m_trainSamplesPtr->end();
    Parameters::ClassSamplesContainerT::const_iterator classSamplesIt;
    Parameters::ClassSamplesContainerT::const_iterator classSamplesItEnd;

    while( classesSamplesIt != classesSamplesItEnd )
    {
      classSamplesIt = classesSamplesIt->second.begin();
      classSamplesItEnd = classesSamplesIt->second.end();

      m_highestTrainClassID = std::max( m_highestTrainClassID,
        classesSamplesIt->first );

      while( classSamplesIt != classSamplesItEnd )
      {
        samples.push_back( *classSamplesIt);
        samplesLabels.push_back( classesSamplesIt->first );

        ++classSamplesIt;
      }

      ++classesSamplesIt;
    }

    if( !samples.empty() )
    {
      const std::size_t featuresNmb = samples[ 0 ].size();

      for( std::size_t featureIdx = 0 ; featureIdx < featuresNmb ; ++featureIdx )
      {
        samplesAttributesIndices.push_back(static_cast<unsigned int>(featureIdx));
      }
    }
  }

  if( ! m_initParams.m_MAPModelParams.m_classLabels.empty() )
  {
    const std::size_t classLabelsSize = m_initParams.m_MAPModelParams.m_classLabels.size();
    
    for( std::size_t classLabelsIdx = 0 ; classLabelsIdx < classLabelsSize ;
      ++classLabelsIdx )
    {
      m_highestTrainClassID = std::max( m_highestTrainClassID,
       m_initParams.m_MAPModelParams.m_classLabels[ classLabelsIdx ] );
    }
  }  
  
  // Initializing the generic classifier

  te::cl::MAP::Parameters genPars;
  genPars.m_prioriProbs = m_initParams.m_prioriProbs;
  genPars.m_prioriCalcSampleStep = m_initParams.m_prioriCalcSampleStep;
  
  if( m_initParams.m_MAPModelParams.m_classLabels.empty() )
  {
    TERP_INSTANCE_TRUE_OR_RETURN_FALSE( m_genClassInstance.initialize( genPars ),
      "Invalid parameters" )
  }
  else
  {
    TERP_INSTANCE_TRUE_OR_RETURN_FALSE( m_genClassInstance.initialize( genPars,
      m_initParams.m_MAPModelParams ), "Invalid parameters" )    
  }
  
  // Training the generic classifier
  
  if( !samples.empty() )
  {
    te::cl::VectorOfVectorsInputAdaptor< double > samplesAdaptor(
      samples );

    TERP_INSTANCE_TRUE_OR_RETURN_FALSE( m_genClassInstance.train( samplesAdaptor, samplesAttributesIndices,
      samplesLabels, false ), "Classifier trainning error" );
  }
  
  m_outParams.m_MAPModelParams = m_genClassInstance.getModelParams();

  // Finalizing
  
  m_isInitialized = true;
  
  return true;
}

bool te::rp::ClassifierMAPStrategy::execute() 
{
  TERP_INSTANCE_TRUE_OR_RETURN_FALSE( m_isInitialized,
    "Classification strategy not initialized" );

  // input raster cache

  te::rst::Raster const * inputRasterPtr = m_inputRasterPtr;
  std::unique_ptr< te::mem::CachedRaster > cachedInputRasterPtr;

  if( m_enableRasterCache  )
  {
    assert( m_inputRasterPtr );
    cachedInputRasterPtr.reset( new te::mem::CachedRaster( *m_inputRasterPtr, 25, 0 ) );
    inputRasterPtr = cachedInputRasterPtr.get();
  }

  // Creating the output raster

  te::rst::Raster* outputRasterPtr = 0;
  std::unique_ptr< te::mem::CachedRaster > cachedOutputRasterPtr;

  {
    std::vector< int > dt;
    dt.push_back( te::dt::UINT32_TYPE );

    std::vector< double > noDataValues;
    noDataValues.push_back( (double)( m_highestTrainClassID + 2 ) );

    TERP_INSTANCE_TRUE_OR_RETURN_FALSE( createOutputRaster( dt, noDataValues ),
      "Output raster creation error" );

    TERP_INSTANCE_TRUE_OR_RETURN_FALSE( setOutputRasterPalette( m_highestTrainClassID + 1 ),
      "Output raster palette creation error" );

    outputRasterPtr = m_outputRasterPtr.get();
    if( m_enableRasterCache  )
    {
      assert( m_outputRasterPtr.get() );
      cachedOutputRasterPtr.reset( new te::mem::CachedRaster( *m_outputRasterPtr, 33, 0 ) );
      outputRasterPtr = cachedOutputRasterPtr.get();
    }
  }

  te::rp::ca::RasterInputAdaptor< double > inputRasterAdp( *inputRasterPtr );
  te::rp::ca::RasterOutputAdaptor< unsigned int > outputRasterAdp( *outputRasterPtr );

  std::vector< double > inputNoDataValues;
  for( unsigned int inputRasterBandsIdx = 0 ; inputRasterBandsIdx < m_inputRasterBands.size() ;
    ++inputRasterBandsIdx )
  {
    inputNoDataValues.push_back( inputRasterPtr->getBand(
      m_inputRasterBands[ inputRasterBandsIdx ] )->getProperty()->m_noDataValue );
  }

  TERP_INSTANCE_TRUE_OR_RETURN_FALSE( m_genClassInstance.classify( inputRasterAdp,
    m_inputRasterBands, inputNoDataValues, outputRasterAdp, 0,
    outputRasterPtr->getBand( 0 )->getProperty()->m_noDataValue,
    m_progressInterfaceEnabled ), "Classifier execution error" );

  return true;
}

te::rp::ClassifierStrategyOutParameters const* te::rp::ClassifierMAPStrategy::getOutputParameters() const
{
  return &m_outParams;
}

//-----------------------------------------------------------------------------

te::rp::ClassifierMAPStrategyFactory::ClassifierMAPStrategyFactory()
  : te::rp::ClassifierStrategyFactory("map")
{
}

te::rp::ClassifierMAPStrategyFactory::~ClassifierMAPStrategyFactory() = default;

te::rp::ClassifierStrategy* te::rp::ClassifierMAPStrategyFactory::build()
{
  return new te::rp::ClassifierMAPStrategy();
}
