/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/rp/ClassifierMAPStrategy.h
  \brief Maximum a posteriori probability strategy.
 */

#ifndef __TERRALIB_RP_INTERNAL_CLASSIFIERMAPSTRATEGY_H
#define __TERRALIB_RP_INTERNAL_CLASSIFIERMAPSTRATEGY_H

#include "ClassifierStrategy.h"
#include "ClassifierStrategyOutParameters.h"
#include "ClassifierStrategyFactory.h"
#include "Config.h"
#include "../classification/MAP.h"
#include "../classification/Adaptors.h"

#include <boost/shared_ptr.hpp>

#include <map>
#include <vector>

namespace te
{
  namespace rp
  {
    class AlgorithmParametersSerializer;

    /*!
      \class ClassifierMAPStrategy

      \brief Maximum a posteriori probability strategy.
      
      \ingroup rp_class
     */
    class TERPEXPORT ClassifierMAPStrategy : public ClassifierStrategy
    {
      public:

        /*!
          \class Parameters

          \brief Classifier Parameters
        */
        class TERPEXPORT Parameters : public ClassifierStrategyParameters
        {
          public:
            
            typedef unsigned int ClassIDT; //!< Class ID type definition (zero means invalid ID).
            
            typedef std::vector< double > ClassSampleT; //!< Class sample type definition.
            
            typedef std::vector< ClassSampleT > ClassSamplesContainerT; //!< Class samples container type definition.
            
            typedef std::map< ClassIDT, ClassSamplesContainerT > MClassesSamplesCT; //!< Multi-classes samples container type definition.

            typedef boost::shared_ptr<MClassesSamplesCT> MClassesSamplesCTPtr; //!< A shared pointer to a multi classes samples container type definition.
            
            MClassesSamplesCTPtr m_trainSamplesPtr; //!< A shared pointer to a always-valid structure where trainning samples are stored.

            std::vector< double > m_prioriProbs; //!< Priori probabilities, one for each class. Values from 0 to 1 (use an empty vector to allow internal calcule of priori probabilities).            
            
            unsigned int m_prioriCalcSampleStep; //!< A positive non-zero sample step used when calculating piori probabilities (default:5 - 1/5 of samples will be used);
            
            te::cl::MAP::ModelParameters m_MAPModelParams; //!< Optional model parameters (further training not required).

            Parameters();

            ~Parameters();

            //overload
            const Parameters& operator=(const Parameters& params);

            //overload
            void reset() ;

            //overload
            AbstractParameters* clone() const;

            //overload
            bool serialize ( AlgorithmParametersSerializer& serializer ) const;
        };
        
        /*!
          \class Output Parameters

          \brief MAP Classifier output parameters
        */
        class TERPEXPORT OutputParameters : public ClassifierStrategyOutParameters
        {
          public:
            
            te::cl::MAP::ModelParameters m_MAPModelParams; //!< The calculated model parameters.

            OutputParameters();

            ~OutputParameters();

            //overload
            const OutputParameters& operator=(const OutputParameters& params);

            //overload
            void reset() ;

            //overload
            AbstractParameters* clone() const;

            //overload
            bool serialize ( AlgorithmParametersSerializer& serializer ) const;
        };        

        ClassifierMAPStrategy();

        ~ClassifierMAPStrategy();

        //overload
        bool initialize(ClassifierStrategyParameters const* const strategyParams) ;

        //overload
        bool execute() ;

        unsigned int getOutputNumberBands(void);
        
        //overload
        ClassifierStrategyOutParameters const* getOutputParameters() const;

      protected :

        bool m_isInitialized; //!< Is this instance initialized?

        Parameters::ClassIDT m_highestTrainClassID; //!< The highest train sample lable.
        
        ClassifierMAPStrategy::Parameters m_initParams; //!< Initialization parameters.

        te::cl::MAP m_genClassInstance;  //!< A generir classifier instance.
        
        OutputParameters m_outParams; //!< Output execution parameters.

    };

    /*!
      \class ClassifierMAPStrategyFactory

      \brief Maximum a posteriori probability strategy factory.

      \note Factory key: RegionGrowing
    */
    class TERPEXPORT ClassifierMAPStrategyFactory : public ClassifierStrategyFactory
    {
      public:

        ClassifierMAPStrategyFactory();

        ~ClassifierMAPStrategyFactory();

        //overload
        te::rp::ClassifierStrategy* build();
    };

  } // end namespace rp
}   // end namespace te

#endif // __TERRALIB_RP_INTERNAL_CLASSIFIERMAPSTRATEGY_H

