/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/rp/ClassifierMaxLikelihoodStrategy.cpp

  \brief Maximum a posteriori probability strategy.
*/

// TerraLib

#include "ClassifierMaxLikelihoodStrategy.h"
#include "Macros.h"
#include "ClassifierAdaptors.h"
#include "AlgorithmParametersSerializer.h"
#include "../raster/Band.h"
#include "../raster/Raster.h"
#include "../memory/CachedRaster.h"
#include "../common/MatrixUtils.h"
#include "../common/progress/TaskProgress.h"

#include <boost/numeric/ublas/matrix.hpp>

#include <cfloat>
#include <algorithm>
#include <cmath>
#include <limits>

#define CLASS_ML_STRAT_BETA_ESTIMATION_MAX_VALUE 100.0
#define CLASS_ML_STRAT_BETA_ESTIMATION_MIN_VALUE -1.0
#define CLASS_ML_STRAT_BETA_ESTIMATION_INITIAL_BETA_VALUE -1.0
#define CLASS_ML_STRAT_BETA_ESTIMATION_TRESHOLD_VALUE 0.01

namespace
{
  static te::rp::ClassifierMaxLikelihoodStrategyFactory classifierMaxLikelihoodStrategyFactoryInstance;
}

//-----------------------------------------------------------------------------

te::rp::ClassifierMaxLikelihoodStrategy::Parameters::Parameters()
{
  reset();
}

te::rp::ClassifierMaxLikelihoodStrategy::Parameters::~Parameters() = default;

const te::rp::ClassifierMaxLikelihoodStrategy::Parameters& te::rp::ClassifierMaxLikelihoodStrategy::Parameters::operator=(const te::rp::ClassifierMaxLikelihoodStrategy::Parameters& rhs)
{
  reset();

  m_trainSamplesPtr = rhs.m_trainSamplesPtr;
  m_applyICM = rhs.m_applyICM;
  m_icmChangesThreshold = rhs.m_icmChangesThreshold;
  m_icmMaxIterations = rhs.m_icmMaxIterations;

  return *this;
}

void te::rp::ClassifierMaxLikelihoodStrategy::Parameters::reset() 
{
  m_trainSamplesPtr.reset();
  m_applyICM = false;
  m_icmChangesThreshold = 10.0;
  m_icmMaxIterations = 10;
}

te::common::AbstractParameters* te::rp::ClassifierMaxLikelihoodStrategy::Parameters::clone() const
{
  return new te::rp::ClassifierMaxLikelihoodStrategy::Parameters( *this );
}

bool te::rp::ClassifierMaxLikelihoodStrategy::Parameters::serialize(
  AlgorithmParametersSerializer& serializer ) const
{
  TERP_TRUE_OR_RETURN_FALSE( ClassifierStrategyParameters::serialize( serializer ),
    "Serialization error" );

  TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "applyICM",
    m_applyICM ), "Error adding parameter applyICM" );

  TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "icmChangesThreshold",
    m_icmChangesThreshold ), "Error adding parameter icmChangesThreshold" );

  TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "icmMaxIterations",
    m_icmMaxIterations ), "Error adding parameter icmMaxIterations" );

  return true;
}

//-----------------------------------------------------------------------------

te::rp::ClassifierMaxLikelihoodStrategy::ClassifierMaxLikelihoodStrategy()
{
  m_isInitialized = false;
  m_highestTrainClassID = 0;
}

te::rp::ClassifierMaxLikelihoodStrategy::~ClassifierMaxLikelihoodStrategy() = default;

bool te::rp::ClassifierMaxLikelihoodStrategy::initialize(
  te::rp::ClassifierStrategyParameters const* const strategyParams) 
{
  m_isInitialized = false;
  m_initParams.reset();
  
  {
    ClassifierMaxLikelihoodStrategy::Parameters const * const castParamsPtr = 
      dynamic_cast< ClassifierMaxLikelihoodStrategy::Parameters const * >( strategyParams );
    TERP_INSTANCE_TRUE_OR_RETURN_FALSE( castParamsPtr, "Invalid parameters" );
    
    // Checking the input parameters
    
    TERP_INSTANCE_TRUE_OR_RETURN_FALSE( castParamsPtr->m_trainSamplesPtr,
      "Invalid classes samples pointer" )
    TERP_INSTANCE_TRUE_OR_RETURN_FALSE( castParamsPtr->m_trainSamplesPtr->size() > 0,
      "Invalid classes samples number" )
    TERP_INSTANCE_TRUE_OR_RETURN_FALSE( castParamsPtr->m_trainSamplesPtr->begin()->second.size() > 0,
      "Invalid classes samples number" )    

    m_initParams = (*castParamsPtr);
  }

  // Groupping samples

  std::vector< std::vector< double > > samples;
  std::vector<unsigned int> samplesLabels;
  std::vector<unsigned int> samplesAttributesIndices;

  {
    Parameters::MClassesSamplesCT::const_iterator classesSamplesIt = m_initParams.m_trainSamplesPtr->begin();
    const Parameters::MClassesSamplesCT::const_iterator classesSamplesItEnd = m_initParams.m_trainSamplesPtr->end();
    Parameters::ClassSamplesContainerT::const_iterator classSamplesIt;
    Parameters::ClassSamplesContainerT::const_iterator classSamplesItEnd;

    m_highestTrainClassID = 0;

    while( classesSamplesIt != classesSamplesItEnd )
    {
      classSamplesIt = classesSamplesIt->second.begin();
      classSamplesItEnd = classesSamplesIt->second.end();

      m_highestTrainClassID = std::max( m_highestTrainClassID,
        classesSamplesIt->first );

      while( classSamplesIt != classSamplesItEnd )
      {
        samples.push_back( *classSamplesIt);
        samplesLabels.push_back( classesSamplesIt->first );

        ++classSamplesIt;
      }

      ++classesSamplesIt;
    }

    if( !samples.empty() )
    {
      const std::size_t featuresNmb = samples[ 0 ].size();

      for( std::size_t featureIdx = 0 ; featureIdx < featuresNmb ; ++featureIdx )
      {
        samplesAttributesIndices.push_back(static_cast<unsigned int>(featureIdx));
      }
    }
  }

  // Initializing the generic classifier

  te::cl::MaxLikelihood::Parameters genPars;

  TERP_INSTANCE_TRUE_OR_RETURN_FALSE( m_genClassInstance.initialize( genPars ),
    "Invalid parameters" )
  
  // Training the generic classifier

  te::cl::VectorOfVectorsInputAdaptor< double > samplesAdaptor(
    samples );

  TERP_INSTANCE_TRUE_OR_RETURN_FALSE( m_genClassInstance.train( samplesAdaptor, samplesAttributesIndices,
    samplesLabels, false ), "Classifier trainning error" );

  // Finalizing
  
  m_isInitialized = true;
  
  return true;
}

bool te::rp::ClassifierMaxLikelihoodStrategy::execute() 
{
  TERP_INSTANCE_TRUE_OR_RETURN_FALSE( m_isInitialized,
    "Classification strategy not initialized" );

  // input raster cache

  te::rst::Raster const * inputRasterPtr = m_inputRasterPtr;
  std::unique_ptr< te::mem::CachedRaster > cachedInputRasterPtr;

  if( m_enableRasterCache  )
  {
    assert( m_inputRasterPtr );
    cachedInputRasterPtr.reset( new te::mem::CachedRaster( *m_inputRasterPtr, 25, 0 ) );
    inputRasterPtr = cachedInputRasterPtr.get();
  }
  
  const unsigned int outputPaletteSize = m_highestTrainClassID + 1;
  const unsigned int outputNoDataValue = (double)( outputPaletteSize + 1 );
  
  // Classfying
  
  std::unique_ptr< te::rp::Matrix< unsigned int > > classifiedMatrixPtr( 
    new te::rp::Matrix< unsigned int >() );
  TERP_INSTANCE_TRUE_OR_RETURN_FALSE( classifiedMatrixPtr->reset( 
    inputRasterPtr->getNumberOfRows(), inputRasterPtr->getNumberOfColumns(),
    te::rp::Matrix< unsigned int >::AutoMemPol ), 
    "Classification output matrix creation error" );
  
  {
    te::rp::ca::RasterInputAdaptor< double > inputRasterAdp( *inputRasterPtr );
    te::rp::ca::MatrixOutputAdaptor< unsigned int > outputAdp( *classifiedMatrixPtr );

    std::vector< double > inputNoDataValues;
    for( unsigned int inputRasterBandsIdx = 0 ; inputRasterBandsIdx < m_inputRasterBands.size() ;
      ++inputRasterBandsIdx )
    {
      inputNoDataValues.push_back( inputRasterPtr->getBand(
        m_inputRasterBands[ inputRasterBandsIdx ] )->getProperty()->m_noDataValue );
    }

    TERP_INSTANCE_TRUE_OR_RETURN_FALSE( m_genClassInstance.classify( inputRasterAdp,
      m_inputRasterBands, inputNoDataValues, outputAdp, 0,
      outputNoDataValue, m_progressInterfaceEnabled ), "Classifier execution error" );  
  }
  
  // apply ICM if required
  
  if( m_initParams.m_applyICM )
  {
    std::string errorMessage;
    TERP_INSTANCE_TRUE_OR_RETURN_FALSE( applyICM( *inputRasterPtr,
      classifiedMatrixPtr, outputNoDataValue, errorMessage ),
      "ICM Filter error - " + errorMessage  );
  }

  // Creating the output raster

  te::rst::Raster* outputRasterPtr = 0;
  std::unique_ptr< te::mem::CachedRaster > cachedOutputRasterPtr;
  
  {
    std::vector< int > dt;
    dt.push_back( te::dt::UINT32_TYPE );

    std::vector< double > noDataValues( 1, outputNoDataValue );

    TERP_INSTANCE_TRUE_OR_RETURN_FALSE( createOutputRaster( dt, noDataValues ),
      "Output raster creation error" );

    TERP_INSTANCE_TRUE_OR_RETURN_FALSE( setOutputRasterPalette( outputPaletteSize ),
      "Output raster palette creation error" );

    outputRasterPtr = m_outputRasterPtr.get();
    if( m_enableRasterCache  )
    {
      assert( m_outputRasterPtr.get() );
      cachedOutputRasterPtr.reset( new te::mem::CachedRaster( *m_outputRasterPtr, 33, 0 ) );
      outputRasterPtr = cachedOutputRasterPtr.get();
    }
  }

  {
    const te::rp::Matrix< unsigned int >& inputMatrix = *classifiedMatrixPtr;
    te::rst::Band& outputBand = *outputRasterPtr->getBand( 0 );
    const unsigned int nRows = inputRasterPtr->getNumberOfRows();
    const unsigned int nCols = inputRasterPtr->getNumberOfColumns();
    unsigned int col = 0;
    
    for( unsigned int row = 0 ; row < nRows ; ++row )
    {
      for( col = 0 ; col < nCols ; ++col )
      {
        outputBand.setValue( col, row, (double)inputMatrix( row, col ) );
      }      
    }
  }

  return true;
}

bool te::rp::ClassifierMaxLikelihoodStrategy::applyICM( 
  const te::rst::Raster& inputRaster,
  std::unique_ptr< te::rp::Matrix< unsigned int > >& matrixPtr,
  const unsigned int classifiedMatrixNoDataValue,
  std::string& errorMessage ) const
{
  if( 
      ( matrixPtr.get() == 0 ) 
      ||
      ( matrixPtr->getLinesNumber() == 0 )
    )
  {
    errorMessage = "Invalid classified matrix";
    return false;
  }
  
  // globals
  
  const unsigned int minChangedPixelsPerIteration = (unsigned int)
    m_initParams.m_icmChangesThreshold *
    (double)( matrixPtr->getLinesNumber() *
    matrixPtr->getColumnsNumber() ) / 100.0;
  const unsigned int rowsNumber = matrixPtr->getLinesNumber();
  const unsigned int colsNumber = matrixPtr->getColumnsNumber();
  const unsigned int lastRowIdx = rowsNumber - 1;
  const unsigned int lastColIdx = colsNumber - 1;
  
  // pre-processing classes data
  // reordering to follow class labels IDs
  
  unsigned int maxClassLabelValue = 0;
  std::vector< double > determinantLn; // index: class ID
  std::vector< std::vector< double > > means; // index: class ID
  std::vector< boost::numeric::ublas::matrix< double > > covMatrixes; // index: class ID
  std::vector< boost::numeric::ublas::matrix< double > > inverseCovMatrixes; // index: class ID
  
  {
    std::vector<unsigned int> originalClassLabels;
    if( ! m_genClassInstance.getLables( originalClassLabels ) )
    {
      errorMessage = "Invalid class labels";
      return false;
    }
    
    for( std::size_t originalClassLabelsIdx = 0 ; originalClassLabelsIdx < 
      originalClassLabels.size(); ++originalClassLabelsIdx )
    {
      if( maxClassLabelValue < originalClassLabels[ originalClassLabelsIdx ] )
      {
        maxClassLabelValue = originalClassLabels[ originalClassLabelsIdx ];
      }
    }
    
    determinantLn.resize( maxClassLabelValue + 1 );
    means.resize( maxClassLabelValue + 1 );
    covMatrixes.resize( maxClassLabelValue + 1 );
    inverseCovMatrixes.resize( maxClassLabelValue + 1 );
    
    std::vector< std::vector< double > > originalMeans;
    if( ! m_genClassInstance.getMeans( originalMeans ) )
    {
      errorMessage = "Invalid class means vectors";
      return false;
    }    
    
    std::vector< boost::numeric::ublas::matrix< double > > originalCovMatrixes;
    if( ! m_genClassInstance.getCovMatrixes( originalCovMatrixes ) )
    {
      errorMessage = "Invalid class covariance matrixes";
      return false;
    }
    
    std::vector< boost::numeric::ublas::matrix< double > > originalInverseCovMatrixes;
    if( ! m_genClassInstance.getInverseCovMatrixes( originalInverseCovMatrixes ) )
    {
      errorMessage = "Invalid class inverse covariance matrixes";
      return false;
    }
    
    double classCovarianceMatrixDet = 0;
    
    for( std::size_t originalCovMatrixesIdx = 0 ; originalCovMatrixesIdx < 
      originalCovMatrixes.size() ; ++originalCovMatrixesIdx )
    {
      const unsigned int classLabel = originalClassLabels[ originalCovMatrixesIdx ];
      
      if( te::common::GetDeterminant( originalCovMatrixes[ originalCovMatrixesIdx ],
        classCovarianceMatrixDet ) )
      {
        if( classCovarianceMatrixDet == 0 )
        {
          determinantLn[ classLabel ] = std::numeric_limits< double >::max();
        }
        else
        {
          determinantLn[ classLabel ] = -1.0 * std::log( classCovarianceMatrixDet );
        }
      }
      else
      {
        determinantLn[ classLabel ] = std::numeric_limits< double >::max();
      }
      
      means[ classLabel ] = originalMeans[ originalCovMatrixesIdx ];
      covMatrixes[ classLabel ] = originalCovMatrixes[ originalCovMatrixesIdx ];
      inverseCovMatrixes[ classLabel ] = originalInverseCovMatrixes[ 
        originalCovMatrixesIdx ];
    }
  }
  
  // iterations loop
  
  unsigned int lastIterationChangedPixelsNumber = 
    minChangedPixelsPerIteration + 1;
  unsigned int currentIteration = 0;
  std::unique_ptr< te::rp::Matrix< unsigned int > > auxMatrixPtr;
  te::rp::Matrix< unsigned int > const* inputMatrixPtr = 0;
  te::rp::Matrix< unsigned int >* ouputMatrixPtr = 0;
  double iterationBetaValue = 0;
  
  while( 
          ( currentIteration < m_initParams.m_icmMaxIterations ) 
          &&
          ( lastIterationChangedPixelsNumber > minChangedPixelsPerIteration )
       )
  {
    lastIterationChangedPixelsNumber = 0;

    
    // defining input and output matrixes
    
    if( auxMatrixPtr.get() == 0 )
    {
      auxMatrixPtr.reset(  new te::rp::Matrix< unsigned int >() );
      if( ! auxMatrixPtr->reset( rowsNumber, 
        colsNumber, 
        te::rp::Matrix< unsigned int >::AutoMemPol ) )
      {
        errorMessage = "Auxiliary matrix creation error";
        return false;
      }
      
      inputMatrixPtr = matrixPtr.get();
      ouputMatrixPtr = auxMatrixPtr.get();      
    }
    else
    {
      if( inputMatrixPtr == auxMatrixPtr.get() )
      {
        inputMatrixPtr = matrixPtr.get();
        ouputMatrixPtr = auxMatrixPtr.get();
      }
      else
      {
        inputMatrixPtr = auxMatrixPtr.get();
        ouputMatrixPtr = matrixPtr.get();      
      }
    }
    
    if( ! getICMBeta( *inputMatrixPtr, classifiedMatrixNoDataValue,
      iterationBetaValue, errorMessage ) )
    {
      errorMessage = "ICM beta value calcule error";
      return false;
    }
    
    // border pixels copy
    
    {
      unsigned int row = 0;
      unsigned int col = 0;
      
      for( row = 0 ; row < rowsNumber ; ++row )
      {
        ouputMatrixPtr->operator()( row, 0 ) = inputMatrixPtr->operator()( row, 0 );
        ouputMatrixPtr->operator()( row, lastColIdx ) = 
          inputMatrixPtr->operator()( row, lastColIdx );        
      }

      for( col = 0 ; col < colsNumber ; ++col )
      {
        ouputMatrixPtr->operator()( 0, col ) = inputMatrixPtr->operator()( 0, col );
        ouputMatrixPtr->operator()( lastRowIdx, col ) = 
          inputMatrixPtr->operator()( lastRowIdx, col );        
      }
    }
    
    // filtering
    
    {
    
      // progress
      
      std::unique_ptr< te::common::TaskProgress > progressPtr;
      if( m_progressInterfaceEnabled )
      {
        progressPtr.reset( new te::common::TaskProgress );
        progressPtr->setTotalSteps( rowsNumber - 2 );
        progressPtr->setMessage( "Applying ICM" );
      }
      
      // globals
      
      const std::size_t inputRasterBandsSize = m_inputRasterBands.size();
      const std::size_t neighborClassesCountSize = maxClassLabelValue + 1;
      
      unsigned int row = 0;
      unsigned int col = 0;
      unsigned int centerClassLabel = 0;
      double bestClassDiscriminantFunctionValue = 0;
      unsigned int bestClassLabel = 0;
      std::vector< unsigned int > neighborClassesCount( neighborClassesCountSize, 
        0 );
      unsigned int neighborClassLabel = 0;
      double currentClassDiscriminantFunctionValue = 0;
      std::size_t inputRasterBandsIdx = 0;
      boost::numeric::ublas::matrix< double > sample( inputRasterBandsSize, 1 );
      boost::numeric::ublas::matrix< double > sampleMinusMean( inputRasterBandsSize, 1 );
      boost::numeric::ublas::matrix< double > sampleMinusMeanT( 1, inputRasterBandsSize );
      boost::numeric::ublas::matrix< double > auxMatrix;
      boost::numeric::ublas::matrix< double > mahalanobisDistanceMatrix;
      
      for( row = 1 ; row < lastRowIdx ; ++row )
      {
        for( col = 1 ; col < lastColIdx ; ++col )
        {
          centerClassLabel = inputMatrixPtr->operator()( row, col );
          
          if( centerClassLabel == classifiedMatrixNoDataValue )
          {
            ouputMatrixPtr->operator()( row, col ) = classifiedMatrixNoDataValue;
          }
          else
          {
            // updating surrundings info
            
            std::fill_n( neighborClassesCount.begin(), 
              neighborClassesCount.size(), 0 );
            
            neighborClassLabel = inputMatrixPtr->operator()( row - 1, col );
            assert( neighborClassLabel < neighborClassesCountSize );
            ++neighborClassesCount[ neighborClassLabel ];

            neighborClassLabel = inputMatrixPtr->operator()( row, col - 1 );
            assert( neighborClassLabel < neighborClassesCountSize );
            ++neighborClassesCount[ neighborClassLabel ];

            neighborClassLabel = inputMatrixPtr->operator()( row, col + 1 );
            assert( neighborClassLabel < neighborClassesCountSize );
            ++neighborClassesCount[ neighborClassLabel ];

            neighborClassLabel = inputMatrixPtr->operator()( row + 1, col );
            assert( neighborClassLabel < neighborClassesCountSize );
            ++neighborClassesCount[ neighborClassLabel ];
            
              
            // get best mininum distance 
            
            bestClassDiscriminantFunctionValue = -1.0 * std::numeric_limits< double >::max();
            bestClassLabel = centerClassLabel;
            
            for( neighborClassLabel = 0 ; neighborClassLabel <
              neighborClassesCountSize ; ++neighborClassLabel )
            {
              if( 
                  ( neighborClassLabel != classifiedMatrixNoDataValue )
                  &&
                  ( neighborClassesCount[ neighborClassLabel ] != 0 )
                )
              {
                // load sample

                for( inputRasterBandsIdx = 0 ; inputRasterBandsIdx < 
                  inputRasterBandsSize ; ++inputRasterBandsIdx )
                {
                  inputRaster.getValue( col, row, sample(
                    inputRasterBandsIdx, 0 ), m_inputRasterBands[ 
                    inputRasterBandsIdx] );
                }
                
                //calculate the mahalanobis distance: (x-mean)T * CovMatrizInvert * (x-mean)
                
                for( inputRasterBandsIdx = 0 ; inputRasterBandsIdx < 
                  inputRasterBandsSize ; ++inputRasterBandsIdx )
                {
                  sampleMinusMean( inputRasterBandsIdx, 0 ) = 
                    sampleMinusMeanT( 0, inputRasterBandsIdx ) =
                    ( sample( inputRasterBandsIdx, 0 ) - 
                    means[ neighborClassLabel ][ inputRasterBandsIdx ] );
                }                

                auxMatrix = boost::numeric::ublas::prod( sampleMinusMeanT,
                  inverseCovMatrixes[ neighborClassLabel ] );

                mahalanobisDistanceMatrix = boost::numeric::ublas::prod( auxMatrix, sampleMinusMean );
                assert( mahalanobisDistanceMatrix.size1() == 1 );
                assert( mahalanobisDistanceMatrix.size2() == 1 );

                currentClassDiscriminantFunctionValue = determinantLn[ neighborClassLabel ]
                  - ( 0.5 * mahalanobisDistanceMatrix( 0, 0 ) );
                
                currentClassDiscriminantFunctionValue += ( iterationBetaValue * 
                  ((double)neighborClassesCount[ neighborClassLabel ] ) );
                
                // checking if this is the best neighbor class
                
                if( 
                    currentClassDiscriminantFunctionValue 
                    > 
                    bestClassDiscriminantFunctionValue 
                  )
                {
                  bestClassDiscriminantFunctionValue = currentClassDiscriminantFunctionValue;
                  bestClassLabel = neighborClassLabel;
                }
              }
            }
            
            // Change the current pixel class if necessary
            
            ouputMatrixPtr->operator()( row, col ) = bestClassLabel;
            
            if( bestClassLabel != centerClassLabel )
            {
              ++lastIterationChangedPixelsNumber;
            }
          }
        }
        
        if( m_progressInterfaceEnabled )
        {
          progressPtr->pulse();
          if( ! progressPtr->isActive() ) 
          {
            errorMessage = "Canceled";
            return false;
          }
        }
      }
    }
    
    // going to the next iteration    

    ++currentIteration;
  }
  
  // switching the output matrix if needed
  
  if( ouputMatrixPtr == auxMatrixPtr.get() )
  {
    matrixPtr.reset( auxMatrixPtr.release() );
  }
    
  return true;
}

bool te::rp::ClassifierMaxLikelihoodStrategy::getICMBeta( 
  const te::rp::Matrix< unsigned int >& matrix,
  const unsigned int matrixNoDataValue, 
  double& betaValue, std::string& errorMessage ) const
{
  if( matrix.getLinesNumber() == 0 )
  {
    errorMessage = "Invalid classified matrix";
    return false;
  }
    
  std::vector<unsigned int> originalClassLabels;
  if( ! m_genClassInstance.getLables( originalClassLabels ) )
  {
    errorMessage = "Invalid class labels";
    return false;
  }
  const unsigned int originalClassLabelsSize = originalClassLabels.size();
  
  // globals
  
  const unsigned int rowsNumber = matrix.getLinesNumber();
  const unsigned int colsNumber = matrix.getColumnsNumber();
  const unsigned int lastRowIdx = rowsNumber - 1;
  const unsigned int lastColIdx = colsNumber - 1;  
  
  if( originalClassLabelsSize == 2 )
  {
    double			pp[10];

    int             val, viz;
    unsigned  int   i, j;
    int             cm1m4 = 0, c1m4 = 0, cm1m2 = 0,
            c1m2 = 0, cm12 = 0,
                    c12 = 0, cm14 = 0, c14 = 0;


  //	int	n_filas   = Classif->NumLin();
  //	int	n_colunas = Classif->NumCol();

    pp[0] = CLASS_ML_STRAT_BETA_ESTIMATION_INITIAL_BETA_VALUE;
    pp[1] = CLASS_ML_STRAT_BETA_ESTIMATION_MAX_VALUE;
    pp[2] = CLASS_ML_STRAT_BETA_ESTIMATION_TRESHOLD_VALUE;

    *(pp + 3) = *(pp + 4) = *(pp + 5) = *(pp + 6) = *(pp + 7) = 0.;

    for (i = 1; i < lastRowIdx; i++)
    {
      for (j = 1; j < lastColIdx; j++)
      {
        //val = (int) x[j + i * n_colunas];
        //viz = (int) x[j + (i - 1) * n_colunas] + (int) x[j + (i + 1) * n_colunas] +
        //	(int) x[(j - 1) + i * n_colunas] + (int) x[(j + 1) + i * n_colunas];

        val = (int)matrix( i, j ) - 1;
        if (val<0) continue;
        int c1=(int)matrix(i-1,j  ) - 1, 
          c2=(int)matrix(i+1,j  ) - 1,
          c3=(int)matrix(i,j-1) - 1,
          c4=(int)matrix(i,j+1) - 1;
        if (c1<0||c2 <0 || c3<0 || c4<0) continue;


        viz = c1+c2+c3+c4;				

        if (val == 0) {
          switch (viz) {
            case 0:	cm1m4++;	break;
            case 1:	cm1m2++;	break;
            case 3:	cm12++;		break;
            case 4:	cm14++;		break;
          }
        } else {
          switch (viz) {
            case 4:	c14++;		break;
            case 3:	c12++;		break;
            case 1:	c1m2++;		break;
            case 0:	c1m4++;		break;
          }
        }

        *(pp + 3) = (double) (2 * (c14 - c1m4) + c12 - c1m2);
        *(pp + 4) = (double) (2 * (cm1m4 + c1m4));
        *(pp + 5) = (double) (cm1m2 + c1m2);
        *(pp + 6) = (double) (cm12 + c12);
        *(pp + 7) = (double) (2 * (cm14 + c14));
      }
    }

    /*
     * padronizacao das constantes envolvidas na equacao de
     * pseudo-verossimilhanca dividindo pelo valor maximo
     */

    double          max = -std::numeric_limits< double >::max();

    for (i = 3; i < 8; i++) {
      max = (*(pp + i) > max ? *(pp + i) : max);
    }
    for (i = 3; i < 8; i++) {
      *(pp + i) /= max;
    }

    /* fim padronizacao */

    betaValue = rtbis(plikehood_2, pp);
  }
  else if( originalClassLabelsSize == 3 )
  {
    double		pp[10];

    unsigned        i, j, k, v_400, v_310, v_220, v_211;

    unsigned char  	valores[3];

    //int        c=-1;

  //	int	n_filas   = Classif->NumLin();
  //	int	n_colunas = Classif->NumCol();

    pp[0] = CLASS_ML_STRAT_BETA_ESTIMATION_INITIAL_BETA_VALUE;
    pp[1] = CLASS_ML_STRAT_BETA_ESTIMATION_MAX_VALUE;
    pp[2] = CLASS_ML_STRAT_BETA_ESTIMATION_TRESHOLD_VALUE;

    Estimacao_geral( &pp[3], matrix );	/* esta funcao calcula a			//DEP
               * parte comum para os casos
               * com mais de dois estados
               * possiveis; o resultado e
               * guardado em pp[3] */

    v_400 = v_310 = v_220 = v_211 = 0;

    for (i = 1; i < lastRowIdx; i++)
    {
      for (j = 1; j < lastColIdx; j++)
      {        
        
        for (k = 0; k < 3; k++)
        {
          valores[k] = 0;
        }

        int center=(int)matrix(i,j)-1;
        if (center<0) continue;
        int c1=(int)matrix(i-1,j  ) - 1, 
          c2=(int)matrix(i+1,j  ) - 1,
          c3=(int)matrix(i,j-1) - 1,
          c4=(int)matrix(i,j+1) - 1;
        if (c1<0||c2 <0 || c3<0 || c4<0) continue;

              valores[c1]++;valores[c2]++;valores[c3]++;valores[c4]++;

        qsort(valores, 3, sizeof(unsigned char), compara_uchar_vesre);

        if (valores[0] == 4) {
          v_400++;
        } else if (valores[0] == 3) {
          v_310++;
        } else if (valores[0] == valores[1]) {
          v_220++;
        } else {
          v_211++;
        }
      }
    }

    pp[4] = (double) v_400;
    pp[5] = (double) v_310;
    pp[6] = (double) v_220;
    pp[7] = (double) v_211;

    betaValue = rtbis(plikehood_3, pp);
  }
  else
  {
    double		pp[10];	// Posicao 9 do vetor armazena o numero de classes + 1

    unsigned        i, j, k,
        v_400 = 0, v_310 = 0, 
        v_220 = 0, v_211 = 0, v_111 = 0;

    //int c=-1;
    unsigned char  *valores=NULL;

    if( ( valores = new(std::nothrow) unsigned char[ originalClassLabelsSize ] ) == NULL ){
      errorMessage = "No memory available";
      return false;
    }

    pp[0] = CLASS_ML_STRAT_BETA_ESTIMATION_INITIAL_BETA_VALUE;
    pp[1] = CLASS_ML_STRAT_BETA_ESTIMATION_MAX_VALUE;
    pp[2] = CLASS_ML_STRAT_BETA_ESTIMATION_TRESHOLD_VALUE;


    Estimacao_geral( &pp[3], matrix );	//DEP


  //	int	n_filas   = Classif->NumLin();
  //	int	n_colunas = Classif->NumCol();

    for (i = 1; i < lastRowIdx; i++)
    {
      for (j = 1; j < lastColIdx; j++)
      {          
        for (k = 0; k < originalClassLabelsSize; k++)
        {
          valores[k] = 0;
        }

        int center=(int)matrix(i,j)-1;
        if (center<0) continue;
        int c1=(int)matrix(i-1,j  ) - 1, 
          c2=(int)matrix(i+1,j  ) - 1,
          c3=(int)matrix(i,j-1) - 1,
          c4=(int)matrix(i,j+1) - 1;
        if (c1<0||c2 <0 || c3<0 || c4<0) continue;

              valores[c1]++;valores[c2]++;valores[c3]++;valores[c4]++;
      
        qsort(valores, originalClassLabelsSize, sizeof(unsigned char), compara_uchar_vesre);

        if (valores[0] == 4) {
          v_400++;
        } else if (valores[0] == 3) {
          v_310++;
        } else if (originalClassLabelsSize > 1 && valores[0] == valores[1]) {
          v_220++;
        } else if (valores[0] == 2) {
          v_211++;
        } else {
          v_111++;
        }
      }
    }

  //Liberate output buffers

    if (valores != NULL)
    {
      delete []valores;	
      valores = NULL;
    }

  //	delete [] valores;

    pp[4] = (double) v_400;
    pp[5] = (double) v_310;
    pp[6] = (double) v_220;
    pp[7] = (double) v_211;
    pp[8] = (double) v_111;
    pp[9] = (double)originalClassLabelsSize;

    betaValue = rtbis( plikehood_4, pp );
  }
    
  return true;
}

double te::rp::ClassifierMaxLikelihoodStrategy::rtbis(
  double (*func) (double x, double *param), double *param) const
{
	double          dx, xmid;
	unsigned int    j;
	double          f = (*func) (*param, param), rtb;
	double          fmid = (*func) (*(param + 1), param);

// 	if (f * fmid >= 0.)
// 		printf("\nO intervalo deve incluir a raiz\n");

	rtb = f < 0. ? (dx = *(param + 1) - *param, *param) : (dx = *param - *(param + 1), *(param + 1));

	for (j = 1; j <= std::numeric_limits< unsigned int>::max(); j++) {

		fmid = (*func) (xmid = rtb + (dx *= .5), param);
		if (fmid <= 0.)
			rtb = xmid;
		if (fabs(dx) < *(param + 2) || fmid == 0.)
			return (rtb);
	}

	//printf("\nUltrapassou o maximo de iteracoes permitido\n");

	return (*(param + 1));
}	

double te::rp::ClassifierMaxLikelihoodStrategy::plikehood_2(double search, double *pp)
{

	double	t2, t4, tm2, tm4;
	
	t2 = exp(2 * (double) search);
	t4 = t2 * t2;
	tm2 = 1 / t2;
	tm4 = 1 / t4;

	return pp[3] + pp[4] * sigmo(tm4) + pp[5] * sigmo(tm2) - pp[6] * sigmo(t2) - pp[7] * sigmo(t4);

}

double te::rp::ClassifierMaxLikelihoodStrategy::plikehood_3(double search, double *pp)
{
	double  t1, t2, t3, t4;

	t1 = exp(search);
	t2 = t1 * t1;
	t3 = t2 * t1;
	t4 = t2 * t2;
	return   pp[3] - ((pp[4] * 4. * t4 / (2. + t4)) + (pp[5] * ((3. * t3 + t1) / (t3 + t1 + 1.))) +
		 (pp[6] * 4. * t2 / (2. * t2 + 1.)) + (pp[7] * (2. * t1 + 2.) / (t1 + 2.)));
}

double te::rp::ClassifierMaxLikelihoodStrategy::plikehood_4(double search, double *pp)
{				/* o numero de classes e passado na posicao
				 * "10" do vetor pp */
	double   t1, t2, t3, t4;

	t1 = exp(search);
	t2 = t1 * t1;
	t3 = t2 * t1;
	t4 = t2 * t2;

	return pp[3] - 
		((pp[4] * 4. * t4 / ((pp[9] - 1.) + t4)) + (pp[5] * ((3. * t3 + t1) / (t3 + t1 + (pp[9] - 2.)))) + (pp[6] * 4. * t2 / (2. * t2 + (pp[9] - 2.))) + (pp[7] * (2. * t2 + 2. * t1) / (t2 + 2. * t1 + (pp[9] - 3.)) + (pp[8] * 4. * t1 / (4. * t1 + (pp[9] - 4.)))));
}

double te::rp::ClassifierMaxLikelihoodStrategy::sigmo( double a )
{
	if( a != -1.)
		return   a  / ( 1. + a );
	else
	{
//		printf("\nDescontinuidade em SIGMOD");
		return fabs(a) / a * std::numeric_limits< double >::max();
	}
}

void te::rp::ClassifierMaxLikelihoodStrategy::Estimacao_geral(  
  double *termo_comum, const te::rp::Matrix< unsigned int >& matrix )
{
	/*
	 * calculo da parte que nao envolve o valor de beta na equacao de
	 * pseudoverossimilhanca
	 */

	*termo_comum = 0.;

	unsigned char   v_ij;
	unsigned        i, j;
  const unsigned int lastRowIdx = matrix.getLinesNumber() - 1;
  const unsigned int lastColIdx = matrix.getColumnsNumber() - 1;
  
	for (i = 1; i < lastRowIdx ; i++)
	{
		for (j = 1; j < lastColIdx ; j++)
		{
			//v_ij = x[j + i * n_colunas];
			//if( x[j + (i - 1) * n_colunas] == v_ij)	*termo_comum++;
			//if( x[j + (i + 1) * n_colunas] == v_ij)	*termo_comum++;
			//if( x[(j - 1) + i * n_colunas] == v_ij)	*termo_comum++;
			//if( x[(j + 1) + i * n_colunas] == v_ij)	*termo_comum++;

			v_ij = (int)matrix(i,j);

			if( matrix(i-1,j   ) == v_ij ) (*termo_comum)++;
			if( matrix(i+1,j   ) == v_ij ) (*termo_comum)++;
			if( matrix(i  ,j-1 ) == v_ij ) (*termo_comum)++;
			if( matrix(i  ,j+1 ) == v_ij ) (*termo_comum)++;
		}
	}
}

int te::rp::ClassifierMaxLikelihoodStrategy::compara_uchar_vesre(const void *x, const void *y)
{
	if (*(unsigned char *) x < *(unsigned char *) y)
		return (1);
	if (*(unsigned char *) x > *(unsigned char *) y)
		return (-1);
	if (*(unsigned char *) x == *(unsigned char *) y)
		return (0);
return 1;
}

  
//-----------------------------------------------------------------------------

te::rp::ClassifierMaxLikelihoodStrategyFactory::ClassifierMaxLikelihoodStrategyFactory()
  : te::rp::ClassifierStrategyFactory("maxlikelihood")
{
}

te::rp::ClassifierMaxLikelihoodStrategyFactory::~ClassifierMaxLikelihoodStrategyFactory() = default;

te::rp::ClassifierStrategy* te::rp::ClassifierMaxLikelihoodStrategyFactory::build()
{
  return new te::rp::ClassifierMaxLikelihoodStrategy();
}
