/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/rp/ClassifierMaxLikelihoodStrategy.h
  \brief A maximum likelihood estimation strategy for classification (a.k.a. MaxVer in portuguese).
 */

#ifndef __TERRALIB_RP_INTERNAL_CLASSIFIERMAXLIKELIHOODSTRATEGY_H
#define __TERRALIB_RP_INTERNAL_CLASSIFIERMAXLIKELIHOODSTRATEGY_H

#include "ClassifierStrategy.h"
#include "ClassifierStrategyFactory.h"
#include "Matrix.h"
#include "Config.h"
#include "../classification/MaxLikelihood.h"
#include "../classification/Adaptors.h"

#include <boost/shared_ptr.hpp>

#include <map>
#include <vector>

namespace te
{
  namespace rp
  {
    class AlgorithmParametersSerializer;

    /*!
      \class ClassifierMaxLikelihoodStrategy

      \brief A maximum likelihood estimation strategy for classification (a.k.a. MaxVer in portuguese).
      
      \ingroup rp_class
      
      \see te::cl::MaxLikelihood class docs
     */
    class TERPEXPORT ClassifierMaxLikelihoodStrategy : public ClassifierStrategy
    {
      public:

        /*!
          \class Parameters

          \brief Classifier Parameters
        */
        class TERPEXPORT Parameters : public ClassifierStrategyParameters
        {
          public:
            
            typedef unsigned int ClassIDT; //!< Class ID type definition (zero means invalid ID).
            
            typedef std::vector< double > ClassSampleT; //!< Class sample type definition.
            
            typedef std::vector< ClassSampleT > ClassSamplesContainerT; //!< Class samples container type definition.
            
            typedef std::map< ClassIDT, ClassSamplesContainerT > MClassesSamplesCT; //!< Multi-classes samples container type definition.

            typedef boost::shared_ptr<MClassesSamplesCT> MClassesSamplesCTPtr; //!< A shared pointer to a multi classes samples container type definition.
            
            MClassesSamplesCTPtr m_trainSamplesPtr; //!< A shared pointer to a always-valid structure where trainning samples are stored.
            
            bool m_applyICM; //!< Apply a Iterated conditional modes iteractive filtering (default:false);
            
            float m_icmChangesThreshold;  //!< Minimum porcentage of number of pixel that changes from one iteration to the other [valid values range:1-100, default:10)].
            
            unsigned int m_icmMaxIterations;  //!< Maximum number of ICM iterations to perform (default 10).

            Parameters();

            ~Parameters();

            //overload
            const Parameters& operator=(const Parameters& params);

            //overload
            void reset() ;

            //overload
            AbstractParameters* clone() const;

            //overload
            bool serialize ( AlgorithmParametersSerializer& serializer ) const;
        };

        ClassifierMaxLikelihoodStrategy();

        ~ClassifierMaxLikelihoodStrategy();

        //overload
        bool initialize(ClassifierStrategyParameters const* const strategyParams) ;

        //overload
        bool execute() ;

        unsigned int getOutputNumberBands(void);

      protected :

        bool m_isInitialized; //!< Is this instance initialized?

        Parameters::ClassIDT m_highestTrainClassID; //!< The highest train sample lable.
        
        ClassifierMaxLikelihoodStrategy::Parameters m_initParams; //!< Initialization parameters.

        te::cl::MaxLikelihood m_genClassInstance;  //!< A generir classifier instance.

        /*!
          \brief Apply an ICM filter
          \param inputRaster Input raster.
          \param matrixPtr A pointer to a input/output classified matrix (the pointer may be reassigned if required).
          \param classifiedMatrixNoDataValue Matrix no-data value.
          \param errorMessage Error message.
          \return true if OK, false on errors;
        */
        bool applyICM(
          const te::rst::Raster& inputRaster,
          std::unique_ptr< te::rp::Matrix< unsigned int > >& matrixPtr,
          const unsigned int classifiedMatrixNoDataValue,
          std::string& errorMessage ) const;
          
        /*!
          \brief Iteration Beta value Calule.
          \param matrix Current classified matrix.
          \param matrixNoDataValue Matrix no-data value.
          \param errorMessage Error message.
          \return true if OK, false on errors;
        */
        bool getICMBeta( const te::rp::Matrix< unsigned int >& matrix,
          const unsigned int matrixNoDataValue, 
          double& betaValue, std::string& errorMessage ) const;
          
        /*!
           \brief Metodo da biseccao para achar os zeros de uma funcao de dupla precisao com
            parametros. Numerical Recipes in C cap 9. Versao del 22 de agosto de 1990.
            INPE. O vetor (double)param contem os seguintes valores: param[0] = o
            minimo do intervalo de procura param[1] = o maximo do intervalo de procura
            param[2] = a tolerancia param[3] ... param[n] = parametros da funcao
            
            \note Imported from SPRING
         */
        double rtbis(double (*func) (double x, double *param), double *param) const;
        
        static double	plikehood_2(double  , double * );
        
        static double	plikehood_3(double  , double * );
        
        static double	plikehood_4(double  , double * );        
        
        static double sigmo( double a );
        
        /*! \brief Compute the common part in pseudo-likelihood equation.
          \param img_box : image to process.
          \param term : comom part
        */
        static void	Estimacao_geral(  double *term, const te::rp::Matrix< unsigned int >& matrix  );        
        
        static int compara_uchar_vesre(const void *x, const void *y);
    };

    /*!
      \class ClassifierMaxLikelihoodStrategyFactory

      \brief Maximum a posteriori probability strategy factory.

      \note Factory key: RegionGrowing
    */
    class TERPEXPORT ClassifierMaxLikelihoodStrategyFactory : public ClassifierStrategyFactory
    {
      public:

        ClassifierMaxLikelihoodStrategyFactory();

        ~ClassifierMaxLikelihoodStrategyFactory();

        //overload
        te::rp::ClassifierStrategy* build();
    };

  } // end namespace rp
}   // end namespace te

#endif // __TERRALIB_RP_INTERNAL_CLASSIFIERMAXLIKELIHOODSTRATEGY_H

