/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/clouddetection/CloudDetection.cpp

  \brief Implementation of the cloud detection.
*/

// Terralib
#include "../raster.h"
#include "../rp/Filter.h"
#include "CloudDetection.h"
#include "Macros.h"

te::rp::CloudDetection::CloudDetection() = default;

te::rp::CloudDetection::~CloudDetection() = default;

te::rst::Raster* te::rp::CloudDetection::executeCloudDetection(te::rst::Raster* cloudRaster, te::rst::Raster* shadowRaster, int cloudBand, int shadowBand, double minCloud, double maxCloud,
  double minShadow, double maxShadow, std::vector<te::rp::Filter::InputParameters::FilterType> vecFilter, std::vector<int> vecIt)
{
  //cloud limiar
  te::rst::Raster* limiarCloudRaster = generateCloudRaster(cloudRaster, cloudBand, minCloud, maxCloud);
  
  //shadow limiar
  te::rst::Raster* limiarShadowRaster = generateShadowRaster(shadowRaster, shadowBand, minShadow, maxShadow);

  //Sum of cloud limiar and shadow limiar
  te::rst::Raster* raster = generateSumRasters(limiarCloudRaster, limiarShadowRaster);

  //Apply morphological filters
  for (unsigned int i = 0; i < vecFilter.size(); i++)
  {
    {
      //run filter
      te::rp::Filter algorithmInstance;

      te::rp::Filter::InputParameters algoInputParams;
      algoInputParams.m_iterationsNumber = static_cast<unsigned int>(vecIt[i]);
      algoInputParams.m_enableProgress = true;
      algoInputParams.m_filterType = vecFilter[i];
      algoInputParams.m_inRasterPtr = raster;
      algoInputParams.m_inRasterBands.push_back(0);

      std::map< std::string, std::string > dummyRInfo;

      te::rp::Filter::OutputParameters algoOutputParams;
      algoOutputParams.m_rInfo = dummyRInfo;
      algoOutputParams.m_rType = "MEM";

      if (algorithmInstance.initialize(algoInputParams))
      {
        if (!algorithmInstance.execute(algoOutputParams))
        {
          return nullptr;
        }
      }

      delete raster;

      raster = algoOutputParams.m_outputRasterPtr.release();
    }
  }

  //adjust dummy value
  std::size_t nBands = raster->getNumberOfBands();

  for (std::size_t t = 0; t < nBands; ++t)
  {
    te::rst::BandProperty* bProp = raster->getBand(t)->getProperty();
    bProp->m_noDataValue = 255.;
  }

  delete limiarCloudRaster;
  delete limiarShadowRaster;

  return raster;
}

te::rst::Raster* te::rp::CloudDetection::generateCloudRaster(te::rst::Raster* raster, int cloudBand, double minCloud, double maxCloud)
{
  //load to memory
  std::vector<te::rst::BandProperty*> bprops;

  te::rst::BandProperty* bp = new te::rst::BandProperty(0, te::dt::UCHAR_TYPE);
  bp->m_noDataValue = 255.;

  bprops.push_back(bp);

  std::map< std::string, std::string > dummyRInfo;

  te::rst::Raster* rasterMem = te::rst::RasterFactory::make("MEM", new te::rst::Grid(*(raster->getGrid())), bprops, dummyRInfo, nullptr, nullptr);

  const te::rst::Band& bin = *raster->getBand(static_cast<size_t>(cloudBand));
  const unsigned int nRows = raster->getNumberOfRows();
  const unsigned int nCols = raster->getNumberOfColumns();

  for (unsigned int row = 0; row < nRows; ++row)
  {
    for (unsigned int col = 0; col < nCols; ++col)
    {
      double val;
      bin.getValue(col, row, val);

      if (val >= minCloud && val <= maxCloud)
        rasterMem->setValue(col, row, 0.);
      else
        rasterMem->setValue(col, row, 255.);
    }
  }

  return rasterMem;
}

te::rst::Raster* te::rp::CloudDetection::generateShadowRaster(te::rst::Raster* raster, int shadowBand, double minShadow, double maxShadow)
{
  //load to memory
  std::vector<te::rst::BandProperty*> bprops;

  te::rst::BandProperty* bp = new te::rst::BandProperty(0, te::dt::UCHAR_TYPE);
  bp->m_noDataValue = 255.;

  bprops.push_back(bp);

  std::map< std::string, std::string > dummyRInfo;

  te::rst::Raster* rasterMem = te::rst::RasterFactory::make("MEM", new te::rst::Grid(*(raster->getGrid())), bprops, dummyRInfo, nullptr, nullptr);

  const te::rst::Band& bin = *raster->getBand(static_cast<size_t>(shadowBand));
  const unsigned int nRows = raster->getNumberOfRows();
  const unsigned int nCols = raster->getNumberOfColumns();

  for (unsigned int row = 0; row < nRows; ++row)
  {
    for (unsigned int col = 0; col < nCols; ++col)
    {
      double val;
      bin.getValue(col, row, val);

      if (val >= minShadow && val <= maxShadow)
        rasterMem->setValue(col, row, 0.);
      else
        rasterMem->setValue(col, row, 255.);
    }
  }

  return rasterMem;
}

te::rst::Raster* te::rp::CloudDetection::generateSumRasters(te::rst::Raster* cloudRaster, te::rst::Raster* shadowRaster)
{
  //load to memory
  std::vector<te::rst::BandProperty*> bprops;

  te::rst::BandProperty* bp = new te::rst::BandProperty(0, te::dt::UCHAR_TYPE);

  bprops.push_back(bp);

  std::map< std::string, std::string > dummyRInfo;

  te::rst::Raster* rasterMem = nullptr;
  try
  {
    rasterMem = te::rst::RasterFactory::make("MEM", 
      new te::rst::Grid(*(cloudRaster->getGrid())), bprops, dummyRInfo, nullptr, nullptr);
  }
  catch(...)
  {
    TERP_LOGERR( "Output raster creation error" );
    return nullptr;
  }   

  const te::rst::Band& cloudBand = *cloudRaster->getBand(0);
  const te::rst::Band& shadowBand = *shadowRaster->getBand(0);
  const unsigned int nRows = cloudRaster->getNumberOfRows();
  const unsigned int nCols = cloudRaster->getNumberOfColumns();

  for (unsigned int row = 0; row < nRows; ++row)
  {
    for (unsigned int col = 0; col < nCols; ++col)
    {
      double cloudValue;
      cloudBand.getValue(col, row, cloudValue);

      if (cloudValue == 255.)
      {
        double shadowValue;
        shadowBand.getValue(col, row, shadowValue);

        rasterMem->setValue(col, row, shadowValue);
      }
      else
      {
        rasterMem->setValue(col, row, cloudValue);
      }
    }
  }

  return rasterMem;
}
