/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

#include "Composer.h"

#include "Macros.h"
#include "../geometry/LinearRing.h"
#include "../geometry/CircularString.h"
#include "../geometry/MultiPoint.h"
#include "../geometry/MultiLineString.h"
#include "../geometry/Point.h"
#include "../geometry/Coord2D.h"
#include "../geometry/CurvePolygon.h"
#include "../geometry/Envelope.h"
#include "../geometry/Enums.h"
#include "../raster/Raster.h"
#include "../raster/Grid.h"
#include "../raster/Band.h"
#include "../raster/BandProperty.h"
#include "../raster/Utils.h"
#include "../raster/SynchronizedRaster.h"
#include "../srs/Converter.h"
#include "../common/PlatformUtils.h"
#include "../common/MathUtils.h"
#include "../common/progress/TaskProgress.h"

#include <boost/graph/graph_concepts.hpp>
#include <boost/scoped_ptr.hpp>
#include <boost/scoped_array.hpp>


#include <complex>
#include <limits>
#include <algorithm>
#include <memory>

#define UPDATER3LIMITSBYCOORD( inputRasterptr, raster3Ptr, targetCol, targetRow ) \
{ \
  double inputRasterX = 0; \
  double inputRasterY = 0; \
  double raster3X = 0; \
  double raster3Y = 0; \
  double raster3Row = 0; \
  double raster3Col = 0; \
  te::srs::Converter conv( inputRasterptr->getSRID(), raster3Ptr->getSRID() ); \
  inputRasterptr->getGrid()->gridToGeo( targetCol, targetRow, inputRasterX, inputRasterY ); \
  conv.convert( inputRasterX, inputRasterY, raster3X, raster3Y ); \
  raster3Ptr->getGrid()->geoToGrid( raster3X, raster3Y, raster3Col, raster3Row ); \
  raster3Col = (double)te::common::Round< double, long int >( raster3Col ); \
  raster3Col = std::max( 0.0, raster3Col ); \
  raster3Col = std::min( (double)( raster3Ptr->getNumberOfColumns() - 1 ), raster3Col ); \
  firstRaster3Col = std::min( firstRaster3Col, (unsigned int)raster3Col ); \
  lastRaster3Col = std::max( lastRaster3Col, (unsigned int)raster3Col ); \
  raster3Row = (double)te::common::Round< double, long int >( raster3Row ); \
  raster3Row = std::max( 0.0, raster3Row ); \
  raster3Row = std::min( (double)( raster3Ptr->getNumberOfRows() - 1 ), raster3Row ); \
  firstRaster3Row = std::min( firstRaster3Row, (unsigned int)raster3Row ); \
  lastRaster3Row = std::max( lastRaster3Row, (unsigned int)raster3Row ); \
}

#define UPDATER3LIMITS( inputRasterptr, raster3Ptr ) \
{ \
  UPDATER3LIMITSBYCOORD( inputRasterptr, raster3Ptr, 0 , 0 ); \
  UPDATER3LIMITSBYCOORD( inputRasterptr, raster3Ptr, 0 , inputRasterptr->getNumberOfRows() - 1 ); \
  UPDATER3LIMITSBYCOORD( inputRasterptr, raster3Ptr, inputRasterptr->getNumberOfColumns() - 1, 0 ); \
  UPDATER3LIMITSBYCOORD( inputRasterptr, raster3Ptr, inputRasterptr->getNumberOfColumns() - 1, inputRasterptr->getNumberOfRows() - 1 ); \
}

// #define updateR3Limits( rasterPtr, row, col )

namespace te
{
  namespace rp
  {
    Composer::Composer()
    {
      initState();
    }
    
    Composer::~Composer()
    {
      clear();
    }    
    
    Composer::Composer( const Composer& )
    {
    }
    
    Composer& Composer::operator=( const Composer& )
    {
      return *this;
    }
    
    bool Composer::initialize( te::rst::Raster& raster1, 
      const std::vector< unsigned int >& raster1Bands, 
      const te::rst::Raster& raster2, 
      const std::vector< unsigned int >& raster2Bands,
      te::rst::Raster& raster3, 
      const std::vector< unsigned int >& raster3Bands,
      const te::rst::Interpolator::Method& interpMethod1,
      const te::rst::Interpolator::Method& interpMethod2,
      const double& noDataValue,
      const bool forceRaster1NoDataValue,
      const bool forceRaster2NoDataValue,
      const std::vector< double >& pixelOffsets1,
      const std::vector< double >& pixelScales1,
      const std::vector< double >& pixelOffsets2,
      const std::vector< double >& pixelScales2,
      const unsigned int threadsNumber,
      const bool enableProgressInterface,
      ComposerRule const * const rulePtr,
      const bool enableRasterCache )
    {
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( 
        raster1.getAccessPolicy() & te::common::RAccess, 
        "Invalid raster 1" );        
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( 
        raster2.getAccessPolicy() & te::common::RAccess, 
        "Invalid raster 2" );
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( 
        raster3.getAccessPolicy() & te::common::RWAccess, 
        "Invalid raster 3" );       
      
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( raster1Bands.size() > 0,
        "Invalid raster bands vector" );      
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( raster1Bands.size() ==
        raster2Bands.size(), "Invalid raster bands vector" );
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( raster3Bands.size() ==
        raster1Bands.size(), "Invalid raster bands vector" );      
      
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( pixelOffsets1.size() ==  
        raster1Bands.size(), "Invalid pixel offsets" );
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( pixelScales1.size() ==  
        raster1Bands.size(), "Invalid pixel scales" );        
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( pixelOffsets2.size() ==  
        raster2Bands.size(), "Invalid pixel offsets" );
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( pixelScales2.size() ==  
        raster2Bands.size(), "Invalid pixel scales" );
      
      clear();
      
      // defining the input rasters
      
      m_raster1Ptr = &raster1;
      m_raster2Ptr = &raster2;
      m_raster3Ptr = &raster3;
      
      m_interpMethod1 = interpMethod1;
      m_interpMethod2 = interpMethod2;
      
      // defining dummy values
      
      m_forceRaster1NoDataValue = forceRaster1NoDataValue;
      m_forceRaster2NoDataValue = forceRaster2NoDataValue;
      
      for( std::vector< unsigned int >::size_type rasterBandsIdx = 0 ; 
        rasterBandsIdx < raster1Bands.size() ; ++rasterBandsIdx )
      {
        TERP_INSTANCE_TRUE_OR_RETURN_FALSE( raster1Bands[ rasterBandsIdx ] <
          raster1.getNumberOfBands(), "Invalid raster 1 band" );
        TERP_INSTANCE_TRUE_OR_RETURN_FALSE( raster2Bands[ rasterBandsIdx ] <
          raster2.getNumberOfBands(), "Invalid raster2 band" );            
        TERP_INSTANCE_TRUE_OR_RETURN_FALSE( raster3Bands[ rasterBandsIdx ] <
          raster3.getNumberOfBands(), "Invalid raster 3 band" );                    
      }
      
      m_outputNoDataValue = noDataValue;
      
      // defining raster bands
      
      m_raster1Bands = raster1Bands;
      m_raster2Bands = raster2Bands;
      m_raster3Bands = raster3Bands;
      
      // checking raster 3 blocking scheme
      
      {
        const te::rst::Band& firstR3Band = *( m_raster3Ptr->getBand(
           m_raster3Bands[ 0 ] ) );        
        
        for( unsigned int raster3BandsIdx = 0 ; raster3BandsIdx < 
          m_raster3Bands.size() ; ++raster3BandsIdx )
        {
          const te::rst::Band& band = *( m_raster3Ptr->getBand(
            m_raster3Bands[ raster3BandsIdx ] ) ); 
          
          if(
              ( band.getProperty()->m_blkh != firstR3Band.getProperty()->m_blkh )
              ||
              ( band.getProperty()->m_blkw != firstR3Band.getProperty()->m_blkw )
              ||
              ( band.getProperty()->m_nblocksx != firstR3Band.getProperty()->m_nblocksx )
              ||
              ( band.getProperty()->m_nblocksy != firstR3Band.getProperty()->m_nblocksy )
            )
          {
            TERP_INSTANCE_LOG_AND_RETURN_FALSE( "Invalid raster 3 bands scheme" );
          }
        }      
      }
      
      // defining pixel offsets
      
      m_pixelOffsets1 = pixelOffsets1;
      m_pixelScales1 = pixelScales1;
      
      m_pixelOffsets2 = pixelOffsets2;
      m_pixelScales2 = pixelScales2;
      
      // threads
      
      if( threadsNumber == 0 )
      {
        m_threadsNumber = te::common::GetPhysProcNumber();
      }
      else
      {
        m_threadsNumber = threadsNumber;
      }
      
      // Rule
      
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( rulePtr != nullptr, "No rule provided" );
      
      m_rulePtr.reset( rulePtr->clone() );
      
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( 
        m_rulePtr->initialize(
          raster1,
          raster1Bands,
          raster2,
          raster2Bands,
          raster3,
          raster3Bands,
          interpMethod1,
          interpMethod2,
          noDataValue,
          forceRaster1NoDataValue,
          forceRaster2NoDataValue,
          pixelOffsets1,
          pixelScales1,
          pixelOffsets2,
          pixelScales2
        ),
        "Rule initialization error - " + m_rulePtr->getErrorMessage() );
      
      // cache 
      
      m_enableRasterCache = enableRasterCache;
      
      // progress interface
      
      m_enableProgressInterface = enableProgressInterface;
      
      return true;
    }
    
    void Composer::initState()
    {
      m_enableProgressInterface = false;
      m_forceRaster1NoDataValue = false;
      m_forceRaster2NoDataValue = false;
      m_threadsNumber = 0;
      m_raster1Ptr = nullptr;
      m_raster2Ptr = nullptr;
      m_raster3Ptr = nullptr;
      m_interpMethod1 = te::rst::NearestNeighbor;
      m_interpMethod2 = te::rst::NearestNeighbor;
      m_outputNoDataValue = 0;
      m_enableRasterCache = true;
    }
    
    void Composer::clear()
    {
      m_raster1Bands.clear();
      m_raster2Bands.clear();
      m_raster3Bands.clear();
      m_pixelOffsets1.clear();
      m_pixelScales1.clear();
      m_pixelOffsets2.clear();
      m_pixelScales2.clear();
      m_rulePtr.reset();
      m_errorMessage.clear();

      initState();
    }

    bool Composer::composeIntoRaster3()
    {
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( m_raster3Ptr->getAccessPolicy() & 
        te::common::WAccess, "Invalid output raster 3 access policy" );      
      
      // Locating the raster 3 area that need processing
      
      unsigned int firstRaster3Col = 0;
      unsigned int lastRaster3Row = 0;
      unsigned int lastRaster3Col = 0;
      unsigned int firstRaster3Row = 0;
      
      try
      {
        UPDATER3LIMITS( m_raster1Ptr, m_raster3Ptr )
        UPDATER3LIMITS( m_raster2Ptr, m_raster3Ptr )
      }  
      catch(...)
      {
        TERP_INSTANCE_LOG_AND_RETURN_FALSE( "Unable to locate the output raster processing area");
      }
      
      // Discovering the raster 1 blocks we need to process
      
      std::vector< RasterBlockInfo > raster3BlocksInfos;
      
      {
        const te::rst::Band& r3FirstBand = *( m_raster3Ptr->getBand(
           m_raster3Bands[ 0 ] ) );        
          
        unsigned int firstBlockX = firstRaster3Col / 
          r3FirstBand.getProperty()->m_blkw;
        unsigned int lastBlockX = lastRaster3Col / 
          r3FirstBand.getProperty()->m_blkw;
        unsigned int firstBlockY = firstRaster3Row / 
          r3FirstBand.getProperty()->m_blkh;
        unsigned int lastBlockY = lastRaster3Row / 
          r3FirstBand.getProperty()->m_blkh;                        
          
        for( unsigned int blkY = firstBlockY ; blkY <= lastBlockY ; ++blkY )
        {
          for( unsigned int blkX = firstBlockX ; blkX <= lastBlockX ; ++blkX )
          {
            raster3BlocksInfos.push_back( RasterBlockInfo() );
            
            RasterBlockInfo& rBInfo = raster3BlocksInfos.back();
            
            rBInfo.m_wasProcessed = false;            
            rBInfo.m_blkX = blkX;
            rBInfo.m_blkY = blkY;
            rBInfo.m_blkTotalPixelsNumber = r3FirstBand.getProperty()->m_blkh *
              r3FirstBand.getProperty()->m_blkw;
            
            rBInfo.m_firstRasterRow2Process = blkY * r3FirstBand.getProperty()->m_blkh;
            rBInfo.m_firstRasterRow2Process = std::max( firstRaster3Row,
              rBInfo.m_firstRasterRow2Process );
            rBInfo.m_firstRasterRow2Process = std::min( lastRaster3Row,
              rBInfo.m_firstRasterRow2Process );
            
            rBInfo.m_rasterRows2ProcessBound = ( blkY + 1 ) * r3FirstBand.getProperty()->m_blkh;
            rBInfo.m_rasterRows2ProcessBound = std::max( firstRaster3Row,
              rBInfo.m_rasterRows2ProcessBound );
            rBInfo.m_rasterRows2ProcessBound = std::min( lastRaster3Row + 1,
              rBInfo.m_rasterRows2ProcessBound );              
            
            rBInfo.m_firstRasterCol2Process = blkX * r3FirstBand.getProperty()->m_blkw;
            rBInfo.m_firstRasterCol2Process = std::max( firstRaster3Col,
              rBInfo.m_firstRasterCol2Process );
            rBInfo.m_firstRasterCol2Process = std::min( lastRaster3Col,
              rBInfo.m_firstRasterCol2Process );              
            
            rBInfo.m_rasterCols2ProcessBound = ( blkX + 1 ) * r3FirstBand.getProperty()->m_blkw;          
            rBInfo.m_rasterCols2ProcessBound = std::max( firstRaster3Col,
              rBInfo.m_rasterCols2ProcessBound );
            rBInfo.m_rasterCols2ProcessBound = std::min( lastRaster3Col + 1,
              rBInfo.m_rasterCols2ProcessBound );                
          }
        }
      }          
      
      // composing
      
      bool returnValue = true;
      
      {
        // Guessing memory resources

        const double totalPhysMem = (double)te::common::GetTotalPhysicalMemory();
        const double usedVMem = (double)te::common::GetUsedVirtualMemory();
        const double totalVMem = ( (double)te::common::GetTotalVirtualMemory() );
        const double maxVMem2Use = 0.75 * std::min( totalPhysMem, ( totalVMem - usedVMem ) );        
        
        // Creating thread exec params
        
        bool abortValue = false;
        std::mutex mutex;
        te::rst::RasterSynchronizer sync1( *((te::rst::Raster*)m_raster1Ptr), 
          te::common::RAccess );
        te::rst::RasterSynchronizer sync2( *((te::rst::Raster*)m_raster2Ptr), 
          te::common::RAccess );
        te::rst::RasterSynchronizer sync3( *m_raster3Ptr, te::common::RWAccess );
        std::mutex blockProcessedSignalMutex;
        std::condition_variable blockProcessedSignal;
        unsigned int runningThreadsCounter = 0;
        unsigned int processedBlocksNmb = 0;
        
        ComposeIntoRaster3ThreadParams auxThreadParams;
        auxThreadParams.m_returnValuePtr = &returnValue;
        auxThreadParams.m_abortValuePtr = &abortValue;
        auxThreadParams.m_sync1Ptr = &sync1;
        auxThreadParams.m_sync2Ptr = &sync2;
        auxThreadParams.m_sync3Ptr = &sync3;
        auxThreadParams.m_raster3BlocksInfosPtr = &raster3BlocksInfos;
        auxThreadParams.m_mutexPtr = &mutex;
        auxThreadParams.m_blockProcessedSignalMutexPtr = &blockProcessedSignalMutex;
        auxThreadParams.m_blockProcessedSignalPtr = &blockProcessedSignal;
        auxThreadParams.m_runningThreadsCounterPtr = &runningThreadsCounter;
        auxThreadParams.m_raster1BandsPtr = &m_raster1Bands;
        auxThreadParams.m_raster2BandsPtr = &m_raster2Bands;
        auxThreadParams.m_raster3BandsPtr = &m_raster3Bands;
        auxThreadParams.m_interpMethod1 = m_interpMethod1;
        auxThreadParams.m_interpMethod2 = m_interpMethod2;
        auxThreadParams.m_noDataValue = m_outputNoDataValue;
        auxThreadParams.m_forceRaster1NoDataValue = m_forceRaster1NoDataValue;
        auxThreadParams.m_forceRaster2NoDataValue = m_forceRaster2NoDataValue;
        auxThreadParams.m_pixelOffsets1Ptr = &m_pixelOffsets1;
        auxThreadParams.m_pixelScales1Ptr = &m_pixelScales1;
        auxThreadParams.m_pixelOffsets2Ptr = &m_pixelOffsets2;
        auxThreadParams.m_pixelScales2Ptr = &m_pixelScales2;
        auxThreadParams.m_maxRasterCachedBlocks = 
          m_enableRasterCache
          ?
            (unsigned int) std::max(
              (double)( 2 * m_raster1Bands.size() )
              , 
              (
                maxVMem2Use /
                (
                  (double)
                  ( 
                    std::max(
                      m_raster1Ptr->getBand( m_raster1Bands[ 0 ] )->getBlockSize(),
                      m_raster2Ptr->getBand( m_raster2Bands[ 0 ] )->getBlockSize()
                    )
                    *
                    m_raster1Bands.size()
                    *
                    m_threadsNumber
                  )
                )
              )
            )
          : ( 2 * m_raster1Bands.size() );        
        auxThreadParams.m_rulePtr = m_rulePtr.get();
        auxThreadParams.m_processedBlocksNmbPtr = &processedBlocksNmb;

        // creating threads
        
        if( m_threadsNumber == 1 )
        {
          runningThreadsCounter = 1;
          
          auxThreadParams.m_useProgress = m_enableProgressInterface;
          
            composeIntoRaster3Thread( &auxThreadParams );
        }
        else
        {
          std::vector< std::unique_ptr< std::thread > > threads;
          runningThreadsCounter = m_threadsNumber;
          
          for( unsigned int threadIdx = 0 ; threadIdx < m_threadsNumber ;
            ++threadIdx )
          {
            auxThreadParams.m_useProgress = false;
            
            threads.push_back( std::unique_ptr< std::thread >( 
              new std::thread( composeIntoRaster3Thread, 
               &auxThreadParams ) ) );
          };    
          
          // progress stuff
          
          std::unique_ptr< te::common::TaskProgress > progressPtr;
          if( m_enableProgressInterface )
          {
            std::unique_lock<std::mutex> locker( mutex );
            
            progressPtr.reset( new te::common::TaskProgress );
            progressPtr->setTotalSteps(static_cast<int>(raster3BlocksInfos.size()));
            progressPtr->setMessage( "Composing" );
            
            locker.unlock();
          
            while( true )
            {
              locker.lock();
              
              if( abortValue )
              {
                locker.unlock();
                break;
              }
              
              if( runningThreadsCounter == 0 )
              {
                locker.unlock();
                break;
              }
              
              if( progressPtr->isActive() )
              {
                if( ((int)processedBlocksNmb) > progressPtr->getCurrentStep() )
                {
                  progressPtr->pulse();
                }
                
                locker.unlock();
                
                std::unique_lock<std::mutex> blockProcessedSignalLocker( 
                  blockProcessedSignalMutex );
                blockProcessedSignal.wait_for( blockProcessedSignalLocker, 
                  std::chrono::seconds( 1 ) );
              }
              else
              {
                abortValue = true;
                
                locker.unlock();
                
                break;
              }
            }
          }
          
          // Joining threads

          for( unsigned int threadIdx = 0 ; threadIdx < m_threadsNumber ;
            ++threadIdx )
          {                    
            threads[ threadIdx ]->join();
          }
        }
      }
      
      return returnValue;
    }
    
    void Composer::composeIntoRaster3Thread( ComposeIntoRaster3ThreadParams* paramsPtr )
    {
      // locker
      
      std::unique_lock<std::mutex> locker(*paramsPtr->m_mutexPtr);      
      
      // local parameters copy
      
      const std::vector< unsigned int > raster1Bands = 
        *paramsPtr->m_raster1BandsPtr;
      const std::vector< unsigned int > raster2Bands = 
        *paramsPtr->m_raster2BandsPtr;              
      const std::vector< unsigned int > raster3Bands = 
        *paramsPtr->m_raster3BandsPtr;
      
      te::rst::SynchronizedRaster raster1( paramsPtr->m_maxRasterCachedBlocks / 
        2, *( paramsPtr->m_sync1Ptr ) );
      te::rst::SynchronizedRaster raster2( paramsPtr->m_maxRasterCachedBlocks /
        2, *( paramsPtr->m_sync2Ptr ) );
      te::rst::SynchronizedRaster raster3( raster3Bands.size(),
        *( paramsPtr->m_sync3Ptr ) );
      
      const unsigned int raster3BlocksInfosSize = static_cast<unsigned int>(
        paramsPtr->m_raster3BlocksInfosPtr->size());
      
      const std::vector< double > pixelOffsets1 =*paramsPtr->m_pixelOffsets1Ptr;
      const std::vector< double > pixelScales1 =*paramsPtr->m_pixelScales1Ptr;
      const std::vector< double > pixelOffsets2 = *paramsPtr->m_pixelOffsets2Ptr;
      const std::vector< double > pixelScales2 = *paramsPtr->m_pixelScales2Ptr;
      
      const te::rst::Interpolator::Method interpMethod1 =
        paramsPtr->m_interpMethod1;
      const te::rst::Interpolator::Method interpMethod2 =
        paramsPtr->m_interpMethod2;

      const double noDataValue = paramsPtr->m_noDataValue;
      
      const bool forceRaster1NoDataValue = paramsPtr->m_forceRaster1NoDataValue;
      const bool forceRaster2NoDataValue = paramsPtr->m_forceRaster2NoDataValue;
      
      std::unique_ptr< te::common::TaskProgress > progressPtr;
      if( paramsPtr->m_useProgress )
      {
        progressPtr.reset( new te::common::TaskProgress );
        progressPtr->setTotalSteps( raster3BlocksInfosSize );
        progressPtr->setMessage( "Composing" );
      }
      
      std::unique_ptr< ComposerRule > rulePtr( paramsPtr->m_rulePtr->clone() );
      
      locker.unlock();
      
      // local blender
      
      Composer composerInstance;      
      if( ! composerInstance.initialize( 
        raster1,
        raster1Bands,
        raster2,
        raster2Bands,
        raster3,
        raster3Bands,        
        interpMethod1,
        interpMethod2,
        noDataValue,
        forceRaster1NoDataValue,
        forceRaster2NoDataValue,
        pixelOffsets1,
        pixelScales1,
        pixelOffsets2,
        pixelScales2,
        1,
        false,
        rulePtr.get(),
        false ) )
      {
        locker.lock();
        
        *(paramsPtr->m_abortValuePtr) = true;
        *(paramsPtr->m_returnValuePtr) = false;
        --( *(paramsPtr->m_runningThreadsCounterPtr) );
        
        return;
      }            
      
      // Guessing the output raster channels ranges
      
      const unsigned int raster3BandsSize = static_cast<unsigned int>(
        raster3Bands.size());
      
      std::vector< double > raster3BandsRangeMin( raster3BandsSize, 0 );
      std::vector< double > raster3BandsRangeMax( raster3BandsSize, 0 ); 
      
      {
        for( unsigned int raster3BandsIdx = 0 ;  raster3BandsIdx < 
          raster3BandsSize ; ++raster3BandsIdx )
        {
          unsigned int bandIdx = raster1Bands[ raster3BandsIdx ];
          
          te::rst::GetDataTypeRanges( raster1.getBand( bandIdx )->getProperty()->m_type,
            raster3BandsRangeMin[ raster3BandsIdx ],
            raster3BandsRangeMax[ raster3BandsIdx ]);             
        }
      }      
      
      // loocking for the next raster block to process
      
      boost::scoped_array< double > composedValuesHandler( 
        new double[ raster3BandsSize ] );
      double* composedValuesHandlerPtr = composedValuesHandler.get();
      
      for( unsigned int raster3BlocksInfosIdx = 0 ; raster3BlocksInfosIdx <
            raster3BlocksInfosSize ; ++raster3BlocksInfosIdx )
      {
        locker.lock();
                
        if( paramsPtr->m_raster3BlocksInfosPtr->operator[]( raster3BlocksInfosIdx ).m_wasProcessed )
        {
          locker.unlock();
        }
        else
        {
          RasterBlockInfo& rBInfo = paramsPtr->m_raster3BlocksInfosPtr->operator[]( 
            raster3BlocksInfosIdx );
          rBInfo.m_wasProcessed = true;
          
          locker.unlock();
          
          // processing block data          
          
          unsigned int raster3Row = 0;
          unsigned int raster3Col = 0;
          unsigned int raster3BandsIdx = 0;
            
          for( raster3Row = rBInfo.m_firstRasterRow2Process ; raster3Row < rBInfo.m_rasterRows2ProcessBound ;
            ++raster3Row )
          {
            for( raster3Col = rBInfo.m_firstRasterCol2Process ; raster3Col < rBInfo.m_rasterCols2ProcessBound ;
              ++raster3Col )
            {
              composerInstance.getComposedValues( (double)raster3Row, 
                (double)raster3Col, composedValuesHandlerPtr );  
              
              for( raster3BandsIdx = 0 ; raster3BandsIdx < raster3BandsSize ; ++raster3BandsIdx )
              {          
                double& composedValue = composedValuesHandlerPtr[ raster3BandsIdx ];
                
                composedValue = std::max( composedValue , 
                  raster3BandsRangeMin[ raster3BandsIdx ] );
                composedValue = std::min( composedValue , 
                  raster3BandsRangeMax[ raster3BandsIdx ] );

                raster3.setValue( raster3Col, raster3Row, composedValue,
                  raster3Bands[ raster3BandsIdx ] );
              }
            }
          }
          
          locker.lock();
          
          ++( *(paramsPtr->m_processedBlocksNmbPtr) );
          
          locker.unlock();
          
          // progress stuff
          
          if( progressPtr.get() )
          {
            if( progressPtr->isActive() )
            {
              progressPtr->pulse();
            }
            else
            {
              locker.lock();
              
              *(paramsPtr->m_abortValuePtr) = true;
              *(paramsPtr->m_returnValuePtr) = false;
              --( *(paramsPtr->m_runningThreadsCounterPtr) );
              
              return;
            }
          }
          else
          {
            // notifying the main thread with the block processed signal
            
            paramsPtr->m_blockProcessedSignalPtr->notify_all();
          }
        }
        
        // do we must continue processing ?
        
        locker.lock();
        
        if( *(paramsPtr->m_abortValuePtr) )
        {
          locker.unlock();
          break;
        }
        else
        {
          locker.unlock();
        }
      }
      
      locker.lock();
      
      --( *(paramsPtr->m_runningThreadsCounterPtr) );
      
      locker.unlock();
    }
    
    const std::string& Composer::getErrorMessage() const
    {
      return m_errorMessage;
    }
    
    void Composer::setErrorMessage( const std::string& newErrorMessage )
    {
      m_errorMessage = newErrorMessage;
    }
    
  } // end namespace rp
}   // end namespace te    

