/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

#include "ComposerRule.h"
#include "Macros.h"

namespace te
{
  namespace rp
  {
    ComposerRule::ComposerRule()
    {
      initState();
    }
    
    ComposerRule::ComposerRule( const ComposerRule& other )
    {
      initState();
      operator=( other );
    }
    
    ComposerRule::~ComposerRule()
    {
      clear();
    }
     
    const ComposerRule& ComposerRule::operator=( const ComposerRule& other )
    {
      clear();
      
      m_forceRaster1NoDataValue = other.m_forceRaster1NoDataValue;
      m_forceRaster2NoDataValue = other.m_forceRaster2NoDataValue;
      
      m_raster1Ptr = other.m_raster1Ptr;
      m_raster2Ptr = other.m_raster2Ptr;
      m_raster3Ptr = other.m_raster3Ptr;
      
      m_interpMethod1 = other.m_interpMethod1;
      m_interpMethod2 = other.m_interpMethod2;
      
      m_outputNoDataValue = other.m_outputNoDataValue;
      
      m_raster1Bands = other.m_raster1Bands;
      m_raster2Bands = other.m_raster2Bands;
      m_raster3Bands = other.m_raster3Bands;
      
      m_pixelOffsets1 = other.m_pixelOffsets1;
      m_pixelScales1 = other.m_pixelScales1;      

      m_pixelOffsets2 = other.m_pixelOffsets2;
      m_pixelScales2 = other.m_pixelScales2;
      
      return *this;
    }
    
    bool ComposerRule::initialize( 
      const te::rst::Raster& raster1, 
      const std::vector< unsigned int >& raster1Bands, 
      const te::rst::Raster& raster2, 
      const std::vector< unsigned int >& raster2Bands,
      te::rst::Raster& raster3, 
      const std::vector< unsigned int >& raster3Bands,      
      const te::rst::Interpolator::Method& interpMethod1,
      const te::rst::Interpolator::Method& interpMethod2,
      const double& noDataValue,
      const bool forceRaster1NoDataValue,
      const bool forceRaster2NoDataValue,
      const std::vector< double >& pixelOffsets1,
      const std::vector< double >& pixelScales1,
      const std::vector< double >& pixelOffsets2,
      const std::vector< double >& pixelScales2 )
    {
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( 
        raster1.getAccessPolicy() & te::common::RAccess, 
        "Invalid raster 1" );        
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( 
        raster2.getAccessPolicy() & te::common::RAccess, 
        "Invalid raster 2" ); 
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( raster1Bands.size() > 0,
        "Invalid raster bands vector" );      
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( raster1Bands.size() ==
        raster2Bands.size(), "Invalid raster bands vector" );
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( raster1Bands.size() ==
        raster3Bands.size(), "Invalid raster bands vector" );      
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( pixelOffsets1.size() ==  
        raster1Bands.size(), "Invalid pixel offsets" );
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( pixelScales1.size() ==  
        raster1Bands.size(), "Invalid pixel scales" );        
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( pixelOffsets2.size() ==  
        raster2Bands.size(), "Invalid pixel offsets" );
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( pixelScales2.size() ==  
        raster2Bands.size(), "Invalid pixel scales" );        
        
      clear();
      
      m_forceRaster1NoDataValue = forceRaster1NoDataValue;
      m_forceRaster2NoDataValue = forceRaster2NoDataValue;
      
      m_raster1Ptr = &raster1;
      m_raster2Ptr = &raster2;
      m_raster3Ptr = &raster3;

      for( std::size_t raster1BandsIdx = 0 ; raster1BandsIdx < 
        raster1Bands.size() ; ++raster1BandsIdx )
      {
        TERP_INSTANCE_TRUE_OR_RETURN_FALSE( raster1Bands[ raster1BandsIdx ] <
          raster1.getNumberOfBands(), "Invalid raster 1 band" );        
      }
      
      for( std::size_t raster2BandsIdx = 0 ; raster2BandsIdx < 
        raster2Bands.size() ; ++raster2BandsIdx )
      {
        TERP_INSTANCE_TRUE_OR_RETURN_FALSE( raster2Bands[ raster2BandsIdx ] <
          raster2.getNumberOfBands(), "Invalid raster 2 band" );                
      }       
      
      for( std::size_t raster3BandsIdx = 0 ; raster3BandsIdx < 
        raster3Bands.size() ; ++raster3BandsIdx )
      {
        TERP_INSTANCE_TRUE_OR_RETURN_FALSE( raster3Bands[ raster3BandsIdx ] <
          raster3.getNumberOfBands(), "Invalid raster 3 band" );                
      }
      
      m_raster1Bands = raster1Bands;
      m_raster2Bands = raster2Bands;
      m_raster3Bands = raster3Bands;
      
      m_interpMethod1 = interpMethod1;
      m_interpMethod2 = interpMethod2;
      
      m_outputNoDataValue = noDataValue;
      
      m_pixelOffsets1 = pixelOffsets1;
      m_pixelScales1 = pixelScales1;
      
      m_pixelOffsets2 = pixelOffsets2;
      m_pixelScales2 = pixelScales2;      
      
      return true;
    }
    
    void ComposerRule::initState()
    {
      m_forceRaster1NoDataValue = false;
      m_forceRaster2NoDataValue = false;
      m_raster1Ptr = nullptr;
      m_raster2Ptr = nullptr;
      m_raster3Ptr = nullptr;
      m_interpMethod1 = te::rst::NearestNeighbor;
      m_interpMethod2 = te::rst::NearestNeighbor;
      m_outputNoDataValue = 0;
    }
    
    void ComposerRule::clear()
    {
      m_raster1Bands.clear();
      m_raster2Bands.clear();
      m_raster3Bands.clear();
      m_pixelOffsets1.clear();
      m_pixelScales1.clear();
      m_pixelOffsets2.clear();
      m_pixelScales2.clear();
      m_errorMessage.clear();

      initState();
    }
    
    const std::string& ComposerRule::getErrorMessage() const
    {
      return m_errorMessage;
    }    
    
    void ComposerRule::setErrorMessage( const std::string& newErrorMessage )
    {
      m_errorMessage = newErrorMessage;
    }    
    
  } // end namespace rp
}   // end namespace te    

