/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

#ifndef __TERRALIB_RP_INTERNAL_COMPOSER_RULE_H
#define __TERRALIB_RP_INTERNAL_COMPOSER_RULE_H

#include "Config.h"
#include "../raster/Raster.h"
#include "../raster/Interpolator.h"
#include "../geometry/MultiPolygon.h"
#include "../geometry/GeometricTransformation.h"

#include <memory>
#include <vector>

namespace te
{
  namespace rp
  {
    class Composer;
    
    /*!
      \class Composer Rule
      \brief A rule used by the te::rst::Composer classe.
      \ingroup rp_optical
     */
    class TERPEXPORT ComposerRule
    {
      public:

        virtual ~ComposerRule();
      
        /*!
          \brief Return the current error message if there is any.

          \return Return the current error message if there is any.
         */
        const std::string& getErrorMessage() const;              

      protected:
        
        friend class Composer;
        
        bool m_forceRaster1NoDataValue; //!<  Use noDataValue as the input raster 1 no-data value (The original rasters no-data values will be ignored)
        bool m_forceRaster2NoDataValue; //!<  Use noDataValue as the input raster 2 no-data value (The original rasters no-data values will be ignored)  
        te::rst::Raster const * m_raster1Ptr; //!< Input raster 1.
        te::rst::Raster const * m_raster2Ptr; //!< Input raster 2.
        te::rst::Raster * m_raster3Ptr; //!< Input raster 3.
        te::rst::Interpolator::Method m_interpMethod1; //!< The interpolation method to use when reading raster 1 data.
        te::rst::Interpolator::Method m_interpMethod2; //!< The interpolation method to use when reading raster 2 data.
        double m_outputNoDataValue; //!< The output raster no-data value.
        std::vector< unsigned int > m_raster1Bands; //!< Input raster 1 band indexes to use.
        std::vector< unsigned int > m_raster2Bands; //!< Input raster 2 band indexes to use.
        std::vector< unsigned int > m_raster3Bands; //!< Input raster 3 band indexes to use.
        std::vector< double > m_pixelOffsets1; //!< The values offset to be applied to raster 1 pixel values before the composed value calcule (one element for each used raster channel/band).
        std::vector< double > m_pixelScales1; //!< The values scale to be applied to raster 1 pixel values before the composed value calcule (one element for each used raster channel/band).
        std::vector< double > m_pixelOffsets2; //!< The values offset to be applied to raster 2 pixel values before the composed value calcule (one element for each used raster channel/band).
        std::vector< double > m_pixelScales2; //!< The values scale to be applied to raster 2 pixel values before the composed value calcule (one element for each used raster channel/band).
        std::string m_errorMessage; //!< Current error message.
       
        ComposerRule();
        
        ComposerRule( const ComposerRule& other );
        
        const ComposerRule& operator=( const ComposerRule& other );
        
      /*!
        \brief Inititate the instance.
        \param raster1 Raster 1.
        \param raster1Bands Raster 1 band indexes to use.
        \param raster2 Raster 2.
        \param raster2Bands Raster 2 band indexes to use (this vector has the same size as raster1Bands).
        \param raster3 Raster 3.
        \param raster3Bands Raster 3 band indexes to use (this vector has the same size as raster1Bands).        
        \param interpMethod1 The interpolation method to use when reading raster 1 data.
        \param interpMethod2 The interpolation method to use when reading raster 2 data.
        \param noDataValue The value returned where there is no pixel data.
        \param forceRaster1NoDataValue Use noDataValue as the input raster 1 no-data value (The original rasters no-data values will be ignored)
        \param forceRaster2NoDataValue Use noDataValue as the input raster 1 no-data value (The original rasters no-data values will be ignored)  
        \param pixelOffsets1 The values offset to be applied to raster 1 pixel values before the composed value calcule (one element for each used raster channel/band).
        \param pixelScales1 The values scale to be applied to raster 1 pixel values before the composed value calcule (one element for each used raster channel/band).
        \param pixelOffsets2 The values offset to be applied to raster 2 pixel values before the composed value calcule (one element for each used raster channel/band).
        \param pixelScales2 The values scale to be applied to raster 2 pixel values before the composed value calcule (one element for each used raster channel/band).
        \return true if ok, false on errors
        \note Aboute scale and offset parametrs:  outValue = ( inputValue * scale ) + offset
      */
      virtual bool initialize( 
        const te::rst::Raster& raster1, 
        const std::vector< unsigned int >& raster1Bands, 
        const te::rst::Raster& raster2, 
        const std::vector< unsigned int >& raster2Bands,
        te::rst::Raster& raster3, 
        const std::vector< unsigned int >& raster3Bands,
        const te::rst::Interpolator::Method& interpMethod1,
        const te::rst::Interpolator::Method& interpMethod2,
        const double& noDataValue,
        const bool forceRaster1NoDataValue,
        const bool forceRaster2NoDataValue,
        const std::vector< double >& pixelOffsets1,
        const std::vector< double >& pixelScales1,
        const std::vector< double >& pixelOffsets2,
        const std::vector< double >& pixelScales2 );
      
      /*!
        \brief Compose a pixel value using the current parameters.
        \param line Line (raster 3 reference).
        \param col Column (raster 3 reference).
        \param composedValuesPtr A pointer to a pre-allocated vector where the composed values will be stored.
        \note The caller of this method must be aware that the returned composed value may be outside the original input rasters valid values range.
        \note Raster 1 values have precedence over raster 2 values (when applicable).
      */        
        virtual void getComposedValues( const double& line, const double& col, 
          double* const composedValuesPtr ) const = 0;      
          
      /*!
        \brief Create a clone of this instance.
        \return A clone of this instance.
        \note The caller of this method must take othe owndershipd of the returned pointer.
      */    
        virtual ComposerRule* clone() const = 0;          
        
        /*!
          \brief Set the current error message.

          \param newErrorMessage New error message;
         */
        void setErrorMessage( const std::string& newErrorMessage );                  
        
        void initState();
        
        void clear();        
        
      private :


    };

  } // end namespace rp
}   // end namespace te

#endif

