/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

#include "ComposerRuleValidNeighborhood.h"
#include "Macros.h"
#include "../raster/Raster.h"
#include "../raster/BandProperty.h"
#include "../raster/Band.h"
#include "../raster/Grid.h"

namespace te
{
  namespace rp
  {
    ComposerRuleValidNeighborhood::ComposerRuleValidNeighborhood()
    {
      initState();
    }
    
    ComposerRuleValidNeighborhood::ComposerRuleValidNeighborhood( 
      const ComposerRuleValidNeighborhood& other )
    {
      initState();
      operator=( other );
    }
    
    ComposerRuleValidNeighborhood::~ComposerRuleValidNeighborhood()
    {
      clear();
    }
     
    const ComposerRuleValidNeighborhood& ComposerRuleValidNeighborhood::operator=( 
      const ComposerRuleValidNeighborhood& other )
    {
      ComposerRule::operator=( other );

      if( m_raster3Ptr != nullptr )
      {
        m_converter1.setSourceSRID( m_raster3Ptr->getSRID() );
        m_converter2.setSourceSRID( m_raster3Ptr->getSRID() );
      }
      
      if( m_raster1Ptr != nullptr )
      {
        m_converter1.setTargetSRID( m_raster1Ptr->getSRID() );
      }
      
      if( m_raster2Ptr != nullptr )
      {
        m_converter2.setTargetSRID( m_raster2Ptr->getSRID() );
      }      
      
      m_raster1NoDataValues = other.m_raster1NoDataValues;
      m_raster2NoDataValues = other.m_raster2NoDataValues;
      
      m_windowRadius = other.m_windowRadius;
      
      m_windowMaxNPixels = other.m_windowMaxNPixels;
      
      m_neiborMinValidPercent = other.m_neiborMinValidPercent;
      
      m_ranges1 = other.m_ranges1;
      
      m_ranges2 = other.m_ranges2;
      
      m_NBands = other.m_NBands;
      
      m_r1NRows = other.m_r1NRows;
      
      m_r1NCols = other.m_r1NCols;
      
      m_r2NRows = other.m_r2NRows;
      
      m_r2NCols = other.m_r2NCols;
      
      return *this;
    }
    
    bool ComposerRuleValidNeighborhood::initialize( 
      const te::rst::Raster& raster1, 
      const std::vector< unsigned int >& raster1Bands, 
      const te::rst::Raster& raster2, 
      const std::vector< unsigned int >& raster2Bands,
      te::rst::Raster& raster3, 
      const std::vector< unsigned int >& raster3Bands,      
      const te::rst::Interpolator::Method& interpMethod1,
      const te::rst::Interpolator::Method& interpMethod2,
      const double& noDataValue,
      const bool forceRaster1NoDataValue,
      const bool forceRaster2NoDataValue,
      const std::vector< double >& pixelOffsets1,
      const std::vector< double >& pixelScales1,
      const std::vector< double >& pixelOffsets2,
      const std::vector< double >& pixelScales2 )
    {
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( ComposerRule::initialize( raster1, 
        raster1Bands, raster2, raster2Bands, raster3,  raster3Bands,      
        interpMethod1, interpMethod2, noDataValue, forceRaster1NoDataValue,
        forceRaster2NoDataValue, pixelOffsets1, pixelScales1, pixelOffsets2,
        pixelScales2 ), ComposerRule::getErrorMessage() );
     
      m_converter1.setSourceSRID( raster3.getSRID() );
      m_converter1.setTargetSRID( raster1.getSRID() );
      
      m_converter2.setSourceSRID( raster3.getSRID() );
      m_converter2.setTargetSRID( raster2.getSRID() );
      
      m_raster1NoDataValues.clear();
      for( std::size_t raster1BandsIdx = 0 ; raster1BandsIdx < 
        raster1Bands.size() ; ++raster1BandsIdx )
      {
        if( forceRaster1NoDataValue )
        {
          m_raster1NoDataValues.push_back( noDataValue );
        }
        else
        {
          m_raster1NoDataValues.push_back( 
            raster1.getBand( raster1Bands[ raster1BandsIdx ] )->getProperty()->m_noDataValue );
        }
      }
      
      m_raster2NoDataValues.clear();
      for( std::size_t raster2BandsIdx = 0 ; raster2BandsIdx < 
        raster2Bands.size() ; ++raster2BandsIdx )
      {
        if( forceRaster2NoDataValue )
        {
          m_raster2NoDataValues.push_back( noDataValue );
        }
        else
        {
          m_raster2NoDataValues.push_back( 
            raster2.getBand( raster2Bands[ raster2BandsIdx ] )->getProperty()->m_noDataValue );
        }
      }      
      
      m_NBands = raster1Bands.size();
      
      m_r1NRows = (long int)raster1.getNumberOfRows();
      m_r1NCols = (long int)raster1.getNumberOfColumns();
      
      m_r2NRows = (long int)raster2.getNumberOfRows();
      m_r2NCols = (long int)raster2.getNumberOfColumns();
      
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( m_ranges1.size() == m_NBands, "Invalid raster 1 ranges" );
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( m_ranges2.size() == m_NBands, "Invalid raster 2 ranges" );
      
      return true;
    }
    
    void ComposerRuleValidNeighborhood::initState()
    {
      ComposerRule::initState();

      m_windowRadius = 0;
      m_windowMaxNPixels = 0;
      m_neiborMinValidPercent = 0;
      m_NBands = 0;
      m_r1NRows = 0;
      m_r1NCols = 0;
      m_r2NRows = 0;
      m_r2NCols = 0;
    }
    
    void ComposerRuleValidNeighborhood::clear()
    {
      ComposerRule::clear();
      initState();
      
      m_raster1NoDataValues.clear();
      m_raster2NoDataValues.clear();
      m_ranges1.clear();
      m_ranges2.clear();
    }
    
    void ComposerRuleValidNeighborhood::getComposedValues( const double& line, 
      const double& col, double* const composedValuesPtr ) const
    {
      m_raster3Ptr->getGrid()->gridToGeo( col, line,
        m_getComposedValues_centerX3, m_getComposedValues_centerY3 );

      // calc of raster 1 center pixel quality
      
      m_converter1.convert( m_getComposedValues_centerX3, 
        m_getComposedValues_centerY3, m_getComposedValues_centerX,
        m_getComposedValues_centerY );
      
      m_raster1Ptr->getGrid()->geoToGrid( m_getComposedValues_centerX,
        m_getComposedValues_centerY, m_getComposedValues_centerCol,
        m_getComposedValues_centerRow );
      
      m_getComposedValues_centerRow1= (long int)(m_getComposedValues_centerRow + 
        0.5);
      m_getComposedValues_centerCol1 = (long int)(m_getComposedValues_centerCol + 
        0.5);
      
      m_getComposedValues_r1PixelQuality = 0;
      
      if( 
          ( m_getComposedValues_centerRow1 >= 0 )
          &&
          ( m_getComposedValues_centerRow1 < m_r1NRows )
          &&
          ( m_getComposedValues_centerCol1 >= 0 )
          &&
          ( m_getComposedValues_centerCol1 < m_r1NCols )
        )
      {
        m_getComposedValues_pixelIsValid = true;
        
        for( m_getComposedValues_bandIdx = 0 ; m_getComposedValues_bandIdx <
          m_NBands ; ++m_getComposedValues_bandIdx )
        {
          m_raster1Ptr->getValue( m_getComposedValues_centerCol1, m_getComposedValues_centerRow1,
            m_getComposedValues_pixelValue, 
            m_raster1Bands[ m_getComposedValues_bandIdx ] );
          
          if( 
              m_getComposedValues_pixelValue 
              ==
              m_raster1NoDataValues[ m_getComposedValues_bandIdx ]
            )
          {
            m_getComposedValues_pixelIsValid = false;
            break;
          }
        }        
        
        if( m_getComposedValues_pixelIsValid )
        {
          m_getComposedValues_startRow = m_getComposedValues_centerRow1;
          m_getComposedValues_startRow = ( m_getComposedValues_startRow > m_windowRadius ) ?
            ( m_getComposedValues_startRow - m_windowRadius ) : 0;      
          m_getComposedValues_rowBound = m_getComposedValues_centerRow1 + m_windowRadius + 1;
          m_getComposedValues_rowBound = std::min( m_r1NRows, m_getComposedValues_rowBound );
          
          m_getComposedValues_startCol = m_getComposedValues_centerCol1;
          m_getComposedValues_startCol = ( m_getComposedValues_startCol > m_windowRadius ) ?
            ( m_getComposedValues_startCol - m_windowRadius ) : 0;      
          m_getComposedValues_colBound = m_getComposedValues_centerCol1 + m_windowRadius + 1;
          m_getComposedValues_colBound = std::min( m_r1NCols, m_getComposedValues_colBound );
          
          for( m_getComposedValues_row = m_getComposedValues_startRow ; m_getComposedValues_row < 
             m_getComposedValues_rowBound ; ++m_getComposedValues_row )
          {
            for( m_getComposedValues_col  = m_getComposedValues_startCol ; m_getComposedValues_col < 
               m_getComposedValues_colBound ; ++m_getComposedValues_col )
            {
              m_getComposedValues_pixelIsValid = true;
              
              for( m_getComposedValues_bandIdx = 0 ; m_getComposedValues_bandIdx <
                m_NBands ; ++m_getComposedValues_bandIdx )
              {
                m_raster1Ptr->getValue( m_getComposedValues_col, m_getComposedValues_row,
                  m_getComposedValues_pixelValue, 
                  m_raster1Bands[ m_getComposedValues_bandIdx ] );
                
                if( 
                    m_getComposedValues_pixelValue 
                    ==
                    m_raster1NoDataValues[ m_getComposedValues_bandIdx ]
                  )
                {
                  m_getComposedValues_pixelIsValid = false;
                }
                else
                {
                  if( 
                      (
                        m_getComposedValues_pixelValue 
                        <
                        m_ranges1[ m_getComposedValues_bandIdx ].first
                      )
                      ||
                      (
                        m_getComposedValues_pixelValue 
                        >
                        m_ranges1[ m_getComposedValues_bandIdx ].second
                      )                  
                    )
                  {
                    m_getComposedValues_pixelIsValid = false;
                  }
                }
              }
              
              if( m_getComposedValues_pixelIsValid )
              {
                ++m_getComposedValues_r1PixelQuality;
              }
            }
          }
        }
      }

      // calc of raster 2 center pixel quality
      
      m_getComposedValues_r2PixelQuality = 0;
      
      m_converter2.convert( m_getComposedValues_centerX3, 
        m_getComposedValues_centerY3, m_getComposedValues_centerX,
        m_getComposedValues_centerY );
      
      m_raster2Ptr->getGrid()->geoToGrid( m_getComposedValues_centerX,
        m_getComposedValues_centerY, m_getComposedValues_centerCol,
        m_getComposedValues_centerRow );
      
      m_getComposedValues_centerRow2= (long int)(m_getComposedValues_centerRow + 
        0.5);
      m_getComposedValues_centerCol2 = (long int)(m_getComposedValues_centerCol + 
        0.5);      
      
      if( 
          ( m_getComposedValues_centerRow2 >= 0 )
          &&
          ( m_getComposedValues_centerRow2 < m_r2NRows )
          &&
          ( m_getComposedValues_centerCol2 >= 0 )
          &&
          ( m_getComposedValues_centerCol2 < m_r2NCols )
        )
      {
        m_getComposedValues_pixelIsValid = true;
        
        for( m_getComposedValues_bandIdx = 0 ; m_getComposedValues_bandIdx <
          m_NBands ; ++m_getComposedValues_bandIdx )
        {
          m_raster2Ptr->getValue( m_getComposedValues_centerCol2, m_getComposedValues_centerRow2,
            m_getComposedValues_pixelValue, 
            m_raster2Bands[ m_getComposedValues_bandIdx ] );
          
          if( 
              m_getComposedValues_pixelValue 
              ==
              m_raster2NoDataValues[ m_getComposedValues_bandIdx ]
            )
          {
            m_getComposedValues_pixelIsValid = false;
            break;
          }
        }                
        
        if( m_getComposedValues_pixelIsValid )
        {
          m_getComposedValues_startRow = m_getComposedValues_centerRow2;
          m_getComposedValues_startRow = ( m_getComposedValues_startRow > m_windowRadius ) ?
            ( m_getComposedValues_startRow - m_windowRadius ) : 0;      
          m_getComposedValues_rowBound = m_getComposedValues_centerRow2 + m_windowRadius + 1;
          m_getComposedValues_rowBound = std::min( m_r2NRows, m_getComposedValues_rowBound );
          
          m_getComposedValues_startCol = m_getComposedValues_centerCol2;
          m_getComposedValues_startCol = ( m_getComposedValues_startCol > m_windowRadius ) ?
            ( m_getComposedValues_startCol - m_windowRadius ) : 0;      
          m_getComposedValues_colBound = m_getComposedValues_centerCol2 + m_windowRadius + 1;
          m_getComposedValues_colBound = std::min( m_r2NCols, m_getComposedValues_colBound );
          
          for( m_getComposedValues_row = m_getComposedValues_startRow ; m_getComposedValues_row < 
            m_getComposedValues_rowBound ; ++m_getComposedValues_row )
          {
            for( m_getComposedValues_col  = m_getComposedValues_startCol ; m_getComposedValues_col < 
              m_getComposedValues_colBound ; ++m_getComposedValues_col )
            {
              m_getComposedValues_pixelIsValid = true;
              
              for( m_getComposedValues_bandIdx = 0 ; m_getComposedValues_bandIdx <
                m_NBands ; ++m_getComposedValues_bandIdx )
              {
                m_raster2Ptr->getValue( m_getComposedValues_col, m_getComposedValues_row,
                  m_getComposedValues_pixelValue, 
                  m_raster2Bands[ m_getComposedValues_bandIdx ] );
                
                if( 
                    m_getComposedValues_pixelValue 
                    ==
                    m_raster2NoDataValues[ m_getComposedValues_bandIdx ]
                  )
                {
                  m_getComposedValues_pixelIsValid = false;
                }
                else
                {
                  if( 
                      (
                        m_getComposedValues_pixelValue 
                        <
                        m_ranges2[ m_getComposedValues_bandIdx ].first
                      )
                      ||
                      (
                        m_getComposedValues_pixelValue 
                        >
                        m_ranges2[ m_getComposedValues_bandIdx ].second
                      )                  
                    )
                  {
                    m_getComposedValues_pixelIsValid = false;
                  }
                }
              }
              
              if( m_getComposedValues_pixelIsValid )
              {
                ++m_getComposedValues_r2PixelQuality;
              }
            }
          }
        }
      } 
      
      // comparing pixels quality
      
      if( 
          ( m_getComposedValues_r2PixelQuality > m_getComposedValues_r1PixelQuality )
          &&
          (
            ( m_windowRadius == 0 )
            ||
            (
              (
                ((double)m_getComposedValues_r2PixelQuality)
                /
                m_windowMaxNPixels
              )
              >=
              m_neiborMinValidPercent
            )
          )
        )
      {
        for( m_getComposedValues_bandIdx = 0 ; m_getComposedValues_bandIdx <
          m_NBands ; ++m_getComposedValues_bandIdx )
        {
          m_raster2Ptr->getValue( m_getComposedValues_centerCol2, m_getComposedValues_centerRow2,
            m_getComposedValues_pixelValue, m_raster2Bands[ m_getComposedValues_bandIdx ] );
        
          m_getComposedValues_pixelValue *= m_pixelScales2[ m_getComposedValues_bandIdx ];
          m_getComposedValues_pixelValue += m_pixelOffsets2[ m_getComposedValues_bandIdx ];
          
          composedValuesPtr[ m_getComposedValues_bandIdx ] = 
                m_getComposedValues_pixelValue;
        }
      }
      else if( 
                ( m_getComposedValues_r1PixelQuality != 0 )
                &&
                (
                  ( m_windowRadius == 0 )
                  ||
                  (
                    (
                      ((double)m_getComposedValues_r1PixelQuality)
                      /
                      m_windowMaxNPixels
                    )
                    >=
                    m_neiborMinValidPercent
                  )
                ) 
              )
      {
        for( m_getComposedValues_bandIdx = 0 ; m_getComposedValues_bandIdx <
          m_NBands ; ++m_getComposedValues_bandIdx )
        {
          m_raster1Ptr->getValue( m_getComposedValues_centerCol1, m_getComposedValues_centerRow1,
            m_getComposedValues_pixelValue, m_raster1Bands[ m_getComposedValues_bandIdx ] );
          
          m_getComposedValues_pixelValue *= m_pixelScales1[ m_getComposedValues_bandIdx ];
          m_getComposedValues_pixelValue += m_pixelOffsets1[ m_getComposedValues_bandIdx ];
          
          composedValuesPtr[ m_getComposedValues_bandIdx ] = 
                m_getComposedValues_pixelValue;          
        }
      }
      else
      {
        for( m_getComposedValues_bandIdx = 0 ; m_getComposedValues_bandIdx <
          m_NBands ; ++m_getComposedValues_bandIdx )
        {
          composedValuesPtr[ m_getComposedValues_bandIdx ] = m_outputNoDataValue;
        }
      }
    }
    
    ComposerRule* ComposerRuleValidNeighborhood::clone() const
    {
      return new ComposerRuleValidNeighborhood( *this );
    }
    
    void ComposerRuleValidNeighborhood::setWindowRadius( const unsigned int newRadius )
    {
      m_windowRadius = (long int)newRadius;
      m_windowMaxNPixels = 1 + newRadius + newRadius;
      m_windowMaxNPixels = m_windowMaxNPixels * m_windowMaxNPixels;
    }
    
    unsigned int ComposerRuleValidNeighborhood::getWindowRadius() const
    {
      return m_windowRadius;
    }    
    
    void ComposerRuleValidNeighborhood::setNeiborMinValidPercent( const double newPercent )
    {
      TERP_TRUE_OR_THROW( ( newPercent >= 0 ) && ( newPercent <= 100 ), "Invalid percent value" );
      m_neiborMinValidPercent = ( newPercent / 100.0 );
    }
    
    double ComposerRuleValidNeighborhood::getNeiborMinValidPercent() const
    {
      return ( m_neiborMinValidPercent * 100.0 );
    }
    
    void ComposerRuleValidNeighborhood::setValidValuesRanges( 
      const std::vector< std::pair< double, double > >& ranges1,
      const std::vector< std::pair< double, double > >& ranges2 )
    {
      m_ranges1 = ranges1;
      m_ranges2 = ranges2;
    }    
    
    void ComposerRuleValidNeighborhood::getValidValuesRanges( 
      std::vector< std::pair< double, double > >& ranges1,
      std::vector< std::pair< double, double > >& ranges2 ) const
    {
      ranges1 = m_ranges1;
      ranges2 = m_ranges2;
    }        
    
  } // end namespace rp
}   // end namespace te    

