/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

#include "ComposerRuleValidPixel.h"
#include "Macros.h"
#include "../raster/Raster.h"
#include "../raster/BandProperty.h"
#include "../raster/Band.h"
#include "../raster/Grid.h"

namespace te
{
  namespace rp
  {
    ComposerRuleValidPixel::ComposerRuleValidPixel()
    {
      initState();
    }
    
    ComposerRuleValidPixel::ComposerRuleValidPixel( 
      const ComposerRuleValidPixel& other )
    {
      initState();
      operator=( other );
    }
    
    ComposerRuleValidPixel::~ComposerRuleValidPixel()
    {
      clear();
    }
     
    const ComposerRuleValidPixel& ComposerRuleValidPixel::operator=( 
      const ComposerRuleValidPixel& other )
    {
      ComposerRule::operator=( other );
      
      m_raster1BandsSize = other.m_raster1BandsSize;

      if( m_raster3Ptr != nullptr )
      {
        m_converter1.setSourceSRID( m_raster3Ptr->getSRID() );
        m_converter2.setSourceSRID( m_raster3Ptr->getSRID() );
      }
      
      if( m_raster1Ptr != nullptr )
      {
        m_converter1.setTargetSRID( m_raster1Ptr->getSRID() );
      }
      
      if( m_raster2Ptr != nullptr )
      {
        m_converter2.setTargetSRID( m_raster2Ptr->getSRID() );
      }      
      
      m_interpolator1Ptr.reset();
      if( other.m_interpolator1Ptr.get() )
      {
        m_interpolator1Ptr.reset( other.m_interpolator1Ptr->clone() );
      }
      
      m_interpolator2Ptr.reset();
      if( other.m_interpolator2Ptr.get() )
      {
        m_interpolator2Ptr.reset( other.m_interpolator2Ptr->clone() );
      }      
      
      m_noDataValues1 = other.m_noDataValues1;
      m_noDataValues2 = other.m_noDataValues2;
      
      return *this;
    }
    
    bool ComposerRuleValidPixel::initialize( 
      const te::rst::Raster& raster1, 
      const std::vector< unsigned int >& raster1Bands, 
      const te::rst::Raster& raster2, 
      const std::vector< unsigned int >& raster2Bands,
      te::rst::Raster& raster3, 
      const std::vector< unsigned int >& raster3Bands,      
      const te::rst::Interpolator::Method& interpMethod1,
      const te::rst::Interpolator::Method& interpMethod2,
      const double& noDataValue,
      const bool forceRaster1NoDataValue,
      const bool forceRaster2NoDataValue,
      const std::vector< double >& pixelOffsets1,
      const std::vector< double >& pixelScales1,
      const std::vector< double >& pixelOffsets2,
      const std::vector< double >& pixelScales2 )
    {
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( ComposerRule::initialize( raster1, 
        raster1Bands, raster2, raster2Bands, raster3,  raster3Bands,      
        interpMethod1, interpMethod2, noDataValue, forceRaster1NoDataValue,
        forceRaster2NoDataValue, pixelOffsets1, pixelScales1, pixelOffsets2,
        pixelScales2 ), ComposerRule::getErrorMessage() );
      
      m_raster1BandsSize = m_raster1Bands.size();
     
      m_converter1.setSourceSRID( raster3.getSRID() );
      m_converter1.setTargetSRID( raster1.getSRID() );
      
      m_converter2.setSourceSRID( raster3.getSRID() );
      m_converter2.setTargetSRID( raster2.getSRID() );
      
      m_noDataValues1.clear();
      for( std::size_t raster1BandsIdx = 0 ; raster1BandsIdx < 
        raster1Bands.size() ; ++raster1BandsIdx )
      {
        if( forceRaster1NoDataValue )
        {
          m_noDataValues1.push_back( noDataValue );
        }
        else
        {
          m_noDataValues1.push_back( 
            raster1.getBand( raster1Bands[ raster1BandsIdx ] )->getProperty()->m_noDataValue );
        }
      }
      
      m_noDataValues2.clear();
      for( std::size_t raster2BandsIdx = 0 ; raster2BandsIdx < 
        raster2Bands.size() ; ++raster2BandsIdx )
      {
        if( forceRaster2NoDataValue )
        {
          m_noDataValues2.push_back( noDataValue );
        }
        else
        {
          m_noDataValues2.push_back( 
            raster2.getBand( raster2Bands[ raster2BandsIdx ] )->getProperty()->m_noDataValue );
        }
      }       
      
      m_interpolator1Ptr.reset( new te::rst::Interpolator( &raster1, 
        interpMethod1, m_noDataValues1 ) );
      m_interpolator2Ptr.reset( new te::rst::Interpolator( &raster2, 
        interpMethod2, m_noDataValues2 ) );
      
      m_raster1BandsSize = m_raster1Bands.size();
      
      return true;
    }
    
    void ComposerRuleValidPixel::initState()
    {
      ComposerRule::initState();
      
      m_raster1BandsSize = 0;
    }
    
    void ComposerRuleValidPixel::clear()
    {
      ComposerRule::clear();
      initState();
      
      m_interpolator1Ptr.reset();
      m_interpolator2Ptr.reset();
      m_noDataValues1.clear();
      m_noDataValues2.clear();
    }
    
    void ComposerRuleValidPixel::getComposedValues( const double& line, 
      const double& col, double* const composedValuesPtr ) const
    {
      m_raster3Ptr->getGrid()->gridToGeo( col, line,
        m_getComposedValues_raster3X, m_getComposedValues_raster3Y );      
      
      m_getComposedValues_pixelIsValid = true;      

      m_converter1.convert( m_getComposedValues_raster3X, 
        m_getComposedValues_raster3Y, m_getComposedValues_rasterX,
        m_getComposedValues_rasterY );
      
      m_raster1Ptr->getGrid()->geoToGrid( m_getComposedValues_rasterX,
        m_getComposedValues_rasterY, m_getComposedValues_rasterCol,
        m_getComposedValues_rasterRow );         
      
      for( m_getComposedValues_rasterBandsIdx = 0 ;
            m_getComposedValues_rasterBandsIdx < m_raster1BandsSize ;
        ++m_getComposedValues_rasterBandsIdx )
      {
        m_interpolator1Ptr->getValue( m_getComposedValues_rasterCol,
          m_getComposedValues_rasterRow, m_getComposedValues_PixelValue,
          m_raster1Bands[ m_getComposedValues_rasterBandsIdx ] );
        
        if( 
            m_getComposedValues_PixelValue
            ==
            m_noDataValues1[ m_getComposedValues_rasterBandsIdx ] 
           )
        {
          m_getComposedValues_pixelIsValid = false;
          break;
        }
        
        composedValuesPtr[ m_getComposedValues_rasterBandsIdx ] =
          (
            m_getComposedValues_PixelValue.real()
            * 
            m_pixelScales1[ m_getComposedValues_rasterBandsIdx ]
          )
          +
          m_pixelOffsets1[ m_getComposedValues_rasterBandsIdx ];
      }
      
      if( m_getComposedValues_pixelIsValid ) return;
      
      m_getComposedValues_pixelIsValid = true;
      
      m_converter2.convert( m_getComposedValues_raster3X, 
        m_getComposedValues_raster3Y, m_getComposedValues_rasterX,
        m_getComposedValues_rasterY );
      
      m_raster2Ptr->getGrid()->geoToGrid( m_getComposedValues_rasterX,
        m_getComposedValues_rasterY, m_getComposedValues_rasterCol,
        m_getComposedValues_rasterRow );         
      
      for( m_getComposedValues_rasterBandsIdx = 0 ;
           m_getComposedValues_rasterBandsIdx < m_raster1BandsSize ;
           ++m_getComposedValues_rasterBandsIdx )
      {
        m_interpolator2Ptr->getValue( m_getComposedValues_rasterCol,
          m_getComposedValues_rasterRow, m_getComposedValues_PixelValue,
          m_raster2Bands[ m_getComposedValues_rasterBandsIdx ] );
        
        if( 
            m_getComposedValues_PixelValue
            ==
            m_noDataValues2[ m_getComposedValues_rasterBandsIdx ] 
           )
        {
          m_getComposedValues_pixelIsValid = false;
          break;
        }
          
        composedValuesPtr[ m_getComposedValues_rasterBandsIdx ] =
          (
            m_getComposedValues_PixelValue.real()
            * 
            m_pixelScales2[ m_getComposedValues_rasterBandsIdx ]
          )
          +
          m_pixelOffsets2[ m_getComposedValues_rasterBandsIdx ];
      }
      
      if( m_getComposedValues_pixelIsValid ) return;
      
      for( m_getComposedValues_rasterBandsIdx = 0 ;
           m_getComposedValues_rasterBandsIdx < m_raster1BandsSize ;
           ++m_getComposedValues_rasterBandsIdx )
      {
        composedValuesPtr[ m_getComposedValues_rasterBandsIdx ] =
          m_outputNoDataValue;
      }      
    }
    
    ComposerRule* ComposerRuleValidPixel::clone() const
    {
      return new ComposerRuleValidPixel( *this );
    }
    
  } // end namespace rp
}   // end namespace te    

