/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

#include "Composition.h"

#include "Macros.h"
#include "Functions.h"
#include "AlgorithmParametersSerializer.h"
#include "../raster/Interpolator.h"
#include "../raster/Enums.h"
#include "../raster/RasterFactory.h"
#include "../raster/Grid.h"
#include "../raster/Band.h"
#include "../raster/BandProperty.h"
#include "../raster/PositionIterator.h"
#include "../raster/Utils.h"
#include "../memory/CachedRaster.h"
#include "../geometry/Envelope.h"
#include "../geometry/GTFactory.h"
#include "../geometry/Polygon.h"
#include "../geometry/LinearRing.h"
#include "../geometry/MultiPolygon.h"
#include "../srs/Converter.h"
#include "../common/progress/TaskProgress.h"
#include "../common/MathUtils.h"

#include <boost/shared_array.hpp>
#include <boost/filesystem.hpp>
#include <boost/lexical_cast.hpp>

#include <climits>
#include <cfloat>
#include <cmath>
#include <memory>
#include <cstdio>

namespace te
{
  namespace rp
  {

    Composition::InputParameters::InputParameters()
    {
      reset();
    }

    Composition::InputParameters::InputParameters( const InputParameters& other )
    {
      reset();
      operator=( other );
    }

    Composition::InputParameters::~InputParameters()
    {
      reset();
    }

    void Composition::InputParameters::reset() _NOEXCEPT_OP(false)
    {
      m_feederRasterPtr = nullptr;
      m_inputRastersBands.clear();
      m_interpMethod = te::rst::NearestNeighbor;
      m_noDataValue = 0.0;
      m_forceInputNoDataValue = false;
      m_autoEqualize = true;
      m_useRasterCache = true;
      m_enableProgress = false;
      m_enableMultiThread = true;
      m_compositionTargetMeans.clear();
      m_compositionTargetStdDevs.clear();
      m_composerRulePtr = nullptr;
    }

    const Composition::InputParameters& Composition::InputParameters::operator=(
      const Composition::InputParameters& params )
    {
      reset();

      m_feederRasterPtr = params.m_feederRasterPtr;
      m_inputRastersBands = params.m_inputRastersBands;
      m_interpMethod = params.m_interpMethod;
      m_noDataValue = params.m_noDataValue;
      m_forceInputNoDataValue = params.m_forceInputNoDataValue;
      m_autoEqualize = params.m_autoEqualize;
      m_useRasterCache = params.m_useRasterCache;
      m_enableProgress = params.m_enableProgress;
      m_enableMultiThread = params.m_enableMultiThread;
      m_compositionTargetMeans = params.m_compositionTargetMeans;
      m_compositionTargetStdDevs = params.m_compositionTargetStdDevs;
      m_composerRulePtr = params.m_composerRulePtr;

      return *this;
    }

    te::common::AbstractParameters* Composition::InputParameters::clone() const
    {
      return new InputParameters( *this );
    }

    bool Composition::InputParameters::serialize( AlgorithmParametersSerializer& serializer ) const
    {
      TERP_TRUE_OR_RETURN_FALSE( AlgorithmInputParameters::serialize( serializer ),
        "Serializarion error" );

      serializer.setAlgorithmName( "Composition" );

      std::vector< std::string > rasterNames;
      m_feederRasterPtr->getRasterNames( rasterNames );
      TERP_TRUE_OR_RETURN_FALSE( serializer.addMultivaluedParameter( "inputRasters",
        rasterNames ), "Error adding parameter inputRasters" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addMultivaluedParameters( "inputRasterBands",
        m_inputRastersBands ), "Error adding parameter inputRasterBands" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "noDataValue",
        m_noDataValue ), "Error adding parameter noDataValue" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "forceInputNoDataValue",
        m_forceInputNoDataValue ), "Error adding parameter forceInputNoDataValue" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "autoEqualize",
        m_autoEqualize ), "Error adding parameter autoEqualize" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addMultivaluedParameter( "compositionTargetMeans",
        m_compositionTargetMeans ), "Error adding parameter compositionTargetMeans" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addMultivaluedParameter( "compositionTargetStdDevs",
        m_compositionTargetStdDevs ), "Error adding parameter compositionTargetStdDevs" );

      return true;
    }

    Composition::OutputParameters::OutputParameters()
    {
      reset();
    }

    Composition::OutputParameters::OutputParameters( const OutputParameters& other )
    {
      reset();
      operator=( other );
    }

    Composition::OutputParameters::~OutputParameters()
    {
      reset();
    }

    void Composition::OutputParameters::reset() _NOEXCEPT_OP(false)
    {
      m_rType.clear();
      m_rInfo.clear();
      m_outputRasterPtr.reset();
    }

    const Composition::OutputParameters& Composition::OutputParameters::operator=(
      const Composition::OutputParameters& params )
    {
      reset();

      m_rType = params.m_rType;
      m_rInfo = params.m_rInfo;
      m_outputRasterPtr.reset();;

      return *this;
    }

    te::common::AbstractParameters* Composition::OutputParameters::clone() const
    {
      return new OutputParameters( *this );
    }

    Composition::Composition()
    {
      reset();
    }

    Composition::~Composition() = default;

    bool Composition::execute( AlgorithmOutputParameters& outputParams )
      _NOEXCEPT_OP(false)
    {
      if( ! m_isInitialized ) return false;
      
      Composition::OutputParameters* outParamsPtr = dynamic_cast<
        Composition::OutputParameters* >( &outputParams );
      TERP_TRUE_OR_THROW( outParamsPtr, "Invalid paramters" );
      
      // progress
      
      std::unique_ptr< te::common::TaskProgress > progressPtr;
      if( m_inputParameters.m_enableProgress )
      {
        progressPtr.reset( new te::common::TaskProgress );
        
        progressPtr->setTotalSteps( static_cast<int>( 2 * 
          m_inputParameters.m_feederRasterPtr->getObjsCount()) );
        
        progressPtr->setMessage( "Composition" );
      }       
      
      // First pass: getting global mosaic info

      double compositionXResolution = 0.0;
      double compositionYResolution = 0.0;
      double compositionLLX = DBL_MAX; // world coords
      double compositionLLY = DBL_MAX; // world coords
      double compositionURX = -1.0 * DBL_MAX; // world coords
      double compositionURY = -1.0 * DBL_MAX; // world coords
      int compositionSRID = 0;     
      te::rst::BandProperty compositionBaseBandProperties( 0, 0, "" );
      
      // a map ordereding the distance to the mosaic center to each raster feeder index
      std::multimap< double, unsigned int > centerDist2RasterIdxMap;

      {
        te::rst::Raster const* inputRasterPtr = nullptr;
        unsigned int inputRasterIdx = 0;
        std::vector< std::unique_ptr< te::gm::Polygon > > rastersBBoxesPols;
        
        // a map from input rasters SRIDS and their frequencies
        std::map< int, unsigned int > sridsFreq;

        m_inputParameters.m_feederRasterPtr->reset();
        while( ( inputRasterPtr = m_inputParameters.m_feederRasterPtr->getCurrentObj() ) )
        {
          inputRasterIdx = m_inputParameters.m_feederRasterPtr->getCurrentOffset();
          TERP_INSTANCE_TRUE_OR_RETURN_FALSE(
            inputRasterPtr->getAccessPolicy() & te::common::RAccess,
            "Invalid raster" );
          
          // checking the input bands

          for( std::vector< unsigned int >::size_type inputRastersBandsIdx = 0 ;
            inputRastersBandsIdx <
            m_inputParameters.m_inputRastersBands[ inputRasterIdx ].size() ;
            ++inputRastersBandsIdx )
          {
            const unsigned int& currBand =
              m_inputParameters.m_inputRastersBands[ inputRasterIdx ][ inputRastersBandsIdx ];

            TERP_INSTANCE_TRUE_OR_RETURN_FALSE( currBand < inputRasterPtr->getNumberOfBands(),
              "Invalid band" )
          }
          
          // get original rasters bboxes

          rastersBBoxesPols.push_back( std::unique_ptr< te::gm::Polygon >(           
            static_cast<te::gm::Polygon*>(te::gm::GetGeomFromEnvelope(
            inputRasterPtr->getGrid()->getExtent(), 
            inputRasterPtr->getGrid()->getSRID() ) ) ) );
          
          // SRIDs
          
          ++sridsFreq[ inputRasterPtr->getSRID() ];

          // move to the next raster
          
          if( m_inputParameters.m_enableProgress )
          {
            progressPtr->pulse();
            if( ! progressPtr->isActive() ) return false;
          }                  

          m_inputParameters.m_feederRasterPtr->moveNext();
        }
        
        // defining the composition SRID
        
        std::map< int, unsigned int >::const_iterator sridsFreqIt = 
          sridsFreq.begin();
        const std::map< int, unsigned int >::const_iterator sridsFreqItEnd = 
          sridsFreq.end();
        unsigned highestSRIDFreq = 0;
        
        while( sridsFreqIt != sridsFreqItEnd )
        {
          if( highestSRIDFreq < sridsFreqIt->second )
          {
            highestSRIDFreq = sridsFreqIt->second;
            compositionSRID = sridsFreqIt->first;
          }
          
          ++sridsFreqIt;
        }
        
        // composition area
        
        te::gm::Polygon* currPolPtr = 0;
          
        for( std::size_t rastersBBoxesPolsIdx = 0 ; rastersBBoxesPolsIdx <
          rastersBBoxesPols.size() ; ++rastersBBoxesPolsIdx )
        {
          currPolPtr = rastersBBoxesPols[ rastersBBoxesPolsIdx ].get();
          
          currPolPtr->transform( compositionSRID );
          
          compositionLLX = std::min( compositionLLX, currPolPtr->getMBR()->getLowerLeftX() );
          compositionLLY = std::min( compositionLLY, currPolPtr->getMBR()->getLowerLeftY() );

          compositionURX = std::max( compositionURX, currPolPtr->getMBR()->getUpperRightX() );
          compositionURY = std::max( compositionURY, currPolPtr->getMBR()->getUpperRightY() );                  
        }
        
        // updating center distances map
        
        const double compositionCenterX = ( compositionURX + compositionLLX ) / 2.0;
        const double compositionCenterY = ( compositionURY + compositionLLY ) / 2.0;
        double currDist = 0;
        double currXDist = 0;
        double currYDist = 0;
        te::gm::Coord2D currCenter;
        
        for( std::size_t rastersBBoxesPolsIdx = 0 ; rastersBBoxesPolsIdx <
          rastersBBoxesPols.size() ; ++rastersBBoxesPolsIdx )
        {
          currPolPtr = rastersBBoxesPols[ rastersBBoxesPolsIdx ].get();
          currCenter = currPolPtr->getMBR()->getCenter();
          currXDist = currCenter.getX() - compositionCenterX;
          currYDist = currCenter.getY() - compositionCenterY;
          currDist = std::sqrt( ( currXDist * currXDist ) + ( currYDist *
            currYDist ) );
          
          centerDist2RasterIdxMap.insert( std::pair< double, unsigned int >(
            currDist, rastersBBoxesPolsIdx ) );
        }        
        
        // defining composition resolution and base band props
        // using the composition center raster as reference
        
        inputRasterIdx = centerDist2RasterIdxMap.begin()->second;
        if( ! m_inputParameters.m_feederRasterPtr->moveTo( inputRasterIdx ) )
        {
          TERP_INSTANCE_LOG_AND_RETURN_FALSE( "Error opening input raster ");
        }
        
        inputRasterPtr = m_inputParameters.m_feederRasterPtr->getCurrentObj();
        
        compositionXResolution = inputRasterPtr->getGrid()->getResolutionX();
        compositionYResolution = inputRasterPtr->getGrid()->getResolutionY();
        
        compositionBaseBandProperties = *inputRasterPtr->getBand( 
          m_inputParameters.m_inputRastersBands[ inputRasterIdx ][ 0 ] )->getProperty();        
        if( m_inputParameters.m_forceInputNoDataValue )
        {
          compositionBaseBandProperties.m_noDataValue = m_inputParameters.m_noDataValue;
        }
      }
      
      TERP_DEBUG_TRUE_OR_THROW( centerDist2RasterIdxMap.size() ==
        m_inputParameters.m_feederRasterPtr->getObjsCount(),
        "Internal error - Mosaic center distances mismatch" );      

      // creating the output raster
      
      te::rst::Raster* outputRasterPtr = nullptr;

      {
        te::rst::Grid* outputGrid = new te::rst::Grid( compositionXResolution,
          compositionYResolution,  new te::gm::Envelope( compositionLLX, 
          compositionLLY, compositionURX, compositionURY ), compositionSRID );
        
        compositionBaseBandProperties.m_blkh = 512;
        compositionBaseBandProperties.m_blkw = 512;
        compositionBaseBandProperties.m_nblocksx = (int)std::ceil(
          ((double)outputGrid->getNumberOfColumns()) / 
          ((double)compositionBaseBandProperties.m_blkw ) );
        compositionBaseBandProperties.m_nblocksy = (int)std::ceil(
          ((double)outputGrid->getNumberOfRows()) / 
          ((double)compositionBaseBandProperties.m_blkh ) );
        
        std::vector< te::rst::BandProperty* > bandsProperties;
        for( std::vector< unsigned int >::size_type bandIdx = 0 ;  bandIdx <
          m_inputParameters.m_inputRastersBands[ 0 ].size() ; ++bandIdx )
        {
          bandsProperties.push_back( new te::rst::BandProperty( compositionBaseBandProperties ) );
          bandsProperties[ bandIdx ]->m_colorInterp = te::rst::GrayIdxCInt;
        }

        AlgorithmParametersSerializer paramsSerializer;
        TERP_INSTANCE_TRUE_OR_LOG(  m_inputParameters.serialize( paramsSerializer ),
          "Parameter serialization error" );

        std::map <std::string, std::string> metadata;
        paramsSerializer.getMetaData( metadata );

        try
        {
          outParamsPtr->m_outputRasterPtr.reset(
            te::rst::RasterFactory::make(
              outParamsPtr->m_rType,
              outputGrid,
              bandsProperties,
              outParamsPtr->m_rInfo,
              metadata,
              nullptr,
              nullptr ) );
        }
        catch(...)
        {
          outParamsPtr->m_outputRasterPtr.reset();
        }
        TERP_INSTANCE_TRUE_OR_RETURN_FALSE( outParamsPtr->m_outputRasterPtr.get(),
          "Output raster creation error" );
          
        outputRasterPtr = outParamsPtr->m_outputRasterPtr.get();
      }

      std::unique_ptr< te::mem::CachedRaster > outputCachedRasterInstancePtr;
      
      if( m_inputParameters.m_useRasterCache )
      {
        outputCachedRasterInstancePtr.reset( new te::mem::CachedRaster(
          *(outParamsPtr->m_outputRasterPtr.get()), 20, 0 ) );   
          
        outputRasterPtr = outputCachedRasterInstancePtr.get();
      }

      // fill output with no data values

      {
        const unsigned int nBands = static_cast<unsigned int>(outputRasterPtr->getNumberOfBands());
        unsigned int bandIdx = 0;

        for( bandIdx = 0 ; bandIdx < nBands ; ++bandIdx )
        {
          te::rst::Band& outBand = ( *( outputRasterPtr->getBand( bandIdx ) ) );
          
          const std::complex<double > bandNoDataValue( 
            outBand.getProperty()->m_noDataValue,
            outBand.getProperty()->m_noDataValue );
          
          te::rp::FillBand( *outputRasterPtr, bandIdx, bandNoDataValue );
        }
      }
      
      // iterating over the all rasters
      
      std::vector< unsigned int > outputRasterBands;
      std::vector< double > dummyRasterOffsets;
      std::vector< double > dummyRasterScales;
      for( unsigned int outputRasterBandsIdx = 0 ; outputRasterBandsIdx <
        outputRasterPtr->getNumberOfBands() ; ++outputRasterBandsIdx )
      {
        outputRasterBands.push_back( outputRasterBandsIdx );
        dummyRasterOffsets.push_back( 0.0 );
        dummyRasterScales.push_back( 1.0 );
      }      
      
      std::vector< double > compositionTargetMeans = m_inputParameters.m_compositionTargetMeans;
      std::vector< double > compositionTargetStdDevs = m_inputParameters.m_compositionTargetStdDevs;

      std::multimap< double, unsigned int >::const_iterator
        centerDist2RasterIdxMapIt = centerDist2RasterIdxMap.begin();
      const std::multimap< double, unsigned int >::const_iterator
        centerDist2RasterIdxMapItEnd = centerDist2RasterIdxMap.end();
      
      while( centerDist2RasterIdxMapIt != centerDist2RasterIdxMapItEnd )
      {
        const unsigned int inputRasterIdx = centerDist2RasterIdxMapIt->second;
        
        if( ! m_inputParameters.m_feederRasterPtr->moveTo( inputRasterIdx ) )
        {
          TERP_INSTANCE_LOG_AND_RETURN_FALSE( "Error opening input raster ");
        }
        
        te::rst::Raster const* inputRasterPtr = m_inputParameters.m_feederRasterPtr->getCurrentObj();

        // Generating the offset and gain info for eath band from the current raster

        std::vector< double > currentRasterBandsOffsets = dummyRasterOffsets;
        std::vector< double > currentRasterBandsScales = dummyRasterScales;

        if( m_inputParameters.m_autoEqualize )
        {
          std::vector< double > currentRasterMeans( 
            m_inputParameters.m_inputRastersBands[ inputRasterIdx ].size(), 0 );
          std::vector< double > currentRasterStdDevs( 
            m_inputParameters.m_inputRastersBands[ inputRasterIdx ].size(), 0 );

          for( unsigned int inputRastersBandsIdx = 0 ; inputRastersBandsIdx <
            m_inputParameters.m_inputRastersBands[ inputRasterIdx ].size() ;
            ++inputRastersBandsIdx )
          {
            const unsigned int inputBandIdx = m_inputParameters.m_inputRastersBands[ inputRasterIdx ][
              inputRastersBandsIdx ];
              
            TERP_INSTANCE_TRUE_OR_RETURN_FALSE( te::rp::GetMeanValue( 
              *inputRasterPtr,
              inputBandIdx, 
              m_inputParameters.m_enableMultiThread ? 0 : 1, // auto number of threads
              m_inputParameters.m_forceInputNoDataValue, 
              m_inputParameters.m_noDataValue, 
              currentRasterMeans[ inputRastersBandsIdx ] ),
              "Unable to calculate the raster mean value" );
            
            TERP_INSTANCE_TRUE_OR_RETURN_FALSE( te::rp::GetStdDevValue( 
              *inputRasterPtr,
              inputBandIdx, 
              m_inputParameters.m_enableMultiThread ? 0 : 1, // auto number of threads
              m_inputParameters.m_forceInputNoDataValue, 
              m_inputParameters.m_noDataValue, 
              &(currentRasterMeans[ inputRastersBandsIdx ]),
              currentRasterStdDevs[ inputRastersBandsIdx ] ),
              "Unable to calculate the raster standart deviation value" );
          }
          
          if( compositionTargetMeans.empty() )
          {
            compositionTargetMeans = currentRasterMeans;
            compositionTargetStdDevs = currentRasterStdDevs;
          }

          for( unsigned int inputRastersBandsIdx = 0 ; inputRastersBandsIdx <
            m_inputParameters.m_inputRastersBands[ inputRasterIdx ].size() ;
            ++inputRastersBandsIdx )
          {
            if( currentRasterStdDevs[ inputRastersBandsIdx ] != 0 )
            {
              currentRasterBandsScales[ inputRastersBandsIdx ] = 
                ( 
                  std::sqrt( compositionTargetStdDevs[ inputRastersBandsIdx ] )
                   /
                  std::sqrt( currentRasterStdDevs[ inputRastersBandsIdx ] ) 
                );
              currentRasterBandsOffsets[ inputRastersBandsIdx ] =
                ( 
                  compositionTargetMeans[ inputRastersBandsIdx ]
                   - 
                  ( 
                    currentRasterBandsScales[ inputRastersBandsIdx ] 
                    * 
                    currentRasterMeans[ inputRastersBandsIdx ] 
                  ) 
                );
            }
          }
        }
        
        // creating the temporary raster where the current compostion will
        // be stored
        // this raster is a patch of the current global composition
        
        std::unique_ptr< te::rst::Raster > tempCompositionRasterPtr;
        std::vector< unsigned int > tempCompositionRasterBands;
        unsigned int tempCompositionRasterLowerLeftRow = 0;
        unsigned int tempCompositionRasterLowerLeftCol = 0;
        unsigned int tempCompositionRasterTopRightRow = 0;
        unsigned int tempCompositionRasterTopRightCol = 0;        
        
        {
          te::gm::Envelope inRasterOriginalEnv( 
            *inputRasterPtr->getGrid()->getExtent() );
          inRasterOriginalEnv.transform( inputRasterPtr->getSRID(), 
            outputRasterPtr->getSRID() );
          
          double lowerLeftRowDouble = 0;
          double lowerLeftColDouble = 0;
          outputRasterPtr->getGrid()->geoToGrid(
            inRasterOriginalEnv.getLowerLeftX(),
            inRasterOriginalEnv.getLowerLeftY(),
            lowerLeftColDouble, lowerLeftRowDouble );
          lowerLeftRowDouble = (double)te::common::Round< double, long int >(
             lowerLeftRowDouble );
          lowerLeftRowDouble = std::max( lowerLeftRowDouble, 0.0 );
          lowerLeftRowDouble = std::min( lowerLeftRowDouble, 
            (double)outputRasterPtr->getNumberOfRows() - 1 );
          lowerLeftColDouble = (double)te::common::Round< double, long int >(
            lowerLeftColDouble );
          lowerLeftColDouble = std::max( lowerLeftColDouble, 0.0 );
          lowerLeftColDouble = std::min( lowerLeftColDouble, 
            (double)outputRasterPtr->getNumberOfColumns() - 1);
          tempCompositionRasterLowerLeftRow = (unsigned int)lowerLeftRowDouble;
          tempCompositionRasterLowerLeftCol = (unsigned int)lowerLeftColDouble;
          
          double topRightRowDouble = 0;
          double topRightColDouble = 0;
          outputRasterPtr->getGrid()->geoToGrid(
            inRasterOriginalEnv.getUpperRightX(),
            inRasterOriginalEnv.getUpperRightY(),
            topRightColDouble, topRightRowDouble );
          topRightRowDouble = (double)te::common::Round< double, long int >(
            topRightRowDouble );
          topRightRowDouble = std::max( topRightRowDouble, 0.0 );
          topRightRowDouble = std::min( topRightRowDouble, 
            (double)outputRasterPtr->getNumberOfRows() - 1 );
          topRightColDouble = (double)te::common::Round< double, long int >(
            topRightColDouble );
          topRightColDouble = std::max( topRightColDouble, 0.0 );
          topRightColDouble = std::min( topRightColDouble, 
            (double)outputRasterPtr->getNumberOfColumns() - 1 );
          tempCompositionRasterTopRightRow = (unsigned int)topRightRowDouble;
          tempCompositionRasterTopRightCol = (unsigned int)topRightColDouble;
          
          TERP_DEBUG_TRUE_OR_THROW( tempCompositionRasterLowerLeftRow >= 
            tempCompositionRasterTopRightRow, "Temporary raster invalid rows" );
          TERP_DEBUG_TRUE_OR_THROW( tempCompositionRasterLowerLeftCol <= 
            tempCompositionRasterTopRightCol, "Temporary raster invalid rows" );          
            
          unsigned int outNRows = tempCompositionRasterLowerLeftRow - 
            tempCompositionRasterTopRightRow + 1;
          unsigned int outNCols = tempCompositionRasterTopRightCol - 
            tempCompositionRasterLowerLeftCol + 1;
            
          double outLLX = 0;
          double outLLY = 0;
          outputRasterPtr->getGrid()->gridToGeo( 
                lowerLeftColDouble - 0.5, lowerLeftRowDouble + 0.5 , outLLX, outLLY );
          
          double outURX = 0;
          double outURY = 0;
          outputRasterPtr->getGrid()->gridToGeo( 
                topRightColDouble + 0.5, topRightRowDouble - 0.5 , outURX, outURY );
          
          te::gm::Envelope* outEnvPtr = new te::gm::Envelope( outLLX, outLLY,
            outURX, outURY );
          
          te::rst::Grid* outGridPtr = new te::rst::Grid( outNCols, outNRows,
            outEnvPtr, outputRasterPtr->getSRID() );
          
          std::vector< te::rst::BandProperty* > bandsProperties;
          for( unsigned int outBandIdx = 0 ; outBandIdx < 
            outputRasterPtr->getNumberOfBands() ; ++outBandIdx )
          {
            bandsProperties.push_back( new te::rst::BandProperty(
              *outputRasterPtr->getBand( 
              outBandIdx )->getProperty() ) );
            bandsProperties.back()->m_blkh = 1;
            bandsProperties.back()->m_blkw = outNCols;
            bandsProperties.back()->m_nblocksx = 1;
            bandsProperties.back()->m_nblocksy = outNRows;
            
            tempCompositionRasterBands.push_back( outBandIdx );
          }

          std::map<std::string, std::string> rinfo;
          
          try
          {
            tempCompositionRasterPtr.reset(
              te::rst::RasterFactory::make(
                "EXPANSIBLE",
                outGridPtr,
                bandsProperties,
                rinfo,
                nullptr,
                nullptr ) );
          }
          catch(...)
          {
            tempCompositionRasterPtr.reset();
          }
          TERP_INSTANCE_TRUE_OR_RETURN_FALSE( tempCompositionRasterPtr.get(),
            "Temporary raster creation error" );          
        }
        
        // composing

        te::rp::Composer composerInstance;

        TERP_INSTANCE_TRUE_OR_RETURN_FALSE( composerInstance.initialize(
          *outputRasterPtr,
          outputRasterBands,
          *inputRasterPtr,
          m_inputParameters.m_inputRastersBands[ inputRasterIdx ],
          *tempCompositionRasterPtr,
          tempCompositionRasterBands,
          m_inputParameters.m_interpMethod,
          m_inputParameters.m_interpMethod,
          m_inputParameters.m_noDataValue,
          false,
          m_inputParameters.m_forceInputNoDataValue,
          dummyRasterOffsets,
          dummyRasterScales,
          currentRasterBandsOffsets,
          currentRasterBandsScales,
          m_inputParameters.m_enableMultiThread ? 0 : 1, // threadsNumber
          false, // progress interface
          m_inputParameters.m_composerRulePtr,
          true ) //enable cache
          , 
          "Composer initiazing error" );
        
        TERP_INSTANCE_TRUE_OR_RETURN_FALSE( composerInstance.composeIntoRaster3(),
          "Error composing images" );

        // copy data to output
        
        {
          unsigned int inRow = 0;
          unsigned int inCol = 0;
          const unsigned int inNRows = tempCompositionRasterPtr->getNumberOfRows();
          const unsigned int inNCols = tempCompositionRasterPtr->getNumberOfColumns();
          double value = 0;
          unsigned int outRow = 0;
          
          for( unsigned int outBandIdx = 0 ; outBandIdx < 
            outputRasterPtr->getNumberOfBands() ; ++outBandIdx )
          {
            const te::rst::Band& inBand = *tempCompositionRasterPtr->getBand(
              outBandIdx );
            te::rst::Band& outBand = *outputRasterPtr->getBand(
              outBandIdx );            
            
            for( inRow = 0 ; inRow < inNRows ; ++inRow )
            {
              outRow = inRow + tempCompositionRasterTopRightRow;
              
              for( inCol = 0 ; inCol < inNCols ; ++inCol )
              {
                inBand.getValue( inCol, inRow, value );
                outBand.setValue( inCol + tempCompositionRasterLowerLeftCol, outRow, 
                  value );
              }
            }
          }          
        }
        
        // moving to the next raster

        if( m_inputParameters.m_enableProgress )
        {
          progressPtr->pulse();
          if( ! progressPtr->isActive() ) return false;
        }
        
        ++centerDist2RasterIdxMapIt;
      }
      
      // reseting the output cache
      
      if( outputCachedRasterInstancePtr.get() ) outputCachedRasterInstancePtr.reset();
      
      return true;
    }

    void Composition::reset() _NOEXCEPT_OP(false)
    {
      Algorithm::reset();
      
      m_inputParameters.reset();
      m_isInitialized = false;
    }

    bool Composition::initialize( const AlgorithmInputParameters& inputParams )
      _NOEXCEPT_OP(false)
    {
      reset();

      Composition::InputParameters const* inputParamsPtr = dynamic_cast<
        Composition::InputParameters const* >( &inputParams );
      TERP_TRUE_OR_THROW( inputParamsPtr, "Invalid paramters pointer" );

      m_inputParameters = *inputParamsPtr;

      // Checking the feeder

      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( m_inputParameters.m_feederRasterPtr,
        "Invalid m_feederRasterPtr" )

      TERP_INSTANCE_TRUE_OR_RETURN_FALSE(
        m_inputParameters.m_feederRasterPtr->getObjsCount() > 0,
        "Invalid number of rasters" )

      // checking m_inputRastersBands

      TERP_INSTANCE_TRUE_OR_RETURN_FALSE(
        (static_cast<unsigned int>(m_inputParameters.m_inputRastersBands.size())) ==
        m_inputParameters.m_feederRasterPtr->getObjsCount(),
        "Bands mismatch" );

      for( std::vector< std::vector< unsigned int > >::size_type
        inputRastersBandsIdx = 0 ;  inputRastersBandsIdx <
        m_inputParameters.m_inputRastersBands.size() ; ++inputRastersBandsIdx )
      {
        TERP_INSTANCE_TRUE_OR_RETURN_FALSE( m_inputParameters.m_inputRastersBands[
          inputRastersBandsIdx ].size() > 0, "Invalid bands number" );

        TERP_INSTANCE_TRUE_OR_RETURN_FALSE( m_inputParameters.m_inputRastersBands[
          inputRastersBandsIdx ].size() ==  m_inputParameters.m_inputRastersBands[
          0 ].size(), "Bands number mismatch" );
      }
      
      // checking target means and std devs
      
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( 
        (
          ( 
            m_inputParameters.m_compositionTargetMeans.empty()
            ||
            ( 
              m_inputParameters.m_compositionTargetMeans.size() 
              == 
              m_inputParameters.m_inputRastersBands[ 0 ].size() 
            )
          )
          &&
          ( 
            m_inputParameters.m_compositionTargetStdDevs.empty()
            ||
            ( 
              m_inputParameters.m_compositionTargetStdDevs.size() 
              == 
              m_inputParameters.m_inputRastersBands[ 0 ].size() 
            )
          )          
          &&
          ( 
            m_inputParameters.m_compositionTargetMeans.size()
            ==
            m_inputParameters.m_compositionTargetStdDevs.size()
          )                    
        )
        , "Invalid target means or target standart deviation values" );
      
      // Composer
      
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( m_inputParameters.m_composerRulePtr 
        != nullptr ,
        "Missing composer rule" );
      
      // finalizing

      m_isInitialized = true;

      return true;
    }

    bool Composition::isInitialized() const
    {
      return m_isInitialized;
    }

  } // end namespace rp
}   // end namespace te

