/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/rp/Filter.cpp
  \brief A series of well-known filtering algorithms for images, linear and non-linear.
*/

#include "Filter.h"
#include "Macros.h"
#include "Functions.h"
#include "AlgorithmParametersSerializer.h"
#include "../statistics/core/SummaryFunctions.h"
#include "../raster/Raster.h"
#include "../raster/RasterFactory.h"
#include "../raster/BandProperty.h"
#include "../raster/Grid.h"
#include "../raster/Band.h"
#include "../raster/BandIterator.h"
#include "../raster/BandIteratorWindow.h"
#include "../common/progress/TaskProgress.h"
#include "../common/PlatformUtils.h"
#include "../memory/CachedRaster.h"

#include <memory>
#include <cmath>

namespace te
{
  namespace rp
  {
    // -------------------------------------------------------------------------------
    
    Filter::InputParameters::InputParameters()
    {
      reset();
    }

    Filter::InputParameters::InputParameters( const InputParameters& other )
    {
      reset();
      operator=( other );
    }

    Filter::InputParameters::~InputParameters()
    {
      reset();
    }

    void Filter::InputParameters::reset() _NOEXCEPT_OP(false)
    {
      m_filterType = InputParameters::InvalidFilterT;
      m_inRasterPtr = nullptr;
      m_inRasterBands.clear();
      m_iterationsNumber = 1;
      m_windowH = 3;
      m_windowW = 3;
      m_enableProgress = false;
      m_window.clear();
      m_enableRasterCache = true;
    }

    const Filter::InputParameters& Filter::InputParameters::operator=(
      const Filter::InputParameters& params )
    {
      reset();

      m_filterType = params.m_filterType;
      m_inRasterPtr = params.m_inRasterPtr;
      m_inRasterBands = params.m_inRasterBands;
      m_iterationsNumber = params.m_iterationsNumber;
      m_windowH = params.m_windowH;
      m_windowW = params.m_windowW;
      m_enableProgress = params.m_enableProgress;
      m_window = params.m_window;
      m_enableRasterCache = params.m_enableRasterCache;

      return *this;
    }

    te::common::AbstractParameters* Filter::InputParameters::clone() const
    {
      return new InputParameters( *this );
    }

    bool Filter::InputParameters::serialize( AlgorithmParametersSerializer& serializer ) const
    {
      TERP_TRUE_OR_RETURN_FALSE( AlgorithmInputParameters::serialize( serializer ),
        "Serializarion error" );

      serializer.setAlgorithmName( "Filter" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "filterType",
        m_filterType ), "Error adding parameter filterType" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "inRasterPtr",
        m_inRasterPtr->getName() ), "Error adding parameter inRasterPtr" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addMultivaluedParameter( "inRasterBands",
        m_inRasterBands ), "Error adding parameter inRasterBands" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "iterationsNumber",
        m_iterationsNumber ), "Error adding parameter iterationsNumber" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "windowH",
        m_windowH ), "Error adding parameter windowH" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "windowW",
        m_windowW ), "Error adding parameter windowW" );

      return true;
    }
    
    // -------------------------------------------------------------------------------

    Filter::OutputParameters::OutputParameters()
    {
      reset();
    }

    Filter::OutputParameters::OutputParameters( const OutputParameters& other )
    {
      reset();
      operator=( other );
    }

    Filter::OutputParameters::~OutputParameters()
    {
      reset();
    }

    void Filter::OutputParameters::reset() _NOEXCEPT_OP(false)
    {
      m_rType.clear();
      m_rInfo.clear();
      m_outputRasterPtr.reset();
    }

    const Filter::OutputParameters& Filter::OutputParameters::operator=(
      const Filter::OutputParameters& params )
    {
      reset();

      m_rType = params.m_rType;
      m_rInfo = params.m_rInfo;

      return *this;
    }

    te::common::AbstractParameters* Filter::OutputParameters::clone() const
    {
      return new OutputParameters( *this );
    }
    
    // -------------------------------------------------------------------------------

    Filter::Filter()
    {
      reset();
    }

    Filter::~Filter()
    {
      ResetConvBuffer( 0, 0 );
      ResetConvMatrix( 0, 0 );
    }

    bool Filter::execute( AlgorithmOutputParameters& outputParams ) _NOEXCEPT_OP(false)
    {
      if( ! m_isInitialized ) return false;

      Filter::OutputParameters* outParamsPtr = dynamic_cast<
        Filter::OutputParameters* >( &outputParams );
      TERP_TRUE_OR_THROW( outParamsPtr, "Invalid paramters" );

      // Initializing the buffer rasters

      std::unique_ptr< te::rst::Raster > bufferRaster1Ptr;
      std::unique_ptr< te::rst::Raster > bufferRaster2Ptr;

      {
        std::vector< te::rst::BandProperty* > bufferRaster1BandsProperties;
        std::vector< te::rst::BandProperty* > bufferRaster2BandsProperties;

        for( unsigned int inRasterBandsIdx = 0 ; inRasterBandsIdx <
          m_inputParameters.m_inRasterBands.size() ; ++inRasterBandsIdx )
        {
          if( m_inputParameters.m_iterationsNumber > 1 )
          {          
            bufferRaster1BandsProperties.push_back( new te::rst::BandProperty(
              *( m_inputParameters.m_inRasterPtr->getBand( 
              m_inputParameters.m_inRasterBands[ 0 ] )->getProperty() ) ) );
            bufferRaster1BandsProperties[ inRasterBandsIdx ]->m_type =
              te::dt::DOUBLE_TYPE;
            bufferRaster1BandsProperties[ inRasterBandsIdx ]->m_blkw =
              m_inputParameters.m_inRasterPtr->getNumberOfColumns();
            bufferRaster1BandsProperties[ inRasterBandsIdx ]->m_blkh = 1;
            bufferRaster1BandsProperties[ inRasterBandsIdx ]->m_nblocksx = 1;
            bufferRaster1BandsProperties[ inRasterBandsIdx ]->m_nblocksy =
              m_inputParameters.m_inRasterPtr->getNumberOfRows();
          }

          if( m_inputParameters.m_iterationsNumber > 2 )
          {        
            bufferRaster2BandsProperties.push_back( new te::rst::BandProperty(
              *bufferRaster1BandsProperties[ inRasterBandsIdx ] ) );
          }
        }

        const double totalPhysMem = (double)te::common::GetTotalPhysicalMemory();
        const double usedVMem = (double)te::common::GetUsedVirtualMemory();
        const double totalVMem = ( (double)te::common::GetTotalVirtualMemory() );
        const double freeVMem = 0.75 * MIN( totalPhysMem, ( totalVMem - usedVMem ) );        
        const double blockRequiredRam = sizeof( double ) *
          m_inputParameters.m_inRasterPtr->getNumberOfColumns();
        const unsigned int maxBlocksPerRaster = 
          (unsigned int)
          ( 
            ( freeVMem / blockRequiredRam )
            /
            ( ( m_inputParameters.m_iterationsNumber > 2 ) ? 2 : 1 ) 
          );
          
        std::map< std::string, std::string > rInfo;
        rInfo[ "MAXNUMBEROFRAMBLOCKS" ] = std::to_string( maxBlocksPerRaster );

        if( m_inputParameters.m_iterationsNumber > 1 )
        {
          try
          {
            bufferRaster1Ptr.reset(
              te::rst::RasterFactory::make(
                "EXPANSIBLE",
                new te::rst::Grid( *( m_inputParameters.m_inRasterPtr->getGrid() ) ),
                bufferRaster1BandsProperties,
                rInfo,
                nullptr,
                nullptr ) );
            TERP_INSTANCE_TRUE_OR_RETURN_FALSE( bufferRaster1Ptr.get(),
              "Output raster creation error" );
          }
          catch(...)
          {
            TERP_INSTANCE_LOG_AND_RETURN_FALSE( 
              "Output raster creation error" );
          }          
        }

        if( m_inputParameters.m_iterationsNumber > 2 )
        {
          try
          {
            bufferRaster2Ptr.reset(
              te::rst::RasterFactory::make(
                "EXPANSIBLE",
                new te::rst::Grid( *( m_inputParameters.m_inRasterPtr->getGrid() ) ),
                bufferRaster2BandsProperties,
                rInfo,
                nullptr,
                nullptr ) );
            TERP_INSTANCE_TRUE_OR_RETURN_FALSE( bufferRaster2Ptr.get(),
              "Output raster creation error" );
          }
          catch(...)
          {
            TERP_INSTANCE_LOG_AND_RETURN_FALSE( 
              "Output raster creation error" );
          }
        }
      }
      
      // Initializing the output raster

      {
        std::vector< te::rst::BandProperty* > outRasterBandsProperties;

        for( unsigned int inRasterBandsIdx = 0 ; inRasterBandsIdx <
          m_inputParameters.m_inRasterBands.size() ; ++inRasterBandsIdx )
        {
          const unsigned int& inRasterBandIndex =
            m_inputParameters.m_inRasterBands[ inRasterBandsIdx ];

          assert( inRasterBandIndex <
            m_inputParameters.m_inRasterPtr->getNumberOfBands() );

          outRasterBandsProperties.push_back( new te::rst::BandProperty(
            *( m_inputParameters.m_inRasterPtr->getBand( inRasterBandIndex )->getProperty() ) ) );
          outRasterBandsProperties[ inRasterBandsIdx ]->m_blkw =
            m_inputParameters.m_inRasterPtr->getNumberOfColumns();
          outRasterBandsProperties[ inRasterBandsIdx ]->m_blkh = 1;
          outRasterBandsProperties[ inRasterBandsIdx ]->m_nblocksx = 1;
          outRasterBandsProperties[ inRasterBandsIdx ]->m_nblocksy =
            m_inputParameters.m_inRasterPtr->getNumberOfRows();          
        }

        AlgorithmParametersSerializer paramsSerializer;
        TERP_INSTANCE_TRUE_OR_LOG(  m_inputParameters.serialize( paramsSerializer ),
          "Parameter serialization error" );

        std::map <std::string, std::string> metadata;
        paramsSerializer.getMetaData( metadata );

        try
        {
          outParamsPtr->m_outputRasterPtr.reset(
            te::rst::RasterFactory::make(
              outParamsPtr->m_rType,
              new te::rst::Grid( *( m_inputParameters.m_inRasterPtr->getGrid() ) ),
              outRasterBandsProperties,
              outParamsPtr->m_rInfo,
              metadata,
              nullptr,
              nullptr ) );
        }
        catch(...)
        {
          outParamsPtr->m_outputRasterPtr.reset(0);
        }
        
        TERP_INSTANCE_TRUE_OR_RETURN_FALSE( outParamsPtr->m_outputRasterPtr.get(),
          "Output raster creation error" );
      }      

      // Filtering

      FilterMethodPointerT filterPointer = nullptr;
      switch( m_inputParameters.m_filterType )
      {
        case InputParameters::RobertsFilterT :
        {
          filterPointer = &Filter::RobertsFilter;
          break;
        }
        case InputParameters::SobelFilterT :
        {
          filterPointer = &Filter::SobelFilter;
          break;
        }
        case InputParameters::MeanFilterT :
        {
          filterPointer = &Filter::MeanFilter;
          break;
        }
        case InputParameters::ModeFilterT :
        {
          filterPointer = &Filter::ModeFilter;
          break;
        }
        case InputParameters::MedianFilterT :
        {
          filterPointer = &Filter::MedianFilter;
          break;
        }
        case InputParameters::DilationFilterT :
        {
          filterPointer = &Filter::DilationFilter;
          break;
        }
        case InputParameters::ErosionFilterT :
        {
          filterPointer = &Filter::ErosionFilter;
          break;
        }
        case InputParameters::UserDefinedWindowT :
        {
          filterPointer = &Filter::UserDefinedFilter;
          break;
        }
        default :
        {
          TERP_LOG_AND_THROW( "Invalid filter type" );
        }
      }

      const bool useGlobalProgress = ( ( m_inputParameters.m_iterationsNumber *
        m_inputParameters.m_inRasterBands.size() ) != 1 );

      std::unique_ptr< te::common::TaskProgress > task;
      if( m_inputParameters.m_enableProgress && useGlobalProgress )
      {
        task.reset( new te::common::TaskProgress(TE_TR("Filtering"),
          te::common::TaskProgress::UNDEFINED,
         static_cast<int>(( m_inputParameters.m_iterationsNumber *
          m_inputParameters.m_inRasterBands.size() ) ) ) );
      }

      std::unique_ptr< te::mem::CachedRaster > inputCachedRasterPtr;
      if( m_inputParameters.m_enableRasterCache )
      {
        inputCachedRasterPtr.reset( new te::mem::CachedRaster(
          *m_inputParameters.m_inRasterPtr, 50, 0 ) );
      }
      
      te::rst::Raster const* srcRasterPtr = nullptr;
      te::rst::Raster* dstRasterPtr = nullptr;
      te::rst::Raster const* auxRasterPtr = nullptr;

      for( unsigned int iteration = 0 ; iteration <
        m_inputParameters.m_iterationsNumber ; ++iteration )
      {
        // defining the source raster and
        // destination raster

        if( iteration == 0 )
        {
          if( m_inputParameters.m_enableRasterCache )
          {          
            srcRasterPtr = inputCachedRasterPtr.get();
          }
          else
          {
            srcRasterPtr = m_inputParameters.m_inRasterPtr;
          }

          if( m_inputParameters.m_iterationsNumber == 1 )
          {
            dstRasterPtr = outParamsPtr->m_outputRasterPtr.get();
          }
          else
          {
            dstRasterPtr = bufferRaster1Ptr.get();
          }
        }
        else if( iteration == ( m_inputParameters.m_iterationsNumber - 1 ) )
        {
          srcRasterPtr = dstRasterPtr;

          dstRasterPtr = outParamsPtr->m_outputRasterPtr.get();
        }
        else if( iteration == 1 )
        {
          srcRasterPtr = dstRasterPtr;

          dstRasterPtr = bufferRaster2Ptr.get();
        }
        else
        {
          auxRasterPtr = srcRasterPtr;
          srcRasterPtr = dstRasterPtr;
          dstRasterPtr = (te::rst::Raster*)(auxRasterPtr);
        }

        for( unsigned int inRasterBandsIdx = 0 ; inRasterBandsIdx <
          m_inputParameters.m_inRasterBands.size() ; ++inRasterBandsIdx )
        {
          if( (this->*(filterPointer))( *srcRasterPtr, ( iteration == 0 ) ?
            m_inputParameters.m_inRasterBands[ inRasterBandsIdx ] :
            inRasterBandsIdx, *dstRasterPtr, inRasterBandsIdx,
            ( m_inputParameters.m_enableProgress && (!useGlobalProgress) ) ) ==
            false )
          {
            TERP_INSTANCE_LOG_AND_RETURN_FALSE( TE_TR( "Filter error" ) );
          }
        }

      }

      return true;
    }

    void Filter::reset() _NOEXCEPT_OP(false)
    {
      Algorithm::reset();

      m_isInitialized = false;
      m_inputParameters.reset();

      m_convBuffer = nullptr;
      m_convBufferLines = 0;
      m_convBufferColumns = 0;
      
      m_convMatrix = nullptr;
      m_convMatrixLines = 0;
      m_convMatrixColumns = 0;
    }

    bool Filter::initialize( const AlgorithmInputParameters& inputParams ) _NOEXCEPT_OP(false)
    {
      reset();

      Filter::InputParameters const* inputParamsPtr = dynamic_cast<
        Filter::InputParameters const* >( &inputParams );
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( inputParamsPtr, TE_TR( "Invalid parameters" ) );

      m_inputParameters = *inputParamsPtr;

      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( m_inputParameters.m_filterType !=
        InputParameters::InvalidFilterT,
        TE_TR( "Invalid filter type" ) );

      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( m_inputParameters.m_inRasterPtr,
        TE_TR( "Invalid raster pointer" ) );
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE(
        m_inputParameters.m_inRasterPtr->getAccessPolicy() & te::common::RAccess,
        TE_TR( "Invalid raster" ) );

      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( m_inputParameters.m_inRasterBands.size() > 0,
        TE_TR( "Invalid raster bands number" ) );
      for( unsigned int inRasterBandsIdx = 0 ; inRasterBandsIdx <
        m_inputParameters.m_inRasterBands.size() ; ++inRasterBandsIdx )
      {
        TERP_INSTANCE_TRUE_OR_RETURN_FALSE(
          m_inputParameters.m_inRasterBands[ inRasterBandsIdx ] <
          m_inputParameters.m_inRasterPtr->getNumberOfBands(),
          TE_TR( "Invalid raster bands" ) );
      }

      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( m_inputParameters.m_iterationsNumber > 0,
        TE_TR( "Invalid iterations number" ) );

      // Checking window sizes

      if( m_inputParameters.m_window.size1() > 0 )
      {
        m_inputParameters.m_windowH = static_cast<unsigned int>(m_inputParameters.m_window.size1());
        m_inputParameters.m_windowW = static_cast<unsigned int>(m_inputParameters.m_window.size2());
      }

      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( ( m_inputParameters.m_windowH > 2 ) &&
        ( ( m_inputParameters.m_windowH % 2 ) != 0 ),
        TE_TR( "Invalid mask window height" ) );
      
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE(
        ( 
          m_inputParameters.m_windowH
          <= 
          m_inputParameters.m_inRasterPtr->getNumberOfRows()
        ) 
        &&
        ( 
          m_inputParameters.m_windowW
          <= 
          m_inputParameters.m_inRasterPtr->getNumberOfColumns()
        ),
        TE_TR( "Invalid mask window sizes" ) );      

      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( ( m_inputParameters.m_windowW > 2 ) &&
        ( ( m_inputParameters.m_windowW % 2 ) != 0 ),
        TE_TR( "Invalid mask window width" ) );

      // Checking the user input filter mask

      if( m_inputParameters.m_filterType ==
         InputParameters::UserDefinedWindowT )
      {
        TERP_INSTANCE_TRUE_OR_RETURN_FALSE( m_inputParameters.m_window.size1() > 2,
          TE_TR( "Invalid convolution window" ) );
      }

      if( m_inputParameters.m_window.size1() > 0 )
      {
        TERP_INSTANCE_TRUE_OR_RETURN_FALSE( m_inputParameters.m_window.size1() ==
          m_inputParameters.m_window.size2(),
          TE_TR( "Invalid convolution matrix" ) );

        TERP_INSTANCE_TRUE_OR_RETURN_FALSE( ( ( m_inputParameters.m_window.size1() % 2 ) != 0 ),
          TE_TR( "Invalid convolution matrix" ) );

        if(
            ( m_inputParameters.m_filterType == InputParameters::ErosionFilterT )
            ||
            ( m_inputParameters.m_filterType == InputParameters::DilationFilterT )
          )
        {
          boost::numeric::ublas::matrix<double>::const_iterator1 it =
            m_inputParameters.m_window.begin1();
          boost::numeric::ublas::matrix<double>::const_iterator1 itEnd =
            m_inputParameters.m_window.end1();

          while( it != itEnd )
          {
            if( ( *it != 0 ) && ( *it != 1.0 ) )

            TERP_INSTANCE_TRUE_OR_RETURN_FALSE(
              ( ( *it == 0.0 ) || ( *it == 1.0 ) ),
              TE_TR( "Invalid convolution matrix values" ) );

            ++it;
          }
        }
      }

      // Finalizing


      m_isInitialized = true;

      return true;
    }

    bool Filter::isInitialized() const
    {
      return m_isInitialized;
    }

    bool Filter::RobertsFilter( const te::rst::Raster& srcRaster,
      const unsigned int srcBandIdx, te::rst::Raster& dstRaster,
      const unsigned int dstBandIdx, const bool useProgress )
    {
      TERP_DEBUG_TRUE_OR_THROW( srcRaster.getNumberOfColumns() ==
        dstRaster.getNumberOfColumns(), "Internal error" );
      TERP_DEBUG_TRUE_OR_THROW( srcRaster.getNumberOfRows() ==
        dstRaster.getNumberOfRows(), "Internal error" );
      TERP_DEBUG_TRUE_OR_THROW( srcBandIdx < srcRaster.getNumberOfBands(),
        "Internal error" );
      TERP_DEBUG_TRUE_OR_THROW( dstBandIdx < dstRaster.getNumberOfBands(),
        "Internal error" );

      const unsigned int nRows = static_cast<unsigned int>( srcRaster.getNumberOfRows() );
      const unsigned int rowsBound = static_cast<unsigned int>( nRows ?
        ( nRows - 1 ) : 0 );

      const unsigned int nCols = static_cast<unsigned int>( srcRaster.getNumberOfColumns() );
      const unsigned int colsBound = static_cast<unsigned int>( nCols ?
        ( nCols - 1 ) : 0 );

      std::unique_ptr< te::common::TaskProgress > task;
      if( useProgress )
      {
        task.reset( new te::common::TaskProgress(TE_TR("Mean Filter"),
          te::common::TaskProgress::UNDEFINED,
         static_cast<int>(srcRaster.getNumberOfRows()) - 2) );
      }

      const te::rst::Band& srcBand = *srcRaster.getBand( srcBandIdx );
      te::rst::Band& dstBand = *dstRaster.getBand( dstBandIdx );

      const double srcNoDataValue = srcBand.getProperty()->m_noDataValue;
      const double dstNoDataValue = dstBand.getProperty()->m_noDataValue;

      double dstBandAllowedMin = 0;
      double dstBandAllowedMax = 0;
      te::rp::GetDataTypeRange( dstBand.getProperty()->getType(), dstBandAllowedMin,
        dstBandAllowedMax );

      unsigned int col = 0;
      double value1diag = 0;
      double value2diag = 0;
      double value1adiag = 0;
      double value2adiag = 0;
      double diagDiff = 0;
      double adiagDiff = 0;
      double outValue = 0;

      for( unsigned int row = 0; row < nRows ; ++row )
      {
        for( col = 0 ; col < nCols ; ++col )
        {
          if( ( row < rowsBound ) && ( col < colsBound ) )
          {
            srcBand.getValue( col, row, value1diag );
            if( value1diag == srcNoDataValue )
            {
              dstBand.setValue( col, row, dstNoDataValue );
              continue;
            }

            srcBand.getValue( col + 1, row + 1, value2diag );
            if( value2diag == srcNoDataValue )
            {
              dstBand.setValue( col, row, dstNoDataValue );
              continue;
            }

            srcBand.getValue( col, row + 1, value1adiag );
            if( value1adiag == srcNoDataValue )
            {
              dstBand.setValue( col, row, dstNoDataValue );
              continue;
            }

            srcBand.getValue( col + 1, row, value2adiag );
            if( value2adiag == srcNoDataValue )
            {
              dstBand.setValue( col, row, dstNoDataValue );
              continue;
            }

            diagDiff = value1diag - value2diag;
            adiagDiff = value1adiag - value2adiag;

            outValue = std::sqrt( ( diagDiff * diagDiff ) +
              ( adiagDiff * adiagDiff ) );
            outValue = std::max( outValue, dstBandAllowedMin );
            outValue = std::min( outValue, dstBandAllowedMax );

            dstBand.setValue( col, row, outValue );
          }
          else
          {
            dstBand.setValue( col, row, dstNoDataValue );
          }
        }

        if( useProgress )
        {
          task->pulse();
          if( !task->isActive() ) return false;
        }
      }

      return true;
    }

    bool Filter::SobelFilter( const te::rst::Raster& srcRaster,
      const unsigned int srcBandIdx, te::rst::Raster& dstRaster,
      const unsigned int dstBandIdx, const bool useProgress )
    {
      TERP_DEBUG_TRUE_OR_THROW( srcRaster.getNumberOfColumns() ==
        dstRaster.getNumberOfColumns(), "Internal error" );
      TERP_DEBUG_TRUE_OR_THROW( srcRaster.getNumberOfRows() ==
        dstRaster.getNumberOfRows(), "Internal error" );
      TERP_DEBUG_TRUE_OR_THROW( srcBandIdx < srcRaster.getNumberOfBands(),
        "Internal error" );
      TERP_DEBUG_TRUE_OR_THROW( dstBandIdx < dstRaster.getNumberOfBands(),
        "Internal error" );

      const unsigned int nRows = static_cast<unsigned int>( srcRaster.getNumberOfRows() );
      const unsigned int rowsBound = static_cast<unsigned int>( nRows - 2 );

      const unsigned int nCols = static_cast<unsigned int>( srcRaster.getNumberOfColumns() );
      const unsigned int colsBound = static_cast<unsigned int>( nCols - 2 );

      ResetConvBuffer(3, nCols);

      std::unique_ptr< te::common::TaskProgress > task;
      if( useProgress )
      {
        task.reset( new te::common::TaskProgress(TE_TR("Sobel Filter"),
          te::common::TaskProgress::UNDEFINED,
         static_cast<int>(srcRaster.getNumberOfRows()) - 2) );
      }

      const te::rst::Band& srcBand = *srcRaster.getBand( srcBandIdx );
      te::rst::Band& dstBand = *dstRaster.getBand( dstBandIdx );

      double dstBandAllowedMin = 0;
      double dstBandAllowedMax = 0;
      te::rp::GetDataTypeRange( dstBand.getProperty()->getType(), dstBandAllowedMin,
        dstBandAllowedMax );

      unsigned int col = 0;
      double gY = 0;
      double gX = 0;
      double outValue = 0;

      /* Fills the convolution buffer with the first "mask_lines" from the raster */

      for (unsigned int line = 0; line < 2; ++line) {
        UpdateConvBuffer(srcBand, line);
      }

      /* raster convolution */

      for( unsigned int row = 0; row < rowsBound ; ++row )
      {
        /* Getting one more line from the source raster and adding to buffer */
        UpdateConvBuffer(srcBand, row + 2);

        for( col = 0 ; col < colsBound ; ++col )
        {
          gX = m_convBuffer[2][col] +
            (2 * m_convBuffer[2][col + 1]) +
            m_convBuffer[2][col + 2] -
            m_convBuffer[0][col] -
            (2 * m_convBuffer[0][col + 1]) -
            m_convBuffer[0][col + 2];

          gY = m_convBuffer[0][col + 2] +
            (2 * m_convBuffer[1][col + 2]) +
            m_convBuffer[2][col + 2] -
            m_convBuffer[0][col] -
            (2 * m_convBuffer[1][col]) -
            m_convBuffer[2][col];

          outValue = std::sqrt( ( gY * gY ) +
            ( gX * gX ) );
          outValue = std::max( outValue, dstBandAllowedMin );
          outValue = std::min( outValue, dstBandAllowedMax );

          dstBand.setValue( col, row, outValue );
        }

        if( useProgress )
        {
          task->pulse();
          if( !task->isActive() ) return false;
        }
      }

      return true;
    }

    bool Filter::MeanFilter( const te::rst::Raster& srcRaster,
      const unsigned int srcBandIdx, te::rst::Raster& dstRaster,
      const unsigned int dstBandIdx, const bool useProgress )
    {
      TERP_DEBUG_TRUE_OR_THROW( srcRaster.getNumberOfColumns() ==
        dstRaster.getNumberOfColumns(), "Internal error" );
      TERP_DEBUG_TRUE_OR_THROW( srcRaster.getNumberOfRows() ==
        dstRaster.getNumberOfRows(), "Internal error" );
      TERP_DEBUG_TRUE_OR_THROW( srcBandIdx < srcRaster.getNumberOfBands(),
        "Internal error" );
      TERP_DEBUG_TRUE_OR_THROW( dstBandIdx < dstRaster.getNumberOfBands(),
        "Internal error" );

      const unsigned int nRows = srcRaster.getNumberOfRows();
      const unsigned int nCols = srcRaster.getNumberOfColumns();
      const unsigned int windowHeight = m_inputParameters.m_windowH;
      const unsigned int windowWidth = m_inputParameters.m_windowW;
      const unsigned int windowRowRadius = ( windowHeight / 2 );
      const unsigned int windowColRadius = ( windowWidth / 2 );
      const unsigned int filterStartRowOffsetBound = nRows - windowHeight + 1;
      const unsigned int filterStartColOffsetBound = nCols - windowWidth + 1;
      const unsigned int validDataRowsBound = nRows - windowRowRadius;
      const unsigned int validDataColsBound = nCols - windowColRadius;
      unsigned int filterStartRow = 0;
      unsigned int filterStartCol = 0;
      unsigned int rowOffset = 0;
      unsigned int colOffset = 0;
      std::map< double , unsigned int > frequencies;
      const te::rst::Band& srcBand = *( srcRaster.getBand( srcBandIdx ) );
      te::rst::Band& dstBand = *( dstRaster.getBand( dstBandIdx ) );
      const double srcNoDataValue = srcBand.getProperty()->m_noDataValue;
      const double dstNoDataValue = dstBand.getProperty()->m_noDataValue;
      double value = 0;
      unsigned int row = 0;
      unsigned int col = 0;
      std::map< double , unsigned int >::iterator frequenciesIt;
      std::map< double , unsigned int >::iterator frequenciesItEnd;
      double meanValue = 0;
      unsigned int meanValidPixelsNmb = 0;
      
      std::unique_ptr< te::common::TaskProgress > task;
      if( useProgress )
      {
        task.reset( new te::common::TaskProgress(TE_TR("Mode Filter"),
          te::common::TaskProgress::UNDEFINED, static_cast<int>(filterStartRowOffsetBound + nRows) ) );
      }      
      
      for( row = 0 ; row < nRows ; ++row )
      {
        for( col = 0 ; col < nCols ; ++col )
        {
          if( ( row < windowRowRadius ) || ( row >= validDataRowsBound ) ||
            ( col < windowColRadius ) || ( col >= validDataColsBound ) )
          {
            srcBand.getValue( col, row, value );
            dstBand.setValue( col, row, value );
          }
        }
        
        if( useProgress )
        {
          task->pulse();
          if( !task->isActive() ) return false;
        }         
      }
      
      for( filterStartRow = 0 ; filterStartRow < filterStartRowOffsetBound ; ++filterStartRow )
      {
        for( filterStartCol = 0 ; filterStartCol < filterStartColOffsetBound ; ++filterStartCol )
        {
          meanValue = 0;
          meanValidPixelsNmb = 0;
          
          for( rowOffset = 0 ; rowOffset < windowHeight ; ++rowOffset )
          {
            row = filterStartRow + rowOffset;
            
            for( colOffset = 0 ; colOffset < windowWidth ; ++colOffset )
            {
              col = filterStartCol + colOffset;
              srcBand.getValue( col, row, value );
              if( value != srcNoDataValue )
              {
                meanValue += value;
                ++meanValidPixelsNmb;
              }
            }
          }
          
          if( meanValidPixelsNmb )
          {
            meanValue /= static_cast<double>(meanValidPixelsNmb);
            dstBand.setValue( filterStartCol + windowColRadius, 
              filterStartRow + windowRowRadius, meanValue );            
          }
          else
          {
            dstBand.setValue( filterStartCol + windowColRadius, 
              filterStartRow + windowRowRadius, dstNoDataValue );
          }
        }
        
        if( useProgress )
        {
          task->pulse();
          if( !task->isActive() ) return false;
        }        
      }

      return true;
    }

    bool Filter::ModeFilter( const te::rst::Raster& srcRaster,
      const unsigned int srcBandIdx, te::rst::Raster& dstRaster,
      const unsigned int dstBandIdx, const bool useProgress )
    {
      TERP_DEBUG_TRUE_OR_THROW( srcRaster.getNumberOfColumns() ==
        dstRaster.getNumberOfColumns(), "Internal error" );
      TERP_DEBUG_TRUE_OR_THROW( srcRaster.getNumberOfRows() ==
        dstRaster.getNumberOfRows(), "Internal error" );
      TERP_DEBUG_TRUE_OR_THROW( srcBandIdx < srcRaster.getNumberOfBands(),
        "Internal error" );
      TERP_DEBUG_TRUE_OR_THROW( dstBandIdx < dstRaster.getNumberOfBands(),
        "Internal error" );

      const unsigned int nRows = srcRaster.getNumberOfRows();
      const unsigned int nCols = srcRaster.getNumberOfColumns();
      const unsigned int windowHeight = m_inputParameters.m_windowH;
      const unsigned int windowWidth = m_inputParameters.m_windowW;
      const unsigned int windowRowRadius = ( windowHeight / 2 );
      const unsigned int windowColRadius = ( windowWidth / 2 );
      const unsigned int filterStartRowOffsetBound = nRows - windowHeight + 1;
      const unsigned int filterStartColOffsetBound = nCols - windowWidth + 1;
      const unsigned int validDataRowsBound = nRows - windowRowRadius;
      const unsigned int validDataColsBound = nCols - windowColRadius;
      unsigned int filterStartRow = 0;
      unsigned int filterStartCol = 0;
      unsigned int rowOffset = 0;
      unsigned int colOffset = 0;
      std::map< double , unsigned int > frequencies;
      const te::rst::Band& srcBand = *( srcRaster.getBand( srcBandIdx ) );
      te::rst::Band& dstBand = *( dstRaster.getBand( dstBandIdx ) );
      const double srcNoDataValue = srcBand.getProperty()->m_noDataValue;
      const double dstNoDataValue = dstBand.getProperty()->m_noDataValue;
      double value = 0;
      unsigned int row = 0;
      unsigned int col = 0;
      std::map< double , unsigned int >::iterator frequenciesIt;
      std::map< double , unsigned int >::iterator frequenciesItEnd;
      unsigned int higherFrequency = 0;
      double higherFrequencyValue = 0;
      
      std::unique_ptr< te::common::TaskProgress > task;
      if( useProgress )
      {
        task.reset( new te::common::TaskProgress(TE_TR("Mode Filter"),
          te::common::TaskProgress::UNDEFINED, static_cast<int>(filterStartRowOffsetBound + nRows) ) );
      }      
      
      for( row = 0 ; row < nRows ; ++row )
      {
        for( col = 0 ; col < nCols ; ++col )
        {
          if( ( row < windowRowRadius ) || ( row >= validDataRowsBound ) ||
            ( col < windowColRadius ) || ( col >= validDataColsBound ) )
          {
            dstBand.setValue( col, row, dstNoDataValue );
          }
        }
        
        if( useProgress )
        {
          task->pulse();
          if( !task->isActive() ) return false;
        }         
      }
      
      for( filterStartRow = 0 ; filterStartRow < filterStartRowOffsetBound ; ++filterStartRow )
      {
        for( filterStartCol = 0 ; filterStartCol < filterStartColOffsetBound ; ++filterStartCol )
        {
          frequencies.clear();
          
          for( rowOffset = 0 ; rowOffset < windowHeight ; ++rowOffset )
          {
            row = filterStartRow + rowOffset;
            
            for( colOffset = 0 ; colOffset < windowWidth ; ++colOffset )
            {
              col = filterStartCol + colOffset;
              srcBand.getValue( col, row, value );
              if( value != srcNoDataValue )
              {
                ++frequencies[ value ];
              }
            }
          }
          
          frequenciesIt = frequencies.begin();
          frequenciesItEnd = frequencies.end();
          higherFrequency = 0;
          while( frequenciesIt != frequenciesItEnd )
          {
            if( frequenciesIt->second > higherFrequency )
            {
              higherFrequency = frequenciesIt->second;
              higherFrequencyValue = frequenciesIt->first;              
            }
            
            ++frequenciesIt;
          }
          
          if( higherFrequency == 0.0 )
          {
            dstBand.setValue( filterStartCol + windowColRadius, 
              filterStartRow + windowRowRadius, dstNoDataValue );
          }
          else
          {
            dstBand.setValue( filterStartCol + windowColRadius, 
              filterStartRow + windowRowRadius, higherFrequencyValue );
          }
        }
        
        if( useProgress )
        {
          task->pulse();
          if( !task->isActive() ) return false;
        }        
      }

      return true;
    }

    bool Filter::MedianFilter( const te::rst::Raster& srcRaster,
      const unsigned int srcBandIdx, te::rst::Raster& dstRaster,
      const unsigned int dstBandIdx, const bool useProgress )
    {
      TERP_DEBUG_TRUE_OR_THROW( srcRaster.getNumberOfColumns() ==
        dstRaster.getNumberOfColumns(), "Internal error" );
      TERP_DEBUG_TRUE_OR_THROW( srcRaster.getNumberOfRows() ==
        dstRaster.getNumberOfRows(), "Internal error" );
      TERP_DEBUG_TRUE_OR_THROW( srcBandIdx < srcRaster.getNumberOfBands(),
        "Internal error" );
      TERP_DEBUG_TRUE_OR_THROW( dstBandIdx < dstRaster.getNumberOfBands(),
        "Internal error" );

      const unsigned int H = m_inputParameters.m_windowH;
      const unsigned int W = m_inputParameters.m_windowW;

      std::unique_ptr< te::common::TaskProgress > task;
      if( useProgress )
      {
        task.reset( new te::common::TaskProgress(TE_TR("Median Filter"),
          te::common::TaskProgress::UNDEFINED, 100 ) );
      }

      unsigned int R = 0;
      unsigned int C = 0;
      const unsigned int MR = srcRaster.getNumberOfRows();
      const unsigned int MC = srcRaster.getNumberOfColumns();

      double dstBandAllowedMin = 0;
      double dstBandAllowedMax = 0;
      te::rp::GetDataTypeRange( dstRaster.getBand( dstBandIdx )->getProperty()->getType(), dstBandAllowedMin,
        dstBandAllowedMax );

      std::vector<double> pixels_in_window;
      double pixel_median = 0.0;

      te::rst::Band const * const band = srcRaster.getBand(srcBandIdx);
      te::rst::BandIteratorWindow<double> it = te::rst::BandIteratorWindow<double>::begin(band, W, H);
      te::rst::BandIteratorWindow<double> itend = te::rst::BandIteratorWindow<double>::end(band, W, H);

      while (it != itend)
      {
        R = it.getRow();
        C = it.getColumn();

        if ((R >= static_cast<unsigned int>(H / 2) && R < MR - (H / 2)) &&
            (C >= static_cast<unsigned int>(W / 2) && C < MC - (W / 2)))
        {
          pixel_median = 0.0;
          pixels_in_window.clear();
          for (int r = -1 * static_cast<int> (H / 2); r <= static_cast<int> (H / 2); r++)
            for (int c = -1 * static_cast<int> (W / 2); c <= static_cast<int> (W / 2); c++)
              pixels_in_window.push_back(it.getValue(c, r));

          std::sort(pixels_in_window.begin(), pixels_in_window.end(), OrderFunction);
          pixel_median = pixels_in_window[pixels_in_window.size() / 2];
        }
        else
          pixel_median = it.getValue();

        pixel_median = std::max( pixel_median, dstBandAllowedMin );
        pixel_median = std::min( pixel_median, dstBandAllowedMax );
        dstRaster.setValue(C, R, pixel_median, dstBandIdx);

        if( useProgress )
        {
          task->setCurrentStep( static_cast<int>(R / srcRaster.getNumberOfRows()) );
          if( !task->isActive() ) return false;
        }

        ++it;
      }

      return true;
    }

    bool Filter::DilationFilter( const te::rst::Raster& srcRaster,
      const unsigned int srcBandIdx, te::rst::Raster& dstRaster,
      const unsigned int dstBandIdx, const bool useProgress )
    {
      TERP_DEBUG_TRUE_OR_THROW( srcRaster.getNumberOfColumns() ==
        dstRaster.getNumberOfColumns(), "Internal error" );
      TERP_DEBUG_TRUE_OR_THROW( srcRaster.getNumberOfRows() ==
        dstRaster.getNumberOfRows(), "Internal error" );
      TERP_DEBUG_TRUE_OR_THROW( srcBandIdx < srcRaster.getNumberOfBands(),
        "Internal error" );
      TERP_DEBUG_TRUE_OR_THROW( dstBandIdx < dstRaster.getNumberOfBands(),
        "Internal error" );
      
      const te::rst::Band& srcBand = *srcRaster.getBand(srcBandIdx);
      te::rst::Band& dstBand = *dstRaster.getBand(dstBandIdx);            
      const unsigned int rasterNCols = srcRaster.getNumberOfColumns(); 
      
      setMorfConvMatrix();
     
      ResetConvBuffer( m_convMatrixLines, rasterNCols);
            
      const unsigned int windowHRadius = m_convMatrixLines / 2;
      const unsigned int windowWRadius = m_convMatrixColumns / 2;
      const unsigned int rasterNRows = srcRaster.getNumberOfRows();
      const unsigned int rasterVirtualRowsBound = rasterNRows + windowHRadius;
      const double srcNoDataValue = srcBand.getProperty()->m_noDataValue;
      const double dstNoDataValue = dstBand.getProperty()->m_noDataValue;
      const unsigned int wCenterLastValidCol = rasterNCols - windowWRadius - 1;
      const unsigned int lastBufferRowIdx = m_convMatrixLines -1;
      unsigned int rasterCol = 0;
      double dstValue = 0;
      unsigned int convBufferCol = 0;
      unsigned int convBufferStartCol = 0;
      unsigned int convBufferRow = 0;
      unsigned int convBufferColsBound = 0;
      unsigned int realRasterRow = 0;
      
      double dstBandAllowedMin = 0;
      double dstBandAllowedMax = 0;
      te::rp::GetDataTypeRange( dstBand.getProperty()->getType(),
        dstBandAllowedMin, dstBandAllowedMax );      

      std::unique_ptr< te::common::TaskProgress > task;
      if( useProgress )
      {
        task.reset( new te::common::TaskProgress(TE_TR("Dilation Filter"),
          te::common::TaskProgress::UNDEFINED, rasterVirtualRowsBound ) );
      }

      for( unsigned int virtualRasterRow = 0 ; virtualRasterRow < rasterVirtualRowsBound ; ++virtualRasterRow )
      {
        if( virtualRasterRow < windowHRadius )
        { // copy data from the last bufer line to output
          UpdateConvBuffer( srcBand, virtualRasterRow );
          
          for( rasterCol = 0 ; rasterCol < rasterNCols ; ++rasterCol )
          {
            dstValue = m_convBuffer[ lastBufferRowIdx ][ rasterCol ];
            
            if( dstValue == srcNoDataValue )
            {
              dstBand.setValue( rasterCol, virtualRasterRow, dstNoDataValue );
            }
            else
            {
              dstValue = std::min( dstValue, dstBandAllowedMax );
              dstBand.setValue( rasterCol, virtualRasterRow, 
                std::max( dstValue, dstBandAllowedMin ) );
            }
          }
        }
        else if( virtualRasterRow < lastBufferRowIdx )
        {
          UpdateConvBuffer( srcBand, virtualRasterRow );
        }            
        else if( virtualRasterRow < rasterNRows )
        {
          UpdateConvBuffer( srcBand, virtualRasterRow );
          
          for( rasterCol = 0 ; rasterCol < rasterNCols ; ++rasterCol )
          {
            dstValue = m_convBuffer[ windowHRadius ][ rasterCol ];
            
            if( dstValue == srcNoDataValue )
            {
              dstBand.setValue( rasterCol, virtualRasterRow - windowHRadius, 
                dstNoDataValue );
            }
            else
            {
              if(
                  ( rasterCol < windowWRadius )
                  ||
                  ( rasterCol > wCenterLastValidCol )
                )
              { 
                dstValue = std::min( dstValue, dstBandAllowedMax );
                dstBand.setValue( rasterCol, virtualRasterRow - windowHRadius, 
                  std::max( dstValue, dstBandAllowedMin ) );
              }
              else
              {
                convBufferStartCol = rasterCol - windowWRadius;
                convBufferColsBound = rasterCol + windowWRadius + 1;
                
                for( convBufferRow = 0 ; convBufferRow < m_convMatrixLines ;
                  ++convBufferRow )
                {
                  for( convBufferCol = convBufferStartCol ; convBufferCol 
                    < convBufferColsBound ; ++convBufferCol )
                  {
                    if( 
                        (
                          m_convMatrix[ convBufferRow ][ convBufferCol -
                            convBufferStartCol ] 
                          != 
                          0 
                        )
                        &&
                        (
                          m_convBuffer[ convBufferRow ][ convBufferCol ]
                          !=
                          srcNoDataValue
                        )
                      )
                    {
                      dstValue = std::max( dstValue, m_convBuffer[ 
                        convBufferRow ][ convBufferCol ] );
                    }
                  }
                }
                
                dstValue = std::min( dstValue, dstBandAllowedMax );
                dstBand.setValue( rasterCol, virtualRasterRow - windowHRadius, 
                  std::max( dstValue, dstBandAllowedMin ) );
              }
            }
          }
        }      
        else // ( virtualRasterRow >= rasterNRows )
        { // just copy data from the bufer to output
          realRasterRow = ( virtualRasterRow - rasterNRows );
          convBufferRow = lastBufferRowIdx - realRasterRow;
          realRasterRow = rasterNRows - 1 - realRasterRow;
          
          for( rasterCol = 0 ; rasterCol < rasterNCols ; ++rasterCol )
          {
            dstValue = m_convBuffer[ convBufferRow ][ rasterCol ];
            
            if( dstValue == srcNoDataValue )
            {
              dstBand.setValue( rasterCol, realRasterRow, dstNoDataValue );
            }
            else
            {
              dstValue = std::min( dstValue, dstBandAllowedMax );
              dstBand.setValue( rasterCol, realRasterRow, 
                std::max( dstValue, dstBandAllowedMin ) );
            }
          }
        }

          
        if( useProgress )
        {
          task->pulse();
          if( !task->isActive() ) return false;
        }          
      }

      return true;
    }

    bool Filter::ErosionFilter( const te::rst::Raster& srcRaster,
      const unsigned int srcBandIdx, te::rst::Raster& dstRaster,
      const unsigned int dstBandIdx, const bool useProgress )
    {
      TERP_DEBUG_TRUE_OR_THROW( srcRaster.getNumberOfColumns() ==
        dstRaster.getNumberOfColumns(), "Internal error" );
      TERP_DEBUG_TRUE_OR_THROW( srcRaster.getNumberOfRows() ==
        dstRaster.getNumberOfRows(), "Internal error" );
      TERP_DEBUG_TRUE_OR_THROW( srcBandIdx < srcRaster.getNumberOfBands(),
        "Internal error" );
      TERP_DEBUG_TRUE_OR_THROW( dstBandIdx < dstRaster.getNumberOfBands(),
        "Internal error" );
      
      const te::rst::Band& srcBand = *srcRaster.getBand(srcBandIdx);
      te::rst::Band& dstBand = *dstRaster.getBand(dstBandIdx);            
      const unsigned int rasterNCols = srcRaster.getNumberOfColumns(); 
      
      setMorfConvMatrix();
     
      ResetConvBuffer( m_convMatrixLines, rasterNCols);
            
      const unsigned int windowHRadius = m_convMatrixLines / 2;
      const unsigned int windowWRadius = m_convMatrixColumns / 2;
      const unsigned int rasterNRows = srcRaster.getNumberOfRows();
      const unsigned int rasterVirtualRowsBound = rasterNRows + windowHRadius;
      const double srcNoDataValue = srcBand.getProperty()->m_noDataValue;
      const double dstNoDataValue = dstBand.getProperty()->m_noDataValue;
      const unsigned int wCenterLastValidCol = rasterNCols - windowWRadius - 1;
      const unsigned int lastBufferRowIdx = m_convMatrixLines -1;
      unsigned int rasterCol = 0;
      double dstValue = 0;
      unsigned int convBufferCol = 0;
      unsigned int convBufferStartCol = 0;
      unsigned int convBufferRow = 0;
      unsigned int convBufferColsBound = 0;
      unsigned int realRasterRow = 0;
      
      double dstBandAllowedMin = 0;
      double dstBandAllowedMax = 0;
      te::rp::GetDataTypeRange( dstBand.getProperty()->getType(),
        dstBandAllowedMin, dstBandAllowedMax );      

      std::unique_ptr< te::common::TaskProgress > task;
      if( useProgress )
      {
        task.reset( new te::common::TaskProgress(TE_TR("Erosion Filter"),
          te::common::TaskProgress::UNDEFINED, rasterVirtualRowsBound ) );
      }

      for( unsigned int virtualRasterRow = 0 ; virtualRasterRow < rasterVirtualRowsBound ; ++virtualRasterRow )
      {
        if( virtualRasterRow < windowHRadius )
        { // copy data from the last bufer line to output
          UpdateConvBuffer( srcBand, virtualRasterRow );
          
          for( rasterCol = 0 ; rasterCol < rasterNCols ; ++rasterCol )
          {
            dstValue = m_convBuffer[ lastBufferRowIdx ][ rasterCol ];
            
            if( dstValue == srcNoDataValue )
            {
              dstBand.setValue( rasterCol, virtualRasterRow, dstNoDataValue );
            }
            else
            {
              dstValue = std::min( dstValue, dstBandAllowedMax );
              dstBand.setValue( rasterCol, virtualRasterRow, 
                std::max( dstValue, dstBandAllowedMin ) );
            }
          }
        }
        else if( virtualRasterRow < lastBufferRowIdx )
        {
          UpdateConvBuffer( srcBand, virtualRasterRow );
        }            
        else if( virtualRasterRow < rasterNRows )
        {
          UpdateConvBuffer( srcBand, virtualRasterRow );
          
          for( rasterCol = 0 ; rasterCol < rasterNCols ; ++rasterCol )
          {
            dstValue = m_convBuffer[ windowHRadius ][ rasterCol ];
            
            if( dstValue == srcNoDataValue )
            {
              dstBand.setValue( rasterCol, virtualRasterRow - windowHRadius, 
                dstNoDataValue );
            }
            else
            {
              if(
                  ( rasterCol < windowWRadius )
                  ||
                  ( rasterCol > wCenterLastValidCol )
                )
              { 
                dstValue = std::min( dstValue, dstBandAllowedMax );
                dstBand.setValue( rasterCol, virtualRasterRow - windowHRadius, 
                  std::max( dstValue, dstBandAllowedMin ) );
              }
              else
              {
                convBufferStartCol = rasterCol - windowWRadius;
                convBufferColsBound = rasterCol + windowWRadius + 1;
                
                for( convBufferRow = 0 ; convBufferRow < m_convMatrixLines ;
                  ++convBufferRow )
                {
                  for( convBufferCol = convBufferStartCol ; convBufferCol 
                    < convBufferColsBound ; ++convBufferCol )
                  {
                    if( 
                        (
                          m_convMatrix[ convBufferRow ][ convBufferCol -
                            convBufferStartCol ] 
                          != 
                          0 
                        )
                        &&
                        (
                          m_convBuffer[ convBufferRow ][ convBufferCol ]
                          !=
                          srcNoDataValue
                        )
                      )
                    {
                      dstValue = std::min( dstValue, m_convBuffer[ 
                        convBufferRow ][ convBufferCol ] );
                    }
                  }
                }
                
                dstValue = std::min( dstValue, dstBandAllowedMax );
                dstBand.setValue( rasterCol, virtualRasterRow - windowHRadius, 
                  std::max( dstValue, dstBandAllowedMin ) );
              }
            }
          }
        }      
        else // ( virtualRasterRow >= rasterNRows )
        { // just copy data from the bufer to output
          realRasterRow = ( virtualRasterRow - rasterNRows );
          convBufferRow = lastBufferRowIdx - realRasterRow;
          realRasterRow = rasterNRows - 1 - realRasterRow;
          
          for( rasterCol = 0 ; rasterCol < rasterNCols ; ++rasterCol )
          {
            dstValue = m_convBuffer[ convBufferRow ][ rasterCol ];
            
            if( dstValue == srcNoDataValue )
            {
              dstBand.setValue( rasterCol, realRasterRow, dstNoDataValue );
            }
            else
            {
              dstValue = std::min( dstValue, dstBandAllowedMax );
              dstBand.setValue( rasterCol, realRasterRow, 
                std::max( dstValue, dstBandAllowedMin ) );
            }
          }
        }

          
        if( useProgress )
        {
          task->pulse();
          if( !task->isActive() ) return false;
        }          
      }

      return true;
    }

    bool Filter::UserDefinedFilter( const te::rst::Raster& srcRaster,
      const unsigned int srcBandIdx, te::rst::Raster& dstRaster,
      const unsigned int dstBandIdx, const bool useProgress )
    {
      TERP_DEBUG_TRUE_OR_THROW( srcRaster.getNumberOfColumns() ==
        dstRaster.getNumberOfColumns(), "Internal error" );
      TERP_DEBUG_TRUE_OR_THROW( srcRaster.getNumberOfRows() ==
        dstRaster.getNumberOfRows(), "Internal error" );
      TERP_DEBUG_TRUE_OR_THROW( srcBandIdx < srcRaster.getNumberOfBands(),
        "Internal error" );
      TERP_DEBUG_TRUE_OR_THROW( dstBandIdx < dstRaster.getNumberOfBands(),
        "Internal error" );

      const boost::numeric::ublas::matrix<double>& window =
        m_inputParameters.m_window;

      const unsigned int windowHeight = static_cast<unsigned int>(window.size1());
      const unsigned int windowWidth = static_cast<unsigned int>(window.size2());
      const unsigned int windowHRadius = windowHeight / 2;
      const unsigned int windowWRadius = windowWidth / 2;
      const unsigned int rowsNmb = srcRaster.getNumberOfRows();
      const unsigned int colsNmb = srcRaster.getNumberOfColumns();
      const unsigned int windowCenterRowBound = rowsNmb - windowHRadius;
      const unsigned int windowCenterColBound = colsNmb - windowWRadius;
      const te::rst::Band& srcBand = *srcRaster.getBand(srcBandIdx);
      te::rst::Band& dstBand = *dstRaster.getBand(dstBandIdx);
      const double srcNoDataValue = srcBand.getProperty()->m_noDataValue;

      double dstBandAllowedMin = 0;
      double dstBandAllowedMax = 0;
      te::rp::GetDataTypeRange( dstBand.getProperty()->getType(),
        dstBandAllowedMin, dstBandAllowedMax );

      unsigned int centerRow = 0;
      unsigned int centerCol = 0;
      double srcValue = 0;
      double dstValue = 0;
      unsigned int row = 0;
      unsigned int col = 0;
      unsigned int startRow = 0;
      unsigned int startCol = 0;
      unsigned int rowBound = 0;
      unsigned int colBound = 0;

      std::unique_ptr< te::common::TaskProgress > task;
      if( useProgress )
      {
        task.reset( new te::common::TaskProgress(TE_TR("User Defined Filter"),
          te::common::TaskProgress::UNDEFINED, static_cast<int>(rowsNmb) ) );
      }

      for( centerRow = 0 ; centerRow < rowsNmb ; ++centerRow )
      {
        if( useProgress )
        {
          task->setCurrentStep( static_cast<int>(centerRow) );
          if( !task->isActive() ) return false;
        }

        for( centerCol = 0 ; centerCol < colsNmb ; ++centerCol )
        {
          if(
              ( centerRow < windowHRadius )
              ||
              ( centerRow >= windowCenterRowBound )
              ||
              ( centerCol < windowWRadius )
              ||
              ( centerCol >= windowCenterColBound )
            )
          {
            srcBand.getValue( centerCol, centerRow, srcValue );
            dstBand.setValue( centerCol, centerRow, srcValue );
          }
          else
          {
            startRow = centerRow - windowHRadius;
            startCol = centerCol - windowWRadius;
            rowBound = centerRow + windowHRadius + 1;
            colBound = centerCol + windowWRadius + 1;
            dstValue = 0;

            for( row = startRow ; row < rowBound ; ++row )
            {
              for( col = startCol ; col < colBound ; ++col )
              {
                srcBand.getValue( col, row, srcValue );

                if( srcValue != srcNoDataValue )
                {
                  dstValue += srcValue * window( row - startRow,
                    col - startCol );
                }
              }
            }

            dstValue = std::max( dstValue, dstBandAllowedMin );
            dstValue = std::min( dstValue, dstBandAllowedMax );
            dstBand.setValue( centerCol, centerRow, dstValue );
          }
        }
      }

      return true;
    }

    bool Filter::OrderFunction(double i, double j)
    {
      return (i < j);
    }

    void Filter::ResetConvBuffer(unsigned int lines, unsigned int columns)
    {
      if (m_convBuffer != nullptr) {
        for (unsigned int line = 0; line < m_convBufferLines; ++line) {
          delete[] m_convBuffer[line];
        }

        delete[] m_convBuffer;

        m_convBuffer = nullptr;
        m_convBufferLines = 0;
        m_convBufferColumns = 0;
      }

      if ((lines > 0) && (columns > 0)) {
        m_convBuffer = new double*[lines];

        TERP_DEBUG_TRUE_OR_THROW(m_convBuffer != nullptr, "Memory allocation error");

        for (unsigned int line = 0; line < lines; ++line) {
          m_convBuffer[line] = new double[columns];

          TERP_DEBUG_TRUE_OR_THROW(m_convBuffer[line] != nullptr, "Memory allocation error");
        }

        m_convBufferLines = lines;
        m_convBufferColumns = columns;
      }
    }

    void Filter::UpdateConvBuffer(const te::rst::Band& inRaster, unsigned int line)
    {
      TERP_DEBUG_TRUE_OR_THROW( inRaster.getRaster()->getNumberOfRows() > line,
        "Trying to get a non existent line from raster");
      TERP_DEBUG_TRUE_OR_THROW( inRaster.getRaster()->getNumberOfColumns() ==
        m_convBufferColumns,
        "Buffer columns number not equal to raster columns");
      TERP_DEBUG_TRUE_OR_THROW( m_convBuffer != 0,
        "Invalid convolution Buffer");
      
      unsigned int bufPos = 1;
      double* firstBufferLinePtr = m_convBuffer[ 0 ];      
      for ( ; bufPos < m_convBufferLines; ++bufPos) 
      {
        m_convBuffer[bufPos - 1] = m_convBuffer[bufPos];
      }      
      m_convBuffer[ m_convBufferLines - 1 ] = firstBufferLinePtr;
      
      const unsigned int convBufferLastLine = m_convBufferLines - 1;
      for( bufPos = 0; bufPos < m_convBufferColumns; ++bufPos) 
      {
        inRaster.getValue(bufPos, line, m_convBuffer[convBufferLastLine][bufPos]);
      }
    }
    
    void Filter::ResetConvMatrix(unsigned int lines, unsigned int columns)
    {
      if (m_convMatrix != nullptr) {
        for (unsigned int line = 0; line < m_convMatrixLines; ++line) {
          delete[] m_convMatrix[line];
        }

        delete[] m_convMatrix;

        m_convMatrix = nullptr;
        m_convMatrixLines = 0;
        m_convMatrixColumns = 0;
      }

      if ((lines > 0) && (columns > 0)) 
      {
        m_convMatrix = new double*[lines];

        TERP_DEBUG_TRUE_OR_THROW(m_convMatrix != nullptr, "Memory allocation error");

        for (unsigned int line = 0; line < lines; ++line) {
          m_convMatrix[line] = new double[columns];

          TERP_DEBUG_TRUE_OR_THROW(m_convMatrix[line] != nullptr, "Memory allocation error");
        }

        m_convMatrixLines = lines;
        m_convMatrixColumns = columns;
      }
    }
    
    void Filter::setMorfConvMatrix()
    {
      if( m_inputParameters.m_window.size1() != 0 )
      {
        ResetConvMatrix( m_inputParameters.m_window.size1(),
          m_inputParameters.m_window.size2() );
        
        unsigned int col = 0;
        for( unsigned int line = 0 ; line < m_convMatrixLines ; ++line )
        {
          for( col = 0 ; col < m_convMatrixColumns ; ++col )
          {
            m_convMatrix[ line ][ col ] = m_inputParameters.m_window( line, col );
          }
        }
      }
      else
      {
        ResetConvMatrix( 3, 3 );
        
        unsigned int col = 0;
        for( unsigned int line = 0 ; line < 3 ; ++line )
        {
          for( col = 0 ; col < 3 ; ++col )
          {
            m_convMatrix[ line ][ col ] = 1;
          }
        }
      }
    }
    
  } // end namespace rp
}   // end namespace te

