/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/rp/GeoMosaic.cpp
  \brief Create a mosaic from a set of geo-referenced rasters.
*/

#include "GeoMosaic.h"

#include "Macros.h"
#include "Functions.h"
#include "AlgorithmParametersSerializer.h"
#include "../raster/Interpolator.h"
#include "../raster/Enums.h"
#include "../raster/RasterFactory.h"
#include "../raster/Grid.h"
#include "../raster/Band.h"
#include "../raster/BandProperty.h"
#include "../raster/PositionIterator.h"
#include "../raster/Utils.h"
#include "../raster/Reprojection.h"
#include "../memory/CachedRaster.h"
#include "../geometry/Envelope.h"
#include "../geometry/GTFactory.h"
#include "../geometry/Polygon.h"
#include "../geometry/LinearRing.h"
#include "../srs/Converter.h"
#include "../common/progress/TaskProgress.h"
#include "../common/MathUtils.h"

#include <boost/shared_array.hpp>
#include <boost/filesystem.hpp>
#include <boost/lexical_cast.hpp>

#include <climits>
#include <cfloat>
#include <cmath>
#include <memory>
#include <cstdio>
#include <map>

namespace te
{
  namespace rp
  {

    GeoMosaic::InputParameters::InputParameters()
    {
      reset();
    }

    GeoMosaic::InputParameters::InputParameters( const InputParameters& other )
    {
      reset();
      operator=( other );
    }

    GeoMosaic::InputParameters::~InputParameters()
    {
      reset();
    }

    void GeoMosaic::InputParameters::reset() _NOEXCEPT_OP(false)
    {
      m_feederRasterPtr = nullptr;
      m_inputRastersBands.clear();
      m_interpMethod = te::rst::NearestNeighbor;
      m_noDataValue = 0.0;
      m_forceInputNoDataValue = false;
      m_blendMethod = te::rp::blender::NoBlendMethod;
      m_equalizationType = CenterEqType;
      m_useRasterCache = true;
      m_enableProgress = false;
      m_enableMultiThread = true;
      m_enableAutoValidAreaDetect = true;
      m_mosaicTargetMeans.clear();
      m_mosaicTargetStdDevs.clear();
      m_blenderRulePtr = nullptr;
      m_abortOnFirstError = true;
      m_skipInsideRasters = false;
    }

    const GeoMosaic::InputParameters& GeoMosaic::InputParameters::operator=(
      const GeoMosaic::InputParameters& params )
    {
      reset();

      m_feederRasterPtr = params.m_feederRasterPtr;
      m_inputRastersBands = params.m_inputRastersBands;
      m_interpMethod = params.m_interpMethod;
      m_noDataValue = params.m_noDataValue;
      m_forceInputNoDataValue = params.m_forceInputNoDataValue;
      m_blendMethod = params.m_blendMethod;
      m_equalizationType = params.m_equalizationType;
      m_useRasterCache = params.m_useRasterCache;
      m_enableProgress = params.m_enableProgress;
      m_enableMultiThread = params.m_enableMultiThread;
      m_enableAutoValidAreaDetect = params.m_enableAutoValidAreaDetect;
      m_mosaicTargetMeans = params.m_mosaicTargetMeans;
      m_mosaicTargetStdDevs = params.m_mosaicTargetStdDevs;
      m_blenderRulePtr = params.m_blenderRulePtr;
      m_abortOnFirstError = params.m_abortOnFirstError;
      m_skipInsideRasters = params.m_skipInsideRasters;

      return *this;
    }

    te::common::AbstractParameters* GeoMosaic::InputParameters::clone() const
    {
      return new InputParameters( *this );
    }

    bool GeoMosaic::InputParameters::serialize( AlgorithmParametersSerializer& serializer ) const
    {
      TERP_TRUE_OR_RETURN_FALSE( AlgorithmInputParameters::serialize( serializer ),
        "Serializarion error" );

      serializer.setAlgorithmName( "GeoMosaic" );

      std::vector< std::string > rasterNames;
      m_feederRasterPtr->getRasterNames( rasterNames );
      TERP_TRUE_OR_RETURN_FALSE( serializer.addMultivaluedParameter( "inputRasters",
        rasterNames ), "Error adding parameter inputRasters" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addMultivaluedParameters( "inputRastersBands",
        m_inputRastersBands ), "Error adding parameter inputRastersBands" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "interpMethod",
        m_interpMethod ), "Error adding parameter interpMethod" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "noDataValue",
        m_noDataValue ), "Error adding parameter noDataValue" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "forceInputNoDataValue",
        m_forceInputNoDataValue ), "Error adding parameter forceInputNoDataValue" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "blendMethod",
        m_blendMethod ), "Error adding parameter blendMethod" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "equalizationType",
        m_equalizationType ), "Error adding parameter equalizationType" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "enableAutoValidAreaDetect",
        m_enableAutoValidAreaDetect ), "Error adding parameter enableAutoValidAreaDetect" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addMultivaluedParameter( "mosaicTargetMeans",
        m_mosaicTargetMeans ), "Error adding parameter mosaicTargetMeans" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addMultivaluedParameter( "mosaicTargetStdDevs",
        m_mosaicTargetStdDevs ), "Error adding parameter mosaicTargetStdDevs" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "abortOnFirstError",
        m_abortOnFirstError ), "Error adding parameter abortOnFirstError" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "skipInsideRasters",
        m_skipInsideRasters ), "Error adding parameter skipInsideRasters" );
      
      return true;
    }
    
    void GeoMosaic::InputParameters::getEqTypesInfo( 
      std::vector< std::pair< EqType, std::string > >& eqTypesInfo  )
    {
      eqTypesInfo.clear();
      eqTypesInfo.push_back( std::pair< EqType, std::string >( NoEqType,
        "NoEqType" ) );
      eqTypesInfo.push_back( std::pair< EqType, std::string >( CenterEqType,
        "CenterEqType" ) );
      eqTypesInfo.push_back( std::pair< EqType, std::string >( TargetEqtype,
        "TargetEqtype" ) );
      eqTypesInfo.push_back( std::pair< EqType, std::string >( LocalEqType,
        "LocalEqType" ) );
      eqTypesInfo.push_back( std::pair< EqType, std::string >( OverLapEqType,
        "OverLapEqType" ) );      
    }

    GeoMosaic::OutputParameters::OutputParameters()
    {
      reset();
    }

    GeoMosaic::OutputParameters::OutputParameters( const OutputParameters& other )
    {
      reset();
      operator=( other );
    }

    GeoMosaic::OutputParameters::~OutputParameters()
    {
      reset();
    }

    void GeoMosaic::OutputParameters::reset() _NOEXCEPT_OP(false)
    {
      m_rType.clear();
      m_rInfo.clear();
      m_outputRasterPtr.reset();
    }

    const GeoMosaic::OutputParameters& GeoMosaic::OutputParameters::operator=(
      const GeoMosaic::OutputParameters& params )
    {
      reset();

      m_rType = params.m_rType;
      m_rInfo = params.m_rInfo;
      m_outputRasterPtr.reset();;

      return *this;
    }

    te::common::AbstractParameters* GeoMosaic::OutputParameters::clone() const
    {
      return new OutputParameters( *this );
    }

    GeoMosaic::GeoMosaic()
    {
      reset();
    }

    GeoMosaic::~GeoMosaic() = default;

    bool GeoMosaic::execute( AlgorithmOutputParameters& outputParams )
      _NOEXCEPT_OP(false)
    {
      if( ! m_isInitialized ) return false;
      
      GeoMosaic::OutputParameters* outParamsPtr = dynamic_cast<
        GeoMosaic::OutputParameters* >( &outputParams );
      TERP_INSTANCE_TRUE_OR_THROW( outParamsPtr, "Invalid paramters" );
      
      assert( m_inputParameters.m_feederRasterPtr->getObjsCount() > 1 );
      
      // First pass: getting global mosaic info

      double mosaicXResolution = 0.0;
      double mosaicYResolution = 0.0;
      double mosaicLLX = DBL_MAX; // world coords
      double mosaicLLY = DBL_MAX; // world coords
      double mosaicURX = -1.0 * DBL_MAX; // world coords
      double mosaicURY = -1.0 * DBL_MAX; // world coords
      int mosaicSRID = 0;     
      te::rst::BandProperty mosaicBaseBandProperties( 0, 0, "" );

      // a map ordering raster feeder indexes to the respective rasters bounding box (under the mosaic world coords).
      std::map< unsigned int, std::unique_ptr< te::gm::Polygon > > rastersBBoxesPolsMap;               
      
      // a map ordereding the distance to the mosaic center to each raster feeder index
      std::multimap< double, unsigned int > centerDist2RasterIdxMap;

      {
        std::unique_ptr< te::common::TaskProgress > progressPtr;
        if( m_inputParameters.m_enableProgress )
        {
          progressPtr.reset( new te::common::TaskProgress );
          
          progressPtr->setTotalSteps( static_cast<int>( 
            m_inputParameters.m_feederRasterPtr->getObjsCount()) );
          
          progressPtr->setMessage( "Mosaic: reading input rasters info" );
        }        
        
        // a map from input rasters SRIDS and their frequencies
        std::map< int, unsigned int > sridsFreq;
        
        // a map from input rasters SRIDS and their best resolutions
        std::map< int, std::pair< double, double > > srid2ResolutionsMap;
        
        {
          te::rst::Raster const* inputRasterPtr = nullptr;
          unsigned int inputRasterIdx = 0;          
          bool isRasterValid = false;

          m_inputParameters.m_feederRasterPtr->reset();
          while( ( inputRasterPtr = m_inputParameters.m_feederRasterPtr->getCurrentObj() ) )
          {
            inputRasterIdx = m_inputParameters.m_feederRasterPtr->getCurrentOffset();
            
            // checking if the current raster is valid 
            
            isRasterValid = true;
            
            if( ( inputRasterPtr->getAccessPolicy() & te::common::RAccess ) == 0 )
            {
              isRasterValid = false;
            }
            
            if( inputRasterPtr->getSRID() == 0 )
            {
              isRasterValid = false;
            }

            for( std::vector< unsigned int >::size_type inputRastersBandsIdx = 0 ;
              inputRastersBandsIdx <
              m_inputParameters.m_inputRastersBands[ inputRasterIdx ].size() ;
              ++inputRastersBandsIdx )
            {
              const unsigned int& currBand =
                m_inputParameters.m_inputRastersBands[ inputRasterIdx ][ inputRastersBandsIdx ];

              if( currBand >= inputRasterPtr->getNumberOfBands() )
              {
                isRasterValid = false;
                break;
              }
            }
            
            if( isRasterValid )
            {
              ++sridsFreq[ inputRasterPtr->getSRID() ];  
            
              rastersBBoxesPolsMap[ inputRasterIdx ] =
                std::unique_ptr< te::gm::Polygon >(           
                static_cast<te::gm::Polygon*>(te::gm::GetGeomFromEnvelope(
                inputRasterPtr->getGrid()->getExtent(), 
                inputRasterPtr->getGrid()->getSRID() ) ) );
                
              std::map< int, std::pair< double, double > >::iterator 
                srid2ResolutionsMapIt = srid2ResolutionsMap.find( 
                inputRasterPtr->getGrid()->getSRID() );
              if( 
                  srid2ResolutionsMapIt
                  ==
                  srid2ResolutionsMap.end()
                )
              {
                srid2ResolutionsMap[ inputRasterPtr->getGrid()->getSRID() ] = 
                  std::pair< double, double >(
                  inputRasterPtr->getGrid()->getResolutionX(),
                  inputRasterPtr->getGrid()->getResolutionY() );
              }
              else
              {
                if( 
                    ( 
                      srid2ResolutionsMapIt->second.first
                      <= 
                      inputRasterPtr->getGrid()->getResolutionX()
                    )
                    &&
                    ( 
                      srid2ResolutionsMapIt->second.second
                      <= 
                      inputRasterPtr->getGrid()->getResolutionY()
                    )
                  )
                {
                  srid2ResolutionsMapIt->second.first = 
                    inputRasterPtr->getGrid()->getResolutionX();
                  srid2ResolutionsMapIt->second.second =
                    inputRasterPtr->getGrid()->getResolutionY();
                }
              }
            }
            else
            {
              const std::string message = "Invalid input raster : " +
                std::to_string( inputRasterIdx );
              if( m_inputParameters.m_abortOnFirstError )
              {
                TERP_INSTANCE_LOG_AND_RETURN_FALSE( message );
              }
              else
              {
                TERP_INSTANCE_LOGWARN( message );
              }
            }

            // move to the next raster
            
            if( m_inputParameters.m_enableProgress )
            {
              progressPtr->pulse();
              if( ! progressPtr->isActive() ) return false;
            }                  

            m_inputParameters.m_feederRasterPtr->moveNext();
          }
        }
        
        // defining the mosaic SRID
        
        {
          std::map< int, unsigned int >::const_iterator sridsFreqIt = 
            sridsFreq.begin();
          const std::map< int, unsigned int >::const_iterator sridsFreqItEnd = 
            sridsFreq.end();
          unsigned highestSRIDFreq = 0;
          
          
          while( sridsFreqIt != sridsFreqItEnd )
          {
            if( highestSRIDFreq < sridsFreqIt->second )
            {
              highestSRIDFreq = sridsFreqIt->second;
              mosaicSRID = sridsFreqIt->first;
            }
            
            ++sridsFreqIt;
          }
        }
        
        // defining mosaic resolution
        
        mosaicXResolution = srid2ResolutionsMap[ mosaicSRID ].first;
        mosaicYResolution = srid2ResolutionsMap[ mosaicSRID ].second;
      
        // mosaic area
        
        {
          te::gm::Polygon* currPolPtr = 0;
            
          std::map< unsigned int, std::unique_ptr< te::gm::Polygon > >::const_iterator 
            rastersBBoxesPolsMapIt = rastersBBoxesPolsMap.begin();
          const std::map< unsigned int, std::unique_ptr< te::gm::Polygon > >::const_iterator 
            rastersBBoxesPolsMapItEnd = rastersBBoxesPolsMap.end();
          
          while( rastersBBoxesPolsMapIt != rastersBBoxesPolsMapItEnd )
          {
            currPolPtr = rastersBBoxesPolsMapIt->second.get();
            
            currPolPtr->transform( mosaicSRID );
            
            mosaicLLX = std::min( mosaicLLX, currPolPtr->getMBR()->getLowerLeftX() );
            mosaicLLY = std::min( mosaicLLY, currPolPtr->getMBR()->getLowerLeftY() );

            mosaicURX = std::max( mosaicURX, currPolPtr->getMBR()->getUpperRightX() );
            mosaicURY = std::max( mosaicURY, currPolPtr->getMBR()->getUpperRightY() );                  
            
            ++rastersBBoxesPolsMapIt;
          }
        }
        
        // updating center distances map
        
        {
          const double mosaicCenterX = ( mosaicURX + mosaicLLX ) / 2.0;
          const double mosaicCenterY = ( mosaicURY + mosaicLLY ) / 2.0;
          double currDist = 0;
          double currXDist = 0;
          double currYDist = 0;
          te::gm::Coord2D currCenter;
          te::gm::Polygon* currPolPtr = 0;
          
          std::map< unsigned int, std::unique_ptr< te::gm::Polygon > >::const_iterator 
            rastersBBoxesPolsMapIt = rastersBBoxesPolsMap.begin();
          const std::map< unsigned int, std::unique_ptr< te::gm::Polygon > >::const_iterator 
            rastersBBoxesPolsMapItEnd = rastersBBoxesPolsMap.end();          
          
          while( rastersBBoxesPolsMapIt != rastersBBoxesPolsMapItEnd )
          {
            currPolPtr = rastersBBoxesPolsMapIt->second.get();
            
            currCenter = currPolPtr->getMBR()->getCenter();
            currXDist = currCenter.getX() - mosaicCenterX;
            currYDist = currCenter.getY() - mosaicCenterY;
            currDist = std::sqrt( ( currXDist * currXDist ) + ( currYDist *
              currYDist ) );
            
            centerDist2RasterIdxMap.insert( std::pair< double, unsigned int >(
              currDist, rastersBBoxesPolsMapIt->first ) );
            
            ++rastersBBoxesPolsMapIt;
          }
        }
        
        // base band props
        // using the mosac center raster as reference
        
        {
          unsigned int centerInputRasterIdx = centerDist2RasterIdxMap.begin()->second;
          if( ! m_inputParameters.m_feederRasterPtr->moveTo( centerInputRasterIdx ) )
          {
            TERP_INSTANCE_LOG_AND_RETURN_FALSE( "Error opening input raster ");
          }
          
          te::rst::Raster const* centerInputRasterPtr = 
            m_inputParameters.m_feederRasterPtr->getCurrentObj();
          
          mosaicBaseBandProperties = *centerInputRasterPtr->getBand( 
            m_inputParameters.m_inputRastersBands[ centerInputRasterIdx ][ 0 ] )->getProperty();
          if( m_inputParameters.m_forceInputNoDataValue )
          {
            mosaicBaseBandProperties.m_noDataValue = m_inputParameters.m_noDataValue;
          }
        }
        
        TERP_INSTANCE_TRUE_OR_RETURN_FALSE( !rastersBBoxesPolsMap.empty(),
          "No input rasters to process" );        
      }

      // creating the output raster
      
      te::rst::Raster* outputRasterPtr = nullptr;

      {
        te::rst::Grid* outputGrid = new te::rst::Grid( mosaicXResolution,
          mosaicYResolution,  new te::gm::Envelope( mosaicLLX, mosaicLLY, mosaicURX,
          mosaicURY ), mosaicSRID );
        
        te::rst::BandProperty bandProperty( mosaicBaseBandProperties );
        bandProperty.m_blkh = 1024;
        bandProperty.m_blkw = 1024;
        bandProperty.m_nblocksx =
          ( outputGrid->getNumberOfColumns() / bandProperty.m_blkw )
          +
          (
            ( ( outputGrid->getNumberOfColumns() % bandProperty.m_blkw ) > 0 )
            ? 1 : 0
          );
        bandProperty.m_nblocksy =
          ( outputGrid->getNumberOfRows() / bandProperty.m_blkh )
          +
          (
            ( ( outputGrid->getNumberOfRows() % bandProperty.m_blkh ) > 0 )
            ? 1 : 0
          );
        bandProperty.m_colorInterp = te::rst::GrayIdxCInt;
        
        std::vector< te::rst::BandProperty* > bandsProperties;
        for( std::vector< unsigned int >::size_type bandIdx = 0 ;  bandIdx <
          m_inputParameters.m_inputRastersBands[ 0 ].size() ; ++bandIdx )
        {
          bandsProperties.push_back( new te::rst::BandProperty( bandProperty ) );
        }

        AlgorithmParametersSerializer paramsSerializer;
        TERP_INSTANCE_TRUE_OR_LOG(  m_inputParameters.serialize( paramsSerializer ),
          "Parameter serialization error" );

        std::map <std::string, std::string> metadata;
        paramsSerializer.getMetaData( metadata );

        try
        {
          outParamsPtr->m_outputRasterPtr.reset(
            te::rst::RasterFactory::make(
              outParamsPtr->m_rType,
              outputGrid,
              bandsProperties,
              outParamsPtr->m_rInfo,
              metadata,
              nullptr,
              nullptr ) );
        }
        catch(...)
        {
          outParamsPtr->m_outputRasterPtr.reset();
        }
        TERP_INSTANCE_TRUE_OR_RETURN_FALSE( outParamsPtr->m_outputRasterPtr.get(),
          "Output raster creation error" );
          
        outputRasterPtr = outParamsPtr->m_outputRasterPtr.get();
      }

      std::unique_ptr< te::mem::CachedRaster > outputCachedRasterInstancePtr;
      
      if( m_inputParameters.m_useRasterCache )
      {
        outputCachedRasterInstancePtr.reset( new te::mem::CachedRaster(
          *(outParamsPtr->m_outputRasterPtr.get()), 50, 0 ) );   
          
        outputRasterPtr = outputCachedRasterInstancePtr.get();
      }

      // fill output with no data values

      {
        const unsigned int nBands = static_cast<unsigned int>(outputRasterPtr->getNumberOfBands());
        unsigned int bandIdx = 0;

        for( bandIdx = 0 ; bandIdx < nBands ; ++bandIdx )
        {
          te::rst::Band& outBand = ( *( outputRasterPtr->getBand( bandIdx ) ) );
          
          const std::complex<double > bandNoDataValue( 
            outBand.getProperty()->m_noDataValue,
            outBand.getProperty()->m_noDataValue );
          
          te::rp::FillBand( *outputRasterPtr, bandIdx, bandNoDataValue );
        }
      }
      
      // iterating over the all rasters
      
      std::unique_ptr< te::common::TaskProgress > progressPtr;
      if( m_inputParameters.m_enableProgress )
      {
        progressPtr.reset( new te::common::TaskProgress );
        
        progressPtr->setTotalSteps( static_cast<int>( 
          centerDist2RasterIdxMap.size() ) );
        
        progressPtr->setMessage( "Mosaic: processing input rasters" );
      }              
      
      // the union of all blended rasters valid areas under mosaic SRID
      std::unique_ptr< te::gm::MultiPolygon > mosaicValidAreasUnionPtr;
      
      // blended rasters infos: valid area, means, stddevs ( for each band )
      BlendedRastersInfosT  blendedRastersInfos;

      // blended rasters infos indexing tree
      te::sam::rtree::Index< std::size_t > blendedRastersInfosTree;
      
      std::vector< unsigned int > outputRasterBands;
      for( unsigned int outputRasterBandsIdx = 0 ; outputRasterBandsIdx <
        outputRasterPtr->getNumberOfBands() ; ++outputRasterBandsIdx )
      {
        outputRasterBands.push_back( outputRasterBandsIdx );
      }      
      
      std::vector< double > mosaicTargetMeans;
      std::vector< double > mosaicTargetStdDevs;      

      std::multimap< double, unsigned int >::const_iterator
        centerDist2RasterIdxMapIt = centerDist2RasterIdxMap.begin();
      const std::multimap< double, unsigned int >::const_iterator
        centerDist2RasterIdxMapItEnd = centerDist2RasterIdxMap.end();
      
      while( centerDist2RasterIdxMapIt != centerDist2RasterIdxMapItEnd )
      {
        if( m_inputParameters.m_enableProgress )
        {
          progressPtr->pulse();
          if( ! progressPtr->isActive() ) return false;
        }        
        
        const unsigned int inputRasterIdx = centerDist2RasterIdxMapIt->second;
        
        // is this an inside raster ?
        
        if( 
            m_inputParameters.m_skipInsideRasters
            &&
            ( mosaicValidAreasUnionPtr.get() != nullptr )
          )
        {
          try
          {
            if( 
                mosaicValidAreasUnionPtr->covers( rastersBBoxesPolsMap[ inputRasterIdx ].get() ) 
                ||
                mosaicValidAreasUnionPtr->contains( rastersBBoxesPolsMap[ inputRasterIdx ].get() )
              )
            {
              const std::string message = "Skipping interior raster " +
                std::to_string( inputRasterIdx );                 
              TERP_INSTANCE_LOGWARN( message );
              ++centerDist2RasterIdxMapIt;
              continue;
            }
          }
          catch(...)
          {
            const std::string message = "Unable to find out raster bounding boxes relation " +
              std::to_string( inputRasterIdx );            
            if( m_inputParameters.m_abortOnFirstError )
            {
              TERP_INSTANCE_LOG_AND_RETURN_FALSE( message );
            }
            else
            {
              TERP_INSTANCE_LOGWARN( message );
              ++centerDist2RasterIdxMapIt;
              continue;
            }
          }
        }
        
        if( ! m_inputParameters.m_feederRasterPtr->moveTo( inputRasterIdx ) )
        {
          const std::string message = "Error opening input raster " +
            std::to_string( inputRasterIdx );
          if( m_inputParameters.m_abortOnFirstError )
          {
            TERP_INSTANCE_LOG_AND_RETURN_FALSE( message );
          }
          else
          {
            TERP_INSTANCE_LOGWARN( message );
            ++centerDist2RasterIdxMapIt;
            continue;
          }
        }
        
        te::rst::Raster const* inputRasterPtr = m_inputParameters.m_feederRasterPtr->getCurrentObj();

        // reprojection 
        
        std::unique_ptr< te::rst::Raster > reprojectedInputRasterPtr;
        
        if( outputRasterPtr->getSRID() != inputRasterPtr->getSRID() )
        {
          std::vector< unsigned int > inputRasterBands;
          for( unsigned int bandIdx = 0 ; bandIdx < inputRasterPtr->getNumberOfBands() ;
            ++bandIdx )
          {
            inputRasterBands.push_back( bandIdx );
          }
          
          const bool repResult =
            te::rst::Reproject(
              outputRasterPtr->getSRID(),
              *inputRasterPtr,
              inputRasterBands,
              m_inputParameters.m_interpMethod,
              1, // interpWindowRadius
              0, // firstRow
              0, // firstColumn
              inputRasterPtr->getNumberOfRows(),
              inputRasterPtr->getNumberOfColumns(),
              inputRasterPtr->getNumberOfRows(),
              inputRasterPtr->getNumberOfColumns(),
              0, // maxThreads
              std::map< std::string, std::string >(),
              "EXPANSIBLE",
              reprojectedInputRasterPtr
            );
          
          if( !repResult )
          {
            const std::string message = "Error reprojecting input raster " +
              std::to_string( inputRasterIdx );
            if( m_inputParameters.m_abortOnFirstError )
            {
              TERP_INSTANCE_LOG_AND_RETURN_FALSE( message );
            }
            else
            {
              TERP_INSTANCE_LOGWARN( message );
              ++centerDist2RasterIdxMapIt;
              continue;
            }
          }
           
          inputRasterPtr = reprojectedInputRasterPtr.get();
        }
        
        // Caching issues 
        
        std::unique_ptr< te::mem::CachedRaster > cachedInputRasterPtr;
        
        if(
            m_inputParameters.m_useRasterCache
           &&
           ( reprojectedInputRasterPtr.get() == nullptr ) // Reprojected rasters already have internal cache
          )
        {
          cachedInputRasterPtr.reset( new te::mem::CachedRaster(
            *inputRasterPtr, 20, 0 ) );
          inputRasterPtr = cachedInputRasterPtr.get();
        }         
        
        // The transformation mapping outputRaster pixels 
        // ( te::gm::GTParameters::TiePoint::first ) to input pixels 
        // ( te::gm::GTParameters::TiePoint::second ) 
        // (Note: all coords are indexed by lines/columns).
        
        std::unique_ptr< te::gm::GeometricTransformation > transPtr;
        
        {
          te::gm::GTParameters transParams;
          double auxX;
          double auxY;
          te::gm::GTParameters::TiePoint auxTP;
          
          auxTP.first.x = 0;
          auxTP.first.y = 0;
          outputRasterPtr->getGrid()->gridToGeo( auxTP.first.x, auxTP.first.y, auxX, auxY );
          inputRasterPtr->getGrid()->geoToGrid( auxX, auxY, auxTP.second.x, auxTP.second.y );
          transParams.m_tiePoints.push_back( auxTP );
          
          auxTP.first.x = 0;
          auxTP.first.y = static_cast<double>( outputRasterPtr->getNumberOfRows() - 1 );
          outputRasterPtr->getGrid()->gridToGeo( auxTP.first.x, auxTP.first.y, auxX, auxY );
          inputRasterPtr->getGrid()->geoToGrid( auxX, auxY, auxTP.second.x, auxTP.second.y );          
          transParams.m_tiePoints.push_back( auxTP );

          auxTP.first.x = static_cast<double>( outputRasterPtr->getNumberOfColumns() - 1);
          auxTP.first.y = 0;
          outputRasterPtr->getGrid()->gridToGeo( auxTP.first.x, auxTP.first.y, auxX, auxY );
          inputRasterPtr->getGrid()->geoToGrid( auxX, auxY, auxTP.second.x, auxTP.second.y );
          transParams.m_tiePoints.push_back( auxTP );
          
          auxTP.first.x = static_cast<double>( outputRasterPtr->getNumberOfColumns() - 1 );
          auxTP.first.y = static_cast<double>( outputRasterPtr->getNumberOfRows() - 1 );
          outputRasterPtr->getGrid()->gridToGeo( auxTP.first.x, auxTP.first.y, auxX, auxY );
          inputRasterPtr->getGrid()->geoToGrid( auxX, auxY, auxTP.second.x, auxTP.second.y );                    
          transParams.m_tiePoints.push_back( auxTP );
          
          transPtr.reset( te::gm::GTFactory::make( "Affine" ) );
          TERP_INSTANCE_TRUE_OR_RETURN_FALSE( transPtr.get(), "Could not instantiate a geometric transformation" );
          
          TERP_INSTANCE_TRUE_OR_RETURN_FALSE( transPtr->initialize( transParams ),
            "Could not initialize a geometric transformation" );
        }

        // Locating the raster valid data area if required
        
        std::unique_ptr< te::gm::MultiPolygon >
          currRasterValidDataDelimiterMultiPolPtr;
        
        {
          std::unique_ptr< te::gm::Polygon > 
            currRasterValidDataDelimiterPolygonPtr;          
          
          if( m_inputParameters.m_enableAutoValidAreaDetect )
          {
            std::vector< double > bandNoDataValues;          
            if( m_inputParameters.m_forceInputNoDataValue )
            {
              bandNoDataValues.resize( m_inputParameters.m_inputRastersBands[ 
                inputRasterIdx ].size(), m_inputParameters.m_noDataValue );
            }
            
            std::unique_ptr< te::gm::Polygon > outPolygonPtr;
            
            if( te::rp::GetExternalValidDataPolygon( *inputRasterPtr, 
              m_inputParameters.m_inputRastersBands[ inputRasterIdx ],
              bandNoDataValues, outPolygonPtr ) )
            {
              TERP_DEBUG_TRUE_OR_THROW( outPolygonPtr->isValid(), "Invalid valid area polygon" );
              
              currRasterValidDataDelimiterPolygonPtr.reset( outPolygonPtr.release() );
            }
            else
            {
              const std::string message = "Error calculating input raster valid data area " +
                std::to_string( inputRasterIdx );
              if( m_inputParameters.m_abortOnFirstError )
              {
                TERP_INSTANCE_LOG_AND_RETURN_FALSE( message );
              }
              else
              {
                TERP_INSTANCE_LOGWARN( message );
                ++centerDist2RasterIdxMapIt;
                continue;
              }
            }
          }
          else
          {
            currRasterValidDataDelimiterPolygonPtr.reset(
              static_cast<te::gm::Polygon*>(te::gm::GetGeomFromEnvelope(
              inputRasterPtr->getGrid()->getExtent(), 
              inputRasterPtr->getGrid()->getSRID() ) ) );
          }
          
          currRasterValidDataDelimiterMultiPolPtr.reset( new te::gm::MultiPolygon( 
            0, te::gm::MultiPolygonType, 
            currRasterValidDataDelimiterPolygonPtr->getSRID(), nullptr ) );
          currRasterValidDataDelimiterMultiPolPtr->add( 
            currRasterValidDataDelimiterPolygonPtr.release() );
          currRasterValidDataDelimiterMultiPolPtr->computeMBR( true );
          TERP_DEBUG_TRUE_OR_THROW( currRasterValidDataDelimiterMultiPolPtr->isValid(),
            "Invalid valid area multi-polygon");
        }
        
        // equalization stuff

        std::vector< double > currentRasterMeans;
        std::vector< double > currentRasterStdDevs;
        std::vector< double > currentRasterBandsOffsets;
        std::vector< double > currentRasterBandsScales;        
        std::vector< double > currentRasterNewMeans;
        std::vector< double > currentRasterNewStdDevs;        
        
        if( 
            ! getEqualizationParams( 
              *inputRasterPtr,
              m_inputParameters.m_inputRastersBands[ inputRasterIdx ],
              blendedRastersInfos,
              blendedRastersInfosTree,
              currRasterValidDataDelimiterMultiPolPtr.get(),
              *outputRasterPtr,
              mosaicValidAreasUnionPtr.get(),
              currentRasterMeans,
              currentRasterStdDevs,
              currentRasterBandsOffsets,
              currentRasterBandsScales, 
              currentRasterNewMeans, 
              currentRasterNewStdDevs )
          )
        {
          TERP_INSTANCE_LOG_AND_RETURN_FALSE( 
            "Error calculating statistics from input raster " +
            std::to_string( inputRasterIdx ) 
          );
        }        
        
        // updating blended raster infos

        {
          std::tuple< std::shared_ptr< te::gm::MultiPolygon >,
            std::vector< double >, std::vector< double > > auxTuple;
          std::get<0 >( auxTuple ).reset( (te::gm::MultiPolygon*)
            currRasterValidDataDelimiterMultiPolPtr->clone() );
          std::get<1 >( auxTuple ) = currentRasterNewMeans;
          std::get<2 >( auxTuple ) = currentRasterNewStdDevs;
          blendedRastersInfos.push_back( auxTuple );

          blendedRastersInfosTree.insert(
            *currRasterValidDataDelimiterMultiPolPtr->getMBR(),
            blendedRastersInfos.size() - 1 );
        }

        // blending
         
        {
          te::rp::Blender blenderInstance;

          if( 
              !
              ( 
                (
                  blenderInstance.initialize(
                  *outputRasterPtr,
                  outputRasterBands,
                  *inputRasterPtr,
                  m_inputParameters.m_inputRastersBands[ inputRasterIdx ],
                  m_inputParameters.m_blendMethod,
                  te::rst::NearestNeighbor,
                  m_inputParameters.m_interpMethod,
                  m_inputParameters.m_noDataValue,
                  false,
                  m_inputParameters.m_forceInputNoDataValue,
                  std::vector< double >( outputRasterBands.size(), 0 ),
                  std::vector< double >( outputRasterBands.size(), 1 ),
                  currentRasterBandsOffsets,
                  currentRasterBandsScales,
                  mosaicValidAreasUnionPtr.get(),
                  currRasterValidDataDelimiterMultiPolPtr.get(),
                  *transPtr,
                  0,
                  false,
                  m_inputParameters.m_blenderRulePtr,
                  (
                    m_inputParameters.m_useRasterCache 
                    &&
                    ( inputRasterPtr == m_inputParameters.m_feederRasterPtr->getCurrentObj() ) )
                  )
                )
                &&
                (
                  blenderInstance.blendIntoRaster1()
                )
              )
            )
          {
            const std::string message = "Error blending input raster " +
              std::to_string( inputRasterIdx );
            if( m_inputParameters.m_abortOnFirstError )
            {
              TERP_INSTANCE_LOG_AND_RETURN_FALSE( message );
            }
            else
            {
              TERP_INSTANCE_LOGWARN( message );
              ++centerDist2RasterIdxMapIt;
              continue;
            }
          }

          // te::rp::Copy2DiskRaster( *outputRasterPtr, std::string(
          //   "GeoMosaicOutputRaster_") +  std::to_string( inputRasterIdx ) +
          //   ".tif" );
        }

        // updating the  gloabal mosaic boxes
        
        if( mosaicValidAreasUnionPtr.get() == nullptr )
        {
          mosaicValidAreasUnionPtr.reset(
            currRasterValidDataDelimiterMultiPolPtr.release() );
        }
        else
        {
          std::unique_ptr< te::gm::Geometry > boxesUnionResultPtr; // under the mosaic SRID

          try
          {
            boxesUnionResultPtr.reset( mosaicValidAreasUnionPtr->Union(
              currRasterValidDataDelimiterMultiPolPtr.get() ) );
          }
          catch( const std::exception& e )
          {
            TERP_INSTANCE_LOG_AND_RETURN_FALSE( "Mosaic bounding boxes union error : "
              + std::string( e.what() ) );
          }
          catch( ... )
          {
            TERP_INSTANCE_LOG_AND_RETURN_FALSE( "Mosaic bounding boxes union error" );
          }          
          
          TERP_INSTANCE_TRUE_OR_RETURN_FALSE( boxesUnionResultPtr.get() !=
            nullptr, "Mosaic bounding boxes union error" );
          
          boxesUnionResultPtr->setSRID( outputRasterPtr->getSRID() );

          if( 
              ( boxesUnionResultPtr->getGeomTypeId() == te::gm::MultiPolygonType )
              ||
              ( boxesUnionResultPtr->getGeomTypeId() == te::gm::MultiPolygonZType )
              ||
              ( boxesUnionResultPtr->getGeomTypeId() == te::gm::MultiPolygonMType )
              ||
              ( boxesUnionResultPtr->getGeomTypeId() == te::gm::MultiPolygonZMType )
            )
          {
             mosaicValidAreasUnionPtr.reset( static_cast<te::gm::MultiPolygon*>(boxesUnionResultPtr.release()) );
          }
          else if( 
                   ( boxesUnionResultPtr->getGeomTypeId() == te::gm::PolygonType )
                   ||
                   ( boxesUnionResultPtr->getGeomTypeId() == te::gm::PolygonZType )
                   ||
                   ( boxesUnionResultPtr->getGeomTypeId() == te::gm::PolygonMType )
                   ||
                   ( boxesUnionResultPtr->getGeomTypeId() == te::gm::PolygonZMType )
                 )
          {
            // transforming it into a te::gm::MultiPolygon
            te::gm::MultiPolygon* auxMultiPol = new te::gm::MultiPolygon( 0,
              te::gm::MultiPolygonType, boxesUnionResultPtr->getSRID(), nullptr );
            auxMultiPol->add( boxesUnionResultPtr.release() );
            auxMultiPol->computeMBR( true );
            assert( auxMultiPol->isValid() );

            mosaicValidAreasUnionPtr.reset( auxMultiPol );
          }
          else
          {
            TERP_INSTANCE_LOG_AND_RETURN_FALSE( "Invalid union geometry type" );
          }
        }
        
        // moving to the next raster

        ++centerDist2RasterIdxMapIt;
      }
      
      // reseting the output cache
      
      if( outputCachedRasterInstancePtr.get() ) outputCachedRasterInstancePtr.reset();
      
      return true;
    }

    void GeoMosaic::reset() _NOEXCEPT_OP(false)
    {
      Algorithm::reset();
      
      m_inputParameters.reset();
      m_isInitialized = false;
    }

    bool GeoMosaic::initialize( const AlgorithmInputParameters& inputParams )
      _NOEXCEPT_OP(false)
    {
      reset();

      GeoMosaic::InputParameters const* inputParamsPtr = dynamic_cast<
        GeoMosaic::InputParameters const* >( &inputParams );
      TERP_INSTANCE_TRUE_OR_THROW( inputParamsPtr, "Invalid paramters pointer" );

      m_inputParameters = *inputParamsPtr;

      // Checking the feeder

      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( m_inputParameters.m_feederRasterPtr,
        "Invalid m_feederRasterPtr" )

      TERP_INSTANCE_TRUE_OR_RETURN_FALSE(
        m_inputParameters.m_feederRasterPtr->getObjsCount() > 0,
        "Invalid number of rasters" )

      // checking m_inputRastersBands

      TERP_INSTANCE_TRUE_OR_RETURN_FALSE(
        (static_cast<unsigned int>(m_inputParameters.m_inputRastersBands.size())) ==
        m_inputParameters.m_feederRasterPtr->getObjsCount(),
        "Bands mismatch" );

      for( std::vector< std::vector< unsigned int > >::size_type
        inputRastersBandsIdx = 0 ;  inputRastersBandsIdx <
        m_inputParameters.m_inputRastersBands.size() ; ++inputRastersBandsIdx )
      {
        TERP_INSTANCE_TRUE_OR_RETURN_FALSE( m_inputParameters.m_inputRastersBands[
          inputRastersBandsIdx ].size() > 0, "Invalid bands number" );

        TERP_INSTANCE_TRUE_OR_RETURN_FALSE( m_inputParameters.m_inputRastersBands[
          inputRastersBandsIdx ].size() ==  m_inputParameters.m_inputRastersBands[
          0 ].size(), "Bands number mismatch" );
      }
      
      // Checking eq type
      
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( m_inputParameters.m_equalizationType
        != InputParameters::InvalidEqType , "Invalid equalization type" );
      
      // checking target means and std devs
      
      if( m_inputParameters.m_equalizationType == InputParameters::TargetEqtype )
      {
        TERP_INSTANCE_TRUE_OR_RETURN_FALSE( 
          (
            ( 
              m_inputParameters.m_mosaicTargetMeans.size() 
              == 
              m_inputParameters.m_inputRastersBands[ 0 ].size() 
            )
            &&
            ( 
              m_inputParameters.m_mosaicTargetStdDevs.size() 
              == 
              m_inputParameters.m_inputRastersBands[ 0 ].size() 
            )          
            &&
            ( 
              m_inputParameters.m_mosaicTargetMeans.size()
              ==
              m_inputParameters.m_mosaicTargetStdDevs.size()
            )                    
          )
          , "Invalid target means or target standart deviation values" );
      }
      
      // Blender
      
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( 
        ( 
          ( 
            m_inputParameters.m_blendMethod 
            ==
            te::rp::blender::RuleMethod 
          ) 
          ? 
          ( 
            m_inputParameters.m_blenderRulePtr 
            !=
            nullptr 
          ) 
          : 
          true 
        ),
        "Missing blender rule" );
      
      // finalizing

      m_isInitialized = true;

      return true;
    }
    
    bool GeoMosaic::isInitialized() const
    {
      return m_isInitialized;
    }

    bool GeoMosaic::getRasterStats(
      const te::rst::Raster& inRaster,
      const std::vector< unsigned int >& inputRastersBands,
      te::gm::MultiPolygon const * const  validAreaDelimiterPolPtr,
      std::vector< double >& rasterMeans,
      std::vector< double >& rasterStdDevs  ) const
    {
      const unsigned int inputRastersBandsSize = inputRastersBands.size();

      rasterMeans = std::vector< double >( inputRastersBandsSize, 0 );
      rasterStdDevs = std::vector< double >( inputRastersBandsSize, 0 );

      if( validAreaDelimiterPolPtr == nullptr )
      {
        std::map< double, unsigned > rHistogram;
        std::map< double, unsigned > iHistogram;
        double min = 0;
        double max = 0;

        for( unsigned int inputRastersBandsIdx = 0 ; inputRastersBandsIdx <
         inputRastersBandsSize ; ++inputRastersBandsIdx )
        {
          if( ! te::rst::getHistograms(
                  inRaster,
                  inputRastersBands[ inputRastersBandsIdx ],
                  0, // rowStart
                  0, // colStart
                  inRaster.getNumberOfRows() - 1,
                  inRaster.getNumberOfColumns() - 1,
                  0, // histoBins
                  1, // sampleStep
                  0, // maxThreads
                  rHistogram,
                  iHistogram
                )
            )
          {
            return false;
          }

          te::rst::getHistogramStats( rHistogram, min, max, rasterMeans[
            inputRastersBandsIdx ], rasterStdDevs[ inputRastersBandsIdx ] );
        }
      }
      else
      {
        std::vector< double > meansSum( inputRastersBandsSize, 0 );
        std::vector< double > stdDevsSum( inputRastersBandsSize, 0 );
        std::vector< te::gm::Geometry* > singleGeoms;
        std::size_t singleGeomsSize = 0;
        std::size_t singleGeomsIdx = 0;
        std::vector< te::gm::Polygon const* > singlePolVec;
        std::map< double, unsigned > rHistogram;
        std::map< double, unsigned > iHistogram;
        double mean = 0;
        double stdDev = 0;
        double min = 0;
        double max = 0;
        double singlePolArea = 0;
        double areaSum = 0;

        singleGeoms.clear();
        te::gm::Multi2Single( validAreaDelimiterPolPtr, singleGeoms );

        singleGeomsSize = singleGeoms.size();

        for( singleGeomsIdx = 0 ; singleGeomsIdx < singleGeomsSize ;
          ++singleGeomsIdx )
        {
          if(
              singleGeoms[ singleGeomsIdx ]->getGeomTypeId()
              ==
              te::gm::PolygonType
            )
          {
            singlePolVec.clear();
            singlePolVec.push_back( (te::gm::Polygon const*)
              singleGeoms[ singleGeomsIdx ] );

            singlePolArea = singlePolVec[ 0 ]->getArea();
            areaSum += singlePolArea;

            for( unsigned int inputRastersBandsIdx = 0 ; inputRastersBandsIdx <
             inputRastersBandsSize ; ++inputRastersBandsIdx )
            {
              if( ! te::rp::getHistograms(
                      inRaster,
                      inputRastersBands[ inputRastersBandsIdx ],
                      singlePolVec,
                      0, // histoBins
                      0, // maxThreads
                      te::rst::PolygonIterator< double >::ScanLineIterationT,
                      rHistogram,
                      iHistogram
                    )
                )
              {
                return false;
              }

              te::rst::getHistogramStats( rHistogram, min, max, mean,
                stdDev );

              meansSum[ inputRastersBandsIdx ] += mean * singlePolArea;
              stdDevsSum[ inputRastersBandsIdx ] += stdDev * singlePolArea;
            }
          }

          delete singleGeoms[ singleGeomsIdx ];
        }

        if( areaSum > 0 )
        {
          for( unsigned int inputRastersBandsIdx = 0 ; inputRastersBandsIdx <
           inputRastersBandsSize ; ++inputRastersBandsIdx )
          {
            rasterMeans[ inputRastersBandsIdx ] = meansSum[ inputRastersBandsIdx ]
              / areaSum;
            rasterStdDevs[ inputRastersBandsIdx ] = stdDevsSum[ inputRastersBandsIdx ]
              / areaSum;
          }
        }
      }

      return true;
    }
    
    bool GeoMosaic::getEqualizationParams(
      const te::rst::Raster& inRaster,
      const std::vector< unsigned int >& inputRastersBands,
      const BlendedRastersInfosT&  blendedRastersInfos,
      const te::sam::rtree::Index< std::size_t >& blendedRastersInfosTree,      
      te::gm::MultiPolygon const * const  validAreaDelimiterPolPtr,
      const te::rst::Raster& mosaicRaster,
      te::gm::MultiPolygon const * const  mosaicValidAreaDelimiterPolPtr,
      std::vector< double >& rasterMeans,
      std::vector< double >& rasterStdDevs,
      std::vector< double >& offsets,
      std::vector< double >& scales,
      std::vector< double >& newRasterMeans,
      std::vector< double >& newRasterStdDevs ) const
    {
      const unsigned int bandsNumber = inputRastersBands.size();

      rasterMeans = std::vector< double >( bandsNumber, 0 );
      rasterStdDevs = std::vector< double >( bandsNumber, 0 );
      offsets = std::vector< double >( bandsNumber, 0 );
      scales = std::vector< double >( bandsNumber, 1 );
      newRasterMeans = std::vector< double >( bandsNumber, 0 );
      newRasterStdDevs = std::vector< double >( bandsNumber, 0 );
      
      std::vector< double > targetMeans;
      std::vector< double > targetStdDevs;
      
      switch( m_inputParameters.m_equalizationType )
      {
        case InputParameters::NoEqType :
        {
          targetMeans = rasterMeans;
          targetStdDevs = rasterStdDevs;
          
          break;
        }          
        case InputParameters::CenterEqType :
        {
          if( ! getRasterStats( 
                  inRaster, 
                  inputRastersBands,
                  ( 
                    m_inputParameters.m_enableAutoValidAreaDetect 
                    ?  
                    validAreaDelimiterPolPtr
                    :
                    nullptr
                  ),
                  rasterMeans, 
                  rasterStdDevs ) 
            )
          {
            return false;
          }

          if( blendedRastersInfos.empty() )
          { // this is the first raster
            targetMeans = rasterMeans;
            targetStdDevs = rasterStdDevs;
          }
          else
          {
            targetMeans = std::get< 1 >( blendedRastersInfos[ 0 ] );
            targetStdDevs = std::get< 2 >( blendedRastersInfos[ 0 ] );
          }
          
          break;
        }
        case InputParameters::TargetEqtype :
        {
          if( ! getRasterStats( 
                  inRaster, 
                  inputRastersBands,
                  ( 
                    m_inputParameters.m_enableAutoValidAreaDetect 
                    ?  
                    validAreaDelimiterPolPtr
                    :
                    nullptr
                  ),
                  rasterMeans, 
                  rasterStdDevs ) 
            )
          {
            return false;
          }

          targetMeans = m_inputParameters.m_mosaicTargetMeans;
          targetStdDevs = m_inputParameters.m_mosaicTargetStdDevs;
          
          break;
        }
        case InputParameters::LocalEqType :
        {
          if( validAreaDelimiterPolPtr == nullptr ) return false;

          if( ! getRasterStats( 
                  inRaster, 
                  inputRastersBands,
                  ( 
                    m_inputParameters.m_enableAutoValidAreaDetect 
                    ?  
                    validAreaDelimiterPolPtr
                    :
                    nullptr
                  ),
                  rasterMeans, 
                  rasterStdDevs ) 
            )
          {
            return false;
          }

          te::gm::Envelope treeSearchEnv( *validAreaDelimiterPolPtr->getMBR() );
          treeSearchEnv.m_llx -=
            validAreaDelimiterPolPtr->getMBR()->getWidth()
            / 2.0;
          treeSearchEnv.m_lly -=
            validAreaDelimiterPolPtr->getMBR()->getHeight()
            / 2.0;
          treeSearchEnv.m_urx +=
            validAreaDelimiterPolPtr->getMBR()->getWidth()
            / 2.0;
          treeSearchEnv.m_ury +=
            validAreaDelimiterPolPtr->getMBR()->getHeight()
            / 2.0;

          std::vector< std::size_t > treeSearchResult;
          blendedRastersInfosTree.search( treeSearchEnv, treeSearchResult );

          if( treeSearchResult.empty() )
          {
            targetMeans = rasterMeans;
            targetStdDevs = rasterStdDevs;
          }
          else
          {
            targetMeans = std::vector< double >( bandsNumber, 0 );
            targetStdDevs = std::vector< double >( bandsNumber, 0 );            
            
            double validAreasSum = 0;
            double nRasterValidArea = 0;
            
            for( std::size_t treeSearchResultIdx = 0 ; treeSearchResultIdx <
              treeSearchResult.size() ; ++treeSearchResultIdx )
            {
              const te::gm::MultiPolygon& nRasterValidAreaPol = *std::get< 0 >(
                blendedRastersInfos[ treeSearchResult[
                treeSearchResultIdx ] ] );
              const std::vector< double >& nRasterMeans = std::get< 1 >(
                blendedRastersInfos[ treeSearchResult[
                treeSearchResultIdx ] ] );
              const std::vector< double >& nRasterStdDevs = std::get< 2 >(
                blendedRastersInfos[ treeSearchResult[
                treeSearchResultIdx ] ] );

              nRasterValidArea = nRasterValidAreaPol.getArea();

              validAreasSum += nRasterValidArea;

              for( unsigned int bandIdx = 0 ; bandIdx < bandsNumber ; ++bandIdx )
              {
                targetMeans[ bandIdx ] +=
                  ( nRasterMeans[ bandIdx ] * nRasterValidArea );
                targetStdDevs[ bandIdx ] +=
                  ( nRasterStdDevs[ bandIdx ] * nRasterValidArea );
              }                        
            }
            
            if( validAreasSum == 0 )
            {
              targetMeans = rasterMeans;
              targetStdDevs = rasterStdDevs;
            }
            else
            {
              for( unsigned int bandIdx = 0 ; bandIdx < bandsNumber ;
                ++bandIdx )
              {
                targetMeans[ bandIdx ] /= validAreasSum;
                targetStdDevs[ bandIdx ] /= validAreasSum;
              }
            }
          }
          
          break;
        }            
        case InputParameters::OverLapEqType :
        {
          if( mosaicValidAreaDelimiterPolPtr == nullptr )
          {
            if( ! getRasterStats( 
                    inRaster, 
                    inputRastersBands,
                    ( 
                      m_inputParameters.m_enableAutoValidAreaDetect 
                      ?  
                      validAreaDelimiterPolPtr
                      :
                      nullptr
                    ),
                    rasterMeans, 
                    rasterStdDevs ) 
              )
            {
              return false;
            }

            targetMeans = rasterMeans;
            targetStdDevs = rasterStdDevs;
          }
          else
          {
            if( validAreaDelimiterPolPtr == nullptr ) return false;
            
            std::unique_ptr< te::gm::Geometry > interPtr;
            try
            {
              interPtr.reset( validAreaDelimiterPolPtr->intersection(
                mosaicValidAreaDelimiterPolPtr ) );
            }
            catch(...)
            {
              interPtr.reset();
            }

            if( interPtr.get() == nullptr )
            {
              if( ! getRasterStats( 
                      inRaster, 
                      inputRastersBands,
                      ( 
                        m_inputParameters.m_enableAutoValidAreaDetect 
                        ?  
                        validAreaDelimiterPolPtr
                        :
                        nullptr
                      ),
                      rasterMeans, 
                      rasterStdDevs ) 
                )
              {
                return false;
              }

              targetMeans = rasterMeans;
              targetStdDevs = rasterStdDevs;
            }
            else
            {
              if( interPtr->getGeomTypeId() == te::gm::PolygonType )
              {
                te::gm::MultiPolygon* auxMultiPol = new te::gm::MultiPolygon( 0,
                  te::gm::MultiPolygonType, interPtr->getSRID(), nullptr );
                auxMultiPol->add( (te::gm::Polygon*)interPtr->clone() );
                auxMultiPol->computeMBR( true );
                assert( auxMultiPol->isValid() );

                interPtr.reset( auxMultiPol );
              }

              if( interPtr->getGeomTypeId() == te::gm::MultiPolygonType )
              {
                if( ! getRasterStats( 
                        inRaster, 
                        inputRastersBands,
                        (te::gm::MultiPolygon*)interPtr.get(), 
                        rasterMeans, 
                        rasterStdDevs ) 
                  )
                {
                  return false;
                }                
                
                std::vector< unsigned int > mosaicBandsIndexes;
                for( unsigned int bandIdx = 0 ; bandIdx < bandsNumber ; ++bandIdx )
                  mosaicBandsIndexes.push_back( bandIdx );
                if( ! getRasterStats( 
                        mosaicRaster, 
                        mosaicBandsIndexes,
                        (te::gm::MultiPolygon*)interPtr.get(), 
                        targetMeans, 
                        targetStdDevs 
                      ) 
                  )
                {
                  return false;
                }
              }
              else
              {
                if( ! getRasterStats( 
                        inRaster, 
                        inputRastersBands,
                        ( 
                          m_inputParameters.m_enableAutoValidAreaDetect 
                          ?  
                          validAreaDelimiterPolPtr
                          :
                          nullptr
                        ),
                        rasterMeans, 
                        rasterStdDevs ) 
                  )
                {
                  return false;
                }                                
                
                targetMeans = rasterMeans;
                targetStdDevs = rasterStdDevs;
              }
            }
          }

          break;
        }
        default :
        {
          TERP_LOG_AND_THROW( "Invalid equalization type");
          break;
        }
      }

      for( unsigned int bandIdx = 0 ; bandIdx <  bandsNumber ;
        ++bandIdx )
      {
        if( rasterStdDevs[ bandIdx ] == 0 )
        {
          scales[ bandIdx ] = 1.0;
          offsets[ bandIdx ] = 0.0;
        }
        else
        {
          scales[ bandIdx ] =
            ( 
              std::sqrt( targetStdDevs[ bandIdx ] )
               /
              std::sqrt( rasterStdDevs[ bandIdx ] ) 
            );
          offsets[ bandIdx ] =
            ( 
              targetMeans[ bandIdx ]
               - 
              ( 
                scales[ bandIdx ] 
                * 
                rasterMeans[ bandIdx ] 
              ) 
            );
        }
        
        newRasterMeans[ bandIdx ] = 
          ( 
            rasterMeans[ bandIdx ]
            *
            scales[ bandIdx ]
          )
          +
          offsets[ bandIdx ]
          ;
          
        newRasterStdDevs[ bandIdx ] = 
          ( 
            scales[ bandIdx ]
            *
            scales[ bandIdx ]
            *
            rasterStdDevs[ bandIdx ]
          );
      }
      
      return true;
    }    

  } // end namespace rp
}   // end namespace te

