/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/rp/IHSFusion.cpp
  \brief Creation of skeleton imagems.
*/

#include "IHSFusion.h"
#include "Macros.h"
#include "Functions.h"
#include "AlgorithmParametersSerializer.h"

#include "../raster/BandProperty.h"
#include "../raster/RasterFactory.h"
#include "../raster/Band.h"
#include "../raster/Grid.h"
#include "../raster/Utils.h"
#include "../raster/Reprojection.h"
#include "../raster/SynchronizedRaster.h"
#include "../geometry/Envelope.h"
#include "../common/progress/TaskProgress.h"
#include "../common/PlatformUtils.h"
#include "../memory/CachedRaster.h"

#include <cmath>
#include <limits>

#ifndef M_PI
  #define M_PI       3.14159265358979323846
#endif
  
namespace te
{
  namespace rp
  {

    IHSFusion::InputParameters::InputParameters()
    {
      reset();
    }

    IHSFusion::InputParameters::InputParameters( const InputParameters& other )
    {
      reset();
      operator=( other );
    }

    IHSFusion::InputParameters::~InputParameters()
    {
      reset();
    }

    void IHSFusion::InputParameters::reset() 
    {
      m_lowResRasterPtr = nullptr;
      m_lowResRasterRedBandIndex = 0;
      m_lowResRasterGreenBandIndex = 1;
      m_lowResRasterBlueBandIndex = 2;
      m_highResRasterPtr = nullptr;
      m_highResRasterBand = 0;
      m_enableProgress = false;
      m_interpMethod = te::rst::NearestNeighbor;
      m_RGBMin = 0.0;
      m_RGBMax = 0.0;
      m_enableThreadedProcessing = true;
      m_autoAlignRasters = true;
      m_enableRasterCache = true;
      m_hrHistoFitMethod = InputParameters::MeanStdDevHistoFitMethod;
    }

    const IHSFusion::InputParameters& IHSFusion::InputParameters::operator=(
      const IHSFusion::InputParameters& params )
    {
      reset();

      m_lowResRasterPtr = params.m_lowResRasterPtr;
      m_lowResRasterRedBandIndex = params.m_lowResRasterRedBandIndex;
      m_lowResRasterGreenBandIndex = params.m_lowResRasterGreenBandIndex;
      m_lowResRasterBlueBandIndex = params.m_lowResRasterBlueBandIndex;
      m_highResRasterPtr = params.m_highResRasterPtr;
      m_highResRasterBand = params.m_highResRasterBand;
      m_enableProgress = params.m_enableProgress;
      m_interpMethod = params.m_interpMethod;
      m_RGBMin = params.m_RGBMin;
      m_RGBMax = params.m_RGBMax;
      m_enableThreadedProcessing = params.m_enableThreadedProcessing;
      m_autoAlignRasters = params.m_autoAlignRasters;
      m_enableRasterCache = params.m_enableRasterCache;
      m_hrHistoFitMethod = params.m_hrHistoFitMethod;

      return *this;
    }

    te::common::AbstractParameters* IHSFusion::InputParameters::clone() const
    {
      return new InputParameters( *this );
    }

    bool IHSFusion::InputParameters::serialize( AlgorithmParametersSerializer& serializer ) const
    {
      TERP_TRUE_OR_RETURN_FALSE( AlgorithmInputParameters::serialize( serializer ),
        "Serializarion error" );

      serializer.setAlgorithmName( "IHSFusion" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "lowResRaster",
        m_lowResRasterPtr->getName() ), "Error adding parameter lowResRaster" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "lowResRasterRedBandIndex",
        m_lowResRasterRedBandIndex ), "Error adding parameter lowResRasterRedBandIndex" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "lowResRasterGreenBandIndex",
        m_lowResRasterGreenBandIndex ), "Error adding parameter lowResRasterGreenBandIndex" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "lowResRasterBlueBandIndex",
        m_lowResRasterBlueBandIndex ), "Error adding parameter lowResRasterBlueBandIndex" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "highResRaster",
        m_highResRasterPtr->getName() ), "Error adding parameter highResRaster" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "highResRasterBand",
        m_highResRasterBand ), "Error adding parameter highResRasterBand" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "interpMethod",
        m_interpMethod ), "Error adding parameter interpMethod" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "RGBMin",
        m_RGBMin ), "Error adding parameter RGBMin" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "RGBMax",
        m_RGBMax ), "Error adding parameter RGBMax" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "autoAlignRasters",
        m_autoAlignRasters ), "Error adding parameter autoAlignRasters" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "hrHistoFitMethod",
        m_hrHistoFitMethod ), "Error adding parameter hrHistoFitMethod" );

      return true;
    }

    IHSFusion::OutputParameters::OutputParameters()
    {
      reset();
    }

    IHSFusion::OutputParameters::OutputParameters( const OutputParameters& other )
    {
      reset();
      operator=( other );
    }

    IHSFusion::OutputParameters::~OutputParameters()
    {
      reset();
    }

    void IHSFusion::OutputParameters::reset() 
    {
      m_rType.clear();
      m_rInfo.clear();
      m_outputRasterPtr.reset();
    }

    const IHSFusion::OutputParameters& IHSFusion::OutputParameters::operator=(
      const IHSFusion::OutputParameters& params )
    {
      reset();

      m_rType = params.m_rType;
      m_rInfo = params.m_rInfo;

      return *this;
    }

    te::common::AbstractParameters* IHSFusion::OutputParameters::clone() const
    {
      return new OutputParameters( *this );
    }

    IHSFusion::IHSFusion()
    {
      reset();
    }

    IHSFusion::~IHSFusion() = default;

    bool IHSFusion::execute( AlgorithmOutputParameters& outputParams )
    {
      if( ! m_isInitialized ) return false;

      IHSFusion::OutputParameters* outParamsPtr = dynamic_cast<
        IHSFusion::OutputParameters* >( &outputParams );
      TERP_TRUE_OR_THROW( outParamsPtr, "Invalid paramters" );
      
      // progress
      
      std::unique_ptr< te::common::TaskProgress > progressPtr;
      if( m_inputParameters.m_enableProgress )
      {
        progressPtr.reset( new te::common::TaskProgress );
        
        progressPtr->setTotalSteps( 5 );
        
        progressPtr->setMessage( "Fusing images" );
      }        
      
      // HR raster auto alignment
      
      std::unique_ptr< te::rst::Raster > alignedHighResRasterHandler;
      te::rst::Raster const * alignedHighResRasterPtr = 
        m_inputParameters.m_highResRasterPtr;      
      
      if( m_inputParameters.m_autoAlignRasters )
      {
        te::gm::Envelope newHREnvelope( *m_inputParameters.m_lowResRasterPtr->getExtent() );
        newHREnvelope.transform( m_inputParameters.m_lowResRasterPtr->getSRID(),
          m_inputParameters.m_highResRasterPtr->getSRID() );
        
        double llCol = 0;
        double llRow = 0;
        m_inputParameters.m_highResRasterPtr->getGrid()->geoToGrid(
          newHREnvelope.getLowerLeftX(), newHREnvelope.getLowerLeftY(),
          llCol, llRow );
        llCol = (double)te::common::Round<double, long int>( llCol );
        llCol = std::max( 0.0, llCol );
        llCol = std::min( llCol, 
          (double)( m_inputParameters.m_highResRasterPtr->getNumberOfColumns() - 1 ) );
        llRow = (double)te::common::Round<double, long int>( llRow );
        llRow = std::max( 0.0, llRow );
        llRow = std::min( llRow, 
          (double)( m_inputParameters.m_highResRasterPtr->getNumberOfRows() - 1 ) );

        double urCol = 0;
        double urRow = 0;
        m_inputParameters.m_highResRasterPtr->getGrid()->geoToGrid(
          newHREnvelope.getUpperRightX(), newHREnvelope.getUpperRightY(),
          urCol, urRow );
        urCol = (double)te::common::Round<double, long int>( urCol );
        urCol = std::max( 0.0, urCol );
        urCol = std::min( urCol, 
          (double)( m_inputParameters.m_highResRasterPtr->getNumberOfColumns() - 1 ) );
        urRow = (double)te::common::Round<double, long int>( urRow );
        urRow = std::max( 0.0, urRow );
        urRow = std::min( urRow, 
          (double)( m_inputParameters.m_highResRasterPtr->getNumberOfRows() - 1 ) );
        
        std::vector< unsigned int > inRasterBands;
        inRasterBands.push_back( m_inputParameters.m_highResRasterBand );
        
        std::map<std::string, std::string> routInfo;
        routInfo[ "MAXMEMPERCENTUSED" ] = "20";
        
        TERP_INSTANCE_TRUE_OR_RETURN_FALSE( te::rst::Reproject(
          m_inputParameters.m_lowResRasterPtr->getSRID(),
          *m_inputParameters.m_highResRasterPtr,
          inRasterBands,
          m_inputParameters.m_interpMethod,
          1, // interpWindowRadius
          urRow,
          llCol,
          llRow - urRow + 1, // height
          urCol - llCol + 1, // width
          llRow - urRow + 1,
          urCol - llCol + 1,
          0, // max threads
          routInfo,
          "EXPANSIBLE",
          alignedHighResRasterHandler
          ), "HR raster auto-alignment error" );
        
        alignedHighResRasterPtr = alignedHighResRasterHandler.get();
      }
      
      if( m_inputParameters.m_enableProgress )
      {
        progressPtr->pulse();
        if( ! progressPtr->isActive() ) return false;
      }  
            
      
      // Getting RGB range
      
      double rgbMin = 0.0;
      double rgbMax = 0.0;
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( getRGBRange( rgbMin, rgbMax ),
        "Unable to get RGB range" );
      
      if( m_inputParameters.m_enableProgress )
      {
        progressPtr->pulse();
        if( ! progressPtr->isActive() ) return false;
      }  
      
      // Loading IHS data    
      
      std::unique_ptr< te::rst::Raster > ihsRasterPtr;
      
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( loadUpsampledIHSData( 
        *alignedHighResRasterPtr->getGrid(), rgbMin, rgbMax, ihsRasterPtr ), 
        "Error loading IHS data" );
        
      if( m_inputParameters.m_enableProgress )
      {
        progressPtr->pulse();
        if( ! progressPtr->isActive() ) return false;
      }          

      // Swapping Itensity
      
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( swapIntensity( *alignedHighResRasterPtr,
        ( m_inputParameters.m_autoAlignRasters ? 0 : m_inputParameters.m_highResRasterBand ), 
        *ihsRasterPtr ), "Intensity channel swap error" );
        
      if( m_inputParameters.m_enableProgress )
      {
        progressPtr->pulse();
        if( ! progressPtr->isActive() ) return false;
      }
      
      alignedHighResRasterHandler.reset();
      
      // Saving RGB data
      
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( saveRGBData( rgbMin, rgbMax, 
        outParamsPtr->m_rType, outParamsPtr->m_rInfo, outParamsPtr->m_outputRasterPtr,
        *ihsRasterPtr ), "RGB raster creation error" );

      if( m_inputParameters.m_enableProgress )
      {
        progressPtr->pulse();
        if( ! progressPtr->isActive() ) return false;
      }              
      
      return true;
    }

    void IHSFusion::reset() 
    {
      Algorithm::reset();
      
      m_inputParameters.reset();
      m_isInitialized = false;
    }

    bool IHSFusion::initialize( const AlgorithmInputParameters& inputParams )
    {
      reset();

      IHSFusion::InputParameters const* inputParamsPtr = dynamic_cast<
        IHSFusion::InputParameters const* >( &inputParams );
      TERP_TRUE_OR_THROW( inputParamsPtr, "Invalid paramters pointer" );

      m_inputParameters = *inputParamsPtr;

      // Checking the m_lowResRasterPtr and m_lowResRasterBands

      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( m_inputParameters.m_lowResRasterPtr,
        "Invalid low Resolution Raster Pointer" )

      TERP_INSTANCE_TRUE_OR_RETURN_FALSE(
        m_inputParameters.m_lowResRasterPtr->getAccessPolicy() & te::common::RAccess, 
        "Invalid raster" );
        
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE(
        m_inputParameters.m_lowResRasterRedBandIndex <
        m_inputParameters.m_lowResRasterPtr->getNumberOfBands(), 
        "Invalid raster band" );   
        
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE(
        m_inputParameters.m_lowResRasterGreenBandIndex <
        m_inputParameters.m_lowResRasterPtr->getNumberOfBands(), 
        "Invalid raster band" );  
        
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE(
        m_inputParameters.m_lowResRasterBlueBandIndex <
        m_inputParameters.m_lowResRasterPtr->getNumberOfBands(), 
        "Invalid raster band" );          
      
      // Checking the m_highResRasterPtr and m_highResRasterBand

      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( m_inputParameters.m_highResRasterPtr,
        "Invalid high resolution Raster Pointer" )

      TERP_INSTANCE_TRUE_OR_RETURN_FALSE(
        m_inputParameters.m_highResRasterPtr->getAccessPolicy() & te::common::RAccess, 
        "Invalid raster" );
        
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE(
        m_inputParameters.m_highResRasterBand <
        m_inputParameters.m_highResRasterPtr->getNumberOfBands(), 
        "Invalid raster band" );   
        
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( m_inputParameters.m_RGBMin <=
        m_inputParameters.m_RGBMax, "Invalid RGB channel ranges" );
      
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( 
        (
          ( m_inputParameters.m_autoAlignRasters == false )
          ||
          (
            ( m_inputParameters.m_autoAlignRasters == true )
            &&
            ( m_inputParameters.m_lowResRasterPtr->getSRID() != 0 )
            &&
            ( m_inputParameters.m_highResRasterPtr->getSRID() != 0 )
          )
        )
        ,
        "Rasters with valid SRS are required for auto-alignment" 
      );
        
      m_isInitialized = true;

      return true;
    }

    bool IHSFusion::isInitialized() const
    {
      return m_isInitialized;
    }
    
    bool IHSFusion::getRGBRange( double& rgbMin, double& rgbMax ) const
    {
      if( 
          ( m_inputParameters.m_RGBMax == 0.0 )
          && 
          ( m_inputParameters.m_RGBMin == 0.0 ) 
        )
      {
        rgbMin = std::numeric_limits< double >::max();
        rgbMax = -1.0 * rgbMin;
        
        double meanR = 0;
        double stdDevR = 0;
        double minR = 0;
        double maxR = 0;
        if( ! getStatistics( *m_inputParameters.m_lowResRasterPtr,
            m_inputParameters.m_lowResRasterRedBandIndex, meanR, stdDevR,
            minR, maxR )
          )
        {
          return false;
        }
        rgbMin = std::min( rgbMin, minR );
        rgbMax = std::max( rgbMax, maxR );
        
        double meanG = 0;
        double stdDevG = 0;
        double minG = 0;
        double maxG = 0;
        if( ! getStatistics( *m_inputParameters.m_lowResRasterPtr,
            m_inputParameters.m_lowResRasterGreenBandIndex, meanG, stdDevG,
            minG, maxG )
          )
        {
          return false;
        }
        rgbMin = std::min( rgbMin, minG );
        rgbMax = std::max( rgbMax, maxG );        
        
        double meanB = 0;
        double stdDevB = 0;
        double minB = 0;
        double maxB = 0;
        if( ! getStatistics( *m_inputParameters.m_lowResRasterPtr,
            m_inputParameters.m_lowResRasterBlueBandIndex, meanB, stdDevB,
            minB, maxB )
          )
        {
          return false;
        }
        rgbMin = std::min( rgbMin, minB );
        rgbMax = std::max( rgbMax, maxB );
      }
      else
      {
        rgbMin = m_inputParameters.m_RGBMin;
        rgbMax = m_inputParameters.m_RGBMax;
      }
        
      return true;
    }
    
    bool IHSFusion::swapIntensity( const te::rst::Raster& highResRaster,
      const unsigned int highResRasterBand, te::rst::Raster& ihsRaster )
    {
      if( 
          ( 
            ihsRaster.getNumberOfRows() 
            !=
            highResRaster.getNumberOfRows() 
          ) 
          || 
          ( 
            ihsRaster.getNumberOfColumns() 
            !=
            highResRaster.getNumberOfColumns()
          ) 
        ) 
      {
        return false;
      }
      
      // original intensity statistics
      
      double originalIntensityMean = 0.0;
      double originalIntensityStdDev = 0.0;
      double originalIntensityMin = 0.0;
      double originalIntensityMax = 0.0;
      if( ! getStatistics( ihsRaster, 0, originalIntensityMean, 
        originalIntensityStdDev, originalIntensityMin, originalIntensityMax ) )
      {
        return false;
      }
      
      // load high resolution statistics
      
      double highResMean = 0;
      double highResStdDev = 0;
      double highResMax = 0;
      double highResMin = 0;
      
      if( ! getStatistics( highResRaster, highResRasterBand, highResMean, 
        highResStdDev, highResMin, highResMax ) )
      {
        return false;
      }      
      
      // defining gains and offsets
      
      double inOffset = 0;
      double outOffset = 0;
      double gain = 1.0;
      
      switch( m_inputParameters.m_hrHistoFitMethod )
      {
        case InputParameters::MeanStdDevHistoFitMethod :
        {
          gain = ( ( highResStdDev == 0.0 ) ? 0.0 : ( originalIntensityStdDev / highResStdDev ) );
          inOffset = -1.0 * highResMean;
          outOffset = originalIntensityMean;
         
          break;
        }
        case InputParameters::MinMaxHistoFitMethod :
        {
          
          gain = ( ( (highResMax - highResMin) == 0.0 ) ? 0.0 : 
            ( (originalIntensityMax - originalIntensityMin) 
            / ( highResMax - highResMin) ) );
          inOffset = -1.0 * highResMin;
          outOffset = originalIntensityMin;
          
          break;
        }
        case InputParameters::NoHistoFitMethod :
        {
          gain = 1.0;
          inOffset = 0;
          outOffset = 0;
          
          break;
        }
        default :
        {
          TERP_INSTANCE_LOG_AND_RETURN_FALSE( 
            "Invalid High resolution image histogram fitting method" );
        }
      }
      
      // output blocks to process
      // std::get< 0 > == processed = true / not processed = false
      // std::get< 1 > == output band index
      // std::get< 2 > == output block X
      // std::get< 3 > == output block Y       
     
      std::vector< std::tuple< bool, unsigned int,  unsigned int > >
        rasterBlocksStatus;
      
      {
        std::tuple< bool, unsigned int,  unsigned int > tuple;
        unsigned int blkY = 0;
        unsigned int blkX = 0;
        
        for( blkY = 0 ; blkY <  (unsigned int)ihsRaster.getBand(
          0 )->getProperty()->m_nblocksy ; ++blkY )
        {      
          for( blkX = 0 ; blkX < (unsigned int)ihsRaster.getBand(
            0 )->getProperty()->m_nblocksx ; ++blkX )
          {
            std::get< 0 >( tuple ) = false;
            std::get< 1 >( tuple ) = blkX;
            std::get< 2 >( tuple ) = blkY;
            
            rasterBlocksStatus.push_back( tuple );
          }
        }
      }      
      
      // process raster blocks
      
      unsigned int threadsNumber = m_inputParameters.m_enableThreadedProcessing 
        ? te::common::GetPhysProcNumber() : 1;
      
      te::rst::RasterSynchronizer inputRasterSync( 
        *(te::rst::Raster*)&highResRaster, 
        te::common::RAccess );
      te::rst::RasterSynchronizer outputRasterSync( ihsRaster,
        te::common::WAccess );  
      std::mutex mutex;      
      
      SwapIntensityThreadParameters params;
      params.m_inputRasterSyncPtr = &inputRasterSync;
      params.m_outputRasterSyncPtr = &outputRasterSync;
      params.m_inputRasterBandIdx = highResRasterBand;
      params.m_outputRasterBandIdx = 0;
      params.m_rasterBlocksStatusPtr = &rasterBlocksStatus;
      params.m_returnStatus = true;
      params.m_mutexPtr = &mutex;
      params.m_inOffset = inOffset;
      params.m_outOffset = outOffset;
      params.m_gain = gain;
      
      if( threadsNumber == 1 )
      {
        swapIntensityThreadEntry( &params );
      }
      else
      {
        std::vector< std::shared_ptr< std::thread > > threads;

        for( unsigned int threadIdx = 0 ; threadIdx < threadsNumber ;
          ++threadIdx )
        {
          std::shared_ptr< std::thread > threadPtr( new std::thread(
            &swapIntensityThreadEntry, &params ) );

          threads.push_back( threadPtr );
        };

        for( unsigned int threadIdx = 0 ; threadIdx < threadsNumber ;
          ++threadIdx )
        {
          threads[ threadIdx ]->join();
        };
      }  
      
      return params.m_returnStatus;
    }
    
    void IHSFusion::swapIntensityThreadEntry( SwapIntensityThreadParameters* paramsPtr )
    {
      std::unique_lock< std::mutex > locker( *paramsPtr->m_mutexPtr );

      std::vector< std::tuple< bool, unsigned int,  unsigned int > >&
        rasterBlocksStatus = *paramsPtr->m_rasterBlocksStatusPtr;
      const std::size_t rasterBlocksStatusSize = rasterBlocksStatus.size();
      const unsigned int inputRasterBandIdx = paramsPtr->m_inputRasterBandIdx;
      const unsigned int outputRasterBandIdx = paramsPtr->m_outputRasterBandIdx;
      const double inOffset = paramsPtr->m_inOffset;
      const double outOffset = paramsPtr->m_outOffset;
      const double gain = paramsPtr->m_gain;      

      locker.unlock();

      const te::rst::SynchronizedRaster inputRaster( 9,
        *paramsPtr->m_inputRasterSyncPtr );      
      te::rst::SynchronizedRaster outputRaster( 1,
        *paramsPtr->m_outputRasterSyncPtr );
      
      const unsigned int outputNRows = outputRaster.getNumberOfRows();
      const unsigned int outputNCols = outputRaster.getNumberOfColumns();
      const te::rst::Band& inBand = *inputRaster.getBand( inputRasterBandIdx );
      const double inNoDataValue = inBand.getProperty()->m_noDataValue;
      
      te::rst::Band& outBand = *outputRaster.getBand( outputRasterBandIdx );
      const double outNoDataValue = outBand.getProperty()->m_noDataValue;
      const te::rst::BandProperty& outBandProp = *outBand.getProperty();
      unsigned int outputBlkX = 0;
      unsigned int outputBlkY = 0;
      unsigned int outputBlkStartRow = 0;
      unsigned int outputBlkStartCol = 0;
      unsigned int outputBlkRowBound = 0;
      unsigned int outputBlkColBound = 0;
      unsigned int outputRow = 0;
      unsigned int outputCol = 0;
      double value = 0;
        
      for( std::size_t rasterBlocksStatusIdx = 0 ; rasterBlocksStatusIdx <
        rasterBlocksStatusSize ; ++rasterBlocksStatusIdx )
      {
        locker.lock();

        // looking for the next unprocessed block

        if(
            std::get< 0 >( rasterBlocksStatus[ rasterBlocksStatusIdx ] )
            ==
            false
          )
        {
          std::get< 0 >( rasterBlocksStatus[ rasterBlocksStatusIdx ] ) = true;

          outputBlkX = std::get< 1 >( rasterBlocksStatus[ rasterBlocksStatusIdx ] );
          outputBlkY = std::get< 2 >( rasterBlocksStatus[ rasterBlocksStatusIdx ] );

          locker.unlock();

          // definng the current block limits

          outputBlkStartRow = outBandProp.m_blkh * outputBlkY;
          outputBlkStartRow = std::min( outputBlkStartRow, outputNRows );
          
          outputBlkStartCol = outBandProp.m_blkw * outputBlkX;
          outputBlkStartCol = std::min( outputBlkStartCol, outputNCols );

          outputBlkRowBound = (outBandProp.m_blkh * outputBlkY) + 
            outBandProp.m_blkh;
          outputBlkRowBound = std::min( outputBlkRowBound, outputNRows );
          
          outputBlkColBound = (outBandProp.m_blkw * outputBlkX) + 
            outBandProp.m_blkw;
          outputBlkColBound = std::min( outputBlkColBound, outputNCols );

          for( outputRow = outputBlkStartRow ; outputRow < outputBlkRowBound ; 
            ++outputRow )
          {
            for( outputCol = outputBlkStartCol ; outputCol < outputBlkColBound ; 
              ++outputCol )
            {
              inBand.getValue( outputCol, outputRow, value );
            
              if( value == inNoDataValue )
              {
                outBand.setValue( outputCol, outputRow, outNoDataValue );
              }
              else
              {
                value += inOffset;
                value *= gain;
                value += outOffset;
                
                outBand.setValue( outputCol, outputRow, value );
              }
            }
          }
        }
        else
        {
          locker.unlock();
        }
      }
    }
    
    bool IHSFusion::saveRGBData( const double& rgbMin, const double& rgbMax, 
      const std::string& rType, const std::map< std::string, std::string >& rInfo,
      std::unique_ptr< te::rst::Raster >& outputRasterPtr,
      const te::rst::Raster& ihsRaster )
    {
      outputRasterPtr.reset();
      
      // create output raster
      
      std::vector< te::rst::BandProperty* > outRasterBandsProperties;
      outRasterBandsProperties.push_back( new te::rst::BandProperty(
        *( m_inputParameters.m_lowResRasterPtr->getBand( 
        m_inputParameters.m_lowResRasterRedBandIndex )->getProperty() ) ) );    
      outRasterBandsProperties[ 0 ]->m_blkw = 
        ihsRaster.getBand( 0 )->getProperty()->m_blkw;
      outRasterBandsProperties[ 0 ]->m_blkh = 
        ihsRaster.getBand( 0 )->getProperty()->m_blkh;
      outRasterBandsProperties[ 0 ]->m_nblocksx =
        ihsRaster.getBand( 0 )->getProperty()->m_nblocksx;
      outRasterBandsProperties[ 0 ]->m_nblocksy =
        ihsRaster.getBand( 0 )->getProperty()->m_nblocksy;
      outRasterBandsProperties.push_back( new te::rst::BandProperty(
        *( m_inputParameters.m_lowResRasterPtr->getBand( 
        m_inputParameters.m_lowResRasterGreenBandIndex )->getProperty() ) ) ); 
      outRasterBandsProperties[ 1 ]->m_blkw = 
        outRasterBandsProperties[ 0 ]->m_blkw;
      outRasterBandsProperties[ 1 ]->m_blkh = 
        outRasterBandsProperties[ 0 ]->m_blkh;
      outRasterBandsProperties[ 1 ]->m_nblocksx =
        outRasterBandsProperties[ 0 ]->m_nblocksx;
      outRasterBandsProperties[ 1 ]->m_nblocksy =
        outRasterBandsProperties[ 0 ]->m_nblocksy;
      outRasterBandsProperties.push_back( new te::rst::BandProperty(
        *( m_inputParameters.m_lowResRasterPtr->getBand( 
        m_inputParameters.m_lowResRasterBlueBandIndex )->getProperty() ) ) );         
      outRasterBandsProperties[ 2 ]->m_blkw = 
        outRasterBandsProperties[ 0 ]->m_blkw;
      outRasterBandsProperties[ 2 ]->m_blkh = 
        outRasterBandsProperties[ 0 ]->m_blkh;
      outRasterBandsProperties[ 2 ]->m_nblocksx =
        outRasterBandsProperties[ 0 ]->m_nblocksx;
      outRasterBandsProperties[ 2 ]->m_nblocksy =
        outRasterBandsProperties[ 0 ]->m_nblocksy;

      AlgorithmParametersSerializer paramsSerializer;
      TERP_INSTANCE_TRUE_OR_LOG(  m_inputParameters.serialize( paramsSerializer ),
        "Parameter serialization error" );

      std::map <std::string, std::string> metadata;
      paramsSerializer.getMetaData( metadata );
      
      try
      {
        outputRasterPtr.reset(
          te::rst::RasterFactory::make(
            rType,
            new te::rst::Grid( *ihsRaster.getGrid() ),
            outRasterBandsProperties,
            rInfo,
            metadata,
            nullptr,
            nullptr ) );
      }
      catch(...)
      {
        outputRasterPtr.reset();
      }
      TERP_TRUE_OR_RETURN_FALSE( outputRasterPtr.get(),
        "Output raster creation error" );  
      
      // output blocks to process
      // std::get< 0 > == processed = true / not processed = false
      // std::get< 1 > == output band index
      // std::get< 2 > == output block X
      // std::get< 3 > == output block Y       
     
      std::vector< std::tuple< bool, unsigned int,  unsigned int > >
        rasterBlocksStatus;
      
      {
        std::tuple< bool, unsigned int,  unsigned int > tuple;
        unsigned int blkY = 0;
        unsigned int blkX = 0;
        
        for( blkY = 0 ; blkY <  (unsigned int)outputRasterPtr->getBand(
          0 )->getProperty()->m_nblocksy ; ++blkY )
        {      
          for( blkX = 0 ; blkX < (unsigned int)outputRasterPtr->getBand(
            0 )->getProperty()->m_nblocksx ; ++blkX )
          {
            std::get< 0 >( tuple ) = false;
            std::get< 1 >( tuple ) = blkX;
            std::get< 2 >( tuple ) = blkY;
            
            rasterBlocksStatus.push_back( tuple );
          }
        }
      }      
      
      // process raster blocks
      
      unsigned int threadsNumber = m_inputParameters.m_enableThreadedProcessing 
        ? te::common::GetPhysProcNumber() : 1;
      
      te::rst::RasterSynchronizer inputRasterSync( 
        *(te::rst::Raster*)&ihsRaster, 
        te::common::RAccess );
      te::rst::RasterSynchronizer outputRasterSync( *outputRasterPtr.get(),
        te::common::WAccess );  
      std::mutex mutex;
      
      SaveRGBDataThreadParameters params;
      params.m_inputRasterSyncPtr = &inputRasterSync;
      params.m_outputRasterSyncPtr = &outputRasterSync;
      params.m_rasterBlocksStatusPtr = &rasterBlocksStatus;
      params.m_returnStatus = true;
      params.m_mutexPtr = &mutex;
      params.m_rgbMin = rgbMin;
      params.m_rgbMax = rgbMax;
      
      if( threadsNumber == 1 )
      {
        saveRGBDataThreadEntry( &params );
      }
      else
      {
        std::vector< std::shared_ptr< std::thread > > threads;

        for( unsigned int threadIdx = 0 ; threadIdx < threadsNumber ;
          ++threadIdx )
        {
          std::shared_ptr< std::thread > threadPtr( new std::thread(
            &saveRGBDataThreadEntry, &params ) );

          threads.push_back( threadPtr );
        };

        for( unsigned int threadIdx = 0 ; threadIdx < threadsNumber ;
          ++threadIdx )
        {
          threads[ threadIdx ]->join();
        };
      }  
      
      return params.m_returnStatus;      
    }
    
    void IHSFusion::saveRGBDataThreadEntry( SaveRGBDataThreadParameters* paramsPtr )
    {
      std::unique_lock< std::mutex > locker( *paramsPtr->m_mutexPtr );

      std::vector< std::tuple< bool, unsigned int,  unsigned int > >&
        rasterBlocksStatus = *paramsPtr->m_rasterBlocksStatusPtr;
      const std::size_t rasterBlocksStatusSize = rasterBlocksStatus.size();
      const double rgbMin = paramsPtr->m_rgbMin;
      const double rgbMax = paramsPtr->m_rgbMax;

      locker.unlock();

      const te::rst::SynchronizedRaster inputRaster( 3,
        *paramsPtr->m_inputRasterSyncPtr );      
      te::rst::SynchronizedRaster outputRaster( 3,
        *paramsPtr->m_outputRasterSyncPtr );
      
      const unsigned int outputNRows = outputRaster.getNumberOfRows();
      const unsigned int outputNCols = outputRaster.getNumberOfColumns();
      const te::rst::BandProperty& outBand0Prop = 
        *outputRaster.getBand( 0 )->getProperty();      
        
      const te::rst::Band& iBand = *inputRaster.getBand( 0 );
      const te::rst::Band& hBand = *inputRaster.getBand( 1 );
      const te::rst::Band& sBand = *inputRaster.getBand( 2 );
      te::rst::Band& rBand = *outputRaster.getBand( 0 );
      te::rst::Band& gBand = *outputRaster.getBand( 1 );
      te::rst::Band& bBand = *outputRaster.getBand( 2 );      
      const double inNoDataValuesI = iBand.getProperty()->m_noDataValue;
      const double inNoDataValuesH = hBand.getProperty()->m_noDataValue;
      const double inNoDataValuesS = sBand.getProperty()->m_noDataValue;
      const double outNoDataValuesR = rBand.getProperty()->m_noDataValue;
      const double outNoDataValuesG = gBand.getProperty()->m_noDataValue;        
      const double outNoDataValuesB = bBand.getProperty()->m_noDataValue;
      const double rgbNormFac = ( rgbMax == rgbMin ) ? 0.0 :
        ( rgbMax - rgbMin );         
      const double pi3 = M_PI / 3.0; // 60
      const double twoPi3 = 2.0 * M_PI / 3.0; // 120
      const double fourPi3 = 4.0 * M_PI / 3.0; // 240                
        
      unsigned int outputBlkX = 0;
      unsigned int outputBlkY = 0;
      unsigned int outputBlkStartRow = 0;
      unsigned int outputBlkStartCol = 0;
      unsigned int outputBlkRowBound = 0;
      unsigned int outputBlkColBound = 0;
      unsigned int outputRow = 0;
      unsigned int outputCol = 0;
      double hue = 0;
      double lig = 0;
      double sat = 0;
      double red = 0;
      double green = 0;
      double blue = 0;        
        
      for( std::size_t rasterBlocksStatusIdx = 0 ; rasterBlocksStatusIdx <
        rasterBlocksStatusSize ; ++rasterBlocksStatusIdx )
      {
        locker.lock();

        // looking for the next unprocessed block

        if(
            std::get< 0 >( rasterBlocksStatus[ rasterBlocksStatusIdx ] )
            ==
            false
          )
        {
          std::get< 0 >( rasterBlocksStatus[ rasterBlocksStatusIdx ] ) = true;

          outputBlkX = std::get< 1 >( rasterBlocksStatus[ rasterBlocksStatusIdx ] );
          outputBlkY = std::get< 2 >( rasterBlocksStatus[ rasterBlocksStatusIdx ] );

          locker.unlock();

          // definng the current block limits

          outputBlkStartRow = outBand0Prop.m_blkh * outputBlkY;
          outputBlkStartRow = std::min( outputBlkStartRow, outputNRows );
          
          outputBlkStartCol = outBand0Prop.m_blkw * outputBlkX;
          outputBlkStartCol = std::min( outputBlkStartCol, outputNCols );

          outputBlkRowBound = (outBand0Prop.m_blkh * outputBlkY) + 
            outBand0Prop.m_blkh;
          outputBlkRowBound = std::min( outputBlkRowBound, outputNRows );
          
          outputBlkColBound = (outBand0Prop.m_blkw * outputBlkX) + 
            outBand0Prop.m_blkw;
          outputBlkColBound = std::min( outputBlkColBound, outputNCols );

          for( outputRow = outputBlkStartRow ; outputRow < outputBlkRowBound ; 
            ++outputRow )
          {
            for( outputCol = outputBlkStartCol ; outputCol < outputBlkColBound ; 
              ++outputCol )
            {
              iBand.getValue( outputCol, outputRow, lig );
              hBand.getValue( outputCol, outputRow, hue );
              sBand.getValue( outputCol, outputRow, sat );
              
              if( 
                  ( lig == inNoDataValuesI )
                  ||
                  ( hue == inNoDataValuesH )
                  ||
                  ( sat == inNoDataValuesS )
                )
              {
                rBand.setValue( outputCol, outputRow, outNoDataValuesR );
                gBand.setValue( outputCol, outputRow, outNoDataValuesG );
                bBand.setValue( outputCol, outputRow, outNoDataValuesB );
              }
              else
              {
                if( ( hue == 0.0 ) && ( sat == 0.0 ) )
                { // Gray scale case
                  red = green = blue = ( lig * rgbNormFac );
                }
                else
                { // color case
                  /* Hue inside RG sector */
                  if( hue < twoPi3 )
                  {
                    blue = lig * ( 1.0 - sat );
                    red = lig * ( 1.0 + ( sat * std::cos( hue ) / 
                      std::cos( pi3 - hue ) ) );
                    green = ( 3.0 * lig ) - ( red + blue );
                  }
                  else if( hue < fourPi3 )
                  { /* Hue inside GB sector */
                  
                    hue -= twoPi3;
                    
                    red = lig * ( 1.0 - sat );
                    green = lig * ( 1.0 + ( sat * std::cos( hue ) / 
                      std::cos( pi3 - hue ) ) );
                    blue = ( 3.0 * lig ) - ( red + green );
                  }
                  else
                  { /* Hue inside BR sector */
                  
                    hue -= fourPi3;
                    
                    green = lig * ( 1.0 - sat );
                    blue = lig * ( 1.0 + ( sat * std::cos( hue ) / 
                      std::cos( pi3 - hue ) ) );
                    red = ( 3.0 * lig ) - ( green + blue );
                  }
                  
                  red = ( red * rgbNormFac ) + rgbMin;
                  green = ( green * rgbNormFac ) + rgbMin;
                  blue = ( blue * rgbNormFac ) + rgbMin;
                }
                
                red = MIN( red, rgbMax );
                green = MIN( green, rgbMax );
                blue = MIN( blue, rgbMax );
                
                red = MAX( red, rgbMin );
                green = MAX( green, rgbMin );
                blue = MAX( blue, rgbMin );           
                
                rBand.setValue( outputCol, outputRow, red );
                gBand.setValue( outputCol, outputRow, green );
                bBand.setValue( outputCol, outputRow, blue );
              }
            }
          }
        }
        else
        {
          locker.unlock();
        }
      }
    }
    
    bool IHSFusion::loadUpsampledIHSData( 
      const te::rst::Grid& upsampledGrid,
      const double& rgbMin, 
      const double rgbMax, 
      std::unique_ptr< te::rst::Raster >& ihsRasterPtr ) const
    {
      ihsRasterPtr.reset();
      
      // create output raster
      
      std::vector< te::rst::BandProperty* > outRasterBandsProperties;
      outRasterBandsProperties.push_back( new te::rst::BandProperty(
        *( m_inputParameters.m_lowResRasterPtr->getBand( 
        m_inputParameters.m_lowResRasterRedBandIndex )->getProperty() ) ) );    
      outRasterBandsProperties[ 0 ]->m_blkw = upsampledGrid.getNumberOfColumns();
      outRasterBandsProperties[ 0 ]->m_blkh = 1;
      outRasterBandsProperties[ 0 ]->m_nblocksx = 1;
      outRasterBandsProperties[ 0 ]->m_nblocksy = upsampledGrid.getNumberOfRows();
      outRasterBandsProperties[ 0 ]->m_type = te::dt::DOUBLE_TYPE;
      outRasterBandsProperties[ 0 ]->m_noDataValue = 
        std::numeric_limits< double >::max();
      outRasterBandsProperties.push_back( new te::rst::BandProperty(
        *( m_inputParameters.m_lowResRasterPtr->getBand( 
        m_inputParameters.m_lowResRasterGreenBandIndex )->getProperty() ) ) ); 
      outRasterBandsProperties[ 1 ]->m_blkw = upsampledGrid.getNumberOfColumns();
      outRasterBandsProperties[ 1 ]->m_blkh = 1;
      outRasterBandsProperties[ 1 ]->m_nblocksx = 1;
      outRasterBandsProperties[ 1 ]->m_nblocksy = upsampledGrid.getNumberOfRows();      
      outRasterBandsProperties[ 1 ]->m_type = te::dt::DOUBLE_TYPE;
      outRasterBandsProperties[ 1 ]->m_noDataValue = 
        std::numeric_limits< double >::max();      
      outRasterBandsProperties.push_back( new te::rst::BandProperty(
        *( m_inputParameters.m_lowResRasterPtr->getBand( 
        m_inputParameters.m_lowResRasterBlueBandIndex )->getProperty() ) ) );         
      outRasterBandsProperties[ 2 ]->m_blkw = upsampledGrid.getNumberOfColumns();
      outRasterBandsProperties[ 2 ]->m_blkh = 1;
      outRasterBandsProperties[ 2 ]->m_nblocksx = 1;
      outRasterBandsProperties[ 2 ]->m_nblocksy = upsampledGrid.getNumberOfRows();      
      outRasterBandsProperties[ 2 ]->m_type = te::dt::DOUBLE_TYPE;
      outRasterBandsProperties[ 2 ]->m_noDataValue = 
        std::numeric_limits< double >::max();      

      try
      {
        std::map<std::string, std::string> routInfo;
        routInfo[ "MAXMEMPERCENTUSED" ] = "75";
        
        ihsRasterPtr.reset(
          te::rst::RasterFactory::make(
            "EXPANSIBLE",
            new te::rst::Grid( upsampledGrid ),
            outRasterBandsProperties,
            routInfo,
            nullptr,
            nullptr ) );
      }
      catch(...)
      {
        ihsRasterPtr.reset();
      }
      TERP_TRUE_OR_RETURN_FALSE( ihsRasterPtr.get(),
        "IHS raster creation error" );
      
      // output blocks to process
      // std::get< 0 > == processed = true / not processed = false
      // std::get< 1 > == output band index
      // std::get< 2 > == output block X
      // std::get< 3 > == output block Y       
     
      std::vector< std::tuple< bool, unsigned int,  unsigned int > >
        rasterBlocksStatus;
      
      {
        std::tuple< bool, unsigned int,  unsigned int > tuple;
        unsigned int blkY = 0;
        unsigned int blkX = 0;
        
        for( blkY = 0 ; blkY <  (unsigned int)ihsRasterPtr->getBand(
          0 )->getProperty()->m_nblocksy ; ++blkY )
        {      
          for( blkX = 0 ; blkX < (unsigned int)ihsRasterPtr->getBand(
            0 )->getProperty()->m_nblocksx ; ++blkX )
          {
            std::get< 0 >( tuple ) = false;
            std::get< 1 >( tuple ) = blkX;
            std::get< 2 >( tuple ) = blkY;
            
            rasterBlocksStatus.push_back( tuple );
          }
        }
      }
      
      // process raster blocks
      
      unsigned int threadsNumber = m_inputParameters.m_enableThreadedProcessing 
        ? te::common::GetPhysProcNumber() : 1;
      
      te::rst::RasterSynchronizer inputRasterSync( 
        *(te::rst::Raster*)m_inputParameters.m_lowResRasterPtr, 
        te::common::RAccess );
      te::rst::RasterSynchronizer outputRasterSync( *ihsRasterPtr.get(),
        te::common::WAccess );  
      std::mutex mutex;
      
      LoadUpsampledIHSDataThreadParameters params;
      params.m_inputRasterSyncPtr = &inputRasterSync;
      params.m_outputRasterSyncPtr = &outputRasterSync;
      params.m_rasterBlocksStatusPtr = &rasterBlocksStatus;
      params.m_returnStatus = true;
      params.m_mutexPtr = &mutex;
      params.m_inputRasterRedBandIndex = m_inputParameters.m_lowResRasterRedBandIndex;
      params.m_inputRasterGreenBandIndex = m_inputParameters.m_lowResRasterGreenBandIndex;
      params.m_inputRasterBlueBandIndex = m_inputParameters.m_lowResRasterBlueBandIndex;
      params.m_interpMethod = m_inputParameters.m_interpMethod;
      params.m_interpWindowRadius = 1;
      params.m_rgbMin = rgbMin;
      params.m_rgbMax = rgbMax;
      
      if( threadsNumber == 1 )
      {
        loadUpsampledIHSDataThreadEntry( &params );
      }
      else
      {
        std::vector< std::shared_ptr< std::thread > > threads;

        for( unsigned int threadIdx = 0 ; threadIdx < threadsNumber ;
          ++threadIdx )
        {
          std::shared_ptr< std::thread > threadPtr( new std::thread(
            &loadUpsampledIHSDataThreadEntry, &params ) );

          threads.push_back( threadPtr );
        };

        for( unsigned int threadIdx = 0 ; threadIdx < threadsNumber ;
          ++threadIdx )
        {
          threads[ threadIdx ]->join();
        };
      }  
      
      return params.m_returnStatus;
    }
    
    void IHSFusion::loadUpsampledIHSDataThreadEntry( LoadUpsampledIHSDataThreadParameters* paramsPtr )
    {
      std::unique_lock< std::mutex > locker( *paramsPtr->m_mutexPtr );

      std::vector< std::tuple< bool, unsigned int,  unsigned int > >&
        rasterBlocksStatus = *paramsPtr->m_rasterBlocksStatusPtr;
      const std::size_t rasterBlocksStatusSize = rasterBlocksStatus.size();
      const te::rst::Interpolator::Method interpMethod = paramsPtr->m_interpMethod;
      const unsigned int interpWindowRadius = paramsPtr->m_interpWindowRadius;
      const unsigned int inputRasterRedBandIndex = paramsPtr->m_inputRasterRedBandIndex;
      const unsigned int inputRasterGreenBandIndex = paramsPtr->m_inputRasterGreenBandIndex;            
      const unsigned int inputRasterBlueBandIndex = paramsPtr->m_inputRasterBlueBandIndex;
      const double rgbMin = paramsPtr->m_rgbMin;
      const double rgbMax = paramsPtr->m_rgbMax;

      locker.unlock();

      const te::rst::SynchronizedRaster inputRaster( 27,
        *paramsPtr->m_inputRasterSyncPtr );      
      te::rst::SynchronizedRaster outputRaster( 3,
        *paramsPtr->m_outputRasterSyncPtr );
      
      const unsigned int outputNRows = outputRaster.getNumberOfRows();
      const unsigned int outputNCols = outputRaster.getNumberOfColumns();
      const te::rst::BandProperty& outBand0Prop = 
        *outputRaster.getBand( 0 )->getProperty();      
      const double colsRescaleFactor =
        ((double)( inputRaster.getNumberOfColumns()-1) /
        ((double)(outputNCols-1)));
      const double rowsRescaleFactor =
        ((double)(inputRaster.getNumberOfRows()-1) /
        ((double)(outputNRows-1)));
      const double twoPi = 2.0 * ((double)M_PI);
      const double pi2 = ((double)M_PI) / 2.0;   
      const double rgbNormFac = ( rgbMax == rgbMin ) ? 0.0 :
        ( 1.0 / ( rgbMax - rgbMin ) );        
      const double inNoDataValuesR = inputRaster.getBand(
        inputRasterRedBandIndex )->getProperty()->m_noDataValue;
      const double inNoDataValuesG = inputRaster.getBand(
        inputRasterGreenBandIndex )->getProperty()->m_noDataValue;
      const double inNoDataValuesB = inputRaster.getBand(
        inputRasterBlueBandIndex )->getProperty()->m_noDataValue;
      const double outNoDataValuesI = outputRaster.getBand(
        0 )->getProperty()->m_noDataValue;
      const double outNoDataValuesH = outputRaster.getBand(
        1 )->getProperty()->m_noDataValue;        
      const double outNoDataValuesS = outputRaster.getBand(
        2 )->getProperty()->m_noDataValue;                        
        
      unsigned int outputBlkX = 0;
      unsigned int outputBlkY = 0;
      unsigned int outputBlkStartRow = 0;
      unsigned int outputBlkStartCol = 0;
      unsigned int outputBlkRowBound = 0;
      unsigned int outputBlkColBound = 0;
      unsigned int outputRow = 0;
      unsigned int outputCol = 0;
      double inRow = 0;
      double inCol = 0;
      std::complex< double > redC = 0;
      std::complex< double > greenC = 0;
      std::complex< double > blueC = 0;
      te::rst::Interpolator interpol( &inputRaster,
        interpMethod, std::vector< std::complex<double> >(),
        interpWindowRadius );
      double teta = 0;
      double redNorm = 0, greenNorm = 0, blueNorm = 0;
      double rMinusG = 0, rMinusB = 0;
      double rgbSum = 0;
      double cosValue = 0;      
        
      for( std::size_t rasterBlocksStatusIdx = 0 ; rasterBlocksStatusIdx <
        rasterBlocksStatusSize ; ++rasterBlocksStatusIdx )
      {
        locker.lock();

        // looking for the next unprocessed block

        if(
            std::get< 0 >( rasterBlocksStatus[ rasterBlocksStatusIdx ] )
            ==
            false
          )
        {
          std::get< 0 >( rasterBlocksStatus[ rasterBlocksStatusIdx ] ) = true;

          outputBlkX = std::get< 1 >( rasterBlocksStatus[ rasterBlocksStatusIdx ] );
          outputBlkY = std::get< 2 >( rasterBlocksStatus[ rasterBlocksStatusIdx ] );

          locker.unlock();

          // definng the current block limits

          outputBlkStartRow = outBand0Prop.m_blkh * outputBlkY;
          outputBlkStartRow = std::min( outputBlkStartRow, outputNRows );
          
          outputBlkStartCol = outBand0Prop.m_blkw * outputBlkX;
          outputBlkStartCol = std::min( outputBlkStartCol, outputNCols );

          outputBlkRowBound = (outBand0Prop.m_blkh * outputBlkY) + 
            outBand0Prop.m_blkh;
          outputBlkRowBound = std::min( outputBlkRowBound, outputNRows );
          
          outputBlkColBound = (outBand0Prop.m_blkw * outputBlkX) + 
            outBand0Prop.m_blkw;
          outputBlkColBound = std::min( outputBlkColBound, outputNCols );

          for( outputRow = outputBlkStartRow ; outputRow < outputBlkRowBound ; 
            ++outputRow )
          {
            inRow = ((double)outputRow) * rowsRescaleFactor;
            
            for( outputCol = outputBlkStartCol ; outputCol < outputBlkColBound ; 
              ++outputCol )
            {
              inCol = ((double)outputCol) * colsRescaleFactor;
              
              interpol.getValue( inCol, inRow, redC, inputRasterRedBandIndex );
              interpol.getValue( inCol, inRow, greenC, inputRasterGreenBandIndex );
              interpol.getValue( inCol, inRow, blueC, inputRasterBlueBandIndex );              

              if( 
                  ( redC.real() == inNoDataValuesR ) 
                  || 
                  ( greenC.real() == inNoDataValuesG )
                  ||
                  ( blueC.real() == inNoDataValuesB ) 
                )
              {
                outputRaster.setValue( outputCol, outputRow, outNoDataValuesI, 
                  0 );
                outputRaster.setValue( outputCol, outputRow, outNoDataValuesH, 
                  1 );
                outputRaster.setValue( outputCol, outputRow, outNoDataValuesS, 
                  2 );
              }            
              else
              {
                if( 
                    ( redC.real() == greenC.real() ) 
                    && 
                    ( greenC.real() == blueC.real() ) 
                  ) 
                { // Gray scale case
                  // From Wikipedia:
                  // h = 0 is used for grays though the hue has no geometric 
                  // meaning there, where the saturation s = 0. Similarly, 
                  // the choice of 0 as the value for s when l is equal to 0 or 1 
                  // is arbitrary.        
                  // or green or blue since they all are the same.
                  outputRaster.setValue( outputCol, outputRow, 
                    ( redC.real() * rgbNormFac ), 0 );
                  outputRaster.setValue( outputCol, outputRow, 0, 
                    1 );
                  outputRaster.setValue( outputCol, outputRow, 0, 
                    2 );                  
                }
                else
                { // Color case
                  redNorm = ( redC.real() - rgbMin ) * rgbNormFac;
                  greenNorm = ( greenC.real() - rgbMin ) * rgbNormFac;
                  blueNorm = ( blueC.real() - rgbMin ) * rgbNormFac;
                  
                  rMinusG = redNorm - greenNorm;
                  rMinusB = redNorm - blueNorm;
                  
                  cosValue = sqrt( ( rMinusG * rMinusG ) + ( rMinusB * 
                    ( greenNorm - blueNorm ) ) );
                    
                  if( cosValue == 0.0 )
                  {
                    teta = pi2;
                  }
                  else
                  {
                    cosValue =  ( 0.5 * ( rMinusG + rMinusB )  ) /
                      cosValue;
                    teta = std::acos( cosValue );  
                  }
                    
                  assert( ( cosValue >= (-1.0) ) && ( cosValue <= (1.0) ) );
                    
                  if( blueNorm > greenNorm )
                  {
                    outputRaster.setValue( outputCol, outputRow, ( twoPi - teta ), 
                      1 );
                  }
                  else
                  {
                    outputRaster.setValue( outputCol, outputRow, teta, 
                      1 );                    
                  }
                    
                  rgbSum = ( redNorm + greenNorm + blueNorm );
                  
                  outputRaster.setValue( outputCol, outputRow, 
                    ( rgbSum / 3.0 ), 0 );
                  
                  outputRaster.setValue( outputCol, outputRow, 
                    ( 
                      1.0 
                      - 
                      ( 
                        3 
                        * 
                        std::min( 
                          std::min( 
                            redNorm
                            , 
                            greenNorm 
                          )
                          , 
                          blueNorm 
                        ) 
                        /
                        rgbSum 
                      ) 
                    )
                    , 
                    2 );
                }
              }              
            }
          }
        }
        else
        {
          locker.unlock();
        }
      }
    }
    
    bool IHSFusion::getStatistics( const te::rst::Raster& inputRaster, 
      const unsigned int inputRasterBand, double& mean, double& stdDev,
      double&min, double& max ) const
    {
      std::map<double, unsigned> rHistogram;
      std::map<double, unsigned> iHistogram;
      
      if( ! te::rst::getHistograms( 
          inputRaster,
          inputRasterBand,
          0, // rowStart
          0, // colStart
          inputRaster.getNumberOfRows() - 1, // finalRow
          inputRaster.getNumberOfColumns() - 1, // finalCol
          0, // histoBins
          1, // sampleStep
          0,  // enable Threads
          rHistogram,
          iHistogram )
        )
      {
        return false;
      }      
      
      te::rst::getHistogramStats( rHistogram, min, max, mean, stdDev );
      
      return true;
    }
  }
}   // end namespace te

