/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/rp/MajorityFilter.cpp
  \brief This file contains a class that represents the method to remove pixels from classified image.
*/

#include "MajorityFilter.h"
#include "Macros.h"
#include "Functions.h"
#include "AlgorithmParametersSerializer.h"
#include "../raster/Raster.h"
#include "../raster/RasterFactory.h"
#include "../raster/BandProperty.h"
#include "../raster/Grid.h"
#include "../raster/Band.h"
#include "../raster/BandIterator.h"
#include "../raster/BandIteratorWindow.h"
#include "../common/progress/TaskProgress.h"

#include <memory>

namespace te
{
  namespace rp
  {

    MajorityFilter::InputParameters::InputParameters()
    {
      reset();
    }

    MajorityFilter::InputParameters::InputParameters( const InputParameters& other )
    {
      reset();
      operator=( other );
    }

    MajorityFilter::InputParameters::~InputParameters()
    {
      reset();
    }

    void MajorityFilter::InputParameters::reset() _NOEXCEPT_OP(false)
    {
      m_filterType = InputParameters::InvalidFilterT;
      m_inRasterPtr = nullptr;
      m_inRasterBands.clear();
      m_npixels = 0;
      m_neighbors = 0;
      m_threshold = InputParameters::InvalidThreshold;
      m_enableProgress = false;
    }

    const MajorityFilter::InputParameters& MajorityFilter::InputParameters::operator=(
      const MajorityFilter::InputParameters& params )
    {
      reset();

      m_filterType = params.m_filterType;
      m_inRasterPtr = params.m_inRasterPtr;
      m_inRasterBands = params.m_inRasterBands;
      m_npixels = params.m_npixels;
      m_neighbors = params.m_neighbors;
      m_threshold = params.m_threshold;
      m_enableProgress = params.m_enableProgress;

      return *this;
    }

    te::common::AbstractParameters* MajorityFilter::InputParameters::clone() const
    {
      return new InputParameters( *this );
    }

    bool MajorityFilter::InputParameters::serialize( AlgorithmParametersSerializer& serializer ) const
    {
      TERP_TRUE_OR_RETURN_FALSE( AlgorithmInputParameters::serialize( serializer ),
        "Serializarion error" );

      serializer.setAlgorithmName( "MajorityFilter" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "filterType",
        m_filterType ), "Error adding parameter filterType" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "inRaster",
        m_inRasterPtr->getName() ), "Error adding parameter inRaster" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addMultivaluedParameter( "inRasterBands",
        m_inRasterBands ), "Error adding parameter inRasterBands" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "npixels",
        m_npixels ), "Error adding parameter npixels" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "neighbors",
        m_neighbors ), "Error adding parameter neighbors" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "threshold",
        m_threshold ), "Error adding parameter threshold" );

      return true;
    }

    MajorityFilter::OutputParameters::OutputParameters()
    {
      reset();
    }

    MajorityFilter::OutputParameters::OutputParameters( const OutputParameters& other )
    {
      reset();
      operator=( other );
    }

    MajorityFilter::OutputParameters::~OutputParameters()
    {
      reset();
    }

    void MajorityFilter::OutputParameters::reset() _NOEXCEPT_OP(false)
    {
      m_rType.clear();
      m_rInfo.clear();
      m_outputRasterPtr.reset();
    }

    const MajorityFilter::OutputParameters& MajorityFilter::OutputParameters::operator=(
      const MajorityFilter::OutputParameters& params )
    {
      reset();

      m_rType = params.m_rType;
      m_rInfo = params.m_rInfo;

      return *this;
    }

    te::common::AbstractParameters* MajorityFilter::OutputParameters::clone() const
    {
      return new OutputParameters( *this );
    }

    MajorityFilter::MajorityFilter()
    {
      reset();
    }

    MajorityFilter::~MajorityFilter()
    {

    }

    bool MajorityFilter::execute( AlgorithmOutputParameters& outputParams ) _NOEXCEPT_OP(false)
    {
      if( ! m_isInitialized ) return false;

      MajorityFilter::OutputParameters* outParamsPtr = dynamic_cast<
        MajorityFilter::OutputParameters* >( &outputParams );
      TERP_TRUE_OR_THROW( outParamsPtr, "Invalid paramters" );

      // Initializing the output rasters

      std::unique_ptr< te::rst::Raster > outRasterPtr;
      std::vector< te::rst::BandProperty* > outRasterBandsProperties;

      for (unsigned int inRasterBandsIdx = 0; inRasterBandsIdx <
        m_inputParameters.m_inRasterBands.size(); ++inRasterBandsIdx)
      {
        const unsigned int& inRasterBandIndex =
          m_inputParameters.m_inRasterBands[inRasterBandsIdx];

        assert(inRasterBandIndex <
          m_inputParameters.m_inRasterPtr->getNumberOfBands());

        outRasterBandsProperties.push_back(new te::rst::BandProperty(
          *(m_inputParameters.m_inRasterPtr->getBand(inRasterBandIndex)->getProperty())));
      }

      AlgorithmParametersSerializer paramsSerializer;
      TERP_INSTANCE_TRUE_OR_LOG(  m_inputParameters.serialize( paramsSerializer ),
        "Parameter serialization error" );

      std::map <std::string, std::string> metadata;
      paramsSerializer.getMetaData( metadata );

      try
      {
        outParamsPtr->m_outputRasterPtr.reset(
          te::rst::RasterFactory::make(
            outParamsPtr->m_rType,
            new te::rst::Grid(*(m_inputParameters.m_inRasterPtr->getGrid())),
            outRasterBandsProperties,
            outParamsPtr->m_rInfo,
            metadata,
            nullptr,
            nullptr));
      }
      catch(...)
      {
        outParamsPtr->m_outputRasterPtr.reset();
      }
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE(outParamsPtr->m_outputRasterPtr.get(),
        "Output raster creation error");

      m_outputRaster = outParamsPtr->m_outputRasterPtr.get();

      //Copy input raster to output raster
      te::rst::Copy(*m_inputParameters.m_inRasterPtr, *m_outputRaster);

      switch (m_inputParameters.m_filterType)
      {
        case InputParameters::SimpleFilterT :
        {
          for (unsigned int inRasterBandsIdx = 0; inRasterBandsIdx <
            m_inputParameters.m_inRasterBands.size(); ++inRasterBandsIdx)
          {
            SimpleFilter(*m_inputParameters.m_inRasterPtr, inRasterBandsIdx, inRasterBandsIdx, m_inputParameters.m_enableProgress);
          }

          break;
        }
        case InputParameters::FullFilterT :
        {
          for (unsigned int inRasterBandsIdx = 0; inRasterBandsIdx <
            m_inputParameters.m_inRasterBands.size(); ++inRasterBandsIdx)
          {
            FullFilter(*m_inputParameters.m_inRasterPtr, inRasterBandsIdx, inRasterBandsIdx, m_inputParameters.m_enableProgress);
          }
          break;
        }
        default :
        {
          TERP_LOG_AND_THROW("Invalid filter type");
        }
      }

      return true;
    }

    void MajorityFilter::reset() _NOEXCEPT_OP(false)
    {
      Algorithm::reset();

      m_isInitialized = false;
      m_inputParameters.reset();
    }

    bool MajorityFilter::initialize( const AlgorithmInputParameters& inputParams ) _NOEXCEPT_OP(false)
    {
      reset();

      MajorityFilter::InputParameters const* inputParamsPtr = dynamic_cast<
        MajorityFilter::InputParameters const* >( &inputParams );
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( inputParamsPtr, TE_TR( "Invalid parameters" ) );

      m_inputParameters = *inputParamsPtr;

      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( m_inputParameters.m_filterType !=
        InputParameters::InvalidFilterT,
        TE_TR( "Invalid filter type" ) );

      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( m_inputParameters.m_inRasterPtr,
        TE_TR( "Invalid raster pointer" ) );
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE(
        m_inputParameters.m_inRasterPtr->getAccessPolicy() & te::common::RAccess,
        TE_TR( "Invalid raster" ) );

      if (m_inputParameters.m_filterType == InputParameters::SimpleFilterT)
      {
        TERP_INSTANCE_TRUE_OR_RETURN_FALSE(m_inputParameters.m_npixels > 0,
          TE_TR("Invalid number of pixels"));
      }
      else
      {
        TERP_INSTANCE_TRUE_OR_RETURN_FALSE(m_inputParameters.m_neighbors > 0,
          TE_TR("Invalid neighbors number"));

        TERP_INSTANCE_TRUE_OR_RETURN_FALSE(m_inputParameters.m_threshold !=
          InputParameters::InvalidThreshold,
          TE_TR("Invalid threshold"));
      }
      int nBands = m_inputParameters.m_inRasterPtr->getNumberOfBands();
      for (unsigned int i = 0; i < nBands; ++i)
      {
        m_inputParameters.m_inRasterBands.push_back(i);
      }

      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( m_inputParameters.m_inRasterBands.size() > 0,
        TE_TR( "Invalid raster bands number" ) );
      for( unsigned int inRasterBandsIdx = 0 ; inRasterBandsIdx <
        m_inputParameters.m_inRasterBands.size() ; ++inRasterBandsIdx )
      {
        TERP_INSTANCE_TRUE_OR_RETURN_FALSE(
          m_inputParameters.m_inRasterBands[ inRasterBandsIdx ] <
          m_inputParameters.m_inRasterPtr->getNumberOfBands(),
          TE_TR( "Invalid raster bands" ) );
      }

      m_isInitialized = true;

      return true;
    }

    bool MajorityFilter::isInitialized() const
    {
      return m_isInitialized;
    }

    bool MajorityFilter::SimpleFilter(const te::rst::Raster& srcRaster,
      const unsigned int srcBandIdx,
      const unsigned int dstBandIdx, const bool useProgress)
    {
      TERP_DEBUG_TRUE_OR_THROW(srcRaster.getNumberOfColumns() ==
        m_outputRaster->getNumberOfColumns(), "Internal error");
      TERP_DEBUG_TRUE_OR_THROW(srcRaster.getNumberOfRows() ==
        m_outputRaster->getNumberOfRows(), "Internal error");
      TERP_DEBUG_TRUE_OR_THROW(srcBandIdx < srcRaster.getNumberOfBands(),
        "Internal error");
      TERP_DEBUG_TRUE_OR_THROW(dstBandIdx < m_outputRaster->getNumberOfBands(),
        "Internal error");

      const unsigned int nRows = static_cast<unsigned int>(m_outputRaster->getNumberOfRows());
      const unsigned int nCols = static_cast<unsigned int>(m_outputRaster->getNumberOfColumns());

      int size = nRows * nCols;

      std::vector<int> vecline;
      std::vector<int> veccol;
      vecline.resize(size);
      veccol.resize(size);

      std::unique_ptr< te::common::TaskProgress > task;
      if (useProgress)
      {
        task.reset(new te::common::TaskProgress(TE_TR("Majority Filter"),
          te::common::TaskProgress::UNDEFINED,
          static_cast<int>(srcRaster.getNumberOfRows()) - 2));
      }

      te::rst::Band& dstBand = *m_outputRaster->getBand(dstBandIdx);

      for (unsigned int row = 0; row < nRows; ++row)
      {
        for (unsigned int col = 0; col < nCols; ++col)
        {
          double value = 0;
          dstBand.getValue(col, row, value);
          
          int contClass[256];
          for (int i = 0; i < 256; i++)
            contClass[i] = 0;

          int totalPixels = 0;
          int index = 0;

          findNeighbors(dstBand, col, row, col, row, index, totalPixels, contClass, veccol, vecline);

          if (totalPixels <= m_inputParameters.m_npixels)
          {
            //define most frequent class as a new one
            double newClass = 0;
            for (int i = 0; i < 255; i++)
              if (contClass[i] > contClass[(int)newClass])
                newClass = i;

            for (int i = 0; i < index; i++)
              dstBand.setValue(veccol[i], vecline[i], newClass);
          }
          else
          {
            for (int i = 0; i < index; i++)
              dstBand.setValue(veccol[i], vecline[i], value);
          }
        }

        if (useProgress)
        {
          task->pulse();
          if (!task->isActive())
            return false;
        }
      }

      return true;
    }

    bool MajorityFilter::FullFilter(const te::rst::Raster& srcRaster,
      const unsigned int srcBandIdx, const unsigned int dstBandIdx, const bool useProgress)
    {
      TERP_DEBUG_TRUE_OR_THROW(srcRaster.getNumberOfColumns() ==
        m_outputRaster->getNumberOfColumns(), "Internal error");
      TERP_DEBUG_TRUE_OR_THROW(srcRaster.getNumberOfRows() ==
        m_outputRaster->getNumberOfRows(), "Internal error");
      TERP_DEBUG_TRUE_OR_THROW(srcBandIdx < srcRaster.getNumberOfBands(),
        "Internal error");
      TERP_DEBUG_TRUE_OR_THROW(dstBandIdx < m_outputRaster->getNumberOfBands(),
        "Internal error");

      const unsigned int nRows = static_cast<unsigned int>(srcRaster.getNumberOfRows());
      const unsigned int nCols = static_cast<unsigned int>(srcRaster.getNumberOfColumns());

      std::unique_ptr< te::common::TaskProgress > task;
      if (useProgress)
      {
        task.reset(new te::common::TaskProgress(TE_TR("Majority Filter"),
          te::common::TaskProgress::UNDEFINED,
          static_cast<int>(srcRaster.getNumberOfRows()) - 2));
      }

      te::rst::Band& dstBand = *m_outputRaster->getBand(dstBandIdx);

      double value = 0;

      for (unsigned int row = 0; row < nRows; ++row)
      {
        for (unsigned int col = 0; col < nCols; ++col)
        {
          dstBand.getValue(col, row, value);

          int contClass[256];
          for (int i = 0; i < 256; i++)
            contClass[i] = 0;

          if (m_inputParameters.m_neighbors == 4)
          {
            find4Neighbors(dstBand, col, row, contClass);
          }
          else
          {
            find8Neighbors(dstBand, col, row, contClass);
          }

          //define most frequent class as a new one
          double newClass = 0;
          for (int i = 0; i < 255; i++)
          {
            if (contClass[i] > contClass[(int)newClass])
              newClass = i;
          }

          int minPixels = m_inputParameters.m_neighbors / 2;
          if (m_inputParameters.m_threshold == InputParameters::MajorityThreshold)//Majority
          {
            if (contClass[(int)newClass] > minPixels)
              dstBand.setValue(col, row, newClass);
          }
          else//Half
          {
            if (contClass[(int)newClass] >= minPixels)
              dstBand.setValue(col, row, newClass);
          }
        }

        if (useProgress)
        {
          task->pulse();
          if (!task->isActive())
            return false;
        }
      }

      return true;
    }

    void MajorityFilter::findNeighbors(te::rst::Band& dstBand, unsigned int lcol, unsigned int lrow, unsigned int col, unsigned int row, int& index, int& totalPixel, int* contClass, std::vector<int>& veccol, std::vector<int>& vecrow)
    {
      if (totalPixel > m_inputParameters.m_npixels)
        return;

      totalPixel++;

      const unsigned int nRows = static_cast<unsigned int>(m_outputRaster->getNumberOfRows());
      const unsigned int nCols = static_cast<unsigned int>(m_outputRaster->getNumberOfColumns());

      double lclass = 0;
      double classValue = 0;

      dstBand.getValue(lcol, lrow, lclass);
      dstBand.getValue(col, row, classValue);
    
      dstBand.setValue(col, row, 255);

      veccol[index] = col;
      vecrow[index++] = row;

      if (col < nCols - 1)
      {
        double value = 0;
        dstBand.getValue(col + 1, row, value); //check right neighbor

        if (value == classValue)
          findNeighbors(dstBand, col, row, col + 1, row, index, totalPixel, contClass, veccol, vecrow);
        else
          contClass[(int)value]++;
      }

      if (row < nRows - 1)
      {
        double value = 0;
        dstBand.getValue(col, row + 1, value); //check down neighbor

        if(value == classValue)
          findNeighbors(dstBand, col, row, col, row + 1, index, totalPixel, contClass, veccol, vecrow);
        else
          contClass[(int)value]++;
      }

      if (col > 0)
      {
        double value = 0;
        dstBand.getValue(col - 1, row, value); //check left neighbor

        if (value == classValue)
          findNeighbors(dstBand, col, row, col - 1, row, index, totalPixel, contClass, veccol, vecrow);
        else
          contClass[(int)value]++;
      }

      if (row > 0)
      {
        double value;
        dstBand.getValue(col, row - 1, value); //check up neighbor

        if (value == classValue)
          findNeighbors(dstBand, col, row, col, row - 1, index, totalPixel, contClass, veccol, vecrow);
        else
          contClass[(int)value]++;
      }
    }

    void MajorityFilter::find4Neighbors(te::rst::Band& dstBand, unsigned int col, unsigned int row, int* contClass)
    {
      const unsigned int nRows = static_cast<unsigned int>(m_outputRaster->getNumberOfRows());
      const unsigned int nCols = static_cast<unsigned int>(m_outputRaster->getNumberOfColumns());

      double classValue = 0;
      dstBand.getValue(col, row, classValue);

      double value = 0;

      if (col > 0)
      {
        dstBand.getValue(col - 1, row, value);

        if (classValue != value)
          contClass[(int)value]++;
      }

      if (row > 0)
      {
        dstBand.getValue(col, row - 1, value);

        if (classValue != value)
          contClass[(int)value]++;
      }

      if (col < nCols - 1)
      {
        dstBand.getValue(col + 1, row, value);
        
        if (classValue != value)
          contClass[(int)value]++;
      }

      if (row < nRows - 1)
      {
        dstBand.getValue(col, row + 1, value);

        if (classValue != value)
          contClass[(int)value]++;
      }
    }

    void MajorityFilter::find8Neighbors(te::rst::Band& dstBand, unsigned int col, unsigned int row, int* contClass)
    {
      const unsigned int nRows = static_cast<unsigned int>(m_outputRaster->getNumberOfRows());
      const unsigned int nCols = static_cast<unsigned int>(m_outputRaster->getNumberOfColumns());

      double classValue = 0;
      dstBand.getValue(col, row, classValue);

      double value = 0;

      if (col > 0)
      {
        dstBand.getValue(col - 1, row, value);

        if (classValue != value)
          contClass[(int)value]++;

        if (row > 0)
        {
          dstBand.getValue(col - 1, row - 1, value);

          if (classValue != value)
            contClass[(int)value]++;
        }

        if (row < nRows - 1)
        {
          dstBand.getValue(col - 1, row + 1, value);

          if (classValue != value)
            contClass[(int)value]++;
        }
      }

      if (row > 0)
      {
        dstBand.getValue(col, row - 1, value);

        if (classValue != value)
          contClass[(int)value]++;

        if (col < nCols - 1)
        {
          dstBand.getValue(col + 1, row - 1, value);

          if (classValue != value)
            contClass[(int)value]++;
        }
      }

      if (col < nCols - 1)
      {
        dstBand.getValue(col + 1, row, value);

        if (classValue != value)
          contClass[(int)value]++;

        if (row < nRows - 1)
        {
          dstBand.getValue(col + 1, row + 1, value);

          if (classValue != value)
            contClass[(int)value]++;
        }
      }

      if (row < (nRows - 1))
      {
        dstBand.getValue(col, row + 1, value);

        if (classValue != value)
          contClass[(int)value]++;
      }
    }
    
  } // end namespace rp
}   // end namespace te

