/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/rp/MixtureModelLinearStrategy.cpp

  \brief Raster linear strategy for mixture model classification.
*/

// TerraLib
#include "../common/MatrixUtils.h"
#include "../common/progress/TaskProgress.h"
#include "../raster/Grid.h"
#include "../raster/Utils.h"
#include "Functions.h"
#include "Macros.h"
#include "MixtureModelLinearStrategy.h"
#include "Functions.h"
#include "AlgorithmParametersSerializer.h"

// Boost
#include <boost/numeric/ublas/io.hpp>
#include <boost/numeric/ublas/matrix.hpp>

namespace
{
  static te::rp::MixtureModelLinearStrategyFactory mixtureModelLinearStrategyFactoryInstance;
}

/* -------------------------------------------------------------------------- */

te::rp::MixtureModelLinearStrategy::Parameters::Parameters()
{
  reset();
}

te::rp::MixtureModelLinearStrategy::Parameters::~Parameters()
{
}

const te::rp::MixtureModelLinearStrategy::Parameters& 
  te::rp::MixtureModelLinearStrategy::Parameters::operator=(
  const te::rp::MixtureModelLinearStrategy::Parameters& other )
{
  reset();
  
  m_methodType = other.m_methodType;

  return *this;
}

void te::rp::MixtureModelLinearStrategy::Parameters::reset() 
{
  m_methodType = LST;
}

te::common::AbstractParameters* te::rp::MixtureModelLinearStrategy::Parameters::clone() const
{
  return new te::rp::MixtureModelLinearStrategy::Parameters(*this);
}

bool te::rp::MixtureModelLinearStrategy::Parameters::serialize( AlgorithmParametersSerializer& serializer ) const
{
  TERP_TRUE_OR_RETURN_FALSE( MixtureModelStrategyParameters::serialize( serializer ),
    "Serialization error" );

  TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "methodType",
    m_methodType ), "Error adding parameter methodType" );

  return true;
}

/* -------------------------------------------------------------------------- */

te::rp::MixtureModelLinearStrategy::MixtureModelLinearStrategy()
{
  m_isInitialized = false;
}

te::rp::MixtureModelLinearStrategy::~MixtureModelLinearStrategy()
{
}

bool te::rp::MixtureModelLinearStrategy::initialize(te::rp::StrategyParameters const* const strategyParams) 
{
  m_isInitialized = false;

  te::rp::MixtureModelLinearStrategy::Parameters const* paramsPtr = 
    dynamic_cast<te::rp::MixtureModelLinearStrategy::Parameters const*>(strategyParams);

  if(!paramsPtr)
    return false;
  
  TERP_INSTANCE_TRUE_OR_RETURN_FALSE( paramsPtr->m_methodType != 
    te::rp::MixtureModelLinearStrategy::Parameters::InvalidMethodT,
    "Invalid method type" );

  m_parameters = *(paramsPtr);

  m_isInitialized = true;

  return true;
}

bool te::rp::MixtureModelLinearStrategy::execute(const te::rst::Raster& inputRaster, 
  const std::vector<unsigned int>& inputRasterBands,
  const std::vector<std::string>& inputSensorBands, 
  const std::map<std::string, std::vector<double> >& components,
  std::vector<te::rst::Raster*>& outputRaster, 
  const bool normalize, 
  const bool enableProgressInterface ) 
{
  TERP_INSTANCE_TRUE_OR_RETURN_FALSE(m_isInitialized, "Instance not initialized")
  TERP_INSTANCE_TRUE_OR_RETURN_FALSE( !inputRasterBands.empty(), "Invalid input raster bands" );
  TERP_INSTANCE_TRUE_OR_RETURN_FALSE( !inputSensorBands.empty(), "Invalid sensor bands" );    
  
  switch( m_parameters.m_methodType )
  {
    case Parameters::RLS3T :
    {
      return executeRLS3T( inputRaster, inputRasterBands, inputSensorBands,
        components, outputRaster, normalize, enableProgressInterface );
      break;
    }
    case Parameters::LST :
    {
      return executeLST( inputRaster, inputRasterBands, inputSensorBands,
        components, outputRaster, normalize, enableProgressInterface );
      break;
    }    
    default :
    {
      TERP_INSTANCE_LOG_AND_RETURN_FALSE( "Invalid method" );
    }
  }
  
  return false;
}

bool te::rp::MixtureModelLinearStrategy::executeRLS3T(const te::rst::Raster& inputRaster, 
  const std::vector<unsigned int>& inputRasterBands,
  const std::vector<std::string>& inputSensorBands, 
  const std::map<std::string, std::vector<double> >& components,
  std::vector<te::rst::Raster*>& outputRaster, 
  const bool normalize, 
  const bool enableProgressInterface ) 
{
  TERP_INSTANCE_TRUE_OR_RETURN_FALSE( components.size() == 3, "3 components are required" );
  TERP_INSTANCE_TRUE_OR_RETURN_FALSE( inputRasterBands.size() == 4, "4 bands are required" );
  
  if (m_AMatrix.size1() == 0)
  {
    TERP_INSTANCE_TRUE_OR_RETURN_FALSE( generateTransformMatrix( inputRasterBands, 
      inputSensorBands, components ), "Unable to calculate the transformation matrix" );
  }
  else
  {
    TERP_INSTANCE_TRUE_OR_RETURN_FALSE( m_AMatrix.size1() == components.size(), 
      "Invalid transformation matrix" );
    TERP_INSTANCE_TRUE_OR_RETURN_FALSE( m_AMatrix.size2() == inputRasterBands.size(), 
      "Invalid transformation matrix" );    
  }
  
  // Create temporary output raster
  
  std::unique_ptr<te::rst::Raster> tmpRaster;
  const double tmpRasterNoDataValue = std::numeric_limits< float >::max();
  
  {
    te::rst::Grid* tmpgrid = new te::rst::Grid(*outputRaster[0]->getGrid());

    std::vector<te::rst::BandProperty*> tmpbandsProperties;
    
    for( unsigned int outBandidx = 0 ; outBandidx < outputRaster[0]->getNumberOfBands() ;
      ++outBandidx )
    {
      te::rst::BandProperty* bprop = new te::rst::BandProperty(
        *(outputRaster[0]->getBand(outBandidx)->getProperty()));
      bprop->m_colorInterp = te::rst::GrayIdxCInt;
      bprop->m_type = te::dt::FLOAT_TYPE;
      bprop->m_noDataValue = tmpRasterNoDataValue;
      tmpbandsProperties.push_back(bprop);
    }

    try
    {
      tmpRaster.reset(te::rst::RasterFactory::make("EXPANSIBLE", tmpgrid, 
        tmpbandsProperties, std::map<std::string, std::string>()));
    }
    catch(...)
    {
      TERP_INSTANCE_LOG_AND_RETURN_FALSE( "Memory raster creation error" );
    }  
  }
  
  // calculating the mix model coeficients
  
  const double A1 =
    (
      ( m_AMatrix( 0, 0 ) * m_AMatrix( 0, 0 ) ) /* a11 * a11 */
      + 
      ( m_AMatrix( 0, 1 ) * m_AMatrix( 0, 1 ) ) /* a21 * a21 */
      + 
      ( m_AMatrix( 0, 2 ) * m_AMatrix( 0, 2 ) ) /* a31 * a31 */
      + 
      ( m_AMatrix( 0, 3 ) * m_AMatrix( 0, 3 ) ) /* a41 * a41 */
      + 
      ( m_AMatrix( 2, 0 ) * m_AMatrix( 2, 0 ) ) /* a13 * a13 */
      + 
      ( m_AMatrix( 2, 1 ) * m_AMatrix( 2, 1 ) ) /* a23 * a23 */ 
      + 
      ( m_AMatrix( 2, 2 ) * m_AMatrix( 2, 2 ) ) /* a33 * a33 */ 
      + 
      ( m_AMatrix( 2, 3 ) * m_AMatrix( 2, 3 ) ) /* a43 * a43 */ 
      - 
      (
        2 
        * 
        ( 
          ( m_AMatrix( 0, 0 ) * m_AMatrix( 2, 0 ) ) /* a11 * a13 */
          + 
          ( m_AMatrix( 0, 1 ) * m_AMatrix( 2, 1 ) ) /* a21 * a23 */
          + 
          ( m_AMatrix( 0, 2 ) * m_AMatrix( 2, 2 ) ) /* a31 * a33 */
          + 
          ( m_AMatrix( 0, 3 ) * m_AMatrix( 2, 3 ) ) /* a41 * a43 */
        )
      )
   );
  
  const double A2 = 
    (
      ( m_AMatrix( 1, 0 ) * m_AMatrix( 1, 0 ) ) /* a12 * a12 */
      + 
      ( m_AMatrix( 1, 1 ) * m_AMatrix( 1, 1 ) ) /* a22 * a22 */
      + 
      ( m_AMatrix( 1, 2 ) * m_AMatrix( 1, 2 ) ) /* a32 * a32 */
      + 
      ( m_AMatrix( 1, 3 ) * m_AMatrix( 1, 3 ) ) /* a42 * a42 */
      + 
      ( m_AMatrix( 2, 0 ) * m_AMatrix( 2, 0 ) ) /* a13 * a13 */
      + 
      ( m_AMatrix( 2, 1 ) * m_AMatrix( 2, 1 ) ) /* a23 * a23 */
      + 
      ( m_AMatrix( 2, 2 ) * m_AMatrix( 2, 2 ) ) /* a33 * a33 */
      + 
      ( m_AMatrix( 2, 3 ) * m_AMatrix( 2, 3 ) ) /* a43 * a43 */
      - 
      (
        2
        *
        ( 
          ( m_AMatrix( 1, 0 ) * m_AMatrix( 2, 0 ) ) /* a12 * a13 */
          + 
          ( m_AMatrix( 1, 1 ) * m_AMatrix( 2, 1 ) ) /* a22 * a23 */
          + 
          ( m_AMatrix( 1, 2 ) * m_AMatrix( 2, 2 ) ) /* a32 * a33 */
          + 
          ( m_AMatrix( 1, 3 ) * m_AMatrix( 2, 3 ) ) /* a42 * a43 */
        )
      )
    );
    
  const double A3 = 
    ( 
      2 
      *
      ( 
        ( m_AMatrix( 2, 0 ) * m_AMatrix( 2, 0 ) ) /* a13 ^2 */
        + 
        ( m_AMatrix( 2, 1 ) * m_AMatrix( 2, 1 ) ) /* a23 ^2 */
        + 
        ( m_AMatrix( 2, 2 ) * m_AMatrix( 2, 2 ) ) /* a33 ^2 */
        + 
        ( m_AMatrix( 2, 3 ) * m_AMatrix( 2, 3 ) ) /* a43 ^2 */
        + 
        ( m_AMatrix( 0, 0 ) * m_AMatrix( 1, 0 ) ) /* a11 * a12 */
        + 
        ( m_AMatrix( 0, 1 ) * m_AMatrix( 1, 1 ) ) /* a21 * a22 */
        + 
        ( m_AMatrix( 0, 2 ) * m_AMatrix( 1, 2 ) ) /* a31 * a32 */
        + 
        ( m_AMatrix( 0, 3 ) * m_AMatrix( 1, 3 ) ) /* a41 * a42 */ 
        -
        ( m_AMatrix( 0, 0 ) * m_AMatrix( 2, 0 ) ) /* a11 * a13 */
        - 
        ( m_AMatrix( 0, 1 ) * m_AMatrix( 2, 1 ) ) /* a21 * a23 */
        - 
        ( m_AMatrix( 0, 2 ) * m_AMatrix( 2, 2 ) ) /* a31 * a33 */
        - 
        ( m_AMatrix( 0, 3 ) * m_AMatrix( 2, 3 ) ) /* a41 * a43 */
        -
        ( m_AMatrix( 1, 0 ) * m_AMatrix( 2, 0 ) ) /* a12 * a13 */
        - 
        ( m_AMatrix( 1, 1 ) * m_AMatrix( 2, 1 ) ) /* a22 * a23 */
        - 
        ( m_AMatrix( 1, 2 ) * m_AMatrix( 2, 2 ) ) /* a32 * a33 */
        - 
        ( m_AMatrix( 1, 3 ) * m_AMatrix( 2, 3 ) ) /* a42 * a43 */
      )
    );

  const unsigned int nComponents = (unsigned int)components.size();
  const unsigned int nBands = (unsigned int)inputRasterBands.size();
  const double inNoDataValue = inputRaster.getBand( inputRasterBands[0] )->getProperty()->m_noDataValue;
  const double outNoDataValue = outputRaster[0]->getBand(0)->getProperty()->m_noDataValue;
  const double outErrorNoDataValue = ( outputRaster.size() > 1 ) ?
   outputRaster[1]->getBand(0)->getProperty()->m_noDataValue : 
   std::numeric_limits< double >::max();
  const bool computErrors = (outputRaster.size() > 1);

   boost::numeric::ublas::matrix<double> matrixR = 
     boost::numeric::ublas::matrix<double>( inputRasterBands.size(), 1);

  std::unique_ptr< te::common::TaskProgress > task;
  if( enableProgressInterface )
  {
    task.reset( new te::common::TaskProgress( TE_TR("Linear Mixture Model"), 
      te::common::TaskProgress::UNDEFINED, 2 * inputRaster.getNumberOfRows() ) );
  }
  
  double inputRasterValue = 0;
  double bandError = 0;
  double x1 = 0; // component 1 proportion
  double x2 = 0; // component 2 proportion
  double x3 = 0; // component 3 proportion
  bool pixelIsValid = true;
  double globalMin = std::numeric_limits< double >::max();
  double globalMax = -1.0 * std::numeric_limits< double >::max();
  
  m_min.clear();
  m_min.resize( nComponents, std::numeric_limits< double >::max() );
  
  m_max.clear();
  m_max.resize( nComponents, -1.0 * std::numeric_limits< double >::max() );  
  
  m_minerror.clear();
  m_minerror.resize( nBands, std::numeric_limits< double >::max() );
  
  m_maxerror.clear();
  m_maxerror.resize( nBands, -1.0 * std::numeric_limits< double >::max() );    

  for (unsigned r = 0; r < inputRaster.getNumberOfRows(); r++)
  {
    for (unsigned c = 0; c < inputRaster.getNumberOfColumns(); c++)
    {
      pixelIsValid = true;
      
      for (unsigned b = 0; b < nBands; b++)
      {
        inputRaster.getValue(c, r, inputRasterValue, inputRasterBands[b]);
        
        if (inputRasterValue == inNoDataValue )
        {
          pixelIsValid = false;
          break;
        }
        else
        {
          matrixR(b, 0) = inputRasterValue;
        }
      }
      
      if( pixelIsValid )
      {
        const double A4 = 
          ( 
            2 
            *
            ( 
              ( -( m_AMatrix( 2, 0 ) * m_AMatrix( 2, 0 ) ) ) /* a13 * a13 */
              - 
              ( m_AMatrix( 2, 1 ) * m_AMatrix( 2, 1 ) ) /* a23 * a23 */
              - 
              ( m_AMatrix( 2, 2 ) * m_AMatrix( 2, 2 ) ) /* a33 * a33 */
              - 
              ( m_AMatrix( 2, 3 ) * m_AMatrix( 2, 3 ) ) /* a43 * a43 */
              + 
              ( m_AMatrix( 0, 0 ) *  m_AMatrix( 2, 0 ) ) /* a11 * a13 */
              + 
              ( m_AMatrix( 0, 1 ) * m_AMatrix( 2, 1 ) ) /* a21 * a23 */
              + 
              ( m_AMatrix( 0, 2 ) * m_AMatrix( 2, 2 ) ) /* a31 * a33 */
              + 
              ( m_AMatrix( 0, 3 ) * m_AMatrix( 2, 3 ) ) /* a41 * a43 */
              - 
              ( m_AMatrix( 0, 0 ) * matrixR( 0, 0 ) ) /* a11 * r1 */
              - 
              ( m_AMatrix( 0, 1 ) * matrixR( 1, 0 ) ) /* a21 * r2 */
              - 
              ( m_AMatrix( 0, 2 ) * matrixR( 2, 0 ) ) /* a31 * r3 */
              - 
              ( m_AMatrix( 0, 3 ) * matrixR( 3, 0 ) ) /* a41 * r4 */
              + 
              ( m_AMatrix( 2, 0 ) * matrixR( 0, 0 ) ) /* a13 * r1 */
              +
              ( m_AMatrix( 2, 1 ) * matrixR( 1, 0 ) ) /* a23 * r2 */
              + 
              ( m_AMatrix( 2, 2 ) * matrixR( 2, 0 ) ) /* a33 * r3 */
              + 
              ( m_AMatrix( 2, 3 ) * matrixR( 3, 0 ) ) /* a43 * r4 */
            )
          );
        
        const double A5 = 
          (
            2 
            *
            ( 
              ( -(m_AMatrix( 2, 0 ) * m_AMatrix( 2, 0 ) ) ) /* a13 * a13 */
              - 
              ( m_AMatrix( 2, 1 ) * m_AMatrix( 2, 1 ) ) /* a23 * a23 */
              - 
              ( m_AMatrix( 2, 2 ) * m_AMatrix( 2, 2 ) ) /* a33 * a33 */
              - 
              ( m_AMatrix( 2, 3 ) * m_AMatrix( 2, 3 ) ) /* a43 * a43 */
              + 
              ( m_AMatrix( 1, 0 ) * m_AMatrix( 2, 0 ) ) /* a12 * a13 */
              +
              ( m_AMatrix( 1, 1 ) * m_AMatrix( 2, 1 ) ) /* a22 * a23 */
              + 
              ( m_AMatrix( 1, 2 ) * m_AMatrix( 2, 2 ) ) /* a32 * a33 */
              + 
              ( m_AMatrix( 1, 3 ) * m_AMatrix( 2, 3 ) ) /* a42 * a43 */
              - 
              ( m_AMatrix( 1, 0 ) * matrixR( 0, 0 ) ) /* a12 * r1 */
              - 
              ( m_AMatrix( 1, 1 ) * matrixR( 1, 0 ) ) /* a22 * r2 */
              - 
              ( m_AMatrix( 1, 2 ) * matrixR( 2, 0 ) ) /* a32 * r3 */
              - 
              ( m_AMatrix( 1, 3 ) * matrixR( 3, 0 ) ) /* a42 * r4 */
              + 
              ( m_AMatrix( 2, 0 ) * matrixR( 0, 0 ) ) /* a13 * r1 */
              +
              ( m_AMatrix( 2, 1 ) * matrixR( 1, 0 ) ) /* a23 * r2 */
              + 
              ( m_AMatrix( 2, 2 ) * matrixR( 2, 0 ) ) /* a33 * r3 */
              + 
              ( m_AMatrix( 2, 3 ) * matrixR( 3, 0 ) ) /* a43 * r4 */
            )
          );
          
//         const double A6 = 
//           ( 
//             ( m_AMatrix( 2, 0 ) * m_AMatrix( 2, 0 ) ) /* a13 * a13 */
//             + 
//             ( m_AMatrix( 2, 1 ) * m_AMatrix( 2, 1 ) ) /* a23 * a23 */
//             + 
//             ( m_AMatrix( 2, 2 ) * m_AMatrix( 2, 2 ) ) /* a33 * a33 */
//             + 
//             ( m_AMatrix( 2, 3 ) * m_AMatrix( 2, 3 ) ) /* a43 * a43 */
//             + 
//             ( matrixR( 0, 0 ) * matrixR( 0, 0 ) ) /* r1 * r1 */
//             + 
//             ( matrixR( 1, 0 ) * matrixR( 1, 0 ) ) /* r2 * r2 */
//             + 
//             ( matrixR( 2, 0 ) * matrixR( 2, 0 ) ) /* r3 * r3 */
//             + 
//             ( matrixR( 3, 0 ) * matrixR( 3, 0 ) ) /* r4 * r4 */
//             - 
//             ( 
//               2 
//               *
//               ( 
//                 ( m_AMatrix( 2, 0 ) * matrixR( 0, 0 ) ) /* a13 * r1 */
//                 + 
//                 ( m_AMatrix( 2, 1 ) * matrixR( 1, 0 ) ) /* a23 * r2 */
//                 + 
//                 ( m_AMatrix( 2, 2 ) * matrixR( 2, 0 ) ) /* a33 * r3 */
//                 + 
//                 ( m_AMatrix( 2, 3 ) * matrixR( 3, 0 ) ) /* a43 * r4 */
//               )
//             )
//           );
        
        // x1
        x1 = 
          ( 
            ( A3 * A5 )
            -
            ( 2.0 * A2 * A4 )
          )
          /
          (
            ( 4.0 * A1 * A2 )
            -
            ( A3 * A3 )
          );
          
        // x2
        x2 = 
          ( 
            ( A3 * A4 )
            -
            ( 2.0 * A1 * A5 )
          )
          /
          (
            ( 4.0 * A1 * A2 )
            -
            ( A3 * A3 )
          );
          
        // checking limits
        if( 
            !
            (
              ( 0.0 <= x1 ) && ( x1 <= 1.0 ) 
              && 
              ( 0.0 <= x2 ) && ( x2 <= 1.0 )
              &&
              ( ( x1 + x2 ) <= 1 )
            )
          )
        {  // minimum outside the interest region
          if( ( x1 >= 0.0 ) && ( x2 >= 0.0 ) )
          {
            x1 = 
              (
                (-1.0 )
                *
                ( A3 + A4 - A5 - ( 2.0 * A2 ) )
              )
              /
              (
                2.0
                *
                ( A1 + A2 - A3 )
              );
            
            if( x1 > 1.0 )
            {
              x1 = 1.0;
            }
            
            if( x1 < 0.0 )
            {
              x1 = 0.0;
            }
            
            x2 = 1.0 - x1;
          }
          else if( ( x1 < 0.0 ) && ( x2 >= 0.0 ) )
          {
            x2 = (-1.0 ) * ( A5 / ( 2.0 * A2 ) );
            
            if( x2 > 1.0 )
            {
              x2 = 1.0;
            }
            
            if( x2 < 0.0 )
            {
              x2 = 0.0;
            }
            
            x1 = 0.0;
          }
          else if( ( x1 < 0.0 ) && ( x2 < 0.0 ) )
          {
            x1 = x2 = 0.0;
          }
          else // ( ( x1 >= 0.0 ) && ( x2 < 0.0 ) )
          {
            x1 = (-1.0 ) * ( A4 / ( 2.0 * A1 ) );
            
            if( x1 > 1.0 )
            {
              x1 = 1.0;
            }
            
            if( x1 < 0.0 )
            {
              x1 = 0.0;
            }
            
            x2 = 0;
          }
        }
        
        x3 = 1.0 - x1 - x2;
          
        tmpRaster->setValue(c, r, x1, 0 );
        tmpRaster->setValue(c, r, x2, 1 );
        tmpRaster->setValue(c, r, x3, 2 );
        
        m_min[0] = std::min( m_min[0], x1 );
        m_max[0] = std::max( m_max[0], x1 );
        globalMin = std::min( globalMin, x1 );
        globalMax = std::max( globalMax, x1 );

        m_min[1] = std::min( m_min[1], x2 );
        m_max[1] = std::max( m_max[1], x2 );
        globalMin = std::min( globalMin, x2 );
        globalMax = std::max( globalMax, x2 );        

        m_min[2] = std::min( m_min[2], x3 );
        m_max[2] = std::max( m_max[2], x3 );
        globalMin = std::min( globalMin, x3 );
        globalMax = std::max( globalMax, x3 );        
        
        if( computErrors )
        {
          for (unsigned b = 0 ; b < nBands; b++ )
          {
            inputRaster.getValue(c, r, inputRasterValue, inputRasterBands[b]);
            
            bandError = inputRasterValue;            
            bandError -= ( m_AMatrix( 0, b ) * x1 );
            bandError -= ( m_AMatrix( 1, b ) * x2 );
            bandError -= ( m_AMatrix( 2, b ) * x3 );
              
            outputRaster[1]->setValue(c, r, bandError, b);
            
            m_minerror[b] = std::min( m_minerror[b], bandError );
            m_maxerror[b] = std::max( m_maxerror[b], bandError );
          }
        }        
      }
      else
      {
        for (unsigned b = 0; b < nComponents; b++)
        {
          tmpRaster->setValue(c, r, tmpRasterNoDataValue, b);
        }
        
        if( computErrors )
        {
          for (unsigned b = 0 ; b < nBands; b++ )
          {
            outputRaster[1]->setValue(c, r, outErrorNoDataValue, b);
          }
        }                
      }
    }
    
    if( enableProgressInterface )
    {
      task->pulse();
    }
  }
  // Normalize to output
  
  {
    std::vector< double > outRasterDataTypeRangeMin( nComponents );
    std::vector< double > outRasterDataTypeRangeMax( nComponents );
    for (unsigned b = 0; b < nComponents; b++)
    {
      te::rp::GetDataTypeRange( 
        outputRaster[0]->getBand( b )->getProperty()->m_type,
        outRasterDataTypeRangeMin[ b ], outRasterDataTypeRangeMax[ b ] );
    }
    
    for (unsigned r = 0; r < tmpRaster->getNumberOfRows(); r++)
    {
      for (unsigned c = 0; c < tmpRaster->getNumberOfColumns(); c++)
      {
        for (unsigned b = 0; b < nComponents; b++)
        {
          tmpRaster->getValue(c, r, inputRasterValue, b);
          
          if (inputRasterValue == tmpRasterNoDataValue )
          {
            outputRaster[0]->setValue(c, r, outNoDataValue, b);
          }
          else
          {
            if (normalize)
            {
              inputRasterValue -= globalMin;
              inputRasterValue /= ( globalMax - globalMin );
              
              inputRasterValue *= ( outRasterDataTypeRangeMax[ b ] -
                outRasterDataTypeRangeMin[ b ] );
              inputRasterValue += outRasterDataTypeRangeMin[ b ];
            }
            
            outputRaster[0]->setValue(c, r, inputRasterValue, b);
          }
        }
      }
      
      if( enableProgressInterface )
      {
        task->pulse();
      }    
    }
    
    if( normalize )
    {
      m_min = outRasterDataTypeRangeMin;
      m_max = outRasterDataTypeRangeMax;
    }    
  }

  return true;
}

bool te::rp::MixtureModelLinearStrategy::executeLST(const te::rst::Raster& inputRaster, 
  const std::vector<unsigned int>& inputRasterBands,
  const std::vector<std::string>& inputSensorBands, 
  const std::map<std::string, std::vector<double> >& components,
  std::vector<te::rst::Raster*>& outputRaster, 
  const bool normalize, 
  const bool enableProgressInterface ) 
{
  TERP_INSTANCE_TRUE_OR_RETURN_FALSE( !components.empty(), "Invalid number of components" );
  TERP_INSTANCE_TRUE_OR_RETURN_FALSE( !inputRasterBands.empty(), "Invalid number of bands" );
  
  if (m_AMatrix.size1() == 0)
  {
    TERP_INSTANCE_TRUE_OR_RETURN_FALSE( generateTransformMatrix( inputRasterBands, 
      inputSensorBands, components ), "Unable to calculate the transformation matrix" );
  }
  else
  {
    TERP_INSTANCE_TRUE_OR_RETURN_FALSE( m_AMatrix.size1() == components.size(), 
      "Invalid transformation matrix" );
    TERP_INSTANCE_TRUE_OR_RETURN_FALSE( m_AMatrix.size2() == inputRasterBands.size(), 
      "Invalid transformation matrix" );    
  }
  
  // Create temporary output raster
  
  std::unique_ptr<te::rst::Raster> tmpRaster;
  const double tmpRasterNoDataValue = std::numeric_limits< double >::max();
  
  {
    te::rst::Grid* tmpgrid = new te::rst::Grid(*outputRaster[0]->getGrid());

    std::vector<te::rst::BandProperty*> tmpbandsProperties;
    
    for( unsigned int outBandidx = 0 ; outBandidx < outputRaster[0]->getNumberOfBands() ;
      ++outBandidx )
    {
      te::rst::BandProperty* bprop = new te::rst::BandProperty(
        *(outputRaster[0]->getBand(outBandidx)->getProperty()));
      bprop->m_colorInterp = te::rst::GrayIdxCInt;
      bprop->m_type = te::dt::DOUBLE_TYPE;
      bprop->m_noDataValue = tmpRasterNoDataValue;
      tmpbandsProperties.push_back(bprop);
    }

    try
    {
      tmpRaster.reset(te::rst::RasterFactory::make("EXPANSIBLE", tmpgrid, 
        tmpbandsProperties, std::map<std::string, std::string>()));
    }
    catch(...)
    {
      TERP_INSTANCE_LOG_AND_RETURN_FALSE( "Memory raster creation error" );
    }  
  }
  
  // by definition, A * X = R, we have A, so that X = inv(A' * A) * A' * R
  //           4Lx3C            3Lx1C    4Lx1C 
  //   | c1_b1 c2_b1 c3_b1 |   | x1 |   | r1 |
  //   | c1_b2 c2_b2 c3_b2 | * | x2 | = | r2 |
  //   | c1_b3 c2_b3 c3_b3 |   | x3 |   | r3 |
  //   | c1_b4 c2_b4 c3_b4 |            | r4 |
  //
  // A is the set of known reflectances for each component
  // R is the reflectances of a specific pixel
  // X is the proportion of each component

  const unsigned int nComponents = (unsigned int)components.size();
  const unsigned int nBands = (unsigned int)inputRasterBands.size();
  const double inNoDataValue = inputRaster.getBand( inputRasterBands[0] )->getProperty()->m_noDataValue;
  const double outNoDataValue = outputRaster[0]->getBand(0)->getProperty()->m_noDataValue;
  const double outErrorNoDataValue = ( outputRaster.size() > 1 ) ?
   outputRaster[1]->getBand(0)->getProperty()->m_noDataValue : 
   std::numeric_limits< double >::max();
  const bool computErrors = (outputRaster.size() > 1);
  boost::numeric::ublas::matrix<double> transposedAMatrix = 
    boost::numeric::ublas::trans( m_AMatrix );  
    
  boost::numeric::ublas::matrix<double> inverseTransAMatrix ;
  TERP_INSTANCE_TRUE_OR_RETURN_FALSE( te::common::GetPseudoInverseMatrix(
    transposedAMatrix, inverseTransAMatrix ), "Unable to calculate the inverse transformation matrix" );
  
  boost::numeric::ublas::matrix<double> matrixR = 
    boost::numeric::ublas::matrix<double>( nBands, 1);     
  boost::numeric::ublas::matrix<double> matrixX;  
  boost::numeric::ublas::matrix<double> matrixREstimated;

  std::unique_ptr< te::common::TaskProgress > task;
  if( enableProgressInterface )
  {
    task.reset( new te::common::TaskProgress( TE_TR("Linear Mixture Model"), 
      te::common::TaskProgress::UNDEFINED, 2 * inputRaster.getNumberOfRows() ) );
  }
  
  double inputRasterValue = 0;
  double bandError = 0;
  bool pixelIsValid = true;
  double globalMin = std::numeric_limits< double >::max();
  double globalMax = -1.0 * std::numeric_limits< double >::max();      
  
  m_min.clear();
  m_min.resize( nComponents, std::numeric_limits< double >::max() );
  
  m_max.clear();
  m_max.resize( nComponents, -1.0 * std::numeric_limits< double >::max() );  
  
  m_minerror.clear();
  m_minerror.resize( nBands, std::numeric_limits< double >::max() );
  
  m_maxerror.clear();
  m_maxerror.resize( nBands, -1.0 * std::numeric_limits< double >::max() );    

  for (unsigned r = 0; r < inputRaster.getNumberOfRows(); r++)
  {
    for (unsigned c = 0; c < inputRaster.getNumberOfColumns(); c++)
    {
      pixelIsValid = true;
      
      for (unsigned b = 0; b < nBands; b++)
      {
        inputRaster.getValue(c, r, inputRasterValue, inputRasterBands[b]);
        
        if (inputRasterValue == inNoDataValue )
        {
          pixelIsValid = false;
          break;
        }
        else
        {
          matrixR(b, 0) = inputRasterValue;
        }
      }
      
      if( pixelIsValid )
      {
        matrixX = boost::numeric::ublas::prod( inverseTransAMatrix, matrixR );
        
        for (unsigned b = 0 ; b < nComponents; b++ )
        {
          tmpRaster->setValue(c, r, matrixX( b, 0 ), b );
          
          m_min[b] = std::min( m_min[b], matrixX( b, 0 ) );
          m_max[b] = std::max( m_max[b], matrixX( b, 0 ) );
          globalMin = std::min( globalMin, matrixX( b, 0 ) );
          globalMax = std::max( globalMax, matrixX( b, 0 ) );
        }        

        if( computErrors )
        {
          matrixREstimated = boost::numeric::ublas::prod(transposedAMatrix, matrixX);
          
          for (unsigned b = 0 ; b < nBands; b++ )
          {
            inputRaster.getValue(c, r, inputRasterValue, inputRasterBands[b]);
              
            bandError = matrixR( b , 0 ) - matrixREstimated( b, 0 );
            
            outputRaster[1]->setValue(c, r, bandError, b);
            
            m_minerror[b] = std::min( m_minerror[b], bandError );
            m_maxerror[b] = std::max( m_maxerror[b], bandError );                      
          }
        }        
      }
      else
      {
        for (unsigned b = 0; b < nComponents; b++)
        {
          tmpRaster->setValue(c, r, tmpRasterNoDataValue, b);
                
        }
        
        if( computErrors )
        {
          for (unsigned b = 0 ; b < nBands; b++ )
          {
            outputRaster[1]->setValue(c, r, outErrorNoDataValue, b);
          }
        }                
      }
    }
    
    if( enableProgressInterface )
    {
      task->pulse();
    }
  }
  // Normalize to output
  
  {
    std::vector< double > outRasterDataTypeRangeMin( nComponents );
    std::vector< double > outRasterDataTypeRangeMax( nComponents );
    for (unsigned b = 0; b < nComponents; b++)
    {
      te::rp::GetDataTypeRange( 
        outputRaster[0]->getBand( b )->getProperty()->m_type,
        outRasterDataTypeRangeMin[ b ], outRasterDataTypeRangeMax[ b ] );
    }
    
    for (unsigned r = 0; r < tmpRaster->getNumberOfRows(); r++)
    {
      for (unsigned c = 0; c < tmpRaster->getNumberOfColumns(); c++)
      {
        for (unsigned b = 0; b < nComponents; b++)
        {
          tmpRaster->getValue(c, r, inputRasterValue, b);
          
          if (inputRasterValue == tmpRasterNoDataValue )
          {
            outputRaster[0]->setValue(c, r, outNoDataValue, b);
          }
          else
          {
            if (normalize)
            {
              inputRasterValue -= globalMin;
              inputRasterValue /= ( globalMax - globalMin );
              
              inputRasterValue *= ( outRasterDataTypeRangeMax[ b ] -
                outRasterDataTypeRangeMin[ b ] );
              inputRasterValue += outRasterDataTypeRangeMin[ b ];
            }
            
            outputRaster[0]->setValue(c, r, inputRasterValue, b);
          }
        }
      }
      
      if( enableProgressInterface )
      {
        task->pulse();
      }    
    }
    
    if( normalize )
    {
      m_min = outRasterDataTypeRangeMin;
      m_max = outRasterDataTypeRangeMax;
    }        
  }

  return true;
}

bool te::rp::MixtureModelLinearStrategy::getTransformMatrix(
  boost::numeric::ublas::matrix<double>& matrix) const
{
  matrix = m_AMatrix;
  return true;
}


bool te::rp::MixtureModelLinearStrategy::setTransformMatrix(
  boost::numeric::ublas::matrix<double>& matrix) 
{
  m_AMatrix = matrix;
  
  return true;
}

bool te::rp::MixtureModelLinearStrategy::generateTransformMatrix(
  const std::vector<unsigned int>& inputRasterBands, 
  const std::vector<std::string>& ,
  const std::map<std::string, std::vector<double> >& components ) 
{
  TERP_INSTANCE_TRUE_OR_RETURN_FALSE( !inputRasterBands.empty(), "Invalid input raster bands" );
  
  const unsigned int nComponents = (unsigned int)components.size();
  const unsigned int nBands = (unsigned int)inputRasterBands.size();
  
  switch( m_parameters.m_methodType )
  {
    case Parameters::RLS3T :
    case Parameters::LST :
    {
      m_AMatrix.resize(nComponents, nBands );
      
      std::map<std::string, std::vector<double> >::const_iterator componentsIt =
        components.begin();

      for (unsigned row = 0; row < nComponents ; ++row )
      {
        TERP_INSTANCE_TRUE_OR_RETURN_FALSE( componentsIt->second.size() == 
          inputRasterBands.size(), "Invalid components" );
        
        for (unsigned bandIdx = 0; bandIdx < nBands; bandIdx++)
        {
          m_AMatrix(row, bandIdx) = componentsIt->second[bandIdx];
        }
        
        ++componentsIt;
      }      
      
      break;
    }
    default :
    {
      TERP_INSTANCE_LOG_AND_RETURN_FALSE( "Invalid method" );
    }
  }

  return true;
}

bool te::rp::MixtureModelLinearStrategy::getMinMax(std::vector<double>& min, std::vector<double>& max) const
{
  min = m_min;
  max = m_max;
  return true;
}

bool te::rp::MixtureModelLinearStrategy::getMinMaxError(std::vector<double>& min, std::vector<double>& max) const
{
  min = m_minerror;
  max = m_maxerror;
  return true;
}

/* -------------------------------------------------------------------------- */

te::rp::MixtureModelLinearStrategyFactory::MixtureModelLinearStrategyFactory()
  : te::rp::MixtureModelStrategyFactory("linear")
{
}

te::rp::MixtureModelLinearStrategyFactory::~MixtureModelLinearStrategyFactory()
{
}

te::rp::MixtureModelStrategy* te::rp::MixtureModelLinearStrategyFactory::build()
{
  return new te::rp::MixtureModelLinearStrategy();
}


