/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/rp/MixtureModelLinearStrategy.h

  \brief Raster linear strategy for mixture model classification.
 */

#ifndef __TERRALIB_RP_INTERNAL_MIXTUREMODELLINEARSTRATEGY_H
#define __TERRALIB_RP_INTERNAL_MIXTUREMODELLINEARSTRATEGY_H

// TerraLib
#include "MixtureModelStrategy.h"
#include "MixtureModelStrategyFactory.h"
#include "MixtureModelStrategyParameters.h"
#include "Config.h"

// STL
#include <map>

// Boost
#include <boost/numeric/ublas/matrix.hpp>

namespace te
{
  namespace rp
  {
    /*!
      \class MixtureModelLinearStrategy

      \ingroup rp_class

      \note The transformation matrix (A) returned by getTransformMatrix by definition: A * X = R, we have A, so that X = inv(A' * A) * A' * R where A is the set of known reflectances for each component, R is the reflectances of a specific pixel and X is the proportion of each component.
      
      \note Reference: Modelo Linear de Mistura Espectral Conceitos Teóricos, Algoritmos e Aplicações em Estudos na Amazônia Legal Linear Spectral Mixing Model, Theoretical Concepts, Algorithms and Applications in Studies in the Legal Amazon, Yosio Edemir Shimabukuro , Andeise Cerqueira Dutra e Egidio Arai.
      
      \note Reference: Utilização de atributos derivados de proporções de classes dentro de um elemento de resolução de imagem pixel na classificação multiespectral de imagens de sensoriamento remoto, Ana Paula Dutra de Aguiar, Junho 1991.
    */
    class TERPEXPORT MixtureModelLinearStrategy : public MixtureModelStrategy
    {
      public:

        /*!
          \class Parameters

          \brief MixtureModel Parameters
        */
        class TERPEXPORT Parameters : public MixtureModelStrategyParameters
        {
          public:
            
            /*!
              \enum Linear method type.
             */    
            enum MethodType
            {
              InvalidMethodT, //!< Invalid type
              LST, //!< Least Square Method.
              RLS3T //!< Resticted Least Square Method (only 3 components, 4 bands).
            };            
            
            MethodType m_methodType;

            Parameters();

            ~Parameters();

            //overload
            const Parameters& operator=(const Parameters& params);

            //overload
            void reset() ;

            //overload
            te::common::AbstractParameters* clone() const;

            //overload
            bool serialize ( AlgorithmParametersSerializer& serializer ) const;
        };

      public:

        MixtureModelLinearStrategy();

        ~MixtureModelLinearStrategy();

        //overload
        bool initialize(StrategyParameters const* const strategyParams) ;

        //overload
        bool execute(const te::rst::Raster& inputRaster, const std::vector<unsigned int>& inputRasterBands,
                     const std::vector<std::string>& inputSensorBands, const std::map<std::string, std::vector<double> >& components,
                     std::vector<te::rst::Raster*>& outputRaster, const bool normalize, const bool enableProgressInterface) ;

        //overload
        bool getTransformMatrix(boost::numeric::ublas::matrix<double>& matrix) const;

        //overload
        bool setTransformMatrix(boost::numeric::ublas::matrix<double>& matrix) ;

        //overload
        bool generateTransformMatrix( 
          const std::vector<unsigned int>& inputRasterBands, 
          const std::vector<std::string>& inputSensorBands,
          const std::map<std::string, std::vector<double> >& components );

        //overload
        bool getMinMax(std::vector<double>&, std::vector<double>&) const;
        bool getMinMaxError(std::vector<double>&, std::vector<double>&) const;               

      protected:

        bool m_isInitialized;                                             //!< True if this instance is initialized.
        MixtureModelLinearStrategy::Parameters m_parameters;              //!< Internal execution parameters.
        boost::numeric::ublas::matrix<double> m_AMatrix;           //! A matrix.
        std::vector<double> m_min;                                    //!< Minimun value
        std::vector<double> m_max;                                    //!< Maximum value
        std::vector<double> m_minerror;                                    //!< Minimun error value
        std::vector<double> m_maxerror;                                    //!< Maximum error value
        
        bool executeRLS3T(
          const te::rst::Raster& inputRaster, 
          const std::vector<unsigned int>& inputRasterBands,
          const std::vector<std::string>& inputSensorBands, 
          const std::map<std::string, std::vector<double> >& components,
           std::vector<te::rst::Raster*>& outputRaster, const bool normalize, const bool enableProgressInterface) ;        
           
        bool executeLST(
          const te::rst::Raster& inputRaster, 
          const std::vector<unsigned int>& inputRasterBands,
          const std::vector<std::string>& inputSensorBands, 
          const std::map<std::string, std::vector<double> >& components,
           std::vector<te::rst::Raster*>& outputRaster, const bool normalize, const bool enableProgressInterface) ;                   
    
    };

    /*!
      \class MixtureModelLinearStrategyFactory

      \brief Raster linear mixture model strategy factory.

      \note Factory key: RegionGrowing
    */
    class TERPEXPORT MixtureModelLinearStrategyFactory : public MixtureModelStrategyFactory
    {
      public:

        MixtureModelLinearStrategyFactory();

        ~MixtureModelLinearStrategyFactory();

        //overload
        te::rp::MixtureModelStrategy* build();
    };

  } // end namespace rp
}   // end namespace te

#endif // __TERRALIB_RP_INTERNAL_MIXTUREMODELLINEARSTRATEGY_H
