/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/rp/MixtureModelPCAStrategy.h

  \brief PCA (Principal Component Analysis) strategy for mixture model.
 */

#ifndef __TERRALIB_RP_INTERNAL_MIXTUREMODELPCASTRATEGY_H
#define __TERRALIB_RP_INTERNAL_MIXTUREMODELPCASTRATEGY_H

// TerraLib
#include "MixtureModelStrategy.h"
#include "MixtureModelStrategyParameters.h"
#include "MixtureModelStrategyFactory.h"
#include "Config.h"

// STL
#include <map>

// Boost
#include <boost/numeric/ublas/matrix.hpp>

namespace te
{
  namespace rp
  {
    /*!
      \class MixtureModelPCAStrategy

      \ingroup rp_class
     */
    class TERPEXPORT MixtureModelPCAStrategy : public MixtureModelStrategy
    {
      public:

        /*!
          \class Parameters

          \brief MixtureModel Parameters
        */
        class TERPEXPORT Parameters : public MixtureModelStrategyParameters
        {
          public:

            Parameters();

            ~Parameters();

            //overload
            const Parameters& operator=(const Parameters& params);

            //overload
            void reset() ;

            //overload
            te::common::AbstractParameters* clone() const;

            //overload
            bool serialize ( AlgorithmParametersSerializer& serializer ) const;
        };

      public:

        MixtureModelPCAStrategy();

        ~MixtureModelPCAStrategy();

        //overload
        bool initialize(StrategyParameters const* const strategyParams) ;

        //overload
        bool execute(const te::rst::Raster& inputRaster, const std::vector<unsigned int>& inputRasterBands,
                     const std::vector<std::string>& inputSensorBands, const std::map<std::string, std::vector<double> >& components,
                     std::vector<te::rst::Raster*>& outputRaster, const bool normalize, const bool enableProgressInterface) ;

        //overload
        bool getTransformMatrix( boost::numeric::ublas::matrix<double>& matrix ) const;

        //overload
        bool setTransformMatrix(boost::numeric::ublas::matrix<double>& matrix);

        //overload
        bool generateTransformMatrix(const std::vector<unsigned int>& inputRasterBands, const std::vector<std::string>& inputSensorBands,
          const std::map<std::string, std::vector<double> >& components);

        //overload
        bool getMinMax(std::vector<double>&, std::vector<double>&) const;
        bool getMinMaxError(std::vector<double>&, std::vector<double>&) const;

      protected:

        bool m_isInitialized;                                             //!< True if this instance is initialized.
        MixtureModelPCAStrategy::Parameters m_parameters;                 //!< Internal execution parameters.
        boost::numeric::ublas::matrix<double> m_transfMatrix;         //! Transformation matrix;
        std::vector<double> m_min;                                    //!< Minimun value
        std::vector<double> m_max;                                    //!< Maximum value
        std::vector<double> m_minerror;                                    //!< Minimun error value
        std::vector<double> m_maxerror;                                    //!< Maximum error value
    };

    /*!
      \class MixtureModelPCAStrategyFactory

      \brief Raster PCA mixture model strategy factory.

      \note Factory key: RegionGrowing
    */
    class TERPEXPORT MixtureModelPCAStrategyFactory : public MixtureModelStrategyFactory
    {
      public:

        MixtureModelPCAStrategyFactory();

        ~MixtureModelPCAStrategyFactory();

        //overload
        te::rp::MixtureModelStrategy* build();
    };

  } // end namespace rp
}   // end namespace te

#endif // __TERRALIB_RP_INTERNAL_MIXTUREMODELPCASTRATEGY_H
