/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/rp/PCAFusion.cpp
  \brief Creation of skeleton imagems.
*/

#include "PCAFusion.h"
#include "Macros.h"
#include "Functions.h"
#include "AlgorithmParametersSerializer.h"

#include "../raster/BandProperty.h"
#include "../raster/RasterFactory.h"
#include "../raster/Band.h"
#include "../raster/Grid.h"
#include "../raster/Reprojection.h"
#include "../raster/SynchronizedRaster.h"
#include "../geometry/Envelope.h"
#include "../common/progress/TaskProgress.h"
#include "../memory/ExpansibleRaster.h"
#include "../memory/CachedRaster.h"

#include <cmath>
#include <limits>
  
namespace te
{
  namespace rp
  {

    PCAFusion::InputParameters::InputParameters()
    {
      reset();
    }

    PCAFusion::InputParameters::InputParameters( const InputParameters& other )
    {
      reset();
      operator=( other );
    }

    PCAFusion::InputParameters::~InputParameters()
    {
      reset();
    }

    void PCAFusion::InputParameters::reset() 
    {
      m_lowResRasterPtr = nullptr;
      m_lowResRasterBands.clear();
      m_highResRasterPtr = nullptr;
      m_highResRasterBand = 0;
      m_enableProgress = false;
      m_enableThreadedProcessing = true;
      m_interpMethod = te::rst::NearestNeighbor;
      m_autoAlignRasters = true;
      m_hrHistoFitMethod = InputParameters::MeanStdDevHistoFitMethod;
    }

    const PCAFusion::InputParameters& PCAFusion::InputParameters::operator=(
      const PCAFusion::InputParameters& params )
    {
      reset();

      m_lowResRasterPtr = params.m_lowResRasterPtr;
      m_lowResRasterBands = params.m_lowResRasterBands;
      m_highResRasterPtr = params.m_highResRasterPtr;
      m_highResRasterBand = params.m_highResRasterBand;
      m_enableProgress = params.m_enableProgress;
      m_enableThreadedProcessing = params.m_enableThreadedProcessing;
      m_interpMethod = params.m_interpMethod;
      m_autoAlignRasters = params.m_autoAlignRasters;
      m_hrHistoFitMethod = params.m_hrHistoFitMethod;
      m_directPCAMatrix = params.m_directPCAMatrix;

      return *this;
    }

    te::common::AbstractParameters* PCAFusion::InputParameters::clone() const
    {
      return new InputParameters( *this );
    }

    bool PCAFusion::InputParameters::serialize( AlgorithmParametersSerializer& serializer ) const
    {
      TERP_TRUE_OR_RETURN_FALSE( AlgorithmInputParameters::serialize( serializer ),
        "Serializarion error" );

      serializer.setAlgorithmName( "PCAFusion" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "lowResRaster",
        m_lowResRasterPtr->getName() ), "Error adding parameter lowResRaster" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addMultivaluedParameter( "lowResRasterBands",
        m_lowResRasterBands ), "Error adding parameter lowResRasterBands" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "highResRaster",
        m_highResRasterPtr->getName() ), "Error adding parameter highResRaster" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "highResRasterBand",
        m_highResRasterBand ), "Error adding parameter highResRasterBand" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "interpMethod",
        m_interpMethod ), "Error adding parameter interpMethod" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "autoAlignRasters",
        m_autoAlignRasters ), "Error adding parameter autoAlignRasters" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "hrHistoFitMethod",
        m_hrHistoFitMethod ), "Error adding parameter hrHistoFitMethod" );

      return true;
    }

    PCAFusion::OutputParameters::OutputParameters()
    {
      reset();
    }

    PCAFusion::OutputParameters::OutputParameters( const OutputParameters& other )
    {
      reset();
      operator=( other );
    }

    PCAFusion::OutputParameters::~OutputParameters()
    {
      reset();
    }

    void PCAFusion::OutputParameters::reset() 
    {
      m_rType.clear();
      m_rInfo.clear();
      m_outputRasterPtr.reset();
    }

    const PCAFusion::OutputParameters& PCAFusion::OutputParameters::operator=(
      const PCAFusion::OutputParameters& params )
    {
      reset();

      m_rType = params.m_rType;
      m_rInfo = params.m_rInfo;

      return *this;
    }

    te::common::AbstractParameters* PCAFusion::OutputParameters::clone() const
    {
      return new OutputParameters( *this );
    }

    PCAFusion::PCAFusion()
    {
      reset();
    }

    PCAFusion::~PCAFusion() = default;

    bool PCAFusion::execute( AlgorithmOutputParameters& outputParams )
    {
      if( ! m_isInitialized ) return false;

      PCAFusion::OutputParameters* outParamsPtr = dynamic_cast<
        PCAFusion::OutputParameters* >( &outputParams );
      TERP_TRUE_OR_THROW( outParamsPtr, "Invalid paramters" );

      outParamsPtr->m_directPCAMatrix = m_inputParameters.m_directPCAMatrix;
      
      // progress
      
      std::unique_ptr< te::common::TaskProgress > progressPtr;
      if( m_inputParameters.m_enableProgress )
      {
        progressPtr.reset( new te::common::TaskProgress );
        
        progressPtr->setTotalSteps( 5 );
        
        progressPtr->setMessage( "Fusing images" );
      }   
      
      // HR raster auto alignment
      
      std::unique_ptr< te::rst::Raster > alignedHighResRasterHandler;
      te::rst::Raster const * alignedHighResRasterPtr = 
        m_inputParameters.m_highResRasterPtr;
      
      if( m_inputParameters.m_autoAlignRasters )
      {
        te::gm::Envelope newHREnvelope( *m_inputParameters.m_lowResRasterPtr->getExtent() );
        newHREnvelope.transform( m_inputParameters.m_lowResRasterPtr->getSRID(),
          m_inputParameters.m_highResRasterPtr->getSRID() );
        
        double llCol = 0;
        double llRow = 0;
        m_inputParameters.m_highResRasterPtr->getGrid()->geoToGrid(
          newHREnvelope.getLowerLeftX(), newHREnvelope.getLowerLeftY(),
          llCol, llRow );
        llCol = (double)te::common::Round<double, long int>( llCol );
        llCol = std::max( 0.0, llCol );
        llCol = std::min( llCol, 
          (double)( m_inputParameters.m_highResRasterPtr->getNumberOfColumns() - 1 ) );
        llRow = (double)te::common::Round<double, long int>( llRow );
        llRow = std::max( 0.0, llRow );
        llRow = std::min( llRow, 
          (double)( m_inputParameters.m_highResRasterPtr->getNumberOfRows() - 1 ) );

        double urCol = 0;
        double urRow = 0;
        m_inputParameters.m_highResRasterPtr->getGrid()->geoToGrid(
          newHREnvelope.getUpperRightX(), newHREnvelope.getUpperRightY(),
          urCol, urRow );
        urCol = (double)te::common::Round<double, long int>( urCol );
        urCol = std::max( 0.0, urCol );
        urCol = std::min( urCol, 
          (double)( m_inputParameters.m_highResRasterPtr->getNumberOfColumns() - 1 ) );
        urRow = (double)te::common::Round<double, long int>( urRow );
        urRow = std::max( 0.0, urRow );
        urRow = std::min( urRow, 
          (double)( m_inputParameters.m_highResRasterPtr->getNumberOfRows() - 1 ) );
        
        std::vector< unsigned int > inRasterBands;
        inRasterBands.push_back( m_inputParameters.m_highResRasterBand );
        
        std::map<std::string, std::string> routInfo;
        routInfo[ "MAXMEMPERCENTUSED" ] = "20";
        
        TERP_INSTANCE_TRUE_OR_RETURN_FALSE( te::rst::Reproject(
          m_inputParameters.m_lowResRasterPtr->getSRID(),
          *m_inputParameters.m_highResRasterPtr,
          inRasterBands,
          m_inputParameters.m_interpMethod,
          1, // interpWindowRadius
          urRow,
          llCol,
          llRow - urRow + 1, // height
          urCol - llCol + 1, // width
          llRow - urRow + 1,
          urCol - llCol + 1,
          0, // max threads
          routInfo,
          "EXPANSIBLE",
          alignedHighResRasterHandler
          ), "HR raster auto-alignment error" );
        
        alignedHighResRasterPtr = alignedHighResRasterHandler.get();
      }
      
      if( m_inputParameters.m_enableProgress )
      {
        progressPtr->pulse();
        if( ! progressPtr->isActive() ) return false;
      }        
      
      const te::rst::Grid hrFinalGrid( *alignedHighResRasterPtr->getGrid() );
      
      // Loading ressampled data
      
      std::unique_ptr< te::rst::Raster > ressampledRasterPtr;
      
      {
        std::map<std::string, std::string> rinfo;
        rinfo[ "MAXMEMPERCENTUSED" ] = "37";
        
        TERP_INSTANCE_TRUE_OR_RETURN_FALSE( te::rst::rasterResample(
          *m_inputParameters.m_lowResRasterPtr,
          m_inputParameters.m_lowResRasterBands,
          m_inputParameters.m_interpMethod,
          1, // interp window radius
          0, // firstRow
          0, // firstColumn
          m_inputParameters.m_lowResRasterPtr->getNumberOfRows(),
          m_inputParameters.m_lowResRasterPtr->getNumberOfColumns(),
          alignedHighResRasterPtr->getNumberOfRows(),
          alignedHighResRasterPtr->getNumberOfColumns(),
          0, // enable thread processing
          rinfo,
          "EXPANSIBLE",
          ressampledRasterPtr ),
          "Raster resampling error" );
      } 
        
//      te::rp::Copy2DiskRaster( *ressampledRasterPtr, "ressampledRaster.tif" );
      
      if( m_inputParameters.m_enableProgress )
      {
        progressPtr->pulse();
        if( ! progressPtr->isActive() ) return false;
      }
      
      // Compute PCA space raster
      
      std::unique_ptr< te::rst::Raster > pcaRasterPtr;
      
      {
        te::rst::Grid* gridPtr = new te::rst::Grid( hrFinalGrid );
        std::vector< te::rst::BandProperty * > bandProperties;
        std::vector< unsigned int > ressampledRasterBands;
          
        for( unsigned int bandIdx = 0 ; bandIdx <
          ressampledRasterPtr->getNumberOfBands() ; ++bandIdx )
        {
          bandProperties.push_back( new te::rst::BandProperty( 
            *ressampledRasterPtr->getBand( bandIdx )->getProperty() ) );
          bandProperties[ bandIdx ]->m_type = te::dt::DOUBLE_TYPE;
          bandProperties[ bandIdx ]->m_noDataValue = std::numeric_limits< double >::max();
          
          ressampledRasterBands.push_back( bandIdx );
        } 
        
        try
        {
          pcaRasterPtr.reset( new te::mem::ExpansibleRaster( 60, gridPtr,
            bandProperties ) );    
        }
        catch( te::common::Exception& e )
        {
          TERP_INSTANCE_LOG_AND_RETURN_FALSE( e.what() );
        }    
        
        TERP_INSTANCE_TRUE_OR_RETURN_FALSE( DirectPrincipalComponents( 
            *ressampledRasterPtr, ressampledRasterBands,
            outParamsPtr->m_directPCAMatrix, *pcaRasterPtr,
            ressampledRasterBands, 0 ), "Principal components generation error" );
          
//        te::rp::Copy2DiskRaster( *pcaRasterPtr, "pcaRaster.tif" );
      }
      
      if( m_inputParameters.m_enableProgress )
      {
        progressPtr->pulse();
        if( ! progressPtr->isActive() ) return false;
      }
      
      // free anused raster
      
      ressampledRasterPtr.reset();
      
      // Swapping the first PCA band by the high resolution raster
      
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( swapBandByHighResRaster( 
        *alignedHighResRasterPtr, 
        ( m_inputParameters.m_autoAlignRasters ? 0 : m_inputParameters.m_highResRasterBand ),
        *pcaRasterPtr, 0 ),
         "Band swap error" );
      
      // free anused raster

      alignedHighResRasterHandler.reset();
         
//      te::rp::Copy2DiskRaster( *pcaRasterPtr, "swappedpcaRaster.tif" );
      
      if( m_inputParameters.m_enableProgress )
      {
        progressPtr->pulse();
        if( ! progressPtr->isActive() ) return false;
      }  
      
      // Generating the output fused raster
      
      {
        te::rst::Grid* gridPtr = new te::rst::Grid( hrFinalGrid );
        
        std::vector< te::rst::BandProperty * > bandProperties;
        std::vector< unsigned int > outputRasterBands;
          
        for( unsigned int bandIdx = 0 ; bandIdx <
          pcaRasterPtr->getNumberOfBands() ; ++bandIdx )
        {
          bandProperties.push_back( new te::rst::BandProperty( 
            *pcaRasterPtr->getBand( bandIdx )->getProperty() ) );
          bandProperties[ bandIdx ]->m_type = 
            m_inputParameters.m_lowResRasterPtr->getBand(
            m_inputParameters.m_lowResRasterBands[ bandIdx ] )->getProperty()->m_type;
          bandProperties[ bandIdx ]->m_noDataValue = 
            m_inputParameters.m_lowResRasterPtr->getBand(
            m_inputParameters.m_lowResRasterBands[ bandIdx ] )->getProperty()->m_noDataValue;
            
          outputRasterBands.push_back( bandIdx );
        } 

        AlgorithmParametersSerializer paramsSerializer;
        TERP_INSTANCE_TRUE_OR_LOG(  m_inputParameters.serialize( paramsSerializer ),
          "Parameter serialization error" );

        std::map <std::string, std::string> metadata;
        paramsSerializer.getMetaData( metadata );
        
        try
        {
          outParamsPtr->m_outputRasterPtr.reset(
            te::rst::RasterFactory::make(
              outParamsPtr->m_rType,
              gridPtr,
              bandProperties,
              outParamsPtr->m_rInfo,
              metadata,
              nullptr,
              nullptr ) );
        }
        catch(...)
        {
          TERP_INSTANCE_LOG_AND_RETURN_FALSE( 
            "Output raster creation error" );
        }
        TERP_INSTANCE_TRUE_OR_RETURN_FALSE( 
          outParamsPtr->m_outputRasterPtr.get(),
          "Output raster creation error" );
          
        TERP_INSTANCE_TRUE_OR_RETURN_FALSE( InversePrincipalComponents( *pcaRasterPtr,
          outParamsPtr->m_directPCAMatrix, *outParamsPtr->m_outputRasterPtr,
          outputRasterBands, 0 ),
          "Inverse PCA error" );
      }                  

      if( m_inputParameters.m_enableProgress )
      {
        progressPtr->pulse();
        if( ! progressPtr->isActive() ) return false;
      }  
      
      return true;
    }

    void PCAFusion::reset() 
    {
      Algorithm::reset();
      
      m_inputParameters.reset();
      m_isInitialized = false;
    }

    bool PCAFusion::initialize( const AlgorithmInputParameters& inputParams )
    {
      reset();

      PCAFusion::InputParameters const* inputParamsPtr = dynamic_cast<
        PCAFusion::InputParameters const* >( &inputParams );
      TERP_TRUE_OR_THROW( inputParamsPtr, "Invalid paramters pointer" );

      m_inputParameters = *inputParamsPtr;

      // Checking the m_lowResRasterPtr and m_lowResRasterBands

      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( m_inputParameters.m_lowResRasterPtr,
        "Invalid low Resolution Raster Pointer" )

      TERP_INSTANCE_TRUE_OR_RETURN_FALSE(
        m_inputParameters.m_lowResRasterPtr->getAccessPolicy() & te::common::RAccess, 
        "Invalid raster" );
        
      for( unsigned int lowResRasterBandsIdx = 0 ; lowResRasterBandsIdx <
        inputParamsPtr->m_lowResRasterBands.size() ; ++lowResRasterBandsIdx )
      {
        TERP_INSTANCE_TRUE_OR_RETURN_FALSE(
          inputParamsPtr->m_lowResRasterBands[ lowResRasterBandsIdx ] <
          m_inputParameters.m_lowResRasterPtr->getNumberOfBands(), 
          "Invalid raster band" );   
      }
      
      // Checking the m_highResRasterPtr and m_highResRasterBand

      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( m_inputParameters.m_highResRasterPtr,
        "Invalid high resolution Raster Pointer" )

      TERP_INSTANCE_TRUE_OR_RETURN_FALSE(
        m_inputParameters.m_highResRasterPtr->getAccessPolicy() & te::common::RAccess, 
        "Invalid raster" );
        
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE(
        m_inputParameters.m_highResRasterBand <
        m_inputParameters.m_highResRasterPtr->getNumberOfBands(), 
        "Invalid raster band" );   
      
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( 
        (
          ( m_inputParameters.m_autoAlignRasters == false )
          ||
          (
            ( m_inputParameters.m_autoAlignRasters == true )
            &&
            ( m_inputParameters.m_lowResRasterPtr->getSRID() != 0 )
            &&
            ( m_inputParameters.m_highResRasterPtr->getSRID() != 0 )
          )
        )
        ,
        "Rasters with valid SRS are required for auto-alignment" 
      );      

      if(
          ( m_inputParameters.m_directPCAMatrix.size1() != 0 )
          ||
          ( m_inputParameters.m_directPCAMatrix.size2() != 0 )
        )
      {
        TERP_INSTANCE_TRUE_OR_RETURN_FALSE(
          ( m_inputParameters.m_directPCAMatrix.size1() ==
          m_inputParameters.m_lowResRasterBands.size() ), "Invalid direct PCA matrix" );
        TERP_INSTANCE_TRUE_OR_RETURN_FALSE(
          ( m_inputParameters.m_directPCAMatrix.size1() ==
          m_inputParameters.m_directPCAMatrix.size2() ), "Invalid direct PCA matrix" );
      }
        
      m_isInitialized = true;

      return true;
    }

    bool PCAFusion::isInitialized() const
    {
      return m_isInitialized;
    }
    
    bool PCAFusion::swapBandByHighResRaster( const te::rst::Raster& highResRaster,
      const unsigned int highResRasterBand, te::rst::Raster& pcaRaster, 
      const unsigned int pcaRasterBandIdx )
    {
      assert( pcaRaster.getNumberOfColumns() == highResRaster.getNumberOfColumns() );
      assert( pcaRaster.getNumberOfRows() == highResRaster.getNumberOfRows() );
      assert( highResRasterBand < highResRaster.getNumberOfBands() );
      assert( pcaRasterBandIdx < pcaRaster.getNumberOfBands() );
      
      // defining gain and offsets
      
      double gain = 0;
      double inOffset = 0;
      double outOffset = 0;
      
      switch( m_inputParameters.m_hrHistoFitMethod )
      {
        case InputParameters::MeanStdDevHistoFitMethod :
        {
          double pcaZeroMean = 0.0;
          if( ! te::rp::GetMeanValue( pcaRaster, pcaRasterBandIdx, 0, false, 0,
            pcaZeroMean ) )
          {
            return false;
          }
          
          double pcaZeroStdDev = 0.0;
          if( ! te::rp::GetStdDevValue( pcaRaster, 
            pcaRasterBandIdx, 0, false, 0, &pcaZeroMean, pcaZeroStdDev ) )
          {
            return false;
          }
          
          double hrMean = 0.0;
          if( ! te::rp::GetMeanValue( highResRaster,  highResRasterBand, 
            0, false, 0, hrMean ) )
          {
            return false;
          }
          
          double hrStdDev = 0.0;
          if( ! te::rp::GetStdDevValue( highResRaster, highResRasterBand, 
            0, false, 0, &hrMean, hrStdDev ) )
          {
            return false;
          }
          
          gain = ( ( hrStdDev == 0.0 ) ? 0.0 : ( pcaZeroStdDev / hrStdDev ) );
          inOffset = -1.0 * hrMean;
          outOffset = pcaZeroMean;
         
          break;
        }
        case InputParameters::MinMaxHistoFitMethod :
        {
          std::map<double, unsigned> rHRHistogram;
          std::map<double, unsigned> iHRHistogram;
          
          TERP_INSTANCE_TRUE_OR_RETURN_FALSE( te::rst::getHistograms( 
            highResRaster, 
            highResRasterBand, 
            0, // rowStart
            0, // colStart
            highResRaster.getNumberOfRows() - 1,
            highResRaster.getNumberOfColumns() - 1,
            0, // histoBins
            1, // sampleStep
            0, // maxThreads
            rHRHistogram,
            iHRHistogram ),
            "High resoloution raster Histogram build error" );

          double hrMin = 0;
          double hrMax = 0;
          double hrMean = 0;
          double hrStdDev = 0;
          te::rst::getHistogramStats( rHRHistogram, hrMin, hrMax, hrMean, hrStdDev );
          
          std::map<double, unsigned> rPcaZeroHistogram;
          std::map<double, unsigned> iPcaZeroHistogram;
          
          TERP_INSTANCE_TRUE_OR_RETURN_FALSE( te::rst::getHistograms( 
            pcaRaster, 
            pcaRasterBandIdx, 
            0, // rowStart
            0, // colStart
            pcaRaster.getNumberOfRows() - 1,
            pcaRaster.getNumberOfColumns() - 1,
            0, // histoBins
            1, // sampleStep
            0, // maxThreads
            rPcaZeroHistogram,
            iPcaZeroHistogram ),
            "PCA raster Histogram build error" );

          double pcaZeroMin = 0;
          double pcaZeroMax = 0;
          double pcaZeroMean = 0;
          double pcaZeroStdDev = 0;
          te::rst::getHistogramStats( rPcaZeroHistogram, pcaZeroMin, pcaZeroMax, 
            pcaZeroMean, pcaZeroStdDev );
          
          gain = ( ( (hrMax - hrMin) == 0.0 ) ? 0.0 : ( (pcaZeroMax - pcaZeroMin) 
            / ( hrMax - hrMin) ) );
          inOffset = -1.0 * hrMin;
          outOffset = pcaZeroMin;
          
          break;
        }
        case InputParameters::NoHistoFitMethod :
        {
          gain = 1.0;
          inOffset = 0;
          outOffset = 0;
          
          break;
        }
        default :
        {
          TERP_INSTANCE_LOG_AND_RETURN_FALSE( "Invalid High resolution image histogram fitting method" );
        }
      }
      
      // output blocks to process
      // std::get< 0 > == processed = true / not processed = false
      // std::get< 1 > == output band index
      // std::get< 2 > == output block X
      // std::get< 3 > == output block Y       
     
      std::vector< std::tuple< bool, unsigned int,  unsigned int > >
        rasterBlocksStatus;
      
      {
        std::tuple< bool, unsigned int,  unsigned int > tuple;
        unsigned int blkY = 0;
        unsigned int blkX = 0;
        
        for( blkY = 0 ; blkY <  (unsigned int)pcaRaster.getBand(
          0 )->getProperty()->m_nblocksy ; ++blkY )
        {      
          for( blkX = 0 ; blkX < (unsigned int)pcaRaster.getBand(
            0 )->getProperty()->m_nblocksx ; ++blkX )
          {
            std::get< 0 >( tuple ) = false;
            std::get< 1 >( tuple ) = blkX;
            std::get< 2 >( tuple ) = blkY;
            
            rasterBlocksStatus.push_back( tuple );
          }
        }
      }            
      
      // process raster blocks
      
      unsigned int threadsNumber = m_inputParameters.m_enableThreadedProcessing 
        ? te::common::GetPhysProcNumber() : 1;
      
      te::rst::RasterSynchronizer inputRasterSync( 
        *(te::rst::Raster*)&highResRaster, 
        te::common::RAccess );
      te::rst::RasterSynchronizer outputRasterSync( pcaRaster,
        te::common::WAccess );  
      std::mutex mutex;      
      
      SwapBandByHighResRasterThreadParameters params;
      params.m_inputRasterSyncPtr = &inputRasterSync;
      params.m_outputRasterSyncPtr = &outputRasterSync;
      params.m_inputRasterBandIdx = highResRasterBand;
      params.m_outputRasterBandIdx = pcaRasterBandIdx;
      params.m_rasterBlocksStatusPtr = &rasterBlocksStatus;
      params.m_returnStatus = true;
      params.m_mutexPtr = &mutex;
      params.m_inOffset = inOffset;
      params.m_outOffset = outOffset;
      params.m_gain = gain;
      
      if( threadsNumber == 1 )
      {
        swapBandByHighResRasterThreadEntry( &params );
      }
      else
      {
        std::vector< std::shared_ptr< std::thread > > threads;

        for( unsigned int threadIdx = 0 ; threadIdx < threadsNumber ;
          ++threadIdx )
        {
          std::shared_ptr< std::thread > threadPtr( new std::thread(
            &swapBandByHighResRasterThreadEntry, &params ) );

          threads.push_back( threadPtr );
        };

        for( unsigned int threadIdx = 0 ; threadIdx < threadsNumber ;
          ++threadIdx )
        {
          threads[ threadIdx ]->join();
        };
      }  
      
      return params.m_returnStatus;
    }
    
    void te::rp::PCAFusion::swapBandByHighResRasterThreadEntry( 
      SwapBandByHighResRasterThreadParameters* paramsPtr )
    {
      std::unique_lock< std::mutex > locker( *paramsPtr->m_mutexPtr );

      std::vector< std::tuple< bool, unsigned int,  unsigned int > >&
        rasterBlocksStatus = *paramsPtr->m_rasterBlocksStatusPtr;
      const std::size_t rasterBlocksStatusSize = rasterBlocksStatus.size();
      const unsigned int inputRasterBandIdx = paramsPtr->m_inputRasterBandIdx;
      const unsigned int outputRasterBandIdx = paramsPtr->m_outputRasterBandIdx;
      const double inOffset = paramsPtr->m_inOffset;
      const double outOffset = paramsPtr->m_outOffset;
      const double gain = paramsPtr->m_gain;      

      locker.unlock();

      const te::rst::SynchronizedRaster inputRaster( 9,
        *paramsPtr->m_inputRasterSyncPtr );      
      te::rst::SynchronizedRaster outputRaster( 1,
        *paramsPtr->m_outputRasterSyncPtr );
      
      const unsigned int outputNRows = outputRaster.getNumberOfRows();
      const unsigned int outputNCols = outputRaster.getNumberOfColumns();
      const te::rst::Band& inBand = *inputRaster.getBand( inputRasterBandIdx );
      const double inNoDataValue = inBand.getProperty()->m_noDataValue;
      
      te::rst::Band& outBand = *outputRaster.getBand( outputRasterBandIdx );
      const double outNoDataValue = outBand.getProperty()->m_noDataValue;
      const te::rst::BandProperty& outBandProp = *outBand.getProperty();
      unsigned int outputBlkX = 0;
      unsigned int outputBlkY = 0;
      unsigned int outputBlkStartRow = 0;
      unsigned int outputBlkStartCol = 0;
      unsigned int outputBlkRowBound = 0;
      unsigned int outputBlkColBound = 0;
      unsigned int outputRow = 0;
      unsigned int outputCol = 0;
      double value = 0;
      
      for( std::size_t rasterBlocksStatusIdx = 0 ; rasterBlocksStatusIdx <
        rasterBlocksStatusSize ; ++rasterBlocksStatusIdx )
      {
        locker.lock();

        // looking for the next unprocessed block

        if(
            std::get< 0 >( rasterBlocksStatus[ rasterBlocksStatusIdx ] )
            ==
            false
          )
        {
          std::get< 0 >( rasterBlocksStatus[ rasterBlocksStatusIdx ] ) = true;

          outputBlkX = std::get< 1 >( rasterBlocksStatus[ rasterBlocksStatusIdx ] );
          outputBlkY = std::get< 2 >( rasterBlocksStatus[ rasterBlocksStatusIdx ] );

          locker.unlock();

          // definng the current block limits

          outputBlkStartRow = outBandProp.m_blkh * outputBlkY;
          outputBlkStartRow = std::min( outputBlkStartRow, outputNRows );
          
          outputBlkStartCol = outBandProp.m_blkw * outputBlkX;
          outputBlkStartCol = std::min( outputBlkStartCol, outputNCols );

          outputBlkRowBound = (outBandProp.m_blkh * outputBlkY) + 
            outBandProp.m_blkh;
          outputBlkRowBound = std::min( outputBlkRowBound, outputNRows );
          
          outputBlkColBound = (outBandProp.m_blkw * outputBlkX) + 
            outBandProp.m_blkw;
          outputBlkColBound = std::min( outputBlkColBound, outputNCols );

          for( outputRow = outputBlkStartRow ; outputRow < outputBlkRowBound ; 
            ++outputRow )
          {
            for( outputCol = outputBlkStartCol ; outputCol < outputBlkColBound ; 
              ++outputCol )
            {
              inBand.getValue( outputCol, outputRow, value );
            
              if( value == inNoDataValue )
              {
                outBand.setValue( outputCol, outputRow, outNoDataValue );
              }
              else
              {
                value += inOffset;
                value *= gain;
                value += outOffset;
                
                outBand.setValue( outputCol, outputRow, value );
              }
            }
          }
        }
        else
        {
          locker.unlock();
        }
      }
    }
    
  }
}   // end namespace te

