/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

This file is part of the TerraLib - a Framework for building GIS enabled applications.

TerraLib is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

TerraLib is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with TerraLib. See COPYING. If not, write to
TerraLib Team at <terralib-team@terralib.org>.
*/

/*!
\file terralib/rp/PostClassification.h
\brief Raster post classification.
*/

#ifndef __TERRALIB_RP_INTERNAL_POSTCLASSIFICATION_H
#define __TERRALIB_RP_INTERNAL_POSTCLASSIFICATION_H

#include "Algorithm.h"
#include "../raster/Raster.h"

namespace te
{
  namespace rp
  {
    /*!
    \class PostClassification
    \brief Raster post classification.
    \ingroup rp_enh
    */
    class TERPEXPORT PostClassification : public Algorithm
    {
    public:

      /*!
      \class InputParameters
      \brief Filter input parameters
      */
      class TERPEXPORT InputParameters : public AlgorithmInputParameters
      {
      public:

        te::rst::Raster const* m_inRasterPtr; //!< Input raster.

        bool m_enableRasterCache;  // Enable/disable input raster cache (default:true).

        unsigned int m_weight; //!< The weight value defines the number of times the frequency of the central point has to be considered.

        unsigned int m_threshold; //!< The threshold value is the frequency value above which the central point is modified.

        InputParameters();

        InputParameters(const InputParameters&);

        ~InputParameters();

        //overload
        void reset() _NOEXCEPT_OP(false);

        //overload
        const  InputParameters& operator=(const InputParameters& params);

        //overload
        AbstractParameters* clone() const;

        //overload
        bool serialize ( AlgorithmParametersSerializer& serializer ) const;
      };

      /*!
      \class OutputParameters
      \brief PostClassification output parameters
      */
      class TERPEXPORT OutputParameters : public AlgorithmOutputParameters
      {
      public:

        std::string m_rType; //!< Output raster data source type (as described in te::raster::RasterFactory ).

        std::map< std::string, std::string > m_rInfo; //!< The necessary information to create the raster (as described in te::raster::RasterFactory).

        std::unique_ptr< te::rst::Raster > m_outputRasterPtr; //!< A pointer the ge generated output raster (label image).

        OutputParameters();

        OutputParameters(const OutputParameters&);

        ~OutputParameters();

        //overload
        void reset() _NOEXCEPT_OP(false);

        //overload
        const  OutputParameters& operator=(const OutputParameters& params);

        //overload
        AbstractParameters* clone() const;
      };

      PostClassification();

      ~PostClassification();

      //overload
      bool execute(AlgorithmOutputParameters& outputParams) _NOEXCEPT_OP(false);

      //overload
      void reset() _NOEXCEPT_OP(false);

      //overload
      bool initialize(const AlgorithmInputParameters& inputParams) _NOEXCEPT_OP(false);

      //overload
      bool isInitialized() const;

    protected:

      bool m_isInitialized; //!< Is this instance already initialized?

      PostClassification::InputParameters m_inputParameters; //!< Input parameters.

      /*!
      \brief Apply the post classification.
      \param srcRaster Source raster.
      \param srcBandIdx Source raster band index.
      \param dstRaster Destination raster.
      \param dstBandIdx Destination raster band index.
      \param weight Weight value.
      \param threshold Threshold value.
      \param useProgress if true, the progress interface must be used.
      */
      bool runPostClassification(const te::rst::Raster& srcRaster, te::rst::Raster& dstRaster,
        unsigned int weight, unsigned int threshold);
    };

  } // end namespace rp
}   // end namespace te

#endif

