/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

#define _USE_MATH_DEFINES

#include "Restoration.h"
#include "Macros.h"
#include "Functions.h"
#include "AlgorithmParametersSerializer.h"
#include "../raster/RasterFactory.h"
#include "../raster/BandProperty.h"
#include "../raster/Grid.h"
#include "../raster/Band.h"
#include "../geometry/Envelope.h"
#include "../memory/CachedRaster.h"
#include "../common/progress/TaskProgress.h"

#include <cmath>
#include <algorithm>

namespace te
{
  namespace rp
  {
    // -------------------------------------------------------------------------------
    
    Restoration::InputParameters::InputParameters()
    {
      reset();
    }

    Restoration::InputParameters::InputParameters( const InputParameters& other )
    {
      reset();
      operator=( other );
    }

    Restoration::InputParameters::~InputParameters()
    {
      reset();
    }

    void Restoration::InputParameters::reset() _NOEXCEPT_OP(false)
    {
      m_sensorType = InputParameters::INVALID_RESTORATION_TYPE;
      m_sensorBandDesignations.clear();
      m_inRasterPtr = nullptr;
      m_inRasterBands.clear();
      m_samplingFactor = INVALID_SAMPLING_FACTOR_TYPE;
      m_enableProgress = false;
      m_enableRasterCache = true;
    }

    const Restoration::InputParameters& Restoration::InputParameters::operator=(
      const Restoration::InputParameters& params )
    {
      reset();

      m_sensorType = params.m_sensorType;
      m_sensorBandDesignations = params.m_sensorBandDesignations;
      m_inRasterPtr = params.m_inRasterPtr;
      m_inRasterBands = params.m_inRasterBands;
      m_samplingFactor = params.m_samplingFactor;
      m_enableProgress = params.m_enableProgress;
      m_enableRasterCache = params.m_enableRasterCache;

      return *this;
    }

    te::common::AbstractParameters* Restoration::InputParameters::clone() const
    {
      return new InputParameters( *this );
    }
    
    bool Restoration::InputParameters::serialize( AlgorithmParametersSerializer& serializer ) const
    {
      TERP_TRUE_OR_RETURN_FALSE( AlgorithmInputParameters::serialize( serializer ),
        "Serializarion error" );

      serializer.setAlgorithmName( "Restoration" );

      std::map< SensorType, std::string > sensorTypes;
      getSupportedSensorTypes( sensorTypes );
      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "sensorType",
        sensorTypes[ m_sensorType ] ), "Error adding parameter sensorType" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addMultivaluedParameter( "sensorBandDesignations",
        m_sensorBandDesignations ), "Error adding parameter sensorBandDesignations" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "inRaster",
        m_inRasterPtr->getName() ), "Error adding parameter inRaster" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addMultivaluedParameter( "inRasterBands",
        m_inRasterBands ), "Error adding parameter inRasterBands" );

      std::map< SamplingFactorType, std::string > samplingFactors;
      getSupportedSamplingFactorsTypes( samplingFactors );
      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "samplingFactor",
        samplingFactors[ m_samplingFactor ] ), "Error adding parameter samplingFactor" );

      return true;
    }

    void Restoration::InputParameters::getSupportedSensorTypes( 
      std::map< SensorType, std::string >& infos )
    {
      infos.clear();
      
      infos[ SPOTP ] = "SPOTP";
      infos[ SPOTX ] = "SPOTX";
      infos[ TM ] = "TM";
      infos[ ETMX ] = "ETMX";
      infos[ ETMP ] = "ETMP";
      infos[ LANDSAT_8 ] = "LANDSAT_8";
      infos[ CBERS_CCD ] = "CBERS_CCD";
      infos[ CBERS_CCRX ] = "CBERS_CCRX";
      infos[ CBERS2_CCD ] = "CBERS2_CCD";
      infos[ CBERS2B_CCD ] = "CBERS2B_CCD";
      infos[ CBERS2B_HRC ] = "CBERS2B_HRC";
      infos[ CBERS_4_MUX ] = "CBERS_4_MUX";
      infos[ CBERS_4_PAN10 ] = "CBERS_4_PAN10";
      infos[ CBERS_4_PAN5 ] = "CBERS_4_PAN5";
      infos[ CBERS_4_WFI ] = "CBERS_4_WFI";
      infos[ RAPIDEYE ] = "RAPIDEYE";
      infos[ QUICKBIRD ] = "QUICKBIRD";
      infos[ WORLDVIEW2 ] = "WORLDVIEW2";
      infos[ MODIS_250 ] = "MODIS_250";
      infos[ MODIS_500 ] = "MODIS_500";
    }
    
    void Restoration::InputParameters::getSupportedSamplingFactorsTypes( 
      std::map< SamplingFactorType, std::string >& infos )
    {
      infos.clear();
      
      infos[ SAMPLING_FACTOR_1_BY_2 ] = "SAMPLING_FACTOR_1_BY_2";
      infos[ SAMPLING_FACTOR_1_BY_3 ] = "SAMPLING_FACTOR_1_BY_3";
      infos[ SAMPLING_FACTOR_1_BY_4 ] = "SAMPLING_FACTOR_1_BY_4";
      infos[ SAMPLING_FACTOR_1_BY_5 ] = "SAMPLING_FACTOR_1_BY_5";
      infos[ SAMPLING_FACTOR_1_BY_6 ] = "SAMPLING_FACTOR_1_BY_6";
      infos[ SAMPLING_FACTOR_2_BY_3 ] = "SAMPLING_FACTOR_2_BY_3";
      infos[ SAMPLING_FACTOR_3_BY_4 ] = "SAMPLING_FACTOR_3_BY_4";
    }    
    
    // -------------------------------------------------------------------------------

    Restoration::OutputParameters::OutputParameters()
    {
      reset();
    }

    Restoration::OutputParameters::OutputParameters( const OutputParameters& other )
    {
      reset();
      operator=( other );
    }

    Restoration::OutputParameters::~OutputParameters()
    {
      reset();
    }

    void Restoration::OutputParameters::reset() _NOEXCEPT_OP(false)
    {
      m_rType.clear();
      m_rInfo.clear();
      m_outputRasterPtr.reset();
    }

    const Restoration::OutputParameters& Restoration::OutputParameters::operator=(
      const Restoration::OutputParameters& params )
    {
      reset();

      m_rType = params.m_rType;
      m_rInfo = params.m_rInfo;

      return *this;
    }

    te::common::AbstractParameters* Restoration::OutputParameters::clone() const
    {
      return new OutputParameters( *this );
    }
    
    // -------------------------------------------------------------------------------

    Restoration::Restoration()
    {
      reset();
    }

    Restoration::~Restoration()
    {
    }

    bool Restoration::execute( AlgorithmOutputParameters& outputParams ) _NOEXCEPT_OP(false)
    {
      if( ! m_isInitialized ) return false;

      Restoration::OutputParameters* outParamsPtr = dynamic_cast<
        Restoration::OutputParameters* >( &outputParams );
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( outParamsPtr, "Invalid paramters" );
      
      // Input raster 
      
      te::rst::Raster const * inputRasterPtr = m_inputParameters.m_inRasterPtr;
      
      std::unique_ptr< te::mem::CachedRaster > inputCachedRasterPtr;
      
      if( m_inputParameters.m_enableRasterCache )
      {
        inputCachedRasterPtr.reset( new te::mem::CachedRaster(
          *m_inputParameters.m_inRasterPtr, 50, 0 ) );
        
        inputRasterPtr = inputCachedRasterPtr.get();
      }      

      // Initializing the output raster

      {
        unsigned int sRUpperValue = 0;
        unsigned int sRLowerValue = 0;
        getDecomposedSamplingFactor( m_inputParameters.m_samplingFactor, sRUpperValue, 
          sRLowerValue );
        
        std::unique_ptr< te::rst::Grid > newGrid( new te::rst::Grid(
          ( inputRasterPtr->getGrid()->getNumberOfColumns() * sRLowerValue ) /
            sRUpperValue,
          ( inputRasterPtr->getGrid()->getNumberOfRows() * sRLowerValue ) /
            sRUpperValue,
          new te::gm::Envelope( *inputRasterPtr->getGrid()->getExtent() ),
          inputRasterPtr->getGrid()->getSRID() ) );        
        
        std::vector< te::rst::BandProperty* > outRasterBandsProperties;

        for( unsigned int inRasterBandsIdx = 0 ; inRasterBandsIdx <
          m_inputParameters.m_inRasterBands.size() ; ++inRasterBandsIdx )
        {
          const unsigned int& inRasterBandIndex =
            m_inputParameters.m_inRasterBands[ inRasterBandsIdx ];

          assert( inRasterBandIndex <
            inputRasterPtr->getNumberOfBands() );

          outRasterBandsProperties.push_back( new te::rst::BandProperty(
            *( inputRasterPtr->getBand( inRasterBandIndex )->getProperty() ) ) );
          outRasterBandsProperties[ inRasterBandsIdx ]->m_blkw =
            newGrid->getNumberOfColumns();
          outRasterBandsProperties[ inRasterBandsIdx ]->m_blkh = 1;
          outRasterBandsProperties[ inRasterBandsIdx ]->m_nblocksx = 1;
          outRasterBandsProperties[ inRasterBandsIdx ]->m_nblocksy =
            newGrid->getNumberOfRows();          
        }

        AlgorithmParametersSerializer paramsSerializer;
        TERP_INSTANCE_TRUE_OR_LOG(  m_inputParameters.serialize( paramsSerializer ),
          "Parameter serialization error" );

        std::map <std::string, std::string> metadata;
        paramsSerializer.getMetaData( metadata );

        try
        {
          outParamsPtr->m_outputRasterPtr.reset(
            te::rst::RasterFactory::make(
              outParamsPtr->m_rType,
              newGrid.release(),
              outRasterBandsProperties,
              outParamsPtr->m_rInfo,
              metadata,
              nullptr,
              nullptr ) );
        }
        catch(...)
        {
          outParamsPtr->m_outputRasterPtr.reset();
        }
        
        TERP_INSTANCE_TRUE_OR_RETURN_FALSE( outParamsPtr->m_outputRasterPtr.get(),
          "Output raster creation error" );
      }
      
      // Filter coefficients
      
      std::vector< std::vector< double > > bandsXCoefs( 
        m_inputParameters.m_inRasterBands.size() );
      std::vector< std::vector< double > > bandsYCoefs(
        m_inputParameters.m_inRasterBands.size() );

      for( std::size_t inRasterBandsIdx = 0 ; inRasterBandsIdx <
        m_inputParameters.m_inRasterBands.size() ; ++inRasterBandsIdx )
      {
        TERP_INSTANCE_TRUE_OR_RETURN_FALSE( getFilterCoefs( 
          m_inputParameters.m_sensorBandDesignations[ inRasterBandsIdx ],
          bandsXCoefs[ inRasterBandsIdx ],
          bandsYCoefs[ inRasterBandsIdx ] ),
          "Cannot get filter coefficients" );        
      }
      
      // Calc output raster values
      
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( applyFilterCoefs( 
        bandsXCoefs,
        bandsYCoefs,
        *inputRasterPtr,
        *outParamsPtr->m_outputRasterPtr ),
        "Error applying filter coefficients" );

      return true;
    }

    void Restoration::reset() _NOEXCEPT_OP(false)
    {
      Algorithm::reset();

      m_isInitialized = false;
      m_inputParameters.reset();
    }

    bool Restoration::initialize( const AlgorithmInputParameters& inputParams ) _NOEXCEPT_OP(false)
    {
      reset();

      Restoration::InputParameters const* inputParamsPtr = dynamic_cast<
        Restoration::InputParameters const* >( &inputParams );
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( inputParamsPtr, 
        TE_TR( "Invalid parameters" ) );

      m_inputParameters = *inputParamsPtr;

      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( m_inputParameters.m_sensorType 
        !=
        InputParameters::INVALID_RESTORATION_TYPE,
        TE_TR( "Invalid restoration type" ) );
      
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE(
        ( m_inputParameters.m_sensorBandDesignations.size() ==
        m_inputParameters.m_inRasterBands.size() ),
        TE_TR( "Invalid sensor band designations" ) );

      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( m_inputParameters.m_inRasterPtr,
        TE_TR( "Invalid raster pointer" ) );
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE(
        m_inputParameters.m_inRasterPtr->getAccessPolicy() & te::common::RAccess,
        TE_TR( "Invalid raster" ) );

      TERP_INSTANCE_TRUE_OR_RETURN_FALSE( 
        m_inputParameters.m_inRasterBands.size() > 0,
        TE_TR( "Invalid raster bands number" ) );
      for( unsigned int inRasterBandsIdx = 0 ; inRasterBandsIdx <
        m_inputParameters.m_inRasterBands.size() ; ++inRasterBandsIdx )
      {
        TERP_INSTANCE_TRUE_OR_RETURN_FALSE(
          m_inputParameters.m_inRasterBands[ inRasterBandsIdx ] <
          m_inputParameters.m_inRasterPtr->getNumberOfBands(),
          TE_TR( "Invalid raster bands" ) );
      }
      
      TERP_INSTANCE_TRUE_OR_RETURN_FALSE(
        ( m_inputParameters.m_samplingFactor != 
        InputParameters::INVALID_SAMPLING_FACTOR_TYPE ),
        TE_TR( "Invalid sampling factor" ) );

      m_isInitialized = true;

      return true;
    }

    bool Restoration::isInitialized() const
    {
      return m_isInitialized;
    }
    
    bool Restoration::getFilterCoefs( const unsigned int Band, 
      std::vector< double >& XCoefs, std::vector< double >& YCoefs ) const
    {
      unsigned int FatM = 0;
      unsigned int FatL = 0;
      getDecomposedSamplingFactor( m_inputParameters.m_samplingFactor, FatM, FatL );
      
      const unsigned int TX = getFilterSize();
      
      int		i,j,ind;
      int		Nx;					// number of the filter coefficients 
      double	fam;				// size of the pixel
      double	PIXEL = 0.;			// size of the pixel
        double	x,f,kx,ky,fwx,fwy,
            coi,cop,fx,integ,
          EifovX,EifovY;

      ky=kx=fwx=fwy = 0.;
      
      switch( m_inputParameters.m_sensorType ){
        case InputParameters::SPOTP:
             PIXEL = 10.;
             kx = 5.237;
             ky = 7.33;
    //			   fwx = 0.36; fwy = 0.3;
             fwx = 0.41; fwy = 0.34;	
             break;
        case InputParameters::SPOTX:
             PIXEL = 20.;
             switch( Band ){
            case 1:
              kx = 3.4;
                ky = 3.8;
    //			   		fwx = 0.45; fwy = 0.42;
                fwx = 0.49; fwy = 0.46;	
              break;
              case 2:
              kx = 3.9;
                ky = 4.315;
    //			   		fwx = 0.42; fwy = 0.4;
                fwx = 0.46; fwy = 0.44;	
              break;
              case 3:
                kx = 5.4;
                ky = 4.6;
    //			   		fwx = 0.35; fwy = 0.39;
                fwx = 0.39; fwy = 0.43;	
              break;
            default:					
              return false;
            }
          break; 
        case InputParameters::TM: 
          PIXEL = 30.;
          if( Band > 0 && Band < 5){
             kx = 5.34;
             ky = 6.36;
             fwx = 0.36;
             fwy = 0.33;
          }
          else{ 
            if( Band == 5 || Band == 7){
              kx = 5.06;
              ky = 6.36;
              fwx = 0.37;
              fwy = 0.33;
                }else{					
              return false;
            }
          }		

          break;

          case InputParameters::ETMX: 
              PIXEL = 30.0;

              if (Band == 1) {
              kx = 3.851073071; fwx = 0.42425;
              /*ky = (float)2.776474426;*/  //AND 03/04
              ky = 3.000000000; fwy = 0.49965;
              }

              if (Band == 2) {
              kx = 4.027377459; fwx = 0.41486;
              ky = 3.376271297; fwy = 0.45310;
              }

              if (Band == 3) {
              kx = 4.309515228; fwx = 0.40105;
              ky = 3.441278051; fwy = 0.44880;
              }

              if (Band == 4) {
              kx = 4.725283972; fwx = 0.38300;
              ky = 3.719301316; fwy = 0.43170;
              }

              if (Band == 5) {
              kx = 3.679984607; fwx = 0.43400;
              ky = 3.718439918; fwy = 0.43175;
              }

              if (Band == 7) {
              kx = 3.437447387; fwx = 0.44905;
              ky = 2.986014207; fwy = 0.48180;
              }
          break;

        case InputParameters::ETMP: 
            PIXEL = 15.0;
            kx = 7.062970; fwx = 0.31327;
            ky = 9.686745; fwy = 0.26750;      
            break;
        case InputParameters::LANDSAT_8:
            PIXEL = 30.0;

            if (Band == 2) {
              kx = 0.405; 
              ky = 0.378; 
            }

            if (Band == 3) {
              kx = 0.387; 
              ky = 0.352;
            }

            if (Band == 4) {
              kx = 0.511; 
              ky = 0.442;
            }

            if (Band == 5) {
              kx = 0.395; 
              ky = 0.385; 
            }

            if (Band == 6) {
              kx = 0.366;
              ky = 0.357; 
            }

            if (Band == 7) {
              kx = 0.431; 
              ky = 0.395;
            }

             fwx = sqrt(log((double)2) / kx);
             fwy = sqrt(log((double)2) / ky);

          break;
        //AND END 04/16
        case InputParameters::CBERS_CCRX: 
            PIXEL = 20.;
            kx = 8.;
            ky = 8.;
            fwx = 0.29; fwy = 0.29;	
            break;

        case InputParameters::CBERS_CCD:
             PIXEL = 20.;
             EifovX = 30.;
             EifovY = 30.;
               fam = 1./PIXEL;

             //kx = (float) 4.0 * ln(2) * sqr(fam) * sqr(EifovX);
             //ky = (float) 4.0 * ln(2) * sqr(fam) * sqr(EifovY);
             kx = 4.0 * log((double)2) * pow(fam,2) * pow(EifovX,2); 
             ky = 4.0 * log((double)2) * pow(fam,2) * pow(EifovY,2);
             fwx = sqrt(log((double)2) / kx);
             fwy = sqrt(log((double)2) / ky);
                     
             break;
        case InputParameters::CBERS2_CCD:
             PIXEL = 20.;
             //EifovX = 30.; //AND 12/10
             //EifovY = 30.; //AND 12/10
               fam = 1./PIXEL;

             if (Band == 2)
             {
               EifovX = 59.;
               EifovY = 35.;
             }
             else{
               if (Band == 3)
               {
                 EifovX = 58.;
                 EifovY = 43.;
               }
               else
               {
                 if (Band == 4)
                 {
                   EifovX = 63.;
                   EifovY = 37.;
                 }
                 else
                 {
                   if (Band == 1)
                   {
                    EifovX = 61.;
                    EifovY = 36.;
                   }
                 }
               }
             }
             
             //kx = (float) 4.0 * ln(2) * sqr(fam) * sqr(EifovX);
             //ky = (float) 4.0 * ln(2) * sqr(fam) * sqr(EifovY);
             kx = 4.0 * log((double)2) * pow(fam,2) * pow(EifovX,2); 
             ky = 4.0 * log((double)2) * pow(fam,2) * pow(EifovY,2);
             fwx = sqrt(log((double)2) / kx);
             fwy = sqrt(log((double)2) / ky);
                     
             break;

        case InputParameters::CBERS2B_CCD:
             PIXEL = 20.;
               fam = 1./PIXEL;
             
             if ((Band == 2)||(Band==1))
             {
               EifovX = 40.;
               EifovY = 35.;
             }
             else{
               if (Band == 3)
               {
                 EifovX = 40.;
                 EifovY = 43.;
               }
               else
                 if (Band == 4)
                 {
                   EifovX = 40.;
                   EifovY = 37.;
                 }
             }
             
             //EifovX = 30.; //AND 12/10
            // EifovY = 30.; //AND 12/10

             //kx = (float) 4.0 * ln(2) * sqr(fam) * sqr(EifovX);
             //ky = (float) 4.0 * ln(2) * sqr(fam) * sqr(EifovY);
             kx = 4.0 * log((double)2) * pow(fam,2) * pow(EifovX,2);
             ky = 4.0 * log((double)2) * pow(fam,2) * pow(EifovY,2);
             fwx = sqrt(log((double)2) / kx);
             fwy = sqrt(log((double)2) / ky);

             break;

        case InputParameters::CBERS2B_HRC:
             PIXEL = 2.5;
             EifovX = 3.75;//10.3;  //AND 12/10
             EifovY = 3.75; //9.64; //AND 12/10
               fam = 1./PIXEL;
                    
             //kx = (float) 4.0 * ln(2) * sqr(fam) * sqr(EifovX);
             //ky = (float) 4.0 * ln(2) * sqr(fam) * sqr(EifovY);
             kx = 4.0 * log((double)2) * pow(fam,2) * pow(EifovX,2);
             ky = 4.0 * log((double)2) * pow(fam,2) * pow(EifovY,2);
             fwx = sqrt(log((double)2) / kx);
             fwy = sqrt(log((double)2) / ky);

             break;
        case InputParameters::CBERS_4_MUX: 
             PIXEL = 20.;
               fam = 1./PIXEL;

             if (Band == 2)
             {
               kx = 0.296;
               ky = 0.286; //Leila (0.266 - Giovani)
             }
             else{
               if (Band == 3)
               {
                 kx = 0.311; //Leila (0.357 - Giovani)
                 ky = 0.292;  //Leila (0.39 - Giovani)
               }
               else
               {
                 if (Band == 4)
                 {
                   kx = 0.33;
                   ky = 0.319;
                 }
                 else
                 {
                   if (Band == 5)
                   {
                    kx = 0.306; //Leila (0.406 - Giovani)
                    ky = 0.297; //Leila (0.318 - Giovani)
                   }
                 }
               }
             }
             
             fwx = sqrt(log((double)2) / kx);
             fwy = sqrt(log((double)2) / ky);
                     
             break;
        case InputParameters::CBERS_4_PAN10: 
             PIXEL = 10.;
               fam = 1./PIXEL;

             if (Band == 2)
             {
               kx = 0.253;
               ky = 0.329;
             }
             else{
               if (Band == 3)
               {
                 kx = 0.267;
                 ky = 0.28;
               }
               else
               {
                 if (Band == 4)
                 {
                  kx = 0.304;
                  ky = 0.337;
                 }
               }
             }
             
             fwx = sqrt(log((double)2) / kx);
             fwy = sqrt(log((double)2) / ky);
                     
             break;

        case InputParameters::CBERS_4_PAN5: 
             PIXEL = 5.;
               fam = 1./PIXEL;

             if (Band == 1)
             {
               kx = 0.317;
               ky = 0.297;
             }

             else  kx = ky = 1;
             
             fwx = sqrt(log((double)2) / kx);
             fwy = sqrt(log((double)2) / ky);
                     
             break;

        case InputParameters::CBERS_4_WFI:
            PIXEL = 64.;
                fam = 1./PIXEL;
             
             if (Band == 13)
             {
               kx = 0.217; //Valor aproximado. Este valor ser\303\241 confirmado mais tarde pelo Giovani
               ky = 0.215; 
             }
             else{
               if (Band == 14)
               {
                 kx = 0.219; 
                 ky = 0.209; 
               }
               else
               {
                 if (Band == 15)
                 {
                   kx = 0.204;
                   ky = 0.2;
                 }
                 else
                 {
                   if (Band == 16)
                   {
                    kx = 0.201; 
                    ky = 0.196; 
                   }
                 }
               }
             }
            break;
        case InputParameters::RAPIDEYE:
             PIXEL = 5.;
             kx = ky = 6.55;
             fwx = fwy = 0.325;	
             break;

        case InputParameters::QUICKBIRD:
             PIXEL = 2.4;
             kx = ky = 4.92;
             fwx = fwy = 0.39;	
             break;

        case InputParameters::WORLDVIEW2:
             PIXEL = 2.;
             kx = ky = 6.66;
             fwx = fwy = 0.322;	
             break;

          case InputParameters::MODIS_500:
            PIXEL = 463.;
            fam = 1./PIXEL;
                
            //if ( (EifovX == 463.0) && (EifovY == 463.0) )
          //	{
              if (Band == 3) { EifovX = 560.;  EifovY = 499.1; }
              if (Band == 4) { EifovX = 546.1; EifovY = 499.3; } 
              if (Band == 5) { EifovX = 540.;  EifovY = 520.;  }
              if (Band == 6) { EifovX = 540.;  EifovY = 520.;  }
              if (Band == 7) { EifovX = 537.8; EifovY = 521.3; }
          //	}
                
            kx = 4.0 * log((double)2) * pow(fam,2) * pow(EifovX,2);
            ky = 4.0 * log((double)2) * pow(fam,2) * pow(EifovY,2);
            fwx = sqrt(log((double)2) / kx);
            fwy = sqrt(log((double)2) / ky);	
          break;

          case InputParameters::MODIS_250:
            PIXEL = 231.;
            fam = 1./PIXEL;

            //if ( (EifovX == 231.0) && (EifovY == 231.0) )
            //{
              if (Band == 1) { EifovX = 299.6; EifovY = 261.5; }
              if (Band == 2) { EifovX = 291.1; EifovY = 253.9; } 
            //}
                
            kx = 4.0 * log((double)2) * pow(fam,2) * pow(EifovX,2);
            ky = 4.0 * log((double)2) * pow(fam,2) * pow(EifovY,2);
            fwx = sqrt(log((double)2) / kx);
            fwy = sqrt(log((double)2) / ky);	
          break;
          
          default :
          {
            TERP_LOG_AND_RETURN_FALSE( TE_TR( "Invalid sensor" ) );
            break;
          }
      }
        
      Nx = ((int)TX)/2 * (int)FatL;   
      
      XCoefs.resize( Nx + 1 );
      YCoefs.resize( Nx + 1 );
      
      std::vector< double > Janela( Nx + 1, 0 );
      std::vector< double > Peso( FatL, 0 );
      std::vector< double > Gaus( 101, 0 );
      std::vector< double > Freq( 101, 0 );
      std::vector< double > Han( 101, 0 );

      fam = 1./PIXEL;
    // Gaussian - sensor response
      Gaus[0] =1.;
      for(i=1;i<=100;i++)
      {
        f= (0.005*(double)i);
        Gaus[i]= exp(kx*f*f);
        Freq[i]=f*fam;
      }
    // desired function
      for(i=1;i<=100;i++)
      {
        f= (0.005*(double)i);
        if(f>=0. && f<=fwx) Han[i] = 1.;
        if(f>fwx && f<=0.5) Han[i]= (0.5*(1. + cos((double)(M_PI*(f-fwx)/(0.5-fwx)))));
      }
    // window

      for(i=0;i<=Nx;i++)
        Janela[i] = (0.5*(1.+cos(M_PI * (double)i/(double)Nx )));
      /*----------------------------------------------------------------
                     calculate the coefficients in the X direction
       ----------------------------------------------------------------*/
      
      for(i=0;i<=Nx;i++)
      {
        x = i*PIXEL/(double)FatL;
        cop=0.;
        coi=0.;
        for( j=1;j<100;j++)
        {
          fx = (Gaus[j]*Han[j]*cos((double)(2.*M_PI*Freq[j]*x)));
          if(j & 1) coi += fx;
          else      cop += fx;		
        }
        fx = (Gaus[100]*Han[100]*cos((double)(2.*M_PI*Freq[100]*x)));		
        integ = ((1. +4.*coi +2.*cop + fx)/300.);
        XCoefs[i] =integ*Janela[i];
      }
    // Calculate the normalization  factor 
      for(j=0;j<(int)FatL;j++)
      {
        integ = 0.;
        for(ind=j;ind<=Nx;ind+=FatL)
          integ = integ + XCoefs[ind];
        Peso[j] = (2.*integ);
      }
      Peso[0] = Peso[0] - XCoefs[0]; 
    // normalize the filter coefficients
      for(j=0;j<(int)FatL;j++)
        for(ind=j;ind<=Nx;ind+=FatL)
          XCoefs[ind] = XCoefs[ind]/Peso[j]; 

    //	Gaussian function	
      for(i=1;i<=100;i++)
      {
        f= (0.005*(double)i);
        Gaus[i]= (exp(ky*f*f));
        Freq[i]=f*fam;
      }
    //	desired function
      for(i=1;i<=100;i++)
      {
        f=(0.005*(double)i);
        if(f>=0. && f<=fwy) Han[i] = 1.;
          if(f>fwy && f<=0.5) Han[i]= (0.5*(1. + cos((double)(M_PI*(f-fwy)/(0.5-fwy)))));
      }
    //	window

      for(i=0;i<=Nx;i++)
        Janela[i] = (0.5*(1.+cos(M_PI * (double)i/(double)Nx )));

      /*----------------------------------------------------------------
                     calculate the coefficients in the Y direction
       ----------------------------------------------------------------*/

      for(i=0;i<=Nx;i++)
      {
        x = i*PIXEL/(double)FatL;
        cop=0.;
        coi=0.;
        for( j=1;j<100;j++)
        {
          fx=(Gaus[j]*Han[j]*cos((double)(2.*M_PI*Freq[j]*x)));
          if(j & 1) coi += fx;
          else      cop += fx;		
        }
        fx= (Gaus[100]*Han[100]*cos((double)(2.*M_PI*Freq[100]*x)));		
        integ = ((1. +4.*coi +2.*cop + fx)/300.);
          YCoefs[i] =integ*Janela[i];
      }

    // Calculate the normalization  factor 
      for(j=0;j<(int)FatL;j++)
      {
        integ = 0.;
        for( ind=j;ind<=Nx;ind+=FatL)
          integ = integ + YCoefs[ind];
        Peso[j] = (2. * integ) ;
      }
      Peso[0] = Peso[0] - YCoefs[0];

      for(j=0;j<(int)FatL;j++)
        for(ind=j;ind<=Nx;ind+=FatL)
          YCoefs[ind] = YCoefs[ind]/Peso[j]; 
    /*-------------------test: print out the filter coefficients 
      for(ind=0;ind<=Nx;ind++)
        printf("\nVetorx[%d] = %f",ind,Vetorx[ind]);
      for(ind=0;ind<=Nx;ind++)
        printf("\nVetory[%d] = %f",ind,YCoefs[ind]);*/

      return true;
    }
    
    bool Restoration::applyFilterCoefs(
      const std::vector< std::vector< double > >& bandsXCoefs,
      const std::vector< std::vector< double > >& bandsYCoefs,
      const te::rst::Raster& inputRaster,
      te::rst::Raster& outputRaster )
    {
      assert( bandsXCoefs.size() == m_inputParameters.m_inRasterBands.size() );
      assert( bandsYCoefs.size() == m_inputParameters.m_inRasterBands.size() );
      
      std::unique_ptr< te::common::TaskProgress > progressPtr;
      if( m_inputParameters.m_enableProgress )
      {
        progressPtr.reset( new te::common::TaskProgress );
        progressPtr->setTotalSteps( m_inputParameters.m_inRasterBands.size() );
        progressPtr->setMessage( "Restoration in progress" );
      }
      
      unsigned int FatM = 0;
      unsigned int FatL = 0;
      getDecomposedSamplingFactor( m_inputParameters.m_samplingFactor, FatM, FatL );
      
      const double Rate = ((double)FatM) / ((double)FatL);
      
      const unsigned int nInputCols = inputRaster.getNumberOfColumns();
      const unsigned int nInputRows = inputRaster.getNumberOfRows();
      
      const unsigned int nOutputCols = outputRaster.getNumberOfColumns();
      const unsigned int nOutputRows = outputRaster.getNumberOfRows();
      const unsigned int maxOutNColsNRows = std::max( nOutputCols, nOutputRows );
      
      std::vector< int > Pos( maxOutNColsNRows, 0 );
      std::vector< int > Delta( maxOutNColsNRows, 0 );
      for (unsigned int m = 0; m < maxOutNColsNRows ; m++)
      {
        Pos[m]   = (int) ((double)(m+1) * Rate) ;  
        Delta[m] = ((int)FatM)*( ((int)m) + 1 ) - Pos[m] * ((int)FatL); 
      }
      
      const unsigned int TX = getFilterSize();
      
      const unsigned int outputColsOffset = te::common::Round< double, unsigned int >(
        ( ((double)TX) / 2.0 ) * ((double)FatL) / ((double)FatM) );
      const unsigned int outputRowsOffset = outputColsOffset;            
      const unsigned int outputColsBound = nOutputCols - outputColsOffset;
      const unsigned int outputRowsBound = nOutputRows - outputRowsOffset;      
      
      for( std::size_t inRasterBandsIdx = 0 ; inRasterBandsIdx <
        m_inputParameters.m_inRasterBands.size() ; ++inRasterBandsIdx )
      {
        if( m_inputParameters.m_enableProgress )
        {
          progressPtr->pulse();
          if( ! progressPtr->isActive() )
          {
            TERP_INSTANCE_LOG_AND_RETURN_FALSE( "Operation canceled" );
          }
        }
        
        if( 
            m_inputParameters.m_inRasterBands[ inRasterBandsIdx] 
            >=
            inputRaster.getNumberOfBands() 
          )
        {
          return false;
        }
        
        const te::rst::Band& inputRasterBand = *inputRaster.getBand( 
          m_inputParameters.m_inRasterBands[ inRasterBandsIdx] );        
        te::rst::Band& outputRasterBand = *outputRaster.getBand( 
          inRasterBandsIdx );
        
        int		viz, dm, N1, N2,m;
        bool	dummy_flag=false; //AND 06/16

        double sum = 0;//APixelAllocator().newpixel(desc.type());

        std::vector< double > bufin( inputRaster.getNumberOfColumns(), 0 );

        double dummyIn = inputRasterBand.getProperty()->m_noDataValue;
        double dummyOut = outputRasterBand.getProperty()->m_noDataValue;

        double mn = 0;
        double mx = 0;
        te::rp::GetDataTypeRange( inputRasterBand.getProperty()->m_type, mn, mx );
        
        double val;
        
        te::rst::FillBand( &outputRasterBand, std::complex< double >( dummyOut, dummyOut ) );

        for (unsigned int lin = 0 ; lin < nOutputRows; lin++) {
      //---------   Resampling the columns 
          viz =  Pos[lin] + ((int)TX)/2 -1;
          dm = Delta[lin];
          
          N1 = (int)(-((float)TX)/2 - (float)dm/(float)FatL);
          N2 = (int)( ((float)TX)/2 - (float)dm/(float)FatL);
          //N1 = (int)(-TX/2 - dm/FatL);
          //N2 = (int)( TX/2 - dm/FatL);
          
          for (unsigned int col = 0 ; col <  nInputCols; col++){
            sum = 0.;
            dummy_flag = false;
            
            for( m=N1;m<=N2;m++)
            {
              if( ((viz-m) < 0) || ((viz-m)>= (int)nInputRows) )
              {
                dummy_flag=true;
                break;
              }
              
              inputRasterBand.getValue( col, viz-m, val );
              if( val == dummyIn  )
              {
                dummy_flag=true;
                break;
              }
              sum = sum + val * bandsYCoefs[inRasterBandsIdx][abs(dm+m*(int)FatL)] ; 
            }
            if(dummy_flag){
              bufin[col] = dummyIn;
              continue;
            }
            bufin[col] = MIN(mx-1, MAX(mn,sum));
            if(dummy_flag) continue;
            bufin[col] = MIN(mx-1, MAX(mn,sum));
          }
          
          for (unsigned int c = 0 ; c < nOutputCols; c++){
      //-----------  Resampling the rows
            viz = Pos[c] + TX/2 -1;
            dm = Delta[c];
            
            N1 = (int)(-((float)TX)/2 - (float)dm/(float)FatL);
            N2 = (int)( ((float)TX)/2 - (float)dm/(float)FatL);
            
            sum = 0.;
            dummy_flag = false;
            for( m=N1;m<=N2;m++)
            {
              if( bufin[viz-m] == dummyIn )
              {
                if( ( c < outputColsBound ) && ( lin < outputRowsBound ) )
                {
                  outputRasterBand.setValue( c+outputColsOffset, lin+outputRowsOffset, dummyOut );
                }
//                outputRasterBand.setValue( c, lin, dummyOut );
                
                dummy_flag=true;
                
                break;
              }
              sum = sum + bufin[viz-m] * bandsXCoefs[ inRasterBandsIdx ][abs(dm+m*(int)FatL)] ;
            }			
            if(dummy_flag) continue;
            val = MIN(MAX(mn,sum),mx-1);
            
            if( ( c < outputColsBound ) && ( lin < outputRowsBound ) )
            {
              outputRasterBand.setValue( c+outputColsOffset, lin+outputRowsOffset, val );
            }
//            outputRasterBand.setValue( c, lin, val );
          }
        }
      }
      
      return true;
    }
    
    void Restoration::getDecomposedSamplingFactor( 
      const InputParameters::SamplingFactorType& sFactortype,
      unsigned int& upperValue,
      unsigned int& lowerValue )
    {
      switch( sFactortype )
      {
        case InputParameters::SAMPLING_FACTOR_1_BY_2 :
        {
          upperValue = 1;
          lowerValue = 2;
          break;
        }
        case InputParameters::SAMPLING_FACTOR_1_BY_3 :
        {
          upperValue = 1;
          lowerValue = 3;
          break;
        }        
        case InputParameters::SAMPLING_FACTOR_1_BY_4 :
        {
          upperValue = 1;
          lowerValue = 4;
          break;
        }        
        case InputParameters::SAMPLING_FACTOR_1_BY_5 :
        {
          upperValue = 1;
          lowerValue = 5;
          break;
        }        
        case InputParameters::SAMPLING_FACTOR_1_BY_6 :
        {
          upperValue = 1;
          lowerValue = 6;
          break;
        }        
        case InputParameters::SAMPLING_FACTOR_2_BY_3 :
        {
          upperValue = 2;
          lowerValue = 3;
          break;
        }        
        case InputParameters::SAMPLING_FACTOR_3_BY_4 :
        {
          upperValue = 3;
          lowerValue = 4;
          break;
        }
        default:
        {
          TERP_LOG_AND_THROW( "Invalid sampling factor" );
          break;
        }
      }
    }
    
    unsigned int Restoration::getFilterSize() const
    {
      switch( m_inputParameters.m_sensorType )
      {
        case InputParameters::SPOTP :
        case InputParameters::SPOTX :
        case InputParameters::TM :
        case InputParameters::ETMX :
        case InputParameters::ETMP :
        case InputParameters::LANDSAT_8 :
        case InputParameters::CBERS_CCD :
        case InputParameters::CBERS_CCRX :
        case InputParameters::CBERS2_CCD :
        case InputParameters::CBERS2B_CCD :
        case InputParameters::CBERS2B_HRC :
        case InputParameters::CBERS_4_MUX :
        case InputParameters::CBERS_4_PAN10 :
        case InputParameters::CBERS_4_PAN5 :
        case InputParameters::CBERS_4_WFI :
        case InputParameters::RAPIDEYE :
        case InputParameters::QUICKBIRD :
        case InputParameters::WORLDVIEW2 :
        case InputParameters::MODIS_250 :
        case InputParameters::MODIS_500 :
        {
          return 11;
          break;
        }
        
        default:
        {
          TERP_LOG_AND_THROW( "Invalid sensor type" );
          break;
        }
      }
      
      return 0;
    }
    
    void Restoration::InputParameters::getResolutionSamplingFactor( 
      const InputParameters::SamplingFactorType& sFactortype,
      double& resSamplingFactor )
    {
      unsigned int upperValue = 0;
      unsigned int lowerValue = 0;
      getDecomposedSamplingFactor( sFactortype, upperValue, lowerValue );
      
      resSamplingFactor = ((double)upperValue) / ((double)lowerValue);
    }
  
  } // end namespace rp
}   // end namespace te

