/*  Copyright (C) 2015 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/rp/SegmenterRegionGrowingBaatzMerger.cpp
  \brief Segmenter Baatz merger used in region growing process.
*/

#include "SegmenterRegionGrowingBaatzMerger.h"
#include "Macros.h"

// Baatz Edge Lengh
#define BAATZ_EL( featPtr ) featPtr[ 0 ]

// Baatz Compactness
#define BAATZ_CO( featPtr ) featPtr[ 1 ]

// Baatz Smoothness
#define BAATZ_SM( featPtr ) featPtr[ 2 ]

// Baatz sums
#define BAATZ_SU( featPtr, band ) featPtr[ 3 + band ]

// Baatz square sums
#define BAATZ_SS( featPtr, bandsNmb, band ) featPtr[ 3 + bandsNmb + band ]

// Baatz stddev
#define BAATZ_ST( featPtr, bandsNmb, band ) featPtr[ 3 + ( 2 * bandsNmb ) + band ]

// Baatz stddev start ponter
#define BAATZ_STSTARTPTR( featPtr, bandsNmb ) ( featPtr + 3 + ( 2 * bandsNmb ) )

// The 8 cardinals directions offsets used by getTouchingEdgeLength method
#define NORTHWEST 0
#define NORTH     1
#define NORTHEAST 2
#define EAST      3
#define SOUTHEAST 4
#define SOUTH     5
#define SOUTHWEST 6
#define WEST      7

namespace te
{
  namespace rp
  {
    SegmenterRegionGrowingBaatzMerger::MergeData::MergeData(
      const unsigned int featuresNumber )
    {
      if( featuresNumber )
      {
        m_features = new rg::BaatzFeatureType[ featuresNumber ];
      }
      else
      {
        m_features = 0;
      }
    }
    
    SegmenterRegionGrowingBaatzMerger::MergeData::~MergeData()
    {
      if( m_features )
      {
        delete[] m_features;
      }
    }    
    
    // ---------------------------------------------------------------------
    
    SegmenterRegionGrowingBaatzMerger::SegmenterRegionGrowingBaatzMerger(
      const double& colorWeight, const double& compactnessWeight,
      const std::vector< double >& bandsWeights,
      const SegmentsIdsMatrixT& segmentsIds )
      : 
        m_segmentsIdsMatrix( segmentsIds ),
        m_allSegsCompactnessOffset( 0 ),
        m_allSegsCompactnessGain( 1.0 ),
        m_allSegsSmoothnessOffset( 0 ),
        m_allSegsSmoothnessGain( 0 ),
        m_colorWeight( (rg::BaatzFeatureType)colorWeight ),
        m_compactnessWeight( (rg::BaatzFeatureType)compactnessWeight )
    {
      m_dirsColOffsVecPtr = new int[ 8 ];
      m_dirsColOffsVecPtr[ NORTHWEST ] = -1;
      m_dirsColOffsVecPtr[ NORTH ] = 0;
      m_dirsColOffsVecPtr[ NORTHEAST ] = 1;
      m_dirsColOffsVecPtr[ EAST ] = 1;
      m_dirsColOffsVecPtr[ SOUTHEAST ] = 1;
      m_dirsColOffsVecPtr[ SOUTH ] = 0;
      m_dirsColOffsVecPtr[ SOUTHWEST ] = -1;
      m_dirsColOffsVecPtr[ WEST ] = -1;
      
      m_dirsRowOffsVecPtr = new int[ 8 ];
      m_dirsRowOffsVecPtr[ NORTHWEST ] = -1;
      m_dirsRowOffsVecPtr[ NORTH ] = -1;
      m_dirsRowOffsVecPtr[ NORTHEAST ] = -1;
      m_dirsRowOffsVecPtr[ EAST ] = 0;
      m_dirsRowOffsVecPtr[ SOUTHEAST ] = 1;
      m_dirsRowOffsVecPtr[ SOUTH ] = 1;
      m_dirsRowOffsVecPtr[ SOUTHWEST ] = 1;
      m_dirsRowOffsVecPtr[ WEST ] = 0;
      
      m_segmentsIdsMatrixNRows = m_segmentsIdsMatrix.getLinesNumber();
      m_segmentsIdsMatrixNCols = m_segmentsIdsMatrix.getColumnsNumber();
      m_bandsNumber = (unsigned int)bandsWeights.size();      
      m_bandsWeights.resize( m_bandsNumber, 1 );
      m_segmentFeaturesSize = 3 + ( 3 * m_bandsNumber );
      
      for( unsigned int band = 0 ; band < m_bandsNumber ; ++band )
      {
        m_bandsWeights[ band ] = (rg::BaatzFeatureType)bandsWeights[ band ];
      }
    }

    SegmenterRegionGrowingBaatzMerger::~SegmenterRegionGrowingBaatzMerger()
    {
      delete[] m_dirsColOffsVecPtr;
      delete[] m_dirsRowOffsVecPtr;
    }

    DissimilarityTypeT
    SegmenterRegionGrowingBaatzMerger::getDissimilarity( SegmenterRegionGrowingSegment< rg::BaatzFeatureType > const * const segment1Ptr, 
      SegmenterRegionGrowingSegment< rg::BaatzFeatureType > const * const segment2Ptr, 
      SegmenterRegionGrowingMergeData * const mergeDataPtr ) const
    {
      assert( segment1Ptr );
      assert( segment1Ptr->m_features );
      assert( segment2Ptr );
      assert( segment2Ptr->m_features );
      assert( mergeDataPtr );

      // globals
      
      m_getDissimilarity_mergeDataPtr = (MergeData*)mergeDataPtr;

      m_getDissimilarity_sizeSeg1D = (rg::BaatzFeatureType)segment1Ptr->m_size;      
      
      m_getDissimilarity_sizeSeg2D = (rg::BaatzFeatureType)segment2Ptr->m_size;      
      
      m_getDissimilarity_sizeUnionD = m_getDissimilarity_sizeSeg1D + m_getDissimilarity_sizeSeg2D;
      TERP_DEBUG_TRUE_OR_THROW( m_getDissimilarity_sizeUnionD, "Internal error" );

      m_getDissimilarity_widthUnionD = std::max( segment1Ptr->m_xBound,
        segment2Ptr->m_xBound ) - std::min( segment1Ptr->m_xStart,
        segment2Ptr->m_xStart );
      m_getDissimilarity_heightUnionD = std::max( segment1Ptr->m_yBound,
        segment2Ptr->m_yBound ) - std::min( segment1Ptr->m_yStart,
        segment2Ptr->m_yStart );
      
      // Finding the form heterogeneity
      
      if( m_colorWeight != 1.0 )
      {
        getTouchingEdgeLength(
          segment1Ptr->m_xStart, 
          segment1Ptr->m_yStart,
          segment1Ptr->m_xBound, 
          segment1Ptr->m_yBound, 
          segment2Ptr->m_xStart, 
          segment2Ptr->m_yStart,
          segment2Ptr->m_xBound, 
          segment2Ptr->m_yBound, 
          segment1Ptr->m_id,
          segment2Ptr->m_id,
          m_getDissimilarity_touchingEdgeLength );
        
        BAATZ_EL( m_getDissimilarity_mergeDataPtr->m_features ) = 
          BAATZ_EL( segment1Ptr->m_features ) - 
          ( (rg::BaatzFeatureType)m_getDissimilarity_touchingEdgeLength )
          +
          BAATZ_EL( segment2Ptr->m_features ) - 
          ( (rg::BaatzFeatureType)m_getDissimilarity_touchingEdgeLength );
        
        BAATZ_CO( m_getDissimilarity_mergeDataPtr->m_features ) = (rg::BaatzFeatureType)(
          BAATZ_EL( m_getDissimilarity_mergeDataPtr->m_features ) /
          std::sqrt( m_getDissimilarity_sizeUnionD ) );
          
        BAATZ_SM( m_getDissimilarity_mergeDataPtr->m_features ) =
          BAATZ_EL( m_getDissimilarity_mergeDataPtr->m_features ) 
          /
          (rg::BaatzFeatureType)(
            (
              2 * m_getDissimilarity_widthUnionD
            )
            +
            (
              2 * m_getDissimilarity_heightUnionD
            )
          );
          
        m_getDissimilarity_hCompact = 
          (
            (
              (
                BAATZ_CO( m_getDissimilarity_mergeDataPtr->m_features )
                +
                m_allSegsCompactnessOffset
              )
              *
              m_allSegsCompactnessGain
            )
            -
            (
              (
                (
                  (
                    (
                      BAATZ_CO( segment1Ptr->m_features )
                      +
                      m_allSegsCompactnessOffset
                    )
                    *
                    m_allSegsCompactnessGain
                  )
                  *
                  m_getDissimilarity_sizeSeg1D
                )
                +
                (
                  (
                    (
                      BAATZ_CO( segment2Ptr->m_features )
                      +
                      m_allSegsCompactnessOffset
                    )
                    *
                    m_allSegsCompactnessGain
                  )
                  *
                  m_getDissimilarity_sizeSeg2D
                )
              )
              /
              m_getDissimilarity_sizeUnionD
            )
          );
        
        m_getDissimilarity_hSmooth =
          (
            (
              (
                BAATZ_SM( m_getDissimilarity_mergeDataPtr->m_features )
                +
                m_allSegsSmoothnessOffset
              )
              *
              m_allSegsSmoothnessGain
            )
            -
            (
              (
                (
                  (
                    (
                      BAATZ_SM( segment1Ptr->m_features )
                      +
                      m_allSegsSmoothnessOffset
                    )
                    *
                    m_allSegsSmoothnessGain
                  )
                  *
                  m_getDissimilarity_sizeSeg1D
                )
                +
                (
                  (
                    (
                      BAATZ_SM( segment2Ptr->m_features )
                      +
                      m_allSegsSmoothnessOffset
                    )
                    *
                    m_allSegsSmoothnessGain
                  )
                  *
                  m_getDissimilarity_sizeSeg2D
                )
              )
              /
              m_getDissimilarity_sizeUnionD
            )
          );
          
        m_getDissimilarity_hForm = 
          (
            (
              m_compactnessWeight 
              *
              m_getDissimilarity_hCompact
            )
            +
            (
              ( 1.0f - m_compactnessWeight )
              *
              m_getDissimilarity_hSmooth
            )
          );
      }
      
      // Finding the color heterogeneity
      
      if( m_colorWeight != 0.0 )
      {
        m_getDissimilarity_hColor = 0;
        
        for( m_getDissimilarity_sumsIdx = 0 ; m_getDissimilarity_sumsIdx < m_bandsNumber ; ++m_getDissimilarity_sumsIdx )
        {
          m_getDissimilarity_sumUnion = 
            BAATZ_SU( segment1Ptr->m_features, m_getDissimilarity_sumsIdx ) 
            + 
            BAATZ_SU( segment2Ptr->m_features, m_getDissimilarity_sumsIdx );
          BAATZ_SU( m_getDissimilarity_mergeDataPtr->m_features, m_getDissimilarity_sumsIdx ) = 
            m_getDissimilarity_sumUnion;        
          
          m_getDissimilarity_squaresSumUnion = 
            BAATZ_SS( segment1Ptr->m_features, m_bandsNumber, m_getDissimilarity_sumsIdx ) 
            +
            BAATZ_SS( segment2Ptr->m_features, m_bandsNumber, m_getDissimilarity_sumsIdx );
          BAATZ_SS( m_getDissimilarity_mergeDataPtr->m_features, m_bandsNumber, m_getDissimilarity_sumsIdx ) = m_getDissimilarity_squaresSumUnion;
          
          m_getDissimilarity_meanUnion = m_getDissimilarity_sumUnion / m_getDissimilarity_sizeUnionD;
          
          m_getDissimilarity_stdDevUnion =
            std::sqrt(
              std::max(
                (rg::BaatzFeatureType)0
                ,
                (
                  (
                    m_getDissimilarity_squaresSumUnion
                    -
                    (
                      ((rg::BaatzFeatureType)2) * m_getDissimilarity_meanUnion * m_getDissimilarity_sumUnion
                    )
                    +
                    (
                      m_getDissimilarity_sizeUnionD * m_getDissimilarity_meanUnion * m_getDissimilarity_meanUnion
                    )
                  )
                  /
                  m_getDissimilarity_sizeUnionD
                )
              )
            );
          BAATZ_ST( m_getDissimilarity_mergeDataPtr->m_features, m_bandsNumber, m_getDissimilarity_sumsIdx ) =
            m_getDissimilarity_stdDevUnion;        
           
          m_getDissimilarity_hColor += 
            ( 
              m_bandsWeights[ m_getDissimilarity_sumsIdx ]
              *
              (
                (
                  (
                    m_getDissimilarity_stdDevUnion
                    +
                    m_allSegsStdDevOffset
                  )
                  *
                  m_allSegsStdDevGain
                )
                -
                (
                  (
                    (
                      (
                        BAATZ_ST( segment1Ptr->m_features, m_bandsNumber, m_getDissimilarity_sumsIdx )
                        +
                        m_allSegsStdDevOffset
                      )
                      *
                      m_allSegsStdDevGain
                    )                    
                    +
                    (
                      (
                        BAATZ_ST( segment2Ptr->m_features, m_bandsNumber, m_getDissimilarity_sumsIdx )
                        +
                        m_allSegsStdDevOffset
                      )
                      *
                      m_allSegsStdDevGain
                    )                    
                  )
                  /
                  m_getDissimilarity_sizeUnionD
                )
              )
            );
        }
      }
      
      m_getDissimilarity_hColor = 
        (
          ( 
            m_getDissimilarity_hColor 
            * 
            m_colorWeight 
          ) 
          + 
          (
            ( 1.0f - m_colorWeight )
            *
            m_getDissimilarity_hForm
          )
        );       
        
      return std::max( (DissimilarityTypeT)m_getDissimilarity_hColor, (DissimilarityTypeT)0 );
    }
    
    void SegmenterRegionGrowingBaatzMerger::mergeFeatures( 
      SegmenterRegionGrowingSegment< rg::BaatzFeatureType > * const segment1Ptr, 
      SegmenterRegionGrowingSegment< rg::BaatzFeatureType > const * const segment2Ptr, 
      SegmenterRegionGrowingMergeData const * const mergeDataPtr ) const
    {
      assert( segment1Ptr );
      assert( segment1Ptr->m_features );
      assert( mergeDataPtr );
      
      // Merging basic features
      
      m_mergeFeatures_mergeDataPtr = (MergeData const*)mergeDataPtr;      
        
      // Merging specific features   
      
      memcpy( segment1Ptr->m_features, m_mergeFeatures_mergeDataPtr->m_features,
        sizeof( rg::BaatzFeatureType ) * m_segmentFeaturesSize );
    }    
    
    void SegmenterRegionGrowingBaatzMerger::update(
      SegmenterRegionGrowingSegment< rg::BaatzFeatureType >* const actSegsListHeadPtr )
    {
      m_update_compactnessMin = 
        std::numeric_limits< rg::BaatzFeatureType >::max();
      m_update_compactnessMax = -1.0f * 
        std::numeric_limits< rg::BaatzFeatureType >::max();

      m_update_smoothnessMin =
        std::numeric_limits< rg::BaatzFeatureType >::max();
      m_update_smoothnessMax = -1.0f *
        std::numeric_limits< rg::BaatzFeatureType >::max();
        
      m_update_stdDevMin =
        std::numeric_limits< rg::BaatzFeatureType >::max();
      m_update_stdDevMax = -1.0f *
        std::numeric_limits< rg::BaatzFeatureType >::max();          
        
      m_update_currentActSegPtr = actSegsListHeadPtr;
      
      while( m_update_currentActSegPtr )
      {
        m_update_featuresPtr = m_update_currentActSegPtr->m_features;
        
        if( m_colorWeight != 1.0 )
        {
          if( m_update_compactnessMin > BAATZ_CO( m_update_featuresPtr ) )
          {
            m_update_compactnessMin = BAATZ_CO( m_update_featuresPtr );        
          }
          if( m_update_compactnessMax < BAATZ_CO( m_update_featuresPtr ) )
          {
            m_update_compactnessMax = BAATZ_CO( m_update_featuresPtr );        
          }
          
          if( m_update_smoothnessMin > BAATZ_SM( m_update_featuresPtr ) )
          {
            m_update_smoothnessMin = BAATZ_SM( m_update_featuresPtr );
          }
          if( m_update_smoothnessMax < BAATZ_SM( m_update_featuresPtr ) )
          {
            m_update_smoothnessMax = BAATZ_SM( m_update_featuresPtr );
          }
        }
        
        if( m_colorWeight != 0.0 )
        {
          m_update_stdDevPtr = BAATZ_STSTARTPTR( m_update_featuresPtr, m_bandsNumber ); 
          
          for( m_update_band = 0 ; m_update_band < m_bandsNumber ; ++m_update_band, ++m_update_stdDevPtr )
          {
            if( m_update_stdDevMin > *m_update_stdDevPtr )
            {
              m_update_stdDevMin = *m_update_stdDevPtr;
            }
            if( m_update_stdDevMax < *m_update_stdDevPtr )
            {
              m_update_stdDevMax = *m_update_stdDevPtr;
            }
          }
        }
        
        m_update_currentActSegPtr = m_update_currentActSegPtr->m_nextActiveSegment;
      }       
      
      if( m_update_compactnessMax == m_update_compactnessMin )
      {
        m_allSegsCompactnessOffset = 0.0;
        
        if( m_update_compactnessMax == 0.0 )
          m_allSegsCompactnessGain = 1.0;
        else
          m_allSegsCompactnessGain = 1.0f / m_update_compactnessMax;
      }
      else
      {
        m_allSegsCompactnessOffset = -1.0f  * m_update_compactnessMin;
        m_allSegsCompactnessGain = 1.0f / ( m_update_compactnessMax - m_update_compactnessMin );
      }
      
      if( m_update_smoothnessMax == m_update_smoothnessMin )
      {
        m_allSegsSmoothnessOffset = 0.0;
        
        if( m_update_smoothnessMax == 0.0 )
          m_allSegsSmoothnessGain = 1.0;
        else
          m_allSegsSmoothnessGain = 1.0f / m_update_smoothnessMax;
      }
      else
      {
        m_allSegsSmoothnessOffset = -1.0f  * m_update_smoothnessMin;
        m_allSegsSmoothnessGain = 1.0f / ( m_update_smoothnessMax - m_update_smoothnessMin );      
      }
      
      if( m_update_stdDevMax == m_update_stdDevMin )
      {
        m_allSegsStdDevOffset = 0.0;
        
        if( m_update_stdDevMax == 0.0 )
          m_allSegsStdDevGain = 1.0;
        else
          m_allSegsStdDevGain = 1.0f / m_update_stdDevMax;
      }
      else
      {
        m_allSegsStdDevOffset = -1.0f  * m_update_stdDevMin;
        m_allSegsStdDevGain = 1.0f / ( m_update_stdDevMax - m_update_stdDevMin );      
      }        
    }
    
    void SegmenterRegionGrowingBaatzMerger::createMergeData( SegmenterRegionGrowingMergeData** mergeDataPtr ) const
    {
      *mergeDataPtr = new MergeData( m_segmentFeaturesSize );
    }

    void SegmenterRegionGrowingBaatzMerger::getTouchingEdgeLength( 
      const unsigned int& colStart1, 
      const unsigned int& rowStart1,
      const unsigned int& colBound1, 
      const unsigned int& rowBound1,
      const unsigned int& colStart2, 
      const unsigned int& rowStart2,
      const unsigned int& colBound2, 
      const unsigned int& rowBound2,          
      const SegmenterSegmentsBlock::SegmentIdDataType& id1,
      const SegmenterSegmentsBlock::SegmentIdDataType& id2,
      unsigned int& touchingEdgeLength ) const
    {
      TERP_DEBUG_TRUE_OR_THROW( colStart1 < m_segmentsIdsMatrix.getColumnsNumber(),
        "Internal Error" )
      TERP_DEBUG_TRUE_OR_THROW( colBound1 <= m_segmentsIdsMatrix.getColumnsNumber(),
        "Internal Error" )
      TERP_DEBUG_TRUE_OR_THROW( rowStart1 < m_segmentsIdsMatrix.getLinesNumber(),
        "Internal Error" )
      TERP_DEBUG_TRUE_OR_THROW( rowBound1 <= m_segmentsIdsMatrix.getLinesNumber(),
        "Internal Error" )
      TERP_DEBUG_TRUE_OR_THROW( colStart2 < m_segmentsIdsMatrix.getColumnsNumber(),
        "Internal Error" )
      TERP_DEBUG_TRUE_OR_THROW( colBound2 <= m_segmentsIdsMatrix.getColumnsNumber(),
        "Internal Error" )
      TERP_DEBUG_TRUE_OR_THROW( rowStart2 < m_segmentsIdsMatrix.getLinesNumber(),
        "Internal Error" )
      TERP_DEBUG_TRUE_OR_THROW( rowBound2 <= m_segmentsIdsMatrix.getLinesNumber(),
        "Internal Error" )      

      touchingEdgeLength = 0;

      // locating the first upper-left segment 1 pixel

      m_getTouchingEdgeLength_seg1ColStart = colBound1;

      for( m_getTouchingEdgeLength_row = (int)rowStart1 ;
        m_getTouchingEdgeLength_row < (int)rowBound1 ;
        ++m_getTouchingEdgeLength_row )
      {
        for( m_getTouchingEdgeLength_col = (int)colStart1 ; 
          m_getTouchingEdgeLength_col < (int)colBound1 ; 
          ++m_getTouchingEdgeLength_col )
        {
          if( 
              m_segmentsIdsMatrix[ m_getTouchingEdgeLength_row ][ m_getTouchingEdgeLength_col ]
              == 
              id1 
            )
          {
            m_getTouchingEdgeLength_seg1RowStart = m_getTouchingEdgeLength_row;
            m_getTouchingEdgeLength_seg1ColStart = m_getTouchingEdgeLength_col;
            
            //break the external loop
            m_getTouchingEdgeLength_row = rowBound1;
            
            break;
          }
        }
      }
      
      TERP_DEBUG_TRUE_OR_THROW( m_getTouchingEdgeLength_seg1ColStart != 
        (int)colBound1, "Segment not found" );      
      
      // going through the border locating seg2 pixels
      
       bool currentSegmentInitialLeavingDirectionWasSet = false;
       short currentDirection = EAST;
       short currentPixelInitialNextTestDirection = EAST;
       short currentPixelNextTestDirection = EAST;
       int currentCol = m_getTouchingEdgeLength_seg1ColStart;
       int currentRow = m_getTouchingEdgeLength_seg1RowStart;
       int nextCol = 0;
       int nextRow = 0;
       bool nextPixelWasFound = false;

      while(true)
      {
        // Do the current pixel touches segment 2 ?

        if( 
            (
              ( currentRow != 0 )
              &&
              ( m_segmentsIdsMatrix[ currentRow - 1 ][ currentCol ] == id2 )
            )
            ||
            (
              ( currentCol != 0 )
              &&
              ( m_segmentsIdsMatrix[ currentRow ][ currentCol - 1 ] == id2 )
            )
            ||
            (
              ( ( currentRow + 1 ) < (int)m_segmentsIdsMatrixNRows )
              &&
              ( m_segmentsIdsMatrix[ currentRow + 1 ][ currentCol ] == id2 )
            )
            ||
            ( 
              ( ( currentCol + 1 ) < (int)m_segmentsIdsMatrixNCols )
              &&
              ( m_segmentsIdsMatrix[ currentRow ][ currentCol + 1 ] == id2 )
            )
          )
        {
          ++touchingEdgeLength;
        }

        // finding the next direction

        nextPixelWasFound = false;
        currentPixelInitialNextTestDirection = currentPixelNextTestDirection = 
          (currentDirection + 6) % 8;

        do
        {
          nextCol = currentCol + m_dirsColOffsVecPtr[currentPixelNextTestDirection];
          nextRow = currentRow + m_dirsRowOffsVecPtr[currentPixelNextTestDirection];

          if(
              (nextCol > -1) 
              &&
              (nextRow > -1)
              &&
              (nextCol < (int)m_segmentsIdsMatrixNCols ) 
              &&
              (nextRow < (int)m_segmentsIdsMatrixNRows)
            )
          {
            if (m_segmentsIdsMatrix( nextRow, nextCol ) == id1)
            {
              nextPixelWasFound = true;
              currentDirection = currentPixelNextTestDirection;
              break;
            }
          }

          currentPixelNextTestDirection = (currentPixelNextTestDirection + 2) % 8;
        } 
        while ( 
                currentPixelInitialNextTestDirection 
                != 
                currentPixelNextTestDirection
              );

        // Setting the initial segment leaving direction

        if( !currentSegmentInitialLeavingDirectionWasSet )
          currentSegmentInitialLeavingDirectionWasSet = true;

        if(
            ( !nextPixelWasFound )
            ||
            (
               ( nextCol == m_getTouchingEdgeLength_seg1ColStart ) 
               && 
               ( nextRow == m_getTouchingEdgeLength_seg1RowStart )
            )
          )
        {
          break;
        }

        // updating the current state

        currentCol = nextCol;
        currentRow = nextRow;
      }
    }
  } // end namespace rp
}   // end namespace te    

