/*  Copyright (C) 2015 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/rp/SegmenterRegionGrowingBaatzMerger.h
  \brief Segmenter Baatz merger used in region growing process.
*/

#ifndef __TERRALIB_RP_INTERNAL_SEGMENTERREGIONGROWINGBAATZMERGER_H
#define __TERRALIB_RP_INTERNAL_SEGMENTERREGIONGROWINGBAATZMERGER_H

#include "SegmenterRegionGrowingFunctions.h"
#include "SegmenterRegionGrowingMerger.h"
#include "SegmenterSegmentsBlock.h"
#include "SegmenterRegionGrowingMergeData.h"

#include "Config.h"

#include <vector>

namespace te
{
  namespace rp
  {
    /*!
      \class SegmenterRegionGrowingBaatzMerger
      \brief Baatz based Segments merger
    */        
    class TERPEXPORT SegmenterRegionGrowingBaatzMerger :
      public SegmenterRegionGrowingMerger< rg::BaatzFeatureType >
    {
      public:
        /*!
          \brief Default constructor.
          \param bandsWeights A reference to an external valid structure where each bands weight are stored.
          \param segmentsIds //!< A reference to an external valid structure where all segments IDs are stored.
          \param segmentsMatrix //!< A reference to an external valid segments matrix.
          \param colorWeight //!< The weight given to the color component, deafult:0.5, valid range: [0,1].
          \param compactnessWeight //!< The weight given to the compactness component, deafult:0.5, valid range: [0,1].
        */
        SegmenterRegionGrowingBaatzMerger( const double& colorWeight,
          const double& compactnessWeight,
          const std::vector< double >& bandsWeights,
          const SegmentsIdsMatrixT& segmentsIds );
      
        ~SegmenterRegionGrowingBaatzMerger();
            
        //overload        
        DissimilarityTypeT
          getDissimilarity( SegmenterRegionGrowingSegment< rg::BaatzFeatureType > const * const segment1Ptr, 
          SegmenterRegionGrowingSegment< rg::BaatzFeatureType > const * const segment2Ptr,
          SegmenterRegionGrowingMergeData * const mergeDataPtr ) const;
              
        //overload                
        void mergeFeatures( SegmenterRegionGrowingSegment< rg::BaatzFeatureType > * const segment1Ptr, 
          SegmenterRegionGrowingSegment< rg::BaatzFeatureType > const * const segment2Ptr, 
          SegmenterRegionGrowingMergeData const * const mergeDataPtr ) const;
              
        //overload
        void update( SegmenterRegionGrowingSegment< rg::BaatzFeatureType >* const actSegsListHeadPtr );
            
        //overload
        inline unsigned int getSegmentFeaturesSize() const { return m_segmentFeaturesSize; };
        
        void createMergeData( SegmenterRegionGrowingMergeData** mergeDataPtr ) const;
              
      protected :
        /*!
          \class MergeData
          \brief Baatz specific merge data.
        */
        class MergeData : public SegmenterRegionGrowingMergeData
        {
          public:
            
            MergeData( const unsigned int featuresNumber );
            
            ~MergeData();

            /*!
              \brief Vctor of segment features (this class will take the ownership and delete[] at destruction time).
            */
            rg::BaatzFeatureType* m_features;
            
          private:
            
            MergeData( const MergeData& other ) {};
            const MergeData& operator=( const MergeData& other ) { return *this; };
        };

        mutable int* m_dirsColOffsVecPtr; //!< pointer to a internal vector of directions column offsets.
        mutable int* m_dirsRowOffsVecPtr; //!< pointer to a internal vector of directions row offsets.
        const SegmentsIdsMatrixT& m_segmentsIdsMatrix; //!< A reference to an external valid structure where each all segments IDs are stored.
        unsigned int m_segmentsIdsMatrixNRows; //!<  Matrix rows numnber;
        unsigned int m_segmentsIdsMatrixNCols; //!<  Matrix cols numnber;
        unsigned int m_bandsNumber; //!< The number of features (bands).
        unsigned int m_segmentFeaturesSize; //< the required segments features vector size (numer of elements).
        rg::BaatzFeatureType m_allSegsCompactnessOffset; //!< The offsets applied to normalize the compactness value.
        rg::BaatzFeatureType m_allSegsCompactnessGain; //!< The gains applied to normalize the compactness value.
        rg::BaatzFeatureType m_allSegsSmoothnessOffset; //!< The offsets applied to normalize the smoothness value.
        rg::BaatzFeatureType m_allSegsSmoothnessGain; //!< The gains applied to normalize the smoothness value.
        rg::BaatzFeatureType m_allSegsStdDevOffset; //!< The offsets applied to normalize the standard deviation value.
        rg::BaatzFeatureType m_allSegsStdDevGain; //!< The gains applied to normalize the standard deviation value.
        rg::BaatzFeatureType m_colorWeight; //!< The weight given to the color component, deafult:0.5, valid range: [0,1].
        rg::BaatzFeatureType m_compactnessWeight; //!< The weight given to the compactness component, deafult:0.5, valid range: [0,1].
        std::vector< rg::BaatzFeatureType > m_bandsWeights; //!< A vector where each bands weight are stored.
        
        // Variables used by the method getDissimilarity            
        mutable MergeData* m_getDissimilarity_mergeDataPtr;
        mutable rg::BaatzFeatureType m_getDissimilarity_sizeUnionD;
        mutable rg::BaatzFeatureType m_getDissimilarity_sizeSeg1D;
        mutable rg::BaatzFeatureType m_getDissimilarity_sizeSeg2D;
        mutable unsigned int m_getDissimilarity_touchingEdgeLength;
        mutable rg::BaatzFeatureType m_getDissimilarity_hCompact; 
        mutable rg::BaatzFeatureType m_getDissimilarity_hSmooth;
        mutable rg::BaatzFeatureType m_getDissimilarity_hForm;
        mutable rg::BaatzFeatureType m_getDissimilarity_hColor;
        mutable rg::BaatzFeatureType m_getDissimilarity_sumUnion;
        mutable rg::BaatzFeatureType m_getDissimilarity_squaresSumUnion;
        mutable rg::BaatzFeatureType m_getDissimilarity_meanUnion;
        mutable rg::BaatzFeatureType m_getDissimilarity_stdDevUnion;
        mutable unsigned int m_getDissimilarity_sumsIdx;
        mutable rg::BaatzFeatureType m_getDissimilarity_widthUnionD;
        mutable rg::BaatzFeatureType m_getDissimilarity_heightUnionD;
        
        //Variables used by the method mergeFeaturs
        mutable MergeData const * m_mergeFeatures_mergeDataPtr;

        // Variables used by the method update
        rg::BaatzFeatureType m_update_compactnessMin;
        rg::BaatzFeatureType m_update_compactnessMax;
        rg::BaatzFeatureType m_update_smoothnessMin;
        rg::BaatzFeatureType m_update_smoothnessMax;
        rg::BaatzFeatureType m_update_stdDevMin;
        rg::BaatzFeatureType m_update_stdDevMax;
        rg::BaatzFeatureType* m_update_featuresPtr;
        SegmenterRegionGrowingSegment< rg::BaatzFeatureType >* m_update_currentActSegPtr;
        unsigned int m_update_band;
        rg::BaatzFeatureType const* m_update_stdDevPtr;

        // Variables used by getTouchingEdgeLength
        mutable int m_getTouchingEdgeLength_row;
        mutable int m_getTouchingEdgeLength_col;
        mutable int m_getTouchingEdgeLength_seg1ColStart;
        mutable int m_getTouchingEdgeLength_seg1RowStart;
        

        /*!
          \brief Returns the count of points from region 1 (with ID1) touching the region 2 (with ID2).
          \param colStart1 The upper left X of the bounding box surrounding region 1.
          \param rowStart1 The upper left Y of the bounding box surrounding region 1.
          \param colBound1 The lower right X bound of the bounding box surrounding region 1.
          \param rowBound1 The lower right Y bound of the bounding box surrounding region 1.
          \param colStart2 The upper left X of the bounding box surrounding region 2.
          \param rowStart2 The upper left Y of the bounding box surrounding region 2.
          \param colBound2 The lower right X bound of the bounding box surrounding region 2.
          \param rowBound2 The lower right Y bound of the bounding box surrounding region 21.          
          \param id1 Region 1 ID.
          \param id2 Region 2 ID.
          \param touchingEdgeLength The touching edge length.
          \note The internal m_segmentsIdsMatrix will be used.
        */
        void getTouchingEdgeLength( 
          const unsigned int& colStart1, 
          const unsigned int& rowStart1,
          const unsigned int& colBound1, 
          const unsigned int& rowBound1,
          const unsigned int& colStart2, 
          const unsigned int& rowStart2,
          const unsigned int& colBound2, 
          const unsigned int& rowBound2,          
          const SegmenterSegmentsBlock::SegmentIdDataType& id1,
          const SegmenterSegmentsBlock::SegmentIdDataType& id2,
          unsigned int& touchingEdgeLength ) const;
    };
  } // end namespace rp
}   // end namespace te

#endif  // __TERRALIB_RP_INTERNAL_SEGMENTERREGIONGROWINGBAATZMERGER_H
