/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/rp/SpectralResponseFunctions.cpp
  \brief Spectral Response Functions.
*/

#include "SpectralResponseFunctions.h"

#define srfFormatConvert( pointerToMatrix, map ) \
  { \
    map.clear(); \
    const unsigned int size = sizeof( pointerToMatrix ) / ( 2 * sizeof( double ) ); \
    for( unsigned int idx = 0 ; idx < size ; ++idx ) \
    { \
      map[ pointerToMatrix[ idx ][ 0 ] ] = std::max( 0.0, std::min( 1.0, pointerToMatrix[ idx ][ 1 ] ) ); \
    } \
  }
  
namespace te
{
  namespace rp
  {
    namespace srf
    {
      namespace 
      {
        const std::map< SensorType, std::pair< std::string, std::map< double, double > > >& 
          srfs = getSRFs();
      }
      
      const std::map< SensorType, std::pair< std::string, std::map< double, double > > >& getSRFs()
      {
        static std::map< SensorType, std::pair< std::string, std::map< double, double > > > srfs;
        
        if( srfs.empty() )
        {
          std::pair< std::string, std::map< double, double > > srfsPair;
          
          srfsPair.first = "CBERS2BCCDB1Sensor";
          srfFormatConvert( CBERS2B_CCDB1_SRF, srfsPair.second );
          srfs[ CBERS2BCCDB1Sensor ] = srfsPair;
          
          srfsPair.first = "CBERS2BCCDB2Sensor";
          srfFormatConvert( CBERS2B_CCDB2_SRF, srfsPair.second );
          srfs[ CBERS2BCCDB2Sensor ] = srfsPair;

          srfsPair.first = "CBERS2BCCDB3Sensor";
          srfFormatConvert( CBERS2B_CCDB3_SRF, srfsPair.second );
          srfs[ CBERS2BCCDB3Sensor ] = srfsPair;

          srfsPair.first = "CBERS2BCCDB4Sensor";
          srfFormatConvert( CBERS2B_CCDB4_SRF, srfsPair.second );
          srfs[ CBERS2BCCDB4Sensor ] = srfsPair;

          srfsPair.first = "CBERS2BCCDB5PANSensor";
          srfFormatConvert( CBERS2B_CCDB5PAN_SRF, srfsPair.second );
          srfs[ CBERS2BCCDB5PANSensor ] = srfsPair;

          srfsPair.first = "CBERS4MUXB5Sensor";
          srfFormatConvert( CBERS4_MUXB5_SRF, srfsPair.second );
          srfs[ CBERS4MUXB5Sensor ] = srfsPair;

          srfsPair.first = "CBERS4MUXB6Sensor";
          srfFormatConvert( CBERS4_MUXB6_SRF, srfsPair.second );
          srfs[ CBERS4MUXB6Sensor ] = srfsPair;        
          
          srfsPair.first = "CBERS4MUXB7Sensor";
          srfFormatConvert( CBERS4_MUXB7_SRF, srfsPair.second );
          srfs[ CBERS4MUXB7Sensor ] = srfsPair;
          
          srfsPair.first = "CBERS4MUXB8Sensor";
          srfFormatConvert( CBERS4_MUXB8_SRF, srfsPair.second );
          srfs[ CBERS4MUXB8Sensor ] = srfsPair;

          srfsPair.first = "CBERS04AWPMPANSensor";
          srfFormatConvert( CBERS04A_WPMPAN_SRF, srfsPair.second );
          srfs[ CBERS04AWPMPANSensor ] = srfsPair;

          srfsPair.first = "CBERS04AWPMB1Sensor";
          srfFormatConvert( CBERS04A_WPMB1_SRF, srfsPair.second );
          srfs[ CBERS04AWPMB1Sensor ] = srfsPair;

          srfsPair.first = "CBERS04AWPMB2Sensor";
          srfFormatConvert( CBERS04A_WPMB2_SRF, srfsPair.second );
          srfs[ CBERS04AWPMB2Sensor ] = srfsPair;

          srfsPair.first = "CBERS04AWPMB3Sensor";
          srfFormatConvert( CBERS04A_WPMB3_SRF, srfsPair.second );
          srfs[ CBERS04AWPMB3Sensor ] = srfsPair;

          srfsPair.first = "CBERS04AWPMB4Sensor";
          srfFormatConvert( CBERS04A_WPMB4_SRF, srfsPair.second );
          srfs[ CBERS04AWPMB4Sensor ] = srfsPair;

          srfsPair.first = "LANDSAT8OLIB1Sensor";
          srfFormatConvert( LANDSAT8_OLI1_SRF, srfsPair.second );
          srfs[ LANDSAT8OLIB1Sensor ] = srfsPair;

          srfsPair.first = "LANDSAT8OLIB2Sensor";
          srfFormatConvert( LANDSAT8_OLI2_SRF, srfsPair.second );
          srfs[ LANDSAT8OLIB2Sensor ] = srfsPair;

          srfsPair.first = "LANDSAT8OLIB3Sensor";
          srfFormatConvert( LANDSAT8_OLI3_SRF, srfsPair.second );
          srfs[ LANDSAT8OLIB3Sensor ] = srfsPair;        
          
          srfsPair.first = "LANDSAT8OLIB4Sensor";
          srfFormatConvert( LANDSAT8_OLI4_SRF, srfsPair.second );
          srfs[ LANDSAT8OLIB4Sensor ] = srfsPair;
          
          srfsPair.first = "LANDSAT8OLIB5Sensor";
          srfFormatConvert( LANDSAT8_OLI5_SRF, srfsPair.second );
          srfs[ LANDSAT8OLIB5Sensor ] = srfsPair;

          srfsPair.first = "LANDSAT8OLIB6Sensor";
          srfFormatConvert( LANDSAT8_OLI6_SRF, srfsPair.second );
          srfs[ LANDSAT8OLIB6Sensor ] = srfsPair;

          srfsPair.first = "LANDSAT8OLIB7Sensor";
          srfFormatConvert( LANDSAT8_OLI7_SRF, srfsPair.second );
          srfs[ LANDSAT8OLIB7Sensor ] = srfsPair;

          srfsPair.first = "LANDSAT8OLIB8Sensor";
          srfFormatConvert( LANDSAT8_OLI8_SRF, srfsPair.second );
          srfs[ LANDSAT8OLIB8Sensor ] = srfsPair;

          srfsPair.first = "LANDSAT8OLIB9Sensor";
          srfFormatConvert( LANDSAT8_OLI9_SRF, srfsPair.second );
          srfs[ LANDSAT8OLIB9Sensor ] = srfsPair;

          srfsPair.first = "LANDSAT8OLIB10Sensor";
          srfFormatConvert( LANDSAT8_OLI10_SRF, srfsPair.second );
          srfs[ LANDSAT8OLIB10Sensor ] = srfsPair;
          
          srfsPair.first = "LANDSAT8OLIB11Sensor";
          srfFormatConvert( LANDSAT8_OLI11_SRF, srfsPair.second );
          srfs[ LANDSAT8OLIB11Sensor ] = srfsPair;

          srfsPair.first = "LANDSAT7ETMB1Sensor";
          srfFormatConvert( LANDSAT7_ETM1_SRF, srfsPair.second );
          srfs[ LANDSAT7ETMB1Sensor ] = srfsPair;        
          
          srfsPair.first = "LANDSAT7ETMB2Sensor";
          srfFormatConvert( LANDSAT7_ETM2_SRF, srfsPair.second );
          srfs[ LANDSAT7ETMB2Sensor ] = srfsPair;
          
          srfsPair.first = "LANDSAT7ETMB3Sensor";
          srfFormatConvert( LANDSAT7_ETM3_SRF, srfsPair.second );
          srfs[ LANDSAT7ETMB3Sensor ] = srfsPair;

          srfsPair.first = "LANDSAT7ETMB4Sensor";
          srfFormatConvert( LANDSAT7_ETM4_SRF, srfsPair.second );
          srfs[ LANDSAT7ETMB4Sensor ] = srfsPair;

          srfsPair.first = "LANDSAT7ETMB5Sensor";
          srfFormatConvert( LANDSAT7_ETM5_SRF, srfsPair.second );
          srfs[ LANDSAT7ETMB5Sensor ] = srfsPair;

          srfsPair.first = "LANDSAT7ETMB7Sensor";
          srfFormatConvert( LANDSAT7_ETM7_SRF, srfsPair.second );
          srfs[ LANDSAT7ETMB7Sensor ] = srfsPair;

          srfsPair.first = "LANDSAT7ETMB8PANSensor";
          srfFormatConvert( LANDSAT7_ETM8PAN_SRF, srfsPair.second );
          srfs[ LANDSAT7ETMB8PANSensor ] = srfsPair;
          
          srfsPair.first = "AMAZONIA1WFIROB1Sensor";
          srfFormatConvert( AMAZONIA1_WFI_RO_B1_SRF, srfsPair.second );
          srfs[ AMAZONIA1WFIROB1Sensor ] = srfsPair;

          srfsPair.first = "AMAZONIA1WFIROB2Sensor";
          srfFormatConvert( AMAZONIA1_WFI_RO_B2_SRF, srfsPair.second );
          srfs[ AMAZONIA1WFIROB2Sensor ] = srfsPair;

          srfsPair.first = "AMAZONIA1WFIROB3Sensor";
          srfFormatConvert( AMAZONIA1_WFI_RO_B3_SRF, srfsPair.second );
          srfs[ AMAZONIA1WFIROB3Sensor ] = srfsPair;

          srfsPair.first = "AMAZONIA1WFIROB4Sensor";
          srfFormatConvert( AMAZONIA1_WFI_RO_B4_SRF, srfsPair.second );
          srfs[ AMAZONIA1WFIROB4Sensor ] = srfsPair;

          srfsPair.first = "AMAZONIA1WFILOB1Sensor";
          srfFormatConvert( AMAZONIA1_WFI_LO_B1_SRF, srfsPair.second );
          srfs[ AMAZONIA1WFILOB1Sensor ] = srfsPair;

          srfsPair.first = "AMAZONIA1WFILOB2Sensor";
          srfFormatConvert( AMAZONIA1_WFI_LO_B2_SRF, srfsPair.second );
          srfs[ AMAZONIA1WFILOB2Sensor ] = srfsPair;

          srfsPair.first = "AMAZONIA1WFILOB3Sensor";
          srfFormatConvert( AMAZONIA1_WFI_LO_B3_SRF, srfsPair.second );
          srfs[ AMAZONIA1WFILOB3Sensor ] = srfsPair;

          srfsPair.first = "AMAZONIA1WFILOB4Sensor";
          srfFormatConvert( AMAZONIA1_WFI_LO_B4_SRF, srfsPair.second );
          srfs[ AMAZONIA1WFILOB4Sensor ] = srfsPair;          
          
          srfsPair.first = "CBERS04AMUXB5Sensor";
          srfFormatConvert( CBERS04A_MUXB5_SRF, srfsPair.second );
          srfs[ CBERS04AMUXB5Sensor ] = srfsPair;

          srfsPair.first = "CBERS04AMUXB6Sensor";
          srfFormatConvert( CBERS04A_MUXB6_SRF, srfsPair.second );
          srfs[ CBERS04AMUXB6Sensor ] = srfsPair;          

          srfsPair.first = "CBERS04AMUXB7Sensor";
          srfFormatConvert( CBERS04A_MUXB7_SRF, srfsPair.second );
          srfs[ CBERS04AMUXB7Sensor ] = srfsPair;          

          srfsPair.first = "CBERS04AMUXB8Sensor";
          srfFormatConvert( CBERS04A_MUXB8_SRF, srfsPair.second );
          srfs[ CBERS04AMUXB8Sensor ] = srfsPair;                    
        }
        
        return srfs;
      }
      
      void getSRF( const SensorType& sensor, std::map< double, double >& container )
      {
        container.clear();
        
        const std::map< SensorType, std::pair< std::string, 
          std::map< double, double > > >& srfs = getSRFs();
        
        std::map< SensorType, std::pair< std::string, 
          std::map< double, double > > >::const_iterator srfsIt = 
          srfs.find( sensor );
          
        if( srfsIt != srfs.end() )
        {
          container = srfsIt->second.second;
        }
      }
      
      double interpolateSRF( const std::map< double, double >& sRFs, 
        const double& frequency )
      {
        std::map< double, double >::const_iterator it2 = sRFs.find( frequency );
        
        if( it2 == sRFs.end() )
        {
          it2 = sRFs.upper_bound( frequency );
          
          if( ( it2 == sRFs.begin() ) || ( it2 == sRFs.end() ) )
          {
            return 0.0;
          }
          else
          {
            std::map< double, double >::const_iterator it1 = it2;
            --it1;
            
            double dist1 = frequency - it1->first;
            assert( dist1 > 0.0 );          
            double dist2 = it2->first - frequency;
            assert( dist2 > 0.0 );
            
            return ( 
                      ( it1->second * dist2 ) 
                      + 
                      ( it2->second * dist1 )
                   ) 
                   / 
                   ( dist1  + dist2 );
          }
        }
        else
        {
          return it2->second;
        }
      }
      
      double getSRFArea( const std::map< double, double >& sRFs )
      {
        double returnValue = 0.0;
        
        std::map< double, double >::const_iterator it = sRFs.begin();
        std::map< double, double >::const_iterator itPrev;
        const std::map< double, double >::const_iterator itEnd = sRFs.end();      
        
        while( it != itEnd )
        {
          if( it == sRFs.begin() )
          {
            itPrev = it;
          }
          else
          {
            returnValue += std::abs( it->second - itPrev->second ) * 
              ( it->first - itPrev->first ) / 2.0;
            returnValue += ( it->first - itPrev->first ) *
              std::min( it->second, itPrev->second );
            
            ++itPrev;
          }
          
          ++it;
        }
        
        return returnValue;
      }
      
      void getIntersectionSRF( const std::map< double, double >& sRF1, 
        const std::map< double, double >& sRF2,
        std::map< double, double >& intersectionSRF )
      {
        getUnionSRF( sRF1, sRF2, intersectionSRF );
        
        std::map< double, double >::const_iterator it = intersectionSRF.begin();
        const std::map< double, double >::const_iterator itEnd = intersectionSRF.end();
        
        while( it != itEnd )
        {
          intersectionSRF[ it->first ] =
            std::min(
              interpolateSRF( sRF1, it->first ),
              interpolateSRF( sRF2, it->first )
            );
            
          ++it;
        }
      }
      
      void getUnionSRF( const std::map< double, double >& sRF1, 
        const std::map< double, double >& sRF2,
        std::map< double, double >& unionSRF )
      {
        unionSRF = sRF2;
        
        std::map< double, double >::const_iterator it1 = sRF1.begin();
        const std::map< double, double >::const_iterator it1End = sRF1.end();
        
        while( it1 != it1End )
        {
          unionSRF[ it1->first ] = 
            std::max( 
              interpolateSRF( unionSRF, it1->first ),
              it1->second
            );
            
          ++it1;
        }
      }
      
    }
  }
}   // end namespace te

