/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/rp/TiePointsLocatorInputParameters.cpp
  \brief Tie points locator input parameters.
*/

#include "TiePointsLocatorInputParameters.h"
#include "TiePointsLocatorMoravecStrategy.h"
#include "AlgorithmParametersSerializer.h"

namespace te
{
  namespace rp
  {
    TiePointsLocatorInputParameters::TiePointsLocatorInputParameters()
    {
      reset();
    }
    
    TiePointsLocatorInputParameters::TiePointsLocatorInputParameters( const TiePointsLocatorInputParameters& other )
    {
      reset();
      operator=( other );
    }    

    TiePointsLocatorInputParameters::~TiePointsLocatorInputParameters()
    {
      reset();
    }

    void TiePointsLocatorInputParameters::reset() 
    {
      m_interesPointsLocationStrategyName = "Moravec";
      m_inRaster1Ptr = nullptr;
      m_inMaskRaster1Ptr = nullptr;
      m_inRaster1Bands.clear();
      m_raster1TargetAreaLineStart = 0;
      m_raster1TargetAreaColStart = 0;
      m_raster1TargetAreaWidth = 0;
      m_raster1TargetAreaHeight = 0;
      m_inRaster2Ptr = nullptr;
      m_inMaskRaster2Ptr = nullptr;
      m_inRaster2Bands.clear();
      m_raster2TargetAreaLineStart = 0;
      m_raster2TargetAreaColStart = 0;
      m_raster2TargetAreaWidth = 0;
      m_raster2TargetAreaHeight = 0;
      m_enableMultiThread = true;
      m_enableProgress = false;
      m_maxTiePoints = 2500;
      m_pixelSizeXRelation = 1;
      m_pixelSizeYRelation = 1;
      m_geomTransfName = "Affine";
      m_geomTransfMaxError = 2;
      m_enableGeometryFilter = true;
      m_subSampleOptimizationRescaleFactor = 1.0;
      m_subSampleOptimizationMinTPAreaCoverage = 25;
      m_interpMethod = te::rst::NearestNeighbor;
      m_tiePointsSubSectorsSplitFactor = 3;
      
      m_specStratParamsPtr.reset( new te::rp::TiePointsLocatorMoravecStrategy::Parameters() );
    }

    const TiePointsLocatorInputParameters& TiePointsLocatorInputParameters::operator=(
      const TiePointsLocatorInputParameters& params )
    {
      reset();
      
      m_interesPointsLocationStrategyName = params.m_interesPointsLocationStrategyName;
      m_inRaster1Ptr = params.m_inRaster1Ptr;
      m_inMaskRaster1Ptr = params.m_inMaskRaster1Ptr;
      m_inRaster1Bands = params.m_inRaster1Bands;
      m_raster1TargetAreaLineStart = params.m_raster1TargetAreaLineStart;
      m_raster1TargetAreaColStart = params.m_raster1TargetAreaColStart;
      m_raster1TargetAreaWidth = params.m_raster1TargetAreaWidth;
      m_raster1TargetAreaHeight = params.m_raster1TargetAreaHeight;
      m_inRaster2Ptr = params.m_inRaster2Ptr;
      m_inMaskRaster2Ptr = params.m_inMaskRaster2Ptr;
      m_inRaster2Bands = params.m_inRaster2Bands;
      m_raster2TargetAreaLineStart = params.m_raster2TargetAreaLineStart;
      m_raster2TargetAreaColStart = params.m_raster2TargetAreaColStart;
      m_raster2TargetAreaWidth = params.m_raster2TargetAreaWidth;
      m_raster2TargetAreaHeight = params.m_raster2TargetAreaHeight;
      m_enableMultiThread = params.m_enableMultiThread;
      m_enableProgress = params.m_enableProgress;
      m_maxTiePoints = params.m_maxTiePoints;
      m_pixelSizeXRelation = params.m_pixelSizeXRelation;
      m_pixelSizeYRelation = params.m_pixelSizeYRelation;
      m_geomTransfName = params.m_geomTransfName;
      m_geomTransfMaxError = params.m_geomTransfMaxError;
      m_enableGeometryFilter = params.m_enableGeometryFilter;
      m_subSampleOptimizationRescaleFactor = params.m_subSampleOptimizationRescaleFactor;
      m_subSampleOptimizationMinTPAreaCoverage = params.m_subSampleOptimizationMinTPAreaCoverage;
      m_interpMethod = params.m_interpMethod;
      m_tiePointsSubSectorsSplitFactor = params.m_tiePointsSubSectorsSplitFactor;
      
      if( params.m_specStratParamsPtr.get() )
      {
        m_specStratParamsPtr.reset( (TiePointsLocatorStrategyParameters*) 
          params.m_specStratParamsPtr->clone() );
      }
      else
      {
        m_specStratParamsPtr.reset();
      }

      return *this;
    }

    te::common::AbstractParameters* TiePointsLocatorInputParameters::clone() const
    {
      return new TiePointsLocatorInputParameters( *this );
    }
    
    bool TiePointsLocatorInputParameters::serialize( AlgorithmParametersSerializer& serializer ) const
    {
      TERP_TRUE_OR_RETURN_FALSE( AlgorithmInputParameters::serialize( serializer ),
         "Serializarion error" );

      if( m_specStratParamsPtr.get() )
      {
        TERP_TRUE_OR_RETURN_FALSE( m_specStratParamsPtr->serialize(
          serializer ), "Strategy serialization error" );
      }

      serializer.setAlgorithmName( "TiePointsLocator" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "interesPointsLocationStrategyName",
       m_interesPointsLocationStrategyName ), "Error adding parameter interesPointsLocationStrategyName" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "inRaster1",
        m_inRaster1Ptr->getName() ), "Error adding parameter inRaster1" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "inMaskRaster1",
        m_inMaskRaster1Ptr->getName() ), "Error adding parameter inMaskRaster1" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addMultivaluedParameter( "inRaster1Bands",
        m_inRaster1Bands ), "Error adding parameter inRaster1Bands" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "raster1TargetAreaLineStart",
        m_raster1TargetAreaLineStart ), "Error adding parameter raster1TargetAreaLineStart" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "raster1TargetAreaColStart",
        m_raster1TargetAreaColStart ), "Error adding parameter raster1TargetAreaColStart" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "m_raster1TargetAreaWidth",
        m_raster1TargetAreaWidth ), "Error adding parameter m_raster1TargetAreaWidth" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "raster1TargetAreaHeight",
        m_raster1TargetAreaHeight ), "Error adding parameter raster1TargetAreaHeight" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "inRaster2",
        m_inRaster2Ptr->getName() ), "Error adding parameter inRaster2" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "inMaskRaster2",
        m_inMaskRaster2Ptr->getName() ), "Error adding parameter inMaskRaster2" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addMultivaluedParameter( "inRaster2Bands",
        m_inRaster2Bands ), "Error adding parameter inRaster2Bands" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "raster2TargetAreaLineStart",
        m_raster2TargetAreaLineStart ), "Error adding parameter raster2TargetAreaLineStart" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "raster2TargetAreaColStart",
        m_raster2TargetAreaColStart ), "Error adding parameter raster2TargetAreaColStart" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "m_raster2TargetAreaWidth",
        m_raster2TargetAreaWidth ), "Error adding parameter m_raster2TargetAreaWidth" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "raster2TargetAreaHeight",
        m_raster2TargetAreaHeight ), "Error adding parameter raster2TargetAreaHeight" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "maxTiePoints",
       m_maxTiePoints ), "Error adding parameter maxTiePoints" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "pixelSizeXRelation",
        m_pixelSizeXRelation ), "Error adding parameter pixelSizeXRelation" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "pixelSizeYRelation",
        m_pixelSizeYRelation ), "Error adding parameter pixelSizeYRelation" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "geomTransfName",
        m_geomTransfName ), "Error adding parameter geomTransfName" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "geomTransfMaxError",
        m_geomTransfMaxError ), "Error adding parameter geomTransfMaxError" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "enableGeometryFilter",
        m_enableGeometryFilter ), "Error adding parameter enableGeometryFilter" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "subSampleOptimizationRescaleFactor",
        m_subSampleOptimizationRescaleFactor ), "Error adding parameter subSampleOptimizationRescaleFactor" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "subSampleOptimizationMinTPAreaCoverage",
        m_subSampleOptimizationMinTPAreaCoverage ), "Error adding parameter subSampleOptimizationMinTPAreaCoverage" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "interpMethod",
        m_interpMethod ), "Error adding parameter interpMethod" );

      TERP_TRUE_OR_RETURN_FALSE( serializer.addParameter( "tiePointsSubSectorsSplitFactor",
        m_tiePointsSubSectorsSplitFactor ), "Error adding parameter tiePointsSubSectorsSplitFactor" );

      return true;
    }

    void TiePointsLocatorInputParameters::setSpecStrategyParams( 
      const TiePointsLocatorStrategyParameters& specStratParams )
    {
      m_specStratParamsPtr.reset( (TiePointsLocatorStrategyParameters*)
        specStratParams.clone() );
    }

    TiePointsLocatorStrategyParameters const* TiePointsLocatorInputParameters::getSpecStrategyParams() const
    {
      return m_specStratParamsPtr.get();      
    }
    
    TiePointsLocatorStrategyParameters* TiePointsLocatorInputParameters::getSpecStrategyParams()
    {
      return m_specStratParamsPtr.get();      
    }    

  } // end namespace rp
}   // end namespace te

