/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

This file is part of the TerraLib - a Framework for building GIS enabled applications.

TerraLib is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

TerraLib is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with TerraLib. See COPYING. If not, write to
TerraLib Team at <terralib-team@terralib.org>.
*/

/*!
\file terralib/rp/radar/GaussianModel.cpp
\brief Gaussian model data manipulation.
*/

// TerraLib includes
#include "GaussianModel.h"
#include "../../raster/PositionIterator.h"
#include "../../common/MatrixUtils.h"
#include "../Macros.h"

#include <fstream>

namespace te {
	namespace rp {
		namespace radar {

			GaussianModel::GaussianModel() {
			}

			GaussianModel::~GaussianModel() {
			}

			// Checking covariance matrix
			boost::numeric::ublas::matrix<double> GaussianModel::checkCovarianceMatrix(boost::numeric::ublas::matrix<double> covMatrix) {

				double value = 0.00000000001;;
				bool status = false;

				// getting the first non zero value in the main diagonal
				for (unsigned int i = 0; i < covMatrix.size1(); ++i) {
					if (covMatrix(i, i) != 0) {
						value = covMatrix(i, i);
						break;
					}
				}

				// getting the minimum non zero value in the main diagonal
				for (unsigned int i = 0; i < covMatrix.size1(); ++i) {
					if ((covMatrix(i, i) < value) && (covMatrix(i, i) != 0))
						value = covMatrix(i, i);
				}

				// finding zero values and changing them in the main diagonal
				for (unsigned int i = 0; i < covMatrix.size1(); ++i) {
					if (covMatrix(i, i) == 0) {
						covMatrix(i, i) = value;
						status = true;
					}
				}

				// setting a very small value in the main diagonal:
				//   if all its values are equal to zero or
				//   if there are linearly dependent columns
				if (!status) {
					for (unsigned int i = 0; i < covMatrix.size1(); ++i) {
						covMatrix(i, i) += 0.00000000001;
						status = true;
					}
				}

				return covMatrix;
			}

			// Loading samples
			bool GaussianModel::LoadingSamples() {

				/*     MClassesSamplesCTPtr allClassesSamples(new MClassesSamplesCT());
				ClassSamplesContainerT classSamples;
				ClassSampleT samples;
				std::cout << "m_polygonSamplesT.size(): " << m_polygonSamplesT.size() << std::endl;

				std::ofstream outFileS ("../data/geotiff/Results/samplesPixels.txt");

				for (unsigned int npol = 0; npol < m_polygonSamplesT.size(); ++npol) {
				te::gm::Polygon* polygon = &m_polygonSamplesT[npol];

				// create iterators for band and polygon
				te::rst::PolygonIterator< double > it = te::rst::PolygonIterator< double >::
				begin(m_inputRasterPtr, polygon);
				const te::rst::PolygonIterator< double > itend = te::rst::PolygonIterator< double >::
				end(m_inputRasterPtr, polygon);

				int numPixels = 0;

				while (it != itend) {

				unsigned int column = it.getColumn();
				unsigned int row = it.getRow();

				if (((row % (m_lagY + 1)) == 0) && ((column % (m_lagX + 1)) == 0)) {  // De-correlating samples input

				for (unsigned int band = 0; band < m_inputRastersBands.size(); ++band)
				{
				samples.push_back((*it)[band]);
				//if (m_classesIDT[npol] == 0)
				outFileS << (*it)[band] << " ";
				}
				//if (m_classesIDT[npol] == 0)
				outFileS << std::endl;
				classSamples.push_back(samples);
				samples.clear();
				}

				++it;
				++numPixels;
				}

				std::cout << "classe : " << m_classesIDT[npol] << " - " << numPixels << std::endl;
				if((npol + 1) == m_polygonSamplesT.size()) {
				allClassesSamples->insert(MClassesSamplesCT::value_type(m_classesIDT[npol], classSamples));
				continue;
				}

				else if(m_classesIDT[npol] != m_classesIDT[npol + 1]) {
				allClassesSamples->insert(MClassesSamplesCT::value_type(m_classesIDT[npol], classSamples));
				classSamples.clear();
				}
				}
				outFileS.close();
				m_trainSamplesPtr = allClassesSamples;

				TERP_TRUE_OR_RETURN_FALSE(m_trainSamplesPtr,
				"Invalid Gaussian classes samples pointer")
				TERP_TRUE_OR_RETURN_FALSE(m_trainSamplesPtr->size() > 0,
				"Invalid Gaussian classes samples number")
				TERP_TRUE_OR_RETURN_FALSE(m_trainSamplesPtr->begin()->second.size() > 0,
				"Invalid Gaussian classes samples number")*/

				MClassesSamplesCTPtr allClassesSamples(new MClassesSamplesCT());
				ClassSamplesContainerT classSamples;
				ClassSampleT samples;

				ClassSamplesContainerT::const_iterator samplesIt =
					m_polygonSamplesTNew.begin();
				const ClassSamplesContainerT::const_iterator samplesItE =
					m_polygonSamplesTNew.end();

				int classID = 0;

				while (samplesIt != samplesItE) {
					for (int i = 0; i < m_numberOfSamples[classID]; ++i) {

						int row = (int)(samplesIt->operator[](i) / m_inputRasterPtr->getNumberOfColumns());
						int column = ((int)samplesIt->operator[](i) % m_inputRasterPtr->getNumberOfColumns());

						if (((row % (m_lagY + 1)) == 0) && ((column % (m_lagX + 1)) == 0)) {  // De-correlating samples input
							double value = 0.0;
							for (unsigned int band = 0; band < m_inputRastersBands.size(); ++band) {
								m_inputRasterPtr->getBand(band)->getValue(column, row, value);
								samples.push_back(value);
							}
							classSamples.push_back(samples);
							samples.clear();
						}
					}

					allClassesSamples->insert(MClassesSamplesCT::value_type(classID, classSamples));
					classSamples.clear();

					++classID;
					++samplesIt;
				}

				m_trainSamplesPtr = allClassesSamples;

				TERP_TRUE_OR_RETURN_FALSE(m_trainSamplesPtr,
					"Invalid Gaussian classes samples pointer")
					TERP_TRUE_OR_RETURN_FALSE(m_trainSamplesPtr->size() > 0,
					"Invalid Gaussian classes samples number")
					TERP_TRUE_OR_RETURN_FALSE(m_trainSamplesPtr->begin()->second.size() > 0,
					"Invalid Gaussian classes samples number")

					return true;
			}

			// Calculating Classes Mu parameter
			bool GaussianModel::CalculatingMuClass() {

				const unsigned int dimsNumber = (unsigned int)
					m_trainSamplesPtr->begin()->second.operator[](0).size();
				TERP_TRUE_OR_RETURN_FALSE(dimsNumber > 0, "Invalid Gaussian classes band number");
//				std::cout << "          Number of bands: " << dimsNumber << std::endl;  // --->>>>> TESTE DO NUMERO DE CLASSES
				MClassesSamplesCT::const_iterator classesIt =
					m_trainSamplesPtr->begin();
				const MClassesSamplesCT::const_iterator classesItE =
					m_trainSamplesPtr->end();
				int classesNum = 0;  // --->>>>> TESTE DO NUMERO DE CLASSES

				while (classesIt != classesItE) {
					classesNum++;  // --->>>>> TESTE DO NUMERO DE CLASSES
					const ClassIDT& classID = classesIt->first;

					const ClassSamplesContainerT& classSamples = classesIt->second;
					TERP_TRUE_OR_RETURN_FALSE(classSamples.size() > 0,
						"Invalid class samples number");

					ClassSampleT classMeans(dimsNumber, 0.0);

					unsigned int dimIdx = 0;

					for (dimIdx = 0; dimIdx < dimsNumber; ++dimIdx) {
						double& dimMean = classMeans[dimIdx];
						dimMean = 0.0;

						ClassSamplesContainerT::const_iterator samplesIt =
							classSamples.begin();
						const ClassSamplesContainerT::const_iterator samplesItE =
							classSamples.end();

						while (samplesIt != samplesItE) {
							TERP_TRUE_OR_RETURN_FALSE(samplesIt->size() == dimsNumber,
								"Sample size mismatch")
								dimMean += samplesIt->operator[](dimIdx);

							++samplesIt;
						}

						dimMean /= static_cast<double>(classSamples.size());
					}

					m_muClassesParameters.push_back(classMeans);
					m_classesIndex2IDMu.push_back(classID);
					m_samplesSize.push_back(classSamples.size());

					++classesIt;
				}

				return true;
			}

			// Calculating Classes Sigma parameter
			bool GaussianModel::CalculatingSigmaClass() {

				const unsigned int dimsNumber = (unsigned int)
					m_trainSamplesPtr->begin()->second.operator[](0).size();
				TERP_TRUE_OR_RETURN_FALSE(dimsNumber > 0, "Invalid Gaussian classes band number");

				MClassesSamplesCT::const_iterator classesIt =
					m_trainSamplesPtr->begin();
				const MClassesSamplesCT::const_iterator classesItE =
					m_trainSamplesPtr->end();
				unsigned int classIdx = 0;

				while (classesIt != classesItE) {
					const ClassSamplesContainerT& classSamples = classesIt->second;
					TERP_TRUE_OR_RETURN_FALSE(classSamples.size() > 0,
						"Invalid class samples number");

					const std::vector< double >& classMeans = m_muClassesParameters[classIdx];

					boost::numeric::ublas::matrix< double > classCovarianceMatrix(dimsNumber, dimsNumber);

					unsigned int dimIdx1 = 0;
					unsigned int dimIdx2 = 0;

					for (dimIdx1 = 0; dimIdx1 < dimsNumber; ++dimIdx1) {
						const double& dimMean1 = classMeans[dimIdx1];

						for (dimIdx2 = 0; dimIdx2 < dimsNumber; ++dimIdx2) {
							const double& dimMean2 = classMeans[dimIdx2];

							double& covariance = classCovarianceMatrix(dimIdx1, dimIdx2);
							covariance = 0.0;

							ClassSamplesContainerT::const_iterator samplesIt =
								classSamples.begin();
							const ClassSamplesContainerT::const_iterator samplesItE =
								classSamples.end();

							while (samplesIt != samplesItE) {
								covariance +=
									(
									(samplesIt->operator[](dimIdx1) - dimMean1)
									*
									(samplesIt->operator[](dimIdx2) - dimMean2)
									);

								++samplesIt;
							}

							covariance /= static_cast<double>(classSamples.size() - 1);
						}
					}

					m_classesCovarianceMatrixes.push_back(classCovarianceMatrix);

					++classIdx;
					++classesIt;
				}

				return true;
			}

			// Calculating Regions Mu and Sigma parameters
			bool GaussianModel::CalculatingMuAndSigmaRegion() {

				TERP_TRUE_OR_RETURN_FALSE(m_polygonsSegImage.size() > 0, "GaussianModel algorithm needs polygons")

					TERP_TRUE_OR_RETURN_FALSE(m_inputRastersBands.size() > 0, "Invalid band number");

				unsigned int column = 0;
				unsigned int row = 0;

//				std::cout << "          Regions " << m_polygonsSegImage.size() << std::endl;

				std::vector< double > region;
				std::vector< std::vector< double > > regionComplete;

				for (unsigned int i = 0; i < m_polygonsSegImage.size(); i++) {
					te::gm::Polygon* polygon = m_polygonsSegImage[i];

					te::rst::PolygonIterator< double > polygonsIt = te::rst::PolygonIterator< double >::
						begin(m_inputRasterPtr, polygon);
					te::rst::PolygonIterator< double > polygonsItE = te::rst::PolygonIterator< double >::
						end(m_inputRasterPtr, polygon);

					unsigned int regionsNum = 0;
					std::vector< double > meanRegion(m_inputRastersBands.size(), 0.0);

					while (polygonsIt != polygonsItE) {

						column = polygonsIt.getColumn();
						row = polygonsIt.getRow();

						if (((row % (m_lagY + 1)) == 0) && ((column % (m_lagX + 1)) == 0)) {  // De-correlating regions

							for (unsigned int j = 0; j < m_inputRastersBands.size(); ++j) {
								region.push_back((*polygonsIt)[j]);
								meanRegion[j] += (*polygonsIt)[j];
							}

							regionComplete.push_back(region);
							region.clear();

							++regionsNum;
						}

						++polygonsIt;
					}

					m_regionsSize.push_back(regionsNum);

					for (unsigned int j = 0; j < m_inputRastersBands.size(); ++j)
						meanRegion[j] = meanRegion[j] / regionsNum;

					m_muRegionsParameter.push_back(meanRegion);

					boost::numeric::ublas::matrix< double > regionCovarianceMatrix(m_inputRastersBands.size(), m_inputRastersBands.size());
					unsigned int dimIdx1 = 0;
					unsigned int dimIdx2 = 0;

					for (dimIdx1 = 0; dimIdx1 < m_inputRastersBands.size(); ++dimIdx1) {
						const double& dimMean1 = meanRegion[dimIdx1];

						for (dimIdx2 = 0; dimIdx2 < m_inputRastersBands.size(); ++dimIdx2) {
							const double& dimMean2 = meanRegion[dimIdx2];

							double& covariance = regionCovarianceMatrix(dimIdx1, dimIdx2);
							covariance = 0.0;

							ClassSamplesContainerT::const_iterator regionsIt =
								regionComplete.begin();
							const ClassSamplesContainerT::const_iterator regionsItE =
								regionComplete.end();

							while (regionsIt != regionsItE) {
								covariance +=
									(
									(regionsIt->operator[](dimIdx1) - dimMean1)
									*
									(regionsIt->operator[](dimIdx2) - dimMean2)
									);

								++regionsIt;
							}

							covariance /= static_cast<double>(regionComplete.size() - 1);
						}
					}

					m_regionsCovarianceMatrixes.push_back(regionCovarianceMatrix);
					meanRegion.clear();
					regionComplete.clear();
				}

				return true;
			}

			bool GaussianModel::GettingStochasticDistances() {

				// Calculating Stochastic Distances

				ClassSamplesContainerT::const_iterator regionsIt =
					m_muRegionsParameter.begin();
				const ClassSamplesContainerT::const_iterator regionsItE =
					m_muRegionsParameter.end();

				unsigned int regionIdx = 0;
				boost::numeric::ublas::matrix< double > auxVector(m_inputRastersBands.size(), 1);
				boost::numeric::ublas::matrix< double > auxVectorTransposed(1, m_inputRastersBands.size());
				boost::numeric::ublas::matrix< double > auxMatrix;
				boost::numeric::ublas::matrix< double > auxMatrix2;
				boost::numeric::ublas::matrix< double > parcialResultMatrix;
				boost::numeric::ublas::matrix< double > sumCovarianceInvMatrix;
				std::vector< double > distances;

				while (regionsIt != regionsItE) {

					ClassSamplesContainerT::const_iterator samplesIt =
						m_muClassesParameters.begin();
					const ClassSamplesContainerT::const_iterator samplesItE =
						m_muClassesParameters.end();

					unsigned int classIdxD = 0;

					if (m_stochasticDistanceType == "Bhattacharyya") {  // Bhattacharyya distance

						while (samplesIt != samplesItE) {

							auxVector.clear();
							auxVectorTransposed.clear();
							auxMatrix.clear();
							auxMatrix2.clear();
							parcialResultMatrix.clear();
							sumCovarianceInvMatrix.clear();
							double classCovarianceMatrixDet = 0.0;
							double regionCovarianceMatrixDet = 0.0;
							double sumCovarianceMatrixDet = 0.0;
							double parcialResult1 = 0.0;
							double parcialResult2 = 0.0;

							for (unsigned int i = 0; i < m_inputRastersBands.size(); ++i) {
								auxVector(i, 0) = regionsIt->operator[](i) - samplesIt->operator[](i);
								auxVectorTransposed(0, i) = regionsIt->operator[](i) - samplesIt->operator[](i);
							}

							auxMatrix = (m_regionsCovarianceMatrixes[regionIdx] + m_classesCovarianceMatrixes[classIdxD]) / 2.0;

							TERP_TRUE_OR_RETURN_FALSE(te::common::GetInverseMatrix(auxMatrix,
								sumCovarianceInvMatrix), "Inverse sum matrix calcule error");

							auxMatrix2 = boost::numeric::ublas::prod(sumCovarianceInvMatrix, auxVector);

							parcialResultMatrix = boost::numeric::ublas::prod(auxVectorTransposed, auxMatrix2);
							parcialResult1 = parcialResultMatrix(0, 0) / 8.0;

							auxMatrix2.clear();

							if (!te::common::GetDeterminant(m_classesCovarianceMatrixes[classIdxD], classCovarianceMatrixDet))
								m_classesCovarianceMatrixes[classIdxD] = checkCovarianceMatrix(m_classesCovarianceMatrixes[classIdxD]);
							te::common::GetDeterminant(m_classesCovarianceMatrixes[classIdxD], classCovarianceMatrixDet);

							if (!te::common::GetDeterminant(m_regionsCovarianceMatrixes[regionIdx], regionCovarianceMatrixDet))
								m_regionsCovarianceMatrixes[regionIdx] = checkCovarianceMatrix(m_regionsCovarianceMatrixes[regionIdx]);
							te::common::GetDeterminant(m_regionsCovarianceMatrixes[regionIdx], regionCovarianceMatrixDet);

							if (!te::common::GetDeterminant(auxMatrix, sumCovarianceMatrixDet))
								auxMatrix = checkCovarianceMatrix(auxMatrix);
							te::common::GetDeterminant(auxMatrix, sumCovarianceMatrixDet);

							parcialResult2 = (double)(log((double)sumCovarianceMatrixDet /
								(sqrt(classCovarianceMatrixDet * regionCovarianceMatrixDet)))) / 2.0;

							distances.push_back(parcialResult1 + parcialResult2);

							++samplesIt;
							++classIdxD;
						}

						m_stochasticDistance.push_back(distances);
						distances.clear();
					}

					if (m_stochasticDistanceType == "KullbackLeibler") {  // Kullback-Leibler distance

						boost::numeric::ublas::matrix< double > regionCovarianceInvMatrix(m_inputRastersBands.size(), m_inputRastersBands.size());
						boost::numeric::ublas::matrix< double > classCovarianceInvMatrix;

						while (samplesIt != samplesItE) {

							auxVector.clear();
							auxVectorTransposed.clear();
							auxMatrix.clear();
							auxMatrix2.clear();
							regionCovarianceInvMatrix.clear();
							classCovarianceInvMatrix.clear();
							double classCovarianceMatrixDet = 0.0;
							double regionCovarianceMatrixDet = 0.0;
							double parcialResult1 = 0.0;
							double parcialResult2 = 0.0;
							double matrixTrace = 0.0;
							double matrixTrace2 = 0.0;
							double matrixTrace3 = 0.0;

							for (unsigned int i = 0; i < m_inputRastersBands.size(); ++i) {
								auxVector(i, 0) = regionsIt->operator[](i) - samplesIt->operator[](i);
								auxVectorTransposed(0, i) = regionsIt->operator[](i) - samplesIt->operator[](i);
							}

							if (!te::common::GetDeterminant(m_regionsCovarianceMatrixes[regionIdx], regionCovarianceMatrixDet))
								m_regionsCovarianceMatrixes[regionIdx] = checkCovarianceMatrix(m_regionsCovarianceMatrixes[regionIdx]);

							te::common::GetDeterminant(m_regionsCovarianceMatrixes[regionIdx], regionCovarianceMatrixDet);
							if (regionCovarianceMatrixDet == 0) {
                TERP_LOG_AND_RETURN_FALSE( "Region " + std::to_string( regionIdx )
                  + " with determinant equal to 0." );
//								std::cout << m_regionsCovarianceMatrixes[regionIdx] << std::endl;
							}

							TERP_TRUE_OR_RETURN_FALSE(te::common::GetInverseMatrix(m_regionsCovarianceMatrixes[regionIdx],
								regionCovarianceInvMatrix), "Inverse region matrix calcule error");

							if (!te::common::GetDeterminant(m_classesCovarianceMatrixes[classIdxD], classCovarianceMatrixDet))
								m_classesCovarianceMatrixes[classIdxD] = checkCovarianceMatrix(m_classesCovarianceMatrixes[classIdxD]);

							TERP_TRUE_OR_RETURN_FALSE(te::common::GetInverseMatrix(m_classesCovarianceMatrixes[classIdxD],
								classCovarianceInvMatrix), "Inverse class matrix calcule error");

							auxMatrix = regionCovarianceInvMatrix + classCovarianceInvMatrix;

							auxMatrix2 = boost::numeric::ublas::prod(auxMatrix, auxVector);
							auxMatrix.clear();

							auxMatrix = boost::numeric::ublas::prod(auxVectorTransposed, auxMatrix2);
							parcialResult1 = auxMatrix(0, 0) / 2.0;

							auxMatrix.clear();
							auxMatrix2.clear();

							auxMatrix = boost::numeric::ublas::prod(m_regionsCovarianceMatrixes[regionIdx],
								classCovarianceInvMatrix);
							auxMatrix2 = boost::numeric::ublas::prod(m_classesCovarianceMatrixes[classIdxD],
								regionCovarianceInvMatrix);
							boost::numeric::ublas::identity_matrix<double> identityMatrix(m_inputRastersBands.size());

							for (unsigned int i = 0; i < m_inputRastersBands.size(); ++i) {
								matrixTrace += auxMatrix(i, i);
								matrixTrace2 += auxMatrix2(i, i);
								matrixTrace3 += identityMatrix(i, i);
							}

							parcialResult2 = (matrixTrace + matrixTrace2 - (2 * matrixTrace3)) / 2;

							distances.push_back(parcialResult1 + parcialResult2);

							++samplesIt;
							++classIdxD;
						}

						m_stochasticDistance.push_back(distances);
						distances.clear();
					}

					++regionsIt;
					++regionIdx;
				}

				return true;
			}

			// Calculating Statistical Tests

			bool GaussianModel::GettingStatisticalTests() {

				double degreesOfFreedom = (m_inputRastersBands.size() * (m_inputRastersBands.size() + 3)) / 2;

				boost::math::chi_squared distChiSquared(degreesOfFreedom);

				ClassSamplesContainerT::const_iterator regionsIt =
					m_muRegionsParameter.begin();
				const ClassSamplesContainerT::const_iterator regionsItE =
					m_muRegionsParameter.end();

				unsigned int regionIdx = 0;
				boost::numeric::ublas::matrix< double > auxVector(m_inputRastersBands.size(), 1);
				boost::numeric::ublas::matrix< double > auxVectorTransposed(1, m_inputRastersBands.size());
				boost::numeric::ublas::matrix< double > auxMatrix;
				boost::numeric::ublas::matrix< double > auxMatrix2;
				boost::numeric::ublas::matrix< double > parcialResultMatrix;
				boost::numeric::ublas::matrix< double > sumCovarianceInvMatrix;
				std::vector< double > tests;
				std::vector< double > valuesP;

				while (regionsIt != regionsItE) {

					ClassSamplesContainerT::const_iterator samples2It =
						m_muClassesParameters.begin();
					const ClassSamplesContainerT::const_iterator samples2ItE =
						m_muClassesParameters.end();

					unsigned int classIdxS = 0;

					if (m_stochasticDistanceType == "Bhattacharyya") {  // Bhattacharyya distance
						while (samples2It != samples2ItE) {

							auxVector.clear();
							auxVectorTransposed.clear();
							auxMatrix.clear();
							auxMatrix2.clear();
							parcialResultMatrix.clear();
							sumCovarianceInvMatrix.clear();
							double classCovarianceMatrixDet = 0.0;
							double regionCovarianceMatrixDet = 0.0;
							double sumCovarianceMatrixDet = 0.0;
							double parcialResult1 = 0.0;
							double parcialResult2 = 0.0;
							double parcialResult3 = 0.0;

							for (unsigned int i = 0; i < m_inputRastersBands.size(); ++i) {
								auxVector(i, 0) = regionsIt->operator[](i) - samples2It->operator[](i);
								auxVectorTransposed(0, i) = regionsIt->operator[](i) - samples2It->operator[](i);
							}

							auxMatrix = (m_regionsCovarianceMatrixes[regionIdx] + m_classesCovarianceMatrixes[classIdxS]) / 2.0;

							TERP_TRUE_OR_RETURN_FALSE(te::common::GetInverseMatrix(auxMatrix,
								sumCovarianceInvMatrix), "Inverse sum matrix calcule error");

							auxMatrix2 = boost::numeric::ublas::prod(sumCovarianceInvMatrix, auxVector);

							parcialResultMatrix = boost::numeric::ublas::prod(auxVectorTransposed, auxMatrix2);
							parcialResult1 = parcialResultMatrix(0, 0);

							auxMatrix2.clear();

							if (!te::common::GetDeterminant(m_classesCovarianceMatrixes[classIdxS], classCovarianceMatrixDet))
								m_classesCovarianceMatrixes[classIdxS] = checkCovarianceMatrix(m_classesCovarianceMatrixes[classIdxS]);
							te::common::GetDeterminant(m_classesCovarianceMatrixes[classIdxS], classCovarianceMatrixDet);

							if (!te::common::GetDeterminant(m_regionsCovarianceMatrixes[regionIdx], regionCovarianceMatrixDet))
								m_regionsCovarianceMatrixes[regionIdx] = checkCovarianceMatrix(m_regionsCovarianceMatrixes[regionIdx]);
							te::common::GetDeterminant(m_regionsCovarianceMatrixes[regionIdx], regionCovarianceMatrixDet);

							if (!te::common::GetDeterminant(auxMatrix, sumCovarianceMatrixDet))
								auxMatrix = checkCovarianceMatrix(auxMatrix);
							te::common::GetDeterminant(auxMatrix, sumCovarianceMatrixDet);

							parcialResult2 = 4.0 * (log(sumCovarianceMatrixDet /
								(sqrt(classCovarianceMatrixDet * regionCovarianceMatrixDet))));

							parcialResult3 = (m_regionsSize[regionIdx] * m_samplesSize[classIdxS])
								/
								(m_regionsSize[regionIdx] + m_samplesSize[classIdxS]);

							double result = parcialResult3 * (parcialResult1 + parcialResult2);

							tests.push_back(result);

							try {
								valuesP.push_back(boost::math::cdf(complement(distChiSquared, result)));
							}
							catch (...) {
								valuesP.push_back(NAN);
							}

							++samples2It;
							++classIdxS;
						}

						m_statisticalTest.push_back(tests);
						m_pValue.push_back(valuesP);
						tests.clear();
						valuesP.clear();
					}

					if (m_stochasticDistanceType == "KullbackLeibler") {  // Kullback-Leibler distance

						boost::numeric::ublas::matrix< double > regionCovarianceInvMatrix(m_inputRastersBands.size(), m_inputRastersBands.size());
						boost::numeric::ublas::matrix< double > classCovarianceInvMatrix;

						while (samples2It != samples2ItE) {

							auxVector.clear();
							auxVectorTransposed.clear();
							auxMatrix.clear();
							auxMatrix2.clear();
							regionCovarianceInvMatrix.clear();
							classCovarianceInvMatrix.clear();
							double classCovarianceMatrixDet = 0.0;
							double regionCovarianceMatrixDet = 0.0;
							double parcialResult1 = 0.0;
							double parcialResult2 = 0.0;
							double parcialResult3 = 0.0;
							double matrixTrace = 0.0;
							double matrixTrace2 = 0.0;
							double matrixTrace3 = 0.0;

							for (unsigned int i = 0; i < m_inputRastersBands.size(); ++i) {
								auxVector(i, 0) = regionsIt->operator[](i) - samples2It->operator[](i);
								auxVectorTransposed(0, i) = regionsIt->operator[](i) - samples2It->operator[](i);
							}

							try {
								if (!te::common::GetDeterminant(m_regionsCovarianceMatrixes[regionIdx], regionCovarianceMatrixDet))
									m_regionsCovarianceMatrixes[regionIdx] = checkCovarianceMatrix(m_regionsCovarianceMatrixes[regionIdx]);

								te::common::GetInverseMatrix(m_regionsCovarianceMatrixes[regionIdx], regionCovarianceInvMatrix);
							}
							catch (...) {
								TERP_LOG_AND_RETURN_FALSE( "Region " + std::to_string( regionIdx )
                  + " with determinant equal to 0." )
							}

							if (!te::common::GetDeterminant(m_classesCovarianceMatrixes[classIdxS], classCovarianceMatrixDet))
								m_classesCovarianceMatrixes[classIdxS] = checkCovarianceMatrix(m_classesCovarianceMatrixes[classIdxS]);

							TERP_TRUE_OR_RETURN_FALSE(te::common::GetInverseMatrix(m_classesCovarianceMatrixes[classIdxS],
								classCovarianceInvMatrix), "Inverse class matrix calcule error");

							auxMatrix = regionCovarianceInvMatrix + classCovarianceInvMatrix;

							auxMatrix2 = boost::numeric::ublas::prod(auxMatrix, auxVector);
							auxMatrix.clear();

							auxMatrix = boost::numeric::ublas::prod(auxVectorTransposed, auxMatrix2);
							parcialResult1 = auxMatrix(0, 0);

							auxMatrix.clear();
							auxMatrix2.clear();

							auxMatrix = boost::numeric::ublas::prod(m_regionsCovarianceMatrixes[regionIdx],
								classCovarianceInvMatrix);
							auxMatrix2 = boost::numeric::ublas::prod(m_classesCovarianceMatrixes[classIdxS],
								regionCovarianceInvMatrix);
							boost::numeric::ublas::identity_matrix<double> identityMatrix(m_inputRastersBands.size());

							for (unsigned int i = 0; i < m_inputRastersBands.size(); ++i) {
								matrixTrace += auxMatrix(i, i);
								matrixTrace2 += auxMatrix2(i, i);
								matrixTrace3 += identityMatrix(i, i);
							}

							parcialResult2 = (matrixTrace + matrixTrace2 - (2 * matrixTrace3));

							parcialResult3 = (m_regionsSize[regionIdx] * m_samplesSize[classIdxS])
								/
								(m_regionsSize[regionIdx] + m_samplesSize[classIdxS]);

							double result = parcialResult3 * (parcialResult1 + parcialResult2);

							tests.push_back(result);

							try {
								valuesP.push_back(boost::math::cdf(complement(distChiSquared, result)));
							}
							catch (...) {
								valuesP.push_back(NAN);
							}

							++samples2It;
							++classIdxS;
						}

						m_statisticalTest.push_back(tests);
						m_pValue.push_back(valuesP);
						tests.clear();
						valuesP.clear();
					}

					++regionsIt;
					++regionIdx;
				}

				return true;
			}

			bool GaussianModel::execute() {
				m_samplesSize.clear();
				m_regionsSize.clear();
				m_muClassesParameters.clear();
				m_classesCovarianceMatrixes.clear();
				m_muRegionsParameter.clear();
				m_regionsCovarianceMatrixes.clear();
				m_stochasticDistance.clear();
				m_statisticalTest.clear();
				m_pValue.clear();
				m_trainSamplesPtr = 0;

				if (!LoadingSamples())
				{
					TERP_LOG_AND_RETURN_FALSE( "LoadingSamples execution error." );
				}
				else
				{
//					std::cout << "LoadingSamples execution OK." << std::endl;
				}

				if (!CalculatingMuClass())
				{
					TERP_LOG_AND_RETURN_FALSE( "CalculatingMuClass execution error." );
				}
				else
				{
//					std::cout << "CalculatingMuClass execution OK." << std::endl;
				}

				if (!CalculatingSigmaClass())
				{
					TERP_LOG_AND_RETURN_FALSE( "CalculatingSigmaClass execution error." )
				}
				else
				{
//					std::cout << "CalculatingSigmaClass execution OK." << std::endl;
				}

				if (!CalculatingMuAndSigmaRegion())
				{
					TERP_LOG_AND_RETURN_FALSE( "CalculatingMuAndSigmaRegion execution error." );
				}
				else
				{
//					std::cout << "CalculatingMuAndSigmaRegion execution OK." << std::endl;
				}

				if (!GettingStatisticalTests())
				{
					TERP_LOG_AND_RETURN_FALSE( "GettingStatisticalTests execution error." );
				}
				else
				{
//					std::cout << "GettingStatisticalTests execution OK." << std::endl;
				}

				if (!GettingStochasticDistances())
				{
					TERP_LOG_AND_RETURN_FALSE( "GettingStochasticDistances execution error." );
				}
				else
				{
//					std::cout << "GettingStochasticDistances execution OK." << std::endl;
				}

				return true;
			}

		} // end namespace radar
	} // end namespace rp
}   // end namespace msc

