/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/rp/radar/GaussianModel.h
  \brief Gaussian model data manipulation.
*/

#ifndef __TERRALIB_RP_RADAR_INTERNAL_GAUSSIANMODEL_H
#define __TERRALIB_RP_RADAR_INTERNAL_GAUSSIANMODEL_H

// TerraLib includes
#include "../Config.h"
#include "../../raster/Raster.h"
#include "../../geometry/Polygon.h"

// Boost includes
#include <boost/numeric/ublas/matrix.hpp>
#include <boost/math/distributions/chi_squared.hpp>

#include <vector>
#include <map>

namespace te
{
  namespace rp
  {
    namespace radar
    {
    /*!
      \class GaussianModel

      \brief Gaussian model data manipulation.
     */
    class TERPEXPORT GaussianModel
    {
      public:

        //std::auto_ptr<te::rst::Raster> outputRasterClassesPtr;

        unsigned int m_sourcesNumber;  //!< Number of sources.

        te::rst::Raster* m_inputRasterPtr;  //!< Input raster.

        std::vector< unsigned int > m_inputRastersBands;  //!< Bands to process.

        std::vector< te::gm::Polygon* > m_polygonsSegImage;  //!< Polygons of segmented image.

        unsigned int m_lagX;  //!< The horizontal correlation value.

        unsigned int m_lagY;  //!< The vertical correlation value.

        typedef unsigned int ClassIDT;  //!< CLass ID type definiton.

        std::vector< unsigned int > m_classesIDT;  //!< Training class ID.

        std::vector< te::gm::Polygon > m_polygonSamplesT;  //!< Training samples polygons.
        std::vector< std::vector< double > > m_polygonSamplesTNew;  //!< Training samples polygons.
        std::vector< int > m_numberOfSamples;

        typedef std::vector< double > ClassSampleT;  //!< Class sample type definition.

        typedef std::vector< ClassSampleT > ClassSamplesContainerT;  //!< Class samples container type definition.

        typedef std::map< ClassIDT, ClassSamplesContainerT > MClassesSamplesCT;  //!< Multi-classes samples container type definition.

        typedef boost::shared_ptr<MClassesSamplesCT> MClassesSamplesCTPtr;  //!< A shared pointer to a multi classes samples container type definition.

        MClassesSamplesCTPtr m_trainSamplesPtr;  //!< A shared pointer to a always-valid structure where training samples are stored.

        std::vector< ClassIDT > m_classesIndex2IDMu;  //!< A class index ordered vector of classes IDs for Mu parameter.

        std::vector< std::vector< double > > m_muClassesParameters;  //!< Mu classes parameter.

        std::vector< boost::numeric::ublas::matrix< double > > m_classesCovarianceMatrixes;  //!< Sigma classes parameter.

        std::vector< std::vector< double > > m_muRegionsParameter;  //!< Mu Regions parameter.

        std::vector< boost::numeric::ublas::matrix< double > > m_regionsCovarianceMatrixes;  //!< Sigma regions parameter.

        std::string m_stochasticDistanceType;  //!< The stochastic distance type definition.

        std::vector< std::vector< double > > m_stochasticDistance; //!< Stochastic distance container.

        std::vector< std::vector< double > > m_statisticalTest; //!< Statistical test container.

        std::vector< std::vector< double > > m_pValue; //!< P-values container.

        double m_significanceLevel;   //!< The significance level.

        std::vector< double > m_regionsSize;  //!< Size of regions.

        std::vector< double > m_samplesSize;  //!< Size of sample containers.

        /*! \brief Constructor. */
        GaussianModel();

        /*! \brief Destructor. */
        ~GaussianModel();

        boost::numeric::ublas::matrix< double > checkCovarianceMatrix(boost::numeric::ublas::matrix< double >);

        bool LoadingSamples(void);

        bool CalculatingMuClass(void);

        bool CalculatingSigmaClass();

        bool CalculatingMuAndSigmaRegion(void);

        bool GettingStatisticalTests(void);

        bool GettingStochasticDistances(void);

        bool execute(void);
    };

    } // end namespace radar
  } // end namespace rp
}   // end namespace msc

#endif

