/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

This file is part of the TerraLib - a Framework for building GIS enabled applications.

TerraLib is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

TerraLib is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with TerraLib. See COPYING. If not, write to
TerraLib Team at <terralib-team@terralib.org>.
*/

/*!
\file terralib/rp/radar/IntensityPairModel.cpp
\brief Intensity Pair model data manipulation.
*/

// TerraLib includes
#include "IntensityPairModel.h"
#include "../Macros.h"
#include "../../raster/Raster.h"
#include "../../raster/Band.h"
#include "../../raster/PositionIterator.h"

// Boost
#include <boost/math/special_functions/bessel.hpp>
#include <boost/math/special_functions/gamma.hpp>

#include <fstream>

namespace te {
	namespace rp {
		namespace radar {

			IntensityPairModel::IntensityPairModel() {
			}

			IntensityPairModel::~IntensityPairModel() {
			}

			// Loading samples
			bool IntensityPairModel::LoadingSamples() {

				/*MClassesSamplesCTPtr allClassesSamples(new MClassesSamplesCT());
				ClassSamplesContainerT classSamples;
				ClassSampleT samples;

				for(unsigned int npol = 0; npol < m_polygonSamplesT.size(); npol++) {
				te::gm::Polygon* polygon = &m_polygonSamplesT[npol];

				// create iterators for band and polygon
				te::rst::PolygonIterator< double > it = te::rst::PolygonIterator< double >::
				begin(m_inputRasterPtr, polygon);
				const te::rst::PolygonIterator< double > itend = te::rst::PolygonIterator< double >::
				end(m_inputRasterPtr, polygon);

				while (it != itend) {

				unsigned int column = it.getColumn();
				unsigned int row = it.getRow();

				if (((row % (m_lagY + 1)) == 0) && ((column % (m_lagX + 1)) == 0)) {  // De-correlating samples input

				for(unsigned int band = 0; band < m_inputRastersBands.size(); band++)  //algoInputParams.m_inputRasterPtr[j]->getNumberOfBands()
				samples.push_back((*it)[band]);

				classSamples.push_back(samples);
				samples.clear();
				}

				++it;
				}

				if((npol + 1) == m_polygonSamplesT.size()) {
				allClassesSamples->insert(MClassesSamplesCT::value_type(m_classesIDT[npol], classSamples));
				continue;
				}

				else if(m_classesIDT[npol] != m_classesIDT[npol + 1]) {
				allClassesSamples->insert(MClassesSamplesCT::value_type(m_classesIDT[npol], classSamples));
				classSamples.clear();
				}
				}

				m_trainSamplesPtr = allClassesSamples;

				TERP_TRUE_OR_RETURN_FALSE(m_trainSamplesPtr,
				"Invalid Pair Intensity classes samples pointer")
				TERP_TRUE_OR_RETURN_FALSE(m_trainSamplesPtr->size() > 0,
				"Invalid Pair Intensity classes samples number")
				TERP_TRUE_OR_RETURN_FALSE(m_trainSamplesPtr->begin()->second.size() > 0,
				"Invalid Pair Intensity classes samples number")*/

				MClassesSamplesCTPtr allClassesSamples(new MClassesSamplesCT());
				ClassSamplesContainerT classSamples;
				ClassSampleT samples;

				ClassSamplesContainerT::const_iterator samplesIt =
					m_polygonSamplesTNew.begin();
				const ClassSamplesContainerT::const_iterator samplesItE =
					m_polygonSamplesTNew.end();

				int classID = 0;

				while (samplesIt != samplesItE) {
					for (int i = 0; i < m_numberOfSamples[classID]; ++i) {

						int row = (int)(samplesIt->operator[](i) / m_inputRasterPtr->getNumberOfColumns());
						int column = ((int)samplesIt->operator[](i) % m_inputRasterPtr->getNumberOfColumns());

						if (((row % (m_lagY + 1)) == 0) && ((column % (m_lagX + 1)) == 0)) {  // De-correlating samples input
							double value = 0.0;
							for (unsigned int band = 0; band < m_inputRastersBands.size(); ++band) {
								m_inputRasterPtr->getBand(band)->getValue(column, row, value);
								samples.push_back(value);
							}
							classSamples.push_back(samples);
							samples.clear();
						}
					}

					allClassesSamples->insert(MClassesSamplesCT::value_type(classID, classSamples));
					classSamples.clear();

					++classID;
					++samplesIt;
				}

				m_trainSamplesPtr = allClassesSamples;

				TERP_TRUE_OR_RETURN_FALSE(m_trainSamplesPtr,
					"Invalid Pair Intensity classes samples pointer")
					TERP_TRUE_OR_RETURN_FALSE(m_trainSamplesPtr->size() > 0,
					"Invalid Pair Intensity classes samples number")
					TERP_TRUE_OR_RETURN_FALSE(m_trainSamplesPtr->begin()->second.size() > 0,
					"Invalid Pair Intensity classes samples number")

					return true;
			}

			// Calculating Classes h11 and h22 parameters
			bool IntensityPairModel::CalculatingHClass() {

				const unsigned int dimsNumber = (unsigned int)
					m_trainSamplesPtr->begin()->second.operator[](0).size();
				TERP_TRUE_OR_RETURN_FALSE(dimsNumber == 2, "Invalid Intensity Pair classes band number");
//				std::cout << "          Number of bands: " << dimsNumber << std::endl;  // --->>>>> TESTE DO NUMERO DE CLASSES
				MClassesSamplesCT::const_iterator classesIt =
					m_trainSamplesPtr->begin();
				const MClassesSamplesCT::const_iterator classesItE =
					m_trainSamplesPtr->end();
				int classesNum = 0;  // --->>>>> TESTE DO NUMERO DE CLASSES
				while (classesIt != classesItE) {
					classesNum++;  // --->>>>> TESTE DO NUMERO DE CLASSES
					const ClassIDT& classID = classesIt->first;

					const ClassSamplesContainerT& classSamples = classesIt->second;
					TERP_TRUE_OR_RETURN_FALSE(classSamples.size() > 0,
						"Invalid class samples number");

					ClassSampleT classMeans(dimsNumber, 0.0);

					unsigned int dimIdx = 0;

					for (dimIdx = 0; dimIdx < dimsNumber; ++dimIdx) {
						double& dimMean = classMeans[dimIdx];
						dimMean = 0.0;

						ClassSamplesContainerT::const_iterator samplesIt =
							classSamples.begin();
						const ClassSamplesContainerT::const_iterator samplesItE =
							classSamples.end();

						while (samplesIt != samplesItE) {
							TERP_TRUE_OR_RETURN_FALSE(samplesIt->size() == dimsNumber,
								"Sample size mismatch")

								if ((samplesIt->operator[](dimIdx) / 1) == samplesIt->operator[](dimIdx))  // avoiding nan
									dimMean += samplesIt->operator[](dimIdx);

							++samplesIt;
						}

						dimMean /= static_cast<double>(classSamples.size());
					}

					m_hClassesParameters.push_back(classMeans);
					m_classesIndex2IDH.push_back(classID);
					m_samplesSize.push_back(classSamples.size());

					++classesIt;
				}
//				std::cout << "          Number of classes: " << classesNum << std::endl;  // --->>>>> TESTE DO NUMERO DE CLASSES

				return true;
			}

			// Calculating Regions h11 and h22 parameters
			bool IntensityPairModel::CalculatingHRegion() {

				TERP_TRUE_OR_RETURN_FALSE(m_polygonsSegImage.size() > 0, "IntensityPairModel algorithm needs polygons")

					TERP_TRUE_OR_RETURN_FALSE(m_inputRastersBands.size() > 0, "Invalid band number");

//				std::cout << "          Regions " << m_polygonsSegImage.size() << std::endl;

				std::vector< double > region;
				std::vector< double > meanRegion;
				std::vector< std::vector< double > > regionComplete;

				for (unsigned int i = 0; i < m_polygonsSegImage.size(); i++) {
					te::gm::Polygon* polygon = m_polygonsSegImage[i];

					te::rst::PolygonIterator< double > polygonsIt = te::rst::PolygonIterator< double >::
						begin(m_inputRasterPtr, polygon);
					te::rst::PolygonIterator< double > polygonsItE = te::rst::PolygonIterator< double >::
						end(m_inputRasterPtr, polygon);

					unsigned int regionsNum = 0;

					for (unsigned int j = 0; j < m_inputRastersBands.size(); ++j)
						meanRegion.push_back(0);

					while (polygonsIt != polygonsItE) {

						unsigned int column = polygonsIt.getColumn();
						unsigned int row = polygonsIt.getRow();

						if (((row % (m_lagY + 1)) == 0) && ((column % (m_lagX + 1)) == 0)) {  // De-correlating regions

							for (unsigned int j = 0; j < m_inputRastersBands.size(); ++j) {

								if (((*polygonsIt)[0] / 1) == (*polygonsIt)[0]) {  // avoiding nan
									region.push_back((*polygonsIt)[j]);
									meanRegion[j] += (*polygonsIt)[j];
								}
								else
									region.push_back(0.0);
							}

							regionComplete.push_back(region);
							region.clear();

							++regionsNum;
						}

						++polygonsIt;
					}

					m_regionsSize.push_back(regionsNum);

					for (unsigned int j = 0; j < m_inputRastersBands.size(); ++j)
						meanRegion[j] = meanRegion[j] / regionsNum;

					m_hRegionsParameters.push_back(meanRegion);

					meanRegion.clear();
				}

				return true;
			}

			// Calculating classes pc parameter
			bool IntensityPairModel::CalculatingClassesPC() {

				ClassSamplesContainerT::const_iterator classesPCIt = m_hClassesParameters.begin();

				MClassesSamplesCT::const_iterator classesIt =
					m_trainSamplesPtr->begin();
				const MClassesSamplesCT::const_iterator classesItE =
					m_trainSamplesPtr->end();

				int classIdx = 0;

				while (classesIt != classesItE) {

					double expectation1 = 0.0;
					double expectation2 = 0.0;
					double expectation3 = 0.0;

					const ClassSamplesContainerT& classSamples = classesIt->second;

					ClassSamplesContainerT::const_iterator samplesIt =
						classSamples.begin();
					const ClassSamplesContainerT::const_iterator samplesItE =
						classSamples.end();

					while (samplesIt != samplesItE) {

						expectation1 += (samplesIt->operator[](0) - classesPCIt->operator[](0))
							*
							(samplesIt->operator[](1) - classesPCIt->operator[](1));

						expectation2 += pow(samplesIt->operator[](0) - classesPCIt->operator[](0), 2);

						expectation3 += pow(samplesIt->operator[](1) - classesPCIt->operator[](1), 2);

						++samplesIt;
					}

					double pc = (expectation1 / m_samplesSize[classIdx])
						/ sqrt((expectation2 / m_samplesSize[classIdx]) * (expectation3 / m_samplesSize[classIdx]));
					if (pc < 0)
						pc = 0.001;
					m_pcClassesParameters.push_back(pc);

					++classesIt;
					++classesPCIt;
					++classIdx;
				}

				return true;
			}

			// Calculating regions pc parameter
			bool IntensityPairModel::CalculatingRegionsPC() {

				ClassSamplesContainerT::const_iterator regionsIt = m_hRegionsParameters.begin();

				for (unsigned int i = 0; i < m_polygonsSegImage.size(); ++i) {

					te::gm::Polygon* polygon = m_polygonsSegImage[i];

					double expectation1 = 0.0;
					double expectation2 = 0.0;
					double expectation3 = 0.0;

					te::rst::PolygonIterator< double > polygonsIt = te::rst::PolygonIterator< double >::
						begin(m_inputRasterPtr, polygon);
					te::rst::PolygonIterator< double > polygonsItE = te::rst::PolygonIterator< double >::
						end(m_inputRasterPtr, polygon);

					while (polygonsIt != polygonsItE) {

						double valueBand0 = 0.0;
						double valueBand1 = 0.0;

						m_inputRasterPtr->getValue(polygonsIt.getColumn(), polygonsIt.getRow(), valueBand0, 0);
						m_inputRasterPtr->getValue(polygonsIt.getColumn(), polygonsIt.getRow(), valueBand1, 1);

						if (((valueBand0 / 1) == valueBand0) && ((valueBand1 / 1) == valueBand1)) { // avoiding nan
							expectation1 += (valueBand0 - regionsIt->operator[](0))
								*
								(valueBand1 - regionsIt->operator[](1));
						}
						if ((valueBand0 / 1) == valueBand0)  // avoiding nan
							expectation2 += pow(valueBand0 - regionsIt->operator[](0), 2);
						if ((valueBand1 / 1) == valueBand1)  // avoiding nan
							expectation3 += pow(valueBand1 - regionsIt->operator[](1), 2);

						++polygonsIt;
					}

					double pc = (expectation1 / m_regionsSize[i]) / sqrt((expectation2 / m_regionsSize[i]) * (expectation3 / m_regionsSize[i]));
					if (pc < 0)
						pc = 0.0001;
					m_pcRegionsParameters.push_back(pc);

					++regionsIt;
				}

				return true;
			}

			bool IntensityPairModel::GettingAttributes() {

				// Preparing parameters for Gaussian quadrature integration

				int numberOfPoints = 48;  // the number of points and multiple (double) integral
				int numberOfIntegrals = 2;  // the number of points and multiple (double) integral
				std::vector< double > inferiorLimits(numberOfIntegrals);
				std::vector< double > superiorLimits(numberOfIntegrals);

				std::vector< long double > variableChangedOne(numberOfIntegrals);
				std::vector< long double > variableChangedTwo(numberOfIntegrals);
				std::vector< long double > x(numberOfIntegrals);
				std::vector< long double > y(numberOfIntegrals);

				// Defining limits

				double prob = 0.0;
				double z = 0.1;

				while (prob < 0.999) {
					prob = boost::math::gamma_p(m_ENL, z);
					z += 0.01;
				}

				double limitParameter = z / m_ENL;

				// Calculating Stochastic Distances, Statistical Tests and P-Values

				unsigned int i = 0;

				std::vector< double > distances;
				std::vector< double > distancesTriangular;
				std::vector< double > tests;
				std::vector< double > testsTriangular;
				std::vector< double > pValues;
				std::vector< double > pValuesTriangular;

				double degreesOfFreedom = 2;

				boost::math::chi_squared distChiSquared(degreesOfFreedom);

				ClassSamplesContainerT::const_iterator regionsIt =
					m_hRegionsParameters.begin();
				const ClassSamplesContainerT::const_iterator regionsItE =
					m_hRegionsParameters.end();

				double parcialSum = 0.0;
				double parcialSumTriangular = 0.0;
				double sumTriangular = 0.0;
				double regionBessel = 0.0;
				double classBessel = 0.0;
				double classBesselTerm = 0.0;
				double regionBesselTerm = 0.0;
				double A = 0.0;
				double aux = 0.0;
				double BI = 0.0;
				double CI = 0.0;
				double aux2 = 0.0;
				double BII = 0.0;
				double CII = 0.0;

				while (regionsIt != regionsItE) {

					ClassSamplesContainerT::const_iterator samplesIt =
						m_hClassesParameters.begin();
					const ClassSamplesContainerT::const_iterator samplesItE =
						m_hClassesParameters.end();

					unsigned int j = 0;

					while (samplesIt != samplesItE) {

						double superiorLimit1 = ((regionsIt->operator[](0) + samplesIt->operator[](0)) / 2.0) * limitParameter;
						double superiorLimit2 = ((regionsIt->operator[](1) + samplesIt->operator[](1)) / 2.0) * limitParameter;

						inferiorLimits[0] = 0.0;
						superiorLimits[0] = superiorLimit1;
						inferiorLimits[1] = 0.0;
						superiorLimits[1] = superiorLimit2;

						// Variable changes
						for (int i = 0; i < numberOfIntegrals; i++) {
							variableChangedOne[i] = (double)(superiorLimits[i] - inferiorLimits[i]) / 2.0;
							variableChangedTwo[i] = (double)(superiorLimits[i] + inferiorLimits[i]) / 2.0;
						}

						m_sum = 0.0;
						sumTriangular = 0.0;

						for (int k = 0; k < numberOfPoints; k++) {

							x[0] = variableChangedOne[0] * m_abcissas[k] + variableChangedTwo[0];

							parcialSum = 0.0;
							parcialSumTriangular = 0.0;

							for (int l = 0; l < numberOfPoints; ++l) {

								y[0] = variableChangedOne[1] * m_abcissas[l] + variableChangedTwo[1];

								regionBessel = 2.0 * m_ENL * (sqrt((x[0] * y[0]) / (regionsIt->operator[](0) * regionsIt->operator[](1)))
									*
									sqrt(m_pcRegionsParameters[i]) / (1 - m_pcRegionsParameters[i]));

								classBessel = 2.0 * m_ENL * (sqrt((x[0] * y[0]) / (samplesIt->operator[](0) * samplesIt->operator[](1)))
									*
									sqrt(m_pcClassesParameters[j]) / (1 - m_pcClassesParameters[j]));

								classBesselTerm = 0.0;
								regionBesselTerm = 0.0;

								A = pow(m_ENL, m_ENL + 1) * pow(x[0] * y[0], (double)(m_ENL - 1) / 2.0);
								aux = (m_ENL)* ((double)(x[0] / regionsIt->operator[](0)) + ((double)y[0] / regionsIt->operator[](1))) / (double)(1 - m_pcRegionsParameters[i]);
								BI = exp(-aux);
								CI = pow(regionsIt->operator[](0) * regionsIt->operator[](1), (double)(m_ENL + 1) / 2.0)
									*
									boost::math::tgamma(m_ENL) * (1 - m_pcRegionsParameters[i]) * pow(sqrt(m_pcRegionsParameters[i]), m_ENL - 1);

								aux2 = m_ENL * ((double)(x[0] / samplesIt->operator[](0)) + ((double)y[0] / samplesIt->operator[](1))) / (double)(1 - m_pcClassesParameters[j]);
								BII = exp(-aux2);
								CII = pow(samplesIt->operator[](0) * samplesIt->operator[](1), (double)(m_ENL + 1) / 2.0)
									*
									boost::math::tgamma(m_ENL) * (1 - m_pcClassesParameters[j]) * pow(sqrt(m_pcClassesParameters[j]), m_ENL - 1);

								try {
									regionBesselTerm = boost::math::cyl_bessel_i(m_ENL - 1, regionBessel);
								}
								catch (...) {
									m_stochasticDistanceType = "invalid";
								}

								try {
									classBesselTerm = boost::math::cyl_bessel_i(m_ENL - 1, classBessel);
								}
								catch (...) {
									m_stochasticDistanceType = "invalid";
								}

								parcialSum = parcialSum + ((double)m_weights[l] * sqrt(((double)A * BI * regionBesselTerm / (double)CI) * ((double)A * BII * classBesselTerm / CII)));

								parcialSumTriangular = parcialSumTriangular + (m_weights[l] * (pow(((double)A * BI * regionBesselTerm / (double)CI) - ((double)A * BII * classBesselTerm / (double)CII), 2)
									/ (((double)A * BI * regionBesselTerm / (double)CI) + ((double)A * BII * classBesselTerm / (double)CII))));
							}

							m_sum = m_sum + (m_weights[k] * variableChangedOne[1] * parcialSum);
							sumTriangular = sumTriangular + (m_weights[k] * variableChangedOne[1] * parcialSumTriangular);
						}

						// Stochastic Distances

						double distance = 0.0;

						distance = -log(m_sum * variableChangedOne[0]);
						if ((distance / 1) == distance)
							distances.push_back(distance);
						else {
							distances.push_back(100000000.0);
							distance = 100000000.0;
						}

						if ((sumTriangular / 1) == sumTriangular)
							distancesTriangular.push_back(sumTriangular * variableChangedOne[0]);
						else {
							distancesTriangular.push_back(100000000.0);
							sumTriangular = 100000000.0;
						}

						// Statistical Tests and P-Values

						double resultS = (distance)* (8 * m_regionsSize[i] * m_samplesSize[j]) / (double)(m_regionsSize[i] + m_samplesSize[j]);
						tests.push_back(resultS);

						try {
							pValues.push_back(boost::math::cdf(complement(distChiSquared, resultS)));
						}
						catch (...) {
							pValues.push_back(NAN);
						}

						resultS = sumTriangular * variableChangedOne[0] * (2 * m_regionsSize[i] * m_samplesSize[j])
							/ (double)(m_regionsSize[i] + m_samplesSize[j]);

						testsTriangular.push_back(resultS);

						try {
							pValuesTriangular.push_back(boost::math::cdf(complement(distChiSquared, resultS)));
						}
						catch (...) {
							pValuesTriangular.push_back(NAN);
						}

						++j;
						++samplesIt;
					}

					m_stochasticDistance.push_back(distances);
					m_stochasticDistanceTriangular.push_back(distancesTriangular);
					m_statisticalTest.push_back(tests);
					m_statisticalTestTriangular.push_back(testsTriangular);
					m_pValue.push_back(pValues);
					m_pValueTriangular.push_back(pValuesTriangular);
					distances.clear();
					distancesTriangular.clear();
					tests.clear();
					testsTriangular.clear();
					pValues.clear();
					pValuesTriangular.clear();

					++i;
					++regionsIt;
				}

				return true;
			}

			bool IntensityPairModel::execute() {
				m_hClassesParameters.clear();
				m_hRegionsParameters.clear();
				m_pcClassesParameters.clear();
				m_pcRegionsParameters.clear();
				m_classesIndex2IDH.clear();
				m_samplesSize.clear();
				m_regionsSize.clear();
				m_stochasticDistance.clear();
				m_statisticalTest.clear();
				m_pValue.clear();
				m_trainSamplesPtr = 0;

				m_abcissas.push_back(0.998771007252426);
				m_abcissas.push_back(0.993530172266350);
				m_abcissas.push_back(0.984124583722826);
				m_abcissas.push_back(0.970591592546247);
				m_abcissas.push_back(0.952987703160430);
				m_abcissas.push_back(0.931386690706554);
				m_abcissas.push_back(0.905879136715569);
				m_abcissas.push_back(0.876572020274247);
				m_abcissas.push_back(0.843588261624393);
				m_abcissas.push_back(0.807066204029442);
				m_abcissas.push_back(0.767159032515740);
				m_abcissas.push_back(0.724034130923814);
				m_abcissas.push_back(0.677872379632663);
				m_abcissas.push_back(0.628867396776513);
				m_abcissas.push_back(0.577224726083972);
				m_abcissas.push_back(0.523160974722233);
				m_abcissas.push_back(0.466902904750958);
				m_abcissas.push_back(0.408686481990716);
				m_abcissas.push_back(0.348755886292160);
				m_abcissas.push_back(0.287362487355455);
				m_abcissas.push_back(0.224763790394689);
				m_abcissas.push_back(0.161222356068891);
				m_abcissas.push_back(0.097004699209462);
				m_abcissas.push_back(0.032380170962869);
				m_abcissas.push_back(-0.032380170962869);
				m_abcissas.push_back(-0.097004699209462);
				m_abcissas.push_back(-0.161222356068891);
				m_abcissas.push_back(-0.224763790394689);
				m_abcissas.push_back(-0.287362487355455);
				m_abcissas.push_back(-0.348755886292160);
				m_abcissas.push_back(-0.408686481990716);
				m_abcissas.push_back(-0.466902904750958);
				m_abcissas.push_back(-0.523160974722233);
				m_abcissas.push_back(-0.577224726083972);
				m_abcissas.push_back(-0.628867396776513);
				m_abcissas.push_back(-0.677872379632663);
				m_abcissas.push_back(-0.724034130923814);
				m_abcissas.push_back(-0.767159032515740);
				m_abcissas.push_back(-0.807066204029442);
				m_abcissas.push_back(-0.843588261624393);
				m_abcissas.push_back(-0.876572020274247);
				m_abcissas.push_back(-0.905879136715569);
				m_abcissas.push_back(-0.931386690706554);
				m_abcissas.push_back(-0.952987703160430);
				m_abcissas.push_back(-0.970591592546247);
				m_abcissas.push_back(-0.984124583722826);
				m_abcissas.push_back(-0.993530172266350);
				m_abcissas.push_back(-0.998771007252426);

				m_weights.push_back(0.003153346052305);
				m_weights.push_back(0.007327553901276);
				m_weights.push_back(0.011477234579234);
				m_weights.push_back(0.015579315722943);
				m_weights.push_back(0.019616160457355);
				m_weights.push_back(0.023570760839324);
				m_weights.push_back(0.027426509708356);
				m_weights.push_back(0.031167227832798);
				m_weights.push_back(0.034777222564770);
				m_weights.push_back(0.038241351065830);
				m_weights.push_back(0.041545082943464);
				m_weights.push_back(0.044674560856694);
				m_weights.push_back(0.047616658492490);
				m_weights.push_back(0.050359035553854);
				m_weights.push_back(0.052890189485193);
				m_weights.push_back(0.055199503699984);
				m_weights.push_back(0.057277292100403);
				m_weights.push_back(0.059114839698395);
				m_weights.push_back(0.060704439165893);
				m_weights.push_back(0.062039423159892);
				m_weights.push_back(0.063114192286254);
				m_weights.push_back(0.063924238584648);
				m_weights.push_back(0.064466164435950);
				m_weights.push_back(0.064737696812683);
				m_weights.push_back(0.064737696812683);
				m_weights.push_back(0.064466164435950);
				m_weights.push_back(0.063924238584648);
				m_weights.push_back(0.063114192286254);
				m_weights.push_back(0.062039423159892);
				m_weights.push_back(0.060704439165893);
				m_weights.push_back(0.059114839698395);
				m_weights.push_back(0.057277292100403);
				m_weights.push_back(0.055199503699984);
				m_weights.push_back(0.052890189485193);
				m_weights.push_back(0.050359035553854);
				m_weights.push_back(0.047616658492490);
				m_weights.push_back(0.044674560856694);
				m_weights.push_back(0.041545082943464);
				m_weights.push_back(0.038241351065830);
				m_weights.push_back(0.034777222564770);
				m_weights.push_back(0.031167227832798);
				m_weights.push_back(0.027426509708356);
				m_weights.push_back(0.023570760839324);
				m_weights.push_back(0.019616160457355);
				m_weights.push_back(0.015579315722943);
				m_weights.push_back(0.011477234579234);
				m_weights.push_back(0.007327553901276);
				m_weights.push_back(0.003153346052305);


				if (!LoadingSamples())
				{
					TERP_LOG_AND_RETURN_FALSE( "LoadingSamples execution error." );
				}
				else
				{
//					std::cout << "LoadingSamples execution OK." << std::endl;
				}
				if (!CalculatingHClass())
				{
					TERP_LOG_AND_RETURN_FALSE( "CalculatingHClass execution error." );
				}
				else
				{
//					std::cout << "CalculatingHClass execution OK." << std::endl;
				}
				if (!CalculatingHRegion())
				{
					TERP_LOG_AND_RETURN_FALSE( "CalculatingHRegion execution error." );
				}
				else
				{
//					std::cout << "CalculatingHRegion execution OK." << std::endl;
				}
				if (!CalculatingClassesPC())
				{
					TERP_LOG_AND_RETURN_FALSE( "CalculatingClassesPC execution error." );
				}
				else
				{
//					std::cout << "CalculatingClassesPC execution OK." << std::endl;
				}
				if (!CalculatingRegionsPC())
				{
					TERP_LOG_AND_RETURN_FALSE( "CalculatingRegionsPC execution error." );
				}
				else
				{
//					std::cout << "CalculatingRegionsPC execution OK." << std::endl;
				}
				if (!GettingAttributes())
				{
					TERP_LOG_AND_RETURN_FALSE( "GettingAttributes execution error." );
				}
				else
				{
//					std::cout << "GettingAttributes execution OK." << std::endl;
				}

				return true;
			}

		} // end namespace radar
	} // end namespace rp
}   // end namespace msc

