/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

This file is part of the TerraLib - a Framework for building GIS enabled applications.

TerraLib is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

TerraLib is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with TerraLib. See COPYING. If not, write to
TerraLib Team at <terralib-team@terralib.org>.
*/

/*!
\file terralib/rp/radar/MultiSourceClassifier.cpp
\brief Performs multi source image classification.
*/

// TerraLib includes
#include "MultiSourceClassifier.h"
#include "GammaModel.h"
#include "GaussianModel.h"
#include "WishartModel.h"
#include "IntensityPairModel.h"
#include "../Macros.h"
#include "../AlgorithmParametersSerializer.h"
#include "../../raster/Band.h"
#include "../../raster/RasterFactory.h"
#include "../../raster/PositionIterator.h"

namespace te
{
	namespace rp
	{
		namespace radar
		{
			MultiSourceClassifier::MultiSourceClassifier() {
				reset();
			}

			MultiSourceClassifier::~MultiSourceClassifier() {
			}

			bool MultiSourceClassifier::execute(te::rp::AlgorithmOutputParameters& outputParams)
			{
				if (!m_isInitialized) return false;

				MultiSourceClassifier::OutputParameters* outParamsPtr = dynamic_cast<
					MultiSourceClassifier::OutputParameters* >(&outputParams);
				TERP_TRUE_OR_RETURN_FALSE(outParamsPtr, "Invalid paramters");

				typedef std::vector< std::vector< double > > Container;
				std::map< unsigned int, Container >::iterator it = m_imagesMaps.begin();
				std::map< unsigned int, Container >::iterator dist = m_distancesMaps.begin();

				for (unsigned int imageNum = 0; imageNum < m_inputParameters.m_sourcesNumber; ++imageNum) {

//					std::cout << std::endl;
//					std::cout << "PROCESSING SOURCE " << imageNum << std::endl;

					if (m_inputParameters.m_imageType[imageNum] == "Gamma") {

						te::rp::radar::GammaModel gammaInstance;

						gammaInstance.m_sourcesNumber = m_inputParameters.m_sourcesNumber;

						gammaInstance.m_polygonSamplesT = m_inputParameters.m_polygonSamplesT;
						gammaInstance.m_polygonSamplesTNew = m_inputParameters.m_polygonSamplesTNew;
						gammaInstance.m_numberOfSamples = m_inputParameters.m_numberOfSamples;

						gammaInstance.m_lagX = m_inputParameters.m_lagX;

						gammaInstance.m_lagY = m_inputParameters.m_lagY;

						gammaInstance.m_classesIDT = m_inputParameters.m_classesIDT;

						gammaInstance.m_inputRasterPtr = m_inputParameters.m_inputRasterPtr[imageNum];

						gammaInstance.m_inputRastersBands = m_inputParameters.m_inputRastersBands[imageNum];

						gammaInstance.m_polygonsSegImage = m_inputParameters.m_polygonsSegImage;

						gammaInstance.m_stochasticDistanceType = m_inputParameters.m_stochasticDistanceType[imageNum];

						gammaInstance.m_ENL = m_inputParameters.m_ENL[imageNum];

						gammaInstance.m_betaValue = m_inputParameters.m_betaValue[imageNum];

						gammaInstance.m_significanceLevel = m_inputParameters.m_significanceLevel;

						if (!gammaInstance.execute())
						{
							TERP_LOG_AND_RETURN_FALSE( "GammaModel execution error." );
						}
						else
						{
//							std::cout << "GammaModel execution OK." << std::endl;
						}

						m_stochasticDistance = gammaInstance.m_stochasticDistance;
						m_statisticalTest = gammaInstance.m_statisticalTest;
						m_pValue = gammaInstance.m_pValue;
						m_inputParameters.m_classesIDT2.clear();
						m_inputParameters.m_classesIDT2 = gammaInstance.m_classesIndex2IDLambda;
						m_degreesOfFreedom.push_back(1);
					}

					if (m_inputParameters.m_imageType[imageNum] == "Gaussian") {

						te::rp::radar::GaussianModel gaussianInstance;

						gaussianInstance.m_sourcesNumber = m_inputParameters.m_sourcesNumber;

						gaussianInstance.m_polygonSamplesT = m_inputParameters.m_polygonSamplesT;
						gaussianInstance.m_polygonSamplesTNew = m_inputParameters.m_polygonSamplesTNew;
						gaussianInstance.m_numberOfSamples = m_inputParameters.m_numberOfSamples;

						gaussianInstance.m_lagX = m_inputParameters.m_lagX;

						gaussianInstance.m_lagY = m_inputParameters.m_lagY;

						gaussianInstance.m_classesIDT = m_inputParameters.m_classesIDT;

						gaussianInstance.m_inputRasterPtr = m_inputParameters.m_inputRasterPtr[imageNum];

						gaussianInstance.m_inputRastersBands = m_inputParameters.m_inputRastersBands[imageNum];

						gaussianInstance.m_polygonsSegImage = m_inputParameters.m_polygonsSegImage;

						gaussianInstance.m_stochasticDistanceType = m_inputParameters.m_stochasticDistanceType[imageNum];

						gaussianInstance.m_significanceLevel = m_inputParameters.m_significanceLevel;

						if (!gaussianInstance.execute())
						{
							TERP_LOG_AND_RETURN_FALSE( "GaussianModel execution error." );
						}
						else
						{
//							std::cout << "GaussianModel execution OK." << std::endl;
						}

						m_stochasticDistance = gaussianInstance.m_stochasticDistance;
						m_statisticalTest = gaussianInstance.m_statisticalTest;
						m_pValue = gaussianInstance.m_pValue;
						m_inputParameters.m_classesIDT2.clear();
						m_inputParameters.m_classesIDT2 = gaussianInstance.m_classesIndex2IDMu;
						m_degreesOfFreedom.push_back((gaussianInstance.m_inputRastersBands.size()
							*
							(gaussianInstance.m_inputRastersBands.size() + 3)) / 2);
					}

					if (m_inputParameters.m_imageType[imageNum] == "Wishart") {

						te::rp::radar::WishartModel wishartInstance;

						wishartInstance.m_polygonSamplesT = m_inputParameters.m_polygonSamplesT;
						wishartInstance.m_polygonSamplesTNew = m_inputParameters.m_polygonSamplesTNew;
						wishartInstance.m_numberOfSamples = m_inputParameters.m_numberOfSamples;

						wishartInstance.m_lagX = m_inputParameters.m_lagX;

						wishartInstance.m_lagY = m_inputParameters.m_lagY;

						wishartInstance.m_classesIDT = m_inputParameters.m_classesIDT;

						wishartInstance.m_inputRasterPtr = m_inputParameters.m_inputRasterPtr[imageNum];

						wishartInstance.m_inputRastersBands = m_inputParameters.m_inputRastersBands[imageNum];

						wishartInstance.m_numberOfColumns = m_inputParameters.m_numberOfColumns;

						wishartInstance.m_numberOfRows = m_inputParameters.m_numberOfRows;

						wishartInstance.m_polygonsSegImage = m_inputParameters.m_polygonsSegImage;

						wishartInstance.m_stochasticDistanceType = m_inputParameters.m_stochasticDistanceType[imageNum];

						wishartInstance.m_radarDataRepresentationType = m_inputParameters.m_radarDataRepresentationType[imageNum];

						wishartInstance.m_vectorOrMatrixOrder = m_inputParameters.m_vectorOrMatrixOrder[imageNum];

						wishartInstance.m_ENL = m_inputParameters.m_ENL[imageNum];

						wishartInstance.m_betaValue = m_inputParameters.m_betaValue[imageNum];

						wishartInstance.m_significanceLevel = m_inputParameters.m_significanceLevel;

						if (!wishartInstance.execute())
						{
							TERP_LOG_AND_RETURN_FALSE( "WishartModel execution error." );
						}
						else
						{
//							std::cout << "WishartModel execution OK." << std::endl;
						}

						m_stochasticDistance = wishartInstance.m_stochasticDistance;
						m_statisticalTest = wishartInstance.m_statisticalTest;
						m_pValue = wishartInstance.m_pValue;
						m_inputParameters.m_classesIDT2.clear();
						m_inputParameters.m_classesIDT2 = wishartInstance.m_classesIndex2ID;
						m_degreesOfFreedom.push_back(m_inputParameters.m_vectorOrMatrixOrder[imageNum]
							*
							m_inputParameters.m_vectorOrMatrixOrder[imageNum]);
					}

					if (m_inputParameters.m_imageType[imageNum] == "IntensityPair") {

						te::rp::radar::IntensityPairModel intensityPairInstance;

						intensityPairInstance.m_polygonSamplesT = m_inputParameters.m_polygonSamplesT;
						intensityPairInstance.m_polygonSamplesTNew = m_inputParameters.m_polygonSamplesTNew;
						intensityPairInstance.m_numberOfSamples = m_inputParameters.m_numberOfSamples;

						intensityPairInstance.m_lagX = m_inputParameters.m_lagX;

						intensityPairInstance.m_lagY = m_inputParameters.m_lagY;

						intensityPairInstance.m_classesIDT = m_inputParameters.m_classesIDT;

						intensityPairInstance.m_inputRasterPtr = m_inputParameters.m_inputRasterPtr[imageNum];

						intensityPairInstance.m_inputRastersBands = m_inputParameters.m_inputRastersBands[imageNum];

						intensityPairInstance.m_polygonsSegImage = m_inputParameters.m_polygonsSegImage;

						intensityPairInstance.m_stochasticDistanceType = m_inputParameters.m_stochasticDistanceType[imageNum];

						intensityPairInstance.m_ENL = m_inputParameters.m_ENL[imageNum];

						intensityPairInstance.m_significanceLevel = m_inputParameters.m_significanceLevel;

						if (!intensityPairInstance.execute())
						{
							TERP_LOG_AND_RETURN_FALSE( "Intensity Pair Model execution error." );
						}
						else
						{
//							std::cout << "Intensity Pair Model execution OK." << std::endl;
						}

						if (m_inputParameters.m_stochasticDistanceType[imageNum] == "Bhattacharyya") {
							m_stochasticDistance = intensityPairInstance.m_stochasticDistance;
							m_statisticalTest = intensityPairInstance.m_statisticalTest;
							m_pValue = intensityPairInstance.m_pValue;
						}
						if (m_inputParameters.m_stochasticDistanceType[imageNum] == "Triangular") {
							m_stochasticDistance = intensityPairInstance.m_stochasticDistanceTriangular;
							m_statisticalTest = intensityPairInstance.m_statisticalTestTriangular;
							m_pValue = intensityPairInstance.m_pValueTriangular;
						}
						m_inputParameters.m_classesIDT2.clear();
						m_inputParameters.m_classesIDT2 = intensityPairInstance.m_classesIndex2IDH;
						m_degreesOfFreedom.push_back(2);
					}

					if ((m_inputParameters.m_combinationType == "Hellinger") && (m_inputParameters.m_sourcesNumber > 1))
						m_imagesMaps.insert(it, std::pair< unsigned int, Container >(imageNum, m_stochasticDistance));
					else
						m_imagesMaps.insert(it, std::pair< unsigned int, Container >(imageNum, m_statisticalTest));

					//if (((m_inputParameters.m_combinationType == "sum") && (m_inputParameters.m_sourcesNumber > 1))
					//	||
					//	((m_inputParameters.m_combinationType == "minimum") && (m_inputParameters.m_sourcesNumber > 1)))
						m_distancesMaps.insert(dist, std::pair< unsigned int, Container >(imageNum, m_stochasticDistance));
				}

				const unsigned int classesNumber = m_inputParameters.m_numberOfClassesOnTraining;

				if ((m_inputParameters.m_sourcesNumber > 1) && (m_inputParameters.m_combinationType != "fuzzy")) {

				    dist = m_distancesMaps.begin();
								unsigned int numberRegionsD = dist->second.size();

								std::vector< double > allDistances;
								std::vector< double > allStatistics;

								for (unsigned int i = 0; i < numberRegionsD; ++i) {

									for (dist = m_distancesMaps.begin(); dist != m_distancesMaps.end(); ++dist) {

										for (unsigned int j = 0; j < classesNumber; ++j) {

											const Container& regionMapD = dist->second;

											Container::const_iterator regionsDIt = regionMapD.begin();

											for (unsigned int k = 1; k <= i; ++k)
												++regionsDIt;

											allDistances.push_back(regionsDIt->operator[](j));
										}
								}

									for (it = m_imagesMaps.begin(); it != m_imagesMaps.end(); ++it) {

										for (unsigned int j = 0; j < classesNumber; ++j) {

											const Container& regionMapS = it->second;

											Container::const_iterator regionsSIt = regionMapS.begin();

											for (unsigned int k = 1; k <= i; ++k)
												++regionsSIt;

											allStatistics.push_back(regionsSIt->operator[](j));
										}
								}

									outParamsPtr->m_outAllDistances.push_back(allDistances);
									allDistances.clear();
									outParamsPtr->m_outAllStatistics.push_back(allStatistics);
									allStatistics.clear();
   				 }
				}

				// Fuzzy Classification

				if ((m_inputParameters.m_sourcesNumber > 1) && (m_inputParameters.m_combinationType == "fuzzy")) {

					// creating the output Fuzzy raster where each band corresponds to a class

					std::auto_ptr<te::rst::Grid>
						outputGridPtr(new te::rst::Grid(*m_inputParameters.m_inputRasterPtr[0]->getGrid()));

					// use the properties from the first band to create the output bands
					te::rst::BandProperty*
						inputBandProperty(new te::rst::BandProperty(*(m_inputParameters.m_inputRasterPtr[0]->getBand
						(0)->getProperty())));

					std::vector< te::rst::BandProperty* > bandsProperties;

					for (unsigned int i = 0; i < m_inputParameters.m_numberOfClassesOnTraining; ++i) {  // +1 considering the band for not classified
						te::rst::BandProperty* bandProperty
							(new te::rst::BandProperty(*inputBandProperty));

						if (i != 0) {
							bandProperty->m_colorInterp = te::rst::UndefCInt;
							bandProperty->m_type = te::dt::DOUBLE_TYPE;
						}
						else {
							bandProperty->m_colorInterp = te::rst::GrayIdxCInt;
							bandProperty->m_type = te::dt::DOUBLE_TYPE;
						}
						bandsProperties.push_back(bandProperty);
					}

					try
					{
            AlgorithmParametersSerializer paramsSerializer;
            TERP_INSTANCE_TRUE_OR_LOG(  m_inputParameters.serialize( paramsSerializer ),
              "Parameter serialization error" );

            std::map <std::string, std::string> metadata;
            paramsSerializer.getMetaData( metadata );

            outParamsPtr->m_outputRasterClassImageFuzzyPtr.reset(
              te::rst::RasterFactory::make("GDAL",
                outputGridPtr.release(),
                bandsProperties,
                outParamsPtr->m_rInfoClassImageFuzzy,
                metadata,
                nullptr,
                nullptr));
          }
          catch(...)
          {
            outParamsPtr->m_outputRasterClassImageFuzzyPtr.reset();
          }
					if (outParamsPtr->m_outputRasterClassImageFuzzyPtr.get() == 0) {
						TERP_INSTANCE_LOG_AND_RETURN_FALSE( "Output fuzzy raster creation error" );
					}

					// creating the output raster for reability map (statistical test)

					typedef std::vector< std::vector< unsigned int > > ContainerClasses;
					ContainerClasses::const_iterator bandsIt = m_inputParameters.m_inputRastersBands.begin();

					std::vector< te::rst::BandProperty* > bandsPropertiesMap;
					bandsPropertiesMap.push_back(new te::rst::BandProperty(
						*(m_inputParameters.m_inputRasterPtr[0]->getBand(
						bandsIt->operator[](0))->getProperty())));
					bandsPropertiesMap[0]->m_colorInterp = te::rst::GrayIdxCInt;
					bandsPropertiesMap[0]->m_noDataValue = 0;
					bandsPropertiesMap[0]->m_type = te::dt::DOUBLE_TYPE;

          try
          {
            AlgorithmParametersSerializer paramsSerializer;
            TERP_INSTANCE_TRUE_OR_LOG(  m_inputParameters.serialize( paramsSerializer ),
              "Parameter serialization error" );

            std::map <std::string, std::string> metadata;
            paramsSerializer.getMetaData( metadata );

            outParamsPtr->m_outputRasterMapPtr.reset(
              te::rst::RasterFactory::make(
              outParamsPtr->m_rType,
              new te::rst::Grid(*m_inputParameters.m_inputRasterPtr[0]->getGrid()),
              bandsPropertiesMap, outParamsPtr->m_rInfoMap,
              metadata,
              nullptr,
              nullptr));
          }
          catch(...)
          {
            outParamsPtr->m_outputRasterMapPtr.reset();
          }
					TERP_TRUE_OR_RETURN_FALSE(outParamsPtr->m_outputRasterMapPtr.get(),
						"Output statistical test map raster creation error");

					te::rst::Raster& outputRasterMap = *outParamsPtr->m_outputRasterMapPtr;
					te::rst::Band& outputBandMap = *outputRasterMap.getBand(0);

					// creating the output raster for reability map (p-value)

					std::vector< te::rst::BandProperty* > bandsPropertiesMapPValue;
					bandsPropertiesMapPValue.push_back(new te::rst::BandProperty(
						*(m_inputParameters.m_inputRasterPtr[0]->getBand(
						bandsIt->operator[](0))->getProperty())));
					bandsPropertiesMapPValue[0]->m_colorInterp = te::rst::GrayIdxCInt;
					bandsPropertiesMapPValue[0]->m_noDataValue = 0;
					bandsPropertiesMapPValue[0]->m_type = te::dt::DOUBLE_TYPE;

          try
          {
            AlgorithmParametersSerializer paramsSerializer;
            TERP_INSTANCE_TRUE_OR_LOG(  m_inputParameters.serialize( paramsSerializer ),
              "Parameter serialization error" );

            std::map <std::string, std::string> metadata;
            paramsSerializer.getMetaData( metadata );

            outParamsPtr->m_outputRasterMapPValuePtr.reset(
              te::rst::RasterFactory::make(
              outParamsPtr->m_rType,
              new te::rst::Grid(*m_inputParameters.m_inputRasterPtr[0]->getGrid()),
              bandsPropertiesMapPValue, outParamsPtr->m_rInfoMapPValue,
              metadata,
              nullptr,
              nullptr));
          }
          catch(...)
          {
            outParamsPtr->m_outputRasterMapPValuePtr.reset();
          }
					TERP_TRUE_OR_RETURN_FALSE(outParamsPtr->m_outputRasterMapPValuePtr.get(),
						"Output statistical test map raster creation error");

					te::rst::Raster& outputRasterMapPValue = *outParamsPtr->m_outputRasterMapPValuePtr;
					te::rst::Band& outputBandMapPValue = *outputRasterMapPValue.getBand(0);

					std::vector< std::vector< double > > valueMapFuzzyFinal;
					std::vector< std::vector< double > > valuePFuzzyFinal;
					std::vector< std::vector< double > > classMapFuzzyFinal;
					std::vector< double > valueMapFuzzy;
					std::vector< double > valuePFuzzy;
					std::vector< double > classMapFuzzy;
					std::vector< double > numeberOfVotes(m_inputParameters.m_numberOfClassesOnTraining, 0.0);
					std::vector< std::vector< double > > totalNumberVotes;
					unsigned int classMap;
					double valueMap;
					double valueStatistic;
					double valueP;
					unsigned int numberRegions = 0;

					int imageNum = 0;

					for (it = m_imagesMaps.begin(); it != m_imagesMaps.end(); ++it) {

						boost::math::chi_squared distChiSquared(m_degreesOfFreedom[imageNum]);

						numberRegions = 0;

						//const Container& regionMap = normalizedIt->second;
						const Container& regionMap = it->second;

						Container::const_iterator regionsIt = regionMap.begin();
						const Container::const_iterator regionsItE = regionMap.end();

						while (regionsIt != regionsItE) {
							valueMap = regionsIt->operator[](0);
							classMap = 0;

							for (unsigned int i = 1; i < m_inputParameters.m_numberOfClassesOnTraining; ++i) {

								if (regionsIt->operator[](i) < valueMap) {
									valueMap = regionsIt->operator[](i);
									classMap = i;
								}
							}

							double pValue = 0.0;
							try {
								pValue = boost::math::cdf(complement(distChiSquared, valueMap));
							}
							catch (...) {
								pValue = NAN;
								classMap = m_inputParameters.m_numberOfClassesOnTraining;
							}

							valueMapFuzzy.push_back(valueMap);
							valuePFuzzy.push_back(pValue);
							classMapFuzzy.push_back((double)classMap);

							++numberRegions;
							++regionsIt;
						}

						valueMapFuzzyFinal.push_back(valueMapFuzzy);
						valuePFuzzyFinal.push_back(valuePFuzzy);
						classMapFuzzyFinal.push_back(classMapFuzzy);
						valueMapFuzzy.clear();
						valuePFuzzy.clear();
						classMapFuzzy.clear();
						++imageNum;
					}

					for (unsigned int i = 0; i < numberRegions; ++i) {

						Container::const_iterator monoSourceIt = classMapFuzzyFinal.begin();
						const Container::const_iterator monoSourceItE = classMapFuzzyFinal.end();

						for (unsigned int j = 0; j < numeberOfVotes.size(); ++j)
							numeberOfVotes[j] = 0;

						// calculating the number of votes for each class

						while (monoSourceIt != monoSourceItE) {
							++numeberOfVotes[(int)monoSourceIt->operator[](i)];
							++monoSourceIt;
						}

						Container::const_iterator statisticIt = valueMapFuzzyFinal.begin();
						const Container::const_iterator statisticItE = valueMapFuzzyFinal.end();

						valueStatistic = 1.0;

						while (statisticIt != statisticItE) {
							valueStatistic *= statisticIt->operator[](i);
							++statisticIt;
						}

						Container::const_iterator pIt = valuePFuzzyFinal.begin();
						const Container::const_iterator pItE = valuePFuzzyFinal.end();

						valueP = 1.0;

						while (pIt != pItE) {
							valueP *= pIt->operator[](i);
							++pIt;
						}

						// Generating Fuzzy classification image

						std::vector< double > valuesBands;

						for (unsigned int j = 0; j < numeberOfVotes.size(); ++j) {
							double value = (double)(numeberOfVotes[j] / m_inputParameters.m_sourcesNumber);
							valuesBands.push_back(value);

							te::rst::PolygonIterator< double > polygonsIt = te::rst::PolygonIterator< double >::
								begin(m_inputParameters.m_inputRasterPtr[0], m_inputParameters.m_polygonsSegImage[i]);
							te::rst::PolygonIterator< double > polygonsItE = te::rst::PolygonIterator< double >::
								end(m_inputParameters.m_inputRasterPtr[0], m_inputParameters.m_polygonsSegImage[i]);

							while (polygonsIt != polygonsItE) {
								outParamsPtr->m_outputRasterClassImageFuzzyPtr->getBand(j)->setValue(polygonsIt.getColumn(),
									polygonsIt.getRow(), value);
								outputBandMap.setValue(polygonsIt.getColumn(), polygonsIt.getRow(), valueStatistic);
								outputBandMapPValue.setValue(polygonsIt.getColumn(), polygonsIt.getRow(), valueP);
								++polygonsIt;
							}
						}

						outParamsPtr->m_valuesS.push_back(valueStatistic);
						outParamsPtr->m_valuesP.push_back(valueP);

						outParamsPtr->m_bandsFuzzy.push_back(valuesBands);
						valuesBands.clear();
					}
				}

				if ((m_inputParameters.m_sourcesNumber == 1) || (m_inputParameters.m_combinationType != "fuzzy")) {

					typedef std::vector< std::vector< unsigned int > > ContainerClasses;
					ContainerClasses::const_iterator bandsIt = m_inputParameters.m_inputRastersBands.begin();

					// creating the output raster for classified image

					std::vector< te::rst::BandProperty* > bandsPropertiesClassImage;
					bandsPropertiesClassImage.push_back(new te::rst::BandProperty(
						*(m_inputParameters.m_inputRasterPtr[0]->getBand(
						bandsIt->operator[](0))->getProperty())));
					bandsPropertiesClassImage[0]->m_colorInterp = te::rst::GrayIdxCInt;
					bandsPropertiesClassImage[0]->m_type = te::dt::INT16_TYPE;

          try
          {
            AlgorithmParametersSerializer paramsSerializer;
            TERP_INSTANCE_TRUE_OR_LOG(  m_inputParameters.serialize( paramsSerializer ),
              "Parameter serialization error" );

            std::map <std::string, std::string> metadata;
            paramsSerializer.getMetaData( metadata );

            outParamsPtr->m_outputRasterClassImagePtr.reset(
              te::rst::RasterFactory::make(
              outParamsPtr->m_rType,
              new te::rst::Grid(*m_inputParameters.m_inputRasterPtr[0]->getGrid()),
              bandsPropertiesClassImage, outParamsPtr->m_rInfoClassImage,
              metadata,
              nullptr,
              nullptr));
          }
          catch(...)
          {
            outParamsPtr->m_outputRasterClassImagePtr.reset();
          }
					TERP_TRUE_OR_RETURN_FALSE(outParamsPtr->m_outputRasterClassImagePtr.get(),
						"Output Classified image raster creation error");

					te::rst::Raster& outputRaster = *outParamsPtr->m_outputRasterClassImagePtr;
					te::rst::Band& outputBand = *outputRaster.getBand(0);

					// creating the output raster for reability map (statistical test)

					std::vector< te::rst::BandProperty* > bandsPropertiesMap;
					bandsPropertiesMap.push_back(new te::rst::BandProperty(
						*(m_inputParameters.m_inputRasterPtr[0]->getBand(
						bandsIt->operator[](0))->getProperty())));
					bandsPropertiesMap[0]->m_colorInterp = te::rst::GrayIdxCInt;
					bandsPropertiesMap[0]->m_noDataValue = 0;
					bandsPropertiesMap[0]->m_type = te::dt::DOUBLE_TYPE;

          try
          {
            AlgorithmParametersSerializer paramsSerializer;
            TERP_INSTANCE_TRUE_OR_LOG(  m_inputParameters.serialize( paramsSerializer ),
              "Parameter serialization error" );

            std::map <std::string, std::string> metadata;
            paramsSerializer.getMetaData( metadata );

            outParamsPtr->m_outputRasterMapPtr.reset(
              te::rst::RasterFactory::make(
              outParamsPtr->m_rType,
              new te::rst::Grid(*m_inputParameters.m_inputRasterPtr[0]->getGrid()),
              bandsPropertiesMap, outParamsPtr->m_rInfoMap,
              metadata,
              nullptr,
              nullptr));
          }
          catch(...)
          {
            outParamsPtr->m_outputRasterMapPtr.reset();
          }
					TERP_TRUE_OR_RETURN_FALSE(outParamsPtr->m_outputRasterMapPtr.get(),
						"Output statistical test map raster creation error");

					te::rst::Raster& outputRasterMap = *outParamsPtr->m_outputRasterMapPtr;
					te::rst::Band& outputBandMap = *outputRasterMap.getBand(0);

					// creating the output raster for reability map (p-value)

					std::vector< te::rst::BandProperty* > bandsPropertiesMapPValue;
					bandsPropertiesMapPValue.push_back(new te::rst::BandProperty(
						*(m_inputParameters.m_inputRasterPtr[0]->getBand(
						bandsIt->operator[](0))->getProperty())));
					bandsPropertiesMapPValue[0]->m_colorInterp = te::rst::GrayIdxCInt;
					bandsPropertiesMapPValue[0]->m_noDataValue = 0;
					bandsPropertiesMapPValue[0]->m_type = te::dt::DOUBLE_TYPE;

          try
          {
            AlgorithmParametersSerializer paramsSerializer;
            TERP_INSTANCE_TRUE_OR_LOG(  m_inputParameters.serialize( paramsSerializer ),
              "Parameter serialization error" );

            std::map <std::string, std::string> metadata;
            paramsSerializer.getMetaData( metadata );

            outParamsPtr->m_outputRasterMapPValuePtr.reset(
              te::rst::RasterFactory::make(
              outParamsPtr->m_rType,
              new te::rst::Grid(*m_inputParameters.m_inputRasterPtr[0]->getGrid()),
              bandsPropertiesMapPValue, outParamsPtr->m_rInfoMapPValue,
              metadata,
              nullptr,
              nullptr));
          }
          catch(...)
          {
            outParamsPtr->m_outputRasterMapPValuePtr.reset();
          }
					TERP_TRUE_OR_RETURN_FALSE(outParamsPtr->m_outputRasterMapPValuePtr.get(),
						"Output statistical test map raster creation error");

					te::rst::Raster& outputRasterMapPValue = *outParamsPtr->m_outputRasterMapPValuePtr;
					te::rst::Band& outputBandMapPValue = *outputRasterMapPValue.getBand(0);

					unsigned int classMap;
					double valueMap;
					double normalizedValueMap;

					// Classification of a single source

					if (m_inputParameters.m_sourcesNumber == 1) {

						outParamsPtr->m_outDistancesPerClass = m_stochasticDistance;
						outParamsPtr->m_outStatisticsPerClass = m_statisticalTest;
						outParamsPtr->m_outPValuesPerClass = m_pValue;

						Container::const_iterator regionIt = m_stochasticDistance.begin();
						const Container::const_iterator regionItE = m_stochasticDistance.end();

						double valueDistance;

						while (regionIt != regionItE) {

							valueDistance = 100000000.0;

							if (regionIt->operator[](0) < valueDistance)
								valueDistance = regionIt->operator[](0);

							for (unsigned int i = 1; i < m_inputParameters.m_numberOfClassesOnTraining; ++i) {

								// Finding minimum stochastic distance value

								if (regionIt->operator[](i) < valueDistance) {
									valueDistance = regionIt->operator[](i);
								}
							}

							outParamsPtr->m_distances.push_back(valueDistance);

							++regionIt;
						}
						// Finding critical statistical test value

						boost::math::chi_squared distChiSquared(m_degreesOfFreedom[0]);
						double criticalS = boost::math::quantile(complement(distChiSquared, m_inputParameters.m_significanceLevel));

						std::vector< double > normalizedS;
						std::vector< std::vector< double >> allNormalizedS;

						// Finding max and min statistical test value

						it = m_imagesMaps.begin();

						const Container& toNormalizeRegionMap = it->second;

						Container::const_iterator toNormalizeRegionsIt = toNormalizeRegionMap.begin();
						const Container::const_iterator toNormalizeRegionsItE = toNormalizeRegionMap.end();

						double minS = toNormalizeRegionsIt->operator[](0) - criticalS;
						double maxS = toNormalizeRegionsIt->operator[](0) - criticalS;

						while (toNormalizeRegionsIt != toNormalizeRegionsItE) {

							for (unsigned int i = 0; i < classesNumber; ++i) {
								double valueMinusCriticalS = toNormalizeRegionsIt->operator[](i) - criticalS;
								normalizedS.push_back(valueMinusCriticalS);
								if (valueMinusCriticalS < minS)
									minS = valueMinusCriticalS;
								if (valueMinusCriticalS > maxS)
									maxS = valueMinusCriticalS;
							}

							allNormalizedS.push_back(normalizedS);
							normalizedS.clear();
							++toNormalizeRegionsIt;
						}

						it = m_imagesMaps.begin();

						const Container& regionMap = it->second;

						Container::const_iterator regionsIt = regionMap.begin();
						const Container::const_iterator regionsItE = regionMap.end();

						Container::const_iterator normalizedRegionsIt = allNormalizedS.begin();
						const Container::const_iterator normalizedRegionsItE = allNormalizedS.end();

						unsigned int regionIdx = 0;

						while (regionsIt != regionsItE) {
							valueMap = regionsIt->operator[](0);
							classMap = 0;
							normalizedValueMap = normalizedRegionsIt->operator[](0);

							for (unsigned int i = 1; i < classesNumber; ++i) {

								// Finding minimum statistical test value

								if (regionsIt->operator[](i) < valueMap) {
									valueMap = regionsIt->operator[](i);
									classMap = i;
									normalizedValueMap = normalizedRegionsIt->operator[](i);
								}
							}

							// Generating classifiyng image and reliability map

							te::rst::PolygonIterator< double > polygonsIt = te::rst::PolygonIterator< double >::
								begin(m_inputParameters.m_inputRasterPtr[0], m_inputParameters.m_polygonsSegImage[regionIdx]);
							te::rst::PolygonIterator< double > polygonsItE = te::rst::PolygonIterator< double >::
								end(m_inputParameters.m_inputRasterPtr[0], m_inputParameters.m_polygonsSegImage[regionIdx]);

							unsigned int regionClassValue = classMap;

							double pValue = 0.0;
							try {
								pValue = boost::math::cdf(complement(distChiSquared, valueMap));
							}
							catch (...) {
								pValue = NAN;
								regionClassValue = m_inputParameters.m_numberOfClassesOnTraining;
							}

							// Normalizing statistical test value
       double valueNormalizedS = double(normalizedValueMap - minS) / (maxS - minS);

							double IDSegment = 0.0;
							m_inputParameters.m_rasterSegPtr->getBand(0)->getValue(polygonsIt.getColumn(), polygonsIt.getRow(), IDSegment);
							outParamsPtr->m_IDSegment.push_back(IDSegment);

							while (polygonsIt != polygonsItE) {
								outputBand.setValue(polygonsIt.getColumn(), polygonsIt.getRow(), regionClassValue);
								outputBandMap.setValue(polygonsIt.getColumn(), polygonsIt.getRow(), valueNormalizedS);
								outputBandMapPValue.setValue(polygonsIt.getColumn(), polygonsIt.getRow(), pValue);
								++polygonsIt;
							}

							outParamsPtr->m_valuesS.push_back(valueMap);
							outParamsPtr->m_valuesP.push_back(pValue);
							++regionsIt;
							++regionIdx;
							++normalizedRegionsIt;
						}
					}

					// MultiSource Classification

					std::map< unsigned int, Container > normalizedImagesMaps;
					std::map< unsigned int, Container > beforeNormalizedImagesMaps;

					if ((m_inputParameters.m_sourcesNumber > 1) && (m_inputParameters.m_combinationType != "fuzzy")) {

						// Special case: Hellinger

						if (m_inputParameters.m_combinationType == "Hellinger") {

							std::vector< double > valuesAttribute;
							std::vector< std::vector< double > > allRegions;

							double globalMinValueS = 10000000000.0;
							double globalMaxValueS = 0.0;
							double globalMinD1 = 10000000000.0;
							double globalMaxD1 = 0.0;
							double globalMinD2 = 10000000000.0;
							double globalMaxD2 = 0.0;

							it = m_imagesMaps.begin();
							unsigned int numberRegions = it->second.size();

							int regionSize = 0;

							for (unsigned int i = 0; i < numberRegions; ++i) {

								te::rst::PolygonIterator< double > polygonsIt = te::rst::PolygonIterator< double >::
									begin(m_inputParameters.m_inputRasterPtr[0], m_inputParameters.m_polygonsSegImage[i]);
								te::rst::PolygonIterator< double > polygonsItE = te::rst::PolygonIterator< double >::
									end(m_inputParameters.m_inputRasterPtr[0], m_inputParameters.m_polygonsSegImage[i]);

								while (polygonsIt != polygonsItE) {
									++regionSize;
									++polygonsIt;
								}

								for (unsigned int j = 0; j < classesNumber; ++j) {

									double productValue = 1.0;

									double imageIdx = 0;

									for (it = m_imagesMaps.begin(); it != m_imagesMaps.end(); ++it) {

										const Container& regionMap = it->second;

										Container::const_iterator regionsIt = regionMap.begin();

										for (unsigned int k = 1; k <= i; ++k)
											++regionsIt;

										productValue *= (1 - regionsIt->operator[](j));

										if (imageIdx == 0) {
											if (regionsIt->operator[](j) < globalMinD1)
												globalMinD1 = regionsIt->operator[](j);

											if (regionsIt->operator[](j) > globalMaxD1)
												globalMaxD1 = regionsIt->operator[](j);
										}

										if (imageIdx == 1) {
											if (regionsIt->operator[](j) < globalMinD2)
												globalMinD2 = regionsIt->operator[](j);

											if (regionsIt->operator[](j) > globalMaxD2)
												globalMaxD2 = regionsIt->operator[](j);
										}

										++imageIdx;
									}

									double valueS = ((2 * regionSize * m_inputParameters.m_numberOfSamples[j] * 4)
										/
										(regionSize + m_inputParameters.m_numberOfSamples[j]))
										*
										(1 - productValue);

									valuesAttribute.push_back(valueS);

									// Finding global max and min statistical test value

									if (valueS < globalMinValueS)
										globalMinValueS = valueS;

									if (valueS > globalMaxValueS)
										globalMaxValueS = valueS;
								}

								allRegions.push_back(valuesAttribute);
								valuesAttribute.clear();
								regionSize = 0;
							}

							// Normalizing and finding minimum statistical test value

							Container::const_iterator regionsIt = allRegions.begin();
							const Container::const_iterator regionsItE = allRegions.end();

							double normalizedS;
							double valueMap;
							unsigned int classMap;
							std::vector< double > everySValue;
							//std::vector< double > everyPValue;

							unsigned int regionIdx = 0;

							while (regionsIt != regionsItE) {

								valueMap = double(regionsIt->operator[](0) - globalMinValueS)
									/
									(globalMaxValueS - globalMinValueS);
								classMap = 0;

								everySValue.push_back(valueMap);

								for (unsigned int i = 1; i < classesNumber; ++i) {

									normalizedS = double(regionsIt->operator[](i) - globalMinValueS)
										/
										(globalMaxValueS - globalMinValueS);

									if (normalizedS < valueMap) {
										valueMap = normalizedS;
										classMap = i;
									}

									everySValue.push_back(normalizedS);
								}

								outParamsPtr->m_outStatisticsPerClass.push_back(everySValue);
								everySValue.clear();

								// Generating classifiyng image and reliability map

								te::rst::PolygonIterator< double > polygonsIt = te::rst::PolygonIterator< double >::
									begin(m_inputParameters.m_inputRasterPtr[0], m_inputParameters.m_polygonsSegImage[regionIdx]);
								te::rst::PolygonIterator< double > polygonsItE = te::rst::PolygonIterator< double >::
									end(m_inputParameters.m_inputRasterPtr[0], m_inputParameters.m_polygonsSegImage[regionIdx]);

								unsigned int regionClassValue = m_inputParameters.m_classesIDT2[classMap];

								double IDSegment = 0.0;
								m_inputParameters.m_rasterSegPtr->getBand(0)->getValue(polygonsIt.getColumn(), polygonsIt.getRow(), IDSegment);
								outParamsPtr->m_IDSegment.push_back(IDSegment);

								while (polygonsIt != polygonsItE) {
									outputBand.setValue(polygonsIt.getColumn(), polygonsIt.getRow(), regionClassValue);
									outputBandMap.setValue(polygonsIt.getColumn(), polygonsIt.getRow(), valueMap);
									++polygonsIt;
								}

								outParamsPtr->m_valuesS.push_back(valueMap);
								++regionsIt;
								++regionIdx;
							}
						}

						// Sum combination

						if (m_inputParameters.m_combinationType == "sum") {

							// for distances output

							{
								dist = m_distancesMaps.begin();
								unsigned int numberRegionsD = dist->second.size();

								std::vector< double > distancesRegions;

								for (unsigned int i = 0; i < numberRegionsD; ++i) {

									double Dtest = 10000000000.0;

									for (unsigned int j = 0; j < classesNumber; ++j) {

										double sumValueD = 0.0;

										for (dist = m_distancesMaps.begin(); dist != m_distancesMaps.end(); ++dist) {

											const Container& regionMapD = dist->second;

											Container::const_iterator regionsDIt = regionMapD.begin();

											for (unsigned int k = 1; k <= i; ++k)
												++regionsDIt;

											sumValueD += regionsDIt->operator[](j);
										}

										distancesRegions.push_back(sumValueD);

										// Finding minimum distance sum value

										if (sumValueD < Dtest)
											Dtest = sumValueD;
									}

									outParamsPtr->m_outDistancesPerClass.push_back(distancesRegions);
									distancesRegions.clear();

									outParamsPtr->m_distances.push_back(Dtest);

									Dtest = 10000000000.0;
								}
							}

							// Finding critical statistical test value

							unsigned int sumDegreesOfFreedom = 0;
							for (unsigned int i = 0; i < m_inputParameters.m_sourcesNumber; ++i)
								sumDegreesOfFreedom += m_degreesOfFreedom[i];
							boost::math::chi_squared distChiSquaredSum(sumDegreesOfFreedom);
							double sumCriticalS = boost::math::quantile(complement(distChiSquaredSum, m_inputParameters.m_significanceLevel));

							std::vector< double > pValuesRegion;
							std::vector< double > valuesAttribute;
							std::vector< std::vector< double > > allRegions;

							double globalMinValueS = 10000000000.0;
							double globalMaxValueS = 0.0;

							it = m_imagesMaps.begin();
							unsigned int numberRegions = it->second.size();

							for (unsigned int i = 0; i < numberRegions; ++i) {

								double Stest = 10000000000.0;

								for (unsigned int j = 0; j < classesNumber; ++j) {

									double sumValue = 0.0;

									for (it = m_imagesMaps.begin(); it != m_imagesMaps.end(); ++it) {

										const Container& regionMap = it->second;

										Container::const_iterator regionsIt = regionMap.begin();

										for (unsigned int k = 1; k <= i; ++k)
											++regionsIt;

										sumValue += regionsIt->operator[](j);
									}

									double newValueS = sumValue - sumCriticalS;

									valuesAttribute.push_back(newValueS);

									// Finding minimum statistical test sum value for computing p-value (before normalization process)

									if (sumValue < Stest)
										Stest = sumValue;

									try {
										pValuesRegion.push_back(boost::math::cdf(complement(distChiSquaredSum, sumValue)));
									}
									catch (...) {
										pValuesRegion.push_back(NAN);
									}

									// Finding global max and min statistical test value

									if (newValueS < globalMinValueS)
										globalMinValueS = newValueS;

									if (newValueS > globalMaxValueS)
										globalMaxValueS = newValueS;
								}

								outParamsPtr->m_outPValuesPerClass.push_back(pValuesRegion);
								pValuesRegion.clear();

								allRegions.push_back(valuesAttribute);
								valuesAttribute.clear();

								// Computing p-value of minimum statistical test sum value

								try {
									outParamsPtr->m_pValuesSum.push_back(boost::math::cdf(complement(distChiSquaredSum, Stest)));
								}
								catch (...) {
									outParamsPtr->m_pValuesSum.push_back(NAN);
								}
								Stest = 10000000000.0;
							}

							// Normalizing and finding minimum statistical test value

							Container::const_iterator regionsIt = allRegions.begin();
							const Container::const_iterator regionsItE = allRegions.end();

							double normalizedS;
							double valueMap;
							unsigned int classMap;
							std::vector< double > everySValue;
							std::vector< double > everyPValue;

							unsigned int regionIdx = 0;

							while (regionsIt != regionsItE) {

								valueMap = double(regionsIt->operator[](0) - globalMinValueS)
									/
									(globalMaxValueS - globalMinValueS);
								classMap = 0;

								everySValue.push_back(valueMap);

								for (unsigned int i = 1; i < classesNumber; ++i) {

									normalizedS = double(regionsIt->operator[](i) - globalMinValueS)
										/
										(globalMaxValueS - globalMinValueS);

									if (normalizedS < valueMap) {
										valueMap = normalizedS;
										classMap = i;
									}

									everySValue.push_back(normalizedS);
								}

								outParamsPtr->m_outStatisticsPerClass.push_back(everySValue);
								everySValue.clear();

								// Generating classifiyng image and reliability map

								te::rst::PolygonIterator< double > polygonsIt = te::rst::PolygonIterator< double >::
									begin(m_inputParameters.m_inputRasterPtr[0], m_inputParameters.m_polygonsSegImage[regionIdx]);
								te::rst::PolygonIterator< double > polygonsItE = te::rst::PolygonIterator< double >::
									end(m_inputParameters.m_inputRasterPtr[0], m_inputParameters.m_polygonsSegImage[regionIdx]);

								// Getting the segment ID

								double IDSegment = 0.0;
								m_inputParameters.m_rasterSegPtr->getBand(0)->getValue(polygonsIt.getColumn(), polygonsIt.getRow(), IDSegment);
								outParamsPtr->m_IDSegment.push_back(IDSegment);

								while (polygonsIt != polygonsItE) {
									outputBand.setValue(polygonsIt.getColumn(), polygonsIt.getRow(), classMap);
									outputBandMap.setValue(polygonsIt.getColumn(), polygonsIt.getRow(), valueMap);
									outputBandMapPValue.setValue(polygonsIt.getColumn(), polygonsIt.getRow(), outParamsPtr->m_pValuesSum[regionIdx]);
									++polygonsIt;
								}

								outParamsPtr->m_valuesS.push_back(valueMap);
								++regionsIt;
								++regionIdx;
							}
						}

						// Other combinations

						else {

							double globalMinD1 = 10000000000.0;
							double globalMaxD1 = 0.0;
							double globalMinD2 = 10000000000.0;
							double globalMaxD2 = 0.0;
							double globalMinD1N = 10000000000.0;
							double globalMaxD1N = 0.0;
							double globalMinD2N = 10000000000.0;
							double globalMaxD2N = 0.0;
							double globalMinProduto = 10000000000.0;
							double globalMaxProduto = 0.0;
							double globalMinProdutoNN = 10000000000.0;
							double globalMaxProdutoNN = 0.0;

							// Finding critical statistical test values

							std::vector< double > otherCriticalS;
							for (unsigned int i = 0; i < m_inputParameters.m_sourcesNumber; ++i) {
								boost::math::chi_squared distChiSquared(m_degreesOfFreedom[i]);
								otherCriticalS.push_back(boost::math::quantile(complement(distChiSquared, m_inputParameters.m_significanceLevel)));
							}

							std::vector< double > normalizedRegions;
							std::vector< std::vector< double > > allNormalizedRegions;

							unsigned int imageNum = 0;

							std::map< unsigned int, Container >::iterator newIterator = beforeNormalizedImagesMaps.begin();
							std::map< unsigned int, Container >::iterator newIt = normalizedImagesMaps.begin();

							double globalMinValueS = 10000000000.0;
							double globalMaxValueS = 0.0;

							for (it = m_imagesMaps.begin(); it != m_imagesMaps.end(); ++it) {

								std::vector< double > tests;
								std::vector< std::vector< double > > statisticalTests;

								const Container& regionMap = it->second;

								Container::const_iterator regionsIt = regionMap.begin();
								const Container::const_iterator regionsItE = regionMap.end();

								while (regionsIt != regionsItE) {

									for (unsigned int j = 0; j < classesNumber; ++j) {

										double newValueS = regionsIt->operator[](j) - otherCriticalS[imageNum];

										tests.push_back(newValueS);

										// Finding global max and min statistical test value

										if (imageNum == 0) {
											if (regionsIt->operator[](j) < globalMinD1)
												globalMinD1 = regionsIt->operator[](j);

											if (regionsIt->operator[](j) > globalMaxD1)
												globalMaxD1 = regionsIt->operator[](j);
										}

										if (imageNum == 1) {
											if (regionsIt->operator[](j) < globalMinD2)
												globalMinD2 = regionsIt->operator[](j);

											if (regionsIt->operator[](j) > globalMaxD2)
												globalMaxD2 = regionsIt->operator[](j);
										}

										if (newValueS < globalMinValueS)
											globalMinValueS = newValueS;

										if (newValueS > globalMaxValueS)
											globalMaxValueS = newValueS;
									}

									statisticalTests.push_back(tests);
									tests.clear();

									++regionsIt;
								}

								beforeNormalizedImagesMaps.insert(newIterator, std::pair< unsigned int, Container >(imageNum, statisticalTests));
								statisticalTests.clear();
								++imageNum;
							}

							imageNum = 0;

							for (newIterator = beforeNormalizedImagesMaps.begin(); newIterator != beforeNormalizedImagesMaps.end(); ++newIterator) {

								// Normalizing statistical test values

								const Container& statisticalTests = newIterator->second;

								Container::const_iterator newRegionsIt = statisticalTests.begin();
								const Container::const_iterator newRegionsItE = statisticalTests.end();

								while (newRegionsIt != newRegionsItE) {

									for (unsigned int j = 0; j < classesNumber; ++j) {

										double newValueS = double(newRegionsIt->operator[](j) - globalMinValueS)
											/
											(globalMaxValueS - globalMinValueS);

										normalizedRegions.push_back(newValueS);

										if (imageNum == 0) {
											if (newValueS < globalMinD1N)
												globalMinD1N = newValueS;

											if (newValueS > globalMaxD1N)
												globalMaxD1N = newValueS;
										}

										if (imageNum == 1) {
											if (newValueS < globalMinD2N)
												globalMinD2N = newValueS;

											if (newValueS > globalMaxD2N)
												globalMaxD2N = newValueS;
										}
									}

									allNormalizedRegions.push_back(normalizedRegions);
									normalizedRegions.clear();

									++newRegionsIt;
								}

								normalizedImagesMaps.insert(newIt, std::pair< unsigned int, Container >(imageNum, allNormalizedRegions));
								allNormalizedRegions.clear();
								++imageNum;
							}

							// Multiplication and minimum combinations

							if ((m_inputParameters.m_combinationType == "multiplication") || (m_inputParameters.m_combinationType == "minimum")) {

								// Distances output for minimum combination
								{
								dist = m_distancesMaps.begin();
								unsigned int numberRegionsD = dist->second.size();

								std::vector< double > distancesRegions;

								for (unsigned int i = 0; i < numberRegionsD; ++i) {

									double minClassValue = 10000000000.0;

									for (unsigned int j = 0; j < classesNumber; ++j) {

          double minClassDistancePerClass = 10000000000.0;

										for (dist = m_distancesMaps.begin(); dist != m_distancesMaps.end(); ++dist) {

											const Container& regionMapD = dist->second;

											Container::const_iterator regionsDIt = regionMapD.begin();

											for (unsigned int k = 1; k <= i; ++k)
												++regionsDIt;

											// Finding non-normalized minimum distance value

           if (regionsDIt->operator[](j) < minClassValue) {

             minClassValue = regionsDIt->operator[](j);
           }

           if (regionsDIt->operator[](j) < minClassDistancePerClass) {
             minClassDistancePerClass = regionsDIt->operator[](j);
           }
										}

										distancesRegions.push_back(minClassDistancePerClass);
								}

								outParamsPtr->m_outDistancesPerClass.push_back(distancesRegions);
								distancesRegions.clear();
								outParamsPtr->m_distances.push_back(minClassValue);
							}
							}

								// Findind p-values from statistical test values without normalization
								// for generating p-value map

								double multiValue = 0.0;

								double pValue = 0.0;

								std::vector< double > everyPValue;

								it = m_imagesMaps.begin();
        unsigned int numberRegions = it->second.size();

								if (m_inputParameters.m_combinationType == "minimum") {

								  std::vector< double > pValuesRegion;

          for (unsigned int i = 0; i < numberRegions; ++i) {

           double minClassValue = 1000000.0;
           int sourceForPValueIdx = 0;

           for (unsigned int j = 0; j < classesNumber; ++j) {

            double minClassValuePerClass = 1000000.0;
            int sourceForPValueIdxPerClass = 0;

            for (it = m_imagesMaps.begin(); it != m_imagesMaps.end(); ++it) {

             int sourceIdx = 0;

             const Container& regionMap = it->second;

             Container::const_iterator regionsIt = regionMap.begin();

             for (unsigned int k = 0; k < i; ++k)
              ++regionsIt;

             // Finding non-normalized minimum statistical test value for final p-values

             if (regionsIt->operator[](j) < minClassValue) {

               minClassValue = regionsIt->operator[](j);
               sourceForPValueIdx = sourceIdx;
             }

             if (regionsIt->operator[](j) < minClassValuePerClass) {
               minClassValuePerClass = regionsIt->operator[](j);
               sourceForPValueIdxPerClass = sourceIdx;
             }

             ++sourceIdx;
            }

            // Finding p-values per class

            boost::math::chi_squared distChiSquared(m_degreesOfFreedom[sourceForPValueIdxPerClass]);
            try {
              pValuesRegion.push_back(boost::math::cdf(complement(distChiSquared, minClassValuePerClass)));
            }
            catch (...) {
              pValuesRegion.push_back(NAN);
            }
           }

           // Computing p-value

           boost::math::chi_squared distChiSquared(m_degreesOfFreedom[sourceForPValueIdx]);
           try {
             pValue = boost::math::cdf(complement(distChiSquared, minClassValue));
           }
           catch (...) {
             pValue = NAN;
           }

           outParamsPtr->m_pValuesMinimum.push_back(pValue);
           outParamsPtr->m_outPValuesPerClass.push_back(pValuesRegion);
           pValuesRegion.clear();
          }
								}

								std::vector< double > valuesAttribute;
								std::vector< double > everySValue;

								std::map< unsigned int, Container >::iterator normalizedIt = normalizedImagesMaps.begin();
								numberRegions = normalizedIt->second.size();

								for (unsigned int i = 0; i < numberRegions; ++i) {

         int sourceIdxForMinimun = 0;

									if (m_inputParameters.m_combinationType == "multiplication") {
										for (unsigned int j = 0; j < classesNumber; ++j)
											valuesAttribute.push_back(1.0);
									}

									valueMap = 1000000.0;
									classMap = 0;

									for (unsigned int j = 0; j < classesNumber; ++j) {

          double minClassValuePerClass = 1000000.0;

										for (normalizedIt = normalizedImagesMaps.begin(); normalizedIt != normalizedImagesMaps.end(); ++normalizedIt) {

											const Container& regionMap = normalizedIt->second;

											Container::const_iterator regionsIt = regionMap.begin();

											for (unsigned int k = 0; k < i; ++k)
												++regionsIt;

											// Multiplication combination

											if (m_inputParameters.m_combinationType == "multiplication")
												valuesAttribute[j] *= regionsIt->operator[](j);

											// Minimum combination

											if (m_inputParameters.m_combinationType == "minimum") {

												if (regionsIt->operator[](j) < valueMap) {
													valueMap = regionsIt->operator[](j);
													classMap = j;
												}

												if (regionsIt->operator[](j) < minClassValuePerClass)
               minClassValuePerClass = regionsIt->operator[](j);
											}
										}

										if (m_inputParameters.m_combinationType == "minimum")
										  everySValue.push_back(minClassValuePerClass);
									}

									// Multiplication combination

									double multiplicationPValue = 1.0;

									if (m_inputParameters.m_combinationType == "multiplication") {

										everySValue.push_back(valuesAttribute[0]);
										valueMap = valuesAttribute[0];
										classMap = 0;

										for (unsigned int j = 1; j < classesNumber; ++j) {

											everySValue.push_back(valuesAttribute[j]);

											// Finding minimum statistical test value

											if (valuesAttribute[j] < valueMap) {
												valueMap = valuesAttribute[j];
												classMap = j;
											}
										}

										std::vector< double > productSNotNormalized;
										for (int j = 0; j < classesNumber; ++j)
										  productSNotNormalized.push_back(1.0);

										for (it = m_imagesMaps.begin(); it != m_imagesMaps.end(); ++it) {

            const Container& regionMap = it->second;

											 Container::const_iterator regionsIt = regionMap.begin();

											 for (unsigned int k = 0; k < i; ++k)
											  	++regionsIt;

											 for (unsigned int j = 0; j < classesNumber; ++j) {

											   productSNotNormalized[j] *= regionsIt->operator[](j);
											 }
										}

										int pValueIndex = 0;
										double minProduct = productSNotNormalized[0];
										for (int j = 1; j < classesNumber; ++j) {

            if (productSNotNormalized[j] < minProduct) {
              pValueIndex = j;
              minProduct = productSNotNormalized[j];
            }
										}

										int sourceIdx = 0;

										std::vector< double > allMultiplicationPValue;
										for (int j = 0; j < classesNumber; ++j)
										  allMultiplicationPValue.push_back(1.0);

										for (it = m_imagesMaps.begin(); it != m_imagesMaps.end(); ++it) {

											const Container& regionMap = it->second;

											Container::const_iterator regionsIt = regionMap.begin();

											for (unsigned int k = 0; k < i; ++k)
												++regionsIt;

											for (int j = 0; j < classesNumber; ++j) {
             boost::math::chi_squared distChiSquared(m_degreesOfFreedom[sourceIdx]);
             try {
               allMultiplicationPValue[j] *= boost::math::cdf(complement(distChiSquared, regionsIt->operator[](j)));
             }
             catch (...) {
               allMultiplicationPValue[j] = NAN;
             }
											}

											// Finding p-values and multiplying them for multiplication combination

											boost::math::chi_squared distChiSquared(m_degreesOfFreedom[sourceIdx]);
											try {
												multiplicationPValue *= boost::math::cdf(complement(distChiSquared, regionsIt->operator[](pValueIndex)));
											}
											catch (...) {
												multiplicationPValue = NAN;
											}

											++sourceIdx;
										}

										outParamsPtr->m_pValuesMultiplication.push_back(multiplicationPValue);
										outParamsPtr->m_outPValuesPerClass.push_back(allMultiplicationPValue);
									 allMultiplicationPValue.clear();
									}

									outParamsPtr->m_outStatisticsPerClass.push_back(everySValue);
									everySValue.clear();

									// Generating classifiyng image and reliability map

									te::rst::PolygonIterator< double > polygonsIt = te::rst::PolygonIterator< double >::
										begin(m_inputParameters.m_inputRasterPtr[0], m_inputParameters.m_polygonsSegImage[i]);
									te::rst::PolygonIterator< double > polygonsItE = te::rst::PolygonIterator< double >::
										end(m_inputParameters.m_inputRasterPtr[0], m_inputParameters.m_polygonsSegImage[i]);

									double IDSegment = 0.0;
									m_inputParameters.m_rasterSegPtr->getBand(0)->getValue(polygonsIt.getColumn(), polygonsIt.getRow(), IDSegment);
									outParamsPtr->m_IDSegment.push_back(IDSegment);

									while (polygonsIt != polygonsItE) {
										outputBand.setValue(polygonsIt.getColumn(), polygonsIt.getRow(), classMap);
										outputBandMap.setValue(polygonsIt.getColumn(), polygonsIt.getRow(), valueMap);
										if (m_inputParameters.m_combinationType == "multiplication")
											outputBandMapPValue.setValue(polygonsIt.getColumn(), polygonsIt.getRow(), multiplicationPValue);

										if (m_inputParameters.m_combinationType == "minimum")
											outputBandMapPValue.setValue(polygonsIt.getColumn(), polygonsIt.getRow(), outParamsPtr->m_pValuesMinimum[i]);
										++polygonsIt;
									}

									outParamsPtr->m_valuesS.push_back(valueMap);
									valuesAttribute.clear();
								}
							}
						}
					}

					std::vector< unsigned int > classesPixels;

					for (unsigned int numClass = 0; numClass < m_inputParameters.m_numberOfClassesOnTraining; ++numClass)
						classesPixels.push_back(0);

					typedef std::vector< std::vector< double > > ClassSamplesContainerT;

					ClassSamplesContainerT::const_iterator samplesIt =
						m_inputParameters.m_polygonSamplesTestNew.begin();
					const ClassSamplesContainerT::const_iterator samplesItE =
						m_inputParameters.m_polygonSamplesTestNew.end();

					int classID = 0;

					while (samplesIt != samplesItE) {
						for (int i = 0; i < m_inputParameters.m_numberOfSamplesTest[classID]; ++i) {

							int row = (int)(samplesIt->operator[](i) / outputRaster.getNumberOfColumns());
							int column = ((int)samplesIt->operator[](i) % outputRaster.getNumberOfColumns());

							double value = 0.0;
							outputBand.getValue(column, row, value);

							++classesPixels[value];
						}

						outParamsPtr->m_confusionMatrix.push_back(classesPixels);
						classesPixels.clear();
						for (unsigned int numClass = 0; numClass < m_inputParameters.m_numberOfClassesOnTraining; ++numClass)
							classesPixels.push_back(0);

						++classID;
						++samplesIt;
					}

					// Validating results with test samples and computing the confusion matrix

					/*double classOfPixel = 0.0;

					std::vector< unsigned int > classesPixels;

					for (unsigned int numClass = 0; numClass < m_inputParameters.m_numberOfClassesOnTraining; ++numClass)
						classesPixels.push_back(0);

					for (unsigned int i = 0; i < m_inputParameters.m_polygonSamplesTest.size(); ++i) {

						int numPixels = 0;

						te::gm::Polygon* polygon = &m_inputParameters.m_polygonSamplesTest[i];

						te::rst::PolygonIterator< double > polygonsIt = te::rst::PolygonIterator< double >::
							begin(&outputRaster, polygon);
						te::rst::PolygonIterator< double > polygonsItE = te::rst::PolygonIterator< double >::
							end(&outputRaster, polygon);

						while (polygonsIt != polygonsItE) {

							//outputRaster.getValue(polygonsIt.getColumn(), polygonsIt.getRow(), classOfPixel, 0);
							outputBand.getValue(polygonsIt.getColumn(), polygonsIt.getRow(), classOfPixel);

							//++classesPixels[(unsigned int) classOfPixel];
							++classesPixels[classOfPixel];

							++polygonsIt;
							++numPixels;
						}

						classOfPixel = 0.0;

						if ((i + 1) != m_inputParameters.m_polygonSamplesTest.size()) {
							if (m_inputParameters.m_classesIDTest[i] != m_inputParameters.m_classesIDTest[i + 1]) {
								outParamsPtr->m_confusionMatrix.push_back(classesPixels);
								classesPixels.clear();
								for (unsigned int numClass = 0; numClass < m_inputParameters.m_numberOfClassesOnTraining; ++numClass)
									classesPixels.push_back(0);
							}
						}
						else {
							outParamsPtr->m_confusionMatrix.push_back(classesPixels);
						}
					}*/

					// Calculating Kappa coefficient

					unsigned int sumOfDiagonal = 0;
					unsigned int numberOfSamples = 0;
					unsigned int sumOfMultiplications = 0;
					std::vector< unsigned int > sumOfRows, sumOfColumns;

					for (unsigned int i = 0; i < m_inputParameters.m_numberOfClassesOnTraining; ++i) {
						sumOfDiagonal += outParamsPtr->m_confusionMatrix[i][i];
						sumOfRows.push_back(0);
						sumOfColumns.push_back(0);
						for (unsigned int j = 0; j < m_inputParameters.m_numberOfClassesOnTraining; ++j) {
							sumOfRows[i] += outParamsPtr->m_confusionMatrix[i][j];
							sumOfColumns[i] += outParamsPtr->m_confusionMatrix[j][i];
						}

						numberOfSamples += sumOfRows[i];
						sumOfMultiplications += sumOfRows[i] * sumOfColumns[i];
					}

					outParamsPtr->m_kappaCoefficient = (double)((numberOfSamples * sumOfDiagonal) - sumOfMultiplications)
						/
						(pow(numberOfSamples, 2) - sumOfMultiplications);

					// Calculating the variance of Kappa coefficient

					double sigma1 = (double)sumOfDiagonal / numberOfSamples;
					double sigma2 = (double)sumOfMultiplications / pow(numberOfSamples, 2);
					double sigma3 = 0.0;
					double sigma4 = 0.0;

					for (unsigned int i = 0; i < m_inputParameters.m_numberOfClassesOnTraining; ++i) {
						sigma3 += (double)outParamsPtr->m_confusionMatrix[i][i] * (sumOfRows[i] + sumOfColumns[i]);
						for (unsigned int j = 0; j < m_inputParameters.m_numberOfClassesOnTraining; ++j) {
							sigma4 += (double)outParamsPtr->m_confusionMatrix[i][j] * pow((sumOfRows[j] + sumOfColumns[i]), 2);
						}
					}
					sigma4 /= (double)pow(numberOfSamples, 3);
					sigma3 /= (double)pow(numberOfSamples, 2);

					double aux1 = 1 - sigma1;
					double aux2 = 1 - sigma2;

					outParamsPtr->m_varianceOfKappa = (double)(((sigma1 * (aux1)) / pow((aux2), 2))
						+
						((2 * (aux1)* ((2 * sigma1 * sigma2) - sigma3)) / pow(aux2, 3))
						+
						((pow(aux1, 2) * (sigma4 - (4 * pow(sigma2, 2)))) / pow(aux2, 4))) / numberOfSamples;

					// Calculating the overall accuracy

					outParamsPtr->m_overallAccuracy = 0.0;

					for (unsigned int i = 0; i < m_inputParameters.m_numberOfClassesOnTraining; ++i)
						outParamsPtr->m_overallAccuracy += ((double)outParamsPtr->m_confusionMatrix[i][i] / numberOfSamples);
				}

				return true;
			}

			void MultiSourceClassifier::reset() {
				m_inputParameters.reset();
				m_isInitialized = false;
			}

			bool MultiSourceClassifier::initialize(const te::rp::AlgorithmInputParameters& inputParams)
			{
				reset();

				MultiSourceClassifier::InputParameters const* inputParamsPtr = dynamic_cast<
					MultiSourceClassifier::InputParameters const* >(&inputParams);
				TERP_TRUE_OR_RETURN_FALSE(inputParamsPtr, "Invalid paramters pointer");

				m_inputParameters = *inputParamsPtr;

				// Checking the input parameters

				TERP_TRUE_OR_RETURN_FALSE(m_inputParameters.m_sourcesNumber > 0, "Invalid sources number");

				TERP_TRUE_OR_RETURN_FALSE(m_inputParameters.m_inputRasterPtr.size() > 0,
					"Invalid m_inputRasterPtr")

        TERP_TRUE_OR_RETURN_FALSE( m_inputParameters.m_imageType.size() ==
          m_inputParameters.m_sourcesNumber, "Invalid number of data models" );
        TERP_TRUE_OR_RETURN_FALSE( m_inputParameters.m_inputRasterPtr.size() ==
          m_inputParameters.m_sourcesNumber, "Invalid number of input rasters" );
        TERP_TRUE_OR_RETURN_FALSE( m_inputParameters.m_inputRastersBands.size() ==
          m_inputParameters.m_sourcesNumber, "Invalid number of input rasters bands" );
        TERP_TRUE_OR_RETURN_FALSE( m_inputParameters.m_vectorOrMatrixOrder.size() ==
          m_inputParameters.m_sourcesNumber, "Invalid number of lexicographic vector/covariance matrix elements" );
        TERP_TRUE_OR_RETURN_FALSE( m_inputParameters.m_radarDataRepresentationType.size() ==
          m_inputParameters.m_sourcesNumber, "Invalid number of radar data representation elements" );
        TERP_TRUE_OR_RETURN_FALSE( m_inputParameters.m_ENL.size() ==
          m_inputParameters.m_sourcesNumber, "Invalid number of Equivalent Number of Looks elements" );
        TERP_TRUE_OR_RETURN_FALSE( m_inputParameters.m_betaValue.size() ==
          m_inputParameters.m_sourcesNumber, "Invalid number of Beta values elements" );

					unsigned int sizeMatch = 0;

				for (unsigned int imageNum = 0; imageNum < m_inputParameters.m_sourcesNumber; ++imageNum) {

					TERP_TRUE_OR_RETURN_FALSE(m_inputParameters.m_inputRasterPtr[imageNum], "Invalid m_inputRasterPtr")

						if ((m_inputParameters.m_inputRasterPtr[imageNum]->getNumberOfColumns() != m_inputParameters.m_numberOfColumns)
							||
							(m_inputParameters.m_inputRasterPtr[imageNum]->getNumberOfRows() != m_inputParameters.m_numberOfRows)) {
							sizeMatch = 1;
						}

					TERP_TRUE_OR_RETURN_FALSE(m_inputParameters.m_inputRastersBands[imageNum].size() > 0,
						"Invalid raster bands number");

					if (m_inputParameters.m_imageType[imageNum] == "Gamma")
						TERP_TRUE_OR_RETURN_FALSE(m_inputParameters.m_inputRastersBands[imageNum].size() == 1,  //m_inputRasterPtr->getNumberOfBands()
						"Invalid m_inputRasterPtr Gamma Model number bands")

						if (m_inputParameters.m_imageType[imageNum] == "Wishart") {
							TERP_TRUE_OR_RETURN_FALSE(m_inputParameters.m_vectorOrMatrixOrder.size() > 0,
								"Invalid lexicographic vector or covariance matrix order");

							TERP_TRUE_OR_RETURN_FALSE(m_inputParameters.m_vectorOrMatrixOrder[imageNum] > 1,
								"Invalid lexicographic vector or covariance matrix order");
						}

					if (m_inputParameters.m_imageType[imageNum] == "Wishart" || m_inputParameters.m_imageType[imageNum] == "Gamma")
						TERP_TRUE_OR_RETURN_FALSE(m_inputParameters.m_betaValue[imageNum] >= 0 && m_inputParameters.m_betaValue[imageNum] <= 1,
						"Beta value must be between 0 and 1.")

						TERP_TRUE_OR_RETURN_FALSE(m_inputParameters.m_inputRasterPtr[imageNum]->getSRID() >= 0,
						"Invalid inputRasterPtr SRID")

						TERP_TRUE_OR_RETURN_FALSE(m_inputParameters.m_inputRasterPtr[imageNum]->getSRID() == m_inputParameters.m_SRID,
						"InputRasterPtr SRID mismatch")
				}

				TERP_TRUE_OR_RETURN_FALSE(sizeMatch == 0, "Input and segmented size image mismatch");

				TERP_TRUE_OR_RETURN_FALSE(m_inputParameters.m_lagX >= 0, "Invalid lag X value")

					TERP_TRUE_OR_RETURN_FALSE(m_inputParameters.m_lagY >= 0, "Invalid lag Y value")

					if (m_inputParameters.m_combinationType == "fuzzy")
						TERP_TRUE_OR_RETURN_FALSE(m_inputParameters.m_sourcesNumber > 1, "Input image number must be > 1 in fuzzy combination");

				m_degreesOfFreedom.clear();

				m_isInitialized = true;

				return true;
			}

			bool MultiSourceClassifier::isInitialized() const {
				return m_isInitialized;
			}

		} // end namespace radar
	} // end namespace rp
}   // end namespace msc
