/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/rp/radar/MultiSourceCLassifier.h
  \brief Performs multi source image classification.
 */

#ifndef __TERRALIB_RP_RADAR_INTERNAL_MULTISOURCECLASSIFIER_H
#define __TERRALIB_RP_RADAR_INTERNAL_MULTISOURCECLASSIFIER_H

#include "../Config.h"
#include "MultiSourceClassifierInputParameters.h"
#include "MultiSourceClassifierOutputParameters.h"

//#include <terralib/raster.h>
#include "../Algorithm.h"

#include <boost/numeric/ublas/matrix.hpp>

#include <vector>
#include <map>

namespace te
{
  namespace rp
  {
    namespace radar
    {
     typedef std::complex< double > WishartFeatureType;
      /*!
      \class MultiSourceClassifier
      \brief Performs multi source image classification.
      \details Reference: B. C. Braga, C. d. C. Freitas and S. J. S. Sant'Anna, "Multisource classification based on uncertainty maps," 2015 IEEE International Geoscience and Remote Sensing Symposium (IGARSS), Milan, Italy, 2015, pp. 1630-1633, doi: 10.1109/IGARSS.2015.7326097.
      */
    class TERPEXPORT MultiSourceClassifier : public te::rp::Algorithm {
      public:

        typedef MultiSourceClassifierInputParameters InputParameters;
        typedef MultiSourceClassifierOutputParameters OutputParameters;

        MultiSourceClassifier();

        ~MultiSourceClassifier();

        // overload
        bool execute(te::rp::AlgorithmOutputParameters& outputParams);

        // overload
        void reset();

        // overload
        bool initialize(const te::rp::AlgorithmInputParameters& inputParams);

        bool isInitialized() const;

       // bool loadSamples(te::rp::radar::MultiSourceClassifier::InputParameters::ClassSampleT,
         //                te::rp::radar::MultiSourceClassifier::InputParameters::ClassSamplesContainerT);

      protected:
        MultiSourceClassifier::InputParameters m_inputParameters;  //!< Input execution parameters.

        MultiSourceClassifier::OutputParameters m_outputParameters;  //!< Output execution parameters.

        bool m_isInitialized;  //!< Tells if this instance is initialized.

        std::vector< std::vector< double > > m_stochasticDistance; //!< Stochastic distances container.

        std::vector< std::vector< double > > m_statisticalTest; //!< Statistical tests container.

        std::vector< std::vector< double > > m_pValue; //!< P-values container.

        std::vector< unsigned int > m_degreesOfFreedom; //!< Degrees of freedom container.

        std::map< unsigned int, std::vector< std::vector< double > > > m_imagesMaps;

        std::map< unsigned int, std::vector< std::vector< double > > > m_distancesMaps;
    };
    } // end namespace radar
  } // end namespace rp
}   // end namespace msc

#endif
