/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

#ifndef __TERRALIB_RP_RADAR_INTERNAL_MULTISOURCECLASSIFIER_INPUT_PARAMETERS_H
#define __TERRALIB_RP_RADAR_INTERNAL_MULTISOURCECLASSIFIER_INPUT_PARAMETERS_H

#include "../Config.h"
#include "../../raster/Raster.h"
#include "../../raster/Grid.h"
#include "../AlgorithmInputParameters.h"

// Boost includes
#include <boost/numeric/ublas/matrix.hpp>

#include <vector>
#include <string>

namespace te
{
  namespace rp
  {
    class AlgorithmParametersSerializer;

    namespace radar
    {
      /*!
        \class MultiSourceClassifierInputParameters
        \brief MultiSourceClassifier input parameters
      */
      class TERPEXPORT MultiSourceClassifierInputParameters : public te::rp::AlgorithmInputParameters
      {
        public:
          
          typedef unsigned int ClassIDT;  //!< CLass ID type definiton.          

          // General input parameters
          std::vector< te::rst::Raster* > m_inputRasterPtr;  //!< Input rasters.

          std::vector< std::vector< unsigned int > > m_inputRastersBands;  //!< Bands to process.

          // // Regions image parameters
          unsigned int m_numberOfColumns;  //!< Number of columns that all input rasters must have.

          unsigned int m_numberOfRows;  //!< Number of rows that all input rasters must have.

          int m_SRID;  //!< The SRID of raster data;

          te::rst::Grid* m_extent; //!< The geographic extension of the raster data.

          std::vector< te::gm::Polygon* > m_polygonsSegImage;  //!< Polygons of segmented image.

          te::rst::Raster const* m_rasterSegPtr;  //!< Input segmented raster.

          unsigned int m_sourcesNumber;   //!< The number of different image sources.

          std::vector< std::string > m_imageType;  //!< The used data model (one for each input image) (Gaussian,Gamma,IntensityPair,Wishart).

          std::vector< std::string > m_stochasticDistanceType;  //!< The stochastic distance type (Bhattacharyya,Triangular).

          bool m_savePerClassDistancesFile;  //!< The flag to indicate if stochastic distance values per class will be saved on disk.

          bool m_savePerClassStatiticsFile;  //!< The flag to indicate if statistic test values per class will be saved on disk.

          bool m_savePerClassPValueFile;  //!< The flag to indicate if p-values per class will be saved on disk.

          bool m_saveDistances;  //!< The flag to indicate if stochastic distance values will be saved on disk.

          bool m_saveStatitics;  //!< The flag to indicate if statistic test values will be saved on disk.

          bool m_savePValues;  //!< The flag to indicate if p-values will be saved on disk.

          bool m_saveReportFile;  //!< The flag to indicate if the report will be saved on disk.

          std::string m_combinationType;    //!< How the build the multi-source classification procedure (sum, multiplication,minimum,Hellinger,fuzzy).

          unsigned int m_lagX;  //!< The horizontal correlation value.

          unsigned int m_lagY;  //!< The vertical correlation value.

          double m_significanceLevel;   //!< The significance level.

          // Radar input parameters
          std::vector< std::string > m_radarDataRepresentationType; //!< The radar data representation container.

          std::vector< unsigned int > m_vectorOrMatrixOrder; //!< The lexicographic vector and covariance matrix order container.

          std::vector< double > m_ENL; //!< Equivalent Number of Looks container.

          std::vector< double > m_betaValue; //!< Beta values for Renyi distance container.

          // Samples input
          std::map< std::string, std::string > m_dirTrainingSamples;  //!< Training samples data directory.

          std::map< std::string, std::string > m_dirTestSamples;  //!< Test samples data directory.

          std::vector< te::gm::Polygon > m_polygonSamplesT;  //!< Training samples polygons.
          std::vector< std::vector< double > > m_polygonSamplesTNew;  //!< Training samples polygons.
          std::vector< std::vector< double > > m_polygonSamplesTestNew;
          std::vector< int > m_numberOfSamples;
          std::vector< int > m_numberOfSamplesTest;
          std::vector< std::string > m_nameOfSamples;
          std::vector< std::string > m_nameOfSamplesTest;

          std::vector< unsigned int > m_classesIDT;  //!< Training class ID.

          std::vector< te::gm::Polygon > m_polygonSamplesTest;  //!< Test samples polygons.

          std::vector< unsigned int > m_classesIDTest;  //!< Test class ID.

          std::vector< unsigned int > m_classesIDT2;  //!< Class ID.

          unsigned int m_numberOfClassesOnTraining;  //!< Number of classes on training.

          unsigned int m_numberOfClassesOnTest;  //!< Number of classes on test.

          MultiSourceClassifierInputParameters();

          MultiSourceClassifierInputParameters(const MultiSourceClassifierInputParameters&);

          ~MultiSourceClassifierInputParameters();

          // overload
          void reset();

          // overload
          const  MultiSourceClassifierInputParameters& operator=(const MultiSourceClassifierInputParameters& params);

          // overload
          AbstractParameters* clone() const;

          // overload
          bool serialize( te::rp::AlgorithmParametersSerializer& serializer ) const;
      };
    } // end namespace radar
  } // end namespace rp
}   // end namespace msc

#endif
