/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/rp/radar/RadarFunctions.cpp

  \brief Radar Raster Processing functions.
*/

#include "RadarFunctions.h"

#include "../../common/StringUtils.h"
#include "../../dataaccess/datasource/DataSource.h"
#include "../../dataaccess/datasource/DataSourceFactory.h"
#include "../../dataaccess/datasource/DataSourceTransactor.h"
#include "../../geometry/MultiPolygon.h"
#include "../../raster/RasterFactory.h"

#include <boost/filesystem.hpp>
#include <boost/lexical_cast.hpp>

#include <fstream>
#include <memory>

namespace te
{
  namespace rp
  {
    namespace radar
    {
      bool LoadSamplesFile(
        const std::string& fileName,
        unsigned int& loadedNumberOfClasses,
        std::vector< unsigned int > loadedClassesID,
        std::vector< std::string >& loadedNameOfSamples,
        std::vector< int >& loadedNumberOfSamples,
        std::vector< std::vector< double > >& loadedSamples,
        std::vector< te::gm::Polygon >& loadedPolygons )
      {
        loadedNumberOfClasses = 0;
        loadedClassesID.clear();
        loadedNameOfSamples.clear();
        loadedNumberOfSamples.clear();
        loadedSamples.clear();
        loadedPolygons.clear();

        const std::string extension = te::common::Convert2UCase(
          boost::filesystem::path( fileName ).extension().string() );

        if( extension == ".TXT" )
        {
          std::ifstream fileInstance( fileName );
          if( ! fileInstance.good() )
          {
            return false;
          }

          fileInstance >> loadedNumberOfClasses;

          std::string auxName;
          double auxDoubleNumber = 0;
          double auxIntNumber = 0;

          for (unsigned int i = 0; i < loadedNumberOfClasses; ++i)
          {
            fileInstance >> auxName;
            loadedNameOfSamples.push_back(auxName);
            fileInstance >> auxIntNumber;
            loadedNumberOfSamples.push_back(auxIntNumber);
          }

          std::vector< double > samplesAddresses;
          for (unsigned int i = 0; i < loadedNumberOfClasses; ++i)
          {
            samplesAddresses.clear();

            for (int j = 0; j < loadedNumberOfSamples[i]; ++j)
            {
              fileInstance >> auxDoubleNumber;
              samplesAddresses.push_back(auxDoubleNumber);
            }

            loadedSamples.push_back(samplesAddresses);
          }
        }
        else if( extension == ".SHP" )
        {
          std::unique_ptr< te::da::DataSource > ds =
            te::da::DataSourceFactory::make("OGR", "file://" +
            fileName );
          ds->open();
          if( ! ds->isValid() )
          {
            return false;
          }

          std::unique_ptr< te::da::DataSourceTransactor > transactor =
            (ds->getTransactor());
          std::vector< std::string > datasets = ds->getDataSetNames();
          loadedNumberOfClasses = datasets.size();

          srand( (unsigned)time(NULL) );
          unsigned int ind = 0;

          for(unsigned int i = 0; i < datasets.size(); ++i)
          {
            std::unique_ptr< te::da::DataSet > dataset = (ds->getDataSet(datasets[i]));

            dataset->moveBeforeFirst();

            size_t nproperties = dataset->getNumProperties();

            while(dataset->moveNext())
            {
              for(std::size_t k = 0; k < nproperties; ++k) {

                // check if value is not null
                if(dataset->isNull(k)) {
                  continue;
                }

                std::string a = dataset->getPropertyName(k);

                if(!a.compare("CODCLASS")) {
                  std::string sval = dataset->getNumeric(a);
                  if(sval.empty())
                    sval = "0.0";

                  unsigned int dval = boost::lexical_cast<unsigned int>(sval);
                  loadedClassesID.push_back(dval);
                  continue;
                }

                if(dataset->getPropertyDataType(k) == te::dt::GEOMETRY_TYPE) {
                  std::unique_ptr<te::gm::Geometry> gpr = dataset->getGeometry(k);

                  if(gpr->getGeomTypeId() == te::gm::PolygonType) {
                    te::gm::Polygon* polygon = dynamic_cast< te::gm::Polygon* >(gpr.get());

                    loadedPolygons.push_back( *polygon );
                    loadedClassesID.push_back(ind);
                  }
                  else {
                    if(gpr->getGeomTypeId() == te::gm::MultiPolygonType) {
                      te::gm::MultiPolygon* mPolygon =
                        dynamic_cast< te::gm::MultiPolygon* >(
                        gpr.get());
                      size_t n_geom = mPolygon->getNumGeometries();

                      for(std::size_t n = 0; n < n_geom; ++n) {
                        te::gm::Polygon* polygon = dynamic_cast<te::gm::Polygon*>(mPolygon->getGeometryN(n));
                        loadedPolygons.push_back( *polygon );
                        loadedClassesID.push_back(ind);
                      }

                    }
                  }
                }
              }
            }

            if(loadedPolygons.size()) ind++;
          }
        }
        else
        {
          return false;
        }

        return true;
      }

      bool SaveOutputReportFile(
        const MultiSourceClassifierInputParameters& inputParameters,
        const MultiSourceClassifierOutputParameters& outputParameters,
        const std::string& reportFileName )
      {
        if( reportFileName.empty() ) return false;

        std::vector< int > modelImage;
        for( std::size_t imageTypeIdx = 0 ; imageTypeIdx <
          inputParameters.m_imageType.size() ; ++imageTypeIdx )
        {
          if(
              (inputParameters.m_imageType[ imageTypeIdx] == "IntensityPair")
              ||
              (inputParameters.m_imageType[ imageTypeIdx] == "Gamma")
            )
          {
            modelImage.push_back(1);
          }
          else if(inputParameters.m_imageType[ imageTypeIdx] == "Wishart")
          {
            modelImage.push_back(2);
          }
          else
          {
            modelImage.push_back(0);
          }
        }

        std::ofstream outFile(reportFileName);
        if( !outFile.good() ) return false;

        outFile << "Classification parameters:" << std::endl;
        outFile << "      Number of Input Images: " << inputParameters.m_sourcesNumber << std::endl;
        if (inputParameters.m_sourcesNumber > 1)
           outFile << "      Combination type: " << inputParameters.m_combinationType << std::endl;
        outFile << "      Confidence Level: " << 100 - (inputParameters.m_significanceLevel * 100) << "%" << std::endl;
        outFile << "      Horizontal lag: " << inputParameters.m_lagX << std::endl;
        outFile << "      Vertical lag: " << inputParameters.m_lagY << std::endl;
        outFile << std::endl;
        for (int i = 0; i < inputParameters.m_sourcesNumber; ++i)
        {
          outFile << "      Input Image " << i + 1 << ": " << std::endl;
          outFile << "           Statistical Model: " << inputParameters.m_imageType[i] << std::endl;
          outFile << "           Stochastic distance: " << inputParameters.m_stochasticDistanceType[i] << std::endl;
          if (inputParameters.m_stochasticDistanceType[i] == "Renyi")
            outFile << "           Beta: " << inputParameters.m_betaValue[i] << std::endl;
          if (modelImage[i] != 0) {
            outFile << "           ENL: " << inputParameters.m_ENL[i] << std::endl;
            if (modelImage[i] == 2)
              outFile << "           Covariance matrix order: " << inputParameters.m_vectorOrMatrixOrder[i] << std::endl;
          }
           outFile << std::endl;
        }
        outFile << std::endl;
        outFile << std::endl;

        if ((inputParameters.m_combinationType == "fuzzy") && (inputParameters.m_sourcesNumber > 1))
        {
          outFile << "Fuzzy module does not have statistic results." << std::endl;
          outFile << std::endl;
          outFile << std::endl;
        }

        outFile << "Classes:" << std::endl;
        for (int i = 0; i < inputParameters.m_numberOfClassesOnTraining; i++)
          outFile << inputParameters.m_nameOfSamples[i] << std::endl;
        outFile << std::endl;
        outFile << std::endl;

        outFile << "Confusion matrix:" << std::endl;
        outFile << "Rows - actual class; Columns - classification" << std::endl;
        for (unsigned int i = 0; i < inputParameters.m_numberOfClassesOnTraining; ++i) {
          for (unsigned int j = 0; j < inputParameters.m_numberOfClassesOnTraining; ++j)
            outFile << outputParameters.m_confusionMatrix[i][j] << " ";
          outFile << std::endl;
        }

        outFile << std::endl;
        outFile.precision(4);
        outFile << "Kappa coefficient: " << outputParameters.m_kappaCoefficient << std::endl;

        outFile << std::endl;
        outFile << "Variance of Kappa coefficient: " << outputParameters.m_varianceOfKappa << std::endl;

        outFile << std::endl;
        outFile.precision(4);
        outFile << "Overall accuracy: " << 100 * outputParameters.m_overallAccuracy << "%" << std::endl;

        return true;
      }

      bool SaveOutputAttFile(
        const MultiSourceClassifierInputParameters& inputParameters,
        const MultiSourceClassifierOutputParameters& outputParameters,
        const bool exportStatData,
        const bool exportPValuesData,
        const bool exportDistData,
        const std::string& attributesFileName )
      {
        if( attributesFileName.empty() ) return false;

        std::ofstream outFile( attributesFileName );
        if( !outFile.good() ) return false;

        outFile << "ID SEGMENT";
        if ((inputParameters.m_sourcesNumber == 1)
          ||
          ((
              (inputParameters.m_sourcesNumber > 1))
              &&
              (inputParameters.m_combinationType == "sum"))
          ||
          ((
            inputParameters.m_sourcesNumber > 1)
            &&
            (inputParameters.m_combinationType == "minimum"))
           )
        {
          if (exportDistData)
            outFile << "   " << "STOCHASTIC DISTANCE";
        }
        if (exportStatData)
          outFile << "   " << "STATISTICAL VALUE";
        if (exportPValuesData)
          outFile << "   " << "P-VALUE";
        outFile << std::endl;

        for (unsigned int i = 0; i < outputParameters.m_valuesS.size(); ++i) {

          if (inputParameters.m_sourcesNumber == 1) {

            outFile << outputParameters.m_IDSegment[i];
            if (exportDistData) {
              outFile.precision(10);
              outFile << " " << outputParameters.m_distances[i];
            }
            if (exportStatData) {
              outFile.precision(10);
              outFile << " " << outputParameters.m_valuesS[i];
            }
            if (exportPValuesData) {
              outFile.precision(10);
              outFile << " " << outputParameters.m_valuesP[i];
            }
            outFile << std::endl;
          }

          if ((inputParameters.m_combinationType == "sum") && (inputParameters.m_sourcesNumber > 1)) {
            outFile << outputParameters.m_IDSegment[i];
            if (exportDistData) {
              outFile.precision(10);
              outFile << " " << outputParameters.m_distances[i];
            }
            if (exportStatData) {
              outFile.precision(10);
              outFile << " " << outputParameters.m_valuesS[i];
            }
            if (exportPValuesData) {
              outFile.precision(10);
              outFile << " " << outputParameters.m_pValuesSum[i];
            }
            outFile << std::endl;
          }

          if ((inputParameters.m_combinationType == "Hellinger") && (inputParameters.m_sourcesNumber > 1)) {
            outFile << outputParameters.m_IDSegment[i];
            if (exportStatData) {
              outFile.precision(10);
              outFile << " " << outputParameters.m_valuesS[i];
            }
            outFile << std::endl;
          }

          if ((inputParameters.m_combinationType == "multiplication") && (inputParameters.m_sourcesNumber > 1)) {
            outFile << outputParameters.m_IDSegment[i];
            if (exportStatData) {
              outFile.precision(10);
              outFile << " " << outputParameters.m_valuesS[i];
            }
            if (exportPValuesData) {
              outFile.precision(10);
              outFile << " " << outputParameters.m_pValuesMultiplication[i];
            }
            outFile << std::endl;
          }

          if ((inputParameters.m_combinationType == "minimum") && (inputParameters.m_sourcesNumber > 1)) {
            outFile << outputParameters.m_IDSegment[i];
            if (exportDistData) {
              outFile.precision(10);
              outFile << " " << outputParameters.m_distances[i];
            }
            if (exportStatData) {
              outFile.precision(10);
              outFile << " " << outputParameters.m_valuesS[i];
            }
            if (exportPValuesData) {
              outFile.precision(10);
              outFile << " " << outputParameters.m_pValuesMinimum[i];
            }
            outFile << std::endl;
          }

          if ((inputParameters.m_combinationType == "fuzzy") && (inputParameters.m_sourcesNumber > 1)) {
            outFile << outputParameters.m_IDSegment[i];
            if (exportStatData) {
              outFile.precision(10);
              outFile << " " << outputParameters.m_valuesS[i];
            }
            if (exportPValuesData) {
              outFile.precision(10);
              outFile << " " << outputParameters.m_valuesP[i];
            }
            outFile << std::endl;
          }
        }

        return true;
      }
    } // end namespace radar
  } // end namespace rp
}   // end namespace msc
