/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

This file is part of the TerraLib - a Framework for building GIS enabled applications.

TerraLib is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

TerraLib is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with TerraLib. See COPYING. If not, write to
TerraLib Team at <terralib-team@terralib.org>.
*/

/*!
\file terralib/rp/radar/WishartModel.cpp
\brief Wishart model data manipulation.
*/

// TerraLib includes
#include "WishartModel.h"
#include "../Macros.h"
#include "../../raster/PositionIterator.h"
#include "../../raster/Grid.h"
#include "../../raster/BandProperty.h"
#include "../../common/MatrixUtils.h"
#include "../../raster/RasterFactory.h"

#include <cmath>
#include <fstream>
#include <memory>

namespace te {
	namespace rp {
		namespace radar {

			WishartModel::WishartModel() {
			}

			WishartModel::~WishartModel() {
			}

			// Checking covariance matrix

			boost::numeric::ublas::matrix< WishartFeatureType > WishartModel::checkCovarianceMatrix(boost::numeric::ublas::matrix< WishartFeatureType > covMatrix) {

				WishartFeatureType value;
				bool status = false;

				value = WishartFeatureType(0.00000000001, 0.00000000001);

				// getting the main diagonal first non zero value

				for (unsigned int i = 0; i < covMatrix.size1(); ++i) {
					if ((std::real(covMatrix(i, i))) != 0 && (std::imag(covMatrix(i, i)) != 0)) {
						value = covMatrix(i, i);
						break;
					}
				}

				// finding zero values and changing them in main diagonal

				for (unsigned int i = 0; i < covMatrix.size1(); ++i) {
					if ((std::real(covMatrix(i, i))) != 0 && (std::imag(covMatrix(i, i)) != 0)) {
						covMatrix(i, i) = value;
						status = true;
					}
				}

				// setting value 1.0 in main diagonal

				if (!status) {
					for (unsigned int i = 0; i < covMatrix.size1(); ++i) {
						covMatrix(i, i) = value;
						status = true;
					}
				}

				return covMatrix;
			}

			// Creating covariance raster from lexicografic vector

			bool WishartModel::CreatingCovarianceRaster() {

				std::unique_ptr<te::rst::Raster> outputRasterPtr;
				// creating the output raster

				{
					std::unique_ptr<te::rst::Grid>
						outputGridPtr(new te::rst::Grid(*m_inputRasterPtr->getGrid()));

					// use the properties from the first band to create the output bands

					te::rst::BandProperty*
						inputBandProperty(new te::rst::BandProperty(*(m_inputRasterPtr->getBand
						(0)->getProperty())));

					std::vector< te::rst::BandProperty* > bandsProperties;

					for (size_t i = 0; i < m_inputRastersBands.size() * m_inputRastersBands.size(); ++i) {
						te::rst::BandProperty* bandProperty
							(new te::rst::BandProperty(*inputBandProperty));

						if (i != 0) {
							bandProperty->m_colorInterp = te::rst::UndefCInt;
						}
						else {
							bandProperty->m_colorInterp = te::rst::GrayIdxCInt;
						}
						bandsProperties.push_back(bandProperty);
					}

					std::map< std::string, std::string > outputRasterInfo;
					outputRasterInfo["URI"] = "/scatteringVector2CovMatrix.tif";
					outputRasterPtr.reset(te::rst::RasterFactory::make("Expansible",
						outputGridPtr.release(),
						bandsProperties, outputRasterInfo, 0, 0));
					if (outputRasterPtr.get() == 0) {
						return false;
					}
				}

				std::vector< WishartFeatureType > inputComplex(m_inputRastersBands.size());
				std::vector< WishartFeatureType > conjugatedComplex(m_inputRastersBands.size());

				// for each pixel (line,column), compute the values

				for (unsigned int j = 0; j < m_numberOfRows; ++j) {
					for (unsigned int k = 0; k < m_numberOfColumns; ++k) {

						// get the input values

						for (size_t i = 0; i < m_inputRastersBands.size(); ++i) {
							std::complex< double > inputComplexValue;
							m_inputRasterPtr->getBand(m_inputRastersBands[i])->getValue(k, j, inputComplexValue);
							inputComplex[i] = inputComplexValue;
							conjugatedComplex[i] = std::conj(inputComplexValue);
						}

						for (size_t l = 0; l < inputComplex.size(); ++l) {
							for (size_t m = 0; m < conjugatedComplex.size(); ++m) {
								size_t pos = (m_inputRastersBands.size() * l) + m;

								std::complex<double> value = inputComplex[l] * conjugatedComplex[m];
								outputRasterPtr->getBand(pos)->setValue(k, j, value);
							}
						}
					}
				}

				return true;
			}

			// Loading samples
			bool WishartModel::LoadingSamples() {

				/*MClassesSamplesCTPtr allClassesSamples(new MClassesSamplesCT());
				ClassSamplesContainerT classSamples;
				ClassSampleT samples;

				std::complex< double > pixelValue;

				for(unsigned int npol = 0; npol < m_polygonSamplesT.size(); npol++) {

				// create iterators for band and polygon
				te::rst::PolygonIterator< double > it = te::rst::PolygonIterator< double >::
				begin(m_inputRasterPtr, &m_polygonSamplesT[npol]);
				const te::rst::PolygonIterator< double > itend = te::rst::PolygonIterator< double >::
				end(m_inputRasterPtr, &m_polygonSamplesT[npol]);

				while (it != itend) {

				unsigned int column = it.getColumn();
				unsigned int row = it.getRow();

				if (((row % (m_lagY + 1)) == 0) && ((column % (m_lagX + 1)) == 0)) {  // De-correlating samples input

				for(unsigned int band = 0; band < m_inputRastersBands.size(); band++) { //algoInputParams.m_inputRasterPtr[j]->getNumberOfBands()
				m_inputRasterPtr->getValue(it.getColumn(), it.getRow(), pixelValue, band);
				samples.push_back(pixelValue);
				}

				classSamples.push_back(samples);
				samples.clear();
				}

				++it;
				}

				if((npol + 1) == m_polygonSamplesT.size()) {
				allClassesSamples->insert(MClassesSamplesCT::value_type(m_classesIDT[npol], classSamples));
				continue;
				}

				else if(m_classesIDT[npol] != m_classesIDT[npol + 1]) {
				allClassesSamples->insert(MClassesSamplesCT::value_type(m_classesIDT[npol], classSamples));
				classSamples.clear();
				}
				}

				m_trainSamplesPtr = allClassesSamples;

				TERP_TRUE_OR_RETURN_FALSE(m_trainSamplesPtr,
				"Invalid Wishart classes training samples pointer")
				TERP_TRUE_OR_RETURN_FALSE(m_trainSamplesPtr->size() > 0,
				"Invalid Wishart classes training samples number")
				TERP_TRUE_OR_RETURN_FALSE(m_trainSamplesPtr->begin()->second.size() > 0,
				"Invalid Wishart classes training samples number")*/

				MClassesSamplesCTPtr allClassesSamples(new MClassesSamplesCT());
				ClassSamplesContainerT classSamples;
				ClassSampleT samples;

				typedef std::vector< std::vector< double > > Container;

				Container::const_iterator samplesIt =
					m_polygonSamplesTNew.begin();
				const Container::const_iterator samplesItE =
					m_polygonSamplesTNew.end();

				int classID = 0;
				WishartFeatureType value;

				while (samplesIt != samplesItE) {
					for (int i = 0; i < m_numberOfSamples[classID]; ++i) {

						int row = (int)(samplesIt->operator[](i) / m_inputRasterPtr->getNumberOfColumns());
						int column = ((int)samplesIt->operator[](i) % m_inputRasterPtr->getNumberOfColumns());

						if (((row % (m_lagY + 1)) == 0) && ((column % (m_lagX + 1)) == 0)) {  // De-correlating samples input

							for (unsigned int band = 0; band < m_inputRastersBands.size(); ++band) {
								m_inputRasterPtr->getBand(band)->getValue(column, row, value);

								samples.push_back(value);
							}

							classSamples.push_back(samples);
							samples.clear();
						}
					}

					allClassesSamples->insert(MClassesSamplesCT::value_type(classID, classSamples));
					classSamples.clear();

					++classID;
					++samplesIt;
				}

				m_trainSamplesPtr = allClassesSamples;

				TERP_TRUE_OR_RETURN_FALSE(m_trainSamplesPtr,
					"Invalid Wishart classes training samples pointer")
					TERP_TRUE_OR_RETURN_FALSE(m_trainSamplesPtr->size() > 0,
					"Invalid Wishart classes training samples number")
					TERP_TRUE_OR_RETURN_FALSE(m_trainSamplesPtr->begin()->second.size() > 0,
					"Invalid Wishart classes training samples number")

					return true;
			}

			// Calculating Classes Sigma parameter
			bool WishartModel::CalculatingSigmaClass() {

				const unsigned int dimsNumber = (unsigned int)
					m_trainSamplesPtr->begin()->second.operator[](0).size();
				TERP_TRUE_OR_RETURN_FALSE(dimsNumber > 0, "Invalid Wishart classes band number");

				MClassesSamplesCT::const_iterator classesIt =
					m_trainSamplesPtr->begin();
				const MClassesSamplesCT::const_iterator classesItE =
					m_trainSamplesPtr->end();
				int classesNum = 0;  // --->>>>> TESTE DO NUMERO DE CLASSES
				while (classesIt != classesItE) {
					classesNum++;  // --->>>>> TESTE DO NUMERO DE CLASSES
					const ClassIDT& classID = classesIt->first;

					const ClassSamplesContainerT& classSamples = classesIt->second;
					TERP_TRUE_OR_RETURN_FALSE(classSamples.size() > 0,
						"Invalid class samples number");

					ClassSampleT classMeans(dimsNumber, 0.0);

					for (unsigned int dimIdx = 0; dimIdx < dimsNumber; ++dimIdx) {

						WishartFeatureType& dimMean = classMeans[dimIdx];

						double realComplex = 0.0;
						double imgComplex = 0.0;

						dimMean = 0.0;

						ClassSamplesContainerT::const_iterator samplesIt =
							classSamples.begin();
						const ClassSamplesContainerT::const_iterator samplesItE =
							classSamples.end();

						while (samplesIt != samplesItE) {

							TERP_TRUE_OR_RETURN_FALSE(samplesIt->size() == dimsNumber, "Sample size mismatch")

								if ((std::real((*samplesIt)[dimIdx]) / 1) == std::real((*samplesIt)[dimIdx])) {  // avoiding nan
									realComplex += std::real((*samplesIt)[dimIdx]);
									imgComplex += std::imag((*samplesIt)[dimIdx]);
								}
								else
                {
//									std::cout << "Class: " << classID << std::endl;
                }

							++samplesIt;
						}

						realComplex /= (double)(classSamples.size());
						imgComplex /= (double)(classSamples.size());
						dimMean = WishartFeatureType(realComplex, imgComplex);
					}

					m_sigmaClassesParameter.push_back(classMeans);
					m_classesIndex2ID.push_back(classID);
					m_samplesSize.push_back(classSamples.size());

					++classesIt;
				}
//				std::cout << "          Number of classes: " << classesNum << std::endl;  // --->>>>> TESTE DO NUMERO DE CLASSES

				return true;
			}

			// Calculating Regions Sigma parameter
			bool WishartModel::CalculatingSigmaRegion() {

				TERP_TRUE_OR_RETURN_FALSE(m_polygonsSegImage.size() > 0, "Wishart Model algorithm needs polygons")

					TERP_TRUE_OR_RETURN_FALSE(m_inputRastersBands.size() > 0, "Invalid band number");

				WishartFeatureType value;

				for (unsigned int i = 0; i < m_polygonsSegImage.size(); i++) {

					unsigned int regionsNum = 0;

					std::vector< WishartFeatureType > regionMean;

					for (unsigned int j = 0; j < m_inputRastersBands.size(); ++j)
						regionMean.push_back(WishartFeatureType(0.0, 0.0));

					te::rst::PolygonIterator< WishartFeatureType > polygonsIt = te::rst::PolygonIterator< WishartFeatureType >::
						begin(m_inputRasterPtr, m_polygonsSegImage[i]);
					te::rst::PolygonIterator< WishartFeatureType > polygonsItE = te::rst::PolygonIterator< WishartFeatureType >::
						end(m_inputRasterPtr, m_polygonsSegImage[i]);

					while (polygonsIt != polygonsItE) {

						unsigned int column = polygonsIt.getColumn();
						unsigned int row = polygonsIt.getRow();

						if (((row % (m_lagY + 1)) == 0) && ((column % (m_lagX + 1)) == 0)) {  // De-correlating regions

							for (unsigned int dimIdx = 0; dimIdx < m_inputRastersBands.size(); ++dimIdx) {

								m_inputRasterPtr->getBand(dimIdx)->getValue(column, row, value);
								if ((std::real(value) / 1) == std::real(value))  // avoiding nan
									regionMean[dimIdx] += value;

							}

							++regionsNum;
						}

						++polygonsIt;
					}

					m_regionsSize.push_back(regionsNum);

					for (unsigned int j = 0; j < m_inputRastersBands.size(); ++j)
						regionMean[j] /= (double)regionsNum;

					m_sigmaRegionsParameter.push_back(regionMean);
					regionMean.clear();
				}

				return true;
			}

			// Getting complete sigma matrix
			bool WishartModel::GettingCompleteSigmaMatrixes() {

				double realComplex = 0.0;
				double imgComplex = 0.0;

				// regions sigma matrixes

				ClassSamplesContainerT::const_iterator regionsIt =
					m_sigmaRegionsParameter.begin();
				const ClassSamplesContainerT::const_iterator regionsItE =
					m_sigmaRegionsParameter.end();

				boost::numeric::ublas::matrix< WishartFeatureType > sigmaBand(m_vectorOrMatrixOrder, m_vectorOrMatrixOrder);

				while (regionsIt != regionsItE) {

					if (m_radarDataRepresentationType == "covarianceMatrix") {
						unsigned int i = 0;
						for (unsigned int j = 0; j < m_vectorOrMatrixOrder; ++j) {
							for (unsigned int k = 0; k < m_vectorOrMatrixOrder; ++k) {
								if ((k > j) || (k == j)) {
									sigmaBand(j, k) = (*regionsIt)[i];
									if (k != j)
										sigmaBand(k, j) = std::conj((*regionsIt)[i]);
									++i;
								}
							}
						}
					}

					if (m_radarDataRepresentationType == "lexicographicVector") {
						unsigned int i = 0;
						for (unsigned int j = 0; j < m_vectorOrMatrixOrder; ++j) {
							for (unsigned int k = 0; k < m_vectorOrMatrixOrder; ++k) {
								realComplex = std::real((*regionsIt)[i]); imgComplex = std::imag((*regionsIt)[i]);
								sigmaBand(j, k) = WishartFeatureType(realComplex, imgComplex);
								++i;
							}
						}
					}

					m_sigmaRegionsParameterMatrix.push_back(sigmaBand);
					sigmaBand.clear();
					++regionsIt;
				}

				// classes Sigma matrixes
				ClassSamplesContainerT::const_iterator classesIt =
					m_sigmaClassesParameter.begin();
				const ClassSamplesContainerT::const_iterator classesItE =
					m_sigmaClassesParameter.end();

				while (classesIt != classesItE) {

					if (m_radarDataRepresentationType == "covarianceMatrix") {
						unsigned int i = 0;
						for (unsigned int j = 0; j < m_vectorOrMatrixOrder; ++j) {
							for (unsigned int k = 0; k < m_vectorOrMatrixOrder; ++k) {
								if ((k > j) || (k == j)) {
									sigmaBand(j, k) = (*classesIt)[i];
									if (k != j)
										sigmaBand(k, j) = std::conj((*classesIt)[i]);
									++i;
								}
							}
						}
					}

					if (m_radarDataRepresentationType == "lexicographicVector") {
						unsigned int i = 0;
						for (unsigned int j = 0; j < m_vectorOrMatrixOrder; ++j) {
							for (unsigned int k = 0; k < m_vectorOrMatrixOrder; ++k) {
								realComplex = std::real((*classesIt)[i]); imgComplex = std::imag((*classesIt)[i]);
								sigmaBand(j, k) = WishartFeatureType(realComplex, imgComplex);
								++i;
							}
						}
					}

					m_sigmaClassesParameterMatrix.push_back(sigmaBand);
					sigmaBand.clear();
					++classesIt;
				}

				return true;
			}

			bool WishartModel::GettingAttributes() {

				boost::numeric::ublas::matrix< std::complex< double > > mInverse;

				std::vector< boost::numeric::ublas::matrix< WishartFeatureType > > sigmaRegionInvMatrix;
				std::vector< boost::numeric::ublas::matrix< WishartFeatureType > > sigmaClassInvMatrix;
				std::vector< WishartFeatureType > sigmaRegionMatrixDet;
				std::vector< WishartFeatureType > sigmaClassMatrixDet;
				sigmaRegionInvMatrix.clear();
				sigmaClassInvMatrix.clear();
				sigmaRegionMatrixDet.clear();
				sigmaClassMatrixDet.clear();

				// Calculating determinant and inverse matrixes
				for (unsigned int i = 0; i < m_sigmaRegionsParameterMatrix.size(); i++) {

					WishartFeatureType determinantM;
					if (!te::common::GetDeterminantComplex(m_sigmaRegionsParameterMatrix[i], determinantM))
						m_sigmaRegionsParameterMatrix[i] = checkCovarianceMatrix(m_sigmaRegionsParameterMatrix[i]);

					TERP_TRUE_OR_RETURN_FALSE(te::common::GetDeterminantComplex(m_sigmaRegionsParameterMatrix[i],
						determinantM), "Determinant region matrix calcule error");
					sigmaRegionMatrixDet.push_back(determinantM);

					TERP_TRUE_OR_RETURN_FALSE(te::common::GetInverseMatrix(m_sigmaRegionsParameterMatrix[i],
						mInverse), "Inverse region matrix calcule error");
					sigmaRegionInvMatrix.push_back(mInverse);

					mInverse.clear();
				}

				for (unsigned int i = 0; i < m_sigmaClassesParameterMatrix.size(); i++) {

					std::complex< double > determinantM;
					if (!te::common::GetDeterminantComplex(m_sigmaClassesParameterMatrix[i], determinantM))
						m_sigmaClassesParameterMatrix[i] = checkCovarianceMatrix(m_sigmaClassesParameterMatrix[i]);

					TERP_TRUE_OR_RETURN_FALSE(te::common::GetDeterminantComplex(m_sigmaClassesParameterMatrix[i],
						determinantM), "Determinant class matrix calcule error");
					sigmaClassMatrixDet.push_back(determinantM);

					boost::numeric::ublas::matrix< std::complex< double > > mInverse;

					TERP_TRUE_OR_RETURN_FALSE(te::common::GetInverseMatrix(m_sigmaClassesParameterMatrix[i],
						mInverse), "Inverse class matrix calcule error");
					sigmaClassInvMatrix.push_back(mInverse);

					mInverse.clear();
				}

				// Calculating Stochastic Distances, Statistical Tests and P-values

				boost::numeric::ublas::matrix< WishartFeatureType > sumMatrixes;
				boost::numeric::ublas::matrix< WishartFeatureType > sumInvMatrixes;
				boost::numeric::ublas::matrix< WishartFeatureType > parcialResultMatrix1;
				boost::numeric::ublas::matrix< WishartFeatureType > parcialResultMatrix2;
				std::vector< double > distances;
				std::vector< double > tests;
				std::vector< double > pValues;
				WishartFeatureType parcialResult1;
				WishartFeatureType parcialResult2;
				WishartFeatureType sumInvMatrixesDet;

				double degreesOfFreedom = pow(m_vectorOrMatrixOrder, 2);

				boost::math::chi_squared distChiSquared(degreesOfFreedom);

				for (unsigned int i = 0; i < m_sigmaRegionsParameterMatrix.size(); i++) {

					distances.clear();
					tests.clear();

					if (m_stochasticDistanceType == "Bhattacharyya") {  // Bhattacharyya distance

						for (unsigned int j = 0; j < m_sigmaClassesParameterMatrix.size(); j++) {

							sumMatrixes.clear();
							parcialResult1 = WishartFeatureType(0.0, 0.0);
							parcialResult2 = WishartFeatureType(0.0, 0.0);
							sumInvMatrixesDet = WishartFeatureType(0.0, 0.0);

							sumMatrixes = (sigmaRegionInvMatrix[i] + sigmaClassInvMatrix[j]) / 2.0;
							if (!te::common::GetDeterminantComplex(sumMatrixes, sumInvMatrixesDet))
								sumMatrixes = checkCovarianceMatrix(sumMatrixes);

							TERP_TRUE_OR_RETURN_FALSE(te::common::GetInverseMatrix(sumMatrixes,
								sumInvMatrixes), "Inverse sum inverse matrix calcule error");
							TERP_TRUE_OR_RETURN_FALSE(te::common::GetDeterminantComplex(sumInvMatrixes,
								sumInvMatrixesDet), "Determinant sum inverse matrix calcule error");

							parcialResult1 = (log(sigmaRegionMatrixDet[i]) + log(sigmaClassMatrixDet[j])) / 2.0;

							parcialResult2 = log(sumInvMatrixesDet);

							double aux = std::real(parcialResult1 - parcialResult2);

							double distance = m_ENL * aux;

							if (distance < 0)
								distance = 0;

							distances.push_back(distance);

							double aux1 = (8 * m_regionsSize[i] * m_samplesSize[j]) / (m_regionsSize[i] + m_samplesSize[j]);

							double statistic = aux1 * (m_ENL * aux);

							if (distance < 0)
								statistic = 0;

							tests.push_back(statistic);

							try {
								pValues.push_back(boost::math::cdf(complement(distChiSquared, statistic)));
							}
							catch (...) {
								pValues.push_back(NAN);
							}
						}

						m_stochasticDistance.push_back(distances);
						m_statisticalTest.push_back(tests);
						m_pValue.push_back(pValues);
						distances.clear();
						tests.clear();
						pValues.clear();
					}

					if (m_stochasticDistanceType == "KullbackLeibler") {  // Kullback-Leibler distance

						for (unsigned int j = 0; j < m_sigmaClassesParameterMatrix.size(); j++) {

							parcialResultMatrix1.clear();
							parcialResultMatrix2.clear();
							sumMatrixes.clear();
							parcialResult1 = WishartFeatureType(0.0, 0.0);

							parcialResultMatrix1 = boost::numeric::ublas::prod(sigmaRegionInvMatrix[i], m_sigmaClassesParameterMatrix[j]);
							parcialResultMatrix2 = boost::numeric::ublas::prod(sigmaClassInvMatrix[j], m_sigmaRegionsParameterMatrix[i]);
							sumMatrixes = parcialResultMatrix1 + parcialResultMatrix2;

							for (unsigned int dim = 0; dim < m_vectorOrMatrixOrder; ++dim)
								parcialResult1 += sumMatrixes(dim, dim);

							WishartFeatureType aux = parcialResult1 / 2.0;

							double distance = m_ENL * (std::real(aux) - m_vectorOrMatrixOrder);

							if (distance < 0)
								distance = 0;

							distances.push_back(distance);

							double aux1 = (2 * m_regionsSize[i] * m_samplesSize[j]) / (m_regionsSize[i] + m_samplesSize[j]);

							double statistic = aux1 * distance;

							if (distance < 0)
								statistic = 0;

							tests.push_back(statistic);

							try {
								pValues.push_back(boost::math::cdf(complement(distChiSquared, statistic)));
							}
							catch (...) {
								pValues.push_back(NAN);
							}
						}

						m_stochasticDistance.push_back(distances);
						m_statisticalTest.push_back(tests);
						m_pValue.push_back(pValues);
						distances.clear();
						tests.clear();
						pValues.clear();
					}

					if (m_stochasticDistanceType == "Hellinger") {  // Hellinger distance

						for (unsigned int j = 0; j < m_sigmaClassesParameterMatrix.size(); j++) {

							boost::numeric::ublas::matrix< WishartFeatureType > aux;
							sumMatrixes.clear();
							sumInvMatrixes.clear();
							parcialResult1 = WishartFeatureType(0.0, 0.0);
							parcialResult2 = WishartFeatureType(0.0, 0.0);
							sumInvMatrixesDet = WishartFeatureType(0.0, 0.0);

							sumMatrixes = sigmaRegionInvMatrix[i] + sigmaClassInvMatrix[j];

							if (!te::common::GetDeterminantComplex(sumMatrixes, sumInvMatrixesDet))
								sumMatrixes = checkCovarianceMatrix(sumMatrixes);

							aux = sumMatrixes / 2.0;

							TERP_TRUE_OR_RETURN_FALSE(te::common::GetInverseMatrix(aux,
								sumInvMatrixes), "Inverse sum inverse matrix calcule error");


							TERP_TRUE_OR_RETURN_FALSE(te::common::GetDeterminantComplex(sumInvMatrixes,
								sumInvMatrixesDet), "Determinant sum inverse matrix calcule error");
							aux.clear();

							double parcial1 = sqrt(std::real(sigmaRegionMatrixDet[i]) * std::real(sigmaClassMatrixDet[j]));

							double aux1 = (double)std::real(sumInvMatrixesDet) / (double)parcial1;
							double parcial2 = pow(aux1, m_ENL);

							double distance = 1.0 - parcial2;

							if (distance < 0)
								distance = 0;

							distances.push_back(distance);

							double termS = (8 * m_regionsSize[i] * m_samplesSize[j]) / (m_regionsSize[i] + m_samplesSize[j]);

							double statistic = termS * distance;

							if (distance < 0)
								statistic = 0;

							tests.push_back(statistic);

							try {
								pValues.push_back(boost::math::cdf(complement(distChiSquared, statistic)));
							}
							catch (...) {
								pValues.push_back(NAN);
							}
						}

						m_stochasticDistance.push_back(distances);
						m_statisticalTest.push_back(tests);
						m_pValue.push_back(pValues);
						distances.clear();
						tests.clear();
						pValues.clear();
					}

					if (m_stochasticDistanceType == "Renyi") {  // Renyi distance

						for (unsigned int j = 0; j < m_sigmaClassesParameterMatrix.size(); j++) {

							sumMatrixes.clear();
							sumInvMatrixes.clear();
							parcialResultMatrix1.clear();
							parcialResultMatrix2.clear();
							WishartFeatureType parcialResult1 = WishartFeatureType(0.0, 0.0);
							WishartFeatureType parcialResult2 = WishartFeatureType(0.0, 0.0);
							WishartFeatureType parcialResult3 = WishartFeatureType(0.0, 0.0);
							WishartFeatureType parcialResult4 = WishartFeatureType(0.0, 0.0);
							WishartFeatureType resultD = WishartFeatureType(0.0, 0.0);
							WishartFeatureType resultS = WishartFeatureType(0.0, 0.0);

							parcialResult1 = pow(sigmaRegionMatrixDet[i], -m_betaValue);
							parcialResult2 = pow(sigmaClassMatrixDet[j], m_betaValue - 1);

							parcialResultMatrix1 = m_betaValue * sigmaRegionInvMatrix[i];

							parcialResultMatrix2 = (1 - m_betaValue) * sigmaClassInvMatrix[j];

							sumMatrixes = parcialResultMatrix1 + parcialResultMatrix2;
							if (!te::common::GetDeterminantComplex(sumMatrixes, sumInvMatrixesDet))
								sumMatrixes = checkCovarianceMatrix(sumMatrixes);

							TERP_TRUE_OR_RETURN_FALSE(te::common::GetInverseMatrix(sumMatrixes,
								sumInvMatrixes), "Inverse sum inverse matrix calcule error");
							TERP_TRUE_OR_RETURN_FALSE(te::common::GetDeterminantComplex(sumInvMatrixes,
								sumInvMatrixesDet), "Determinant sum inverse matrix calcule error");

							parcialResult3 = pow(parcialResult1 * parcialResult2 * sumInvMatrixesDet, m_ENL);

							parcialResult1 = pow(sigmaRegionMatrixDet[i], m_betaValue - 1);
							parcialResult2 = pow(sigmaClassMatrixDet[j], -m_betaValue);

							parcialResultMatrix1 = m_betaValue * sigmaClassInvMatrix[j];

							parcialResultMatrix2 = (1 - m_betaValue) * sigmaRegionInvMatrix[i];

							sumMatrixes = parcialResultMatrix1 + parcialResultMatrix2;
							if (!te::common::GetDeterminantComplex(sumMatrixes, sumInvMatrixesDet))
								sumMatrixes = checkCovarianceMatrix(sumMatrixes);

							TERP_TRUE_OR_RETURN_FALSE(te::common::GetInverseMatrix(sumMatrixes,
								sumInvMatrixes), "Inverse sum inverse matrix calcule error");
							TERP_TRUE_OR_RETURN_FALSE(te::common::GetDeterminantComplex(sumInvMatrixes,
								sumInvMatrixesDet), "Determinant sum inverse matrix calcule error");

							parcialResult4 = pow(parcialResult1 * parcialResult2 * sumInvMatrixesDet, m_ENL);

							double term1 = (double)log(2) / (1 - m_betaValue);
							double term2 = (double) 1.0 / (m_betaValue - 1);

							resultD = term1 + term2 * log(parcialResult3 + parcialResult4);

							if (std::real(resultD) < 0)
								resultD = WishartFeatureType(0.0, 0.0);

							distances.push_back(std::real(resultD));

							double termS = (double)(2 * m_regionsSize[i] * m_samplesSize[j]) / (double)(m_betaValue * (m_regionsSize[i] + m_samplesSize[j]));

							resultS = termS * resultD;

							if (std::real(resultS) < 0)
								resultS = WishartFeatureType(0.0, 0.0);

							tests.push_back(std::real(resultS));

							try {
								pValues.push_back(boost::math::cdf(complement(distChiSquared, std::real(resultS))));
							}
							catch (...) {
								pValues.push_back(NAN);
							}
						}

						m_stochasticDistance.push_back(distances);
						m_statisticalTest.push_back(tests);
						m_pValue.push_back(pValues);
						distances.clear();
						tests.clear();
						pValues.clear();
					}

					if (m_stochasticDistanceType == "ChiSquare") {  // Chi Square distance

						for (unsigned int j = 0; j < m_sigmaClassesParameterMatrix.size(); j++) {

							sumMatrixes.clear();
							sumInvMatrixes.clear();
							parcialResultMatrix1.clear();
							parcialResultMatrix2.clear();
							parcialResult1 = WishartFeatureType(0.0, 0.0);
							parcialResult2 = WishartFeatureType(0.0, 0.0);;
							WishartFeatureType parcialResult3;

							parcialResult1 = sigmaRegionMatrixDet[i] / pow(sigmaClassMatrixDet[j], 2);

							parcialResultMatrix1 = sigmaClassInvMatrix[j] * 2;

							sumMatrixes = parcialResultMatrix1 - sigmaRegionInvMatrix[i];
							if (!te::common::GetDeterminantComplex(sumMatrixes, sumInvMatrixesDet))
								sumMatrixes = checkCovarianceMatrix(sumMatrixes);

							TERP_TRUE_OR_RETURN_FALSE(te::common::GetInverseMatrix(sumMatrixes,
								sumInvMatrixes), "Inverse sum inverse matrix calcule error");
							TERP_TRUE_OR_RETURN_FALSE(te::common::GetDeterminantComplex(sumInvMatrixes,
								sumInvMatrixesDet), "Determinant sum inverse matrix calcule error");

							parcialResult2 = pow(parcialResult1 * abs(sumInvMatrixesDet), m_ENL);

							parcialResult1 = sigmaClassMatrixDet[j] / pow(sigmaRegionMatrixDet[i], 2);

							parcialResultMatrix1 = sigmaRegionInvMatrix[i] * 2;

							sumMatrixes = parcialResultMatrix1 - sigmaClassInvMatrix[j];
							if (!te::common::GetDeterminantComplex(sumMatrixes, sumInvMatrixesDet))
								sumMatrixes = checkCovarianceMatrix(sumMatrixes);

							TERP_TRUE_OR_RETURN_FALSE(te::common::GetInverseMatrix(sumMatrixes,
								sumInvMatrixes), "Inverse sum inverse matrix calcule error");
							TERP_TRUE_OR_RETURN_FALSE(te::common::GetDeterminantComplex(sumInvMatrixes,
								sumInvMatrixesDet), "Determinant sum inverse matrix calcule error");

							parcialResult3 = pow(parcialResult1 * abs(sumInvMatrixesDet), m_ENL);

							WishartFeatureType distance = (parcialResult2 + parcialResult3) - WishartFeatureType(2.0, 0.0);

							if (std::real(distance) < 0)
								distance = WishartFeatureType(0.0, 0.0);

							distances.push_back(std::real(distance) / 4.0);

							double termS = ((double)m_regionsSize[i] * m_samplesSize[j]) / (2 * (m_regionsSize[i] + m_samplesSize[j]));

							double statistic = termS * std::real(distance);

							if (std::real(distance) < 0)
								statistic = 0;

							tests.push_back(statistic);

							try {
								pValues.push_back(boost::math::cdf(complement(distChiSquared, statistic)));
							}
							catch (...) {
								pValues.push_back(NAN);
							}
						}

						m_stochasticDistance.push_back(distances);
						m_statisticalTest.push_back(tests);
						m_pValue.push_back(pValues);
						distances.clear();
						tests.clear();
						pValues.clear();
					}
				}

				return true;
			}

			bool WishartModel::execute() {
				m_samplesSize.clear();
				m_regionsSize.clear();
				m_sigmaClassesParameter.clear();
				m_sigmaClassesParameterMatrix.clear();
				m_classesIndex2ID.clear();
				m_sigmaClassesParameter.clear();
				m_sigmaClassesParameterMatrix.clear();
				m_stochasticDistance.clear();
				m_statisticalTest.clear();
				m_pValue.clear();
				m_trainSamplesPtr = 0;

				if (m_radarDataRepresentationType == "lexicographicVector") {
					if (!CreatingCovarianceRaster())
					{
						TERP_LOG_AND_RETURN_FALSE( "CreatingCovarianceRaster execution error." );
					}
					else
					{
//						std::cout << "CreatingCovarianceRaster execution OK." << std::endl;
					}
				}

				if (!LoadingSamples())
				{
					TERP_LOG_AND_RETURN_FALSE( "LoadingSamples execution error." );
				}
				else
				{
//					std::cout << "LoadingSamples execution OK." << std::endl;
				}
				if (!CalculatingSigmaClass())
				{
					TERP_LOG_AND_RETURN_FALSE( "CalculatingSigmaClass execution error." );
				}
				else
				{
//					std::cout << "CalculatingSigmaClass execution OK." << std::endl;
				}
				if (!CalculatingSigmaRegion())
				{
					TERP_LOG_AND_RETURN_FALSE( "CalculatingSigmaRegion execution error." );
				}
				else
				{
//					std::cout << "CalculatingSigmaRegion execution OK." << std::endl;
				}
				if (!GettingCompleteSigmaMatrixes())
				{
					TERP_LOG_AND_RETURN_FALSE( "GettingCompleteSigmaMatrixes execution error." );
				}
				else
				{
//					std::cout << "GettingCompleteSigmaMatrixes execution OK." << std::endl;
				}
				if (!GettingAttributes())
				{
					TERP_LOG_AND_RETURN_FALSE( "GettingAttributes execution error." );
				}
				else
				{
//					std::cout << "GettingAttributes execution OK." << std::endl;
				}

				return true;
			}

		} // end namespace radar
	} // end namespace rp
}   // end namespace msc


