/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/rp/radar/WishartModel.h
  \brief Wishart model data manipulation.
*/

#ifndef __TERRALIB_RP_RADAR_INTERNAL_WISHARTMODEL_H
#define __TERRALIB_RP_RADAR_INTERNAL_WISHARTMODEL_H

// TerraLib includes
#include "../Config.h"
#include "../../raster/Raster.h"
#include "../../geometry/Polygon.h"

// Boost includes
#include <boost/numeric/ublas/matrix.hpp>
#include <boost/math/distributions/chi_squared.hpp>

#include <vector>
#include <map>

namespace te
{
  namespace rp
  {
    namespace radar
    {
    typedef std::complex< double > WishartFeatureType;
    /*!
      \class WishartModel

      \brief Wishart model data manipulation.
     */
    class TERPEXPORT WishartModel
    {
      public:

        te::rst::Raster* m_inputRasterPtr;  //!< Input raster.

        std::vector< unsigned int > m_inputRastersBands;  //!< Bands to process.

        unsigned int m_numberOfColumns;  //!< Number of columns that all input rasters must have.

        unsigned int m_numberOfRows;  //!< Number of rows that all input rasters must have.

        std::vector< te::gm::Polygon* > m_polygonsSegImage;  //!< Polygons of segmented image.

        std::vector< te::gm::Polygon > m_polygonSamplesT;  //!< Training samples polygons.
        std::vector< std::vector< double > > m_polygonSamplesTNew;  //!< Training samples polygons.
        std::vector< int > m_numberOfSamples;

        std::vector< unsigned int > m_classesIDT;  //!< Training class ID.

        typedef unsigned int ClassIDT;  //!< Class ID type definiton.

        typedef std::vector< WishartFeatureType > ClassSampleT;  //!< Class sample type definition.

        typedef std::vector< ClassSampleT > ClassSamplesContainerT;  //!< Class samples container type definition.

        typedef std::map< ClassIDT, ClassSamplesContainerT > MClassesSamplesCT;  //!< Multi-classes samples container type definition.

        typedef boost::shared_ptr< MClassesSamplesCT > MClassesSamplesCTPtr;  //!< A shared pointer to a multi classes samples container type definition.

        MClassesSamplesCTPtr m_trainSamplesPtr;  //!< A shared pointer to a always-valid structure where training samples are stored.

        ClassSamplesContainerT m_sigmaClassesParameter;  //!< Sigma classes parameter.

        std::vector< boost::numeric::ublas::matrix< WishartFeatureType > > m_sigmaClassesParameterMatrix;
        //!< Sigma classes parameter matrix.

        std::vector< unsigned int > m_classesIndex2ID;  //!< A class index ordered vector of classes IDs for Sigma parameter.

        ClassSamplesContainerT m_sigmaRegionsParameter;  //!< Sigma regions parameter.

        std::vector< boost::numeric::ublas::matrix< WishartFeatureType > > m_sigmaRegionsParameterMatrix; //!< Sigma regions parameter matrix.

        std::string m_radarDataRepresentationType; //!< The radar data representation container (covarianceMatrix,lexicographicVector).

        unsigned int m_vectorOrMatrixOrder; //!< The lexicographic vector and covariance matrix order.

        double m_ENL;  //!< Equivalent Number Look type.

        unsigned int m_lagX;  //!< The horizontal correlation value.

        unsigned int m_lagY;  //!< The vertical correlation value.

        double m_betaValue;  //!< Beta value for Renyi distance type.

        double m_significanceLevel;   //!< The significance level.

        std::string m_stochasticDistanceType;
        //!< The stochastic distance type definition.

        std::vector < std::vector< double > > m_stochasticDistance; //!< Stochastic distance container.

        std::vector < std::vector< double > > m_statisticalTest; //!< Statistical test container.

        std::vector< std::vector< double > > m_pValue; //!< P-values container.

        std::vector< double > m_regionsSize;  //!< Size of regions.

        std::vector< double > m_samplesSize;  //!< Size of sample containers.

        /*! \brief Constructor. */
        WishartModel();

        /*! \brief Destructor. */
        ~WishartModel();

        boost::numeric::ublas::matrix< WishartFeatureType > checkCovarianceMatrix(boost::numeric::ublas::matrix< WishartFeatureType >);

        bool CreatingCovarianceRaster(void);

        bool LoadingSamples(void);

        bool CalculatingSigmaClass(void);

        bool CalculatingSigmaRegion(void);

        bool GettingCompleteSigmaMatrixes(void);

        bool GettingAttributes(void);

        bool execute(void);
    };

    } // end namespace radar
  } // end namespace rp
}   // end namespace msc

#endif


