/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*! 
  \file terralib/srs/Converter.cpp
*/

// TerraLib
#include "../core/translator/Translator.h"
#include "Config.h"
#include "Converter.h"
#include "Exception.h"
#include "Module.h"
#include "SpatialReferenceSystemManager.h"


#ifdef TERRALIB_PROJ4_ENABLED
  // Global proj header common to all versions
  #include <proj.h>

  #ifndef PROJ_VERSION_MAJOR
    #error PROJ_VERSION_MAJOR not declared
  #endif

  #if PROJ_VERSION_MAJOR < 8
    //this must be added to allow the include of the file proj_api.h. This file is deprecated in proj 6 and will be removed in proj 7.
    #ifndef ACCEPT_USE_OF_DEPRECATED_PROJ_API_H
      #define ACCEPT_USE_OF_DEPRECATED_PROJ_API_H
    #endif

    #include <proj_api.h>
  #endif
#endif

// STL
#include <cassert>
#include <cstring>

// Boost
#include <boost/thread/mutex.hpp>

te::srs::Converter::Converter():
  m_targetSRID(TE_UNKNOWN_SRS),
  m_sourceSRID(TE_UNKNOWN_SRS),
  m_sourcePj4Handler(nullptr),
  m_targetPj4Handler(nullptr),
  m_transfPjHandler(nullptr),
  m_transfThreadContextHandler(nullptr),
  m_nonNormalizedTransfPjHandler(nullptr),
  m_sourceIsLatLong( false ),
  m_targetIsLatLong( false )
{
  #ifndef TERRALIB_PROJ4_ENABLED
    throw te::srs::Exception(TE_TR("PROJ4 library has to be enabled in order to support coordinate conversion."));
  #endif
}

te::srs::Converter::Converter(int sourceSRID, int targetSRID):
  m_targetSRID(targetSRID),
  m_sourceSRID(sourceSRID),
  m_sourcePj4Handler(nullptr),
  m_targetPj4Handler(nullptr),
  m_transfPjHandler(nullptr),
  m_transfThreadContextHandler(nullptr),
  m_nonNormalizedTransfPjHandler(nullptr),
  m_sourceIsLatLong( false ),
  m_targetIsLatLong( false )
{
  #ifdef TERRALIB_PROJ4_ENABLED
    const std::string sDescription =
      te::srs::SpatialReferenceSystemManager::getInstance().getP4Txt(sourceSRID, "");
    if (sDescription.empty())
      throw te::srs::Exception(TE_TR("Source SRS ID not recognized."));

    const std::string tDescription =
      te::srs::SpatialReferenceSystemManager::getInstance().getP4Txt(targetSRID, "");
    if (tDescription.empty())
      throw te::srs::Exception(TE_TR("Target SRS ID not recognized."));

    m_sourceCRS = sDescription;
    m_targetCRS = tDescription;

    std::string errorString;
    if( ! update( errorString ) )
    {
      throw te::srs::Exception( errorString );
    }
  #else
    throw te::srs::Exception(TE_TR("PROJ4 library has to be enabled in order to support coordinate conversion."));
  #endif
}

te::srs::Converter::~Converter()
{
  m_sourceSRID = TE_UNKNOWN_SRS;
  m_targetSRID= TE_UNKNOWN_SRS;
  m_sourceCRS.clear();
  m_targetCRS.clear();

  std::string errorString;
  update( errorString );
}

void 
te::srs::Converter::setSourceSRID(int sourceSRID)
{
  boost::unique_lock< boost::mutex > lockGuard(getStaticMutex());

  if( sourceSRID == m_sourceSRID ) return;

  std::string description =
    te::srs::SpatialReferenceSystemManager::getInstance().getP4Txt(sourceSRID, "");
  if (description.empty())
  {
    throw te::srs::Exception(TE_TR("SRS ID not recognized."));
  }

  m_sourceSRID = sourceSRID;
  m_sourceCRS = description;

  std::string errorString;
  if( ! update( errorString ) )
  {
    throw te::srs::Exception( errorString );
  }
}

void te::srs::Converter::setSourcePJ4txt(const std::string& pj4txt)
{
  boost::unique_lock< boost::mutex > lockGuard(getStaticMutex());

  if( pj4txt == m_sourceCRS ) return;

  if (pj4txt.empty())
  {
    throw te::srs::Exception(TE_TR("SRS ID not recognized."));
  }

  m_sourceCRS = pj4txt;
  m_sourceSRID = te::srs::SpatialReferenceSystemManager::getInstance().getIdFromP4Txt(
    pj4txt ).second;

  std::string errorString;
  if( ! update( errorString ) )
  {
    throw te::srs::Exception( errorString );
  }
}

int 
te::srs::Converter::getSourceSRID() const
{
  boost::unique_lock< boost::mutex > lockGuard(getStaticMutex());
  return m_sourceSRID;
}

void 
te::srs::Converter::setTargetSRID(int targetSRID)
{
  boost::unique_lock< boost::mutex > lockGuard(getStaticMutex());

  if( targetSRID == m_targetSRID ) return;

  std::string description =
    te::srs::SpatialReferenceSystemManager::getInstance().getP4Txt(targetSRID, "");
  if (description.empty())
  {
    throw te::srs::Exception(TE_TR("SRS ID not recognized."));
  }

  m_targetCRS = description;
  m_targetSRID = targetSRID;

  std::string errorString;
  if( ! update( errorString ) )
  {
    throw te::srs::Exception( errorString );
  }
}

void te::srs::Converter::setTargetPJ4txt(const std::string& pj4txt)
{
  boost::unique_lock< boost::mutex > lockGuard(getStaticMutex());

  if( pj4txt == m_targetCRS ) return;

  if (pj4txt.empty())
  {
    throw te::srs::Exception(TE_TR("SRS ID not recognized."));
  }

  m_targetCRS = pj4txt;
  m_targetSRID = te::srs::SpatialReferenceSystemManager::getInstance().getIdFromP4Txt(
    pj4txt ).second;

  std::string errorString;
  if( ! update( errorString ) )
  {
    throw te::srs::Exception( errorString );
  }
}

int 
te::srs::Converter::getTargetSRID() const
{
  boost::unique_lock< boost::mutex > lockGuard(getStaticMutex());
  return m_targetSRID;
}

bool te::srs::Converter::convert(double *xIn, double *yIn, double *xOut,
  double* yOut, long numCoord, int coordOffset) const
{
  memcpy(xOut, xIn, numCoord*sizeof(double));
  memcpy(yOut, yIn, numCoord*sizeof(double));

  return convert( xOut, yOut, numCoord, coordOffset );
}

bool te::srs::Converter::convert(double *x, double* y, long numCoord,
  int coordOffset) const
{
  for (long i = 0 ; i < numCoord; ++i )
  {
    if( ! convert( x[ i * coordOffset ], y[ i * coordOffset ] ) )
    {
      return false;
    }
  }

  return true;
}

bool te::srs::Converter::convert(double &x, double &y) const
{
  #if PROJ_VERSION_MAJOR < 8
    if(
        ( m_sourcePj4Handler != nullptr )
        &&
        ( m_targetPj4Handler != nullptr )
      )
    {
      if (m_sourceIsLatLong)
      {
        x *= DEG_TO_RAD;
        y *= DEG_TO_RAD;
      }

      m_convert_res = pj_transform((projPJ)m_sourcePj4Handler, (projPJ)m_targetPj4Handler,  1, 1,
        &x, &y, nullptr);

      if( m_convert_res == 0 )
      {
        if (m_targetIsLatLong)
        {
          x *= RAD_TO_DEG;
          y *= RAD_TO_DEG;
        }

        return true;
      }
      else
      {
        return false;
      }
    }
    else
    {
      return false;
    }
  #else
    if( m_transfPjHandler )
    {
      PJ_COORD projCoord;

      projCoord.lpzt.z = 0.0;
      projCoord.lpzt.t = HUGE_VAL;
      projCoord.lpzt.lam = x;
      projCoord.lpzt.phi = y;

      projCoord = proj_trans( (PJ*)m_transfPjHandler, PJ_FWD, projCoord);

      if( proj_context_errno( (PJ_CONTEXT*)m_transfThreadContextHandler ) )
      {
        return false;
      }

      x = projCoord.xy.x;
      y = projCoord.xy.y;

      return true;
    }
    else
    {
      return false;
    }
  #endif
}

bool te::srs::Converter::invert(double *xIn, double *yIn, double *xOut, double* yOut,
  long numCoord, int coordOffset) const
{
  memcpy(xOut, xIn, numCoord*sizeof(double));
  memcpy(yOut, yIn, numCoord*sizeof(double));

  return invert( xOut, yOut, numCoord, coordOffset );
}

bool te::srs::Converter::invert(double *x, double* y, long numCoord, int coordOffset) const
{
  for (long i = 0 ; i < numCoord; ++i )
  {
    if( ! invert( x[ i * coordOffset ], y[ i * coordOffset ] ) )
    {
      return false;
    }
  }

  return true;
}

bool te::srs::Converter::invert(double &x, double &y) const
{
  #if PROJ_VERSION_MAJOR < 8
    if(
        ( m_sourcePj4Handler != nullptr )
        &&
        ( m_targetPj4Handler != nullptr )
      )
    {
      if (m_targetIsLatLong)
      {
        x *= DEG_TO_RAD;
        y *= DEG_TO_RAD;
      }

      m_convert_res = pj_transform((projPJ)m_targetPj4Handler, (projPJ)m_sourcePj4Handler, 1, 1,
        &x, &y, nullptr);

      if( m_convert_res == 0 )
      {
        if (m_sourceIsLatLong)
        {
          x *= RAD_TO_DEG;
          y *= RAD_TO_DEG;
        }

        return true;
      }
      else
      {
        return false;
      }
    }
    else
    {
      return false;
    }
  #else
    if( m_transfPjHandler )
    {
      PJ_COORD projCoord;

      projCoord.lpzt.z = 0.0;
      projCoord.lpzt.t = HUGE_VAL;
      projCoord.lpzt.lam = x;
      projCoord.lpzt.phi = y;

      projCoord = proj_trans( (PJ*)m_transfPjHandler, PJ_INV, projCoord);

      if( proj_context_errno( (PJ_CONTEXT*)m_transfThreadContextHandler ) )
      {
        return false;
      }

      x = projCoord.xy.x;
      y = projCoord.xy.y;

      return true;
    }
    else
    {
      return false;
    }
  #endif
}

bool te::srs::Converter::convertToGeographic(double &x, double &y, int SRID)
{
  #ifdef TERRALIB_PROJ4_ENABLED
    #if PROJ_VERSION_MAJOR < 8
      std::string description = te::srs::SpatialReferenceSystemManager::getInstance().getP4Txt(SRID);
      if (description.empty())
        throw te::common::Exception(TE_TR("Source SRS ID not recognized."));

      projPJ pjhProj = pj_init_plus(description.c_str());
      if (!pjhProj)
      {
        std::string exceptionTxt = TE_TR("srs Source SRS description is not valid: ");
        char* pjError = pj_strerrno(*(pj_get_errno_ref()));
        exceptionTxt += std::string(pjError);
        throw te::common::Exception(exceptionTxt);
      }

      if (pj_is_latlong(pjhProj))
      {
        pj_free(pjhProj);
        return true;
      }

      projPJ pjhGeog = pj_latlong_from_proj(pjhProj);
      int res = pj_transform(pjhProj, pjhGeog, 1, 1, &x, &y, nullptr);
      pj_free(pjhProj);
      pj_free(pjhGeog);

      if (res == 0)
      {
        x *= RAD_TO_DEG;
        y *= RAD_TO_DEG;
        return true;
      }
      return false;
    #else
      throw te::srs::Exception(TE_TR("Operation not supported by PROJ library."));
    #endif
  #else
    throw te::srs::Exception(TE_TR("PROJ4 library has to be enabled in order to support coordinate conversion."));
    return false;
  #endif
}

bool te::srs::Converter::convertToProjected(double &lon, double &lat, int SRID)
{
  #ifdef TERRALIB_PROJ4_ENABLED
    #if PROJ_VERSION_MAJOR < 8
      std::string description = te::srs::SpatialReferenceSystemManager::getInstance().getP4Txt(SRID);
      if (description.empty())
        throw te::srs::Exception(TE_TR("Source SRS ID not recognized."));

      projPJ pjhProj = pj_init_plus(description.c_str());
      if (!pjhProj)
      {
        std::string exceptionTxt = TE_TR("Source SRS description is not valid: ");
        char* pjError = pj_strerrno(*(pj_get_errno_ref()));
        exceptionTxt += std::string(pjError);
        throw te::common::Exception(exceptionTxt);
      }
      projPJ pjhGeog = pj_latlong_from_proj(pjhProj);

      lon *= DEG_TO_RAD;
      lat *= DEG_TO_RAD;
      int res = pj_transform(pjhGeog, pjhProj, 1, 1, &lon, &lat, nullptr);

      pj_free(pjhProj);
      pj_free(pjhGeog);
      return (res == 0);
    #else
      throw te::srs::Exception(TE_TR("Operation not supported by PROJ library."));
      return false;
    #endif
  #else
    throw te::srs::Exception(TE_TR("PROJ4 library has to be enabled in order to support coordinate conversion."));
    return false;
  #endif
}

bool te::srs::Converter::update( std::string& errorString )
{
  errorString.clear();

  m_sourceIsLatLong = false;
  m_targetIsLatLong = false;

  #ifdef TERRALIB_PROJ4_ENABLED
    #if PROJ_VERSION_MAJOR < 8
      if (m_sourcePj4Handler) pj_free((projPJ)m_sourcePj4Handler);
      m_sourcePj4Handler = nullptr;

      if (m_targetPj4Handler) pj_free((projPJ)m_targetPj4Handler);
      m_targetPj4Handler = nullptr;

      if( ! m_sourceCRS.empty() )
      {
        m_sourcePj4Handler = pj_init_plus(m_sourceCRS.c_str());
        if (!m_sourcePj4Handler)
        {
          errorString = TE_TR("Source SRS description is not valid: ");
          char* pjError = pj_strerrno(*(pj_get_errno_ref()));
          errorString += std::string(pjError);
          return false;
        }

        m_sourceIsLatLong = ( pj_is_latlong( (projPJ)m_sourcePj4Handler ) != 0 );
      }

      if( ! m_targetCRS.empty() )
      {
        m_targetPj4Handler = pj_init_plus(m_targetCRS.c_str());
        if (!m_targetPj4Handler)
        {
          errorString = TE_TR("Target SRS description is not valid: ");
          char* pjError = pj_strerrno(*(pj_get_errno_ref()));
          errorString += std::string(pjError);
          return false;
        }

        m_targetIsLatLong = ( pj_is_latlong( (projPJ)m_targetPj4Handler ) != 0 );
      }
    #else

      if (m_transfPjHandler)
      {
        proj_destroy((PJ*)m_transfPjHandler);
        m_transfPjHandler = nullptr;
      }

      if (m_nonNormalizedTransfPjHandler)
      {
        proj_destroy((PJ*)m_nonNormalizedTransfPjHandler);
        m_nonNormalizedTransfPjHandler = nullptr;
      }

      if (m_transfThreadContextHandler)
      {
        proj_context_destroy((PJ_CONTEXT*)m_transfThreadContextHandler);
        m_transfThreadContextHandler = nullptr;
      }

      if( ( ! m_sourceCRS.empty() ) && ( ! m_targetCRS.empty() ) )
      {
        m_transfThreadContextHandler = proj_context_create();
        if(!m_transfThreadContextHandler)
        {
          errorString = TE_TR("Source/Target SRS descriptions are not valid: ");
          return false;
        }

        m_nonNormalizedTransfPjHandler = proj_create_crs_to_crs(
          (PJ_CONTEXT*)m_transfThreadContextHandler,
          m_sourceCRS.c_str(), m_targetCRS.c_str(), 0 );
        if (!m_nonNormalizedTransfPjHandler)
        {
          errorString = TE_TR("Source/Target SRS descriptions are not valid: ");
          char const * const pjError = proj_context_errno_string(
            (PJ_CONTEXT*)m_transfThreadContextHandler, proj_context_errno(
            (PJ_CONTEXT*)m_transfThreadContextHandler ) );
          errorString += std::string(pjError);
          return false;
        }

        m_transfPjHandler = proj_normalize_for_visualization(
          (PJ_CONTEXT*)m_transfThreadContextHandler, (PJ*)m_nonNormalizedTransfPjHandler );
        if (!m_transfPjHandler)
        {
          errorString = TE_TR("Source/Target SRS descriptions are not valid: ");
          char const * const pjError = proj_context_errno_string(
            (PJ_CONTEXT*)m_transfThreadContextHandler, proj_context_errno(
            (PJ_CONTEXT*)m_transfThreadContextHandler ) );
          errorString += std::string(pjError);
          return false;
        }
      }
    #endif

    return true;
  #endif

  return false;
}

te::srs::Converter* te::srs::Converter::clone() const
{
  return new te::srs::Converter( m_sourceSRID, m_targetSRID );
}