/*
  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

  This file is part of the TerraLib - a Framework for building GIS enabled
  applications.

  TerraLib is free software: you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation, either version 3 of the License,
  or (at your option) any later version.

  TerraLib is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with TerraLib. See COPYING. If not, write to
  TerraLib Team at <terralib-team@terralib.org>.
  */

/*!
  \file terralib/vm/lua/VirtualMachine.cpp

  \brief Implementation of a virtual machine for the Lua Programming Language.

  \author Frederico Augusto Bedê
  \author Gilberto Ribeiro de Queiroz
  */

// TerraLib
#include "VirtualMachine.h"
#include "../../core/translator/Translator.h"
#include "../core/Exception.h"

// Lua
#include <QtLua/State>
#include <QtLua/TableTreeModel>
#include <QtLua/Console>
#include <lua.hpp>

// QT
#include <QFile>
#include <QTextStream>
#include <QThread>

// STL                   
#include <cstdio>

void stdinHack(lua_State *L, FILE *in, FILE *out, FILE *err) {
  FILE **pf;
  lua_getglobal(L, "io"); /* Get the IO library */
  lua_pushstring(L, "open");
  lua_gettable(L, -2); /* io, io.open */
  lua_getuservalue(L, -1); /* io, io.open, io.open's env */
  lua_getglobal(L, "io"); /* io, open, env, io */

  lua_pushstring(L, "stdin");
  lua_gettable(L, -2); /* io, open, env, io, io.stdin */
  pf = (FILE **)lua_touserdata(L, -1);
  *pf = in;
  lua_rawseti(L, -3, 1); /* IO_INPUT = 1 */

  lua_pushstring(L, "stdout");
  lua_gettable(L, -2); /* io, open, env, io, io.stdout */
  pf = (FILE **)lua_touserdata(L, -1);
  *pf = out;
  lua_rawseti(L, -3, 2); /* IO_OUTPUT = 2 */

  lua_pushstring(L, "stderr");
  lua_gettable(L, -2); /* io, open, env, io, io.stderr */
  pf = (FILE **)lua_touserdata(L, -1);
  *pf = err;
  lua_pop(L, 5);
}

struct te::vm::lua::VirtualMachine::Impl
{
  std::unique_ptr<QtLua::State> state;  //!< The Lua environment (or state).
  FILE* m_stdin;
  QString m_filename;
  Impl() : state(nullptr)
  {
    m_stdin = tmpfile();
  }
  ~Impl()
  {
//    fclose(m_stdin);
  }
};

te::vm::lua::VirtualMachine::VirtualMachine() : m_pimpl(nullptr)
{
  m_pimpl.reset(new Impl);
}

te::vm::lua::VirtualMachine::~VirtualMachine() = default;

void te::vm::lua::VirtualMachine::setFileName(QString file)
{
  m_pimpl->m_filename = file;
}

void te::vm::lua::VirtualMachine::execute()
{
  m_pimpl->state.reset(new QtLua::State());
  m_pimpl->state->openlib(QtLua::AllLibs);
//  stdinHack(m_pimpl->state->get_lua_state(), m_pimpl->m_stdin, stdout, stderr);
  connect(m_pimpl->state.get(), SIGNAL(output(const QString&)), this,
          SLOT(write(const QString&)), Qt::UniqueConnection);
  QFile script(m_pimpl->m_filename);
  script.open(QIODevice::ReadOnly);
  QTextStream in(&script);
  m_pimpl->state->exec(in.readAll());
}

void te::vm::lua::VirtualMachine::write(const QString& str)
{
  emit output(str);
  QThread::currentThread()->usleep(1);
}

void te::vm::lua::VirtualMachine::readline(QString /*msg*/)
{
//  msg = msg + '\0';
//  if(m_pimpl->state)
//  {
//    std::fputs(msg.toUtf8().data(), m_pimpl->m_stdin);
//  }
}

std::string te::vm::lua::VirtualMachine::readline()
{
  return "";
}

void te::vm::lua::VirtualMachine::stop()
{
}

std::string te::vm::lua::VirtualMachine::getName() const
{
  return "Lua";
}

std::string te::vm::lua::VirtualMachine::getTitle() const
{
  std::string msg(
      TE_TR("TerraLib Virtual Machine for the Lua Programming Language"));

  return msg;
}

std::string te::vm::lua::VirtualMachine::getDescription() const
{
  std::string msg(
      TE_TR("This Virtual Machine allows applications to execute lua scripts."
            "\nAll the TerraLib API will be available for TerraLib programmers "
            "in the Lua environment"));

  return msg;
}
