/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralin/vp/AbstractDataHandler.h

   \brief This abstract class represents handler to help reading and writing data from/to a dataSource by dividing it into smaller groups
 */

#ifndef __TERRALIB_VP_INTERNAL_ABSTRACTDATAHANDLER_H
#define __TERRALIB_VP_INTERNAL_ABSTRACTDATAHANDLER_H

#include "../common/AbstractFactory.h"
#include "Config.h"


#include <vector>

namespace te
{
  namespace common
  {
    class TaskProgress;
  }
  namespace da
  {
    class DataAccess;
    class DataSetType;
  }

  namespace gm
  {
    class Envelope;
  }

  namespace vp
  {
    class AbstractParameters;
    class FeatureSet;

    /*!
      \class AbstractDataHandler

      \brief This abstract class represents handler to help reading and writing data from/to a dataSource by dividing it into smaller groups
     */
    class TEVPEXPORT AbstractDataHandler
    {
    public:
      /*! \brief Default constructor. */
      AbstractDataHandler();

      /*! \brief Virtual destructor. */
      virtual ~AbstractDataHandler()= default;

      //!< Initializes the data handler
      virtual void initialize(te::vp::AbstractParameters* absParams, te::common::TaskProgress* taskProgress);

      //!< Finalizes the data handler
      virtual void finalize(te::common::TaskProgress* taskProgress);

      //!< Gets the datatype associated with the expected output
      virtual const te::da::DataSetType* getOutputDataType() const;

      //!< Returns the number of groups of the data handler
      virtual std::size_t getGroupCount() const = 0;

      //!< Returns the number of inputs of the data handler
      virtual std::size_t getInputCount() const;

      //!< Returns the number of output of the data handler
      virtual std::size_t getOutputCount() const;

      //!< Returns the inputIndex-th featureSet og the groupIndex-th group considering the given filter
      virtual te::vp::FeatureSet getFeatureSet(std::size_t groupIndex, std::size_t inputIndex, const te::gm::Envelope& filter) = 0;

      //!< Returns the inputIndex-th featureSet og the groupIndex-th group
      virtual te::vp::FeatureSet getFeatureSet(std::size_t groupIndex, std::size_t inputIndex);

      //!< Adds a featureSet related to groupIndex-th group the to be added to the outputIndex-th output data access
      virtual void addFeatureSet(std::size_t groupIndex, std::size_t outputIndex, const te::vp::FeatureSet& featureSet) = 0;

      virtual const std::vector<te::da::DataAccess*>& getInputDataAccessVector() const;

      virtual const std::vector<te::da::DataAccess*>& getOutputDataAccessVector() const;

    protected:
      std::vector<te::da::DataAccess*> m_vecInputData;
      std::vector<te::da::DataAccess*> m_vecOutputData;
      std::size_t m_maxCoordinates;
    };

    /*!
      \class DataHandlerFactory

      \brief Abstract factory used to create Data Handler.
    */
    class TEVPEXPORT AbstractDataHandlerFactory : public te::common::AbstractFactory < te::vp::AbstractDataHandler, std::string >
    {
    protected:

      /*!
        \brief Constructor.

        \param factoryKey The key that identifies the factory.
      */
      AbstractDataHandlerFactory(const std::string& factoryKey);

      //!< Destructor
      virtual ~AbstractDataHandlerFactory();
    };
  }
}

#endif //__TERRALIB_VP_INTERNAL_ABSTRACTDATAHANDLER_H
