/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */
/*!
/*!
\file terralib/vp/AbstractOperation.cpp

\brief Abstract Class to represent an abstract vector operation.
*/
#include "AbstractOperation.h"

#include "../common/StringUtils.h"
#include "../common/STLUtils.h"
#include "../core/translator/Translator.h"
#include "../datatype/SimpleProperty.h"
#include "../dataaccess/dataset/DataSetAdapter.h"
#include "../dataaccess/dataset/DataSetTypeConverter.h"
#include "../dataaccess/utils/DataAccess.h"
#include "../dataaccess/utils/Utils.h"
#include "../geometry/GeometryProperty.h"
#include "../geometry/Utils.h"
#include "Feature.h"
#include "OperationReport.h"
#include "OperationService.h"


//Class AbstractParameters
te::vp::AbstractParameters::AbstractParameters()
  : m_swapInput(false)
  , m_snapGeometries(true)
  , m_operationResult(OperationResult::OP_SUCCESS)
  , m_maxCoordinates(1000)
{

}

bool te::vp::AbstractParameters::isValid(std::string& errorMessage)
{
  if (m_inputDataVector.empty() == true)
  {
    errorMessage = TE_TR("Input layer list cannot be empty");
    return false;
  }

  if (m_outputDataVector.empty() == true)
  {
    errorMessage = TE_TR("Output layer list cannot be empty");
    return false;
  }

  for (std::size_t i = 0; i < m_inputDataVector.size(); ++i)
  {
    if (m_inputDataVector.at(i)->getSRID() == TE_UNKNOWN_SRS)
    {
      errorMessage = TE_TR("All input layers must have a valid SRID");
      return false;
    }
  }

  if (m_maxCoordinates > 0 && m_maxCoordinates < 5)
  {
    errorMessage = TE_TR("Max coordinates must be zero or higher than 4");
    return false;
  }

  return true;
}

void te::vp::AbstractParameters::reportParams(OperationReport* report)
{
  report->addSubTitle(TE_TR("Operation Input Parameters"));
  report->addEmptyLine();

  for (std::size_t t = 0; t < m_inputDataVector.size(); ++t)
  {
    te::da::DataAccess* dataAccess = m_inputDataVector.at(t);
    report->addKeyValueMessage(TE_TR("Input Data Access"), std::to_string(t));
    report->addKeyValueMessage(TE_TR("Data Source URI"), dataAccess->getDataSourceConnectionInfo());
    report->addKeyValueMessage(TE_TR("Data Set Name"), dataAccess->getDataSetName());
    report->addKeyValueMessage(TE_TR("Data Set SRID"), te::common::Convert2String(static_cast<boost::int32_t>(dataAccess->getSRID())));
    report->addKeyValueMessage(TE_TR("Query Restriction"), dataAccess->getStrQuery());
    report->addEmptyLine();
  }

  report->addEmptyLine();
  report->addSubTitle(TE_TR("Operation Output Parameters"));
  report->addEmptyLine();

  for (std::size_t t = 0; t < m_outputDataVector.size(); ++t)
  {
    te::da::DataAccess* dataAccess = m_outputDataVector.at(t);
    report->addKeyValueMessage(TE_TR("Output Data Access"), std::to_string(t));
    report->addKeyValueMessage(TE_TR("Data Source URI"), dataAccess->getDataSourceConnectionInfo());
    report->addKeyValueMessage(TE_TR("Data Set Name"), dataAccess->getDataSetName());
    report->addKeyValueMessage(TE_TR("Query Restriction"), dataAccess->getStrQuery());
    report->addEmptyLine();
  }

  report->addEmptyLine();
  report->addSubTitle(TE_TR("Operation Specific Parameters"));
  report->addEmptyLine();

  reportSpecificParams(report);
}

void te::vp::AbstractParameters::addInputDataAccess(te::da::DataAccess* dataAccess)
{
  m_inputDataVector.push_back(dataAccess);
}

void te::vp::AbstractParameters::addOutputDataAccess(te::da::DataAccess* dataAccess)
{
  m_outputDataVector.push_back(dataAccess);
}

const std::vector<te::da::DataAccess*>& te::vp::AbstractParameters::getInputDataAccessVector() const
{
  return m_inputDataVector;
}

const std::vector<te::da::DataAccess*>& te::vp::AbstractParameters::getOutputDataAccessVector() const
{
  return m_outputDataVector;
}

void te::vp::AbstractParameters::setInputDataAccessVector(const std::vector<te::da::DataAccess*>& vecDataAccess)
{
  m_inputDataVector = vecDataAccess;
}

void te::vp::AbstractParameters::setOutputDataAccessVector(const std::vector<te::da::DataAccess*>& vecDataAccess)
{
  m_outputDataVector = vecDataAccess;
}

void te::vp::AbstractParameters::setMaxCoordinates(std::size_t maxCoordinates)
{
  m_maxCoordinates = maxCoordinates;
}

std::size_t te::vp::AbstractParameters::getMaxCoordinates() const
{
  return m_maxCoordinates;
}

void te::vp::AbstractParameters::setSwapInput(bool swap)
{
  m_swapInput = swap;
}

bool te::vp::AbstractParameters::getSwapInput() const
{
  return m_swapInput;
}

te::vp::OperationResult te::vp::AbstractParameters::getOperationResult() const
{
  return m_operationResult;
}

void te::vp::AbstractParameters::setOperationResult(OperationResult result)
{
  m_operationResult = result;
}

void te::vp::AbstractParameters::reportSpecificParams(OperationReport* report)
{

}

//Class AbstractOperationCapabilities
te::vp::AbstractOperationCapabilities::AbstractOperationCapabilities(const std::string& operationName, std::size_t numberOfInputs, std::size_t numberOfOutputs, const std::string& dataHandlerName, bool snapGeometries, bool allowCoordinateSubdivision)
  : m_operationName(operationName)
  , m_numberOfInputs(numberOfInputs)
  , m_numberOfOutputs(numberOfOutputs)
  , m_dataHandlerName(dataHandlerName)
  , m_snapGeometries(snapGeometries)
  , m_allowCoordinateSubdivision(allowCoordinateSubdivision)
{

}

te::vp::AbstractOperationCapabilities::~AbstractOperationCapabilities() = default;

std::size_t te::vp::AbstractOperationCapabilities::getNumberOfInputs() const
{
  return m_numberOfInputs;
}

std::size_t te::vp::AbstractOperationCapabilities::getNumberOfOutputs() const
{
  return m_numberOfOutputs;
}

const std::string& te::vp::AbstractOperationCapabilities::getName() const
{
  return m_operationName;
}

const std::string& te::vp::AbstractOperationCapabilities::getDataHandlerName() const
{
  return m_dataHandlerName;
}

bool te::vp::AbstractOperationCapabilities::getSnapGeometries() const
{
  return m_snapGeometries;
}

bool te::vp::AbstractOperationCapabilities::getAllowCoordinateSubdivision() const
{
  return m_allowCoordinateSubdivision;
}

std::unique_ptr<te::vp::AbstractParameters> te::vp::AbstractOperationCapabilities::createParameters() const
{
  std::unique_ptr<te::vp::AbstractParameters> params(new te::vp::AbstractParameters());
  return params;
}

struct te::vp::AbstractOperationImpl
{
  std::string m_operationName;
  std::unique_ptr<te::vp::AbstractParameters> m_params;
  std::unique_ptr<te::vp::OperationReport> m_report;
  std::vector<te::da::DataSetTypePtr> m_outputDataSetTypeVector;
  bool m_isInitialized;
  int m_outputSrid;
  te::gm::GeomType m_outputGeomType;

  AbstractOperationImpl(const std::string& operationName)
    : m_operationName(operationName)
    , m_report(new te::vp::OperationReport())
    , m_isInitialized(false)
    , m_outputSrid(0)
    , m_outputGeomType(te::gm::UnknownGeometryType)  
  {
    
  }

  void initialize(te::vp::AbstractParameters* params)
  {
    m_params.reset(params);

    const std::vector<te::da::DataAccess*>& vecInput = m_params->getInputDataAccessVector();
    te::da::DataAccess* baseInput = vecInput.at(0);
    const te::da::DataSetType* dataSetType = baseInput->getDataSetType();
    te::gm::GeometryProperty* geometryProperty = te::da::GetFirstGeomProperty(dataSetType);

    //the srid and geometry type will be always based on the ones from the first input dataAccess
    m_outputSrid = baseInput->getSRID();
    m_outputGeomType = te::gm::GetMultiType( geometryProperty->getGeometryType() );

    m_report->addTitle(m_operationName);
    m_params->reportParams(m_report.get());
  }
};

te::vp::AbstractOperation::AbstractOperation(const std::string& operationName)
  : m_impl(new AbstractOperationImpl(operationName))
  , m_aborted(false)
{
  
}

te::vp::AbstractOperation::~AbstractOperation()
{
  delete m_impl;
}

void te::vp::AbstractOperation::initialize()
{
  m_aborted = false;

  if (m_impl->m_isInitialized == false)
  {
    if (initializeInput() == false)
    {
      throw te::common::Exception(TE_TR("Error initializing the operation"));
    }
    if (initializeOutput() == false)
    {
      throw te::common::Exception(TE_TR("Error initializing the operation"));
    }
    m_impl->m_isInitialized = true;
  }
}

std::vector<te::vp::FeatureSet> te::vp::AbstractOperation::execute(const std::vector<te::vp::FeatureSet>& vecInput)
{
  if (m_impl->m_isInitialized == false)
  {
    this->getParameters()->setOperationResult(te::vp::OperationResult::OP_ERROR);
    throw te::common::Exception(TE_TR("Operation must be initialized before calling the execute method"));
  }

  std::vector<te::vp::FeatureSet> vecOutputFeatureSet;
  try
  {
    if (this->getCapabilities()->getSnapGeometries() == false)
    {
      vecOutputFeatureSet = executeImpl(vecInput);
    }
    else
    {
      std::vector<te::vp::FeatureSet> vecInputSnapped = te::vp::OperationService::SnapGeometries(vecInput);
      vecOutputFeatureSet = executeImpl(vecInputSnapped);

      for (std::size_t i = 0; i < vecInputSnapped.size(); ++i)
      {
        te::common::FreeContents(vecInputSnapped.at(i));
      }
    }
  }
  catch (const std::exception&)
  {
    this->getParameters()->setOperationResult(te::vp::OperationResult::OP_ERROR);
    abort();
    throw;
  }
  catch (...)
  {
    this->getParameters()->setOperationResult(te::vp::OperationResult::OP_ERROR);
    abort();
    throw te::common::Exception(TE_TR("Unkown exception during operation execution"));
  }

  return vecOutputFeatureSet;
}

te::vp::Feature* te::vp::AbstractOperation::createOutputFeature(std::size_t outputIndex)
{
  te::vp::Feature* feature = new te::vp::Feature(getDataSetType(outputIndex));
  return feature;
}

const te::da::DataSetType* te::vp::AbstractOperation::getDataSetType(std::size_t outputIndex)
{
  te::da::DataAccess* outputDataAccess = m_impl->m_params->getOutputDataAccessVector().at(outputIndex);
  return outputDataAccess->getDataSetType();
}

void te::vp::AbstractOperation::setParameters(te::vp::AbstractParameters* params)
{
  m_impl->initialize(params);
}

te::vp::AbstractParameters* te::vp::AbstractOperation::getParameters() const
{
  return m_impl->m_params.get();
}

te::vp::OperationReport* te::vp::AbstractOperation::getReport() const
{
  return m_impl->m_report.get();
}

const std::string& te::vp::AbstractOperation::getName() const
{
  return m_impl->m_operationName;
}

te::gm::GeomType te::vp::AbstractOperation::getOutputGeometryType() const
{
  return m_impl->m_outputGeomType;
}

void te::vp::AbstractOperation::setOutputGeometryType( 
  const te::gm::GeomType newType )
{
  m_impl->m_outputGeomType = newType;
}

void te::vp::AbstractOperation::abort()
{
  m_aborted = true;
}

bool te::vp::AbstractOperation::isAborted() const
{
  return m_aborted;
}

std::unique_ptr<te::vp::AbstractOperationCapabilities> te::vp::AbstractOperation::getCapabilities() const
{
  std::unique_ptr<te::vp::AbstractOperationCapabilities> capabilities(new te::vp::AbstractOperationCapabilities(getName(), 2, 1, "SubdivideDataHandler", true, true));
  return capabilities;
}

bool te::vp::AbstractOperation::prepareInputData()
{
  std::string errorMessage;
  if (m_impl->m_params == nullptr || m_impl->m_params->isValid(errorMessage) == false)
  {
    return false;
  }

  //we must create adapters to the input data so all the columns will have the unique name
  std::set<std::string> setColumnNames;

  const std::vector<te::da::DataAccess*>& vecInput = m_impl->m_params->getInputDataAccessVector();
  te::gm::GeometryProperty* geomProperty = te::da::GetFirstGeomProperty(vecInput.at(0)->getDataSetType());

  //we must protect the names that will be in the output dataSet
  std::unique_ptr<te::da::DataSetType> outputDataSetType(this->createBasicOutputDataSetType("", m_impl->m_outputSrid, m_impl->m_outputGeomType, geomProperty->getName()));
  for (std::size_t i = 0; i < outputDataSetType->getProperties().size(); ++i)
  {
    setColumnNames.insert(te::common::Convert2LCase(outputDataSetType->getProperty(i)->getName()));
  }

  //and now we adapt the input data access
  for (std::size_t i = 0; i < vecInput.size(); ++i)
  {
    te::da::DataAccess* dataAccess = vecInput.at(i);

    //we create a conveter just to ensure that the property names are unique
    te::da::DataSetTypeConverter* converter = nullptr;
    if (dataAccess->getConverter() == nullptr)
    {
      converter = new te::da::DataSetTypeConverter(dataAccess->getDataSetType(), dataAccess->getDataSource()->getCapabilities());
      dataAccess->setConverter(converter);
    }
    else
    {
      converter = dataAccess->getConverter();
    }

    te::da::DataSetType* dataSetTypeConverted = converter->getResult();
    te::da::PrimaryKey* primaryKey = dataSetTypeConverted->getPrimaryKey();

    for (std::size_t j = 0; j < dataSetTypeConverted->getProperties().size(); ++j)
    {
      //we must ensure that the names are unique considering ALL the input dataAccess
      te::dt::Property* currentProperty = dataSetTypeConverted->getProperty(j);

      if (primaryKey != nullptr && primaryKey->has(currentProperty->getName()) == true)
      {
        continue;
      }

      //as the geometry columns will not be part of the output layer, we must not change it
      if (currentProperty->getType() == te::dt::GEOMETRY_TYPE)
      {
        continue;
      }

      te::dt::SimpleProperty* currentSimpleProperty = dynamic_cast<te::dt::SimpleProperty*>(currentProperty);
      if (currentSimpleProperty != nullptr)
      {
        currentSimpleProperty->setRequired(false);
        currentSimpleProperty->setAutoNumber(false);
        currentSimpleProperty->setDefaultValue(nullptr);
      }

      std::unique_ptr<te::dt::Property> newProperty(currentProperty->clone());
      std::string baseName = te::common::Convert2LCase(currentProperty->getName());
      std::string newName = baseName;
      std::size_t count = 1;
      while (setColumnNames.find(newName) != setColumnNames.end())
      {
        //we must adapt the name of the column
        newName = baseName + te::common::Convert2String(static_cast<boost::uint32_t>(count));
        ++count;
      }

      //count > 1 means that this property had to be renamed. In this case, we replace it in the dataSet converted
      if (count > 1)
      {
        newProperty->setName(newName);
        dataSetTypeConverted->replace(currentProperty, newProperty.release());

        delete currentProperty;
      }

      setColumnNames.insert(newName);
    }
  }

  return true;
}

bool te::vp::AbstractOperation::prepareOutputData()
{
  std::string errorMessage;
  if (m_impl->m_params == nullptr || m_impl->m_params->isValid(errorMessage) == false)
  {
    return false;
  }

  const std::vector<te::da::DataAccess*>& vecInput = m_impl->m_params->getInputDataAccessVector();
  te::gm::GeometryProperty* geomProperty = te::da::GetFirstGeomProperty(vecInput.at(0)->getDataSetType());

  const std::vector<te::da::DataAccess*>& vecOutput = m_impl->m_params->getOutputDataAccessVector();
  for (std::size_t i = 0; i < vecOutput.size(); ++i)
  {
    te::da::DataAccess* outputDataAccess = vecOutput.at(i);

    //check if the output dataSet needs to be created
    if (outputDataAccess->dataSetExists() == true)
    {
      continue;
    }

    //if it does not exist yet, we create the basic dataSetType
    std::string outputDataSetName = outputDataAccess->getDataSetName();
    std::unique_ptr<te::da::DataSetType> outputDataSetType(createBasicOutputDataSetType(outputDataSetName, m_impl->m_outputSrid, m_impl->m_outputGeomType, geomProperty->getName()));

    //then we populate it if complementary properties
    addOutputColumns(outputDataSetType.get());

    //and we create the output dataSet
    outputDataAccess->createDataSet(outputDataSetType.get());

    //as some dataSources change the name of the created properties, we must handle it
    //we check for changes in the name of the properties, and update this name in the input adapter so all the properties will have the same name as the output dataSet
    const te::da::DataSetType* actualOutputDataSetType = outputDataAccess->getDataSetType();

    if (handleOutputPropertyNamesChanges(outputDataSetType.get(), actualOutputDataSetType) == false)
    {
      errorMessage = TE_TR("Error handling the property names for the current selected output");
      return false;
    }
  }

  return true;
}

bool te::vp::AbstractOperation::handleOutputPropertyNamesChanges(const te::da::DataSetType* baseOutputDataSetType, const te::da::DataSetType* actualOutputDataSetType)
{
  //creates a list of all columns discarding FID and Geometry properties of baseOutputDataSetType
  std::vector<std::string> vecBaseNames;
  for (std::size_t i = 0; i < baseOutputDataSetType->size(); ++i)
  {
    const te::dt::Property* property = baseOutputDataSetType->getProperty(i);
    if (te::common::IsStringEqual(property->getName(), "FID", false) || property->getType() == te::dt::GEOMETRY_TYPE)
    {
      continue;
    }

    vecBaseNames.push_back(property->getName());
  }

  //creates a list of all columns discarding FID and Geometry properties of actualOutputDataSetType
  std::vector<std::string> vecActualNames;
  for (std::size_t i = 0; i < actualOutputDataSetType->size(); ++i)
  {
    const te::dt::Property* property = actualOutputDataSetType->getProperty(i);
    if (te::common::IsStringEqual(property->getName(), "FID", false) || property->getType() == te::dt::GEOMETRY_TYPE)
    {
      continue;
    }

    vecActualNames.push_back(property->getName());
  }

  if (vecBaseNames.size() != vecActualNames.size())
  {
    return false;
  }

  const std::vector<te::da::DataAccess*>& vecInput = m_impl->m_params->getInputDataAccessVector();

  //if any name has changed, we mus handle it by changing the adapters of the input data
  for (std::size_t i = 0; i < vecBaseNames.size(); ++i)
  {
    const std::string baseName = vecBaseNames.at(i);
    const std::string actualName = vecActualNames.at(i);

    if (baseName != actualName)
    {
      bool changed = false;
      for (std::size_t j = 0; j < vecInput.size(); ++j)
      {
        te::da::DataAccess* dataAccess = vecInput.at(j);
        te::da::DataSetTypeConverter* converter = dataAccess->getConverter();
        te::da::DataSetType* dataSetTypeConverted = converter->getResult();
        for (std::size_t k = 0; k < dataSetTypeConverted->size(); ++k)
        {
          te::dt::Property* currentProperty = dataSetTypeConverted->getProperty(k);
          if (currentProperty->getName() == baseName)
          {
            std::unique_ptr<te::dt::Property> newProperty(currentProperty->clone());
            newProperty->setName(actualName);

            dataSetTypeConverted->replace(currentProperty, newProperty.release());

            delete currentProperty;

            changed = true;
            break;
          }
        }
        if (changed == true)
        {
          break;
        }
      }
      if (changed == false)
      {
        return false;
      }
    }
  }

  return true;
}

te::da::DataSetType* te::vp::AbstractOperation::createBasicOutputDataSetType(const std::string& dataSetName, int srid, te::gm::GeomType geometryType, const std::string& geometryColumnName) const
{
  te::da::DataSetType* outputDataSetType = new te::da::DataSetType(dataSetName);
  outputDataSetType->setDatasetName(dataSetName);
  outputDataSetType->setTitle(dataSetName);

  //id
  {
    te::dt::SimpleProperty* absProperty = new te::dt::SimpleProperty("gid", te::dt::INT32_TYPE, true);
    absProperty->setAutoNumber(true);
    outputDataSetType->add(absProperty);

    //this also adds the the primeryKey to the outputDataSetType
    std::string pkName = dataSetName +  + "_id_pk";
    te::da::PrimaryKey* primaryKey = new te::da::PrimaryKey(pkName, outputDataSetType);
    primaryKey->add(absProperty);
  }

  //geom
  {
    te::gm::GeometryProperty* absProperty = new te::gm::GeometryProperty(geometryColumnName, 0, geometryType, true);
    absProperty->setSRID(srid);
    outputDataSetType->add(absProperty);
  }

  return outputDataSetType;
}

void te::vp::AbstractOperation::addOutputColumns(te::da::DataSetType* outputDataSetType)
{
  const std::vector<te::da::DataAccess*>& vecInput = m_impl->m_params->getInputDataAccessVector();
  for (std::size_t i = 0; i < vecInput.size(); ++i)
  {
    te::da::DataAccess* dataAccess = vecInput.at(i);
    const te::da::DataSetType* dataSetType = dataAccess->getDataSetType();
    te::da::PrimaryKey* primaryKey = dataSetType->getPrimaryKey();

    for (std::size_t j = 0; j < dataSetType->getProperties().size(); ++j)
    {
      std::unique_ptr<te::dt::Property> newProperty(dataSetType->getProperty(j)->clone());
      if (primaryKey != nullptr && primaryKey->has(newProperty->getName()) == true)
      {
        continue;
      }

      //we must exclude the all geometry columns as the output must have only one column of this type
      if (newProperty->getType() == te::dt::GEOMETRY_TYPE)
      {
        continue;
      }

      newProperty->setName(te::common::Convert2LCase(newProperty->getName()));

      outputDataSetType->add(newProperty.release());
    }
  }
}

bool te::vp::AbstractOperation::initializeInput()
{
  return true;
}

bool te::vp::AbstractOperation::initializeOutput()
{
  return true;
}

te::vp::OperationFactory::OperationFactory(const std::string& factoryKey)
  : te::common::AbstractFactory<te::vp::AbstractOperation, std::string>(factoryKey)
{

}

te::vp::OperationFactory::~OperationFactory() = default;
