/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralin/vp/AttributeGroupingDataHandler.cpp

   \brief This class represents handler to help reading and writing data from/to a dataSource by dividing it into smaller groups
 */

#include "AttributeGroupingDataHandler.h"

#include "../common/StringUtils.h"
#include "../common/STLUtils.h"
#include "../dataaccess/dataset/DataSet.h"
#include "../dataaccess/dataset/DataSetType.h"
#include "../dataaccess/utils/BasicQueries.h"
#include "../dataaccess/utils/DataAccess.h"
#include "../dataaccess/utils/Utils.h"
#include "../memory/DataSet.h"
#include "../memory/PreparedItemCopy.h"
#include "DissolveSubdivisionOperation.h"
#include "SubdivideDataHandler.h"
#include "Feature.h"
#include "Utils.h"

 //Boost include files
#include <boost/thread/mutex.hpp>

namespace
{
  static te::vp::AttributeGroupingDataHandlerFactory AttributeGroupingDataHandlerFactoryInstance;
}

class te::vp::AttributeGroupingDataHandlerImpl
{
public:
 
  AttributeGroupingDataHandlerImpl()
    : m_originalInputData(nullptr)
  {
  }

  ~AttributeGroupingDataHandlerImpl()
  {
    te::vp::DropTemporaryDataSource(m_tempDataSource);
  }

  te::da::DataAccess* m_originalInputData;
  te::da::DataSourcePtr m_tempDataSource;
  std::unique_ptr<te::da::DataAccess> m_tempDataAccess;
  std::vector<std::string> m_vecGroupColumns;
  std::vector<te::vp::SubdivideKey> m_vecGroupValues;
  boost::mutex m_mutexRead;
  boost::mutex m_mutexWrite;
};

te::vp::AttributeGroupingDataHandler::AttributeGroupingDataHandler()
    : AbstractDataHandler()
    , m_impl(new AttributeGroupingDataHandlerImpl())
{

}

te::vp::AttributeGroupingDataHandler::~AttributeGroupingDataHandler() = default;

te::da::DataAccess* CreateOutputDataAccess(const te::da::DataSetType* inputDataType, te::da::DataSourcePtr outputDataSource, const std::string& outputDataSetName)
{
  if (outputDataSource->isOpened() == false)
  {
    outputDataSource->open();
  }

  std::string lowerName = te::common::Convert2LCase(outputDataSetName);

  std::unique_ptr<te::da::DataSetType> outputDataSetType(static_cast<te::da::DataSetType*>(inputDataType->clone()));
  outputDataSetType->setName(lowerName);
  outputDataSetType->setTitle(lowerName);
  outputDataSetType->setDatasetName(lowerName);

  te::da::DataAccess* outputDataAccess = new te::da::DataAccess(outputDataSource, outputDataSetType->getName());
  outputDataAccess->createDataSet(outputDataSetType.get());
  return outputDataAccess;
}

te::da::DataAccess* CreateTemporaryCopy(te::da::DataAccess* inputDataAccess, te::da::DataSourcePtr tempDataSource)
{
  //copies the data access
  const te::da::DataSetType* inputDataSetType = inputDataAccess->getDataSetType();

  te::da::DataAccess* outputDataAccess = CreateOutputDataAccess(inputDataSetType, tempDataSource, inputDataSetType->getName());
  const te::da::DataSetType* outputDataSetType = outputDataAccess->getDataSetType();

  std::unique_ptr<te::da::DataSet> inputDataSet = inputDataAccess->getDataSet();
  std::unique_ptr<te::mem::DataSet> outputDataSet(new te::mem::DataSet(outputDataSetType));

  te::mem::PreparedItemCopy preparedItemCopy;

  inputDataSet->moveBeforeFirst();
  while (inputDataSet->moveNext())
  {
    std::unique_ptr<te::da::DataSetItem> source = inputDataSet->getItem();
    te::mem::DataSetItem* destination = new te::mem::DataSetItem(outputDataSetType);

    preparedItemCopy.copyValues(source.get(), destination);

    outputDataSet->add(destination);
  }

  outputDataAccess->saveDataSet(outputDataSet.release());

  return outputDataAccess;
}

void te::vp::AttributeGroupingDataHandler::initialize(te::vp::AbstractParameters* absParams, te::common::TaskProgress* taskProgress)
{
  AbstractDataHandler::initialize(absParams, taskProgress);

  if (m_vecInputData.size() != 1)
  {
    return;
  }

  te::vp::DissolveSubdivisionOperationParameters* params = dynamic_cast<te::vp::DissolveSubdivisionOperationParameters*>(absParams);
  if (params == nullptr)
  {
    return;
  }

  //we must copy the data from the input dataaccess so we can do queries and use the data correctly

  te::da::DataAccess* originalInputDataAccess = m_vecInputData.at(0);
  m_impl->m_originalInputData = originalInputDataAccess;

  m_impl->m_tempDataSource = te::vp::CreateTemporaryDataSource();
  m_impl->m_tempDataAccess.reset(CreateTemporaryCopy(originalInputDataAccess, m_impl->m_tempDataSource));
  this->m_vecInputData.at(0) = m_impl->m_tempDataAccess.get();
  
  te::da::DataAccess* dataAccess = this->m_vecInputData.at(0);

  const te::da::DataSetType* dataSetType = dataAccess->getDataSetType();
  const te::da::PrimaryKey* primaryKey = dataSetType->getPrimaryKey();
  for (std::size_t i = 0; i < dataSetType->size(); ++i)
  {
    te::dt::Property* property = dataSetType->getProperty(i);
    if (property->getType() == te::dt::GEOMETRY_TYPE)
    {
      continue;
    }

    std::string propertyName = property->getName();
    if (primaryKey != nullptr && primaryKey->has(propertyName) == true)
    {
      continue;
    }

    if (te::common::IsStringEqual(propertyName, params->getRowPropertyName(), false) == true)
    {
      continue;
    }

    if (te::common::IsStringEqual(propertyName, params->getColumnPropertyName(), false) == true)
    {
      continue;
    }

    m_impl->m_vecGroupColumns.push_back(propertyName);
  }

  std::string query = "SELECT DISTINCT " + te::common::Convert2String(m_impl->m_vecGroupColumns, ",") + " FROM " + dataSetType->getName();
  std::unique_ptr<te::da::DataSet> dataSet = dataAccess->getDataSource()->query(query);
  dataSet->moveBeforeFirst();
  while (dataSet->moveNext())
  {
    std::vector<te::dt::AbstractData*> vecKeys;
    for (std::size_t j = 0; j < m_impl->m_vecGroupColumns.size(); ++j)
    {
      te::dt::AbstractData* absData = nullptr;
      if (dataSet->isNull(j) == false)
      {
        absData = dataSet->getValue(j).release();
      }

      vecKeys.push_back(absData);
    }

    te::vp::SubdivideKey key(vecKeys);
    m_impl->m_vecGroupValues.push_back(key);
  }
}

void te::vp::AttributeGroupingDataHandler::finalize(te::common::TaskProgress* taskProgress)
{
  this->m_vecInputData.at(0) = m_impl->m_originalInputData;
}

std::size_t te::vp::AttributeGroupingDataHandler::getGroupCount() const
{
  return m_impl->m_vecGroupValues.size();
}

te::vp::FeatureSet te::vp::AttributeGroupingDataHandler::getFeatureSet(std::size_t groupIndex, std::size_t inputIndex, const te::gm::Envelope& filter)
{
  std::unique_ptr<te::da::DataSet> dataSet;

  const std::vector<std::string>& vecGroupColumns = m_impl->m_vecGroupColumns;
  const std::vector<te::dt::AbstractData*>& vecGroupKeys = m_impl->m_vecGroupValues.at(groupIndex).m_vecKeys;
  te::da::DataAccess* dataAccess = m_vecInputData.at(inputIndex);

  const te::da::DataSetType* dataSetType = dataAccess->getDataSetType();

  std::vector<std::string> vecColumns;
  for (std::size_t i = 0; i < dataSetType->size(); ++i)
  {
    vecColumns.push_back(dataSetType->getProperty(i)->getName());
  }

  std::unique_ptr<te::da::Select> selectGroup = te::da::BasicQueries::GetQuery(false, vecColumns, m_vecInputData.at(inputIndex)->getDataSetName(), vecGroupColumns, vecGroupKeys);

  //creates a scope so the locker can be used just in the query of data
  {
    boost::mutex::scoped_lock lockGuard(m_impl->m_mutexRead);
    dataSet = dataAccess->query(*selectGroup);
  }

  //declares the resulting featureset
  te::vp::FeatureSet featureSet;
  if (dataSet == nullptr)
  {
    return featureSet;
  }

  if (dataSet->size() == 0)
  {
    return featureSet;
  }

  dataSet->moveBeforeFirst();
  while (dataSet->moveNext())
  {
    te::vp::Feature* feature = te::vp::Feature::createFeatureFromCurrent(dataSet.get());
    featureSet.push_back(feature);
  }

  return featureSet;
}

void te::vp::AttributeGroupingDataHandler::addFeatureSet(std::size_t groupIndex, std::size_t outputIndex, const te::vp::FeatureSet& featureSet)
{
  te::da::DataAccess* dataAccess = m_vecOutputData.at(outputIndex);
  std::unique_ptr<te::da::DataSet> dataSet = te::vp::CreateDataSetFromFeatureSet(dataAccess->getDataSetName(), dataAccess->getDataSetType(), featureSet);

  //the lock is only used in the saving of the geometries
  boost::mutex::scoped_lock lockGuard(m_impl->m_mutexWrite);
  dataAccess->saveDataSet(dataSet.release(), false);
}

te::vp::AttributeGroupingDataHandlerFactory::AttributeGroupingDataHandlerFactory()
  : te::vp::AbstractDataHandlerFactory("AttributeGroupingDataHandler")
{

}

te::vp::AttributeGroupingDataHandlerFactory::~AttributeGroupingDataHandlerFactory() = default;

te::vp::AbstractDataHandler* te::vp::AttributeGroupingDataHandlerFactory::build()
{
  return new te::vp::AttributeGroupingDataHandler();
}