/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralib/vp/CommonDataStructures.cpp

   \brief Utility classes, structures and definitions for Vector Processing.
 */

#include "CommonDataStructures.h"
#include "CommonFunctions.h"

#include "RTreeIndex.h"
#include "../geometry/LineString.h"

#include <memory>

te::vp::SegmentInfo::SegmentInfo(const te::gm::Coord2D& coord1, const te::gm::Coord2D& coord2, const te::gm::Envelope& envelope, std::size_t geometryIndex)
  : m_coord1(coord1)
  , m_coord2(coord2)
  , m_envelope(envelope)
  , m_geometryIndex(geometryIndex)
{

}

te::vp::SegmentInfo::~SegmentInfo()
{
}

te::vp::SegmentInfo* te::vp::SegmentInfo::clone() const
{
  SegmentInfo* newSegmentInfo = new SegmentInfo(m_coord1, m_coord2, m_envelope, m_geometryIndex);
  return newSegmentInfo;
}

const te::gm::Coord2D& te::vp::SegmentInfo::getCoord1() const
{
  return m_coord1;
}

const te::gm::Coord2D& te::vp::SegmentInfo::getCoord2() const
{
  return m_coord2;
}

const te::gm::Envelope& te::vp::SegmentInfo::getEnvelope() const
{
  return m_envelope;
}

std::size_t te::vp::SegmentInfo::getGeometryIndex() const
{
  return m_geometryIndex;
}

void te::vp::SegmentInfo::setGeometryIndex(std::size_t geometryIndex)
{
  m_geometryIndex = geometryIndex;
}

struct te::vp::GeometryInfoImpl
{
  GeometryInfoImpl(std::size_t geometryIndex, te::gm::Geometry* geometry)
    : m_geometryIndex(geometryIndex)
    , m_geometry(geometry)
  {
    init(te::gm::Envelope());
  }

  GeometryInfoImpl(std::size_t geometryIndex, te::gm::Geometry* geometry, const te::gm::Envelope& filter)
    : m_geometryIndex(geometryIndex)
    , m_geometry(geometry)
  {
    init(filter);
  }

  ~GeometryInfoImpl()
  {
    m_index.clear();
  }

  void init(const te::gm::Envelope& filter)
  {
    m_index.clear();

    if (m_geometry == nullptr)
    {
      return;
    }

    std::vector<te::vp::SegmentInfo> vecSegmentInfo;
    GetAllSegments(m_geometry.get(), filter, vecSegmentInfo);

    for (std::size_t i = 0; i < vecSegmentInfo.size(); ++i)
    {
      te::vp::SegmentInfo& segmentInfo = vecSegmentInfo.at(i);
      segmentInfo.setGeometryIndex(m_geometryIndex);
      m_index.insert(vecSegmentInfo.at(i).getEnvelope(), segmentInfo);
    }
  }

  std::size_t m_geometryIndex;
  std::unique_ptr<te::gm::Geometry> m_geometry;
  te::vp::RTreeIndex<te::vp::SegmentInfo> m_index;
};

te::vp::GeometryInfo::GeometryInfo(std::size_t geometryIndex, te::gm::Geometry* geometry)
  : m_impl(new GeometryInfoImpl(geometryIndex, geometry))
{
}

te::vp::GeometryInfo::GeometryInfo(std::size_t geometryIndex, te::gm::Geometry* geometry, const te::gm::Envelope& filter)
  : m_impl(new GeometryInfoImpl(geometryIndex, geometry, filter))
{

}

te::vp::GeometryInfo::~GeometryInfo()
{
  delete m_impl;
}

std::size_t te::vp::GeometryInfo::search(const te::gm::Envelope& filter, std::vector<SegmentInfo>& vecResult) const
{
  std::unique_ptr<te::vp::IndexReport<SegmentInfo>> indexReport = m_impl->m_index.search(filter);
  vecResult = indexReport->getAllData();
  return vecResult.size();
}

std::vector<te::vp::SegmentInfo> te::vp::GeometryInfo::getAllSegments() const
{
  std::unique_ptr<te::vp::IndexReport<te::vp::SegmentInfo>> indexReport = m_impl->m_index.getData();
  return indexReport->getAllData();
}

const te::gm::Geometry* te::vp::GeometryInfo::getGeometry() const
{
  return m_impl->m_geometry.get();
}

void te::vp::GeometryInfo::setGeometry(te::gm::Geometry* geometry)
{
  m_impl->m_geometry.reset(geometry);
  m_impl->init(te::gm::Envelope());
}
