/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralib/vp/CommonDataStructures.cpp

   \brief Utility classes, structures and definitions for Vector Processing.
 */

#include "CommonFunctions.h"
#include "CommonDataStructures.h"
#include "Feature.h"
#include "Utils.h"

#include "../geometry/Coord2D.h"
#include "../geometry/Envelope.h"
#include "../geometry/Geometry.h"
#include "../geometry/GeometryCollection.h"
#include "../geometry/GeometryFactory.h"
#include "../geometry/LineString.h"
#include "../geometry/Polygon.h"
#include "../geometry/Utils.h"

#include <memory>

void te::vp::GetAllSegments(const te::gm::Geometry* geometry, const te::gm::Envelope& filter, std::vector<te::vp::SegmentInfo>& vecSegmentInfo)
{
  if (geometry->isCollection() == true)
  {
    const te::gm::GeometryCollection* collection = static_cast<const te::gm::GeometryCollection*>(geometry);
    for (std::size_t i = 0; i < collection->getNumGeometries(); ++i)
    {
      GetAllSegments(collection->getGeometryN(i), filter, vecSegmentInfo);
    }
  }
  else if (geometry->get2DGeomTypeId() == te::gm::PolygonType)
  {
    const te::gm::Polygon* polygon = static_cast<const te::gm::Polygon*>(geometry);
    for (std::size_t i = 0; i < polygon->getNumRings(); ++i)
    {
      GetAllSegments(polygon->getRingN(i), filter, vecSegmentInfo);
    }
  }
  else if (geometry->get2DGeomTypeId() == te::gm::LineStringType)
  {
    const te::gm::LineString* line = static_cast<const te::gm::LineString*>(geometry);
    for (std::size_t i = 1; i < line->size(); ++i)
    {
      const double& x1 = line->getX(i - 1);
      const double& y1 = line->getY(i - 1);

      const double& x2 = line->getX(i);
      const double& y2 = line->getY(i);

      te::gm::Envelope envelope(te::gm::CreateEnvelope(x1, y1, x2, y2));
      if (filter.isValid() == true)
      {
        if (envelope.intersects(filter) == false)
        {
          continue;
        }
      }

      te::gm::Coord2D coord1(x1, y1);
      te::gm::Coord2D coord2(x2, y2);

      vecSegmentInfo.emplace_back(coord1, coord2, envelope, 0);
    }
  }
}

te::gm::SpatialRelation te::vp::FastRelationCalculator(const te::gm::Geometry* geometryA, const te::gm::Geometry* geometryB)
{
  const te::gm::Envelope* mbrA = geometryA->getMBR();
  const te::gm::Envelope* mbrB = geometryB->getMBR();

  if (mbrA->intersects(*mbrB) == false)
  {
    return te::gm::SpatialRelation::DISJOINT;
  }

  te::gm::Envelope intersectionEnvelope = mbrA->intersection(*mbrB);

  GeometryInfo geometryInfoA(0, static_cast<te::gm::Geometry*>(geometryA->clone()), intersectionEnvelope);
  GeometryInfo geometryInfoB(1, static_cast<te::gm::Geometry*>(geometryB->clone()), intersectionEnvelope);

  return te::vp::FastRelationCalculator(&geometryInfoA, &geometryInfoB);
}

te::gm::SpatialRelation te::vp::FastRelationCalculator(const te::vp::GeometryInfo* geometryInfoA, const te::vp::GeometryInfo* geometryInfoB)
{
  const te::gm::Envelope* mbrA = geometryInfoA->getGeometry()->getMBR();
  const te::gm::Envelope* mbrB = geometryInfoB->getGeometry()->getMBR();

  //if one geometry box is entirely inside the other, we must consider the "smaller" one as geometry A for performance purposes
  if (mbrB->within(*mbrA) == true && mbrB->equals(*mbrA) == false)
  {
    return FastRelationCalculator(geometryInfoB, geometryInfoA);
  }

  std::vector<te::vp::SegmentInfo> vecDataA;
  geometryInfoA->search(*mbrB, vecDataA);

  if (vecDataA.empty() == true)
  {
    //no overlap
    return te::gm::SpatialRelation::DISJOINT;
  }

  //we test one interior point of A to check if it is inside B. if so, the overlap
  te::gm::Coord2D coordA = te::gm::GetInteriorPoint(geometryInfoA->getGeometry());

  //if geometryA have one point inside geometryB, they overlap
  if (te::gm::IsPointOnPolygon(coordA, geometryInfoB->getGeometry()) == true)
  {
    //overlap
    return te::gm::SpatialRelation::OVERLAPS;
  }

  te::gm::SpatialRelation overallRelation = te::gm::SpatialRelation::DISJOINT;

  for (std::size_t i = 0; i < vecDataA.size(); ++i)
  {
    const te::vp::SegmentInfo& segmentInfoA = vecDataA.at(i);
    const te::gm::Envelope& envelopeA = segmentInfoA.getEnvelope();

    //for each envelope, we search for envelopes in geometryInfoB
    std::vector<te::vp::SegmentInfo> vecDataB;
    geometryInfoB->search(envelopeA, vecDataB);

    te::gm::SpatialRelation currentSpatialRelation = te::gm::SpatialRelation::UNKNOWN_SPATIAL_RELATION;
    if (vecDataB.empty() == true)
    {
      continue;
    }

    std::size_t touchesCount = 0;
    for (std::size_t j = 0; j < vecDataB.size(); ++j)
    {
      const te::vp::SegmentInfo& segmentInfoB = vecDataB.at(j);

      std::size_t equalityCount = 0;
      //start of A, start of B
      if (segmentInfoA.getCoord1() == segmentInfoB.getCoord1())
      {
        ++equalityCount;
      }
      //end of A, end of B
      if (segmentInfoA.getCoord2() == segmentInfoB.getCoord2())
      {
        ++equalityCount;
      }
      //start of A, end of B
      if (segmentInfoA.getCoord1() == segmentInfoB.getCoord2())
      {
        ++equalityCount;
      }
      //end of A, start of B
      if (segmentInfoA.getCoord2() == segmentInfoB.getCoord1())
      {
        ++equalityCount;
      }

      if (equalityCount == 0 && mbrA->touches(*mbrB) == false)
      {
        //if none of the coordinates are equal, we cannot determinate if the segments are disjoint or crossing without calculating the intersections
        return te::gm::SpatialRelation::OVERLAPS;
      }
      else if (equalityCount == 1)
      {
        ++touchesCount;
        if (currentSpatialRelation == te::gm::SpatialRelation::UNKNOWN_SPATIAL_RELATION)
        {
          currentSpatialRelation = te::gm::SpatialRelation::TOUCHES;
        }
        continue;
      }
      else if (equalityCount == 2)
      {
        currentSpatialRelation = te::gm::SpatialRelation::EQUALS;
        break;
      }

      return te::gm::SpatialRelation::OVERLAPS;
    }

    if (currentSpatialRelation == te::gm::SpatialRelation::EQUALS)
    {
      continue;
    }
    else if (currentSpatialRelation == te::gm::SpatialRelation::DISJOINT)
    {
    }
    else if (currentSpatialRelation == te::gm::SpatialRelation::TOUCHES)
    {
      if (overallRelation == te::gm::SpatialRelation::DISJOINT)
      {
        overallRelation = te::gm::SpatialRelation::TOUCHES;
      }

      //when one segment from A touches more than one segment from B, we expect the relation to be EQUALS. 
      //If it is not, we must check if the centroid of the current segment from A is within the polygon B
      if (touchesCount > 1)
      {
        //if geometryA have one point inside geometryB, they overlap
        if (te::gm::IsPointOnPolygon(envelopeA.getCenter(), geometryInfoB->getGeometry()) == true)
        {
          return te::gm::SpatialRelation::OVERLAPS;
        }
      }
    }
  }

  //if we got here, the geometries relation can either be disjoint, touches, but they do not overlap
  return overallRelation;
}

void te::vp::DiscardFeaturesWithNullAndEmptyGeometries(te::vp::FeatureSet& featureSet)
{
  te::vp::FeatureSet::iterator it = featureSet.begin();
  while (it != featureSet.end())
  {
    if ((*it)->getFirstGeometryPtr() == nullptr)
    {
      it = featureSet.erase(it);
    }
    else if ((*it)->getFirstGeometryPtr()->isEmpty() == true)
    {
      te::vp::Feature* feature = (*it);
      it = featureSet.erase(it);

      delete feature;
    }
    else
    {
      ++it;
    }
  }
}

void te::vp::Multi2Single(const te::vp::Feature* feature, te::vp::FeatureSet& singleFeatureSet)
{
  if (feature == nullptr)
  {
    return;
  }

  const te::gm::Geometry* geometry = feature->getFirstGeometryPtr();
  if (te::gm::IsMultiType(geometry->get2DGeomTypeId()) == false)
  {
    singleFeatureSet.push_back(static_cast<te::vp::Feature*>(feature->clone()));
    return;
  }

  te::gm::GeometryVector vecSingleGeometries;
  te::gm::Multi2Single(geometry, vecSingleGeometries);

  for (std::size_t i = 0; i < vecSingleGeometries.size(); ++i)
  {
    std::unique_ptr<te::vp::Feature> singleFeature(static_cast<te::vp::Feature*>(feature->clone()));
    singleFeature->setFirstGeometry(vecSingleGeometries.at(i));

    singleFeatureSet.push_back(singleFeature.release());
  }
}

void te::vp::Multi2Single(const te::vp::FeatureSet& featureSet, te::vp::FeatureSet& singleFeatureSet)
{
  for (std::size_t i = 0; i < featureSet.size(); ++i)
  {
    const te::vp::Feature* feature = featureSet.at(i);
    te::vp::Multi2Single(feature, singleFeatureSet);
  }
}

te::gm::GeometryVector te::vp::AdjustToGeomType(const te::gm::Geometry* geometry, te::gm::GeomType geomType)
{
  te::gm::GeometryVector vecAdjusted;
  if (te::gm::IsNullOrEmpty(geometry) == true)
  {
    return vecAdjusted;
  }

  te::gm::GeomType outputSingleGeomType = geomType;

  bool expectedIsMulti = te::gm::IsMultiType(geomType);
  if (expectedIsMulti == true)
  {
    outputSingleGeomType = te::gm::GetSimpleType(geomType);
  }

  std::unique_ptr<te::gm::Geometry> geometryOutput(te::gm::NormalizeGeometryByType(geometry, outputSingleGeomType));
  if (geometryOutput == nullptr)
  {
    return vecAdjusted;
  }

  if (geometryOutput->getGeomTypeId() == geomType)
  {
    vecAdjusted.push_back(geometryOutput.release());
  }
  else
  {
    bool currentIsMulti = te::gm::IsMultiType(geometryOutput->getGeomTypeId());
    if (currentIsMulti == true && expectedIsMulti == false)
    {
      te::gm::Multi2Single(geometryOutput.get(), vecAdjusted);
    }
    else if (currentIsMulti == false && expectedIsMulti == true)
    {
      if (geomType == te::gm::GeometryCollectionType)
      {
        std::unique_ptr<te::gm::GeometryCollection> newCollection(static_cast<te::gm::GeometryCollection*>(te::gm::GeometryFactory::make(te::gm::GeometryCollectionType, geometry->getSRID())));
        newCollection->add(geometryOutput.release());
        vecAdjusted.push_back(newCollection.release());
      }
      else
      {
        vecAdjusted.push_back(te::vp::SetGeomAsMulti(*geometryOutput.get()));
      }
    }
  }

  return vecAdjusted;
}

te::gm::GeometryVector te::vp::AdjustToGeomType(const te::gm::GeometryVector& vecGeometries, te::gm::GeomType geomType)
{
  te::gm::GeometryVector result;
  for (std::size_t i = 0; i < vecGeometries.size(); ++i)
  {
    te::gm::GeometryVector currentResult = te::vp::AdjustToGeomType(vecGeometries.at(i), geomType);
    result.insert(result.end(), currentResult.begin(), currentResult.end());
  }

  return result;
}
