/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralib/vp/CommonDataStructures.h

   \brief Utility classes, structures and definitions for Vector Processing.
 */

#ifndef __TERRALIB_VP_INTERNAL_COMMONFUNCTIONS_H
#define __TERRALIB_VP_INTERNAL_COMMONFUNCTIONS_H

//Terralib include files
#include "Config.h"

#include "../geometry/CommonDataStructures.h"
#include "../geometry/Enums.h"


//STL include files
#include <vector>

namespace te
{
  namespace gm
  {
    struct Coord2D;
    class Envelope;
    class Geometry;
  }

  namespace vp
  {
    class Feature;
    class FeatureSet;
    class GeometryInfo;
    class SegmentInfo;
    
    

    /*!
      * \brief Gets the segment information related to all segments from the given geometry. An Optional filter can be used to filter the segments that will be returned
      *
      * \param geometry The geometry to be analysed
      * \param filter The optional filter
      * \param vecSegmentInfo [Return] The segments from the given geometry. The return is a parameter to make easier to create a list containing segments from several geometries
      */
    TEVPEXPORT void GetAllSegments(const te::gm::Geometry* geometry, const te::gm::Envelope& filter, std::vector<SegmentInfo>& vecSegmentInfo);

    /*!
      \brief Algorithm to fast detection of spatial relations between geometries. Very useful to fase dectect if two geomtries dont intercept themselves
              Sometimes can generate false positives for overlaps, but never generates false positives for disjoint and touches

      \param geometryA The input geometry A
      \param geometryB The input geometry B

      \return The detected spatial relation
    */
    TEVPEXPORT te::gm::SpatialRelation FastRelationCalculator(const te::vp::GeometryInfo * geometryInfoA, const te::vp::GeometryInfo* geometryInfoB);

    /*!
      \brief Algorithm to fast detection of spatial relations between geometries. Very useful to fase dectect if two geomtries dont intercept themselves
              Sometimes can generate false positives for overlaps, but never generates false positives for disjoint and touches

      \param geometryA The input geometry A
      \param geometryB The input geometry B

      \return The detected spatial relation
    */
    TEVPEXPORT te::gm::SpatialRelation FastRelationCalculator(const te::gm::Geometry* geometryA, const te::gm::Geometry* geometryB);

    //!< Discards all the features from the set that have null or empty geometries
    TEVPEXPORT void DiscardFeaturesWithNullAndEmptyGeometries(te::vp::FeatureSet& featureSet);

    //!< Splits all feature that have geometry collections in to features that have single type geometries. All the attributes are copied
    TEVPEXPORT void Multi2Single(const te::vp::Feature* feature, te::vp::FeatureSet& singleFeatureSet);

    //!< Splits all the features that have geometry collections in to features that have single type geometries. All the attributes are copied
    TEVPEXPORT void Multi2Single(const te::vp::FeatureSet& featureSet, te::vp::FeatureSet& singleFeatureSet);

    //!< Adjust the given geometry to the given geometry type. If the conversion is not possible, the geometry is discarded
    TEVPEXPORT te::gm::GeometryVector AdjustToGeomType(const te::gm::Geometry* geometry, te::gm::GeomType geomType);

    //!< Adjust the given geometry to the given geometry type. If the conversion is not possible, the geometry is discarded
    TEVPEXPORT te::gm::GeometryVector AdjustToGeomType(const te::gm::GeometryVector& vecGeometries, te::gm::GeomType geomType);
  }
}



#endif //__TERRALIB_VP_INTERNAL_COMMONFUNCTIONS_H
