/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralin/vp/DifferenceOperation.cpp

   \brief Represents a difference operation between two sets of features
 */
#include "DifferenceOperation.h"

#include "../common/STLUtils.h"
#include "../geometry/algorithms/SafeOverlay.h"
#include "../geometry/CommonDataStructures.h"
#include "../geometry/Geometry.h"
#include "../geometry/Utils.h"
#include "../geometry/Validation.h"
#include "../memory/PreparedItemCopy.h"
#include "../sam/rtree/Index.h"
#include "CommonFunctions.h"
#include "Feature.h"
#include "GeometrySnapper.h"
#include "OperationService.h"

namespace te
{
  namespace vp
  {
    namespace difference
    {
      const std::string OPERATION_NAME = "Difference";
    }
  }
}

namespace
{
  static te::vp::DifferenceOperationFactory
    DifferenceOperationFactoryInstance;
}

te::vp::DifferenceOperation::DifferenceOperation()
    : AbstractOperation(te::vp::difference::OPERATION_NAME)
  {

  }

std::vector<te::vp::FeatureSet> te::vp::DifferenceOperation::executeImpl(const std::vector<te::vp::FeatureSet>& vecInput)
{
  std::vector<te::vp::FeatureSet> vecOutput;
  if (vecInput.size() != 2)
  {
    vecOutput.resize(1);
    return vecOutput;
  }

  te::gm::GeomType outputSingleGeomType = getOutputGeometryType();
  if (te::gm::IsMultiType(outputSingleGeomType) == true)
  {
    outputSingleGeomType = te::gm::GetSimpleType(outputSingleGeomType);
  }

  //snaps the geometries
  te::vp::FeatureSet fsRed = vecInput.at(0);
  te::vp::FeatureSet fsBlue = vecInput.at(1);

  if (this->getParameters()->getSwapInput() == true)
  {
    fsRed = vecInput.at(1);
    fsBlue = vecInput.at(0);
  }

  if (fsRed.empty() == true)
  {
    vecOutput.resize(1);
    return vecOutput;
  }

  te::mem::PreparedItemCopy copy;

  if (fsBlue.empty() == true)
  {
    te::vp::FeatureSet outputFeatureSet;

    for (std::size_t i = 0; i < fsRed.size(); ++i)
    {
      const te::vp::Feature* featureRed = fsRed.at(i);

      const te::gm::Geometry* geometryRed = fsRed.at(i)->getFirstGeometryPtr();
      if (te::gm::IsNullOrEmpty(geometryRed) == true)
      {
        continue;
      }

      te::gm::GeometryVector vecAdjustedOutput = te::vp::AdjustToGeomType(geometryRed, getOutputGeometryType());
      for (std::size_t k = 0; k < vecAdjustedOutput.size(); ++k)
      {
        te::vp::Feature* featureOutput = createOutputFeature(0);

        copy.copyValues(featureRed, featureOutput, false);
        featureOutput->setFirstGeometry(vecAdjustedOutput.at(k));

        outputFeatureSet.push_back(featureOutput);
      }
    }

    vecOutput.push_back(outputFeatureSet);
    return vecOutput;
  }

  //then we index the geometries from the blue featureSet
  std::unique_ptr< te::sam::rtree::Index<std::size_t> > indexBlue = te::vp::OperationService::CreateIndex(fsBlue);
  
  //finally, for each feature in the red list, we make the intersection to all candidades in the blue list
  te::vp::FeatureSet outputFeatureSet;
  for (std::size_t i = 0; i < fsRed.size(); ++i)
  {
    //operation was aborted
    if (isAborted() == true)
    {
      return vecOutput;
    }

    const te::vp::Feature* featureRed = fsRed.at(i);
    const te::gm::Geometry* geometryRed = featureRed->getFirstGeometryPtr();
    if (te::gm::IsNullOrEmpty(geometryRed) == true)
    {
      continue;
    }

    //we search for candidadtes
    std::vector<std::size_t> vecCandidatesBlue;
    indexBlue->search(*geometryRed->getMBR(), vecCandidatesBlue);

    te::gm::GeometryVector vecForDifference;
    for (std::size_t j = 0; j < vecCandidatesBlue.size(); ++j)
    {
      std::size_t candidateIndexBlue = vecCandidatesBlue.at(j);
      const te::vp::Feature* featureBlue = fsBlue.at(candidateIndexBlue);
      const te::gm::Geometry* geometryBlue = featureBlue->getFirstGeometryPtr();

      //we can discard the blue geometries which envelopes that just touches the envelope of the red geometry
      if (geometryRed->getMBR()->touches(*geometryBlue->getMBR()) == true)
      {
        continue;
      }

      vecForDifference.push_back(static_cast<te::gm::Geometry*>(geometryBlue->clone()));
    }

    te::gm::GeometryPtr geometryOutput;
    //if we dont have any candidates, the clone the input geometry
    if (vecForDifference.empty() == true)
    {
      geometryOutput.reset(static_cast<te::gm::Geometry*>(geometryRed->clone()));
    }
    else
    {
      //if any candidate, we must ensure that that are dissolved before making the difference
      //this is done to avoid slivers and unexpected output geometries located in the borders of the input geometries
      te::gm::GeometryPtr candidatesUnited;
      if (vecForDifference.size() == 1)
      {
        candidatesUnited.reset(vecForDifference.at(0));
      }
      else if (vecForDifference.size() > 1)
      {
        //then we first make a union of all actual candidates
        candidatesUnited.reset(te::gm::CascadedUnion(vecForDifference));

        te::common::FreeContents(vecForDifference);

        candidatesUnited.reset(te::gm::NormalizeGeometryByType(candidatesUnited.get(), outputSingleGeomType));
      }

      if (te::gm::IsNullOrEmpty(candidatesUnited) == false)
      {
        //then we calculates the difference
        geometryOutput.reset(te::gm::SafeOverlay::Difference(geometryRed, candidatesUnited.get()).release());
      }
      else
      {
        geometryOutput.reset(static_cast<te::gm::Geometry*>(geometryRed->clone()));
      }
    }

    if (te::gm::IsNullOrEmpty(geometryOutput) == true)
    {
      continue;
    }

    //we filter the results, discarding anything that is not a polygon or multi polygon
    geometryOutput.reset(te::gm::NormalizeGeometryByType(geometryOutput.get(), outputSingleGeomType));
    if (te::gm::IsNullOrEmpty(geometryOutput) == true)
    {
      continue;
    }

    std::vector<te::gm::Geometry*> vecSingleGeometries;
    te::gm::Multi2Single(geometryOutput.get(), vecSingleGeometries);

    for (std::size_t j = 0; j < vecSingleGeometries.size(); ++j)
    {
      //create the output feature based on the resulting geometry and the input attributes
      te::gm::GeometryVector vecAdjustedOutput = te::vp::AdjustToGeomType(vecSingleGeometries.at(j), getOutputGeometryType());
      for (std::size_t k = 0; k < vecAdjustedOutput.size(); ++k)
      {
        te::vp::Feature* featureOutput = createOutputFeature(0);

        copy.copyValues(featureRed, featureOutput, false);
        featureOutput->setFirstGeometry(vecAdjustedOutput.at(k));

        //and add it to the output list
        outputFeatureSet.push_back(featureOutput);
      }
    }

    te::common::FreeContents(vecSingleGeometries);
  }
  vecOutput.push_back(outputFeatureSet);
  return vecOutput;
}

te::vp::DifferenceOperationFactory::DifferenceOperationFactory()
  : OperationFactory(te::vp::difference::OPERATION_NAME)
{
}

te::vp::DifferenceOperationFactory::~DifferenceOperationFactory() = default;

te::vp::AbstractOperation* te::vp::DifferenceOperationFactory::build()
{
  return new DifferenceOperation();
}
