/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

#include "DissolveCellSubdivision.h"

#include "../common/STLUtils.h"
#include "../common/ThreadGroup.h"
#include "../geometry/CommonDataStructures.h"
#include "../geometry/Geometry.h"
#include "../geometry/GeometryCollection.h"
#include "../geometry/Envelope.h"
#include "../geometry/LinearRing.h"
#include "../geometry/Polygon.h"
#include "../geometry/Utils.h"
#include "../memory/PreparedItemCopy.h"
#include "../sam/rtree/Index.h"
#include "CommonFunctions.h"
#include "Feature.h"
#include "Utils.h"

//Boost include files
#include <boost/numeric/ublas/matrix.hpp>
#include <boost/thread.hpp>
#include <boost/asio/io_service.hpp>

//STL include files
#include <map>
#include <memory>


struct CellReference
{
  CellReference()
    : m_column(0)
    , m_row(0)
  {

  }

  CellReference(std::size_t column, std::size_t row, const te::vp::FeatureSet& featureSet)
    : m_column(column)
    , m_row(row)
    , m_featureSet(featureSet)
  {

  }

  CellReference(const CellReference& rhs)
  {
    *this = rhs;
  }

  ~CellReference()
  {
    te::common::FreeContents(m_featureSet);
    m_featureSet.clear();
  }

  CellReference& operator =(const CellReference& rhs)
  {
    this->m_column = rhs.m_column;
    this->m_row = rhs.m_row;

    te::common::FreeContents(this->m_featureSet);
    this->m_featureSet.clear();

    te::common::Clone(rhs.m_featureSet, m_featureSet);

    return *this;
  }

  std::size_t m_column;
  std::size_t m_row;
  te::vp::FeatureSet m_featureSet;
};

te::gm::Envelope GetMBR(const te::gm::GeometryVector& vecGeometries)
{
  te::gm::Envelope mbr;
  for (std::size_t i = 0; i < vecGeometries.size(); ++i)
  {
    mbr.Union(*vecGeometries.at(i)->getMBR());
  }

  return mbr;
}

te::gm::Envelope GetMBR(const te::vp::FeatureSet& fs)
{
  te::gm::Envelope mbr;
  for (std::size_t i = 0; i < fs.size(); ++i)
  {
    mbr.Union(*fs.at(i)->getFirstGeometryPtr()->getMBR());
  }

  return mbr;
}

void ExtractHoles(te::gm::Geometry* geometry, te::gm::GeometryVector& extractedHoles)
{
  if (geometry->isCollection() == true)
  {
    te::gm::GeometryCollection* collection = static_cast<te::gm::GeometryCollection*>(geometry);
    for (std::size_t i = 0; i < collection->getNumGeometries(); ++i)
    {
      ExtractHoles(collection->getGeometryN(i), extractedHoles);
    }
  }
  else if (geometry->get2DGeomTypeId() == te::gm::PolygonType)
  {
    te::gm::Polygon* polygon = static_cast<te::gm::Polygon*>(geometry);
    while (polygon->getNumRings() > 1)
    {
      extractedHoles.push_back(static_cast<te::gm::Geometry*>(polygon->getRingN(1)->clone()));
      polygon->removeRingN(1);
    }
  }
}

void ExtractHoles(te::gm::GeometryVector& vecGeometries, std::vector<te::gm::GeometryVector>& extractedHoles)
{
  for (std::size_t i = 0; i < vecGeometries.size(); ++i)
  {
    te::gm::GeometryVector vecHoles;
    ExtractHoles(vecGeometries.at(i), vecHoles);

    extractedHoles.push_back(vecHoles);
  }
}

void AddHolesBack(te::gm::GeometryVector& vecPolygons, std::vector<te::gm::GeometryVector>& vecAllHoles)
{
  //we first create an index for tghe polygons
  te::sam::rtree::Index<std::size_t> index;
  for (std::size_t i = 0; i < vecPolygons.size(); ++i)
  {
    te::gm::Geometry* geometry = vecPolygons.at(i);
    index.insert(*geometry->getMBR(), i);
  }

  //and then we distribute the holes
  for (std::size_t i = 0; i < vecAllHoles.size(); ++i)
  {
    te::gm::GeometryVector& vecHoles = vecAllHoles.at(i);

    //for each grup of holes, we must find their parent
    std::size_t parentIndex = std::string::npos;
    for (std::size_t j = 0; j < vecHoles.size(); ++j)
    {
      te::gm::LinearRing* ring = static_cast<te::gm::LinearRing*>(vecHoles.at(j));

      std::vector<std::size_t> vecCandidates;
      index.search(*ring->getMBR(), vecCandidates);

      if (vecCandidates.size() == 1)
      {
        parentIndex = vecCandidates.at(0);
      }
      else
      {
        te::gm::Coord2D coord = ring->getCentroid();
        for (std::size_t k = 0; k < vecCandidates.size(); ++k)
        {
          std::size_t candidateIndex = vecCandidates.at(k);
          if (te::gm::IsPointOnPolygon(coord, vecPolygons.at(candidateIndex)) == true)
          {
            parentIndex = candidateIndex;
            break;
          }
        }
      }

      if (parentIndex != std::string::npos)
      {
        break;
      }
    }

    for (std::size_t j = 0; j < vecHoles.size(); ++j)
    {
      te::gm::LinearRing* ring = static_cast<te::gm::LinearRing*>(vecHoles.at(j));

      te::gm::Polygon* polygon = static_cast<te::gm::Polygon*>(vecPolygons.at(parentIndex));
      polygon->add(ring);
    }
  }
}

te::vp::FeatureSet DissolveCellBorder(const te::vp::FeatureSet& fs1, const te::vp::FeatureSet& fs2)
{
  te::vp::FeatureSet featureSetResult;
  if (fs1.empty() == true && fs2.empty() == true)
  {
    return featureSetResult;
  }
  else if (fs1.empty() == false && fs2.empty() == true)
  {
    te::common::Clone(fs1, featureSetResult);
    return featureSetResult;
  }
  else if (fs1.empty() == true && fs2.empty() == false)
  {
    te::common::Clone(fs2, featureSetResult);
    return featureSetResult;
  }

  te::gm::Envelope mbr1 = GetMBR(fs1);
  te::gm::Envelope mbr2 = GetMBR(fs2);

  if (mbr1.intersects(mbr2) == false)
  {
    te::common::Clone(fs1, featureSetResult);
    te::common::Clone(fs2, featureSetResult);

    return featureSetResult;
  }

  te::gm::GeometryVectorConst vecGeometries1 = te::vp::GetGeometriesFromFeatureSet(fs1);
  te::gm::GeometryVectorConst vecGeometries2 = te::vp::GetGeometriesFromFeatureSet(fs2);

  te::gm::GeometryVectorConst vecAllInput;
  vecAllInput.insert(vecAllInput.end(), vecGeometries1.begin(), vecGeometries1.end());
  vecAllInput.insert(vecAllInput.end(), vecGeometries2.begin(), vecGeometries2.end());

  te::gm::Envelope interEnvelope = mbr1.intersection(mbr2);

  te::gm::GeometryVector vecResult;
  te::gm::GeometryVectorConst vecToDissolve;
  for (std::size_t i = 0; i < vecAllInput.size(); ++i)
  {
    const te::gm::Geometry* currentGeometry = vecAllInput.at(i);
    if (currentGeometry->getMBR()->intersects(interEnvelope) == true)
    {
      vecToDissolve.push_back(currentGeometry);
    }
    else
    {
      vecResult.push_back(static_cast<te::gm::Geometry*>(currentGeometry->clone()));
    }
  }

  if (vecToDissolve.empty() == false)
  {
    te::gm::GeometryVector vecDissolved = te::vp::DissolveCellSubdivision::GetGeometryUnionNoOverlaps(vecToDissolve);
    vecResult.insert(vecResult.end(), vecDissolved.begin(), vecDissolved.end());
  }

  te::vp::Feature* seed = fs1.at(0);

  for (std::size_t i = 0; i < vecResult.size(); ++i)
  {
    te::vp::Feature* featureResult = static_cast<te::vp::Feature*>(seed->clone());
    featureResult->setFirstGeometry(vecResult.at(i));

    featureSetResult.push_back(featureResult);
  }

  return featureSetResult;
}

void DissolveQuadrant(std::size_t r, std::size_t c, const boost::numeric::ublas::matrix<CellReference>* matrix, te::vp::FeatureSet* result)
{
  te::common::FreeContents(*result);
  result->clear();

  te::vp::FeatureSet r1c1;
  te::vp::FeatureSet r1c2;
  te::vp::FeatureSet r2c1;
  te::vp::FeatureSet r2c2;

  //we first get the featureSets from all 4 cells
  r1c1 = matrix->operator()(r, c).m_featureSet;

  if ((c + 1) < matrix->size2())
  {
    r1c2 = matrix->operator()(r, c + 1).m_featureSet;
  }

  if ((r + 1) < matrix->size1())
  {
    r2c1 = matrix->operator()(r + 1, c).m_featureSet;
  }

  if ((r + 1) < matrix->size1() && (c + 1) < matrix->size2())
  {
    r2c2 = matrix->operator()(r + 1, c + 1).m_featureSet;
  }

  //dissolves the first line
  te::vp::FeatureSet row1 = DissolveCellBorder(r1c1, r1c2);

  //dissolves the second line
  te::vp::FeatureSet row2 = DissolveCellBorder(r2c1, r2c2);

  //dissolves both lines
  te::vp::FeatureSet fsResult = DissolveCellBorder(row1, row2);
  result->insert(result->end(), fsResult.begin(), fsResult.end());

  te::common::FreeContents(row1);
  te::common::FreeContents(row2);
}

void DissolveCell(te::vp::FeatureSet* featureSet)
{
  if (featureSet->size() <= 1)
  {
    return;
  }

  te::gm::GeometryVectorConst vecGeometriesConst = te::vp::GetGeometriesFromFeatureSet(*featureSet);
  std::unique_ptr<te::gm::Geometry> unionResult = te::gm::GetGeometryUnion(vecGeometriesConst);

  te::gm::GeometryVector vecDissolved;
  te::gm::Multi2Single(unionResult.get(), vecDissolved);

  const te::vp::Feature* seed = featureSet->at(0);

  te::vp::FeatureSet resultFeatureSet;
  resultFeatureSet.reserve(vecDissolved.size());
  for (std::size_t i = 0; i < vecDissolved.size(); ++i)
  {
    te::vp::Feature* featureResult = static_cast<te::vp::Feature*>(seed->clone());
    featureResult->setFirstGeometry(vecDissolved.at(i));

    resultFeatureSet.push_back(featureResult);
  }

  te::common::FreeContents(*featureSet);
  featureSet->clear();

  featureSet->insert(featureSet->end(), resultFeatureSet.begin(), resultFeatureSet.end());
}

te::gm::GeometryVector te::vp::DissolveCellSubdivision::GetGeometryUnionNoOverlaps(const te::gm::GeometryVectorConst& vecGeometries)
{
  te::gm::GeometryVector vecToDissolve;
  if (vecGeometries.empty() == true)
  {
    return vecToDissolve;
  }

  te::common::Clone(vecGeometries, vecToDissolve);

  if (vecGeometries.size() == 1)
  {
    return vecToDissolve;
  }

  std::vector<te::gm::GeometryVector> vecAllHoles;
  ExtractHoles(vecToDissolve, vecAllHoles);

  std::unique_ptr<te::gm::Geometry> unionResult = te::gm::GetGeometryUnion(te::gm::ToConstVector(vecToDissolve));

  te::gm::GeometryVector vecDissolved;
  te::gm::Multi2Single(unionResult.get(), vecDissolved);

  AddHolesBack(vecDissolved, vecAllHoles);

  te::common::FreeContents(vecToDissolve);

  return vecDissolved;
}

te::vp::FeatureSet te::vp::DissolveCellSubdivision::GetGeometryUnionNoOverlaps(const te::vp::FeatureSet& fs)
{
  te::gm::GeometryVectorConst vecGeometriesConst = te::vp::GetGeometriesFromFeatureSet(fs);
  te::gm::GeometryVector vecDissolved = te::vp::DissolveCellSubdivision::GetGeometryUnionNoOverlaps(vecGeometriesConst);

  const te::vp::Feature* seed = fs.at(0);
  
  te::vp::FeatureSet fsResult;
  fsResult.reserve(vecDissolved.size());
  for (std::size_t i = 0; i < vecDissolved.size(); ++i)
  {
    te::vp::Feature* featureResult = static_cast<te::vp::Feature*>(seed->clone());
    featureResult->setFirstGeometry(vecDissolved.at(i));

    fsResult.push_back(featureResult);
  }

  return fsResult;
}

te::vp::FeatureSet te::vp::DissolveCellSubdivision::Dissolve(const te::vp::FeatureSet& fs, std::size_t rowIndex, std::size_t columnIndex, bool dissolveCoordinateSubdivision)
{
  te::vp::FeatureSet outputFeatureSet;
  if (fs.empty() == true)
  {
    return outputFeatureSet;
  }

  if (fs.size() == 1)
  {
    te::common::Clone(fs, outputFeatureSet);
    return outputFeatureSet;
  }

  std::set<std::size_t> setColumn;
  std::set<std::size_t> setRow;
  for (std::size_t i = 0; i < fs.size(); ++i)
  {
    const te::vp::Feature* feature = fs.at(i);

    std::size_t row = static_cast<std::size_t>(feature->getInt32(rowIndex));
    std::size_t column = static_cast<std::size_t>(feature->getInt32(columnIndex));

    setColumn.insert(column);
    setRow.insert(row);
  }

  std::size_t columnStart = *setColumn.begin();
  std::size_t columnEnd = *setColumn.rbegin();

  std::size_t rowStart = *setRow.begin();
  std::size_t rowEnd = *setRow.rbegin();

  std::size_t numColumns = columnEnd - columnStart + 1;
  std::size_t numRows = rowEnd - rowStart + 1;

  if (numRows == 1 && numColumns == 1 && dissolveCoordinateSubdivision == false)
  {
    te::common::Clone(fs, outputFeatureSet);
    return outputFeatureSet;
  }

  //we create a matrix and add the features to each cell
  boost::numeric::ublas::matrix<CellReference> matrix(numRows, numColumns);

  for (std::size_t i = 0; i < fs.size(); ++i)
  {
    te::vp::Feature* clonedFeature = static_cast<te::vp::Feature*>(fs.at(i)->clone());

    std::size_t row = static_cast<std::size_t>(clonedFeature->getInt32(rowIndex));
    std::size_t column = static_cast<std::size_t>(clonedFeature->getInt32(columnIndex));
    
    std::size_t normalizedRow = row - rowStart;
    std::size_t normalizedColumn = column - columnStart;

    CellReference& cellReference = matrix(normalizedRow, normalizedColumn);
    cellReference.m_row = normalizedRow;
    cellReference.m_column = normalizedColumn;
    cellReference.m_featureSet.push_back(clonedFeature);
  }

  if (dissolveCoordinateSubdivision == true)
  {
    te::common::ThreadGroup threadStep1;

    //Step 1 - dissolves the features within each cell
    for (std::size_t r = 0; r < matrix.size1(); ++r)
    {
      for (std::size_t c = 0; c < matrix.size2(); ++c)
      {
        CellReference& cellReference = matrix(r, c);

        if (cellReference.m_featureSet.size() <= 1)
        {
          continue;
        }

        //adds the jobs to the manager
        threadStep1.addJob(std::bind(DissolveCell, &cellReference.m_featureSet));

        //no thread. This code is used to run without using multithread. Useful for debug
        //DissolveCell(&cellReference.m_featureSet);
      }
    }

    //wait for the jobs to be finished
    threadStep1.waitToFinish();
  }

  //if (numRows > 1 || numColumns > 1)
  {
    //Step 2 - dissolves cells of the matrix
    while (true)
    {
      te::common::ThreadGroup threadStep2;

      std::size_t newRowCount = static_cast<std::size_t>(std::ceil(matrix.size1() / 2.));
      std::size_t newColumnCount = static_cast<std::size_t>(std::ceil(matrix.size2() / 2.));

      boost::numeric::ublas::matrix<CellReference> subMatrix(newRowCount, newColumnCount);
      for (std::size_t r = 0; r < subMatrix.size1(); ++r)
      {
        for (std::size_t c = 0; c < subMatrix.size2(); ++c)
        {
          CellReference& cellReference = subMatrix(r, c);
          cellReference.m_column = c;
          cellReference.m_row = r;

          //adds the jobs to the manager
          threadStep2.addJob(std::bind(DissolveQuadrant, r * 2, c * 2, &matrix, &cellReference.m_featureSet));

          //no thread. This code is used to run without using multithread. Useful for debug
          //DissolveQuadrant(r * 2, c * 2, &matrix, &cellReference.m_featureSet);
        }
      }

      //wait for the jobs to be finished
      threadStep2.waitToFinish();

      matrix = subMatrix;

      if (newRowCount == 1 && newColumnCount == 1)
      {
        break;
      }
    }
  }

  const te::vp::FeatureSet& dissolvedFS = matrix(0, 0).m_featureSet;

  te::vp::FeatureSet resultFS;

  te::common::Clone(dissolvedFS, resultFS);
  
  return resultFS;
}
