/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralin/vp/DissolveOperation.cpp

   \brief Represents a dissolve operation
 */
#include "DissolveOperation.h"

#include "../common/STLUtils.h"
#include "../dataaccess/dataset/DataSetType.h"
#include "../dataaccess/utils/DataAccess.h"
#include "../geometry/CommonDataStructures.h"
#include "../geometry/Geometry.h"
#include "../geometry/Utils.h"
#include "../memory/PreparedItemCopy.h"
#include "CommonFunctions.h"
#include "Feature.h"
#include "Utils.h"

namespace te
{
  namespace vp
  {
    namespace dissolve
    {
      const std::string OPERATION_NAME = "Dissolve";
    }
  }
}

namespace
{
  static te::vp::DissolveOperationFactory
    DissolveOperationFactoryInstance;
}

te::vp::DissolveOperationParameters::DissolveOperationParameters()
  : te::vp::AbstractParameters()
{

}
te::vp::DissolveOperationParameters::~DissolveOperationParameters()
{

}

bool te::vp::DissolveOperationParameters::isValid(std::string& errorMessage)
{
  for (std::size_t i = 0; i < this->m_inputDataVector.size(); ++i)
  {
    te::da::DataAccess* dataAccess = this->m_inputDataVector.at(i);
    const te::da::DataSetType* dataSetType = dataAccess->getDataSetType();
    for (std::size_t j = 0; j < dataSetType->size(); ++j)
    {
      te::dt::Property* property = dataSetType->getProperty(j);
      if (property->getType() == te::dt::FLOAT_TYPE || property->getType() == te::dt::DOUBLE_TYPE)
      {
        errorMessage = TE_TR("This operation cannot handle floating point properties yet. Remove them and try again!");
        return false;
      }
    }
  }

  return te::vp::AbstractParameters::isValid(errorMessage);
}

te::vp::DissolveOperationCapabilities::DissolveOperationCapabilities(const std::string& operationName, std::size_t numberOfInputs, std::size_t numberOfOutputs, const std::string& dataHandlerName, bool snapGeometries, bool allowCoordinateSubdivision)
  : te::vp::AbstractOperationCapabilities(operationName, numberOfInputs, numberOfOutputs, dataHandlerName, snapGeometries, allowCoordinateSubdivision)
{

}

std::unique_ptr<te::vp::AbstractParameters> te::vp::DissolveOperationCapabilities::createParameters() const
{
  std::unique_ptr<te::vp::AbstractParameters> params(new te::vp::DissolveOperationParameters());
  return params;
}

te::vp::DissolveOperation::DissolveOperation()
    : AbstractOperation(te::vp::dissolve::OPERATION_NAME)
 {

 }

std::unique_ptr<te::vp::AbstractOperationCapabilities> te::vp::DissolveOperation::getCapabilities() const
{
  std::unique_ptr<te::vp::AbstractOperationCapabilities> capabilities(new te::vp::DissolveOperationCapabilities(getName(), 1, 1, "DissolveSubdivideDataHandler", false, false));
  return capabilities;
}

std::vector<te::vp::FeatureSet> te::vp::DissolveOperation::executeImpl(const std::vector<te::vp::FeatureSet>& vecInput)
{
  std::vector<te::vp::FeatureSet> vecOutput;

  if (vecInput.size() != 1)
  {
    vecOutput.resize(1);
    return vecOutput;
  }

  if (vecInput.at(0).empty() == true)
  {
    vecOutput.resize(1);
    return vecOutput;
  }

  te::vp::FeatureSet outputFeatureSet;

  const te::vp::FeatureSet& fsRed = vecInput.at(0);
  if (fsRed.empty() == true)
  {
    vecOutput.resize(1);
    return vecOutput;
  }

  const te::vp::Feature* seedFeature = fsRed.at(0);

  te::gm::GeometryVectorConst vecGeometries = te::vp::GetGeometriesFromFeatureSet(fsRed);
  std::unique_ptr<te::gm::Geometry> result = te::gm::GetGeometryUnion(vecGeometries);

  te::gm::GeometryVector vecSingleResult;
  te::gm::Multi2Single(result.get(), vecSingleResult);

  te::mem::PreparedItemCopy copy;
  for (std::size_t g = 0; g < vecSingleResult.size(); ++g)
  {
    //create the output feature based on the resulting geometry and the input attributes
    te::gm::GeometryVector vecAdjustedOutput = te::vp::AdjustToGeomType(vecSingleResult.at(g), getOutputGeometryType());
    for (std::size_t k = 0; k < vecAdjustedOutput.size(); ++k)
    {
      //create the output feature based on the resulting geometry and the input attributes
      te::vp::Feature* featureOutput = createOutputFeature(0);
      copy.copyValues(seedFeature, featureOutput, false);
      featureOutput->setFirstGeometry(vecAdjustedOutput.at(k));

      //and add it to the output list
      outputFeatureSet.push_back(featureOutput);
    }
  }
  
  te::common::FreeContents(vecSingleResult);

  vecOutput.push_back(outputFeatureSet);
  return vecOutput;
}

te::vp::DissolveOperationFactory::DissolveOperationFactory()
  : OperationFactory(te::vp::dissolve::OPERATION_NAME)
{
}

te::vp::DissolveOperationFactory::~DissolveOperationFactory() = default;

te::vp::AbstractOperation* te::vp::DissolveOperationFactory::build()
{
  return new DissolveOperation();
}
