/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralin/vp/DissolveSubdivisionOperation.cpp

   \brief Represents a dissolve operation
 */
#include "DissolveSubdivisionOperation.h"

#include "../common/STLUtils.h"
#include "../dataaccess/dataset/DataSetType.h"
#include "../dataaccess/utils/DataAccess.h"
#include "../geometry/CommonDataStructures.h"
#include "../geometry/Geometry.h"
#include "../geometry/Utils.h"
#include "../memory/PreparedItemCopy.h"
#include "CommonFunctions.h"
#include "DissolveCellSubdivision.h"
#include "Feature.h"
#include "Utils.h"

namespace te
{
  namespace vp
  {
    namespace dissolve_subdivision
    {
      const std::string OPERATION_NAME = "DissolveSubdivision";
    }
  }
}

namespace
{
  static te::vp::DissolveSubdivisionOperationFactory
    DissolveSubdivisionOperationFactoryInstance;
}

te::vp::DissolveSubdivisionOperationParameters::DissolveSubdivisionOperationParameters()
  : te::vp::AbstractParameters()
  , m_rowPropertyName("")
  , m_columnPropertyName("")
  , m_dissolveMaxCoordinates(0)
{

}

te::vp::DissolveSubdivisionOperationParameters::DissolveSubdivisionOperationParameters(const std::string& rowPropertyName, const std::string& columnPropertyName, bool dissolveMaxCoordinates)
  : te::vp::AbstractParameters()
  , m_rowPropertyName(rowPropertyName)
  , m_columnPropertyName(columnPropertyName)
  , m_dissolveMaxCoordinates(dissolveMaxCoordinates)
{

}

te::vp::DissolveSubdivisionOperationParameters::~DissolveSubdivisionOperationParameters() = default;

bool te::vp::DissolveSubdivisionOperationParameters::isValid(std::string& errorMessage)
{
  if (m_rowPropertyName.empty() == true)
  {
    errorMessage = TE_TR("Row property name cannot be empty");
    return false;
  }

  if (m_columnPropertyName.empty() == true)
  {
    errorMessage = TE_TR("Column property name cannot be empty");
    return false;
  }
  if (m_rowPropertyName == m_columnPropertyName)
  {
    errorMessage = TE_TR("Row property name cannot be equal to the column property name");
    return false;
  }

  for (std::size_t i = 0; i < this->m_inputDataVector.size(); ++i)
  {
    te::da::DataAccess* dataAccess = this->m_inputDataVector.at(i);
    const te::da::DataSetType* dataSetType = dataAccess->getDataSetType();
    for (std::size_t j = 0; j < dataSetType->size(); ++j)
    {
      te::dt::Property* property = dataSetType->getProperty(j);
      if (property->getType() == te::dt::FLOAT_TYPE || property->getType() == te::dt::DOUBLE_TYPE)
      {
        errorMessage = TE_TR("This operation cannot handle floating point properties yet. Remove them and try again!");
        return false;
      }
    }
  }

  return te::vp::AbstractParameters::isValid(errorMessage);
}

void te::vp::DissolveSubdivisionOperationParameters::setRowPropertyName(const std::string& rowPropertyName)
{
  m_rowPropertyName = rowPropertyName;
}

void te::vp::DissolveSubdivisionOperationParameters::setColumnPropertyName(const std::string& columnPropertyName)
{
  m_columnPropertyName = columnPropertyName;
}

void te::vp::DissolveSubdivisionOperationParameters::setDissolveMaxCoordinates(bool dissolveMaxCoordinates)
{
  m_dissolveMaxCoordinates = dissolveMaxCoordinates;
}

const std::string& te::vp::DissolveSubdivisionOperationParameters::getRowPropertyName() const
{
  return m_rowPropertyName;
}

const std::string& te::vp::DissolveSubdivisionOperationParameters::getColumnPropertyName() const
{
  return m_columnPropertyName;
}

bool te::vp::DissolveSubdivisionOperationParameters::getDissolveMaxCoordinates() const
{
  return m_dissolveMaxCoordinates;
}

te::vp::DissolveSubdivisionOperationCapabilities::DissolveSubdivisionOperationCapabilities(const std::string& operationName, std::size_t numberOfInputs, std::size_t numberOfOutputs, const std::string& dataHandlerName, bool snapGeometries, bool allowCoordinateSubdivision)
  : te::vp::AbstractOperationCapabilities(operationName, numberOfInputs, numberOfOutputs, dataHandlerName, snapGeometries, allowCoordinateSubdivision)
{

}

std::unique_ptr<te::vp::AbstractParameters> te::vp::DissolveSubdivisionOperationCapabilities::createParameters() const
{
  std::unique_ptr<te::vp::AbstractParameters> params(new te::vp::DissolveSubdivisionOperationParameters());
  return params;
}

te::vp::DissolveSubdivisionOperation::DissolveSubdivisionOperation()
    : AbstractOperation(te::vp::dissolve_subdivision::OPERATION_NAME)
 {

 }

std::unique_ptr<te::vp::AbstractOperationCapabilities> te::vp::DissolveSubdivisionOperation::getCapabilities() const
{
  std::unique_ptr<te::vp::AbstractOperationCapabilities> capabilities(new te::vp::DissolveSubdivisionOperationCapabilities(getName(), 1, 1, "AttributeGroupingDataHandler", false, false));
  return capabilities;
}

std::vector<te::vp::FeatureSet> te::vp::DissolveSubdivisionOperation::executeImpl(const std::vector<te::vp::FeatureSet>& vecInput)
{
  std::vector<te::vp::FeatureSet> vecOutput;

  if (vecInput.size() != 1)
  {
    vecOutput.resize(1);
    return vecOutput;
  }

  if (vecInput.at(0).empty() == true)
  {
    vecOutput.resize(1);
    return vecOutput;
  }

  te::vp::FeatureSet outputFeatureSet;

  const te::vp::FeatureSet& fsRed = vecInput.at(0);
  if (fsRed.empty() == true)
  {
    vecOutput.resize(1);
    return vecOutput;
  }

  te::vp::DissolveSubdivisionOperationParameters* params = dynamic_cast<te::vp::DissolveSubdivisionOperationParameters*>(this->getParameters());
  if (params == nullptr)
  {
    vecOutput.resize(1);
    return vecOutput;
  }

  const std::string& rowPropertyName = params->getRowPropertyName();
  const std::string& columnPropertyName = params->getColumnPropertyName();
  bool dissolveMaxCoordinatesSubdivision = params->getDissolveMaxCoordinates();

  const  te::vp::Feature* seedFeature = fsRed.at(0);
  std::size_t rowIndex = seedFeature->getPropertyPos(rowPropertyName);
  std::size_t columnIndex = seedFeature->getPropertyPos(columnPropertyName);

  te::vp::FeatureSet dissolvedFS = te::vp::DissolveCellSubdivision::Dissolve(fsRed, rowIndex, columnIndex, dissolveMaxCoordinatesSubdivision);

  for (std::size_t i = 0; i < dissolvedFS.size(); ++i)
  {
    te::gm::GeometryVector vecSingleResult;
    te::gm::Multi2Single(dissolvedFS.at(i)->getFirstGeometryPtr(), vecSingleResult);

    te::mem::PreparedItemCopy copy;
    for (std::size_t g = 0; g < vecSingleResult.size(); ++g)
    {
      //create the output feature based on the resulting geometry and the input attributes
      te::gm::GeometryVector vecAdjustedOutput = te::vp::AdjustToGeomType(vecSingleResult.at(g), getOutputGeometryType());
      for (std::size_t k = 0; k < vecAdjustedOutput.size(); ++k)
      {
        //create the output feature based on the resulting geometry and the input attributes
        te::vp::Feature* featureOutput = createOutputFeature(0);
        copy.copyValues(seedFeature, featureOutput, false);
        featureOutput->setFirstGeometry(vecAdjustedOutput.at(k));

        //and add it to the output list
        outputFeatureSet.push_back(featureOutput);
      }
    }

    te::common::FreeContents(vecSingleResult);
  }

  te::common::FreeContents(dissolvedFS);

  vecOutput.push_back(outputFeatureSet);
  return vecOutput;
}

te::vp::DissolveSubdivisionOperationFactory::DissolveSubdivisionOperationFactory()
  : OperationFactory(te::vp::dissolve_subdivision::OPERATION_NAME)
{
}

te::vp::DissolveSubdivisionOperationFactory::~DissolveSubdivisionOperationFactory() = default;

te::vp::AbstractOperation* te::vp::DissolveSubdivisionOperationFactory::build()
{
  return new DissolveSubdivisionOperation();
}
