/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

#include "Feature.h"

#include "../dataaccess/dataset/DataSetItem.h"
#include "../common/STLUtils.h"
#include "../dataaccess/utils/Utils.h"
#include "../datatype/Enums.h"
#include "../geometry/Geometry.h"
#include "../memory/DataSetItem.h"

te::vp::Feature::Feature(const te::da::DataSet* dataSet)
  : te::mem::DataSetItem(dataSet)
  , m_firstGeometryPropertyPos(std::string::npos)
{
  m_firstGeometryPropertyPos = getFirstGeometryPropertyPos();
}

te::vp::Feature::Feature(const te::da::DataSetType* dataSetType)
  : te::mem::DataSetItem(dataSetType)
  , m_firstGeometryPropertyPos(std::string::npos)
{
  m_firstGeometryPropertyPos = getFirstGeometryPropertyPos();
}

te::vp::Feature::Feature(const te::da::DataSetItem& dataSetItem)
  : te::mem::DataSetItem(dataSetItem.getPropertyNames(), dataSetItem.getPropertyTypes())
  , m_firstGeometryPropertyPos(std::string::npos)
{
  for (std::size_t i = 0; i < m_vecNames.size(); ++i)
  {
    if (dataSetItem.isNull(i) == true)
    {
      this->m_data.replace(i, nullptr);
    }
    else
    {
      this->m_data.replace(i, dataSetItem.getValue(i).release());
    }
  }

  m_firstGeometryPropertyPos = getFirstGeometryPropertyPos();
}

te::vp::Feature::Feature(const Feature& rhs)
  : te::mem::DataSetItem(rhs)
  , m_firstGeometryPropertyPos(rhs.m_firstGeometryPropertyPos)
{
  
}

te::vp::Feature::~Feature() = default;

te::da::DataSetItem* te::vp::Feature::clone() const
{
  return new te::vp::Feature(*this);
}

std::unique_ptr<te::gm::Geometry> te::vp::Feature::getFirstGeometry() const
{
  if (m_firstGeometryPropertyPos != std::string::npos && isNull(m_firstGeometryPropertyPos) == false)
  {
    return getGeometry(m_firstGeometryPropertyPos);
  }

  return std::unique_ptr<te::gm::Geometry>();
}

const te::gm::Geometry* te::vp::Feature::getFirstGeometryPtr() const
{
  if (m_firstGeometryPropertyPos != std::string::npos && isNull(m_firstGeometryPropertyPos) == false)
  {
    const te::gm::Geometry* geometry = static_cast<const te::gm::Geometry*>(&m_data.at(m_firstGeometryPropertyPos));
    return geometry;
  }

  return nullptr;
}

te::gm::Geometry* te::vp::Feature::getFirstGeometryPtr()
{
  if (m_firstGeometryPropertyPos != std::string::npos && isNull(m_firstGeometryPropertyPos) == false)
  {
    te::gm::Geometry* geometry = static_cast<te::gm::Geometry*>(&m_data.at(m_firstGeometryPropertyPos));
    return geometry;
  }

  return nullptr;
}

void te::vp::Feature::setFirstGeometry(te::gm::Geometry* geometry)
{
  if (m_firstGeometryPropertyPos == std::string::npos)
  {
    return;
  }

  this->setGeometry(m_firstGeometryPropertyPos, geometry);
}

void te::vp::Feature::transform(int srid)
{
  for (std::size_t i = 0; i < m_vecTypes.size(); ++i)
  {
    if (m_vecTypes.at(i) == te::dt::GEOMETRY_TYPE && isNull(i) == false)
    {
      te::gm::Geometry* geometry = static_cast<te::gm::Geometry*>(&m_data.at(i));
      if (geometry->getSRID() != srid)
      {
        geometry->transform(srid);
      }
    }
  }
}

void te::vp::Feature::transform(te::srs::Converter* converter)
{
  for (std::size_t i = 0; i < m_vecTypes.size(); ++i)
  {
    if (m_vecTypes.at(i) == te::dt::GEOMETRY_TYPE && isNull(i) == false)
    {
      te::gm::Geometry* geometry = static_cast<te::gm::Geometry*>(&m_data.at(i));
      geometry->convert(converter);
    }
  }
}

void te::vp::Feature::setSRID(int srid)
{
  for (std::size_t i = 0; i < m_vecTypes.size(); ++i)
  {
    if (m_vecTypes.at(i) == te::dt::GEOMETRY_TYPE && isNull(i) == false)
    {
      te::gm::Geometry* geometry = static_cast<te::gm::Geometry*>(&m_data.at(i));
      geometry->setSRID(srid);
    }
  }
}

std::size_t te::vp::Feature::getFirstGeometryPropertyPos() const
{
  std::size_t firstGeometryPos = std::string::npos;
  for (std::size_t i = 0; i < m_vecTypes.size(); ++i)
  {
    if (m_vecTypes.at(i) == te::dt::GEOMETRY_TYPE)
    {
      firstGeometryPos = i;
      break;
    }
  }

  return firstGeometryPos;
}

te::vp::Feature* te::vp::Feature::createFeatureFromCurrent(const te::da::DataSet* dataSet)
{
  if (dataSet == nullptr)
    return nullptr;

  std::size_t numProperties = dataSet->getNumProperties();

  te::vp::Feature* newItem = new te::vp::Feature(dataSet);
  for (std::size_t i = 0; i < numProperties; ++i)
  {
    if (dataSet->isNull(i) == false)
    {
      newItem->m_data.replace(i, dataSet->getValue(i).release());
    }
  }

  return newItem;
}
