/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralib/geometry/Feature.h

   \brief A feature is a composition of a geometry and its attributes
 */

#ifndef __TERRALIB_VP_INTERNAL_FEATURE_H
#define __TERRALIB_VP_INTERNAL_FEATURE_H

 // TerraLib
#include "Config.h"

#include "../memory/DataSetItem.h"

#include <memory>
#include <vector>

namespace te
{
  namespace da
  {
    class DataSetItem;
    class DataSetType;
  }

  namespace gm
  {
    class Geometry;
  }

  namespace srs
  {
    class Converter;
  }

  namespace vp
  {
    class Feature;

    //!< An alias for a set of Features
    class TEVPEXPORT FeatureSet : public std::vector<Feature*> {};

    /*!
      \class Feature

      \brief A feature is a composition of a geometry and its attributes

      \ingroup geometry
     */
    class TEVPEXPORT Feature : public te::mem::DataSetItem
    {
    public:

      explicit Feature(const te::da::DataSet* dataSet);

      explicit Feature(const te::da::DataSetType* dataSetType);

      /*!
      * \brief Constructor
        \param dataSetItem The dataset item to be associated with this feature. Feature will take the ownership of the given dataset item
      */
      explicit Feature(const te::da::DataSetItem& dataSetItem);

      /*!
        \brief It creates a new feature by cloning the values in the source feature (rhs).

        \param rhs The right-hand-side object to copy its values.
      */
      explicit Feature(const Feature& rhs);

      /*!
      * \brief Destructor
      */
      virtual ~Feature();

      /*!
      * \brief Clones the feature
      */
      virtual te::da::DataSetItem* clone() const override;

      /*!
      * \brief Gets the first geometry associated to this feature. The caller will take the ownership of the object
      *
      * \return The first geometry associated to this feature
      */
      std::unique_ptr<te::gm::Geometry> getFirstGeometry() const;

      /*!
      * \brief Gets the first geometry associated to this feature. The caller will NOT take the ownership of the object
      *
      * \return The first geometry associated to this feature
      */
      const te::gm::Geometry* getFirstGeometryPtr() const;

      /*!
      * \brief Gets the first geometry associated to this feature. The caller will NOT take the ownership of the object
      *
      * \return The first geometry associated to this feature
      */
      te::gm::Geometry* getFirstGeometryPtr();

      /*!
      * \brief Sets the first geometry value associated to this feature. This will take the ownership of the object
      *
      * \param The geometry to be associated to this feature
      */
      void setFirstGeometry(te::gm::Geometry* geometry);

      /*!
      * \brief It converts the coordinate values of the feature to the new spatial reference system.
      *
      * \param The new Spatial Reference System ID used to transform the coordinates of the feature.
      */
      void transform(int srid);

      /*!
      * \brief It converts the coordinate values of the feature to the new spatial reference system.
      *
      * \param The new Spatial Reference System ID used to transform the coordinates of the feature.
      */
      void transform(te::srs::Converter* converter);

      //!< Sets the srid of the feature. The given value will be applied to all geometries
      void setSRID(int srid);

      /*!
        \brief Creates a Feature based on the current row of the givem DataSet
      */
      static te::vp::Feature* createFeatureFromCurrent(const te::da::DataSet* dataSet);

    protected:

      std::size_t getFirstGeometryPropertyPos() const;

    protected:

      std::size_t m_firstGeometryPropertyPos;
    };
  }
}

#endif //__TERRALIB_VP_INTERNAL_FEATURE_H
