/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
 \file GeometrySnapper.cpp
 */

#include "GeometrySnapper.h"

#include "../common/Exception.h"
#include "../common/STLUtils.h"
#include "../core/translator/Translator.h"
#include "../geometry/CommonDataStructures.h"
#include "../geometry/CoordinateSnapper.h"
#include "../geometry/GeometryFunctions.h"
#include "../geometry/Utils.h"
#include "../geometry/Validation.h"

#include "CommonDataStructures.h"
#include "CommonFunctions.h"
#include "Feature.h"
#include "Utils.h"

te::gm::GeometryVector te::vp::GeometrySnapper::snapGeometries(const te::gm::GeometryVectorConst& vecGeometries, double tolerance)
{
  bool wasChanged1 = false;
  te::gm::GeometryVector vecAddedProjectedCoordinates = te::gm::Distance::AddProjectedCoordinates(vecGeometries, tolerance, wasChanged1);

  te::gm::GeometryVector vecSnapTempResult = te::gm::CoordinateSnapper::snapCoordinates(te::gm::ToConstVector(vecAddedProjectedCoordinates), tolerance);
  //free the memory
  te::common::FreeContents(vecAddedProjectedCoordinates);

  bool wasChanged2 = false;
  te::gm::GeometryVector vecPrepareResult = te::gm::PrepareGeometriesToIntersection(te::gm::ToConstVector(vecSnapTempResult), wasChanged2, tolerance);

  //free the memory
  te::common::FreeContents(vecSnapTempResult);

  return vecPrepareResult;
}

te::gm::Geometry* te::vp::GeometrySnapper::snapGeometries(const te::gm::Geometry* geometry, double tolerance)
{
  te::gm::GeometryVectorConst geometryVector;
  geometryVector.push_back(geometry);

  te::gm::GeometryVector outputGeometryVector = snapGeometries(geometryVector, tolerance);
  return outputGeometryVector.at(0);
}

te::vp::FeatureSet te::vp::GeometrySnapper::snapFeatures(const te::vp::FeatureSet& featureSet, double tolerance, bool discardCollapsedFeatures)
{
  te::gm::GeometryVectorConst vecGeometries = te::vp::GetGeometriesFromFeatureSet(featureSet);
  
  te::gm::GeometryVector vecResult = snapGeometries(vecGeometries, tolerance);
     
  if (featureSet.size() != vecResult.size())
  {
    te::common::FreeContents(vecResult);

    throw te::common::Exception(TE_TR("Unexpected result while executing the snap of features"));
  }
   
  //we replace the geometry inside the feature for the new prepared geometry. That why we dont need to free the 'vecPrepareResult' content
  te::vp::FeatureSet featureSetResult;
  featureSetResult.reserve(featureSet.size());

  for (std::size_t i = 0; i < featureSet.size(); ++i)
  {
    const te::vp::Feature* inputFeature = featureSet.at(i);

    std::unique_ptr<te::gm::Geometry> outputGeometry(vecResult.at(i));
    if (te::gm::IsNullOrEmpty(outputGeometry.get()) == false)
    {
      if (outputGeometry->isValid() == false)
      {
        outputGeometry = te::gm::Validation::MakeValidWithSnap(outputGeometry.get());
      }
    }


    te::vp::Feature* outputFeature = new te::vp::Feature(*inputFeature);
    outputFeature->setFirstGeometry(outputGeometry.release());
    featureSetResult.push_back(outputFeature);
  }

  if (discardCollapsedFeatures == true)
  {
    te::vp::DiscardFeaturesWithNullAndEmptyGeometries(featureSetResult);
  }

  return featureSetResult;
}

te::gm::Geometry* te::vp::GeometrySnapper::removeSlivers(te::gm::Geometry* geometry, double tolerance)
{
  te::gm::GeometryVectorConst geometyVector;
  geometyVector.push_back(geometry);

  te::gm::GeometryVector resultGeometryVector = removeSlivers(geometyVector, tolerance);
  return resultGeometryVector.at(0);
}

te::gm::GeometryVector te::vp::GeometrySnapper::removeSlivers(const te::gm::GeometryVectorConst& geometryVector, double tolerance)
{
  bool wasChanged = false;
  te::gm::GeometryVector outputGeometryVector = te::gm::Distance::AddProjectedCoordinates(geometryVector, tolerance, wasChanged);
  te::gm::GeometryVector outputGeometryVector2 = te::gm::CoordinateSnapper::snapCoordinates(te::gm::ToConstVector(outputGeometryVector), tolerance);

  te::common::FreeContents(outputGeometryVector);

  return outputGeometryVector2;
}

te::vp::FeatureSet te::vp::GeometrySnapper::removeSlivers(const te::vp::FeatureSet& featureSet, double tolerance, bool discardCollapsedFeatures)
{
  te::vp::FeatureSet featureSetResult;
  te::common::Clone(featureSet, featureSetResult);

  te::gm::GeometryVectorConst vecGeometries = te::vp::GetGeometriesFromFeatureSet(featureSetResult);

  te::gm::GeometryVector vecResult = te::vp::GeometrySnapper::removeSlivers(vecGeometries, tolerance);

  if (featureSetResult.size() != vecResult.size())
  {
    te::common::FreeContents(featureSetResult);
    te::common::FreeContents(vecResult);

    throw te::common::Exception(TE_TR("Unexpected result while executing the snap of features"));
  }

  //we replace the geometry inside the feature for the new prepared geometry. That why we dont need to free the 'vecPrepareResult' content
  for (std::size_t i = 0; i < featureSetResult.size(); ++i)
  {
    std::unique_ptr<te::gm::Geometry> geometry(vecResult.at(i));
    if (te::gm::IsNullOrEmpty(geometry.get()) == true)
    {
      geometry.reset();
    }
    else if (geometry->isValid() == false)
    {
      geometry = te::gm::Validation::MakeValidWithSnap(geometry.get());
    }

    te::vp::Feature* feature = featureSetResult.at(i);
    feature->setFirstGeometry(geometry.release());
  }

  //free the memory
  te::common::FreeContents(vecResult);

  if (discardCollapsedFeatures == true)
  {
    te::vp::DiscardFeaturesWithNullAndEmptyGeometries(featureSetResult);
  }

  return featureSetResult;
}

