/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
 \file GeometrySnapper.h
 
 \brief Algorithm to snap geometries to each other based on a given tolerance
 
 \ingroup vp
 */

#ifndef __TERRALIB_VP_INTERNAL_GEOMETRYSNAPPER_H
#define __TERRALIB_VP_INTERNAL_GEOMETRYSNAPPER_H

#include "Config.h"

#include "../geometry/CommonDataStructures.h"

namespace te
{
  namespace gm
  {
    class Geometry;
  }

  namespace vp
  {
    class FeatureSet;

    /*!
      \class GeometrySnapper

      \brief Algorithm to snap geometries to each other based on a given tolerance

      \ingroup vp

      \sa CoordinateSnapper
    */
    class TEVPEXPORT GeometrySnapper
    {
    public:
      
      /*!
          \brief Algorithm to snap given geometries to each other based on a given tolerance. 
                This algorithm will also detect intersection segments between geometries and create new coordinates when necessary

          \param geometryVector The geometry list to be analysed
          \param tolerance The tolerance to be considered for the snapping of the coordinates.

          \return The resulting snapped geometries
        */
      static te::gm::GeometryVector snapGeometries(const te::gm::GeometryVectorConst& geometryVector, double tolerance);

      /*!
          \brief Algorithm to snap given geometries to each other based on a given tolerance.
                This algorithm will also detect intersection segments between geometries and create new coordinates when necessary

          \param geometryVector The geometry list to be analysed
          \param tolerance The tolerance to be considered for the snapping of the coordinates.

          \return The resulting snapped geometries
        */
      static te::gm::Geometry* snapGeometries(const te::gm::Geometry* geometry, double tolerance);

      /*!
          \brief Algorithm to snap given features to each other based on a given tolerance.
                This algorithm will also detect intersection segments between features and create new coordinates when necessary

          \param vecGeometries The geometry list to be analysed
          \param tolerance The tolerance to be considered for the snapping of the coordinates.
          \param discardCollapsedFeatures IF TRUE, all collapsed features will be discarded from the resulting featureSet

          \return The resulting snapped geometries
        */
      static te::vp::FeatureSet snapFeatures(const te::vp::FeatureSet& featureSet, double tolerance, bool discardCollapsedFeatures);

      /*!
          \brief Algorithm to remove slivers from a geometry.

          \param geometry The geometry to be analysed
          \param tolerance The tolerance to be considered for detecting the slivers.

          \return The resulting geometry
        */
      static te::gm::Geometry* removeSlivers(te::gm::Geometry* geometry, double tolerance);

      /*!
          \brief Algorithm to remove slivers from geometries.

          \param geometryVector The geometry list to be analysed
          \param tolerance The tolerance to be considered for detecting the slivers.

          \return The resulting geometries
        */
      static te::gm::GeometryVector removeSlivers(const te::gm::GeometryVectorConst& geometryVector, double tolerance);

      /*!
          \brief Algorithm to remove slivers from features.

          \param vecGeometries The geometry list to be analysed
          \param tolerance The tolerance to be considered for detecting the slivers.

          \param discardCollapsedFeatures IF TRUE, all collapsed features will be discarded from the resulting featureSet

          \return The resulting features
        */
      static te::vp::FeatureSet removeSlivers(const te::vp::FeatureSet& featureSet, double tolerance, bool discardCollapsedFeatures);

      //!< Discard all the features that are null or empty
      static void discardNullAndEmptyFeatures(te::vp::FeatureSet& featureSet);
    };
  }
}
#endif // __TERRALIB_VP_INTERNAL_GEOMETRYSNAPPER_H
