/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralin/vp/IdentityOperation.cpp

   \brief Represents a identity operation between two sets of features
 */
#include "IdentityOperation.h"

#include "../geometry/Geometry.h"
#include "../geometry/Utils.h"
#include "../memory/PreparedItemCopy.h"
#include "../sam/rtree/Index.h"
#include "Feature.h"
#include "OperationService.h"
#include "DifferenceOperation.h"
#include "IntersectionOperation.h"

namespace te
{
  namespace vp
  {
    namespace identityOperation
    {
      const std::string OPERATION_NAME = "Identity";
    }
  }
}

namespace
{
  static te::vp::IdentityOperationFactory
    IdentityOperationFactoryInstance;
}

te::vp::IdentityOperation::IdentityOperation()
    : AbstractOperation(te::vp::identityOperation::OPERATION_NAME)
{

}

te::vp::IdentityOperation::~IdentityOperation() = default;

void te::vp::IdentityOperation::initialize()
{
  AbstractOperation::initialize();

  //initializes the intersection operation
  m_intersectionOperation.reset(new IntersectionOperation());
  m_intersectionOperation->setParameters(new AbstractParameters(*this->getParameters()));
  m_intersectionOperation->initialize();

  //initializes the difference operation
  m_differenceOperation.reset(new DifferenceOperation());

  AbstractParameters* differenceParams = new AbstractParameters(*this->getParameters());
  m_differenceOperation->setParameters(differenceParams);
  m_differenceOperation->initialize();
}

std::vector<te::vp::FeatureSet> te::vp::IdentityOperation::executeImpl(const std::vector<te::vp::FeatureSet>& vecInput)
{
  std::vector<te::vp::FeatureSet> vecOutput;
  if (vecInput.size() != 2)
  {
    vecOutput.resize(1);
    return vecOutput;
  }

  te::vp::FeatureSet fullResult;

  //intersection (A & B)
  {
    std::vector<te::vp::FeatureSet> vecSingleOutput = m_intersectionOperation->executeImpl(vecInput);
    fullResult.insert(fullResult.end(), vecSingleOutput.at(0).begin(), vecSingleOutput.at(0).end());
  }

  //difference (A - B)
  {
    std::vector<te::vp::FeatureSet> vecSingleOutput = m_differenceOperation->executeImpl(vecInput);
    fullResult.insert(fullResult.end(), vecSingleOutput.at(0).begin(), vecSingleOutput.at(0).end());
  }

  te::mem::PreparedItemCopy copy;

  //we must convert the geometries to the Identity operation output type
  for (std::size_t i = 0; i < fullResult.size(); ++i)
  {
    te::vp::Feature* feature = fullResult.at(i);
    te::vp::Feature* convertedFeature = createOutputFeature(0);
    copy.copyValues(feature, convertedFeature, false);
    convertedFeature->setFirstGeometry(feature->getFirstGeometry().release());

    delete feature;
    fullResult[i] = convertedFeature;
  }

  vecOutput.push_back(fullResult);
  
  return vecOutput;
}

te::vp::IdentityOperationFactory::IdentityOperationFactory()
  : OperationFactory(te::vp::identityOperation::OPERATION_NAME)
{
}

te::vp::IdentityOperationFactory::~IdentityOperationFactory() = default;

te::vp::AbstractOperation* te::vp::IdentityOperationFactory::build()
{
  return new IdentityOperation();
}
