/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralin/vp/IntersectionOperation.cpp

   \brief Represents a intersection operation between two sets of features
 */
#include "IntersectionOperation.h"

#include "../common/STLUtils.h"
#include "../geometry/algorithms/SafeOverlay.h"
#include "../geometry/CommonDataStructures.h"
#include "../geometry/Geometry.h"
#include "../geometry/Utils.h"
#include "../geometry/Validation.h"
#include "../memory/PreparedItemCopy.h"
#include "../sam/rtree/Index.h"
#include "CommonFunctions.h"
#include "Feature.h"
#include "GeometrySnapper.h"
#include "OperationService.h"

namespace te
{
  namespace vp
  {
    namespace intersection
    {
      const std::string OPERATION_NAME = "Intersection";
    }
  }
}

namespace
{
  static te::vp::IntersectionOperationFactory
    IntersectionOperationFactoryInstance;
}

te::vp::IntersectionOperation::IntersectionOperation()
    : AbstractOperation(te::vp::intersection::OPERATION_NAME)
  {

  }

std::vector<te::vp::FeatureSet> te::vp::IntersectionOperation::executeImpl(const std::vector<te::vp::FeatureSet>& vecInput)
{
  std::vector<te::vp::FeatureSet> vecOutput;

  if (vecInput.size() != 2)
  {
    vecOutput.resize(1);
    return vecOutput;
  }

  if (vecInput.at(0).empty() == true || vecInput.at(1).empty() == true)
  {
    vecOutput.resize(1);
    return vecOutput;
  }

  te::vp::FeatureSet outputFeatureSet;

  const te::vp::FeatureSet& fsRed = vecInput.at(0);
  const te::vp::FeatureSet& fsBlue = vecInput.at(1);

  //then we index the geometries from the blue featureSet
  std::unique_ptr< te::sam::rtree::Index<std::size_t> > indexBlue = te::vp::OperationService::CreateIndex(fsBlue);

  te::mem::PreparedItemCopy copyRed;
  te::mem::PreparedItemCopy copyBlue;

  //finally, for each feature in the red list, we make the intersection to all candidades in the blue list
  for (std::size_t i = 0; i < fsRed.size(); ++i)
  {
    //operation was aborted
    if (isAborted() == true)
    {
      return vecOutput;
    }

    const te::vp::Feature* featureRed = fsRed.at(i);
    const te::gm::Geometry* geometryRed = featureRed->getFirstGeometryPtr();
    if (te::gm::IsNullOrEmpty(geometryRed) == true)
    {
      continue;
    }

    //we search for candidadtes
    std::vector<std::size_t> vecCandidatesBlue;
    indexBlue->search(*geometryRed->getMBR(), vecCandidatesBlue);

    for (std::size_t j = 0; j < vecCandidatesBlue.size(); ++j)
    {
      std::size_t candidateIndexBlue = vecCandidatesBlue.at(j);
      const te::vp::Feature* featureBlue = fsBlue.at(candidateIndexBlue);
      const te::gm::Geometry* geometryBlue = featureBlue->getFirstGeometryPtr();

      //we can discard the blue geometries which envelopes that just touches the envelope of the red geometry
      if (geometryRed->getMBR()->touches(*geometryBlue->getMBR()) == true)
      {
        continue;
      }

      //then we calculates the intersection
      te::gm::GeometryPtr geometryOutput = te::gm::SafeOverlay::Intersection(geometryRed, geometryBlue);
      if (te::gm::IsNullOrEmpty(geometryOutput) == true)
      {
        continue;
      }

      std::vector<te::gm::Geometry*> vecSingleGeometries;
      te::gm::Multi2Single(geometryOutput.get(), vecSingleGeometries);

      for (std::size_t g = 0; g < vecSingleGeometries.size(); ++g)
      {
        //create the output feature based on the resulting geometry and the input attributes
        te::gm::GeometryVector vecAdjustedOutput = te::vp::AdjustToGeomType(vecSingleGeometries.at(g), getOutputGeometryType());
        for (std::size_t k = 0; k < vecAdjustedOutput.size(); ++k)
        {
          //create the output feature based on the resulting geometry and the input attributes
          te::vp::Feature* featureOutput = createOutputFeature(0);
          copyRed.copyValues(featureRed, featureOutput, false);
          copyBlue.copyValues(featureBlue, featureOutput, false);
          featureOutput->setFirstGeometry(vecAdjustedOutput.at(k));

          //and add it to the output list
          outputFeatureSet.push_back(featureOutput);
        }
      }

      te::common::FreeContents(vecSingleGeometries);
    }
  }
  vecOutput.push_back(outputFeatureSet);
  return vecOutput;
}

te::vp::IntersectionOperationFactory::IntersectionOperationFactory()
  : OperationFactory(te::vp::intersection::OPERATION_NAME)
{
}

te::vp::IntersectionOperationFactory::~IntersectionOperationFactory() = default;

te::vp::AbstractOperation* te::vp::IntersectionOperationFactory::build()
{
  return new IntersectionOperation();
}
