/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
 \file KDTreeIndex.h
 
 \brief This class makes easier and more generic the use of the te::sam::rtree::Index
 
 \ingroup vp
 */

#ifndef __TERRALIB_VP_INTERNAL_KDTREEMULTIINDEX_H
#define __TERRALIB_VP_INTERNAL_KDTREEMULTIINDEX_H

// Terralib
#include "Config.h"

#include "../common/STLUtils.h"
#include "../sam/kdtree/Index.h"
#include "CommonDataStructures.h"

//STL
#include <memory>
#include <vector>

namespace te
{
  namespace vp
  {
    /*!
      \class KDTreeIndex

      \brief This class makes easier and more generic the use of the te::sam::kdtree::Index

      \ingroup vp
    */
    template<class T>
    class KDTreeMultiIndex
    {
      using KD_NODE = te::sam::kdtree::Node<te::gm::Coord2D, std::size_t, std::vector<std::size_t>, te::sam::kdtree::kd_node_m_dataset_tag>;

    public:
      
      //!< Constructor
      KDTreeMultiIndex();
      
      //!< Destructor
      ~KDTreeMultiIndex();

      //!< Clears the index
      void clear();

      //!< Inserts the given data considering the given envelope as key
      void insert(const te::gm::Coord2D& coord, const T& data);

      //!< Removes the given data considering  the given envelope as key
      void remove(const te::gm::Coord2D& coord, std::size_t dataIndex);

      //!< Gets all data from the KDTreeIndex
      std::unique_ptr<te::vp::IndexReport<T>> getData() const;

      //!< Searches from all data consdering the given mbr
      std::unique_ptr<te::vp::IndexReport<T>> search(const te::gm::Envelope& mbr) const;

    private:
      te::sam::kdtree::Index<KD_NODE> m_index; //!< The internal rtree to store the data indexes
      IndexContainer<T> m_indexContainer; //!< The indexed data container
    };
  }
}

//Implementation
template<class T>
te::vp::KDTreeMultiIndex<T>::KDTreeMultiIndex()
  : m_index(te::gm::Envelope())
{

}
template<class T>
te::vp::KDTreeMultiIndex<T>::~KDTreeMultiIndex()
{
  clear();
}

template<class T>
void te::vp::KDTreeMultiIndex<T>::clear()
{
  m_index.clear();
  m_indexContainer.clear();
}


template<class T>
void te::vp::KDTreeMultiIndex<T>::insert(const te::gm::Coord2D& coord, const T& data)
{
  std::size_t newDataIndex = m_indexContainer.insert(data);
  m_index.insert(coord, newDataIndex);
}

template<class T>
void te::vp::KDTreeMultiIndex<T>::remove(const te::gm::Coord2D& coord, std::size_t dataIndex)
{
  m_indexContainer.remove(dataIndex);
}

template<class T>
std::unique_ptr<te::vp::IndexReport<T>> te::vp::KDTreeMultiIndex<T>::getData() const
{
  std::vector<std::size_t> vecValidIndexes = m_indexContainer.getValidIndexes();

  std::unique_ptr<te::vp::IndexReport<T>> indexReport(new te::vp::IndexReport<T>(&m_indexContainer, vecValidIndexes));
  return indexReport;
}
 
template<class T>
std::unique_ptr<te::vp::IndexReport<T>> te::vp::KDTreeMultiIndex<T>::search(const te::gm::Envelope& mbr) const
{
  std::vector<KD_NODE*> vecNodesReport;
  m_index.search(mbr, vecNodesReport);

  std::vector<std::size_t> vecFullIndexesReport;
  for (std::size_t i = 0; i < vecNodesReport.size(); ++i)
  {
    const std::vector<std::size_t>& vecNodeIndexes = vecNodesReport.at(i)->getData();
    vecFullIndexesReport.insert(vecFullIndexesReport.end(), vecNodeIndexes.begin(), vecNodeIndexes.end());
  }

  std::vector<std::size_t> vecNetIndexesReport = m_indexContainer.getValidIndexes(vecFullIndexesReport);

  std::unique_ptr<te::vp::IndexReport<T>> indexReport(new te::vp::IndexReport<T>(&m_indexContainer, vecNetIndexesReport));
  return indexReport;
}

#endif // __TERRALIB_VP_INTERNAL_KDTREEINDEX_H
