/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
 \file FixGeometry.cpp
 */

#include "../common/progress/TaskProgress.h"
#include "../core/logger/Logger.h"
#include "../core/translator/Translator.h"
#include "../dataaccess/dataset/DataSet.h"
#include "../dataaccess/query/DataSetName.h"
#include "../dataaccess/query/EqualTo.h"
#include "../dataaccess/query/Field.h"
#include "../dataaccess/query/Fields.h"
#include "../dataaccess/query/From.h"
#include "../dataaccess/query/FromItem.h"
#include "../dataaccess/query/LiteralString.h"
#include "../dataaccess/query/PropertyName.h"
#include "../dataaccess/query/Select.h"
#include "../dataaccess/query/ST_Buffer.h"
#include "../dataaccess/query/ST_IsValid.h"
#include "../dataaccess/query/ST_MakeValid.h"
#include "../dataaccess/query/ST_Multi.h"
#include "../dataaccess/query/Where.h"

#include "../dataaccess/datasource/DataSourceCapabilities.h"
#include "../dataaccess/datasource/DataSourceTransactor.h"

#include "../dataaccess/utils/DataAccess.h"
#include "../dataaccess/utils/Utils.h"

#include "../geometry/FixGeometryTopology.h"
#include "../geometry/Geometry.h"
#include "../geometry/GeometryProperty.h"
#include "../geometry/Utils.h"
#include "../geometry/Validation.h"

#include "../memory/DataSet.h"
#include "../memory/DataSetItem.h"

#include "MakeGeometryValid.h"
#include "Utils.h"

#include "boost/lexical_cast.hpp"

bool te::vp::MakeGeometryValid::makeValid(const te::da::DataAccessPtr& dataAccess,
                                          std::string& errorMessage,
                                          te::da::ObjectIdSet* oidSet)
{
  
  const te::da::DataSourceCapabilities dsCapabilities = dataAccess->getDataSource()->getCapabilities();
  te::da::QueryCapabilities queryCapabilities = dsCapabilities.getQueryCapabilities();

  if(queryCapabilities.supportsSpatialSQLDialect())
    return makeValidQuery(dataAccess, errorMessage, oidSet);
  else
    return makeValidMemory(dataAccess, errorMessage, oidSet);
}

bool te::vp::MakeGeometryValid::makeValidQuery(const te::da::DataAccessPtr& dataAccess,
    std::string& errorMessage, da::ObjectIdSet* /*oidSet*/)
{
  // Creating and executing a query to fix geometries.
  //
  // NOTE: This code creates and executes a query using ST_MakeValid clause.
  //       The result is a te::da::DataSet with same attributes from input table
  //       plus fixed geometries.
  //

  const te::da::DataSetType* dsType = dataAccess->getDataSetType();

  te::gm::GeometryProperty* geometryProperty =
      te::da::GetFirstGeomProperty(dsType);

  te::da::Fields fields;

  std::vector<te::dt::Property*> props = dsType->getProperties();
  for(std::size_t p = 0; p < props.size(); ++p)
  {
    if(props[p]->getType() == te::dt::GEOMETRY_TYPE)
      continue;

    te::da::Field* field = new te::da::Field(props[p]->getName());
    fields.push_back(field);
  }

  te::da::ST_MakeValid st_makeValid(
      te::da::PropertyName(geometryProperty->getName()));
  te::da::Expression* e_makeValid = new te::da::ST_MakeValid(st_makeValid);

  te::da::Expression* e_buffer = new te::da::ST_Buffer(*e_makeValid, 0);

  te::da::Field* f_geometry;

  if(te::gm::IsMultiType(geometryProperty->getGeometryType()))
  {
    te::da::Expression* e_multi = new te::da::ST_Multi(*e_buffer);
    f_geometry = new te::da::Field(*e_multi, geometryProperty->getName());
  }
  else
  {
    f_geometry = new te::da::Field(*e_buffer, geometryProperty->getName());
  }

  fields.push_back(f_geometry);

  te::da::From from;
  te::da::FromItem* fromItem = new te::da::DataSetName(dataAccess->getDataSetName());
  from.push_back(fromItem);

  te::da::ST_IsValid st_isValid(
      te::da::PropertyName(geometryProperty->getName()));
  te::da::Expression* e_isValid = new te::da::ST_IsValid(st_isValid);

  te::da::Expression* e_false = new te::da::LiteralString("f");

  te::da::EqualTo equal(e_isValid, e_false);

  te::da::Where where(equal);

  te::da::Select select(fields, from, where);

  std::unique_ptr<te::da::DataSet> ds;

  try
  {
    ds.reset(dataAccess->getDataSource()->query(select).release());
  }
  catch (const te::common::Exception& e)
  {
    TE_LOG_ERROR(TE_TR("Vector Processing") + " - " +
                 TE_TR("Fix Geometry") + ": " + e.what());

    errorMessage = e.what();
    return false;
  }
  catch (const std::exception& e)
  {
    TE_LOG_ERROR(TE_TR("Vector Processing") + " - " +
                 TE_TR("Fix Geometry") + ": " + e.what());

    errorMessage = e.what();
    return false;
  }
  catch (...)
  {
    TE_LOG_ERROR(TE_TR("Vector Processing") + " - " +
                 TE_TR("Fix Geometry") + ": Unknown Exception.");

    errorMessage = "Unknown Exception.";
    return false;
  }


  // Updating input table.
  //
  // NOTE: This code uptades the invalid geometries.
  //

  // Get position of attr used as pk
  te::da::PrimaryKey* pk = dsType->getPrimaryKey();

  size_t numProps = pk->getProperties().size();
  std::vector<std::size_t> ids;

  for (size_t i = 0; i < numProps; ++i)
  {
    std::string propName = pk->getProperties()[i]->getName();
    ids.push_back(dsType->getPropertyPosition(propName));
  }

  std::set<int> posSet;
  posSet.insert((int)dsType->getPropertyPosition(geometryProperty->getName()));
  std::vector< std::set<int> > propsPos;

  for(std::size_t p = 0; p < ds->size(); ++p)
    propsPos.push_back(posSet);

  std::unique_ptr<te::da::DataSourceTransactor> transactor = dataAccess->getDataSource()->getTransactor();

  try
  {
    transactor->update(dataAccess->getDataSetName(), ds.get(), propsPos, ids);
  }
  catch (const te::common::Exception& e)
  {
    TE_LOG_ERROR(TE_TR("Vector Processing") + " - " +
                 TE_TR("Fix Geometry") + ": " + e.what());

    errorMessage = e.what();

    transactor->rollBack();
    return false;
  }
  catch (const std::exception& e)
  {
    TE_LOG_ERROR(TE_TR("Vector Processing") + " - " +
                 TE_TR("Fix Geometry") + ": " + e.what());

    errorMessage = e.what();

    transactor->rollBack();
    return false;
  }
  catch (...)
  {
    TE_LOG_ERROR(TE_TR("Vector Processing") + " - " +
                 TE_TR("Fix Geometry") + ": Unknown Exception.");

    errorMessage = "Unknown Exception.";

    transactor->rollBack();
    return false;
  }

  transactor->commit();

  return true;
}

bool te::vp::MakeGeometryValid::makeValidMemory(const te::da::DataAccessPtr& dataAccess,
                                                std::string& errorMessage,
                                                te::da::ObjectIdSet* oidSet)
{
  std::unique_ptr<te::da::DataSourceTransactor> transactor = dataAccess->getDataSource()->getTransactor();

  const te::da::DataSetType* dataSetType = dataAccess->getDataSetType();

  te::gm::GeometryProperty* geometryProperty =
      te::da::GetFirstGeomProperty(dataSetType);

  std::unique_ptr<te::da::DataSet> dataSetToFix;
  std::unique_ptr<te::mem::DataSet> fixedDataSet(new te::mem::DataSet(dataSetType));

  try
  {
    // Get the object id properties.
    std::vector<std::string> oidprops;
    te::da::GetOIDPropertyNames(dataSetType, oidprops);

    // Get dataSet.

    dataSetToFix = dataAccess->getDataSet();
    if (dataSetToFix == nullptr)
    {
      errorMessage = TE_TR("Error querying data!");
      return false;
    }

    std::unique_ptr<te::da::ObjectIdSet> oidSetToRemove(new te::da::ObjectIdSet());

    // Scrolls through the dataset.
    dataSetToFix->moveBeforeFirst();

    te::common::TaskProgress task("Fixing...");
    task.setTotalSteps((int)dataSetToFix->size());

    while(dataSetToFix->moveNext())
    {
      if (!task.isActive())
      {
        errorMessage = TE_TR("Fix geometry canceled!");
        return false;
      }

      task.pulse();

      //we just try to fix the geometries thar are not null
      //for the null geometries, we just remove them from the dataset
      if (dataSetToFix->isNull(geometryProperty->getName()) == false)
      {
        std::unique_ptr<te::gm::Geometry> geometryToFix =
          dataSetToFix->getGeometry(geometryProperty->getName());

        std::vector<te::gm::Geometry*> fixedGeometryVec;

        te::gm::TopologyValidationError topologyError;

        bool isValid = te::gm::Validation::CheckValidity(geometryToFix.get(), topologyError);
        if (isValid == true)
        {
          continue;
        }

        std::string makeValidErrorMessage;

        te::gm::GeometryPtr fixedGeometry;
        try
        {
          fixedGeometry.reset(te::gm::Validation::MakeValidWithSnap(geometryToFix.get()).release());
          if (te::gm::IsNullOrEmpty(fixedGeometry) == false)
          {
            isValid = te::gm::Validation::CheckValidity(fixedGeometry.get(), topologyError);
          }
          else
          {
            isValid = true;
          }
        }
        catch (const std::exception & e)
        {
          makeValidErrorMessage = e.what();
        }

        if (isValid == false)
        {
          TE_LOG_ERROR(TE_TR("Vector Processing") + " - " +
            TE_TR("Fix Geometry") + ": " +
            TE_TR("Could not fix geometry") + ": " + topologyError.m_message + "; " +
            TE_TR("Coordinate") + ": (" + boost::lexical_cast<std::string>(topologyError.m_coordinate.getX()) + "," +
            boost::lexical_cast<std::string>(topologyError.m_coordinate.getY()) + "); " +
            TE_TR("Geometry WKT") + ": " + geometryToFix->asText() + 
            TE_TR("ERROR MESSAGE") + ": " + makeValidErrorMessage );

          errorMessage = "Some geometries could not be fixed. Check the log file for more information.";

          continue;
        }

        if (te::gm::IsNullOrEmpty(fixedGeometry) == false)
        {
          te::gm::Multi2Single(fixedGeometry.get(), fixedGeometryVec);
        }

        for (std::size_t i = 0; i < fixedGeometryVec.size(); ++i)
        {
          te::mem::DataSetItem* dataSetItemMemory = GetFilledItemMemory(dataSetToFix.get());

          dataSetItemMemory->setGeometry(geometryProperty->getName(), fixedGeometryVec[i]);

          fixedDataSet->add(dataSetItemMemory);
        }
      }

      // Add the current object id, to remove from input dataset.
      te::da::ObjectId* oid = te::da::GenerateOID(dataSetToFix.get(), oidprops);
      oidSetToRemove->add(oid);

      for (std::string p : oidprops)
      {
        te::dt::Property* prop = dataSetType->getProperty(p);
        oidSetToRemove->addProperty(p, te::da::GetPropertyPos(dataSetType, p), prop->getType());
      }
    }

    std::map<std::string, std::string> options;

    transactor->begin();
    transactor->remove(dataAccess->getDataSetName(), oidSetToRemove.get());
    transactor->add(dataAccess->getDataSetName(), fixedDataSet.get(), options);
    transactor->commit();
  }
  catch (const te::common::Exception& e)
  {
    std::string logMessage = TE_TR("Vector Processing");
    logMessage += std::string(" - ") + TE_TR("Fix Geometry") + ": ";
    logMessage += TE_TR("Could not fix geometry") + std::string(": ") + e.what() + "; ";

    if (dataSetToFix != nullptr && dataSetToFix->isPositionValid() == true)
    {
      std::unique_ptr<te::gm::Geometry> geometryToFix = dataSetToFix->getGeometry(geometryProperty->getName());
      logMessage += TE_TR("Geometry WKT") + std::string(": ") + geometryToFix->asText();
    }

    TE_LOG_ERROR(logMessage);

    errorMessage = e.what();
    if (transactor->isInTransaction() == true)
    {
      transactor->rollBack();
    }
    return false;
  }
  catch (const std::exception& e)
  {
    std::string logMessage = TE_TR("Vector Processing");
    logMessage += std::string(" - ") + TE_TR("Fix Geometry") + ": ";
    logMessage += TE_TR("Could not fix geometry") + std::string(": ") + e.what() + "; ";

    if (dataSetToFix != nullptr && dataSetToFix->isPositionValid() == true)
    {
      std::unique_ptr<te::gm::Geometry> geometryToFix = dataSetToFix->getGeometry(geometryProperty->getName());
      logMessage += TE_TR("Geometry WKT") + std::string(": ") + geometryToFix->asText();
    }

    TE_LOG_ERROR(logMessage);

    errorMessage = e.what();

    if (transactor->isInTransaction() == true)
    {
      transactor->rollBack();
    }

    return false;
  }
  catch (...)
  {
    std::unique_ptr<te::gm::Geometry> geometryToFix = dataSetToFix->getGeometry(geometryProperty->getName());

    TE_LOG_ERROR(TE_TR("Vector Processing") + " - " +
                 TE_TR("Fix Geometry") + ": " +
                 TE_TR("Could not fix geometry") + ": Unknown Exception; " +
                 TE_TR("Geometry WKT") + ": " + geometryToFix->asText());

    errorMessage = "Unknown Exception.";

    if (transactor->isInTransaction() == true)
    {
      transactor->rollBack();
    }

    return false;
  }

  return true;
}
