/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralin/vp/MakeValidOperation.cpp

   \brief Represents a make valid operation which analyses and fixes features of a set
 */
#include "MakeValidOperation.h"

#include "../geometry/Geometry.h"
#include "../geometry/Utils.h"
#include "../geometry/Validation.h"
#include "../memory/PreparedItemCopy.h"
#include "CommonFunctions.h"
#include "Feature.h"
#include "OperationService.h"

namespace te
{
  namespace vp
  {
    namespace makeValidOperation
    {
      const std::string OPERATION_NAME = "MakeValid";
    }
  }
}

namespace
{
  static te::vp::MakeValidOperationFactory
    MakeValidOperationFactoryInstance;
}

te::vp::MakeValidOperation::MakeValidOperation()
    : AbstractOperation(te::vp::makeValidOperation::OPERATION_NAME)
{

}

te::vp::MakeValidOperation::~MakeValidOperation() = default;

std::unique_ptr<te::vp::AbstractOperationCapabilities> te::vp::MakeValidOperation::getCapabilities() const
{
  std::unique_ptr<te::vp::AbstractOperationCapabilities> capabilities(new te::vp::AbstractOperationCapabilities(getName(), 1, 2, "UnorderedGroupingDataHandler", false, false));
  return capabilities;
}

std::vector<te::vp::FeatureSet> te::vp::MakeValidOperation::executeImpl(const std::vector<te::vp::FeatureSet>& vecInput)
{
  std::vector<te::vp::FeatureSet> vecOutput;
  if (vecInput.size() != 1)
  {
    vecOutput.resize(2);
    return vecOutput;
  }

  const te::vp::FeatureSet& inputFeatureSet = vecInput.at(0);

  te::mem::PreparedItemCopy copy1;
  te::mem::PreparedItemCopy copy2;

  te::vp::FeatureSet outputFeatureSet;
  te::vp::FeatureSet errorsFeatureSet;

  //we must convert the geometries to the Identity operation output type
  for (std::size_t i = 0; i < inputFeatureSet.size(); ++i)
  {
    const te::vp::Feature* inputFeature = inputFeatureSet.at(i);
    const te::gm::Geometry* geometry = inputFeature->getFirstGeometryPtr();
    if (geometry == nullptr)
    {
      continue;
    }

    te::gm::GeometryPtr fixedGeometry;
    if (geometry->isValid() == true)
    {
      fixedGeometry.reset(static_cast<te::gm::Geometry*>(geometry->clone()));
    }
    else
    {
      fixedGeometry.reset(te::gm::Validation::MakeValidWithSnap(geometry).release());
    }

    if (te::gm::IsNullOrEmpty(fixedGeometry) == true)
    {
      //we store the geometries collapsed or with fix problem in the second output
      te::gm::GeometryVector vecAdjustedOutput = te::vp::AdjustToGeomType(geometry, getOutputGeometryType());
      for (std::size_t k = 0; k < vecAdjustedOutput.size(); ++k)
      {
        te::gm::GeometryPtr adjustedOutputGeometry(vecAdjustedOutput.at(k));

        te::vp::Feature* convertedFeature = createOutputFeature(1);
        copy2.copyValues(inputFeature, convertedFeature, false);
        convertedFeature->setFirstGeometry(adjustedOutputGeometry.release());

        errorsFeatureSet.push_back(convertedFeature);
      }
      continue;
    }

    te::gm::GeometryVector vecAdjustedOutput = te::vp::AdjustToGeomType(fixedGeometry.get(), getOutputGeometryType());
    for (std::size_t k = 0; k < vecAdjustedOutput.size(); ++k)
    {
      te::gm::GeometryPtr adjustedOutputGeometry(vecAdjustedOutput.at(k));

      te::vp::Feature* convertedFeature = createOutputFeature(0);
      copy1.copyValues(inputFeature, convertedFeature, false);
      convertedFeature->setFirstGeometry(adjustedOutputGeometry.release());

      outputFeatureSet.push_back(convertedFeature);
    }
  }

  vecOutput.push_back(outputFeatureSet);
  vecOutput.push_back(errorsFeatureSet);
  
  return vecOutput;
}

te::vp::MakeValidOperationFactory::MakeValidOperationFactory()
  : OperationFactory(te::vp::makeValidOperation::OPERATION_NAME)
{
}

te::vp::MakeValidOperationFactory::~MakeValidOperationFactory() = default;

te::vp::AbstractOperation* te::vp::MakeValidOperationFactory::build()
{
  return new MakeValidOperation();
}
