/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file MultipartToSinglepart.h

  \brief Multipart to Singlepart Vector Processing functions.
*/

#include "MultipartToSinglepart.h"

//Terralib
#include "../common/progress/TaskProgress.h"
#include "../core/logger/Logger.h"
#include "../core/translator/Translator.h"

#include "../dataaccess/dataset/DataSet.h"
#include "../dataaccess/dataset/ObjectId.h"
#include "../dataaccess/dataset/ObjectIdSet.h"
#include "../dataaccess/utils/Utils.h"
#include "../datatype/Property.h"
#include "../datatype/SimpleProperty.h"
#include "../datatype/StringProperty.h"

#include "../geometry/Geometry.h"
#include "../geometry/GeometryProperty.h"
#include "../geometry/Utils.h"

#include "../memory/DataSet.h"
#include "../memory/DataSetItem.h"
#include "../memory/PreparedItemCopy.h"

#include "Config.h"
#include "Exception.h"
#include "Feature.h"
#include "Utils.h"

// STL
#include <map>
#include <cmath>
#include <string>
#include <vector>

te::vp::MultipartToSinglepart::MultipartToSinglepart()
  : m_inputDataAccess(nullptr)
{
}

te::vp::MultipartToSinglepart::~MultipartToSinglepart() = default;

bool te::vp::MultipartToSinglepart::run()
{
  if (m_inputDataAccess == nullptr)
  {
    throw te::common::Exception(TE_TR("Input not set!"));
  }

  const te::da::DataSetType* inputDataSetType = m_inputDataAccess->getDataSetType();

  te::dt::Property* prop = te::da::GetFirstSpatialProperty(inputDataSetType);
  te::gm::GeometryProperty* geomProp = dynamic_cast<te::gm::GeometryProperty*>(prop);

  if(!te::gm::IsMultiType(geomProp->getGeometryType()))
  {
    throw te::common::Exception(TE_TR("This layer has not a multipart geometry!"));
  }

  return runAll();
}

bool te::vp::MultipartToSinglepart::runAll()
{
  //gets the input dataSet
  std::unique_ptr<te::da::DataSet> inputDataSet = m_inputDataAccess->getDataSet();
  std::size_t geomPos = te::da::GetFirstSpatialPropertyPos(inputDataSet.get());

  //creates the output dataSet
  std::unique_ptr<te::da::DataSetType> outDst = getOutDst();
  std::unique_ptr<te::mem::DataSet> outDs(new te::mem::DataSet(outDst.get()));

  inputDataSet->moveBeforeFirst();

  int pkCount = 0;

  te::mem::PreparedItemCopy copy;

  while(inputDataSet->moveNext())
  {
    try
    {
      if (inputDataSet->isNull(geomPos))
      {
        continue;
      }

      std::unique_ptr<te::gm::Geometry> geom = inputDataSet->getGeometry(geomPos);
      std::unique_ptr<te::da::DataSetItem> inputItem = inputDataSet->getItem();

      std::vector<te::gm::Geometry*> geoms;
      te::gm::Multi2Single(geom.get(), geoms);

      for (std::size_t g = 0; g < geoms.size(); ++g)
      {
        //we first copy all the values from the input item to the output item
        //we ignore the copy of the geometry as we will replace it
        te::vp::Feature* outputItem = new te::vp::Feature(outDs.get());
        copy.copyValues(inputItem.get(), outputItem, false);
        
        //then we set the value of the primary key
        outputItem->setInt32(0, static_cast<int32_t>(pkCount));
        ++pkCount;

        //we set the single part geometry
        outputItem->setFirstGeometry(geoms.at(g));

        //and finnaly we add the created item to the output dataSet
        outDs->add(outputItem);
      }
    }
    catch(te::common::Exception& e)
    {
      TE_LOG_ERROR("Vector Processing - Multipart to Singlepart - " + e.what());
      continue;
    }
  }

  te::vp::Save(m_outDsrc.get(), outDs.get(), outDst.get());

  return true;
}

void te::vp::MultipartToSinglepart::setInput(te::da::DataAccessPtr inputDataAccess)
{
  m_inputDataAccess = std::move(inputDataAccess);
}

void te::vp::MultipartToSinglepart::setOutput(te::da::DataSourcePtr outDsrc, const std::string& outDsName)
{
  m_outDsrc = outDsrc;
  m_outDsName = outDsName;
}

std::unique_ptr<te::da::DataSetType> te::vp::MultipartToSinglepart::getOutDst(bool onlySelected)
{
  std::unique_ptr<te::da::DataSetType> outDst(new te::da::DataSetType(m_outDsName));

  te::dt::Property* newId = new te::dt::SimpleProperty("newid", te::dt::INT32_TYPE, true);

  outDst->add(newId);

  te::da::PrimaryKey* pk = new te::da::PrimaryKey(m_outDsName + "mpk", outDst.get());

  pk->add(newId);

  outDst->setPrimaryKey(pk);

  const te::da::DataSetType* inputDataSetType = m_inputDataAccess->getDataSetType();
  for(std::size_t i = 0; i < inputDataSetType->size(); ++i)
  {
    te::dt::Property* oldP = inputDataSetType->getProperty(i);
    te::dt::Property* newP = oldP->clone();

    if(newP->getType() == te::dt::GEOMETRY_TYPE)
    {
      te::gm::GeometryProperty* geomP = dynamic_cast<te::gm::GeometryProperty*>(newP);

      geomP->setGeometryType(te::gm::GetSimpleType(geomP->getGeometryType()));
    }

    outDst->add(newP);
  }

  return outDst;
}
