/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralin/vp/AbstractDataHandler.cpp

   \brief This class represents handler to help reading and writing data from/to a dataSource by dividing it into smaller groups
 */

#include "NoGroupingDataHandler.h"

#include "../dataaccess/dataset/DataSetType.h"
#include "../dataaccess/utils/DataAccess.h"
#include "../dataaccess/utils/Utils.h"
#include "Feature.h"
#include "Utils.h"

namespace
{
  static te::vp::NoGroupingDataHandlerFactory NoGroupingDataHandlerFactoryInstance;
}

te::vp::NoGroupingDataHandler::NoGroupingDataHandler()
    : AbstractDataHandler()
  {

  }

std::size_t te::vp::NoGroupingDataHandler::getGroupCount() const
{
  return 1;
}


te::vp::FeatureSet te::vp::NoGroupingDataHandler::getFeatureSet(std::size_t groupIndex, std::size_t inputIndex, const te::gm::Envelope& filter)
{
  te::da::DataAccess* dataAccess = m_vecInputData.at(inputIndex);

  const te::da::DataSetType* dataSetType = dataAccess->getDataSetType();
  std::unique_ptr<te::da::DataSet> dataSet = dataAccess->getDataSet();

  dataSet->moveBeforeFirst();

  te::vp::FeatureSet featureSet;
  while (dataSet->moveNext())
  {
    std::unique_ptr<te::vp::Feature> feature(te::vp::Feature::createFeatureFromCurrent(dataSet.get()));
    featureSet.push_back(feature.release());
  }

  return featureSet;
}

void te::vp::NoGroupingDataHandler::addFeatureSet(std::size_t groupIndex, std::size_t outputIndex, const te::vp::FeatureSet& featureSet)
{
  te::da::DataAccess* dataAccess = m_vecOutputData.at(outputIndex);

  std::unique_ptr<te::da::DataSet> dataSet = te::vp::CreateDataSetFromFeatureSet(dataAccess->getDataSetName(), dataAccess->getDataSetType(), featureSet);
  dataAccess->saveDataSet(dataSet.release());
}

te::vp::NoGroupingDataHandlerFactory::NoGroupingDataHandlerFactory()
  : te::vp::AbstractDataHandlerFactory("NoGroupingDataHandler")
{

}

te::vp::NoGroupingDataHandlerFactory::~NoGroupingDataHandlerFactory() = default;

te::vp::AbstractDataHandler* te::vp::NoGroupingDataHandlerFactory::build()
{
  return new te::vp::NoGroupingDataHandler();
}