/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */


/*!
  \file terralib/vp/OperationReport.cpp

  \brief Class used as a report logger for Operations.
*/

#include "OperationReport.h"

// TerraLib
#include "../common/STLUtils.h"
#include "../core/logger/Logger.h"

// STL
#include <fstream>
#include <iostream>

te::vp::OperationReport::OperationReport() 
{
  m_reportStr << "----------------------------------------------------------------------------" << std::endl;
}

te::vp::OperationReport::~OperationReport()
{
  m_clockTime.clear();
}

void te::vp::OperationReport::addTitle(const std::string& message)
{
  m_reportStr << "------------------- " << message << " -------------------" <<std::endl;
  m_reportStr << std::endl;
}

void te::vp::OperationReport::addSubTitle(const std::string& message)
{
  m_reportStr << std::endl;
  m_reportStr << "------- " << message << " -------" << std::endl;
  m_reportStr << std::endl;
}

void te::vp::OperationReport::addMessage(const std::string& message, const std::string& severity, bool appendTime)
{
  std::string time;

  if (appendTime)
  {
    time = te::common::GetCurrentTime();
  }

  m_reportStr << "--- " << "[ " + time + " ]" << "( " + severity + " ) "<< message << std::endl;
}

void te::vp::OperationReport::addStartMessage(const std::string& message, int id)
{
  m_clockTime[id] = std::clock();

  std::string time = te::common::GetCurrentTime();

  m_reportStr << std::endl;
  m_reportStr << "--- ## START ## " << "[ " + time + " ] " << "( TIME ) " << message << std::endl;
  m_reportStr << std::endl;
}

void te::vp::OperationReport::addEndMessage(const std::string& message, int id)
{
  std::clock_t start = m_clockTime[id];

  double timeInSeconds = (std::clock() - start) / (double)CLOCKS_PER_SEC;

  std::string time = te::common::GetCurrentTime();

  m_reportStr << std::endl;
  m_reportStr << "--- ## END ## " << "[ " + time + " ] " << "( TIME ) " << message << std::endl;
  m_reportStr << "--- ## ELAPSED TIME IN SECONDS ## " << "[ " + std::to_string(timeInSeconds) + " ]" << std::endl;
  m_reportStr << std::endl;
}

void te::vp::OperationReport::addKeyValueMessage(const std::string& key, const std::string& value)
{
  m_reportStr << "--- PARAMETER: " << key << "\t\t" << "VALUE: " << value << std::endl;
}

void te::vp::OperationReport::addEmptyLine()
{
  m_reportStr << std::endl;
}

void te::vp::OperationReport::toFile(const std::string& reportPath)
{
  std::ofstream reportFile(reportPath, std::ios::app | std::ios::out);

  reportFile << m_reportStr.str();

  reportFile.close();
}

std::string te::vp::OperationReport::toString()
{
  return m_reportStr.str();
}
