/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralin/vp/OperationService.cpp

   \brief Represents a service containing useful functions related to an operation
 */
#include "OperationService.h"

#include "../geometry/CommonDataStructures.h"
#include "../geometry/Geometry.h"
#include "../geometry/Utils.h"
#include "../geometry/Validation.h"
#include "CommonFunctions.h"
#include "Feature.h"
#include "GeometrySnapper.h"
#include "Utils.h"

#include <memory>

std::vector<te::vp::FeatureSet> te::vp::OperationService::SnapGeometries(const std::vector<te::vp::FeatureSet>& vecInput)
{

  //we must first join all the featuresSets so they can be handled as one
  //after that, we must split them into the original sets
  //and finally we must erase the null and the empty features
  te::vp::FeatureSet allFeatures;
  for (std::size_t i = 0; i < vecInput.size(); ++i)
  {
    const te::vp::FeatureSet& inputFeatureSet = vecInput.at(i);
    allFeatures.insert(allFeatures.end(), inputFeatureSet.begin(), inputFeatureSet.end());
  }

  if (allFeatures.empty() == true)
  {
    std::vector<te::vp::FeatureSet> vecOutput;
    vecOutput.resize(vecInput.size());
    return vecOutput;
  }

  //now we snap all the features from all featureset at once
  std::unique_ptr<te::gm::Geometry> firstGeometry = allFeatures.at(0)->getFirstGeometry();
  double precision = te::gm::GetAppropriatePrecision(firstGeometry->getSRID());

  //we first ensure that the geometries are snapped inside each featureSet to avoid the creation of small and invalid artefacts between geometries
  te::vp::FeatureSet allFeaturesSnapped = te::vp::GeometrySnapper::snapFeatures(allFeatures, precision, false);

  //then we split all the snapped geometries into their original sets
  std::vector<te::vp::FeatureSet> vecOutput;
  std::size_t firstIndexFromCurrentSet = 0;
  for (std::size_t i = 0; i < vecInput.size(); ++i)
  {
    const te::vp::FeatureSet& inputFeatureSet = vecInput.at(i);
    std::size_t originalSize = inputFeatureSet.size();
    std::size_t finalIndexFromCurrentSet = firstIndexFromCurrentSet + originalSize;

    te::vp::FeatureSet outputFeatureSet;
    outputFeatureSet.insert(outputFeatureSet.end(), allFeaturesSnapped.begin() + firstIndexFromCurrentSet, allFeaturesSnapped.begin() + finalIndexFromCurrentSet);

    //now we clear the empty and null features
    te::vp::DiscardFeaturesWithNullAndEmptyGeometries(outputFeatureSet);

    //and finally we split the multi type features into single type feature
    te::vp::FeatureSet outputSingleFeatureSet;
    te::vp::Multi2Single(outputFeatureSet, outputSingleFeatureSet);

    te::common::FreeContents(outputFeatureSet);

    vecOutput.push_back(outputSingleFeatureSet);

    firstIndexFromCurrentSet += originalSize;
  }

  return vecOutput;
}

std::unique_ptr< te::sam::rtree::Index<std::size_t> > te::vp::OperationService::CreateIndex(const te::vp::FeatureSet& inputFeatureSet)
{
  std::unique_ptr< te::sam::rtree::Index<std::size_t> > index(new te::sam::rtree::Index<std::size_t>());
  for (std::size_t i = 0; i < inputFeatureSet.size(); ++i)
  {
    const te::vp::Feature* feature = inputFeatureSet.at(i);
    const te::gm::Geometry* geometry = feature->getFirstGeometryPtr();
    if (geometry == nullptr || geometry->isEmpty() == true)
    {
      continue;
    }

    index->insert(*geometry->getMBR(), i);
  }

  return index;
}
