/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralin/vp/OperationThreadManager.cpp

   \brief This class represents a vector processing engine to help the processing of a large number of geometries by dividing it into smallers groups
 */

#include "OperationThreadManager.h"

#include "../common/StringUtils.h"
#include "../common/STLUtils.h"
#include "../common/ThreadGroup.h"
#include "../common/progress/TaskProgress.h"
#include "../core/translator/Translator.h"
#include "../geometry/Geometry.h"

#include "AbstractOperation.h"
#include "AbstractDataHandler.h"
#include "Feature.h"
#include "OperationReport.h"
#include "OperationThreadManagerListener.h"

#include <chrono>
#include <vector>


te::vp::OperationThreadManager::OperationThreadManager(AbstractOperation* operation, AbstractDataHandler* dataHandler)
  : m_operation(operation)
  , m_dataHandler(dataHandler)
  , m_thread(nullptr)
  , m_task(nullptr)
{
  
}

te::vp::OperationThreadManager::~OperationThreadManager()
{
  stopThread();
}

void ExecuteOperation(te::vp::AbstractOperation* operation, te::vp::AbstractDataHandler* dataHandler, std::size_t groupIndex, te::common::TaskProgress* taskProgress)
{
  std::string currentTile = te::common::Convert2String(static_cast<boost::uint32_t>(groupIndex));
  std::string totalTiles = te::common::Convert2String(static_cast<boost::uint32_t>(dataHandler->getGroupCount()));
  std::string tileMessage = currentTile + "/ " + totalTiles;

  if (taskProgress != nullptr && taskProgress->isActive() == false)
  {
    return;
  }

  if (operation->isAborted() == true)
  {
    return;
  }

  std::chrono::time_point<std::chrono::steady_clock> start = std::chrono::steady_clock::now();

  try
  {
    std::size_t inputCount = dataHandler->getInputCount();

    //gets the features from the data handler
    std::vector<te::vp::FeatureSet> vecInput;
    vecInput.reserve(inputCount);
    std::size_t numFeatures = 0;
    for (std::size_t dataIndex = 0; dataIndex < inputCount; ++dataIndex)
    {
      vecInput.push_back(dataHandler->getFeatureSet(groupIndex, dataIndex));
      numFeatures += vecInput.at(dataIndex).size();
    }

    //executes the operatin
    std::vector<te::vp::FeatureSet> vecOutput = operation->execute(vecInput);

    //free the contents of the input
    for (std::size_t dataIndex = 0; dataIndex < inputCount; ++dataIndex)
    {
      te::common::FreeContents(vecInput.at(dataIndex));
    }

    //adds the resulting features the data handler
    std::size_t outputCount = vecOutput.size();
    for (std::size_t dataIndex = 0; dataIndex < outputCount; ++dataIndex)
    {
      const te::vp::FeatureSet& featureSet = vecOutput.at(dataIndex);
      dataHandler->addFeatureSet(groupIndex, dataIndex, featureSet);

      //after storing the data, we free the memory
      te::common::FreeContents(featureSet);
    }

    if (taskProgress != nullptr)
    {
      taskProgress->pulse();
    }
  }
  catch (const std::exception& e)
  {
    operation->getReport()->addMessage(TE_TR("Error processing tile: ") + tileMessage + TE_TR(". Error: ") + std::string(e.what()), "ERROR", true);
    if (taskProgress != nullptr && taskProgress->isActive() == true)
    {
      taskProgress->cancel();
    }
  }
  catch (...)
  {
    operation->getReport()->addMessage(TE_TR("Error processing tile: ") + tileMessage + TE_TR(". Error: Unknown"), "ERROR", true);
    if (taskProgress != nullptr && taskProgress->isActive() == true)
    {
      taskProgress->cancel();
    }
  }

  std::chrono::time_point<std::chrono::steady_clock> end = std::chrono::steady_clock::now();

  static std::map<std::string, long long> mapTimeCount;
  mapTimeCount[tileMessage] = std::chrono::duration_cast<std::chrono::milliseconds>(end - start).count();
}

void te::vp::OperationThreadManager::start( const bool useThreads )
{
  startImpl(nullptr, useThreads );
}

void te::vp::OperationThreadManager::startAsync(te::vp::OperationThreadManagerListener* listener)
{
  m_thread = new std::thread(&te::vp::OperationThreadManager::startImpl, this, listener, true);
}

void te::vp::OperationThreadManager::stop()
{
  m_operation->getReport()->addMessage(TE_TR("Cancell Requested"), "INFO", true);
  m_operation->getParameters()->setOperationResult(te::vp::OperationResult::OP_CANCELLED);
  m_operation->abort();

  stopThread();
}

void te::vp::OperationThreadManager::setTask(te::common::TaskProgress* task)
{
  m_task = task;
}

te::vp::AbstractOperation* te::vp::OperationThreadManager::getOperation() const
{
  return m_operation.get();
}

void te::vp::OperationThreadManager::startImpl(te::vp::OperationThreadManagerListener* listener, const bool useThreads )
{
  m_operation->getReport()->addStartMessage(TE_TR("Operation: Started"), 0);

  m_operation->getReport()->addStartMessage(TE_TR("Initialization Step: Started"), 1);

  try
  {
    if (m_operation->prepareInputData() == false)
    {
      throw te::common::Exception(TE_TR("Error preparing the input data"));
    }

    if (m_operation->prepareOutputData() == false)
    {
      throw te::common::Exception(TE_TR("Error preparing the output data"));
    }
  }
  catch (const std::exception & e)
  {
    m_operation->getReport()->addEndMessage(TE_TR("Operation: Aborted due error in the execution: ") + std::string(e.what()), 0);
    
    if (listener != nullptr)
      listener->processingFinished(te::vp::OperationResult::OP_ERROR);
    return;
  }

  if (listener != nullptr)
  {
    listener->processingStateChanged(te::vp::ProgressState::PRE_PROCESSING_STARTED);
  }

  //and then we initialize the dataHandler
  std::size_t maxCoordinates = m_operation->getParameters()->getMaxCoordinates();
  if (m_operation->getCapabilities()->getAllowCoordinateSubdivision() == false)
  {
    m_operation->getParameters()->setMaxCoordinates(0);

  }
  m_dataHandler->initialize(m_operation->getParameters(), m_task);

  //temporary. We must set the changed output to the operation
  m_operation->getParameters()->setInputDataAccessVector(m_dataHandler->getInputDataAccessVector());
  m_operation->getParameters()->setOutputDataAccessVector(m_dataHandler->getOutputDataAccessVector());

  //we first initialize the operation to ensure that the output dataAccess will be valid from now on  
  m_operation->initialize();

  if (listener != nullptr)
  {
    listener->processingStateChanged(te::vp::ProgressState::PRE_PROCESSING_FINISHED);
  }

  m_operation->getReport()->addEndMessage(TE_TR("Initialization Step: Finished"), 1);

  if (listener != nullptr)
  {
    listener->processingStateChanged(te::vp::ProgressState::PROCESSING_STARTED);
  }

  m_operation->getReport()->addStartMessage(TE_TR("Processing Step: Started"), 1);

  //then we initialize the task progress
  if (m_task != nullptr)
  {
    m_task->setTotalSteps(static_cast<int>(m_dataHandler->getGroupCount()));
    m_task->setCurrentStep(0);
    m_task->useMultiThread(useThreads);
  }

  {
    m_operation->getReport()->addMessage(TE_TR("Number of groups: ") + te::common::Convert2String(static_cast<boost::uint32_t>(m_dataHandler->getGroupCount())), "INFO", true);

    // we create a thread pool so we can use all the available cores
    te::common::ThreadGroup threadGroup;
    for (std::size_t groupIndex = 0; groupIndex < m_dataHandler->getGroupCount(); ++groupIndex)
    {
      if( useThreads )
      {
        //adds the jobs to the manager
        threadGroup.addJob(std::bind(ExecuteOperation, m_operation.get(), m_dataHandler.get(), groupIndex, m_task));
      }
      else
      {
        //no thread
        ExecuteOperation(m_operation.get(), m_dataHandler.get(), groupIndex, m_task );
      }
    }

    //wait for the jobs to be finished
    if( useThreads ) threadGroup.waitToFinish();
  }

  m_operation->getReport()->addEndMessage(TE_TR("Processing Step: Finished"), 1);
  if (listener != nullptr)
  {
    listener->processingStateChanged(te::vp::ProgressState::PROCESSING_FINISHED);
  }

  //if the operation was aborted, we must check if there was an error or if it was cancelled by the user.
  //in the case of error, we do not generate the result
  //in the case it was aborted, we must finish the execution
  if (m_operation->isAborted() == true)
  {
    if (m_operation->getParameters()->getOperationResult() == te::vp::OperationResult::OP_ERROR)
    {
      if (listener != nullptr)
      {
        listener->processingFinished(m_operation->getParameters()->getOperationResult());
      }
      
      m_operation->getReport()->addEndMessage(TE_TR("Operation: Aborted due error in the execution"), 0);
      return;
    }
  }

  if (listener != nullptr)
  {
    listener->processingStateChanged(te::vp::ProgressState::POST_PROCESSING_STARTED);
  }

  m_operation->getReport()->addStartMessage(TE_TR("Finalization Step: Started"), 1);

  try
  {
    m_dataHandler->finalize(m_task);
  }
  catch (const std::exception&)
  {
    m_operation->getReport()->addEndMessage(TE_TR("Operation: Aborted due error in the execution"), 0);
    listener->processingFinished(te::vp::OperationResult::OP_ERROR);
    return;
  }

  m_operation->getReport()->addEndMessage(TE_TR("Finalization Step: Finished"), 1);

  if (listener != nullptr)
  {
    listener->processingStateChanged(te::vp::ProgressState::POST_PROCESSING_FINISHED);
  }

  m_operation->getReport()->addEndMessage(TE_TR("Operation: Finished"), 0);

  if (m_thread != nullptr)
  {
    m_thread->detach();
  }

  if (listener != nullptr)
  {
    listener->processingFinished(m_operation->getParameters()->getOperationResult());
  }
}

void te::vp::OperationThreadManager::stopThread()
{
  if (m_thread == nullptr)
  {
    return;
  }

  try
  {
    if (m_thread->joinable() == true)
    {
      m_thread->join();
    }

    delete m_thread;
    m_thread = nullptr;
  }
  catch (...)
  {

  }
}
