/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
 \file OverlapFixer.cpp
 */

#include "OverlapFixer.h"

#include "CommonFunctions.h"
#include "GeometrySnapper.h"
#include "RTreeIndex.h"
#include "../common/STLUtils.h"
#include "../common/ThreadGroup.h"
#include "../geometry/CommonDataStructures.h"
#include "../geometry/Geometry.h"
#include "../geometry/GeometryFactory.h"
#include "../geometry/LineString.h"
#include "../geometry/MultiPolygon.h"
#include "../geometry/Polygon.h"
#include "../geometry/Utils.h"
#include "../sam/rtree/Index.h"

//STL include files
#include <memory>
#include <thread>

class GeometryInfoOverlap : public te::vp::GeometryInfo
{

  public:
    GeometryInfoOverlap(std::size_t geometryIndex, te::gm::Geometry* geometry, const te::gm::Envelope& filter)
      : GeometryInfo(geometryIndex, geometry, filter)
      , m_fixResult(te::vp::OverlapFixerReport::FixResult::UNCHANGED)
    {

    }

    GeometryInfoOverlap(std::size_t geometryIndex, te::gm::Geometry* geometry)
      : GeometryInfo(geometryIndex, geometry, te::gm::Envelope())
      , m_fixResult(te::vp::OverlapFixerReport::FixResult::UNCHANGED)
    {

    }

public:

    te::vp::OverlapFixerReport::FixResult m_fixResult;
};

struct SingleFixReport
{
  SingleFixReport()
  {

  }

  ~SingleFixReport()
  {
  }

  std::unique_ptr<te::gm::Geometry> m_newGeometry;
};

struct OverlapReport
{
  OverlapReport()
  {

  }

  ~OverlapReport()
  {
  }

  bool neededFix() const
  {
    return (m_vecFragments.empty() == false);
  }

  te::gm::GeometryVector m_vecFragments;
};

//OverlapFixerReport
te::vp::OverlapFixerReport::OverlapFixerReport()
  : m_geometry(nullptr)
  , m_result(FixResult::UNCHANGED)
{

}

te::vp::OverlapFixerReport::OverlapFixerReport(te::gm::Geometry* geometry, FixResult result)
  : m_geometry(geometry)
  , m_result(result)
{

}

void CalculateIntersectionThread(std::size_t index, const std::vector<GeometryInfoOverlap*>& vecGeometries, const te::sam::rtree::Index<std::size_t>& rtree, OverlapReport* report)
{
  GeometryInfoOverlap* inputGeometryInfo = vecGeometries.at(index);
  const te::gm::Geometry* inputGeometry = inputGeometryInfo->getGeometry();

  const te::gm::Envelope* mbr = inputGeometry->getMBR();

  std::vector<std::size_t> vecCandidatesRawIndexes;
  rtree.search(*mbr, vecCandidatesRawIndexes);

  //a geometry containing all the intersection fragments
  te::gm::GeometryVector vecFragments;

  //we first prepare the threads to execute the analysis and fix of each candidate
  std::vector<std::size_t> vecCandidatesNetIndexes;
  te::gm::GeometryVector vecCandidates;
  for (std::size_t j = 0; j < vecCandidatesRawIndexes.size(); ++j)
  {
    std::size_t candidateIndex = vecCandidatesRawIndexes.at(j);
    if (candidateIndex <= index)
    {
      //here we discard the current geoemtry AND all other geometries that have already been processed
      //as we fix both geometries in each i-iteraction, we must not reprocess the already processed indexes
      continue;
    }

    GeometryInfoOverlap* candidateGeometryInfo = vecGeometries.at(candidateIndex);
    const te::gm::Geometry* candidateGeometry = candidateGeometryInfo->getGeometry();

    te::gm::SpatialRelation relation = te::vp::FastRelationCalculator(inputGeometryInfo, candidateGeometryInfo);
    if (relation != te::gm::SpatialRelation::OVERLAPS)
    {
      //if we get here, we can be sure that there is no overlap
      continue;
    }

    std::unique_ptr<te::gm::Geometry> fragmentIntersection(inputGeometry->intersection(candidateGeometry));
    if (fragmentIntersection == nullptr || fragmentIntersection->isEmpty() == true)
    {
      continue;
    }

    //then we filter them because we just need polygons
    fragmentIntersection.reset(NormalizeGeometryByType(fragmentIntersection.get(), te::gm::PolygonType));
    if (fragmentIntersection == nullptr || fragmentIntersection->isEmpty() == true)
    {
      continue;
    }

    te::gm::Multi2Single(fragmentIntersection.get(), vecFragments);
  }

  if (vecFragments.empty() == true)
  {
    return;
  }

  report->m_vecFragments = vecFragments;
}

void FixOverlapThread(const te::gm::Geometry* geometry, const te::gm::GeometryVector& vecFragments, SingleFixReport* report)
{
  //now we make a union of the fragments
  std::unique_ptr<te::gm::Geometry> fragmentsUnion(te::gm::CascadedUnion(vecFragments));
  fragmentsUnion.reset(NormalizeGeometryByType(fragmentsUnion.get(), te::gm::PolygonType));

  //and fix the input geometry
  std::unique_ptr<te::gm::Geometry> inputGeometryFixed(geometry->difference(fragmentsUnion.get()));
  inputGeometryFixed.reset(NormalizeGeometryByType(inputGeometryFixed.get(), te::gm::PolygonType));

  report->m_newGeometry.reset(inputGeometryFixed.release());
}

te::gm::GeometryVector te::vp::OverlapFixer::detectOverlaps(const std::vector<GeometryInfoOverlap*>& vecGeometries)
{
  //we first create an index of the geometries
  te::sam::rtree::Index<std::size_t> rtree;
  for (std::size_t i = 0; i < vecGeometries.size(); ++i)
  {
    const te::gm::Geometry* geometry = vecGeometries.at(i)->getGeometry();
    const te::gm::Envelope* mbr = geometry->getMBR();

    rtree.insert(*mbr, i);
  }

  // we create a thread pool so we can use all the available cores
  te::common::ThreadGroup threadGroup;

  //then we detect the overlaps for each geometry
  //for performance purposes, we want to first calculate the intersection fragments
  //the current geometry will be fixed later
  std::vector<OverlapReport*> vecOverlapReports;
  vecOverlapReports.reserve(vecGeometries.size());
  for (std::size_t i = 0; i < vecGeometries.size(); ++i)
  {
    //creates a thread to execute the intersection calculation
    OverlapReport* overlapReport = new OverlapReport();

    //adds the jobs to the manager
    threadGroup.addJob(boost::bind(CalculateIntersectionThread, i, std::ref(vecGeometries), std::ref(rtree), overlapReport));

    vecOverlapReports.push_back(overlapReport);
  }

  //wait for the jobs to be finished
  threadGroup.waitToFinish();

  te::gm::GeometryVector vecAllFragments;
  for (std::size_t index = 0; index < vecOverlapReports.size(); ++index)
  {
    //then we store the intersection fragments
    OverlapReport* overlapReport = vecOverlapReports.at(index);
    if (overlapReport->neededFix() == false)
    {
      continue;
    }

    //we create a vector containing all the detected fragments
    vecAllFragments.insert(vecAllFragments.end(), overlapReport->m_vecFragments.begin(), overlapReport->m_vecFragments.end());
  }

  te::gm::GeometryVector vecAllFragmentsFixed = fixFragmentsOverlaps(vecAllFragments);

  te::common::FreeContents(vecOverlapReports);
  te::common::FreeContents(vecAllFragments);

  return vecAllFragmentsFixed;
}

void te::vp::OverlapFixer::fixOverlaps(std::vector<GeometryInfoOverlap*>& vecGeometries, const te::gm::GeometryVector& vecFragments)
{
  //transform the fragments into single geometries
  te::vp::RTreeIndex<te::gm::Geometry*> rtreeIndex;
  for (std::size_t i = 0; i < vecFragments.size(); ++i)
  {
    te::gm::Geometry* geometry = vecFragments.at(i);
    const te::gm::Envelope* mbr = geometry->getMBR();
    rtreeIndex.insert(*mbr, geometry);
  }

  te::common::ThreadGroup threadGroup;

  //then we fix all overlaps
  std::vector<SingleFixReport*> vecFixReports;
  std::vector<std::size_t> vecIndexes;
  vecFixReports.reserve(vecGeometries.size());

  for (std::size_t index = 0; index < vecGeometries.size(); ++index)
  {
    //then we store the intersection fragments
    GeometryInfoOverlap* geometryInfo = vecGeometries.at(index);
    const te::gm::Geometry* geometry = geometryInfo->getGeometry();
    const te::gm::Envelope* mbr = geometry->getMBR();
    
    std::vector<te::gm::Geometry*> vecCandidates = rtreeIndex.search(*mbr)->getAllData();
    if (vecCandidates.empty() == true)
    {
      continue;
    }

    te::gm::GeometryVector geometryVectorCandidates;
    geometryVectorCandidates.insert(geometryVectorCandidates.begin(), vecCandidates.begin(), vecCandidates.end());

    //creates a thread to execute the intersection calculation
    SingleFixReport* fixReport = new SingleFixReport();

    //adds the jobs to the manager
    threadGroup.addJob(boost::bind(FixOverlapThread, geometry, geometryVectorCandidates, fixReport));

    vecIndexes.push_back(index);
    vecFixReports.push_back(fixReport);
  }

  //wait for the jobs to be finished
  threadGroup.waitToFinish();

  for (std::size_t i = 0; i < vecFixReports.size(); ++i)
  {
    std::size_t geometryInfoIndex = vecIndexes.at(i);
    SingleFixReport* fixReport = vecFixReports.at(i);

    GeometryInfoOverlap* geometryInfo = vecGeometries.at(geometryInfoIndex);
    geometryInfo->setGeometry(fixReport->m_newGeometry.release());
    if (geometryInfo->m_fixResult == te::vp::OverlapFixerReport::FixResult::UNCHANGED)
    {
      geometryInfo->m_fixResult = te::vp::OverlapFixerReport::FixResult::FIXED;
      if (geometryInfo->getGeometry() == nullptr)
      {
        geometryInfo->m_fixResult = te::vp::OverlapFixerReport::FixResult::REMOVED;
      }
    }
  }

  te::common::FreeContents(vecFixReports);
}

te::gm::GeometryVector te::vp::OverlapFixer::fixOverlaps(std::vector<GeometryInfoOverlap*>& vecGeometries)
{
  //we first create an index of the geometries
  te::gm::GeometryVector vecFragments(detectOverlaps(vecGeometries));
  if (vecFragments.empty() == true)
  {
    return vecFragments;
  }

  //now we fix the overlaps
  fixOverlaps(vecGeometries, vecFragments);

  return vecFragments;
}

te::gm::GeometryVector te::vp::OverlapFixer::fixFragmentsOverlaps(const te::gm::GeometryVector& vecFragments)
{
  te::gm::GeometryVector vecFragmentsWithoutOverlaps;

  te::vp::RTreeIndex<te::gm::Geometry*> indexToFix;
  for (std::size_t i = 0; i < vecFragments.size(); ++i)
  {
    te::gm::Geometry* geometry = vecFragments.at(i);

    indexToFix.insert(*geometry->getMBR(), geometry);
  }

  bool maybeStillHaveOverlaps = false;
  for (std::size_t currentIndex = 0; currentIndex < vecFragments.size(); ++currentIndex)
  {
    te::gm::Geometry* geometry = vecFragments.at(currentIndex);

    //for fixing, we must add the candidates that have not been processed and also the ones that have already been processed
    te::gm::GeometryVector vecCandidates;
    std::unique_ptr<te::vp::IndexReport<te::gm::Geometry*>> indexReport = indexToFix.search(*geometry->getMBR());

    std::unique_ptr<te::gm::Geometry> geometryFixed(static_cast<te::gm::Geometry*>(geometry->clone()));
    for (std::size_t j = 0; j < indexReport->size(); ++j)
    {
      //discard the current geometry
      te::vp::IndexData<te::gm::Geometry*>* indexData = indexReport->getData(j);
      if (indexData->m_dataIndex == currentIndex)
      {
        continue;
      }

      te::gm::SpatialRelation relation = te::vp::FastRelationCalculator(geometryFixed.get(), indexData->m_data);
      if (relation == te::gm::SpatialRelation::OVERLAPS)
      {
        geometryFixed.reset(geometryFixed->difference(indexData->m_data));
        maybeStillHaveOverlaps = true;
      }
    }

    if (geometryFixed != nullptr)
    {
      geometryFixed.reset(te::gm::NormalizeGeometryByType(geometryFixed.get(), te::gm::PolygonType));
      if (geometryFixed != nullptr)
      {
        te::gm::Multi2Single(geometryFixed.get(), vecFragmentsWithoutOverlaps);
      }
    }
  }

  if (maybeStillHaveOverlaps == true)
  {
    //after fixing all the fragments without intersection, we fix the one with intersection
    std::unique_ptr<te::gm::Geometry> originalFragmentsUnion(te::gm::CascadedUnion(vecFragments));
    std::unique_ptr<te::gm::Geometry> fragmentsWithoutOverlapUnion(te::gm::CascadedUnion(vecFragmentsWithoutOverlaps));

    std::unique_ptr<te::gm::Geometry> newFragmentsWithoutOverlap(originalFragmentsUnion->difference(fragmentsWithoutOverlapUnion.get()));
    if (newFragmentsWithoutOverlap != nullptr)
    {
      newFragmentsWithoutOverlap.reset(te::gm::NormalizeGeometryByType(newFragmentsWithoutOverlap.get(), te::gm::PolygonType));
      if (newFragmentsWithoutOverlap != nullptr)
      {
        te::gm::Multi2Single(newFragmentsWithoutOverlap.get(), vecFragmentsWithoutOverlaps);
      }
    }
  }

  return vecFragmentsWithoutOverlaps;
}

//OverlapFixer
std::vector<te::vp::OverlapFixerReport*> te::vp::OverlapFixer::fix(const te::gm::GeometryVector& vecGeometries)
{
  //then we must fix the bigger overlaps
  //creates the geometry spatial index
  te::sam::rtree::Index<std::size_t> rtree;
  std::vector<GeometryInfoOverlap*> vecGeometryInfo;
  vecGeometryInfo.reserve(vecGeometries.size());
  for (std::size_t i = 0; i < vecGeometries.size(); ++i)
  {
    te::gm::Geometry* geometry = vecGeometries.at(i);
    const te::gm::Envelope* mbr = geometry->getMBR();

    GeometryInfoOverlap* geometryInfo = new GeometryInfoOverlap(i, static_cast<te::gm::Geometry*>(geometry->clone()));

    rtree.insert(*mbr, i);
    vecGeometryInfo.push_back(geometryInfo);
  }

  //then, we the intersection fragments until there are no more intersections
  te::gm::GeometryVector vecFragments = fixOverlaps(vecGeometryInfo);

  //after fixing everything, we prepare the output
  std::vector<te::vp::OverlapFixerReport*> vecFixReport;

  //we begin handling the fixed geometries
  for (std::size_t i = 0; i < vecGeometryInfo.size(); ++i)
  {
    GeometryInfoOverlap* geometryInfo = vecGeometryInfo.at(i);
    if (geometryInfo->m_fixResult == te::vp::OverlapFixerReport::FixResult::REMOVED)
    {
      continue;
    }

    te::vp::OverlapFixerReport* report = new te::vp::OverlapFixerReport();
    report->m_geometry.reset(static_cast<te::gm::Geometry*>(geometryInfo->getGeometry()->clone()));
    report->m_result = geometryInfo->m_fixResult;

    vecFixReport.push_back(report);
  }

  te::common::FreeContents(vecGeometryInfo);

  //the we handle the created fragments. We must discard the fragments that became null during the process
  for (std::size_t i = 0; i < vecFragments.size(); ++i)
  {
    te::gm::Geometry* fragment = vecFragments.at(i);
    if (fragment == nullptr)
    {
      continue;
    }

    te::vp::OverlapFixerReport* report = new te::vp::OverlapFixerReport();
    report->m_geometry.reset(fragment);
    report->m_result = te::vp::OverlapFixerReport::FixResult::NEW;

    vecFixReport.push_back(report);
  }

  return vecFixReport;
}

te::gm::GeometryVector te::vp::OverlapFixer::detectOverlaps(const te::gm::GeometryVector& vecGeometries)
{
  //then we must fix the bigger overlaps
  //creates the geometry spatial index
  te::sam::rtree::Index<std::size_t> rtree;
  std::vector<GeometryInfoOverlap*> vecGeometryInfo;
  vecGeometryInfo.reserve(vecGeometries.size());
  for (std::size_t i = 0; i < vecGeometries.size(); ++i)
  {
    te::gm::Geometry* geometry = vecGeometries.at(i);
    const te::gm::Envelope* mbr = geometry->getMBR();

    GeometryInfoOverlap* geometryInfo = new GeometryInfoOverlap(i, static_cast<te::gm::Geometry*>(geometry->clone()));

    rtree.insert(*mbr, i);
    vecGeometryInfo.push_back(geometryInfo);
  }

  //then, we the intersection fragments until there are no more intersections
  te::gm::GeometryVector vecFragments = detectOverlaps(vecGeometryInfo);

  te::common::FreeContents(vecGeometryInfo);

  return vecFragments;
}

bool te::vp::OverlapFixer::hasOverlaps(const te::gm::Geometry* geometryA, const te::gm::Geometry* geometryB)
{
  const te::gm::Envelope* envelopeA = geometryA->getMBR();
  const te::gm::Envelope* envelopeB = geometryB->getMBR();

  if (envelopeA->intersects(*envelopeB) == false)
  {
    //no overlap
    return false;
  }
  if (envelopeA->touches(*envelopeB) == true)
  {
    //no overlap
    return false;
  }

  std::string actualRelation = geometryA->relate(geometryB);

  //check if the relation is disjoint
  std::string disjointRelation = "FF*FF****";
  if (te::gm::RelationMatches(actualRelation, disjointRelation) == true)
  {
    //no overlap
    return false;
  }

  //check if the relation is touches
  std::string touchesRelation1 = "FT*******";
  if (te::gm::RelationMatches(actualRelation, touchesRelation1) == true)
  {
    //no overlap
    return false;
  }

  std::string touchesRelation2 = "F**T*****";
  if (te::gm::RelationMatches(actualRelation, touchesRelation2) == true)
  {
    //no overlap
    return false;
  }

  std::string touchesRelation3 = "F***T****";
  if (te::gm::RelationMatches(actualRelation, touchesRelation3) == true)
  {
    //no overlap
    return false;
  }

  return true;
}
