/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
 \file OverlapFixer.h
 
 \brief Algorithm to detect and fix overlaps in a list of polygons
 
 \ingroup vp
 */

#ifndef __TERRALIB_VP_INTERNAL_OVERLAPFIXER_H
#define __TERRALIB_VP_INTERNAL_OVERLAPFIXER_H

// Terralib
#include "Config.h"

#include "../geometry/CommonDataStructures.h"
#include "../geometry/Geometry.h"

#include <memory>
#include <vector>

class GeometryInfoOverlap;
struct OverlapReport;
struct SingleFixReport;

namespace te
{
  namespace vp
  {
    /*!
      \class OverlapFixerReport

      \brief OverlapFixerReport represents the report of the fix operation, for each input geometry and for the generated intersection fragments.
    */
    class TEVPEXPORT OverlapFixerReport
    {
      public:
        enum class FixResult
        {
          UNCHANGED, FIXED, REMOVED, NEW
        };

      public:
        //!< Constructor
        OverlapFixerReport();

        //!< Constructor
        OverlapFixerReport(te::gm::Geometry* geometry, FixResult result);

        //!< Destructor
        ~OverlapFixerReport() = default;

        std::unique_ptr<te::gm::Geometry> m_geometry; //!< The resulting geometry of the fixer operation
        FixResult m_result; //!< The result of the fixer operation
    };

    /*!
      \class OverlapFixer

      \brief OverlapFixerReport is a class containing algorithms to detect and fix overlaps from a list of polygons or multipolygons
    */
    class TEVPEXPORT OverlapFixer
    {
      public:

        /*!
          \brief Fix the existing overlaps of the polygons ou multipolygons from the given vecGeometries vector

          \param vecGeometries The input geometries to be analysed
          
          \return A vector containing the fixed geometries
        */
        static std::vector<OverlapFixerReport*> fix(const te::gm::GeometryVector& vecGeometries);

        /*!
          \brief Detects all the existing overlaps from the given geometries

          \param vecGeometries The input geometries to be analysed

          \return A vector containing the intersection fragments
        */
        static te::gm::GeometryVector detectOverlaps(const te::gm::GeometryVector& vecGeometries);

        /*!
          \brief Algorithm to detect overlaps. This algorith does a deep and exausting geometry analysis and doesnt generate false positives

          \param geometryA The input geometry A
          \param geometryB The input geometry B

          \return TRUE if there is at least one overlap between the given geometries. FALSE otherwise
        */
        static bool hasOverlaps(const te::gm::Geometry* geometryA, const te::gm::Geometry* geometryB);

      protected:

        /*!
          \brief Internal method that detects all the existing overlaps from the given geometries

          \param vecGeometries The input geometries to be analysed

          \return A vector containing the intersection fragments
        */
        static te::gm::GeometryVector detectOverlaps(const std::vector<GeometryInfoOverlap*>& vecGeometries);

        /*!
          \brief Internal method to fix the existing overlaps of the polygons ou multipolygons from the given vecGeometries vector considering the given intersection fragments

          \param vecGeometries The input geometries to be fixed
          \param vecFragments The intersection fragments to be used to fix the geometries
        */
        static void fixOverlaps(std::vector<GeometryInfoOverlap*>& vecGeometries, const te::gm::GeometryVector& vecFragments);

        /*!
          \brief Internal method to fix the existing overlaps of the polygons ou multipolygons from the given vecGeometries vector

          \param vecGeometries The input geometries to be analysed

          \return A vector containing the fixed geometries
        */
        static te::gm::GeometryVector fixOverlaps(std::vector<GeometryInfoOverlap*>& vecGeometries);

        /*!
          \brief Internal method to fix the possible overlaps within the intersection fragments themselves

          \param vecFragments A vector containing all detected intersection fragments

          \return A vector containing the intersection fragments without any overlaps between the geometries
        */
        static te::gm::GeometryVector fixFragmentsOverlaps(const te::gm::GeometryVector& vecFragments);
    };
  }
}
#endif // __TERRALIB_VP_INTERNAL_OVERLAPFIXER_H
