/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

#include "PolygonSubdivider.h"

#include "../common/STLUtils.h"
#include "../common/ThreadGroup.h"
#include "../geometry/algorithms/SafeOverlay.h"
#include "../geometry/CommonDataStructures.h"
#include "../geometry/Geometry.h"
#include "../geometry/GeometryCollection.h"
#include "../geometry/GeometryFunctions.h"
#include "../geometry/LineString.h"
#include "../geometry/MultiPolygon.h"
#include "../geometry/Polygon.h"
#include "../geometry/Utils.h"
#include "../geometry/Validation.h"
#include "../geometry/validation/FixNoding.h"
#include "CommonDataStructures.h"
#include "CommonFunctions.h"
#include "GeometrySnapper.h"
#include "OperationService.h"
#include "SubdividerTilling.h"
#include "Utils.h"

 //Boost
#include <boost/bimap.hpp>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/connected_components.hpp>

te::gm::GeometryVector te::vp::PolygonSubdivider::subdivide(const te::gm::Geometry* geometry, std::size_t maxCoordinates)
{
  const te::gm::Envelope* envelope = geometry->getMBR();
  return subdivide(geometry, *envelope, maxCoordinates);
}

te::gm::GeometryVector te::vp::PolygonSubdivider::subdivide(const te::gm::Geometry* geometry, const te::gm::Envelope& envelope, std::size_t maxCoordinates)
{
  te::gm::GeometryVector vecResult;
  te::gm::GeometryPtr result;

  bool tryAgain = false;
  
  if (geometry->getMBR()->within(envelope) == true)
  {
    result.reset(static_cast<te::gm::Geometry*>(geometry->clone()));
  }
  else
  {
    try
    {
      result.reset(te::gm::SafeOverlay::ClipGeometry(geometry, envelope));
    }
    catch (const std::exception&)
    {
      throw;
    }
  }
  
  if (te::gm::IsNullOrEmpty(result) == false)
  {
    te::gm::GeometryVector vecSingleResult;
    te::gm::Multi2Single(result.get(), vecSingleResult);

    for (std::size_t i = 0; i < vecSingleResult.size(); ++i)
    {
      te::gm::GeometryPtr singleGeometry(vecSingleResult.at(i));
      if (maxCoordinates == 0 || singleGeometry->getNPoints() < maxCoordinates)
      {
        if (singleGeometry->getDimension() == te::gm::A)
        {
          //we must ensure that all geometries are valid
          //in some cases, this fast algorithm can lead to self-intersectiong geometries.
          if (singleGeometry->isValid() == false)
          {
            singleGeometry.reset(te::gm::Validation::MakeValidWithSnap(singleGeometry.get()).release());
          }

          if (te::gm::IsNullOrEmpty(singleGeometry) == false)
          {
            te::gm::Multi2Single(singleGeometry.get(), vecResult);
          }
        }
        else
        {
          vecResult.push_back(singleGeometry.release());
        }
        continue;
      }

      double precision = te::gm::GetAppropriatePrecision(geometry->getSRID());

      std::size_t splitCount = 2;

      te::gm::Envelope adjustedEnvelope(envelope);
      adjustedEnvelope.m_llx -= precision;
      adjustedEnvelope.m_lly -= precision;
      adjustedEnvelope.m_urx += precision;
      adjustedEnvelope.m_urx += precision;

      double width = adjustedEnvelope.getWidth();
      double height = adjustedEnvelope.getHeight();
      double widthStep = width / splitCount;
      double heightStep = height / splitCount;

      for (std::size_t row = 0; row < splitCount; ++row)
      {
        for (std::size_t column = 0; column < splitCount; ++column)
        {
          double x1 = adjustedEnvelope.m_llx + (widthStep * column);
          double y1 = adjustedEnvelope.m_lly + (heightStep * row);
          double x2 = adjustedEnvelope.m_llx + (widthStep * (column + 1));
          double y2 = adjustedEnvelope.m_lly + (heightStep * (row + 1));

          te::gm::Envelope subEnvelope(x1, y1, x2, y2);
          te::gm::GeometryVector vecTempSingleResult = subdivide(singleGeometry.get(), subEnvelope, maxCoordinates);
          vecResult.insert(vecResult.end(), vecTempSingleResult.begin(), vecTempSingleResult.end());
        }
      }
    }
  }

  return vecResult;
}

te::gm::GeometryVector te::vp::PolygonSubdivider::subdivide(const te::gm::GeometryVector& vecInputGeometries, std::size_t maxCoordinates, std::vector<std::size_t>& vecParentIndexes)
{
  //then we divide the geometries based on each tile
  te::gm::GeometryVector vecResult;
  for (std::size_t i = 0; i < vecInputGeometries.size(); ++i)
  {
    const te::gm::Geometry* geometry = vecInputGeometries.at(i);
    te::gm::GeometryVector currentResult = subdivide(geometry, *geometry->getMBR(), maxCoordinates);

    for (std::size_t j = 0; j < currentResult.size(); ++j)
    {
      vecResult.push_back(currentResult.at(j));
      vecParentIndexes.push_back(i);
    }
  }

  return vecResult;
}

te::gm::GeometryVector te::vp::PolygonSubdivider::subdivide(const te::gm::GeometryVector& vecInputGeometries, const te::gm::Envelope& envelope, std::size_t maxCoordinates)
{
  //then we divide the geometries based on each tile
  te::gm::GeometryVector vecResult;
  for (std::size_t i = 0; i < vecInputGeometries.size(); ++i)
  {
    const te::gm::Geometry* geometry = vecInputGeometries.at(i);
    te::gm::GeometryVector currentResult = subdivide(geometry, envelope, maxCoordinates);

    vecResult.insert(vecResult.end(), currentResult.begin(), currentResult.end());
  }

  return vecResult;
}

te::gm::GeometryVector te::vp::PolygonSubdivider::subdivide(const te::gm::Geometry* geometry, const SubdividerTilling& tilling, std::size_t maxCoordinates)
{
  te::gm::GeomType expectedGeometryType = geometry->get2DGeomTypeId();
  const te::gm::Envelope* mbr = geometry->getMBR();

  std::size_t initRow = tilling.getTileRow(mbr->getLowerLeft());
  std::size_t initColumn = tilling.getTileColumn(mbr->getLowerLeft());

  std::size_t finalRow = tilling.getTileRow(mbr->getUpperRight());
  std::size_t finalColumn = tilling.getTileColumn(mbr->getUpperRight());

  te::gm::GeometryVector vecFragments;
  for (std::size_t row = initRow; row <= finalRow; ++row)
  {
    for (std::size_t column = initColumn; column <= finalColumn; ++column)
    {
      te::gm::Envelope clipArea = tilling.getTile(column, row);
      te::gm::GeometryVector vecClipped = te::vp::PolygonSubdivider::subdivide(geometry, clipArea, maxCoordinates);

      for (std::size_t f = 0; f < vecClipped.size(); ++f)
      {
        te::gm::GeometryVector vecAdjusted = te::vp::AdjustToGeomType(vecClipped.at(f), expectedGeometryType);
        vecFragments.insert(vecFragments.end(), vecAdjusted.begin(), vecAdjusted.end());
      }

      te::common::FreeContents(vecClipped);
    }
  }

  return vecFragments;
}

te::gm::GeometryVector te::vp::PolygonSubdivider::subdivide(const te::gm::GeometryVector& vecInputGeometries, const SubdividerTilling& tilling, std::size_t maxCoordinates)
{
  //then we divide the geometries based on each tile
  te::gm::GeometryVector vecResult;
  for (std::size_t i = 0; i < vecInputGeometries.size(); ++i)
  {
    const te::gm::Geometry* geometry = vecInputGeometries.at(i);
    te::gm::GeometryVector currentResult = subdivide(geometry, tilling, maxCoordinates);

    vecResult.insert(vecResult.end(), currentResult.begin(), currentResult.end());
  }

  return vecResult;
}

bool te::vp::PolygonSubdivider::filterSegmentsInBorder(const te::gm::Geometry* geometry, std::size_t geometryIndex, const te::gm::LineString* splitLine, std::vector<te::vp::SegmentInfo*>& vecSegmentsInBorder)
{
  const te::gm::Envelope* mbrSplitLine = splitLine->getMBR();
  if (geometry->getMBR()->intersects(*mbrSplitLine) == false)
  {
    //nothing to be done
    return false;
  }

  te::gm::Dimensionality dimensionality = geometry->getDimension();
  //if the geometry has dimension of 0, we do not analyse it
  if (dimensionality == te::gm::P)
  {
    return false;
  }

  //we now index all the segments that overlap the split line
  std::unique_ptr<GeometryInfo> geometryInfo(new GeometryInfo(geometryIndex, static_cast<te::gm::Geometry*>(geometry->clone()), *mbrSplitLine));
  const std::vector<te::vp::SegmentInfo>& vecSegments = geometryInfo->getAllSegments();

  bool atLeastOneSegmentInBorder = false;
  for (std::size_t j = 0; j < vecSegments.size(); ++j)
  {
    const te::vp::SegmentInfo& segmentInfo = vecSegments.at(j);
    const te::gm::Coord2D& coord1 = segmentInfo.getCoord1();
    const te::gm::Coord2D& coord2 = segmentInfo.getCoord2();

    //we check if the first coordinate intercepts the splitLine
    bool hasIntersection = false;
    te::gm::Envelope envelopeCoord1 = te::gm::CreateEnvelope(coord1, 0.);
    if (mbrSplitLine->intersects(envelopeCoord1) == false)
    {
      //if we have a polygon, both coordinates must be in the border.
      //So if the current coordinate is not in the border, we continue to the next candidate
      if (dimensionality == te::gm::A)
      {
        continue;
      }
    }
    else
    {
      hasIntersection = true;
    }

    //we check if the second coordinate intercepts the splitLine
    te::gm::Envelope envelopeCoord2 = te::gm::CreateEnvelope(coord2, 0.);
    if (mbrSplitLine->intersects(envelopeCoord2) == false)
    {
      //if we have a polygon, both coordinates must be in the border.
      //So if the current coordinate is not in the border, we continue to the next candidate
      if (dimensionality == te::gm::A)
      {
        continue;
      }
    }
    else 
    {
      hasIntersection = true;
    }

    if (hasIntersection == false)
    {
      continue;
    }

    //if both intercept, we add the segment to the index
    vecSegmentsInBorder.push_back(segmentInfo.clone());
    atLeastOneSegmentInBorder = true;
  }

  return atLeastOneSegmentInBorder;
}

std::vector<te::vp::SegmentInfo*> te::vp::PolygonSubdivider::filterSegmentsInBorder(const te::gm::GeometryVector& vecFragments, const te::gm::LineString* splitLine, std::vector<std::size_t>& vecIndexesNotInBorder)
{
  std::vector<te::vp::SegmentInfo*> vecSegmentOnBorder;

  //we must first index all the border segments that are overlapping the split line
  //we also discard all the geometries that do not touche the split line
  for (std::size_t i = 0; i < vecFragments.size(); ++i)
  {
    //if the geometry has dimension of 0, we do not analyse it
    if (vecFragments.at(i)->getDimension() == te::gm::P)
    {
      continue;
    }
    const te::gm::Geometry* geometry = vecFragments.at(i);
    bool hasOnSegmentOnBorder = filterSegmentsInBorder(geometry, i, splitLine, vecSegmentOnBorder);
    if (hasOnSegmentOnBorder == false)
    {
      vecIndexesNotInBorder.push_back(i);
    }
  }

  return vecSegmentOnBorder;
}

bool te::vp::PolygonSubdivider::checkForGeometrySubdivision(const te::gm::Envelope& currentTile, const te::gm::Geometry* geometry, std::size_t geometryIndex, std::vector<te::vp::SegmentInfo*>& vecSegmentsInBorder)
{
  bool found = false;

  //we check if we need to stop the evaluation. If the geometry box is not within the given envelope it means that the envelope is smalled than the possible current subdivision
  if (geometry->getMBR()->within(currentTile) == false)
  {
    return found;
  }

  te::gm::LineString middleVerticalBorder = te::vp::SubdividerTilling::getSplitLine(currentTile, te::vp::SubdividerTilling::SplitLinePosition::MIDDLE_VERTICAL);
  te::gm::LineString middleHorizontalBorder = te::vp::SubdividerTilling::getSplitLine(currentTile, te::vp::SubdividerTilling::SplitLinePosition::MIDDLE_HORIZONTAL);

  //now we get the middle lines from the current box
  found = found | te::vp::PolygonSubdivider::filterSegmentsInBorder(geometry, geometryIndex, &middleVerticalBorder, vecSegmentsInBorder);
  found = found | te::vp::PolygonSubdivider::filterSegmentsInBorder(geometry, geometryIndex, &middleHorizontalBorder, vecSegmentsInBorder);

  std::size_t splitCount = 2;

  double width = currentTile.getWidth();
  double height = currentTile.getHeight();
  double widthStep = width / splitCount;
  double heightStep = height / splitCount;

  for (std::size_t row = 0; row < splitCount; ++row)
  {
    for (std::size_t column = 0; column < splitCount; ++column)
    {
      double x1 = currentTile.m_llx + (widthStep * column);
      double x2 = currentTile.m_llx + (widthStep * (column + 1));
      double y1 = currentTile.m_lly + (heightStep * row);
      double y2 = currentTile.m_lly + (heightStep * (row + 1));

      te::gm::Envelope subEnvelope(x1, y1, x2, y2);
      found = found | checkForGeometrySubdivision(subEnvelope, geometry, geometryIndex, vecSegmentsInBorder);
    }
  }

  return found;
}

te::vp::PolygonSubdivider::GeometryAssociations te::vp::PolygonSubdivider::calculateGeometryAssociations(const te::gm::GeometryVector& vecFragments, const te::gm::LineString* splitLine)
{
  if (vecFragments.empty() == true)
  {
    te::vp::PolygonSubdivider::GeometryAssociations geometryAssociations;
    return geometryAssociations;
  }

  std::vector<std::size_t> vecIndexesNotInBorder;
  std::vector<te::vp::SegmentInfo*> vecBorderSegments = te::vp::PolygonSubdivider::filterSegmentsInBorder(vecFragments, splitLine, vecIndexesNotInBorder);

  if (vecBorderSegments.empty() == true)
  {
    te::vp::PolygonSubdivider::GeometryAssociations geometryAssociations;
    return geometryAssociations;
  }

  //after indexing all the border segments, we must find the correspondent segments from the other geometries
  te::gm::Dimensionality dimensionality = vecFragments.at(0)->getDimension();
  te::vp::PolygonSubdivider::GeometryAssociations geometryAssociations = calculateGeometryAssociations(vecBorderSegments, dimensionality);

  te::common::FreeContents(vecBorderSegments);

  return geometryAssociations;
}

struct DissolveResult
{
  DissolveResult(std::size_t seedIndex)
    : m_seedIndex(seedIndex)
  {

  }

  ~DissolveResult()
  {
    te::common::FreeContents(m_geometryVector);
  }

  std::size_t m_seedIndex;
  te::gm::GeometryVector m_geometryVector;
  std::unique_ptr<te::gm::Geometry> m_outputGeometry;
};

void DissolveThread(DissolveResult* dissolveResult)
{
  std::unique_ptr<te::gm::Geometry> geometryDissolved(te::gm::CascadedUnion(dissolveResult->m_geometryVector));
  dissolveResult->m_outputGeometry = std::move(geometryDissolved);
}

te::gm::GeometryVector te::vp::PolygonSubdivider::dissolveFragments(const te::gm::GeometryVector& vecFragments, const Clusters& clusters)
{
  te::gm::GeometryVector vecResult;
  vecResult.reserve(clusters.size());

  std::vector<DissolveResult*> vecDissolveResult;
  vecDissolveResult.reserve(clusters.size());

  // we create a thread pool so we can use all the available cores
  te::common::ThreadGroup threadGroup;

  std::set<std::size_t> setUsed;
  for(std::size_t i = 0; i < clusters.size(); ++i)
  {
    const ClusterIndexes& clusterIndexes = clusters.at(i);

    //we first create a vector containing all the geometries in a group that must be dissolved
    te::gm::GeometryVector vecAllSingleGeometries;
    for (std::size_t j = 0; j < clusterIndexes.size(); ++j)
    {
      std::size_t index = clusterIndexes.at(j);

      te::gm::GeometryVector vecSingleGeometries;
      te::gm::Multi2Single(vecFragments.at(index), vecSingleGeometries);

      vecAllSingleGeometries.insert(vecAllSingleGeometries.end(), vecSingleGeometries.begin(), vecSingleGeometries.end());
    }

    //for the union operation using geometries with few intersections, it is faster to use buffer(0)
    DissolveResult* dissolveResult = new DissolveResult(i);
    dissolveResult->m_geometryVector = vecAllSingleGeometries;

    //adds the jobs to the manager
    threadGroup.addJob(boost::bind(DissolveThread, dissolveResult));

    //no thread
    //DissolveThread(dissolveResult);

    vecDissolveResult.push_back(dissolveResult);
  }

  //wait for the jobs to be finished
  threadGroup.waitToFinish();

  //then we join the results
  for (std::size_t i = 0; i < vecDissolveResult.size(); ++i)
  {
    DissolveResult* dissolveResult = vecDissolveResult.at(i);

    vecResult.push_back(dissolveResult->m_outputGeometry.release());
  }

  te::common::FreeContents(vecDissolveResult);

  return vecResult;
}

te::gm::GeometryVector te::vp::PolygonSubdivider::merge(const te::gm::GeometryVector& vecFragments, const te::gm::LineString* splitLine)
{
  //we first calculate all the indexes of the objects that must be dissolved
  GeometryAssociations geometryAssociations = calculateGeometryAssociations(vecFragments, splitLine);
  SetIndexes setUniqueIndexes = getUniqueIndexes(geometryAssociations);
  Clusters clusters = calculateGeometryClusters(geometryAssociations);

  //then we get the map containing the matches between the border polygons and dissolve them
  std::vector<std::size_t> vecSeedIndexes;
  te::gm::GeometryVector mergeResult = dissolveFragments(vecFragments, clusters);

  //then we just complete the list with the geometries that havent needed to be dissolved
  for (std::size_t i = 0; i < vecFragments.size(); ++i)
  {
    if (setUniqueIndexes.find(i) == setUniqueIndexes.end())
    {
      mergeResult.push_back(static_cast<te::gm::Geometry*>(vecFragments.at(i)->clone()));
    }
  }

  return mergeResult;
}

te::gm::GeometryVector te::vp::PolygonSubdivider::merge(const te::gm::GeometryVector& vecFragments, const SubdividerTilling& tilling)
{
  te::gm::Dimensionality dimensionality = te::gm::A;

  //we create a vector to store all segments on border from all geometries
  std::vector<te::vp::SegmentInfo*> vecSegmentsInBorder;

  te::gm::GeometryVector vecResult;
  te::gm::GeometryVector vecNeedToDissolve;
  for (std::size_t i = 0; i < vecFragments.size(); ++i)
  {
    const te::gm::Geometry* geometry = vecFragments.at(i);
    if (geometry == nullptr)
    {
      continue;
    }

    dimensionality = geometry->getDimension();
    te::gm::Coord2D internalCoord = geometry->getMBR()->getCenter();

    //we first aquire information to make the merge
    std::size_t column = tilling.getTileColumn(internalCoord);
    std::size_t row = tilling.getTileRow(internalCoord);
    te::gm::Envelope currentTile = tilling.getTile(column, row);

    te::gm::LineString leftBorder = tilling.getSplitLine(column, row, te::vp::SubdividerTilling::SplitLinePosition::LEFT);
    te::gm::LineString rightBorder = tilling.getSplitLine(column, row, te::vp::SubdividerTilling::SplitLinePosition::RIGHT);
    te::gm::LineString topBorder = tilling.getSplitLine(column, row, te::vp::SubdividerTilling::SplitLinePosition::TOP);
    te::gm::LineString bottomBorder = tilling.getSplitLine(column, row, te::vp::SubdividerTilling::SplitLinePosition::BOTTOM);

    bool isOnLeftBorder = te::vp::PolygonSubdivider::filterSegmentsInBorder(geometry, i, &leftBorder, vecSegmentsInBorder);
    bool isOnRightBorder = te::vp::PolygonSubdivider::filterSegmentsInBorder(geometry, i, &rightBorder, vecSegmentsInBorder);
    bool isOnTopBorder = te::vp::PolygonSubdivider::filterSegmentsInBorder(geometry, i, &topBorder, vecSegmentsInBorder);
    bool isOnBottomBorder = te::vp::PolygonSubdivider::filterSegmentsInBorder(geometry, i, &bottomBorder, vecSegmentsInBorder);

    bool isOnMiddleLinesBorder = checkForGeometrySubdivision(currentTile, geometry, i, vecSegmentsInBorder);

    if (isOnLeftBorder == false && isOnRightBorder == false && isOnTopBorder == false && isOnBottomBorder == false && isOnMiddleLinesBorder == false)
    {
      vecResult.push_back(static_cast<te::gm::Geometry*>(geometry->clone()));

      vecNeedToDissolve.push_back(nullptr);
      continue;
    }

    vecNeedToDissolve.push_back(static_cast<te::gm::Geometry*>(geometry->clone()));
  }

  //now we must analyse the geometries that have segments of the cell border
  te::vp::PolygonSubdivider::GeometryAssociations geometryAssociations = te::vp::PolygonSubdivider::calculateGeometryAssociations(vecSegmentsInBorder, dimensionality);
  te::vp::PolygonSubdivider::Clusters clusters = te::vp::PolygonSubdivider::calculateGeometryClusters(geometryAssociations);

  std::vector<std::size_t> vecSeedIndexes;
  te::gm::GeometryVector vecDissolved = te::vp::PolygonSubdivider::dissolveFragments(vecNeedToDissolve, clusters);
  vecResult.insert(vecResult.end(), vecDissolved.begin(), vecDissolved.end());

  te::common::FreeContents(vecNeedToDissolve);

  return vecResult;
}

te::vp::PolygonSubdivider::Clusters te::vp::PolygonSubdivider::calculateGeometryClusters(const GeometryAssociations& geometryAssociations)
{
  typedef
    boost::adjacency_list<
    boost::vecS           // edge list
    , boost::vecS         // vertex list
    , boost::undirectedS  // directedness
    , float               // property associated with vertices
    >
    Graph;

  typedef boost::bimap< std::size_t, std::size_t > Bimap;
  typedef Bimap::value_type BimapItem;
  
  Graph graph;              // create a graph
  Bimap mapIndexes;
  
  //we first create a set containing the unique indexes
  std::set<std::size_t> setUniqueIndexes = getUniqueIndexes(geometryAssociations);  
  std::size_t numVertexes = setUniqueIndexes.size();

  //then we must add new vertexes to tge graph, onde for each unique index
  //we must also map the values from boost indexes to my indexes
  std::set<std::size_t>::iterator itUniqueIndexes = setUniqueIndexes.begin();
  while (itUniqueIndexes != setUniqueIndexes.end())
  {
    std::size_t boostIndex = boost::add_vertex(graph);
    std::size_t myIndex = (*itUniqueIndexes);

    mapIndexes.insert(BimapItem(myIndex, boostIndex));
    ++itUniqueIndexes;
  }

  //then we must add the edges to the graph representing the associations of the geometries
  for(std::size_t i = 0; i < geometryAssociations.size(); ++i)
  {
    const GeometryAssociation& geometryAssociation = geometryAssociations.at(i);

    std::size_t indexBoostRed = mapIndexes.left.at(geometryAssociation.first);
    std::size_t indexBoostBlue = mapIndexes.left.at(geometryAssociation.second);

    //add edge
    boost::add_edge(indexBoostRed, indexBoostBlue, graph);
  }
  
  std::vector<int> component(numVertexes);
  size_t num_components = boost::connected_components(graph, &component[0]);

  //now we normalize the geometryMatchMap
  Clusters clusters;
  clusters.resize(num_components);
  for (std::size_t i = 0; i < numVertexes; ++i)
  {
    std::size_t myIndex = mapIndexes.right.at(i);
    clusters.at(component.at(i)).push_back(myIndex);
  }

  return clusters;
}

te::vp::PolygonSubdivider::GeometryAssociations te::vp::PolygonSubdivider::calculateGeometryAssociations(const std::vector<te::vp::SegmentInfo*>& vecBorderSegments, te::gm::Dimensionality dimensionality)
{
  //we must first index all the border segments that are overlapping the split line
 //we also discard all the geometries that do not touches the split line
  te::sam::rtree::Index<std::size_t> indexSegments;
  for (std::size_t i = 0; i < vecBorderSegments.size(); ++i)
  {
    const te::vp::SegmentInfo* segmentInfo = vecBorderSegments.at(i);
    indexSegments.insert(segmentInfo->getEnvelope(), i);
  }

  //then we try to find a match for each border segment
  GeometryAssociations geometryAssociations;
  for (std::size_t i = 0; i < vecBorderSegments.size(); ++i)
  {
    const te::vp::SegmentInfo* currentSegmentInfo = vecBorderSegments.at(i);

    //we must ensure that we will be part of the output
    geometryAssociations.push_back(GeometryAssociation(currentSegmentInfo->getGeometryIndex(), currentSegmentInfo->getGeometryIndex()));

    // we first find the candidate segments in the index
    std::vector<std::size_t> vecCandidadtes;
    indexSegments.search(currentSegmentInfo->getEnvelope(), vecCandidadtes);

    for (std::size_t j = 0; j < vecCandidadtes.size(); ++j)
    {
      const std::size_t& candidateIndex = vecCandidadtes.at(j);
      const te::vp::SegmentInfo* candidateSegmentInfo = vecBorderSegments.at(candidateIndex);

      //we the candidate is from the current polygon, we discard it
      if (currentSegmentInfo->getGeometryIndex() == candidateSegmentInfo->getGeometryIndex())
      {
        continue;
      }

      //if the candidate segment is from another polygon, we must check if it is equal from the current segment and in opposite position
      te::gm::Envelope envelopeCoord1 = te::gm::CreateEnvelope(currentSegmentInfo->getCoord1(), 0.);
      te::gm::Envelope envelopeCoord2 = te::gm::CreateEnvelope(currentSegmentInfo->getCoord2(), 0.);

      bool addToMap = false;

      if (dimensionality == te::gm::A)
      {
        if (envelopeCoord1.intersects(candidateSegmentInfo->getEnvelope()) == true && envelopeCoord2.intersects(candidateSegmentInfo->getEnvelope()) == true)
        {
          addToMap = true;
        }
      }
      else if (dimensionality == te::gm::L)
      {
        if (currentSegmentInfo->getCoord1() == candidateSegmentInfo->getCoord2() || currentSegmentInfo->getCoord2() == candidateSegmentInfo->getCoord1())
        {
          addToMap = true;
        }
      }

      if(addToMap == true)
      {
        //we must store the geometry associations
        geometryAssociations.push_back(GeometryAssociation(candidateSegmentInfo->getGeometryIndex(), currentSegmentInfo->getGeometryIndex()));
      }
    }
  }

  return geometryAssociations;
}


te::vp::PolygonSubdivider::SetIndexes te::vp::PolygonSubdivider::getUniqueIndexes(const GeometryAssociations& geometryAssociations)
{
  SetIndexes setUniqueIndexes;
  for(std::size_t i = 0; i < geometryAssociations.size(); ++i)
  {
    const GeometryAssociation& geometryAssociation = geometryAssociations.at(i);

    setUniqueIndexes.insert(geometryAssociation.first);
    setUniqueIndexes.insert(geometryAssociation.second);
  }

  return setUniqueIndexes;
}
