/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
 \file RTreeIndex.h
 
 \brief This class makes easier and more generic the use of the te::sam::rtree::Index
 
 \ingroup vp
 */

#ifndef __TERRALIB_VP_INTERNAL_RTREEINDEX_H
#define __TERRALIB_VP_INTERNAL_RTREEINDEX_H

// Terralib
#include "Config.h"

#include "../common/STLUtils.h"
#include "../sam/rtree/Index.h"
#include "CommonDataStructures.h"

//STL
#include <memory>
#include <vector>

namespace te
{
  namespace vp
  {
    /*!
      \class RTreeIndex

      \brief This class makes easier and more generic the use of the te::sam::rtree::Index

      \ingroup vp
    */
    template<class T>
    class RTreeIndex
    {
    public:
      
      //!< Constructor
      RTreeIndex();
      
      //!< Destructor
      ~RTreeIndex();

      //!< Clears the index
      void clear();

      //!< Inserts the given data considering the given envelope as key
      void insert(const te::gm::Envelope& envelope, const T& data);

      //!< Removes the given data considering  the given envelope as key
      void remove(const te::gm::Envelope& envelope, std::size_t dataIndex);

      //!< Gets all data from the RTreeIndex
      std::unique_ptr<te::vp::IndexReport<T>> getData() const;

      //!< Searches from all data consdering the given mbr
      std::unique_ptr<te::vp::IndexReport<T>> search(const te::gm::Envelope& mbr) const;

    private:
      te::sam::rtree::Index<std::size_t> m_index; //!< The internal rtree to store the data indexes
      IndexContainer<T> m_indexContainer; //!< The indexed data container
    };
  }
}

//Implementation
template<class T>
te::vp::RTreeIndex<T>::RTreeIndex()
{

}
template<class T>
te::vp::RTreeIndex<T>::~RTreeIndex()
{
  clear();
}

template<class T>
void te::vp::RTreeIndex<T>::clear()
{
  m_index.clear();
  m_indexContainer.clear();
}


template<class T>
void te::vp::RTreeIndex<T>::insert(const te::gm::Envelope& envelope, const T& data)
{
  std::size_t newDataIndex = m_indexContainer.insert(data);
  m_index.insert(envelope, newDataIndex);
}

template<class T>
void te::vp::RTreeIndex<T>::remove(const te::gm::Envelope& envelope, std::size_t dataIndex)
{
  m_index.remove(envelope, dataIndex);
  m_indexContainer.remove(dataIndex);
}

template<class T>
std::unique_ptr<te::vp::IndexReport<T>> te::vp::RTreeIndex<T>::getData() const
{
  std::vector<std::size_t> vecValidIndexes = m_indexContainer.getValidIndexes();

  std::unique_ptr<te::vp::IndexReport<T>> indexReport(new te::vp::IndexReport<T>(&m_indexContainer, vecValidIndexes));
  return indexReport;
}
 
template<class T>
std::unique_ptr<te::vp::IndexReport<T>> te::vp::RTreeIndex<T>::search(const te::gm::Envelope& mbr) const
{
  std::vector<std::size_t> vecIndexesReport;
  m_index.search(mbr, vecIndexesReport);

  std::unique_ptr<te::vp::IndexReport<T>> indexReport(new te::vp::IndexReport<T>(&m_indexContainer, vecIndexesReport));
  return indexReport;
}

#endif // __TERRALIB_VP_INTERNAL_RTREEINDEX_H
