/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

#include "Register.h"
//#include "../common/PlatformUtils.h"
//#include "../sam/rtree.h"
#include "../common/progress/TaskProgress.h"
#include "../dataaccess/utils/Utils.h"
#include "../sam/rtree/Index.h"
#include "../geometry/GeometryCollection.h"
#include "../geometry/GeometryProperty.h"
#include "../geometry/Envelope.h"
//#include "../dataaccess/dataset/DataSetType.h"
//#include "../dataaccess/utils/Utils.h"
#include "../srs/Converter.h"
#include "../memory/DataSet.h"
#include "../memory/DataSetItem.h"

//#include <thread>
//#include <mutex>
#include <memory>

namespace te
{
  namespace vp
  {
    namespace Register
    {
      bool TranslateCoord(
        const std::vector< TiePointT >& refPts,
        const te::sam::rtree::Index<std::size_t >& refPtsTree,
        const double& treeSearchBoxIncrementSize,
        const unsigned int& minNeighborTiePoints,
        double& xCoord, double& yCoord )
      {
        const te::gm::Envelope& treeMBR = refPtsTree.getMBR();
        std::vector< std::size_t > treeResult;
        te::gm::Envelope env( xCoord, yCoord, xCoord, yCoord );
        std::size_t treeResultSize = 0;

        while( true )
        {
          env.m_llx -= treeSearchBoxIncrementSize;
          env.m_lly -= treeSearchBoxIncrementSize;
          env.m_urx += treeSearchBoxIncrementSize;
          env.m_ury += treeSearchBoxIncrementSize;

          treeResult.clear();
          refPtsTree.search( env, treeResult );
          treeResultSize = treeResult.size();

          if( treeResultSize >= minNeighborTiePoints )
          {
            double meanXOff = 0;
            double meanYOff = 0;
            double xDiff = 0;
            double yDiff = 0;
            double weightsSum = 0;
            double weight = 0;
            double xRefOff = 0;
            double yRefOff = 0;            

            for( std::size_t treeResultIdx = 0; treeResultIdx < treeResultSize ;
              ++treeResultIdx )
            {
              const TiePointT& currRefPtTuple =
                refPts[ treeResult[ treeResultIdx ] ];
                
              const double& srcX = std::get< 1 >( currRefPtTuple );
              const double& srcY = std::get< 2 >( currRefPtTuple );
              const double& trgX = std::get< 4 >( currRefPtTuple );
              const double& trgY = std::get< 5 >( currRefPtTuple );              
              
              xRefOff = trgX - srcX;
              yRefOff = trgY - srcY;

              xDiff = srcX - xCoord;
              yDiff = srcY - yCoord;

              if( ( xDiff == 0 ) && ( yDiff == 0 ) )
              {
                meanXOff = xRefOff;
                meanYOff = yRefOff;

                weightsSum = 1;

                break;
              }
              else
              {
                weight = 1.0 / std::sqrt( ( xDiff * xDiff ) + ( yDiff * yDiff ) );

                weightsSum += weight;

                meanXOff += ( xRefOff * weight );
                meanYOff += ( yRefOff * weight );
              }
            }

            meanXOff /= weightsSum;
            meanYOff /= weightsSum;

            xCoord += meanXOff;
            yCoord += meanYOff;

            return true;
          }

          if( env.contains( treeMBR ) )
          {
            break;
          }
        }

        return false;
      }
      
      bool TranslateGeometry(
        const std::vector< TiePointT >& refPts,
        const te::sam::rtree::Index<std::size_t >& refPtsTree,
        const double& treeSearchBoxIncrementSize,
        const unsigned int& minNeighborTiePoints,
        te::gm::Geometry* geomPtr )
      {
        assert( !refPtsTree.isEmpty() );

        te::gm::GeometryCollection* gc =
          dynamic_cast< te::gm::GeometryCollection*>(geomPtr);

        if (gc)
        {
          const std::size_t nGeoms = gc->getNumGeometries();

          for (std::size_t i = 0; i < nGeoms; ++i)
          {
            if( ! TranslateGeometry( refPts, refPtsTree, treeSearchBoxIncrementSize,
              minNeighborTiePoints, gc->getGeometryN(i) ) )
            {
              return false;
            }
          }

          gc->computeMBR( true );
        }
        else
        {
          if( geomPtr->getGeomTypeId() == te::gm::PolygonType )
          {
            te::gm::Polygon const * polPtr = ( te::gm::Polygon*)geomPtr;
            const std::size_t nGeoms = polPtr->getNumRings();

            for (std::size_t i = 0; i < nGeoms; ++i)
            {
              if( ! TranslateGeometry( refPts, refPtsTree, treeSearchBoxIncrementSize,
                minNeighborTiePoints, polPtr->getRingN(i) ) )
              {
                return false;
              }
            }

            polPtr->computeMBR( true );
          }
          else if( geomPtr->getGeomTypeId() == te::gm::LineStringType )
          {
            te::gm::LineString const * castGeomPtr = ( te::gm::LineString*)geomPtr;
            const std::size_t nElements = castGeomPtr->getNPoints();
            te::gm::Coord2D* coordsVectorPtr = castGeomPtr->getCoordinates();

            for( std::size_t elementIdx = 0 ; elementIdx < nElements ; ++elementIdx )
            {
              if( ! TranslateCoord( refPts, refPtsTree, treeSearchBoxIncrementSize,
                 minNeighborTiePoints, coordsVectorPtr[ elementIdx ].x,
                 coordsVectorPtr[ elementIdx ].y ) )
              {
                return false;
              }
            }

            castGeomPtr->computeMBR( true );
          }
          else
          {
            return false;
          }
        }

        return true;
      }

      
      bool RegisterDatasetTransByN(
        const std::vector< TiePointT >& tiePoints,
        const unsigned int& minNeighborTiePoints,
        const bool enableThreads,
        te::da::DataSet& inDataset,
        const te::da::DataSetType& inDsType,
        const std::string& outDataSetName,
        te::da::DataAccess& outDataAccess )
      {
        if( tiePoints.empty() ) return false;
        if( outDataSetName.empty() ) return false;
        
        te::gm::GeometryProperty* inGeomPropPtr = te::da::GetFirstGeomProperty(
          &inDsType );
        if( inGeomPropPtr == nullptr )
        {
          return false;
        }
        
        const int inSRID = inGeomPropPtr->getSRID();
        if( inSRID == 0 ) return false;
        
        // reprojected tiePoints
        
        std::vector< TiePointT > reprojTiePoints( tiePoints.size() );
        
        {
          const std::size_t tiePointsSize = tiePoints.size();
          
          te::srs::Converter conv;
          conv.setTargetSRID( inSRID );
          
          for( std::size_t tiePointsIdx = 0 ; tiePointsIdx < tiePointsSize ;
            ++tiePointsIdx )
          {
            if( std::get< 0 >( tiePoints[ tiePointsIdx ] ) == 0 ) return false;
            conv.setSourceSRID( std::get< 0 >( tiePoints[ tiePointsIdx ] ) );
            conv.convert(  
              std::get< 1 >( tiePoints[ tiePointsIdx ] ),
              std::get< 2 >( tiePoints[ tiePointsIdx ] ),
              std::get< 1 >( reprojTiePoints[ tiePointsIdx ] ),
              std::get< 2 >( reprojTiePoints[ tiePointsIdx ] )
            );
            
            if( std::get< 3 >( tiePoints[ tiePointsIdx ] ) == 0 ) return false;
            conv.setSourceSRID( std::get< 3 >( tiePoints[ tiePointsIdx ] ) );
            conv.convert(  
              std::get< 4 >( tiePoints[ tiePointsIdx ] ),
              std::get< 5 >( tiePoints[ tiePointsIdx ] ),
              std::get< 4 >( reprojTiePoints[ tiePointsIdx ] ),
              std::get< 5 >( reprojTiePoints[ tiePointsIdx ] )
            );            
          }
        }

        // index

        te::sam::rtree::Index<std::size_t > reprojTiePointsTree;
        
        {
          const std::size_t reprojTiePointsSize = reprojTiePoints.size();

          te::gm::Envelope auxEnvelope;

          for( std::size_t reprojTiePointsIdx = 0 ; reprojTiePointsIdx < 
            reprojTiePointsSize ; ++reprojTiePointsIdx )
          {
            auxEnvelope.m_llx = auxEnvelope.m_urx = std::get< 1 >( 
              reprojTiePoints[ reprojTiePointsIdx ] );
            auxEnvelope.m_lly = auxEnvelope.m_ury = std::get< 2 >( 
              reprojTiePoints[ reprojTiePointsIdx ] );

            reprojTiePointsTree.insert( auxEnvelope, reprojTiePointsIdx );
          }
        }
        
        
        // creating a memory translated dataset

        std::unique_ptr<te::da::DataSetType> memDsType( 
          (te::da::DataSetType*)inDsType.clone() );
        memDsType->setName( outDataSetName );

        std::unique_ptr<te::mem::DataSet> memDataSetPtr(
          new te::mem::DataSet( memDsType.get() ) );
        
        {
          const double treeSearchBoxIncrementSize =
            (
              std::min( 
                reprojTiePointsTree.getMBR().getWidth(), 
                reprojTiePointsTree.getMBR().getHeight() 
              )
              /
              std::sqrt( (double)reprojTiePointsTree.size() )
            );          
          
          std::unique_ptr< te::mem::DataSetItem > dsItemPtr;

          inDataset.moveBeforeFirst();

          std::size_t inPropsNumber = inDataset.getNumProperties();
          std::size_t inPropIdx = 0;
          std::unique_ptr<te::dt::AbstractData> inAbsDataPtr;
          std::unique_ptr<te::gm::Geometry> inGeomPtr;
          bool geomTranslationOK = false;

          while( inDataset.moveNext() )
          {
            geomTranslationOK = false;
            dsItemPtr.reset( new te::mem::DataSetItem( memDsType.get() ) );

            for( inPropIdx = 0 ; inPropIdx < inPropsNumber ; ++inPropIdx )
            {
              if(
                  inDataset.getPropertyDataType( inPropIdx )
                  ==
                  te::dt::GEOMETRY_TYPE
                )
              {
                inGeomPtr = inDataset.getGeometry( inPropIdx );

                if( inGeomPtr.get() != nullptr )
                {
                  if( TranslateGeometry( reprojTiePoints, reprojTiePointsTree, 
                    treeSearchBoxIncrementSize,
                    minNeighborTiePoints, inGeomPtr.get() ) )
                  {
                    dsItemPtr->setGeometry( inDataset.getPropertyName( inPropIdx ),
                      inGeomPtr.release() );

                    geomTranslationOK = true;
                  }
                }
              }
              else
              {
                inAbsDataPtr = inDataset.getValue( inPropIdx );
                dsItemPtr->setValue( inDataset.getPropertyName( inPropIdx ),
                  inAbsDataPtr.release() );
              }
            }

            if( geomTranslationOK )
            {
              memDataSetPtr->add( dsItemPtr.release() );
            }
          }        
        }
        
        // saving to output dataset

        try
        {
          std::unique_ptr<te::da::DataSetType> outDsType( 
            (te::da::DataSetType*)inDsType.clone() );
          outDsType->setName( outDataSetName );
          
          outDataAccess.createDataSet( outDsType.get() );
          if( ! outDataAccess.dataSetExists() ) return false;
          
          memDataSetPtr->moveBeforeFirst();
          outDataAccess.saveDataSet( memDataSetPtr.release(), true );
        }
        catch(...)
        {
          return false;
        }        
        
        return true;
      }      
      
      bool RegisterDataset(
        const RegisterMethod& method,
        const std::vector< TiePointT >& tiePoints,
        const unsigned int& minNeighborTiePoints,
        const bool enableThreads,
        te::da::DataSet& inDataset,
        const te::da::DataSetType& inDsType,
        const std::string& outDataSetName,
        te::da::DataAccess& outDataAccess )
      {
        switch( method )
        {
          case TransByNMethod :
          {
            return RegisterDatasetTransByN( tiePoints, minNeighborTiePoints,
              enableThreads, inDataset, inDsType, outDataSetName, outDataAccess);
            break;
          }
          default :
          {
            return false;
          }
        }
        
        
        return true;
      }      
    } // end namespace Register
  } // end namespace vp
}   // end namespace te

