/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

#ifndef __TERRALIB_VP_INTERNAL_REGISTER_H
#define __TERRALIB_VP_INTERNAL_REGISTER_H

#include "Config.h"
#include "../dataaccess/dataset/DataSet.h"
#include "../dataaccess/dataset/DataSetType.h"
#include "../dataaccess/utils/DataAccess.h"

#include <tuple>
#include <vector>
#include <string>
namespace te
{
  namespace vp
  {
    /*!
      \namespace Register A namespace for functions to register vector data.
    */
    namespace Register
    {
      /*!
        \enum RegisterMethod
        \brief Register method.
       */    
      enum RegisterMethod
      {
        InvalidRegMethod = 0, //!< Invalid type
        TransByNMethod = 1 //!< Aplly a translation transform by using neighor tie-points.
      };      
      
      /*!
        \typedef TiePointT
        \brief Tie poit type definition [ SRID1, x1, y2, SRID2, x2, y2 ].
      */
      typedef  std::tuple< int, double, double, int, double, double > TiePointT;

      /*!
        \brief Register dataset geometries using supplied tie-points.
        \param method Method to use.
        \param tiePoints Input tie-points.
        \param minNeighborTiePoints Minimum number of neighbor tie-points to use when registering each geometry point.
        \param enableThreads Enable/disable the use of threads.
        \param inDataset Input dataset.
        \param inDsType Input dataset type.
        \param outDataSetName Output dataset name.
        \param outDataAccess Output data access where the registered dataset will be created.
        \return true if ok, false on errors.
        \note Tie points x1,y1 are source coords ( incorrect/adjust coord );
        \note Tie points x2,y2 are target coords ( correct/reference coord );
        \note NEW_COORD(target) = OLD_COORD(source) + OFFSET
        \note OFFSET = NEW_COORD(target) - OLD_COORD(source)        
      */
      TEVPEXPORT bool RegisterDataset(
        const RegisterMethod& method,
        const std::vector< TiePointT >& tiePoints,
        const unsigned int& minNeighborTiePoints,
        const bool enableThreads,
        te::da::DataSet& inDataset,
        const te::da::DataSetType& inDsType,
        const std::string& outDataSetName,
        te::da::DataAccess& outDataAccess );      

    }  // end namespace Register
  } // end namespace vp
}   // end namespace te

#endif  // __TERRALIB_VP_INTERNAL_REGISTER_H
