/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralin/vp/AbstractDataHandler.cpp

   \brief This class represents handler to help reading and writing data from/to a dataSource by dividing it into smaller groups
 */

#include "SubdivideDataHandler.h"

#include "../common/progress/TaskProgress.h"
#include "../common/StringUtils.h"
#include "../common/STLUtils.h"
#include "../core/translator/Translator.h"
#include "../datatype/AbstractData.h"
#include "../datatype/SimpleData.h"
#include "../dataaccess/dataset/DataSetAdapter.h"
#include "../dataaccess/dataset/DataSetTypeConverter.h"
#include "../dataaccess/utils/DataAccess.h"
#include "../dataaccess/utils/BasicQueries.h"
#include "../dataaccess/utils/DataAccess.h"
#include "../dataaccess/utils/Utils.h"
#include "../geometry/Envelope.h"
#include "../memory/DataSet.h" 
#include "../memory/PreparedItemCopy.h" 
#include "Feature.h"
#include "SubdivideByBorderTillingStrategy.h"
#include "Utils.h"

//Boost include files
#include <boost/thread/lock_guard.hpp>
#include <boost/thread/mutex.hpp>

namespace
{
  static te::vp::SubdivideDataHandlerFactory SubdivideDataHandlerFactoryInstance;
}

te::vp::SubdivideKey::SubdivideKey(const std::vector<te::dt::AbstractData*>& vecKey)
  : m_vecKeys(vecKey)
{
  init();
}

te::vp::SubdivideKey::SubdivideKey(const te::vp::SubdivideKey& other)
{
  *this = other;
}

te::vp::SubdivideKey::~SubdivideKey()
{
  clear();
}

te::vp::SubdivideKey& te::vp::SubdivideKey::operator=(const te::vp::SubdivideKey& other)
{
  clear();

  this->m_keyStr = other.m_keyStr;
  te::common::Clone(other.m_vecKeys, this->m_vecKeys);

  return *this;
}

bool te::vp::SubdivideKey::operator<(const te::vp::SubdivideKey& other) const
{
  if (m_keyStr < other.m_keyStr)
  {
    return true;
  }

  return false;
}

void te::vp::SubdivideKey::clear()
{
  te::common::FreeContents(m_vecKeys);
}

void te::vp::SubdivideKey::init()
{
  m_keyStr = "";
  for (std::size_t i = 0; i < m_vecKeys.size(); ++i)
  {
    if (i != 0)
    {
      m_keyStr += ";";
    }

    if (m_vecKeys.at(i) == nullptr)
    {
      m_keyStr += "null";
    }
    else
    {
      m_keyStr += m_vecKeys.at(i)->toString();
    }
  }
}

class SubdivideDataHandlerImpl
{
public:

  SubdivideDataHandlerImpl() = default;
  ~SubdivideDataHandlerImpl() = default;

  std::vector<te::da::DataAccess*> m_vecOriginalInputData;
  std::vector<te::da::DataAccess*> m_vecOriginalOutputData;
  std::vector< std::vector<std::string> > m_vecSubdivideColumns;
  std::vector< std::vector<std::string> > m_vecMergeColumns;
  std::vector<te::vp::SubdivideKey> m_vecGroups;
  std::unique_ptr<te::vp::SubdivideByBorderTillingStrategy> m_strategy;
  boost::mutex m_mutexRead;
  boost::mutex m_mutexWrite;
};


te::vp::SubdivideDataHandler::SubdivideDataHandler()
  : AbstractDataHandler()
  , m_impl(new SubdivideDataHandlerImpl())
{
}

//declaration just for the unique_ptr to be deleted on the cpp file
te::vp::SubdivideDataHandler::~SubdivideDataHandler() = default;

void te::vp::SubdivideDataHandler::initialize(te::vp::AbstractParameters* absParams, te::common::TaskProgress* taskProgress)
{
  AbstractDataHandler::initialize(absParams, taskProgress);

  m_impl->m_vecOriginalInputData = m_vecInputData;
  m_impl->m_vecOriginalOutputData = m_vecOutputData;

  std::size_t totalGeometries = 0;

  int srid = m_impl->m_vecOriginalInputData.at(0)->getSRID();
  te::gm::Envelope fullEnvelope;
  for (std::size_t i = 0; i < m_impl->m_vecOriginalInputData.size(); ++i)
  {
    const te::da::DataAccess* dataAccess = m_impl->m_vecOriginalInputData.at(i);
    std::unique_ptr<te::da::DataSet> dataSet = dataAccess->getDataSet();
    totalGeometries += dataSet->size();

    int currentSRID = dataAccess->getSRID();
    te::gm::Envelope currentEnvelope = m_impl->m_vecOriginalInputData.at(i)->getMBR();

    if (currentSRID != srid)
    {
      currentEnvelope.transform(currentSRID, srid);
    }
    
    fullEnvelope.Union(currentEnvelope);
  }

  double goodCountPerCel = static_cast<double>(totalGeometries) / 500.;
  double goodMatrixAxisSize = std::sqrt(goodCountPerCel);
  std::size_t goodMatrixAxisSizeRounded = static_cast<std::size_t>(std::ceil(goodMatrixAxisSize));

  std::size_t numRows = 10;
  std::size_t numColumns = 10;
  
  if (numRows < goodMatrixAxisSizeRounded)
  {
    numRows = goodMatrixAxisSizeRounded;
  }
  if (numColumns < goodMatrixAxisSizeRounded)
  {
    numColumns = goodMatrixAxisSizeRounded;
  }

  double resolutionX = fullEnvelope.getWidth() / static_cast<double>(numColumns);
  double resolutionY = fullEnvelope.getHeight() / static_cast<double>(numRows);

  //creates the strategy
  m_impl->m_strategy.reset(new te::vp::SubdivideByBorderTillingStrategy(fullEnvelope, srid, resolutionX, resolutionY, taskProgress));

  //subdivides
  m_vecInputData = m_impl->m_strategy->subdivide(m_impl->m_vecOriginalInputData, m_maxCoordinates);
  m_vecOutputData.clear();

  //prepares the subdivided output data access
  m_vecOutputData.push_back(m_impl->m_strategy->createOutputDataAccessForSubdivision(m_impl->m_vecOriginalOutputData.at(0)->getDataSetType(), m_vecInputData, "result"));

  initGroups();
}

void te::vp::SubdivideDataHandler::initGroups()
{
  std::set<SubdivideKey> setKeys;
  std::vector< std::vector<std::string> > vecSubdivideColumns;
  std::vector<std::string> vecAllColumns;
  for (std::size_t i = 0; i < m_vecInputData.size(); ++i)
  {
    te::da::DataAccess* dataAccess = m_vecInputData.at(i);

    const te::da::DataSetType* dataSetType = dataAccess->getDataSetType();
    std::string rowPropertyName = te::vp::SubdivideByBorderTillingStrategy::getRowPropertyName();
    std::string columnPropertyName = te::vp::SubdivideByBorderTillingStrategy::getColumnPropertyName();

    std::vector<std::string> vecCurrentSubdivideColumns;
    vecCurrentSubdivideColumns.push_back(rowPropertyName);
    vecCurrentSubdivideColumns.push_back(columnPropertyName);

    vecSubdivideColumns.push_back(vecCurrentSubdivideColumns);

    std::string query = "SELECT DISTINCT " + te::common::Convert2String(vecCurrentSubdivideColumns, ",") + " FROM " + dataSetType->getName();
    std::unique_ptr<te::da::DataSet> dataSet = dataAccess->getDataSource()->query(query);
    dataSet->moveBeforeFirst();
    while (dataSet->moveNext())
    {
      std::vector<te::dt::AbstractData*> vecKeys;
      for (std::size_t j = 0; j < vecCurrentSubdivideColumns.size(); ++j)
      {
        te::dt::AbstractData* absData = nullptr;
        if (dataSet->isNull(j) == false)
        {
          absData = dataSet->getValue(j).release();
        }

        vecKeys.push_back(absData);
      }

      SubdivideKey key(vecKeys);
      setKeys.insert(key);
    }
  }

  std::vector< std::vector<std::string> > vecMergeColumns;

  for (std::size_t i = 0; i < m_vecOutputData.size(); ++i)
  {
    te::da::DataAccess* dataAccess = m_vecOutputData.at(i);
    const te::da::DataSetType* dataSetType = dataAccess->getDataSetType();

    std::vector<std::string> vecCurrentMergeColumns;
    for (std::size_t i = 0; i < dataSetType->getProperties().size(); ++i)
    {
      std::string propertyName = dataSetType->getProperty(i)->getName();
      if (propertyName.find(te::vp::SubdivideByBorderTillingStrategy::getParentPropertyName()) != std::string::npos)
      {
        vecCurrentMergeColumns.push_back(propertyName);
      }
    }

    vecMergeColumns.push_back(vecCurrentMergeColumns);
  }

  std::vector<te::vp::SubdivideKey> vecGroups;
  vecGroups.insert(vecGroups.end(), setKeys.begin(), setKeys.end());

  setGroupInfo(vecGroups, vecSubdivideColumns, vecMergeColumns);
}

void te::vp::SubdivideDataHandler::setGroupInfo(const std::vector<te::vp::SubdivideKey>& vecGroups, const std::vector< std::vector<std::string> >& vecSubdivideColumns, const std::vector< std::vector<std::string> >& vecMergeColumns)
{
  m_impl->m_vecGroups = vecGroups;
  m_impl->m_vecSubdivideColumns = vecSubdivideColumns;
  m_impl->m_vecMergeColumns = vecMergeColumns;
}

void te::vp::SubdivideDataHandler::finalize(te::common::TaskProgress* taskProgress)
{
  bool dissolveGeometryCoordinateSubdivision = true;
  if (this->m_maxCoordinates == 0)
  {
    dissolveGeometryCoordinateSubdivision = false;
  }

  //now we must merge the tiles
  std::vector<te::da::DataAccess*> vecMergedDataAccess = m_impl->m_strategy->merge(m_vecOutputData, m_impl->m_vecMergeColumns, dissolveGeometryCoordinateSubdivision);

  te::mem::PreparedItemCopy copy;

  //and then copy the merge data access to the output dataccess
  for (std::size_t i = 0; i < vecMergedDataAccess.size(); ++i)
  {
    te::da::DataSourcePtr dataSource = vecMergedDataAccess.at(i)->getDataSource();
    std::unique_ptr<te::da::DataSet> dataSet = vecMergedDataAccess.at(i)->getDataSet();

    //check for empty results
    if (dataSet == nullptr)
    {
      continue;
    }

    const te::da::DataSetType* inputDataSetType = vecMergedDataAccess.at(i)->getDataSetType();
    std::unique_ptr<te::da::DataSetType> outputDataSetTypeWithoutKey(static_cast<te::da::DataSetType*>(m_impl->m_vecOriginalOutputData.at(i)->getDataSetType()->clone()));

    te::da::PrimaryKey* primaryKey = outputDataSetTypeWithoutKey->getPrimaryKey();
    if (primaryKey != nullptr)
    {
      if (primaryKey->getProperties().empty() == false)
      {
        const std::string& pkName = primaryKey->getProperties().at(0)->getName();
        te::dt::Property* keyProperty = outputDataSetTypeWithoutKey->getProperty(pkName);
        if (keyProperty != nullptr)
        {
          outputDataSetTypeWithoutKey->remove(keyProperty);
        }
      }
    }

    if (taskProgress != nullptr)
    {
      taskProgress->setMessage(TE_TR("Saving Results"));
      taskProgress->setTotalSteps(static_cast<int>(0));
      taskProgress->useMultiThread(true);
    }

    std::unique_ptr<te::mem::DataSet> outputDataSet(new te::mem::DataSet(outputDataSetTypeWithoutKey.get()));

    dataSet->moveBeforeFirst();
    while (dataSet->moveNext())
    {
      te::vp::Feature* feature = new te::vp::Feature(outputDataSetTypeWithoutKey.get());
      copy.copyValues(dataSet->getItem().get(), feature);

      outputDataSet->add(feature);
    }

    if (outputDataSet->size() > 0)
    {
      m_impl->m_vecOriginalOutputData.at(i)->saveDataSet(outputDataSet.release(), true);
    }
  }

  te::common::FreeContents(vecMergedDataAccess);

  for (std::size_t i = 0; i < m_vecInputData.size(); ++i)
  {
    delete m_vecInputData.at(i);
  }

  for (std::size_t i = 0; i < m_vecOutputData.size(); ++i)
  {
    delete m_vecOutputData.at(i);
  }
}


std::size_t te::vp::SubdivideDataHandler::getGroupCount() const
{
  return m_impl->m_vecGroups.size();
}


te::vp::FeatureSet te::vp::SubdivideDataHandler::getFeatureSet(std::size_t groupIndex, std::size_t inputIndex, const te::gm::Envelope& filter)
{
  std::unique_ptr<te::da::DataSet> dataSet;

  const std::vector<std::string>& vecSubdivideColumns = m_impl->m_vecSubdivideColumns.at(inputIndex);
  const std::vector<te::dt::AbstractData*>& vecSubdivideKeys = m_impl->m_vecGroups.at(groupIndex).m_vecKeys;
  te::da::DataAccess* dataAccess = m_vecInputData.at(inputIndex);

  const te::da::DataSetType* dataSetType = dataAccess->getDataSetType();

  std::vector<std::string> vecColumns;
  for (std::size_t i = 0; i < dataSetType->size(); ++i)
  {
    vecColumns.push_back(dataSetType->getProperty(i)->getName());
  }

  std::unique_ptr<te::da::Select> selectGroup = te::da::BasicQueries::GetQuery(false, vecColumns, m_vecInputData.at(inputIndex)->getDataSetName(), vecSubdivideColumns, vecSubdivideKeys);

  //creates a scope so the locker can be used just in the query of data
  {
    boost::lock_guard<boost::mutex> lockGuard(m_impl->m_mutexRead);

    dataSet = dataAccess->query(*selectGroup);
  }

  //declares the resulting featureset
  te::vp::FeatureSet featureSet;
  if (dataSet == nullptr)
  {
    return featureSet;
  }

  dataSet->moveBeforeFirst();
  while (dataSet->moveNext())
  {
    te::vp::Feature* feature = te::vp::Feature::createFeatureFromCurrent(dataSet.get());
    featureSet.push_back(feature);
  }

  return featureSet;
}

void te::vp::SubdivideDataHandler::addFeatureSet(std::size_t groupIndex, std::size_t outputIndex, const te::vp::FeatureSet& featureSet)
{
  if (featureSet.empty() == true)
  {
    return;
  }

  te::da::DataAccess* dataAccess = m_vecOutputData.at(outputIndex);
  std::unique_ptr<te::da::DataSet> dataSet = te::vp::CreateDataSetFromFeatureSet(dataAccess->getDataSetName(), dataAccess->getDataSetType(), featureSet);

  //the lock is only used in the saving of the geometries
  boost::lock_guard<boost::mutex> lockGuard(m_impl->m_mutexWrite);
  dataAccess->saveDataSet(dataSet.release(), false);
}


te::vp::SubdivideDataHandlerFactory::SubdivideDataHandlerFactory()
  : te::vp::AbstractDataHandlerFactory("SubdivideDataHandler")
{

}

te::vp::SubdivideDataHandlerFactory::~SubdivideDataHandlerFactory() = default;

te::vp::AbstractDataHandler* te::vp::SubdivideDataHandlerFactory::build()
{
  return new te::vp::SubdivideDataHandler();
}
