/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralin/vp/SubdivideOperation.cpp

   \brief Represents a make valid operation which analyses and fixes features of a set
 */
#include "SubdivideOperation.h"

#include "../datatype/SimpleProperty.h"
#include "../dataaccess/dataset/DataSetType.h"
#include "../dataaccess/utils/Utils.h"
#include "../geometry/algorithms/CellTilling.h"
#include "../geometry/algorithms/GeometrySubdivider.h"
#include "../geometry/Geometry.h"
#include "../geometry/GeometryCollection.h"
#include "../geometry/Utils.h"
#include "../memory/PreparedItemCopy.h"
#include "CommonFunctions.h"
#include "Feature.h"
#include "PolygonSubdivider.h"
#include "SubdivideByBorderTillingStrategy.h"

namespace te
{
  namespace vp
  {
    namespace subdivideOperation
    {
      const std::string OPERATION_NAME = "Subdivide";
    }
  }
}

namespace
{
  static te::vp::SubdivideOperationFactory
    SubdivideOperationFactoryInstance;
}

te::vp::SubdivideOperationParameters::SubdivideOperationParameters(std::size_t maxCoordinates, te::gm::CellTilling* cellTilling)
  : te::vp::AbstractParameters()
  , m_maxCoordinates(maxCoordinates)
  , m_cellTilling(cellTilling)
{

}
te::vp::SubdivideOperationParameters::~SubdivideOperationParameters() = default;

void te::vp::SubdivideOperationParameters::setCellTilling(te::gm::CellTilling* cellTilling)
{
  m_cellTilling.reset(cellTilling);
}

const te::gm::CellTilling* te::vp::SubdivideOperationParameters::getCellTilling() const
{
  return m_cellTilling.get();
}

void te::vp::SubdivideOperationParameters::setMaxCoordinates(std::size_t maxCoordinates)
{
  m_maxCoordinates = maxCoordinates;
}

std::size_t te::vp::SubdivideOperationParameters::getMaxCoordinates() const
{
  return m_maxCoordinates;
}

te::vp::SubdivideOperationCapabilities::SubdivideOperationCapabilities(const std::string& operationName, std::size_t numberOfInputs, std::size_t numberOfOutputs, const std::string& dataHandlerName, bool snapGeometries, bool allowCoordinateSubdivision)
  : te::vp::AbstractOperationCapabilities(operationName, numberOfInputs, numberOfOutputs, dataHandlerName, snapGeometries, allowCoordinateSubdivision)
{

}

std::unique_ptr<te::vp::AbstractParameters> te::vp::SubdivideOperationCapabilities::createParameters() const
{
  std::unique_ptr<te::vp::AbstractParameters> params(new te::vp::SubdivideOperationParameters());
  return params;
}


te::vp::SubdivideOperation::SubdivideOperation()
    : AbstractOperation(te::vp::subdivideOperation::OPERATION_NAME)
    , m_parentId(0)
{

}

te::vp::SubdivideOperation::~SubdivideOperation() = default;

std::unique_ptr<te::vp::AbstractOperationCapabilities> te::vp::SubdivideOperation::getCapabilities() const
{
  std::unique_ptr<te::vp::AbstractOperationCapabilities> capabilities(new te::vp::SubdivideOperationCapabilities(getName(), 1, 1, "UnorderedGroupingDataHandler", false, true));
  return capabilities;
}

std::vector<te::vp::FeatureSet> te::vp::SubdivideOperation::executeImpl(const std::vector<te::vp::FeatureSet>& vecInput)
{
  std::vector<te::vp::FeatureSet> vecOutput;
  if (vecInput.size() != 1)
  {
    vecOutput.resize(1);
    return vecOutput;
  }

  te::vp::SubdivideOperationParameters* subdivideParameters = dynamic_cast<te::vp::SubdivideOperationParameters*>(this->getParameters());
  if (subdivideParameters == nullptr)
  {
    vecOutput.resize(1);
    return vecOutput;
  }

  const te::vp::FeatureSet& inputFeatureSet = vecInput.at(0);
  if (inputFeatureSet.empty() == true)
  {
    vecOutput.resize(1);
    return vecOutput;
  }

  //initialize property positions
  std::string rowPropertyName = te::vp::SubdivideByBorderTillingStrategy::getRowPropertyName();
  std::string columnPropertyName = te::vp::SubdivideByBorderTillingStrategy::getColumnPropertyName();
  std::string parentPropertyName = te::vp::SubdivideByBorderTillingStrategy::getParentPropertyName();

  const te::da::DataSetType* outputDataSetType = getDataSetType(0);

  std::size_t rowIndex = outputDataSetType->getPropertyPosition(rowPropertyName);
  std::size_t columnIndex = outputDataSetType->getPropertyPosition(columnPropertyName);
  std::size_t parentIdIndex = outputDataSetType->getPropertyPosition(parentPropertyName);

  const te::gm::CellTilling* cellTilling = subdivideParameters->getCellTilling();
  std::size_t maxCoordinates = subdivideParameters->getMaxCoordinates();  

  te::mem::PreparedItemCopy copy;

  te::vp::FeatureSet outputFeatureSet;

  int srid = inputFeatureSet.at(0)->getFirstGeometryPtr()->getSRID();
  double precision = te::gm::GetAppropriatePrecision(srid);

  //we must convert the geometries to the Identity operation output type
  for (std::size_t i = 0; i < inputFeatureSet.size(); ++i)
  {
    const te::vp::Feature* inputFeature = inputFeatureSet.at(i);
    const te::gm::Geometry* geometry = inputFeature->getFirstGeometryPtr();
    if (geometry == nullptr)
    {
      continue;
    }

    if (this->isAborted() == true)
    {
      vecOutput.push_back(outputFeatureSet);
      return vecOutput;
    }

    std::vector<te::gm::GeometrySubdivider::ColumnRow> vecColumnRow;
    te::gm::GeometryVector vecSubdivided;

    //check if we must sudivide the geometry by a cell tilling
    if (cellTilling != nullptr)
    {
      vecSubdivided = te::gm::GeometrySubdivider::subdivideByCellTilling(geometry, *cellTilling, vecColumnRow);
    }
    else
    {
      vecSubdivided.push_back(static_cast<te::gm::Geometry*>(geometry->clone()));
    }

    if (maxCoordinates > 0)
    {
      std::vector<te::gm::GeometrySubdivider::ColumnRow> vecColumnRowStep2;
      te::gm::GeometryVector vecSubdividedStep2;
      for (std::size_t j = 0; j < vecSubdivided.size(); ++j)
      {
        const te::gm::Geometry* currentGeometry = vecSubdivided.at(j);
        te::gm::GeometryVector vecCurrentSubdivided = te::gm::GeometrySubdivider::subdivideByMaxCoordinates(currentGeometry, maxCoordinates);
        for (std::size_t k = 0; k < vecCurrentSubdivided.size(); ++k)
        {
          vecSubdividedStep2.push_back(vecCurrentSubdivided.at(k));
          
          //geometry may hot have been splitted by cell
          if (vecColumnRow.empty() == false)
          {
            const te::gm::GeometrySubdivider::ColumnRow& currentColumnRow = vecColumnRow.at(j);
            vecColumnRowStep2.push_back(currentColumnRow);
          }
        }
      }

      te::common::FreeContents(vecSubdivided);

      vecColumnRow = vecColumnRowStep2;
      vecSubdivided = vecSubdividedStep2;
    }

    std::size_t parentId = safeCreateParentId();

    for (std::size_t j = 0; j < vecSubdivided.size(); ++j)
    {
      const te::gm::Geometry* currentGeometry = vecSubdivided.at(j);

      te::gm::GeometryVector vecSingleOutput;
      te::gm::Multi2Single(currentGeometry, vecSingleOutput);

      te::gm::GeometryVector vecAdjustedOutput = te::vp::AdjustToGeomType(vecSingleOutput, getOutputGeometryType());

      te::common::FreeContents(vecSingleOutput);

      for (std::size_t k = 0; k < vecAdjustedOutput.size(); ++k)
      {
        te::gm::Geometry* adjustedOutputGeometry = vecAdjustedOutput.at(k);

        te::vp::Feature* convertedFeature = createOutputFeature(0);
        copy.copyValues(inputFeature, convertedFeature, false);
        convertedFeature->setFirstGeometry(adjustedOutputGeometry);

        if (parentIdIndex != std::string::npos)
        {
          convertedFeature->setInt32(parentIdIndex, static_cast<boost::int32_t>(parentId));
        }

        if (rowIndex != std::string::npos)
        {
          const te::gm::GeometrySubdivider::ColumnRow& currentColumnRow = vecColumnRow.at(j);
          convertedFeature->setInt32(rowIndex, static_cast<boost::int32_t>(currentColumnRow.m_row));
        }

        if (columnIndex != std::string::npos)
        {
          const te::gm::GeometrySubdivider::ColumnRow& currentColumnRow = vecColumnRow.at(j);
          convertedFeature->setInt32(columnIndex, static_cast<boost::int32_t>(currentColumnRow.m_column));
        }

        outputFeatureSet.push_back(convertedFeature);
      }
    }

    te::common::FreeContents(vecSubdivided);

  }

  vecOutput.push_back(outputFeatureSet);
  
  return vecOutput;
}

te::da::DataSetType* te::vp::SubdivideOperation::createBasicOutputDataSetType(const std::string& dataSetName, int srid, te::gm::GeomType geometryType, const std::string& geometryColumnName) const
{
  te::da::DataSetType* basicDataSetType = AbstractOperation::createBasicOutputDataSetType(dataSetName, srid, geometryType, geometryColumnName);

  te::vp::SubdivideOperationParameters* subdivideParameters = dynamic_cast<te::vp::SubdivideOperationParameters*>(this->getParameters());
  if (subdivideParameters == nullptr)
  {
    return basicDataSetType;
  }

  if (subdivideParameters->getCellTilling() != nullptr)
  {
    std::string rowPropertyName = te::vp::SubdivideByBorderTillingStrategy::getRowPropertyName();
    if (te::da::GetPropertyPos(basicDataSetType, rowPropertyName) == std::string::npos)
    {
      te::dt::SimpleProperty* rowProperty = new te::dt::SimpleProperty(rowPropertyName, te::dt::INT32_TYPE, true);
      basicDataSetType->add(rowProperty);
    }    

    std::string columnPropertyName = te::vp::SubdivideByBorderTillingStrategy::getColumnPropertyName();
    if (te::da::GetPropertyPos(basicDataSetType, columnPropertyName) == std::string::npos)
    {
      te::dt::SimpleProperty* columnProperty = new te::dt::SimpleProperty(columnPropertyName, te::dt::INT32_TYPE, true);
      basicDataSetType->add(columnProperty);
    }    
  }

  if(subdivideParameters->getMaxCoordinates() > 0)
  {
    std::string parentPropertyName = te::vp::SubdivideByBorderTillingStrategy::getParentPropertyName();
    if (te::da::GetPropertyPos(basicDataSetType, parentPropertyName) == std::string::npos)
    {
      te::dt::SimpleProperty* absProperty = new te::dt::SimpleProperty(parentPropertyName, te::dt::INT32_TYPE, true);
      basicDataSetType->add(absProperty);
    }
  }

  return basicDataSetType;
}

std::size_t te::vp::SubdivideOperation::safeCreateParentId()
{
  const std::lock_guard<std::mutex> lock(m_mutex);
  ++m_parentId;

  return m_parentId;
}

te::vp::SubdivideOperationFactory::SubdivideOperationFactory()
  : OperationFactory(te::vp::subdivideOperation::OPERATION_NAME)
{
}

te::vp::SubdivideOperationFactory::~SubdivideOperationFactory() = default;

te::vp::AbstractOperation* te::vp::SubdivideOperationFactory::build()
{
  return new SubdivideOperation();
}
