/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

#include "SubdividerTilling.h"

#include "../common/MathUtils.h"
#include "../geometry/LineString.h"

#include <cmath>

te::vp::SubdividerTilling::SubdividerTilling(const te::gm::Envelope& envelope, double resolutionX, double resolutionY)
  : m_envelope(envelope)
  , m_resolutionX(resolutionX)
  , m_resolutionY(resolutionY)
{
  m_numColumns = static_cast<std::size_t>(std::ceil(m_envelope.getWidth() / m_resolutionX));
  m_numRows = static_cast<std::size_t>(std::ceil(m_envelope.getHeight() / m_resolutionY));
}

std::size_t te::vp::SubdividerTilling::getNumColumns() const
{
  return m_numColumns;
}

std::size_t te::vp::SubdividerTilling::getNumRows() const
{
  return m_numRows;
}

std::size_t te::vp::SubdividerTilling::getNumTiles() const
{
  std::size_t numTiles = m_numColumns * m_numRows;
  return numTiles;
}

te::gm::Envelope te::vp::SubdividerTilling::getTile(std::size_t column, std::size_t row) const
{
  double x1 = m_envelope.m_llx + (column * m_resolutionX);
  double y1 = m_envelope.m_lly + (row * m_resolutionY);

  double x2 = m_envelope.m_llx + ((column + 1) * m_resolutionX);
  double y2 = m_envelope.m_lly + ((row + 1) * m_resolutionY);

  te::gm::Envelope envelope(x1, y1, x2, y2);
  return envelope;
}

std::size_t te::vp::SubdividerTilling::getTileColumn(const te::gm::Coord2D& coord) const
{
  double distanceFromOrigin = coord.x - m_envelope.m_llx;
  //a handle for the cases where the mbr is truncated and so the geometry is "moved" slightly outside of the mbr
  if (distanceFromOrigin < 0)
  {
    distanceFromOrigin = 0;
  }
  std::size_t columnIndex = static_cast<std::size_t>(std::floor(distanceFromOrigin / m_resolutionX));
  return columnIndex;
}

std::size_t te::vp::SubdividerTilling::getTileRow(const te::gm::Coord2D& coord) const
{
  double distanceFromOrigin = coord.y - m_envelope.m_lly;
  //a handle for the cases where the mbr is truncated and so the geometry is "moved" slightly outside of the mbr
  if (distanceFromOrigin < 0)
  {
    distanceFromOrigin = 0;
  }
  std::size_t rowIndex = static_cast<std::size_t>(std::floor(distanceFromOrigin / m_resolutionY));
  return rowIndex;
}

te::gm::LineString te::vp::SubdividerTilling::getSplitLine(std::size_t column, std::size_t row, SplitLinePosition position) const
{
  te::gm::Envelope tile = getTile(column, row);
  return  te::vp::SubdividerTilling::getSplitLine(tile, position);
}

te::gm::LineString te::vp::SubdividerTilling::getSplitLine(const te::gm::Envelope& tile, SplitLinePosition position)
{
  te::gm::LineString splitLineString(2, te::gm::LineStringType);
  if (position == SplitLinePosition::TOP)
  {
    splitLineString.setX(0, tile.m_llx);
    splitLineString.setY(0, tile.m_ury);
    splitLineString.setX(1, tile.m_urx);
    splitLineString.setY(1, tile.m_ury);
  }
  else if (position == SplitLinePosition::BOTTOM)
  {
    splitLineString.setX(0, tile.m_llx);
    splitLineString.setY(0, tile.m_lly);
    splitLineString.setX(1, tile.m_urx);
    splitLineString.setY(1, tile.m_lly);
  }
  else if (position == SplitLinePosition::LEFT)
  {
    splitLineString.setX(0, tile.m_llx);
    splitLineString.setY(0, tile.m_lly);
    splitLineString.setX(1, tile.m_llx);
    splitLineString.setY(1, tile.m_ury);
  }
  else if (position == SplitLinePosition::RIGHT)
  {
    splitLineString.setX(0, tile.m_urx);
    splitLineString.setY(0, tile.m_lly);
    splitLineString.setX(1, tile.m_urx);
    splitLineString.setY(1, tile.m_ury);
  }
  else if (position == SplitLinePosition::MIDDLE_VERTICAL)
  {
    splitLineString.setX(0, tile.m_llx + (tile.getWidth() / 2.));
    splitLineString.setY(0, tile.m_lly);
    splitLineString.setX(1, tile.m_llx + (tile.getWidth() / 2.));
    splitLineString.setY(1, tile.m_ury);
  }
  else if (position == SplitLinePosition::MIDDLE_HORIZONTAL)
  {
    splitLineString.setX(0, tile.m_llx);
    splitLineString.setY(0, tile.m_lly + (tile.getHeight() / 2.));
    splitLineString.setX(1, tile.m_urx);
    splitLineString.setY(1, tile.m_lly + (tile.getHeight() / 2.));
  }

  return splitLineString;
}

te::gm::Envelope te::vp::SubdividerTilling::getTileN(std::size_t i) const
{
  std::size_t column = static_cast<std::size_t>(i % m_numColumns);
  std::size_t row = static_cast<std::size_t>(i / m_numColumns);

  return getTile(column, row);
}