/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

#include "TiePointsLocator.h"
#include "../common/PlatformUtils.h"
#include "../sam/rtree.h"
#include "../common/progress/TaskProgress.h"
#include "../geometry/GeometryCollection.h"
#include "../dataaccess/utils/Utils.h"

#include <thread>
#include <mutex>
#include <vector>
#include <memory>

namespace te
{
  namespace vp
  {
    namespace TiePointsLocator
    {
      struct MatchGeometriesThreadParams
      {
        std::mutex* m_mutexPtr;
        double m_maxXOffset;
        double m_maxYOffset;
        double m_maxPercentDiff;
        std::vector< te::gm::Geometry const* > const * m_geometries1Ptr;
        std::vector< te::gm::Geometry const* > const * m_geometries2Ptr;
        std::vector< TiePointT >* m_tiePointsPtr;
        bool* m_returnValuePtr;
        std::size_t* m_currGeometries1IdxPtr;
        te::common::TaskProgress* m_progressPtr;
      };

      void GetGeomInfos( te::gm::Geometry const * const gmPtr, double& area,
        double& length )
      {
        assert( gmPtr != nullptr );

        te::gm::GeometryCollection const * gc =
          dynamic_cast< te::gm::GeometryCollection const *>(gmPtr);

        if (gc)
        {
          const std::size_t nGeoms = gc->getNumGeometries();
          double localLengh = 0;
          double localArea = 0;

          for (std::size_t i = 0; i < nGeoms; ++i)
          {
            GetGeomInfos( gc->getGeometryN(i), localArea, localLengh );

            area += localArea;
            length += localLengh;
          }
        }
        else
        {
          if( dynamic_cast< te::gm::Curve const* >( gmPtr ) != nullptr )
          {
            area = 0;
            length = ((te::gm::Curve const*)gmPtr)->getLength();
          }
          else if( dynamic_cast< te::gm::Surface const* >( gmPtr ) != nullptr )
          {
            area = ((te::gm::Surface const*)gmPtr)->getArea();
            length = ((te::gm::Surface const*)gmPtr)->getPerimeter();
          }
          else
          {
            area = length = 0;
          }
        }
      }

      bool applyOffsets( te::gm::Geometry* geomPtr,
          const double& xOff, const double& yOff )
      {
        assert( geomPtr != nullptr );

        te::gm::GeometryCollection* gc =
          dynamic_cast< te::gm::GeometryCollection*>(geomPtr);

        if (gc)
        {
          const std::size_t nGeoms = gc->getNumGeometries();

          for (std::size_t i = 0; i < nGeoms; ++i)
          {
            if( ! applyOffsets( gc->getGeometryN(i), xOff, yOff ) )
            {
              return false;
            }
          }

          gc->computeMBR( true );
        }
        else
        {
          if( geomPtr->getGeomTypeId() == te::gm::PolygonType )
          {
            te::gm::Polygon const * polPtr = ( te::gm::Polygon*)geomPtr;
            const std::size_t nGeoms = polPtr->getNumRings();

            for (std::size_t i = 0; i < nGeoms; ++i)
            {
              if( ! applyOffsets( polPtr->getRingN(i), xOff, yOff ) )
              {
                return false;
              }
            }

            polPtr->computeMBR( true );
          }
          else if( geomPtr->getGeomTypeId() == te::gm::LineStringType )
          {
            te::gm::LineString const * castGeomPtr = ( te::gm::LineString*)geomPtr;
            const std::size_t nElements = castGeomPtr->getNPoints();
            te::gm::Coord2D* coordsVectorPtr = castGeomPtr->getCoordinates();

            for( std::size_t elementIdx = 0 ; elementIdx < nElements ; ++elementIdx )
            {
              coordsVectorPtr[ elementIdx ].x += xOff;
              coordsVectorPtr[ elementIdx ].y += yOff;
            }

            castGeomPtr->computeMBR( true );
          }
          else
          {
            return false;
          }
        }

        return true;
      }

      void MatchGeometriesThreadEntry( MatchGeometriesThreadParams* paramsPtr )
      {
        std::unique_lock<std::mutex> locker(*paramsPtr->m_mutexPtr );

        const double maxXOffset = paramsPtr->m_maxXOffset;
        const double maxYOffset = paramsPtr->m_maxYOffset;
        const double maxPercentDiff = paramsPtr->m_maxPercentDiff;
        const std::vector< te::gm::Geometry const* >& geometries1 =
          *paramsPtr->m_geometries1Ptr;
        const std::vector< te::gm::Geometry const* >& geometries2 =
          *paramsPtr->m_geometries2Ptr;
        std::vector< TiePointT >& tiePoints = *paramsPtr->m_tiePointsPtr;
        bool& returnValue = *paramsPtr->m_returnValuePtr;
        std::size_t& currGeometries1Idx = *paramsPtr->m_currGeometries1IdxPtr;
        const std::size_t geometries1Size = geometries1.size();
        te::common::TaskProgress& progress =*paramsPtr->m_progressPtr;

        // index

        te::sam::rtree::Index<std::size_t > geoms2Tree;

        const std::size_t geometries2Size = geometries2.size();

        for( std::size_t geometries2Idx = 0 ; geometries2Idx < geometries2Size ; ++geometries2Idx )
        {
          geoms2Tree.insert( *geometries2[ geometries2Idx ]->getMBR(), geometries2Idx );
        }

        locker.unlock();

        // MatchGeometries

        std::size_t geometries1Idx = 0;
        std::vector< std::size_t > treeResult;
        te::gm::Geometry const * geom1Ptr;
        te::gm::Geometry const * geom2Ptr;
        std::size_t treeResultSize = 0;
        std::size_t treeResultIdx = 0;
        te::gm::Coord2D centroid1;
        te::gm::Coord2D centroid2;
        double offX = 0;  // offX = centroid1.x - centroid2.x;
        double offY = 0;  // offY = centroid1.y - centroid2.y;
        double area1 = 0;
        double lengh1 = 0;
        double area2 = 0;
        double lengh2 = 0;
        double areaIntersection = 0;
        double lenghIntersection = 0;
        te::gm::GeomType type1 = te::gm::GeometryType;
        te::gm::GeomType type2 = te::gm::GeometryType;
        std::multimap< double, te::gm::Geometry const* > selectedGeoms2; // key = intersection area
        std::multimap< double, te::gm::Geometry const* >::iterator selectedGeoms2It;
        std::unique_ptr< te::gm::Geometry > transGeom2Ptr;
        std::unique_ptr< te::gm::Geometry > intersectionGeomPtr;
        TiePointT auxTiePoint;
        te::gm::Envelope env1Expanded;
        double diffPercentArea1And2 = 0;
        double diffPercentLenght1And2 = 0;
        double diffPercentArea1AndIntersect = 0;
        double diffPercentArea2AndIntersect = 0;
        double diffPercentLength1AndIntersect = 0;
        double diffPercentLength2AndIntersect = 0;
        double currDiffPercent = 0;

        while( true )
        {
          locker.lock();

          if( currGeometries1Idx >= geometries1Size ) break;
          geometries1Idx = currGeometries1Idx;
          ++currGeometries1Idx;

          progress.pulse();
          if( !progress.isActive() )
          {
            returnValue = false;
            return;
          }

          locker.unlock();

          geom1Ptr = geometries1[ geometries1Idx ];

          env1Expanded = *geom1Ptr->getMBR();
          env1Expanded.m_llx -= maxXOffset;
          env1Expanded.m_lly -= maxYOffset;
          env1Expanded.m_urx += maxXOffset;
          env1Expanded.m_ury += maxYOffset;

          centroid1 = geom1Ptr->getCentroid();
          type1 = geom1Ptr->getGeomTypeId();

          area1 = lengh1 = 0;
          try
          {
            GetGeomInfos( geom1Ptr, area1, lengh1 );
          }
          catch(...)
          {
            continue;
          }

          selectedGeoms2.clear();

          treeResult.clear();
          geoms2Tree.search( env1Expanded, treeResult );
          treeResultSize = treeResult.size();

          for( treeResultIdx = 0 ; treeResultIdx < treeResultSize ; ++treeResultIdx )
          {
            geom2Ptr = geometries2[ treeResult[ treeResultIdx] ];

            type2 = geom2Ptr->getGeomTypeId();

            // Checking both geometries types

            if( type1 != type2 ) continue;

            // Checking the max offsets between both geometries

            centroid2 = geom2Ptr->getCentroid();

            offX = centroid1.x - centroid2.x;
            offY = centroid1.y - centroid2.y;

            if(
                ( std::abs( offX ) > maxXOffset )
                ||
                ( std::abs( offY ) > maxYOffset )
              )
            {
              continue;
            }

            // Checking the area difference between both geometries

            area2 = lengh2 = 0;
            try
            {
              GetGeomInfos( geom2Ptr, area2, lengh2 );
            }
            catch(...)
            {
              continue;
            }

            if( ( area1 > 0 ) || ( area2 > 0 ) )
            {
              diffPercentArea1And2 =
                (
                  1.0
                  -
                  (
                    std::min( area1 , area2 )
                    /
                    std::max( area1 , area2 )
                  )
                );
            }
            else
            {
              diffPercentArea1And2 = 0;
            }

            if( diffPercentArea1And2 > maxPercentDiff )
            {
              continue;
            }

            if( ( lengh1 > 0 ) || ( lengh2 > 0 ) )
            {
              diffPercentLenght1And2 =
                (
                  1.0
                  -
                  (
                    std::min( lengh1 , lengh2 )
                    /
                    std::max( lengh1 , lengh2 )
                  )
                );
            }
            else
            {
              diffPercentLenght1And2 = 0;
            }

            if( diffPercentLenght1And2 > maxPercentDiff )
            {
              continue;
            }

            // checking the intersection area

            transGeom2Ptr.reset( (te::gm::Geometry*) geom2Ptr->clone() );
            if( !applyOffsets( transGeom2Ptr.get(), offX, offY ) ) continue;

            try
            {
              intersectionGeomPtr.reset( transGeom2Ptr->intersection( geom1Ptr ) );
            }
            catch(...)
            {
              continue;
            }
            if( intersectionGeomPtr.get() == nullptr ) continue;

            areaIntersection = lenghIntersection = 0;
            try
            {
              GetGeomInfos( intersectionGeomPtr.get(), areaIntersection, lenghIntersection );
            }
            catch(...)
            {
              continue;
            }

            if( ( areaIntersection == 0 ) && ( lenghIntersection == 0 ) )
              continue;

            if( ( areaIntersection > 0 ) || ( area1 > 0 ) )
            {
              diffPercentArea1AndIntersect =
                (
                  1.0
                  -
                  (
                    std::min( areaIntersection , area1 )
                    /
                    std::max( areaIntersection , area1 )
                  )
                );
            }
            else
            {
              diffPercentArea1AndIntersect = 0;
            }

            if( diffPercentArea1AndIntersect > maxPercentDiff )
            {
              continue;
            }

            if( ( areaIntersection > 0 ) || ( area2 > 0 ) )
            {
              diffPercentArea2AndIntersect =
                (
                  1.0
                  -
                  (
                    std::min( areaIntersection , area2 )
                    /
                    std::max( areaIntersection , area2 )
                  )
                );
            }
            else
            {
              diffPercentArea2AndIntersect = 0;
            }

            if( diffPercentArea2AndIntersect > maxPercentDiff )
            {
              continue;
            }

            if( ( areaIntersection > 0 ) || ( lengh1 > 0 ) )
            {
              diffPercentLength1AndIntersect =
                (
                  1.0
                  -
                  (
                    std::min( lenghIntersection , lengh1 )
                    /
                    std::max( lenghIntersection , lengh1 )
                  )
                );
            }
            else
            {
              diffPercentLength1AndIntersect = 0;
            }

            if( diffPercentLength1AndIntersect > maxPercentDiff )
            {
              continue;
            }

            if( ( areaIntersection > 0 ) || ( lengh2 > 0 ) )
            {
              diffPercentLength2AndIntersect =
                (
                  1.0
                  -
                  (
                    std::min( lenghIntersection , lengh2 )
                    /
                    std::max( lenghIntersection , lengh2 )
                  )
                );
            }
            else
            {
              diffPercentLength2AndIntersect = 0;
            }

            if( diffPercentLength2AndIntersect > maxPercentDiff )
            {
              continue;
            }

            currDiffPercent =
              std::max(
                diffPercentArea1And2,
                std::max(
                  diffPercentLenght1And2,
                  std::max(
                    diffPercentArea1AndIntersect,
                    std::max(
                      diffPercentArea2AndIntersect,
                      std::max(
                        diffPercentLength1AndIntersect,
                        diffPercentLength2AndIntersect
                      )
                    )
                  )
                )
              );

            selectedGeoms2.insert( std::pair< double, te::gm::Geometry const* >(
              currDiffPercent, geom2Ptr ) );
          }

          if( !selectedGeoms2.empty() )
          {
            selectedGeoms2It = selectedGeoms2.begin();

            std::get< 0 >( auxTiePoint ) = geom1Ptr->getSRID();
            std::get< 1 >( auxTiePoint ) = centroid1.x;
            std::get< 2 >( auxTiePoint ) = centroid1.y;
            std::get< 3 >( auxTiePoint ) = std::get< 0 >( auxTiePoint );
            std::get< 4 >( auxTiePoint ) = selectedGeoms2It->second->getCentroid().x;
            std::get< 5 >( auxTiePoint ) = selectedGeoms2It->second->getCentroid().y;

            locker.lock();

            tiePoints.push_back( auxTiePoint );

            locker.unlock();
          }
        }
      }

      TEVPEXPORT bool MatchGeometries(
        const std::vector< te::gm::Geometry const* >& geometries1,
        const std::vector< te::gm::Geometry const* >& geometries2,
        const bool enableThreads,
        const double maxXOffset,
        const double maxYOffset,
        const double maxPercentDiff,
        std::vector< TiePointT >& tiePoints )
      {
        tiePoints.clear();

        if( ( maxPercentDiff < 0 ) || ( maxPercentDiff > 1 ) )
        {
          return false;
        }

        if( geometries1.empty() ) return true;

        const int baseSRID = geometries1[ 0 ]->getSRID();

        std::vector< te::gm::Geometry const* > auxGeometries1Ptrs( geometries1 );
        std::vector< std::unique_ptr< te::gm::Geometry > > reprojGeometries1Handler;
        {
          const std::size_t geometriesSize = geometries1.size();
          te::gm::Geometry* reprojGeometryPtr = nullptr;

          for( std::size_t geometriesIdx = 0 ; geometriesIdx < geometriesSize ;
            ++geometriesIdx )
          {
            if( geometries1[ geometriesIdx ]->getSRID() != baseSRID )
            {
              reprojGeometryPtr = (te::gm::Geometry*)geometries1[
                geometriesIdx ]->clone();
              reprojGeometryPtr->transform( baseSRID );

              auxGeometries1Ptrs[ geometriesIdx ] = reprojGeometryPtr;
              reprojGeometries1Handler.push_back( std::unique_ptr< te::gm::Geometry >(
                reprojGeometryPtr ) );
            }
          }
        }

        std::vector< te::gm::Geometry const* > auxGeometries2Ptrs( geometries2 );
        std::vector< std::unique_ptr< te::gm::Geometry > > reprojGeometries2Handler;
        {
          const std::size_t geometriesSize = geometries2.size();
          te::gm::Geometry* reprojGeometryPtr = nullptr;

          for( std::size_t geometriesIdx = 0 ; geometriesIdx < geometriesSize ;
            ++geometriesIdx )
          {
            if( geometries2[ geometriesIdx ]->getSRID() != baseSRID )
            {
              reprojGeometryPtr = (te::gm::Geometry*)geometries2[
                geometriesIdx ]->clone();
              reprojGeometryPtr->transform( baseSRID );

              auxGeometries2Ptrs[ geometriesIdx ] = reprojGeometryPtr;
              reprojGeometries2Handler.push_back( std::unique_ptr< te::gm::Geometry >(
                reprojGeometryPtr ) );
            }
          }
        }

        bool returnValue = true;
        std::mutex mutex;
        std::size_t currGeometries1Idx = 0;

        te::common::TaskProgress progress( "Geometries processed",
          te::common::TaskProgress::UNDEFINED, geometries1.size() );
        progress.useMultiThread( enableThreads );

        MatchGeometriesThreadParams threadsParameters;
        threadsParameters.m_mutexPtr = &mutex;
        threadsParameters.m_maxXOffset = maxXOffset;
        threadsParameters.m_maxYOffset = maxYOffset;
        threadsParameters.m_maxPercentDiff = maxPercentDiff;
        threadsParameters.m_geometries1Ptr = &auxGeometries1Ptrs;
        threadsParameters.m_geometries2Ptr = &auxGeometries2Ptrs;
        threadsParameters.m_tiePointsPtr = &tiePoints;
        threadsParameters.m_returnValuePtr = &returnValue;
        threadsParameters.m_currGeometries1IdxPtr = &currGeometries1Idx;
        threadsParameters.m_progressPtr = &progress;

        if( enableThreads )
        {
          std::vector< std::shared_ptr< std::thread > > threadsVec;

          const unsigned int procsNumber = te::common::GetPhysProcNumber();

          for( unsigned int threadIdx = 0 ; threadIdx < procsNumber ; ++threadIdx )
          {
            threadsVec.push_back( std::shared_ptr< std::thread >(
              new std::thread( MatchGeometriesThreadEntry, &threadsParameters ) ) );
          }

          for( unsigned int threadIdx = 0 ; threadIdx < procsNumber ; ++threadIdx )
          {
            threadsVec[ threadIdx ]->join();
          }
        }
        else
        {
          MatchGeometriesThreadEntry( &threadsParameters );
        }

        return returnValue;
      }

      bool MatchDataSets(
        te::da::DataSet& dataset1,
        const te::da::DataSetType& dsType1,
        te::da::DataSet& dataset2,
        const te::da::DataSetType& dsType2,
        const bool enableThreads,
        const double maxXOffset,
        const double maxYOffset,
        const double maxPercentDiff,
        std::vector< TiePointT >& tiePoints )
      {
        int baseSRID = 0;
        {
          const std::size_t geomPropPos = te::da::GetFirstPropertyPos(
            &dataset1,
            te::dt::GEOMETRY_TYPE );

          dataset1.moveBeforeFirst();

          std::unique_ptr<te::gm::Geometry> inGeomPtr;

          if( !dataset1.moveNext() ) return false;

          inGeomPtr = dataset1.getGeometry( geomPropPos );

          if( inGeomPtr.get() == nullptr ) return false;

          baseSRID = inGeomPtr->getSRID();
        }

        std::vector< te::gm::Geometry const* > geometries1;
        {
          const std::size_t geomPropPos = te::da::GetFirstPropertyPos(
            &dataset1,
            te::dt::GEOMETRY_TYPE );

          dataset1.moveBeforeFirst();

          std::unique_ptr<te::gm::Geometry> inGeomPtr;

          while( dataset1.moveNext() )
          {
            inGeomPtr = dataset1.getGeometry( geomPropPos );

            if( inGeomPtr.get() != nullptr )
            {
              inGeomPtr->transform( baseSRID );

              geometries1.push_back( inGeomPtr.release() );
            }
          }
        }

        std::vector< te::gm::Geometry const* > geometries2;
        {
          const std::size_t geomPropPos = te::da::GetFirstPropertyPos(
            &dataset2,
            te::dt::GEOMETRY_TYPE );

          dataset2.moveBeforeFirst();

          std::unique_ptr<te::gm::Geometry> inGeomPtr;

          while( dataset2.moveNext() )
          {
            inGeomPtr = dataset2.getGeometry( geomPropPos );

            if( inGeomPtr.get() != nullptr )
            {
              inGeomPtr->transform( baseSRID );

              geometries2.push_back( inGeomPtr.release() );
            }
          }
        }

        bool returnValue = MatchGeometries( geometries1, geometries2, enableThreads,
          maxXOffset, maxYOffset, maxPercentDiff, tiePoints );

        std::size_t geometries1Size = geometries1.size();
        for( std::size_t geometries1Idx = 0 ; geometries1Idx < geometries1Size ;
          ++geometries1Idx )
        {
          delete geometries1[ geometries1Idx ];
        }

        std::size_t geometries2Size = geometries2.size();
        for( std::size_t geometries2Idx = 0 ; geometries2Idx < geometries2Size ;
          ++geometries2Idx )
        {
          delete geometries2[ geometries2Idx ];
        }

        return returnValue;
      }

    } // end namespace TiePointsLocator
  } // end namespace vp
}   // end namespace te

