/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

#ifndef __TERRALIB_VP_INTERNAL_TIEPOINTSLOCATOR_H
#define __TERRALIB_VP_INTERNAL_TIEPOINTSLOCATOR_H

#include "Config.h"
#include "../dataaccess/dataset/DataSet.h"
#include "../dataaccess/dataset/DataSetType.h"

#include<tuple>

namespace te
{
  namespace vp
  {
    /*!
      \namespace TiePointsLocator A namespace for functions to locate tie-points in vector data.
    */
    namespace TiePointsLocator
    {
      /*!
        \typedef TiePointT
        \brief Tie poit type definition [ SRID1, x1, y2, SRID2, x2, y2 ].
      */
      typedef  std::tuple< int, double, double, int, double, double > TiePointT;

      /*!
        \brief Match geometries creating tie-points for the mathed ones.
        \param geometries1 Input Geometries 1 vector.
        \param geometries2 INput Geometries 2 vector.
        \param enableThreads Enable/disable the use of threads.
        \param maxXOffset Maximum X offset betweeen Geometries 1 and Geometries 2.
        \param maxYOffset Maximum Y offset betweeen Geometries 1 and Geometries 2.
        \param maxPercentDiff Maximum percentual difference betweeen features of Geometries 1 and Geometries 2.
        \param tiePoints Output tie-points.
        \return true if ok, false on errors.
      */
      TEVPEXPORT bool MatchGeometries(
        const std::vector< te::gm::Geometry const* >& geometries1,
        const std::vector< te::gm::Geometry const* >& geometries2,
        const bool enableThreads,
        const double maxXOffset,
        const double maxYOffset,
        const double maxPercentDiff,
        std::vector< TiePointT >& tiePoints );

      /*!
        \brief Match geometries from the given datasets creating tie-points for the mathed ones.
        \param dataset1 Input dataset 1.
        \param dsType1 Input dataset type 1.
        \param dataset2 Input dataset 2.
        \param dsType2 Input dataset type 2.
        \param enableThreads Enable/disable the use of threads.
        \param maxXOffset Maximum X offset betweeen Geometries 1 and Geometries 2.
        \param maxYOffset Maximum Y offset betweeen Geometries 1 and Geometries 2.
        \param maxPercentDiff Maximum percentual difference betweeen features of Geometries 1 and Geometries 2.
        \param tiePoints Output tie-points.
        \return true if ok, false on errors.
      */
      TEVPEXPORT bool MatchDataSets(
        te::da::DataSet& dataset1,
        const te::da::DataSetType& dsType1,
        te::da::DataSet& dataset2,
        const te::da::DataSetType& dsType2,
        const bool enableThreads,
        const double maxXOffset,
        const double maxYOffset,
        const double maxPercentDiff,
        std::vector< TiePointT >& tiePoints );

    }  // end namespace TiePointsLocator
  } // end namespace vp
}   // end namespace te

#endif  // __TERRALIB_VP_INTERNAL_TIEPOINTSLOCATOR_H
