/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralin/vp/UnionOperation.cpp

   \brief Represents a union operation between two sets of features
 */
#include "UnionOperation.h"

#include "../geometry/Geometry.h"
#include "../geometry/Utils.h"
#include "../memory/PreparedItemCopy.h"
#include "../sam/rtree/Index.h"
#include "Feature.h"
#include "OperationService.h"
#include "IdentityOperation.h"
#include "DifferenceOperation.h"

namespace te
{
  namespace vp
  {
    namespace unionOperation
    {
      const std::string OPERATION_NAME = "Union";
    }
  }
}

namespace
{
  static te::vp::UnionOperationFactory
    UnionOperationFactoryInstance;
}

te::vp::UnionOperation::UnionOperation()
    : AbstractOperation(te::vp::unionOperation::OPERATION_NAME)
{

}

te::vp::UnionOperation::~UnionOperation() = default;

void te::vp::UnionOperation::initialize()
{
  AbstractOperation::initialize();

  //initializes the identity operation
  m_identityOperation.reset(new IdentityOperation());
  m_identityOperation->setParameters(new AbstractParameters(*this->getParameters()));
  m_identityOperation->initialize();

  //initializes the difference operation
  m_differenceOperation.reset(new DifferenceOperation());

  AbstractParameters* differenceParams = new AbstractParameters(*this->getParameters());
  differenceParams->setSwapInput(true);
  m_differenceOperation->setParameters(differenceParams);
  m_differenceOperation->initialize();
}

std::vector<te::vp::FeatureSet> te::vp::UnionOperation::executeImpl(const std::vector<te::vp::FeatureSet>& vecInput)
{
  std::vector<te::vp::FeatureSet> vecOutput;
  if (vecInput.size() != 2)
  {
    vecOutput.resize(1);
    return vecOutput;
  }

  te::vp::FeatureSet fullResult;

  //intersection (A & B) or (A - B)
  {
    std::vector<te::vp::FeatureSet> vecSingleOutput = m_identityOperation->executeImpl(vecInput);
    fullResult.insert(fullResult.end(), vecSingleOutput.at(0).begin(), vecSingleOutput.at(0).end());
  }
  
  //difference (B - A)
  {
    std::vector<te::vp::FeatureSet> vecSingleOutput = m_differenceOperation->executeImpl(vecInput);
    fullResult.insert(fullResult.end(), vecSingleOutput.at(0).begin(), vecSingleOutput.at(0).end());
  }

  //we must convert the geometries to the union operation output type

  te::mem::PreparedItemCopy copy;
  for (std::size_t i = 0; i < fullResult.size(); ++i)
  {
    te::vp::Feature* feature = fullResult.at(i);
    te::vp::Feature* convertedFeature = createOutputFeature(0);

    copy.copyValues(feature, convertedFeature, false);
    convertedFeature->setFirstGeometry(feature->getFirstGeometry().release());

    delete feature;
    fullResult[i] = convertedFeature;
  }

  vecOutput.push_back(fullResult);
  
  return vecOutput;
}

te::vp::UnionOperationFactory::UnionOperationFactory()
  : OperationFactory(te::vp::unionOperation::OPERATION_NAME)
{
}

te::vp::UnionOperationFactory::~UnionOperationFactory() = default;

te::vp::AbstractOperation* te::vp::UnionOperationFactory::build()
{
  return new UnionOperation();
}
