/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralin/vp/UnorderedGroupingDataHandler.cpp

   \brief This class represents handler to help reading and writing data from/to a dataSource by dividing it into smaller groups
 */

#include "UnorderedGroupingDataHandler.h"

#include "../common/STLUtils.h"
#include "../dataaccess/dataset/DataSet.h"
#include "../dataaccess/dataset/DataSetType.h"
#include "../dataaccess/utils/DataAccess.h"
#include "../dataaccess/utils/Utils.h"
#include "AbstractOperation.h"
#include "Feature.h"
#include "Utils.h"

 //Boost include files
#include <boost/thread/lock_guard.hpp>
#include <boost/thread/mutex.hpp>

namespace
{
  static te::vp::UnorderedGroupingDataHandlerFactory UnorderedGroupingDataHandlerFactoryInstance;
}

struct GroupInfo
{
  GroupInfo(te::da::DataSet* dataSet)
    : m_dataSet(dataSet)
  {
    m_dataSet->moveBeforeFirst();
    m_dataSetSize = m_dataSet->size();
  }

  std::size_t m_dataSetSize;
  std::unique_ptr<te::da::DataSet> m_dataSet;
};

class te::vp::UnorderedGroupingDataHandlerImpl
{
public:

  UnorderedGroupingDataHandlerImpl() 
  : m_chunckSize(0)
  , m_groupCount(0)
  {}

  ~UnorderedGroupingDataHandlerImpl()
  {
    te::common::FreeContents(m_groupInfo);
  }
  
  std::size_t m_chunckSize;
  std::size_t m_groupCount;
  std::vector<GroupInfo*> m_groupInfo;
  boost::mutex m_mutexRead;
  boost::mutex m_mutexWrite;
};

te::vp::UnorderedGroupingDataHandler::UnorderedGroupingDataHandler()
    : AbstractDataHandler()
    , m_impl(new UnorderedGroupingDataHandlerImpl())
{

}

te::vp::UnorderedGroupingDataHandler::~UnorderedGroupingDataHandler() = default;

void te::vp::UnorderedGroupingDataHandler::initialize(te::vp::AbstractParameters* absParams, te::common::TaskProgress* taskProgress)
{
  AbstractDataHandler::initialize(absParams, taskProgress);

  std::size_t totalDataSize = 0;
  for (std::size_t i = 0; i < m_vecInputData.size(); ++i)
  {
    GroupInfo* info = new GroupInfo(m_vecInputData.at(i)->getDataSet().release());

    totalDataSize += info->m_dataSetSize;

    m_impl->m_groupInfo.push_back(info);
  }

  double idealChuckFraction = static_cast<double>(totalDataSize) / 100.;
  if (idealChuckFraction < 10)
  {
    idealChuckFraction = 10;
  }

  double idealGroupCount = static_cast<double>(totalDataSize) / idealChuckFraction;
  
  m_impl->m_chunckSize = static_cast<std::size_t>(std::ceil(idealChuckFraction));
  m_impl->m_groupCount = static_cast<std::size_t>(std::ceil(idealGroupCount));

  //if we get any problem related to the calculation, we define 1 group for all the data
  if (m_impl->m_chunckSize == 0 || m_impl->m_groupCount == 0)
  {
    m_impl->m_chunckSize = std::string::npos;
    m_impl->m_groupCount = 1;
  }
}

void te::vp::UnorderedGroupingDataHandler::finalize(te::common::TaskProgress* taskProgress)
{

}

std::size_t te::vp::UnorderedGroupingDataHandler::getGroupCount() const
{
  return m_impl->m_groupCount;
}


te::vp::FeatureSet te::vp::UnorderedGroupingDataHandler::getFeatureSet(std::size_t groupIndex, std::size_t inputIndex, const te::gm::Envelope& filter)
{
  boost::mutex::scoped_lock lockGuard(m_impl->m_mutexRead);

  GroupInfo* groupInfo = m_impl->m_groupInfo.at(inputIndex);

  te::vp::FeatureSet featureSet;
  while (groupInfo->m_dataSet->moveNext())
  {
    std::unique_ptr<te::vp::Feature> feature(te::vp::Feature::createFeatureFromCurrent(groupInfo->m_dataSet.get()));
    featureSet.push_back(feature.release());

    if (featureSet.size() == m_impl->m_chunckSize)
    {
      break;
    }
  }

  return featureSet;
}

void te::vp::UnorderedGroupingDataHandler::addFeatureSet(std::size_t groupIndex, std::size_t outputIndex, const te::vp::FeatureSet& featureSet)
{
  boost::mutex::scoped_lock lockGuard(m_impl->m_mutexWrite);

  te::da::DataAccess* dataAccess = m_vecOutputData.at(outputIndex);

  std::unique_ptr<te::da::DataSet> dataSet = te::vp::CreateDataSetFromFeatureSet(dataAccess->getDataSetName(), dataAccess->getDataSetType(), featureSet);
  dataAccess->saveDataSet(dataSet.release());
}

te::vp::UnorderedGroupingDataHandlerFactory::UnorderedGroupingDataHandlerFactory()
  : te::vp::AbstractDataHandlerFactory("UnorderedGroupingDataHandler")
{

}

te::vp::UnorderedGroupingDataHandlerFactory::~UnorderedGroupingDataHandlerFactory() = default;

te::vp::AbstractDataHandler* te::vp::UnorderedGroupingDataHandlerFactory::build()
{
  return new te::vp::UnorderedGroupingDataHandler();
}