/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/vp/Utils.h

  \brief Utility functions for Vector Processing.  
*/

#ifndef __TERRALIB_VP_INTERNAL_UTILS_H
#define __TERRALIB_VP_INTERNAL_UTILS_H

// Terralib
#include "../dataaccess/datasource/DataSource.h"

#include "../geometry/CommonDataStructures.h"
#include "../geometry/Enums.h"

#include "../sam/rtree.h"

#include "Config.h"
#include "CommonDataStructures.h"

//STL
#include <map>
#include <string>

namespace te
{
  namespace da
  {
    class DataSet;
    class DataSetType;
  }

  namespace gm
  {
    class Geometry;
    class GeometryCollection;
  }

  namespace mem
  {
    class DataSetItem;
  }

  namespace vp
  {
    class AlgorithmParams;
    class AbstractDataHandler;
    class AbstractOperation;
    class FeatureSet;

    enum Strategy
    {
      MEMORY,
      QUERY
    };

    /*!
      \brief It returns the union of a geometry vector.

      \param items  Vector of itens that represents a group.

      \return Union of the geometry.
    */
    TEVPEXPORT te::gm::Geometry* GetGeometryUnion(const std::vector<te::mem::DataSetItem*>& items, size_t geomIdx, te::gm::GeomType outGeoType);

    TEVPEXPORT te::gm::Geometry* GetGeometryUnion(const std::vector<te::mem::DataSetItem*>& items, size_t geomIdx);

    TEVPEXPORT void SplitGeometryCollection(te::gm::GeometryCollection* geomIn, te::gm::GeometryCollection* gcOut);

    TEVPEXPORT std::string GetSimpleTableName(std::string fullName);

    TEVPEXPORT std::unique_ptr<te::da::DataSet> PrepareAdd(te::da::DataSet* ds, te::da::DataSetType* dt);

    TEVPEXPORT void Save(te::da::DataSource* source, te::da::DataSet* result, te::da::DataSetType* outDsType, const bool& enableProgress = true);

    TEVPEXPORT te::da::DataSourcePtr CreateOGRDataSource(std::string repository);

    TEVPEXPORT void ValidateAlgorithmParams(AlgorithmParams* mainParams, Strategy st);

    TEVPEXPORT te::sam::rtree::Index<size_t, 8>* GetRtree(te::da::DataSet* data);

    TEVPEXPORT te::gm::Geometry* SetGeomAsMulti(const te::gm::Geometry& geom);

    TEVPEXPORT std::string GetDistinctName(const std::string& name, std::vector<std::string> names, std::size_t maxSize = 0);

    TEVPEXPORT bool IsPointType(const te::gm::GeomType& geomType);

    TEVPEXPORT bool IsLineStringType(const te::gm::GeomType& geomType);

    TEVPEXPORT bool IsPolygonType(const te::gm::GeomType& geomType);

    /*!
      \brief It returns a memory item with all properties filled with values
      from the current position dataset.

      \param dataSet      DataSet in the current position to get values.

      \return A memory item with attributes filled with dataSet values.
    */
    TEVPEXPORT te::mem::DataSetItem* GetFilledItemMemory(te::da::DataSet* dataSet);

    /*!
      \brief Creates a memory dataSet based on the given featureSet

      \param dataSetName The name of the dataSet
      \param dataSetType The dataSetType of the dataSet
      \param featureSet The featureSet to be converted

      \return A memory dataSet containing all the given features
    */
    TEVPEXPORT std::unique_ptr<te::da::DataSet> CreateDataSetFromFeatureSet(const std::string& dataSetName, const te::da::DataSetType* dataSetType, const te::vp::FeatureSet& featureSet);

    //!< Extracts the GeometryVector from the given FeatureSet. The caller must free the memory after using the resulting geometry vector
    TEVPEXPORT te::gm::GeometryVector ExtractGeometriesFromFeatureSet(const te::vp::FeatureSet& featureSet);

    //!< Gets  the GeometryVector from the given FeatureSet. The caller must NOT free the memory after using the resulting geometry vector as this method returns the reference to the geometries
    TEVPEXPORT te::gm::GeometryVectorConst GetGeometriesFromFeatureSet(const te::vp::FeatureSet& featureSet);

    /*!
      \brief Creates a data source located in the temporary folder of the user filesystem
      \return A a data source located in the temporary folder of the user filesystem
    */
    TEVPEXPORT te::da::DataSourcePtr CreateTemporaryDataSource();

    TEVPEXPORT void DropTemporaryDataSource(te::da::DataSourcePtr dataSource);

    TEVPEXPORT te::vp::AbstractOperation* CreateOperation(const std::string& operationName);

    TEVPEXPORT te::vp::AbstractDataHandler* CreateDataHandler(const std::string& dataHandlerName);

  } // end namespace vp
}   // end namespace te

#endif  // __TERRALIB_VP_INTERNAL_UTILS_H
