/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

This file is part of the TerraLib - a Framework for building GIS enabled applications.

TerraLib is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

TerraLib is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with TerraLib. See COPYING. If not, write to
TerraLib Team at <terralib-team@terralib.org>.
*/

/*!
 * \file src/terralib/vp/qt/DissolveSubdivisionParametersWidget.cpp
 *
 * \brief Implements a widget used to return the specific parameters for dissolve subdivision operation.
 */

#include "DissolveSubdivisionParametersWidget.h"
#include "ui/ui_DissolveSubdivisionParametersWidgetForm.h"

#include "../../common/Exception.h"
#include "../../dataaccess/dataset/DataSetType.h"
#include "../../dataaccess/utils/DataAccess.h"
#include "../../qt/widgets/Utils.h"
#include "../SubdivideByBorderTillingStrategy.h"
#include "../DissolveSubdivisionOperation.h"


te::vp::qt::DissolveSubdivisionParametersWidget::DissolveSubdivisionParametersWidget(QWidget* parent, Qt::WindowFlags f)
  : QWidget(parent, f)
  , te::vp::qt::SpecificParametersWidget()
  , m_ui(new Ui::DissolveSubdivisionParametersWidgetForm)
{
  m_ui->setupUi(this);
}

te::vp::qt::DissolveSubdivisionParametersWidget::~DissolveSubdivisionParametersWidget() = default;

void te::vp::qt::DissolveSubdivisionParametersWidget::setInputData(const std::vector<te::da::DataAccessPtr>& vecDataAccess)
{
  if (vecDataAccess.size() != 1)
  {
    return;
  }

  m_ui->m_rowPropertyNameComboBox->clear();
  m_ui->m_columnPropertyNameComboBox->clear();

  const te::da::DataSetType* dataSetType = vecDataAccess.at(0)->getDataSetType();

  std::size_t defaultRowPropertyIndex = std::string::npos;
  std::size_t defaultColumnPropertyIndex = std::string::npos;
  for (std::size_t i = 0; i < dataSetType->size(); ++i)
  {
    te::dt::Property* property = dataSetType->getProperty(i);
    if (property->getType() != te::dt::INT16_TYPE && property->getType() != te::dt::INT32_TYPE)
    {
      continue;
    }

    const std::string& propertyName = property->getName();
    if (propertyName == te::vp::SubdivideByBorderTillingStrategy::getRowPropertyName())
    {
      defaultRowPropertyIndex = i;
    }
    if (propertyName == te::vp::SubdivideByBorderTillingStrategy::getColumnPropertyName())
    {
      defaultColumnPropertyIndex = i;
    }

    QString qPropertyName = te::qt::widgets::Convert2QString(propertyName);

    m_ui->m_rowPropertyNameComboBox->addItem(qPropertyName);
    m_ui->m_columnPropertyNameComboBox->addItem(qPropertyName);
  }

  if (defaultRowPropertyIndex != std::string::npos)
  {
    m_ui->m_rowPropertyNameComboBox->setCurrentIndex(static_cast<int>(defaultRowPropertyIndex));
  }

  if (defaultColumnPropertyIndex != std::string::npos)
  {
    m_ui->m_columnPropertyNameComboBox->setCurrentIndex(static_cast<int>(defaultColumnPropertyIndex));
  }
}

QWidget* te::vp::qt::DissolveSubdivisionParametersWidget::getWidget()
{
  return this;
}

void te::vp::qt::DissolveSubdivisionParametersWidget::setValuesToParameters(te::vp::AbstractParameters* parameters)
{
  te::vp::DissolveSubdivisionOperationParameters* specificParameters = dynamic_cast<te::vp::DissolveSubdivisionOperationParameters*>(parameters);
  if (specificParameters == nullptr)
  {
    throw te::common::Exception(TE_TR("Invalid parameter type. Call the developer"));
  }

  if(m_ui->m_rowPropertyNameComboBox->currentText().isEmpty() == true)
  {
    throw te::common::Exception(TE_TR("Please select a valid row property name"));
  }

  if (m_ui->m_columnPropertyNameComboBox->currentText().isEmpty() == true)
  {
    throw te::common::Exception(TE_TR("Please select a valid column property name"));
  }

  std::string rowPropertyname = te::qt::widgets::Convert2StdString(m_ui->m_rowPropertyNameComboBox->currentText());
  std::string columnPropertyname = te::qt::widgets::Convert2StdString(m_ui->m_columnPropertyNameComboBox->currentText());

  if (rowPropertyname == columnPropertyname)
  {
    throw te::common::Exception(TE_TR("Row property name and column property name cannot be the same"));
  }

  bool dissolveMaxCoordinates = m_ui->m_dissolveMaxCoordinatesCheckBox->isChecked();

  specificParameters->setRowPropertyName(rowPropertyname);
  specificParameters->setColumnPropertyName(columnPropertyname);
  specificParameters->setDissolveMaxCoordinates(dissolveMaxCoordinates);
}