/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/vp/qt/IdentityDialog.cpp

  \brief Identity operation dialog.
*/

// TerraLib
#include "../../core/filesystem/FileSystem.h"
#include "../../core/logger/Logger.h"
#include "../../core/translator/Translator.h"
#include "../../common/progress/ProgressManager.h"
#include "../../common/StringUtils.h"

#include "../../dataaccess/dataset/DataSet.h"
#include "../../dataaccess/dataset/DataSetType.h"

#include "../../dataaccess/datasource/DataSourceCapabilities.h"
#include "../../dataaccess/datasource/DataSourceInfo.h"
#include "../../dataaccess/datasource/DataSourceInfoManager.h"
#include "../../dataaccess/datasource/DataSourceManager.h"
#include "../../dataaccess/datasource/DataSourceFactory.h"

#include "../../dataaccess/utils/Utils.h"

#include "../../datatype/Property.h"
#include "../../datatype/SimpleData.h"

#include "../../geometry/GeometryProperty.h"

#include "../../qt/widgets/datasource/selector/DataSourceSelectorDialog.h"
#include "../../qt/widgets/layer/utils/DataSet2Layer.h"
#include "../../qt/widgets/progress/ProgressViewerDialog.h"
#include "../../qt/widgets/utils/DoubleInputListWidget.h"
#include "../../qt/widgets/utils/FileDialog.h"

#include "../ComplexData.h"
#include "../Exception.h"
#include "../Identity.h"
#include "../Utils.h"

#include "IdentityDialog.h"
#include "ui/ui_IdentityDialogForm.h"
#include "Utils.h"

// Qt
#include <QFileDialog>
#include <QGridLayout>
#include <QMessageBox>

// BOOST
#include <boost/filesystem.hpp>
#include <boost/uuid/random_generator.hpp>
#include <boost/uuid/uuid_io.hpp>

Q_DECLARE_METATYPE(te::map::AbstractLayerPtr)

te::vp::IdentityDialog::IdentityDialog(QWidget* parent, Qt::WindowFlags f)
  : QDialog(parent, f),
    m_ui(new Ui::IdentityDialogForm),
    m_layers(std::list<te::map::AbstractLayerPtr>())
{
  // add controls
  m_ui->setupUi(this);

  m_ui->m_imgLabel->setPixmap(
      QIcon::fromTheme("vp-identity-hint").pixmap(112, 48));
  m_ui->m_targetDatasourceToolButton->setIcon(QIcon::fromTheme("datasource"));

  //add double list widget to this form
  m_inputDoubleInputListWidget.reset(new te::qt::widgets::DoubleInputListWidget(
      m_ui->m_inputSelectionGroupBox));
  m_inputDoubleInputListWidget->setFirstLayerLabel("");
  m_inputDoubleInputListWidget->setSecondLayerLabel("");
  m_inputDoubleInputListWidget->setOutputLabel("Output attributes");

  QGridLayout* inputLayout = new QGridLayout(m_ui->m_inputSelectionGroupBox);
  inputLayout->addWidget(m_inputDoubleInputListWidget.get());
  inputLayout->setContentsMargins(0, 0, 0, 0);

  connect(m_ui->m_inputLayerComboBox, SIGNAL(currentIndexChanged(int)), this,
          SLOT(onInputLayerComboBoxChanged(int)));
  connect(m_ui->m_intersectionLayerComboBox, SIGNAL(currentIndexChanged(int)),
          this, SLOT(onIntersectionLayerComboBoxChanged(int)));
  connect(m_ui->m_targetDatasourceToolButton, SIGNAL(pressed()), this,
          SLOT(onTargetDatasourceToolButtonPressed()));
  connect(m_ui->m_targetFileToolButton, SIGNAL(pressed()), this,
          SLOT(onTargetFileToolButtonPressed()));
  connect(m_ui->m_okPushButton, SIGNAL(clicked()), this,
          SLOT(onOkPushButtonClicked()));
  connect(m_ui->m_cancelPushButton, SIGNAL(clicked()), this,
          SLOT(onCancelPushButtonClicked()));

  m_ui->m_helpPushButton->setNameSpace("dpi.inpe.br.plugins");
  m_ui->m_helpPushButton->setPageReference("plugins/vp/vp_identity.html");
}

te::vp::IdentityDialog::~IdentityDialog() = default;

void te::vp::IdentityDialog::setLayers(
    std::list<te::map::AbstractLayerPtr> layers)
{
  std::list<te::map::AbstractLayerPtr>::iterator it = layers.begin();

  while (it != layers.end())
  {
    std::unique_ptr<te::da::DataSetType> dsType = it->get()->getSchema();
    if (dsType->hasGeom())
    {
      m_layers.push_back(*it);
    }

    ++it;
  }

  updateInputLayerComboBox();

  updateIntersectionLayerComboBox();

  updateDoubleListWidget();
}

te::map::AbstractLayerPtr te::vp::IdentityDialog::getLayer()
{
  return m_layerResult;
}

std::vector<std::string> te::vp::IdentityDialog::getWarnings()
{
  return m_warnings;
}

std::map<int, std::vector<std::string> >
te::vp::IdentityDialog::getSelectedProperties()
{
  std::map<int, std::vector<std::string> > propertiesMap;

  std::vector<std::string> propertiesName =
      m_inputDoubleInputListWidget->getOutputValues();

  std::vector<int> layerIds =
      m_inputDoubleInputListWidget->getOutputDataValues();

  if(propertiesName.size() != layerIds.size())
    return propertiesMap;

  std::vector<std::string> inputProperties;
  std::vector<std::string> intersectionProperties;

  for (std::size_t i = 0; i < layerIds.size(); ++i)
  {
    if(layerIds[i] == 0)
      inputProperties.push_back(propertiesName[i]);
    else
      intersectionProperties.push_back(propertiesName[i]);
  }

  if(!inputProperties.empty())
    propertiesMap.insert(
        std::pair<int, std::vector<std::string> >(0, inputProperties));

  if(!intersectionProperties.empty())
    propertiesMap.insert(
        std::pair<int, std::vector<std::string> >(1, intersectionProperties));

  return propertiesMap;
}

void te::vp::IdentityDialog::updateInputLayerComboBox()
{
  std::list<te::map::AbstractLayerPtr>::iterator it = m_layers.begin();

  disconnect(m_ui->m_inputLayerComboBox, SIGNAL(currentIndexChanged(int)), this,
             SLOT(onInputLayerComboBoxChanged(int)));

  while (it != m_layers.end())
  {
    m_ui->m_inputLayerComboBox->addItem(QString(it->get()->getTitle().c_str()),
                                        QVariant::fromValue(*it));
    ++it;
  }

  connect(m_ui->m_inputLayerComboBox, SIGNAL(currentIndexChanged(int)), this,
          SLOT(onInputLayerComboBoxChanged(int)));

  QVariant varLayer = m_ui->m_inputLayerComboBox->itemData(
      m_ui->m_inputLayerComboBox->currentIndex(), Qt::UserRole);
  te::map::AbstractLayerPtr layer = varLayer.value<te::map::AbstractLayerPtr>();

  m_inputSelectedLayer = layer;
}

void te::vp::IdentityDialog::updateIntersectionLayerComboBox()
{
  int currIndex = m_ui->m_inputLayerComboBox->currentIndex();

  std::list<te::map::AbstractLayerPtr>::iterator it = m_layers.begin();

  disconnect(m_ui->m_intersectionLayerComboBox,
             SIGNAL(currentIndexChanged(int)), this,
             SLOT(onIntersectionLayerComboBoxChanged(int)));

  while (it != m_layers.end())
  {
    m_ui->m_intersectionLayerComboBox->addItem(
        QString(it->get()->getTitle().c_str()), QVariant::fromValue(*it));
    ++it;
  }

  m_ui->m_intersectionLayerComboBox->removeItem(currIndex);

  connect(m_ui->m_intersectionLayerComboBox, SIGNAL(currentIndexChanged(int)),
          this, SLOT(onIntersectionLayerComboBoxChanged(int)));

  QVariant varLayer = m_ui->m_intersectionLayerComboBox->itemData(
      m_ui->m_intersectionLayerComboBox->currentIndex(), Qt::UserRole);
  te::map::AbstractLayerPtr layer = varLayer.value<te::map::AbstractLayerPtr>();

  m_intersectionSelectedLayer = layer;
}

void te::vp::IdentityDialog::updateDoubleListWidget()
{
  std::vector<std::string> values;

  std::unique_ptr<te::da::DataSetType> inputSchema;
  std::unique_ptr<te::da::DataSetType> intersectionSchema;

  std::vector<te::dt::Property*> inputProps;
  std::vector<te::dt::Property*> intersectionProps;

  std::string inputName;
  std::string intersectionName;

  if (m_inputSelectedLayer)
  {
    inputSchema = m_inputSelectedLayer->getSchema();
    inputProps = inputSchema->getProperties();
    inputName = m_inputSelectedLayer->getTitle();
    m_inputDoubleInputListWidget->setFirstLayerLabel(
        QString::fromUtf8(inputName.c_str()));
  }
  if (m_intersectionSelectedLayer)
  {
    intersectionSchema = m_intersectionSelectedLayer->getSchema();
    intersectionProps = intersectionSchema->getProperties();
    intersectionName = m_intersectionSelectedLayer->getTitle();
    m_inputDoubleInputListWidget->setSecondLayerLabel(
        QString::fromUtf8(intersectionName.c_str()));
  }

  values.clear();
  for (std::size_t i = 0; i < inputProps.size(); ++i)
  {
    if (inputProps[i]->getType() != te::dt::GEOMETRY_TYPE)
      values.push_back(inputProps[i]->getName());
  }
  m_inputDoubleInputListWidget->setFirstInputValues(values);

  values.clear();
  for (std::size_t i = 0; i < intersectionProps.size(); ++i)
  {
    if (intersectionProps[i]->getType() != te::dt::GEOMETRY_TYPE)
      values.push_back(intersectionProps[i]->getName());
  }
  m_inputDoubleInputListWidget->setSecondInputValues(values);
}

void te::vp::IdentityDialog::onInputLayerComboBoxChanged(int index)
{
  QVariant varLayer = m_ui->m_inputLayerComboBox->itemData(index, Qt::UserRole);
  te::map::AbstractLayerPtr layer = varLayer.value<te::map::AbstractLayerPtr>();

  m_ui->m_intersectionLayerComboBox->clear();
  m_inputDoubleInputListWidget->clearOutputValues();
  
  m_inputSelectedLayer = layer;

  updateIntersectionLayerComboBox();

  updateDoubleListWidget();
}

void te::vp::IdentityDialog::onIntersectionLayerComboBoxChanged(int index)
{
  QVariant varLayer =
      m_ui->m_intersectionLayerComboBox->itemData(index, Qt::UserRole);
  te::map::AbstractLayerPtr layer = varLayer.value<te::map::AbstractLayerPtr>();

  m_inputDoubleInputListWidget->clearOutputValues();

  m_intersectionSelectedLayer = layer;

  updateDoubleListWidget();
}

void te::vp::IdentityDialog::onTargetDatasourceToolButtonPressed()
{
  m_ui->m_newLayerNameLineEdit->clear();
  m_ui->m_newLayerNameLineEdit->setEnabled(true);
  te::qt::widgets::DataSourceSelectorDialog dlg(this);
  dlg.exec();

  std::list<te::da::DataSourceInfoPtr> dsPtrList = dlg.getSelecteds();

  if (dsPtrList.empty())
    return;

  std::list<te::da::DataSourceInfoPtr>::iterator it = dsPtrList.begin();

  m_ui->m_repositoryLineEdit->setText(QString(it->get()->getTitle().c_str()));

  m_outputDatasource = *it;

  m_toFile = false;
}

void te::vp::IdentityDialog::onTargetFileToolButtonPressed()
{
  m_ui->m_newLayerNameLineEdit->clear();
  m_ui->m_repositoryLineEdit->clear();

  te::qt::widgets::FileDialog fileDialog(this, te::qt::widgets::FileDialog::VECTOR);

  try {
    fileDialog.exec();
  }
  catch (te::common::Exception& ex) {
    QMessageBox::warning(this, tr("File information"), ex.what());
    return;
  }

  m_ui->m_repositoryLineEdit->setText(fileDialog.getPath().c_str());
  m_ui->m_newLayerNameLineEdit->setText(fileDialog.getFileName().c_str());

  m_toFile = true;
  m_ui->m_newLayerNameLineEdit->setEnabled(false);
}

void te::vp::IdentityDialog::onOkPushButtonClicked()
{
  // Validate Input Layer.
  if (m_ui->m_inputLayerComboBox->currentText().isEmpty())
  {
    QMessageBox::warning(this, tr("Identity"),
                         tr("Select an input layer."));
    return;
  }

  // Validate DataSource.
  te::da::DataSourcePtr inputDataSource =
      te::da::GetDataSource(m_inputSelectedLayer->getDataSourceId(), true);
  if (!inputDataSource.get())
  {
    QMessageBox::information(
        this, "Identity",
        "The selected input data source can not be accessed.");
    return;
  }

  // Validate Identity Layer.
  if (m_ui->m_intersectionLayerComboBox->currentText().isEmpty())
  {
    QMessageBox::warning(this, tr("Identity"),
                         tr("Select a layer to execute the operation."));
    return;
  }

  // Validate DataSource.
  te::da::DataSourcePtr intersectionDataSource = te::da::GetDataSource(
      m_intersectionSelectedLayer->getDataSourceId(), true);
  if (!intersectionDataSource.get())
  {
    QMessageBox::information(
        this, "Identity",
        "The selected intersection data source can not be accessed.");
    return;
  }

  // Get output attributes.
  std::map<int, std::vector<std::string> > attributesMap =
      this->getSelectedProperties();

  std::map<std::string, te::dt::AbstractData*> specificParams;

  if (!attributesMap.empty())
  {
    std::map<int, std::vector<std::string> >::iterator it;

    it = attributesMap.find(0);
    if(it != attributesMap.end())
    {
      std::unique_ptr<te::da::DataSetType> inputSchema =
          m_inputSelectedLayer->getSchema();

      specificParams[inputSchema->getName()] =
          new te::vp::ComplexData<std::vector<std::string> >(it->second);
    }

    it = attributesMap.find(1);
    if(it != attributesMap.end())
    {
      std::unique_ptr<te::da::DataSetType> intersectionSchema =
          m_intersectionSelectedLayer->getSchema();

      specificParams[intersectionSchema->getName()] =
          new te::vp::ComplexData<std::vector<std::string> >(it->second);
    }
  }

  // Validade output repository.
  if(m_ui->m_repositoryLineEdit->text().isEmpty())
  {
    QMessageBox::warning(this, tr("Identity"),
                         tr("Select a repository for the resulting layer."));
    return;
  }
  
  if(m_ui->m_newLayerNameLineEdit->text().isEmpty())
  {
    QMessageBox::warning(this, tr("Identity"),
                         tr("Define a name for the resulting layer."));
    return;
  }

  if(m_inputSelectedLayer->getSRID() == TE_UNKNOWN_SRS ||
    m_intersectionSelectedLayer->getSRID() == TE_UNKNOWN_SRS)
  {
    int ret = QMessageBox::question(
        this, tr("Identity"),
        tr("The two layers have incompatible SRS. The result might be "
           "incorrect. Do you wish to continue?"),
        QMessageBox::No, QMessageBox::Yes);
    if (ret == QMessageBox::No)
      return;
  }

  // Verify if "Input Only Selected objects" is checked.
  bool inputIsChecked = false;

  if (m_ui->m_inputOnlySelectedCheckBox->isChecked())
    inputIsChecked = true;

  // Verify if "Intersection only selected objects" is checked.
  bool intersectionIsChecked = false;

  if (m_ui->m_diffOnlySelectedCheckBox->isChecked())
    intersectionIsChecked = true;


  // Progress
  te::qt::widgets::ProgressViewerDialog v(this);
  
  try
  {
    // Declare the input parameters
    te::vp::InputParams structInputParams;
    te::vp::InputParams structIntersectionParams;

    // Set the inputLayer parameters
    structInputParams.m_inputDataSource = inputDataSource;
    structInputParams.m_inputDataSetType =
        m_inputSelectedLayer->getSchema().release();

    // Set the intersectionLayer parameters
    structIntersectionParams.m_inputDataSource = intersectionDataSource;
    structIntersectionParams.m_inputDataSetType =
        m_intersectionSelectedLayer->getSchema().release();

    int inputSRID = m_inputSelectedLayer->getSRID();
    int intersectionSRID = m_intersectionSelectedLayer->getSRID();


    // Get DataSet and DataSetType using AbstractLayerPtr to process by
    // memory, using GEOS.
    te::vp::DataStruct inputData = te::vp::GetDataStructFromLayer(
        m_inputSelectedLayer, inputIsChecked, inputSRID);

    if (inputData.m_dataSet)
      structInputParams.m_inputDataSet = inputData.m_dataSet;

    if (inputData.m_dataSetType)
      structInputParams.m_inputDataSetType = inputData.m_dataSetType;

    te::vp::DataStruct intersectionData = te::vp::GetDataStructFromLayer(
        m_intersectionSelectedLayer, intersectionIsChecked, intersectionSRID);

    if (intersectionData.m_dataSet)
      structIntersectionParams.m_inputDataSet = intersectionData.m_dataSet;

    if (intersectionData.m_dataSetType)
      structIntersectionParams.m_inputDataSetType =
          intersectionData.m_dataSetType;

    m_inputParams.push_back(structInputParams);
    m_inputParams.push_back(structIntersectionParams);

    // Get the output dataset name.
    std::string outputdataset =
        m_ui->m_newLayerNameLineEdit->text().toUtf8().data();

    // Return of operation result.
    bool res = true;
    
    if (m_toFile)
    {
      boost::filesystem::path uri(
          m_ui->m_repositoryLineEdit->text().toUtf8().data());

      if (te::core::FileSystem::exists(uri.string()))
      {
        QMessageBox::information(
            this, tr("Identity"),
            tr("Output file already exists. Remove it and try again. "));

        return;
      }

      std::size_t idx = outputdataset.find(".");
      if(idx != std::string::npos)
        outputdataset = outputdataset.substr(0, idx);

      std::string dsinfo("file://");
      dsinfo += uri.string();

      te::da::DataSourcePtr dsOGR(
          te::da::DataSourceFactory::make("OGR", dsinfo).release());

      dsOGR->open();

      this->setCursor(Qt::WaitCursor);

      // Set parameters (Input/Output).
      m_params = new te::vp::AlgorithmParams();
      m_params->setInputParams(m_inputParams);
      m_params->setOutputDataSource(dsOGR);
      m_params->setOutputDataSetName(outputdataset);
      m_params->setSpecificParams(specificParams);

      te::vp::Identity identity;

      res = identity.executeMemory(m_params);

      delete m_params;

      if(!res)
      {
        dsOGR->close();

        QMessageBox::information(
            this, tr("Identity"),
            tr("Error: could not generate the difference."));

        reject();
      }

      // let's include the new datasource in the managers
      boost::uuids::basic_random_generator<boost::mt19937> gen;
      boost::uuids::uuid u = gen();
      std::string ds_id = boost::uuids::to_string(u);

      te::da::DataSourceInfoPtr ds(new te::da::DataSourceInfo);
      ds->setConnInfo(dsinfo);
      ds->setTitle(uri.stem().string());
      ds->setAccessDriver("OGR");
      ds->setType("OGR");
      ds->setDescription(uri.string());
      ds->setId(ds_id);

      te::da::DataSourcePtr newds =
          te::da::DataSourceManager::getInstance().get(ds_id, "OGR",
                                                       ds->getConnInfo());

      newds->open();
      te::da::DataSourceInfoManager::getInstance().add(ds);
      m_outputDatasource = ds;

      if (!m_outputDatasource)
      {
        QMessageBox::information(
            this, tr("Identity"),
            tr("The output data source can not be accessed."));

        return;
      }
    }
    else
    {
      te::da::DataSourcePtr aux =
          te::da::GetDataSource(m_outputDatasource->getId());

      if (!aux.get())
      {
        QMessageBox::information(
            this, tr("Identity"),
            tr("The output data source can not be accessed."));

        return;
      }

      std::string name = te::common::Convert2LCase(outputdataset);

      if (aux->dataSetExists(name))
      {
        QMessageBox::information(this, tr("Identity"),
                                 tr("Dataset already exists. Remove it or "
                                    "select a new name and try again."));

        return;
      }

      this->setCursor(Qt::WaitCursor);

      // Set parameters (Input/Output).
      m_params = new te::vp::AlgorithmParams();
      m_params->setInputParams(m_inputParams);
      m_params->setOutputDataSource(aux);
      m_params->setOutputDataSetName(outputdataset);
      m_params->setSpecificParams(specificParams);

      te::vp::Identity identity;

      res = identity.executeMemory(m_params);

      delete m_params;

      if(!res)
      {
        this->setCursor(Qt::ArrowCursor);
        QMessageBox::information(this, tr("Identity"),
                                 tr("Error: could not generate the identity."));
        reject();
      }
    }

    // Creating a layer for the result
    te::da::DataSourcePtr outDataSource =
        te::da::GetDataSource(m_outputDatasource->getId());

    te::qt::widgets::DataSet2Layer converter(m_outputDatasource->getId());

    te::da::DataSetTypePtr dt(
        outDataSource->getDataSetType(outputdataset).release());

    m_layerResult = converter(dt);
  }
  catch(const std::exception& e)
  {
    this->setCursor(Qt::ArrowCursor);
    QMessageBox::warning(this, tr("Identity"), e.what());

    std::string str = "Vector Processing - Identity - ";
    str += e.what();
    TE_LOG_ERROR(str);

    return;
  }

  this->setCursor(Qt::ArrowCursor);
  accept();
}

void te::vp::IdentityDialog::onCancelPushButtonClicked()
{
  reject();
}

