/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

This file is part of the TerraLib - a Framework for building GIS enabled applications.

TerraLib is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

TerraLib is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with TerraLib. See COPYING. If not, write to
TerraLib Team at <terralib-team@terralib.org>.
*/

/*!
 * \file src/terralib/vp/qt/InputDataAccessItemWidget.cpp
 *
 * \brief Implements a widget used to return a Data Access object.
 */

#include "InputDataAccessItemWidget.h"
#include "ui/ui_InputDataAccessItemWidgetForm.h"

// TerraLib Includes
#include "../../dataaccess/datasource/DataSource.h"
#include "../../dataaccess/datasource/DataSourceManager.h"
#include "../../dataaccess/utils/DataAccess.h"
#include "../../maptools/QueryLayer.h"
#include "../../maptools/Utils.h"
#include "../../qt/widgets/Utils.h"
#include "../../qt/widgets/dataset/utils/DataSetTypeFilter.h"
#include "../../qt/widgets/srs/SRSManagerDialog.h"
 
// Qt Includes
#include <QGridLayout>
#include <QMessageBox>

Q_DECLARE_METATYPE(te::map::AbstractLayerPtr)

te::vp::qt::InputDataAccessItemWidget::InputDataAccessItemWidget(QWidget* parent, Qt::WindowFlags f)
  : QWidget(parent, f),
  m_ui(new Ui::InputDataAccessItemWidgetForm)
{
  m_ui->setupUi(this);

  //set icons
  m_ui->m_selObjsToolButton->setIcon(QIcon::fromTheme("pointer-selection"));
  m_ui->m_converterToolButton->setIcon(QIcon::fromTheme("view-data-table"));

  //start datasetType filter widget to this widget
  m_dsTypeFiltertWidget.reset(new te::qt::widgets::DataSetTypeFilter(nullptr));
  
  //connects
  connect(m_ui->m_inputLayerComboBox, SIGNAL(activated(int)), this, SLOT(onInputLayerComboBoxActivated(int)));
  connect(m_ui->m_converterToolButton, SIGNAL(clicked()), this, SLOT(onConverterToolButtonClicked()));
}

te::vp::qt::InputDataAccessItemWidget::~InputDataAccessItemWidget() = default;
 
Ui::InputDataAccessItemWidgetForm* te::vp::qt::InputDataAccessItemWidget::getForm() const
{
  return m_ui.get();
}

void te::vp::qt::InputDataAccessItemWidget::setLayerList(const std::list<te::map::AbstractLayerPtr>& layers)
{
  std::list<te::map::AbstractLayerPtr>::const_iterator it = layers.begin();

  while (it != layers.end())
  {
    std::unique_ptr<te::da::DataSetType> dsType = std::move(it->get()->getSchema());

    if (dsType->hasGeom())
    {
      m_ui->m_inputLayerComboBox->addItem(te::qt::widgets::Convert2QString(it->get()->getTitle()), QVariant::fromValue(*it));
    }

    ++it;
  }

  m_ui->m_inputLayerComboBox->model()->sort(0);
  
  //activate the first layer
  if (m_ui->m_inputLayerComboBox->count() > 0)
  {
    m_ui->m_inputLayerComboBox->setCurrentIndex(0);
    onInputLayerComboBoxActivated(0);
  }
}

te::da::DataAccessPtr te::vp::qt::InputDataAccessItemWidget::getDataAccess()
{
  //get layer
  int index = m_ui->m_inputLayerComboBox->currentIndex();
  if (index == -1)
  {
    QString qMessage = tr("There is not a valid current item in the input layer list");
    throw te::common::Exception(te::qt::widgets::Convert2StdString(qMessage));
  }
  QVariant varLayer = m_ui->m_inputLayerComboBox->itemData(index, Qt::UserRole);

  te::map::AbstractLayerPtr layer = varLayer.value<te::map::AbstractLayerPtr>();

  //create dataaccess
  te::da::DataAccessPtr dataAccess = te::map::GetDataAccess(layer);

  //if the converter already exists, we just append the new attribute convertions
  m_dsTypeFiltertWidget->applyFilterToDataAccess(dataAccess.get());

  if (m_ui->m_selObjsToolButton->isChecked() == true)
  {
    dataAccess->setObjectIdSetFilter(layer->getSelected());
  }

  return dataAccess;
}

void te::vp::qt::InputDataAccessItemWidget::setFilterPropertiesEnabled(bool enabled)
{
  m_ui->m_converterToolButton->setVisible(enabled);
}

void te::vp::qt::InputDataAccessItemWidget::onInputLayerComboBoxActivated(int index)
{
  //get layer
  QVariant varLayer = m_ui->m_inputLayerComboBox->itemData(index, Qt::UserRole);

  te::map::AbstractLayerPtr layer = varLayer.value<te::map::AbstractLayerPtr>();

  //sets the input DataAccess to be adapted
  te::da::DataAccessPtr dataAccess = te::map::GetDataAccess(layer);
  m_dsTypeFiltertWidget->init(dataAccess->getDataSetType(), true);

  //we try to suggest the the layer selection filter by checking if there is a selection in the layer
  bool hasSelection = (layer->getSelected() != nullptr);
  m_ui->m_selObjsToolButton->setEnabled(hasSelection);
  m_ui->m_selObjsToolButton->setChecked(hasSelection);

  emit dataAccessSelectionChanged();
}

void te::vp::qt::InputDataAccessItemWidget::onConverterToolButtonClicked()
{
  QDialog dummyDlg(this, Qt::Tool);
 
  dummyDlg.setWindowTitle(tr("Property Filter"));

  QGridLayout* gridLayout = new QGridLayout(&dummyDlg);

  gridLayout->addWidget(m_dsTypeFiltertWidget.get());

  dummyDlg.exec();

  m_dsTypeFiltertWidget->setParent(nullptr);
}
