/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/vp/MergeDialog.cpp

  \brief A dialog for merge operation
*/

// TerraLib
#include "../../core/logger/Logger.h"
#include "../../core/filesystem/FileSystem.h"
#include "../../core/translator/Translator.h"
#include "../../common/progress/ProgressManager.h"
#include "../../common/StringUtils.h"
#include "../../common/progress/TaskProgress.h"
#include "../../dataaccess/dataset/DataSetAdapter.h"
#include "../../dataaccess/dataset/DataSetTypeConverter.h"
#include "../../dataaccess/datasource/DataSource.h"
#include "../../dataaccess/datasource/DataSourceCapabilities.h"
#include "../../dataaccess/datasource/DataSourceInfoManager.h"
#include "../../dataaccess/datasource/DataSourceFactory.h"
#include "../../dataaccess/datasource/DataSourceManager.h"
#include "../../dataaccess/datasource/DataSourceTransactor.h"
#include "../../dataaccess/utils/Utils.h"
#include "../../maptools/QueryLayer.h"
#include "../../qt/widgets/datasource/selector/DataSourceSelectorDialog.h"
#include "../../qt/widgets/layer/utils/DataSet2Layer.h"
#include "../../qt/widgets/progress/ProgressViewerDialog.h"
#include "../../qt/widgets/utils/FileDialog.h"

#include "../../vp/Merge.h"

#include "MergeDialog.h"
#include "ui/ui_MergeDialogForm.h"

#include "../../vp/AlgorithmParams.h"
#include "../../vp/ComplexData.h"
#include "../../vp/InputParams.h"

// Qt
#include <QComboBox>
#include <QFileDialog>
#include <QMessageBox>

// Boost
#include <boost/filesystem.hpp>
#include <boost/uuid/random_generator.hpp>
#include <boost/uuid/uuid_io.hpp>

Q_DECLARE_METATYPE(te::map::AbstractLayerPtr)

te::vp::MergeDialog::MergeDialog(QWidget* parent, Qt::WindowFlags f)
  : QDialog(parent, f),
    m_ui(new Ui::MergeDialogForm),
    m_layers(std::list<te::map::AbstractLayerPtr>()),
    m_toFile(false)
{
// add controls
  m_ui->setupUi(this);

  connect(m_ui->m_targetLayerComboBox, SIGNAL(currentIndexChanged(int)), this, SLOT(onTargetLayerComboBoxChanged(int)));
  connect(m_ui->m_originLayerComboBox, SIGNAL(currentIndexChanged(int)), this, SLOT(onOriginLayerComboBoxChanged(int)));
  connect(m_ui->m_okPushButton, SIGNAL(clicked()), this, SLOT(onOkPushButtonClicked()));
  connect(m_ui->m_targetDatasourceToolButton, SIGNAL(pressed()), this, SLOT(onTargetDatasourceToolButtonPressed()));
  connect(m_ui->m_targetFileToolButton, SIGNAL(pressed()), this, SLOT(onTargetFileToolButtonPressed()));
  connect(m_ui->m_outputGroupBox, SIGNAL(toggled(bool)), this, SLOT(onOutputGroupBoxToggled(bool)));

// add icons
  m_ui->m_imgLabel->setPixmap(QIcon::fromTheme("vp-merge-hint").pixmap(112,48));
  m_ui->m_targetDatasourceToolButton->setIcon(QIcon::fromTheme("datasource"));

  m_ui->m_helpPushButton->setNameSpace("dpi.inpe.br.plugins"); 
  m_ui->m_helpPushButton->setPageReference("plugins/vp/vp_merge.html");

  m_outputDatasource = te::da::DataSourceInfoPtr();
  m_ui->m_newLayerNameLineEdit->setEnabled(true);
}

te::vp::MergeDialog::~MergeDialog() = default;

void te::vp::MergeDialog::setLayers(std::list<te::map::AbstractLayerPtr> layers)
{
  std::list<te::map::AbstractLayerPtr>::iterator it = layers.begin();

  while (it != layers.end())
  {
    std::unique_ptr<te::da::DataSetType> dsType = it->get()->getSchema();
    if (dsType->hasGeom())
    {
      m_layers.push_back(*it);
    }

    ++it;
  }

  updateTargetLayerComboBox();

  updateOriginLayerComboBox();

  updateAttrTableWidget();
}

void te::vp::MergeDialog::updateTargetLayerComboBox()
{
  std::list<te::map::AbstractLayerPtr>::iterator it = m_layers.begin();

  disconnect(m_ui->m_targetLayerComboBox, SIGNAL(currentIndexChanged(int)), this, SLOT(onTargetLayerComboBoxChanged(int)));

  while (it != m_layers.end())
  {
    m_ui->m_targetLayerComboBox->addItem(QString(it->get()->getTitle().c_str()), QVariant::fromValue(*it));
    ++it;
  }

  connect(m_ui->m_targetLayerComboBox, SIGNAL(currentIndexChanged(int)), this, SLOT(onTargetLayerComboBoxChanged(int)));

  QVariant varLayer = m_ui->m_targetLayerComboBox->itemData(m_ui->m_targetLayerComboBox->currentIndex(), Qt::UserRole);
  te::map::AbstractLayerPtr layer = varLayer.value<te::map::AbstractLayerPtr>();

  m_targetSelectedLayer = layer;
}

void te::vp::MergeDialog::updateOriginLayerComboBox()
{
  int currIndex = m_ui->m_targetLayerComboBox->currentIndex();

  std::list<te::map::AbstractLayerPtr>::iterator it = m_layers.begin();

  disconnect(m_ui->m_originLayerComboBox, SIGNAL(currentIndexChanged(int)), this, SLOT(onOriginLayerComboBoxChanged(int)));

  while (it != m_layers.end())
  {
    m_ui->m_originLayerComboBox->addItem(QString(it->get()->getTitle().c_str()), QVariant::fromValue(*it));
    ++it;
  }

  m_ui->m_originLayerComboBox->removeItem(currIndex);

  connect(m_ui->m_originLayerComboBox, SIGNAL(currentIndexChanged(int)), this, SLOT(onOriginLayerComboBoxChanged(int)));

  QVariant varLayer = m_ui->m_originLayerComboBox->itemData(m_ui->m_originLayerComboBox->currentIndex(), Qt::UserRole);
  te::map::AbstractLayerPtr layer = varLayer.value<te::map::AbstractLayerPtr>();

  m_originSelectedLayer = layer;
}

te::map::AbstractLayerPtr te::vp::MergeDialog::getLayer()
{
  return m_outputLayer;
}

void te::vp::MergeDialog::onTargetDatasourceToolButtonPressed()
{
  m_ui->m_newLayerNameLineEdit->clear();
  m_ui->m_newLayerNameLineEdit->setEnabled(true);
  te::qt::widgets::DataSourceSelectorDialog dlg(this);
  dlg.exec();

  std::list<te::da::DataSourceInfoPtr> dsPtrList = dlg.getSelecteds();

  if(dsPtrList.empty())
    return;

  std::list<te::da::DataSourceInfoPtr>::iterator it = dsPtrList.begin();

  m_ui->m_repositoryLineEdit->setText(QString(it->get()->getTitle().c_str()));

  m_outputDatasource = *it;
  
  m_toFile = false;
}

void te::vp::MergeDialog::onTargetFileToolButtonPressed()
{
  m_ui->m_newLayerNameLineEdit->clear();
  m_ui->m_repositoryLineEdit->clear();
  
  te::qt::widgets::FileDialog fileDialog(this, te::qt::widgets::FileDialog::VECTOR);

  try {
    fileDialog.exec();
  }
  catch (te::common::Exception& ex) {
    QMessageBox::warning(this, tr("File information"), ex.what());
    return;
  }

  m_ui->m_repositoryLineEdit->setText(fileDialog.getPath().c_str());
  m_ui->m_newLayerNameLineEdit->setText(fileDialog.getFileName().c_str());
  
  m_toFile = true;
  m_ui->m_newLayerNameLineEdit->setEnabled(false);
}

void te::vp::MergeDialog::onOkPushButtonClicked()
{
  TE_LOG_INFO(TE_TR("Merge operation started"));
  
  bool isUpdate = !m_ui->m_outputGroupBox->isChecked();

  if (isUpdate)
  {
    int res = QMessageBox::question(this, tr("Merge"), tr("The origin layer information will be inserted into the target layer.\n Do you want to continue the operation?"), QMessageBox::Yes, QMessageBox::Cancel);
    if(res == QMessageBox::Cancel)
      return;
  }
  
  if (m_ui->m_targetLayerComboBox->currentText().isEmpty() || m_ui->m_originLayerComboBox->currentText().isEmpty())
  {
    QMessageBox::information(this, tr("Merge"), tr("It is necessary at least two layer to operate!"));
    return;
  }

  // Checking consistency of output paramenters
  if (!isUpdate && m_ui->m_repositoryLineEdit->text().isEmpty())
  {
    QMessageBox::information(this, tr("Merge"), tr("Select a repository for the resulting layer!"));
    return;
  }

  if (!isUpdate && m_ui->m_newLayerNameLineEdit->text().isEmpty())
  {
    QMessageBox::information(this, tr("Merge"), tr("Define a name for the resulting layer!"));
    return;
  }

  int targetSrid = m_targetSelectedLayer->getSRID();
  int originSrid = m_originSelectedLayer->getSRID();

  if (targetSrid <= 0 || originSrid <= 0)
  {
    QMessageBox::information(this, tr("Merge"), tr("All layers must have SRID!"));
    return;
  }

  te::qt::widgets::ProgressViewerDialog v(this);

  std::string outputdataset = m_ui->m_newLayerNameLineEdit->text().toUtf8().data();

  std::string targetSourceId = m_targetSelectedLayer->getDataSourceId();
  std::string originSourceId = m_originSelectedLayer->getDataSourceId();

  te::da::DataSourcePtr targetSource = te::da::DataSourceManager::getInstance().find(targetSourceId);
  te::da::DataSourcePtr originSource = te::da::DataSourceManager::getInstance().find(originSourceId);

  std::unique_ptr<te::da::DataSetTypeConverter> targetConverter(new te::da::DataSetTypeConverter(m_targetSelectedLayer->getSchema().get(), targetSource->getCapabilities(), targetSource->getEncoding()));
  te::da::AssociateDataSetTypeConverterSRID(targetConverter.get(), m_targetSelectedLayer->getSRID());

  std::unique_ptr<te::da::DataSetTypeConverter> originConverter(new te::da::DataSetTypeConverter(m_originSelectedLayer->getSchema().get(), originSource->getCapabilities(), originSource->getEncoding()));
  te::da::AssociateDataSetTypeConverterSRID(originConverter.get(), m_originSelectedLayer->getSRID(), m_targetSelectedLayer->getSRID());

  te::da::DataSetType* targetDst(targetConverter->getResult());
  te::da::DataSetType* originDst(originConverter->getResult());

  std::unique_ptr<te::common::TaskProgress> task(new te::common::TaskProgress(TE_TR("Preparing Merge...")));
  task->setTotalSteps(2);

  std::string logMsg = TE_TR("Getting Target Layer Data...");
  task->setMessage(logMsg);
  TE_LOG_INFO(logMsg);
  const te::da::DataSourceCapabilities& targetCap = targetSource->getCapabilities();
  const te::da::DataSetCapabilities& targetDsCap = targetCap.getDataSetCapabilities();
  std::unique_ptr<te::da::DataSourceTransactor> targetTransactor = targetSource->getTransactor();
  std::unique_ptr<te::da::DataSet> targetConnDs = targetTransactor->getDataSet(targetDst->getName(), te::common::FORWARDONLY, targetDsCap.isConnected());
  TE_LOG_INFO(TE_TR("Target Layer DataSet ") + (targetDsCap.isConnected() ? TE_TR("CONNECTED") : ("NOT CONNECTED")));

  task->pulse();

  logMsg = TE_TR("Getting Origin Layer Data...");
  task->setMessage(logMsg);
  TE_LOG_INFO(logMsg);
  const te::da::DataSourceCapabilities& originCap = originSource->getCapabilities();
  const te::da::DataSetCapabilities& originDsCap = originCap.getDataSetCapabilities();
  std::unique_ptr<te::da::DataSourceTransactor> originTransactor = originSource->getTransactor();

  std::unique_ptr<te::da::DataSet> originConnDs;

  te::map::QueryLayer* queryLayer = dynamic_cast<te::map::QueryLayer*>(m_originSelectedLayer.get());
  
  if (queryLayer)
  {
    te::da::Select* q = queryLayer->getQuery();

    originConnDs = originTransactor->query(q);
  }
  else
  {
    originConnDs = originTransactor->getDataSet(originDst->getName(), te::common::FORWARDONLY, originDsCap.isConnected());
    TE_LOG_INFO(TE_TR("Origin: Query Layer"));
  }

  TE_LOG_INFO(TE_TR("Origin Layer DataSet ") + (originDsCap.isConnected() ? TE_TR("CONNECTED") : ("NOT CONNECTED")));

  task->pulse();

  task.reset(nullptr);
  
  te::da::DataSetAdapter* targetAdapter = te::da::CreateAdapter(targetConnDs.release(), targetConverter.get());
  te::da::DataSetAdapter* originAdapter = te::da::CreateAdapter(originConnDs.release(), originConverter.get());

  std::unique_ptr<te::da::DataSet> targetDs(targetAdapter);
  std::unique_ptr<te::da::DataSet> originDs(originAdapter);
  
  try
  {

    te::vp::InputParams inputParam;
    inputParam.m_inputDataSet = targetDs.get();
    inputParam.m_inputDataSetType = targetDst;
    inputParam.m_inputDataSource = targetSource;

    te::vp::InputParams mergeParam;
    mergeParam.m_inputDataSet = originDs.get();
    mergeParam.m_inputDataSetType = originDst;
    mergeParam.m_inputDataSource = originSource;

    std::vector<te::vp::InputParams> params;
    params.push_back(inputParam);
    params.push_back(mergeParam);

    te::vp::ComplexData<std::vector<std::pair<std::string, std::string> > >* attrs = new te::vp::ComplexData<std::vector<std::pair<std::string, std::string> > >(getTablePropertiesNames());

    te::vp::ComplexData<bool>* isUp = new te::vp::ComplexData<bool>(isUpdate);

    std::map<std::string, te::dt::AbstractData*> specificParams;
    specificParams["ATTRIBUTES"] = attrs;
    specificParams["ISUPDATE"] = isUp;

    te::vp::AlgorithmParams* aParams = new te::vp::AlgorithmParams();
    aParams->setInputParams(params);
    aParams->setOutputDataSetName(outputdataset);
    aParams->setSpecificParams(specificParams);

    te::da::DataSourcePtr auxSource;
    std::string ogrDsinfo("file://");
    boost::filesystem::path ogrUri;

    if (m_toFile)
    {
      ogrUri = m_ui->m_repositoryLineEdit->text().toUtf8().data();

      if (te::core::FileSystem::exists(ogrUri.string()))
      {
        QMessageBox::information(this, tr("Merge"), tr("Output file already exists. Remove it or select a new name and try again."));
        return;
      }

      std::size_t idx = outputdataset.find(".");
      if (idx != std::string::npos)
        outputdataset = outputdataset.substr(0, idx);

      ogrDsinfo += ogrUri.string();

      auxSource.reset(te::da::DataSourceFactory::make("OGR", ogrDsinfo).release());
      auxSource->open();

      if (auxSource->dataSetExists(outputdataset))
      {
        QMessageBox::information(this, tr("Merge"), tr("There is already a dataset with the requested name in the output data source. Remove it or select a new name and try again."));
        return;
      }
    }
    else
    {
      if (!isUpdate)
      {
        auxSource = te::da::GetDataSource(m_outputDatasource->getId());
        if (!auxSource)
        {
          QMessageBox::information(this, tr("Merge"), tr("The selected output datasource can not be accessed."));
          return;
        }
        if (auxSource->dataSetExists(outputdataset))
        {
          QMessageBox::information(this, tr("Merge"), tr("Dataset already exists. Remove it or select a new name and try again."));
          return;
        }
      }
    }

  
    if (!isUpdate)
    {
      aParams->setOutputDataSource(auxSource);
    }
    else
    {
      aParams->setOutputDataSource(targetSource);
    }    

    

    te::vp::Merge merge;
    merge.executeMemory(aParams);

    delete aParams;

    if (m_toFile)
    {
      // let's include the new datasource in the managers
      boost::uuids::basic_random_generator<boost::mt19937> gen;
      boost::uuids::uuid u = gen();
      std::string id = boost::uuids::to_string(u);

      te::da::DataSourceInfoPtr ds(new te::da::DataSourceInfo);
      ds->setConnInfo(ogrDsinfo);
      ds->setTitle(ogrUri.stem().string());
      ds->setAccessDriver("OGR");
      ds->setType("OGR");
      ds->setDescription(ogrUri.string());
      ds->setId(id);

      te::da::DataSourcePtr newds = te::da::DataSourceManager::getInstance().get(id, "OGR", ds->getConnInfo());
      newds->open();
      te::da::DataSourceInfoManager::getInstance().add(ds);
      m_outputDatasource = ds;
    }

    if (!isUpdate)
    {
      // creating a layer for the result
      te::da::DataSourcePtr outDataSource = te::da::GetDataSource(m_outputDatasource->getId());

      te::qt::widgets::DataSet2Layer converter(m_outputDatasource->getId());

      te::da::DataSetTypePtr dt(outDataSource->getDataSetType(outputdataset).release());
      m_outputLayer = converter(dt);
    }
    else
    {
      te::gm::Envelope env1 = m_targetSelectedLayer->getExtent();
      te::gm::Envelope env2 = m_originSelectedLayer->getExtent();

      env1.Union(env2);

      m_targetSelectedLayer->setExtent(env1);

      QMessageBox::information(this, tr("Merge"), tr("Successfully Merge!"));
    }
  }
  catch (const te::common::Exception& e)
  {
    QMessageBox::warning(this, tr("Merge"), e.what());
    return;
  }
  catch (const std::exception& e)
  {
    QMessageBox::warning(this, tr("Merge"), e.what());
    return;
  }

  TE_LOG_INFO(TE_TR("Merge operation finished"));
  accept();
}

void te::vp::MergeDialog::onCancelPushButtonClicked()
{
  reject();
}

void te::vp::MergeDialog::onTargetLayerComboBoxChanged(int index)
{
  QVariant varLayer = m_ui->m_targetLayerComboBox->itemData(index, Qt::UserRole);
  te::map::AbstractLayerPtr layer = varLayer.value<te::map::AbstractLayerPtr>();

  m_ui->m_originLayerComboBox->clear();

  m_targetSelectedLayer = layer;

  updateOriginLayerComboBox();

  updateAttrTableWidget();

  // The update only works with datasetlayer
  if (m_targetSelectedLayer->getType() != "DATASETLAYER")
  {
    m_ui->m_outputGroupBox->setChecked(true);
  }
}

void te::vp::MergeDialog::onOriginLayerComboBoxChanged(int index)
{
  QVariant varLayer = m_ui->m_originLayerComboBox->itemData(index, Qt::UserRole);
  te::map::AbstractLayerPtr layer = varLayer.value<te::map::AbstractLayerPtr>();

  m_originSelectedLayer = layer;

  updateAttrTableWidget();
}

void te::vp::MergeDialog::updateAttrTableWidget()
{
  m_ui->m_attributeTableWidget->setRowCount(0);

  std::unique_ptr<te::da::DataSetType> targetSchema;
  std::unique_ptr<te::da::DataSetType> originSchema;

  if (m_targetSelectedLayer)
    targetSchema = m_targetSelectedLayer->getSchema();
  else
    return;

  if (m_originSelectedLayer)
    originSchema = m_originSelectedLayer->getSchema();
  else
    return;

  std::vector<te::dt::Property*> originProps = originSchema->getProperties();

  std::vector<te::dt::Property*> targetProps = targetSchema->getProperties();

  int rowCount = -1;
  for (std::size_t i = 0; i < targetProps.size(); ++i)
  {
    if (targetProps[i]->getType() != te::dt::GEOMETRY_TYPE)
    {
      ++rowCount;
      m_ui->m_attributeTableWidget->insertRow(rowCount);

      QTableWidgetItem* item = new QTableWidgetItem(targetProps[i]->getName().c_str());
      item->setFlags(Qt::ItemIsEnabled);
      m_ui->m_attributeTableWidget->setItem(rowCount, 0, item);

      QComboBox* cb = getPossibleAttributesComboBox(targetProps[i], originProps);
      cb->setParent(m_ui->m_attributeTableWidget);

      m_ui->m_attributeTableWidget->setCellWidget(rowCount, 1, cb);
    }
  }

  std::size_t fSize = targetProps.size();
  std::size_t sSize = originProps.size();
  if (targetProps.size() < originProps.size())
  {
    int diference = static_cast<int>(sSize - fSize);

    for (int i = 0; i < diference; ++i)
    {
      ++rowCount;
      m_ui->m_attributeTableWidget->insertRow(rowCount);

      QTableWidgetItem* item = new QTableWidgetItem();
      m_ui->m_attributeTableWidget->setItem(rowCount, 0, item);

      QComboBox* cb = getAllAttributesComboBox(originProps);
      cb->setParent(m_ui->m_attributeTableWidget);

      m_ui->m_attributeTableWidget->setCellWidget(rowCount, 1, cb);
    }
  }

#if (QT_VERSION >= 0x050000)
  m_ui->m_attributeTableWidget->horizontalHeader()->setSectionResizeMode(QHeaderView::Stretch);
#else
  m_ui->m_attributeTableWidget->horizontalHeader()->setResizeMode(QHeaderView::Stretch);
#endif
}

QComboBox* te::vp::MergeDialog::getPossibleAttributesComboBox(te::dt::Property* mainProp, std::vector<te::dt::Property*> props)
{
  QComboBox* result = new QComboBox();
  result->addItem("");

  int sameName = -1;
  int count = 1;

  for (std::size_t i = 0; i < props.size(); ++i)
  {
    if (mainProp->getType() == props[i]->getType())
    {
      if (te::common::Convert2LCase(mainProp->getName()) == te::common::Convert2LCase(props[i]->getName()))
      {
        sameName = count;
      }

      result->addItem(props[i]->getName().c_str());
      ++count;
    }
  }

  if (sameName > -1)
  {
    result->setCurrentIndex(sameName);
  }

  return result;
}

QComboBox* te::vp::MergeDialog::getAllAttributesComboBox(std::vector<te::dt::Property*> props)
{
  QComboBox* result = new QComboBox();
  result->addItem("");

  for (std::size_t i = 0; i < props.size(); ++i)
  {
    if (props[i]->getType() != te::dt::GEOMETRY_TYPE)
    {
      result->addItem(props[i]->getName().c_str());
    }
  }

  return result;
}

std::vector<std::pair<std::string, std::string> > te::vp::MergeDialog::getTablePropertiesNames()
{
  std::vector<std::pair<std::string, std::string> > result;

  int rowCount = m_ui->m_attributeTableWidget->rowCount();

  for (std::size_t i = 0; i < static_cast<std::size_t>(rowCount); ++i)
  {
    std::string fp = m_ui->m_attributeTableWidget->item(static_cast<int>(i), 0)->text().toUtf8().data();;

    QComboBox* cb = dynamic_cast<QComboBox*>(m_ui->m_attributeTableWidget->cellWidget(static_cast<int>(i), 1));

    std::string sp = cb->currentText().toUtf8().data();

    result.push_back(std::pair<std::string, std::string>(fp, sp));
  }

  return result;
}

void te::vp::MergeDialog::onOutputGroupBoxToggled(bool on)
{
  if (!on)
  {
    if (m_targetSelectedLayer->getType() != "DATASETLAYER")
    {
      QMessageBox::warning(this, tr("Merge"), tr("To use target layer as output (update), it must be a DataSet Layer!"));
      m_ui->m_outputGroupBox->setChecked(true);
      return;
    }
  }
}
