/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

This file is part of the TerraLib - a Framework for building GIS enabled applications.

TerraLib is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

TerraLib is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with TerraLib. See COPYING. If not, write to
TerraLib Team at <terralib-team@terralib.org>.
*/

/*!
 * \file src/terralib/vp/qt/OutputDataAccessItemWidget.cpp
 *
 * \brief Implements a widget used to return a Data Access object.
 */

#include "OutputDataAccessItemWidget.h"
#include "ui/ui_OutputDataAccessItemWidgetForm.h"

// TerraLib Includes
#include "../../core/filesystem/FileSystem.h"
#include "../../common/Exception.h"
#include "../../common/StringUtils.h"
#include "../../dataaccess/datasource/DataSourceFactory.h"
#include "../../dataaccess/datasource/DataSourceInfoManager.h"
#include "../../dataaccess/datasource/DataSourceManager.h"
#include "../../dataaccess/utils/DataAccess.h"
#include "../../dataaccess/utils/Utils.h"
#include "../../qt/widgets/Utils.h"
#include "../../qt/widgets/datasource/selector/DataSourceSelectorDialog.h"
#include "../../qt/widgets/utils/FileDialog.h"
 
// Qt Includes
#include <QMessageBox> 

te::vp::qt::OutputDataAccessItemWidget::OutputDataAccessItemWidget(QWidget* parent, Qt::WindowFlags f)
  : QWidget(parent, f),
  m_ui(new Ui::OutputDataAccessItemWidgetForm),
  m_toFile(false)
{
  m_ui->setupUi(this);

  m_ui->m_targetFileToolButton->setIcon(QIcon::fromTheme("folder-open"));
  m_ui->m_targetDatasourceToolButton->setIcon(QIcon::fromTheme("datasource"));
  m_ui->m_newLayerNameLineEdit->setReadOnly(true);

  //connects
  connect(m_ui->m_targetDatasourceToolButton, SIGNAL(pressed()), this, SLOT(onTargetDatasourceToolButtonPressed()));
  connect(m_ui->m_targetFileToolButton, SIGNAL(pressed()), this, SLOT(onTargetFileToolButtonPressed()));
}

te::vp::qt::OutputDataAccessItemWidget::~OutputDataAccessItemWidget() = default;

Ui::OutputDataAccessItemWidgetForm* te::vp::qt::OutputDataAccessItemWidget::getForm() const
{
  return m_ui.get();
}

te::da::DataSourcePtr te::vp::qt::OutputDataAccessItemWidget::getOutputDataSorce()
{
  if (m_ui->m_newLayerNameLineEdit->text().isEmpty())
  {
    throw(te::common::Exception("Output layer name not defined."));
  }

  if (m_ui->m_repositoryLabel->text().isEmpty())
  {
    throw(te::common::Exception("Output repository not defined."));
  }

  if (m_outputDatasourceInfo.get() == nullptr && m_toFile == false)
  {
    throw(te::common::Exception("Output Data Souce information not defined."));
  }

  if (m_outputDatasource.get() != nullptr)
  {
    return m_outputDatasource;
  }

  std::string outputdataset = te::qt::widgets::Convert2StdString(m_ui->m_newLayerNameLineEdit->text());

  if (m_toFile)
  {
    std::string filePath = te::qt::widgets::Convert2StdString(m_ui->m_repositoryLabel->text());

    if (te::core::FileSystem::exists(filePath))
    {
      throw(te::common::Exception("Output file already exists. Remove it or select a new name and try again."));
    }

    //get dataset name without extension
    outputdataset = te::core::FileSystem::stem(outputdataset);

    //create URI
    std::string dsInfoStr = "file://" + filePath;

    //create data source
    m_outputDatasource = te::da::DataSourceFactory::make("OGR", dsInfoStr);
    m_outputDatasource->open();

    if (m_outputDatasource->dataSetExists(outputdataset))
    {
      throw(te::common::Exception("Dataset already exists. Remove it or select a new name and try again."));
    }

    //register new data source in terralib datasource manager
    te::da::DataSourceInfoPtr dsInfo = registerFileDataSource(dsInfoStr, outputdataset);

    m_outputDatasource = te::da::DataSourceManager::getInstance().get(dsInfo->getId(), "OGR", dsInfo->getConnInfo());
  }
  else
  {
    m_outputDatasource = te::da::GetDataSource(m_outputDatasourceInfo->getId());

    if (!m_outputDatasource)
    {
      throw(te::common::Exception("The selected output datasource can not be accessed."));
    }

    if (m_outputDatasource->dataSetExists(outputdataset))
    {
      throw(te::common::Exception("Dataset already exists. Remove it or select a new name and try again."));
    }
  }

  return m_outputDatasource;
}

te::da::DataAccessPtr te::vp::qt::OutputDataAccessItemWidget::getDataAccess()
{
  //get dataset name
  std::string layerName = te::qt::widgets::Convert2StdString(m_ui->m_newLayerNameLineEdit->text());
  std::string outputdataset = te::core::FileSystem::stem(layerName);

  //create dataaccess
  te::da::DataAccessPtr dataAccess(new te::da::DataAccess(getOutputDataSorce(), outputdataset));

  return dataAccess;
}

void te::vp::qt::OutputDataAccessItemWidget::reset()
{
  m_ui->m_newLayerNameLineEdit->clear();
  m_ui->m_repositoryLabel->clear();
  m_ui->m_newLayerNameLineEdit->setEnabled(true);

  m_toFile = false;
  m_outputDatasource.reset();
  m_outputDatasource.reset();
}

te::da::DataSourceInfoPtr te::vp::qt::OutputDataAccessItemWidget::registerFileDataSource(const std::string& filePath, const std::string& dataSetName)
{
  // let's include the new datasource in the managers
  std::string id = te::common::GetUniqueID();

  te::da::DataSourceInfoPtr ds(new te::da::DataSourceInfo);
  ds->setConnInfo(filePath);
  ds->setTitle(dataSetName);
  ds->setAccessDriver("OGR");
  ds->setType("OGR");
  ds->setDescription(filePath);
  ds->setId(id);

  te::da::DataSourceInfoManager::getInstance().add(ds);

  return ds;
}

void te::vp::qt::OutputDataAccessItemWidget::onTargetDatasourceToolButtonPressed()
{
  m_ui->m_newLayerNameLineEdit->clear();
  m_ui->m_newLayerNameLineEdit->setEnabled(true);

  te::qt::widgets::DataSourceSelectorDialog dlg(this);
  dlg.exec();

  std::list<te::da::DataSourceInfoPtr> dsPtrList = dlg.getSelecteds();

  if (dsPtrList.empty())
    return;

  std::list<te::da::DataSourceInfoPtr>::iterator it = dsPtrList.begin();

  m_ui->m_repositoryLabel->setText(te::qt::widgets::Convert2QString(it->get()->getTitle()));
  m_ui->m_newLayerNameLineEdit->setReadOnly(false);

  m_outputDatasourceInfo = *it;

  m_toFile = false;
}

void te::vp::qt::OutputDataAccessItemWidget::onTargetFileToolButtonPressed()
{
  m_ui->m_newLayerNameLineEdit->clear();
  m_ui->m_repositoryLabel->clear();

  te::qt::widgets::FileDialog fileDialog(this, te::qt::widgets::FileDialog::VECTOR);

  try
  {
    fileDialog.exec();
  }
  catch (te::common::Exception& ex)
  {
    QMessageBox::warning(this, tr("Vector Processing"), ex.what());
    return;
  }

  m_ui->m_repositoryLabel->setText(te::qt::widgets::Convert2QString(fileDialog.getPath()));
  m_ui->m_newLayerNameLineEdit->setText(te::qt::widgets::Convert2QString(fileDialog.getFileName()));
  m_ui->m_newLayerNameLineEdit->setReadOnly(true);

  m_toFile = true;
}
