/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

#include "RegisterDialog.h"
#include "ui/ui_RegisterDialogForm.h"
#include "TiePointsLocatorDialog.h"
#include "../Register.h"
#include "../../common/progress/ProgressManager.h"
#include "../../common/STLUtils.h"
#include "../../common/StringUtils.h"
#include "../../common/UnitsOfMeasureManager.h"
#include "../../dataaccess/dataset/DataSetType.h"
#include "../../dataaccess/datasource/DataSourceCapabilities.h"
#include "../../dataaccess/datasource/DataSourceInfo.h"
#include "../../dataaccess/datasource/DataSourceInfoManager.h"
#include "../../dataaccess/datasource/DataSourceManager.h"
#include "../../dataaccess/datasource/DataSourceFactory.h"
#include "../../dataaccess/utils/Utils.h"
#include "../../maptools/AbstractLayer.h"
#include "../../maptools/DataSetLayer.h"
#include "../../qt/widgets/datasource/selector/DataSourceSelectorDialog.h"
#include "../../qt/widgets/layer/utils/DataSet2Layer.h"
#include "../../qt/widgets/progress/ProgressViewerDialog.h"
#include "../../qt/widgets/srs/SRSManagerDialog.h"
#include "../../qt/widgets/utils/FileDialog.h"
#include "../../qt/widgets/utils/ScopedCursor.h"

#include <QMessageBox>

#include <boost/filesystem.hpp>

te::vp::RegisterDialog::RegisterDialog(QWidget* parent, Qt::WindowFlags f)
  : QDialog(parent, f),
    m_ui(new Ui::RegisterDialogForm),
    m_layers(std::list<te::map::AbstractLayerPtr>()),
    m_selectedLayer(nullptr)
{
// add controls
  m_ui->setupUi(this);

// add icons
  m_ui->m_imgLabel->setPixmap(QIcon::fromTheme("register").pixmap(112,48));

  QSize iconSize(96, 48);

  m_ui->m_targetDatasourceToolButton->setIcon(QIcon::fromTheme("datasource"));

  connect(m_ui->m_layersComboBox, SIGNAL(currentIndexChanged(int)), this, SLOT(onLayerComboBoxChanged(int)));
  connect(m_ui->m_targetDatasourceToolButton, SIGNAL(pressed()), this, SLOT(onTargetDatasourceToolButtonPressed()));
  connect(m_ui->m_targetFileToolButton, SIGNAL(pressed()), this,  SLOT(onTargetFileToolButtonPressed()));
  connect(m_ui->m_okPushButton, SIGNAL(clicked()), this, SLOT(onOkPushButtonClicked()));
  connect(m_ui->m_cancelPushButton, SIGNAL(clicked()), this, SLOT(onCancelPushButtonClicked()));
  connect(m_ui->m_acquireTiepointsPushButton, SIGNAL(clicked()), this, SLOT(onAcquirePushButtonClicked()));
  
  m_ui->m_helpPushButton->setNameSpace("dpi.inpe.br.plugins"); 
  m_ui->m_helpPushButton->setPageReference("plugins/vp/vp_register.html");
}

te::vp::RegisterDialog::~RegisterDialog() = default;

void te::vp::RegisterDialog::setLayers(std::list<te::map::AbstractLayerPtr> layers)
{
  m_layers = layers;
  
  std::list<te::map::AbstractLayerPtr>::iterator it = m_layers.begin();

  while(it != m_layers.end())
  {
    std::unique_ptr<te::da::DataSetType> dsType = it->get()->getSchema();

    if(
        ( it->get()->getType() == "DATASETLAYER" )
        &&
        ( dsType.get() && dsType->hasGeom() )
      )
    {
      m_ui->m_layersComboBox->addItem(QString(it->get()->getTitle().c_str()), QVariant(it->get()->getId().c_str()));
    }

    ++it;
  }
}

te::map::AbstractLayerPtr te::vp::RegisterDialog::getLayer()
{
  return m_outputLayer;
}

void te::vp::RegisterDialog::onLayerComboBoxChanged(int index)
{
  std::list<te::map::AbstractLayerPtr>::iterator it = m_layers.begin();
  std::string layerID = m_ui->m_layersComboBox->itemData(index, Qt::UserRole).toString().toUtf8().data();

  while(it != m_layers.end())
  {
    if(layerID == it->get()->getId())
    {
      te::map::AbstractLayerPtr selectedLayer = it->get();
      m_selectedLayer = selectedLayer;

      return;
    }

    ++it;
  }
}

void te::vp::RegisterDialog::onTargetDatasourceToolButtonPressed()
{
  m_ui->m_newLayerNameLineEdit->clear();
  m_ui->m_newLayerNameLineEdit->setEnabled(true);
  te::qt::widgets::DataSourceSelectorDialog dlg(this);
  dlg.exec();

  std::list<te::da::DataSourceInfoPtr> dsPtrList = dlg.getSelecteds();

  if(dsPtrList.empty())
    return;

  std::list<te::da::DataSourceInfoPtr>::iterator it = dsPtrList.begin();

  m_ui->m_repositoryLineEdit->setText(QString(it->get()->getTitle().c_str()));

  m_outputDatasource = *it;

  m_toFile = false;
}

void te::vp::RegisterDialog::onTargetFileToolButtonPressed()
{
  m_ui->m_newLayerNameLineEdit->clear();
  m_ui->m_repositoryLineEdit->clear();

  te::qt::widgets::FileDialog fileDialog(this, te::qt::widgets::FileDialog::VECTOR);

  try {
    fileDialog.exec();
  }
  catch (te::common::Exception& ex) {
    QMessageBox::critical(this, tr("File information"), ex.what());
    return;
  }

  boost::filesystem::path filePath( fileDialog.getPath() );

  m_ui->m_repositoryLineEdit->setText(fileDialog.getPath().c_str());
  m_ui->m_newLayerNameLineEdit->setText( filePath.stem().string().c_str());
  
  m_toFile = true;
  m_ui->m_newLayerNameLineEdit->setEnabled(false);
}

void te::vp::RegisterDialog::onOkPushButtonClicked()
{
  if( m_selectedLayer.get() == nullptr )
  {
    QMessageBox::critical(this, "Register", tr("No input layer selected"));
    return;
  }

  if( m_tiePoints.empty() )
  {
    QMessageBox::critical(this, "Register", tr("No tie-point acquired"));
    return;
  }

  te::map::DataSetLayer* dsLayer = dynamic_cast<te::map::DataSetLayer*>(m_selectedLayer.get());
  if(!dsLayer)
  {
    QMessageBox::critical(this, tr("Register"), tr("Can not execute this operation on this type of layer."));
    return;
  }

  // Checking consistency of output paramenters
  if(m_ui->m_repositoryLineEdit->text().isEmpty())
  {
    QMessageBox::critical(this, tr("Register"), tr("Select a repository for the resulting layer."));
    return;
  }

  if(m_ui->m_newLayerNameLineEdit->text().isEmpty())
  {
    QMessageBox::critical(this, tr("Register"), tr("Define a name for the resulting layer."));
    return;
  }

  std::unique_ptr<te::da::DataSet> inDataset = dsLayer->getData();
  if( inDataset.get() == nullptr )
  {
    QMessageBox::critical(this, tr("Register"), tr("Invalid input dataset"));
    return;
  }

  std::unique_ptr<te::map::LayerSchema > inDatasetType = dsLayer->getSchema();
  if( inDatasetType.get() == nullptr )
  {
    QMessageBox::critical(this, tr("Register"), tr("Invalid input dataset type"));
    return;
  }

  std::unique_ptr< te::da::DataAccess > outDataAccessPtr;
  std::string outDataSetName =
     m_ui->m_newLayerNameLineEdit->text().toUtf8().data();
  if( m_toFile )
  {
    const std::string filePath = m_ui->m_repositoryLineEdit->text().toUtf8().data();

    te::core::URI outURI( "file://" + filePath );

    te::da::DataSourceInfoPtr dsInfoPtr(new te::da::DataSourceInfo);
    dsInfoPtr->setConnInfo( outURI.uri() );
    dsInfoPtr->setTitle( outDataSetName );
    dsInfoPtr->setAccessDriver("OGR");
    dsInfoPtr->setType("OGR");
    dsInfoPtr->setDescription( outDataSetName );
    dsInfoPtr->setId(te::common::GetUniqueID());

    te::da::DataSourceInfoManager::getInstance().add(dsInfoPtr);

    te::da::DataSourcePtr outputDataSourcePtr =
      te::da::DataSourceManager::getInstance().get( dsInfoPtr->getId(),
      dsInfoPtr->getType(), outURI );

    outDataAccessPtr.reset( new te::da::DataAccess( outputDataSourcePtr,
      outDataSetName ) );
  }
  else
  {
    te::da::DataSourcePtr outDataSource = te::da::GetDataSource(
      m_outputDatasource->getId());

    outDataAccessPtr.reset( new te::da::DataAccess( outDataSource, "" ) );
  }

  bool result = false;
  {
    te::qt::widgets::ScopedCursor cursor(Qt::WaitCursor);

    result =
      te::vp::Register::RegisterDataset(
        te::vp::Register::TransByNMethod,
        m_tiePoints,
        m_ui->m_minNeighborsSpinBox->text().toUInt(),
        true, // enable threads
        *inDataset,
        *inDatasetType,
        outDataSetName,
        *outDataAccessPtr
      );
  }

  if( result )
  {
    te::map::DataSetLayerPtr outDSL = te::qt::widgets::CreateLayerFromDataSet(
      outDataAccessPtr->getDataSource()->getId(), outDataSetName, 0 );

    m_outputLayer.reset( outDSL.detach() );

    QMessageBox::information(this, tr("Register"),
      tr("Register finished successfully") );
  }
  else
  {
    QMessageBox::critical(this, tr("Register"),
      tr("Register error") );
    return;
  }

  accept();
}

void te::vp::RegisterDialog::onCancelPushButtonClicked()
{
  reject();
}

void te::vp::RegisterDialog::onAcquirePushButtonClicked()
{
  if( m_selectedLayer.get() != nullptr )
  {
    TiePointsLocatorDialog tpLocDiag( this );
    tpLocDiag.setLayers( m_layers );
    tpLocDiag.setAdjLayer( m_selectedLayer );

    if( tpLocDiag.exec() == QDialog::Accepted )
    {
      tpLocDiag.getTiePoints( m_tiePoints );

      m_ui->m_tiePointsNumberLineEdit->setText( QString::number( m_tiePoints.size() ) );
    }
  }
}