/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

#include "TiePointsLocatorDialog.h"
#include "ui/ui_TiePointsLocatorDialogForm.h"
#include "../../maptools/DataSetLayer.h"
#include "../../qt/widgets/utils/ScopedCursor.h"
#include "../../qt/widgets/Utils.h"
#include "../../common/StringUtils.h"

#include <QDoubleValidator>
#include <QMessageBox>
#include <QFileDialog>

#include <limits>
#include <fstream>

#include <boost/filesystem.hpp>
#include <boost/lexical_cast.hpp>

#define TPLOCATORWINDOWNAME "Tie-points locator"
#define TPFILEHEADER "SRID1;X1;Y1;SRID2;X2;Y2"
#define TPFILEPATHALIASSTR "vp_tie_points_file_directory"

te::vp::TiePointsLocatorDialog::TiePointsLocatorDialog(QWidget* parent, Qt::WindowFlags f)
  : QDialog(parent, f),
    m_ui(new Ui::TiePointsLocatorDialogForm)
{
  m_ui->setupUi(this);

  m_ui->m_imgLabel->setPixmap(QIcon::fromTheme("register").pixmap(112,48));

  QSize iconSize(96, 48);

  m_ui->m_maxXOffset->setValidator( new QDoubleValidator( 0,
    std::numeric_limits< double >::max(), 10, nullptr ) );
  m_ui->m_maxYOffset->setValidator( new QDoubleValidator( 0,
    std::numeric_limits< double >::max(), 10, nullptr ) );
  m_ui->m_maxPercentDiff->setValidator( new QDoubleValidator( 0, 100, 10, nullptr ) );

  m_ui->m_maxYOffset->text().toDouble(),
  m_ui->m_maxPercentDiff->text().toDouble(),

  connect(m_ui->m_referenceLayerComboBox, SIGNAL(currentIndexChanged(int)), this,
    SLOT(onRefLayerComboBoxChanged(int)));
  connect(m_ui->m_adjustLayerComboBox, SIGNAL(currentIndexChanged(int)), this,
    SLOT(onAdjLayerComboBoxChanged(int)));
  connect(m_ui->m_locateTPPushButton, SIGNAL(clicked()), this, SLOT(onLocatePushButtonClicked()));
  connect(m_ui->m_clearTPPushButton, SIGNAL(clicked()), this, SLOT(onClearPushButtonClicked()));
  connect(m_ui->m_loadTPPushButton, SIGNAL(clicked()), this, SLOT(onLoadPushButtonClicked()));
  connect(m_ui->m_saveTPPushButton, SIGNAL(clicked()), this, SLOT(onSavePushButtonClicked()));
  connect(m_ui->m_closePushButton, SIGNAL(clicked()), this, SLOT(onClosePushButtonClicked()));
  
  m_ui->m_helpPushButton->setNameSpace("dpi.inpe.br.plugins"); 
  m_ui->m_helpPushButton->setPageReference("plugins/vp/vp_tiepointslocator.html");
}

te::vp::TiePointsLocatorDialog::~TiePointsLocatorDialog() = default;

void te::vp::TiePointsLocatorDialog::setLayers(std::list<te::map::AbstractLayerPtr> layers)
{
  m_ui->m_referenceLayerComboBox->clear();
  m_ui->m_adjustLayerComboBox->clear();

  m_layers = layers;
  
  std::list<te::map::AbstractLayerPtr>::iterator it = m_layers.begin();

  while(it != m_layers.end())
  {
    std::unique_ptr<te::da::DataSetType> dsType = it->get()->getSchema();

    if(
        ( it->get()->getType() == "DATASETLAYER" )
        &&
        ( dsType.get() && dsType->hasGeom() )
      )
    {
      m_ui->m_referenceLayerComboBox->addItem(QString(it->get()->getTitle().c_str()),
        QVariant(it->get()->getId().c_str()));
      m_ui->m_adjustLayerComboBox->addItem(QString(it->get()->getTitle().c_str()),
        QVariant(it->get()->getId().c_str()));
    }

    ++it;
  }

  m_ui->m_referenceLayerComboBox->setDisabled( false );
  m_ui->m_adjustLayerComboBox->setDisabled( false );
}

void te::vp::TiePointsLocatorDialog::onRefLayerComboBoxChanged(int index)
{
  std::list<te::map::AbstractLayerPtr>::iterator it = m_layers.begin();
  std::string layerID = m_ui->m_referenceLayerComboBox->itemData(index,
    Qt::UserRole).toString().toUtf8().data();

  while(it != m_layers.end())
  {
    if(layerID == it->get()->getId())
    {
      te::map::AbstractLayerPtr selectedLayer = it->get();
      m_selectedRefLayer = selectedLayer;

      return;
    }

    ++it;
  }
}

void te::vp::TiePointsLocatorDialog::onAdjLayerComboBoxChanged(int index)
{
  std::list<te::map::AbstractLayerPtr>::iterator it = m_layers.begin();
  std::string layerID = m_ui->m_adjustLayerComboBox->itemData(index,
    Qt::UserRole).toString().toUtf8().data();

  while(it != m_layers.end())
  {
    if(layerID == it->get()->getId())
    {
      te::map::AbstractLayerPtr selectedLayer = it->get();
      m_selectedAdjLayer = selectedLayer;

      return;
    }

    ++it;
  }
}

void te::vp::TiePointsLocatorDialog::onLocatePushButtonClicked()
{
  te::map::DataSetLayer* rafDsLayer = dynamic_cast<te::map::DataSetLayer*>(m_selectedRefLayer.get());
  if( rafDsLayer == nullptr ) return;

  te::map::DataSetLayer* adjDsLayer = dynamic_cast<te::map::DataSetLayer*>(m_selectedAdjLayer.get());
  if( adjDsLayer == nullptr ) return;

  std::unique_ptr<te::da::DataSet> refDs = rafDsLayer->getData();
  if( refDs.get() == nullptr ) return;

  std::unique_ptr<te::da::DataSet> adjfDs = adjDsLayer->getData();
  if( adjfDs.get() == nullptr ) return;

  std::unique_ptr<te::map::LayerSchema > refDsType = rafDsLayer->getSchema();
  if( refDsType.get() == nullptr ) return;

  std::unique_ptr<te::map::LayerSchema > adjDsType = adjDsLayer->getSchema();
  if( adjDsType.get() == nullptr ) return;

  std::vector< te::vp::TiePointsLocator::TiePointT > tiePoints;

  bool result = false;
  {
    te::qt::widgets::ScopedCursor cursor(Qt::WaitCursor);
    result =
      te::vp::TiePointsLocator::MatchDataSets(
        *adjfDs,
        *adjDsType,
        *refDs,
        *refDsType,
        true, // enable threads
        m_ui->m_maxXOffset->text().toDouble(),
        m_ui->m_maxYOffset->text().toDouble(),
        m_ui->m_maxPercentDiff->text().toDouble(),
        tiePoints
      );
  }

  if( result )
  {
    m_tiePoints.insert( m_tiePoints.end(), tiePoints.begin(), tiePoints.end() );

    m_ui->m_currentTPNmbLineEdit->setText( QString::number( m_tiePoints.size() ) );

    QMessageBox::information(this, tr(TPLOCATORWINDOWNAME),
      tr("Found tie-points:") + QString::number( tiePoints.size() ) );
  }
  else
  {
    QMessageBox::critical(this, tr(TPLOCATORWINDOWNAME),
      tr("Tie-points location error ") );
  }
}

void te::vp::TiePointsLocatorDialog::onClearPushButtonClicked()
{
  m_tiePoints.clear();
  m_ui->m_currentTPNmbLineEdit->setText("0");
}

void te::vp::TiePointsLocatorDialog::onLoadPushButtonClicked()
{
  QString tiePointsFileName = QFileDialog::getOpenFileName(this, tr("Select the input file name"), 
    te::qt::widgets::GetFilePathFromSettings( TPFILEPATHALIASSTR ), 
    tr("Tie-points file (*.txt)"), 0 ,QFileDialog::ReadOnly);
  
  if( !tiePointsFileName.isEmpty() )
  {
    boost::filesystem::path fullFilePath( tiePointsFileName.toUtf8().data() );
    te::qt::widgets::AddFilePathToSettings(fullFilePath.parent_path().string().c_str(), 
      TPFILEPATHALIASSTR );
    
    std::ifstream tiePointsFile;
    tiePointsFile.open( tiePointsFileName.toUtf8().data(), std::ofstream::in );
    if( !tiePointsFile.good() )
    {
      QMessageBox::critical(this, tr(TPLOCATORWINDOWNAME), tr("Tie points file open error"));
      return;
    }
   
    try
    {
      std::string lineStr;
      bool headerOk = false;
      std::vector< std::string > tokens;
      te::vp::TiePointsLocator::TiePointT auxTP;
      
      while( !tiePointsFile.eof() )
      {
        std::getline( tiePointsFile, lineStr );
        
        if( headerOk )
        {
          tokens.clear();
          te::common::Tokenize( lineStr, tokens, ";" );
          
          if( tokens.size() == 6 )
          {
            std::get< 0 >( auxTP ) = boost::lexical_cast< int >( tokens[ 0 ] );
            std::get< 1 >( auxTP ) = boost::lexical_cast< double >( tokens[ 1 ] );
            std::get< 2 >( auxTP ) = boost::lexical_cast< double >( tokens[ 2 ] );
            std::get< 3 >( auxTP ) = boost::lexical_cast< int >( tokens[ 3 ] );
            std::get< 4 >( auxTP ) = boost::lexical_cast< double >( tokens[ 4 ] );
            std::get< 5 >( auxTP ) = boost::lexical_cast< double >( tokens[ 5 ] );
            
            m_tiePoints.push_back( auxTP );
          }
          else
          {
            QMessageBox::critical(this, tr(TPLOCATORWINDOWNAME), tr("Tie points file read error"));
            return;
          }
        }
        else
        {
          if( lineStr == TPFILEHEADER )
          {
            headerOk = true;
          }
          else
          {
            QMessageBox::critical(this, tr(TPLOCATORWINDOWNAME), tr("Tie points file read error"));
            return;
          }
        }
      }
    }
    catch(...)
    {
      QMessageBox::critical(this, tr(TPLOCATORWINDOWNAME), tr("Tie points file read error"));
      return;
    }
    
    m_ui->m_currentTPNmbLineEdit->setText( QString::number( m_tiePoints.size() ) );
  }
}

void te::vp::TiePointsLocatorDialog::onSavePushButtonClicked()
{
  if( m_tiePoints.empty() )
  {
    QMessageBox::critical(this, tr(TPLOCATORWINDOWNAME),
      tr("There are no tie-points to save") );
    return;
  }
  
  QString tiePointsFileName = QFileDialog::getSaveFileName(this, 
    tr("Select the output file name"), 
    te::qt::widgets::GetFilePathFromSettings( TPFILEPATHALIASSTR ), 
    tr("Tie-points file (*.txt)"), 0, 0);
  
  if( !tiePointsFileName.isEmpty() )
  {
    boost::filesystem::path fullFilePath( tiePointsFileName.toUtf8().data() );
    te::qt::widgets::AddFilePathToSettings(fullFilePath.parent_path().string().c_str(), 
      TPFILEPATHALIASSTR );   
    
    std::ofstream tiePointsFile;
    tiePointsFile.open( tiePointsFileName.toUtf8().data(), std::ofstream::out | 
      std::ofstream::trunc );
  
    if( !tiePointsFile.good() )
    {
      QMessageBox::critical(this, tr(TPLOCATORWINDOWNAME), tr("Tie points file creation error"));
      return;
    }
    
    tiePointsFile.precision( 40 );

    tiePointsFile << TPFILEHEADER;

    const std::size_t tiePointsSize = m_tiePoints.size();
    
    for( std::size_t tiePointsIdx = 0 ; tiePointsIdx < tiePointsSize ;
      ++tiePointsIdx )
    {
      const te::vp::TiePointsLocator::TiePointT tp = m_tiePoints[ tiePointsIdx ];
      
      tiePointsFile 
        << std::endl
        << std::get< 0 >( tp )
        << ";" << std::get< 1 >( tp )
        << ";" << std::get< 2 >( tp )
        << ";" << std::get< 3 >( tp )
        << ";" << std::get< 4 >( tp )
        << ";" << std::get< 5 >( tp );
    }    
  }
}

void te::vp::TiePointsLocatorDialog::onClosePushButtonClicked()
{
  accept();
}

void te::vp::TiePointsLocatorDialog::setRefLayer( te::map::AbstractLayerPtr newReflayer)
{
  if(
      ( newReflayer.get() != nullptr )
      &&
      ( newReflayer.get()->getType() == "DATASETLAYER" )
      &&
      ( newReflayer.get()->getSchema().get() != nullptr )
      &&
      ( newReflayer.get()->getSchema()->hasGeom() )
    )
  {
    m_selectedRefLayer = newReflayer;
    m_ui->m_referenceLayerComboBox->setDisabled( true );
    m_ui->m_referenceLayerComboBox->clear();
    m_ui->m_referenceLayerComboBox->addItem(QString(newReflayer->getTitle().c_str()),
      QVariant(newReflayer->getId().c_str()));
  }
}

void te::vp::TiePointsLocatorDialog::setAdjLayer( te::map::AbstractLayerPtr newAdjlayer)
{
  if(
      ( newAdjlayer.get() != nullptr )
      &&
      ( newAdjlayer.get()->getType() == "DATASETLAYER" )
      &&
      ( newAdjlayer.get()->getSchema().get() != nullptr )
      &&
      ( newAdjlayer.get()->getSchema()->hasGeom() )
    )
  {
    m_selectedAdjLayer = newAdjlayer;
    m_ui->m_adjustLayerComboBox->setDisabled( true );
    m_ui->m_adjustLayerComboBox->clear();
    m_ui->m_adjustLayerComboBox->addItem(QString(newAdjlayer->getTitle().c_str()),
      QVariant(newAdjlayer->getId().c_str()));
  }
}

void te::vp::TiePointsLocatorDialog::getTiePoints(
  std::vector< te::vp::TiePointsLocator::TiePointT >& tiePoints ) const
{
  tiePoints = m_tiePoints;
}
