/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

This file is part of the TerraLib - a Framework for building GIS enabled applications.

TerraLib is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

TerraLib is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with TerraLib. See COPYING. If not, write to
TerraLib Team at <terralib-team@terralib.org>.
*/

/*!
 * \file src/terralib/vp/qt/VectorProcessingOperationDialog.h
 *
 * \brief Implements a dialog used to execute a vector processing operation
 */

#ifndef __TERRALIB_VP_QT_INTERNAL_VECTORPROCESSINGDIALOGCAPABILITIES_H
#define __TERRALIB_VP_QT_INTERNAL_VECTORPROCESSINGDIALOGCAPABILITIES_H

#include "../Config.h"

#include "../../common/AbstractFactory.h"

//QT Includes
#include <QIcon>
#include <QPixmap>

//STL includes
#include <memory>
#include <string>

class QWidget;

namespace te
{
  namespace vp
  {
    namespace qt
    {
      class SpecificParametersWidget;

      /*!
        \class VectorProcessingDialogCapabilities

        \brief A class used to define the base capabilities of an operation, including number of inputs and outputs.
      */
      class TEVPEXPORT VectorProcessingDialogCapabilities
      {
      public:
        VectorProcessingDialogCapabilities( 
          const std::string& operationName
          , const std::string& caption
          , const std::string& description
          , const std::string& icon
          , const std::string& hintPixmap
          , const std::string& helpPageNamespace
          , const std::string& helpPageReference
          , bool enablePropertySelection = true
          );

        //!< Destructor
        virtual ~VectorProcessingDialogCapabilities();

        //!< Returns the name of the operation
        const std::string& getName() const;

        //!< Returns the caption of the operation
        const std::string& getCaption() const;

        //!< Returns the description of the operation
        const std::string& getDescription() const;

        //!< Returns the icon of the operation
        const QIcon& getIcon() const;

        //!< Returns the hint pixmap of the operation
        const QPixmap& getHintPixmap() const;

        //!< Returns help page namespace
        const std::string& getHelpPageNamespace() const;

        //!< Returns help page referece
        const std::string& getHelpPageReference() const;

        //!< Returns the configuration enable property selection
        bool isPropertySelectionEnable() const;

        //!< Returns the widget contiaing the specific inputs of the operation. Returns nullptr if there aren't specific parameters for the operation
        virtual SpecificParametersWidget* getParametersWidget(QWidget* parent = nullptr) const;

      protected:

        std::string m_operationName; //!< The name of the operation
        std::string m_caption; //!< The caption of the widget
        std::string m_description; //!< The caption of the widget
        std::string m_icon; //!< The icon of the operation
        std::string m_hintPixmap; //!< The hint pixmap to be used in the widget
        std::string m_helpPageNamespace; //!< The help page namespace
        std::string m_helpPageReference; //!< The help page referece
        QIcon       m_qIcon; //!< The icon of the operation
        QPixmap     m_qHintPixmap; //!< The hint pixmap to be used in the widget
        bool        m_enablePropertySelection; //!< If TRUE, it allows the user to select the properties that will be used to create the output DataSet
      };


      /*!
        \class VectorProcessingDialogCapabilitiesFactory

        \brief Abstract factory used to create VectorProcessingDialogCapabilities.
      */
      class TEVPEXPORT VectorProcessingDialogCapabilitiesFactory : public te::common::AbstractFactory < VectorProcessingDialogCapabilities, std::string >
      {
      protected:

        /*!
          \brief Constructor.

          \param factoryKey The key that identifies the factory.
        */
        VectorProcessingDialogCapabilitiesFactory(const std::string& factoryKey);

        //!< Destructor
        virtual ~VectorProcessingDialogCapabilitiesFactory();
      };

    } //end namespace qt
  }//end namespace vp
}//end namespace te


#endif //__TERRALIB_VP_QT_INTERNAL_VECTORPROCESSINGDIALOGCAPABILITIES_H
