/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

This file is part of the TerraLib - a Framework for building GIS enabled applications.

TerraLib is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

TerraLib is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with TerraLib. See COPYING. If not, write to
TerraLib Team at <terralib-team@terralib.org>.
*/

/*!
 * \file src/terralib/vp/qt/VectorProcessingOperationDialog.cpp
 *
 * \brief Implements a dialog used to execute a vector processing operation
 */

#include "VectorProcessingOperationDialog.h"
#include "InputDataAccessWidget.h"
#include "OutputDataAccessWidget.h"
#include "SpecificParametersWidget.h"
#include "ui/ui_VectorProcessingOperationDialogForm.h"

// TerraLib Includes
#include "../../common/Exception.h"
#include "../../core/translator/Translator.h"
#include "../../core/utils/Platform.h"
#include "../../dataaccess/utils/DataAccess.h" 
#include "../../qt/widgets/Utils.h"
#include "../../qt/widgets/layer/utils/DataSet2Layer.h"
#include "../../qt/widgets/progress/ProgressViewerTaskWidget.h"
#include "../../vp/qt/Utils.h"
#include "../../vp/qt/VectorProcessingDialogCapabilities.h"
#include "../../vp/AbstractDataHandler.h"
#include "../../vp/AbstractOperation.h"
#include "../../vp/AlgorithmParams.h"
#include "../../vp/AbstractOperation.h"
#include "../../vp/OperationThreadManager.h"
#include "../../vp/OperationReport.h"
#include "../../vp/NoGroupingDataHandler.h"
#include "../../vp/SubdivideDataHandler.h"
#include "../../vp/Utils.h"

// Qt Includes
#include <QGridLayout>
#include <QMessageBox>
#include <QMetaType>
#include <QTextStream>

Q_DECLARE_METATYPE(te::vp::ProgressState)
Q_DECLARE_METATYPE(te::vp::OperationResult)

class CurrentExecutionImpl
{
public:

  ~CurrentExecutionImpl()
  {
    te::common::FreeContents(m_vecInputDataAccess);
    te::common::FreeContents(m_vecOutputDataAccess);
  }

  std::unique_ptr<te::vp::OperationThreadManager> m_operationTM;

  std::unique_ptr< te::common::TaskProgress > m_task;

  std::vector<te::da::DataAccess*> m_vecInputDataAccess;

  std::vector<te::da::DataAccess*> m_vecOutputDataAccess;
};

te::vp::qt::VectorProcessingOperationDialog::VectorProcessingOperationDialog(QWidget* parent, Qt::WindowFlags f)
  : QDialog(parent, f | Qt::WindowMaximizeButtonHint)
  , m_ui(new Ui::VectorProcessingOperationDialogForm)
  , m_specificParamsWidget(nullptr)
{
  qRegisterMetaType<te::vp::ProgressState>();
  qRegisterMetaType<te::vp::OperationResult>();
   
  m_ui->setupUi(this); 

  m_ui->m_executeOperationToolButton->setIcon(QIcon::fromTheme("media-playback-start-green"));

  m_ui->m_parametersWidget->setVisible(false);

  //build input data widget
  m_inputDataWidget = new te::vp::qt::InputDataAccessWidget(m_ui->m_inputWidget);

  QGridLayout* layoutIn = new QGridLayout(m_ui->m_inputWidget);
  layoutIn->addWidget(m_inputDataWidget);
  layoutIn->setContentsMargins(0, 0, 0, 0);

  //build output data widget
  m_outputDataWidget = new te::vp::qt::OutputDataAccessWidget(m_ui->m_outputWidget);

  QGridLayout* layoutOut = new QGridLayout(m_ui->m_outputWidget);
  layoutOut->addWidget(m_outputDataWidget);
  layoutOut->setContentsMargins(0, 0, 0, 0);

  //build progress widget
  m_progressViewer = new te::qt::widgets::ProgressViewerTaskWidget(m_ui->m_progressViewerWidget, false, TE_TR("Vector Processing Progress Bar"), "item-remove");
  
  QGridLayout* layout = new QGridLayout(m_ui->m_progressViewerWidget);
  layout->setContentsMargins(0, 0, 0, 0);
  layout->addWidget(m_progressViewer);
  
  // fill geom out type combo box
  m_ui->m_OutGeomTypeComboBox->addItem( " ", QVariant( te::gm::UnknownGeometryType ) );
  m_ui->m_OutGeomTypeComboBox->addItem( "PointType", QVariant( te::gm::PointType ) );
  m_ui->m_OutGeomTypeComboBox->addItem( "LineStringType", QVariant( te::gm::LineStringType ) );
  m_ui->m_OutGeomTypeComboBox->addItem( "CircularStringType", QVariant( te::gm::CircularStringType ) );
  m_ui->m_OutGeomTypeComboBox->addItem( "CompoundCurveType", QVariant( te::gm::CompoundCurveType ) );
  m_ui->m_OutGeomTypeComboBox->addItem( "PolygonType", QVariant( te::gm::PolygonType ) );
  m_ui->m_OutGeomTypeComboBox->addItem( "CurvePolygonType", QVariant( te::gm::CurvePolygonType ) );
  m_ui->m_OutGeomTypeComboBox->addItem( "MultiPointType", QVariant( te::gm::MultiPointType ) );
  m_ui->m_OutGeomTypeComboBox->addItem( "MultiLineStringType", QVariant( te::gm::MultiLineStringType ) );
  m_ui->m_OutGeomTypeComboBox->addItem( "MultiPolygonType", QVariant( te::gm::MultiPolygonType ) );
  m_ui->m_OutGeomTypeComboBox->addItem( "MultiSurfaceType", QVariant( te::gm::MultiSurfaceType ) );
  m_ui->m_OutGeomTypeComboBox->addItem( "PolyhedralSurfaceType", QVariant( te::gm::PolyhedralSurfaceType ) );
  m_ui->m_OutGeomTypeComboBox->addItem( "TINType", QVariant( te::gm::TINType ) );
  m_ui->m_OutGeomTypeComboBox->addItem( "TriangleType", QVariant( te::gm::TriangleType ) );  

  //connects
  connect(m_ui->m_executeOperationToolButton, SIGNAL(pressed()), this, SLOT(onExecuteOperationToolButtonPressed()));
  connect(&m_processListener, SIGNAL(processingStateChangedSignal(const te::vp::ProgressState&)), this, SLOT(onProcessingStateChanged(const te::vp::ProgressState&)));
  connect(&m_processListener, SIGNAL(processingFinishedSignal(const te::vp::OperationResult&)), this, SLOT(onProcessingFinished(const te::vp::OperationResult&)));
  connect(m_progressViewer, SIGNAL(taskCancelled()), this, SLOT(taskCancelled()));
  connect(m_ui->m_closePushButton, SIGNAL(released()), this, SLOT(onClose()));
  connect(m_inputDataWidget, SIGNAL(dataAccessSelectionChanged()), this, SLOT(onDataAccessSelectionChanged()));
}

te::vp::qt::VectorProcessingOperationDialog::~VectorProcessingOperationDialog()
{
}

Ui::VectorProcessingOperationDialogForm* te::vp::qt::VectorProcessingOperationDialog::getForm() const
{
  return m_ui.get();
}

void te::vp::qt::VectorProcessingOperationDialog::init(const std::string& operationName, const std::list<te::map::AbstractLayerPtr>& layers)
{
  m_operationName = operationName;

  std::unique_ptr<te::vp::AbstractOperation> operation(te::vp::CreateOperation(m_operationName));
  std::unique_ptr<te::vp::AbstractOperationCapabilities> operationCapabilities = operation->getCapabilities();
  std::unique_ptr<te::vp::qt::VectorProcessingDialogCapabilities> dialogCapabilities(te::vp::CreateDialogCapabilities(operationName));

  //build operation parameters
  m_specificParamsWidget = dialogCapabilities->getParametersWidget(m_ui->m_parametersWidget);
  if (m_specificParamsWidget != nullptr)
  { 
    QWidget* paramsWidget = m_specificParamsWidget->getWidget();
    
    m_ui->m_parametersWidget->setVisible(true);

    QGridLayout* paramsLayout = new QGridLayout(m_ui->m_parametersWidget);
    paramsLayout->setContentsMargins(0, 0, 0, 0);
    paramsLayout->addWidget(paramsWidget);
  }

  //updates the dialog
  m_inputDataWidget->setNumbersOfInputDataAccess(operationCapabilities->getNumberOfInputs(), layers);
  m_outputDataWidget->setNumbersOfOutputDataAccess(operationCapabilities->getNumberOfOutputs());

  bool enablePropertySelection = dialogCapabilities->isPropertySelectionEnable();
  m_inputDataWidget->setFilterPropertiesEnabled(enablePropertySelection);

  m_ui->m_operationTitleLabel->setText(te::qt::widgets::Convert2QString(dialogCapabilities->getCaption()));
  m_ui->m_operationDescLabel->setText(te::qt::widgets::Convert2QString(dialogCapabilities->getDescription()));

  QPixmap pixmap = dialogCapabilities->getHintPixmap();
  m_ui->m_operationImageLabel->setPixmap(pixmap);

  std::string helpRef = dialogCapabilities->getHelpPageReference();
  std::string helpFileName = te::core::FindInTerraLibPath(helpRef);

  if (helpFileName.empty() == false)
  {
    QUrl url(te::qt::widgets::Convert2QString("file://" + helpFileName));

    m_ui->m_textBrowser->setSource(url);

    m_ui->m_textBrowser->setReadOnly(true);
  }
}

te::vp::OperationResult te::vp::qt::VectorProcessingOperationDialog::getResult() const
{
  if (m_run == nullptr || m_run->m_operationTM == nullptr || m_run->m_operationTM->getOperation() == nullptr)
  {
    return te::vp::OperationResult::OP_NOT_STARTED;
  }

  return (m_run->m_operationTM->getOperation()->getParameters()->getOperationResult());
}

std::list<te::map::AbstractLayerPtr> te::vp::qt::VectorProcessingOperationDialog::getOutputLayers()
{
  std::list<te::map::AbstractLayerPtr> outputLayerList;

  for (std::size_t i = 0; i < m_vecOverallOutputDataAccess.size(); ++i)
  {
    te::da::DataSourcePtr dataSource = m_vecOverallOutputDataAccess.at(i)->getDataSource();
    te::da::DataSetTypePtr dsType(static_cast<te::da::DataSetType*>(m_vecOverallOutputDataAccess.at(i)->getDataSetType()->clone()));

    te::qt::widgets::DataSet2Layer ds2layer(dataSource->getId());

    te::map::AbstractLayerPtr layer = ds2layer(dsType);
    outputLayerList.push_back(layer);
  }

  return outputLayerList;
}

void te::vp::qt::VectorProcessingOperationDialog::keyPressEvent(QKeyEvent* event)
{
  if (event->key() == Qt::Key_Escape)
  {
    event->ignore();
  }
  else
  {
    event->accept();
  }
}

void te::vp::qt::VectorProcessingOperationDialog::setDialogOnExecution(bool onExecution)
{
  m_inputDataWidget->setEnabled(!onExecution);
  m_outputDataWidget->setEnabled(!onExecution);
  if (onExecution == false)
  {
    m_ui->m_executeOperationToolButton->setChecked(false);
    m_ui->m_executeOperationToolButton->raise();
  }

  m_ui->m_executeOperationToolButton->setEnabled(!onExecution);
  m_ui->m_closePushButton->setEnabled(!onExecution);
}

te::vp::AbstractParameters* te::vp::qt::VectorProcessingOperationDialog::createParametersForOperation(te::vp::AbstractOperation* operation, QString& errorMessage)
{
  //we first create the parameters based on the operation
  std::unique_ptr<te::vp::AbstractParameters> parameters = operation->getCapabilities()->createParameters();

  //then we get the input and output dataaccess
  std::vector<te::da::DataAccessPtr> vecInput;
  std::vector<te::da::DataAccessPtr> vecOutput;
  try
  {
    //get input dataaccess
    vecInput = m_inputDataWidget->getDataAccessVector();

    //get output dataaccess
    vecOutput = m_outputDataWidget->getDataAccessVector();
  }
  catch (const std::exception& e)
  {
    errorMessage = tr("Could not initialize the execution for the operation: '") + te::qt::widgets::Convert2QString(m_operationName) + "'";
    errorMessage += "\n" + tr("Error: ") + te::qt::widgets::Convert2QString(e.what());
    return nullptr;
  }

  //we must check if the input layers are not the same and add them to the parameters
  std::set<std::string> setInputNames;
  for (std::size_t i = 0; i < vecInput.size(); ++i)
  {
    std::string dataSetName = vecInput.at(i)->getDataSetName();
    if (setInputNames.find(dataSetName) != setInputNames.end())
    {
      errorMessage = tr("The selected layers must not be the same!");
      return nullptr;
    }
    setInputNames.insert(dataSetName);

    parameters->addInputDataAccess(vecInput.at(i).release());
  }

  //we must check if the output layers are not the same and add them to the parameters
  std::set<std::string> setOutputNames;
  for (std::size_t i = 0; i < vecOutput.size(); ++i)
  {
    std::string dataSetName = vecOutput.at(i)->getDataSetName();
    if (setOutputNames.find(dataSetName) != setOutputNames.end())
    {
      errorMessage = tr("The selected layers must not be the same!");
      return nullptr;
    }
    setOutputNames.insert(dataSetName);

    parameters->addOutputDataAccess(vecOutput.at(i).release());
  }

  //we continue the initialization with the specific parameters, if applyable
  try
  {
    if (m_specificParamsWidget != nullptr)
    {
      m_specificParamsWidget->setValuesToParameters(parameters.get());
    }
  }
  catch (const te::common::Exception& exception)
  {
    errorMessage = exception.what();
    return nullptr;
  }

  return parameters.release();
}

void te::vp::qt::VectorProcessingOperationDialog::onProcessingStateChanged(const te::vp::ProgressState& newState)
{
  switch (newState)
  {
  case te::vp::ProgressState::PRE_PROCESSING_STARTED:
    m_progressViewer->setMessage(TE_TR("Initializing Process"));
    m_run->m_task->setTotalSteps(0); //makes the bar to show some motion
    break;
  case te::vp::ProgressState::PROCESSING_STARTED:
    m_progressViewer->setMessage(TE_TR("Processing"));
    break;
  case te::vp::ProgressState::POST_PROCESSING_STARTED:
    m_progressViewer->setMessage(TE_TR("Finalizing Process"));
    m_run->m_task->setTotalSteps(0); //makes the bar to show some motion
    break;
  case te::vp::ProgressState::POST_PROCESSING_FINISHED:
    m_progressViewer->setMessage(TE_TR("Process Finalized"));
    break;
  case te::vp::ProgressState::CANCEL_REQUESTED:
    m_progressViewer->setMessage(TE_TR("Cancel Requested. Stopping the execution. Please Wait!"));
    break;

  default:
    break;
  }

  m_progressViewer->setTask(m_run->m_task.get());
  m_progressViewer->setTotalValues(m_run->m_task->getId());

  //report results
  m_ui->m_reportPlainTextEdit->clear();
  m_ui->m_reportPlainTextEdit->appendPlainText(te::qt::widgets::Convert2QString(m_run->m_operationTM->getOperation()->getReport()->toString()));
}

void te::vp::qt::VectorProcessingOperationDialog::onProcessingFinished(const te::vp::OperationResult& result)
{
  //report results
  m_ui->m_reportPlainTextEdit->clear();
  m_ui->m_reportPlainTextEdit->appendPlainText(te::qt::widgets::Convert2QString(m_run->m_operationTM->getOperation()->getReport()->toString()));

  m_progressViewer->removeTask(m_run->m_task->getId());
  m_run->m_task.reset();

  setDialogOnExecution(false);
  m_outputDataWidget->reset();

  bool addOutput = false;

  if (result == OperationResult::OP_SUCCESS)
  {
    addOutput = true;
    QMessageBox::information(this, tr("Vector Processing"), tr("Finished with success."));
  }
  else if (result == OperationResult::OP_CANCELLED)
  {
    addOutput = true;
    QMessageBox::warning(this, tr("Vector Processing"), tr("Operation cancelled by the user. Check the report for more details"));
  }
  else if (result == OperationResult::OP_ERROR)
  {
    QMessageBox::critical(this, tr("Vector Processing"), tr("Error executing the operation. Check the report for more details"));
  }
  
  //adds the output from the current execution to the overall output
  if (addOutput == true)
  {
    for (std::size_t i = 0; i < m_run->m_vecOutputDataAccess.size(); ++i)
    {
      te::da::DataAccessPtr dataAccess(m_run->m_vecOutputDataAccess.at(i));
      m_vecOverallOutputDataAccess.push_back(std::move(dataAccess));
    }

    m_run->m_vecOutputDataAccess.clear();
  }
}

void te::vp::qt::VectorProcessingOperationDialog::taskCancelled()
{
  onProcessingStateChanged(te::vp::ProgressState::CANCEL_REQUESTED);
  m_run->m_operationTM->stop();
}

void te::vp::qt::VectorProcessingOperationDialog::onClose()
{
  accept();
}

void te::vp::qt::VectorProcessingOperationDialog::onExecuteOperationToolButtonPressed()
{
  //create operation
  std::unique_ptr<te::vp::AbstractOperation> operation(te::vp::CreateOperation(m_operationName));
  if (operation == nullptr)
  {
    setDialogOnExecution(false);
    QString message = tr("Could not initialize the execution for the operation: '") + te::qt::widgets::Convert2QString(m_operationName) + "'";
    message += "\n" + tr("Error: No operation was found with the given name");

    QMessageBox::information(this, tr("Vector Processing"), message);
    m_ui->m_executeOperationToolButton->setChecked(false);
    return;
  }

  QString qErrorMessage;
  std::unique_ptr<te::vp::AbstractParameters> parameters(createParametersForOperation(operation.get(), qErrorMessage));
  if (parameters == nullptr)
  {
    setDialogOnExecution(false);
    QMessageBox::critical(this, tr("Vector Processing"), qErrorMessage);
    m_ui->m_executeOperationToolButton->setChecked(false);
    return;
  }

  std::string errorMessage;
  if (parameters->isValid(errorMessage) == false)
  {
    QMessageBox::information(this, tr("Vector Processing"), tr("Some of the parameters are invalid. The error message is: ") + te::qt::widgets::Convert2QString(errorMessage));
    m_ui->m_executeOperationToolButton->setChecked(false);
    return;
  }

  m_run.reset(new CurrentExecutionImpl());
  m_run->m_vecInputDataAccess = parameters->getInputDataAccessVector();
  m_run->m_vecOutputDataAccess = parameters->getOutputDataAccessVector();

  operation->setParameters(parameters.release());
  
  if( 
      ((te::gm::GeomType)m_ui->m_OutGeomTypeComboBox->currentData().toInt())
      !=  
      te::gm::UnknownGeometryType
    )
  {
    operation->setOutputGeometryType( 
      ((te::gm::GeomType)m_ui->m_OutGeomTypeComboBox->currentData().toInt()) );
  }

  setDialogOnExecution(true);

  //run operation
  try
  {
    te::vp::AbstractDataHandler* dataHandler = te::vp::CreateDataHandler(operation->getCapabilities()->getDataHandlerName());

    m_run->m_operationTM.reset(new te::vp::OperationThreadManager(operation.release(), dataHandler));

    m_run->m_task.reset(new te::common::TaskProgress());
    m_run->m_operationTM->setTask(m_run->m_task.get());
    m_run->m_operationTM->startAsync(&m_processListener);
  }
  catch (...)
  {
    QMessageBox::warning(this, tr("Vector Processing"), tr("Unexpected Error... Sorry :("));
    m_ui->m_executeOperationToolButton->setChecked(false);
  }
}

void te::vp::qt::VectorProcessingOperationDialog::onDataAccessSelectionChanged()
{
  //get specific parameters
  if (m_specificParamsWidget != nullptr)
  {
    std::vector<te::da::DataAccessPtr> vecInputData = m_inputDataWidget->getDataAccessVector();
    m_specificParamsWidget->setInputData(vecInputData);
  }
}
