/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

This file is part of the TerraLib - a Framework for building GIS enabled applications.

TerraLib is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

TerraLib is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with TerraLib. See COPYING. If not, write to
TerraLib Team at <terralib-team@terralib.org>.
*/

/*!
 * \file src/terralib/vp/qt/VectorProcessingOperationDialog.h
 *
 * \brief Implements a dialog used to execute a vector processing operation
 */

#ifndef __TERRALIB_VP_QT_INTERNAL_VECTORPROCESSINGOPERATIONDIALOG_H
#define __TERRALIB_VP_QT_INTERNAL_VECTORPROCESSINGOPERATIONDIALOG_H

#include "../Config.h"

 // TerraLib Includes
#include "../../maptools/AbstractLayer.h"
#include "../OperationThreadManagerListener.h"

// STL Includes
#include <memory>

 // Qt Includes
#include <QDialog>
#include <QPixmap>

namespace Ui { class VectorProcessingOperationDialogForm; }

class CurrentExecutionImpl;

namespace te
{
  namespace common
  {
    class TaskProgress;
  }

  namespace da
  {
    class DataAccess;
    class DataAccessPtr;
  }

  namespace qt
  {
    namespace widgets
    {
      class ProgressViewerTaskWidget;
    }
  }

  namespace vp
  {
    class AbstractOperation;
    class AbstractParameters;
    class OperationThreadManager;
    enum class ProgressState;
    enum class OperationResult;

    namespace qt
    {
      class InputDataAccessWidget;
      class OutputDataAccessWidget;
      class SpecificParametersWidget;

      class TEVPEXPORT QtOperationThreadManagerListener : public QObject, public te::vp::OperationThreadManagerListener
      {
        Q_OBJECT

      signals:

        void processingStateChangedSignal(const te::vp::ProgressState& newState);

        void processingFinishedSignal(const te::vp::OperationResult& result);

      public:

        void processingStateChanged(const te::vp::ProgressState& newState) override
        {
          emit processingStateChangedSignal(newState);
        } 

        void processingFinished(const te::vp::OperationResult& result) override
        {
          emit processingFinishedSignal(result);
        }
      };
      

      /*!
        \class VectorProcessingOperationDialog

        \brief  Implements a dialog used to execute a vector processing operation
      */
      class TEVPEXPORT VectorProcessingOperationDialog : public QDialog
      {
        Q_OBJECT

        public:

          /*!
            \brief Constructor.
          */
          VectorProcessingOperationDialog(QWidget* parent = 0, Qt::WindowFlags f = 0);

          /*!
            \brief Destructor.
          */
          ~VectorProcessingOperationDialog();

        public:

          /*!
            \brief Returns the internal widget

            \return Pointer to Qt Widget InputDataAccessItemWidgetForm
          */
          Ui::VectorProcessingOperationDialogForm* getForm() const;

          /*!
            \brief Inits the dialog by setting the the operation to be executed in interface and the list of available input layers

            \param operation Pointer to a TerraLib VP operation
            \param layers The list of available input layers

            \note NOT take the ownership of this object
          */
          void init(const std::string& operationName, const std::list<te::map::AbstractLayerPtr>& layers);

          te::vp::OperationResult getResult() const;

          std::list<te::map::AbstractLayerPtr> getOutputLayers();

          void keyPressEvent(QKeyEvent* event);

        protected:

          void setDialogOnExecution(bool onExecution);

          te::vp::AbstractParameters* createParametersForOperation(te::vp::AbstractOperation* operation, QString& errorMessage);

        protected slots:

          void onProcessingStateChanged(const te::vp::ProgressState& newState);

          void onProcessingFinished(const te::vp::OperationResult& result);

          void taskCancelled();

          void onClose();

          /*!
            \brief Interface tool button to execute the vectorial process operation
          */
          void onExecuteOperationToolButtonPressed();

          void onDataAccessSelectionChanged();

        private:

          std::unique_ptr<Ui::VectorProcessingOperationDialogForm> m_ui;  //!< Pointer to Qt widget

          te::vp::qt::InputDataAccessWidget* m_inputDataWidget;           //!< TerraLib Qt component used to get input data access

          te::vp::qt::OutputDataAccessWidget* m_outputDataWidget;         //!< TerraLib Qt component used to get output data access

          te::vp::qt::SpecificParametersWidget* m_specificParamsWidget;   //!< TerraLib Qt component with specific parameters for each operation

          std::string m_operationName;                                    //!< The name of the operation to be created and executed

          te::qt::widgets::ProgressViewerTaskWidget* m_progressViewer;    //!< TerraLib Qt component to represent a progress viewer

          QtOperationThreadManagerListener m_processListener; //!< A listener to the status of the execution

          std::unique_ptr< CurrentExecutionImpl > m_run; //!< The implementation of one execution

          std::vector<te::da::DataAccessPtr> m_vecOverallOutputDataAccess; //!< Stores all the outputs considering all the executions
      };
    } // end namespace qt
  }  // end namespace qt
}  // end namespace qt


#endif // __TERRALIB_VP_QT_INTERNAL_VECTORPROCESSINGOPERATIONDIALOG_H

