/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/wfs/DataSource.cpp

  \brief Implementation of the data source for the WFS driver.
*/

// TerraLib
#include "../dataaccess/datasource/DataSourceTransactor.h"
#include "../dataaccess/query/SQLDialect.h"
#include "../common/StringUtils.h"
#include "../core/translator/Translator.h"
#include "../core/uri/Proxy.h"
#include "../core/uri/ProxyInfo.h"
#include "../ogr/Utils.h"
#include "DataSource.h"
#include "Exception.h"
#include "Globals.h"
#include "Transactor.h"
#include "Utils.h"

// OGR
#include <ogrsf_frmts.h>
#include <ogr_core.h>

te::da::DataSourceCapabilities te::wfs::DataSource::sm_capabilities;

te::wfs::DataSource::DataSource(const std::string& connInfo)
  : te::da::DataSource(connInfo),
    m_ogrDS(nullptr),
    m_isOpened(false)
{
}

te::wfs::DataSource::DataSource(const te::core::URI& uri)
  : te::da::DataSource(uri),
  m_ogrDS(nullptr),
  m_isOpened(false)
{
}

te::wfs::DataSource::~DataSource()
{
  close();
}

std::string te::wfs::DataSource::getType() const
{
  return TE_WFS_DRIVER_IDENTIFIER;
}

std::unique_ptr<te::da::DataSourceTransactor> te::wfs::DataSource::getTransactor()
{
  if(!m_isOpened)
    throw Exception(TE_TR("The data source is not opened!"));

  return std::unique_ptr<te::da::DataSourceTransactor>(new Transactor(this));
}

void te::wfs::DataSource::open()
{
  if(m_isOpened)
    return;

  verifyConnectionInfo();

  setupEnvironment();

  m_ogrDS = (GDALDataset*)GDALOpenEx(m_uri.uri().c_str(), GDAL_OF_READONLY, nullptr, nullptr, nullptr);

  if (m_ogrDS == nullptr)
  {
    const te::core::ProxyInfo& proxyInfo = te::core::Proxy::instance().getAuthenticatedProxyInfo();
    if (proxyInfo.isEnabled() == true)
    {
      te::core::Proxy::instance().resetAuthentication();
    }
    throw Exception(TE_TR("Could not open the WFS data source!"));
  }

  m_isOpened = true;
}

void te::wfs::DataSource::close()
{
  if (m_ogrDS)
  {
    GDALClose(m_ogrDS);
  }

  m_ogrDS = nullptr;

  m_isOpened = false;
}

bool te::wfs::DataSource::isOpened() const
{
  return m_isOpened;
}

bool te::wfs::DataSource::isValid() const
{
  if (m_isOpened)
    return true;

  verifyConnectionInfo();

  setupEnvironment();

  GDALDataset* gds = static_cast<GDALDataset*>(GDALOpenEx(m_uri.uri().c_str(), GDAL_OF_READONLY, nullptr, nullptr, nullptr));
  if (gds == nullptr)
  {
    const te::core::ProxyInfo& proxyInfo = te::core::Proxy::instance().getAuthenticatedProxyInfo();
    if (proxyInfo.isEnabled() == true)
    {
      te::core::Proxy::instance().resetAuthentication();
    }
    return false;
  }

  GDALClose(gds);

  return true;
}

const te::da::DataSourceCapabilities& te::wfs::DataSource::getCapabilities() const
{
  return sm_capabilities;
}

void te::wfs::DataSource::setCapabilities(const te::da::DataSourceCapabilities& capabilities)
{
  sm_capabilities = capabilities;
}

const te::da::SQLDialect* te::wfs::DataSource::getDialect() const
{
  return te::wfs::Globals::sm_queryDialect;;
}

GDALDataset* te::wfs::DataSource::getOGRDataSource()
{
  return m_ogrDS;
}

const std::vector<te::wfs::WFSLayerInfo>& te::wfs::DataSource::getLayersInfo()
{
  if(!m_layersInfo.empty())
    return m_layersInfo;

  buildLayersInfo();

  return m_layersInfo;
}

void te::wfs::DataSource::create(const std::string& /*connInfo*/)
{
  throw Exception(TE_TR("The create() method is not supported by the WFS driver!"));
}

void te::wfs::DataSource::drop(const std::string& /*connInfo*/)
{
  throw Exception(TE_TR("The drop() method is not supported by the WFS driver!"));
}

bool te::wfs::DataSource::exists(const std::string& connInfo)
{
  if (connInfo.empty())
    return false;

  const te::core::URI aux(connInfo);
  if (!aux.isValid())
    return false;

  std::string path = aux.path();
  if (path.empty())
    return false;

  setupEnvironment();

  GDALDataset* gds = static_cast<GDALDataset*>(GDALOpen(path.c_str(), GA_ReadOnly));
  if (gds == nullptr)
  {
    const te::core::ProxyInfo& proxyInfo = te::core::Proxy::instance().getAuthenticatedProxyInfo();
    if (proxyInfo.isEnabled() == true)
    {
      te::core::Proxy::instance().resetAuthentication();
    }
    return false;
  }

  GDALClose(gds);

  return true;
}

std::vector<std::string> te::wfs::DataSource::getDataSourceNames(const std::string& /*connInfo*/)
{
  return std::vector<std::string>();
}

void te::wfs::DataSource::verifyConnectionInfo() const
{
  if (!m_uri.isValid())
    throw Exception(TE_TR("The connection information is invalid!"));

  if (m_uri.path().empty())
    throw Exception(TE_TR("The connection information is invalid. Missing the path parameter!"));
}

void te::wfs::DataSource::setupEnvironment() const
{
  std::string authentication = "";
  std::string hostPort = "";
  std::string authenticationType = "";

  // if the proxy configuration is set, we configure curl
  const te::core::ProxyInfo& proxyInfo = te::core::Proxy::instance().getAuthenticatedProxyInfo();
  if(proxyInfo.isEnabled() == true)
  {
    authentication = proxyInfo.getAuthentication();
    hostPort = proxyInfo.getHostPort();
    authenticationType = "ANY";
  }

  CPLSetConfigOption("GDAL_HTTP_PROXY", hostPort.c_str());
  CPLSetConfigOption("GDAL_HTTP_PROXYUSERPWD", authentication.c_str());
  CPLSetConfigOption("GDAL_HTTP_UNSAFESSL", "YES");
  CPLSetConfigOption("TRUST_CAPABILITIES_BOUNDS", "YES");
  CPLSetConfigOption("OGR_WFS_TRUST_CAPABILITIES_BOUNDS", "YES");
  CPLSetConfigOption("OGR_WFS_PAGING_ALLOWED", "NO");
  CPLSetConfigOption("OGR_WFS_USE_STREAMING", "NO");
}

void te::wfs::DataSource::buildLayersInfo()
{
  open();

  OGRLayer* wfsMetadata = m_ogrDS->GetLayerByName("WFSLayerMetadata");

  if(wfsMetadata == nullptr)
    throw Exception(TE_TR("Could not retrieve the metadata from WFS server!"));

  OGRFeatureDefn* featDef = wfsMetadata->GetLayerDefn();

  OGRFeature* f;
  wfsMetadata->ResetReading();
  while((f = wfsMetadata->GetNextFeature()) != nullptr)
  {
    WFSLayerInfo info;
    info.m_name = f->GetFieldAsString("layer_name");
    info.m_title = f->GetFieldAsString("title");
    info.m_abstract = f->GetFieldAsString("abstract");

    m_layersInfo.push_back(info);

    OGRFeature::DestroyFeature(f);
  }
}
