/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/ws/ogc/wcs/qt/WCSConnectorDialog.cpp

  \brief A dialog window for showing the WCS connector widget.
*/

#include "WCSConnectorDialog.h"
#include "ui/ui_WCSConnectorDialogForm2.h"


// TerraLib
#include "../../../../core/translator/Translator.h"
#include "../../../../core/uri/URI.h"
#include "../../../../core/uri/Utils.h"
#include "../../../../dataaccess/datasource/DataSource.h"
#include "../../../../dataaccess/datasource/DataSourceFactory.h"
#include "../../../../dataaccess/datasource/DataSourceInfo.h"
#include "../../../../dataaccess/datasource/DataSourceManager.h"
#include "../dataaccess/Exception.h"
#include "../../../core/Exception.h"
#include "../../../../qt/af/ApplicationController.h"


// Boost
#include <boost/algorithm/string/case_conv.hpp>
#include <boost/uuid/random_generator.hpp>
#include <boost/uuid/uuid_io.hpp>
#include <boost/lexical_cast.hpp>

//Qt
#include <QMessageBox>
#include <QFileDialog>

te::ws::ogc::wcs::qt::WCSConnectorDialog::WCSConnectorDialog(QWidget *parent, Qt::WindowFlags f)
  : QDialog(parent, f),
    m_ui(new Ui::WCSConnectorDialogForm2)
{
  // add controls
    m_ui->setupUi(this);

    // set icons
    m_ui->m_wcsDirToolButton->setIcon(QIcon::fromTheme("folder"));

  // connect signal and slots
    connect(m_ui->m_authCheckBox, SIGNAL(stateChanged(int)), this, SLOT(authCheckBoxStateChanged(int)));
    connect(m_ui->m_openPushButton, SIGNAL(pressed()), this, SLOT(openPushButtonPressed()));
    connect(m_ui->m_testPushButton, SIGNAL(pressed()), this, SLOT(testPushButtonPressed()));
    connect(m_ui->m_helpPushButton, SIGNAL(pressed()), this, SLOT(helpPushButtonPressed()));
    connect(m_ui->m_wcsDirToolButton, SIGNAL(clicked()), this, SLOT(wcsDirToolButtonClicked()));

    m_ui->m_authCheckBox->setCheckState(Qt::Unchecked);
    m_ui->m_authUsernameLineEdit->setEnabled(false);
    m_ui->m_authPasswordLineEdit->setEnabled(false);
}

te::ws::ogc::wcs::qt::WCSConnectorDialog::~WCSConnectorDialog() = default;

const te::da::DataSourceInfoPtr& te::ws::ogc::wcs::qt::WCSConnectorDialog::getDataSource() const
{
  return m_datasource;
}

const te::da::DataSourcePtr& te::ws::ogc::wcs::qt::WCSConnectorDialog::getDriver() const
{
  return m_driver;
}

void te::ws::ogc::wcs::qt::WCSConnectorDialog::set(const te::da::DataSourceInfoPtr &ds)
{
  m_datasource = ds;

  if(m_datasource.get() != nullptr)
  {
    const te::core::URI& connInfo = m_datasource->getConnInfo();

    std::map<std::string, std::string> kvp = te::core::Expand(connInfo.query());

    std::string serviceURL = kvp["URI"];

    te::core::URI serviceURI (serviceURL);

    if(!serviceURI.user().empty() && !serviceURI.password().empty())
    {
      std::string baseURL = serviceURI.scheme() + std::string("://") + serviceURI.host();

      if(!serviceURI.port().empty())
      {
        baseURL = baseURL + std::string(":") + serviceURI.port();
      }

      baseURL = baseURL + serviceURI.path();

      if(!serviceURI.query().empty())
      {
        baseURL = baseURL + std::string("?") + serviceURI.query();
      }

      if(!serviceURI.fragment().empty())
      {
        baseURL = baseURL + std::string("#") + serviceURI.fragment();
      }

      m_ui->m_serverLineEdit->setText(QString::fromUtf8(baseURL.c_str()));


      m_ui->m_authCheckBox->setCheckState(Qt::Checked);
      m_ui->m_authUsernameLineEdit->setEnabled(true);
      m_ui->m_authPasswordLineEdit->setEnabled(true);

      m_ui->m_authUsernameLineEdit->setText(QString::fromUtf8(serviceURI.user().c_str()));
      m_ui->m_authPasswordLineEdit->setText(QString::fromUtf8(serviceURI.password().c_str()));
    }
    else
    {
      m_ui->m_serverLineEdit->setText(QString::fromUtf8(serviceURL.c_str()));
    }

    m_ui->m_datasourceTitleLineEdit->setText(QString::fromUtf8(m_datasource->getTitle().c_str()));

    m_ui->m_datasourceDescriptionTextEdit->setText(QString::fromUtf8(m_datasource->getDescription().c_str()));

    std::string usrDataDir = kvp["USERDATADIR"];

    m_ui->m_wcsDirLineEdit->setText(QString::fromUtf8(usrDataDir.c_str()));
  }
}

void te::ws::ogc::wcs::qt::WCSConnectorDialog::openPushButtonPressed()
{
  try
  {
    // Check if driver is loaded
    if(te::da::DataSourceFactory::find("WCS2") == 0)
      throw te::ws::ogc::wcs::da::Exception(TE_TR("Sorry! No data access driver loaded for WCS data sources!"));

    // Get the data source connection info based on form data
    std::string dsInfo = getConnectionInfo();

    // Perform connection
    std::unique_ptr<te::da::DataSource> ds = te::da::DataSourceFactory::make("WCS2", dsInfo);
    ds->open();
    m_driver.reset(ds.release());

    if(m_driver.get() == nullptr)
      throw te::ws::ogc::wcs::da::Exception(TE_TR("Could not open WCS data source due to an unknown error!"));

    QString title = m_ui->m_datasourceTitleLineEdit->text().trimmed();

    if(title.isEmpty())
      title = m_ui->m_serverLineEdit->text().trimmed();

    if(m_datasource.get() == nullptr)
    {
      // Create a new data source based on the form data
      m_datasource.reset(new te::da::DataSourceInfo);

      m_datasource->setConnInfo(dsInfo);

      boost::uuids::basic_random_generator<boost::mt19937> gen;
      boost::uuids::uuid u = gen();
      std::string dsId = boost::uuids::to_string(u);

      m_datasource->setId(dsId);
      m_driver->setId(dsId);
      m_datasource->setTitle(title.toUtf8().data());
      m_datasource->setDescription(m_ui->m_datasourceDescriptionTextEdit->toPlainText().trimmed().toUtf8().data());
      m_datasource->setAccessDriver("WCS2");
      m_datasource->setType("WCS2");
    }
    else
    {
      m_driver->setId(m_datasource->getId());
      m_datasource->setConnInfo(dsInfo);
      m_datasource->setTitle(title.toUtf8().data());
      m_datasource->setDescription(m_ui->m_datasourceDescriptionTextEdit->toPlainText().trimmed().toUtf8().data());
    }
  }
  catch(te::ws::core::Exception& e)
  {
    if(const std::string* d =
           boost::get_error_info<te::ErrorDescription>(e))
    {
      QMessageBox::warning(this,
                         tr("TerraLib Qt Components"),
                         tr(d->c_str()));
    }
    return;
  }
  catch(...)
  {
    QMessageBox::warning(this,
                         tr("TerraLib Qt Components"),
                         tr("Unknown error while opening WCS data source!"));
    return;
  }

  accept();
}

void te::ws::ogc::wcs::qt::WCSConnectorDialog::testPushButtonPressed()
{
  try
  {
    // Check if driver is loaded
    if(te::da::DataSourceFactory::find("WCS2") == 0)
      throw te::ws::ogc::wcs::da::Exception(TE_TR("Sorry! No data access driver loaded for WCS data sources!"));

    // Perform connection
    std::unique_ptr<te::da::DataSource> ds(te::da::DataSourceFactory::make("WCS2", getConnectionInfo()));

    if(ds.get() == nullptr)
      throw te::ws::core::Exception() << te::ErrorDescription(TE_TR("Could not open WCS server!"));

    ds->open();

    QMessageBox::information(this,
                             tr("TerraLib Qt Components"),
                             tr("Data source is ok!"));

    ds->close();
  }
  catch(te::ws::core::Exception& e)
  {
    if(const std::string* d =
           boost::get_error_info<te::ErrorDescription>(e))
    {
      QMessageBox::warning(this,
                         tr("TerraLib Qt Components"),
                         tr(d->c_str()));
    }
  }
  catch(...)
  {
    QMessageBox::warning(this,
                         tr("TerraLib Qt Components"),
                         tr("Unknown error while testing WCS data source!"));
  }
}

void te::ws::ogc::wcs::qt::WCSConnectorDialog::helpPushButtonPressed()
{
  QMessageBox::warning(this,
                       tr("TerraLib Qt Components"),
                       tr("Not implemented yet!\nWe will provide it soon!"));
}

void te::ws::ogc::wcs::qt::WCSConnectorDialog::authCheckBoxStateChanged(const int &state)
{
  Qt::CheckState checkState = (Qt::CheckState) state;

  if(checkState == Qt::Checked)
  {
    m_ui->m_authUsernameLineEdit->setEnabled(true);
    m_ui->m_authPasswordLineEdit->setEnabled(true);
  }
  else
  {
    m_ui->m_authUsernameLineEdit->setEnabled(false);
    m_ui->m_authPasswordLineEdit->setEnabled(false);
    m_ui->m_authUsernameLineEdit->clear();
    m_ui->m_authPasswordLineEdit->clear();
  }
}

void te::ws::ogc::wcs::qt::WCSConnectorDialog::wcsDirToolButtonClicked()
{
  QString folderName = QFileDialog::getExistingDirectory(this, tr("Select a directory"), "");

  if (folderName.isEmpty())
    return;

  m_ui->m_wcsDirLineEdit->setText(folderName);
}

const std::string te::ws::ogc::wcs::qt::WCSConnectorDialog::getConnectionInfo() const
{
  std::string strURI("wcs://");

  // Get the server URL
  std::string serviceURL = m_ui->m_serverLineEdit->text().trimmed().toUtf8().constData();

  if (serviceURL.empty())
    throw te::ws::core::Exception() << te::ErrorDescription(TE_TR("Please define the server address first!"));

  std::string usrDataDir = m_ui->m_wcsDirLineEdit->text().trimmed().toUtf8().constData();

  if (usrDataDir.empty())
    throw te::ws::core::Exception() << te::ErrorDescription(TE_TR("Please define output directory to WCS files."));

  if(m_ui->m_authCheckBox->checkState() == Qt::Checked)
  {

    if(m_ui->m_authUsernameLineEdit->text().isEmpty() || m_ui->m_authPasswordLineEdit->text().isEmpty())
    {
      throw te::ws::core::Exception() << te::ErrorDescription(TE_TR("Username or password is not defined!"));
    }

    std::string username = m_ui->m_authUsernameLineEdit->text().toUtf8().constData();
    std::string password = m_ui->m_authPasswordLineEdit->text().toUtf8().constData();
    std::string userAndPassword = username + std::string(":") + password;

    te::core::URI aux(serviceURL);

    serviceURL = aux.scheme() + std::string("://") + userAndPassword + std::string("@")
        + aux.host();

    if(!aux.port().empty())
    {
      serviceURL = serviceURL + std::string(":") + aux.port();
    }

    serviceURL = serviceURL + aux.path();

    if(!aux.query().empty())
    {
      serviceURL = serviceURL + std::string("?") + aux.query();
    }

    if(!aux.fragment().empty())
    {
      serviceURL = serviceURL + std::string("#") + aux.fragment();
    }
  }

  std::string encodedURL = te::core::URIEncode(serviceURL);

  strURI = strURI + "?URI="+ encodedURL + "&VERSION=2.0.1" + "&USERDATADIR=" + usrDataDir;

  return strURI;
}
