#include "Utils.h"
#include "../../../../common/StringUtils.h"
#include "../../../../common/Enums.h"
#include "../../../core/Exception.h"
#include "../../../../core/translator/Translator.h"
#include "../../../../dataaccess/datasource/DataSource.h"
#include "../../../../dataaccess/datasource/DataSourceFactory.h"
#include "../../../../dataaccess/datasource/DataSourceTransactor.h"
#include "../../../../dataaccess/dataset/DataSet.h"
#include "../../../../raster/BandProperty.h"
#include "../../../../raster/Grid.h"
#include "../../../../raster/Raster.h"
#include "../../../../raster/RasterFactory.h"
#include "../../../../geometry/Envelope.h"
#include "../../../../gdal/Utils.h"

// GDAL
#include <gdal_priv.h>
#include <gdalwarper.h>
#include <ogr_api.h>
#include <ogr_spatialref.h>
#include <ogrsf_frmts.h>

#include <string>

te::ws::ogc::wms::WMSGetMapResponse te::ws::ogc::wms::GetMap( WMSLayer* wmsLayer, const te::gm::Envelope &box,
                                                             const int &width, const int &height)
{
  //Preparing request acording to size and bounding box.
  te::ws::ogc::wms::WMSGetMapRequest request = wmsLayer->getRequest();

  request.m_width = width;
  request.m_height = height;

  te::ws::ogc::wms::BoundingBox boundingBox;
  boundingBox.m_crs = request.m_boundingBox.m_crs;
  boundingBox.m_minX = box.getLowerLeftX();
  boundingBox.m_minY = box.getLowerLeftY();
  boundingBox.m_maxX = box.getUpperRightX();
  boundingBox.m_maxY = box.getUpperRightY();

  request.m_boundingBox = boundingBox;

  wmsLayer->setGetMapRequest(request);

  te::ws::ogc::wms::WMSGetMapResponse response = wmsLayer->getMap();

  return response;
}

te::rst::Raster* te::ws::ogc::wms::GetMapRaster(te::ws::ogc::wms::WMSLayer *wmsLayer, const int& srid, const te::gm::Envelope &box, const int &width, const int &height)
{

  //Preparing request acording to size and bounding box.
  te::ws::ogc::wms::WMSGetMapRequest request = wmsLayer->getRequest();

  request.m_width = width;
  request.m_height = height;

  std::string epsgCode = "EPSG:" + std::to_string(srid);

  te::ws::ogc::wms::BoundingBox boundingBox;
  boundingBox.m_crs = epsgCode;
  boundingBox.m_minX = box.getLowerLeftX();
  boundingBox.m_minY = box.getLowerLeftY();
  boundingBox.m_maxX = box.getUpperRightX();
  boundingBox.m_maxY = box.getUpperRightY();

  request.m_boundingBox = boundingBox;

  request.m_srs = epsgCode;

  wmsLayer->setGetMapRequest(request);

  std::string format = request.m_format;

  std::vector<std::string> formatSplit = te::common::SplitString(format, '/');

  if(formatSplit.size() != 2)
  {
    throw te::ws::core::Exception() << te::ErrorDescription(TE_TR("Image format not supported."));
  }

  format = formatSplit[1];


  std::string filename = wmsLayer->getId() + "." + format;

  std::string filePath = wmsLayer->saveGetMap(filename);

  std::map<std::string, std::string> memRInfo;
  memRInfo["URI"] = filePath;

  te::rst::Raster* raster = te::rst::RasterFactory::open(std::string("GDAL"), memRInfo, te::common::RAccess);

  if (!raster)
  {
    throw te::ws::core::Exception() << te::ErrorDescription(TE_TR("Failed on create a Raster from WMS Server."));
  }

  te::gm::Envelope* mbr = new te::gm::Envelope(boundingBox.m_minX, boundingBox.m_minY, boundingBox.m_maxX, boundingBox.m_maxY);

  te::rst::Grid* grid = new te::rst::Grid((unsigned int)raster->getNumberOfColumns(), (unsigned int)raster->getNumberOfRows(),
                                          mbr, srid);

  double geoTrans[6];
  geoTrans[0] = grid->getGeoreference()[0];
  geoTrans[1] = grid->getGeoreference()[1];
  geoTrans[2] = grid->getGeoreference()[2];
  geoTrans[3] = grid->getGeoreference()[3];
  geoTrans[4] = grid->getGeoreference()[4];
  geoTrans[5] = grid->getGeoreference()[5];

  raster->getGrid()->setGeoreference(geoTrans, srid);

  delete grid;

  return raster;
}

te::rst::Raster* te::ws::ogc::wms::GetLayerMapRaster(const WMSLayer* wmsLayer)
{
  //Preparing request acording to size and bounding box.
  te::ws::ogc::wms::WMSGetMapRequest request = wmsLayer->getRequest();

  std::string format = request.m_format;

  std::vector<std::string> formatSplit = te::common::SplitString(format, '/');

  if (formatSplit.size() != 2)
  {
    throw te::ws::core::Exception() << te::ErrorDescription(TE_TR("Image format not supported."));
  }

  format = formatSplit[1];


  std::string filename = wmsLayer->getId() + "." + format;

  std::string filePath = wmsLayer->saveGetMap(filename);

  std::map<std::string, std::string> memRInfo;
  memRInfo["URI"] = filePath;

  te::rst::Raster* raster = te::rst::RasterFactory::open(std::string("GDAL"), memRInfo, te::common::RAccess);

  if (!raster)
  {
    throw te::ws::core::Exception() << te::ErrorDescription(TE_TR("Failed on create a Raster from WMS Server."));
  }

  te::ws::ogc::wms::BoundingBox boundingBox = request.m_boundingBox;

  te::gm::Envelope* mbr = new te::gm::Envelope(boundingBox.m_minX, boundingBox.m_minY, boundingBox.m_maxX, boundingBox.m_maxY);

  te::rst::Grid* grid = new te::rst::Grid((unsigned int)raster->getNumberOfColumns(), (unsigned int)raster->getNumberOfRows(), mbr, wmsLayer->getSRID());

  double geoTrans[6];
  geoTrans[0] = grid->getGeoreference()[0];
  geoTrans[1] = grid->getGeoreference()[1];
  geoTrans[2] = grid->getGeoreference()[2];
  geoTrans[3] = grid->getGeoreference()[3];
  geoTrans[4] = grid->getGeoreference()[4];
  geoTrans[5] = grid->getGeoreference()[5];

  raster->getGrid()->setGeoreference(geoTrans, wmsLayer->getSRID());

  delete grid;

  return raster;
}
