#include "WMSLayerRenderer.h"

#include "WMSLayer.h"
#include "Utils.h"

#include "../../../../core/translator/Translator.h"
#include "../../../../maptools/AbstractLayer.h"
#include "../../../../maptools/Canvas.h"
#include "../../../../maptools/Utils.h"
#include "../../../../dataaccess/utils/Utils.h"
#include "../../../../geometry/Envelope.h"
#include "../../../../se/Utils.h"
#include "../../../../se/CoverageStyle.h"
#include "../../../../raster/RasterProperty.h"
#include "../../../core/Exception.h"
#include "../../../core/Utils.h"
#include "../../../../srs/Config.h"
#include "../../../../srs/SpatialReferenceSystemManager.h"

// Boost
#include <boost/format.hpp>

te::ws::ogc::wms::WMSLayerRenderer::WMSLayerRenderer() = default;

te::ws::ogc::wms::WMSLayerRenderer::~WMSLayerRenderer() = default;

void te::ws::ogc::wms::WMSLayerRenderer::draw(te::map::AbstractLayer *layer, te::map::Canvas *canvas, const te::gm::Envelope &bbox, int srid, const double &scale, bool *cancel)
{

  te::ws::ogc::wms::WMSLayer* wmsLayer = dynamic_cast<te::ws::ogc::wms::WMSLayer*>(layer);

  if(wmsLayer == nullptr)
  {
    throw te::ws::core::Exception() << te::ErrorDescription(TE_TR("Wrong type render type for this layer!"));
  }

  te::gm::Envelope reprojectedBBOX(bbox);

  if(srid != TE_UNKNOWN_SRS)
  {
    reprojectedBBOX.transform(srid, wmsLayer->getSRID());
  }
  else if(layer->getSRID() != srid)
  {
    throw te::ws::core::Exception() << te::ErrorDescription(TE_TR("The layer or map has no valid SRID!"));
  }

  if(!reprojectedBBOX.intersects(wmsLayer->getExtent()))
    return;

  //update request

  te::ws::ogc::wms::WMSGetMapRequest request = wmsLayer->getRequest();

  request.m_width = canvas->getWidth();
  request.m_height = canvas->getHeight();

  std::string epsgCode = "EPSG:" + std::to_string(wmsLayer->getSRID());

  te::ws::ogc::wms::BoundingBox boundingBox;
  boundingBox.m_crs = epsgCode;
  boundingBox.m_minX = reprojectedBBOX.getLowerLeftX();
  boundingBox.m_minY = reprojectedBBOX.getLowerLeftY();
  boundingBox.m_maxX = reprojectedBBOX.getUpperRightX();
  boundingBox.m_maxY = reprojectedBBOX.getUpperRightY();

  request.m_boundingBox = boundingBox;

  request.m_srs = epsgCode;

  wmsLayer->setGetMapRequest(request);

  std::unique_ptr<te::rst::Raster> memRaster(te::map::GetRaster(wmsLayer));

  std::unique_ptr<te::da::DataSetType> dsType = wmsLayer->getSchema();

  te::rst::RasterProperty* rasterProperty = te::da::GetFirstRasterProperty(dsType.get());

  // get the associated layer style
  te::se::Style* style = wmsLayer->getStyle();
  if(style == nullptr)
  {
    // try create an appropriate style
    style = te::se::CreateCoverageStyle(rasterProperty->getBandProperties());

    if(style == nullptr)
      throw te::ws::core::Exception() << te::ErrorDescription((boost::format(TE_TR("Could not create a default coverage style to the layer %1%.")) % wmsLayer->getTitle()).str());

    wmsLayer->setStyle(style);
  }

  // should I render this style?
  te::se::CoverageStyle* cs = dynamic_cast<te::se::CoverageStyle*>(style);
  if(cs == nullptr)
    throw te::ws::core::Exception() << te::ErrorDescription(TE_TR("The layer style is not a Coverage Style!"));

  te::map::DrawRaster(memRaster.get(), canvas, reprojectedBBOX, wmsLayer->getSRID(), bbox, srid, cs, layer->getRasterContrast(), scale, cancel);
}
