/*  Copyright (C) 2017 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/ws/ogc/wms/qt/WMSGetFeatureInfoMaker.cpp

  \brief WMSGetFeatureInfoMaker is a WMS implementation of AbstractFeatureInfoMaker
         that could be used to fill a QTreeWidgetItem with GetFeatureInfo WMS operation based information.
*/

#include "WMSGetFeatureInfoMaker.h"

//TerraLib
#include "../../../../qt/widgets/canvas/MapDisplay.h"
#include "../../../../maptools/WorldDeviceTransformer.h"
#include "../dataaccess/WMSLayer.h"
#include "../client/DataTypes.h"

//QT
#include <QTreeWidget>

te::ws::ogc::wms::qt::WMSGetFeatureInfoMaker::WMSGetFeatureInfoMaker() =
    default;

te::ws::ogc::wms::qt::WMSGetFeatureInfoMaker::~WMSGetFeatureInfoMaker() =
    default;

void te::ws::ogc::wms::qt::WMSGetFeatureInfoMaker::getInfo(QTreeWidgetItem *layerItem, te::qt::widgets::MapDisplay* display, const te::map::AbstractLayerPtr &layer, const te::gm::Envelope &e)
{
  if(layer->getVisibility() != te::map::VISIBLE)
    return;

  te::ws::ogc::wms::WMSLayer* wmsLayer = dynamic_cast<te::ws::ogc::wms::WMSLayer*>(layer.get());

  if(wmsLayer == nullptr)
    return;

  te::gm::Envelope reprojectedEnvelope(e);

  if((layer->getSRID() != TE_UNKNOWN_SRS) &&
     (display->getSRID() != TE_UNKNOWN_SRS) &&
     (layer->getSRID() != display->getSRID()))
  {
    reprojectedEnvelope.transform(display->getSRID(), layer->getSRID());
  }

  if(!reprojectedEnvelope.intersects(layer->getExtent()))
    return;

  te::gm::Envelope displayExtent = display->getExtent();

  te::map::WorldDeviceTransformer transformer (displayExtent.getLowerLeftX(), displayExtent.getLowerLeftY(),
                                               displayExtent.getUpperRightX(), displayExtent.getUpperRightY(),
                                               display->width(), display->height());

  double x1, y1, x2, y2;

  transformer.world2Device(e.getLowerLeftX(), e.getLowerLeftY(), x1, y1);
  transformer.world2Device(e.getUpperRightX(), e.getUpperRightY(), x2, y2);

  te::gm::Envelope auxEnv (x1, y1, x2, y2);

  te::gm::Coord2D center = auxEnv.getCenter();

  te::ws::ogc::wms::WMSGetMapRequest getMap = wmsLayer->getRequest();

  te::ws::ogc::wms::WMSGetFeatureInfoRequest infoRequest;
  infoRequest.m_queryLayers.push_back(wmsLayer->getDataSetName());
  infoRequest.m_getMap = getMap;
  infoRequest.m_i = static_cast<int>(center.getX());
  infoRequest.m_j = static_cast<int>(center.getY());

  te::ws::ogc::wms::WMSGetFeatureInfoResponse response = wmsLayer->getFeatureInfo(infoRequest);

  QTreeWidgetItem* getFeatureInfoItem = new QTreeWidgetItem(layerItem);
  getFeatureInfoItem->setText(0, QString("GetFeatureInfo"));

  getFeatureInfoItem->setText(1, response.m_featureInfoContent.c_str());

  QTreeWidget* treeWidget = layerItem->treeWidget();

  if(treeWidget != nullptr)
  {
    treeWidget->expandAll();
    treeWidget->resizeColumnToContents(1);
  }

}
