/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of TerraView - A Free and Open Source GIS Application.

    TerraView is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraView is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with TerraView. See COPYING. If not, write to
    TerraLib Team at <terralib-team@dpi.inpe.br>.
 */

/*!
  \file terraview/main.cpp

  \brief It contains the main routine of TerraView.
*/


// TerraView
#include "Config.h"
#include "../terralib/Defines.h"
#include "TerraView.h"
#include "TerraViewApp.h"

// TerraLib
#include <terralib/Exception.h>
#include <terralib/common/Version.h>
#include <terralib/core/filesystem/FileSystem.h>
#include <terralib/core/utils/Platform.h>
#include <terralib/qt/af/Utils.h>
#include <terralib/qt/af/SplashScreenManager.h>
#include <terralib/qt/widgets/Utils.h>
#include <terralib/Version.h>

// STL
#include <cstdlib>
#include <exception>
#include <locale>

// Qt
#include <QApplication>
#include <QDir>
#include <QFileInfo>
#include <QLibraryInfo>
#include <QLocale>
#include <QMessageBox>
#include <QSplashScreen>
#include <QTextCodec>
#include <QTranslator>
#include <QPainter>


#if TE_PLATFORM == TE_PLATFORMCODE_APPLE
#include <CoreFoundation/CoreFoundation.h>
#endif


int main(int argc, char** argv)
{
  TerraViewApp app(argc, argv);
  app.setApplicationVersion(QString::fromUtf8(te::common::Version::asString().c_str()));

  QDir dir(QLibraryInfo::location(QLibraryInfo::TranslationsPath));

  QStringList filters;
  filters <<"*" + QLocale::system().name().toLower() + ".qm";

  QFileInfoList lst = dir.entryInfoList(filters, QDir::Files);

  //search directly in the share/terralib/translations folder as a plan B
  if (dir.exists() == false || lst.isEmpty() == true)
  {
    std::string translationDir = te::core::FindInTerraLibPath("/share/terralib/translations");
    translationDir = te::core::FileSystem::adjustPathToNativeForm(translationDir);
    dir = QDir(te::qt::widgets::Convert2QString(translationDir));

    lst = dir.entryInfoList(filters, QDir::Files);
  }

  //search directly in the translation folder for the cases when running from the build folder
  if (dir.exists() == false || lst.isEmpty() == true)
  {
    std::string translationDir = te::core::FindInTerraLibPath("/translations");
    translationDir = te::core::FileSystem::adjustPathToNativeForm(translationDir);
    dir = QDir(te::qt::widgets::Convert2QString(translationDir));

    lst = dir.entryInfoList(filters, QDir::Files);
  }

  for(int i=0; i<lst.size(); ++i)
  {
    QTranslator* trans = new QTranslator;
    trans->load(lst.at(i).baseName(), dir.path());
    app.installTranslator(trans);
  }

  setlocale(LC_ALL,"C"); // This force to use "." as decimal separator.

#if QT_VERSION >= 0x050000
  QTextCodec::setCodecForLocale(QTextCodec::codecForLocale());
#else
  QTextCodec::setCodecForCStrings(QTextCodec::codecForLocale());
#endif

  int waitVal = EXIT_FAILURE;

  const int RESTART_CODE = 1000;
  try
  {
    QString terraviewVersionStr( "TerraView : " TERRALIB_VERSION_STRING );
    QString terralibVersionStr( "TerraLib : " TERRALIB_VERSION_STRING );
    QString buildDateStr( ( "Build : " + te::common::Version::buildDate() ).c_str() );

    std::string splash_pix = te::core::FindInTerraLibPath(TVIEW_SPLASH_SCREEN_PIXMAP);
    QPixmap pixmap(splash_pix.c_str());

    QPainter painter( &pixmap );
    painter.setFont( QFont("Arial", 10 ) );
    painter.drawText( QPoint(8, 275), terraviewVersionStr );
    painter.drawText( QPoint(8, 290), terralibVersionStr );
    painter.drawText( QPoint(8, 305), buildDateStr );
    painter.drawText( QPoint(8, 320), "Copyright 2010 INPE" );

    do
    {
      QSplashScreen* splash = new QSplashScreen(pixmap/*, Qt::WindowStaysOnTopHint*/);

//      splash.setAttribute(Qt::WA_DeleteOnClose, true);

      splash->setStyleSheet("QWidget { font-size: 12px; font-weight: bold }");

      te::qt::af::SplashScreenManager::getInstance().set(splash, Qt::AlignBottom | Qt::AlignHCenter, Qt::white);

      splash->show();

      TerraView tview;

      //tview.resetTerraLib(waitVal != RESTART_CODE);

#if TE_PLATFORM == TE_PLATFORMCODE_APPLE
      CFBundleRef mainBundle = CFBundleGetMainBundle();
      CFURLRef execPath = CFBundleCopyBundleURL(mainBundle);

      char path[PATH_MAX];

      if (!CFURLGetFileSystemRepresentation(execPath, TRUE, (UInt8 *)path, PATH_MAX))
        throw; // error!

      CFRelease(execPath);

      QDir dPath(path);

      dPath.cd("Contents");

      chdir(dPath.path().toUtf8().data());
#endif

      tview.init(te::core::FindInTerraLibPath(TERRAVIEW_APPLICATION_CONFIG_FILE).c_str());

      splash->setParent(&tview);

      splash->finish(&tview);

      tview.showMaximized();

      //tview.resetState();

      // Start TerraView from project file(.tview)
      if(argc > 1)
      {
        std::string projPath = argv[1];

        if(!projPath.empty())
          tview.startProject(projPath.c_str());
      }

      waitVal = app.exec();

    } while(waitVal == RESTART_CODE);
  }
  catch(const boost::exception& e)
  {
    if(const std::string* d =
           boost::get_error_info<te::ErrorDescription>(e))
      QMessageBox::warning(nullptr, "TerraView", d->c_str());
    else
      QMessageBox::warning(nullptr,"TerraView", "An unknown error has occurred");
    return EXIT_FAILURE;
  }
  catch(const std::exception& e)
  {
    QMessageBox::warning(nullptr, "TerraView", e.what());
    return EXIT_FAILURE;
  }
  catch(...)
  {
    return EXIT_FAILURE;
  }

  return waitVal;
}
