/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file TsISOData.cpp
 
  \brief A test suit for the SynchronizedRaster Class.
 */

// TerraLib
#include <terralib_buildconfig.h>

#include <terralib/classification.h>
#include <terralib/raster.h>
#include "Config.h"

// STL
#include <memory>
#include <vector>

// Boost
#include <boost/test/unit_test.hpp>

    /*!
      \class RasterInputAdaptor
      \brief Classifiers raster input data adaptor.

    */
    template<typename DataType>
    class RasterInputAdaptor : public te::cl::InputAdaptor< DataType >
    {
      public:

        RasterInputAdaptor( const te::rst::Raster& raster )
        : m_raster( raster ), m_nRows( raster.getNumberOfRows() ),
          m_nCols( raster.getNumberOfColumns() )
        {};

        ~RasterInputAdaptor() {};

        // overload
        void getFeature(const unsigned int& elementIndex, const unsigned int& featureIndex,
          DataType& featureValue ) const
        {
          assert( elementIndex < ( m_nCols * m_nRows ) );
          assert( featureIndex < ( m_raster.getNumberOfBands() ) );
          m_raster.getValue( elementIndex % m_nCols, elementIndex / m_nCols, featureValue, featureIndex );
        };

        // overload
        unsigned int getElementsCount() const
        {
          return ( m_nRows * m_nCols );
        };

        // overload
        unsigned int getFeaturesCount() const
        {
          return m_raster.getNumberOfBands();
        };

      protected :

        const te::rst::Raster& m_raster;
        unsigned int m_nRows;
        unsigned int m_nCols;
    };

    /*!
      \class RasterInputOutputAdaptor
      \brief Classifiers raster input/output data adaptor.

    */
    template<typename DataType>
    class RasterInputOutputAdaptor : public te::cl::InputOutputAdaptor< DataType >
    {
      public:

        RasterInputOutputAdaptor( te::rst::Raster& raster )
        : m_raster( raster ), m_nRows( raster.getNumberOfRows() ),
          m_nCols( raster.getNumberOfColumns() )
        {};

        ~RasterInputOutputAdaptor() {};

        // overload
        void getFeature(const unsigned int& elementIndex, const unsigned int& featureIndex,
          DataType& featureValue ) const
        {
          assert( elementIndex < ( m_nCols * m_nRows ) );
          assert( featureIndex < ( m_raster.getNumberOfBands() ) );
          m_raster.getValue( elementIndex % m_nCols,
            elementIndex / m_nCols, m_value, featureIndex );
          featureValue = (DataType)m_value;
        };

        // overload
        void setFeature(const unsigned int& elementIndex,
          const unsigned int& featureIndex, const DataType& value )
        {
          assert( elementIndex < ( m_nCols * m_nRows ) );
          assert( featureIndex < ( m_raster.getNumberOfBands() ) );
          m_raster.setValue( elementIndex % m_nCols, elementIndex / m_nCols, (double)value,
            featureIndex );
        };

        // overload
        unsigned int getElementsCount() const
        {
          return ( m_nRows * m_nCols );
        };

        // overload
        unsigned int getFeaturesCount() const
        {
          return m_raster.getNumberOfBands();
        };

      protected :

        te::rst::Raster& m_raster;
        unsigned int m_nRows;
        unsigned int m_nCols;
        mutable double m_value;
    };

BOOST_AUTO_TEST_SUITE ( classification_isodata_tests )

BOOST_AUTO_TEST_CASE (isodata_test_1)
{
  /* Openning input raster */

  std::unique_ptr< te::rst::Raster > inputRasterPtrPointer;
  {
    std::map<std::string, std::string> auxRasterInfo;
    auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers_rgb342_crop1.tif";

    inputRasterPtrPointer.reset( te::rst::RasterFactory::open(
      auxRasterInfo ) );
    BOOST_REQUIRE( inputRasterPtrPointer.get() );
  }

  // creating the Memory output raster

  std::unique_ptr< te::rst::Raster > outputMEMRasterPtr;
  {
    std::vector< te::rst::BandProperty* > bProps;
    bProps.push_back( new te::rst::BandProperty( *inputRasterPtrPointer->getBand( 0 )->getProperty() ) );

    std::map<std::string, std::string> auxRasterInfo;

    outputMEMRasterPtr.reset( te::rst::RasterFactory::make(
      "EXPANSIBLE",
      new te::rst::Grid( *inputRasterPtrPointer->getGrid() ),
      bProps,
      auxRasterInfo,
      0,
      0 ) );
    BOOST_REQUIRE( outputMEMRasterPtr.get() );
  }

  // classifying

  std::vector<unsigned int> samplesAttributesIndices;
  samplesAttributesIndices.push_back( 0 );
  samplesAttributesIndices.push_back( 1 );
  samplesAttributesIndices.push_back( 2 );

  te::cl::ISOData classifierInstance;

  te::cl::ISOData::Parameters params;
  params.m_clustersNumber = 6;

  BOOST_REQUIRE( classifierInstance.initialize( params ) );

  RasterInputAdaptor< double > inputRasterAdp( *inputRasterPtrPointer );
  RasterInputOutputAdaptor< unsigned int > outputRasterAdp( *outputMEMRasterPtr );

  std::vector< double > inputNoDataValues(
    inputRasterPtrPointer->getNumberOfBands(),
    257 );

  std::vector< te::cl::ISOData::ClusterData > clustersData;
  BOOST_REQUIRE( classifierInstance.classify(
    inputRasterAdp,
    samplesAttributesIndices,
    inputNoDataValues,
    outputRasterAdp,
    0,
    255,
    true,
    &clustersData ) );
  
  // print Metadata
  
  std::vector< te::cl::ISOData::ClusterData >::const_iterator clustersDataIt =
    clustersData.begin();
  const std::vector< te::cl::ISOData::ClusterData >::const_iterator clustersDataItEnd =
    clustersData.end();
  while( clustersDataIt != clustersDataItEnd )
  {
    std::cout << std::endl << "IsoData Clusters Data: name=[" << 
      clustersDataIt->m_label << "] size=[" << 
      clustersDataIt->m_size << "]" << std::endl;
    ++clustersDataIt;
  }  

  BOOST_REQUIRE( ( clustersData.size() >= 3 ) && ( clustersData.size() <= 9 ) )  ;

  // Creating the disk output raster

  std::unique_ptr< te::rst::Raster > outputDiskRasterPtr;
  {
    std::vector< te::rst::BandProperty* > bProps;
    bProps.push_back( new te::rst::BandProperty(
      *outputMEMRasterPtr->getBand( 0 )->getProperty() ) );

    std::map<std::string, std::string> auxRasterInfo;
    auxRasterInfo["URI"] = "terralib_unittest_classification_ISOData_Test.tif";

    outputDiskRasterPtr.reset( te::rst::RasterFactory::make(
      "GDAL",
      new te::rst::Grid( *outputMEMRasterPtr->getGrid() ),
      bProps,
      auxRasterInfo,
      0,
      0 ) );
    BOOST_REQUIRE( outputMEMRasterPtr.get() );
  }

  const unsigned int nRows = outputMEMRasterPtr->getNumberOfRows();
  const unsigned int nCols = outputMEMRasterPtr->getNumberOfColumns();
  unsigned int row = 0;
  unsigned int col = 0;
  double value = 0;

  for( row = 0 ; row < nRows ; ++row )
  {
    for( col = 0 ; col < nCols ; ++col )
    {
      outputMEMRasterPtr->getValue( col, row, value, 0 );
      outputDiskRasterPtr->setValue( col, row, value, 0 );
    }
  }

  return;
}

BOOST_AUTO_TEST_SUITE_END()
