/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file TsMAP.cpp
 
  \brief A test suit for the SynchronizedRaster Class.
 */

// TerraLib
#include <terralib_buildconfig.h>

#include <terralib/classification.h>
#include <terralib/raster.h>
#include "Config.h"

// STL
#include <memory>
#include <vector>

// Boost
#include <boost/test/unit_test.hpp>

BOOST_AUTO_TEST_SUITE ( classification_kmeans_tests )

BOOST_AUTO_TEST_CASE (kmeans_test_1)
{
  std::vector<unsigned int> attributesIndices;
  attributesIndices.push_back( 0 );
  attributesIndices.push_back( 1 );
  attributesIndices.push_back( 2 );
  
  std::vector< unsigned int > bands;
  bands.push_back( 0 );
  bands.push_back( 1 );
  bands.push_back( 2 );   

  /* Openning input raster */

  std::map<std::string, std::string> auxRasterInfo;

  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers_rgb342_crop1.tif";
  std::unique_ptr< te::rst::Raster > inputRasterPtrPointer ( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_CHECK( inputRasterPtrPointer.get() );

  // creating the output raster

  std::vector< te::rst::BandProperty* > bProps;
  bProps.push_back( new te::rst::BandProperty( *inputRasterPtrPointer->getBand( 0 )->getProperty() ) );

  auxRasterInfo["URI"] = "terralib_unittest_kmeans_test_1.tif";

  std::unique_ptr< te::rst::Raster > outputRasterPtr( te::rst::RasterFactory::make(
    "GDAL",
    new te::rst::Grid( *inputRasterPtrPointer->getGrid() ),
    bProps,
    auxRasterInfo,
    0,
    0 ) );

  // training  
  
  te::rst::RasterIterator< double > rasterIt = 
    te::rst::RasterIterator< double >::begin( inputRasterPtrPointer.get(), bands );
  te::rst::RasterIterator< double > rasterItEnd = 
    te::rst::RasterIterator< double >::end( inputRasterPtrPointer.get(), bands );

  te::cl::KMeans< te::rst::RasterIterator< double >,
    te::rst::RasterIterator< double > > classifierInstance;

  te::cl::KMeans< te::rst::RasterIterator< double >,
    te::rst::RasterIterator< double > >::Parameters params;
  params.m_K = 2;
  params.m_maxIterations = 10;
  params.m_epsilon = 10;

  BOOST_CHECK( classifierInstance.initialize( params ) );
  
  BOOST_CHECK( classifierInstance.train( rasterIt, rasterItEnd,
    attributesIndices, std::vector<unsigned int>(), true ) );
  
  // Classifying  
  
  rasterIt = 
    te::rst::RasterIterator< double >::begin( inputRasterPtrPointer.get(), bands );
  rasterItEnd = 
    te::rst::RasterIterator< double >::end( inputRasterPtrPointer.get(), bands );  

  std::vector<unsigned int> classification;

  BOOST_CHECK( classifierInstance.classify( rasterIt, rasterItEnd, 
    attributesIndices, classification, true ) );
  
  const std::size_t classificationSize = classification.size();
  const unsigned int outputNRows = outputRasterPtr->getNumberOfRows();
  const unsigned int outputNCols = outputRasterPtr->getNumberOfColumns();
  
  for( std::size_t classificationIdx = 0 ; classificationIdx < classificationSize ;
    ++classificationIdx )
  {
    outputRasterPtr->setValue( classificationIdx % outputNCols,
      classificationIdx / outputNCols, classification[ classificationIdx ], 0 );
  }
}

// Testing with samples with one zero range feature
BOOST_AUTO_TEST_CASE (kmeans_test_2)
{
  // creating samples

  std::vector< std::vector< double > > samples;
  std::vector<unsigned int> samplesAttributesIndices;
  {
    samplesAttributesIndices.push_back( 0 );
    samplesAttributesIndices.push_back( 1 );
    samplesAttributesIndices.push_back( 2 );

    std::vector< double > sample;

    // vegetation

    sample.clear();
    sample.push_back( 41 );
    sample.push_back( 212 );
    sample.push_back( 255 );
    samples.push_back( sample );

    sample.clear();
    sample.push_back( 38 );
    sample.push_back( 213 );
    sample.push_back( 255 );
    samples.push_back( sample );

    sample.clear();
    sample.push_back( 42 );
    sample.push_back( 128 );
    sample.push_back( 255 );
    samples.push_back( sample );

    sample.clear();
    sample.push_back( 31 );
    sample.push_back( 159 );
    sample.push_back( 255 );
    samples.push_back( sample );

    // water

    sample.clear();
    sample.push_back( 39 );
    sample.push_back( 36 );
    sample.push_back( 255 );
    samples.push_back( sample );

    sample.clear();
    sample.push_back( 35 );
    sample.push_back( 41 );
    sample.push_back( 255 );
    samples.push_back( sample );

    sample.clear();
    sample.push_back( 39 );
    sample.push_back( 42 );
    sample.push_back( 255 );
    samples.push_back( sample );

    sample.clear();
    sample.push_back( 38 );
    sample.push_back( 38 );
    sample.push_back( 255 );
    samples.push_back( sample ); 
  }

  // training

  te::cl::KMeans< std::vector< std::vector< double > >::iterator,
    std::vector< std::vector< double > >::iterator > classifierInstance;

  te::cl::KMeans< std::vector< std::vector< double > >::iterator,
    std::vector< std::vector< double > >::iterator >::Parameters params;
  params.m_K = 2;
  params.m_maxIterations = 10;
  params.m_epsilon = 10;

  BOOST_CHECK( classifierInstance.initialize( params ) );
  
  std::vector< std::vector< double > >::iterator samplesIt = samples.begin();
  std::vector< std::vector< double > >::iterator samplesItEnd = samples.end();

  BOOST_CHECK( classifierInstance.train( samplesIt, samplesItEnd,
    samplesAttributesIndices, std::vector<unsigned int>(), true ) );
  
  // Classifying
  
  samplesIt = samples.begin();
  samplesItEnd = samples.end();  
  
  std::vector<unsigned int> classification;
  
  BOOST_CHECK( classifierInstance.classify( samplesIt, samplesItEnd, 
    samplesAttributesIndices, classification, true ) );
  
  BOOST_CHECK_EQUAL( classification.size(), samples.size() );
}

BOOST_AUTO_TEST_SUITE_END()
