/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file TsMaxLikelihood.cpp
 
  \brief A test suit for the MaxLikelihood Class.
 */

// TerraLib
#include <terralib_buildconfig.h>

#include <terralib/classification.h>
#include <terralib/raster.h>
#include "Config.h"

// STL
#include <memory>
#include <vector>

// Boost
#include <boost/test/unit_test.hpp>

    /*!
      \class RasterInputAdaptor
      \brief Classifiers raster input data adaptor.

    */
    template<typename DataType>
    class RasterInputAdaptor : public te::cl::InputAdaptor< DataType >
    {
      public:

        RasterInputAdaptor( const te::rst::Raster& raster )
        : m_raster( raster ), m_nRows( raster.getNumberOfRows() ),
          m_nCols( raster.getNumberOfColumns() )
        {};

        ~RasterInputAdaptor() {};

        // overload
        void getFeature(const unsigned int& elementIndex, const unsigned int& featureIndex,
          DataType& featureValue ) const
        {
          m_raster.getValue( elementIndex % m_nCols, elementIndex / m_nCols, featureValue, featureIndex );
        };

        // overload
        unsigned int getElementsCount() const
        {
          return ( m_nRows * m_nCols );
        };

        // overload
        unsigned int getFeaturesCount() const
        {
          return m_raster.getNumberOfBands();
        };

      protected :

        const te::rst::Raster& m_raster;
        unsigned int m_nRows;
        unsigned int m_nCols;
    };

    /*!
      \class RasterOutputAdaptor
      \brief Classifiers raster output data adaptor.

    */
    template<typename DataType>
    class RasterOutputAdaptor : public te::cl::OutputAdaptor< DataType >
    {
      public:

        RasterOutputAdaptor( te::rst::Raster& raster )
        : m_raster( raster ), m_nRows( raster.getNumberOfRows() ),
          m_nCols( raster.getNumberOfColumns() )
        {};

        ~RasterOutputAdaptor() {};

        // overload
        void setFeature(const unsigned int& elementIndex,
          const unsigned int& featureIndex, const DataType& value )
        {
          m_raster.setValue( elementIndex % m_nCols, elementIndex / m_nCols, (double)value,
            featureIndex );
        };

        // overload
        unsigned int getElementsCount() const
        {
          return ( m_nRows * m_nCols );
        };

        // overload
        unsigned int getFeaturesCount() const
        {
          return m_raster.getNumberOfBands();
        };

      protected :

        te::rst::Raster& m_raster;
        unsigned int m_nRows;
        unsigned int m_nCols;
        double m_value;
    };

BOOST_AUTO_TEST_SUITE ( classification_maxlikelihood_tests )

BOOST_AUTO_TEST_CASE (maxlikelihood_test_1)
{
  // creating samples

  std::vector< std::vector< double > > samples;
  std::vector<unsigned int> samplesAttributesIndices;
  std::vector<unsigned int> samplesLabels;
  {
    samplesAttributesIndices.push_back( 0 );
    samplesAttributesIndices.push_back( 1 );
    samplesAttributesIndices.push_back( 2 );

    std::vector< double > sample;

    // vegetation

    sample.clear();
    sample.push_back( 41 );
    sample.push_back( 212 );
    sample.push_back( 81 );
    samples.push_back( sample );
    samplesLabels.push_back( 1 );

    sample.clear();
    sample.push_back( 38 );
    sample.push_back( 213 );
    sample.push_back( 76 );
    samples.push_back( sample );
    samplesLabels.push_back( 1 );

    sample.clear();
    sample.push_back( 42 );
    sample.push_back( 128 );
    sample.push_back( 70 );
    samples.push_back( sample );
    samplesLabels.push_back( 1 );

    sample.clear();
    sample.push_back( 31 );
    sample.push_back( 159 );
    sample.push_back( 59 );
    samples.push_back( sample );
    samplesLabels.push_back( 1 );

    // water

    sample.clear();
    sample.push_back( 39 );
    sample.push_back( 36 );
    sample.push_back( 81 );
    samples.push_back( sample );
    samplesLabels.push_back( 20 );

    sample.clear();
    sample.push_back( 35 );
    sample.push_back( 41 );
    sample.push_back( 72 );
    samples.push_back( sample );
    samplesLabels.push_back( 20 );

    sample.clear();
    sample.push_back( 39 );
    sample.push_back( 42 );
    sample.push_back( 76 );
    samples.push_back( sample );
    samplesLabels.push_back( 20 );

    sample.clear();
    sample.push_back( 38 );
    sample.push_back( 38 );
    sample.push_back( 77 );
    samples.push_back( sample );
    samplesLabels.push_back( 20 );
  }

  /* Openning input raster */

  std::map<std::string, std::string> auxRasterInfo;

  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers_rgb342_crop1.tif";
  std::unique_ptr< te::rst::Raster > inputRasterPtrPointer ( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_CHECK( inputRasterPtrPointer.get() );

  std::vector< unsigned int > bands;
  bands.push_back( 0 );
  bands.push_back( 1 );
  bands.push_back( 2 );

  // creating the output raster

  std::vector< te::rst::BandProperty* > bProps;
  bProps.push_back( new te::rst::BandProperty( *inputRasterPtrPointer->getBand( 0 )->getProperty() ) );

  auxRasterInfo["URI"] = "terralib_unittest_maxlikelihood_test_1.tif";

  std::unique_ptr< te::rst::Raster > outputRasterPtr( te::rst::RasterFactory::make(
    "GDAL",
    new te::rst::Grid( *inputRasterPtrPointer->getGrid() ),
    bProps,
    auxRasterInfo,
    0,
    0 ) );

  // training

  te::cl::VectorOfVectorsInputAdaptor< double > samplesAdaptor( samples );

  te::cl::MaxLikelihood classifierInstance;

  te::cl::MaxLikelihood::Parameters params;

  BOOST_CHECK( classifierInstance.initialize( params ) );

  BOOST_CHECK( classifierInstance.train( samplesAdaptor, samplesAttributesIndices,
    samplesLabels, true ) );

  RasterInputAdaptor< double > inputRasterAdp( *inputRasterPtrPointer );
  RasterOutputAdaptor< unsigned int > outputRasterAdp( *outputRasterPtr );

  std::vector< double > inputNoDataValues( inputRasterPtrPointer->getNumberOfBands(),
    std::numeric_limits< double >::max() );

  BOOST_CHECK( classifierInstance.classify( inputRasterAdp, samplesAttributesIndices,
    inputNoDataValues, outputRasterAdp, 0, 255, true ) );
}

BOOST_AUTO_TEST_CASE (maxlikelihood_icm_test)
{
  // creating samples

  std::vector< std::vector< double > > samples;
  std::vector<unsigned int> samplesAttributesIndices;
  std::vector<unsigned int> samplesLabels;
  {
    samplesAttributesIndices.push_back( 0 );
    samplesAttributesIndices.push_back( 1 );
    samplesAttributesIndices.push_back( 2 );

    std::vector< double > sample;

    // vegetation

    sample.clear();
    sample.push_back( 41 );
    sample.push_back( 212 );
    sample.push_back( 81 );
    samples.push_back( sample );
    samplesLabels.push_back( 1 );

    sample.clear();
    sample.push_back( 38 );
    sample.push_back( 213 );
    sample.push_back( 76 );
    samples.push_back( sample );
    samplesLabels.push_back( 1 );

    sample.clear();
    sample.push_back( 42 );
    sample.push_back( 128 );
    sample.push_back( 70 );
    samples.push_back( sample );
    samplesLabels.push_back( 1 );

    sample.clear();
    sample.push_back( 31 );
    sample.push_back( 159 );
    sample.push_back( 59 );
    samples.push_back( sample );
    samplesLabels.push_back( 1 );

    // water

    sample.clear();
    sample.push_back( 39 );
    sample.push_back( 36 );
    sample.push_back( 81 );
    samples.push_back( sample );
    samplesLabels.push_back( 20 );

    sample.clear();
    sample.push_back( 35 );
    sample.push_back( 41 );
    sample.push_back( 72 );
    samples.push_back( sample );
    samplesLabels.push_back( 20 );

    sample.clear();
    sample.push_back( 39 );
    sample.push_back( 42 );
    sample.push_back( 76 );
    samples.push_back( sample );
    samplesLabels.push_back( 20 );

    sample.clear();
    sample.push_back( 38 );
    sample.push_back( 38 );
    sample.push_back( 77 );
    samples.push_back( sample );
    samplesLabels.push_back( 20 );
  }

  /* Openning input raster */

  std::map<std::string, std::string> auxRasterInfo;

  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers_rgb342_crop1.tif";
  std::unique_ptr< te::rst::Raster > inputRasterPtrPointer ( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_CHECK( inputRasterPtrPointer.get() );

  std::vector< unsigned int > bands;
  bands.push_back( 0 );
  bands.push_back( 1 );
  bands.push_back( 2 );

  // creating the output raster

  std::vector< te::rst::BandProperty* > bProps;
  bProps.push_back( new te::rst::BandProperty( *inputRasterPtrPointer->getBand( 0 )->getProperty() ) );

  auxRasterInfo["URI"] = "terralib_unittest_maxlikelihood_icm_test.tif";

  std::unique_ptr< te::rst::Raster > outputRasterPtr( te::rst::RasterFactory::make(
    "GDAL",
    new te::rst::Grid( *inputRasterPtrPointer->getGrid() ),
    bProps,
    auxRasterInfo,
    0,
    0 ) );

  // training

  te::cl::VectorOfVectorsInputAdaptor< double > samplesAdaptor( samples );

  te::cl::MaxLikelihood classifierInstance;

  te::cl::MaxLikelihood::Parameters params;

  BOOST_CHECK( classifierInstance.initialize( params ) );

  BOOST_CHECK( classifierInstance.train( samplesAdaptor, samplesAttributesIndices,
    samplesLabels, true ) );

  RasterInputAdaptor< double > inputRasterAdp( *inputRasterPtrPointer );
  RasterOutputAdaptor< unsigned int > outputRasterAdp( *outputRasterPtr );

  std::vector< double > inputNoDataValues( inputRasterPtrPointer->getNumberOfBands(),
    std::numeric_limits< double >::max() );

  BOOST_CHECK( classifierInstance.classify( inputRasterAdp, samplesAttributesIndices,
    inputNoDataValues, outputRasterAdp, 0, 255, true ) );
}

BOOST_AUTO_TEST_SUITE_END()
