/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

// TerraLib
#include <terralib_buildconfig.h>
#include "../Config.h"

#include <terralib/dataaccess.h>
#include <terralib/raster.h>

#include <boost/test/unit_test.hpp>

#include <memory>

BOOST_AUTO_TEST_SUITE( raster_tests )

BOOST_AUTO_TEST_CASE( multiple_tif_instances_rasterfactory_test )
{
  std::map<std::string, std::string> auxRasterInfo;

  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers_rgb342_crop1.tif";
  
  std::unique_ptr< te::rst::Raster > inputRaster1Ptr ( 
    te::rst::RasterFactory::open( auxRasterInfo ) );
  BOOST_REQUIRE( inputRaster1Ptr.get() );
  
  std::unique_ptr< te::rst::Raster > inputRaster2Ptr ( 
    te::rst::RasterFactory::open( auxRasterInfo ) );
  BOOST_REQUIRE( inputRaster2Ptr.get() );
  
  std::unique_ptr< te::rst::Raster > inputRaster3Ptr ( 
    te::rst::RasterFactory::open( auxRasterInfo ) );
  BOOST_REQUIRE( inputRaster3Ptr.get() );  
}

BOOST_AUTO_TEST_CASE( multiple_tif_instances_dataset_test )
{
  te::core::URI uri( "file://" TERRALIB_DATA_DIR "/geotiff/cbers_rgb342_crop1.tif" );
  BOOST_REQUIRE( uri.isValid() );
  std::unique_ptr< te::da::DataSource > dataSourcePtr = te::da::DataSourceFactory::make(
    "GDAL", uri );
  BOOST_REQUIRE( dataSourcePtr.get() != nullptr );
  dataSourcePtr->open();
  BOOST_REQUIRE( dataSourcePtr->isValid() );
  
  std::vector<std::string> dsNames1 = dataSourcePtr->getDataSetNames();
  BOOST_REQUIRE( dsNames1.size() == 1 );
  
  std::size_t rpos = 0;
  
  std::unique_ptr<te::da::DataSet> dataset1Ptr = dataSourcePtr->getDataSet(
    dsNames1[ 0 ] );
  BOOST_REQUIRE( dataset1Ptr.get() != nullptr );
  rpos = te::da::GetFirstPropertyPos( dataset1Ptr.get(), te::dt::RASTER_TYPE);
  std::unique_ptr<te::rst::Raster> raster1Ptr = dataset1Ptr->getRaster(rpos);
  BOOST_REQUIRE( raster1Ptr.get() != nullptr );

  std::unique_ptr<te::da::DataSet> dataset2Ptr = dataSourcePtr->getDataSet(
    dsNames1[ 0 ] );
  BOOST_REQUIRE( dataset2Ptr.get() != nullptr );
  rpos = te::da::GetFirstPropertyPos( dataset2Ptr.get(), te::dt::RASTER_TYPE);
  std::unique_ptr<te::rst::Raster> raster2Ptr = dataset2Ptr->getRaster(rpos);
  BOOST_REQUIRE( raster2Ptr.get() != nullptr );  
}

BOOST_AUTO_TEST_CASE( multiple_hdf_instances_rasterfactory_test )
{
  std::map<std::string, std::string> auxRasterInfo;

  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/hdf/MCD12Q1.A2001001.h13v11.051.2014287162617.hdf";
  
  std::unique_ptr< te::rst::Raster > inputRaster1Ptr ( 
    te::rst::RasterFactory::open( auxRasterInfo ) );
  BOOST_REQUIRE( inputRaster1Ptr.get() );
  
  std::unique_ptr< te::rst::Raster > inputRaster2Ptr ( 
    te::rst::RasterFactory::open( auxRasterInfo ) );
  BOOST_REQUIRE( inputRaster2Ptr.get() );
  
  std::unique_ptr< te::rst::Raster > inputRaster3Ptr ( 
    te::rst::RasterFactory::open( auxRasterInfo ) );
  BOOST_REQUIRE( inputRaster3Ptr.get() );  
}

BOOST_AUTO_TEST_CASE( multiple_hdf_instances_dataset_test )
{
  te::core::URI uri( "file://" TERRALIB_DATA_DIR "/hdf/MCD12Q1.A2001001.h13v11.051.2014287162617.hdf" );
  BOOST_REQUIRE( uri.isValid() );
  std::unique_ptr< te::da::DataSource > dataSourcePtr = te::da::DataSourceFactory::make(
    "GDAL", uri );
  BOOST_REQUIRE( dataSourcePtr.get() != nullptr );
  dataSourcePtr->open();
  BOOST_REQUIRE( dataSourcePtr->isValid() );
  
  std::vector<std::string> dsNames1 = dataSourcePtr->getDataSetNames();
  BOOST_REQUIRE( dsNames1.size() == 16 );
  
  std::size_t rpos = 0;
  
  std::unique_ptr<te::da::DataSet> dataset1Ptr = dataSourcePtr->getDataSet(
    dsNames1[ 0 ] );
  BOOST_REQUIRE( dataset1Ptr.get() != nullptr );
  rpos = te::da::GetFirstPropertyPos( dataset1Ptr.get(), te::dt::RASTER_TYPE);
  std::unique_ptr<te::rst::Raster> raster1Ptr = dataset1Ptr->getRaster(rpos);
  BOOST_REQUIRE( raster1Ptr.get() != nullptr );

  std::unique_ptr<te::da::DataSet> dataset2Ptr = dataSourcePtr->getDataSet(
    dsNames1[ 0 ] );
  BOOST_REQUIRE( dataset2Ptr.get() != nullptr );
  rpos = te::da::GetFirstPropertyPos( dataset2Ptr.get(), te::dt::RASTER_TYPE);
  std::unique_ptr<te::rst::Raster> raster2Ptr = dataset2Ptr->getRaster(rpos);
  BOOST_REQUIRE( raster2Ptr.get() != nullptr );  
}

BOOST_AUTO_TEST_SUITE_END ()
