/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

// Unit-Test TerraLib
#include "TsValidation.h"
#include "../Config.h"

#include "../../src/terralib/dataaccess.h"
#include "../../src/terralib/geometry.h"
#include "../../src/terralib/memory.h"

CPPUNIT_TEST_SUITE_REGISTRATION( TsValidation );

bool TsValidation::loadSHPFile( const std::string& shpFileName,
  std::vector< te::gm::Geometry* >& geomPtrs )
{
  std::string connInfo("file://");
  connInfo += shpFileName;
  std::unique_ptr< te::da::DataSource > ogrDataSourcePtr(
    te::da::DataSourceFactory::make("OGR", connInfo) );
  CPPUNIT_ASSERT( ogrDataSourcePtr.get() != 0 );

  ogrDataSourcePtr->open();
  
  CPPUNIT_ASSERT( ogrDataSourcePtr->isOpened() );
  
  std::vector<std::string> dataSetNames = ogrDataSourcePtr->getDataSetNames();
  std::unique_ptr< te::da::DataSourceTransactor > transactorPtr =
    ogrDataSourcePtr->getTransactor();
  
  for( unsigned int dataSetNamesIdx = 0 ; dataSetNamesIdx < dataSetNames.size() ;
    ++dataSetNamesIdx )
  {
    std::unique_ptr<te::da::DataSet> datasetPtr = transactorPtr->getDataSet(
      dataSetNames[ dataSetNamesIdx ] );
    
    const std::size_t nProperties = datasetPtr->getNumProperties();
    
    CPPUNIT_ASSERT( datasetPtr->moveBeforeFirst() );
    
    while( datasetPtr->moveNext() )
    {
      for( std::size_t propIdx = 0 ; propIdx < nProperties ; ++propIdx )
      {
        if( datasetPtr->getPropertyDataType( propIdx ) == te::dt::GEOMETRY_TYPE )
        {
          std::unique_ptr< te::gm::Geometry > geomPtr = datasetPtr->getGeometry(
            propIdx );
          
          if( geomPtr->getGeomTypeId() == te::gm::PolygonType )
          {
            geomPtrs.push_back( geomPtr.release() );
          }
          else if( geomPtr->getGeomTypeId() == te::gm::MultiPolygonType )
          {
            te::gm::MultiPolygon const * const mPolPtr = (te::gm::MultiPolygon const *)
              geomPtr.get();
              
            for( std::size_t gIdx = 0 ; gIdx < mPolPtr->getNumGeometries() ;
              ++gIdx )
            {
              geomPtrs.push_back( new te::gm::Polygon( *( (te::gm::Polygon*)
                mPolPtr->getGeometryN( gIdx ) ) ) );
            }
          }
        }
      }
    }
  }
  
  return true;
}

void TsValidation::saveToShp( std::vector<te::gm::Polygon*> polygons, 
  std::vector< double >& polygonsValues,
  const std::string& shpBaseFileName )
{
  std::unique_ptr<te::da::DataSetType> dataSetTypePtr1(new te::da::DataSetType(shpBaseFileName));
  
  dataSetTypePtr1->add( new te::dt::SimpleProperty("value", te::dt::DOUBLE_TYPE, true) );  
  dataSetTypePtr1->add( new te::dt::SimpleProperty("id", te::dt::DOUBLE_TYPE, true) );
  dataSetTypePtr1->add( new te::gm::GeometryProperty("polygon", polygons[ 0 ]->getSRID(), 
    te::gm::PolygonType, true) );  
  
  std::unique_ptr<te::da::DataSetType> dataSetTypePtr2( new te::da::DataSetType( *dataSetTypePtr1 ) );
  
  std::unique_ptr< te::mem::DataSet > memDataSetPtr( new te::mem::DataSet( dataSetTypePtr1.get()) );
  
  for( unsigned int polygonsIdx = 0 ; polygonsIdx < polygons.size() ; ++polygonsIdx )
  {
    te::mem::DataSetItem* dsItemPtr = new te::mem::DataSetItem(memDataSetPtr.get());
    dsItemPtr->setDouble( 0, polygonsValues[ polygonsIdx ] );
    dsItemPtr->setDouble( 1, polygonsIdx );
    dsItemPtr->setGeometry( 2, (te::gm::Geometry*)polygons[ polygonsIdx ]->clone() );
    
    memDataSetPtr->add( dsItemPtr );
  }

  remove( ( shpBaseFileName + ".shx" ).c_str() );
  remove( ( shpBaseFileName + ".shp" ).c_str() );
  remove( ( shpBaseFileName + ".prj" ).c_str() );
  remove( ( shpBaseFileName + ".dbf" ).c_str() );
  
  std::string connInfo("file://" + shpBaseFileName + ".shp");
  
  std::unique_ptr<te::da::DataSource> dsOGR( te::da::DataSourceFactory::make("OGR", connInfo) );
  dsOGR->open();
  
  memDataSetPtr->moveBeforeFirst();
  
  te::da::Create(dsOGR.get(), dataSetTypePtr2.get(), memDataSetPtr.get());

  dsOGR->close();
} 

void TsValidation::setUp()
{
}

void TsValidation::tearDown()
{
}

void TsValidation::tcMakeValid_hole_lies_outside_shell()
{
  std::vector< te::gm::Geometry* > geomPtrs;
  
  CPPUNIT_ASSERT( loadSHPFile( TERRALIB_DATA_DIR 
    "/shape/invalid_geometries_hole_lies_outside_shell.shp", geomPtrs ) );
  CPPUNIT_ASSERT( geomPtrs.size() == 2 );
  
  te::gm::TopologyValidationError error1;
  CPPUNIT_ASSERT( !te::gm::Validation::CheckValidity( geomPtrs[ 0 ], error1 ) );
  
  te::gm::TopologyValidationError error2;
  CPPUNIT_ASSERT( !te::gm::Validation::CheckValidity( geomPtrs[ 1 ], error2 ) );
  
  te::gm::GeometryPtr fixedGeomPtr1 = te::gm::Validation::MakeValid( geomPtrs[ 0 ] );
  CPPUNIT_ASSERT( fixedGeomPtr1.get() != nullptr );
  CPPUNIT_ASSERT( te::gm::Validation::CheckValidity( fixedGeomPtr1.get(), error1 ) );
  
  te::gm::GeometryPtr fixedGeomPtr2 = te::gm::Validation::MakeValid( geomPtrs[ 1 ] );
  CPPUNIT_ASSERT( fixedGeomPtr2.get() != nullptr );
  CPPUNIT_ASSERT( te::gm::Validation::CheckValidity( fixedGeomPtr2.get(), error1 ) );  
  
  delete geomPtrs[ 0 ];
  delete geomPtrs[ 1 ];  
  
  std::vector<te::gm::Polygon*> polygons;
  polygons.push_back( (te::gm::Polygon*)
    (((te::gm::MultiPolygon*)fixedGeomPtr1.get())->getGeometryN( 0 ) ) );
  polygons.push_back( (te::gm::Polygon*)
    (((te::gm::MultiPolygon*)fixedGeomPtr2.get())->getGeometryN( 0 ) ) );
  
  std::vector< double > polygonsValues( polygons.size(), 0 );
  
  saveToShp( polygons, polygonsValues, "TsValidation_tcMakeValid_hole_lies_outside_shell" );
}

void TsValidation::tcValidate_hole_lies_outside_shell()
{
  std::vector< te::gm::Geometry* > geomPtrs;
  
  CPPUNIT_ASSERT( loadSHPFile( TERRALIB_DATA_DIR 
    "/shape/invalid_geometries_hole_lies_outside_shell.shp", geomPtrs ) );
  CPPUNIT_ASSERT( geomPtrs.size() == 2 );
  
  te::gm::TopologyValidationError error1;
  CPPUNIT_ASSERT( !te::gm::Validation::CheckValidity( geomPtrs[ 0 ], error1 ) );
  
  te::gm::TopologyValidationError error2;
  CPPUNIT_ASSERT( !te::gm::Validation::CheckValidity( geomPtrs[ 1 ], error2 ) );
  
  te::gm::GeometryPtr fixedGeomPtr1( te::gm::Validation::Validate( geomPtrs[ 0 ] ) );
  CPPUNIT_ASSERT( fixedGeomPtr1.get() != nullptr );
  CPPUNIT_ASSERT( te::gm::Validation::CheckValidity( fixedGeomPtr1.get(), error1 ) );
  
  te::gm::GeometryPtr fixedGeomPtr2( te::gm::Validation::Validate( geomPtrs[ 1 ] ) );
  CPPUNIT_ASSERT( fixedGeomPtr2.get() != nullptr );
  CPPUNIT_ASSERT( te::gm::Validation::CheckValidity( fixedGeomPtr2.get(), error1 ) );  
  
  delete geomPtrs[ 0 ];
  delete geomPtrs[ 1 ];
  
  std::vector<te::gm::Polygon*> polygons;
  polygons.push_back( (te::gm::Polygon*)
    (((te::gm::MultiPolygon*)fixedGeomPtr1.get())->getGeometryN( 0 ) ) );
  polygons.push_back( (te::gm::Polygon*)
    (((te::gm::MultiPolygon*)fixedGeomPtr2.get())->getGeometryN( 0 ) ) );
  
  std::vector< double > polygonsValues( polygons.size(), 0 );
  
  saveToShp( polygons, polygonsValues, "TsValidation_tcValidate_hole_lies_outside_shell" );  
}

void TsValidation::tcMakeValidWithSnap_hole_lies_outside_shell()
{
  std::vector< te::gm::Geometry* > geomPtrs;
  
  CPPUNIT_ASSERT( loadSHPFile( TERRALIB_DATA_DIR 
    "/shape/invalid_geometries_hole_lies_outside_shell.shp", geomPtrs ) );
  CPPUNIT_ASSERT( geomPtrs.size() == 2 );
  
  te::gm::TopologyValidationError error1;
  CPPUNIT_ASSERT( !te::gm::Validation::CheckValidity( geomPtrs[ 0 ], error1 ) );
  
  te::gm::TopologyValidationError error2;
  CPPUNIT_ASSERT( !te::gm::Validation::CheckValidity( geomPtrs[ 1 ], error2 ) );
  
  te::gm::GeometryPtr fixedGeomPtr1 = te::gm::Validation::MakeValidWithSnap( geomPtrs[ 0 ] );
  CPPUNIT_ASSERT( fixedGeomPtr1.get() != nullptr );
  CPPUNIT_ASSERT( te::gm::Validation::CheckValidity( fixedGeomPtr1.get(), error1 ) );
  
  te::gm::GeometryPtr fixedGeomPtr2 = te::gm::Validation::MakeValidWithSnap( geomPtrs[ 1 ] );
  CPPUNIT_ASSERT( fixedGeomPtr2.get() != nullptr );
  CPPUNIT_ASSERT( te::gm::Validation::CheckValidity( fixedGeomPtr2.get(), error1 ) );  
  
  delete geomPtrs[ 0 ];
  delete geomPtrs[ 1 ];
  
  std::vector<te::gm::Polygon*> polygons;
  polygons.push_back( (te::gm::Polygon*)
    (((te::gm::MultiPolygon*)fixedGeomPtr1.get())->getGeometryN( 0 ) ) );
  polygons.push_back( (te::gm::Polygon*)
    (((te::gm::MultiPolygon*)fixedGeomPtr2.get())->getGeometryN( 0 ) ) );
  
  std::vector< double > polygonsValues( polygons.size(), 0 );
  
  saveToShp( polygons, polygonsValues, "TsValidation_tcMakeValidWithSnap_hole_lies_outside_shell" );  
}