/*  Copyright (C) 2017 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
  \file terralib/unittest/ogc-wms/TsWMSClient.cpp

  \brief A test suit for the WS OGC WMSClient.

  \author Emerson Moraes
 */

 //TerraLib
 #include <terralib/ws/ogc/wms/client/WMSClient.h>
 #include <terralib/ws/ogc/wms/client/DataTypes.h>
 #include <terralib/ws/core/Exception.h>
 #include <terralib/qt/af/ApplicationController.h>

 //Boost
 #include <boost/test/unit_test.hpp>

//STL
#include <memory>
#include <iostream>

BOOST_AUTO_TEST_CASE(getcapabilities_test)
{
  te::qt::af::AppCtrlSingleton::getInstance().initialize();
  
  // Directory where test WMS test data is stored.
  std::string dataDir = TERRALIB_DATA_DIR;

  //WMS GeoSolutions Demo server
  std::string url = "http://demo.geo-solutions.it/geoserver/ows";

  //Version that will be used on WMS Requests.
  std::string version = "1.3.0"; 

  te::ws::ogc::WMSClient client (dataDir, url, version);

  // Makes GetCapabilities Test Request.
  BOOST_CHECK_NO_THROW(client.updateCapabilities());

  te::ws::ogc::wms::WMSCapabilities capabilities = client.getCapabilities();

  te::ws::ogc::wms::Capability capability = capabilities.m_capability;

  std::vector<te::ws::ogc::wms::Layer> layers = capability.m_layer.m_layers;

  BOOST_CHECK(layers.size() > 0);

  return;
}

BOOST_AUTO_TEST_CASE(getcapabilities_exception_test)
{

  te::qt::af::AppCtrlSingleton::getInstance().initialize();

  // Directory where test WMS test data is stored.
  std::string dataDir = TERRALIB_DATA_DIR;

  // A wrong WMS server URL.
  std::string url = "http://wrongaddress.org/geoserver/ows";

  // Version that will be used on WMS Requests.
  std::string version = "1.3.0";

  te::ws::ogc::WMSClient client (dataDir, url, version);

  // Makes GetCapabilities Request.
  BOOST_CHECK_THROW(client.updateCapabilities(), te::ws::core::Exception);

  return;
}


BOOST_AUTO_TEST_CASE(not_supported_version_test)
{
  te::qt::af::AppCtrlSingleton::getInstance().initialize();

  // Directory where test WMS test data is stored.
  std::string dataDir = TERRALIB_DATA_DIR;

  // WMS server URL.
  std::string url = "http://demo.geo-solutions.it/geoserver/ows";

  // Version not supported yet.
  std::string version = "1.0.0";

  te::ws::ogc::WMSClient client (dataDir, url, version);

  BOOST_CHECK_THROW(client.updateCapabilities(), te::ws::core::Exception);

  return;
}

BOOST_AUTO_TEST_CASE(getmap_test)
{
  te::qt::af::AppCtrlSingleton::getInstance().initialize();

  // Directory where test WMS test data is stored.
  std::string dataDir = TERRALIB_DATA_DIR;

  // WMS server URL.
  std::string url = "http://demo.geo-solutions.it/geoserver/ows";

  std::string version = "1.3.0";

  te::ws::ogc::WMSClient client (dataDir, url, version);

  client.updateCapabilities();

  te::ws::ogc::wms::WMSCapabilities capabilities = client.getCapabilities();

  te::ws::ogc::wms::Capability capability = capabilities.m_capability;

  std::vector<te::ws::ogc::wms::Layer> layers = capability.m_layer.m_layers;

  te::ws::ogc::wms::WMSGetMapRequest request;

  for(size_t i = 0; i < layers.size(); i++)
  {
    if(layers[i].m_name == "tasmania")
    {
      request.m_layers.push_back(layers[i].m_name);

      if(layers[i].m_styles.size() > 0)
      {
        request.m_styles.push_back(layers[i].m_styles[0].m_name);
      }

      request.m_srs = "EPSG:4326";

      te::ws::ogc::wms::BoundingBox bbox;
      bbox.m_crs = request.m_srs;
      bbox.m_maxX = layers[i].m_geoBoundingBox.m_eastBoundLongitude;
      bbox.m_maxY = layers[i].m_geoBoundingBox.m_northBoundLatitude;
      bbox.m_minX = layers[i].m_geoBoundingBox.m_westBoundLongitude;
      bbox.m_minY = layers[i].m_geoBoundingBox.m_southBoundLatitude;

      request.m_boundingBox = bbox;

      break;
    }
  }

  BOOST_CHECK(request.m_layers.size() == 1);

  te::ws::ogc::wms::WMSGetMapResponse response = client.getMap(request);

  BOOST_CHECK(response.m_buffer.empty() == false);
  BOOST_CHECK(response.m_size > 0);
  BOOST_CHECK(response.m_format == "image/png");

  return;
}


BOOST_AUTO_TEST_CASE(savegetmap_test)
{
    te::qt::af::AppCtrlSingleton::getInstance().initialize();

  // Directory where test WMS test data is stored.
  std::string dataDir = TERRALIB_DATA_DIR;

  // WMS server URL.
  std::string url = "http://demo.geo-solutions.it/geoserver/ows";

  std::string version = "1.3.0";

  te::ws::ogc::WMSClient client (dataDir, url, version);

  client.updateCapabilities();

  te::ws::ogc::wms::WMSCapabilities capabilities = client.getCapabilities();

  te::ws::ogc::wms::Capability capability = capabilities.m_capability;

  std::vector<te::ws::ogc::wms::Layer> layers = capability.m_layer.m_layers;

  te::ws::ogc::wms::WMSGetMapRequest request;

  for(size_t i = 0; i < layers.size(); i++)
  {
    if(layers[i].m_name == "tasmania")
    {
      request.m_layers.push_back(layers[i].m_name);

      if(layers[i].m_styles.size() > 0)
      {
        request.m_styles.push_back(layers[i].m_styles[0].m_name);
      }

      request.m_srs = "EPSG:4326";

      te::ws::ogc::wms::BoundingBox bbox;
      bbox.m_crs = request.m_srs;
      bbox.m_maxX = layers[i].m_geoBoundingBox.m_eastBoundLongitude;
      bbox.m_maxY = layers[i].m_geoBoundingBox.m_northBoundLatitude;
      bbox.m_minX = layers[i].m_geoBoundingBox.m_westBoundLongitude;
      bbox.m_minY = layers[i].m_geoBoundingBox.m_southBoundLatitude;

      request.m_boundingBox = bbox;

      break;
    }
  }

  BOOST_CHECK(request.m_layers.size() == 1);

  std::string imageDiskPath = client.saveGetMap(request, "test_image.png");

  std::string expectedResult = dataDir + std::string("/wms/") + "test_image.png";

  BOOST_CHECK(imageDiskPath == expectedResult);
}
