/*  Copyright (C) 2017 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

  /*!
  \file terralib/unittest/ogc-wms/TsWMSDataAccess.cpp

  \brief A test suit for the WS OGC WMS DataAccess.

  \author Emerson Moraes
 */

//TerraLib


#include <terralib/ws/ogc/wms/client/WMSClient.h>
#include <terralib/ws/ogc/wms/client/WMSClient.h>
#include <terralib/ws/ogc/wms/client/DataTypes.h>
#include <terralib/ws/core/Exception.h>
#include <terralib/ws/ogc/wms/dataaccess/Config.h>
#include <terralib/ws/ogc/wms/dataaccess/DataSourceFactory.h>
#include <terralib/ws/ogc/wms/dataaccess/DataSource.h>
#include <terralib/dataaccess/datasource/DataSource.h>
#include <terralib/dataaccess/datasource/DataSourceFactory.h>
#include <terralib/dataaccess/datasource/DataSourceManager.h>
#include <terralib/qt/af/ApplicationController.h>
#include <terralib/core/uri/Utils.h>


// Boost
#include <boost/test/unit_test.hpp>



BOOST_AUTO_TEST_CASE(datasource_open_test)
{
  te::qt::af::AppCtrlSingleton::getInstance().initialize();

  te::da::DataSourceFactory::add(TE_OGC_WMS_DRIVER_IDENTIFIER, te::ws::ogc::wms::da::Build);

  // Directory where temporary WMS data will be stored.
  std::string usrDataDir = te::qt::af::AppCtrlSingleton::getInstance().getUserDataDir().toUtf8().data();

  // WMS server URL.
  std::string url = "http://demo.geo-solutions.it/geoserver/ows";

  // Version that will be used on WMS Requests.
  std::string version = "1.3.0";

  std::string encodedURL = te::core::URIEncode(url);

  std::string connInfo("wms://?URI=" + encodedURL + "&VERSION=" + version + "&USERDATADIR=" + usrDataDir);

  // Perform connection
  std::unique_ptr<te::da::DataSource> ds = te::da::DataSourceFactory::make("WMS2", connInfo);
  ds->open();

  BOOST_CHECK(ds->isOpened());

  return;
}

BOOST_AUTO_TEST_CASE(datasource_open_exception_test)
{
  te::qt::af::AppCtrlSingleton::getInstance().initialize();

  // Directory where temporary WMS data will be stored.
  std::string usrDataDir = te::qt::af::AppCtrlSingleton::getInstance().getUserDataDir().toUtf8().data();

  // WMS server URL.
  std::string url = "http://demo.geo-solutions.it/geoserver/ows";

  std::string encodedURL = te::core::URIEncode(url);

  // Version that will be used on WMS Requests.
  std::string version = "1.3.0";

  std::string connInfo;

  // Perform connection
  std::unique_ptr<te::da::DataSource> ds = te::da::DataSourceFactory::make("WMS2", connInfo);

  // Test without any correct connection parameter
  BOOST_CHECK(ds.get() == nullptr);

  connInfo = "wms://?URI=" + encodedURL;
  ds.reset(te::da::DataSourceFactory::make("WMS2", connInfo).release());

  // Test with URI correct connection parameter
  BOOST_CHECK_THROW(ds->open(), te::ws::core::Exception);

  connInfo = "wms://?VERSION=" + version;
  ds.reset(te::da::DataSourceFactory::make("WMS2", connInfo).release());

  // Test with VERSION correct connection parameter
  BOOST_CHECK_THROW(ds->open(), te::ws::core::Exception);

  connInfo = "wms://?USERDATADIR=" + usrDataDir;
  ds.reset(te::da::DataSourceFactory::make("WMS2", connInfo).release());

  // Test with USERDATADIR correct connection parameter
  BOOST_CHECK_THROW(ds->open(), te::ws::core::Exception);

  connInfo = "wms://?URI=" + encodedURL + "&VERSION=" + version;
  ds.reset(te::da::DataSourceFactory::make("WMS2", connInfo).release());

  // Test with URI and VERSION correct connection parameters
  BOOST_CHECK_THROW(ds->open(), te::ws::core::Exception);

  connInfo = "wms://?URI=" + encodedURL + "&USERDATADIR=" + usrDataDir;
  ds.reset(te::da::DataSourceFactory::make("WMS2", connInfo).release());

  // Test with URI and USERDATADIR correct connection parameters
  BOOST_CHECK_THROW(ds->open(), te::ws::core::Exception);

  connInfo = "wms://?VERSION=" + version + "&USERDATADIR=" + usrDataDir;
  ds.reset(te::da::DataSourceFactory::make("WMS2", connInfo).release());

  // Test with VERSION and USERDATADIR correct connection parameters
  BOOST_CHECK_THROW(ds->open(), te::ws::core::Exception);

  return;
}

BOOST_AUTO_TEST_CASE(datasource_getmap_test)
{
  te::qt::af::AppCtrlSingleton::getInstance().initialize();

  // Directory where temporary WMS data will be stored.
  std::string usrDataDir = te::qt::af::AppCtrlSingleton::getInstance().getUserDataDir().toUtf8().data();

  // WMS server URL.
  std::string url = "http://demo.geo-solutions.it/geoserver/ows";

  // Version that will be used on WMS Requests.
  std::string version = "1.3.0";

  std::string encodedURL = te::core::URIEncode(url);

  std::string connInfo("wms://?URI=" + encodedURL + "&VERSION=" + version + "&USERDATADIR=" + usrDataDir);

  // Perform connection
  std::unique_ptr<te::da::DataSource> ds = te::da::DataSourceFactory::make("WMS2", connInfo);
  ds->open();

  te::ws::ogc::wms::da::DataSource* wmsDs = dynamic_cast<te::ws::ogc::wms::da::DataSource*>(ds.get());

  std::shared_ptr<te::ws::ogc::WMSClient> client = wmsDs->getWMSClient();

  te::ws::ogc::wms::Capability capability = client->getCapabilities().m_capability;

  std::vector<te::ws::ogc::wms::Layer> layers = capability.m_layer.m_layers;

  te::ws::ogc::wms::WMSGetMapRequest request;

  for(size_t i = 0; i < layers.size(); i++)
  {
    if(layers[i].m_name == "tasmania")
    {
      request.m_layers.push_back(layers[i].m_name);

      if(layers[i].m_styles.size() > 0)
      {
        request.m_styles.push_back(layers[i].m_styles[0].m_name);
      }

      request.m_srs = "EPSG:4326";

      te::ws::ogc::wms::BoundingBox bbox;
      bbox.m_crs = request.m_srs;
      bbox.m_maxX = layers[i].m_geoBoundingBox.m_eastBoundLongitude;
      bbox.m_maxY = layers[i].m_geoBoundingBox.m_northBoundLatitude;
      bbox.m_minX = layers[i].m_geoBoundingBox.m_westBoundLongitude;
      bbox.m_minY = layers[i].m_geoBoundingBox.m_southBoundLatitude;

      request.m_boundingBox = bbox;

      break;
    }
  }

  BOOST_CHECK(request.m_layers.size() == 1);

  te::ws::ogc::wms::WMSGetMapResponse response = wmsDs->getMap(request);

  BOOST_CHECK(response.m_buffer.empty() == false);
  BOOST_CHECK(response.m_size > 0);
  BOOST_CHECK(response.m_format == "image/png");

  return;
}

BOOST_AUTO_TEST_CASE(datasource_savegetmap_test)
{
  te::qt::af::AppCtrlSingleton::getInstance().initialize();

  // Directory where temporary WMS data will be stored.
  std::string usrDataDir = te::qt::af::AppCtrlSingleton::getInstance().getUserDataDir().toUtf8().data();

  // WMS server URL.
  std::string url = "http://demo.geo-solutions.it/geoserver/ows";

  // Version that will be used on WMS Requests.
  std::string version = "1.3.0";

  std::string encodedURL = te::core::URIEncode(url);

  std::string connInfo("wms://?URI=" + encodedURL + "&VERSION=" + version + "&USERDATADIR=" + usrDataDir);

  // Perform connection
  std::unique_ptr<te::da::DataSource> ds = te::da::DataSourceFactory::make("WMS2", connInfo);
  ds->open();

  te::ws::ogc::wms::da::DataSource* wmsDs = dynamic_cast<te::ws::ogc::wms::da::DataSource*>(ds.get());

  std::shared_ptr<te::ws::ogc::WMSClient> client = wmsDs->getWMSClient();

  te::ws::ogc::wms::Capability capability = client->getCapabilities().m_capability;

  std::vector<te::ws::ogc::wms::Layer> layers = capability.m_layer.m_layers;

  te::ws::ogc::wms::WMSGetMapRequest request;

  for(size_t i = 0; i < layers.size(); i++)
  {
    if(layers[i].m_name == "tasmania")
    {
      request.m_layers.push_back(layers[i].m_name);

      if(layers[i].m_styles.size() > 0)
      {
        request.m_styles.push_back(layers[i].m_styles[0].m_name);
      }

      request.m_srs = "EPSG:4326";

      te::ws::ogc::wms::BoundingBox bbox;
      bbox.m_crs = request.m_srs;
      bbox.m_maxX = layers[i].m_geoBoundingBox.m_eastBoundLongitude;
      bbox.m_maxY = layers[i].m_geoBoundingBox.m_northBoundLatitude;
      bbox.m_minX = layers[i].m_geoBoundingBox.m_westBoundLongitude;
      bbox.m_minY = layers[i].m_geoBoundingBox.m_southBoundLatitude;

      request.m_boundingBox = bbox;

      break;
    }
  }

  BOOST_CHECK(request.m_layers.size() == 1);

  std::string imageDiskPath = wmsDs->saveGetMap(request, "test_image.png");

  std::string expectedResult = usrDataDir + std::string("/wms/") + "test_image.png";

  BOOST_CHECK(imageDiskPath == expectedResult);
}

