/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */


// TerraLib
#include "../../Config.h"
#include <terralib/ogr/DataSet.h>
#include <terralib/dataaccess.h>
#include <terralib/dataaccess/utils/DataAccess.h>
#include <terralib/datatype.h>
#include <terralib/geometry.h>

// STL
#include <iostream>
#include <exception>
#include <cstdio>

// Boost
#include <boost/test/unit_test.hpp>

BOOST_AUTO_TEST_SUITE( dataaccess_tests )

BOOST_AUTO_TEST_CASE( dataset_copy_to_shp_test )
{
  std::remove( "dataset_copy_to_shp_test.shx" );
  std::remove( "dataset_copy_to_shp_test.shp" );
  std::remove( "dataset_copy_to_shp_test.dbf" );
  std::remove( "dataset_copy_to_shp_test.prj" );

  const std::string inputFileName(TERRALIB_DATA_DIR "/shape/munic_goias.shp");
  const std::string inDsetName = "munic_goias";
  const std::string outDataSetName( "dataset_copy_to_shp_test");

  const std::string outputFileName( outDataSetName + ".shp");

  te::da::DataSourcePtr inputDataSourcePtr(te::da::DataSourceFactory::make(
    "OGR", "file://" + inputFileName));
  inputDataSourcePtr->open();
  BOOST_REQUIRE( inputDataSourcePtr->isValid() );

  std::unique_ptr<te::da::DataSet> inputDataSetPtr =
    inputDataSourcePtr->getDataSet( inDsetName );
  BOOST_REQUIRE( inputDataSetPtr.get() != nullptr );
  BOOST_REQUIRE( !inputDataSetPtr->isEmpty() );
  BOOST_REQUIRE( inputDataSetPtr->moveBeforeFirst() );

  std::unique_ptr<te::da::DataSetType> outDataSetTypePtr =
    inputDataSourcePtr->getDataSetType( inDsetName );
  BOOST_REQUIRE( outDataSetTypePtr.get() != nullptr );
  outDataSetTypePtr->setName( outDataSetName );
  outDataSetTypePtr->setTitle( outDataSetName );

  te::da::DataSourcePtr outputDataSourcePtr(te::da::DataSourceFactory::make(
    "OGR", "file://" + outputFileName));
  outputDataSourcePtr->open();
  BOOST_REQUIRE( outputDataSourcePtr->isValid() );

  te::da::DataAccess dataAccess(outputDataSourcePtr, outDataSetName);

  dataAccess.beginTransaction();

  dataAccess.createDataSet(outDataSetTypePtr.get());

  dataAccess.saveDataSet(inputDataSetPtr.release());

  dataAccess.commitTransaction();
}

BOOST_AUTO_TEST_CASE( dataset_copy_to_gpkg_test )
{
  const std::string inputFileName(TERRALIB_DATA_DIR "/shape/munic_goias.shp");
  const std::string inDsetName = "munic_goias";
  const std::string outDataSetName( "dataset_copy_to_gpkg_test");

  const std::string outputFileName( outDataSetName + ".gpkg");

  std::remove( outputFileName.c_str() );

  te::da::DataSourcePtr inputDataSourcePtr(te::da::DataSourceFactory::make(
    "OGR", "file://" + inputFileName));
  inputDataSourcePtr->open();
  BOOST_REQUIRE( inputDataSourcePtr->isValid() );

  std::unique_ptr<te::da::DataSet> inputDataSetPtr =
    inputDataSourcePtr->getDataSet( inDsetName );
  BOOST_REQUIRE( inputDataSetPtr.get() != nullptr );
  BOOST_REQUIRE( !inputDataSetPtr->isEmpty() );
  BOOST_REQUIRE( inputDataSetPtr->moveBeforeFirst() );

  std::unique_ptr<te::da::DataSetType> outDataSetTypePtr =
    inputDataSourcePtr->getDataSetType( inDsetName );
  BOOST_REQUIRE( outDataSetTypePtr.get() != nullptr );
  outDataSetTypePtr->setName( outDataSetName );
  outDataSetTypePtr->setTitle( outDataSetName );

  te::da::DataSourcePtr outputDataSourcePtr(te::da::DataSourceFactory::make(
    "OGR", "file://" + outputFileName));
  outputDataSourcePtr->open();
  BOOST_REQUIRE( outputDataSourcePtr->isValid() );

  te::da::DataAccess dataAccess(outputDataSourcePtr, outDataSetName);

  dataAccess.beginTransaction();

  dataAccess.createDataSet(outDataSetTypePtr.get());

  dataAccess.saveDataSet(inputDataSetPtr.release());

  dataAccess.commitTransaction();
}

BOOST_AUTO_TEST_CASE( dataset_copy_to_sqlite_test )
{
  const std::string inputFileName(TERRALIB_DATA_DIR "/shape/munic_goias.shp");
  const std::string inDsetName = "munic_goias";
  const std::string outDataSetName( "dataset_copy_to_sqlite_test");

  const std::string outputFileName( outDataSetName + ".sqlite");

  std::remove( outputFileName.c_str() );

  te::da::DataSourcePtr inputDataSourcePtr(te::da::DataSourceFactory::make(
    "OGR", "file://" + inputFileName));
  inputDataSourcePtr->open();
  BOOST_REQUIRE( inputDataSourcePtr->isValid() );

  std::unique_ptr<te::da::DataSet> inputDataSetPtr =
    inputDataSourcePtr->getDataSet( inDsetName );
  BOOST_REQUIRE( inputDataSetPtr.get() != nullptr );
  BOOST_REQUIRE( !inputDataSetPtr->isEmpty() );
  BOOST_REQUIRE( inputDataSetPtr->moveBeforeFirst() );

  std::unique_ptr<te::da::DataSetType> outDataSetTypePtr =
    inputDataSourcePtr->getDataSetType( inDsetName );
  BOOST_REQUIRE( outDataSetTypePtr.get() != nullptr );
  outDataSetTypePtr->setName( outDataSetName );
  outDataSetTypePtr->setTitle( outDataSetName );

  te::da::DataSourcePtr outputDataSourcePtr(te::da::DataSourceFactory::make(
    "OGR", "file://" + outputFileName));
  outputDataSourcePtr->open();
  BOOST_REQUIRE( outputDataSourcePtr->isValid() );

  te::da::DataAccess dataAccess(outputDataSourcePtr, outDataSetName);

  dataAccess.beginTransaction();

  dataAccess.createDataSet(outDataSetTypePtr.get());

  dataAccess.saveDataSet(inputDataSetPtr.release());

  dataAccess.commitTransaction();
}

BOOST_AUTO_TEST_CASE( dataset_copy_selected_to_shp_test )
{
  std::remove( "dataset_copy_selected_to_shp_test.shx" );
  std::remove( "dataset_copy_selected_to_shp_test.shp" );
  std::remove( "dataset_copy_selected_to_shp_test.dbf" );
  std::remove( "dataset_copy_selected_to_shp_test.prj" );

  const std::string inputFileName(TERRALIB_DATA_DIR "/shape/munic_goias.shp");
  const std::string inDsetName = "munic_goias";
  const std::string outDataSetName( "dataset_copy_selected_to_shp_test");

  const std::string outputFileName( outDataSetName + ".shp");

  te::da::DataSourcePtr inputDataSourcePtr(te::da::DataSourceFactory::make(
    "OGR", "file://" + inputFileName));
  inputDataSourcePtr->open();
  BOOST_REQUIRE( inputDataSourcePtr->isValid() );

  std::unique_ptr<te::da::Select> selectPtr = te::da::BuildSelect( inDsetName );
  BOOST_REQUIRE( selectPtr.get() != nullptr );

  std::unique_ptr<te::da::DataSet> inputDataSetPtr =
    inputDataSourcePtr->query( *selectPtr );
  BOOST_REQUIRE( inputDataSetPtr.get() != nullptr );
  BOOST_REQUIRE( !inputDataSetPtr->isEmpty() );
  BOOST_REQUIRE( inputDataSetPtr->moveBeforeFirst() );

  std::unique_ptr<te::da::DataSetType> outDataSetTypePtr =
    inputDataSourcePtr->getDataSetType( inDsetName );
  BOOST_REQUIRE( outDataSetTypePtr.get() != nullptr );
  outDataSetTypePtr->setName( outDataSetName );
  outDataSetTypePtr->setTitle( outDataSetName );

  te::da::DataSourcePtr outputDataSourcePtr(te::da::DataSourceFactory::make(
    "OGR", "file://" + outputFileName));
  outputDataSourcePtr->open();
  BOOST_REQUIRE( outputDataSourcePtr->isValid() );

  te::da::DataAccess dataAccess(outputDataSourcePtr, outDataSetName);

  dataAccess.beginTransaction();

  dataAccess.createDataSet(outDataSetTypePtr.get());

  dataAccess.saveDataSet(inputDataSetPtr.release());

  dataAccess.commitTransaction();
}

BOOST_AUTO_TEST_CASE( dataset_copy_selected_to_gpkg_test )
{
  const std::string inputFileName(TERRALIB_DATA_DIR "/shape/munic_goias.shp");
  const std::string inDsetName = "munic_goias";
  const std::string outDataSetName( "dataset_copy_selected_to_gpkg_test");

  const std::string outputFileName( outDataSetName + ".gpkg");

  std::remove( outputFileName.c_str() );

  te::da::DataSourcePtr inputDataSourcePtr(te::da::DataSourceFactory::make(
    "OGR", "file://" + inputFileName));
  inputDataSourcePtr->open();
  BOOST_REQUIRE( inputDataSourcePtr->isValid() );

  std::unique_ptr<te::da::Select> selectPtr = te::da::BuildSelect( inDsetName );
  BOOST_REQUIRE( selectPtr.get() != nullptr );

  std::unique_ptr<te::da::DataSet> inputDataSetPtr =
    inputDataSourcePtr->query( *selectPtr );
  BOOST_REQUIRE( inputDataSetPtr.get() != nullptr );
  BOOST_REQUIRE( !inputDataSetPtr->isEmpty() );
  BOOST_REQUIRE( inputDataSetPtr->moveBeforeFirst() );

  std::unique_ptr<te::da::DataSetType> outDataSetTypePtr =
    inputDataSourcePtr->getDataSetType( inDsetName );
  BOOST_REQUIRE( outDataSetTypePtr.get() != nullptr );
  outDataSetTypePtr->setName( outDataSetName );
  outDataSetTypePtr->setTitle( outDataSetName );

  te::da::DataSourcePtr outputDataSourcePtr(te::da::DataSourceFactory::make(
    "OGR", "file://" + outputFileName));
  outputDataSourcePtr->open();
  BOOST_REQUIRE( outputDataSourcePtr->isValid() );

  te::da::DataAccess dataAccess(outputDataSourcePtr, outDataSetName);

  dataAccess.beginTransaction();

  dataAccess.createDataSet(outDataSetTypePtr.get());

  dataAccess.saveDataSet(inputDataSetPtr.release());

  dataAccess.commitTransaction();
}

BOOST_AUTO_TEST_CASE( dataset_copy_selected_to_sqlite_test )
{
  const std::string inputFileName(TERRALIB_DATA_DIR "/shape/munic_goias.shp");
  const std::string inDsetName = "munic_goias";
  const std::string outDataSetName( "dataset_copy_selected_to_sqlite_test");

  const std::string outputFileName( outDataSetName + ".sqlite");

  std::remove( outputFileName.c_str() );

  te::da::DataSourcePtr inputDataSourcePtr(te::da::DataSourceFactory::make(
    "OGR", "file://" + inputFileName));
  inputDataSourcePtr->open();
  BOOST_REQUIRE( inputDataSourcePtr->isValid() );

  std::unique_ptr<te::da::Select> selectPtr = te::da::BuildSelect( inDsetName );
  BOOST_REQUIRE( selectPtr.get() != nullptr );

  std::unique_ptr<te::da::DataSet> inputDataSetPtr =
    inputDataSourcePtr->query( *selectPtr );
  BOOST_REQUIRE( inputDataSetPtr.get() != nullptr );
  BOOST_REQUIRE( !inputDataSetPtr->isEmpty() );
  BOOST_REQUIRE( inputDataSetPtr->moveBeforeFirst() );

  std::unique_ptr<te::da::DataSetType> outDataSetTypePtr =
    inputDataSourcePtr->getDataSetType( inDsetName );
  BOOST_REQUIRE( outDataSetTypePtr.get() != nullptr );
  outDataSetTypePtr->setName( outDataSetName );
  outDataSetTypePtr->setTitle( outDataSetName );

  te::da::DataSourcePtr outputDataSourcePtr(te::da::DataSourceFactory::make(
    "OGR", "file://" + outputFileName));
  outputDataSourcePtr->open();
  BOOST_REQUIRE( outputDataSourcePtr->isValid() );

  te::da::DataAccess dataAccess(outputDataSourcePtr, outDataSetName);

  dataAccess.beginTransaction();

  dataAccess.createDataSet(outDataSetTypePtr.get());

  dataAccess.saveDataSet(inputDataSetPtr.release());

  dataAccess.commitTransaction();
}

BOOST_AUTO_TEST_SUITE_END()

