/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/unittest/ogr/dataset/TsDataSet.cpp

  \brief A test suite for the DataSet Class.

 */

// TerraLib
#include "../../Config.h"
#include <terralib/ogr/DataSet.h>
#include <terralib/postgis/Transactor.h>
#include <terralib/dataaccess.h>
#include <terralib/dataaccess/utils/DataAccess.h>
#include <terralib/datatype.h>
#include <terralib/geometry.h>

// STL
#include <iostream>
#include <exception>

// Boost
#include <boost/test/unit_test.hpp>

std::unique_ptr<te::da::DataSet> GetDataSet()
{
  std::string data_dir = TERRALIB_DATA_DIR;

  std::string filename(data_dir + "/shape/munic_goias.shp");

  std::string srcInfo ("file://" + filename);

  te::da::DataSourcePtr dataSource(te::da::DataSourceFactory::make("OGR", srcInfo));
  dataSource->open();

  std::string inDsetName = "munic_goias";
  return dataSource->getDataSet(inDsetName);
}



BOOST_AUTO_TEST_SUITE( dataSet_tests )

BOOST_AUTO_TEST_CASE( moveBeforeFirst_test )
{
  std::unique_ptr<te::da::DataSet> dataSet = GetDataSet();
  dataSet->moveBeforeFirst();
  BOOST_CHECK(dataSet->isBeforeBegin());
}

BOOST_AUTO_TEST_CASE( moveFirst_test )
{
  std::unique_ptr<te::da::DataSet> dataSet = GetDataSet();
  dataSet->moveFirst();
  BOOST_CHECK(dataSet->isAtBegin());
}

BOOST_AUTO_TEST_CASE( moveLast_test )
{
  std::unique_ptr<te::da::DataSet> dataSet = GetDataSet();
  dataSet->moveLast();
  BOOST_CHECK(dataSet->isAtEnd());
}

BOOST_AUTO_TEST_CASE( move_test )
{
  std::unique_ptr<te::da::DataSet> dataSet = GetDataSet();

  if(dataSet->move(8))
  {
    int id = dataSet->getInt32(0);
    std::cout << std::endl
              << "ID = " << boost::lexical_cast<std::string>(id)
              << std::endl;
  }
  else
  {
    std::cout << std::endl
              << "ID not found."
              << std::endl;
  }

  if(dataSet->move(20))
  {
    int id = dataSet->getInt32(0);
    std::cout << std::endl
              << "ID = " << boost::lexical_cast<std::string>(id)
              << std::endl;
  }
  else
  {
    std::cout << std::endl
              << "ID not found."
              << std::endl;
  }

  if(dataSet->move(5000))
  {
    int id = dataSet->getInt32(0);
    std::cout << std::endl
              << "ID = " << boost::lexical_cast<std::string>(id)
              << std::endl;
  }
  else
  {
    std::cout << std::endl
              << "ID not found."
              << std::endl;
  }
}

BOOST_AUTO_TEST_CASE( moveNext_test )
{
  std::unique_ptr<te::da::DataSet> dataSet = GetDataSet();

  dataSet->moveBeforeFirst();

  BOOST_CHECK(dataSet->isBeforeBegin());

  while(dataSet->moveNext())
  {
    BOOST_CHECK(dataSet->isPositionValid());
  }
}

BOOST_AUTO_TEST_CASE( movePrevious_test )
{
  std::unique_ptr<te::da::DataSet> dataSet = GetDataSet();

  dataSet->moveLast();

  BOOST_CHECK(dataSet->isAtEnd());

  do
  {
    BOOST_CHECK(dataSet->isPositionValid());
  }while(dataSet->movePrevious());
}

BOOST_AUTO_TEST_CASE( save_selected_all_shp_test )
{
  const std::string inputFileName(TERRALIB_DATA_DIR "/shape/munic_goias.shp");
  const std::string inDsetName = "munic_goias";  
  const std::string outDataSetName( "save_selected_all_shp_test");
  const std::string outputFileName( outDataSetName + ".shp");

  std::remove( std::string( outDataSetName + ".shp" ).c_str() );
  std::remove( std::string( outDataSetName + ".shx" ).c_str() );
  std::remove( std::string( outDataSetName + ".dbf" ).c_str() );
  std::remove( std::string( outDataSetName + ".prj" ).c_str() );
  std::remove( std::string( outDataSetName + ".qix" ).c_str() );

  te::da::DataSourcePtr inputDataSourcePtr(te::da::DataSourceFactory::make(
    "OGR", "file://" + inputFileName));
  inputDataSourcePtr->open();
  BOOST_REQUIRE( inputDataSourcePtr->isValid() );
  
  std::unique_ptr<te::da::Select> selectPtr = te::da::BuildSelect( inDsetName );
  BOOST_REQUIRE( selectPtr.get() != nullptr );
  
  std::unique_ptr<te::da::DataSet> inputDataSetPtr = 
    inputDataSourcePtr->query( *selectPtr );
  BOOST_REQUIRE( inputDataSetPtr.get() != nullptr );
  BOOST_REQUIRE( !inputDataSetPtr->isEmpty() );
  BOOST_REQUIRE( inputDataSetPtr->moveBeforeFirst() );
  
  std::unique_ptr<te::da::DataSetType> outDataSetTypePtr = 
    inputDataSourcePtr->getDataSetType( inDsetName );
  BOOST_REQUIRE( outDataSetTypePtr.get() != nullptr );
  outDataSetTypePtr->setName( outDataSetName );
  outDataSetTypePtr->setTitle( outDataSetName );
  
  te::da::DataSourcePtr outputDataSourcePtr(te::da::DataSourceFactory::make(
    "OGR", "file://" + outputFileName));
  inputDataSourcePtr->open();
  BOOST_REQUIRE( inputDataSourcePtr->isValid() );  
  
  te::da::DataSourceTransactorPtr outTransPtr = outputDataSourcePtr->getTransactor();
  BOOST_REQUIRE( outTransPtr.get() != nullptr );
  
  outTransPtr->createDataSet(outDataSetTypePtr.get(), 
    std::map<std::string, std::string>() );
  
  outTransPtr->add(outDataSetName, inputDataSetPtr.get(), 
    std::map<std::string, std::string>(), 0, false );
}

BOOST_AUTO_TEST_CASE( save_selected_by_oids_shp_test )
{
  const std::string inputFileName(TERRALIB_DATA_DIR "/shape/munic_goias.shp");
  const std::string inDsetName = "munic_goias";  
  const std::string outDataSetName( "save_selected_by_oids_shp_test");
  const std::string outputFileName( outDataSetName + ".shp");

  std::remove( std::string( outDataSetName + ".shp" ).c_str() );
  std::remove( std::string( outDataSetName + ".shx" ).c_str() );
  std::remove( std::string( outDataSetName + ".dbf" ).c_str() );
  std::remove( std::string( outDataSetName + ".prj" ).c_str() );
  std::remove( std::string( outDataSetName + ".qix" ).c_str() );

  te::da::DataSourcePtr inputDataSourcePtr(te::da::DataSourceFactory::make(
    "OGR", "file://" + inputFileName));
  inputDataSourcePtr->open();
  BOOST_REQUIRE( inputDataSourcePtr->isValid() );
  
  std::unique_ptr<te::da::DataSet> inputDataSetPtr = 
    inputDataSourcePtr->getDataSet( inDsetName );
  BOOST_REQUIRE( inputDataSetPtr.get() != nullptr );
  BOOST_REQUIRE( !inputDataSetPtr->isEmpty() );
  BOOST_REQUIRE( inputDataSetPtr->moveBeforeFirst() );
  
  std::unique_ptr<te::da::DataSetType> inDataSetTypePtr = 
    inputDataSourcePtr->getDataSetType( inDsetName );  
  
  std::unique_ptr< te::da::ObjectIdSet > objIDsSetPtr( te::da::GenerateOIDSet(
    inputDataSetPtr.get(), inDataSetTypePtr.get() ) );
  
  std::unique_ptr<te::da::DataSet> selectedInputDataSetPtr = 
    inputDataSourcePtr->getDataSet( inDsetName, objIDsSetPtr.get() );
  BOOST_REQUIRE( selectedInputDataSetPtr.get() != nullptr );
  BOOST_REQUIRE( !selectedInputDataSetPtr->isEmpty() );
  BOOST_REQUIRE( selectedInputDataSetPtr->moveBeforeFirst() );
  
  std::unique_ptr<te::da::DataSetType> outDataSetTypePtr = 
    inputDataSourcePtr->getDataSetType( inDsetName );
  BOOST_REQUIRE( outDataSetTypePtr.get() != nullptr );
  outDataSetTypePtr->setName( outDataSetName );
  outDataSetTypePtr->setTitle( outDataSetName );
  
  te::da::DataSourcePtr outputDataSourcePtr(te::da::DataSourceFactory::make(
    "OGR", "file://" + outputFileName));
  inputDataSourcePtr->open();
  BOOST_REQUIRE( inputDataSourcePtr->isValid() );  
  
  te::da::DataSourceTransactorPtr outTransPtr = outputDataSourcePtr->getTransactor();
  BOOST_REQUIRE( outTransPtr.get() != nullptr );
  
  outTransPtr->createDataSet(outDataSetTypePtr.get(), 
    std::map<std::string, std::string>() );
  
  outTransPtr->add(outDataSetName, selectedInputDataSetPtr.get(), 
    std::map<std::string, std::string>(), 0, false );
}

BOOST_AUTO_TEST_CASE( save_selected_all_gpkg_test )
{
  const std::string inputFileName(TERRALIB_DATA_DIR "/shape/munic_goias.shp");
  const std::string inDsetName = "munic_goias";
  const std::string outDataSetName( "save_selected_all_gpkg_test");
  const std::string outputFileName( outDataSetName + ".gpkg");

  std::remove( std::string( outDataSetName + ".gpkg" ).c_str() );

  te::da::DataSourcePtr inputDataSourcePtr(te::da::DataSourceFactory::make(
    "OGR", "file://" + inputFileName));
  inputDataSourcePtr->open();
  BOOST_REQUIRE( inputDataSourcePtr->isValid() );

  std::unique_ptr<te::da::Select> selectPtr = te::da::BuildSelect( inDsetName );
  BOOST_REQUIRE( selectPtr.get() != nullptr );

  std::unique_ptr<te::da::DataSet> inputDataSetPtr =
    inputDataSourcePtr->query( *selectPtr );
  BOOST_REQUIRE( inputDataSetPtr.get() != nullptr );
  BOOST_REQUIRE( !inputDataSetPtr->isEmpty() );
  BOOST_REQUIRE( inputDataSetPtr->moveBeforeFirst() );

  std::unique_ptr<te::da::DataSetType> outDataSetTypePtr =
    inputDataSourcePtr->getDataSetType( inDsetName );
  BOOST_REQUIRE( outDataSetTypePtr.get() != nullptr );
  outDataSetTypePtr->setName( outDataSetName );
  outDataSetTypePtr->setTitle( outDataSetName );

  te::da::DataSourcePtr outputDataSourcePtr(te::da::DataSourceFactory::make(
    "OGR", "file://" + outputFileName));
  inputDataSourcePtr->open();
  BOOST_REQUIRE( inputDataSourcePtr->isValid() );

  te::da::DataSourceTransactorPtr outTransPtr = outputDataSourcePtr->getTransactor();
  BOOST_REQUIRE( outTransPtr.get() != nullptr );

  outTransPtr->createDataSet(outDataSetTypePtr.get(),
    std::map<std::string, std::string>() );

  outTransPtr->add(outDataSetName, inputDataSetPtr.get(),
    std::map<std::string, std::string>(), 0, false );
}

BOOST_AUTO_TEST_CASE( save_selected_by_oids_gpkg_test )
{
  const std::string inputFileName(TERRALIB_DATA_DIR "/shape/munic_goias.shp");
  const std::string inDsetName = "munic_goias";
  const std::string outDataSetName( "save_selected_by_oids_gpkg_test");
  const std::string outputFileName( outDataSetName + ".gpkg");

  std::remove( std::string( outDataSetName + ".gpkg" ).c_str() );

  te::da::DataSourcePtr inputDataSourcePtr(te::da::DataSourceFactory::make(
    "OGR", "file://" + inputFileName));
  inputDataSourcePtr->open();
  BOOST_REQUIRE( inputDataSourcePtr->isValid() );

  std::unique_ptr<te::da::DataSet> inputDataSetPtr =
    inputDataSourcePtr->getDataSet( inDsetName );
  BOOST_REQUIRE( inputDataSetPtr.get() != nullptr );
  BOOST_REQUIRE( !inputDataSetPtr->isEmpty() );
  BOOST_REQUIRE( inputDataSetPtr->moveBeforeFirst() );

  std::unique_ptr<te::da::DataSetType> inDataSetTypePtr =
    inputDataSourcePtr->getDataSetType( inDsetName );

  std::unique_ptr< te::da::ObjectIdSet > objIDsSetPtr( te::da::GenerateOIDSet(
    inputDataSetPtr.get(), inDataSetTypePtr.get() ) );

  std::unique_ptr<te::da::DataSet> selectedInputDataSetPtr =
    inputDataSourcePtr->getDataSet( inDsetName, objIDsSetPtr.get() );
  BOOST_REQUIRE( selectedInputDataSetPtr.get() != nullptr );
  BOOST_REQUIRE( !selectedInputDataSetPtr->isEmpty() );
  BOOST_REQUIRE( selectedInputDataSetPtr->moveBeforeFirst() );

  std::unique_ptr<te::da::DataSetType> outDataSetTypePtr =
    inputDataSourcePtr->getDataSetType( inDsetName );
  BOOST_REQUIRE( outDataSetTypePtr.get() != nullptr );
  outDataSetTypePtr->setName( outDataSetName );
  outDataSetTypePtr->setTitle( outDataSetName );

  te::da::DataSourcePtr outputDataSourcePtr(te::da::DataSourceFactory::make(
    "OGR", "file://" + outputFileName));
  inputDataSourcePtr->open();
  BOOST_REQUIRE( inputDataSourcePtr->isValid() );

  te::da::DataSourceTransactorPtr outTransPtr = outputDataSourcePtr->getTransactor();
  BOOST_REQUIRE( outTransPtr.get() != nullptr );

  outTransPtr->createDataSet(outDataSetTypePtr.get(),
    std::map<std::string, std::string>() );

  outTransPtr->add(outDataSetName, selectedInputDataSetPtr.get(),
    std::map<std::string, std::string>(), 0, false );
}

BOOST_AUTO_TEST_CASE( save_selected_all_reprojection_gpkg_test )
{
  const std::string inputFileName(TERRALIB_DATA_DIR "/shape/munic_goias.shp");
  const std::string inDsetName = "munic_goias";
  const std::string outDataSetName( "save_selected_all_reprojection_gpkg_test");
  const std::string outputFileName( outDataSetName + ".gpkg");

  std::remove( std::string( outDataSetName + ".gpkg" ).c_str() );

  te::da::DataSourcePtr inputDataSourcePtr(te::da::DataSourceFactory::make(
    "OGR", "file://" + inputFileName));
  inputDataSourcePtr->open();
  BOOST_REQUIRE( inputDataSourcePtr->isValid() );

  std::unique_ptr<te::da::DataSet> inputDataSetPtr = inputDataSourcePtr->getDataSet(
    inDsetName, te::common::FORWARDONLY, te::common::RAccess );
  BOOST_REQUIRE( inputDataSetPtr.get() != nullptr );
  BOOST_REQUIRE( !inputDataSetPtr->isEmpty() );
  BOOST_REQUIRE( inputDataSetPtr->moveBeforeFirst() );

  std::unique_ptr<te::da::DataSetType> inDataSetTypePtr =
    inputDataSourcePtr->getDataSetType( inDsetName );

  te::da::DataSourcePtr outputDataSourcePtr(te::da::DataSourceFactory::make(
    "OGR", "file://" + outputFileName));
  outputDataSourcePtr->open();
  BOOST_REQUIRE( outputDataSourcePtr->isValid() );

  std::unique_ptr<te::da::DataSetTypeConverter> converter(new
    te::da::DataSetTypeConverter(inDataSetTypePtr.get(),
    outputDataSourcePtr->getCapabilities(), outputDataSourcePtr->getEncoding()));

  te::da::AssociateDataSetTypeConverterSRID(converter.get(), 4618, 950022);

  te::da::DataSetType* outDataSetTypePtr = converter->getResult();
  outDataSetTypePtr->setName(outDataSetName);

  te::da::DataSourceTransactorPtr outTransPtr = outputDataSourcePtr->getTransactor();
  BOOST_REQUIRE( outTransPtr.get() != nullptr );

  outTransPtr->createDataSet(outDataSetTypePtr,
    std::map<std::string, std::string>() );
  
  std::unique_ptr<te::da::DataSetAdapter> inputDataSetAdapterPtr(
    te::da::CreateAdapter(inputDataSetPtr.get(), converter.get()));
  
  BOOST_REQUIRE( inputDataSetPtr->moveBeforeFirst() );    

  outTransPtr->add(outDataSetName, inputDataSetAdapterPtr.get(),
    std::map<std::string, std::string>(), 0, false );
}

BOOST_AUTO_TEST_CASE( save_selected_all_reprojection_sqlite_test )
{
  const std::string inputFileName(TERRALIB_DATA_DIR "/shape/munic_goias.shp");
  const std::string inDsetName = "munic_goias";
  const std::string outDataSetName( "save_selected_all_reprojection_gpkg_test");
  const std::string outputFileName( outDataSetName + ".sqlite");

  std::remove( std::string( outDataSetName + ".sqlite" ).c_str() );

  te::da::DataSourcePtr inputDataSourcePtr(te::da::DataSourceFactory::make(
    "OGR", "file://" + inputFileName));
  inputDataSourcePtr->open();
  BOOST_REQUIRE( inputDataSourcePtr->isValid() );

  std::unique_ptr<te::da::DataSet> inputDataSetPtr = inputDataSourcePtr->getDataSet(
    inDsetName, te::common::FORWARDONLY, te::common::RAccess );
  BOOST_REQUIRE( inputDataSetPtr.get() != nullptr );
  BOOST_REQUIRE( !inputDataSetPtr->isEmpty() );
  BOOST_REQUIRE( inputDataSetPtr->moveBeforeFirst() );

  std::unique_ptr<te::da::DataSetType> inDataSetTypePtr =
    inputDataSourcePtr->getDataSetType( inDsetName );

  te::da::DataSourcePtr outputDataSourcePtr(te::da::DataSourceFactory::make(
    "OGR", "file://" + outputFileName));
  outputDataSourcePtr->open();
  BOOST_REQUIRE( outputDataSourcePtr->isValid() );

  std::unique_ptr<te::da::DataSetTypeConverter> converter(new
    te::da::DataSetTypeConverter(inDataSetTypePtr.get(),
    outputDataSourcePtr->getCapabilities(), outputDataSourcePtr->getEncoding()));

  te::da::AssociateDataSetTypeConverterSRID(converter.get(), 4618, 950022);

  te::da::DataSetType* outDataSetTypePtr = converter->getResult();
  outDataSetTypePtr->setName(outDataSetName);

  te::da::DataSourceTransactorPtr outTransPtr = outputDataSourcePtr->getTransactor();
  BOOST_REQUIRE( outTransPtr.get() != nullptr );

  outTransPtr->createDataSet(outDataSetTypePtr,
    std::map<std::string, std::string>() );

  std::unique_ptr<te::da::DataSetAdapter> inputDataSetAdapterPtr(
    te::da::CreateAdapter(inputDataSetPtr.get(), converter.get()));

  BOOST_REQUIRE( inputDataSetPtr->moveBeforeFirst() );

  outTransPtr->add(outDataSetName, inputDataSetAdapterPtr.get(),
    std::map<std::string, std::string>(), 0, false );
}

BOOST_AUTO_TEST_CASE( save_selected_all_reprojection_shp_test )
{
  const std::string inputFileName(TERRALIB_DATA_DIR "/shape/munic_goias.shp");
  const std::string inDsetName = "munic_goias";
  const std::string outDataSetName( "save_selected_all_reprojection_shp_test");
  const std::string outputFileName( outDataSetName + ".shp");

  std::remove( std::string( outDataSetName + ".shp" ).c_str() );
  std::remove( std::string( outDataSetName + ".shx" ).c_str() );
  std::remove( std::string( outDataSetName + ".dbf" ).c_str() );
  std::remove( std::string( outDataSetName + ".prj" ).c_str() );
  std::remove( std::string( outDataSetName + ".qix" ).c_str() );

  te::da::DataSourcePtr inputDataSourcePtr(te::da::DataSourceFactory::make(
    "OGR", "file://" + inputFileName));
  inputDataSourcePtr->open();
  BOOST_REQUIRE( inputDataSourcePtr->isValid() );

  std::unique_ptr<te::da::DataSet> inputDataSetPtr = inputDataSourcePtr->getDataSet(
    inDsetName, te::common::FORWARDONLY, te::common::RAccess );
  BOOST_REQUIRE( inputDataSetPtr.get() != nullptr );
  BOOST_REQUIRE( !inputDataSetPtr->isEmpty() );
  BOOST_REQUIRE( inputDataSetPtr->moveBeforeFirst() );

  std::unique_ptr<te::da::DataSetType> inDataSetTypePtr =
    inputDataSourcePtr->getDataSetType( inDsetName );

  te::da::DataSourcePtr outputDataSourcePtr(te::da::DataSourceFactory::make(
    "OGR", "file://" + outputFileName));
  outputDataSourcePtr->open();
  BOOST_REQUIRE( outputDataSourcePtr->isValid() );

  std::unique_ptr<te::da::DataSetTypeConverter> converter(new
    te::da::DataSetTypeConverter(inDataSetTypePtr.get(),
    outputDataSourcePtr->getCapabilities(), outputDataSourcePtr->getEncoding()));

  te::da::AssociateDataSetTypeConverterSRID(converter.get(), 4618, 950022);

  te::da::DataSetType* outDataSetTypePtr = converter->getResult();
  outDataSetTypePtr->setName(outDataSetName);

  te::da::DataSourceTransactorPtr outTransPtr = outputDataSourcePtr->getTransactor();
  BOOST_REQUIRE( outTransPtr.get() != nullptr );

  outTransPtr->createDataSet(outDataSetTypePtr,
    std::map<std::string, std::string>() );
  
  std::unique_ptr<te::da::DataSetAdapter> inputDataSetAdapterPtr(
    te::da::CreateAdapter(inputDataSetPtr.get(), converter.get()));
  
  BOOST_REQUIRE( inputDataSetPtr->moveBeforeFirst() );  

  outTransPtr->add(outDataSetName, inputDataSetAdapterPtr.get(),
    std::map<std::string, std::string>(), 0, false );
}

BOOST_AUTO_TEST_SUITE_END()

