/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */


// TerraLib
#include "../../Config.h"
#include <terralib/ogr/DataSet.h>
#include <terralib/dataaccess.h>
#include <terralib/dataaccess/utils/DataAccess.h>
#include <terralib/datatype.h>
#include <terralib/geometry.h>

// STL
#include <iostream>
#include <exception>
#include <cstdio>

// Boost
#include <boost/test/unit_test.hpp>

BOOST_AUTO_TEST_SUITE( transactor_tests )

BOOST_AUTO_TEST_CASE( query_string_datetime_shp_test )
{
  const std::string outDataSetName( "query_string_datetime_shp_test");

  std::remove( ( outDataSetName + ".shx" ).c_str() );
  std::remove( ( outDataSetName + ".shp" ).c_str() );
  std::remove( ( outDataSetName + ".dbf" ).c_str() );
  std::remove( ( outDataSetName + ".prj" ).c_str() );

  const std::string inputFileName(TERRALIB_DATA_DIR "/shape/determaio.shp");
  const std::string inDsetName = "determaio";

  const std::string outputFileName( outDataSetName + ".shp");

  te::da::DataSourcePtr inputDataSourcePtr(te::da::DataSourceFactory::make(
    "OGR", "file://" + inputFileName));
  inputDataSourcePtr->open();
  BOOST_REQUIRE( inputDataSourcePtr->isValid() );

  std::unique_ptr<te::da::DataSourceTransactor> transactorPtr =
    inputDataSourcePtr->getTransactor();

  std::unique_ptr<te::da::DataSet> inputDataSetPtr =
    transactorPtr->query(
    "SELECT * FROM \"determaio\" WHERE ( \"VIEW_DATE\" >="
    " cast ( \'2022-05-01\' as date ) )",
    te::common::FORWARDONLY, false, te::common::RAccess );
  BOOST_REQUIRE( inputDataSetPtr.get() != nullptr );
  BOOST_REQUIRE( !inputDataSetPtr->isEmpty() );
  BOOST_REQUIRE( inputDataSetPtr->moveBeforeFirst() );

  std::unique_ptr<te::da::DataSetType> outDataSetTypePtr =
    inputDataSourcePtr->getDataSetType( inDsetName );
  BOOST_REQUIRE( outDataSetTypePtr.get() != nullptr );
  outDataSetTypePtr->setName( outDataSetName );
  outDataSetTypePtr->setTitle( outDataSetName );

  te::da::DataSourcePtr outputDataSourcePtr(te::da::DataSourceFactory::make(
    "OGR", "file://" + outputFileName));
  outputDataSourcePtr->open();
  BOOST_REQUIRE( outputDataSourcePtr->isValid() );

  te::da::DataAccess dataAccess(outputDataSourcePtr, outDataSetName);

  dataAccess.beginTransaction();

  dataAccess.createDataSet(outDataSetTypePtr.get());

  dataAccess.saveDataSet(inputDataSetPtr.release());

  dataAccess.commitTransaction();
}

BOOST_AUTO_TEST_CASE( query_datetime_shp_test )
{
  const std::string outDataSetName( "query_datetime_shp_test");

  std::remove( ( outDataSetName + ".shx" ).c_str() );
  std::remove( ( outDataSetName + ".shp" ).c_str() );
  std::remove( ( outDataSetName + ".dbf" ).c_str() );
  std::remove( ( outDataSetName + ".prj" ).c_str() );

  const std::string inputFileName(TERRALIB_DATA_DIR "/shape/determaio.shp");
  const std::string inDsetName = "determaio";

  const std::string outputFileName( outDataSetName + ".shp");

  te::da::DataSourcePtr inputDataSourcePtr(te::da::DataSourceFactory::make(
    "OGR", "file://" + inputFileName));
  inputDataSourcePtr->open();
  BOOST_REQUIRE( inputDataSourcePtr->isValid() );

  std::unique_ptr<te::da::DataSourceTransactor> transactorPtr =
    inputDataSourcePtr->getTransactor();

  boost::gregorian::date bDate = boost::gregorian::from_string( "2022-05-01" );

  te::dt::Date tDate( bDate );

  te::da::LiteralDateTime l( tDate );

  te::da::Fields fields;
  fields.push_back( new te::da::Field( "*" ) );
  te::da::From from;
  from.push_back( new te::da::DataSetName("determaio", "t1") );
  te::da::Where where( new te::da::EqualTo( te::da::PropertyName( "VIEW_DATE" ),
    l ) );
  te::da::Select select(fields, from, where );

  std::unique_ptr<te::da::DataSet> inputDataSetPtr = transactorPtr->query( select );
  BOOST_REQUIRE( inputDataSetPtr.get() != nullptr );
  BOOST_REQUIRE( !inputDataSetPtr->isEmpty() );
  BOOST_REQUIRE( inputDataSetPtr->moveBeforeFirst() );

  std::unique_ptr<te::da::DataSetType> outDataSetTypePtr =
    inputDataSourcePtr->getDataSetType( inDsetName );
  BOOST_REQUIRE( outDataSetTypePtr.get() != nullptr );
  outDataSetTypePtr->setName( outDataSetName );
  outDataSetTypePtr->setTitle( outDataSetName );

  te::da::DataSourcePtr outputDataSourcePtr(te::da::DataSourceFactory::make(
    "OGR", "file://" + outputFileName));
  outputDataSourcePtr->open();
  BOOST_REQUIRE( outputDataSourcePtr->isValid() );

  te::da::DataAccess dataAccess(outputDataSourcePtr, outDataSetName);

  dataAccess.beginTransaction();

  dataAccess.createDataSet(outDataSetTypePtr.get());

  dataAccess.saveDataSet(inputDataSetPtr.release());

  dataAccess.commitTransaction();
}

BOOST_AUTO_TEST_SUITE_END()

