/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/unittest/dataaccess/TsConnectedDataSet.cpp

  \brief A test suite for the ConnectedDataSet Class.

 */

// TerraLib
//#include <terralib/dataaccess/dataset/DataSet.h>
//#include <terralib/dataaccess/datasource/DataSource.h>
//#include <terralib/dataaccess/datasource/DataSourceFactory.h>
//#include <terralib/dataaccess/utils/Utils.h>
#include <terralib/dataaccess.h>
#include <terralib/datatype.h>
#include <terralib/geometry.h>
//#include <terralib/postgis/Transactor.h>
#include "../Utils.h"

// STL
#include <iostream>
#include <exception>

// Boost
#include <boost/test/unit_test.hpp>

std::unique_ptr<te::da::DataSet> GetConnectedDataSet(te::da::DataSource* ds,
   const te::gm::Envelope* e = nullptr)
{
  std::unique_ptr<te::da::DataSet> dataSet;

  try
  {
    if (!ds)
      return dataSet;

    ds->open();

    std::string name = "munic_goias";

    std::unique_ptr<te::da::DataSourceTransactor> transactor = ds->getTransactor();
    const te::da::DataSourceCapabilities& cap = ds->getCapabilities();
    const te::da::DataSetCapabilities& dsCap = cap.getDataSetCapabilities();
    const bool isConnected = dsCap.isConnected();

    if( e )
    {
      dataSet = transactor->getDataSet(name, "ogr_geometry", e, te::gm::INTERSECTS, 
        te::common::RANDOM, isConnected, te::common::RAccess );      
    }
    else
    {
      dataSet = transactor->getDataSet(name, te::common::RANDOM, isConnected);
    }      
  }
  catch(const std::exception& e)
  {
    std::cout << std::endl
              << "An exception has occurred in the PostGIS Example: "
              << e.what() << std::endl;
  }
  catch(...)
  {
    std::cout << std::endl
              << "An unexpected exception has occurred in the PostGIS Example!"
              << std::endl;
  }

  return dataSet;
}



BOOST_AUTO_TEST_SUITE( connectedDataSet_tests )

BOOST_AUTO_TEST_CASE( moveBeforeFirst_test )
{
  std::unique_ptr<te::da::DataSource> ds = GetPostGISConnection();
  std::unique_ptr<te::da::DataSet> dataSet = GetConnectedDataSet(ds.get());
  dataSet->moveBeforeFirst();
  BOOST_CHECK(dataSet->isBeforeBegin());
}

BOOST_AUTO_TEST_CASE( moveFirst_test )
{
  std::unique_ptr<te::da::DataSource> ds = GetPostGISConnection();
  std::unique_ptr<te::da::DataSet> dataSet = GetConnectedDataSet(ds.get());
  dataSet->moveFirst();

  BOOST_CHECK(dataSet->isAtBegin());

  int id = dataSet->getInt32(0);
  std::cout << std::endl
            << "First ID = " << boost::lexical_cast<std::string>(id)
            << std::endl;
}

BOOST_AUTO_TEST_CASE( moveLast_test )
{
  std::unique_ptr<te::da::DataSource> ds = GetPostGISConnection();
  std::unique_ptr<te::da::DataSet> dataSet = GetConnectedDataSet(ds.get());
  dataSet->moveLast();

  BOOST_CHECK(dataSet->isAtEnd());

  int id = dataSet->getInt32(0);
  std::cout << std::endl
            << "Last ID = " << boost::lexical_cast<std::string>(id)
            << std::endl;
}

BOOST_AUTO_TEST_CASE( move_test )
{
  std::unique_ptr<te::da::DataSource> ds = GetPostGISConnection();
  std::unique_ptr<te::da::DataSet> dataSet = GetConnectedDataSet(ds.get());
  if(dataSet->move(8))
  {
    int id = dataSet->getInt32(0);
    std::cout << std::endl
              << "ID = " << boost::lexical_cast<std::string>(id)
              << std::endl;
  }
  else
  {
    std::cout << std::endl
              << "ID not found."
              << std::endl;
  }

  if(dataSet->move(20))
  {
    int id = dataSet->getInt32(0);
    std::cout << std::endl
              << "ID = " << boost::lexical_cast<std::string>(id)
              << std::endl;
  }
  else
  {
    std::cout << std::endl
              << "ID not found."
              << std::endl;
  }

  if(dataSet->move(5000))
  {
    int id = dataSet->getInt32(0);
    std::cout << std::endl
              << "ID = " << boost::lexical_cast<std::string>(id)
              << std::endl;
  }
  else
  {
    std::cout << std::endl
              << "ID not found."
              << std::endl;
  }
}

BOOST_AUTO_TEST_CASE( moveNext_test )
{
  std::unique_ptr<te::da::DataSource> ds = GetPostGISConnection();
  std::unique_ptr<te::da::DataSet> dataSet = GetConnectedDataSet(ds.get());
  dataSet->moveBeforeFirst();

  BOOST_CHECK(dataSet->isBeforeBegin());

  while(dataSet->moveNext())
  {
    int id = dataSet->getInt32(0);
    std::cout << std::endl
              << "Current ID = " << boost::lexical_cast<std::string>(id)
              << std::endl;
  }
}

BOOST_AUTO_TEST_CASE( movePrevious_test )
{
  std::unique_ptr<te::da::DataSource> ds = GetPostGISConnection();
  std::unique_ptr<te::da::DataSet> dataSet = GetConnectedDataSet(ds.get());
  dataSet->moveLast();

  BOOST_CHECK(dataSet->isAtEnd());

  do
  {
    int id = dataSet->getInt32(0);
    std::cout << std::endl
              << "Current ID = " << boost::lexical_cast<std::string>(id)
              << std::endl;
  }while(dataSet->movePrevious());
}

BOOST_AUTO_TEST_CASE( restriction_move_test )
{
  {
    te::da::DataSourcePtr ds = GetPostGISConnectionFromDSManager();
    std::unique_ptr<te::da::DataSet> dataSet = GetConnectedDataSet(ds.get(), nullptr);

    dataSet->moveBeforeFirst();

    BOOST_CHECK(dataSet->isBeforeBegin());

    while(dataSet->moveNext())
    {
      int id = dataSet->getInt32(0);
      std::cout << std::endl
                << "Current ID = " << boost::lexical_cast<std::string>(id)
                << std::endl;
    }
  }

  {
    te::gm::Envelope e( -49.914891, -16.455826, -49.160085, -16.058560 );
    te::da::DataSourcePtr ds = GetPostGISConnectionFromDSManager();
    std::unique_ptr<te::da::DataSet> dataSet = GetConnectedDataSet(ds.get(), &e);

    dataSet->moveBeforeFirst();

    BOOST_CHECK(dataSet->isBeforeBegin());

    while(dataSet->moveNext())
    {
      int id = dataSet->getInt32(0);
      std::cout << std::endl
                << "Current ID = " << boost::lexical_cast<std::string>(id)
                << std::endl;
    }
  }
}

BOOST_AUTO_TEST_SUITE_END()

