/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/unittest/dataaccess/TsDataSet.cpp

  \brief A test suite for the DataSet Class.

 */

// TerraLib
#include <terralib/postgis/DataSet.h>
#include <terralib/dataaccess/datasource/DataSource.h>
#include <terralib/dataaccess/datasource/DataSourceFactory.h>
#include <terralib/dataaccess/utils/Utils.h>
#include <terralib/datatype.h>
#include <terralib/geometry/GeometryProperty.h>
#include <terralib/postgis/Transactor.h>
#include "../Utils.h"

// STL
#include <iostream>
#include <exception>

// Boost
#include <boost/test/unit_test.hpp>

std::unique_ptr<te::da::DataSet> GetDataSet(te::da::DataSource* ds)
{
  std::unique_ptr<te::da::DataSet> dataSet;

  try
  {
    if (!ds)
      return dataSet;

    ds->open();

    std::string name = "munic_goias";

    std::unique_ptr<te::da::DataSourceTransactor> transactor = ds->getTransactor();

    dataSet = transactor->getDataSet(name, te::common::RANDOM, false);
  }
  catch(const std::exception& e)
  {
    std::cout << std::endl
              << "An exception has occurred in the PostGIS Example: "
              << e.what() << std::endl;
  }
  catch(...)
  {
    std::cout << std::endl
              << "An unexpected exception has occurred in the PostGIS Example!"
              << std::endl;
  }

  return dataSet;
}



BOOST_AUTO_TEST_SUITE( dataSet_tests )

BOOST_AUTO_TEST_CASE( moveBeforeFirst_test )
{
  std::unique_ptr<te::da::DataSource> ds = GetPostGISConnection();
  std::unique_ptr<te::da::DataSet> dataSet = GetDataSet(ds.get());
  dataSet->moveBeforeFirst();
  BOOST_CHECK(dataSet->isBeforeBegin());
}

BOOST_AUTO_TEST_CASE( moveFirst_test )
{
  std::unique_ptr<te::da::DataSource> ds = GetPostGISConnection();
  std::unique_ptr<te::da::DataSet> dataSet = GetDataSet(ds.get());
  dataSet->moveFirst();
  BOOST_CHECK(dataSet->isAtBegin());
}

BOOST_AUTO_TEST_CASE( moveLast_test )
{
  std::unique_ptr<te::da::DataSource> ds = GetPostGISConnection();
  std::unique_ptr<te::da::DataSet> dataSet = GetDataSet(ds.get());
  dataSet->moveLast();
  BOOST_CHECK(dataSet->isAtEnd());
}

BOOST_AUTO_TEST_CASE( move_test )
{
  std::unique_ptr<te::da::DataSource> ds = GetPostGISConnection();
  std::unique_ptr<te::da::DataSet> dataSet = GetDataSet(ds.get());
  if(dataSet->move(8))
  {
    int id = dataSet->getInt32(0);
    std::cout << std::endl
              << "ID = " << boost::lexical_cast<std::string>(id)
              << std::endl;
  }
  else
  {
    std::cout << std::endl
              << "ID not found."
              << std::endl;
  }

  if(dataSet->move(20))
  {
    int id = dataSet->getInt32(0);
    std::cout << std::endl
              << "ID = " << boost::lexical_cast<std::string>(id)
              << std::endl;
  }
  else
  {
    std::cout << std::endl
              << "ID not found."
              << std::endl;
  }

  if(dataSet->move(5000))
  {
    int id = dataSet->getInt32(0);
    std::cout << std::endl
              << "ID = " << boost::lexical_cast<std::string>(id)
              << std::endl;
  }
  else
  {
    std::cout << std::endl
              << "ID not found."
              << std::endl;
  }
}

BOOST_AUTO_TEST_CASE( moveNext_test )
{
  std::unique_ptr<te::da::DataSource> ds = GetPostGISConnection();
  std::unique_ptr<te::da::DataSet> dataSet = GetDataSet(ds.get());
  dataSet->moveBeforeFirst();

  BOOST_CHECK(dataSet->isBeforeBegin());

  while(dataSet->moveNext())
  {
    BOOST_CHECK(dataSet->isPositionValid());
  }
}

BOOST_AUTO_TEST_CASE( movePrevious_test )
{
  std::unique_ptr<te::da::DataSource> ds = GetPostGISConnection();
  std::unique_ptr<te::da::DataSet> dataSet = GetDataSet(ds.get());
  dataSet->moveLast();

  BOOST_CHECK(dataSet->isAtEnd());

  do
  {
    BOOST_CHECK(dataSet->isPositionValid());
  }while(dataSet->movePrevious());
}

BOOST_AUTO_TEST_SUITE_END()

