/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/unittest/raster/TsReprojection.cpp
 
  \brief A test suit for the Raster Reprojection class.
 */

// TerraLib
#include <terralib_buildconfig.h>

#include <terralib/raster.h>
#include "../Config.h"

// Boost
#include <boost/test/unit_test.hpp>
#include <boost/shared_ptr.hpp>

BOOST_AUTO_TEST_SUITE ( reprojection_tests )

BOOST_AUTO_TEST_CASE (reprojection1_test)
{
  /* Openning input raster */

  std::map<std::string, std::string> auxRasterInfo;

  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers_rgb342_crop1.tif";
  boost::shared_ptr< te::rst::Raster > inputRasterPtr ( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_CHECK( inputRasterPtr.get() );

  /* Reprojecting */
  
  auxRasterInfo["URI"] = "TsReprojection_tcReprojection1.tif";
  boost::shared_ptr< te::rst::Raster > outputRasterPtr(
    inputRasterPtr->transform( 32621, auxRasterInfo, 1 ) );
  BOOST_CHECK( outputRasterPtr.get() );
}

BOOST_AUTO_TEST_CASE (reprojection_with_custom_roi_test)
{
  /* Openning input raster */

  std::map<std::string, std::string> auxRasterInfo;

  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers_rgb342_crop1.tif";
  boost::shared_ptr< te::rst::Raster > inputRasterPtr ( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_CHECK( inputRasterPtr.get() );

  // Defining the ROI

  te::gm::Envelope envelope;
  {
    const double xOff = ( ((double)inputRasterPtr->getNumberOfColumns()) / 2 )
      * inputRasterPtr->getResolutionX();
    const double yOff = ( ((double)inputRasterPtr->getNumberOfRows()) / 2 )
      * inputRasterPtr->getResolutionY();

    envelope = *inputRasterPtr->getExtent();
    envelope.m_llx -= xOff;
    envelope.m_urx -= xOff;
    envelope.m_lly -= yOff;
    envelope.m_ury -= yOff;
  }

  /* Reprojecting */

  auxRasterInfo["URI"] = "TsReprojection_reprojection_with_custom_roi_test.tif";
  boost::shared_ptr< te::rst::Raster > outputRasterPtr(
    inputRasterPtr->transform( 32621, envelope.m_llx, envelope.m_lly,
    envelope.m_urx, envelope.m_ury, auxRasterInfo, 1 ) );
  BOOST_CHECK( outputRasterPtr.get() );
}

BOOST_AUTO_TEST_CASE (reprojection_with_clip_and_ressample_same_srid_test)
{
  /* Openning input raster */

  std::map<std::string, std::string> auxRasterInfo;

  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers_rgb342_crop3_EPSG_22522.tif";
  boost::shared_ptr< te::rst::Raster > inputRasterPtr ( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_REQUIRE( inputRasterPtr.get() );
  
  auxRasterInfo["URI"] = "TsReprojection_reprojection_with_clip_and_ressample_same_srid_test.tif";
  
  std::vector< unsigned int > inputRasterBands;
  inputRasterBands.push_back( 1 );
  
  std::unique_ptr< te::rst::Raster > outputRasterPtr;
  BOOST_REQUIRE( te::rst::Reproject( 
    inputRasterPtr->getSRID(), 
    *inputRasterPtr,
    inputRasterBands,
    te::rst::NearestNeighbor,
    0, //interpWindowRadius
    0, //firstRow
    0, //firstColumn
    inputRasterPtr->getNumberOfRows(), //height
    inputRasterPtr->getNumberOfColumns(), //width
    inputRasterPtr->getNumberOfRows(), //newheight
    inputRasterPtr->getNumberOfColumns(), //newwidth
    1, //maxThreads
    auxRasterInfo,
    "GDAL",
    outputRasterPtr ) );
}

BOOST_AUTO_TEST_CASE (reprojection_with_clip_and_ressample_test)
{
  /* Openning input raster */

  std::map<std::string, std::string> auxRasterInfo;

  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers_rgb342_crop1.tif";
  boost::shared_ptr< te::rst::Raster > inputRasterPtr ( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_REQUIRE( inputRasterPtr.get() );
  
  auxRasterInfo["URI"] = "TsReprojection_reprojection_with_clip_and_ressample_test.tif";
  
  std::vector< unsigned int > inputRasterBands;
  inputRasterBands.push_back( 0 );
  inputRasterBands.push_back( 1 );
  inputRasterBands.push_back( 2 );
  
  std::unique_ptr< te::rst::Raster > outputRasterPtr;
  BOOST_REQUIRE( te::rst::Reproject( 
    4326, 
    *inputRasterPtr,
    inputRasterBands,
    te::rst::NearestNeighbor,
    0, //interpWindowRadius
    100, //firstRow
    100, //firstColumn
    100, //height
    100, //width
    200, //newheight
    200, //newwidth
    1, //maxThreads
    auxRasterInfo,
    "GDAL",
    outputRasterPtr ) );
}

BOOST_AUTO_TEST_CASE (reprojection_with_clip_and_ressample_with_threads_test)
{
  /* Openning input raster */

  std::map<std::string, std::string> auxRasterInfo;

  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers_rgb342_crop1.tif";
  boost::shared_ptr< te::rst::Raster > inputRasterPtr ( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_REQUIRE( inputRasterPtr.get() );
  
  auxRasterInfo["URI"] = "TsReprojection_reprojection_with_clip_and_ressample_with_threads_test.tif";
  
  std::vector< unsigned int > inputRasterBands;
  inputRasterBands.push_back( 0 );
  inputRasterBands.push_back( 1 );
  inputRasterBands.push_back( 2 );
  
  std::unique_ptr< te::rst::Raster > outputRasterPtr;
  BOOST_REQUIRE( te::rst::Reproject( 
    4326, 
    *inputRasterPtr,
    inputRasterBands,
    te::rst::NearestNeighbor,
    0, //interpWindowRadius
    100, //firstRow
    100, //firstColumn
    100, //height
    100, //width
    200, //newheight
    200, //newwidth
    4, //maxThreads
    auxRasterInfo,
    "GDAL",
    outputRasterPtr ) );
}

BOOST_AUTO_TEST_SUITE_END()
