/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/unittest/raster/TsRasterSummaryManager.cpp
 
  \brief A test suit for the Raster Summary Manager class.
 */

// TerraLib
#include <terralib_buildconfig.h>

#include <terralib/raster.h>
#include "../Config.h"

// Boost
#include <boost/test/unit_test.hpp>
#include <boost/shared_ptr.hpp>

BOOST_AUTO_TEST_SUITE ( rasterSummaryManager_tests )


BOOST_AUTO_TEST_CASE (get_test)
{
  // create summary
  
  {
    std::map<std::string, std::string> auxRasterInfo;

    auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers_rgb342_crop1.tif";
    boost::shared_ptr< te::rst::Raster > inputRasterPtr ( te::rst::RasterFactory::open(
      auxRasterInfo ) );
    BOOST_REQUIRE( inputRasterPtr.get() );  
  
    te::rst::RasterSummaryManager::getInstance().removeAll();
    te::rst::RasterSummaryManager::getInstance().enableDeletedRastersCache( true );    
    te::rst::RasterSummaryManager::getInstance().setDefaultRasterHistSize( 256 );
    
    te::rst::RasterSummary summary;
    te::rst::RasterSummaryManager::getInstance().get( inputRasterPtr.get(), true, summary );    
    
    BOOST_REQUIRE( summary.size() == 3 );
    BOOST_REQUIRE_CLOSE( summary.at( 0 ).m_minVal->real(), 24, 0 );
    BOOST_REQUIRE_CLOSE( summary.at( 0 ).m_maxVal->real(), 184, 0 );
    BOOST_REQUIRE_CLOSE( summary.at( 0 ).m_stdVal->real(), 7.0794198247739137, 0.1 );
    BOOST_REQUIRE_CLOSE( summary.at( 0 ).m_meanVal->real(), 42.587113202466334, 0.1 );
    
    BOOST_REQUIRE_CLOSE( summary.at( 1 ).m_minVal->real(), 34, 0 );
    BOOST_REQUIRE_CLOSE( summary.at( 1 ).m_maxVal->real(), 255, 0 );
    BOOST_REQUIRE_CLOSE( summary.at( 1 ).m_stdVal->real(), 38.286828510115711, 0.1 );
    BOOST_REQUIRE_CLOSE( summary.at( 1 ).m_meanVal->real(), 154.93450714993628, 0.1 );
    
    BOOST_REQUIRE_CLOSE( summary.at( 2 ).m_minVal->real(), 42, 0 );
    BOOST_REQUIRE_CLOSE( summary.at( 2 ).m_maxVal->real(), 255, 0 );
    BOOST_REQUIRE_CLOSE( summary.at( 2 ).m_stdVal->real(), 11.17826961209258, 0.1 );
    BOOST_REQUIRE_CLOSE( summary.at( 2 ).m_meanVal->real(), 77.114384335934588, 0.1 );    
  }

  // find test
  
  {
    std::map<std::string, std::string> auxRasterInfo;

    auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers_rgb342_crop1.tif";
    boost::shared_ptr< te::rst::Raster > inputRasterPtr ( te::rst::RasterFactory::open(
      auxRasterInfo ) );
    BOOST_REQUIRE( inputRasterPtr.get() );  
    
    te::rst::RasterSummary summary;
    BOOST_REQUIRE( te::rst::RasterSummaryManager::getInstance().find( 
      inputRasterPtr.get(), summary ) );
    
    BOOST_REQUIRE( summary.size() == inputRasterPtr->getNumberOfBands() );
  }  
  
  // using cache
  
  {
    std::map<std::string, std::string> auxRasterInfo;

    auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers_rgb342_crop1.tif";
    boost::shared_ptr< te::rst::Raster > inputRasterPtr ( te::rst::RasterFactory::open(
      auxRasterInfo ) );
    BOOST_REQUIRE( inputRasterPtr.get() );  
  
    te::rst::RasterSummary summary;
    te::rst::RasterSummaryManager::getInstance().get( inputRasterPtr.get(), false, summary );    
    BOOST_REQUIRE( summary.size() == inputRasterPtr->getNumberOfBands() );
  }
}

BOOST_AUTO_TEST_CASE (get_double_pixel_type_test)
{
  // create summary
  
  {
    std::map<std::string, std::string> auxRasterInfo;

    auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers2b_rgb342_crop_linearMixtureModel_double.tif";
    boost::shared_ptr< te::rst::Raster > inputRasterPtr ( te::rst::RasterFactory::open(
      auxRasterInfo ) );
    BOOST_REQUIRE( inputRasterPtr.get() );  
  
    te::rst::RasterSummaryManager::getInstance().removeAll();
    te::rst::RasterSummaryManager::getInstance().enableDeletedRastersCache( true );    
    te::rst::RasterSummaryManager::getInstance().setDefaultRasterHistSize( 256 );
    
    te::rst::RasterSummary summary;
    te::rst::RasterSummaryManager::getInstance().get( inputRasterPtr.get(), true, summary );    
    
    BOOST_REQUIRE( summary.size() == 6 );
    BOOST_REQUIRE_CLOSE( summary.at( 0 ).m_minVal->real(), -4.1985388617416337, 0.1 );
    BOOST_REQUIRE_CLOSE( summary.at( 0 ).m_maxVal->real(), 2.0537622690155928, 0.1 );
    BOOST_REQUIRE_CLOSE( summary.at( 0 ).m_stdVal->real(), 1.2018360106534016, 0.1 );
    BOOST_REQUIRE_CLOSE( summary.at( 0 ).m_meanVal->real(), -1.428157779296426, 0.1 );
  }

  // find test
  
  {
    std::map<std::string, std::string> auxRasterInfo;

    auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers2b_rgb342_crop_linearMixtureModel_double.tif";
    boost::shared_ptr< te::rst::Raster > inputRasterPtr ( te::rst::RasterFactory::open(
      auxRasterInfo ) );
    BOOST_REQUIRE( inputRasterPtr.get() );  
    
    te::rst::RasterSummary summary;
    BOOST_REQUIRE( te::rst::RasterSummaryManager::getInstance().find( 
      inputRasterPtr.get(), summary ) );
    
    BOOST_REQUIRE( summary.size() == inputRasterPtr->getNumberOfBands() );
  }  
  
  // using cache
  
  {
    std::map<std::string, std::string> auxRasterInfo;

    auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers2b_rgb342_crop_linearMixtureModel_double.tif";
    boost::shared_ptr< te::rst::Raster > inputRasterPtr ( te::rst::RasterFactory::open(
      auxRasterInfo ) );
    BOOST_REQUIRE( inputRasterPtr.get() );  
  
    te::rst::RasterSummary summary;
    te::rst::RasterSummaryManager::getInstance().get( inputRasterPtr.get(), false, summary );    
    BOOST_REQUIRE( summary.size() == inputRasterPtr->getNumberOfBands() );
  }
}

BOOST_AUTO_TEST_CASE (get_complex_test)
{
  // create summary
  
  {
    std::map<std::string, std::string> auxRasterInfo;

    auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/te_complex.tif";
    boost::shared_ptr< te::rst::Raster > inputRasterPtr ( te::rst::RasterFactory::open(
      auxRasterInfo ) );
    BOOST_REQUIRE( inputRasterPtr.get() );  
  
    te::rst::RasterSummaryManager::getInstance().removeAll();
    te::rst::RasterSummaryManager::getInstance().enableDeletedRastersCache( true );    
    te::rst::RasterSummaryManager::getInstance().setDefaultRasterHistSize( 256 );
    
    te::rst::RasterSummary summary;
    te::rst::RasterSummaryManager::getInstance().get( inputRasterPtr.get(), true, summary );    
    
    BOOST_REQUIRE( summary.size() == 1 );
    BOOST_REQUIRE_CLOSE( summary.at( 0 ).m_minVal->real(), 11, 0 );
    BOOST_REQUIRE_CLOSE( summary.at( 0 ).m_maxVal->real(), 239, 0 );
    BOOST_REQUIRE_CLOSE( summary.at( 0 ).m_stdVal->real(), 27.867499487010917, 0.1 );
    BOOST_REQUIRE_CLOSE( summary.at( 0 ).m_meanVal->real(), 131.85372926770705, 0.1 );
    
    BOOST_REQUIRE_CLOSE( summary.at( 0 ).m_minVal->imag(), 3, 0.1 );
    BOOST_REQUIRE_CLOSE( summary.at( 0 ).m_maxVal->imag(), 242, 0.1 );
    BOOST_REQUIRE_CLOSE( summary.at( 0 ).m_stdVal->imag(), 26.596501099937015, 0.1 );
    BOOST_REQUIRE_CLOSE( summary.at( 0 ).m_meanVal->imag(), 133.56333390956388, 0.1 );
  }

  // find test
  
  {
    std::map<std::string, std::string> auxRasterInfo;

    auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/te_complex.tif";
    boost::shared_ptr< te::rst::Raster > inputRasterPtr ( te::rst::RasterFactory::open(
      auxRasterInfo ) );
    BOOST_REQUIRE( inputRasterPtr.get() );  
    
    te::rst::RasterSummary summary;
    BOOST_REQUIRE( te::rst::RasterSummaryManager::getInstance().find( 
      inputRasterPtr.get(), summary ) );
    
    BOOST_REQUIRE( summary.size() == inputRasterPtr->getNumberOfBands() );
  }  
  
  // using cache
  
  {
    std::map<std::string, std::string> auxRasterInfo;

    auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/te_complex.tif";
    boost::shared_ptr< te::rst::Raster > inputRasterPtr ( te::rst::RasterFactory::open(
      auxRasterInfo ) );
    BOOST_REQUIRE( inputRasterPtr.get() );  
  
    te::rst::RasterSummary summary;
    te::rst::RasterSummaryManager::getInstance().get( inputRasterPtr.get(), false, summary );    
    BOOST_REQUIRE( summary.size() == inputRasterPtr->getNumberOfBands() );
  }  
}

BOOST_AUTO_TEST_SUITE_END ()
