/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/unittest/raster/TsBlockUtils.cpp
 
  \brief A test suit for the BlockUtils class.
 */

// TerraLib
#include <terralib_buildconfig.h>

#include <terralib/raster.h>
#include "../Config.h"

// Boost
#include <boost/test/unit_test.hpp>
#include <boost/shared_ptr.hpp>

BOOST_AUTO_TEST_SUITE (utils_tests)

BOOST_AUTO_TEST_CASE (getHistograms_SingleThread_test)
{
  /* Openning input raster */

  std::map<std::string, std::string> auxRasterInfo;

  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers_rgb342_crop1.tif";
  boost::shared_ptr< te::rst::Raster > inputRasterPtr ( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_CHECK( inputRasterPtr.get() );

  // Single thread test

  std::map<double, unsigned> rHistogram;
  std::map<double, unsigned> iHistogram;

  BOOST_CHECK( te::rst::getHistograms( *inputRasterPtr, 0, 0, 0,
    inputRasterPtr->getNumberOfRows() - 1,
    inputRasterPtr->getNumberOfColumns() - 1, 0, 1, 1, rHistogram, iHistogram ) );
  BOOST_CHECK_EQUAL( rHistogram.size(), 161 );
  BOOST_CHECK_EQUAL( iHistogram.size(), 1 );
}

BOOST_AUTO_TEST_CASE (getHistograms_SampleStep_test)
{
  /* Openning input raster */

  std::map<std::string, std::string> auxRasterInfo;

  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers_rgb342_crop1.tif";
  boost::shared_ptr< te::rst::Raster > inputRasterPtr ( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_CHECK( inputRasterPtr.get() );

  // Single thread test

  std::map<double, unsigned> rHistogram;
  std::map<double, unsigned> iHistogram;
  
  unsigned int sampleStep = 2 * std::min(
    inputRasterPtr->getBand( 0 )->getProperty()->m_blkw,
    inputRasterPtr->getBand( 0 )->getProperty()->m_blkh );

  BOOST_CHECK( te::rst::getHistograms( *inputRasterPtr, 
    0, // bandIndex
    0, // rowStart
    0, // colStart
    inputRasterPtr->getNumberOfRows() - 1,
    inputRasterPtr->getNumberOfColumns() - 1, 
    0, // histoBins
    sampleStep, 
    1, // maxThreads
    rHistogram, iHistogram ) );
  BOOST_CHECK_EQUAL( rHistogram.size(), 155 );
  BOOST_CHECK_EQUAL( iHistogram.size(), 1 );
}

BOOST_AUTO_TEST_CASE (getHistograms_MultiThread_test)
{
  /* Openning input raster */

  std::map<std::string, std::string> auxRasterInfo;

  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers_rgb342_crop1.tif";
  boost::shared_ptr< te::rst::Raster > inputRasterPtr ( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_CHECK( inputRasterPtr.get() );

  // Muilti thread test

  std::map<double, unsigned> rHistogram;
  std::map<double, unsigned> iHistogram;

  BOOST_CHECK( te::rst::getHistograms( *inputRasterPtr, 0, 0, 0,
    inputRasterPtr->getNumberOfRows() - 1,
    inputRasterPtr->getNumberOfColumns() - 1, 0, 1, 0, rHistogram, iHistogram ) );
  BOOST_CHECK_EQUAL( rHistogram.size(), 161 );
  BOOST_CHECK_EQUAL( iHistogram.size(), 1 );
}

BOOST_AUTO_TEST_CASE (getHistograms_PredefinedBins_test)
{
  /* Openning input raster */

  std::map<std::string, std::string> auxRasterInfo;

  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers_rgb342_crop1.tif";
  boost::shared_ptr< te::rst::Raster > inputRasterPtr ( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_CHECK( inputRasterPtr.get() );

  // predefined bins test

  std::map<double, unsigned> rHistogram;
  std::map<double, unsigned> iHistogram;

  BOOST_CHECK( te::rst::getHistograms( *inputRasterPtr, 0, 0, 0,
    inputRasterPtr->getNumberOfRows() - 1,
    inputRasterPtr->getNumberOfColumns() - 1, 100, 1, 1, rHistogram, iHistogram ) );
  BOOST_CHECK_EQUAL( rHistogram.size(), 100 );
  BOOST_CHECK_EQUAL( iHistogram.size(), 1 );
}

BOOST_AUTO_TEST_CASE (getHistograms_DoubleData_test)
{
  /* Openning input raster */

  std::map<std::string, std::string> auxRasterInfo;

  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers2b_rgb342_crop_linearMixtureModel_double.tif";
  boost::shared_ptr< te::rst::Raster > inputRasterPtr ( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_CHECK( inputRasterPtr.get() );

  // Single thread test

  std::map<double, unsigned> rHistogram;
  std::map<double, unsigned> iHistogram;

  BOOST_CHECK( te::rst::getHistograms( 
    *inputRasterPtr, 
    0, 
    0, 
    0,
    inputRasterPtr->getNumberOfRows() - 1,
    inputRasterPtr->getNumberOfColumns() - 1, 
    0, /* histoBins */
    1, /* sampleStep */
    1, /* maxThreads */
    rHistogram, 
    iHistogram ) );
  BOOST_CHECK_EQUAL( rHistogram.size(), 256 );
  BOOST_CHECK_EQUAL( iHistogram.size(), 1 );
}

BOOST_AUTO_TEST_CASE (getHistograms_Uint32Data_4096bands_singlethread_test)
{
  std::map<std::string, std::string> auxRasterInfo;

  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/raster_16rows_16cols_4096bands_uint32.tif";
  boost::shared_ptr< te::rst::Raster > inputRasterPtr ( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_CHECK( inputRasterPtr.get() );

  const unsigned int nBands = inputRasterPtr->getNumberOfBands();
  std::map<double, unsigned> rHistogram;
  std::map<double, unsigned> iHistogram;  
  
  for( unsigned int bandIdx = 0 ; bandIdx < nBands ; ++bandIdx )
  {
    rHistogram.clear();
    iHistogram.clear();
    
    BOOST_CHECK( te::rst::getHistograms( 
      *inputRasterPtr, 
      bandIdx, 
      0, 
      0,
      inputRasterPtr->getNumberOfRows() - 1,
      inputRasterPtr->getNumberOfColumns() - 1, 
      0, /* histoBins */
      1, /* sampleStep */
      1, /* maxThreads */
      rHistogram, 
      iHistogram ) );
    BOOST_CHECK_EQUAL( rHistogram.size(), 256 );
    BOOST_CHECK_EQUAL( iHistogram.size(), 1 );
  }
}

BOOST_AUTO_TEST_CASE (getHistograms_DoubleData_4096bands_singlethread_test)
{
  std::map<std::string, std::string> auxRasterInfo;

  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/raster_16rows_16cols_4096bands_double.tif";
  boost::shared_ptr< te::rst::Raster > inputRasterPtr ( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_CHECK( inputRasterPtr.get() );

  const unsigned int nBands = inputRasterPtr->getNumberOfBands();
  std::map<double, unsigned> rHistogram;
  std::map<double, unsigned> iHistogram;  
  
  for( unsigned int bandIdx = 0 ; bandIdx < nBands ; ++bandIdx )
  {
    rHistogram.clear();
    iHistogram.clear();
    
    BOOST_CHECK( te::rst::getHistograms( 
      *inputRasterPtr, 
      bandIdx, 
      0, 
      0,
      inputRasterPtr->getNumberOfRows() - 1,
      inputRasterPtr->getNumberOfColumns() - 1, 
      0, /* histoBins */
      1, /* sampleStep */
      1, /* maxThreads */
      rHistogram, 
      iHistogram ) );
    BOOST_CHECK_EQUAL( rHistogram.size(), 256 );
    BOOST_CHECK_EQUAL( iHistogram.size(), 1 );
  }
}

BOOST_AUTO_TEST_CASE (getHistograms_DoubleData_4096bands_multithread_test)
{
  std::map<std::string, std::string> auxRasterInfo;

  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/raster_16rows_16cols_4096bands_double.tif";
  boost::shared_ptr< te::rst::Raster > inputRasterPtr ( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_CHECK( inputRasterPtr.get() );

  const unsigned int nBands = inputRasterPtr->getNumberOfBands();
  std::map<double, unsigned> rHistogram;
  std::map<double, unsigned> iHistogram;  
  
  for( unsigned int bandIdx = 0 ; bandIdx < nBands ; ++bandIdx )
  {
    rHistogram.clear();
    iHistogram.clear();
    
    BOOST_CHECK( te::rst::getHistograms( 
      *inputRasterPtr, 
      bandIdx, 
      0, 
      0,
      inputRasterPtr->getNumberOfRows() - 1,
      inputRasterPtr->getNumberOfColumns() - 1, 
      0, /* histoBins */
      1, /* sampleStep */
      2, /* maxThreads */
      rHistogram, 
      iHistogram ) );
    BOOST_CHECK_EQUAL( rHistogram.size(), 256 );
    BOOST_CHECK_EQUAL( iHistogram.size(), 1 );
  }
}

BOOST_AUTO_TEST_CASE (getHistograms_DoubleDataPreDefinedBins_test)
{
  /* Openning input raster */

  std::map<std::string, std::string> auxRasterInfo;

  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers2b_rgb342_crop_linearMixtureModel_double.tif";
  boost::shared_ptr< te::rst::Raster > inputRasterPtr ( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_CHECK( inputRasterPtr.get() );

  // Single thread test

  std::map<double, unsigned> rHistogram;
  std::map<double, unsigned> iHistogram;

  BOOST_CHECK( te::rst::getHistograms( 
    *inputRasterPtr, 
    0, 
    0, 
    0,
    inputRasterPtr->getNumberOfRows() - 1,
    inputRasterPtr->getNumberOfColumns() - 1, 
    100, /* histoBins */
    1, /* sampleStep */
    1, /* maxThreads */
    rHistogram, 
    iHistogram ) );
  BOOST_CHECK_EQUAL( rHistogram.size(), 100 );
  BOOST_CHECK_EQUAL( iHistogram.size(), 1 );
}

BOOST_AUTO_TEST_CASE (getHistograms_ComplexData_test)
{
  /* Openning input raster */

  std::map<std::string, std::string> auxRasterInfo;

  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/te_complex.tif";
  boost::shared_ptr< te::rst::Raster > inputRasterPtr ( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_CHECK( inputRasterPtr.get() );

  // Single thread test

  std::map<double, unsigned> rHistogram;
  std::map<double, unsigned> iHistogram;

  BOOST_CHECK( te::rst::getHistograms( 
    *inputRasterPtr, 
    0, 
    0, 
    0,
    inputRasterPtr->getNumberOfRows() - 1,
    inputRasterPtr->getNumberOfColumns() - 1, 
    0, /* histoBins */
    1, /* sampleStep */
    1, /* maxThreads */
    rHistogram, 
    iHistogram ) );
  BOOST_CHECK_EQUAL( rHistogram.size(), 229 );
  BOOST_CHECK_EQUAL( iHistogram.size(), 240 );
}

BOOST_AUTO_TEST_CASE (getHistogramStats_test)
{
  /* Openning input raster */

  std::map<std::string, std::string> auxRasterInfo;

  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers_rgb342_crop1.tif";
  boost::shared_ptr< te::rst::Raster > inputRasterPtr ( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_CHECK( inputRasterPtr.get() );

  // Single thread test

  std::map<double, unsigned> rHistogram;
  std::map<double, unsigned> iHistogram;

  BOOST_CHECK( te::rst::getHistograms( *inputRasterPtr, 0, 0, 0,
    inputRasterPtr->getNumberOfRows() - 1,
    inputRasterPtr->getNumberOfColumns() - 1, 0, 1, 0, rHistogram, iHistogram ) );
  BOOST_CHECK_EQUAL( rHistogram.size(), 161 );

  double min = 0;
  double max = 0;
  double mean = 0;
  double stdDev = 0;

  te::rst::getHistogramStats( rHistogram, min, max, mean, stdDev );
  BOOST_REQUIRE_CLOSE( min, 24, 0.1 );
  BOOST_REQUIRE_CLOSE( max, 184, 0.1 );
  BOOST_REQUIRE_CLOSE( mean, 42.26, 0.1 );
  BOOST_REQUIRE_CLOSE( stdDev, 7.07, 0.1 );

  te::rst::getHistogramStats( iHistogram, min, max, mean, stdDev );
  BOOST_REQUIRE_CLOSE( min, 0, 0.00001 );
  BOOST_REQUIRE_CLOSE( max, 0, 0.00001 );
  BOOST_REQUIRE_CLOSE( mean, 0, 0.00001 );
  BOOST_REQUIRE_CLOSE( stdDev, 0, 0.00001 );
}

BOOST_AUTO_TEST_CASE (rasterResample_samesize_test)
{
  /* Openning input raster */
  
  const std::string outFileName = "terralib_unittest_rst_utils_rasterResample_samesize_test.tif";

  std::map<std::string, std::string> auxRasterInfo;

  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers_rgb342_crop1.tif";
  boost::shared_ptr< te::rst::Raster > inputRasterPtr ( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_REQUIRE( inputRasterPtr.get() );

  {
    std::vector< unsigned int > inputRasterBands;
    inputRasterBands.push_back( 0 );
    inputRasterBands.push_back( 1 );
    inputRasterBands.push_back( 2 );
    
    std::map<std::string, std::string> rinfo;
    rinfo["URI"] = outFileName;
    
    std::unique_ptr< te::rst::Raster > resampledRasterPtr;

    BOOST_REQUIRE( te::rst::rasterResample( 
      *inputRasterPtr, 
      inputRasterBands,
      te::rst::NearestNeighbor,
      1, // interp window radius
      0, // first row
      0, // first col
      inputRasterPtr->getNumberOfRows(),
      inputRasterPtr->getNumberOfColumns(),
      inputRasterPtr->getNumberOfRows(),
      inputRasterPtr->getNumberOfColumns(),
      1, // maxThreads
      rinfo, 
      "GDAL",
      resampledRasterPtr ) );
    
    BOOST_REQUIRE( resampledRasterPtr.get() != 0  );
  }
  
  auxRasterInfo["URI"] = outFileName;
  boost::shared_ptr< te::rst::Raster > inputRasterPtr2 ( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_REQUIRE( inputRasterPtr2.get() );  
  
  BOOST_REQUIRE( inputRasterPtr2->getNumberOfBands() == 3 );
  BOOST_REQUIRE( inputRasterPtr2->getNumberOfColumns() == 
    inputRasterPtr->getNumberOfColumns() );
  BOOST_REQUIRE( inputRasterPtr2->getNumberOfRows() == 
    inputRasterPtr->getNumberOfRows() );
  BOOST_REQUIRE( inputRasterPtr2->getGrid()->getExtent()->operator==(
    *inputRasterPtr->getGrid()->getExtent() ) );  
}

BOOST_AUTO_TEST_CASE (rasterResample_SingleThread_test)
{
  const std::string outFileName = "terralib_unittest_rst_utils_rasterResample_SingleThread_test.tif";
  
  /* Openning input raster */

  std::map<std::string, std::string> auxRasterInfo;

  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers_rgb342_crop1.tif";
  boost::shared_ptr< te::rst::Raster > inputRasterPtr ( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_REQUIRE( inputRasterPtr.get() );

  {
    std::vector< unsigned int > inputRasterBands;
    inputRasterBands.push_back( 0 );
    inputRasterBands.push_back( 1 );
    inputRasterBands.push_back( 2 );
    
    std::map<std::string, std::string> rinfo;
    rinfo["URI"] = outFileName;
    
    std::unique_ptr< te::rst::Raster > resampledRasterPtr;

    BOOST_REQUIRE( te::rst::rasterResample( 
      *inputRasterPtr, 
      inputRasterBands,
      te::rst::NearestNeighbor,
      1, // interp window radius
      0, // first row
      0, // first col
      inputRasterPtr->getNumberOfRows(),
      inputRasterPtr->getNumberOfColumns(),
      inputRasterPtr->getNumberOfRows() * 2,
      inputRasterPtr->getNumberOfColumns() * 2,
      1, // maxThreads
      rinfo, 
      "GDAL",
      resampledRasterPtr ) );
    
    BOOST_REQUIRE( resampledRasterPtr.get() != 0  );
  }
  
  auxRasterInfo["URI"] = outFileName;
  boost::shared_ptr< te::rst::Raster > inputRasterPtr2 ( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_REQUIRE( inputRasterPtr2.get() );  
  
  BOOST_REQUIRE( inputRasterPtr2->getNumberOfBands() == 3 );
  BOOST_REQUIRE( inputRasterPtr2->getNumberOfColumns() == 
    2 * inputRasterPtr->getNumberOfColumns() );
  BOOST_REQUIRE( inputRasterPtr2->getNumberOfRows() == 
    2* inputRasterPtr->getNumberOfRows() );
  BOOST_REQUIRE( inputRasterPtr2->getGrid()->getExtent()->operator==(
    *inputRasterPtr->getGrid()->getExtent() ) );    
}

BOOST_AUTO_TEST_CASE (rasterResample_MultiThread_test)
{
  const std::string outFileName = "terralib_unittest_rst_utils_rasterResample_MultiThread_test.tif";
  
  /* Openning input raster */

  std::map<std::string, std::string> auxRasterInfo;

  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers_rgb342_crop1.tif";
  boost::shared_ptr< te::rst::Raster > inputRasterPtr ( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_REQUIRE( inputRasterPtr.get() );

  {
    std::vector< unsigned int > inputRasterBands;
    inputRasterBands.push_back( 0 );
    inputRasterBands.push_back( 1 );
    inputRasterBands.push_back( 2 );
    
    std::map<std::string, std::string> rinfo;
    rinfo["URI"] = outFileName;
    
    std::unique_ptr< te::rst::Raster > resampledRasterPtr;

    BOOST_REQUIRE( te::rst::rasterResample( 
      *inputRasterPtr, 
      inputRasterBands,
      te::rst::NearestNeighbor,
      1, // interp window radius
      0, // first row
      0, // first col
      inputRasterPtr->getNumberOfRows(),
      inputRasterPtr->getNumberOfColumns(),
      inputRasterPtr->getNumberOfRows() * 2,
      inputRasterPtr->getNumberOfColumns() * 2,
      4, // maxThreads
      rinfo, 
      "GDAL",
      resampledRasterPtr ) );
    
    BOOST_REQUIRE( resampledRasterPtr.get() != 0  );
  }
  
  auxRasterInfo["URI"] = outFileName;
  boost::shared_ptr< te::rst::Raster > inputRasterPtr2 ( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_REQUIRE( inputRasterPtr2.get() );  
  
  BOOST_REQUIRE( inputRasterPtr2->getNumberOfBands() == 3 );
  BOOST_REQUIRE( inputRasterPtr2->getNumberOfColumns() == 
    2* inputRasterPtr->getNumberOfColumns() );
  BOOST_REQUIRE( inputRasterPtr2->getNumberOfRows() == 
    2 * inputRasterPtr->getNumberOfRows() );
  BOOST_REQUIRE( inputRasterPtr2->getGrid()->getExtent()->operator==(
    *inputRasterPtr->getGrid()->getExtent() ) );    
}

BOOST_AUTO_TEST_SUITE_END()
