/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/unittest/rp/blender/TsBlender.cpp

  \brief A test suit for the Blender interface.
*/

// TerraLib
#include "../Config.h"
#include <terralib/rp.h>
#include <terralib/raster.h>
#include <terralib/geometry.h>
#include <terralib/common.h>

// STL
#include <memory>

// Boost
#define BOOST_TEST_NO_MAIN
#include <boost/test/unit_test.hpp>
#include <boost/scoped_array.hpp>

BOOST_AUTO_TEST_SUITE (composer_tests)

BOOST_AUTO_TEST_CASE(composer_fullRasterCompose_ComposerRuleValidPixel_test)
{
  /* Progress interface */

  te::common::ConsoleProgressViewer progressViewerInstance;
  
  /* Openning input rasters */

  std::map<std::string, std::string> auxRasterInfo;
  
  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers_rgb342_crop2.tif";
  std::unique_ptr< te::rst::Raster > inputRaster1Pointer ( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_REQUIRE( inputRaster1Pointer.get() );
  
  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers_rgb342_crop1.tif";
  std::unique_ptr< te::rst::Raster > inputRaster2Pointer ( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_REQUIRE( inputRaster2Pointer.get() );
  
  /* Creating the output image */

  std::unique_ptr< te::rst::Raster > outRasterPtr;
  
  {
    std::map<std::string, std::string> outputRasterInfo;
    outputRasterInfo["URI"] = "composer_fullRasterCompose_ComposerRuleValidPixel_test.tif";
    
    te::gm::Envelope* envPtr = new te::gm::Envelope( 
      inputRaster1Pointer->getGrid()->getExtent()->getLowerLeftX(),
      inputRaster1Pointer->getGrid()->getExtent()->getLowerLeftY() - (
        inputRaster1Pointer->getGrid()->getExtent()->getUpperRightY() -
        inputRaster1Pointer->getGrid()->getExtent()->getLowerLeftY() ),
      inputRaster1Pointer->getGrid()->getExtent()->getUpperRightX() + (
        inputRaster1Pointer->getGrid()->getExtent()->getUpperRightX() -
        inputRaster1Pointer->getGrid()->getExtent()->getLowerLeftX() ),
      inputRaster1Pointer->getGrid()->getExtent()->getUpperRightY() );
    
    te::rst::Grid* gridPtr = new te::rst::Grid( 
      inputRaster1Pointer->getGrid()->getNumberOfColumns() * 2,
      inputRaster1Pointer->getGrid()->getNumberOfRows() * 2,
      envPtr,
      inputRaster1Pointer->getSRID() );
    
    std::vector< te::rst::BandProperty* > bandsProperties;
    for( unsigned int inRasterBandsIdx = 0 ; inRasterBandsIdx <
      inputRaster1Pointer->getNumberOfBands() ; ++inRasterBandsIdx )
    {
      bandsProperties.push_back( new te::rst::BandProperty(
        *( inputRaster1Pointer->getBand( 0 )->getProperty() ) ) );
      bandsProperties[ inRasterBandsIdx ]->m_blkh = 1;
      bandsProperties[ inRasterBandsIdx ]->m_blkw = gridPtr->getNumberOfColumns();
      bandsProperties[ inRasterBandsIdx ]->m_nblocksx = 1;
      bandsProperties[ inRasterBandsIdx ]->m_nblocksy = gridPtr->getNumberOfRows();
      bandsProperties[ inRasterBandsIdx ]->m_noDataValue = 0.0;
    }
    
    outRasterPtr.reset( te::rst::RasterFactory::make(
      "GDAL",
      gridPtr,
      bandsProperties,
      outputRasterInfo,
      0,
      0 ) );
    BOOST_REQUIRE( outRasterPtr.get() );
    
    const unsigned int outputNRows = outRasterPtr->getNumberOfRows();
    const unsigned int outputNCols = outRasterPtr->getNumberOfColumns();    
    unsigned int row = 0;
    unsigned int col = 0;
    double outNoDataValue = 0;
    
    for( unsigned int band = 0; band < inputRaster1Pointer->getNumberOfBands();
      ++band )
    {
      outNoDataValue = outRasterPtr->getBand( band )->getProperty()->m_noDataValue;
      
      for( row = 0 ; row < outputNRows ; ++row )
      {
        for( col = 0; col < outputNCols ; ++col )
        {
          outRasterPtr->setValue( col, row, outNoDataValue, band );
        }
      }      
    }
  }
  
  /* Creating composer parameters */

  std::vector< unsigned int > raster1Bands;
  raster1Bands.push_back( 0 );
  raster1Bands.push_back( 1 );
  raster1Bands.push_back( 2 );
  
  std::vector< unsigned int > raster2Bands;
  raster2Bands.push_back( 1 );
  raster2Bands.push_back( 2 );
  raster2Bands.push_back( 0 );
  
  std::vector< unsigned int > raster3Bands;
  raster3Bands.push_back( 0 );
  raster3Bands.push_back( 1 );
  raster3Bands.push_back( 2 );
  
  const std::vector< double > pixelOffsets( 3, 0.0 );
  const std::vector< double > pixelScales( 3, 1.0 );
  
  te::rp::ComposerRuleValidPixel ruleInstance;
  
  /* Initiating the blender instance */

  te::rp::Composer composerInstance;
  
  BOOST_REQUIRE_MESSAGE( composerInstance.initialize(
    *inputRaster1Pointer,
    raster1Bands,
    *inputRaster2Pointer,
    raster2Bands,
    *outRasterPtr,
    raster3Bands,
    te::rst::NearestNeighbor,
    te::rst::NearestNeighbor,
    0.0,
    false,
    false,
    pixelOffsets,
    pixelScales,
    pixelOffsets,
    pixelScales,
    1,
    true,
    &ruleInstance,
    true ),
    composerInstance.getErrorMessage() );
  
  BOOST_REQUIRE_MESSAGE( composerInstance.composeIntoRaster3(),
    composerInstance.getErrorMessage() );
}

BOOST_AUTO_TEST_CASE(composer_fullRasterCompose_ComposerRuleValidPixelMultiThread_test)
{
  /* Progress interface */

  te::common::ConsoleProgressViewer progressViewerInstance;
  
  /* Openning input rasters */

  std::map<std::string, std::string> auxRasterInfo;
  
  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers_rgb342_crop2.tif";
  std::unique_ptr< te::rst::Raster > inputRaster1Pointer ( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_REQUIRE( inputRaster1Pointer.get() );
  
  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers_rgb342_crop1.tif";
  std::unique_ptr< te::rst::Raster > inputRaster2Pointer ( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_REQUIRE( inputRaster2Pointer.get() );
  
  /* Creating the output image */

  std::unique_ptr< te::rst::Raster > outRasterPtr;
  
  {
    std::map<std::string, std::string> outputRasterInfo;
    outputRasterInfo["URI"] = "composer_fullRasterCompose_ComposerRuleValidPixelMultiThread_test.tif";
    
    te::gm::Envelope* envPtr = new te::gm::Envelope( 
      inputRaster1Pointer->getGrid()->getExtent()->getLowerLeftX(),
      inputRaster1Pointer->getGrid()->getExtent()->getLowerLeftY() - (
        inputRaster1Pointer->getGrid()->getExtent()->getUpperRightY() -
        inputRaster1Pointer->getGrid()->getExtent()->getLowerLeftY() ),
      inputRaster1Pointer->getGrid()->getExtent()->getUpperRightX() + (
        inputRaster1Pointer->getGrid()->getExtent()->getUpperRightX() -
        inputRaster1Pointer->getGrid()->getExtent()->getLowerLeftX() ),
      inputRaster1Pointer->getGrid()->getExtent()->getUpperRightY() );
    
    te::rst::Grid* gridPtr = new te::rst::Grid( 
      inputRaster1Pointer->getGrid()->getNumberOfColumns() * 2,
      inputRaster1Pointer->getGrid()->getNumberOfRows() * 2,
      envPtr,
      inputRaster1Pointer->getSRID() );
    
    std::vector< te::rst::BandProperty* > bandsProperties;
    for( unsigned int inRasterBandsIdx = 0 ; inRasterBandsIdx <
      inputRaster1Pointer->getNumberOfBands() ; ++inRasterBandsIdx )
    {
      bandsProperties.push_back( new te::rst::BandProperty(
        *( inputRaster1Pointer->getBand( 0 )->getProperty() ) ) );
      bandsProperties[ inRasterBandsIdx ]->m_blkh = 1;
      bandsProperties[ inRasterBandsIdx ]->m_blkw = gridPtr->getNumberOfColumns();
      bandsProperties[ inRasterBandsIdx ]->m_nblocksx = 1;
      bandsProperties[ inRasterBandsIdx ]->m_nblocksy = gridPtr->getNumberOfRows();
      bandsProperties[ inRasterBandsIdx ]->m_noDataValue = 0.0;
    }
    
    outRasterPtr.reset( te::rst::RasterFactory::make(
      "GDAL",
      gridPtr,
      bandsProperties,
      outputRasterInfo,
      0,
      0 ) );
    BOOST_REQUIRE( outRasterPtr.get() );
    
    const unsigned int outputNRows = outRasterPtr->getNumberOfRows();
    const unsigned int outputNCols = outRasterPtr->getNumberOfColumns();    
    unsigned int row = 0;
    unsigned int col = 0;
    double outNoDataValue = 0;
    
    for( unsigned int band = 0; band < inputRaster1Pointer->getNumberOfBands();
      ++band )
    {
      outNoDataValue = outRasterPtr->getBand( band )->getProperty()->m_noDataValue;
      
      for( row = 0 ; row < outputNRows ; ++row )
      {
        for( col = 0; col < outputNCols ; ++col )
        {
          outRasterPtr->setValue( col, row, outNoDataValue, band );
        }
      }      
    }
  }
  
  /* Creating composer parameters */

  std::vector< unsigned int > raster1Bands;
  raster1Bands.push_back( 0 );
  raster1Bands.push_back( 1 );
  raster1Bands.push_back( 2 );
  
  std::vector< unsigned int > raster2Bands;
  raster2Bands.push_back( 1 );
  raster2Bands.push_back( 2 );
  raster2Bands.push_back( 0 );
  
  std::vector< unsigned int > raster3Bands;
  raster3Bands.push_back( 0 );
  raster3Bands.push_back( 1 );
  raster3Bands.push_back( 2 );
  
  const std::vector< double > pixelOffsets( 3, 0.0 );
  const std::vector< double > pixelScales( 3, 1.0 );
  
  te::rp::ComposerRuleValidPixel ruleInstance;
  
  /* Initiating the blender instance */

  te::rp::Composer composerInstance;
  
  BOOST_REQUIRE_MESSAGE( composerInstance.initialize(
    *inputRaster1Pointer,
    raster1Bands,
    *inputRaster2Pointer,
    raster2Bands,
    *outRasterPtr,
    raster3Bands,
    te::rst::NearestNeighbor,
    te::rst::NearestNeighbor,
    0.0,
    false,
    false,
    pixelOffsets,
    pixelScales,
    pixelOffsets,
    pixelScales,
    4,
    true,
    &ruleInstance,
    true ),
    composerInstance.getErrorMessage() );
  
  BOOST_REQUIRE_MESSAGE( composerInstance.composeIntoRaster3(),
    composerInstance.getErrorMessage() );
}

BOOST_AUTO_TEST_CASE(composer_fullRasterCompose_ComposerRuleValidNeighborhood_test)
{
  /* Progress interface */

  te::common::ConsoleProgressViewer progressViewerInstance;
  
  /* Openning input rasters */

  std::map<std::string, std::string> auxRasterInfo;
  
  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers_rgb342_crop2.tif";
  std::unique_ptr< te::rst::Raster > inputRaster1Pointer ( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_REQUIRE( inputRaster1Pointer.get() );
  
  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers_rgb342_crop1.tif";
  std::unique_ptr< te::rst::Raster > inputRaster2Pointer ( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_REQUIRE( inputRaster2Pointer.get() );
  
  /* Creating the output image */

  std::unique_ptr< te::rst::Raster > outRasterPtr;
  
  {
    std::map<std::string, std::string> outputRasterInfo;
    outputRasterInfo["URI"] = "composer_fullRasterCompose_ComposerRuleValidNeighborhood_test.tif";
    
    te::gm::Envelope* envPtr = new te::gm::Envelope( 
      inputRaster1Pointer->getGrid()->getExtent()->getLowerLeftX(),
      inputRaster1Pointer->getGrid()->getExtent()->getLowerLeftY() - (
        inputRaster1Pointer->getGrid()->getExtent()->getUpperRightY() -
        inputRaster1Pointer->getGrid()->getExtent()->getLowerLeftY() ),
      inputRaster1Pointer->getGrid()->getExtent()->getUpperRightX() + (
        inputRaster1Pointer->getGrid()->getExtent()->getUpperRightX() -
        inputRaster1Pointer->getGrid()->getExtent()->getLowerLeftX() ),
      inputRaster1Pointer->getGrid()->getExtent()->getUpperRightY() );
    
    te::rst::Grid* gridPtr = new te::rst::Grid( 
      inputRaster1Pointer->getGrid()->getNumberOfColumns() * 2,
      inputRaster1Pointer->getGrid()->getNumberOfRows() * 2,
      envPtr,
      inputRaster1Pointer->getSRID() );
    
    std::vector< te::rst::BandProperty* > bandsProperties;
    for( unsigned int inRasterBandsIdx = 0 ; inRasterBandsIdx <
      inputRaster1Pointer->getNumberOfBands() ; ++inRasterBandsIdx )
    {
      bandsProperties.push_back( new te::rst::BandProperty(
        *( inputRaster1Pointer->getBand( 0 )->getProperty() ) ) );
      bandsProperties[ inRasterBandsIdx ]->m_blkh = 1;
      bandsProperties[ inRasterBandsIdx ]->m_blkw = gridPtr->getNumberOfColumns();
      bandsProperties[ inRasterBandsIdx ]->m_nblocksx = 1;
      bandsProperties[ inRasterBandsIdx ]->m_nblocksy = gridPtr->getNumberOfRows();
      bandsProperties[ inRasterBandsIdx ]->m_noDataValue = 0.0;
    }
    
    outRasterPtr.reset( te::rst::RasterFactory::make(
      "GDAL",
      gridPtr,
      bandsProperties,
      outputRasterInfo,
      0,
      0 ) );
    BOOST_REQUIRE( outRasterPtr.get() );
    
    const unsigned int outputNRows = outRasterPtr->getNumberOfRows();
    const unsigned int outputNCols = outRasterPtr->getNumberOfColumns();    
    unsigned int row = 0;
    unsigned int col = 0;
    double outNoDataValue = 0;
    
    for( unsigned int band = 0; band < inputRaster1Pointer->getNumberOfBands();
      ++band )
    {
      outNoDataValue = outRasterPtr->getBand( band )->getProperty()->m_noDataValue;
      
      for( row = 0 ; row < outputNRows ; ++row )
      {
        for( col = 0; col < outputNCols ; ++col )
        {
          outRasterPtr->setValue( col, row, outNoDataValue, band );
        }
      }      
    }
  }
  
  /* Creating composer parameters */

  std::vector< unsigned int > raster1Bands;
  raster1Bands.push_back( 0 );
  raster1Bands.push_back( 1 );
  raster1Bands.push_back( 2 );
  
  std::vector< unsigned int > raster2Bands;
  raster2Bands.push_back( 1 );
  raster2Bands.push_back( 2 );
  raster2Bands.push_back( 0 );
  
  std::vector< unsigned int > raster3Bands;
  raster3Bands.push_back( 0 );
  raster3Bands.push_back( 1 );
  raster3Bands.push_back( 2 );
  
  const std::vector< double > pixelOffsets( 3, 0.0 );
  const std::vector< double > pixelScales( 3, 1.0 );
  
  std::vector< std::pair< double, double > > ranges1;
  ranges1.resize( raster1Bands.size(), std::pair< double, double >( 0, 250 ) );
  
  std::vector< std::pair< double, double > > ranges2 = ranges1;
  
  te::rp::ComposerRuleValidNeighborhood ruleInstance;
  ruleInstance.setValidValuesRanges( ranges1, ranges2 );
  ruleInstance.setWindowRadius( 1 );
  ruleInstance.setNeiborMinValidPercent( 50 );
  
  /* Initiating the blender instance */

  te::rp::Composer composerInstance;
  
  BOOST_REQUIRE_MESSAGE( composerInstance.initialize(
    *inputRaster1Pointer,
    raster1Bands,
    *inputRaster2Pointer,
    raster2Bands,
    *outRasterPtr,
    raster3Bands,
    te::rst::NearestNeighbor,
    te::rst::NearestNeighbor,
    0.0,
    false,
    false,
    pixelOffsets,
    pixelScales,
    pixelOffsets,
    pixelScales,
    1,
    true,
    &ruleInstance,
    true ),
    composerInstance.getErrorMessage() );
  
  BOOST_REQUIRE_MESSAGE( composerInstance.composeIntoRaster3(),
    composerInstance.getErrorMessage() );
}

BOOST_AUTO_TEST_SUITE_END()
