/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/unittest/rp/mosaic/TsMosaic.cpp

  \brief A test suit for the Mosaic interface.
*/

// TerraLib
#include "../Config.h"
#include <terralib/rp.h>
#include <terralib/raster.h>
#include <terralib/dataaccess.h>
#include <terralib/common.h>

// Boost
#define BOOST_TEST_NO_MAIN
#include <boost/test/unit_test.hpp>
#include <boost/shared_ptr.hpp>

BOOST_AUTO_TEST_SUITE (composition_tests)

BOOST_AUTO_TEST_CASE(composition_basic_test)
{
  /* Progress interface */

  te::common::ConsoleProgressViewer progressViewerInstance;  
  
  /* Openning input rasters */
  
  std::map<std::string, std::string> auxRasterInfo;
  
  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers_rgb342_crop1.tif";
  boost::shared_ptr< te::rst::Raster > inputRaster1Pointer ( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_REQUIRE( inputRaster1Pointer.get() );
  
  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers_rgb342_crop2.tif";
  boost::shared_ptr< te::rst::Raster > inputRaster2Pointer ( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_REQUIRE( inputRaster2Pointer.get() );
  
  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers_rgb342_crop3_EPSG_22522.tif";
  boost::shared_ptr< te::rst::Raster > inputRaster3Pointer ( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_REQUIRE( inputRaster3Pointer.get() );
  
  /* Creating the algorithm parameters */
  
  te::rp::Composition::InputParameters algoInputParams;
  
  std::vector< const te::rst::Raster* > rasters;
  rasters.push_back( inputRaster1Pointer.get() );
  rasters.push_back( inputRaster2Pointer.get() );
  rasters.push_back( inputRaster3Pointer.get() );
  te::rp::FeederConstRasterVector feeder( rasters );
  algoInputParams.m_feederRasterPtr = &feeder;
  
  std::vector< unsigned int > bands;
  bands.push_back( 0 );
  bands.push_back( 1 );
  bands.push_back( 2 );
  algoInputParams.m_inputRastersBands.push_back( bands );
  
  bands[ 0 ] = 1;
  bands[ 1 ] = 2;
  bands[ 2 ] = 0;
  algoInputParams.m_inputRastersBands.push_back( bands );
  
  bands[ 2 ] = 2;
  bands[ 0 ] = 1;
  bands[ 1 ] = 0;
  algoInputParams.m_inputRastersBands.push_back( bands );
  
  algoInputParams.m_interpMethod = te::rst::NearestNeighbor;
  algoInputParams.m_noDataValue = 0;
  algoInputParams.m_forceInputNoDataValue = true;
  algoInputParams.m_autoEqualize = true;
  algoInputParams.m_useRasterCache = true;
  algoInputParams.m_enableProgress = true;
  algoInputParams.m_enableMultiThread = false;
  algoInputParams.m_compositionTargetMeans.clear();
  algoInputParams.m_compositionTargetStdDevs.clear();
  
  te::rp::ComposerRuleValidPixel ruleInstance;
  algoInputParams.m_composerRulePtr = &ruleInstance;

  te::rp::Composition::OutputParameters algoOutputParams;
  
  algoOutputParams.m_rInfo["URI"] =
    "terralib_unittest_rp_composition_basic_test.tif";
  algoOutputParams.m_rType = "GDAL";
  
  /* Executing the algorithm */
  
  te::rp::Composition algorithmInstance;
  
  BOOST_REQUIRE( algorithmInstance.initialize( algoInputParams ) );
  BOOST_REQUIRE( algorithmInstance.execute( algoOutputParams ) );
}

BOOST_AUTO_TEST_CASE(composition_multithread_test)
{
  /* Progress interface */

  te::common::ConsoleProgressViewer progressViewerInstance;  
  
  /* Openning input rasters */
  
  std::map<std::string, std::string> auxRasterInfo;
  
  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers_rgb342_crop1.tif";
  boost::shared_ptr< te::rst::Raster > inputRaster1Pointer ( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_REQUIRE( inputRaster1Pointer.get() );
  
  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers_rgb342_crop2.tif";
  boost::shared_ptr< te::rst::Raster > inputRaster2Pointer ( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_REQUIRE( inputRaster2Pointer.get() );
  
  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers_rgb342_crop3_EPSG_22522.tif";
  boost::shared_ptr< te::rst::Raster > inputRaster3Pointer ( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_REQUIRE( inputRaster3Pointer.get() );
  
  /* Creating the algorithm parameters */
  
  te::rp::Composition::InputParameters algoInputParams;
  
  std::vector< const te::rst::Raster* > rasters;
  rasters.push_back( inputRaster1Pointer.get() );
  rasters.push_back( inputRaster2Pointer.get() );
  rasters.push_back( inputRaster3Pointer.get() );
  te::rp::FeederConstRasterVector feeder( rasters );
  algoInputParams.m_feederRasterPtr = &feeder;
  
  std::vector< unsigned int > bands;
  bands.push_back( 0 );
  bands.push_back( 1 );
  bands.push_back( 2 );
  algoInputParams.m_inputRastersBands.push_back( bands );
  
  bands[ 0 ] = 1;
  bands[ 1 ] = 2;
  bands[ 2 ] = 0;
  algoInputParams.m_inputRastersBands.push_back( bands );
  
  bands[ 2 ] = 2;
  bands[ 0 ] = 1;
  bands[ 1 ] = 0;
  algoInputParams.m_inputRastersBands.push_back( bands );
  
  algoInputParams.m_interpMethod = te::rst::NearestNeighbor;
  algoInputParams.m_noDataValue = 0;
  algoInputParams.m_forceInputNoDataValue = true;
  algoInputParams.m_autoEqualize = true;
  algoInputParams.m_useRasterCache = true;
  algoInputParams.m_enableProgress = true;
  algoInputParams.m_enableMultiThread = true;
  algoInputParams.m_compositionTargetMeans.clear();
  algoInputParams.m_compositionTargetStdDevs.clear();
  
  te::rp::ComposerRuleValidPixel ruleInstance;
  algoInputParams.m_composerRulePtr = &ruleInstance;

  te::rp::Composition::OutputParameters algoOutputParams;
  
  algoOutputParams.m_rInfo["URI"] =
    "terralib_unittest_rp_composition_multithread_test.tif";
  algoOutputParams.m_rType = "GDAL";
  
  /* Executing the algorithm */
  
  te::rp::Composition algorithmInstance;
  
  BOOST_REQUIRE( algorithmInstance.initialize( algoInputParams ) );
  BOOST_REQUIRE( algorithmInstance.execute( algoOutputParams ) );
}


BOOST_AUTO_TEST_SUITE_END()
