/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/unittest/rp/functions/TsFunctions.cpp

  \brief A test suit for the Functions interface.
*/

// TerraLib
#include "../Config.h"
#include <terralib/rp.h>
#include <terralib/raster.h>
#include <terralib/dataaccess.h>
#include <terralib/memory.h>

// STL
#include <memory>
#include <string>

// Boost
#define BOOST_TEST_NO_MAIN
#include <boost/test/unit_test.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/timer.hpp>

BOOST_AUTO_TEST_SUITE (functions_tests)

void saveToShp( std::vector<te::gm::Polygon*> polygons, 
  std::vector< double >& polygonsValues,
  const std::string& shpBaseFileName )
{
  std::unique_ptr<te::da::DataSetType> dataSetTypePtr1(new te::da::DataSetType(shpBaseFileName));
  
  dataSetTypePtr1->add( new te::dt::SimpleProperty("value", te::dt::DOUBLE_TYPE, true) );  
  dataSetTypePtr1->add( new te::dt::SimpleProperty("id", te::dt::DOUBLE_TYPE, true) );
  dataSetTypePtr1->add( new te::gm::GeometryProperty("polygon", polygons[ 0 ]->getSRID(), 
    te::gm::PolygonType, true) );  
  
  std::unique_ptr<te::da::DataSetType> dataSetTypePtr2( new te::da::DataSetType( *dataSetTypePtr1 ) );
  
  std::unique_ptr< te::mem::DataSet > memDataSetPtr( new te::mem::DataSet( dataSetTypePtr1.get()) );
  
  for( unsigned int polygonsIdx = 0 ; polygonsIdx < polygons.size() ; ++polygonsIdx )
  {
    te::mem::DataSetItem* dsItemPtr = new te::mem::DataSetItem(memDataSetPtr.get());
    dsItemPtr->setDouble( 0, polygonsValues[ polygonsIdx ] );
    dsItemPtr->setDouble( 1, polygonsIdx );
    dsItemPtr->setGeometry( 2, (te::gm::Geometry*)polygons[ polygonsIdx ]->clone() );
    
    memDataSetPtr->add( dsItemPtr );
  }

  remove( ( shpBaseFileName + ".shx" ).c_str() );
  remove( ( shpBaseFileName + ".shp" ).c_str() );
  remove( ( shpBaseFileName + ".prj" ).c_str() );
  remove( ( shpBaseFileName + ".dbf" ).c_str() );
  
  std::string connInfo("file://" + shpBaseFileName + ".shp");
  
  std::unique_ptr<te::da::DataSource> dsOGR( te::da::DataSourceFactory::make("OGR", connInfo) );
  dsOGR->open();
  
  memDataSetPtr->moveBeforeFirst();
  
  te::da::Create(dsOGR.get(), dataSetTypePtr2.get(), memDataSetPtr.get());

  dsOGR->close();
}  

BOOST_AUTO_TEST_CASE( createFixedStepPalette_test )
{
  std::vector< te::rst::BandProperty * > bandsProps;
  bandsProps.push_back( new te::rst::BandProperty( 0, 
    te::dt::UCHAR_TYPE ) );    
  bandsProps.push_back( new te::rst::BandProperty( 1, 
    te::dt::UCHAR_TYPE ) );    
  bandsProps.push_back( new te::rst::BandProperty( 2, 
    te::dt::UCHAR_TYPE ) );   
  bandsProps[ 0 ]->m_colorInterp = te::rst::RedCInt;
  bandsProps[ 1 ]->m_colorInterp = te::rst::GreenCInt;
  bandsProps[ 2 ]->m_colorInterp = te::rst::BlueCInt;  
  bandsProps[ 0 ]->m_noDataValue = std::numeric_limits< double >::max();
  bandsProps[ 1 ]->m_noDataValue = std::numeric_limits< double >::max();
  bandsProps[ 2 ]->m_noDataValue = std::numeric_limits< double >::max();  
  
  std::map< std::string, std::string > rInfo;
  rInfo["URI"] = "terralib_unittest_rp_functions_CreateFixedStepPalette.tif";
  
  std::unique_ptr< te::rst::Raster > rasterPointer( te::rst::RasterFactory::make( "GDAL", 
    new te::rst::Grid( 256, 256 ), bandsProps, 
    rInfo, 0, 0 ) );
  
  std::vector< te::rst::BandProperty::ColorEntry > palette;
  te::rp::CreateFixedStepPalette( 256, true, palette );
    
  unsigned int line = 0;
  unsigned int col = 0;
  
  for( line = 0 ; line < 256 ; ++line )
    for( col = 0 ; col < 256 ; ++col )
    {
      rasterPointer->setValue( col, line, (double)palette[ line ].c1, 0 );
      rasterPointer->setValue( col, line, (double)palette[ line ].c2, 1 );
      rasterPointer->setValue( col, line, (double)palette[ line ].c3, 2 );
    }
}

BOOST_AUTO_TEST_CASE(rasterSlicing_test)
{
  /* Openning input raster */
  
  std::map<std::string, std::string> auxRasterInfo;
  
  std::vector< double > limits;
  for (int lim = 0; lim < 255; lim += 25)
    limits.push_back((double)lim);
  
  std::vector< double > slicesOutputValues;
  for(std::size_t limitsIdx = 0; limitsIdx < limits.size(); ++limitsIdx )
    slicesOutputValues.push_back( 1.489 * (double)limitsIdx );  
  
  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers2b_hrc_crop.tif";
  std::unique_ptr< te::rst::Raster > diskRasterPtr( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_CHECK( diskRasterPtr.get() );

  // RBG no pallete test
  
  {
    std::unique_ptr< te::rst::Raster > outRasterPtr;
    std::map<std::string, std::string> auxRasterInfo2;
    auxRasterInfo2["URI"] = "terralib_unittest_rp_functions_RasterSlicing_RGB.tif";
    BOOST_CHECK( te::rp::RasterSlicing(
      *diskRasterPtr.get(), 
      0, 
      false,
      limits.size(),
      true,
      auxRasterInfo2,
      "GDAL",
      true,
      0,
      outRasterPtr, 
      &limits,
      0 ) );
  }
  
  // single band pallete test
  
  {
    std::unique_ptr< te::rst::Raster > outRasterPtr;
    std::map<std::string, std::string> auxRasterInfo2;
    auxRasterInfo2["URI"] = "terralib_unittest_rp_functions_RasterSlicing_palette.tif";
    BOOST_CHECK( te::rp::RasterSlicing(
      *diskRasterPtr.get(), 
      0, 
      true,
      limits.size(),
      true,
      auxRasterInfo2,
      "GDAL",
      true,
      0,
      outRasterPtr,
      &limits,
      0 ) );
  }
  
  // single band pallete no limits test
  
  {
    std::unique_ptr< te::rst::Raster > outRasterPtr;
    std::map<std::string, std::string> auxRasterInfo2;
    auxRasterInfo2["URI"] = "terralib_unittest_rp_functions_RasterSlicing_nolimits_palette.tif";
    BOOST_CHECK( te::rp::RasterSlicing(
      *diskRasterPtr.get(), 
      0, 
      true,
      limits.size(),
      true,
      auxRasterInfo2,
      "GDAL",
      true,
      0,
      outRasterPtr,
      0,
      0 ) );
  }  
  
  // single band double output values
  
  {
    std::unique_ptr< te::rst::Raster > outRasterPtr;
    std::map<std::string, std::string> auxRasterInfo2;
    auxRasterInfo2["URI"] = "terralib_unittest_rp_functions_RasterSlicing_doubleOutput.tif";
    BOOST_CHECK( te::rp::RasterSlicing(
      *diskRasterPtr.get(), 
      0, 
      true,
      limits.size(),
      true,
      auxRasterInfo2,
      "GDAL",
      true,
      0,
      outRasterPtr,
      &limits,
      &slicesOutputValues ) );
  }  
}

BOOST_AUTO_TEST_CASE(ihs_test)
{
  /* Openning input raster */
  
  std::map<std::string, std::string> auxRasterInfo;
  
  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers2b_rgb342_crop.tif";
  std::unique_ptr< te::rst::Raster > diskRasterPtr( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_CHECK( diskRasterPtr.get() );
  
  auxRasterInfo.clear();
  auxRasterInfo["MEM_RASTER_NROWS"] = boost::lexical_cast< std::string >( 
    diskRasterPtr->getNumberOfRows() );
  auxRasterInfo["MEM_RASTER_NCOLS"] = boost::lexical_cast< std::string >( 
    diskRasterPtr->getNumberOfColumns() );
  auxRasterInfo["MEM_RASTER_DATATYPE"] = boost::lexical_cast< std::string >((int)te::dt::DOUBLE_TYPE);
  auxRasterInfo["MEM_RASTER_NBANDS"] = "3";
   
  std::unique_ptr< te::rst::Raster > ihsRasterPtr( te::rst::RasterFactory::make(
    "MEM", 0, std::vector<te::rst::BandProperty*>(), auxRasterInfo) );
  BOOST_CHECK( ihsRasterPtr.get() );

  BOOST_CHECK( te::rp::ConvertRGB2IHS( *diskRasterPtr, 0, 1, 2, 0, 255, *ihsRasterPtr ) );
  
  BOOST_CHECK( te::rp::Copy2DiskRaster( *ihsRasterPtr, "terralib_unittest_rp_functions_rgb2ihs.tif" ) ) ;
  
  auxRasterInfo.clear();
  auxRasterInfo["MEM_RASTER_NROWS"] = boost::lexical_cast< std::string >( 
    diskRasterPtr->getNumberOfRows() );
  auxRasterInfo["MEM_RASTER_NCOLS"] = boost::lexical_cast< std::string >( 
    diskRasterPtr->getNumberOfColumns() );
  auxRasterInfo["MEM_RASTER_DATATYPE"] = boost::lexical_cast< std::string >((int)te::dt::UCHAR_TYPE);
  auxRasterInfo["MEM_RASTER_NBANDS"] = "3";
   
  std::unique_ptr< te::rst::Raster > rgbRasterPtr( te::rst::RasterFactory::make(
    "MEM", 0, std::vector<te::rst::BandProperty*>(), auxRasterInfo) );
  BOOST_CHECK( rgbRasterPtr.get() );
  
  BOOST_CHECK( te::rp::ConvertIHS2RGB( *ihsRasterPtr, 0, 1, 2, 0, 255, *rgbRasterPtr ) );
  
  std::vector< te::rst::BandProperty * > bandProps;
  bandProps.push_back( new te::rst::BandProperty( *( diskRasterPtr->getBand( 0 )->getProperty() ) ) );
  bandProps.push_back( new te::rst::BandProperty( *( diskRasterPtr->getBand( 1 )->getProperty() ) ) );
  bandProps.push_back( new te::rst::BandProperty( *( diskRasterPtr->getBand( 2 )->getProperty() ) ) );
  
  auxRasterInfo["URI"] = "terralib_unittest_rp_functions_ihs2rgb.tif";
  
  std::unique_ptr< te::rst::Raster > outDiskRasterPtr( te::rst::RasterFactory::make(
    "GDAL", new te::rst::Grid( *( diskRasterPtr->getGrid() ) ), bandProps, 
    auxRasterInfo, 0, 0 ) );
  BOOST_CHECK( outDiskRasterPtr.get() );
  
  const size_t nBands = outDiskRasterPtr->getNumberOfBands();
  const unsigned int nCols = outDiskRasterPtr->getNumberOfColumns();
  const unsigned int nRows = outDiskRasterPtr->getNumberOfRows();
  unsigned int col = 0;
  unsigned int row = 0;
  double value = 0;
  
  for( size_t band = 0 ; band < nBands ; ++band )
  {
    for( row = 0 ; row < nRows ; ++row )
    {
      for( col = 0 ; col < nCols ; ++col )
      {
        rgbRasterPtr->getValue( col, row, value, band );
        outDiskRasterPtr->setValue( col, row, value, band );
      }
    }
  }
}

BOOST_AUTO_TEST_CASE(hls_test)
{
  /* Openning input raster */

  std::map<std::string, std::string> auxRasterInfo;

  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers2b_rgb342_crop.tif";
  std::unique_ptr< te::rst::Raster > diskRasterPtr(te::rst::RasterFactory::open(
    auxRasterInfo));
  BOOST_CHECK(diskRasterPtr.get());

  auxRasterInfo.clear();
  auxRasterInfo["MEM_RASTER_NROWS"] = boost::lexical_cast< std::string >(
    diskRasterPtr->getNumberOfRows());
  auxRasterInfo["MEM_RASTER_NCOLS"] = boost::lexical_cast< std::string >(
    diskRasterPtr->getNumberOfColumns());
  auxRasterInfo["MEM_RASTER_DATATYPE"] = boost::lexical_cast< std::string >((int)te::dt::DOUBLE_TYPE);
  auxRasterInfo["MEM_RASTER_NBANDS"] = "3";

  std::unique_ptr< te::rst::Raster > hlsRasterPtr(te::rst::RasterFactory::make(
    "MEM", 0, std::vector<te::rst::BandProperty*>(), auxRasterInfo));
  BOOST_CHECK(hlsRasterPtr.get());

  BOOST_CHECK(te::rp::ConvertRGB2HLS(*diskRasterPtr, 0, 1, 2, 0, 255, *hlsRasterPtr));

  BOOST_CHECK(te::rp::Copy2DiskRaster(*hlsRasterPtr, "terralib_unittest_rp_functions_rgb2hls.tif"));

  auxRasterInfo.clear();
  auxRasterInfo["MEM_RASTER_NROWS"] = boost::lexical_cast< std::string >(
    diskRasterPtr->getNumberOfRows());
  auxRasterInfo["MEM_RASTER_NCOLS"] = boost::lexical_cast< std::string >(
    diskRasterPtr->getNumberOfColumns());
  auxRasterInfo["MEM_RASTER_DATATYPE"] = boost::lexical_cast< std::string >((int)te::dt::UCHAR_TYPE);
  auxRasterInfo["MEM_RASTER_NBANDS"] = "3";

  std::unique_ptr< te::rst::Raster > rgbRasterPtr(te::rst::RasterFactory::make(
    "MEM", 0, std::vector<te::rst::BandProperty*>(), auxRasterInfo));
  BOOST_CHECK(rgbRasterPtr.get());

  BOOST_CHECK(te::rp::ConvertHLS2RGB(*hlsRasterPtr, 0, 1, 2, 0, 255, *rgbRasterPtr));

  std::vector< te::rst::BandProperty * > bandProps;
  bandProps.push_back(new te::rst::BandProperty(*(diskRasterPtr->getBand(0)->getProperty())));
  bandProps.push_back(new te::rst::BandProperty(*(diskRasterPtr->getBand(1)->getProperty())));
  bandProps.push_back(new te::rst::BandProperty(*(diskRasterPtr->getBand(2)->getProperty())));

  auxRasterInfo["URI"] = "terralib_unittest_rp_functions_hls2rgb.tif";

  std::unique_ptr< te::rst::Raster > outDiskRasterPtr(te::rst::RasterFactory::make(
    "GDAL", new te::rst::Grid(*(diskRasterPtr->getGrid())), bandProps,
    auxRasterInfo, 0, 0));
  BOOST_CHECK(outDiskRasterPtr.get());

  const size_t nBands = outDiskRasterPtr->getNumberOfBands();
  const unsigned int nCols = outDiskRasterPtr->getNumberOfColumns();
  const unsigned int nRows = outDiskRasterPtr->getNumberOfRows();
  unsigned int col = 0;
  unsigned int row = 0;
  double value = 0;

  for (size_t band = 0; band < nBands; ++band)
  {
    for (row = 0; row < nRows; ++row)
    {
      for (col = 0; col < nCols; ++col)
      {
        rgbRasterPtr->getValue(col, row, value, band);
        outDiskRasterPtr->setValue(col, row, value, band);
      }
    }
  }
}

BOOST_AUTO_TEST_CASE(getMeanValue_test)
{
  /* Openning input raster */
  
  std::map<std::string, std::string> auxRasterInfo;
  
  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers2b_hrc_crop.tif";
  std::unique_ptr< te::rst::Raster > diskRasterPtr( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_CHECK( diskRasterPtr.get() );
  
  double meanValue = 0;
  
//  boost::timer timer;
  BOOST_CHECK( te::rp::GetMeanValue( *diskRasterPtr, 0, 1, false, 0, meanValue) );
//  std::cout << std::endl << "Elapsed:" << timer.elapsed() << std::endl;
  BOOST_CHECK_CLOSE( 181.340256531345, meanValue, 0.0001 );
}

BOOST_AUTO_TEST_CASE(getMeanValueOptimized_test)
{
  /* Openning input raster */
  
  std::map<std::string, std::string> auxRasterInfo;
  
  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers2b_hrc_crop.tif";
  std::unique_ptr< te::rst::Raster > diskRasterPtr( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_CHECK( diskRasterPtr.get() );
  
  double meanValue = 0;
  
//  boost::timer timer;
  BOOST_CHECK( te::rp::GetMeanValue( *diskRasterPtr, 0, 4, false, 0, meanValue) );
//  std::cout << std::endl << "Elapsed:" << timer.elapsed() << std::endl;
  BOOST_CHECK_CLOSE( 181.340256531345, meanValue, 0.0001 );
}

BOOST_AUTO_TEST_CASE(getCovarianceValue_test)
{
  // no threads test
  
  {
    /* Openning input raster */
    
    std::map<std::string, std::string> auxRasterInfo;
    
    auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers2b_rgb342_crop.tif";
    std::unique_ptr< te::rst::Raster > diskRasterPtr( te::rst::RasterFactory::open(
      auxRasterInfo ) );
    BOOST_CHECK( diskRasterPtr.get() );
    
    double covarianceValue = 0;
    

    BOOST_CHECK( te::rp::GetCovarianceValue( 
      *diskRasterPtr, 
      0, // band 1
      *diskRasterPtr, 
      1, // band 2
      1, // max threads
      0, // mean value 1 pointer
      0, // mean value 2 pointer
      covarianceValue) );

    BOOST_CHECK_CLOSE( -135.17885990762366, covarianceValue, 0.0001 );
  }
  
  // with with threads
  
  {
    /* Openning input raster */
    
    std::map<std::string, std::string> auxRasterInfo;
    
    auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers2b_rgb342_crop.tif";
    std::unique_ptr< te::rst::Raster > diskRasterPtr( te::rst::RasterFactory::open(
      auxRasterInfo ) );
    BOOST_CHECK( diskRasterPtr.get() );
    
    double covarianceValue = 0;
    
    BOOST_CHECK( te::rp::GetCovarianceValue( 
      *diskRasterPtr, 
      0, // band 1
      *diskRasterPtr, 
      1, // band 2
      0, // max threads
      0, // mean value 1 pointer
      0, // mean value 2 pointer
      covarianceValue) );

    BOOST_CHECK_CLOSE( -135.178859, covarianceValue, 0.01 );
  }
}

BOOST_AUTO_TEST_CASE(principalComponents_test)
{
  /* Openning input raster */
  
  std::map<std::string, std::string> auxRasterInfo;
  
  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers2b_rgb342_crop.tif";
  std::unique_ptr< te::rst::Raster > diskRasterPtr( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_REQUIRE( diskRasterPtr.get() );
  
  std::vector< te::rst::BandProperty * > bandProps1;
  bandProps1.push_back( new te::rst::BandProperty( *( diskRasterPtr->getBand( 0 )->getProperty() ) ) );
  bandProps1[ 0 ]->m_blkh = 1;
  bandProps1[ 0 ]->m_blkw = diskRasterPtr->getNumberOfColumns();
  bandProps1[ 0 ]->m_nblocksx = 1;
  bandProps1[ 0 ]->m_nblocksy = diskRasterPtr->getNumberOfRows();
  bandProps1[ 0 ]->m_type = te::dt::DOUBLE_TYPE;

  bandProps1.push_back( new te::rst::BandProperty( *( diskRasterPtr->getBand( 1 )->getProperty() ) ) );
  bandProps1[ 1 ]->m_blkh = 1;
  bandProps1[ 1 ]->m_blkw = diskRasterPtr->getNumberOfColumns();
  bandProps1[ 1 ]->m_nblocksx = 1;
  bandProps1[ 1 ]->m_nblocksy = diskRasterPtr->getNumberOfRows();
  bandProps1[ 1 ]->m_type = te::dt::DOUBLE_TYPE;
  
  bandProps1.push_back( new te::rst::BandProperty( *( diskRasterPtr->getBand( 2 )->getProperty() ) ) );
  bandProps1[ 2 ]->m_blkh = 1;
  bandProps1[ 2 ]->m_blkw = diskRasterPtr->getNumberOfColumns();
  bandProps1[ 2 ]->m_nblocksx = 1;
  bandProps1[ 2 ]->m_nblocksy = diskRasterPtr->getNumberOfRows();
  bandProps1[ 2 ]->m_type = te::dt::DOUBLE_TYPE;
  
  std::unique_ptr< te::rst::Raster > pcaRasterPtr( te::rst::RasterFactory::make(
    "MEM", new te::rst::Grid( *( diskRasterPtr->getGrid() ) ), bandProps1, 
    std::map<std::string, std::string>(), 0, 0 ) );
  BOOST_REQUIRE( pcaRasterPtr.get() );
  
  std::vector< unsigned int > inputRasterBands;
  inputRasterBands.push_back( 0 );
  inputRasterBands.push_back( 1 );
  inputRasterBands.push_back( 2 );

  std::vector< unsigned int > pcaRasterBands;
  pcaRasterBands.push_back( 0 );
  pcaRasterBands.push_back( 1 );
  pcaRasterBands.push_back( 2 );
  
  boost::numeric::ublas::matrix< double > pcaMatrix;
  
  BOOST_REQUIRE( te::rp::DirectPrincipalComponents(
    *diskRasterPtr, 
    inputRasterBands,
    pcaMatrix,
    *pcaRasterPtr,
    pcaRasterBands,
    1 ) );

  BOOST_REQUIRE( pcaMatrix.size1() == 3 );
  BOOST_REQUIRE( pcaMatrix.size2() == 3 );
    
  BOOST_REQUIRE( te::rp::Copy2DiskRaster( *pcaRasterPtr, "terralib_unittest_rp_functions_DirectPrincipalComponents.tif" ) ) ;
  
  auxRasterInfo.clear();
  auxRasterInfo["URI"] = "terralib_unittest_rp_functions_InversePrincipalComponents.tif";
  
  std::vector< te::rst::BandProperty * > bandProps2;
  bandProps2.push_back( new te::rst::BandProperty( *( diskRasterPtr->getBand( 0 )->getProperty() ) ) );
  bandProps2.push_back( new te::rst::BandProperty( *( diskRasterPtr->getBand( 1 )->getProperty() ) ) );
  bandProps2.push_back( new te::rst::BandProperty( *( diskRasterPtr->getBand( 2 )->getProperty() ) ) );

  std::vector< unsigned int > outputRasterBands;
  outputRasterBands.push_back( 0 );
  outputRasterBands.push_back( 1 );
  outputRasterBands.push_back( 2 );
  
  std::unique_ptr< te::rst::Raster > outDiskRasterPtr( te::rst::RasterFactory::make(
    "GDAL", new te::rst::Grid( *( diskRasterPtr->getGrid() ) ), bandProps2, 
    auxRasterInfo, 0, 0 ) );
  BOOST_REQUIRE( outDiskRasterPtr.get() );
  
  BOOST_REQUIRE( te::rp::InversePrincipalComponents(
    *pcaRasterPtr,
    pcaMatrix,
    *outDiskRasterPtr,
    outputRasterBands,
    1 ) );  
}

BOOST_AUTO_TEST_CASE(getStdDevValue_test)
{
  /* Openning input raster */
  
  std::map<std::string, std::string> auxRasterInfo;
  
  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers2b_hrc_crop.tif";
  std::unique_ptr< te::rst::Raster > diskRasterPtr( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_CHECK( diskRasterPtr.get() );
  
  double stdDevValue = 0;
  
//  boost::timer timer;
  BOOST_CHECK( te::rp::GetStdDevValue( *diskRasterPtr, 0, 1, false, 0, 0, stdDevValue) );
//  std::cout << std::endl << "Elapsed:" << timer.elapsed() << std::endl;
  BOOST_CHECK_CLOSE( 46.302240075, stdDevValue, 0.0001 );
}

BOOST_AUTO_TEST_CASE(getStdDevValueOptimized_test)
{
  /* Openning input raster */
  
  std::map<std::string, std::string> auxRasterInfo;
  
  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers2b_hrc_crop.tif";
  std::unique_ptr< te::rst::Raster > diskRasterPtr( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_CHECK( diskRasterPtr.get() );
  
  double stdDevValue = 0;
  
//  boost::timer timer;
  BOOST_CHECK( te::rp::GetStdDevValue( *diskRasterPtr, 0, 4, false, 0, 0, stdDevValue) );
//  std::cout << std::endl << "Elapsed:" << timer.elapsed() << std::endl;
  BOOST_CHECK_CLOSE( 46.302240075, stdDevValue, 0.0001 );
}

BOOST_AUTO_TEST_CASE(decomposeBands_test)
{
  /* Openning input raster */
  
  std::map<std::string, std::string> auxRasterInfo;
  
  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers2b_rgb342_crop.tif";
  std::unique_ptr< te::rst::Raster > diskRasterPtr( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_CHECK( diskRasterPtr.get() );
  
  std::vector< unsigned int > inputRasterBands;
  inputRasterBands.push_back( 2 );
  inputRasterBands.push_back( 0 );
  
  std::vector< std::map<std::string, std::string> > outputRastersInfos( 2 );
  outputRastersInfos[ 0 ][ "URI" ] = "terralib_unittest_rp_functions_DecomposeBands_2.tif";
  outputRastersInfos[ 1 ][ "URI" ] = "terralib_unittest_rp_functions_DecomposeBands_0.tif";

  std::vector< boost::shared_ptr< te::rst::Raster > > outputRastersPtrs;
  
  BOOST_CHECK( te::rp::DecomposeBands( *diskRasterPtr, inputRasterBands,
    outputRastersInfos, "GDAL", outputRastersPtrs ) );  
}

BOOST_AUTO_TEST_CASE(composeBandsSameSRID_test)
{
  /* Openning input rasters */
  
  std::map<std::string, std::string> auxRasterInfo;
  
  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers_rgb342_crop2.tif";
  std::unique_ptr< te::rst::Raster > diskRasterPtr1( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_CHECK( diskRasterPtr1.get() );

  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers_rgb342_crop3.tif";
  std::unique_ptr< te::rst::Raster > diskRasterPtr2( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_CHECK( diskRasterPtr2.get() );
  
  std::vector< te::rst::Raster const * > rasters;
  rasters.push_back( diskRasterPtr1.get() );
  rasters.push_back( diskRasterPtr2.get() );

  std::vector< unsigned int > inputRasterBands;
  inputRasterBands.push_back( 2 );
  inputRasterBands.push_back( 0 );
  
  std::map<std::string, std::string> outputRasterInfo;
  outputRasterInfo["URI"] = "terralib_unittest_rp_functions_ComposeBandsSameSRID.tif";
  
  std::unique_ptr< te::rst::Raster > outputRasterPtr;
  
  BOOST_CHECK( te::rp::ComposeBands( rasters, inputRasterBands,
    te::rst::NearestNeighbor, outputRasterInfo,
    "GDAL", 1, false, true, outputRasterPtr ) );
}

BOOST_AUTO_TEST_CASE(composeBandsDifSRID_test)
{
  /* Openning input rasters */
  
  std::map<std::string, std::string> auxRasterInfo;
  
  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers_rgb342_crop2.tif";
  std::unique_ptr< te::rst::Raster > diskRasterPtr1( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_CHECK( diskRasterPtr1.get() );

  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers_rgb342_crop3_EPSG_22522.tif";
  std::unique_ptr< te::rst::Raster > diskRasterPtr2( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_CHECK( diskRasterPtr2.get() );
  
  std::vector< te::rst::Raster const * > rasters;
  rasters.push_back( diskRasterPtr1.get() );
  rasters.push_back( diskRasterPtr2.get() );
  
  std::vector< unsigned int > inputRasterBands;
  inputRasterBands.push_back( 2 );
  inputRasterBands.push_back( 0 );
  
  std::map<std::string, std::string> outputRasterInfo;
  outputRasterInfo["URI"] = "terralib_unittest_rp_functions_ComposeBandsDifSRID.tif";
  
  std::unique_ptr< te::rst::Raster > outputRasterPtr;
  
  BOOST_CHECK( te::rp::ComposeBands( rasters, inputRasterBands,
    te::rst::NearestNeighbor, outputRasterInfo,
    "GDAL", 1, false, true, outputRasterPtr ) );  
}

BOOST_AUTO_TEST_CASE(composeBandsZeroSRID_test)
{
  /* Openning input rasters */

  std::map<std::string, std::string> auxRasterInfo;

  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers_b2_crop_contrast_halfsampled_90rotation.tif";
  std::unique_ptr< te::rst::Raster > diskRasterPtr1( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_CHECK( diskRasterPtr1.get() );

  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers_b2_crop_contrast_halfsampled_90rotation.tif";
  std::unique_ptr< te::rst::Raster > diskRasterPtr2( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_CHECK( diskRasterPtr2.get() );

  std::vector< te::rst::Raster const * > rasters;
  rasters.push_back( diskRasterPtr1.get() );
  rasters.push_back( diskRasterPtr2.get() );

  std::vector< unsigned int > inputRasterBands;
  inputRasterBands.push_back( 0 );
  inputRasterBands.push_back( 0 );

  std::map<std::string, std::string> outputRasterInfo;
  outputRasterInfo["URI"] = "terralib_unittest_rp_functions_composeBandsZeroSRID_test.tif";

  std::unique_ptr< te::rst::Raster > outputRasterPtr;

  BOOST_CHECK( te::rp::ComposeBands( rasters, inputRasterBands,
    te::rst::NearestNeighbor, outputRasterInfo,
    "GDAL", 1, false, true, outputRasterPtr ) );
}

BOOST_AUTO_TEST_CASE(composeBandsDifSRIDMultiThread_test)
{
  /* Openning input rasters */
  
  std::map<std::string, std::string> auxRasterInfo;
  
  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers_rgb342_crop2.tif";
  std::unique_ptr< te::rst::Raster > diskRasterPtr1( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_CHECK( diskRasterPtr1.get() );

  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers_rgb342_crop3_EPSG_22522.tif";
  std::unique_ptr< te::rst::Raster > diskRasterPtr2( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_CHECK( diskRasterPtr2.get() );
  
  std::vector< te::rst::Raster const * > rasters;
  rasters.push_back( diskRasterPtr1.get() );
  rasters.push_back( diskRasterPtr2.get() );

  std::vector< unsigned int > inputRasterBands;
  inputRasterBands.push_back( 2 );
  inputRasterBands.push_back( 0 );
  
  std::map<std::string, std::string> outputRasterInfo;
  outputRasterInfo["URI"] = "terralib_unittest_rp_functions_ComposeBandsDifSRIDMultiThread.tif";
  
  std::unique_ptr< te::rst::Raster > outputRasterPtr;
  
  BOOST_CHECK( te::rp::ComposeBands( rasters, inputRasterBands,
    te::rst::NearestNeighbor, outputRasterInfo,
    "GDAL", 0, false, true, outputRasterPtr ) );  
}

BOOST_AUTO_TEST_CASE(getDetailedExtent_test)
{
  te::rst::Grid grid( 2u, 2u, new te::gm::Envelope( 0.0, 0.0, 2.0, 2.0 ), 12345 );
  
  te::gm::LinearRing lr( te::gm::LineStringType, 0, 0 );
  
  BOOST_CHECK( te::rp::GetDetailedExtent( grid, lr ) );
  BOOST_CHECK( lr.size() == 9 );
  
  BOOST_CHECK_CLOSE( lr.getX( 0 ), 0.0 , 0.0000000001 );
  BOOST_CHECK_CLOSE( lr.getY( 0 ), 2.0 , 0.0000000001 );
  
  BOOST_CHECK_CLOSE( lr.getX( 1 ), 1.0 , 0.0000000001 );
  BOOST_CHECK_CLOSE( lr.getY( 1 ), 2.0 , 0.0000000001 );

  BOOST_CHECK_CLOSE( lr.getX( 2 ), 2.0 , 0.0000000001 );
  BOOST_CHECK_CLOSE( lr.getY( 2 ), 2.0 , 0.0000000001 );
  
  BOOST_CHECK_CLOSE( lr.getX( 3 ), 2.0 , 0.0000000001 );
  BOOST_CHECK_CLOSE( lr.getY( 3 ), 1.0 , 0.0000000001 );
  
  BOOST_CHECK_CLOSE( lr.getX( 4 ), 2.0 , 0.0000000001 );
  BOOST_CHECK_CLOSE( lr.getY( 4 ), 0.0 , 0.0000000001 );
  
  BOOST_CHECK_CLOSE( lr.getX( 5 ), 1.0 , 0.0000000001 );
  BOOST_CHECK_CLOSE( lr.getY( 5 ), 0.0 , 0.0000000001 );
  
  BOOST_CHECK_CLOSE( lr.getX( 6 ), 0.0 , 0.0000000001 );
  BOOST_CHECK_CLOSE( lr.getY( 6 ), 0.0 , 0.0000000001 );
  
  BOOST_CHECK_CLOSE( lr.getX( 7 ), 0.0 , 0.0000000001 );
  BOOST_CHECK_CLOSE( lr.getY( 7 ), 1.0 , 0.0000000001 );
  
  BOOST_CHECK_CLOSE( lr.getX( 8 ), 0.0 , 0.0000000001 );
  BOOST_CHECK_CLOSE( lr.getY( 8 ), 2.0 , 0.0000000001 );
  
  BOOST_CHECK_CLOSE( grid.getExtent()->m_llx, 0.0 , 0.0000000001 );
  BOOST_CHECK_CLOSE( grid.getExtent()->m_lly, 0.0 , 0.0000000001 );
  BOOST_CHECK_CLOSE( grid.getExtent()->m_urx, 2.0 , 0.0000000001 );
  BOOST_CHECK_CLOSE( grid.getExtent()->m_ury, 2.0 , 0.0000000001 );
}

BOOST_AUTO_TEST_CASE(getIndexedDetailedExtent_test)
{
  te::rst::Grid grid( 2u, 2u, new te::gm::Envelope( 0.0, 0.0, 2.0, 2.0 ), 12345 );
  
  te::gm::LinearRing lr( te::gm::LineStringType, 0, 0 );
  
  BOOST_CHECK( te::rp::GetIndexedDetailedExtent( grid, lr ) );
  BOOST_CHECK( lr.size() == 9 );
  
  BOOST_CHECK_CLOSE( lr.getX( 0 ), -0.5 , 0.0000000001 );
  BOOST_CHECK_CLOSE( lr.getY( 0 ), -0.5 , 0.0000000001 );
  
  BOOST_CHECK_CLOSE( lr.getX( 1 ), 0.5 , 0.0000000001 );
  BOOST_CHECK_CLOSE( lr.getY( 1 ), -0.5 , 0.0000000001 );

  BOOST_CHECK_CLOSE( lr.getX( 2 ), 1.5 , 0.0000000001 );
  BOOST_CHECK_CLOSE( lr.getY( 2 ), -0.5 , 0.0000000001 );
  
  BOOST_CHECK_CLOSE( lr.getX( 3 ), 1.5 , 0.0000000001 );
  BOOST_CHECK_CLOSE( lr.getY( 3 ), 0.5 , 0.0000000001 );
  
  BOOST_CHECK_CLOSE( lr.getX( 4 ), 1.5 , 0.0000000001 );
  BOOST_CHECK_CLOSE( lr.getY( 4 ), 1.5 , 0.0000000001 );
  
  BOOST_CHECK_CLOSE( lr.getX( 5 ), 0.5 , 0.0000000001 );
  BOOST_CHECK_CLOSE( lr.getY( 5 ), 1.5 , 0.0000000001 );
  
  BOOST_CHECK_CLOSE( lr.getX( 6 ), -0.5 , 0.0000000001 );
  BOOST_CHECK_CLOSE( lr.getY( 6 ), 1.5 , 0.0000000001 );
  
  BOOST_CHECK_CLOSE( lr.getX( 7 ), -0.5 , 0.0000000001 );
  BOOST_CHECK_CLOSE( lr.getY( 7 ), 0.5 , 0.0000000001 );
  
  BOOST_CHECK_CLOSE( lr.getX( 8 ), -0.5 , 0.0000000001 );
  BOOST_CHECK_CLOSE( lr.getY( 8 ), -0.5 , 0.0000000001 );
  
  BOOST_CHECK_CLOSE( grid.getExtent()->m_llx, 0.0 , 0.0000000001 );
  BOOST_CHECK_CLOSE( grid.getExtent()->m_lly, 0.0 , 0.0000000001 );
  BOOST_CHECK_CLOSE( grid.getExtent()->m_urx, 2.0 , 0.0000000001 );
  BOOST_CHECK_CLOSE( grid.getExtent()->m_ury, 2.0 , 0.0000000001 );  
}

BOOST_AUTO_TEST_CASE(waveletAtrous_test)
{
  const unsigned int waveletDecompLevelsNumb = 3;
  
  /* Openning input raster */
  
  std::map<std::string, std::string> auxRasterInfo;
  
  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers2b_rgb342_crop.tif";
  std::unique_ptr< te::rst::Raster > diskRasterPtr( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_CHECK( diskRasterPtr.get() );
  
  std::vector< te::rst::BandProperty * > waveletRasterBandProps;
  for( unsigned int bandIdx = 0 ; bandIdx < diskRasterPtr->getNumberOfBands() ;
    ++bandIdx )
  {
    for( unsigned int levelIdx = 0 ; levelIdx < waveletDecompLevelsNumb ;
      ++levelIdx )
    {
      waveletRasterBandProps.push_back( new te::rst::BandProperty( 
        *( diskRasterPtr->getBand( 0 )->getProperty() ) ) );
      waveletRasterBandProps.back()->m_blkh = diskRasterPtr->getNumberOfRows();
      waveletRasterBandProps.back()->m_blkw = diskRasterPtr->getNumberOfColumns();
      waveletRasterBandProps.back()->m_nblocksx = 1;
      waveletRasterBandProps.back()->m_nblocksy = 1;
      waveletRasterBandProps.back()->m_type = te::dt::DOUBLE_TYPE;
      
      waveletRasterBandProps.push_back( new te::rst::BandProperty( 
        *( diskRasterPtr->getBand( 0 )->getProperty() ) ) );
      waveletRasterBandProps.back()->m_blkh = diskRasterPtr->getNumberOfRows();
      waveletRasterBandProps.back()->m_blkw = diskRasterPtr->getNumberOfColumns();
      waveletRasterBandProps.back()->m_nblocksx = 1;
      waveletRasterBandProps.back()->m_nblocksy = 1;
      waveletRasterBandProps.back()->m_type = te::dt::DOUBLE_TYPE;      
    }
  }
  
  std::unique_ptr< te::rst::Raster > waveletRasterPtr( te::rst::RasterFactory::make(
    "MEM", new te::rst::Grid( *( diskRasterPtr->getGrid() ) ), waveletRasterBandProps, 
    std::map<std::string, std::string>(), 0, 0 ) );
  BOOST_CHECK( waveletRasterPtr.get() );
  
  std::vector< unsigned int > inputRasterBands;
  inputRasterBands.push_back( 0 );
  inputRasterBands.push_back( 1 );
  inputRasterBands.push_back( 2 );
  
  boost::numeric::ublas::matrix< double > filter = te::rp::CreateWaveletAtrousFilter( 
    te::rp::B3SplineWAFilter );  
  
  BOOST_CHECK( te::rp::DirectWaveletAtrous(
    *diskRasterPtr, 
    inputRasterBands,
    *waveletRasterPtr,
    waveletDecompLevelsNumb,
    filter ) );
    
  BOOST_CHECK( te::rp::Copy2DiskRaster( *waveletRasterPtr, "terralib_unittest_rp_functions_DirectWaveletAtrous.tif" ) ) ;
  
  auxRasterInfo.clear();
  auxRasterInfo["URI"] = "terralib_unittest_rp_functions_InverseWaveletAtrous.tif";
  
  std::vector< te::rst::BandProperty * > bandProps2;
  bandProps2.push_back( new te::rst::BandProperty( *( diskRasterPtr->getBand( 0 )->getProperty() ) ) );
  bandProps2.push_back( new te::rst::BandProperty( *( diskRasterPtr->getBand( 1 )->getProperty() ) ) );
  bandProps2.push_back( new te::rst::BandProperty( *( diskRasterPtr->getBand( 2 )->getProperty() ) ) );
  
  std::unique_ptr< te::rst::Raster > outDiskRasterPtr( te::rst::RasterFactory::make(
    "GDAL", new te::rst::Grid( *( diskRasterPtr->getGrid() ) ), bandProps2, 
    auxRasterInfo, 0, 0 ) );
  BOOST_CHECK( outDiskRasterPtr.get() );
  
  BOOST_CHECK( te::rp::InverseWaveletAtrous(
    *waveletRasterPtr,
    waveletDecompLevelsNumb,
    *outDiskRasterPtr,
    inputRasterBands ) );  
}

BOOST_AUTO_TEST_CASE(SelectiveReplaceValues_test)
{
  /* Openning input raster */

  std::map<std::string, std::string> auxRasterInfo;

  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers2b_rgb342_crop.tif";
  std::unique_ptr< te::rst::Raster > inputRasterPtr( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_CHECK( inputRasterPtr.get() );

  // no restriction polygons test

  {
    std::vector< te::rst::BandProperty * > bandProps;
    bandProps.push_back( new te::rst::BandProperty(
      *( inputRasterPtr->getBand( 0 )->getProperty() ) ) );

    std::unique_ptr< te::rst::Raster > memRasterPtr( te::rst::RasterFactory::make(
      "MEM", new te::rst::Grid( *( inputRasterPtr->getGrid() ) ),
      bandProps, std::map<std::string, std::string>(), 0, 0 ) );
    BOOST_CHECK( memRasterPtr.get() );

    std::vector< std::pair< double, double > > targetValues;
    targetValues.push_back( std::pair< double, double >( 0, 0 ) );

    BOOST_CHECK( te::rp::SelectiveReplaceValues( *inputRasterPtr, 0, targetValues,
      true, std::vector< te::gm::Polygon* >(), *memRasterPtr, 0 ) );

    BOOST_CHECK( te::rp::Copy2DiskRaster( *memRasterPtr,
      "terralib_unittest_rp_functions_SelectiveReplaceValuesNoRestriction.tif" ) ) ;
  }

  // with restriction polygon test

  {
    std::unique_ptr< te::gm::Polygon > polPtr( (te::gm::Polygon*)
      te::gm::GetGeomFromEnvelope( inputRasterPtr->getGrid()->getExtent(),
      inputRasterPtr->getSRID() ) );

    std::vector< te::gm::Polygon* > restrictionPols;
    restrictionPols.push_back( polPtr.get() );

    std::vector< te::rst::BandProperty * > bandProps;
    bandProps.push_back( new te::rst::BandProperty(
      *( inputRasterPtr->getBand( 0 )->getProperty() ) ) );

    std::unique_ptr< te::rst::Raster > memRasterPtr( te::rst::RasterFactory::make(
      "MEM", new te::rst::Grid( *( inputRasterPtr->getGrid() ) ),
      bandProps, std::map<std::string, std::string>(), 0, 0 ) );
    BOOST_CHECK( memRasterPtr.get() );

    std::vector< std::pair< double, double > > targetValues;
    targetValues.push_back( std::pair< double, double >( 0, 0 ) );

    BOOST_CHECK( te::rp::SelectiveReplaceValues( *inputRasterPtr, 0, targetValues,
      true, restrictionPols, *memRasterPtr, 0 ) );

    BOOST_CHECK( te::rp::Copy2DiskRaster( *memRasterPtr,
      "terralib_unittest_rp_functions_SelectiveReplaceValuesWithRestriction.tif" ) ) ;
  }
}

BOOST_AUTO_TEST_CASE (getHistograms_SingleThread_test)
{
  /* Openning input raster */

  std::map<std::string, std::string> auxRasterInfo;

  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers_rgb342_crop1.tif";
  boost::shared_ptr< te::rst::Raster > inputRasterPtr ( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_CHECK( inputRasterPtr.get() );
  
  // Restriction polygon 
  
  std::unique_ptr< te::gm::Polygon > polPtr( (te::gm::Polygon*)
    te::gm::GetGeomFromEnvelope( inputRasterPtr->getGrid()->getExtent(),
    inputRasterPtr->getSRID() ) );

  std::vector< te::gm::Polygon const* > restrictionPols;
  restrictionPols.push_back( polPtr.get() );    

  // Single thread test

  std::map<double, unsigned> rHistogram;
  std::map<double, unsigned> iHistogram;

  BOOST_CHECK( te::rp::getHistograms( 
    *inputRasterPtr, 
    0, // bandIndex
    restrictionPols, 
    0, // histoBins
    1, // maxThreads
    te::rst::PolygonIterator< double >::ScanLineIterationT,
    rHistogram, 
    iHistogram ) );
  BOOST_CHECK_EQUAL( rHistogram.size(), 161 );
  BOOST_CHECK_EQUAL( iHistogram.size(), 1 );
}

BOOST_AUTO_TEST_CASE (getHistograms_MultiThread_test)
{
  /* Openning input raster */

  std::map<std::string, std::string> auxRasterInfo;

  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers_rgb342_crop1.tif";
  boost::shared_ptr< te::rst::Raster > inputRasterPtr ( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_CHECK( inputRasterPtr.get() );
  
  // Restriction polygon 
  
  std::unique_ptr< te::gm::Polygon > polPtr( (te::gm::Polygon*)
    te::gm::GetGeomFromEnvelope( inputRasterPtr->getGrid()->getExtent(),
    inputRasterPtr->getSRID() ) );

  std::vector< te::gm::Polygon const* > restrictionPols;
  restrictionPols.push_back( polPtr.get() );      

  // Muilti thread test

  std::map<double, unsigned> rHistogram;
  std::map<double, unsigned> iHistogram;

  BOOST_CHECK( te::rp::getHistograms( 
    *inputRasterPtr, 
    0,  // bandIndex
    restrictionPols, 
    0, // histoBins
    1, //maxThreads
    te::rst::PolygonIterator< double >::ScanLineIterationT,
    rHistogram, 
    iHistogram ) );
  BOOST_CHECK_EQUAL( rHistogram.size(), 161 );
  BOOST_CHECK_EQUAL( iHistogram.size(), 1 );
}

BOOST_AUTO_TEST_CASE (getHistograms_PredefinedBins_test)
{
  /* Openning input raster */

  std::map<std::string, std::string> auxRasterInfo;

  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers_rgb342_crop1.tif";
  boost::shared_ptr< te::rst::Raster > inputRasterPtr ( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_CHECK( inputRasterPtr.get() );
  
  // Restriction polygon 
  
  std::unique_ptr< te::gm::Polygon > polPtr( (te::gm::Polygon*)
    te::gm::GetGeomFromEnvelope( inputRasterPtr->getGrid()->getExtent(),
    inputRasterPtr->getSRID() ) );

  std::vector< te::gm::Polygon const* > restrictionPols;
  restrictionPols.push_back( polPtr.get() );      

  // predefined bins test

  std::map<double, unsigned> rHistogram;
  std::map<double, unsigned> iHistogram;

  BOOST_CHECK( te::rp::getHistograms( 
    *inputRasterPtr, 
    0, // bandIndex
    restrictionPols, 
    100, /* histoBins */
    1, // maxThreads
    te::rst::PolygonIterator< double >::ScanLineIterationT,
    rHistogram, 
    iHistogram ) );
  BOOST_CHECK_EQUAL( rHistogram.size(), 100 );
  BOOST_CHECK_EQUAL( iHistogram.size(), 1 );
}

BOOST_AUTO_TEST_CASE (getHistograms_DoubleData_test)
{
  /* Openning input raster */

  std::map<std::string, std::string> auxRasterInfo;

  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers2b_rgb342_crop_linearMixtureModel_double.tif";
  boost::shared_ptr< te::rst::Raster > inputRasterPtr ( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_CHECK( inputRasterPtr.get() );
  
  // Restriction polygon 
  
  std::unique_ptr< te::gm::Polygon > polPtr( (te::gm::Polygon*)
    te::gm::GetGeomFromEnvelope( inputRasterPtr->getGrid()->getExtent(),
    inputRasterPtr->getSRID() ) );

  std::vector< te::gm::Polygon const* > restrictionPols;
  restrictionPols.push_back( polPtr.get() );      

  // Single thread test

  std::map<double, unsigned> rHistogram;
  std::map<double, unsigned> iHistogram;

  BOOST_CHECK( te::rp::getHistograms( 
    *inputRasterPtr, 
    0, // bandIndex
    restrictionPols, 
    0, /* histoBins */
    1, /* maxThreads */
    te::rst::PolygonIterator< double >::ScanLineIterationT,
    rHistogram, 
    iHistogram ) );
  BOOST_CHECK_EQUAL( rHistogram.size(), 256 );
  BOOST_CHECK_EQUAL( iHistogram.size(), 1 );
}

BOOST_AUTO_TEST_CASE (getHistograms_DoubleDataPreDefinedBins_test)
{
  /* Openning input raster */

  std::map<std::string, std::string> auxRasterInfo;

  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers2b_rgb342_crop_linearMixtureModel_double.tif";
  boost::shared_ptr< te::rst::Raster > inputRasterPtr ( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_CHECK( inputRasterPtr.get() );
  
  // Restriction polygon 
  
  std::unique_ptr< te::gm::Polygon > polPtr( (te::gm::Polygon*)
    te::gm::GetGeomFromEnvelope( inputRasterPtr->getGrid()->getExtent(),
    inputRasterPtr->getSRID() ) );

  std::vector< te::gm::Polygon const* > restrictionPols;
  restrictionPols.push_back( polPtr.get() );      

  // Single thread test

  std::map<double, unsigned> rHistogram;
  std::map<double, unsigned> iHistogram;

  BOOST_CHECK( te::rp::getHistograms( 
    *inputRasterPtr, 
    0, // bandIndex
    restrictionPols, 
    100, /* histoBins */
    1, /* maxThreads */
    te::rst::PolygonIterator< double >::ScanLineIterationT,
    rHistogram, 
    iHistogram ) );
  BOOST_CHECK_EQUAL( rHistogram.size(), 100 );
  BOOST_CHECK_EQUAL( iHistogram.size(), 1 );
}

BOOST_AUTO_TEST_CASE (getHistograms_ComplexData_test)
{
  /* Openning input raster */

  std::map<std::string, std::string> auxRasterInfo;

  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/te_complex.tif";
  boost::shared_ptr< te::rst::Raster > inputRasterPtr ( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_CHECK( inputRasterPtr.get() );
  
  // Restriction polygon 
  
  std::unique_ptr< te::gm::Polygon > polPtr( (te::gm::Polygon*)
    te::gm::GetGeomFromEnvelope( inputRasterPtr->getGrid()->getExtent(),
    inputRasterPtr->getSRID() ) );

  std::vector< te::gm::Polygon const* > restrictionPols;
  restrictionPols.push_back( polPtr.get() );      

  // Single thread test

  std::map<double, unsigned> rHistogram;
  std::map<double, unsigned> iHistogram;

  BOOST_CHECK( te::rp::getHistograms( 
    *inputRasterPtr, 
    0, // bandIndex
    restrictionPols, 
    0, // histoBins
    1, /* maxThreads */
    te::rst::PolygonIterator< double >::ScanLineIterationT,
    rHistogram, 
    iHistogram ) );
  BOOST_CHECK_EQUAL( rHistogram.size(), 229 );
  BOOST_CHECK_EQUAL( iHistogram.size(), 240 );
}

BOOST_AUTO_TEST_CASE (getHistogramStats_test)
{
  /* Openning input raster */

  std::map<std::string, std::string> auxRasterInfo;

  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers_rgb342_crop1.tif";
  boost::shared_ptr< te::rst::Raster > inputRasterPtr ( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_CHECK( inputRasterPtr.get() );
  
  // Restriction polygon 
  
  std::unique_ptr< te::gm::Polygon > polPtr( (te::gm::Polygon*)
    te::gm::GetGeomFromEnvelope( inputRasterPtr->getGrid()->getExtent(),
    inputRasterPtr->getSRID() ) );

  std::vector< te::gm::Polygon const* > restrictionPols;
  restrictionPols.push_back( polPtr.get() );       

  // Single thread test

  std::map<double, unsigned> rHistogram;
  std::map<double, unsigned> iHistogram;
  std::vector< te::gm::Polygon const* > emptyROIS;

  BOOST_CHECK( te::rp::getHistograms( 
    *inputRasterPtr, 
    2, // bandIndex
    restrictionPols, 
    0, // histoBins
    1, // maxThreads
    te::rst::PolygonIterator< double >::ScanLineIterationT,
    rHistogram, 
    iHistogram ) );
  BOOST_CHECK_EQUAL( rHistogram.size(), 214 );

  double min = 0;
  double max = 0;
  double mean = 0;
  double stdDev = 0;
  double mode = 0;
  double entropy = 0;
  double sum = 0;
  double sum2 = 0;
  double sum3 = 0;
  double sum4 = 0;
  double variance = 0;
  double median = 0;

  te::rp::getHistogramStats( rHistogram, min, max, mean, stdDev, mode, entropy,
    sum, sum2, sum3, sum4, variance, median );
  BOOST_REQUIRE_CLOSE( min, 42, 0.1 );
  BOOST_REQUIRE_CLOSE( max, 255, 0.1 );
  BOOST_REQUIRE_CLOSE( mean, 76.6986, 0.1 );
  BOOST_REQUIRE_CLOSE( stdDev, 11.16, 0.1 );  
  BOOST_REQUIRE_CLOSE( mode, 75, 0.1 );
  BOOST_REQUIRE_CLOSE( entropy, 5.0542, 0.1 );
  BOOST_REQUIRE_CLOSE( sum, 67715341, 0.1 );
  BOOST_REQUIRE_CLOSE( sum2, 5303786357, 0.1 );
  BOOST_REQUIRE_CLOSE( sum3, 430682720623, 0.1 );
  BOOST_REQUIRE_CLOSE( sum4, 37675811615417, 0.1 );
  BOOST_REQUIRE_CLOSE( variance, 124.7170378541913, 0.1 );
  BOOST_REQUIRE_CLOSE( median, 149, 0.1 );

  te::rp::getHistogramStats( iHistogram, min, max, mean, stdDev, mode, entropy,
    sum, sum2, sum3, sum4, variance, median );
  BOOST_REQUIRE_CLOSE( min, 0, 0.00001 );
  BOOST_REQUIRE_CLOSE( max, 0, 0.00001 );
  BOOST_REQUIRE_CLOSE( mean, 0, 0.00001 );
  BOOST_REQUIRE_CLOSE( stdDev, 0, 0.00001 );
  BOOST_REQUIRE_CLOSE( mode, 0, 0.1 );
  BOOST_REQUIRE_CLOSE( entropy, 0, 0.1 );
  BOOST_REQUIRE_CLOSE( sum, 0, 0.1 );
  BOOST_REQUIRE_CLOSE( sum2, 0, 0.1 );
  BOOST_REQUIRE_CLOSE( sum3, 0, 0.1 );
  BOOST_REQUIRE_CLOSE( sum4, 0, 0.1 );
  BOOST_REQUIRE_CLOSE( variance, 0, 0.1 );
  BOOST_REQUIRE_CLOSE( median, 0, 0.1 );
}

BOOST_AUTO_TEST_CASE (getJoinHistograms_SingleThread_test)
{
  /* Openning input raster */

  std::map<std::string, std::string> auxRasterInfo;

  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers_rgb342_crop1.tif";
  boost::shared_ptr< te::rst::Raster > inputRasterPtr ( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_CHECK( inputRasterPtr.get() );
  
  // Restriction polygon 
  
  std::unique_ptr< te::gm::Polygon > polPtr( (te::gm::Polygon*)
    te::gm::GetGeomFromEnvelope( inputRasterPtr->getGrid()->getExtent(),
    inputRasterPtr->getSRID() ) );

  std::vector< te::gm::Polygon const* > restrictionPols;
  restrictionPols.push_back( polPtr.get() );    

  // Single thread test

  std::map< std::pair< double, double >, unsigned int> realJointHistogram;
  std::map< std::pair< double, double >, unsigned int> imagJointHistogram;

  BOOST_CHECK( te::rp::getJointHistograms( 
    *inputRasterPtr, 
    0, // bandIndex1
    1, // bandIndex2
    restrictionPols, 
    0, // histoBins
    1, // maxThreads
    te::rst::PolygonIterator< double >::ScanLineIterationT,
    realJointHistogram, 
    imagJointHistogram,
    0, // rasterRealHistogram1ptr
    0, // rasterImagHistogram1Ptr
    0, // rasterRealHistogram2Ptr
    0 // rasterImagHistogram2Ptr
  ) );
  
  BOOST_CHECK_EQUAL( realJointHistogram.size(), 8344 );
  BOOST_CHECK_EQUAL( imagJointHistogram.size(), 1 );
}

BOOST_AUTO_TEST_CASE (getJoinHistograms_MultiThread_test)
{
  /* Openning input raster */

  std::map<std::string, std::string> auxRasterInfo;

  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers_rgb342_crop1.tif";
  boost::shared_ptr< te::rst::Raster > inputRasterPtr ( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_CHECK( inputRasterPtr.get() );
  
  // Restriction polygon 
  
  std::unique_ptr< te::gm::Polygon > polPtr( (te::gm::Polygon*)
    te::gm::GetGeomFromEnvelope( inputRasterPtr->getGrid()->getExtent(),
    inputRasterPtr->getSRID() ) );

  std::vector< te::gm::Polygon const* > restrictionPols;
  restrictionPols.push_back( polPtr.get() );    

  // Single thread test

  std::map< std::pair< double, double >, unsigned int> realJointHistogram;
  std::map< std::pair< double, double >, unsigned int> imagJointHistogram;

  BOOST_CHECK( te::rp::getJointHistograms( 
    *inputRasterPtr, 
    0, // bandIndex1
    1, // bandIndex2
    restrictionPols, 
    0, // histoBins
    4, // maxThreads
    te::rst::PolygonIterator< double >::ScanLineIterationT,
    realJointHistogram, 
    imagJointHistogram,
    0, // rasterRealHistogram1ptr
    0, // rasterImagHistogram1Ptr
    0, // rasterRealHistogram2Ptr
    0 // rasterImagHistogram2Ptr
  ) );
  
  BOOST_CHECK_EQUAL( realJointHistogram.size(), 8344 );
  BOOST_CHECK_EQUAL( imagJointHistogram.size(), 1 );
}

BOOST_AUTO_TEST_CASE (getJointHistogramStats_test)
{
  /* Openning input raster */

  std::map<std::string, std::string> auxRasterInfo;

  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers_rgb342_crop1.tif";
  boost::shared_ptr< te::rst::Raster > inputRasterPtr ( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_CHECK( inputRasterPtr.get() );
  
  std::unique_ptr< te::gm::Polygon > polPtr( (te::gm::Polygon*)
    te::gm::GetGeomFromEnvelope( inputRasterPtr->getGrid()->getExtent(),
    inputRasterPtr->getSRID() ) );

  std::vector< te::gm::Polygon const* > restrictionPols;
  restrictionPols.push_back( polPtr.get() );    

  // regular histograms
  
  std::map<double, unsigned> rHistogram1;
  std::map<double, unsigned> iHistogram1;
  BOOST_CHECK( te::rp::getHistograms( 
    *inputRasterPtr, 
    0, // bandIndex
    restrictionPols, 
    256, // histoBins
    1, // maxThreads
    te::rst::PolygonIterator< double >::ScanLineIterationT,
    rHistogram1, 
    iHistogram1 ) );
  
  std::map<double, unsigned> rHistogram2;
  std::map<double, unsigned> iHistogram2;
  BOOST_CHECK( te::rp::getHistograms( 
    *inputRasterPtr, 
    1, // bandIndex
    restrictionPols, 
    256, // histoBins
    1, // maxThreads
    te::rst::PolygonIterator< double >::ScanLineIterationT,
    rHistogram2, 
    iHistogram2 ) );
  
  // joint histograms

  std::map< std::pair< double, double >, unsigned int> realJointHistogram;
  std::map< std::pair< double, double >, unsigned int> imagJointHistogram;

  BOOST_CHECK( te::rp::getJointHistograms( 
    *inputRasterPtr, 
    0, // bandIndex1
    1, // bandIndex2
    restrictionPols, 
    256, // histoBins
    1, // maxThreads
    te::rst::PolygonIterator< double >::ScanLineIterationT,
    realJointHistogram, 
    imagJointHistogram,
    &rHistogram1,
    &iHistogram1,
    &rHistogram2,
    &iHistogram2
  ) );

  double covariance = 0;
  double corrlationCoef = 0;

  te::rp::getJointHistogramStats( realJointHistogram, rHistogram1, rHistogram2, 
    covariance, corrlationCoef );
  BOOST_REQUIRE_CLOSE( covariance, 47.57555053051211, 0.1 );
  BOOST_REQUIRE_CLOSE( corrlationCoef, 0.17552410279503561, 0.1 );
}

BOOST_AUTO_TEST_CASE(RemapValues_test)
{
  /* Openning input raster */
  
  std::map<std::string, std::string> auxRasterInfo;
  
  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers2b_rgb342_crop.tif";
  std::unique_ptr< te::rst::Raster > diskRasterPtr( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_CHECK( diskRasterPtr.get() );
  
  auxRasterInfo.clear();
  auxRasterInfo["MEM_RASTER_NROWS"] = boost::lexical_cast< std::string >( 
    diskRasterPtr->getNumberOfRows() );
  auxRasterInfo["MEM_RASTER_NCOLS"] = boost::lexical_cast< std::string >( 
    diskRasterPtr->getNumberOfColumns() );
  auxRasterInfo["MEM_RASTER_DATATYPE"] = boost::lexical_cast< std::string >((int)te::dt::DOUBLE_TYPE);
  auxRasterInfo["MEM_RASTER_NBANDS"] = "3";
   
  std::unique_ptr< te::rst::Raster > outputRasterPtr( te::rst::RasterFactory::make(
    "MEM", 0, std::vector<te::rst::BandProperty*>(), auxRasterInfo) );
  BOOST_REQUIRE( outputRasterPtr.get() );
  
  boost::numeric::ublas::matrix< double > remapMatrix( 3, 3 );
  remapMatrix( 0, 0 ) = 1;
  remapMatrix( 0, 1 ) = 0;
  remapMatrix( 0, 2 ) = 0;
  remapMatrix( 1, 0 ) = 0;
  remapMatrix( 1, 1 ) = 1;
  remapMatrix( 1, 2 ) = 0;
  remapMatrix( 2, 0 ) = 0;
  remapMatrix( 2, 1 ) = 0;
  remapMatrix( 2, 2 ) = 1;
  
  std::vector< unsigned int > inputRasterBands;
  inputRasterBands.push_back( 0 );
  inputRasterBands.push_back( 1 );
  inputRasterBands.push_back( 2 );

  BOOST_CHECK( te::rp::RemapValues( *diskRasterPtr, inputRasterBands,
    remapMatrix, *outputRasterPtr, inputRasterBands, 1 ) );
  
  BOOST_CHECK( te::rp::Copy2DiskRaster( *outputRasterPtr, 
    "terralib_unittest_rp_functions_RemapValues.tif" ) ) ;
}

BOOST_AUTO_TEST_CASE (GetExternalValidDataPolygon_test)
{
  /* Openning input raster */

  std::map<std::string, std::string> auxRasterInfo;

  auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/L71218076_07620060814_r3g2b1.tif";
  boost::shared_ptr< te::rst::Raster > inputRasterPtr ( te::rst::RasterFactory::open(
    auxRasterInfo ) );
  BOOST_REQUIRE( inputRasterPtr.get() );
  
  std::vector< unsigned int > inputRasterBandsIdx;
  inputRasterBandsIdx.push_back( 0 );
  inputRasterBandsIdx.push_back( 1 );
  inputRasterBandsIdx.push_back( 2 );
  
  std::vector< double > bandNoDataValues;
  bandNoDataValues.push_back( 0 );
  bandNoDataValues.push_back( 0 );
  bandNoDataValues.push_back( 0 );
  
  std::unique_ptr< te::gm::Polygon > outPolygonPtr;
  BOOST_REQUIRE( te::rp::GetExternalValidDataPolygon( *inputRasterPtr, inputRasterBandsIdx,
    bandNoDataValues, outPolygonPtr ) );
  BOOST_REQUIRE( outPolygonPtr.get() != nullptr );
  BOOST_REQUIRE( outPolygonPtr->isValid() );
  
  std::vector<te::gm::Polygon*> polygons;
  polygons.push_back( outPolygonPtr.get() );
  
  std::vector< double > polygonsValues( 1, 0 );
  
  saveToShp( polygons, polygonsValues, "GetExternalValidDataPolygon_test.shp");
}

BOOST_AUTO_TEST_SUITE_END()
