/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/unittest/rp/mixture_model/TsMixtureModel.cpp

  \brief A test suit for the Mixture Model interface.
*/

// TerraLib
#include "../Config.h"
#include <terralib/rp.h>
#include <terralib/raster.h>

// Boost
#define BOOST_TEST_NO_MAIN
#include <boost/test/unit_test.hpp>

#include <memory>

BOOST_AUTO_TEST_SUITE (mixtureModel_tests)

BOOST_AUTO_TEST_CASE(linear_LST_test)
{
  /* First open the input image */

  std::map<std::string, std::string> rinfo;
  rinfo["URI"] = TERRALIB_DATA_DIR"/geotiff/AMAZONIA_1_WFI_20210903_035_019_L4_BANDS1234_clip_sjc.tif";

  std::unique_ptr< te::rst::Raster > rin( te::rst::RasterFactory::open(rinfo) );

  /* Create output raster for linear mixture model */

  std::map<std::string, std::string> orinfo;
  orinfo["URI"] = "mixtureModel_tests_linear_LST_test.tif";
  std::map<std::string, std::string> orinfo_error;
  orinfo_error["URI"] = "mixtureModel_tests_linear_LST_test_error.tif";

  /* Create algorithm parameters */
  
  te::rp::MixtureModelLinearStrategy::Parameters stratParams;
  stratParams.m_methodType = te::rp::MixtureModelLinearStrategy::Parameters::LST;

  te::rp::MixtureModel::InputParameters mmInputParameters;
  
  mmInputParameters.setMixtureModelStrategyParams(stratParams);

  mmInputParameters.m_inputRasterPtr = rin.get();
  
  for (unsigned b = 0; b < rin->getNumberOfBands(); b++)
    mmInputParameters.m_inputRasterBands.push_back(b);
  
  mmInputParameters.m_strategyName = "linear";

  // Defining sensors
  mmInputParameters.m_inputSensorBands.push_back("AMAZONIA1_WFI_B1");
  mmInputParameters.m_inputSensorBands.push_back("AMAZONIA1_WFI_B2");
  mmInputParameters.m_inputSensorBands.push_back("AMAZONIA1_WFI_B3");
  mmInputParameters.m_inputSensorBands.push_back("AMAZONIA1_WFI_B4");

  mmInputParameters.m_components["clouds"].push_back(625);
  mmInputParameters.m_components["clouds"].push_back(700);
  mmInputParameters.m_components["clouds"].push_back(1023);
  mmInputParameters.m_components["clouds"].push_back(716);
  
  mmInputParameters.m_components["shadow"].push_back(237);
  mmInputParameters.m_components["shadow"].push_back(233);
  mmInputParameters.m_components["shadow"].push_back(314);
  mmInputParameters.m_components["shadow"].push_back(328);
  
  mmInputParameters.m_components["vegetation"].push_back(152);
  mmInputParameters.m_components["vegetation"].push_back(122);
  mmInputParameters.m_components["vegetation"].push_back(87);
  mmInputParameters.m_components["vegetation"].push_back(328);

  te::rp::MixtureModel::OutputParameters mmOutputParameters;
  mmOutputParameters.m_rInfo = orinfo;
  mmOutputParameters.m_rType = "GDAL";
  mmOutputParameters.m_createErrorRaster = true;
  mmOutputParameters.m_rInfoError = orinfo_error;

  /* Execute the algorithm */

  te::rp::MixtureModel algorithmInstance;
  boost::numeric::ublas::matrix<double> transfMatrix;

  BOOST_CHECK( algorithmInstance.initialize(mmInputParameters) );
  BOOST_CHECK(algorithmInstance.generateTransformMatrix(transfMatrix));
  BOOST_CHECK(transfMatrix.size1() == 3 );
  BOOST_CHECK(transfMatrix.size2() == 4 );
  BOOST_CHECK( algorithmInstance.execute(mmOutputParameters) );
}

BOOST_AUTO_TEST_CASE(linear_LST_normalize_test)
{
  /* First open the input image */

  std::map<std::string, std::string> rinfo;
  rinfo["URI"] = TERRALIB_DATA_DIR"/geotiff/AMAZONIA_1_WFI_20210903_035_019_L4_BANDS1234_clip_sjc.tif";

  std::unique_ptr< te::rst::Raster > rin( te::rst::RasterFactory::open(rinfo) );

  /* Create output raster for linear mixture model */

  std::map<std::string, std::string> orinfo;
  orinfo["URI"] = "mixtureModel_tests_linear_LST_normalize_test.tif";
  std::map<std::string, std::string> orinfo_error;
  orinfo_error["URI"] = "mixtureModel_tests_linear_LST_normalize_test_error.tif";

  /* Create algorithm parameters */
  
  te::rp::MixtureModelLinearStrategy::Parameters stratParams;
  stratParams.m_methodType = te::rp::MixtureModelLinearStrategy::Parameters::LST;

  te::rp::MixtureModel::InputParameters mmInputParameters;
  
  mmInputParameters.setMixtureModelStrategyParams(stratParams);

  mmInputParameters.m_inputRasterPtr = rin.get();
  
  for (unsigned b = 0; b < rin->getNumberOfBands(); b++)
    mmInputParameters.m_inputRasterBands.push_back(b);
  
  mmInputParameters.m_strategyName = "linear";

  // Defining sensors
  mmInputParameters.m_inputSensorBands.push_back("AMAZONIA1_WFI_B1");
  mmInputParameters.m_inputSensorBands.push_back("AMAZONIA1_WFI_B2");
  mmInputParameters.m_inputSensorBands.push_back("AMAZONIA1_WFI_B3");
  mmInputParameters.m_inputSensorBands.push_back("AMAZONIA1_WFI_B4");

  mmInputParameters.m_components["clouds"].push_back(625);
  mmInputParameters.m_components["clouds"].push_back(700);
  mmInputParameters.m_components["clouds"].push_back(1023);
  mmInputParameters.m_components["clouds"].push_back(716);
  
  mmInputParameters.m_components["shadow"].push_back(237);
  mmInputParameters.m_components["shadow"].push_back(233);
  mmInputParameters.m_components["shadow"].push_back(314);
  mmInputParameters.m_components["shadow"].push_back(328);
  
  mmInputParameters.m_components["vegetation"].push_back(152);
  mmInputParameters.m_components["vegetation"].push_back(122);
  mmInputParameters.m_components["vegetation"].push_back(87);
  mmInputParameters.m_components["vegetation"].push_back(328);

  te::rp::MixtureModel::OutputParameters mmOutputParameters;
  mmOutputParameters.m_rInfo = orinfo;
  mmOutputParameters.m_rType = "GDAL";
  mmOutputParameters.m_createErrorRaster = true;
  mmOutputParameters.m_rInfoError = orinfo_error;
  mmOutputParameters.m_normalizeOutput = true;

  /* Execute the algorithm */

  te::rp::MixtureModel algorithmInstance;
  boost::numeric::ublas::matrix<double> transfMatrix;

  BOOST_CHECK( algorithmInstance.initialize(mmInputParameters) );
  BOOST_CHECK(algorithmInstance.generateTransformMatrix(transfMatrix));
  BOOST_CHECK(transfMatrix.size1() == 3 );
  BOOST_CHECK(transfMatrix.size2() == 4 );
  BOOST_CHECK( algorithmInstance.execute(mmOutputParameters) );
}

BOOST_AUTO_TEST_CASE(linear_RLS3T_test)
{
  /* First open the input image */

  std::map<std::string, std::string> rinfo;
  rinfo["URI"] = TERRALIB_DATA_DIR"/geotiff/AMAZONIA_1_WFI_20210903_035_019_L4_BANDS1234_clip_sjc.tif";

  std::unique_ptr< te::rst::Raster > rin( te::rst::RasterFactory::open(rinfo) );

  /* Create output raster for linear mixture model */

  std::map<std::string, std::string> orinfo;
  orinfo["URI"] = "mixtureModel_tests_linear_RLS3T_test.tif";
  std::map<std::string, std::string> orinfo_error;
  orinfo_error["URI"] = "mixtureModel_tests_linear_RLS3T_test_error.tif";

  /* Create algorithm parameters */
  
  te::rp::MixtureModelLinearStrategy::Parameters stratParams;
  stratParams.m_methodType = te::rp::MixtureModelLinearStrategy::Parameters::RLS3T;

  te::rp::MixtureModel::InputParameters mmInputParameters;
  
  mmInputParameters.setMixtureModelStrategyParams(stratParams);

  mmInputParameters.m_inputRasterPtr = rin.get();
  
  for (unsigned b = 0; b < rin->getNumberOfBands(); b++)
    mmInputParameters.m_inputRasterBands.push_back(b);
  
  mmInputParameters.m_strategyName = "linear";

  // Defining sensors
  mmInputParameters.m_inputSensorBands.push_back("AMAZONIA1_WFI_B1");
  mmInputParameters.m_inputSensorBands.push_back("AMAZONIA1_WFI_B2");
  mmInputParameters.m_inputSensorBands.push_back("AMAZONIA1_WFI_B3");
  mmInputParameters.m_inputSensorBands.push_back("AMAZONIA1_WFI_B4");

  mmInputParameters.m_components["clouds"].push_back(625);
  mmInputParameters.m_components["clouds"].push_back(700);
  mmInputParameters.m_components["clouds"].push_back(1023);
  mmInputParameters.m_components["clouds"].push_back(716);
  
  mmInputParameters.m_components["shadow"].push_back(237);
  mmInputParameters.m_components["shadow"].push_back(233);
  mmInputParameters.m_components["shadow"].push_back(314);
  mmInputParameters.m_components["shadow"].push_back(328);
  
  mmInputParameters.m_components["vegetation"].push_back(152);
  mmInputParameters.m_components["vegetation"].push_back(122);
  mmInputParameters.m_components["vegetation"].push_back(87);
  mmInputParameters.m_components["vegetation"].push_back(328);

  te::rp::MixtureModel::OutputParameters mmOutputParameters;
  mmOutputParameters.m_rInfo = orinfo;
  mmOutputParameters.m_rType = "GDAL";
  mmOutputParameters.m_createErrorRaster = true;
  mmOutputParameters.m_rInfoError = orinfo_error;

  /* Execute the algorithm */

  te::rp::MixtureModel algorithmInstance;
  boost::numeric::ublas::matrix<double> transfMatrix;

  BOOST_CHECK( algorithmInstance.initialize(mmInputParameters) );
  BOOST_CHECK(algorithmInstance.generateTransformMatrix(transfMatrix));
  BOOST_CHECK(transfMatrix.size1() == 3 );
  BOOST_CHECK(transfMatrix.size2() == 4 );
  BOOST_CHECK( algorithmInstance.execute(mmOutputParameters) );
}

BOOST_AUTO_TEST_CASE(linear_RLS3T_normalize_test)
{
  /* First open the input image */

  std::map<std::string, std::string> rinfo;
  rinfo["URI"] = TERRALIB_DATA_DIR"/geotiff/AMAZONIA_1_WFI_20210903_035_019_L4_BANDS1234_clip_sjc.tif";

  std::unique_ptr< te::rst::Raster > rin( te::rst::RasterFactory::open(rinfo) );

  /* Create output raster for linear mixture model */

  std::map<std::string, std::string> orinfo;
  orinfo["URI"] = "mixtureModel_tests_linear_RLS3T_normalize_test.tif";
  std::map<std::string, std::string> orinfo_error;
  orinfo_error["URI"] = "mixtureModel_tests_linear_RLS3T_normalize_test_error.tif";

  /* Create algorithm parameters */
  
  te::rp::MixtureModelLinearStrategy::Parameters stratParams;
  stratParams.m_methodType = te::rp::MixtureModelLinearStrategy::Parameters::RLS3T;

  te::rp::MixtureModel::InputParameters mmInputParameters;
  
  mmInputParameters.setMixtureModelStrategyParams(stratParams);

  mmInputParameters.m_inputRasterPtr = rin.get();
  
  for (unsigned b = 0; b < rin->getNumberOfBands(); b++)
    mmInputParameters.m_inputRasterBands.push_back(b);
  
  mmInputParameters.m_strategyName = "linear";

  // Defining sensors
  mmInputParameters.m_inputSensorBands.push_back("AMAZONIA1_WFI_B1");
  mmInputParameters.m_inputSensorBands.push_back("AMAZONIA1_WFI_B2");
  mmInputParameters.m_inputSensorBands.push_back("AMAZONIA1_WFI_B3");
  mmInputParameters.m_inputSensorBands.push_back("AMAZONIA1_WFI_B4");

  mmInputParameters.m_components["clouds"].push_back(625);
  mmInputParameters.m_components["clouds"].push_back(700);
  mmInputParameters.m_components["clouds"].push_back(1023);
  mmInputParameters.m_components["clouds"].push_back(716);
  
  mmInputParameters.m_components["shadow"].push_back(237);
  mmInputParameters.m_components["shadow"].push_back(233);
  mmInputParameters.m_components["shadow"].push_back(314);
  mmInputParameters.m_components["shadow"].push_back(328);
  
  mmInputParameters.m_components["vegetation"].push_back(152);
  mmInputParameters.m_components["vegetation"].push_back(122);
  mmInputParameters.m_components["vegetation"].push_back(87);
  mmInputParameters.m_components["vegetation"].push_back(328);

  te::rp::MixtureModel::OutputParameters mmOutputParameters;
  mmOutputParameters.m_rInfo = orinfo;
  mmOutputParameters.m_rType = "GDAL";
  mmOutputParameters.m_createErrorRaster = true;
  mmOutputParameters.m_rInfoError = orinfo_error;
  mmOutputParameters.m_normalizeOutput = true;

  /* Execute the algorithm */

  te::rp::MixtureModel algorithmInstance;
  boost::numeric::ublas::matrix<double> transfMatrix;

  BOOST_CHECK( algorithmInstance.initialize(mmInputParameters) );
  BOOST_CHECK(algorithmInstance.generateTransformMatrix(transfMatrix));
  BOOST_CHECK(transfMatrix.size1() == 3 );
  BOOST_CHECK(transfMatrix.size2() == 4 );
  BOOST_CHECK( algorithmInstance.execute(mmOutputParameters) );
}

BOOST_AUTO_TEST_CASE(pca_test)
{
  /* First open the input image */

  std::map<std::string, std::string> rinfo;
  rinfo["URI"] = TERRALIB_DATA_DIR"/geotiff/cbers2b_rgb342_crop.tif";

  std::unique_ptr< te::rst::Raster > rin( te::rst::RasterFactory::open(rinfo) );

  /* Create output raster for PCA mixture model */

  std::map<std::string, std::string> orinfo;
  orinfo["URI"] = "cbers2b_rgb342_crop_PCAMixtureModel.tif";
  std::map<std::string, std::string> orinfo_error;
  orinfo_error["URI"] = "cbers2b_rgb342_crop_PCAMixtureModel_error.tif";

  /* Create algorithm parameters */

  te::rp::MixtureModel::InputParameters mmInputParameters;
  te::rp::MixtureModel::OutputParameters mmOutputParameters;

  /* Defining input parameters */

  mmInputParameters.m_inputRasterPtr = rin.get();
  for (unsigned b = 0; b < rin->getNumberOfBands(); b++)
    mmInputParameters.m_inputRasterBands.push_back(b);
  mmInputParameters.m_strategyName = "pca";

  // Defining sensors
  mmInputParameters.m_inputSensorBands.push_back("CBERS2_CCD_1_BLUE");
  mmInputParameters.m_inputSensorBands.push_back("CBERS2_CCD_1_GREEN");
  mmInputParameters.m_inputSensorBands.push_back("CBERS2_CCD_1_RED");

  /* Link specific parameters with chosen implementation */

  te::rp::MixtureModelPCAStrategy::Parameters pcammParameters;
  mmInputParameters.setMixtureModelStrategyParams(pcammParameters);

  /* Defining endmembers */

  mmInputParameters.m_components["clouds"].push_back(255.0);
  mmInputParameters.m_components["clouds"].push_back(184.875);
  mmInputParameters.m_components["clouds"].push_back(255.0);

  mmInputParameters.m_components["shadow"].push_back(42.075);
  mmInputParameters.m_components["shadow"].push_back(24.99);
  mmInputParameters.m_components["shadow"].push_back(49.98);

  mmInputParameters.m_components["vegetation"].push_back(99.96);
  mmInputParameters.m_components["vegetation"].push_back(64.005);
  mmInputParameters.m_components["vegetation"].push_back(154.02);

  /* Defining output parameters */

  mmOutputParameters.m_rInfo = orinfo;
  mmOutputParameters.m_rType = "GDAL";
  mmOutputParameters.m_createErrorRaster = true;
  mmOutputParameters.m_rInfoError = orinfo_error;

  /* Execute the algorithm */

  te::rp::MixtureModel algorithmInstance;
  boost::numeric::ublas::matrix<double> transfMatrix;

  BOOST_CHECK( algorithmInstance.initialize(mmInputParameters) );
  BOOST_CHECK(algorithmInstance.generateTransformMatrix(transfMatrix));
  BOOST_CHECK(transfMatrix.size1() == 3);
  BOOST_CHECK(transfMatrix.size2() == 3);
  BOOST_CHECK( algorithmInstance.execute(mmOutputParameters) );
}

BOOST_AUTO_TEST_SUITE_END()
