/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/unittest/rp/radar_functions/TsRadarFunctions.cpp

  \brief A test suit for raster processing Radar Functions.
*/

// TerraLib
#include "../Config.h"
#include <terralib/rp.h>
#include <terralib/raster.h>
#include <terralib/geometry.h>

// Boost
#define BOOST_TEST_NO_MAIN
#include <boost/test/unit_test.hpp>
#include <boost/shared_ptr.hpp>

BOOST_AUTO_TEST_SUITE (multisourceclassifier_tests)

BOOST_AUTO_TEST_CASE(multisourceclassifier_one_image_test)
{
  te::rp::radar::MultiSourceClassifier::InputParameters algoInputParams;

  std::unique_ptr< te::rst::Raster > inputRasterPtr;
  {
    std::map<std::string, std::string> auxRasterInfo;
    auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/Multivarada_em_amplitude.tif";
    inputRasterPtr.reset( te::rst::RasterFactory::open( auxRasterInfo ) );
    BOOST_REQUIRE( inputRasterPtr.get() );

    std::vector< unsigned int > bandsIndexes;
    for( std::size_t bandIdx = 0 ; bandIdx < inputRasterPtr->getNumberOfBands() ;
      ++bandIdx )
    {
      bandsIndexes.push_back( bandIdx );
    }
    algoInputParams.m_inputRastersBands.push_back( bandsIndexes );
  }

  std::unique_ptr< te::rst::Raster > inputSegRasterPtr;
  {
    std::map<std::string, std::string> auxRasterInfo;
    auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/imagem_segmentada.tif";
    inputSegRasterPtr.reset( te::rst::RasterFactory::open( auxRasterInfo ) );
    BOOST_REQUIRE( inputSegRasterPtr.get() );
  }

  algoInputParams.m_dirTrainingSamples["URI"] =
    TERRALIB_DATA_DIR "/txt/AMOSTRAS_DE_TREINAMENTO.txt";
  algoInputParams.m_dirTestSamples["URI"] =
    TERRALIB_DATA_DIR "/txt/AMOSTRAS_DE_TESTE.txt";
  algoInputParams.m_inputRasterPtr.push_back( inputRasterPtr.get() );
  algoInputParams.m_sourcesNumber = 1;
  algoInputParams.m_combinationType = "none";
  algoInputParams.m_significanceLevel = 0.95;
  algoInputParams.m_lagX = 0;
  algoInputParams.m_lagY = 0;
  algoInputParams.m_imageType.push_back( "Gaussian" );
  algoInputParams.m_stochasticDistanceType.push_back( "Bhattacharyya" );
  algoInputParams.m_radarDataRepresentationType.push_back( "covarianceMatrix" );
  algoInputParams.m_vectorOrMatrixOrder.push_back( 0 );
  algoInputParams.m_ENL.push_back( 0 );
  algoInputParams.m_betaValue.push_back( 0 );
  algoInputParams.m_rasterSegPtr = inputSegRasterPtr.get();
  algoInputParams.m_numberOfRows = algoInputParams.m_rasterSegPtr->getNumberOfRows();
  algoInputParams.m_numberOfColumns = algoInputParams.m_rasterSegPtr->getNumberOfColumns();
  algoInputParams.m_SRID = algoInputParams.m_rasterSegPtr->getSRID();
  algoInputParams.m_extent = inputSegRasterPtr->getGrid();

  te::rp::radar::PointersVector< te::gm::Geometry* > inputSegRasterGeoms;
  inputSegRasterPtr->vectorize(inputSegRasterGeoms,0);
  {
    std::size_t inputSegRasterGeomsSize = inputSegRasterGeoms.size();

    for (unsigned i = 0; i < inputSegRasterGeomsSize; i++)
    {
      algoInputParams.m_polygonsSegImage.push_back(
        static_cast<te::gm::Polygon*> (inputSegRasterGeoms[i]));
    }
  }

  BOOST_REQUIRE( te::rp::radar::LoadSamplesFile(
    algoInputParams.m_dirTrainingSamples["URI"],
    algoInputParams.m_numberOfClassesOnTraining,
    algoInputParams.m_classesIDT,
    algoInputParams.m_nameOfSamples,
    algoInputParams.m_numberOfSamples,
    algoInputParams.m_polygonSamplesTNew,
    algoInputParams.m_polygonSamplesT ) );

  BOOST_REQUIRE( te::rp::radar::LoadSamplesFile(
    algoInputParams.m_dirTestSamples["URI"],
    algoInputParams.m_numberOfClassesOnTest,
    algoInputParams.m_classesIDTest,
    algoInputParams.m_nameOfSamplesTest,
    algoInputParams.m_numberOfSamplesTest,
    algoInputParams.m_polygonSamplesTestNew,
    algoInputParams.m_polygonSamplesTest ) );

  te::rp::radar::MultiSourceClassifier::OutputParameters algoOutputParams;
  algoOutputParams.m_rInfoClassImage["URI"] =
    "multisourceclassifier_one_image_test_outclasses.tif";
  algoOutputParams.m_rType = "GDAL";
  algoOutputParams.m_rInfoMap["URI"] =
    "multisourceclassifier_one_image_test_stat.tif";
  algoOutputParams.m_rInfoMapPValue["URI"] =
    "multisourceclassifier_one_image_test_pvalue.tif";

  te::rp::radar::MultiSourceClassifier algorithmInstance;
  BOOST_REQUIRE( algorithmInstance.initialize( algoInputParams ) );
  BOOST_REQUIRE( algorithmInstance.execute( algoOutputParams ) );

  BOOST_REQUIRE( te::rp::radar::SaveOutputReportFile(
    algoInputParams, algoOutputParams,
    "multisourceclassifier_one_image_test_stat.txt" ) );
  BOOST_REQUIRE( te::rp::radar::SaveOutputAttFile(
    algoInputParams, algoOutputParams, true, true, true,
    "multisourceclassifier_one_image_test_att.txt" ) );
}

BOOST_AUTO_TEST_CASE(multisourceclassifier_two_image_test)
{
  te::rp::radar::MultiSourceClassifier::InputParameters algoInputParams;

  std::unique_ptr< te::rst::Raster > inputRasterPtr;
  {
    std::map<std::string, std::string> auxRasterInfo;
    auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/Multivarada_em_amplitude.tif";
    inputRasterPtr.reset( te::rst::RasterFactory::open( auxRasterInfo ) );
    BOOST_REQUIRE( inputRasterPtr.get() );

    std::vector< unsigned int > bandsIndexes;
    for( std::size_t bandIdx = 0 ; bandIdx < inputRasterPtr->getNumberOfBands() ;
      ++bandIdx )
    {
      bandsIndexes.push_back( bandIdx );
    }
    algoInputParams.m_inputRasterPtr.push_back( inputRasterPtr.get() );
    algoInputParams.m_inputRastersBands.push_back( bandsIndexes );
  }

  std::unique_ptr< te::rst::Raster > inputPolRasterPtr;
  {
    std::map<std::string, std::string> auxRasterInfo;
    auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/Polarimetrica.tif";
    inputPolRasterPtr.reset( te::rst::RasterFactory::open( auxRasterInfo ) );
    BOOST_REQUIRE( inputPolRasterPtr.get() );

    std::vector< unsigned int > bandsIndexes;
    for( std::size_t bandIdx = 0 ; bandIdx < inputPolRasterPtr->getNumberOfBands() ;
      ++bandIdx )
    {
      bandsIndexes.push_back( bandIdx );
    }
    algoInputParams.m_inputRasterPtr.push_back( inputPolRasterPtr.get() );
    algoInputParams.m_inputRastersBands.push_back( bandsIndexes );
  }

  std::unique_ptr< te::rst::Raster > inputSegRasterPtr;
  {
    std::map<std::string, std::string> auxRasterInfo;
    auxRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/imagem_segmentada.tif";
    inputSegRasterPtr.reset( te::rst::RasterFactory::open( auxRasterInfo ) );
    BOOST_REQUIRE( inputSegRasterPtr.get() );
  }

  algoInputParams.m_dirTrainingSamples["URI"] =
    TERRALIB_DATA_DIR "/txt/AMOSTRAS_DE_TREINAMENTO.txt";
  algoInputParams.m_dirTestSamples["URI"] =
    TERRALIB_DATA_DIR "/txt/AMOSTRAS_DE_TESTE.txt";
  algoInputParams.m_sourcesNumber = 2;
  algoInputParams.m_combinationType = "none";
  algoInputParams.m_significanceLevel = 0.95;
  algoInputParams.m_lagX = 0;
  algoInputParams.m_lagY = 0;
  algoInputParams.m_imageType.push_back( "Gaussian" );
  algoInputParams.m_imageType.push_back( "Wishart" );
  algoInputParams.m_stochasticDistanceType.push_back( "Bhattacharyya" );
  algoInputParams.m_stochasticDistanceType.push_back( "Bhattacharyya" );
  algoInputParams.m_radarDataRepresentationType.push_back( "covarianceMatrix" );
  algoInputParams.m_radarDataRepresentationType.push_back( "covarianceMatrix" );
  algoInputParams.m_vectorOrMatrixOrder.push_back( 0 );
  algoInputParams.m_vectorOrMatrixOrder.push_back( 3 );
  algoInputParams.m_ENL.push_back( 0 );
  algoInputParams.m_ENL.push_back( 4 );
  algoInputParams.m_betaValue.push_back( 0 );
  algoInputParams.m_betaValue.push_back( 0 );
  algoInputParams.m_rasterSegPtr = inputSegRasterPtr.get();
  algoInputParams.m_numberOfRows = algoInputParams.m_rasterSegPtr->getNumberOfRows();
  algoInputParams.m_numberOfColumns = algoInputParams.m_rasterSegPtr->getNumberOfColumns();
  algoInputParams.m_SRID = algoInputParams.m_rasterSegPtr->getSRID();
  algoInputParams.m_extent = inputSegRasterPtr->getGrid();

  te::rp::radar::PointersVector< te::gm::Geometry* > inputSegRasterGeoms;
  inputSegRasterPtr->vectorize(inputSegRasterGeoms,0);
  {
    std::size_t inputSegRasterGeomsSize = inputSegRasterGeoms.size();

    for (unsigned i = 0; i < inputSegRasterGeomsSize; i++)
    {
      algoInputParams.m_polygonsSegImage.push_back(
        static_cast<te::gm::Polygon*> (inputSegRasterGeoms[i]));
    }
  }

  BOOST_REQUIRE( te::rp::radar::LoadSamplesFile(
    algoInputParams.m_dirTrainingSamples["URI"],
    algoInputParams.m_numberOfClassesOnTraining,
    algoInputParams.m_classesIDT,
    algoInputParams.m_nameOfSamples,
    algoInputParams.m_numberOfSamples,
    algoInputParams.m_polygonSamplesTNew,
    algoInputParams.m_polygonSamplesT ) );

  BOOST_REQUIRE( te::rp::radar::LoadSamplesFile(
    algoInputParams.m_dirTestSamples["URI"],
    algoInputParams.m_numberOfClassesOnTest,
    algoInputParams.m_classesIDTest,
    algoInputParams.m_nameOfSamplesTest,
    algoInputParams.m_numberOfSamplesTest,
    algoInputParams.m_polygonSamplesTestNew,
    algoInputParams.m_polygonSamplesTest ) );

  te::rp::radar::MultiSourceClassifier::OutputParameters algoOutputParams;
  algoOutputParams.m_rInfoClassImage["URI"] =
    "multisourceclassifier_two_image_test_outclasses.tif";
  algoOutputParams.m_rType = "GDAL";
  algoOutputParams.m_rInfoMap["URI"] =
    "multisourceclassifier_two_image_test_stat.tif";
  algoOutputParams.m_rInfoMapPValue["URI"] =
    "multisourceclassifier_two_image_test_pvalue.tif";

  te::rp::radar::MultiSourceClassifier algorithmInstance;
  BOOST_REQUIRE( algorithmInstance.initialize( algoInputParams ) );
  BOOST_REQUIRE( algorithmInstance.execute( algoOutputParams ) );

  BOOST_REQUIRE( te::rp::radar::SaveOutputReportFile(
    algoInputParams, algoOutputParams,
    "multisourceclassifier_two_image_test_stat.txt" ) );
  BOOST_REQUIRE( te::rp::radar::SaveOutputAttFile(
    algoInputParams, algoOutputParams, true, true, true,
    "multisourceclassifier_two_image_test_att.txt" ) );
}

BOOST_AUTO_TEST_SUITE_END()
