/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/unittest/rp/segmenter/TsSegmenter.cpp

  \brief A test suit for Segmenter interface.
*/

// TerraLib
#include "../Config.h"
#include <terralib/rp.h>
#include <terralib/raster.h>
#include <terralib/common.h>
#include <terralib/dataaccess.h>

// STL
#include <cmath>

// Boost
#define BOOST_TEST_NO_MAIN
#include <boost/test/unit_test.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/timer.hpp>

BOOST_AUTO_TEST_SUITE (segmenter_tests)

BOOST_AUTO_TEST_CASE(dummyStrategy_test)
{
  /* Progress interface */

  te::common::ConsoleProgressViewer progressViewerInstance;
  
  /* Open input raster */
  
  std::map<std::string, std::string> inputRasterInfo;
  inputRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers2b_rgb342_crop.tif";
  
  boost::shared_ptr< te::rst::Raster > inputRasterPointer ( te::rst::RasterFactory::open(
    inputRasterInfo ) );
  BOOST_CHECK( inputRasterPointer.get() );

  /* Access a raster datasource to create the output raster */

  std::map<std::string, std::string> outputRasterInfo;
  outputRasterInfo["URI"] = "terralib_unittest_rp_Segmenter_DummyStrategy_Test.tif";
  
  /* Creating the algorithm parameters */
  
  te::rp::SegmenterRegionGrowingMeanStrategy::Parameters strategyParameters;
  strategyParameters.m_minSegmentSize = 50;
  strategyParameters.m_segmentsSimilarityThreshold = 0.1;
  
  te::rp::Segmenter::InputParameters algoInputParams;
  algoInputParams.m_inputRasterPtr = inputRasterPointer.get();
  algoInputParams.m_inputRasterBands.push_back( 0 );
  algoInputParams.m_inputRasterBands.push_back( 1 );
  algoInputParams.m_inputRasterBands.push_back( 2 );
  algoInputParams.m_enableThreadedProcessing = false;
  algoInputParams.m_maxSegThreads = 0;
  algoInputParams.m_enableBlockProcessing = false;
  algoInputParams.m_blocksOverlapPercent = 0;
  algoInputParams.m_maxBlockSize = 0;
  algoInputParams.m_strategyName = "Dummy";
  algoInputParams.setSegStrategyParams( strategyParameters );
  algoInputParams.m_enableProgress = true;
  
  te::rp::Segmenter::OutputParameters algoOutputParams;
  algoOutputParams.m_rInfo = outputRasterInfo;
  algoOutputParams.m_rType = "GDAL";
  
  /* Executing the algorithm */
  
  te::rp::Segmenter algorithmInstance;
  
  BOOST_CHECK( algorithmInstance.initialize( algoInputParams ) );
  BOOST_CHECK( algorithmInstance.execute( algoOutputParams ) );
}

BOOST_AUTO_TEST_CASE(blockProcessingWithoutMerging_test)
{
  /* Progress interface */

  te::common::ConsoleProgressViewer progressViewerInstance;
  
  /* Open input raster */
  
  std::map<std::string, std::string> inputRasterInfo;
  inputRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers2b_rgb342_crop.tif";
  
  boost::shared_ptr< te::rst::Raster > inputRasterPointer ( te::rst::RasterFactory::open(
    inputRasterInfo ) );
  BOOST_CHECK( inputRasterPointer.get() );
  
  /* Access a raster datasource to create the output raster */
  
  std::map<std::string, std::string> outputRasterInfo;
  outputRasterInfo["URI"] = "terralib_unittest_rp_Segmenter_BlockProcessingWithoutMerging_Test.tif";
  
  /* Creating the algorithm parameters */
  
  te::rp::SegmenterRegionGrowingMeanStrategy::Parameters strategyParameters;
  strategyParameters.m_minSegmentSize = 50;
  strategyParameters.m_segmentsSimilarityThreshold = 0.1;
  
  te::rp::Segmenter::InputParameters algoInputParams;
  algoInputParams.m_inputRasterPtr = inputRasterPointer.get();
  algoInputParams.m_inputRasterBands.push_back( 0 );
  algoInputParams.m_inputRasterBands.push_back( 1 );
  algoInputParams.m_inputRasterBands.push_back( 2 );
  algoInputParams.m_enableThreadedProcessing = false;
  algoInputParams.m_maxSegThreads = 0;
  algoInputParams.m_enableBlockProcessing = true;
  algoInputParams.m_blocksOverlapPercent = 0;
  algoInputParams.m_maxBlockSize = 100;
  algoInputParams.m_strategyName = "Dummy";
  algoInputParams.setSegStrategyParams( strategyParameters );
  algoInputParams.m_enableProgress = true;
  
  te::rp::Segmenter::OutputParameters algoOutputParams;
  algoOutputParams.m_rInfo = outputRasterInfo;
  algoOutputParams.m_rType = "GDAL";
  
  /* Executing the algorithm */
  
  te::rp::Segmenter algorithmInstance;
  
  BOOST_CHECK( algorithmInstance.initialize( algoInputParams ) );
  BOOST_CHECK( algorithmInstance.execute( algoOutputParams ) );
}

BOOST_AUTO_TEST_CASE(blockProcessingWithMerging_test)
{
  /* Progress interface */

  te::common::ConsoleProgressViewer progressViewerInstance;
  
  /* Open input raster */
  
  std::map<std::string, std::string> inputRasterInfo;
  inputRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers2b_rgb342_crop.tif";
  
  boost::shared_ptr< te::rst::Raster > inputRasterPointer ( te::rst::RasterFactory::open(
    inputRasterInfo ) );
  BOOST_CHECK( inputRasterPointer.get() );
  
  /* Access a raster datasource to create the output raster */
  
  std::map<std::string, std::string> outputRasterInfo;
  outputRasterInfo["URI"] = "terralib_unittest_rp_Segmenter_BlockProcessingWithMerging_Test.tif";
  
  /* Creating the algorithm parameters */
  
  te::rp::SegmenterRegionGrowingMeanStrategy::Parameters strategyParameters;
  strategyParameters.m_minSegmentSize = 50;
  strategyParameters.m_segmentsSimilarityThreshold = 0.1;
  
  te::rp::Segmenter::InputParameters algoInputParams;
  algoInputParams.m_inputRasterPtr = inputRasterPointer.get();
  algoInputParams.m_inputRasterBands.push_back( 0 );
  algoInputParams.m_inputRasterBands.push_back( 1 );
  algoInputParams.m_inputRasterBands.push_back( 2 );
  algoInputParams.m_enableThreadedProcessing = false;
  algoInputParams.m_maxSegThreads = 0;
  algoInputParams.m_enableBlockProcessing = true;
  algoInputParams.m_blocksOverlapPercent = 20;
  algoInputParams.m_maxBlockSize = 100;
  algoInputParams.m_strategyName = "Dummy";
  algoInputParams.setSegStrategyParams( strategyParameters );
  algoInputParams.m_enableProgress = true;
  
  te::rp::Segmenter::OutputParameters algoOutputParams;
  algoOutputParams.m_rInfo = outputRasterInfo;
  algoOutputParams.m_rType = "GDAL";
  
  /* Executing the algorithm */
  
  te::rp::Segmenter algorithmInstance;
  
  BOOST_CHECK( algorithmInstance.initialize( algoInputParams ) );
  BOOST_CHECK( algorithmInstance.execute( algoOutputParams ) );
}

BOOST_AUTO_TEST_CASE(threadedProcessing_test)
{
  /* Progress interface */

  te::common::ConsoleProgressViewer progressViewerInstance;
  
  /* Open input raster */
  
  std::map<std::string, std::string> inputRasterInfo;
  inputRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers2b_rgb342_crop.tif";
  
  boost::shared_ptr< te::rst::Raster > inputRasterPointer ( te::rst::RasterFactory::open(
    inputRasterInfo ) );
  BOOST_CHECK( inputRasterPointer.get() );
  
  /* Access a raster datasource to create the output raster */
  
  std::map<std::string, std::string> outputRasterInfo;
  outputRasterInfo["URI"] = "terralib_unittest_rp_Segmenter_ThreadedProcessing_Test.tif";
  
  /* Creating the algorithm parameters */
  
  te::rp::SegmenterRegionGrowingMeanStrategy::Parameters strategyParameters;
  strategyParameters.m_minSegmentSize = 50;
  strategyParameters.m_segmentsSimilarityThreshold = 0.1;
  
  te::rp::Segmenter::InputParameters algoInputParams;
  algoInputParams.m_inputRasterPtr = inputRasterPointer.get();
  algoInputParams.m_inputRasterBands.push_back( 0 );
  algoInputParams.m_inputRasterBands.push_back( 1 );
  algoInputParams.m_inputRasterBands.push_back( 2 );
  algoInputParams.m_enableThreadedProcessing = true;
  algoInputParams.m_maxSegThreads = 4;
  algoInputParams.m_enableBlockProcessing = true;
  algoInputParams.m_blocksOverlapPercent = 20;
  algoInputParams.m_maxBlockSize = 100;
  algoInputParams.m_strategyName = "Dummy";
  algoInputParams.setSegStrategyParams( strategyParameters );
  algoInputParams.m_enableProgress = true;
  
  te::rp::Segmenter::OutputParameters algoOutputParams;
  algoOutputParams.m_rInfo = outputRasterInfo;
  algoOutputParams.m_rType = "GDAL";
  
  /* Executing the algorithm */
  
  te::rp::Segmenter algorithmInstance;
  
  BOOST_CHECK( algorithmInstance.initialize( algoInputParams ) );
  BOOST_CHECK( algorithmInstance.execute( algoOutputParams ) );
}

BOOST_AUTO_TEST_CASE(regionGrowingMeanStrategy_test)
{
  /* Progress interface */

  te::common::ConsoleProgressViewer progressViewerInstance;
  
  /* Open input raster */
  
  std::map<std::string, std::string> inputRasterInfo;
  inputRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers_rgb342_crop1.tif";
  
  boost::shared_ptr< te::rst::Raster > inputRasterPointer ( te::rst::RasterFactory::open(
    inputRasterInfo ) );
  BOOST_CHECK( inputRasterPointer.get() );
  
  /* Access a raster datasource to create the output raster */
  
  std::map<std::string, std::string> outputRasterInfo;
  outputRasterInfo["URI"] = "terralib_unittest_rp_Segmenter_RegionGrowingMeanStrategy_Test.tif";
  
  /* Creating the algorithm parameters */
  
  te::rp::SegmenterRegionGrowingMeanStrategy::Parameters strategyParameters;
  strategyParameters.m_minSegmentSize = 100;
  strategyParameters.m_segmentsSimilarityThreshold = 0.025;
  strategyParameters.m_segmentsSimIncreaseSteps = 3;
  strategyParameters.m_enableLocalMutualBestFitting = true;
  
  te::rp::Segmenter::InputParameters algoInputParams;
  algoInputParams.m_inputRasterPtr = inputRasterPointer.get();
  algoInputParams.m_inputRasterBands.push_back( 0 );
  algoInputParams.m_inputRasterBands.push_back( 1 );
  algoInputParams.m_inputRasterBands.push_back( 2 );
  algoInputParams.m_enableThreadedProcessing = false;
  algoInputParams.m_maxSegThreads = 0;
  algoInputParams.m_enableBlockProcessing = false;
  algoInputParams.m_blocksOverlapPercent = 0;
  algoInputParams.m_maxBlockSize = 0;
  algoInputParams.m_strategyName = "RegionGrowingMean";
  algoInputParams.setSegStrategyParams( strategyParameters );
  algoInputParams.m_enableProgress = true;
  
  te::rp::Segmenter::OutputParameters algoOutputParams;
  algoOutputParams.m_rInfo = outputRasterInfo;
  algoOutputParams.m_rType = "GDAL";
  
  /* Executing the algorithm */
  
  te::rp::Segmenter algorithmInstance;
  
  BOOST_CHECK( algorithmInstance.initialize( algoInputParams ) );
  
  boost::timer timerInstance;
  BOOST_CHECK( algorithmInstance.execute( algoOutputParams ) );
  std::cout << std::endl << "Elapsed time: " << timerInstance.elapsed() <<
    std::endl;
}

BOOST_AUTO_TEST_CASE(regionGrowingMeanStrategy_forced_nodata_test)
{
  /* Progress interface */

  te::common::ConsoleProgressViewer progressViewerInstance;

  /* Open input raster */

  std::map<std::string, std::string> inputRasterInfo;
  inputRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers_rgb342_crop1.tif";

  boost::shared_ptr< te::rst::Raster > inputRasterPointer ( te::rst::RasterFactory::open(
    inputRasterInfo ) );
  BOOST_CHECK( inputRasterPointer.get() );

  /* Access a raster datasource to create the output raster */

  std::map<std::string, std::string> outputRasterInfo;
  outputRasterInfo["URI"] = "terralib_unittest_rp_Segmenter_RegionGrowingMeanStrategy_Forced_no_data_Test.tif";

  /* Creating the algorithm parameters */

  te::rp::SegmenterRegionGrowingMeanStrategy::Parameters strategyParameters;
  strategyParameters.m_minSegmentSize = 100;
  strategyParameters.m_segmentsSimilarityThreshold = 0.025;
  strategyParameters.m_segmentsSimIncreaseSteps = 3;
  strategyParameters.m_enableLocalMutualBestFitting = true;

  te::rp::Segmenter::InputParameters algoInputParams;
  algoInputParams.m_inputRasterPtr = inputRasterPointer.get();
  algoInputParams.m_inputRasterBands.push_back( 0 );
  algoInputParams.m_inputRasterBands.push_back( 1 );
  algoInputParams.m_inputRasterBands.push_back( 2 );
  algoInputParams.m_inputRasterNoDataValues.push_back( 36 );
  algoInputParams.m_inputRasterNoDataValues.push_back( 40 );
  algoInputParams.m_inputRasterNoDataValues.push_back( 74 );
  algoInputParams.m_enableThreadedProcessing = false;
  algoInputParams.m_maxSegThreads = 0;
  algoInputParams.m_enableBlockProcessing = false;
  algoInputParams.m_blocksOverlapPercent = 0;
  algoInputParams.m_maxBlockSize = 0;
  algoInputParams.m_strategyName = "RegionGrowingMean";
  algoInputParams.setSegStrategyParams( strategyParameters );
  algoInputParams.m_enableProgress = true;

  te::rp::Segmenter::OutputParameters algoOutputParams;
  algoOutputParams.m_rInfo = outputRasterInfo;
  algoOutputParams.m_rType = "GDAL";

  /* Executing the algorithm */

  te::rp::Segmenter algorithmInstance;

  BOOST_CHECK( algorithmInstance.initialize( algoInputParams ) );

  boost::timer timerInstance;
  BOOST_CHECK( algorithmInstance.execute( algoOutputParams ) );
  std::cout << std::endl << "Elapsed time: " << timerInstance.elapsed() <<
    std::endl;
}

BOOST_AUTO_TEST_CASE(regionGrowingMeanStrategy_no_data_test)
{
  /* Progress interface */

  te::common::ConsoleProgressViewer progressViewerInstance;

  /* Open input raster */

  std::map<std::string, std::string> inputRasterInfo;
  inputRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers2b_rgb342_crop_with_color_table_and_nodata.tif";

  boost::shared_ptr< te::rst::Raster > inputRasterPointer ( te::rst::RasterFactory::open(
    inputRasterInfo ) );
  BOOST_CHECK( inputRasterPointer.get() );

  /* Access a raster datasource to create the output raster */

  std::map<std::string, std::string> outputRasterInfo;
  outputRasterInfo["URI"] = "terralib_unittest_rp_Segmenter_RegionGrowingMeanStrategy_no_data_Test.tif";

  /* Creating the algorithm parameters */

  te::rp::SegmenterRegionGrowingMeanStrategy::Parameters strategyParameters;
  strategyParameters.m_minSegmentSize = 100;
  strategyParameters.m_segmentsSimilarityThreshold = 0.025;
  strategyParameters.m_segmentsSimIncreaseSteps = 3;
  strategyParameters.m_enableLocalMutualBestFitting = true;

  te::rp::Segmenter::InputParameters algoInputParams;
  algoInputParams.m_inputRasterPtr = inputRasterPointer.get();
  algoInputParams.m_inputRasterBands.push_back( 0 );
  algoInputParams.m_enableThreadedProcessing = false;
  algoInputParams.m_maxSegThreads = 0;
  algoInputParams.m_enableBlockProcessing = false;
  algoInputParams.m_blocksOverlapPercent = 0;
  algoInputParams.m_maxBlockSize = 0;
  algoInputParams.m_strategyName = "RegionGrowingMean";
  algoInputParams.setSegStrategyParams( strategyParameters );
  algoInputParams.m_enableProgress = true;

  te::rp::Segmenter::OutputParameters algoOutputParams;
  algoOutputParams.m_rInfo = outputRasterInfo;
  algoOutputParams.m_rType = "GDAL";

  /* Executing the algorithm */

  te::rp::Segmenter algorithmInstance;

  BOOST_CHECK( algorithmInstance.initialize( algoInputParams ) );

  boost::timer timerInstance;
  BOOST_CHECK( algorithmInstance.execute( algoOutputParams ) );
  std::cout << std::endl << "Elapsed time: " << timerInstance.elapsed() <<
    std::endl;
}

BOOST_AUTO_TEST_CASE(regionGrowingMeanStrategyBlockProcessing_test)
{
  /* Progress interface */

  te::common::ConsoleProgressViewer progressViewerInstance;

  /* Open input raster */
  
  std::map<std::string, std::string> inputRasterInfo;
  inputRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers2b_rgb342_crop.tif";
  
  boost::shared_ptr< te::rst::Raster > inputRasterPointer ( te::rst::RasterFactory::open(
    inputRasterInfo ) );
  BOOST_CHECK( inputRasterPointer.get() );
  
  /* Access a raster datasource to create the output raster */
  
  std::map<std::string, std::string> outputRasterInfo;
  outputRasterInfo["URI"] = "terralib_unittest_rp_Segmenter_RegionGrowingMeanStrategyBlockProcessing_Test.tif";
  
  /* Creating the algorithm parameters */
  
  te::rp::SegmenterRegionGrowingMeanStrategy::Parameters strategyParameters;
  strategyParameters.m_minSegmentSize = 100;
  strategyParameters.m_segmentsSimilarityThreshold = 0.1;
  
  te::rp::Segmenter::InputParameters algoInputParams;
  algoInputParams.m_inputRasterPtr = inputRasterPointer.get();
  algoInputParams.m_inputRasterBands.push_back( 0 );
  algoInputParams.m_inputRasterBands.push_back( 1 );
  algoInputParams.m_inputRasterBands.push_back( 2 );
  algoInputParams.m_enableThreadedProcessing = false;
  algoInputParams.m_maxSegThreads = 0;
  algoInputParams.m_enableBlockProcessing = true;
  algoInputParams.m_blocksOverlapPercent = 20;
  algoInputParams.m_maxBlockSize = 100;
  algoInputParams.m_strategyName = "RegionGrowingMean";
  algoInputParams.setSegStrategyParams( strategyParameters );
  algoInputParams.m_enableProgress = true;
  
  te::rp::Segmenter::OutputParameters algoOutputParams;
  algoOutputParams.m_rInfo = outputRasterInfo;
  algoOutputParams.m_rType = "GDAL";
  
  /* Executing the algorithm */
  
  te::rp::Segmenter algorithmInstance;
  
  BOOST_CHECK( algorithmInstance.initialize( algoInputParams ) );
  BOOST_CHECK( algorithmInstance.execute( algoOutputParams ) );
}

BOOST_AUTO_TEST_CASE(regionGrowingBaatzStrategy_test)
{
  /* Progress interface */

  te::common::ConsoleProgressViewer progressViewerInstance;
  
  /* Open input raster */
  
  std::map<std::string, std::string> inputRasterInfo;
  inputRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers_rgb342_crop1.tif";
  
  boost::shared_ptr< te::rst::Raster > inputRasterPointer ( te::rst::RasterFactory::open(
    inputRasterInfo ) );
  BOOST_CHECK( inputRasterPointer.get() );
  
  /* Access a raster datasource to create the output raster */
  
  std::map<std::string, std::string> outputRasterInfo;
  outputRasterInfo["URI"] = "terralib_unittest_rp_Segmenter_RegionGrowingBaatzStrategy_Test.tif";
  
  /* Creating the algorithm parameters */
  
  te::rp::SegmenterRegionGrowingBaatzStrategy::Parameters strategyParameters;
  strategyParameters.m_minSegmentSize = 100;
  strategyParameters.m_segmentsSimilarityThreshold = 0.25;
  strategyParameters.m_bandsWeights.resize( 3, 1.0 / 3.0 );
  strategyParameters.m_colorWeight = 0.75;
  strategyParameters.m_compactnessWeight = 0.5;
  strategyParameters.m_segmentsSimIncreaseSteps = 3;
  strategyParameters.m_enableLocalMutualBestFitting = false;
  
  te::rp::Segmenter::InputParameters algoInputParams;
  algoInputParams.m_inputRasterPtr = inputRasterPointer.get();
  algoInputParams.m_inputRasterBands.push_back( 0 );
  algoInputParams.m_inputRasterBands.push_back( 1 );
  algoInputParams.m_inputRasterBands.push_back( 2 );
  algoInputParams.m_enableThreadedProcessing = false;
  algoInputParams.m_maxSegThreads = 0;
  algoInputParams.m_enableBlockProcessing = false;
  algoInputParams.m_blocksOverlapPercent = 0;
  algoInputParams.m_maxBlockSize = 0;
  algoInputParams.m_strategyName = "RegionGrowingBaatz";
  algoInputParams.setSegStrategyParams( strategyParameters );
  algoInputParams.m_enableProgress = true;
  
  te::rp::Segmenter::OutputParameters algoOutputParams;
  algoOutputParams.m_rInfo = outputRasterInfo;
  algoOutputParams.m_rType = "GDAL";
  
  /* Executing the algorithm */
  
  te::rp::Segmenter algorithmInstance;
  
  BOOST_CHECK( algorithmInstance.initialize( algoInputParams ) );
  boost::timer timerInstance;
  BOOST_CHECK( algorithmInstance.execute( algoOutputParams ) );
  std::cout << std::endl << "Elapsed time: " << timerInstance.elapsed() <<
    std::endl;
}

BOOST_AUTO_TEST_CASE(regionGrowingBaatzStrategy_forced_no_data_test)
{
  /* Progress interface */

  te::common::ConsoleProgressViewer progressViewerInstance;

  /* Open input raster */

  std::map<std::string, std::string> inputRasterInfo;
  inputRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers_rgb342_crop1.tif";

  boost::shared_ptr< te::rst::Raster > inputRasterPointer ( te::rst::RasterFactory::open(
    inputRasterInfo ) );
  BOOST_CHECK( inputRasterPointer.get() );

  /* Access a raster datasource to create the output raster */

  std::map<std::string, std::string> outputRasterInfo;
  outputRasterInfo["URI"] = "terralib_unittest_rp_Segmenter_RegionGrowingBaatzStrategy_Forced_no_data_Test.tif";

  /* Creating the algorithm parameters */

  te::rp::SegmenterRegionGrowingBaatzStrategy::Parameters strategyParameters;
  strategyParameters.m_minSegmentSize = 100;
  strategyParameters.m_segmentsSimilarityThreshold = 0.25;
  strategyParameters.m_bandsWeights.resize( 3, 1.0 / 3.0 );
  strategyParameters.m_colorWeight = 0.75;
  strategyParameters.m_compactnessWeight = 0.5;
  strategyParameters.m_segmentsSimIncreaseSteps = 3;
  strategyParameters.m_enableLocalMutualBestFitting = false;

  te::rp::Segmenter::InputParameters algoInputParams;
  algoInputParams.m_inputRasterPtr = inputRasterPointer.get();
  algoInputParams.m_inputRasterBands.push_back( 0 );
  algoInputParams.m_inputRasterBands.push_back( 1 );
  algoInputParams.m_inputRasterBands.push_back( 2 );
  algoInputParams.m_inputRasterNoDataValues.push_back( 36 );
  algoInputParams.m_inputRasterNoDataValues.push_back( 40 );
  algoInputParams.m_inputRasterNoDataValues.push_back( 74 );
  algoInputParams.m_enableThreadedProcessing = false;
  algoInputParams.m_maxSegThreads = 0;
  algoInputParams.m_enableBlockProcessing = false;
  algoInputParams.m_blocksOverlapPercent = 0;
  algoInputParams.m_maxBlockSize = 0;
  algoInputParams.m_strategyName = "RegionGrowingBaatz";
  algoInputParams.setSegStrategyParams( strategyParameters );
  algoInputParams.m_enableProgress = true;

  te::rp::Segmenter::OutputParameters algoOutputParams;
  algoOutputParams.m_rInfo = outputRasterInfo;
  algoOutputParams.m_rType = "GDAL";

  /* Executing the algorithm */

  te::rp::Segmenter algorithmInstance;

  BOOST_CHECK( algorithmInstance.initialize( algoInputParams ) );
  boost::timer timerInstance;
  BOOST_CHECK( algorithmInstance.execute( algoOutputParams ) );
  std::cout << std::endl << "Elapsed time: " << timerInstance.elapsed() <<
    std::endl;
}

BOOST_AUTO_TEST_CASE(regionGrowingBaatzStrategy_no_data_test)
{
  /* Progress interface */

  te::common::ConsoleProgressViewer progressViewerInstance;

  /* Open input raster */

  std::map<std::string, std::string> inputRasterInfo;
  inputRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers2b_rgb342_crop_with_color_table_and_nodata.tif";

  boost::shared_ptr< te::rst::Raster > inputRasterPointer ( te::rst::RasterFactory::open(
    inputRasterInfo ) );
  BOOST_CHECK( inputRasterPointer.get() );

  /* Access a raster datasource to create the output raster */

  std::map<std::string, std::string> outputRasterInfo;
  outputRasterInfo["URI"] = "terralib_unittest_rp_Segmenter_RegionGrowingBaatzStrategy_no_data_Test.tif";

  /* Creating the algorithm parameters */

  te::rp::SegmenterRegionGrowingBaatzStrategy::Parameters strategyParameters;
  strategyParameters.m_minSegmentSize = 100;
  strategyParameters.m_segmentsSimilarityThreshold = 0.25;
  strategyParameters.m_bandsWeights.resize( 1, 1.0 );
  strategyParameters.m_colorWeight = 0.75;
  strategyParameters.m_compactnessWeight = 0.5;
  strategyParameters.m_segmentsSimIncreaseSteps = 3;
  strategyParameters.m_enableLocalMutualBestFitting = false;

  te::rp::Segmenter::InputParameters algoInputParams;
  algoInputParams.m_inputRasterPtr = inputRasterPointer.get();
  algoInputParams.m_inputRasterBands.push_back( 0 );
  algoInputParams.m_enableThreadedProcessing = false;
  algoInputParams.m_maxSegThreads = 0;
  algoInputParams.m_enableBlockProcessing = false;
  algoInputParams.m_blocksOverlapPercent = 0;
  algoInputParams.m_maxBlockSize = 0;
  algoInputParams.m_strategyName = "RegionGrowingBaatz";
  algoInputParams.setSegStrategyParams( strategyParameters );
  algoInputParams.m_enableProgress = true;

  te::rp::Segmenter::OutputParameters algoOutputParams;
  algoOutputParams.m_rInfo = outputRasterInfo;
  algoOutputParams.m_rType = "GDAL";

  /* Executing the algorithm */

  te::rp::Segmenter algorithmInstance;

  BOOST_CHECK( algorithmInstance.initialize( algoInputParams ) );
  boost::timer timerInstance;
  BOOST_CHECK( algorithmInstance.execute( algoOutputParams ) );
  std::cout << std::endl << "Elapsed time: " << timerInstance.elapsed() <<
    std::endl;
}

BOOST_AUTO_TEST_CASE(regionGrowingBaatzStrategyBlockProcessing_test)
{
  /* Progress interface */

  te::common::ConsoleProgressViewer progressViewerInstance;
  
  /* Open input raster */
  
  std::map<std::string, std::string> inputRasterInfo;
  inputRasterInfo["URI"] = TERRALIB_DATA_DIR "/geotiff/cbers2b_rgb342_crop.tif";
  
  boost::shared_ptr< te::rst::Raster > inputRasterPointer ( te::rst::RasterFactory::open(
    inputRasterInfo ) );
  BOOST_CHECK( inputRasterPointer.get() );
  
  /* Access a raster datasource to create the output raster */
  
  std::map<std::string, std::string> outputRasterInfo;
  outputRasterInfo["URI"] = "terralib_unittest_rp_Segmenter_RegionGrowingBaatzStrategyBlockProcessing_Test.tif";
  
  /* Creating the algorithm parameters */
  
  te::rp::SegmenterRegionGrowingBaatzStrategy::Parameters strategyParameters;
  strategyParameters.m_minSegmentSize = 100;
  strategyParameters.m_segmentsSimilarityThreshold = 0.03;
  strategyParameters.m_bandsWeights.resize(
    (unsigned int)inputRasterPointer->getNumberOfBands(),
    1.0 / ((double)inputRasterPointer->getNumberOfBands()) );
  strategyParameters.m_colorWeight = 1.0;
  strategyParameters.m_compactnessWeight = 0.5;
  strategyParameters.m_segmentsSimIncreaseSteps = 10;
  
  te::rp::Segmenter::InputParameters algoInputParams;
  algoInputParams.m_inputRasterPtr = inputRasterPointer.get();
  algoInputParams.m_inputRasterBands.push_back( 0 );
  algoInputParams.m_inputRasterBands.push_back( 1 );
  algoInputParams.m_inputRasterBands.push_back( 2 );
  algoInputParams.m_enableThreadedProcessing = false;
  algoInputParams.m_maxSegThreads = 0;
  algoInputParams.m_enableBlockProcessing = true;
  algoInputParams.m_blocksOverlapPercent = 20;
  algoInputParams.m_maxBlockSize = 100;
  algoInputParams.m_strategyName = "RegionGrowingBaatz";
  algoInputParams.setSegStrategyParams( strategyParameters );
  algoInputParams.m_enableProgress = true;
  
  te::rp::Segmenter::OutputParameters algoOutputParams;
  algoOutputParams.m_rInfo = outputRasterInfo;
  algoOutputParams.m_rType = "GDAL";
  
  /* Executing the algorithm */
  
  te::rp::Segmenter algorithmInstance;
  
  BOOST_CHECK( algorithmInstance.initialize( algoInputParams ) );
  BOOST_CHECK( algorithmInstance.execute( algoOutputParams ) );
}

BOOST_AUTO_TEST_SUITE_END()
