/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/unittest/srs/TsSrs.cpp

  \brief A test suite for the Spatial Reference System Class.

 */

#include <terralib/srs.h>

#include <boost/test/unit_test.hpp>

BOOST_AUTO_TEST_SUITE( srs_tests )

BOOST_AUTO_TEST_CASE( converter_EPSG_32721_and_EPSG_31981_test )
{
  te::srs::Converter conv( 32721, 31981 );

  double xIn[ 1 ];
  xIn[ 0 ] = 1645118.892241;
  double yIn[ 1 ];
  yIn[ 0 ]= 7389079.630977;

  double xOut1[ 1 ];
  xOut1[ 0 ]= 0;
  double yOut1[ 1 ];
  yOut1[ 0 ]= 0;

  BOOST_REQUIRE( conv.convert( xIn, yIn, xOut1, yOut1, 1, 1 ) );

  BOOST_CHECK_CLOSE( xOut1[ 0 ], 1645118.8922440868, 0.0001 );
  BOOST_CHECK_CLOSE( yOut1[ 0 ], 7389079.631054824, 0.0001 );

  double xOut2[ 1 ];
  xOut2[ 0 ]= 0;
  double yOut2[ 1 ];
  yOut2[ 0 ]= 0;

  BOOST_REQUIRE( conv.invert( xOut1, yOut1, xOut2, yOut2, 1, 1 ) );

  BOOST_CHECK_CLOSE( xOut2[ 0 ], xIn[ 0 ], 0.0001 );
  BOOST_CHECK_CLOSE( yOut2[ 0 ], yIn[ 0 ], 0.0001 );
}

BOOST_AUTO_TEST_CASE( converter_EPSG_32721_and_EPSG_4326_test )
{
  te::srs::Converter conv( 32721, 4326 );

  double xIn[ 1 ];
  xIn[ 0 ] = 1645118.892241;
  double yIn[ 1 ];
  yIn[ 0 ]= 7389079.630977;

  double xOut1[ 1 ];
  xOut1[ 0 ]= 0;
  double yOut1[ 1 ];
  yOut1[ 0 ]= 0;

  BOOST_REQUIRE( conv.convert( xIn, yIn, xOut1, yOut1, 1, 1 ) );

  BOOST_CHECK_CLOSE( xOut1[ 0 ], -45.8577055, 0.0001 );
  BOOST_CHECK_CLOSE( yOut1[ 0 ], -23.209032, 0.0001 );

  double xOut2[ 1 ];
  xOut2[ 0 ]= 0;
  double yOut2[ 1 ];
  yOut2[ 0 ]= 0;

  BOOST_REQUIRE( conv.invert( xOut1, yOut1, xOut2, yOut2, 1, 1 ) );

  BOOST_CHECK_CLOSE( xOut2[ 0 ], xIn[ 0 ], 0.0001 );
  BOOST_CHECK_CLOSE( yOut2[ 0 ], yIn[ 0 ], 0.0001 );
}

BOOST_AUTO_TEST_CASE( converter_EPSG_4225_and_EPSG_4326_test )
{
  te::srs::Converter conv( 4225, 4326 );

  double xIn[ 1 ];
  xIn[ 0 ] = -45.8574008;
  double yIn[ 1 ];
  yIn[ 0 ]= -23.2086862;

  double xOut1[ 1 ];
  xOut1[ 0 ]= 0;
  double yOut1[ 1 ];
  yOut1[ 0 ]= 0;

  BOOST_REQUIRE( conv.convert( xIn, yIn, xOut1, yOut1, 1, 1 ) );

  BOOST_CHECK_CLOSE( xOut1[ 0 ], -45.8577054, 0.001 );
  BOOST_CHECK_CLOSE( yOut1[ 0 ], -23.2090321, 0.001 );

  double xOut2[ 1 ];
  xOut2[ 0 ]= 0;
  double yOut2[ 1 ];
  yOut2[ 0 ]= 0;

  BOOST_REQUIRE( conv.invert( xOut1, yOut1, xOut2, yOut2, 1, 1 ) );

  BOOST_CHECK_CLOSE( xOut2[ 0 ], xIn[ 0 ], 0.0001 );
  BOOST_CHECK_CLOSE( yOut2[ 0 ], yIn[ 0 ], 0.0001 );
}

BOOST_AUTO_TEST_CASE( converter_EPSG_4326_and_EPSG_4225_test )
{
  te::srs::Converter conv( 4326, 4225 );

  double xIn[ 1 ];
  xIn[ 0 ] = -45.8577054;
  double yIn[ 1 ];
  yIn[ 0 ]= -23.2090321;

  double xOut1[ 1 ];
  xOut1[ 0 ]= 0;
  double yOut1[ 1 ];
  yOut1[ 0 ]= 0;

  BOOST_REQUIRE( conv.convert( xIn, yIn, xOut1, yOut1, 1, 1 ) );

  BOOST_CHECK_CLOSE( xOut1[ 0 ], -45.8574008, 0.001 );
  BOOST_CHECK_CLOSE( yOut1[ 0 ], -23.2086862, 0.001 );

  double xOut2[ 1 ];
  xOut2[ 0 ]= 0;
  double yOut2[ 1 ];
  yOut2[ 0 ]= 0;

  BOOST_REQUIRE( conv.invert( xOut1, yOut1, xOut2, yOut2, 1, 1 ) );

  BOOST_CHECK_CLOSE( xOut2[ 0 ], xIn[ 0 ], 0.0001 );
  BOOST_CHECK_CLOSE( yOut2[ 0 ], yIn[ 0 ], 0.0001 );
}

BOOST_AUTO_TEST_CASE( converter_EPSG_4326_and_EPSG_4225_memory_test )
{
  for( unsigned long int step = 0 ; step < std::numeric_limits< unsigned long int >::max() ;
    ++step )
  {
    te::srs::Converter conv;

    conv.setSourceSRID( 4326 );
    conv.setTargetSRID( 4225 );

    double xIn[ 1 ];
    xIn[ 0 ] = -45.8577054;
    double yIn[ 1 ];
    yIn[ 0 ]= -23.2090321;

    double xOut1[ 1 ];
    xOut1[ 0 ]= 0;
    double yOut1[ 1 ];
    yOut1[ 0 ]= 0;

    BOOST_REQUIRE( conv.convert( xIn, yIn, xOut1, yOut1, 1, 1 ) );

    BOOST_CHECK_CLOSE( xOut1[ 0 ], -45.8574008, 0.001 );
    BOOST_CHECK_CLOSE( yOut1[ 0 ], -23.2086862, 0.001 );

    double xOut2[ 1 ];
    xOut2[ 0 ]= 0;
    double yOut2[ 1 ];
    yOut2[ 0 ]= 0;

    BOOST_REQUIRE( conv.invert( xOut1, yOut1, xOut2, yOut2, 1, 1 ) );

    BOOST_CHECK_CLOSE( xOut2[ 0 ], xIn[ 0 ], 0.0001 );
    BOOST_CHECK_CLOSE( yOut2[ 0 ], yIn[ 0 ], 0.0001 );
  }
}

BOOST_AUTO_TEST_SUITE_END()
