/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

//TerraLib
#include <terralib_buildconfig.h>
#include <terralib/common.h>
#include <terralib/common/STLUtils.h>
#include <terralib/vp.h>
#include <terralib/geometry.h>
#include <terralib/dataaccess.h>
#include <terralib/dataaccess/utils/DataAccess.h>
#include "../../src/terralib/geometry.h"
#include "../../src/terralib/memory.h"

// Boost
#include <boost/test/unit_test.hpp>
#include <boost/uuid/uuid_io.hpp>

#include <cstdio>

void differenceoperation_test(
  const bool useThreads,
  const std::string& outDataset1Name,
  const std::string& outDataset2Name,
  const std::string& outFileExtension,
  const te::gm::GeomType outGeomType )
{
  std::string outDataset1FileName =  outDataset1Name + "." +
    outFileExtension;
  std::string outDataset2FileName =  outDataset2Name + "." +
    outFileExtension;

  {
    const std::string inDataset1Name = "munic_2001";
    const std::string inDataset2Name = "munic_goias";

    const std::string inFile1Name = TERRALIB_DATA_DIR
      "/shape/" + inDataset1Name + ".shp";
    const std::string inFile2Name = TERRALIB_DATA_DIR
      "/shape/" + inDataset2Name + ".shp";

    const std::string inConnInfo1("file://" + inFile1Name );
    const std::string inConnInfo2("file://" + inFile2Name );

    const te::core::URI inputURI1( inConnInfo1 );
    const te::core::URI inputURI2( inConnInfo2 );

    te::da::DataSourcePtr inputDataSourcePtr1(
      te::da::DataSourceFactory::make("OGR", inputURI1) );
    BOOST_REQUIRE( inputDataSourcePtr1.get() != 0 );
    inputDataSourcePtr1->open();
    BOOST_REQUIRE( inputDataSourcePtr1->isOpened() );

    te::da::DataSourcePtr inputDataSourcePtr2(
      te::da::DataSourceFactory::make("OGR", inputURI2) );
    BOOST_REQUIRE( inputDataSourcePtr2.get() != 0 );
    inputDataSourcePtr2->open();
    BOOST_REQUIRE( inputDataSourcePtr2->isOpened() );

    const te::core::URI outputURI1( "file://" + outDataset1FileName );
    const te::core::URI outputURI2( "file://" + outDataset2FileName );

    const std::string outId1 = te::common::GetUniqueID();
    const std::string outId2 = te::common::GetUniqueID();

    te::da::DataSourceInfoPtr outDsInfo1(new te::da::DataSourceInfo);
    outDsInfo1->setConnInfo( outputURI1.uri() );
    outDsInfo1->setTitle( outDataset1Name );
    outDsInfo1->setAccessDriver("OGR");
    outDsInfo1->setType("OGR");
    outDsInfo1->setDescription( outDataset1Name );
    outDsInfo1->setId(outId1);

    te::da::DataSourceInfoPtr outDsInfo2(new te::da::DataSourceInfo);
    outDsInfo2->setConnInfo( outputURI2.uri() );
    outDsInfo2->setTitle( outDataset2Name );
    outDsInfo2->setAccessDriver("OGR");
    outDsInfo2->setType("OGR");
    outDsInfo2->setDescription( outDataset2Name );
    outDsInfo2->setId(outId2);

    te::da::DataSourceInfoManager::getInstance().add(outDsInfo1);
    te::da::DataSourceInfoManager::getInstance().add(outDsInfo2);

    te::da::DataSourcePtr outputDataSourcePtr1 =
      te::da::DataSourceManager::getInstance().get( outId1, "OGR",
      outputURI1.uri() );
    BOOST_REQUIRE( outputDataSourcePtr1.get() != nullptr );
    te::da::DataSourcePtr outputDataSourcePtr2 =
      te::da::DataSourceManager::getInstance().get( outId2, "OGR",
      outputURI2.uri() );
    BOOST_REQUIRE( outputDataSourcePtr2.get() != nullptr );

    te::da::DataAccessPtr inputDataAccessPtr1( new te::da::DataAccess(
      inputDataSourcePtr1, inDataset1Name ) );
    te::da::DataAccessPtr inputDataAccessPtr2( new te::da::DataAccess(
      inputDataSourcePtr2, inDataset2Name ) );

    te::da::DataAccessPtr outputDataAccessPtr1( new te::da::DataAccess(
      outputDataSourcePtr1, outDataset1Name ) );
    te::da::DataAccessPtr outputDataAccessPtr2( new te::da::DataAccess(
      outputDataSourcePtr2, outDataset2Name ) );

    const std::string operationName( "Difference" );

    std::unique_ptr<te::vp::AbstractOperation> operation(
      te::vp::CreateOperation( operationName ) );
    BOOST_REQUIRE( operation.get() != nullptr );

    std::unique_ptr<te::vp::AbstractParameters> parameters =
      operation->getCapabilities()->createParameters();
    BOOST_REQUIRE( parameters.get() != nullptr );
    parameters->addInputDataAccess( inputDataAccessPtr1.get() );
    parameters->addInputDataAccess( inputDataAccessPtr2.get() );
    parameters->addOutputDataAccess( outputDataAccessPtr1.get() );
    parameters->addOutputDataAccess( outputDataAccessPtr2.get() );

    std::unique_ptr< te::common::TaskProgress > task(
      new te::common::TaskProgress() );

    operation->setParameters(parameters.release());
    if( outGeomType != te::gm::UnknownGeometryType )
      operation->setOutputGeometryType( outGeomType );

    te::vp::AbstractDataHandler* dataHandler = te::vp::CreateDataHandler(
      operation->getCapabilities()->getDataHandlerName());
    BOOST_REQUIRE( dataHandler != nullptr );

    te::vp::OperationThreadManager operationTM( operation.release(),
      dataHandler );
    operationTM.setTask( task.get() );
    operationTM.start( useThreads );

    std::cout << std::endl << "test report : ["
      << operationTM.getOperation()->getReport()->toString() << "]"
      << std::endl;

    BOOST_REQUIRE( operationTM.getOperation()->getParameters() != nullptr );
    BOOST_REQUIRE(
      operationTM.getOperation()->getParameters()->getOperationResult()
      ==
      te::vp::OperationResult::OP_SUCCESS );
  }
}

BOOST_AUTO_TEST_SUITE( difference_tests )


BOOST_AUTO_TEST_CASE( differenceoperation_no_threads_shp_test )
{
  std::string outDataset1Name = "differenceoperation_no_threads_shp_test_1";
  std::string outDataset2Name = "differenceoperation_no_threads_shp_test_2";

  remove( ( outDataset1Name + ".shp" ).c_str() );
  remove( ( outDataset1Name + ".shx" ).c_str() );
  remove( ( outDataset1Name + ".dbf" ).c_str() );
  remove( ( outDataset1Name + ".prj" ).c_str() );

  remove( ( outDataset2Name + ".shp" ).c_str() );
  remove( ( outDataset2Name + ".shx" ).c_str() );
  remove( ( outDataset2Name + ".dbf" ).c_str() );
  remove( ( outDataset2Name + ".prj" ).c_str() );

  differenceoperation_test( false, outDataset1Name, outDataset2Name, "shp",
    te::gm::UnknownGeometryType );
}

BOOST_AUTO_TEST_SUITE_END()
