/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

//TerraLib
#include <terralib_buildconfig.h>
#include <terralib/common.h>
#include <terralib/common/STLUtils.h>
#include <terralib/vp.h>
#include <terralib/geometry.h>
#include <terralib/dataaccess.h>
#include <terralib/dataaccess/utils/DataAccess.h>
#include "../../src/terralib/geometry.h"
#include "../../src/terralib/memory.h"

// Boost
#include <boost/test/unit_test.hpp>
#include <boost/uuid/uuid_io.hpp>

#include <cstdio>

bool loadSHPFile( const std::string& shpFileName,
  std::vector< te::gm::Geometry* >& geomPtrs )
{
  std::string connInfo("file://");
  connInfo += shpFileName;
  std::unique_ptr< te::da::DataSource > ogrDataSourcePtr(
    te::da::DataSourceFactory::make("OGR", connInfo) );
  if( ogrDataSourcePtr.get() == 0 ) return false;

  ogrDataSourcePtr->open();
  
  if( ! ogrDataSourcePtr->isOpened() ) return false;
  
  std::vector<std::string> dataSetNames = ogrDataSourcePtr->getDataSetNames();
  std::unique_ptr< te::da::DataSourceTransactor > transactorPtr =
    ogrDataSourcePtr->getTransactor();
  
  for( unsigned int dataSetNamesIdx = 0 ; dataSetNamesIdx < dataSetNames.size() ;
    ++dataSetNamesIdx )
  {
    std::unique_ptr<te::da::DataSet> datasetPtr = transactorPtr->getDataSet(
      dataSetNames[ dataSetNamesIdx ] );
    
    const std::size_t nProperties = datasetPtr->getNumProperties();
    
    if( !datasetPtr->moveBeforeFirst() ) return false;
    
    while( datasetPtr->moveNext() )
    {
      for( std::size_t propIdx = 0 ; propIdx < nProperties ; ++propIdx )
      {
        if( datasetPtr->getPropertyDataType( propIdx ) == te::dt::GEOMETRY_TYPE )
        {
          std::unique_ptr< te::gm::Geometry > geomPtr = datasetPtr->getGeometry(
            propIdx );
          
          if( geomPtr->getGeomTypeId() == te::gm::PolygonType )
          {
            geomPtrs.push_back( geomPtr.release() );
          }
          else if( geomPtr->getGeomTypeId() == te::gm::MultiPolygonType )
          {
            te::gm::MultiPolygon const * const mPolPtr = (te::gm::MultiPolygon const *)
              geomPtr.get();
              
            for( std::size_t gIdx = 0 ; gIdx < mPolPtr->getNumGeometries() ;
              ++gIdx )
            {
              geomPtrs.push_back( new te::gm::Polygon( *( (te::gm::Polygon*)
                mPolPtr->getGeometryN( gIdx ) ) ) );
            }
          }
        }
      }
    }
  }
  
  return true;
}

bool checkValidity( std::vector< te::gm::Geometry* >& geomPtrs )
{
  if( geomPtrs.empty() )
  {
    return false;    
  }
  else
  {
    te::gm::TopologyValidationError error1;
    for( std::size_t geomPtrsIdx = 0 ; geomPtrsIdx < geomPtrs.size() ; ++geomPtrsIdx )
    {
      if( !te::gm::Validation::CheckValidity( geomPtrs[ geomPtrsIdx ], error1 ) )
      {
        return false;
      }
    }
    
    return true;
  }
}

void makevalidoperation_test( bool useThreads )
{
  std::string outDatasetNameOK = 
    "makevalidoperation_tests_makevalidoperation_test_OK";
  if( useThreads ) outDatasetNameOK = outDatasetNameOK + "_threads";
  std::string outFileNameOK =  outDatasetNameOK + ".shp";
  
  std::string outDatasetNameError = 
    "makevalidoperation_tests_makevalidoperation_test_Error";      
  if( useThreads ) outDatasetNameError = outDatasetNameError + "_threads";
  std::string outFileNameError =  outDatasetNameError + ".shp";  
  
  {
    const std::string inDatasetName = 
      "invalid_geometries_hole_lies_outside_shell";    
      
    const std::string inFileName = TERRALIB_DATA_DIR 
      "/shape/" + inDatasetName + ".shp";  
    
    
    const std::string inConnInfo("file://" + inFileName );
    
    const te::core::URI inputURI( inConnInfo );
    
    std::remove( ( outDatasetNameOK + ".qix" ).c_str() );
    std::remove( ( outDatasetNameOK + ".shx" ).c_str() );
    std::remove( ( outDatasetNameOK + ".shp" ).c_str() );
    std::remove( ( outDatasetNameOK + ".dbf" ).c_str() );
    std::remove( ( outDatasetNameOK + ".prj" ).c_str() );
    
    std::remove( ( outDatasetNameError + ".qix" ).c_str() );
    std::remove( ( outDatasetNameError + ".shx" ).c_str() );
    std::remove( ( outDatasetNameError + ".shp" ).c_str() );
    std::remove( ( outDatasetNameError + ".dbf" ).c_str() );
    std::remove( ( outDatasetNameError + ".prj" ).c_str() );    
    
    te::da::DataSourcePtr inputDataSourcePtr(
      te::da::DataSourceFactory::make("OGR", inputURI) );
    BOOST_REQUIRE( inputDataSourcePtr.get() != 0 );
    inputDataSourcePtr->open();
    BOOST_REQUIRE( inputDataSourcePtr->isOpened() );
    
    const te::core::URI outputURIOK( "file://" + outFileNameOK );
    const te::core::URI outputURIError( "file://" + outFileNameError );
    
    const std::string idOK = te::common::GetUniqueID();
    const std::string idError = te::common::GetUniqueID();

    te::da::DataSourceInfoPtr dsInfoOK(new te::da::DataSourceInfo);
    dsInfoOK->setConnInfo( outputURIOK.uri() );
    dsInfoOK->setTitle( outDatasetNameOK );
    dsInfoOK->setAccessDriver("OGR");
    dsInfoOK->setType("OGR");
    dsInfoOK->setDescription( outDatasetNameOK );
    dsInfoOK->setId(idOK);
    
    te::da::DataSourceInfoPtr dsInfoError(new te::da::DataSourceInfo);
    dsInfoError->setConnInfo( outputURIError.uri() );
    dsInfoError->setTitle( outDatasetNameError );
    dsInfoError->setAccessDriver("OGR");
    dsInfoError->setType("OGR");
    dsInfoError->setDescription( outDatasetNameError );
    dsInfoError->setId(idError);    

    te::da::DataSourceInfoManager::getInstance().add(dsInfoOK);
    te::da::DataSourceInfoManager::getInstance().add(dsInfoError);
    
    te::da::DataSourcePtr outputDataSourceOKPtr = 
      te::da::DataSourceManager::getInstance().get( idOK, "OGR", outputURIOK.uri() );
    te::da::DataSourcePtr outputDataSourceErrorPtr = 
      te::da::DataSourceManager::getInstance().get( idError, "OGR", outputURIError.uri() );      

    te::da::DataAccessPtr inputDataAccessPtr( new te::da::DataAccess( 
      inputDataSourcePtr, inDatasetName ) );
    
    te::da::DataAccessPtr outputDataAccessOKPtr( new te::da::DataAccess( 
      outputDataSourceOKPtr, outDatasetNameOK ) );
    te::da::DataAccessPtr outputDataAccessErrorPtr( new te::da::DataAccess( 
      outputDataSourceErrorPtr, outDatasetNameError ) );    
    
    const std::string operationName( "MakeValid" );
    
    std::unique_ptr<te::vp::AbstractOperation> operation(
      te::vp::CreateOperation( operationName ) );
    BOOST_REQUIRE( operation.get() != nullptr );
    
    std::unique_ptr<te::vp::AbstractParameters> parameters = 
      operation->getCapabilities()->createParameters();    
    BOOST_REQUIRE( parameters.get() != nullptr );
    parameters->addInputDataAccess( inputDataAccessPtr.get() );
    parameters->addOutputDataAccess( outputDataAccessOKPtr.get() );
    parameters->addOutputDataAccess( outputDataAccessErrorPtr.get() );
    
    std::unique_ptr< te::common::TaskProgress > task( new te::common::TaskProgress() );
    
    operation->setParameters(parameters.release());
    operation->setOutputGeometryType( te::gm::PolygonType );
    
    te::vp::AbstractDataHandler* dataHandler = te::vp::CreateDataHandler(
      operation->getCapabilities()->getDataHandlerName());
    BOOST_REQUIRE( dataHandler != nullptr );  
    
    std::unique_ptr<te::vp::OperationThreadManager> operationTM(
      new te::vp::OperationThreadManager(operation.release(), dataHandler) );
    operationTM->setTask( task.get() );
    operationTM->start( useThreads );
    
    std::cout << std::endl << "makevalidoperation_no_threads_test report : [" 
      << operationTM->getOperation()->getReport()->toString() << "]" 
      << std::endl;
  }
    
  {
    std::vector< te::gm::Geometry* > geomPtrs;
    BOOST_REQUIRE( loadSHPFile(  outFileNameOK, geomPtrs ) );
    BOOST_REQUIRE( checkValidity( geomPtrs ) );
    te::common::FreeContents( geomPtrs ); 
  }
}

BOOST_AUTO_TEST_SUITE( makevalidoperation_tests )

BOOST_AUTO_TEST_CASE( makevalidoperation_no_threads_test )
{
  makevalidoperation_test( false );
}

BOOST_AUTO_TEST_CASE( makevalidoperation_with_threads_test )
{
  makevalidoperation_test( true );
}

BOOST_AUTO_TEST_SUITE_END()
