/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

 This file is part of the TerraLib - a Framework for building GIS enabled applications.

 TerraLib is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, either version 3 of the License,
 or (at your option) any later version.

 TerraLib is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with TerraLib. See COPYING. If not, write to
 TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
 \file terralib/itest/raster/ITsRaster.cpp

 \brief A integration test suit for the TerraLib VP Module.

 \author Mario Rocco Pettinati
 */

// TerraLib
#include <terralib_buildconfig.h>

// TerraLib
#include <terralib/common/TerraLib.h>
#include <terralib/common/STLUtils.h>
#include <terralib/geometry.h>
#include <terralib/dataaccess.h>
#include <terralib/geometry/GeometryFunctions.h>
#include <terralib/geometry/GeometryProperty.h>
#include <terralib/vp/CommonDataStructures.h>
#include <terralib/vp/Feature.h>
#include <terralib/vp/GeometrySnapper.h>
#include <terralib/vp/OperationService.h>
#include <terralib/vp/Utils.h>
#include "TestUtils.h"

// GTest and GMock
#include <gmock/gmock.h>
#include <gtest/gtest.h>

//STL include files
#include <thread>

namespace
{
  class ItsPrepareToOverlay : public ::testing::Test
  {
    virtual void SetUp()
    {
    }

    virtual void TearDown()
    {
    }
  };

  TEST_F(ItsPrepareToOverlay, PerpendicularDistance)
  {
    {
      te::gm::Coord2D c1(0, 0);
      te::gm::Coord2D c2(10, 0);
      te::gm::Coord2D cInput(5, 5);
      te::gm::Coord2D cInter;
      bool isOnSegment = false;
      double distance = te::gm::Distance::PerpendicularDistance(c1, c2, cInput, cInter, isOnSegment);
      ASSERT_EQ(distance, 5.0);

      te::gm::Coord2D expectedCInter(5, 0);
      ASSERT_TRUE(cInter == expectedCInter);

      ASSERT_TRUE(te::gm::Distance::IsOnLine(cInter, c1, c2));
    }

    {
      te::gm::Coord2D c1(0, 0);
      te::gm::Coord2D c2(0, 0);
      te::gm::Coord2D cInput(0, 5);
      te::gm::Coord2D cInter;
      bool isOnSegment = false;
      double distance = te::gm::Distance::PerpendicularDistance(c1, c2, cInput, cInter, isOnSegment);
      ASSERT_EQ(distance, 5.0);

      te::gm::Coord2D expectedCInter(0, 0);
      ASSERT_TRUE(cInter == expectedCInter);

      ASSERT_TRUE(te::gm::Distance::IsOnLine(cInter, c1, c2));
    }
  }

  TEST_F(ItsPrepareToOverlay, AddPerpendicularPoints)
  {
    std::string wkt("LINESTRING(0 0, 2 1, 4 1, 6 1, 10 5, 10 0, 7 0, 5 0, 3 0, 0 0)");
    std::unique_ptr<te::gm::Geometry> geometry(te::gm::WKTReader::read(wkt.c_str()));

    te::gm::LineString* lineString = static_cast<te::gm::LineString*>(geometry.get());

    for (std::size_t i = 1; i < lineString->getNPoints(); ++i)
    {
      te::gm::Coord2D c1(lineString->getX(i - 1), lineString->getY(i - 1));
      te::gm::Coord2D c2(lineString->getX(i), lineString->getY(i));
    }
  }

  TEST_F(ItsPrepareToOverlay, PrepareGeometriesAlmostEqual)
  {
    std::string wktA("MULTIPOLYGON (((-69.25521841581619 -3.7223770767775806, -69.25521840431624 -3.722337128949591, -69.25521835223174 -3.7221561921404955, -69.2552183001498 -3.721975255330656, -69.25521824807042 -3.7217943185200726, -69.25521819599356 -3.7216133817087447, -69.25539830642957 -3.721613329375788, -69.25557841686302 -3.7216132770059143, -69.2557585272939 -3.721613224599121, -69.25593863772225 -3.7216131721554113, -69.25611874814804 -3.7216131196747826, -69.25629885857126 -3.7216130671572363, -69.25647896899191 -3.7216130146027724, -69.2564789166604 -3.7214320778085965, -69.25647886433141 -3.7212511410136773, -69.25647881200503 -3.7210702042180137, -69.25647875968117 -3.7208892674216067, -69.25647870735989 -3.7207083306244555, -69.25647865504116 -3.720527393826561, -69.25647860272501 -3.720346457027922, -69.25629849256154 -3.720346509564452, -69.2562984402847 -3.720165572762507, -69.2561183301554 -3.720165625259572, -69.25593822002355 -3.720165677719733, -69.25575810988914 -3.720165730142991, -69.25557799975219 -3.720165782529345, -69.25557794762483 -3.719984845716427, -69.25539783752205 -3.7199848980633248, -69.25539778543397 -3.7198039612471083, -69.25539773334845 -3.7196230244301502, -69.25539768126548 -3.719442087612447, -69.25539762918504 -3.7192611507940008, -69.25521751922663 -3.7192612030937915, -69.25503740926567 -3.719261255356688, -69.25503735726124 -3.7190803185323986, -69.25503730525936 -3.7188993817073643, -69.25503725326006 -3.718718444881588, -69.25485714340671 -3.7187184970999474, -69.25485709144665 -3.7185375602708786, -69.25467698162748 -3.7185376124498024, -69.25449687180578 -3.718537664591839, -69.25431676198153 -3.718537716696988, -69.25413665215474 -3.71853776876525, -69.25395654232541 -3.718537820796625, -69.25377643249355 -3.7185378727911123, -69.25359632265915 -3.7185379247487114, -69.25341621282222 -3.7185379766694244, -69.25323610298275 -3.718538028553249, -69.25305599314075 -3.7185380804001866, -69.25287588329624 -3.7185381322102367, -69.25269577344919 -3.7185381839833997, -69.25251566359962 -3.7185382357196746, -69.25233555374753 -3.718538287419063, -69.25233550230406 -3.7183573505537755, -69.2523354508631 -3.718176413687743, -69.25233539942468 -3.7179954768209678, -69.25233534798879 -3.717814539953449, -69.2523352965554 -3.717633603085186, -69.25233524512454 -3.717452666216182, -69.2523351936962 -3.717271729346433, -69.25215508409858 -3.717271780991289, -69.25197497449842 -3.7172718325992706, -69.25197492314602 -3.7170908957237403, -69.25179481358006 -3.71709094729233, -69.25161470401159 -3.717090998824047, -69.2516146527351 -3.71691006194274, -69.25161460146113 -3.7167291250606906, -69.25161455018966 -3.7165481881778986, -69.25143444072879 -3.7165482396652028, -69.2512543312654 -3.7165482911156382, -69.25107422179948 -3.7165483425292067, -69.25089411233108 -3.7165483939059074, -69.25071400286016 -3.7165484452457402, -69.25053389338673 -3.7165484965487057, -69.25035378391082 -3.7165485478148033, -69.25035383492538 -3.716729484715151, -69.25017372541024 -3.7167295359468815, -69.25 -3.7167295853271054, -69.25 -3.719262702125634, -69.25017443936267 -3.719262652508731, -69.25035454939179 -3.71926260124199, -69.25035460044386 -3.7194435381311886, -69.25035465149843 -3.719624475019645, -69.2503547025555 -3.7198054119073563, -69.25035475361508 -3.719986348794325, -69.25035480467716 -3.720167285680549, -69.25035485574173 -3.72034822256603, -69.25035490680881 -3.720529159450767, -69.25035495787839 -3.7207100963347597, -69.25035500895046 -3.7208910332180096, -69.25053511930768 -3.720890981891849, -69.2507152296624 -3.7208909305287787, -69.25071528081047 -3.7210718674062773, -69.25071533196103 -3.7212528042830315, -69.25089544238674 -3.7212527528780375, -69.25107555280994 -3.7212527014361294, -69.25125566323062 -3.721252649957306, -69.2514357736488 -3.7212525984415694, -69.25161588406446 -3.7212525468889175, -69.2517959944776 -3.721252495299351, -69.25197610488826 -3.7212524436728707, -69.25215621529637 -3.7212523920094758, -69.25233632570198 -3.721252340309167, -69.25251643610505 -3.7212522885719426, -69.25251648762558 -3.72143322542281, -69.25251653914866 -3.7216141622729344, -69.2526966496227 -3.7216141104937472, -69.25287676009424 -3.721614058677643, -69.25305687056324 -3.72161400682462, -69.25305692219908 -3.721794943666422, -69.25305697383745 -3.721975880507479, -69.25305702547834 -3.722156817347792, -69.25305707712177 -3.7223377541873623, -69.25305712876771 -3.722518691026187, -69.25323723941796 -3.7225186391235985, -69.25323728024316 -3.7226615577434363, -69.25325695484409 -3.7226587325710683, -69.25328737940644 -3.722654363759059, -69.25328746846691 -3.7226543509704295, -69.25331646292378 -3.7226501875142755, -69.25331650686756 -3.7226501812041723, -69.25331667746092 -3.7226501567078367, -69.25334695011307 -3.722645809709358, -69.25335128887949 -3.7226451866846246, -69.2533535821216 -3.7226448573867503, -69.25335683682063 -3.722644390028567, -69.25347726575063 -3.7226270970483246, -69.25510896895015 -3.7223927927894396, -69.25521841581619 -3.7223770767775806)),((-69.2532881836721 -3.722654248270612, -69.25328996344369 -3.722653992704488, -69.25329583115104 -3.7226531501316353, -69.2532986310817 -3.722652748075867, -69.2532881836721 -3.722654248270612)),((-69.25330717747454 -3.722651520857438, -69.25331018947243 -3.7226510883499, -69.25331265494788 -3.722650734320199, -69.25331600750248 -3.7226502529104533, -69.25331612142243 -3.722650236552128, -69.25330717747454 -3.722651520857438)),((-69.25336452745606 -3.722643285692538, -69.2533768394287 -3.7226415177560495, -69.2533785869394 -3.7226412668224227, -69.25337872240767 -3.7226412473698707, -69.25336452745606 -3.722643285692538)))");
    std::string wktB("POLYGON ((-69.25 -3.719262702125634, -69.25017443936267 -3.719262652508731, -69.25035454939179 -3.71926260124199, -69.25035460044386 -3.7194435381311886, -69.25035465149843 -3.719624475019645, -69.2503547025555 -3.7198054119073563, -69.25035475361508 -3.719986348794325, -69.25035480467716 -3.720167285680549, -69.25035485574173 -3.72034822256603, -69.25035490680881 -3.720529159450767, -69.25035495787839 -3.7207100963347597, -69.25035500895046 -3.7208910332180096, -69.25053511930768 -3.720890981891849, -69.2507152296624 -3.7208909305287787, -69.25071528081047 -3.7210718674062773, -69.25071533196103 -3.7212528042830315, -69.25089544238674 -3.7212527528780375, -69.25107555280994 -3.7212527014361294, -69.25125566323062 -3.721252649957306, -69.2514357736488 -3.7212525984415694, -69.25161588406446 -3.7212525468889175, -69.2517959944776 -3.721252495299351, -69.25197610488826 -3.7212524436728707, -69.25215621529637 -3.7212523920094758, -69.25233632570198 -3.721252340309167, -69.25251643610505 -3.7212522885719426, -69.25251648762558 -3.72143322542281, -69.25251653914866 -3.7216141622729344, -69.2526966496227 -3.7216141104937472, -69.25287676009424 -3.721614058677643, -69.25305687056324 -3.72161400682462, -69.25305692219908 -3.721794943666422, -69.25305697383745 -3.721975880507479, -69.25305702547834 -3.722156817347792, -69.25305707712177 -3.7223377541873623, -69.25305712876771 -3.722518691026187, -69.25323723941796 -3.7225186391235985, -69.25323728024316 -3.7226615577434363, -69.25325695484409 -3.7226587325710683, -69.25331018947243 -3.7226510883499, -69.25333315897272 -3.722647790046786, -69.25336452745606 -3.722643285692538, -69.2533768394287 -3.7226415177560495, -69.25337958716291 -3.7226411231954275, -69.25347726575063 -3.7226270970483246, -69.25510896895015 -3.7223927927894396, -69.25521296660388 -3.7223778592566816, -69.25521841581619 -3.7223770767775806, -69.25521840431624 -3.722337128949591, -69.25521835223174 -3.7221561921404955, -69.2552183001498 -3.721975255330656, -69.25521824807042 -3.7217943185200726, -69.25521819599356 -3.7216133817087447, -69.25539830642957 -3.721613329375788, -69.25557841686302 -3.7216132770059143, -69.2557585272939 -3.721613224599121, -69.25593863772225 -3.7216131721554113, -69.25611874814804 -3.7216131196747826, -69.25629885857126 -3.7216130671572363, -69.25647896899191 -3.7216130146027724, -69.2564789166604 -3.7214320778085965, -69.25647886433141 -3.7212511410136773, -69.25647881200503 -3.7210702042180137, -69.25647875968117 -3.7208892674216067, -69.25647870735989 -3.7207083306244555, -69.25647865504116 -3.720527393826561, -69.25647860272501 -3.720346457027922, -69.25629849256154 -3.720346509564452, -69.2562984402847 -3.720165572762507, -69.2561183301554 -3.720165625259572, -69.25593822002355 -3.720165677719733, -69.25575810988914 -3.720165730142991, -69.25557799975219 -3.720165782529345, -69.25557794762483 -3.719984845716427, -69.25539783752205 -3.7199848980633248, -69.25539778543397 -3.7198039612471083, -69.25539773334845 -3.7196230244301502, -69.25539768126548 -3.719442087612447, -69.25539762918504 -3.7192611507940008, -69.25521751922663 -3.7192612030937915, -69.25503740926567 -3.719261255356688, -69.25503735726124 -3.7190803185323986, -69.25503730525936 -3.7188993817073643, -69.25503725326006 -3.718718444881588, -69.25485714340671 -3.7187184970999474, -69.25485709144665 -3.7185375602708786, -69.25467698162748 -3.7185376124498024, -69.25449687180578 -3.718537664591839, -69.25431676198153 -3.718537716696988, -69.25413665215474 -3.71853776876525, -69.25395654232541 -3.718537820796625, -69.25377643249355 -3.7185378727911123, -69.25359632265915 -3.7185379247487114, -69.25341621282222 -3.7185379766694244, -69.25323610298275 -3.718538028553249, -69.25305599314075 -3.7185380804001866, -69.25287588329624 -3.7185381322102367, -69.25269577344919 -3.7185381839833997, -69.25251566359962 -3.7185382357196746, -69.25233555374753 -3.718538287419063, -69.25233550230406 -3.7183573505537755, -69.2523354508631 -3.718176413687743, -69.25233539942468 -3.7179954768209678, -69.25233534798879 -3.717814539953449, -69.2523352965554 -3.717633603085186, -69.25233524512454 -3.717452666216182, -69.2523351936962 -3.717271729346433, -69.25215508409858 -3.717271780991289, -69.25197497449842 -3.7172718325992706, -69.25197492314602 -3.7170908957237403, -69.25179481358006 -3.71709094729233, -69.25161470401159 -3.717090998824047, -69.2516146527351 -3.71691006194274, -69.25161460146113 -3.7167291250606906, -69.25161455018966 -3.7165481881778986, -69.25143444072879 -3.7165482396652028, -69.2512543312654 -3.7165482911156382, -69.25107422179948 -3.7165483425292067, -69.25089411233108 -3.7165483939059074, -69.25071400286016 -3.7165484452457402, -69.25053389338673 -3.7165484965487057, -69.25035378391082 -3.7165485478148033, -69.25035383492538 -3.716729484715151, -69.25017372541024 -3.7167295359468815, -69.25 -3.7167295853271054, -69.24999361589262 -3.716729587141743, -69.2498135063725 -3.7167296382997344, -69.24963339684989 -3.7167296894208577, -69.24945328732477 -3.716729740505111, -69.24927317779718 -3.716729791552494, -69.24909306826707 -3.7167298425630078, -69.24909301750942 -3.7165489056451926, -69.24909296675425 -3.716367968726635, -69.24909291600156 -3.716187031807332, -69.24909286525137 -3.716006094887289, -69.24909281450365 -3.7158251579665005, -69.24927292385027 -3.7158251069684365, -69.2494530331944 -3.7158250559335113, -69.24963314253601 -3.7158250048617267, -69.24963309168072 -3.7156440679476725, -69.24963304082793 -3.715463131032874, -69.24963298997763 -3.715282194117333, -69.24963293912982 -3.715101257201051, -69.2496328882845 -3.714920320284026, -69.24945277912632 -3.714920371343343, -69.24927266996563 -3.714920422365809, -69.24909256080245 -3.7149204733514227, -69.24909251006967 -3.714739536426179, -69.24891240094068 -3.714739587372453, -69.24873229180923 -3.7147396382818783, -69.2485862841725 -3.7147396795223435, -69.24900097109597 -3.7166872042824477, -69.2494271261306 -3.7186881351788372, -69.2494909371432 -3.718987747799282, -69.24954952372677 -3.719262830094811, -69.24963410926034 -3.719262806087589, -69.24981421929695 -3.719262754931531, -69.24999432933106 -3.7192627037385786, -69.25 -3.719262702125634))");
    
    std::unique_ptr<te::gm::Geometry> geometryA(te::gm::WKTReader::read(wktA.c_str()));
    std::unique_ptr<te::gm::Geometry> geometryB(te::gm::WKTReader::read(wktB.c_str()));

    te::gm::GeometryVectorConst input;
    input.push_back(geometryA.get());
    input.push_back(geometryB.get());

    double tolerance = te::gm::GetAppropriatePrecision(4674);

    te::gm::GeometryVector output = te::vp::GeometrySnapper::snapGeometries(input, tolerance);
    std::string wktResultA = output.at(0)->toString();
    std::string wktResultB = output.at(1)->toString();
    

    te::common::FreeContents(output);
  }

  TEST_F(ItsPrepareToOverlay, PrepareGeometriesAlmostEqual2)
  {
    std::string wktA("multipolygon(((-64.1451981793532298 3.55296681855199648,-64.1451981793532298 3.57002712739561545,-64.1450918067335323 3.57002712739561545,-64.1450918067335323 3.56991904724431652,-64.1450205926935695 3.56987987952233476,-64.1449340157781194 3.56983226221883632,-64.1448969415567518 3.56983226221883632,-64.1448969415567518 3.56981187139708744,-64.1442254151352813 3.56944253186527805,-64.1439226156728637 3.56944253186527805,-64.1438738370612214 3.56936936394781235,-64.1432828936233506 3.56936936394781235,-64.1432828936233506 3.56924666279269376,-64.1434622515854613 3.56924666279269376,-64.1434622515854613 3.56906730483057677,-64.1436416095475863 3.56888794686845978,-64.1438209675096971 3.56870858890634279,-64.1440003254718079 3.56852923094422581,-64.1441796834339328 3.56834987298210882,-64.1443590413960436 3.56817051501999138,-64.1445383993581686 3.5679911570578744,-64.1447177573202794 3.56781179909575741,-64.1448971152824043 3.56781179909575741,-64.1448971152824043 3.56772508189847093,-64.1450179233212907 3.5677249309444643,-64.1447478551621373 3.56772526840438031,-64.1447475193541408 3.5674539140783379,-64.144477451248477 3.56745425143296746,-64.1444771155453566 3.56718289707956604,-64.1444767798679436 3.56691154272453037,-64.1444764442162381 3.5666401883678609,-64.1442063763223587 3.56664052556578071,-64.1442060407755008 3.56636917118175845,-64.1442057052543362 3.56609781679610194,-64.1439356374930867 3.56609815386305984,-64.143665569706144 3.56609849085049202,-64.143395501893508 3.56609882775839937,-64.1431254340552215 3.56609916458678056,-64.1428553661912417 3.56609950133563647,-64.1426822505643059 3.56609971714330687,-64.1426822656591042 3.56609971712448992,-64.1426822665188041 3.56634902966981304,-64.1426822415900517 3.56609971715449436,-64.1425852983016114 3.56609983800496666,-64.1426559905859506 3.5660997498792586,-64.1410292575303345 3.56772653716020249,-64.1425654617748791 3.56619028170800689,-64.1425654617748791 3.56619757743670363,-64.1425654617748791 3.56637693539882061,-64.1425654617748791 3.56655629336093805,-64.1425654617748791 3.56673565132305503,-64.1425654617748791 3.56691500928517202,-64.1425654617748791 3.56709436724728901,-64.1423861038127541 3.567273725209406,-64.1423861038127541 3.56745308317152343,-64.1423861038127541 3.56763244113364042,-64.1423861038127541 3.56781179909575741,-64.1423861038127541 3.5679911570578744,-64.1422067458506433 3.56817051501999138,-64.1420273878885325 3.56834987298210882,-64.1420273878885325 3.56852923094422581,-64.1418480299264075 3.56852923094422581,-64.1418480299264075 3.56870858890634279,-64.1416686719642968 3.56870858890634279,-64.1414893140021718 3.56888794686845978,-64.141309956040061 3.56906730483057677,-64.14113061601374 3.56906732276637317,-64.141029229980191 3.5687631697348996,-64.141029229980191 3.56882851560236425,-64.141029229980191 3.56910401710044223,-64.1413047314782716 3.56937951859852021,-64.1413047314782716 3.56973836072058504,-64.1412973950113638 3.56974202895403891,-64.1409247300051391 3.5699283614571522,-64.1405520649989143 3.57011469396026548,-64.1403657324957948 3.57030102646337877,-64.1401793999926895 3.57030102646337877,-64.1401793999926895 3.57119631845628938,-64.1409996380211709 3.57119631845628938,-64.1409996380211709 3.57139118363306851,-64.1417790987282928 3.57139118363306851,-64.1419739639050732 3.57158604880984765,-64.1421688290818537 3.57158604880984765,-64.1421688290818537 3.57178091398662678,-64.1423636942586342 3.57178091398662678,-64.1423636942586342 3.57179856698692122,-64.1423636942586342 3.57197577916340592,-64.1425585594354004 3.57197577916340592,-64.1425585594354004 3.57217064434018461,-64.1427534246121809 3.57217064434018461,-64.1427534246121809 3.57236550951696374,-64.1431331824776265 3.57236550951696374,-64.1439226156728637 3.57236550951696374,-64.1439226156728637 3.57253701650073907,-64.1444650475642959 3.57253701650073907,-64.1444650475642959 3.57256037469374288,-64.1450918067335323 3.57256037469374288,-64.1450918067335323 3.57272334900385236,-64.1451981793532298 3.57272334900385236,-64.1451981793532298 3.58362832807414966,-64.1412566261937798 3.58362832807414966,-64.141256427765839 3.58346821562181761,-64.1412566261937656 3.58362832807414966,-64.1409074366366809 3.58362832807414966,-64.1390851976311609 3.58319957415217605,-64.1388254332019585 3.58319989845089504,-64.1385553599296827 3.58320023553958356,-64.1385550244855267 3.58292888071676607,-64.1385552049578962 3.58307487237782496,-64.1377232395534094 3.58287911955448868,-64.13697398108485 3.58362832807414966,-64.1006565550055711 3.58362832807414966,-64.1006565550055711 3.58026179905184838,-64.1007412861745252 3.58026169691702023,-64.1007409621365554 3.57999033855325699,-64.1007403141347254 3.57944762182081755,-64.1010103898147037 3.57944729626604907,-64.1010100657714332 3.57917593792211441,-64.1010097417528755 3.57890457957654284,-64.1012798172492353 3.57890425399143997,-64.10127949317598 3.57863289566898057,-64.1012792875608426 3.57846071350074002,-64.1012791691274515 3.5783615373448816,-64.1018193197282358 3.57836088603425972,-64.1020893949915234 3.57836056025924254,-64.1020890707294484 3.57808920200778591,-64.1023591458885704 3.57808887617773674,-64.1026292210229229 3.57808855026788963,-64.1026288966267686 3.57781719206434889,-64.1026279235868799 3.5770031174439052,-64.102627599289761 3.57673175923381681,-64.1023575245524171 3.57673208501977324,-64.1018173750035061 3.57673273635238065,-64.1018170509693164 3.57646137806633924,-64.101816726959882 3.57619001977866269,-64.1015466523070785 3.57619034527579505,-64.1015437380496849 3.57374812039024459,-64.1020838859022319 3.57374746976196089,-64.1023539597913867 3.57374714432826757,-64.1023536358707418 3.57347578607377026,-64.1023523404356723 3.57239035303944208,-64.1042428541905025 3.5723880736398379,-64.1045129274849046 3.57238774769264156,-64.1045126030537062 3.57211638962836897,-64.1045113055769349 3.57103095735492948,-64.1045109812697547 3.57075959928248521,-64.1042409084510041 3.57075992508075846,-64.1034306898459647 3.57076090199777285,-64.103160616928065 3.57076122747751068,-64.1031609408389471 3.57103258567399706,-64.1028908678169955 3.57103291109888854,-64.1026207947703028 3.57103323644414283,-64.1026211185474182 3.57130459468857042,-64.1026220900273671 3.57211866941204859,-64.1006565550055711 3.57212103502532541,-64.1006565550055711 3.57049288263196241,-64.1007296363799952 3.5704927947905909,-64.1009997094413819 3.57049247005230397,-64.1009993862140277 3.57022111165439737,-64.1012694591714478 3.57022078686122901,-64.101269179589579 3.56998610966720209,-64.1006565550055711 3.56998610966720209,-64.1006565550055711 3.55296681855199648,-64.1451981793532298 3.55296681855199648),(-64.1018091105973724 3.56980532454323507,-64.1017681342339642 3.56980532454323507,-64.1016244577027834 3.56994900107442081,-64.1018092815717466 3.56994877871074889,-64.1018091105973724 3.56980532454323507),(-64.1090338280357628 3.57143239065893736,-64.1096431754630629 3.57143239065893736,-64.1096430117577967 3.57129610918266405,-64.1100402282328332 3.57129562723856964,-64.1100842499364347 3.57125160553497079,-64.1100842499364347 3.57052846503910271,-64.110265035060408 3.57034767991513569,-64.110445820184367 3.57016689479116911,-64.110992018472615 3.57016689479116911,-64.1109907647325628 3.56912361134993583,-64.1109904386977973 3.56885225386356719,-64.1107203670306234 3.56885258139828299,-64.1104502953385094 3.568852908853406,-64.1104499694871066 3.56858155131547772,-64.1101798978492781 3.56858187866605814,-64.1101792463797011 3.56803916353539385,-64.1101789206823014 3.56776780596761256,-64.1099088492571241 3.56776813316376451,-64.1096387778070351 3.56776846028034988,-64.1096384522929412 3.56749710266104714,-64.1093683808971093 3.56749742967312944,-64.1093680554871241 3.56722607202725728,-64.1093677301020364 3.56695471437975442,-64.1090976588396444 3.56695504126241891,-64.1090973335586511 3.56668368358835153,-64.1090970083025411 3.56641232591265389,-64.1090338280357628 3.56641240235310386,-64.108826937173589 3.56641265266592233,-64.1088262868944128 3.5658699372597833,-64.1088259617921636 3.56559857955426596,-64.1085558908757633 3.56559890615324626,-64.1085555658775661 3.56532754842117949,-64.1085552409042521 3.56505619068748114,-64.1082851701212206 3.56505651715712002,-64.1082845204075369 3.5645138016350093,-64.1082841955880269 3.56424244387150679,-64.108014125017462 3.56424277018692326,-64.1080134756113011 3.56370005460521488,-64.1080131509455384 3.56342869681191265,-64.1077430805874116 3.5634290229731449,-64.1061226579164156 3.5634309782716751,-64.1058525873842768 3.56343130387661944,-64.1058529114172728 3.56370266186894291,-64.104772628724092 3.56370396359338404,-64.1045025579887664 3.56370428882580503,-64.1045028816510722 3.56397564694068869,-64.1035759854736256 3.56397676255716167,-64.1035759854736256 3.56777578262268813,-64.1047774874502352 3.56777433477297024,-64.1047778115637641 3.56804569283855288,-64.1047781357021194 3.56831705090250439,-64.105048207758216 3.56831672516845755,-64.1050485320005805 3.56858808320594356,-64.1053186041110479 3.56858775736747935,-64.1063988923046679 3.56858645321776446,-64.10640149030624 3.57075731646482586,-64.1064018151682404 3.5710286743633568,-64.1066718878676625 3.57102834790314416,-64.1066722128337716 3.57129970577517142,-64.1066725378247355 3.5715710636455662,-64.1069426106578533 3.57157073705596373,-64.1069448868455112 3.57347024192882756");
    wktA += ", -64.1061346666064225 3.57347122198093059, -64.105864593143707 3.57347154850557658, -64.1058649180955342 3.5737429064374826, -64.1058658930999883 3.57455698022338808, -64.1055958192951181 3.57455730676774319, -64.1055961442669258 3.57482866471795724, -64.1055980946191681 3.57645681238488944, -64.1055984197647746 3.57672817032365176, -64.1058684942045716 3.57672784358048634, -64.1058688194544004 3.57699920149276185, -64.1058691447290556 3.57727055940339911, -64.1061392193028041 3.57727023253075282, -64.1074895917988101 3.57726859697083288, -64.1074886146202658 3.57645452368171268, -64.1077586888067117 3.57645419640507223, -64.1077583630508911 3.5761828386636525, -64.1077557578998523 3.57401197667339598, -64.1088187540686647 3.57401068831169777, -64.1088187540686647 3.57161317578290438, -64.1089995391926379 3.57143239065893736, -64.1090338280357628 3.57143239065893736), (-64.1158563111628581 3.57264534341541351, -64.1158559833764627 3.57237398640068982, -64.1158556556151069 3.57210262938432965, -64.115855327878819 3.57183127236633435, -64.1158550001675991 3.57155991534670481, -64.1158546724814187 3.57128855832544101, -64.1158543448203062 3.57101720130254119, -64.1158540171842475 3.57074584427800845, -64.1158536895732425 3.57047448725183925, -64.1161237612407433 3.57047415805442236, -64.1161234335755523 3.57020280105170151, -64.1163935051387455 3.57020247179974559, -64.1163931774193685 3.56993111482047842, -64.1163928497250879 3.56965975783957656, -64.1163925220558468 3.5693884008570409, -64.1163921944117021 3.56911704387287054, -64.1163918667926254 3.56884568688706771, -64.1163915391986023 3.56857432989963064, -64.1163912116296473 3.56830297291056064, -64.1163908840857744 3.56803161591985729, -64.1161208131563427 3.56803194497111376, -64.1158507422018431 3.5680322739427992, -64.1155806712222898 3.5680326028349123, -64.1155803439410477 3.56776124576733089, -64.1155800166848735 3.56748988869811701, -64.1153099458386322 3.56749021746052053, -64.1150398749673514 3.56749054614336414, -64.1147698040710452 3.56749087474664828, -64.1144997331496853 3.5674912032703725, -64.1142296622033143 3.56749153171453681, -64.1139595912319038 3.56749186007914165, -64.1139599180129807 3.5677632172986784, -64.1139602448190686 3.56803457451658357, -64.1139605716501677 3.56830593173285493, -64.1139608985062637 3.56857728894749293, -64.1139612253874134 3.56884864616049757, -64.1139615522935458 3.56912000337186841, -64.1136914808219274 3.56912033180710964, -64.1134214093252979 3.56912066016275498, -64.1131513378036715 3.56912098843880354, -64.1131516644971668 3.56939234572362185, -64.112881592871318 3.56939267394508919, -64.1128819195105706 3.56966403125328746, -64.1126118477805136 3.56966435942016069, -64.1123417760254597 3.56966468750742516, -64.1120717042454231 3.56966501551508131, -64.1118016324404181 3.56966534344312825, -64.1118019587877228 3.56993670084969139, -64.1118022851587739 3.57020805722533474, -64.1112108414566251 3.57020877556755378, -64.1113497458042048 3.57034767991513569, -64.1118024531013475 3.57034767991513569, -64.1118026115572661 3.57047941565791449, -64.1118029379794763 3.57075077305957667, -64.1115328658325438 3.57075110100795223, -64.1112627936606287 3.57075142887669505, -64.1109927214637594 3.57075175666580513, -64.1107226492419642 3.57075208437528202, -64.1104525769952005 3.57075241200512572, -64.1104529030460952 3.57102376953001288, -64.1104532291219158 3.57129512705326668, -64.1104535552226764 3.57156648457488579, -64.1104538813483771 3.57183784209486976, -64.1104542074990178 3.57210919961322038, -64.1107242801421222 3.57210887185857651, -64.1107246063969853 3.5723802293503315, -64.1107249326767743 3.57265158684045137, -64.1107252589815175 3.57292294432893653, -64.1107255853112008 3.57319430181578612, -64.1107259116658525 3.57346565930100191, -64.1107262380454159 3.57373701678458078, -64.1107265644499194 3.57400837426652451, -64.1107268908793912 3.57427973174683267, -64.110727217333789 3.57455108922550524, -64.110727543813141 3.57482244670254223, -64.1107275263300522 3.57480791540661258, -64.1108042241763627 3.57480790773606083, -64.1109835506832866 3.57462858122913829, -64.1111628771902105 3.57462858122913829, -64.1113422036971343 3.57462858122913829, -64.1113422036971343 3.57444925472221531, -64.1115215302040582 3.57444925472221531, -64.1117008567109821 3.57444925472221531, -64.1117008567109821 3.57426992821529277, -64.1118801832179059 3.57426992821529277, -64.1120595097248298 3.57426992821529277, -64.1122388362317537 3.57409060170837023, -64.1124181627386633 3.57409060170837023, -64.1124181627386633 3.57391127520144725, -64.1125974892455872 3.57391127520144725, -64.1127768157525111 3.57373194869452471, -64.112956142259435 3.57355262218760217, -64.1131354687663588 3.57337329568067918, -64.1133147952732827 3.57337329568067918, -64.1133147952732827 3.57319396917375665, -64.1134941217802066 3.57319396917375665, -64.1136734482871304 3.57319396917375665, -64.1136734482871304 3.57301464266683411, -64.1138527747940543 3.57301464266683411, -64.1138527747940543 3.57283531615991112, -64.1140321013009782 3.57283531615991112, -64.1142114278079021 3.57265598965298858, -64.1143907543148259 3.57265598965298858, -64.1145700808217356 3.57247666314606604, -64.1147494073286595 3.57247666314606604, -64.1149287338355833 3.57247666314606604, -64.1152873868494311 3.57247666314606604, -64.1158253663702027 3.57247666314606604, -64.1158561074050937 3.57247666314606604, -64.1158563111628581 3.57264534341541351), (-64.1301576819611654 3.56150218999404267, -64.1300104044266703 3.5613549124595405, -64.1300104044266703 3.56117555449742307, -64.1300104044266703 3.56099619653530608, -64.1300104044266703 3.56081683857318909, -64.1300104044266703 3.5606374806110721, -64.1300104044266703 3.56045812264895511, -64.1300104044266703 3.56009940672472069, -64.1300104044266703 3.55992004876260371, -64.1300104044266703 3.55987423430453598, -64.1298844651548734 3.55987438928665645, -64.1296143978593847 3.55987472155473705, -64.1293443305385296 3.55987505374343005, -64.1290742631923081 3.559875385852735, -64.1290745936978084 3.56014674170645318, -64.1288045262472366 3.56014707376174266, -64.1288048566990625 3.5604184296392023, -64.128534789144112 3.560418761640463, -64.1285351195422635 3.56069011754165965, -64.1282650518829485 3.56069044948888047, -64.1279949841983097 3.56069078135669503, -64.1277249164883187 3.56069111314510378, -64.1274548487530183 3.56069144485410538, -64.1271847809924083 3.56069177648370161, -64.1269147132064603 3.56069210803389069, -64.1266446453952312 3.56069243950467351, -64.1263745775586926 3.56069277089605007, -64.1263742477927536 3.56042141479214402, -64.1261041800099463 3.56042174607880302, -64.1258341122018578 3.56042207728606108, -64.125833782619253 3.56015072112990927, -64.1255637148648958 3.56015105223246753, -64.1252936470852859 3.56015138325563152, -64.1250235792803807 3.5601517141994008, -64.1247535114502369 3.56015204506377536, -64.1244834435948263 3.5601523758487561, -64.124213375714163 3.56015270655434168, -64.1242137048226652 3.56042406286219837, -64.1239436368377369 3.56042439351365214, -64.1239481370993616 3.56042438800386174, -64.1239481370993616 3.5605634706600271, -64.1242236385974422 3.56083897215810508, -64.1242236385974422 3.56111447365618305, -64.1242236385974422 3.56138997515426103, -64.1242236385974422 3.56166547665233901, -64.1242236385974422 3.56194097815041655, -64.1242236385974422 3.56221647964849453, -64.1244991400955229 3.56221647964849453, -64.1244991400955229 3.56249198114657251, -64.1244991400955229 3.56276748264465049, -64.1244991400955229 3.56304298414272846, -64.1244991400955229 3.56331848564080644, -64.1247746415936035 3.56331848564080644, -64.1247746415936035 3.56359398713888442, -64.12505014309167 3.56359398713888442, -64.1252978282889359 3.56359401190492742, -64.1252976018478762 3.56340765762894751, -64.125297272145346 3.56313630144014182, -64.1255673407945181 3.56313597013875238, -64.1255670110381857 3.56286461397360865, -64.1258370795830075 3.56286428261805677, -64.1261071481025482 3.56286395118305066, -64.1263772165968078 3.56286361966858856, -64.1266472850657721 3.56286328807467179, -64.1269173535094268 3.56286295640130035, -64.1271874219278004 3.56286262464847381, -64.1274574903208503 3.56286229281619171, -64.1277275586885764 3.56286196090445539, -64.127997627030993 3.56286162891326397, -64.1279972965882479 3.56259027297440189, -64.1279969661708407 3.56231891703390868, -64.1282670343297667 3.56231858501397047, -64.128537102463369 3.56231825291458959, -64.1288071705716334 3.56231792073576603, -64.1288068399423565 3.56204656486972082, -64.1290769079462137 3.56204623263682674, -64.1293469759247046 3.56204590032449664, -64.1296170438778432 3.56204556793273008, -64.1296167130367536 3.56177421214118928, -64.1298867808854709 3.56177387969537618, -64.1298864499906927 3.56150252392759414, -64.1301565177349886 3.56150219142774782, -64.1301576819611654 3.56150218999404267), (-64.1404421776975084 3.58021297307141984, -64.1404418419783724 3.57994161840982539, -64.1404415062848443 3.57967026374659003, -64.1404411706169384 3.57939890908171598, -64.1404408349746404 3.57912755441520192, -64.1404404993579504 3.57885619974704738, -64.1401704275109239 3.57885653690513772, -64.1399003556383178 3.57885687398341723, -64.1396302837401038 3.57885721098188547, -64.1393602118162818 3.57885754790054289, -64.1390901398668944 3.5788578847393886, -64.1388200678919134 3.57885822149842348, -64.1388204030319855 3.57912957632025375, -64.138820738197623 3.579400931140444, -64.1388209455597149 3.579568801591547, -64.1388252179955742 3.57957307402740366, -64.1391007194936407 3.57984857552548164, -64.1391007194936407 3.58012407702355961, -64.1393762209917213 3.58012407702355961, -64.1393762209917213 3.5802143038506129, -64.1393618885668815 3.58021432173729082, -64.1396319608879253 3.58021398469058516, -64.1399020331833896 3.58021364756403848, -64.1401721054532459 3.58021331035764989, -64.1404421776975084 3.58021297307141984), (-64.1447498705499157 3.56935339432630583, -64.1450199391842943 3.56935305671200753, -64.1450196031428561 3.56908170242150424, -64.1450198775148976 3.569303258522313, -64.1449700169170995 3.56935311912011066, -64.1447498705499157 3.56935339432630583)))";
    std::string wktB("multipolygon(((-64.1451981793532298 3.55296681855199648,-64.1451981793532298 3.57002712739561545,-64.1450918067335323 3.57002712739561545,-64.1450918067335323 3.56991904724431652,-64.1450205926935695 3.56987987952233476,-64.1449340157781194 3.56983226221883632,-64.1448969415567518 3.56983226221883632,-64.1448969415567518 3.56981187139708744,-64.1442254151352813 3.56944253186527805,-64.1439226156728637 3.56944253186527805,-64.1438738370612214 3.56936936394781235,-64.1432828936233506 3.56936936394781235,-64.1432828936233506 3.56924666279269376,-64.1434622515854613 3.56924666279269376,-64.1434622515854613 3.56906730483057677,-64.1436416095475863 3.56888794686845978,-64.1438209675096971 3.56870858890634279,-64.1440003254718079 3.56852923094422581,-64.1441796834339328 3.56834987298210882,-64.1443590413960436 3.56817051501999138,-64.1445383993581686 3.5679911570578744,-64.1447177573202794 3.56781179909575741,-64.1448971152824043 3.56781179909575741,-64.1448971152824043 3.56772508189847093,-64.1447478551621373 3.56772526840438031,-64.1447475193541408 3.5674539140783379,-64.144477451248477 3.56745425143296746,-64.1444771155453566 3.56718289707956604,-64.1444767798679436 3.56691154272453037,-64.1444764442162381 3.5666401883678609,-64.1442063763223587 3.56664052556578071,-64.1442060407755008 3.56636917118175845,-64.1442057052543362 3.56609781679610194,-64.1439356374930867 3.56609815386305984,-64.143665569706144 3.56609849085049202,-64.143395501893508 3.56609882775839937,-64.1431254340552215 3.56609916458678056,-64.1428553661912417 3.56609950133563647,-64.1426822656591042 3.56609971712448992,-64.1426822665188041 3.56634902966981304,-64.1426822415900517 3.56609971715449436,-64.1426559905859506 3.5660997498792586,-64.1425654617748791 3.56619028170800689,-64.1425654617748791 3.56709436724728901,-64.1423861038127541 3.567273725209406,-64.1423861038127541 3.5679911570578744,-64.1422067458506433 3.56817051501999138,-64.1420273878885325 3.56834987298210882,-64.1420273878885325 3.56852923094422581,-64.1418480299264075 3.56852923094422581,-64.1418480299264075 3.56870858890634279,-64.1416686719642968 3.56870858890634279,-64.1414893140021718 3.56888794686845978,-64.141309956040061 3.56906730483057677,-64.14113061601374 3.56906732276637317,-64.141029229980191 3.5687631697348996,-64.141029229980191 3.56910401710044223,-64.1413047314782716 3.56937951859852021,-64.1413047314782716 3.56973836072058504,-64.1412973950113638 3.56974202895403891,-64.1409247300051391 3.5699283614571522,-64.1405520649989143 3.57011469396026548,-64.1403657324957948 3.57030102646337877,-64.1401793999926895 3.57030102646337877,-64.1401793999926895 3.57119631845628938,-64.1409996380211709 3.57119631845628938,-64.1409996380211709 3.57139118363306851,-64.1417790987282928 3.57139118363306851,-64.1419739639050732 3.57158604880984765,-64.1421688290818537 3.57158604880984765,-64.1421688290818537 3.57178091398662678,-64.1423636942586342 3.57178091398662678,-64.1423636942586342 3.57197577916340592,-64.1425585594354004 3.57197577916340592,-64.1425585594354004 3.57217064434018461,-64.1427534246121809 3.57217064434018461,-64.1427534246121809 3.57236550951696374,-64.1439226156728637 3.57236550951696374,-64.1439226156728637 3.57253701650073907,-64.1444650475642959 3.57253701650073907,-64.1444650475642959 3.57256037469374288,-64.1450918067335323 3.57256037469374288,-64.1450918067335323 3.57272334900385236,-64.1451981793532298 3.57272334900385236,-64.1451981793532298 3.58362832807414966,-64.1409074366366809 3.58362832807414966,-64.1390851976311609 3.58319957415217605,-64.1388254332019585 3.58319989845089504,-64.1385553599296827 3.58320023553958356,-64.1385552049578962 3.58307487237782496,-64.1377232395534094 3.58287911955448868,-64.13697398108485 3.58362832807414966,-64.1006565550055711 3.58362832807414966,-64.1006565550055711 3.58026179905184838,-64.1007412861745252 3.58026169691702023,-64.1007409621365554 3.57999033855325699,-64.1007403141347254 3.57944762182081755,-64.1010103898147037 3.57944729626604907,-64.1010100657714332 3.57917593792211441,-64.1010097417528755 3.57890457957654284,-64.1012798172492353 3.57890425399143997,-64.10127949317598 3.57863289566898057,-64.1012791691274515 3.5783615373448816,-64.1018193197282358 3.57836088603425972,-64.1020893949915234 3.5783605602592421,-64.1020890707294484 3.57808920200778591,-64.1023591458885704 3.57808887617773674,-64.1026292210229229 3.57808855026788963,-64.1026288966267686 3.57781719206434889,-64.1026279235868657 3.5770031174439052,-64.102627599289761 3.57673175923381681,-64.1023575245524171 3.57673208501977324,-64.1018173750035061 3.57673273635238065,-64.1018170509693164 3.57646137806633924,-64.101816726959882 3.57619001977866269,-64.1015466523070785 3.57619034527579505,-64.1015437380496849 3.57374812039024459,-64.1020838859022319 3.57374746976196089,-64.1023539597913867 3.57374714432826757,-64.1023536358707418 3.57347578607377026,-64.1023523404356723 3.57239035303944208,-64.1042428541905025 3.5723880736398379,-64.1045129274849046 3.57238774769264156,-64.1045126030537062 3.57211638962836897,-64.1045113055769349 3.57103095735492948,-64.1045109812697547 3.57075959928248521,-64.1042409084510041 3.57075992508075846,-64.1034306898459647 3.57076090199777285,-64.103160616928065 3.57076122747751068,-64.1031609408389471 3.57103258567399706,-64.1028908678169955 3.57103291109888854,-64.1026207947703028 3.57103323644414283,-64.1026211185474182 3.57130459468857042,-64.1026220900273671 3.57211866941204859,-64.1006565550055711 3.57212103502532541,-64.1006565550055711 3.57049288263196241,-64.1007296363799952 3.5704927947905909,-64.1009997094413819 3.57049247005230397,-64.1009993862140277 3.57022111165439737,-64.1012694591714478 3.57022078686122901,-64.101269179589579 3.56998610966720209,-64.1006565550055711 3.56998610966720209,-64.1006565550055711 3.55296681855199648,-64.1451981793532298 3.55296681855199648),(-64.1018091105973724 3.56980532454323507,-64.1017681342339642 3.56980532454323507,-64.1016244577027692 3.56994900107442081,-64.1018092815717466 3.56994877871074889,-64.1018091105973724 3.56980532454323507),(-64.110992018472615 3.57016689479116911,-64.1109907647325628 3.56912361134993583,-64.1109904386977973 3.56885225386356719,-64.1107203670306234 3.56885258139828299,-64.1104502953385094 3.568852908853406,-64.1104499694871066 3.56858155131547772,-64.1101798978492781 3.56858187866605814,-64.1101792463797011 3.56803916353539385,-64.1101789206823014 3.56776780596761256,-64.1099088492571241 3.56776813316376451,-64.1096387778070351 3.56776846028034988,-64.1096384522929412 3.56749710266104714,-64.1093683808971093 3.56749742967312944,-64.1093680554871241 3.56722607202725728,-64.1093677301020364 3.56695471437975442,-64.1090976588396444 3.56695504126241847,-64.1090973335586511 3.56668368358835153,-64.1090970083025411 3.56641232591265389,-64.108826937173589 3.56641265266592233,-64.1088262868944128 3.5658699372597833,-64.1088259617921636 3.56559857955426596,-64.1085558908757633 3.56559890615324626,-64.1085555658775661 3.56532754842117949,-64.1085552409042521 3.56505619068748114,-64.1082851701212206 3.56505651715712002,-64.1082845204075369 3.5645138016350093,-64.1082841955880269 3.56424244387150635,-64.108014125017462 3.56424277018692326,-64.1080134756113011 3.56370005460521488,-64.1080131509455384 3.56342869681191265,-64.1077430805874116 3.5634290229731449,-64.1061226579164156 3.5634309782716751,-64.1058525873842768 3.56343130387661944,-64.1058529114172728 3.56370266186894291,-64.104772628724092 3.56370396359338404,-64.1045025579887664 3.56370428882580503,-64.1045028816510722 3.56397564694068869,-64.1035759854736256 3.56397676255716167,-64.1035759854736256 3.56777578262268813,-64.1047774874502352 3.56777433477296979,-64.1047778115637641 3.56804569283855288,-64.1047781357021194 3.56831705090250439,-64.105048207758216 3.56831672516845755,-64.1050485320005805 3.56858808320594356,-64.1053186041110479 3.56858775736747935,-64.1063988923046679 3.56858645321776446,-64.10640149030624 3.57075731646482586,-64.1064018151682404 3.5710286743633568,-64.1066718878676625 3.57102834790314416,-64.1066722128337716 3.57129970577517142,-64.1066725378247355 3.5715710636455662,-64.1069426106578533 3.57157073705596373,-64.1069448868455112 3.57347024192882756,-64.1061346666064225 3.57347122198093059,-64.105864593143707 3.57347154850557658,-64.1058649180955342 3.5737429064374826,-64.1058658930999883 3.57455698022338808,-64.1055958192951181 3.57455730676774319,-64.1055961442669258 3.57482866471795724,-64.1055980946191681 3.57645681238488944,-64.1055984197647746 3.57672817032365176,-64.1058684942045716 3.57672784358048634,-64.1058688194544004 3.57699920149276185,-64.1058691447290556 3.57727055940339911,-64.1061392193028041 3.57727023253075282,-64.1074895917988101 3.57726859697083288,-64.1074886146202658 3.57645452368171268,-64.1077586888067117 3.57645419640507223,-64.1077583630508911 3.5761828386636525,-64.1077557578998523 3.57401197667339598,-64.1088187540686647 3.57401068831169777,-64.1088187540686647 3.57161317578290438,-64.1089995391926379 3.57143239065893736,-64.1096431754630629 3.57143239065893736,-64.1096430117577967 3.57129610918266405,-64.1100402282328332 3.57129562723856964,-64.1100842499364347 3.57125160553497079,-64.1100842499364347 3.57052846503910271,-64.110265035060408 3.57034767991513569,-64.110445820184367 3.57016689479116911,-64.110992018472615 3.57016689479116911),(-64.1163935051387455 3.57020247179974559");
    wktB += ", -64.1163931774193685 3.56993111482047842, -64.1163928497250879 3.56965975783957656, -64.1163925220558468 3.5693884008570409, -64.1163921944117021 3.56911704387287054, -64.1163918667926254 3.56884568688706771, -64.1163915391986023 3.56857432989963064, -64.1163912116296473 3.56830297291056064, -64.1163908840857744 3.56803161591985729, -64.1161208131563427 3.56803194497111376, -64.1158507422018431 3.5680322739427992, -64.1155806712222898 3.5680326028349123, -64.1155803439410477 3.56776124576733089, -64.1155800166848735 3.56748988869811701, -64.1153099458386322 3.56749021746052053, -64.1150398749673514 3.56749054614336414, -64.1147698040710452 3.56749087474664828, -64.1144997331496853 3.5674912032703725, -64.1142296622033143 3.56749153171453681, -64.1139595912319038 3.56749186007914165, -64.1139599180129807 3.5677632172986784, -64.1139602448190686 3.56803457451658357, -64.1139605716501677 3.56830593173285493, -64.1139608985062637 3.56857728894749293, -64.1139612253874134 3.56884864616049757, -64.1139615522935458 3.56912000337186841, -64.1136914808219274 3.56912033180710964, -64.1134214093252979 3.56912066016275498, -64.1131513378036715 3.56912098843880354, -64.1131516644971668 3.56939234572362185, -64.112881592871318 3.56939267394508919, -64.1128819195105706 3.56966403125328746, -64.1126118477805136 3.56966435942016069, -64.1123417760254597 3.56966468750742516, -64.1120717042454231 3.56966501551508131, -64.1118016324404181 3.56966534344312825, -64.1118019587877228 3.56993670084969139, -64.1118022851587739 3.57020805722533474, -64.1112108414566251 3.57020877556755378, -64.1113497458042048 3.57034767991513569, -64.1118024531013475 3.57034767991513569, -64.1118026115572661 3.57047941565791449, -64.1118029379794763 3.57075077305957667, -64.1115328658325438 3.57075110100795223, -64.1112627936606287 3.57075142887669505, -64.1109927214637594 3.57075175666580513, -64.1107226492419642 3.57075208437528202, -64.1104525769952005 3.57075241200512572, -64.1104529030460952 3.57102376953001288, -64.1104532291219158 3.57129512705326668, -64.1104535552226764 3.57156648457488579, -64.1104538813483771 3.57183784209486976, -64.1104542074990178 3.57210919961322038, -64.1107242801421222 3.57210887185857651, -64.1107246063969853 3.5723802293503315, -64.1107249326767743 3.57265158684045137, -64.1107252589815175 3.57292294432893653, -64.1107255853112008 3.57319430181578612, -64.1107259116658525 3.57346565930100191, -64.1107262380454159 3.57373701678458078, -64.1107265644499194 3.57400837426652451, -64.1107268908793912 3.57427973174683267, -64.110727217333789 3.57455108922550524, -64.1107275263300522 3.57480791540661258, -64.1108042241763627 3.57480790773606083, -64.1109835506832866 3.57462858122913829, -64.1113422036971343 3.57462858122913829, -64.1113422036971343 3.57444925472221531, -64.1117008567109821 3.57444925472221531, -64.1117008567109821 3.57426992821529277, -64.1120595097248298 3.57426992821529277, -64.1122388362317537 3.57409060170837023, -64.1124181627386633 3.57409060170837023, -64.1124181627386633 3.57391127520144725, -64.1125974892455872 3.57391127520144725, -64.1127768157525111 3.57373194869452471, -64.112956142259435 3.57355262218760217, -64.1131354687663588 3.57337329568067918, -64.1133147952732827 3.57337329568067918, -64.1133147952732827 3.57319396917375665, -64.1136734482871304 3.57319396917375665, -64.1136734482871304 3.57301464266683411, -64.1138527747940543 3.57301464266683411, -64.1138527747940543 3.57283531615991112, -64.1140321013009782 3.57283531615991112, -64.1142114278079021 3.57265598965298858, -64.1143907543148259 3.57265598965298858, -64.1145700808217356 3.57247666314606604, -64.1158561074050937 3.57247666314606604, -64.1158559833764627 3.57237398640068982, -64.1158556556151069 3.57210262938432965, -64.115855327878819 3.57183127236633435, -64.1158550001675991 3.57155991534670481, -64.1158546724814187 3.57128855832544101, -64.1158543448203062 3.57101720130254119, -64.1158540171842475 3.57074584427800845, -64.1158536895732425 3.57047448725183925, -64.1161237612407433 3.57047415805442236, -64.1161234335755523 3.57020280105170151, -64.1163935051387455 3.57020247179974559), (-64.1301576819611654 3.56150218999404267, -64.1300104044266703 3.5613549124595405, -64.1300104044266703 3.55987423430453598, -64.1298844651548734 3.55987438928665645, -64.1296143978593847 3.55987472155473705, -64.1293443305385296 3.55987505374343005, -64.1290742631923081 3.559875385852735, -64.1290745936978084 3.56014674170645318, -64.1288045262472366 3.56014707376174266, -64.1288048566990625 3.5604184296392023, -64.128534789144112 3.560418761640463, -64.1285351195422635 3.56069011754165965, -64.1282650518829485 3.56069044948888047, -64.1279949841983097 3.56069078135669503, -64.1277249164883187 3.56069111314510378, -64.1274548487530183 3.56069144485410538, -64.1271847809924083 3.56069177648370161, -64.1269147132064603 3.56069210803389069, -64.1266446453952312 3.56069243950467351, -64.1263745775586926 3.56069277089605007, -64.1263742477927536 3.56042141479214402, -64.1261041800099463 3.56042174607880302, -64.1258341122018578 3.56042207728606108, -64.125833782619253 3.56015072112990927, -64.1255637148648958 3.56015105223246753, -64.1252936470852859 3.56015138325563152, -64.1250235792803807 3.5601517141994008, -64.1247535114502369 3.56015204506377536, -64.1244834435948263 3.5601523758487561, -64.124213375714163 3.56015270655434168, -64.1242137048226652 3.56042406286219837, -64.1239481370993616 3.56042438800386174, -64.1239481370993616 3.5605634706600271, -64.1242236385974422 3.56083897215810508, -64.1242236385974422 3.56221647964849453, -64.1244991400955229 3.56221647964849453, -64.1244991400955229 3.56331848564080644, -64.1247746415936035 3.56331848564080644, -64.1247746415936035 3.56359398713888442, -64.12505014309167 3.56359398713888442, -64.1252978282889359 3.56359401190492742, -64.1252976018478762 3.56340765762894751, -64.125297272145346 3.56313630144014182, -64.1255673407945181 3.56313597013875238, -64.1255670110381857 3.56286461397360865, -64.1258370795830075 3.56286428261805677, -64.1261071481025482 3.56286395118305066, -64.1263772165968078 3.56286361966858856, -64.1266472850657721 3.56286328807467179, -64.1269173535094268 3.56286295640130035, -64.1271874219278004 3.56286262464847381, -64.1274574903208503 3.56286229281619171, -64.1277275586885764 3.56286196090445539, -64.127997627030993 3.56286162891326397, -64.1279972965882479 3.56259027297440189, -64.1279969661708407 3.56231891703390868, -64.1282670343297667 3.56231858501397047, -64.128537102463369 3.56231825291458959, -64.1288071705716334 3.56231792073576603, -64.1288068399423565 3.56204656486972082, -64.1290769079462137 3.56204623263682674, -64.1293469759247046 3.56204590032449664, -64.1296170438778432 3.56204556793273008, -64.1296167130367536 3.56177421214118928, -64.1298867808854709 3.56177387969537618, -64.1298864499906927 3.56150252392759414, -64.1301565177349886 3.56150219142774782, -64.1301576819611654 3.56150218999404267), (-64.1404421776975084 3.58021297307141984, -64.1404418419783724 3.57994161840982539, -64.1404415062848443 3.57967026374659003, -64.1404411706169384 3.57939890908171598, -64.1404408349746404 3.57912755441520192, -64.1404404993579504 3.57885619974704738, -64.1401704275109239 3.57885653690513772, -64.1399003556383178 3.57885687398341723, -64.1396302837401038 3.57885721098188547, -64.1393602118162818 3.57885754790054289, -64.1390901398668944 3.5788578847393886, -64.1388200678919134 3.57885822149842348, -64.1388204030319855 3.57912957632025375, -64.138820738197623 3.579400931140444, -64.1388209455597149 3.579568801591547, -64.1388252179955742 3.57957307402740366, -64.1391007194936407 3.57984857552548164, -64.1391007194936407 3.58012407702355961, -64.1393762209917213 3.58012407702355961, -64.1393762209917213 3.5802143038506129, -64.1396319608879253 3.58021398469058516, -64.1399020331833896 3.58021364756403848, -64.1401721054532459 3.58021331035764989, -64.1404421776975084 3.58021297307141984), (-64.1450199391842943 3.56935305671200753, -64.1450198775148976 3.569303258522313, -64.1449700169170995 3.56935311912011066, -64.1450199391842943 3.56935305671200753)))";

    std::unique_ptr<te::gm::Geometry> geometryA(te::gm::WKTReader::read(wktA.c_str()));
    std::unique_ptr<te::gm::Geometry> geometryB(te::gm::WKTReader::read(wktB.c_str()));

    te::gm::GeometryVectorConst input;
    input.push_back(geometryA.get());
    input.push_back(geometryB.get());

    double tolerance = te::gm::GetAppropriatePrecision(4674);

    te::gm::GeometryVector output = te::vp::GeometrySnapper::snapGeometries(input, tolerance);
    std::string wktResultA = output.at(0)->toString();
    std::string wktResultB = output.at(1)->toString();

    te::common::FreeContents(output);
  }

  TEST_F(ItsPrepareToOverlay, PrepareGeometries)
  {
    std::vector<std::string> vecWKTs;
    vecWKTs.push_back("multipolygon(((-62.7751625341312618 2.99058438231863155,-62.77516258939918 2.99031297242929561,-62.7754325548757421 2.99031302795702425,-62.7757025203573491 2.99031308341804269,-62.7759724858439654 2.99031313881234828,-62.7759725409077802 2.99004172890653708,-62.7762425063330696 2.9900417842291005,-62.7762425074046604 2.99004178422931988,-62.7762424524249099 2.9903131934352305,-62.7762423974224575 2.99058460334471965,-62.7759724318644672 2.99058454801208073,-62.7751625352206588 2.99058438161385842,-62.7751625341312618 2.99058438231863155)))");
    vecWKTs.push_back("multipolygon(((-62.7740816715439109 2.99546953693347584,-62.7740817271683937 2.99519812708864608,-62.7740817827878175 2.99492671724246939,-62.7738118162075764 2.99492666129483798,-62.7738118718883911 2.99465525145239075,-62.7738119275641395 2.99438384160859616,-62.7738119832348076 2.99411243176345288,-62.7738120389004237 2.99384102191696089,-62.7738120843762815 2.99361927420808005,-62.7741858991690407 2.99361927420808005,-62.7744269214919157 2.99386029653095287,-62.7746679438147908 2.99386029653095287,-62.7746679438147908 2.99058428040207236,-62.7748925685933798 2.99058432671913588,-62.774892569682784 2.99058432601434898,-62.7751625341312618 2.99058438231863155,-62.7751625352206588 2.99058438161385842,-62.7759724318644672 2.99058454801208073,-62.7762423974224575 2.99058460334471965,-62.7762424524249099 2.9903131934352305,-62.7762425074046604 2.99004178422931988,-62.7765124717633825 2.99004183948495994,-62.7765125266894728 2.98977042956774897,-62.7773224228113875 2.98977059492005415,-62.7775923881953304 2.98977064990408969,-62.7775923335346278 2.99004205984133753,-62.7775921148419513 2.99112769957687208,-62.7775920601563229 2.99139910950738841,-62.7773220943741919 2.99139905449335197,-62.7773218205392993 2.99275610410073645,-62.777591786653403 2.99275615913977289,-62.7778617527725018 2.9927562141120454,-62.7778616981233881 2.99302762403947531,-62.7778613154401413 2.99492749349375043,-62.7778612607511874 2.99519890341039652,-62.7775912940341527 2.99519884839317907,-62.7767813939129624 2.99519868294059943,-62.7765114272158939 2.99519862765609846,-62.7765114822371473 2.99492721776448212,-62.7765115372534055 2.99465580787151753,-62.7762415706942676 2.9946557525302433,-62.7762416257719522 2.99438434264095488,-62.7748917933838442 2.9943840649576714,-62.7748917379739027 2.99465547482174532,-62.7746217714449415 2.99465541907961974,-62.7746216604770026 2.99519823879360247,-62.7746216049854482 2.99546964864857257,-62.7743516382621536 2.99546959282443792,-62.7740816715439109 2.99546953693347584)))");

    vecWKTs.push_back("multipolygon(((-62.7762425074046604 2.99004178422931988,-62.7762424524249099 2.9903131934352305,-62.7762423974224575 2.99058460334471965,-62.7759724318644672 2.99058454801208073,-62.7751625352206588 2.99058438161385842,-62.7751625341312618 2.99058438231863155,-62.77516258939918 2.99031297242929561,-62.7754325548757421 2.99031302795702425,-62.7757025203573491 2.99031308341804269,-62.7759724858439654 2.99031313881234828,-62.7759725409077802 2.99004172890653708,-62.7762425063330696 2.9900417842291005,-62.7762425074046604 2.99004178422931988)))");
    vecWKTs.push_back("multipolygon(((-62.7765125266894728 2.98977042956774897,-62.7773224228113875 2.98977059492005415,-62.7775923881953304 2.98977064990408969,-62.7775923335346278 2.99004205984133753,-62.7775921148419513 2.99112769957687208,-62.7775920601563229 2.99139910950738841,-62.7773220943741919 2.99139905449335197,-62.7773218205392993 2.992756104100736,-62.777591786653403 2.99275615913977289,-62.7778617527725018 2.99275621411204495,-62.7778616981233881 2.99302762403947531,-62.7778613154401413 2.99492749349375043,-62.7778612607511874 2.99519890341039652,-62.7775912940341527 2.99519884839317907,-62.7767813939129624 2.99519868294059943,-62.7765114272158939 2.99519862765609846,-62.7765114822371473 2.99492721776448212,-62.7765115372534055 2.99465580787151753,-62.7762415706942676 2.9946557525302433,-62.7762416257719522 2.99438434264095488,-62.7748917933838442 2.9943840649576714,-62.7748917379739027 2.99465547482174532,-62.7746217714449415 2.99465541907961974,-62.7746216604770026 2.99519823879360247,-62.7746216049854482 2.99546964864857257,-62.7743516382621536 2.99546959282443792,-62.7740816715439109 2.99546953693347584,-62.7740817271683937 2.99519812708864608,-62.7740817827878175 2.99492671724246939,-62.7738118162075764 2.99492666129483798,-62.7738118718883911 2.99465525145239075,-62.7738119275641395 2.99438384160859616,-62.7738119832348076 2.99411243176345288,-62.7738120389004237 2.99384102191696089,-62.7738120843762815 2.99361927420808005,-62.7741858991690336 2.99361927420808005,-62.7744269214919086 2.99386029653095287,-62.7746679438147837 2.99386029653095287,-62.7746679438147837 2.99058428040207236,-62.7748925685933798 2.99058432671913588,-62.774892569682784 2.99058432601434898,-62.7751625341312618 2.99058438231863155,-62.7751625352206588 2.99058438161385842,-62.7759724318644672 2.99058454801208073,-62.7762423974224575 2.99058460334471965,-62.7762424524249099 2.9903131934352305,-62.7762425074046604 2.99004178422931988,-62.7765124717633825 2.99004183948495994,-62.7765125266894728 2.98977042956774897)))");

    te::gm::GeometryVectorConst input;
    for (std::size_t i = 0; i < vecWKTs.size(); ++i)
    {
      input.push_back(te::gm::WKTReader::read(vecWKTs.at(i).c_str()));
    }

    double tolerance = te::gm::GetAppropriatePrecision(4674);
    te::gm::GeometryVector output = te::vp::GeometrySnapper::snapGeometries(input, tolerance);


    for (std::size_t i = 0; i < vecWKTs.size(); ++i)
    {
      std::string wktResultA = output.at(i)->toString();
      ASSERT_TRUE(output.at(i)->isValid());
      
    }
    te::common::FreeContents(input);
    te::common::FreeContents(output);
  }
  

  TEST_F(ItsPrepareToOverlay, PrepareGeometries2)
  {
    te::gtest::DataInfo redDataInfo = te::gtest::GetRedDataInfo();

    te::vp::FeatureSet featureSet;
    te::gtest::GetFeaturesFromShapefile(redDataInfo.m_uri, featureSet);

    ASSERT_FALSE(featureSet.empty());

    te::gm::GeometryVectorConst inputGeometryVector = te::vp::GetGeometriesFromFeatureSet(featureSet);

    bool wasChanged = false;
    te::gm::GeometryVector vecPrepared = te::gm::PrepareGeometriesToIntersection(inputGeometryVector, wasChanged, te::gm::GetAppropriatePrecision(inputGeometryVector.at(0)->getSRID()));

    te::gtest::SimpleSaveGeometriesToShapefile(te::gtest::GetOutputDir(), redDataInfo.m_datasetName + "_prepared", te::gm::ToConstVector(vecPrepared));

    te::common::FreeContents(featureSet);
    te::common::FreeContents(vecPrepared);
  }

  TEST_F(ItsPrepareToOverlay, PrepareGeometriesCase1)
  {
    std::vector<std::string> vecWKTs;

    std::string wkt1 = "multipolygon(((-48.781067432571902 -22.7785387732640956,-48.781067432571902 -22.7785387732640956,-48.7809840381884925 -22.7778382632009979,-48.7804183814698433 -22.7767015559232746,-48.7804183814698433 -22.7767015559232746,-48.7803881602657015 -22.7763953433808055,-48.7803881602657015 -22.7763953433808055,-48.7806285511635025 -22.7752786109551977,-48.7804672944828965 -22.7746268844778221,-48.7804672944828965 -22.7746268844778221,-48.7797637459356963 -22.773503690007594,-48.7796274317220053 -22.7728878913769961,-48.779269107394498 -22.7720127497969003,-48.7793196038735033 -22.7715317954117964,-48.7793196038735033 -22.7715317954117964,-48.7793196038735033 -22.7715317954117964,-48.778929966619998 -22.7709923729626951,-48.7788927591797972 -22.7704095181785959,-48.7783494507602882 -22.7696841108660948,-48.7783494506366395 -22.7696841096418829,-48.778255382290503 -22.7687527574069009,-48.778255382290503 -22.7687527574069009,-48.7777792405830226 -22.7680547487745955,-48.7777792405830226 -22.7680547487745955,-48.7773816145576973 -22.7659880326378108,-48.7773816137097924 -22.7659880282307014,-48.7773289353368042 -22.7652489096596042,-48.7773289353368042 -22.7652489096596042,-48.7776619540510978 -22.7639767368425012,-48.778058027846086 -22.76355476964169,-48.778058027846086 -22.76355476964169,-48.7783309250336004 -22.7634502192278987,-48.7783309250336004 -22.7634502192278987,-48.7783309250336004 -22.7634502192278987,-48.7781987428535899 -22.7620107146700974,-48.7773549329140934 -22.7617824222034955,-48.7770363710401966 -22.7612705281587964,-48.7768447926258943 -22.7611589380556936,-48.7766051644324961 -22.7612087483180972,-48.7766884210484903 -22.7615292944400949,-48.7766122068620973 -22.7616627654279995,-48.7757800778210964 -22.7609495462482982,-48.7753209277060904 -22.7608796489179994,-48.7743698990553014 -22.761035904900595,-48.7742349525664025 -22.7608190154761942,-48.7743035836249987 -22.7605268334669972,-48.7736740272736995 -22.7599948857105971,-48.7736603507051996 -22.7594596203150985,-48.7734274304048938 -22.7594358147034015,-48.7730231754636918 -22.7596126029187005,-48.7722789220754009 -22.7596099263971965,-48.7705443247627954 -22.7583945414403956,-48.7702373497338897 -22.7582944842497952,-48.7698052172694929 -22.7583082884664947,-48.7696294253789944 -22.7584213975273997,-48.7694422896857986 -22.7587295151460012,-48.7691541279863969 -22.7597951417873006,-48.7687976158564993 -22.7598951455739993,-48.768590956978997 -22.7600765731138992,-48.7685648323440972 -22.7602503905737947,-48.7688368707646021 -22.760627523121201,-48.7687831334688013 -22.7608355531387971,-48.7685111734645886 -22.7609410180287952,-48.7681106755936895 -22.7608406029599948,-48.7678962300659009 -22.7609452440566002,-48.7679984109214928 -22.761509507732697,-48.7679284608078945 -22.7617593247431955,-48.7668904013650888 -22.7618469650166944,-48.7665028836955941 -22.7623502121359991,-48.7654993710839904 -22.762782909076698,-48.7651753452396974 -22.7633836648216956,-48.7643593016723003 -22.7637036320120956,-48.7642180159857901 -22.7643409700837012,-48.7639945667080923 -22.7646298839705992,-48.7632726699327037 -22.7650609647138964,-48.7632126492335942 -22.7652291700749956,-48.7633100096108905 -22.7658680063616927,-48.7629905038085951 -22.7663500555680969,-48.7627482085838864 -22.7665456137098019,-48.7621744858157911 -22.7667960589740943,-48.7616774856110027 -22.7672822036468929,-48.760542065523893 -22.7675448817372938,-48.7600697308381967 -22.7678445531654958,-48.7597353068453998 -22.768257678336397,-48.7591418017391902 -22.7683372965472941,-48.7587525176075047 -22.7688089474720954,-48.7584646619174862 -22.7689471845357971,-48.758091029631899 -22.7689844749320969,-48.7578708595916908 -22.7694670603133957,-48.7577454982138931 -22.7695601948831978,-48.7575330268870886 -22.7696126629537972,-48.7572086847366961 -22.7694822982830019,-48.7569876017007857 -22.7695560921405971,-48.7567200627906914 -22.7702577105387931,-48.7559866386014988 -22.7706829707954981,-48.7559166586914898 -22.7713397360637941,-48.7557248782232904 -22.7715387355894947,-48.7547051046155957 -22.7716809041382007,-48.754547527456694 -22.7720043701368979,-48.7548035250961931 -22.7723927031708016,-48.7547812009946 -22.7725684404789988,-48.7545948778579969 -22.7726757812363942,-48.7537035375193852 -22.7726699900770981,-48.7533235153789946 -22.7725431948592956,-48.7522697556455 -22.772732851557798,-48.7516824966764943 -22.7725218312247932,-48.7508777841260041 -22.7725692830840956,-48.7506431146658912 -22.7723365040380017,-48.750689416763791 -22.7720183991854981,-48.7505730254076894 -22.771572514123001,-48.7508624926218914 -22.7713164008444977,-48.7508417842548027 -22.7712211857418012,-48.7504809519170976 -22.7709752678782991,-48.7504362906499011 -22.7707424150811946,-48.7506540648636957 -22.7703993193088934,-48.7502220803880988 -22.7700016116507982,-48.7501323101636928 -22.7697762836075981,-48.7496025277605867 -22.7694897126118008,-48.7494461445098892 -22.7692991146355013,-48.7493212557817941 -22.7688970679021949,-48.749056369809594 -22.7685885686745948,-48.7488178115144848 -22.7684573998070974,-48.7479729404344937 -22.7684855170097933,-48.7478364694790969 -22.7686395019161978,-48.7477060402497955 -22.7696265017400989,-48.7469726046203959 -22.769848239311095,-48.7469022156952008 -22.7701844589693962,-48.7467005616338867 -22.7704119329848993,-48.7461690241586894 -22.7702981592781981,-48.7457308398088855 -22.770316195535699,-48.7453829767540014 -22.7705703823245003,-48.7452392518575977 -22.7709131976495982,-48.7450580178683026 -22.7709639720910957,-48.7445605464634895 -22.7708765202825987,-48.7440237742543019 -22.7709498381527951,-48.7438671622888933 -22.7707907400241965,-48.7439300136996962 -22.7701688681142969,-48.7437970276859005 -22.7699781332867985,-48.7431270249118995 -22.7695626939547964,-48.7426981249519997 -22.7690345220758985,-48.7421402840712972 -22.7687686162618945,-48.7408028019796973 -22.7686589766445948,-48.7394798674365859 -22.7683229197781927,-48.7391044593473879 -22.7680512924954002,-48.7389704511363959 -22.767760584376596,-48.7389910676237008 -22.7670211871801982,-48.7388121936079983 -22.7665795499998964,-48.738443654892194 -22.7662307190807951,-48.7374281241539933 -22.7659587675360982,-48.7369251944191006 -22.7654308539885974,-48.7366761318454991 -22.7653452414274966,-48.7356638581978956 -22.765318277176199,-48.7342941677406998 -22.7655910708539011,-48.7326893114471034 -22.7653915496488963,-48.7317154781633946 -22.7653784291938948,-48.7307621159954891 -22.7649599294710967,-48.7299897916536935 -22.7649768362470937,-48.7294289589464995 -22.764663076901499,-48.7294289589464995 -22.764663076901499,-48.7287349799703904 -22.7645177965445171,-48.7287349799703904 -22.7645177965445171,-48.7277984910411988 -22.7638891613040002,-48.7268463157213958 -22.7638452145638013,-48.7268463157213958 -22.7638452145638013,-48.7260309630633941 -22.7635374715381005,-48.7257643283390962 -22.7637312647210948,-48.7257643283390962 -22.7637312647210948,-48.7256288752390958 -22.7643642315372006,-48.7256288752390958 -22.7643642315372006,-48.7251340825435761 -22.7645622146060589,-48.7251340825435761 -22.7645622146060589,-48.724584139476292 -22.7650788932430004,-48.7239256124523976 -22.7652273636601699,-48.7239256124523976 -22.7652273636601699,-48.7235723565314913 -22.7659494867045993,-48.7235723565314913 -22.7659494867045993,-48.7231706528736481 -22.7660830050614749,-48.7231706528736481 -22.7660830050614749,-48.7229452698611851 -22.7660404673428971,-48.7224607194467012 -22.7655932507124987,-48.7221968237586793 -22.7654882216165291,-48.7221968237586793 -22.7654882216165291,-48.7210356199003058 -22.7653311140413983,-48.7210356199003058 -22.7653311140413983,-48.7208624992437933 -22.7651966632829001,-48.7204560964949991 -22.7645170353997983,-48.7202667836601364 -22.7643982562216252,-48.7185812108386926 -22.764397608710496,-48.7185812108386926 -22.764397608710496,-48.7166554032806971 -22.764108122844199,-48.7163537837443954 -22.7641729000841977,-48.7163537837443954 -22.7641729000841977,-48.7154458106146961 -22.7646246140321935,-48.7154458106146961 -22.7646246140321935,-48.7151204157180899 -22.7646489906862968,-48.7141979465846902 -22.7644664008499973,-48.7136094106139979 -22.7647449809341964,-48.7131079926231934 -22.7648616776317994,-48.7129033343807976 -22.7649890684024001,-48.7126689346585948 -22.7653315508350005,-48.7121749511982927 -22.7654574926278954,-48.7120309707507886 -22.7655778820506001,-48.7115893494250969 -22.7665932521721004,-48.7115893494250969 -22.7665932521721004,-48.7117315041760008 -22.766999482189199,-48.7117315041760008 -22.766999482189199,-48.7116000401815015 -22.7674011796339713,-48.7117392798426891 -22.7679585649508951,-48.7116058382070065 -22.7685627658017964,-48.7116058382070065 -22.7685627658017964,-48.7116052248425007 -22.7690831224034937,-48.7109173156262898 -22.7703946586854968,-48.7109173156262898 -22.7703946586854968,-48.710831028434697 -22.7710139198393833,-48.710831028434697 -22.7710139198393833,-48.7104995390633917 -22.7713720964290971,-48.7091962223217863 -22.7716891649517734,-48.7091962223217863 -22.7716891649517734,-48.7087434997544904 -22.7722089873764943,-48.7081749003280891 -22.7724999032419007,-48.7078589695206858 -22.7729090037779969,-48.7079813364827885 -22.7736251866894008,-48.7079044198527953 -22.7738252127485978,-48.7069141527330913 -22.7741598060599983,-48.7069141527330913 -22.7741598060599983,-48.7063631203039407 -22.7747267862649494,-48.7063631203039407 -22.7747267862649494,-48.7057599969921 -22.7748265865267925,"; 
    wkt1 += "-48.7054746796655564 -22.7749981131646564,-48.7054746796655564 -22.7749981131646564,-48.7049658527592868 -22.7749524786500075,-48.7049658527592868 -22.7749524786500075,-48.7020571784881895 -22.7751551677680908,-48.7020571784881895 -22.7751551677680908,-48.7018294149386861 -22.7752745735546966,-48.7015108399859002 -22.7758519253964948,-48.7005527936166942 -22.7757058832682091,-48.7005527936166942 -22.7757058832682091,-48.70005181743619 -22.7761871648501,-48.6994769069212907 -22.7762626653553824,-48.6994769069212907 -22.7762626653553824,-48.6991673742184901 -22.776606233741596,-48.6991673742184901 -22.776606233741596,-48.6984216610692897 -22.7767643041514987,-48.6980287937450953 -22.7773220808247956,-48.6980287937450953 -22.7773220808247956,-48.6976431555730471 -22.7774668615990876,-48.6976431555730471 -22.7774668615990876,-48.6974280170066862 -22.7774111212009984,-48.6970150728733984 -22.7770535339632971,-48.6965300330522695 -22.7771734128705781,-48.6965300330522695 -22.7771734128705781,-48.6960394529135101 -22.7770842393811996,-48.6951937341906955 -22.777283095183499,-48.6949523209907937 -22.7772264811558998,-48.6945096688370995 -22.7774593937335013,-48.6943286626664928 -22.777402979780895,-48.6941967521744132 -22.776884544192292,-48.6941967521744132 -22.776884544192292,-48.693430251361093 -22.7767240999853939,-48.6927308699115144 -22.77626518138457,-48.6927308699115144 -22.77626518138457,-48.6919925428903895 -22.7759814254718975,-48.6915057195786005 -22.7759211728861111,-48.6915057195786005 -22.7759211728861111,-48.6908415597725934 -22.7759360647743989,-48.6895173126740985 -22.7763882081542022,-48.6887673248349984 -22.7764803696785947,-48.6853859280262924 -22.7765590329139975,-48.6853859280262924 -22.7765590329139975,-48.6837749319996931 -22.7767414270364981,-48.6812735515148916 -22.777193122179181,-48.6812735515148916 -22.777193122179181,-48.6796320558953965 -22.7777084608413958,-48.6796320558953965 -22.7777084608413958,-48.6791638984446777 -22.7779449805771499,-48.6791638984446777 -22.7779449805771499,-48.678376294680703 -22.7786030122767968,-48.678376294680703 -22.7786030122767968,-48.6760071381985497 -22.7811395343785463,-48.6760071381985497 -22.7811395343785463,-48.6741457288433992 -22.7833219904127944,-48.6741457288433992 -22.7833219904127944,-48.6731550164683995 -22.7843144313815955,-48.6731550164683995 -22.7843144313815955,-48.670614055977893 -22.7863274738113972,-48.6687220981484217 -22.7875140357497763,-48.6687220981484217 -22.7875140357497763,-48.6667289336621565 -22.7882461678231749,-48.6667289336621565 -22.7882461678231749,-48.6655095460602922 -22.7884278969938947,-48.6631568015054938 -22.7881847211268962,-48.6599110356213913 -22.7876515484208113,-48.6599110356213913 -22.7876515484208113,-48.6548287061717843 -22.7873333726708012,-48.6537385066531982 -22.7875256266712967,-48.652493536941428 -22.7882366449661795,-48.652493536941428 -22.7882366449661795,-48.6517329750741965 -22.7889206265371982,-48.6501956712838961 -22.7908758929211004,-48.6497001497824897 -22.7910618835027989,-48.6497001497824897 -22.7910618835027989,-48.6490242795673993 -22.7910500913317975,-48.6482874360565773 -22.7907535819360376,-48.6482874360565773 -22.7907535819360376,-48.6475000742098018 -22.7899800535357997,-48.6449014030098184 -22.7851659834128384,-48.6449014030098184 -22.7851659834128384,-48.643119950875132 -22.7825299063892679,-48.643119950875132 -22.7825299063892679,-48.6409057530036932 -22.779972060478002,-48.6401926453420401 -22.7796187054598178,-48.6401926453420401 -22.7796187054598178,-48.6389315690917954 -22.7794477530965978,-48.6179240468165972 -22.779250725991794,-48.6179240468165972 -22.779250725991794,-48.6153065100928927 -22.7737317746600993,-48.6148587873719009 -22.7730748732986967,-48.6143503668177885 -22.7727967173875996,-48.6131814771309934 -22.7726871905337944,-48.6123785201547918 -22.7728861326711964,-48.6117349941341956 -22.7732182632296016,-48.6111720018569997 -22.7737817519837975,-48.6108499332831911 -22.7745369867535956,-48.61082559183059 -22.7752293233465011,-48.6110057003413942 -22.7759015340390008,-48.611734542554089 -22.7768676716605967,-48.6122114920604957 -22.7772771084757935,-48.6123600428837932 -22.7778600373437001,-48.6127844962437905 -22.7785386684588929,-48.6129609675769956 -22.7794520115277948,-48.6129114604731001 -22.7808330597506981,-48.6130492755915924 -22.7825795921452965,-48.6129347185763976 -22.7837811111964967,-48.6129722833073998 -22.7848068077206989,-48.6131531240289974 -22.785782413108393,-48.613089440544087 -22.7866075580246985,-48.6127596544585003 -22.7874588555231981,-48.6128409141659006 -22.7887110735684999,-48.6125537369910035 -22.7904595064082969,-48.6114717798386948 -22.7925421624495925,-48.6107889743147012 -22.7931511141034946,-48.6104260706122915 -22.7937537313407006,-48.6096897331313897 -22.7944842013368003,-48.6092572039307953 -22.7950727668503959,-48.6075046818441905 -22.7969068417606948,-48.6058412387576908 -22.7984333250355959,-48.6055607965245002 -22.7988469634961959,-48.6051948823798909 -22.799030870154894,-48.6047353209034014 -22.7995131838155984,-48.6032886834871931 -22.8002862770917005,-48.6031102055869013 -22.8006349337330008,-48.6032059892786918 -22.8011710792403974,-48.6030640464108927 -22.8023221404068011,-48.6034512129255916 -22.802562948757199,-48.6042058435802034 -22.8026550145950964,-48.6046342385456001 -22.8028143929944989,-48.6052102663752947 -22.8034558523733963,-48.605188894791695 -22.8042193981284953,-48.6054315196871869 -22.804936933990497,-48.6054376852954846 -22.8056942069197959,-48.6053507903192923 -22.8060189481386999,-48.6048281290974984 -22.8069384259770018,-48.6048191788313915 -22.807510648497896,-48.6049552245343932 -22.8077924970620955,-48.6053298098044877 -22.8080067738114991,-48.6056697928428889 -22.8079372213346971,-48.6059273010058916 -22.8080557287462966,-48.6059903253731989 -22.8083603200442013,-48.6057823113158918 -22.8089384333823944,-48.6058676057111967 -22.8093454979563006,-48.6063608517123953 -22.8098131292664945,-48.6068262329944929 -22.8100484692815968,-48.6064450066154023 -22.8102984726181006,-48.6054992336451903 -22.8106275574295978,-48.6052481245621948 -22.8108405365655997,-48.6045294662632017 -22.8111691728120007,-48.603979935797696 -22.8115836284470959,-48.6025008783707904 -22.8119154054672961,-48.6020183825612975 -22.8123105104643997,-48.6011160549151953 -22.8133152752883959,-48.5986923174957894 -22.814636835813296,-48.5983395445607869 -22.814955290571497,-48.5971775365306939 -22.8153283250751926,-48.5957386278008983 -22.8151365784749984,-48.5953072355644906 -22.8151508063614941,-48.5948318668778967 -22.815315678067293,-48.5944121678163015 -22.8153546036218984,-48.5935993609451913 -22.8150561732311985,-48.5921228059879979 -22.8155445607109968,-48.5913858375077865 -22.8154233192983966,-48.5908228177683981 -22.815438484392196,-48.5905213304370989 -22.8156920399457945,-48.5901964096717904 -22.8163894911140019,-48.5893976459835955 -22.817163628584197,-48.5889855565283 -22.8177483046782967,-48.5883225654449973 -22.8181705579848995,-48.588030513302698 -22.8185792570614012,-48.5873126396315911 -22.8202382914509982,-48.5873680614669894 -22.8209549082029994,-48.5872759209911891 -22.8216071181450957,-48.5874738125375956 -22.821901873748196,-48.5876081738236962 -22.8224222236326,-48.5876034278617936 -22.8226216311662249,-48.7805839406553119 -22.8226216311662249,-48.7805839406553119 -22.8226216311662249,-48.7805839406553119 -22.8226216311662249,-48.7805462353790062 -22.8214948881798989,-48.780301490333585 -22.8209106572390965,-48.7790127856201963 -22.8201743862837958,-48.778755729720487 -22.819886875361,-48.778286157748397 -22.8189882730542948,-48.7780992532984925 -22.8182681730282013,-48.7778728074926846 -22.8178430039835973,-48.7764568028833594 -22.8169702063427415,-48.7764568028833594 -22.8169702063427415,-48.7761596379381928 -22.816686784797696,-48.7760915505917012 -22.8164027442728994,-48.7765419516445888 -22.814921943428395,-48.776402211381793 -22.8140647272331947,-48.7756621018719585 -22.8128754688271194,-48.7756621018719585 -22.8128754688271194,-48.7742479535055935 -22.8114715047064962,-48.7741443478451941 -22.8100266289704976,-48.773828741377045 -22.8094103675903739,-48.773828741377045 -22.8094103675903739,-48.773871051204793 -22.8089921785983947,-48.7742107124063864 -22.8082658758091981,-48.7742633492551789 -22.8078579279585938,-48.7742633492551789 -22.8078579279585938,-48.7741694422143013 -22.8072552140739973,-48.7737938959860529 -22.8064479177540136,-48.7737938959860529 -22.8064479177540136,-48.7738688173024997 -22.8060433993551968,-48.7740559408064911 -22.8056866649538961,-48.7751538138790863 -22.8047060149880991,-48.7753371805921745 -22.8041159597776897,-48.7753371805921745 -22.8041159597776897,-48.7752078443276957 -22.8026297265434081,-48.7752702718306921 -22.8023733594421003,-48.7752702718306921 -22.8023733594421003,-48.7756559076855538 -22.8019987954352459,-48.7756559076855538 -22.8019987954352459,-48.775774479356194 -22.8010383712025799,-48.775774479356194 -22.8010383712025799,-48.7759018329165883 -22.800842647939696,-48.7759018329165883 -22.800842647939696,-48.7767003621045632 -22.8006040138890071,-48.7767003621045632 -22.8006040138890071,-48.7770946446630944 -22.8003854916663968,-48.7773831733147816 -22.7997718540063445,-48.7777647412583519 -22.7995799256886187,-48.7777647412583519 -22.7995799256886187,-48.7791450609998947 -22.7983464473645938,-48.7795350123199967 -22.7979603065852956,-48.7795350123199967 -22.7979603065852956,-48.7798494507450044 -22.7974843643538989,-48.7795580896270025 -22.796922027743399,-48.779534598020895 -22.7966439459660961,-48.779534598020895 -22.7966439459660961,-48.7797231186188895 -22.7962242200202994,-48.7797138591247972 -22.7959051853619954,-48.779481760445293 -22.7948882929900023,-48.7795000063342812 -22.7945890636886972,-48.7795000063342812 -22.7945890636886972,-48.7797286774921659 -22.7940364914764828,-48.7797286774921659 -22.7940364914764828,-48.7794414045479101 -22.7932906128472261,-48.7794414045479101 -22.7932906128472261,-48.7793947046146883 -22.7926759344086989,-48.7797284400858899 -22.7890898261172978,-48.7797284400858899 -22.7890898261172978,-48.7797284400858899 -22.7890898261172978,-48.7795887326042958 -22.7883586666993132,-48.7795887326042958 -22.7883586666993132,-48.7792385042406949 -22.7873730372512,-48.7792150302316969 -22.7869923136470049,-48.7792574980513933 -22.7866200436888988,-48.7795092528371868 -22.7859727830869971,-48.779593133230577 -22.7852813302853932,-48.779593133230577 -22.7852813302853932,-48.7791466426613312 -22.7844690354698756,-48.7791466426613312 -22.7844690354698756,-48.7792556338890932 -22.7835280979775021,-48.7792556338890932 -22.7835280979775021,-48.7800952233800018 -22.7822868515923957,-48.7801584793685947 -22.7805520906703975,-48.7804727830174869 -22.7792351854799975,-48.7805927844761982 -22.7790013987824977,-48.781067432571902 -22.7785387732640956,-48.781067432571902 -22.7785387732640956)))";

    std::string wkt2 = "multipolygon(((-48.8025679611872931 -22.8453336950107015,-48.8024555064786938 -22.8451251955265988,-48.8023430517701939 -22.844916696042695,-48.802074872663745 -22.8446630613183466,-48.802074872663745 -22.8446630613183466,-48.8005941206087925 -22.8436883614352979,-48.8001140797639295 -22.8428803884638398,-48.8001140797639295 -22.8428803884638398,-48.7996782339786961 -22.8418082375605991,-48.7991682807375824 -22.841222587260205,-48.7991682807375824 -22.841222587260205,-48.7970994737853445 -22.839794729742831,-48.7970994737853445 -22.839794729742831,-48.7969746464586933 -22.8392846446533966,-48.7971283203265926 -22.8380155961410978,-48.7971283203265926 -22.8380155961410978,-48.7971283203265926 -22.8380155961410978,-48.7970409262914941 -22.8369583922272952,-48.7970409262914941 -22.8369583922272952,-48.7966932301130072 -22.8362852819354956,-48.796020898694195 -22.8354593367232965,-48.7938238323515918 -22.8335122557985954,-48.7932232065540887 -22.832851609751998,-48.7914775868922987 -22.834045355838299,-48.7914775868922987 -22.834045355838299,-48.7908910387105976 -22.834752845917798,-48.7905012129603932 -22.8355162681566952,-48.7901770960641983 -22.8366266784800942,-48.7899404323153476 -22.8368251491972458,-48.7899404323153476 -22.8368251491972458,-48.7896502039242961 -22.8369201293130999,-48.7894508291090006 -22.8369334357875928,-48.7894508291090006 -22.8369334357875928,-48.7891672720565239 -22.8368026363494572,-48.7891672720565239 -22.8368026363494572,-48.7868620865337022 -22.8346511604386961,-48.7861794395314945 -22.8334845038064955,-48.7855787426023966 -22.8321377431054984,-48.784322548930092 -22.8307515819504943,-48.7832148045441585 -22.8298177481311484,-48.7832148045441585 -22.8298177481311484,-48.7829264693012732 -22.8294553878965054,-48.7829264693012732 -22.8294553878965054,-48.7824819109233587 -22.8286152295395119,-48.7824819109233587 -22.8286152295395119,-48.7810261891573518 -22.8275286659431451,-48.7810261891573518 -22.8275286659431451,-48.7801416576308924 -22.8267236571661982,-48.7800121954465027 -22.8263619804936937,-48.7802694985226992 -22.8258274578718954,-48.7806049326898972 -22.8232489340312981,-48.7806049326898972 -22.8232489340312981,-48.7806049326898972 -22.8232489340312981,-48.7805839406552479 -22.8226216311662284,-48.7805839406552479 -22.8226216311662284,-48.5876034278617936 -22.8226216311662249,-48.5875862099759956 -22.8233450622029963,-48.5878665950115902 -22.8270911911818004,-48.5877982834801969 -22.8281501879225956,-48.5875057820997966 -22.8289873515482995,-48.5865157364965938 -22.830722673050893,-48.5861981953371895 -22.8314779682654958,-48.5854582728016879 -22.8357530419670951,-48.5851037936860877 -22.836529585082296,-48.5833272571606898 -22.8390906851480011,-48.5817347986432964 -22.8408199376087993,-48.5806340592332901 -22.842514536611894,-48.5786985916864893 -22.8447524091865972,-48.577012798670296 -22.8461282629357996,-48.5755521795163929 -22.8468915259248959,-48.5738011934135869 -22.8475764983778973,-48.5728109321879913 -22.8481567709252005,-48.5712375358352944 -22.8489107549197001,-48.5696286445010941 -22.8499434633259995,-48.5685515666249898 -22.8509888343856957,-48.567314908655689 -22.8534447134898997,-48.5682392556108908 -22.853712891607799,-48.5689430617428002 -22.853765629606194,-48.5716037037907995 -22.8536472400511954,-48.5726301835360914 -22.8538972615523974,-48.5734460147183995 -22.8543444571873948,-48.5738947099260088 -22.8543642020518973,-48.5749684939391955 -22.8536463405635928,-48.5757326282231929 -22.8534571534301989,-48.5769605079687921 -22.8534484386607986,-48.5803883824271026 -22.8538960502535957,-48.5830944741396991 -22.8544001293359997,-48.5859087427873959 -22.8544640464074007,-48.587240669704201 -22.8548062491545991,-48.589103252442996 -22.8551178112377968,-48.5899613185119961 -22.8554358370782964,-48.5908517369075952 -22.8559214880612984,-48.5919977674174959 -22.8560593190443981,-48.5957438153918986 -22.8536241969876954,-48.5959654274591912 -22.8535677821259995,-48.5975738082108961 -22.8536509990796972,-48.5985754101515965 -22.8535210405736962,-48.5996620875167906 -22.8531878086127946,-48.6017322148059918 -22.8520992420392943,-48.6041443875941894 -22.851482197657198,-48.6054442473417936 -22.8513784713720014,-48.6060169910490032 -22.8514662286291994,-48.6071540747056048 -22.8518295948533989,-48.6088882532038014 -22.8519610528217001,-48.6100073963104933 -22.8522112699985982,-48.6102939885036989 -22.8521633281657941,-48.6111092013832007 -22.8517902449040022,-48.6119698203313959 -22.8517976736568968,-48.6136512547903976 -22.8520056191204972,-48.6150453043049993 -22.8516646358112965,-48.6167991698112019 -22.8516679428621998,-48.6178766552156958 -22.8514918044135982,-48.6224799727095984 -22.8503685493036954,-48.623451575480189 -22.8505281866189023,-48.624702187752689 -22.8510545935935951,-48.6252427250702937 -22.8510691825334007,-48.625536471560892 -22.8507405853643988,-48.6255601742783909 -22.8497682489153959,-48.6256658169897946 -22.8495368468322013,-48.625793428534692 -22.8493169863483985,-48.6267485845236962 -22.8486073396358975,-48.6269679100682026 -22.8476433818943931,-48.6271792890536005 -22.8473264129521993,-48.6275698011283026 -22.8470901861591926,-48.6285434953602902 -22.8470725168848006,-48.6291889784687896 -22.8468232533871003,-48.6300233639478989 -22.8461590334556988,-48.6307916543783847 -22.8453161430355962,-48.6310504118664966 -22.8451573979692952,-48.6316517835134903 -22.844913841728598,-48.6327347180514948 -22.8447458768559954,-48.6330974733557895 -22.8445456105777964,-48.6335366209555957 -22.8435188023603999,-48.6335239618684909 -22.8424777173135993,-48.6337266360125966 -22.8420344121763961,-48.6340976055695933 -22.8416498719489951,-48.6349856826358007 -22.8415275562784998,-48.6353459612818995 -22.8412659844587012,-48.635536347757494 -22.8409365920648995,-48.6357515758841998 -22.8400994049786945,-48.6360920200557985 -22.8397966545433988,-48.6369415076643037 -22.8393660525529967,-48.6373604062433031 -22.8388777840591999,-48.637667688212602 -22.8388232797625932,-48.6382933607450951 -22.8390765343824,-48.638880923122997 -22.8391088520655003,-48.6402326685974913 -22.8383007793556985,-48.6410075842627947 -22.8379810802923942,-48.6414835572812905 -22.8380744621748946,-48.6421882486640982 -22.8387732669424963,-48.6427524441945849 -22.8388047868376987,-48.6430511323657981 -22.8386959696303933,-48.6441078813173959 -22.8380013022513992,-48.6451865833028876 -22.8376908627502999,-48.6459404618044857 -22.8370373445351973,-48.6465617685962002 -22.8370293525419967,-48.6468389458988923 -22.8369468136960947,-48.6469286078812928 -22.8368480903079956,-48.6469529048077902 -22.8365518254487014,-48.6467312859972907 -22.8359637476825981,-48.6469412900581943 -22.8356800053236952,-48.6471524509963018 -22.8355655462180991,-48.6473376923196952 -22.8355402425288929,-48.6481061331926981 -22.8358072435456947,-48.6484075837146008 -22.8357768280444944,-48.6495102386614917 -22.8352276930232989,-48.6500814312182968 -22.8351549912333951,-48.6510877603781893 -22.8346008183355949,-48.6512628713464892 -22.8342618860957955,-48.6512722843671952 -22.8337481622823972,-48.6514841119462886 -22.8335418939038988,-48.6517222002761969 -22.8335858714458979,-48.6520136190756887 -22.8341843830569999,-48.6523942265633025 -22.8339747588679955,-48.6526615101579978 -22.8340449189335004,-48.6529088136841921 -22.8339532662681961,-48.6531331578730999 -22.8336762947854979,-48.6535244526671988 -22.8336416725977998,-48.6538655245441873 -22.8334955393927004,-48.6544016329446976 -22.8327752701318971,-48.6547606503926886 -22.8325442113166943,-48.655157896976192 -22.8326313169888948,-48.6553605456564924 -22.8328334476408941,-48.6556351119386932 -22.8329416565733965,-48.6557734056671904 -22.832920173864494,-48.6560174825838914 -22.8326618619580941,-48.6570161594076893 -22.8326754464196,-48.6579043450708895 -22.8322991242261963,-48.6588814600901998 -22.8321661352788965,-48.6598670631532002 -22.8317912558730995,-48.660546709986491 -22.8316618203270991,-48.6603573409293944 -22.8320669091924984,-48.6603763709404973 -22.8324276459993953,-48.6610537820274942 -22.8331029604229983,-48.6612320354392978 -22.8336049547741951,-48.6610452657711932 -22.8355630679527977,-48.6609428101248938 -22.8388788779899983,-48.6609975695940022 -22.8393479390528995,-48.6617494892624904 -22.8403372337690982,-48.6619046747675981 -22.8408060493107001,-48.6621118245306974 -22.8422389726887971,-48.6629892584833925 -22.8433098807593993,-48.663175815897489 -22.8445467617281963,-48.6630486414864976 -22.8454049576419962,-48.6634190973152982 -22.8461528421540976,-48.6636412841090902 -22.8474260240045979,-48.6637951847202004 -22.847604014360698,-48.6646372521646953 -22.8479850428402926,-48.6675125124320971 -22.8475075888386954,-48.6690376365761921 -22.8476242134838969,-48.6696510685540886 -22.8477689207335999,-48.6704839366962005 -22.8484106902946991,-48.6714105769295884 -22.8486738732363932,-48.6718235788195983 -22.8491575881898967,-48.672118712591697 -22.8498633372499995,-48.6721910136811999 -22.8507327035699959,-48.6717818033467893 -22.8519731206022954,-48.6717047279207975 -22.8525233112647967,-48.6717979961887011 -22.8529558487887954,-48.6721667803816871 -22.8534236855818946,-48.6727731183762984 -22.8537256950959957,-48.673452828138096 -22.8538717031712935,-48.6744856307294924 -22.854263552797498,-48.6746922471820085 -22.8544153856746988,-48.6749807710649947 -22.8548166227266947,-48.6754109397268024 -22.8558599718113982,-48.675776290819698 -22.8564411226789019,-48.6764296568607904 -22.8568626248014013"; 
    wkt2 += ",-48.6771526002127928 -22.8576996678021018,-48.677776445287094 -22.8582282143188991,-48.678757162829001 -22.8587065338100004,-48.6797246940047899 -22.8590232602917993,-48.6801330745480954 -22.8590521715705961,-48.6809686353991964 -22.8589530695236007,-48.6823579446357968 -22.858155786865197,-48.6830864910831878 -22.8580504079443969,-48.6833634423101955 -22.8581991723039977,-48.6833870476753958 -22.8585438635934963,-48.6824348520953976 -22.8609077870458961,-48.6821747443494957 -22.8625511274830018,-48.6822388417457006 -22.863157346629599,-48.6825037961936999 -22.8641384684445015,-48.6830397955804983 -22.8653950995445996,-48.6838216776136932 -22.8667976613881976,-48.6862129247377879 -22.8701632960022962,-48.6871020867395856 -22.8716666528087984,-48.6886189809778926 -22.8734267736951935,-48.6890237048101966 -22.8737256318595001,-48.6903276357828929 -22.8743461007033986,-48.6909146667123025 -22.8743710325503002,-48.6914284219341909 -22.8742494794321942,-48.6923034566122936 -22.8736150746945981,-48.6927946569827981 -22.8729210854010958,-48.6930701469744918 -22.872207317227101,-48.6941485730261974 -22.8707315921285996,-48.6952563657364976 -22.8697780986419019,-48.6957090817329998 -22.8695707730161999,-48.696235810931185 -22.8694910453985969,-48.6970244695851022 -22.8695460590057991,-48.6975199302674895 -22.8694679890208974,-48.6977542524046925 -22.8692849117232981,-48.6989089172975866 -22.8677522562330964,-48.6990816799410027 -22.8676976741068998,-48.6992653123135 -22.8677145757822942,-48.6995371855219901 -22.8680198000067989,-48.6997636940932921 -22.868538687445195,-48.7001989284598977 -22.8701088449315932,-48.7006550819821911 -22.8706902752086947,-48.7021457289862951 -22.8717003762843021,-48.7030822857289962 -22.8721149881232009,-48.7037304921097984 -22.872296753037098,-48.7047034031313899 -22.872353344599702,-48.7051251701437877 -22.8720891460711968,-48.7053788019801885 -22.8716600157135943,-48.706053737391187 -22.8692470971829991,-48.706288489723093 -22.8687264077289996,-48.7067139351167953 -22.8682165430675965,-48.7084333693655935 -22.8671526793934987,-48.7096250164278928 -22.8662207523313938,-48.7103586099791954 -22.8657597999861011,-48.7112456588949954 -22.8654398735680999,-48.7131041947837957 -22.8653123736352981,-48.713861558271887 -22.8651681974852963,-48.7153309622505901 -22.8648918378874981,-48.7161028335978017 -22.8650020177762983,-48.7164441586084962 -22.8649502936882953,-48.7190577561032967 -22.8621458758046998,-48.7195250497602927 -22.8619380550545941,-48.7200512769221916 -22.8619455559457947,-48.7204671087732919 -22.8621096602054976,-48.7208294001143969 -22.8625735981881988,-48.7216779986566024 -22.8649316715791002,-48.7219697984895959 -22.8660829190203998,-48.7233365492061949 -22.8675506043001988,-48.7246185852211937 -22.868587831613894,-48.7254615981524921 -22.869001928574292,-48.7267871223790934 -22.8692942152041994,-48.727266669299695 -22.8692740451992975,-48.7290913017605902 -22.8687382993715005,-48.7316929608781919 -22.8675997528871946,-48.7320656574018898 -22.8675184198972019,-48.7333328511770958 -22.8675791437976983,-48.7334183096483997 -22.8677899389465935,-48.7333660279213916 -22.868965840721998,-48.7335290195548936 -22.8695420258950008,-48.7339326270119955 -22.8703512321117941,-48.7343706273912005 -22.8708257506868975,-48.7358730649994953 -22.8716378921512984,-48.7377926469279004 -22.8724808837528002,-48.7411954545992927 -22.8733475306367993,-48.7427171966791946 -22.8733832645910979,-48.7447007206443885 -22.8722925638353942,-48.7460099416670971 -22.8716946302224997,-48.7479783538224893 -22.8714497040267943,-48.749817218255302 -22.8710744427630956,-48.7517850756146913 -22.8709446931149998,-48.7525472372680895 -22.8706249224018947,-48.7547696464548892 -22.8689486030782945,-48.7558864227366939 -22.8683577575511947,-48.7566541752376992 -22.8677653412322996,-48.7581624775368994 -22.8663239013124979,-48.7587166257821991 -22.8655551908917971,-48.7591737548604982 -22.8647229686820026,-48.7594246724350882 -22.8645115204947018,-48.7599373538206038 -22.864389696596696,-48.7625235411497897 -22.8641261340252981,-48.7632313605158885 -22.8637216868418989,-48.7635014468881991 -22.8632902724176006,-48.7635715643777985 -22.862480450554898,-48.7638207674242921 -22.8613235625096962,-48.7642245901466964 -22.8605832076928976,-48.7645688027867976 -22.8602523599972969,-48.7651368437939965 -22.8601863531708958,-48.7654973310274897 -22.8598632394832961,-48.7667180758094929 -22.8593036195493013,-48.7676972412230043 -22.8584839707722978,-48.7684474194834934 -22.8573047698155989,-48.7688445010251854 -22.8557988687602958,-48.7694223368928022 -22.8541809387380965,-48.7699461453340888 -22.8534372887189008,-48.7710011056045971 -22.8526732140882984,-48.7714092139895001 -22.8525056035966969,-48.7725419139835878 -22.8523581114002994,-48.7771529887211983 -22.8526242354874967,-48.7788125704630886 -22.8530260495890971,-48.7795779180912916 -22.8528475876104977,-48.7803820797796988 -22.8524191860106924,-48.7826157939905016 -22.8508111793826956,-48.7839917405808023 -22.8502168580440994,-48.7852576851980899 -22.8498728380750933,-48.7869553644291898 -22.849864336216001,-48.787629095105487 -22.8499394494180024,-48.7880373458966972 -22.8498438272765014,-48.7887192881856038 -22.8491682850225999,-48.7891808132463893 -22.8488520032749989,-48.7914699482940932 -22.8480857156322941,-48.7917562873480932 -22.8477357997243971,-48.7924538302508921 -22.8475955752264994,-48.7929009944852936 -22.8471698655797937,-48.7931135665671007 -22.8470921594134992,-48.7950866257805984 -22.8470196117472,-48.796176756271592 -22.8472703270085944,-48.7965837968719001 -22.8472826883118998,-48.7991703713471878 -22.8469977758375933,-48.7995964399787994 -22.8468945982543943,-48.8000569300425013 -22.8466556802530931,-48.8017669011885999 -22.8455048130715994,-48.8021013536504995 -22.8453752358185014,-48.8025679611872931 -22.8453336950107015)))";

    vecWKTs.push_back(wkt1);
    vecWKTs.push_back(wkt2);

    te::gm::GeometryVectorConst input;
    for (std::size_t i = 0; i < vecWKTs.size(); ++i)
    {
      input.push_back(te::gm::WKTReader::read(vecWKTs.at(i).c_str()));
    }

    double tolerance = te::gm::GetAppropriatePrecision(4674);
    te::gm::GeometryVector output = te::vp::GeometrySnapper::snapGeometries(input, tolerance);


    for (std::size_t i = 0; i < vecWKTs.size(); ++i)
    {
      std::string wktResultA = output.at(i)->toString();
      ASSERT_TRUE(output.at(i)->isValid());

    }
    te::common::FreeContents(input);
    te::common::FreeContents(output);
  }

  TEST_F(ItsPrepareToOverlay, SnapGeometries)
  {
    te::gtest::DataInfo redDataInfo = te::gtest::GetRedDataInfo();

    te::vp::FeatureSet featureSet;
    te::gtest::GetFeaturesFromShapefile(redDataInfo.m_uri, featureSet);

    ASSERT_FALSE(featureSet.empty());

    te::gm::GeometryVectorConst inputGeometryVector = te::vp::GetGeometriesFromFeatureSet(featureSet);

    bool wasChanged = false;
    te::gm::GeometryVector vecPrepared = te::vp::GeometrySnapper::snapGeometries(inputGeometryVector, te::gm::GetAppropriatePrecision(inputGeometryVector.at(0)->getSRID()));

    te::gtest::SimpleSaveGeometriesToShapefile(te::gtest::GetOutputDir(), redDataInfo.m_datasetName + "_snapped_geometries", te::gm::ToConstVector(vecPrepared));

    te::common::FreeContents(featureSet);
    te::common::FreeContents(vecPrepared);
  }

  TEST_F(ItsPrepareToOverlay, SnapFeatures)
  {
    te::gtest::DataInfo redDataInfo = te::gtest::GetRedDataInfo();

    te::vp::FeatureSet featureSet;
    te::gtest::GetFeaturesFromShapefile(redDataInfo.m_uri, featureSet);

    ASSERT_FALSE(featureSet.empty());

    te::vp::FeatureSet outputFeatureSet = te::vp::GeometrySnapper::snapFeatures(featureSet, te::gm::GetAppropriatePrecision(featureSet.at(0)->getFirstGeometry()->getSRID()), true);

    te::gm::GeometryVectorConst inputGeometryVector = te::vp::GetGeometriesFromFeatureSet(outputFeatureSet);

    te::gtest::SimpleSaveGeometriesToShapefile(te::gtest::GetOutputDir(), redDataInfo.m_datasetName + "_snapped_features", inputGeometryVector);

    te::common::FreeContents(featureSet);
    te::common::FreeContents(outputFeatureSet);
  }

  TEST_F(ItsPrepareToOverlay, SnapAllFeatures)
  {
    te::gtest::DataInfo redDataInfo = te::gtest::GetRedDataInfo();
    te::gtest::DataInfo blueDataInfo = te::gtest::GetBlueDataInfo();

    te::vp::FeatureSet redFeatureSet;
    te::gtest::GetFeaturesFromShapefile(redDataInfo.m_uri, redFeatureSet);
    te::gtest::GetFeaturesFromShapefile(redDataInfo.m_uri, redFeatureSet);
    te::gtest::GetFeaturesFromShapefile(redDataInfo.m_uri, redFeatureSet);

    ASSERT_FALSE(redFeatureSet.empty());

    te::vp::FeatureSet blueFeatureSet;
    te::gtest::GetFeaturesFromShapefile(blueDataInfo.m_uri, blueFeatureSet);

    ASSERT_FALSE(blueFeatureSet.empty());

    std::vector<te::vp::FeatureSet> vecFeatureSets;
    vecFeatureSets.push_back(redFeatureSet);
    vecFeatureSets.push_back(blueFeatureSet);

    std::vector<te::vp::FeatureSet> vecOutputFeatureSets = te::vp::OperationService::SnapGeometries(vecFeatureSets);

    for (std::size_t i = 0; i < vecFeatureSets.size(); ++i)
    {
      ASSERT_FALSE(vecOutputFeatureSets.at(i).empty());

      te::common::FreeContents(vecFeatureSets.at(i));
      te::common::FreeContents(vecOutputFeatureSets.at(i));
    }
  }

  
}

