/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

 This file is part of the TerraLib - a Framework for building GIS enabled applications.

 TerraLib is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, either version 3 of the License,
 or (at your option) any later version.

 TerraLib is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with TerraLib. See COPYING. If not, write to
 TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
 \file terralib/itest/raster/ITsValidation.cpp

 \brief A integration test suit for the TerraLib Vector Processing Module.

 \author Mario Rocco Pettinati
 */

// TerraLib
#include <terralib_buildconfig.h>
#include <terralib/core/filesystem/FileSystem.h>
#include <terralib/core/utils/Platform.h>
#include <terralib/common.h>
#include <terralib/dataaccess.h>
#include <terralib/dataaccess/utils/DataAccess.h>
#include <terralib/core/plugin.h>
#include <terralib/geometry/CommonDataStructures.h>
#include <terralib/geometry/CoordinateSnapper.h>
#include <terralib/geometry/GEOSReader.h>
#include <terralib/geometry/GEOSWriter.h>
#include <terralib/geometry/GeometryCollection.h>
#include <terralib/geometry/GeometryFunctions.h>
#include <terralib/geometry/GeometryProperty.h>
#include <terralib/geometry/GEOSWriter.h>
#include <terralib/geometry/InternalConfig.h>
#include <terralib/geometry/Utils.h>
#include <terralib/geometry/Validation.h>
#include <terralib/geometry/validation/AddProjectedCoordinates.h>
#include <terralib/geometry/validation/FixNoding.h>
#include <terralib/geometry/validation/RemoveCollapsedSegments.h>
#include <terralib/geometry/WKTReader.h>
#include <terralib/sam/kdtree.h>
#include <terralib/memory/DataSet.h>
#include <terralib/memory/DataSetItem.h>
#include <terralib/srs/Converter.h>
#include <terralib/srs/SpatialReferenceSystemManager.h>
#include <terralib/vp/CommonFunctions.h>
#include <terralib/vp/GeometrySnapper.h>
#include "TestUtils.h"

// STL
#include <cmath>
#include <fstream>
#include <iostream>
#include <locale>
#include <memory>
#include <map>
#include <string>


//BOOST
#include <boost/bimap.hpp>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/connected_components.hpp>

#ifdef TERRALIB_GEOS_ENABLED

#include <geos/geom/Geometry.h>
#include <geos/operation/overlay/snap/GeometrySnapper.h>
#include <geos/simplify/TopologyPreservingSimplifier.h>
#include <geos/util/GEOSException.h>

#if GEOS_FULLVERSION >= 30800
  #include <geos/operation/valid/MakeValid.h>
#endif
#endif

// GTest and GMock
#include <gmock/gmock.h>
#include <gtest/gtest.h>

namespace
{
  class ITsValidation : public ::testing::Test
  {
    virtual void SetUp()
    {
    }

    virtual void TearDown()
    {
    }
  };

  te::gtest::DataInfo getOutputDataInfo(const std::string& fileName)
  {
    std::string outputPath = te::core::FileSystem::tempDirectoryPath() + "/terralib/vp/test_results";
    outputPath = te::core::FileSystem::adjustPathToGenericForm(outputPath);
    outputPath += "/" + fileName;

    te::gtest::DataInfo outputFile(outputPath);
    return outputFile;
  }

  TEST_F(ITsValidation, AddProjectedCoords)
  {
    std::string wkt = "multipolygon(((-64.1451981793532298 3.55296681855199648, -64.1451981793532298 3.57002712739561545, -64.1450918067335323 3.57002712739561545, -64.1450918067335323 3.56991904724431652, -64.1450205926935695 3.56987987952233476, -64.1449340157781194 3.56983226221883632, -64.1448969415567518 3.56983226221883632, -64.1448969415567518 3.56981187139708744, -64.1442254151352813 3.56944253186527805, -64.1439226156728637 3.56944253186527805, -64.1438738370612214 3.56936936394781235, -64.1432828936233506 3.56936936394781235, -64.1432828936233506 3.56924666279269376, -64.1434622515854613 3.56924666279269376, -64.1434622515854613 3.56906730483057677, -64.1436416095475863 3.56888794686845978, -64.1438209675096971 3.56870858890634279, -64.1440003254718079 3.56852923094422581, -64.1441796834339328 3.56834987298210882, -64.1443590413960436 3.56817051501999138, -64.1445383993581686 3.5679911570578744, -64.1447177573202794 3.56781179909575741, -64.1448971152824043 3.56781179909575741, -64.1448971152824043 3.56772508189847093, -64.1450179233212907 3.5677249309444643, -64.1447478551621373 3.56772526840438031, -64.1447475193541408 3.5674539140783379, -64.144477451248477 3.56745425143296746, -64.1444771155453566 3.56718289707956604, -64.1444767798679436 3.56691154272453037, -64.1444764442162381 3.5666401883678609, -64.1442063763223587 3.56664052556578071, -64.1442060407755008 3.56636917118175845, -64.1442057052543362 3.56609781679610194, -64.1439356374930867 3.56609815386305984, -64.143665569706144 3.56609849085049202, -64.143395501893508 3.56609882775839937, -64.1431254340552215 3.56609916458678056, -64.1428553661912417 3.56609950133563647, -64.1426822505643059 3.56609971714330687, -64.1426822656591042 3.56609971712448992, -64.1426822665188041 3.56634902966981304, -64.1426822415900517 3.56609971715449436, -64.1425852983016114 3.56609983800496666, -64.1426559905859506 3.5660997498792586, -64.1410292575303345 3.56772653716020249, -64.1425654617748791 3.56619028170800689, -64.1425654617748791 3.56619757743670363, -64.1425654617748791 3.56637693539882061, -64.1425654617748791 3.56655629336093805, -64.1425654617748791 3.56673565132305503, -64.1425654617748791 3.56691500928517202, -64.1425654617748791 3.56709436724728901, -64.1423861038127541 3.567273725209406, -64.1423861038127541 3.56745308317152343, -64.1423861038127541 3.56763244113364042, -64.1423861038127541 3.56781179909575741, -64.1423861038127541 3.5679911570578744, -64.1422067458506433 3.56817051501999138, -64.1420273878885325 3.56834987298210882, -64.1420273878885325 3.56852923094422581, -64.1418480299264075 3.56852923094422581, -64.1418480299264075 3.56870858890634279, -64.1416686719642968 3.56870858890634279, -64.1414893140021718 3.56888794686845978, -64.141309956040061 3.56906730483057677, -64.14113061601374 3.56906732276637317, -64.141029229980191 3.5687631697348996, -64.141029229980191 3.56882851560236425, -64.141029229980191 3.56910401710044223, -64.1413047314782716 3.56937951859852021, -64.1413047314782716 3.56973836072058504, -64.1412973950113638 3.56974202895403891, -64.1409247300051391 3.5699283614571522, -64.1405520649989143 3.57011469396026548, -64.1403657324957948 3.57030102646337877, -64.1401793999926895 3.57030102646337877, -64.1401793999926895 3.57119631845628938, -64.1409996380211709 3.57119631845628938, -64.1409996380211709 3.57139118363306851, -64.1417790987282928 3.57139118363306851, -64.1419739639050732 3.57158604880984765, -64.1421688290818537 3.57158604880984765, -64.1421688290818537 3.57178091398662678, -64.1423636942586342 3.57178091398662678, -64.1423636942586342 3.57179856698692122, -64.1423636942586342 3.57197577916340592, -64.1425585594354004 3.57197577916340592, -64.1425585594354004 3.57217064434018461, -64.1427534246121809 3.57217064434018461, -64.1427534246121809 3.57236550951696374, -64.1431331824776265 3.57236550951696374, -64.1439226156728637 3.57236550951696374, -64.1439226156728637 3.57253701650073907, -64.1444650475642959 3.57253701650073907, -64.1444650475642959 3.57256037469374288, -64.1450918067335323 3.57256037469374288, -64.1450918067335323 3.57272334900385236, -64.1451981793532298 3.57272334900385236, -64.1451981793532298 3.58362832807414966, -64.1412566261937798 3.58362832807414966, -64.141256427765839 3.58346821562181761, -64.1412566261937656 3.58362832807414966, -64.1409074366366809 3.58362832807414966, -64.1390851976311609 3.58319957415217605, -64.1388254332019585 3.58319989845089504, -64.1385553599296827 3.58320023553958356, -64.1385550244855267 3.58292888071676607, -64.1385552049578962 3.58307487237782496, -64.1377232395534094 3.58287911955448868, -64.13697398108485 3.58362832807414966, -64.1006565550055711 3.58362832807414966, -64.1006565550055711 3.58026179905184838, -64.1007412861745252 3.58026169691702023, -64.1007409621365554 3.57999033855325699, -64.1007403141347254 3.57944762182081755, -64.1010103898147037 3.57944729626604907, -64.1010100657714332 3.57917593792211441, -64.1010097417528755 3.57890457957654284, -64.1012798172492353 3.57890425399143997, -64.10127949317598 3.57863289566898057, -64.1012792875608426 3.57846071350074002, -64.1012791691274515 3.5783615373448816, -64.1018193197282358 3.57836088603425972, -64.1020893949915234 3.57836056025924254, -64.1020890707294484 3.57808920200778591, -64.1023591458885704 3.57808887617773674, -64.1026292210229229 3.57808855026788963, -64.1026288966267686 3.57781719206434889, -64.1026279235868799 3.5770031174439052, -64.102627599289761 3.57673175923381681, -64.1023575245524171 3.57673208501977324, -64.1018173750035061 3.57673273635238065, -64.1018170509693164 3.57646137806633924, -64.101816726959882 3.57619001977866269, -64.1015466523070785 3.57619034527579505, -64.1015437380496849 3.57374812039024459, -64.1020838859022319 3.57374746976196089, -64.1023539597913867 3.57374714432826757, -64.1023536358707418 3.57347578607377026, -64.1023523404356723 3.57239035303944208, -64.1042428541905025 3.5723880736398379, -64.1045129274849046 3.57238774769264156, -64.1045126030537062 3.57211638962836897, -64.1045113055769349 3.57103095735492948, -64.1045109812697547 3.57075959928248521, -64.1042409084510041 3.57075992508075846, -64.1034306898459647 3.57076090199777285, -64.103160616928065 3.57076122747751068, -64.1031609408389471 3.57103258567399706, -64.1028908678169955 3.57103291109888854, -64.1026207947703028 3.57103323644414283, -64.1026211185474182 3.57130459468857042, -64.1026220900273671 3.57211866941204859, -64.1006565550055711 3.57212103502532541, -64.1006565550055711 3.57049288263196241, -64.1007296363799952 3.5704927947905909, -64.1009997094413819 3.57049247005230397, -64.1009993862140277 3.57022111165439737, -64.1012694591714478 3.57022078686122901, -64.101269179589579 3.56998610966720209, -64.1006565550055711 3.56998610966720209, -64.1006565550055711 3.55296681855199648, -64.1451981793532298 3.55296681855199648), (-64.1018091105973724 3.56980532454323507, -64.1017681342339642 3.56980532454323507, -64.1016244577027834 3.56994900107442081, -64.1018092815717466 3.56994877871074889, -64.1018091105973724 3.56980532454323507), (-64.1090338280357628 3.57143239065893736, -64.1096431754630629 3.57143239065893736, -64.1096430117577967 3.57129610918266405, -64.1100402282328332 3.57129562723856964, -64.1100842499364347 3.57125160553497079, -64.1100842499364347 3.57052846503910271, -64.110265035060408 3.57034767991513569, -64.110445820184367 3.57016689479116911, -64.110992018472615 3.57016689479116911, -64.1109907647325628 3.56912361134993583, -64.1109904386977973 3.56885225386356719, -64.1107203670306234 3.56885258139828299, -64.1104502953385094 3.568852908853406, -64.1104499694871066 3.56858155131547772, -64.1101798978492781 3.56858187866605814, -64.1101792463797011 3.56803916353539385, -64.1101789206823014 3.56776780596761256, -64.1099088492571241 3.56776813316376451, -64.1096387778070351 3.56776846028034988, -64.1096384522929412 3.56749710266104714, -64.1093683808971093 3.56749742967312944, -64.1093680554871241 3.56722607202725728, -64.1093677301020364 3.56695471437975442, -64.1090976588396444 3.56695504126241891, -64.1090973335586511 3.56668368358835153, -64.1090970083025411 3.56641232591265389, -64.1090338280357628 3.56641240235310386, -64.108826937173589 3.56641265266592233, -64.1088262868944128 3.5658699372597833, -64.1088259617921636 3.56559857955426596, -64.1085558908757633 3.56559890615324626, -64.1085555658775661 3.56532754842117949, -64.1085552409042521 3.56505619068748114, -64.1082851701212206 3.56505651715712002, -64.1082845204075369 3.5645138016350093, -64.1082841955880269 3.56424244387150679, -64.108014125017462 3.56424277018692326, -64.1080134756113011 3.56370005460521488, -64.1080131509455384 3.56342869681191265, -64.1077430805874116 3.5634290229731449, -64.1061226579164156 3.5634309782716751, -64.1058525873842768 3.56343130387661944, -64.1058529114172728 3.56370266186894291, -64.104772628724092 3.56370396359338404, -64.1045025579887664 3.56370428882580503, -64.1045028816510722 3.56397564694068869, -64.1035759854736256 3.56397676255716167, -64.1035759854736256 3.56777578262268813, -64.1047774874502352 3.56777433477297024, -64.1047778115637641 3.56804569283855288, -64.1047781357021194 3.56831705090250439, -64.105048207758216 3.56831672516845755, -64.1050485320005805 3.56858808320594356, -64.1053186041110479 3.56858775736747935, -64.1063988923046679 3.56858645321776446, -64.10640149030624 3.57075731646482586, -64.1064018151682404 3.5710286743633568, -64.1066718878676625 3.57102834790314416, -64.1066722128337716 3.57129970577517142, -64.1066725378247355 3.5715710636455662, -64.1069426106578533 3.57157073705596373, -64.1069448868455112 3.57347024192882756, -64.1061346666064225 3.57347122198093059";
    wkt += ", -64.105864593143707 3.57347154850557658, -64.1058649180955342 3.5737429064374826, -64.1058658930999883 3.57455698022338808, -64.1055958192951181 3.57455730676774319, -64.1055961442669258 3.57482866471795724, -64.1055980946191681 3.57645681238488944, -64.1055984197647746 3.57672817032365176, -64.1058684942045716 3.57672784358048634, -64.1058688194544004 3.57699920149276185, -64.1058691447290556 3.57727055940339911, -64.1061392193028041 3.57727023253075282, -64.1074895917988101 3.57726859697083288, -64.1074886146202658 3.57645452368171268, -64.1077586888067117 3.57645419640507223, -64.1077583630508911 3.5761828386636525, -64.1077557578998523 3.57401197667339598, -64.1088187540686647 3.57401068831169777, -64.1088187540686647 3.57161317578290438, -64.1089995391926379 3.57143239065893736, -64.1090338280357628 3.57143239065893736), (-64.1158563111628581 3.57264534341541351, -64.1158559833764627 3.57237398640068982, -64.1158556556151069 3.57210262938432965, -64.115855327878819 3.57183127236633435, -64.1158550001675991 3.57155991534670481, -64.1158546724814187 3.57128855832544101, -64.1158543448203062 3.57101720130254119, -64.1158540171842475 3.57074584427800845, -64.1158536895732425 3.57047448725183925, -64.1161237612407433 3.57047415805442236, -64.1161234335755523 3.57020280105170151, -64.1163935051387455 3.57020247179974559, -64.1163931774193685 3.56993111482047842, -64.1163928497250879 3.56965975783957656, -64.1163925220558468 3.5693884008570409, -64.1163921944117021 3.56911704387287054, -64.1163918667926254 3.56884568688706771, -64.1163915391986023 3.56857432989963064, -64.1163912116296473 3.56830297291056064, -64.1163908840857744 3.56803161591985729, -64.1161208131563427 3.56803194497111376, -64.1158507422018431 3.5680322739427992, -64.1155806712222898 3.5680326028349123, -64.1155803439410477 3.56776124576733089, -64.1155800166848735 3.56748988869811701, -64.1153099458386322 3.56749021746052053, -64.1150398749673514 3.56749054614336414, -64.1147698040710452 3.56749087474664828, -64.1144997331496853 3.5674912032703725, -64.1142296622033143 3.56749153171453681, -64.1139595912319038 3.56749186007914165, -64.1139599180129807 3.5677632172986784, -64.1139602448190686 3.56803457451658357, -64.1139605716501677 3.56830593173285493, -64.1139608985062637 3.56857728894749293, -64.1139612253874134 3.56884864616049757, -64.1139615522935458 3.56912000337186841, -64.1136914808219274 3.56912033180710964, -64.1134214093252979 3.56912066016275498, -64.1131513378036715 3.56912098843880354, -64.1131516644971668 3.56939234572362185, -64.112881592871318 3.56939267394508919, -64.1128819195105706 3.56966403125328746, -64.1126118477805136 3.56966435942016069, -64.1123417760254597 3.56966468750742516, -64.1120717042454231 3.56966501551508131, -64.1118016324404181 3.56966534344312825, -64.1118019587877228 3.56993670084969139, -64.1118022851587739 3.57020805722533474, -64.1112108414566251 3.57020877556755378, -64.1113497458042048 3.57034767991513569, -64.1118024531013475 3.57034767991513569, -64.1118026115572661 3.57047941565791449, -64.1118029379794763 3.57075077305957667, -64.1115328658325438 3.57075110100795223, -64.1112627936606287 3.57075142887669505, -64.1109927214637594 3.57075175666580513, -64.1107226492419642 3.57075208437528202, -64.1104525769952005 3.57075241200512572, -64.1104529030460952 3.57102376953001288, -64.1104532291219158 3.57129512705326668, -64.1104535552226764 3.57156648457488579, -64.1104538813483771 3.57183784209486976, -64.1104542074990178 3.57210919961322038, -64.1107242801421222 3.57210887185857651, -64.1107246063969853 3.5723802293503315, -64.1107249326767743 3.57265158684045137, -64.1107252589815175 3.57292294432893653, -64.1107255853112008 3.57319430181578612, -64.1107259116658525 3.57346565930100191, -64.1107262380454159 3.57373701678458078, -64.1107265644499194 3.57400837426652451, -64.1107268908793912 3.57427973174683267, -64.110727217333789 3.57455108922550524, -64.110727543813141 3.57482244670254223, -64.1107275263300522 3.57480791540661258, -64.1108042241763627 3.57480790773606083, -64.1109835506832866 3.57462858122913829, -64.1111628771902105 3.57462858122913829, -64.1113422036971343 3.57462858122913829, -64.1113422036971343 3.57444925472221531, -64.1115215302040582 3.57444925472221531, -64.1117008567109821 3.57444925472221531, -64.1117008567109821 3.57426992821529277, -64.1118801832179059 3.57426992821529277, -64.1120595097248298 3.57426992821529277, -64.1122388362317537 3.57409060170837023, -64.1124181627386633 3.57409060170837023, -64.1124181627386633 3.57391127520144725, -64.1125974892455872 3.57391127520144725, -64.1127768157525111 3.57373194869452471, -64.112956142259435 3.57355262218760217, -64.1131354687663588 3.57337329568067918, -64.1133147952732827 3.57337329568067918, -64.1133147952732827 3.57319396917375665, -64.1134941217802066 3.57319396917375665, -64.1136734482871304 3.57319396917375665, -64.1136734482871304 3.57301464266683411, -64.1138527747940543 3.57301464266683411, -64.1138527747940543 3.57283531615991112, -64.1140321013009782 3.57283531615991112, -64.1142114278079021 3.57265598965298858, -64.1143907543148259 3.57265598965298858, -64.1145700808217356 3.57247666314606604, -64.1147494073286595 3.57247666314606604, -64.1149287338355833 3.57247666314606604, -64.1152873868494311 3.57247666314606604, -64.1158253663702027 3.57247666314606604, -64.1158561074050937 3.57247666314606604, -64.1158563111628581 3.57264534341541351), (-64.1301576819611654 3.56150218999404267, -64.1300104044266703 3.5613549124595405, -64.1300104044266703 3.56117555449742307, -64.1300104044266703 3.56099619653530608, -64.1300104044266703 3.56081683857318909, -64.1300104044266703 3.5606374806110721, -64.1300104044266703 3.56045812264895511, -64.1300104044266703 3.56009940672472069, -64.1300104044266703 3.55992004876260371, -64.1300104044266703 3.55987423430453598, -64.1298844651548734 3.55987438928665645, -64.1296143978593847 3.55987472155473705, -64.1293443305385296 3.55987505374343005, -64.1290742631923081 3.559875385852735, -64.1290745936978084 3.56014674170645318, -64.1288045262472366 3.56014707376174266, -64.1288048566990625 3.5604184296392023, -64.128534789144112 3.560418761640463, -64.1285351195422635 3.56069011754165965, -64.1282650518829485 3.56069044948888047, -64.1279949841983097 3.56069078135669503, -64.1277249164883187 3.56069111314510378, -64.1274548487530183 3.56069144485410538, -64.1271847809924083 3.56069177648370161, -64.1269147132064603 3.56069210803389069, -64.1266446453952312 3.56069243950467351, -64.1263745775586926 3.56069277089605007, -64.1263742477927536 3.56042141479214402, -64.1261041800099463 3.56042174607880302, -64.1258341122018578 3.56042207728606108, -64.125833782619253 3.56015072112990927, -64.1255637148648958 3.56015105223246753, -64.1252936470852859 3.56015138325563152, -64.1250235792803807 3.5601517141994008, -64.1247535114502369 3.56015204506377536, -64.1244834435948263 3.5601523758487561, -64.124213375714163 3.56015270655434168, -64.1242137048226652 3.56042406286219837, -64.1239436368377369 3.56042439351365214, -64.1239481370993616 3.56042438800386174, -64.1239481370993616 3.5605634706600271, -64.1242236385974422 3.56083897215810508, -64.1242236385974422 3.56111447365618305, -64.1242236385974422 3.56138997515426103, -64.1242236385974422 3.56166547665233901, -64.1242236385974422 3.56194097815041655, -64.1242236385974422 3.56221647964849453, -64.1244991400955229 3.56221647964849453, -64.1244991400955229 3.56249198114657251, -64.1244991400955229 3.56276748264465049, -64.1244991400955229 3.56304298414272846, -64.1244991400955229 3.56331848564080644, -64.1247746415936035 3.56331848564080644, -64.1247746415936035 3.56359398713888442, -64.12505014309167 3.56359398713888442, -64.1252978282889359 3.56359401190492742, -64.1252976018478762 3.56340765762894751, -64.125297272145346 3.56313630144014182, -64.1255673407945181 3.56313597013875238, -64.1255670110381857 3.56286461397360865, -64.1258370795830075 3.56286428261805677, -64.1261071481025482 3.56286395118305066, -64.1263772165968078 3.56286361966858856, -64.1266472850657721 3.56286328807467179, -64.1269173535094268 3.56286295640130035, -64.1271874219278004 3.56286262464847381, -64.1274574903208503 3.56286229281619171, -64.1277275586885764 3.56286196090445539, -64.127997627030993 3.56286162891326397, -64.1279972965882479 3.56259027297440189, -64.1279969661708407 3.56231891703390868, -64.1282670343297667 3.56231858501397047, -64.128537102463369 3.56231825291458959, -64.1288071705716334 3.56231792073576603, -64.1288068399423565 3.56204656486972082, -64.1290769079462137 3.56204623263682674, -64.1293469759247046 3.56204590032449664, -64.1296170438778432 3.56204556793273008, -64.1296167130367536 3.56177421214118928, -64.1298867808854709 3.56177387969537618, -64.1298864499906927 3.56150252392759414, -64.1301565177349886 3.56150219142774782, -64.1301576819611654 3.56150218999404267), (-64.1404421776975084 3.58021297307141984, -64.1404418419783724 3.57994161840982539, -64.1404415062848443 3.57967026374659003, -64.1404411706169384 3.57939890908171598, -64.1404408349746404 3.57912755441520192, -64.1404404993579504 3.57885619974704738, -64.1401704275109239 3.57885653690513772, -64.1399003556383178 3.57885687398341723, -64.1396302837401038 3.57885721098188547, -64.1393602118162818 3.57885754790054289, -64.1390901398668944 3.5788578847393886, -64.1388200678919134 3.57885822149842348, -64.1388204030319855 3.57912957632025375, -64.138820738197623 3.579400931140444, -64.1388209455597149 3.579568801591547, -64.1388252179955742 3.57957307402740366, -64.1391007194936407 3.57984857552548164, -64.1391007194936407 3.58012407702355961, -64.1393762209917213 3.58012407702355961, -64.1393762209917213 3.5802143038506129, -64.1393618885668815 3.58021432173729082, -64.1396319608879253 3.58021398469058516, -64.1399020331833896 3.58021364756403848, -64.1401721054532459 3.58021331035764989, -64.1404421776975084 3.58021297307141984), (-64.1447498705499157 3.56935339432630583, -64.1450199391842943 3.56935305671200753, -64.1450196031428561 3.56908170242150424, -64.1450198775148976 3.569303258522313, -64.1449700169170995 3.56935311912011066, -64.1447498705499157 3.56935339432630583)))";
  
    std::unique_ptr<te::gm::Geometry> geometryA(te::gm::WKTReader::read(wkt.c_str()));

    double tolerance = te::gm::GetAppropriatePrecision(4674);
    bool wasChanged = false;

    te::gm::AbstractValidatorPtr validator(new te::gm::AddProjectedCoordinates());
    validator->setPrecision(tolerance);

    te::gm::GeometryPtr result(validator->validate(geometryA.get(), wasChanged));

    ASSERT_FALSE(result->isEmpty());

    std::string resultWKT = result->toString();
  }

  void RemoveEqualSegments(const std::string& wkt)
  {
    std::unique_ptr<te::gm::Geometry> geometryA(te::gm::WKTReader::read(wkt.c_str()));

    bool wasChanged = false;

    te::gm::RemoveCollapsedSegments validator;
    te::gm::GeometryPtr result = validator.validate(geometryA.get(), wasChanged);

    ASSERT_TRUE(wasChanged);

    if (result != nullptr)
    {
      std::string outputWkt = result->toString();
      ASSERT_NE(geometryA->getNPoints(), result->getNPoints());
    }
  }

  TEST_F(ITsValidation, RemoveEqualSegments)
  {
    std::string wkt0 = "LINESTRING(0 0, 5 0, 10 0, 15 0, 10 0, 5 0, 5 5)";
    std::string wkt1 = "LINESTRING (-64.14519817935323 3.552966818551996, -64.14519817935323 3.5700271273956155, -64.14509180673353 3.5700271273956155, -64.14509180673353 3.569919047244317, -64.14502059269357 3.5698798795223348, -64.14493401577812 3.5698322622188363, -64.14489694155675 3.5698322622188363, -64.14489694155675 3.5698118713970874, -64.14422541513528 3.569442531865278, -64.14392261567286 3.569442531865278, -64.14387383706122 3.5693693639478123, -64.14328289362335 3.5693693639478123, -64.14328289362335 3.5692466627926938, -64.14346225158546 3.5692466627926938, -64.14346225158546 3.5690673048305768, -64.14364160954759 3.5688879468684593, -64.1438209675097 3.5687085889063432, -64.14400032547181 3.568529230944226, -64.14417968343393 3.568349872982109, -64.14435904139604 3.568170515019992, -64.14453839935817 3.5679911570578744, -64.14471775732028 3.5678117990957574, -64.1448971152824 3.5678117990957574, -64.1448971152824 3.567725081898471, -64.14501792332129 3.5677249309444643, -64.1448971152824 3.567725081898471, -64.14474785516214 3.5677252684043803, -64.14474751935414 3.567453914078338, -64.14447745124848 3.5674542514329675, -64.14447711554536 3.567182897079566, -64.14447677986794 3.5669115427245304, -64.14447644421624 3.566640188367861, -64.14420637632236 3.5666405255657803, -64.1442060407755 3.566369171181759, -64.14420570525434 3.566097816796102, -64.14393563749309 3.56609815386306, -64.14366556970614 3.566098490850492, -64.14339550189351 3.5660988277583994, -64.14312543405522 3.566099164586781, -64.14285536619124 3.5660995013356365, -64.1426822656591 3.5660997171244904, -64.1426822505643 3.566099717143307, -64.1426822656591 3.5660997171244904, -64.1426822665188 3.566349029669813, -64.14268224159005 3.566099717154494, -64.14265599058595 3.5660997498792586, -64.14258529830161 3.566099838004966, -64.14265599058595 3.5660997498792586, -64.14256546177488 3.566190281708007, -64.14102925753033 3.567726537160202, -64.14256546177488 3.566190281708007, -64.14256546177488 3.566197577436703, -64.14256546177488 3.5663769353988206, -64.14256546177488 3.566556293360938, -64.14256546177488 3.5667356513230546, -64.14256546177488 3.566915009285172, -64.14256546177488 3.567094367247289, -64.14238610381275 3.567273725209406, -64.14238610381275 3.567453083171523, -64.14238610381275 3.5676324411336404, -64.14238610381275 3.5678117990957574, -64.14238610381275 3.5679911570578744, -64.14220674585064 3.568170515019992, -64.14202738788853 3.568349872982109, -64.14202738788853 3.568529230944226, -64.14184802992641 3.568529230944226, -64.14184802992641 3.5687085889063432, -64.1416686719643 3.5687085889063432, -64.14148931400217 3.5688879468684593, -64.14130995604006 3.5690673048305768, -64.14113061601374 3.569067322766373, -64.14102922998019 3.5687631697348996, -64.14102922998019 3.5688285156023642, -64.14102922998019 3.5691040171004422, -64.14130473147827 3.56937951859852, -64.14130473147827 3.569738360720585, -64.14129739501136 3.569742028954039, -64.14092473000514 3.569928361457152, -64.14055206499891 3.5701146939602655, -64.1403657324958 3.5703010264633788, -64.14017939999269 3.5703010264633788, -64.14017939999269 3.5711963184562894, -64.14099963802117 3.5711963184562894, -64.14099963802117 3.571391183633069, -64.14177909872829 3.571391183633069, -64.14197396390507 3.5715860488098476, -64.14216882908185 3.5715860488098476, -64.14216882908185 3.571780913986627, -64.14236369425863 3.571780913986627, -64.14236369425863 3.5717985669869208, -64.14236369425863 3.571975779163406, -64.1425585594354 3.571975779163406, -64.1425585594354 3.572170644340184, -64.14275342461218 3.572170644340184, -64.14275342461218 3.5723655095169637, -64.14313318247763 3.5723655095169637, -64.14392261567286 3.5723655095169637, -64.14392261567286 3.572537016500739, -64.1444650475643 3.572537016500739, -64.1444650475643 3.572560374693743, -64.14509180673353 3.572560374693743, -64.14509180673353 3.5727233490038524, -64.14519817935323 3.5727233490038524, -64.14519817935323 3.5836283280741497, -64.14125662619378 3.5836283280741497, -64.14125662619378 3.5836283280741497, -64.14125642776584 3.5834682156218176, -64.14125662619378 3.5836283280741497, -64.14125662619378 3.5836283280741497, -64.14090743663668 3.5836283280741497, -64.13908519763116 3.583199574152176, -64.13882543320196 3.5831998984508955, -64.13855535992968 3.5832002355395836, -64.1385552049579 3.5830748723778254, -64.13855502448553 3.582928880716766, -64.1385552049579 3.5830748723778254, -64.13772323955341 3.582879119554489, -64.13697398108485 3.5836283280741497, -64.10065655500557 3.5836283280741497, -64.10065655500557 3.5802617990518484, -64.10074128617453 3.5802616969170202, -64.10074096213656 3.579990338553257, -64.10074031413473 3.579447621820817, -64.1010103898147 3.579447296266049, -64.10101006577143 3.5791759379221144, -64.10100974175288 3.578904579576543, -64.10127981724924 3.57890425399144, -64.10127949317598 3.5786328956689806, -64.10127928756084 3.57846071350074, -64.10127916912745 3.5783615373448816, -64.10181931972824 3.5783608860342597, -64.10208939499152 3.578360560259242, -64.10208907072945 3.5780892020077855, -64.10235914588857 3.5780888761777367, -64.10262922102292 3.57808855026789, -64.10262889662677 3.577817192064349, -64.10262792358688 3.577003117443905, -64.10262759928976 3.576731759233817, -64.10235752455242 3.5767320850197732, -64.1018173750035 3.5767327363523806, -64.10181705096932 3.5764613780663392, -64.10181672695988 3.5761900197786622, -64.10154665230708 3.576190345275795, -64.10154373804968 3.5737481203902446, -64.10208388590223 3.573747469761961, -64.10235395979139 3.5737471443282676, -64.10235363587074 3.5734757860737703, -64.10235234043567 3.572390353039442, -64.1042428541905 3.572388073639838, -64.1045129274849 3.572387747692641, -64.1045126030537 3.5721163896283694, -64.10451130557693 3.5710309573549295, -64.10451098126975 3.570759599282485, -64.104240908451 3.5707599250807585, -64.10343068984596 3.570760901997773, -64.10316061692806 3.5707612274775102, -64.10316094083895 3.5710325856739966, -64.102890867817 3.5710329110988885, -64.1026207947703 3.571033236444143, -64.10262111854742 3.5713045946885704, -64.10262209002737 3.5721186694120486, -64.10065655500557 3.5721210350253254, -64.10065655500557 3.5704928826319624, -64.10072963638 3.570492794790591, -64.10099970944138 3.570492470052304, -64.10099938621403 3.5702211116543974, -64.10126945917145 3.570220786861229, -64.10126917958958 3.569986109667202, -64.10065655500557 3.569986109667202, -64.10065655500557 3.552966818551996, -64.14519817935323 3.552966818551996)";
    std::string wkt2 = "LINESTRING (-64.14044217769751 3.5802129730714203, -64.14044184197837 3.5799416184098254, -64.14044150628484 3.57967026374659, -64.14044117061694 3.579398909081716, -64.14044083497464 3.579127554415202, -64.14044049935795 3.578856199747048, -64.14017042751092 3.5788565369051377, -64.13990035563832 3.578856873983417, -64.1396302837401 3.5788572109818855, -64.13936021181628 3.578857547900543, -64.1390901398669 3.5788578847393886, -64.13882006789191 3.5788582214984235, -64.13882040303199 3.5791295763202533, -64.13882073819762 3.579400931140444, -64.13882094555971 3.579568801591547, -64.13882521799557 3.5795730740274037, -64.13910071949364 3.5798485755254816, -64.13910071949364 3.5801240770235596, -64.13937622099172 3.5801240770235596, -64.13937622099172 3.580214303850613, -64.13936188856688 3.580214321737291, -64.13937622099172 3.580214303850613, -64.13963196088793 3.580213984690585, -64.13990203318339 3.5802136475640385, -64.14017210545325 3.58021331035765, -64.14044217769751 3.5802129730714203)";
    
    RemoveEqualSegments(wkt0);
    RemoveEqualSegments(wkt1);
    RemoveEqualSegments(wkt2);
  }

  void MakeValid(const std::string& wkt, bool expectEmpty = false, bool expectRaiseOfArea = false)
  {
    te::gm::GeometryPtr geometry(te::gm::WKTReader::read(wkt.c_str()));

    //we first apply the validation method
    bool wasChanged = false;
    te::gm::GeometryPtr geometryFixed = te::gm::Validation::MakeValidWithSnap(geometry.get()).release();
    if (te::gm::IsNullOrEmpty(geometryFixed.get()) != expectEmpty)
    {
      int a = 0;
    }

    //the we check if the result is as expected (null or valid)
    ASSERT_EQ(te::gm::IsNullOrEmpty(geometryFixed.get()), expectEmpty);

    if (te::gm::IsNullOrEmpty(geometryFixed.get()) == false)
    {
      //we we confirm that the geometry has been fixed
      std::string resultSnap2WKT = geometryFixed->toString();
      ASSERT_TRUE(geometryFixed->isValid());

      //and finally we check it the resulting area is compatible with the original area
      double originalArea = te::gm::GetArea(geometry.get());
      double fixedArea = te::gm::GetArea(geometryFixed.get());
      double tolerance = te::gm::GetAppropriatePrecision(geometry->getSRID());

      if (expectRaiseOfArea == true)
      {
        ASSERT_GT(fixedArea, originalArea);
      }
      else
      {
        double areaDelta = std::fabs(originalArea - fixedArea);
        if (areaDelta >= tolerance)
        {
          int a = 0;
        }
        ASSERT_LT(areaDelta, tolerance);
      }
    }    
  }

  TEST_F(ITsValidation, MakeValid)
  {
    std::string wkt011 = "MULTIPOLYGON (((-66.59839682922409 -1.8130360313269605, -66.59839682922404 -1.8135693092002314, -66.59893482922402 -1.8135693092001992, -66.59893482922398 -1.8141073092001812, -66.59947282922391 -1.8141073092001494, -66.59947282922388 -1.8146453092001307, -66.59866582922398 -1.814645309200178, -66.59866582922396 -1.814862020117877, -66.59636279898247 -1.8131731312741173, -66.59457104557704 -1.8128147805930317, -66.59457104557704 -1.8119189038903183, -66.59446005668178 -1.811863409442692, -66.59424757811581 -1.8119848257661049, -66.59419698798105 -1.8117318750923264, -66.59170424012835 -1.8104855011659766, -66.59128356514455 -1.8100648261821672, -66.59109954277707 -1.8100648261821672, -66.59055720655952 -1.8097032687037986, -66.59047197337912 -1.809589624463263, -66.59027083740402 -1.809589624463263, -66.58945368044812 -1.8089767567463326, -66.58947429102079 -1.8089748830579075, -66.58945368044812 -1.8089767567463326, -66.58868868233161 -1.8090463020296526, -66.58887696346449 -1.8085442190086187, -66.58883743467968 -1.8085145724200067, -66.58704568127425 -1.8079770463983786, -66.5843580511661 -1.8056477669713233, -66.5828057899743 -1.8046599643947228, -66.58260294203542 -1.8048222427458258, -66.58229760972803 -1.8044151330026417, -66.58229760978627 -1.804614202966421, -66.58229637239212 -1.804614202966788, -66.58229695397277 -1.8046250379095632, -66.58228787025153 -1.8046548844007442, -66.5822691262409 -1.8046798241137614, -66.5822429827452 -1.8046968489511541, -66.58221259305586 -1.8047039054662743, -66.58218162261791 -1.8047001425392535, -66.58060494458562 -1.8042202840082044, -66.58060527605532 -1.8042191948941515, -66.58059624031277 -1.8042171364412625, -66.58034097797032 -1.8041096575603348, -66.5803414655405 -1.8041084995820276, -66.58033092198103 -1.8041046620451093, -66.58031417426238 -1.8040906090539621, -66.58031387849215 -1.8040906090540503, -66.58031387849208 -1.8040903608734284, -66.58031344314466 -1.8040899955739296, -66.58030943749829 -1.8040904992138604, -66.58027901866014 -1.804083569429741, -66.57989480520067 -1.8039217953417044, -66.5796584403708 -1.8038222733081783, -66.57965858467637 -1.8038219305826997, -66.57965634830349 -1.8038213609788398, -66.5796305722746 -1.8038037847514108, -66.57961236215642 -1.803778452582206, -66.57960391435802 -1.8037484199046925, -66.57959134902579 -1.803622766582253, -66.57958458022924 -1.8035550786166068, -66.57957393294289 -1.8034486057530574, -66.57935644444042 -1.803448314619341, -66.57935662972972 -1.8033081621450575, -66.57954718490384 -1.8031811253623053, -66.57957393294289 -1.8034486057530574, -66.57958458022924 -1.8035550786166068, -66.57957393294289 -1.8034486057530574, -66.57954718490384 -1.8031811253623053, -66.57953844701291 -1.8030937464527748, -66.57953936594286 -1.8030936545598402, -66.57953936593572 -1.8030692088088787, -66.57955003633984 -1.8030398920981674, -66.57957009014902 -1.8030159928932636, -66.57959710857787 -1.803000393790995, -66.57962783280514 -1.8029949762733113, -66.57982625541344 -1.8029949762145239, -66.57982625541347 -1.8029950783558917, -66.57987866765254 -1.802960136863183, -66.579867529887 -1.8030046879253632, -66.5798608060445 -1.8030018863276478, -66.579867529887 -1.8030046879253632, -66.5819439164299 -1.803869847957365, -66.58194405797151 -1.8038693642338033, -66.58196253206793 -1.803844223928034, -66.58198849053403 -1.8038269182790525, -66.58201880239568 -1.8038195346035284, -66.58204981159501 -1.8038229634817011, -66.58207777796494 -1.8038367913402589, -66.58209932834853 -1.8038593503353804, -66.58214570905177 -1.8039289213212915, -66.58224138455705 -1.8037375703107208, -66.5827948292259 -1.8034608479762992, -66.58279482922589 -1.803616309201491, -66.58314472179063 -1.8036163092014703, -66.58333282922587 -1.8036163092014592, -66.58333282922587 -1.803867119115112, -66.58333282922585 -1.8041543092014403, -66.58343785896591 -1.804154309201434, -66.5838708292258 -1.804154309201408, -66.58387082922577 -1.8046923092013905, -66.58440882922571 -1.8046923092013591, -66.58440882922568 -1.8052303092013406, -66.5854848292256 -1.8052303092012767, -66.58548482922559 -1.805768309201259, -66.58656082922546 -1.8057683092011954, -66.58656082922543 -1.8063063092011773, -66.58709882922541 -1.806306309201145, -66.58709882922537 -1.8068443092011268, -66.58817482922528 -1.806844309201063, -66.58817482922525 -1.807616437730428, -66.58817482922525 -1.8076513092010358, -66.58823759787234 -1.807651309201032, -66.58925082922514 -1.807651309200973, -66.58925082922511 -1.808189309200954, -66.58978882922506 -1.8081893092009234, -66.58978882922504 -1.8087273092009035, -66.59086482922496 -1.8087273092008402, -66.59086482922493 -1.809121813560888, -66.59086482922491 -1.8092653092008233, -66.591273503968 -1.809265309200799, -66.59194082922484 -1.8096657043548938, -66.59194082922481 -1.8097658759519788, -66.59194082922481 -1.8098033092007402, -66.59247882922476 -1.8098033092007086, -66.5924788292247 -1.8103413092006904, -66.5940928292246 -1.8103413092005944, -66.59409282922456 -1.8108793092005773, -66.5946308292245 -1.8108793092005457, -66.59463082922449 -1.8111766717346622, -66.59463082922449 -1.811417309200527, -66.59495167917895 -1.811417309200508, -66.59570682922438 -1.811417309200463, -66.59570682922435 -1.8119553092004457, -66.59624482922432 -1.8119553092004135, -66.59624482922429 -1.8124933092003948, -66.59732082922419 -1.8124933092003308, -66.59732082922416 -1.8130313092003139, -66.59838974599467 -1.81303130920025, -66.59839682922409 -1.8130360313269605)))";

    std::string wkt010 = "polygon((-41.953125 -20.7499611377716064,-41.953125 -20.7416633772931576,-41.9524337242890084 -20.7416633772931576,-41.9524044875611253 -20.7416738067273165,-41.9523101971411734 -20.741755375165905,-41.952299656052368 -20.7418242470247058,-41.9522798345419687 -20.7418814133756761,-41.95219467830902 -20.7418770220736306,-41.9520358034401042 -20.7418313146224698,-41.9518322007519515 -20.7417674517228185,-41.9516514351100938 -20.7416934181958581,-41.9515851811921792 -20.7416633772931576,-41.9512254144794312 -20.7416633772931576,-41.9510906671868185 -20.741636427834635,-41.9510162556849266 -20.7415823103787105,-41.9510050027306463 -20.7415837356253547,-41.9508340887097262 -20.7416180419564995,-41.9506716068189007 -20.7416710094343202,-41.9505026903292944 -20.7417203795935663,-41.9504063824995583 -20.7417161184139722,-41.9503899812651966 -20.7417981245857774,-41.9502821834310993 -20.7419059224198676,-41.9501474361384865 -20.7419328718783937,-41.9501204866799497 -20.7417981245857774,-41.9499386625061561 -20.7415960977259957,-41.9498535524633809 -20.7415708502798317,-41.9497224431925346 -20.7415232168694885,-41.9496036212320504 -20.7414575194725046,-41.9495248881470602 -20.7413938827079214,-41.9483957213344496 -20.7413938827079214,-41.9481531762077324 -20.7413669332493988,-41.9480184289151197 -20.7412321859567825,-41.9480026066359031 -20.7410106740477751,-41.9479868201267081 -20.7410030828682324,-41.9479762151055482 -20.7409540391356231,-41.9479949703910364 -20.7409037666196063,-41.9479914794565971 -20.740854893537449,-41.9479645299980604 -20.7407201462448327,-41.9478297827054547 -20.7406123484107354,-41.9474524902861248 -20.7405853989522129,-41.9474524902861248 -20.7404943614953865,-41.9473524081878466 -20.7404249270423726,-41.9472019042298356 -20.7403271715102626,-41.9470661648003897 -20.7402470255345435,-41.9469308743230229 -20.7401850030755739,-41.9468266550253475 -20.7401166865616169,-41.9467527341590127 -20.7400735380908081,-41.9466986867707519 -20.7400715101794368,-41.9466519921961236 -20.7401111635993693,-41.9466151040368942 -20.740158115150404,-41.9465559631929494 -20.7402026420273167,-41.9464408273338734 -20.7402219384429785,-41.9462964618323824 -20.7402322041509741,-41.9461415132291791 -20.7402372042663323,-41.9460050359837169 -20.740249545364442,-41.9458946277942459 -20.7402487193255034,-41.9457980111580753 -20.7402319735587142,-41.9457073969339547 -20.7401954488535019,-41.9456339172806594 -20.7401613680107353,-41.9455657222164433 -20.7401451020753065,-41.9455108289877074 -20.7401299473515763,-41.9454966351853074 -20.7400837576193133,-41.945523821400684 -20.7400173835418222,-41.9455266114359873 -20.7399605395047288,-41.9454741472439068 -20.7399367070221636,-41.9454077399362788 -20.7399506671428888,-41.9453408545922386 -20.7399723289176734,-41.9452709459944018 -20.7399817312493937,-41.945186317650915 -20.7399632649300258,-41.9450782326646916 -20.7399410051319784,-41.9449661214096068 -20.7398861036685425,-41.9448510816111408 -20.7398124421493577,-41.9447466884221143 -20.7397284891097122,-41.9446577874530462 -20.7396636359790563,-41.9445834583542378 -20.7395984330370133,-41.9444957560393448 -20.7395310014165268,-41.9444060968596872 -20.7394632758641073,-41.9443238356911863 -20.7393753503925211,-41.9442445293234059 -20.7392945998573097,-41.9441824390592402 -20.7392308060079955,-41.9441225746618898 -20.7391710024053815,-41.9440487426237212 -20.7391032652242195,-41.9439881236888752 -20.7390194779848471,-41.9439204525543445 -20.7389715406218897,-41.9438238115880395 -20.7389363664266462,-41.9436950560399495 -20.7388619885273258,-41.9435753835463956 -20.7387797850959252,-41.9435083690274482 -20.7386867488287621,-41.9434554555374746 -20.738575597253643,-41.9434700839699488 -20.7384589908041903,-41.9435220715421337 -20.7384038183281554,-41.9435405744826326 -20.738417568581248,-41.9435446131072496 -20.7384533324730853,-41.9435944465197892 -20.7384972701543795,-41.9437042664422606 -20.7385110999911362,-41.943798255438935 -20.7384831302599935,-41.9438606440100585 -20.738421236892659,-41.9438782879377925 -20.7383375450630538,-41.9438604265276851 -20.7382424857725347,-41.9438364248955082 -20.7381402151228649,-41.9438109216146628 -20.7380444803177042,-41.9437898096338628 -20.7379740160020454,-41.9438215251492252 -20.7379081618772076,-41.9438707360196261 -20.7378404773556042,-41.9438953925304787 -20.7377433050750888,-41.9438473619534946 -20.7376128587606381,-41.9437642599179341 -20.7374244685609312,-41.9437085025503578 -20.7372412158689059,-41.9436691118635423 -20.7370523446240291,-41.9436602431223875 -20.7369006466908701,-41.943655571798601 -20.7367738297661361,-41.9436751455411354 -20.7366563976967626,-41.9437311279058136 -20.7365694680386525,-41.9438314652532753 -20.7365313959275248,-41.9440189442764364 -20.7364954352823929,-41.9440814402546067 -20.7364449510946969,-41.9441288224577988 -20.7364079354703144,-41.9441815327598277 -20.7363614546653814,-41.9442481422009905 -20.7363167765966097,-41.9443329072288265 -20.7362859997537292,-41.9444730399135395 -20.736297938716465,-41.944688176622293 -20.7363887644517249,-41.9448722046862699 -20.7364787103834267,-41.9450329172835268 -20.7365012680982872,-41.9451606851434491 -20.7364452123279648,-41.9452726755860681 -20.7363873484152208,-41.9453807650291921 -20.7363330524884084,-41.945488968602838 -20.7362867324712319,-41.9455838110994108 -20.7362644023192111,-41.9456121115774891 -20.7363151872670848,-41.9455500925240372 -20.7364405737746402,-41.9454213506969182 -20.7365963135627673,-41.9452282095513809 -20.7367224100965473,-41.9449785722957458 -20.7368248113741842,-41.9447004227772879 -20.7369058358661924,-41.944447343838263 -20.7369994546095278,-41.944274962672182 -20.7371179213642947,-41.9442436874258746 -20.7372366131873207,-41.9443699436201385 -20.7372653440409778,-41.9445905461643775 -20.7371962346621181,-41.9448514528572147 -20.7371198136547328,-41.945118510436231 -20.7370792644865993,-41.9453644261203493 -20.7370517518317996,-41.9455779900550212 -20.7370192963426199,-41.9457856241397238 -20.7370071432633694,-41.9459757346227349 -20.7369681805627906,-41.9460908069312453 -20.7369081418390238,-41.9461203894427044 -20.7368509306656321,-41.9460756233984497 -20.7368211429289282,-41.9460264331468196 -20.7367810894616937,-41.946032837407671 -20.7367124654018831,-41.9461100911472968 -20.7366274401482151,-41.94624587141071 -20.736549640488473,-41.9464560197064955 -20.7365052187367596,-41.9467154445952559 -20.7364817032188,-41.946985053180299 -20.7364793279886577,-41.9472291643852913 -20.7365227008380089,-41.9474730923195409 -20.7365725240593797,-41.9476640925140103 -20.7366148609042966,-41.9478194368283255 -20.7366613920604799,-41.9479629380459258 -20.7367145366013403,-41.9480906383190089 -20.7367752995790511,-41.9482170194532458 -20.7368371688992319,-41.9483484184324809 -20.7369135948282839,-41.9484614149673618 -20.7369857985370025,-41.9485939684129505 -20.7370390212217153,-41.9487612701300634 -20.7370846356399987,-41.9489854932219117 -20.7371126665924912,-41.949211092989799 -20.7371022941050853,-41.949403566867467 -20.7370629672224354,-41.9495308879361417 -20.7370449832924884,-41.9496160118090344 -20.7370543085114143,-41.9497062216244458 -20.7370565380919167,-41.9497856029821818 -20.7370265691188891,-41.9498149386122776 -20.7369809457240386,-41.9498202076023787 -20.7369411025403636,-41.9498571508808311 -20.7369062735059657,-41.9499351670769514 -20.7368743458293672,-41.9500128779678079 -20.7368511534651674,-41.950074112625245 -20.7368491097550951,-41.9500802457006046 -20.7368759675254388,-41.9500628807298739 -20.7369125099121518,-41.9500517881106916 -20.7369453990480359,-41.9500758203764263 -20.7369641414445347,-41.9501513522283034 -20.7369593880325844,-41.9502282950196701 -20.7369470633359043,-41.9502859325038173 -20.7369805671929406,-41.950301732165002 -20.7370867799252103,-41.9502566641137875 -20.7372183591434442,-41.9501760482728869 -20.7373471713469151,-41.9500794967877866 -20.737460445754369,-41.9500104603690218 -20.737557132896427,-41.9499535755984283 -20.7376613097283915,-41.9498903425786693 -20.7377959152141571,-41.9498325006474033 -20.7379614852672489,-41.9497988200881053 -20.7381646683368182,-41.9498436841019711 -20.7383963799205411,-41.9499996216490203 -20.7386078673037098,-41.9501474361384865 -20.7387164422904604,-41.9501474361384865 -20.7386989368555632,-41.9505247285578236 -20.7386719873970371,-41.9506594758504292 -20.7385372401044208,-41.9506864253089589 -20.7381599476850909,-41.9509559198941915 -20.7381599476850909,-41.9513601617720369 -20.7381599476850909,-41.951602706898754 -20.7381868971436134,-41.9517374541913597 -20.7383216444362333,-41.9517913531084119 -20.738806734689657,-41.9519261004010247 -20.7389414819822733,-41.9521416960692122 -20.7389414819822768,-41.9524381401129673 -20.7387258863140858,-41.9525728874056085 -20.7386989368555632,-41.9529501798249385 -20.7387258863140858,-41.9530849271175512 -20.7388336841481831,-41.9531118765760809 -20.7389684314407994,-41.953125 -20.7391521593756742,-41.953125 -20.7300000000000004,-41.9389133703759498 -20.7300000000000004,-41.9375 -20.7300000000000004,-41.9375 -20.7313847079211584,-41.9377115256449997 -20.7311708581369736,-41.9379144094720928 -20.7309802086842758,-41.938129763680621 -20.7308077127228074,-41.9383330317239 -20.7306218061469707,-41.9384940748289381 -20.7304552925919339,-41.9386375247091507 -20.7303134349638327,-41.9387633659239398 -20.7301791258990349,-41.9389133703759498 -20.7300000000000004,-41.9400781034231329 -20.7300901392814048,-41.9404434419129686 -20.7302099581854833,-41.9406107350025437 -20.7303850832858885,-41.9406459988541656 -20.7305601068660792,-41.9406082528660917 -20.7307294193028255,-41.9405197470217743 -20.7308429545407975";
    wkt010 += ",-41.9403751559380069 -20.7309051737597692,-41.9401723921359846 -20.7309068416022804,-41.9399284428062771 -20.7308867102537739,-41.9396482308727201 -20.7309144340512184,-41.9393709025491006 -20.7310397353799267,-41.9391320706352886 -20.7312155800013436,-41.9389266369774845 -20.7313827575953624,-41.938744896874411 -20.7315436441749057,-41.9386128010688708 -20.7316593759654566,-41.9385173384854895 -20.7317652406309101,-41.9384135132785261 -20.7318797418279885,-41.9382811828152384 -20.7320149962475107,-41.9381433027757993 -20.7321253875099387,-41.9380135920910817 -20.7322257918950861,-41.9378776454576183 -20.7323410665017001,-41.9377628791904939 -20.7324781736288806,-41.9376567944820167 -20.7326004152103067,-41.9375570794433159 -20.7326965328509836,-41.9375 -20.732725528772459,-41.9375 -20.7401702537608514,-41.9375796543721862 -20.7401526558970097,-41.9377219398451118 -20.7401432039517459,-41.9378382758713784 -20.7401676325852122,-41.9379568999670553 -20.7402098010221678,-41.9380977024409844 -20.7402564770741371,-41.9382165125736321 -20.7403007145336318,-41.9383304891069244 -20.7403523020198968,-41.9384440262904192 -20.7404226464778638,-41.9385682797126051 -20.7405091283514977,-41.9387013985649375 -20.7406158763029467,-41.9388331523516698 -20.7407291109468837,-41.9389540594750585 -20.7408696215795061,-41.939060758391058 -20.7410447150998891,-41.9391538959792172 -20.7411820084252909,-41.9392391904526463 -20.7412818641166368,-41.9393355512432464 -20.7413609097989884,-41.9394521051045075 -20.7414180631599407,-41.9396023354031513 -20.7414397617487616,-41.9397878405695508 -20.7414277659847883,-41.9399746602604893 -20.7414232291446723,-41.9401470667045757 -20.7414227422270336,-41.9403153479529252 -20.7413991399921827,-41.9405153119181762 -20.7413634571900616,-41.9407275941831799 -20.7412957358550294,-41.9409488996187463 -20.7412258097712154,-41.9411406638116873 -20.7411582504648138,-41.9413320431423102 -20.7411014400213141,-41.9415035764137087 -20.7410507253779741,-41.9416685090577204 -20.7410039945744451,-41.9418078256992217 -20.7409634608481248,-41.941930507963697 -20.7409471712245512,-41.9420346363268166 -20.7409514824972234,-41.9421071553849529 -20.7409753418529021,-41.9421775211918515 -20.7410050719269954,-41.942268902224626 -20.7410311294119509,-41.94235450354698 -20.7410628837766282,-41.942428832515759 -20.7411009995347122,-41.9425179590135926 -20.7411158295817621,-41.9426197866953459 -20.7410881329513686,-41.9427410090788229 -20.741045213419202,-41.9428707573766744 -20.7409779820861253,-41.942992166922572 -20.7408882298854138,-41.9430905771303273 -20.7408487146001477,-41.9431584226308942 -20.7408829907753045,-41.9432051984282737 -20.7409626027906882,-41.9432704050933083 -20.7410512770624926,-41.9433152828102678 -20.7411519963265647,-41.9433151668749318 -20.7412669578142399,-41.9433083913572133 -20.7413640061793174,-41.9433035384399133 -20.7414412251333644,-41.9433004011778365 -20.7415139056057356,-41.9432897181724584 -20.7415991273361762,-41.9433000317260394 -20.7417047659949496,-41.9433297933576839 -20.7418356998989779,-41.9433924975031189 -20.7419935588621733,-41.9434870632454349 -20.7421737176347385,-41.9435873289552248 -20.7422967766447286,-41.9436737623248206 -20.7423495654687109,-41.9437399101572126 -20.7423046207509465,-41.9437892343150764 -20.7422391214652002,-41.9438248228005293 -20.7421700910315963,-41.9438557914346362 -20.7420958654392926,-41.9438676793555558 -20.7420153410943868,-41.9438844122829266 -20.7419173400524031,-41.9439238270784287 -20.7418023650084429,-41.9439565788723741 -20.7416713480872836,-41.943987535114168 -20.7415244372456158,-41.9440194338744874 -20.7413625122929552,-41.9440822578726298 -20.7412420467835723,-41.9441713165087009 -20.7412089921770075,-41.9442638930566147 -20.7412615742425572,-41.9443576436274981 -20.7413321711323029,-41.9444433385776279 -20.7413939686472304,-41.9445144372190413 -20.7414583143881757,-41.9445619502502751 -20.7415192393780075,-41.9445788259910017 -20.7415789841826808,-41.9445560773222752 -20.7416319983866231,-41.9445198806932993 -20.74167108390008,-41.9445006291620928 -20.7417359681122946,-41.9445163986972958 -20.7418196372369934,-41.9445561833319402 -20.7419049221335463,-41.9445616383757738 -20.7419943344873232,-41.9445143577866375 -20.7421091082070248,-41.9444558006454358 -20.7422516476068886,-41.9443802207358871 -20.7423712007644596,-41.9443054259675421 -20.7424457407216103,-41.9442408830290034 -20.7425001050304516,-41.9441814091740639 -20.7425478289241489,-41.9441209309621357 -20.7425982387169157,-41.9440796992564131 -20.7426384797060841,-41.9440704749263773 -20.7426822210807273,-41.9440961567806454 -20.7427399775545744,-41.9441151292521184 -20.7428114831126322,-41.944118479518778 -20.7428949940771403,-41.9440964485421333 -20.7430050056800823,-41.9440771398409922 -20.7431501376147658,-41.9440742594253493 -20.7433176711701783,-41.9440710104263843 -20.7434634733460506,-41.944098339358959 -20.743563522690998,-41.9441662677681535 -20.7436188803548767,-41.9442479814037057 -20.7436200710092251,-41.9443310573847228 -20.7436019056585366,-41.9444140183381577 -20.743588452778404,-41.9444763627005131 -20.7435894699562695,-41.9445132576936714 -20.7436312282423323,-41.9445172349598394 -20.7437339892557553,-41.9444939929018332 -20.7438481659848328,-41.9444472746644053 -20.7439724071359386,-41.9444560207680226 -20.7440834760007533,-41.944508029050688 -20.7441433444408574,-41.9445750364820213 -20.7441971720698319,-41.9446266691713348 -20.7442664960940526,-41.9446808870893051 -20.744325090426873,-41.9447591214937674 -20.7443831021709393,-41.944859709927421 -20.7444824796315608,-41.9449905549291557 -20.7445708869372112,-41.9451450185493115 -20.7446454772938615,-41.945359180548742 -20.7447089550880115,-41.945587195350484 -20.744798154336511,-41.9458366889774865 -20.744949683332063,-41.9460386724128824 -20.7450590623270124,-41.9461966961043586 -20.7451365900739617,-41.9462836240535424 -20.7452064078961094,-41.9463132487178996 -20.7452672636377322,-41.946343243021424 -20.7453101366117281,-41.9464056086703607 -20.7453302562293693,-41.9464694782099912 -20.7453690239496318,-41.9465310980175445 -20.7454482235647006,-41.9465977218322905 -20.7455558809777756,-41.9466276429613529 -20.7456450331197644,-41.9466111034656493 -20.7457064853029394,-41.9465748664363218 -20.7457650173398847,-41.9465305818931569 -20.7458374863914266,-41.9464699671737549 -20.7458921766672937,-41.9464047205591726 -20.7459221897152659,-41.946376615655879 -20.7459741034791065,-41.9463850598151211 -20.74607108473948,-41.9463668419646396 -20.7461735302801209,-41.9463123918491831 -20.7462822444084622,-41.9462229344409963 -20.7463800108039642,-41.9461139329845096 -20.7464512340286156,-41.945974961232416 -20.7464902860701415,-41.945819347141061 -20.7465175958671715,-41.945668736159277 -20.7465517745867487,-41.9455383642925881 -20.7466046651047407,-41.9454103283713877 -20.7466483873305627,-41.9452730959961002 -20.7466731372918751,-41.9451741778841125 -20.7467210697054192,-41.9450947909869427 -20.7467952328175187,-41.9449712532106673 -20.7468780968725213,-41.9448487025288017 -20.7469647961803325,-41.9447109259333573 -20.7470300296314392,-41.9445898717079828 -20.7470603135146234,-41.9444915314502822 -20.7470431417907299,-41.9444279492560028 -20.7469693555213048,-41.9443956319011448 -20.746882220721087,-41.944375409034663 -20.7468030343807506,-41.944351935351186 -20.7467375510580609,-41.9443170768121192 -20.7466932005273819,-41.9442700924591847 -20.7466309880716189,-41.9442286196057808 -20.7467237633740886,-41.9441348906930358 -20.7468220940920247,-41.944013227377333 -20.7469095013730396,-41.9438840580369927 -20.7469652382055827,-41.943764958045918 -20.7469639985420073,-41.9436909857360121 -20.7469430367619907,-41.9436833886290259 -20.7469061452891914,-41.9437095568507914 -20.7468647549607326,-41.943700957594352 -20.746829406191317,-41.9436512871538767 -20.7468148457084354,-41.9435764703171543 -20.7467995369303431,-41.9435004452287785 -20.7468122306155252,-41.9434378157321675 -20.7468895147122865,-41.9434122520964436 -20.7470081612323405,-41.9434285327295271 -20.7471174563676115,-41.9434590289750773 -20.7472166634985697,-41.9434852881391222 -20.7473095265594409,-41.943488942298238 -20.7474055965945325,-41.9434823805540091 -20.7474937582139098,-41.9434314032798738 -20.7475637166632154,-41.9433541814039756 -20.7476030972254541,-41.9432913103131142 -20.7476312782354775,-41.9432650523657031 -20.747698548687179,-41.9432820473481272 -20.7478037503830102,-41.9433138839772255 -20.7479156650433794,-41.9433661064840919 -20.7480228898191825,-41.9434634393630219 -20.7480948057913857,-41.9435621703408472 -20.7481248740711166,-41.9436560013144515 -20.7480961634063412,-41.9437404948867609 -20.7480224227673204,-41.9438179210825979 -20.7480067545259317,-41.943888499273335 -20.7480535815368121,-41.9439748464656219 -20.7481329782781359,-41.9440543830392087 -20.7482144137596478,-41.9441286143161776 -20.7482913548682717,-41.9442062428866578 -20.7483707979977261,-41.9442909825417232 -20.7484435861621783,-41.9443522318914219 -20.7485040864368102,-41.9443470727825485 -20.7485530706726209,-41.9443069374750976 -20.7485916471051404,-41.9442808967144742 -20.7486593313654346,-41.9442633267060145 -20.7487393710871899,-41.9442267443424939 -20.7488099236095849,-41.9441651940181757 -20.7488771594670389,-41.9440680631448757 -20.7489505382985335,-41.9439282062094847 -20.7490352204002235,-41.9437770069834812 -20.7491021838991792,-41.9436798448119603 -20.7491805546242283,-41.9436146937214289 -20.74926662977051,-41.9435254322468722 -20.7493646230302247,-41.9434399053869242 -20.7494871109277774,-41.9433819161072847 -20.7496065354955306,-41.9433881618466273 -20.7497065491870281,-41.9434817698403819 -20.7497882508618581,-41.9435954907426591 -20.7498366720580769,-41.9436657507007382 -20.7498749998211487,-41.9436671247637989 -20.74991374773251,-41.9436126816753259 -20.7499604376447273,-41.9436120165190331 -20.7499611377716064,-41.9479873165050847 -20.7499611377716064,-41.9479873165050847 -20.7498821916845344,-41.948210772426286 -20.7498821916845344,-41.9484342283474945 -20.7498821916845344,-41.9486576842686958 -20.7498821916845344,-41.9488811401899042 -20.7498821916845344,-41.9488811401899042 -20.7499611377716064,-41.9506462133406686 -20.7499611377716064,-41.9506222886907949 -20.7498911930895957,-41.9505856942795248 -20.7497630395758179,-41.9505551283095102 -20.7495944460785573,-41.9504930743378353 -20.7493559077412932,-41.950393750044455 -20.7490970938616854,-41.950251795638124 -20.7486896826235316,-41.9503370781133214 -20.7485247931656787,-41.9503829333621994 -20.7483837586798714,-41.950427114752074 -20.7482408736715911,-41.9504885487192212 -20.748100275784445,-41.9505770225641825 -20.7479780929024216,-41.9507178443222699 -20.7479360378813418,-41.950787915293148 -20.7479913283334518,-41.9507460471736877 -20.7481240215313711,-41.9506840273004613 -20.748245011783613,-41.9506711583149183 -20.7483428547434414,-41.950676061084188 -20.7484348567047832,-41.9506565158567994 -20.7485343429011344,-41.9506107454047168 -20.7486259437612865,-41.9505791765189571 -20.7487022551634226,-41.95059036823951 -20.7487742464600693,-41.9506757780440154 -20.7488535931907876,-41.9507853376393456 -20.7489340914325702,-41.9508979012064529 -20.7489989296786739,-41.9510158572530543 -20.7490655370988044,-41.9511674288283984 -20.7491316863986484,-41.9513088708992825 -20.749183072159429,-41.9514510060200791 -20.7491876652775531,-41.9515920493426435 -20.7491531156855729,-41.9517286285769799 -20.7491208683572879,-41.9518472338497261 -20.749099358151593,-41.9519657959139494 -20.7491226041631336,-41.9520794070504195 -20.749212686405837,-41.9521179246458473 -20.7493153398566612,-41.9521089303996035 -20.7494194578949553,-41.9520653792769238 -20.7495224415489297,-41.9520134407092584 -20.7496137171161905,-41.9519598120969661 -20.7496917482031655,-41.951904088336569 -20.7497640142572308,-41.9518396135185938 -20.7498173121838612,-41.9517552838304226 -20.7498704557767901,-41.9516393340567717 -20.7499351549064208,-41.9515515182415086 -20.7499611377716064,-41.953125 -20.7499611377716064))";
    std::string wkt09 = "multipolygon(((-60.0761253048836323 0.750689105289661263,-60.0761253048836323 0.750868245093666786,-60.0761253048836323 0.750941858502082837,-60.0761568501982666 0.750941858502082837,-60.0761568501982666 0.750897636501200338,-60.0763359683324794 0.750897636501200338,-60.076515086466685 0.750897636501200338,-60.0766942046008978 0.750897636501200338,-60.0768733227351106 0.750897636501200338,-60.0770524408693234 0.750897636501200338,-60.0772315590035291 0.750897636501200338,-60.0774106771377419 0.750897636501200338,-60.0774106771377419 0.751076754635411237,-60.0775897952719546 0.751076754635411237,-60.0775897952719546 0.751255872769622246,-60.0777689134061674 0.751255872769622246,-60.0779480315403731 0.751434990903833255,-60.0779480315403731 0.751614109038044265,-60.0781271496745859 0.751614109038044265,-60.0781271496745859 0.751793227172255274,-60.0783062678087987 0.751793227172255274,-60.0783062678087987 0.751972345306466172,-60.0784853859430044 0.751972345306466172,-60.0784853859430044 0.752151463440677182,-60.0786645040772171 0.752330581574888191,-60.0786645040772171 0.7525096997090992,-60.0788436222114299 0.7525096997090992,-60.0788436222114299 0.752688817843310209,-60.0790227403456427 0.752688817843310209,-60.0790227403456427 0.752867935977521219,-60.0792018584798484 0.753047054111732117,-60.0793809766140612 0.753047054111732117,-60.0793809766140612 0.753226172245943126,-60.079560094748274 0.753226172245943126,-60.079560094748274 0.753405290380154136,-60.0797392128824868 0.753405290380154136,-60.0797392128824868 0.753584408514365145,-60.0799183310166924 0.753763526648576154,-60.0800974491509052 0.753763526648576154,-60.0800974491509052 0.753942644782787164,-60.080276567285118 0.753942644782787164,-60.080276567285118 0.754121762916998062,-60.0804556854193308 0.754300881051209071,-60.0804556854193308 0.75447999918542008,-60.0806348035535365 0.75447999918542008,-60.0806348035535365 0.75465911731963109,-60.0808139216877493 0.75465911731963109,-60.0808139216877493 0.754838235453842099,-60.080993039821962 0.754838235453842099,-60.080993039821962 0.755017353588053108,-60.080993039821962 0.755196471722264007,-60.080993039821962 0.755375589856475016,-60.080993039821962 0.755554707990686025,-60.0808139216877493 0.755554707990686025,-60.0806348035535365 0.755554707990686025,-60.0804556854193308 0.755554707990686025,-60.0803439655965832 0.755554707990686025,-60.0804030808496634 0.755761611376456033,-60.0804964752807251 0.755730479899434759,-60.0805102752906919 0.75578567993928869,-60.0804099575818995 0.75578567993928869,-60.080557120236989 0.756300749232102443,-60.0815057932769179 0.757012254012048635,-60.0813512760903805 0.757166771198584887,-60.0815303942245933 0.756987653064373989,-60.0817095123588061 0.756987653064373989,-60.0818886304930189 0.756987653064373989,-60.0820677486272245 0.756987653064373989,-60.0822468667614373 0.756808534930162979,-60.0824259848956501 0.756808534930162979,-60.0826051030298629 0.756808534930162979,-60.0827842211640686 0.756808534930162979,-60.0827842211640686 0.756987653064373989,-60.0829633392982814 0.757166771198584887,-60.0831424574324942 0.757166771198584887,-60.0833215755667069 0.757345889332795896,-60.0835006937009126 0.757345889332795896,-60.0836798118351254 0.757345889332795896,-60.0836798118351254 0.757525007467006906,-60.0836798118351254 0.757704125601217915,-60.0838589299693382 0.757704125601217915,-60.0838589299693382 0.757883243735428924,-60.084038048103551 0.757883243735428924,-60.0842171841495727 0.758062379781453299,-60.0863666017601048 0.761286506197251134,-60.0855549691134669 0.761802987625890493,-60.0858922791764627 0.762244084602212491,-60.0849952785288366 0.762782284990789483,-60.0846364782697862 0.762423484731738155,-60.0848158604592939 0.762423466791725124,-60.0851746786583618 0.762064684472686826,-60.0846364782697862 0.76134708395458417,-60.0851646773047747 0.762051349334572636,-60.0843962843719694 0.762540315224914611,-60.0842171841495727 0.762540333136727977,-60.083096830832659 0.760808883566007177,-60.0830218771040521 0.760808883566007177,-60.0830218771040521 0.760693046357103775,-60.0822468667614373 0.759495306943327786,-60.0822468667614373 0.759316188809116777,-60.0820677665390406 0.759316206720930253,-60.0813512760903805 0.759495306943327786,-60.0813512760903805 0.759316188809116777,-60.0811721579561748 0.759137070674905767,-60.0811721579561748 0.758957952540694869,-60.0811721579561748 0.75877883440648386,-60.080993039821962 0.75859971627227285,-60.080993039821962 0.758420598138061841,-60.080993039821962 0.758415802716399745,-60.0806260282026372 0.758369922249845452,-60.080603817897746 0.758392134775881188,-60.0804246601797658 0.758392116861900778,-60.0795289790737144 0.758392134775881188,-60.0795289611597383 0.758212977057895143,-60.0795289611597383 0.75803383725388962,-60.0795289611597383 0.757874944330069966,-60.0789445028920852 0.756121569527112958,-60.0790510714164583 0.755375589856475016,-60.0790227403456427 0.755375589856475016,-60.0788436222114299 0.755196471722264007,-60.0786645040772171 0.755196471722264007,-60.0784853859430044 0.755196471722264007,-60.0784853859430044 0.755017353588053108,-60.0783062678087987 0.755017353588053108,-60.0781271496745859 0.755017353588053108,-60.0781271496745859 0.754838235453842099,-60.0779480494521891 0.754838253365655465,-60.0778318325775729 0.754700906342291034,-60.0778192733478065 0.754709279162134705,-60.0778065048205718 0.754670973580421678,-60.0773835163154786 0.754171078773557713,-60.0772810729592308 0.754171078773557713,-60.0772144849313321 0.75397131468986367,-60.0759777320640538 0.7525096997090992,-60.0759777499758698 0.752330599486701668,-60.075082141392997 0.752330581574888191,-60.075082141392997 0.752151463440677182,-60.075082141392997 0.751972345306466172,-60.0752612595272097 0.751793227172255274,-60.0752612595272097 0.751614109038044265,-60.0754403776614225 0.751614109038044265,-60.0754403776614225 0.751434990903833255,-60.0756194957956353 0.751434990903833255,-60.075798613929841 0.751255872769622246,-60.0758279619977884 0.751226524701677945,-60.0757670252756171 0.751226524701677945,-60.075587885471613 0.751226524701677945,-60.0754087456676089 0.751226524701677945,-60.0752296058635977 0.751226524701677945,-60.0750504660595936 0.751226524701677945,-60.0748713262555896 0.751226524701677945,-60.0746922043655616 0.751226542615658355,-60.0745130466475814 0.751047384897672421,-60.0745130466475814 0.750868245093666786,-60.0745130466475814 0.750689105289661263,-60.0745130466475814 0.750509965485655628,-60.0745130466475814 0.750330825681650104,-60.0745130466475814 0.75015168587764447,-60.0746921864515855 0.75015168587764447,-60.0746921864515855 0.75,-60.0759461650796212 0.75,-60.0759461650796212 0.75015168587764447,-60.0759461650796212 0.750330825681650104,-60.0759461650796212 0.750509965485655628,-60.0761253048836323 0.750689105289661263)))";
    std::string wkt08 = "MULTIPOLYGON (((-61.52603809951475 3.539035163901769, -61.52603985855592 3.539026920528163, -61.5270374638152 3.534351854904043, -61.527195111095025 3.533613074337468, -61.52728342264327 3.5336132149590775, -61.527283850500304 3.5333418961532117, -61.52728408033498 3.533196138849774, -61.52728755592775 3.533179851220846, -61.52731086468585 3.533070619665862, -61.52735488908668 3.5330706897438784, -61.52735488908668 3.5331089205321113, -61.52950688908655 3.5331089205322193, -61.52950688908659 3.5336469205318672, -61.530582889086496 3.533646920531921, -61.530582889086546 3.534722920531215, -61.53112088908652 3.5347229205312414, -61.531120889086544 3.535260920530888, -61.5316588890865 3.535260920530914, -61.53165888908647 3.534722920531268, -61.532196889086435 3.5347229205312938, -61.532196889086485 3.5357989205305875, -61.531658889086536 3.535798920530561, -61.5316588890866 3.5368749205298546, -61.53112088908662 3.536874920529829, -61.53112088908666 3.5374129205294778, -61.530582889086695 3.5374129205294484, -61.53058288908667 3.5368749205298013, -61.530044889086696 3.5368749205297765, -61.53004488908667 3.5363369205301285, -61.52896888908673 3.536336920530074, -61.52896888908679 3.537412920529368, -61.52950688908678 3.5374129205293956, -61.5295068890868 3.5379509205290423, -61.52896888908682 3.5379509205290134, -61.52896888908687 3.539026920528308, -61.526039859592444 3.539026920528163, -61.52603757608213 3.5390376168577844, -61.52603809951475 3.539035163901769)))";
    std::string wkt07 = te::gtest::GetWktFromFile("huge_polygon_wkt_2.txt");
    std::string wkt06 = "POLYGON ((-41.6740995562981 -20.745600050613398, -41.6740995562981 -20.733491993771562, -41.67409105510578 -20.733416821114638, -41.67407582089285 -20.733285742658868, -41.67403310227144 -20.733164476117643, -41.673973540773574 -20.73306662727492, -41.673895914590766 -20.732978234146046, -41.67385669607207 -20.732890638914103, -41.67383457102919 -20.73281958977272, -41.673845984082334 -20.73276229265076, -41.67390051598791 -20.732713518653227, -41.67395874761044 -20.732654772731873, -41.674006152590735 -20.73261724100094, -41.67404137089003 -20.73261018430541, -41.6740554982949 -20.73264885641104, -41.674038267043635 -20.732713366056227, -41.67402089907773 -20.7327919934479, -41.67402983908341 -20.73287612070905, -41.67409045649446 -20.732969715367386, -41.6740995562981 -20.73297873533571, -41.6740995562981 -20.731876980796113, -41.674081955935875 -20.73188419581297, -41.67398032244037 -20.731934611966278, -41.673892674142046 -20.731975472453428, -41.67379223731035 -20.732016699113075, -41.67365052238735 -20.732055535752902, -41.67351006436853 -20.732077695510593, -41.67339338980015 -20.73203556449504, -41.67334835955849 -20.731906493376393, -41.673362964699116 -20.73173018209115, -41.673380531005485 -20.731557547782106, -41.67340919584607 -20.731392438986457, -41.67343011415253 -20.731230984935223, -41.67342799063069 -20.73108224687638, -41.67342485524358 -20.730945215820828, -41.6734468167455 -20.730791400743104, -41.673483078663445 -20.730616269163953, -41.673520282317504 -20.730467642767074, -41.673533246170834 -20.730344024039105, -41.6735469404956 -20.73023366838646, -41.67355433094802 -20.730149903531053, -41.67355724768763 -20.730084493672198, -41.67355566814717 -20.73, -41.6740995562981 -20.730009316050857, -41.6740995562981 -20.73, -41.67355566814717 -20.73, -41.65779336060545 -20.73, -41.65779336060545 -20.73968851739311, -41.65788976756544 -20.739646481747585, -41.65797384030815 -20.739628495244848, -41.658020387956626 -20.739644706581213, -41.658077503518825 -20.73967507064123, -41.65815938829908 -20.73970442751074, -41.65826727405857 -20.7397170009175, -41.658363271624694 -20.739722964201544, -41.65844435535949 -20.739706109783672, -41.658500636386975 -20.73967704754427, -41.65851661229904 -20.739642119666332, -41.6585034747605 -20.739600318576034, -41.65851345354411 -20.73955020659965, -41.65853924645495 -20.739503133052526, -41.658589738471754 -20.73946091717925, -41.65867035846403 -20.73941883586581, -41.6587392347667 -20.739353524275558, -41.658775592056344 -20.73924585946963, -41.65879258891239 -20.739121727347612, -41.65880342829737 -20.73900631362428, -41.658821798390726 -20.738901480744907, -41.658807152215545 -20.738804654460395, -41.658723854055886 -20.738704731209708, -41.65859143073352 -20.738585175118793, -41.658486061438474 -20.73842196001131, -41.658492606391704 -20.738292634911176, -41.658565269310515 -20.738208657202954, -41.65867467230837 -20.738163841430058, -41.65879911868914 -20.738163846401278, -41.65893848524003 -20.73823098314222, -41.65907237179505 -20.738339717194286, -41.65921249172882 -20.73845655146933, -41.65937153730851 -20.73857178957698, -41.65953687305215 -20.738658288487198, -41.65968172771462 -20.73870872562912, -41.65982420558871 -20.738682858646182, -41.659895507363736 -20.73862104734091, -41.65988839412362 -20.738564922540505, -41.65985216867654 -20.738501679375045, -41.65980617416939 -20.738414943092017, -41.65974074619186 -20.738330443129154, -41.659657981623994 -20.738248180683506, -41.659563351492885 -20.738162913729493, -41.65948348610358 -20.73807197469139, -41.659431270678766 -20.737966385406462, -41.65941173131076 -20.737827491263666, -41.659428715057835 -20.73767125431366, -41.65945879971138 -20.737513361898856, -41.65950563669814 -20.737365792608013, -41.6595634066552 -20.73723716049561, -41.659620178459484 -20.73713024095506, -41.65972549780441 -20.737049701212268, -41.659878032361014 -20.73702502789743, -41.660068965123656 -20.737085381561613, -41.66024289008979 -20.737185147773143, -41.66039488280839 -20.7372621609512, -41.66055071111209 -20.737299334157044, -41.6607031922039 -20.737283341121117, -41.66086546771372 -20.737246069301346, -41.661032846285885 -20.737206320921256, -41.66121330464213 -20.737178392448694, -41.66137905893 -20.737168667170877, -41.66153399802118 -20.73712596461651, -41.66168945302378 -20.737035300630122, -41.66182908135767 -20.736939881085064, -41.661949601346485 -20.73685193969027, -41.662065566502164 -20.736791999739467, -41.66216080450839 -20.73671648417607, -41.662244465113474 -20.736592882479147, -41.66231816547098 -20.736439936829314, -41.66238125757288 -20.736288271763897, -41.662482004705396 -20.73618615445542, -41.66255341093798 -20.73614417475682, -41.66259921464129 -20.73614280520661, -41.66263470148699 -20.736166403964, -41.662658388619356 -20.736191771365593, -41.66267617700891 -20.736215442958024, -41.662693198445105 -20.736237977709223, -41.662714487371744 -20.73626631887808, -41.66273976550635 -20.736300104677326, -41.66277351326135 -20.736346057017478, -41.66280637518939 -20.736393746448368, -41.66283150814967 -20.736431278160893, -41.66285514038391 -20.736467606392107, -41.66288029031955 -20.73650782658011, -41.662912081159824 -20.736557256105804, -41.662948626254845 -20.736620022742876, -41.66299018474294 -20.736698498586318, -41.663035006713145 -20.736785262766364, -41.66308568549434 -20.73685811861815, -41.663136173365274 -20.73693094493679, -41.66319152530636 -20.737029639395978, -41.66325458919019 -20.737159857488514, -41.663307574037056 -20.737318543571725, -41.66334499581228 -20.73747951009685, -41.663351651599534 -20.737648525668963, -41.66331582174926 -20.737827713328052, -41.6632291161066 -20.737968638837362, -41.66315187440895 -20.738055240190327, -41.66311611428895 -20.73811540004539, -41.66314697166038 -20.738156964372326, -41.66322123792936 -20.738201183288663, -41.663268884986856 -20.738277108009708, -41.66330534737446 -20.738381197740395, -41.66335678818802 -20.738506663496988, -41.663398707167936 -20.738622190544604, -41.66343794482469 -20.73871757216326, -41.66346998754223 -20.738790290045035, -41.66348853797489 -20.738834142128134, -41.6634982208145 -20.738863328475954, -41.66350389362424 -20.738885336686494, -41.663509755277076 -20.738910958205, -41.663518641066396 -20.738947746367188, -41.66352930520392 -20.738983362720003, -41.66354827169181 -20.73903166262252, -41.66358860169031 -20.739120110064142, -41.663665919340346 -20.739273004471105, -41.66376700480718 -20.739465413307293, -41.66385039353499 -20.739629287300666, -41.663934021814605 -20.739766514256665, -41.66402199040936 -20.739901717119928, -41.66410172770936 -20.74002977242815, -41.66417004443055 -20.74013771599914, -41.66423731218604 -20.74023950539956, -41.66430415758875 -20.740309182152775, -41.66435448959927 -20.74035776649414, -41.66438738863196 -20.740389807119392, -41.664414180163206 -20.7404179220998, -41.66443964990927 -20.740446534620414, -41.66447053935933 -20.740479905954434, -41.66450788569031 -20.74052140647822, -41.664560201192984 -20.740574836650378, -41.664626517605896 -20.740617538666633, -41.66468750278318 -20.740661801236463, -41.66473993322903 -20.740708499340975, -41.66479127371887 -20.740757216555835, -41.664859423613635 -20.74081332724961, -41.6649577049868 -20.74088397444214, -41.6650940307163 -20.7409684040325, -41.66522407208405 -20.74103603482589, -41.665313326175806 -20.741088189033494, -41.66537081134694 -20.741128914394963, -41.66541236780966 -20.74116634888121, -41.66544517772148 -20.741199184031395, -41.665479336844136 -20.741233614372575, -41.665507929147985 -20.741259854165072, -41.66553883579167 -20.741283729491734, -41.66558151688054 -20.741312945024635, -41.6656284762241 -20.7413430124965, -41.66568732726392 -20.74137960437048, -41.66574452525463 -20.741416136191912, -41.665785369182366 -20.7414464883692, -41.6658145014877 -20.74147825279789, -41.665823904340236 -20.74151231176642, -41.66582719834204 -20.741547862738976, -41.665857396262346 -20.741580464151518, -41.66590147134738 -20.74160919064725, -41.66594538526733 -20.741650351549257, -41.66598802171214 -20.74168997803539, -41.666019354018545 -20.741728295808215, -41.66604290855543 -20.741774624805835, -41.66605871488473 -20.74181821450484, -41.66606971854619 -20.741868100810887, -41.666070059508755 -20.74192189140484, -41.66605956184815 -20.74199733364187, -41.666046866725914 -20.74210808812874, -41.666042260122225 -20.74222171518585, -41.66605099749692 -20.74232947270873, -41.666107669651524 -20.742463921910126, -41.6661509800077 -20.742605112240405, -41.66616073585033 -20.742705078656297, -41.66614961137413 -20.742776847822395, -41.666171644154325 -20.742835574550426, -41.66622456675712 -20.742883942745934, -41.66627219976713 -20.742944647134983, -41.66632104533599 -20.74299369646512, -41.66635391002918 -20.743036853751452, -41.66637572131684 -20.743082446716883, -41.6663887350306 -20.74311833217469, -41.66639827420984 -20.743149035680673, -41.66640750462737 -20.74318062370072, -41.66641463700828 -20.743202496291943, -41.66642110230559 -20.74321719645025, -41.666428270707904 -20.74322829829467, -41.66643747978365 -20.743239185294627, -41.666451077509855 -20.743254249263753, -41.66646959045726 -20.743276398205165, -41.66648736734992 -20.743302832124353, -41.66650352461627 -20.743336606048747, -41.66651718903165 -20.743388249409957, -41.66652805975965 -20.743460553921985, -41.666541415769636 -20.743530631073405, -41.666587136080864 -20.743589271446304, -41.6666411491548 -20.743641317882677, -41.666697573697505 -20.74370753757461, -41.66675962585613 -20.743788708324793, -41.66683259037531 -20.74389335955665, -41.666919321818064 -20.744006112962325, -41.6670026793475 -20.74412250037767, -41.66708576322706 -20.744244699019557, -41.6671577157299 -20.74436024142325, -41.66720160776921 -20.744437847364388, -41.66718692373089 -20.744444787158002, -41.66711123388558 -20.744385294175398, -41.66701990992864 -20.744315467427118, -41.6669243422888 -20.744249902850864, -41.66682240909358 -20.744190061307663, -41.66670967657497 -20.74413523928303, -41.66659465415284 -20.744085652118354, -41.66647686135543 -20.74404236211309, -41.666358814346964 -20.744002187357523, -41.666243805313144 -20.743961490691387, -41.66613471212809 -20.743916398068123, -41.66602935480872 -20.74386330096185, -41.665925189514894 -20.743803705927817, -41.66581956693486 -20.74374309815451, -41.665708359241926 -20.743681175969208, -41.66558489990845 -20.74361641668781, -41.6654419709614 -20.743550735780723, -41.66530579972402 -20.7435001052763, -41.66517979646259 -20.74346504074705, -41.66506160695816 -20.743445427319916, -41.66494423169032 -20.743439148284132, -41.66482533200942 -20.74344205210552, -41.66470040759681 -20.74345196977205, -41.664575645878585 -20.743465216774432, -41.66445719928149 -20.743484300200656, -41.664339929973444 -20.743509383336637, -41.66421950436183 -20.74354004564574, -41.664104113547985 -20.743570459174087, -41.664021900228704 -20.74359299783159, -41.66395300974537 -20.743617991785893, -41.66389933310019 -20.743647207994474, -41.66384123163142 -20.743680043131505, -41.66376147822641 -20.743723432733965, -41.66367112277529 -20.743773304190206, -41.663579439114386 -20.743826320242658, -41.66349300905173 -20.743885560172846, -41.66341310686646 -20.74395036733392, -41.66334316849959 -20.744015027564295, -41.66329189350601 -20.74405682283946, -41.663252340365155 -20.744064859437852, -41.663210501523885 -20.744045594047126, -41.66316479633053 -20.74401767845569, -41.66312190908126 -20.74399144708039, -41.66305322588513 -20.74400865406191, -41.663014477609345 -20.744054959238404, -41.66300412957307 -20.744104747574802, -41.66297732807527 -20.74415498698874, -41.66290015433389 -20.744192380448595, -41.66276961808227 -20.744197388114408, -41.66263384601438 -20.744184554937128, -41.66255401752988 -20.744203261268357, -41.66254289604653 -20.744272986792367, -41.662531954761356 -20.74434342026285, -41.66242669900611 -20.74436214142157, -41.66227456284552 -20.744296778165353, -41.66211758419672 -20.744199082490965, -41.66196584582835 -20.74409720555975, -41.66183598410192 -20.74401702796449, -41.66176581295311 -20.7439533177971, -41.66177369717066 -20.743902473999608, -41.66184757118275 -20.743873743371733, -41.661929243796834 -20.743856359364596, -41.66201695135318 -20.743847020248364, -41.66210904274652 -20.74383445490231, -41.66216259705952 -20.743796747616244, -41.66213537101618 -20.74375284988343, -41.66206940279838 -20.743707029979937, -41.66199424002524 -20.743655151937837, -41.66200153964465 -20.74360079500266, -41.662074916289306 -20.74355092794823, -41.66215392669088 -20.743505451243625, -41.66219853123093 -20.74346538436886, -41.662159558153824 -20.743425305325143, -41.662023778112385 -20.743428272234162, -41.661871509352835 -20.743455860515464, -41.66174457753231 -20.743510154453, -41.66163724872388 -20.743558568685017, -41.66155139870923 -20.74361111710784, -41.661471070811565 -20.743662352783193, -41.66136999308336 -20.743715036033876, -41.661236254500956 -20.743749328903945, -41.66109901132243 -20.743749231105326, -41.66097450991582 -20.743719535380727, -41.660838330847085 -20.743640678799903, -41.66069084872692 -20.74353373480132, -41.66052042716857 -20.743453487594056, -41.66037866984225 -20.743394470467887, -41.66025382012387 -20.74331482554144, -41.66015333173947 -20.743215541050898, -41.66006196114941 -20.743115057553638, -41.6599656398833 -20.74302011544512, -41.65986155457061 -20.742939276255914, -41.6596836479663 -20.742808575210187, -41.65957922080054 -20.742701660168603, -41.65950733828344 -20.74262269581319, -41.65943106248444 -20.742548333594257, -41.6593795883404 -20.74247154684245, -41.659352808953436 -20.742386739581843, -41.659351831412394 -20.742289412336362, -41.65933886471565 -20.74220695669768, -41.65932478842483 -20.74211660891175, -41.65929889134633 -20.742017630086465, -41.65924562041527 -20.74190393861797, -41.659141423343065 -20.741766930518413, -41.65899469117973 -20.741611236089998, -41.658835417446134 -20.741456725766234, -41.65867206170953 -20.741353966531307, -41.65852133543315 -20.741294940211784, -41.65840573833769 -20.74124501032948, -41.658354035646724 -20.741190240847803, -41.65835282727461 -20.74112474950174, -41.65839029901174 -20.74105204202933, -41.658435644524936 -20.740987562316647, -41.65846786561637 -20.740928869730876, -41.65846202020989 -20.740876725385725, -41.65837384075368 -20.740843752947647, -41.65819599957164 -20.740845442787958, -41.65796784041449 -20.74084712944595, -41.65779336060545 -20.740828939809568, -41.65779336060545 -20.74377488388866, -41.657925202033695 -20.7438221944967, -41.65805809906931 -20.74389390084637, -41.65817608982547 -20.743983372223386, -41.65835090196588 -20.744066267063975, -41.65856863722696 -20.74413698995043, -41.658839840625234 -20.744195582726388, -41.659106646985734 -20.7442379194713, -41.659316107527644 -20.74428692241058, -41.6594687448173 -20.744324256293623, -41.65960021941174 -20.74432445598591, -41.659718048234566 -20.744306688004635, -41.6598131227372 -20.744335739298815, -41.659889701801596 -20.744413878411336, -41.66000844317093 -20.744491633696413, -41.66015280529539 -20.74453580066796, -41.66029323159196 -20.74456374048525, -41.66040356469429 -20.744591309986514, -41.66048676180369 -20.74464025119477, -41.6605393363845 -20.74472283189573, -41.66055277263561 -20.744814102599733, -41.660508859517726 -20.74491908222078, -41.660408139740184 -20.745030267217235, -41.66030109270642 -20.745131054198417, -41.66024876616071 -20.74521641474934, -41.66027360120346 -20.74524917651008, -41.66035157204393 -20.745230758486002, -41.66041481750088 -20.74523940164157, -41.66040403975954 -20.74528596519911, -41.660347371474906 -20.745352083664937, -41.66025725289538 -20.74543588543198, -41.6601577575571 -20.7455347961252, -41.66009553873006 -20.745600050613398, -41.6740995562981 -20.745600050613398))";
    std::string wkt05 = "MULTIPOLYGON (((-65.99037861046052 -4.091811995361438, -65.99031457950917 -4.091787984855196, -65.99037869926688 -4.09178822641072, -65.99037802418253 -4.09196891220657, -65.99037734906811 -4.092149598001494, -65.99055729114389 -4.092150275908573, -65.99073723324972 -4.092150953775266, -65.9909171753856 -4.092151631601576, -65.99109711755156 -4.092152309387501, -65.99127705974756 -4.092152987133043, -65.99127773466091 -4.091972301187985, -65.99145767684676 -4.091972978863121, -65.99163761906264 -4.091973656497875, -65.9916382938655 -4.091792970491849, -65.99163896863831 -4.0916122844848966, -65.99163919179084 -4.091552527578432, -65.991731456216 -4.091552527578432, -65.99199817666266 -4.091285807131765, -65.99199817666266 -4.091019086685098, -65.99199817666266 -4.090752366238432, -65.99199817666266 -4.090485645791765, -65.99199817666266 -4.0902189253450985, -65.99199817666266 -4.089952204898432, -65.99199817666266 -4.089685484451765, -65.991731456216 -4.089685484451765, -65.991731456216 -4.0894187640050985, -65.991731456216 -4.089152043558432, -65.99146473576933 -4.089152043558432, -65.99146473576933 -4.088885323111766, -65.99146473576933 -4.0886186026650995, -65.99146473576933 -4.088351882218433, -65.99146473576933 -4.088085161771766, -65.99119801532267 -4.088085161771766, -65.99119801532267 -4.0878184413250995, -65.99119801532267 -4.087551720878433, -65.99119801532267 -4.087285000431766, -65.99119801532267 -4.0870182799851, -65.99119801532267 -4.086751559538433, -65.99119801532267 -4.086484839091766, -65.99093129487599 -4.086484839091766, -65.99093129487599 -4.0862181186451005, -65.99066457442933 -4.0862181186451005, -65.99039785398266 -4.0862181186451005, -65.990131133536 -4.0862181186451005, -65.98986441308934 -4.0862181186451005, -65.98959769264266 -4.0862181186451005, -65.989330972196 -4.086484839091766, -65.989330972196 -4.086751559538433, -65.98906425174933 -4.086751559538433, -65.98906425174933 -4.0870182799851, -65.98906425174933 -4.087285000431766, -65.98906425174933 -4.087446758511437, -65.9891353058997 -4.087447026019222, -65.98895536510948 -4.087446348570234, -65.98877542434934 -4.087445671080907, -65.98877474962492 -4.087626356628611, -65.98859480885469 -4.0876256790689, -65.98841486811455 -4.08762500146885, -65.98823492740449 -4.087624323828461, -65.98823425252947 -4.087805009285091, -65.98823357762437 -4.087985694740795, -65.98823290268918 -4.088166380195575, -65.98805296188867 -4.088165702424687, -65.98805228688323 -4.088346387848488, -65.9880516118477 -4.088527073271363, -65.98805093678209 -4.088707758693315, -65.98787099589113 -4.0887070807919175, -65.98787032075525 -4.0888877661828875, -65.98786964558927 -4.089068451572935, -65.98786897039322 -4.089249136962054, -65.98786829516709 -4.089429822350248, -65.98786761991084 -4.089610507737519, -65.98786694462454 -4.089791193123863, -65.98786626930814 -4.089971878509281, -65.98786559396166 -4.090152563893775, -65.98786491858509 -4.090333249277343, -65.98786424317842 -4.0905139346599855, -65.98786356774168 -4.0906946200417025, -65.98786289227485 -4.090875305422493, -65.98804283364812 -4.090875983684555, -65.9880421581914 -4.091056669094476, -65.98804148270462 -4.091237354503471, -65.98804080718772 -4.091418039911541, -65.98804046549061 -4.091509432050502, -65.98799736996267 -4.091552527578432, -65.987730649516 -4.0918192480250974, -65.98746392906934 -4.092085968471764, -65.98746392906934 -4.092352688918431, -65.98746392906934 -4.092619409365097, -65.98746392906934 -4.092886129811764, -65.98746392906934 -4.093152850258431, -65.98773067618804 -4.093152823586386, -65.98986441308934 -4.0918192480250974, -65.990131133536 -4.0918192480250974, -65.99037858345532 -4.09181922328258, -65.99037861046052 -4.091811995361438)))";
    std::string wkt04 = "MULTIPOLYGON (((-69.49189569631487 -10.575214625320914, -69.49189569631487 -10.574990507165179, -69.49189569631487 -10.574766389009447, -69.49189569631487 -10.574542270853714, -69.49189569631487 -10.574318152697982, -69.49189569631487 -10.574094034542249, -69.49189569631487 -10.573869916386515, -69.49189569631487 -10.573645798230782, -69.49189569631487 -10.57342168007505, -69.49189569631487 -10.573197561919317, -69.49189569631487 -10.57301003020227, -69.49182241590574 -10.57264362815721, -69.49109847929196 -10.57264362815721, -69.49109847929196 -10.572525207452117, -69.49122334184767 -10.572525207452117, -69.49167157815914 -10.572525207452117, -69.49189569631487 -10.572525207452117, -69.49189569631487 -10.572525207452117, -69.49167157815914 -10.572525207452117, -69.49122334184767 -10.572525207452117, -69.49109847929196 -10.572525207452117, -69.49099922369194 -10.572525207452117, -69.49055098738047 -10.572525207452117, -69.49032686922475 -10.572525207452117, -69.490102751069 -10.572525207452117, -69.48987863291327 -10.572525207452117, -69.48965451475755 -10.572525207452117, -69.48943039660182 -10.572525207452117, -69.48920627844608 -10.572525207452117, -69.48898216029035 -10.572525207452117, -69.4887580421346 -10.57274932560785, -69.4887580421346 -10.572973443763583, -69.4887580421346 -10.573197561919317, -69.48898216029035 -10.573197561919317, -69.48898216029035 -10.57342168007505, -69.48920627844608 -10.57342168007505, -69.48920627844608 -10.573645798230782, -69.48943039660182 -10.573869916386515, -69.48943039660182 -10.574094034542249, -69.48943039660182 -10.574318152697982, -69.49010277348084 -10.575214602909098, -69.49032686922475 -10.575214625320914, -69.49055098738047 -10.575214625320914, -69.4907751055362 -10.575214625320914, -69.49099922369194 -10.575214625320914, -69.49122334184767 -10.575214625320914, -69.49144746000341 -10.575214625320914, -69.49167157815914 -10.575214625320914, -69.49189569631487 -10.575214625320914)))";
    std::string wkt03 = "MULTIPOLYGON (((-67.43861429516198 -9.456244395390115, -67.43866512068917 -9.456193569862876, -67.43871776004721 -9.456140930504944, -67.4390667452991 -9.456140930504944, -67.43941573055099 -9.456140930504944, -67.4401137010549 -9.456140930504944, -67.44016634041282 -9.456193569862876, -67.4404252389952 -9.456452468445285, -67.4404252389952 -9.456193569862876, -67.44016634041282 -9.456193569862876, -67.4401137010549 -9.456140930504944, -67.43941573055099 -9.456140930504944, -67.4390667452991 -9.456140930504944, -67.43871776004721 -9.456140930504944, -67.43866512068917 -9.456193569862876, -67.43861429516198 -9.456193569862876, -67.43861429516198 -9.456244395390115)))";
    std::string wkt02 = "MULTIPOLYGON (((-66.11486785868165 -9.503638882458754, -66.11486785868165 -9.503638882458763, -66.1142127734077 -9.507241851464755, -66.11390105603675 -9.507137945674344, -66.11363496141007 -9.507137945674344, -66.11363496141007 -9.506969300546643, -66.11309696141006 -9.506969300546643, -66.11309696141006 -9.507144995131911, -66.1129414171897 -9.507110429749623, -66.1130969614101 -9.507144995131933, -66.11309696141011 -9.506969300546741, -66.11363496141006 -9.506969300546707, -66.11363496141004 -9.50723830054653, -66.11390396141002 -9.507238300546515, -66.11390396141002 -9.507324328465248, -66.11418636725249 -9.507387085319133, -66.11486785868165 -9.503638882458754)))";

    std::string wkt01 = "MULTIPOLYGON (((-66.00159729478753 -8.971313169434826, -66.00159729478767 -8.971313169434257, -66.00159729478767 -8.970506315894122, -66.00224172096182 -8.968573037371646, -66.00224172096182 -8.968573037371584, -66.002223694946 -8.96862711541911, -66.00193738440328 -8.968913425961716, -66.00212825809842 -8.968913425961716, -66.00159729478753 -8.970506315894113, -66.00159729478753 -8.971313169434325, -66.00159729478753 -8.971313169434826)))";

    std::string wkt0 = "MULTIPOLYGON (((-49.99900526775495 -1.4686859061674484, -49.99900496263029 -1.4693714802591222, -49.99900484180583 -1.469642856178192, -49.99927449092923 -1.4696429777712592, -49.999813789243014 -1.4696432208591554, -49.99981353890423 -1.470205737193789, -49.999862130921386 -1.470205737193789, -49.999929136187156 -1.4698372082320597, -49.999999999821114 -1.469447456277538, -50 -1.4698372082320597, -50.00078345332571 -1.4698372082320597, -50.00096771780658 -1.4700214727129246, -50.00133624676831 -1.4700214727129246, -50.00133624676831 -1.4702057371937893, -50.001520511249176 -1.4702057371937893, -50.001520511249176 -1.4703900016746543, -50.001889040210905 -1.4703900016746543, -50.001889040210905 -1.470574266155519, -50.00224021988365 -1.470574266155519, -50.00224015080054 -1.470729817355422, -50.00331874901851 -1.4707303028419765, -50.003588398628686 -1.4707304240005676, -50.003588518988984 -1.4704590477031514, -50.003588879936494 -1.469644918806872, -50.003589000207846 -1.4693735425067684, -50.00331935076041 -1.4693734214600043, -50.00331959130148 -1.4688306689025135, -50.00331965544069 -1.4686859061674484, -50.00331971153865 -1.468559292622761, -50.00305006221102 -1.4685591716103705, -50.00278041290564 -1.468559050565258, -50.002780533185614 -1.4682876743295827, -50.0027806534433 -1.4680162980932376, -50.00251100422519 -1.4680161770601703, -50.00089310938418 -1.4680154501748583, -50.00089322984711 -1.4677440740946528, -50.000893350287754 -1.4674726980137767, -50.00062370129058 -1.4674725767965924, -50.000623821741414 -1.4672012007374706, -50.00035417279902 -1.4672010795100199, -50.00008452387893 -1.4672009582498788, -50.00008464437243 -1.4669295822349566, -50 -1.4669295441674648, -50 -1.467994563423411, -49.99912507299793 -1.4679945634234113, -49.99930933747879 -1.4678102989425466, -49.99949360195966 -1.4674417699808169, -49.99967786644053 -1.4674417699808169, -49.99967786644053 -1.4670732410190872, -49.99982164350521 -1.4669294639544013, -49.99981499550714 -1.4669294609645673, -49.99981535695115 -1.46611533298311, -49.997388519045224 -1.466114240685562, -49.9971188705008 -1.4661141191558316, -49.99711874971715 -1.466385494925549, -49.99711862891112 -1.466656870694596, -49.996848980324145 -1.4666567490871771, -49.99684885946327 -1.4669281248330432, -49.996848254823114 -1.468285003552313, -49.996848133827925 -1.4685563792941534, -49.99711778264235 -1.468556501059146, -49.99900532496962 -1.468557352497884, -49.99900526775495 -1.4686859061674484),(-50.00013846376376 -1.4686859061674484, -50 -1.4694474552936656, -50 -1.4686859061674484, -50 -1.4679945644234111, -50.00004639540225 -1.467994563423411, -50.00004639540225 -1.4681788279042762, -50.000230659883115 -1.4681788279042762, -50.00013846376376 -1.4686859061674484)))";

    std::string wkt1 = "multipolygon(((-64.1451981793532298 3.55296681855199648, -64.1451981793532298 3.57002712739561545, -64.1450918067335323 3.57002712739561545, -64.1450918067335323 3.56991904724431652, -64.1450205926935695 3.56987987952233476, -64.1449340157781194 3.56983226221883632, -64.1448969415567518 3.56983226221883632, -64.1448969415567518 3.56981187139708744, -64.1442254151352813 3.56944253186527805, -64.1439226156728637 3.56944253186527805, -64.1438738370612214 3.56936936394781235, -64.1432828936233506 3.56936936394781235, -64.1432828936233506 3.56924666279269376, -64.1434622515854613 3.56924666279269376, -64.1434622515854613 3.56906730483057677, -64.1436416095475863 3.56888794686845978, -64.1438209675096971 3.56870858890634279, -64.1440003254718079 3.56852923094422581, -64.1441796834339328 3.56834987298210882, -64.1443590413960436 3.56817051501999138, -64.1445383993581686 3.5679911570578744, -64.1447177573202794 3.56781179909575741, -64.1448971152824043 3.56781179909575741, -64.1448971152824043 3.56772508189847093, -64.1450179233212907 3.5677249309444643, -64.1447478551621373 3.56772526840438031, -64.1447475193541408 3.5674539140783379, -64.144477451248477 3.56745425143296746, -64.1444771155453566 3.56718289707956604, -64.1444767798679436 3.56691154272453037, -64.1444764442162381 3.5666401883678609, -64.1442063763223587 3.56664052556578071, -64.1442060407755008 3.56636917118175845, -64.1442057052543362 3.56609781679610194, -64.1439356374930867 3.56609815386305984, -64.143665569706144 3.56609849085049202, -64.143395501893508 3.56609882775839937, -64.1431254340552215 3.56609916458678056, -64.1428553661912417 3.56609950133563647, -64.1426822505643059 3.56609971714330687, -64.1426822656591042 3.56609971712448992, -64.1426822665188041 3.56634902966981304, -64.1426822415900517 3.56609971715449436, -64.1425852983016114 3.56609983800496666, -64.1426559905859506 3.5660997498792586, -64.1410292575303345 3.56772653716020249, -64.1425654617748791 3.56619028170800689, -64.1425654617748791 3.56619757743670363, -64.1425654617748791 3.56637693539882061, -64.1425654617748791 3.56655629336093805, -64.1425654617748791 3.56673565132305503, -64.1425654617748791 3.56691500928517202, -64.1425654617748791 3.56709436724728901, -64.1423861038127541 3.567273725209406, -64.1423861038127541 3.56745308317152343, -64.1423861038127541 3.56763244113364042, -64.1423861038127541 3.56781179909575741, -64.1423861038127541 3.5679911570578744, -64.1422067458506433 3.56817051501999138, -64.1420273878885325 3.56834987298210882, -64.1420273878885325 3.56852923094422581, -64.1418480299264075 3.56852923094422581, -64.1418480299264075 3.56870858890634279, -64.1416686719642968 3.56870858890634279, -64.1414893140021718 3.56888794686845978, -64.141309956040061 3.56906730483057677, -64.14113061601374 3.56906732276637317, -64.141029229980191 3.5687631697348996, -64.141029229980191 3.56882851560236425, -64.141029229980191 3.56910401710044223, -64.1413047314782716 3.56937951859852021, -64.1413047314782716 3.56973836072058504, -64.1412973950113638 3.56974202895403891, -64.1409247300051391 3.5699283614571522, -64.1405520649989143 3.57011469396026548, -64.1403657324957948 3.57030102646337877, -64.1401793999926895 3.57030102646337877, -64.1401793999926895 3.57119631845628938, -64.1409996380211709 3.57119631845628938, -64.1409996380211709 3.57139118363306851, -64.1417790987282928 3.57139118363306851, -64.1419739639050732 3.57158604880984765, -64.1421688290818537 3.57158604880984765, -64.1421688290818537 3.57178091398662678, -64.1423636942586342 3.57178091398662678, -64.1423636942586342 3.57179856698692122, -64.1423636942586342 3.57197577916340592, -64.1425585594354004 3.57197577916340592, -64.1425585594354004 3.57217064434018461, -64.1427534246121809 3.57217064434018461, -64.1427534246121809 3.57236550951696374, -64.1431331824776265 3.57236550951696374, -64.1439226156728637 3.57236550951696374, -64.1439226156728637 3.57253701650073907, -64.1444650475642959 3.57253701650073907, -64.1444650475642959 3.57256037469374288, -64.1450918067335323 3.57256037469374288, -64.1450918067335323 3.57272334900385236, -64.1451981793532298 3.57272334900385236, -64.1451981793532298 3.58362832807414966, -64.1412566261937798 3.58362832807414966, -64.141256427765839 3.58346821562181761, -64.1412566261937656 3.58362832807414966, -64.1409074366366809 3.58362832807414966, -64.1390851976311609 3.58319957415217605, -64.1388254332019585 3.58319989845089504, -64.1385553599296827 3.58320023553958356, -64.1385550244855267 3.58292888071676607, -64.1385552049578962 3.58307487237782496, -64.1377232395534094 3.58287911955448868, -64.13697398108485 3.58362832807414966, -64.1006565550055711 3.58362832807414966, -64.1006565550055711 3.58026179905184838, -64.1007412861745252 3.58026169691702023, -64.1007409621365554 3.57999033855325699, -64.1007403141347254 3.57944762182081755, -64.1010103898147037 3.57944729626604907, -64.1010100657714332 3.57917593792211441, -64.1010097417528755 3.57890457957654284, -64.1012798172492353 3.57890425399143997, -64.10127949317598 3.57863289566898057, -64.1012792875608426 3.57846071350074002, -64.1012791691274515 3.5783615373448816, -64.1018193197282358 3.57836088603425972, -64.1020893949915234 3.57836056025924254, -64.1020890707294484 3.57808920200778591, -64.1023591458885704 3.57808887617773674, -64.1026292210229229 3.57808855026788963, -64.1026288966267686 3.57781719206434889, -64.1026279235868799 3.5770031174439052, -64.102627599289761 3.57673175923381681, -64.1023575245524171 3.57673208501977324, -64.1018173750035061 3.57673273635238065, -64.1018170509693164 3.57646137806633924, -64.101816726959882 3.57619001977866269, -64.1015466523070785 3.57619034527579505, -64.1015437380496849 3.57374812039024459, -64.1020838859022319 3.57374746976196089, -64.1023539597913867 3.57374714432826757, -64.1023536358707418 3.57347578607377026, -64.1023523404356723 3.57239035303944208, -64.1042428541905025 3.5723880736398379, -64.1045129274849046 3.57238774769264156, -64.1045126030537062 3.57211638962836897, -64.1045113055769349 3.57103095735492948, -64.1045109812697547 3.57075959928248521, -64.1042409084510041 3.57075992508075846, -64.1034306898459647 3.57076090199777285, -64.103160616928065 3.57076122747751068, -64.1031609408389471 3.57103258567399706, -64.1028908678169955 3.57103291109888854, -64.1026207947703028 3.57103323644414283, -64.1026211185474182 3.57130459468857042, -64.1026220900273671 3.57211866941204859, -64.1006565550055711 3.57212103502532541, -64.1006565550055711 3.57049288263196241, -64.1007296363799952 3.5704927947905909, -64.1009997094413819 3.57049247005230397, -64.1009993862140277 3.57022111165439737, -64.1012694591714478 3.57022078686122901, -64.101269179589579 3.56998610966720209, -64.1006565550055711 3.56998610966720209, -64.1006565550055711 3.55296681855199648, -64.1451981793532298 3.55296681855199648), (-64.1018091105973724 3.56980532454323507, -64.1017681342339642 3.56980532454323507, -64.1016244577027834 3.56994900107442081, -64.1018092815717466 3.56994877871074889, -64.1018091105973724 3.56980532454323507), (-64.1090338280357628 3.57143239065893736, -64.1096431754630629 3.57143239065893736, -64.1096430117577967 3.57129610918266405, -64.1100402282328332 3.57129562723856964, -64.1100842499364347 3.57125160553497079, -64.1100842499364347 3.57052846503910271, -64.110265035060408 3.57034767991513569, -64.110445820184367 3.57016689479116911, -64.110992018472615 3.57016689479116911, -64.1109907647325628 3.56912361134993583, -64.1109904386977973 3.56885225386356719, -64.1107203670306234 3.56885258139828299, -64.1104502953385094 3.568852908853406, -64.1104499694871066 3.56858155131547772, -64.1101798978492781 3.56858187866605814, -64.1101792463797011 3.56803916353539385, -64.1101789206823014 3.56776780596761256, -64.1099088492571241 3.56776813316376451, -64.1096387778070351 3.56776846028034988, -64.1096384522929412 3.56749710266104714, -64.1093683808971093 3.56749742967312944, -64.1093680554871241 3.56722607202725728, -64.1093677301020364 3.56695471437975442, -64.1090976588396444 3.56695504126241891, -64.1090973335586511 3.56668368358835153, -64.1090970083025411 3.56641232591265389, -64.1090338280357628 3.56641240235310386, -64.108826937173589 3.56641265266592233, -64.1088262868944128 3.5658699372597833, -64.1088259617921636 3.56559857955426596, -64.1085558908757633 3.56559890615324626, -64.1085555658775661 3.56532754842117949, -64.1085552409042521 3.56505619068748114, -64.1082851701212206 3.56505651715712002, -64.1082845204075369 3.5645138016350093, -64.1082841955880269 3.56424244387150679, -64.108014125017462 3.56424277018692326, -64.1080134756113011 3.56370005460521488, -64.1080131509455384 3.56342869681191265, -64.1077430805874116 3.5634290229731449, -64.1061226579164156 3.5634309782716751, -64.1058525873842768 3.56343130387661944, -64.1058529114172728 3.56370266186894291, -64.104772628724092 3.56370396359338404, -64.1045025579887664 3.56370428882580503, -64.1045028816510722 3.56397564694068869, -64.1035759854736256 3.56397676255716167, -64.1035759854736256 3.56777578262268813, -64.1047774874502352 3.56777433477297024, -64.1047778115637641 3.56804569283855288, -64.1047781357021194 3.56831705090250439, -64.105048207758216 3.56831672516845755, -64.1050485320005805 3.56858808320594356, -64.1053186041110479 3.56858775736747935, -64.1063988923046679 3.56858645321776446, -64.10640149030624 3.57075731646482586, -64.1064018151682404 3.5710286743633568, -64.1066718878676625 3.57102834790314416, -64.1066722128337716 3.57129970577517142, -64.1066725378247355 3.5715710636455662, -64.1069426106578533 3.57157073705596373, -64.1069448868455112 3.57347024192882756, -64.1061346666064225 3.57347122198093059";
    wkt1 += ", -64.105864593143707 3.57347154850557658, -64.1058649180955342 3.5737429064374826, -64.1058658930999883 3.57455698022338808, -64.1055958192951181 3.57455730676774319, -64.1055961442669258 3.57482866471795724, -64.1055980946191681 3.57645681238488944, -64.1055984197647746 3.57672817032365176, -64.1058684942045716 3.57672784358048634, -64.1058688194544004 3.57699920149276185, -64.1058691447290556 3.57727055940339911, -64.1061392193028041 3.57727023253075282, -64.1074895917988101 3.57726859697083288, -64.1074886146202658 3.57645452368171268, -64.1077586888067117 3.57645419640507223, -64.1077583630508911 3.5761828386636525, -64.1077557578998523 3.57401197667339598, -64.1088187540686647 3.57401068831169777, -64.1088187540686647 3.57161317578290438, -64.1089995391926379 3.57143239065893736, -64.1090338280357628 3.57143239065893736), (-64.1158563111628581 3.57264534341541351, -64.1158559833764627 3.57237398640068982, -64.1158556556151069 3.57210262938432965, -64.115855327878819 3.57183127236633435, -64.1158550001675991 3.57155991534670481, -64.1158546724814187 3.57128855832544101, -64.1158543448203062 3.57101720130254119, -64.1158540171842475 3.57074584427800845, -64.1158536895732425 3.57047448725183925, -64.1161237612407433 3.57047415805442236, -64.1161234335755523 3.57020280105170151, -64.1163935051387455 3.57020247179974559, -64.1163931774193685 3.56993111482047842, -64.1163928497250879 3.56965975783957656, -64.1163925220558468 3.5693884008570409, -64.1163921944117021 3.56911704387287054, -64.1163918667926254 3.56884568688706771, -64.1163915391986023 3.56857432989963064, -64.1163912116296473 3.56830297291056064, -64.1163908840857744 3.56803161591985729, -64.1161208131563427 3.56803194497111376, -64.1158507422018431 3.5680322739427992, -64.1155806712222898 3.5680326028349123, -64.1155803439410477 3.56776124576733089, -64.1155800166848735 3.56748988869811701, -64.1153099458386322 3.56749021746052053, -64.1150398749673514 3.56749054614336414, -64.1147698040710452 3.56749087474664828, -64.1144997331496853 3.5674912032703725, -64.1142296622033143 3.56749153171453681, -64.1139595912319038 3.56749186007914165, -64.1139599180129807 3.5677632172986784, -64.1139602448190686 3.56803457451658357, -64.1139605716501677 3.56830593173285493, -64.1139608985062637 3.56857728894749293, -64.1139612253874134 3.56884864616049757, -64.1139615522935458 3.56912000337186841, -64.1136914808219274 3.56912033180710964, -64.1134214093252979 3.56912066016275498, -64.1131513378036715 3.56912098843880354, -64.1131516644971668 3.56939234572362185, -64.112881592871318 3.56939267394508919, -64.1128819195105706 3.56966403125328746, -64.1126118477805136 3.56966435942016069, -64.1123417760254597 3.56966468750742516, -64.1120717042454231 3.56966501551508131, -64.1118016324404181 3.56966534344312825, -64.1118019587877228 3.56993670084969139, -64.1118022851587739 3.57020805722533474, -64.1112108414566251 3.57020877556755378, -64.1113497458042048 3.57034767991513569, -64.1118024531013475 3.57034767991513569, -64.1118026115572661 3.57047941565791449, -64.1118029379794763 3.57075077305957667, -64.1115328658325438 3.57075110100795223, -64.1112627936606287 3.57075142887669505, -64.1109927214637594 3.57075175666580513, -64.1107226492419642 3.57075208437528202, -64.1104525769952005 3.57075241200512572, -64.1104529030460952 3.57102376953001288, -64.1104532291219158 3.57129512705326668, -64.1104535552226764 3.57156648457488579, -64.1104538813483771 3.57183784209486976, -64.1104542074990178 3.57210919961322038, -64.1107242801421222 3.57210887185857651, -64.1107246063969853 3.5723802293503315, -64.1107249326767743 3.57265158684045137, -64.1107252589815175 3.57292294432893653, -64.1107255853112008 3.57319430181578612, -64.1107259116658525 3.57346565930100191, -64.1107262380454159 3.57373701678458078, -64.1107265644499194 3.57400837426652451, -64.1107268908793912 3.57427973174683267, -64.110727217333789 3.57455108922550524, -64.110727543813141 3.57482244670254223, -64.1107275263300522 3.57480791540661258, -64.1108042241763627 3.57480790773606083, -64.1109835506832866 3.57462858122913829, -64.1111628771902105 3.57462858122913829, -64.1113422036971343 3.57462858122913829, -64.1113422036971343 3.57444925472221531, -64.1115215302040582 3.57444925472221531, -64.1117008567109821 3.57444925472221531, -64.1117008567109821 3.57426992821529277, -64.1118801832179059 3.57426992821529277, -64.1120595097248298 3.57426992821529277, -64.1122388362317537 3.57409060170837023, -64.1124181627386633 3.57409060170837023, -64.1124181627386633 3.57391127520144725, -64.1125974892455872 3.57391127520144725, -64.1127768157525111 3.57373194869452471, -64.112956142259435 3.57355262218760217, -64.1131354687663588 3.57337329568067918, -64.1133147952732827 3.57337329568067918, -64.1133147952732827 3.57319396917375665, -64.1134941217802066 3.57319396917375665, -64.1136734482871304 3.57319396917375665, -64.1136734482871304 3.57301464266683411, -64.1138527747940543 3.57301464266683411, -64.1138527747940543 3.57283531615991112, -64.1140321013009782 3.57283531615991112, -64.1142114278079021 3.57265598965298858, -64.1143907543148259 3.57265598965298858, -64.1145700808217356 3.57247666314606604, -64.1147494073286595 3.57247666314606604, -64.1149287338355833 3.57247666314606604, -64.1152873868494311 3.57247666314606604, -64.1158253663702027 3.57247666314606604, -64.1158561074050937 3.57247666314606604, -64.1158563111628581 3.57264534341541351), (-64.1301576819611654 3.56150218999404267, -64.1300104044266703 3.5613549124595405, -64.1300104044266703 3.56117555449742307, -64.1300104044266703 3.56099619653530608, -64.1300104044266703 3.56081683857318909, -64.1300104044266703 3.5606374806110721, -64.1300104044266703 3.56045812264895511, -64.1300104044266703 3.56009940672472069, -64.1300104044266703 3.55992004876260371, -64.1300104044266703 3.55987423430453598, -64.1298844651548734 3.55987438928665645, -64.1296143978593847 3.55987472155473705, -64.1293443305385296 3.55987505374343005, -64.1290742631923081 3.559875385852735, -64.1290745936978084 3.56014674170645318, -64.1288045262472366 3.56014707376174266, -64.1288048566990625 3.5604184296392023, -64.128534789144112 3.560418761640463, -64.1285351195422635 3.56069011754165965, -64.1282650518829485 3.56069044948888047, -64.1279949841983097 3.56069078135669503, -64.1277249164883187 3.56069111314510378, -64.1274548487530183 3.56069144485410538, -64.1271847809924083 3.56069177648370161, -64.1269147132064603 3.56069210803389069, -64.1266446453952312 3.56069243950467351, -64.1263745775586926 3.56069277089605007, -64.1263742477927536 3.56042141479214402, -64.1261041800099463 3.56042174607880302, -64.1258341122018578 3.56042207728606108, -64.125833782619253 3.56015072112990927, -64.1255637148648958 3.56015105223246753, -64.1252936470852859 3.56015138325563152, -64.1250235792803807 3.5601517141994008, -64.1247535114502369 3.56015204506377536, -64.1244834435948263 3.5601523758487561, -64.124213375714163 3.56015270655434168, -64.1242137048226652 3.56042406286219837, -64.1239436368377369 3.56042439351365214, -64.1239481370993616 3.56042438800386174, -64.1239481370993616 3.5605634706600271, -64.1242236385974422 3.56083897215810508, -64.1242236385974422 3.56111447365618305, -64.1242236385974422 3.56138997515426103, -64.1242236385974422 3.56166547665233901, -64.1242236385974422 3.56194097815041655, -64.1242236385974422 3.56221647964849453, -64.1244991400955229 3.56221647964849453, -64.1244991400955229 3.56249198114657251, -64.1244991400955229 3.56276748264465049, -64.1244991400955229 3.56304298414272846, -64.1244991400955229 3.56331848564080644, -64.1247746415936035 3.56331848564080644, -64.1247746415936035 3.56359398713888442, -64.12505014309167 3.56359398713888442, -64.1252978282889359 3.56359401190492742, -64.1252976018478762 3.56340765762894751, -64.125297272145346 3.56313630144014182, -64.1255673407945181 3.56313597013875238, -64.1255670110381857 3.56286461397360865, -64.1258370795830075 3.56286428261805677, -64.1261071481025482 3.56286395118305066, -64.1263772165968078 3.56286361966858856, -64.1266472850657721 3.56286328807467179, -64.1269173535094268 3.56286295640130035, -64.1271874219278004 3.56286262464847381, -64.1274574903208503 3.56286229281619171, -64.1277275586885764 3.56286196090445539, -64.127997627030993 3.56286162891326397, -64.1279972965882479 3.56259027297440189, -64.1279969661708407 3.56231891703390868, -64.1282670343297667 3.56231858501397047, -64.128537102463369 3.56231825291458959, -64.1288071705716334 3.56231792073576603, -64.1288068399423565 3.56204656486972082, -64.1290769079462137 3.56204623263682674, -64.1293469759247046 3.56204590032449664, -64.1296170438778432 3.56204556793273008, -64.1296167130367536 3.56177421214118928, -64.1298867808854709 3.56177387969537618, -64.1298864499906927 3.56150252392759414, -64.1301565177349886 3.56150219142774782, -64.1301576819611654 3.56150218999404267), (-64.1404421776975084 3.58021297307141984, -64.1404418419783724 3.57994161840982539, -64.1404415062848443 3.57967026374659003, -64.1404411706169384 3.57939890908171598, -64.1404408349746404 3.57912755441520192, -64.1404404993579504 3.57885619974704738, -64.1401704275109239 3.57885653690513772, -64.1399003556383178 3.57885687398341723, -64.1396302837401038 3.57885721098188547, -64.1393602118162818 3.57885754790054289, -64.1390901398668944 3.5788578847393886, -64.1388200678919134 3.57885822149842348, -64.1388204030319855 3.57912957632025375, -64.138820738197623 3.579400931140444, -64.1388209455597149 3.579568801591547, -64.1388252179955742 3.57957307402740366, -64.1391007194936407 3.57984857552548164, -64.1391007194936407 3.58012407702355961, -64.1393762209917213 3.58012407702355961, -64.1393762209917213 3.5802143038506129, -64.1393618885668815 3.58021432173729082, -64.1396319608879253 3.58021398469058516, -64.1399020331833896 3.58021364756403848, -64.1401721054532459 3.58021331035764989, -64.1404421776975084 3.58021297307141984), (-64.1447498705499157 3.56935339432630583, -64.1450199391842943 3.56935305671200753, -64.1450196031428561 3.56908170242150424, -64.1450198775148976 3.569303258522313, -64.1449700169170995 3.56935311912011066, -64.1447498705499157 3.56935339432630583)))";

    std::string wkt2 = "LINESTRING(-64.14474987054992 3.569353394326306, -64.1450199391843 3.5693530567120075, -64.14501960314286 3.5690817024215042, -64.1450198775149 3.569303258522313, -64.1449700169171 3.5693531191201107, -64.14474987054992 3.569353394326306)";

    std::string wkt3 = "POLYGON ((-50.93278988955943 2.4925693563677256, -50.93278988955943 2.4925693563677256, -50.93278988955943 2.4925693563677256, -50.93278490173396 2.492592800778822, -50.93278988955943 2.4925693563677256, -50.93278988955943 2.4925693563677256))";

    std::string wkt4 = "POLYGON ((0 0, 1 1, 2 2, 0 2, 1 1, 2 0, 0 0))";
    std::string wkt5 = "POLYGON ((0 0, 2 0, 1 1, 0 2, 2 2, 1 1, 0 0))";
    std::string wkt6 = "POLYGON ((0 0, 1 1, 0 2, 2 2, 1 1, 2 0, 0 0))";
    std::string wkt7 = "POLYGON ((0 0, 2 0, 1 1, 2 2, 0 2, 1 1, 0 0))";

    std::string wkt8 = "multipolygon(((-53.8727899626359061 -2.22231725152106963,-53.872754330395999 -2.22232238651748704,-53.8701843756027543 -2.22269274517839932,-53.8644116607571988 -2.22352530609850785,-53.8612953160829946 -2.22397475611617512,-53.8553468693720134 -2.22483199250333996,-53.855440979511279 -2.22483219442738189,-53.8557103750219568 -2.22483277239672006,-53.8559797706027865 -2.2248333503167399,-53.8562491662537752 -2.22483392818744186,-53.8565185619748945 -2.22483450600882504,-53.8567879577661657 -2.2248350837808899,-53.8567873834126303 -2.22510608802104182,-53.8567868089889714 -2.22537709226004532,-53.8567862344951749 -2.22564809649790263,-53.8567856599312478 -2.2259191007346133,-53.8567850852971901 -2.22619010497017511,-53.8570544814037859 -2.22619068304511991,-53.8570539067486536 -2.22646168734997518,-53.8570533320233977 -2.22673269165368204,-53.8570527572280184 -2.22700369595624137,-53.8570521823625086 -2.22727470025765317,-53.8570521384787781 -2.22729538549341433,-53.8579241349876909 -2.22781858339876049,-53.8581286187264752 -2.22781902235351925,-53.8583980154780804 -2.22781960060417861,-53.8585568300280997 -2.22781994146482676,-53.8585568300281494 -2.22675330987466324,-53.8574808300281873 -2.22675330987471964,-53.8574808300282228 -2.22621530987473371,-53.8601708300281032 -2.22621530987459426,-53.8601708300280677 -2.22675330987457887,-53.861246830026289 -2.22675330987452336,-53.8617848300279931 -2.22675330987449538,-53.861784830027986 -2.22729130987448176,-53.8628608300279339 -2.22729130987442492,-53.8628608300278913 -2.22782930987441175,-53.8623228300279209 -2.22782930987443972,-53.8623228300278996 -2.22836730987442699,-53.8617848300279149 -2.22836730987445408,-53.8617848300278865 -2.22890530987444135,-53.8655508300277148 -2.22890530987424329,-53.8666268300276627 -2.22890530987418645,-53.8666268300276485 -2.22944330987417239,-53.8682408300275739 -2.2294433098740889,-53.8682408300275952 -2.22890530987410296,-53.8687788300275798 -2.22890530987407454,-53.8687788300275372 -2.22944330987406136,-53.8698548300274993 -2.22944330987400452,-53.8698548300275206 -2.22890530987401858,-53.8720068300274235 -2.22890530987390623,-53.8720068300274448 -2.22836730987392073,-53.8752348300273027 -2.2283673098737502,-53.8752348300274733 -2.22513930987383324,-53.8757728300274366 -2.22513930987380437,-53.8757728300274579 -2.22460130987381888,-53.8752348300274875 -2.22460130987384774,-53.8752348300276012 -2.22244930987390221,-53.8757728300275716 -2.22244930987387557,-53.8757728300276 -2.22191130987388696,-53.875606866085775 -2.22191130609898302,-53.879073435111799 -2.22141173424965066,-53.8735136078370189 -2.22221296631404419,-53.8727899626359061 -2.22231725152106963),(-53.8727899626359061 -2.22231725152106963,-53.8730828300276983 -2.22227504736687287,-53.8730828300276556 -2.2232563098739937,-53.8722758300276894 -2.22325630987403589,-53.8722758300277036 -2.22298730987404358,-53.8720068300277148 -2.22298730987405735,-53.8720068300277291 -2.2227183098740646,-53.8717378300277474 -2.22271830987407881,-53.8717378300277616 -2.22246887661258752,-53.8718736317274107 -2.22244930609882552,-53.8727899626359061 -2.22231725152106963),(-53.8701843756027543 -2.22269274517839932,-53.8703689626360145 -2.22266614541112784,-53.870392830027825 -2.22266270585830306,-53.8703928300277965 -2.22298730987414128,-53.8703928300277752 -2.22325630987413403,-53.8701238300277936 -2.22325630987414824,-53.8701238300277794 -2.22352530987414232,-53.8698548300277906 -2.22352530987415697,-53.869854830027549 -2.22836730987403175,-53.8687788300276083 -2.2283673098740886,-53.8687788300276438 -2.22782930987410221,-53.868240830027645 -2.22782930987412975,-53.8682408300276805 -2.22729130987414381,-53.8677028300277101 -2.22729130987417312,-53.8677028300278025 -2.22513930987422626,-53.8666268300278617 -2.22513930987428266,-53.8666268300278261 -2.22567730987426948,-53.8650128300279007 -2.2256773098743543,-53.8650128300278723 -2.22621530987434069,-53.8644748300278948 -2.22621530987436778,-53.8644748300279161 -2.22567730987438184,-53.8639368300279529 -2.22567730987440981,-53.8639368300279884 -2.22513930987442432,-53.8644748300279446 -2.22513930987439545,-53.8644748300280085 -2.22406330987442402,-53.863936830028031 -2.22406330987445156,-53.8639368300280665 -2.22359378907853023,-53.8644116694130659 -2.22352530609850829,-53.8701843756027543 -2.22269274517839932)))";

    MakeValid(wkt011);
    MakeValid(wkt010);
    MakeValid(wkt09);
    MakeValid(wkt08);
    MakeValid(wkt07);
    MakeValid(wkt06);
    MakeValid(wkt05);
    MakeValid(wkt04);
    MakeValid(wkt03);
    MakeValid(wkt02);
    MakeValid(wkt01);
    MakeValid(wkt0);
    MakeValid(wkt1);
    MakeValid(wkt2);
    MakeValid(wkt3, true);
    MakeValid(wkt4, false, true);
    MakeValid(wkt5, false, true);
    MakeValid(wkt6);
    MakeValid(wkt7);
    MakeValid(wkt8);
  }

  void FixPolygon8(const std::string& wkt)
  {
    te::gm::GeometryPtr geometry(te::gm::WKTReader::read(wkt.c_str()));
    te::gm::FixNoding fixNoding;
    te::gm::GeometryPtr geometryFixed = fixNoding.validate(geometry.get());
    std::string wkt1 = geometryFixed->toString();
  }

  TEST_F(ITsValidation, FixPolygon8)
  {    
    std::string wkt4 = "POLYGON ((0 0, 1 1, 2 2, 0 2, 1 1, 2 0, 0 0))";
    std::string wkt5 = "POLYGON ((0 0, 2 0, 1 1, 0 2, 2 2, 1 1, 0 0))";
    std::string wkt6 = "POLYGON ((0 0, 1 1, 0 2, 2 2, 1 1, 2 0, 0 0))";
    std::string wkt7 = "POLYGON ((0 0, 2 0, 1 1, 2 2, 0 2, 1 1, 0 0))";
    std::string wkt8 = "POLYGON((300 200, 350 200, 400 200, 450 200, 500 200, 500 250, 500 300, 450 300, 450 250, 400 250, 400 300, 450 300, 450 350, 400 350, 350 350, 300 350, 300 300, 300 250, 300 200))";
    std::string wkt9 = "MULTIPOLYGON (((-41.953125 -20.749961137771606, -41.95155151824151 -20.749961137771606, -41.95163933405677 -20.74993515490642, -41.95175528383042 -20.74987045577679, -41.951839613518594 -20.74981731218386, -41.95190408833657 -20.74976401425723, -41.951959812096966 -20.749691748203166, -41.95201344070926 -20.74961371711619, -41.952065379276924 -20.74952244154893, -41.9521089303996 -20.749419457894955, -41.95211792464585 -20.74931533985666, -41.95207940705042 -20.749212686405837, -41.95196579591395 -20.749122604163134, -41.951847233849726 -20.749099358151593, -41.95172862857698 -20.749120868357288, -41.95159204934264 -20.749153115685573, -41.95145100602008 -20.749187665277553, -41.95130887089928 -20.74918307215943, -41.9511674288284 -20.74913168639865, -41.951015857253054 -20.749065537098804, -41.95089790120645 -20.748998929678674, -41.950785337639346 -20.74893409143257, -41.950675778044015 -20.748853593190788, -41.95059036823951 -20.74877424646007, -41.95057917651896 -20.748702255163423, -41.95061074540472 -20.748625943761287, -41.9506565158568 -20.748534342901134, -41.95067606108419 -20.748434856704783, -41.95067115831492 -20.74834285474344, -41.95068402730046 -20.748245011783613, -41.95074604717369 -20.74812402153137, -41.95078791529315 -20.747991328333452, -41.95071784432227 -20.74793603788134, -41.95057702256418 -20.74797809290242, -41.95048854871922 -20.748100275784445, -41.950427114752074 -20.74824087367159, -41.9503829333622 -20.74838375867987, -41.95033707811332 -20.74852479316568, -41.950251795638124 -20.74868968262353, -41.950393750044455 -20.749097093861685, -41.950493074337835 -20.749355907741293, -41.95055512830951 -20.749594446078557, -41.950585694279525 -20.749763039575818, -41.950622288690795 -20.749891193089596, -41.95064621334067 -20.749961137771606, -41.948881140189904 -20.749961137771606, -41.948881140189904 -20.749882191684534, -41.948657684268696 -20.749882191684534, -41.948434228347494 -20.749882191684534, -41.948210772426286 -20.749882191684534, -41.947987316505085 -20.749882191684534, -41.947987316505085 -20.749961137771606, -41.94361201651903 -20.749961137771606, -41.943612681675326 -20.749960437644727, -41.9436671247638 -20.74991374773251, -41.94366575070074 -20.74987499982115, -41.94359549074266 -20.749836672058077, -41.94348176984038 -20.749788250861858, -41.94338816184663 -20.749706549187028, -41.943381916107285 -20.74960653549553, -41.943439905386924 -20.749487110927777, -41.94352543224687 -20.749364623030225, -41.94361469372143 -20.74926662977051, -41.94367984481196 -20.74918055462423, -41.94377700698348 -20.74910218389918, -41.943928206209485 -20.749035220400224, -41.944068063144876 -20.748950538298534, -41.944165194018176 -20.74887715946704, -41.944226744342494 -20.748809923609585, -41.944263326706015 -20.74873937108719, -41.944280896714474 -20.748659331365435, -41.9443069374751 -20.74859164710514, -41.94434707278255 -20.74855307067262, -41.94435223189142 -20.74850408643681, -41.94429098254172 -20.74844358616218, -41.94420624288666 -20.748370797997726, -41.94412861431618 -20.74829135486827, -41.94405438303921 -20.748214413759648, -41.94397484646562 -20.748132978278136, -41.943888499273335 -20.748053581536812, -41.9438179210826 -20.74800675452593, -41.94374049488676 -20.74802242276732, -41.94365600131445 -20.74809616340634, -41.94356217034085 -20.748124874071117, -41.94346343936302 -20.748094805791386, -41.94336610648409 -20.748022889819183, -41.943313883977225 -20.74791566504338, -41.94328204734813 -20.74780375038301, -41.9432650523657 -20.74769854868718, -41.943291310313114 -20.747631278235477, -41.943354181403976 -20.747603097225454, -41.943431403279874 -20.747563716663215, -41.94348238055401 -20.74749375821391, -41.94348894229824 -20.747405596594533, -41.94348528813912 -20.74730952655944, -41.94345902897508 -20.74721666349857, -41.94342853272953 -20.74711745636761, -41.943412252096444 -20.74700816123234, -41.94343781573217 -20.746889514712286, -41.94350044522878 -20.746812230615525, -41.943576470317154 -20.746799536930343, -41.94365128715388 -20.746814845708435, -41.94370095759435 -20.746829406191317, -41.94370955685079 -20.746864754960733, -41.943683388629026 -20.74690614528919, -41.94369098573601 -20.74694303676199, -41.94376495804592 -20.746963998542007, -41.94388405803699 -20.746965238205583, -41.94401322737733 -20.74690950137304, -41.944134890693036 -20.746822094092025, -41.94422861960578 -20.74672376337409, -41.944270092459185 -20.74663098807162, -41.94431707681212 -20.746693200527382, -41.944351935351186 -20.74673755105806, -41.94437540903466 -20.74680303438075, -41.944395631901145 -20.746882220721087, -41.944427949256 -20.746969355521305, -41.94449153145028 -20.74704314179073, -41.94458987170798 -20.747060313514623, -41.94471092593336 -20.74703002963144, -41.9448487025288 -20.746964796180333, -41.94497125321067 -20.74687809687252, -41.94509479098694 -20.74679523281752, -41.94517417788411 -20.74672106970542, -41.9452730959961 -20.746673137291875, -41.94541032837139 -20.746648387330563, -41.94553836429259 -20.74660466510474, -41.94566873615928 -20.74655177458675, -41.94581934714106 -20.74651759586717, -41.945974961232416 -20.74649028607014, -41.94611393298451 -20.746451234028616, -41.946222934440996 -20.746380010803964, -41.94631239184918 -20.746282244408462, -41.94636684196464 -20.74617353028012, -41.94638505981512 -20.74607108473948, -41.94637661565588 -20.745974103479107, -41.94640472055917 -20.745922189715266, -41.946469967173755 -20.745892176667294, -41.94653058189316 -20.745837486391427, -41.94657486643632 -20.745765017339885, -41.94661110346565 -20.74570648530294, -41.94662764296135 -20.745645033119764, -41.94659772183229 -20.745555880977776, -41.946531098017545 -20.7454482235647, -41.94646947820999 -20.745369023949632, -41.94640560867036 -20.74533025622937, -41.946343243021424 -20.745310136611728, -41.9463132487179 -20.745267263637732, -41.94628362405354 -20.74520640789611, -41.94619669610436 -20.74513659007396, -41.94603867241288 -20.745059062327012, -41.94583668897749 -20.744949683332063, -41.945587195350484 -20.74479815433651, -41.94535918054874 -20.74470895508801, -41.94514501854931 -20.74464547729386, -41.944990554929156 -20.74457088693721, -41.94485970992742 -20.74448247963156, -41.94475912149377 -20.74438310217094, -41.944680887089305 -20.744325090426873, -41.944626669171335 -20.744266496094053, -41.94457503648202 -20.744197172069832, -41.94450802905069 -20.744143344440857, -41.94445602076802 -20.744083476000753, -41.944447274664405 -20.74397240713594, -41.94449399290183 -20.743848165984833, -41.94451723495984 -20.743733989255755, -41.94451325769367 -20.743631228242332, -41.94447636270051 -20.74358946995627, -41.94441401833816 -20.743588452778404, -41.94433105738472 -20.743601905658537, -41.944247981403706 -20.743620071009225, -41.94416626776815 -20.743618880354877, -41.94409833935896 -20.743563522690998, -41.944071010426384 -20.74346347334605, -41.94407425942535 -20.74331767117018, -41.94407713984099 -20.743150137614766, -41.94409644854213 -20.743005005680082, -41.94411847951878 -20.74289499407714, -41.94411512925212 -20.742811483112632, -41.944096156780645 -20.742739977554574, -41.94407047492638 -20.742682221080727, -41.94407969925641 -20.742638479706084, -41.944120930962136 -20.742598238716916, -41.944181409174064 -20.74254782892415, -41.944240883029 -20.74250010503045, -41.94430542596754 -20.74244574072161, -41.94438022073589 -20.74237120076446, -41.944455800645436 -20.74225164760689, -41.94451435778664 -20.742109108207025, -41.944561638375774 -20.741994334487323, -41.94455618333194 -20.741904922133546, -41.944516398697296 -20.741819637236993, -41.94450062916209 -20.741735968112295, -41.9445198806933 -20.74167108390008, -41.944556077322275 -20.741631998386623, -41.944578825991 -20.74157898418268, -41.944561950250275 -20.741519239378007, -41.94451443721904 -20.741458314388176, -41.94444333857763 -20.74139396864723, -41.9443576436275 -20.741332171132303, -41.944263893056615 -20.741261574242557, -41.9441713165087 -20.741208992177008, -41.94408225787263 -20.741242046783572, -41.94401943387449 -20.741362512292955, -41.94398753511417 -20.741524437245616, -41.943956578872374 -20.741671348087284, -41.94392382707843 -20.741802365008443, -41.94388441228293 -20.741917340052403, -41.943867679355556 -20.742015341094387, -41.943855791434636 -20.742095865439293, -41.94382482280053 -20.742170091031596, -41.943789234315076 -20.7422391214652, -41.94373991015721 -20.742304620750946, -41.94367376232482 -20.74234956546871, -41.943587328955225 -20.74229677664473, -41.943487063245435 -20.74217371763474, -41.94339249750312 -20.741993558862173, -41.943329793357684 -20.741835699898978, -41.94330003172604 -20.74170476599495, -41.94328971817246 -20.741599127336176, -41.94330040117784 -20.741513905605736, -41.94330353843991 -20.741441225133364, -41.94330839135721 -20.741364006179317, -41.94331516687493 -20.74126695781424, -41.94331528281027 -20.741151996326565, -41.94327040509331 -20.741051277062493, -41.943205198428274 -20.740962602790688, -41.943158422630894 -20.740882990775305, -41.94309057713033 -20.740848714600148, -41.94299216692257 -20.740888229885414, -41.942870757376674 -20.740977982086125, -41.94274100907882 -20.741045213419202, -41.942619786695346 -20.74108813295137, -41.94251795901359 -20.741115829581762, -41.94242883251576 -20.741100999534712, -41.94235450354698 -20.741062883776628, -41.942268902224626 -20.74103112941195, -41.94217752119185 -20.741005071926995, -41.94210715538495 -20.740975341852902, -41.94203463632682 -20.740951482497223, -41.9419305079637 -20.74094717122455, -41.94180782569922 -20.740963460848125, -41.94166850905772 -20.741003994574445, -41.94150357641371 -20.741050725377974, -41.94133204314231 -20.741101440021314, -41.94114066381169 -20.741158250464814, -41.940948899618746 -20.741225809771215, -41.94072759418318 -20.74129573585503, -41.940515311918176 -20.74136345719006, -41.940315347952925 -20.741399139992183, -41.940147066704576 -20.741422742227034, -41.93997466026049 -20.741423229144672, -41.93978784056955 -20.74142776598479, -41.93960233540315 -20.74143976174876, -41.93945210510451 -20.74141806315994, -41.939335551243246 -20.74136090979899, -41.939239190452646 -20.741281864116637, -41.93915389597922 -20.74118200842529, -41.93906075839106 -20.74104471509989, -41.93895405947506 -20.740869621579506, -41.93883315235167 -20.740729110946884, -41.93870139856494 -20.740615876302947, -41.938568279712605 -20.740509128351498, -41.93844402629042 -20.740422646477864, -41.938330489106924 -20.740352302019897, -41.93821651257363 -20.74030071453363, -41.938097702440984 -20.740256477074137, -41.937956899967055 -20.740209801022168, -41.93783827587138 -20.740167632585212, -41.93772193984511 -20.740143203951746, -41.937579654372186 -20.74015265589701, -41.9375 -20.74017025376085, -41.9375 -20.73272552877246, -41.937557079443316 -20.732696532850984, -41.93765679448202 -20.732600415210307, -41.937762879190494 -20.73247817362888, -41.93787764545762 -20.7323410665017, -41.93801359209108 -20.732225791895086, -41.9381433027758 -20.73212538750994, -41.93828118281524 -20.73201499624751, -41.938413513278526 -20.73187974182799";
    wkt9 += ", -41.93851733848549 -20.73176524063091, -41.93861280106887 -20.731659375965457, -41.93874489687441 -20.731543644174906, -41.938926636977484 -20.731382757595362, -41.93913207063529 -20.731215580001344, -41.9393709025491 -20.731039735379927, -41.93964823087272 -20.73091443405122, -41.93992844280628 -20.730886710253774, -41.940172392135985 -20.73090684160228, -41.94037515593801 -20.73090517375977, -41.940519747021774 -20.730842954540798, -41.94060825286609 -20.730729419302826, -41.940645998854166 -20.73056010686608, -41.940610735002544 -20.73038508328589, -41.94044344191297 -20.730209958185483, -41.94007810342313 -20.730090139281405, -41.93891337037595 -20.73, -41.93876336592394 -20.730179125899035, -41.93863752470915 -20.730313434963833, -41.93849407482894 -20.730455292591934, -41.9383330317239 -20.73062180614697, -41.93812976368062 -20.730807712722807, -41.93791440947209 -20.730980208684276, -41.937711525645 -20.731170858136974, -41.9375 -20.73138470792116, -41.9375 -20.73, -41.93891337037595 -20.73, -41.953125 -20.73, -41.953125 -20.739152159375674, -41.95311187657608 -20.7389684314408, -41.95308492711755 -20.738833684148183, -41.95295017982494 -20.738725886314086, -41.95257288740561 -20.738698936855563, -41.95243814011297 -20.738725886314086, -41.95214169606921 -20.738941481982277, -41.951926100401025 -20.738941481982273, -41.95179135310841 -20.738806734689657, -41.95173745419136 -20.738321644436233, -41.951602706898754 -20.738186897143613, -41.95136016177204 -20.73815994768509, -41.95095591989419 -20.73815994768509, -41.95068642530896 -20.73815994768509, -41.95065947585043 -20.73853724010442, -41.95052472855782 -20.738671987397037, -41.95014743613849 -20.738698936855563, -41.95014743613849 -20.73871644229046, -41.94999962164902 -20.73860786730371, -41.94984368410197 -20.73839637992054, -41.949798820088105 -20.738164668336818, -41.9498325006474 -20.73796148526725, -41.94989034257867 -20.737795915214157, -41.94995357559843 -20.73766130972839, -41.95001046036902 -20.737557132896427, -41.95007949678779 -20.73746044575437, -41.95017604827289 -20.737347171346915, -41.95025666411379 -20.737218359143444, -41.950301732165 -20.73708677992521, -41.95028593250382 -20.73698056719294, -41.95022829501967 -20.736947063335904, -41.9501513522283 -20.736959388032584, -41.950075820376426 -20.736964141444535, -41.95005178811069 -20.736945399048036, -41.950062880729874 -20.736912509912152, -41.950080245700605 -20.73687596752544, -41.950074112625245 -20.736849109755095, -41.95001287796781 -20.736851153465167, -41.94993516707695 -20.736874345829367, -41.94985715088083 -20.736906273505966, -41.94982020760238 -20.736941102540364, -41.94981493861228 -20.73698094572404, -41.94978560298218 -20.73702656911889, -41.949706221624446 -20.737056538091917, -41.949616011809034 -20.737054308511414, -41.94953088793614 -20.73704498329249, -41.94940356686747 -20.737062967222435, -41.9492110929898 -20.737102294105085, -41.94898549322191 -20.73711266659249, -41.94876127013006 -20.73708463564, -41.94859396841295 -20.737039021221715, -41.94846141496736 -20.736985798537003, -41.94834841843248 -20.736913594828284, -41.948217019453246 -20.736837168899232, -41.94809063831901 -20.73677529957905, -41.947962938045926 -20.73671453660134, -41.947819436828325 -20.73666139206048, -41.94766409251401 -20.736614860904297, -41.94747309231954 -20.73657252405938, -41.94722916438529 -20.73652270083801, -41.9469850531803 -20.736479327988658, -41.946715444595256 -20.7364817032188, -41.946456019706496 -20.73650521873676, -41.94624587141071 -20.736549640488473, -41.9461100911473 -20.736627440148215, -41.94603283740767 -20.736712465401883, -41.94602643314682 -20.736781089461694, -41.94607562339845 -20.736821142928928, -41.946120389442704 -20.736850930665632, -41.946090806931245 -20.736908141839024, -41.945975734622735 -20.73696818056279, -41.945785624139724 -20.73700714326337, -41.94557799005502 -20.73701929634262, -41.94536442612035 -20.7370517518318, -41.94511851043623 -20.7370792644866, -41.944851452857215 -20.737119813654733, -41.94459054616438 -20.737196234662118, -41.94436994362014 -20.737265344040978, -41.944243687425875 -20.73723661318732, -41.94427496267218 -20.737117921364295, -41.94444734383826 -20.736999454609528, -41.94470042277729 -20.736905835866192, -41.944978572295746 -20.736824811374184, -41.94522820955138 -20.736722410096547, -41.94542135069692 -20.736596313562767, -41.94555009252404 -20.73644057377464, -41.94561211157749 -20.736315187267085, -41.94558381109941 -20.73626440231921, -41.94548896860284 -20.736286732471232, -41.94538076502919 -20.73633305248841, -41.94527267558607 -20.73638734841522, -41.94516068514345 -20.736445212327965, -41.94503291728353 -20.736501268098287, -41.94487220468627 -20.736478710383427, -41.94468817662229 -20.736388764451725, -41.94447303991354 -20.736297938716465, -41.94433290722883 -20.73628599975373, -41.94424814220099 -20.73631677659661, -41.94418153275983 -20.73636145466538, -41.9441288224578 -20.736407935470314, -41.94408144025461 -20.736444951094697, -41.944018944276436 -20.736495435282393, -41.943831465253275 -20.736531395927525, -41.943731127905814 -20.736569468038653, -41.943675145541135 -20.736656397696763, -41.9436555717986 -20.736773829766136, -41.94366024312239 -20.73690064669087, -41.94366911186354 -20.73705234462403, -41.94370850255036 -20.737241215868906, -41.943764259917934 -20.73742446856093, -41.943847361953495 -20.737612858760638, -41.94389539253048 -20.73774330507509, -41.943870736019626 -20.737840477355604, -41.943821525149225 -20.737908161877208, -41.94378980963386 -20.737974016002045, -41.94381092161466 -20.738044480317704, -41.94383642489551 -20.738140215122865, -41.943860426527685 -20.738242485772535, -41.94387828793779 -20.738337545063054, -41.94386064401006 -20.73842123689266, -41.943798255438935 -20.738483130259993, -41.94370426644226 -20.738511099991136, -41.94359444651979 -20.73849727015438, -41.94354461310725 -20.738453332473085, -41.94354057448263 -20.738417568581248, -41.943522071542134 -20.738403818328155, -41.94347008396995 -20.73845899080419, -41.943455455537475 -20.738575597253643, -41.94350836902745 -20.738686748828762, -41.943575383546396 -20.738779785095925, -41.94369505603995 -20.738861988527326, -41.94382381158804 -20.738936366426646, -41.943920452554345 -20.73897154062189, -41.943988123688875 -20.739019477984847, -41.94404874262372 -20.73910326522422, -41.94412257466189 -20.73917100240538, -41.94418243905924 -20.739230806007996, -41.944244529323406 -20.73929459985731, -41.944323835691186 -20.73937535039252, -41.94440609685969 -20.739463275864107, -41.944495756039345 -20.739531001416527, -41.94458345835424 -20.739598433037013, -41.944657787453046 -20.739663635979056, -41.944746688422114 -20.739728489109712, -41.94485108161114 -20.739812442149358, -41.94496612140961 -20.739886103668542, -41.94507823266469 -20.73994100513198, -41.945186317650915 -20.739963264930026, -41.9452709459944 -20.739981731249394, -41.94534085459224 -20.739972328917673, -41.94540773993628 -20.73995066714289, -41.94547414724391 -20.739936707022164, -41.94552661143599 -20.73996053950473, -41.945523821400684 -20.740017383541822, -41.94549663518531 -20.740083757619313, -41.94551082898771 -20.740129947351576, -41.94556572221644 -20.740145102075306, -41.94563391728066 -20.740161368010735, -41.945707396933955 -20.740195448853502, -41.945798011158075 -20.740231973558714, -41.945894627794246 -20.740248719325503, -41.94600503598372 -20.740249545364442, -41.94614151322918 -20.740237204266332, -41.94629646183238 -20.740232204150974, -41.94644082733387 -20.74022193844298, -41.94655596319295 -20.740202642027317, -41.946615104036894 -20.740158115150404, -41.946651992196124 -20.74011116359937, -41.94669868677075 -20.740071510179437, -41.94675273415901 -20.740073538090808, -41.94682665502535 -20.740116686561617, -41.94693087432302 -20.740185003075574, -41.94706616480039 -20.740247025534543, -41.947201904229836 -20.740327171510263, -41.94735240818785 -20.740424927042373, -41.947452490286125 -20.740494361495386, -41.947452490286125 -20.740585398952213, -41.947829782705455 -20.740612348410735, -41.94796452999806 -20.740720146244833, -41.9479914794566 -20.74085489353745, -41.947994970391036 -20.740903766619606, -41.94797621510555 -20.740954039135623, -41.94798682012671 -20.741003082868232, -41.9480026066359 -20.741010674047775, -41.94801842891512 -20.741232185956783, -41.94815317620773 -20.7413669332494, -41.94839572133445 -20.74139388270792, -41.94952488814706 -20.74139388270792, -41.94960362123205 -20.741457519472505, -41.949722443192535 -20.74152321686949, -41.94985355246338 -20.74157085027983, -41.949938662506156 -20.741596097725996, -41.95012048667995 -20.741798124585777, -41.95014743613849 -20.741932871878394, -41.9502821834311 -20.741905922419868, -41.9503899812652 -20.741798124585777, -41.95040638249956 -20.741716118413972, -41.950502690329294 -20.741720379593566, -41.9506716068189 -20.74167100943432, -41.950834088709726 -20.7416180419565, -41.951005002730646 -20.741583735625355, -41.95101625568493 -20.74158231037871, -41.95109066718682 -20.741636427834635, -41.95122541447943 -20.741663377293158, -41.95158518119218 -20.741663377293158, -41.951651435110094 -20.741693418195858, -41.95183220075195 -20.74176745172282, -41.952035803440104 -20.74183131462247, -41.95219467830902 -20.74187702207363, -41.95227983454197 -20.741881413375676, -41.95229965605237 -20.741824247024706, -41.95231019714117 -20.741755375165905, -41.952404487561125 -20.741673806727317, -41.95243372428901 -20.741663377293158, -41.953125 -20.741663377293158, -41.953125 -20.749961137771606)))";

    FixPolygon8(wkt4);
    FixPolygon8(wkt5);
    FixPolygon8(wkt6);
    FixPolygon8(wkt7);
    FixPolygon8(wkt8);
    FixPolygon8(wkt9);
  }


  TEST_F(ITsValidation, RemoveSliversGeos)
  {
    std::string wkt = "POLYGON ((0 0, 0 2, 10 2, 10 6, 15 6, 15 -4, 10 -4, 10 0, 0 0))";
    std::unique_ptr<te::gm::Geometry> geometryA(te::gm::WKTReader::read(wkt.c_str()));


    std::unique_ptr<geos::geom::Geometry> geosGeometry(te::gm::GEOSWriter::write(geometryA.get()));

    double tolerance = geos::operation::overlay::snap::GeometrySnapper::computeOverlaySnapTolerance(*geosGeometry.get());
    geos::operation::overlay::snap::GeometrySnapper::GeomPtr geosResult = geos::operation::overlay::snap::GeometrySnapper::snapToSelf(*geosGeometry.get(), 3, true);

    std::string resultWKT = geosResult->toString();
  }

  void RemoveSliversTest(const std::string& wkt, double tolerance, std::size_t expectedPoints)
  {
    std::unique_ptr<te::gm::Geometry> geometryA(te::gm::WKTReader::read(wkt.c_str()));

    te::gm::GeometryVector geometryVector;
    geometryVector.push_back(geometryA.get());

    te::gm::GeometryVector resultGeometryVector = te::vp::GeometrySnapper::removeSlivers(te::gm::ToConstVector(geometryVector), tolerance);

    ASSERT_FALSE(resultGeometryVector.empty());

    if (resultGeometryVector.at(0) != nullptr)
    {
      ASSERT_EQ(resultGeometryVector.at(0)->getNPoints(), expectedPoints);
    }

    te::common::FreeContents(resultGeometryVector);
  }

  TEST_F(ITsValidation, RemoveSliversTerralib)
  {
    RemoveSliversTest("POLYGON ((0 0, 0 2, 10 2, 10 6, 15 6, 15 -4, 10 -4, 10 0, 0 0))", 2.0, 6);
    RemoveSliversTest("POLYGON((0 0, 0 2, 2 2, 2 0, 0 0))", 2.0, 0);
    RemoveSliversTest("POLYGON ((0 0, 0 -4, -5 -4, -5 6, 0 6, 0 2, 10 2, 10 6, 15 6, 15 -4, 10 -4, 10 0, 0 0))", 2.0, 12);
    RemoveSliversTest("POLYGON ((0 0, 0 -4, -5 -4, -5 6, 0 6, 0 2, 10 2, 10 6, 15 6, 15 -4, 10 -4, 10 0, 0 0))", 3.0, 12);
    RemoveSliversTest("POLYGON ((0 0, 0 -4, -5 -4, -5 6, 0 6, 0 2, 10 2, 10 6, 15 6, 15 -4, 10 -4, 10 0, 0 0))", 5.0, 10);
    //RemoveSliversTest("MULTIPOLYGON (((-66.81199084185886 -8.654496203832796, -66.80495180736521 -8.650976686585977, -66.80253213925803 -8.655816022800353, -66.80275210908596 -8.656475932284131, -66.80341199657275 -8.656475954281115, -66.80363198839765 -8.656475932284131, -66.80539174702106 -8.656915871939983, -66.80583168667692 -8.657795751251689, -66.8053135034081 -8.659350301058128, -66.80604096140377 -8.65935030105808, -66.80604096140371 -8.660157301057438, -66.80657896140367 -8.660157301057522, -66.8065789614036 -8.661233301056805, -66.80667533281874 -8.6612333010568, -66.80694433225393 -8.661233301056782, -66.80711696140354 -8.661233301056772, -66.80711696140352 -8.661495871603396, -66.80711696140352 -8.661502301056599, -66.8074823311243 -8.661502301056574, -66.80765496140349 -8.661502301056563, -66.80765496140347 -8.661764870778242, -66.80765496140347 -8.661771301056385, -66.80796701334815 -8.661771301056367, -66.80803138495618 -8.661755208154359, -66.81022090049781 -8.657741096328037, -66.809868597544 -8.65832826791773, -66.8082482080007 -8.65832826791773, -66.80801672378024 -8.656476394153966, -66.80662781845741 -8.656013425713025, -66.8082482080007 -8.652772646626438, -66.81183153430872 -8.654788267674697, -66.81199084185886 -8.654496203832796),(-66.80604096140371 -8.660157301057438, -66.80604096140371 -8.66011257274107, -66.8061483093631 -8.660157301057438, -66.80604096140371 -8.660157301057438)))", 0, 33);
    RemoveSliversTest("MULTIPOLYGON (((-66.28592467107359 -9.538236084631137, -66.28592467107359 -9.538236084631137, -66.28592467107359 -9.537908206831275, -66.2859246710736 -9.535924372505992, -66.28493393730562 -9.536254617095317, -66.28361295894832 -9.53658486168464, -66.28427344812695 -9.538236084631137, -66.28592467107359 -9.538236084631137, -66.28592467107359 -9.538236084631137),(-66.28379599639544 -9.537042455302412, -66.28392572728245 -9.537172186189423, -66.28410973244291 -9.537172186189423, -66.28410973244291 -9.537356191349886, -66.28429373760343 -9.537540196510294, -66.2844777427639 -9.537540196510294, -66.2844777427639 -9.537724201670812, -66.28466174792425 -9.537724201670812, -66.28495615618112 -9.537871405799194, -66.28502975824529 -9.538092211991739, -66.28558177372679 -9.538092211991739, -66.28558177372679 -9.537908206831275, -66.28592467107359 -9.537908206831275, -66.28592467107359 -9.538236084631137, -66.28427344812695 -9.538236084631137, -66.28427344812695 -9.538236084631137, -66.28379599639544 -9.537042455302412)))", 0, 17);
  }

  TEST_F(ITsValidation, RemoveSliversTerralib2)
  {
    std::string wkt = "multipolygon(((-64.1451981793532298 3.55296681855199648, -64.1451981793532298 3.57002712739561545, -64.1450918067335323 3.57002712739561545, -64.1450918067335323 3.56991904724431652, -64.1450205926935695 3.56987987952233476, -64.1449340157781194 3.56983226221883632, -64.1448969415567518 3.56983226221883632, -64.1448969415567518 3.56981187139708744, -64.1442254151352813 3.56944253186527805, -64.1439226156728637 3.56944253186527805, -64.1438738370612214 3.56936936394781235, -64.1432828936233506 3.56936936394781235, -64.1432828936233506 3.56924666279269376, -64.1434622515854613 3.56924666279269376, -64.1434622515854613 3.56906730483057677, -64.1436416095475863 3.56888794686845978, -64.1438209675096971 3.56870858890634279, -64.1440003254718079 3.56852923094422581, -64.1441796834339328 3.56834987298210882, -64.1443590413960436 3.56817051501999138, -64.1445383993581686 3.5679911570578744, -64.1447177573202794 3.56781179909575741, -64.1448971152824043 3.56781179909575741, -64.1448971152824043 3.56772508189847093, -64.1450179233212907 3.5677249309444643, -64.1447478551621373 3.56772526840438031, -64.1447475193541408 3.5674539140783379, -64.144477451248477 3.56745425143296746, -64.1444771155453566 3.56718289707956604, -64.1444767798679436 3.56691154272453037, -64.1444764442162381 3.5666401883678609, -64.1442063763223587 3.56664052556578071, -64.1442060407755008 3.56636917118175845, -64.1442057052543362 3.56609781679610194, -64.1439356374930867 3.56609815386305984, -64.143665569706144 3.56609849085049202, -64.143395501893508 3.56609882775839937, -64.1431254340552215 3.56609916458678056, -64.1428553661912417 3.56609950133563647, -64.1426822505643059 3.56609971714330687, -64.1426822656591042 3.56609971712448992, -64.1426822665188041 3.56634902966981304, -64.1426822415900517 3.56609971715449436, -64.1425852983016114 3.56609983800496666, -64.1426559905859506 3.5660997498792586, -64.1410292575303345 3.56772653716020249, -64.1425654617748791 3.56619028170800689, -64.1425654617748791 3.56619757743670363, -64.1425654617748791 3.56637693539882061, -64.1425654617748791 3.56655629336093805, -64.1425654617748791 3.56673565132305503, -64.1425654617748791 3.56691500928517202, -64.1425654617748791 3.56709436724728901, -64.1423861038127541 3.567273725209406, -64.1423861038127541 3.56745308317152343, -64.1423861038127541 3.56763244113364042, -64.1423861038127541 3.56781179909575741, -64.1423861038127541 3.5679911570578744, -64.1422067458506433 3.56817051501999138, -64.1420273878885325 3.56834987298210882, -64.1420273878885325 3.56852923094422581, -64.1418480299264075 3.56852923094422581, -64.1418480299264075 3.56870858890634279, -64.1416686719642968 3.56870858890634279, -64.1414893140021718 3.56888794686845978, -64.141309956040061 3.56906730483057677, -64.14113061601374 3.56906732276637317, -64.141029229980191 3.5687631697348996, -64.141029229980191 3.56882851560236425, -64.141029229980191 3.56910401710044223, -64.1413047314782716 3.56937951859852021, -64.1413047314782716 3.56973836072058504, -64.1412973950113638 3.56974202895403891, -64.1409247300051391 3.5699283614571522, -64.1405520649989143 3.57011469396026548, -64.1403657324957948 3.57030102646337877, -64.1401793999926895 3.57030102646337877, -64.1401793999926895 3.57119631845628938, -64.1409996380211709 3.57119631845628938, -64.1409996380211709 3.57139118363306851, -64.1417790987282928 3.57139118363306851, -64.1419739639050732 3.57158604880984765, -64.1421688290818537 3.57158604880984765, -64.1421688290818537 3.57178091398662678, -64.1423636942586342 3.57178091398662678, -64.1423636942586342 3.57179856698692122, -64.1423636942586342 3.57197577916340592, -64.1425585594354004 3.57197577916340592, -64.1425585594354004 3.57217064434018461, -64.1427534246121809 3.57217064434018461, -64.1427534246121809 3.57236550951696374, -64.1431331824776265 3.57236550951696374, -64.1439226156728637 3.57236550951696374, -64.1439226156728637 3.57253701650073907, -64.1444650475642959 3.57253701650073907, -64.1444650475642959 3.57256037469374288, -64.1450918067335323 3.57256037469374288, -64.1450918067335323 3.57272334900385236, -64.1451981793532298 3.57272334900385236, -64.1451981793532298 3.58362832807414966, -64.1412566261937798 3.58362832807414966, -64.141256427765839 3.58346821562181761, -64.1412566261937656 3.58362832807414966, -64.1409074366366809 3.58362832807414966, -64.1390851976311609 3.58319957415217605, -64.1388254332019585 3.58319989845089504, -64.1385553599296827 3.58320023553958356, -64.1385550244855267 3.58292888071676607, -64.1385552049578962 3.58307487237782496, -64.1377232395534094 3.58287911955448868, -64.13697398108485 3.58362832807414966, -64.1006565550055711 3.58362832807414966, -64.1006565550055711 3.58026179905184838, -64.1007412861745252 3.58026169691702023, -64.1007409621365554 3.57999033855325699, -64.1007403141347254 3.57944762182081755, -64.1010103898147037 3.57944729626604907, -64.1010100657714332 3.57917593792211441, -64.1010097417528755 3.57890457957654284, -64.1012798172492353 3.57890425399143997, -64.10127949317598 3.57863289566898057, -64.1012792875608426 3.57846071350074002, -64.1012791691274515 3.5783615373448816, -64.1018193197282358 3.57836088603425972, -64.1020893949915234 3.57836056025924254, -64.1020890707294484 3.57808920200778591, -64.1023591458885704 3.57808887617773674, -64.1026292210229229 3.57808855026788963, -64.1026288966267686 3.57781719206434889, -64.1026279235868799 3.5770031174439052, -64.102627599289761 3.57673175923381681, -64.1023575245524171 3.57673208501977324, -64.1018173750035061 3.57673273635238065, -64.1018170509693164 3.57646137806633924, -64.101816726959882 3.57619001977866269, -64.1015466523070785 3.57619034527579505, -64.1015437380496849 3.57374812039024459, -64.1020838859022319 3.57374746976196089, -64.1023539597913867 3.57374714432826757, -64.1023536358707418 3.57347578607377026, -64.1023523404356723 3.57239035303944208, -64.1042428541905025 3.5723880736398379, -64.1045129274849046 3.57238774769264156, -64.1045126030537062 3.57211638962836897, -64.1045113055769349 3.57103095735492948, -64.1045109812697547 3.57075959928248521, -64.1042409084510041 3.57075992508075846, -64.1034306898459647 3.57076090199777285, -64.103160616928065 3.57076122747751068, -64.1031609408389471 3.57103258567399706, -64.1028908678169955 3.57103291109888854, -64.1026207947703028 3.57103323644414283, -64.1026211185474182 3.57130459468857042, -64.1026220900273671 3.57211866941204859, -64.1006565550055711 3.57212103502532541, -64.1006565550055711 3.57049288263196241, -64.1007296363799952 3.5704927947905909, -64.1009997094413819 3.57049247005230397, -64.1009993862140277 3.57022111165439737, -64.1012694591714478 3.57022078686122901, -64.101269179589579 3.56998610966720209, -64.1006565550055711 3.56998610966720209, -64.1006565550055711 3.55296681855199648, -64.1451981793532298 3.55296681855199648), (-64.1018091105973724 3.56980532454323507, -64.1017681342339642 3.56980532454323507, -64.1016244577027834 3.56994900107442081, -64.1018092815717466 3.56994877871074889, -64.1018091105973724 3.56980532454323507), (-64.1090338280357628 3.57143239065893736, -64.1096431754630629 3.57143239065893736, -64.1096430117577967 3.57129610918266405, -64.1100402282328332 3.57129562723856964, -64.1100842499364347 3.57125160553497079, -64.1100842499364347 3.57052846503910271, -64.110265035060408 3.57034767991513569, -64.110445820184367 3.57016689479116911, -64.110992018472615 3.57016689479116911, -64.1109907647325628 3.56912361134993583, -64.1109904386977973 3.56885225386356719, -64.1107203670306234 3.56885258139828299, -64.1104502953385094 3.568852908853406, -64.1104499694871066 3.56858155131547772, -64.1101798978492781 3.56858187866605814, -64.1101792463797011 3.56803916353539385, -64.1101789206823014 3.56776780596761256, -64.1099088492571241 3.56776813316376451, -64.1096387778070351 3.56776846028034988, -64.1096384522929412 3.56749710266104714, -64.1093683808971093 3.56749742967312944, -64.1093680554871241 3.56722607202725728";
    wkt += ",-64.1093677301020364 3.56695471437975442, -64.1090976588396444 3.56695504126241891, -64.1090973335586511 3.56668368358835153, -64.1090970083025411 3.56641232591265389, -64.1090338280357628 3.56641240235310386, -64.108826937173589 3.56641265266592233, -64.1088262868944128 3.5658699372597833, -64.1088259617921636 3.56559857955426596, -64.1085558908757633 3.56559890615324626, -64.1085555658775661 3.56532754842117949, -64.1085552409042521 3.56505619068748114, -64.1082851701212206 3.56505651715712002, -64.1082845204075369 3.5645138016350093, -64.1082841955880269 3.56424244387150679, -64.108014125017462 3.56424277018692326, -64.1080134756113011 3.56370005460521488, -64.1080131509455384 3.56342869681191265, -64.1077430805874116 3.5634290229731449, -64.1061226579164156 3.5634309782716751, -64.1058525873842768 3.56343130387661944, -64.1058529114172728 3.56370266186894291, -64.104772628724092 3.56370396359338404, -64.1045025579887664 3.56370428882580503, -64.1045028816510722 3.56397564694068869, -64.1035759854736256 3.56397676255716167, -64.1035759854736256 3.56777578262268813, -64.1047774874502352 3.56777433477297024, -64.1047778115637641 3.56804569283855288, -64.1047781357021194 3.56831705090250439, -64.105048207758216 3.56831672516845755, -64.1050485320005805 3.56858808320594356, -64.1053186041110479 3.56858775736747935, -64.1063988923046679 3.56858645321776446, -64.10640149030624 3.57075731646482586, -64.1064018151682404 3.5710286743633568, -64.1066718878676625 3.57102834790314416, -64.1066722128337716 3.57129970577517142, -64.1066725378247355 3.5715710636455662, -64.1069426106578533 3.57157073705596373, -64.1069448868455112 3.57347024192882756, -64.1061346666064225 3.57347122198093059, -64.105864593143707 3.57347154850557658, -64.1058649180955342 3.5737429064374826, -64.1058658930999883 3.57455698022338808, -64.1055958192951181 3.57455730676774319, -64.1055961442669258 3.57482866471795724, -64.1055980946191681 3.57645681238488944, -64.1055984197647746 3.57672817032365176, -64.1058684942045716 3.57672784358048634, -64.1058688194544004 3.57699920149276185, -64.1058691447290556 3.57727055940339911, -64.1061392193028041 3.57727023253075282, -64.1074895917988101 3.57726859697083288, -64.1074886146202658 3.57645452368171268, -64.1077586888067117 3.57645419640507223, -64.1077583630508911 3.5761828386636525, -64.1077557578998523 3.57401197667339598, -64.1088187540686647 3.57401068831169777, -64.1088187540686647 3.57161317578290438, -64.1089995391926379 3.57143239065893736, -64.1090338280357628 3.57143239065893736), (-64.1158563111628581 3.57264534341541351, -64.1158559833764627 3.57237398640068982, -64.1158556556151069 3.57210262938432965, -64.115855327878819 3.57183127236633435, -64.1158550001675991 3.57155991534670481, -64.1158546724814187 3.57128855832544101, -64.1158543448203062 3.57101720130254119, -64.1158540171842475 3.57074584427800845, -64.1158536895732425 3.57047448725183925, -64.1161237612407433 3.57047415805442236, -64.1161234335755523 3.57020280105170151, -64.1163935051387455 3.57020247179974559, -64.1163931774193685 3.56993111482047842, -64.1163928497250879 3.56965975783957656, -64.1163925220558468 3.5693884008570409, -64.1163921944117021 3.56911704387287054, -64.1163918667926254 3.56884568688706771, -64.1163915391986023 3.56857432989963064, -64.1163912116296473 3.56830297291056064, -64.1163908840857744 3.56803161591985729, -64.1161208131563427 3.56803194497111376, -64.1158507422018431 3.5680322739427992, -64.1155806712222898 3.5680326028349123, -64.1155803439410477 3.56776124576733089, -64.1155800166848735 3.56748988869811701, -64.1153099458386322 3.56749021746052053, -64.1150398749673514 3.56749054614336414, -64.1147698040710452 3.56749087474664828, -64.1144997331496853 3.5674912032703725, -64.1142296622033143 3.56749153171453681, -64.1139595912319038 3.56749186007914165, -64.1139599180129807 3.5677632172986784, -64.1139602448190686 3.56803457451658357, -64.1139605716501677 3.56830593173285493, -64.1139608985062637 3.56857728894749293, -64.1139612253874134 3.56884864616049757, -64.1139615522935458 3.56912000337186841, -64.1136914808219274 3.56912033180710964, -64.1134214093252979 3.56912066016275498, -64.1131513378036715 3.56912098843880354, -64.1131516644971668 3.56939234572362185, -64.112881592871318 3.56939267394508919, -64.1128819195105706 3.56966403125328746, -64.1126118477805136 3.56966435942016069, -64.1123417760254597 3.56966468750742516, -64.1120717042454231 3.56966501551508131, -64.1118016324404181 3.56966534344312825, -64.1118019587877228 3.56993670084969139, -64.1118022851587739 3.57020805722533474, -64.1112108414566251 3.57020877556755378, -64.1113497458042048 3.57034767991513569, -64.1118024531013475 3.57034767991513569, -64.1118026115572661 3.57047941565791449, -64.1118029379794763 3.57075077305957667, -64.1115328658325438 3.57075110100795223, -64.1112627936606287 3.57075142887669505, -64.1109927214637594 3.57075175666580513, -64.1107226492419642 3.57075208437528202, -64.1104525769952005 3.57075241200512572, -64.1104529030460952 3.57102376953001288, -64.1104532291219158 3.57129512705326668, -64.1104535552226764 3.57156648457488579, -64.1104538813483771 3.57183784209486976, -64.1104542074990178 3.57210919961322038, -64.1107242801421222 3.57210887185857651, -64.1107246063969853 3.5723802293503315, -64.1107249326767743 3.57265158684045137, -64.1107252589815175 3.57292294432893653, -64.1107255853112008 3.57319430181578612, -64.1107259116658525 3.57346565930100191, -64.1107262380454159 3.57373701678458078, -64.1107265644499194 3.57400837426652451, -64.1107268908793912 3.57427973174683267, -64.110727217333789 3.57455108922550524, -64.110727543813141 3.57482244670254223, -64.1107275263300522 3.57480791540661258, -64.1108042241763627 3.57480790773606083, -64.1109835506832866 3.57462858122913829, -64.1111628771902105 3.57462858122913829, -64.1113422036971343 3.57462858122913829, -64.1113422036971343 3.57444925472221531, -64.1115215302040582 3.57444925472221531, -64.1117008567109821 3.57444925472221531, -64.1117008567109821 3.57426992821529277, -64.1118801832179059 3.57426992821529277, -64.1120595097248298 3.57426992821529277, -64.1122388362317537 3.57409060170837023, -64.1124181627386633 3.57409060170837023, -64.1124181627386633 3.57391127520144725, -64.1125974892455872 3.57391127520144725, -64.1127768157525111 3.57373194869452471, -64.112956142259435 3.57355262218760217, -64.1131354687663588 3.57337329568067918, -64.1133147952732827 3.57337329568067918, -64.1133147952732827 3.57319396917375665, -64.1134941217802066 3.57319396917375665, -64.1136734482871304 3.57319396917375665, -64.1136734482871304 3.57301464266683411, -64.1138527747940543 3.57301464266683411, -64.1138527747940543 3.57283531615991112, -64.1140321013009782 3.57283531615991112, -64.1142114278079021 3.57265598965298858, -64.1143907543148259 3.57265598965298858, -64.1145700808217356 3.57247666314606604, -64.1147494073286595 3.57247666314606604, -64.1149287338355833 3.57247666314606604, -64.1152873868494311 3.57247666314606604, -64.1158253663702027 3.57247666314606604, -64.1158561074050937 3.57247666314606604, -64.1158563111628581 3.57264534341541351), (-64.1301576819611654 3.56150218999404267, -64.1300104044266703 3.5613549124595405, -64.1300104044266703 3.56117555449742307, -64.1300104044266703 3.56099619653530608, -64.1300104044266703 3.56081683857318909, -64.1300104044266703 3.5606374806110721, -64.1300104044266703 3.56045812264895511, -64.1300104044266703 3.56009940672472069, -64.1300104044266703 3.55992004876260371, -64.1300104044266703 3.55987423430453598, -64.1298844651548734 3.55987438928665645, -64.1296143978593847 3.55987472155473705, -64.1293443305385296 3.55987505374343005, -64.1290742631923081 3.559875385852735, -64.1290745936978084 3.56014674170645318, -64.1288045262472366 3.56014707376174266, -64.1288048566990625 3.5604184296392023, -64.128534789144112 3.560418761640463, -64.1285351195422635 3.56069011754165965, -64.1282650518829485 3.56069044948888047, -64.1279949841983097 3.56069078135669503, -64.1277249164883187 3.56069111314510378, -64.1274548487530183 3.56069144485410538, -64.1271847809924083 3.56069177648370161, -64.1269147132064603 3.56069210803389069, -64.1266446453952312 3.56069243950467351, -64.1263745775586926 3.56069277089605007, -64.1263742477927536 3.56042141479214402, -64.1261041800099463 3.56042174607880302, -64.1258341122018578 3.56042207728606108, -64.125833782619253 3.56015072112990927, -64.1255637148648958 3.56015105223246753, -64.1252936470852859 3.56015138325563152, -64.1250235792803807 3.5601517141994008, -64.1247535114502369 3.56015204506377536, -64.1244834435948263 3.5601523758487561, -64.124213375714163 3.56015270655434168, -64.1242137048226652 3.56042406286219837, -64.1239436368377369 3.56042439351365214, -64.1239481370993616 3.56042438800386174, -64.1239481370993616 3.5605634706600271, -64.1242236385974422 3.56083897215810508, -64.1242236385974422 3.56111447365618305, -64.1242236385974422 3.56138997515426103, -64.1242236385974422 3.56166547665233901, -64.1242236385974422 3.56194097815041655, -64.1242236385974422 3.56221647964849453, -64.1244991400955229 3.56221647964849453, -64.1244991400955229 3.56249198114657251, -64.1244991400955229 3.56276748264465049, -64.1244991400955229 3.56304298414272846, -64.1244991400955229 3.56331848564080644, -64.1247746415936035 3.56331848564080644, -64.1247746415936035 3.56359398713888442, -64.12505014309167 3.56359398713888442, -64.1252978282889359 3.56359401190492742, -64.1252976018478762 3.56340765762894751, -64.125297272145346 3.56313630144014182, -64.1255673407945181 3.56313597013875238, -64.1255670110381857 3.56286461397360865, -64.1258370795830075 3.56286428261805677, -64.1261071481025482 3.56286395118305066, -64.1263772165968078 3.56286361966858856, -64.1266472850657721 3.56286328807467179, -64.1269173535094268 3.56286295640130035, -64.1271874219278004 3.56286262464847381, -64.1274574903208503 3.56286229281619171, -64.1277275586885764 3.56286196090445539, -64.127997627030993 3.56286162891326397, -64.1279972965882479 3.56259027297440189, -64.1279969661708407 3.56231891703390868, -64.1282670343297667 3.56231858501397047, -64.128537102463369 3.56231825291458959, -64.1288071705716334 3.56231792073576603, -64.1288068399423565 3.56204656486972082, -64.1290769079462137 3.56204623263682674, -64.1293469759247046 3.56204590032449664, -64.1296170438778432 3.56204556793273008, -64.1296167130367536 3.56177421214118928, -64.1298867808854709 3.56177387969537618, -64.1298864499906927 3.56150252392759414, -64.1301565177349886 3.56150219142774782, -64.1301576819611654 3.56150218999404267), (-64.1404421776975084 3.58021297307141984, -64.1404418419783724 3.57994161840982539, -64.1404415062848443 3.57967026374659003, -64.1404411706169384 3.57939890908171598, -64.1404408349746404 3.57912755441520192, -64.1404404993579504 3.57885619974704738, -64.1401704275109239 3.57885653690513772, -64.1399003556383178 3.57885687398341723, -64.1396302837401038 3.57885721098188547, -64.1393602118162818 3.57885754790054289, -64.1390901398668944 3.5788578847393886, -64.1388200678919134 3.57885822149842348, -64.1388204030319855 3.57912957632025375, -64.138820738197623 3.579400931140444, -64.1388209455597149 3.579568801591547, -64.1388252179955742 3.57957307402740366, -64.1391007194936407 3.57984857552548164, -64.1391007194936407 3.58012407702355961, -64.1393762209917213 3.58012407702355961, -64.1393762209917213 3.5802143038506129, -64.1393618885668815 3.58021432173729082, -64.1396319608879253 3.58021398469058516, -64.1399020331833896 3.58021364756403848, -64.1401721054532459 3.58021331035764989, -64.1404421776975084 3.58021297307141984), (-64.1447498705499157 3.56935339432630583, -64.1450199391842943 3.56935305671200753, -64.1450196031428561 3.56908170242150424, -64.1450198775148976 3.569303258522313, -64.1449700169170995 3.56935311912011066, -64.1447498705499157 3.56935339432630583)))";
    RemoveSliversTest(wkt, 0, 468);
  }

  TEST_F(ITsValidation, CreatePolygonFromEnvelope)
  {
    te::gm::Envelope envelope(0, 0, 10, 10);
    te::gm::Polygon polygon = te::gm::CreatePolygon(envelope, 0, 1);

    ASSERT_EQ(polygon.getNumRings(), 1);

    ASSERT_TRUE(polygon.getRingN(0)->isValid());

    ASSERT_TRUE(polygon.getRingN(0)->isRing());

    te::gm::LineString* lineString = static_cast<te::gm::LineString*>(polygon.getRingN(0));

    ASSERT_EQ(lineString->size(), 9);
  }

  TEST_F(ITsValidation, Validation)
  {
    te::gtest::DataInfo redDataInfo = te::gtest::GetRedDataInfo();

    std::unique_ptr<te::da::DataSource> dataSource = te::da::DataSourceFactory::make("OGR", redDataInfo.m_uri);
    dataSource->open();

    std::unique_ptr<te::da::DataSetType> dataSetType = dataSource->getDataSetType(redDataInfo.m_datasetName);

    std::unique_ptr<te::da::DataSet> dataSet = dataSource->getDataSet(redDataInfo.m_datasetName);

    ASSERT_FALSE(dataSet == nullptr);

    std::size_t geomIdx = te::da::GetFirstSpatialPropertyPos(dataSet.get());

    //get data
    bool isValid = true;

    dataSet->moveBeforeFirst();

    while (dataSet->moveNext())
    {
      std::unique_ptr<te::gm::Geometry> geom = dataSet->getGeometry(geomIdx);
      if (geom->isValid() == true)
      {
        continue;
      }

#ifdef TERRALIB_GEOS_ENABLED
      try
      {
        te::gm::GeometryPtr geomFixed = te::gm::Validation::MakeValidWithSnap(geom.get());
        if (geomFixed->isValid() == false)
        {
          isValid = false;
          break;
        }
      }
      catch (const geos::util::GEOSException& e)
      {
        std::cout << e.what();
      }
#endif

    }

    ASSERT_TRUE(isValid);
  }

  TEST_F(ITsValidation, Snap)
  {
    te::gtest::DataInfo blueDataInfo = te::gtest::GetBlueDataInfo();

    std::unique_ptr<te::da::DataSource> dataSource = te::da::DataSourceFactory::make("OGR", blueDataInfo.m_uri);
    dataSource->open();

    std::unique_ptr<te::da::DataSetType> dataSetType = dataSource->getDataSetType(blueDataInfo.m_datasetName);

    std::unique_ptr<te::da::DataSet> dataSet = dataSource->getDataSet(blueDataInfo.m_datasetName);

    te::gm::GeometryProperty* geomProp = te::da::GetFirstGeomProperty(dataSetType.get());

    //get data
    std::unique_ptr<te::gm::GeometryCollection> geomCollection(new te::gm::GeometryCollection(0, te::gm::GeometryCollectionType, geomProp->getSRID()));

    dataSet->moveBeforeFirst();

    while (dataSet->moveNext())
    {
      std::unique_ptr<te::gm::Geometry> geom = dataSet->getGeometry(geomProp->getName());

      geomCollection->add(geom.release());
    }

    //snap
    bool result = true;

    std::unique_ptr<te::gm::Geometry> geomResult;

#ifdef TERRALIB_GEOS_ENABLED
    try
    {
      std::unique_ptr<geos::geom::Geometry> geomGeos(te::gm::GEOSWriter::write(static_cast<te::gm::Geometry*>(geomCollection.get())));

      double newTolerance = geos::operation::overlay::snap::GeometrySnapper::computeOverlaySnapTolerance(*geomGeos.get());

      geos::operation::overlay::snap::GeometrySnapper snapper(*geomGeos.get());

      std::unique_ptr<geos::geom::Geometry> resultGeom = snapper.snapToSelf(newTolerance, true);

      resultGeom->setSRID(geomCollection->getSRID());

      geomResult.reset(te::gm::GEOSReader::read(resultGeom.get()));
    }
    catch (geos::util::GEOSException& e)
    {
      std::cout << e.what();

      result = false;
    }
#endif

    te::gm::GeometryVector geometries;
    if (geomResult != nullptr)
    {
      te::gm::GeometryCollection* geomCollectionResult = dynamic_cast<te::gm::GeometryCollection*>(geomResult.get());
      if (geomCollectionResult)
      {
        std::size_t size = geomCollectionResult->getGeometries().size();
        for (std::size_t t = 0; t < size; ++t)
        {
          te::gm::Geometry* geomItem = geomCollectionResult->getGeometryN(t);

          std::vector<te::gm::Geometry*> inputGeomVec;
          te::gm::Multi2Single(geomItem, inputGeomVec);

          geometries.insert(geometries.end(), inputGeomVec.begin(), inputGeomVec.end());
        }
      }
    }

    //create output data
    te::gtest::DataInfo outputDataInfo = getOutputDataInfo(blueDataInfo.m_datasetName);
    std::string filePath = outputDataInfo.m_path;
    te::gtest::SimpleSaveGeometriesToShapefile(filePath, blueDataInfo.m_datasetName + "_validate_snap", te::gm::ToConstVector(geometries));

    te::common::FreeContents(geometries);

    ASSERT_TRUE(result);
  }

  TEST_F(ITsValidation, Prepare_Snap)
  {
    te::gtest::DataInfo blueDataInfo = te::gtest::GetBlueDataInfo();

    std::unique_ptr<te::da::DataSource> dataSource = te::da::DataSourceFactory::make("OGR", blueDataInfo.m_uri);
    dataSource->open();

    std::unique_ptr<te::da::DataSetType> dataSetType = dataSource->getDataSetType(blueDataInfo.m_datasetName);

    std::unique_ptr<te::da::DataSet> dataSet = dataSource->getDataSet(blueDataInfo.m_datasetName);

    te::gm::GeometryProperty* geomProp = te::da::GetFirstGeomProperty(dataSetType.get());

    //get data
    std::unique_ptr<te::sam::rtree::Index<std::size_t, 8> >rtree(new te::sam::rtree::Index<size_t, 8>);

    std::size_t index = 0;

    te::gm::GeometryVector geometries;

    dataSet->moveBeforeFirst();

    while (dataSet->moveNext())
    {
      std::unique_ptr<te::gm::Geometry> geom = dataSet->getGeometry(geomProp->getName());

      if (geom->isValid() == false)
        continue;

      double tolerance = te::gm::GetAppropriatePrecision(geom->getSRID());

#ifdef TERRALIB_GEOS_ENABLED
      std::unique_ptr<geos::geom::Geometry> geomGeos(te::gm::GEOSWriter::write(geom.get()));

      std::unique_ptr<geos::geom::Geometry> geomGeosBuffer(geomGeos->buffer(0));

      std::unique_ptr<geos::geom::Geometry> simplifiedGeom = geos::simplify::TopologyPreservingSimplifier::simplify(geomGeosBuffer.get(), tolerance);

      geos::operation::overlay::snap::GeometrySnapper snapper(*simplifiedGeom.get());

      double newTolerance = geos::operation::overlay::snap::GeometrySnapper::computeSizeBasedSnapTolerance(*simplifiedGeom.get());

      if (newTolerance > tolerance)
      {
        tolerance = newTolerance;
      }

      std::unique_ptr<geos::geom::Geometry> resultGeom = snapper.snapToSelf(tolerance, true);

      int geomSRID = geom->getSRID();
      geom.reset(te::gm::GEOSReader::read(resultGeom.get()));
      geom->setSRID(geomSRID);
#endif

      rtree->insert(*geom->getMBR(), index);

      geometries.push_back(geom.release());

      ++index;
    }

    for (int i = 0; i < 2; ++i)
    {
      for (std::size_t t = 0; t < geometries.size(); ++t)
      {
        te::gm::Geometry* geom = geometries[t];

        te::gm::Envelope geomMBR = *geom->getMBR();

        std::vector<std::size_t> report;
        rtree->search(*geom->getMBR(), report);

        std::vector<te::gm::Geometry*> reportGeom;

        for (std::size_t p = 0; p < report.size(); ++p)
        {
          if (report[p] == t)
          {
            continue;
          }

          te::gm::Geometry* candidate = geometries[report[p]];

          double tolerance = te::gm::GetAppropriatePrecision(candidate->getSRID());

          //snap geos
#ifdef TERRALIB_GEOS_ENABLED
          try
          {
            std::unique_ptr<geos::geom::Geometry> geomGeosA(te::gm::GEOSWriter::write(geom));

            std::unique_ptr<geos::geom::Geometry> geomGeosB(te::gm::GEOSWriter::write(candidate));

            geos::operation::overlay::snap::GeometrySnapper snapperA(*geomGeosA.get());

            double newTolerance = geos::operation::overlay::snap::GeometrySnapper::computeOverlaySnapTolerance(*geomGeosA.get(), *geomGeosB.get());

            if (newTolerance > tolerance)
            {
              tolerance = newTolerance;
            }

            std::unique_ptr<geos::geom::Geometry> resultGeomA = snapperA.snapTo(*geomGeosB.get(), newTolerance);

            int geomSRID = geom->getSRID();
            delete geom;
            geom = te::gm::GEOSReader::read(resultGeomA.get());
            geom->setSRID(geomSRID);
          }
          catch (geos::util::GEOSException& e)
          {
            throw te::common::Exception(e.what());
          }
#endif
          bool wasChanged = false;

          std::unique_ptr<te::gm::Geometry> geomResultFixA(te::gm::PrepareGeometriesToIntersection(geom, candidate, wasChanged, tolerance));

          std::unique_ptr<te::gm::Geometry> geomResultFixB(te::gm::PrepareGeometriesToIntersection(candidate, geomResultFixA.get(), wasChanged, tolerance));

          delete geom;
          geom = geomResultFixA.release();
          geometries[t] = geom;

          delete candidate;
          candidate = geomResultFixB.release();
          geometries[report[p]] = candidate;
        }
      }
    }

    //create output data
    te::gtest::DataInfo outputDataInfo = getOutputDataInfo(blueDataInfo.m_datasetName);
    std::string filePath = outputDataInfo.m_path;
    te::gtest::SimpleSaveGeometriesToShapefile(filePath, blueDataInfo.m_datasetName + "_validate_prepareSnap", te::gm::ToConstVector(geometries));

    te::common::FreeContents(geometries);

    ASSERT_TRUE(true);
  }

  TEST_F(ITsValidation, Prepare_Prepare)
  {
    te::gtest::DataInfo blueDataInfo = te::gtest::GetBlueDataInfo();

    std::unique_ptr<te::da::DataSource> dataSource = te::da::DataSourceFactory::make("OGR", blueDataInfo.m_uri);
    dataSource->open();

    std::unique_ptr<te::da::DataSetType> dataSetType = dataSource->getDataSetType(blueDataInfo.m_datasetName);

    std::unique_ptr<te::da::DataSet> dataSet = dataSource->getDataSet(blueDataInfo.m_datasetName);

    te::gm::GeometryProperty* geomProp = te::da::GetFirstGeomProperty(dataSetType.get());

    //get data
    std::unique_ptr<te::sam::rtree::Index<std::size_t, 8> >rtree(new te::sam::rtree::Index<size_t, 8>);

    std::size_t index = 0;

    te::gm::GeometryVector geometries;

    dataSet->moveBeforeFirst();

    while (dataSet->moveNext())
    {
      std::unique_ptr<te::gm::Geometry> geom = dataSet->getGeometry(geomProp->getName());

      if (geom->isValid() == false)
        continue;

      double tolerance = te::gm::GetAppropriatePrecision(geom->getSRID());

#ifdef TERRALIB_GEOS_ENABLED
      std::unique_ptr<geos::geom::Geometry> geomGeos(te::gm::GEOSWriter::write(geom.get()));

      std::unique_ptr<geos::geom::Geometry> geomGeosBuffer(geomGeos->buffer(0));

      std::unique_ptr<geos::geom::Geometry> simplifiedGeom = geos::simplify::TopologyPreservingSimplifier::simplify(geomGeosBuffer.get(), tolerance);

      geos::operation::overlay::snap::GeometrySnapper snapper(*simplifiedGeom.get());

      double newTolerance = geos::operation::overlay::snap::GeometrySnapper::computeSizeBasedSnapTolerance(*simplifiedGeom.get());

      if (newTolerance > tolerance)
      {
        tolerance = newTolerance;
      }

      std::unique_ptr<geos::geom::Geometry> resultGeom = snapper.snapToSelf(tolerance, true);

      int geomSRID = geom->getSRID();
      geom.reset(te::gm::GEOSReader::read(resultGeom.get()));
      geom->setSRID(geomSRID);
#endif

      rtree->insert(*geom->getMBR(), index);

      geometries.push_back(geom.release());

      ++index;
    }

    for (std::size_t t = 0; t < geometries.size(); ++t)
    {
      te::gm::Geometry* geom = geometries[t];

      te::gm::Envelope geomMBR = *geom->getMBR();

      std::vector<std::size_t> report;
      rtree->search(*geom->getMBR(), report);

      te::gm::GeometryVector reportGeom;

      for (std::size_t p = 0; p < report.size(); ++p)
      {
        if (report[p] == t)
        {
          continue;
        }

        te::gm::Geometry* candidate = geometries[report[p]];

        reportGeom.push_back(candidate);
      }

      double tolerance = te::gm::GetAppropriatePrecision(geom->getSRID());

      bool wasChanged = true;


      std::unique_ptr<te::gm::Geometry> geomResultFix(te::gm::PrepareGeometriesToIntersection(geom, reportGeom, wasChanged, tolerance));

      if (wasChanged)
      {
        delete geom;
        geom = geomResultFix.release();
        geometries[t] = geom;
      }
    }

    //create output data
    te::gtest::DataInfo outputDataInfo = getOutputDataInfo(blueDataInfo.m_datasetName);
    std::string filePath = outputDataInfo.m_path;

    te::gtest::SimpleSaveGeometriesToShapefile(filePath, blueDataInfo.m_datasetName + "_validate_prepare", te::gm::ToConstVector(geometries));

    te::common::FreeContents(geometries);

    ASSERT_TRUE(true);
  }

  te::gm::Geometry* MakeValidGEOS(te::gm::Geometry * geometry)
  {
#if GEOS_FULLVERSION >= 30800
    std::unique_ptr<geos::geom::Geometry> geosGeometry(te::gm::GEOSWriter::write(geometry));
    geos::operation::valid::MakeValid makeValid;
    std::unique_ptr<geos::geom::Geometry> geosResult = makeValid.build(geosGeometry.get());
    geosResult->setSRID(geometry->getSRID());
    return te::gm::GEOSReader::read(geosResult.get());
#else
    return static_cast<te::gm::Geometry*>(geometry->clone());
#endif
  }

  struct comma_separator : std::numpunct<char> 
  {
    virtual char do_decimal_point() const override 
    { 
      return ','; 
    }
  };

  void executeFix(te::da::DataAccess* dataAccess, int method, const std::string& outputFileName)
  {
    std::unique_ptr<te::da::DataSet> dataSet = dataAccess->getDataSet();
    std::size_t geometryColumnPos = te::da::GetFirstSpatialPropertyPos(dataSet.get());

    //double areaTotalBefore = 0;
    //double areaTotalAfter = 0;

    //double areaFixedBefore = 0;
    //double areaFixedAfter = 0;

    std::size_t index = 0;
    std::size_t numFixes = 0;

    std::ofstream file;
    file.open(outputFileName.c_str(), std::ofstream::out | std::ofstream::trunc);
    file.imbue(std::locale(std::cout.getloc(), new comma_separator));
    file << "index;was_valid_before;fix_result;area_before;area_after;delta;percent;wkt_before;wkt_after" << std::endl;
    while (dataSet->moveNext() == true)
    {
      std::unique_ptr<te::gm::Geometry> geometry = dataSet->getGeometry(geometryColumnPos);
      geometry->transform(29101);

      double areaBefore = te::gm::GetArea(geometry.get());
      //areaTotalBefore += areaBefore;

      te::gm::GeometryPtr result;
      std::string fixResult = "unchanged";
      bool wasValidBefore = geometry->isValid();
      if (wasValidBefore == true)
      {
        result.reset(static_cast<te::gm::Geometry*>(geometry->clone()));
      }
      else
      {
        //applies the 3 methods and sum the resulting areas
        if (method == 1)
        {
          //Fix 1 - buffer(0)
          result.reset(geometry->buffer(0));
        }
        else if (method == 2)
        {
          //Fix 2 - current terralib makevalid algorithm
          result.reset(te::gm::Validation::MakeValidWithSnap(geometry.get()).release());
        }
        else if (method == 3)
        {
          //Fix 3 - new geos makeValid algorithm
          result.reset(MakeValidGEOS(geometry.get()));
        }
      }

      std::string wktBefore;
      std::string wktAfter;

      double areaAfter = 0;
      if (te::gm::IsNullOrEmpty(result) == true)
      {
        wktBefore = geometry->toString();
        fixResult = "discarded";
      }
      else
      {
        areaAfter = te::gm::GetArea(result.get());
        
        if (wasValidBefore == false)
        {
          wktBefore = geometry->toString();
          wktAfter = result->toString();
          if (result->isValid() == true)
          {
            fixResult = "fixed";
          }
          else
          {
            fixResult = "not fixed";
          }
        }
      }
      
      //areaFixedBefore += areaBefore;
      //areaFixedAfter += areaAfter;
      
      double areaDelta = std::fabs(areaBefore - areaAfter);
      double changePercent = areaDelta * 100 / areaBefore;

      if (changePercent > 20)
      {
        int a = 0;
      }

      //index;is_valid_before;is_valid_after;area_before;area_after;delta;percent
      file << index;
      file << ";" << wasValidBefore;
      file << ";" << fixResult;
      file << ";" << std::fixed << std::setprecision(20) << areaBefore;
      file << ";" << std::fixed << std::setprecision(20) << areaAfter;
      file << ";" << std::fixed << std::setprecision(20) << areaDelta;
      file << ";" << std::fixed << std::setprecision(10) << changePercent;
      file << ";" << wktBefore;
      file << ";" << wktAfter;
      file << std::endl;

      ++numFixes;
      ++index;
    }
  }

  TEST_F(ITsValidation, CompareAlgorithms)
  {
    te::gtest::DataInfo dataInfo = te::gtest::DataInfo("C:\\temp\\inpe\\avancine\\makevalid\\deter_publicPolygon.shp");

    std::unique_ptr<te::da::DataAccess> dataAccess(new te::da::DataAccess(te::core::URI(dataInfo.m_uri), "OGR", dataInfo.m_datasetName));
   
    std::string path = "C:\\temp\\inpe\\avancine\\makevalid\\";

    //Fix Method 1 - geometry->buffer(0)
    executeFix(dataAccess.get(), 1, path + "fix_method_1.csv");

    //Fix Method 2 - te::gm::validation::MakeValid
    executeFix(dataAccess.get(), 2, path + "fix_method_2.csv");

    //Fix Method 3 - geos::operation::valid::MakeValid
    executeFix(dataAccess.get(), 3, path + "fix_method_3.csv");
  }

  int getPlanarSRID(const te::gm::Envelope& envelope)
  {
    int srid = 3395;
    return srid;
  }

  te::gm::Coord2D transform(const te::gm::Coord2D& coord, int oldsrid, int newsrid)
  {
    if (oldsrid == newsrid)
      return coord;

    std::unique_ptr<te::srs::Converter> converter(new te::srs::Converter());

    try
    {
      converter->setSourceSRID(oldsrid);
      converter->setTargetSRID(newsrid);
    }
    catch (te::common::Exception& /* ex */)
    {
      return coord;
    }

    double newX = 0.;
    double newY = 0.;

    // convert the four corners
    converter->convert(coord.x, coord.y, newX, newY);

    te::gm::Coord2D newCoord(newX, newY);
    return newCoord;
  }

  te::gm::Envelope transformEnvelope(const te::gm::Envelope& envelope, int oldsrid, int newsrid)
  {
    te::gm::Coord2D coordLowerLeft = envelope.getLowerLeft();
    te::gm::Coord2D coordUpperRight = envelope.getUpperRight();

    coordLowerLeft = transform(coordLowerLeft, oldsrid, newsrid);
    coordUpperRight = transform(coordUpperRight, oldsrid, newsrid);

    te::gm::Envelope newEnvelope(envelope);
    newEnvelope.transform(oldsrid, newsrid);
    return newEnvelope;
  }

  double getScale(double widthMM, double heightMM, const te::gm::Envelope& envelope, int envelopeSRID)
  {
    double widthMeters = widthMM / 1000.;
    double heightMeters = heightMM / 1000.;

    te::gm::Envelope planarEnvelope(envelope);
    int planarSRID = envelopeSRID;
    te::common::UnitOfMeasurePtr unitPtr = te::srs::SpatialReferenceSystemManager::getInstance().getUnit(envelopeSRID);
    std::string unit = unitPtr->getName();
    if (unit == "DEGREE")
    {
      planarSRID = getPlanarSRID(envelope);
      planarEnvelope = transformEnvelope(envelope, envelopeSRID, planarSRID);
    }

    double wdx = planarEnvelope.getWidth();
    double wdy = planarEnvelope.getHeight();

    double scaleX = wdx / widthMeters;
    double scaleY = wdy / heightMeters;

    return scaleX > scaleY ? scaleY : scaleX;
  }

  double calculateErrorFactor(const te::gm::Envelope& referenceEnvelope, int referenceSRID, int testSRID)
  {
    te::gm::Coord2D referenceLowerLeft = referenceEnvelope.getLowerLeft();
    te::gm::Coord2D referenceUpperRight(referenceEnvelope.m_urx, referenceEnvelope.m_lly);

    te::gm::Coord2D testLowerLeft = transform(referenceLowerLeft, referenceSRID, testSRID);
    te::gm::Coord2D testUpperRight = transform(referenceUpperRight, referenceSRID, testSRID);

    te::gm::Coord2D conferenceLowerLeft = transform(testLowerLeft, testSRID, referenceSRID);
    te::gm::Coord2D conferenceUpperRight = transform(testUpperRight, testSRID, referenceSRID);
    
    double referenceDistance = te::gm::Distance::CalculateDistance(referenceLowerLeft, referenceUpperRight);
    double conferenceDistance = te::gm::Distance::CalculateDistance(conferenceLowerLeft, conferenceUpperRight);

    double transformErrorFactor = conferenceDistance / referenceDistance;
    return transformErrorFactor;
  }

  te::gm::Envelope createEnvelopeForScale(double scale, double widthMM, double heightMM, const te::gm::Coord2D& planarCenter)
  {
    double centerX = planarCenter.x;
    double centerY = planarCenter.y;

    //calculates the planar envelope based on the given scale
    double envelopeWidth = widthMM / 1000. * scale;
    double envelopeHeight = heightMM / 1000. * scale;
    double halfWidth = envelopeWidth / 2.;
    double halfHeight = envelopeHeight / 2.;

    te::gm::Coord2D c1Planar(centerX - halfWidth, centerY - halfHeight);
    te::gm::Coord2D c2Planar(centerX + halfWidth, centerY + halfHeight);

    te::gm::Envelope resultEnvelope = te::gm::CreateEnvelope(c1Planar.x, c1Planar.y, c2Planar.x, c2Planar.y);
    return resultEnvelope;
  }

  te::gm::Envelope adjustEnvelopeAspectRatio(const te::gm::Envelope& envelope, double widthByHeightFactor)
  {
    te::gm::Envelope resultEnvelope(envelope);
    if (widthByHeightFactor > 0)
    {
      //the envelope has its width higher than its hight
      //in this case we must keep the width, and adjust its height
      te::gm::Coord2D resultCenter = resultEnvelope.getCenter();
      double resultCenterY = resultCenter.getY();
      double newHeight = resultEnvelope.getWidth() / widthByHeightFactor;
      resultEnvelope.m_lly = resultCenterY - (newHeight * 0.5);
      resultEnvelope.m_ury = resultEnvelope.m_lly + newHeight;
    }
    else if (widthByHeightFactor < 0)
    {
      //the envelope has its height higher than its width
      //in this case we must keep the height, and adjust its width
      te::gm::Coord2D resultCenter = resultEnvelope.getCenter();
      double resultCenterX = resultCenter.getX();
      double newWidth = resultEnvelope.getHeight() * widthByHeightFactor;
      resultEnvelope.m_llx = resultCenterX - (newWidth * 0.5);
      resultEnvelope.m_urx = resultEnvelope.m_llx + newWidth;
    }

    return resultEnvelope;
  }

  te::gm::Envelope adjustEnvelopeForScale(double scale, const te::gm::Envelope& envelope, int envelopeSRID, double widthMM, double heightMM)
  {
    //calculates the initial parameters based on the given envelope and srid
    //we must ensure that they are projected in a planar coordinate system
    int planarSRID = envelopeSRID;
    te::gm::Coord2D centerPlanar = envelope.getCenter();
    double widthByHeightFactor = widthMM / heightMM;
    te::common::UnitOfMeasurePtr unitPtr = te::srs::SpatialReferenceSystemManager::getInstance().getUnit(envelopeSRID);
    std::string unit = unitPtr->getName();
    if (unit == "DEGREE")
    {
      planarSRID = getPlanarSRID(envelope);
      centerPlanar = transform(envelope.getCenter(), envelopeSRID, planarSRID);
    }

    te::gm::Envelope resultEnvelope = createEnvelopeForScale(scale, widthMM, heightMM, centerPlanar);
    if (unit == "DEGREE")
    {
      double distanceErrorFactor = calculateErrorFactor(resultEnvelope, planarSRID, envelopeSRID);
      //std::cout << std::endl << "Distance Error Factor: " << distanceErrorFactor;

      //calculates the equivalent geographic envelope
      resultEnvelope = transformEnvelope(resultEnvelope, planarSRID, envelopeSRID);
      resultEnvelope = adjustEnvelopeAspectRatio(resultEnvelope, widthByHeightFactor);
    }

    return resultEnvelope;
  }

  TEST_F(ITsValidation, ScaleCalculation)
  {
    double x1 = -51.211752809306248;
    double y1 = -20.321073215508001;
    double x2 = -50.337769198313119;
    double y2 = -19.765172075182441;
    te::gm::Envelope geographicEnvelope(x1, y1, x2, y2);
    

    int geographicSRID = 4674;
    int planarSRID = getPlanarSRID(geographicEnvelope);

    double width = 1470;
    double height = 935;
    double widthMM = 457;
    double heightMM = 290;

    double factor = width / height;
    heightMM = widthMM / factor;
    
    double factorMM = widthMM / heightMM;
    double factorWorld = geographicEnvelope.getWidth() / geographicEnvelope.getHeight();

    double power = 0;
    double scale = 0;
    while (scale <= 1000000)
    {
      //initializes the scale to be tested
      scale = std::pow(2, power++);

      //adjusts the envelope
      te::gm::Envelope adjustedEnvelope = adjustEnvelopeForScale(scale, geographicEnvelope, geographicSRID, widthMM, heightMM);
      
      //tests the resulting scale
      double adjustedScale = getScale(widthMM, heightMM, adjustedEnvelope, geographicSRID);
      adjustedScale = std::round(adjustedScale);
      ASSERT_EQ(scale, adjustedScale);

      //tests the aspect ratio
      double adjustedFactor = adjustedEnvelope.getWidth() / adjustedEnvelope.getHeight();
      ASSERT_TRUE(std::fabs(factorMM - adjustedFactor) < 0.00001);
    }
  }
}
