/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/rp/Functions.cpp

  \brief Raster Processing functions.
*/

// TerraLib

#ifdef WIN32
#define NOMINMAX
#endif

#include "Functions.h"

#include "../common/SystemApplicationSettings.h"
#include "../common/Version.h"
#include "../common/MathUtils.h"
#include "../core/utils/Platform.h"
#include "../dataaccess/dataset/DataSetType.h"
#include "../dataaccess/datasource/DataSourceFactory.h"
#include "../dataaccess/utils/Utils.h"
#include "../datatype/Enums.h"
#include "../raster/BandProperty.h"
#include "../raster/Grid.h"
#include "../raster/RasterFactory.h"
#include "../raster/RasterProperty.h"
#include "../raster/RasterSummaryManager.h"
#include "../raster/RasterIterator.h"
#include "../raster/Utils.h"
#include "../raster/PositionIterator.h"
#include "../raster/RasterSynchronizer.h"
#include "../raster/SynchronizedRaster.h"
#include "../memory/CachedRaster.h"
#include "../geometry/Point.h"
#include "../geometry/Validation.h"
#include "../common/MatrixUtils.h"
#include "../common/progress/TaskProgress.h"
#include "Exception.h"
#include "Macros.h"
#include "RasterHandler.h"
#include "Contrast.h"

// Boost
#include <boost/filesystem.hpp>
#include <boost/numeric/ublas/io.hpp>
#include <boost/numeric/ublas/matrix.hpp>
#include <boost/property_tree/ptree.hpp>
#include <boost/property_tree/json_parser.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/shared_array.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/thread.hpp>
#include <boost/numeric/ublas/lu.hpp>
#include <boost/numeric/ublas/matrix.hpp>

// STL
#include <cstring>
#include <string>
#include <limits>
#include <map>
#include <memory>
#include <algorithm>
#include <mutex>
#include <stack>

#ifndef M_PI
  #define M_PI       3.14159265358979323846
#endif

namespace te
{
  namespace rp
  {
    struct GetMeanValueThreadParams
    {
      te::rst::RasterSynchronizer* m_rasterSyncPtr;
      unsigned int m_inputBandIndex;
      Matrix< bool >* m_rasterBlocksStatusPtr;
      double* m_meanValuePtr;
      double* m_meanElementsNumberPtr;
      double m_noDataValue;
      bool m_returnStatus;
      std::mutex* m_mutexPtr;
    };  
    
    struct GetStdDevValueThreadParams
    {
      te::rst::RasterSynchronizer* m_rasterSyncPtr;
      unsigned int m_inputBandIndex;
      Matrix< bool >* m_rasterBlocksStatusPtr;
      double* m_stdDevValuePtr;
      double* m_pixelsNumberValuePtr;
      double m_meanValue;
      double m_noDataValue;
      bool m_returnStatus;
      std::mutex* m_mutexPtr;
    };     
    
    struct RemapValuesThreadParams
    {
      te::rst::RasterSynchronizer* m_inputRasterSync;
      te::rst::RasterSynchronizer* m_outputRasterSync;
      std::vector< unsigned int > const * m_inputRasterBandsPtr;
      std::vector< unsigned int > const * m_outputRasterBandsPtr;
      boost::numeric::ublas::matrix< double > const * m_remapMatrixPtr;
      Matrix< bool > * m_outputRasterBlocksStatusPtr;
      bool m_returnStatus;
      std::mutex* m_mutexPtr;
      unsigned int m_inputRasterMaxNBlocksX;
    };     
    
    struct ComposeBandsThreadParams
    {
      std::vector< std::unique_ptr< te::rst::RasterSynchronizer > >*
        m_inRastersSyncsPtrsPtr;
      std::vector< unsigned int > m_inputRasterBands;
      te::rst::RasterSynchronizer* m_outRasterSyncPtr;
      Matrix< bool >* m_rasterBlocksStatusPtr;
      bool* m_returnStatusPtr;
      std::mutex* m_mutexPtr;
      te::rst::Interpolator::Method m_interpMethod;
      bool m_allowNoDataPixels;
      te::common::TaskProgress* m_progressPtr;
    };     
    
    struct GetHistogramsThreadParameters
    {
      te::rst::RasterSynchronizer* m_rasterSyncPtr;
      unsigned int m_inputBandIndex;
      std::vector< te::gm::Polygon const* > const* m_roisPtr;
      std::vector< bool >* m_roisProcStatusPtr;
      bool m_returnStatus;
      std::mutex* m_mutexPtr;
      std::map<double, unsigned>* m_rHistogramPtr;
      std::map<double, unsigned>* m_iHistogramPtr;
      bool m_preReadRasterMinMax;
      double m_inBandMinValueR;
      double m_inBandMaxValueR;
      double m_inBandMinValueI;
      double m_inBandMaxValueI;
      unsigned int m_maxRasterBlocksPerThread;
      te::rst::PolygonIterator< double >::IterationType m_roisItType;

      // 0 = calc min and max for each block
      // 1 = calc histograms
      unsigned int m_threadExecMode;
    };    
   
    struct GetJointHistogramsThreadParameters
    {
      te::rst::RasterSynchronizer* m_rasterSyncPtr;
      unsigned int m_inputBandIndex1;
      unsigned int m_inputBandIndex2;
      std::vector< te::gm::Polygon const* > const* m_roisPtr;
      std::vector< bool >* m_roisProcStatusPtr;
      bool m_returnStatus;
      std::mutex* m_mutexPtr;
      std::map<double, unsigned int> const * m_rRHist1Ptr;
      std::map<double, unsigned int> const * m_rIHist1Ptr;
      std::map<double, unsigned int> const * m_rRHist2Ptr;
      std::map<double, unsigned int> const * m_rIHist2Ptr;      
      std::map< std::pair< double, double >, unsigned int>* m_realJointHistogramPtr;
      std::map< std::pair< double, double >, unsigned int>* m_imagJointHistogramPtr;
      unsigned int m_maxRasterBlocksPerThread;
      te::rst::PolygonIterator< double >::IterationType m_roisItType;
    };
    
    struct GetCovarianceValueThreadParameters
    {
      te::rst::RasterSynchronizer* m_rasterSyncPtr1;
      te::rst::RasterSynchronizer* m_rasterSyncPtr2;
      unsigned int m_inputBandIndex1;
      unsigned int m_inputBandIndex2;
      bool m_returnStatus;
      std::mutex* m_mutexPtr;
      double m_mean1Value;
      double m_mean2Value;
      double* m_covarianceValuePtr;
      std::vector< std::tuple< bool,  unsigned int, unsigned int > >*
        m_raster1BlocksStatusPtr;
      unsigned int m_raster2NBlocksX;
      unsigned int* m_elementsNumberPtr;
    };         
    
    bool CreateNewRaster( const te::rst::Grid& rasterGrid,
      const std::vector< te::rst::BandProperty* >& bandsProperties,
      const std::string& outDataSetName,
      const std::string& dataSourceType,
      RasterHandler& outRasterHandler )
    {
      // Creating a new datasource

      std::unique_ptr< te::da::DataSource > dataSourcePtr(
        te::da::DataSourceFactory::make( dataSourceType, ("file://" + outDataSetName)));
      if( dataSourcePtr.get() == nullptr ) return false;      
      
      RasterHandler internalRasterHandler;
      
      if( CreateNewRaster( rasterGrid, bandsProperties, 
        outDataSetName, *dataSourcePtr, internalRasterHandler ) )
      {
        std::unique_ptr< te::da::DataSource > dummyDataSourcePtr;
        std::unique_ptr< te::da::DataSourceTransactor > transactorPtr;
        std::unique_ptr< te::da::DataSet > dataSetPtr;
        std::unique_ptr< te::rst::Raster > rasterPtr;
        
        internalRasterHandler.release( dummyDataSourcePtr, transactorPtr,
          dataSetPtr, rasterPtr );
          
        outRasterHandler.reset( dataSourcePtr.release(), transactorPtr.release(), 
          dataSetPtr.release(), rasterPtr.release() );
        
        return true;
      }
      else
      {
        return false;
      }      
    }
        
    bool CreateNewRaster( const te::rst::Grid& rasterGrid,
      const std::vector< te::rst::BandProperty* >& bandsProperties,
      const std::string& outDataSetName,
      te::da::DataSource& outDataSource,
      RasterHandler& outRasterHandler )
    {
      // Defining the raster properties

      std::unique_ptr< te::rst::RasterProperty > rasterPropertyPtr(
        new te::rst::RasterProperty( new te::rst::Grid( rasterGrid ),
        bandsProperties, std::map< std::string, std::string >(),
        false, 0, nullptr ) );

      // acquiring a transactor instance

      std::unique_ptr< te::da::DataSourceTransactor > transactorPtr(
       outDataSource.getTransactor() );

      if( transactorPtr.get() == nullptr )
      {
       return false;
      }

      // Creating a data set instance
      
      if( transactorPtr->dataSetExists( outDataSetName ) )
      {
        transactorPtr->dropDataSet( outDataSetName );
      }

      std::unique_ptr< te::da::DataSetType > dataSetTypePtr(
        new te::da::DataSetType( outDataSetName ) );
      dataSetTypePtr->add( rasterPropertyPtr.release() );

      try
      {
        transactorPtr->createDataSet( dataSetTypePtr.get(),
          std::map< std::string, std::string >() );
      }
      catch( ... )
      {
        return false;
      }
      
       if( ! transactorPtr->dataSetExists( outDataSetName ) ) return false;

      std::unique_ptr< te::da::DataSet > dataSetPtr( transactorPtr->getDataSet(
        outDataSetName, te::common::FORWARDONLY, true, te::common::RWAccess ) );

      if( dataSetPtr.get() == nullptr )
      {
        return false;
      }

      // Creating a raster instance

      std::unique_ptr< te::rst::Raster > rasterPtr( dataSetPtr->getRaster( 0 ) );

      if( rasterPtr.get() )
      {
        outRasterHandler.reset( transactorPtr.release(), dataSetPtr.release(), rasterPtr.release() );
        return true;
      }
      else
      {
        return false;
      }
    }
    
    bool TERPEXPORT CreateNewRaster( const te::rst::Grid& rasterGrid,
      const std::vector< te::rst::BandProperty* >& bandsProperties,
      const std::map< std::string, std::string>& rasterInfo,
      const std::string& rasterType,
      std::unique_ptr< te::rst::Raster >& outRasterPtr )
    {
      try
      {
        outRasterPtr.reset( te::rst::RasterFactory::make( rasterType,
          new te::rst::Grid( rasterGrid ), bandsProperties, rasterInfo, nullptr, nullptr ) );
      }
      catch(...)
      {
        outRasterPtr.reset();
        return false;
      }
      
      if( outRasterPtr.get() == nullptr )
      {
        return false;
      }      
      
      return true;
    }

    bool CreateNewMemRaster( const te::rst::Grid& rasterGrid,
      std::vector< te::rst::BandProperty* > bandsProperties,
      RasterHandler& outRasterHandler )
    {
      std::string dataSetName = std::string( "createNewMemRaster" ) +
        boost::lexical_cast< std::string >( &outRasterHandler );
        
      return CreateNewRaster( rasterGrid, bandsProperties, dataSetName, 
        "MEM", outRasterHandler );
    }

    bool CreateNewGdalRaster( const te::rst::Grid& rasterGrid,
      std::vector< te::rst::BandProperty* > bandsProperties,
      const std::string& fileName,
      RasterHandler& outRasterHandler )
    {
      std::unique_ptr< te::rst::Raster > outRasterPtr;
      if( CreateNewGdalRaster( rasterGrid, bandsProperties, fileName, outRasterPtr ) )
      {
        outRasterHandler.reset( outRasterPtr.release() ); 
        return true;
      }
      else
      {
        outRasterHandler.reset();
        return false;
      }
    }
    
    bool CreateNewGdalRaster( const te::rst::Grid& rasterGrid,
      std::vector< te::rst::BandProperty* > bandsProperties,
      const std::string& fileName,
      std::unique_ptr< te::rst::Raster >& outRasterPtr )
    {
      outRasterPtr.reset();
      
      std::map< std::string, std::string > rInfo;
      rInfo[ "URI" ] = fileName;
      
      try
      {
        outRasterPtr.reset( te::rst::RasterFactory::make(
          "GDAL",
          new te::rst::Grid( rasterGrid ),
          bandsProperties,
          rInfo,
          nullptr,
          nullptr ) );
      }
      catch(...)
      {
        outRasterPtr.reset();
      }
      
      if( outRasterPtr.get() )
      {
        return true;
      }
      else
      {
        return false;
      }
    }
    
    bool TERPEXPORT Copy2DiskRaster( const te::rst::Raster& inputRaster,
      const std::string& fileName )
    {
      if( !(inputRaster.getAccessPolicy() & te::common::RAccess ) )
      {
        return false;
      };
      
      const unsigned int nBands = static_cast<unsigned int>(inputRaster.getNumberOfBands());
      const unsigned int nCols = inputRaster.getNumberOfColumns();
      const unsigned int nRows = inputRaster.getNumberOfRows();
      unsigned int bandIdx =0;
      unsigned int col = 0;
      unsigned int row = 0;
      
      std::vector< te::rst::BandProperty* > bandsProperties;
      for( bandIdx = 0 ; bandIdx < nBands ; ++bandIdx )
      {
        bandsProperties.push_back( new te::rst::BandProperty(
          *( inputRaster.getBand( bandIdx )->getProperty() ) ) );        
      }
      
      RasterHandler outRasterHandler;
      
      if( ! CreateNewGdalRaster( *( inputRaster.getGrid() ), bandsProperties,
        fileName, outRasterHandler ) )
      {
        return false;
      }
      
      double value = 0;
      
      for( bandIdx = 0 ; bandIdx < nBands ; ++bandIdx )
      {
        const te::rst::Band& inBand = *inputRaster.getBand( bandIdx );
        te::rst::Band& outBand = *outRasterHandler.getRasterPtr()->getBand( bandIdx );
        
        for( row = 0 ; row < nRows ; ++row )
        {
          for( col = 0 ; col < nCols ; ++col )
          {
            inBand.getValue( col, row, value );
            outBand.setValue( col, row, value );
          }
        }
      }
      
      return true;
    }

    void Convert2DoublesVector( void* inputVector, const int inputVectorDataType,
      unsigned int inputVectorSize, double* outputVector )
    {
      switch( inputVectorDataType )
      {
        case te::dt::CHAR_TYPE :
        {
          char* vPtr = static_cast<char*>(inputVector);
          for( unsigned int idx = 0 ; idx < inputVectorSize ; ++idx )
            outputVector[ idx ] = static_cast<double>(vPtr[ idx ]);
          break;
        }
        case te::dt::BIT_TYPE :
        case te::dt::UCHAR_TYPE :
        {
          unsigned char* vPtr = static_cast<unsigned char*>(inputVector);
          for( unsigned int idx = 0 ; idx < inputVectorSize ; ++idx )
            outputVector[ idx ] = static_cast<double>(vPtr[ idx ]);
          break;
        }
        case te::dt::INT16_TYPE :
        {
          short int* vPtr = static_cast<short int*>(inputVector);
          for( unsigned int idx = 0 ; idx < inputVectorSize ; ++idx )
            outputVector[ idx ] = static_cast<double>(vPtr[ idx ]);
          break;
        }
        case te::dt::CINT16_TYPE :
        {
          std::complex< short int >* vPtr = static_cast<std::complex< short int >*>(inputVector);
          for( unsigned int idx = 0 ; idx < inputVectorSize ; ++idx )
            outputVector[ idx ] = static_cast<double>(vPtr[ idx ].real());
          break;
        }
        case te::dt::UINT16_TYPE :
        {
          unsigned short int* vPtr = static_cast<unsigned short int*>(inputVector);
          for( unsigned int idx = 0 ; idx < inputVectorSize ; ++idx )
            outputVector[ idx ] = static_cast<double>(vPtr[ idx ]);
          break;
        }
        case te::dt::INT32_TYPE :
        {
          int* vPtr = static_cast<int*>(inputVector);
          for( unsigned int idx = 0 ; idx < inputVectorSize ; ++idx )
            outputVector[ idx ] = static_cast<double>(vPtr[ idx ]);
          break;
        }
        case te::dt::CINT32_TYPE :
        {
          std::complex< int >* vPtr = static_cast<std::complex< int >*>(inputVector);
          for( unsigned int idx = 0 ; idx < inputVectorSize ; ++idx )
            outputVector[ idx ] = static_cast<double>(vPtr[ idx ].real());
          break;
        }
        case te::dt::UINT32_TYPE :
        {
          unsigned int* vPtr = static_cast<unsigned int*>(inputVector);
          for( unsigned int idx = 0 ; idx < inputVectorSize ; ++idx )
            outputVector[ idx ] = static_cast<double>(vPtr[ idx ]);
          break;
        }
        case te::dt::INT64_TYPE :
        {
          long int* vPtr = static_cast<long int*>(inputVector);
          for( unsigned int idx = 0 ; idx < inputVectorSize ; ++idx )
            outputVector[ idx ] = static_cast<double>(vPtr[ idx ]);
          break;
        }
        case te::dt::UINT64_TYPE :
        {
          unsigned long int* vPtr = static_cast<unsigned long int*>(inputVector);
          for( unsigned int idx = 0 ; idx < inputVectorSize ; ++idx )
            outputVector[ idx ] = static_cast<double>(vPtr[ idx ]);
          break;
        }
        case te::dt::FLOAT_TYPE :
        {
          float* vPtr = static_cast<float*>(inputVector);
          for( unsigned int idx = 0 ; idx < inputVectorSize ; ++idx )
            outputVector[ idx ] = static_cast<double>(vPtr[ idx ]);
          break;
        }
        case te::dt::CFLOAT_TYPE :
        {
          std::complex< float >* vPtr = static_cast<std::complex< float >*>(inputVector);
          for( unsigned int idx = 0 ; idx < inputVectorSize ; ++idx )
            outputVector[ idx ] = static_cast<double>(vPtr[ idx ].real());
          break;
        }
        case te::dt::DOUBLE_TYPE :
        {
          memcpy( outputVector, inputVector, inputVectorSize * sizeof( double ) );
          break;
        }
        case te::dt::CDOUBLE_TYPE :
        {
          std::complex< double >* vPtr = static_cast<std::complex< double >*>(inputVector);
          for( unsigned int idx = 0 ; idx < inputVectorSize ; ++idx )
            outputVector[ idx ] = static_cast<double>(vPtr[ idx ].real());
          break;
        }
        default :
          throw te::rp::Exception( "Invalid data type" );
      };
    }

    void ConvertDoublesVector( double* inputVector,
      unsigned int inputVectorSize, const int outputVectorDataType,
      void* outputVector )
    {
      switch( outputVectorDataType )
      {
        case te::dt::CHAR_TYPE :
        {
          char* vPtr = static_cast<char*>(outputVector);
          for(unsigned int idx = 0 ; idx < inputVectorSize ; ++idx )
            vPtr[ idx ] = static_cast<char>(inputVector[ idx ]);
          break;
        }
        case te::dt::BIT_TYPE :
        case te::dt::UCHAR_TYPE :
        {
          unsigned char* vPtr = static_cast<unsigned char*>(outputVector);
          for(unsigned int idx = 0 ; idx < inputVectorSize ; ++idx )
            vPtr[ idx ] = static_cast<unsigned char>(inputVector[ idx ]);
          break;
        }
        case te::dt::INT16_TYPE :
        {
          short int* vPtr = static_cast<short int*>(outputVector);
          for(unsigned int idx = 0 ; idx < inputVectorSize ; ++idx )
            vPtr[ idx ] = static_cast<short int>(inputVector[ idx ]);
          break;
        }
        case te::dt::CINT16_TYPE :
        {
          std::complex< short int >* vPtr = static_cast<std::complex< short int >*>(outputVector);
          for(unsigned int idx = 0 ; idx < inputVectorSize ; ++idx )
            vPtr[ idx ]= ( static_cast<short int>(inputVector[ idx ]) );
          break;
        }
        case te::dt::UINT16_TYPE :
        {
          unsigned short int* vPtr = static_cast<unsigned short int*>(outputVector);
          for(unsigned int idx = 0 ; idx < inputVectorSize ; ++idx )
            vPtr[ idx ] = static_cast<unsigned short int>(inputVector[ idx ]);
          break;
        }
        case te::dt::INT32_TYPE :
        {
          int* vPtr = static_cast<int*>(outputVector);
          for(unsigned int idx = 0 ; idx < inputVectorSize ; ++idx )
            vPtr[ idx ] = static_cast<int>(inputVector[ idx ]);
          break;
        }
        case te::dt::CINT32_TYPE :
        {
          std::complex< int >* vPtr = static_cast<std::complex< int >*>(outputVector);
          for(unsigned int idx = 0 ; idx < inputVectorSize ; ++idx )
            vPtr[ idx ] = ( static_cast<int>(inputVector[ idx ]) );
          break;
        }
        case te::dt::UINT32_TYPE :
        {
          unsigned int* vPtr = static_cast<unsigned int*>(outputVector);
          for(unsigned int idx = 0 ; idx < inputVectorSize ; ++idx )
            vPtr[ idx ] = static_cast<unsigned int>(inputVector[ idx ]);
          break;
        }
        case te::dt::INT64_TYPE :
        {
          long int* vPtr = static_cast<long int*>(outputVector);
          for(unsigned int idx = 0 ; idx < inputVectorSize ; ++idx )
            vPtr[ idx ] = static_cast<long int>(inputVector[ idx ]);
          break;
        }
        case te::dt::UINT64_TYPE :
        {
          unsigned long int* vPtr = static_cast<unsigned long int*>(outputVector);
          for(unsigned int idx = 0 ; idx < inputVectorSize ; ++idx )
            vPtr[ idx ] = static_cast<unsigned long int>(inputVector[ idx ]);
          break;
        }
        case te::dt::FLOAT_TYPE :
        {
          float* vPtr = static_cast<float*>(outputVector);
          for(unsigned int idx = 0 ; idx < inputVectorSize ; ++idx )
            vPtr[ idx ] = static_cast<float>(inputVector[ idx ]);
          break;
        }
        case te::dt::CFLOAT_TYPE :
        {
          std::complex< float >* vPtr = static_cast<std::complex< float >*>(outputVector);
          for(unsigned int idx = 0 ; idx < inputVectorSize ; ++idx )
            vPtr[ idx ] = ( static_cast<float>(inputVector[ idx ]) );
          break;
        }
        case te::dt::DOUBLE_TYPE :
        {
          memcpy( outputVector, inputVector, inputVectorSize * sizeof( double ) );
          break;
        }
        case te::dt::CDOUBLE_TYPE :
        {
          std::complex< double >* vPtr = static_cast<std::complex< double >*>(outputVector);
          for( unsigned int idx = 0 ; idx < inputVectorSize ; ++idx )
            vPtr[ idx ] = ( static_cast<double>(inputVector[ idx ]) );
          break;
        }
        default :
          throw te::rp::Exception( "Invalid data type" );
      };
    }

    std::string GetSensorFilename()
    {
      std::string datalocalPath = te::core::GetAppLocalDataLocation();
      std::string app = te::common::SystemApplicationSettings::getInstance().getValue("Application.Name");
      std::string Organization = te::common::SystemApplicationSettings::getInstance().getValue("Application.Organization");
      std::string vers = te::common::Version::asString();

      static std::string sens_file = datalocalPath + "/" + Organization + "/" + app + "-" + vers + "/SpectralSensor.json";

      return sens_file;
    }
    
    std::vector<std::string> GetBandNames()
    {
      std::map<std::string, struct SpectralSensorParams > specBandInfos;
      GetAllSpectralBandInfos( specBandInfos );
      
      std::map<std::string, struct SpectralSensorParams >::const_iterator 
        specBandInfosIt = specBandInfos.begin();
      const std::map<std::string, struct SpectralSensorParams >::const_iterator 
        specBandInfosItEnd = specBandInfos.end();        
      std::vector<std::string> bandsNames;
        
      while( specBandInfosIt != specBandInfosItEnd )
      {
        bandsNames.push_back( specBandInfosIt->first );
        
        ++specBandInfosIt;
      }

      return bandsNames;
    }

    std::map<std::string, SpectralSensorParams > getSensorParams()
    {
      std::string sens_file = GetSensorFilename();

      std::map<std::string, SpectralSensorParams > SensorParams;

      if (!boost::filesystem::exists(sens_file)) //SpectralSensor.json doesn't exists, then uses the original sensors
      {
        GetAllSpectralBandInfos( SensorParams );
      }
      else
      {
        boost::property_tree::ptree pt;
        boost::property_tree::json_parser::read_json(sens_file, pt);

        for(boost::property_tree::ptree::value_type &v: pt.get_child("Spectral Sensors"))
        {
          std::string name = v.second.get<std::string>("name");
          std::string band = v.second.get<std::string>("band");
          std::string lower = v.second.get<std::string>("lower");
          std::string upper = v.second.get<std::string>("upper");
          std::string min = v.second.get<std::string>("min");
          std::string max = v.second.get<std::string>("max");
          SensorParams[name] = SpectralSensorParams(std::atoi(band.c_str()), std::atof(lower.c_str()), std::atof(upper.c_str()), std::atof(min.c_str()), std::atof(max.c_str()));
        }
      }

      return SensorParams;
    }

    void SaveSensorParams(std::map<std::string, SpectralSensorParams > &params)
    {
      std::string filename = GetSensorFilename();

      boost::property_tree::ptree pt;
      boost::property_tree::ptree sensors;

      for (std::map<std::string, te::rp::SpectralSensorParams > ::iterator it = params.begin(); it != params.end(); ++it)
      {
        boost::property_tree::ptree sensor;

        sensor.put("name", it->first);
        sensor.put("band", it->second.m_band);
        sensor.put("lower", it->second.m_lower);
        sensor.put("upper", it->second.m_upper);
        sensor.put("min", it->second.m_min);
        sensor.put("max", it->second.m_max);

        sensors.push_back(std::make_pair("Spectral Sensor", sensor));
      }

      pt.add_child("Spectral Sensors", sensors);

      boost::property_tree::json_parser::write_json(filename, pt);

    }
    
    bool GetSpectralBandInfo( const std::string& bandName,
      SpectralSensorParams& specBandInfo )
    {
      std::map<std::string, struct SpectralSensorParams > specBandInfos;
      GetAllSpectralBandInfos( specBandInfos );
      
      if (specBandInfos.find(bandName) == specBandInfos.end())
      {
        return false;
      }
      else
      {
        specBandInfo = specBandInfos[bandName];
        return true;      
      }
    }

    void GetAllSpectralBandInfos( 
      std::map<std::string, struct SpectralSensorParams >& specBandInfos )
    {
      static std::map<std::string, struct SpectralSensorParams > staticBandsInfos;
      
      if( staticBandsInfos.empty() )
      {
        staticBandsInfos["CBERS2_CCD_1_BLUE"] = SpectralSensorParams(1, 0.45, 0.52, 0.0, 255.0);
        staticBandsInfos["CBERS2_CCD_2_GREEN"] = SpectralSensorParams(2, 0.52, 0.59, 0.0, 255.0);
        staticBandsInfos["CBERS2_CCD_3_RED"] = SpectralSensorParams(3, 0.63, 0.69, 0.0, 255.0);
        staticBandsInfos["CBERS2_CCD_4_NIR"] = SpectralSensorParams(4, 0.77, 0.89, 0.0, 255.0);
        staticBandsInfos["CBERS2_CCD_5_PAN"] = SpectralSensorParams(5, 0.51, 0.73, 0.0, 255.0);

        staticBandsInfos["LANDSAT5_TM_1_BLUE"] = SpectralSensorParams(1, 0.45, 0.52, 0.0, 255.0);
        staticBandsInfos["LANDSAT5_TM_2_GREEN"] = SpectralSensorParams(2, 0.52, 0.60, 0.0, 255.0);
        staticBandsInfos["LANDSAT5_TM_3_RED"] = SpectralSensorParams(3, 0.63, 0.69, 0.0, 255.0);
        staticBandsInfos["LANDSAT5_TM_4_NIR"] = SpectralSensorParams(4, 0.76, 0.90, 0.0, 255.0);
        staticBandsInfos["LANDSAT5_TM_5_SWIR"] = SpectralSensorParams(5, 1.55, 1.75, 0.0, 255.0);
        staticBandsInfos["LANDSAT5_TM_6_THERMAL"] = SpectralSensorParams(6, 10.40, 12.50, 0.0, 255.0);
        staticBandsInfos["LANDSAT5_TM_7_MIR"] = SpectralSensorParams(7, 2.08, 2.35, 0.0, 255.0);

        staticBandsInfos["LANDSAT7_ETM+_1_BLUE"] = SpectralSensorParams(1, 0.45, 0.515, 0.0, 255.0);
        staticBandsInfos["LANDSAT7_ETM+_2_GREEN"] = SpectralSensorParams(2, 0.525, 0.605, 0.0, 255.0);
        staticBandsInfos["LANDSAT7_ETM+_3_RED"] = SpectralSensorParams(3, 0.63, 0.69, 0.0, 255.0);
        staticBandsInfos["LANDSAT7_ETM+_4_NIR"] = SpectralSensorParams(4, 0.75, 0.90, 0.0, 255.0);
        staticBandsInfos["LANDSAT7_ETM+_5_SWIR"] = SpectralSensorParams(5, 1.55, 1.75, 0.0, 255.0);
        staticBandsInfos["LANDSAT7_ETM+_6_THERMAL"] = SpectralSensorParams(6, 10.40, 12.5, 0.0, 255.0);
        staticBandsInfos["LANDSAT7_ETM+_7_MIR"] = SpectralSensorParams(7, 2.09, 2.35, 0.0, 255.0);
        staticBandsInfos["LANDSAT7_ETM+_8_PAN"] = SpectralSensorParams(8, 0.52, 0.90, 0.0, 255.0);

        staticBandsInfos["WV2_MUL_1_COASTAL"] = SpectralSensorParams(1, 0.4, 0.45, 0.0, 2048.0);
        staticBandsInfos["WV2_MUL_2_BLUE"] = SpectralSensorParams(2, 0.45, 0.51, 0.0, 2048.0);
        staticBandsInfos["WV2_MUL_3_GREEN"] = SpectralSensorParams(3, 0.51, 0.58, 0.0, 2048.0);
        staticBandsInfos["WV2_MUL_4_YELLOW"] = SpectralSensorParams(4, 0.585, 0.625, 0.0, 2048.0);
        staticBandsInfos["WV2_MUL_5_RED"] = SpectralSensorParams(5, 0.66, 0.69, 0.0, 2048.0);
        staticBandsInfos["WV2_MUL_6_REDEDGE"] = SpectralSensorParams(6, 0.705, 0.745, 0.0, 2048.0);
        staticBandsInfos["WV2_MUL_7_NIR1"] = SpectralSensorParams(7, 0.77, 0.895, 0.0, 2048.0);
        staticBandsInfos["WV2_MUL_8_NIR2"] = SpectralSensorParams(8, 0.86, 0.104, 0.0, 2048.0);
        staticBandsInfos["WV2_PAN"] = SpectralSensorParams(9, 0.45, 0.8, 0.0, 2048.0);

        staticBandsInfos["AWIFS_RED"] = SpectralSensorParams(1, 0.62, 0.68, 0.0, 4096.0);
        staticBandsInfos["AWIFS_GREEN"] = SpectralSensorParams(2, 0.52, 0.59, 0.0, 4096.0);
        staticBandsInfos["AWIFS_NIR"] = SpectralSensorParams(3, 0.77, 0.86, 0.0, 4096.0);
        staticBandsInfos["AWIFS_MIR"] = SpectralSensorParams(4, 1.55, 1.70, 0.0, 4096.0);

        staticBandsInfos["RAPID_EYE_BLUE"] = SpectralSensorParams(1, 440.00, 510.00, 0.0, 4096.0);
        staticBandsInfos["RAPID_EYE_GREEN"] = SpectralSensorParams(2, 520.00, 590.00, 0.0, 4096.0);
        staticBandsInfos["RAPID_EYE_RED"] = SpectralSensorParams(3, 630.00, 685.00, 0.0, 4096.0);
        staticBandsInfos["RAPID_EYE_REDEDGE"] = SpectralSensorParams(4, 690.00, 730.00, 0.0, 4096.0);
        staticBandsInfos["RAPID_EYE_NIR"] = SpectralSensorParams(5, 760.00, 850.00, 0.0, 4096.0);

        staticBandsInfos["LISS3_GREEN"] = SpectralSensorParams(1, 0.52, 0.59, 0.0, 128.0);
        staticBandsInfos["LISS3_RED"] = SpectralSensorParams(2, 0.62, 0.68, 0.0, 128.0);
        staticBandsInfos["LISS3_NEAR_INFRARED"] = SpectralSensorParams(3, 0.77, 0.86, 0.0, 128.0);
        staticBandsInfos["LISS3_SHORTWAVE_INFRARED"] = SpectralSensorParams(4, 1.55, 1.70, 0.0, 128.0);

        staticBandsInfos["DMC_NIR"] = SpectralSensorParams(1, 0.77, 0.90, 0.0, 1024.0);
        staticBandsInfos["DMC_RED"] = SpectralSensorParams(2, 0.63, 0.69, 0.0, 1024.0);
        staticBandsInfos["DMC_GREEN"] = SpectralSensorParams(3, 0.52, 0.60, 0.0, 1024.0);

        staticBandsInfos["LANDSAT8-OLI1_COASTAL_AEROSOL"] = SpectralSensorParams(1, 0.43, 0.45, 0.0, 65536.0);
        staticBandsInfos["LANDSAT8-OLI2_BLUE"] = SpectralSensorParams(2, 0.45, 0.51, 0.0, 65536.0);
        staticBandsInfos["LANDSAT8-OLI3_GREEN"] = SpectralSensorParams(3, 0.53, 0.59, 0.0, 65536.0);
        staticBandsInfos["LANDSAT8-OLI4_RED"] = SpectralSensorParams(4, 0.64, 0.67, 0.0, 65536.0);
        staticBandsInfos["LANDSAT8-OLI5_NIR"] = SpectralSensorParams(5, 0.85, 0.88, 0.0, 65536.0);
        staticBandsInfos["LANDSAT8-OLI6_SWIR1"] = SpectralSensorParams(6, 1.57, 1.65, 0.0, 65536.0);
        staticBandsInfos["LANDSAT8-OLI7_SWIR2"] = SpectralSensorParams(7, 2.11, 2.29, 0.0, 65536.0);
        staticBandsInfos["LANDSAT8-OLI8_PAN"] = SpectralSensorParams(8, 0.50, 0.68, 0.0, 65536.0);
        staticBandsInfos["LANDSAT8-OLI9_CIRRUS"] = SpectralSensorParams(9, 1.36, 1.38, 0.0, 65536.0);

        staticBandsInfos["LANDSAT8-TIRS1"] = SpectralSensorParams(10, 10.60, 11.19, 0.0, 65536.0);
        staticBandsInfos["LANDSAT8-TIRS2"] = SpectralSensorParams(11, 11.50, 12.51, 0.0, 65536.0);

        staticBandsInfos["CBERS4_PAN_PAN"] = SpectralSensorParams(1, 0.51, 0.85, 0.0, 255.0);
        staticBandsInfos["CBERS4_PAN_GREEN"] = SpectralSensorParams(2, 0.52, 0.59, 0.0, 255.0);
        staticBandsInfos["CBERS4_PAN_RED"] = SpectralSensorParams(3, 0.63, 0.69, 0.0, 255.0);
        staticBandsInfos["CBERS4_PAN_NIR"] = SpectralSensorParams(4, 0.77, 0.89, 0.0, 255.0);

        staticBandsInfos["CBERS4_MUX_BLUE"] = SpectralSensorParams(5, 0.45, 0.52, 0.0, 255.0);
        staticBandsInfos["CBERS4_MUX_GREEN"] = SpectralSensorParams(6, 0.52, 0.59, 0.0, 255.0);
        staticBandsInfos["CBERS4_MUX_RED"] = SpectralSensorParams(7, 0.63, 0.69, 0.0, 255.0);
        staticBandsInfos["CBERS4_MUX_NIR"] = SpectralSensorParams(8, 0.77, 0.89, 0.0, 255.0);

        staticBandsInfos["CBERS4_IRS_PAN"] = SpectralSensorParams(9, 0.50, 0.90, 0.0, 255.0);
        staticBandsInfos["CBERS4_IRS_SWIR_B10"] = SpectralSensorParams(10, 1.55, 1.75, 0.0, 255.0);
        staticBandsInfos["CBERS4_IRS_SWIR_B11"] = SpectralSensorParams(11, 2.08, 2.35, 0.0, 255.0);
        staticBandsInfos["CBERS4_IRS_TH"] = SpectralSensorParams(12, 10.40, 12.50, 0.0, 255.0);

        staticBandsInfos["CBERS4_WFI_BLUE"] = SpectralSensorParams(13, 0.45, 0.52, 0.0, 1024.0);
        staticBandsInfos["CBERS4_WFI_GREEN"] = SpectralSensorParams(14, 0.52, 0.59, 0.0, 1024.0);
        staticBandsInfos["CBERS4_WFI_RED"] = SpectralSensorParams(15, 0.63, 0.69, 0.0, 1024.0);
        staticBandsInfos["CBERS4_WFI_NIR"] = SpectralSensorParams(16, 0.77, 0.89, 0.0, 1024.0);
        
        staticBandsInfos["SENTINEL2A_MSI_BAND1"] = SpectralSensorParams(
          1, 0.442 - 0.021, 0.442 + 0.021, 0, 4095 );        
        staticBandsInfos["SENTINEL2A_MSI_BAND2"] = SpectralSensorParams(
          2, 0.492 - 0.066, 0.492 + 0.066, 0, 4095 );
        staticBandsInfos["SENTINEL2A_MSI_BAND3"] = SpectralSensorParams(
          3, 0.559 - 0.036, 0.559 + 0.036, 0, 4095 );
        staticBandsInfos["SENTINEL2A_MSI_BAND4"] = SpectralSensorParams(
          4, 0.664 - 0.031, 0.664 + 0.031, 0, 4095 );
        staticBandsInfos["SENTINEL2A_MSI_BAND5"] = SpectralSensorParams(
          5, 0.704 - 0.015, 0.704 + 0.015, 0, 4095 );        
        staticBandsInfos["SENTINEL2A_MSI_BAND6"] = SpectralSensorParams(
          6, 0.740 - 0.015, 0.740 + 0.015, 0, 4095 );        
        staticBandsInfos["SENTINEL2A_MSI_BAND7"] = SpectralSensorParams(
          7, 0.782 - 0.020, 0.782 + 0.020, 0, 4095 );        
        staticBandsInfos["SENTINEL2A_MSI_BAND8"] = SpectralSensorParams(
          8, 0.832 - 0.106, 0.832 + 0.106, 0, 4095 );
        staticBandsInfos["SENTINEL2A_MSI_BAND9"] = SpectralSensorParams(
          9, 0.945 - 0.020, 0.945 + 0.020, 0, 4095 );
        staticBandsInfos["SENTINEL2A_MSI_BAND10"] = SpectralSensorParams(
          10, 1.373 - 0.031, 1.373 + 0.031, 0, 4095 );      
        staticBandsInfos["SENTINEL2A_MSI_BAND11"] = SpectralSensorParams(
          11, 1.613 - 0.091, 1.613 + 0.091, 0, 4095 );
        staticBandsInfos["SENTINEL2A_MSI_BAND12"] = SpectralSensorParams(
          12, 2.202 - 0.175, 2.202 + 0.175, 0, 4095 );
        
        staticBandsInfos["AMAZONIA1_WFI_B1"] = SpectralSensorParams(
          1, 0.45, 0.52, 0, 1023 );
        staticBandsInfos["AMAZONIA1_WFI_B2"] = SpectralSensorParams(
          2, 0.52, 0.59, 0, 1023 );
        staticBandsInfos["AMAZONIA1_WFI_B3"] = SpectralSensorParams(
          3, 0.63, 0.69, 0, 1023 );
        staticBandsInfos["AMAZONIA1_WFI_B4"] = SpectralSensorParams(
          4, 0.77, 0.89, 0, 1023 );
      }
      
      specBandInfos = staticBandsInfos;
    }

    std::pair<double, double> GetDigitalNumberBandInfo(std::string bandName)
    {
      std::map<std::string, SpectralSensorParams > params = getSensorParams();
      std::map<std::string, SpectralSensorParams > ::iterator it = params.find(bandName);
      if (it == params.end())
        return std::pair<double, double>(0.0, 255.0);

      return std::pair<double, double>(it->second.m_min, it->second.m_max);
    }

    double GetDigitalNumberBandMax(std::string bandName)
    {
      return GetDigitalNumberBandInfo(bandName).second;
    }

    double GetDigitalNumberBandMin(std::string bandName)
    {
      return GetDigitalNumberBandInfo(bandName).first;
    }

    bool NormalizeRaster(te::rst::Raster& inraster, double nmin, double nmax)
    {
      if (nmin > nmax)
        return false;

    // computing old maximuns and minimums for each band
      std::vector<double> omins;
      std::vector<double> omaxs;
      std::vector<unsigned int> bands;
      std::vector<double> dummy;
      
      te::rst::RasterSummary summary;
      te::rst::RasterSummaryManager::getInstance().get(&inraster, false, summary );      

      for (unsigned int b = 0; b < inraster.getNumberOfBands(); b++)
      {
        const std::complex<double>* cmin = summary.at(b).m_minVal;
        const std::complex<double>* cmax = summary.at(b).m_maxVal;

        omins.push_back(cmin->real());
        omaxs.push_back(cmax->real());
        dummy.push_back(inraster.getBand(b)->getProperty()->m_noDataValue);

        bands.push_back(b);
      }

    // calculating amplitudes to avoid multiple subtractions
      double value;
      const double namplitude = nmax - nmin;
      std::vector<double> oamplitude;
      for (unsigned int b = 0; b < inraster.getNumberOfBands(); b++)
        oamplitude.push_back(omaxs[b] - omins[b]);

    // iterating over raster to normalize pixel values
      te::rst::RasterIterator<double> it = te::rst::RasterIterator<double>::begin(&inraster, bands);
      te::rst::RasterIterator<double> itend = te::rst::RasterIterator<double>::end(&inraster, bands);

      while (it != itend)
      {
        for (unsigned int b = 0; b < inraster.getNumberOfBands(); b++)
        {
          value = (*it)[b];
          if (value != dummy[b])
            value = ((*it)[b] - omins[b]) * namplitude / oamplitude[b] + nmin;

          inraster.setValue(it.getColumn(), it.getRow(), value, b);
        }

        ++it;
      }

      return true;
    }

    te::rst::Raster* NormalizeRaster(const te::rst::Raster* inputRaster)
    {
      double nmin = 0.0;
      double nmax = 255.0;

      // computing old maximuns and minimums for each band
      std::vector<double> omins;
      std::vector<double> omaxs;
      std::vector<unsigned int> bands;
      std::vector<double> dummy;
      
      te::rst::RasterSummary summary;
      te::rst::RasterSummaryManager::getInstance().get(inputRaster, false, summary );

      for (unsigned int b = 0; b < inputRaster->getNumberOfBands(); b++)
      {
        const std::complex<double>* cmin = summary.at(b).m_minVal;
        const std::complex<double>* cmax = summary.at(b).m_maxVal;

        omins.push_back(cmin->real());
        omaxs.push_back(cmax->real());
        dummy.push_back(inputRaster->getBand(b)->getProperty()->m_noDataValue);

        bands.push_back(b);
      }

      // calculating amplitudes to avoid multiple subtractions
      double value;
      const double namplitude = nmax - nmin;
      std::vector<double> oamplitude;
      for (unsigned int b = 0; b < inputRaster->getNumberOfBands(); b++)
        oamplitude.push_back(omaxs[b] - omins[b]);

      // iterating over raster to normalize pixel values
      te::rst::RasterIterator<double> it = te::rst::RasterIterator<double>::begin((te::rst::Raster*)inputRaster, bands);
      te::rst::RasterIterator<double> itend = te::rst::RasterIterator<double>::end((te::rst::Raster*)inputRaster, bands);

      std::vector<te::rst::BandProperty*> bandsProperties;

      for (unsigned i = 0; i < inputRaster->getNumberOfBands(); i++)
      {
        te::rst::BandProperty* bprop = new te::rst::BandProperty(*(inputRaster->getBand(i)->getProperty()));
        bprop->m_type = te::dt::UCHAR_TYPE;
        bprop->m_noDataValue = inputRaster->getBand(i)->getProperty()->m_noDataValue;
        bandsProperties.push_back(bprop);
      }

      std::map< std::string, std::string > rInfo;

      te::rst::Raster* outputRaster = nullptr;
      try
      {
        outputRaster = te::rst::RasterFactory::make("MEM", new te::rst::Grid(*inputRaster->getGrid()), bandsProperties, rInfo);
      }
      catch(...)
      {
        TERP_LOGERR( "Raster allocation error" );
        return nullptr;
      }

      while (it != itend)
      {
        for (unsigned int b = 0; b < inputRaster->getNumberOfBands(); b++)
        {
          value = (*it)[b];
          if (value != dummy[b])
            value = ((*it)[b] - omins[b]) * namplitude / oamplitude[b] + nmin;

          outputRaster->setValue(it.getColumn(), it.getRow(), value, b);
        }

        ++it;
      }

      double rasterValue;

      for (unsigned r = 0; r < inputRaster->getNumberOfRows(); r++)
      {
        for (unsigned c = 0; c < inputRaster->getNumberOfColumns(); c++)
        {
          for (unsigned b = 0; b < inputRaster->getNumberOfBands(); b++)
          {
            inputRaster->getValue(c, r, rasterValue, b);

            if (rasterValue != inputRaster->getBand(b)->getProperty()->m_noDataValue)
               rasterValue = 255 * (rasterValue - omins[b]) / (omaxs[b] - omins[b]);

            outputRaster->setValue(c, r, rasterValue, b);
          }
        }
      }

      return outputRaster;
    }
  
    bool ConvertRGB2IHS( const te::rst::Raster& inputRGBRaster, 
      const unsigned int redBandIdx, const unsigned int greenBandIdx,
      const unsigned int blueBandIdx, const double rgbRangeMin, 
      const double rgbRangeMax, te::rst::Raster& outputIHSRaster ) 
    {
      return ConvertRGB2IHS( 
        inputRGBRaster, 
        redBandIdx, 
        inputRGBRaster,
        greenBandIdx,
        inputRGBRaster,
        blueBandIdx,
        rgbRangeMin,
        rgbRangeMax,
        outputIHSRaster );
    }

    bool ConvertRGB2IHS(const te::rst::Raster& inputRedRaster, const unsigned int redBandIdx,
      const te::rst::Raster& inputGreenRaster, const unsigned int greenBandIdx,
      const te::rst::Raster& inputBlueRaster, const unsigned int blueBandIdx,
      const double rgbRangeMin, const double rgbRangeMax, te::rst::Raster& outputIHSRaster)
    {
      if ((inputRedRaster.getAccessPolicy() & te::common::RAccess) == 0 ||
          (inputGreenRaster.getAccessPolicy() & te::common::RAccess) == 0 ||
          (inputBlueRaster.getAccessPolicy() & te::common::RAccess) == 0)
        return false;
      if (redBandIdx > inputRedRaster.getNumberOfBands()) return false;
      if (greenBandIdx > inputGreenRaster.getNumberOfBands()) return false;
      if (blueBandIdx > inputBlueRaster.getNumberOfBands()) return false;
      if ((outputIHSRaster.getAccessPolicy() & te::common::WAccess) == 0)
        return false;
      if (outputIHSRaster.getNumberOfBands() < 3) return false;
      if (inputRedRaster.getNumberOfColumns() != inputGreenRaster.getNumberOfColumns() ||
          inputRedRaster.getNumberOfColumns() != inputBlueRaster.getNumberOfColumns())
        return false;
      if (inputRedRaster.getNumberOfRows() != inputGreenRaster.getNumberOfRows() ||
          inputRedRaster.getNumberOfRows() != inputBlueRaster.getNumberOfRows())
        return false;
      if (inputRedRaster.getNumberOfColumns() != outputIHSRaster.getNumberOfColumns())
        return false;
      if (inputRedRaster.getNumberOfRows() != outputIHSRaster.getNumberOfRows())
        return false;

      const te::rst::Band& redBand = *inputRedRaster.getBand(redBandIdx);
      const te::rst::Band& greenBand = *inputGreenRaster.getBand(greenBandIdx);
      const te::rst::Band& blueBand = *inputBlueRaster.getBand(blueBandIdx);

      te::rst::Band& iBand = *outputIHSRaster.getBand(0);
      te::rst::Band& hBand = *outputIHSRaster.getBand(1);
      te::rst::Band& sBand = *outputIHSRaster.getBand(2);

      const unsigned int outNCols = outputIHSRaster.getNumberOfColumns();
      const unsigned int outNRows = outputIHSRaster.getNumberOfRows();
      const double redNoData = redBand.getProperty()->m_noDataValue;
      const double greenNoData = greenBand.getProperty()->m_noDataValue;
      const double blueNoData = blueBand.getProperty()->m_noDataValue;
      const double iNoData = iBand.getProperty()->m_noDataValue;
      const double hNoData = hBand.getProperty()->m_noDataValue;
      const double sNoData = sBand.getProperty()->m_noDataValue;

      unsigned int outRow = 0;
      unsigned int outCol = 0;
      double red = 0;
      double green = 0;
      double blue = 0;
      double teta = 0;
      double redNorm = 0, greenNorm = 0, blueNorm = 0;
      double rMinusG = 0, rMinusB = 0;
      double rgbSum = 0;
      double cosValue = 0;
      const double twoPi = 2.0 * (static_cast<double>(M_PI));
      const double pi2 = (static_cast<double>(M_PI)) / 2.0;
      const double rgbNormFac = (rgbRangeMax == rgbRangeMin) ? 0.0 :
        (1.0 / (rgbRangeMax - rgbRangeMin));
      double intensity = 0;
      double hue = 0;
      double saturation = 0;

      for (outRow = 0; outRow < outNRows; ++outRow)
      {
        for (outCol = 0; outCol < outNCols; ++outCol)
        {
          redBand.getValue(outCol, outRow, red);
          greenBand.getValue(outCol, outRow, green);
          blueBand.getValue(outCol, outRow, blue);

          if ((red == redNoData) || (green == greenNoData) ||
            (blue == blueNoData))
          {
            iBand.setValue(outCol, outRow, iNoData);
            hBand.setValue(outCol, outRow, hNoData);
            sBand.setValue(outCol, outRow, sNoData);
          }
          else
          {
            if ((red == green) && (green == blue))
            { // Gray scale case
              // From Wikipedia:
              // h = 0 is used for grays though the hue has no geometric 
              // meaning there, where the saturation s = 0. Similarly, 
              // the choice of 0 as the value for s when l is equal to 0 or 1 
              // is arbitrary.        

              hue = 0.0;
              saturation = 0.0;
              intensity = (red * rgbNormFac); // or green or blue since they all are the same.
            }
            else
            { // Color case
              redNorm = (red - rgbRangeMin) * rgbNormFac;
              greenNorm = (green - rgbRangeMin) * rgbNormFac;
              blueNorm = (blue - rgbRangeMin) * rgbNormFac;

              rMinusG = redNorm - greenNorm;
              rMinusB = redNorm - blueNorm;

              cosValue = std::sqrt((rMinusG * rMinusG) + (rMinusB *
                (greenNorm - blueNorm)));

              if (cosValue == 0.0)
              {
                teta = pi2;
              }
              else
              {
                cosValue = (0.5 * (rMinusG + rMinusB)) /
                  cosValue;
                teta = std::acos(cosValue);
              }

              assert((cosValue >= (-1.0)) && (cosValue <= (1.0)));

              if (blueNorm > greenNorm)
              {
                hue = (twoPi - teta);
              }
              else
              {
                hue = teta;
              }

              rgbSum = (redNorm + greenNorm + blueNorm);

              saturation = (1.0 - (3 * std::min(std::min(redNorm, greenNorm), blueNorm) /
                rgbSum));

              intensity = (rgbSum / 3.0);
            }

            iBand.setValue(outCol, outRow, intensity);
            hBand.setValue(outCol, outRow, hue);
            sBand.setValue(outCol, outRow, saturation);
          }
        }
      }

      return true;
    }
    
    bool ConvertIHS2RGB( const te::rst::Raster& inputIHSRaster, 
      const unsigned int intensityBandIdx, const unsigned int hueBandIdx,
      const unsigned int saturationBandIdx, const double rgbRangeMin, 
      const double rgbRangeMax, te::rst::Raster& outputRGBRaster )
    {
      return ConvertIHS2RGB(
        inputIHSRaster,
        intensityBandIdx,
        inputIHSRaster,
        hueBandIdx,
        inputIHSRaster,
        saturationBandIdx,
        rgbRangeMin,
        rgbRangeMax,
        outputRGBRaster
      );
    }

    bool ConvertIHS2RGB(const te::rst::Raster& inputIRaster, const unsigned int intensityBandIdx,
      const te::rst::Raster& inputHRaster, const unsigned int hueBandIdx,
      const te::rst::Raster& inputSRaster, const unsigned int saturationBandIdx,
      const double rgbRangeMin, const double rgbRangeMax, te::rst::Raster& outputRGBRaster)
    {
      if ((inputIRaster.getAccessPolicy() & te::common::RAccess) == 0 ||
          (inputHRaster.getAccessPolicy() & te::common::RAccess) == 0 ||
          (inputSRaster.getAccessPolicy() & te::common::RAccess) == 0)
        return false;
      if (intensityBandIdx > inputIRaster.getNumberOfBands()) return false;
      if (hueBandIdx > inputHRaster.getNumberOfBands()) return false;
      if (saturationBandIdx > inputSRaster.getNumberOfBands()) return false;
      if (inputIRaster.getNumberOfColumns() != inputHRaster.getNumberOfColumns() ||
        inputIRaster.getNumberOfColumns() != inputSRaster.getNumberOfColumns())
        return false;
      if (inputIRaster.getNumberOfRows() != inputHRaster.getNumberOfRows() ||
        inputIRaster.getNumberOfRows() != inputSRaster.getNumberOfRows())
        return false;
      if ((outputRGBRaster.getAccessPolicy() & te::common::WAccess) == 0)
        return false;
      if (outputRGBRaster.getNumberOfBands() < 3) return false;
      if (inputIRaster.getNumberOfColumns() != outputRGBRaster.getNumberOfColumns())
        return false;
      if (inputIRaster.getNumberOfRows() != outputRGBRaster.getNumberOfRows())
        return false;

      const te::rst::Band& intensityBand = *inputIRaster.getBand(intensityBandIdx);
      const te::rst::Band& hueBand = *inputHRaster.getBand(hueBandIdx);
      const te::rst::Band& saturationBand = *inputSRaster.getBand(saturationBandIdx);
      te::rst::Band& redBand = *outputRGBRaster.getBand(0);
      te::rst::Band& greenBand = *outputRGBRaster.getBand(1);
      te::rst::Band& blueBand = *outputRGBRaster.getBand(2);

      const unsigned int nCols = outputRGBRaster.getNumberOfColumns();
      const unsigned int nRows = outputRGBRaster.getNumberOfRows();

      const double intensityNoData = intensityBand.getProperty()->m_noDataValue;
      const double hueNoData = hueBand.getProperty()->m_noDataValue;
      const double saturationNoData = saturationBand.getProperty()->m_noDataValue;

      const double rNoData = redBand.getProperty()->m_noDataValue;
      const double gNoData = greenBand.getProperty()->m_noDataValue;
      const double bNoData = blueBand.getProperty()->m_noDataValue;

      const double rgbNormFac = (rgbRangeMax == rgbRangeMin) ? 0.0 :
        (rgbRangeMax - rgbRangeMin);
      const double pi3 = M_PI / 3.0; // 60
      const double twoPi3 = 2.0 * M_PI / 3.0; // 120
      const double fourPi3 = 4.0 * M_PI / 3.0; // 240        
      unsigned int row = 0;
      unsigned int col = 0;
      double hue = 0;
      double lig = 0;
      double sat = 0;
      double red = 0;
      double green = 0;
      double blue = 0;


      for (row = 0; row < nRows; ++row)
      {
        for (col = 0; col < nCols; ++col)
        {
          intensityBand.getValue(col, row, lig);
          hueBand.getValue(col, row, hue);
          saturationBand.getValue(col, row, sat);

          if ((lig == intensityNoData) || (hue == hueNoData) ||
            (sat == saturationNoData))
          {
            redBand.setValue(col, row, rNoData);
            greenBand.setValue(col, row, gNoData);
            blueBand.setValue(col, row, bNoData);
          }
          else
          {
            if ((hue == 0.0) && (sat == 0.0))
            { // Gray scale case
              red = green = blue = (lig * rgbNormFac);
            }
            else
            { // color case
              /* Hue inside RG sector */
              if (hue < twoPi3)
              {
                blue = lig * (1.0 - sat);
                red = lig * (1.0 + (sat * std::cos(hue) /
                  std::cos(pi3 - hue)));
                green = (3.0 * lig) - (red + blue);
              }
              else if (hue < fourPi3)
              { /* Hue inside GB sector */

                hue -= twoPi3;

                red = lig * (1.0 - sat);
                green = lig * (1.0 + (sat * std::cos(hue) /
                  std::cos(pi3 - hue)));
                blue = (3.0 * lig) - (red + green);
              }
              else
              { /* Hue inside BR sector */

                hue -= fourPi3;

                green = lig * (1.0 - sat);
                blue = lig * (1.0 + (sat * std::cos(hue) /
                  std::cos(pi3 - hue)));
                red = (3.0 * lig) - (green + blue);
              }

              red = (red * rgbNormFac) + rgbRangeMin;
              green = (green * rgbNormFac) + rgbRangeMin;
              blue = (blue * rgbNormFac) + rgbRangeMin;
            }

            red = MIN(red, rgbRangeMax);
            green = MIN(green, rgbRangeMax);
            blue = MIN(blue, rgbRangeMax);

            red = MAX(red, rgbRangeMin);
            green = MAX(green, rgbRangeMin);
            blue = MAX(blue, rgbRangeMin);

            redBand.setValue(col, row, red);
            greenBand.setValue(col, row, green);
            blueBand.setValue(col, row, blue);
          }
        }
      }

      return true;
    }

    bool ConvertRGB2HLS(const te::rst::Raster& inputRGBRaster,
      const unsigned int redBandIdx, const unsigned int greenBandIdx,
      const unsigned int blueBandIdx, const double rgbRangeMin,
      const double rgbRangeMax, te::rst::Raster& outputHLSRaster)
    {
      return ConvertRGB2HLS(
        inputRGBRaster,
        redBandIdx,
        inputRGBRaster,
        greenBandIdx,
        inputRGBRaster,
        blueBandIdx,
        rgbRangeMin,
        rgbRangeMax,
        outputHLSRaster
      );
    }

    bool ConvertRGB2HLS(const te::rst::Raster& inputRedRaster, const unsigned int redBandIdx,
      const te::rst::Raster& inputGreenRaster, const unsigned int greenBandIdx,
      const te::rst::Raster& inputBlueRaster, const unsigned int blueBandIdx,
      const double /*rgbRangeMin*/, const double rgbRangeMax, te::rst::Raster& outputHLSRaster)
    {
      if ((inputRedRaster.getAccessPolicy() & te::common::RAccess) == 0 ||
          (inputGreenRaster.getAccessPolicy() & te::common::RAccess) == 0 ||
          (inputBlueRaster.getAccessPolicy() & te::common::RAccess) == 0)
        return false;
      if (redBandIdx > inputRedRaster.getNumberOfBands()) return false;
      if (greenBandIdx > inputGreenRaster.getNumberOfBands()) return false;
      if (blueBandIdx > inputBlueRaster.getNumberOfBands()) return false;
      if ((outputHLSRaster.getAccessPolicy() & te::common::WAccess) == 0)
        return false;
      if (outputHLSRaster.getNumberOfBands() < 3) return false;
      if (inputRedRaster.getNumberOfRows() != inputGreenRaster.getNumberOfRows() ||
        inputRedRaster.getNumberOfRows() != inputBlueRaster.getNumberOfRows())
        return false;
      if (inputRedRaster.getNumberOfColumns() != outputHLSRaster.getNumberOfColumns())
        return false;
      if (inputRedRaster.getNumberOfRows() != outputHLSRaster.getNumberOfRows())
        return false;

      const te::rst::Band& redBand = *inputRedRaster.getBand(redBandIdx);
      const te::rst::Band& greenBand = *inputGreenRaster.getBand(greenBandIdx);
      const te::rst::Band& blueBand = *inputBlueRaster.getBand(blueBandIdx);

      te::rst::Band& hBand = *outputHLSRaster.getBand(0);
      te::rst::Band& lBand = *outputHLSRaster.getBand(1);
      te::rst::Band& sBand = *outputHLSRaster.getBand(2);

      const unsigned int outNCols = outputHLSRaster.getNumberOfColumns();
      const unsigned int outNRows = outputHLSRaster.getNumberOfRows();
      const double redNoData = redBand.getProperty()->m_noDataValue;
      const double greenNoData = greenBand.getProperty()->m_noDataValue;
      const double blueNoData = blueBand.getProperty()->m_noDataValue;
      const double hNoData = hBand.getProperty()->m_noDataValue;
      const double lNoData = lBand.getProperty()->m_noDataValue;
      const double sNoData = sBand.getProperty()->m_noDataValue;      

      unsigned int outRow = 0;
      unsigned int outCol = 0;
      double red = 0;
      double green = 0;
      double blue = 0;
      double light = 0;
      double hue = 0;
      double saturation = 0;

      for (outRow = 0; outRow < outNRows; ++outRow)
      {
        for (outCol = 0; outCol < outNCols; ++outCol)
        {
          redBand.getValue(outCol, outRow, red);
          greenBand.getValue(outCol, outRow, green);
          blueBand.getValue(outCol, outRow, blue);
          
          if ((red == redNoData) || (green == greenNoData) ||
            (blue == blueNoData))
          {
            hBand.setValue(outCol, outRow, hNoData);
            lBand.setValue(outCol, outRow, lNoData);
            sBand.setValue(outCol, outRow, sNoData);
          }
          else
          {
            red /= rgbRangeMax;
            green /= rgbRangeMax;
            blue /= rgbRangeMax;

            double vmax = MAX(red, green);
            vmax = MAX(vmax, blue);

            double vmin = MIN(red, green);
            vmin = MIN(vmin, blue);

            hue = (vmax + vmin) / 2.;
            saturation = (vmax + vmin) / 2.;
            light = (vmax + vmin) / 2.;

            if (vmax == vmin)
            {// achromatic
              hue = 0.;
              saturation = 0.;
            }
            else
            {
              double diff = vmax - vmin;

              saturation = light > 0.5 ? diff / (2. - vmax - vmin) : diff / (vmax + vmin);

              if (vmax == red)
              {
                hue = (green - blue) / diff + (green < blue ? 6. : 0.);
              }
              else if (vmax == green)
              {
                hue = (blue - red) / diff + 2.;
              }
              else if (vmax == blue)
              {
                hue = (red - green) / diff + 4.;
              }

              hue *= 60.;
            }

            hBand.setValue(outCol, outRow, hue);
            lBand.setValue(outCol, outRow, light);
            sBand.setValue(outCol, outRow, saturation);
          }
        }
      }

      return true;
    }

    bool ConvertHLS2RGB(const te::rst::Raster& inputHLSRaster,
      const unsigned int hueBandIdx, const unsigned int lightBandIdx,
      const unsigned int saturationBandIdx, const double rgbRangeMin,
      const double rgbRangeMax, te::rst::Raster& outputRGBRaster)
    {
      return ConvertHLS2RGB( 
        inputHLSRaster,
        hueBandIdx,
        inputHLSRaster,
        lightBandIdx,
        inputHLSRaster,
        saturationBandIdx,
        rgbRangeMin,
        rgbRangeMax,
        outputRGBRaster
      );
    }

    bool ConvertHLS2RGB(const te::rst::Raster& inputHRaster, const unsigned int hueBandIdx,
      const te::rst::Raster& inputLRaster, const unsigned int lightBandIdx,
      const te::rst::Raster& inputSRaster, const unsigned int saturationBandIdx,
      const double /*rgbRangeMin*/, const double rgbRangeMax, te::rst::Raster& outputRGBRaster)
    {
      if ((inputHRaster.getAccessPolicy() & te::common::RAccess) == 0 ||
          (inputLRaster.getAccessPolicy() & te::common::RAccess) == 0 ||
          (inputSRaster.getAccessPolicy() & te::common::RAccess) == 0)
        return false;
      if (hueBandIdx > inputHRaster.getNumberOfBands()) return false;
      if (lightBandIdx > inputLRaster.getNumberOfBands()) return false;
      if (saturationBandIdx > inputSRaster.getNumberOfBands()) return false;
      if (inputHRaster.getNumberOfColumns() != inputLRaster.getNumberOfColumns() ||
          inputHRaster.getNumberOfColumns() != inputSRaster.getNumberOfColumns())
        return false;
      if (inputHRaster.getNumberOfRows() != inputLRaster.getNumberOfRows() ||
        inputHRaster.getNumberOfRows() != inputSRaster.getNumberOfRows())
        return false;
      if ((outputRGBRaster.getAccessPolicy() & te::common::WAccess) == 0)
        return false;
      if (outputRGBRaster.getNumberOfBands() < 3) return false;
      if (inputHRaster.getNumberOfColumns() != outputRGBRaster.getNumberOfColumns())
        return false;
      if (inputHRaster.getNumberOfRows() != outputRGBRaster.getNumberOfRows())
        return false;

      const te::rst::Band& hueBand = *inputHRaster.getBand(hueBandIdx);
      const te::rst::Band& lightBand = *inputLRaster.getBand(lightBandIdx);
      const te::rst::Band& saturationBand = *inputSRaster.getBand(saturationBandIdx);
      te::rst::Band& redBand = *outputRGBRaster.getBand(0);
      te::rst::Band& greenBand = *outputRGBRaster.getBand(1);
      te::rst::Band& blueBand = *outputRGBRaster.getBand(2);

      const unsigned int nCols = outputRGBRaster.getNumberOfColumns();
      const unsigned int nRows = outputRGBRaster.getNumberOfRows();

      const double hueNoData = hueBand.getProperty()->m_noDataValue;
      const double lightNoData = lightBand.getProperty()->m_noDataValue;
      const double saturationNoData = saturationBand.getProperty()->m_noDataValue;

      const double rNoData = redBand.getProperty()->m_noDataValue;
      const double gNoData = greenBand.getProperty()->m_noDataValue;
      const double bNoData = blueBand.getProperty()->m_noDataValue;
   
      unsigned int row = 0;
      unsigned int col = 0;
      double hue = 0;
      double light = 0;
      double saturation = 0;

      for (row = 0; row < nRows; ++row)
      {
        for (col = 0; col < nCols; ++col)
        {
          hueBand.getValue(col, row, hue);
          lightBand.getValue(col, row, light);
          saturationBand.getValue(col, row, saturation);

          if ((hue == hueNoData) || (light == lightNoData) ||
            (saturation == saturationNoData))
          {
            redBand.setValue(col, row, rNoData);
            greenBand.setValue(col, row, gNoData);
            blueBand.setValue(col, row, bNoData);
          }
          else
          {
            std::vector< double > rgb;
            hue = hue / 360.;

            if (saturation == 0)
            {// achromatic
              rgb.push_back(light * rgbRangeMax);
              rgb.push_back(light * rgbRangeMax);
              rgb.push_back(light * rgbRangeMax);
            }
            else
            {
              for (unsigned int i = 0; i < 3; ++i)
              {
                double q = light < 0.5 ? light * (1. + saturation) : light + saturation - light * saturation;
                double p = 2. * light - q;

                double t = 0;

                switch (i)
                {
                case 0://red
                {
                  t = hue + 1. / 3.;
                  break;
                }
                case 1://green
                {
                  t = hue;
                  break;
                }
                case 2://blue
                {
                  t = hue - 1. / 3.;
                  break;
                }
                }

                if (t < 0)
                {
                  t += 1.;
                }
                if (t > 1)
                {
                  t -= 1.;
                }

                if (t < (1. / 6.))
                {
                  p = (p + (q - p) * 6. * t);
                }
                else if (t < (1. / 2.))
                {
                  p = q;
                }
                else if (t < (2. / 3.))
                {
                  p = (p + (q - p) * (2. / 3. - t) * 6.);
                }

                rgb.push_back(p * rgbRangeMax);
              }
            }

            redBand.setValue(col, row, rgb[0]);
            greenBand.setValue(col, row, rgb[1]);
            blueBand.setValue(col, row, rgb[2]);
          }
        }
      }

      return true;
    }
    
    void GetMeanValueThread( GetMeanValueThreadParams* paramsPtr )
    {
      std::unique_lock< std::mutex > locker( *(paramsPtr->m_mutexPtr) );

      const unsigned int nBlocksX = paramsPtr->m_rasterBlocksStatusPtr->getColumnsNumber();
      const unsigned int nBlocksY = paramsPtr->m_rasterBlocksStatusPtr->getLinesNumber();
      const unsigned int& inputBandIndex = paramsPtr->m_inputBandIndex;
      const double noDataValue = paramsPtr->m_noDataValue;

      locker.unlock();

      const te::rst::SynchronizedRaster inputRaster( 1,
        *paramsPtr->m_rasterSyncPtr );
      const te::rst::Band& band = *inputRaster.getBand( inputBandIndex );
      const unsigned int maxBlkWidth = band.getProperty()->m_blkw;
      const unsigned int maxBlkHeight = band.getProperty()->m_blkh;
      const unsigned int nRows = inputRaster.getNumberOfRows();
      const unsigned int nCols = inputRaster.getNumberOfColumns();
      unsigned blkX = 0;
      double sum = 0;
      double elementsNumber = 0;
      unsigned int blkStartRow = 0;
      unsigned int blkStartCol = 0;
      unsigned int blkRowBound = 0;
      unsigned int blkColBound = 0;
      unsigned int row = 0 ;
      unsigned int col = 0;
      double value = 0;
      
      for( unsigned blkY = 0 ; blkY < nBlocksY ; ++blkY )
      {
        for( blkX = 0 ; blkX < nBlocksX ; ++blkX )
        {
          locker.lock();
          
          if( paramsPtr->m_rasterBlocksStatusPtr->operator()( blkY, blkX ) )
          {
            locker.unlock();
          }
          else
          {
            paramsPtr->m_rasterBlocksStatusPtr->operator()( blkY, blkX ) = true;
            
            locker.unlock();

            blkStartRow = std::min( maxBlkHeight * blkY, nRows );
            blkRowBound = std::min( maxBlkHeight * ( 1 + blkY ), nRows );
            blkStartCol = std::min( maxBlkWidth * blkX, nCols );
            blkColBound = std::min( maxBlkWidth * ( 1 + blkX ), nCols );

            sum = 0;
            elementsNumber = 0;

            for( row = blkStartRow ; row < blkRowBound ; ++row )
            {
              for( col = blkStartCol ; col < blkColBound ; ++col )
              {
                band.getValue( col, row, value );

                if( value != noDataValue )
                {
                  sum += value;
                  ++elementsNumber;
                }
              }
            }

            if( elementsNumber != 0 )
            {
              locker.lock();

              ( *(paramsPtr->m_meanValuePtr) ) += sum;
              ( *(paramsPtr->m_meanElementsNumberPtr) ) += elementsNumber;

              locker.unlock();
            }
          }
        }
      }
    }
    
    bool GetMeanValue( const te::rst::Raster& inputRaster,
      const unsigned int inputBandIndex,
      const unsigned int maxThreads, 
      const bool forceNoDataValue,
      const double noDataValue,      
      double& meanValue )
    {
      meanValue = 0;

      Matrix< bool > rasterBlocksStatus;
      if( ! rasterBlocksStatus.reset( static_cast<unsigned int>(
        inputRaster.getBand( inputBandIndex )->getProperty()->m_nblocksy),
        static_cast<unsigned int>(inputRaster.getBand(
        inputBandIndex )->getProperty()->m_nblocksx ), Matrix< bool >::RAMMemPol ) )
      {
        return false;
      }
      for( unsigned int row = 0 ; row < rasterBlocksStatus.getLinesNumber() ;
        ++row )
      {
        for( unsigned int col = 0 ; col < rasterBlocksStatus.getColumnsNumber() ;
          ++col )
        {
          rasterBlocksStatus( row, col ) = false;
        }
      }
      
      te::rst::RasterSynchronizer rasterSync(
        *(te::rst::Raster*)(&inputRaster), te::common::RAccess );
      std::mutex mutex;
      double meanElementsNumber = 0;

      GetMeanValueThreadParams params;
      params.m_rasterSyncPtr = &rasterSync;
      params.m_inputBandIndex = inputBandIndex;
      params.m_rasterBlocksStatusPtr = &rasterBlocksStatus;
      params.m_meanValuePtr = &meanValue;
      params.m_meanElementsNumberPtr = &meanElementsNumber;
      params.m_returnStatus = true;
      params.m_mutexPtr = &mutex;
      params.m_noDataValue = forceNoDataValue ? noDataValue : 
        inputRaster.getBand( inputBandIndex )->getProperty()->m_noDataValue;
      
      if( maxThreads == 1 )
      {
        GetMeanValueThread( &params );
      }
      else
      {
        unsigned int threadsNumber = maxThreads ? maxThreads : te::common::GetPhysProcNumber();
        
        boost::thread_group threads;
          
        for( unsigned int threadIdx = 0 ; threadIdx < threadsNumber ;
          ++threadIdx )
        {
          threads.add_thread( new boost::thread( GetMeanValueThread, 
             &params ) );
        };  
        
        threads.join_all();
      }

      if( meanElementsNumber != 0 )
      {
        meanValue /= meanElementsNumber;
      }
      
      return params.m_returnStatus;
    }
    
    void GetStdDevValueThread( GetStdDevValueThreadParams* paramsPtr )
    {
      std::unique_lock< std::mutex > locker( *paramsPtr->m_mutexPtr );

      const unsigned int nBlocksX = paramsPtr->m_rasterBlocksStatusPtr->getColumnsNumber();
      const unsigned int nBlocksY = paramsPtr->m_rasterBlocksStatusPtr->getLinesNumber();
      const unsigned int& inputBandIndex = paramsPtr->m_inputBandIndex;
      const double noDataValue = paramsPtr->m_noDataValue;

      locker.unlock();

      const te::rst::SynchronizedRaster inputRaster( 1,
        *paramsPtr->m_rasterSyncPtr );
      const te::rst::Band& band = *inputRaster.getBand( inputBandIndex );
      const double& meanValue = paramsPtr->m_meanValue;

      unsigned blkX = 0;
      double diff = 0;
      double squaresDifsSum = 0;
      double elementsNumber = 0;
      const unsigned int maxBlkWidth = band.getProperty()->m_blkw;
      const unsigned int maxBlkHeight = band.getProperty()->m_blkh;
      const unsigned int nRows = inputRaster.getNumberOfRows();
      const unsigned int nCols = inputRaster.getNumberOfColumns();
      unsigned int blkStartRow = 0;
      unsigned int blkStartCol = 0;
      unsigned int blkRowBound = 0;
      unsigned int blkColBound = 0;
      unsigned int row = 0 ;
      unsigned int col = 0;
      double value = 0;
      
      for( unsigned blkY = 0 ; blkY < paramsPtr->m_rasterBlocksStatusPtr->getLinesNumber() ;
        ++blkY )
      {
        for( blkX = 0 ; blkX < paramsPtr->m_rasterBlocksStatusPtr->getColumnsNumber() ;
          ++blkX )
        {
          locker.lock();
          
          if( paramsPtr->m_rasterBlocksStatusPtr->operator()( blkY, blkX ) )
          {
            locker.unlock();
          }
          else
          {
            paramsPtr->m_rasterBlocksStatusPtr->operator()( blkY, blkX ) = true;
            
            locker.unlock();

            blkStartRow = std::min( maxBlkHeight * blkY, nRows );
            blkRowBound = std::min( maxBlkHeight * ( 1 + blkY ), nRows );
            blkStartCol = std::min( maxBlkWidth * blkX, nCols );
            blkColBound = std::min( maxBlkWidth * ( 1 + blkX ), nCols );

            for( row = blkStartRow ; row < blkRowBound ; ++row )
            {
              for( col = blkStartCol ; col < blkColBound ; ++col )
              {
                band.getValue( col, row, value );

                if( value != noDataValue )
                {
                  diff = value - meanValue;
                  diff *= diff;
                  squaresDifsSum += diff;
                  ++elementsNumber;
                }
              }
            }
          }
        }
      }
      
      locker.lock();

      ( *(paramsPtr->m_stdDevValuePtr) ) += squaresDifsSum;
      ( *( paramsPtr->m_pixelsNumberValuePtr ) ) += elementsNumber;

      locker.unlock();
    }
    
    bool GetStdDevValue( const te::rst::Raster& inputRaster,
      const unsigned int inputBandIndex, const unsigned int maxThreads, 
      const bool forceNoDataValue, const double noDataValue,
      double const * const meanValuePtr, 
      double& stdDevValue )
    {
      stdDevValue = 0;

      double mean = 0;
      if( meanValuePtr )
      {
        mean = (*meanValuePtr);
      }
      else
      {
        if( ! GetMeanValue( inputRaster, inputBandIndex, maxThreads, false, 0, 
          mean ) )
        {
          return false;
        }
      }
            
      Matrix< bool > rasterBlocksStatus;
      if( ! rasterBlocksStatus.reset( static_cast<unsigned int>(
        inputRaster.getBand( inputBandIndex )->getProperty()->m_nblocksy),
        static_cast<unsigned int>(inputRaster.getBand(
        inputBandIndex )->getProperty()->m_nblocksx), Matrix< bool >::RAMMemPol ) )
      {
        return false;
      }
      for( unsigned int row = 0 ; row < rasterBlocksStatus.getLinesNumber() ;
        ++row )
      {
        for( unsigned int col = 0 ; col < rasterBlocksStatus.getColumnsNumber() ;
          ++col )
        {
          rasterBlocksStatus( row, col ) = false;
        }
      }
      
      te::rst::RasterSynchronizer rasterSync( *(te::rst::Raster*)(&inputRaster),
        te::common::RAccess );
      double pixelsNumber = 0.0;
      std::mutex mutex;
      
      GetStdDevValueThreadParams params;
      params.m_rasterSyncPtr = &rasterSync;
      params.m_inputBandIndex = inputBandIndex;
      params.m_rasterBlocksStatusPtr = &rasterBlocksStatus;
      params.m_stdDevValuePtr = &stdDevValue;
      params.m_pixelsNumberValuePtr = &pixelsNumber;
      params.m_meanValue = mean;
      params.m_returnStatus = true;
      params.m_mutexPtr = &mutex;
      params.m_noDataValue = forceNoDataValue ? noDataValue : 
        inputRaster.getBand( inputBandIndex )->getProperty()->m_noDataValue;      
      
      if( maxThreads == 1 )
      {
        GetStdDevValueThread( &params );
      }
      else
      {
        unsigned int threadsNumber = maxThreads ? maxThreads : te::common::GetPhysProcNumber();
        
        boost::thread_group threads;
          
        for( unsigned int threadIdx = 0 ; threadIdx < threadsNumber ;
          ++threadIdx )
        {
          threads.add_thread( new boost::thread( GetStdDevValueThread, 
             &params ) );
        };  
        
        threads.join_all();
      }
      
      stdDevValue = std::sqrt( stdDevValue / pixelsNumber );
      
      return params.m_returnStatus;
    }    
    
    void GetCovarianceValueThreadEntry( te::rp::GetCovarianceValueThreadParameters* paramsPtr )
    {
      std::unique_lock< std::mutex > locker( *paramsPtr->m_mutexPtr );
      
      const unsigned int inputBandIndex1 = paramsPtr->m_inputBandIndex1;
      const unsigned int inputBandIndex2 = paramsPtr->m_inputBandIndex2;
      const double mean1 = paramsPtr->m_mean1Value;
      const double mean2 = paramsPtr->m_mean2Value;
      std::vector< std::tuple< bool,  unsigned int, unsigned int > >&
        raster1BlocksStatus = *paramsPtr->m_raster1BlocksStatusPtr;
      const unsigned int raster2NBlocksX = paramsPtr->m_raster2NBlocksX;
      const std::size_t raster1BlocksStatusSize = raster1BlocksStatus.size();

      locker.unlock();

      const te::rst::SynchronizedRaster inputRaster1( 1,
        *paramsPtr->m_rasterSyncPtr1 );
      const te::rst::SynchronizedRaster inputRaster2( raster2NBlocksX,
        *paramsPtr->m_rasterSyncPtr2 );      
      
      const te::rst::Band& band1 = *inputRaster1.getBand( inputBandIndex1 );
      const te::rst::Band& band2 = *inputRaster2.getBand( inputBandIndex2 );
      
      const double noDataValue1 = band1.getProperty()->m_noDataValue;
      const double noDataValue2 = band2.getProperty()->m_noDataValue;
      
      const unsigned int maxRaster1BlkWidth = band1.getProperty()->m_blkw;
      const unsigned int maxRaster1BlkHeight = band1.getProperty()->m_blkh;      
      
      const unsigned int raster1NRows = inputRaster1.getNumberOfRows();
      const unsigned int raster1NCols = inputRaster1.getNumberOfColumns();      
      
      unsigned int currRaster1BlkX = 0;
      unsigned int currRaster1BlkY = 0;
      unsigned int blkStartRow = 0;
      unsigned int blkStartCol = 0;
      unsigned int blkRowBound = 0;
      unsigned int blkColBound = 0;
      double diffSum = 0;
      unsigned int elementsNumber = 0;
      unsigned int row = 0 ;
      unsigned int col = 0;
      double value1 = 0;
      double value2 = 0;      
      
      for( std::size_t raster1BlocksStatusIdx = 0 ; raster1BlocksStatusIdx <
        raster1BlocksStatusSize ; ++raster1BlocksStatusIdx )
      {
        locker.lock();

        // looking for the next unprocessed roi

        if( !std::get< 0 >( raster1BlocksStatus[ raster1BlocksStatusIdx ] ) )
        {
          std::get< 0 >( raster1BlocksStatus[ raster1BlocksStatusIdx ] ) = true;
          currRaster1BlkX = std::get< 1 >( raster1BlocksStatus[ 
            raster1BlocksStatusIdx ] );
          currRaster1BlkY = std::get< 2 >( raster1BlocksStatus[ 
            raster1BlocksStatusIdx ] );

          locker.unlock();
          
          blkStartRow = std::min( maxRaster1BlkHeight * currRaster1BlkY, 
            raster1NRows );
          blkRowBound = std::min( maxRaster1BlkHeight * ( 1 + currRaster1BlkY ), 
            raster1NRows );
          
          blkStartCol = std::min( maxRaster1BlkWidth * currRaster1BlkX, 
            raster1NCols );
          blkColBound = std::min( maxRaster1BlkWidth * ( 1 + currRaster1BlkX ), 
            raster1NCols );                

          diffSum = 0;
          elementsNumber = 0;

          for( row = blkStartRow ; row < blkRowBound ; ++row )
          {
            for( col = blkStartCol ; col < blkColBound ; ++col )
            {
              band1.getValue( col, row, value1 );
              band2.getValue( col, row, value2 );

              if(
                  ( value1 != noDataValue1 )
                  &&
                  ( value2 != noDataValue2 )
                )
              {
                diffSum += ( ( value1 - mean1 ) * ( value2 - mean2 ) );
                ++elementsNumber;
              }
            }
          }

          if( elementsNumber != 0 )
          {
            locker.lock();

            ( *(paramsPtr->m_covarianceValuePtr) ) += diffSum;
            ( *(paramsPtr->m_elementsNumberPtr) ) += elementsNumber;

            locker.unlock();
          }
        }
        else
        {
          locker.unlock();
        }
      }
    }    
    
    bool GetCovarianceValue( const te::rst::Raster& inputRaster1,
      const unsigned int inputBandIndex1, const te::rst::Raster& inputRaster2,
      const unsigned int inputBandIndex2, const unsigned int maxThreads,
      double const * const mean1ValuePtr, double const * const mean2ValuePtr,  
      double& covarianceValue )
    {
      covarianceValue = 0;
      
      if( inputBandIndex1 >= inputRaster1.getNumberOfBands() ) return false;
      if( inputBandIndex2 >= inputRaster2.getNumberOfBands() ) return false;

      if( inputRaster1.getNumberOfColumns()
          !=
          inputRaster2.getNumberOfColumns()
        )
      {
        return false;
      }
      if( inputRaster1.getNumberOfRows()
          !=
          inputRaster2.getNumberOfRows()
        )
      {
        return false;
      }
      
      // means

      double mean1 = 0;
      if( mean1ValuePtr )
      {
        mean1 = (*mean1ValuePtr);
      }
      else
      {
        if( ! GetMeanValue( inputRaster1, inputBandIndex1, maxThreads, false, 0, 
          mean1 ) )
        {
          return false;
        }
      }
      
      double mean2 = 0;
      if( mean2ValuePtr )
      {
        mean2 = (*mean2ValuePtr);
      }
      else
      {
        if( ! GetMeanValue( inputRaster2, inputBandIndex2, maxThreads, 
          false, 0, mean2 ) )
        {
          return false;
        }
      }

      // defining the raster blocks to process
      // std::get< 0 > == processed status;
      // std::get< 1 > == block X;
      // std::get< 2 > == block Y;
      std::vector< std::tuple< bool,  unsigned int, unsigned int > >
        raster1BlocksStatus;
        
      { 
        const te::rst::Band& raster1Band = *inputRaster1.getBand( 
          inputBandIndex1 );
        std::tuple< bool,  unsigned int, unsigned int > tuple;
        std::get< 0 >( tuple ) = false;
        
        for( unsigned int blkX = 0 ; 
          blkX < (unsigned int)raster1Band.getProperty()->m_nblocksx ; 
          ++blkX )
        {
          for( unsigned int blkY = 0 ; 
            blkY < (unsigned int)raster1Band.getProperty()->m_nblocksy ; 
            ++blkY )
          {
            std::get< 1 >( tuple ) = blkX;
            std::get< 2 >( tuple ) = blkY;

            raster1BlocksStatus.push_back( tuple );
          }
        }
      }
      
      // processing each block
        
      te::rst::RasterSynchronizer rasterSync1(
        *(te::rst::Raster*)(&inputRaster1), te::common::RAccess );
      te::rst::RasterSynchronizer rasterSync2(
        *(te::rst::Raster*)(&inputRaster2), te::common::RAccess );      
      std::mutex mutex;
      unsigned int elementsNumber = 0;
      
      GetCovarianceValueThreadParameters params;
      params.m_rasterSyncPtr1 = &rasterSync1;
      params.m_rasterSyncPtr2 = &rasterSync2;
      params.m_inputBandIndex1 = inputBandIndex1;
      params.m_inputBandIndex2 = inputBandIndex2;
      params.m_returnStatus = true;
      params.m_mutexPtr = &mutex;
      params.m_mean1Value = mean1;
      params.m_mean2Value = mean2;
      params.m_covarianceValuePtr = &covarianceValue;
      params.m_raster1BlocksStatusPtr = &raster1BlocksStatus;
      params.m_raster2NBlocksX = inputRaster2.getBand( 
        inputBandIndex2 )->getProperty()->m_nblocksx;
      params.m_elementsNumberPtr = &elementsNumber;

      if( maxThreads == 1 )
      {
        GetCovarianceValueThreadEntry( &params );
      }
      else
      {
        unsigned int threadsNumber = maxThreads ? maxThreads :
          te::common::GetPhysProcNumber();        
        std::vector< std::shared_ptr< std::thread > > threads;

        for( unsigned int threadIdx = 0 ; threadIdx < threadsNumber ;
          ++threadIdx )
        {
          std::shared_ptr< std::thread > threadPtr( new std::thread(
            &GetCovarianceValueThreadEntry, &params ) );

          threads.push_back( threadPtr );
        };

        for( unsigned int threadIdx = 0 ; threadIdx < threadsNumber ;
          ++threadIdx )
        {
          threads[ threadIdx ]->join();
        };
      }
      
      if( elementsNumber )
      {
        covarianceValue /= ((double)elementsNumber);
      }
      
      return params.m_returnStatus;
    }    
    
    bool DirectPrincipalComponents( 
      const te::rst::Raster& inputRaster,
      const std::vector< unsigned int >& inputRasterBands,
      boost::numeric::ublas::matrix< double >& pcaMatrix,
      te::rst::Raster& pcaRaster,
      const std::vector< unsigned int >& pcaRasterBands,
      const unsigned int maxThreads )
    {
      if( ( inputRaster.getAccessPolicy() & te::common::RAccess ) == 0 )
      {
        return false;
      }
      if( ( pcaRaster.getAccessPolicy() & te::common::WAccess ) == 0 )
      {
        return false;
      }   
      if( pcaRaster.getNumberOfBands() != inputRasterBands.size() )
      {
        return false;
      }
      if( inputRaster.getNumberOfRows() != pcaRaster.getNumberOfRows() )
      {
        return false;
      }
      if( inputRaster.getNumberOfColumns() != pcaRaster.getNumberOfColumns() )
      {
        return false;
      }    

      // PCA Matrix

      if( pcaMatrix.size1() == 0 )
      {
        // Covariance matrix

        boost::numeric::ublas::matrix< double > covarianceMatrix;

        covarianceMatrix.resize( inputRasterBands.size(), inputRasterBands.size() );

        for( unsigned int covMatrixIdx1 = 0 ; covMatrixIdx1 < inputRasterBands.size() ;
          ++covMatrixIdx1 )
        {
          if( inputRasterBands[ covMatrixIdx1 ] >= inputRaster.getNumberOfBands() )
          {
            return false;
          }

          for( unsigned int covMatrixIdx2 = 0 ; covMatrixIdx2 < inputRasterBands.size() ;
          ++covMatrixIdx2 )
          {
            if( inputRasterBands[ covMatrixIdx2 ] >= inputRaster.getNumberOfBands() )
            {
              return false;
            }

            if( covMatrixIdx1 > covMatrixIdx2 )
            {
              covarianceMatrix( covMatrixIdx1, covMatrixIdx2 ) =
                covarianceMatrix( covMatrixIdx2, covMatrixIdx1 );
            }
            else
            {
              if( ! GetCovarianceValue(
                      inputRaster,
                      inputRasterBands[ covMatrixIdx1 ],
                      inputRaster,
                      inputRasterBands[ covMatrixIdx2 ],
                      maxThreads, // max threads
                      nullptr, // mean 1 pointer
                      nullptr, // mean 2 pointer
                      covarianceMatrix( covMatrixIdx1, covMatrixIdx2 ) ) )
              {
                return false;
              }
            }
          }
        }

        //std::cout << std::endl << "Covariance matrix:" << covarianceMatrix << std::endl;

        // Eigen stuff

        boost::numeric::ublas::matrix< double > eigenValues;
        boost::numeric::ublas::matrix< double > eigenVectors;

        if( ! te::common::EigenVectors( covarianceMatrix, eigenVectors, eigenValues ) )
        {
          return false;
        }

        pcaMatrix = boost::numeric::ublas::trans( eigenVectors );
      }
      
      return RemapValues( inputRaster, inputRasterBands, pcaMatrix, pcaRaster, 
        pcaRasterBands, maxThreads );
    }
    
    bool InversePrincipalComponents( 
      const te::rst::Raster& pcaRaster,
      const boost::numeric::ublas::matrix< double >& pcaMatrix,
      te::rst::Raster& outputRaster,
      const std::vector< unsigned int >& outputRasterBands,
      const unsigned int maxThreads )    
    {
      boost::numeric::ublas::matrix< double > inversePcaMatrix;
      if( ! te::common::GetInverseMatrix( pcaMatrix, inversePcaMatrix ) )
      {
        return false;
      }
      
      std::vector< unsigned int > inputRasterBands;
      
      for( unsigned int bandIdx = 0 ; bandIdx < pcaRaster.getNumberOfBands() ;
        ++bandIdx )
      {
        inputRasterBands.push_back( bandIdx );
      }
      
      return RemapValues( pcaRaster, inputRasterBands, inversePcaMatrix, 
        outputRaster, outputRasterBands, maxThreads );
    }
    
    void RemapValuesThread( RemapValuesThreadParams* paramsPtr )
    {
      std::unique_lock< std::mutex > locker( *paramsPtr->m_mutexPtr );
      
      const std::vector< unsigned int >  inputRasterBands = 
        *( paramsPtr->m_inputRasterBandsPtr );
      const std::vector< unsigned int >  outputRasterBands = 
        *( paramsPtr->m_outputRasterBandsPtr );
      const boost::numeric::ublas::matrix< double > remapMatrix =
        *paramsPtr->m_remapMatrixPtr;
      Matrix< bool >& outputRasterBlocksStatus = 
        *paramsPtr->m_outputRasterBlocksStatusPtr;
      const unsigned int inputRasterMaxNBlocksX = paramsPtr->m_inputRasterMaxNBlocksX;
      
      locker.unlock();
      
      const te::rst::SynchronizedRaster inputRaster( inputRasterMaxNBlocksX *
        inputRasterBands.size(), *paramsPtr->m_inputRasterSync );      
      te::rst::SynchronizedRaster outputRaster( outputRasterBands.size(),
        *paramsPtr->m_outputRasterSync );            
      
      std::vector< double > inputBandsNoDataValues( inputRasterBands.size(), 
        0.0 );
      std::vector< double > outputBandsNoDataValues( inputRasterBands.size(), 
        0.0 );
      std::vector< double > outputMinValue( outputRasterBands.size(), 0 );
      std::vector< double > outputMaxValue( outputRasterBands.size(), 0 );        
      unsigned int inputRasterBandsIdx = 0;
      unsigned int outputBlkW = 0;
      unsigned int outputBlkH = 0;
      
      for( inputRasterBandsIdx = 0 ; inputRasterBandsIdx < outputRasterBands.size() ;
        ++inputRasterBandsIdx )
      {
        const unsigned int& inBandIdx = inputRasterBands[ inputRasterBandsIdx ];
        const unsigned int& outBandIdx = outputRasterBands[ inputRasterBandsIdx ];
        
        const te::rst::Band& inBand = *inputRaster.getBand( inBandIdx );
        const te::rst::Band& outBand = *outputRaster.getBand( outBandIdx );
        
        inputBandsNoDataValues[ inputRasterBandsIdx ] = 
          inBand.getProperty()->m_noDataValue;
        outputBandsNoDataValues[ inputRasterBandsIdx ] = 
          outBand.getProperty()->m_noDataValue;
          
        te::rst::GetDataTypeRanges( outBand.getProperty()->getType(), 
          outputMinValue[ inputRasterBandsIdx ],
          outputMaxValue[ inputRasterBandsIdx ] );
        
        if( inputRasterBandsIdx == 0 )
        {
          outputBlkW = outBand.getProperty()->m_blkw;
          outputBlkH = outBand.getProperty()->m_blkh;
        }
      }

      const unsigned int inputRasterBandsSize = inputRasterBands.size();
      const unsigned int nRows = outputRaster.getNumberOfRows();
      const unsigned int nCols = outputRaster.getNumberOfColumns();
      boost::numeric::ublas::matrix< double > pixelValues( 
        inputRasterBands.size(), 1 );
      boost::numeric::ublas::matrix< double > remappedPixelValues( 
        inputRasterBands.size(), 1 );
      bool elementIsValid = false;
      unsigned int blkStartRow = 0;
      unsigned int blkStartCol = 0;
      unsigned int blkRowBound = 0;
      unsigned int blkColBound = 0;
      unsigned int row = 0;
      unsigned int col = 0;
      double pixelValue = 0;
      
      for( unsigned blkY = 0 ; blkY < 
        paramsPtr->m_outputRasterBlocksStatusPtr->getLinesNumber() ;
        ++blkY )
      {
        for( unsigned int blkX = 0 ; 
          blkX < paramsPtr->m_outputRasterBlocksStatusPtr->getColumnsNumber() ;
          ++blkX )
        {
          locker.lock();
          
          if( paramsPtr->m_outputRasterBlocksStatusPtr->operator()( blkY, blkX ) )
          {
            locker.unlock();
          }
          else
          {
            paramsPtr->m_outputRasterBlocksStatusPtr->operator()( blkY, blkX ) = true;
            
            locker.unlock();
            
            blkStartRow = std::min( outputBlkH * blkY, nRows );
            blkRowBound = std::min( outputBlkH * ( 1 + blkY ), nRows );
            
            blkStartCol = std::min( outputBlkW * blkX, nCols );
            blkColBound = std::min( outputBlkW * ( 1 + blkX ), nCols );        
            
            for( row = blkStartRow ; row < blkRowBound ; ++row )
            {
              for( col = blkStartCol ; col < blkColBound ; ++col )
              {
                elementIsValid = true;
                
                for( inputRasterBandsIdx = 0 ; inputRasterBandsIdx < 
                  inputRasterBandsSize ; ++inputRasterBandsIdx )
                {    
                  inputRaster.getValue( col, row, pixelValue, inputRasterBands[ 
                    inputRasterBandsIdx ]);
                  
                  if( pixelValue == inputBandsNoDataValues[ inputRasterBandsIdx ] )
                  {
                    elementIsValid = false;
                    break;
                  }
                  
                  pixelValues( inputRasterBandsIdx, 0 ) = pixelValue;
                }
                
                if( elementIsValid )
                {
                  remappedPixelValues = boost::numeric::ublas::prod( remapMatrix, 
                    pixelValues );

                  for( inputRasterBandsIdx = 0 ; inputRasterBandsIdx < 
                    inputRasterBandsSize ; ++inputRasterBandsIdx )
                  { 
                    pixelValue = remappedPixelValues( inputRasterBandsIdx, 0 );
                    pixelValue = std::max( pixelValue, outputMinValue[ 
                      inputRasterBandsIdx ] );
                    pixelValue = std::min( pixelValue, outputMaxValue[ 
                      inputRasterBandsIdx ] );
                      
                    outputRaster.setValue( col, row, pixelValue, outputRasterBands[ 
                      inputRasterBandsIdx ] );
                  }             
                }
                else
                {
                  for( inputRasterBandsIdx = 0 ; inputRasterBandsIdx < 
                    inputRasterBandsSize ; ++inputRasterBandsIdx )
                  {    
                    outputRaster.setValue( col, row, outputBandsNoDataValues[ 
                      inputRasterBandsIdx ], outputRasterBands[ 
                      inputRasterBandsIdx ] );
                  }
                }
              }
            }
          }
        }
      }     
    }

    bool SelectiveReplaceValues(
      const te::rst::Raster& inputRaster,
      const unsigned int& inputRasterBandIdx,
      const std::vector< std::pair< double, double > >& targetValues,
      const bool enableProgress,
      const std::vector< te::gm::Polygon* >& restrictionPols,
      te::rst::Raster& outputRaster,
      const unsigned int& outputRasterBandIdx )
    {
      if( ( inputRaster.getAccessPolicy() & te::common::RAccess ) == 0 )
      {
        return false;
      }
      if( ( outputRaster.getAccessPolicy() & te::common::WAccess ) == 0 )
      {
        return false;
      }
      if( inputRaster.getNumberOfColumns() != outputRaster.getNumberOfColumns() )
      {
        return false;
      }
      if( inputRaster.getNumberOfRows() != outputRaster.getNumberOfRows() )
      {
        return false;
      }
      if( targetValues.empty() )
      {
        return false;
      }
      if( inputRasterBandIdx >= inputRaster.getNumberOfBands() )
      {
        return false;
      }
      if( outputRasterBandIdx >= outputRaster.getNumberOfBands() )
      {
        return false;
      }

      // remapping

      if( restrictionPols.empty() )
      {
        const te::rst::Band& inputBand = *inputRaster.getBand( inputRasterBandIdx );
        te::rst::Band& outputBand = *outputRaster.getBand( outputRasterBandIdx );
        const unsigned int blkw = (unsigned int)inputBand.getProperty()->m_blkw;
        const unsigned int blkh = (unsigned int)inputBand.getProperty()->m_blkh;
        const unsigned int nblksx = (unsigned int)inputBand.getProperty()->m_nblocksx;
        const unsigned int nblksy = (unsigned int)inputBand.getProperty()->m_nblocksy;
        const unsigned int nCols = inputRaster.getNumberOfColumns();
        const unsigned int nRows = inputRaster.getNumberOfRows();
        const std::size_t targetValuesSize = targetValues.size();

        unsigned int blkYIdx = 0;
        unsigned int blkStartRow = 0;
        unsigned int blkRowBound = 0;
        unsigned int blkStartCol = 0;
        unsigned int blkColBound = 0;
        unsigned int col = 0;
        unsigned int row = 0;
        std::size_t targetValuesIdx = 0;
        double value = 0;

        std::unique_ptr< te::common::TaskProgress > progressPtr;
        if( enableProgress )
        {
          progressPtr.reset( new te::common::TaskProgress );
          progressPtr->setMessage( "Remapping" );
          progressPtr->setTotalSteps( static_cast<int>( nblksx * nblksy ) );
        }

        for( unsigned int blkXIdx = 0 ; blkXIdx < nblksx ; ++blkXIdx )
        {
          blkStartCol = std::min( blkXIdx * blkw, nCols );

          blkColBound = std::min( blkStartCol + blkw, nCols );

          for( blkYIdx = 0 ; blkYIdx < nblksy ; ++blkYIdx )
          {
            blkStartRow = blkYIdx * blkh;

            blkRowBound = std::min( blkStartRow + blkh, nRows );

            for( col = blkStartCol ; col < blkColBound ; ++col )
            {
              for( row = blkStartRow ; row < blkRowBound ; ++row )
              {
                inputBand.getValue( col, row, value );

                for( targetValuesIdx = 0 ; targetValuesIdx < targetValuesSize ;
                  ++targetValuesIdx )
                {
                  if( value == targetValues[ targetValuesIdx ].first )
                  {
                    value = targetValues[ targetValuesIdx ].second;
                    break;
                  }
                }

                outputBand.setValue( col, row, value );
              }
            }

            if( enableProgress )
            {
              progressPtr->pulse();
              if( ! progressPtr->isActive() ) return false;
            }
          }
        }
      }
      else
      {
        const std::size_t restrictionPolsSize = restrictionPols.size();
        const te::rst::Band& inputBand = *inputRaster.getBand( inputRasterBandIdx );
        const int srid = inputRaster.getSRID();
        const std::size_t targetValuesSize = targetValues.size();
        const unsigned int nCols = inputRaster.getNumberOfColumns();
        const unsigned int nRows = inputRaster.getNumberOfRows();
        const unsigned int blkw = (unsigned int)inputBand.getProperty()->m_blkw;
        const unsigned int blkh = (unsigned int)inputBand.getProperty()->m_blkh;
        const unsigned int nblksx = (unsigned int)inputBand.getProperty()->m_nblocksx;
        const unsigned int nblksy = (unsigned int)inputBand.getProperty()->m_nblocksy;

        te::rst::Band& outputBand = *outputRaster.getBand( outputRasterBandIdx );
        te::rst::PolygonIterator< double > polIt;
        te::rst::PolygonIterator< double > polItEnd;
        te::gm::Polygon* polPtr;
        std::unique_ptr< te::gm::Polygon > reprojPolPtr;
        unsigned int col = 0;
        unsigned int row = 0;
        double value = 0;
        unsigned int blkYIdx = 0;
        unsigned int blkStartRow = 0;
        unsigned int blkRowBound = 0;
        unsigned int blkStartCol = 0;
        unsigned int blkColBound = 0;
        std::size_t targetValuesIdx = 0;

        std::unique_ptr< te::common::TaskProgress > progressPtr;
        if( enableProgress )
        {
          progressPtr.reset( new te::common::TaskProgress );
          progressPtr->setMessage( "Remapping" );
          progressPtr->setTotalSteps( static_cast<int>( nblksx * nblksy ) +
            restrictionPolsSize );
        }

        try
        {
          // copying original data

          for( unsigned int blkXIdx = 0 ; blkXIdx < nblksx ; ++blkXIdx )
          {
            blkStartCol = std::min( blkXIdx * blkw, nCols );

            blkColBound = std::min( blkStartCol + blkw, nCols );

            for( blkYIdx = 0 ; blkYIdx < nblksy ; ++blkYIdx )
            {
              blkStartRow = blkYIdx * blkh;

              blkRowBound = std::min( blkStartRow + blkh, nRows );

              for( col = blkStartCol ; col < blkColBound ; ++col )
              {
                for( row = blkStartRow ; row < blkRowBound ; ++row )
                {
                  inputBand.getValue( col, row, value );
                  outputBand.setValue( col, row, value );
                }
              }

              if( enableProgress )
              {
                progressPtr->pulse();
                if( ! progressPtr->isActive() ) return false;
              }
            }
          }

          for( std::size_t restrictionPolsIdx = 0 ; restrictionPolsIdx <
            restrictionPolsSize ; ++restrictionPolsIdx )
          {
            polPtr = restrictionPols[ restrictionPolsIdx ];
            if( polPtr->getSRID() != srid )
            {
              reprojPolPtr.reset( (te::gm::Polygon*)polPtr->clone() );
              polPtr = reprojPolPtr.get();
              polPtr->transform( srid );
            }

            polIt = te::rst::PolygonIterator< double >::begin( &inputRaster,
              polPtr,te::rst::PolygonIterator< double >::ScanLineIterationT );
            polItEnd = te::rst::PolygonIterator< double >::end( &inputRaster,
              polPtr, te::rst::PolygonIterator< double >::ScanLineIterationT );

            while( polIt != polItEnd )
            {
              col = polIt.getColumn();
              row = polIt.getRow();

              inputBand.getValue( col, row, value );

              for( targetValuesIdx = 0 ; targetValuesIdx < targetValuesSize ;
                ++targetValuesIdx )
              {
                if( value == targetValues[ targetValuesIdx ].first )
                {
                  value = targetValues[ targetValuesIdx ].second;
                  break;
                }
              }

              outputBand.setValue( col, row, value );

              ++polIt;
            }

            if( enableProgress )
            {
              progressPtr->pulse();
              if( ! progressPtr->isActive() ) return false;
            }
          }
        }
        catch(...)
        {
          return false;
        }
      }

      return true;
    }

    bool RemapValues( 
      const te::rst::Raster& inputRaster,
      const std::vector< unsigned int >& inputRasterBands, 
      const boost::numeric::ublas::matrix< double >& remapMatrix,
      te::rst::Raster& outputRaster,
      const std::vector< unsigned int >& outputRasterBands,
      const unsigned int maxThreads )
    {
      if( ( inputRaster.getAccessPolicy() & te::common::RAccess ) == 0 )
      {
        return false;
      }
      if( ( outputRaster.getAccessPolicy() & te::common::WAccess ) == 0 )
      {
        return false;
      }   
      if( remapMatrix.size1() != inputRasterBands.size() )
      {
        return false;
      }
      if( remapMatrix.size2() != inputRasterBands.size() )
      {
        return false;
      }      
      if( outputRaster.getNumberOfBands() != inputRasterBands.size() )
      {
        return false;
      }
      if( inputRaster.getNumberOfRows() != outputRaster.getNumberOfRows() )
      {
        return false;
      }
      if( inputRaster.getNumberOfColumns() != outputRaster.getNumberOfColumns() )
      {
        return false;
      }    
      if( remapMatrix.size1() != outputRasterBands.size() )
      {
        return false;
      }
      if( remapMatrix.size2() != outputRasterBands.size() )
      {
        return false;
      }     
      for( unsigned int inputRasterBandsIdx = 0 ; inputRasterBandsIdx < inputRasterBands.size() ;
        ++inputRasterBandsIdx )
      {
        if( inputRasterBands[ inputRasterBandsIdx ] >= inputRaster.getNumberOfBands() )
        {
          return false;
        }    
      }
      for( unsigned int outputRasterBandsIdx = 0 ; outputRasterBandsIdx < outputRasterBands.size() ;
        ++outputRasterBandsIdx )
      {
        if( outputRasterBands[ outputRasterBandsIdx ] >= outputRaster.getNumberOfBands() )
        {
          return false;
        }    
      }      
      
      // remapping
      
      Matrix< bool > outputRasterBlocksStatus;
      if( ! outputRasterBlocksStatus.reset( 
        static_cast<unsigned int>(outputRaster.getBand( 
          outputRasterBands[ 0 ] )->getProperty()->m_nblocksy),
        static_cast<unsigned int>(outputRaster.getBand( 
          outputRasterBands[ 0 ] )->getProperty()->m_nblocksx),
        Matrix< bool >::RAMMemPol ) )
      {
        return false;
      }
      for( unsigned int row = 0 ; row < outputRasterBlocksStatus.getLinesNumber() ;
        ++row )
      {
        for( unsigned int col = 0 ; col < outputRasterBlocksStatus.getColumnsNumber() ;
          ++col )
        {
          outputRasterBlocksStatus( row, col ) = false;
        }
      }
      
      te::rst::RasterSynchronizer inputRasterSync(
        *(te::rst::Raster*)(&inputRaster), te::common::RAccess );      
      te::rst::RasterSynchronizer outputRasterSync(
        *(te::rst::Raster*)(&outputRaster), te::common::WAccess );        
      std::mutex mutex;
      
      unsigned int inputRasterMaxNBlocksX = 0;
      for( unsigned int inputRasterBandsIdx = 0 ; inputRasterBandsIdx < 
        inputRasterBands.size() ; ++inputRasterBandsIdx )
      {
        inputRasterMaxNBlocksX = std::max( inputRasterMaxNBlocksX,
          (unsigned int)inputRaster.getBand( inputRasterBands[ 
          inputRasterBandsIdx ] )->getProperty()->m_nblocksx );
      }      
      
      RemapValuesThreadParams params;
      params.m_inputRasterSync = &inputRasterSync;
      params.m_outputRasterSync = &outputRasterSync;
      params.m_inputRasterBandsPtr = &inputRasterBands;
      params.m_outputRasterBandsPtr = &outputRasterBands;
      params.m_remapMatrixPtr = &remapMatrix;
      params.m_outputRasterBlocksStatusPtr = &outputRasterBlocksStatus;
      params.m_returnStatus = true;
      params.m_mutexPtr = &mutex;
      params.m_inputRasterMaxNBlocksX = inputRasterMaxNBlocksX;
      
      if( maxThreads == 1 )
      {
        RemapValuesThread( &params );
      }
      else
      {
        unsigned int threadsNumber = maxThreads ? maxThreads : te::common::GetPhysProcNumber();
        
        boost::thread_group threads;
          
        for( unsigned int threadIdx = 0 ; threadIdx < threadsNumber ;
          ++threadIdx )
        {
          threads.add_thread( new boost::thread( RemapValuesThread, 
             &params ) );
        };  
        
        threads.join_all();
      }
      
      return params.m_returnStatus;        
    }
    
    bool DecomposeBands( 
      const te::rst::Raster& inputRaster,
      const std::vector< unsigned int >& inputRasterBands,
      const std::vector< std::map<std::string, std::string> > & outputRastersInfos,
      const std::string& outputDataSourceType,
      std::vector< boost::shared_ptr< te::rst::Raster > > & outputRastersPtrs )
    {
      outputRastersPtrs.clear();
      
      if( !( inputRaster.getAccessPolicy() & te::common::RAccess ) )
      {
        return false;
      }
      if( outputRastersInfos.size() != inputRasterBands.size() )
      {
        return false;
      }
      if( outputDataSourceType.empty() )
      {
        return false;
      }
      
      te::mem::CachedRaster cachedInputRaster( inputRaster,
        50, 0 );
      
      outputRastersPtrs.resize( inputRasterBands.size() );
      
      const unsigned int nRows = cachedInputRaster.getNumberOfRows();
      const unsigned int nCols = cachedInputRaster.getNumberOfColumns();
      
      for( unsigned int inputRasterBandsIdx = 0 ; inputRasterBandsIdx <
        inputRasterBands.size() ; ++inputRasterBandsIdx )
      {
        const unsigned int bandIdx = inputRasterBands[ inputRasterBandsIdx ];
        if( bandIdx >= cachedInputRaster.getNumberOfBands() )
        {
          return false;
        }
        const te::rst::Band& inBand = *(cachedInputRaster.getBand( bandIdx ));
        
        std::vector< te::rst::BandProperty* > bandsProperties;
        bandsProperties.push_back( new te::rst::BandProperty( 
          *( cachedInputRaster.getBand( bandIdx )->getProperty()) ) );
        bandsProperties[ 0 ]->m_blkw = nCols;
        bandsProperties[ 0 ]->m_blkh = 1;
        bandsProperties[ 0 ]->m_nblocksx = 1;
        bandsProperties[ 0 ]->m_nblocksy = nRows;
        
        try
        {
          outputRastersPtrs[ inputRasterBandsIdx].reset( 
            te::rst::RasterFactory::make( outputDataSourceType,
            new te::rst::Grid( *cachedInputRaster.getGrid() ), bandsProperties, 
            outputRastersInfos[ inputRasterBandsIdx ], nullptr, nullptr ) );
        }
        catch(...)
        {
          outputRastersPtrs[ inputRasterBandsIdx].reset();
        }
        if( outputRastersPtrs[ inputRasterBandsIdx].get() == nullptr ) return false;        
        
        
        te::rst::Band& outBand = *(outputRastersPtrs[ inputRasterBandsIdx]->getBand( 0 ));        
        const unsigned int nBlocksX = (unsigned int)outBand.getProperty()->m_nblocksx;
        const unsigned int nBlocksY = (unsigned int)outBand.getProperty()->m_nblocksy;
        const unsigned int blockW = (unsigned int)outBand.getProperty()->m_blkw;
        const unsigned int blockH = (unsigned int)outBand.getProperty()->m_blkh;
        unsigned int blockStartRow = 0 ;
        unsigned int blockStartCol = 0 ;
        unsigned int blockRowBound = 0 ;
        unsigned int blockColBound = 0 ;        
        unsigned int col = 0;
        unsigned int row = 0;
        double value = 0;
        
        for( unsigned int blockYIdx = 0 ; blockYIdx < nBlocksY ; ++blockYIdx )
        {
          blockStartRow = blockH * blockYIdx;
          blockRowBound = blockStartRow + blockH;
          blockStartRow = std::min( blockStartRow, nRows - 1 );
          blockRowBound = std::min( blockRowBound, nRows );
        
          for( unsigned int blockXIdx = 0 ; blockXIdx < nBlocksX ; ++blockXIdx )
          {
            blockStartCol = blockW * blockXIdx;
            blockColBound = blockStartCol + blockW;
            blockStartCol = std::min( blockStartCol, nCols - 1 );
            blockColBound = std::min( blockColBound, nCols );
            
            for( row = blockStartRow ; row < blockRowBound ; ++row )
            {
              for( col = blockStartCol ; col < blockColBound ; ++col )
              {
                inBand.getValue( col, row, value );
                outBand.setValue( col, row, value );
              }
            }
          }
        }
      }
      
      return true;
    }
    
    void ComposeBandsThread( ComposeBandsThreadParams* paramsPtr )
    {
      std::unique_lock< std::mutex > locker( *paramsPtr->m_mutexPtr );
      
      std::vector< std::unique_ptr< te::rst::RasterSynchronizer > >&
        inRastersSyncsPtrs = *paramsPtr->m_inRastersSyncsPtrsPtr;
      const std::vector< unsigned int > inputRasterBands = paramsPtr->m_inputRasterBands;
      Matrix< bool >& rasterBlocksStatus = *paramsPtr->m_rasterBlocksStatusPtr;
      const te::rst::Interpolator::Method interpMethod = paramsPtr->m_interpMethod;
      const bool allowNoDataPixels = paramsPtr->m_allowNoDataPixels;
      te::common::TaskProgress * const progressPtr = paramsPtr->m_progressPtr;;

      locker.unlock();

      te::rst::SynchronizedRaster outRaster( inputRasterBands.size(), 
        *paramsPtr->m_outRasterSyncPtr );
      
      std::vector< std::unique_ptr< te::rst::SynchronizedRaster > > inRastersPtrs;
      for( std::size_t inRastersSyncsPtrsIdx = 0 ; inRastersSyncsPtrsIdx <
        inRastersSyncsPtrs.size() ; ++inRastersSyncsPtrsIdx )
      {
        inRastersPtrs.push_back( std::unique_ptr< te::rst::SynchronizedRaster >(
          new te::rst::SynchronizedRaster( 9 * outRaster.getNumberOfBands(), 
          *inRastersSyncsPtrs[ inRastersSyncsPtrsIdx ] ) ) );
      }
      
      std::vector< std::unique_ptr< te::rst::Interpolator > > interpolatorsPtrs;
      for( std::size_t inRastersPtrsIdx = 0 ; inRastersPtrsIdx <
        inRastersPtrs.size() ; ++inRastersPtrsIdx )
      {
        interpolatorsPtrs.push_back( std::unique_ptr< te::rst::Interpolator >(
          new te::rst::Interpolator( inRastersPtrs[ inRastersPtrsIdx ].get(), 
          interpMethod ) ) );
      }
      
      std::vector< double> inNoDataValues;
      for( std::size_t inRastersPtrsIdx = 0 ; inRastersPtrsIdx <
        inRastersPtrs.size() ; ++inRastersPtrsIdx )
      {
        inNoDataValues.push_back( inRastersPtrs[ inRastersPtrsIdx ]->getBand(
          inputRasterBands[ inRastersPtrsIdx ] )->getProperty()->m_noDataValue );
      }
      
      std::vector< int > inSRIDS;
      for( std::size_t inRastersPtrsIdx = 0 ; inRastersPtrsIdx <
        inRastersPtrs.size() ; ++inRastersPtrsIdx )
      {
        inSRIDS.push_back( inRastersPtrs[ inRastersPtrsIdx ]->getSRID() );
      }      
      
      std::vector< std::unique_ptr< te::rst::Grid > > inGridsPtrs;
      for( std::size_t inRastersPtrsIdx = 0 ; inRastersPtrsIdx <
        inRastersPtrs.size() ; ++inRastersPtrsIdx )
      {
        inGridsPtrs.push_back( std::unique_ptr< te::rst::Grid >(
          new te::rst::Grid( *inRastersPtrs[ inRastersPtrsIdx ]->getGrid() ) ) );
      }      
      
      std::vector< std::unique_ptr< te::srs::Converter > > convertersPtrs;
      for( std::size_t inRastersPtrsIdx = 0 ; inRastersPtrsIdx <
        inRastersPtrs.size() ; ++inRastersPtrsIdx )
      {      
        if( outRaster.getSRID() == inSRIDS[ inRastersPtrsIdx ] )
        {
          convertersPtrs.push_back( std::unique_ptr< te::srs::Converter >() );
        }
        else
        {
          try
          {
            convertersPtrs.push_back( std::unique_ptr< te::srs::Converter >(
              new te::srs::Converter( outRaster.getSRID(), 
              inSRIDS[ inRastersPtrsIdx ] ) ) );
          }
          catch(...)
          {
            locker.lock();
            *paramsPtr->m_returnStatusPtr = false;
            return;
          }
        }
      }

      const int outSRID = outRaster.getSRID();
      const unsigned int outBlkW = outRaster.getBand( 0 )->getProperty()->m_blkw;
      const unsigned int outBlkH = outRaster.getBand( 0 )->getProperty()->m_blkh;
      const unsigned int outNRows = outRaster.getNumberOfRows();
      const unsigned int outNCols = outRaster.getNumberOfColumns();
      const unsigned int inRastersNmb = inRastersPtrs.size();
      const std::complex< double > outNoDataValue(
        outRaster.getBand( 0 )->getProperty()->m_noDataValue,
        outRaster.getBand( 0 )->getProperty()->m_noDataValue );
      const te::rst::Grid& outGrid = *outRaster.getGrid();
      unsigned blkX = 0;
      unsigned int outBlkStartRow = 0;
      unsigned int outBlkStartCol = 0;
      unsigned int outBlkRowBound = 0;
      unsigned int outBlkColBound = 0;
      unsigned int outRow = 0 ;
      unsigned int outCol = 0;
      double xOutCoord = 0;
      double yOutCoord = 0;
      double xInCoord = 0;
      double yInCoord = 0;      
      double inRow = 0;
      double inCol = 0;
      std::vector< std::complex< double > > inValues( inRastersNmb );
      std::size_t inRasterIdx = 0;
      bool isValidPixel = false;
      
      try
      {
        for( unsigned blkY = 0 ; blkY < paramsPtr->m_rasterBlocksStatusPtr->getLinesNumber() ;
          ++blkY )
        {
          outBlkStartRow = std::min( outBlkH * blkY, outNRows );
          outBlkRowBound = std::min( outBlkH * ( 1 + blkY ), outNRows );
          
          for( blkX = 0 ; blkX < paramsPtr->m_rasterBlocksStatusPtr->getColumnsNumber() ;
            ++blkX )
          {
            locker.lock();
            
            if( progressPtr )
            {
              if( progressPtr->isActive() )
              {
                progressPtr->pulse();
              }
              else
              {
                *paramsPtr->m_returnStatusPtr = false;
                return;
              }
            }
            
            if( paramsPtr->m_rasterBlocksStatusPtr->operator()( blkY, blkX ) )
            {
              locker.unlock();
            }
            else
            {
              paramsPtr->m_rasterBlocksStatusPtr->operator()( blkY, blkX ) = true;
              
              locker.unlock();

              outBlkStartCol = std::min( outBlkW * blkX, outNCols );
              outBlkColBound = std::min( outBlkW * ( 1 + blkX ), outNCols );
              
              for( outRow = outBlkStartRow ; outRow < outBlkRowBound ; ++outRow )
              {
                for( outCol = outBlkStartCol ; outCol < outBlkColBound ; ++outCol )
                {
                  outGrid.gridToGeo( (double)outCol, (double)outRow, xOutCoord, yOutCoord );
                  
                  isValidPixel = true;
                  
                  for( inRasterIdx = 0 ; inRasterIdx < inRastersNmb ; ++inRasterIdx )
                  {
                    if( outSRID == inSRIDS[ inRasterIdx ] )
                    {
                      inGridsPtrs[ inRasterIdx ]->geoToGrid( xOutCoord, yOutCoord, 
                        inCol, inRow );   
                    }
                    else
                    {
                      convertersPtrs[ inRasterIdx ]->convert( xOutCoord, yOutCoord, 
                        xInCoord, yInCoord );
                      inGridsPtrs[ inRasterIdx ]->geoToGrid( xInCoord, yInCoord, 
                        inCol, inRow );                    
                    }

                    interpolatorsPtrs[ inRasterIdx ]->getValue( inCol, inRow, 
                      inValues[ inRasterIdx ], inputRasterBands[ inRasterIdx ] );
                    
                    if( inValues[ inRasterIdx ].real() == inNoDataValues[ inRasterIdx ] )
                    {
                      isValidPixel = false;
                    }
                  }
                  
                  if( isValidPixel || allowNoDataPixels )
                  {
                    for( inRasterIdx = 0 ; inRasterIdx < inRastersNmb ; ++inRasterIdx )
                    {
                      outRaster.setValue( outCol, outRow, inValues[ inRasterIdx ], 
                        inRasterIdx );
                    }
                  }
                  else
                  {
                    for( inRasterIdx = 0 ; inRasterIdx < inRastersNmb ; ++inRasterIdx )
                    {
                      outRaster.setValue( outCol, outRow, outNoDataValue, 
                        inRasterIdx );
                    }
                  }
                }
              }
            }
          }
        }
      }
      catch(...)
      {
        locker.lock();
        *paramsPtr->m_returnStatusPtr = false;
        return;
      }
    }
    
    bool ComposeBands( 
      const std::vector< te::rst::Raster const * >& inputRasters,
      const std::vector< unsigned int >& inputRasterBands,
      const te::rst::Interpolator::Method& interpMethod,
      const std::map<std::string, std::string>& outputRasterInfo,
      const std::string& outputDataSourceType,
      const unsigned int maxThreads, 
      const bool allowNoDataPixels,
      const bool enableProgress,
      std::unique_ptr< te::rst::Raster >& outputRasterPtr )
    {
      outputRasterPtr.reset();
      
      if( inputRasterBands.size() != inputRasters.size() )
      {
        return false;
      }
      if( outputDataSourceType.empty() )
      {
        return false;
      }
      
      unsigned int threadsNumber = maxThreads ? maxThreads : 
        te::common::GetPhysProcNumber();
      
      // creating the output raster
      
      {
        te::rst::Raster const * inputRasterPtr = nullptr;
        std::unique_ptr< te::rst::Grid > outputGridPtr;
        std::vector< te::rst::BandProperty* > bandsProperties;
        
        for( std::size_t inputRastersIdx = 0 ; inputRastersIdx < inputRasters.size() ;
          ++inputRastersIdx )
        {
          inputRasterPtr = inputRasters[ inputRastersIdx ];
          
          if( inputRasterBands[ inputRastersIdx ] >= inputRasterPtr->getNumberOfBands() )
          {
            return false;
          }
          
          if( inputRastersIdx == 0 )
          {
            outputGridPtr.reset( new te::rst::Grid( *inputRasterPtr->getGrid() ) );
          }          

          bandsProperties.push_back( new te::rst::BandProperty( 
            *( inputRasterPtr->getBand( inputRasterBands[ inputRastersIdx ] )->getProperty()) ) );
          bandsProperties.back()->m_blkw = outputGridPtr->getNumberOfColumns();
          bandsProperties.back()->m_blkh = 1;
          bandsProperties.back()->m_nblocksx = 1;
          bandsProperties.back()->m_nblocksy = outputGridPtr->getNumberOfRows();
          bandsProperties.back()->m_noDataValue = bandsProperties[ 0 ]->m_noDataValue;
        }
        
        try
        {
          outputRasterPtr.reset( te::rst::RasterFactory::make( outputDataSourceType,
            outputGridPtr.release(), bandsProperties, outputRasterInfo, nullptr, nullptr ) );
        }
        catch(...)
        {
          outputRasterPtr.reset();
        }
        
        if( outputRasterPtr.get() == nullptr ) return false;
      }
      
      std::unique_ptr< te::common::TaskProgress > progressPtr;
      if( enableProgress )
      {
        te::rst::Band& outBand = *outputRasterPtr->getBand( 0 );
        
        progressPtr.reset( new te::common::TaskProgress );
        progressPtr->useMultiThread( threadsNumber != 1 );
        progressPtr->setMessage( "Composing" );
        progressPtr->setTotalSteps( threadsNumber * outBand.getProperty()->m_nblocksy
          * outBand.getProperty()->m_nblocksx );
      }                     
      
      // processing data
        
      {
        te::rst::RasterSynchronizer outRasterSync( *outputRasterPtr, 
          te::common::WAccess );
        
        std::vector< std::unique_ptr< te::rst::RasterSynchronizer > >
          inRastersSyncsPtrs;
        for( std::size_t inputRastersIdx = 0 ; inputRastersIdx < inputRasters.size() ;
          ++inputRastersIdx )
        {
          inRastersSyncsPtrs.push_back( std::unique_ptr< te::rst::RasterSynchronizer >(
            new te::rst::RasterSynchronizer( 
            *( ( te::rst::Raster*)inputRasters[ inputRastersIdx ] ),
            te::common::RAccess ) ) );
        }
        
        Matrix< bool > rasterBlocksStatus;
        {
          te::rst::Band& outBand = *outputRasterPtr->getBand( 0 );
          
          if( ! rasterBlocksStatus.reset( static_cast<unsigned int>(
            outBand.getProperty()->m_nblocksy),
            static_cast<unsigned int>(outBand.getProperty()->m_nblocksx ), 
            Matrix< bool >::RAMMemPol ) )
          {
            return false;
          }
          
          for( unsigned int row = 0 ; row < rasterBlocksStatus.getLinesNumber() ;
            ++row )
          {
            for( unsigned int col = 0 ; col < rasterBlocksStatus.getColumnsNumber() ;
              ++col )
            {
              rasterBlocksStatus( row, col ) = false;
            }
          }
        }        
        
        {
          bool returnStatus = true;
          std::mutex mutex;
          
          ComposeBandsThreadParams params;
          params.m_inRastersSyncsPtrsPtr = &inRastersSyncsPtrs;
          params.m_inputRasterBands = inputRasterBands;
          params.m_outRasterSyncPtr = &outRasterSync;
          params.m_rasterBlocksStatusPtr = &rasterBlocksStatus;
          params.m_returnStatusPtr = &returnStatus;
          params.m_mutexPtr = &mutex;
          params.m_interpMethod = interpMethod;
          params.m_allowNoDataPixels = allowNoDataPixels;
          params.m_progressPtr = progressPtr.get();
          
          if( maxThreads == 1 )
          {
            ComposeBandsThread( &params );
          }
          else
          {
            boost::thread_group threads;
              
            for( unsigned int threadIdx = 0 ; threadIdx < threadsNumber ;
              ++threadIdx )
            {
              threads.add_thread( new boost::thread( ComposeBandsThread, 
                 &params ) );
            };  
            
            threads.join_all();
          }
          
          return returnStatus;
        }
      }
      
      return true;
    }
    
    bool GetDetailedExtent( const te::rst::Grid& grid, te::gm::LinearRing& detailedExtent )
    {
      const unsigned int nCols = grid.getNumberOfColumns();
      const unsigned int nRows = grid.getNumberOfRows();
      if( ( nCols == 0 ) || ( nRows == 0 ) )
      {
        return false;
      }
      
      const unsigned int ringSize = ( 2 * ( nCols + 1 ) ) +
        ( 2 * ( nRows - 1 ) ) + 1;
      
      te::gm::LinearRing ring( ringSize , te::gm::LineStringType,
        grid.getSRID(), nullptr );

      const double lLX = grid.getExtent()->getLowerLeftX();
      const double lLY = grid.getExtent()->getLowerLeftY();
      const double uRX = grid.getExtent()->getUpperRightX();
      const double uRY = grid.getExtent()->getUpperRightY();
      const double& resX = grid.getResolutionX();
      const double& resY = grid.getResolutionY();        
      unsigned int ringIdx = 0;
      unsigned int row = 0;
      unsigned int col = 0;
      
      ring.setPoint( 0, lLX, uRY ); 
      
      for( col = 1 ; col <= nCols ; ++col )
      {
        ring.setPoint( ++ringIdx, lLX + ( (static_cast<double>( col ) ) * resX ), uRY );
      }

      for( row = 1 ; row <= nRows ; ++row )
      {
        ring.setPoint( ++ringIdx, uRX, uRY - ( (static_cast<double>( row ) ) * resY ) );
      } 
      
      for( col = 1 ; col <= nCols ; ++col )
      {
        ring.setPoint( ++ringIdx, uRX - ( (static_cast<double>( col ) ) * resX ), lLY );
      }    
      
      for( row = 1 ; row <= nRows ; ++row )
      {
        ring.setPoint( ++ringIdx, lLX, lLY + ( (static_cast<double>( row ) ) * resY ) );
      }

      // Ensuring correct closing
      ring.setPoint( ringSize - 1, lLX, uRY );
      
      detailedExtent = ring;
      
      return true;
    }
    
    bool GetIndexedDetailedExtent( const te::rst::Grid& grid, 
      te::gm::LinearRing& indexedDetailedExtent )
    {
      const unsigned int nCols = grid.getNumberOfColumns();
      const unsigned int nRows = grid.getNumberOfRows();
      if( ( nCols == 0 ) || ( nRows == 0 ) )
      {
        return false;
      }
      
      const unsigned int ringSize = ( 2 * ( nCols + 1 ) ) +
        ( 2 * ( nRows - 1 ) ) + 1;
      
      te::gm::LinearRing ring( ringSize , te::gm::LineStringType,
        0, nullptr );

      const double lLY = (static_cast<double>(nRows)) - 0.5;
      const double uRX = (static_cast<double>(nCols)) - 0.5;
      unsigned int ringIdx = 0;
      unsigned int row = 0;
      unsigned int col = 0;
      
      ring.setPoint( 0, -0.5, -0.5 );
      
      for( col = 0 ; col < nCols ; ++col )
      {
        ring.setPoint( ++ringIdx, 0.5 + (static_cast<double>(col)), (-0.5) );
      }

      for( row = 0 ; row < nRows ; ++row )
      {
        ring.setPoint( ++ringIdx, uRX, 0.5 + (static_cast<double>(row)) );
      } 
      
      for( col = 1; col <= nCols ; ++col )
      {
        ring.setPoint( ++ringIdx, uRX - (static_cast<double>(col)), lLY );
      }    
      
      for( row = 1 ; row <= nRows ; ++row )
      {
        ring.setPoint( ++ringIdx, (-0.5), lLY - (static_cast<double>(row)) );
      }       

      // Ensuring correct closing
      ring.setPoint( ringSize - 1, -0.5, -0.5 ); 
      
      indexedDetailedExtent = ring;
      
      return true;
    }
    
    boost::numeric::ublas::matrix< double > CreateWaveletAtrousFilter( 
      const WaveletAtrousFilterType& filterType )
    {
      boost::numeric::ublas::matrix< double > emptyFilter;
      
      switch( filterType )
      {
        case B3SplineWAFilter :
        {
          boost::numeric::ublas::matrix< double > internalFilter( 5, 5 );
          const double weight = 256;
          
          internalFilter(0,0) = 1/weight; internalFilter(0,1) = 4/weight; internalFilter(0,2) = 6/weight; internalFilter(0,3) = 4/weight; internalFilter(0,4) = 1/weight;
          internalFilter(1,0) = 4/weight; internalFilter(1,1) = 16/weight; internalFilter(1,2) = 24/weight; internalFilter(1,3) = 16/weight; internalFilter(1,4) = 4/weight;
          internalFilter(2,0) = 6/weight; internalFilter(2,1) = 24/weight; internalFilter(2,2) = 36/weight; internalFilter(2,3) = 24/weight; internalFilter(2,4) = 6/weight;
          internalFilter(3,0) = 4/weight; internalFilter(3,1) = 16/weight; internalFilter(3,2) = 24/weight; internalFilter(3,3) = 16/weight; internalFilter(3,4) = 4/weight;
          internalFilter(4,0) = 1/weight; internalFilter(4,1) = 4/weight; internalFilter(4,2) = 6/weight; internalFilter(4,3) = 4/weight; internalFilter(4,4) = 1/weight;
          
          return internalFilter;
        }
        case TriangleWAFilter :
        {
          boost::numeric::ublas::matrix< double > internalFilter( 3, 3 );
          const double weight = 16;
          
          internalFilter(0,0) = 1/weight; internalFilter(0,1) = 2/weight; internalFilter(0,2) = 1/weight;
          internalFilter(1,0) = 2/weight; internalFilter(1,1) = 4/weight; internalFilter(1,2) = 2/weight;
          internalFilter(2,0) = 1/weight; internalFilter(2,1) = 2/weight; internalFilter(2,2) = 1/weight;
          
          return internalFilter;
        }  
        default :
        {
          throw te::rp::Exception( "Invalid filter type" );
        }
      }
      
      return emptyFilter;
    }
    
    bool DirectWaveletAtrous( 
      const te::rst::Raster& inputRaster,
      const std::vector< unsigned int >& inputRasterBands,
      te::rst::Raster& waveletRaster,
      const unsigned int levelsNumber,
      const boost::numeric::ublas::matrix< double >& filter )
    {
      if( ! ( inputRaster.getAccessPolicy() & te::common::RAccess ) )
      {
        return false;
      }
      for( unsigned int inputRasterBandsIdx = 0 ; inputRasterBandsIdx <
        inputRasterBands.size() ; ++inputRasterBandsIdx )
      {
        if( inputRasterBands[ inputRasterBandsIdx ] >= inputRaster.getNumberOfBands() )
        {
          return false;
        }
      }
      if( 
          ( ( waveletRaster.getAccessPolicy() & te::common::WAccess ) == 0 )
          ||
          ( waveletRaster.getNumberOfColumns() != inputRaster.getNumberOfColumns() )
          ||
          ( waveletRaster.getNumberOfRows() != inputRaster.getNumberOfRows() )
          ||
          ( waveletRaster.getNumberOfBands() < ( 2 *  levelsNumber * inputRasterBands.size() ) )
        )
      {
        return false;
      }
      if( levelsNumber == 0 )
      {
        return false;
      }
      if( ( filter.size1() == 0 ) || ( filter.size2() == 0 ) || 
        ( filter.size1() != filter.size2() )
      )
      {
        return false;
      }
      
      // Creating the coeficients and the resitual planes for each required
      // wavelet level

      const int filterWidth = static_cast<int>(filter.size1());
      const int offset = filterWidth / 2;
      const int nLines = static_cast<int>(inputRaster.getNumberOfRows());
      const int nCols = static_cast<int>(inputRaster.getNumberOfColumns());
      
      for( unsigned int inputRasterBandsIdx = 0 ; inputRasterBandsIdx < inputRasterBands.size() ;
        ++inputRasterBandsIdx )
      {
        for( unsigned int levelIndex = 0; levelIndex < levelsNumber; ++levelIndex)
        {
          const unsigned int currentSmoothBandIdx = ( 2 * levelsNumber * 
            inputRasterBandsIdx ) + ( 2 * levelIndex );
          te::rst::Band& currentSmoothBand = *waveletRaster.getBand( currentSmoothBandIdx );
          
          const unsigned int currentWaveletBandIdx = currentSmoothBandIdx + 1;
          te::rst::Band& currentWaveletBand = *waveletRaster.getBand(
            currentWaveletBandIdx );
          
          const unsigned int prevSmoothBandIdx =  currentSmoothBandIdx - 2;
          te::rst::Band const* prevSmoothBandPtr = nullptr;
          if( levelIndex == 0 )
          {
            prevSmoothBandPtr = inputRaster.getBand( inputRasterBands[ inputRasterBandsIdx ] );
          }
          else
          {
            prevSmoothBandPtr = waveletRaster.getBand( prevSmoothBandIdx );
          }          
          const te::rst::Band& prevSmoothBand = *prevSmoothBandPtr;          
          
          const int filterScale = static_cast<int>(std::pow(2.0, static_cast<double>(levelIndex)));
          
          int col = 0;
          int row = 0;
          int convolutionCenterCol = 0;
          int convolutionCenterRow = 0;
          int filterCol = 0;
          int filterRow = 0; 
          double valueOriginal = 0.0;
          double valueNew = 0.0;          

          for (convolutionCenterRow = 0; convolutionCenterRow < nLines; convolutionCenterRow++)
          {
            for (convolutionCenterCol = 0; convolutionCenterCol < nCols; convolutionCenterCol++)
            {
              valueNew = 0.0;
              
              for (filterRow = 0; filterRow < filterWidth; filterRow++)
              {
                for (filterCol = 0; filterCol < filterWidth; filterCol++)
                {
                  col = convolutionCenterCol+(filterCol-offset)*filterScale;
                  row = convolutionCenterRow+(filterRow-offset)*filterScale;
                  
                  if (col < 0)
                    col = nCols + col;
                  else if (col >= nCols)
                    col = col - nCols;
                  if (row < 0)
                    row = nLines + row;
                  else if (row >= nLines)
                    row = row - nLines;
                  
                  prevSmoothBand.getValue( static_cast<unsigned int>(col), static_cast<unsigned int>(row), valueOriginal );
                  
                  valueNew += valueOriginal * filter( static_cast<size_t>(filterRow), static_cast<size_t>(filterCol) );
                }
              }
              
              currentSmoothBand.setValue( static_cast<unsigned int>(convolutionCenterCol), static_cast<unsigned int>(convolutionCenterRow),
                valueNew );
              prevSmoothBand.getValue( static_cast<unsigned int>(convolutionCenterCol), static_cast<unsigned int>(convolutionCenterRow),
                valueOriginal );
              currentWaveletBand.setValue( static_cast<unsigned int>(convolutionCenterCol), static_cast<unsigned int>(convolutionCenterRow),
                valueOriginal - valueNew );
            }
          }
        }
      }
      
      return true;
    }
    
    bool InverseWaveletAtrous( 
      const te::rst::Raster& waveletRaster,
      const unsigned int levelsNumber,
      te::rst::Raster& outputRaster,
      const std::vector< unsigned int >& outputRasterBands )
    {
      if( 
          ( ! ( waveletRaster.getAccessPolicy() & te::common::RAccess ) )
          ||
          ( waveletRaster.getNumberOfBands() < ( 2 *  levelsNumber * outputRasterBands.size() ) )
        )
      {
        return false;
      }
      if( levelsNumber == 0 )
      {
        return false;
      }
      if( 
          ( ( outputRaster.getAccessPolicy() & te::common::WAccess ) == 0 )
          ||
          ( waveletRaster.getNumberOfColumns() != outputRaster.getNumberOfColumns() )
          ||
          ( waveletRaster.getNumberOfRows() != outputRaster.getNumberOfRows() )
        )
      {
        return false;
      }
      for( unsigned int outputRasterBandsIdx = 0 ; outputRasterBandsIdx <
        outputRasterBands.size() ; ++outputRasterBandsIdx )
      {
        if( outputRasterBands[ outputRasterBandsIdx ] >= outputRaster.getNumberOfBands() )
        {
          return false;
        }
      }
      
      for( unsigned int outputRasterBandsIdx = 0 ; outputRasterBandsIdx <
        outputRasterBands.size() ; ++outputRasterBandsIdx )
      {
        const unsigned int outputRasterBandIdx = outputRasterBands[ outputRasterBandsIdx ];
        const unsigned int nRows = waveletRaster.getNumberOfRows();
        const unsigned int nCols = waveletRaster.getNumberOfColumns();        
        const unsigned int firstSmoothBandIdx = outputRasterBandsIdx *
          levelsNumber * 2;
        const unsigned int firstWaveletBandIdx = firstSmoothBandIdx
          + 1;          
        const unsigned int lastSmoothBandIdx = firstSmoothBandIdx
          + ( 2 * levelsNumber ) - 2;
        const unsigned int lastWaveletBandIdx = lastSmoothBandIdx + 1;
        unsigned int col = 0;
        unsigned int row = 0;        
        double value = 0;
        double sum = 0;
        unsigned int waveletRasterBand = 0;
        
        double bandAllowedMinValue = 0;
        double bandAllowedMaxValue = 0;
        te::rst::GetDataTypeRanges( outputRaster.getBand( 
          outputRasterBandIdx )->getProperty()->m_type, bandAllowedMinValue,
          bandAllowedMaxValue );
        
        
        for( row = 0 ; row < nRows ; ++row )
        {
          for( col = 0 ; col < nCols ; ++col )
          {
            sum = 0.0;
            
            for( waveletRasterBand = firstWaveletBandIdx ; 
              waveletRasterBand <= lastWaveletBandIdx ; 
              waveletRasterBand += 2)
            {
              waveletRaster.getValue( col, row, value, waveletRasterBand );
              sum += value;
            }
            
            waveletRaster.getValue( col, row, value, lastSmoothBandIdx );
            sum += value;
            
            sum = std::max( sum, bandAllowedMinValue );
            sum = std::min( sum, bandAllowedMaxValue );
            
            outputRaster.setValue( col, row, sum, outputRasterBandIdx );
          }
        }
      }
           
      return true;
    }
    
    bool RasterResample( 
      const te::rst::Raster& inputRaster,
      const std::vector< unsigned int >& inputRasterBands,
      const te::rst::Interpolator::Method interpMethod, 
      const unsigned int interpWindowRadius, 
      const unsigned int firstRow,
      const unsigned int firstColumn, 
      const unsigned int height, 
      const unsigned int width,
      const unsigned int newheight, 
      const unsigned int newwidth, 
      const std::map<std::string, std::string>& rinfo,
      const std::string& dataSourceType,
      std::unique_ptr< te::rst::Raster >& resampledRasterPtr )
    {
      return te::rst::rasterResample( 
        inputRaster, 
        inputRasterBands, 
        interpMethod, 
        interpWindowRadius,
        firstRow,
        firstColumn,
        height, 
        width, 
        newheight, 
        newwidth, 
        0, // maxThreads
        rinfo, 
        dataSourceType,
        resampledRasterPtr );
    }
    
    void CreateFixedStepPalette( 
      const unsigned int paletteSize, const bool randomize,
      std::vector< te::rst::BandProperty::ColorEntry >& palette )
    {
      palette.resize( paletteSize );
      if( paletteSize == 0 ) return;

      const short int step = static_cast<short int>( 256.0 /
        std::cbrt( static_cast<double>(paletteSize) ) );
      short int green = 0;
      short int blue = 0;
      unsigned int paletteIdx = 0;

      for( short int red = 0 ; red < 256 ; red += step )
      {
        for( green = 0 ; green < 256 ; green += step )
        {
          for( blue = 0 ; blue < 256 ; blue += step )
          {
            if( paletteIdx >= paletteSize )
            {
              green = 255;
              red = 255;
              break;
            }

            palette[ paletteIdx ].c1 = red;
            palette[ paletteIdx ].c2 = green;
            palette[ paletteIdx ].c3 = blue;
            palette[ paletteIdx ].c4 = 255;

            ++paletteIdx;
          }
        }
      }

      if( randomize )
      {
        std::random_shuffle( palette.begin(), palette.end() );
      }
    }
    
    bool RasterSlicing( 
      const te::rst::Raster& inputRaster,
      const unsigned int inputRasterBand,
      const bool createPaletteRaster,
      const unsigned int slicesNumber,
      const bool eqHistogram,
      const std::map< std::string, std::string >& rasterInfo,
      const std::string& rasterType,
      const bool enableProgress,
      std::vector< te::rst::BandProperty::ColorEntry > const * const palettePtr,
      std::unique_ptr< te::rst::Raster >& outRasterPtr,
      std::vector< double > const* slicesLimitsPtr,
      std::vector< double > const* slicesOutputValuesPtr )
    {
      TERP_TRUE_OR_RETURN_FALSE( inputRasterBand < inputRaster.getNumberOfBands(),
        "Invalid raster band index" );
      TERP_TRUE_OR_RETURN_FALSE( slicesNumber > 0,
        "Invalid slices number" );      
      TERP_TRUE_OR_RETURN_FALSE( ( slicesLimitsPtr == 0 ) ||
         ( slicesLimitsPtr->size() == slicesNumber ),
        "Invalid slices limits" );
      TERP_TRUE_OR_RETURN_FALSE( ( slicesOutputValuesPtr == 0 ) ||
         ( slicesOutputValuesPtr->size() == slicesNumber ),
        "Invalid slices limits" );      
      
      te::rst::Raster const * internalInRasterPtr = &inputRaster;
      unsigned int internalInRasterBand = inputRasterBand;
      
      // slices output values
      
      std::vector< double > slicesOutputValues;
      double maxOutputSliceValue = -1.0 * std::numeric_limits< double >::max();
      bool isPositiveIntergetOutputSliceValues = true;
      
      if( slicesOutputValuesPtr )
      {
        slicesOutputValues = *slicesOutputValuesPtr;
        
        for( unsigned int sliceIdx = 0 ; sliceIdx < slicesOutputValues.size() ; ++sliceIdx )
        {
          maxOutputSliceValue = std::max( maxOutputSliceValue, 
            slicesOutputValues[ sliceIdx ] );
          
          if( slicesOutputValues[ sliceIdx ] != std::abs( std::trunc( 
            slicesOutputValues[ sliceIdx ] ) ) )
          {
            isPositiveIntergetOutputSliceValues = false;
          }
        }
      }
      else
      {
        for( unsigned int sliceIdx = 0 ; sliceIdx < slicesNumber ; ++sliceIdx )
        {
          slicesOutputValues.push_back( (double)sliceIdx );
        }
        
        maxOutputSliceValue = (double)( slicesNumber - 1 );
      }      
      
      // slices limits to slices output values map
      
      std::map< double, double > slicesLimits2OutputValues;
      
      if( slicesLimitsPtr )
      {
        for( unsigned int slicesLimitsIdx = 0 ; slicesLimitsIdx < slicesLimitsPtr->size() ;
          ++slicesLimitsIdx )
        {
          slicesLimits2OutputValues[ slicesLimitsPtr->operator[]( slicesLimitsIdx ) ]
            = slicesOutputValues[ slicesLimitsIdx ];
        }
      }
      else
      {
        std::map<double, unsigned> iHistogram;
        std::map<double, unsigned> rHistogram;
        
        if( ! te::rst::getHistograms( 
          inputRaster, 
          inputRasterBand, 
          0, // rowStart
          0, // colStart
          inputRaster.getNumberOfRows() - 1, // finalrow
          inputRaster.getNumberOfColumns() - 1, //finalCol
          0, // histoBins
          1, // sampleStep
          0, // maxthreads
          rHistogram, 
          iHistogram ) )
        {
          return false;
        }

        double minReal = 0;
        double maxReal = 0;
        double meanReal = 0;
        double stdDevReal = 0;
        
        te::rst::getHistogramStats( rHistogram, minReal, maxReal, meanReal, stdDevReal );
    
        const double sliceSize = 
            ( ( maxReal == minReal ) || ( slicesNumber == 1 ) )
          ? 
            0 
          :
            ( ( maxReal - minReal ) / ( slicesNumber - 1 ) );
          
        for( unsigned int sliceIdx = 0 ; sliceIdx < slicesNumber ; ++sliceIdx )
        {
          slicesLimits2OutputValues[ minReal + (((double)sliceIdx) * sliceSize) ]
            = slicesOutputValues[ sliceIdx ];            
        }
      }
      
      // Perform raster equalizatiion if required
      
      std::unique_ptr< te::rst::Raster > eqRasterPtr;
      
      if( eqHistogram )
      {
        double min = 0;
        double max = 0;
        GetDataTypeRange( inputRaster.getBand( inputRasterBand )->getProperty()->m_type,
          min, max );
        
        Contrast::InputParameters contInputPars;
        contInputPars.m_type = Contrast::InputParameters::HistogramEqualizationContrastT;
        contInputPars.m_inRasterPtr = &inputRaster;
        contInputPars.m_inRasterBands.push_back( inputRasterBand );
        contInputPars.m_outRangeMin = min;
        contInputPars.m_outRangeMax = max;
        contInputPars.m_hECMaxInput.push_back( max );
        contInputPars.m_enableProgress = enableProgress;
        
        Contrast contInstance;
        TERP_TRUE_OR_RETURN_FALSE( contInstance.initialize( contInputPars ),
          "Contrast init error" );        
        
        Contrast::OutputParameters contOutPars;
        contOutPars.m_createdOutRasterDSType = "EXPANSIBLE";
        TERP_TRUE_OR_RETURN_FALSE( contInstance.execute( contOutPars ),
          "Contrast exec error" );  
        
        eqRasterPtr.reset( contOutPars.m_createdOutRasterPtr.release() );
        internalInRasterPtr = eqRasterPtr.get();
        internalInRasterBand = 0;
      }
      
      // Pallete

      std::vector< te::rst::BandProperty::ColorEntry > rasterPalette;
      
      if( isPositiveIntergetOutputSliceValues )
      {
        if( palettePtr )
        {
          rasterPalette = *palettePtr;
        }
        else
        {
          CreateFixedStepPalette( maxOutputSliceValue + 1, true, rasterPalette );
        }
        
        // limiting pallete values so 255 can be used ass nodata value
        
        for( std::size_t rasterPaletteIdx = 0 ; rasterPaletteIdx <
          rasterPalette.size() ; ++rasterPaletteIdx )
        {
          rasterPalette[ rasterPaletteIdx ].c1 = std::min( (short int)254,
            rasterPalette[ rasterPaletteIdx ].c1 );
          rasterPalette[ rasterPaletteIdx ].c2 = std::min( (short int)254,
            rasterPalette[ rasterPaletteIdx ].c2 );
          rasterPalette[ rasterPaletteIdx ].c3 = std::min( (short int)254,
            rasterPalette[ rasterPaletteIdx ].c3 );
          rasterPalette[ rasterPaletteIdx ].c4 = std::min( (short int)254,
            rasterPalette[ rasterPaletteIdx ].c4 );          
        }
      }
      
      // Creating the output raster

      {
        // Band properties
        
        std::vector< te::rst::BandProperty* > bandsProperties;
        
        bandsProperties.push_back( new te::rst::BandProperty( 
          * internalInRasterPtr->getBand( internalInRasterBand )->getProperty() ) );
        
        if( createPaletteRaster && isPositiveIntergetOutputSliceValues )
        {
          bandsProperties[ 0 ]->m_noDataValue = maxOutputSliceValue + 1;
          
          bandsProperties[ 0 ]->m_colorInterp = te::rst::PaletteIdxCInt;
          bandsProperties[ 0 ]->m_paletteInterp = te::rst::RGBPalInt;
          bandsProperties[ 0 ]->m_palette = rasterPalette;          
          
          if( maxOutputSliceValue <= (double)std::numeric_limits< unsigned char >::max() )
          {
            bandsProperties[ 0 ]->m_type = te::dt::UCHAR_TYPE;
          }
          else if( maxOutputSliceValue <= (double)std::numeric_limits< unsigned short int >::max() )
          {
            bandsProperties[ 0 ]->m_type = te::dt::UINT16_TYPE;
          }
          else if( maxOutputSliceValue <= (double)std::numeric_limits< unsigned int >::max() )
          {
            bandsProperties[ 0 ]->m_type = te::dt::UINT32_TYPE;
          }
          else
          {
            TERP_LOG_AND_RETURN_FALSE( "Invalid palette size" );
          }
        }
        else
        {
          if( isPositiveIntergetOutputSliceValues )
          {
            bandsProperties[ 0 ]->m_noDataValue = 255;
            
            if( maxOutputSliceValue <= (double)std::numeric_limits< unsigned char >::max() )
            {
              bandsProperties[ 0 ]->m_type = te::dt::UCHAR_TYPE;
            }
            else if( maxOutputSliceValue <= (double)std::numeric_limits< unsigned short int >::max() )
            {
              bandsProperties[ 0 ]->m_type = te::dt::UINT16_TYPE;
            }
            else if( maxOutputSliceValue <= (double)std::numeric_limits< unsigned int >::max() )
            {
              bandsProperties[ 0 ]->m_type = te::dt::UINT32_TYPE;
            }
            else
            {
              TERP_LOG_AND_RETURN_FALSE( "Invalid palette size" );
            }
            
            bandsProperties.push_back( new te::rst::BandProperty( 
              *bandsProperties[ 0 ] ) );
            bandsProperties.push_back( new te::rst::BandProperty( 
              *bandsProperties[ 0 ] ) );
            
            bandsProperties[ 0 ]->m_colorInterp = te::rst::RedCInt;
            bandsProperties[ 1 ]->m_colorInterp = te::rst::GreenCInt;
            bandsProperties[ 2 ]->m_colorInterp = te::rst::BlueCInt;                
          }
          else
          {
            bandsProperties[ 0 ]->m_noDataValue = maxOutputSliceValue + 1;
            bandsProperties[ 0 ]->m_type = te::dt::DOUBLE_TYPE;
          }
        }
        
        TERP_TRUE_OR_RETURN_FALSE( CreateNewRaster( *internalInRasterPtr->getGrid(),
          bandsProperties, rasterInfo, rasterType, outRasterPtr ),
          "Output raster creation error" );
      }
        
        // fill output raster values
        
      {
        std::unique_ptr< te::common::TaskProgress > progressPtr;
        if( enableProgress )
        {
          progressPtr.reset( new te::common::TaskProgress );
          progressPtr->setMessage( "Slicing" );
          progressPtr->setTotalSteps( static_cast<int>(2 * internalInRasterPtr->getNumberOfRows()) );
        }               
          
        const bool isOneBandOutRaster = ( outRasterPtr->getNumberOfBands() == 1 );
        const te::rst::Band& inBand = (*internalInRasterPtr->getBand( internalInRasterBand ));
        te::rst::Band& outBandRed = (*outRasterPtr->getBand( 0 ));
        te::rst::Band& outBandGreen = ( *outRasterPtr->getBand( 
                                            isOneBandOutRaster ? 0 : 1 ) );
        te::rst::Band& outBandBlue = (*outRasterPtr->getBand( 
                                           isOneBandOutRaster ? 0 : 2 ) );
        const unsigned int nRows = internalInRasterPtr->getNumberOfRows();
        const unsigned int nCols = internalInRasterPtr->getNumberOfColumns();
        const double inputNoDataValue = inBand.getProperty()->m_noDataValue;
        const double outputNoDataValue = outBandRed.getProperty()->m_noDataValue;
        unsigned int col = 0;
        double inputValue = 0;
        std::map< double, double >::const_iterator slicesLimits2OutputValuesIt;
        const std::map< double, double >::const_iterator 
          slicesLimits2OutputValuesItEnd = slicesLimits2OutputValues.end();
        te::rst::BandProperty::ColorEntry const* colorEntryPtr = 0;
        double outValue = 0;
        const double lastOutValue = slicesLimits2OutputValues.rbegin()->second;
        
        for( unsigned int row = 0 ; row < nRows ; ++row )
        {
          for( col = 0 ; col < nCols ; ++col )
          {
            inBand.getValue( col, row, inputValue );
            
            if( inputValue == inputNoDataValue )
            {
              if( isOneBandOutRaster )
              {
                outBandRed.setValue( col, row, outputNoDataValue );
              }
              else
              {
                outBandRed.setValue( col, row, outputNoDataValue );
                outBandGreen.setValue( col, row, outputNoDataValue );
                outBandBlue.setValue( col, row, outputNoDataValue );
              }
            }
            else
            {
              slicesLimits2OutputValuesIt = slicesLimits2OutputValues.upper_bound( inputValue );
              if( slicesLimits2OutputValuesIt == slicesLimits2OutputValuesItEnd )
              {
                outValue = lastOutValue;
              }
              else
              {
                --slicesLimits2OutputValuesIt;
                outValue = slicesLimits2OutputValuesIt->second;
              }
              
              if( isOneBandOutRaster )
              {
                outBandRed.setValue( col, row, outValue );
              }
              else
              {
                assert( outValue < (double)rasterPalette.size() );
                colorEntryPtr =  &rasterPalette[ (std::size_t)outValue ];
                
                outBandRed.setValue( col, row, colorEntryPtr->c1 );
                outBandGreen.setValue( col, row, colorEntryPtr->c2 );
                outBandBlue.setValue( col, row, colorEntryPtr->c3 );
              }
            }
          }
          
          if( enableProgress )
          {
            progressPtr->pulse();
            if( ! progressPtr->isActive() ) return false;
          }   
        }
      }
      
      return true;
    }

    bool ReplaceContiguousSegmentValues(
      te::rst::Raster& inputRaster,
      const unsigned int& inputRasterBandIdx,
      const unsigned targetRow,
      const unsigned targetCol,
      const double& outputValue )
    {
      // checking parameters
      if( ( inputRaster.getAccessPolicy() & te::common::WAccess ) == 0 )
      {
        return false;
      }
      if( inputRasterBandIdx >= inputRaster.getNumberOfBands() )
      {
        return false;
      }
      if( targetRow >= inputRaster.getNumberOfRows() ) return false;
      if( targetCol >= inputRaster.getNumberOfColumns() ) return false;

      // main loop

      std::pair< unsigned int, unsigned int > currentPixelCoods;
      std::size_t targetValuesIdx = 0;
      double pixelValue = 0;
      te::rst::Band& inBand = *inputRaster.getBand( inputRasterBandIdx );
      const double inNoDataValue = inBand.getProperty()->m_noDataValue;
      const unsigned int lastRowIdx = inputRaster.getNumberOfRows() - 1;
      const unsigned int lastColIdx = inputRaster.getNumberOfColumns() - 1;

      double segmentOriginalValue = 0;
      inBand.getValue( targetCol, targetRow, segmentOriginalValue );
      if( segmentOriginalValue == inNoDataValue ) return true;
      if( segmentOriginalValue == outputValue ) return true;

      // std::pair< col, row >
      std::stack< std::pair< unsigned int, unsigned int > > pixels2Process;
      pixels2Process.push( std::pair< unsigned int, unsigned int >( targetCol,
        targetRow ) );

      while( !pixels2Process.empty() )
      {
        currentPixelCoods = pixels2Process.top();
        pixels2Process.pop();

        inBand.setValue( currentPixelCoods.first, currentPixelCoods.second,
          outputValue );

        // add neighbors

        // do i have a left neighbor ?
        if( currentPixelCoods.first )
        {
          inBand.getValue( currentPixelCoods.first - 1, currentPixelCoods.second,
            pixelValue );

          if( pixelValue == segmentOriginalValue )
          {
            pixels2Process.push( std::pair< unsigned int, unsigned int >(
              currentPixelCoods.first - 1, currentPixelCoods.second ) );
          }
        }

        // do i have a right neighbor ?
        if( currentPixelCoods.first < lastColIdx )
        {
          inBand.getValue( currentPixelCoods.first + 1, currentPixelCoods.second,
            pixelValue );

          if( pixelValue == segmentOriginalValue )
          {
            pixels2Process.push( std::pair< unsigned int, unsigned int >(
              currentPixelCoods.first + 1, currentPixelCoods.second ) );
          }
        }

        // do i have a top neighbor ?
        if( currentPixelCoods.second )
        {
          inBand.getValue( currentPixelCoods.first, currentPixelCoods.second - 1,
            pixelValue );

          if( pixelValue == segmentOriginalValue )
          {
            pixels2Process.push( std::pair< unsigned int, unsigned int >(
              currentPixelCoods.first, currentPixelCoods.second - 1 ) );
          }
        }

        // do i have a bottom neighbor ?
        if( currentPixelCoods.second < lastRowIdx )
        {
          inBand.getValue( currentPixelCoods.first, currentPixelCoods.second + 1,
            pixelValue );

          if( pixelValue == segmentOriginalValue )
          {
            pixels2Process.push( std::pair< unsigned int, unsigned int >(
              currentPixelCoods.first, currentPixelCoods.second + 1 ) );
          }
        }
      }

      return true;
    }

    bool ReplaceContiguousSegmentValues(
      const te::rst::Raster& inputRaster,
      const unsigned int& inputRasterBandIdx,
      te::rst::Raster& outputRaster,
      const unsigned int& outputRasterBandIdx,
      const unsigned targetInputRow,
      const unsigned targetInputCol,
      const double& outputValue )
    {
      // checking parameters
      if( ( outputRaster.getAccessPolicy() & te::common::WAccess ) == 0 )
      {
        return false;
      }
      if( inputRasterBandIdx >= inputRaster.getNumberOfBands() )
      {
        return false;
      }
      if( targetInputRow >= inputRaster.getNumberOfRows() ) return false;
      if( targetInputCol >= inputRaster.getNumberOfColumns() ) return false;
      if( outputRasterBandIdx >= outputRaster.getNumberOfBands() )
      {
        return false;
      }

      // main loop

      std::pair< unsigned int, unsigned int > currentPixelInputCoods;
      std::pair< unsigned int, unsigned int > neighborPixelInputCoods;
      std::size_t targetValuesIdx = 0;
      double pixelValue = 0;
      const te::rst::Band& inBand = *inputRaster.getBand( inputRasterBandIdx );
      te::rst::Band& outBand = *outputRaster.getBand( inputRasterBandIdx );
      const te::rst::Grid& inGrid = *inputRaster.getGrid();
      const te::rst::Grid& outGrid = *outputRaster.getGrid();
      const double inNoDataValue = inBand.getProperty()->m_noDataValue;
      const unsigned int lastInputRowIdx = inputRaster.getNumberOfRows() - 1;
      const unsigned int lastInputColIdx = inputRaster.getNumberOfColumns() - 1;
      const unsigned int numberOfOutputRows = outputRaster.getNumberOfRows();
      const unsigned int numberOfOutputCols = outputRaster.getNumberOfColumns();
      double inputXCoords = 0;
      double inputYCoords = 0;
      double outputXCoords = 0;
      double outputYCoords = 0;
      double outputColDouble = 0;
      double outputRowDouble = 0;
      unsigned int outputCol = 0;
      unsigned int outputRow = 0;
      te::srs::Converter conv( inputRaster.getSRID(), outputRaster.getSRID() );

      double inputSegmentOriginalValue = 0;
      inBand.getValue( targetInputCol, targetInputRow, inputSegmentOriginalValue );
      if( inputSegmentOriginalValue == inNoDataValue ) return true;

      // std::pair< col, row >
      std::stack< std::pair< unsigned int, unsigned int > > pixels2ProcessInputCoords;
      pixels2ProcessInputCoords.push( std::pair< unsigned int, unsigned int >( targetInputCol,
        targetInputRow ) );

      std::set< std::pair< unsigned int, unsigned int > > processedPixels;

      while( !pixels2ProcessInputCoords.empty() )
      {
        currentPixelInputCoods = pixels2ProcessInputCoords.top();
        pixels2ProcessInputCoords.pop();

        inGrid.gridToGeo( (double)currentPixelInputCoods.first,
          (double)currentPixelInputCoods.second, inputXCoords, inputYCoords );

        conv.convert( inputXCoords, inputYCoords, outputXCoords, outputYCoords );

        outGrid.geoToGrid( outputXCoords, outputYCoords, outputColDouble,
          outputRowDouble );

        if( ( outputColDouble >= 0) && ( outputRowDouble >= 0 ) )
        {
          outputCol = te::common::Round< double, unsigned int >( outputColDouble );
          outputRow = te::common::Round< double, unsigned int >( outputRowDouble );

          if( (outputCol < numberOfOutputCols) && ( outputRow < numberOfOutputRows) )
          {
            outBand.setValue( outputCol, outputRow, outputValue );
          }
        }

        // add neighbors for post processing

        // do i have a left neighbor ?
        if( currentPixelInputCoods.first )
        {
          neighborPixelInputCoods.first = currentPixelInputCoods.first - 1;
          neighborPixelInputCoods.second = currentPixelInputCoods.second;

          inBand.getValue( neighborPixelInputCoods.first, neighborPixelInputCoods.second,
            pixelValue );

          if(
              ( pixelValue == inputSegmentOriginalValue )
              &&
              ( processedPixels.find( neighborPixelInputCoods) == processedPixels.end() )
            )
          {
            pixels2ProcessInputCoords.push( neighborPixelInputCoods );
          }
        }

        // do i have a right neighbor ?
        if( currentPixelInputCoods.first < lastInputColIdx )
        {
          neighborPixelInputCoods.first = currentPixelInputCoods.first + 1;
          neighborPixelInputCoods.second = currentPixelInputCoods.second;

          inBand.getValue( neighborPixelInputCoods.first, neighborPixelInputCoods.second,
            pixelValue );

          if(
              ( pixelValue == inputSegmentOriginalValue )
              &&
              ( processedPixels.find( neighborPixelInputCoods) == processedPixels.end() )
            )
          {
            pixels2ProcessInputCoords.push( neighborPixelInputCoods );
          }
        }

        // do i have a top neighbor ?
        if( currentPixelInputCoods.second )
        {
          neighborPixelInputCoods.first = currentPixelInputCoods.first;
          neighborPixelInputCoods.second = currentPixelInputCoods.second - 1;

          inBand.getValue( neighborPixelInputCoods.first, neighborPixelInputCoods.second,
            pixelValue );

          if(
              ( pixelValue == inputSegmentOriginalValue )
              &&
              ( processedPixels.find( neighborPixelInputCoods) == processedPixels.end() )
            )
          {
            pixels2ProcessInputCoords.push( neighborPixelInputCoods );
          }
        }

        // do i have a bottom neighbor ?
        if( currentPixelInputCoods.second < lastInputRowIdx )
        {
          neighborPixelInputCoods.first = currentPixelInputCoods.first;
          neighborPixelInputCoods.second = currentPixelInputCoods.second + 1;

          inBand.getValue( neighborPixelInputCoods.first, neighborPixelInputCoods.second,
            pixelValue );

          if(
              ( pixelValue == inputSegmentOriginalValue )
              &&
              ( processedPixels.find( neighborPixelInputCoods) == processedPixels.end() )
            )
          {
            pixels2ProcessInputCoords.push( neighborPixelInputCoods );
          }
        }

        processedPixels.insert( currentPixelInputCoods );
      }

      return true;
    }
    
    void getHistogramsThreadEntry( te::rp::GetHistogramsThreadParameters* paramsPtr )
    {
      std::unique_lock< std::mutex > locker( *paramsPtr->m_mutexPtr );

      const unsigned int inputBandIndex = paramsPtr->m_inputBandIndex;
      const std::vector< te::gm::Polygon const* >& rois = *paramsPtr->m_roisPtr;
      std::vector< bool >& roisProcStatus = *paramsPtr->m_roisProcStatusPtr;
      const unsigned int threadExecMode = paramsPtr->m_threadExecMode;
      const bool preReadRasterMinMax = paramsPtr->m_preReadRasterMinMax;
      std::map<double, unsigned>& rHistogram = *paramsPtr->m_rHistogramPtr;
      std::map<double, unsigned>& iHistogram = *paramsPtr->m_iHistogramPtr;
      const std::size_t roisProcStatusSize = roisProcStatus.size();
      const unsigned int maxRasterBlocksPerThread = paramsPtr->m_maxRasterBlocksPerThread;
      const te::rst::PolygonIterator< double >::IterationType roisItType =
        paramsPtr->m_roisItType;

      locker.unlock();

      const te::rst::SynchronizedRaster inputRaster( maxRasterBlocksPerThread,
        *paramsPtr->m_rasterSyncPtr );
      const double noDataValue = inputRaster.getBand( inputBandIndex 
        )->getProperty()->m_noDataValue;

      std::complex< double > complexValue = 0;
      double realValue = 0;
      double imagValue = 0;
      te::gm::Polygon const* currentROIPtr = 0;
      te::rst::PolygonIterator< double > polIt;
      te::rst::PolygonIterator< double > polItEnd;

      if( threadExecMode == 0 )
      {
        // dicovering band values limits

        double roiMinRValue = 0;
        double roiMinIValue = 0;
        double roiMaxRValue = 0;
        double roiMaxIValue = 0;

        for( std::size_t roisProcStatusIdx = 0 ; roisProcStatusIdx <
          roisProcStatusSize ; ++roisProcStatusIdx )
        {
          locker.lock();

          // looking for the next unprocessed roi

          if( !roisProcStatus[ roisProcStatusIdx ] )
          {
            roisProcStatus[ roisProcStatusIdx ] = true;
            currentROIPtr = rois[ roisProcStatusIdx ];

            locker.unlock();
            
            roiMinRValue = std::numeric_limits< double >::max();
            roiMinIValue = roiMinRValue;
            roiMaxRValue = -1.0 * roiMinRValue;
            roiMaxIValue = roiMaxRValue;            

            assert( inputRaster.getSRID() == currentROIPtr->getSRID() );
            polIt = 
              te::rst::PolygonIterator< double >::begin( &inputRaster, 
              currentROIPtr, roisItType );
            polItEnd = 
              te::rst::PolygonIterator< double >::end( &inputRaster, 
              currentROIPtr, roisItType );
            
            while( polIt != polItEnd )
            {
              complexValue = polIt( inputBandIndex );
              realValue = complexValue.real();
              imagValue = complexValue.imag();
              
              if( realValue != noDataValue )
              {
                roiMinRValue = std::min( roiMinRValue, realValue );
                roiMaxRValue = std::max( roiMaxRValue, realValue );

                roiMinIValue = std::min( roiMinIValue, imagValue );
                roiMaxIValue = std::max( roiMaxIValue, imagValue );
              }
              
              ++polIt;
            }
            
            locker.lock();

            // update global min and max
            
            paramsPtr->m_inBandMinValueR = std::min( paramsPtr->m_inBandMinValueR, 
              roiMinRValue );
            paramsPtr->m_inBandMaxValueR = std::max( paramsPtr->m_inBandMaxValueR, 
              roiMaxRValue );

            paramsPtr->m_inBandMinValueI = std::min( paramsPtr->m_inBandMinValueI, 
              roiMinIValue );
            paramsPtr->m_inBandMaxValueI = std::max( paramsPtr->m_inBandMaxValueI, 
              roiMaxIValue );            

            locker.unlock();
          }
          else
          {
            locker.unlock();
          }
        }
      }
      else // threadExecMode == 1
      {
        locker.lock();

        std::map<double, unsigned> roiRHistogram =
          *paramsPtr->m_rHistogramPtr;
        std::map<double, unsigned> roiIHistogram =
          *paramsPtr->m_iHistogramPtr;
        std::map<double, unsigned>::iterator hIt;
        std::map<double, unsigned>::iterator hItEnd;

        locker.unlock();

        // generating histograms
        
        for( std::size_t roisProcStatusIdx = 0 ; roisProcStatusIdx <
          roisProcStatusSize ; ++roisProcStatusIdx )
        {
          locker.lock();

          // looking for the next unprocessed roi

          if( !roisProcStatus[ roisProcStatusIdx ] )
          {
            roisProcStatus[ roisProcStatusIdx ] = true;
            currentROIPtr = rois[ roisProcStatusIdx ];

            locker.unlock();        

            // zero roi histograms frequencies

            if( preReadRasterMinMax )
            {
              hIt = roiRHistogram.begin();
              hItEnd = roiRHistogram.end();
              while( hIt != hItEnd )
              {
                hIt->second = 0;
                ++hIt;
              }

              hIt = roiIHistogram.begin();
              hItEnd = roiIHistogram.end();
              while( hIt != hItEnd )
              {
                hIt->second = 0;
                ++hIt;
              }
            }
            else
            {
              roiIHistogram.clear();
              roiRHistogram.clear();
            }

            
            if( preReadRasterMinMax )
            {
              assert( inputRaster.getSRID() == currentROIPtr->getSRID() );
              polIt = 
                te::rst::PolygonIterator< double >::begin( &inputRaster, 
                currentROIPtr, roisItType );
              polItEnd = 
                te::rst::PolygonIterator< double >::end( &inputRaster, 
                currentROIPtr, roisItType );
              
              while( polIt != polItEnd )
              {
                complexValue = polIt( inputBandIndex );
                realValue = complexValue.real();
                imagValue = complexValue.imag();
                
                if( realValue != noDataValue )
                {
                  ++( roiRHistogram.lower_bound( realValue )->second );
                  ++( roiIHistogram.lower_bound( imagValue )->second );
                }
                
                ++polIt;
              }
            }
            else
            {
              assert( inputRaster.getSRID() == currentROIPtr->getSRID() );
              polIt = 
                te::rst::PolygonIterator< double >::begin( &inputRaster, 
                currentROIPtr, roisItType );
              polItEnd = 
                te::rst::PolygonIterator< double >::end( &inputRaster, 
                currentROIPtr, roisItType );
              
              while( polIt != polItEnd )
              {
                complexValue = polIt( inputBandIndex );
                realValue = complexValue.real();
                imagValue = complexValue.imag();
                
                if( realValue != noDataValue )
                {
                  ++roiRHistogram[ realValue ];
                  ++roiIHistogram[ imagValue ];
                }
                
                ++polIt;
              }
            }

            // transfering values to the main histograms

            locker.lock();

            hIt = roiRHistogram.begin();
            hItEnd = roiRHistogram.end();
            while( hIt != hItEnd )
            {
              rHistogram[ hIt->first ] += hIt->second;
              ++hIt;
            }

            hIt = roiIHistogram.begin();
            hItEnd = roiIHistogram.end();
            while( hIt != hItEnd )
            {
              iHistogram[ hIt->first ] += hIt->second;
              ++hIt;
            }

            locker.unlock();
          }
          else
          {
            locker.unlock();
          }
        }
      }
    }
    
    bool getHistograms(
      const te::rst::Raster& raster,
      const unsigned int bandIndex,
      const std::vector< te::gm::Polygon const* >& rois,
      const unsigned int histoBins,
      const unsigned int maxThreads,
      const te::rst::PolygonIterator< double >::IterationType roisItType,
      std::map<double, unsigned>& rHistogram,
      std::map<double, unsigned>& iHistogram )
    {
      // The case where all image must be processed
      if( rois.empty() )
      {
        return te::rst::getHistograms( 
          raster, 
          bandIndex, 
          0, // rowStart
          0, // colStart
          raster.getNumberOfRows() - 1, // finalRow
          raster.getNumberOfColumns() - 1, // finalCol
          histoBins,
          1, // sampleStep
          maxThreads,
          rHistogram,
          iHistogram );
      }
      
      rHistogram.clear();
      iHistogram.clear();

      if( bandIndex >= raster.getNumberOfBands() )
      {
        return false;
      }

      const te::rst::Band& band = *raster.getBand( bandIndex );

      bool isFloatData = false;  
      if(
          ( band.getProperty()->m_type == te::dt::FLOAT_TYPE )
          ||
          ( band.getProperty()->m_type == te::dt::DOUBLE_TYPE )
          ||
          ( band.getProperty()->m_type == te::dt::CFLOAT_TYPE )
          ||
          ( band.getProperty()->m_type == te::dt::CDOUBLE_TYPE )
        )
      {
        isFloatData = true;
      }

      unsigned int threadsNumber = maxThreads ? maxThreads :
        te::common::GetPhysProcNumber();      
      const double totalPhysMem = (double)te::common::GetTotalPhysicalMemory();
      const double usedVMem = (double)te::common::GetUsedVirtualMemory();
      const double totalVMem = ( (double)te::common::GetTotalVirtualMemory() );
      const double freeVMem = std::min( totalPhysMem, ( totalVMem - usedVMem ) );
      const double rasterBlockRequiredRam = band.getBlockSize();
      const double maxSimultaneousRasterBlocks = 0.75 * freeVMem / rasterBlockRequiredRam;
      const unsigned int maxRasterBlocksPerThread = std::max( 1u,
        ( (unsigned int)( maxSimultaneousRasterBlocks ) ) / threadsNumber );
      
      std::vector< bool > roisProcStatus( rois.size(), false );

      te::rst::RasterSynchronizer rasterSync(
        *(te::rst::Raster*)(&raster), te::common::RAccess );
      std::mutex mutex;

      te::rp::GetHistogramsThreadParameters params;
      params.m_rasterSyncPtr = &rasterSync;
      params.m_inputBandIndex = bandIndex;
      params.m_roisPtr = &rois;
      params.m_roisProcStatusPtr = &roisProcStatus;
      params.m_returnStatus = true;
      params.m_mutexPtr = &mutex;
      params.m_rHistogramPtr = &rHistogram;
      params.m_iHistogramPtr = &iHistogram;
      params.m_preReadRasterMinMax = isFloatData;
      params.m_inBandMinValueR = std::numeric_limits< double >::max();
      params.m_inBandMaxValueR = -1.0 * params.m_inBandMinValueR;
      params.m_inBandMinValueI = std::numeric_limits< double >::max();
      params.m_inBandMaxValueI = -1.0 * params.m_inBandMinValueI;
      params.m_maxRasterBlocksPerThread = maxRasterBlocksPerThread;
      params.m_roisItType = roisItType;

      // step 1 - min and max calcule

      if( isFloatData )
      {
        params.m_threadExecMode = 0;

        if( maxThreads == 1 )
        {
          getHistogramsThreadEntry( &params );
        }
        else
        {
          std::vector< std::shared_ptr< std::thread > > threads;

          for( unsigned int threadIdx = 0 ; threadIdx < threadsNumber ;
            ++threadIdx )
          {
            std::shared_ptr< std::thread > threadPtr( new std::thread(
              &getHistogramsThreadEntry, &params ) );

            threads.push_back( threadPtr );
          };

          for( unsigned int threadIdx = 0 ; threadIdx < threadsNumber ;
            ++threadIdx )
          {
            threads[ threadIdx ]->join();
          };
        }

        if( params.m_returnStatus == false ) return false;
      }

      // step 2 - set initial float histograms

      if( isFloatData )
      {
        const unsigned int finalHistoBins = histoBins ? histoBins : 256;
        
        const double binRValue =
          ( finalHistoBins == 1 ) ? 0 :
          ( 
            ( params.m_inBandMaxValueR - params.m_inBandMinValueR ) 
            / (double)( finalHistoBins - 1 ) 
          );
        const double binIValue =
          ( finalHistoBins == 1 ) ? 0 :
          ( 
            ( params.m_inBandMaxValueI - params.m_inBandMinValueI ) 
            / 
            (double)( finalHistoBins - 1 ) 
          );

        for( unsigned int binIdx = 0 ; binIdx < finalHistoBins ; ++binIdx )
        {
          rHistogram[ ( ((double)binIdx) * binRValue ) + params.m_inBandMinValueR ] = 0;
          iHistogram[ ( ((double)binIdx) * binIValue ) + params.m_inBandMinValueI ] = 0;
        }
      }

      // step 3 - histograms frequencies calcule
      
      {
        // mark all blocks as not processed
        
        std::fill( roisProcStatus.begin(), roisProcStatus.end(), false );
        
        params.m_threadExecMode = 1;

        if( maxThreads == 1 )
        {
          getHistogramsThreadEntry( &params );
        }
        else
        {
          std::vector< std::shared_ptr< std::thread > > threads;

          for( unsigned int threadIdx = 0 ; threadIdx < threadsNumber ;
            ++threadIdx )
          {
            std::shared_ptr< std::thread > threadPtr( new std::thread(
              &getHistogramsThreadEntry, &params ) );

            threads.push_back( threadPtr );
          };

          for( unsigned int threadIdx = 0 ; threadIdx < threadsNumber ;
            ++threadIdx )
          {
            threads[ threadIdx ]->join();
          };
        }
      }
      
      // step 4 - histogram user bins interpolation
      
      if( 
          ( ! isFloatData ) 
          && 
          ( histoBins != 0 ) 
        )
      {
        std::map<double, unsigned> auxHistogram;
        
        if( ! te::rst::interpolateHistogram( rHistogram, auxHistogram, histoBins, 0, 0 ) )
        {
          return false;
        }
        rHistogram = auxHistogram;
        
        if( ! te::rst::interpolateHistogram( iHistogram, auxHistogram, histoBins, 0, 0 ) )
        {
          return false;
        }
        iHistogram = auxHistogram;    
      }

      // step 5 - fill integer histogram gaps

      if( 
          ( ! isFloatData )
          &&
          ( histoBins == 0 )
        )
      {
        std::map<double, unsigned> auxHistogram;      
        
        te::rst::fillHistogramGaps( rHistogram, auxHistogram );
        rHistogram = auxHistogram;

        te::rst::fillHistogramGaps( iHistogram, auxHistogram );
        iHistogram = auxHistogram;
      }

      return params.m_returnStatus;
    }
    
    void getHistogramStats(
      const std::map<double, unsigned int>& histogram,
      double& min,
      double& max,
      double& mean,
      double& stdDev,
      double& mode,
      double& entropy,
      double& sum,
      double& sum2,
      double& sum3,
      double& sum4,
      double& variance,
      double& median
      )
    {
      min = max = mean = stdDev = mode = entropy = sum = sum2 = sum3 =
        sum4 = variance = median = 0;

      if( ! histogram.empty() )
      {
        std::map<double, unsigned int>::const_iterator it = 
          histogram.begin();
        std::map<double, unsigned int>::const_reverse_iterator rit = 
          histogram.rbegin();
        const std::map<double, unsigned int>::const_iterator itEnd = 
          histogram.end();
        const std::size_t medianPosition = histogram.size() / 2;
        unsigned int elementsNumber = 0;
        unsigned int highestFrequencyFound = 0;
        double probability = 0;
        std::size_t histogramIdx = 0;

        min = it->first;
        max = rit->first;
        
        while( it != itEnd )
        {
          mean += ( it->first * ((double)it->second) );
          
          elementsNumber += it->second;
          
          if( it->second > highestFrequencyFound )
          {
            highestFrequencyFound = it->second;
            mode = it->first;
          }
          
          sum += it->first * (double)it->second;
          
          sum2 += ( it->first * it->first 
            * (double)it->second );
          
          sum3 += ( it->first * it->first * it->first 
            * (double)it->second );
          
          sum4 += ( it->first * it->first * it->first * it->first 
            * (double)it->second );
          
          if( histogramIdx == medianPosition )
          {
            median = it->first;
          }
          
          ++histogramIdx;
          ++it;
        }
        
        if( elementsNumber )
        {
          mean /= (double)elementsNumber;
        }

        double diff = 0;
        
        it = histogram.begin();
        while( it != itEnd )
        {
          diff = ( it->first - mean );
          
          variance += ( ( diff * diff ) * ((double)it->second) );
          
          if( elementsNumber )
          {
            probability = ( (double)(it->second) ) / (double)elementsNumber;
            
            if( probability > 0.0 ) 
            {
              entropy += ( probability * ( std::log( probability ) /
                log( 2.0 ) ) );
            }
          }
          
          ++it;
        }
        
        entropy = (-1.0) * entropy;
        
        if( elementsNumber )
        {
          variance /= (double)elementsNumber;
        }
        
        stdDev = std::sqrt( variance );
      }
    }
    
    void getJointHistogramsThreadEntry( te::rp::GetJointHistogramsThreadParameters* paramsPtr )
    {
      std::unique_lock< std::mutex > locker( *paramsPtr->m_mutexPtr );

      const unsigned int inputBandIndex1 = paramsPtr->m_inputBandIndex1;
      const unsigned int inputBandIndex2 = paramsPtr->m_inputBandIndex2;
      
      const std::vector< te::gm::Polygon const* >& rois = *paramsPtr->m_roisPtr;
      std::vector< bool >& roisProcStatus = *paramsPtr->m_roisProcStatusPtr;
      
      const std::map<double, unsigned int> rRHist1 = *paramsPtr->m_rRHist1Ptr;
      const std::map<double, unsigned int> rIHist1 = *paramsPtr->m_rIHist1Ptr;
      const std::map<double, unsigned int> rRHist2 = *paramsPtr->m_rRHist2Ptr;
      const std::map<double, unsigned int> rIHist2 = *paramsPtr->m_rIHist2Ptr;         
      
      const std::size_t roisProcStatusSize = roisProcStatus.size();
      const unsigned int maxRasterBlocksPerThread = 
        paramsPtr->m_maxRasterBlocksPerThread;
      
      std::map< std::pair< double, double >, unsigned int> roiRJointHistogram =
        *paramsPtr->m_realJointHistogramPtr;
      std::map< std::pair< double, double >, unsigned int> roiIJointHistogram =
        *paramsPtr->m_imagJointHistogramPtr;
        
      const te::rst::PolygonIterator< double >::IterationType roisItType =
        paramsPtr->m_roisItType;        

      locker.unlock();

      const te::rst::SynchronizedRaster inputRaster( maxRasterBlocksPerThread,
        *paramsPtr->m_rasterSyncPtr );
      
      const double noDataValue1 = inputRaster.getBand( inputBandIndex1 
        )->getProperty()->m_noDataValue;
      const double noDataValue2 = inputRaster.getBand( inputBandIndex2 
        )->getProperty()->m_noDataValue;
        
      std::complex< double > complexValue1 = 0;
      double realValue1 = 0;
      double imagValue1 = 0;
      std::complex< double > complexValue2 = 0;
      double realValue2 = 0;
      double imagValue2 = 0;
      
      te::gm::Polygon const* currentROIPtr = 0;
      te::rst::PolygonIterator< double > polIt;
      te::rst::PolygonIterator< double > polItEnd;
      
      std::map<double, unsigned int>::const_iterator rRHist1It;
      std::map<double, unsigned int>::const_iterator rIHist1It;
      std::map<double, unsigned int>::const_iterator rRHist2It;
      std::map<double, unsigned int>::const_iterator rIHist2It;
        
      std::map< std::pair< double, double >, unsigned int>::iterator 
        roiRJointHistogramIt;        
      const std::map< std::pair< double, double >, unsigned int>::iterator 
        roiRJointHistogramItEnd = roiRJointHistogram.end();
      std::map< std::pair< double, double >, unsigned int>::iterator 
        roiIJointHistogramIt;
      const std::map< std::pair< double, double >, unsigned int>::iterator 
        roiIJointHistogramItEnd = roiIJointHistogram.end();
      std::pair< double, double > auxPair;
      
      for( std::size_t roisProcStatusIdx = 0 ; roisProcStatusIdx <
        roisProcStatusSize ; ++roisProcStatusIdx )
      {
        locker.lock();

        // looking for the next unprocessed roi

        if( !roisProcStatus[ roisProcStatusIdx ] )
        {
          roisProcStatus[ roisProcStatusIdx ] = true;
          currentROIPtr = rois[ roisProcStatusIdx ];

          locker.unlock();        

          // zero roi histograms frequencies

          roiRJointHistogramIt = roiRJointHistogram.begin();
          while( roiRJointHistogramIt != roiRJointHistogramItEnd )
          {
            roiRJointHistogramIt->second = 0;
            ++roiRJointHistogramIt;
          }

          roiIJointHistogramIt = roiIJointHistogram.begin();
          while( roiIJointHistogramIt != roiIJointHistogramItEnd )
          {
            roiIJointHistogramIt->second = 0;
            ++roiIJointHistogramIt;
          }
          
          assert( inputRaster.getSRID() == currentROIPtr->getSRID() );
          polIt = 
            te::rst::PolygonIterator< double >::begin( &inputRaster, 
            currentROIPtr, roisItType );
          polItEnd = 
            te::rst::PolygonIterator< double >::end( &inputRaster, 
            currentROIPtr, roisItType );
          
          while( polIt != polItEnd )
          {
            complexValue1 = polIt( inputBandIndex1 );
            realValue1 = complexValue1.real();
            imagValue1 = complexValue1.imag();
            
            complexValue2 = polIt( inputBandIndex2 );
            realValue2 = complexValue2.real();
            imagValue2 = complexValue2.imag();
            
            if( 
                ( realValue1 != noDataValue1 ) 
                && 
                ( realValue2 != noDataValue2 )
              )
            {
              rRHist1It = rRHist1.lower_bound( realValue1 );
              rIHist1It = rIHist1.lower_bound( imagValue1 );
              
              rRHist2It = rRHist2.lower_bound( realValue2 );
              rIHist2It = rIHist2.lower_bound( imagValue2 );
              
              auxPair.first = rRHist1It->first;
              auxPair.second = rRHist2It->first;
              ++roiRJointHistogram[ auxPair ];
              
              auxPair.first = rIHist1It->first;
              auxPair.second = rIHist2It->first;
              ++roiIJointHistogram[ auxPair ];
            }
            
            ++polIt;
          }

          // transfering values to the main histograms

          locker.lock();

          roiRJointHistogramIt = roiRJointHistogram.begin();
          while( roiRJointHistogramIt != roiRJointHistogramItEnd )
          {
            paramsPtr->m_realJointHistogramPtr->operator[]( 
              roiRJointHistogramIt->first ) += roiRJointHistogramIt->second;
            ++roiRJointHistogramIt;
          }

          roiIJointHistogramIt = roiIJointHistogram.begin();
          while( roiIJointHistogramIt != roiIJointHistogramItEnd )
          {
            paramsPtr->m_imagJointHistogramPtr->operator[]( 
              roiIJointHistogramIt->first ) += roiIJointHistogramIt->second;
            ++roiIJointHistogramIt;
          }

          locker.unlock();
        }
        else
        {
          locker.unlock();
        }
      }
    }
    
    bool getJointHistograms(
      const te::rst::Raster& raster,
      const unsigned int bandIndex1,
      const unsigned int bandIndex2,      
      const std::vector< te::gm::Polygon const* >& rois,
      const unsigned int histoBins,
      const unsigned int maxThreads,
      const te::rst::PolygonIterator< double >::IterationType roisItType,
      std::map< std::pair< double, double >, unsigned int>& realJointHistogram,
      std::map< std::pair< double, double >, unsigned int>& imagJointHistogram,
      std::map<double, unsigned int> const * const rasterRealHistogram1ptr,
      std::map<double, unsigned int> const * const rasterImagHistogram1Ptr,
      std::map<double, unsigned int> const * const rasterRealHistogram2Ptr,
      std::map<double, unsigned int> const * const rasterImagHistogram2Ptr )
    {
      realJointHistogram.clear();
      imagJointHistogram.clear();

      if( bandIndex1 >= raster.getNumberOfBands() )
      {
        return false;
      }
      
      if( bandIndex2 >= raster.getNumberOfBands() )
      {
        return false;
      }
      
      // input raster hitograms

      std::map<double, unsigned int> rRHist1;
      std::map<double, unsigned int> rIHist1; 
      if( ( rasterRealHistogram1ptr != 0 ) && ( rasterImagHistogram1Ptr != 0) )
      {
        rRHist1 = *rasterRealHistogram1ptr;
        rIHist1 = *rasterImagHistogram1Ptr;
      }
      else
      {
        if( !getHistograms( raster, bandIndex1, rois, histoBins, maxThreads,
          roisItType, rRHist1, rIHist1 ) )
        {
          return false;
        }
      }
      
      std::map<double, unsigned int> rRHist2;
      std::map<double, unsigned int> rIHist2; 
      if( ( rasterRealHistogram2Ptr != 0 ) && ( rasterImagHistogram2Ptr != 0) )
      {
        rRHist2 = *rasterRealHistogram2Ptr;
        rIHist2 = *rasterImagHistogram2Ptr;
      }
      else
      {
        if( !getHistograms( raster, bandIndex2, rois, histoBins, maxThreads,
          roisItType, rRHist2, rIHist2 ) )
        {
          return false;
        }
      }
      
      // Checking if all image must be processed
      std::vector< te::gm::Polygon const* > internalRois;
      std::unique_ptr< te::gm::Polygon > rasterExtentPolPtr;      
      if( rois.empty() )
      {
        rasterExtentPolPtr.reset( (te::gm::Polygon*)
          te::gm::GetGeomFromEnvelope( raster.getGrid()->getExtent(),
          raster.getSRID() ) );

        internalRois.push_back( rasterExtentPolPtr.get() ); 
      }
      else
      {
        internalRois = rois;
      }
      
      unsigned int threadsNumber = maxThreads ? maxThreads :
        te::common::GetPhysProcNumber();      
      const double totalPhysMem = (double)te::common::GetTotalPhysicalMemory();
      const double usedVMem = (double)te::common::GetUsedVirtualMemory();
      const double totalVMem = ( (double)te::common::GetTotalVirtualMemory() );
      const double freeVMem = std::min( totalPhysMem, ( totalVMem - usedVMem ) );
      const double rasterBlockRequiredRam = std::max(
        raster.getBand( bandIndex1 )->getBlockSize(),
        raster.getBand( bandIndex2 )->getBlockSize() );
      const double maxSimultaneousRasterBlocks = 0.75 * freeVMem / rasterBlockRequiredRam;
      const unsigned int maxRasterBlocksPerThread = std::max( 1u,
        ( (unsigned int)( maxSimultaneousRasterBlocks ) ) / threadsNumber );
      
      std::vector< bool > roisProcStatus( internalRois.size(), false );

      te::rst::RasterSynchronizer rasterSync(
        *(te::rst::Raster*)(&raster), te::common::RAccess );
      std::mutex mutex;

      te::rp::GetJointHistogramsThreadParameters params;
      params.m_rasterSyncPtr = &rasterSync;
      params.m_inputBandIndex1 = bandIndex1;
      params.m_inputBandIndex2 = bandIndex2;
      params.m_roisPtr = &internalRois;
      params.m_roisProcStatusPtr = &roisProcStatus;
      params.m_returnStatus = true;
      params.m_mutexPtr = &mutex;
      params.m_rRHist1Ptr = &rRHist1;
      params.m_rIHist1Ptr = &rIHist1;
      params.m_rRHist2Ptr = &rRHist2;
      params.m_rIHist2Ptr = &rIHist2;
      params.m_realJointHistogramPtr = &realJointHistogram;
      params.m_imagJointHistogramPtr = &imagJointHistogram;
      params.m_maxRasterBlocksPerThread = maxRasterBlocksPerThread;
      params.m_roisItType = roisItType;

      {
        // mark all blocks as not processed
        
        std::fill( roisProcStatus.begin(), roisProcStatus.end(), false );

        if( maxThreads == 1 )
        {
          getJointHistogramsThreadEntry( &params );
        }
        else
        {
          std::vector< std::shared_ptr< std::thread > > threads;

          for( unsigned int threadIdx = 0 ; threadIdx < threadsNumber ;
            ++threadIdx )
          {
            std::shared_ptr< std::thread > threadPtr( new std::thread(
              &getJointHistogramsThreadEntry, &params ) );

            threads.push_back( threadPtr );
          };

          for( unsigned int threadIdx = 0 ; threadIdx < threadsNumber ;
            ++threadIdx )
          {
            threads[ threadIdx ]->join();
          };
        }
      }

      return params.m_returnStatus;
    }
    
    bool getJointHistogramStats(
      const std::map< std::pair< double, double >, unsigned int>& jointHistogram,
      const std::map<double, unsigned int>& histogram1,
      const std::map<double, unsigned int>& histogram2,
      double& covariance, double& corrlationCoef )
    {
      covariance = 0;
      corrlationCoef = 0;
      
      if( jointHistogram.empty() ) return false;
      if( histogram1.empty() ) return false;
      if( histogram2.empty() ) return false;
      
      double min1 = 0;
      double max1 = 0;
      double mean1 = 0;
      double stdDev1 = 0;
      double mode1 = 0;
      double entropy1 = 0;
      double sum1 = 0;
      double sum2_1 = 0;
      double sum3_1 = 0;
      double sum4_1 = 0;
      double variance1 = 0;
      double median1 = 0;
      getHistogramStats( histogram1, min1, max1, mean1, stdDev1, mode1,
        entropy1, sum1, sum2_1, sum3_1, sum4_1, variance1, median1 );
      
      double min2 = 0;
      double max2 = 0;
      double mean2 = 0;
      double stdDev2 = 0;
      double mode2 = 0;
      double entropy2 = 0;
      double sum2 = 0;
      double sum2_2 = 0;
      double sum3_2 = 0;
      double sum4_2 = 0;
      double variance2 = 0;
      double median2 = 0;
      getHistogramStats( histogram2, min2, max2, mean2, stdDev2, mode2,
        entropy2, sum2, sum2_2, sum3_2, sum4_2, variance2, median2 );      
      
      std::map< std::pair< double, double >, unsigned int>::const_iterator
        jointHistogramIt = jointHistogram.begin();
      const std::map< std::pair< double, double >, unsigned int>::const_iterator
        jointHistogramItEnd = jointHistogram.end();        
        
      unsigned int elementsNumber = 0;
      
      while( jointHistogramIt != jointHistogramItEnd )
      {
        covariance += 
          ( 
            ( jointHistogramIt->first.first - mean1 ) 
            *
            ( jointHistogramIt->first.second - mean2 ) 
            *
            ( (double)jointHistogramIt->second )
          );
        
        elementsNumber += jointHistogramIt->second;
        
        ++jointHistogramIt;
      }
      
      if( elementsNumber )
      {
        covariance /= (double)elementsNumber;
        
        corrlationCoef = ( stdDev1 * stdDev2 );
        
        if( corrlationCoef != 0.0 )
        {
          corrlationCoef = covariance / corrlationCoef;
        }        
      }
        
      return true;
    }    
    
    bool FillBand( te::rst::Raster& raster, const unsigned int bandIndex,
      const std::complex< double >& value )
    {
      if( bandIndex >= raster.getNumberOfBands() ) return false;
      
      te::rst::Band& band = *raster.getBand( bandIndex );
      
      double bandMax = 0;
      double bandMin = 0;
      te::rp::GetDataTypeRange( band.getProperty()->m_type, bandMin, bandMax );
      
      std::complex< double > normValue = value;
      normValue.real( std::min( bandMax, normValue.real() ) );
      normValue.real( std::max( bandMin, normValue.real() ) );
      normValue.imag( std::min( bandMax, normValue.imag() ) );
      normValue.imag( std::max( bandMin, normValue.imag() ) );
      
      const unsigned int maxBlkWidth = band.getProperty()->m_blkw;
      const unsigned int maxBlkHeight = band.getProperty()->m_blkh;
      const unsigned int nRows = raster.getNumberOfRows();
      const unsigned int nCols = raster.getNumberOfColumns();
      const unsigned int nBlocksX = band.getProperty()->m_nblocksx;
      const unsigned int nBlocksY = band.getProperty()->m_nblocksy;
      
      unsigned blkX = 0;
      unsigned int blkStartRow = 0;
      unsigned int blkStartCol = 0;
      unsigned int blkRowBound = 0;
      unsigned int blkColBound = 0;
      unsigned int row = 0 ;
      unsigned int col = 0;
      
      for( unsigned blkY = 0 ; blkY < nBlocksY ; ++blkY )
      {
        for( blkX = 0 ; blkX < nBlocksX ; ++blkX )
        {
          blkStartRow = std::min( maxBlkHeight * blkY, nRows );
          blkRowBound = std::min( maxBlkHeight * ( 1 + blkY ), nRows );
          blkStartCol = std::min( maxBlkWidth * blkX, nCols );
          blkColBound = std::min( maxBlkWidth * ( 1 + blkX ), nCols );
         
          for( row = blkStartRow ; row < blkRowBound ; ++row )
          {
            for( col = blkStartCol ; col < blkColBound ; ++col )
            {
              band.setValue( col, row, normValue );
            }
          }
        }
      }
      
      return true;
    }
    
    bool GetExternalValidDataPolygon(
      const te::rst::Raster& inputRaster,
      const std::vector< unsigned int >& inputRasterBandsIdx,
      const std::vector< double >& bandNoDataValues,
      std::unique_ptr< te::gm::Polygon >& outPolygonPtr )
    {
      outPolygonPtr.reset();
      
      // bands indexes
      
      if( inputRasterBandsIdx.empty() )
      {
        return false;
      }
      else
      {
        for( std::size_t inputRasterBandsIdxIdx = 0 ; inputRasterBandsIdxIdx <
          inputRasterBandsIdx.size() ; ++inputRasterBandsIdxIdx )
        {
          if( 
              inputRasterBandsIdx[ inputRasterBandsIdxIdx ] 
              >= 
              inputRaster.getNumberOfBands() 
            )
          {
            return false;
          }
        }
      }
      
      // no-data values
      
      std::vector< double > internalBandNoDataValues;
      
      if( bandNoDataValues.empty() )
      {
        for( std::size_t inputRasterBandsIdxIdx = 0 ; inputRasterBandsIdxIdx <
          inputRasterBandsIdx.size() ; ++inputRasterBandsIdxIdx )
        {
          internalBandNoDataValues.push_back( inputRaster.getBand( 
            inputRasterBandsIdx[ 
            inputRasterBandsIdxIdx ] )->getProperty()->m_noDataValue );
        }
      }
      else
      {
        if( bandNoDataValues.size() != inputRasterBandsIdx.size() )
        {
          return false;
        }
        
        internalBandNoDataValues = bandNoDataValues;
      }
      
      // creating edge profiles
      
      std::vector< unsigned int > tProfile;
      std::vector< unsigned int > bProfile;
      std::vector< unsigned int > lProfile;
      std::vector< unsigned int > rProfile;

      {
        const unsigned int nRows = inputRaster.getNumberOfRows();
        const unsigned int nCols = inputRaster.getNumberOfColumns();
        const unsigned int inputRasterBandsIdxSize = (unsigned int)inputRasterBandsIdx.size();
        unsigned int col = 0;
        unsigned int inputRasterBandsIdxIdx = 0;
        bool isValidPixel = false;
        double value = 0;
        
        tProfile.resize( nCols, nRows );
        bProfile.resize( nCols, nRows );
        lProfile.resize( nRows, nCols );
        rProfile.resize( nRows, nCols );      
        
        for( unsigned int row = 0 ; row < nRows ; ++row )
        {
          for( col = 0 ; col < nCols ; ++col )
          {
            isValidPixel = true;
            
            for( inputRasterBandsIdxIdx = 0 ; inputRasterBandsIdxIdx <
              inputRasterBandsIdxSize ; ++inputRasterBandsIdxIdx )
            {
              inputRaster.getValue( col, row, value, inputRasterBandsIdx[
                inputRasterBandsIdxIdx ] );
              if( value == internalBandNoDataValues[ inputRasterBandsIdxIdx ] )
              {
                isValidPixel = false;
                break;
              }
            }
            
            if( isValidPixel )
            {
              if( tProfile[ col ] == nRows )
              {
                tProfile[ col ] = row;
              }
              else
              {
                bProfile[ col ] = row;
              }
              
              if( lProfile[ row ] == nCols )
              {
                lProfile[ row ] = col;
              }
              else
              {
                rProfile[ row ] = col;
              }
            }
          }
        }
      }
      
      // creating grid polygon points
      
      // first = col, second = row
      std::list< std::pair< unsigned int, unsigned int > > gridPoints;
      
      {
        // first = col, second = row
        std::pair< unsigned int, unsigned int > auxPair;
        
        const unsigned int nRows = inputRaster.getNumberOfRows();
        const unsigned int nCols = inputRaster.getNumberOfColumns();
        
        for( unsigned int col = 0 ; col < nCols ; ++col )
        {
          if( tProfile[ col ] != nRows )
          {
            if( 
                ( lProfile[ tProfile[ col ] ] == col )
                ||
                ( rProfile[ tProfile[ col ] ] == col )
              )
            {
              auxPair.first = col;
              auxPair.second = tProfile[ col ];
              
              if( 
                  std::find( gridPoints.begin(), gridPoints.end(), auxPair ) 
                  ==
                  gridPoints.end()
                )
              {
                gridPoints.push_back( auxPair );
              }
            }
          }
          
          if( bProfile[ col ] != nRows )
          {
            if( 
                ( lProfile[ bProfile[ col ] ] == col )
                ||
                ( rProfile[ bProfile[ col ] ] == col )
              )
            {
              auxPair.first = col;
              auxPair.second = bProfile[ col ];
              
              if( 
                  std::find( gridPoints.begin(), gridPoints.end(), auxPair ) 
                  ==
                  gridPoints.end()
                )
              {
                gridPoints.push_back( auxPair );
              }
            }
          }          
        }
      }
      
      if( gridPoints.size() < 3 )
      {
        return false;
      }
      
      // calculating the convexhull
      
      std::unique_ptr< te::gm::CurvePolygon > convexHullPtr;
      
      {
        te::gm::MultiPoint pointsGeom( 0, te::gm::MultiPointType, 
          inputRaster.getSRID(), 0 );        
        std::list< std::pair< unsigned int, unsigned int > >::iterator 
          gridPointsIt = gridPoints.begin();
        const std::list< std::pair< unsigned int, unsigned int > >::iterator 
          gridPointsItEnd = gridPoints.end();
        te::gm::Point* pointPtr = 0;
        double xCoord = 0;
        double yCoord = 0;        
          
        while( gridPointsIt != gridPointsItEnd )
        {
          inputRaster.getGrid()->gridToGeo( (double)gridPointsIt->first,
            (double)gridPointsIt->second, xCoord, yCoord );                    
          
          pointPtr = new te::gm::Point( xCoord, yCoord, inputRaster.getSRID(), 
            te::gm::PointType, 0 );
          
          pointsGeom.add( pointPtr );
          
          ++gridPointsIt;
        }
        
        try
        {
          convexHullPtr.reset( (te::gm::CurvePolygon*)pointsGeom.convexHull() );
        }
        catch(...)
        {
          convexHullPtr.reset();
        }
        
        if( convexHullPtr.get() == nullptr ) return false;
        if( convexHullPtr->getExteriorRing() == nullptr ) return false;
      }
      
      // creating the output polygon
      
      outPolygonPtr.reset( new te::gm::Polygon( 0, te::gm::PolygonType, 
        inputRaster.getSRID(), 0 ) );
      outPolygonPtr->add( (te::gm::Curve*)convexHullPtr->getExteriorRing()->clone() );
      outPolygonPtr->computeMBR( true );
      
      return true;
    }
    
  } // end namespace rp
}   // end namespace te
