/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/rp/SpectralResponseFunctions.h
  \brief Spectral Response Functions.
 */

#ifndef __TERRALIB_RP_INTERNAL_SPECTRALRESPONSEFUNCTIONS_H
#define __TERRALIB_RP_INTERNAL_SPECTRALRESPONSEFUNCTIONS_H

#include "Config.h"
#include "Macros.h"

#include <map>

namespace te
{
  namespace rp
  {
    /*!
      \brief Namespace related to Spectral Response Functions.
      \note Reference unit: nm (nanometer)
    */    
    namespace srf
    {
      
      /*!
        \brief Satellite sensors    
       */            
      enum SensorType
      {
        InvalidSensor, //!< Invalid sensor.
        CBERS2BCCDB1Sensor, //!< CBERS 2B Band 1.
        CBERS2BCCDB2Sensor, //!< CBERS 2B Band 2.
        CBERS2BCCDB3Sensor, //!< CBERS 2B Band 3.
        CBERS2BCCDB4Sensor, //!< CBERS 2B Band 4.
        CBERS2BCCDB5PANSensor, //!< CBERS 2B Band 5.
        CBERS4MUXB5Sensor, //!< CBERS 4 MUX Band 5.
        CBERS4MUXB6Sensor, //!< CBERS 4 MUX Band 6.
        CBERS4MUXB7Sensor, //!< CBERS 4 MUX Band 7.
        CBERS4MUXB8Sensor, //!< CBERS 4 MUX Band 8.        
        CBERS04AWPMPANSensor, //!< CBERS 04A WPM PAN Band.
        CBERS04AWPMB1Sensor, //!< CBERS 04A WPM Band 1 Band.
        CBERS04AWPMB2Sensor, //!< CBERS 04A WPM Band 2 Band.
        CBERS04AWPMB3Sensor, //!< CBERS 04A WPM Band 3 Band.
        CBERS04AWPMB4Sensor, //!< CBERS 04A WPM Band 4 Band.        
        LANDSAT8OLIB1Sensor, //!< LANDSAT 8 OLI Band 1
        LANDSAT8OLIB2Sensor, //!< LANDSAT 8 OLI Band 2
        LANDSAT8OLIB3Sensor, //!< LANDSAT 8 OLI Band 3
        LANDSAT8OLIB4Sensor, //!< LANDSAT 8 OLI Band 4
        LANDSAT8OLIB5Sensor, //!< LANDSAT 8 OLI Band 5
        LANDSAT8OLIB6Sensor, //!< LANDSAT 8 OLI Band 6
        LANDSAT8OLIB7Sensor, //!< LANDSAT 8 OLI Band 7
        LANDSAT8OLIB8Sensor, //!< LANDSAT 8 OLI Band 8
        LANDSAT8OLIB9Sensor, //!< LANDSAT 8 OLI Band 9
        LANDSAT8OLIB10Sensor, //!< LANDSAT 8 OLI Band 10
        LANDSAT8OLIB11Sensor, //!< LANDSAT 8 OLI Band 11
        LANDSAT7ETMB1Sensor, //!< LANDSAT 7 ETM+ Band 1
        LANDSAT7ETMB2Sensor, //!< LANDSAT 7 ETM+ Band 2
        LANDSAT7ETMB3Sensor, //!< LANDSAT 7 ETM+ Band 3
        LANDSAT7ETMB4Sensor, //!< LANDSAT 7 ETM+ Band 4
        LANDSAT7ETMB5Sensor, //!< LANDSAT 7 ETM+ Band 5
        LANDSAT7ETMB7Sensor, //!< LANDSAT 7 ETM+ Band 7
        LANDSAT7ETMB8PANSensor, //!< LANDSAT 7 ETM+ Band 8 PAN
        AMAZONIA1WFIROB1Sensor, //!< AMAZONIA-1 WFI Right Optics Band 1.
        AMAZONIA1WFIROB2Sensor, //!< AMAZONIA-1 WFI Right Optics Band 2.
        AMAZONIA1WFIROB3Sensor, //!< AMAZONIA-1 WFI Right Optics Band 3.
        AMAZONIA1WFIROB4Sensor, //!< AMAZONIA-1 WFI Right Optics Band 4.
        AMAZONIA1WFILOB1Sensor, //!< AMAZONIA-1 WFI Left Optics Band 1.
        AMAZONIA1WFILOB2Sensor, //!< AMAZONIA-1 WFI Left Optics Band 2.
        AMAZONIA1WFILOB3Sensor, //!< AMAZONIA-1 WFI Left Optics Band 3.
        AMAZONIA1WFILOB4Sensor, //!< AMAZONIA-1 WFI Left Optics Band 4.
        CBERS04AMUXB5Sensor, //!< CBERS 04A MUX B5 Band.
        CBERS04AMUXB6Sensor, //!< CBERS 04A MUX B6 Band.
        CBERS04AMUXB7Sensor, //!< CBERS 04A MUX B7 Band.
        CBERS04AMUXB8Sensor //!< CBERS 04A MUX B8 Band.
      };
      
      static const double CBERS2B_CCDB1_SRF[][2] = 
      {  
        {424.385,0.001},
        {424.715,0.001},
        {425.045,0.001},
        {425.375,0.001},
        {425.705,0.001},
        {426.035,0.001},
        {426.365,0.001},
        {426.695,0.002},
        {427.025,0.002},
        {427.355,0.002},
        {427.685,0.002},
        {428.015,0.002},
        {428.345,0.002},
        {428.675,0.002},
        {429.005,0.002},
        {429.335,0.003},
        {429.665,0.003},
        {429.995,0.003},
        {430.325,0.003},
        {430.655,0.003},
        {430.985,0.003},
        {431.315,0.003},
        {431.645,0.004},
        {431.975,0.004},
        {432.305,0.004},
        {432.635,0.004},
        {432.965,0.004},
        {433.295,0.005},
        {433.625,0.005},
        {433.955,0.005},
        {434.285,0.005},
        {434.615,0.005},
        {434.945,0.006},
        {435.275,0.006},
        {435.605,0.006},
        {435.935,0.006},
        {436.265,0.006},
        {436.595,0.007},
        {436.925,0.007},
        {437.255,0.007},
        {437.585,0.007},
        {437.915,0.007},
        {438.245,0.008},
        {438.575,0.008},
        {438.905,0.008},
        {439.235,0.009},
        {439.565,0.009},
        {439.895,0.009},
        {440.225,0.010},
        {440.555,0.010},
        {440.885,0.010},
        {441.215,0.011},
        {441.545,0.011},
        {441.875,0.011},
        {442.205,0.011},
        {442.535,0.012},
        {442.865,0.012},
        {443.195,0.013},
        {443.525,0.013},
        {443.855,0.014},
        {444.185,0.014},
        {444.515,0.014},
        {444.845,0.015},
        {445.175,0.015},
        {445.505,0.016},
        {445.835,0.016},
        {446.165,0.017},
        {446.495,0.017},
        {446.825,0.017},
        {447.155,0.018},
        {447.485,0.018},
        {447.815,0.019},
        {448.145,0.019},
        {448.475,0.020},
        {448.805,0.021},
        {449.135,0.023},
        {449.465,0.024},
        {449.795,0.025},
        {450.125,0.026},
        {450.455,0.030},
        {450.785,0.035},
        {451.115,0.040},
        {451.445,0.043},
        {451.775,0.045},
        {452.105,0.047},
        {452.435,0.049},
        {452.765,0.051},
        {453.095,0.053},
        {453.425,0.056},
        {453.755,0.060},
        {454.085,0.064},
        {454.415,0.068},
        {454.745,0.071},
        {455.075,0.075},
        {455.405,0.079},
        {455.735,0.083},
        {456.065,0.087},
        {456.395,0.091},
        {456.725,0.095},
        {457.055,0.099},
        {457.385,0.103},
        {457.715,0.107},
        {458.045,0.111},
        {458.375,0.115},
        {458.705,0.119},
        {459.035,0.123},
        {459.365,0.128},
        {459.695,0.137},
        {460.025,0.146},
        {460.355,0.155},
        {460.685,0.165},
        {461.015,0.176},
        {461.345,0.186},
        {461.675,0.196},
        {462.005,0.207},
        {462.335,0.217},
        {462.665,0.227},
        {462.995,0.237},
        {463.325,0.248},
        {463.655,0.258},
        {463.985,0.268},
        {464.315,0.279},
        {464.645,0.289},
        {464.975,0.299},
        {465.305,0.310},
        {465.635,0.319},
        {465.965,0.329},
        {466.295,0.339},
        {466.625,0.349},
        {466.955,0.358},
        {467.285,0.368},
        {467.615,0.378},
        {467.945,0.388},
        {468.275,0.397},
        {468.605,0.407},
        {468.935,0.417},
        {469.265,0.427},
        {469.595,0.434},
        {469.925,0.440},
        {470.255,0.446},
        {470.585,0.453},
        {470.915,0.459},
        {471.245,0.465},
        {471.575,0.471},
        {471.905,0.477},
        {472.235,0.483},
        {472.565,0.489},
        {472.895,0.494},
        {473.225,0.500},
        {473.555,0.506},
        {473.885,0.512},
        {474.215,0.518},
        {474.545,0.523},
        {474.875,0.529},
        {475.205,0.535},
        {475.535,0.541},
        {475.865,0.547},
        {476.195,0.552},
        {476.525,0.558},
        {476.855,0.564},
        {477.185,0.570},
        {477.515,0.576},
        {477.845,0.582},
        {478.175,0.588},
        {478.505,0.594},
        {478.835,0.600},
        {479.165,0.606},
        {479.495,0.612},
        {479.825,0.618},
        {480.155,0.624},
        {480.485,0.630},
        {480.815,0.637},
        {481.145,0.643},
        {481.475,0.650},
        {481.805,0.657},
        {482.135,0.663},
        {482.465,0.670},
        {482.795,0.676},
        {483.125,0.683},
        {483.455,0.690},
        {483.785,0.696},
        {484.115,0.703},
        {484.445,0.709},
        {484.775,0.714},
        {485.105,0.719},
        {485.435,0.723},
        {485.765,0.728},
        {486.095,0.733},
        {486.425,0.738},
        {486.755,0.742},
        {487.085,0.747},
        {487.415,0.752},
        {487.745,0.756},
        {488.075,0.761},
        {488.405,0.766},
        {488.735,0.768},
        {489.065,0.770},
        {489.395,0.772},
        {489.725,0.775},
        {490.055,0.777},
        {490.385,0.779},
        {490.715,0.781},
        {491.045,0.783},
        {491.375,0.785},
        {491.705,0.787},
        {492.035,0.789},
        {492.365,0.791},
        {492.695,0.793},
        {493.025,0.794},
        {493.355,0.796},
        {493.685,0.798},
        {494.015,0.800},
        {494.345,0.802},
        {494.675,0.804},
        {495.005,0.806},
        {495.335,0.808},
        {495.665,0.810},
        {495.995,0.812},
        {496.325,0.814},
        {496.655,0.816},
        {496.985,0.818},
        {497.315,0.820},
        {497.645,0.822},
        {497.975,0.824},
        {498.305,0.826},
        {498.635,0.828},
        {498.965,0.829},
        {499.295,0.830},
        {499.625,0.831},
        {499.955,0.832},
        {500.285,0.833},
        {500.615,0.834},
        {500.945,0.835},
        {501.275,0.836},
        {501.605,0.837},
        {501.935,0.838},
        {502.265,0.839},
        {502.595,0.840},
        {502.925,0.841},
        {503.255,0.842},
        {503.585,0.843},
        {503.915,0.844},
        {504.245,0.845},
        {504.575,0.846},
        {504.905,0.847},
        {505.235,0.848},
        {505.565,0.849},
        {505.895,0.850},
        {506.225,0.851},
        {506.555,0.852},
        {506.885,0.853},
        {507.215,0.854},
        {507.545,0.855},
        {507.875,0.856},
        {508.205,0.857},
        {508.535,0.858},
        {508.865,0.856},
        {509.195,0.854},
        {509.525,0.852},
        {509.855,0.849},
        {510.185,0.847},
        {510.515,0.845},
        {510.845,0.843},
        {511.175,0.840},
        {511.505,0.838},
        {511.835,0.836},
        {512.165,0.833},
        {512.495,0.831},
        {512.825,0.829},
        {513.155,0.827},
        {513.485,0.824},
        {513.815,0.822},
        {514.145,0.819},
        {514.475,0.817},
        {514.805,0.814},
        {515.135,0.812},
        {515.465,0.809},
        {515.795,0.807},
        {516.125,0.804},
        {516.455,0.802},
        {516.785,0.799},
        {517.115,0.797},
        {517.445,0.794},
        {517.775,0.792},
        {518.105,0.789},
        {518.435,0.787},
        {518.765,0.781},
        {519.095,0.773},
        {519.425,0.764},
        {519.755,0.756},
        {520.085,0.748},
        {520.415,0.739},
        {520.745,0.731},
        {521.075,0.722},
        {521.405,0.714},
        {521.735,0.705},
        {522.065,0.697},
        {522.395,0.688},
        {522.725,0.679},
        {523.055,0.670},
        {523.385,0.661},
        {523.715,0.653},
        {524.045,0.644},
        {524.375,0.635},
        {524.705,0.625},
        {525.035,0.612},
        {525.365,0.599},
        {525.695,0.586},
        {526.025,0.572},
        {526.355,0.559},
        {526.685,0.547},
        {527.015,0.540},
        {527.345,0.534},
        {527.675,0.527},
        {528.005,0.520},
        {528.335,0.514},
        {528.665,0.507},
        {528.995,0.501},
        {529.325,0.494},
        {529.655,0.487},
        {529.985,0.481},
        {530.315,0.474},
        {530.645,0.468},
        {530.975,0.442},
        {531.305,0.415},
        {531.635,0.389},
        {531.965,0.379},
        {532.295,0.371},
        {532.625,0.362},
        {532.955,0.354},
        {533.285,0.346},
        {533.615,0.337},
        {533.945,0.329},
        {534.275,0.321},
        {534.605,0.312},
        {534.935,0.300},
        {535.265,0.287},
        {535.595,0.274},
        {535.925,0.261},
        {536.255,0.249},
        {536.585,0.236},
        {536.915,0.223},
        {537.245,0.210},
        {537.575,0.197},
        {537.905,0.184},
        {538.235,0.171},
        {538.565,0.158},
        {538.895,0.151},
        {539.225,0.146},
        {539.555,0.141},
        {539.885,0.137},
        {540.215,0.132},
        {540.545,0.128},
        {540.875,0.123},
        {541.205,0.119},
        {541.535,0.114},
        {541.865,0.110},
        {542.195,0.105},
        {542.525,0.101},
        {542.855,0.096},
        {543.185,0.092},
        {543.515,0.087},
        {543.845,0.083},
        {544.175,0.079},
        {544.505,0.075},
        {544.835,0.071},
        {545.165,0.067},
        {545.495,0.063},
        {545.825,0.059},
        {546.155,0.055},
        {546.485,0.051},
        {546.815,0.047},
        {547.145,0.042},
        {547.475,0.038},
        {547.805,0.034},
        {548.135,0.030},
        {548.465,0.026},
        {548.795,0.023},
        {549.125,0.023},
        {549.455,0.023},
        {549.785,0.022},
        {550.115,0.022},
        {550.445,0.022},
        {550.775,0.022},
        {551.105,0.022},
        {551.435,0.022},
        {551.765,0.021},
        {552.095,0.021},
        {552.425,0.021},
        {552.755,0.021},
        {553.085,0.021},
        {553.415,0.020},
        {553.745,0.020},
        {554.075,0.020},
        {554.405,0.020},
        {554.735,0.020},
        {555.065,0.020},
        {555.395,0.019},
        {555.725,0.019},
        {556.055,0.018},
        {556.385,0.018},
        {556.715,0.017},
        {557.045,0.016},
        {557.375,0.015},
        {557.705,0.014},
        {558.035,0.013},
        {558.365,0.013},
        {558.695,0.012},
        {559.025,0.012},
        {559.355,0.013},
        {559.685,0.013},
        {560.015,0.014},
        {560.345,0.014},
        {560.675,0.015},
        {561.005,0.016},
        {561.335,0.016},
        {561.665,0.017},
        {561.995,0.018},
        {562.325,0.018},
        {562.655,0.019},
        {562.985,0.019},
        {563.315,0.019},
        {563.645,0.019},
        {563.975,0.018},
        {564.305,0.018},
        {564.635,0.018},
        {564.965,0.018},
        {565.295,0.017},
        {565.625,0.017},
        {565.955,0.017},
        {566.285,0.017},
        {566.615,0.016},
        {566.945,0.016},
        {567.275,0.016},
        {567.605,0.016},
        {567.935,0.016},
        {568.265,0.019},
        {568.595,0.021},
        {568.925,0.023},
        {569.255,0.022},
        {569.585,0.022},
        {569.915,0.021},
        {570.245,0.021},
        {570.575,0.020},
        {570.905,0.020},
        {571.235,0.019},
        {571.565,0.019},
        {571.895,0.018},
        {572.225,0.018},
        {572.555,0.017},
        {572.885,0.017},
        {573.215,0.016},
        {573.545,0.016},
        {573.875,0.015},
        {574.205,0.015},
        {574.535,0.015},
        {574.865,0.015},
        {575.195,0.015},
        {575.525,0.014},
        {575.855,0.014},
        {576.185,0.014},
        {576.515,0.014},
        {576.845,0.014},
        {577.175,0.014},
        {577.505,0.013},
        {577.835,0.013},
        {578.165,0.013},
        {578.495,0.013},
        {578.825,0.013},
        {579.155,0.012},
        {579.485,0.012},
        {579.815,0.012},
        {580.145,0.012},
        {580.475,0.012},
        {580.805,0.012},
        {581.135,0.011},
        {581.465,0.011},
        {581.795,0.011},
        {582.125,0.011},
        {582.455,0.011},
        {582.785,0.011},
        {583.115,0.010},
        {583.445,0.010},
        {583.775,0.010},
        {584.105,0.010},
        {584.435,0.010},
        {584.765,0.010},
        {585.095,0.009},
        {585.425,0.009},
        {585.755,0.009},
        {586.085,0.009},
        {586.415,0.009},
        {586.745,0.009},
        {587.075,0.009},
        {587.405,0.008},
        {587.735,0.008},
        {588.065,0.008},
        {588.395,0.008},
        {588.725,0.008},
        {589.055,0.008},
        {589.385,0.007},
        {589.715,0.007},
        {590.045,0.007},
        {590.375,0.007},
        {590.705,0.007},
        {591.035,0.007},
        {591.365,0.007},
        {591.695,0.006},
        {592.025,0.006},
        {592.355,0.006},
        {592.685,0.006},
        {593.015,0.006},
        {593.345,0.006},
        {593.675,0.005},
        {594.005,0.005},
        {594.335,0.005},
        {594.665,0.005},
        {594.995,0.005},
        {595.325,0.005},
        {595.655,0.004},
        {595.985,0.004},
        {596.315,0.004},
        {596.645,0.004},
        {596.975,0.004},
        {597.305,0.004},
        {597.635,0.004},
        {597.965,0.004},
        {598.295,0.004},
        {598.625,0.004},
        {598.955,0.004},
        {599.285,0.004},
        {599.615,0.004},
        {599.945,0.004},
        {600.275,0.004},
        {600.605,0.004},
        {600.935,0.004},
        {601.265,0.004},
        {601.595,0.004},
        {601.925,0.004},
        {602.255,0.004},
        {602.585,0.004},
        {602.915,0.004},
        {603.245,0.004},
        {603.575,0.004},
        {603.905,0.004},
        {604.235,0.004},
        {604.565,0.004},
        {604.895,0.004},
        {605.225,0.004},
        {605.555,0.004},
        {605.885,0.004}
      };

      static const double CBERS2B_CCDB2_SRF[][2] = 
      {  
        {504.575,0.008},
        {504.905,0.008},
        {505.235,0.009},
        {505.565,0.010},
        {505.895,0.010},
        {506.225,0.011},
        {506.555,0.012},
        {506.885,0.012},
        {507.215,0.013},
        {507.545,0.013},
        {507.875,0.014},
        {508.205,0.015},
        {508.535,0.015},
        {508.865,0.018},
        {509.195,0.020},
        {509.525,0.023},
        {509.855,0.026},
        {510.185,0.029},
        {510.515,0.032},
        {510.845,0.035},
        {511.175,0.038},
        {511.505,0.041},
        {511.835,0.044},
        {512.165,0.047},
        {512.495,0.050},
        {512.825,0.054},
        {513.155,0.057},
        {513.485,0.060},
        {513.815,0.063},
        {514.145,0.066},
        {514.475,0.069},
        {514.805,0.072},
        {515.135,0.075},
        {515.465,0.078},
        {515.795,0.081},
        {516.125,0.084},
        {516.455,0.087},
        {516.785,0.090},
        {517.115,0.094},
        {517.445,0.097},
        {517.775,0.100},
        {518.105,0.103},
        {518.435,0.106},
        {518.765,0.112},
        {519.095,0.122},
        {519.425,0.131},
        {519.755,0.141},
        {520.085,0.150},
        {520.415,0.160},
        {520.745,0.169},
        {521.075,0.179},
        {521.405,0.188},
        {521.735,0.198},
        {522.065,0.208},
        {522.395,0.217},
        {522.725,0.227},
        {523.055,0.236},
        {523.385,0.246},
        {523.715,0.255},
        {524.045,0.265},
        {524.375,0.274},
        {524.705,0.284},
        {525.035,0.293},
        {525.365,0.303},
        {525.695,0.313},
        {526.025,0.323},
        {526.355,0.333},
        {526.685,0.344},
        {527.015,0.354},
        {527.345,0.365},
        {527.675,0.375},
        {528.005,0.386},
        {528.335,0.396},
        {528.665,0.407},
        {528.995,0.419},
        {529.325,0.431},
        {529.655,0.443},
        {529.985,0.456},
        {530.315,0.468},
        {530.645,0.480},
        {530.975,0.492},
        {531.305,0.504},
        {531.635,0.516},
        {531.965,0.529},
        {532.295,0.541},
        {532.625,0.553},
        {532.955,0.565},
        {533.285,0.577},
        {533.615,0.589},
        {533.945,0.602},
        {534.275,0.614},
        {534.605,0.626},
        {534.935,0.639},
        {535.265,0.652},
        {535.595,0.664},
        {535.925,0.677},
        {536.255,0.690},
        {536.585,0.703},
        {536.915,0.716},
        {537.245,0.729},
        {537.575,0.742},
        {537.905,0.755},
        {538.235,0.768},
        {538.565,0.781},
        {538.895,0.790},
        {539.225,0.798},
        {539.555,0.805},
        {539.885,0.812},
        {540.215,0.818},
        {540.545,0.825},
        {540.875,0.831},
        {541.205,0.837},
        {541.535,0.844},
        {541.865,0.850},
        {542.195,0.856},
        {542.525,0.862},
        {542.855,0.869},
        {543.185,0.875},
        {543.515,0.881},
        {543.845,0.888},
        {544.175,0.894},
        {544.505,0.900},
        {544.835,0.906},
        {545.165,0.913},
        {545.495,0.919},
        {545.825,0.925},
        {546.155,0.932},
        {546.485,0.938},
        {546.815,0.944},
        {547.145,0.950},
        {547.475,0.955},
        {547.805,0.961},
        {548.135,0.967},
        {548.465,0.972},
        {548.795,0.977},
        {549.125,0.978},
        {549.455,0.979},
        {549.785,0.980},
        {550.115,0.981},
        {550.445,0.981},
        {550.775,0.982},
        {551.105,0.983},
        {551.435,0.984},
        {551.765,0.985},
        {552.095,0.985},
        {552.425,0.985},
        {552.755,0.986},
        {553.085,0.986},
        {553.415,0.987},
        {553.745,0.987},
        {554.075,0.987},
        {554.405,0.988},
        {554.735,0.988},
        {555.065,0.988},
        {555.395,0.989},
        {555.725,0.989},
        {556.055,0.989},
        {556.385,0.990},
        {556.715,0.990},
        {557.045,0.991},
        {557.375,0.991},
        {557.705,0.991},
        {558.035,0.992},
        {558.365,0.992},
        {558.695,0.992},
        {559.025,0.993},
        {559.355,0.993},
        {559.685,0.993},
        {560.015,0.993},
        {560.345,0.994},
        {560.675,0.994},
        {561.005,0.994},
        {561.335,0.994},
        {561.665,0.995},
        {561.995,0.995},
        {562.325,0.995},
        {562.655,0.995},
        {562.985,0.996},
        {563.315,0.996},
        {563.645,0.996},
        {563.975,0.996},
        {564.305,0.997},
        {564.635,0.997},
        {564.965,0.997},
        {565.295,0.997},
        {565.625,0.998},
        {565.955,0.998},
        {566.285,0.998},
        {566.615,0.998},
        {566.945,0.999},
        {567.275,0.999},
        {567.605,0.999},
        {567.935,0.999},
        {568.265,1.000},
        {568.595,1.000},
        {568.925,1.000},
        {569.255,1.000},
        {569.585,0.999},
        {569.915,0.999},
        {570.245,0.998},
        {570.575,0.998},
        {570.905,0.997},
        {571.235,0.997},
        {571.565,0.997},
        {571.895,0.996},
        {572.225,0.996},
        {572.555,0.995},
        {572.885,0.995},
        {573.215,0.995},
        {573.545,0.994},
        {573.875,0.994},
        {574.205,0.993},
        {574.535,0.993},
        {574.865,0.992},
        {575.195,0.991},
        {575.525,0.990},
        {575.855,0.989},
        {576.185,0.988},
        {576.515,0.987},
        {576.845,0.986},
        {577.175,0.985},
        {577.505,0.984},
        {577.835,0.983},
        {578.165,0.982},
        {578.495,0.981},
        {578.825,0.980},
        {579.155,0.979},
        {579.485,0.978},
        {579.815,0.977},
        {580.145,0.975},
        {580.475,0.973},
        {580.805,0.970},
        {581.135,0.968},
        {581.465,0.966},
        {581.795,0.964},
        {582.125,0.962},
        {582.455,0.961},
        {582.785,0.959},
        {583.115,0.957},
        {583.445,0.955},
        {583.775,0.954},
        {584.105,0.952},
        {584.435,0.950},
        {584.765,0.949},
        {585.095,0.947},
        {585.425,0.945},
        {585.755,0.944},
        {586.085,0.942},
        {586.415,0.940},
        {586.745,0.938},
        {587.075,0.936},
        {587.405,0.934},
        {587.735,0.932},
        {588.065,0.930},
        {588.395,0.929},
        {588.725,0.927},
        {589.055,0.925},
        {589.385,0.923},
        {589.715,0.921},
        {590.045,0.918},
        {590.375,0.912},
        {590.705,0.905},
        {591.035,0.899},
        {591.365,0.893},
        {591.695,0.886},
        {592.025,0.880},
        {592.355,0.874},
        {592.685,0.867},
        {593.015,0.861},
        {593.345,0.855},
        {593.675,0.849},
        {594.005,0.842},
        {594.335,0.836},
        {594.665,0.830},
        {594.995,0.823},
        {595.325,0.817},
        {595.655,0.811},
        {595.985,0.805},
        {596.315,0.798},
        {596.645,0.792},
        {596.975,0.786},
        {597.305,0.778},
        {597.635,0.770},
        {597.965,0.762},
        {598.295,0.754},
        {598.625,0.746},
        {598.955,0.738},
        {599.285,0.730},
        {599.615,0.722},
        {599.945,0.714},
        {600.275,0.705},
        {600.605,0.696},
        {600.935,0.687},
        {601.265,0.677},
        {601.595,0.668},
        {601.925,0.659},
        {602.255,0.650},
        {602.585,0.640},
        {602.915,0.631},
        {603.245,0.623},
        {603.575,0.614},
        {603.905,0.606},
        {604.235,0.597},
        {604.565,0.589},
        {604.895,0.581},
        {605.225,0.572},
        {605.555,0.564},
        {605.885,0.555},
        {606.215,0.544},
        {606.545,0.530},
        {606.875,0.516},
        {607.205,0.503},
        {607.535,0.489},
        {607.865,0.475},
        {608.195,0.464},
        {608.525,0.455},
        {608.855,0.447},
        {609.185,0.438},
        {609.515,0.430},
        {609.845,0.422},
        {610.175,0.413},
        {610.505,0.405},
        {610.835,0.396},
        {611.165,0.388},
        {611.495,0.379},
        {611.825,0.370},
        {612.155,0.361},
        {612.485,0.353},
        {612.815,0.344},
        {613.145,0.335},
        {613.475,0.326},
        {613.805,0.317},
        {614.135,0.309},
        {614.465,0.302},
        {614.795,0.294},
        {615.125,0.287},
        {615.455,0.280},
        {615.785,0.272},
        {616.115,0.265},
        {616.445,0.258},
        {616.775,0.250},
        {617.105,0.243},
        {617.435,0.236},
        {617.765,0.228},
        {618.095,0.221},
        {618.425,0.214},
        {618.755,0.206},
        {619.085,0.199},
        {619.415,0.192},
        {619.745,0.184},
        {620.075,0.177},
        {620.405,0.173},
        {620.735,0.169},
        {621.065,0.165},
        {621.395,0.161},
        {621.725,0.158},
        {622.055,0.154},
        {622.385,0.150},
        {622.715,0.147},
        {623.045,0.143},
        {623.375,0.140},
        {623.705,0.136},
        {624.035,0.132},
        {624.365,0.129},
        {624.695,0.125},
        {625.025,0.122},
        {625.355,0.118},
        {625.685,0.115},
        {626.015,0.111},
        {626.345,0.108},
        {626.675,0.104},
        {627.005,0.100},
        {627.335,0.097},
        {627.665,0.093},
        {627.995,0.090},
        {628.325,0.087},
        {628.655,0.084},
        {628.985,0.082},
        {629.315,0.080},
        {629.645,0.078},
        {629.975,0.077},
        {630.305,0.076},
        {630.635,0.074},
        {630.965,0.073},
        {631.295,0.072},
        {631.625,0.070},
        {631.955,0.069},
        {632.285,0.068},
        {632.615,0.066},
        {632.945,0.065},
        {633.275,0.064},
        {633.605,0.062},
        {633.935,0.061},
        {634.265,0.059},
        {634.595,0.058},
        {634.925,0.057},
        {635.255,0.055},
        {635.585,0.054},
        {635.915,0.053},
        {636.245,0.051},
        {636.575,0.050},
        {636.905,0.049},
        {637.235,0.047},
        {637.565,0.046},
        {637.895,0.045},
        {638.225,0.043},
        {638.555,0.042},
        {638.885,0.040},
        {639.215,0.039},
        {639.545,0.038},
        {639.875,0.036},
        {640.205,0.035},
        {640.535,0.034},
        {640.865,0.032},
        {641.195,0.031},
        {641.525,0.030},
        {641.855,0.028},
        {642.185,0.027},
        {642.515,0.026},
        {642.845,0.026},
        {643.175,0.025},
        {643.505,0.025},
        {643.835,0.024},
        {644.165,0.024},
        {644.495,0.023},
        {644.825,0.023},
        {645.155,0.022},
        {645.485,0.021},
        {645.815,0.021},
        {646.145,0.020},
        {646.475,0.020},
        {646.805,0.019},
        {647.135,0.019},
        {647.465,0.018},
        {647.795,0.018},
        {648.125,0.017},
        {648.455,0.017},
        {648.785,0.016},
        {649.115,0.016},
        {649.445,0.015},
        {649.775,0.015},
        {650.105,0.015},
        {650.435,0.014},
        {650.765,0.014},
        {651.095,0.014},
        {651.425,0.014},
        {651.755,0.013},
        {652.085,0.013},
        {652.415,0.013},
        {652.745,0.013},
        {653.075,0.012},
        {653.405,0.012},
        {653.735,0.012},
        {654.065,0.012},
        {654.395,0.011},
        {654.725,0.011},
        {655.055,0.011},
        {655.385,0.011},
        {655.715,0.010},
        {656.045,0.010},
        {656.375,0.010},
        {656.705,0.010},
        {657.035,0.009},
        {657.365,0.009},
        {657.695,0.009},
        {658.025,0.009},
        {658.355,0.008},
        {658.685,0.008},
        {659.015,0.008},
        {659.345,0.008},
        {659.675,0.007},
        {660.005,0.007},
        {660.335,0.007},
        {660.665,0.007},
        {660.995,0.007},
        {661.325,0.007},
        {661.655,0.007},
        {661.985,0.007},
        {662.315,0.006},
        {662.645,0.006},
        {662.975,0.006},
        {663.305,0.006},
        {663.635,0.006},
        {663.965,0.006},
        {664.295,0.006},
        {664.625,0.005},
        {664.955,0.005},
        {665.285,0.005},
        {665.615,0.005},
        {665.945,0.005},
        {666.275,0.005},
        {666.605,0.005},
        {666.935,0.004},
        {667.265,0.004},
        {667.595,0.004},
        {667.925,0.004},
        {668.255,0.004},
        {668.585,0.004},
        {668.915,0.004},
        {669.245,0.004},
        {669.575,0.003},
        {669.905,0.003},
        {670.235,0.003},
        {670.565,0.003},
        {670.895,0.003},
        {671.225,0.003},
        {671.555,0.003},
        {671.885,0.003},
        {672.215,0.003},
        {672.545,0.003},
        {672.875,0.003},
        {673.205,0.002},
        {673.535,0.002},
        {673.865,0.002},
        {674.195,0.002},
        {674.525,0.002},
        {674.855,0.002},
        {675.185,0.002},
        {675.515,0.002},
        {675.845,0.002},
        {676.175,0.002},
        {676.505,0.001},
        {676.835,0.001},
        {677.165,0.001},
        {677.495,0.001},
        {677.825,0.001},
        {678.155,0.001},
        {678.485,0.001},
        {678.815,0.001},
        {679.145,0.001},
        {679.475,0.001}
      };
      
      static const double CBERS2B_CCDB3_SRF[][2] = 
      {  
        {603.905,0.001},
        {604.235,0.001},
        {604.565,0.001},
        {604.895,0.002},
        {605.225,0.002},
        {605.555,0.002},
        {605.885,0.002},
        {606.215,0.003},
        {606.545,0.003},
        {606.875,0.003},
        {607.205,0.004},
        {607.535,0.004},
        {607.865,0.004},
        {608.195,0.004},
        {608.525,0.005},
        {608.855,0.005},
        {609.185,0.005},
        {609.515,0.006},
        {609.845,0.006},
        {610.175,0.006},
        {610.505,0.006},
        {610.835,0.007},
        {611.165,0.007},
        {611.495,0.007},
        {611.825,0.007},
        {612.155,0.008},
        {612.485,0.009},
        {612.815,0.009},
        {613.145,0.010},
        {613.475,0.011},
        {613.805,0.011},
        {614.135,0.012},
        {614.465,0.013},
        {614.795,0.014},
        {615.125,0.014},
        {615.455,0.015},
        {615.785,0.016},
        {616.115,0.017},
        {616.445,0.017},
        {616.775,0.018},
        {617.105,0.019},
        {617.435,0.019},
        {617.765,0.020},
        {618.095,0.021},
        {618.425,0.022},
        {618.755,0.022},
        {619.085,0.023},
        {619.415,0.028},
        {619.745,0.033},
        {620.075,0.038},
        {620.405,0.044},
        {620.735,0.049},
        {621.065,0.055},
        {621.395,0.061},
        {621.725,0.066},
        {622.055,0.072},
        {622.385,0.078},
        {622.715,0.085},
        {623.045,0.092},
        {623.375,0.098},
        {623.705,0.105},
        {624.035,0.111},
        {624.365,0.118},
        {624.695,0.125},
        {625.025,0.131},
        {625.355,0.138},
        {625.685,0.144},
        {626.015,0.151},
        {626.345,0.157},
        {626.675,0.162},
        {627.005,0.167},
        {627.335,0.172},
        {627.665,0.177},
        {627.995,0.182},
        {628.325,0.189},
        {628.655,0.198},
        {628.985,0.206},
        {629.315,0.219},
        {629.645,0.235},
        {629.975,0.252},
        {630.305,0.268},
        {630.635,0.285},
        {630.965,0.301},
        {631.295,0.316},
        {631.625,0.330},
        {631.955,0.344},
        {632.285,0.358},
        {632.615,0.372},
        {632.945,0.385},
        {633.275,0.403},
        {633.605,0.426},
        {633.935,0.450},
        {634.265,0.475},
        {634.595,0.500},
        {634.925,0.525},
        {635.255,0.547},
        {635.585,0.561},
        {635.915,0.575},
        {636.245,0.589},
        {636.575,0.603},
        {636.905,0.617},
        {637.235,0.631},
        {637.565,0.652},
        {637.895,0.672},
        {638.225,0.691},
        {638.555,0.699},
        {638.885,0.708},
        {639.215,0.717},
        {639.545,0.726},
        {639.875,0.735},
        {640.205,0.743},
        {640.535,0.750},
        {640.865,0.756},
        {641.195,0.762},
        {641.525,0.766},
        {641.855,0.770},
        {642.185,0.774},
        {642.515,0.778},
        {642.845,0.781},
        {643.175,0.785},
        {643.505,0.788},
        {643.835,0.791},
        {644.165,0.794},
        {644.495,0.796},
        {644.825,0.799},
        {645.155,0.802},
        {645.485,0.805},
        {645.815,0.808},
        {646.145,0.811},
        {646.475,0.813},
        {646.805,0.816},
        {647.135,0.819},
        {647.465,0.823},
        {647.795,0.827},
        {648.125,0.831},
        {648.455,0.836},
        {648.785,0.840},
        {649.115,0.844},
        {649.445,0.848},
        {649.775,0.852},
        {650.105,0.856},
        {650.435,0.859},
        {650.765,0.859},
        {651.095,0.860},
        {651.425,0.861},
        {651.755,0.862},
        {652.085,0.862},
        {652.415,0.863},
        {652.745,0.864},
        {653.075,0.864},
        {653.405,0.865},
        {653.735,0.866},
        {654.065,0.866},
        {654.395,0.867},
        {654.725,0.868},
        {655.055,0.868},
        {655.385,0.869},
        {655.715,0.870},
        {656.045,0.871},
        {656.375,0.871},
        {656.705,0.872},
        {657.035,0.873},
        {657.365,0.873},
        {657.695,0.874},
        {658.025,0.875},
        {658.355,0.875},
        {658.685,0.876},
        {659.015,0.877},
        {659.345,0.877},
        {659.675,0.877},
        {660.005,0.877},
        {660.335,0.876},
        {660.665,0.876},
        {660.995,0.875},
        {661.325,0.875},
        {661.655,0.874},
        {661.985,0.874},
        {662.315,0.874},
        {662.645,0.873},
        {662.975,0.873},
        {663.305,0.872},
        {663.635,0.872},
        {663.965,0.872},
        {664.295,0.871},
        {664.625,0.871},
        {664.955,0.870},
        {665.285,0.870},
        {665.615,0.869},
        {665.945,0.868},
        {666.275,0.867},
        {666.605,0.866},
        {666.935,0.865},
        {667.265,0.864},
        {667.595,0.863},
        {667.925,0.862},
        {668.255,0.861},
        {668.585,0.861},
        {668.915,0.860},
        {669.245,0.859},
        {669.575,0.859},
        {669.905,0.859},
        {670.235,0.860},
        {670.565,0.861},
        {670.895,0.861},
        {671.225,0.862},
        {671.555,0.863},
        {671.885,0.863},
        {672.215,0.864},
        {672.545,0.864},
        {672.875,0.865},
        {673.205,0.866},
        {673.535,0.866},
        {673.865,0.867},
        {674.195,0.868},
        {674.525,0.868},
        {674.855,0.869},
        {675.185,0.869},
        {675.515,0.870},
        {675.845,0.869},
        {676.175,0.869},
        {676.505,0.869},
        {676.835,0.868},
        {677.165,0.868},
        {677.495,0.868},
        {677.825,0.867},
        {678.155,0.867},
        {678.485,0.867},
        {678.815,0.867},
        {679.145,0.866},
        {679.475,0.866},
        {679.805,0.866},
        {680.135,0.866},
        {680.465,0.866},
        {680.795,0.866},
        {681.125,0.866},
        {681.455,0.866},
        {681.785,0.866},
        {682.115,0.866},
        {682.445,0.866},
        {682.775,0.866},
        {683.105,0.866},
        {683.435,0.866},
        {683.765,0.866},
        {684.095,0.866},
        {684.425,0.866},
        {684.755,0.866},
        {685.085,0.866},
        {685.415,0.866},
        {685.745,0.866},
        {686.075,0.866},
        {686.405,0.866},
        {686.735,0.866},
        {687.065,0.866},
        {687.395,0.866},
        {687.725,0.866},
        {688.055,0.866},
        {688.385,0.866},
        {688.715,0.866},
        {689.045,0.866},
        {689.375,0.866},
        {689.705,0.861},
        {690.035,0.855},
        {690.365,0.848},
        {690.695,0.842},
        {691.025,0.836},
        {691.355,0.830},
        {691.685,0.824},
        {692.015,0.820},
        {692.345,0.815},
        {692.675,0.811},
        {693.005,0.806},
        {693.335,0.801},
        {693.665,0.797},
        {693.995,0.792},
        {694.325,0.788},
        {694.655,0.782},
        {694.985,0.775},
        {695.315,0.768},
        {695.645,0.761},
        {695.975,0.754},
        {696.305,0.747},
        {696.635,0.740},
        {696.965,0.734},
        {697.295,0.728},
        {697.625,0.722},
        {697.955,0.715},
        {698.285,0.709},
        {698.615,0.700},
        {698.945,0.688},
        {699.275,0.675},
        {699.605,0.663},
        {699.935,0.650},
        {700.265,0.637},
        {700.595,0.625},
        {700.925,0.611},
        {701.255,0.597},
        {701.585,0.583},
        {701.915,0.569},
        {702.245,0.555},
        {702.575,0.542},
        {702.905,0.529},
        {703.235,0.517},
        {703.565,0.504},
        {703.895,0.491},
        {704.225,0.479},
        {704.555,0.466},
        {704.885,0.452},
        {705.215,0.438},
        {705.545,0.425},
        {705.875,0.411},
        {706.205,0.397},
        {706.535,0.383},
        {706.865,0.359},
        {707.195,0.335},
        {707.525,0.311},
        {707.855,0.304},
        {708.185,0.297},
        {708.515,0.290},
        {708.845,0.283},
        {709.175,0.276},
        {709.505,0.269},
        {709.835,0.253},
        {710.165,0.235},
        {710.495,0.218},
        {710.825,0.210},
        {711.155,0.205},
        {711.485,0.200},
        {711.815,0.195},
        {712.145,0.190},
        {712.475,0.185},
        {712.805,0.177},
        {713.135,0.166},
        {713.465,0.156},
        {713.795,0.151},
        {714.125,0.147},
        {714.455,0.144},
        {714.785,0.141},
        {715.115,0.137},
        {715.445,0.134},
        {715.775,0.130},
        {716.105,0.127},
        {716.435,0.123},
        {716.765,0.120},
        {717.095,0.116},
        {717.425,0.113},
        {717.755,0.108},
        {718.085,0.101},
        {718.415,0.094},
        {718.745,0.087},
        {719.075,0.080},
        {719.405,0.073},
        {719.735,0.068},
        {720.065,0.065},
        {720.395,0.063},
        {720.725,0.060},
        {721.055,0.058},
        {721.385,0.055},
        {721.715,0.053},
        {722.045,0.050},
        {722.375,0.048},
        {722.705,0.045},
        {723.035,0.043},
        {723.365,0.040},
        {723.695,0.038},
        {724.025,0.037},
        {724.355,0.035},
        {724.685,0.034},
        {725.015,0.032},
        {725.345,0.031},
        {725.675,0.029},
        {726.005,0.028},
        {726.335,0.026},
        {726.665,0.025},
        {726.995,0.023},
        {727.325,0.022},
        {727.655,0.020},
        {727.985,0.019},
        {728.315,0.017},
        {728.645,0.016},
        {728.975,0.014},
        {729.305,0.013},
        {729.635,0.012},
        {729.965,0.011},
        {730.295,0.010},
        {730.625,0.009},
        {730.955,0.008},
        {731.285,0.007},
        {731.615,0.007},
        {731.945,0.006},
        {732.275,0.005},
        {732.605,0.004},
        {732.935,0.004},
        {733.265,0.003},
        {733.595,0.003},
        {733.925,0.003},
        {734.255,0.003},
        {734.585,0.003},
        {734.915,0.002},
        {735.245,0.002},
        {735.575,0.002},
        {735.905,0.002},
        {736.235,0.002},
        {736.565,0.001},
        {736.895,0.001},
        {737.225,0.001},
        {737.555,0.001},
        {737.885,0.001}
      };

      static const double CBERS2B_CCDB4_SRF[][2] = 
      {  
        {744.815,0.001},
        {745.145,0.001},
        {745.475,0.001},
        {745.805,0.001},
        {746.135,0.001},
        {746.465,0.001},
        {746.795,0.002},
        {747.125,0.002},
        {747.455,0.002},
        {747.785,0.002},
        {748.115,0.002},
        {748.445,0.003},
        {748.775,0.003},
        {749.105,0.003},
        {749.435,0.003},
        {749.765,0.003},
        {750.095,0.003},
        {750.425,0.004},
        {750.755,0.004},
        {751.085,0.004},
        {751.415,0.004},
        {751.745,0.005},
        {752.075,0.005},
        {752.405,0.005},
        {752.735,0.005},
        {753.065,0.006},
        {753.395,0.006},
        {753.725,0.006},
        {754.055,0.006},
        {754.385,0.007},
        {754.715,0.007},
        {755.045,0.007},
        {755.375,0.007},
        {755.705,0.008},
        {756.035,0.008},
        {756.365,0.008},
        {756.695,0.008},
        {757.025,0.008},
        {757.355,0.008},
        {757.685,0.008},
        {758.015,0.009},
        {758.345,0.009},
        {758.675,0.009},
        {759.005,0.009},
        {759.335,0.009},
        {759.665,0.009},
        {759.995,0.009},
        {760.325,0.010},
        {760.655,0.010},
        {760.985,0.010},
        {761.315,0.010},
        {761.645,0.010},
        {761.975,0.010},
        {762.305,0.010},
        {762.635,0.011},
        {762.965,0.011},
        {763.295,0.011},
        {763.625,0.011},
        {763.955,0.011},
        {764.285,0.011},
        {764.615,0.011},
        {764.945,0.012},
        {765.275,0.012},
        {765.605,0.013},
        {765.935,0.014},
        {766.265,0.014},
        {766.595,0.015},
        {766.925,0.015},
        {767.255,0.016},
        {767.585,0.017},
        {767.915,0.017},
        {768.245,0.018},
        {768.575,0.019},
        {768.905,0.020},
        {769.235,0.022},
        {769.565,0.024},
        {769.895,0.027},
        {770.225,0.029},
        {770.555,0.031},
        {770.885,0.033},
        {771.215,0.036},
        {771.545,0.038},
        {771.875,0.040},
        {772.205,0.043},
        {772.535,0.045},
        {772.865,0.047},
        {773.195,0.050},
        {773.525,0.052},
        {773.855,0.054},
        {774.185,0.056},
        {774.515,0.059},
        {774.845,0.061},
        {775.175,0.065},
        {775.505,0.069},
        {775.835,0.072},
        {776.165,0.076},
        {776.495,0.080},
        {776.825,0.084},
        {777.155,0.087},
        {777.485,0.091},
        {777.815,0.095},
        {778.145,0.099},
        {778.475,0.102},
        {778.805,0.106},
        {779.135,0.112},
        {779.465,0.118},
        {779.795,0.124},
        {780.125,0.132},
        {780.455,0.141},
        {780.785,0.150},
        {781.115,0.162},
        {781.445,0.175},
        {781.775,0.188},
        {782.105,0.201},
        {782.435,0.215},
        {782.765,0.228},
        {783.095,0.241},
        {783.425,0.253},
        {783.755,0.266},
        {784.085,0.278},
        {784.415,0.291},
        {784.745,0.304},
        {785.075,0.314},
        {785.405,0.322},
        {785.735,0.330},
        {786.065,0.338},
        {786.395,0.346},
        {786.725,0.354},
        {787.055,0.362},
        {787.385,0.370},
        {787.715,0.378},
        {788.045,0.388},
        {788.375,0.401},
        {788.705,0.415},
        {789.035,0.429},
        {789.365,0.443},
        {789.695,0.457},
        {790.025,0.469},
        {790.355,0.475},
        {790.685,0.482},
        {791.015,0.488},
        {791.345,0.494},
        {791.675,0.500},
        {792.005,0.507},
        {792.335,0.513},
        {792.665,0.519},
        {792.995,0.526},
        {793.325,0.532},
        {793.655,0.538},
        {793.985,0.544},
        {794.315,0.551},
        {794.645,0.558},
        {794.975,0.565},
        {795.305,0.572},
        {795.635,0.579},
        {795.965,0.586},
        {796.295,0.593},
        {796.625,0.600},
        {796.955,0.607},
        {797.285,0.614},
        {797.615,0.621},
        {797.945,0.627},
        {798.275,0.635},
        {798.605,0.642},
        {798.935,0.650},
        {799.265,0.656},
        {799.595,0.661},
        {799.925,0.666},
        {800.255,0.671},
        {800.585,0.676},
        {800.915,0.681},
        {801.245,0.682},
        {801.575,0.681},
        {801.905,0.680},
        {802.235,0.680},
        {802.565,0.679},
        {802.895,0.678},
        {803.225,0.678},
        {803.555,0.678},
        {803.885,0.678},
        {804.215,0.677},
        {804.545,0.677},
        {804.875,0.677},
        {805.205,0.677},
        {805.535,0.677},
        {805.865,0.676},
        {806.195,0.676},
        {806.525,0.676},
        {806.855,0.676},
        {807.185,0.676},
        {807.515,0.675},
        {807.845,0.675},
        {808.175,0.675},
        {808.505,0.675},
        {808.835,0.675},
        {809.165,0.675},
        {809.495,0.676},
        {809.825,0.678},
        {810.155,0.679},
        {810.485,0.681},
        {810.815,0.682},
        {811.145,0.684},
        {811.475,0.685},
        {811.805,0.687},
        {812.135,0.688},
        {812.465,0.690},
        {812.795,0.691},
        {813.125,0.693},
        {813.455,0.694},
        {813.785,0.695},
        {814.115,0.697},
        {814.445,0.698},
        {814.775,0.700},
        {815.105,0.701},
        {815.435,0.703},
        {815.765,0.704},
        {816.095,0.705},
        {816.425,0.706},
        {816.755,0.708},
        {817.085,0.709},
        {817.415,0.710},
        {817.745,0.711},
        {818.075,0.713},
        {818.405,0.714},
        {818.735,0.715},
        {819.065,0.716},
        {819.395,0.718},
        {819.725,0.719},
        {820.055,0.720},
        {820.385,0.722},
        {820.715,0.724},
        {821.045,0.726},
        {821.375,0.728},
        {821.705,0.730},
        {822.035,0.732},
        {822.365,0.734},
        {822.695,0.736},
        {823.025,0.738},
        {823.355,0.740},
        {823.685,0.742},
        {824.015,0.744},
        {824.345,0.746},
        {824.675,0.748},
        {825.005,0.750},
        {825.335,0.753},
        {825.665,0.755},
        {825.995,0.758},
        {826.325,0.760},
        {826.655,0.763},
        {826.985,0.765},
        {827.315,0.768},
        {827.645,0.770},
        {827.975,0.773},
        {828.305,0.775},
        {828.635,0.778},
        {828.965,0.780},
        {829.295,0.782},
        {829.625,0.782},
        {829.955,0.783},
        {830.285,0.783},
        {830.615,0.784},
        {830.945,0.784},
        {831.275,0.784},
        {831.605,0.785},
        {831.935,0.785},
        {832.265,0.786},
        {832.595,0.786},
        {832.925,0.786},
        {833.255,0.787},
        {833.585,0.787},
        {833.915,0.787},
        {834.245,0.788},
        {834.575,0.788},
        {834.905,0.789},
        {835.235,0.789},
        {835.565,0.789},
        {835.895,0.790},
        {836.225,0.790},
        {836.555,0.790},
        {836.885,0.791},
        {837.215,0.791},
        {837.545,0.791},
        {837.875,0.792},
        {838.205,0.792},
        {838.535,0.792},
        {838.865,0.793},
        {839.195,0.793},
        {839.525,0.792},
        {839.855,0.792},
        {840.185,0.791},
        {840.515,0.790},
        {840.845,0.789},
        {841.175,0.789},
        {841.505,0.788},
        {841.835,0.787},
        {842.165,0.786},
        {842.495,0.786},
        {842.825,0.785},
        {843.155,0.784},
        {843.485,0.783},
        {843.815,0.783},
        {844.145,0.782},
        {844.475,0.781},
        {844.805,0.780},
        {845.135,0.779},
        {845.465,0.778},
        {845.795,0.777},
        {846.125,0.776},
        {846.455,0.775},
        {846.785,0.774},
        {847.115,0.772},
        {847.445,0.771},
        {847.775,0.770},
        {848.105,0.769},
        {848.435,0.768},
        {848.765,0.767},
        {849.095,0.766},
        {849.425,0.765},
        {849.755,0.764},
        {850.085,0.763},
        {850.415,0.762},
        {850.745,0.761},
        {851.075,0.759},
        {851.405,0.758},
        {851.735,0.757},
        {852.065,0.756},
        {852.395,0.754},
        {852.725,0.753},
        {853.055,0.752},
        {853.385,0.751},
        {853.715,0.749},
        {854.045,0.748},
        {854.375,0.747},
        {854.705,0.746},
        {855.035,0.744},
        {855.365,0.743},
        {855.695,0.741},
        {856.025,0.739},
        {856.355,0.737},
        {856.685,0.735},
        {857.015,0.733},
        {857.345,0.732},
        {857.675,0.730},
        {858.005,0.728},
        {858.335,0.726},
        {858.665,0.724},
        {858.995,0.722},
        {859.325,0.720},
        {859.655,0.721},
        {859.985,0.722},
        {860.315,0.723},
        {860.645,0.724},
        {860.975,0.725},
        {861.305,0.725},
        {861.635,0.726},
        {861.965,0.727},
        {862.295,0.728},
        {862.625,0.729},
        {862.955,0.730},
        {863.285,0.731},
        {863.615,0.731},
        {863.945,0.732},
        {864.275,0.733},
        {864.605,0.734},
        {864.935,0.735},
        {865.265,0.736},
        {865.595,0.736},
        {865.925,0.737},
        {866.255,0.737},
        {866.585,0.738},
        {866.915,0.738},
        {867.245,0.739},
        {867.575,0.739},
        {867.905,0.740},
        {868.235,0.740},
        {868.565,0.741},
        {868.895,0.741},
        {869.225,0.742},
        {869.555,0.742},
        {869.885,0.743},
        {870.215,0.743},
        {870.545,0.745},
        {870.875,0.747},
        {871.205,0.749},
        {871.535,0.751},
        {871.865,0.753},
        {872.195,0.755},
        {872.525,0.757},
        {872.855,0.759},
        {873.185,0.761},
        {873.515,0.764},
        {873.845,0.766},
        {874.175,0.769},
        {874.505,0.771},
        {874.835,0.774},
        {875.165,0.776},
        {875.495,0.779},
        {875.825,0.781},
        {876.155,0.784},
        {876.485,0.786},
        {876.815,0.789},
        {877.145,0.792},
        {877.475,0.795},
        {877.805,0.798},
        {878.135,0.801},
        {878.465,0.804},
        {878.795,0.807},
        {879.125,0.810},
        {879.455,0.811},
        {879.785,0.802},
        {880.115,0.793},
        {880.445,0.785},
        {880.775,0.778},
        {881.105,0.771},
        {881.435,0.765},
        {881.765,0.758},
        {882.095,0.752},
        {882.425,0.745},
        {882.755,0.738},
        {883.085,0.732},
        {883.415,0.725},
        {883.745,0.719},
        {884.075,0.712},
        {884.405,0.705},
        {884.735,0.697},
        {885.065,0.689},
        {885.395,0.680},
        {885.725,0.672},
        {886.055,0.664},
        {886.385,0.655},
        {886.715,0.647},
        {887.045,0.638},
        {887.375,0.630},
        {887.705,0.621},
        {888.035,0.611},
        {888.365,0.601},
        {888.695,0.592},
        {889.025,0.582},
        {889.355,0.572},
        {889.685,0.563},
        {890.015,0.553},
        {890.345,0.543},
        {890.675,0.532},
        {891.005,0.520},
        {891.335,0.508},
        {891.665,0.496},
        {891.995,0.484},
        {892.325,0.472},
        {892.655,0.462},
        {892.985,0.454},
        {893.315,0.445},
        {893.645,0.436},
        {893.975,0.427},
        {894.305,0.418},
        {894.635,0.409},
        {894.965,0.401},
        {895.295,0.392},
        {895.625,0.376},
        {895.955,0.351},
        {896.285,0.325},
        {896.615,0.306},
        {896.945,0.294},
        {897.275,0.282},
        {897.605,0.270},
        {897.935,0.258},
        {898.265,0.246},
        {898.595,0.236},
        {898.925,0.229},
        {899.255,0.223},
        {899.585,0.216},
        {899.915,0.209},
        {900.245,0.201},
        {900.575,0.194},
        {900.905,0.186},
        {901.235,0.179},
        {901.565,0.172},
        {901.895,0.169},
        {902.225,0.166},
        {902.555,0.162},
        {902.885,0.159},
        {903.215,0.156},
        {903.545,0.153},
        {903.875,0.148},
        {904.205,0.143},
        {904.535,0.138},
        {904.865,0.133},
        {905.195,0.128},
        {905.525,0.123},
        {905.855,0.119},
        {906.185,0.116},
        {906.515,0.112},
        {906.845,0.109},
        {907.175,0.105},
        {907.505,0.102},
        {907.835,0.098},
        {908.165,0.095},
        {908.495,0.092},
        {908.825,0.088},
        {909.155,0.085},
        {909.485,0.081},
        {909.815,0.079},
        {910.145,0.078},
        {910.475,0.076},
        {910.805,0.075},
        {911.135,0.073},
        {911.465,0.072},
        {911.795,0.070},
        {912.125,0.069},
        {912.455,0.067},
        {912.785,0.066},
        {913.115,0.064},
        {913.445,0.063},
        {913.775,0.061},
        {914.105,0.060},
        {914.435,0.058},
        {914.765,0.056},
        {915.095,0.054},
        {915.425,0.052},
        {915.755,0.050},
        {916.085,0.048},
        {916.415,0.046},
        {916.745,0.044},
        {917.075,0.042},
        {917.405,0.040},
        {917.735,0.038},
        {918.065,0.036},
        {918.395,0.034},
        {918.725,0.032},
        {919.055,0.030},
        {919.385,0.028},
        {919.715,0.027},
        {920.045,0.026},
        {920.375,0.026},
        {920.705,0.025},
        {921.035,0.024},
        {921.365,0.024},
        {921.695,0.023},
        {922.025,0.023},
        {922.355,0.022},
        {922.685,0.022},
        {923.015,0.021},
        {923.345,0.021},
        {923.675,0.020},
        {924.005,0.020},
        {924.335,0.019},
        {924.665,0.019},
        {924.995,0.018},
        {925.325,0.017},
        {925.655,0.017},
        {925.985,0.016},
        {926.315,0.016},
        {926.645,0.015},
        {926.975,0.015},
        {927.305,0.015},
        {927.635,0.014},
        {927.965,0.014},
        {928.295,0.014},
        {928.625,0.014},
        {928.955,0.013},
        {929.285,0.013},
        {929.615,0.013},
        {929.945,0.013},
        {930.275,0.012},
        {930.605,0.012},
        {930.935,0.012},
        {931.265,0.011},
        {931.595,0.011},
        {931.925,0.011},
        {932.255,0.011},
        {932.585,0.010},
        {932.915,0.010},
        {933.245,0.010},
        {933.575,0.009},
        {933.905,0.009},
        {934.235,0.009},
        {934.565,0.009},
        {934.895,0.008},
        {935.225,0.008},
        {935.555,0.008},
        {935.885,0.008},
        {936.215,0.007},
        {936.545,0.007},
        {936.875,0.007},
        {937.205,0.007},
        {937.535,0.006},
        {937.865,0.006},
        {938.195,0.006},
        {938.525,0.006},
        {938.855,0.005},
        {939.185,0.005},
        {939.515,0.005},
        {939.845,0.005},
        {940.175,0.005},
        {940.505,0.004},
        {940.835,0.004},
        {941.165,0.004},
        {941.495,0.004},
        {941.825,0.003},
        {942.155,0.003},
        {942.485,0.003},
        {942.815,0.003},
        {943.145,0.002},
        {943.475,0.002},
        {943.805,0.002},
        {944.135,0.002},
        {944.465,0.002},
        {944.795,0.001},
        {945.125,0.001},
        {945.455,0.001},
        {945.785,0.001}
      };

      static const double CBERS2B_CCDB5PAN_SRF[][2] = 
      {  
        {462.005,0.001},
        {462.335,0.001},
        {462.665,0.001},
        {462.995,0.002},
        {463.325,0.002},
        {463.655,0.003},
        {463.985,0.003},
        {464.315,0.003},
        {464.645,0.004},
        {464.975,0.004},
        {465.305,0.004},
        {465.635,0.005},
        {465.965,0.005},
        {466.295,0.005},
        {466.625,0.006},
        {466.955,0.006},
        {467.285,0.007},
        {467.615,0.007},
        {467.945,0.007},
        {468.275,0.008},
        {468.605,0.008},
        {468.935,0.007},
        {469.265,0.007},
        {469.595,0.007},
        {469.925,0.007},
        {470.255,0.007},
        {470.585,0.006},
        {470.915,0.006},
        {471.245,0.006},
        {471.575,0.006},
        {471.905,0.006},
        {472.235,0.006},
        {472.565,0.005},
        {472.895,0.005},
        {473.225,0.005},
        {473.555,0.005},
        {473.885,0.005},
        {474.215,0.004},
        {474.545,0.004},
        {474.875,0.004},
        {475.205,0.004},
        {475.535,0.004},
        {475.865,0.004},
        {476.195,0.004},
        {476.525,0.004},
        {476.855,0.004},
        {477.185,0.004},
        {477.515,0.004},
        {477.845,0.004},
        {478.175,0.004},
        {478.505,0.004},
        {478.835,0.004},
        {479.165,0.004},
        {479.495,0.004},
        {479.825,0.004},
        {480.155,0.004},
        {480.485,0.004},
        {480.815,0.004},
        {481.145,0.004},
        {481.475,0.004},
        {481.805,0.004},
        {482.135,0.004},
        {482.465,0.004},
        {482.795,0.004},
        {483.125,0.004},
        {483.455,0.004},
        {483.785,0.004},
        {484.115,0.004},
        {484.445,0.004},
        {484.775,0.004},
        {485.105,0.004},
        {485.435,0.004},
        {485.765,0.004},
        {486.095,0.004},
        {486.425,0.004},
        {486.755,0.004},
        {487.085,0.004},
        {487.415,0.004},
        {487.745,0.004},
        {488.075,0.004},
        {488.405,0.004},
        {488.735,0.004},
        {489.065,0.004},
        {489.395,0.004},
        {489.725,0.004},
        {490.055,0.004},
        {490.385,0.004},
        {490.715,0.004},
        {491.045,0.004},
        {491.375,0.004},
        {491.705,0.004},
        {492.035,0.005},
        {492.365,0.005},
        {492.695,0.005},
        {493.025,0.005},
        {493.355,0.005},
        {493.685,0.005},
        {494.015,0.005},
        {494.345,0.005},
        {494.675,0.005},
        {495.005,0.005},
        {495.335,0.005},
        {495.665,0.005},
        {495.995,0.005},
        {496.325,0.005},
        {496.655,0.005},
        {496.985,0.005},
        {497.315,0.005},
        {497.645,0.005},
        {497.975,0.005},
        {498.305,0.005},
        {498.635,0.006},
        {498.965,0.006},
        {499.295,0.006},
        {499.625,0.006},
        {499.955,0.006},
        {500.285,0.006},
        {500.615,0.006},
        {500.945,0.006},
        {501.275,0.006},
        {501.605,0.006},
        {501.935,0.006},
        {502.265,0.006},
        {502.595,0.006},
        {502.925,0.006},
        {503.255,0.006},
        {503.585,0.006},
        {503.915,0.006},
        {504.245,0.006},
        {504.575,0.006},
        {504.905,0.006},
        {505.235,0.007},
        {505.565,0.007},
        {505.895,0.007},
        {506.225,0.007},
        {506.555,0.007},
        {506.885,0.007},
        {507.215,0.007},
        {507.545,0.007},
        {507.875,0.007},
        {508.205,0.007},
        {508.535,0.007},
        {508.865,0.007},
        {509.195,0.007},
        {509.525,0.007},
        {509.855,0.007},
        {510.185,0.007},
        {510.515,0.007},
        {510.845,0.007},
        {511.175,0.007},
        {511.505,0.007},
        {511.835,0.008},
        {512.165,0.008},
        {512.495,0.008},
        {512.825,0.008},
        {513.155,0.008},
        {513.485,0.009},
        {513.815,0.009},
        {514.145,0.010},
        {514.475,0.010},
        {514.805,0.011},
        {515.135,0.011},
        {515.465,0.011},
        {515.795,0.012},
        {516.125,0.012},
        {516.455,0.013},
        {516.785,0.013},
        {517.115,0.013},
        {517.445,0.014},
        {517.775,0.014},
        {518.105,0.015},
        {518.435,0.015},
        {518.765,0.017},
        {519.095,0.021},
        {519.425,0.025},
        {519.755,0.029},
        {520.085,0.032},
        {520.415,0.036},
        {520.745,0.040},
        {521.075,0.044},
        {521.405,0.047},
        {521.735,0.051},
        {522.065,0.055},
        {522.395,0.059},
        {522.725,0.063},
        {523.055,0.066},
        {523.385,0.070},
        {523.715,0.074},
        {524.045,0.078},
        {524.375,0.083},
        {524.705,0.087},
        {525.035,0.091},
        {525.365,0.095},
        {525.695,0.100},
        {526.025,0.104},
        {526.355,0.108},
        {526.685,0.113},
        {527.015,0.117},
        {527.345,0.121},
        {527.675,0.125},
        {528.005,0.130},
        {528.335,0.134},
        {528.665,0.138},
        {528.995,0.143},
        {529.325,0.148},
        {529.655,0.154},
        {529.985,0.166},
        {530.315,0.179},
        {530.645,0.191},
        {530.975,0.204},
        {531.305,0.216},
        {531.635,0.229},
        {531.965,0.242},
        {532.295,0.255},
        {532.625,0.269},
        {532.955,0.282},
        {533.285,0.295},
        {533.615,0.308},
        {533.945,0.318},
        {534.275,0.328},
        {534.605,0.338},
        {534.935,0.347},
        {535.265,0.357},
        {535.595,0.367},
        {535.925,0.376},
        {536.255,0.386},
        {536.585,0.396},
        {536.915,0.414},
        {537.245,0.437},
        {537.575,0.460},
        {537.905,0.472},
        {538.235,0.478},
        {538.565,0.484},
        {538.895,0.490},
        {539.225,0.497},
        {539.555,0.503},
        {539.885,0.508},
        {540.215,0.512},
        {540.545,0.515},
        {540.875,0.519},
        {541.205,0.523},
        {541.535,0.527},
        {541.865,0.531},
        {542.195,0.534},
        {542.525,0.538},
        {542.855,0.542},
        {543.185,0.544},
        {543.515,0.547},
        {543.845,0.550},
        {544.175,0.553},
        {544.505,0.556},
        {544.835,0.559},
        {545.165,0.562},
        {545.495,0.565},
        {545.825,0.568},
        {546.155,0.571},
        {546.485,0.574},
        {546.815,0.577},
        {547.145,0.580},
        {547.475,0.583},
        {547.805,0.586},
        {548.135,0.589},
        {548.465,0.591},
        {548.795,0.595},
        {549.125,0.598},
        {549.455,0.602},
        {549.785,0.606},
        {550.115,0.610},
        {550.445,0.613},
        {550.775,0.617},
        {551.105,0.621},
        {551.435,0.625},
        {551.765,0.628},
        {552.095,0.632},
        {552.425,0.636},
        {552.755,0.639},
        {553.085,0.643},
        {553.415,0.646},
        {553.745,0.650},
        {554.075,0.654},
        {554.405,0.657},
        {554.735,0.661},
        {555.065,0.664},
        {555.395,0.668},
        {555.725,0.672},
        {556.055,0.675},
        {556.385,0.679},
        {556.715,0.682},
        {557.045,0.686},
        {557.375,0.690},
        {557.705,0.693},
        {558.035,0.697},
        {558.365,0.700},
        {558.695,0.704},
        {559.025,0.706},
        {559.355,0.706},
        {559.685,0.707},
        {560.015,0.708},
        {560.345,0.709},
        {560.675,0.709},
        {561.005,0.710},
        {561.335,0.711},
        {561.665,0.712},
        {561.995,0.712},
        {562.325,0.713},
        {562.655,0.714},
        {562.985,0.715},
        {563.315,0.715},
        {563.645,0.716},
        {563.975,0.717},
        {564.305,0.719},
        {564.635,0.720},
        {564.965,0.722},
        {565.295,0.723},
        {565.625,0.725},
        {565.955,0.726},
        {566.285,0.728},
        {566.615,0.729},
        {566.945,0.731},
        {567.275,0.732},
        {567.605,0.734},
        {567.935,0.735},
        {568.265,0.737},
        {568.595,0.738},
        {568.925,0.739},
        {569.255,0.738},
        {569.585,0.736},
        {569.915,0.735},
        {570.245,0.733},
        {570.575,0.732},
        {570.905,0.730},
        {571.235,0.729},
        {571.565,0.727},
        {571.895,0.726},
        {572.225,0.724},
        {572.555,0.723},
        {572.885,0.721},
        {573.215,0.720},
        {573.545,0.718},
        {573.875,0.717},
        {574.205,0.715},
        {574.535,0.713},
        {574.865,0.712},
        {575.195,0.710},
        {575.525,0.709},
        {575.855,0.707},
        {576.185,0.706},
        {576.515,0.704},
        {576.845,0.703},
        {577.175,0.701},
        {577.505,0.700},
        {577.835,0.698},
        {578.165,0.697},
        {578.495,0.695},
        {578.825,0.694},
        {579.155,0.692},
        {579.485,0.691},
        {579.815,0.689},
        {580.145,0.687},
        {580.475,0.686},
        {580.805,0.684},
        {581.135,0.682},
        {581.465,0.681},
        {581.795,0.679},
        {582.125,0.678},
        {582.455,0.676},
        {582.785,0.674},
        {583.115,0.673},
        {583.445,0.671},
        {583.775,0.669},
        {584.105,0.668},
        {584.435,0.666},
        {584.765,0.664},
        {585.095,0.663},
        {585.425,0.661},
        {585.755,0.660},
        {586.085,0.658},
        {586.415,0.656},
        {586.745,0.655},
        {587.075,0.653},
        {587.405,0.651},
        {587.735,0.650},
        {588.065,0.648},
        {588.395,0.646},
        {588.725,0.645},
        {589.055,0.643},
        {589.385,0.642},
        {589.715,0.641},
        {590.045,0.640},
        {590.375,0.640},
        {590.705,0.641},
        {591.035,0.641},
        {591.365,0.641},
        {591.695,0.642},
        {592.025,0.642},
        {592.355,0.642},
        {592.685,0.643},
        {593.015,0.643},
        {593.345,0.643},
        {593.675,0.643},
        {594.005,0.644},
        {594.335,0.644},
        {594.665,0.644},
        {594.995,0.645},
        {595.325,0.645},
        {595.655,0.645},
        {595.985,0.645},
        {596.315,0.646},
        {596.645,0.646},
        {596.975,0.646},
        {597.305,0.646},
        {597.635,0.647},
        {597.965,0.647},
        {598.295,0.647},
        {598.625,0.647},
        {598.955,0.648},
        {599.285,0.650},
        {599.615,0.652},
        {599.945,0.654},
        {600.275,0.657},
        {600.605,0.659},
        {600.935,0.661},
        {601.265,0.663},
        {601.595,0.666},
        {601.925,0.668},
        {602.255,0.670},
        {602.585,0.673},
        {602.915,0.675},
        {603.245,0.677},
        {603.575,0.680},
        {603.905,0.682},
        {604.235,0.684},
        {604.565,0.687},
        {604.895,0.689},
        {605.225,0.691},
        {605.555,0.694},
        {605.885,0.697},
        {606.215,0.700},
        {606.545,0.702},
        {606.875,0.705},
        {607.205,0.708},
        {607.535,0.711},
        {607.865,0.714},
        {608.195,0.716},
        {608.525,0.719},
        {608.855,0.722},
        {609.185,0.725},
        {609.515,0.727},
        {609.845,0.730},
        {610.175,0.733},
        {610.505,0.734},
        {610.835,0.736},
        {611.165,0.738},
        {611.495,0.740},
        {611.825,0.741},
        {612.155,0.743},
        {612.485,0.745},
        {612.815,0.747},
        {613.145,0.748},
        {613.475,0.750},
        {613.805,0.752},
        {614.135,0.754},
        {614.465,0.755},
        {614.795,0.757},
        {615.125,0.759},
        {615.455,0.762},
        {615.785,0.764},
        {616.115,0.767},
        {616.445,0.769},
        {616.775,0.772},
        {617.105,0.774},
        {617.435,0.777},
        {617.765,0.779},
        {618.095,0.782},
        {618.425,0.784},
        {618.755,0.787},
        {619.085,0.789},
        {619.415,0.789},
        {619.745,0.788},
        {620.075,0.787},
        {620.405,0.786},
        {620.735,0.786},
        {621.065,0.785},
        {621.395,0.784},
        {621.725,0.783},
        {622.055,0.783},
        {622.385,0.782},
        {622.715,0.781},
        {623.045,0.780},
        {623.375,0.780},
        {623.705,0.779},
        {624.035,0.778},
        {624.365,0.777},
        {624.695,0.777},
        {625.025,0.776},
        {625.355,0.775},
        {625.685,0.774},
        {626.015,0.774},
        {626.345,0.773},
        {626.675,0.772},
        {627.005,0.771},
        {627.335,0.771},
        {627.665,0.770},
        {627.995,0.769},
        {628.325,0.768},
        {628.655,0.767},
        {628.985,0.767},
        {629.315,0.767},
        {629.645,0.767},
        {629.975,0.768},
        {630.305,0.769},
        {630.635,0.769},
        {630.965,0.770},
        {631.295,0.770},
        {631.625,0.771},
        {631.955,0.772},
        {632.285,0.772},
        {632.615,0.773},
        {632.945,0.774},
        {633.275,0.774},
        {633.605,0.775},
        {633.935,0.775},
        {634.265,0.776},
        {634.595,0.776},
        {634.925,0.777},
        {635.255,0.777},
        {635.585,0.778},
        {635.915,0.778},
        {636.245,0.779},
        {636.575,0.779},
        {636.905,0.780},
        {637.235,0.780},
        {637.565,0.781},
        {637.895,0.781},
        {638.225,0.782},
        {638.555,0.782},
        {638.885,0.783},
        {639.215,0.783},
        {639.545,0.783},
        {639.875,0.784},
        {640.205,0.784},
        {640.535,0.785},
        {640.865,0.785},
        {641.195,0.786},
        {641.525,0.787},
        {641.855,0.789},
        {642.185,0.790},
        {642.515,0.792},
        {642.845,0.793},
        {643.175,0.795},
        {643.505,0.796},
        {643.835,0.798},
        {644.165,0.800},
        {644.495,0.801},
        {644.825,0.803},
        {645.155,0.804},
        {645.485,0.806},
        {645.815,0.809},
        {646.145,0.811},
        {646.475,0.813},
        {646.805,0.815},
        {647.135,0.817},
        {647.465,0.820},
        {647.795,0.822},
        {648.125,0.824},
        {648.455,0.826},
        {648.785,0.828},
        {649.115,0.831},
        {649.445,0.831},
        {649.775,0.830},
        {650.105,0.829},
        {650.435,0.828},
        {650.765,0.827},
        {651.095,0.826},
        {651.425,0.825},
        {651.755,0.824},
        {652.085,0.822},
        {652.415,0.821},
        {652.745,0.820},
        {653.075,0.819},
        {653.405,0.818},
        {653.735,0.817},
        {654.065,0.816},
        {654.395,0.815},
        {654.725,0.814},
        {655.055,0.813},
        {655.385,0.812},
        {655.715,0.811},
        {656.045,0.810},
        {656.375,0.809},
        {656.705,0.808},
        {657.035,0.807},
        {657.365,0.806},
        {657.695,0.805},
        {658.025,0.804},
        {658.355,0.803},
        {658.685,0.802},
        {659.015,0.801},
        {659.345,0.800},
        {659.675,0.799},
        {660.005,0.798},
        {660.335,0.797},
        {660.665,0.793},
        {660.995,0.789},
        {661.325,0.785},
        {661.655,0.783},
        {661.985,0.780},
        {662.315,0.778},
        {662.645,0.775},
        {662.975,0.773},
        {663.305,0.770},
        {663.635,0.768},
        {663.965,0.765},
        {664.295,0.763},
        {664.625,0.760},
        {664.955,0.758},
        {665.285,0.755},
        {665.615,0.752},
        {665.945,0.748},
        {666.275,0.745},
        {666.605,0.741},
        {666.935,0.738},
        {667.265,0.735},
        {667.595,0.731},
        {667.925,0.728},
        {668.255,0.724},
        {668.585,0.721},
        {668.915,0.717},
        {669.245,0.714},
        {669.575,0.713},
        {669.905,0.712},
        {670.235,0.712},
        {670.565,0.712},
        {670.895,0.712},
        {671.225,0.711},
        {671.555,0.711},
        {671.885,0.711},
        {672.215,0.711},
        {672.545,0.710},
        {672.875,0.710},
        {673.205,0.710},
        {673.535,0.710},
        {673.865,0.709},
        {674.195,0.709},
        {674.525,0.709},
        {674.855,0.709},
        {675.185,0.708},
        {675.515,0.708},
        {675.845,0.708},
        {676.175,0.708},
        {676.505,0.707},
        {676.835,0.707},
        {677.165,0.707},
        {677.495,0.707},
        {677.825,0.706},
        {678.155,0.706},
        {678.485,0.706},
        {678.815,0.705},
        {679.145,0.705},
        {679.475,0.705},
        {679.805,0.706},
        {680.135,0.707},
        {680.465,0.707},
        {680.795,0.708},
        {681.125,0.709},
        {681.455,0.710},
        {681.785,0.710},
        {682.115,0.711},
        {682.445,0.712},
        {682.775,0.712},
        {683.105,0.713},
        {683.435,0.714},
        {683.765,0.715},
        {684.095,0.715},
        {684.425,0.716},
        {684.755,0.717},
        {685.085,0.717},
        {685.415,0.718},
        {685.745,0.719},
        {686.075,0.720},
        {686.405,0.720},
        {686.735,0.721},
        {687.065,0.722},
        {687.395,0.723},
        {687.725,0.723},
        {688.055,0.724},
        {688.385,0.725},
        {688.715,0.726},
        {689.045,0.727},
        {689.375,0.727},
        {689.705,0.728},
        {690.035,0.729},
        {690.365,0.730},
        {690.695,0.731},
        {691.025,0.731},
        {691.355,0.732},
        {691.685,0.733},
        {692.015,0.734},
        {692.345,0.734},
        {692.675,0.735},
        {693.005,0.736},
        {693.335,0.737},
        {693.665,0.738},
        {693.995,0.738},
        {694.325,0.739},
        {694.655,0.740},
        {694.985,0.741},
        {695.315,0.741},
        {695.645,0.742},
        {695.975,0.743},
        {696.305,0.744},
        {696.635,0.744},
        {696.965,0.745},
        {697.295,0.746},
        {697.625,0.747},
        {697.955,0.748},
        {698.285,0.748},
        {698.615,0.749},
        {698.945,0.750},
        {699.275,0.751},
        {699.605,0.751},
        {699.935,0.750},
        {700.265,0.749},
        {700.595,0.749},
        {700.925,0.748},
        {701.255,0.747},
        {701.585,0.746},
        {701.915,0.745},
        {702.245,0.745},
        {702.575,0.744},
        {702.905,0.743},
        {703.235,0.742},
        {703.565,0.742},
        {703.895,0.741},
        {704.225,0.740},
        {704.555,0.739},
        {704.885,0.739},
        {705.215,0.738},
        {705.545,0.737},
        {705.875,0.736},
        {706.205,0.736},
        {706.535,0.735},
        {706.865,0.734},
        {707.195,0.733},
        {707.525,0.733},
        {707.855,0.732},
        {708.185,0.731},
        {708.515,0.730},
        {708.845,0.730},
        {709.175,0.729},
        {709.505,0.728},
        {709.835,0.725},
        {710.165,0.721},
        {710.495,0.718},
        {710.825,0.714},
        {711.155,0.711},
        {711.485,0.707},
        {711.815,0.704},
        {712.145,0.700},
        {712.475,0.697},
        {712.805,0.693},
        {713.135,0.690},
        {713.465,0.686},
        {713.795,0.683},
        {714.125,0.679},
        {714.455,0.676},
        {714.785,0.672},
        {715.115,0.668},
        {715.445,0.664},
        {715.775,0.661},
        {716.105,0.657},
        {716.435,0.653},
        {716.765,0.649},
        {717.095,0.646},
        {717.425,0.642},
        {717.755,0.638},
        {718.085,0.634},
        {718.415,0.630},
        {718.745,0.627},
        {719.075,0.625},
        {719.405,0.622},
        {719.735,0.620},
        {720.065,0.617},
        {720.395,0.615},
        {720.725,0.609},
        {721.055,0.597},
        {721.385,0.585},
        {721.715,0.573},
        {722.045,0.561},
        {722.375,0.549},
        {722.705,0.537},
        {723.035,0.525},
        {723.365,0.513},
        {723.695,0.501},
        {724.025,0.489},
        {724.355,0.477},
        {724.685,0.465},
        {725.015,0.452},
        {725.345,0.439},
        {725.675,0.425},
        {726.005,0.412},
        {726.335,0.399},
        {726.665,0.386},
        {726.995,0.378},
        {727.325,0.369},
        {727.655,0.361},
        {727.985,0.353},
        {728.315,0.344},
        {728.645,0.336},
        {728.975,0.327},
        {729.305,0.319},
        {729.635,0.311},
        {729.965,0.298},
        {730.295,0.286},
        {730.625,0.273},
        {730.955,0.267},
        {731.285,0.262},
        {731.615,0.257},
        {731.945,0.252},
        {732.275,0.247},
        {732.605,0.242},
        {732.935,0.237},
        {733.265,0.232},
        {733.595,0.227},
        {733.925,0.222},
        {734.255,0.217},
        {734.585,0.212},
        {734.915,0.207},
        {735.245,0.203},
        {735.575,0.198},
        {735.905,0.193},
        {736.235,0.188},
        {736.565,0.184},
        {736.895,0.179},
        {737.225,0.174},
        {737.555,0.170},
        {737.885,0.165},
        {738.215,0.160},
        {738.545,0.155},
        {738.875,0.152},
        {739.205,0.148},
        {739.535,0.145},
        {739.865,0.142},
        {740.195,0.139},
        {740.525,0.136},
        {740.855,0.133},
        {741.185,0.130},
        {741.515,0.126},
        {741.845,0.123},
        {742.175,0.120},
        {742.505,0.117},
        {742.835,0.114},
        {743.165,0.111},
        {743.495,0.108},
        {743.825,0.104},
        {744.155,0.101},
        {744.485,0.098},
        {744.815,0.095},
        {745.145,0.092},
        {745.475,0.089},
        {745.805,0.086},
        {746.135,0.083},
        {746.465,0.080},
        {746.795,0.077},
        {747.125,0.074},
        {747.455,0.071},
        {747.785,0.068},
        {748.115,0.065},
        {748.445,0.062},
        {748.775,0.059},
        {749.105,0.056},
        {749.435,0.053},
        {749.765,0.050},
        {750.095,0.049},
        {750.425,0.048},
        {750.755,0.047},
        {751.085,0.045},
        {751.415,0.044},
        {751.745,0.043},
        {752.075,0.042},
        {752.405,0.041},
        {752.735,0.040},
        {753.065,0.039},
        {753.395,0.038},
        {753.725,0.037},
        {754.055,0.036},
        {754.385,0.035},
        {754.715,0.034},
        {755.045,0.033},
        {755.375,0.031},
        {755.705,0.030},
        {756.035,0.029},
        {756.365,0.028},
        {756.695,0.027},
        {757.025,0.026},
        {757.355,0.026},
        {757.685,0.025},
        {758.015,0.024},
        {758.345,0.024},
        {758.675,0.023},
        {759.005,0.023},
        {759.335,0.022},
        {759.665,0.021},
        {759.995,0.021},
        {760.325,0.020},
        {760.655,0.019},
        {760.985,0.019},
        {761.315,0.018},
        {761.645,0.018},
        {761.975,0.017},
        {762.305,0.016},
        {762.635,0.016},
        {762.965,0.015},
        {763.295,0.014},
        {763.625,0.014},
        {763.955,0.013},
        {764.285,0.013},
        {764.615,0.012},
        {764.945,0.011},
        {765.275,0.011},
        {765.605,0.011},
        {765.935,0.010},
        {766.265,0.010},
        {766.595,0.010},
        {766.925,0.009},
        {767.255,0.009},
        {767.585,0.008},
        {767.915,0.008},
        {768.245,0.008},
        {768.575,0.007},
        {768.905,0.007},
        {769.235,0.007},
        {769.565,0.006},
        {769.895,0.006},
        {770.225,0.006},
        {770.555,0.005},
        {770.885,0.005},
        {771.215,0.005},
        {771.545,0.004},
        {771.875,0.004},
        {772.205,0.004},
        {772.535,0.004},
        {772.865,0.004},
        {773.195,0.005},
        {773.525,0.005},
        {773.855,0.005},
        {774.185,0.005},
        {774.515,0.005},
        {774.845,0.005},
        {775.175,0.006},
        {775.505,0.006},
        {775.835,0.006},
        {776.165,0.006},
        {776.495,0.006},
        {776.825,0.006},
        {777.155,0.007},
        {777.485,0.007},
        {777.815,0.007},
        {778.145,0.007},
        {778.475,0.007},
        {778.805,0.007},
        {779.135,0.008},
        {779.465,0.008},
        {779.795,0.008},
        {780.125,0.008},
        {780.455,0.008},
        {780.785,0.008},
        {781.115,0.009},
        {781.445,0.009},
        {781.775,0.009},
        {782.105,0.009},
        {782.435,0.009},
        {782.765,0.009},
        {783.095,0.010},
        {783.425,0.010},
        {783.755,0.010},
        {784.085,0.010},
        {784.415,0.010},
        {784.745,0.011},
        {785.075,0.011},
        {785.405,0.011},
        {785.735,0.011},
        {786.065,0.011},
        {786.395,0.011},
        {786.725,0.012},
        {787.055,0.012},
        {787.385,0.012},
        {787.715,0.012},
        {788.045,0.012},
        {788.375,0.012},
        {788.705,0.013},
        {789.035,0.013},
        {789.365,0.013},
        {789.695,0.013},
        {790.025,0.013},
        {790.355,0.013},
        {790.685,0.014},
        {791.015,0.014},
        {791.345,0.014},
        {791.675,0.014},
        {792.005,0.014},
        {792.335,0.014},
        {792.665,0.015},
        {792.995,0.015},
        {793.325,0.015},
        {793.655,0.015},
        {793.985,0.015},
        {794.315,0.015},
        {794.645,0.015},
        {794.975,0.015},
        {795.305,0.015},
        {795.635,0.015},
        {795.965,0.015},
        {796.295,0.015},
        {796.625,0.015},
        {796.955,0.015},
        {797.285,0.015},
        {797.615,0.015},
        {797.945,0.014},
        {798.275,0.014},
        {798.605,0.014},
        {798.935,0.014},
        {799.265,0.014},
        {799.595,0.014},
        {799.925,0.014},
        {800.255,0.014},
        {800.585,0.014},
        {800.915,0.014},
        {801.245,0.014},
        {801.575,0.014},
        {801.905,0.014},
        {802.235,0.014},
        {802.565,0.014},
        {802.895,0.013},
        {803.225,0.013},
        {803.555,0.013},
        {803.885,0.013},
        {804.215,0.013},
        {804.545,0.013},
        {804.875,0.013},
        {805.205,0.013},
        {805.535,0.013},
        {805.865,0.013},
        {806.195,0.013},
        {806.525,0.013},
        {806.855,0.013},
        {807.185,0.013},
        {807.515,0.012},
        {807.845,0.012},
        {808.175,0.012},
        {808.505,0.012},
        {808.835,0.012},
        {809.165,0.012},
        {809.495,0.012},
        {809.825,0.012},
        {810.155,0.012},
        {810.485,0.012},
        {810.815,0.012},
        {811.145,0.012},
        {811.475,0.012},
        {811.805,0.012},
        {812.135,0.011},
        {812.465,0.011},
        {812.795,0.011},
        {813.125,0.011},
        {813.455,0.011},
        {813.785,0.011},
        {814.115,0.011},
        {814.445,0.011},
        {814.775,0.011},
        {815.105,0.011},
        {815.435,0.011},
        {815.765,0.010},
        {816.095,0.010},
        {816.425,0.010},
        {816.755,0.010},
        {817.085,0.010},
        {817.415,0.010},
        {817.745,0.010},
        {818.075,0.010},
        {818.405,0.010},
        {818.735,0.010},
        {819.065,0.009},
        {819.395,0.009},
        {819.725,0.009},
        {820.055,0.009},
        {820.385,0.009},
        {820.715,0.009},
        {821.045,0.009},
        {821.375,0.009},
        {821.705,0.009},
        {822.035,0.009},
        {822.365,0.008},
        {822.695,0.008},
        {823.025,0.008},
        {823.355,0.008},
        {823.685,0.008},
        {824.015,0.008},
        {824.345,0.008},
        {824.675,0.008},
        {825.005,0.008},
        {825.335,0.008},
        {825.665,0.008},
        {825.995,0.007},
        {826.325,0.007},
        {826.655,0.007},
        {826.985,0.007},
        {827.315,0.007},
        {827.645,0.007},
        {827.975,0.007},
        {828.305,0.007},
        {828.635,0.007},
        {828.965,0.007},
        {829.295,0.007},
        {829.625,0.007},
        {829.955,0.007},
        {830.285,0.007},
        {830.615,0.007},
        {830.945,0.006},
        {831.275,0.006},
        {831.605,0.006},
        {831.935,0.006},
        {832.265,0.006},
        {832.595,0.006},
        {832.925,0.006},
        {833.255,0.006},
        {833.585,0.006},
        {833.915,0.006},
        {834.245,0.006},
        {834.575,0.006},
        {834.905,0.006},
        {835.235,0.006},
        {835.565,0.005},
        {835.895,0.005},
        {836.225,0.005},
        {836.555,0.005},
        {836.885,0.005},
        {837.215,0.005},
        {837.545,0.005},
        {837.875,0.005},
        {838.205,0.005},
        {838.535,0.005},
        {838.865,0.005},
        {839.195,0.005},
        {839.525,0.005},
        {839.855,0.005},
        {840.185,0.004},
        {840.515,0.004},
        {840.845,0.004},
        {841.175,0.004},
        {841.505,0.004},
        {841.835,0.004},
        {842.165,0.004},
        {842.495,0.004},
        {842.825,0.004},
        {843.155,0.004},
        {843.485,0.004},
        {843.815,0.004},
        {844.145,0.004},
        {844.475,0.003},
        {844.805,0.003},
        {845.135,0.003},
        {845.465,0.003},
        {845.795,0.003},
        {846.125,0.003},
        {846.455,0.003},
        {846.785,0.003},
        {847.115,0.003},
        {847.445,0.002},
        {847.775,0.002},
        {848.105,0.002},
        {848.435,0.002},
        {848.765,0.002},
        {849.095,0.002},
        {849.425,0.002},
        {849.755,0.002},
        {850.085,0.002},
        {850.415,0.002},
        {850.745,0.001},
        {851.075,0.001},
        {851.405,0.001},
        {851.735,0.001},
        {852.065,0.001},
        {852.395,0.001},
        {852.725,0.001},
        {853.055,0.001},
        {853.385,0.001},
        {853.715,0.000},
        {854.045,0.000},
        {854.375,0.000},
        {854.705,0.000},
        {855.035,0.000},
        {855.365,0.000},
        {855.695,0.000},
        {856.025,0.000},
        {856.355,0.000},
        {856.685,0.001},
        {857.015,0.001},
        {857.345,0.001},
        {857.675,0.001},
        {858.005,0.001},
        {858.335,0.001},
        {858.665,0.001},
        {858.995,0.002},
        {859.325,0.002},
        {859.655,0.002},
        {859.985,0.002},
        {860.315,0.002},
        {860.645,0.002},
        {860.975,0.002},
        {861.305,0.003},
        {861.635,0.003},
        {861.965,0.003},
        {862.295,0.003},
        {862.625,0.003},
        {862.955,0.003},
        {863.285,0.003},
        {863.615,0.004},
        {863.945,0.004},
        {864.275,0.004},
        {864.605,0.004},
        {864.935,0.004},
        {865.265,0.003},
        {865.595,0.003},
        {865.925,0.003},
        {866.255,0.003},
        {866.585,0.003},
        {866.915,0.003},
        {867.245,0.003},
        {867.575,0.003},
        {867.905,0.002},
        {868.235,0.002},
        {868.565,0.002},
        {868.895,0.002},
        {869.225,0.002},
        {869.555,0.002},
        {869.885,0.002},
        {870.215,0.002},
        {870.545,0.001},
        {870.875,0.001},
        {871.205,0.001},
        {871.535,0.001},
        {871.865,0.001},
        {872.195,0.001},
        {872.525,0.001},
        {872.855,0.001}
      };
      
      static const double LANDSAT8_OLI1_SRF[][2] = 
      {
        {427,0.000073},
        {428,0.000609},
        {429,0.001628},
        {430,0.003421},
        {431,0.008019},
        {432,0.024767},
        {433,0.085688},
        {434,0.254149},
        {435,0.517821},
        {436,0.765117},
        {437,0.908749},
        {438,0.958204},
        {439,0.977393},
        {440,0.98379},
        {441,0.989052},
        {442,0.986713},
        {443,0.993683},
        {444,0.993137},
        {445,1},
        {446,0.996969},
        {447,0.98278},
        {448,0.972692},
        {449,0.905808},
        {450,0.745606},
        {451,0.471329},
        {452,0.226412},
        {453,0.09286},
        {454,0.036603},
        {455,0.014537},
        {456,0.005829},
        {457,0.002414},
        {458,0.000984},
        {459,0.000255}
      };

      static const double LANDSAT8_OLI2_SRF[][2] = 
      {
        {436,0.00001},
        {437,0.000061},
        {438,0.000117},
        {439,0.000241},
        {440,0.000349},
        {441,0.000455},
        {442,0.000756},
        {443,0.001197},
        {444,0.00207},
        {445,0.003712},
        {446,0.006869},
        {447,0.013212},
        {448,0.02717},
        {449,0.058606},
        {450,0.130876},
        {451,0.27137},
        {452,0.493542},
        {453,0.723971},
        {454,0.85751},
        {455,0.894222},
        {456,0.903034},
        {457,0.910928},
        {458,0.90988},
        {459,0.899475},
        {460,0.897977},
        {461,0.889667},
        {462,0.883542},
        {463,0.877453},
        {464,0.881011},
        {465,0.874721},
        {466,0.879688},
        {467,0.886569},
        {468,0.891913},
        {469,0.88768},
        {470,0.861157},
        {471,0.848533},
        {472,0.840828},
        {473,0.828339},
        {474,0.844202},
        {475,0.865864},
        {476,0.868497},
        {477,0.890253},
        {478,0.912538},
        {479,0.910385},
        {480,0.918822},
        {481,0.931726},
        {482,0.931813},
        {483,0.954248},
        {484,0.955545},
        {485,0.96242},
        {486,0.956424},
        {487,0.953352},
        {488,0.978564},
        {489,0.989104},
        {490,0.985615},
        {491,0.989469},
        {492,0.982262},
        {493,0.968801},
        {494,0.967332},
        {495,0.976836},
        {496,0.988729},
        {497,0.980826},
        {498,0.967361},
        {499,0.954754},
        {500,0.964132},
        {501,0.966125},
        {502,0.966772},
        {503,0.981834},
        {504,0.98232},
        {505,0.965685},
        {506,0.963135},
        {507,0.972261},
        {508,0.996498},
        {509,1},
        {510,0.9556},
        {511,0.844893},
        {512,0.534592},
        {513,0.190738},
        {514,0.048329},
        {515,0.013894},
        {516,0.005328},
        {517,0.002611},
        {518,0.001557},
        {519,0.0011},
        {520,0.000785},
        {521,0.000516},
        {522,0.000321},
        {523,0.000162},
        {524,0.000072},
        {525,0.000057},
        {526,0.000023},
        {527,0.000032}
      };

      static const double LANDSAT8_OLI3_SRF[][2] = 
      {
        {513,0.000016},
        {514,0.00011},
        {515,0.000247},
        {516,0.000362},
        {517,0.000648},
        {518,0.000935},
        {519,0.001332},
        {520,0.001816},
        {521,0.002515},
        {522,0.003446},
        {523,0.00488},
        {524,0.007024},
        {525,0.010441},
        {526,0.016247},
        {527,0.025513},
        {528,0.041451},
        {529,0.070551},
        {530,0.123444},
        {531,0.21168},
        {532,0.353885},
        {533,0.545856},
        {534,0.741205},
        {535,0.865225},
        {536,0.927396},
        {537,0.954627},
        {538,0.954163},
        {539,0.959215},
        {540,0.961328},
        {541,0.964902},
        {542,0.969873},
        {543,0.952489},
        {544,0.961397},
        {545,0.97827},
        {546,0.977533},
        {547,0.977001},
        {548,0.980884},
        {549,0.990784},
        {550,1},
        {551,0.992264},
        {552,0.982642},
        {553,0.983832},
        {554,0.977765},
        {555,0.965081},
        {556,0.957314},
        {557,0.946245},
        {558,0.947871},
        {559,0.959038},
        {560,0.966534},
        {561,0.977656},
        {562,0.966447},
        {563,0.953399},
        {564,0.958314},
        {565,0.970039},
        {566,0.978607},
        {567,0.983397},
        {568,0.98096},
        {569,0.974522},
        {570,0.967229},
        {571,0.979406},
        {572,0.978208},
        {573,0.975818},
        {574,0.974392},
        {575,0.979973},
        {576,0.968827},
        {577,0.969181},
        {578,0.967838},
        {579,0.982956},
        {580,0.979598},
        {581,0.963811},
        {582,0.968886},
        {583,0.983655},
        {584,0.986657},
        {585,0.974207},
        {586,0.946407},
        {587,0.904478},
        {588,0.809275},
        {589,0.684974},
        {590,0.525304},
        {591,0.345364},
        {592,0.190467},
        {593,0.087833},
        {594,0.035393},
        {595,0.014077},
        {596,0.005944},
        {597,0.002574},
        {598,0.001046},
        {599,0.000394},
        {600,0.000085}
      };

      static const double LANDSAT8_OLI4_SRF[][2] = 
      {
        {626,0.00027},
        {627,0.000895},
        {628,0.00185},
        {629,0.003648},
        {630,0.007197},
        {631,0.014515},
        {632,0.030432},
        {633,0.066861},
        {634,0.148518},
        {635,0.299778},
        {636,0.526812},
        {637,0.764443},
        {638,0.905473},
        {639,0.947949},
        {640,0.950823},
        {641,0.947418},
        {642,0.951831},
        {643,0.962705},
        {644,0.975075},
        {645,0.984173},
        {646,0.983613},
        {647,0.983434},
        {648,0.982911},
        {649,0.973636},
        {650,0.959441},
        {651,0.955641},
        {652,0.955548},
        {653,0.953337},
        {654,0.956628},
        {655,0.981688},
        {656,1},
        {657,0.992388},
        {658,0.984615},
        {659,0.981568},
        {660,0.97696},
        {661,0.97298},
        {662,0.98108},
        {663,0.996804},
        {664,0.992142},
        {665,0.980678},
        {666,0.964002},
        {667,0.962154},
        {668,0.970778},
        {669,0.96718},
        {670,0.966928},
        {671,0.949928},
        {672,0.848855},
        {673,0.609359},
        {674,0.31635},
        {675,0.123946},
        {676,0.046033},
        {677,0.017702},
        {678,0.007333},
        {679,0.003205},
        {680,0.001402},
        {681,0.000554},
        {682,0.000117}
      };

      static const double LANDSAT8_OLI5_SRF[][2] = 
      {
        {830,0.000011},
        {831,0.00005},
        {832,0.0001},
        {833,0.000239},
        {834,0.000314},
        {835,0.000495},
        {836,0.000719},
        {837,0.000986},
        {838,0.001445},
        {839,0.002107},
        {840,0.00316},
        {841,0.004744},
        {842,0.007059},
        {843,0.0109},
        {844,0.017346},
        {845,0.028332},
        {846,0.048191},
        {847,0.084363},
        {848,0.145365},
        {849,0.249733},
        {850,0.403526},
        {851,0.582623},
        {852,0.745037},
        {853,0.890315},
        {854,0.960215},
        {855,0.986833},
        {856,0.973133},
        {857,0.980606},
        {858,0.99612},
        {859,1},
        {860,0.989777},
        {861,0.980733},
        {862,0.975935},
        {863,0.972043},
        {864,0.957357},
        {865,0.951209},
        {866,0.947044},
        {867,0.953162},
        {868,0.951499},
        {869,0.94845},
        {870,0.940094},
        {871,0.950632},
        {872,0.956079},
        {873,0.96646},
        {874,0.969821},
        {875,0.93661},
        {876,0.891066},
        {877,0.788733},
        {878,0.63532},
        {879,0.448364},
        {880,0.288847},
        {881,0.174619},
        {882,0.100343},
        {883,0.058265},
        {884,0.034532},
        {885,0.02072},
        {886,0.01244},
        {887,0.007601},
        {888,0.004702},
        {889,0.002944},
        {890,0.00187},
        {891,0.001192},
        {892,0.000743},
        {893,0.000423},
        {894,0.000241},
        {895,0.000116},
        {896,0.000044}
      };

      static const double LANDSAT8_OLI6_SRF[][2] = 
      {
        {1516,0.000067},
        {1517,0.000151},
        {1518,0.000249},
        {1519,0.000348},
        {1520,0.000466},
        {1521,0.000585},
        {1522,0.000758},
        {1523,0.000932},
        {1524,0.00115},
        {1525,0.001369},
        {1526,0.001613},
        {1527,0.001859},
        {1528,0.002172},
        {1529,0.002488},
        {1530,0.002881},
        {1531,0.003277},
        {1532,0.003772},
        {1533,0.004271},
        {1534,0.004898},
        {1535,0.005528},
        {1536,0.006421},
        {1537,0.007319},
        {1538,0.008459},
        {1539,0.009606},
        {1540,0.010989},
        {1541,0.012372},
        {1542,0.014303},
        {1543,0.016248},
        {1544,0.019029},
        {1545,0.021831},
        {1546,0.02589},
        {1547,0.029952},
        {1548,0.035171},
        {1549,0.040434},
        {1550,0.047864},
        {1551,0.055352},
        {1552,0.065732},
        {1553,0.076166},
        {1554,0.089024},
        {1555,0.101893},
        {1556,0.12015},
        {1557,0.138642},
        {1558,0.163127},
        {1559,0.187803},
        {1560,0.220261},
        {1561,0.252894},
        {1562,0.291359},
        {1563,0.329939},
        {1564,0.375648},
        {1565,0.42147},
        {1566,0.47356},
        {1567,0.525681},
        {1568,0.578787},
        {1569,0.631645},
        {1570,0.676683},
        {1571,0.721282},
        {1572,0.75477},
        {1573,0.788248},
        {1574,0.821155},
        {1575,0.854065},
        {1576,0.87337},
        {1577,0.89183},
        {1578,0.899817},
        {1579,0.907252},
        {1580,0.913009},
        {1581,0.918685},
        {1582,0.922953},
        {1583,0.927163},
        {1584,0.92686},
        {1585,0.926413},
        {1586,0.925059},
        {1587,0.923683},
        {1588,0.923953},
        {1589,0.924259},
        {1590,0.922828},
        {1591,0.921383},
        {1592,0.922061},
        {1593,0.922756},
        {1594,0.924648},
        {1595,0.926605},
        {1596,0.934552},
        {1597,0.942525},
        {1598,0.944351},
        {1599,0.945872},
        {1600,0.946175},
        {1601,0.946432},
        {1602,0.947006},
        {1603,0.947589},
        {1604,0.950194},
        {1605,0.952859},
        {1606,0.951303},
        {1607,0.94967},
        {1608,0.953047},
        {1609,0.956494},
        {1610,0.959047},
        {1611,0.961545},
        {1612,0.960048},
        {1613,0.958335},
        {1614,0.959835},
        {1615,0.96147},
        {1616,0.960857},
        {1617,0.960175},
        {1618,0.960813},
        {1619,0.961486},
        {1620,0.964703},
        {1621,0.967943},
        {1622,0.969314},
        {1623,0.970589},
        {1624,0.973713},
        {1625,0.976906},
        {1626,0.9791},
        {1627,0.981258},
        {1628,0.981285},
        {1629,0.981372},
        {1630,0.988609},
        {1631,0.995869},
        {1632,0.998021},
        {1633,1},
        {1634,0.999848},
        {1635,0.999642},
        {1636,0.99659},
        {1637,0.993439},
        {1638,0.986217},
        {1639,0.978989},
        {1640,0.967125},
        {1641,0.95512},
        {1642,0.936199},
        {1643,0.917239},
        {1644,0.879424},
        {1645,0.840967},
        {1646,0.796545},
        {1647,0.751893},
        {1648,0.694313},
        {1649,0.636542},
        {1650,0.573232},
        {1651,0.509946},
        {1652,0.451966},
        {1653,0.39403},
        {1654,0.34275},
        {1655,0.291752},
        {1656,0.251309},
        {1657,0.211153},
        {1658,0.180823},
        {1659,0.15054},
        {1660,0.128463},
        {1661,0.106664},
        {1662,0.090735},
        {1663,0.074941},
        {1664,0.06379},
        {1665,0.052752},
        {1666,0.045028},
        {1667,0.03731},
        {1668,0.031821},
        {1669,0.02635},
        {1670,0.022504},
        {1671,0.018724},
        {1672,0.016045},
        {1673,0.013394},
        {1674,0.011483},
        {1675,0.009587},
        {1676,0.008227},
        {1677,0.006882},
        {1678,0.00591},
        {1679,0.00494},
        {1680,0.004257},
        {1681,0.003576},
        {1682,0.003055},
        {1683,0.002541},
        {1684,0.00216},
        {1685,0.001781},
        {1686,0.001512},
        {1687,0.001244},
        {1688,0.00104},
        {1689,0.000837},
        {1690,0.000677},
        {1691,0.000517},
        {1692,0.000409},
        {1693,0.000301},
        {1694,0.000206},
        {1695,0.000112},
        {1696,0.00004}
      };

      static const double LANDSAT8_OLI7_SRF[][2] = 
      {
        {2038,0.000037},
        {2039,0.000083},
        {2040,0.000131},
        {2041,0.000179},
        {2042,0.00024},
        {2043,0.000305},
        {2044,0.000368},
        {2045,0.00043},
        {2046,0.000512},
        {2047,0.000599},
        {2048,0.000704},
        {2049,0.000814},
        {2050,0.000947},
        {2051,0.001085},
        {2052,0.001222},
        {2053,0.00136},
        {2054,0.001546},
        {2055,0.001745},
        {2056,0.001964},
        {2057,0.002187},
        {2058,0.002439},
        {2059,0.002696},
        {2060,0.00301},
        {2061,0.003339},
        {2062,0.003733},
        {2063,0.004141},
        {2064,0.004627},
        {2065,0.005137},
        {2066,0.005728},
        {2067,0.006337},
        {2068,0.007139},
        {2069,0.007996},
        {2070,0.008903},
        {2071,0.009824},
        {2072,0.011005},
        {2073,0.012261},
        {2074,0.01361},
        {2075,0.014987},
        {2076,0.016872},
        {2077,0.018899},
        {2078,0.021},
        {2079,0.023121},
        {2080,0.025897},
        {2081,0.028847},
        {2082,0.032071},
        {2083,0.035363},
        {2084,0.040206},
        {2085,0.045432},
        {2086,0.050903},
        {2087,0.056429},
        {2088,0.06327},
        {2089,0.070409},
        {2090,0.079382},
        {2091,0.088907},
        {2092,0.10064},
        {2093,0.11283},
        {2094,0.128292},
        {2095,0.144668},
        {2096,0.162055},
        {2097,0.179714},
        {2098,0.20278},
        {2099,0.227234},
        {2100,0.253732},
        {2101,0.280925},
        {2102,0.311347},
        {2103,0.342526},
        {2104,0.377044},
        {2105,0.412621},
        {2106,0.45047},
        {2107,0.488816},
        {2108,0.522458},
        {2109,0.554715},
        {2110,0.593227},
        {2111,0.633521},
        {2112,0.663067},
        {2113,0.689664},
        {2114,0.722284},
        {2115,0.756529},
        {2116,0.776463},
        {2117,0.792667},
        {2118,0.813716},
        {2119,0.836001},
        {2120,0.846344},
        {2121,0.853714},
        {2122,0.867845},
        {2123,0.883615},
        {2124,0.886411},
        {2125,0.886127},
        {2126,0.895232},
        {2127,0.906527},
        {2128,0.909739},
        {2129,0.911091},
        {2130,0.917985},
        {2131,0.926131},
        {2132,0.929693},
        {2133,0.932227},
        {2134,0.936544},
        {2135,0.941406},
        {2136,0.942571},
        {2137,0.942952},
        {2138,0.943112},
        {2139,0.943194},
        {2140,0.945168},
        {2141,0.947537},
        {2142,0.948776},
        {2143,0.949694},
        {2144,0.949643},
        {2145,0.9494},
        {2146,0.952551},
        {2147,0.956635},
        {2148,0.953083},
        {2149,0.947423},
        {2150,0.949094},
        {2151,0.952773},
        {2152,0.950874},
        {2153,0.947477},
        {2154,0.947014},
        {2155,0.947085},
        {2156,0.951812},
        {2157,0.957717},
        {2158,0.953332},
        {2159,0.946412},
        {2160,0.947778},
        {2161,0.951119},
        {2162,0.951641},
        {2163,0.951518},
        {2164,0.948644},
        {2165,0.944956},
        {2166,0.942515},
        {2167,0.940311},
        {2168,0.9434},
        {2169,0.947923},
        {2170,0.94501},
        {2171,0.940213},
        {2172,0.938737},
        {2173,0.93806},
        {2174,0.941859},
        {2175,0.947019},
        {2176,0.946554},
        {2177,0.944482},
        {2178,0.947414},
        {2179,0.951661},
        {2180,0.949283},
        {2181,0.945318},
        {2182,0.939939},
        {2183,0.934142},
        {2184,0.935493},
        {2185,0.93882},
        {2186,0.939253},
        {2187,0.938955},
        {2188,0.934675},
        {2189,0.929162},
        {2190,0.927085},
        {2191,0.925692},
        {2192,0.930508},
        {2193,0.936899},
        {2194,0.933908},
        {2195,0.927984},
        {2196,0.930981},
        {2197,0.936472},
        {2198,0.935776},
        {2199,0.933523},
        {2200,0.935132},
        {2201,0.937592},
        {2202,0.946217},
        {2203,0.956567},
        {2204,0.955661},
        {2205,0.951991},
        {2206,0.956666},
        {2207,0.963135},
        {2208,0.964442},
        {2209,0.964365},
        {2210,0.963523},
        {2211,0.962434},
        {2212,0.962905},
        {2213,0.963685},
        {2214,0.962473},
        {2215,0.960741},
        {2216,0.959239},
        {2217,0.957814},
        {2218,0.957781},
        {2219,0.958041},
        {2220,0.953274},
        {2221,0.94716},
        {2222,0.951706},
        {2223,0.958833},
        {2224,0.960212},
        {2225,0.960339},
        {2226,0.954785},
        {2227,0.947696},
        {2228,0.951965},
        {2229,0.95906},
        {2230,0.960554},
        {2231,0.960764},
        {2232,0.95575},
        {2233,0.949261},
        {2234,0.953245},
        {2235,0.95997},
        {2236,0.963736},
        {2237,0.966786},
        {2238,0.964315},
        {2239,0.960173},
        {2240,0.965473},
        {2241,0.973416},
        {2242,0.977637},
        {2243,0.980904},
        {2244,0.98276},
        {2245,0.984155},
        {2246,0.984693},
        {2247,0.985056},
        {2248,0.991486},
        {2249,0.9996},
        {2250,0.997654},
        {2251,0.993153},
        {2252,0.992469},
        {2253,0.992675},
        {2254,0.995894},
        {2255,1},
        {2256,0.999279},
        {2257,0.997261},
        {2258,0.994356},
        {2259,0.991127},
        {2260,0.987747},
        {2261,0.984349},
        {2262,0.986037},
        {2263,0.989132},
        {2264,0.984536},
        {2265,0.978024},
        {2266,0.975019},
        {2267,0.972794},
        {2268,0.974168},
        {2269,0.97654},
        {2270,0.976199},
        {2271,0.975206},
        {2272,0.974409},
        {2273,0.973662},
        {2274,0.967502},
        {2275,0.959895},
        {2276,0.956943},
        {2277,0.955095},
        {2278,0.955085},
        {2279,0.955588},
        {2280,0.947195},
        {2281,0.93665},
        {2282,0.922405},
        {2283,0.907109},
        {2284,0.89494},
        {2285,0.883588},
        {2286,0.855489},
        {2287,0.823876},
        {2288,0.78488},
        {2289,0.744025},
        {2290,0.69852},
        {2291,0.651673},
        {2292,0.602539},
        {2293,0.552647},
        {2294,0.502693},
        {2295,0.452698},
        {2296,0.403993},
        {2297,0.355569},
        {2298,0.315712},
        {2299,0.27826},
        {2300,0.244645},
        {2301,0.212213},
        {2302,0.186151},
        {2303,0.161749},
        {2304,0.141435},
        {2305,0.122015},
        {2306,0.106531},
        {2307,0.092029},
        {2308,0.080268},
        {2309,0.069276},
        {2310,0.060703},
        {2311,0.052702},
        {2312,0.046332},
        {2313,0.040405},
        {2314,0.035634},
        {2315,0.031213},
        {2316,0.027516},
        {2317,0.024},
        {2318,0.021262},
        {2319,0.018688},
        {2320,0.016562},
        {2321,0.014545},
        {2322,0.01293},
        {2323,0.011426},
        {2324,0.010155},
        {2325,0.008959},
        {2326,0.007992},
        {2327,0.007088},
        {2328,0.006348},
        {2329,0.005643},
        {2330,0.005019},
        {2331,0.004415},
        {2332,0.003903},
        {2333,0.003421},
        {2334,0.003025},
        {2335,0.002651},
        {2336,0.00234},
        {2337,0.002047},
        {2338,0.001795},
        {2339,0.001554},
        {2340,0.001345},
        {2341,0.001145},
        {2342,0.000974},
        {2343,0.000811},
        {2344,0.00068},
        {2345,0.00056},
        {2346,0.00044},
        {2347,0.00032},
        {2348,0.000217},
        {2349,0.000119},
        {2350,0.000028}
      };

      static const double LANDSAT8_OLI8_SRF[][2] = 
      {
        {488,0.000216},
        {489,0.000514},
        {490,0.001013},
        {491,0.001587},
        {492,0.002619},
        {493,0.003841},
        {494,0.006195},
        {495,0.009003},
        {496,0.015515},
        {497,0.02327},
        {498,0.042723},
        {499,0.066556},
        {500,0.124269},
        {501,0.196006},
        {502,0.323232},
        {503,0.472496},
        {504,0.598302},
        {505,0.714816},
        {506,0.776009},
        {507,0.816555},
        {508,0.831881},
        {509,0.837994},
        {510,0.848975},
        {511,0.861668},
        {512,0.862705},
        {513,0.85964},
        {514,0.858532},
        {515,0.858244},
        {516,0.857148},
        {517,0.855759},
        {518,0.858455},
        {519,0.862834},
        {520,0.860609},
        {521,0.855993},
        {522,0.852785},
        {523,0.850183},
        {524,0.852788},
        {525,0.857317},
        {526,0.85913},
        {527,0.859876},
        {528,0.861508},
        {529,0.863494},
        {530,0.860265},
        {531,0.855102},
        {532,0.864423},
        {533,0.879249},
        {534,0.885593},
        {535,0.888663},
        {536,0.894758},
        {537,0.902017},
        {538,0.906294},
        {539,0.90929},
        {540,0.911734},
        {541,0.914},
        {542,0.909776},
        {543,0.902939},
        {544,0.908919},
        {545,0.919859},
        {546,0.92162},
        {547,0.919456},
        {548,0.91302},
        {549,0.905134},
        {550,0.892702},
        {551,0.878599},
        {552,0.876838},
        {553,0.879443},
        {554,0.878303},
        {555,0.875692},
        {556,0.872666},
        {557,0.869478},
        {558,0.875361},
        {559,0.884759},
        {560,0.889903},
        {561,0.893427},
        {562,0.885948},
        {563,0.874097},
        {564,0.875674},
        {565,0.882402},
        {566,0.891375},
        {567,0.901194},
        {568,0.903528},
        {569,0.902826},
        {570,0.907551},
        {571,0.91435},
        {572,0.914932},
        {573,0.91319},
        {574,0.915508},
        {575,0.919554},
        {576,0.920802},
        {577,0.92104},
        {578,0.924431},
        {579,0.929105},
        {580,0.930068},
        {581,0.929551},
        {582,0.937549},
        {583,0.948863},
        {584,0.947084},
        {585,0.940273},
        {586,0.940813},
        {587,0.944136},
        {588,0.945674},
        {589,0.946566},
        {590,0.942462},
        {591,0.936298},
        {592,0.939621},
        {593,0.946659},
        {594,0.942374},
        {595,0.933839},
        {596,0.93425},
        {597,0.93784},
        {598,0.940838},
        {599,0.943626},
        {600,0.952415},
        {601,0.963664},
        {602,0.967381},
        {603,0.968241},
        {604,0.968233},
        {605,0.967878},
        {606,0.965083},
        {607,0.961367},
        {608,0.957232},
        {609,0.952939},
        {610,0.949345},
        {611,0.946006},
        {612,0.948081},
        {613,0.952465},
        {614,0.954526},
        {615,0.955814},
        {616,0.959149},
        {617,0.963216},
        {618,0.964158},
        {619,0.963953},
        {620,0.965943},
        {621,0.968789},
        {622,0.972688},
        {623,0.977026},
        {624,0.978028},
        {625,0.977901},
        {626,0.974158},
        {627,0.969078},
        {628,0.969583},
        {629,0.972297},
        {630,0.972906},
        {631,0.972708},
        {632,0.969732},
        {633,0.96578},
        {634,0.965561},
        {635,0.966835},
        {636,0.966641},
        {637,0.965899},
        {638,0.972067},
        {639,0.981077},
        {640,0.980971},
        {641,0.977722},
        {642,0.974405},
        {643,0.971123},
        {644,0.963139},
        {645,0.95326},
        {646,0.953494},
        {647,0.957536},
        {648,0.963851},
        {649,0.970928},
        {650,0.969453},
        {651,0.96475},
        {652,0.966541},
        {653,0.970613},
        {654,0.973941},
        {655,0.977055},
        {656,0.982894},
        {657,0.98977},
        {658,0.988302},
        {659,0.983542},
        {660,0.987968},
        {661,0.995538},
        {662,0.998579},
        {663,1},
        {664,0.99981},
        {665,0.999061},
        {666,0.997891},
        {667,0.99658},
        {668,0.992555},
        {669,0.987369},
        {670,0.985862},
        {671,0.985761},
        {672,0.955679},
        {673,0.913945},
        {674,0.78014},
        {675,0.610197},
        {676,0.438556},
        {677,0.266714},
        {678,0.167313},
        {679,0.094013},
        {680,0.057564},
        {681,0.034787},
        {682,0.022882},
        {683,0.015178},
        {684,0.010755},
        {685,0.007663},
        {686,0.005756},
        {687,0.004292},
        {688,0.00322},
        {689,0.002283},
        {690,0.001545},
        {691,0.000879},
        {692,0.000214}
      };

      static const double LANDSAT8_OLI9_SRF[][2] = 
      {
        {1341,0.000086},
        {1342,0.000194},
        {1343,0.000321},
        {1344,0.000469},
        {1345,0.000647},
        {1346,0.000856},
        {1347,0.001119},
        {1348,0.001429},
        {1349,0.001828},
        {1350,0.002318},
        {1351,0.00303},
        {1352,0.004011},
        {1353,0.005398},
        {1354,0.007194},
        {1355,0.011124},
        {1356,0.016985},
        {1357,0.026321},
        {1358,0.042756},
        {1359,0.06922},
        {1360,0.115351},
        {1361,0.182833},
        {1362,0.316333},
        {1363,0.45703},
        {1364,0.596852},
        {1365,0.772118},
        {1366,0.862394},
        {1367,0.871575},
        {1368,0.930308},
        {1369,0.962664},
        {1370,0.931247},
        {1371,0.938676},
        {1372,0.983748},
        {1373,0.999626},
        {1374,0.986968},
        {1375,1},
        {1376,0.983239},
        {1377,0.990577},
        {1378,0.963584},
        {1379,0.964054},
        {1380,0.871343},
        {1381,0.872689},
        {1382,0.703198},
        {1383,0.606578},
        {1384,0.452434},
        {1385,0.29792},
        {1386,0.190978},
        {1387,0.113055},
        {1388,0.066174},
        {1389,0.041513},
        {1390,0.026084},
        {1391,0.016218},
        {1392,0.009774},
        {1393,0.006839},
        {1394,0.004358},
        {1395,0.00278},
        {1396,0.001809},
        {1397,0.001351},
        {1398,0.000887},
        {1399,0.000481},
        {1400,0.00022},
        {1401,0.000114},
        {1402,0.000029}
      };

      static const double LANDSAT8_OLI10_SRF[][2] = 
      {
        {10300,0.00779829},
        {10350,0.01622384},
        {10400,0.031957661},
        {10450,0.066079988},
        {10500,0.138395248},
        {10550,0.278406982},
        {10600,0.505662398},
        {10650,0.725864199},
        {10700,0.808512638},
        {10750,0.841080412},
        {10800,0.843348592},
        {10850,0.873433585},
        {10900,0.915442775},
        {10950,0.978028907},
        {11000,1},
        {11050,0.990887898},
        {11100,0.871548375},
        {11150,0.653390796},
        {11200,0.409641209},
        {11250,0.241318408},
        {11300,0.135921801}
      };

      static const double LANDSAT8_OLI11_SRF[][2] = 
      {
        {11500,0.478180944},
        {11550,0.677969452},
        {11600,0.856502695},
        {11650,0.966787346},
        {11700,0.977679663},
        {11750,0.906317161},
        {11800,0.862368041},
        {11850,0.829035771},
        {11900,0.859573529},
        {11950,0.896279126},
        {12000,0.972061678},
        {12050,0.997528121},
        {12100,0.987537925},
        {12150,0.987631547},
        {12200,0.954109591},
        {12250,0.923587577},
        {12300,0.961260942},
        {12350,1},
        {12400,0.977224356},
        {12450,0.80101634},
        {12500,0.524503731}
      };
      
      static const double LANDSAT7_ETM1_SRF[][2] = 
      {
        {435,0.016},
        {436,0.027},
        {437,0.048},
        {438,0.094},
        {439,0.167},
        {440,0.287},
        {441,0.459},
        {442,0.605},
        {443,0.728},
        {444,0.769},
        {445,0.792},
        {446,0.821},
        {447,0.857},
        {448,0.857},
        {449,0.862},
        {450,0.839},
        {451,0.845},
        {452,0.81},
        {453,0.802},
        {454,0.804},
        {455,0.779},
        {456,0.798},
        {457,0.816},
        {458,0.876},
        {459,0.888},
        {460,0.901},
        {461,0.918},
        {462,0.896},
        {463,0.903},
        {464,0.888},
        {465,0.89},
        {466,0.863},
        {467,0.86},
        {468,0.842},
        {469,0.866},
        {470,0.875},
        {471,0.881},
        {472,0.888},
        {473,0.898},
        {474,0.879},
        {475,0.884},
        {476,0.907},
        {477,0.928},
        {478,0.932},
        {479,0.955},
        {480,0.958},
        {481,0.948},
        {482,0.952},
        {483,0.956},
        {484,0.98},
        {485,0.98},
        {486,0.975},
        {487,0.973},
        {488,0.977},
        {489,0.958},
        {490,0.965},
        {491,0.957},
        {492,0.952},
        {493,0.973},
        {494,0.974},
        {495,0.995},
        {496,0.986},
        {497,0.986},
        {498,0.994},
        {499,1},
        {500,0.99},
        {501,0.99},
        {502,0.976},
        {503,0.983},
        {504,0.976},
        {505,0.983},
        {506,0.96},
        {507,0.973},
        {508,0.964},
        {509,0.975},
        {510,0.96},
        {511,0.932},
        {512,0.853},
        {513,0.684},
        {514,0.486},
        {515,0.293},
        {516,0.15},
        {517,0.073},
        {518,0.036},
        {519,0.019},
        {520,0.009}
      };      

      static const double LANDSAT7_ETM2_SRF[][2] = 
      {
        {500,0.0006},
        {501,0.0009},
        {502,0.001},
        {503,0.002},
        {504,0.002},
        {505,0.003},
        {506,0.005},
        {507,0.009},
        {508,0.0145},
        {509,0.0236},
        {510,0.026},
        {511,0.041},
        {512,0.06},
        {513,0.088},
        {514,0.126},
        {515,0.174},
        {516,0.236},
        {517,0.308},
        {518,0.388},
        {519,0.472},
        {520,0.552},
        {521,0.621},
        {522,0.676},
        {523,0.716},
        {524,0.743},
        {525,0.759},
        {526,0.769},
        {527,0.779},
        {528,0.79},
        {529,0.805},
        {530,0.822},
        {531,0.842},
        {532,0.861},
        {533,0.878},
        {534,0.893},
        {535,0.905},
        {536,0.916},
        {537,0.924},
        {538,0.933},
        {539,0.942},
        {540,0.947},
        {541,0.951},
        {542,0.953},
        {543,0.952},
        {544,0.951},
        {545,0.952},
        {546,0.951},
        {547,0.951},
        {548,0.952},
        {549,0.952},
        {550,0.953},
        {551,0.951},
        {552,0.95},
        {553,0.95},
        {554,0.951},
        {555,0.954},
        {556,0.96},
        {557,0.966},
        {558,0.968},
        {559,0.965},
        {560,0.959},
        {561,0.951},
        {562,0.944},
        {563,0.937},
        {564,0.932},
        {565,0.933},
        {566,0.935},
        {567,0.937},
        {568,0.94},
        {569,0.945},
        {570,0.951},
        {571,0.955},
        {572,0.957},
        {573,0.956},
        {574,0.957},
        {575,0.955},
        {576,0.952},
        {577,0.954},
        {578,0.958},
        {579,0.963},
        {580,0.973},
        {581,0.981},
        {582,0.988},
        {583,0.995},
        {584,1},
        {585,1},
        {586,0.994},
        {587,0.983},
        {588,0.969},
        {589,0.954},
        {590,0.942},
        {591,0.936},
        {592,0.932},
        {593,0.928},
        {594,0.924},
        {595,0.912},
        {596,0.883},
        {597,0.834},
        {598,0.763},
        {599,0.674},
        {600,0.574},
        {601,0.473},
        {602,0.38},
        {603,0.3},
        {604,0.235},
        {605,0.185},
        {606,0.146},
        {607,0.117},
        {608,0.094},
        {609,0.077},
        {610,0.062},
        {611,0.052},
        {612,0.042},
        {613,0.033},
        {614,0.026},
        {615,0.021},
        {616,0.0158},
        {617,0.01235},
        {618,0.0094},
        {619,0.00717},
        {620,0.0048},
        {621,0.0037},
        {622,0.0026},
        {623,0.0017},
        {624,0.0014}
      };      

      static const double LANDSAT7_ETM3_SRF[][2] = 
      {
        {614,0.0005},
        {615,0.0001},
        {616,0},
        {617,0.0005},
        {618,0.0006},
        {619,0.0014},
        {620,0.002},
        {621,0.003},
        {622,0.006},
        {623,0.013},
        {624,0.025},
        {625,0.047},
        {626,0.083},
        {627,0.137},
        {628,0.211},
        {629,0.306},
        {630,0.419},
        {631,0.545},
        {632,0.674},
        {633,0.788},
        {634,0.873},
        {635,0.921},
        {636,0.941},
        {637,0.943},
        {638,0.942},
        {639,0.939},
        {640,0.937},
        {641,0.935},
        {642,0.935},
        {643,0.938},
        {644,0.943},
        {645,0.949},
        {646,0.953},
        {647,0.961},
        {648,0.968},
        {649,0.971},
        {650,0.973},
        {651,0.974},
        {652,0.972},
        {653,0.969},
        {654,0.963},
        {655,0.958},
        {656,0.956},
        {657,0.955},
        {658,0.955},
        {659,0.956},
        {660,0.962},
        {661,0.969},
        {662,0.977},
        {663,0.983},
        {664,0.988},
        {665,0.993},
        {666,0.996},
        {667,0.997},
        {668,0.999},
        {669,1},
        {670,1},
        {671,0.998},
        {672,0.996},
        {673,0.995},
        {674,0.993},
        {675,0.992},
        {676,0.991},
        {677,0.989},
        {678,0.988},
        {679,0.984},
        {680,0.977},
        {681,0.97},
        {682,0.96},
        {683,0.949},
        {684,0.94},
        {685,0.932},
        {686,0.919},
        {687,0.898},
        {688,0.863},
        {689,0.809},
        {690,0.729},
        {691,0.625},
        {692,0.506},
        {693,0.382},
        {694,0.272},
        {695,0.183},
        {696,0.12},
        {697,0.079},
        {698,0.053},
        {699,0.036},
        {700,0.025},
        {701,0.0196},
        {702,0.0142},
        {703,0.0101},
        {704,0.0075}
      };      

      static const double LANDSAT7_ETM4_SRF[][2] = 
      {
        {740,0.0004},
        {745,0.0037},
        {750,0.001},
        {751,0.0144},
        {752,0.0182},
        {753,0.0216},
        {754,0.0269},
        {755,0.0315},
        {756,0.0377},
        {757,0.0474},
        {758,0.0562},
        {759,0.0688},
        {760,0.069},
        {761,0.083},
        {762,0.099},
        {763,0.121},
        {764,0.146},
        {765,0.175},
        {766,0.209},
        {767,0.248},
        {768,0.294},
        {769,0.346},
        {770,0.402},
        {771,0.463},
        {772,0.523},
        {773,0.588},
        {774,0.649},
        {775,0.705},
        {776,0.757},
        {777,0.797},
        {778,0.827},
        {779,0.853},
        {780,0.871},
        {781,0.884},
        {782,0.892},
        {783,0.899},
        {784,0.903},
        {785,0.908},
        {786,0.911},
        {787,0.916},
        {788,0.92},
        {789,0.925},
        {790,0.926},
        {791,0.927},
        {792,0.927},
        {793,0.929},
        {794,0.932},
        {795,0.93},
        {796,0.926},
        {797,0.926},
        {798,0.925},
        {799,0.928},
        {800,0.925},
        {801,0.926},
        {802,0.928},
        {803,0.928},
        {804,0.928},
        {805,0.923},
        {806,0.92},
        {807,0.919},
        {808,0.914},
        {809,0.91},
        {810,0.908},
        {811,0.905},
        {812,0.903},
        {813,0.904},
        {814,0.902},
        {815,0.909},
        {816,0.917},
        {817,0.92},
        {818,0.928},
        {819,0.938},
        {820,0.946},
        {821,0.953},
        {822,0.962},
        {823,0.969},
        {824,0.971},
        {825,0.971},
        {826,0.97},
        {827,0.969},
        {828,0.969},
        {829,0.97},
        {830,0.967},
        {831,0.969},
        {832,0.968},
        {833,0.963},
        {834,0.965},
        {835,0.967},
        {836,0.965},
        {837,0.963},
        {838,0.958},
        {839,0.95},
        {840,0.949},
        {841,0.943},
        {842,0.933},
        {843,0.929},
        {844,0.928},
        {845,0.925},
        {846,0.924},
        {847,0.927},
        {848,0.932},
        {849,0.934},
        {850,0.943},
        {851,0.952},
        {852,0.956},
        {853,0.966},
        {854,0.977},
        {855,0.985},
        {856,0.99},
        {857,0.992},
        {858,0.993},
        {859,0.994},
        {860,0.998},
        {861,0.996},
        {862,0.992},
        {863,0.991},
        {864,0.992},
        {865,0.994},
        {866,0.993},
        {867,0.997},
        {868,0.997},
        {869,0.996},
        {870,0.998},
        {871,0.999},
        {872,1},
        {873,0.999},
        {874,0.996},
        {875,0.991},
        {876,0.99},
        {877,0.991},
        {878,0.985},
        {879,0.978},
        {880,0.969},
        {881,0.955},
        {882,0.937},
        {883,0.916},
        {884,0.892},
        {885,0.868},
        {886,0.845},
        {887,0.824},
        {888,0.811},
        {889,0.807},
        {890,0.819},
        {891,0.841},
        {892,0.868},
        {893,0.892},
        {894,0.892},
        {895,0.854},
        {896,0.77},
        {897,0.644},
        {898,0.501},
        {899,0.365},
        {900,0.256},
        {901,0.177},
        {902,0.122},
        {903,0.085},
        {904,0.061},
        {905,0.044},
        {906,0.032},
        {907,0.025},
        {908,0.019},
        {909,0.014},
        {910,0.011},
        {911,0.0107},
        {912,0.008},
        {913,0.0061},
        {914,0.0052}
      };      

      static const double LANDSAT7_ETM5_SRF[][2] = 
      {
        {1508,0.001},
        {1510,0.001},
        {1512,0.013},
        {1514,0.006},
        {1515,0.012},
        {1517,0.003},
        {1519,0.015},
        {1521,0.012},
        {1523,0.024},
        {1525,0.04},
        {1526,0.041},
        {1528,0.057},
        {1530,0.076},
        {1532,0.097},
        {1534,0.12},
        {1536,0.176},
        {1538,0.215},
        {1540,0.274},
        {1542,0.339},
        {1543,0.393},
        {1545,0.462},
        {1547,0.499},
        {1549,0.558},
        {1551,0.598},
        {1553,0.634},
        {1555,0.667},
        {1557,0.704},
        {1559,0.724},
        {1561,0.75},
        {1563,0.778},
        {1565,0.808},
        {1567,0.825},
        {1569,0.851},
        {1571,0.867},
        {1572,0.872},
        {1574,0.884},
        {1576,0.902},
        {1578,0.901},
        {1580,0.896},
        {1582,0.897},
        {1584,0.89},
        {1586,0.899},
        {1588,0.884},
        {1590,0.876},
        {1592,0.867},
        {1594,0.873},
        {1596,0.872},
        {1598,0.879},
        {1600,0.874},
        {1602,0.861},
        {1604,0.859},
        {1606,0.877},
        {1608,0.879},
        {1610,0.899},
        {1613,0.893},
        {1615,0.9},
        {1617,0.897},
        {1619,0.917},
        {1621,0.921},
        {1623,0.926},
        {1625,0.929},
        {1627,0.945},
        {1629,0.947},
        {1631,0.948},
        {1633,0.955},
        {1635,0.952},
        {1637,0.969},
        {1639,0.96},
        {1641,0.962},
        {1643,0.959},
        {1645,0.978},
        {1647,0.96},
        {1649,0.955},
        {1651,0.952},
        {1653,0.951},
        {1656,0.952},
        {1658,0.956},
        {1660,0.944},
        {1662,0.935},
        {1664,0.933},
        {1666,0.928},
        {1668,0.942},
        {1670,0.948},
        {1672,0.942},
        {1674,0.933},
        {1676,0.944},
        {1678,0.948},
        {1680,0.945},
        {1682,0.943},
        {1684,0.951},
        {1687,0.964},
        {1689,0.967},
        {1691,0.971},
        {1693,0.974},
        {1695,0.991},
        {1697,0.995},
        {1699,0.999},
        {1701,0.996},
        {1703,0.994},
        {1705,1},
        {1707,0.994},
        {1709,0.983},
        {1711,0.99},
        {1714,0.987},
        {1716,0.992},
        {1718,0.986},
        {1720,0.981},
        {1722,0.983},
        {1724,0.976},
        {1725,0.978},
        {1726,0.97},
        {1728,0.968},
        {1730,0.96},
        {1732,0.944},
        {1734,0.921},
        {1736,0.883},
        {1738,0.845},
        {1740,0.791},
        {1742,0.711},
        {1744,0.638},
        {1747,0.547},
        {1749,0.462},
        {1751,0.393},
        {1753,0.325},
        {1755,0.267},
        {1757,0.212},
        {1759,0.175},
        {1761,0.142},
        {1763,0.111},
        {1765,0.084},
        {1767,0.077},
        {1769,0.058},
        {1771,0.049},
        {1773,0.042},
        {1775,0.039},
        {1777,0.034},
        {1779,0.02},
        {1781,0.021},
        {1783,0.022},
        {1785,0.011},
        {1787,0.012},
        {1789,0.004},
        {1791,0.008}
      };      
      
      static const double LANDSAT7_ETM7_SRF[][2] = 
      {
        {2015,0.002},
        {2017,0.002},
        {2018,0.002},
        {2020,0.012},
        {2022,0.009},
        {2024,0.007},
        {2026,0.011},
        {2028,0.02},
        {2030,0.017},
        {2032,0.03},
        {2034,0.035},
        {2035,0.037},
        {2037,0.044},
        {2039,0.051},
        {2041,0.065},
        {2043,0.08},
        {2045,0.088},
        {2047,0.102},
        {2049,0.133},
        {2051,0.165},
        {2052,0.188},
        {2054,0.22},
        {2056,0.264},
        {2058,0.316},
        {2060,0.367},
        {2062,0.421},
        {2064,0.484},
        {2066,0.554},
        {2067,0.59},
        {2069,0.67},
        {2071,0.683},
        {2073,0.73},
        {2075,0.756},
        {2077,0.767},
        {2079,0.794},
        {2081,0.774},
        {2083,0.776},
        {2085,0.789},
        {2086,0.775},
        {2088,0.784},
        {2090,0.778},
        {2092,0.768},
        {2094,0.762},
        {2096,0.761},
        {2099,0.775},
        {2100,0.775},
        {2102,0.764},
        {2104,0.784},
        {2106,0.792},
        {2108,0.814},
        {2110,0.794},
        {2112,0.825},
        {2114,0.817},
        {2116,0.806},
        {2117,0.819},
        {2119,0.821},
        {2121,0.852},
        {2123,0.832},
        {2125,0.836},
        {2127,0.85},
        {2129,0.855},
        {2131,0.862},
        {2133,0.853},
        {2135,0.871},
        {2136,0.848},
        {2138,0.882},
        {2140,0.875},
        {2142,0.86},
        {2144,0.856},
        {2146,0.887},
        {2148,0.85},
        {2150,0.872},
        {2151,0.879},
        {2153,0.857},
        {2155,0.865},
        {2157,0.867},
        {2159,0.871},
        {2161,0.882},
        {2163,0.87},
        {2165,0.869},
        {2166,0.873},
        {2168,0.877},
        {2170,0.868},
        {2172,0.88},
        {2174,0.877},
        {2176,0.87},
        {2178,0.878},
        {2180,0.88},
        {2182,0.868},
        {2183,0.881},
        {2185,0.87},
        {2187,0.856},
        {2189,0.863},
        {2191,0.863},
        {2193,0.857},
        {2195,0.844},
        {2197,0.859},
        {2199,0.857},
        {2201,0.852},
        {2203,0.866},
        {2205,0.868},
        {2207,0.856},
        {2209,0.856},
        {2210,0.847},
        {2212,0.861},
        {2214,0.862},
        {2216,0.84},
        {2218,0.856},
        {2220,0.838},
        {2222,0.856},
        {2223,0.837},
        {2226,0.842},
        {2227,0.826},
        {2229,0.844},
        {2231,0.827},
        {2233,0.842},
        {2235,0.822},
        {2237,0.843},
        {2239,0.823},
        {2241,0.854},
        {2242,0.839},
        {2244,0.853},
        {2246,0.854},
        {2248,0.865},
        {2250,0.873},
        {2252,0.869},
        {2254,0.865},
        {2256,0.893},
        {2258,0.89},
        {2259,0.89},
        {2261,0.906},
        {2263,0.924},
        {2265,0.92},
        {2267,0.922},
        {2269,0.939},
        {2271,0.916},
        {2273,0.94},
        {2274,0.93},
        {2276,0.942},
        {2278,0.957},
        {2280,0.954},
        {2282,0.951},
        {2284,0.954},
        {2286,0.966},
        {2288,0.975},
        {2290,0.985},
        {2292,0.971},
        {2293,0.973},
        {2295,0.97},
        {2297,0.993},
        {2299,0.996},
        {2301,0.983},
        {2303,0.972},
        {2305,1},
        {2307,0.998},
        {2309,0.971},
        {2310,0.968},
        {2312,0.967},
        {2314,0.962},
        {2316,0.949},
        {2318,0.923},
        {2320,0.929},
        {2322,0.917},
        {2323,0.934},
        {2325,0.903},
        {2327,0.926},
        {2329,0.916},
        {2331,0.942},
        {2333,0.924},
        {2335,0.92},
        {2337,0.863},
        {2339,0.824},
        {2340,0.775},
        {2342,0.684},
        {2344,0.583},
        {2346,0.48},
        {2348,0.378},
        {2350,0.275},
        {2352,0.233},
        {2354,0.171},
        {2355,0.131},
        {2357,0.111},
        {2359,0.081},
        {2361,0.069},
        {2363,0.046},
        {2365,0.029},
        {2367,0.038}
      };      
      
      static const double LANDSAT7_ETM8PAN_SRF[][2] = 
      {
        {502,0.001},
        {504,0.008},
        {506,0.027},
        {508,0.066},
        {510,0.15},
        {512,0.289},
        {514,0.447},
        {516,0.556},
        {518,0.594},
        {520,0.599},
        {522,0.6},
        {524,0.607},
        {526,0.612},
        {528,0.615},
        {530,0.61},
        {532,0.604},
        {534,0.602},
        {536,0.605},
        {538,0.614},
        {540,0.627},
        {542,0.637},
        {544,0.643},
        {546,0.646},
        {548,0.643},
        {550,0.632},
        {552,0.627},
        {554,0.623},
        {556,0.626},
        {558,0.634},
        {560,0.642},
        {562,0.655},
        {564,0.672},
        {566,0.683},
        {568,0.692},
        {570,0.697},
        {572,0.7},
        {574,0.702},
        {576,0.708},
        {578,0.712},
        {580,0.713},
        {582,0.716},
        {584,0.718},
        {586,0.718},
        {588,0.71},
        {590,0.707},
        {592,0.705},
        {594,0.701},
        {596,0.705},
        {598,0.718},
        {600,0.73},
        {602,0.742},
        {604,0.75},
        {606,0.757},
        {608,0.763},
        {610,0.764},
        {612,0.755},
        {614,0.748},
        {616,0.742},
        {618,0.733},
        {620,0.729},
        {622,0.728},
        {624,0.73},
        {626,0.732},
        {628,0.733},
        {630,0.734},
        {632,0.742},
        {634,0.748},
        {636,0.751},
        {638,0.755},
        {640,0.758},
        {642,0.76},
        {644,0.767},
        {646,0.769},
        {648,0.773},
        {650,0.779},
        {652,0.781},
        {654,0.783},
        {656,0.787},
        {658,0.791},
        {660,0.793},
        {662,0.792},
        {664,0.791},
        {666,0.792},
        {668,0.796},
        {670,0.798},
        {672,0.801},
        {674,0.804},
        {676,0.808},
        {678,0.815},
        {680,0.82},
        {682,0.827},
        {684,0.829},
        {686,0.834},
        {688,0.838},
        {690,0.842},
        {692,0.847},
        {694,0.853},
        {696,0.856},
        {698,0.862},
        {700,0.868},
        {702,0.871},
        {704,0.876},
        {706,0.883},
        {708,0.886},
        {710,0.892},
        {712,0.894},
        {714,0.896},
        {716,0.896},
        {718,0.894},
        {720,0.894},
        {722,0.894},
        {724,0.895},
        {726,0.896},
        {728,0.898},
        {730,0.9},
        {732,0.903},
        {734,0.904},
        {736,0.903},
        {738,0.902},
        {740,0.9},
        {742,0.897},
        {744,0.896},
        {746,0.891},
        {748,0.88},
        {750,0.879},
        {752,0.88},
        {754,0.873},
        {756,0.873},
        {758,0.877},
        {760,0.88},
        {762,0.88},
        {764,0.885},
        {766,0.888},
        {768,0.892},
        {770,0.893},
        {772,0.902},
        {774,0.908},
        {776,0.913},
        {778,0.92},
        {780,0.92},
        {782,0.918},
        {784,0.916},
        {786,0.916},
        {788,0.919},
        {790,0.918},
        {792,0.917},
        {794,0.916},
        {796,0.915},
        {798,0.918},
        {800,0.927},
        {802,0.928},
        {804,0.932},
        {806,0.943},
        {808,0.952},
        {810,0.957},
        {812,0.962},
        {814,0.969},
        {816,0.97},
        {818,0.972},
        {820,0.977},
        {822,0.979},
        {824,0.981},
        {826,0.987},
        {828,0.992},
        {830,0.997},
        {832,0.998},
        {834,1},
        {836,0.996},
        {838,0.995},
        {840,0.994},
        {842,0.992},
        {844,0.992},
        {846,0.994},
        {848,0.992},
        {850,0.987},
        {852,0.978},
        {854,0.968},
        {856,0.962},
        {858,0.957},
        {860,0.949},
        {862,0.937},
        {864,0.928},
        {866,0.92},
        {868,0.911},
        {870,0.907},
        {872,0.905},
        {874,0.906},
        {876,0.911},
        {878,0.922},
        {880,0.94},
        {882,0.962},
        {884,0.977},
        {886,0.979},
        {888,0.957},
        {890,0.895},
        {892,0.787},
        {894,0.647},
        {896,0.496},
        {898,0.363},
        {900,0.258},
        {902,0.181},
        {904,0.127},
        {906,0.089},
        {908,0.062},
        {910,0.044}
      };

      static const double CBERS4_MUXB5_SRF[][2] = 
      {
        { 421 , 0.0000526890  },
        { 422 , 0.0001053780  },
        { 423 , 0.0000526890  },
        { 424 , 0.0000000000  },
        { 425 , 0.0000000000  },
        { 426 , 0.0000000000  },
        { 427 , 0.0000482633  },
        { 428 , 0.0000965267  },
        { 429 , 0.0002842930  },
        { 430 , 0.0004720593  },
        { 431 , 0.0007384815  },
        { 432 , 0.0010049037  },
        { 433 , 0.0012079852  },
        { 434 , 0.0014110666  },
        { 435 , 0.0021316968  },
        { 436 , 0.0028523269  },
        { 437 , 0.0062826055  },
        { 438 , 0.0097128841  },
        { 439 , 0.0174169915  },
        { 440 , 0.0251210990  },
        { 441 , 0.0412177823  },
        { 442 , 0.0573144655  },
        { 443 , 0.0996880363  },
        { 444 , 0.1420616070  },
        { 445 , 0.2614926777  },
        { 446 , 0.3809237485  },
        { 447 , 0.5118168683  },
        { 448 , 0.6427099880  },
        { 449 , 0.6630050389  },
        { 450 , 0.6833000897  },
        { 451 , 0.6990656859  },
        { 452 , 0.7148312821  },
        { 453 , 0.7186192155  },
        { 454 , 0.7224071490  },
        { 455 , 0.7160350045  },
        { 456 , 0.7096628600  },
        { 457 , 0.7048414589  },
        { 458 , 0.7000200579  },
        { 459 , 0.7103798710  },
        { 460 , 0.7207396841  },
        { 461 , 0.7294953875  },
        { 462 , 0.7382510908  },
        { 463 , 0.7386517593  },
        { 464 , 0.7390524277  },
        { 465 , 0.7458904778  },
        { 466 , 0.7527285279  },
        { 467 , 0.7464043090  },
        { 468 , 0.7400800900  },
        { 469 , 0.7384613032  },
        { 470 , 0.7368425164  },
        { 471 , 0.7365073626  },
        { 472 , 0.7361722087  },
        { 473 , 0.7423525612  },
        { 474 , 0.7485329137  },
        { 475 , 0.7548914794  },
        { 476 , 0.7612500451  },
        { 477 , 0.7540900394  },
        { 478 , 0.7469300337  },
        { 479 , 0.7466014164  },
        { 480 , 0.7462727991  },
        { 481 , 0.7468420947  },
        { 482 , 0.7474113903  },
        { 483 , 0.7624332544  },
        { 484 , 0.7774551185  },
        { 485 , 0.7909564755  },
        { 486 , 0.8044578326  },
        { 487 , 0.8065016050  },
        { 488 , 0.8085453774  },
        { 489 , 0.8141045230  },
        { 490 , 0.8196636686  },
        { 491 , 0.8281524107  },
        { 492 , 0.8366411529  },
        { 493 , 0.8576382745  },
        { 494 , 0.8786353962  },
        { 495 , 0.8980385305  },
        { 496 , 0.9174416649  },
        { 497 , 0.9262161262  },
        { 498 , 0.9349905876  },
        { 499 , 0.9418006266  },
        { 500 , 0.9486106656  },
        { 501 , 0.9464323568  },
        { 502 , 0.9442540480  },
        { 503 , 0.9580565440  },
        { 504 , 0.9718590401  },
        { 505 , 0.9778238584  },
        { 506 , 0.9837886768  },
        { 507 , 0.9918943384  },
        { 508 , 1.0000000000  },
        { 509 , 0.9913978556  },
        { 510 , 0.9827957111  },
        { 511 , 0.9810986527  },
        { 512 , 0.9794015943  },
        { 513 , 0.9786207175  },
        { 514 , 0.9778398408  },
        { 515 , 0.9519019650  },
        { 516 , 0.9259640891  },
        { 517 , 0.8805665337  },
        { 518 , 0.8351689782  },
        { 519 , 0.8106135127  },
        { 520 , 0.7860580472  },
        { 521 , 0.7792866818  },
        { 522 , 0.7725153165  },
        { 523 , 0.7038166352  },
        { 524 , 0.6351179538  },
        { 525 , 0.5085662730  },
        { 526 , 0.3820145922  },
        { 527 , 0.2971561215  },
        { 528 , 0.2122976509  },
        { 529 , 0.1719771473  },
        { 530 , 0.1316566437  },
        { 531 , 0.1062382517  },
        { 532 , 0.0808198597  },
        { 533 , 0.0639739208  },
        { 534 , 0.0471279818  },
        { 535 , 0.0366205601  },
        { 536 , 0.0261131384  },
        { 537 , 0.0209366477  },
        { 538 , 0.0157601570  },
        { 539 , 0.0129941639  },
        { 540 , 0.0102281709  },
        { 541 , 0.0092722109  },
        { 542 , 0.0083162510  },
        { 543 , 0.0068270039  },
        { 544 , 0.0053377569  },
        { 545 , 0.0045159565  },
        { 546 , 0.0036941562  },
        { 547 , 0.0028124182  },
        { 548 , 0.0019306803  },
        { 549 , 0.0014976711  },
        { 550 , 0.0010646619  },
        { 551 , 0.0009583648  },
        { 552 , 0.0008520677  },
        { 553 , 0.0009616556  },
        { 554 , 0.0010712436  },
        { 555 , 0.0010452893  },
        { 556 , 0.0010193350  },
        { 557 , 0.0010173016  },
        { 558 , 0.0010152681  },
        { 559 , 0.0010614673  },
        { 560 , 0.0011076665  },
        { 561 , 0.0009727444  },
        { 562 , 0.0008378222  },
        { 563 , 0.0007978795  },
        { 564 , 0.0007579368  },
        { 565 , 0.0007975094  },
        { 566 , 0.0008370820  },
        { 567 , 0.0007764541  },
        { 568 , 0.0007158263  },
        { 569 , 0.0008400645  },
        { 570 , 0.0009643026  },
        { 571 , 0.0009223650  },
        { 572 , 0.0008804274  },
        { 573 , 0.0008212484  },
        { 574 , 0.0007620693  },
        { 575 , 0.0008380575  },
        { 576 , 0.0009140457  },
        { 577 , 0.0008347905  },
        { 578 , 0.0007555352  },
        { 579 , 0.0007550691  },
        { 580 , 0.0007546030  },
        { 581 , 0.0008222925  },
        { 582 , 0.0008899819  },
        { 583 , 0.0008421498  },
        { 584 , 0.0007943177  },
        { 585 , 0.0007129434  },
        { 586 , 0.0006315691  },
        { 587 , 0.0006103182  },
        { 588 , 0.0005890673  },
        { 589 , 0.0006142982  },
        { 590 , 0.0006395290  },
        { 591 , 0.0006619936  },
        { 592 , 0.0006844582  },
        { 593 , 0.0006171849  },
        { 594 , 0.0005499117  },
        { 595 , 0.0005728040  },
        { 596 , 0.0005956964  },
        { 597 , 0.0004453629  },
        { 598 , 0.0002950293  },
        { 599 , 0.0001475147  }
      };
      
      static const double CBERS4_MUXB6_SRF[][2] = 
      {
        { 403 , 0.0000771482  },
        { 404 , 0.0001542964  },
        { 405 , 0.0000771482  },
        { 406 , 0.0000000000  },
        { 407 , 0.0000000000  },
        { 408 , 0.0000000000  },
        { 409 , 0.0000681768  },
        { 410 , 0.0001363537  },
        { 411 , 0.0002663343  },
        { 412 , 0.0003963149  },
        { 413 , 0.0002611546  },
        { 414 , 0.0001259944  },
        { 415 , 0.0000629972  },
        { 416 , 0.0000000000  },
        { 417 , 0.0000000000  },
        { 418 , 0.0000000000  },
        { 419 , 0.0000567467  },
        { 420 , 0.0001134933  },
        { 421 , 0.0000567467  },
        { 422 , 0.0000000000  },
        { 423 , 0.0000000000  },
        { 424 , 0.0000000000  },
        { 425 , 0.0000000000  },
        { 426 , 0.0000000000  },
        { 427 , 0.0000000000  },
        { 428 , 0.0000000000  },
        { 429 , 0.0001979817  },
        { 430 , 0.0003959633  },
        { 431 , 0.0001979817  },
        { 432 , 0.0000000000  },
        { 433 , 0.0000924690  },
        { 434 , 0.0001849379  },
        { 435 , 0.0000924690  },
        { 436 , 0.0000000000  },
        { 437 , 0.0001771123  },
        { 438 , 0.0003542247  },
        { 439 , 0.0001771123  },
        { 440 , 0.0000000000  },
        { 441 , 0.0002535621  },
        { 442 , 0.0005071242  },
        { 443 , 0.0003765948  },
        { 444 , 0.0002460654  },
        { 445 , 0.0005677972  },
        { 446 , 0.0008895290  },
        { 447 , 0.0007642153  },
        { 448 , 0.0006389015  },
        { 449 , 0.0009023015  },
        { 450 , 0.0011657015  },
        { 451 , 0.0013116384  },
        { 452 , 0.0014575753  },
        { 453 , 0.0012141850  },
        { 454 , 0.0009707947  },
        { 455 , 0.0008557714  },
        { 456 , 0.0007407481  },
        { 457 , 0.0006962294  },
        { 458 , 0.0006517106  },
        { 459 , 0.0006475191  },
        { 460 , 0.0006433276  },
        { 461 , 0.0007035538  },
        { 462 , 0.0007637799  },
        { 463 , 0.0008515242  },
        { 464 , 0.0009392684  },
        { 465 , 0.0008039145  },
        { 466 , 0.0006685606  },
        { 467 , 0.0005295266  },
        { 468 , 0.0003904926  },
        { 469 , 0.0006752285  },
        { 470 , 0.0009599644  },
        { 471 , 0.0008539232  },
        { 472 , 0.0007478820  },
        { 473 , 0.0008605483  },
        { 474 , 0.0009732146  },
        { 475 , 0.0013297029  },
        { 476 , 0.0016861912  },
        { 477 , 0.0011063315  },
        { 478 , 0.0005264718  },
        { 479 , 0.0006999780  },
        { 480 , 0.0008734841  },
        { 481 , 0.0008054835  },
        { 482 , 0.0007374830  },
        { 483 , 0.0005653555  },
        { 484 , 0.0003932280  },
        { 485 , 0.0009460479  },
        { 486 , 0.0014988679  },
        { 487 , 0.0013475584  },
        { 488 , 0.0011962489  },
        { 489 , 0.0017856869  },
        { 490 , 0.0023751249  },
        { 491 , 0.0022235215  },
        { 492 , 0.0020719181  },
        { 493 , 0.0027046968  },
        { 494 , 0.0033374755  },
        { 495 , 0.0041784129  },
        { 496 , 0.0050193503  },
        { 497 , 0.0063658308  },
        { 498 , 0.0077123114  },
        { 499 , 0.0080908673  },
        { 500 , 0.0084694232  },
        { 501 , 0.0076180854  },
        { 502 , 0.0067667475  },
        { 503 , 0.0077359118  },
        { 504 , 0.0087050760  },
        { 505 , 0.0106940435  },
        { 506 , 0.0126830110  },
        { 507 , 0.0185655267  },
        { 508 , 0.0244480423  },
        { 509 , 0.0362467750  },
        { 510 , 0.0480455077  },
        { 511 , 0.0637188744  },
        { 512 , 0.0793922410  },
        { 513 , 0.1010815379  },
        { 514 , 0.1227708348  },
        { 515 , 0.1656339224  },
        { 516 , 0.2084970099  },
        { 517 , 0.2902464499  },
        { 518 , 0.3719958899  },
        { 519 , 0.4608187031  },
        { 520 , 0.5496415164  },
        { 521 , 0.6043723771  },
        { 522 , 0.6591032379  },
        { 523 , 0.6947529368  },
        { 524 , 0.7304026357  },
        { 525 , 0.7626603909  },
        { 526 , 0.7949181460  },
        { 527 , 0.8110202337  },
        { 528 , 0.8271223214  },
        { 529 , 0.8305024548  },
        { 530 , 0.8338825883  },
        { 531 , 0.8338910281  },
        { 532 , 0.8338994680  },
        { 533 , 0.8406958425  },
        { 534 , 0.8474922170  },
        { 535 , 0.8583185900  },
        { 536 , 0.8691449630  },
        { 537 , 0.8819671277  },
        { 538 , 0.8947892923  },
        { 539 , 0.8896460241  },
        { 540 , 0.8845027558  },
        { 541 , 0.8904884703  },
        { 542 , 0.8964741847  },
        { 543 , 0.8896889756  },
        { 544 , 0.8829037664  },
        { 545 , 0.8802537731  },
        { 546 , 0.8776037797  },
        { 547 , 0.8775616279  },
        { 548 , 0.8775194762  },
        { 549 , 0.8888734034  },
        { 550 , 0.9002273306  },
        { 551 , 0.9195023295  },
        { 552 , 0.9387773283  },
        { 553 , 0.9586765476  },
        { 554 , 0.9785757669  },
        { 555 , 0.9847531264  },
        { 556 , 0.9909304858  },
        { 557 , 0.9927681552  },
        { 558 , 0.9946058246  },
        { 559 , 0.9973029123  },
        { 560 , 1.0000000000  },
        { 561 , 0.9923749228  },
        { 562 , 0.9847498455  },
        { 563 , 0.9895018620  },
        { 564 , 0.9942538786  },
        { 565 , 0.9916121072  },
        { 566 , 0.9889703358  },
        { 567 , 0.9888507410  },
        { 568 , 0.9887311461  },
        { 569 , 0.9822953972  },
        { 570 , 0.9758596482  },
        { 571 , 0.9672311345  },
        { 572 , 0.9586026208  },
        { 573 , 0.9594515553  },
        { 574 , 0.9603004898  },
        { 575 , 0.9512578547  },
        { 576 , 0.9422152196  },
        { 577 , 0.9520260441  },
        { 578 , 0.9618368687  },
        { 579 , 0.9609578104  },
        { 580 , 0.9600787521  },
        { 581 , 0.9610983521  },
        { 582 , 0.9621179521  },
        { 583 , 0.9475162760  },
        { 584 , 0.9329146000  },
        { 585 , 0.9294455638  },
        { 586 , 0.9259765277  },
        { 587 , 0.9166426797  },
        { 588 , 0.9073088318  },
        { 589 , 0.9040602796  },
        { 590 , 0.9008117274  },
        { 591 , 0.8770514083  },
        { 592 , 0.8532910891  },
        { 593 , 0.8060367365  },
        { 594 , 0.7587823839  },
        { 595 , 0.6988231488  },
        { 596 , 0.6388639136  },
        { 597 , 0.5606053172  },
        { 598 , 0.4823467208  },
        { 599 , 0.4159953669  },
        { 600 , 0.3496440130  },
        { 601 , 0.2868209112  },
        { 602 , 0.2239978094  },
        { 603 , 0.1820797303  },
        { 604 , 0.1401616512  },
        { 605 , 0.1079018209  },
        { 606 , 0.0756419907  },
        { 607 , 0.0589459155  },
        { 608 , 0.0422498402  },
        { 609 , 0.0328977174  },
        { 610 , 0.0235455946  },
        { 611 , 0.0187631386  },
        { 612 , 0.0139806827  },
        { 613 , 0.0115482076  },
        { 614 , 0.0091157325  },
        { 615 , 0.0076211909  },
        { 616 , 0.0061266494  },
        { 617 , 0.0053185632  },
        { 618 , 0.0045104770  },
        { 619 , 0.0043796062  },
        { 620 , 0.0042487354  },
        { 621 , 0.0041641945  },
        { 622 , 0.0040796537  },
        { 623 , 0.0038788311  },
        { 624 , 0.0036780086  },
        { 625 , 0.0034315303  },
        { 626 , 0.0031850521  },
        { 627 , 0.0030122434  },
        { 628 , 0.0028394347  },
        { 629 , 0.0027917746  },
        { 630 , 0.0027441146  },
        { 631 , 0.0027664110  },
        { 632 , 0.0027887074  },
        { 633 , 0.0026410223  },
        { 634 , 0.0024933372  },
        { 635 , 0.0024518637  },
        { 636 , 0.0024103903  },
        { 637 , 0.0025039227  },
        { 638 , 0.0025974550  },
        { 639 , 0.0022707526  },
        { 640 , 0.0019440503  },
        { 641 , 0.0017330255  },
        { 642 , 0.0015220008  },
        { 643 , 0.0015867857  },
        { 644 , 0.0016515707  },
        { 645 , 0.0017203518  },
        { 646 , 0.0017891328  },
        { 647 , 0.0018323435  },
        { 648 , 0.0018755541  },
        { 649 , 0.0015840185  },
        { 650 , 0.0012924828  },
        { 651 , 0.0015366886  },
        { 652 , 0.0017808944  },
        { 653 , 0.0017614276  },
        { 654 , 0.0017419608  },
        { 655 , 0.0014175192  },
        { 656 , 0.0010930776  },
        { 657 , 0.0010472244  },
        { 658 , 0.0010013712  },
        { 659 , 0.0012221707  },
        { 660 , 0.0014429701  },
        { 661 , 0.0014804600  },
        { 662 , 0.0015179499  },
        { 663 , 0.0016186097  },
        { 664 , 0.0017192695  },
        { 665 , 0.0014060819  },
        { 666 , 0.0010928944  },
        { 667 , 0.0013672745  },
        { 668 , 0.0016416547  },
        { 669 , 0.0015860668  },
        { 670 , 0.0015304790  },
        { 671 , 0.0014794227  },
        { 672 , 0.0014283664  },
        { 673 , 0.0015946891  },
        { 674 , 0.0017610119  },
        { 675 , 0.0016470766  },
        { 676 , 0.0015331413  },
        { 677 , 0.0015498049  },
        { 678 , 0.0015664686  },
        { 679 , 0.0014756692  },
        { 680 , 0.0013848698  },
        { 681 , 0.0014111267  },
        { 682 , 0.0014373836  },
        { 683 , 0.0014086630  },
        { 684 , 0.0013799424  },
        { 685 , 0.0012446914  },
        { 686 , 0.0011094404  },
        { 687 , 0.0010076393  },
        { 688 , 0.0009058381  },
        { 689 , 0.0006943493  },
        { 690 , 0.0004828605  },
        { 691 , 0.0002414302  }
      };            

      static const double CBERS4_MUXB7_SRF[][2] = 
      {
        { 405 , 0.0000730916  },
        { 406 , 0.0001461832  },
        { 407 , 0.0000730916  },
        { 408 , 0.0000000000  },
        { 409 , 0.0000675381  },
        { 410 , 0.0001350763  },
        { 411 , 0.0000675381  },
        { 412 , 0.0000000000  },
        { 413 , 0.0000000000  },
        { 414 , 0.0000000000  },
        { 415 , 0.0000000000  },
        { 416 , 0.0000000000  },
        { 417 , 0.0000582542  },
        { 418 , 0.0001165083  },
        { 419 , 0.0000582542  },
        { 420 , 0.0000000000  },
        { 421 , 0.0000000000  },
        { 422 , 0.0000000000  },
        { 423 , 0.0000000000  },
        { 424 , 0.0000000000  },
        { 425 , 0.0000000000  },
        { 426 , 0.0000000000  },
        { 427 , 0.0000501300  },
        { 428 , 0.0001002601  },
        { 429 , 0.0000501300  },
        { 430 , 0.0000000000  },
        { 431 , 0.0000000000  },
        { 432 , 0.0000000000  },
        { 433 , 0.0000000000  },
        { 434 , 0.0000000000  },
        { 435 , 0.0000000000  },
        { 436 , 0.0000000000  },
        { 437 , 0.0000000000  },
        { 438 , 0.0000000000  },
        { 439 , 0.0000000000  },
        { 440 , 0.0000000000  },
        { 441 , 0.0000000000  },
        { 442 , 0.0000000000  },
        { 443 , 0.0000000000  },
        { 444 , 0.0000000000  },
        { 445 , 0.0002403260  },
        { 446 , 0.0004806521  },
        { 447 , 0.0002403260  },
        { 448 , 0.0000000000  },
        { 449 , 0.0000384927  },
        { 450 , 0.0000769854  },
        { 451 , 0.0000384927  },
        { 452 , 0.0000000000  },
        { 453 , 0.0000739769  },
        { 454 , 0.0001479538  },
        { 455 , 0.0001840482  },
        { 456 , 0.0002201425  },
        { 457 , 0.0002535391  },
        { 458 , 0.0002869356  },
        { 459 , 0.0002142790  },
        { 460 , 0.0001416224  },
        { 461 , 0.0002427713  },
        { 462 , 0.0003439202  },
        { 463 , 0.0002716532  },
        { 464 , 0.0001993862  },
        { 465 , 0.0000996931  },
        { 466 , 0.0000000000  },
        { 467 , 0.0000000000  },
        { 468 , 0.0000000000  },
        { 469 , 0.0000633981  },
        { 470 , 0.0001267961  },
        { 471 , 0.0000633981  },
        { 472 , 0.0000000000  },
        { 473 , 0.0000000000  },
        { 474 , 0.0000000000  },
        { 475 , 0.0009246824  },
        { 476 , 0.0018493649  },
        { 477 , 0.0023444291  },
        { 478 , 0.0028394933  },
        { 479 , 0.0015351201  },
        { 480 , 0.0002307469  },
        { 481 , 0.0001434724  },
        { 482 , 0.0000561980  },
        { 483 , 0.0001672219  },
        { 484 , 0.0002782457  },
        { 485 , 0.0002491100  },
        { 486 , 0.0002199742  },
        { 487 , 0.0001638526  },
        { 488 , 0.0001077311  },
        { 489 , 0.0001608143  },
        { 490 , 0.0002138976  },
        { 491 , 0.0002648340  },
        { 492 , 0.0003157703  },
        { 493 , 0.0001841249  },
        { 494 , 0.0000524795  },
        { 495 , 0.0001298299  },
        { 496 , 0.0002071803  },
        { 497 , 0.0001799907  },
        { 498 , 0.0001528012  },
        { 499 , 0.0001526740  },
        { 500 , 0.0001525469  },
        { 501 , 0.0001507551  },
        { 502 , 0.0001489634  },
        { 503 , 0.0001488224  },
        { 504 , 0.0001486814  },
        { 505 , 0.0001230391  },
        { 506 , 0.0000973968  },
        { 507 , 0.0001465530  },
        { 508 , 0.0001957091  },
        { 509 , 0.0002189008  },
        { 510 , 0.0002420925  },
        { 511 , 0.0001687120  },
        { 512 , 0.0000953315  },
        { 513 , 0.0001420184  },
        { 514 , 0.0001887054  },
        { 515 , 0.0001880658  },
        { 516 , 0.0001874262  },
        { 517 , 0.0002335300  },
        { 518 , 0.0002796339  },
        { 519 , 0.0003713187  },
        { 520 , 0.0004630036  },
        { 521 , 0.0004615195  },
        { 522 , 0.0004600355  },
        { 523 , 0.0003912080  },
        { 524 , 0.0003223804  },
        { 525 , 0.0002979514  },
        { 526 , 0.0002735224  },
        { 527 , 0.0003394760  },
        { 528 , 0.0004054297  },
        { 529 , 0.0003601558  },
        { 530 , 0.0003148820  },
        { 531 , 0.0003369081  },
        { 532 , 0.0003589342  },
        { 533 , 0.0003586098  },
        { 534 , 0.0003582855  },
        { 535 , 0.0004700673  },
        { 536 , 0.0005818491  },
        { 537 , 0.0005622458  },
        { 538 , 0.0005426424  },
        { 539 , 0.0005602521  },
        { 540 , 0.0005778618  },
        { 541 , 0.0005828908  },
        { 542 , 0.0005879198  },
        { 543 , 0.0005644090  },
        { 544 , 0.0005408981  },
        { 545 , 0.0005412986  },
        { 546 , 0.0005416991  },
        { 547 , 0.0004713849  },
        { 548 , 0.0004010707  },
        { 549 , 0.0004438201  },
        { 550 , 0.0004865696  },
        { 551 , 0.0005972941  },
        { 552 , 0.0007080186  },
        { 553 , 0.0006878121  },
        { 554 , 0.0006676057  },
        { 555 , 0.0006205503  },
        { 556 , 0.0005734950  },
        { 557 , 0.0006162899  },
        { 558 , 0.0006590848  },
        { 559 , 0.0005507939  },
        { 560 , 0.0004425030  },
        { 561 , 0.0005693422  },
        { 562 , 0.0006961814  },
        { 563 , 0.0005449036  },
        { 564 , 0.0003936258  },
        { 565 , 0.0004576503  },
        { 566 , 0.0005216747  },
        { 567 , 0.0005669895  },
        { 568 , 0.0006123043  },
        { 569 , 0.0006545344  },
        { 570 , 0.0006967646  },
        { 571 , 0.0006749822  },
        { 572 , 0.0006531998  },
        { 573 , 0.0006564099  },
        { 574 , 0.0006596199  },
        { 575 , 0.0006103140  },
        { 576 , 0.0005610081  },
        { 577 , 0.0006074862  },
        { 578 , 0.0006539642  },
        { 579 , 0.0005882451  },
        { 580 , 0.0005225259  },
        { 581 , 0.0005253783  },
        { 582 , 0.0005282307  },
        { 583 , 0.0005463657  },
        { 584 , 0.0005645007  },
        { 585 , 0.0005227823  },
        { 586 , 0.0004810640  },
        { 587 , 0.0004153465  },
        { 588 , 0.0003496290  },
        { 589 , 0.0004405201  },
        { 590 , 0.0005314113  },
        { 591 , 0.0004656550  },
        { 592 , 0.0003998987  },
        { 593 , 0.0004196342  },
        { 594 , 0.0004393697  },
        { 595 , 0.0003964666  },
        { 596 , 0.0003535635  },
        { 597 , 0.0002862247  },
        { 598 , 0.0002188859  },
        { 599 , 0.0001983922  },
        { 600 , 0.0001778985  },
        { 601 , 0.0001769870  },
        { 602 , 0.0001760756  },
        { 603 , 0.0002224941  },
        { 604 , 0.0002689126  },
        { 605 , 0.0002659952  },
        { 606 , 0.0002630778  },
        { 607 , 0.0003115562  },
        { 608 , 0.0003600346  },
        { 609 , 0.0006474077  },
        { 610 , 0.0009347808  },
        { 611 , 0.0018300226  },
        { 612 , 0.0027252644  },
        { 613 , 0.0038130739  },
        { 614 , 0.0049008834  },
        { 615 , 0.0068015597  },
        { 616 , 0.0087022359  },
        { 617 , 0.0110985823  },
        { 618 , 0.0134949286  },
        { 619 , 0.0167772573  },
        { 620 , 0.0200595860  },
        { 621 , 0.0267631435  },
        { 622 , 0.0334667010  },
        { 623 , 0.0463827471  },
        { 624 , 0.0592987932  },
        { 625 , 0.0875800325  },
        { 626 , 0.1158612718  },
        { 627 , 0.1711471779  },
        { 628 , 0.2264330839  },
        { 629 , 0.3307878966  },
        { 630 , 0.4351427093  },
        { 631 , 0.5502363726  },
        { 632 , 0.6653300360  },
        { 633 , 0.7573173453  },
        { 634 , 0.8493046547  },
        { 635 , 0.8805380370  },
        { 636 , 0.9117714192  },
        { 637 , 0.9245153131  },
        { 638 , 0.9372592070  },
        { 639 , 0.9489334486  },
        { 640 , 0.9606076901  },
        { 641 , 0.9477576695  },
        { 642 , 0.9349076488  },
        { 643 , 0.9377774478  },
        { 644 , 0.9406472468  },
        { 645 , 0.9492515788  },
        { 646 , 0.9578559108  },
        { 647 , 0.9688161267  },
        { 648 , 0.9797763426  },
        { 649 , 0.9754004750  },
        { 650 , 0.9710246074  },
        { 651 , 0.9790642940  },
        { 652 , 0.9871039805  },
        { 653 , 0.9835478195  },
        { 654 , 0.9799916585  },
        { 655 , 0.9899958293  },
        { 656 , 1.0000000000  },
        { 657 , 0.9983509122  },
        { 658 , 0.9967018245  },
        { 659 , 0.9903930162  },
        { 660 , 0.9840842079  },
        { 661 , 0.9778969303  },
        { 662 , 0.9717096528  },
        { 663 , 0.9731648434  },
        { 664 , 0.9746200339  },
        { 665 , 0.9471679769  },
        { 666 , 0.9197159200  },
        { 667 , 0.9190707407  },
        { 668 , 0.9184255614  },
        { 669 , 0.9044446037  },
        { 670 , 0.8904636460  },
        { 671 , 0.8917066605  },
        { 672 , 0.8929496749  },
        { 673 , 0.8900062837  },
        { 674 , 0.8870628925  },
        { 675 , 0.8968215830  },
        { 676 , 0.9065802735  },
        { 677 , 0.9063717706  },
        { 678 , 0.9061632677  },
        { 679 , 0.9151610301  },
        { 680 , 0.9241587926  },
        { 681 , 0.9343593118  },
        { 682 , 0.9445598309  },
        { 683 , 0.9339372335  },
        { 684 , 0.9233146361  },
        { 685 , 0.8969815127  },
        { 686 , 0.8706483892  },
        { 687 , 0.8156789802  },
        { 688 , 0.7607095712  },
        { 689 , 0.6609391675  },
        { 690 , 0.5611687639  },
        { 691 , 0.4676477290  },
        { 692 , 0.3741266940  },
        { 693 , 0.2933220123  },
        { 694 , 0.2125173306  },
        { 695 , 0.1645987898  },
        { 696 , 0.1166802491  },
        { 697 , 0.0895750906  },
        { 698 , 0.0624699321  },
        { 699 , 0.0491561681  },
        { 700 , 0.0358424040  },
        { 701 , 0.0287112113  },
        { 702 , 0.0215800187  },
        { 703 , 0.0177826795  },
        { 704 , 0.0139853402  },
        { 705 , 0.0117016724  },
        { 706 , 0.0094180045  },
        { 707 , 0.0087090409  },
        { 708 , 0.0080000773  },
        { 709 , 0.0065178001  },
        { 710 , 0.0050355229  },
        { 711 , 0.0049195497  },
        { 712 , 0.0048035764  },
        { 713 , 0.0045671403  },
        { 714 , 0.0043307042  },
        { 715 , 0.0043393498  },
        { 716 , 0.0043479954  },
        { 717 , 0.0040195543  },
        { 718 , 0.0036911133  },
        { 719 , 0.0034662032  },
        { 720 , 0.0032412932  },
        { 721 , 0.0025613006  },
        { 722 , 0.0018813079  },
        { 723 , 0.0014513077  },
        { 724 , 0.0010213074  },
        { 725 , 0.0008171475  },
        { 726 , 0.0006129875  },
        { 727 , 0.0004357622  },
        { 728 , 0.0002585369  },
        { 729 , 0.0002594938  },
        { 730 , 0.0002604506  },
        { 731 , 0.0002892635  },
        { 732 , 0.0003180765  },
        { 733 , 0.0002631083  },
        { 734 , 0.0002081401  },
        { 735 , 0.0002969162  },
        { 736 , 0.0003856923  },
        { 737 , 0.0003018657  },
        { 738 , 0.0002180391  },
        { 739 , 0.0002224598  },
        { 740 , 0.0002268805  },
        { 741 , 0.0002554651  },
        { 742 , 0.0002840496  },
        { 743 , 0.0002590500  },
        { 744 , 0.0002340503  },
        { 745 , 0.0003215005  },
        { 746 , 0.0004089507  },
        { 747 , 0.0003802538  },
        { 748 , 0.0003515569  },
        { 749 , 0.0003853473  },
        { 750 , 0.0004191378  }
      };            

      static const double CBERS4_MUXB8_SRF[][2] = 
      {
        { 407 , 0.0002520625  },
        { 408 , 0.0005041250  },
        { 409 , 0.0002520625  },
        { 410 , 0.0000000000  },
        { 411 , 0.0000000000  },
        { 412 , 0.0000000000  },
        { 413 , 0.0001877856  },
        { 414 , 0.0003755711  },
        { 415 , 0.0005492769  },
        { 416 , 0.0007229827  },
        { 417 , 0.0010977070  },
        { 418 , 0.0014724314  },
        { 419 , 0.0012775072  },
        { 420 , 0.0010825831  },
        { 421 , 0.0006071618  },
        { 422 , 0.0001317405  },
        { 423 , 0.0005757241  },
        { 424 , 0.0010197076  },
        { 425 , 0.0005098538  },
        { 426 , 0.0000000000  },
        { 427 , 0.0002715184  },
        { 428 , 0.0005430369  },
        { 429 , 0.0004780726  },
        { 430 , 0.0004131084  },
        { 431 , 0.0004920772  },
        { 432 , 0.0005710461  },
        { 433 , 0.0002855231  },
        { 434 , 0.0000000000  },
        { 435 , 0.0001620862  },
        { 436 , 0.0003241724  },
        { 437 , 0.0001620862  },
        { 438 , 0.0000000000  },
        { 439 , 0.0000516541  },
        { 440 , 0.0001033081  },
        { 441 , 0.0000516541  },
        { 442 , 0.0000000000  },
        { 443 , 0.0001711466  },
        { 444 , 0.0003422932  },
        { 445 , 0.0001952517  },
        { 446 , 0.0000482101  },
        { 447 , 0.0000955228  },
        { 448 , 0.0001428355  },
        { 449 , 0.0003957313  },
        { 450 , 0.0006486271  },
        { 451 , 0.0007359283  },
        { 452 , 0.0008232295  },
        { 453 , 0.0007900346  },
        { 454 , 0.0007568398  },
        { 455 , 0.0005109037  },
        { 456 , 0.0002649676  },
        { 457 , 0.0003915045  },
        { 458 , 0.0005180415  },
        { 459 , 0.0005573245  },
        { 460 , 0.0005966075  },
        { 461 , 0.0009399243  },
        { 462 , 0.0012832410  },
        { 463 , 0.0012615815  },
        { 464 , 0.0012399220  },
        { 465 , 0.0009587511  },
        { 466 , 0.0006775802  },
        { 467 , 0.0005327905  },
        { 468 , 0.0003880008  },
        { 469 , 0.0004610752  },
        { 470 , 0.0005341496  },
        { 471 , 0.0004156967  },
        { 472 , 0.0002972438  },
        { 473 , 0.0004568545  },
        { 474 , 0.0006164652  },
        { 475 , 0.0007749598  },
        { 476 , 0.0009334544  },
        { 477 , 0.0009200912  },
        { 478 , 0.0009067279  },
        { 479 , 0.0007831698  },
        { 480 , 0.0006596116  },
        { 481 , 0.0006003696  },
        { 482 , 0.0005411276  },
        { 483 , 0.0006054655  },
        { 484 , 0.0006698035  },
        { 485 , 0.0008975273  },
        { 486 , 0.0011252512  },
        { 487 , 0.0011299192  },
        { 488 , 0.0011345872  },
        { 489 , 0.0011143772  },
        { 490 , 0.0010941671  },
        { 491 , 0.0008954784  },
        { 492 , 0.0006967896  },
        { 493 , 0.0006958039  },
        { 494 , 0.0006948182  },
        { 495 , 0.0006591166  },
        { 496 , 0.0006234150  },
        { 497 , 0.0007255147  },
        { 498 , 0.0008276144  },
        { 499 , 0.0007351216  },
        { 500 , 0.0006426288  },
        { 501 , 0.0004707270  },
        { 502 , 0.0002988252  },
        { 503 , 0.0002985423  },
        { 504 , 0.0002982595  },
        { 505 , 0.0001784369  },
        { 506 , 0.0000586143  },
        { 507 , 0.0002059765  },
        { 508 , 0.0003533386  },
        { 509 , 0.0002786548  },
        { 510 , 0.0002039709  },
        { 511 , 0.0002310710  },
        { 512 , 0.0002581711  },
        { 513 , 0.0002994326  },
        { 514 , 0.0003406940  },
        { 515 , 0.0002549431  },
        { 516 , 0.0001691923  },
        { 517 , 0.0002949539  },
        { 518 , 0.0004207156  },
        { 519 , 0.0002242898  },
        { 520 , 0.0000278640  },
        { 521 , 0.0002769429  },
        { 522 , 0.0005260219  },
        { 523 , 0.0003877326  },
        { 524 , 0.0002494434  },
        { 525 , 0.0001521564  },
        { 526 , 0.0000548694  },
        { 527 , 0.0001629855  },
        { 528 , 0.0002711015  },
        { 529 , 0.0002032289  },
        { 530 , 0.0001353563  },
        { 531 , 0.0001621825  },
        { 532 , 0.0001890087  },
        { 533 , 0.0001888379  },
        { 534 , 0.0001886671  },
        { 535 , 0.0002020757  },
        { 536 , 0.0002154844  },
        { 537 , 0.0002438119  },
        { 538 , 0.0002721394  },
        { 539 , 0.0003634529  },
        { 540 , 0.0004547664  },
        { 541 , 0.0003770744  },
        { 542 , 0.0002993824  },
        { 543 , 0.0002310706  },
        { 544 , 0.0001627587  },
        { 545 , 0.0002036292  },
        { 546 , 0.0002444997  },
        { 547 , 0.0003233899  },
        { 548 , 0.0004022801  },
        { 549 , 0.0004540318  },
        { 550 , 0.0005057836  },
        { 551 , 0.0003993609  },
        { 552 , 0.0002929383  },
        { 553 , 0.0002536082  },
        { 554 , 0.0002142780  },
        { 555 , 0.0002929807  },
        { 556 , 0.0003716834  },
        { 557 , 0.0003444990  },
        { 558 , 0.0003173147  },
        { 559 , 0.0003184388  },
        { 560 , 0.0003195629  },
        { 561 , 0.0003692656  },
        { 562 , 0.0004189684  },
        { 563 , 0.0004332114  },
        { 564 , 0.0004474544  },
        { 565 , 0.0003283768  },
        { 566 , 0.0002092991  },
        { 567 , 0.0003678570  },
        { 568 , 0.0005264148  },
        { 569 , 0.0004466596  },
        { 570 , 0.0003669044  },
        { 571 , 0.0002358658  },
        { 572 , 0.0001048271  },
        { 573 , 0.0001582710  },
        { 574 , 0.0002117149  },
        { 575 , 0.0002486966  },
        { 576 , 0.0002856784  },
        { 577 , 0.0003265013  },
        { 578 , 0.0003673243  },
        { 579 , 0.0003408926  },
        { 580 , 0.0003144609  },
        { 581 , 0.0003956511  },
        { 582 , 0.0004768412  },
        { 583 , 0.0003690828  },
        { 584 , 0.0002613244  },
        { 585 , 0.0003280545  },
        { 586 , 0.0003947846  },
        { 587 , 0.0003025973  },
        { 588 , 0.0002104100  },
        { 589 , 0.0002651091  },
        { 590 , 0.0003198082  },
        { 591 , 0.0003470863  },
        { 592 , 0.0003743643  },
        { 593 , 0.0003193905  },
        { 594 , 0.0002644168  },
        { 595 , 0.0003449862  },
        { 596 , 0.0004255556  },
        { 597 , 0.0004630601  },
        { 598 , 0.0005005646  },
        { 599 , 0.0004108738  },
        { 600 , 0.0003211829  },
        { 601 , 0.0003592738  },
        { 602 , 0.0003973647  },
        { 603 , 0.0004818920  },
        { 604 , 0.0005664193  },
        { 605 , 0.0004547259  },
        { 606 , 0.0003430325  },
        { 607 , 0.0003746464  },
        { 608 , 0.0004062603  },
        { 609 , 0.0003772558  },
        { 610 , 0.0003482513  },
        { 611 , 0.0004833228  },
        { 612 , 0.0006183943  },
        { 613 , 0.0005549803  },
        { 614 , 0.0004915662  },
        { 615 , 0.0003800673  },
        { 616 , 0.0002685685  },
        { 617 , 0.0004873874  },
        { 618 , 0.0007062063  },
        { 619 , 0.0007034080  },
        { 620 , 0.0007006097  },
        { 621 , 0.0005689271  },
        { 622 , 0.0004372444  },
        { 623 , 0.0005475306  },
        { 624 , 0.0006578168  },
        { 625 , 0.0006639972  },
        { 626 , 0.0006701775  },
        { 627 , 0.0007300729  },
        { 628 , 0.0007899683  },
        { 629 , 0.0008114108  },
        { 630 , 0.0008328534  },
        { 631 , 0.0009000767  },
        { 632 , 0.0009673001  },
        { 633 , 0.0010371176  },
        { 634 , 0.0011069352  },
        { 635 , 0.0010963362  },
        { 636 , 0.0010857373  },
        { 637 , 0.0010913056  },
        { 638 , 0.0010968739  },
        { 639 , 0.0011113734  },
        { 640 , 0.0011258729  },
        { 641 , 0.0010976384  },
        { 642 , 0.0010694039  },
        { 643 , 0.0011090630  },
        { 644 , 0.0011487220  },
        { 645 , 0.0011410079  },
        { 646 , 0.0011332938  },
        { 647 , 0.0011596059  },
        { 648 , 0.0011859181  },
        { 649 , 0.0011624893  },
        { 650 , 0.0011390606  },
        { 651 , 0.0011517627  },
        { 652 , 0.0011644648  },
        { 653 , 0.0011880281  },
        { 654 , 0.0012115913  },
        { 655 , 0.0012212536  },
        { 656 , 0.0012309159  },
        { 657 , 0.0012684143  },
        { 658 , 0.0013059127  },
        { 659 , 0.0013074999  },
        { 660 , 0.0013090871  },
        { 661 , 0.0012880141  },
        { 662 , 0.0012669412  },
        { 663 , 0.0012519924  },
        { 664 , 0.0012370435  },
        { 665 , 0.0011886299  },
        { 666 , 0.0011402163  },
        { 667 , 0.0011068185  },
        { 668 , 0.0010734208  },
        { 669 , 0.0010517904  },
        { 670 , 0.0010301601  },
        { 671 , 0.0010118180  },
        { 672 , 0.0009934760  },
        { 673 , 0.0009678311  },
        { 674 , 0.0009421861  },
        { 675 , 0.0009281004  },
        { 676 , 0.0009140146  },
        { 677 , 0.0008987089  },
        { 678 , 0.0008834032  },
        { 679 , 0.0008920395  },
        { 680 , 0.0009006758  },
        { 681 , 0.0009040050  },
        { 682 , 0.0009073342  },
        { 683 , 0.0008899378  },
        { 684 , 0.0008725414  },
        { 685 , 0.0008772150  },
        { 686 , 0.0008818885  },
        { 687 , 0.0008582437  },
        { 688 , 0.0008345989  },
        { 689 , 0.0008131173  },
        { 690 , 0.0007916356  },
        { 691 , 0.0007797842  },
        { 692 , 0.0007679327  },
        { 693 , 0.0007534624  },
        { 694 , 0.0007389921  },
        { 695 , 0.0007073292  },
        { 696 , 0.0006756664  },
        { 697 , 0.0006854750  },
        { 698 , 0.0006952837  },
        { 699 , 0.0007134538  },
        { 700 , 0.0007316239  },
        { 701 , 0.0006486983  },
        { 702 , 0.0005657727  },
        { 703 , 0.0006075973  },
        { 704 , 0.0006494219  },
        { 705 , 0.0006822617  },
        { 706 , 0.0007151016  },
        { 707 , 0.0006739336  },
        { 708 , 0.0006327656  },
        { 709 , 0.0007222432  },
        { 710 , 0.0008117208  },
        { 711 , 0.0008367063  },
        { 712 , 0.0008616917  },
        { 713 , 0.0008091736  },
        { 714 , 0.0007566555  },
        { 715 , 0.0008191787  },
        { 716 , 0.0008817019  },
        { 717 , 0.0008965117  },
        { 718 , 0.0009113215  },
        { 719 , 0.0009508233  },
        { 720 , 0.0009903250  },
        { 721 , 0.0010016679  },
        { 722 , 0.0010130108  },
        { 723 , 0.0010135770  },
        { 724 , 0.0010141432  },
        { 725 , 0.0010604239  },
        { 726 , 0.0011067047  },
        { 727 , 0.0011134761  },
        { 728 , 0.0011202475  },
        { 729 , 0.0012497867  },
        { 730 , 0.0013793259  },
        { 731 , 0.0021253232  },
        { 732 , 0.0028713206  },
        { 733 , 0.0034554879  },
        { 734 , 0.0040396553  },
        { 735 , 0.0042746416  },
        { 736 , 0.0045096279  },
        { 737 , 0.0059945253  },
        { 738 , 0.0074794227  },
        { 739 , 0.0085015019  },
        { 740 , 0.0095235811  },
        { 741 , 0.0110183299  },
        { 742 , 0.0125130787  },
        { 743 , 0.0146901497  },
        { 744 , 0.0168672206  },
        { 745 , 0.0193327817  },
        { 746 , 0.0217983428  },
        { 747 , 0.0238754776  },
        { 748 , 0.0259526124  },
        { 749 , 0.0297683494  },
        { 750 , 0.0335840863  },
        { 751 , 0.0377510202  },
        { 752 , 0.0419179541  },
        { 753 , 0.0479866675  },
        { 754 , 0.0540553809  },
        { 755 , 0.0640325911  },
        { 756 , 0.0740098014  },
        { 757 , 0.0866366210  },
        { 758 , 0.0992634406  },
        { 759 , 0.1192702221  },
        { 760 , 0.1392770037  },
        { 761 , 0.1706502568  },
        { 762 , 0.2020235100  },
        { 763 , 0.2555851432  },
        { 764 , 0.3091467763  },
        { 765 , 0.3809128951  },
        { 766 , 0.4526790138  },
        { 767 , 0.5463936978  },
        { 768 , 0.6401083819  },
        { 769 , 0.7356179421  },
        { 770 , 0.8311275023  },
        { 771 , 0.8927650470  },
        { 772 , 0.9544025917  },
        { 773 , 0.9772012958  },
        { 774 , 1.0000000000  },
        { 775 , 0.9912554688  },
        { 776 , 0.9825109376  },
        { 777 , 0.9704397645  },
        { 778 , 0.9583685913  },
        { 779 , 0.9385729389  },
        { 780 , 0.9187772865  },
        { 781 , 0.9148533801  },
        { 782 , 0.9109294736  },
        { 783 , 0.8806939570  },
        { 784 , 0.8504584403  },
        { 785 , 0.8688212367  },
        { 786 , 0.8871840330  },
        { 787 , 0.8786906037  },
        { 788 , 0.8701971744  },
        { 789 , 0.8815710088  },
        { 790 , 0.8929448431  },
        { 791 , 0.8833767069  },
        { 792 , 0.8738085707  },
        { 793 , 0.9013163887  },
        { 794 , 0.9288242067  },
        { 795 , 0.9210035398  },
        { 796 , 0.9131828728  },
        { 797 , 0.9042637048  },
        { 798 , 0.8953445368  },
        { 799 , 0.8951113226  },
        { 800 , 0.8948781084  },
        { 801 , 0.9064692998  },
        { 802 , 0.9180604912  },
        { 803 , 0.9253564569  },
        { 804 , 0.9326524226  },
        { 805 , 0.9222030268  },
        { 806 , 0.9117536310  },
        { 807 , 0.8993698581  },
        { 808 , 0.8869860853  },
        { 809 , 0.8803205249  },
        { 810 , 0.8736549646  },
        { 811 , 0.8727665928  },
        { 812 , 0.8718782210  },
        { 813 , 0.8714667029  },
        { 814 , 0.8710551849  },
        { 815 , 0.8742071805  },
        { 816 , 0.8773591762  },
        { 817 , 0.8592877413  },
        { 818 , 0.8412163064  },
        { 819 , 0.8506686998  },
        { 820 , 0.8601210933  },
        { 821 , 0.8573355695  },
        { 822 , 0.8545500457  },
        { 823 , 0.8630622162  },
        { 824 , 0.8715743868  },
        { 825 , 0.8849589713  },
        { 826 , 0.8983435558  },
        { 827 , 0.8779408009  },
        { 828 , 0.8575380461  },
        { 829 , 0.8618316080  },
        { 830 , 0.8661251699  },
        { 831 , 0.8627130047  },
        { 832 , 0.8593008395  },
        { 833 , 0.8659228118  },
        { 834 , 0.8725447840  },
        { 835 , 0.8740134203  },
        { 836 , 0.8754820566  },
        { 837 , 0.8577014740  },
        { 838 , 0.8399208914  },
        { 839 , 0.8341294275  },
        { 840 , 0.8283379636  },
        { 841 , 0.8273919944  },
        { 842 , 0.8264460252  },
        { 843 , 0.8325915386  },
        { 844 , 0.8387370521  },
        { 845 , 0.8324376887  },
        { 846 , 0.8261383252  },
        { 847 , 0.8222648410  },
        { 848 , 0.8183913568  },
        { 849 , 0.8192727667  },
        { 850 , 0.8201541767  },
        { 851 , 0.8021773027  },
        { 852 , 0.7842004286  },
        { 853 , 0.7736202348  },
        { 854 , 0.7630400411  },
        { 855 , 0.7756645144  },
        { 856 , 0.7882889876  },
        { 857 , 0.7803992835  },
        { 858 , 0.7725095794  },
        { 859 , 0.7529858519  },
        { 860 , 0.7334621244  },
        { 861 , 0.7301641781  },
        { 862 , 0.7268662317  },
        { 863 , 0.7092277719  },
        { 864 , 0.6915893120  },
        { 865 , 0.6870818366  },
        { 866 , 0.6825743611  },
        { 867 , 0.6637813781  },
        { 868 , 0.6449883951  },
        { 869 , 0.6335219749  },
        { 870 , 0.6220555547  },
        { 871 , 0.6006326692  },
        { 872 , 0.5792097837  },
        { 873 , 0.5598527157  },
        { 874 , 0.5404956477  },
        { 875 , 0.5023819139  },
        { 876 , 0.4642681802  },
        { 877 , 0.4119742876  },
        { 878 , 0.3596803951  },
        { 879 , 0.3209956347  },
        { 880 , 0.2823108742  },
        { 881 , 0.2421723105  },
        { 882 , 0.2020337469  },
        { 883 , 0.1763732458  },
        { 884 , 0.1507127447  },
        { 885 , 0.1258409697  },
        { 886 , 0.1009691946  },
        { 887 , 0.0864324777  },
        { 888 , 0.0718957608  },
        { 889 , 0.0629060508  },
        { 890 , 0.0539163408  },
        { 891 , 0.0446591243  },
        { 892 , 0.0354019077  },
        { 893 , 0.0315467118  },
        { 894 , 0.0276915158  },
        { 895 , 0.0247977046  },
        { 896 , 0.0219038934  },
        { 897 , 0.0187226125  },
        { 898 , 0.0155413315  },
        { 899 , 0.0138250384  },
        { 900 , 0.0121087453  },
        { 901 , 0.0104114340  },
        { 902 , 0.0087141227  },
        { 903 , 0.0082140179  },
        { 904 , 0.0077139131  },
        { 905 , 0.0069977977  },
        { 906 , 0.0062816822  },
        { 907 , 0.0046733149  },
        { 908 , 0.0030649475  },
        { 909 , 0.0028929349  },
        { 910 , 0.0027209222  },
        { 911 , 0.0025294824  },
        { 912 , 0.0023380425  },
        { 913 , 0.0023010487  },
        { 914 , 0.0022640549  },
        { 915 , 0.0023222069  },
        { 916 , 0.0023803589  },
        { 917 , 0.0023183543  },
        { 918 , 0.0022563496  },
        { 919 , 0.0022057582  },
        { 920 , 0.0021551669  },
        { 921 , 0.0021736855  },
        { 922 , 0.0021922041  },
        { 923 , 0.0021730662  },
        { 924 , 0.0021539284  },
        { 925 , 0.0021142671  },
        { 926 , 0.0020746057  },
        { 927 , 0.0021553647  },
        { 928 , 0.0022361237  },
        { 929 , 0.0019363280  },
        { 930 , 0.0016365322  },
        { 931 , 0.0018496103  },
        { 932 , 0.0020626883  },
        { 933 , 0.0019023388  },
        { 934 , 0.0017419893  },
        { 935 , 0.0017315593  },
        { 936 , 0.0017211294  },
        { 937 , 0.0016604520  },
        { 938 , 0.0015997746  },
        { 939 , 0.0017267970  },
        { 940 , 0.0018538194  },
        { 941 , 0.0015674353  },
        { 942 , 0.0012810512  },
        { 943 , 0.0012810512  },
        { 944 , 0.0012810512  },
        { 945 , 0.0012810512  },
        { 946 , 0.0012810512  },
        { 947 , 0.0012810512  },
        { 948 , 0.0012810512  },
        { 949 , 0.0012810512  },
        { 950 , 0.0012810512  },
        { 951 , 0.0012810512  },
        { 952 , 0.0012810512  },
        { 953 , 0.0012810512  },
        { 954 , 0.0012810512  },
        { 955 , 0.0012810512  },
        { 956 , 0.0012810512  },
        { 957 , 0.0012810512  },
        { 958 , 0.0012810512  },
        { 959 , 0.0012810512  },
        { 960 , 0.0012810512  },
        { 961 , 0.0012810512  },
        { 962 , 0.0012810512  },
        { 963 , 0.0012810512  },
        { 964 , 0.0012810512  },
        { 965 , 0.0012810512  },
        { 966 , 0.0012810512  },
        { 967 , 0.0012810512  },
        { 968 , 0.0012810512  },
        { 969 , 0.0012810512  },
        { 970 , 0.0012810512  },
        { 971 , 0.0012810512  },
        { 972 , 0.0012810512  },
        { 973 , 0.0012810512  },
        { 974 , 0.0012810512  },
        { 975 , 0.0012810512  },
        { 976 , 0.0012810512  },
        { 977 , 0.0012810512  },
        { 978 , 0.0012810512  },
        { 979 , 0.0012810512  },
        { 980 , 0.0012810512  },
        { 981 , 0.0012810512  },
        { 982 , 0.0012810512  },
        { 983 , 0.0012810512  },
        { 984 , 0.0012810512  },
        { 985 , 0.0012810512  },
        { 986 , 0.0012810512  },
        { 987 , 0.0012810512  },
        { 988 , 0.0012810512  },
        { 989 , 0.0012810512  },
        { 990 , 0.0012810512  },
        { 991 , 0.0012810512  },
        { 992 , 0.0012810512  },
        { 993 , 0.0012810512  },
        { 994 , 0.0012810512  },
        { 995 , 0.0012810512  },
        { 996 , 0.0012810512  },
        { 997 , 0.0012810512  },
        { 998 , 0.0012810512  },
        { 999 , 0.0012810512  }
      };                  

      static const double CBERS04A_WPMPAN_SRF[][2] = 
      {
        {	350	,	0.0000000000	},
        {	351	,	0.0001480000	},
        {	352	,	0.0002820000	},
        {	353	,	0.0004030000	},
        {	354	,	0.0005120000	},
        {	355	,	0.0006110000	},
        {	356	,	0.0006870000	},
        {	357	,	0.0007550000	},
        {	358	,	0.0008180000	},
        {	359	,	0.0008750000	},
        {	360	,	0.0009260000	},
        {	361	,	0.0007700000	},
        {	362	,	0.0006230000	},
        {	363	,	0.0004860000	},
        {	364	,	0.0003580000	},
        {	365	,	0.0002400000	},
        {	366	,	0.0005010000	},
        {	367	,	0.0007420000	},
        {	368	,	0.0009660000	},
        {	369	,	0.0011700000	},
        {	370	,	0.0013600000	},
        {	371	,	0.0012200000	},
        {	372	,	0.0011000000	},
        {	373	,	0.0009740000	},
        {	374	,	0.0008610000	},
        {	375	,	0.0007530000	},
        {	376	,	0.0007810000	},
        {	377	,	0.0008040000	},
        {	378	,	0.0008240000	},
        {	379	,	0.0008430000	},
        {	380	,	0.0008590000	},
        {	381	,	0.0007890000	},
        {	382	,	0.0007240000	},
        {	383	,	0.0006640000	},
        {	384	,	0.0006090000	},
        {	385	,	0.0005580000	},
        {	386	,	0.0005150000	},
        {	387	,	0.0004750000	},
        {	388	,	0.0004360000	},
        {	389	,	0.0003990000	},
        {	390	,	0.0003630000	},
        {	391	,	0.0003190000	},
        {	392	,	0.0002760000	},
        {	393	,	0.0002360000	},
        {	394	,	0.0001970000	},
        {	395	,	0.0001590000	},
        {	396	,	0.0002050000	},
        {	397	,	0.0002490000	},
        {	398	,	0.0002890000	},
        {	399	,	0.0003270000	},
        {	400	,	0.0003630000	},
        {	401	,	0.0004740000	},
        {	402	,	0.0005780000	},
        {	403	,	0.0006760000	},
        {	404	,	0.0007680000	},
        {	405	,	0.0008550000	},
        {	406	,	0.0008110000	},
        {	407	,	0.0007710000	},
        {	408	,	0.0007320000	},
        {	409	,	0.0006940000	},
        {	410	,	0.0006580000	},
        {	411	,	0.0006030000	},
        {	412	,	0.0005490000	},
        {	413	,	0.0004990000	},
        {	414	,	0.0004510000	},
        {	415	,	0.0004040000	},
        {	416	,	0.0003960000	},
        {	417	,	0.0003880000	},
        {	418	,	0.0003810000	},
        {	419	,	0.0003730000	},
        {	420	,	0.0003670000	},
        {	421	,	0.0004010000	},
        {	422	,	0.0004330000	},
        {	423	,	0.0004640000	},
        {	424	,	0.0004940000	},
        {	425	,	0.0005230000	},
        {	426	,	0.0005070000	},
        {	427	,	0.0004910000	},
        {	428	,	0.0004760000	},
        {	429	,	0.0004610000	},
        {	430	,	0.0004470000	},
        {	431	,	0.0004700000	},
        {	432	,	0.0004920000	},
        {	433	,	0.0005140000	},
        {	434	,	0.0005350000	},
        {	435	,	0.0005540000	},
        {	436	,	0.0006790000	},
        {	437	,	0.0007990000	},
        {	438	,	0.0009150000	},
        {	439	,	0.0010300000	},
        {	440	,	0.0011400000	},
        {	441	,	0.0022300000	},
        {	442	,	0.0032900000	},
        {	443	,	0.0043100000	},
        {	444	,	0.0053000000	},
        {	445	,	0.0062600000	},
        {	446	,	0.0221000000	},
        {	447	,	0.0375000000	},
        {	448	,	0.0524000000	},
        {	449	,	0.0667000000	},
        {	450	,	0.0806000000	},
        {	451	,	0.1210000000	},
        {	452	,	0.1610000000	},
        {	453	,	0.1990000000	},
        {	454	,	0.2360000000	},
        {	455	,	0.2710000000	},
        {	456	,	0.2810000000	},
        {	457	,	0.2890000000	},
        {	458	,	0.2980000000	},
        {	459	,	0.3070000000	},
        {	460	,	0.3150000000	},
        {	461	,	0.3180000000	},
        {	462	,	0.3220000000	},
        {	463	,	0.3250000000	},
        {	464	,	0.3280000000	},
        {	465	,	0.3310000000	},
        {	466	,	0.3360000000	},
        {	467	,	0.3400000000	},
        {	468	,	0.3440000000	},
        {	469	,	0.3480000000	},
        {	470	,	0.3530000000	},
        {	471	,	0.3580000000	},
        {	472	,	0.3630000000	},
        {	473	,	0.3690000000	},
        {	474	,	0.3730000000	},
        {	475	,	0.3780000000	},
        {	476	,	0.3830000000	},
        {	477	,	0.3880000000	},
        {	478	,	0.3920000000	},
        {	479	,	0.3960000000	},
        {	480	,	0.3990000000	},
        {	481	,	0.4050000000	},
        {	482	,	0.4100000000	},
        {	483	,	0.4150000000	},
        {	484	,	0.4200000000	},
        {	485	,	0.4250000000	},
        {	486	,	0.4270000000	},
        {	487	,	0.4300000000	},
        {	488	,	0.4330000000	},
        {	489	,	0.4360000000	},
        {	490	,	0.4380000000	},
        {	491	,	0.4400000000	},
        {	492	,	0.4420000000	},
        {	493	,	0.4440000000	},
        {	494	,	0.4460000000	},
        {	495	,	0.4490000000	},
        {	496	,	0.4490000000	},
        {	497	,	0.4500000000	},
        {	498	,	0.4510000000	},
        {	499	,	0.4520000000	},
        {	500	,	0.4530000000	},
        {	501	,	0.4550000000	},
        {	502	,	0.4570000000	},
        {	503	,	0.4580000000	},
        {	504	,	0.4600000000	},
        {	505	,	0.4620000000	},
        {	506	,	0.4630000000	},
        {	507	,	0.4650000000	},
        {	508	,	0.4660000000	},
        {	509	,	0.4680000000	},
        {	510	,	0.4690000000	},
        {	511	,	0.4690000000	},
        {	512	,	0.4700000000	},
        {	513	,	0.4700000000	},
        {	514	,	0.4700000000	},
        {	515	,	0.4710000000	},
        {	516	,	0.4720000000	},
        {	517	,	0.4740000000	},
        {	518	,	0.4750000000	},
        {	519	,	0.4770000000	},
        {	520	,	0.4780000000	},
        {	521	,	0.4800000000	},
        {	522	,	0.4820000000	},
        {	523	,	0.4840000000	},
        {	524	,	0.4860000000	},
        {	525	,	0.4880000000	},
        {	526	,	0.4900000000	},
        {	527	,	0.4920000000	},
        {	528	,	0.4940000000	},
        {	529	,	0.4960000000	},
        {	530	,	0.4980000000	},
        {	531	,	0.4990000000	},
        {	532	,	0.5000000000	},
        {	533	,	0.5010000000	},
        {	534	,	0.5020000000	},
        {	535	,	0.5030000000	},
        {	536	,	0.5010000000	},
        {	537	,	0.4990000000	},
        {	538	,	0.4970000000	},
        {	539	,	0.4950000000	},
        {	540	,	0.4930000000	},
        {	541	,	0.4910000000	},
        {	542	,	0.4890000000	},
        {	543	,	0.4870000000	},
        {	544	,	0.4850000000	},
        {	545	,	0.4840000000	},
        {	546	,	0.4820000000	},
        {	547	,	0.4800000000	},
        {	548	,	0.4780000000	},
        {	549	,	0.4770000000	},
        {	550	,	0.4750000000	},
        {	551	,	0.4750000000	},
        {	552	,	0.4750000000	},
        {	553	,	0.4750000000	},
        {	554	,	0.4750000000	},
        {	555	,	0.4750000000	},
        {	556	,	0.4790000000	},
        {	557	,	0.4830000000	},
        {	558	,	0.4870000000	},
        {	559	,	0.4910000000	},
        {	560	,	0.4950000000	},
        {	561	,	0.5050000000	},
        {	562	,	0.5140000000	},
        {	563	,	0.5230000000	},
        {	564	,	0.5330000000	},
        {	565	,	0.5420000000	},
        {	566	,	0.5540000000	},
        {	567	,	0.5660000000	},
        {	568	,	0.5780000000	},
        {	569	,	0.5900000000	},
        {	570	,	0.6020000000	},
        {	571	,	0.6100000000	},
        {	572	,	0.6190000000	},
        {	573	,	0.6280000000	},
        {	574	,	0.6370000000	},
        {	575	,	0.6460000000	},
        {	576	,	0.6500000000	},
        {	577	,	0.6540000000	},
        {	578	,	0.6590000000	},
        {	579	,	0.6640000000	},
        {	580	,	0.6680000000	},
        {	581	,	0.6700000000	},
        {	582	,	0.6710000000	},
        {	583	,	0.6720000000	},
        {	584	,	0.6740000000	},
        {	585	,	0.6750000000	},
        {	586	,	0.6780000000	},
        {	587	,	0.6800000000	},
        {	588	,	0.6820000000	},
        {	589	,	0.6850000000	},
        {	590	,	0.6870000000	},
        {	591	,	0.6910000000	},
        {	592	,	0.6950000000	},
        {	593	,	0.7000000000	},
        {	594	,	0.7040000000	},
        {	595	,	0.7090000000	},
        {	596	,	0.7150000000	},
        {	597	,	0.7210000000	},
        {	598	,	0.7280000000	},
        {	599	,	0.7350000000	},
        {	600	,	0.7410000000	},
        {	601	,	0.7740000000	},
        {	602	,	0.7330000000	},
        {	603	,	0.7490000000	},
        {	604	,	0.7670000000	},
        {	605	,	0.7850000000	},
        {	606	,	0.7910000000	},
        {	607	,	0.7980000000	},
        {	608	,	0.8050000000	},
        {	609	,	0.8140000000	},
        {	610	,	0.8220000000	},
        {	611	,	0.8270000000	},
        {	612	,	0.8320000000	},
        {	613	,	0.8370000000	},
        {	614	,	0.8430000000	},
        {	615	,	0.8500000000	},
        {	616	,	0.8530000000	},
        {	617	,	0.8570000000	},
        {	618	,	0.8610000000	},
        {	619	,	0.8650000000	},
        {	620	,	0.8690000000	},
        {	621	,	0.8730000000	},
        {	622	,	0.8770000000	},
        {	623	,	0.8810000000	},
        {	624	,	0.8850000000	},
        {	625	,	0.8890000000	},
        {	626	,	0.8930000000	},
        {	627	,	0.8980000000	},
        {	628	,	0.9030000000	},
        {	629	,	0.9080000000	},
        {	630	,	0.9130000000	},
        {	631	,	0.9150000000	},
        {	632	,	0.9170000000	},
        {	633	,	0.9220000000	},
        {	634	,	0.9280000000	},
        {	635	,	0.9350000000	},
        {	636	,	0.9440000000	},
        {	637	,	0.9510000000	},
        {	638	,	0.9540000000	},
        {	639	,	0.9540000000	},
        {	640	,	0.9530000000	},
        {	641	,	0.9570000000	},
        {	642	,	0.9620000000	},
        {	643	,	0.9660000000	},
        {	644	,	0.9700000000	},
        {	645	,	0.9740000000	},
        {	646	,	0.9770000000	},
        {	647	,	0.9800000000	},
        {	648	,	0.9830000000	},
        {	649	,	0.9860000000	},
        {	650	,	0.9900000000	},
        {	651	,	0.9920000000	},
        {	652	,	0.9940000000	},
        {	653	,	0.9960000000	},
        {	654	,	0.9980000000	},
        {	655	,	1.0000000000	},
        {	656	,	0.9990000000	},
        {	657	,	0.9980000000	},
        {	658	,	0.9980000000	},
        {	659	,	0.9980000000	},
        {	660	,	0.9980000000	},
        {	661	,	0.9940000000	},
        {	662	,	0.9910000000	},
        {	663	,	0.9880000000	},
        {	664	,	0.9850000000	},
        {	665	,	0.9830000000	},
        {	666	,	0.9800000000	},
        {	667	,	0.9760000000	},
        {	668	,	0.9730000000	},
        {	669	,	0.9700000000	},
        {	670	,	0.9670000000	},
        {	671	,	0.9640000000	},
        {	672	,	0.9600000000	},
        {	673	,	0.9570000000	},
        {	674	,	0.9550000000	},
        {	675	,	0.9520000000	},
        {	676	,	0.9490000000	},
        {	677	,	0.9470000000	},
        {	678	,	0.9440000000	},
        {	679	,	0.9420000000	},
        {	680	,	0.9400000000	},
        {	681	,	0.9360000000	},
        {	682	,	0.9320000000	},
        {	683	,	0.9290000000	},
        {	684	,	0.9250000000	},
        {	685	,	0.9230000000	},
        {	686	,	0.9200000000	},
        {	687	,	0.9170000000	},
        {	688	,	0.9140000000	},
        {	689	,	0.9120000000	},
        {	690	,	0.9090000000	},
        {	691	,	0.9080000000	},
        {	692	,	0.9070000000	},
        {	693	,	0.9070000000	},
        {	694	,	0.9060000000	},
        {	695	,	0.9060000000	},
        {	696	,	0.9070000000	},
        {	697	,	0.9090000000	},
        {	698	,	0.9110000000	},
        {	699	,	0.9120000000	},
        {	700	,	0.9140000000	},
        {	701	,	0.9140000000	},
        {	702	,	0.9140000000	},
        {	703	,	0.9130000000	},
        {	704	,	0.9130000000	},
        {	705	,	0.9130000000	},
        {	706	,	0.9120000000	},
        {	707	,	0.9110000000	},
        {	708	,	0.9110000000	},
        {	709	,	0.9100000000	},
        {	710	,	0.9100000000	},
        {	711	,	0.9130000000	},
        {	712	,	0.9170000000	},
        {	713	,	0.9210000000	},
        {	714	,	0.9250000000	},
        {	715	,	0.9300000000	},
        {	716	,	0.9370000000	},
        {	717	,	0.9450000000	},
        {	718	,	0.9530000000	},
        {	719	,	0.9610000000	},
        {	720	,	0.9700000000	},
        {	721	,	0.9750000000	},
        {	722	,	0.9800000000	},
        {	723	,	0.9850000000	},
        {	724	,	0.9910000000	},
        {	725	,	0.9960000000	},
        {	726	,	0.9960000000	},
        {	727	,	0.9960000000	},
        {	728	,	0.9960000000	},
        {	729	,	0.9960000000	},
        {	730	,	0.9960000000	},
        {	731	,	0.9910000000	},
        {	732	,	0.9860000000	},
        {	733	,	0.9810000000	},
        {	734	,	0.9750000000	},
        {	735	,	0.9700000000	},
        {	736	,	0.9640000000	},
        {	737	,	0.9580000000	},
        {	738	,	0.9520000000	},
        {	739	,	0.9450000000	},
        {	740	,	0.9390000000	},
        {	741	,	0.9350000000	},
        {	742	,	0.9310000000	},
        {	743	,	0.9260000000	},
        {	744	,	0.9220000000	},
        {	745	,	0.9170000000	},
        {	746	,	0.9140000000	},
        {	747	,	0.9120000000	},
        {	748	,	0.9090000000	},
        {	749	,	0.9060000000	},
        {	750	,	0.9030000000	},
        {	751	,	0.9010000000	},
        {	752	,	0.8980000000	},
        {	753	,	0.8950000000	},
        {	754	,	0.8920000000	},
        {	755	,	0.8900000000	},
        {	756	,	0.8870000000	},
        {	757	,	0.8840000000	},
        {	758	,	0.8820000000	},
        {	759	,	0.8800000000	},
        {	760	,	0.8780000000	},
        {	761	,	0.8760000000	},
        {	762	,	0.8740000000	},
        {	763	,	0.8710000000	},
        {	764	,	0.8680000000	},
        {	765	,	0.8650000000	},
        {	766	,	0.8600000000	},
        {	767	,	0.8560000000	},
        {	768	,	0.8520000000	},
        {	769	,	0.8480000000	},
        {	770	,	0.8440000000	},
        {	771	,	0.8390000000	},
        {	772	,	0.8340000000	},
        {	773	,	0.8290000000	},
        {	774	,	0.8240000000	},
        {	775	,	0.8180000000	},
        {	776	,	0.8160000000	},
        {	777	,	0.8130000000	},
        {	778	,	0.8110000000	},
        {	779	,	0.8080000000	},
        {	780	,	0.8050000000	},
        {	781	,	0.8060000000	},
        {	782	,	0.8070000000	},
        {	783	,	0.8070000000	},
        {	784	,	0.8080000000	},
        {	785	,	0.8080000000	},
        {	786	,	0.8110000000	},
        {	787	,	0.8150000000	},
        {	788	,	0.8180000000	},
        {	789	,	0.8210000000	},
        {	790	,	0.8240000000	},
        {	791	,	0.8300000000	},
        {	792	,	0.8370000000	},
        {	793	,	0.8430000000	},
        {	794	,	0.8500000000	},
        {	795	,	0.8560000000	},
        {	796	,	0.8600000000	},
        {	797	,	0.8640000000	},
        {	798	,	0.8680000000	},
        {	799	,	0.8710000000	},
        {	800	,	0.8750000000	},
        {	801	,	0.8780000000	},
        {	802	,	0.8810000000	},
        {	803	,	0.8840000000	},
        {	804	,	0.8860000000	},
        {	805	,	0.8890000000	},
        {	806	,	0.8920000000	},
        {	807	,	0.8950000000	},
        {	808	,	0.8970000000	},
        {	809	,	0.9000000000	},
        {	810	,	0.9030000000	},
        {	811	,	0.9050000000	},
        {	812	,	0.9070000000	},
        {	813	,	0.9080000000	},
        {	814	,	0.9090000000	},
        {	815	,	0.9100000000	},
        {	816	,	0.9130000000	},
        {	817	,	0.9140000000	},
        {	818	,	0.9150000000	},
        {	819	,	0.9170000000	},
        {	820	,	0.9170000000	},
        {	821	,	0.9180000000	},
        {	822	,	0.9190000000	},
        {	823	,	0.9200000000	},
        {	824	,	0.9200000000	},
        {	825	,	0.9200000000	},
        {	826	,	0.9180000000	},
        {	827	,	0.9160000000	},
        {	828	,	0.9140000000	},
        {	829	,	0.9110000000	},
        {	830	,	0.9080000000	},
        {	831	,	0.9020000000	},
        {	832	,	0.8960000000	},
        {	833	,	0.8890000000	},
        {	834	,	0.8820000000	},
        {	835	,	0.8750000000	},
        {	836	,	0.8670000000	},
        {	837	,	0.8600000000	},
        {	838	,	0.8510000000	},
        {	839	,	0.8430000000	},
        {	840	,	0.8340000000	},
        {	841	,	0.8280000000	},
        {	842	,	0.8220000000	},
        {	843	,	0.8150000000	},
        {	844	,	0.8080000000	},
        {	845	,	0.8010000000	},
        {	846	,	0.7990000000	},
        {	847	,	0.7960000000	},
        {	848	,	0.7930000000	},
        {	849	,	0.7900000000	},
        {	850	,	0.7870000000	},
        {	851	,	0.7850000000	},
        {	852	,	0.7840000000	},
        {	853	,	0.7820000000	},
        {	854	,	0.7790000000	},
        {	855	,	0.7770000000	},
        {	856	,	0.7760000000	},
        {	857	,	0.7750000000	},
        {	858	,	0.7740000000	},
        {	859	,	0.7720000000	},
        {	860	,	0.7700000000	},
        {	861	,	0.7680000000	},
        {	862	,	0.7650000000	},
        {	863	,	0.7630000000	},
        {	864	,	0.7600000000	},
        {	865	,	0.7570000000	},
        {	866	,	0.7540000000	},
        {	867	,	0.7510000000	},
        {	868	,	0.7480000000	},
        {	869	,	0.7450000000	},
        {	870	,	0.7420000000	},
        {	871	,	0.7390000000	},
        {	872	,	0.7370000000	},
        {	873	,	0.7340000000	},
        {	874	,	0.7320000000	},
        {	875	,	0.7290000000	},
        {	876	,	0.7260000000	},
        {	877	,	0.7240000000	},
        {	878	,	0.7210000000	},
        {	879	,	0.7180000000	},
        {	880	,	0.7160000000	},
        {	881	,	0.7130000000	},
        {	882	,	0.7110000000	},
        {	883	,	0.7080000000	},
        {	884	,	0.7050000000	},
        {	885	,	0.7030000000	},
        {	886	,	0.7040000000	},
        {	887	,	0.7040000000	},
        {	888	,	0.7040000000	},
        {	889	,	0.7050000000	},
        {	890	,	0.7050000000	},
        {	891	,	0.7050000000	},
        {	892	,	0.7050000000	},
        {	893	,	0.7050000000	},
        {	894	,	0.7050000000	},
        {	895	,	0.7050000000	},
        {	896	,	0.6760000000	},
        {	897	,	0.6490000000	},
        {	898	,	0.6210000000	},
        {	899	,	0.5950000000	},
        {	900	,	0.5680000000	},
        {	901	,	0.5080000000	},
        {	902	,	0.4490000000	},
        {	903	,	0.3910000000	},
        {	904	,	0.3340000000	},
        {	905	,	0.2780000000	},
        {	906	,	0.2400000000	},
        {	907	,	0.2030000000	},
        {	908	,	0.1670000000	},
        {	909	,	0.1310000000	},
        {	910	,	0.0960000000	},
        {	911	,	0.0831000000	},
        {	912	,	0.0704000000	},
        {	913	,	0.0580000000	},
        {	914	,	0.0458000000	},
        {	915	,	0.0339000000	},
        {	916	,	0.0295000000	},
        {	917	,	0.0251000000	},
        {	918	,	0.0209000000	},
        {	919	,	0.0167000000	},
        {	920	,	0.0126000000	},
        {	921	,	0.0111000000	},
        {	922	,	0.0096400000	},
        {	923	,	0.0081800000	},
        {	924	,	0.0067500000	},
        {	925	,	0.0053400000	},
        {	926	,	0.0047500000	},
        {	927	,	0.0041700000	},
        {	928	,	0.0035900000	},
        {	929	,	0.0030300000	},
        {	930	,	0.0024700000	},
        {	931	,	0.0022000000	},
        {	932	,	0.0019300000	},
        {	933	,	0.0016700000	},
        {	934	,	0.0014100000	},
        {	935	,	0.0011500000	},
        {	936	,	0.0010200000	},
        {	937	,	0.0008960000	},
        {	938	,	0.0007720000	},
        {	939	,	0.0006500000	},
        {	940	,	0.0005310000	},
        {	941	,	0.0004710000	},
        {	942	,	0.0004120000	},
        {	943	,	0.0003530000	},
        {	944	,	0.0002950000	},
        {	945	,	0.0002380000	},
        {	946	,	0.0002070000	},
        {	947	,	0.0001760000	},
        {	948	,	0.0001450000	},
        {	949	,	0.0001150000	},
        {	950	,	0.0000858000	},
        {	951	,	0.0001260000	},
        {	952	,	0.0001660000	},
        {	953	,	0.0002060000	},
        {	954	,	0.0002450000	},
        {	955	,	0.0002840000	},
        {	956	,	0.0002820000	},
        {	957	,	0.0002810000	},
        {	958	,	0.0002790000	},
        {	959	,	0.0002770000	},
        {	960	,	0.0002760000	},
        {	961	,	0.0002630000	},
        {	962	,	0.0002500000	},
        {	963	,	0.0002370000	},
        {	964	,	0.0002250000	},
        {	965	,	0.0002130000	},
        {	966	,	0.0002090000	},
        {	967	,	0.0002060000	},
        {	968	,	0.0002030000	},
        {	969	,	0.0002000000	},
        {	970	,	0.0001970000	},
        {	971	,	0.0001650000	},
        {	972	,	0.0001320000	},
        {	973	,	0.0001000000	},
        {	974	,	0.0000686000	},
        {	975	,	0.0000370000	},
        {	976	,	0.0000421000	},
        {	977	,	0.0000472000	},
        {	978	,	0.0000522000	},
        {	979	,	0.0000573000	},
        {	980	,	0.0000622000	},
        {	981	,	0.0000637000	},
        {	982	,	0.0000652000	},
        {	983	,	0.0000666000	},
        {	984	,	0.0000680000	},
        {	985	,	0.0000695000	},
        {	986	,	0.0000554000	},
        {	987	,	0.0000414000	},
        {	988	,	0.0000275000	},
        {	989	,	0.0000137000	},
        {	990	,	0.0000000000	},
        {	991	,	0.0000050500	},
        {	992	,	0.0000101000	},
        {	993	,	0.0000151000	},
        {	994	,	0.0000200000	},
        {	995	,	0.0000250000	},
        {	996	,	0.0000696000	},
        {	997	,	0.0001140000	},
        {	998	,	0.0001590000	},
        {	999	,	0.0002030000	},
        {	1000	,	0.0002480000	}
      };

      static const double CBERS04A_WPMB1_SRF[][2] = 
      {
        {	350	,	0.0000000000	},
        {	351	,	0.0002200000	},
        {	352	,	0.0004180000	},
        {	353	,	0.0005970000	},
        {	354	,	0.0007590000	},
        {	355	,	0.0009070000	},
        {	356	,	0.0011600000	},
        {	357	,	0.0013900000	},
        {	358	,	0.0016000000	},
        {	359	,	0.0017900000	},
        {	360	,	0.0019700000	},
        {	361	,	0.0020700000	},
        {	362	,	0.0021600000	},
        {	363	,	0.0022400000	},
        {	364	,	0.0023100000	},
        {	365	,	0.0023800000	},
        {	366	,	0.0024300000	},
        {	367	,	0.0024700000	},
        {	368	,	0.0025100000	},
        {	369	,	0.0025400000	},
        {	370	,	0.0025600000	},
        {	371	,	0.0025900000	},
        {	372	,	0.0026100000	},
        {	373	,	0.0026300000	},
        {	374	,	0.0026400000	},
        {	375	,	0.0026500000	},
        {	376	,	0.0025400000	},
        {	377	,	0.0024400000	},
        {	378	,	0.0023400000	},
        {	379	,	0.0022400000	},
        {	380	,	0.0021500000	},
        {	381	,	0.0020700000	},
        {	382	,	0.0019900000	},
        {	383	,	0.0019300000	},
        {	384	,	0.0018700000	},
        {	385	,	0.0018200000	},
        {	386	,	0.0018200000	},
        {	387	,	0.0018200000	},
        {	388	,	0.0018300000	},
        {	389	,	0.0018300000	},
        {	390	,	0.0018300000	},
        {	391	,	0.0017900000	},
        {	392	,	0.0017500000	},
        {	393	,	0.0017000000	},
        {	394	,	0.0016600000	},
        {	395	,	0.0016100000	},
        {	396	,	0.0016600000	},
        {	397	,	0.0017100000	},
        {	398	,	0.0017600000	},
        {	399	,	0.0017900000	},
        {	400	,	0.0018300000	},
        {	401	,	0.0018700000	},
        {	402	,	0.0019000000	},
        {	403	,	0.0019400000	},
        {	404	,	0.0019700000	},
        {	405	,	0.0019900000	},
        {	406	,	0.0019800000	},
        {	407	,	0.0019600000	},
        {	408	,	0.0019400000	},
        {	409	,	0.0019200000	},
        {	410	,	0.0019000000	},
        {	411	,	0.0018900000	},
        {	412	,	0.0018800000	},
        {	413	,	0.0018700000	},
        {	414	,	0.0018600000	},
        {	415	,	0.0018500000	},
        {	416	,	0.0018100000	},
        {	417	,	0.0017600000	},
        {	418	,	0.0017200000	},
        {	419	,	0.0016800000	},
        {	420	,	0.0016400000	},
        {	421	,	0.0016500000	},
        {	422	,	0.0016500000	},
        {	423	,	0.0016600000	},
        {	424	,	0.0016600000	},
        {	425	,	0.0016600000	},
        {	426	,	0.0016400000	},
        {	427	,	0.0016200000	},
        {	428	,	0.0016000000	},
        {	429	,	0.0015800000	},
        {	430	,	0.0015600000	},
        {	431	,	0.0015600000	},
        {	432	,	0.0015700000	},
        {	433	,	0.0015700000	},
        {	434	,	0.0015800000	},
        {	435	,	0.0015800000	},
        {	436	,	0.0015500000	},
        {	437	,	0.0015200000	},
        {	438	,	0.0015000000	},
        {	439	,	0.0014700000	},
        {	440	,	0.0014500000	},
        {	441	,	0.0014600000	},
        {	442	,	0.0014700000	},
        {	443	,	0.0014800000	},
        {	444	,	0.0015000000	},
        {	445	,	0.0015100000	},
        {	446	,	0.0014500000	},
        {	447	,	0.0013900000	},
        {	448	,	0.0013300000	},
        {	449	,	0.0012800000	},
        {	450	,	0.0012200000	},
        {	451	,	0.1290000000	},
        {	452	,	0.2530000000	},
        {	453	,	0.3720000000	},
        {	454	,	0.4890000000	},
        {	455	,	0.6010000000	},
        {	456	,	0.5920000000	},
        {	457	,	0.5830000000	},
        {	458	,	0.5740000000	},
        {	459	,	0.5650000000	},
        {	460	,	0.5570000000	},
        {	461	,	0.5760000000	},
        {	462	,	0.5940000000	},
        {	463	,	0.6120000000	},
        {	464	,	0.6300000000	},
        {	465	,	0.6460000000	},
        {	466	,	0.6540000000	},
        {	467	,	0.6620000000	},
        {	468	,	0.6700000000	},
        {	469	,	0.6770000000	},
        {	470	,	0.6850000000	},
        {	471	,	0.6770000000	},
        {	472	,	0.6690000000	},
        {	473	,	0.6610000000	},
        {	474	,	0.6530000000	},
        {	475	,	0.6450000000	},
        {	476	,	0.6790000000	},
        {	477	,	0.7110000000	},
        {	478	,	0.7420000000	},
        {	479	,	0.7710000000	},
        {	480	,	0.7990000000	},
        {	481	,	0.8080000000	},
        {	482	,	0.8160000000	},
        {	483	,	0.8230000000	},
        {	484	,	0.8310000000	},
        {	485	,	0.8370000000	},
        {	486	,	0.8490000000	},
        {	487	,	0.8590000000	},
        {	488	,	0.8700000000	},
        {	489	,	0.8810000000	},
        {	490	,	0.8920000000	},
        {	491	,	0.8960000000	},
        {	492	,	0.9000000000	},
        {	493	,	0.9050000000	},
        {	494	,	0.9090000000	},
        {	495	,	0.9150000000	},
        {	496	,	0.9200000000	},
        {	497	,	0.9260000000	},
        {	498	,	0.9320000000	},
        {	499	,	0.9390000000	},
        {	500	,	0.9440000000	},
        {	501	,	0.9480000000	},
        {	502	,	0.9520000000	},
        {	503	,	0.9560000000	},
        {	504	,	0.9600000000	},
        {	505	,	0.9640000000	},
        {	506	,	0.9720000000	},
        {	507	,	0.9790000000	},
        {	508	,	0.9860000000	},
        {	509	,	0.9930000000	},
        {	510	,	1.0000000000	},
        {	511	,	0.9930000000	},
        {	512	,	0.9850000000	},
        {	513	,	0.9780000000	},
        {	514	,	0.9710000000	},
        {	515	,	0.9640000000	},
        {	516	,	0.9170000000	},
        {	517	,	0.8700000000	},
        {	518	,	0.8240000000	},
        {	519	,	0.7790000000	},
        {	520	,	0.7340000000	},
        {	521	,	0.5910000000	},
        {	522	,	0.4500000000	},
        {	523	,	0.3110000000	},
        {	524	,	0.1730000000	},
        {	525	,	0.0367000000	},
        {	526	,	0.0365000000	},
        {	527	,	0.0363000000	},
        {	528	,	0.0361000000	},
        {	529	,	0.0359000000	},
        {	530	,	0.0357000000	},
        {	531	,	0.0286000000	},
        {	532	,	0.0215000000	},
        {	533	,	0.0145000000	},
        {	534	,	0.0076100000	},
        {	535	,	0.0007470000	},
        {	536	,	0.0006900000	},
        {	537	,	0.0006340000	},
        {	538	,	0.0005790000	},
        {	539	,	0.0005240000	},
        {	540	,	0.0004700000	},
        {	541	,	0.0004860000	},
        {	542	,	0.0005020000	},
        {	543	,	0.0005180000	},
        {	544	,	0.0005330000	},
        {	545	,	0.0005490000	},
        {	546	,	0.0005590000	},
        {	547	,	0.0005700000	},
        {	548	,	0.0005800000	},
        {	549	,	0.0005910000	},
        {	550	,	0.0006010000	},
        {	551	,	0.0006840000	},
        {	552	,	0.0007660000	},
        {	553	,	0.0008480000	},
        {	554	,	0.0009290000	},
        {	555	,	0.0010100000	},
        {	556	,	0.0010000000	},
        {	557	,	0.0009980000	},
        {	558	,	0.0009930000	},
        {	559	,	0.0009870000	},
        {	560	,	0.0009810000	},
        {	561	,	0.0009770000	},
        {	562	,	0.0009740000	},
        {	563	,	0.0009700000	},
        {	564	,	0.0009670000	},
        {	565	,	0.0009640000	},
        {	566	,	0.0008970000	},
        {	567	,	0.0008300000	},
        {	568	,	0.0007630000	},
        {	569	,	0.0006970000	},
        {	570	,	0.0006310000	},
        {	571	,	0.0006350000	},
        {	572	,	0.0006380000	},
        {	573	,	0.0006420000	},
        {	574	,	0.0006460000	},
        {	575	,	0.0006500000	},
        {	576	,	0.0006640000	},
        {	577	,	0.0006790000	},
        {	578	,	0.0006940000	},
        {	579	,	0.0007090000	},
        {	580	,	0.0007240000	},
        {	581	,	0.0008380000	},
        {	582	,	0.0009520000	},
        {	583	,	0.0010700000	},
        {	584	,	0.0011800000	},
        {	585	,	0.0013000000	},
        {	586	,	0.0013000000	},
        {	587	,	0.0013100000	},
        {	588	,	0.0013200000	},
        {	589	,	0.0013200000	},
        {	590	,	0.0013300000	},
        {	591	,	0.0013800000	},
        {	592	,	0.0014300000	},
        {	593	,	0.0014800000	},
        {	594	,	0.0015300000	},
        {	595	,	0.0015800000	},
        {	596	,	0.0013800000	},
        {	597	,	0.0011800000	},
        {	598	,	0.0009770000	},
        {	599	,	0.0007710000	},
        {	600	,	0.0005640000	},
        {	601	,	0.0005670000	},
        {	602	,	0.0005170000	},
        {	603	,	0.0005100000	},
        {	604	,	0.0005040000	},
        {	605	,	0.0004990000	},
        {	606	,	0.0004770000	},
        {	607	,	0.0004560000	},
        {	608	,	0.0004350000	},
        {	609	,	0.0004150000	},
        {	610	,	0.0003950000	},
        {	611	,	0.0004360000	},
        {	612	,	0.0004760000	},
        {	613	,	0.0005170000	},
        {	614	,	0.0005580000	},
        {	615	,	0.0005980000	},
        {	616	,	0.0005980000	},
        {	617	,	0.0005980000	},
        {	618	,	0.0005980000	},
        {	619	,	0.0005980000	},
        {	620	,	0.0005990000	},
        {	621	,	0.0006390000	},
        {	622	,	0.0006800000	},
        {	623	,	0.0007210000	},
        {	624	,	0.0007620000	},
        {	625	,	0.0008040000	},
        {	626	,	0.0008150000	},
        {	627	,	0.0008260000	},
        {	628	,	0.0008380000	},
        {	629	,	0.0008500000	},
        {	630	,	0.0008620000	},
        {	631	,	0.0008200000	},
        {	632	,	0.0007780000	},
        {	633	,	0.0007370000	},
        {	634	,	0.0006970000	},
        {	635	,	0.0006560000	},
        {	636	,	0.0006390000	},
        {	637	,	0.0006200000	},
        {	638	,	0.0005990000	},
        {	639	,	0.0005760000	},
        {	640	,	0.0005520000	},
        {	641	,	0.0005480000	},
        {	642	,	0.0005450000	},
        {	643	,	0.0005420000	},
        {	644	,	0.0005380000	},
        {	645	,	0.0005350000	},
        {	646	,	0.0005290000	},
        {	647	,	0.0005230000	},
        {	648	,	0.0005160000	},
        {	649	,	0.0005100000	},
        {	650	,	0.0005040000	},
        {	651	,	0.0005570000	},
        {	652	,	0.0006090000	},
        {	653	,	0.0006600000	},
        {	654	,	0.0007120000	},
        {	655	,	0.0007620000	},
        {	656	,	0.0007220000	},
        {	657	,	0.0006810000	},
        {	658	,	0.0006420000	},
        {	659	,	0.0006020000	},
        {	660	,	0.0005630000	},
        {	661	,	0.0005690000	},
        {	662	,	0.0005760000	},
        {	663	,	0.0005820000	},
        {	664	,	0.0005880000	},
        {	665	,	0.0005940000	},
        {	666	,	0.0005990000	},
        {	667	,	0.0006050000	},
        {	668	,	0.0006100000	},
        {	669	,	0.0006150000	},
        {	670	,	0.0006200000	},
        {	671	,	0.0006330000	},
        {	672	,	0.0006450000	},
        {	673	,	0.0006580000	},
        {	674	,	0.0006700000	},
        {	675	,	0.0006830000	},
        {	676	,	0.0006910000	},
        {	677	,	0.0007000000	},
        {	678	,	0.0007080000	},
        {	679	,	0.0007180000	},
        {	680	,	0.0007260000	},
        {	681	,	0.0007240000	},
        {	682	,	0.0007230000	},
        {	683	,	0.0007220000	},
        {	684	,	0.0007200000	},
        {	685	,	0.0007190000	},
        {	686	,	0.0007110000	},
        {	687	,	0.0007030000	},
        {	688	,	0.0006950000	},
        {	689	,	0.0006870000	},
        {	690	,	0.0006790000	},
        {	691	,	0.0006640000	},
        {	692	,	0.0006490000	},
        {	693	,	0.0006350000	},
        {	694	,	0.0006200000	},
        {	695	,	0.0006060000	},
        {	696	,	0.0006410000	},
        {	697	,	0.0006760000	},
        {	698	,	0.0007120000	},
        {	699	,	0.0007480000	},
        {	700	,	0.0007840000	},
        {	701	,	0.0007860000	},
        {	702	,	0.0007880000	},
        {	703	,	0.0007900000	},
        {	704	,	0.0007920000	},
        {	705	,	0.0007950000	},
        {	706	,	0.0007970000	},
        {	707	,	0.0008000000	},
        {	708	,	0.0008030000	},
        {	709	,	0.0008060000	},
        {	710	,	0.0008090000	},
        {	711	,	0.0007410000	},
        {	712	,	0.0006730000	},
        {	713	,	0.0006040000	},
        {	714	,	0.0005340000	},
        {	715	,	0.0004640000	},
        {	716	,	0.0004680000	},
        {	717	,	0.0004720000	},
        {	718	,	0.0004760000	},
        {	719	,	0.0004810000	},
        {	720	,	0.0004850000	},
        {	721	,	0.0004890000	},
        {	722	,	0.0004940000	},
        {	723	,	0.0004990000	},
        {	724	,	0.0005040000	},
        {	725	,	0.0005100000	},
        {	726	,	0.0005190000	},
        {	727	,	0.0005280000	},
        {	728	,	0.0005370000	},
        {	729	,	0.0005470000	},
        {	730	,	0.0005560000	},
        {	731	,	0.0005570000	},
        {	732	,	0.0005570000	},
        {	733	,	0.0005580000	},
        {	734	,	0.0005580000	},
        {	735	,	0.0005590000	},
        {	736	,	0.0005670000	},
        {	737	,	0.0005750000	},
        {	738	,	0.0005830000	},
        {	739	,	0.0005920000	},
        {	740	,	0.0006000000	},
        {	741	,	0.0005960000	},
        {	742	,	0.0005930000	},
        {	743	,	0.0005890000	},
        {	744	,	0.0005850000	},
        {	745	,	0.0005810000	},
        {	746	,	0.0006050000	},
        {	747	,	0.0006290000	},
        {	748	,	0.0006530000	},
        {	749	,	0.0006780000	},
        {	750	,	0.0007030000	},
        {	751	,	0.0007280000	},
        {	752	,	0.0007550000	},
        {	753	,	0.0007810000	},
        {	754	,	0.0008080000	},
        {	755	,	0.0008350000	},
        {	756	,	0.0008410000	},
        {	757	,	0.0008460000	},
        {	758	,	0.0008530000	},
        {	759	,	0.0008590000	},
        {	760	,	0.0008660000	},
        {	761	,	0.0008730000	},
        {	762	,	0.0008800000	},
        {	763	,	0.0008880000	},
        {	764	,	0.0008940000	},
        {	765	,	0.0009000000	},
        {	766	,	0.0009070000	},
        {	767	,	0.0009140000	},
        {	768	,	0.0009210000	},
        {	769	,	0.0009280000	},
        {	770	,	0.0009350000	},
        {	771	,	0.0009070000	},
        {	772	,	0.0008780000	},
        {	773	,	0.0008480000	},
        {	774	,	0.0008190000	},
        {	775	,	0.0007880000	},
        {	776	,	0.0008500000	},
        {	777	,	0.0009130000	},
        {	778	,	0.0009770000	},
        {	779	,	0.0010400000	},
        {	780	,	0.0011100000	},
        {	781	,	0.0010900000	},
        {	782	,	0.0010600000	},
        {	783	,	0.0010400000	},
        {	784	,	0.0010200000	},
        {	785	,	0.0009930000	},
        {	786	,	0.0009610000	},
        {	787	,	0.0009280000	},
        {	788	,	0.0008940000	},
        {	789	,	0.0008600000	},
        {	790	,	0.0008250000	},
        {	791	,	0.0008380000	},
        {	792	,	0.0008510000	},
        {	793	,	0.0008640000	},
        {	794	,	0.0008770000	},
        {	795	,	0.0008910000	},
        {	796	,	0.0009960000	},
        {	797	,	0.0011000000	},
        {	798	,	0.0012100000	},
        {	799	,	0.0013200000	},
        {	800	,	0.0014300000	},
        {	801	,	0.0014400000	},
        {	802	,	0.0014500000	},
        {	803	,	0.0014500000	},
        {	804	,	0.0014600000	},
        {	805	,	0.0014700000	},
        {	806	,	0.0014200000	},
        {	807	,	0.0013600000	},
        {	808	,	0.0013000000	},
        {	809	,	0.0012500000	},
        {	810	,	0.0011900000	},
        {	811	,	0.0011400000	},
        {	812	,	0.0010900000	},
        {	813	,	0.0010400000	},
        {	814	,	0.0009950000	},
        {	815	,	0.0009440000	},
        {	816	,	0.0009530000	},
        {	817	,	0.0009610000	},
        {	818	,	0.0009690000	},
        {	819	,	0.0009760000	},
        {	820	,	0.0009840000	},
        {	821	,	0.0009980000	},
        {	822	,	0.0010100000	},
        {	823	,	0.0010300000	},
        {	824	,	0.0010400000	},
        {	825	,	0.0010500000	},
        {	826	,	0.0010800000	},
        {	827	,	0.0011100000	},
        {	828	,	0.0011400000	},
        {	829	,	0.0011700000	},
        {	830	,	0.0012000000	},
        {	831	,	0.0011500000	},
        {	832	,	0.0011100000	},
        {	833	,	0.0010600000	},
        {	834	,	0.0010100000	},
        {	835	,	0.0009630000	},
        {	836	,	0.0009740000	},
        {	837	,	0.0009840000	},
        {	838	,	0.0009940000	},
        {	839	,	0.0010000000	},
        {	840	,	0.0010100000	},
        {	841	,	0.0010100000	},
        {	842	,	0.0010100000	},
        {	843	,	0.0010100000	},
        {	844	,	0.0010100000	},
        {	845	,	0.0010100000	},
        {	846	,	0.0009760000	},
        {	847	,	0.0009400000	},
        {	848	,	0.0009050000	},
        {	849	,	0.0008710000	},
        {	850	,	0.0008360000	},
        {	851	,	0.0007240000	},
        {	852	,	0.0006140000	},
        {	853	,	0.0005050000	},
        {	854	,	0.0003970000	},
        {	855	,	0.0002920000	},
        {	856	,	0.0002530000	},
        {	857	,	0.0002160000	},
        {	858	,	0.0001780000	},
        {	859	,	0.0001420000	},
        {	860	,	0.0001060000	},
        {	861	,	0.0001050000	},
        {	862	,	0.0001040000	},
        {	863	,	0.0001030000	},
        {	864	,	0.0001020000	},
        {	865	,	0.0001010000	},
        {	866	,	0.0001680000	},
        {	867	,	0.0002340000	},
        {	868	,	0.0002990000	},
        {	869	,	0.0003620000	},
        {	870	,	0.0004240000	},
        {	871	,	0.0003870000	},
        {	872	,	0.0003510000	},
        {	873	,	0.0003160000	},
        {	874	,	0.0002820000	},
        {	875	,	0.0002480000	},
        {	876	,	0.0002080000	},
        {	877	,	0.0001690000	},
        {	878	,	0.0001300000	},
        {	879	,	0.0000930000	},
        {	880	,	0.0000563000	},
        {	881	,	0.0001120000	},
        {	882	,	0.0001660000	},
        {	883	,	0.0002190000	},
        {	884	,	0.0002710000	},
        {	885	,	0.0003220000	},
        {	886	,	0.0002710000	},
        {	887	,	0.0002220000	},
        {	888	,	0.0001730000	},
        {	889	,	0.0001260000	},
        {	890	,	0.0000793000	},
        {	891	,	0.0000727000	},
        {	892	,	0.0000662000	},
        {	893	,	0.0000598000	},
        {	894	,	0.0000536000	},
        {	895	,	0.0000475000	},
        {	896	,	0.0000453000	},
        {	897	,	0.0000430000	},
        {	898	,	0.0000409000	},
        {	899	,	0.0000387000	},
        {	900	,	0.0000366000	},
        {	901	,	0.0001140000	},
        {	902	,	0.0001890000	},
        {	903	,	0.0002630000	},
        {	904	,	0.0003360000	},
        {	905	,	0.0004070000	},
        {	906	,	0.0004030000	},
        {	907	,	0.0003990000	},
        {	908	,	0.0003950000	},
        {	909	,	0.0003910000	},
        {	910	,	0.0003880000	},
        {	911	,	0.0003840000	},
        {	912	,	0.0003800000	},
        {	913	,	0.0003760000	},
        {	914	,	0.0003730000	},
        {	915	,	0.0003690000	},
        {	916	,	0.0003240000	},
        {	917	,	0.0002790000	},
        {	918	,	0.0002340000	},
        {	919	,	0.0001910000	},
        {	920	,	0.0001490000	},
        {	921	,	0.0001440000	},
        {	922	,	0.0001390000	},
        {	923	,	0.0001340000	},
        {	924	,	0.0001290000	},
        {	925	,	0.0001250000	},
        {	926	,	0.0001240000	},
        {	927	,	0.0001230000	},
        {	928	,	0.0001220000	},
        {	929	,	0.0001220000	},
        {	930	,	0.0001210000	},
        {	931	,	0.0001200000	},
        {	932	,	0.0001190000	},
        {	933	,	0.0001180000	},
        {	934	,	0.0001180000	},
        {	935	,	0.0001170000	},
        {	936	,	0.0001160000	},
        {	937	,	0.0001150000	},
        {	938	,	0.0001140000	},
        {	939	,	0.0001130000	},
        {	940	,	0.0001120000	},
        {	941	,	0.0001110000	},
        {	942	,	0.0001110000	},
        {	943	,	0.0001100000	},
        {	944	,	0.0001090000	},
        {	945	,	0.0001090000	},
        {	946	,	0.0001080000	},
        {	947	,	0.0001070000	},
        {	948	,	0.0001070000	},
        {	949	,	0.0001060000	},
        {	950	,	0.0001050000	},
        {	951	,	0.0000971000	},
        {	952	,	0.0000888000	},
        {	953	,	0.0000805000	},
        {	954	,	0.0000724000	},
        {	955	,	0.0000645000	},
        {	956	,	0.0000606000	},
        {	957	,	0.0000567000	},
        {	958	,	0.0000529000	},
        {	959	,	0.0000492000	},
        {	960	,	0.0000455000	},
        {	961	,	0.0000453000	},
        {	962	,	0.0000451000	},
        {	963	,	0.0000449000	},
        {	964	,	0.0000448000	},
        {	965	,	0.0000446000	},
        {	966	,	0.0000420000	},
        {	967	,	0.0000394000	},
        {	968	,	0.0000368000	},
        {	969	,	0.0000343000	},
        {	970	,	0.0000317000	},
        {	971	,	0.0000302000	},
        {	972	,	0.0000286000	},
        {	973	,	0.0000271000	},
        {	974	,	0.0000256000	},
        {	975	,	0.0000241000	},
        {	976	,	0.0000221000	},
        {	977	,	0.0000201000	},
        {	978	,	0.0000182000	},
        {	979	,	0.0000163000	},
        {	980	,	0.0000144000	},
        {	981	,	0.0000143000	},
        {	982	,	0.0000143000	},
        {	983	,	0.0000142000	},
        {	984	,	0.0000142000	},
        {	985	,	0.0000141000	},
        {	986	,	0.0000134000	},
        {	987	,	0.0000127000	},
        {	988	,	0.0000120000	},
        {	989	,	0.0000113000	},
        {	990	,	0.0000106000	},
        {	991	,	0.0000116000	},
        {	992	,	0.0000126000	},
        {	993	,	0.0000136000	},
        {	994	,	0.0000146000	},
        {	995	,	0.0000156000	},
        {	996	,	0.0000125000	},
        {	997	,	0.0000093400	},
        {	998	,	0.0000062200	},
        {	999	,	0.0000031100	},
        {	1000	,	0.0000000000	}
      };

      static const double CBERS04A_WPMB2_SRF[][2] = 
      {
        {	350	,	0.0000000000	},
        {	351	,	0.0001600000	},
        {	352	,	0.0003050000	},
        {	353	,	0.0004350000	},
        {	354	,	0.0005530000	},
        {	355	,	0.0006600000	},
        {	356	,	0.0007640000	},
        {	357	,	0.0008580000	},
        {	358	,	0.0009440000	},
        {	359	,	0.0010200000	},
        {	360	,	0.0010900000	},
        {	361	,	0.0011300000	},
        {	362	,	0.0011700000	},
        {	363	,	0.0012000000	},
        {	364	,	0.0012200000	},
        {	365	,	0.0012500000	},
        {	366	,	0.0012200000	},
        {	367	,	0.0011800000	},
        {	368	,	0.0011500000	},
        {	369	,	0.0011200000	},
        {	370	,	0.0010900000	},
        {	371	,	0.0010200000	},
        {	372	,	0.0009580000	},
        {	373	,	0.0008950000	},
        {	374	,	0.0008350000	},
        {	375	,	0.0007770000	},
        {	376	,	0.0008260000	},
        {	377	,	0.0008700000	},
        {	378	,	0.0009090000	},
        {	379	,	0.0009450000	},
        {	380	,	0.0009780000	},
        {	381	,	0.0009350000	},
        {	382	,	0.0008950000	},
        {	383	,	0.0008580000	},
        {	384	,	0.0008250000	},
        {	385	,	0.0007960000	},
        {	386	,	0.0007790000	},
        {	387	,	0.0007630000	},
        {	388	,	0.0007470000	},
        {	389	,	0.0007310000	},
        {	390	,	0.0007150000	},
        {	391	,	0.0007630000	},
        {	392	,	0.0008070000	},
        {	393	,	0.0008470000	},
        {	394	,	0.0008850000	},
        {	395	,	0.0009200000	},
        {	396	,	0.0008960000	},
        {	397	,	0.0008730000	},
        {	398	,	0.0008500000	},
        {	399	,	0.0008280000	},
        {	400	,	0.0008060000	},
        {	401	,	0.0007920000	},
        {	402	,	0.0007770000	},
        {	403	,	0.0007650000	},
        {	404	,	0.0007520000	},
        {	405	,	0.0007400000	},
        {	406	,	0.0007300000	},
        {	407	,	0.0007210000	},
        {	408	,	0.0007120000	},
        {	409	,	0.0007030000	},
        {	410	,	0.0006940000	},
        {	411	,	0.0006740000	},
        {	412	,	0.0006560000	},
        {	413	,	0.0006380000	},
        {	414	,	0.0006210000	},
        {	415	,	0.0006040000	},
        {	416	,	0.0005820000	},
        {	417	,	0.0005610000	},
        {	418	,	0.0005420000	},
        {	419	,	0.0005230000	},
        {	420	,	0.0005040000	},
        {	421	,	0.0005050000	},
        {	422	,	0.0005070000	},
        {	423	,	0.0005070000	},
        {	424	,	0.0005080000	},
        {	425	,	0.0005090000	},
        {	426	,	0.0005010000	},
        {	427	,	0.0004940000	},
        {	428	,	0.0004860000	},
        {	429	,	0.0004790000	},
        {	430	,	0.0004730000	},
        {	431	,	0.0004630000	},
        {	432	,	0.0004540000	},
        {	433	,	0.0004450000	},
        {	434	,	0.0004370000	},
        {	435	,	0.0004290000	},
        {	436	,	0.0004230000	},
        {	437	,	0.0004160000	},
        {	438	,	0.0004100000	},
        {	439	,	0.0004040000	},
        {	440	,	0.0003980000	},
        {	441	,	0.0003900000	},
        {	442	,	0.0003830000	},
        {	443	,	0.0003760000	},
        {	444	,	0.0003690000	},
        {	445	,	0.0003620000	},
        {	446	,	0.0003600000	},
        {	447	,	0.0003580000	},
        {	448	,	0.0003560000	},
        {	449	,	0.0003550000	},
        {	450	,	0.0003530000	},
        {	451	,	0.0003440000	},
        {	452	,	0.0003350000	},
        {	453	,	0.0003260000	},
        {	454	,	0.0003180000	},
        {	455	,	0.0003100000	},
        {	456	,	0.0003110000	},
        {	457	,	0.0003120000	},
        {	458	,	0.0003130000	},
        {	459	,	0.0003140000	},
        {	460	,	0.0003150000	},
        {	461	,	0.0003120000	},
        {	462	,	0.0003090000	},
        {	463	,	0.0003070000	},
        {	464	,	0.0003040000	},
        {	465	,	0.0003010000	},
        {	466	,	0.0002970000	},
        {	467	,	0.0002940000	},
        {	468	,	0.0002900000	},
        {	469	,	0.0002860000	},
        {	470	,	0.0002830000	},
        {	471	,	0.0002780000	},
        {	472	,	0.0002730000	},
        {	473	,	0.0002680000	},
        {	474	,	0.0002630000	},
        {	475	,	0.0002580000	},
        {	476	,	0.0002570000	},
        {	477	,	0.0002550000	},
        {	478	,	0.0002530000	},
        {	479	,	0.0002510000	},
        {	480	,	0.0002480000	},
        {	481	,	0.0002450000	},
        {	482	,	0.0002410000	},
        {	483	,	0.0002380000	},
        {	484	,	0.0002340000	},
        {	485	,	0.0002310000	},
        {	486	,	0.0002290000	},
        {	487	,	0.0002280000	},
        {	488	,	0.0002260000	},
        {	489	,	0.0002250000	},
        {	490	,	0.0002240000	},
        {	491	,	0.0002220000	},
        {	492	,	0.0002200000	},
        {	493	,	0.0002190000	},
        {	494	,	0.0002170000	},
        {	495	,	0.0002150000	},
        {	496	,	0.0002120000	},
        {	497	,	0.0002100000	},
        {	498	,	0.0002070000	},
        {	499	,	0.0002050000	},
        {	500	,	0.0002020000	},
        {	501	,	0.0002030000	},
        {	502	,	0.0002040000	},
        {	503	,	0.0002050000	},
        {	504	,	0.0002060000	},
        {	505	,	0.0002070000	},
        {	506	,	0.0002060000	},
        {	507	,	0.0002040000	},
        {	508	,	0.0002030000	},
        {	509	,	0.0002010000	},
        {	510	,	0.0002000000	},
        {	511	,	0.0206000000	},
        {	512	,	0.0406000000	},
        {	513	,	0.0603000000	},
        {	514	,	0.0799000000	},
        {	515	,	0.0990000000	},
        {	516	,	0.1820000000	},
        {	517	,	0.2630000000	},
        {	518	,	0.3440000000	},
        {	519	,	0.4230000000	},
        {	520	,	0.5020000000	},
        {	521	,	0.4990000000	},
        {	522	,	0.4960000000	},
        {	523	,	0.4930000000	},
        {	524	,	0.4910000000	},
        {	525	,	0.4880000000	},
        {	526	,	0.5280000000	},
        {	527	,	0.5670000000	},
        {	528	,	0.6060000000	},
        {	529	,	0.6440000000	},
        {	530	,	0.6820000000	},
        {	531	,	0.6850000000	},
        {	532	,	0.6870000000	},
        {	533	,	0.6890000000	},
        {	534	,	0.6910000000	},
        {	535	,	0.6930000000	},
        {	536	,	0.6910000000	},
        {	537	,	0.6890000000	},
        {	538	,	0.6860000000	},
        {	539	,	0.6840000000	},
        {	540	,	0.6820000000	},
        {	541	,	0.6800000000	},
        {	542	,	0.6780000000	},
        {	543	,	0.6760000000	},
        {	544	,	0.6740000000	},
        {	545	,	0.6720000000	},
        {	546	,	0.6720000000	},
        {	547	,	0.6720000000	},
        {	548	,	0.6720000000	},
        {	549	,	0.6720000000	},
        {	550	,	0.6720000000	},
        {	551	,	0.6700000000	},
        {	552	,	0.6680000000	},
        {	553	,	0.6670000000	},
        {	554	,	0.6650000000	},
        {	555	,	0.6630000000	},
        {	556	,	0.6700000000	},
        {	557	,	0.6760000000	},
        {	558	,	0.6820000000	},
        {	559	,	0.6890000000	},
        {	560	,	0.6950000000	},
        {	561	,	0.7080000000	},
        {	562	,	0.7220000000	},
        {	563	,	0.7350000000	},
        {	564	,	0.7480000000	},
        {	565	,	0.7620000000	},
        {	566	,	0.7780000000	},
        {	567	,	0.7950000000	},
        {	568	,	0.8110000000	},
        {	569	,	0.8280000000	},
        {	570	,	0.8450000000	},
        {	571	,	0.8570000000	},
        {	572	,	0.8680000000	},
        {	573	,	0.8800000000	},
        {	574	,	0.8920000000	},
        {	575	,	0.9040000000	},
        {	576	,	0.9210000000	},
        {	577	,	0.9380000000	},
        {	578	,	0.9550000000	},
        {	579	,	0.9730000000	},
        {	580	,	0.9900000000	},
        {	581	,	0.9920000000	},
        {	582	,	0.9930000000	},
        {	583	,	0.9950000000	},
        {	584	,	0.9970000000	},
        {	585	,	1.0000000000	},
        {	586	,	0.9610000000	},
        {	587	,	0.9230000000	},
        {	588	,	0.8840000000	},
        {	589	,	0.8450000000	},
        {	590	,	0.8060000000	},
        {	591	,	0.6480000000	},
        {	592	,	0.4890000000	},
        {	593	,	0.3280000000	},
        {	594	,	0.1670000000	},
        {	595	,	0.0037400000	},
        {	596	,	0.0037500000	},
        {	597	,	0.0037500000	},
        {	598	,	0.0037600000	},
        {	599	,	0.0037700000	},
        {	600	,	0.0037700000	},
        {	601	,	0.0030500000	},
        {	602	,	0.0021100000	},
        {	603	,	0.0014200000	},
        {	604	,	0.0007420000	},
        {	605	,	0.0000810000	},
        {	606	,	0.0000781000	},
        {	607	,	0.0000755000	},
        {	608	,	0.0000728000	},
        {	609	,	0.0000703000	},
        {	610	,	0.0000679000	},
        {	611	,	0.0000638000	},
        {	612	,	0.0000597000	},
        {	613	,	0.0000557000	},
        {	614	,	0.0000518000	},
        {	615	,	0.0000478000	},
        {	616	,	0.0000423000	},
        {	617	,	0.0000367000	},
        {	618	,	0.0000312000	},
        {	619	,	0.0000256000	},
        {	620	,	0.0000201000	},
        {	621	,	0.0000223000	},
        {	622	,	0.0000245000	},
        {	623	,	0.0000267000	},
        {	624	,	0.0000288000	},
        {	625	,	0.0000311000	},
        {	626	,	0.0000709000	},
        {	627	,	0.0001110000	},
        {	628	,	0.0001510000	},
        {	629	,	0.0001920000	},
        {	630	,	0.0002320000	},
        {	631	,	0.0003560000	},
        {	632	,	0.0004800000	},
        {	633	,	0.0006080000	},
        {	634	,	0.0007380000	},
        {	635	,	0.0008710000	},
        {	636	,	0.0008710000	},
        {	637	,	0.0008680000	},
        {	638	,	0.0008620000	},
        {	639	,	0.0008540000	},
        {	640	,	0.0008440000	},
        {	641	,	0.0008920000	},
        {	642	,	0.0009410000	},
        {	643	,	0.0009880000	},
        {	644	,	0.0010300000	},
        {	645	,	0.0010800000	},
        {	646	,	0.0010800000	},
        {	647	,	0.0010700000	},
        {	648	,	0.0010700000	},
        {	649	,	0.0010700000	},
        {	650	,	0.0010700000	},
        {	651	,	0.0009490000	},
        {	652	,	0.0008330000	},
        {	653	,	0.0007170000	},
        {	654	,	0.0006030000	},
        {	655	,	0.0004910000	},
        {	656	,	0.0004810000	},
        {	657	,	0.0004710000	},
        {	658	,	0.0004620000	},
        {	659	,	0.0004520000	},
        {	660	,	0.0004430000	},
        {	661	,	0.0004570000	},
        {	662	,	0.0004700000	},
        {	663	,	0.0004830000	},
        {	664	,	0.0004970000	},
        {	665	,	0.0005100000	},
        {	666	,	0.0004420000	},
        {	667	,	0.0003740000	},
        {	668	,	0.0003070000	},
        {	669	,	0.0002400000	},
        {	670	,	0.0001730000	},
        {	671	,	0.0001700000	},
        {	672	,	0.0001680000	},
        {	673	,	0.0001650000	},
        {	674	,	0.0001630000	},
        {	675	,	0.0001610000	},
        {	676	,	0.0001610000	},
        {	677	,	0.0001620000	},
        {	678	,	0.0001620000	},
        {	679	,	0.0001630000	},
        {	680	,	0.0001640000	},
        {	681	,	0.0001920000	},
        {	682	,	0.0002210000	},
        {	683	,	0.0002490000	},
        {	684	,	0.0002780000	},
        {	685	,	0.0003070000	},
        {	686	,	0.0002950000	},
        {	687	,	0.0002840000	},
        {	688	,	0.0002720000	},
        {	689	,	0.0002600000	},
        {	690	,	0.0002490000	},
        {	691	,	0.0002550000	},
        {	692	,	0.0002620000	},
        {	693	,	0.0002690000	},
        {	694	,	0.0002760000	},
        {	695	,	0.0002830000	},
        {	696	,	0.0002840000	},
        {	697	,	0.0002850000	},
        {	698	,	0.0002870000	},
        {	699	,	0.0002880000	},
        {	700	,	0.0002900000	},
        {	701	,	0.0002910000	},
        {	702	,	0.0002910000	},
        {	703	,	0.0002920000	},
        {	704	,	0.0002930000	},
        {	705	,	0.0002940000	},
        {	706	,	0.0003180000	},
        {	707	,	0.0003430000	},
        {	708	,	0.0003670000	},
        {	709	,	0.0003920000	},
        {	710	,	0.0004160000	},
        {	711	,	0.0004070000	},
        {	712	,	0.0003980000	},
        {	713	,	0.0003890000	},
        {	714	,	0.0003800000	},
        {	715	,	0.0003710000	},
        {	716	,	0.0004480000	},
        {	717	,	0.0005250000	},
        {	718	,	0.0006030000	},
        {	719	,	0.0006810000	},
        {	720	,	0.0007600000	},
        {	721	,	0.0007880000	},
        {	722	,	0.0008170000	},
        {	723	,	0.0008460000	},
        {	724	,	0.0008750000	},
        {	725	,	0.0009040000	},
        {	726	,	0.0009200000	},
        {	727	,	0.0009360000	},
        {	728	,	0.0009520000	},
        {	729	,	0.0009680000	},
        {	730	,	0.0009850000	},
        {	731	,	0.0009850000	},
        {	732	,	0.0009850000	},
        {	733	,	0.0009850000	},
        {	734	,	0.0009850000	},
        {	735	,	0.0009860000	},
        {	736	,	0.0010000000	},
        {	737	,	0.0010200000	},
        {	738	,	0.0010400000	},
        {	739	,	0.0010600000	},
        {	740	,	0.0010800000	},
        {	741	,	0.0009670000	},
        {	742	,	0.0008550000	},
        {	743	,	0.0007420000	},
        {	744	,	0.0006280000	},
        {	745	,	0.0005110000	},
        {	746	,	0.0004640000	},
        {	747	,	0.0004170000	},
        {	748	,	0.0003680000	},
        {	749	,	0.0003190000	},
        {	750	,	0.0002700000	},
        {	751	,	0.0002690000	},
        {	752	,	0.0002680000	},
        {	753	,	0.0002670000	},
        {	754	,	0.0002670000	},
        {	755	,	0.0002660000	},
        {	756	,	0.0002680000	},
        {	757	,	0.0002700000	},
        {	758	,	0.0002720000	},
        {	759	,	0.0002740000	},
        {	760	,	0.0002770000	},
        {	761	,	0.0002790000	},
        {	762	,	0.0002810000	},
        {	763	,	0.0002840000	},
        {	764	,	0.0002860000	},
        {	765	,	0.0002880000	},
        {	766	,	0.0002530000	},
        {	767	,	0.0002170000	},
        {	768	,	0.0001800000	},
        {	769	,	0.0001430000	},
        {	770	,	0.0001060000	},
        {	771	,	0.0001330000	},
        {	772	,	0.0001600000	},
        {	773	,	0.0001880000	},
        {	774	,	0.0002170000	},
        {	775	,	0.0002450000	},
        {	776	,	0.0003360000	},
        {	777	,	0.0004280000	},
        {	778	,	0.0005220000	},
        {	779	,	0.0006170000	},
        {	780	,	0.0007140000	},
        {	781	,	0.0007160000	},
        {	782	,	0.0007180000	},
        {	783	,	0.0007200000	},
        {	784	,	0.0007220000	},
        {	785	,	0.0007240000	},
        {	786	,	0.0006820000	},
        {	787	,	0.0006390000	},
        {	788	,	0.0005960000	},
        {	789	,	0.0005520000	},
        {	790	,	0.0005070000	},
        {	791	,	0.0004380000	},
        {	792	,	0.0003680000	},
        {	793	,	0.0002970000	},
        {	794	,	0.0002250000	},
        {	795	,	0.0001510000	},
        {	796	,	0.0001500000	},
        {	797	,	0.0001490000	},
        {	798	,	0.0001470000	},
        {	799	,	0.0001460000	},
        {	800	,	0.0001440000	},
        {	801	,	0.0001260000	},
        {	802	,	0.0001070000	},
        {	803	,	0.0000884000	},
        {	804	,	0.0000692000	},
        {	805	,	0.0000498000	},
        {	806	,	0.0000454000	},
        {	807	,	0.0000410000	},
        {	808	,	0.0000365000	},
        {	809	,	0.0000319000	},
        {	810	,	0.0000273000	},
        {	811	,	0.0000278000	},
        {	812	,	0.0000283000	},
        {	813	,	0.0000287000	},
        {	814	,	0.0000292000	},
        {	815	,	0.0000297000	},
        {	816	,	0.0000246000	},
        {	817	,	0.0000194000	},
        {	818	,	0.0000142000	},
        {	819	,	0.0000088700	},
        {	820	,	0.0000035400	},
        {	821	,	0.0000061400	},
        {	822	,	0.0000087500	},
        {	823	,	0.0000114000	},
        {	824	,	0.0000140000	},
        {	825	,	0.0000167000	},
        {	826	,	0.0000134000	},
        {	827	,	0.0000101000	},
        {	828	,	0.0000068500	},
        {	829	,	0.0000035600	},
        {	830	,	0.0000002580	},
        {	831	,	0.0000017000	},
        {	832	,	0.0000031400	},
        {	833	,	0.0000045800	},
        {	834	,	0.0000060100	},
        {	835	,	0.0000074400	},
        {	836	,	0.0000123000	},
        {	837	,	0.0000172000	},
        {	838	,	0.0000221000	},
        {	839	,	0.0000269000	},
        {	840	,	0.0000317000	},
        {	841	,	0.0000381000	},
        {	842	,	0.0000444000	},
        {	843	,	0.0000507000	},
        {	844	,	0.0000569000	},
        {	845	,	0.0000630000	},
        {	846	,	0.0000741000	},
        {	847	,	0.0000850000	},
        {	848	,	0.0000958000	},
        {	849	,	0.0001060000	},
        {	850	,	0.0001170000	},
        {	851	,	0.0001090000	},
        {	852	,	0.0001020000	},
        {	853	,	0.0000949000	},
        {	854	,	0.0000877000	},
        {	855	,	0.0000807000	},
        {	856	,	0.0000801000	},
        {	857	,	0.0000794000	},
        {	858	,	0.0000787000	},
        {	859	,	0.0000780000	},
        {	860	,	0.0000773000	},
        {	861	,	0.0000666000	},
        {	862	,	0.0000560000	},
        {	863	,	0.0000456000	},
        {	864	,	0.0000354000	},
        {	865	,	0.0000254000	},
        {	866	,	0.0000298000	},
        {	867	,	0.0000341000	},
        {	868	,	0.0000383000	},
        {	869	,	0.0000424000	},
        {	870	,	0.0000463000	},
        {	871	,	0.0000591000	},
        {	872	,	0.0000716000	},
        {	873	,	0.0000838000	},
        {	874	,	0.0000957000	},
        {	875	,	0.0001070000	},
        {	876	,	0.0000862000	},
        {	877	,	0.0000655000	},
        {	878	,	0.0000453000	},
        {	879	,	0.0000255000	},
        {	880	,	0.0000061100	},
        {	881	,	0.0000083800	},
        {	882	,	0.0000106000	},
        {	883	,	0.0000128000	},
        {	884	,	0.0000149000	},
        {	885	,	0.0000170000	},
        {	886	,	0.0000159000	},
        {	887	,	0.0000148000	},
        {	888	,	0.0000138000	},
        {	889	,	0.0000128000	},
        {	890	,	0.0000118000	},
        {	891	,	0.0000122000	},
        {	892	,	0.0000125000	},
        {	893	,	0.0000127000	},
        {	894	,	0.0000130000	},
        {	895	,	0.0000133000	},
        {	896	,	0.0000276000	},
        {	897	,	0.0000415000	},
        {	898	,	0.0000552000	},
        {	899	,	0.0000686000	},
        {	900	,	0.0000816000	},
        {	901	,	0.0000808000	},
        {	902	,	0.0000800000	},
        {	903	,	0.0000791000	},
        {	904	,	0.0000783000	},
        {	905	,	0.0000775000	},
        {	906	,	0.0000767000	},
        {	907	,	0.0000760000	},
        {	908	,	0.0000753000	},
        {	909	,	0.0000745000	},
        {	910	,	0.0000739000	},
        {	911	,	0.0000684000	},
        {	912	,	0.0000630000	},
        {	913	,	0.0000577000	},
        {	914	,	0.0000526000	},
        {	915	,	0.0000475000	},
        {	916	,	0.0000456000	},
        {	917	,	0.0000437000	},
        {	918	,	0.0000419000	},
        {	919	,	0.0000400000	},
        {	920	,	0.0000383000	},
        {	921	,	0.0000379000	},
        {	922	,	0.0000376000	},
        {	923	,	0.0000373000	},
        {	924	,	0.0000370000	},
        {	925	,	0.0000367000	},
        {	926	,	0.0000365000	},
        {	927	,	0.0000362000	},
        {	928	,	0.0000360000	},
        {	929	,	0.0000357000	},
        {	930	,	0.0000355000	},
        {	931	,	0.0000353000	},
        {	932	,	0.0000351000	},
        {	933	,	0.0000348000	},
        {	934	,	0.0000346000	},
        {	935	,	0.0000344000	},
        {	936	,	0.0000341000	},
        {	937	,	0.0000338000	},
        {	938	,	0.0000335000	},
        {	939	,	0.0000332000	},
        {	940	,	0.0000329000	},
        {	941	,	0.0000327000	},
        {	942	,	0.0000325000	},
        {	943	,	0.0000324000	},
        {	944	,	0.0000322000	},
        {	945	,	0.0000320000	},
        {	946	,	0.0000304000	},
        {	947	,	0.0000289000	},
        {	948	,	0.0000274000	},
        {	949	,	0.0000259000	},
        {	950	,	0.0000244000	},
        {	951	,	0.0000229000	},
        {	952	,	0.0000214000	},
        {	953	,	0.0000199000	},
        {	954	,	0.0000185000	},
        {	955	,	0.0000170000	},
        {	956	,	0.0000178000	},
        {	957	,	0.0000185000	},
        {	958	,	0.0000193000	},
        {	959	,	0.0000200000	},
        {	960	,	0.0000207000	},
        {	961	,	0.0000212000	},
        {	962	,	0.0000218000	},
        {	963	,	0.0000223000	},
        {	964	,	0.0000228000	},
        {	965	,	0.0000233000	},
        {	966	,	0.0000223000	},
        {	967	,	0.0000214000	},
        {	968	,	0.0000204000	},
        {	969	,	0.0000194000	},
        {	970	,	0.0000185000	},
        {	971	,	0.0000183000	},
        {	972	,	0.0000181000	},
        {	973	,	0.0000179000	},
        {	974	,	0.0000177000	},
        {	975	,	0.0000175000	},
        {	976	,	0.0000174000	},
        {	977	,	0.0000174000	},
        {	978	,	0.0000173000	},
        {	979	,	0.0000172000	},
        {	980	,	0.0000172000	},
        {	981	,	0.0000163000	},
        {	982	,	0.0000154000	},
        {	983	,	0.0000146000	},
        {	984	,	0.0000137000	},
        {	985	,	0.0000129000	},
        {	986	,	0.0000127000	},
        {	987	,	0.0000126000	},
        {	988	,	0.0000124000	},
        {	989	,	0.0000123000	},
        {	990	,	0.0000122000	},
        {	991	,	0.0000096900	},
        {	992	,	0.0000072500	},
        {	993	,	0.0000048200	},
        {	994	,	0.0000024000	},
        {	995	,	0.0000000000	},
        {	996	,	0.0000964000	},
        {	997	,	0.0001930000	},
        {	998	,	0.0002890000	},
        {	999	,	0.0003840000	},
        {	1000	,	0.0004800000	}
      };
      
      static const double CBERS04A_WPMB3_SRF[][2] = 
      {
        {	350	,	0.0000000000	},
        {	351	,	0.0007720000	},
        {	352	,	0.0014700000	},
        {	353	,	0.0020900000	},
        {	354	,	0.0026600000	},
        {	355	,	0.0031800000	},
        {	356	,	0.0033700000	},
        {	357	,	0.0035500000	},
        {	358	,	0.0037100000	},
        {	359	,	0.0038600000	},
        {	360	,	0.0039800000	},
        {	361	,	0.0040500000	},
        {	362	,	0.0040900000	},
        {	363	,	0.0041300000	},
        {	364	,	0.0041700000	},
        {	365	,	0.0042000000	},
        {	366	,	0.0041400000	},
        {	367	,	0.0040900000	},
        {	368	,	0.0040400000	},
        {	369	,	0.0039800000	},
        {	370	,	0.0039200000	},
        {	371	,	0.0038200000	},
        {	372	,	0.0037300000	},
        {	373	,	0.0036300000	},
        {	374	,	0.0035400000	},
        {	375	,	0.0034500000	},
        {	376	,	0.0034600000	},
        {	377	,	0.0034700000	},
        {	378	,	0.0034700000	},
        {	379	,	0.0034700000	},
        {	380	,	0.0034700000	},
        {	381	,	0.0032900000	},
        {	382	,	0.0031200000	},
        {	383	,	0.0029700000	},
        {	384	,	0.0028300000	},
        {	385	,	0.0027100000	},
        {	386	,	0.0027100000	},
        {	387	,	0.0027100000	},
        {	388	,	0.0027200000	},
        {	389	,	0.0027200000	},
        {	390	,	0.0027200000	},
        {	391	,	0.0026600000	},
        {	392	,	0.0026000000	},
        {	393	,	0.0025300000	},
        {	394	,	0.0024700000	},
        {	395	,	0.0024100000	},
        {	396	,	0.0023600000	},
        {	397	,	0.0023200000	},
        {	398	,	0.0022700000	},
        {	399	,	0.0022300000	},
        {	400	,	0.0021800000	},
        {	401	,	0.0021400000	},
        {	402	,	0.0020900000	},
        {	403	,	0.0020500000	},
        {	404	,	0.0020100000	},
        {	405	,	0.0019700000	},
        {	406	,	0.0019200000	},
        {	407	,	0.0018800000	},
        {	408	,	0.0018300000	},
        {	409	,	0.0017900000	},
        {	410	,	0.0017500000	},
        {	411	,	0.0017100000	},
        {	412	,	0.0016800000	},
        {	413	,	0.0016500000	},
        {	414	,	0.0016200000	},
        {	415	,	0.0015900000	},
        {	416	,	0.0015400000	},
        {	417	,	0.0015000000	},
        {	418	,	0.0014600000	},
        {	419	,	0.0014200000	},
        {	420	,	0.0013800000	},
        {	421	,	0.0013600000	},
        {	422	,	0.0013300000	},
        {	423	,	0.0013100000	},
        {	424	,	0.0012800000	},
        {	425	,	0.0012600000	},
        {	426	,	0.0012300000	},
        {	427	,	0.0012100000	},
        {	428	,	0.0011900000	},
        {	429	,	0.0011600000	},
        {	430	,	0.0011400000	},
        {	431	,	0.0011200000	},
        {	432	,	0.0011100000	},
        {	433	,	0.0010900000	},
        {	434	,	0.0010700000	},
        {	435	,	0.0010600000	},
        {	436	,	0.0010400000	},
        {	437	,	0.0010300000	},
        {	438	,	0.0010100000	},
        {	439	,	0.0009930000	},
        {	440	,	0.0009780000	},
        {	441	,	0.0009580000	},
        {	442	,	0.0009390000	},
        {	443	,	0.0009220000	},
        {	444	,	0.0009050000	},
        {	445	,	0.0008870000	},
        {	446	,	0.0008750000	},
        {	447	,	0.0008640000	},
        {	448	,	0.0008520000	},
        {	449	,	0.0008410000	},
        {	450	,	0.0008300000	},
        {	451	,	0.0008190000	},
        {	452	,	0.0008090000	},
        {	453	,	0.0007990000	},
        {	454	,	0.0007900000	},
        {	455	,	0.0007800000	},
        {	456	,	0.0007690000	},
        {	457	,	0.0007580000	},
        {	458	,	0.0007470000	},
        {	459	,	0.0007370000	},
        {	460	,	0.0007270000	},
        {	461	,	0.0007170000	},
        {	462	,	0.0007070000	},
        {	463	,	0.0006980000	},
        {	464	,	0.0006880000	},
        {	465	,	0.0006790000	},
        {	466	,	0.0006730000	},
        {	467	,	0.0006670000	},
        {	468	,	0.0006610000	},
        {	469	,	0.0006560000	},
        {	470	,	0.0006510000	},
        {	471	,	0.0006420000	},
        {	472	,	0.0006340000	},
        {	473	,	0.0006260000	},
        {	474	,	0.0006180000	},
        {	475	,	0.0006090000	},
        {	476	,	0.0006020000	},
        {	477	,	0.0005940000	},
        {	478	,	0.0005860000	},
        {	479	,	0.0005780000	},
        {	480	,	0.0005700000	},
        {	481	,	0.0005640000	},
        {	482	,	0.0005590000	},
        {	483	,	0.0005530000	},
        {	484	,	0.0005480000	},
        {	485	,	0.0005420000	},
        {	486	,	0.0005320000	},
        {	487	,	0.0005230000	},
        {	488	,	0.0005140000	},
        {	489	,	0.0005050000	},
        {	490	,	0.0004960000	},
        {	491	,	0.0004930000	},
        {	492	,	0.0004900000	},
        {	493	,	0.0004880000	},
        {	494	,	0.0004850000	},
        {	495	,	0.0004830000	},
        {	496	,	0.0004880000	},
        {	497	,	0.0004940000	},
        {	498	,	0.0004990000	},
        {	499	,	0.0005040000	},
        {	500	,	0.0005090000	},
        {	501	,	0.0005070000	},
        {	502	,	0.0005040000	},
        {	503	,	0.0005020000	},
        {	504	,	0.0005000000	},
        {	505	,	0.0004980000	},
        {	506	,	0.0004950000	},
        {	507	,	0.0004920000	},
        {	508	,	0.0004880000	},
        {	509	,	0.0004850000	},
        {	510	,	0.0004820000	},
        {	511	,	0.0004800000	},
        {	512	,	0.0004780000	},
        {	513	,	0.0004750000	},
        {	514	,	0.0004740000	},
        {	515	,	0.0004710000	},
        {	516	,	0.0004650000	},
        {	517	,	0.0004590000	},
        {	518	,	0.0004520000	},
        {	519	,	0.0004460000	},
        {	520	,	0.0004400000	},
        {	521	,	0.0004630000	},
        {	522	,	0.0004850000	},
        {	523	,	0.0005070000	},
        {	524	,	0.0005290000	},
        {	525	,	0.0005510000	},
        {	526	,	0.0005720000	},
        {	527	,	0.0005920000	},
        {	528	,	0.0006120000	},
        {	529	,	0.0006320000	},
        {	530	,	0.0006510000	},
        {	531	,	0.0007060000	},
        {	532	,	0.0007600000	},
        {	533	,	0.0008140000	},
        {	534	,	0.0008680000	},
        {	535	,	0.0009210000	},
        {	536	,	0.0008610000	},
        {	537	,	0.0008030000	},
        {	538	,	0.0007440000	},
        {	539	,	0.0006870000	},
        {	540	,	0.0006300000	},
        {	541	,	0.0006960000	},
        {	542	,	0.0007620000	},
        {	543	,	0.0008270000	},
        {	544	,	0.0008920000	},
        {	545	,	0.0009560000	},
        {	546	,	0.0009520000	},
        {	547	,	0.0009480000	},
        {	548	,	0.0009440000	},
        {	549	,	0.0009400000	},
        {	550	,	0.0009360000	},
        {	551	,	0.0008200000	},
        {	552	,	0.0007040000	},
        {	553	,	0.0005880000	},
        {	554	,	0.0004740000	},
        {	555	,	0.0003600000	},
        {	556	,	0.0003950000	},
        {	557	,	0.0004310000	},
        {	558	,	0.0004660000	},
        {	559	,	0.0005020000	},
        {	560	,	0.0005370000	},
        {	561	,	0.0005340000	},
        {	562	,	0.0005320000	},
        {	563	,	0.0005300000	},
        {	564	,	0.0005270000	},
        {	565	,	0.0005250000	},
        {	566	,	0.0005660000	},
        {	567	,	0.0006060000	},
        {	568	,	0.0006470000	},
        {	569	,	0.0006870000	},
        {	570	,	0.0007280000	},
        {	571	,	0.0007670000	},
        {	572	,	0.0008070000	},
        {	573	,	0.0008470000	},
        {	574	,	0.0008870000	},
        {	575	,	0.0009270000	},
        {	576	,	0.0009170000	},
        {	577	,	0.0009090000	},
        {	578	,	0.0008990000	},
        {	579	,	0.0008900000	},
        {	580	,	0.0008820000	},
        {	581	,	0.0008830000	},
        {	582	,	0.0008850000	},
        {	583	,	0.0008860000	},
        {	584	,	0.0008880000	},
        {	585	,	0.0008910000	},
        {	586	,	0.0008520000	},
        {	587	,	0.0008150000	},
        {	588	,	0.0007760000	},
        {	589	,	0.0007380000	},
        {	590	,	0.0007000000	},
        {	591	,	0.0006730000	},
        {	592	,	0.0006470000	},
        {	593	,	0.0006200000	},
        {	594	,	0.0005940000	},
        {	595	,	0.0005670000	},
        {	596	,	0.0005330000	},
        {	597	,	0.0005000000	},
        {	598	,	0.0004660000	},
        {	599	,	0.0004310000	},
        {	600	,	0.0003970000	},
        {	601	,	0.0003960000	},
        {	602	,	0.0003590000	},
        {	603	,	0.0003520000	},
        {	604	,	0.0003460000	},
        {	605	,	0.0003410000	},
        {	606	,	0.0003360000	},
        {	607	,	0.0003320000	},
        {	608	,	0.0003280000	},
        {	609	,	0.0003240000	},
        {	610	,	0.0003210000	},
        {	611	,	0.0003190000	},
        {	612	,	0.0003170000	},
        {	613	,	0.0003160000	},
        {	614	,	0.0003150000	},
        {	615	,	0.0003130000	},
        {	616	,	0.0005790000	},
        {	617	,	0.0008450000	},
        {	618	,	0.0011100000	},
        {	619	,	0.0013800000	},
        {	620	,	0.0016400000	},
        {	621	,	0.0016400000	},
        {	622	,	0.0016400000	},
        {	623	,	0.0016500000	},
        {	624	,	0.0016500000	},
        {	625	,	0.0016500000	},
        {	626	,	0.0408000000	},
        {	627	,	0.0801000000	},
        {	628	,	0.1200000000	},
        {	629	,	0.1590000000	},
        {	630	,	0.2000000000	},
        {	631	,	0.2320000000	},
        {	632	,	0.2660000000	},
        {	633	,	0.3000000000	},
        {	634	,	0.3360000000	},
        {	635	,	0.3720000000	},
        {	636	,	0.4980000000	},
        {	637	,	0.6230000000	},
        {	638	,	0.7460000000	},
        {	639	,	0.8670000000	},
        {	640	,	0.9870000000	},
        {	641	,	0.9800000000	},
        {	642	,	0.9740000000	},
        {	643	,	0.9680000000	},
        {	644	,	0.9620000000	},
        {	645	,	0.9560000000	},
        {	646	,	0.9650000000	},
        {	647	,	0.9740000000	},
        {	648	,	0.9830000000	},
        {	649	,	0.9920000000	},
        {	650	,	1.0000000000	},
        {	651	,	0.9940000000	},
        {	652	,	0.9880000000	},
        {	653	,	0.9830000000	},
        {	654	,	0.9770000000	},
        {	655	,	0.9720000000	},
        {	656	,	0.9680000000	},
        {	657	,	0.9630000000	},
        {	658	,	0.9590000000	},
        {	659	,	0.9550000000	},
        {	660	,	0.9510000000	},
        {	661	,	0.9480000000	},
        {	662	,	0.9450000000	},
        {	663	,	0.9420000000	},
        {	664	,	0.9400000000	},
        {	665	,	0.9370000000	},
        {	666	,	0.9360000000	},
        {	667	,	0.9340000000	},
        {	668	,	0.9330000000	},
        {	669	,	0.9320000000	},
        {	670	,	0.9310000000	},
        {	671	,	0.9300000000	},
        {	672	,	0.9290000000	},
        {	673	,	0.9290000000	},
        {	674	,	0.9280000000	},
        {	675	,	0.9270000000	},
        {	676	,	0.9270000000	},
        {	677	,	0.9270000000	},
        {	678	,	0.9270000000	},
        {	679	,	0.9280000000	},
        {	680	,	0.9280000000	},
        {	681	,	0.9290000000	},
        {	682	,	0.9300000000	},
        {	683	,	0.9310000000	},
        {	684	,	0.9320000000	},
        {	685	,	0.9340000000	},
        {	686	,	0.9350000000	},
        {	687	,	0.9360000000	},
        {	688	,	0.9380000000	},
        {	689	,	0.9390000000	},
        {	690	,	0.9410000000	},
        {	691	,	0.8540000000	},
        {	692	,	0.7670000000	},
        {	693	,	0.6790000000	},
        {	694	,	0.5900000000	},
        {	695	,	0.5020000000	},
        {	696	,	0.4110000000	},
        {	697	,	0.3200000000	},
        {	698	,	0.2280000000	},
        {	699	,	0.1360000000	},
        {	700	,	0.0435000000	},
        {	701	,	0.0357000000	},
        {	702	,	0.0278000000	},
        {	703	,	0.0199000000	},
        {	704	,	0.0119000000	},
        {	705	,	0.0038700000	},
        {	706	,	0.0031100000	},
        {	707	,	0.0023400000	},
        {	708	,	0.0015700000	},
        {	709	,	0.0007860000	},
        {	710	,	0.0000009060	},
        {	711	,	0.0000016000	},
        {	712	,	0.0000022900	},
        {	713	,	0.0000029900	},
        {	714	,	0.0000037000	},
        {	715	,	0.0000044100	},
        {	716	,	0.0000043700	},
        {	717	,	0.0000043200	},
        {	718	,	0.0000042800	},
        {	719	,	0.0000042400	},
        {	720	,	0.0000042000	},
        {	721	,	0.0000051400	},
        {	722	,	0.0000061000	},
        {	723	,	0.0000070700	},
        {	724	,	0.0000080400	},
        {	725	,	0.0000090300	},
        {	726	,	0.0000085300	},
        {	727	,	0.0000080400	},
        {	728	,	0.0000075400	},
        {	729	,	0.0000070300	},
        {	730	,	0.0000065100	},
        {	731	,	0.0000067700	},
        {	732	,	0.0000070300	},
        {	733	,	0.0000072900	},
        {	734	,	0.0000075500	},
        {	735	,	0.0000078200	},
        {	736	,	0.0000064700	},
        {	737	,	0.0000051100	},
        {	738	,	0.0000037400	},
        {	739	,	0.0000023400	},
        {	740	,	0.0000009340	},
        {	741	,	0.0000031200	},
        {	742	,	0.0000053500	},
        {	743	,	0.0000075900	},
        {	744	,	0.0000098700	},
        {	745	,	0.0000122000	},
        {	746	,	0.0000101000	},
        {	747	,	0.0000079900	},
        {	748	,	0.0000058600	},
        {	749	,	0.0000036900	},
        {	750	,	0.0000015000	},
        {	751	,	0.0000037700	},
        {	752	,	0.0000060700	},
        {	753	,	0.0000084000	},
        {	754	,	0.0000108000	},
        {	755	,	0.0000132000	},
        {	756	,	0.0000114000	},
        {	757	,	0.0000096400	},
        {	758	,	0.0000078500	},
        {	759	,	0.0000060200	},
        {	760	,	0.0000041700	},
        {	761	,	0.0000060400	},
        {	762	,	0.0000079500	},
        {	763	,	0.0000098900	},
        {	764	,	0.0000118000	},
        {	765	,	0.0000138000	},
        {	766	,	0.0000486000	},
        {	767	,	0.0000838000	},
        {	768	,	0.0001200000	},
        {	769	,	0.0001560000	},
        {	770	,	0.0001930000	},
        {	771	,	0.0001650000	},
        {	772	,	0.0001370000	},
        {	773	,	0.0001090000	},
        {	774	,	0.0000801000	},
        {	775	,	0.0000507000	},
        {	776	,	0.0000463000	},
        {	777	,	0.0000418000	},
        {	778	,	0.0000373000	},
        {	779	,	0.0000326000	},
        {	780	,	0.0000279000	},
        {	781	,	0.0000680000	},
        {	782	,	0.0001090000	},
        {	783	,	0.0001500000	},
        {	784	,	0.0001930000	},
        {	785	,	0.0002350000	},
        {	786	,	0.0002670000	},
        {	787	,	0.0002990000	},
        {	788	,	0.0003310000	},
        {	789	,	0.0003640000	},
        {	790	,	0.0003980000	},
        {	791	,	0.0003260000	},
        {	792	,	0.0002530000	},
        {	793	,	0.0001790000	},
        {	794	,	0.0001030000	},
        {	795	,	0.0000264000	},
        {	796	,	0.0001050000	},
        {	797	,	0.0001850000	},
        {	798	,	0.0002660000	},
        {	799	,	0.0003480000	},
        {	800	,	0.0004310000	},
        {	801	,	0.0004370000	},
        {	802	,	0.0004420000	},
        {	803	,	0.0004480000	},
        {	804	,	0.0004530000	},
        {	805	,	0.0004590000	},
        {	806	,	0.0004610000	},
        {	807	,	0.0004620000	},
        {	808	,	0.0004640000	},
        {	809	,	0.0004660000	},
        {	810	,	0.0004680000	},
        {	811	,	0.0004260000	},
        {	812	,	0.0003830000	},
        {	813	,	0.0003400000	},
        {	814	,	0.0002970000	},
        {	815	,	0.0002530000	},
        {	816	,	0.0002520000	},
        {	817	,	0.0002510000	},
        {	818	,	0.0002500000	},
        {	819	,	0.0002490000	},
        {	820	,	0.0002480000	},
        {	821	,	0.0002330000	},
        {	822	,	0.0002180000	},
        {	823	,	0.0002030000	},
        {	824	,	0.0001870000	},
        {	825	,	0.0001720000	},
        {	826	,	0.0001720000	},
        {	827	,	0.0001720000	},
        {	828	,	0.0001730000	},
        {	829	,	0.0001730000	},
        {	830	,	0.0001730000	},
        {	831	,	0.0001890000	},
        {	832	,	0.0002040000	},
        {	833	,	0.0002190000	},
        {	834	,	0.0002350000	},
        {	835	,	0.0002500000	},
        {	836	,	0.0002710000	},
        {	837	,	0.0002920000	},
        {	838	,	0.0003130000	},
        {	839	,	0.0003340000	},
        {	840	,	0.0003550000	},
        {	841	,	0.0003570000	},
        {	842	,	0.0003600000	},
        {	843	,	0.0003620000	},
        {	844	,	0.0003630000	},
        {	845	,	0.0003650000	},
        {	846	,	0.0006150000	},
        {	847	,	0.0008610000	},
        {	848	,	0.0011000000	},
        {	849	,	0.0013500000	},
        {	850	,	0.0015800000	},
        {	851	,	0.0012600000	},
        {	852	,	0.0009500000	},
        {	853	,	0.0006400000	},
        {	854	,	0.0003350000	},
        {	855	,	0.0000347000	},
        {	856	,	0.0000358000	},
        {	857	,	0.0000368000	},
        {	858	,	0.0000379000	},
        {	859	,	0.0000389000	},
        {	860	,	0.0000398000	},
        {	861	,	0.0000507000	},
        {	862	,	0.0000614000	},
        {	863	,	0.0000719000	},
        {	864	,	0.0000822000	},
        {	865	,	0.0000922000	},
        {	866	,	0.0000742000	},
        {	867	,	0.0000565000	},
        {	868	,	0.0000392000	},
        {	869	,	0.0000222000	},
        {	870	,	0.0000056100	},
        {	871	,	0.0000065900	},
        {	872	,	0.0000075600	},
        {	873	,	0.0000084900	},
        {	874	,	0.0000094100	},
        {	875	,	0.0000103000	},
        {	876	,	0.0000098900	},
        {	877	,	0.0000094700	},
        {	878	,	0.0000090600	},
        {	879	,	0.0000086600	},
        {	880	,	0.0000082700	},
        {	881	,	0.0000088500	},
        {	882	,	0.0000094100	},
        {	883	,	0.0000099600	},
        {	884	,	0.0000105000	},
        {	885	,	0.0000110000	},
        {	886	,	0.0000201000	},
        {	887	,	0.0000289000	},
        {	888	,	0.0000376000	},
        {	889	,	0.0000460000	},
        {	890	,	0.0000542000	},
        {	891	,	0.0002580000	},
        {	892	,	0.0004560000	},
        {	893	,	0.0006500000	},
        {	894	,	0.0008400000	},
        {	895	,	0.0010300000	},
        {	896	,	0.0010500000	},
        {	897	,	0.0010700000	},
        {	898	,	0.0010900000	},
        {	899	,	0.0011100000	},
        {	900	,	0.0011300000	},
        {	901	,	0.0010400000	},
        {	902	,	0.0009540000	},
        {	903	,	0.0008690000	},
        {	904	,	0.0007860000	},
        {	905	,	0.0007040000	},
        {	906	,	0.0005680000	},
        {	907	,	0.0004350000	},
        {	908	,	0.0003040000	},
        {	909	,	0.0001750000	},
        {	910	,	0.0000493000	},
        {	911	,	0.0000396000	},
        {	912	,	0.0000301000	},
        {	913	,	0.0000208000	},
        {	914	,	0.0000117000	},
        {	915	,	0.0000027500	},
        {	916	,	0.0000031600	},
        {	917	,	0.0000035700	},
        {	918	,	0.0000039800	},
        {	919	,	0.0000043700	},
        {	920	,	0.0000047600	},
        {	921	,	0.0000062800	},
        {	922	,	0.0000077700	},
        {	923	,	0.0000092400	},
        {	924	,	0.0000107000	},
        {	925	,	0.0000121000	},
        {	926	,	0.0000128000	},
        {	927	,	0.0000135000	},
        {	928	,	0.0000142000	},
        {	929	,	0.0000148000	},
        {	930	,	0.0000155000	},
        {	931	,	0.0000159000	},
        {	932	,	0.0000162000	},
        {	933	,	0.0000166000	},
        {	934	,	0.0000169000	},
        {	935	,	0.0000173000	},
        {	936	,	0.0000178000	},
        {	937	,	0.0000184000	},
        {	938	,	0.0000189000	},
        {	939	,	0.0000194000	},
        {	940	,	0.0000199000	},
        {	941	,	0.0000194000	},
        {	942	,	0.0000189000	},
        {	943	,	0.0000184000	},
        {	944	,	0.0000179000	},
        {	945	,	0.0000174000	},
        {	946	,	0.0000179000	},
        {	947	,	0.0000184000	},
        {	948	,	0.0000189000	},
        {	949	,	0.0000194000	},
        {	950	,	0.0000199000	},
        {	951	,	0.0000165000	},
        {	952	,	0.0000132000	},
        {	953	,	0.0000099100	},
        {	954	,	0.0000066600	},
        {	955	,	0.0000034400	},
        {	956	,	0.0000028300	},
        {	957	,	0.0000022300	},
        {	958	,	0.0000016300	},
        {	959	,	0.0000010300	},
        {	960	,	0.0000004480	},
        {	961	,	0.0000008460	},
        {	962	,	0.0000012400	},
        {	963	,	0.0000016300	},
        {	964	,	0.0000020200	},
        {	965	,	0.0000024000	},
        {	966	,	0.0000023300	},
        {	967	,	0.0000022600	},
        {	968	,	0.0000021900	},
        {	969	,	0.0000021300	},
        {	970	,	0.0000020600	},
        {	971	,	0.0000021900	},
        {	972	,	0.0000023300	},
        {	973	,	0.0000024600	},
        {	974	,	0.0000025900	},
        {	975	,	0.0000027200	},
        {	976	,	0.0000033900	},
        {	977	,	0.0000040600	},
        {	978	,	0.0000047200	},
        {	979	,	0.0000053800	},
        {	980	,	0.0000060300	},
        {	981	,	0.0000058200	},
        {	982	,	0.0000056100	},
        {	983	,	0.0000054000	},
        {	984	,	0.0000051900	},
        {	985	,	0.0000049800	},
        {	986	,	0.0000043100	},
        {	987	,	0.0000036400	},
        {	988	,	0.0000029700	},
        {	989	,	0.0000023100	},
        {	990	,	0.0000016600	},
        {	991	,	0.0000014100	},
        {	992	,	0.0000011700	},
        {	993	,	0.0000009310	},
        {	994	,	0.0000006930	},
        {	995	,	0.0000004570	},
        {	996	,	0.0000003650	},
        {	997	,	0.0000002730	},
        {	998	,	0.0000001820	},
        {	999	,	0.0000000909	},
        {	1000	,	0.0000000000	}
      };
      
      static const double CBERS04A_WPMB4_SRF[][2] = 
      {
        {	350	,	0.0000000000	},
        {	351	,	0.0000124000	},
        {	352	,	0.0000236000	},
        {	353	,	0.0000337000	},
        {	354	,	0.0000428000	},
        {	355	,	0.0000511000	},
        {	356	,	0.0001030000	},
        {	357	,	0.0001510000	},
        {	358	,	0.0001950000	},
        {	359	,	0.0002350000	},
        {	360	,	0.0002720000	},
        {	361	,	0.0002930000	},
        {	362	,	0.0003120000	},
        {	363	,	0.0003290000	},
        {	364	,	0.0003440000	},
        {	365	,	0.0003590000	},
        {	366	,	0.0003550000	},
        {	367	,	0.0003510000	},
        {	368	,	0.0003480000	},
        {	369	,	0.0003430000	},
        {	370	,	0.0003390000	},
        {	371	,	0.0003530000	},
        {	372	,	0.0003650000	},
        {	373	,	0.0003770000	},
        {	374	,	0.0003860000	},
        {	375	,	0.0003950000	},
        {	376	,	0.0004000000	},
        {	377	,	0.0004040000	},
        {	378	,	0.0004070000	},
        {	379	,	0.0004090000	},
        {	380	,	0.0004110000	},
        {	381	,	0.0004140000	},
        {	382	,	0.0004160000	},
        {	383	,	0.0004180000	},
        {	384	,	0.0004210000	},
        {	385	,	0.0004260000	},
        {	386	,	0.0004260000	},
        {	387	,	0.0004260000	},
        {	388	,	0.0004270000	},
        {	389	,	0.0004260000	},
        {	390	,	0.0004260000	},
        {	391	,	0.0004560000	},
        {	392	,	0.0004850000	},
        {	393	,	0.0005110000	},
        {	394	,	0.0005350000	},
        {	395	,	0.0005580000	},
        {	396	,	0.0005690000	},
        {	397	,	0.0005800000	},
        {	398	,	0.0005890000	},
        {	399	,	0.0005970000	},
        {	400	,	0.0006040000	},
        {	401	,	0.0005930000	},
        {	402	,	0.0005810000	},
        {	403	,	0.0005710000	},
        {	404	,	0.0005600000	},
        {	405	,	0.0005500000	},
        {	406	,	0.0005180000	},
        {	407	,	0.0004880000	},
        {	408	,	0.0004590000	},
        {	409	,	0.0004310000	},
        {	410	,	0.0004040000	},
        {	411	,	0.0004220000	},
        {	412	,	0.0004390000	},
        {	413	,	0.0004540000	},
        {	414	,	0.0004690000	},
        {	415	,	0.0004830000	},
        {	416	,	0.0005050000	},
        {	417	,	0.0005260000	},
        {	418	,	0.0005470000	},
        {	419	,	0.0005660000	},
        {	420	,	0.0005840000	},
        {	421	,	0.0005610000	},
        {	422	,	0.0005390000	},
        {	423	,	0.0005180000	},
        {	424	,	0.0004980000	},
        {	425	,	0.0004790000	},
        {	426	,	0.0005130000	},
        {	427	,	0.0005450000	},
        {	428	,	0.0005760000	},
        {	429	,	0.0006060000	},
        {	430	,	0.0006350000	},
        {	431	,	0.0007890000	},
        {	432	,	0.0009380000	},
        {	433	,	0.0010800000	},
        {	434	,	0.0012200000	},
        {	435	,	0.0013500000	},
        {	436	,	0.0020800000	},
        {	437	,	0.0027900000	},
        {	438	,	0.0034700000	},
        {	439	,	0.0041300000	},
        {	440	,	0.0047600000	},
        {	441	,	0.0050400000	},
        {	442	,	0.0053100000	},
        {	443	,	0.0055800000	},
        {	444	,	0.0058400000	},
        {	445	,	0.0060800000	},
        {	446	,	0.0063300000	},
        {	447	,	0.0065700000	},
        {	448	,	0.0068000000	},
        {	449	,	0.0070300000	},
        {	450	,	0.0072400000	},
        {	451	,	0.0074800000	},
        {	452	,	0.0077100000	},
        {	453	,	0.0079300000	},
        {	454	,	0.0081500000	},
        {	455	,	0.0083600000	},
        {	456	,	0.0082200000	},
        {	457	,	0.0081000000	},
        {	458	,	0.0079700000	},
        {	459	,	0.0078500000	},
        {	460	,	0.0077400000	},
        {	461	,	0.0076500000	},
        {	462	,	0.0075500000	},
        {	463	,	0.0074600000	},
        {	464	,	0.0073700000	},
        {	465	,	0.0072700000	},
        {	466	,	0.0072300000	},
        {	467	,	0.0071900000	},
        {	468	,	0.0071500000	},
        {	469	,	0.0071100000	},
        {	470	,	0.0070800000	},
        {	471	,	0.0069000000	},
        {	472	,	0.0067300000	},
        {	473	,	0.0065600000	},
        {	474	,	0.0064000000	},
        {	475	,	0.0062300000	},
        {	476	,	0.0064700000	},
        {	477	,	0.0067000000	},
        {	478	,	0.0069300000	},
        {	479	,	0.0071400000	},
        {	480	,	0.0073400000	},
        {	481	,	0.0066800000	},
        {	482	,	0.0060400000	},
        {	483	,	0.0054100000	},
        {	484	,	0.0048100000	},
        {	485	,	0.0042200000	},
        {	486	,	0.0056100000	},
        {	487	,	0.0069700000	},
        {	488	,	0.0082900000	},
        {	489	,	0.0095900000	},
        {	490	,	0.0109000000	},
        {	491	,	0.0107000000	},
        {	492	,	0.0106000000	},
        {	493	,	0.0105000000	},
        {	494	,	0.0104000000	},
        {	495	,	0.0103000000	},
        {	496	,	0.0111000000	},
        {	497	,	0.0119000000	},
        {	498	,	0.0126000000	},
        {	499	,	0.0134000000	},
        {	500	,	0.0142000000	},
        {	501	,	0.0140000000	},
        {	502	,	0.0139000000	},
        {	503	,	0.0138000000	},
        {	504	,	0.0137000000	},
        {	505	,	0.0135000000	},
        {	506	,	0.0134000000	},
        {	507	,	0.0133000000	},
        {	508	,	0.0132000000	},
        {	509	,	0.0131000000	},
        {	510	,	0.0130000000	},
        {	511	,	0.0120000000	},
        {	512	,	0.0109000000	},
        {	513	,	0.0099200000	},
        {	514	,	0.0089100000	},
        {	515	,	0.0079100000	},
        {	516	,	0.0079300000	},
        {	517	,	0.0079500000	},
        {	518	,	0.0079600000	},
        {	519	,	0.0079800000	},
        {	520	,	0.0080000000	},
        {	521	,	0.0081900000	},
        {	522	,	0.0083800000	},
        {	523	,	0.0085700000	},
        {	524	,	0.0087500000	},
        {	525	,	0.0089300000	},
        {	526	,	0.0097400000	},
        {	527	,	0.0105000000	},
        {	528	,	0.0113000000	},
        {	529	,	0.0121000000	},
        {	530	,	0.0129000000	},
        {	531	,	0.0132000000	},
        {	532	,	0.0136000000	},
        {	533	,	0.0140000000	},
        {	534	,	0.0144000000	},
        {	535	,	0.0147000000	},
        {	536	,	0.0147000000	},
        {	537	,	0.0146000000	},
        {	538	,	0.0145000000	},
        {	539	,	0.0145000000	},
        {	540	,	0.0144000000	},
        {	541	,	0.0116000000	},
        {	542	,	0.0087500000	},
        {	543	,	0.0059500000	},
        {	544	,	0.0031800000	},
        {	545	,	0.0004260000	},
        {	546	,	0.0026500000	},
        {	547	,	0.0048600000	},
        {	548	,	0.0070600000	},
        {	549	,	0.0092600000	},
        {	550	,	0.0114000000	},
        {	551	,	0.0114000000	},
        {	552	,	0.0114000000	},
        {	553	,	0.0113000000	},
        {	554	,	0.0113000000	},
        {	555	,	0.0113000000	},
        {	556	,	0.0109000000	},
        {	557	,	0.0105000000	},
        {	558	,	0.0101000000	},
        {	559	,	0.0097600000	},
        {	560	,	0.0093800000	},
        {	561	,	0.0103000000	},
        {	562	,	0.0112000000	},
        {	563	,	0.0120000000	},
        {	564	,	0.0129000000	},
        {	565	,	0.0138000000	},
        {	566	,	0.0134000000	},
        {	567	,	0.0131000000	},
        {	568	,	0.0127000000	},
        {	569	,	0.0123000000	},
        {	570	,	0.0119000000	},
        {	571	,	0.0110000000	},
        {	572	,	0.0100000000	},
        {	573	,	0.0090700000	},
        {	574	,	0.0081200000	},
        {	575	,	0.0071600000	},
        {	576	,	0.0077200000	},
        {	577	,	0.0082800000	},
        {	578	,	0.0088400000	},
        {	579	,	0.0094100000	},
        {	580	,	0.0099800000	},
        {	581	,	0.0104000000	},
        {	582	,	0.0109000000	},
        {	583	,	0.0114000000	},
        {	584	,	0.0118000000	},
        {	585	,	0.0123000000	},
        {	586	,	0.0120000000	},
        {	587	,	0.0117000000	},
        {	588	,	0.0113000000	},
        {	589	,	0.0110000000	},
        {	590	,	0.0107000000	},
        {	591	,	0.0102000000	},
        {	592	,	0.0096800000	},
        {	593	,	0.0091800000	},
        {	594	,	0.0086700000	},
        {	595	,	0.0081700000	},
        {	596	,	0.0081000000	},
        {	597	,	0.0080400000	},
        {	598	,	0.0079700000	},
        {	599	,	0.0079100000	},
        {	600	,	0.0078500000	},
        {	601	,	0.0078300000	},
        {	602	,	0.0070900000	},
        {	603	,	0.0069500000	},
        {	604	,	0.0068200000	},
        {	605	,	0.0067100000	},
        {	606	,	0.0086800000	},
        {	607	,	0.0106000000	},
        {	608	,	0.0126000000	},
        {	609	,	0.0145000000	},
        {	610	,	0.0164000000	},
        {	611	,	0.0164000000	},
        {	612	,	0.0163000000	},
        {	613	,	0.0163000000	},
        {	614	,	0.0163000000	},
        {	615	,	0.0162000000	},
        {	616	,	0.0141000000	},
        {	617	,	0.0119000000	},
        {	618	,	0.0098100000	},
        {	619	,	0.0076800000	},
        {	620	,	0.0055500000	},
        {	621	,	0.0058400000	},
        {	622	,	0.0061300000	},
        {	623	,	0.0064300000	},
        {	624	,	0.0067300000	},
        {	625	,	0.0070300000	},
        {	626	,	0.0106000000	},
        {	627	,	0.0142000000	},
        {	628	,	0.0178000000	},
        {	629	,	0.0214000000	},
        {	630	,	0.0251000000	},
        {	631	,	0.0252000000	},
        {	632	,	0.0253000000	},
        {	633	,	0.0255000000	},
        {	634	,	0.0258000000	},
        {	635	,	0.0260000000	},
        {	636	,	0.0233000000	},
        {	637	,	0.0205000000	},
        {	638	,	0.0176000000	},
        {	639	,	0.0147000000	},
        {	640	,	0.0117000000	},
        {	641	,	0.0133000000	},
        {	642	,	0.0149000000	},
        {	643	,	0.0165000000	},
        {	644	,	0.0181000000	},
        {	645	,	0.0196000000	},
        {	646	,	0.0223000000	},
        {	647	,	0.0250000000	},
        {	648	,	0.0277000000	},
        {	649	,	0.0303000000	},
        {	650	,	0.0329000000	},
        {	651	,	0.0327000000	},
        {	652	,	0.0325000000	},
        {	653	,	0.0324000000	},
        {	654	,	0.0322000000	},
        {	655	,	0.0321000000	},
        {	656	,	0.0282000000	},
        {	657	,	0.0244000000	},
        {	658	,	0.0206000000	},
        {	659	,	0.0168000000	},
        {	660	,	0.0130000000	},
        {	661	,	0.0130000000	},
        {	662	,	0.0130000000	},
        {	663	,	0.0129000000	},
        {	664	,	0.0129000000	},
        {	665	,	0.0129000000	},
        {	666	,	0.0152000000	},
        {	667	,	0.0174000000	},
        {	668	,	0.0197000000	},
        {	669	,	0.0220000000	},
        {	670	,	0.0243000000	},
        {	671	,	0.0205000000	},
        {	672	,	0.0168000000	},
        {	673	,	0.0131000000	},
        {	674	,	0.0094000000	},
        {	675	,	0.0057000000	},
        {	676	,	0.0057100000	},
        {	677	,	0.0057200000	},
        {	678	,	0.0057200000	},
        {	679	,	0.0057300000	},
        {	680	,	0.0057400000	},
        {	681	,	0.0052500000	},
        {	682	,	0.0047600000	},
        {	683	,	0.0042800000	},
        {	684	,	0.0037900000	},
        {	685	,	0.0033000000	},
        {	686	,	0.0035900000	},
        {	687	,	0.0038700000	},
        {	688	,	0.0041600000	},
        {	689	,	0.0044500000	},
        {	690	,	0.0047400000	},
        {	691	,	0.0068600000	},
        {	692	,	0.0089900000	},
        {	693	,	0.0111000000	},
        {	694	,	0.0133000000	},
        {	695	,	0.0154000000	},
        {	696	,	0.0127000000	},
        {	697	,	0.0099000000	},
        {	698	,	0.0071100000	},
        {	699	,	0.0043100000	},
        {	700	,	0.0014900000	},
        {	701	,	0.0015000000	},
        {	702	,	0.0015000000	},
        {	703	,	0.0015100000	},
        {	704	,	0.0015100000	},
        {	705	,	0.0015200000	},
        {	706	,	0.0013900000	},
        {	707	,	0.0012700000	},
        {	708	,	0.0011400000	},
        {	709	,	0.0010100000	},
        {	710	,	0.0008830000	},
        {	711	,	0.0008750000	},
        {	712	,	0.0008660000	},
        {	713	,	0.0008580000	},
        {	714	,	0.0008500000	},
        {	715	,	0.0008420000	},
        {	716	,	0.0008490000	},
        {	717	,	0.0008560000	},
        {	718	,	0.0008630000	},
        {	719	,	0.0008700000	},
        {	720	,	0.0008770000	},
        {	721	,	0.0009190000	},
        {	722	,	0.0009610000	},
        {	723	,	0.0010000000	},
        {	724	,	0.0010500000	},
        {	725	,	0.0010900000	},
        {	726	,	0.0012300000	},
        {	727	,	0.0013700000	},
        {	728	,	0.0015200000	},
        {	729	,	0.0016600000	},
        {	730	,	0.0018100000	},
        {	731	,	0.0018200000	},
        {	732	,	0.0018300000	},
        {	733	,	0.0018400000	},
        {	734	,	0.0018500000	},
        {	735	,	0.0018600000	},
        {	736	,	0.0019900000	},
        {	737	,	0.0021200000	},
        {	738	,	0.0022600000	},
        {	739	,	0.0024000000	},
        {	740	,	0.0025400000	},
        {	741	,	0.0020800000	},
        {	742	,	0.0016000000	},
        {	743	,	0.0011300000	},
        {	744	,	0.0006440000	},
        {	745	,	0.0001540000	},
        {	746	,	0.0009200000	},
        {	747	,	0.0017000000	},
        {	748	,	0.0024800000	},
        {	749	,	0.0032800000	},
        {	750	,	0.0040900000	},
        {	751	,	0.0043300000	},
        {	752	,	0.0045700000	},
        {	753	,	0.0048100000	},
        {	754	,	0.0050600000	},
        {	755	,	0.0053100000	},
        {	756	,	0.0076900000	},
        {	757	,	0.0101000000	},
        {	758	,	0.0126000000	},
        {	759	,	0.0151000000	},
        {	760	,	0.0176000000	},
        {	761	,	0.0275000000	},
        {	762	,	0.0376000000	},
        {	763	,	0.0478000000	},
        {	764	,	0.0582000000	},
        {	765	,	0.0686000000	},
        {	766	,	0.1130000000	},
        {	767	,	0.1570000000	},
        {	768	,	0.2030000000	},
        {	769	,	0.2490000000	},
        {	770	,	0.2960000000	},
        {	771	,	0.3720000000	},
        {	772	,	0.4500000000	},
        {	773	,	0.5290000000	},
        {	774	,	0.6090000000	},
        {	775	,	0.6910000000	},
        {	776	,	0.7140000000	},
        {	777	,	0.7380000000	},
        {	778	,	0.7620000000	},
        {	779	,	0.7860000000	},
        {	780	,	0.8110000000	},
        {	781	,	0.8180000000	},
        {	782	,	0.8260000000	},
        {	783	,	0.8330000000	},
        {	784	,	0.8400000000	},
        {	785	,	0.8470000000	},
        {	786	,	0.8540000000	},
        {	787	,	0.8610000000	},
        {	788	,	0.8680000000	},
        {	789	,	0.8760000000	},
        {	790	,	0.8830000000	},
        {	791	,	0.8890000000	},
        {	792	,	0.8940000000	},
        {	793	,	0.9000000000	},
        {	794	,	0.9060000000	},
        {	795	,	0.9110000000	},
        {	796	,	0.9170000000	},
        {	797	,	0.9230000000	},
        {	798	,	0.9290000000	},
        {	799	,	0.9350000000	},
        {	800	,	0.9410000000	},
        {	801	,	0.9440000000	},
        {	802	,	0.9460000000	},
        {	803	,	0.9480000000	},
        {	804	,	0.9500000000	},
        {	805	,	0.9520000000	},
        {	806	,	0.9570000000	},
        {	807	,	0.9620000000	},
        {	808	,	0.9670000000	},
        {	809	,	0.9720000000	},
        {	810	,	0.9770000000	},
        {	811	,	0.9780000000	},
        {	812	,	0.9790000000	},
        {	813	,	0.9790000000	},
        {	814	,	0.9800000000	},
        {	815	,	0.9800000000	},
        {	816	,	0.9840000000	},
        {	817	,	0.9890000000	},
        {	818	,	0.9930000000	},
        {	819	,	0.9970000000	},
        {	820	,	1.0000000000	},
        {	821	,	1.0000000000	},
        {	822	,	1.0000000000	},
        {	823	,	0.9990000000	},
        {	824	,	0.9980000000	},
        {	825	,	0.9970000000	},
        {	826	,	0.9870000000	},
        {	827	,	0.9770000000	},
        {	828	,	0.9670000000	},
        {	829	,	0.9560000000	},
        {	830	,	0.9450000000	},
        {	831	,	0.9430000000	},
        {	832	,	0.9410000000	},
        {	833	,	0.9390000000	},
        {	834	,	0.9360000000	},
        {	835	,	0.9330000000	},
        {	836	,	0.9310000000	},
        {	837	,	0.9290000000	},
        {	838	,	0.9270000000	},
        {	839	,	0.9240000000	},
        {	840	,	0.9220000000	},
        {	841	,	0.9220000000	},
        {	842	,	0.9230000000	},
        {	843	,	0.9230000000	},
        {	844	,	0.9220000000	},
        {	845	,	0.9220000000	},
        {	846	,	0.9180000000	},
        {	847	,	0.9130000000	},
        {	848	,	0.9070000000	},
        {	849	,	0.9020000000	},
        {	850	,	0.8960000000	},
        {	851	,	0.9060000000	},
        {	852	,	0.9160000000	},
        {	853	,	0.9250000000	},
        {	854	,	0.9330000000	},
        {	855	,	0.9410000000	},
        {	856	,	0.9410000000	},
        {	857	,	0.9400000000	},
        {	858	,	0.9380000000	},
        {	859	,	0.9370000000	},
        {	860	,	0.9350000000	},
        {	861	,	0.9260000000	},
        {	862	,	0.9170000000	},
        {	863	,	0.9090000000	},
        {	864	,	0.9000000000	},
        {	865	,	0.8910000000	},
        {	866	,	0.8820000000	},
        {	867	,	0.8720000000	},
        {	868	,	0.8630000000	},
        {	869	,	0.8540000000	},
        {	870	,	0.8440000000	},
        {	871	,	0.8440000000	},
        {	872	,	0.8440000000	},
        {	873	,	0.8430000000	},
        {	874	,	0.8430000000	},
        {	875	,	0.8420000000	},
        {	876	,	0.8380000000	},
        {	877	,	0.8340000000	},
        {	878	,	0.8300000000	},
        {	879	,	0.8260000000	},
        {	880	,	0.8220000000	},
        {	881	,	0.8190000000	},
        {	882	,	0.8150000000	},
        {	883	,	0.8110000000	},
        {	884	,	0.8070000000	},
        {	885	,	0.8030000000	},
        {	886	,	0.7350000000	},
        {	887	,	0.6690000000	},
        {	888	,	0.6040000000	},
        {	889	,	0.5400000000	},
        {	890	,	0.4780000000	},
        {	891	,	0.4070000000	},
        {	892	,	0.3380000000	},
        {	893	,	0.2700000000	},
        {	894	,	0.2050000000	},
        {	895	,	0.1400000000	},
        {	896	,	0.1130000000	},
        {	897	,	0.0866000000	},
        {	898	,	0.0607000000	},
        {	899	,	0.0353000000	},
        {	900	,	0.0104000000	},
        {	901	,	0.0089600000	},
        {	902	,	0.0075300000	},
        {	903	,	0.0061300000	},
        {	904	,	0.0047600000	},
        {	905	,	0.0034100000	},
        {	906	,	0.0033800000	},
        {	907	,	0.0033500000	},
        {	908	,	0.0033200000	},
        {	909	,	0.0032800000	},
        {	910	,	0.0032500000	},
        {	911	,	0.0028200000	},
        {	912	,	0.0023900000	},
        {	913	,	0.0019700000	},
        {	914	,	0.0015600000	},
        {	915	,	0.0011600000	},
        {	916	,	0.0010400000	},
        {	917	,	0.0009170000	},
        {	918	,	0.0008000000	},
        {	919	,	0.0006860000	},
        {	920	,	0.0005740000	},
        {	921	,	0.0004740000	},
        {	922	,	0.0003760000	},
        {	923	,	0.0002800000	},
        {	924	,	0.0001850000	},
        {	925	,	0.0000922000	},
        {	926	,	0.0000838000	},
        {	927	,	0.0000754000	},
        {	928	,	0.0000671000	},
        {	929	,	0.0000590000	},
        {	930	,	0.0000510000	},
        {	931	,	0.0000515000	},
        {	932	,	0.0000521000	},
        {	933	,	0.0000526000	},
        {	934	,	0.0000531000	},
        {	935	,	0.0000536000	},
        {	936	,	0.0000516000	},
        {	937	,	0.0000496000	},
        {	938	,	0.0000477000	},
        {	939	,	0.0000457000	},
        {	940	,	0.0000439000	},
        {	941	,	0.0000351000	},
        {	942	,	0.0000265000	},
        {	943	,	0.0000181000	},
        {	944	,	0.0000096700	},
        {	945	,	0.0000013700	},
        {	946	,	0.0000117000	},
        {	947	,	0.0000219000	},
        {	948	,	0.0000319000	},
        {	949	,	0.0000419000	},
        {	950	,	0.0000517000	},
        {	951	,	0.0000517000	},
        {	952	,	0.0000517000	},
        {	953	,	0.0000517000	},
        {	954	,	0.0000516000	},
        {	955	,	0.0000516000	},
        {	956	,	0.0000533000	},
        {	957	,	0.0000548000	},
        {	958	,	0.0000564000	},
        {	959	,	0.0000580000	},
        {	960	,	0.0000596000	},
        {	961	,	0.0000586000	},
        {	962	,	0.0000577000	},
        {	963	,	0.0000569000	},
        {	964	,	0.0000560000	},
        {	965	,	0.0000551000	},
        {	966	,	0.0000462000	},
        {	967	,	0.0000373000	},
        {	968	,	0.0000285000	},
        {	969	,	0.0000199000	},
        {	970	,	0.0000112000	},
        {	971	,	0.0000125000	},
        {	972	,	0.0000138000	},
        {	973	,	0.0000151000	},
        {	974	,	0.0000164000	},
        {	975	,	0.0000177000	},
        {	976	,	0.0000219000	},
        {	977	,	0.0000261000	},
        {	978	,	0.0000302000	},
        {	979	,	0.0000343000	},
        {	980	,	0.0000384000	},
        {	981	,	0.0000383000	},
        {	982	,	0.0000383000	},
        {	983	,	0.0000382000	},
        {	984	,	0.0000382000	},
        {	985	,	0.0000382000	},
        {	986	,	0.0000462000	},
        {	987	,	0.0000541000	},
        {	988	,	0.0000620000	},
        {	989	,	0.0000699000	},
        {	990	,	0.0000777000	},
        {	991	,	0.0000846000	},
        {	992	,	0.0000915000	},
        {	993	,	0.0000984000	},
        {	994	,	0.0001050000	},
        {	995	,	0.0001120000	},
        {	996	,	0.0000895000	},
        {	997	,	0.0000670000	},
        {	998	,	0.0000446000	},
        {	999	,	0.0000223000	},
        {	1000	,	0.0000000000	}
      };
      
      static const double AMAZONIA1_WFI_RO_B1_SRF[][2] = 
      {
        { 420.000000 , 0.005245 },
        { 425.000000 , 0.004232 },
        { 430.000000 , 0.005150 },
        { 435.000000 , 0.004619 },
        { 440.000000 , 0.006628 },
        { 445.000000 , 0.014616 },
        { 450.000000 , 0.072971 },
        { 455.000000 , 0.215572 },
        { 460.000000 , 0.330873 },
        { 465.000000 , 0.391840 },
        { 470.000000 , 0.456304 },
        { 475.000000 , 0.504440 },
        { 480.000000 , 0.541912 },
        { 485.000000 , 0.575420 },
        { 490.000000 , 0.621048 },
        { 495.000000 , 0.725572 },
        { 500.000000 , 0.820970 },
        { 505.000000 , 0.890073 },
        { 510.000000 , 1.000000 },
        { 515.000000 , 0.997193 },
        { 520.000000 , 0.687880 },
        { 525.000000 , 0.147085 },
        { 530.000000 , 0.046065 },
        { 535.000000 , 0.016336 },
        { 540.000000 , 0.010258 },
        { 545.000000 , 0.006614 },
        { 550.000000 , 0.005782 },
        { 555.000000 , 0.005354 },
        { 560.000000 , 0.004729 },
        { 565.000000 , 0.003866 },
        { 570.000000 , 0.004264 },
        { 575.000000 , 0.003950 },
        { 580.000000 , 0.004032 },
        { 585.000000 , 0.004700 },
        { 590.000000 , 0.004283 },
        { 595.000000 , 0.004835 },
        { 600.000000 , 0.003226 },
        { 605.000000 , 0.000000 },
        { 610.000000 , 0.000000 },
        { 615.000000 , 0.000000 },
        { 620.000000 , 0.000000 },
        { 625.000000 , 0.000000 },
        { 630.000000 , 0.000000 },
        { 635.000000 , 0.000000 },
        { 640.000000 , 0.000000 },
        { 645.000000 , 0.000000 },
        { 650.000000 , 0.000000 },
        { 655.000000 , 0.000000 },
        { 660.000000 , 0.000000 },
        { 665.000000 , 0.000000 },
        { 670.000000 , 0.000000 },
        { 675.000000 , 0.000000 },
        { 680.000000 , 0.000000 },
        { 685.000000 , 0.000000 },
        { 690.000000 , 0.000000 },
        { 695.000000 , 0.000000 },
        { 700.000000 , 0.000000 },
        { 705.000000 , 0.000000 },
        { 710.000000 , 0.000000 },
        { 715.000000 , 0.000000 },
        { 720.000000 , 0.000000 },
        { 725.000000 , 0.000000 },
        { 730.000000 , 0.000000 },
        { 735.000000 , 0.000000 },
        { 740.000000 , 0.000000 },
        { 745.000000 , 0.000000 },
        { 750.000000 , 0.000000 },
        { 755.000000 , 0.000000 },
        { 760.000000 , 0.000000 },
        { 765.000000 , 0.000000 },
        { 770.000000 , 0.000000 },
        { 775.000000 , 0.000000 },
        { 780.000000 , 0.000000 },
        { 785.000000 , 0.000000 },
        { 790.000000 , 0.000000 },
        { 795.000000 , 0.000000 },
        { 800.000000 , 0.000000 },
        { 805.000000 , 0.000000 },
        { 810.000000 , 0.000000 },
        { 815.000000 , 0.000000 },
        { 820.000000 , 0.000000 },
        { 825.000000 , 0.000000 },
        { 830.000000 , 0.000000 },
        { 835.000000 , 0.000000 },
        { 840.000000 , 0.000000 },
        { 845.000000 , 0.000000 },
        { 850.000000 , 0.000000 },
        { 855.000000 , 0.000000 },
        { 860.000000 , 0.000000 },
        { 865.000000 , 0.000000 },
        { 870.000000 , 0.000000 },
        { 875.000000 , 0.000000 },
        { 880.000000 , 0.000000 },
        { 885.000000 , 0.000000 },
        { 890.000000 , 0.000000 },
        { 895.000000 , 0.000000 },
        { 900.000000 , 0.000000 },
        { 905.000000 , 0.000000 },
        { 910.000000 , 0.000000 },
        { 915.000000 , 0.000000 },
        { 920.000000 , 0.000000 }
      };        

      static const double AMAZONIA1_WFI_RO_B2_SRF[][2] = 
      {
        { 420.000000 , 0.000000 },
        { 425.000000 , 0.000000 },
        { 430.000000 , 0.000000 },
        { 435.000000 , 0.000000 },
        { 440.000000 , 0.000000 },
        { 445.000000 , 0.000000 },
        { 450.000000 , 0.005011 },
        { 455.000000 , 0.004380 },
        { 460.000000 , 0.005650 },
        { 465.000000 , 0.005098 },
        { 470.000000 , 0.005203 },
        { 475.000000 , 0.005492 },
        { 480.000000 , 0.005377 },
        { 485.000000 , 0.005686 },
        { 490.000000 , 0.006945 },
        { 495.000000 , 0.010684 },
        { 500.000000 , 0.031308 },
        { 505.000000 , 0.034874 },
        { 510.000000 , 0.110801 },
        { 515.000000 , 0.599999 },
        { 520.000000 , 1.000000 },
        { 525.000000 , 0.981791 },
        { 530.000000 , 0.959314 },
        { 535.000000 , 0.946222 },
        { 540.000000 , 0.939856 },
        { 545.000000 , 0.926986 },
        { 550.000000 , 0.921742 },
        { 555.000000 , 0.916062 },
        { 560.000000 , 0.918618 },
        { 565.000000 , 0.919429 },
        { 570.000000 , 0.918989 },
        { 575.000000 , 0.918646 },
        { 580.000000 , 0.918921 },
        { 585.000000 , 0.925055 },
        { 590.000000 , 0.936700 },
        { 595.000000 , 0.947902 },
        { 600.000000 , 0.964414 },
        { 605.000000 , 0.419652 },
        { 610.000000 , 0.112106 },
        { 615.000000 , 0.030544 },
        { 620.000000 , 0.013897 },
        { 625.000000 , 0.008100 },
        { 630.000000 , 0.005466 },
        { 635.000000 , 0.003506 },
        { 640.000000 , 0.001999 },
        { 645.000000 , 0.000000 },
        { 650.000000 , 0.000000 },
        { 655.000000 , 0.000000 },
        { 660.000000 , 0.000000 },
        { 665.000000 , 0.000000 },
        { 670.000000 , 0.000000 },
        { 675.000000 , 0.000000 },
        { 680.000000 , 0.000000 },
        { 685.000000 , 0.000000 },
        { 690.000000 , 0.000000 },
        { 695.000000 , 0.000000 },
        { 700.000000 , 0.000000 },
        { 705.000000 , 0.000000 },
        { 710.000000 , 0.000000 },
        { 715.000000 , 0.000000 },
        { 720.000000 , 0.000000 },
        { 725.000000 , 0.000000 },
        { 730.000000 , 0.000000 },
        { 735.000000 , 0.000000 },
        { 740.000000 , 0.000000 },
        { 745.000000 , 0.000000 },
        { 750.000000 , 0.000000 },
        { 755.000000 , 0.000000 },
        { 760.000000 , 0.000000 },
        { 765.000000 , 0.000000 },
        { 770.000000 , 0.000000 },
        { 775.000000 , 0.000000 },
        { 780.000000 , 0.000000 },
        { 785.000000 , 0.000000 },
        { 790.000000 , 0.000000 },
        { 795.000000 , 0.000000 },
        { 800.000000 , 0.000000 },
        { 805.000000 , 0.000000 },
        { 810.000000 , 0.000000 },
        { 815.000000 , 0.000000 },
        { 820.000000 , 0.000000 },
        { 825.000000 , 0.000000 },
        { 830.000000 , 0.000000 },
        { 835.000000 , 0.000000 },
        { 840.000000 , 0.000000 },
        { 845.000000 , 0.000000 },
        { 850.000000 , 0.000000 },
        { 855.000000 , 0.000000 },
        { 860.000000 , 0.000000 },
        { 865.000000 , 0.000000 },
        { 870.000000 , 0.000000 },
        { 875.000000 , 0.000000 },
        { 880.000000 , 0.000000 },
        { 885.000000 , 0.000000 },
        { 890.000000 , 0.000000 },
        { 895.000000 , 0.000000 },
        { 900.000000 , 0.000000 },
        { 905.000000 , 0.000000 },
        { 910.000000 , 0.000000 },
        { 915.000000 , 0.000000 },
        { 920.000000 , 0.000000 }
      };        

      static const double AMAZONIA1_WFI_RO_B3_SRF[][2] = 
      {
        { 420.000000 , 0.000000 },
        { 425.000000 , 0.000000 },
        { 430.000000 , 0.000000 },
        { 435.000000 , 0.000000 },
        { 440.000000 , 0.000000 },
        { 445.000000 , 0.000000 },
        { 450.000000 , 0.000000 },
        { 455.000000 , 0.000000 },
        { 460.000000 , 0.000000 },
        { 465.000000 , 0.000000 },
        { 470.000000 , 0.000000 },
        { 475.000000 , 0.000000 },
        { 480.000000 , 0.000000 },
        { 485.000000 , 0.000000 },
        { 490.000000 , 0.000000 },
        { 495.000000 , 0.000000 },
        { 500.000000 , 0.000000 },
        { 505.000000 , 0.000000 },
        { 510.000000 , 0.000000 },
        { 515.000000 , 0.000000 },
        { 520.000000 , 0.000000 },
        { 525.000000 , 0.000000 },
        { 530.000000 , 0.000000 },
        { 535.000000 , 0.000000 },
        { 540.000000 , 0.000000 },
        { 545.000000 , 0.000000 },
        { 550.000000 , 0.000000 },
        { 555.000000 , 0.000000 },
        { 560.000000 , 0.000000 },
        { 565.000000 , 0.000000 },
        { 570.000000 , 0.000000 },
        { 575.000000 , 0.000000 },
        { 580.000000 , 0.005534 },
        { 585.000000 , 0.006031 },
        { 590.000000 , 0.005592 },
        { 595.000000 , 0.004937 },
        { 600.000000 , 0.000088 },
        { 605.000000 , 0.001147 },
        { 610.000000 , 0.014641 },
        { 615.000000 , 0.052963 },
        { 620.000000 , 0.397154 },
        { 625.000000 , 0.766724 },
        { 630.000000 , 0.917151 },
        { 635.000000 , 0.944761 },
        { 640.000000 , 0.948591 },
        { 645.000000 , 0.957919 },
        { 650.000000 , 1.000000 },
        { 655.000000 , 0.975593 },
        { 660.000000 , 0.907217 },
        { 665.000000 , 0.879429 },
        { 670.000000 , 0.873367 },
        { 675.000000 , 0.853074 },
        { 680.000000 , 0.530974 },
        { 685.000000 , 0.198709 },
        { 690.000000 , 0.034638 },
        { 695.000000 , 0.012552 },
        { 700.000000 , 0.001593 },
        { 705.000000 , 0.001438 },
        { 710.000000 , 0.000521 },
        { 715.000000 , 0.000000 },
        { 720.000000 , 0.000000 },
        { 725.000000 , 0.000000 },
        { 730.000000 , 0.000000 },
        { 735.000000 , 0.000000 },
        { 740.000000 , 0.000000 },
        { 745.000000 , 0.000000 },
        { 750.000000 , 0.000000 },
        { 755.000000 , 0.000000 },
        { 760.000000 , 0.000000 },
        { 765.000000 , 0.000000 },
        { 770.000000 , 0.000000 },
        { 775.000000 , 0.000000 },
        { 780.000000 , 0.000000 },
        { 785.000000 , 0.000000 },
        { 790.000000 , 0.000000 },
        { 795.000000 , 0.000000 },
        { 800.000000 , 0.000000 },
        { 805.000000 , 0.000000 },
        { 810.000000 , 0.000000 },
        { 815.000000 , 0.000000 },
        { 820.000000 , 0.000000 },
        { 825.000000 , 0.000000 },
        { 830.000000 , 0.000000 },
        { 835.000000 , 0.000000 },
        { 840.000000 , 0.000000 },
        { 845.000000 , 0.000000 },
        { 850.000000 , 0.000000 },
        { 855.000000 , 0.000000 },
        { 860.000000 , 0.000000 },
        { 865.000000 , 0.000000 },
        { 870.000000 , 0.000000 },
        { 875.000000 , 0.000000 },
        { 880.000000 , 0.000000 },
        { 885.000000 , 0.000000 },
        { 890.000000 , 0.000000 },
        { 895.000000 , 0.000000 },
        { 900.000000 , 0.000000 },
        { 905.000000 , 0.000000 },
        { 910.000000 , 0.000000 },
        { 915.000000 , 0.000000 },
        { 920.000000 , 0.000000 }
      };        

      static const double AMAZONIA1_WFI_RO_B4_SRF[][2] = 
      {
        { 420.000000 , 0.000000 },
        { 425.000000 , 0.000000 },
        { 430.000000 , 0.000000 },
        { 435.000000 , 0.000000 },
        { 440.000000 , 0.000000 },
        { 445.000000 , 0.000000 },
        { 450.000000 , 0.000000 },
        { 455.000000 , 0.000000 },
        { 460.000000 , 0.000000 },
        { 465.000000 , 0.000000 },
        { 470.000000 , 0.000000 },
        { 475.000000 , 0.000000 },
        { 480.000000 , 0.000000 },
        { 485.000000 , 0.000000 },
        { 490.000000 , 0.000000 },
        { 495.000000 , 0.000000 },
        { 500.000000 , 0.000000 },
        { 505.000000 , 0.000000 },
        { 510.000000 , 0.000000 },
        { 515.000000 , 0.000000 },
        { 520.000000 , 0.000000 },
        { 525.000000 , 0.000000 },
        { 530.000000 , 0.000000 },
        { 535.000000 , 0.000000 },
        { 540.000000 , 0.000000 },
        { 545.000000 , 0.000000 },
        { 550.000000 , 0.000000 },
        { 555.000000 , 0.000000 },
        { 560.000000 , 0.000000 },
        { 565.000000 , 0.000000 },
        { 570.000000 , 0.000000 },
        { 575.000000 , 0.000000 },
        { 580.000000 , 0.000000 },
        { 585.000000 , 0.000000 },
        { 590.000000 , 0.000000 },
        { 595.000000 , 0.000000 },
        { 600.000000 , 0.000000 },
        { 605.000000 , 0.000000 },
        { 610.000000 , 0.000000 },
        { 615.000000 , 0.000000 },
        { 620.000000 , 0.000000 },
        { 625.000000 , 0.000000 },
        { 630.000000 , 0.000000 },
        { 635.000000 , 0.000000 },
        { 640.000000 , 0.000000 },
        { 645.000000 , 0.000000 },
        { 650.000000 , 0.000000 },
        { 655.000000 , 0.000000 },
        { 660.000000 , 0.000000 },
        { 665.000000 , 0.000000 },
        { 670.000000 , 0.000000 },
        { 675.000000 , 0.000000 },
        { 680.000000 , 0.000000 },
        { 685.000000 , 0.000000 },
        { 690.000000 , 0.000000 },
        { 695.000000 , 0.000000 },
        { 700.000000 , 0.004458 },
        { 705.000000 , 0.005212 },
        { 710.000000 , 0.005294 },
        { 715.000000 , 0.005838 },
        { 720.000000 , 0.005805 },
        { 725.000000 , 0.006247 },
        { 730.000000 , 0.007434 },
        { 735.000000 , 0.009071 },
        { 740.000000 , 0.014273 },
        { 745.000000 , 0.023602 },
        { 750.000000 , 0.037796 },
        { 755.000000 , 0.058161 },
        { 760.000000 , 0.100610 },
        { 765.000000 , 0.222729 },
        { 770.000000 , 0.607334 },
        { 775.000000 , 1.000000 },
        { 780.000000 , 0.939732 },
        { 785.000000 , 0.828428 },
        { 790.000000 , 0.778250 },
        { 795.000000 , 0.767656 },
        { 800.000000 , 0.760926 },
        { 805.000000 , 0.754079 },
        { 810.000000 , 0.757492 },
        { 815.000000 , 0.756476 },
        { 820.000000 , 0.739288 },
        { 825.000000 , 0.737127 },
        { 830.000000 , 0.723470 },
        { 835.000000 , 0.719693 },
        { 840.000000 , 0.710535 },
        { 845.000000 , 0.687836 },
        { 850.000000 , 0.671049 },
        { 855.000000 , 0.615753 },
        { 860.000000 , 0.582881 },
        { 865.000000 , 0.527103 },
        { 870.000000 , 0.468958 },
        { 875.000000 , 0.417982 },
        { 880.000000 , 0.381754 },
        { 885.000000 , 0.322474 },
        { 890.000000 , 0.210976 },
        { 895.000000 , 0.105473 },
        { 900.000000 , 0.047633 },
        { 905.000000 , 0.023805 },
        { 910.000000 , 0.001395 },
        { 915.000000 , 0.000000 },
        { 920.000000 , 0.000000 }
      };        

      static const double AMAZONIA1_WFI_LO_B1_SRF[][2] = 
      {
        { 420.000000 , 0.000000 },
        { 425.000000 , 0.000000 },
        { 430.000000 , 0.000000 },
        { 435.000000 , 0.000000 },
        { 440.000000 , 0.000343 },
        { 445.000000 , 0.027120 },
        { 450.000000 , 0.159359 },
        { 455.000000 , 0.734836 },
        { 460.000000 , 1.000000 },
        { 465.000000 , 0.967427 },
        { 470.000000 , 0.938777 },
        { 475.000000 , 0.916297 },
        { 480.000000 , 0.898144 },
        { 485.000000 , 0.882801 },
        { 490.000000 , 0.868241 },
        { 495.000000 , 0.861915 },
        { 500.000000 , 0.847617 },
        { 505.000000 , 0.826272 },
        { 510.000000 , 0.820392 },
        { 515.000000 , 0.805033 },
        { 520.000000 , 0.776328 },
        { 525.000000 , 0.195792 },
        { 530.000000 , 0.052785 },
        { 535.000000 , 0.013238 },
        { 540.000000 , 0.003898 },
        { 545.000000 , 0.000437 },
        { 550.000000 , 0.000000 },
        { 555.000000 , 0.000000 },
        { 560.000000 , 0.000000 },
        { 565.000000 , 0.000000 },
        { 570.000000 , 0.000000 },
        { 575.000000 , 0.000000 },
        { 580.000000 , 0.000000 },
        { 585.000000 , 0.000000 },
        { 590.000000 , 0.000000 },
        { 595.000000 , 0.000000 },
        { 600.000000 , 0.000000 },
        { 605.000000 , 0.000000 },
        { 610.000000 , 0.000000 },
        { 615.000000 , 0.000000 },
        { 620.000000 , 0.000000 },
        { 625.000000 , 0.000000 },
        { 630.000000 , 0.000000 },
        { 635.000000 , 0.000000 },
        { 640.000000 , 0.000000 },
        { 645.000000 , 0.000000 },
        { 650.000000 , 0.000000 },
        { 655.000000 , 0.000000 },
        { 660.000000 , 0.000000 },
        { 665.000000 , 0.000000 },
        { 670.000000 , 0.000000 },
        { 675.000000 , 0.000000 },
        { 680.000000 , 0.000000 },
        { 685.000000 , 0.000000 },
        { 690.000000 , 0.000000 },
        { 695.000000 , 0.000000 },
        { 700.000000 , 0.000000 },
        { 705.000000 , 0.000000 },
        { 710.000000 , 0.000000 },
        { 715.000000 , 0.000000 },
        { 720.000000 , 0.000000 },
        { 725.000000 , 0.000000 },
        { 730.000000 , 0.000000 },
        { 735.000000 , 0.000000 },
        { 740.000000 , 0.000000 },
        { 745.000000 , 0.000000 },
        { 750.000000 , 0.000000 },
        { 755.000000 , 0.000000 },
        { 760.000000 , 0.000000 },
        { 765.000000 , 0.000000 },
        { 770.000000 , 0.000000 },
        { 775.000000 , 0.000000 },
        { 780.000000 , 0.000000 },
        { 785.000000 , 0.000000 },
        { 790.000000 , 0.000000 },
        { 795.000000 , 0.000000 },
        { 800.000000 , 0.000000 },
        { 805.000000 , 0.000000 },
        { 810.000000 , 0.000000 },
        { 815.000000 , 0.000000 },
        { 820.000000 , 0.000000 },
        { 825.000000 , 0.000000 },
        { 830.000000 , 0.000000 },
        { 835.000000 , 0.000000 },
        { 840.000000 , 0.000000 },
        { 845.000000 , 0.000000 },
        { 850.000000 , 0.000000 },
        { 855.000000 , 0.000000 },
        { 860.000000 , 0.000000 },
        { 865.000000 , 0.000000 },
        { 870.000000 , 0.000000 },
        { 875.000000 , 0.000000 },
        { 880.000000 , 0.000000 },
        { 885.000000 , 0.000000 },
        { 890.000000 , 0.000000 },
        { 895.000000 , 0.000000 },
        { 900.000000 , 0.000000 },
        { 905.000000 , 0.000000 },
        { 910.000000 , 0.000000 },
        { 915.000000 , 0.000000 },
        { 920.000000 , 0.000000 }
      };        

      static const double AMAZONIA1_WFI_LO_B2_SRF[][2] = 
      {
        { 420.000000 , 0.000000 },
        { 425.000000 , 0.000000 },
        { 430.000000 , 0.000000 },
        { 435.000000 , 0.000000 },
        { 440.000000 , 0.000000 },
        { 445.000000 , 0.000000 },
        { 450.000000 , 0.000000 },
        { 455.000000 , 0.000000 },
        { 460.000000 , 0.000000 },
        { 465.000000 , 0.000000 },
        { 470.000000 , 0.000000 },
        { 475.000000 , 0.000000 },
        { 480.000000 , 0.000000 },
        { 485.000000 , 0.000000 },
        { 490.000000 , 0.000000 },
        { 495.000000 , 0.000000 },
        { 500.000000 , 0.002787 },
        { 505.000000 , 0.003877 },
        { 510.000000 , 0.050973 },
        { 515.000000 , 0.257939 },
        { 520.000000 , 0.775467 },
        { 525.000000 , 0.831703 },
        { 530.000000 , 0.862457 },
        { 535.000000 , 0.895887 },
        { 540.000000 , 0.868734 },
        { 545.000000 , 0.886492 },
        { 550.000000 , 0.893894 },
        { 555.000000 , 0.958189 },
        { 560.000000 , 1.000000 },
        { 565.000000 , 0.956081 },
        { 570.000000 , 0.959031 },
        { 575.000000 , 0.896924 },
        { 580.000000 , 0.886359 },
        { 585.000000 , 0.852023 },
        { 590.000000 , 0.802688 },
        { 595.000000 , 0.655466 },
        { 600.000000 , 0.373130 },
        { 605.000000 , 0.121867 },
        { 610.000000 , 0.023353 },
        { 615.000000 , 0.001012 },
        { 620.000000 , 0.000000 },
        { 625.000000 , 0.000000 },
        { 630.000000 , 0.000000 },
        { 635.000000 , 0.000000 },
        { 640.000000 , 0.000000 },
        { 645.000000 , 0.000000 },
        { 650.000000 , 0.000000 },
        { 655.000000 , 0.000000 },
        { 660.000000 , 0.000000 },
        { 665.000000 , 0.000000 },
        { 670.000000 , 0.000000 },
        { 675.000000 , 0.000000 },
        { 680.000000 , 0.000000 },
        { 685.000000 , 0.000000 },
        { 690.000000 , 0.000000 },
        { 695.000000 , 0.000000 },
        { 700.000000 , 0.000000 },
        { 705.000000 , 0.000000 },
        { 710.000000 , 0.000000 },
        { 715.000000 , 0.000000 },
        { 720.000000 , 0.000000 },
        { 725.000000 , 0.000000 },
        { 730.000000 , 0.000000 },
        { 735.000000 , 0.000000 },
        { 740.000000 , 0.000000 },
        { 745.000000 , 0.000000 },
        { 750.000000 , 0.000000 },
        { 755.000000 , 0.000000 },
        { 760.000000 , 0.000000 },
        { 765.000000 , 0.000000 },
        { 770.000000 , 0.000000 },
        { 775.000000 , 0.000000 },
        { 780.000000 , 0.000000 },
        { 785.000000 , 0.000000 },
        { 790.000000 , 0.000000 },
        { 795.000000 , 0.000000 },
        { 800.000000 , 0.000000 },
        { 805.000000 , 0.000000 },
        { 810.000000 , 0.000000 },
        { 815.000000 , 0.000000 },
        { 820.000000 , 0.000000 },
        { 825.000000 , 0.000000 },
        { 830.000000 , 0.000000 },
        { 835.000000 , 0.000000 },
        { 840.000000 , 0.000000 },
        { 845.000000 , 0.000000 },
        { 850.000000 , 0.000000 },
        { 855.000000 , 0.000000 },
        { 860.000000 , 0.000000 },
        { 865.000000 , 0.000000 },
        { 870.000000 , 0.000000 },
        { 875.000000 , 0.000000 },
        { 880.000000 , 0.000000 },
        { 885.000000 , 0.000000 },
        { 890.000000 , 0.000000 },
        { 895.000000 , 0.000000 },
        { 900.000000 , 0.000000 },
        { 905.000000 , 0.000000 },
        { 910.000000 , 0.000000 },
        { 915.000000 , 0.000000 },
        { 920.000000 , 0.000000 }
      };        

      static const double AMAZONIA1_WFI_LO_B3_SRF[][2] = 
      {
        { 420.000000 , 0.000000 },
        { 425.000000 , 0.000000 },
        { 430.000000 , 0.000000 },
        { 435.000000 , 0.000000 },
        { 440.000000 , 0.000000 },
        { 445.000000 , 0.000000 },
        { 450.000000 , 0.000000 },
        { 455.000000 , 0.000000 },
        { 460.000000 , 0.000000 },
        { 465.000000 , 0.000000 },
        { 470.000000 , 0.000000 },
        { 475.000000 , 0.000000 },
        { 480.000000 , 0.000000 },
        { 485.000000 , 0.000000 },
        { 490.000000 , 0.000000 },
        { 495.000000 , 0.000000 },
        { 500.000000 , 0.000000 },
        { 505.000000 , 0.000000 },
        { 510.000000 , 0.000000 },
        { 515.000000 , 0.000000 },
        { 520.000000 , 0.000000 },
        { 525.000000 , 0.000000 },
        { 530.000000 , 0.000000 },
        { 535.000000 , 0.000000 },
        { 540.000000 , 0.000000 },
        { 545.000000 , 0.000000 },
        { 550.000000 , 0.000000 },
        { 555.000000 , 0.000000 },
        { 560.000000 , 0.000000 },
        { 565.000000 , 0.000000 },
        { 570.000000 , 0.000000 },
        { 575.000000 , 0.000000 },
        { 580.000000 , 0.000000 },
        { 585.000000 , 0.000000 },
        { 590.000000 , 0.000000 },
        { 595.000000 , 0.000000 },
        { 600.000000 , 0.000065 },
        { 605.000000 , 0.007917 },
        { 610.000000 , 0.041560 },
        { 615.000000 , 0.187565 },
        { 620.000000 , 0.975987 },
        { 625.000000 , 0.982465 },
        { 630.000000 , 0.960772 },
        { 635.000000 , 0.950276 },
        { 640.000000 , 0.939569 },
        { 645.000000 , 0.934049 },
        { 650.000000 , 0.935253 },
        { 655.000000 , 0.936887 },
        { 660.000000 , 0.948125 },
        { 665.000000 , 0.961116 },
        { 670.000000 , 0.976604 },
        { 675.000000 , 0.992030 },
        { 680.000000 , 1.000000 },
        { 685.000000 , 0.837685 },
        { 690.000000 , 0.201914 },
        { 695.000000 , 0.056993 },
        { 700.000000 , 0.021484 },
        { 705.000000 , 0.011616 },
        { 710.000000 , 0.006300 },
        { 715.000000 , 0.000522 },
        { 720.000000 , 0.000000 },
        { 725.000000 , 0.000000 },
        { 730.000000 , 0.000000 },
        { 735.000000 , 0.000000 },
        { 740.000000 , 0.000000 },
        { 745.000000 , 0.000000 },
        { 750.000000 , 0.000000 },
        { 755.000000 , 0.000000 },
        { 760.000000 , 0.000000 },
        { 765.000000 , 0.000000 },
        { 770.000000 , 0.000000 },
        { 775.000000 , 0.000000 },
        { 780.000000 , 0.000000 },
        { 785.000000 , 0.000000 },
        { 790.000000 , 0.000000 },
        { 795.000000 , 0.000000 },
        { 800.000000 , 0.000000 },
        { 805.000000 , 0.000000 },
        { 810.000000 , 0.000000 },
        { 815.000000 , 0.000000 },
        { 820.000000 , 0.000000 },
        { 825.000000 , 0.000000 },
        { 830.000000 , 0.000000 },
        { 835.000000 , 0.000000 },
        { 840.000000 , 0.000000 },
        { 845.000000 , 0.000000 },
        { 850.000000 , 0.000000 },
        { 855.000000 , 0.000000 },
        { 860.000000 , 0.000000 },
        { 865.000000 , 0.000000 },
        { 870.000000 , 0.000000 },
        { 875.000000 , 0.000000 },
        { 880.000000 , 0.000000 },
        { 885.000000 , 0.000000 },
        { 890.000000 , 0.000000 },
        { 895.000000 , 0.000000 },
        { 900.000000 , 0.000000 },
        { 905.000000 , 0.000000 },
        { 910.000000 , 0.000000 },
        { 915.000000 , 0.000000 },
        { 920.000000 , 0.000000 }
      };        

      static const double AMAZONIA1_WFI_LO_B4_SRF[][2] = 
      {
        { 420.000000 , 0.000000 },
        { 425.000000 , 0.000000 },
        { 430.000000 , 0.000000 },
        { 435.000000 , 0.000000 },
        { 440.000000 , 0.000000 },
        { 445.000000 , 0.000000 },
        { 450.000000 , 0.000000 },
        { 455.000000 , 0.000000 },
        { 460.000000 , 0.000000 },
        { 465.000000 , 0.000000 },
        { 470.000000 , 0.000000 },
        { 475.000000 , 0.000000 },
        { 480.000000 , 0.000000 },
        { 485.000000 , 0.000000 },
        { 490.000000 , 0.000000 },
        { 495.000000 , 0.000000 },
        { 500.000000 , 0.000000 },
        { 505.000000 , 0.000000 },
        { 510.000000 , 0.000000 },
        { 515.000000 , 0.000000 },
        { 520.000000 , 0.000000 },
        { 525.000000 , 0.000000 },
        { 530.000000 , 0.000000 },
        { 535.000000 , 0.000000 },
        { 540.000000 , 0.000000 },
        { 545.000000 , 0.000000 },
        { 550.000000 , 0.000000 },
        { 555.000000 , 0.000000 },
        { 560.000000 , 0.000000 },
        { 565.000000 , 0.000000 },
        { 570.000000 , 0.000000 },
        { 575.000000 , 0.000000 },
        { 580.000000 , 0.000000 },
        { 585.000000 , 0.000000 },
        { 590.000000 , 0.000000 },
        { 595.000000 , 0.000000 },
        { 600.000000 , 0.000000 },
        { 605.000000 , 0.000000 },
        { 610.000000 , 0.000000 },
        { 615.000000 , 0.000000 },
        { 620.000000 , 0.000000 },
        { 625.000000 , 0.000000 },
        { 630.000000 , 0.000000 },
        { 635.000000 , 0.000000 },
        { 640.000000 , 0.000000 },
        { 645.000000 , 0.000000 },
        { 650.000000 , 0.000000 },
        { 655.000000 , 0.000000 },
        { 660.000000 , 0.000000 },
        { 665.000000 , 0.000000 },
        { 670.000000 , 0.000000 },
        { 675.000000 , 0.000000 },
        { 680.000000 , 0.000000 },
        { 685.000000 , 0.000000 },
        { 690.000000 , 0.000000 },
        { 695.000000 , 0.000000 },
        { 700.000000 , 0.000000 },
        { 705.000000 , 0.000000 },
        { 710.000000 , 0.000000 },
        { 715.000000 , 0.000000 },
        { 720.000000 , 0.000000 },
        { 725.000000 , 0.000000 },
        { 730.000000 , 0.000000 },
        { 735.000000 , 0.000123 },
        { 740.000000 , 0.005265 },
        { 745.000000 , 0.014336 },
        { 750.000000 , 0.030327 },
        { 755.000000 , 0.051471 },
        { 760.000000 , 0.103373 },
        { 765.000000 , 0.282363 },
        { 770.000000 , 0.695441 },
        { 775.000000 , 1.000000 },
        { 780.000000 , 0.932699 },
        { 785.000000 , 0.854035 },
        { 790.000000 , 0.823080 },
        { 795.000000 , 0.827009 },
        { 800.000000 , 0.830941 },
        { 805.000000 , 0.836814 },
        { 810.000000 , 0.853611 },
        { 815.000000 , 0.860114 },
        { 820.000000 , 0.855556 },
        { 825.000000 , 0.867904 },
        { 830.000000 , 0.867007 },
        { 835.000000 , 0.863776 },
        { 840.000000 , 0.854108 },
        { 845.000000 , 0.838051 },
        { 850.000000 , 0.804612 },
        { 855.000000 , 0.770379 },
        { 860.000000 , 0.721606 },
        { 865.000000 , 0.656192 },
        { 870.000000 , 0.589744 },
        { 875.000000 , 0.543369 },
        { 880.000000 , 0.506475 },
        { 885.000000 , 0.315572 },
        { 890.000000 , 0.157839 },
        { 895.000000 , 0.064853 },
        { 900.000000 , 0.022707 },
        { 905.000000 , 0.006739 },
        { 910.000000 , 0.000000 },
        { 915.000000 , 0.000000 },
        { 920.000000 , 0.000000 }
      };
      
      static const double CBERS04A_MUXB5_SRF[][2] = 
      {
        {360,0.000000},
        {361,0.000000},
        {362,0.000000},
        {363,0.000000},
        {364,0.000000},
        {365,0.000000},
        {366,0.000000},
        {367,0.000000},
        {368,0.000000},
        {369,0.000000},
        {370,0.000000},
        {371,0.000000},
        {372,0.000000},
        {373,0.000000},
        {374,0.000000},
        {375,0.000000},
        {376,0.000000},
        {377,0.000000},
        {378,0.000000},
        {379,0.000000},
        {380,0.000000},
        {381,0.000000},
        {382,0.000000},
        {383,0.000000},
        {384,0.000000},
        {385,0.000000},
        {386,0.000000},
        {387,0.000000},
        {388,0.000000},
        {389,0.000000},
        {390,0.000000},
        {391,0.000000},
        {392,0.000000},
        {393,0.000000},
        {394,0.000000},
        {395,0.000000},
        {396,0.000000},
        {397,0.000000},
        {398,0.000000},
        {399,0.000000},
        {400,0.000000},
        {401,0.000000},
        {402,0.000000},
        {403,0.000088},
        {404,0.000177},
        {405,0.000088},
        {406,0.000000},
        {407,0.000000},
        {408,0.000000},
        {409,0.000595},
        {410,0.001190},
        {411,0.000595},
        {412,0.000000},
        {413,0.000072},
        {414,0.000143},
        {415,0.000140},
        {416,0.000136},
        {417,0.000068},
        {418,0.000000},
        {419,0.000132},
        {420,0.000265},
        {421,0.000132},
        {422,0.000000},
        {423,0.000122},
        {424,0.000244},
        {425,0.000239},
        {426,0.000234},
        {427,0.000406},
        {428,0.000578},
        {429,0.000289},
        {430,0.000000},
        {431,0.000324},
        {432,0.000648},
        {433,0.000324},
        {434,0.000000},
        {435,0.000347},
        {436,0.000694},
        {437,0.000685},
        {438,0.000676},
        {439,0.000571},
        {440,0.000466},
        {441,0.002548},
        {442,0.004631},
        {443,0.007909},
        {444,0.011188},
        {445,0.017990},
        {446,0.024792},
        {447,0.035768},
        {448,0.046744},
        {449,0.075321},
        {450,0.103898},
        {451,0.156567},
        {452,0.209236},
        {453,0.270533},
        {454,0.331829},
        {455,0.377406},
        {456,0.422984},
        {457,0.456530},
        {458,0.490076},
        {459,0.523851},
        {460,0.557626},
        {461,0.578408},
        {462,0.599189},
        {463,0.602330},
        {464,0.605470},
        {465,0.606610},
        {466,0.607750},
        {467,0.617594},
        {468,0.627439},
        {469,0.638078},
        {470,0.648717},
        {471,0.657317},
        {472,0.665917},
        {473,0.671048},
        {474,0.676180},
        {475,0.686741},
        {476,0.697303},
        {477,0.703910},
        {478,0.710518},
        {479,0.729951},
        {480,0.749384},
        {481,0.748872},
        {482,0.748361},
        {483,0.759258},
        {484,0.770155},
        {485,0.779815},
        {486,0.789476},
        {487,0.801857},
        {488,0.814237},
        {489,0.826101},
        {490,0.837964},
        {491,0.849017},
        {492,0.860071},
        {493,0.871003},
        {494,0.881934},
        {495,0.897109},
        {496,0.912283},
        {497,0.918861},
        {498,0.925439},
        {499,0.941151},
        {500,0.956862},
        {501,0.969419},
        {502,0.981976},
        {503,0.990988},
        {504,1.0},
        {505,0.994021},
        {506,0.988042},
        {507,0.987170},
        {508,0.986297},
        {509,0.991499},
        {510,0.996701},
        {511,0.983983},
        {512,0.971265},
        {513,0.935965},
        {514,0.900665},
        {515,0.852920},
        {516,0.805175},
        {517,0.776565},
        {518,0.747954},
        {519,0.686618},
        {520,0.625282},
        {521,0.491865},
        {522,0.358447},
        {523,0.269545},
        {524,0.180643},
        {525,0.140602},
        {526,0.100561},
        {527,0.081846},
        {528,0.063132},
        {529,0.052699},
        {530,0.042266},
        {531,0.033774},
        {532,0.025282},
        {533,0.020260},
        {534,0.015238},
        {535,0.012193},
        {536,0.009148},
        {537,0.007792},
        {538,0.006436},
        {539,0.005440},
        {540,0.004443},
        {541,0.003955},
        {542,0.003468},
        {543,0.003032},
        {544,0.002597},
        {545,0.001999},
        {546,0.001401},
        {547,0.000952},
        {548,0.000503},
        {549,0.000408},
        {550,0.000314},
        {551,0.000278},
        {552,0.000242},
        {553,0.000278},
        {554,0.000315},
        {555,0.000296},
        {556,0.000277},
        {557,0.000258},
        {558,0.000239},
        {559,0.000221},
        {560,0.000204},
        {561,0.000169},
        {562,0.000135},
        {563,0.000101},
        {564,0.000067},
        {565,0.000151},
        {566,0.000236},
        {567,0.000219},
        {568,0.000202},
        {569,0.000271},
        {570,0.000340},
        {571,0.000219},
        {572,0.000099},
        {573,0.000133},
        {574,0.000167},
        {575,0.000132},
        {576,0.000098},
        {577,0.000098},
        {578,0.000098},
        {579,0.000131},
        {580,0.000165},
        {581,0.000148},
        {582,0.000131},
        {583,0.000097},
        {584,0.000064},
        {585,0.000064},
        {586,0.000064},
        {587,0.000115},
        {588,0.000165},
        {589,0.000197},
        {590,0.000229},
        {591,0.000196},
        {592,0.000164},
        {593,0.000097},
        {594,0.000031},
        {595,0.000015},
        {596,0.000000},
        {597,0.000000},
        {598,0.000000},
        {599,0.000000},
        {600,0.000000},
        {601,0.000000},
        {602,0.000000},
        {603,0.000015},
        {604,0.000030},
        {605,0.000015},
        {606,0.000000},
        {607,0.000000},
        {608,0.000000},
        {609,0.000065},
        {610,0.000131},
        {611,0.000065},
        {612,0.000000},
        {613,0.000000},
        {614,0.000000},
        {615,0.000000},
        {616,0.000000},
        {617,0.000000},
        {618,0.000000},
        {619,0.000000},
        {620,0.000000},
        {621,0.000000},
        {622,0.000000},
        {623,0.000000},
        {624,0.000000},
        {625,0.000000},
        {626,0.000000},
        {627,0.000136},
        {628,0.000271},
        {629,0.000136},
        {630,0.000000},
        {631,0.000000},
        {632,0.000000},
        {633,0.000053},
        {634,0.000106},
        {635,0.000107},
        {636,0.000107},
        {637,0.000144},
        {638,0.000180},
        {639,0.000126},
        {640,0.000071},
        {641,0.000164},
        {642,0.000258},
        {643,0.000184},
        {644,0.000111},
        {645,0.000148},
        {646,0.000186},
        {647,0.000185},
        {648,0.000185},
        {649,0.000128},
        {650,0.000071},
        {651,0.000053},
        {652,0.000035},
        {653,0.000094},
        {654,0.000153},
        {655,0.000218},
        {656,0.000282},
        {657,0.000181},
        {658,0.000079},
        {659,0.000162},
        {660,0.000245},
        {661,0.000123},
        {662,0.000000},
        {663,0.000076},
        {664,0.000151},
        {665,0.000147},
        {666,0.000143},
        {667,0.000072},
        {668,0.000000},
        {669,0.000116},
        {670,0.000231},
        {671,0.000211},
        {672,0.000190},
        {673,0.000140},
        {674,0.000091},
        {675,0.000178},
        {676,0.000266},
        {677,0.000218},
        {678,0.000170},
        {679,0.000140},
        {680,0.000111},
        {681,0.000208},
        {682,0.000304},
        {683,0.000152},
        {684,0.000000},
        {685,0.000026},
        {686,0.000052},
        {687,0.000053},
        {688,0.000054},
        {689,0.000027},
        {690,0.000000},
        {691,0.000000},
        {692,0.000000},
        {693,0.000000},
        {694,0.000000},
        {695,0.000000},
        {696,0.000000},
        {697,0.000000},
        {698,0.000000},
        {699,0.000000},
        {700,0.000000},
        {701,0.000000},
        {702,0.000000},
        {703,0.000000},
        {704,0.000000},
        {705,0.000057},
        {706,0.000114},
        {707,0.000084},
        {708,0.000055},
        {709,0.000027},
        {710,0.000000},
        {711,0.000000},
        {712,0.000000},
        {713,0.000000},
        {714,0.000000},
        {715,0.000000},
        {716,0.000000},
        {717,0.000000},
        {718,0.000000},
        {719,0.000000},
        {720,0.000000},
        {721,0.000000},
        {722,0.000000},
        {723,0.000000},
        {724,0.000000},
        {725,0.000000},
        {726,0.000000},
        {727,0.000000},
        {728,0.000000},
        {729,0.000000},
        {730,0.000000},
        {731,0.000000},
        {732,0.000000},
        {733,0.000000},
        {734,0.000000},
        {735,0.000000},
        {736,0.000000},
        {737,0.000000},
        {738,0.000000},
        {739,0.000000},
        {740,0.000000},
        {741,0.000085},
        {742,0.000170},
        {743,0.000085},
        {744,0.000000},
        {745,0.000000},
        {746,0.000000},
        {747,0.000000},
        {748,0.000000},
        {749,0.000000},
        {750,0.000000},
        {751,0.000053},
        {752,0.000107},
        {753,0.000053},
        {754,0.000000},
        {755,0.000000},
        {756,0.000000},
        {757,0.000034},
        {758,0.000069},
        {759,0.000051},
        {760,0.000034},
        {761,0.000017},
        {762,0.000000},
        {763,0.000000},
        {764,0.000000},
        {765,0.000000},
        {766,0.000000},
        {767,0.000000},
        {768,0.000000},
        {769,0.000000},
        {770,0.000000},
        {771,0.000000},
        {772,0.000000},
        {773,0.000000},
        {774,0.000000},
        {775,0.000000},
        {776,0.000000},
        {777,0.000000},
        {778,0.000000},
        {779,0.000000},
        {780,0.000000},
        {781,0.000000},
        {782,0.000000},
        {783,0.000082},
        {784,0.000164},
        {785,0.000082},
        {786,0.000000},
        {787,0.000020},
        {788,0.000040},
        {789,0.000020},
        {790,0.000000},
        {791,0.000000},
        {792,0.000000},
        {793,0.000087},
        {794,0.000174},
        {795,0.000087},
        {796,0.000000},
        {797,0.000021},
        {798,0.000042},
        {799,0.000021},
        {800,0.000000},
        {801,0.000114},
        {802,0.000227},
        {803,0.000114},
        {804,0.000000},
        {805,0.000140},
        {806,0.000280},
        {807,0.000140},
        {808,0.000000},
        {809,0.000071},
        {810,0.000142},
        {811,0.000169},
        {812,0.000195},
        {813,0.000145},
        {814,0.000095},
        {815,0.000047},
        {816,0.000000},
        {817,0.000000},
        {818,0.000000},
        {819,0.000000},
        {820,0.000000},
        {821,0.000000},
        {822,0.000000},
        {823,0.000000},
        {824,0.000000},
        {825,0.000000},
        {826,0.000000},
        {827,0.000000},
        {828,0.000000},
        {829,0.000051},
        {830,0.000101},
        {831,0.000101},
        {832,0.000101},
        {833,0.000154},
        {834,0.000207},
        {835,0.000257},
        {836,0.000306},
        {837,0.000203},
        {838,0.000100},
        {839,0.000153},
        {840,0.000206},
        {841,0.000236},
        {842,0.000266},
        {843,0.000133},
        {844,0.000000},
        {845,0.000237},
        {846,0.000475},
        {847,0.000262},
        {848,0.000050},
        {849,0.000132},
        {850,0.000215},
        {851,0.000214},
        {852,0.000213},
        {853,0.000183},
        {854,0.000153},
        {855,0.000128},
        {856,0.000103},
        {857,0.000076},
        {858,0.000050},
        {859,0.000025},
        {860,0.000000},
        {861,0.000108},
        {862,0.000216},
        {863,0.000161},
        {864,0.000105},
        {865,0.000053},
        {866,0.000000},
        {867,0.000105},
        {868,0.000210},
        {869,0.000293},
        {870,0.000377},
        {871,0.000268},
        {872,0.000159},
        {873,0.000326},
        {874,0.000493},
        {875,0.000408},
        {876,0.000323},
        {877,0.000561},
        {878,0.000798},
        {879,0.000693},
        {880,0.000587},
        {881,0.000474},
        {882,0.000361},
        {883,0.000631},
        {884,0.000900},
        {885,0.000800},
        {886,0.000699},
        {887,0.000401},
        {888,0.000102},
        {889,0.000157},
        {890,0.000212},
        {891,0.000182},
        {892,0.000152},
        {893,0.000283},
        {894,0.000414},
        {895,0.000366},
        {896,0.000319},
        {897,0.000369},
        {898,0.000419},
        {899,0.000343},
        {900,0.000267},
        {901,0.000299},
        {902,0.000330},
        {903,0.000165},
        {904,0.000000},
        {905,0.000246},
        {906,0.000492},
        {907,0.000615},
        {908,0.000738},
        {909,0.000933},
        {910,0.001128},
        {911,0.000936},
        {912,0.000744},
        {913,0.000825},
        {914,0.000906},
        {915,0.001069},
        {916,0.001233},
        {917,0.001487},
        {918,0.001741},
        {919,0.001772},
        {920,0.001803},
        {921,0.001685},
        {922,0.001567},
        {923,0.001628},
        {924,0.001689},
        {925,0.001878},
        {926,0.002067},
        {927,0.002089},
        {928,0.002112},
        {929,0.002273},
        {930,0.002434},
        {931,0.002660},
        {932,0.002886},
        {933,0.003077},
        {934,0.003269},
        {935,0.003276},
        {936,0.003283},
        {937,0.003490},
        {938,0.003697},
        {939,0.003748},
        {940,0.003799},
        {941,0.003796},
        {942,0.003793},
        {943,0.003751},
        {944,0.003710},
        {945,0.003711},
        {946,0.003712},
        {947,0.003723},
        {948,0.003734},
        {949,0.004066},
        {950,0.004397},
        {951,0.004223},
        {952,0.004049},
        {953,0.004704},
        {954,0.005358},
        {955,0.005138},
        {956,0.004919},
        {957,0.004936},
        {958,0.004954},
        {959,0.005029},
        {960,0.005104},
        {961,0.005118},
        {962,0.005133},
        {963,0.005155},
        {964,0.005178},
        {965,0.005278},
        {966,0.005378},
        {967,0.005460},
        {968,0.005542},
        {969,0.005556},
        {970,0.005570},
        {971,0.005453},
        {972,0.005336},
        {973,0.005502},
        {974,0.005668},
        {975,0.005778},
        {976,0.005887},
        {977,0.005931},
        {978,0.005975},
        {979,0.006297},
        {980,0.006619},
        {981,0.006667},
        {982,0.006715},
        {983,0.006949},
        {984,0.007184},
        {985,0.007349},
        {986,0.007514},
        {987,0.008080},
        {988,0.008647},
        {989,0.008924},
        {990,0.009201},
        {991,0.009568},
        {992,0.009934},
        {993,0.010568},
        {994,0.011202},
        {995,0.011724},
        {996,0.012246},
        {997,0.012158},
        {998,0.012070},
        {999,0.006035},
        {1000,0.000000}
      };
      
      static const double CBERS04A_MUXB6_SRF[][2] = 
      {
        {360,0.000000},
        {361,0.000000},
        {362,0.000000},
        {363,0.000000},
        {364,0.000000},
        {365,0.000000},
        {366,0.000000},
        {367,0.000000},
        {368,0.000000},
        {369,0.000000},
        {370,0.000000},
        {371,0.000000},
        {372,0.000000},
        {373,0.000000},
        {374,0.000000},
        {375,0.000000},
        {376,0.000000},
        {377,0.000000},
        {378,0.000000},
        {379,0.000000},
        {380,0.000000},
        {381,0.000000},
        {382,0.000000},
        {383,0.000000},
        {384,0.000000},
        {385,0.000000},
        {386,0.000000},
        {387,0.000000},
        {388,0.000000},
        {389,0.000000},
        {390,0.000000},
        {391,0.000000},
        {392,0.000000},
        {393,0.000000},
        {394,0.000000},
        {395,0.000000},
        {396,0.000000},
        {397,0.000000},
        {398,0.000000},
        {399,0.000000},
        {400,0.000000},
        {401,0.000000},
        {402,0.000000},
        {403,0.000209},
        {404,0.000418},
        {405,0.000209},
        {406,0.000000},
        {407,0.000195},
        {408,0.000389},
        {409,0.000267},
        {410,0.000144},
        {411,0.000573},
        {412,0.001002},
        {413,0.000671},
        {414,0.000339},
        {415,0.000169},
        {416,0.000000},
        {417,0.000250},
        {418,0.000500},
        {419,0.000250},
        {420,0.000000},
        {421,0.000584},
        {422,0.001167},
        {423,0.000584},
        {424,0.000000},
        {425,0.000293},
        {426,0.000586},
        {427,0.000342},
        {428,0.000098},
        {429,0.000170},
        {430,0.000241},
        {431,0.000526},
        {432,0.000811},
        {433,0.000449},
        {434,0.000087},
        {435,0.000612},
        {436,0.001138},
        {437,0.000867},
        {438,0.000596},
        {439,0.001023},
        {440,0.001450},
        {441,0.001677},
        {442,0.001904},
        {443,0.001879},
        {444,0.001853},
        {445,0.002122},
        {446,0.002390},
        {447,0.002641},
        {448,0.002892},
        {449,0.002632},
        {450,0.002372},
        {451,0.002403},
        {452,0.002435},
        {453,0.002235},
        {454,0.002035},
        {455,0.001556},
        {456,0.001078},
        {457,0.001159},
        {458,0.001240},
        {459,0.001411},
        {460,0.001582},
        {461,0.001377},
        {462,0.001173},
        {463,0.000929},
        {464,0.000685},
        {465,0.000895},
        {466,0.001105},
        {467,0.000552},
        {468,0.000000},
        {469,0.000476},
        {470,0.000951},
        {471,0.001019},
        {472,0.001086},
        {473,0.000801},
        {474,0.000515},
        {475,0.001008},
        {476,0.001501},
        {477,0.001414},
        {478,0.001327},
        {479,0.000762},
        {480,0.000196},
        {481,0.000445},
        {482,0.000694},
        {483,0.000970},
        {484,0.001247},
        {485,0.001341},
        {486,0.001435},
        {487,0.001762},
        {488,0.002089},
        {489,0.002505},
        {490,0.002921},
        {491,0.002901},
        {492,0.002881},
        {493,0.003862},
        {494,0.004842},
        {495,0.005480},
        {496,0.006117},
        {497,0.008580},
        {498,0.011043},
        {499,0.010900},
        {500,0.010757},
        {501,0.010204},
        {502,0.009650},
        {503,0.010261},
        {504,0.010873},
        {505,0.014424},
        {506,0.017975},
        {507,0.028313},
        {508,0.038651},
        {509,0.058785},
        {510,0.078918},
        {511,0.107101},
        {512,0.135284},
        {513,0.175280},
        {514,0.215276},
        {515,0.282255},
        {516,0.349233},
        {517,0.443724},
        {518,0.538215},
        {519,0.617065},
        {520,0.695915},
        {521,0.728827},
        {522,0.761740},
        {523,0.759479},
        {524,0.757218},
        {525,0.761411},
        {526,0.765605},
        {527,0.765936},
        {528,0.766268},
        {529,0.779257},
        {530,0.792247},
        {531,0.805359},
        {532,0.818471},
        {533,0.827680},
        {534,0.836890},
        {535,0.831630},
        {536,0.826370},
        {537,0.830763},
        {538,0.835156},
        {539,0.838075},
        {540,0.840994},
        {541,0.851884},
        {542,0.862775},
        {543,0.871613},
        {544,0.880452},
        {545,0.880241},
        {546,0.880030},
        {547,0.885484},
        {548,0.890939},
        {549,0.885211},
        {550,0.879483},
        {551,0.877130},
        {552,0.874777},
        {553,0.886654},
        {554,0.898532},
        {555,0.908330},
        {556,0.918129},
        {557,0.929224},
        {558,0.940320},
        {559,0.943460},
        {560,0.946601},
        {561,0.951758},
        {562,0.956916},
        {563,0.962154},
        {564,0.967392},
        {565,0.965531},
        {566,0.963671},
        {567,0.965432},
        {568,0.967193},
        {569,0.972763},
        {570,0.978333},
        {571,0.977361},
        {572,0.976390},
        {573,0.976102},
        {574,0.975815},
        {575,0.976892},
        {576,0.977968},
        {577,0.983664},
        {578,0.989360},
        {579,0.993406},
        {580,0.997452},
        {581,0.998726},
        {582,1.0},
        {583,0.998557},
        {584,0.997113},
        {585,0.992439},
        {586,0.987765},
        {587,0.972028},
        {588,0.956290},
        {589,0.931981},
        {590,0.907672},
        {591,0.881162},
        {592,0.854653},
        {593,0.807615},
        {594,0.760577},
        {595,0.690585},
        {596,0.620593},
        {597,0.537942},
        {598,0.455291},
        {599,0.389490},
        {600,0.323690},
        {601,0.267297},
        {602,0.210905},
        {603,0.172901},
        {604,0.134897},
        {605,0.106833},
        {606,0.078768},
        {607,0.062850},
        {608,0.046933},
        {609,0.037085},
        {610,0.027236},
        {611,0.022045},
        {612,0.016853},
        {613,0.014417},
        {614,0.011981},
        {615,0.009940},
        {616,0.007899},
        {617,0.006773},
        {618,0.005647},
        {619,0.004798},
        {620,0.003950},
        {621,0.003830},
        {622,0.003710},
        {623,0.003684},
        {624,0.003659},
        {625,0.003680},
        {626,0.003701},
        {627,0.003553},
        {628,0.003406},
        {629,0.003418},
        {630,0.003431},
        {631,0.003290},
        {632,0.003148},
        {633,0.002717},
        {634,0.002287},
        {635,0.002099},
        {636,0.001910},
        {637,0.001597},
        {638,0.001284},
        {639,0.001096},
        {640,0.000907},
        {641,0.000763},
        {642,0.000619},
        {643,0.000600},
        {644,0.000582},
        {645,0.000728},
        {646,0.000875},
        {647,0.000726},
        {648,0.000576},
        {649,0.000645},
        {650,0.000714},
        {651,0.000548},
        {652,0.000382},
        {653,0.000516},
        {654,0.000649},
        {655,0.000610},
        {656,0.000570},
        {657,0.000285},
        {658,0.000000},
        {659,0.000126},
        {660,0.000252},
        {661,0.000410},
        {662,0.000568},
        {663,0.000630},
        {664,0.000692},
        {665,0.000770},
        {666,0.000847},
        {667,0.000532},
        {668,0.000216},
        {669,0.000276},
        {670,0.000335},
        {671,0.000329},
        {672,0.000322},
        {673,0.000276},
        {674,0.000231},
        {675,0.000363},
        {676,0.000495},
        {677,0.000335},
        {678,0.000175},
        {679,0.000304},
        {680,0.000433},
        {681,0.000375},
        {682,0.000316},
        {683,0.000243},
        {684,0.000170},
        {685,0.000085},
        {686,0.000000},
        {687,0.000000},
        {688,0.000000},
        {689,0.000000},
        {690,0.000000},
        {691,0.000000},
        {692,0.000000},
        {693,0.000000},
        {694,0.000000},
        {695,0.000000},
        {696,0.000000},
        {697,0.000000},
        {698,0.000000},
        {699,0.000000},
        {700,0.000000},
        {701,0.000000},
        {702,0.000000},
        {703,0.000000},
        {704,0.000000},
        {705,0.000000},
        {706,0.000000},
        {707,0.000000},
        {708,0.000000},
        {709,0.000000},
        {710,0.000000},
        {711,0.000000},
        {712,0.000000},
        {713,0.000000},
        {714,0.000000},
        {715,0.000000},
        {716,0.000000},
        {717,0.000000},
        {718,0.000000},
        {719,0.000000},
        {720,0.000000},
        {721,0.000000},
        {722,0.000000},
        {723,0.000000},
        {724,0.000000},
        {725,0.000000},
        {726,0.000000},
        {727,0.000000},
        {728,0.000000},
        {729,0.000000},
        {730,0.000000},
        {731,0.000000},
        {732,0.000000},
        {733,0.000000},
        {734,0.000000},
        {735,0.000000},
        {736,0.000000},
        {737,0.000000},
        {738,0.000000},
        {739,0.000000},
        {740,0.000000},
        {741,0.000000},
        {742,0.000000},
        {743,0.000000},
        {744,0.000000},
        {745,0.000000},
        {746,0.000000},
        {747,0.000000},
        {748,0.000000},
        {749,0.000000},
        {750,0.000000},
        {751,0.000000},
        {752,0.000000},
        {753,0.000000},
        {754,0.000000},
        {755,0.000000},
        {756,0.000000},
        {757,0.000000},
        {758,0.000000},
        {759,0.000000},
        {760,0.000000},
        {761,0.000000},
        {762,0.000000},
        {763,0.000000},
        {764,0.000000},
        {765,0.000000},
        {766,0.000000},
        {767,0.000000},
        {768,0.000000},
        {769,0.000000},
        {770,0.000000},
        {771,0.000000},
        {772,0.000000},
        {773,0.000000},
        {774,0.000000},
        {775,0.000000},
        {776,0.000000},
        {777,0.000000},
        {778,0.000000},
        {779,0.000000},
        {780,0.000000},
        {781,0.000000},
        {782,0.000000},
        {783,0.000000},
        {784,0.000000},
        {785,0.000000},
        {786,0.000000},
        {787,0.000000},
        {788,0.000000},
        {789,0.000000},
        {790,0.000000},
        {791,0.000000},
        {792,0.000000},
        {793,0.000019},
        {794,0.000037},
        {795,0.000128},
        {796,0.000218},
        {797,0.000190},
        {798,0.000161},
        {799,0.000081},
        {800,0.000000},
        {801,0.000000},
        {802,0.000000},
        {803,0.000000},
        {804,0.000000},
        {805,0.000113},
        {806,0.000226},
        {807,0.000166},
        {808,0.000105},
        {809,0.000362},
        {810,0.000619},
        {811,0.000723},
        {812,0.000827},
        {813,0.000858},
        {814,0.000889},
        {815,0.002263},
        {816,0.003638},
        {817,0.004899},
        {818,0.006161},
        {819,0.006254},
        {820,0.006347},
        {821,0.005262},
        {822,0.004178},
        {823,0.002662},
        {824,0.001146},
        {825,0.000573},
        {826,0.000000},
        {827,0.000000},
        {828,0.000000},
        {829,0.000000},
        {830,0.000000},
        {831,0.000000},
        {832,0.000000},
        {833,0.000000},
        {834,0.000000},
        {835,0.000000},
        {836,0.000000},
        {837,0.000000},
        {838,0.000000},
        {839,0.000000},
        {840,0.000000},
        {841,0.000000},
        {842,0.000000},
        {843,0.000059},
        {844,0.000117},
        {845,0.000059},
        {846,0.000000},
        {847,0.000000},
        {848,0.000000},
        {849,0.000000},
        {850,0.000000},
        {851,0.000000},
        {852,0.000000},
        {853,0.000125},
        {854,0.000251},
        {855,0.000218},
        {856,0.000186},
        {857,0.000093},
        {858,0.000000},
        {859,0.000266},
        {860,0.000532},
        {861,0.000434},
        {862,0.000336},
        {863,0.000914},
        {864,0.001491},
        {865,0.001188},
        {866,0.000885},
        {867,0.000747},
        {868,0.000609},
        {869,0.000363},
        {870,0.000117},
        {871,0.000059},
        {872,0.000000},
        {873,0.000000},
        {874,0.000000},
        {875,0.000000},
        {876,0.000000},
        {877,0.000000},
        {878,0.000000},
        {879,0.000000},
        {880,0.000000},
        {881,0.000056},
        {882,0.000112},
        {883,0.000056},
        {884,0.000000},
        {885,0.000000},
        {886,0.000000},
        {887,0.000000},
        {888,0.000000},
        {889,0.000000},
        {890,0.000000},
        {891,0.000000},
        {892,0.000000},
        {893,0.000000},
        {894,0.000000},
        {895,0.000000},
        {896,0.000000},
        {897,0.000000},
        {898,0.000000},
        {899,0.000000},
        {900,0.000000},
        {901,0.000000},
        {902,0.000000},
        {903,0.000000},
        {904,0.000000},
        {905,0.000000},
        {906,0.000000},
        {907,0.000000},
        {908,0.000000},
        {909,0.000000},
        {910,0.000000},
        {911,0.000000},
        {912,0.000000},
        {913,0.000000},
        {914,0.000000},
        {915,0.000000},
        {916,0.000000},
        {917,0.000000},
        {918,0.000000},
        {919,0.000000},
        {920,0.000000},
        {921,0.000000},
        {922,0.000000},
        {923,0.000000},
        {924,0.000000},
        {925,0.000000},
        {926,0.000000},
        {927,0.000000},
        {928,0.000000},
        {929,0.000000},
        {930,0.000000},
        {931,0.000000},
        {932,0.000000},
        {933,0.000000},
        {934,0.000000},
        {935,0.000000},
        {936,0.000000},
        {937,0.000023},
        {938,0.000045},
        {939,0.000023},
        {940,0.000000},
        {941,0.000000},
        {942,0.000000},
        {943,0.000000},
        {944,0.000000},
        {945,0.000000},
        {946,0.000000},
        {947,0.000000},
        {948,0.000000},
        {949,0.000000},
        {950,0.000000},
        {951,0.000000},
        {952,0.000000},
        {953,0.000000},
        {954,0.000000},
        {955,0.000000},
        {956,0.000000},
        {957,0.000000},
        {958,0.000000},
        {959,0.000000},
        {960,0.000000},
        {961,0.000000},
        {962,0.000000},
        {963,0.000000},
        {964,0.000000},
        {965,0.000000},
        {966,0.000000},
        {967,0.000000},
        {968,0.000000},
        {969,0.000000},
        {970,0.000000},
        {971,0.000000},
        {972,0.000000},
        {973,0.000000},
        {974,0.000000},
        {975,0.000000},
        {976,0.000000},
        {977,0.000000},
        {978,0.000000},
        {979,0.000000},
        {980,0.000000},
        {981,0.000000},
        {982,0.000000},
        {983,0.000000},
        {984,0.000000},
        {985,0.000000},
        {986,0.000000},
        {987,0.000000},
        {988,0.000000},
        {989,0.000000},
        {990,0.000000},
        {991,0.000000},
        {992,0.000000},
        {993,0.000000},
        {994,0.000000},
        {995,0.000000},
        {996,0.000000},
        {997,0.000000},
        {998,0.000000},
        {999,0.000000},
        {1000,0.000000}
      };
      
      static const double CBERS04A_MUXB7_SRF[][2] = 
      {
        {360,0.000000},
        {361,0.000000},
        {362,0.000000},
        {363,0.000000},
        {364,0.000000},
        {365,0.000000},
        {366,0.000000},
        {367,0.000000},
        {368,0.000000},
        {369,0.000000},
        {370,0.000000},
        {371,0.000000},
        {372,0.000000},
        {373,0.000000},
        {374,0.000000},
        {375,0.000000},
        {376,0.000000},
        {377,0.000000},
        {378,0.000000},
        {379,0.000000},
        {380,0.000000},
        {381,0.000000},
        {382,0.000000},
        {383,0.000000},
        {384,0.000000},
        {385,0.000000},
        {386,0.000000},
        {387,0.000000},
        {388,0.000000},
        {389,0.000000},
        {390,0.000000},
        {391,0.000000},
        {392,0.000000},
        {393,0.000000},
        {394,0.000000},
        {395,0.000000},
        {396,0.000000},
        {397,0.000000},
        {398,0.000000},
        {399,0.000222},
        {400,0.000445},
        {401,0.000222},
        {402,0.000000},
        {403,0.000320},
        {404,0.000639},
        {405,0.000320},
        {406,0.000000},
        {407,0.000000},
        {408,0.000000},
        {409,0.000000},
        {410,0.000000},
        {411,0.000380},
        {412,0.000761},
        {413,0.000380},
        {414,0.000000},
        {415,0.000146},
        {416,0.000291},
        {417,0.000146},
        {418,0.000000},
        {419,0.000042},
        {420,0.000084},
        {421,0.000262},
        {422,0.000440},
        {423,0.000220},
        {424,0.000000},
        {425,0.001105},
        {426,0.002210},
        {427,0.001105},
        {428,0.000000},
        {429,0.000000},
        {430,0.000000},
        {431,0.000033},
        {432,0.000067},
        {433,0.000033},
        {434,0.000000},
        {435,0.000165},
        {436,0.000329},
        {437,0.000165},
        {438,0.000000},
        {439,0.000219},
        {440,0.000439},
        {441,0.000372},
        {442,0.000305},
        {443,0.000240},
        {444,0.000176},
        {445,0.000292},
        {446,0.000408},
        {447,0.000204},
        {448,0.000000},
        {449,0.000360},
        {450,0.000721},
        {451,0.000495},
        {452,0.000269},
        {453,0.000135},
        {454,0.000000},
        {455,0.000282},
        {456,0.000563},
        {457,0.000282},
        {458,0.000000},
        {459,0.000000},
        {460,0.000000},
        {461,0.000000},
        {462,0.000000},
        {463,0.000000},
        {464,0.000000},
        {465,0.000000},
        {466,0.000000},
        {467,0.000000},
        {468,0.000000},
        {469,0.001060},
        {470,0.002120},
        {471,0.002134},
        {472,0.002148},
        {473,0.001074},
        {474,0.000000},
        {475,0.000000},
        {476,0.000000},
        {477,0.000247},
        {478,0.000495},
        {479,0.000378},
        {480,0.000261},
        {481,0.000131},
        {482,0.000000},
        {483,0.000124},
        {484,0.000249},
        {485,0.000211},
        {486,0.000174},
        {487,0.000311},
        {488,0.000448},
        {489,0.000342},
        {490,0.000236},
        {491,0.000268},
        {492,0.000300},
        {493,0.000462},
        {494,0.000625},
        {495,0.000425},
        {496,0.000225},
        {497,0.000380},
        {498,0.000536},
        {499,0.000502},
        {500,0.000468},
        {501,0.000371},
        {502,0.000274},
        {503,0.000453},
        {504,0.000631},
        {505,0.000506},
        {506,0.000382},
        {507,0.000585},
        {508,0.000788},
        {509,0.000666},
        {510,0.000544},
        {511,0.000313},
        {512,0.000082},
        {513,0.000307},
        {514,0.000532},
        {515,0.000306},
        {516,0.000081},
        {517,0.000323},
        {518,0.000566},
        {519,0.000295},
        {520,0.000025},
        {521,0.000135},
        {522,0.000245},
        {523,0.000296},
        {524,0.000348},
        {525,0.000425},
        {526,0.000503},
        {527,0.000678},
        {528,0.000853},
        {529,0.000822},
        {530,0.000791},
        {531,0.000664},
        {532,0.000537},
        {533,0.000611},
        {534,0.000686},
        {535,0.000954},
        {536,0.001222},
        {537,0.001048},
        {538,0.000874},
        {539,0.000768},
        {540,0.000661},
        {541,0.000811},
        {542,0.000961},
        {543,0.000908},
        {544,0.000855},
        {545,0.000655},
        {546,0.000455},
        {547,0.000311},
        {548,0.000167},
        {549,0.000542},
        {550,0.000916},
        {551,0.000796},
        {552,0.000676},
        {553,0.000514},
        {554,0.000353},
        {555,0.000445},
        {556,0.000537},
        {557,0.000649},
        {558,0.000761},
        {559,0.000737},
        {560,0.000712},
        {561,0.000781},
        {562,0.000850},
        {563,0.000738},
        {564,0.000627},
        {565,0.000553},
        {566,0.000478},
        {567,0.000409},
        {568,0.000340},
        {569,0.000524},
        {570,0.000708},
        {571,0.000843},
        {572,0.000979},
        {573,0.000909},
        {574,0.000840},
        {575,0.000812},
        {576,0.000785},
        {577,0.000740},
        {578,0.000695},
        {579,0.000672},
        {580,0.000649},
        {581,0.000849},
        {582,0.001050},
        {583,0.000984},
        {584,0.000917},
        {585,0.000986},
        {586,0.001054},
        {587,0.000852},
        {588,0.000649},
        {589,0.000689},
        {590,0.000729},
        {591,0.000731},
        {592,0.000734},
        {593,0.000760},
        {594,0.000786},
        {595,0.000983},
        {596,0.001181},
        {597,0.001199},
        {598,0.001217},
        {599,0.001443},
        {600,0.001669},
        {601,0.002011},
        {602,0.002352},
        {603,0.002776},
        {604,0.003199},
        {605,0.003541},
        {606,0.003882},
        {607,0.005590},
        {608,0.007298},
        {609,0.008868},
        {610,0.010439},
        {611,0.012869},
        {612,0.015298},
        {613,0.019036},
        {614,0.022773},
        {615,0.029037},
        {616,0.035301},
        {617,0.048376},
        {618,0.061450},
        {619,0.085476},
        {620,0.109502},
        {621,0.157513},
        {622,0.205525},
        {623,0.282238},
        {624,0.358951},
        {625,0.454161},
        {626,0.549372},
        {627,0.624941},
        {628,0.700511},
        {629,0.751429},
        {630,0.802346},
        {631,0.837077},
        {632,0.871807},
        {633,0.890203},
        {634,0.908599},
        {635,0.918331},
        {636,0.928063},
        {637,0.928965},
        {638,0.929867},
        {639,0.934336},
        {640,0.938805},
        {641,0.939611},
        {642,0.940417},
        {643,0.949866},
        {644,0.959316},
        {645,0.955100},
        {646,0.950884},
        {647,0.949815},
        {648,0.948746},
        {649,0.941605},
        {650,0.934463},
        {651,0.948301},
        {652,0.962139},
        {653,0.972099},
        {654,0.982059},
        {655,0.991029},
        {656,1.0},
        {657,0.988619},
        {658,0.977237},
        {659,0.973996},
        {660,0.970755},
        {661,0.968356},
        {662,0.965958},
        {663,0.955100},
        {664,0.944241},
        {665,0.935509},
        {666,0.926778},
        {667,0.927118},
        {668,0.927459},
        {669,0.918239},
        {670,0.909019},
        {671,0.908467},
        {672,0.907914},
        {673,0.907296},
        {674,0.906678},
        {675,0.900698},
        {676,0.894718},
        {677,0.881502},
        {678,0.868286},
        {679,0.859020},
        {680,0.849755},
        {681,0.811625},
        {682,0.773495},
        {683,0.694240},
        {684,0.614984},
        {685,0.498664},
        {686,0.382344},
        {687,0.297868},
        {688,0.213392},
        {689,0.157202},
        {690,0.101012},
        {691,0.077027},
        {692,0.053043},
        {693,0.041347},
        {694,0.029652},
        {695,0.023894},
        {696,0.018137},
        {697,0.015042},
        {698,0.011948},
        {699,0.010298},
        {700,0.008648},
        {701,0.007603},
        {702,0.006558},
        {703,0.006376},
        {704,0.006193},
        {705,0.005469},
        {706,0.004745},
        {707,0.004184},
        {708,0.003622},
        {709,0.003565},
        {710,0.003508},
        {711,0.003398},
        {712,0.003289},
        {713,0.003129},
        {714,0.002969},
        {715,0.002914},
        {716,0.002858},
        {717,0.002438},
        {718,0.002017},
        {719,0.001589},
        {720,0.001160},
        {721,0.001230},
        {722,0.001300},
        {723,0.001207},
        {724,0.001114},
        {725,0.000926},
        {726,0.000737},
        {727,0.000833},
        {728,0.000929},
        {729,0.000819},
        {730,0.000709},
        {731,0.000759},
        {732,0.000810},
        {733,0.000810},
        {734,0.000811},
        {735,0.000835},
        {736,0.000860},
        {737,0.000922},
        {738,0.000985},
        {739,0.001021},
        {740,0.001056},
        {741,0.001163},
        {742,0.001270},
        {743,0.001196},
        {744,0.001123},
        {745,0.001090},
        {746,0.001057},
        {747,0.001137},
        {748,0.001217},
        {749,0.001141},
        {750,0.001064},
        {751,0.001231},
        {752,0.001398},
        {753,0.001169},
        {754,0.000941},
        {755,0.001224},
        {756,0.001507},
        {757,0.001102},
        {758,0.000697},
        {759,0.000908},
        {760,0.001118},
        {761,0.001046},
        {762,0.000973},
        {763,0.000783},
        {764,0.000592},
        {765,0.000413},
        {766,0.000233},
        {767,0.000616},
        {768,0.000999},
        {769,0.001172},
        {770,0.001345},
        {771,0.001323},
        {772,0.001302},
        {773,0.001324},
        {774,0.001346},
        {775,0.001235},
        {776,0.001124},
        {777,0.001178},
        {778,0.001233},
        {779,0.001055},
        {780,0.000878},
        {781,0.001235},
        {782,0.001592},
        {783,0.001512},
        {784,0.001432},
        {785,0.001537},
        {786,0.001641},
        {787,0.001446},
        {788,0.001252},
        {789,0.001105},
        {790,0.000959},
        {791,0.001041},
        {792,0.001122},
        {793,0.001257},
        {794,0.001392},
        {795,0.001199},
        {796,0.001006},
        {797,0.001140},
        {798,0.001274},
        {799,0.001402},
        {800,0.001529},
        {801,0.001460},
        {802,0.001391},
        {803,0.001443},
        {804,0.001495},
        {805,0.001428},
        {806,0.001361},
        {807,0.001640},
        {808,0.001918},
        {809,0.001792},
        {810,0.001667},
        {811,0.001649},
        {812,0.001631},
        {813,0.001429},
        {814,0.001226},
        {815,0.000928},
        {816,0.000630},
        {817,0.000570},
        {818,0.000511},
        {819,0.000444},
        {820,0.000377},
        {821,0.000929},
        {822,0.001482},
        {823,0.001459},
        {824,0.001435},
        {825,0.001143},
        {826,0.000851},
        {827,0.000954},
        {828,0.001058},
        {829,0.001149},
        {830,0.001240},
        {831,0.000813},
        {832,0.000385},
        {833,0.000563},
        {834,0.000741},
        {835,0.000906},
        {836,0.001070},
        {837,0.000831},
        {838,0.000592},
        {839,0.000488},
        {840,0.000385},
        {841,0.000426},
        {842,0.000467},
        {843,0.000234},
        {844,0.000000},
        {845,0.000764},
        {846,0.001528},
        {847,0.001219},
        {848,0.000909},
        {849,0.000948},
        {850,0.000988},
        {851,0.001165},
        {852,0.001342},
        {853,0.000757},
        {854,0.000172},
        {855,0.000569},
        {856,0.000966},
        {857,0.000902},
        {858,0.000837},
        {859,0.000418},
        {860,0.000000},
        {861,0.000793},
        {862,0.001586},
        {863,0.001178},
        {864,0.000770},
        {865,0.000615},
        {866,0.000461},
        {867,0.000570},
        {868,0.000680},
        {869,0.000466},
        {870,0.000252},
        {871,0.000434},
        {872,0.000616},
        {873,0.000436},
        {874,0.000256},
        {875,0.000473},
        {876,0.000691},
        {877,0.000899},
        {878,0.001106},
        {879,0.001038},
        {880,0.000969},
        {881,0.000954},
        {882,0.000940},
        {883,0.000771},
        {884,0.000602},
        {885,0.001078},
        {886,0.001554},
        {887,0.001260},
        {888,0.000965},
        {889,0.001007},
        {890,0.001049},
        {891,0.000751},
        {892,0.000452},
        {893,0.000942},
        {894,0.001432},
        {895,0.001489},
        {896,0.001547},
        {897,0.001356},
        {898,0.001165},
        {899,0.001582},
        {900,0.001999},
        {901,0.001913},
        {902,0.001827},
        {903,0.001480},
        {904,0.001133},
        {905,0.001432},
        {906,0.001731},
        {907,0.001272},
        {908,0.000813},
        {909,0.001042},
        {910,0.001271},
        {911,0.001188},
        {912,0.001105},
        {913,0.001142},
        {914,0.001179},
        {915,0.000678},
        {916,0.000177},
        {917,0.000543},
        {918,0.000909},
        {919,0.000911},
        {920,0.000913},
        {921,0.001125},
        {922,0.001338},
        {923,0.001307},
        {924,0.001275},
        {925,0.001404},
        {926,0.001532},
        {927,0.001218},
        {928,0.000904},
        {929,0.001173},
        {930,0.001442},
        {931,0.001307},
        {932,0.001172},
        {933,0.001640},
        {934,0.002109},
        {935,0.001465},
        {936,0.000821},
        {937,0.001337},
        {938,0.001853},
        {939,0.001729},
        {940,0.001605},
        {941,0.001342},
        {942,0.001080},
        {943,0.000776},
        {944,0.000472},
        {945,0.000438},
        {946,0.000405},
        {947,0.000550},
        {948,0.000695},
        {949,0.000769},
        {950,0.000843},
        {951,0.000914},
        {952,0.000986},
        {953,0.001044},
        {954,0.001103},
        {955,0.001157},
        {956,0.001211},
        {957,0.001170},
        {958,0.001128},
        {959,0.000853},
        {960,0.000579},
        {961,0.000289},
        {962,0.000000},
        {963,0.000602},
        {964,0.001204},
        {965,0.001200},
        {966,0.001196},
        {967,0.001495},
        {968,0.001794},
        {969,0.001541},
        {970,0.001288},
        {971,0.000817},
        {972,0.000346},
        {973,0.000582},
        {974,0.000818},
        {975,0.000503},
        {976,0.000188},
        {977,0.000265},
        {978,0.000343},
        {979,0.000535},
        {980,0.000728},
        {981,0.000842},
        {982,0.000957},
        {983,0.000692},
        {984,0.000426},
        {985,0.000431},
        {986,0.000435},
        {987,0.000278},
        {988,0.000121},
        {989,0.000284},
        {990,0.000448},
        {991,0.000362},
        {992,0.000277},
        {993,0.000423},
        {994,0.000569},
        {995,0.000615},
        {996,0.000660},
        {997,0.000553},
        {998,0.000445},
        {999,0.000222},
        {1000,0.000000}
      };
      
      static const double CBERS04A_MUXB8_SRF[][2] = 
      {
        {360,0.000000},
        {361,0.000000},
        {362,0.000000},
        {363,0.000000},
        {364,0.000000},
        {365,0.000000},
        {366,0.000000},
        {367,0.000000},
        {368,0.000000},
        {369,0.000000},
        {370,0.000000},
        {371,0.000000},
        {372,0.000000},
        {373,0.000000},
        {374,0.000000},
        {375,0.000000},
        {376,0.000000},
        {377,0.000000},
        {378,0.000000},
        {379,0.000000},
        {380,0.000000},
        {381,0.000000},
        {382,0.000000},
        {383,0.000000},
        {384,0.000000},
        {385,0.000000},
        {386,0.000000},
        {387,0.000000},
        {388,0.000000},
        {389,0.000000},
        {390,0.000000},
        {391,0.000000},
        {392,0.000000},
        {393,0.000000},
        {394,0.000000},
        {395,0.000000},
        {396,0.000000},
        {397,0.000000},
        {398,0.000000},
        {399,0.000000},
        {400,0.000000},
        {401,0.000000},
        {402,0.000000},
        {403,0.000090},
        {404,0.000181},
        {405,0.000090},
        {406,0.000000},
        {407,0.000000},
        {408,0.000000},
        {409,0.000000},
        {410,0.000000},
        {411,0.000000},
        {412,0.000000},
        {413,0.000000},
        {414,0.000000},
        {415,0.000000},
        {416,0.000000},
        {417,0.000000},
        {418,0.000000},
        {419,0.000064},
        {420,0.000129},
        {421,0.000064},
        {422,0.000000},
        {423,0.000126},
        {424,0.000253},
        {425,0.000126},
        {426,0.000000},
        {427,0.000000},
        {428,0.000000},
        {429,0.000000},
        {430,0.000000},
        {431,0.000000},
        {432,0.000000},
        {433,0.000103},
        {434,0.000205},
        {435,0.000103},
        {436,0.000000},
        {437,0.000000},
        {438,0.000000},
        {439,0.000000},
        {440,0.000000},
        {441,0.000000},
        {442,0.000000},
        {443,0.000000},
        {444,0.000000},
        {445,0.000000},
        {446,0.000000},
        {447,0.000000},
        {448,0.000000},
        {449,0.000000},
        {450,0.000000},
        {451,0.000000},
        {452,0.000000},
        {453,0.000000},
        {454,0.000000},
        {455,0.000000},
        {456,0.000000},
        {457,0.000000},
        {458,0.000000},
        {459,0.000000},
        {460,0.000000},
        {461,0.000033},
        {462,0.000066},
        {463,0.000033},
        {464,0.000000},
        {465,0.000000},
        {466,0.000000},
        {467,0.000063},
        {468,0.000127},
        {469,0.000063},
        {470,0.000000},
        {471,0.000000},
        {472,0.000000},
        {473,0.000027},
        {474,0.000055},
        {475,0.000027},
        {476,0.000000},
        {477,0.000000},
        {478,0.000000},
        {479,0.000056},
        {480,0.000112},
        {481,0.000081},
        {482,0.000051},
        {483,0.000025},
        {484,0.000000},
        {485,0.000052},
        {486,0.000105},
        {487,0.000185},
        {488,0.000266},
        {489,0.000556},
        {490,0.000847},
        {491,0.000447},
        {492,0.000047},
        {493,0.000046},
        {494,0.000046},
        {495,0.000046},
        {496,0.000045},
        {497,0.000023},
        {498,0.000000},
        {499,0.000046},
        {500,0.000093},
        {501,0.000068},
        {502,0.000043},
        {503,0.000021},
        {504,0.000000},
        {505,0.000000},
        {506,0.000000},
        {507,0.000020},
        {508,0.000041},
        {509,0.000020},
        {510,0.000000},
        {511,0.000000},
        {512,0.000000},
        {513,0.000000},
        {514,0.000000},
        {515,0.000000},
        {516,0.000000},
        {517,0.000000},
        {518,0.000000},
        {519,0.000000},
        {520,0.000000},
        {521,0.000000},
        {522,0.000000},
        {523,0.000000},
        {524,0.000000},
        {525,0.000018},
        {526,0.000037},
        {527,0.000018},
        {528,0.000000},
        {529,0.000000},
        {530,0.000000},
        {531,0.000000},
        {532,0.000000},
        {533,0.000000},
        {534,0.000000},
        {535,0.000000},
        {536,0.000000},
        {537,0.000000},
        {538,0.000000},
        {539,0.000000},
        {540,0.000000},
        {541,0.000017},
        {542,0.000034},
        {543,0.000017},
        {544,0.000000},
        {545,0.000017},
        {546,0.000033},
        {547,0.000017},
        {548,0.000000},
        {549,0.000016},
        {550,0.000033},
        {551,0.000016},
        {552,0.000000},
        {553,0.000000},
        {554,0.000000},
        {555,0.000000},
        {556,0.000000},
        {557,0.000000},
        {558,0.000000},
        {559,0.000000},
        {560,0.000000},
        {561,0.000000},
        {562,0.000000},
        {563,0.000016},
        {564,0.000032},
        {565,0.000032},
        {566,0.000032},
        {567,0.000016},
        {568,0.000000},
        {569,0.000000},
        {570,0.000000},
        {571,0.000000},
        {572,0.000000},
        {573,0.000000},
        {574,0.000000},
        {575,0.000000},
        {576,0.000000},
        {577,0.000000},
        {578,0.000000},
        {579,0.000000},
        {580,0.000000},
        {581,0.000016},
        {582,0.000031},
        {583,0.000016},
        {584,0.000000},
        {585,0.000000},
        {586,0.000000},
        {587,0.000000},
        {588,0.000000},
        {589,0.000015},
        {590,0.000031},
        {591,0.000031},
        {592,0.000031},
        {593,0.000015},
        {594,0.000000},
        {595,0.000000},
        {596,0.000000},
        {597,0.000033},
        {598,0.000065},
        {599,0.000033},
        {600,0.000000},
        {601,0.000000},
        {602,0.000000},
        {603,0.000015},
        {604,0.000031},
        {605,0.000015},
        {606,0.000000},
        {607,0.000000},
        {608,0.000000},
        {609,0.000016},
        {610,0.000031},
        {611,0.000031},
        {612,0.000031},
        {613,0.000015},
        {614,0.000000},
        {615,0.000000},
        {616,0.000000},
        {617,0.000000},
        {618,0.000000},
        {619,0.000015},
        {620,0.000031},
        {621,0.000048},
        {622,0.000066},
        {623,0.000033},
        {624,0.000000},
        {625,0.000016},
        {626,0.000031},
        {627,0.000032},
        {628,0.000032},
        {629,0.000069},
        {630,0.000106},
        {631,0.000145},
        {632,0.000184},
        {633,0.000166},
        {634,0.000149},
        {635,0.000149},
        {636,0.000149},
        {637,0.000149},
        {638,0.000149},
        {639,0.000150},
        {640,0.000151},
        {641,0.000171},
        {642,0.000191},
        {643,0.000153},
        {644,0.000115},
        {645,0.000115},
        {646,0.000114},
        {647,0.000114},
        {648,0.000114},
        {649,0.000094},
        {650,0.000073},
        {651,0.000134},
        {652,0.000195},
        {653,0.000198},
        {654,0.000200},
        {655,0.000162},
        {656,0.000123},
        {657,0.000276},
        {658,0.000429},
        {659,0.000364},
        {660,0.000299},
        {661,0.000211},
        {662,0.000123},
        {663,0.000282},
        {664,0.000440},
        {665,0.000371},
        {666,0.000302},
        {667,0.000204},
        {668,0.000107},
        {669,0.000139},
        {670,0.000171},
        {671,0.000201},
        {672,0.000231},
        {673,0.000261},
        {674,0.000290},
        {675,0.000221},
        {676,0.000153},
        {677,0.000165},
        {678,0.000177},
        {679,0.000235},
        {680,0.000293},
        {681,0.000320},
        {682,0.000347},
        {683,0.000245},
        {684,0.000143},
        {685,0.000112},
        {686,0.000082},
        {687,0.000054},
        {688,0.000026},
        {689,0.000013},
        {690,0.000000},
        {691,0.000000},
        {692,0.000000},
        {693,0.000013},
        {694,0.000025},
        {695,0.000026},
        {696,0.000026},
        {697,0.000013},
        {698,0.000000},
        {699,0.000000},
        {700,0.000000},
        {701,0.000014},
        {702,0.000027},
        {703,0.000027},
        {704,0.000027},
        {705,0.000058},
        {706,0.000088},
        {707,0.000072},
        {708,0.000057},
        {709,0.000057},
        {710,0.000058},
        {711,0.000043},
        {712,0.000027},
        {713,0.000058},
        {714,0.000089},
        {715,0.000074},
        {716,0.000059},
        {717,0.000044},
        {718,0.000028},
        {719,0.000076},
        {720,0.000124},
        {721,0.000092},
        {722,0.000061},
        {723,0.000062},
        {724,0.000062},
        {725,0.000095},
        {726,0.000128},
        {727,0.000147},
        {728,0.000166},
        {729,0.000300},
        {730,0.000434},
        {731,0.000832},
        {732,0.001231},
        {733,0.001661},
        {734,0.002091},
        {735,0.002649},
        {736,0.003208},
        {737,0.003317},
        {738,0.003425},
        {739,0.003841},
        {740,0.004257},
        {741,0.005741},
        {742,0.007225},
        {743,0.008940},
        {744,0.010654},
        {745,0.012552},
        {746,0.014451},
        {747,0.016514},
        {748,0.018577},
        {749,0.020399},
        {750,0.022222},
        {751,0.024722},
        {752,0.027222},
        {753,0.030597},
        {754,0.033971},
        {755,0.038271},
        {756,0.042570},
        {757,0.047825},
        {758,0.053081},
        {759,0.062215},
        {760,0.071349},
        {761,0.085534},
        {762,0.099720},
        {763,0.123384},
        {764,0.147048},
        {765,0.183884},
        {766,0.220721},
        {767,0.282534},
        {768,0.344347},
        {769,0.416241},
        {770,0.488134},
        {771,0.582517},
        {772,0.676900},
        {773,0.772149},
        {774,0.867399},
        {775,0.904229},
        {776,0.941058},
        {777,0.970529},
        {778,1.0},
        {779,0.968928},
        {780,0.937855},
        {781,0.925289},
        {782,0.912723},
        {783,0.875339},
        {784,0.837955},
        {785,0.836936},
        {786,0.835917},
        {787,0.828990},
        {788,0.822063},
        {789,0.814755},
        {790,0.807447},
        {791,0.799989},
        {792,0.792531},
        {793,0.797652},
        {794,0.802773},
        {795,0.813029},
        {796,0.823285},
        {797,0.826962},
        {798,0.830639},
        {799,0.828741},
        {800,0.826842},
        {801,0.818331},
        {802,0.809821},
        {803,0.812974},
        {804,0.816128},
        {805,0.810388},
        {806,0.804647},
        {807,0.811876},
        {808,0.819104},
        {809,0.813334},
        {810,0.807563},
        {811,0.805787},
        {812,0.804011},
        {813,0.796793},
        {814,0.789574},
        {815,0.784041},
        {816,0.778508},
        {817,0.781810},
        {818,0.785112},
        {819,0.782458},
        {820,0.779803},
        {821,0.770303},
        {822,0.760804},
        {823,0.761772},
        {824,0.762740},
        {825,0.746772},
        {826,0.730804},
        {827,0.730303},
        {828,0.729802},
        {829,0.736810},
        {830,0.743817},
        {831,0.738771},
        {832,0.733725},
        {833,0.729064},
        {834,0.724403},
        {835,0.711984},
        {836,0.699564},
        {837,0.700354},
        {838,0.701145},
        {839,0.701163},
        {840,0.701182},
        {841,0.705485},
        {842,0.709787},
        {843,0.711417},
        {844,0.713047},
        {845,0.698679},
        {846,0.684310},
        {847,0.690605},
        {848,0.696899},
        {849,0.695887},
        {850,0.694874},
        {851,0.689262},
        {852,0.683651},
        {853,0.669994},
        {854,0.656338},
        {855,0.660359},
        {856,0.664380},
        {857,0.667846},
        {858,0.671312},
        {859,0.655698},
        {860,0.640085},
        {861,0.646025},
        {862,0.651965},
        {863,0.645562},
        {864,0.639159},
        {865,0.624486},
        {866,0.609814},
        {867,0.602325},
        {868,0.594837},
        {869,0.592166},
        {870,0.589496},
        {871,0.581508},
        {872,0.573521},
        {873,0.574610},
        {874,0.575700},
        {875,0.567022},
        {876,0.558344},
        {877,0.546279},
        {878,0.534213},
        {879,0.526024},
        {880,0.517834},
        {881,0.492070},
        {882,0.466307},
        {883,0.447254},
        {884,0.428202},
        {885,0.392737},
        {886,0.357272},
        {887,0.323391},
        {888,0.289510},
        {889,0.247737},
        {890,0.205963},
        {891,0.177736},
        {892,0.149509},
        {893,0.126356},
        {894,0.103203},
        {895,0.088285},
        {896,0.073367},
        {897,0.064268},
        {898,0.055169},
        {899,0.045996},
        {900,0.036822},
        {901,0.032975},
        {902,0.029129},
        {903,0.025682},
        {904,0.022235},
        {905,0.018526},
        {906,0.014817},
        {907,0.013007},
        {908,0.011198},
        {909,0.009714},
        {910,0.008230},
        {911,0.006963},
        {912,0.005696},
        {913,0.005645},
        {914,0.005593},
        {915,0.005353},
        {916,0.005112},
        {917,0.004550},
        {918,0.003987},
        {919,0.003337},
        {920,0.002686},
        {921,0.001680},
        {922,0.000674},
        {923,0.000734},
        {924,0.000793},
        {925,0.000657},
        {926,0.000520},
        {927,0.000399},
        {928,0.000277},
        {929,0.000280},
        {930,0.000283},
        {931,0.000227},
        {932,0.000171},
        {933,0.000142},
        {934,0.000113},
        {935,0.000115},
        {936,0.000116},
        {937,0.000112},
        {938,0.000107},
        {939,0.000080},
        {940,0.000052},
        {941,0.000052},
        {942,0.000052},
        {943,0.000026},
        {944,0.000000},
        {945,0.000000},
        {946,0.000000},
        {947,0.000000},
        {948,0.000000},
        {949,0.000000},
        {950,0.000000},
        {951,0.000025},
        {952,0.000051},
        {953,0.000052},
        {954,0.000053},
        {955,0.000054},
        {956,0.000054},
        {957,0.000027},
        {958,0.000000},
        {959,0.000000},
        {960,0.000000},
        {961,0.000000},
        {962,0.000000},
        {963,0.000000},
        {964,0.000000},
        {965,0.000000},
        {966,0.000000},
        {967,0.000027},
        {968,0.000053},
        {969,0.000027},
        {970,0.000000},
        {971,0.000000},
        {972,0.000000},
        {973,0.000000},
        {974,0.000000},
        {975,0.000000},
        {976,0.000000},
        {977,0.000000},
        {978,0.000000},
        {979,0.000000},
        {980,0.000000},
        {981,0.000000},
        {982,0.000000},
        {983,0.000000},
        {984,0.000000},
        {985,0.000000},
        {986,0.000000},
        {987,0.000000},
        {988,0.000000},
        {989,0.000000},
        {990,0.000000},
        {991,0.000000},
        {992,0.000000},
        {993,0.000000},
        {994,0.000000},
        {995,0.000000},
        {996,0.000000},
        {997,0.000000},
        {998,0.000000},
        {999,0.000000},
        {1000,0.000000}
      };
      
      /*!
        \brief Returns a reference to the global static Spectral Response Functions.
        \brief Returns a reference to the global static Spectral Response Functions.
        \note The SRFs container is static and the dada follows the format: sensor type, sensor type string and SRF[ wavelength , response ].
        \note Reference unit: nm (nanometer)
       */            
      TERPEXPORT const std::map< SensorType, std::pair< std::string, std::map< double, double > > >& getSRFs();
      
       /*!
-        \brief Returns a Spectral Response Function from the given sensor.
         \param sensor The Sensor.
-        \param container The container where the SRF will be stored.
-        \note std::pair< double, double >( wavelength , response }
         \note Reference unit: nm (nanometer)
        */            
      TERPEXPORT void getSRF( const SensorType& sensor, std::map< double, double >& container );
      
      /*!
        \brief Return a SRF interpolated from the given SRFs.
        \param SRFs Input SRFs.
        \param frequency The desired interpolated frequency.
        \return Return a SRF interpolated from the given SRFs.
       */         
      TERPEXPORT double interpolateSRF( const std::map< double, double >& sRFs, const double& frequency );
      
      /*!
        \brief Return the SRF area.
        \param SRFs Input SRFs.
        \return Return the SRF area.
       */         
      TERPEXPORT double getSRFArea( const std::map< double, double >& sRFs );        
      
      /*!
        \brief Return the intersetction SRF.
        \param sRF1 Input SRFs 1.
        \param sRF2 Input SRFs 2.
        \param intersectionSRF The resultant intersection SRF.
        \return Return a SRF interpolated from the given SRFs.
       */         
      TERPEXPORT void getIntersectionSRF( const std::map< double, double >& sRF1, 
        const std::map< double, double >& sRF2,
        std::map< double, double >& intersectionSRF );        
        
      /*!
        \brief Return the union SRF.
        \param sRF1 Input SRFs 1.
        \param sRF2 Input SRFs 2.
        \param unionSRF The resultant intersection SRF.
        \return Return a SRF interpolated from the given SRFs.
       */         
      TERPEXPORT void getUnionSRF( const std::map< double, double >& sRF1, 
        const std::map< double, double >& sRF2,
        std::map< double, double >& unionSRF );        
    } // end namespace spf
  } // end namespace rp
}   // end namespace te

#endif

